/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*-
 * Copyright (c) 1982, 1986, 1990, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)socketvar.h    8.1 (Berkeley) 6/2/93
 */

/*
 * Kernel structure per socket.
 * Contains send and receive buffer queues,
 * handle on protocol and pointer to protocol
 * private data and error information.
 */
struct sockbuf;

typedef struct socket {
    /* CUnknown... */
    const struct IEndpointVtbl *v;    /* IEndpoint dispatch */
    UINT RefCnt;                /* Reference count */
    /* ...CUnknown */
    PCONDITION pCond;           /* Signal when status changes */

    UINT16 so_type;            /* generic type, see socket.h */
    UINT16 so_options;         /* from socket call, see socket.h */
    INT16 so_linger;           /* time to linger while closing */
    UINT16 so_state;           /* internal state flags SS_*, below */
    PTR so_pcb;                /* protocol control block */
    const struct protosw *so_proto;  /* protocol handle */
/*
 * Variables for connection queueing.
 * Socket where accepts occur is so_head in all subsidiary sockets.
 * If so_head is 0, socket is not related to an accept.
 * For head socket so_q0 queues partially completed connections,
 * while so_q is a queue of connections ready to be accepted.
 * If a connection is aborted and it has so_head set, then
 * it has to be pulled out of either so_q0 or so_q.
 * We allow connections to queue up based on current queue lengths
 * and limit on number of queued connections for this socket.
 */
    struct socket *so_head;  /* back pointer to accept socket */
    struct socket *so_q0;    /* queue of partial connections */
    struct socket *so_q;     /* queue of incoming connections */
    INT16 so_q0len;          /* partials on so_q0 */
    INT16 so_qlen;           /* number of connections on so_q */
    INT16 so_qlimit;         /* max number queued connections */
    UINT16 so_error;         /* error affecting connection */
    TIME so_timeo;           /* connection timeout */
    pid_t so_pgid;           /* pgid for signals */
    UINT32 so_oobmark;       /* chars to oob mark */
/*
 * Variables for socket buffering.
 */
    struct sockbuf {
        UINT32 sb_cc;        /* actual chars in buffer */
        UINT32 sb_hiwat;     /* max actual char count */
        UINT32 sb_mbcnt;     /* chars of mbufs used */
        UINT32 sb_mbmax;     /* max chars of mbufs to use */
        UINT32 sb_lowat;     /* low water mark */
        struct mbuf *sb_mb;  /* the mbuf chain */
        TIME sb_timeo;       /* timeout for read/write */
        UINT16 sb_flags;     /* flags, see below */
    } so_rcv, so_snd;
#define SB_MAX (256L*1024L)    /* default for max chars in sockbuf */
#define SB_LOCK   0x01       /* lock on data queue */
#define SB_WANT   0x02       /* someone is waiting to lock */
#define SB_WAIT   0x04       /* someone is waiting for data/space */
#define SB_SEL    0x08       /* someone is selecting */
#define SB_ASYNC  0x10       /* ASYNC I/O, need signals */
#define SB_NOTIFY (SB_WAIT|SB_SEL|SB_ASYNC)
#define SB_NOINTR 0x40       /* operations not interruptible */

    PTR so_tpcb;             /* Wisc. protocol control block XXX */
    void (*so_upcall) (struct socket *so, PTR arg, INT waitf);
    PTR so_upcallarg;        /* Arg for above */
} *PISOCKET;

/*
 * Socket state bits.
 */
#define SS_NOFDREF         0x001  /* no file table ref any more */
#define SS_ISCONNECTED     0x002  /* socket connected to a peer */
#define SS_ISCONNECTING    0x004  /* in process of connecting to peer */
#define SS_ISDISCONNECTING 0x008  /* in process of disconnecting */
#define SS_CANTSENDMORE    0x010  /* can't send more data to peer */
#define SS_CANTRCVMORE     0x020  /* can't receive more data from peer */
#define SS_RCVATMARK       0x040  /* at mark on input */

#define SS_PRIV            0x080  /* privileged for broadcast, raw... */
#define SS_NBIO            0x100  /* non-blocking ops */
#define SS_ASYNC           0x200  /* async i/o notify */
#define SS_ISCONFIRMING    0x400  /* deciding to accept connection req */


/*
 * Macros for sockets and socket buffering.
 */

/*
 * How much space is there in a socket buffer (so->so_snd or so->so_rcv)?
 * This is problematical if the fields are unsigned, as the space might
 * still be negative (cc > hiwat or mbcnt > mbmax).  Should detect
 * overflow and return 0.  Should use "lmin" but it doesn't exist now.
 */
#define sbspace(sb) \
    ((INT) imin((INT)((sb)->sb_hiwat - (sb)->sb_cc), \
     (INT)((sb)->sb_mbmax - (sb)->sb_mbcnt)))

/* do we have to send all at once on a socket? */
#define sosendallatonce(so) \
    ((so)->so_proto->pr_flags & PR_ATOMIC)

/* can we read something from so? */
#define soreadable(so) \
    ((so)->so_rcv.sb_cc >= (so)->so_rcv.sb_lowat || \
    ((so)->so_state & SS_CANTRCVMORE) || \
    (so)->so_qlen || (so)->so_error)

/* can we write something to so? */
#define sowriteable(so) \
    ((sbspace(&(so)->so_snd) >= (long)(so)->so_snd.sb_lowat && \
    (((so)->so_state&SS_ISCONNECTED) || \
      ((so)->so_proto->pr_flags&PR_CONNREQUIRED)==0)) || \
     ((so)->so_state & SS_CANTSENDMORE) || \
     (so)->so_error)

/* adjust counters in sb reflecting allocation of m */
#define sballoc(sb, m) { \
    (sb)->sb_cc += (m)->m_len; \
    (sb)->sb_mbcnt += MSIZE; \
    if ((m)->m_flags & M_EXT) \
        (sb)->sb_mbcnt += MCLBYTES; \
}

/* adjust counters in sb reflecting freeing of m */
#define sbfree(sb, m) { \
    (sb)->sb_cc -= (m)->m_len; \
    (sb)->sb_mbcnt -= MSIZE; \
    if ((m)->m_flags & M_EXT) \
        (sb)->sb_mbcnt -= MCLBYTES; \
}

/*
 * Set lock on sockbuf sb; sleep if lock is already held.
 * Unless SB_NOINTR is set on sockbuf, sleep is interruptible.
 * Returns error without lock if sleep is interrupted.
 */
#define sblock(sb, wf) (((sb)->sb_flags & SB_LOCK) ? \
        (((wf) == M_WAITOK) ? sb_lock(sb) : EWOULDBLOCK) : \
        (((sb)->sb_flags |= SB_LOCK), 0))

/* release lock on sockbuf sb */
#define sbunlock(sb) { \
    (sb)->sb_flags &= ~SB_LOCK; \
    if ((sb)->sb_flags & SB_WANT) { \
        (sb)->sb_flags &= ~SB_WANT; \
        WakeUp((PTR)&(sb)->sb_flags); \
    } \
}

#define sorwakeup(so) { sowakeup((so), &(so)->so_rcv); \
            if ((so)->so_upcall) \
                (*((so)->so_upcall))((so), (so)->so_upcallarg, M_DONTWAIT); \
            }

#define sowwakeup(so) sowakeup((so), &(so)->so_snd)

#ifdef KERNEL
struct socket *sonewconn(struct socket *head, INT connstatus);
extern INT soreserve(struct socket *so, UINT32 sndcc, UINT32 rcvcc);
extern void sofree(struct socket *so);
extern INT soqremque(struct socket *so, INT q);
extern void soisconnecting(struct socket *so);
extern void soisconnected(struct socket *so);
extern void soisdisconnecting(struct socket *so);
extern void soisdisconnected(struct socket *so);
extern void socantsendmore(struct socket *so);
extern void socantrcvmore(struct socket *so);
extern void sowakeup(struct socket *so, struct sockbuf *sb);

extern void sbappend(struct sockbuf *sb, struct mbuf *m);
extern INT sbappendaddr(struct sockbuf *sb, struct sockaddr *asa,
             struct mbuf *m0, struct mbuf *control);
extern void sbdrop(struct sockbuf *sb, INT len);
extern void sbdroprecord(struct sockbuf *sb);
extern void sbrelease(struct sockbuf *sb);
extern INT sbreserve(struct sockbuf *sb, UINT32 cc);
extern INT sb_lock(struct sockbuf *sb);
extern INT sbwait(struct sockbuf *sb);
extern void sbflush(struct sockbuf *sb);

extern INT soreceive(struct socket *so, struct mbuf **paddr, struct uio *uio,
                     struct mbuf **mp0, struct mbuf **controlp, UINT32 *flagsp);
extern INT sosend2(struct socket *so, struct mbuf *addr, struct uio *uio,
                   struct mbuf *send_m, UINT flags);
extern INT soclose(struct socket *so);
extern INT soabort(struct socket *so);
extern void sohasoutofband(register struct socket *so);

/* strings for sleep message: */
extern const char netio[], netcon[], netcls[];
#endif
