/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <arm_asm.h>

/*
 * Startup code for ARM MMLite.
 * This code is shared by several ARM boards.
 * Separate board specific startup code jumps to ArmStart
 * 
 * Machine specific code #includes this file because the local
 * labels _start and ArmStart do not link otherwise.
 *
 * Not in C for a variety of reasons.
 * Also contains other assembly misc.
 */

#if !defined (ADS)
    _IMPORT(_end)
#endif
    _IMPORT(BaseInit)
    _IMPORT(TrapInit)
    _IMPORT(ExitUserMode)
    _IMPORT(BoardInit)
    _IMPORT(DebugBreak)

    _IMPORT(IntStackPtr)

/*
 * Board independent ARM startup code.
 * This code can be executed in three ways:
 * 1) Executed from Flash/ROM linked to RAM
 *    -> Need to copy .text, .data, and zero .bss
 *    -> All the sections are contiguous
 * 2) Executed from Flash/ROM linked to Flash/ROM
 *    -> Need to copy .data and zero .bss
 *    -> Copying .text is not ok (bus error or some such to be expected).
 *    -> .data immediately follows .text in Flash/ROM but link addresses
 *       are discontiguous. .bss immediately follows .data link addresses.
 * 3) Executed from RAM linked to RAM
 *    -> Need to copy .bss
 *    -> All the sections are contiguous
 *    -> Copying .text and .data is benign but useless
 *
 * 3 can be treated as either 1 or 2 so we treat it as 2 (less copying).
 */

_LABEL(ArmStart)
    /* Copy this image to RAM, where its supposed to go.
     * If already in RAM we harmlessly overcopy but still check out the code.
     * Source = r1
     */
    adr     r1, _L(_start)     /* Zero, really, but make it general */

    /* Dst = r2 */
    ldr     r2, _L(TextStart)+0

    /* EndPtr = r3 */
    ldr     r3, _L(DataEnd)+0

    /* If r1 != r2 then we are running in Flash/ROM but should run in RAM (2)
     * In that case copy .text. Otherwise skip to .data.
     */
    cmp     r1, r2
    bne     CpyLoop

    /* Adjust copy start to that of .data */
    ldr     r1, _L(DataStartROM)+0 /* source */
    ldr     r2, _L(DataStartRAM)+0 /* dst */
    /* r3 stays the same */

_LABEL(CpyLoop)
    /* loop copying 4 words at a time */
    ldmia   r1!, {r4,r5,r6,r7}
    stmia   r2!, {r4,r5,r6,r7}
    cmp     r2, r3
    blo     CpyLoop

    /* zero the bss
     */
    sub     r1, r1, r1
    sub     r2, r2, r2
    ldr     r4, _L(BssEnd)+0
_LABEL(ZroLoop)
    cmp     r3, r4
    stmloia r3!, {r1,r2}
    blo     ZroLoop

    /*
     * Make sure we're in SVC mode
     * The ROM monitor (ANGEL) starts us in USER mode.
     * If we are running from FLASH we can still do this, its harmless.
     * [because the SRAM is on, that is!]
     */
    bl      ExitUserMode

    /* Jump to RAM copy (below)
     */
    ldr     pc, _L(RelocToRAM)

_LABEL(RelocToRAM)
    _DCD    InRAM

    /* More static data, since we are at it.
     */

_LABEL(TextStart)
    /* Text start at final link address */
    _DCD    _start

_LABEL(DataStartROM)
    /* Data start in ROM == text end */
#if defined(__GNUC__) || defined(_MSC_VER)
    _IMPORT(_etext)
    _DCD    _etext
#else /* one of the ARM toolsets */
    _IMPORT(|Image$$RO$$Limit|)
    _DCD    |Image$$RO$$Limit|
#endif

_LABEL(DataStartRAM)
    /* Data start in RAM */
#if defined(__GNUC__) || defined(_MSC_VER)
    _IMPORT(__data_start)
    _DCD    __data_start
#else /* one of the ARM toolsets */
    _IMPORT(|Image$$RW$$Base|)
    _DCD    |Image$$RW$$Base|
#endif

_LABEL(DataEnd)
    /* Data end == bss start */
#if defined(__GNUC__)
    _IMPORT(_edata)
    _DCD    _edata
#elif defined(_MSC_VER)
     _IMPORT(__bss_start)       /* or __bss_start__ */
     _DCD   __bss_start
#else /* one of the ARM toolsets */
    _IMPORT(|Image$$ZI$$Base|)
    _DCD    |Image$$ZI$$Base|
#endif

_LABEL(BssEnd)
    /* Bss end */
#if defined (ADS)
    _IMPORT(|Image$$ZI$$Limit|)
    _DCD    |Image$$ZI$$Limit|
#else
    _DCD    _end
#endif

    /*
     * Execute from here after copying out of FLASH into RAM
     */
_LABEL(InRAM)

    /*
     * Set stack pointer for SVC mode
     * BUGBUG Save some memory by using SRAM
     * [but what if its all needed for apps]
     */
    ldr     sp, = _L(IntStackPtr)
    ldr     sp, [sp]

    /*
     * Initialize the board
     */
    bl      _F(BoardInit)

    /*
     * Initialize interrupt dispatching
     */
    bl      _F(TrapInit)

    /*
     * Drop into debugger if required.
     */
#if 0
    bl      _F(DebugBreak)
#endif

    /*
     * Go straight for MI init
     */
    b       _F(BaseInit)
    b       _start
