/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <arm_asm.h>

/*
 * Startup code for ARM MMLite.
 * Version for the Atmel EB63 board.
 *
 * Not in C for a variety of reasons.
 * Also contains other assembly misc.
 */

    _CODE(C$$code)

    _EXPORT(_start)

/*
 * Entry point
 *
 * This startup code is executed at chip powerup/reset.
 * Lives in FLASH, is visible at address 0x0 on powerup,
 * and at address 0x1000000 after TheEbi.RemapRegister
 * has been written to.
 *
 * One desireable feature of this code is to be executable
 * at both of the addresses indicated above. This way a
 * system reset can be performed by jumping at 0x1000000.
 * [NB: we could use the watchdog timer instead]
 *
 * This very same code can also be downloaded at its intended
 * execution place in RAM, and started and debugged in there.
 */
    _ENTRY

_FUNCTION(_start)
    msr     cpsr_c, #0xD3       /* disable ints, SVC mode */
                                /* NB: under Angel only turns intrs off */
    /* Setup the EBI:
     * - enable FLASH by writing to CS0
     * - enable RAM by writing to CS1
     *   and setting TheEbi->MemoryControl = EBIMC_ALE_CS4_6;
     * - stop remapping 
     */
    adr     r1, _L(TheEbi)     /* Setup regs with Ebi ptr and values */
    ldmia   r1, {r1-r10}

    stmia   r1, {r2-r9}        /* Set CS0&CS1 */
    str     r10, [r1, #+0x24]   /* Set MemoryControl */

    mov     r2, #1             /* for RemapControl and 0x1000000 */
    adr     r3, _L(Relocate)
    orr     r3, r3, r2, lsl #24/* r3 = Relocate+0x1000000 */
                               /* This is in FLASH, ONLY good at boot */
    movs    r4, pc, lsr #24    /* FLASH is lower than 0x1000000 at reset */
    str     r2, [r1, #+0x20]   /* stop remapping (if we were doing it) */
    moveq   pc, r3             /* .. and use new addresses if we have to */

    /* At this point:
     * - Memory mapping is the normal one, with FLASH at 1mil,
     *   RAM at 2mil, and SRAM at 0.
     * - If this code is in FLASH, we are running with pc=0x010000xx
     * - If this code is debugged in RAM, pc=0x020080xx (or similar).
     */
_LABEL(Relocate)
#define FRAM 1
#if FRAM
    /* We need to turn on power to the FRAM chips, as we are going
     * to either copy ourselves there or at least use them as heap.
     * To this we need to turn PB16 on and wait a bit.
     */
    adr     r1, _L(ThePio)     /* Setup regs with Pio ptr and values */
    ldmia   r1, {r1-r2}
    str     r2, [r1, #+0x00]   /* Set Enable for PB16 */
    str     r2, [r1, #+0x10]   /* Set OutEnable for PB16 */
    str     r2, [r1, #+0x30]   /* Set SetData for PB16 (e.g. PB16=1=ON) */
    /* Not clear to me if this is absolutely necessary.. */
_LABEL(PowerupSpin)
    subs    r2,r2,#1
    bne     PowerupSpin
#endif	

	/* Continue initializing the regular way
     */     
    b       _L(ArmStart)        /* PC relative: does not matter where we are */
    b       _start

    /* Static data needed above.
     */
_LABEL(TheEbi)
    _DCD    0xffe00000
    _DCD    0x01002529         /* CS0 for FLASH (see _fredEb63.c) */
    _DCD    0x30002125         /* CS1 for RAM   (see _fredEb63.c) */
    _DCD    0x20002362         /* CS2 for ???   (see _fredEb63.c) */
    _DCD    0x03002f3d         /* CS3 extRAM  (see _fredEb63.c) */
    _DCD    0x40002125         /* CS4 for MMU   (see _fredEb63.c) */
    _DCD    0x50000000         /* CS4 for MMU   (see _fredEb63.c) */
    _DCD    0x60000000         /* CS4 for MMU   (see _fredEb63.c) */
    _DCD    0x70000000         /* CS4 for MMU   (see _fredEb63.c) */
    _DCD    0x00000004         /* MemoryControl (see _fredEb63.c) */
_LABEL(ThePio)
    _DCD    0xffff0000         /* Pointer to PIO-B */
    _DCD    0x00010000         /* Bit for PB16     */
#include "../_armstart.s"
    _END
