#include <arm_asm.h>

/*
 * Startup code for ARM MMLite.
 * Version for the Atmel AT91FR4081 chip on MSFT Zenith board.
 *
 * Not in C for a variety of reasons.
 * Also contains other assembly misc.
 */

    _CODE(C$$code)

#if !defined (ADS)
    _IMPORT(_end)
#endif
    _IMPORT(BaseInit)
    _IMPORT(TrapInit)
    _IMPORT(ExitUserMode)
    _IMPORT(Z1Init)
    _IMPORT(DebugBreak)

    _IMPORT(IntStackPtr)

    _EXPORT(_start)

/*
 * Entry point
 *
 * This startup code is executed at chip powerup/reset.
 * Lives in FLASH, is visible at address 0x0 on powerup,
 * and at address 0x01000000 after TheEbi.RemapRegister
 * has been written to.  On chip RAM is at 0x00100000.
 *
 * One desireable feature of this code is to be executable
 * at both of the addresses indicated above. This way a
 * system reset can be performed by jumping at 0x1000000.
 * [NB: we could use the watchdog timer instead]
 * This very same code can also be downloaded at its intended
 * execution place in RAM, and started and debugged in there.
 * Finally, we want to be able to just execute in-place;
 * meaning code out of flash and data+bss in RAM. (TODO).
 */
    _ENTRY

_FUNCTION(_start)
//    b _start /* allow ICE debug if in Flash */

    msr     cpsr_c, #0xD3       /* disable ints, SVC mode */
                                /* NB: under Angel only turns intrs off */
    /* Setup the EBI:
     * - enable FLASH by writing to CS0
     * - enable RAM by writing to CS1
     *   and setting TheEbi->MemoryControl = EBIMC_ALE_CS4_6;
     * - stop remapping 
     */
    adr     r1, _L(TheEbi)     /* Setup regs with Ebi ptr and values */
    ldmia   r1, {r1,r2,r3,r4,r5}

    stmia   r1, {r2,r3}        /* Set CS0&CS1 */
    str     r4, [r1, #+0x24]   /* Set MemoryControl */

    mov     r2, #1             /* for RemapControl, RAMWU and 0x1000000 */
    str     r2, [r5, #0]       /* Fs->MemoryMode = 1 (RAM writeable) */
    adr     r3, _L(Relocate)   /* PC-relative version of "Relocate" */
    orr     r3, r3, r2, lsl #24/* r3 = Relocate+0x1000000 */
                               /* This is in FLASH, ONLY good at boot */
    movs    r4, pc, lsr #20    /* FLASH is lower than 0x1000000 at reset */
                               /* and on-chip SRAM is at 0x100000 */
    str     r2, [r1, #+0x20]   /* stop remapping (if we were doing it) */
    moveq   pc, r3             /* .. and use new addresses if we have to */

    /* At this point:
     * - Memory mapping is the normal one, with FLASH at 1mil,
     *   RAM at 100k, and SRAM at 0.
     * - If this code is in FLASH, we are running with pc=0x010000xx
     * - If this code is debugged in RAM, pc=0x001000xx (or similar).
     */
_LABEL(Relocate)
    /* Copy this image to RAM, where its supposed to go.
     * If already in RAM we harmlessly overcopy but still check out the code.
     * Source = r1
     */
    adr     r1, _L(_start)     /* Zero, really, but make it general */

    /* Dst = r2
     */
    ldr     r2, _L(ImageLimits)+0

    /* EndCount = r3
     */
    ldr     r3, _L(ImageLimits)+4

    /* loop copying 4 words at a time */
_LABEL(CpyLoop)
    ldmia   r1!, {r4,r5,r6,r7}
    stmia   r2!, {r4,r5,r6,r7}
    cmp     r2, r3
    blo     CpyLoop

    /* zero the bss
     */
    sub     r1, r1, r1
    sub     r2, r2, r2
    ldr     r4, _L(ImageLimits)+8
_LABEL(ZroLoop)
    cmp     r3, r4
    stmloia r3!, {r1,r2}
    blo     ZroLoop

    /*
     * Make sure we're in SVC mode
     * The ROM monitor (ANGEL) starts us in USER mode.
     * If we are running from FLASH we can still do this, its harmless.
     * [because the SRAM is on, that is!]
     */
    bl      ExitUserMode

    /* Jump to RAM copy (below)
     */
    ldr     pc, _L(RelocToRAM)

_LABEL(RelocToRAM)
    _DCD    InRAM

    /* More static data, since we are at it.
     */
_LABEL(TheEbi)
    _DCD    0xffe00000
    _DCD    0x01003429         /* CS0 for FLASH (see _fredZ1.c) */
                               /* 1M, conservative, if 40MHz */
    _DCD    0x20000000         /* CS1 for extRAM (see _fredZ1.c) */
    _DCD    0x00000016         /* MemoryControl (see _fredZ1.c) */
    _DCD    0xfff0000c         /* FS:MemoryMode (see _fredZ1.c) */
    
_LABEL(ImageLimits)
    _DCD    _start

#if defined(__GNUC__)
    _IMPORT(_edata)            /* or __bss_start, __bss_start__ */
    _DCD    _edata
#elif defined(_MSC_VER)
     _IMPORT(__bss_start)
     _DCD   __bss_start
#else /* one of the ARM toolsets */
    IMPORT  |Image$$ZI$$Base|
    _DCD    |Image$$ZI$$Base|
#endif

#if defined (ADS)
    IMPORT  |Image$$ZI$$Limit|
    _DCD    |Image$$ZI$$Limit|
#else
    _DCD    _end
#endif

    /*
     * Execute from here after copying out of FLASH into RAM
     */
_LABEL(InRAM)

    /*
     * Set stack pointer for SVC mode
     * BUGBUG Save some memory by using SRAM
     * [but what if its all needed for apps]
     */
    ldr     sp, = _L(IntStackPtr)
    ldr     sp, [sp]

    /*
     * Initialize the board
     */
    bl      _F(Z1Init)

    /*
     * Initialize interrupt dispatching
     */
    bl      _F(TrapInit)

    /*
     * Drop into debugger if required.
     */
#if 0
    bl      _F(DebugBreak)
#endif

    /*
     * Go straight for MI init
     */
    b       _F(BaseInit)

    b       _start

    _END
