/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <string.h>
#include <fred.h>
#include <x86_cpu.h>
#include <_dosx.h>
#include "_pic.h"

DOSX_INTERFACE Dosx;
extern INT *bstack;     /* in main.c */

void InitializeDosx( PDOSX_INTERFACE pIface)
{
    Dosx = *pIface;
}

/* Return a pointer to the info DOSX passed to us.
 */
PTR /*PDOSX_INTERFACE*/ DosxPointer(void)
{
    return (PTR) &Dosx;
}

/* Extract the command line */
const char *BootCmdLine(void)
{
    return (const char*) DosxCmdLine();
}

/* Say we are NOT a simulator
 */
UINT CurrentProcessorOEM(void)
{
    /* Somewhere in the BIOS I suppose.. */
    return 0;
}

/* Create the system heap.  Left to machdep code because messy otherwise.
 * The heap should account for all available memory, carved out of memory
 * that is already used or unavailable (like the 640K-1meg hole on PCs).
 * It does not have to account for all the memory in the system.
 * Returns a PIHEAP pointer.
 */
PIHEAP MachineHeapCreate( void )
{
    ADDRESS MemBase, MemTop;
    ADDR_SIZE MemSize;
    PIHEAP pIHeap;
#define _HOLE_STARTS 0xa0000
#define _HOLE_SIZE (0x100000-0xa0000)
#define _RELOCATED_HOLE (_HOLE_STARTS-Dosx.RelocationBase)

#define _USE_ALL_MEMORY 1
#if _USE_ALL_MEMORY
    DOSX_CALLBACK_ARGUMENTS Args;
    /* Find out how much extended memory we got, if any
     */
    Args.gp.eax.r8.rh = (UINT8)0x88;
    SafelyCallback(0x15, &Args, &Args);
    if ((Args.eflags & X86_FLAG_CF) == 0) {
        /* Add low mem back in, careful about overflows here */
        MemTop = (Args.gp.eax.r32.erx + 1024);
        /* Take relocation away */
        MemTop -= (Dosx.RelocationBase / 1024);
        /* Turn into an address now */
        MemTop = (MemTop * 1024) + (Dosx.RelocationBase % 1024);
    } else
        MemTop = _RELOCATED_HOLE;
#else
    /* Only use the low 640 K
     */
    MemTop = _RELOCATED_HOLE;
#endif

    MemBase = Dosx.ImageSize;
    MemSize = MemTop - MemBase;

    pIHeap = CreateHeapFrom(MemBase, 0, MemSize, MemSize);

    /* Carve out "the hole"
     */
    if (MemTop > _RELOCATED_HOLE) {
        pIHeap->v->Extract(pIHeap, 0,
                           (PTR)_RELOCATED_HOLE,
                           _HOLE_SIZE);
    }

    return pIHeap;
}

static int *saved_esp;
__declspec(naked) void CallBackHelper(void)
{
    /* Switch to bios stack and call dos extender.  Args are already pushed */
    __asm {
        pushad;
        mov saved_esp,esp;
        mov esp,bstack;
        /* Dosx.CallbackRoutine(IntNo, pIn, pOut); */
        call Dosx.CallbackRoutine;
        mov esp,saved_esp;
        popad;
        ret;
    }
}

/* Make a call to DOS/BIOS, i.e. execute an INT<N> instruction
 * in real mode, with the PIN register arguments.
 * Return the resulting register state in POUT.
 * In between, do anything necessary to restore the machine state
 * to what the DOS extender had it.
 */
extern INT16 __cdecl PITRead (void);
extern void PITWrite (UINT nTicks);
extern void DisableTimers( void ), EnableTimers( void );

void SafelyCallback( UINT IntNo,
                     PDOSX_CALLBACK_ARGUMENTS pIn,
                     PDOSX_CALLBACK_ARGUMENTS pOut )
{
    UINT Eflags;
    INT16 oPIT;
    IDT_POINTER Kidt;
    INT32 *in, *out, *oldbstack;

    /* No interrupts: the bios stack is global.
     */
    Eflags = PicToBios();

    /* Put the PIT/RTC back the way BIOS had them
     */
    oPIT = PITRead();
    DisableTimers();

    __asm {
        sidt    Kidt;
    }

    /* Push arguments to bios stack
     */
    oldbstack = bstack;
    bstack -= sizeof(*pIn) / sizeof(int);
    memmove(bstack, pIn, sizeof(*pIn));
    in = bstack;

    bstack -= sizeof(*pOut) / sizeof(int);
    out = bstack;

    *--bstack = (int) out; /* BUGBUG: compensate for relo base? */
    *--bstack = (int) in;
    *--bstack = IntNo;

    /* Make the call
     */
    CallBackHelper();

    /* Copy return args back to our stack
     */
    if (pOut)
      memmove(pOut, out, sizeof(*pOut));

    /* Restore bios stack
     */
    bstack = oldbstack;

    __asm {
        lidt    Kidt;
    }

    /* Restart PIT/RTC from where they left (more or less)
     */
    EnableTimers();
    PITWrite(oPIT);

    /* And we are back in business.
     */
    PicFromBios(Eflags);
}
