/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mips_asm.h>

/*
 * Startup code for MIPS MMLite.
 * This code is shared by several MIPS boards.
 * Separate board specific startup code jumps to MipsStart
 * 
 * Machine specific code #includes this file because the local
 * labels _start and MipsStart do not link otherwise.
 *
 * Not in C for a variety of reasons.
 * Also contains other assembly misc.
 */


    _IMPORT(BaseInit)
    _IMPORT(TrapInit)
    _IMPORT(BoardInit)

    _IMPORT(_etext)
    _IMPORT(__data_start)
    _IMPORT(_edata)
    _IMPORT(_end)

/*
 * Board independent MIPS startup code.
 * This code can be executed in three ways:
 * 1) Executed from Flash/ROM linked to RAM
 *    -> Need to copy .text, .data, and zero .bss
 *    -> All the sections are contiguous
 * 2) Executed from Flash/ROM linked to Flash/ROM
 *    -> Need to copy .data and zero .bss
 *    -> Copying .text is not ok (bus error or some such to be expected).
 *    -> .data immediately follows .text in Flash/ROM but link addresses
 *       are discontiguous. .bss immediately follows .data link addresses.
 * 3) Executed from RAM linked to RAM
 *    -> Need to zero .bss
 *    -> All the sections are contiguous
 *    -> Copying .text and .data is benign but useless
 *
 * 3 can be treated as either 1 or 2 so we treat it as 2 (less copying).
 */

/*
 * void MipsStart(ADDRESS ActualTextStart)
 * ActualTextStart is where "_start" landed, e.g. 0xbfc00000 probably.
 */
ENTRY(MipsStart)
    /* Copy this image to RAM, where it is supposed to go.
     * If already in RAM we harmlessly overcopy but still check out the code.
     * Source = $a0
     */

    /* If we had PC-relative addressing... sigh.
     */
	.set noreorder
    bgezal  $zero, _L(MipsStart2) /* Always jumps */
    nop
        
    /* All of the static data, since we are at it.
     */
_LABEL(TextStart)                         /* + 0 */
    /* Text start at final link address */
    _DCD    _start

_LABEL(DataStartROM)                      /* + 4 */
    /* Data start in ROM == text end */
    _DCD    _etext

_LABEL(DataStartRAM)                      /* + 8 */
    /* Data start in RAM */
    _DCD    __data_start

_LABEL(DataEnd)                           /* +12 */
    /* Data end == bss start */
    _DCD    _edata

_LABEL(BssEnd)                            /* +16 */
    /* Bss end */
    _DCD    _end

_LABEL(RelocToRAM)                        /* +20 */
    _DCD    InRAM

_LABEL(MipsStart2)

    /* Dst = $t2 */
    lw      $t2, 0($ra)     /* _L(TextStart) */

    /* EndPtr = $t3 */
     /* in bdelay slot */

    /* If $a0 != $t2 then we are running in Flash/ROM but should run in RAM (2)
     * In that case copy .text. Otherwise skip to .data.
     */
    bne     $a0,$t2,CpyLoop
    lw      $t3, 12($ra)    /* _L(DataEnd)   */

    /* Adjust copy start to that of .data */
    lw      $a0, 4($ra)     /* _L(DataStartROM)  is source */
    lw      $t2, 8($ra)     /* _L(DataStartRAM)  is dst */
    /* $t3 stays the same */
#if 1
    /* For simulation's sake: .data might be ok too */
    beq     $a0,$t2,ZroLoop-4
    nop
    
#endif

_LABEL(CpyLoop)
    /* loop copying 2 words at a time */
    lw      $t4,0($a0)
    lw      $t5,4($a0)
    addiu   $a0,$a0,8
    sw      $t4,0($t2)
    addiu   $t2,$t2,8
    sltu    $t1,$t2,$t3
    bne     $t1,$zero,CpyLoop
    sw      $t5,-4($t2)

    /* zero the bss
     */
    lw      $t4, 16($ra)   /* _L(BssEnd)  */
_LABEL(ZroLoop)
    sltu    $t1,$t3,$t4
    sw      $zero,0($t3)
    bne     $t1,$zero,ZroLoop
    addiu   $t3,$t3,4

    /* Jump to RAM copy (below)
     */
    lw      $t1, 20($ra)   /* _L(RelocToRAM) */
    jr      $t1
    nop

    /*
     * Execute from here after copying out of FLASH into RAM
     */
_LABEL(InRAM)

    /*
     * Set stack pointer to somewhat after the end of the BSS
     */
#define BOOT_STACK_SIZE (8*1024)   /* 4 Kilobytes */
    li      $t1,BOOT_STACK_SIZE

    /*
     * Initialize the board
     */
    jal     _F(BoardInit)
    addu    $sp,$t3,$t1

    /*
     * Install the exception handlers
     */
    jal     _F(TrapInit)
    nop

    mfc0    $s0, $cp0_status
    li      $a0, ~CP0_SR_BEV
    and     $s0,$s0,$a0

    /*
     * Done, do the MI init and dont come back
     */
    j       _F(BaseInit)
    mtc0    $s0, $cp0_status

END(MipsStart)

