/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <ppc_asm.h>
#define __ASSEMBLER__ 1
#include "eb63.h"

	_CODE(text)

/*
 *
 * VOID
 * FlushCache(
 *     IN ADDRESS BaseAddress,
 *     IN ADDR_SIZE nBytes
 *     )
 *
 * Routine Description:
 *
 *     This function flushes a range of the instruction cache of all processors.
 *
 * Arguments:
 *
 *     BaseAddress - Specifies the starting virtual address to begin the flush.
 *
 *     Length - Specifies the number of bytes to flush.
 *
 * Return Value:
 *
 *     None.
 *
 */

ENTRY(FlushCache)

.errnz  ICACHE_BLOCK_SIZE - DCACHE_BLOCK_SIZE

/*
 * Exit early if the number of bytes is zero.
 */

        cmplwi  r4,0
        beqlr-

/*
 * Compute the number of bytes spanned by the cache block given the base address
 * and length and align the base address to a cache block boundary.
 */

        clrlwi  r5,r3,32-DCACHE_BLOCK_SIZE_BITS
        add     r4,r4,r5
        addi    r4,r4,DCACHE_BLOCK_SIZE-1
        srwi    r4,r4,DCACHE_BLOCK_SIZE_BITS
        clrrwi  r3,r3,DCACHE_BLOCK_SIZE_BITS

/*
 * Flush each cache block from the data cache.
 */

        mtctr   r4
        mr      r5,r3

SweepNextDcacheBlock:
        dcbf    r0,r3
        addi    r3,r3,DCACHE_BLOCK_SIZE
        bdnz+   SweepNextDcacheBlock
        sync

/*
 * Flush each cache block from the instruction cache.
 */

        mtctr   r4

SweepNextIcacheBlock:
        icbi    r0,r5
        addi    r5,r5,DCACHE_BLOCK_SIZE
        bdnz+   SweepNextIcacheBlock
        sync
        isync

        blr

END(FlushCache)

/*
 *
 * VOID
 * FlushCurrentEntireTb(
 *     IN BOOLEAN ZeroEntireTb
 *     )
 *
 * Routine Description:
 *
 *     This function flushes the entire translation lookaside buffer of the
 *     current processor.
 *
 * Arguments:
 *
 *     ZeroEntireTb - Specifies whether hardware page table should be zeroed.
 *                    When this argument is FALSE, user-mode entries are not
 *                    invalidated from the hardware page table.
 *
 * Return Value:
 *
 *     None.
 *
 */

ENTRY(FlushCurrentEntireTb)

/*
 * Disable interrupts.
 */

        li      r4,MSR_EE_MASK
        mfmsr   r12
        and     r12,r12,r4                  // find current EE state
	
        li      r4,0
        mtmsree r4                          // disable interrupts
        mfmsr   r8
        or      r12,r8,r12                  // add previous EE state
        isync                               // context synchronize

/*
 * Disable instruction and data relocation so that we don't need to worry about
 * modifications to the hashed page table from the trap handlers while we're
 * modifying the table here.
 */

        lis     r5,RelocationsDisabled
        addi    r5,r5,RelocationsDisabled
        clrlwi  r5,r5,1                     // clear high order address bit
        mtsrr0  r5
        mfmsr   r5
        li      r8,MSR_IR_MASK+MSR_DR_MASK
        andc    r5,r5,r8
        mtsrr1  r5
        rfid
RelocationsDisabled:

/*
 * Walk over an effective address range and flush all sets of the TLB.
 */

        li      r8,0
        li      r9,1024                     // flush 1024 sets
        mtctr   r9

FlushNextEffectiveAddress:

        tlbiel  r8,0
        addi    r8,r8,1000h                 // advance to next TLB set
        bdnz+   FlushNextEffectiveAddress

        ptesync

        mfsdr1  r8
        clrrwi  r8,r8,16                    // clear HTABSIZE and reserved bits
        li      r9,LOWWORD(HASHED_PAGE_TABLE_PHYSICAL_SIZE SHR 7)
        mtctr   r9

        cmpwi   r3,0                        // zero entire page table?
        beq-    FlushNextHashedPageTableEntryGroup

/*
 * Zero the contents of the hashed page table.
 */

ZeroNextHashedPageTableCacheBlock:
        dcbz128 r0,r8
        addi    r8,r8,128
        bdnz+   ZeroNextHashedPageTableCacheBlock
        b       FlushComplete

/*
 * Flush each PTEG by invalidating kernel-mode entries and moving any user-mode
 * entries to the beginning.  The PTEG will end up with either no valid entries
 * or one or more valid user-mode entries followed by zero or more invalid entries.
 */

FlushNextHashedPageTableEntryGroup:
        mr      r10,r8                      // trailing pointer for user mode entries
FlushNextHashedPageTableEntry:
        ld      r5,0(r8)
        andi.   r11,r5,1                    // PTE marked as valid?
        beq+    FlushNextHashedPageTableEntryGroupAdvance
        rlwinm. r11,r5,0,15,15              // user mode bit on?
        beq+    FlushNextHashedPageTableEntryInvalidate
        cmplw   r10,r8                      // is there a gap?
        addi    r10,r10,16                  // advance trailing pointer
        beq+    FlushNextHashedPageTableEntryAdvance
        std     r5,-16(r10)                 // move PTE, then invalidate entry
        ld      r5,8(r8)
        std     r5,-8(r10)
FlushNextHashedPageTableEntryInvalidate:
        li      r5,0                        // turn off valid bit
        std     r5,0(r8)
FlushNextHashedPageTableEntryAdvance:
        addi    r8,r8,16
        andi.   r11,r8,127
        bne+    FlushNextHashedPageTableEntry
        bdnz+   FlushNextHashedPageTableEntryGroup
        b       FlushComplete
FlushNextHashedPageTableEntryGroupAdvance:
        rlwinm  r8,r8,0,0,24                // back up to beginning of PTEG
        addi    r8,r8,128                   // advance to next PTEG
        bdnz+   FlushNextHashedPageTableEntryGroup

FlushComplete:

/*
 * Restore instruction and data relocation and interrupts and return to the
 * caller.
 */

        mflr    r6
        mtsrr0  r6
        mtsrr1  r12
        rfid

END(FlushCurrentEntireTb)

        _ENDCODE
        _END
