/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define OPTIMIZE        /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <string.h>
#include <base/cons.h>

#undef DBGOUT
//#define DBGOUT(a,b,c) c;
#define DBGOUT(a,b,c);

#define KimageToUnmapped(a) (a)

/* Memory-Device Device Driver
 */

typedef struct {
    const struct IFileVtbl *v;
    UINT     RefCnt;
    PIHEAP   pIHeap;
    UINT8    *BaseOfMemory;
    UINT     SizeOfMemory;
    UINT     FsType;
#define FS_ISA_FAT 1
#define FS_ISA_ROM 2
} MEMORY_DEVICE;

/* Magic header.
 */
#define SIGNATURE_SIZE 4 /* words */

struct {
    UINT32 Signature[SIGNATURE_SIZE];
    UINT32 FileSystemType;
    UINT32 FileSystemSize;
    UINT8  *FileSystemPointer;
} MemoryFileSystemData = {
    /*
     * Signature, needed by tools to find this array
     * in an arbitrary image (any CPU, any format).
     */
    { 0x11131719, 0x01020304, 0x04030201, 0x19171311 },
    0, /* nothing in here */
    0,
    NULL
};

/* Alternative runtime way of specifying the file system params.
 */
void MemDeviceSet( UINT32 FSSize, ADDRESS FSPointer )
{
    MemoryFileSystemData.FileSystemType = FS_ISA_ROM;
    MemoryFileSystemData.FileSystemSize = FSSize;
    MemoryFileSystemData.FileSystemPointer = (UINT8 *) FSPointer;
}

BOOL MemDeviceParams( MEMORY_DEVICE *pDevice )
{
    /*
     * mkmemfs sets FileSystemPointer to a boot image value.  Since only the kernel portion
     * of the boot image is mapped, we must convert the pointer to an unmapped address.
     */
    if (MemoryFileSystemData.FileSystemPointer != NULL)
        pDevice->BaseOfMemory = (UINT8 *)
                KimageToUnmapped((ADDRESS) MemoryFileSystemData.FileSystemPointer);
    else
        pDevice->BaseOfMemory = NULL;

    pDevice->SizeOfMemory = MemoryFileSystemData.FileSystemSize;
    pDevice->FsType = MemoryFileSystemData.FileSystemType;
    if (pDevice->BaseOfMemory == NULL ||
        pDevice->SizeOfMemory == 0)
        return FALSE;

    DBGOUT(DBGFS,1,dprintf("InMemoryFilesystem at x%x, size x%x bytes\n",
                           pDevice->BaseOfMemory, pDevice->SizeOfMemory));
    return TRUE;
}

/* IFile methods
 */

UINT MCT MemDeviceAddRef(PIFILE pThis)
{
    MEMORY_DEVICE * File = (MEMORY_DEVICE *) pThis;

    return AtomicInc(&File->RefCnt);
}

UINT MCT
MemDeviceRelease( PIFILE pThis)
{
    MEMORY_DEVICE * pDevice = (MEMORY_DEVICE *) pThis;
    UINT Refs;

    Refs = AtomicDec(&pDevice->RefCnt);
    if (Refs == 0)
        (void) pDevice->pIHeap->v->Free(pDevice->pIHeap, 0, pDevice);

    return Refs;
}

SCODE MCT MemDeviceQueryInterface(PIFILE pThis, REFIID pIid, void* *ppNew)
{
    if (UuidCmp(&IID_IFile, pIid) || UuidCmp(&IID_IUnknown, pIid)) {
        MemDeviceAddRef(pThis);
        *ppNew = pThis;
        return NO_ERROR;
    }
    *ppNew = NULL;
    return E_NO_INTERFACE;
}

SCODE MCT
MemDeviceReadAt( PIFILE pThis, UINT64 Offset, BYTE *vBuf, UINT Size, PUINT pSizeRead)
{
    MEMORY_DEVICE * pDevice = (MEMORY_DEVICE *) pThis;

    DBGOUT(DBGFS,0,dprintf("MemDeviceReadAt(x%x,x%x,x%x,x%lx)\n", pThis,vBuf,Size,Offset));

    *pSizeRead = Size;
    memcpy((PTR)vBuf,
           pDevice->BaseOfMemory + Int64ToInt32(Offset),
           Size);
    return NO_ERROR;
}

SCODE MCT
MemDeviceWriteAt( PIFILE pThis, UINT64 Offset, const BYTE *vBuf, UINT Size, PUINT pSizeWritten)
{
    DBGOUT(DBGFS,2,dprintf("MemDeviceWriteAt(x%x,x%x,x%x,x%x)\n", pThis,vBuf,Size,Offset));

    *pSizeWritten = Size;
#if 0
    MemCpy(pDevice->BaseOfMemory + (UINT) (*((PINT64)&Offset)),
           (PTR)vBuf,
           Size);
#else
    UnusedParameter(pThis);
    UnusedParameter(Offset);
    UnusedParameter(vBuf);
#endif
    return NO_ERROR;
}

/*
 */
SCODE MCT
MemDeviceGetSize( PIFILE pThis, PUINT64 pSize)
{
    UINT64 Bs;

    Int64FromHighAndLow(Bs, 0, ((MEMORY_DEVICE *)pThis)->SizeOfMemory);
    *pSize = Bs;

    return NO_ERROR;
}


/*
 */
SCODE MCT
MemDeviceSetSize( PIFILE pThis, UINT64 Size)
{
    UnusedParameter(pThis);
    UnusedParameter(Size);

    return E_INVALID_PARAMETER;
}

const struct IFileVtbl MemDeviceVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    MemDeviceQueryInterface,
    MemDeviceAddRef,
    MemDeviceRelease,
    MemDeviceReadAt,
    MemDeviceWriteAt,
    MemDeviceSetSize,
    MemDeviceGetSize
};

/* Constructor
 */
MEMORY_DEVICE *MemDeviceNew( void )
{
    PIHEAP      pIHeap = CurrentHeap();
    MEMORY_DEVICE       *pDevice;

    pDevice = (MEMORY_DEVICE *) pIHeap->v->Alloc(pIHeap,
                                  HEAP_DEBUG_NAME("memdev") | HEAP_ZERO_MEMORY,
                                  sizeof(MEMORY_DEVICE), 0);
    if (!pDevice)
        return NULL;
    pDevice->v = &MemDeviceVtbl;
    pDevice->RefCnt = 1;
    pDevice->pIHeap = pIHeap;

    if (!MemDeviceParams( pDevice )) {
        (void) pDevice->v->Release((PIFILE)pDevice);
        pDevice = NULL;
    }

    return pDevice;
}

#define RootFsStrName TEXT("RomFs")

/*
 * How the kernel uses this device
 */
PINAMESPACE InitMemoryFileSystem(void)
{
    MEMORY_DEVICE *pMemFile;
    PINAMESPACE Ns;

    /* Do we have a Memory FS
     */
    pMemFile = MemDeviceNew();

    if (pMemFile == NULL)
        return NULL;

    /* Is it structured correctly
     */
    switch (pMemFile->FsType) {
    case FS_ISA_FAT:
        DBGOUT(DBGFS,2,dprintf("No FAT support, cannot use this memory FS\n"));
        return NULL;
    case FS_ISA_ROM:
        Ns = RomFsNew(pMemFile->BaseOfMemory, pMemFile->SizeOfMemory);       
        return Ns;
    default:
        return NULL;
    }

}
