/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <file.h>
#include <fred.h>
typedef struct IFile *PIFILE;

#undef DBGOUT
#define DBGOUT(x,y,z) z

/* Uses of mutable functions:
 *
 * - Specialization
 * - Interposition & Migration
 * - Replacement  (e.g. floppy driver)
 * - Mutual exclusion
 */

typedef struct _TEST_FILE {
    const struct IFileVtbl *TestFileV;
    UINT    ActiveCnt;
    UINT    RefCnt;

    char *pData;
    UINT Size;

} *PTESTFILE;

#define pTF(_x_) ((PTESTFILE)_x_)

/* Entry for a mutable (repluggable) method:
 * Increment ActiveCnt *or* Branch to replugged method
 */
/* This macro assumes that callee saved registers are not clobbered.
 * This ought to be true as the compiler is conservative around __asm.
 *
 * It is not recommended to put much code before the macro call.
 * BUGBUG: Unfotunately the macro can't contain a brace as the
 * compiler's parser doesn't like it.
 */
#define MUTABLE_ENTRY(_this_,_active_,_v_,_method_,_function_) \
{ __asm { \
    __asm mov eax, _this_ \
    __asm lock inc DWORD PTR [eax._active_] \
    __asm mov ecx, [eax._v_] \
    __asm cmp [ecx._method_], OFFSET _function_ \
    __asm je SHORT $goahead \
    __asm lock dec DWORD PTR [eax._active_] \
    __asm leave \
    __asm jmp [ecx._method_] \
    __asm $goahead: \
}}

/* Same as above but with default names: less typing */
#define MUTABLE_ENTRY2(_method_,_function_) \
    __asm mov eax, pThis \
    __asm lock inc DWORD PTR [eax.ActiveCnt] \
    __asm mov ecx, [eax] \
    __asm cmp [ecx._method_], OFFSET _function_ \
    __asm je SHORT $goahead \
    __asm lock dec DWORD PTR [eax.ActiveCnt] \
    __asm leave \
    __asm jmp [ecx._method_] \
    __asm $goahead:


/* Exit code for a mutable (repluggable) method:  Decrement ActiveCnt */
#define MUTABLE_EXIT(_this_,_active_,_rv_) \
{ __asm { \
    __asm mov ecx, _this_ \
    __asm lock dec DWORD PTR [ecx._active_] \
}} \
return _rv_;}


SCODE Test3FileReadAt(PIFILE pThis, UINT64 Position, 
                         BYTE *Buffer, UINT ByteCount, PUINT pSizeRead)
{
    SCODE sc;
    MUTABLE_ENTRY(pThis,ActiveCnt,TestFileV,ReadAt,Test3FileReadAt);
//    MUTABLE_ENTRY2(ReadAt,Test3FileReadAt);
    {

    UINT BytesRead, CurOffset;
    PTESTFILE File = (PTESTFILE) pThis;
    CurOffset = (UINT) (*((PINT64) &Position));


    DBGOUT(DBGFS, 0, printf("TestFileReadAt(%x, %x, %x, %x)\n", 
           File, CurOffset, Buffer, ByteCount));

    if (CurOffset >= File->Size) {
        if (CurOffset != File->Size)
            return E_INVALID_PARAMETER;

        BytesRead = 0;  /* Read at EOF is not an error */
        goto Exit;
    }

    if (CurOffset + ByteCount > File->Size)
        ByteCount = File->Size - CurOffset;

    MemCpy(Buffer, File->pData + CurOffset, ByteCount);

    BytesRead = ByteCount;
Exit:
    if (pSizeRead != NULL)
        *pSizeRead = BytesRead;
    sc = (CurOffset == File->Size) ? S_FALSE : S_OK;

    MUTABLE_EXIT(pThis,ActiveCnt,sc);
}

static const struct IFileVtbl Test3FileVtbl = {
    NULL, 
    NULL, 
    NULL, 
    Test3FileReadAt, 
    NULL, 
    NULL, 
    NULL, 
    NULL
};

/* This holding function spins by jumping to the current method repeatedly.
 * While it spins, it will come back to itself.  XXX sleep on mutex instead?
 */
#define HOLDINGFUNC(_n_) \
void __declspec(naked) HoldingFunc ## _n_ (PTR pInterface) \
{ \
    __asm mov eax, pInterface \
    __asm mov eax, [eax] \
    __asm jmp DWORD PTR [eax + 4 * _n_ ] \
}

/* Assume no object has more than 20 methods.  Too small?  What would do? */
HOLDINGFUNC(0)
HOLDINGFUNC(1)
HOLDINGFUNC(2)
HOLDINGFUNC(3)
HOLDINGFUNC(4)
HOLDINGFUNC(5)
HOLDINGFUNC(6)
HOLDINGFUNC(7)
HOLDINGFUNC(8)
HOLDINGFUNC(9)
HOLDINGFUNC(10)
HOLDINGFUNC(11)
HOLDINGFUNC(12)
HOLDINGFUNC(13)
HOLDINGFUNC(14)
HOLDINGFUNC(15)
HOLDINGFUNC(16)
HOLDINGFUNC(17)
HOLDINGFUNC(18)
HOLDINGFUNC(19)

const struct {
    void (*a)(PTR pInterface);
    void (*b)(PTR pInterface);
    void (*c)(PTR pInterface);
    void (*d)(PTR pInterface);
    void (*e)(PTR pInterface);
    void (*f)(PTR pInterface);
    void (*g)(PTR pInterface);
    void (*h)(PTR pInterface);
    void (*i)(PTR pInterface);
    void (*j)(PTR pInterface);
    void (*k)(PTR pInterface);
    void (*l)(PTR pInterface);
    void (*m)(PTR pInterface);
    void (*n)(PTR pInterface);
    void (*o)(PTR pInterface);
    void (*p)(PTR pInterface);
    void (*q)(PTR pInterface);
    void (*r)(PTR pInterface);
    void (*s)(PTR pInterface);
    void (*t)(PTR pInterface);
} HoldingV = {HoldingFunc0, HoldingFunc1, HoldingFunc2, HoldingFunc3,
                HoldingFunc4, HoldingFunc5, HoldingFunc6, HoldingFunc7,
                HoldingFunc8, HoldingFunc9, HoldingFunc10, HoldingFunc11,
                HoldingFunc12, HoldingFunc13, HoldingFunc14, HoldingFunc15,
                HoldingFunc16, HoldingFunc17, HoldingFunc18, HoldingFunc19
          };

void Mutator(PTESTFILE pObj)
{
    /*
       1 check activecnt (optional)
       2 set v to holding
       3 spin on active
       
       4 mutate

       5 set v to new class
       6 wake up holding
     */

    pObj->TestFileV = (PTR) &HoldingV;
    while (pObj->ActiveCnt != 0)
      ;

    /* ... */

    pObj->TestFileV = &Test3FileVtbl;
}

























#if 0
#pragma warning(disable:4035)
_declspec(naked) SCODE TestFileReadAt(PIFILE pThis, UINT64 Position, 
                         BYTE *Buffer, UINT ByteCount, PUINT pSizeRead)
{
    UINT BytesRead, CurOffset;
    PTESTFILE File;
    SCODE sc;

    __asm {
        mov ecx, [esp + 4]
        lock inc DWORD PTR [ecx + 8];

        push ebp;
        mov ebp, esp;
        sub esp, __LOCAL_SIZE;
        push ebx;
        push esi;
        push edi;
    }

    File = (PTESTFILE) pThis;
    CurOffset = (UINT) (*((PINT64) &Position));



    if (CurOffset + ByteCount > File->Size)
        ByteCount = File->Size - CurOffset;

    MemCpy(Buffer, File->pData + CurOffset, ByteCount);

    BytesRead = ByteCount;
//Exit:
    if (pSizeRead != NULL)
        *pSizeRead = BytesRead;
    sc = (CurOffset == File->Size) ? S_FALSE : S_OK;

    __asm {
        pop edi;
        pop esi;
        pop ebx;
          
        mov ecx, pThis
        lock dec DWORD PTR [ecx + 8]
        mov eax, sc;

        leave
        ret
    }
}


SCODE Test2FileReadAt(PIFILE pThis, UINT64 Position, 
                         BYTE *Buffer, UINT ByteCount, PUINT pSizeRead)
{
    UINT BytesRead, CurOffset;
    PTESTFILE File;
    register SCODE sc;

    __asm mov eax, pThis;
    __asm lock inc DWORD PTR [eax.ActiveCnt];

    __asm mov ecx, [eax.TestFileV];
    __asm cmp [ecx.ReadAt], OFFSET Test2FileReadAt;
    __asm je SHORT $goahead;

    __asm lock dec DWORD PTR [eax.ActiveCnt];
    __asm leave
    __asm jmp [ecx.ReadAt];
    __asm $goahead:

    File = (PTESTFILE) pThis;
    CurOffset = (UINT) (*((PINT64) &Position));



    if (CurOffset + ByteCount > File->Size)
        ByteCount = File->Size - CurOffset;

    MemCpy(Buffer, File->pData + CurOffset, ByteCount);

    BytesRead = ByteCount;
//Exit:
    if (pSizeRead != NULL)
        *pSizeRead = BytesRead;
    sc = (CurOffset == File->Size) ? S_FALSE : S_OK;

    __asm mov ecx, pThis
    __asm lock dec DWORD PTR [ecx.ActiveCnt];

    return sc;
}
#endif
