/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define OPTIMIZE        /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <fred.h>
#include <base/schedul.h>
#include <schedulif.h>

/* Mutex is a concrete class (doesn't have a vtbl)
 */
typedef struct _MX { /* Internal definition */
    PTR pThdOwner;   /* Owner, or NULL if free. LSB set if there are waiters */
} MX, *PMX;

#if _UINTSIZE == 16
#define BAD_MUTEX 0xFFFF
#endif
#if _UINTSIZE == 32
#define BAD_MUTEX 0xFFFFBAD0
#endif
#if _UINTSIZE == 64
#define BAD_MUTEX 0xFFFFFFFFFFFFBAD0
#endif

/* pThdOwner points to the thread that owns the mutex.
 * The list of waiters hangs off of owner's thread structure (see thread.c).
 * If a thread owns more then one mutex the list contains all threads waiting
 * on all mutexes.  The LSB of pThdOwner is set if the list is not empty.  
 * That causes Mutex_Unlock to call the scheduler, which wakes up all waiters.
 *
 */
#define pMX(pMutex) ((PMX)(pMutex))

/* *** TakeMutex
 *
 * Try to acquire a mutex on behalf of the current thread.
 * Returns S_OK if mutex is taken, S_FALSE if already owned.
 * Scheduler must be locked.
 *
 */
PRIVATE SCODE TakeMutex(PMX pMX, PITHREAD pThdThis)
{
    ADDRESS pThdOwner;

    assert(!pTheScheduler->v->PreemptionIsEnabled(pTheScheduler));

    DBGOUT(DBGMUTEX,MXLVL,printf(
           "TakeMutex   ThdThis %08x MX %08x MX.Owner %08x ThdOwner %08x\n",
           pThdThis,pMX,pMX->pThdOwner,pThdOwner));

    /* MP BUGBUG On a multiprocessor, a spinlock protecting
     * the thread queue is needed. This spinlock should reside
     * in the owner thread object.
     */

    /* If mutex is (now) free then claim it and return success.
     */
    pThdOwner = (ADDRESS) pMX->pThdOwner & ~1;
    if (pThdOwner == 0) {
        pMX->pThdOwner = (PTR) pThdThis;
        return S_OK;
    }

    DBGOUT(DBGMUTEX,2,if (pThdThis == (PITHREAD) pThdOwner)
           printf("TakeMutex: deadlock!\n"
           "    ThdThis %08x MX %08x MX.Owner %08x\n",
           pThdThis,pMX,pMX->pThdOwner));

    assert(pThdThis != (PITHREAD)(PTR) pThdOwner);

    /* Set flag indicating presence of waiters.  Do this before the thread is
     * queued, to make sure copy on write, or auto commit faults restart
     * correctly (they shouldn't happen under normal usage of a mutex, but
     * test, or buggy code, may trigger this).
     */
    pMX->pThdOwner = (PTR) (pThdOwner | 1);

    /* Add this thread to the list of threads blocked by the thread
     * that owns this mutex.
     */
    pTheScheduler->v->WaitForMutex(pTheScheduler,
                                   pThdThis,
                                   (PMUTEX)pMX,
                                   (PITHREAD)pThdOwner);

    return S_FALSE;
}


/* *** UntakeMutex
 *
 * Release a mutex that has threads waiting for it.
 * Scheduler must be locked.
 *
 * Caller must be the mutex's real owner.
 */
PRIVATE void UntakeMutex(PMX pMX)
{
    ADDRESS pThdOwner;

    assert(!pTheScheduler->v->PreemptionIsEnabled(pTheScheduler));

    DBGOUT(DBGMUTEX,MXLVL,printf(
           "UntakeMutex MX %08x MX.Owner %08x Self %08x\n",
           pMX,pMX->pThdOwner,CurrentThread()));

    /* MP BUGBUG On a multiprocessor, a spinlock protecting
     * the thread queue is needed. This spinlock should reside
     * in the owner thread object.
     */

    pThdOwner = (ADDRESS) pMX->pThdOwner;
    pMX->pThdOwner = NULL;

    /* Recheck there are infact waiters
     */
    if (pThdOwner & 1) /* Wake all waiters */
        pTheScheduler->v->WakeupOnMutex(pTheScheduler,
                                        (PMUTEX)pMX,
                                        (PITHREAD)(PTR)(pThdOwner & ~1));
}


/* ======== Exported Interface Functions ======== 
 */

void Mutex_Init(PMUTEX pMutex)
{
    pMX(pMutex)->pThdOwner = NULL;
}

void Mutex_Destroy(PMUTEX pMutex)
{
    pMX(pMutex)->pThdOwner = (PTR) BAD_MUTEX;
}

BOOL Mutex_TryLock(PMUTEX pMutex)
{
    PITHREAD pIThread;

    pIThread = CurrentThread();

/* BUGBUG needs new AtomicCmpAndSwapAddress() */
/* BUGBUG needs new AtomicCmpAndSwapAddress() */
    return AtomicCmpAndSwap((PUINT) &pMX(pMutex)->pThdOwner,0,
                            (UINT) (PTR) pIThread);
}

void Mutex_Lock(PMUTEX pMutex)
{
    PITHREAD Self;
    SCODE sc;

    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler));

    /* Scheduler lock *might* have been taken already, e.g. from
     * within the condition code XXX NEVER HAPPENS XXX
     */
    Self = CurrentThread();

    do {
/* BUGBUG needs new AtomicCmpAndSwapAddress() */
/* BUGBUG needs new AtomicCmpAndSwapAddress() */
        if (AtomicCmpAndSwap((PUINT) &pMX(pMutex)->pThdOwner,0,
                             (UINT) /*(PTR)*/ Self))
            break;

        /* Mutex taken, take slow path */

#ifdef _DEBUG
        /* sanity */
        if ((UINT) pMX(pMutex)->pThdOwner == BAD_MUTEX) {
            printf("Mutex_Lock: thread x%x used a bad mutex %x !\n",
                   (UINT) Self, (UINT) pMutex);
        }
#endif
        /* Try to acquire a mutex on behalf of the current thread.
         * Returns S_OK if mutex is taken, S_FALSE if already owned.
         */
        pTheScheduler->v->DisablePreemption(pTheScheduler,TRUE);

        sc = TakeMutex(pMX(pMutex), Self);

        pTheScheduler->v->EnablePreemption(pTheScheduler,TRUE);

    } while (sc != S_OK);
}

void Mutex_Unlock(PMUTEX pMutex)
{
    PITHREAD Self;

    //assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler));

    /* Scheduler lock *might* have been taken already, e.g. from
     * within the condition code
     YYY DOESHAPPEN from WaitCond from Condition_Wait -> this is OK XXX NEVER HAPPENS XXX
     */
    Self = CurrentThread();

    /* Is anybody waiting */
/* BUGBUG needs new AtomicCmpAndSwapAddress() */
/* BUGBUG needs new AtomicCmpAndSwapAddress() */
    if (AtomicCmpAndSwap((PUINT) &pMX(pMutex)->pThdOwner,
                         (UINT)/*(PTR)*/ Self,0))
        return;

    /* Yes, take slower path */

#ifdef _DEBUG
    /* sanity */
    if ((((UINT) pMX(pMutex)->pThdOwner) & ~1) != (UINT)Self) {
        printf("Mutex_Unlock: thread %x used a bad mutex %x(%x) !\n",
               (UINT) Self, (UINT) pMutex, (UINT) pMX(pMutex)->pThdOwner);
    }
#endif


    pTheScheduler->v->DisablePreemption(pTheScheduler,TRUE);

    UntakeMutex(pMX(pMutex));

    pTheScheduler->v->EnablePreemption(pTheScheduler,TRUE);

}

/* Return TRUE if the current thread owns the mutex.
 */
BOOL Mutex_Locked(const MUTEX *pMutex)
{
    return (ADDRESS)(PTR)CurrentThread() == 
           ((ADDRESS)pMX(pMutex)->pThdOwner &~ 1);
}
