/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* C Source Code for NameSpace Interface
 */

#define OPTIMIZE        /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <util/list-double2.h>
#include <tchar.h>
#include <string.h>
#include <base/cons.h>

#define __WANTS_NAMERANGES 1
#define Round4(_a_) (__ROUND((_a_), 4))

/* Status codes returned by local function ScanNameEntries() */
#define FOUND_NAMED_OBJECT S_FALSE
#define EMPTY_NAME E_INVALID_PARAMETER
//      E_NAME_NOT_FOUND
#define NAME_RECURSED S_OK

/* An entry in the linked list of object-name/object-pointer
 * pairs within a namespace
 */
typedef struct _NAMEENTRY {
    PIUNKNOWN pObj;    /* pointer to object */
    NAME_SPACE_FLAGS Flags;     /* Flags associated with object */
    ListNodeDeclare(_NAMEENTRY, Link);
#define Link_off offsetof(NAMEENTRY, Link)
    UINT16 StrLen;     /* length of name string */
    _TCHAR Str[1];     /* first character of name string */
} NAMEENTRY;

/* If the entry is a range, this struct follows Str (rounded to word bdry) */
typedef struct _RANGEENTRY {
    ADDRESS Start;
    UINT Size;
    struct IID iid;
} RANGEENTRY;

/* This flag is set if it is a range */
#define NAME_SPACE_REGION NAME_SPACE_CREATE /* reuse value */
/* Can't be set by app */
#define STORE_FLAGS_MASK ~(NAME_SPACE_CREATE|NAME_SPACE_FAILIFEXIST)

/* A namespace containing a list of entries, each of which
 * associates a name with an object.  This is a COM object.
 */

typedef struct {
    const struct INameRangesVtbl *Vtbl; /* virtual function table */
    UINT Refs;              /* reference count */
    MUTEX Mutex;            /* protect namespace data */
    NAMEENTRY *NameListHead;
} NAMESPC, *PNAMESPC;

/* The v-table, forward declaration */
PRIVATE const struct INameRangesVtbl NameObjVtbl;

/* **** Constructor:  NameSpaceCreate()
 */
PINAMESPACE NameSpaceNew(const _TCHAR *Name)
{
    PNAMESPC pNmSpc;
    IHeap *pHeap = CurrentHeap();

    pNmSpc = (PNAMESPC) pHeap->v->Alloc(pHeap,
                                        HEAP_DEBUG_NAME(Name ? Name
                                                             : _T("rootns")),
                                        sizeof(NAMESPC), 0);
    if (!pNmSpc)
        return NULL;

    pNmSpc->Vtbl = &NameObjVtbl;
    pNmSpc->Refs = 1;
    Mutex_Init(&pNmSpc->Mutex);
    pNmSpc->NameListHead = NULL;

    return (PINAMESPACE)pNmSpc;
}

/* Add pName to namespace.
 * Consumes pObj ref.
 */
PRIVATE NAMEENTRY *AddEntry(PNAMESPC pNmSpc, const _TCHAR *pName, UINT Len,
                            PIUNKNOWN pObj, NAME_SPACE_FLAGS Flags)
{
    PIHEAP pHeap = CurrentHeap();
    NAMEENTRY *pEntry;
    UINT tmp;

    tmp = sizeof(NAMEENTRY)+Len*sizeof(_TCHAR);
#ifdef __WANTS_NAMERANGES
    if (Flags & NAME_SPACE_REGION)
        tmp = Round4(tmp) + sizeof(RANGEENTRY);
#endif
    pEntry = (NAMEENTRY *)pHeap->v->Alloc(pHeap, HEAP_DEBUG_NAME("nsentry"),
                                          tmp, 0);
    if (!pEntry)
        return NULL;

    /* Initialize the new entry and insert it into the namespace.
     */
    pObj->v->AddRef(pObj);      /* account for the stored pointer */
    pEntry->pObj = pObj;
    pEntry->Flags = Flags;
    pEntry->StrLen = (UINT16) Len;
    memcpy(pEntry->Str, pName, (Len+1)*sizeof(_TCHAR));   /* copy EOS too */
    ListAddLast(&pNmSpc->NameListHead, Link_off, pEntry);
    return pEntry;
}

/* Scan the namespace pointed to by pNmSpc for an entry whose name matches
 * the string pointed to by pStr.  If found, output a pointer to the entry
 * and also output the length of the string.
 */
PRIVATE SCODE ScanNameEntries(PNAMESPC pNmSpc, const _TCHAR * pStr,
                              REFIID Iid,
                              NAMEENTRY **ppEntryOut, UINT *pLenOut)
{
    UINT Len, Separator;
    NAMEENTRY *pEntry;

    /* Calculate length of this segment */
    Len = Separator = 0;
    if (pStr) {
        for ( ; pStr[Len] != _EOS; Len++)
            if (IS_PATH_SEPARATOR(pStr[Len])) {
                Separator++;
                break;
            }
    }

    if (Len == 0)
        return EMPTY_NAME;    /* name string is empty */

    *pLenOut = Len + Separator; /* output length of consumed name string */

    /* Ensure mutually exclusive access to namespace data structures.
     * Note that the caller is responsible for UNlocking the mutex.
     */
    Mutex_Lock(&pNmSpc->Mutex);

    /* The loop below searches the current namespace for an entry
     * whose name matches the substring from the pathname.
     */
    pEntry = pNmSpc->NameListHead;
    if (pEntry) do {
        /* Does this entry's name match the substring? */
        if (pEntry->StrLen == Len && !memcmp(pEntry->Str,
                                             pStr,
                                             Len * sizeof(_TCHAR))) 
        {
            /* Yes, we've found an entry with a name that matches.
             */
            if (pStr[Len] == _EOS) {
                /* No recursion, return it */
                *ppEntryOut = pEntry;    /* output pointer to matching entry */
                return FOUND_NAMED_OBJECT;
            } else {
                /* Recursion: common code */
                PINAMESPACE ns;
                SCODE sc;

                sc = pEntry->pObj->v->QueryInterface(pEntry->pObj,
                                                     Iid,
                                                     (void **)&ns);
                Mutex_Unlock(&pNmSpc->Mutex);
                if (FAILED(sc))
                    return sc;

                *ppEntryOut = (void *) ns;
                return NAME_RECURSED;
            }
        }
        pEntry = ListNodeNext(Link_off, pEntry);
    } while (pEntry != pNmSpc->NameListHead);

    /* No, could not find entry that matches specified name.
     */
    return E_NAME_NOT_FOUND;
}



/* **** INameSpace::Register()
 */
PRIVATE SCODE MCT NameObjRegister(PINAMERANGES pThis, const _TCHAR * pName,
       PIUNKNOWN pObject, NAME_SPACE_FLAGS Flags, PINAMESPACE pServerNameSpace)
{
    SCODE sc;
    UINT Len;
    NAMEENTRY *pEntry;
    PNAMESPC pNmSpc = (PNAMESPC) pThis;
    PIUNKNOWN pUnk;
    PINAMESPACE ns;

    UnusedParameter(pServerNameSpace);

    sc = ScanNameEntries(pNmSpc, pName, &IID_INameSpace, &pEntry, &Len);
    switch (sc) {

    case E_NAME_NOT_FOUND:
        /* None of the entries in the namespace contains the specified
         * object name.  That means we need to create a new entry to
         * hold the object's name and pointer.
         */

        /* Disallow recursive creates. */
        if (pName[Len] != _EOS) {
            break;
        }

        if (AddEntry(pNmSpc, pName, Len, pObject, Flags & STORE_FLAGS_MASK))
            sc = S_OK;
        else
            sc = E_NOT_ENOUGH_MEMORY;
        break;

    case NAME_RECURSED:
        ns = (void *) pEntry;
        sc = ns->v->Register(ns, &pName[Len], pObject, Flags,
                             pServerNameSpace);
        ns->v->Release(ns);
        return sc;
    case FOUND_NAMED_OBJECT:
        /* Found namespace entry that matches specified object name.
         * This was the last segment.
         */
        if (Flags & NAME_SPACE_FAILIFEXIST) {
            sc = E_ALREADY_EXISTS;
            break;
        }
        /* Replace the entry's old object pointer with the new one.
         */
        pUnk = pEntry->pObj;
        pUnk->v->Release(pUnk);
        pObject->v->AddRef(pObject);
        pEntry->pObj = pObject;
        sc = S_ALREADY_EXISTS;
        break;

    default: /* Recursion failed or empty name. Lock was dropped. */
        return sc;
    }

    /* Unlock namespace and return status code to caller.
     */
    Mutex_Unlock(&pNmSpc->Mutex);
    return sc;
}

/* **** INameSpace::Bind()
 */
PRIVATE SCODE MCT NameObjBind(PINAMERANGES pThis, const _TCHAR * pName,
                                   NAME_SPACE_FLAGS Flags, PIUNKNOWN *ppObject)
{
    SCODE sc;
    UINT Len;
    NAMEENTRY *pEntry;
    PNAMESPC pNmSpc = (PNAMESPC) pThis;
    PIUNKNOWN pObj;

    sc = ScanNameEntries(pNmSpc, pName, &IID_INameSpace, &pEntry, &Len);

    switch (sc) {

    case E_NAME_NOT_FOUND:
        /* The namespace contains no entry that matches the specified
         * object name.  Disallow recursive creates.
         */
        if (pName[Len] != _EOS)
            break;

        /* Should we create such an entry? */
        if (!(Flags & NAME_SPACE_CREATE))
            break; /* no, don't create entry */

        /* The caller wants to create an object and enter it into the
         * namespace.  We can create only a namespace object.  Allocate
         * storage for both the object and an entry to point to it.
         */
        pObj = (PIUNKNOWN) NameSpaceNew(pName);
        if (!pObj) {
            sc = E_NOT_ENOUGH_MEMORY;
        } else {
            if (AddEntry(pNmSpc, pName, Len, pObj, Flags & STORE_FLAGS_MASK)) {
                *ppObject = pObj;       /* output pointer to named object */
                sc = S_OK;
            } else {
                CurrentHeap()->v->Free(CurrentHeap(), 0, pObj);
                sc = E_NOT_ENOUGH_MEMORY;
            }
        }
        break;

    case FOUND_NAMED_OBJECT:
        /* Found existing entry that matches specified object name.
         */
        if (Flags & NAME_SPACE_FAILIFEXIST) {
            sc = E_ALREADY_EXISTS;
            break;
        }
        if (Flags & NAME_SPACE_CREATE) {

            /* Create a new namespace object to replace the old object
             * with the same name.
             */
            pObj = (PIUNKNOWN) NameSpaceNew(pName);
            if (!pObj) {
                sc = E_NOT_ENOUGH_MEMORY;
                break;                /* unable to create new object */
            }
            pEntry->pObj->v->Release(pEntry->pObj);   /* release old object */
            pEntry->pObj = pObj;
        } else {
            pObj = pEntry->pObj;
        }
        pObj->v->AddRef(pObj);  /* add caller's reference */
        *ppObject = pObj;       /* output pointer to named object */
        sc = S_OK;
        break;

    case NAME_RECURSED: {
        PINAMESPACE ns = (void *) pEntry;
        sc = ns->v->Bind(ns, &pName[Len], Flags, ppObject);
        ns->v->Release(ns);
        return sc;
    }
    case EMPTY_NAME: /* empty, return self. Lock was dropped. */
        pThis->v->AddRef(pThis);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    default: /* Recursion failed. Lock was dropped. */
        return sc;
    }

    /* Unlock namespace and return status code to caller.
     */
    Mutex_Unlock(&pNmSpc->Mutex);
    return sc;
}

/* **** INameSpace::Unregister()
 */
PRIVATE SCODE MCT NameObjUnregister(PINAMERANGES pThis, const _TCHAR * pName)
{
    SCODE sc;
    UINT Len;
    NAMEENTRY *pEntry;
    PNAMESPC pNmSpc = (PNAMESPC) pThis;
    IHeap *pHeap;
    PIUNKNOWN pUnk;
    PINAMESPACE ns;

    sc = ScanNameEntries(pNmSpc, pName, &IID_INameSpace, &pEntry, &Len);

    switch (sc) {

    case E_NAME_NOT_FOUND:
        /* No namespace entry matching the object name was found.
         */
        break;

    case NAME_RECURSED:
        ns = (void *) pEntry;
        sc = ns->v->Unregister(ns, &pName[Len]);
        ns->v->Release(ns);
        return sc;

    case FOUND_NAMED_OBJECT:
        /* Found existing entry that has specified object name.
         */
        ListRemoveItem(&pNmSpc->NameListHead, Link_off, pEntry);
        /* Drop lock asap because pUnk.Release() might access the name space */
        Mutex_Unlock(&pNmSpc->Mutex);
        pUnk = pEntry->pObj;
        pUnk->v->Release(pUnk);    /* release namespace's ref */
        pHeap = CurrentHeap();
        (void)pHeap->v->Free(pHeap, 0, pEntry);
        return S_OK;
    default: /* Recursion failed or empty name. Lock was dropped. */
        return sc;
    }

    /* Unlock namespace and return status code to caller.
     */
    Mutex_Unlock(&pNmSpc->Mutex);
    return sc;
}

/* **** INameSpace::FindNext()
 */
PRIVATE SCODE MCT NameObjFindNext(PINAMERANGES pThis, const _TCHAR * pPrefix,
       const _TCHAR * pPrevious, _TCHAR * pBuffer, UINT BufSize, PUINT pStrLen)
{
    SCODE sc;
    UINT Len, Size;
    NAMEENTRY *pEntry;
    PNAMESPC pNmSpc = (PNAMESPC) pThis;

    /* If pPrefix is set, "cd" to it and do the rest there */
    if (pPrefix && pPrefix[0] != _EOS) {
        PIUNKNOWN pUnk;
        sc = pThis->v->Bind(pThis, pPrefix, NAME_SPACE_READ, &pUnk);
        if (FAILED(sc))
            return sc;
        sc = pUnk->v->QueryInterface(pUnk, &IID_INameSpace, (void **) &pThis);
        pUnk->v->Release(pUnk);
        if (FAILED(sc))
            return sc;
        return pThis->v->FindNext(pThis, NULL, pPrevious,
                                  pBuffer, BufSize, pStrLen);
    }

    /* Scan the entries in the designated namespace to find the entry
     * that contains the object name specified by pPrevious.
     */
    sc = ScanNameEntries(pNmSpc, pPrevious, &IID_INameSpace, &pEntry, &Len);

    switch (sc) {

    case E_NAME_NOT_FOUND:
        /* Unable to find entry containing specified object name.
         */
        break;

    case NAME_RECURSED: {
        PINAMESPACE ns = (void *) pEntry;
        /* Support recursive pPrevious:
         * Copy beginning of pPrevious to pBuffer.
         */
        if ((Len+2) * sizeof(_TCHAR) >= BufSize) {
            sc = E_BUFFER_TOO_SMALL;
        } else {
            _tcsncpy(pBuffer, pPrevious, Len);
            pBuffer[Len] = PATH_SEPARATOR1;
            pBuffer[Len+1] = _EOS;
            Len++; /* account for slash */
            sc = ns->v->FindNext(ns, pPrefix, &pPrevious[Len], &pBuffer[Len],
                                 BufSize - Len * sizeof(_TCHAR), pStrLen);
            if (pStrLen)
                *pStrLen += Len;
        }
        ns->v->Release(ns);
        return sc;
    }
    case EMPTY_NAME:
        /* Need to take lock now */
        Mutex_Lock(&pNmSpc->Mutex);

        /* pPrevious points to an empty string, in which case, a call to
         * FindNext() is equivalent to a call to FindFirst().  Just below,
         * pEntry is set to point to the head of the namespace's list of
         * entries.  pPtr will later be set to point to the first entry.
         */
        pEntry = pNmSpc->NameListHead;
        /* Fall through to "case FOUND_NAMED_OBJECT" below
         */
        if (!pEntry) {
            sc = S_NO_MORE_ENTRIES;
            if (pStrLen)
                *pStrLen = 0;
            break;
        }
        goto blahblah; /* ... but don't skip the first entry... */

    case FOUND_NAMED_OBJECT:
        /* Found existing entry that has specified object name.
         */
        pEntry = ListNodeNext(Link_off, pEntry);
        if (pEntry == pNmSpc->NameListHead) {
            sc = S_NO_MORE_ENTRIES;
            if (pStrLen)
                *pStrLen = 0;
            break;
        }
blahblah:
        /* We know the next entry is valid.  Make sure the caller's buffer
         * is big enough to hold the object name contained in the entry.
         */
        Len = pEntry->StrLen;
        if (pStrLen)
            *pStrLen = Len;             /* character count, excl. EOS */
        Size = (Len+1)*sizeof(_TCHAR);  /* size in bytes, incl. EOS */
        if (Size > BufSize) {
            sc = S_BUFFER_TOO_SMALL;
            break;
        }
        memcpy(pBuffer, pEntry->Str, Size);   /* copy EOS too */

        /* Is this the last name registered in the namespace?
         */
        if (ListNodeNext(Link_off, pEntry) != pNmSpc->NameListHead)
            sc = S_OK;                 /* more names after this one */
        else
            sc = S_NO_MORE_ENTRIES;
        break;
    default: /* Recursion failed. Lock was dropped. */
        return sc;
    }

    /* Unlock namespace and return status code to caller.
     */
    Mutex_Unlock(&pNmSpc->Mutex);
    return sc;
}

#ifdef __WANTS_NAMERANGES /* dunno if this method is needed by anything */
/* **** INameSpace::GetCapabilities()
 */
PRIVATE SCODE MCT NameObjGetCapabilities(PINAMERANGES pThis,
                                const _TCHAR * pName, NAME_SPACE_FLAGS *pFlags)
{
    SCODE sc;
    UINT Len;
    NAMEENTRY *pEntry;
    PNAMESPC pNmSpc = (PNAMESPC) pThis;
    PINAMESPACE ns;

    sc = ScanNameEntries(pNmSpc, pName, &IID_INameSpace, &pEntry, &Len);

    switch (sc) {

    case E_NAME_NOT_FOUND:
        /* No namespace entry matching the object name was found. */
        break;

    case NAME_RECURSED:
        ns = (void *) pEntry;
        sc = ns->v->GetCapabilities(ns, &pName[Len], pFlags);
        ns->v->Release(ns);
        return sc;

    case FOUND_NAMED_OBJECT:
        /* Found existing entry that has specified object name. */
        if (pFlags)
            *pFlags = pEntry->Flags; /* return argument */
        sc = S_OK;
        break;

    default: /* Recursion failed or empty name. Lock was dropped. */
        return sc;
    }

    /* Unlock namespace and return status code to caller.
     */
    Mutex_Unlock(&pNmSpc->Mutex);
    return (sc);
}
#endif

/* **** INameSpace::QueryInterface()
 */
PRIVATE SCODE MCT NameObjQueryInterface(PINAMERANGES pThis, REFIID pIid,
                                        void **ppUnk)
{
    SCODE sc;
    sc = GenericQueryInterface((IUnknown *)pThis,pIid,ppUnk,&IID_INameSpace);
#ifdef __WANTS_NAMERANGES
    if (sc == E_NO_INTERFACE) {
        sc = GenericQueryInterface((IUnknown *)pThis,pIid,ppUnk,
                                   &IID_INameRanges);
    }
#endif
    return sc;
}

/* **** INameSpace::AddRef()
 */
PRIVATE UINT MCT NameObjAddRef(PINAMERANGES pThis)
{
    return AtomicInc(&((PNAMESPC)pThis)->Refs);
}

/* **** INameSpace::Release()
 */
PRIVATE UINT MCT NameObjRelease(PINAMERANGES pThis)
{
    UINT RefCnt;
    PNAMESPC pNmSpc = (PNAMESPC) pThis;

    RefCnt = AtomicDec(&pNmSpc->Refs);
    if (RefCnt == 0) {
        NAMEENTRY *pEntry;
        IUnknown *pUnk;
        IHeap *pHeap = CurrentHeap();

    Next:
        /* Note: Zap entry before pObj.Release as cobmgr might re-enter here */
        pEntry = ListRemoveFirst(&pNmSpc->NameListHead, Link_off);
        if (pEntry) {
            /* Free the storage for the entry in the namespace.
             */
            pUnk = (IUnknown *)pEntry->pObj;
            pUnk->v->Release(pUnk);
            (void)pHeap->v->Free(pHeap, 0, pEntry);

            goto Next;
        }
        /* Free the storage for the namespace object itself.
         */
        (void)pHeap->v->Free(pHeap, 0, pNmSpc);
    }
    return RefCnt;
}

/*---------- CNameRanges ----------*/

#ifdef __WANTS_NAMERANGES
PRIVATE SCODE NameObjRegisterRange (
    /*in*/ PINAMERANGES pThis,
    /*in*/ const _TCHAR * pName,
    /*in*/ PIUNKNOWN pObj,
    /*in*/ NAME_SPACE_FLAGS Flags,
    /*in*/ ADDRESS Start,
    /*in*/ UINT Size,
    /*in*/ REFIID Iid)
{
    SCODE sc;
    UINT Len;
    NAMEENTRY *pEntry, *pOld = NULL;
    RANGEENTRY *re;
    PNAMESPC pNmSpc = (PNAMESPC) pThis;
    IHeap *pHeap;
    PIUNKNOWN pUnk;
    PINAMERANGES ns;

    sc = ScanNameEntries(pNmSpc, pName, &IID_INameRanges, &pEntry, &Len);
    switch (sc) {

    case E_NAME_NOT_FOUND:
        /* None of the entries in the namespace contains the specified
         * object name.  That means we need to create a new entry to
         * hold the object's name and pointer.
         */

        /* Disallow recursive creates of intermediate path components. */
        if (pName[Len] != _EOS)
            break;

CreateIt:
        pEntry = AddEntry(pNmSpc, pName, Len, pObj,
                          (Flags & STORE_FLAGS_MASK) | NAME_SPACE_REGION);
        if (!pEntry) {
            sc = E_NOT_ENOUGH_MEMORY;
            break;
        }
        /* Initialize the range part of the entry */
        re = (PTR)((PUINT8) pEntry
                   + Round4(sizeof(NAMEENTRY)+Len*sizeof(_TCHAR)));
        re->Start = Start;
        re->Size = Size;
        re->iid = *Iid;
        sc = S_OK;
        break;

    case NAME_RECURSED:
        ns = (void *) pEntry;
        sc = ns->v->RegisterRange(ns, &pName[Len], pObj, Flags,
                                  Start, Size, Iid);
        ns->v->Release(ns);
        return sc;

    case FOUND_NAMED_OBJECT:
        /* Found namespace entry that matches specified object name.
         */
        if (Flags & NAME_SPACE_FAILIFEXIST) {
            sc = E_ALREADY_EXISTS;
            break;
        }
        /* Delete old entry and create a new one.
         */
        ListRemoveItem(&pNmSpc->NameListHead, Link_off, pEntry);
        pOld = pEntry;
        sc = S_ALREADY_EXISTS;
        goto CreateIt;

    default: /* Recursion failed or empty name. Lock was dropped. */
        return sc;
    }

    /* Unlock namespace and return status code to caller.
     */
    Mutex_Unlock(&pNmSpc->Mutex);

    if (pOld != NULL) {
        /* Get rid of old entry if any */
        pUnk = pOld->pObj;
        pUnk->v->Release(pUnk);
        pHeap = CurrentHeap();
        (void)pHeap->v->Free(pHeap, 0, pOld);
    }
    return sc;
}

PRIVATE SCODE NameObjLookupRange (
    /*in*/ PINAMERANGES pThis,
    /*in*/ const _TCHAR * pName,
    /*in*/ REFIID iid,
    /*out*/ ADDRESS * pStart,
    /*out*/ PUINT pSize)
{
    PNAMESPC x = (PNAMESPC) pThis;

    SCODE sc;
    UINT Len;
    NAMEENTRY *pEntry;
    PINAMERANGES ns;

    sc = ScanNameEntries(x, pName, &IID_INameRanges, &pEntry, &Len);

    switch (sc) {

    case E_NAME_NOT_FOUND:
        /* No namespace entry matching the object name was found.
         */
        break;

    case NAME_RECURSED:
        sc = pEntry->pObj->v->QueryInterface(pEntry->pObj,
                                             &IID_INameRanges,
                                             (void **)&ns);
        Mutex_Unlock(&x->Mutex);
        if (FAILED(sc))
            return sc;
        sc = ns->v->LookupRange(ns, &pName[Len], iid, pStart, pSize);
        ns->v->Release(ns);
        return sc;

    case FOUND_NAMED_OBJECT:
        /* Found existing entry that has specified object name.
         */
        if (pEntry->Flags & NAME_SPACE_REGION) {
            UINT tmp;
            RANGEENTRY *re;

            tmp = sizeof(NAMEENTRY) + Len * sizeof(_TCHAR);
            tmp = Round4(tmp);
            re = (PTR)((PUINT8) pEntry + tmp); /* the RANGEENTRY is here */
            *pStart = re->Start;
            *pSize = re->Size;
            if (UuidCmp(iid, &re->iid))
                sc = S_OK;
            else
                sc = E_NO_INTERFACE;
        } else {
            /* This entry is not a region */
            sc = E_OBJECT_NOT_INITIALIZED; /* XXX should have better code */
        }
        break;

    default: /* Recursion failed or empty name. Lock was dropped. */
        return sc;
    }

    /* Unlock namespace and return status code to caller.
     */
    Mutex_Unlock(&x->Mutex);
    return sc;
}

/* Scan the namespace pointed to by pNmSpc for an entry whose name matches
 * the string pointed to by pStr.  If found, return the object (if any).
 */

PRIVATE SCODE NameObjBindRange (
    /*in*/ PINAMERANGES pThis,
    /*in*/ ADDRESS Address,
    /*out*/ PIUNKNOWN *ppObj)
{
    PNAMESPC pNmSpc = (PNAMESPC) pThis;
    PIUNKNOWN pObj;
    NAMEENTRY *pEntry;
    RANGEENTRY *re;
    UINT tmp;

    Mutex_Lock(&pNmSpc->Mutex);

    pEntry = pNmSpc->NameListHead;
    if (pEntry) do {
        /* If this is not a range at all, skip it */
        if ((pEntry->Flags & NAME_SPACE_REGION) == 0)
            continue;

        tmp = sizeof(NAMEENTRY) + pEntry->StrLen * sizeof(_TCHAR);
        tmp = Round4(tmp);
        re = (PTR)((PUINT8) pEntry + tmp); /* the RANGEENTRY is here */

        /* Is the address inside the range? */
        if (Address <= re->Start)
            continue;           /* no, the range is later */
        if (Address > re->Start + re->Size)
            continue;           /* no the range ends before */
        /* Yes, we've found an entry that matches.
         */
        pObj = pEntry->pObj;
        pObj->v->AddRef(pObj);
        Mutex_Unlock(&pNmSpc->Mutex);
        *ppObj = pObj;
        return S_OK;

        pEntry = ListNodeNext(Link_off, pEntry);
    } while (pEntry != pNmSpc->NameListHead);

    /* No, could not find entry that matches specified name.
     */

    Mutex_Unlock(&pNmSpc->Mutex);
    return E_NAME_NOT_FOUND;
}
#else

PRIVATE SCODE Unimplemented()
{
    return E_NOT_IMPLEMENTED;
}
#endif /* __WANTS_NAMERANGES */

/* Virtual function table for namespace object
 */
PRIVATE const struct INameRangesVtbl NameObjVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NameObjQueryInterface,
    NameObjAddRef,
    NameObjRelease,
    NameObjRegister,
    NameObjUnregister,
    NameObjBind,
    NameObjFindNext,
#ifdef __WANTS_NAMERANGES
    NameObjGetCapabilities,

    NameObjRegisterRange,
    NameObjLookupRange,
    NameObjBindRange
#else
    (_INameRanges_GetCapabilities_METHOD_TYPE) Unimplemented,
    (_INameRanges_RegisterRange_METHOD_TYPE) Unimplemented,
    (_INameRanges_LookupRange_METHOD_TYPE) Unimplemented,
    (_INameRanges_BindRange_METHOD_TYPE) Unimplemented
#endif /* __WANTS_NAMERANGES */
};
