/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * This file implements two abstractions in one, the IProcess
 * and the IModule.  This is in order to keep things simpler
 * and save on dynamic data structures.
 */

#define OPTIMIZE 1  /* enable Base.h optimizations, since we're the "base" */
#define DEATH_WATCH 0

#include <mmlite.h>
#include <mmmacros.h>
#include <fred.h>
#include <base/cons.h>
#include "process.h"
#include <base/schedul.h>
#include <schedulif.h>
#include <base/debugger.h>

/* Imports
 */
extern struct IBaseRtlVtbl BaseRtlVtbl;

/* Store count of waiters in high half of RefCnt */
#define WAITER_SHIFT (_UINTSIZE/2)
#define WAITER_MASK ((1<<WAITER_SHIFT)-1)

/* Local globals
 */
#if MODLIST
PRIVATE PMODULE ModuleList = NULL;
#else
PRIVATE PLISTNODE ModuleList = NULL;
#endif
PRIVATE MUTEX ModuleListLock = MUTEX_INITIALIZER;
PRIVATE PROCESS TheProcess;

/* ++++++++++++++++++++++++++++++++++++++++++
 * IModule implementation
 */

PRIVATE SCODE MCT ModuleQueryInterface ( 
            IModule *pThis,
            /* [in] */ REFIID pIid,
            /* [out] */ void **ppObject)
{
    SCODE sc;

    sc = GenericQueryInterface((IUnknown *)pThis,pIid,ppObject,&IID_IModule);

    /* Special handling for IProcess interface if not recognized already
     */
    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IProcess, pIid)) {
        pThis->v->AddRef(pThis);
        *ppObject = (void*) &pMD(pThis)->pv;
        sc = S_OK;
    }

    return sc;
}
        
PRIVATE UINT MCT ModuleAddRef ( 
            IModule *pThis)
{
    UINT RefCnt;

    RefCnt = AtomicInc(&pMD(pThis)->RefCnt);

    return RefCnt;
}

#if _DEBUG && !defined(_UNICODE)
#ifdef MODLIST
PRIVATE void plist(char *b)
{
    PMODULE pMod;

    printf("ModList-%s",b);
    pMod = ModuleList;
    if (pMod) {
        do {
            _tprintf(_T(" %s"),pMod->pName);

            pMod = ListNodeNext(Mod_off, pMod);

        } while (pMod != ModuleList);
    }
    printf("\n");
}
#else
PRIVATE void plist(char *b)
{
    PMODULE pMod;
    PLISTNODE pNext;

    printf("ModList-%s",b);
    pNext = ModuleList;
    if (pNext) {
        do {
            pMod = GetModuleFrom(pNext,Link);

            _tprintf(_T(" %s"),pMod->pName);

            pNext = pNext->pNext;

        } while (pNext != ModuleList);
    }
    printf("\n");
}
#endif
#else
#define plist(x)
#endif

PRIVATE UINT MCT ModuleRelease ( 
            IModule *pThis)
{
    UINT RefCnt;
    PMODULE pMod = pMD(pThis);

    RefCnt = AtomicDec(&pMod->RefCnt);
    assert((INT) RefCnt >= 0);

    if (RefCnt == 0) {
        PIHEAP pIHeap = CurrentHeap();

        assert(pMod != &TheProcess);

        /* Take module out of list of all modules
         */
        Mutex_Lock(&ModuleListLock);
        plist("release-before");
        RefCnt = pMod->RefCnt;
        if (RefCnt != 0) {
            /* Cover for race with LookupModule:
             * Somebody looked it up before we got the lock.
             */
            Mutex_Unlock(&ModuleListLock);
            return RefCnt;
        }
        /* Lets make sure the list stays ordered */
#ifdef MODLIST
        if (ModuleList == pMod)
            ModuleList = ListNodeNext(Mod_off, pMod);
#else
        if (ModuleList == &pMod->Link)
            ModuleList = pMod->Link.pNext;
#endif
        (void) ListNodeRemove(Mod_off, pMod);
        plist("release-after");
        Mutex_Unlock(&ModuleListLock);

#if PRINT_IDLE_TIME
        if (pTheScheduler) {
            TIME IdleIs = pTheScheduler->v->GetIdleTime(pTheScheduler);
            TIME TimeIs = CurrentTime();
            printf("  -- module %s Idle %lld of %lld elapsed\n", pMod->pName, 
                   Int64Subtract(IdleIs,pMod->IdleWas),
                   Int64Subtract(TimeIs,pMod->TimeWas));
        }
#endif

        /* Notify the debugger
         */
        DebuggerNotifyUnload(pMod->pName,pMod->Start,pMod->DebugToken);
        /* Release all memory resources
         */
#ifdef notyet
        for (i = 0; i < pMod->nImportedModules; i++)
            pMod->ImportedModules[i]->v->Release(pMod->ImportedModules[i]);
        pMod->nImportedModules = 0;
#endif

#if DEATH_WATCH
        printf("Module %x dead\n", pMod);
#endif
        Condition_Destroy(&pMod->_Termination);
        RELEASE(pMod->pNameSpace);
        if (pMod->Start)
            pIHeap->v->Free(pIHeap,0,(PTR) pMod->Start);
        pIHeap->v->Free(pIHeap,0,pThis);
    } else if ((RefCnt & WAITER_MASK) == (RefCnt >> WAITER_SHIFT)) {
        /* Someone is waiting for this process to terminate. */

        /* Ok to do one at a time: the next guy will Release and come here */
        Mutex_Lock(pMod->pTerminationMx);
        /* Barrier to keep us from getting into the read-test-block window
         * in ModuleWaitFor. This avoids the short race opportunity.
         */
        Mutex_Unlock(pMod->pTerminationMx);
        Condition_Signal(pMod->pTermination);
    }



    return (RefCnt & WAITER_MASK);
}

SCODE MCT PeGetFunctionAddress (
        PTR ExportData,
        UINT ExportSize,
        const _TCHAR * ModName,
        UINT Ordinal,
        /* [out] */ ADDRESS *pAddress);

PRIVATE SCODE MCT ModuleGetFunctionAddress ( 
            IModule *pThis,
            const _TCHAR * Name,
            UINT Ordinal,
            /* [out] */ ADDRESS *pAddress)
{
#ifdef i386
    if (pMD(pThis)->Exports != (ADDRESS *) &BaseRtlVtbl)
      return PeGetFunctionAddress(pMD(pThis)->Exports,
                                  pMD(pThis)->ExportsCount,
                                  Name,
                                  Ordinal,
                                  pAddress);
#endif
       
    /* BUGBUG: This is the simplest possible implementation that only
     * works for zero based ordinals.
     */
    if (Name != NULL)
      return E_INVALID_PARAMETER; /* not yet supported */

    if (Ordinal > pMD(pThis)->ExportsCount)
      return E_INVALID_PARAMETER;

    *pAddress = pMD(pThis)->Exports[Ordinal];

    return NO_ERROR;

}
        
PRIVATE SCODE MCT ModuleGetName ( 
            IModule *pThis,
            /* [size_is][out] */ _TCHAR *Buffer,
            /* [in] */ UINT BufSize)
{
    if (Buffer)
        _tcsncpy(Buffer,pMD(pThis)->pName,BufSize);
    return NO_ERROR;
}
        
PRIVATE SCODE MCT ModuleGetArgs ( 
            IModule *pThis,
            /* [size_is][out] */ _TCHAR * Buffer,
            /* [in] */ UINT BufSize)
{
    if (Buffer)
        _tcsncpy(Buffer,pMD(pThis)->pArgs,BufSize);
    return NO_ERROR;
}
        
PRIVATE SCODE MCT ModuleGetEntryPoint ( 
            IModule *pThis,
            /* [out] */ ADDRESS *pEntry)
{
    if (pEntry)
        *pEntry = (ADDRESS)pMD(pThis)->Entry;
    return NO_ERROR;
}

#if 0
PRIVATE SCODE MCT ModuleRelocateRva ( 
            IModule *pThis,
            ADDRESS Rva,
            /* [out] */ ADDRESS *pAddress,
            /* [out] */ UINT *pSize)
{
    /* XXX Naive and incorrect */
    *pAddress = (ADDRESS)pMD(pThis)->Start + Rva;

    if (pSize)
        *pSize = pMD(pThis)->End - pMD(pThis)->Start;

    return S_OK;
}
#endif

PRIVATE UINT MCT PrcSetValue (
            PIPROCESS iThis,
            UINT Index,
            UINT Value)
{
    PMODULE This = pPR(iThis);
    
    switch (Index) {
    case PROCESS_VALUE_EXIT_CODE:
        return AtomicSwap(&This->ExitCode, Value);
    case PROCESS_VALUE_STDIN:
        /* If there already was a value, don't do it but return new value */
        if (AtomicCmpAndSwap(&This->io0, 0, Value))
            return 0;
        else
            return Value;
    case PROCESS_VALUE_STDOUT:
        if (AtomicCmpAndSwap(&This->io1, 0, Value))
            return 0;
        else
            return Value;
    case PROCESS_VALUE_STDERR:
        if (AtomicCmpAndSwap(&This->io2, 0, Value))
            return 0;
        else
            return Value;
    default:
        return 0;
    }
}

PRIVATE UINT MCT PrcGetValue (
            PIPROCESS iThis,
            UINT Index)
{
    PMODULE This = pPR(iThis);
    
    switch (Index) {
    case PROCESS_VALUE_EXIT_CODE:
        return This->ExitCode;
    case PROCESS_VALUE_STDIN:
        return This->io0;
    case PROCESS_VALUE_STDOUT:
        return This->io1;
    case PROCESS_VALUE_STDERR:
        return This->io2;
    default:
        return 0;
    }
}

/* We only wait for Modules to unload.
 * Which implies processes.
 */
PRIVATE SCODE MCT PrcWaitFor(PIPROCESS pThis, TIME timeout)
{
    PMODULE pMod = pPR(pThis);
    UINT Refs;
    SCODE sc = S_OK;

    Refs = AtomicAdd(&pMod->RefCnt, 1 << WAITER_SHIFT);
    for (;;) {
        /* If all the refs are waiters then stop waiting */
        if ((Refs & WAITER_MASK) == (Refs >> WAITER_SHIFT)) {
            AtomicSub(&pMod->RefCnt, 1 << WAITER_SHIFT);
            return S_OK;
        }
        assert((Refs & WAITER_MASK) > (Refs >> WAITER_SHIFT)); /* Sanity */
        /* No lock needed above: only waiters had refs. */
        /* Cover for race: another ref was dropped and the signal gets lost */
        Mutex_Lock(pMod->pTerminationMx);
        Refs = pMod->RefCnt;
        if ((Refs & WAITER_MASK) != (Refs >> WAITER_SHIFT)) {
            sc = Condition_Wait(pMod->pTermination, pMod->pTerminationMx,
                                &timeout);
        }
        Mutex_Unlock(pMod->pTerminationMx);
        if (FAILED(sc)) {
            AtomicSub(&pMod->RefCnt, 1 << WAITER_SHIFT);
            return sc;          /* timed out */
        }
        Refs = pMod->RefCnt;
    }
}

/* This one was defined as an RTL. tzk tzk tzk
 */
#ifdef MODLIST
SCODE GetModuleContainsAddr(ADDRESS Addr, PIMODULE *ppIModule)
{
    PMODULE pMod;

    if (ppIModule)
        *ppIModule = NULL; /* allow sloppy usage */

    Mutex_Lock(&ModuleListLock);

    pMod = ModuleList;
    if (pMod) {

        /* Simple list scan, we dont expect too many modules at once
         */
        do {
            if (pMod->Start <= Addr &&
                Addr < pMod->End)
                goto Found;
            pMod = ListNodeNext(Mod_off, pMod);

        } while (pMod != ModuleList);

        /* Mumble..whats he looking for ?
         */
    }
    /* If nothing found, return the Base. Better than NULL usually... */
    pMod = &TheProcess;

 Found:
    if (ppIModule && pMod) {
        pMod->v->AddRef((PIMODULE)pMod);
        *ppIModule = (PIMODULE)pMod;
    }

    Mutex_Unlock(&ModuleListLock);

    return (pMod) ? NO_ERROR : E_OBJECT_NOT_FOUND;
}
#else
SCODE GetModuleContainsAddr(ADDRESS Addr, PIMODULE *ppIModule)
{
    PMODULE pMod;
    PLISTNODE pNext;

    if (ppIModule)
        *ppIModule = NULL; /* allow sloppy usage */

    Mutex_Lock(&ModuleListLock);

    pNext = ModuleList;
    if (pNext) {

        /* Simple list scan, we dont expect too many modules at once
         */
        do {
            pMod = GetModuleFrom(pNext,Link);

            if (pMod->Start <= Addr &&
                Addr < pMod->End)
                goto Found;
            pNext = pNext->pNext;

        } while (pNext != ModuleList);

        /* Mumble..whats he looking for ?
         */
    }
    /* If nothing found, return the Base. Better than NULL usually... */
    pMod = &TheProcess;

 Found:
    if (ppIModule && pMod) {
        pMod->v->AddRef((PIMODULE)pMod);
        *ppIModule = (PIMODULE)pMod;
    }

    Mutex_Unlock(&ModuleListLock);

    return (pMod) ? NO_ERROR : E_OBJECT_NOT_FOUND;
}
#endif

/* Get module objects one at a time.  Return the module that follows the
 * one specified by Previous.  If Previous is NULL return the first entry.
 * Return S_OK if found a module, S_FALSE if Previous was the last one, or
 * an error code.   Specifically return E_MOD_NOT_FOUND if Previous not valid,
 * or if there are no loaded modules at all. [not implem for mmlite]
 * This function takes a reference on the object it returns, if any.
 */
#ifdef MODLIST
SCODE GetModuleNext(PIMODULE Previous, PIMODULE *ppIModule)
{
    PMODULE pMod;
    SCODE sc = S_OK;

    Mutex_Lock(&ModuleListLock);

    /* Note that the list is never empty because the kernel does not go away.
     */

    if (Previous) {
        /* Wants the next, if any.
         * Note: since the module list is insert-at-front, any new
         * module added after this function was first called (with
         * Previous==NULL) will be found last.
         */
        pMod = ListNodeNext(Mod_off, Previous);

        /* Check for end of list
         */
        if (pMod == ModuleList) {
            pMod = NULL;
            sc = S_FALSE;
        }
    } else {
        pMod = ModuleList;
    }

    if (pMod)
        (void) pMod->v->AddRef(iMD(pMod));

    Mutex_Unlock(&ModuleListLock);

    *ppIModule = iMD(pMod);
    return sc;
}
#else
SCODE GetModuleNext(PIMODULE Previous, PIMODULE *ppIModule)
{
    PIMODULE pIMod;
    PLISTNODE pNext;
    SCODE sc = S_OK;

    Mutex_Lock(&ModuleListLock);

    /* Note that the list is never empty because the kernel does not go away.
     */
    pNext = ModuleList;

    if (Previous != NULL) {

        /* Wants the next, if any.
         * Note: since the module list is insert-at-front, any new
         * module added after this function was first called (with
         * Previous==NULL) will be found last.
         */
        pNext = ((PMODULE)Previous)->Link.pNext;

        /* Check for end of list
         */
        if (pNext == ModuleList) {
            pNext = NULL;
            sc = S_FALSE;
        }
    }

    pIMod = NULL;
    if (pNext) {
        pIMod = (PIMODULE) GetModuleFrom(pNext,Link);
        (void) pIMod->v->AddRef(pIMod);
    }

    Mutex_Unlock(&ModuleListLock);

    *ppIModule = pIMod;
    return sc;
}
#endif

/* Find module by name */
#ifdef MODLIST
PIMODULE LookupModule(const _TCHAR *pModName, UINT VersionCheck)
{
    PMODULE pMod;

    /* Time stamp or crc or whatever the linker gives to check versions */
    UnusedParameter(VersionCheck);

    Mutex_Lock(&ModuleListLock);

    pMod = ModuleList;
    if (pMod) {

        /* Simple list scan, we dont expect too many modules at once
         */
        do {
            if (!_tcsicmp(pMod->pName, pModName))
                goto Found;
            pMod = ListNodeNext(Mod_off, pMod);

        } while (pMod != ModuleList);

        /* Mumble..whats he looking for ?
         */
    }
    pMod = NULL;
 Found:
    if (pMod) {
        pMod->v->AddRef((PIMODULE)pMod);
    }

    Mutex_Unlock(&ModuleListLock);
    return (PIMODULE) pMod;
}
#else
PIMODULE LookupModule(const _TCHAR *pModName, UINT VersionCheck)
{
    PMODULE pMod;
    PLISTNODE pNext;

    /* Time stamp or crc or whatever the linker gives to check versions */
    UnusedParameter(VersionCheck);

    Mutex_Lock(&ModuleListLock);

    pNext = ModuleList;
    if (pNext) {

        /* Simple list scan, we dont expect too many modules at once
         */
        do {
            pMod = GetModuleFrom(pNext,Link);

            if (!_tcsicmp(pMod->pName, pModName))
                goto Found;
            pNext = pNext->pNext;

        } while (pNext != ModuleList);

        /* Mumble..whats he looking for ?
         */
    }
    pMod = NULL;
 Found:
    if (pMod) {
        pMod->v->AddRef((PIMODULE)pMod);
    }

    Mutex_Unlock(&ModuleListLock);
    return (PIMODULE) pMod;
}
#endif

/* The debugger asks for all loaded modules
 * NB: This is called with interrupts off,
 * and possibly on a very confused system.
 * Dont try to take locks or anything.
 */
#ifdef MODLIST
PUBLIC void DebuggerNotifyModuleList(void)
{
    PMODULE pMod;

    /* Walk the list of loaded modules
     * NB: Initially there are no modules
     */
    pMod = ModuleList;
    while (pMod) {
        /* Notify of this one (BUGBUG exclude mmlite.exe ?)
         */
        pMod->DebugToken = DebuggerNotifyLoad(pMod->pName,
                                              pMod->Start,
                                              pMod->End,
                                              pMod->Exports);
        
        /* Move on to next, if any
         */
        pMod = ListNodeNext(Mod_off, pMod);
        if (pMod == ModuleList)
            break;
    }
}
#else
PUBLIC void DebuggerNotifyModuleList(void)
{
    PMODULE pModule;
    PLISTNODE pNext;

    /* Walk the list of loaded modules
     * NB: Initially there are no modules
     */
    pNext = ModuleList;
    for (;pNext;) {

        /* Get the proper pointer
         */
        pModule = GetModuleFrom(pNext,Link);

        /* Notify of this one (BUGBUG exclude mmlite.exe ?)
         */
        pModule->DebugToken = DebuggerNotifyLoad(pModule->pName,
                                                 pModule->Start,
                                                 pModule->End,
                                                 pModule->Exports);
        
        /* Move on to next, if any
         */
        pNext = pModule->Link.pNext;
        if (pNext == ModuleList)
            break;
    }
}
#endif

/* ++++++++++++++++++++++++++++++++++++++++++
 * IProcess implementation
 */

PRIVATE SCODE MCT PrcQueryInterface(IProcess *pThis,
                              /* [in] */ REFIID pIid,
                              /* [out] */ void **ppObject)
{
    return ModuleQueryInterface ( (IModule *)(pPR(pThis)), pIid, ppObject);
}
        
PRIVATE UINT MCT PrcAddRef ( 
            IProcess *pThis)
{
    return ModuleAddRef( (PIMODULE) pPR(pThis) );
}
        
PRIVATE UINT MCT PrcRelease ( 
            IProcess *pThis)
{
    return ModuleRelease( (PIMODULE) pPR(pThis) );
}
        
PRIVATE SCODE MCT PrcCreateThread( 
            PIPROCESS pThis,
            CREATE_THREAD_FLAGS Flags,
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg,
            ADDR_SIZE StackSize,
            /*optional,out*/ PTR *pStackTop,
            /*optional,out*/ IThread **ppNewThread)
{
    PMODULE This = pPR(pThis);
    if (StackSize == 0)
        StackSize = This->DefaultStackSize;
    Flags |= This->CreationFlags;

    /* Create the thread
     */
    return ThreadCreate( pThis, pStart, Arg, StackSize, Flags, NULL,
                        pStackTop, ppNewThread);
}

PRIVATE SCODE MCT PrcClone( 
            PIPROCESS iThis,
            /*out*/ PIPROCESS *ppClone)
{
    PMODULE This = pPR(iThis);
    PIMODULE mod;
    SCODE sc;

    /* XXX Should probably set dependency of parent module */
    sc = ModuleCreate(_T("#clone"), NULL,
                      0, 0, 0, NULL, 0, /* set module stuff to NULL */
                      NULL, 0, 0, 0, This->CreationFlags, &mod);
    if (SUCCEEDED(sc))
        sc = mod->v->QueryInterface(mod, &IID_IProcess, (void**) ppClone);

    RELEASE(mod);
    return sc;
}

/* Image-specific loader code is elsewhere
 */
#include <base/loader.h>

PRIVATE SCODE MCT PrcLoadImage(IProcess *pThis,
                       const _TCHAR * pImage,
                       const _TCHAR * pArgs,
                       UINT Flags,
                       /* [out] */ IModule **ppIModule)
{
    UnusedParameter(pThis);
    return LdrLoadImage(GetImageNameSpace(), pImage, pArgs, Flags, ppIModule);
}
        
PRIVATE SCODE MCT PrcInit( IProcess *pThis,
                           ADDR_SIZE StackSize,
                           PINAMESPACE ns,
                           PCONDITION pCond,
                           PMUTEX pMx)
{
    PMODULE This = pPR(pThis);

    if (StackSize) {
        This->DefaultStackSize = StackSize;
    }

    if (ns) {
        ns = (PTR) AtomicSwap((ADDRESS*) &This->pNameSpace, (ADDRESS) ns);
        RELEASE(ns);
    }

    if (pCond) {
        pCond = (PTR) AtomicSwap((ADDRESS*) &This->pTermination,
                                 (ADDRESS) pCond);
        if (pCond)
            Condition_Broadcast(pCond);
    }

    return S_OK;
}

PRIVATE SCODE MCT PrcMachineInfo( IProcess *pThis,
                       MACHINE_INFO_FLAVOR Kind,
                       /* [out] */ PMACHINE_INFO pInfo)
{
    SCODE sc;
    ADDR_SIZE Total, Used;

    /* NB: On MPs, invoking thread should be bound to a processor
     */

    UnusedParameter(pThis);

    switch (Kind) {
    case MACHINE_KIND_GENERIC:
        sc = NO_ERROR;
        if (pInfo) {
            PIHEAP pHp = CurrentHeap();
            memset(pInfo,0,sizeof *pInfo);

            pInfo->ProcessorKind = CurrentProcessorKind();
            pInfo->OEMId = CurrentProcessorOEM();
            sc = pHp->v->Status(pHp,
                                &Total,
                                NULL,
                                &Used,
                                NULL);
            pInfo->TotalPhysicalMemory = Total;
            pInfo->AvailablePhysicalMemory = Total - Used;
            pInfo->ProcessorSpeed = CurrentProcessorSpeed();
        }
        break;
    default:
        sc = MachineSpecificInfo(Kind,pInfo);
    }

    return sc;
}

#include "s_CProcess_v.c"    /* Instantiate v-tables */

PRIVATE PROCESS TheProcess = {
    &ModuleCProcessVtbl,
    &PrcCProcessVtbl,
    1, /* wont go away on refcount */
#ifdef MODLIST
    { &TheProcess, &TheProcess },
#else
    { &TheProcess.Link, &TheProcess.Link },
#endif
    0, 0, /* BUGBUG */
    NULL, /* would be cosmetic anyways */
    DEFAULT_STACK_SIZE,
#if PRINT_IDLE_TIME
#if defined(__NO_BUILTIN_INT64)
    Int64Initializer(0,0),
    Int64Initializer(0,0),
#else
    TIME_ORIGIN,TIME_ORIGIN,
#endif
#endif
    0,
    CONDITION_INITIALIZER, &TheProcess._Termination, &ModuleListLock,
    0,0,0,

    NULL, NULL, NULL,

#ifdef __NO_EXECUTABLES
    NULL, 0,
#else
    (ADDRESS*) &BaseRtlVtbl, sizeof(BaseRtlVtbl) / sizeof(PTR),
#endif
    NULL,                 /* no dependents */
    0,
    0,                    /* DebugToken */
    0,                    /* CreationFlags */
    _TEXT("mmlite.exe"),  /* Name */
    _TEXT("")             /* Args */
};

/* More RTL calls
 */
_TCHAR * RTLCALLTYPE ProcessImageFromPC( ADDRESS Pc )
{
    PIMODULE pIMod;
    _TCHAR * Ret;

    if (FAILED(GetModuleContainsAddr( Pc, &pIMod)))
        return NULL;

    /* Mumble, this returned pointer is not refcounted properly.
     */
    Ret = (_TCHAR *) pMD(pIMod)->pName;
    pIMod->v->Release(pIMod);

    return Ret;
}

_TCHAR * RTLCALLTYPE ProcessArgsFromPC( ADDRESS Pc )
{
    PIMODULE pIMod;
    _TCHAR * Ret;

    if (FAILED(GetModuleContainsAddr( Pc, &pIMod)))
        return NULL;

    /* Mumble, this returned pointer is not refcounted properly.
     */
    Ret = (_TCHAR *) pMD(pIMod)->pArgs;
    pIMod->v->Release(pIMod);

    return Ret;
}

/********* XXX temporary *************/
SCODE MCT ModuleGetVtable(
             IModule *pThis,
             ADDRESS **pExportData,
             UINT *pExportSize)
{
    *pExportData = pMD(pThis)->Exports;
    *pExportSize = pMD(pThis)->ExportsCount;
    return NO_ERROR;
}

/********* XXX temporary *************/


/* ++++++++++++++++++++++++++++++++++++++++++
 * Constructors
 */

/* Constructor for IModule
 */
SCODE ModuleCreate(const _TCHAR * pName,
                   const _TCHAR * pArgs,
                   PBYTE StartAddress,
                   UINT AddressSize,
                   MODULEENTRY EntryPoint,
                   PTR ExportData,
                   UINT ExportSize,
                   PIMODULE *ImportedModules,
                   UINT nImportedModules,
                   ADDRESS NotAnyMore__Metadata,
                   UINT32 NotAnyMore__EntryPointToken,
                   UINT32 Flags,
                   IModule **ppIModule)
{
    PIHEAP pIHeap = CurrentHeap();
    PMODULE pMod;
    int NameSize;
    int ArgsSize;
    _TCHAR * pCmdLine;

    /* Allocate a new descriptor.
     */
    if (!pName) pName = _T("");
    NameSize = _tcslen(pName) + 1;

    if (!pArgs) pArgs = _T("");
    ArgsSize = _tcslen(pArgs) + 1;

    pMod = (PMODULE) pIHeap->v->Alloc(pIHeap, HEAP_DEBUG_NAME("IModule"),
        (sizeof(MODULE)
         + nImportedModules * sizeof(PIMODULE)
         + ((NameSize + ArgsSize) * sizeof(_TCHAR))), 0);

    if (pMod == NULL)
        return E_NOT_ENOUGH_MEMORY;

    /* Initialize all fields
     */
    pMod->RefCnt = 1;
    pMod->v = &ModuleCProcessVtbl;
    pMod->pv = &PrcCProcessVtbl;

    pMod->Start = (ADDRESS) StartAddress;
    pMod->End   = (ADDRESS)(StartAddress + AddressSize);
    pMod->Entry = EntryPoint;
    Condition_Init(&pMod->_Termination);
    pMod->pTermination = &pMod->_Termination;
    pMod->pTerminationMx = &ModuleListLock;
    pMod->Exports = (ADDRESS *) ExportData;
    pMod->ExportsCount = ExportSize;
    /* XXX. Use kernel values for now */
    pMod->pVmView = 0;
    pMod->pHeap = CurrentHeap();
    pMod->pNameSpace = CurrentNameSpace();
        pMod->pNameSpace->v->AddRef(pMod->pNameSpace);
    pMod->ExitCode = 0;
    pMod->io0 = 0;
    pMod->io1 = 0;
    pMod->io2 = 0;

#if PRINT_IDLE_TIME
    if (pTheScheduler) {
        pMod->IdleWas = pTheScheduler->v->GetIdleTime(pTheScheduler);
        pMod->TimeWas = CurrentTime();
    }
#endif

    /* Use kernel's default stack size. This allows for
     * globally changing the stack size from the default.
     */
    pMod->DefaultStackSize = TheProcess.DefaultStackSize;

    /* Dependency list is placed immediately after module struct in extra
     * memory that was allocated with the module struct
     */
    pMod->nImportedModules = nImportedModules;
    pMod->ImportedModules = (PIMODULE *) (pMod + 1);
    memcpy(pMod->ImportedModules, ImportedModules,
           nImportedModules * sizeof(PIMODULE));

    pMod->CreationFlags = Flags;

    /* Name & Args are placed after dependency list in same extra memory */
    pCmdLine = (_TCHAR *) (pMod->ImportedModules + nImportedModules);
    pMod->pName = _tcscpy(&pCmdLine[       0],pName);
    pMod->pArgs = _tcscpy(&pCmdLine[NameSize],pArgs);

#if DEATH_WATCH
    printf("  ++ module %s(%x) at %x..%x\n", pMod->pName, pMod,
           pMod->Start, pMod->End);
#endif

    /* Add to the list of all active modules
     */
#ifdef MODLIST
    ListNodeInit(Mod_off, pMod);
#else
    ListInitialize(pMod->Link);
#endif
    Mutex_Lock(&ModuleListLock);
    plist("before-insert");
    /* Lets make sure it goes at the front */
    {
#ifdef MODLIST
        ListAddFirst(&ModuleList, Mod_off, pMod);
#else
        PLISTNODE New = &pMod->Link;
        New->pNext = ModuleList;
        New->pPrev = ModuleList->pPrev;
        ModuleList->pPrev = New;
        New->pPrev->pNext = New;
        ModuleList = New;
#endif
    }
    plist("after-insert");
    Mutex_Unlock(&ModuleListLock);

    /* Notify the debugger
     */
    pMod->DebugToken = DebuggerNotifyLoad(pMod->pName,
                                          pMod->Start,
                                          pMod->End,
                                          pMod->Exports);

    /* Done.
     */
    if (ppIModule)
        *ppIModule = (PIMODULE)pMod;

    return NO_ERROR;

}

/* Initialization
 */
PIPROCESS InitializeIModule(PIHEAP pHeap, PINAMESPACE pNameSpace)
{
    PMODULE pMod;

    Mutex_Init(&ModuleListLock);
    /* Add kernel module */
    pMod = &TheProcess;
#ifdef MODLIST
    ListNodeInit(Mod_off, pMod);
#else
    ListInitialize(pMod->Link);
#endif
    pMod->pHeap = pHeap;
    pMod->pNameSpace = pNameSpace;
#if !defined(_MSC_VER) || defined(arm)
    pMod->End = (ADDRESS)_end;
#endif
#ifdef MODLIST
    ModuleList = pMod;
#else
    ModuleList = &pMod->Link;
#endif
    return (PIPROCESS) &TheProcess.pv;
}

#ifdef mips /* not needed on other platforms */
/* XXX */
void SetBaseExports(ADDRESS *Exports,
                    UINT ExportsCount,
                    const _TCHAR *pArgs)
{
    TheProcess.Exports = Exports;
    TheProcess.ExportsCount = ExportsCount;
    if (pArgs)
        TheProcess.pArgs = pArgs;
}
#endif
