/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * This file implements two abstractions in one, the IProcess
 * and the IModule.  This is in order to keep things simpler
 * and save on dynamic data structures.
 */

#define OPTIMIZE 1  /* enable Base.h optimizations, since we're the "base" */
#define DEATH_WATCH 0

#include <mmlite.h>
#include <fred.h>
#include <base/cons.h>
#include <base/process2.h>
#include <base/thread2.h>
#include <base/schedul.h>
#include <schedulif.h>
#include <base/debugger.h>
#include <util/list-single.h>

/* Imports
 */
extern struct IBaseRtlVtbl BaseRtlVtbl;

/* Store count of waiters in high half of RefCnt */
#define WAITER_SHIFT (_UINTSIZE/2)
#define WAITER_MASK ((1<<WAITER_SHIFT)-1)

/* Local globals
 */
PRIVATE PNODE ModuleList = NULL;
PRIVATE MUTEX ModuleListLock = MUTEX_INITIALIZER;
PRIVATE PROCESS TheProcess;

/* ++++++++++++++++++++++++++++++++++++++++++
 * IModule implementation
 */

PRIVATE SCODE MCT ModuleQueryInterface ( 
            IModule *pThis,
            /* [in] */ REFIID pIid,
            /* [out] */ void **ppObject)
{
    SCODE sc;

    sc = GenericQueryInterface((IUnknown *)pThis,pIid,ppObject,&IID_IModule);

    /* Special handling for IProcess interface if not recognized already
     */
    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IProcess, pIid)) {
        pThis->v->AddRef(pThis);
        *ppObject = (void*) &pMD(pThis)->pv;
        sc = S_OK;
    }

    return sc;
}
        
PRIVATE UINT MCT ModuleAddRef ( 
            IModule *pThis)
{
    UINT RefCnt;

    RefCnt = AtomicInc(&pMD(pThis)->RefCnt);

    return RefCnt;
}

PRIVATE UINT MCT ModuleRelease ( 
            IModule *pThis)
{
    UINT RefCnt;
    PMODULE pMod = pMD(pThis);

    RefCnt = AtomicDec(&pMod->RefCnt);
    assert((INT) RefCnt >= 0);

    if (RefCnt == 0) {
        PIHEAP pIHeap = CurrentHeap();

        assert(pMod != &TheProcess);

        /* Take module out of list of all modules
         */
        Mutex_Lock(&ModuleListLock);
        RefCnt = pMod->RefCnt;
        if (RefCnt != 0) {
            /* Cover for race with LookupModule:
             * Somebody looked it up before we got the lock.
             */
            Mutex_Unlock(&ModuleListLock);
            return RefCnt;
        }
        ModuleList = Node_StackRemove(ModuleList, ModuleGetNode(pMod));
        Mutex_Unlock(&ModuleListLock);

#if PRINT_IDLE_TIME
        {
        TIME IdleIs = pTheScheduler->v->GetIdleTime(pTheScheduler);
        TIME TimeIs = CurrentTime();
        printf("  -- module %s Idle %lld of %lld elapsed\n", pMod->pName, 
               Int64Subtract(IdleIs,pMod->IdleWas),
               Int64Subtract(TimeIs,pMod->TimeWas));
        }
#endif

        /* Notify the debugger
         */
        DebuggerNotifyUnload(pMod->pName,pMod->Start,pMod->DebugToken);
        /* Release all memory resources
         */
#ifdef notyet
        for (i = 0; i < pMod->nImportedModules; i++)
            pMod->ImportedModules[i]->v->Release(pMod->ImportedModules[i]);
        pMod->nImportedModules = 0;
#endif

#if DEATH_WATCH
        printf("Module %x dead\n", pMod);
#endif
        Condition_Destroy(&pMod->Termination);
        pIHeap->v->Free(pIHeap,0,(PTR) pMod->Start);
        pIHeap->v->Free(pIHeap,0,pThis);
    } else if ((RefCnt & WAITER_MASK) == (RefCnt >> WAITER_SHIFT)) {
        /* Someone is waiting for this process to terminate. */

        /* Ok to do one at a time: the next guy will Release and come here */
        Mutex_Lock(&ModuleListLock);
        /* Barrier to keep us from getting into the read-test-block window
         * in ModuleWaitFor. This avoids the short race opportunity.
         */
        Mutex_Unlock(&ModuleListLock);
        Condition_Signal(&pMod->Termination);
    }



    return (RefCnt & WAITER_MASK);
}

SCODE MCT PeGetFunctionAddress (
        PTR ExportData,
        UINT ExportSize,
        const _TCHAR * ModName,
        UINT Ordinal,
        /* [out] */ ADDRESS *pAddress);

PRIVATE SCODE MCT ModuleGetFunctionAddress ( 
            IModule *pThis,
            const _TCHAR * Name,
            UINT Ordinal,
            /* [out] */ ADDRESS *pAddress)
{
#ifdef i386
    if (pMD(pThis)->Exports != (ADDRESS *) &BaseRtlVtbl)
      return PeGetFunctionAddress(pMD(pThis)->Exports,
                                  pMD(pThis)->ExportsCount,
                                  Name,
                                  Ordinal,
                                  pAddress);
#endif
       
    /* BUGBUG: This is the simplest possible implementation that only
     * works for zero based ordinals.
     */
    if (Name != NULL)
      return E_INVALID_PARAMETER; /* not yet supported */

    if (Ordinal > pMD(pThis)->ExportsCount)
      return E_INVALID_PARAMETER;

    *pAddress = pMD(pThis)->Exports[Ordinal];

    return NO_ERROR;

}
        
PRIVATE SCODE MCT ModuleGetName ( 
            IModule *pThis,
            /* [size_is][out] */ _TCHAR *Buffer,
            /* [in] */ UINT BufSize)
{
    if (Buffer)
        _tcsncpy(Buffer,pMD(pThis)->pName,BufSize);
    return NO_ERROR;
}
        
PRIVATE SCODE MCT ModuleGetArgs ( 
            IModule *pThis,
            /* [size_is][out] */ _TCHAR * Buffer,
            /* [in] */ UINT BufSize)
{
    if (Buffer)
        _tcsncpy(Buffer,pMD(pThis)->pArgs,BufSize);
    return NO_ERROR;
}
        
PRIVATE SCODE MCT ModuleGetEntryPoint ( 
            IModule *pThis,
            /* [out] */ ADDRESS *pEntry)
{
    if (pEntry)
        *pEntry = (ADDRESS)pMD(pThis)->Entry;
    return NO_ERROR;
}

#if 0
PRIVATE SCODE MCT ModuleRelocateRva ( 
            IModule *pThis,
            ADDRESS Rva,
            /* [out] */ ADDRESS *pAddress,
            /* [out] */ UINT *pSize)
{
    /* XXX Naive and incorrect */
    *pAddress = (ADDRESS)pMD(pThis)->Start + Rva;

    if (pSize)
        *pSize = pMD(pThis)->End - pMD(pThis)->Start;

    return S_OK;
}
#endif

PRIVATE SCODE MCT ModuleSetExitCode (
            IModule *pThis,
            /* [in] */ INT ExitCode)
{
    pMD(pThis)->ExitCode = ExitCode;
    return S_OK;
}

PRIVATE SCODE MCT ModuleGetExitCode (
            IModule *pThis,
            /* [out] */ INT *pExitCode)
{
    *pExitCode = pMD(pThis)->ExitCode;
    return S_OK;
}

/* We only wait for Modules to unload.
 * Which implies processes.
 */
PRIVATE SCODE MCT ModuleWaitFor(IModule *pThis)
{
    PMODULE pMod = pMD(pThis);
    UINT Refs;

    Refs = AtomicAdd(&pMod->RefCnt, 1 << WAITER_SHIFT);
    for (;;) {
        /* If all the refs are waiters then stop waiting */
        if ((Refs & WAITER_MASK) == (Refs >> WAITER_SHIFT)) {
            AtomicSub(&pMod->RefCnt, 1 << WAITER_SHIFT);
            return S_OK;
        }
        assert((Refs & WAITER_MASK) > (Refs >> WAITER_SHIFT)); /* Sanity */
        /* No lock needed above: only waiters had refs. */
        /* Cover for race: another ref was dropped and the signal gets lost */
        Mutex_Lock(&ModuleListLock);
        Refs = pMod->RefCnt;
        if ((Refs & WAITER_MASK) != (Refs >> WAITER_SHIFT)) {
            Condition_Wait(&pMod->Termination, &ModuleListLock, NULL);
        }
        Mutex_Unlock(&ModuleListLock);
        Refs = pMod->RefCnt;
    }
}

/* This one was defined as an RTL. tzk tzk tzk
 */
SCODE GetModuleContainsAddr(ADDRESS Addr, PIMODULE *ppIModule)
{
    PMODULE pMod;
    PNODE pnNext;

    if (ppIModule)
        *ppIModule = NULL; /* allow sloppy usage */

    Mutex_Lock(&ModuleListLock);

    pnNext = ModuleList;

    /* Simple list scan, we dont expect too many modules at once
     */
    while (pnNext != NULL) {
        pMod = (PMODULE) Node_GetData(pnNext);

        if ((ModuleGetStartAddress(pMod) <= Addr) &&
            (Addr < ModuleGetEndAddress(pMod))) {
            goto Found;
        }
        pnNext = Node_GetNext(pnNext);
    }
    /* If nothing found, return the Base. Better than NULL usually... */
    pMod = &TheProcess;

 Found:
    if (ppIModule && pMod) {
        pMod->v->AddRef((PIMODULE)pMod);
        *ppIModule = (PIMODULE)pMod;
    }

    Mutex_Unlock(&ModuleListLock);

    return (pMod) ? NO_ERROR : E_OBJECT_NOT_FOUND;
}

/* Get module objects one at a time.  Return the module that follows the
 * one specified by Previous.  If Previous is NULL return the first entry.
 * Return S_OK if found a module, S_FALSE if Previous was the last one, or
 * an error code.   Specifically return E_MOD_NOT_FOUND if Previous not valid,
 * or if there are no loaded modules at all. [not implem for mmlite]
 * This function takes a reference on the object it returns, if any.
 */
SCODE GetModuleNext(PIMODULE Previous, PIMODULE *ppIModule)
{
    PIMODULE pIMod;
    PNODE pnNext;
    SCODE sc = S_OK;

    Mutex_Lock(&ModuleListLock);

    /* Note that the list is never empty because the kernel does not go away.
     */
    pnNext = ModuleList;

    if (Previous != NULL) {

        /* Wants the next, if any.
         * Note: since the module list is insert-at-front, any new
         * module added after this function was first called (with
         * Previous==NULL) will be found last.
         */
        pnNext = Node_GetNext(ModuleGetNode(Previous));

        /* Check for end of list
         */
        if (pnNext == NULL)
            sc = S_FALSE;
    }

    pIMod = NULL;
    if (pnNext) {
        pIMod = (PIMODULE) Node_GetData(pnNext);
        (void) pIMod->v->AddRef(pIMod);
    }

    Mutex_Unlock(&ModuleListLock);

    *ppIModule = pIMod;
    return sc;
}

/* Find module by name */
PIMODULE LookupModule(const _TCHAR *pModName, UINT VersionCheck)
{
    PMODULE pMod;
    PNODE pnNext;

    /* Time stamp or crc or whatever the linker gives to check versions */
    UnusedParameter(VersionCheck);

    Mutex_Lock(&ModuleListLock);

    pnNext = ModuleList;

    /* Simple list scan, we dont expect too many modules at once
     */
    while (pnNext != NULL) {
        pMod = (PMODULE) Node_GetData(pnNext);

        if (!_tcsicmp(ModuleNameGet(pMod), pModName))
            goto Found;

        pnNext = Node_GetNext(pnNext);
    }

    /* Mumble..whats he looking for ?
     */
    pMod = NULL;
 Found:
    if (pMod) {
        pMod->v->AddRef((PIMODULE)pMod);
    }

    Mutex_Unlock(&ModuleListLock);
    return (PIMODULE) pMod;
}

/* The debugger asks for all loaded modules
 * NB: This is called with interrupts off,
 * and possibly on a very confused system.
 * Dont try to take locks or anything.
 */
PUBLIC void DebuggerNotifyModuleList(void)
{
    PMODULE pModule;
    PNODE pnNext;

    /* Walk the list of loaded modules
     * NB: Initially there are no modules
     */
    pnNext = ModuleList;
    while (pnNext != NULL) {
        /* Get the proper pointer
         */
        pModule = (PMODULE) Node_GetData(pnNext);

        /* Notify of this one (BUGBUG exclude mmlite.exe ?)
         */
        pModule->DebugToken = DebuggerNotifyLoad(pModule->pName,
                                                 pModule->Start,
                                                 pModule->End,
                                                 pModule->Exports);
        
        /* Move on to next, if any
         */
        pnNext = Node_GetNext(pnNext);
    }
}

/* ++++++++++++++++++++++++++++++++++++++++++
 * IProcess implementation
 */

PRIVATE SCODE MCT PrcQueryInterface(IProcess *pThis,
                              /* [in] */ REFIID pIid,
                              /* [out] */ void **ppObject)
{
    return ModuleQueryInterface ( (IModule *)(pPR(pThis)), pIid, ppObject);
}
        
PRIVATE UINT MCT PrcAddRef ( 
            IProcess *pThis)
{
    return ModuleAddRef( (PIMODULE) pPR(pThis) );
}
        
PRIVATE UINT MCT PrcRelease ( 
            IProcess *pThis)
{
    return ModuleRelease( (PIMODULE) pPR(pThis) );
}
        
PRIVATE SCODE MCT PrcCreateThread( 
            IProcess *pThis,
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg,
            ADDR_SIZE StackSize,
            /*optional*/     PIVMVIEW pVmView,
            /*optional,out*/ PTR *pStackTop,
            /*optional,out*/ IThread **ppNewThread)
{
    PMODULE This = pPR(pThis);
    if (StackSize == 0)
        StackSize = This->DefaultStackSize;
    Flags |= This->CreationFlags;

    /* Create the thread
     */
    return ThreadCreate( pThis, pStart, Arg, StackSize, Flags, NULL,
                        pStackTop, ppNewThread);
}

/* Image-specific loader code is elsewhere
 */
#include <base/loader.h>

PRIVATE SCODE MCT PrcLoadImage(IProcess *pThis,
                       const _TCHAR * pImage,
                       const _TCHAR * pArgs,
                       UINT Flags,
                       /* [out] */ IModule **ppIModule)
{
    UnusedParameter(pThis);
    return LdrLoadImage(GetImageNameSpace(), pImage, pArgs, Flags, ppIModule);
}
        
PRIVATE SCODE MCT PrcInit( IProcess *pThis,
                   ADDR_SIZE StackSize,
                   void *compat)
{
    UnusedParameter(compat);
    pPR(pThis)->DefaultStackSize = StackSize;
    return NO_ERROR;
}

PRIVATE SCODE MCT PrcMachineInfo( IProcess *pThis,
                       MACHINE_INFO_FLAVOR Kind,
                       /* [out] */ PMACHINE_INFO pInfo)
{
    SCODE sc;
    ADDR_SIZE Total, Used;

    /* NB: On MPs, invoking thread should be bound to a processor
     */

    UnusedParameter(pThis);

    switch (Kind) {
    case MACHINE_KIND_GENERIC:
        sc = NO_ERROR;
        if (pInfo) {
            PIHEAP pHp = CurrentHeap();
            memset(pInfo,0,sizeof *pInfo);

            pInfo->ProcessorKind = CurrentProcessorKind();
            pInfo->OEMId = CurrentProcessorOEM();
            sc = pHp->v->Status(pHp,
                                &Total,
                                NULL,
                                &Used,
                                NULL);
            pInfo->TotalPhysicalMemory = Total;
            pInfo->AvailablePhysicalMemory = Total - Used;
            pInfo->ProcessorSpeed = CurrentProcessorSpeed();
        }
        break;
    default:
        sc = MachineSpecificInfo(Kind,pInfo);
    }

    return sc;
}

#include "s_CProcess_v.c"    /* Instantiate v-tables */

PRIVATE PROCESS TheProcess = {
    &ModuleCProcessVtbl,
    &PrcCProcessVtbl,
    1, /* wont go away on refcount */
    { NULL, &TheProcess }, /* initializing node */
    0, 0, /* BUGBUG */
    NULL, /* would be cosmetic anyways */
    DEFAULT_STACK_SIZE,
#if PRINT_IDLE_TIME
#if defined(__NO_BUILTIN_INT64)
    Int64Initializer(0,0),
    Int64Initializer(0,0),
#else
    TIME_ORIGIN,TIME_ORIGIN,
#endif
#endif
    0,
    CONDITION_INITIALIZER,
    NULL,
    NULL,

    NULL,
    (ADDRESS*) &BaseRtlVtbl, sizeof(BaseRtlVtbl) / sizeof(PTR),
    NULL,                 /* no dependents */
    0,
    0,
    _TEXT("mmlite.exe"),  /* Name */
    _TEXT("")             /* Args */
};

/* More RTL calls
 */
_TCHAR * RTLCALLTYPE ProcessImageFromPC( ADDRESS Pc )
{
    PIMODULE pIMod;
    _TCHAR * Ret;

    if (FAILED(GetModuleContainsAddr( Pc, &pIMod)))
        return NULL;

    /* Mumble, this returned pointer is not refcounted properly.
     */
    Ret = (_TCHAR *) pMD(pIMod)->pName;
    pIMod->v->Release(pIMod);

    return Ret;
}

_TCHAR * RTLCALLTYPE ProcessArgsFromPC( ADDRESS Pc )
{
    PIMODULE pIMod;
    _TCHAR * Ret;

    if (FAILED(GetModuleContainsAddr( Pc, &pIMod)))
        return NULL;

    /* Mumble, this returned pointer is not refcounted properly.
     */
    Ret = (_TCHAR *) pMD(pIMod)->pArgs;
    pIMod->v->Release(pIMod);

    return Ret;
}

/********* XXX temporary *************/
SCODE MCT ModuleGetVtable(
             IModule *pThis,
             ADDRESS **pExportData,
             UINT *pExportSize)
{
    *pExportData = pMD(pThis)->Exports;
    *pExportSize = pMD(pThis)->ExportsCount;
    return NO_ERROR;
}

/********* XXX temporary *************/


/* ++++++++++++++++++++++++++++++++++++++++++
 * Constructors
 */

/* Constructor for IModule
 */
SCODE ModuleCreate(const _TCHAR * pName,
                   const _TCHAR * pArgs,
                   PBYTE StartAddress,
                   UINT AddressSize,
                   MODULEENTRY EntryPoint,
                   PTR ExportData,
                   UINT ExportSize,
                   PIMODULE *ImportedModules,
                   UINT nImportedModules,
                   ADDRESS NotAnyMore__Metadata,
                   UINT32 NotAnyMore__EntryPointToken,
                   UINT32 Flags,
                   IModule **ppIModule)
{
    PIHEAP pIHeap = CurrentHeap();
    PMODULE pMod;
    int NameSize;
    int ArgsSize;
    _TCHAR * pCmdLine;

    /* Allocate a new descriptor.
     */
    if (!pName) pName = _T("");
    NameSize = _tcslen(pName) + 1;

    if (!pArgs) pArgs = _T("");
    ArgsSize = _tcslen(pArgs) + 1;

    pMod = (PMODULE) pIHeap->v->Alloc(pIHeap, HEAP_DEBUG_NAME("IModule"),
        (sizeof(MODULE)
         + nImportedModules * sizeof(PIMODULE)
         + ((NameSize + ArgsSize) * sizeof(_TCHAR))), 0);

    if (pMod == NULL)
        return E_NOT_ENOUGH_MEMORY;

    /* Initialize all fields
     */
    pMod->RefCnt = 1;
    pMod->v = &ModuleCProcessVtbl;
    pMod->pv = &PrcCProcessVtbl;

    pMod->Start = (ADDRESS) StartAddress;
    pMod->End   = (ADDRESS)(StartAddress + AddressSize);
    pMod->Entry = EntryPoint;
    Condition_Init(&pMod->Termination);
    pMod->Exports = (ADDRESS *) ExportData;
    pMod->ExportsCount = ExportSize;
    /* XXX. Use kernel values for now */
    pMod->pVmView = 0;
    pMod->pHeap = CurrentHeap();
    pMod->pNameSpace = CurrentNameSpace();
    pMod->ExitCode = 0;

#if PRINT_IDLE_TIME
    pMod->IdleWas = pTheScheduler->v->GetIdleTime(pTheScheduler);
    pMod->TimeWas = CurrentTime();
#endif

    /* Use kernel's default stack size. This allows for
     * globally changing the stack size from the default.
     */
    pMod->DefaultStackSize = TheProcess.DefaultStackSize;

    /* Dependency list is placed immediately after module struct in extra
     * memory that was allocated with the module struct
     */
    pMod->nImportedModules = nImportedModules;
    pMod->ImportedModules = (PIMODULE *) (pMod + 1);
    memcpy(pMod->ImportedModules, ImportedModules,
           nImportedModules * sizeof(PIMODULE));

    pMod->CreationFlags = Flags;

    /* Name & Args are placed after dependency list in same extra memory */
    pCmdLine = (_TCHAR *) (pMod->ImportedModules + nImportedModules);
    pMod->pName = _tcscpy(&pCmdLine[       0],pName);
    pMod->pArgs = _tcscpy(&pCmdLine[NameSize],pArgs);

#if DEATH_WATCH
    printf("  ++ module %s(%x) at %x..%x\n", pMod->pName, pMod,
           pMod->Start, pMod->End);
#endif

    /* Add to the list of all active modules
     */
    ModuleNodeInitialize(pMod);
    Mutex_Lock(&ModuleListLock);
    ModuleList = Node_Push(ModuleList, ModuleGetNode(pMod));
    Mutex_Unlock(&ModuleListLock);

    /* Notify the debugger
     */
    pMod->DebugToken = DebuggerNotifyLoad(pMod->pName,
                                          pMod->Start,
                                          pMod->End,
                                          pMod->Exports);

    /* Done.
     */
    if (ppIModule)
        *ppIModule = (PIMODULE)pMod;

    return NO_ERROR;

}

/* Initialization
 */
PIPROCESS InitializeIModule(PIHEAP pHeap, PINAMESPACE pNameSpace)
{
    PMODULE pMod;

    Mutex_Init(&ModuleListLock);
    /* Add kernel module */
    pMod = &TheProcess;
    /* Already initialized in struct filler above */
/*      ModuleNodeInitialize(pMod); */
    pMod->pHeap = pHeap;
    pMod->pNameSpace = pNameSpace;
#if !defined(_MSC_VER) || defined(arm)
    pMod->End = (ADDRESS)_end;
#endif
    ModuleList = ModuleGetNode(pMod);
    return (PIPROCESS) &TheProcess.pv;
}

#ifdef mips /* not needed on other platforms */
/* XXX */
void SetBaseExports(ADDRESS *Exports,
                    UINT ExportsCount,
                    const _TCHAR *pArgs)
{
    TheProcess.Exports = Exports;
    TheProcess.ExportsCount = ExportsCount;
    if (pArgs)
        TheProcess.pArgs = pArgs;
}
#endif
