/* thread2.c
 * Implementation of kernel execution threads.
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#define OPTIMIZE
/* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <vmemif.h>
#include <fred.h>
#include <base/thread2.h>
#include <base/schedul.h>
#include <schedulif.h>
#include <diagnostics.h>

/* Machdep code */
#include "_thread.c"

PTHDINFO TheCurrentThread = NULL;
PIVMVIEW TheCurrentVmView = NULL;

PRIVATE SCODE MCT ThreadQueryInterface( 
    PITHREAD pThis,
    /* [in] */ REFIID pIid,
    /* [out] */ void **ppUnk
    )
{
    return GenericQueryInterface((IUnknown *)pThis, pIid, ppUnk, &IID_IThread);
}
        
PRIVATE UINT MCT ThreadAddRef(
    PITHREAD pThis
    )
{
    UINT RefCnt;

    RefCnt = AtomicInc(&pTH(pThis)->RefCnt);

    return RefCnt;
}
        
PRIVATE UINT MCT ThreadRelease (PITHREAD pThis)
{
    UINT RefCnt;

    RefCnt = AtomicDec(&pTH(pThis)->RefCnt);

#if defined(STACK_STATS)
    printf("ThreadRelease (0x%x)->%d stack usage: %d/%d\n",
           (ADDRESS)pThis, RefCnt,
           CheckStackUsage(pThis), pTH(pThis)->StackSize);
#endif

    if (RefCnt == 0) {
        PIPROCESS pPrc;
        DBG(("Thread (0x%x) dead.\n", (ADDRESS)pThis));
        CurrentHeapDelayedFree(pThis);


        /* If ThreadExit did not zero StartedIn, this must be a thread that
         * never ran and thus this is from normal thread context so no
         * need to worry about locks etc. In any case the ref needs to go.
         */
        pPrc = pTH(pThis)->StartedIn;
        if (pPrc)
            pPrc->v->Release(pPrc);
    }

    return RefCnt;
}
        
void ThreadExit(void)
{
    PITHREAD pThis = CurrentThread();
    PIPROCESS pPrc;

    /* Release virtual memory.
     * The order of events is significant.
     * We might get preempted at any time.
     * 1. Set thread view to zero (physical)
     *    This will make sure we don't switch back to the view that is
     *    terminating.
     * 2. Switch to physical space.
     *    If we get preempted it is ok to switch to anybody else, but
     *    from now on we are off the old one.
     * 3. Set CurrentVmView to reflect the change.
     *    May cause unnecessary switch if whoever preempted us blocks
     *    and then runs again. But that's ok.
     * 4. Release the ref.
     */
    if (pTH(pThis)->pVmView) {
        PIVMVIEW vv = pTH(pThis)->pVmView;
        DBG(("vvRele(%x)\n", (UINT) vv));
        pTH(pThis)->pVmView = NULL; /* 1 */
        vv->v->SwitchTo(0); /* 2 */
        TheCurrentVmView = 0; /* 3 */
        vv->v->Release(vv); /* 4 */
    }

    /* Release ref on process.  Do it here, else we might get
     * in deadlock later: if the only two threads are this and
     * the one waiting-for the process to terminate then this
     * thread wont find an undertaker.
     * Yes it would have been better to do this in ThreadRelease.
     */
    pPrc = pTH(pThis)->StartedIn;
    pTH(pThis)->StartedIn = NULL; /* well, we no longer have a ref... */

    pPrc->v->Release(pPrc);

    /* Let scheduler know this thread is gone
     */
    pTheScheduler->v->RemoveThread(pTheScheduler, pThis);

    /* We never get back here.
     */
}
        
/* Disagreeable call this one.
 */
PRIVATE SCODE MCT ThreadGetProcess ( 
            IThread *pThis,
            /* [out] */ IProcess **ppIProcess)
{
    if (ppIProcess) {
        IProcess *pPrc = pTH(pThis)->StartedIn;
        pPrc->v->AddRef(pPrc);
        *ppIProcess = pPrc;
    }

    return S_OK;
}
        
/* Get partially initialized thread going.
 */
PRIVATE SCODE ThreadStartWithConstraint (
    PITHREAD pThis,
    PTIME_CONSTRAINT pcval,
    THREAD_FUNCTION pStart,
    THREAD_ARGUMENT Arg,
    PTR pStackTop,
    UINT ArgCount)
{
    PTR Top = pStackTop;
    /* Make sure this was only partially initialized */
    assert(pTH(pThis)->CxtInfo == NULL);

    /* If user did not specify stack offset, use default (empty stack) */
    if (Top == NULL)
        Top = pTH(pThis)->StackBase + pTH(pThis)->StackSize - sizeof(UINT);

    /* Setup processor state for new thread.
     * Push default frame.
     */
    MachineThreadCreate(pThis, pStart, Arg, Top, pStackTop ? ArgCount : -1, 0);

    /* One ref for the scheduler... */
    pThis->v->AddRef(pThis);

    /* Let scheduler know about this new thread
     */
    pTheScheduler->v->AddThread(pTheScheduler, pThis, pcval);

    return S_OK;
}

PRIVATE UINT ThreadGetValue(PITHREAD pTh, THREAD_VALUE Index)
{
    switch (Index) {
    case THREAD_VALUE_LAST_ERROR:
        return pTH(pTh)->LastError;
    case THREAD_VALUE_RPC_CONTINUATION:
        return (UINT) pTH(pTh)->RpcContinuation;
    default:
        return 0;
    }
}

PRIVATE void ThreadSetValue(PITHREAD pTh, THREAD_VALUE Index, UINT Value)
{
    switch (Index) {
    case THREAD_VALUE_LAST_ERROR:
        pTH(pTh)->LastError = Value;
        break;
    case THREAD_VALUE_RPC_CONTINUATION:
        pTH(pTh)->RpcContinuation = (PTR) Value;
        break;
    default:
        ;
    }
}

#include <s_CThread_v.c>        

#define STACK_OVERFLOW_SIGNATURE 0x91827365

#if defined(CHECK_STACK_OVERFLOW)

BOOL CheckStackOverflow(IThread *pThis)
{
    if (pThis &&
        *(PUINT32) pTH(pThis)->StackBase != STACK_OVERFLOW_SIGNATURE) {
        DBGOUT(DBGTHREAD,3,dprintf("Stack Overflow: %x\n", pThis));
#if defined(STACK_STATS)
        printf("Thread=0x%x stack usage: %d/%d\n",
               (ADDRESS)pThis, CheckStackUsage(pThis), pTH(pThis)->StackSize);
#endif
        return FALSE;
    } else
        return TRUE;
}
#endif

#if defined(STACK_STATS)

void InitStackStats(IThread *pThis)
{
    int c;
    PUINT32 p = (PUINT32) pTH(pThis)->StackBase;

    for (c = pTH(pThis)->StackSize; c > 0; c -= sizeof(UINT32))
        *p++ = STACK_OVERFLOW_SIGNATURE;
}

INT CheckStackUsage(IThread *pThis)
{
    INT c = -1;
    PUINT pStk;

    if (pThis) {
        pStk = (PUINT) pTH(pThis)->StackBase;

        for (c = pTH(pThis)->StackSize; c > 0; c -= sizeof *pStk)
            if (STACK_OVERFLOW_SIGNATURE != *pStk++)
                break;
    }

    return c;
}

#endif


/* Constructor
 */
SCODE MCT ThreadCreate( 
            IProcess *pProcess,
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg,
            ADDR_SIZE StackSize,
            UINT32 Flags,
            /*optional*/     PIVMVIEW pVmView,
            /*optional,out*/ PTR *pStackTop,
            /*optional,out*/ IThread **ppNewThread)
{
    PIHEAP pIHeap = CurrentHeap();
    PITHREAD pThdNew;
    PBYTE pStack;

    /* Check for no-op case */
    if (pStackTop && !ppNewThread)
        return E_INVALID_PARAMETER;

    /* Allocate a stack and the thread descriptor
     */
    pThdNew = (PITHREAD) pIHeap->v->Alloc(pIHeap,
                                          HEAP_DEBUG_NAME("IThread"),
					  sizeof(THDINFO) + StackSize,
					  0);
    if (pThdNew == NULL)
        goto ErrorExit;

    memset(pThdNew,0,sizeof(THDINFO));

    /* Remember where the stack is
     */
    pStack = (PBYTE)((PTHDINFO) pThdNew + 1);
    pTH(pThdNew)->StackBase = pStack;
    pTH(pThdNew)->StackSize = StackSize;
    DBG(("NewThread (0x%x) start %x stack 0x%x..0x%x\n", (ADDRESS)pThdNew,
         (ADDRESS)pStart, (ADDRESS)pStack, (ADDRESS)(pStack+StackSize)));
#if defined(CHECK_STACK_OVERFLOW)
    *(PUINT32) pTH(pThdNew)->StackBase = STACK_OVERFLOW_SIGNATURE;
#endif
#if defined(STACK_STATS)
    InitStackStats(pThdNew);
#endif

    /* Our stacks grow downwards
     */
    pStack += StackSize - sizeof(UINT);

    /* Initialize vtable, reference count, etc.
     */
    Thread_Init((PTHDINFO)pThdNew);
    pTH(pThdNew)->v = &CThreadVtbl;
    pTH(pThdNew)->RefCnt = (ppNewThread && !pStackTop) ? 2 : 1;
    pTH(pThdNew)->StartedIn = pProcess;
    /* Address space */
    if (pVmView) {
        /* Take a ref as we save the pointer */
        pVmView->v->AddRef(pVmView);
        pTH(pThdNew)->pVmView = pVmView;
    }
    /* rest is NULL/0 */

    /* Take a reference on the process for this thread.
     */
    if (pProcess)
        (void) pProcess->v->AddRef(pProcess);

    if (pStackTop) {
      /* Return address where to push frames */
      *pStackTop = (PTR) pStack;
    } else {
      /* Setup processor state for new thread.
       * Push default frame.
       */
      MachineThreadCreate(pThdNew, pStart, Arg, pStack, -1, Flags);


      /* Let scheduler know about this new thread
       */
      pTheScheduler->v->AddThread(pTheScheduler, pThdNew, NULL);
    }

    /* Done ok
     */
    if (ppNewThread) {
        /* ref taken above */
        *ppNewThread = pThdNew;
    }

    /*printf("ThreadCreate: ok, pThdNew=0x%x vv=%x\n",pThdNew,pVmView);*/
    return S_OK;

 ErrorExit:
    DBGOUT(DBGTHREAD,3,dprintf("Failed CreateThread(%x): %x\n",
                               pStart, BaseGetLastError()));
    return E_NOT_ENOUGH_MEMORY;
}

void Thread_Init(
     PTHDINFO pThd
     )
{
    Thread_MutexListInitialize(pThd);
    Thread_ClearIdle(pThd);
    Thread_SetWaiting(pThd, THREAD_SLEEPING, NULL);
/*     ThreadSetWaitState(pThd, WAIT_NONE); */
    Thread_SetCurrentConstraint(pThd, NULL);
}

void Thread_UnlockMutexes(
     PTHDINFO pThd
     )
{
    assert(pThd != NULL);
    List_TransformData(Thread_GetMutexList(pThd), Thread_MutexUnlock);
}

BOOL Thread_UrgencyGT(
     PITHREAD pThd1,
     PITHREAD pThd2
     )
{
    return KConstraint_UrgencyGT(Thread_GetCurrentConstraint(pThd1),
				Thread_GetCurrentConstraint(pThd2));
}

PKCONSTRAINT Thread_FindMostUrgentMutexWaiter(
     PTHDINFO pThd
     )
{
    PMX pMutexMostUrgent;

    assert(pThd != NULL);

    pMutexMostUrgent = List_FindDataMax(Thread_GetMutexList(pThd),
					 Thread_MutexWaiterCompareGT);

    if (pMutexMostUrgent != NULL) {
      PLIST plWaiters = Mutex_GetWaiterQueue(pMutexMostUrgent);
      if (!List_IsEmpty(plWaiters))
      return (PKCONSTRAINT)Node_GetData(List_GetHead(plWaiters));
    }

    return NULL;
}

BOOL Thread_MutexWaiterCompareGT(
     PTR pMX1,
     PTR pMX2
     )
{
    PNODE pn1 = List_GetHead(Mutex_GetWaiterQueue(pMX1));
    PNODE pn2 = List_GetHead(Mutex_GetWaiterQueue(pMX2));
    if (pn2 == NULL) {
      return TRUE;
    }
    if (pn1 == NULL) {
      return FALSE;
    }

    return KConstraint_UrgencyGT((PKCONSTRAINT)Node_GetData(pn1),
				 (PKCONSTRAINT)Node_GetData(pn2));
}

void Thread_MutexUnlock(
     PTR pMutex
     )
{
    Mutex_Unlock((PMUTEX)pMutex);
}
