/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>
#include <assert.h>
#include <diagnostics.h>
#include <base/schedul.h>
#include <base/workitem.h>
#include <base/condition2.h>
#include <base/mutex2.h>
#include <base/kconstraint2.h>
#include <base/activity.h>
#include <util/util.h>
#include <util/list-single.h>

/* Limits of the current thread pool */
#define MAX_POOL_RETRIES 5
#define MAX_TOTAL_POOL_COUNT 22

/* Keep an array of all thread IDs b/c surrogate threads can never be the
 * CurrentActivity and so will never know their addresses otherwise. */
PRIVATE PACTINFO ThreadIds[MAX_TOTAL_POOL_COUNT];
PRIVATE UINT ThreadIndex = 0;

PRIVATE UINT TotalPoolCount;      // The number of total threads in the pool
PRIVATE UINT AvailablePoolCount;  // Number of threads not working

/* Synchronization structures for the thread pool */
PRIVATE CONDITION ThreadPoolCond;
PRIVATE MUTEX ThreadCreateLock;

PRIVATE TIME_CONSTRAINT ConstraintUrgent;

/* Signal variable for the base to shut the thread pool down */
PRIVATE BOOL KillEverything;

/******************************************************************************
 * Private function prototypes 
 */

PRIVATE void 
CreateSurrogateThread();

/******************************************************************************
 * Private function definitions
 */

UINT
ThreadPool_GetAvailableCount()
{
  return AvailablePoolCount;
}

void
ThreadPool_Create()
{
  int i;

  TotalPoolCount = 0;
  AvailablePoolCount = 0;
  KillEverything = FALSE;

  Condition_Init(&ThreadPoolCond);
  Mutex_Init(&ThreadCreateLock);

  Int32ToInt64(ConstraintUrgent.Start, 0);
  Int32ToInt64(ConstraintUrgent.Estimate, TIME_MILLIS(0));

  for (i = 0; i < MIN_AVAIL_POOL_COUNT; i++) {
    CreateSurrogateThread();
  }
                                        
}

/* Signal all the surrogate threads to exit. We are only killing available,
 * waiting threads here, b/c all working threads have already exited in order
 * for the scheduler to call BaseDelete. Don't worry about being preempted,
 * There is only the "base thread" when it's time to die b/c the threads we
 * are trying to kill are blocked. */
void
ThreadPool_Destroy()
{
  assert(AvailablePoolCount == TotalPoolCount);
  KillEverything = TRUE;
  Condition_Broadcast(&ThreadPoolCond);
  while (TotalPoolCount > 0) {
    MilliSleep(1);
  }
}

/* Main thread pool function where surrogate threads hang out in between
 * doing work items. */
PRIVATE void THREAD_LINKAGE
ThreadPoolFunc(
    THREAD_ARGUMENT arg
    )
{
  PWORKINFO pLocalWorkItem;
  WORK_FUNCTION WorkItemFunc;
  WORK_ARGUMENT WorkItemArg;
  UINT MyIndex = (UINT)arg;

  AtomicInc(&TotalPoolCount);

  while (TRUE) {
    DBG(("Surrogate Thread available %x\n", ThreadIds[MyIndex]));
    /* Reaching here means you are ready for work. */
    AtomicInc(&AvailablePoolCount);

    if (AvailablePoolCount > MAX_AVAIL_POOL_COUNT) {
      /* we have more threads than the specified maximum,
       * so commit suicide by returning from this function */
      AtomicDec(&AvailablePoolCount);
      break;
    }

    DBG(("Surrogate thread %x waiting\n", ThreadIds[MyIndex]));
    /* wait here for the scheduler */
    Condition_Wait(&ThreadPoolCond, NULL, NULL);
    DBG(("Surrogate thread %x woken\n", ThreadIds[MyIndex]));

    /* reaching here means you have a ready work item to process
     * and are no longer available. Doesn't have to be atomic b/c of lock.
     */
    AtomicDec(&AvailablePoolCount);

    if (KillEverything == TRUE) {
      break;
    }

    if (AvailablePoolCount < MIN_AVAIL_POOL_COUNT &&
        TotalPoolCount <= MAX_TOTAL_POOL_COUNT) {
      // If we are the last available thread, allocate one more and wait it
      CreateSurrogateThread();
    }

    // make sure we have a work item to do and copy it locally
    pLocalWorkItem = (PWORKINFO)Thread_GetWard(ThreadIds[MyIndex]);
    assert(pLocalWorkItem != NULL);

    DBG(("Surrogate Thread %x picking up %x\n", ThreadIds[MyIndex],
         pLocalWorkItem));

    // Extract params to local vars
    WorkItemFunc = WorkItem_GetFunc(pLocalWorkItem);
    WorkItemArg = WorkItem_GetArg(pLocalWorkItem);

    // Do the work
    WorkItemArg = WorkItemFunc(WorkItemArg);
    // Write back return value
    WorkItem_SetArg(pLocalWorkItem, WorkItemArg);
    // Set the work item back to having a pseudo thread b/c it's done running
    // and the scheduler expects it in order to identify a dead work item.
    WorkItem_SetPseudoThread(pLocalWorkItem);
    pTheScheduler->v->RemoveActivity(pTheScheduler,
                                     (PIACTIVITY)pLocalWorkItem);
    DBG(("Surrogate Thread %x done with %x\n", ThreadIds[MyIndex],
         pLocalWorkItem));
  }

  TotalPoolCount--;
  DBG(("Surrogate Thread %x dying\n", ThreadIds[MyIndex]));
}

PRIVATE void 
CreateSurrogateThread() 
{
  PIACTIVITY pNewThd;
  PIACTIVITY pCurrentActivity = CurrentActivity();
  SCODE sc;
  UINT relativeDeadline;

  Mutex_Lock(&ThreadCreateLock);
  if (ThreadIndex >= MAX_TOTAL_POOL_COUNT) {
    /* we cannot create more total threads than this maximum */
    DBG(("Max thread pool ceiling of %d reached.\n", MAX_TOTAL_POOL_COUNT));
    Mutex_Unlock(&ThreadCreateLock);
    return;
  }
  ThreadIndex++;
  Mutex_Unlock(&ThreadCreateLock);

  /* Because we pass in a pointer (pNewThd) here, call Release so that when
   * the thread exits, it deallocates itself. */
  pCurrentActivity->v->CreateThread(pCurrentActivity, CurrentProcess(),
                                    (THREAD_FUNCTION)ThreadPoolFunc,
                                    (THREAD_ARGUMENT)ThreadIndex, 0, NULL,
                                    &pNewThd);
  assert(pNewThd != NULL);
  pNewThd->v->Release(pNewThd);
  ThreadIds[ThreadIndex] = (PACTINFO)pNewThd;

  relativeDeadline = 1;
  /* Loop until our constraint has been excepted, with linear backoff.
   * Stop trying after MAX_POOL_RETRIES so we don't block. */
  do {
    ConstraintUrgent.Deadline = 
      Int64Add(CurrentTime(), TIME_MILLIS(relativeDeadline));
    sc = pTheScheduler->v->AddActivity(pTheScheduler, pNewThd, 
                                       &ConstraintUrgent);
    relativeDeadline++;
    if (relativeDeadline > MAX_POOL_RETRIES) {
      DBG(("Unable to schedule new surrogate thread.\n"));
      break;
    }
  } while (sc != S_OK);
  DBG(("CreateSurrogateThread %x %d\n", pNewThd, (UINT)sc));
}

/* This function is called every time a native work item is ready to run 
 * and the scheduler wants to "context switch" it in. */
PACTINFO THREAD_LINKAGE
ThreadPool_Switch(
    PWORKINFO pNewWork,  /* New work item to get a surrogate thread for */
    PPWORKINFO ppOldWork /* Unused; native work items map directly to threads*/
    )
{
  PACTINFO pThd;
  assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);
  /* We should probably abstract this somehow into Condition_CancelWaiter */
  if (Condition_WaiterQueueIsEmpty(&ThreadPoolCond)) {
    /* there are no surrogate threads to work us, so return ourselves for
       rescheduling */
    return (PACTINFO)pNewWork;
  }

  pThd = Condition_WaiterQueueRemoveMostUrgent(&ThreadPoolCond);

  /* There is a surrogate thread; copy over parameters and set it running */
  Activity_ClearWaiting(pThd);
  Activity_SetState(pThd, ACTIVITY_RUNNING);
  Activity_SetThread(pNewWork, pThd);
  Activity_CopyConstraint(pThd, pNewWork);
  Thread_SetWard(pThd, (PACTINFO)pNewWork);
  return (PACTINFO)pThd;
}

PRIVATE void
WorkItem_Exit(
    PIACTIVITY pAct
    )
{
  PACTINFO pThdSurrogate;

  pThdSurrogate = Activity_GetThread(pAct);
  assert(pThdSurrogate != NULL);

  Activity_Exit((PACTINFO)pAct);
}

/******************************************************************************
 * Vtable
 */

CONST_VTBL struct IActivityVtbl WorkItem_Vtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    Activity_QueryInterface,               
    Activity_AddRef,                       
    Activity_Release,                      
    Activity_GetProcess,
    Activity_CreateThread,
    Activity_CreateWorkItem,
    Activity_GetTimeUsage,
    WorkItem_Exit
};

PRIVATE INLINE void
WorkItem_Init(
    PWORKINFO pWork,
    WORK_FUNCTION Func,
    WORK_ARGUMENT Arg,
    WORK_SWITCH_FUNCTION SwitchFunction
    )
{
  Activity_Init((PACTINFO)pWork, CurrentProcess(), &WorkItem_Vtbl);
  WorkItem_SetFunc(pWork, Func);
  WorkItem_SetArg(pWork, Arg);
  WorkItem_SetPseudoThread(pWork);
  if (SwitchFunction == NULL) {
    WorkItem_SetSwitchFunction(pWork, ThreadPool_Switch);
  }
  else {
    WorkItem_SetSwitchFunction(pWork, SwitchFunction);
  }
}

SCODE MCT
Activity_CreateWorkItem( 
    PIACTIVITY pAct,
    PIPROCESS pProcess,
    WORK_FUNCTION pStart,
    WORK_ARGUMENT Arg,
    WORK_SWITCH_FUNCTION SwitchFunction,
    /* [out] */ PIACTIVITY* ppNewActivity
    )
{
  PIHEAP pIHeap = CurrentHeap();
  PWORKINFO pWorkNew;

  /* Allocate a work item descriptor  */
  pWorkNew = (PWORKINFO) pIHeap->v->Alloc(pIHeap,
                                          HEAP_DEBUG_NAME("WorkItem"),
					  sizeof(WORKINFO), 0);
  WorkItem_Init(pWorkNew, pStart, Arg, SwitchFunction);
  Activity_SetRefCount(pWorkNew, 1);

  if (ppNewActivity != NULL) {
    *ppNewActivity = (PIACTIVITY)pWorkNew;
    Activity_SetRefCount(pWorkNew, 2);
  }
  
  return NO_ERROR;
}

