/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* CobMgr cob manager */
#define DBGLVL 3
#include <mmlite.h>
#include <mmmacros.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <base/cons.h>

extern PIBASERTL pTheBaseRtl;


/********************************************************/
/* Implementation of COB namespace                   */
/********************************************************/

/* Original namespace v-table   */
static CONST_VTBL struct INameSpaceVtbl *pRealNsVtbl = NULL;

/* Subclassed namespace v-table */
static struct INameSpaceVtbl CobNsVtbl = {__VTABLE_COMPATIBILITY_FILLER_INITIALIZER NULL,};

/* COB Namespace Object      */
static PINAMESPACE pCobNs = NULL;

/********************************************************/
/* Subclassing of COB methods */
/********************************************************/
SCODE MCT CobMgrCobQueryInterface(PIUNKNOWN This, REFIID Iid, void **ppObject);
UINT  MCT CobMgrCobAddRef(PIUNKNOWN This);
UINT  MCT CobMgrCobRelease(PIUNKNOWN This);

static CONST_VTBL struct IUnknownVtbl CobMgrCobVtbl = 
{
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    CobMgrCobQueryInterface,
    CobMgrCobAddRef,
    CobMgrCobRelease
};

typedef struct _COBINFO
{
    const struct IUnknownVtbl *v;
    PIMODULE pIModule;
    PIUNKNOWN pICob;
} COBINFO;

SCODE MCT CobMgrCobQueryInterface(PIUNKNOWN pThis, REFIID pIid, void **ppObject)
{
    PIUNKNOWN pICob;
    SCODE sc;

    DBGME(0,printf("CobMgrCobQueryInterface"));
    
    pICob = ((COBINFO *)pThis)->pICob;
    sc = pICob->v->QueryInterface(pICob,pIid,ppObject);

    /* If caller asked for IUnknown, replace the real one with with ours */
    if (UuidCmp(&IID_IUnknown, pIid))
    {
        *ppObject = (void *)pThis;
    }

    return sc;
}

UINT MCT CobMgrCobAddRef(PIUNKNOWN pThis)
{
    PIUNKNOWN pICob;

    DBGME(0,printf("CobMgrCobAddRef"));
    
    pICob = ((COBINFO *)pThis)->pICob;
    return pICob->v->AddRef(pICob);
}

UINT MCT CobMgrCobRelease(PIUNKNOWN pThis)
{
    UINT RefCnt;
    COBINFO *pDI;
    PIUNKNOWN pICob;

    DBGME(0,printf("CobMgrCobRelease"));
    
    pDI = (COBINFO *)pThis;
    pICob = pDI->pICob;

    DBGME(1,printf("CobMgrCobRelease: About to call Cob Release() for pDI= 0x%x, Cob 0x%x\n", (ADDRESS) pDI, (ADDRESS) pICob));

    RefCnt = pICob->v->Release(pICob);

    DBGME(2,printf("CobMgrCobRelease: RefCnt = %d\n",RefCnt));
    if (RefCnt==1) {
        PIMODULE pIModule;
        _TCHAR ModuleName[20];
        PIHEAP pIHeap;

        pIModule = pDI->pIModule;
        if (pIModule == NULL) {
            DBGME(1,printf("CobMgrCobRelease: Preloaded COB released. Ignored.\n"));
            /* Take the entry back. We don't really want to unload... */
            /* XXX alternatively search the precobtable and delete the right name */
            return pICob->v->AddRef(pICob);
            
        }
        DBGME(1,printf("CobMgrCobRelease: Removing Cob from namespace\n"));

        pIModule->v->GetName (pIModule,ModuleName,20);

        DBGME(1,printf("CobMgrCobRelease: IModule=0x%x, Cob name is %s\n", (ADDRESS) pDI->pIModule, ModuleName));

        DBGME(1,printf("CobMgrCobRelease: Unregistering Cob from namespace\n"));
        pCobNs->v->Unregister(pCobNs,ModuleName);

        DBGME(1,printf("CobMgrCobRelease: Freeing Cob with IModule->v->Release()\n"));
        RefCnt = pIModule->v->Release(pIModule);
        DBGME(2,printf("CobMgrCobRelease: IModule->v->Release() returned %d\n",RefCnt));

        pIHeap = CurrentHeap();
        pIHeap->v->Free(pIHeap,0,pDI);

        return 0;
    }
    
    return RefCnt;
}

typedef PIUNKNOWN  (*COB_ENTRY)(PIBASERTL);
PIUNKNOWN CreateCobObject(const _TCHAR * CobName)
{
    PIHEAP    pIHeap;
    PIPROCESS pIProcess;
    COBINFO  *pDI;
    COB_ENTRY CobEntry = NULL;
    SCODE sc;

    DBGME(0,printf("CreateCobObject"));
    
    /* Allocate context info */
    DBGME(1,printf("CreateCobObject: Allocating context...\n"));
    pIHeap = CurrentHeap();
    pDI = pIHeap->v->Alloc(pIHeap,HEAP_DEBUG_NAME("cobinfo"),sizeof(COBINFO),0);
    if (pDI==NULL)
    {
        DBGME(3,printf("CreateCobObject: Couldn't allocate memory for Cob context struct\n"));
        return 0;
    }

    DBGME(1,printf("CreateCobObject: Allocated context at 0x%x\n", (ADDRESS) pDI));
    pDI->v = &CobMgrCobVtbl;

    sc = CobPreLookup(CobName, &pDI->pICob);
    if (SUCCEEDED(sc)) {
            pDI->pIModule = NULL;
            goto GotCob;        
    }

#ifndef h8
    /* Load Cob and get its IUnknown interface */
    DBGME(1,printf("CreateCobObject: Loading Cob %s\n",CobName));
    pIProcess = CurrentProcess();
    sc = pIProcess->v->LoadImage(pIProcess, CobName, NULL, 0, &pDI->pIModule);
    if (FAILED(sc)) {
        DBGME(3,printf("CreateCobObject: Cob %s not on disk\n",CobName));
        pIHeap->v->Free(pIHeap,0,pDI);
        return 0;
    }

    DBGME(1,printf("CreateCobObject: Cob %s loaded from disk\n",CobName));

    pDI->pIModule->v->GetEntryPoint(pDI->pIModule,(ADDRESS *)&CobEntry);
    DBGME(1,printf("CreateCobObject: Cob %s entry point=0x%x\n",
         CobName, (ADDRESS) CobEntry));

    pDI->pICob = CobEntry(pTheBaseRtl);
    DBGME(1,printf("CreateCobObject: Successfully invoked entry point 0x%x\n",
         (ADDRESS) CobEntry));
#endif

GotCob:
    DBGME(1,printf("CreateCobObject: Cob %s done with initialization,pICob=0x%x\n",
         CobName, (ADDRESS) pDI->pICob));
    if (pDI->pICob == 0) {
        DBGME(3,printf("CreateCobObject: Cob %s didn't initialize interface\n",CobName));
        if (pDI->pIModule)
            pDI->pIModule->v->Release(pDI->pIModule);
        pIHeap->v->Free(pIHeap,0,pDI);
        return 0;
    }

    return (PIUNKNOWN )pDI;
}

/* Subclass namespace 'bind' to load Cob if not already loaded */
SCODE MCT CobMgrNsBind(PINAMESPACE pThis, const _TCHAR * CobName, NAME_SPACE_FLAGS Flags, PIUNKNOWN *ppUnk)
{
    SCODE sc;
    PIUNKNOWN pUnk;

    DBGME(0,printf("CobMgrNsBind"));
    
    DBGME(1,printf("CobMgrNsBind: Trying to bind to Cob %s\n",CobName));

    /* Attempt to bind to the Cob in the Cob namespace */
    DBGME(1,printf("CobMgrNsBind: Attempting to bind to Cob...\n"));
    sc = pRealNsVtbl->Bind(pThis,CobName,Flags,ppUnk);
    if (sc!=E_NAME_NOT_FOUND)
        return sc;

    /* name not found - look on disk */
    DBGME(1,printf("CobMgrNsBind: Name not found, attempting to load from disk...\n"));
    pUnk = CreateCobObject(CobName);
    if (pUnk==NULL)
    {
        DBGME(3,printf("CobMgrNsBind: Couldn't load from disk, returning sc=0x%x\n",sc));
        return sc;
    }

    /* Register Cob with the Cob namespace */
    DBGME(1,printf("CreateCobObject: Registering Cob %s in Cob namespace as object 0x%x\n", CobName, (ADDRESS) pUnk));
    pRealNsVtbl->Register(pThis,CobName, pUnk, NAME_SPACE_FAILIFEXIST, NULL);
    
    *ppUnk = pUnk;

    DBGME(1,printf("CobMgrNsBind: Returning S_OK, pUnk=0x%x\n", (ADDRESS) pUnk));
    return S_OK;
}

SCODE CobInit(void)
{
    PINAMESPACE pSysNs;                 /* System namespace object */
    SCODE sc;

    DBGME(0,printf("CobInit"));
    
    /* Create a new namespace object for loaded Cobs */
    pSysNs = CurrentNameSpace();
    sc = pSysNs->v->Bind(pSysNs, _T("COB"), NAME_SPACE_CREATE,
                         (PIUNKNOWN *)&pCobNs);
    if (sc!=S_OK)
    {
        DBGME(3,printf("Failed to create COB namespace"));
        return sc;
    }
    
    pRealNsVtbl = pCobNs->v;    /* Get ptr to orig NS vtable       */
    CobNsVtbl   = *pRealNsVtbl; /* Copy it to subclassed NS vtable */
    CobNsVtbl.Bind = CobMgrNsBind; /* Subclass Bind function          */
    pCobNs->v   = &CobNsVtbl;   /* Point NS to subclassed vtable   */

    return S_OK;
}

void CobMgrExit(void)
{
    INameSpace *pSysNs;

    pSysNs = CurrentNameSpace();
    pSysNs->v->Unregister(pSysNs,_T("COB"));
    pCobNs->v = pRealNsVtbl;
    pCobNs->v->Release(pCobNs);

    return;
}

#if 0
#define TESTCob "cobgen.cob"
BOOL TestCobMgr(void)
{
    INameSpace *pSysNs;         /* System namespace object  */
    INameSpace *pCobNs;         /* Cob Namespace         */
    PIUNKNOWN pICob;
    UINT RefCnt;
    SCODE sc;

    DBGME(1,printf("TestCobMgr: Ready to try loading Cob\n"));

    /* Get ptr to system namespace */
    pSysNs = CurrentNameSpace();

    /* Get ptr to Cob namespace */
    DBGME(1,printf("TestCobMgr: System Namespace = 0x%x, binding to Cob namespace\n",pSysNs));
    sc = pSysNs->v->Bind(pSysNs,"Cobs",0,(PIUNKNOWN *)&pCobNs);
    if (sc!=S_OK)
    {
        DBGME(3,printf("TestCobMgr: Failed to bind to Cob namespace\n"));
        return FALSE;
    }

    DBGME(1,printf("TestCobMgr: Cob Namespace = 0x%x, binding to Cob object\n",pCobNs));
    sc = pCobNs->v->Bind(pCobNs,TESTCob,0,(PIUNKNOWN *)&pICob);
    if (sc!=S_OK)
    {
        DBGME(3,printf("TestCobMgr: Failed to bind to Cob interface\n"));
        return FALSE;
    }

    RefCnt = pCobNs->v->Release(pCobNs);
    DBGME(1,printf("TestCobMgr: Released Cob namespace, RefCnt = %d\n",RefCnt));

    DBGME(1,printf("TestCobMgr: Cob object = 0x%x\n",pICob));
    DBGME(1,printf("TestCobMgr: Releasing Cob object, Release at 0x%x\n",pICob->v->Release));
    RefCnt = pICob->v->Release(pICob);
    DBGME(1,printf("TestCobMgr: Released Cob, RefCnt = %d\n",RefCnt));

    return TRUE;
}

int main()
{
    INameSpace *pSysNs;                 /* System namespace object */
    SCODE sc;
    UINT RefCnt;

    CobMgrInit();

    TestCobMgr();

    CobMgrExit();

    return (0);
}
#endif
