/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* cmdline.c
 *
 * Code to parse a "command line" e.g. a string into the
 * standard C-main-style argc,argv pair.
 *
 * Uses dynamic memory, and makes it hard to rid of it.
 */

#include <mmlite.h>
#include <stdlib.h>
#include <tchar.h>

static void *__pCmdLine = NULL;
static void *__pArgv = NULL;

SCODE ParseCmdLine(INT *pArgc, _TCHAR *  **pArgv)
/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
{
    INT  Argc;
    BOOL fInQuote;
    _TCHAR * pCmdLineStart;
    _TCHAR * pCmdLineCur;
    INT  index;
    const _TCHAR * pCmdLine = ProcessArgs();
    const _TCHAR * pImage = ProcessImage();

#if defined(ANGEL_SEMIHOST)
    /* Special debug hack. If we got started weird from emulator..
     */
    if (pImage == NULL) {
        extern _TCHAR * ArmCmdLine(void);
        pCmdLine = ArmCmdLine();
        /* image is included */
    }
#endif

    /* initialize */
    *pArgc = Argc = 0;
    fInQuote = FALSE;

    if ( NULL == pCmdLine || _TEXT('\0') == *pCmdLine ) {
        pCmdLineStart = (_TCHAR *) malloc( sizeof(_TCHAR) );
        goto noargs;
    }

    /* alloc memory for the new command line */
    pCmdLineStart = (_TCHAR *) malloc( (_tcslen(pCmdLine)+1)*sizeof(_TCHAR) );

    /* check allocated memory */
    if ( NULL == pCmdLineStart ) {
        return E_NOT_ENOUGH_MEMORY;
    }

    /* init new command-line and set our 'walking' ptr */
    _tcscpy( pCmdLineStart, pCmdLine );
    pCmdLineCur = pCmdLineStart;
    
    /* parse each argument and insert '\0' after each arg */
    while ( _TEXT('\0') != *pCmdLineCur ) {
        /* eat leading spaces */
        while ( _TEXT(' ') == *pCmdLineCur ||
               _TEXT('\t') == *pCmdLineCur ) {
            ++pCmdLineCur ;
        }

        /* are we at the end */
        if ( _TEXT('\0') == *pCmdLineCur ) {
            break;
        }
    
        /* nope, increment our argument count */
        ++Argc;

        /* walk to end of current arg to insert the NULL if needed */
        while ( *pCmdLineCur && ( _TEXT(' ') != *pCmdLineCur || fInQuote ) ) {

            /* if we are on a quote, toggle flag */
            if ( _TEXT('\"') == *pCmdLineCur ) {
                fInQuote = ( fInQuote ? FALSE : TRUE );
            }

            /* continue to end of this arg */
            ++pCmdLineCur;
        }

        /* if this is the very end, we dont have to set the null */
        if ( _TEXT('\0') == *pCmdLineCur ) {
            break;
        }
        
        /* null-terminate the arg at this postion */
        *pCmdLineCur = _TEXT('\0');
        ++pCmdLineCur;
    }

    /* allocate the argv array and
     * walk the new command-line again and fill in our argv array
     * JOEBAR: alloc Argc+2 for argv[0] + Argc + NULL
     */
noargs:
    *pArgv = (_TCHAR **) malloc( (Argc + 2)*sizeof(_TCHAR *) );

    /* check allocated memory */
    if ( *pArgv == NULL ) {
        /* free our buffer and return */
        free(pCmdLineStart);
        return E_NOT_ENOUGH_MEMORY;
    }

    __pCmdLine = pCmdLineStart;
    __pArgv = *pArgv;

    /* JOEBAR: set Argv[0] to NULL */
    (*pArgv)[0] = NULL;

    /* did we actually find anything */
    /* JOEBAR: we always return pArgv holding argv[0] and trailing NULL */
    if ( 0 == Argc ) {
        goto out;
    }

    pCmdLineCur = pCmdLineStart;
    
    /* eat leading spaces */
    while ( _TEXT(' ') == *pCmdLineCur || _TEXT('\t') == *pCmdLineCur ) {
        ++pCmdLineCur ;
    }

    /* at beginning of first arg */
    /* JOEBAR: index of first is 1, not 0, on account of arg[0] */
    /* set the ptr (argv[1]) */
    (*pArgv)[1] = pCmdLineCur;
    
    for (index=1; index<Argc; index++) {

        /* find next '\0' */
        while ( _TEXT('\0') != *pCmdLineCur ) {
            ++pCmdLineCur;
        }
        ++pCmdLineCur;
        
        /* eat leading spaces which will leave us on the
         * first char of the arg
         */
        while (_TEXT(' ') == *pCmdLineCur || _TEXT('\t') == *pCmdLineCur) {
            ++pCmdLineCur;
        }

        /* at beginning of next arg; set the ptr (argv[index]) */
        /* JOEBAR: add one on account of argv[0] */
        (*pArgv)[index + 1] = pCmdLineCur;
    }

    /* JOEBAR: Argv[Argc+1] is NULL */
out:
    (*pArgv)[Argc + 1] = NULL;
#if defined(ANGEL_SEMIHOST)
    if (pImage == NULL) {
        /* see above */
        Argc--;
        *pArgv = *pArgv + 1;
    } else
#endif
    (*pArgv)[0] = (_TCHAR *) pImage; /* drop the const qualifier */
    *pArgc = Argc + 1; /* Argc args + argv[0] */ 

    return NO_ERROR;
}

SCODE FreeCmdLine( INT *pArgc, _TCHAR * **pArgv )
{
    /* We dont actually use pArgc,pArgv but some othe rimplementation might
     */
    UnusedParameter(pArgc);
    UnusedParameter(pArgv);

    if (__pCmdLine)
        free(__pCmdLine);
    if (__pArgv)
        free(__pArgv);
    __pArgv = __pCmdLine = NULL;

    return NO_ERROR;
}
