/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Portable 64-bit integer math functions for MMLite
 */
#define __NO_BUILTIN_INT64 1
#if defined(MMLITE)
#include <mmlite.h>
#include <machdep.h>
#else
#define INT64 MYINT64
#define UINT64 MYUINT64
#include <windows.h>
#undef INT64
#undef UINT64
typedef struct _INT64 INT64;
typedef struct _UINT64 UINT64;
#define EXTERN_C extern
#include "int64.h"
#endif

/* Cannot (and should not) compile this if there is compiler support.
 * Dont think you can have it both ways because of alignment issues.
 * Someday the unions might get smarter.
 */
/* Well then, we'll give it another shot.  If there is compiler support
 * we'll just make sure the compiler support routines work.  So the code
 * in there is only used by the test programs e.g. tstdiv64.c & friends.
 */

/* Nullify macros for functions we define. If we get compiled
 * it means that those macros are needed as functions.
 */
#undef Int64DividedByInt64
#undef Int64DividedByInt32
#undef Uint64DividedByUint64
#undef Uint64DividedByUint32

#if !defined(__TCS__)
#define MUX(a,b,c)     ((a) ? (b) : (c))
#define izero(a,b) ((a) ? (b) : 0)
#define carry(a,b) (((UINT32)(a) + (UINT32)(b)) < (a))
#define borrow(a,b) ((a) < (b))
#endif

#if defined(_MSC_VER)
#pragma warning(disable:4146)  /* unary minus to unsigned type, result still unsigned */
#endif

extern const UINT64 INT64_MIN;

/* Worker function for all of the div/mod functions.
 * Computes both the result and the remainder of the
 * division of X/Y, returns them by reference.
 * Funny order of arguments to minimize register swaps,
 * esp in the division case (more popular than modulus).
 */
void Uint64DivMod(UINT64 *pq, UINT64 x, UINT64 y, UINT64 *pr)
{
    int i, j, n, x1ne0, y1ne0, nlt32, xgey;
    UINT32 x32, y32;
    UINT64 z;

    /* Check for divide by zero.
     */
    if (!(y.LowPart | y.HighPart)) {
        if (!(x.LowPart | x.HighPart)) {
            z = UINT64_ZERO;
        } else {
            z = INT64_MIN;
        }
        x = UINT64_ZERO;
        goto out;
    }

    /* Calculate n-bit misalignment of dividend y
     * with respect to divisor x.  Align y to x.
     * z = x1.x0 / y1.y0;
     */

    /* Get the non-zero parts. They both exist, see above */
    x1ne0 = (x.HighPart != 0);
    y1ne0 = (y.HighPart != 0);
    x32 = MUX(x.HighPart, x.HighPart, x.LowPart);
    y32 = MUX(y.HighPart, y.HighPart, y.LowPart);

    /* Compute the index of the leftmost bit, into I for x32
     * and into J for y32.
     */
    i = MUX(x32 >> 16, 24, 8);
    j = MUX(y32 >> 16, 24, 8);
    i = MUX(x32 >> i, i+4, i-4);
    j = MUX(y32 >> j, j+4, j-4);
    i = MUX(x32 >> i, i+2, i-2);
    j = MUX(y32 >> j, j+2, j-2);
    i = MUX(x32 >> i, i+1, i-1);
    j = MUX(y32 >> j, j+1, j-1);
    i -= !(x32 >> i);
    j -= !(y32 >> j);

    /* Compute their disalignment, could be negative.
     */
    n = i - j;

    /* Align y (sort of)
     */
    y.HighPart = (y.HighPart << (n&31)) |
                 izero(i != j, y.LowPart >> ((32-n)&31));
    y.LowPart <<= (n&31);

    /* how many bits will we really get
     */
    n += (x1ne0 - y1ne0) << 5;

    /* really align y, swapping if necessary
     */
    nlt32 = (n < 32);
    y.HighPart = MUX(nlt32, y.HighPart, y.LowPart);
    y.LowPart = izero(nlt32, y.LowPart);

    /* Init result */
    z.LowPart = z.HighPart = 0;

    /* Calculate one bit of quotient z per iteration.
     */
    while (n-- >= 0) {

        xgey = MUX(x.HighPart != y.HighPart, x.HighPart > y.HighPart, x.LowPart >= y.LowPart);
        z.HighPart += z.HighPart + carry(z.LowPart, z.LowPart);
        z.LowPart += z.LowPart + xgey;
        x.HighPart -= izero(xgey, y.HighPart + borrow(x.LowPart, y.LowPart));
        x.LowPart -= izero(xgey, y.LowPart);
        y.LowPart = (y.LowPart >> 1) | (y.HighPart << 31);
        y.HighPart >>= 1;
    }

 out:
    /* Quotient, if desired
     */
    if (pq)
        *pq = z;
    /* Remainder, if desired
     */
    if (pr)
        *pr = x;
}

/* Same as above, only in signed integer terms.
 * At least one compiler uses this directly.
 */
void Int64DivMod(INT64 *pq, INT64 x, INT64 y, INT64 *pr)
{
    int sign, xneg, yneg;
    UINT64 ux, uy, uz, ur;
    INT64 z, r;

    /* Remember sign of result.
     */
    sign = (x.HighPart ^ y.HighPart) >> 31;

    /* Take absolute values of dividend x and divisor y.
     */
    xneg = (x.HighPart < 0);
    ux.HighPart = MUX(xneg, ~x.HighPart + !x.LowPart, x.HighPart);
    ux.LowPart = MUX(xneg, -x.LowPart, x.LowPart);
    yneg = (y.HighPart < 0);
    uy.HighPart = MUX(yneg, ~y.HighPart + !y.LowPart, y.HighPart);
    uy.LowPart = MUX(yneg, -y.LowPart, y.LowPart);

    /* Perform the division, cast the results.
     */
    Uint64DivMod(&uz,ux,uy,&ur);
    z.LowPart = uz.LowPart;
    z.HighPart = (UINT32) uz.HighPart;
    r.LowPart = ur.LowPart;
    r.HighPart = (UINT32) ur.HighPart;

    /* If necessary, negate quotient/remainder
     */
    z.HighPart = MUX(sign, ~z.HighPart + !z.LowPart, z.HighPart);
    z.LowPart = MUX(sign, -z.LowPart, z.LowPart);
    r.HighPart = MUX(xneg, ~r.HighPart + !r.LowPart, r.HighPart);
    r.LowPart = MUX(xneg, -r.LowPart, r.LowPart);

    /* Quotient, if desired
     */
    if (pq)
        *pq = z;
    /* Remainder, if desired
     */
    if (pr)
        *pr = r;
}

/* MMLite runtime library function -- Divide an unsigned, 64-bit
 * integer by another unsigned, 64-bit integer.  The first
 * argument is the dividend and the second is the divisor.
 * The return value (the quotient) is an unsigned 64-bit integer.
 */
UINT64 Uint64DividedByUint64(UINT64 x, UINT64 y)
{
    UINT64 z;
    Uint64DivMod(&z,x,y,NULL);
    return z;
}

/* MMLite runtime library function -- Divide an unsigned, 64-bit
 * integer by an unsigned, 32-bit integer.  The first argument
 * is the dividend and the second is the divisor.  The
 * return value (the quotient) is an unsigned 64-bit integer.
 */
UINT64 Uint64DividedByUint32(UINT64 x, UINT32 y)
{
    UINT64 y64;

    y64.LowPart = y;
    y64.HighPart = 0;
    return Uint64DividedByUint64(x, y64);
}


/* MMLite runtime library function -- Divide a signed, 64-bit
 * integer by another signed, 64-bit integer.  The first
 * argument is the dividend and the second is the divisor.
 * The return value (the quotient) is a signed 64-bit integer.
 */
INT64 Int64DividedByInt64(INT64 x, INT64 y)
{
    INT64 z;
    Int64DivMod(&z,x,y,NULL);
    return z;
}

/* MMLite runtime library function -- Divide a signed, 64-bit
 * integer by a signed, 32-bit integer.  The first argument
 * is the dividend and the second is the divisor.  The
 * return value (the quotient) is a signed 64-bit integer.
 */
INT64 Int64DividedByInt32(INT64 x, INT32 y)
{
    INT64 y64;

    y64.LowPart = y;
    y64.HighPart = y >> 31;
    return Int64DividedByInt64(x, y64);
}

