/* Copyright (c) Microsoft Corporation. All rights reserved. */
        #include <arm_asm.h>

        _CODE(C$$code)

/*
 * Source code for 64-bit integer support routines
 *
 * C type signatures for these functions are:
 *
 *  INT64 Int64Add(INT64 a, INT64 b);
 *  INT64 Int64Subtract(INT64 a, INT64 b);
 *  INT64 Int64LShift(INT64 a, INT32 n);            left shift
 *  INT64 Int64RShift(INT64 a, INT32 n);            signed right shift
 *  UINT64 Uint64RShift(UINT64 a, INT32 n);         unsigned right shift
 *  INT64 Int64TimesInt32( INT64 a, INT32 b);       multiply 64x32
 *  INT64 Int64TimesInt64( INT64 a, INT64 b);       multiply 64x64
 *  INT64 Int64DividedByInt32( INT64 a, INT32 b);   divide 64x32
 *  INT64 Int64DividedByInt64( INT64 a, INT64 b);   divide 64x64
 *
 * The ARM C compiler uses the following conventions for handling
 * int64 data, which we declared as a structure of two int32.
 *
 *       - If a function returns an int64, it does so via a pointer.
 *
 *       - A pointer to the return value is passed to the callee, in
 *         register R0 (aka a1).  This is the first register-argument, hence
 *         the logic here is to add one extra hidden argument (at front).
 *         Note this same register is used as the/any function's result.
 *
 *       - The caller (de)allocates room for the return value on the stack.
 *
 *       - If an argument is an int64 it is passed inline in registers,
 *         when appropriate.  So if a function takes as its sole argument
 *         an int64 it will be passed in registers R0/R1 (a1/a2)).
 *
 *       - It is possible for an int64 to be passed half in register R3 (a4)
 *         and half on the stack (sp+0).
 *
 * Finally, remember that the HighPart of an int64 is the one at [ptr]+4
 *
 */

        _EXPORT(Int64Subtract)
        _EXPORT(Int64Add)
        _EXPORT(Int64AddInt32)
        _EXPORT(Uint64SubtractUint32)
        _EXPORT(Uint64AddUint32)
        _EXPORT(Int64LShift)
        _EXPORT(Int64RShift)
        _EXPORT(Uint64RShift)
        _EXPORT(Int64TimesInt32)
        _EXPORT(Int64DividedByInt32)
        _EXPORT(Int64DividedByInt64)
         _EXPORT(__divdi3)         /* alias for GCC when builtin */
        _EXPORT(Int64TimesInt64)
         _EXPORT(__muldi3)         /* alias for GCC when builtin */

/*
 * Entry point for Int64Add
 */
_FUNCTION(Int64Add)
        /*
         * Pick up HighPart from stack.
         * In most cases this hits in the cache. Even if it misses,
         * better issue it now before we start the ADD.
         */
        ldr     r12,[sp,#0]

        /*
         * Add LowParts, set carry.  Reuse r1 cuz scratch
         */
        adds    r1,r1,r3

        /*
         * Nothing else to do here, everything is interlocked.
         * So wait for everything and issue the ADC.
         * We dont care for the resulting flags.
         */
        adc     r2,r2,r12

        /*
         * Wait for the ADC to complete, save results.
         */
        stmia   r0,{r1,r2}

        /*
         * Done.
         */
        mov     pc,lr


/*
 * Entry point for Int64AddInt32
 */
_FUNCTION(Int64AddInt32)
        /*
         * Add LowParts, set carry.  Reuse r1 cuz scratch
         */
        adds    r1,r1,r3

        /*
         * Sign extend the int32 to an int64 and issue the ADC.
         * We dont care for the resulting flags.
         */
        adc     r2,r2,r3,asr #32

        /*
         * Wait for the ADC to complete, save results.
         */
        stmia   r0,{r1,r2}

        /*
         * Done.
         */
        mov     pc,lr


/*
 * Entry point for Uint64AddUint32
 */
_FUNCTION(Uint64AddUint32)
        /*
         * Add LowParts, set carry.  Reuse r1 cuz scratch
         */
        adds    r1,r1,r3

        /*
         * Add in the carry.
         * We dont care for the resulting flags.
         */
        adc     r2,r2,#0

        /*
         * Wait for the ADC to complete, save results.
         */
        stmia   r0,{r1,r2}

        /*
         * Done.
         */
        mov     pc,lr


/*
 * Entry point for Int64Subtract
 */
_FUNCTION(Int64Subtract)
        /*
         * Pick up HighPart from stack.
         * In most cases this hits in the cache. Even if it misses,
         * better issue it now before we start the SUB.
         */
        ldr     r12,[sp,#0]

        /*
         * Subtract LowParts, set borrow(carry).  Reuse r1 cuz scratch
         */
        subs    r1,r1,r3

        /*
         * Nothing else to do here, everything is interlocked.
         * So wait for everything and issue the SBC.
         * We dont care for the resulting flags.
         */
        sbc     r2,r2,r12

        /*
         * Wait for the SBC to complete, save results.
         */
        stmia   r0,{r1,r2}

        /*
         * Done.
         */
        mov     pc,lr


/*
 * Entry point for Uint64SubtractUint32
 */
_FUNCTION(Uint64SubtractUint32)
        /*
         * Subtract LowParts, set borrow(carry).  Reuse r1 cuz scratch
         */
        subs    r1,r1,r3

        /*
         * Subtract the borrow.
         * We dont care for the resulting flags.
         */
        sbc     r2,r2,#0

        /*
         * Wait for the SBC to complete, save results.
         */
        stmia   r0,{r1,r2}

        /*
         * Done.
         */
        mov     pc,lr

/*
 * Entry point for Int64LShift
 */

_FUNCTION(Int64LShift)

        cmp      r3,#63          /* is n > 63 ? */
        movhi    r3,#64          /* if so, set n = 64 */

        cmp      r3,#32          /* is n > 32 ? */
        bls      Le32            /* jump if n <= 32 */

        mov      r2,r1           /* val.msbs = val.lsbs */
        mov      r2,#0           /* val.lsbs = 0 */
        sub      r3,r3,#32       /* n -= 32 */

_LABEL(Le32)
        rsb      r12,r3,#32       /* 32 - n */
        mov      r2,r2, lsl r3   /* val.msbs <<= n */
        orr      r2,r2,r1, lsr r12  /* val.msbs |= val.lsbs>>32-n */
        mov      r1,r1, lsl r3   /* val.lsbs <<= n */
        stmia    r0,{r1,r2}      /* save return value = val */

        mov     pc,lr


/*
 * Entry point for Int64RShift
 */

_FUNCTION(Int64RShift)

        cmp      r3,#63          /* is n > 63 ? */
        movhi    r3,#64          /* if so, set n = 64 */

        cmp      r3,#32          /* is n > 32 ? */
        bls      Le32a           /* jump if n <= 32 */

        mov      r1,r2           /* val.lsbs = val.msbs */
        mov      r2,r2,asr #31   /* val.msbs = 32x sign bit */
        sub      r3,r3,#32       /* n -= 32 */

_LABEL(Le32a)
        rsb      r12,r3,#32       /* 32 - n */
        mov      r12,r2,lsl r12    /* tmp = val.msbs << 32-n */
        orr      r1,r12,r1,lsr r3 /* val.lsbs = tmp|(val.lsbs>>n) */
        mov      r2,r2,asr r3    /* val.msbs >>= n */
        stmia    r0,{r1,r2}      /* save return value = val */

        mov     pc,lr


/*
 * Entry point for Uint64RShift
 */

_FUNCTION(Uint64RShift)

        cmp      r3,#63          /* is n > 63 ? */
        movhi    r3,#64          /* if so, set n = 64 */

        cmp      r3,#32          /* is n > 32 ? */
        bls      Le32b           /* jump if n <= 32 */

        mov      r1,r2           /* val.lsbs = val.msbs */
        mov      r2,#0           /* val.msbs = 0 */
        sub      r3,r3,#32       /* n -= 32 */

_LABEL(Le32b)
        rsb      r12,r3,#32       /* 32 - n */
        mov      r12,r2,lsl r12    /* tmp = val.msbs << 32-n */
        orr      r1,r12,r1,lsr r3 /* val.lsbs = tmp|(val.lsbs>>n) */
        mov      r2,r2,lsr r3    /* val.msbs >>= n */
        stmia    r0,{r1,r2}      /* save return value = val */

        mov     pc,lr


/*
 * Entry point for Int64TimesInt32
 */
_FUNCTION(Int64TimesInt32)

       stmdb  sp!,{r4-r7,lr}      /* save registers on stack */

       rsb    r5,r1,#0            /* z.msbs = -x.lsbs */
       and    r5,r5,r3, asr #31   /* if y < 0, z.msbs = 0 */

       mov    r6,r1,lsl #16
       mov    r6,r6,lsr #16       /* x0 = x.lsbs & 0xffff */
       mov    r7,r3,lsl #16
       mov    r7,r7,lsr #16       /* y0 = y.lsbs & 0xffff */
       mul    r4,r6,r7            /* z.lsbs = x0*y0 */

       mov    r1,r1,lsr #16       /* x1 = x.lsbs >> 16 */
       mov    r3,r3,lsr #16       /* y1 = y.lsbs >> 16 */
       mla    r5,r1,r3,r5         /* z.msbs += x1*y1 */

       mul    lr,r1,r7            /* tmp = x1*y0 */
       adds   r4,r4,lr,lsl #16    /* z.lsbs += tmp << 16 */
       adc    r5,r5,lr,lsr #16    /* z.msbs += tmp >> 16 */

       mul    lr,r6,r3            /* tmp = x0*y1 */
       adds   r4,r4,lr,lsl #16    /* z.lsbs += tmp << 16 */
       adc    r5,r5,lr,lsr #16    /* z.msbs += tmp >> 16 */

       mov    lr,r2,lsl #16
       mov    lr,lr,lsr #16       /* x2 = x.msbs & 0xffff */
       mla    r5,lr,r7,r5         /* z.msbs += x2*y0 */
       mul    lr,r3,lr            /* tmp = x2*y1 */

       mov    r2,r2,lsr #16       /* x3 = x.msbs >> 16 */
       mla    lr,r2,r7,lr         /* tmp += x3*y0 */
       add    r5,r5,lr,lsl #16    /* z.msbs += tmp << 16 */

       stmia  r0,{r4,r5}          /* set return value = z */
       ldmia  sp!,{r4-r7,pc}      /* return from subroutine */


/*
 * Entry point for Int64DividedByInt32
 */

_FUNCTION(Int64DividedByInt32)

       stmdb  sp!,{r4-r8,lr}  /* save registers on stack */
       mov    r4,r3, asr #31  /* extend y (arg2) sign bits */
       b      DivCommon       /* jump to common divide routine */

/*
 * Entry point for Int64DividedByInt64
 */

_FUNCTION(Int64DividedByInt64)
_FUNCTION(__divdi3)

       stmdb  sp!,{r4-r8,lr}  /* save registers on stack */
       ldr    r4,[sp,#24]     /* load y.msbs (32 MSBs of arg2) */

_LABEL(DivCommon)
       eor    lr,r2,r4        /* remember sign of result */
       stmdb  sp!,{r0,lr}     /* save sign and ret val pointer */

/*
 * If divide by zero, handle as special case.
 */
       orrs   r5,r4,r3        /* is divisor y == 0 ? */
       beq    DivByZero       /* if so, divide by zero */

/*
 * Convert dividend x and divisor y to positive numbers
 * so that division operation will be unsigned.
 */
       adds   r2,r2,#0        /* check sign of x */
       bge    XNotNegative    /* jump if x >= 0 */

       rsbs   r1,r1,#0        /* negate x.lsbs */
       rsc    r2,r2,#0        /* negate x.msbs */
_LABEL(XNotNegative)

       adds   r4,r4,#0        /* check sign of y */
       bge    YNotNegative    /* jump if y >= 0 */

       rsbs   r3,r3,#0        /* negate y.lsbs */
       rsc    r4,r4,#0        /* negate y.msbs */
_LABEL(YNotNegative)

/*
 * If dividend x < divisor y, return quotient z = 0.
 */
       subs   r5,r1,r3        /* x.lsbs - y.lsbs */
       sbcs   r5,r2,r4        /* x.msbs - y.msbs */
       blo    ReturnZero      /* if x < y, return value z = 0 */

/*
 * Align MSB of divisor y to MSB of dividend x by shifting y.
 */
       mov     r5,#32         /* i = 32 (assume x.msbs != 0) */
       movs    r6,r2          /* is x.msbs == 0 ? */
       moveq   r5,#0          /* if (x.msbs == 0) i = 0 */
       moveq   r6,r1          /* if (x.msbs == 0) w = x.lsbs */
       bl      LeftMostBit    /* i = bit # of leftmost 1 in x */
       mov     r0,r5          /* save i */

       mov     r5,#32         /* j = 32 (assume y.msbs != 0) */
       movs    r6,r4          /* is y.msbs == 0 ? */
       moveq   r5,#0          /* if (y.msbs == 0) j = 0 */
       moveq   r6,r3          /* if (y.msbs == 0) w = y.lsbs */
       bl      LeftMostBit    /* j = bit # of leftmost 1 in y */

       rsb     r0,r5,r0       /* left shift count n = i - j */
       cmp     r0,#32         /* is shift count n >= 32 ? */
       movge   r4,r3          /* shift y.lsbs into y.msbs */
       movge   r3,#0          /* shift 0s into y.lsbs */
       and     r5,r0,#31      /* left shift count = n & 31 */
       rsb     r6,r5,#32      /* right shift count = 32-(n&31) */
       mov     r4,r4, lsl r5  /* align y.msbs to x.msbs */
       orr     r4,r4,r3, lsr r6  /* merge two parts of y.msbs */
       mov     r3,r3, lsl r5  /* align y.lsbs to x.lsbs */

/*
 * Use nonrestoring division algorithm.  Each iteration processes
 * 4 bits at a time to reduce effective loop overhead.  Enter
 * loop at one of four starting points, depending on remainder
 * of n/4, where n = loop count - 1.
 */
       mov    r5,#0           /* initialize z.lsbs = 0 */
       mov    r6,#0           /* initialize z.msbs = 0 */

       ands    r7,r0,#3       /* n & 3 = remainder of n/4 */
       beq     Sub3A          /* jump if remainder = 0 */
       cmp     r7,#1          /* is remainder == 1 ? */
       beq     Sub2A          /* jump if remainder == 1 */
       cmp     r7,#2          /* is remainder == 2 ? */
       beq     Sub1A          /* jump if remainder == 2 */

/*
 * The main loop for the nonrestoring division algorithm follows
 * below.  The loop body has two halves, designated by labels
 * "SubFrom" and "AddBack".  Remain in the SubFrom half of the
 * loop as long as accumulator value x remains nonnegative.  Any
 * time x becomes negative, jump to corresponding position in
 * AddBack half of loop (see below).
 */
_LABEL(SubFrom)
       mov     r6,r6, lsl #4  /* z.msbs <<= 4 */
       orr     r6,r6,r5, lsr #28  /* z.msbs |= z.lsbs >> 28 */
       mov     r5,r5, lsl #4  /* z.lsbs <<= 4 */
_LABEL(Sub0A)
       subs    r1,r1,r3       /* x.lsbs -= y.lsbs */
       sbcs    r2,r2,r4       /* x.msbs -= y.msbs */
       add     r5,r5,#8       /* z.lsbs -= 8 */
       blo     Add1           /* jump if x >= 0 */
_LABEL(Sub1)
       adds    r1,r1,r1       /* x.lsbs <<= 1 */
       adc     r2,r2,r2       /* x.msbs <<= 1 */
_LABEL(Sub1A)
       subs    r1,r1,r3       /* x.lsbs -= y.lsbs */
       sbcs    r2,r2,r4       /* x.msbs -= y.msbs */
       add     r5,r5,#4       /* z.lsbs -= 4 */
       blo     Add2           /* jump if x >= 0 */
_LABEL(Sub2)
       adds    r1,r1,r1       /* x.lsbs <<= 1 */
       adc     r2,r2,r2       /* x.msbs <<= 1 */
_LABEL(Sub2A)
       subs    r1,r1,r3       /* x.lsbs -= y.lsbs */
       sbcs    r2,r2,r4       /* x.msbs -= y.msbs */
       add     r5,r5,#2       /* z.lsbs -= 2 */
       blo     Add3           /* jump if x >= 0 */
_LABEL(Sub3)
       adds    r1,r1,r1       /* x.lsbs <<= 1 */
       adc     r2,r2,r2       /* x.msbs <<= 1 */
_LABEL(Sub3A)
       subs    r1,r1,r3       /* x.lsbs -= y.lsbs */
       sbcs    r2,r2,r4       /* x.msbs -= y.msbs */
       add     r5,r5,#1       /* z.lsbs -= 1 */
       blo     Add4           /* jump if x >= 0 */
_LABEL(Sub4)
       adds    r1,r1,r1       /* x.lsbs <<= 1 */
       adc     r2,r2,r2       /* x.msbs <<= 1 */
       subs    r0,r0,#4       /* n -= 4 */
       bge     SubFrom        /* loop while n >= 0 */
       b       DivDone        /* divide is complete */

/*
 * This is the second half of the division loop.  Remain
 * in the AddBack half below as long as accumulator value
 * x remains negative. Any time x becomes nonnegative,
 * jump to corresponding position in SubFrom half above.
 */
_LABEL(AddBack)
       mov     r6,r6, lsl #4  /* z.msbs <<= 4 */
       orr     r6,r6,r5, lsr #28 /* z.msbs |= z.lsbs >> 28 */
       mov     r5,r5, lsl #4  /* z.lsbs <<= 4 */
_LABEL(Add0)
       adds    r1,r1,r3       /* x.lsbs += y.lsbs */
       adcs    r2,r2,r4       /* x.msbs += y.msbs */
       sub     r5,r5,#8       /* z.lsbs -= 8 */
       bhs     Sub1           /* jump if x >= 0 */
_LABEL(Add1)
       adds    r1,r1,r1       /* x.lsbs <<= 1 */
       adc     r2,r2,r2       /* x.msbs <<= 1 */
       adds    r1,r1,r3       /* x.lsbs += y.lsbs */
       adcs    r2,r2,r4       /* x.msbs += y.msbs */
       sub     r5,r5,#4       /* z.lsbs -= 4 */
       bhs     Sub2           /* jump if x >= 0 */
_LABEL(Add2)
       adds    r1,r1,r1       /* x.lsbs <<= 1 */
       adc     r2,r2,r2       /* x.msbs <<= 1 */
       adds    r1,r1,r3       /* x.lsbs += y.lsbs */
       adcs    r2,r2,r4       /* x.msbs += y.msbs */
       sub     r5,r5,#2       /* z.lsbs -= 2 */
       bhs     Sub3           /* jump if x >= 0 */
_LABEL(Add3)
       adds    r1,r1,r1       /* x.lsbs <<= 1 */
       adc     r2,r2,r2       /* x.msbs <<= 1 */
       adds    r1,r1,r3       /* x.lsbs += y.lsbs */
       adcs    r2,r2,r4       /* x.msbs += y.msbs */
       sub     r5,r5,#1       /* z.lsbs -= 1 */
       bhs     Sub4           /* jump if x >= 0 */
_LABEL(Add4)
       adds    r1,r1,r1       /* x.lsbs <<= 1 */
       adc     r2,r2,r2       /* x.msbs <<= 1 */
       subs    r0,r0,#4       /* n -= 4 */
       bge     AddBack        /* loop while n > 0 */
/*
 * Quotient z is almost complete, but is slightly
 * off because it is currently truncated toward +z
 * direction.  Adjust by truncating toward -z.
 */
       subs    r5,r5,#1       /* z.lsbs */
       sbc     r6,r6,#0       /* propagate into z.msbs */

/*
 * Division operation complete.  If necessary, fix up sign
 * of quotient.  Then save quotient and return to caller.
 */
_LABEL(DivDone)
       ldmia   sp!,{r0,lr}    /* restore sign, ret val pointer */
       adds    lr,lr,#0       /* test sign of quotient */
       bge     SignOK         /* jump if quotient nonnegative */

       rsbs    r5,r5,#0       /* negate z.lsbs */
       rsc     r6,r6,#0       /* negate z.msbs */
_LABEL(SignOK)

       stmia   r0,{r5,r6}     /* save return value z */
       ldmia   sp!,{r4-r8,pc} /* restore registers, return */

/*
 * Return a value of zero because divisor > dividend.
 */
_LABEL(ReturnZero)

       mov     r5,#0          /* x.lsbs = 0 */
       mov     r6,#0          /* x.msbs = 0 */
       b       DivDone        /* jump to return code */

/*
 * Leftmost-one function returns bit number of leftmost 1
 * in specified 32-bit value.  At entry, specified value
 * resides in r6 and 5 LSBs of r5 are cleared to hold
 * 5-bit return value (bit number of leftmost 1 in r6).
 */
_LABEL(LeftMostBit)
       mov     r7,#16         /* k = 16 */

_LABEL(LMOLoop)
       movs    r8,r6, lsr r7  /* is (val >> k) != 0 ? */
       movne   r6,r8          /* if so, val >>= k */
       orrne   r5,r5,r7       /* if so, n |= k */
       movs    r7,r7, lsr #1  /* k >>= 1 */
       bne     LMOLoop        /* loop while k > 0 */

       mov     pc,lr          /* return n in r5 */

/*
 * Handle divide by zero as special case.
 */
_FUNCTION(DivByZero)
       orrs    r6,r4,r3       /* is dividend x == 0 ? */
       movne   r6,#0x80000000 /* if x != 0, set z = "infinity" */
       b       DivDone        /* jump to return code */


/*
 * Entry point for Int64TimesInt64
 */
_FUNCTION(Int64TimesInt64)
_FUNCTION(__muldi3)

       stmdb  sp!,{r4-r8,lr}      /* save registers on stack */

       mov    r7,r1,lsl #16
       mov    r7,r7,lsr #16       /* x0 = x.lsbs & 0xffff */
       mov    r8,r3,lsl #16
       mov    r8,r8,lsr #16       /* y0 = y.lsbs & 0xffff */
       mul    r5,r7,r8            /* z.lsbs = x0*y0 */

       mov    r1,r1,lsr #16       /* x1 = x.lsbs >> 16 */
       mov    r3,r3,lsr #16       /* y1 = y.lsbs >> 16 */
       mul    r6,r1,r3            /* z.msbs = x1*y1 */

       mul    lr,r1,r8            /* tmp = x1*y0 */
       adds   r5,r5,lr,lsl #16    /* z.lsbs += tmp << 16 */
       adc    r6,r6,lr,lsr #16    /* z.msbs += tmp >> 16 */

       mul    lr,r7,r3            /* tmp = x0*y1 */
       adds   r5,r5,lr,lsl #16    /* z.lsbs += tmp << 16 */
       adc    r6,r6,lr,lsr #16    /* z.msbs += tmp >> 16 */

       mov    lr,r2,lsl #16
       mov    lr,lr,lsr #16       /* x2 = x.msbs & 0xffff */
       mla    r6,lr,r8,r6         /* z.msbs += x2*y0 */
       mul    lr,r3,lr            /* tmp = x2*y1 */

       ldr    r4,[sp,#24]         /* load y.msbs */
       mov    r3,r4,lsl #16
       mov    r3,r3,lsr #16       /* y2 = y.msbs & 0xffff */
       mla    r6,r7,r3,r6         /* z.msbs += x0*y2 */
       mla    lr,r1,r3,lr         /* tmp += x1*y2 */

       mov    r2,r2,lsr #16       /* x3 = x.msbs >> 16 */
       mla    lr,r2,r8,lr         /* tmp += x3*y0 */
       mov    r4,r4,lsr #16       /* y3 = y.msbs >> 16 */
       mla    lr,r7,r4,lr         /* tmp += x0*y3 */
       add    r6,r6,lr,lsl #16    /* z.msbs += tmp << 16 */

       stmia  r0,{r5,r6}          /* set return value = z */
       ldmia  sp!,{r4-r8,pc}      /* return from subroutine */

       _END
