/* Copyright (c) Microsoft Corporation. All rights reserved. */
	#include <arm_asm.h>

/*
 * MMLITE versions of string functions memcmp, memcpy,
 * memmove, and memset.  Written in ARM assembly language.
 * Similar to standard ANSI C string.h functions.
 *
 * History:
 *   01/26/96...Original version written.......... J. Van Aken
 *   01/26/96...Consolidated byte-serial loops.... A. Raffman
 */

        _EXPORT(memcmp)
	_EXPORT(__rt_memcmp)
        _EXPORT(memcpy)
        _EXPORT(_memcpy)
	_EXPORT(__rt_memcpy)
	_EXPORT(__rt_memcpy_w)
        _EXPORT(memmove)
        _EXPORT(__rt_memmove)
	_EXPORT(__rt_memclr)
	_EXPORT(__rt_memclr_w)
	_EXPORT(memset)
	_EXPORT(__rt_memset)

/*
 * Compare n bytes of string p1 to string p2.
 *
 * int memcmp(const void *p1, const void *p2, int n)
 */

	_CODE(C$$code)

_FUNCTION(__rt_memcmp)
_FUNCTION(memcmp)
        stmdb    sp!,{r4-r6,lr}    /* save registers on stack */
        mov      r3,r0             /* copy 1st arg, p1 */
        mov      r0,#0             /* set return value = 0 */
        teq      r2,#0             /* is n == 0 ? */
        ldmeqia  sp!,{r4-r6,pc}    /* return if n == 0 */

        cmp      r2,#12            /* is n > 12 */
        ble      ByteLoop          /* if n <= 12, jump */

        sub      r0,r3,r1          /* p1 - p2 */
        ands     r0,r0,#3          /* (p1 - p2) & 3 */
        beq      WordCompare       /* jump if byte offsets match */

/*
 * The strings begin at different byte offsets WRT word boundaries.
 * Loop below processes only a single pair of bytes per iteration.
 */
_LABEL(ByteLoop)
        ldrb     r0,[r3],#1        /* b1 = next byte from string 1 */
        ldrb     r4,[r1],#1        /* b2 = next byte from string 2 */
        subs     r0,r0,r4          /* b1 - b2 */
        bne      ByteBreak         /* if b1 != b2, break out of loop */
        subs     r2,r2,#1          /* --n (decrement loop counter) */
        bne      ByteLoop          /* loop again if n > 0 */

_LABEL(ByteBreak)
        ldmia    sp!,{r4-r6,pc}    /* return to caller (ret val = b1-b2) */

/*
 * The two strings have same starting byte alignment WRT word boundary.
 * Set up inner loop that compares a pair of words per iteration.
 */
_LABEL(WordCompare)
        add      r5,r3,r2          /* e1 = p1 + n (point to trailing byte) */
        and      lr,r3,#3          /* align = p1 & 3 (initial byte offset) */
        bic      r3,r3,#3          /* p1 &= ~3 (point to word boundary) */
        bic      r1,r1,#3          /* p2 &= ~3 */
        add      r2,r5,#3          /* e1 + 3 */
        sub      r2,r2,r3          /* e1 + 3 - p1 */
        mov      r2,r2,lsr #2      /* nWords = (e1 + 3 - p1) >> 2 */


        mvn      r6,#0             /* initialize mask to all 1s */
        mov      r0,lr,lsl #3      /* convert byte offset to bit offset */
        mov      r6,r6,lsl r0      /* poke holes in mask for invalid bytes */
        ldr      r0,[r3],#4        /* w1 = *p1++ */
        and      r0,r0,r6          /* isolate starting bytes in 1st string */
        ldr      r4,[r1],#4        /* w2 = *p2++ */
        and      r4,r4,r6          /* isolate starting bytes in 2nd string */

/*
 * Inner loop:  Compare the two strings one word at a time to look for
 * a mismatch.  If the two strings match, return 0.
 */
_LABEL(WordLoop)
        subs     r0,r0,r4          /* w1 - w2 */
        bne      WordBreak         /* if w1 != w2, break out of loop */
        ldr      r0,[r3],#4        /* w1 = *p1++ */
        subs     r2,r2,#1          /* --nWords */
        ldr      r4,[r1],#4        /* w2 = *p2++ */
        bne      WordLoop          /* loop again if more words in string */

        mov      r0,#0             /* set return argument = 0 */
        ldmia    sp!,{r4-r6,pc}    /* all done -- return to caller */

/*
 * The strings may still match if the apparent mismatch happened in
 * the final pair of words from the two strings (in trailing bytes).
 */
_LABEL(WordBreak)
        teq      r2,#1             /* nWords == 1? (mismatch at EOS?) */
        bne      FindMismatch      /* jump if nWords != 1 */

        ands     r5,r5,#3          /* is trailing byte word-aligned? */
        beq      FindMismatch      /* jump if word-aligned (real mismatch) */

        mvn      r6,#0             /* initialize mask to all 1s */
        mov      r5,r5,lsl #3      /* convert byte offset to bit offset */
        mvn      r6,r6,lsl r5      /* poke holes in mask for invalid bytes */
        ands     r0,r0,r6          /* mask off trailing bytes, string 1 */
        ldmeqia  sp!,{r4-r6,pc}    /* if w1 == w2, return val = 0 */

/*
 * We detected a mismatch in the current pair of words from the strings.
 * But in which byte position within the words did the mismatch occur?
 */
_LABEL(FindMismatch)
        add      r3,r0,r4          /* restore value w1 */

_LABEL(NextByte)
        and      r0,r3,#0xff       /* b1 = w1 & 0xff (isolate byte) */
        and      r2,r4,#0xff       /* b2 = w2 & 0xff */
        subs     r0,r0,r2          /* return val = b1 - b2 ? */

        ldmneia  sp!,{r4-r6,pc}    /* if val != 0, return to caller */

        mov      r3,r3,lsr #8      /* w1 >>= 8 (position next byte) */
        mov      r4,r4,lsr #8      /* w2 >>= 8 */
        b        NextByte          /* if b1 != b2, loop again */



/*
 * Fast copy n bytes from source p2 to destination p1.
 *
 *     void *memcpy(void *p1, const void *p2, int n)
 */

_FUNCTION(__rt_memcpy)
_FUNCTION(__rt_memcpy_w)
_FUNCTION(memcpy)
_FUNCTION(_memcpy)
        teq      r2,#0       	/* is arg n == 0 ? */
        moveq    pc,lr       	/* if n == 0, return */

        stmdb    sp!,{lr}    	/* push return address */
        mov      r12,r0       	/* copy pointer p1 */
        cmp      r2,#0x8     	/* is string long or short? */
        ble      ByteSerial  	/* jump if long string */

        sub      r3,r0,r1    	/* compare pointers p1, p2 */
        tst      r3,#3       	/* strings aligned same? */
        bne      ByteSerial  	/* jump if strings not aligned */

/*
 * Both strings are similarly aligned WRT word boundaries.
 * At least a portion of the data can be copied an entire
 * word at a time, which is faster than copying bytes.
 */
_LABEL(WordSerial)
        ands     r3,r0,#3    	/* check byte alignment */
        beq      WordAligned 	/* jump if p1, p2 word-aligned */

        rsb      r3,r3,#4    	/* m = no. of odd initial bytes */
        sub      r2,r2,r3    	/* n = n - m */

/*
 * If the two strings do not begin on word boundaries, begin
 * by copying the odd bytes that precede the first full word.
 */
_LABEL(PreLoop)
        ldrb     lr,[r1],#1  	/* read byte from string 2 */
        subs     r3,r3,#1    	/* --m (decrement loop count) */
        strb     lr,[r12],#1  	/* write byte to string 1 */
        bne      PreLoop     	/* loop if more bytes to move */

_LABEL(WordAligned)
        movs     r3,r2,asr #5 	/* any chunks of 8 words? */
        beq      OctsDone    	/* jump if no 8-word chunks */

        and      r2,r2,#0x1f 	/* subtract chunks from n */
        stmdb    sp!,{r4-r10} 	/* save registers on stack */

/*
 * The strings are long enough that we can transfer at least
 * some portion of the data in 8-word chunks.
 */
_LABEL(OctLoop)
        ldmia    r1!,{r4-r10,lr} /* load 8 words from string 2 */
        subs     r3,r3,#1    	/* more 8-word chunks to move? */
        stmia    r12!,{r4-r10,lr} /* write 8 words to string 1 */
        bne      OctLoop     	/* loop if more chunks */

        ldmia    sp!,{r4-r10} 	/* restore registers from stack */

_LABEL(OctsDone)
        movs     r3,r2,asr #2 	/* any more whole words to move? */
        beq      WordsDone   	/* jump if no more whole words */

/*
 * Copy as much of the remaining data as possible one word at
 * a time.
 */
_LABEL(WordLoop2)
        ldr      lr,[r1],#4  	/* read next word from string 2 */
        subs     r3,r3,#1    	/* decrement word count */
        str      lr,[r12],#4  	/* write next word to string 1 */
        bne      WordLoop2   	/* loop while more words to move */

_LABEL(WordsDone)
        ands     r2,r2,#3    	/* any last bytes to transfer? */
        ldmeqia  sp!,{pc}    	/* return if already done */

/*
 * The two strings do not end on word boundaries.
 * Copy the remaining data one byte at a time.
 */
_LABEL(ByteSerial)
        ldrb     lr,[r1],#1  	/* read byte from string 2 */
        subs     r2,r2,#1    	/* --n (decrement loop count) */
        strb     lr,[r12],#1  	/* write byte to string 1 */
        bne      ByteSerial  	/* loop if more bytes to move */

        ldmia    sp!,{pc}    	/* return to caller */


/*
 * Safely copy c bytes from source s to destination d.
 *
 * void *memmove(void *d, const void *s, unsigned c);
 */

_FUNCTION(__rt_memmove)
_FUNCTION(memmove)
        cmp      r0,r1          /* is d > s ? */
        bls      _F(memcpy)     /* jump to memcpy if d <= s */
/*
 * Need to copy backwards, starting at tail ends of source and
 * destination arrays.  Copy a word or a byte at a time?
 */

        orr      r3,r1,r0       /* tmp = s | d */
        orr      r3,r3,r2       /* tmp = s | d | c */
        ands     r3,r3,#3       /* is tmp even multiple of 4? */

        add      r1,r1,r2       /* s + c (end of source buffer) */
        add      r2,r2,r0       /* d + c (end of dest'n buffer) */

        beq      MOVE1          /* jump if tmp is multiple of 4 */
        b        MOVE2          /*  */
/*
 * because count c is an even multiple of 4 and the source
 * and destination arrays begin on even word boundaries, move
 * an entire word at a time from source to destination.
 */

_LABEL(MOVE3)
        ldrb     r3,[r1,#-1]!   /* load next byte from source */
        strb     r3,[r2,#-1]!   /* store next byte to dest'n */
_LABEL(MOVE2)
        teq      r0,r2          /* more bytes to move? */
        bne      MOVE3          /* jump if more bytes */

        mov      pc,lr          /* all done */
/*
 * Because the source and destination arrays are not aligned to even
 * word boundaries in memory, transfer only a byte at a time.
 */

_LABEL(MOVE4)
        ldr      r3,[r1,#-4]!   /* load next word from source */
        str      r3,[r2,#-4]!   /* store next word to dest'n */
_LABEL(MOVE1)
        teq      r0,r2          /* more words to move? */
        bne      MOVE4          /* jump if more words */

        mov      pc,lr          /* all done */



/*
 * Set n bytes of buffer p1 to value v.
 *
 *     void *memset(void *p1, int v, int n)
 */

_FUNCTION(__rt_memclr_w)          /* memclr(s, l) -> memset(s, 0, l) */
_FUNCTION(__rt_memclr)            /* memclr(s, l) -> memset(s, 0, l) */
        mov     r2, r1
        mov     r1, #0
_FUNCTION(__rt_memset)
_FUNCTION(memset)
        teq      r2,#0            /* is arg n == 0? */
        moveq    pc,lr            /* if n == 0, return */

        mov      r12,r0            /* copy pointer p */
        cmp      r2,#8            /* is n > 8 ? */
        ble      ByteSerial2      /* jump if n <= 8 */

/*
 * The string is long enough that at least of portion of it can
 * be set an entire word at a time.
 */
_LABEL(WordSerial2)
        ands     r3,r0,#3         /* is string word-aligned? */
        beq      WordAligned2     /* jump if string word-aligned */

        rsb      r3,r3,#4         /* m = no. of odd initial bytes */
        sub      r2,r2,r3         /* n = n - m */

/*
 * The string does not begin on a word boundary.  Set the first
 * few odd bytes of the string that precede the first word.
 */
_LABEL(PreLoop2)
        strb     r1,[r12],#1       /* set next memory byte to v */
        subs     r3,r3,#1         /* --m (decrement loop count) */
        bne      PreLoop2         /* loop if more odd bytes to set */

_LABEL(WordAligned2)
        and      r1,r1,#0xff      /* isolate arg v in 8 LSBs */
        orr      r1,r1,r1,lsl #8  /* duplicate v */
        orr      r1,r1,r1,lsl #16 /* replicate v across word */
        movs     r3,r2,lsr #5     /* any chunks of 8 words to set? */
        beq      OctsDone2        /* jump if no 8-word chunks */

/*
 * The string is long enough that we can set at least some portion
 * of the string an entire 8-word chunk at a time.
 */
        and      r2,r2,#0x1f      /* n &= 31 */
        stmdb    sp!,{r4-r10,lr}  /* save registers on stack */
        mov      r4,r1            /* copy v to registers */
        mov      r5,r1
        mov      r6,r1
        mov      r7,r1
        mov      r8,r1
        mov      r9,r1
        mov      r10,r1
        mov      lr,r1

_LABEL(OctLoop2)
        subs     r3,r3,#1         /* --nOcts (decrement loop count) */
        stmia    r12!,{r4-r10,lr}  /* set 8 memory words to value v */
        bne      OctLoop2         /* loop if more 8-word chunks to set */

        ldmia    sp!,{r4-r10,lr}  /* restore registers from stack */

_LABEL(OctsDone2)
        movs     r3,r2,lsr #2     /* any more whole words to set? */
        beq      WordsDone2       /* jump if no more whole words */

/*
 * At least some portion of the remainder of the string can be set
 * an entire word at a time.
 */
_LABEL(WordLoop3)
        subs     r3,r3,#1         /* --nWords (decrement loop count) */
        str      r1,[r12],#4       /* set next memory word to value v */
        bne      WordLoop3        /* loop if more words to set */

_LABEL(WordsDone2)
        ands     r2,r2,#3         /* any last bytes to transfer? */
        moveq    pc,lr            /* return if already done */

/*
 * The string does not end on a word boundary.  Set the last few
 * bytes in byte-serial fashion.
 */
_LABEL(ByteSerial2)
        subs     r2,r2,#1         /* --n (decrement loop count) */
        strb     r1,[r12],#1       /* store next byte to memory */
        bne      ByteSerial2      /* loop if more bytes to set */

        mov      pc,lr            /* return to caller */

        _END
