/* Copyright (c) Microsoft Corporation. All rights reserved. */

/*
 * Null cypher. e.g. identity function
 */

#include <mmlite.h>
#include <assert.h>
#define DBGLVL 0
#include <diagnostics.h>
#include <crypto.h>

#define ARBITRARY_BLOCK_LENGTH 16 /* meaning.. cyphers use small buffers */

/* State for our Cypher
 */
typedef struct _NULL_CYPHER {
    CONST_VTBL struct ICypherVtbl *v;
    UINT RefCnt;
} NULL_CYPHER, *PNULL_CYPHER;

#define pNullCypher(_i_) ((PNULL_CYPHER)((_i_)))
#define iNullCypher(_p_) ((PICYPHER)((_p_)))

/* Forward decls
 */
PRIVATE
PICYPHER
NullCypherNew(
       UINT KeyStrength);

PRIVATE
BOOL
NullCypherDestroy(
           PICYPHER pThis);

PIUNKNOWN NullCypherCobMain(
                     void);

/* Implementations for the exported methods
 */

PRIVATE SCODE NullCypherQueryInterface (
                         /*in*/ PICYPHER pThis,
                         /*in*/ REFIID iid,
                         /*out*/ void** ppObject
)
{
    PIUNKNOWN pUnk;

    DBGME(0,_tprintf(_TEXT("NullCypherQI %x %x [%x %x %x %x]\n"),
                     (ADDRESS)pThis, (ADDRESS)ppObject, *iid));

    /* If one wants to build more null cyphers out of one
     */
    if (UuidCmp(&IID_ICypherConstructor,iid)) {
        DBGME(0,_tprintf(_TEXT("NullCypherQI constructor %x\n"),(ADDRESS)ppObject));
        pUnk = NullCypherCobMain();
        if (pUnk)
            pUnk->v->AddRef(pUnk);
        *ppObject = pUnk;
        return NO_ERROR;
    }

    /* Otherwise..
     */
    return GenericQueryInterface((PIUNKNOWN) pThis, iid, ppObject, &IID_ICypher);
}

PRIVATE UINT NullCypherAddRef (
        /*in*/ PICYPHER pThis
)
{
    PNULL_CYPHER NullCypher = (PNULL_CYPHER) pThis;
    DBGME(0,_tprintf(_TEXT("NullCypherAR %x\n"),(ADDRESS)pThis));
    return AtomicInc(&NullCypher->RefCnt);
}

PRIVATE UINT NullCypherRelease (
                 /*in*/ PICYPHER pThis
)
{
    PNULL_CYPHER NullCypher = (PNULL_CYPHER) pThis;
    UINT Refs;

    Refs = AtomicDec(&NullCypher->RefCnt);
    DBGME(0,_tprintf(_TEXT("NullCypherR %x : %d\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
        (void) NullCypherDestroy(pThis);
    }

    return Refs;
}

PRIVATE SCODE NullCypherGetBlockLengths (
                                 /*in*/ PICYPHER pThis,
                                 /*inout*/ CYPHER_OPS * pOps,
                                 /*inout*/ UINT * pInputBlockLength,
                                 /*inout*/ UINT * pOutputBlockLength
)
{
    UnusedParameter(pThis);

    if (pInputBlockLength != NULL) {
        if (*pInputBlockLength == 0)
            *pInputBlockLength = ARBITRARY_BLOCK_LENGTH;
    }

    if (pOutputBlockLength != NULL) {
        if (*pOutputBlockLength == 0)
            *pOutputBlockLength = ARBITRARY_BLOCK_LENGTH;
    }

    /* But we dont do magic */
    if (pInputBlockLength && pOutputBlockLength &&
        (*pInputBlockLength != *pOutputBlockLength)) {
        return E_INVALID_PARAMETER;
    }

    return S_OK;
}


PRIVATE SCODE NullCypherSetKey (
                 /*in*/ PICYPHER pThis,
                 /*in*/ const BYTE * Key,
                 /*in*/ UINT Size,
                 /*in*/ CYPHER_OPS Operations
)
{
    DBGME(0,_tprintf(_TEXT("NullCypherSK %x %x %d %x\n"),
                     (ADDRESS)pThis, (ADDRESS)Key, Size, Operations));
    return S_OK;
}

PRIVATE SCODE NullCypherCypher (
                 /*in*/ PICYPHER pThis,
                 /*in*/ CYPHER_OPS Operation,
                 /*in*/ const BYTE * pInBuffer,
                 /*in*/ UINT InSize,
                 /*out*/ UINT * pSizeConsumed,
                 /*out*/ BYTE * pOutBuffer,
                 /*in*/ UINT OutSize,
                 /*out*/ UINT * pSizeProduced
)
{
    DBGME(0,_tprintf(_TEXT("NullCypherC %x %d %x %d %x %d %x\n"),
                     (ADDRESS)pThis, Operation,
                     (ADDRESS)pInBuffer, InSize,
                     (ADDRESS)pOutBuffer, OutSize, (ADDRESS)pSizeProduced));

    if (pInBuffer != pOutBuffer) {
        memmove(pOutBuffer,pInBuffer,InSize);
    }
    if (pSizeConsumed)
        *pSizeConsumed = InSize;
    *pSizeProduced = InSize;
    return S_OK;
}

PRIVATE SCODE NullCypherGenKey (
    PICYPHER pThis,
    UINT KeyStrength,
    /*out*/ const BYTE ** Key, UINT* KeySize,
    /*out*/ const BYTE ** PublicKey, UINT* PublicKeySize)
{
    PNULL_CYPHER This = pNullCypher(pThis);

    UnusedParameter(This);
    UnusedParameter(KeyStrength);
    UnusedParameter(Key);
    UnusedParameter(KeySize);
    UnusedParameter(PublicKey);
    UnusedParameter(PublicKeySize);

    return E_NOT_IMPLEMENTED;
}

PRIVATE void NullCypherDeleteKey (
    PICYPHER pThis,
    BYTE * Key)
{
    PNULL_CYPHER This = pNullCypher(pThis);

    UnusedParameter(This);
    UnusedParameter(Key);
}

#include "s_NullCypher_v.c"

/* Internal constructor/desctructor
 */
PRIVATE
PICYPHER
NullCypherNew(
       UINT KeyStrength)
{
    PNULL_CYPHER NullCypher;
    PIHEAP pHeap = CurrentHeap();
    NullCypher = pHeap->v->Alloc(pHeap,HEAP_ZERO_MEMORY, sizeof(*NullCypher), 0);
    if (NullCypher) {
        NullCypher->v = &NullCypherVtbl;
        NullCypher->RefCnt = 1;
    }
    DBGME(0,_tprintf(_TEXT("NullCypherNew -> %x\n"),(ADDRESS)NullCypher));
    return (PICYPHER) NullCypher;
}

PRIVATE
BOOL
NullCypherDestroy(
           PICYPHER pThis)
{
    PIHEAP pHeap = CurrentHeap();
    DBGME(0,_tprintf(_TEXT("NullCypherDestroy %x\n"),(ADDRESS)pThis));
    return pHeap->v->Free(pHeap,0,pThis);
}

/*
 * This is the exported constructor
 */
/********************** ICypherConstructor **********************/

/*BUGBUG these are quite generic */

PRIVATE SCODE NullCypherConstructorQueryInterface (
                         /*in*/ PICYPHERCONSTRUCTOR pThis,
                         /*in*/ REFIID iid,
                         /*out*/ void** ppObject
)
{
    DBGME(0,_tprintf(_TEXT("NullCypherConstructorQI %x %x [%x %x %x %x]\n"),
                     (ADDRESS)pThis, (ADDRESS)ppObject, *iid));

    return GenericQueryInterface((PIUNKNOWN) pThis, iid, ppObject, &IID_ICypherConstructor);
}

PRIVATE UINT NullCypherConstructorAddRef (
        /*in*/ PICYPHERCONSTRUCTOR pThis
)
{
    PCYPHERCOB NullCypherConstructor = (PCYPHERCOB) pThis;
    DBGME(0,_tprintf(_TEXT("NullCypherConstructorAR %x\n"),(ADDRESS)pThis));
    return AtomicInc(&NullCypherConstructor->RefCnt);
}

PRIVATE UINT NullCypherConstructorRelease (
                 /*in*/ PICYPHERCONSTRUCTOR pThis
)
{
    PCYPHERCOB NullCypherConstructor = (PCYPHERCOB) pThis;
    UINT Refs;

    Refs = AtomicDec(&NullCypherConstructor->RefCnt);
    DBGME(0,_tprintf(_TEXT("NullCypherConstructorR %x : %d\n"),(ADDRESS)pThis,Refs));
    /* Static object, no free */
    return Refs;
}

PRIVATE SCODE __cdecl NullCypherConstructorConstructor(
                             PICYPHERCONSTRUCTOR pThis,
                             /*out*/ PIUNKNOWN* ppUnk,
                             const _TCHAR * Name,
                             UINT KeyStrength)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Name);

    DBGME(0,_tprintf(_T("NullCypherConstructor(%x %x %s %d)\n"),
                     (ADDRESS) pThis, (ADDRESS) ppUnk, Name, KeyStrength));

    pUnk = (PIUNKNOWN) NullCypherNew(KeyStrength);

    DBGME(0,_tprintf(_T("NullCypherConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    }

    DBGME(3,_tprintf(_T("NullCypherConstructor failed\n")));
    return E_FAIL;
}

#include "s_NullCypherConstructor_v.c"

PRIVATE struct CYPHERCOB NullCypherConstructorObject = {
    (struct ICypherConstructorVtbl *) &NullCypherConstructorVtbl, 1 };

PIUNKNOWN NullCypherCobMain(void)
{
    DBGME(0,_tprintf(_T("NullCypherCobMain()\n")));

    return (PIUNKNOWN) &NullCypherConstructorObject;
}

