/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    gsound.c
 *
 * Purpose:
 *
 *    Device driver for the Giano simulated audio device.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */
#define DEVICE_ORDER LITTLE_ENDIAN
#include <mmlite.h>
#include <drivers/drivers.h>
#include <tchar.h>
#include <diagnostics.h>
#include <assert.h>

/* General definitions for the hw programming interface
 */
#define GSOUND_NBUFFERS 2
typedef struct {
    /* Physical addresses of the output audio buffers
     * NB: '0' is understood as not-valid.
     *                                 (RW)
     */
    volatile UINT32 BufferPointers[GSOUND_NBUFFERS];

    /* Same size for all
     */
    volatile UINT32 BufferSize;

    /* Commands here                   (RW)
     */
    volatile UINT32 Command;
#define GSOUND_CMD_STOP    0
#define GSOUND_CMD_START   1
#define GSOUND_CMD_PAUSE   2
#define GSOUND_CMD_RESUME  3
#define GSOUND_CMD_INTEN   4
#define GSOUND_CMD_INTDIS  5

    /* Status here, no side-effects    (RO)
     * Low byte same as above, then..
     */
    volatile UINT32 Status;
#define GSOUND_ST_ERROR    0x80000000
#define GSOUND_ST_ACT_MASK 0xff
#define GSOUND_ST_ACT0     0x01
#define GSOUND_ST_ACT1     0x02
#define GSOUND_ST_INTEN    0x40000000

    /* Interrupt status here, stacked  (RO)
     */
    volatile UINT32 InterruptStatus;
#define GSOUND_IST_ERROR      GSOUND_ST_ERROR
#define GSOUND_IST_DONE_MASK  0xff
#define GSOUND_IST_DONE0      0x01
#define GSOUND_IST_DONE1      0x02

    /* Identification
     */
    volatile UINT32 ChipId;
#define GSOUND_ID         0xaf00babe

    /* Volume control
     */
    volatile UINT16 VolumeLeft;
    volatile UINT16 VolumeRight;

    /* Same as WAVEFORMATEX in Win32 (caveat emptor!)
     *                                 (RW)
     */
    volatile UINT16 wFormatTag;      // format type
#define WAVE_FORMAT_PCM     1
    volatile UINT16 nChannels;       // 1=mono, 2=stereo...
    volatile UINT32 nSamplesPerSec;  // sample rate
    volatile UINT32 nAvgBytesPerSec; // for buffer estimation
    volatile UINT16 nBlockAlign;     // block size of data
    volatile UINT16 wBitsPerSample;  // one channel worth of bits
    volatile UINT32 mbz[4];          // pad, must-be-zero

} GSOUND_STATE, *PGSOUND_STATE;


#define STATISTICS 0
#if STATISTICS
#define KEEP_STATS(x) x
#else
#define KEEP_STATS(x)
#endif

/*
 * Concrete class definition for our device
 */
typedef struct CGsoundDevice {

    struct CDevice CDevice;

    /* Our own state
     */
    PGSOUND_STATE regs;

    MUTEX Lock;

    UINT16 OutputSampleRate, InputSampleRate;
    UINT16 VolumeLeft, VolumeRight;
    UINT8 MasterVolume;
    UINT8 Channels;
    UINT8 BitsPerSample;
    UINT8 NextBuffer;
    UINT Reset;
    UINT Errors;

} CGsoundDevice, *PCGsoundDevice;


/* Forward refs
 */
INT  MCT GsoundInitialize(PCGsoundDevice this);

/* Read SIZE bytes from device.
 * Returns whats available, if anything.
 * Else waits until at least one sample is in.
 */
SCODE MCT GsoundReadAt (PCGsoundDevice this, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    DBGME(0,_tprintf(_T("GsoundReadAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeRead));

    return E_NOT_IMPLEMENTED;
}

/* Write SIZE bytes to Audio output path.
 * Does not return until all bytes are sent (bad!).
 */
SCODE MCT GsoundWriteAt (PCGsoundDevice this, UINT64 Position, const BYTE *pBuffer,
                            UINT Size, PUINT pSizeWritten)
{
    PGSOUND_STATE gs = this->regs;
    SCODE sc = NO_ERROR;

    UnusedParameter(Position);

    DBGME(0,_tprintf(_T("GsoundWriteAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeWritten));

    /* Get ahold of send capability.  If someone else is
     * currently writing we wait here.
     */
    Mutex_Lock( &this->Lock );

    while (gs->BufferPointers[this->NextBuffer] != 0) {

        /* Sleep with device held
         *
         * An explicit wakeup on the condvar is required to break this.
         */
        KEEP_STATS(this->AudioOut.nSleepy++;)
        sc = Condition_Wait( &this->CDevice.IntrPoint,NULL,NULL);

        if (this->Reset)
            break;
    }

    /* Backoff if a reset has been requested
     */
    if (this->Reset) {
        Size = 0;
        sc = E_RESTART;
        goto Done;
    }

    /* Post the buffer
     */
    gs->BufferSize = _bs32(Size);
    gs->BufferPointers[this->NextBuffer] = _bs32((ADDRESS) pBuffer);
    gs->Command = _bs32(GSOUND_CMD_START);
    KEEP_STATS(this->AudioOut.nIn += Size;)

    /* Get ready for next buffer
     */
    if (++this->NextBuffer == GSOUND_NBUFFERS)
        this->NextBuffer = 0;

 Done:
    Mutex_Unlock( &this->Lock );

    return sc;
}

/* Interrupt service routine.
 * Acts as a DMA engine of sorts, pulling bytes
 * in and out of the queues as appropriate.
 */
BOOL MCT GsoundIsr(PCGsoundDevice this, PBOOL pNotMyInterrupt)
{
    PGSOUND_STATE gs = this->regs;
    UINT Intr;

    /* Look at the cause of interrupt
     */
    Intr = gs->InterruptStatus;

    /* Default: yes, its ours */
    UnusedParameter(pNotMyInterrupt);

    if (Intr & _bs32(GSOUND_IST_ERROR)) {

        this->Errors++;
    }

    KEEP_STATS(this->AudioOut.nSignalled++;)

    return Condition_InterruptSignal( & this->CDevice.IntrPoint );
}

/* (Re)Initialization code.
 * Leaves interrupts enabled, chip functional.
 */
INT  MCT GsoundInitialize(PCGsoundDevice this)
{
    PGSOUND_STATE gs = this->regs;
    UINT32 Value32;
    UINT16 Value16;
    int i;

    DBGME(0,_tprintf(_T("GsoundInit(%x)\n"), (ADDRESS) this));

    /* Flag that reset is pending.  This will abort any transmitting that is
     * in progress via WriteAt.
     */
    AtomicInc(&this->Reset);

    /* Wakeup any writer
     */
    Condition_Signal(&this->CDevice.IntrPoint);

    /* Wait for any transmitting threads to get out.  This gives us exclusive
     * access to the hardware.
     */
    Mutex_Lock(&this->Lock);

    /* Stop it dead
     */
    gs->Command = _bs32(GSOUND_CMD_INTDIS);
    gs->Command = _bs32(GSOUND_CMD_STOP);

    /* Cleanup buffers
     */
    this->NextBuffer = 0;
    gs->BufferSize = 0;
    for (i = 0; i < GSOUND_NBUFFERS; i++)
        gs->BufferPointers[i] = 0;

    /* Initialize things
     */
    this->MasterVolume = 1;

    /* Output rate and gain
     */
    gs->wFormatTag = _bs16(WAVE_FORMAT_PCM);
    gs->nChannels = _bs16(this->Channels);
    gs->nSamplesPerSec = _bs32(this->OutputSampleRate);
    Value16 = this->Channels * (this->BitsPerSample / 8);
    gs->nBlockAlign = _bs16(Value16);
    Value32 = this->OutputSampleRate *
        this->Channels * (this->BitsPerSample / 8);        
    gs->nAvgBytesPerSec = _bs32(Value32);
    gs->wBitsPerSample = _bs16(this->BitsPerSample);

    Value16 = this->VolumeLeft * this->MasterVolume;
    gs->VolumeLeft = _bs16(Value16);
    Value16 = this->VolumeRight * this->MasterVolume;
    gs->VolumeRight = _bs16(Value16);

    /* We're done.
     */
    AtomicDec(&this->Reset);

    gs->Command = _bs32(GSOUND_CMD_INTEN);

    /* Let other threads back in.
     */
    Mutex_Unlock( &this->Lock );
    
    return 0;
}

/* SetSize method
 */
SCODE MCT GsoundSetSize(PCGsoundDevice this, UINT64 Arg)
{
    PGSOUND_STATE gs = this->regs;
    UINT BitsPerSample, Rate;
    UINT16 Value16;
    UINT16 Volume;
    UINT8 Cmd;

    /* Decrit cmd
     */
    Cmd = (UINT8) (Uint64High32Bits(Arg) >> 24);

    switch (Cmd) {
    case 0: /* Sample rate & Channels */
        Rate = Uint64Low32Bits(Arg);
        BitsPerSample = Uint64High32Bits(Arg);
        this->Channels = (UINT8) ((BitsPerSample >> 16) & 0xff);
        this->BitsPerSample = (UINT8) BitsPerSample;
        this->OutputSampleRate = (UINT16) Rate;

        (void) GsoundInitialize(this);

        break;

    case 1: /* Master Volume */
        this->MasterVolume = (UINT8) Uint64Low32Bits(Arg);
        /* Change possibly while playing */
    SetVol:
        Value16 = this->VolumeLeft * this->MasterVolume;
        gs->VolumeLeft = _bs16(Value16);
        Value16 = this->VolumeLeft * this->MasterVolume;
        gs->VolumeRight = _bs32(Value16);
        break;

    case 2: /* Voice Channel Volume(s) */
        Volume = (UINT16) Uint64Low32Bits(Arg); /* not nuf bits, chop off */
        this->VolumeRight = Volume & 0xff;
        this->VolumeLeft = Volume >> 8;
        goto SetVol;
        break;

    default:
        return E_INVALID_PARAMETER;

    }

    return NO_ERROR;
}

struct CDeviceVtbl GsoundDriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)) GsoundReadAt,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Position, const BYTE *pBuffer, UINT Size, PUINT pSizeWritten)) GsoundWriteAt,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Size)) GsoundSetSize,
    NULL,
    (BOOL      (MCT *)(PCDEVICE pThis, PBOOL pWakeup)) GsoundIsr,
    NULL,
    (INT       (MCT *)(PCDEVICE pThis)) GsoundInitialize,
    NULL
};

/* Constructor
 */
BOOL
CGsoundDeviceInit (PCGsoundDevice this, ADDRESS Base, UINT *pIrq)
{
    DBGME(0,_tprintf(_T("GsoundDeviceInit(%x %x %x)\r\n"),
                     (ADDRESS) this, Base, (ADDRESS) pIrq));

    this->regs = (PGSOUND_STATE) Base;

    if (this->regs->ChipId != _bs32(GSOUND_ID))
        return FALSE;

    Mutex_Init(&this->Lock);
    Condition_Init(&this->CDevice.IntrPoint);

    this->OutputSampleRate = this->InputSampleRate = 11025;
    this->BitsPerSample = 16;
    this->Channels = 2;
    this->VolumeLeft = 1;
    this->VolumeRight = 1;

    this->Reset = FALSE;

    this->CDevice.v = &GsoundDriverVtable;
    this->CDevice.RefCnt = 1;

    return TRUE;
}


/* Exported constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE GsoundNew( ADDRESS * pPortBase, PUINT pIrq)
{
    PCDEVICE pAud;
    PIHEAP pH = CurrentHeap();

    pAud = (PCDEVICE) pH->v->Alloc( pH, HEAP_ZERO_MEMORY, sizeof(CGsoundDevice), 0 );
    if (pAud) {
        if (!CGsoundDeviceInit( (PCGsoundDevice)pAud,
                                *pPortBase,
                                pIrq))
        {
            pH->v->Free( pH, 0, pAud);
            pAud = NULL;
        }
    }

    return pAud;
}

/* ** 
*/
SCODE __cdecl GsoundConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("GsoundConstructor(%x %s %x %d %x)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) GsoundNew(&RegisterBase,&Interrupt);

    DBGME(0,_tprintf(_T("GsoundConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("GsoundConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl GsoundDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) GsoundConstructor
};

static struct _DRIVERCOB GsoundDriverCobObject = { &GsoundDriverCobVtable, 1 };

PIUNKNOWN GsoundCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &GsoundDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &GsoundDriverVtable, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    /* Null the thread loop, to save a useless thread
     */
    GsoundDriverVtable.ThreadLoop = NULL;

    return (PIUNKNOWN) &GsoundDriverCobObject;
}
