/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    ucb1100.h
 *
 * Purpose:
 *
 *    Portable device driver for Philips UCB1100 codec and compatibles.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#ifndef _MMLITE_H_   /* If customization code has not */
#include <mmlite.h>
#include <mmhal.h>
#include <drivers/drivers.h>
#endif
#include <tchar.h>
#include <diagnostics.h>
#include <assert.h>

/* General definitions for the chip programming interface
 */

/* Register names
 */
#define UCB_IO_DATA            0
#define UCB_IO_DIR             1
#define UCB_RISING_INTEN       2
#define UCB_FALLING_INTEN      3
#define UCB_IO_INT             4
#define UCB_TELCOM_A           5
#define UCB_TELCOM_B           6
#define UCB_AUDIO_A            7
#define UCB_AUDIO_B            8
#define UCB_TOUCH_SCREEN       9
#define UCB_ADC_CONTROL        10
#define UCB_ADC_DATA           11
#define UCB_ID                 12
#define UCB_MODE               13
#define UCB_reserved           14
#define UCB_NULL               15

/* Register 0: IO port data register
 */
#define UCBIO_MASK             0x03ff   /* 10 pins */
#define UCBIO_PIN0             (1<<0)
#define UCBIO_PIN1             (1<<1)
#define UCBIO_PIN2             (1<<2)
#define UCBIO_PIN3             (1<<3)
#define UCBIO_PIN4             (1<<4)
#define UCBIO_PIN5             (1<<5)
#define UCBIO_PIN6             (1<<6)
#define UCBIO_PIN7             (1<<7)
#define UCBIO_PIN8             (1<<8)
#define UCBIO_PIN9             (1<<9)

/* Register 1: IO port direction register
 */
#define UCBIO_OUT              0x1      /* pin is output */
#define UCBIO_IN               0x0      /* pin is input */
#define UCBIO_SIB_ZERO         (1<<15)  /* 1 -> sibdout forced 0 in 2nd word */

/* Register 2: Rising edge interrupt enable register
 * One bit per IO pin plus..
 */
#define UCBINT_ADC             (1<<11)
#define UCBINT_TSPX            (1<<12)
#define UCBINT_TSMX            (1<<13)
#define UCBINT_TCLIP           (1<<14)
#define UCBINT_ACLIP           (1<<15)

/* Register 3: Falling edge interrupt enable register
 * One bit per IO pin plus the above
 */

/* Register 4: Interrupt clear/status register
 * One bit per IO pin plus the above
 */

/* Register 5: Telecom control register A
 */
#define UCBTA_DIVISOR_MASK     0x007f     /* valid values: 16..127 */
#define UCBTA_LOOPBACK         (1<<7)     /* 1100 & 1200 */
#define UCBTA_AMP              (1<<7)     /* 1300 : +6 dB for weak input */

/* Register 6: Telecom control register B
 */
#define UCBTB_HIGH_PASS_ENABLE (1<<3)
#define UCBTB_TEL_CLIP         (1<<4)
#define UCBTB_TEL_ATT          (1<<6)     /* 6 dB attenuation */
#define UCBTB_SIDE_SUP_ENA     (1<<11)
#define UCBTB_TEL_MUTE         (1<<13)
#define UCBTB_TEL_IN_ENA       (1<<14)
#define UCBTB_TEL_OUT_ENA      (1<<15)

/* Register 7: Audio control register A
 */
#define UCBAA_DIVISOR_MASK     0x007f     /* valid values: 16..127 */
#define UCBAA_GAIN_MASK        0x0f80     /* valid values: 0..31 but:  */
#define UCBAA_MAX_GAIN_1100    31         /*     1100,1300 -> 46.5dB   */
#define UCBAA_MAX_GAIN_1200    15         /*     1200 -> 22.5dB        */

# define UCBAA_GET_GAIN(_x_)   (((_x_)&UCBAA_GAIN_MASK)>>7)
# define UCBAA_SET_GAIN(_x_)   (((_x_)<<7)&UCBAA_GAIN_MASK)

/* Register 8: Audio control register B
 */
#define UCBAB_ATT_MASK         0x001f     /* valid values: 0..31 = 69 dB */
#define UCBAB_CLIP_ENA         (1<<6)
#define UCBAB_LOOP             (1<<8)
#define UCBAB_MUTE             (1<<13)
#define UCBAB_IN_ENA           (1<<14)
#define UCBAB_OUT_ENA          (1<<15)

/* Register 9: Touch screen control register
 * NB: In the 1100 specs the 'P' and 'M' bits are swapped.
 */
#define UCBTS_TSMX_POWER       (1<<0)
#define UCBTS_TSPX_POWER       (1<<1)
#define UCBTS_TSMY_POWER       (1<<2)
#define UCBTS_TSPY_POWER       (1<<3)
#define UCBTS_TSMX_GROUND      (1<<4)
#define UCBTS_TSPX_GROUND      (1<<5)
#define UCBTS_TSMY_GROUND      (1<<6)
#define UCBTS_TSPY_GROUND      (1<<7)
#define UCBTS_INTERRUPT_MODE   0x0000
#define UCBTS_PRESSURE_MODE    0x0100
#define UCBTS_POSITION_MODE    0x0200
#define UCBTS_BIAS_ENA         (1<<11)
#define UCBTS_TSPX_STATE       (1<<12) /* 1 is pen down */
#define UCBTS_TSMX_STATE       (1<<13) /* 1 is pen down */

/* Register 10: ADC control register
 */
#define UCBAC_SYNC_ENA         (1<<0)
#define UCBAC_VREF_BYPASS      (1<<1)  /* 1 -> use ext vrefbyp pin for Vref */
#define UCBAC_TSPX             0x0000  /* select one of these as input */
#define UCBAC_TSMX             0x0004
#define UCBAC_TSPY             0x0008
#define UCBAC_TSMY             0x000c
#define UCBAC_AD0              0x0010
#define UCBAC_AD1              0x0014
#define UCBAC_AD2              0x0018
#define UCBAC_AD3              0x001c
#define UCBAC_EXT_ENA          (1<<5)
#define UCBAC_START            (1<<7)
#define UCBAC_ENA              (1<<15)

/* Register 11: ADC data register
 */
#define UCBAD_MASK             0x7ff0   /* 10 bits (4..14) */
#define UCBAD_VALID            (1<<15)  /* conversion is complete */
# define UCBAD_GET_DATA(_x_)   (((_x_)&UCBAD_DATA_MASK)>>4)
# define UCBAD_SET_DATA(_x_)   (((_x_)<<4)&UCBAD_DATA_MASK)

/* Register 12: ID register
 */
#define UCBID_GET_VERSION(x)   (UINT8)((x)&0x3f)
#define UCBID_GET_DEVICE(x)    (UINT8)(((x)&0x0fc0)>>6)
#define UCBID_GET_SUPPLIER(x)  (UINT8)(((x)&0xf000)>>12)

#define UCBID_1100             3  /* as a VERSION field */
#define UCBID_1200             4
#define UCBID_1300             5
#define UCBID_UCB              0  /* device is UCB compatible */
#define UCBID_PHILIPS          1  /* supplier */

/* Register 13: Mode register
 */
#define UCBM_AUDIO_TEST        (1<<0)
#define UCBM_TELCO_TEST        (1<<1)
#define UCBM_PROD_TEST_MASK    0x003c
#define UCBM_DYN_VFLAG         (1<<12)
#define UCBM_AUDIO_OFFSET_OFF  (1<<13)

/* Register 15: Null register
 */
#define UCBNULL_VALUE          0xffff

#define STATISTICS 0
#if STATISTICS
#define KEEP_STATS(x) x
#else
#define KEEP_STATS(x)
#endif

/*
 * Byte queue data type.
 *
 * The count of bytes in the queue is computed by using the Head and Tail
 * values.  Because there are Size+1 "states" (possible different counts
 * including empty), the maximum number of bytes that a queue can hold is
 * Size-1.
 *
 * The exception to this rule is that when a queue is initialized, Tail is
 * set to -1.  This means that a queue can be initialized "full".
 */
typedef struct {
    CONDITION WaitPoint;
    MUTEX Lock;
    PUINT16 pBuffer;
    volatile INT Head, Tail;
    UINT Size;
#if STATISTICS
    UINT nIn, nOut, nSignalled, nLoopy, nSleepy, nCally;
#endif
} BYTE_QUEUE;

/* Is the queue empty ?
 */
#define CQueueEmpty(_q_) \
    ((_q_).Head == (_q_).Tail)

/* How many bytes in the queue ?
 * Non-opaque, but interrupt-safe eval
 */
#define CQueueCountEval(_h_,_t_,_s_) \
    (((INT)(_h_) > (INT)(_t_)) ? \
     ((INT)(_h_) - (INT)(_t_)) : \
     ((_s_) - ((INT)(_t_) - (INT)(_h_))))

/* How many bytes in the queue ?
 * Opaque, but beware of multiple evals
 */
#define CQueueCount(_q_) \
    CQueueCountEval(_q_.Head,_q_.Tail,_q_.Size)

/* How much room left in the queue ?
 * Non-opaque, but interrupt-safe eval
 */
#define CQueueFreeEval(_h_,_t_,_s_) \
    (((INT)(_h_) > (INT)(_t_)) ? \
     ((_s_)-1 - ((INT)(_h_) - (INT)(_t_))) : \
     (((INT)(_t_) - (INT)(_h_))-1))

/* How much room left in the queue ?
 * Opaque, but beware of multiple evals
 */
#define CQueueFree(_q_) \
    CQueueFreeEval(_q_.Head,_q_.Tail,_q_.Size)

/* Initialize a byte queue
 */
#define CQueueInit(_q_) \
  { \
    Condition_Init(& (_q_).WaitPoint); \
    Mutex_Init(& (_q_).Lock); \
    (_q_).Tail = (_q_).Head = -1; \
  }

#define SAMPLE_SIZE 2
#ifndef UCB_OUTBUFFER_SIZE
#define UCB_OUTBUFFER_SIZE (1024-8)  /* /2 == 508 */
#endif

/*
 * Concrete class definition for our device
 */
typedef struct CUcbDevice {

    struct CDevice CDevice;

    /* Our own state
     */
    IO_PORT PortBase;
    IO_PORT ControlBase;
    UINT16 ChipVersion;
    UINT16 AudioSampleRate, TelcoSampleRate;
    UINT16 AudioGain, AudioVolume;
    UINT Errors;
    BOOL Reset, AudioIsActive;

    BYTE_QUEUE AudioOut;

} CUcbDevice, *PCUcbDevice;


/* Forward refs
 */
INT  MCT UcbInitialize(PCUcbDevice this);

/* Utilities
 */
#ifdef _DEBUG
typedef _TCHAR *REGNAME;
void UcbDump(PCUcbDevice this, BOOL DoRegs)
{
    int i;
    _TCHAR Name[] = _T("UCB1100");
    static REGNAME Regnames[16] = {
        _T("IO_DATA"),_T("IO_DIR"),_T("RISING_INTEN"),_T("FALLING_INTEN"),
        _T("IO_INT"),_T("TELCOM_A"),_T("TELCOM_B"),_T("AUDIO_A"),
        _T("AUDIO_B"),_T("TOUCH_SCREEN"),_T("ADC_CONTROL"),_T("ADC_DATA"),
        _T("ID\t"),_T("MODE\t"),0,_T("NULL\t")
    };

    if ((this->ChipVersion >= UCBID_1100) &&
        (this->ChipVersion < 9))
        Name[4] = (_TCHAR)(_T('1') + (this->ChipVersion - UCBID_1100));
    else
        Name[4] = _T('?');

    _tprintf(_T("@x%x  "),(ADDRESS)this);
    UCB_COMM_DUMP(this,Name);

    if (!DoRegs)
        return;

    for (i = 0; i < sizeof Regnames/sizeof Regnames[0]; i++) {
        UINT n;
        if (!Regnames[i]) continue;
        UCB_REGREAD(this,i,n);
        _tprintf(_T("\t%d %s\tx%x"),i,Regnames[i],n);
        if (i&1) _tprintf(_T("\r\n"));
    }
}

#define nSam 256
UINT16 Samples[nSam];

void Test(PCUcbDevice this)
{
#define DUMPIT(_x_)
    TIME Now, End;
    int i, j, k;

    /* In those 256 samples we can fit 8 periods of a saw (~~sin)
     * Each zero is one period.  First zero is at 32, the length in samples.
     * At a 11KHz sampling rate (90us period) this is a 347.2 Hz wave.
     * Its audible (and a bit annoying too)
     */
    j = 128;
    i = 0;
    DUMPIT(_tprintf(_T("Zeroes at: "));)
    for (k = 0; k < nSam; k++) {
        DUMPIT(if (i == 0) _tprintf(_T("%d "), k);)
        Samples[k] = i << 4;

        i += j;
        if ((i >= ((1<<11)-1)) || (i <= 0)) {
            j = -j;
        }
    }
    DUMPIT(
           _tprintf(_T("\nSamples: %d %d %d %d ...%d %d\n"),
                    Samples[0], Samples[1], Samples[2], Samples[3], 
                    Samples[nSam-2],Samples[nSam-1]);)
    for (k = 0; k < nSam; k++) {
        DUMPIT(_tprintf(_T("%d "), Samples[k]);)
    }
    DUMPIT(_tprintf(_T("\n"));)

    Int32ToInt64(End,TIME_SECONDS(5));
    Now = CurrentTime();
    End = TimeAdd(End,Now);

    UCB_WRITE_XMIT_BUFFER(this, Samples[0]);
    k = 1;
    UCB_REGWRITE(this,UCB_AUDIO_B,UCBAB_OUT_ENA|this->AudioVolume);
    this->AudioIsActive = TRUE;

    for (;;) {

        /* Play one buffer
         */
        do {
            i = 0;

            while (UCB_XMIT_BUFFER_READY(this)) {
                UCB_WRITE_XMIT_BUFFER(this, Samples[k++]);
                if (k >= nSam)
                    break;
            }

            /* Are we done yet ?
             */
            Now = CurrentTime();
            if (TimeLess(End,Now))
                break;

        } while (k < nSam);
        k = 0;

        if (TimeLess(End,Now))
            break;
    }
    
}

#else
#define UcbDump(x,y)
#define Test(x)
#endif

/* Read SIZE bytes from device.
 * Returns whats available, if anything.
 * Else waits until at least one sample is in.
 */
SCODE MCT UcbReadAt (PCUcbDevice this, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    DBGME(0,_tprintf(_T("UcbReadAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeRead));

    Test(this);
    return E_NOT_IMPLEMENTED;/*yet*/
}

/* Write SIZE bytes to Audio output path.
 * Does not return until all bytes are sent (bad!).
 */
SCODE MCT UcbWriteAt (PCUcbDevice this, UINT64 Position, const BYTE *pBuffer,
                            UINT Size, PUINT pSizeWritten)
{
    UINT16 c = 0;
    SCODE sc = NO_ERROR;
    UINT Avail;

    UnusedParameter(Position);

    DBGME(0,_tprintf(_T("UcbWriteAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeWritten));

    /* Get ahold of send capability.  If someone else is
     * currently writing we wait here.
     */
    Mutex_Lock( &this->AudioOut.Lock );

    /* Sleep with device held until below threshold (half)
     *
     * An explicit wakeup on the condvar is required to break this.
     */
    KEEP_STATS(this->AudioOut.nCally++;)
    for (;;) {
        /* How much can we put ?
         */
        KEEP_STATS(this->AudioOut.nLoopy++;)
        if (CQueueEmpty(this->AudioOut)) {
            Avail = this->AudioOut.Size-1;
        } else {
            Avail = this->AudioOut.Tail; /* cache against interrupts */
            Avail = CQueueFreeEval(this->AudioOut.Head,Avail,this->AudioOut.Size);
        }

        /* If we have to wait, lets wait until much room is available.
         * NB: the ISR still wakes up only on last one out. BUGBUG fixme.
         */
        if ((Avail > (this->AudioOut.Size>>1)) || this->Reset)
            break;

        /* Not enough, sleep
         */
        KEEP_STATS(this->AudioOut.nSleepy++;)
        sc = Condition_Wait( &this->AudioOut.WaitPoint,NULL,NULL);
    }

    /* Backoff if a reset has been requested
     */
    if (this->Reset) {
        Size = 0;
        sc = E_RESTART;
        goto Done;
    }

    /* Trim how much we post
     */
    Size /= SAMPLE_SIZE;/* nope cant do 1 byte */
    Size = (Size > Avail) ? Avail : Size;
    if (pSizeWritten)
        *pSizeWritten = Size * SAMPLE_SIZE;
    KEEP_STATS(this->AudioOut.nIn += Size;)

    /* copy data, first part (towards the end of buffer)
     */
    Avail = this->AudioOut.Size - 1 - this->AudioOut.Head;
    if (Avail > Size)
        Avail = Size;
    if (Avail) {
        memcpy( & this->AudioOut.pBuffer[ this->AudioOut.Head + 1],
                pBuffer,
                Avail * SAMPLE_SIZE);
        pBuffer += Avail * SAMPLE_SIZE;
        this->AudioOut.Head += Avail;
        Size -= Avail;
    }

    /* copy data, second part (start of buf)
     */
    if (Size) {
        memcpy( this->AudioOut.pBuffer,
                pBuffer,
                Size * SAMPLE_SIZE );
        this->AudioOut.Head = Size - 1;
    }

    /* If the Audio codec isnt running get it started now.
     * But first, prime the xmit buffer/FIFOs.
     */
    if (!UCB_AUDIOACTIVE(this)) {

        /* Get the transmission process started.
         *
         * There are two schemes used for xmit interrupts.
         * In one case, the chip interrupts any time its
         * xmit buffer is empty (including the idle state).
         * In the other case, an interrupt is generated only on
         * successful transmission of one (or the last) byte.
         */
#if UCB_INTERRUPTS_ON_XMIT_EMPTY
        /* Send as many samples as we can now.  
         * Enable interrupts iff necessary to finish the job.
         */
        while (UCB_XMIT_BUFFER_READY(this) && !CQueueEmpty(this->AudioOut)) {

            /* Interrupts are disabled, ISR cannot run. 
             * We are therefore safe to touch Tail.
             */
            Avail = this->AudioOut.Tail + 1;
            if (Avail >= this->AudioOut.Size)
                Avail = 0;

            c = this->AudioOut.pBuffer[Avail];
            this->AudioOut.Tail = Avail;

            UCB_WRITE_XMIT_BUFFER(this, c);
        }

#else /* UCB_INTERRUPTS_ON_XMIT_EMPTY not */
        /* Must send the first byte out. Xmit interrupts are
         * always on.
         */
        if (UCB_XMIT_BUFFER_READY(this)) {

            /* Interrupts are disabled, ISR cannot run. 
             * We are therefore safe to touch Tail.
             */
            Avail = this->AudioOut.Tail + 1;
            if (Avail >= this->AudioOut.Size)
                Avail = 0;

            c = this->AudioOut.pBuffer[Avail];
            this->AudioOut.Tail = Avail;

            UCB_WRITE_XMIT_BUFFER(this, c);
        }

#endif /* UCB_INTERRUPTS_ON_XMIT_EMPTY */

        /* Get the codec going.
         */
        UCB_REGWRITE(this,UCB_AUDIO_B,UCBAB_OUT_ENA|this->AudioVolume);
        this->AudioIsActive = TRUE;
    }

    /* In the presence of a FIFO we might have disabled interrupts
     * after loading the last few samples. Better check they are on if
     * there is somehting left to send out.  
     */
#if UCB_INTERRUPTS_ON_XMIT_EMPTY
    if (!CQueueEmpty(this->AudioOut)) {
        UCB_ENABLE_TRANSMIT_INTERRUPTS(this);
    }
#endif

 Done:
    Mutex_Unlock( &this->AudioOut.Lock );

    return sc;
}

/* Interrupt service routine.
 * Acts as a DMA engine of sorts, pulling bytes
 * in and out of the queues as appropriate.
 */
BOOL MCT UcbIsr(PCUcbDevice this, PBOOL pNotMyInterrupt)
{
    INT Avail; /* signed !! */
    UINT16 c;
    BOOL Wakeup = FALSE, Signal;
    UINT Intr;

    /* Look at the cause of interrupt
     */
    UCB_READ_INTR_PORT(this,Intr);

    /* Default: yes its ours */
    UnusedParameter(pNotMyInterrupt);

    Signal = FALSE;
    if (UCB_ISR_TRANSMITTED_DATA(this,Intr)) {

        do {
            /* Have we got another sample to send ?
             */
            if (!CQueueEmpty(this->AudioOut)) {

                if (!this->Reset) {
                    /* Pick up next byte, update ptrs
                     */
                    Avail = this->AudioOut.Tail + 1;
                    if (Avail >= (INT)this->AudioOut.Size)
                        Avail = 0;

                    c = this->AudioOut.pBuffer[Avail];

                    /* Send sample out
                     */
                    UCB_WRITE_XMIT_BUFFER(this, c);
                    KEEP_STATS(this->AudioOut.nOut++;)

                    /* Set updated ptr
                     */
                    this->AudioOut.Tail = Avail;

                    /* Signal once at the halfpoint
                     */
                    Avail = CQueueFree(this->AudioOut);
                    if (Avail == (INT)(this->AudioOut.Size>>1))
                        Signal = TRUE;

                } else {
                    /* Reset is pending.  Dump the rest of the buffer so
                     * the reset can proceed.
                     */
                    this->AudioOut.Tail = this->AudioOut.Head;
                    Signal = TRUE;

                }

            } else {

#if UCB_INTERRUPTS_ON_XMIT_EMPTY
                /* Time to turn ints off
                 */
                UCB_DISABLE_TRANSMIT_INTERRUPTS(this);        
#endif
                /* We just sent the last character, wakeup writer
                 */
                Signal = TRUE;
                break;
            }

        } while (UCB_XMIT_BUFFER_READY(this));

    }

    /* Check for errors
     */
    if (UCB_ERRORS(this,Intr)) {

        this->Errors++;
        UCB_REGWRITE(this,UCB_AUDIO_B,UCBAB_MUTE);
        this->AudioIsActive = FALSE;
        Signal = TRUE;

        /* NB: Something fishy with that FIFO, seems to pick from somewhere
         * other than the last sample when in underrun. Filling up all four
         * slots seems to cure it.  This shouldnt be a problem elsewhere.
         */
        UCB_CLEAR_ERROR(this,Intr);
        UCB_AUDIOWRITE(this,0);
        UCB_AUDIOWRITE(this,0);
        UCB_AUDIOWRITE(this,0);
        UCB_AUDIOWRITE(this,0);
    }

    if (Signal) {
        Wakeup |= Condition_InterruptSignal( & this->AudioOut.WaitPoint );
        KEEP_STATS(this->AudioOut.nSignalled++;)
    }

    UCB_ACK_INTERRUPTS(this);
    return Wakeup;
}

/* (Re)Initialization code.
 * Leaves interrupts enabled, chip functional.
 */
INT  MCT UcbInitialize(PCUcbDevice this)
{
    UINT n;
    INT r = 0;

    DBGME(0,_tprintf(_T("UcbInit(%x)\n"), (ADDRESS) this));

    /* Flag that reset is pending.  This will abort any transmitting that is
     * in progress via WriteAt.
     */
    AtomicInc(&this->Reset);

    /* Wakeup any writer
     */
    Condition_Signal(&this->AudioOut.WaitPoint);

    /* Wait for any transmitting threads to get out.  This gives us exclusive
     * access to the hardware.
     */
    Mutex_Lock(&this->AudioOut.Lock);

    /* Init comm interface
     */
    UCB_COMM_INIT(this);

    /* See what type it is, if any
     */
    UCB_REGREAD(this,UCB_NULL,n);
    if (n != UCBNULL_VALUE) {
        DBGME(3,_tprintf(_T("UCB not detected (%x %x)\r\n"),
                         this->PortBase, this->ControlBase));
        UcbDump(this,FALSE);
        r = -1;
        goto Done;
    }

    UCB_REGREAD(this,UCB_ID,n);
    this->ChipVersion = UCBID_GET_VERSION(n);
    DBGME(0,_tprintf(_T("UCB ID %d.%d.%d (x%x)\r\n"), 
           UCBID_GET_VERSION(n),
           UCBID_GET_DEVICE(n),
           UCBID_GET_SUPPLIER(n),
           n));

    /* Phase 1: RESET
     */

    /* Stop and reset both codecs
     */
    UCB_REGWRITE(this,UCB_TELCOM_B,UCBTB_TEL_CLIP|UCBTB_TEL_MUTE);
    UCB_REGWRITE(this,UCB_TELCOM_A,16);
    UCB_REGWRITE(this,UCB_AUDIO_B,UCBAB_MUTE);
    UCB_REGWRITE(this,UCB_AUDIO_A,6);

    /* Bring touch screen back to reset state
     */
    UCB_REGWRITE(this,UCB_TOUCH_SCREEN,0);
    
    /* Ditto for ADC
     */
    UCB_REGWRITE(this,UCB_ADC_CONTROL,0);    

    /* Disable all interrupts
     */
    UCB_REGWRITE(this,UCB_RISING_INTEN,0);
    UCB_REGWRITE(this,UCB_FALLING_INTEN,0);

    /* Clear all pending interrupts (unnecessary but)
     */
    UCB_REGWRITE(this,UCB_IO_INT,~0);

    /* Say where we are (DEBUG)
     */
    UcbDump(this,TRUE);

    /* Phase 2: INITIALIZE
     */

    /* Audio codec rate and gain
     */
    UCB_DIVISOR(this,this->AudioSampleRate,n);
    UCB_COMM_AUDIORATE(this,this->AudioSampleRate,n);
    UCB_REGWRITE(this,UCB_AUDIO_A,
                 (UINT16)(UCBAA_SET_GAIN(this->AudioGain)|n));

    /* Telco codec rate
     */
    UCB_DIVISOR(this,this->TelcoSampleRate,n);
    UCB_COMM_TELCORATE(this,this->TelcoSampleRate,n);

 Done:
    /* We're done.
     */
    AtomicDec(&this->Reset);

    /* Let other threads back in.
     */
    Mutex_Unlock( &this->AudioOut.Lock );
    
    return r;
}

/* SetSize method
 */
SCODE MCT UcbSetSize(PCUcbDevice this, UINT64 Arg)
{
    UINT Channels, BitsPerSample, Rate;
    UINT8 Cmd;

    /* Decrit cmd
     */
    Cmd = (UINT8) (Uint64High32Bits(Arg) >> 24);

    switch (Cmd) {
    case 0: /* Sample rate & Channels */
        Rate = Uint64Low32Bits(Arg);
        BitsPerSample = Uint64High32Bits(Arg);
        Channels = (BitsPerSample >> 16) & 0xff;
        BitsPerSample &= 0xffff;

        /* what we cant do we cant */
        if ((BitsPerSample != 16) ||
            (Channels != 1) ||
            (Rate > 44100))
            return E_INVALID_PARAMETER;

        this->AudioSampleRate = (UINT16) Rate;

        (void) UcbInitialize(this);

        break;

    case 1: /* Master Volume */
        this->AudioVolume = (UINT8) Uint64Low32Bits(Arg);
        /* Change possibly while playing */
        UCB_REGREAD(this,UCB_AUDIO_B,Rate);
        Rate = (Rate & ~UCBAB_ATT_MASK);
        UCB_REGWRITE(this,UCB_AUDIO_B,Rate|this->AudioVolume);

        break;

#if notyet
    case 2: /* Voice Channel Volume(s) */
        Volume = (UINT8) Arg; /* not nuf bits, chop off */
        WriteMixer(MIX_VOICE_VOLIDX_R, Volume);
        Volume = (UINT8) (Arg >> 8);
        WriteMixer(MIX_VOICE_VOLIDX_L, Volume);
        break;
#endif
    default:
        return E_INVALID_PARAMETER;

    }

    return NO_ERROR;
}

struct CDeviceVtbl UcbDriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)) UcbReadAt,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Position, const BYTE *pBuffer, UINT Size, PUINT pSizeWritten)) UcbWriteAt,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Size)) UcbSetSize,
    NULL,
    (BOOL      (MCT *)(PCDEVICE pThis, PBOOL pWakeup)) UcbIsr,
    NULL,
    (INT       (MCT *)(PCDEVICE pThis)) UcbInitialize,
    NULL
};

/* Constructor
 */
void
CUcbDeviceInit (PCUcbDevice this, IO_PORT Base, UINT *pIrq, IO_PORT ControlBase )
{
    PIHEAP pH = CurrentHeap();
    PUINT16 pBuffer;

    DBGME(0,_tprintf(_T("UcbDeviceInit(%x %x %x %x)\r\n"),
                     (ADDRESS) this, Base, (ADDRESS) pIrq, ControlBase));

    this->AudioSampleRate = this->TelcoSampleRate = 11025;
    this->AudioGain = 0;
    this->AudioVolume = 2;

    pBuffer = (PUINT16) pH->v->Alloc ( pH,0,UCB_OUTBUFFER_SIZE,0);

    this->AudioOut.Size = UCB_OUTBUFFER_SIZE / SAMPLE_SIZE;
    CQueueInit( this->AudioOut );
    this->AudioOut.pBuffer = pBuffer;
    this->Reset = 0;

    this->ChipVersion = 0;/* unk for now */
    this->PortBase = Base;
    this->ControlBase = ControlBase;

    this->CDevice.v = &UcbDriverVtable;
    this->CDevice.RefCnt = 1;

}


/* Exported constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE UcbNew( ADDRESS * pPortBase, PUINT pIrq, PUINT pCtlBase)
{
    PCDEVICE pAud;
    PIHEAP pH = CurrentHeap();

    pAud = (PCDEVICE) pH->v->Alloc( pH, HEAP_ZERO_MEMORY, sizeof(CUcbDevice), 0 );
    if (pAud)
        CUcbDeviceInit( (PCUcbDevice)pAud,
                        (IO_PORT)*pPortBase,
                        pIrq,
                        (IO_PORT)*pCtlBase);

    return pAud;
}

/* ** 
*/
SCODE __cdecl UcbConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        PTR NoIdeaWhatsThisFor,
        PTR NoIdeaWhatsThisForNeither,
        UINT ControlBase,
        ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("UcbConstructor(%x %s %d %x %x %x)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk, ControlBase));

    pUnk = (PIUNKNOWN) UcbNew(&RegisterBase,
                              &Interrupt,
                              &ControlBase);

    DBGME(0,_tprintf(_T("UcbConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UcbConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UcbDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UcbConstructor
};

static struct _DRIVERCOB UcbDriverCobObject = { &UcbDriverCobVtable, 1 };

PIUNKNOWN UcbCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UcbDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UcbDriverVtable, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    /* Null the thread loop, to save a useless thread
     */
    UcbDriverVtable.ThreadLoop = NULL;

    return (PIUNKNOWN) &UcbDriverCobObject;
}
