/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    ucb_cerf.c
 *
 * Purpose:
 *
 *    Portable device driver for Philips UCB1100 codec and compatibles.
 *    Customization code for StrongARM 1110 on Intrinsyc CERF board.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */


/* Use notes for SA-1110:
 * - The SIB interface is on USART-4, MCP mode
 * - MCP is at INT18 (SSP would be at INT19)
 *
 * The MCP interface runs with either (3.6864*13) /5 or /4,
 * which is 9.58464 MHz or 11.9808 MHz.  The CODEC's sample rate
 * is defined by this base clock, using an internal 7 bit divisor.
 * The CODEC takes the SCLK clock, doubles it, 64-oversamples and divides.
 * Which means it divides it by 32 times the programmed divisor value.
 * For instance, to get an audio sample rate of 44.100 KHz:
 * SCLK   DIV   RATE    ERROR
 *  9.x     6   49.920  +13.2%
 *  9.x     7   42.789  - 3.0%
 * 11.x     8   46.800  + 6.1%
 * 11.x     9   41.600  - 5.7%
 * A lower rate of 11.025 KHz gets
 * SCLK   DIV   RATE    ERROR
 *  9.x    27   11.093  + 0.6%
 *  9.x    28   10.697  - 3.0%
 * 11.x    33   11.345  + 2.9%
 * 11.x    34   11.011  - 0.12%
 * ==> the errors are huge.
 */

/* Use notes for Intrinsyc's BoB board:
 *
 * Touch screen is on connector J18: TSPY,1 TSMX,2 TSMY,3 TSPX,4
 * AD0-3,ADSYNC are on connector J20: AD0,7 AD1,5 AD2,3 AD3,1 ADSYNC,9
 * Microphone is on J17: MICP on outside, GNDANALOG tip
 * Speaker is on J16: SPRKP on outside, GNDANALOG tip
 * Telco modem is unconnected.
 * IO0-7 are on J15 (IO7,1 IO6,3 .. IO0,15), IO8,9 grounded.
 * RESET comes from the button
 */

/* In USART space, starting at 0x80060000 */
#define CONTROL0(_this_)      ((IO_PORT)((_this_)->PortBase + 0x0))
#define DATA0(_this_)         ((IO_PORT)((_this_)->PortBase + 0x8))
#define DATA1(_this_)         ((IO_PORT)((_this_)->PortBase + 0xc))
#define DATA2(_this_)         ((IO_PORT)((_this_)->PortBase + 0x10))
#define __DATA2(_base_)         ((IO_PORT)((_base_) + 0x10))
#define STATUS(_this_)        ((IO_PORT)((_this_)->PortBase + 0x18))
#define __STATUS(_base_)        ((IO_PORT)((_base_) + 0x18))

/* in PPC space, which starts at 0x90060000 */
#define CONTROL1(_this_)      ((IO_PORT)((_this_)->ControlBase + 0x30))

/* Control register 0 */
#define ASD   0x0000007f    /* audio sample rate divisor (6..127 cuz codec) */
#define TSD   0x00007f00    /* telecom sample rate divisor (16..127) */
#define MCE   0x00010000    /* MCP enable */
#define ECS   0x00020000    /* external clock select */
#define ADM   0x00040000    /* A/D data sampling mode (1 -> use ASD/TSD) */
#define TTE   0x00080000    /* Telecom Transmit fifo interrupt  enable */
#define TRE   0x00100000    /* Telecom Receive fifo interrupt  enable */
#define ATE   0x00200000    /* Audio Transmit fifo interrupt  enable */
#define ARE   0x00400000    /* Audio Receive fifo interrupt  enable */
#define LBM   0x00800000    /* Loopback mode */
#define ECP   0x03000000    /* External clock prescaler - 1 */

/* Control register 1 */
#define CFS   0x00100000    /* 0 -> 11.981 MHz, 1 -> 9.585 MHz */
# define CFS_9MHz   CFS
# define CFS_11MHz  0

/* Data register 0
 * READ -> bottom of audio receive FIFO
 * WRITE -> top of audio xmit FIFO
 * NB: 4 lsb are zero
 */
#define AUDIO_DATA_MASK  0xfff0

/* Data register 1
 * READ -> bottom of telecom receive FIFO
 * WRITE -> top of telecom xmit FIFO
 * NB: 2 lsb are zero
 */
#define TELCO_DATA_MASK  0xfffc

/* Data register 2
 * READ -> last value returned by codec for given register number
 * WRITE -> codec register number and value to write to
 */
#define CODEC_REGNO_MASK 0x001e0000
# define CODEC_REGNO_SET(_x_)  (((_x_)<<17)&CODEC_REGNO_MASK)
# define CODEC_REGNO_GET(_x_)  (((_x_)&CODEC_REGNO_MASK)>>17)
#define CODEC_WRITE      0x00010000
#define CODEC_READ       0x00000000
#define CODEC_VALUE_MASK 0x0000ffff

/* Status register */
#define ATS   0x00000001    /* audio xmit fifo service request */
#define ARS   0x00000002    /* audio receive fifo service request */
#define TTS   0x00000004    /* telco xmit fifo service request */
#define TRS   0x00000008    /* telco receive fifo service request */
#define ATU   0x00000010    /* audio xmit fifo underrun */
#define ARO   0x00000020    /* audio receive fifo overrun */
#define TTU   0x00000040    /* telco xmit fifo underrun */
#define TRO   0x00000080    /* telco receive fifo overrun */
#define ANF   0x00000100    /* audio xmit not full */
#define ANE   0x00000200    /* audio recv not empty */
#define TNF   0x00000400    /* telco xmit not full */
#define TNE   0x00000800    /* telco recv not empty */
#define CWC   0x00001000    /* codec write completed */
#define CRC   0x00002000    /* codec read completed */
#define ACE   0x00004000    /* audio codec is enabled */
#define TCE   0x00008000    /* telco codec is enabled */

/* Customization macros
 */

/* (Re)Initialization
 */
#define UCB_COMM_INIT(_this_) \
{ \
   IOSpaceWriteInt32(CONTROL0(_this_),0); \
   Delay(1000); \
   IOSpaceWriteInt32(CONTROL1(_this_),CFS_11MHz); \
   IOSpaceWriteInt32(STATUS(_this_),0); \
   IOSpaceWriteInt32(CONTROL0(_this_),MCE); \
}

/* Debug print
 */
#define UCB_COMM_DUMP(_this_,_name_) \
{ \
    _tprintf(_T("%s: ctl0 x%x  ctl1 x%x  status x%x  data2 x%x\r\n"), \
             _name_, \
             IOSpaceReadInt32(CONTROL0(_this_)), \
             IOSpaceReadInt32(CONTROL1(_this_)), \
             IOSpaceReadInt32(STATUS(_this_)), \
             IOSpaceReadInt32(DATA2(_this_))); \
}

/* Is the Audio codec active ?
 */
#define UCB_AUDIOACTIVE(_this_) \
   (IOSpaceReadInt32(STATUS(_this_)) & ACE)

/* Is the Telco codec active ?
 */
#define UCB_TELCOACTIVE(_this_) \
   (IOSpaceReadInt32(STATUS(_this_)) & TCE)

/* Write one sample to the audio codec FIFO
 */
#define UCB_AUDIOWRITE(_this_,_value_) \
{ \
   IOSpaceWriteInt32(DATA0(_this_),((_value_)<<4)& AUDIO_DATA_MASK); \
}

#define UCB_WRITE_XMIT_BUFFER(_this_,_value_) \
{ \
   IOSpaceWriteInt32(DATA0(_this_),(_value_) & AUDIO_DATA_MASK); \
}


/* Read one sample from the audio codec FIFO
 */
#define UCB_AUDIOREAD(_this_,_value_) \
{ \
   _value_ = (IOSpaceReadInt32(DATA0(_this_))&AUDIO_DATA_MASK)>>4; \
}

/* Write one sample to the telco codec FIFO
 */
#define UCB_TELCOWRITE(_this_,_value_) \
{ \
   IOSpaceWriteInt32(DATA1(_this_),((_value_)<<4)& TELCO_DATA_MASK); \
}

/* Read one sample from the telco codec FIFO
 */
#define UCB_TELCOREAD(_this_,_value_) \
{ \
   _value_ = (IOSpaceReadInt32(DATA1(_this_))&TELCO_DATA_MASK)>>4; \
}

/* Can we post one audio sample
 */
#define UCB_COMM_AUDIO_CANWRITE(_this_) \
   (IOSpaceReadInt32(STATUS(_this_)) & ANF)

#define UCB_XMIT_BUFFER_READY UCB_COMM_AUDIO_CANWRITE
#define UCB_INTERRUPTS_ON_XMIT_EMPTY 1

/* Can we take one audio sample
 */
#define UCB_COMM_AUDIO_CANREAD(_this_) \
   (IOSpaceReadInt32(STATUS(_this_)) & ANE)

/* Can we post one telco sample
 */
#define UCB_COMM_TELCO_CANWRITE(_this_) \
   (IOSpaceReadInt32(STATUS(_this_)) & TNF)

/* Can we take one telco sample
 */
#define UCB_COMM_TELCO_CANREAD(_this_) \
   (IOSpaceReadInt32(STATUS(_this_)) & TNE)

/* Interrupt status
 */
#define UCB_READ_INTR_PORT(_this_,_value_) \
   _value_ = IOSpaceReadInt32(STATUS(_this_))

#define UCB_ACK_INTERRUPTS(_this_) /* nothing */

#define UCB_ISR_TRANSMITTED_DATA(_this_,_value_) \
   ((_value_) & ATS)

/* Interrupt enable/disable
 */
#define UCB_ENABLE_TRANSMIT_INTERRUPTS(_this_) \
{ \
   IOSpaceWriteInt32(CONTROL0(_this_), ATE | \
                     IOSpaceReadInt32(CONTROL0(_this_))); \
}

#define UCB_DISABLE_TRANSMIT_INTERRUPTS(_this_) \
{ \
   IOSpaceWriteInt32(CONTROL0(_this_), \
                     IOSpaceReadInt32(CONTROL0(_this_)) & ~ATE); \
}

/* Errors
 */
#define UCB_ERRORS(_this_,_intr_) \
   ((IOSpaceReadInt32(STATUS(_this_))) & (ATU|ARO|TTU|TRO))

#define UCB_CLEAR_ERROR(_this_,_intr_) \
{ \
   IOSpaceWriteInt32(STATUS(_this_),_intr_); \
   IOSpaceWriteInt32(STATUS(_this_),0); \
}

/* Compute best divisor value for sample rate
 * NB: doing this right could take quite a bit of code.
 * NB: in certain cases the lower frequency could be better!
 */
#define UCB_DIVISOR(_this_,_rate_,_divisor_) \
{ \
   UINT d = 11980800 * 2; /* 3.6864*13/4 plus one bit for rounding */ \
   d = d / (32 * _rate_); /* divide */ \
   _divisor_ = (d + 1) >> 1; /* roundup */ \
}

/* Set the audio sample rate
 */
#define UCB_COMM_AUDIORATE(_this_,_rate_,_divisor_) \
{ \
   UINT v; \
   v = IOSpaceReadInt32(CONTROL0(_this_)) & ~ASD; \
   IOSpaceWriteInt32(CONTROL0(_this_),v|((_divisor_)&ASD)|ADM); \
}

/* Set the telco modem sample rate
 */
#define UCB_COMM_TELCORATE(_this_,_rate_,_divisor_) \
{ \
   UINT v; \
   v = IOSpaceReadInt32(CONTROL0(_this_)) & ~TSD; \
   IOSpaceWriteInt32(CONTROL0(_this_),v|(((_divisor_)<<8)&TSD)|ADM); \
}

/* Needs real code from here on
 */
#include <mmlite.h>
#include <mmhal.h>
#include <drivers/drivers.h>

/* Read from codec register
 */
UINT16 UcbRegRead(IO_PORT Base, UINT Regno)
{
   int i;

   /* Say what we want
    */
   IOSpaceWriteInt32(__DATA2(Base),CODEC_REGNO_SET(Regno));

   /* Wait until its done reading (next frame).
    * Dont get stuck here waiting though.
    */
   for (i = 0; i < 100; i++) {
       Delay(10); /* at 11 MHz 128 bits is about 10us */
       if (IOSpaceReadInt32(__STATUS(Base)) & CRC)
           break;
   }
   return (UINT16)(IOSpaceReadInt32(__DATA2(Base)) & CODEC_VALUE_MASK);
}

#define UCB_REGREAD(_this_,_regno_,_value_) \
{ \
   _value_ = UcbRegRead((_this_)->PortBase,_regno_); \
}

/* Write to codec register
 */
BOOL UcbRegWrite(IO_PORT Base, UINT Regno, UINT16 Value)
{
   int i;

   /* Say what we want
    */
   IOSpaceWriteInt32(__DATA2(Base),
                     CODEC_REGNO_SET(Regno)|CODEC_WRITE|
                       ((Value)&CODEC_VALUE_MASK));

   /* Wait until its done writing (next frame).
    * Dont get stuck here waiting though.
    * NB: Yes we dont *have* to wait, except if this is called right back.
    *     Because of the way the CRC/CWC bits work.. there is no easy way
    *     to detect that the previous write had not completed yet. Sigh.
    */
   for (i = 0; i < 100; i++) {
       Delay(10); /* at 11 MHz 128 bits is about 10us */
       if (IOSpaceReadInt32(__STATUS(Base)) & CWC)
           break;
   }
   return (i < 100);
}

#define UCB_REGWRITE(_this_,_regno_,_value_) \
{ \
   (void) UcbRegWrite((_this_)->PortBase,_regno_,(UINT16)(_value_)); \
}

/*  The actual driver code
 */
#include "ucb1100.h"
