/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    null.cpp
 *
 * Purpose:
 *
 *    The null device.  Useful as data sink, or zero-length input file.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <diagnostics.h>
#include <drivers/drivers.h>

/*
 * Concrete class definition for a null device
 */

typedef struct CDevice CNullDevice, *PCNullDevice;


/* Read SIZE bytes from null device, which always fails with EOF.
 */
PRIVATE SCODE MCT NullReadAt (PCNullDevice this, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    UnusedParameter(this);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);

    if (pSizeRead)
        *pSizeRead = 0;
    return S_FALSE; /* EOF */

}

/* Write SIZE bytes to serial line, which always succeeds.
 */
PRIVATE SCODE MCT NullWriteAt (PCNullDevice this, UINT64 Position, const BYTE *pBuffer,
                            UINT Size, PUINT pSizeWritten)
{
    UnusedParameter(this);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);

    if (pSizeWritten)
        *pSizeWritten = Size;
    return NO_ERROR;
}

/* GetSize method on null device
 */
PRIVATE SCODE MCT NullGetSize (PCNullDevice this, PUINT64 pSize)
{
    UnusedParameter(this);
    if (pSize)
        Uint32ToUint64(*pSize,0);
    return NO_ERROR;
}

struct CDeviceVtbl NullDriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    NullReadAt,
    NullWriteAt,
    NULL,
    NullGetSize,
    NULL,
    NULL,
    NULL,
    NULL
};

/* Constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE NullNew(void)
{
    PCDEVICE pNull;
    PIHEAP pH = CurrentHeap();

    pNull = (PCDEVICE) pH->v->Alloc( pH, HEAP_ZERO_MEMORY, sizeof(CNullDevice), 0 );
    if (pNull) {
        pNull->v = &NullDriverVtable;
        pNull->RefCnt = 1;
    }
    return pNull;
}

/* ** 
*/
SCODE __cdecl NullConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Name);
    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("NullConstructor(%x %s %x %x %x)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) NullNew();

    DBGME(0,_tprintf(_T("NullConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("NullConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl NullDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) NullConstructor
};

static struct _DRIVERCOB NullDriverCobObject = { &NullDriverCobVtable, 1 };

PIUNKNOWN NullCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &NullDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &NullDriverVtable, _T("CDevice"),
                          &IID_IDevice);
    if (FAILED(sc))
        return NULL;

    /* Null the thread loop, to save a useless thread
     */
    NullDriverVtable.ThreadLoop = NULL;

    return (PIUNKNOWN) &NullDriverCobObject;
}

