/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    config.cpp
 *
 * Purpose:
 *
 *    Configuration code for MMOSA device drivers.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "drivers.h"
#include "config.h"

EXTERN_C UINT __cdecl atoh(const char *);

#undef Free     //BUGBUG conflicts with IHeap->Free()

#define JVHPLSFIXME 1
#if JVHPLSFIXME==0
typedef PCDEVICE (*DEVICE_FACTORY)(PUINT, ...);
#else
extern PIBASERTL pTheBaseRtl;
typedef PCDEVICE (*DEVICE_FACTORY)(PIBASERTL, ...);
#endif

#define MAX_ARGS 16
#define MAX_LINE 512 /* bytes!! */

#define COMMENT_MARK '!'
#define OBJECT_MARK ':'

/* Read a new line of input, eats CR/LF.
 * Returns TRUE iff there is something in pLine.
 * Works for both plain ASCII and UNICODE.
 * The unicode parameter is false when reading an ASCII file.
 */
BOOL ReadLine(IFile * Input, PUINT pOffset, BOOL unicode, _TCHAR *pLine)
{
    SCODE Sc;
    UINT nBytes, BytesSoFar;
    _TCHAR c;
    BOOL Ret;
    UINT64 Offset;
    UINT charsize = unicode ? sizeof(_TCHAR) : sizeof(char);

    Ret = FALSE;
    BytesSoFar = 0;

    do {
        Uint32ToUint64(Offset, *pOffset);

        Sc = Input->v->ReadAt( Input, Offset, (PBYTE) &c, charsize, &nBytes);
        if (FAILED(Sc) || nBytes == 0)
            return Ret;

        *pOffset += nBytes;
#ifdef _UNICODE
        if (! unicode) c = (_TCHAR)*(char *)&c;
#endif
        *pLine++ = c;

        if ((c == '\n') || (c == '\r')) {
            *--pLine = _EOS;

            /* If the line has only one of CR/LF all is fine.
             * If it has both (either order) the first is
             * eaten now and we return, the second is eaten on
             * the next call and we continue.
             * We also take care of empty lines.
             */
            if (Ret)
                return Ret;
            else
                continue;
        }

        /* Got something
         */
        Ret = TRUE;

        /* Make sure we dont get too much, leave a hole
         * at the end for an EOS to truncate wild lines.
         */
        BytesSoFar += nBytes;

    } while (BytesSoFar < (MAX_LINE - sizeof c));

    /* If we get here we got a wild line. Terminate it.
     */
    *pLine = _EOS;

    return Ret;
}

/*
 * Skip until separator, starting from non-separator.
 * Return NULL if string exhausted, else pointer to separator
 */
_TCHAR *SkipToBlank( _TCHAR *String)
{
    _TCHAR c;

    if (String == NULL)
        return NULL;

    do {
        c = *++String;
    } while (c != OBJECT_MARK && c != '\t' && c != ' ' && c != _EOS);

    return (c == _EOS) ? NULL : String;
}

/*
 * Skip all separators, start anywhere
 * Return NULL if string exhausted, else pointer to non-separator
 */
_TCHAR *SkipBlanks( _TCHAR *String)
{
    _TCHAR c;

    if (String == NULL)
        return NULL;

    c = *String;
    while ((c == OBJECT_MARK || c == '\t' || c == ' ') && c != _EOS)
        c = *++String;

    return (c == _EOS) ? NULL : String;
}

/*
 * Parse a configuration line.
 *
 * Syntax
 *
 * Line == Comment | Config
 * Comment = ! [ Anything ]* NewLine
 * Config == [ Name ] :: Constructor [ : Option ]* [ Blank Arg ]*
 * Option == NetworkDriver | NoInterrupts | NakedIsr | Dll = String
 * Blank == [ space | tab ]*
 * Arg == HexNumber | String
 *
 * Examples:
 *      ! comment to remember what the args are
 *      Wave::BlasterNew 220 7 1 6 300
 *      ::LanceNew:NetworkDriver 360 5 7
 *      Mouse::MouseNew:NoInterrupts Com1
 */

_TCHAR *ParseLine( _TCHAR *String, _TCHAR **ppName, _TCHAR **ppConstr,
                PUINT pOptions, _TCHAR **ppDllName)
{
    _TCHAR *Opt, c;

    /* Initialize ifnot
     */
    *ppName = *ppConstr = *ppDllName = NULL;
    *pOptions = 0;

    /* Do we have a device name
     */
    if (*String != OBJECT_MARK) {

        /* Yes we do
         */
        *ppName = String;

        /* Look for end of name
         */
        String = SkipToBlank(String);

        /* Terminate name
         */
        *String++ = _EOS;

    } else
        *ppName = NULL;


    /* We must have a constructor
     */
    String = SkipBlanks(String);
    *ppConstr = String;

    String = SkipToBlank(String);
    if (String == NULL)
        return NULL;

    c = *String;
    *String++ = _EOS;


    /* Do we have any options
     */
    *pOptions = DEVICE_FLAGS_INTERRUPTS;
    while (c == OBJECT_MARK && String) {

        Opt = SkipBlanks(String);
        String = SkipToBlank(Opt);
        if (String) {
            c = *String;
            *String++ = _EOS;
        }

        if (!_tcscmp(Opt, _T("NoInterrupts")))
            *pOptions = (*pOptions | DEVICE_FLAGS_STACKED_DRIVER)
                        & ~DEVICE_FLAGS_INTERRUPTS;

        else if (!_tcscmp(Opt, _T("NetworkDriver")))
            *pOptions |= DEVICE_FLAGS_NET_DRIVER;
        
        else if (!_tcscmp(Opt, _T("NakedIsr")))
            *pOptions |= DEVICE_FLAGS_NAKED_ISR;
        
        else if (!_tcscmp(Opt, _T("LevelTrigger")))
            *pOptions |= DEVICE_FLAGS_LEVEL_TRIGGER;

        else if (!_tcscmp(Opt, _T("PositiveLevel")))
            *pOptions |= DEVICE_FLAGS_POSITIVE_LEVEL;

        else if (!_tcsncmp(Opt, _T("Dll="), 4)) {
            *ppDllName = Opt + 4;
        }
        
        else printf(_T("Unrecognized device option \"%s\"\n"), Opt);

    }

    return SkipBlanks(String);
}

/*
 * Parse arguments to be passed to a new device.
 * Arguments are of three types, identified by
 * their leading char: strings, hex, decimals.
 */
void ParseArgs( _TCHAR *String, PUINT pArgs)
{
    while ((String = SkipBlanks(String)) != NULL) {
        _TCHAR *Arg, c;

        /* Isolate this argument
         */
        Arg = String;
        String = SkipToBlank(Arg);
        if (String)
            *String++ = _EOS;

        c = *Arg;

        /* A hex ?
         */
        if (c == '0' && (Arg[1] == 'x' || Arg[1] == 'X')) {
            *pArgs++ = atoh(Arg+2); /* XXX Unicode */
        }

        /* A decimal number ?
         */
        else if (c >= '0' && c <= '9') {
            *pArgs++ = _ttoi(Arg);
        }

        /* Must be a string then
         */
        else {
            *pArgs++ = (UINT) Arg;
        }
    }
}

#if !USE_DLLS
EXTERN_C PCDEVICE RTLCALLTYPE SerialNew(PUINT pPortBase, PUINT pIrq, PUINT pBaudRate);
EXTERN_C PCDEVICE RTLCALLTYPE Lan9000New(PUINT pPortBase, PUINT pIrq, PUINT pAttribBase, PUINT pPhyType);

/*
 * Search for the constructor function address from the list
 * of statically loaded drivers.
 */
SCODE GetConstructorAddress ( _TCHAR *NameOfNew, ADDRESS *ConsAddress )
{
    int i;
    static first = TRUE;
    static struct {
        _TCHAR  *Name;
        ADDRESS  Addr;
    } ConTbl[] = {
        //{"Lan9000New", 0},
        {"SerialNew", 0},
        {0,0}
    };

    /*
     * Arm compiler doesn't let us statically initialize ConTbl.
     */
    if (first) {
        first = FALSE;
        //ConTbl[0].Addr = (ADDRESS) Lan9000New;
        ConTbl[0].Addr = (ADDRESS) SerialNew;
    }

    for (i = 0; ConTbl[i].Name; ++i)
        if (0 == strcmp(ConTbl[i].Name, NameOfNew)) {
            *ConsAddress = ConTbl[i].Addr;
            return NO_ERROR;
        }

    return E_INVALID_PARAMETER;
}
#endif

/*
 * Scan the configuration file.
 * Create each device it mentions,
 * allocate resources to it,
 * register it in the name space.
 *
 * At the end, returns a count of devices that failed to
 * startup properly.  Returns 0 if all was well.
 * [An empty config file counts as success]
 */
UINT ParseFile ( IModule *pModule, IFile * pConfig, BOOL LoadDlls )
{
    UINT ret = 0;
    SCODE Sc;
    _TCHAR *Name, *NameOfNew, *pArgs, *pLine, *DllName;
    UINT Options, Offset = 0;
    UINT Arguments[MAX_ARGS];
    ADDRESS ConsAddress = 0;
    INameSpace *pIName = NULL;
    PCDEVICE NewDevice;
    DEVICE_FACTORY Constructor;
    BOOL unicode = FALSE;

    /* Check the handle on our image */
    if (pModule == NULL) {
            printf(_T("Donno what module I am ?!?\n"));
            return 1;
    }

    /* Allocate input buffer */
    pLine = (_TCHAR*) CurrentHeap()->v->Alloc( CurrentHeap(), 0, MAX_LINE, 0);
    if (pLine == NULL) {
            printf(_T("Memory allocation failure\n"));
            return 1;
    }

#if defined(_UNICODE)
    {
        _TCHAR firs_TCHAR;
        UINT count;

        /* Unicode stream starts with a Byte-Order-Mark (BOM) 0xFEFF.
         * Otherwise it's an ASCII stream.
         */
        Sc = pConfig->ReadAt((UINT64)0, (BYTE *)&firs_TCHAR, sizeof(_TCHAR), &count);
        if (SUCCEEDED(Sc) && count == sizeof(_TCHAR)
            &&  firs_TCHAR == UNICODE_BYTE_ORDER_MARK)
        {
            unicode = TRUE;
            Offset = sizeof(_TCHAR);
        }
#if 0
        printf(_T("config: %hs drivers.dat file\n"),
               unicode ? "Unicode" : "ASCII");
#endif
    }
#endif

    /* Do the scanning proper */
    while (ReadLine(pConfig, &Offset, unicode, pLine)) {

        /* Skip comments */
        if (*pLine == COMMENT_MARK)
            continue;

        /* Got something to do, parse it */
#if 0
        printf(_T("config: Parsing line \"%s\"\n"), pLine);
#endif
        pArgs = ParseLine( pLine, &Name, &NameOfNew, &Options, &DllName );
        if (NameOfNew == NULL)
            /* garbled line */
            continue;

        /* During initialization of *this* DLL the loader prevents
         * us from loading other DLLs that depend on us. Catch-22.
         *
         * The workaround for this limitation (why cant we just let
         * the loader know we are done initializing ?) is to split
         * driver creation in two: static and dynamic.
         *
         * All drivers that have been statically linked into this DLL
         * can be 'statically' created during this DLL's initialization.
         * The configuration file can specify 'dynamic' drivers, e.g.
         * drivers which live in a separate DLL which we have to load
         * ourselves.  These drivers *must* be created after this DLL
         * has completed initialization.  A separate program will have
         * to invoke us again, with proper arguments.
         *
         * Therefore, LoadDlls and DllName decide what we do.
         */
        if (LoadDlls && DllName == NULL)
            continue;
        if (!LoadDlls && DllName != NULL)
            continue;

        /* Load a separate DLL if we have to (despised) */
        if (DllName != NULL) {
            IModule *pNewModule;
            IProcess *pPrc = CurrentProcess();

            Sc = pPrc->v->LoadImage( pPrc, DllName, NULL, 0, &pNewModule );
            if (FAILED(Sc)) {
                printf(_T("%x Could not load DLL \"%s\"\n"), Sc, DllName);
                ConsAddress = (ADDRESS) NULL; /* Quiet warning */
            } else {
                Sc = pNewModule->v->GetFunctionAddress( pNewModule, NameOfNew,
                                                        0, &ConsAddress);
#if 0
                /* DONT. We are holding the only reference */
                (void) pNewModule->Release();
#endif
            }
        } else {

#if !USE_DLLS
            Sc = GetConstructorAddress(NameOfNew, &ConsAddress);
#else
            /* Lookup the constructor function for this device */
            DllName = _T("this dll");
            Sc = pModule->v->GetFunctionAddress( pModule, NameOfNew, 0,
                                                 &ConsAddress);
#endif
        }

        if (FAILED(Sc)) {
            _tprintf(_T("Could not find constructor \"%s\" in \"%s\"\n"),
                     NameOfNew, DllName);
            ret++;
            continue;
        }

        /* Do the argument parsing now */
        memset(Arguments, 0, sizeof Arguments);
        ParseArgs(pArgs, Arguments);

        /* Cast properly */
        Constructor = (DEVICE_FACTORY) ConsAddress;

        /* Create the new device */
#if JVHPLSFIXME==0
        _tprintf(_T("config: About to call %s (0x%x)\n"),
                 NameOfNew, ConsAddress);

        NewDevice = Constructor(Arguments + 0, Arguments + 1, Arguments + 2,
                                Arguments + 3, Arguments + 4, Arguments + 5,
                                Arguments + 6, Arguments + 7, Arguments + 8,
                                Arguments + 9, Arguments + 10, Arguments + 11,
                                Arguments + 12, Arguments + 13,
                                Arguments + 14, Arguments + 15);
#else
        
        NewDevice = Constructor(pTheBaseRtl, 
                                Arguments + 0, Arguments + 1, Arguments + 2,
                                Arguments + 3, Arguments + 4, Arguments + 5,
                                Arguments + 6, Arguments + 7, Arguments + 8,
                                Arguments + 9, Arguments + 10, Arguments + 11,
                                Arguments + 12, Arguments + 13,
                                Arguments + 14, Arguments + 15);
#endif

        if (NewDevice == NULL) {
            ret++;
            continue;
        }

        /* All is well so far. Connect to the OS. */
        Sc = ConnectDevice( NewDevice, Name, Options, Arguments[0],
                            Arguments[1]);
        if (FAILED(Sc)) {
            printf(_T("Could not connect %s\n"), NameOfNew);
            NewDevice->v->Release(NewDevice);
            /* RemoveDevice( NewDevice ); */
            /* NewDevice->Unload() */
            ret++;
            continue;
        }

        /* Do we have to register the new device in the name space */
        if (Name) {
            if (pIName == NULL) 
                pIName = CurrentNameSpace();
            Sc = pIName->v->Register(pIName, Name, (IUnknown *) NewDevice, 
                                     NAME_SPACE_READ | NAME_SPACE_WRITE, NULL);
            if (FAILED(Sc))
                printf(_T("Failed to register %s as %s, error code %x\n"), NameOfNew, Name, Sc);
        }

        /* Say what happened */
        printf(_T("%s::%s started ok.\n"), (Name) ? Name : _T(""), NameOfNew);
    }

    CurrentHeap()->v->Free(CurrentHeap(), 0, pLine);

    return ret;
}


/*
 * Do the configuration
 */
void ConfigureDrivers( PIMODULE pModule,
                       _TCHAR *ConfigurationFileName,
                       BOOL LoadDllDrivers )
{
    UINT ret;
    IUnknown * pUnk;
    IFile * pFile;
    SCODE Sc;

    /* Lookup the configuration file */
    Sc = CurrentNameSpace()->v->Bind( CurrentNameSpace(),
                                      ConfigurationFileName, NAME_SPACE_READ,
                                      &pUnk);
    if (FAILED(Sc)) {
        _tprintf(_T("Could not bind to %s (x%x)\n"),
                 ConfigurationFileName, Sc);
        return;
    }

    /* Make sure its a file and not something else */
    Sc = pUnk->v->QueryInterface( pUnk, &IID_IFile, (void**)&pFile);
    if (FAILED(Sc)) {
        _tprintf(_T("%s is not an IFile object (x%x)\n"),
                 ConfigurationFileName, Sc);                
        return;
    }

    pUnk->v->Release(pUnk);

    /* Do the parsing */
    ret = ParseFile(pModule, pFile, LoadDllDrivers );

    /* Let user know if anything went wrong
     * BUGBUG need logging facility...
     */
    if (ret > 0)
        printf(_T(" %d device driver(s) failed to initialize\n"), ret);

    /* Done, let the file go and return */
    pFile->v->Release(pFile);
}
