/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    connect.c
 *
 * Purpose:
 *
 *    Connect a device driver to the rest of the system.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <diagnostics.h>
#include <mmhal.h>
#define DLL_LINKAGE DLL_EXPORT
#include <drivers/drivers.h>
#include "config.h"

#ifdef _MSC_VER
#pragma warning(disable:4054)   /* cast from function ptr to data ptr (/Za) */
#endif

extern void THREAD_LINKAGE DeviceThread(THREAD_ARGUMENT);

SCODE ConnectDevice(PCDEVICE Device, _TCHAR * Name, UINT Flags,
                    UINT RegisterBase, UINT Irq)
{
    SCODE sc;
    PIPROCESS pIPrc;
    PIMODULE pIMod; 
    PITHREAD pThd;
    PTR pIsr = (PTR) Device->v->Isr;
    PIENDPOINTFACTORY mbf;
    PCNETDRIVER NetDev = (PCNETDRIVER) Device;

    /* Make sure this is initialized properly.
     * Do it before we enable ISRs, which might use it
     */
    Condition_Init(&Device->IntrPoint);

    /* Make MBufFactory accessible before installing ISR */
    if (Flags & DEVICE_FLAGS_NET_DRIVER) {

        /* BUGBUG Assume that the protocol COB has been loaded already */
        sc = BindToObject(CurrentNameSpace(), _T("COB/protocol.cob"),
                          NAME_SPACE_READ, &IID_IEndpointFactory,
                          (void**) &mbf);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("MBufFactory bind failed sc=%x\n"), sc));
            return sc;
        }

        NetDev->MBufFact = mbf;
    }

    if (Flags & DEVICE_FLAGS_INTERRUPTS) {
#if 0
        /* For now this is mips-specific, but some other
         * future architecture will probably use it as well
         */
        if (Flags & DEVICE_FLAGS_NAKED_ISR)
            pIsr = (PTR)((ADDRESS)pIsr | 1);
#endif
        sc = AddDevice( Device, pIsr, Flags, Irq, NULL );
        if (FAILED(sc))
            return sc;

    }

    /* Synchronously initialize device */
    (void) Device->v->Initialize(Device);

    if (Device->v->ThreadLoop && !(Flags & DEVICE_FLAGS_NO_THREAD)) {

        /* Use the device's module (via the ISR) rather than CurrentProcess()
         * to account the new thread against the driver's dll rather than
         * against the calling process.  This makes the calling process
         * go away properly.  Explanation: CurrentProcess() returns the
         * current thread's process, i.e. dynamic scope.  The current
         * module, however is statically scoped.
         */
        sc = GetModuleContainsAddr((ADDRESS)pIsr, &pIMod);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("No module for ISR ?!?\n")));
            return sc;
        }

        sc = pIMod->v->QueryInterface(pIMod, &IID_IProcess, (void**)&pIPrc);
        pIMod->v->Release(pIMod);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("No process for ISR ?!?\n")));
            return sc;
        }

        /* Start handler thread
         */
        sc = pIPrc->v->CreateThread(pIPrc, CREATE_THREAD_STATIC_SCOPING,
                                    (THREAD_FUNCTION) Device->v->ThreadLoop,
                                    (THREAD_ARGUMENT) Device,
                                    0, NULL, &pThd);
        pIPrc->v->Release(pIPrc);
        DBGME(2,_tprintf(_T("drivers: Created %s thread %x sc=%x\n"),Name,(ADDRESS)pThd,sc));
        if (FAILED(sc))
            return sc;
    }

    /* Finally, if a network interface, notify protocol stack */
    if (Flags & DEVICE_FLAGS_NET_DRIVER) {
        sc = NetDev->MBufFact->v->IfAttach(NetDev->MBufFact,
                                           (PINETDRIVER) Device);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("IfAttach failed sc=%x\n"),sc));
            return sc;
        }
    }

#if 0
    if (Flags & DEVICE_FLAGS_NET_DRIVER)
    {
        PIUNKNOWN pUnk = NULL;
        PIENDPOINTFACTORY pStack = NULL;
        PIENDPOINTFACTORY pDriver = NULL;

        sc = Device->v->QueryInterface(Device, &IID_IEndpointFactory, (void**) &pDriver);
        if (FAILED(sc)) DebugBreak();
        sc = BindToObject(CurrentNameSpace(),
                          "COB/protocol.cob", NAME_SPACE_READ,
                          &IID_IEndpointFactory, (void**) &pStack);
        if (FAILED(sc)) DebugBreak();
        sc = pStack->v->ConnectComponent(pStack, (PTR) pDriver, 0x1);
        if (FAILED(sc)) DebugBreak();
        pDriver->v->Release(pDriver);
        pUnk->v->Release(pUnk);
        pStack->v->Release(pStack);
  }
#endif
    return S_OK;
}
