/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    drivers.cpp
 *
 * Purpose:
 *
 *    Initialization and control code for MMLite device drivers.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <mmhal.h>
#include <assert.h>
#define DLL_LINKAGE DLL_EXPORT
#include <drivers/drivers.h>

/*
 * --------------------------------------------------------------------------
 * Shared default methods
 * True implementation is in C to allow sharing from both C and C++.
 */
/* Memory allocation
 */
EXTERN_C DLL_EXPORT void * operator_new( UINT size)
{
    PIHEAP pHeap = CurrentHeap();
    return pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, size, 0);
}

/* Memory release
 */
EXTERN_C DLL_EXPORT void operator_delete( void *pThis)
{
    PIHEAP pHeap = CurrentHeap();
    (void) pHeap->v->Free(pHeap, 0, pThis);
}

/* Methods for the CDevice class
 */

/* This truly is IFile::QueryInterface, but the decl for
 * the IFile class does not allow it.
 * The other catch is reference counting.
 */
EXTERN_C DLL_EXPORT SCODE MCT CDevice_QueryInterface (struct CDevice *pThis, REFIID Iid, void* *ppNew)
{
    if (UuidCmp(&IID_IFile,Iid) || UuidCmp(&IID_IUnknown,Iid)
        || UuidCmp(&IID_IDevice,Iid))
    {
        AtomicInc(&pThis->RefCnt);
        *ppNew = (PIUNKNOWN) pThis;
        return NO_ERROR;
    }
    *ppNew = NULL;
    return E_NO_INTERFACE;
}

EXTERN_C SCODE MCT CNetDriver_QueryInterface (struct CNetDriver *pThis, REFIID Iid, void* *ppNew)
{
    if (UuidCmp(&IID_IFile,Iid) || UuidCmp(&IID_IUnknown,Iid)
        || UuidCmp(&IID_IDevice,Iid) || UuidCmp(&IID_INetDriver,Iid))
    {
        AtomicInc(&pThis->RefCnt);
        *ppNew = (PIUNKNOWN) pThis;
        return NO_ERROR;
    }
    *ppNew = NULL;
    return E_NO_INTERFACE;
}

/* Add a reference
 */
EXTERN_C DLL_EXPORT UINT MCT CDevice_AddRef ( struct CDevice *pThis )
{
    return AtomicInc(&pThis->RefCnt);
}

/* Release a reference
 */
EXTERN_C DLL_EXPORT UINT MCT CDevice_Release ( struct CDevice *pThis )
{
    UINT Refs;

    Refs = AtomicDec(&pThis->RefCnt);
    if (Refs == 0)
        operator_delete(pThis);
    return Refs;
}

/* Invalid default methods
 */
EXTERN_C DLL_EXPORT SCODE MCT CDevice_ReadAt ( struct CDevice *pThis, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    UnusedParameter(pThis);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);
    UnusedParameter(pSizeRead);

    return E_ACCESS_DENIED; /* ? */
}

EXTERN_C DLL_EXPORT SCODE MCT CDevice_WriteAt ( struct CDevice *pThis, UINT64 Position, const BYTE *pBuffer,
                               UINT Size, PUINT pSizeWritten)
{
    UnusedParameter(pThis);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);
    UnusedParameter(pSizeWritten);

    return E_ACCESS_DENIED; /* ? */
}

EXTERN_C DLL_EXPORT SCODE MCT CDevice_SetSize(  struct CDevice *pThis, UINT64 Size)
{
    UnusedParameter(pThis);
    UnusedParameter(Size);
    return E_ACCESS_DENIED; /* ? */
}

EXTERN_C DLL_EXPORT SCODE MCT CDevice_GetSize(  struct CDevice *pThis, PUINT64 pSize)
{
    UnusedParameter(pThis);
    UnusedParameter(pSize);
    return E_ACCESS_DENIED; /* ? */
}

EXTERN_C DLL_EXPORT BOOL MCT CDevice_Isr( struct CDevice *pThis, PBOOL pWakeup)
{
    UnusedParameter(pThis);
    UnusedParameter(pWakeup);
    return FALSE;
}

EXTERN_C DLL_EXPORT BOOL MCT CDevice_HandleInterrupt( struct CDevice *pThis )
{
    UnusedParameter(pThis);
    return FALSE;
}

EXTERN_C DLL_EXPORT INT MCT CDevice_Initialize(  struct CDevice *pThis )
{
    UnusedParameter(pThis);
    return 0;
}

EXTERN_C DLL_EXPORT void MCT CDevice_ThreadLoop(  struct CDevice *pThis )
{
    BOOL Handled;
    TIME_CONSTRAINT TimeConstraint;

    /* Construct a very urgent TimeConstraint and do an
     * initial BeginConstraint
     */
    Int32ToInt64(TimeConstraint.Start, 0);/* can start immediately after wait*/
    Int32ToInt64(TimeConstraint.Estimate, 0); /* does not really matter */
    Int32ToInt64(TimeConstraint.Deadline,0); /* immediately if not sooner */
    TimeConstraint.Criticality = TRUE;

    (void) BeginConstraint(FALSE, &TimeConstraint, NULL);

    for (;;) {
        Int32ToInt64(TimeConstraint.Start, 0);
        Int32ToInt64(TimeConstraint.Estimate, 0);
        Int32ToInt64(TimeConstraint.Deadline,0);
        TimeConstraint.Criticality = TRUE;

        (void) Condition_WaitAndBeginConstraint(&pThis->IntrPoint, NULL, NULL,
                                                TRUE, &TimeConstraint, NULL);
        Handled = pThis->v->HandleInterrupt(pThis);
    }
}

/* Methods for the CNetDriver class
 */
EXTERN_C INT  MCT CNetDriver_Output(PINETDRIVER pThis,
                                    PIBUFFER Pkt,
                                    struct sockaddr *Addr,
                                    struct route_in *rti)
{
    PCNETDRIVER This = (PCNETDRIVER) pThis;
    return This->MBufFact->v->EtherOutput(This->MBufFact,
                                          pThis, Pkt, Addr, rti);
}

EXTERN_C void MCT CNetDriver_Start(struct CNetDriver *pThis)
{
    UnusedParameter(pThis);
}

EXTERN_C INT  MCT CNetDriver_Ioctl( struct CNetDriver *pThis,
                                    UINT Cmd, PTR Data)
{
    UnusedParameter(pThis);
    UnusedParameter(Cmd);
    UnusedParameter(Data);
    return 0;
}

EXTERN_C INT  MCT CNetDriver_Watchdog(PINETDRIVER This)
{
    UnusedParameter(This);
    return 0;
}

/* XXX This is here temporarily to make drivers link */
EXTERN_C DLL_EXPORT PIDMA CreateDmaChannel( PTR Description, PUINT nomr)
{
    UnusedParameter(Description);
    UnusedParameter(nomr);
    return NULL;
}


/*
 * --------------------------------------------------------------------------
 * Default vtables and constructor objects
 */
const struct IDeviceVtbl TheCDeviceVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    (_IDevice_QueryInterface_METHOD_TYPE) CDevice_QueryInterface,
    (_IDevice_AddRef_METHOD_TYPE) CDevice_AddRef,
    (_IDevice_Release_METHOD_TYPE) CDevice_Release,
    (_IDevice_ReadAt_METHOD_TYPE) CDevice_ReadAt,
    (_IDevice_WriteAt_METHOD_TYPE) CDevice_WriteAt,
    (_IDevice_SetSize_METHOD_TYPE) CDevice_SetSize,
    (_IDevice_GetSize_METHOD_TYPE) CDevice_GetSize,
    (_IDevice_Isr_METHOD_TYPE) CDevice_Isr,
    (_IDevice_HandleInterrupt_METHOD_TYPE) CDevice_HandleInterrupt,
    (_IDevice_Initialize_METHOD_TYPE) CDevice_Initialize,
    (_IDevice_ThreadLoop_METHOD_TYPE) CDevice_ThreadLoop
};

const struct INetDriverVtbl TheCNetDriverVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    (_INetDriver_QueryInterface_METHOD_TYPE) CNetDriver_QueryInterface,
    (_INetDriver_AddRef_METHOD_TYPE) CDevice_AddRef,
    (_INetDriver_Release_METHOD_TYPE) CDevice_Release,
    (_INetDriver_ReadAt_METHOD_TYPE) CDevice_ReadAt,
    (_INetDriver_WriteAt_METHOD_TYPE) CDevice_WriteAt,
    (_INetDriver_SetSize_METHOD_TYPE) CDevice_SetSize,
    (_INetDriver_GetSize_METHOD_TYPE) CDevice_GetSize,
    (_INetDriver_Isr_METHOD_TYPE) CDevice_Isr,
    (_INetDriver_HandleInterrupt_METHOD_TYPE) CDevice_HandleInterrupt,
    (_INetDriver_Initialize_METHOD_TYPE) CDevice_Initialize,
    (_INetDriver_ThreadLoop_METHOD_TYPE) CDevice_ThreadLoop,
    (_INetDriver_Output_METHOD_TYPE) CNetDriver_Output,
    (_INetDriver_Start_METHOD_TYPE) CNetDriver_Start,
    (_INetDriver_Ioctl_METHOD_TYPE) CNetDriver_Ioctl,
    (_INetDriver_Watchdog_METHOD_TYPE) CNetDriver_Watchdog
};

/* Default Drivers COB constructor object */

/* IUnknown::QI.  Return a pointer of the right kind. */
SCODE MCT DriverCobQueryInterface(PIDEVICECONSTRUCTOR pThis, REFIID Iid,
                                       void **ppObj)
{
    return GenericQueryInterface((PIUNKNOWN) pThis, Iid, ppObj,
                                 &IID_IDeviceConstructor);
}

/* IUnknown::AddRef */
UINT MCT DriverCobAddRef(PIDEVICECONSTRUCTOR pThis)
{
    PDRIVERCOB x = pDriverCob(pThis);

    return AtomicInc(&x->RefCnt);
}

/* IUnknown::Release */
UINT MCT DriverCobRelease(PIDEVICECONSTRUCTOR pThis)
{
    PDRIVERCOB x = pDriverCob(pThis);

    /* NOTE: If driver has global destructors, it will have to override
     * this method with one that calls CrtExit() when RefCnt == 0.
     */
    return AtomicDec(&x->RefCnt);
}

SCODE MCT UnimplementedMethod()
{
    return E_NOT_IMPLEMENTED;
}

static const struct IDeviceConstructorVtbl TheCDeviceConstructorVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    DriverCobQueryInterface,
    DriverCobAddRef,
    DriverCobRelease,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UnimplementedMethod
};

/* My COB Object */
typedef struct _DEFCOB {
    const struct IUnknownVtbl *v;
    UINT RefCnt;
} *PDEFCOB;

/* IUnknown::<this> points to v-table
 * pDC (pointer of My Cob) converts interface ptr to implementation.
 * iDC (interface of DC)  converts implementation ptr to interface ptr.
 */
#define pMyCob(_i_) ((PDEFCOB)(_i_))
#define iMyCob(_p_) ((PIUNKNOWN)(_p_))

/* IUnknown::QI.  Return a pointer of the right kind. */
static SCODE MCT MyCobQueryInterface(PIUNKNOWN pThis, REFIID Iid,
                                       void **ppObj)
{
    return GenericQueryInterface(pThis, Iid, ppObj, &IID_IUnknown);
}

/* IUnknown::AddRef */
static UINT MCT MyCobAddRef(PIUNKNOWN pThis)
{
    PDEFCOB x = pMyCob(pThis);

    return AtomicInc(&x->RefCnt);
}

/* IUnknown::Release */
static UINT MCT MyCobRelease(PIUNKNOWN pThis)
{
    PDEFCOB x = pMyCob(pThis);
    UINT RefCnt = AtomicDec(&x->RefCnt);

    if (RefCnt == 0) {
        /* Unregister registered classes before terminating */
        BaseRegisterClass(NULL, 0, _T("CNetDriver"), NULL);
        BaseRegisterClass(NULL, 0, _T("CDevice"), NULL);
        BaseRegisterClass(NULL, 0, _T("CDeviceConstructor"), NULL);
        /* And give CRT a chance to clean up */
        // CrtInit(DLL_PROCESS_DETACH);
    }
    return RefCnt;
}

static const struct IUnknownVtbl MyCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    MyCobQueryInterface,
    MyCobAddRef,
    MyCobRelease
};

struct _DEFCOB _MyCobObject = { &MyCobVtable, 1 };

extern struct _DEFCOB _MyCobObject;
/*
 * --------------------------------------------------------------------------
 * Library initialization, called at dll load time.
 */
PIUNKNOWN DriversCobMain(void)
{
    /* Export default classes for driver's use and configure */
    SCODE sc;
    sc = BaseRegisterClass((ADDRESS *) &TheCNetDriverVtbl,
                           sizeof(CNetDriverVtbl)/sizeof(ADDRESS),
                           _T("CNetDriver"),
                           &IID_INetDriver);
    assert(sc == S_OK);
    if (FAILED(sc))
        return NULL;

    sc = BaseRegisterClass((ADDRESS *) &TheCDeviceVtbl,
                           sizeof(struct IDeviceVtbl)/sizeof(ADDRESS),
                           _T("CDevice"),
                           &IID_IDevice);
    assert(sc == S_OK);
    if (FAILED(sc))
        return NULL;

    sc = BaseRegisterClass((ADDRESS *) &TheCDeviceConstructorVtable,
                         sizeof(struct IDeviceConstructorVtbl)/sizeof(ADDRESS),
                           _T("CDeviceConstructor"),
                           &IID_IDeviceConstructor);
    assert(sc == S_OK);
    if (FAILED(sc))
        return NULL;

//    PerformConfiguration( USE_DLLS );

    return (PIUNKNOWN) &_MyCobObject;
}
