/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    pnp.c
 *
 * Purpose:
 *
 *    Plug aNd Play manager.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <string.h>
//#define _DEBUG 1
//#define DBGLVL 0
#include <diagnostics.h>
#include <assert.h>

//#define SMALL_CUSTOM_TABLE 1
#define DYNAMIC_OBJ 0
#define USE_DELEGATE 1

/*
 * Dynamic driver startup requires some run-time lookup tables
 * to associate a physical device with the component (driver)
 * that will handle it.  This is currently known by the catch phrase
 * "plug and play", to underline the desired simplicity of use.
 * Various busses use different ways to identify their devices,
 * but it general comes down to two or three small integers that
 * identify the manufacturer and device type and perhaps revision.
 * In more standardized cases it might be a class and protocol
 * identifiers pair.  Whatever the semantic of those small integers,
 * it does not really concern us; all we need to provide is an
 * association between that triple and the name of a component.
 *
 * One idea for this is to use a namespace, using a string-ified
 * form of the triplet for lookup.  However it is the implemented
 * inside, the user will just do something like this:
 *
 *      sprintf(DriverName,"/Pnp/Vid%04x_Pid%04x_Dev%04x",
 *              VendorID,ProductID,DeviceID);
 *      BindToObject(CurrentNameSpace(),DriverName,&IID_IFile,&File);
 *	    File->ReadAt(File,..,&CobName);
 *
 * In other words, the triple is used to lookup an object in the
 * namespace (which could be anything). At this time we are just
 * asking this object to export a way to return the corresponding
 * device driver name, which we can easily do with an IFile.
 * Perhaps in the future this could become a different object,
 * but for today that will do, and it has advantages.
 *
 * Partial matching can be done with asterisks:
 *  	"/Pnp/Vid0452_Pid0666_*"
 * which would work for USB where there are usually only two IDs.
 * But getting too fancy is an overkill, I would NOT use something like this:
 *	    BindToObject( , "/Pnp/Vid0452_*", &IID_INameSpace, &Ns);
 *      Dir(Ns) -> all driver names for Microsoft devices
 * especially since we can simply impose lexical ordering of strings,
 * and use FindNext as customary.
 *
 * Two simple advantages of this scheme is that a "dir \PnP" will
 * show the list of all IDs the manager knows about, a simple practical
 * problem that gets nonetheless complicated in Windows.
 * And to find which driver handles a certain mouse you'd do just
 * a "type \PnP\Vid045e_Pid0009_Dev*" and get "COB\usbmouse.cob".
 *
 * Another advantage is that the table can be easily extended at runtime:
 *      Register(CurrentNameSpace(),"\PnP\VenNNNN_PidMMMM_DevKKKK",File);
 * where "File" only has to implement ReadAt.
 * In that vein, we might have subspaces under \PnP. Say some future
 * bus "SuperBus" uses not integers but some other form of identification.
 * It could be implemented under say "\PnP\SuperBus\<superids>"
 *
 * Another idea could be to directly start up the driver, without even
 * knowing its name or how it is implemented.
 * For instance, instead of QI for IFile one might QI for something else
 * like say IDriver.  I mean, whatever you QI for we could start the
 * component and then QI *it* for that IID.
 *
 * But thats for tomorrow, today we keep it simple.
 *
 * A possible implementation is as follows.
 */

/* Format of a database entry, to keep it small.
 */
typedef struct _PNP_ENTRY {
    /* 16 bits each for the triple
     */
    UINT16 Vid;
    UINT16 Pid;
    UINT16 Rev;
    /* Distinguished value for partial matches
     */
#define ANY_ID  0xffff
    /* The index in the string table for the driver.
     */
    UINT16 DriverIndex;
} PNP_ENTRY, *PPNP_ENTRY;

/*
 * The table itself is kept separate
 */
extern const PNP_ENTRY PnpTable[];
#define nPnpTable (sizeof PnpTable / sizeof PnpTable[0])

/* And so is the string table
 */
typedef _TCHAR DRIVER_NAME[12];
extern const DRIVER_NAME DriverTable[];
#define nDriverTable (sizeof DriverTable / sizeof DriverTable[0])

#include "pnp_table.c"

typedef struct _PNPFILE {
    const struct IFileVtbl *v;
    UINT RefCnt;
    UINT16 TableIndex;
    UINT16 StrSize; /* in bytes! */
} PNPFILE, *PPNPFILE;

/* Convert between interface and instance pointer and vice versa */
#define pPnpFile(_i_) ((PPNPFILE)((_i_)))
#define iPnpFile(_p_) ((PIFILE)((_p_)))

/* IFile::IUnknown
 */
PRIVATE SCODE PnpFileQueryInterface (
    PIFILE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("PnpFileQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));

    /* IFile only
     */
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_IFile,iid))
    {
        AtomicInc(&pPnpFile(pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT PnpFileAddRef (
    PIFILE pThis)
{
    DBGME(0,_tprintf(_T("PnpFileAR(%x,%d)\n"),(ADDRESS)pThis,pPnpFile(pThis)->RefCnt));
    return AtomicInc(&pPnpFile(pThis)->RefCnt);
}

PRIVATE UINT PnpFileRelease (
    PIFILE pThis)
{
    PPNPFILE this = pPnpFile(pThis);
    UINT Refs = AtomicDec(&this->RefCnt);
    DBGME(0,_tprintf(_T("PnpFileRel(%x,%d)\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        pHeap->v->Free(pHeap, 0, (void *) this);
    }
    return Refs;
}

PRIVATE SCODE PnpFileWriteAt (
    PIFILE pThis,
    UINT64 Position,
    const BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeWritten)
{
    DBGME(0,_tprintf(_T("PnpFileWriteAt(%x)\n"),(ADDRESS)pThis));

    UnusedParameter(pThis);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);
    UnusedParameter(pSizeWritten);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE PnpFileReadAt (
    PIFILE pThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PPNPFILE this = pPnpFile(pThis);
    UINT64 Eof;
    UINT Offset;
    const _TCHAR *String;

    DBGME(0,_tprintf(_T("PnpFileReadAt(%x)\n"),(ADDRESS)pThis));

    /* Check for EOF
     */
    Uint32ToUint64(Eof,this->StrSize);
    if (!Uint64Less(Position,Eof)) {
        if (Int64Equal(Position,Eof)) {
            if (pSizeRead)
                *pSizeRead = 0;
            return S_FALSE;
        }
        return E_INVALID_PARAMETER;
    }

    /* Make sure we do not overcopy
     */
    Offset = Uint64ToUint32(Position);
    if ((Offset+Size) > this->StrSize)
        Size = this->StrSize - Offset;

    String = &DriverTable[PnpTable[this->TableIndex].DriverIndex][0];

    memcpy(pBuffer,
           String+Offset,
           Size);

    if (pSizeRead)
        *pSizeRead = Size;

    return S_OK;
}

PRIVATE SCODE PnpFileSetSize (
    PIFILE pThis,
    UINT64 Size)
{
    DBGME(0,_tprintf(_T("PnpFileSetSize(%x)\n"),(ADDRESS)pThis));

    UnusedParameter(pThis);
    UnusedParameter(Size);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE PnpFileGetSize (
    PIFILE pThis,
    /*out*/ UINT64 * Size)
{
    PPNPFILE this = pPnpFile(pThis);

    DBGME(0,_tprintf(_T("PnpFileGetSize(%x)\n"),(ADDRESS)pThis));

    Uint32ToUint64(*Size,this->StrSize);
    return S_OK;
}

const struct IFileVtbl PnpFileVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    PnpFileQueryInterface,
    PnpFileAddRef,
    PnpFileRelease,
    PnpFileReadAt,
    PnpFileWriteAt,
    PnpFileSetSize,
    PnpFileGetSize
};

PRIVATE PIUNKNOWN PnpFileNew(
    UINT16 TableIndex)
{
    const _TCHAR *String;
    UINT16 Len;
    PPNPFILE this = malloc(sizeof(PNPFILE));

    DBGME(0,_tprintf(_T("PnpFileNew(%x) -> %x\n"),TableIndex, (ADDRESS)this));

    if (this == NULL)
        goto Out;

    this->v = &PnpFileVtable;
    this->RefCnt = 1;
    this->TableIndex = TableIndex;
    /* Make sure we do not include zeroes, and we do not collapse contiguous strings
     */
    String = &DriverTable[PnpTable[this->TableIndex].DriverIndex][0];
    for (Len = 0; (Len < sizeof(DRIVER_NAME)) && (String[Len] != 0);)
        Len++;
    this->StrSize = Len * sizeof(_TCHAR);/* in bytes! */
 Out:
    return (PIUNKNOWN) this;
}

/* INameSpace::IUnknown
 */
typedef struct _PNPNAME {
    const struct INameSpaceVtbl *v;
    UINT RefCnt;
    PINAMESPACE Delegate;
} PNPNAME, *PPNPNAME;

const _TCHAR ThePnpName[] = _T("pnp");

/* Convert between interface and instance pointer and vice versa */
#define pPnpName(_i_) ((PPNPNAME)((_i_)))
#define iPnpName(_p_) ((PINAMESPACE)((_p_)))

PRIVATE SCODE PnpNameQueryInterface (
    PINAMESPACE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    PPNPNAME this = pPnpName(pThis);

    DBGME(0,_tprintf(_T("PnpNameQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));

    /* INameSpace only
     */
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_INameSpace,iid))
    {
        AtomicInc(&pPnpName(pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }

    /* Otherwise, delegate
     */
    if (this->Delegate)
        return this->Delegate->v->QueryInterface(this->Delegate,
                                              iid,
                                              ppObject);

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT PnpNameAddRef (
    PINAMESPACE pThis)
{
    DBGME(0,_tprintf(_T("PnpNameAR(%x,%d)\n"),(ADDRESS)pThis,pPnpName(pThis)->RefCnt));
    return AtomicInc(&pPnpName(pThis)->RefCnt);
}

PRIVATE UINT PnpNameRelease (
    PINAMESPACE pThis)
{
    PPNPNAME this = pPnpName(pThis);
    UINT Refs = AtomicDec(&this->RefCnt);
    DBGME(0,_tprintf(_T("PnpNameRel(%x,%d)\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
#if DYNAMIC_OBJ
        PIHEAP pHeap = CurrentHeap();
#endif
        PINAMESPACE Ns = CurrentNameSpace();

        /* Finalize object */

        (void) Ns->v->Unregister(Ns,ThePnpName);

        if (this->Delegate)
            this->Delegate->v->Release(this->Delegate);

#if DYNAMIC_OBJ
        pHeap->v->Free(pHeap, 0, (void *) this);
#endif
    }
    return Refs;
}

PRIVATE SCODE PnpNameRegister (
    PINAMESPACE pThis,
    const _TCHAR *pName,
    PIUNKNOWN pObj,
    NAME_SPACE_FLAGS Flags,
    PINAMESPACE pServerNameSpace)
{
    PPNPNAME this = pPnpName(pThis);

    DBGME(0,_tprintf(_T("PnpNameRegister(%x,%s)\n"),(ADDRESS)pThis,pName));

    if (this->Delegate)
        return this->Delegate->v->Register(this->Delegate,
                                           pName,
                                           pObj,
                                           Flags,
                                           pServerNameSpace);
    return E_NOT_IMPLEMENTED;
}

/* Utils
 */
PRIVATE UINT16 a4x(const _TCHAR *A4)
{
    int i, b;
    UINT16 r = ANY_ID;

    if (A4[0] == _T('*'))
        goto Out;

    r = 0;
    for (i = 0; i < 4; i++) {
        _TCHAR c = A4[i];

        if ((c >= _T('0')) && (c <= _T('9')))
            b = c - _T('0');
        else

        if ((c >= _T('a')) && (c <= _T('f')))
            b = c - _T('a') + 10;
        else

        if ((c >= _T('A')) && (c <= _T('F')))
            b = c - _T('A') + 10;
        else

            break;

        r = (r << 4) + b;
    }
 Out:
    DBGME(0,_tprintf(_T("a4x(%s) -> %x\n"),A4,r));
    return r;
}

PRIVATE SCODE PnpNameLookup(
    PPNPNAME this,
    const _TCHAR *Name,
    UINT16 *Index)
{
    UINT16 Vid, Pid, Rev;
    int First, Middle, Last, Cmp;

    DBGME(0,_tprintf(_T("PnpNameLookup(%x,%s)\n"),(ADDRESS)this,Name));

    /* See Pattern below!!
     */
    Vid = a4x(Name+3);
    Pid = a4x(Name+11);
    Rev = a4x(Name+19);

    /* Binary search for Vid
     */
    First = 0;
    Last = nPnpTable;
    while (First < Last) {
        Middle = (First + Last) / 2;

        DBGME(0,_tprintf(_T("Cmp[%d].Vid = %x\n"),Middle,PnpTable[Middle].Vid));
        Cmp = PnpTable[Middle].Vid - Vid;

        if (Cmp == 0)
            goto FoundVid;

        if (Cmp < 0)
            First = Middle + 1;
        else
            Last = Middle;
    }

    return E_NAME_NOT_FOUND;

 FoundVid:
    /* Scan sequentually for Pid
     */
    DBGME(0,_tprintf(_T("Cmp[%d].Pid = %x\n"),Middle,PnpTable[Middle].Pid));
    Cmp = PnpTable[Middle].Pid - Pid;
    if (Cmp == 0)
        goto FoundPid;

    if (Cmp < 0)
        /* Forward */
        while ((++Middle < nPnpTable) && (PnpTable[Middle].Vid == Vid)) {

            DBGME(0,_tprintf(_T("Cmp[%d].Pid = %x\n"),Middle,PnpTable[Middle].Pid));
            Cmp = PnpTable[Middle].Pid - Pid;
            if (Cmp == 0)
                goto FoundPid;

        }
    else
        /* Backward */
        while ((--Middle > 0) && (PnpTable[Middle].Vid == Vid)) {

            DBGME(0,_tprintf(_T("Cmp[%d].Pid = %x\n"),Middle,PnpTable[Middle].Pid));
            Cmp = PnpTable[Middle].Pid - Pid;
            if (Cmp == 0)
                goto FoundPid;

        }
        
    return E_NAME_NOT_FOUND;

 FoundPid:
    /* Scan sequentually for Rev
     */
    DBGME(0,_tprintf(_T("Cmp[%d].Rev = %x\n"),Middle,PnpTable[Middle].Rev));
    Cmp = PnpTable[Middle].Rev - Rev;
    if ((Cmp == 0) || (Rev == ANY_ID) ||
        (PnpTable[Middle].Rev == ANY_ID))
        goto FoundRev;

    if (Cmp < 0)
        /* Forward */
        while ((++Middle < nPnpTable) &&
               (PnpTable[Middle].Vid == Vid) &&
               (PnpTable[Middle].Pid == Pid)) {

            DBGME(0,_tprintf(_T("Cmp[%d].Rev = %x\n"),Middle,PnpTable[Middle].Rev));
            Cmp = PnpTable[Middle].Rev - Rev;
            if ((Cmp == 0) || (PnpTable[Middle].Rev == ANY_ID))
                goto FoundRev;

        }
    else
        /* Backward */
        while ((--Middle > 0) &&
               (PnpTable[Middle].Vid == Vid) &&
               (PnpTable[Middle].Pid == Pid)) {

            DBGME(0,_tprintf(_T("Cmp[%d].Rev = %x\n"),Middle,PnpTable[Middle].Rev));
            Cmp = PnpTable[Middle].Rev - Rev;
            if ((Cmp == 0) || (PnpTable[Middle].Rev == ANY_ID))
                goto FoundRev;

        }
        
    return E_NAME_NOT_FOUND;

 FoundRev:
    *Index = (UINT16) Middle;
    return S_OK;
}


PRIVATE SCODE PnpNameUnregister (
    PINAMESPACE pThis,
    const _TCHAR *pName)
{
    PPNPNAME this = pPnpName(pThis);

    DBGME(0,_tprintf(_T("PnpNameUnregister(%x,%s)\n"),(ADDRESS)pThis,pName));

    if (this->Delegate)
        return this->Delegate->v->Unregister(this->Delegate,
                                             pName);
    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE PnpNameBind (
    PINAMESPACE pThis,
    const _TCHAR *pName,
    NAME_SPACE_FLAGS Flags,
    /*out*/ PIUNKNOWN* ppUnk)
{
    PPNPNAME this = pPnpName(pThis);
    PIUNKNOWN File;
    UINT16 Index;
    SCODE sc;

    DBGME(0,_tprintf(_T("PnpNameBind(%x,%s)\n"),(ADDRESS)pThis,pName));

    /* See if its ours
     */
    sc = PnpNameLookup(this,pName,&Index);
    if (FAILED(sc)) {
        if (this->Delegate)
            return this->Delegate->v->Bind(this->Delegate,
                                           pName,
                                           Flags,
                                           ppUnk);
        return sc;
    }

    File = PnpFileNew(Index);
    *ppUnk = (PIUNKNOWN) File;
    return (File) ? S_OK : E_NOT_ENOUGH_MEMORY;
}

PRIVATE SCODE PnpNameFindNext (
    PINAMESPACE pThis,
    const _TCHAR * pPrefix,
    const _TCHAR * pPrevious,
    /*out*/ _TCHAR * pBuffer,
    UINT BufSize,
    /*out*/ UINT * pStrLen)
{
    PPNPNAME this = pPnpName(pThis);
    UINT16 Index;
    SCODE sc;

    DBGME(0,_tprintf(_T("PnpNameFindNext(%x,%s)\n"),(ADDRESS)pThis,pPrevious));

    /* Find first?
     */
    if ((pPrevious == NULL) || (pPrevious[0] == 0)) {
        Index = 0;
        goto Doit;
    }

    /* See if its ours
     */
    sc = PnpNameLookup(this,pPrevious,&Index);
    if (FAILED(sc)) {
        if (this->Delegate)
            return this->Delegate->v->FindNext(this->Delegate,
                                               pPrefix,
                                               pPrevious,
                                               pBuffer,
                                               BufSize,
                                               pStrLen);
        return sc;
    }

    /* Last one?
     */
    if (Index >= (nPnpTable-1)) {
        if (pStrLen)
            *pStrLen = 0;
        return S_NO_MORE_ENTRIES;
    }
    Index++;

    /* Our pattern takes 3+4+4+4+4+4+1=24 _TCHARs
     * Less if we put a star at the end butsowhat.
     */
#define Pattern _T("Vid%04x_Pid%04x_Dev%04x")

 Doit:
    if (BufSize < sizeof Pattern)
        return S_BUFFER_TOO_SMALL;

    if (pStrLen)
        *pStrLen = sizeof Pattern; /* in bytes! incl EOS */

    _stprintf(pBuffer,
              Pattern,
              PnpTable[Index].Vid,
              PnpTable[Index].Pid,
              PnpTable[Index].Rev);

    /* Is *this* one the last one
     */
    return ((Index == (nPnpTable-1)) ? S_NO_MORE_ENTRIES : S_OK);
}

PRIVATE SCODE PnpNameGetCapabilities (
    PINAMESPACE pThis,
    const _TCHAR * pName,
    /*out*/ NAME_SPACE_FLAGS * pFlags)
{
    DBGME(0,_tprintf(_T("PnpNameGetCapabilities(%x,pName)\n"),(ADDRESS)pThis));

    UnusedParameter(pThis);

    *pFlags = NAME_SPACE_READ;

    return S_OK;
}
	
const struct INameSpaceVtbl PnpNameVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    PnpNameQueryInterface,
    PnpNameAddRef,
    PnpNameRelease,
    PnpNameRegister,
    PnpNameUnregister,
    PnpNameBind,
    PnpNameFindNext,
    PnpNameGetCapabilities
};

/* Constructor */
#if DYNAMIC_OBJ
PINAMESPACE PnpNameNew(void)
{
    PIHEAP pHeap = CurrentHeap();
    PPNPNAME this = (PPNPNAME) pHeap->v->Alloc(pHeap, 0, sizeof *this, 0);

    if (!this)
        return NULL;

    this->v = &PnpNameVtable;
    this->RefCnt = 1;
    this->Delegate = NULL;

    return iPnpName(this);
}

#else
PRIVATE PNPNAME MyPnp = {
    &PnpNameVtable,
    1,
    NULL
};
#endif

PIUNKNOWN PnpCobMain(void)
{
    PPNPNAME pPnp;
    PINAMESPACE Ns = CurrentNameSpace();
    SCODE sc;

    /* Ready to go
     */
#if DYNAMIC_OBJ
    pPnp = PnpNameNew();
#else
    pPnp = &MyPnp;
#endif

#if USE_DELEGATE
    /* Namespace stuff
     * We need to do this to create the hybrid namespace
     */
    sc = BindToObject(Ns,
                      ThePnpName,
                      NAME_SPACE_READ | 
                      NAME_SPACE_WRITE |
                      NAME_SPACE_CREATE,
                      &IID_INameSpace,
                      (void **) &pPnp->Delegate);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("PnpMgr: cant Bind %s, sc=%x\n"),ThePnpName,sc));
        /* ignore?? */
    }

    DBGME(0,_tprintf(_T("PnpMgr:New NameSpace %x, sc=%x\n"),
                     (ADDRESS)pPnp->Delegate,sc));

#endif

    sc = Ns->v->Register(Ns,
                         ThePnpName,
                         (PIUNKNOWN) pPnp,
                         NAME_SPACE_READ | 
                         NAME_SPACE_WRITE |
                         NAME_SPACE_CREATE,
                         NULL);

    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("PnpMgr: cant Register %s, sc=%x\n"),ThePnpName,sc));
        /* ignore?? */
    }

    DBGME(0,_tprintf(_T("pnp manager started, sc=%x\n"),sc));
    return (PIUNKNOWN) pPnp;
}
