/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    SysAce.c
 *
 * Purpose:
 *
 *    Driver for the Xilinx System ACE CompactFlash Solution
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * References:
 *    "System ACE CompactFlash Solution", Advance Product Specification
 *    Document DS080 Version 1.5 April 5, 2002.  Xilinx Corp.
 *    Available at http://www.xilinx.com
 *
 *    "CF+ and CompactFlash Specification", Revision 4.1, 02/16/2007.
 *    CompactFlash Association.
 *    Available at http://www.compactflash.org
 * --------------------------------------------------------------------------
 */

#define DEVICE_ORDER LITTLE_ENDIAN

#include <mmlite.h>
//#define _DEBUG 1
#define DBGLVL 3
#include <diagnostics.h>
#include <assert.h>
#include <drivers/drivers.h>
#include <mips/ml40x.h>

/*
 * Concrete class definition for a SysAce device
 */

typedef struct CSysace {
    struct CDevice CDevice;
    /* Our state
     */
    struct _Sac *Interface;
    BOOL MediaHasChanged;
    UINT32 NumberOfSectors;  /* in bytes */
#define CF_SECBITS     9
#define CF_SECTOR_SIZE (1 << CF_SECBITS)
    UINT64 LastOffset;
#if _DEBUG
    int IntCnt;
#endif
    MUTEX Lock;
} CSysace, *PCSysace;

#define pSt(x) (PCSysace)x

/* Structure returned by the Identify command (see CFlash specs)
 * NB: We only care for the first sector so that is what we define here.
 * NB: Beware of mis-alignment for all 32bit things
 */
typedef struct _CFLASH_IDENTIFY {
    UINT16 Signature;                              /* Word 0 */
#define CFLASH_SIGNATURE 0x848a
    UINT16 DefaultNumberOfCylinders;               /* Word 1 */
    UINT16 Reserved1;                              /* Word 2 */
    UINT16 DefaultNumberOfHeads;                   /* Word 3 */
    UINT16 Obsolete1[2];                           /* Word 4 */
    UINT16 DefaultSectorsPerTrack;                 /* Word 6 */
    UINT16 SectorsPerCard[2];                      /* Word 7 */
    UINT16 Obsolete2;                              /* Word 9 */
    UINT8  SerialNumber[20]; /* padded, right-justified Word 10 */
    UINT16 Obsolete3[2];                           /* Word 20 */
    UINT16 EccBytesInRWLong;                       /* Word 22 */
    UINT8  FirmwareRevision[8];                    /* Word 23 */
    UINT8  ModelNumber[40];                        /* Word 27 */
    UINT16 SectorsInRWMultiple;                    /* Word 47 */
    UINT16 Reserved2;                              /* Word 48 */
    UINT16 Capabilities;                           /* Word 49 */
    UINT16 Reserved3;                              /* Word 50 */
    UINT16 PioMode;                                /* Word 51 */
    UINT16 Obsolete4;                              /* Word 52 */
    UINT16 FieldValidity;                          /* Word 53 */
    UINT16 CurrentNumberOfCylinders;               /* Word 54 */
    UINT16 CurrentNumberOfHeads;                   /* Word 55 */
    UINT16 CurrentSectorsPerTrack;                 /* Word 56 */
    UINT16 CurrentCapacity[2];                     /* Word 57 */
    UINT16 MultiSectorSettings;                    /* Word 59 */
    UINT16 NumberOfAddressableSectors[2];          /* Word 60 */
    UINT16 Reserved4;                              /* Word 62 */
    UINT16 MultiWordDmaTransfer;                   /* Word 63 */
    UINT16 AdvancedPioModes;                       /* Word 64 */
    UINT16 MinimumMultiWordDmaTiming;              /* Word 65 */
    UINT16 RecommendedMultiWordDmaTiming;          /* Word 66 */
    UINT16 PioTimingNoFlowControl;                 /* Word 67 */
    UINT16 PioTimingWithFlowControl;               /* Word 68 */
    UINT16 Reserved5[13];                          /* Word 69 */
    UINT16 FeaturesSupported[3];                   /* Word 82 */
    UINT16 FeaturesEnabled[3];                     /* Word 85 */
    UINT16 UdmaMode;                               /* Word 88 */
    UINT16 SecurityEraseTime;                      /* Word 89 */
    UINT16 EnhancedSecurityEraseTime;              /* Word 90 */
    UINT16 CurrentPowerManagementValue;            /* Word 91 */
    UINT8  Reserved6[72];                          /* Word 92-127 */
    UINT8  SecondHalf[256];                        /* Word 128-255 */
} CFLASH_IDENTIFY, *PCFLASH_IDENTIFY;

#define SIZEOF_IDENTIFY CF_SECTOR_SIZE /* must be a sector multiple */

/* Utilities
 */
/* Sleep for a few milliseconds
 */
PRIVATE void MilliSleep(int mSecs)
{
    TIME t;
    Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(mSecs)));
    SleepUntil(t);
}

#if _DEBUG
PRIVATE void SysaceDumpIdentify(PCFLASH_IDENTIFY Data)
{
    UINT16 Saved;
    _tprintf(_T("Sysace::Identify data @%x:\n"),(ADDRESS)Data);
    _tprintf(_T("\tSignature %x\n"), Data->Signature);
    _tprintf(_T("\tDefaultNumberOfCylinders %x\n"), Data->DefaultNumberOfCylinders);
    _tprintf(_T("\tDefaultNumberOfHeads %x\n"), Data->DefaultNumberOfHeads);
    _tprintf(_T("\tDefaultSectorsPerTrack %x\n"), Data->DefaultSectorsPerTrack);
    _tprintf(_T("\tSectorsPerCard %04x.%04x\n"), 
             Data->SectorsPerCard[0], Data->SectorsPerCard[1]);

    Saved = Data->Obsolete3[0];
    Data->Obsolete3[0] = 0;
    _tprintf(_T("\tSerialNumber '%s'\n"), Data->SerialNumber);
    Data->Obsolete3[0] = Saved;

    _tprintf(_T("\tEccBytesInRWLong %x\n"), Data->EccBytesInRWLong);

    Saved = Data->ModelNumber[0];
    Data->ModelNumber[0] = 0;
    _tprintf(_T("\tFirmwareRevision '%s'\n"), Data->FirmwareRevision);
    Data->ModelNumber[0] = (UINT8)Saved;

    Saved = Data->SectorsInRWMultiple;
    Data->SectorsInRWMultiple = 0;
    _tprintf(_T("\tModelNumber '%s'\n"), Data->ModelNumber);
    Data->SectorsInRWMultiple = Saved;

    _tprintf(_T("\tSectorsInRWMultiple %x\n"), Data->SectorsInRWMultiple);
    _tprintf(_T("\tCapabilities %x\n"), Data->Capabilities);
    _tprintf(_T("\tPioMode %x\n"), Data->PioMode);
    _tprintf(_T("\tFieldValidity %x\n"), Data->FieldValidity);
    _tprintf(_T("\tCurrentNumberOfCylinders %x\n"), Data->CurrentNumberOfCylinders);
    _tprintf(_T("\tCurrentNumberOfHeads %x\n"), Data->CurrentNumberOfHeads);
    _tprintf(_T("\tCurrentSectorsPerTrack %x\n"), Data->CurrentSectorsPerTrack);
    _tprintf(_T("\tCurrentCapacity %04x.%04x\n"),
             Data->CurrentCapacity[0], Data->CurrentCapacity[1]);
    _tprintf(_T("\tMultiSectorSettings %x\n"), Data->MultiSectorSettings);
    _tprintf(_T("\tNumberOfAddressableSectors %04x.%04x\n"),
             Data->NumberOfAddressableSectors[0], Data->NumberOfAddressableSectors[1]);
    _tprintf(_T("\tMultiWordDmaTransfer %x\n"), Data->MultiWordDmaTransfer);
    _tprintf(_T("\tAdvancedPioModes %x\n"), Data->AdvancedPioModes);
    _tprintf(_T("\tMinimumMultiWordDmaTiming %x\n"), Data->MinimumMultiWordDmaTiming);
    _tprintf(_T("\tRecommendedMultiWordDmaTiming %x\n"), Data->RecommendedMultiWordDmaTiming);
    _tprintf(_T("\tPioTimingNoFlowControl %x\n"), Data->PioTimingNoFlowControl);
    _tprintf(_T("\tPioTimingWithFlowControl %x\n"), Data->PioTimingWithFlowControl);
    _tprintf(_T("\tFeaturesSupported %04x.%04x.%04x\n"),
             Data->FeaturesSupported[0], Data->FeaturesSupported[1], Data->FeaturesSupported[2]);
    _tprintf(_T("\tFeaturesEnabled %04x.%04x.%04x\n"),
             Data->FeaturesEnabled[0], Data->FeaturesEnabled[1], Data->FeaturesEnabled[2]);
    _tprintf(_T("\tUdmaMode %x\n"), Data->UdmaMode);
    _tprintf(_T("\tSecurityEraseTime %x\n"), Data->SecurityEraseTime);
    _tprintf(_T("\tEnhancedSecurityEraseTime %x\n"), Data->EnhancedSecurityEraseTime);
    _tprintf(_T("\tCurrentPowerManagementValue %x\n"), Data->CurrentPowerManagementValue);
}

typedef _TCHAR *NAME;
typedef struct _REGDESC {
    NAME RegisterName;
    NAME BitNames[32];
} REGDESC, *PREGDESC;

PRIVATE void SysacePrintRegister(const REGDESC *Desc, UINT32 Value)
{
    int i;
    _tprintf(_T("\t%s %x ="), Desc->RegisterName, Value);
    for (i = 31; i >= 0; i--) {
        if (Value & (1 << i))
            _tprintf(_T(" %s"), 
                     (Desc->BitNames[i]) ? Desc->BitNames[i] : _T("?"));
    }
    _tprintf(_T("\n"));            
}

PRIVATE UINT32 SysaceDumpRegisters(struct _Sac *Interface)
{
	const REGDESC Control_Names =
        { _T("Control"),
          {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
          _T("RST"), //			0x00010000
          _T("BUS8"), //		0x00020000
          _T("BUS16"), //		0x00040000
          _T("BUS32"), //		0x00080000
          _T("IRQ"), //			0x00100000
          _T("BRDY"), //		0x00200000
          _T("IMSK0"), //		0x00400000
          _T("IMSK1"), //		0x00800000
          _T("TD0"), //	    	0x0f000000
          _T("TD1"), //		    0x0f000000
          _T("TD2"), //	    	0x0f000000
          _T("TD3"), //	    	0x0f000000
          _T("BUFW8"), //		0x10000000
          _T("BUFW16"), //		0x20000000
          _T("BUFW32"), //		0x40000000
          _T("DEBUG")} //		0x80000000
        };

	const REGDESC STATUS_Names =
        { _T("STATUS"),
          {_T("CFGLOCK"), //	0x00000001
          _T("MPULOCK"), //		0x00000002
          _T("CFGERROR"), //	0x00000004
          _T("CFCERROR"), //	0x00000008
          _T("CFDETECT"), //	0x00000010
          _T("DATABUFRDY"), //	0x00000020
          _T("DATABUFWRITE"), //0x00000040
          _T("CFGDONE"), //		0x00000080
          _T("RDYFORCFCMD"), //	0x00000100
          _T("CFGMODEPIN"), //	0x00000200
          0,0,0,
          _T("CFGADDRPIN0"), //	0x0000e000
          _T("CFGADDRPIN1"), //	0x0000e000
          _T("CFGADDRPIN2"), //	0x0000e000
          0,
          _T("CFBSY"), //		0x00020000
          _T("CFRDY"), //		0x00040000
          _T("CFDWF"), //		0x00080000
          _T("CFDSC"), //		0x00100000
          _T("CFDRQ"), //		0x00200000
          _T("CFCORR"), //		0x00400000
          _T("CFERR"), //		0x00800000
          0,}
        };

	const REGDESC ERRORREG_Names =
        { _T("ERRORREG"),
          {_T("CARDRESETERR"), //			0x00000001
          _T("CARDRDYERR"), //				0x00000002
          _T("CARDREADERR"), //				0x00000004
          _T("CARDWRITEERR"), //			0x00000008
          _T("SECTORRDYERR"), //			0x00000010
          _T("CFGADDRERR"), //				0x00000020
          _T("CFGFAILED"), //				0x00000040
          _T("CFGREADERR"), //				0x00000080
          _T("CFGINSTRERR"), //				0x00000100
          _T("CFGINITERR"), //				0x00000200
          0,
          _T("CFBBK"), //					0x00000800
          _T("CFUNC"), //					0x00001000
          _T("CFIDNF"), //					0x00002000
          _T("CFABORT"), //					0x00004000
          _T("CFAMNF"), //					0x00008000
           0,}
        };

    const NAME CommandNames[8] =
        { _T("0"), //     				0x0000
          _T("RESETMEMCARD"), //   		0x0100
          _T("IDENTIFYMEMCARD"), //		0x0200
          _T("READMEMCARDDATA"), //		0x0300
          _T("WRITEMEMCARDDATA"), //  	0x0400
          _T("5"), //     				0x0500
          _T("ABORT"), //			    0x0600
          _T("7") //     			    0x0700
        };

	const REGDESC CONTROLREG_Names =
        { _T("CONTROLREG"),
          {_T("FORCELOCKREQ"), //			0x00000001
          _T("LOCKREQ"), //					0x00000002
          _T("FORCECFGADDR"), //			0x00000004
          _T("FORCECFGMODE"), //			0x00000008
          _T("CFGMODE"), //					0x00000010
          _T("CFGSTART"), //				0x00000020
          _T("CFGSEL_MPU"), //  			0x00000040
          _T("CFGRESET"), //				0x00000080
          _T("DATABUFRDYIRQ"), //			0x00000100
          _T("ERRORIRQ"), //				0x00000200
          _T("CFGDONEIRQ"), //				0x00000400
          _T("RESETIRQ"), //				0x00000800
          _T("CFGPROG"), //					0x00001000
          _T("CFGADDR_B0"), //				0x00002000
          _T("CFGADDR_B1"), //				0x00004000
          _T("CFGADDR_B2"), //				0x00008000
          0,}
        };

	const REGDESC FATSTATREG_Names =
        { _T("FATSTATREG"),
          {_T("MBRVALID"), //				0x00000001
          _T("PBRVALID"), //				0x00000002
          _T("MBRFAT12"), //				0x00000004
          _T("PBRFAT12"), //				0x00000008
          _T("MBRFAT16"), //				0x00000010
          _T("PBRFAT16"), //				0x00000020
          _T("CALCFAT12"), //				0x00000040
          _T("CALCFAT16"), //				0x00000080
          0, }
        };

    _tprintf(_T("Sysace@%x:\n"), (ADDRESS)Interface);
    _tprintf(_T("\tTag %x\n"), Interface->Tag);
    SysacePrintRegister(&Control_Names, Interface->Control);
    _tprintf(_T("\tBUSMODEREG %x\n"), Interface->BUSMODEREG);
    SysacePrintRegister(&STATUS_Names, Interface->STATUS);
    SysacePrintRegister(&ERRORREG_Names, Interface->ERRORREG);
    _tprintf(_T("\tCFGLBAREG %x\n"), Interface->CFGLBAREG);
    _tprintf(_T("\tMPULBAREG %x\n"), Interface->MPULBAREG);
    _tprintf(_T("\tVERSIONREG %x\n"), Interface->VERSIONREG);
    _tprintf(_T("\tSECCNTCMDREG %x = %s cnt=%d\n"), Interface->SECCNTCMDREG,
             CommandNames[(Interface->SECCNTCMDREG >> 8)&7],
             Interface->SECCNTCMDREG & SAC_SECCCNT);
    SysacePrintRegister(&CONTROLREG_Names, Interface->CONTROLREG);
    SysacePrintRegister(&FATSTATREG_Names, Interface->FATSTATREG);

    return 1;
}

#else
#define SysaceDumpIdentify(_i_)
#define SysaceDumpRegisters(_c_)
#endif

/* Reset the device and the interface
 */
PRIVATE INT SysaceInitialize(PCDEVICE pThis)
{
    PCSysace pSysace = pSt(pThis);
    struct _Sac *Interface = pSysace->Interface;

    /* 16bit etc etc */
	UINT32 BusMode, Control;

    /* reset our interface */
    Interface->Control = SAC_RST;
    Delay(200);

    /* repeat on both byte lanes */
	Interface->BUSMODEREG = SAC_MODE16 | (SAC_MODE16 << 8);
    Delay(1);

    /* check what our interface does and what the SysACE expects */
	Control = Interface->Control;
	BusMode = Interface->BUSMODEREG;

    /* get them to agree */
	if (BusMode & SAC_MODE16)
	{
		Interface->Control = Control | SAC_BUS16;
		Interface->Control = Interface->Control & ~SAC_BUS8;
	}
	else
	{
		Interface->Control = Control | SAC_BUS8;
		Interface->Control = Interface->Control & ~SAC_BUS16;
	}

    /* check that it worked */
	BusMode = Interface->BUSMODEREG;
	Control = Interface->Control;

	if (((BusMode & SAC_MODE16) == 0) && ((Control & SAC_BUS8) == 0)) return FALSE;
	if (((BusMode & SAC_MODE16) > 0) && ((Control & SAC_BUS16) == 0)) return FALSE;

    /* interrupts off for now */
    Interface->Control &= ~SAC_INTMASK;
#define SAC_INTERRUPTS (SAC_DATABUFRDYIRQ |	SAC_ERRORIRQ )// | SAC_CFGDONEIRQ)
    Control = Interface->CONTROLREG;
    Control = (Control & ~SAC_INTERRUPTS) | SAC_RESETIRQ | SAC_FORCECFGMODE;
    Interface->CONTROLREG = Control;
    Interface->CONTROLREG = Control & ~SAC_RESETIRQ;

    /* no command */
    Interface->MPULBAREG = 0;

    return 0;
}

/* Take control of the ACE datapath
 */
PRIVATE BOOL SysaceLock(PCSysace pSysace)
{
    UINT32 Status;
    int i;

    DBGME(0,_tprintf(_T("Sysace::Lock %x\n"), (ADDRESS)pSysace));

    /* Locked already?
     */
    Status = pSysace->Interface->STATUS;
    if (Status & SAC_MPULOCK)
        return TRUE;

    /* Request lock
     */
    pSysace->Interface->CONTROLREG |= SAC_LOCKREQ;

    /* Spin a bit until we get it
     */
    for (i = 0; i < 200; i++) {
        Delay(100);
        Status = pSysace->Interface->STATUS;
        if (Status & SAC_MPULOCK)
            return TRUE;
        DBGME(0,_tprintf(_T("Sysace::Lock loops.. (st=%x)\n"),Status));
    }

    /* oopsie!
     */
    DBGME(3,_tprintf(_T("Sysace::Lock timeout (st=%x)\n"),Status));
    SysaceDumpRegisters(pSysace->Interface);
    return FALSE;
}

/* Release control of the ACE datapath
 */
PRIVATE BOOL SysaceUnlock(PCSysace pSysace)
{
    UINT32 Status;
    int i;

    DBGME(0,_tprintf(_T("Sysace::Unlock %x\n"), (ADDRESS)pSysace));

    /* Clear reset 
     */
    pSysace->Interface->CONTROLREG &= ~SAC_CFGRESET;

    /* Unlocked already?
     */
    Status = pSysace->Interface->STATUS;
    if (0 == (Status & SAC_MPULOCK))
        return TRUE;

    /* Request unlock
     */
    pSysace->Interface->CONTROLREG &= ~SAC_LOCKREQ;

    /* Spin a bit until we get it
     */
    for (i = 0; i < 200; i++) {
        Delay(100);
        Status = pSysace->Interface->STATUS;
        if (0 == (Status & SAC_MPULOCK))
            return TRUE;
        DBGME(0,_tprintf(_T("Sysace::Unlock loops.. (st=%x)\n"),Status));
    }

    /* oopsie!
     */
    DBGME(3,_tprintf(_T("Sysace::Unlock timeout (st=%x)\n"),Status));
    SysaceDumpRegisters(pSysace->Interface);
    return FALSE;
}

/* Check if the ACE is waiting for a comamnd
 */
#define SysaceReadyForCommand(_s_) ((_s_)->Interface->STATUS & SAC_RDYFORCFCMD)

/* Check if the ACE is executing a comamnd
 */
#define SysaceBusyWithCommand(_s_) ((_s_)->Interface->STATUS & SAC_CFBSY)

/* Turn on interrupts from the ACE
 */
#define SysaceInton(_s_) { \
          (_s_)->Interface->CONTROLREG |= SAC_INTERRUPTS; \
          (_s_)->Interface->Control |= SAC_INTMASK; \
}

/* Turn off interrupts from the ACE
 */
#define SysaceIntoff(_s_) { \
          (_s_)->Interface->CONTROLREG &= ~SAC_INTERRUPTS; \
          (_s_)->Interface->Control &= ~SAC_INTMASK; \
}

/* Start a command on the ACE, such as read or identify.
 */
PRIVATE SCODE SysaceStartCommand(PCSysace pSysace,
                                 UINT32 Command,
                                 UINT32 Lba,
                                 UINT32 nSectors)
{
    SCODE sc = E_DISK_RESET_FAILED;

    /* Lock it if not already
     */
    if (!SysaceLock(pSysace)) {
        /* printed already */
        return sc;
    }

    /* Is there a CF inserted
     */
    if (! (pSysace->Interface->STATUS & SAC_CFDETECT)) {
        /* NB: Not a failure state */
        DBGME(2,_tprintf(_T("Sysace:: no media (st=%x)\n"),pSysace->Interface->STATUS));
        if (pSysace->NumberOfSectors) {
            pSysace->MediaHasChanged = TRUE;
            pSysace->NumberOfSectors = 0;
        }
        return E_NO_MEDIA_IN_DRIVE;
    }

    /* Is it ready for a command
     */
    if (!SysaceReadyForCommand(pSysace)) {
        DBGME(3,_tprintf(_T("Sysace:: not ready (st=%x)\n"),pSysace->Interface->STATUS));
        SysaceDumpRegisters(pSysace->Interface);
        return sc;
    }

    /* sector number and command
     */
    pSysace->Interface->MPULBAREG = Lba;
    pSysace->Interface->SECCNTCMDREG = (UINT16)(Command | (nSectors & SAC_SECCCNT));

    /* re-route the chip
     * NB: The "RESET" is actually not much of a misnomer.
     * The chip was designed for a one-shot execution at reset time,
     * namely loading the configuration data into the FPGA. So.
     */
    pSysace->Interface->CONTROLREG |= SAC_CFGRESET;
    return S_OK;
}

/* Identify the (size of the) CompactFlash card inserted in the slot.
 */
PRIVATE SCODE SysaceIdentifyCard(PCSysace pSysace)
{
    PCFLASH_IDENTIFY Identify;
    UINT32 Status;
    SCODE sc;
    int i, j;

    /* Make sure we are talking sectors.. */
    assert(sizeof(*Identify) == SIZEOF_IDENTIFY);

    /* No need to permanently waste any extra memory */
    Identify = malloc(sizeof *Identify);
    if (Identify == NULL)
        return E_NOT_ENOUGH_MEMORY;

    /* Lock out others
     */
    Mutex_Lock(&pSysace->Lock);

    /* Turn on interrupts before we start the command
     */
    SysaceInton(pSysace); /* BUGBUG should we add polling mode */

    /* This will invalidate the ACE's current sector data 
     */
    Int64FromHighAndLow(pSysace->LastOffset,~0,~0);

    /* Get it going
     */
    sc = SysaceStartCommand(pSysace,
                            SAC_CMD_IDENTIFYMEMCARD,
                            0,
                            1);

    /* Wait until its done
     */
    if (!FAILED(sc)) {

        DBGME(1,_tprintf(_T("Sysace:: cwait.. (st=%x) sizeof=%d\n"),pSysace->Interface->STATUS, sizeof(*Identify)));
        Condition_Wait(&pSysace->CDevice.IntrPoint, NULL, NULL);/*...*/

        /* Did it work?
         */
        Status = pSysace->Interface->STATUS;

        if (Status & SAC_DATABUFRDY) {

            /* Yes, pull out all the data.
             * NB: Until we do so the chip will not be ready for another command
             */
            for (i = 0; i < sizeof(*Identify); i += 4) {

                /* Verify the (32-bytes) FIFO has reloaded 
                 */
                for (j = 0; j < 10; j++) {
                    Status = pSysace->Interface->STATUS;
                    if (Status & SAC_DATABUFRDY)
                        break;
                    Delay(10);
                }
                if (Status & SAC_DATABUFRDY)
                {
                    UINT32 Data32;

                    /* This pulls two 16-bit words out of the FIFO.
                     * They are ordered in LE.
                     */
                    Data32 = pSysace->Interface->DATABUFREG[0];
#if (BYTE_ORDER == BIG_ENDIAN)
                    Data32 = (Data32 >> 16) | (Data32 << 16);
#endif
                    memcpy(((BYTE*)Identify)+i,&Data32,4);
                }
                else
                /* Ooops, what's going on here?
                 */
                {
                    DBGME(3,_tprintf(_T("Sysace::!DATABUFRDY %x\n"),Status));
                    break;
                }
            }

            /* Make sure we did ok and pick up the relevant info
             */
            if ((Status & SAC_DATABUFRDY) &&
                (Identify->Signature == CFLASH_SIGNATURE))
            {
                DBGME(1,_tprintf(_T("Sysace::Card is %s::%s\n"),
                                 Identify->ModelNumber, Identify->SerialNumber));
                SysaceDumpIdentify(Identify);

                pSysace->NumberOfSectors =
                    (Identify->SectorsPerCard[0] << 16) | Identify->SectorsPerCard[1];
                DBGME(1,_tprintf(_T("Sysace::NumberOfSectors %x\n"), pSysace->NumberOfSectors));
            } else {
                DBGME(3,_tprintf(_T("Sysace::Bad card signature? %x != %x\n"),
                                 Identify->Signature, CFLASH_SIGNATURE));
                pSysace->NumberOfSectors = 0;
            }
        }
        else
        /* No, it did not work. Maybe there is no card inserted
         */
        {
            DBGME(3,_tprintf(_T("Sysace::Identify failed, missing CFLASH card?\n")));
            SysaceDumpRegisters(pSysace->Interface);
            /* BUGBUG Fix the error code accordingly */
            sc = E_TIMED_OUT;
        }
    }

    /* Free the ACE for the JTAG, just in case */
    SysaceUnlock(pSysace);

    /* Let others in
     */
    Mutex_Unlock(&pSysace->Lock);

    /* Release memory
     */
    free(Identify);

    /* Done
     */
    return sc;
}

/* Interrupt service routine
 */
PRIVATE BOOL SysaceDeviceIsr (
    PCDEVICE pThis,
    /*out*/ BOOL * pNotOurInterrupt)
{
    PCSysace pSysace = pSt(pThis);
    UINT32 Control;

#if _DEBUG
    pSysace->IntCnt++;
#endif

    /* Turn off interrupts and ACK them
     */
    SysaceIntoff(pSysace);

    Control = pSysace->Interface->CONTROLREG & (~(SAC_RESETIRQ|SAC_INTERRUPTS));
    pSysace->Interface->CONTROLREG = Control | SAC_RESETIRQ;
    pSysace->Interface->CONTROLREG = Control;

    /* ... read status and do whatever ... */

    return Condition_InterruptSignal(&pSysace->CDevice.IntrPoint);
}


/* Common code for read&write argument validation
 */
PRIVATE SCODE SysaceValidateArguments(PCSysace pSysace,
                                     UINT64 Position,
                                     UINT *pOffset,
                                     UINT *pSize)
{
    UINT32 StartSector, Offset, Size;

    /* Compute which sector and the offset within
     */
    Offset = Uint64ToUint32(Position) & (CF_SECTOR_SIZE-1);
    StartSector = Uint64ToUint32( Uint64RShift(Position,CF_SECBITS) );

    /* We can only really read/write full sectors.
     * To help debugging, make it that we operate in a buffer worth,
     * meaning 32bytes chunks on 32bytes boundaries
     */
    if (((Offset & 31) != 0) ||
        ((*pSize & 31) != 0))
        return E_INVALID_PARAMETER;

    /* Verify that we know the real size
     */
    if (pSysace->NumberOfSectors == 0) {
        SCODE sc = SysaceIdentifyCard(pSysace);
        if (FAILED(sc))
            return sc;
    }

    /* Validate args
     */
    if (StartSector >= pSysace->NumberOfSectors) {
        *pSize = 0;
        DBGME(1,_tprintf(_T("Sysace::ValidateArg(%d) EOF\n"),Offset));
        return S_FALSE; /* EOF */
    }

    /* Return the starting sector
     */
    *pOffset = StartSector;

    /* Adjust size if necessary
     */
    Position = Uint64AddUint32(Position,*pSize+(CF_SECTOR_SIZE-1));
    Size = Uint64ToUint32( Uint64RShift(Position,CF_SECBITS) );
    if (Size > pSysace->NumberOfSectors) {
        /* At most this many sectors 
         */
        Size = pSysace->NumberOfSectors - StartSector;
        /* If it overflows 32bits we trim it, and that's ok
         */
        *pSize = Size << CF_SECBITS;
    }

    DBGME(0,_tprintf(_T("Sysace::FlashValidateArg %d %d\n"),
                     *pOffset,*pSize));
    return NO_ERROR;
}

/* Read SIZE bytes from sysace device, at offset Position
 */
PRIVATE SCODE MCT SysaceReadAt (PCDEVICE this, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    PCSysace pSysace = pSt(this);
    SCODE sc;
    UINT32 StartSector, SizeThisTime;
    UINT32 Status = 0, SizeRead = 0;
    UINT i, j;

    DBGME(1,_tprintf(_T("SysaceReadAt(%x %lx %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeRead));

#define TZK_HACK 1
#if TZK_HACK
    if (Size == 127) Size = 64; /* tzk hack */
#endif

    /* Validate & trim arguments
     */
    sc = SysaceValidateArguments(pSysace,
                                Position,
                                &StartSector,
                                &Size);

    /* Repeat until we are done or error
     */
    while (sc == NO_ERROR) {

        /* .. one sector at a time
         * BUGBUG Supposedly we can do up to 256 sectors?
         */
        SizeThisTime = Size;
        if (SizeThisTime > CF_SECTOR_SIZE)
            SizeThisTime = CF_SECTOR_SIZE;

        /* Give someone else a chance
         * BUGBUG Not doing this would be faster
         */
        Mutex_Lock(&pSysace->Lock);

        /* Are we getting outside of the ACE's cached sector?
         */
        if ((!Int64Equal(Position,pSysace->LastOffset)) || /* non-sequential? */
            /* or at a sector boundary? */
            (0 == (Uint64ToUint32(Position) & (CF_SECTOR_SIZE-1))) ||
            /* Or the cached data was not for a READ? */
            ((pSysace->Interface->SECCNTCMDREG & SAC_CMD) != SAC_CMD_READMEMCARDDATA))
        {
            /* Yes, start a sector read
             */
            SysaceInton(pSysace);
            sc = SysaceStartCommand(pSysace,
                                    SAC_CMD_READMEMCARDDATA,
                                    StartSector,
                                    1);
            /* And wait until done, if ok
             */
            if (!FAILED(sc)) {
                pSysace->LastOffset = Position;
                StartSector += 1;
                /* BUGBUG timeouts! */
                sc = Condition_Wait(&pSysace->CDevice.IntrPoint, NULL, NULL);/*...*/
            }
        }

        /* Are we doing ok
         */
        if (!FAILED(sc)) {

            /* Get the data out of the ACE
             */
            for (i = 0; i < SizeThisTime; i += 4) {

                /* Make sure the FIFO is ready
                 */
                for (j = 0; j < 10; j++) {
                    Status = pSysace->Interface->STATUS;
                    if (Status & SAC_DATABUFRDY)
                        break;
                    Delay(1000);
                }

                /* Got it?
                 */
                if (Status & SAC_DATABUFRDY) {
                    UINT32 Data32;
                    
                    Data32 = pSysace->Interface->DATABUFREG[0];
                    Data32 = _bs32(Data32);
                    memcpy(pBuffer+i,&Data32,4);
                }
                else
                {
                    /* Ooops, get out of here
                     */
                    DBGME(3,_tprintf(_T("Sysace::READ timeout\n")));
                    SysaceDumpRegisters(pSysace->Interface);
                    sc = E_TIMED_OUT;
                    break;
                }
            }

            /* Still doing ok?
             */
            if (!FAILED(sc)) {
                Position            = Uint64AddUint32(Position,SizeThisTime);
                pSysace->LastOffset = Position;
                pBuffer            += SizeThisTime;
                SizeRead           += SizeThisTime;
                Size               -= SizeThisTime;
            }
        }

        /* Free the ACE for the JTAG, just in case */
        SysaceUnlock(pSysace);

        /* Let others in
         */
        Mutex_Unlock(&pSysace->Lock);

        /* Are we done yet?
         */
        if (Size == 0)
            break;
    }

    if (pSizeRead)
        *pSizeRead = SizeRead;
    return sc;
}

/* Write SIZE bytes to device.
 */
PRIVATE SCODE MCT SysaceWriteAt (PCDEVICE this, UINT64 Position, const BYTE *pBuffer,
                                 UINT Size, PUINT pSizeWritten)
{
    PCSysace pSysace = pSt(this);
    SCODE sc;
    UINT32 StartSector, SizeThisTime;
    UINT32 Status = 0, SizeWritten = 0;
    UINT i, j;

    DBGME(1,_tprintf(_T("SysaceWriteAt(%x %lx %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeWritten));

#if TZK_HACK
    if (Size == 127) Size = 64; /* tzk hack */
#endif


    /* Validate & trim arguments
     */
    sc = SysaceValidateArguments(pSysace,
                                Position,
                                &StartSector,
                                &Size);

    /* Repeat until we are done or error
     */
    while (sc == NO_ERROR) {

        /* .. one sector at a time
         * BUGBUG Supposedly we can do up to 256 sectors?
         */
        SizeThisTime = Size;
        if (SizeThisTime > CF_SECTOR_SIZE)
            SizeThisTime = CF_SECTOR_SIZE;

        /* Give someone else a chance
         * BUGBUG Not doing this would be faster
         */
        Mutex_Lock(&pSysace->Lock);

        /* Are we getting outside of the ACE's cached sector?
         */
        if ((!Int64Equal(Position,pSysace->LastOffset)) || /* non-sequential? */
            /* or at a sector boundary? */
            (0 == (Uint64ToUint32(Position) & (CF_SECTOR_SIZE-1))) ||
            /* Or the cached data was not for a WRITE? */
            ((pSysace->Interface->SECCNTCMDREG & SAC_CMD) != SAC_CMD_WRITEMEMCARDDATA))
        {
            /* Yes, start a sector write
             */
            SysaceInton(pSysace);
            sc = SysaceStartCommand(pSysace,
                                    SAC_CMD_WRITEMEMCARDDATA,
                                    StartSector,
                                    1);
            /* And wait until done, if ok
             */
            if (!FAILED(sc)) {
                pSysace->LastOffset = Position;
                StartSector += 1;
                /* BUGBUG timeouts! */
                sc = Condition_Wait(&pSysace->CDevice.IntrPoint, NULL, NULL);/*...*/
            }
        }

        /* Are we doing ok
         */
        if (!FAILED(sc)) {

            /* Get the data out of the ACE
             */
            for (i = 0; i < SizeThisTime; i += 4) {

                /* Make sure the FIFO is ready
                 */
                for (j = 0; j < 10; j++) {
                    Status = pSysace->Interface->STATUS;
                    if (Status & SAC_DATABUFRDY)
                        break;
                    Delay(1000);
                }

                /* Got it?
                 */
                if (Status & SAC_DATABUFRDY) {
                    UINT32 Data32;

                    memcpy(&Data32,pBuffer+i,4);
                    Data32 = _bs32(Data32);
                    pSysace->Interface->DATABUFREG[0] = Data32;
                }
                else
                {
                    /* Ooops, get out of here
                     */
                    DBGME(3,_tprintf(_T("Sysace::WRITE timeout\n")));
                    SysaceDumpRegisters(pSysace->Interface);
                    sc = E_TIMED_OUT;
                    break;
                }
            }

            /* Still doing ok?
             */
            if (!FAILED(sc)) {
                Position            = Uint64AddUint32(Position,SizeThisTime);
                pSysace->LastOffset = Position;
                pBuffer            += SizeThisTime;
                SizeWritten        += SizeThisTime;
                Size               -= SizeThisTime;
            }
        }

        /* We need to wait until the device is ready for the next command
         * Experimentation shows that it can take longer than 10msec.
         */
        if (!FAILED(sc)) {
            for (j = 0; j < 300; j++) {
                Status = pSysace->Interface->STATUS;
                if (Status & SAC_RDYFORCFCMD)
                    break;
                MilliSleep(2);
            }
            if (! (Status & SAC_RDYFORCFCMD)) {
                DBGME(3,_tprintf(_T("Sysace::WRITE-COMPLETE timeout St=%x\n"), Status));
                SysaceDumpRegisters(pSysace->Interface);
                /* Ignore,  we'll handle it the next time around
                 * BUGBUG To be revised along with non-existant error handling
                 */
            }
        }

        /* Free the ACE for the JTAG, just in case */
        SysaceUnlock(pSysace);

        /* Let others in
         */
        Mutex_Unlock(&pSysace->Lock);

        /* Are we done yet?
         */
        if (Size == 0)
            break;
    }

    if (pSizeWritten)
        *pSizeWritten = SizeWritten;
    return sc;
}

/* GetSize method on sysace device
 * BUGBUG Test media was not removed
 */
PRIVATE SCODE MCT SysaceGetSize (PCDEVICE this, PUINT64 pSize)
{
    PCSysace pSysace = pSt(this);

#if _DEBUG
    DBGME(1,_tprintf(_T("Sysace.IntCnt @%x = %d\n"),(ADDRESS)&pSysace->IntCnt,pSysace->IntCnt));
#endif

#if _DEBUG
    SysaceDumpRegisters(pSysace->Interface);
#endif

    /* Verify that we know the real size
     */
    if (pSysace->MediaHasChanged ||
        (pSysace->NumberOfSectors &&
         (! (pSysace->Interface->STATUS & SAC_CFDETECT)) )) {
        /* NB: Not a failure state */
        DBGME(2,_tprintf(_T("Sysace:: no media (st=%x)\n"),pSysace->Interface->STATUS));
        pSysace->NumberOfSectors = 0;
        pSysace->MediaHasChanged = FALSE;
        return E_NO_MEDIA_IN_DRIVE;
    }

    if (pSysace->NumberOfSectors == 0) {
        /* We dont, do an IDENTIFY on the card
         */
        SCODE sc = SysaceIdentifyCard(pSysace);
        if (FAILED(sc))
            return sc;
        SysaceDumpRegisters(pSysace->Interface);
    }

    /* Return result
     */
    if (pSize) {
        Uint32ToUint64(*pSize,pSysace->NumberOfSectors);
        *pSize = Uint64LShift(*pSize,9);
    }
    return NO_ERROR;
}

/* SetSize method on sysace device
 * We dont truncate, but we do tell if the media has changed.
 */
PRIVATE SCODE MCT SysaceSetSize( PCDEVICE this, UINT64 Size)
{
    PCSysace pSysace = pSt(this);
    UINT64 MediaChangedSignal;

    DBGME(0,_tprintf(_T("SysaceSetSize(%x %lx)\n"),
                     (ADDRESS) this, Size));

    Int64FromHighAndLow(MediaChangedSignal,~0,~0);
    if (Int64Equal(Size,MediaChangedSignal)) {

        /* Only warn once
         */
        if (pSysace->MediaHasChanged ||
            (pSysace->NumberOfSectors &&
             (! (pSysace->Interface->STATUS & SAC_CFDETECT)) )) {
            /* NB: Not a failure state */
            DBGME(2,_tprintf(_T("Sysace:: no media (st=%x)\n"),pSysace->Interface->STATUS));
            pSysace->NumberOfSectors = 0;
            pSysace->MediaHasChanged = FALSE;
            return E_NO_MEDIA_IN_DRIVE;
        }
    }

    return E_INVALID_PARAMETER;
}



struct CDeviceVtbl SysaceDriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    SysaceReadAt,
    SysaceWriteAt,
    SysaceSetSize,
    SysaceGetSize,
    SysaceDeviceIsr,
    NULL,
    SysaceInitialize,
    NULL
};

/* Constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE SysaceNew(ADDRESS Base)
{
    PCSysace pSysace;
    struct _Sac * Sac = (struct _Sac *)Base;
    PIHEAP pH = CurrentHeap();

    /* Check the tag to see if its there
     */
    if ((Sac->Tag & SAC_TAG) != PMTTAG_SYSTEM_ACE) {
        DBGME(3,_tprintf(_T("SysaceNew: bad tag (%x != %x) @x%x\n"),
                         Sac->Tag, PMTTAG_SYSTEM_ACE, Base));
        return NULL;
    }

    /* Allocate a descriptor
     */
    pSysace = (PCSysace) pH->v->Alloc( pH, 0, 
                                       sizeof(CSysace), 0 );

    /* Initialize all fields
     */
    if (pSysace) {
        pSysace->CDevice.v = &SysaceDriverVtable;
        pSysace->CDevice.RefCnt = 1;
        Mutex_Init(&pSysace->Lock);
        pSysace->Interface = Sac;
        pSysace->MediaHasChanged = FALSE;
        pSysace->NumberOfSectors = 0; /* IDENTIFY later */
        Int64FromHighAndLow(pSysace->LastOffset,~0,~0);
#if _DEBUG
        pSysace->IntCnt = 0;
#endif
        /* Reset the interface
         */
        if (pSysace->CDevice.v->Initialize((PCDEVICE)pSysace) < 0) {
            DBGME(3,_tprintf(_T("SysaceNew: reset fails %x ? %x\n"), 
                             Sac->BUSMODEREG, Sac->Control));
            pH->v->Free( pH, 0, pSysace );
            return NULL;
        }

        DBGME(1,_tprintf(_T("SysaceNew(%x)\n"),(ADDRESS)pSysace));
    }

    return (PCDEVICE)pSysace;
}

#if 1 /* I am in a hurry */
struct _Sac *DefaultInterface = NULL;

BOOL SysaceSendConfig(struct _Sac *Interface, UINT32 *Data, UINT nBytes)
{
    UINT i, j, nWords;
    UINT32 CtlWas;
    UINT32 Status;

    if (Interface == NULL)
        Interface = DefaultInterface;
    if (Interface == NULL)
        return FALSE;

    CtlWas = Interface->CONTROLREG;

    /* Set the bits but in RESET (pag 49-50 of specs)*/
#define CFGCMD (SAC_FORCELOCKREQ | SAC_LOCKREQ | SAC_CFGSEL | \
                SAC_FORCECFGMODE |/* SAC_CFGMODE |*/ SAC_CFGSTART)

    Interface->CONTROLREG = CFGCMD | SAC_CFGRESET;

    /* Take it out of RESET */
    Interface->CONTROLREG = CFGCMD;

    /* Must wait till it says READY
     * It can take a looong time
     */
    for (j = 0; j < 1000; j++) {
        Status = Interface->STATUS;
        if (Status & SAC_RDYFORCFCMD)
            break;
        Delay(1000);
    }

    if (0 == (Status & SAC_RDYFORCFCMD)) {
        DBGME(3,_tprintf(_T("Sysace::CMD error %x (j=%d)\n"), Status, j));
        goto Error;
    }

    /* Get the data out to the ACE
     */
#define ACEROUNDUP 32
    nBytes = (nBytes + ACEROUNDUP - 1) & ~(ACEROUNDUP-1);
    nWords = (nBytes + 3) / 4;
    DBGME(1,_tprintf(_T("Sending %d bytes (as %d words) to %x "), 
                     nBytes, nWords, (ADDRESS)Interface));
    for (i = 0; i < nWords; i += 1/*word*/) {

        /* Stop on errors */
        Status = Interface->ERRORREG;
        if (Status) {
            /* Ooops, get out of here
             */
            DBGME(3,_tprintf(_T("Sysace::CFG error %x (i=%d)\n"), Status, i));
            goto Error;
        }

        /* Make sure the FIFO is ready
         */
        for (j = 0; j < 100; j++) {
            Status = Interface->STATUS;
            if (Status & SAC_DATABUFRDY)
                break;
            Delay(1000);
        }

        /* Got it?
         */
        if (Status & SAC_DATABUFRDY) {
            UINT32 Data32;

            Data32 = Data[i];
            Data32 = _bs32(Data32);
            Interface->DATABUFREG[0] = Data32;
        }
        else
        {
            /* Ooops, get out of here
             */
            DBGME(3,_tprintf(_T("Sysace::WRITE timeout %x (i=%d)\n"), Status, i));
            goto Error;
        }
    }
    DBGME(1,_tprintf(_T("done ok.\n")));

    /* Put it back the way it was (try to.. :-( )*/
    Interface->CONTROLREG = CtlWas;
    return TRUE;

 Error:
    SysaceDumpRegisters(Interface);
    Interface->CONTROLREG = CtlWas;
    return (FALSE);
}

#endif

/* ** 
*/
SCODE __cdecl SysaceConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Name);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("SysaceConstructor(%x %s %x %x %x)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk));

#if 1 /* I am in a hurry */
    if (DefaultInterface == NULL) DefaultInterface = (struct _Sac *)RegisterBase;
#endif
    pUnk = (PIUNKNOWN) SysaceNew(RegisterBase);

    DBGME(0,_tprintf(_T("SysaceConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("SysaceConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl SysaceDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) SysaceConstructor
};

static struct _DRIVERCOB SysaceDriverCobObject = { &SysaceDriverCobVtable, 1 };

PIUNKNOWN SysaceCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &SysaceDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &SysaceDriverVtable, _T("CDevice"),
                          &IID_IDevice);
    if (FAILED(sc))
        return NULL;

    /* Void the thread loop, to save a useless thread
     */
    SysaceDriverVtable.ThreadLoop = NULL;

    return (PIUNKNOWN) &SysaceDriverCobObject;
}

