/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    at24c512.c
 *
 * Purpose:
 *
 *    Driver for the Atmel AT24C512 Serial EEPROM device.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <diagnostics.h>
#include <assert.h>
#include <drivers/drivers.h>
#include <arm\at91m63200.h>

#define ALLOW_WRITE_OPERATIONS 1

/*
 * Concrete class definition for a c512 device
 */

typedef struct CC512 {
    struct CDevice CDevice;
    /* Our state
     */
    MUTEX Lock;
    UINT32 Size;  /* in bytes */
    UINT32 CurPos;/* ditto */
} CC512, *PCC512;

#define pCc(x) (PCC512)x

/* The board connects the uC to this I2C device as follows:
 * - It uses an Atmel AT24C512 8-pin PDIP chip (64KB)
 * - ROM-1,2,3,7 are not-connected.  That's A0,A1,NC,WP.
 * - ROM-4 is GND and ROM-8 is VCC.
 * - ROM-5 (SDA) connects to CPU-93 (PA21/TXD2) 
 * - ROM-6 (SCL) connects to CPU-92 (PA20/SCK2)
 *
 * This means that A0=A1=0, and yes its writeable.
 */

/* Other defines
 */
#define AT24C_ADDRESS 0xa0
#define EE_WRITE 0
#define EE_READ 1
#define EE_PAGESIZE 128 /* bytes */
#define INVALID_OFFSET (~0)

/* PIO Interface
 */
#define ThePio PioA
#define SCL    PIOA_SCLK2
#define SDA    PIOA_TXD2
#define _WireMakeInput(_p_)   ThePio->OutDisable = (_p_)
#define _WireMakeOutput(_p_)  ThePio->OutEnable = (_p_)
#define _WireLow(_p_)         ThePio->ClearData = (_p_)
#define _WireHigh(_p_)        ThePio->SetData = (_p_)
#define _WireIsHigh(_p_)      ((ThePio->PinDataStatus & (_p_)) != 0)
#define WireIsLow(_p_)       (!WireIsHigh(_p_)) 
#define _WireEnable(_p_)      ThePio->Enable = (_p_)

#ifdef _DEBUG
PRIVATE void sayso(_TCHAR *m, UINT Bits)
{
#if (DBGLVL==0)
    _tprintf(m);
    if (Bits & SCL) _tprintf(_T(" SCL"));
    if (Bits & SDA) _tprintf(_T(" SDA"));
    _tprintf(_T("\n"));
#endif
}
#define sayit(f,v) _tprintf(f,v)
#else
#define sayso(a,b)
#define sayit(f,v)
#endif

PRIVATE void WireMakeInput(UINT Bits)
{
    sayso(_T("WireMakeInput"), Bits);
    _WireMakeInput(Bits);
}

PRIVATE void WireMakeOutput(UINT Bits)
{
    sayso(_T("WireMakeOutput"), Bits);
    _WireMakeOutput(Bits);
}

PRIVATE void WireLow(UINT Bits)
{
    sayso(_T("WireLow"), Bits);
    _WireLow(Bits);
}

PRIVATE void WireHigh(UINT Bits)
{
    sayso(_T("WireHigh"), Bits);
    _WireHigh(Bits);
}

PRIVATE void WireEnable(UINT Bits)
{
    sayso(_T("WireEnable"), Bits);
    _WireEnable(Bits);
}

PRIVATE BOOL WireIsHigh(UINT Bit)
{
    sayit(_T("%x "), ThePio->PinDataStatus & Bit);
    sayso(_T("="), Bit);
    return ((ThePio->PinDataStatus & (Bit)) != 0);
}

/* Initialization/Finalization
 */
PRIVATE void InitWires(void)
{
    DBGME(0,_tprintf(_T("C512::InitWires\n")));

    /* Before we enable them pins, set safe values.
     */
    WireHigh(SCL|SDA);

    /* Set directions
     */
    WireMakeOutput(SDA|SCL);
 
    /* Make sure the Pio owns them
     */
    WireEnable(SCL|SDA);

    /* Done
     */
}

/* Save power by not driving them pins
 */
#define FinitWires()  WireMakeInput(SCL|SDA)

/* Indicate the START of an I2C bus transaction
 */
PRIVATE void EepromStart(void)
{
    DBGME(0,_tprintf(_T("C512::EepromStart\n")));

    /* assert( WireIsHigh(SCL) ) */

    /* NB: Since there must be an external pullup for SDA
     * this one is not really needed. But the EB63 doesnt have it.
     */
    WireHigh(SDA);

    /* Pull clock high (valid) and wait for raise time (T-r=1us)
     */
    WireHigh(SCL);
    Delay(1);

    /* BUGBUG Check someone else is not driving the bus
     */

    /* This transition properly defines the START condition
     */
    WireLow(SDA);

    /* T-f=300ns is a given, but T-low=4.7us is a must.
     * Notice we did 1us already up above.
     */
    Delay(4);

    /* Exit condition: SCL=SDA=LOW
     */
    WireLow(SCL);
    Delay(1);
}

PRIVATE void EepromStop(void)
{
    DBGME(0,_tprintf(_T("C512::EepromStop\n")));

    /* assert( WireIsLow(SCL) ) */

    /* SDA must be low initially, then data setup time
     */
    WireLow(SDA);
    Delay(1);

    /* Pull clock high (valid) and wait for raise time (T-r=1us)
     */
    WireHigh(SCL);
    Delay(1);

    /* This transition properly defines the STOP condition
     */
    WireHigh(SDA);

    /* T-f=300ns is a given, but T-low=4.7us is a must.
     * Notice we did 1us already up above.
     */
    Delay(4);

    /* Exit condition SCL=SDA=HIGH
     */
}

/* Needed after an interruption in protocol, power loss, or system reset.
 */
PRIVATE void EEpromReset(void)
{
    int i;
    BOOL SawAnACK = FALSE;

    DBGME(2,_tprintf(_T("C512::EEpromReset\n")));

    /* I am confused by the "MEMORY RESET" discussion in the specs.
     * It does not say "drive SDA low and then (a) Clock up to...".
     * So it makes no sense, SDA is supposed to be normally pulled HIGH.
     * I think they mean to say that if the CPU was reset in the middle
     * of a transaction with the AT24 we have to complete it.
     * That transaction could be a READ or a WRITE.
     * We would check for a WRITE by clocking and watching for an ACK.
     * If the ACK never happens, great we are done, device wasnt selected.
     * If we do see the ACK we complete with a STOP condition.
     * [I am not clear why they say to create a START instead]
     * As for a READ transaction, we are the ones that are supposed to NACK.
     * And then generate a STOP condition.
     */
    for (i = 0; (i < 9) && (!SawAnACK); i++)
    {
        /* Enable clock, keep it high for half of T-high
         */
        WireHigh(SCL);
        Delay(2);

        /* Sample SDA.  If LOW this is an ACK (*we* are driving it HIGH)
         */
        SawAnACK = WireIsLow(SDA);

        /* Finish with T-high
         */
        Delay(2);

        /* Flip and keep low
         */
        WireLow(SCL);
        Delay(4);
    }

    /* Either way issue the STOP
     */
    EepromStop();

    DBGME(1,_tprintf(_T("C512::EEpromReset %d %d\n"),i,SawAnACK));

    /* Done.
     * Exit condition SCL=SDA=HIGH
     */
}

/* Write a given number of bytes to the bus
 */
PRIVATE BOOL EepromOutput(const UINT8 *Bits,
                          UINT nBytes,
                          BOOL Start)
{
    UINT i, j;
    UINT8 c = 0;
    BOOL ok = TRUE;

    DBGME(0,_tprintf(_T("C512::EepromOutput(%x %d)\n"),(ADDRESS)Bits,nBytes));

    /* START condition?
     */
    if (Start)
        EepromStart();

    DBGME(0,_tprintf(_T("C512::EepromOutput now\n")));

    /* Transfer one byte, get ACK, repeat
     */
    for (i = 0; ok && (i < nBytes); i++) {

        /* Fetch next byte 
         */
        c = *Bits++;

        for (j = 0; j < 8; j++)
        {
            /* Set the value before we toggle the clock
             * NB: Must be done MSB first, remember
             */
            if (c & 0x80) {
                WireHigh(SDA);
            } else {
                WireLow(SDA);
            }
            Delay(1);

            /* Clock this bit for T-high
             */
            WireHigh(SCL);
            Delay(4);

            /* Toggle for T-low (..)
             */
            WireLow(SCL);
            Delay(4);

            /* Next bit
             */
            c <<= 1;
        }

        /* Make sure we get the ACK
         */
        WireHigh(SDA);
        Delay(1);
        WireHigh(SCL);

        /* According to specs, it should be low already (no delay then, but..)
         */
        Delay(2);
        ok = WireIsLow(SDA);

        /* Toggle
         */
        WireLow(SCL);
        Delay(4);
    }    

    if (!ok)
    {
        DBGME(3,_tprintf(_T("C512::EepromOutput NAK %d %d\n"),i,j));
        return FALSE;
    }

    /* NB: We dont do the STOP in here
     * Exit condition SCL=LOW  SDA=HIGH
     */
    return TRUE;
}

/* Read a given number of bits from the device
 */
PRIVATE void EepromInput(UINT8 *Bits,
                         UINT nBytes)
{
    UINT i, j;
    UINT8 c = 0;

    DBGME(0,_tprintf(_T("C512::EepromInput(%x %d)\n"),(ADDRESS)Bits,nBytes));

    /* Entry condition: SCL=LOW, SDA=HIGH
     */

    /* Transfer one byte, send ACK, repeat
     */
    for (i = nBytes; i-- > 0;) {

        /* Next byte
         */
        c = 0;

        for (j = 0; j < 8; j++)
        {
            /* Next bit
             */
            c <<= 1;

            /* Toggle for half T-high, then sample
             */
            WireHigh(SCL);
            Delay(2);

            /* Get the value before we toggle the clock
             * NB: Must be done MSB first, remember
             */
            c |= (WireIsHigh(SDA)) ? 1 : 0;

            /* Before we can toggle SCL again we need to finish T-high
             */
            Delay(2);

            /* Toggle low, hold for T-low
             */
            WireLow(SCL);
            Delay(4);
        }

        /* Store next byte
         */
        *Bits++ = c;

        /* send ACK, unless its the last byte
         */
        if (i)
        {
            /* ACK */
            WireLow(SDA);
        }
        else
        {
            /* NACK */
            WireHigh(SDA);
        }

        /* Regardless, it needs the 9-th cycle
         */
        WireHigh(SCL);
        Delay(4);
        WireLow(SCL);
        Delay(2/*+2*/);
        WireHigh(SDA);
    }    

    /* Exit condition: SDA=SCL=HIGH
     */
}


/* Read data from memory
 */
PRIVATE BOOL EepromRead(PCC512 pC512,
                        UINT Offset,
                        UINT8 *Data,
                        UINT nBytes)
{
    UINT8 CommandBits[3];

    DBGME(1,_tprintf(_T("C512::EepromRead(%d,%x,%d)\n"),
                     Offset,(ADDRESS)Data,nBytes));

    /* Can we skip the addressing part?
     */
    if (Offset != pC512->CurPos)
    {
        /* Nope
         */
        CommandBits[0] = AT24C_ADDRESS | EE_WRITE; /* dummy write cycle */
        CommandBits[1] = (UINT8)(Offset >> 8);
        CommandBits[2] = (UINT8)Offset;

        /* Send it out
         */
        if (!EepromOutput(CommandBits, 3, TRUE))
        {
            DBGME(3,_tprintf(_T("C512::EepromRead no out0\n")));
            return FALSE;
        }
    }

    /* Either way, proceed with a Sequential Read
     */
    CommandBits[0] = AT24C_ADDRESS | EE_READ;
    if (!EepromOutput(CommandBits, 1, TRUE))
    {
        DBGME(3,_tprintf(_T("C512::EepromRead no out1\n")));
        return FALSE;
    }

    EepromInput(Data,nBytes);
    EepromStop();

    pC512->CurPos = Offset + nBytes;

    return TRUE;
}

#if ALLOW_WRITE_OPERATIONS

/* See if a write has completed
 */
#define EepromReady(_pc512_,_offset_,_pbyte_) \
     EepromRead(_pc512_,_offset_,_pbyte_,1)


/* Write data the EEPROM
 */
PRIVATE SCODE EepromWrite(PCC512 pC512,
                          UINT Offset,
                          const UINT8 *Data,
                          UINT nBytes)
{
    int i;
    TIME t;
    UINT8 CommandBits[3];

    DBGME(0,_tprintf(_T("C512::EepromBufferWrite(%d,%x,%d)\n"),
                     Offset,(ADDRESS)Data,nBytes));

    /* Not sure where the address will be left at
     */
    pC512->CurPos = INVALID_OFFSET;

    /* Select starting address
     */
    CommandBits[0] = AT24C_ADDRESS | EE_WRITE;
    CommandBits[1] = (UINT8)(Offset >> 8);
    CommandBits[2] = (UINT8)Offset;

    /* Issue command
     */
    if (!EepromOutput(CommandBits, 3, TRUE))
    {
        DBGME(3,_tprintf(_T("C512::EepromWrite no out0\n")));
        return E_FAIL;
    }
    
    if (!EepromOutput(Data,nBytes,FALSE))
    {
        DBGME(3,_tprintf(_T("C512::EepromWrite no out2\n")));
        return E_FAIL;
    }
    
    EepromStop();

    /* Then wait till its done. Says it should take about 5 or 10-20 msecs
     * Dont wait forever
     */
#define MAXPOLL 10
    for (i = 0; i < MAXPOLL; i++) {

        /* Sleep a bit
         */
        Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(5)));
        SleepUntil(t);

        /* Is it done yet ?
         */
        if (EepromReady(pC512,Offset,CommandBits))
            break;
    }

    return ((i < MAXPOLL) && (CommandBits[0] == Data[0])) ? NO_ERROR : E_FAIL;
#undef MAXPOLL
}

#endif

/* Common code for read&write argument validation
 */
PRIVATE SCODE EepromValidateArguments(PCC512 pC512,
                                      UINT Offset,
                                      UINT *pSize,
                                      UINT *pSizeWrite)
{
    UINT End, PageEnd;

    /* Validate args
     */
    if (Offset >= pC512->Size) {
        *pSize = 0;
        DBGME(1,_tprintf(_T("C512::EepromValidateArg(%d) EOF\n"),Offset));
        return S_FALSE; /* EOF */
    }

    /* See where we want to go
     */
    End = Offset + *pSize;
    if (End > pC512->Size)
    {
        End = pC512->Size;
        *pSize = End - Offset;
    }

    /* Writes would roll-over at page boundaries, trim ifso
     */
    if (pSizeWrite)
    {
        PageEnd = (Offset / EE_PAGESIZE) + 1;
        PageEnd = PageEnd * EE_PAGESIZE;
        if (End > PageEnd)
        {
            *pSizeWrite = *pSize - (End - PageEnd);
        }
    }

    DBGME(0,_tprintf(_T("C512::EepromValidateArg %d %d %d\n"),
                     Offset,*pSize,
                     (pSizeWrite) ? *pSizeWrite : *pSize));
    return NO_ERROR;
}

/* Read SIZE bytes from c512 device, at offset Position
 */
PRIVATE SCODE MCT C512ReadAt (PCDEVICE this,
                              UINT64 Position,
                              BYTE *pBuffer,
                              UINT Size,
                              PUINT pSizeRead)
{
    PCC512 pC512 = pCc(this);
    UINT Offset;
    SCODE sc;

    DBGME(1,_tprintf(_T("C512ReadAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeRead));

    Offset = Uint64ToUint32(Position);
    sc = EepromValidateArguments(pC512,
                                 Offset,
                                &Size,
                                 NULL);

    if (sc == NO_ERROR) {
        Mutex_Lock(&pC512->Lock);
        (void) EepromRead(pC512,Offset,pBuffer,Size);
        Mutex_Unlock(&pC512->Lock);
    }

    if (pSizeRead)
        *pSizeRead = Size;
    return sc;
}

/* Write SIZE bytes to device.
 */
PRIVATE SCODE MCT C512WriteAt (PCDEVICE this,
                               UINT64 Position,
                               const BYTE *pBuffer,
                               UINT Size,
                               PUINT pSizeWritten)
{
#if ALLOW_WRITE_OPERATIONS

    PCC512 pC512 = pCc(this);
    UINT Offset;
    SCODE sc;

    DBGME(1,_tprintf(_T("C512WriteAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeWritten));


    Offset = Uint64ToUint32(Position);
    sc = EepromValidateArguments(pC512,
                                 Offset,
                                &Size,
                                &Size);

    if (sc == NO_ERROR) {

        Mutex_Lock(&pC512->Lock);

        /* Hopefully we'll do one full page
         */
        sc = EepromWrite(pC512,Offset,pBuffer,Size);

        Mutex_Unlock(&pC512->Lock);
    }

    if (pSizeWritten)
        *pSizeWritten = Size;
    return sc;

#else
    /* Keep eeprom read-only */
    UnusedParameter(this);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);

    if (pSizeWritten)
        *pSizeWritten = 0;
    return E_ACCESS_DENIED;
#endif
}

/* GetSize method on c512 device
 */
PRIVATE SCODE MCT C512GetSize (PCDEVICE this,
                               PUINT64 pSize)
{
    PCC512 pC512 = pCc(this);

    if (pSize)
        Uint32ToUint64(*pSize,pC512->Size);
    return NO_ERROR;
}

#define C512ThreadLoop NULL

struct CDeviceVtbl C512DriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    C512ReadAt,
    C512WriteAt,
    NULL,
    C512GetSize,
    NULL,
    NULL,
    NULL,
    C512ThreadLoop
};

/* Constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE C512New(void)
{
    PCC512 pC512;
    UINT8 b;
    PIHEAP pH = CurrentHeap();

    /* Initialize things
     */
    InitWires();
    EEpromReset();

    /* Allocate resources
     */
    pC512 = (PCC512) pH->v->Alloc( pH, HEAP_ZERO_MEMORY, sizeof(CC512), 0 );
    if (pC512) {
        pC512->CDevice.v = &C512DriverVtable;
        pC512->CDevice.RefCnt = 1;
        Mutex_Init(&pC512->Lock);
        pC512->CurPos = ~0;
        /* BUGBUG To handle other chips this should be an argument
         */
        pC512->Size = 524288/8;

        DBGME(1,_tprintf(_T("C512New(%x %x)\n"),(ADDRESS)pC512,pC512->Size));

        /* Make sure the device is there
         */
        if (EepromReady(pC512,0,&b))
        {
            printf("b is %x\n", b);
            return (PCDEVICE)pC512;
        }

        /* Nope, release
         */
        pH->v->Free( pH, 0, pC512 );
    }

    FinitWires();
    return NULL;
}

/* ** 
*/
SCODE __cdecl C512Constructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Name);
    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("C512Constructor(%x %s %x %x %x)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) C512New();

    DBGME(0,_tprintf(_T("C512Constructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("C512Constructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl C512DriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) C512Constructor
};

static struct _DRIVERCOB C512DriverCobObject = { &C512DriverCobVtable, 1 };

PIUNKNOWN C512CobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &C512DriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &C512DriverVtable, _T("CDevice"),
                          &IID_IDevice);
    if (FAILED(sc))
        return NULL;

    /* Void the thread loop, to save a useless thread
     */
    C512DriverVtable.ThreadLoop = C512ThreadLoop;

    return (PIUNKNOWN) &C512DriverCobObject;
}

