/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    tat24c512.c
 *
 * Purpose:
 *
 *    Test program for Driver for the Atmel AT24C512 Serial EEPROM device.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include "at24c512.c"
#include <stdlib.h>

#define Nbuf EE_PAGESIZE*2
UINT8 buf[Nbuf];

static void calibrate(_TCHAR t)
{
    UINT pin = 0;
    UINT i;
    _TCHAR *pinname = _T("");

    switch (t) {
    case 'c':
    case 'C':
        pinname = _T("SCL: PA20/SCK2 P3-38");
        pin = SCL; 
        break;
    case 'd':
    case 'D':
        pinname = _T("SDA: PA21/TXD2 P3-39");
        pin = SDA;
        break;
    default:
        _tprintf(_T("Usage: [c|d]\n"));
        return;
    }

    _tprintf(_T("Calibrating pin %s\n"), pinname);
    ThePio->Enable = pin;
    ThePio->OutEnable = pin;
    ThePio->SetData = pin;
    for (i = 0; i < 8; i++) {
        _tprintf(_T("tic "));
        if (i & 1)
            ThePio->SetData = pin;
        else
            ThePio->ClearData = pin;
    }

    ThePio->OutDisable = pin;
    _tprintf(_T("\n"));
}

int _tmain(INT argc, _TCHAR **argv)
{
    PCDEVICE pCC;
    PIUNKNOWN pCob;
    int i;
    _TCHAR *msg = NULL;
    SCODE sc;

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_TEXT("No command line args (%x), test FAILED\r\n"), sc);
        exit(EXIT_FAILURE);
    }

    /* Special: analyzer calibration
     */
    if (argc > 1) {
        calibrate(argv[1][0]);
        goto Done;
    }

    sc = E_FAIL; /*unless*/
#define failure(_m_)  {msg = _m_; goto Done;}

    pCob = C512CobMain();
    pCC = C512New();
    if ((pCC == NULL) || pCob == NULL)
    {
        _tprintf(_T("New failed %x %x\n"),(ADDRESS)pCob,(ADDRESS)pCC);
        failure(_T("no chip?"));
        goto Done;
    }


#if 1
    /* First test: make sure reading one byte at a time 
     * gives the same results as reading all at once
     */
    memset(buf,0,sizeof buf);
#define N (Nbuf/2)
    if (!EepromRead((PCC512)pCC,0,buf,N))
        failure(_T("er0 failed\n"));

    for (i = 0; i < N; i++)
    {
        if (!EepromRead((PCC512)pCC,i,buf+N+i,1))
            failure(_T("er1 failed\n"));
    }

    if (memcmp(buf,buf+N,N)) {
        _tprintf(_T("Different!!\n"));
        for (i = 0; i < N; i++)
            _tprintf(_T("%x ? %x\n"), buf[i], buf[N+i]);
        failure(_T("cmp0 failed"));
    }
#endif

#if 1
    /* Second test: see if there is anything that is not FF in the EEPROM
     */
    {
        int j;
        UINT8 b = ~buf[0];
        for (i = 0; i < 0x10000/EE_PAGESIZE; i++) {
            if (!EepromRead((PCC512)pCC,i*EE_PAGESIZE,buf,EE_PAGESIZE))
                failure(_T("er2 failed\n"));
            for (j = 0; j < EE_PAGESIZE; j++)
                if (buf[j] != b) {
                    b = buf[j];
                    _tprintf(_T("[%d,%d,x%x]\n"), i, j, b);
                }
        }        
    }

#endif

#if 1
    /* Third test: write page #1 with a given pattern,
     * read it back and see if it takes
     */
    for (i = 0; i < EE_PAGESIZE; i++)
        buf[i] = (UINT8) i;
    sc = EepromWrite((PCC512)pCC,EE_PAGESIZE,buf,EE_PAGESIZE);
    if (FAILED(sc))
        failure(_T("cant write page1"));

    if (!EepromRead((PCC512)pCC,EE_PAGESIZE,buf+EE_PAGESIZE,EE_PAGESIZE))
        failure(_T("cant readback page1"));

    if (memcmp(buf,buf+EE_PAGESIZE,EE_PAGESIZE)) {
        _tprintf(_T("Different!!\n"));
        for (i = 0; i < EE_PAGESIZE; i++)
            _tprintf(_T("%x ? %x\n"), buf[i], buf[EE_PAGESIZE+i]);
        failure(_T("cmp0 failed"));
    }
#endif

    pCC->v->Release(pCC);
    sc = S_OK;
 Done:
    _tprintf(_T("Done %s\n"), (FAILED(sc)) ? msg : _T("ok"));
    (void) FreeCmdLine(&argc,&argv);
    exit(EXIT_SUCCESS);
    return 0;
}

