/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    bv16.c
 *
 * Purpose:
 *
 *    Driver for the Atmel AT49BV1604(T) AT49BV/LV8011(T) FLASH memory devices.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <diagnostics.h>
#include <assert.h>
#include <drivers/drivers.h>
#include <arm\at91m63200.h>

#define ALLOW_WRITE_OPERATIONS 0

/*
 * Concrete class definition for a bv16 device
 */
typedef const struct _SMAP *PSMAP;

typedef struct CBv16 {
    struct CDevice CDevice;
    /* Our state
     */
    UINT32 Size;  /* in bytes */
    PSMAP Map;
    UINT8 *Base;
    ADDRESS Skip;
#if ALLOW_WRITE_OPERATIONS
    MUTEX Lock;
    UINT CachedSectorStart;
#define INVALID_SECTOR_START (~0)
    UINT CachedSectorSize;
    UINT8 *SectorData;
    TIME LastWriteTime;
#endif
} CBv16, *PCBv16;

#define pBv(x) (PCBv16)x

/* Product ID codes
 */
#define MANUF_ATMEL  0x1f
#define DEVICE_16x4  0xc0
#define DEVICE_16x4T 0xc2
#define DEVICE_8011  0xcb
#define DEVICE_8011T 0x4a

/* Sector size maps
 */
#define SECTOR_UNIT (1024)  /* to compress maps */
#define nDELTAS 6
typedef const struct _SMAP {
    struct {
        UINT8 nSectors;
        UINT8 SizeInUnits;
    } Deltas[nDELTAS];
    UINT8 DeviceCode;
    UINT8 TotalSectors;
    _TCHAR *DeviceName;
} SMAP;

#define nMAPS (sizeof SectorMaps / sizeof SectorMaps[0] )
const struct _SMAP SectorMaps[] = {

    /* If the FLASH is mapped normally [E7=="STD" on EB63/55 boards]
     */
    {
     {{8,8},{2,32},{30,64},{0,0},{0,0},{0,0}},
     DEVICE_16x4, 40,
     _T("AT49BV16x4")
    },
    {
     {{30,64},{2,32},{8,8},{0,0},{0,0},{0,0}},
     DEVICE_16x4T, 40,
     _T("AT49BV16x4T")
    },

    /* With the funny toggle to the A20 line [E7=="USER" on EB63/55 boards]
     */
    {
     {{16,64},{8,8},{2,32},{14,64},{0,0},{0,0}},
     DEVICE_16x4, 40,
     _T("AT49BV16x4 ~A20")
    },
    {
     {{14,64},{2,32},{8,8},{16,64},{0,0},{0,0}},
     DEVICE_16x4T, 40,
     _T("AT49BV16x4T ~A20")
    },

    /* This one is on-chip and we dont play any tricks
     */
    {
     {{1,16},{1,32},{4,8},{1,32},{1,16},{14,64}},
     DEVICE_8011, 22,
     _T("AT49xV8011")
    },
    {
     {{14,64},{1,16},{1,32},{4,8},{1,32},{1,16}},
     DEVICE_8011T, 22,
     _T("AT49xV8011T")
    }
};

/* Returns the size in SECTOR_UNITs of a given sector,
 * or -1 for bad arguments.
 */
PRIVATE int UnitsInSector(PSMAP SecMap, UINT SecNo)
{
    int i;

    for (i = 0; i < nDELTAS; i++) {
        if (SecNo < SecMap->Deltas[i].nSectors)
            return SecMap->Deltas[i].SizeInUnits;
        SecNo -= SecMap->Deltas[i].nSectors;
    }

    return -1;
}

#define SectorSize(_map_,_sector_) (SECTOR_UNIT * UnitsInSector(_map_,_sector_))

/* Whats the starting offset of sector N
 */
PRIVATE ADDRESS SectorStart(PSMAP SecMap, int SecNo)
{
    int i;
    ADDRESS Offset = 0;

    for (i = 0; i < nDELTAS; i++) {
        if (SecNo < SecMap->Deltas[i].nSectors)
            return SECTOR_UNIT * (Offset + (SecMap->Deltas[i].SizeInUnits * SecNo));
        SecNo -= SecMap->Deltas[i].nSectors;
        Offset += SecMap->Deltas[i].nSectors * SecMap->Deltas[i].SizeInUnits;
    }

    return ~0;
}

/* Whats the ending offset (exclusive) of sector N
 */
PRIVATE ADDRESS SectorEnd(PSMAP SecMap, int SecNo)
{
    ADDRESS Offset = SectorStart(SecMap,SecNo);
    if (Offset != ~0)
        return Offset + (SECTOR_UNIT * UnitsInSector(SecMap,SecNo));

    return ~0;
}

#define FlashSize(_map_) SectorEnd(_map_,(_map_)->TotalSectors-1)

#if ALLOW_WRITE_OPERATIONS

/* What is the size of the largest sector
 */
UINT LargestSectorSize(PSMAP SecMap)
{
    int i;
    UINT Units = 0;

    for (i = 0; i < nDELTAS; i++) {
        if (Units < SecMap->Deltas[i].SizeInUnits)
            Units = SecMap->Deltas[i].SizeInUnits;
    }

    return Units * SECTOR_UNIT;
}

/* All info for the sector starting at the given sector:
 * sector number (return value), size and starting offset
 */
PRIVATE UINT SectorInfo(PSMAP SecMap, ADDRESS Offset,
                        UINT *pSize, UINT *pStart)
{
    UINT i;
    UINT SecNo = 0, Temp;
    ADDRESS Start = 0;

    Offset /= SECTOR_UNIT;
    for (i = 0; i < nDELTAS; i++) {

        /* How big is this run of sectors
         */
        Temp = SecMap->Deltas[i].nSectors * SecMap->Deltas[i].SizeInUnits;

        /* Are we smack in the middle of it ?
         */
        if (Offset < Temp) {

            /* So that is the size then
             */
            *pSize = SECTOR_UNIT * SecMap->Deltas[i].SizeInUnits;

            /* How many sectors do we skip from the start of this chunk
             */
            Temp = Offset / SecMap->Deltas[i].SizeInUnits;

            /* Return start and sector number 
             */
            *pStart = SECTOR_UNIT * (Start +
                                     (Temp * SecMap->Deltas[i].SizeInUnits));
            return SecNo + Temp;
        }
        /* Nope, next chunk
         */
        SecNo += SecMap->Deltas[i].nSectors;
        Offset -= Temp;
        Start += Temp;
    }

    return ~0;
}
#endif

/* Use 8/32bit accesses. If not StrongARM, the ARM processor doesnt have the
 * 16bit load/store operations. Which means a 16bit store gets compiled
 * into two 8bit stores. A 16bit load becomes a 32bit load, meaning
 * two 16bit loads if the bus is a 16bit bus.
 * This can make for very confusing results and/or strange code.
 *
 * Also, to make things even more confusing. The EB63 board uses the part
 * in 16bit mode. Addresses are shifted down one line, the processor's
 * A0 line (pA0) is not connected, pA1 goes to A0, pA2 to A1...pA20->A19.
 * The data bus is mapped normally, pD0->D0..pD15->D15.
 *
 * Keeping this all in mind.. it can be done. Sigh.
 */
typedef volatile UINT8 *WRITE_POINTER;

#if defined(arm) && defined(__GNUC__)
#define WriteDamnIt(_where_,_what_) \
  __asm__ volatile ("strh %1,[%0]" :: "r" (_where_), "r" (_what_) )
#else
#define WriteDamnIt(_where_,_what_) *(_where_) = _what_
#endif

/* Enter the Product ID mode for the Atmel part
 */
PRIVATE void ProductIdEnter(WRITE_POINTER pBase)
{
    pBase[0x5555*2] = 0xaa;
    pBase[0x2aaa*2] = 0x55;
    pBase[0x5555*2] = 0x90;
}

/* Exit the Product ID mode for the Atmel part
 */
PRIVATE void ProductIdExit(WRITE_POINTER pBase)
{
    pBase[0x5555*2] = 0xf0;
}

/* Identify the sector map for a given chip.
 * Argument is the base address of the device
 * [or suitably aligned alias thereof].
 * Returns NULL if failed
 */
PRIVATE PSMAP FlashIdentify(ADDRESS FlashBase, BOOL UserStrap)
{
    WRITE_POINTER pBase = (WRITE_POINTER) FlashBase;
    UINT8 Mid, Did;
    int i;

    ProductIdEnter(pBase);
    Mid = pBase[0*2];
    Did = pBase[1*2];
    ProductIdExit(pBase);

    DBGME(0,_tprintf(_T("ManufacturerId: x%x  DeviceId: x%x\r\n"), Mid, Did));

    for (i = 0; i < nMAPS; i++)
        if (SectorMaps[i].DeviceCode == Did)
            return SectorMaps+i+((UserStrap) ? 2 : 0);
    return NULL;
}

#if ALLOW_WRITE_OPERATIONS

/* Erase one sector of the Atmel part, do not wait till done
 */
PRIVATE void SectorEraseStart(PCBv16 Bv16, UINT SectorStart)
{
    WRITE_POINTER pBase = (WRITE_POINTER) Bv16->Base;

    DBGME(1,_tprintf(_T("Erasing sector at offset x%x ..."), SectorStart));

    pBase[0x5555*2] = 0xaa;
    pBase[0x2aaa*2] = 0x55;
    pBase[0x5555*2] = 0x80;
    pBase[0x5555*2] = 0xaa;
    pBase[0x2aaa*2] = 0x55;
    pBase[SectorStart] = 0x30;
}

/* Wait until the erase of the part is complete
 */
PRIVATE BOOL SectorEraseComplete(PCBv16 Bv16, UINT SectorStart)
{
    WRITE_POINTER pBase = (WRITE_POINTER) Bv16->Base;
    volatile UINT8 Data, Data1;
    UINT16 i;
    TIME t;

    DBGME(1,_tprintf(_T("EraseComplete? at offset x%x ..."), SectorStart));

    /* Wait until the erase is actually completed
     * Use the TOGGLE BIT method.
     */
    Data = 0;
    Data1 = ~Data;
    for (i = 0; ; i++) {
        /* Are we there yet
         */
        Data  = pBase[SectorStart];
        Data1 = pBase[SectorStart];
        if (Data == Data1)
            break;

        /* Dont wait too long
         */
        if (i >= 100) break;

        /* Supposedly an erase takes 200 msec, average,
         * but we do not know how long ago it was started.
         * Check at 10msec intervals, timeout in one second or so.
         */
        Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(10)));
        SleepUntil(t);
    }

    if (Data == Data1) {
        DBGME(1,_tprintf(_T("done.\r\n")));
        return TRUE;
    } else {
        DBGME(3,_tprintf(_T("Sector at %x NOT erased [%x != %x]\r\n"), 
                         SectorStart, Data, Data1));
        return FALSE;
    }
}

/* Write one word.  
 * Since Vpp is not provided this is the only way to write to it.
 */
PRIVATE BOOL
WordProgram(PCBv16 Bv16,
            ADDRESS Offset,
            UINT16 Value)
{
    WRITE_POINTER pBase = (WRITE_POINTER) Bv16->Base;
    UINT8 Data, Data1;
    UINT16 i, Data16;

#if _DEBUG
    Data16 = pBase[Offset] | (pBase[Offset+1] << 8);
    if (Data16 != 0xffff)
        DBGME(3,_tprintf(_T("Offset %x not ERASED, wont take.\r\n"),Offset));
#endif

    pBase[0x5555*2] = 0xaa;
    pBase[0x2aaa*2] = 0x55;
    pBase[0x5555*2] = 0xa0;

    WriteDamnIt((UINT16*)(pBase+Offset),Value);

    /* Wait until the operation is completed
     * It should take about 20/50 usecs.
     */
    Delay(50);

    /* Use the BIT TOGGLE method
     */
    Data = 0;
    Data1 = ~Data;
    for (i = 0; Data != Data1; i++) {
        Data  = pBase[Offset];
        Data1 = pBase[Offset];
        /* Do not wait too long 
         */
        if (i == 0xffff) break;
    }

    /* Check to see if it sticks
     */
    Data16 = pBase[Offset] | (pBase[Offset+1] << 8);
             
    if (Data16 != Value) {
        DBGME(3,_tprintf(_T("WordProgram failed to write @%x [%x != %x]\r\n"), 
                 Offset, Data16, Value));
        return FALSE;
    }

    return TRUE;
}

/* Writeback the cached sector
 */
PRIVATE SCODE
FlashSectorWrite(PCBv16 Bv16)
{
    UINT i, nTries;
    UINT16 *Buffer = (UINT16 *)Bv16->SectorData;
    UINT Offset = Bv16->CachedSectorStart;

    if (!SectorEraseComplete(Bv16,Offset))
        return E_FAIL;

    for (i = 0; i < Bv16->CachedSectorSize/2; i++) {

        for (nTries = 0; nTries < 5; nTries++)
            if (WordProgram(Bv16,Offset,Buffer[i]))
                break;
        Offset += 2;
    }

    DBGME(1,_tprintf(_T("SectorWrite complete, offset %x\r\n"),Offset));

    return NO_ERROR;
}
#endif

/* Read SIZE bytes from a bv16 device, at offset Position
 */
PRIVATE SCODE MCT Bv16ReadAt (PCDEVICE this, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    PCBv16 Bv16 = pBv(this);
    UINT Offset;
    SCODE sc;

    DBGME(0,_tprintf(_T("Bv16ReadAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeRead));

    Offset = Uint64ToUint32(Position) + Bv16->Skip;
    if (Offset >= Bv16->Size) {
        sc = S_FALSE;
        Size = 0;
        goto Done;
    }

    if ((Size > Bv16->Size) || ((Offset + Size) > Bv16->Size))
        Size = Bv16->Size - Offset;

    /* BUGBUG Cannot read from the same memory plane if an erase is in progress
     */
    memcpy(pBuffer,Bv16->Base + Offset, Size);

    sc = NO_ERROR;
 Done:
    if (pSizeRead)
        *pSizeRead = Size;
    return sc;
}

/* Write SIZE bytes to device.
 */
PRIVATE SCODE MCT Bv16WriteAt (PCDEVICE this, UINT64 Position, const BYTE *pBuffer,
                            UINT Size, PUINT pSizeWritten)
{
#if ALLOW_WRITE_OPERATIONS

    PCBv16 Bv16 = pBv(this);
    SCODE sc;
    UINT Offset, SectorNumber, SectorSize, SectorStart;

    DBGME(0,_tprintf(_T("Bv16WriteAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeWritten));

    /* Validate args
     */
    Offset = Uint64ToUint32(Position) + Bv16->Skip;
    if (Offset >= Bv16->Size) {
        DBGME(1,_tprintf(_T("Bv16WriteAt:: (%d) EOF\n"),Offset));
        Size = 0;
        sc = S_FALSE; /* EOF */
        goto Done;
    }

    /* See where we want to go
     */
    SectorNumber = SectorInfo(Bv16->Map,Offset,&SectorSize,&SectorStart);
    Offset = Offset - SectorStart;

    /* We wont do more than what is in one sector
     */
    if (Size > (SectorSize - Offset))
        Size = SectorSize - Offset;

    Mutex_Lock(&Bv16->Lock);

    /* Are we missing in the sector cache ?
     */
    if (Bv16->CachedSectorStart != SectorStart) {
        /* Yes, do we have to writeback
         */
        if (Bv16->CachedSectorStart != INVALID_SECTOR_START) {
            sc = FlashSectorWrite(Bv16);
            /* BUGBUG async error reporting !! */
        }

        /* Cache & erase the sector, in preparation for writing
         */
        memcpy(Bv16->SectorData, Bv16->Base + SectorStart, SectorSize);
        Bv16->CachedSectorSize = SectorSize;
        Bv16->CachedSectorStart = SectorStart;

        /* Get the erase started
         */
        SectorEraseStart(Bv16,SectorStart);
    }

    /* Modify the cached sector data, timestampp it.
     */
    memcpy(Bv16->SectorData+Offset,pBuffer,Size);
    Bv16->LastWriteTime = CurrentTime();

    Mutex_Unlock(&Bv16->Lock);

    /* Trigger the writeback timeout
     */
    Condition_Signal(&Bv16->CDevice.IntrPoint);

    sc = NO_ERROR;
 Done:
    if (pSizeWritten)
        *pSizeWritten = Size;
    return sc;

#else
    /* Keep flash read-only */
    UnusedParameter(this);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);
    if (pSizeWritten)
        *pSizeWritten = 0;
    return E_ACCESS_DENIED;
#endif
}

/* GetSize method on bv16 device
 */
PRIVATE SCODE MCT Bv16GetSize (PCDEVICE this, PUINT64 pSize)
{
    PCBv16 Bv16 = pBv(this);

    if (pSize)
        Uint32ToUint64(*pSize,Bv16->Size);
    return NO_ERROR;
}

#if ALLOW_WRITE_OPERATIONS

PRIVATE void Bv16ThreadLoop(PCDEVICE this)
{
    PCBv16 Bv16 = pBv(this);
    static const TIME CacheLifeTime = Int64Initializer(0,TIME_MILLIS(500));
    TIME t;

    Mutex_Lock(&Bv16->Lock);
    for (;;) {
        /* Wait till someone writes
         */
        Condition_Wait(&Bv16->CDevice.IntrPoint,
                       &Bv16->Lock,
                       NULL);

        /* Now wait for the cache lifetime
         */
        for (;;) {
            t = CurrentTime();
            t = TimeSubtract(t,Bv16->LastWriteTime);

            /* Someone refreshed/wrote the cache ?
             */
            if (TimeLess(t,CacheLifeTime)) {

                Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(500)));

                Mutex_Unlock(&Bv16->Lock);
                SleepUntil(t);
                Mutex_Lock(&Bv16->Lock);
            } else {
                break;
            }
        }

        /* Write back & invalidate the buffer, if it is still valid
         */
        if (Bv16->CachedSectorStart != INVALID_SECTOR_START) {
            (void) FlashSectorWrite(Bv16);
            Bv16->CachedSectorStart = INVALID_SECTOR_START;
        }
    }
    Mutex_Unlock(&Bv16->Lock);
}

#else
#define Bv16ThreadLoop NULL
#endif

struct CDeviceVtbl Bv16DriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    Bv16ReadAt,
    Bv16WriteAt,
    NULL,
    Bv16GetSize,
    NULL,
    NULL,
    NULL,
    Bv16ThreadLoop
};

/* Constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE Bv16New(ADDRESS Base, BOOL FlipA20, ADDRESS Skip)
{
    PCBv16 Bv16;
    PSMAP Map;
    UINT MaxSectorSize;
    PIHEAP pH = CurrentHeap();

    /* Verify args and identify device
     */
    Map = FlashIdentify(Base,FlipA20);
    if (Map == NULL) {
        DBGME(3,_tprintf(_T("Bv16New: unk @x%x\n"),Base));
        return NULL;
    }

    /* Allocate descriptor and, if writing, room for one sector cache
     */
#if ALLOW_WRITE_OPERATIONS
    MaxSectorSize = LargestSectorSize(Map);
#else
    MaxSectorSize = 0;
#endif

    Bv16 = (PCBv16) pH->v->Alloc( pH, 0, sizeof(CBv16) + MaxSectorSize, 0 );
    if (Bv16) {
        Bv16->CDevice.v = &Bv16DriverVtable;
        Bv16->CDevice.RefCnt = 1;
        Bv16->Map = Map;
        Bv16->Base = (UINT8 *)Base;
#if ALLOW_WRITE_OPERATIONS
        Mutex_Init(&Bv16->Lock);
        Bv16->CachedSectorStart = INVALID_SECTOR_START;
        Bv16->CachedSectorSize = 0;
        Bv16->LastWriteTime = CurrentTime();
        Bv16->SectorData = (UINT8*) (Bv16 + 1);
#endif
        Bv16->Size = FlashSize(Map);
        /* Might want to protect the first part of it */
        Bv16->Skip = Skip;
        DBGME(1,_tprintf(_T("Bv16New(%x %s %x)\n"),
                         (ADDRESS)Bv16,
                         Map->DeviceName,
                         Bv16->Size - Skip));
    }
    return (PCDEVICE)Bv16;
}

/* ** 
*/
PRIVATE SCODE __cdecl 
Bv16Constructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        PTR NoIdeaWhatsThisFor,
        PTR NoIdeaWhatsThisForNeither,
        ADDRESS SizeOfHiddenFirstPortion,
        BOOL A20AddressLineIsFlipped,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Name);
    UnusedParameter(Interrupt);

    DBGME(3,_tprintf(_T("Bv16Constructor(%x %s %x %x %x %x %x)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     SizeOfHiddenFirstPortion, A20AddressLineIsFlipped,
                     (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) Bv16New(RegisterBase,
                               A20AddressLineIsFlipped,
                               SizeOfHiddenFirstPortion);

    DBGME(0,_tprintf(_T("Bv16Constructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("Bv16Constructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl Bv16DriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) Bv16Constructor
};

static struct _DRIVERCOB Bv16DriverCobObject = { &Bv16DriverCobVtable, 1 };

PIUNKNOWN Bv16CobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &Bv16DriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &Bv16DriverVtable, _T("CDevice"),
                          &IID_IDevice);
    if (FAILED(sc))
        return NULL;

    /* Void the thread loop, to save a useless thread
     */
    Bv16DriverVtable.ThreadLoop = Bv16ThreadLoop;

    return (PIUNKNOWN) &Bv16DriverCobObject;
}

