/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    db161.c
 *
 * Purpose:
 *
 *    Driver for the Atmel AT45DB161 Serial FLASH memory device.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
//#define _DEBUG 1
//#define DBGLVL 1
#include <diagnostics.h>
#include <assert.h>
#include <drivers/drivers.h>
#include <arm\at91m63200.h>

#define ALLOW_WRITE_OPERATIONS 1
#define USE_WRITE_CACHE 1

/*
 * Concrete class definition for a db161 device
 */

typedef struct CDb161 {
    struct CDevice CDevice;
    /* Our state
     */
    UINT32 Size;  /* in bytes */
    MUTEX Lock;
#if USE_WRITE_CACHE
    UINT CachedPage;
    TIME LastWriteTime;
#endif
} CDb161, *PCDb161;

#define pDb(x) (PCDb161)x

/* The board connects the uC to the device as follows:
 * nCS == PA26 aka NPCS0
 *  SO == PA24 aka MOSI
 *  SI == PA25 aka MISO
 * SCK == PA23 aka SPCK
 * nRESET = NRST
 * RDY == Vcc
 * nWP == Vcc
 * Additionally, a strap E12 can pull nCS high. It must be
 * installed in the 2-3 position to enable the device.
 * This is *not* always the factory setting.
 *
 * There are two ways to talk to the device, using the SPI interface
 * or directly manipulating the I/O pins.  
 * The SPI interface is limited to 1.2 Mbps, the device to 13 Mbps max.
 * Depending on the uC' performance direct manipulation could be faster.
 */

/* Opcodes
 */
#define SF_PAGE_READ            0x52
#define SF_BUFFER1_READ         0x54
#define SF_BUFFER2_READ         0x56
#define SF_PAGE_CACHE1          0x53
#define SF_PAGE_CACHE2          0x55
#define SF_PAGE_COMPARE1        0x60
#define SF_PAGE_COMPARE2        0x61
#define SF_BUFFER1_WRITE        0x84
#define SF_BUFFER2_WRITE        0x87
#define SF_BUFFER1_FLUSH_PERASE 0x83
#define SF_BUFFER2_FLUSH_PERASE 0x86
#define SF_BUFFER1_FLUSH        0x88
#define SF_BUFFER2_FLUSH        0x89
#define SF_PAGE_ERASE           0x81
#define SF_BLOCK_ERASE          0x50
#define SF_PAGE_PROGRAM1        0x82
#define SF_PAGE_PROGRAM2        0x85
#define SF_PAGE_REWRITE1        0x58
#define SF_PAGE_REWRITE2        0x59
#define SF_READ_STATUS          0x57

/* Status bits (MSB order)
 */
#define SF_ST_READY 0x80
#define SF_ST_COMP  0x40
#define SF_ST_SIZE  0x38  /* should be 0x28 for db161 */
#define SF_ST_xx    0x07

/* Other defines
 */
#define SF_PAGESIZE 528 /* bytes */
#define INVALID_PAGE (~0)

/* PIO Interface
 */
#define ThePio PioA
#define nCS PIOA_NPCS0
#define SO  PIOA_MOSI
#define SI  PIOA_MISO
#define SCK PIOA_SPCK
#define WireIsIn(_p_)   ThePio->OutDisable = (_p_)
#define WireIsOut(_p_)  ThePio->OutEnable = (_p_)
#define WireLow(_p_)    ThePio->ClearData = (_p_)
#define WireHigh(_p_)   ThePio->SetData = (_p_)
#define WireIs(_p_)     ((ThePio->PinDataStatus & (_p_)) != 0)
#define WireEnable(_p_) ThePio->Enable = (_p_)

/* Initialization
 */
PRIVATE void InitWires(void)
{
    DBGME(0,_tprintf(_T("Db161::InitWires\n")));

    /* Before we enable them pins, set safe values.
     * All we really need is to keep the device disabled.
     */
    WireHigh(nCS);

    /* We will use the clock in SPI mode 3
     */
    WireHigh(SCK);

    /* Set proper directions
     */
    WireIsIn(SI);
    WireIsOut(nCS|SO|SCK);
 
    /* Make sure the Pio owns them
     */
    WireEnable(nCS|SI|SCK|SO);

    /* Done
     */
}

/* De/Select the device
 */
#define FlashEnable() WireLow(nCS)
#define FlashDisable() WireHigh(nCS)

/* Write a given number of bits to the device
 */
PRIVATE void FlashOutput(const UINT8 *Bits, UINT nBits)
{
    UINT i;
    UINT8 c = 0;

    DBGME(0,_tprintf(_T("Db161::FlashOutput(%x %d)\n"),(ADDRESS)Bits,nBits));

    /* Enable the device.
     * NB: If already enabled this is a no-op.
     */
    FlashEnable();

    /* Should wait 250 ns before lowering SCK.
     * Its a given I suppose, but..
     */
    Delay(0);

    /* Ok, do it one bit at a time
     */
    for (i = 0; i < nBits; i++) {
        /* Fetch next byte if necessary
         */
        if ((i % 8) == 0)
            c = *Bits++;
        
        /* Set the value before we toggle the clock
         * NB: Must be done MSB first, remember
         */
        if (c & 0x80) {
            WireHigh(SO);
        } else {
            WireLow(SO);
        }

        /* Toggle
         * SCK should stay down a minimum of 35 ns, again a given.
         */
        WireLow(SCK);
        WireHigh(SCK);

        /* The hold time is zero, we dont have to wait.
         * Before we can toggle SCK again we need 35 ns, again..
         */

        /* Next bit
         */
        c <<= 1;
    }    
}

/* Read a given number of bits from the device
 */
PRIVATE void FlashInput(UINT8 *Bits, UINT nBits)
{
    UINT i;
    UINT8 c = 0;

    DBGME(0,_tprintf(_T("Db161::FlashInput(%x %d)\n"),(ADDRESS)Bits,nBits));

    /* Device should already be selected or else
     * [it only talks to us if we send it a command]
     */
    assert(WireIs(nCS) == 0);

    /* Ok, do it one bit at a time
     */
    for (i = 0; i < nBits; i++) {

        /* Next bit
         */
        c <<= 1;

        /* Toggle
         * SCK should stay down a minimum of 10 ns, again a given.
         */
        WireLow(SCK);

        /* Get the value before we toggle the clock
         * NB: Must be done MSB first, remember
         */
        c |= (WireIs(SI)) ? 1 : 0;
        WireHigh(SCK);

        /* Before we can toggle SCK again we need 20 ns, again..
         */

        /* Store next byte if we are done
         */
        if ((i % 8) == 7)
            *Bits++ = c;
    }    

    /* In the unlikely case we want a weird number of bits..
     */
    if ((i % 8) != 0)
        *Bits++ = c;
}


/* Read the device's status
 */
PRIVATE UINT8 FlashStatus(void)
{
    UINT8 Status;

    Status = SF_READ_STATUS;
    FlashOutput(&Status, 8);
    FlashInput(&Status,8);
    FlashDisable();

    DBGME(1,_tprintf(_T("Db161::FlashStatus %x\n"),Status));

    return Status;
}

/* Decode the device's size in bits
 */
PRIVATE UINT32 FlashSize(UINT8 Status)
{
    Status = (Status & SF_ST_SIZE) >> 3;
    return (512*1024) << Status;
}

#if ALLOW_WRITE_OPERATIONS

/* Write data to Buffer #1
 */
PRIVATE void FlashBufferWrite(UINT Offset, const UINT8 *Data, UINT nBits)
{
    UINT8 CommandBits[4];

    DBGME(0,_tprintf(_T("Db161::FlashBufferWrite(%d,%x,%d)\n"),
                     Offset,(ADDRESS)Data,nBits));

    CommandBits[0] = SF_BUFFER1_WRITE;
    CommandBits[1] = 0;
    CommandBits[2] = (UINT8)(Offset >> 8);
    CommandBits[3] = (UINT8)Offset;
    FlashOutput(CommandBits,4*8);
    FlashOutput(Data,nBits);
    FlashDisable();
}

/* Write one page of data
 */
PRIVATE SCODE FlashPageWrite(UINT PageNumber, const UINT8 *Data)
{
    UINT8 CommandBits[4];
    int i, c = 0;

    DBGME(1,_tprintf(_T("Db161::FlashPageWrite(%d,%x)\n"),
                     PageNumber,(ADDRESS)Data));

    /* First xfer the bits to Buffer 1, unless they are there already
     */
    if (Data != NULL) {
        FlashBufferWrite(0,Data,SF_PAGESIZE*8);
    }

    /* Next issue the erase&program command
     */
    CommandBits[0] = SF_BUFFER1_FLUSH_PERASE;
    CommandBits[1] = (UINT8)(PageNumber >> 6);
    CommandBits[2] = (UINT8)(PageNumber << 2);
    CommandBits[3] = 0;

    FlashOutput(CommandBits,4*8);
    FlashDisable();

    /* Then wait till its done. Says it should take about 10-20 msecs
     * Dont wait forever
     */
    for (i = 0; i < 10; i++) {
        TIME t;

        /* Is it done yet ?
         */
        c = FlashStatus();
        if (c & SF_ST_READY)
            break;

        /* Nope, sleep a bit
         */
        Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(10)));
        SleepUntil(t);
    }

    return (c & SF_ST_READY) ? NO_ERROR : E_FAIL;
}

/* Cache one page of data to Buffer #1
 */
PRIVATE BOOL FlashPageCache(UINT PageNumber)
{
    UINT8 CommandBits[4];
    int i;

    DBGME(1,_tprintf(_T("Db161::FlashPageCache(%d)\n"),PageNumber));

    CommandBits[0] = SF_PAGE_CACHE1;
    CommandBits[1] = (UINT8)(PageNumber >> 6);
    CommandBits[2] = (UINT8)(PageNumber << 2);
    CommandBits[3] = 0;

    /* Issue the command
     */
    FlashOutput(CommandBits,4*8);
    FlashDisable();

    /* Wait for it to complete
     * Txfr is speced between 250 and 350 usecs
     */
    for (i = 0; i < 10; i++) {
        if (FlashStatus() & SF_ST_READY)
            return TRUE;
        Delay(250);
    }

    return FALSE;
}

#endif

/* Read one page of data from memory
 */
PRIVATE BOOL FlashPageRead(UINT PageNumber, UINT Offset, UINT8 *Data, UINT nBytes)
{
    UINT8 CommandBits[8];

    DBGME(1,_tprintf(_T("Db161::FlashPageRead(%d,%d,%x,%d)\n"),
                     PageNumber,Offset,(ADDRESS)Data,nBytes));

    assert(Offset < 2048);

    CommandBits[0] = SF_PAGE_READ;
    CommandBits[1] = (UINT8)(PageNumber >> 6);
    CommandBits[2] = (UINT8)((PageNumber << 2) | (Offset >> 8));
    CommandBits[3] = (UINT8)Offset;
    /* the last 32 bits are dont-care so we dont care */

    FlashOutput(CommandBits,8*8);
    FlashInput(Data,nBytes*8);
    FlashDisable();
    return TRUE;
}

/* Common code for read&write argument validation
 */
PRIVATE SCODE FlashValidateArguments(PCDb161 pDb161,
                                     UINT64 Position,
                                     UINT *pPage,
                                     UINT *pOffset,
                                     UINT *pSize)
{
    UINT Offset;

    /* Validate args
     */
    Offset = Uint64ToUint32(Position);
    if (Offset >= pDb161->Size) {
        *pSize = 0;
        DBGME(1,_tprintf(_T("Db161::FlashValidateArg(%d) EOF\n"),Offset));
        return S_FALSE; /* EOF */
    }

    /* See where we want to go
     * BUGBUG: Should use the two buffers as a cache
     */
    *pPage = Offset / SF_PAGESIZE;
    Offset = Offset - (*pPage * SF_PAGESIZE);

    /* We wont do more than whats in one page
     * BUGBUG: but the continuous read command...
     */
    if (*pSize > (SF_PAGESIZE - Offset))
        *pSize = SF_PAGESIZE - Offset;

    *pOffset = Offset;

    DBGME(0,_tprintf(_T("Db161::FlashValidateArg %d %d %d\n"),
                     *pPage,*pOffset,*pSize));
    return NO_ERROR;
}

/* Read SIZE bytes from db161 device, at offset Position
 */
PRIVATE SCODE MCT Db161ReadAt (PCDEVICE this, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    PCDb161 pDb161 = pDb(this);
    UINT Page, Offset;
    SCODE sc;

    DBGME(1,_tprintf(_T("Db161ReadAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeRead));

    sc = FlashValidateArguments(pDb161,
                                Position,
                                &Page,
                                &Offset,
                                &Size);

    if (sc == NO_ERROR) {
        Mutex_Lock(&pDb161->Lock);
        (void) FlashPageRead(Page,Offset,pBuffer,Size);
        Mutex_Unlock(&pDb161->Lock);
    }

    if (pSizeRead)
        *pSizeRead = Size;
    return sc;
}

/* Write SIZE bytes to device.
 */
PRIVATE SCODE MCT Db161WriteAt (PCDEVICE this, UINT64 Position, const BYTE *pBuffer,
                            UINT Size, PUINT pSizeWritten)
{
#if ALLOW_WRITE_OPERATIONS

    PCDb161 pDb161 = pDb(this);
    UINT Page, Offset;
    SCODE sc;

    DBGME(1,_tprintf(_T("Db161WriteAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeWritten));


    sc = FlashValidateArguments(pDb161,
                                Position,
                                &Page,
                                &Offset,
                                &Size);

    if (sc == NO_ERROR) {

#if USE_WRITE_CACHE
        Mutex_Lock(&pDb161->Lock);

        if (pDb161->CachedPage != Page) {
            if (pDb161->CachedPage != INVALID_PAGE) {
                sc = FlashPageWrite(pDb161->CachedPage,NULL);
                /* BUGBUG async error reporting !! */
            }
            FlashPageCache(Page);
            pDb161->CachedPage = Page;
        }
        FlashBufferWrite(Offset,pBuffer,Size*8);
        pDb161->LastWriteTime = CurrentTime();

        Mutex_Unlock(&pDb161->Lock);

        Condition_Signal(&pDb161->CDevice.IntrPoint);
#else

        Mutex_Lock(&pDb161->Lock);

        /* Hopefully we'll do one full page
         */
        if (Size == SF_PAGESIZE) {
            sc = FlashPageWrite(Page,pBuffer);
        } else {
            /* Urgh.. a wasteful non-intelligent client.
             * Do the read-modify-write. Sigh.
             */
            FlashPageCache(Page);
            FlashBufferWrite(Offset,pBuffer,Size*8);
            sc = FlashPageWrite(Page,NULL);
        }

        Mutex_Unlock(&pDb161->Lock);
#endif
    }

    if (pSizeWritten)
        *pSizeWritten = Size;
    return sc;

#else
    /* Keep flash read-only */
    UnusedParameter(this);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);

    if (pSizeWritten)
        *pSizeWritten = 0;
    return E_ACCESS_DENIED;
#endif
}

/* GetSize method on db161 device
 */
PRIVATE SCODE MCT Db161GetSize (PCDEVICE this, PUINT64 pSize)
{
    PCDb161 pDb161 = pDb(this);

    if (pSize)
        Uint32ToUint64(*pSize,pDb161->Size);
    return NO_ERROR;
}

#if USE_WRITE_CACHE

PRIVATE void Db161ThreadLoop(PCDEVICE this)
{
    PCDb161 pDb161 = pDb(this);
    static const TIME CacheLifeTime = Int64Initializer(0,TIME_MILLIS(500));
    TIME t;

    Mutex_Lock(&pDb161->Lock);
    for (;;) {
        /* Wait till someone writes
         */
        Condition_Wait(&pDb161->CDevice.IntrPoint,
                       &pDb161->Lock,
                       NULL);

        /* Now wait for the cache lifetime
         */
        for (;;) {
            t = CurrentTime();
            t = TimeSubtract(t,pDb161->LastWriteTime);

            /* Someone refreshed/wrote the cache ?
             */
            if (TimeLess(t,CacheLifeTime)) {

                Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(500)));

                Mutex_Unlock(&pDb161->Lock);
                SleepUntil(t);
                Mutex_Lock(&pDb161->Lock);
            } else {
                break;
            }
        }

        /* Write back & invalidate the buffer, if it is still valid
         */
        if (pDb161->CachedPage != INVALID_PAGE) {
            (void) FlashPageWrite(pDb161->CachedPage,NULL);
            pDb161->CachedPage = INVALID_PAGE;
        }
    }
    Mutex_Unlock(&pDb161->Lock);
}

#else
#define Db161ThreadLoop NULL
#endif

struct CDeviceVtbl Db161DriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    Db161ReadAt,
    Db161WriteAt,
    NULL,
    Db161GetSize,
    NULL,
    NULL,
    NULL,
    Db161ThreadLoop
};

/* Constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE Db161New(void)
{
    PCDb161 pDb161;
    PIHEAP pH = CurrentHeap();

    pDb161 = (PCDb161) pH->v->Alloc( pH, HEAP_ZERO_MEMORY, sizeof(CDb161), 0 );
    if (pDb161) {
        pDb161->CDevice.v = &Db161DriverVtable;
        pDb161->CDevice.RefCnt = 1;
        Mutex_Init(&pDb161->Lock);
#if USE_WRITE_CACHE
        pDb161->CachedPage = INVALID_PAGE;
        pDb161->LastWriteTime = CurrentTime();
#endif
        InitWires();
        pDb161->Size = FlashSize(FlashStatus()) / 8;
        DBGME(1,_tprintf(_T("Db161New(%x %x)\n"),(ADDRESS)pDb161,pDb161->Size));
    }
    return (PCDEVICE)pDb161;
}

/* ** 
*/
SCODE __cdecl Db161Constructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Name);
    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("Db161Constructor(%x %s %x %x %x)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) Db161New();

    DBGME(0,_tprintf(_T("Db161Constructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("Db161Constructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl Db161DriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) Db161Constructor
};

static struct _DRIVERCOB Db161DriverCobObject = { &Db161DriverCobVtable, 1 };

PIUNKNOWN Db161CobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &Db161DriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &Db161DriverVtable, _T("CDevice"),
                          &IID_IDevice);
    if (FAILED(sc))
        return NULL;

    /* Void the thread loop, to save a useless thread
     */
    Db161DriverVtable.ThreadLoop = Db161ThreadLoop;

    return (PIUNKNOWN) &Db161DriverCobObject;
}

