/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    db161.c
 *
 * Purpose:
 *
 *    Driver for the Intel 28F128 (J3A150) StrataFlash memory device.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#define _DEBUG 1
#define DBGLVL 1
#include <diagnostics.h>
#include <assert.h>
#include <drivers/drivers.h>

#define ALLOW_WRITE_OPERATIONS 1

/*
 * Concrete class definition for a db161 device
 */

typedef struct CStrata {
    struct CDevice CDevice;
    /* Our state
     */
    UINT8 *Base;
    UINT32 Size;  /* in bytes */
    UINT32 SectorSize;  /* in bytes */
    MUTEX Lock;
#if ALLOW_WRITE_OPERATIONS
    UINT CachedSectorStart;
#define INVALID_SECTOR_START (~0)
    UINT8 *SectorData;
    TIME LastWriteTime;
#endif
} CStrata, *PCStrata;

#define pSt(x) (PCStrata)x

/* Utilities
 */
/* Product ID codes
 */
#define MANUF_INTEL  0x89
#define DEVICE_320   0x16
#define DEVICE_640   0x17
#define DEVICE_128   0x18

/* Defines fo rthe STATUS register
 */
#define ST_reserved          0x01
#define ST_BLOCK_LOCKED      0x02
#define ST_PROGRAM_SUSPENDED 0x04
#define ST_LOW_VOLTAGE       0x08
#define ST_LOCK_BIT_ERROR    0x10
#define ST_ERASE_ERROR       0x20
#define ST_ERASE_SUSPENDED   0x40
#define ST_READY             0x80
#define ST_ERASE_MASK        0xee  /* bits to check after erase command */
#define ST_MASK              0xfe  /* ignore reserved */

/* Command set (what we use of it)
 */
#define CMD_READ_ARRAY    0xff
#define CMD_READ_ID       0x90
#define CMD_READ_STATUS   0x70
#define CMD_CLEAR_STATUS  0x50
#define CMD_WRITE_WORD    0x40
#define CMD_ERASE_SETUP   0x20
#define CMD_ERASE_CONFIRM 0xd0
#define CMD_SET_PREFIX    0x60  /* set read config, lock bits */
#define CMD_LOCK          0x01
#define CMD_UNLOCK        0xd0

/* Enter the Product ID mode (Read Identifier Codes)
 */
#define ProductIdEnter(_pBase_) \
{ \
      ((volatile UINT8 *)_pBase_)[0] = CMD_READ_ID; \
}

/* Exit the Product ID mode (enter Read Array mode)
 */
#define ProductIdExit(_pBase_) \
{ \
      ((volatile UINT8 *)_pBase_)[0] = CMD_READ_ARRAY; \
}

PRIVATE BOOL
FlashIdentify(ADDRESS Base,
              UINT32 *pSize,
              UINT32 *pSectorSize)
{
    static struct {
        UINT8 DeviceCode;
        UINT8 nKBperSector;
        UINT16 nSectors;
    } KnownDevices[] = {
        {  DEVICE_320, 32, 32 },
        {  DEVICE_640, 64, 64 },
        {  DEVICE_128, 128, 128 }
    };
#define nKNOWN (sizeof KnownDevices / sizeof KnownDevices[0] )
    int i;
    UINT8 Mid, Did;

    ProductIdEnter(Base);
    Mid = ((volatile UINT8 *)Base)[0*2];
    Did = ((volatile UINT8 *)Base)[1*2];
    ProductIdExit(Base);    

    for (i = 0; i < nKNOWN; i++) {
        if (KnownDevices[i].DeviceCode = Did) {
            *pSectorSize = 1024 * KnownDevices[i].nKBperSector;
            *pSize = *pSectorSize * KnownDevices[i].nSectors;
            return TRUE;
        }
    }

    DBGME(3,_tprintf(_T("FlashIdentify: unk ID x%x\r\n"),Did));
    return FALSE;
}

#if ALLOW_WRITE_OPERATIONS
/* Erase one sector of the chip, do not wait till done
 */
PRIVATE void SectorEraseStart(PCStrata Strata, UINT SectorStart)
{
}

/* Wait until the erase of the part is complete
 */
PRIVATE BOOL SectorEraseComplete(PCStrata Strata, UINT SectorStart)
{
    return FALSE;
}

/* Writeback the cached sector
 */
PRIVATE SCODE
FlashSectorWrite(PCStrata Strata)
{
    return E_FAIL;
}

#endif

/* Common code for read&write argument validation
 */
PRIVATE SCODE FlashValidateArguments(PCStrata pStrata,
                                     UINT64 Position,
                                     UINT *pOffset,
                                     UINT *pSize)
{
    UINT Offset;

    /* Validate args
     */
    Offset = Uint64ToUint32(Position);
    if (Offset >= pStrata->Size) {
        *pSize = 0;
        DBGME(1,_tprintf(_T("Strata::FlashValidateArg(%d) EOF\n"),Offset));
        return S_FALSE; /* EOF */
    }

    /* Adjust size
     */
    if (*pSize > (pStrata->Size - Offset))
        *pSize = pStrata->Size - Offset;

    *pOffset = Offset;

    DBGME(0,_tprintf(_T("Strata::FlashValidateArg %d %d\n"),
                     *pOffset,*pSize));
    return NO_ERROR;
}

/* Read SIZE bytes from db161 device, at offset Position
 */
PRIVATE SCODE MCT StrataReadAt (PCDEVICE this, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    PCStrata pStrata = pSt(this);
    UINT Offset;
    SCODE sc;

    DBGME(1,_tprintf(_T("StrataReadAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeRead));

    sc = FlashValidateArguments(pStrata,
                                Position,
                                &Offset,
                                &Size);

    if (sc == NO_ERROR) {
        Mutex_Lock(&pStrata->Lock);
        memcpy(pBuffer,pStrata->Base + Offset, Size);
        Mutex_Unlock(&pStrata->Lock);
    }

    if (pSizeRead)
        *pSizeRead = Size;
    return sc;
}

/* Write SIZE bytes to device.
 */
PRIVATE SCODE MCT StrataWriteAt (PCDEVICE this, UINT64 Position, const BYTE *pBuffer,
                                 UINT Size, PUINT pSizeWritten)
{
#if ALLOW_WRITE_OPERATIONS

    PCStrata pStrata = pSt(this);
    SCODE sc;
    UINT Offset, SectorStart;

    DBGME(1,_tprintf(_T("StrataWriteAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeWritten));


    sc = FlashValidateArguments(pStrata,
                                Position,
                                &Offset,
                                &Size);

    if (sc == NO_ERROR) {

        /* Are we missing in the sector cache ?
         */
        SectorStart = Offset / pStrata->SectorSize;

        if (pStrata->CachedSectorStart != SectorStart) {
            /* Yes, do we have to writeback
             */
            if (pStrata->CachedSectorStart != INVALID_SECTOR_START) {
                sc = FlashSectorWrite(pStrata);
                /* BUGBUG async error reporting !! */
            }

            /* Cache & erase the sector, in preparation for writing
             */
            memcpy(pStrata->SectorData, pStrata->Base + SectorStart, pStrata->SectorSize);
            pStrata->CachedSectorStart = SectorStart;

            /* Get the erase started
             */
            SectorEraseStart(pStrata,SectorStart);
        }

        /* Modify the cached sector data, timestamp it.
         */
        memcpy(pStrata->SectorData+Offset,pBuffer,Size);
        pStrata->LastWriteTime = CurrentTime();

        Mutex_Unlock(&pStrata->Lock);

        /* Trigger the writeback timeout
         */
        Condition_Signal(&pStrata->CDevice.IntrPoint);

    }

    if (pSizeWritten)
        *pSizeWritten = Size;
    return sc;

#else
    /* Keep flash read-only */
    UnusedParameter(this);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);

    if (pSizeWritten)
        *pSizeWritten = 0;
    return E_ACCESS_DENIED;
#endif
}

/* GetSize method on db161 device
 */
PRIVATE SCODE MCT StrataGetSize (PCDEVICE this, PUINT64 pSize)
{
    PCStrata pStrata = pSt(this);

    if (pSize)
        Uint32ToUint64(*pSize,pStrata->Size);
    return NO_ERROR;
}

#if ALLOW_WRITE_OPERATIONS

PRIVATE void StrataThreadLoop(PCDEVICE this)
{
    PCStrata pStrata = pSt(this);
    static const TIME CacheLifeTime = Int64Initializer(0,TIME_MILLIS(500));
    TIME t;

    Mutex_Lock(&pStrata->Lock);
    for (;;) {
        /* Wait till someone writes
         */
        Condition_Wait(&pStrata->CDevice.IntrPoint,
                       &pStrata->Lock,
                       NULL);

        /* Now wait for the cache lifetime
         */
        for (;;) {
            t = CurrentTime();
            t = TimeSubtract(t,pStrata->LastWriteTime);

            /* Someone refreshed/wrote the cache ?
             */
            if (TimeLess(t,CacheLifeTime)) {

                Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(500)));

                Mutex_Unlock(&pStrata->Lock);
                SleepUntil(t);
                Mutex_Lock(&pStrata->Lock);
            } else {
                break;
            }
        }

        /* Write back & invalidate the buffer, if it is still valid
         */
        if (pStrata->CachedSectorStart != INVALID_SECTOR_START) {
            (void) FlashSectorWrite(pStrata);
            pStrata->CachedSectorStart = INVALID_SECTOR_START;
        }
    }
    Mutex_Unlock(&pStrata->Lock);
}

#else
#define StrataThreadLoop NULL
#endif

struct CDeviceVtbl StrataDriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    StrataReadAt,
    StrataWriteAt,
    NULL,
    StrataGetSize,
    NULL,
    NULL,
    NULL,
    StrataThreadLoop
};

/* Constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE StrataNew(ADDRESS Base)
{
    PCStrata pStrata;
    PIHEAP pH = CurrentHeap();
    UINT32 Size, SectorSize;

    if (!FlashIdentify(Base,&Size,&SectorSize)) {
        DBGME(3,_tprintf(_T("StrataNew: unk @x%x\n"),Base));
        return NULL;
    }

    /* Allocate descriptor and, if writing, room for one sector cache
     */
    pStrata = (PCStrata) pH->v->Alloc( pH, 0, 
#if ALLOW_WRITE_OPERATIONS
                                       SectorSize + 
#endif
                                       sizeof(CStrata), 0 );
    if (pStrata) {
        pStrata->CDevice.v = &StrataDriverVtable;
        pStrata->CDevice.RefCnt = 1;
        Mutex_Init(&pStrata->Lock);
        pStrata->SectorSize = SectorSize;
#if ALLOW_WRITE_OPERATIONS
        pStrata->CachedSectorStart = INVALID_SECTOR_START;
        pStrata->LastWriteTime = CurrentTime();
        pStrata->SectorData = (UINT8*) (pStrata + 1);
#endif

        pStrata->Size = Size;
        DBGME(1,_tprintf(_T("StrataNew(%x %x)\n"),(ADDRESS)pStrata,pStrata->Size));
    }
    return (PCDEVICE)pStrata;
}

/* ** 
*/
SCODE __cdecl StrataConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Name);
    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("StrataConstructor(%x %s %x %x %x)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) StrataNew(RegisterBase);

    DBGME(0,_tprintf(_T("StrataConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("StrataConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl StrataDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) StrataConstructor
};

static struct _DRIVERCOB StrataDriverCobObject = { &StrataDriverCobVtable, 1 };

PIUNKNOWN StrataCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &StrataDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &StrataDriverVtable, _T("CDevice"),
                          &IID_IDevice);
    if (FAILED(sc))
        return NULL;

    /* Void the thread loop, to save a useless thread
     */
    StrataDriverVtable.ThreadLoop = StrataThreadLoop;

    return (PIUNKNOWN) &StrataDriverCobObject;
}

