/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Driver for a keyboard pseudo-device available for the Giano simulator.
 * Implements an IFile for keyboard input.
 */
#define DEVICE_ORDER LITTLE_ENDIAN
#include <mmlite.h>
#include <stddef.h> /* for offsetof() */
//#define _DEBUG 1
#define DBGLVL 3
#include <diagnostics.h>
#include <assert.h>
#include <drivers\drivers.h>
#include <drivers\keybd_hw.h>

/* Simple Keyboard
 * Register interface of the peripheral
 */
typedef struct {
    volatile UINT32  Status;
#define SKS_INTEN           0x00000001
#define SKS_SIGNATURE       0x474b6200    /* GKb */
#define SKS_SIGNATURE_MASK  0xffffff00

    volatile UINT32  Count;

    volatile UINT32  Data;
#define SKD_CODE_MASK       0x000000ff /* see frame.cpp */
#define SKD_KEYUP           0x00010000
#define SKD_VKEY_MASK       0xfffe0000

    volatile UINT32  Pad;
} KEYBOARD_STATE, *PKEYBOARD_STATE;

typedef struct _GIANOKBD {
    struct CDevice CDevice;
    MUTEX Lock;
    UINT Reset;
    PKEYBOARD_STATE Regs;
} GIANOKBD, *PGIANOKBD;

/* Convert between interface and instance pointer and vice versa */
#define pGianoKbd(_i_)  ((PGIANOKBD)((_i_)))
#define iGianoKbd(_p_)  ((PIDEVICE)((_p_)))

/* IFile implementation for the keyboard
 */
PRIVATE SCODE GianoKbdReadAt (
    PIDEVICE iThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PGIANOKBD This = pGianoKbd(iThis);
    PKEYBOARD_INPUT_DATA Report;
    UINT32 Key;
    SCODE sc = NO_ERROR;

    UnusedParameter(Position);

    if (pSizeRead)
        *pSizeRead = 0;

    if (Size < sizeof(KEYBOARD_INPUT_DATA))
        return E_INVALID_PARAMETER;

    /* Single-use device */
    Mutex_Lock( &This->Lock );

    while (This->Regs->Count == 0) {
        /* Wait for data
         *
         * An explicit wakeup on the condvar is required to break this.
         */
        sc = Condition_Wait( &This->CDevice.IntrPoint,NULL,NULL);

        if (This->Reset)
            break;
    }

    /* Backoff if a reset has been requested
     */
    if (This->Reset) {
        sc = E_RESTART;
        goto Done;
    }

    /* Pick up all the data that fits
     */
    for (; Size >= sizeof(KEYBOARD_INPUT_DATA); 
           Size -= sizeof(KEYBOARD_INPUT_DATA)) {

        Key = This->Regs->Data;
        Key = _bs32(Key);
        Report = (PKEYBOARD_INPUT_DATA)pBuffer;
        Report->ReportTime = CurrentTime();
        Report->MakeCode = (UINT16) (Key & 0xffff); /* scancode */
        if (Key & 0x10000)
            Report->Flags = KEYBOARD_FLAGS_UP_STROKE;
        else
            Report->Flags = KEYBOARD_FLAGS_DOWN_STROKE;

        if (pSizeRead)
            *pSizeRead += sizeof(KEYBOARD_INPUT_DATA);

        /* If we have picked up the last one re-enable interrupts
         * NB: Interrupt is level-triggered
         */
        if (This->Regs->Count == 0) {
            This->Regs->Status = _bs32(SKS_INTEN);
            break;
        }
    }

 Done:
    Mutex_Unlock( &This->Lock );

    return sc;
}

PRIVATE SCODE GianoKbdGetSize (
    PIDEVICE iThis,
    /*out*/ UINT64 * Size)
{
    UnusedParameter(iThis);

    Int64FromHighAndLow(*Size,0,sizeof(KEYBOARD_INPUT_DATA));
    return S_OK;
}

PRIVATE BOOL GianoKbdIsr (
    PIDEVICE iThis,
    /*out*/ BOOL * pNotOurInterrupt)
{
    PGIANOKBD This = pGianoKbd(iThis);

    UnusedParameter(pNotOurInterrupt);

    /* Disable interrupts for now */
    This->Regs->Status = 0;

    return Condition_InterruptSignal(&This->CDevice.IntrPoint);
}

PRIVATE INT GianoKbdInitialize (
    PIDEVICE iThis)
{
    PGIANOKBD This = pGianoKbd(iThis);

    /* Flag that reset is pending.  This will abort any reading that is
     * in progress via ReadAt.
     */
    AtomicInc(&This->Reset);

    /* Wakeup any reader
     */
    Condition_Signal(&This->CDevice.IntrPoint);

    /* Wait for any reading threads to get out.  This gives us exclusive
     * access to the hardware.
     */
    Mutex_Lock(&This->Lock);

    /* Stop it dead
     */
    This->Regs->Status = 0;
    /* Umbug.. should we clear up the queue? */

    /* We're done.
     */
    AtomicDec(&This->Reset);

    /* Enable interrupts */
    This->Regs->Status = _bs32(SKS_INTEN);

    /* Let other threads back in.
     */
    Mutex_Unlock( &This->Lock );
    
    return 0;
}
/* Auto-generated Vtables */
#include "s_GianoKbd_v.c"

/* Constructor */
PIUNKNOWN GianoKbdNew(
    ADDRESS BaseAddress)
{
    PKEYBOARD_STATE Regs = (PKEYBOARD_STATE) BaseAddress;
    UINT32 Key;

    /* We can only do one (although..)
     */
    static struct _GIANOKBD TheKbd = {
        {0,},
    };

    /* Sanity check the signature
     */
    Key = Regs->Status;
    if ((_bs32(Key) & SKS_SIGNATURE_MASK) != SKS_SIGNATURE)
        return NULL;

    TheKbd.CDevice.v = (struct CDeviceVtbl *) &GianoKbdVtbl;
    TheKbd.CDevice.RefCnt = 1;
    Condition_Init(&TheKbd.CDevice.IntrPoint);

    Mutex_Init(&TheKbd.Lock);
    TheKbd.Reset = 0;
    TheKbd.Regs = Regs;

    return (PIUNKNOWN) iGianoKbd(&TheKbd);
}
