/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Driver for a mouse pseudo-device available for the Giano simulator.
 * Implements an IFile for mouse input.
 */
#define DEVICE_ORDER LITTLE_ENDIAN
#include <mmlite.h>
#include <stddef.h> /* for offsetof() */
//#define _DEBUG 1
#define DBGLVL 3
#include <diagnostics.h>
#include <assert.h>
#include <drivers\drivers.h>
#include <drivers\mouse_hw.h>

/* Simple Mouse
 * Register interface of the peripheral
 */
typedef struct {
    UINT32  Status;
#define SKS_INTEN           0x00000001
#define SKS_SIGNATURE       0x474d7300    /* GMs */
#define SKS_SIGNATURE_MASK  0xffffff00

    UINT32  Count;

    UINT32  Buttons;
#define MK_LBUTTON          0x0001
#define MK_RBUTTON          0x0002
#define MK_SHIFT            0x0004
#define MK_CONTROL          0x0008
#define MK_MBUTTON          0x0010
#define MK_XBUTTON1         0x0020
#define MK_XBUTTON2         0x0040
    UINT32  Coordinates;
} MOUSE_STATE, *PMOUSE_STATE;

typedef struct _GIANOMS {
    struct CDevice CDevice;
    MUTEX Lock;
    UINT Reset;
    PMOUSE_STATE Regs;
} GIANOMS, *PGIANOMS;

/* Convert between interface and instance pointer and vice versa */
#define pGianoMs(_i_)  ((PGIANOMS)((_i_)))
#define iGianoMs(_p_)  ((PIDEVICE)((_p_)))

/* IFile implementation for the mouse
 */
PRIVATE SCODE GianoMsReadAt (
    PIDEVICE iThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PGIANOMS This = pGianoMs(iThis);
    PMOUSE_INPUT_DATA Report;
    SCODE sc = NO_ERROR;

    UnusedParameter(Position);

    if (pSizeRead)
        *pSizeRead = 0;

    if (Size < sizeof(MOUSE_INPUT_DATA))
        return E_INVALID_PARAMETER;

    /* Single-use device */
    Mutex_Lock( &This->Lock );

    while (This->Regs->Count == 0) {
        /* Wait for data
         *
         * An explicit wakeup on the condvar is required to break this.
         */
        sc = Condition_Wait( &This->CDevice.IntrPoint,NULL,NULL);

        if (This->Reset)
            break;
    }

    /* Backoff if a reset has been requested
     */
    if (This->Reset) {
        sc = E_RESTART;
        goto Done;
    }

    /* Pick up all the data that fits
     */
    for (Report = (PMOUSE_INPUT_DATA)pBuffer;
         Size >= sizeof(MOUSE_INPUT_DATA); 
         Size -= sizeof(MOUSE_INPUT_DATA)) {
        UINT32 Key;

        Report->ReportTime = CurrentTime();

        Key = This->Regs->Buttons;
        Key = _bs32(Key);
        Report->RawButtons = 0;
        if (Key & MK_LBUTTON) Report->RawButtons |= MOUSE_LEFT_BUTTON_DOWN;
        if (Key & MK_MBUTTON) Report->RawButtons |= MOUSE_MIDDLE_BUTTON_DOWN;
        if (Key & MK_RBUTTON) Report->RawButtons |= MOUSE_RIGHT_BUTTON_DOWN;

        /* NB: This pops the FIFO */
        Key = This->Regs->Coordinates;
        Key = _bs32(Key);
        Report->LastX = (UINT16)(Key);
        Report->LastY = (UINT16)(Key >> 16);
        Report->LastZ = 0;

        Report++;
        if (pSizeRead)
            *pSizeRead += sizeof(MOUSE_INPUT_DATA);

        /* If we have picked up the last one re-enable interrupts
         * NB: Interrupt is level-triggered
         */
        if (This->Regs->Count == 0) {
            This->Regs->Status = _bs32(SKS_INTEN);
            break;
        }
    }

 Done:
    Mutex_Unlock( &This->Lock );

    return sc;
}

PRIVATE SCODE GianoMsGetSize (
    PIDEVICE iThis,
    /*out*/ UINT64 * Size)
{
    UnusedParameter(iThis);

    Int64FromHighAndLow(*Size,0,sizeof(MOUSE_INPUT_DATA));
    return S_OK;
}

PRIVATE BOOL GianoMsIsr (
    PIDEVICE iThis,
    /*out*/ BOOL * pNotOurInterrupt)
{
    PGIANOMS This = pGianoMs(iThis);

    UnusedParameter(pNotOurInterrupt);

    /* Disable interrupts for now */
    This->Regs->Status = 0;

    return Condition_InterruptSignal(&This->CDevice.IntrPoint);
}

PRIVATE INT GianoMsInitialize (
    PIDEVICE iThis)
{
    PGIANOMS This = pGianoMs(iThis);

    /* Flag that reset is pending.  This will abort any reading that is
     * in progress via ReadAt.
     */
    AtomicInc(&This->Reset);

    /* Wakeup any reader
     */
    Condition_Signal(&This->CDevice.IntrPoint);

    /* Wait for any reading threads to get out.  This gives us exclusive
     * access to the hardware.
     */
    Mutex_Lock(&This->Lock);

    /* Stop it dead
     */
    This->Regs->Status = 0;
    /* Umbug.. should we clear up the queue? */

    /* We're done.
     */
    AtomicDec(&This->Reset);

    /* Enable interrupts */
    This->Regs->Status = _bs32(SKS_INTEN);

    /* Let other threads back in.
     */
    Mutex_Unlock( &This->Lock );
    
    return 0;
}
/* Auto-generated Vtables */
#include "s_GianoMs_v.c"

/* Constructor */
PIUNKNOWN GianoMsNew(
    ADDRESS BaseAddress)
{
    PMOUSE_STATE Regs = (PMOUSE_STATE) BaseAddress;
    UINT32 Key;

    /* We can only do one (although..)
     */
    static struct _GIANOMS TheMs = {
        {0,},
    };

    /* Sanity check the signature
     */
    Key = Regs->Status;
    if ((_bs32(Key) & SKS_SIGNATURE_MASK) != SKS_SIGNATURE)
        return NULL;

    TheMs.CDevice.v = (struct CDeviceVtbl *) &GianoMsVtbl;
    TheMs.CDevice.RefCnt = 1;
    Condition_Init(&TheMs.CDevice.IntrPoint);

    Mutex_Init(&TheMs.Lock);
    TheMs.Reset = 0;
    TheMs.Regs = Regs;

    return (PIUNKNOWN) iGianoMs(&TheMs);
}
