/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*++

Module Name:

    Read.c

Abstract:


Author:


Environment:

    Kernel mode only.

Notes:


Future:



Revision History:

    Updated for Windows 2000 - Eliyas Yakub June, 1999

--*/

#include "ntddk.h"
#include "ndis.h"
#include "packet.h"


NTSTATUS
PacketRead(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )

/*++

Routine Description:

    This is the dispatch routine for read requests.

Arguments:

    DeviceObject - Pointer to the device object.

    Irp - Pointer to the request packet.

Return Value:

    NT status code.
--*/

{
    POPEN_INSTANCE      open;
    PNDIS_PACKET        pPacket;
    NDIS_STATUS         status;
    NTSTATUS            ntStatus;
    PIO_STACK_LOCATION  irpSp;

//    DebugPrint(("Read\n"));

    open = DeviceObject->DeviceExtension;

    //
    // Check to see whether you are still bound to the adapter
    //

    IoIncrement(open);
    
    if(!open->Bound)
    {
        ntStatus = STATUS_DEVICE_NOT_READY;
        goto ERROR;
    }
    
    irpSp = IoGetCurrentIrpStackLocation(Irp);

    //
    //  See if the buffer is at least big enough to hold the
    //  ethernet header
    //
    if (irpSp->Parameters.Read.Length < ETHERNET_HEADER_LENGTH) {

        ntStatus = STATUS_BUFFER_TOO_SMALL;
        goto ERROR;
    }

    //
    // Try to get a packet from our list of free ones. 
    //

    NdisAllocatePacket(
        &status,
        &pPacket,
        open->PacketPool
        );

    if (status != NDIS_STATUS_SUCCESS) {
        DebugPrint(("Packet: Read- No free packets\n"));

        ntStatus = STATUS_INSUFFICIENT_RESOURCES;
        goto ERROR;
    }

    //
    //  Get a pointer to the packet itself.
    //

    RESERVED(pPacket)->Irp=Irp;
    RESERVED(pPacket)->pMdl=NULL;

    IoMarkIrpPending(Irp);

    //
    // Set the cancel routine so that we can cancel the requests
    // pending in the queue if the application terminates.
    //
    
    IoSetCancelRoutine(Irp, PacketCancelRoutine);

    //
    //  Put this packet in a list of pending reads.
    //  The PacketReceiveIndicate or PacketReceivePacket handler will 
    //  attempt to remove packets from this list.
    //
    
    ExInterlockedInsertTailList(
        &open->RcvList,
        &RESERVED(pPacket)->ListElement,
        &open->RcvQSpinLock);

  
    return STATUS_PENDING;
    
ERROR:
    Irp->IoStatus.Status = ntStatus;
    IoCompleteRequest (Irp, IO_NO_INCREMENT);
    IoDecrement(open);
    return ntStatus;

}

#define FILTER_PACKETS  1
#define FILTER_NON_ETHERNET_PACKETS 0
#define FILTER_ETHERNET_PACKETS 1

#if FILTER_PACKETS

#pragma pack(1)

typedef struct _ETHERNET
{
    unsigned char   Dst[6];
    unsigned char   Src[6];
    USHORT          Type;
} ETHERNET, *PETHERNET;

#pragma pack()

#define LOBYTE(w)       ((w) & 0x00FF)
#define HIBYTE(w)       ((w) >> 8)
#define SWAPWORD(w)     ((LOBYTE(w) << 8) | HIBYTE(w))

#define ETHERNET_802_3_PACKET   0x0600
#define ETHERTYPE_IP            0x0800
#define ETHERTYPE_ARP           0x0806

#endif


//
// True if the promiscuous bit is set.
//
#define FILTER_IS_PROMISCUOUS(filter) \
    (NDIS_PACKET_TYPE_PROMISCUOUS & (filter))


//
// True if any multicast address is accepted.
// This behaves like the NDIS_PACKET_TYPE_ALL_MULTICAST bit is set,
// even though the client may intend to use the multicast address list.
// (Hack.)
//
#define FILTER_ALLOWS_MULTICAST(filter)                 \
    (                                                   \
     (NDIS_PACKET_TYPE_ALL_MULTICAST & (filter)) ||     \
     (NDIS_PACKET_TYPE_MULTICAST & (filter))            \
    )


// 
// True if the broadcast bit is set (correct implementation).
//
#define FILTER_ALLOWS_BROADCAST(filter)     \
    (NDIS_PACKET_TYPE_BROADCAST & (filter))


//
// True if the address is a link-level multicast address, 
// which requires the lowest bit to be set.
//
#define MAC_ADDRESS_IS_MULTICAST(address)   \
    (((UCHAR)0x1) & (address)[5])


//
// True if the address is the broadcast address, with all bits set.
//
#define MAC_ADDRESS_IS_BROADCAST(address)   \
    ((0xFF & (address)[0]) &&               \
     (0xFF & (address)[1]) &&               \
     (0xFF & (address)[2]) &&               \
     (0xFF & (address)[3]) &&               \
     (0xFF & (address)[4]) &&               \
     (0xFF & (address)[5]))


NDIS_STATUS
PacketFilterEthernet(
    IN  POPEN_INSTANCE      Open,
    IN  PETHERNET           Ethernet
    )
{
    USHORT          eType;
    ULONG           filter;

    eType    = SWAPWORD(Ethernet->Type);

#if FILTER_NON_ETHERNET_PACKETS
    //
    //  If this is not an ethernet packet (i.e. an 802.3 packet) then
    //  drop it on the floor. Note: 0x600 is the first ETYPE defined 
    //  and it is slightly larger then the max MTU size of 1500 bytes.
    //

    if ( eType < ETHERNET_802_3_PACKET )
    {
        //
        //  It's an 802.3 packet, pitch it.
        //

        return NDIS_STATUS_NOT_ACCEPTED;
    }
#endif

#if FILTER_ETHERNET_PACKETS
    //
    // Apply the virtual packet filter that replaces the filter in 
    // the physical NIC. (The physical NIC must run in promiscuous
    // mode...)
    //
    // This is a partial implementation - there is no multicast 
    // address list, etc. 
    //

    filter = Open->VirtualNicFilter;

    if (! FILTER_IS_PROMISCUOUS(filter))
    {
        if (MAC_ADDRESS_IS_MULTICAST(Ethernet->Dst))
        {
            //
            // Let multicast packet pass if allowed.
            //
            return FILTER_ALLOWS_MULTICAST(filter)
                ? NDIS_STATUS_SUCCESS
                : NDIS_STATUS_NOT_ACCEPTED;
        }
        else if (MAC_ADDRESS_IS_BROADCAST(Ethernet->Dst))
        {
            //
            // Let broadcast packets pass if allowed.
            //
            return FILTER_ALLOWS_BROADCAST(filter)
                ? NDIS_STATUS_SUCCESS
                : NDIS_STATUS_NOT_ACCEPTED;
        }
        else 
        {
            //
            // The destination address is a unicast address. 
            // If it doesn't match the virtual MAC address, 
            // then drop the packet.
            //

            INT i = 0;

            for (i = 0; i < 6; ++i)
            {
                if (Ethernet->Dst[i] != Open->VirtualMacAddress[i])
                {
                    return NDIS_STATUS_NOT_ACCEPTED;
                }
            }

            //
            // Let unicast packet directed at the virtual MAC address pass.
            //
            return NDIS_STATUS_SUCCESS;
        }
    }
#endif

    //
    // Indiscriminately let everything that gets here pass.
    //
    return NDIS_STATUS_SUCCESS;
}


NDIS_STATUS
PacketFilterPacket(
    IN    POPEN_INSTANCE    Open,
    IN    PNDIS_PACKET      Packet
    )
{
#if FILTER_PACKETS
    PVOID           virtualAddress;
    PNDIS_BUFFER    firstBuffer, nextBuffer;
    ULONG           totalLength;

    NdisQueryPacket(
                Packet, 
                NULL,
                NULL, 
                &firstBuffer, 
                &totalLength
                );

    NdisQueryBufferSafe(
                firstBuffer, 
                &virtualAddress, 
                &totalLength, 
                NormalPagePriority
                );

    return PacketFilterEthernet(Open, (PETHERNET) virtualAddress);
#else
    return NDIS_STATUS_SUCCESS;
#endif
}


NDIS_STATUS
PacketReceiveIndicate (
    IN NDIS_HANDLE ProtocolBindingContext,
    IN NDIS_HANDLE MacReceiveContext,
    IN PVOID       HeaderBuffer,
    IN UINT        HeaderBufferSize,
    IN PVOID       LookAheadBuffer,
    IN UINT        LookaheadBufferSize,
    IN UINT        PacketSize
    )
/*++

Routine Description:

    PacketReceiveIndicate is called with a pointer to a 
    lookahead buffer. If this buffer contains less than 
    the full, received network packet, it 
    calls NdisTransferData with a protocol-allocated packet 
    descriptor specifying protocol-allocated buffer(s) to obtain 
    the remainder of the received packet. 

Arguments:


Return Value:


--*/
{
    POPEN_INSTANCE      open;
    PIO_STACK_LOCATION  irpSp;
    PIRP                irp;
    PLIST_ENTRY         packetListEntry;
    PNDIS_PACKET        pPacket;
    ULONG               sizeToTransfer;
    NDIS_STATUS         status;
    UINT                bytesTransfered = 0;
    ULONG               bufferLength;
    PPACKET_RESERVED    reserved;
    PMDL                pMdl;

//    DebugPrint(("-->ReceiveIndicate\n"));

    open= (POPEN_INSTANCE)ProtocolBindingContext;

    if (HeaderBufferSize > ETHERNET_HEADER_LENGTH) {

        return NDIS_STATUS_SUCCESS;
    }

    //
    // Throw away packets we don't want.
    //

    if (NDIS_STATUS_NOT_ACCEPTED == PacketFilterEthernet(open, (PETHERNET) HeaderBuffer))
    {
        return NDIS_STATUS_NOT_ACCEPTED;
    }

    //
    //  See if there are any pending read that we can satisfy
    //
    packetListEntry=ExInterlockedRemoveHeadList(
                        &open->RcvList,
                        &open->RcvQSpinLock
                        );

    if (packetListEntry == NULL) {
//        DebugPrint(("No pending read, dropping packets\n"));
        return NDIS_STATUS_NOT_ACCEPTED;
    }


    reserved=CONTAINING_RECORD(packetListEntry,PACKET_RESERVED,ListElement);
    pPacket=CONTAINING_RECORD(reserved,NDIS_PACKET,ProtocolReserved);

    irp=RESERVED(pPacket)->Irp;
    irpSp = IoGetCurrentIrpStackLocation(irp);

    //
    // We don't have to worry about the situation where the IRP is cancelled
    // after we remove it from the queue and before we reset the cancel
    // routine because the cancel routine has been coded to cancel an IRP
    // only if it's in the queue.
    //
    
    IoSetCancelRoutine(irp, NULL);
    
    //
    //  This is the length of our partial MDL
    //
    bufferLength=irpSp->Parameters.Read.Length-ETHERNET_HEADER_LENGTH;

    //
    //  Find out how much to transfer
    //
    sizeToTransfer = (PacketSize < bufferLength) ?
                       PacketSize : bufferLength;

    //
    //  copy the ethernet header into the actual readbuffer
    //
    NdisMoveMappedMemory(
        MmGetSystemAddressForMdlSafe(irp->MdlAddress, NormalPagePriority),
        HeaderBuffer,
        HeaderBufferSize
        );

    //
    //  Allocate an MDL to map the portion of the buffer following the
    //  header
    //
    pMdl=IoAllocateMdl(
              MmGetMdlVirtualAddress(irp->MdlAddress),
              MmGetMdlByteCount(irp->MdlAddress),
              FALSE,
              FALSE,
              NULL
              );


    if (pMdl == NULL) {
        DebugPrint(("Packet: Read-Failed to allocate Mdl\n"));
        status = NDIS_STATUS_RESOURCES;
        goto ERROR;
    }

    //
    //  Build the mdl to point to the the portion of the buffer following
    //  the header
    //
    IoBuildPartialMdl(
        irp->MdlAddress,
        pMdl,
        ((PUCHAR)MmGetMdlVirtualAddress(irp->MdlAddress))+ETHERNET_HEADER_LENGTH,
        0
        );

    //
    //  Clear the next link in the new MDL
    //

    pMdl->Next=NULL;

    RESERVED(pPacket)->pMdl=pMdl;

    //
    //  Attach our partial MDL to the packet
    //

    NdisChainBufferAtFront(pPacket,pMdl);

    //
    //  Call the Mac to transfer the packet
    //

    NdisTransferData(
        &status,
        open->AdapterHandle,
        MacReceiveContext,
        0,
        sizeToTransfer,
        pPacket,
        &bytesTransfered);
    if (status == NDIS_STATUS_PENDING) {
        return NDIS_STATUS_SUCCESS;
    }

ERROR:
    
    //
    //  If it didn't pend, call the completeion routine now
    //
    PacketTransferDataComplete(
                                open,
                                pPacket,
                                status,
                                bytesTransfered);
    return NDIS_STATUS_SUCCESS;

}



VOID
PacketTransferDataComplete (
    IN NDIS_HANDLE   ProtocolBindingContext,
    IN PNDIS_PACKET  pPacket,
    IN NDIS_STATUS   Status,
    IN UINT          BytesTransfered
    )
/*++

Routine Description:

    This is a required function unless the protocol binds itself 
    exclusively to underlying NIC driver(s) that indicate packets 
    with NdisMIndicateReceivePacket. ProtocolTransferDataComplete 
    is called when a previous call to NdisTransferData returned 
    NDIS_STATUS_PENDING and the remaining data has been copied 
    into the protocol-supplied buffers chained to a given packet 
    descriptor. 

Arguments:


Return Value:


--*/

{
    PIO_STACK_LOCATION   irpSp;
    POPEN_INSTANCE       open;
    PIRP                 irp;
    PMDL                 pMdl;

    DebugPrint(("Packet: TransferDataComplete\n"));

    open= (POPEN_INSTANCE)ProtocolBindingContext;

    irp=RESERVED(pPacket)->Irp;

    irpSp = IoGetCurrentIrpStackLocation(irp);

    pMdl=RESERVED(pPacket)->pMdl;

    
    //
    //  Free the partial MDL that we allocated
    //
    if(pMdl)
        IoFreeMdl(pMdl);

    //
    //  Put the packet on the free queue
    //
    NdisFreePacket(pPacket);

    if(Status == NDIS_STATUS_SUCCESS)
    {
        irp->IoStatus.Status = STATUS_SUCCESS;
        irp->IoStatus.Information = BytesTransfered+ETHERNET_HEADER_LENGTH;
    } else {
    
        irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
        irp->IoStatus.Information = 0;
    }

    DebugPrint(("BytesTransfered:%d\n", irp->IoStatus.Information));

    IoCompleteRequest(irp, IO_NO_INCREMENT);

    IoDecrement(open);
    return;


}

VOID
PacketReceiveComplete(
    IN NDIS_HANDLE  ProtocolBindingContext
    )
/*++

Routine Description:

    This is a required function. PacketReceiveComplete 
    is called to indicate that any received packets previously 
    indicated to PacketReceivePacket can now be postprocessed. 
    
Arguments:


Return Value:


--*/
{
    return;
}


INT
PacketReceivePacket(
    IN    NDIS_HANDLE         ProtocolBindingContext,
    IN    PNDIS_PACKET        Packet
    )
/*++

Routine Description:

    ReceivePacket handler. Called up by the miniport below 
    when it supports NDIS 4.0 style receives.
    A ProtocolReceivePacket function should be provided if 
    the protocol driver might be bound to a NIC driver that 
    indicates an array of one or more packets by calling 
    NdisMIndicateReceivePacket. 

Arguments:

    ProtocolBindingContext    Pointer to our adapter structure.
    Packet - Pointer to the packet

Return Value:

    == 0 -> We are done with the packet
    != 0 -> We will keep the packet and call NdisReturnPackets() 
            this many times when done.
--*/
{
    UINT                bytesTransfered = 0;
    POPEN_INSTANCE      open;
    PIRP                irp;
    PNDIS_PACKET        myPacket;
    PLIST_ENTRY         packetListEntry;
    ULONG               bufferLength;
    PPACKET_RESERVED    reserved;
    PIO_STACK_LOCATION  irpSp;
    PMDL                mdl;
    NTSTATUS            status = STATUS_SUCCESS;

    DebugPrint(("PacketReceivePacket\n"));

    open= (POPEN_INSTANCE)ProtocolBindingContext;

    //
    // Throw away packets we don't want.
    //

    if (NDIS_STATUS_NOT_ACCEPTED == PacketFilterPacket(open, Packet))
    {
        return NDIS_STATUS_NOT_ACCEPTED;
    }

    //
    //  See if there are any pending read that we can satisfy
    //
    
    packetListEntry=ExInterlockedRemoveHeadList(
                        &open->RcvList,
                        &open->RcvQSpinLock
                        );

    if (packetListEntry == NULL) {
        //
        // No pending reads so just the dropping the packet
        // on the floor.
        //
//        DebugPrint(("No pending read, dropping packets\n"));
        return 0;
    }


    reserved=CONTAINING_RECORD(packetListEntry,PACKET_RESERVED,ListElement);
    myPacket=CONTAINING_RECORD(reserved,NDIS_PACKET,ProtocolReserved);

    irp = RESERVED(myPacket)->Irp;
    irpSp = IoGetCurrentIrpStackLocation(irp);

    //
    // We don't have to worry about the situation where the IRP is cancelled
    // after we remove it from the queue and before we reset the cancel
    // routine because the cancel routine has been coded to cancel an IRP
    // only if it's in the queue.
    //

    IoSetCancelRoutine(irp, NULL);

    //
    // Following block of code locks the destination packet
    // MDLs in a safe manner. This is a temporary workaround
    // for NdisCopyFromPacketToPacket that currently doesn't use
    // safe functions to lock pages of MDL. This is required to
    // prevent system from bugchecking under low memory resources.
    //
    {
        PVOID           virtualAddress;
        PNDIS_BUFFER    firstBuffer, nextBuffer;
        ULONG           totalLength;

        NdisQueryPacket(Packet, NULL, NULL, &firstBuffer, &totalLength);
        while( firstBuffer != NULL)
        {
            NdisQueryBufferSafe(firstBuffer, &virtualAddress, 
                                    &totalLength, NormalPagePriority );
            if(!virtualAddress) {
                //
                // System is running low on memory resources. 
                // So fail the read.
                // 
                status = STATUS_INSUFFICIENT_RESOURCES;
                goto CleanExit;
            }
            NdisGetNextBuffer(firstBuffer,  &nextBuffer);
            firstBuffer = nextBuffer;
        }
    }
    
    //
    //
    //  Attach the actual MDL to the packet
    //

    NdisChainBufferAtFront(myPacket, irp->MdlAddress);  

    bufferLength=irpSp->Parameters.Read.Length;

    NdisCopyFromPacketToPacket(myPacket, 0, bufferLength, Packet, 0, 
                                               &bytesTransfered);    

CleanExit:
    //                                 
    //  Put the packet on the free queue
    //

    NdisFreePacket(myPacket);

    irp->IoStatus.Status = status;
    irp->IoStatus.Information = bytesTransfered;
    IoCompleteRequest(irp, IO_NO_INCREMENT);
    DebugPrint(("BytesTransfered:%d\n", bytesTransfered));
    IoDecrement(open);

    return 0;
}


VOID
PacketCancelRoutine (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PIRP             Irp
    )

/*++

Routine Description:

    The cancel routine. It will remove the IRP from the queue 
    and will complete it. The cancel spin lock is already acquired 
    when this routine is called.

Arguments:

    DeviceObject - pointer to the device object.
    
    Irp - pointer to the IRP to be cancelled.
    
    
Return Value:

    VOID.

--*/
{

    POPEN_INSTANCE      open = DeviceObject->DeviceExtension; 
    KIRQL               oldIrql;
    PIRP                irpToComplete = NULL;
    PLIST_ENTRY         thisEntry, listHead;
    PIRP                pendingIrp;
    PNDIS_PACKET        myPacket = NULL;
    PPACKET_RESERVED    reserved;
    PMDL                mdl;

    //
    // Don't assume that the IRP being cancelled is in the queue.
    // Only complete the IRP if it IS in the queue.
    //
    //
    // Must acquire the local spinlock before releasing 
    // the global cancel spinlock
    //
    DebugPrint(("PacketCancelRoutine\n"));

    oldIrql = Irp->CancelIrql;

    //
    // One should not intermix KeAcquireSpinLock(AtDpcLevel) 
    // and ExInterlocked...List() functions on the same spinlock if the
    // routines that use the lock run at IRQL > DISPATCH_LEVEL.
    // After acquiring the lock using Ke function, if we got interrupted
    // and entered into an ISR and tried to manipulate the list using
    // ExInterlocked...List function with the same lock, we deadlock.
    // In this sample we can safely do that because none of our routines 
    // will be called at IRQL > DISPATCH_LEVEL.
    //
    
    KeAcquireSpinLockAtDpcLevel(&open->RcvQSpinLock);

    IoReleaseCancelSpinLock( KeGetCurrentIrql() );

    //
    // Remove the IRP from the queue
    //
  
    listHead = &open->RcvList;
    
    for(thisEntry = listHead->Flink; 
        thisEntry != listHead;
        thisEntry = thisEntry->Flink)
    {

        reserved=CONTAINING_RECORD(thisEntry,PACKET_RESERVED,ListElement);
        myPacket=CONTAINING_RECORD(reserved,NDIS_PACKET,ProtocolReserved);
        pendingIrp = RESERVED(myPacket)->Irp;

        if (pendingIrp == Irp) 
        {
                RemoveEntryList(thisEntry);
                irpToComplete = pendingIrp;
                break;
        }
      
    }
    
    //
    // Release the queue spinlock
    //

    KeReleaseSpinLock(&open->RcvQSpinLock, oldIrql);

    //
    // Complete the IRP with status canclled
    //
    
    if(irpToComplete) {
    
        DebugPrint(("Cancelling IRP\n"));

        ASSERT(myPacket);       

        //
        //  Put the packet on the free queue
        //
        NdisFreePacket(myPacket);

        irpToComplete->IoStatus.Status = STATUS_CANCELLED;
        irpToComplete->IoStatus.Information = 0;
        IoCompleteRequest(irpToComplete, IO_NO_INCREMENT);
        IoDecrement(open);
    }

    return;
 
}


NTSTATUS
PacketCancelReadIrps(
    IN PDEVICE_OBJECT DeviceObject
)
{
    POPEN_INSTANCE      open = DeviceObject->DeviceExtension; 
    PLIST_ENTRY         thisEntry;
    PIRP                pendingIrp;
    PNDIS_PACKET        myPacket = NULL;
    PPACKET_RESERVED    reserved;
    PMDL                mdl;

    DebugPrint(("PacketCancelReadIrps\n"));

    //
    // Walk through the RcvList and cancel all read IRPs.
    //

    while(thisEntry=ExInterlockedRemoveHeadList(
                        &open->RcvList,
                        &open->RcvQSpinLock
                        ))
    {
        reserved=CONTAINING_RECORD(thisEntry,PACKET_RESERVED,ListElement);
        myPacket=CONTAINING_RECORD(reserved,NDIS_PACKET,ProtocolReserved);

        ASSERT(myPacket);

        pendingIrp = RESERVED(myPacket)->Irp;

        //
        //  Put the packet on the free queue
        //
        NdisFreePacket(myPacket);
        
        DebugPrint(("Cancelled : 0%0x\n", pendingIrp));
        
        IoSetCancelRoutine(pendingIrp, NULL);

        //
        // Cancel the Irp
        //
        pendingIrp->IoStatus.Information = 0;
        pendingIrp->IoStatus.Status = STATUS_CANCELLED;
        IoCompleteRequest(pendingIrp, IO_NO_INCREMENT);
        IoDecrement(open);
    }

    return STATUS_SUCCESS;
}

