/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Notes
 * - specs pag 14: IOCHRDY, and nSBHE (reset!)
 *   - try writing 1 byte to PPTR+1
 * - dump EEPROM to see how its inited
 */
#define USE_INTRINSICS
#include <mmlite.h>
#include <mmmacros.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <sock.h>
#include <if.h>
#include <if_types.h>
#include <ip/in.h>
#include <ip/arp.h>

#include <drivers/drivers.h>

#include <cs8900.h>

#define LOGIT 0
#if LOGIT
#define LOGSIZE 128
UINT LogP = 0;
struct {
    UINT Key;
    UINT Info;
    TIME When;
} LogBuf[LOGSIZE];
char *LogNames[] = {
/* 0*/    "unused",
/* 1*/    "rcv St=",
/* 2*/    "erc Ln=",
/* 3*/    "drp Ln=",
/* 4*/    "xmt fl=",
/* 5*/    "bid ln=",
/* 6*/    "later  ",
/* 7*/    "mbf ln=",
/* 8*/    "out Ln=",
/* 9*/    "xin St=",
};

void LogIt(UINT Key, UINT Info)
{
    TIME Now = CurrentTime();
    INT i;

    for (;;) {
        /* Disabled ?
         */
        if (LogP == ~0)
            return;
        /* Get a number, make sure its good.
         */
        i = AtomicInc(&LogP) - 1;
        if (i >= LOGSIZE) {
            /* Roundabout, with care
             */
            if (!AtomicCmpAndSwap(&LogP,
                                  i+1,
                                  0))
                DebugBreak();
        } else
            /* Good one.
             */
            break;
    }
    LogBuf[i].When = Now;
    LogBuf[i].Key = Key;
    LogBuf[i].Info  = Info;
#if 0
    printf("LOG %hs %08x %ld\n",
           LogNames[Key],
           LogBuf[i].Info,
           LogBuf[i].When);

#endif
}

void PrintZeLog(void)
{
    UINT index, LogIs;
    UINT key;
    TIME Now = CurrentTime(), t;

    /* Block the log while we print it
     */
    for (;;) {
        LogIs = LogP;
        if (AtomicCmpAndSwap(&LogP,
                             LogIs,
                             ~0))
            break;
    }
    index = LogIs;
    Int32ToInt64(t,0);
    do {

        if (++index >= LOGSIZE) index = 0;

        key = LogBuf[index].Key;
        if (key == 0)
            continue;

        printf("%s %08x  %ld (+%ld)\n",
               LogNames[key],
               LogBuf[index].Info,
               LogBuf[index].When,
               TimeSubtract(LogBuf[index].When,t));

        t = LogBuf[index].When;

    } while (index != LogIs);
    printf(" now  %ld (+%ld)\n", Now, TimeSubtract(Now, t) );
    /* Let the log go
     */
    LogP = LogIs;
}
#else
#define LogIt(x,y)
#endif

typedef struct _CS8900 { /* Inherits from CDevice */

    struct arpcom ac;            /* standard ether interface header */
#define is_if ac.ac_if
#define is_addr ac.ac_enaddr
#define CollisionCount ac.ac_if.if_data.ifi_collisions
#define MissedPackets ac.ac_if.if_data.ifi_errors[IFI_ERR_MISS]
#define CompleteResets ac.ac_if.if_data.ifi_errors[IFI_ERR_RESET]
#define XmitTimeouts ac.ac_if.if_data.ifi_errors[IFI_ERR_XTIM]
#define MemoryErrors ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM]

    /* Driver's private data
     */
    IO_PORT IoBase;       /* Passed to driver                     */
    UINT16  ISQHead;      /* ISR pulls first entry from ISQ       */
    BOOL    TxBidding;    /* True is waiting for xmit allocation  */
    UINT    MissOverflow; /* # of overflows of the 12 bit counter */
    UINT    CollOverflow; /* # of overflows of the 12 bit counter */
    MUTEX   Lock;

} CS8900, *PCS8900;

/* Macros to access the chip, in IO mode
 * BUGBUG Should go somewhere else for portability
 */
#define CsRegLock(_cs_) \
    Mutex_Lock(&(_cs_)->Lock)

#define CsRegUnlock(_cs_) \
    Mutex_Unlock(&(_cs_)->Lock)

#define CsGetPPMem(_base_,_offset_,_reg_) { \
    IOSpaceWriteInt16((IO_PORT)(_base_+CS_IO_PP_PTR),(UINT16)_offset_); \
    _reg_ = IOSpaceReadInt16((IO_PORT)(_base_+CS_IO_PP_DATA)); \
}

#define CsGetReg(_base_,_id_,_reg_) \
    CsGetPPMem(_base_,_id_,_reg_)

#define CsSetReg(_base_,_id_,_reg_) { \
    IOSpaceWriteInt16((IO_PORT)(_base_+CS_IO_PP_PTR),_id_); \
    IOSpaceWriteInt16((IO_PORT)(_base_+CS_IO_PP_DATA),_reg_); \
}

#define CsSetTxReq(_base_,_txcmd_,_len_) { \
    IOSpaceWriteInt16((IO_PORT)(_base_+CS_IO_TxCMD),(_txcmd_)); \
    IOSpaceWriteInt16((IO_PORT)(_base_+CS_IO_TxLength),(UINT16)(_len_)); \
}

#define CsGetISQ(_base_,_isq_) \
    _isq_ = IOSpaceReadInt16((IO_PORT)(_base_+CS_IO_ISQ))

#define CsGetRecvData(_base_) \
    IOSpaceReadInt16((IO_PORT)(_base_+CS_IO_DATA))

#define CsReadRecvFrame(_base_,_dst_,_n_) \
    IOSpaceReadFIFO16((IO_PORT)(_base_+CS_IO_DATA),_dst_,_n_)

#if _DEBUG && !defined(m68k)
typedef struct {
    _TCHAR* Regname;
    IO_PORT Offset;
    BOOL IsReadDestructive;
} REGGIE;

void CsDumpSomeRegs(IO_PORT IoBase, REGGIE *Regs, BOOL IoMode, BOOL Safely)
{
    UINT RegNum;
    UINT16 Reg = 0;

    for (;;) {
        for (RegNum = 0; RegNum < 4; RegNum++) {
            if (Regs[RegNum].Regname == NULL)
                goto NewlineAndReturn;
            if (Regs[RegNum].IsReadDestructive && Safely)
                continue;
            if (IoMode) {
                Reg = IOSpaceReadInt16((IO_PORT)(IoBase+Regs[RegNum].Offset));
            } else {
                CsGetReg(IoBase,Regs[RegNum].Offset,Reg);
            }
            printf(_T("%s=%x "),Regs[RegNum].Regname,Reg);
        }
        printf(_T("\n"));
        Regs += RegNum;
    }
 NewlineAndReturn:
    printf(_T("\n"));
}

static REGGIE IOModeRegs[] = {
        { _T("IO_DATA"),        0x0000, TRUE },
        { _T("IO_DATA1"),       0x0002, TRUE },
        { _T("IO_TxCMD"),       0x0004, FALSE },
        { _T("IO_TxLength"),    0x0006, FALSE },
        { _T("IO_ISQ"),         0x0008, TRUE },
        { _T("IO_PP_PTR"),      0x000a, FALSE },
        { _T("IO_PP_DATA"),     0x000c, TRUE },
        { _T("IO_PP_DATA1"),    0x000e, TRUE },
        { NULL, }
};

static REGGIE InternalRegs[] = {
        { _T("PID"),            0x0000, FALSE },
        { _T("REV"),            0x0002, FALSE },
        { _T("IO_BASE"),        0x0020, FALSE },
        { _T("INTR"),           0x0022, FALSE },
        { _T("DMA_CH"),         0x0024, FALSE },
        { _T("DMA_SOF"),        0x0026, FALSE },
        { _T("DMA_FCOUNT"),     0x0028, FALSE },
        { _T("DMA_BCOUNT"),     0x002a, FALSE },
        { _T("MEMORY_BASE"),    0x002c, FALSE },
        { _T("PROM_BASE01"),    0x0030, FALSE },
        { _T("PROM_BASE23"),    0x0032, FALSE },
        { _T("PROM_SIZE01"),    0x0034, FALSE },
        { _T("PROM_SIZE23"),    0x0036, FALSE },
        { _T("EE_CMD"),         0x0040, FALSE },
        { _T("EE_DATA"),        0x0042, FALSE },
        { _T("RECV_BCOUNT"),    0x0044, FALSE },
        { _T("RxCFG"),          0x0102, FALSE },
        { _T("RxCTL"),          0x0104, FALSE },
        { _T("TxCFG"),          0x0106, FALSE },
        { _T("TxCMD"),          0x0108, FALSE },
        { _T("BufCFG"),         0x010A, FALSE },
        { _T("LineCTL"),        0x0112, FALSE },
        { _T("SelfCTL"),        0x0114, FALSE },
        { _T("BusCTL"),         0x0116, FALSE },
        { _T("TestCTL"),        0x0118, FALSE },
        { _T("ISQ"),            0x0120, TRUE },
        { _T("RxEvent"),        0x0124, FALSE },
        { _T("TxEvent"),        0x0128, FALSE },
        { _T("BufEvent"),       0x012c, FALSE },
        { _T("RxMISS"),         0x0130, FALSE },
        { _T("TxCOL"),          0x0132, FALSE },
        { _T("LineST"),         0x0134, FALSE },
        { _T("SelfST"),         0x0136, FALSE },
        { _T("BusST"),          0x0138, FALSE },
        { _T("TDR"),            0x013c, FALSE },
        { _T("LAFIlter01"),     0x0150, FALSE },
        { _T("LAFIlter23"),     0x0152, FALSE },
        { _T("LAFIlter45"),     0x0154, FALSE },
        { _T("LAFIlter67"),     0x0156, FALSE },
        { _T("StationAddress01"),0x0158, FALSE },
        { _T("StationAddress23"),0x015A, FALSE },
        { _T("StationAddress45"),0x015C, FALSE },
        { _T("RxStatus"),       0x0400, FALSE },
        { _T("RxLength"),       0x0402, FALSE },
        { NULL, }
};

void CsDumpAllRegs(PCS8900 pCs, BOOL Safely)
{
    IO_PORT IoBase = pCs->IoBase;

    printf(_T("CS8900(%x) Base = %x\n I/O Mode regs:\n"),(ADDRESS)pCs,IoBase);
    CsDumpSomeRegs(IoBase,IOModeRegs,TRUE,Safely);

    printf(_T(" Internal Regs:\n"));
    CsDumpSomeRegs(IoBase,InternalRegs,FALSE,Safely);
    
    /* Example output:
    CS8900(c000f448) Base = 8000300
     I/O Mode regs:
    IO_DATA=0 IO_DATA1=0 IO_TxCMD=c9 IO_TxLength=90
    IO_ISQ=0 IO_PP_PTR=315c IO_PP_DATA=2191 IO_PP_DATA1=2191
     Internal Regs:
    PID=630e REV=800 IO_BASE=300 INTR=5554
    DMA_CH=5553 DMA_SOF=0 DMA_FCOUNT=0 DMA_BCOUNT=0
    MEMORY_BASE=0 PROM_BASE01=0 PROM_BASE23=0 PROM_SIZE01=0
    PROM_SIZE23=0 EE_CMD=21e EE_DATA=2191 RECV_BCOUNT=0
    RxCFG=3 RxCTL=f05 TxCFG=7 TxCMD=c9
    BufCFG=b LineCTL=d3 SelfCTL=15 BusCTL=1017
    TestCTL=19 ISQ=0 RxEvent=eb04 TxEvent=8
    BufEvent=c RxMISS=50 TxCOL=12 LineST=1294
    SelfST=2d6 BusST=18 TDR=1c LAFIlter01=ffff
    LAFIlter23=ffff LAFIlter45=ffff LAFIlter67=ffff StationAddress01=d000
    StationAddress23=f1ca StationAddress45=2191 RxStatus=eb04 RxLength=3c
    */

}

UINT16 CsReadEEPROM(IO_PORT IoBase, UINT8 Offset)
{
    UINT16 Reg;
    int i;

    /* This line makes m68k gcc produce and.w instruction that it is invalid
     * according to the assembler.
     */
    CsSetReg(IoBase, (UINT16) CS_EE_CMD,
             (UINT16) ((((UINT16) Offset) & CS_EE_CMD_ADD_MASK)|CS_EE_CMD_READ_REG));
    /* Wait for SIBUSY to clear
     */
    for (i = 0; i < 100; i++) {
        Delay(50); /* takes at least 25us */
        CsGetReg(IoBase,CS_SelfST,Reg);
        if ((Reg & CS_SelfST_SIBUSY) == 0) break;
    }
    CsGetReg(IoBase,CS_EE_DATA,Reg);
    return Reg;
}

void CsDumpEEPROM(PCS8900 pCs)
{
    IO_PORT IoBase = pCs->IoBase;
    UINT16 Offset, Reg;

    printf(_T("CS8900(%x) Base = %x\n EEPROM Content:"),(ADDRESS)pCs,IoBase);
    /* Bit weird: its 16bit WORD addressed (not BYTE addressed)
     */
    for (Offset = 0; Offset < 0x100; Offset += 1) {
        if ((Offset & 0xf) == 0)
            printf(_T("\n%02x: "), Offset);
        Reg = CsReadEEPROM(IoBase,(UINT8)Offset);
        printf(_T("%04x "), Reg);
    }
    /* gets:
 EEPROM Content:
00: ffff ffff ffff ffff d000 f1ca 2191 ffff ffff ffff ffff ffff ffff ffff ffff ffff
10: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff d000 f1ca 2191 ffff
20: ffff ffff ffff ffff ffff ffff ffff ffff d000 f1ca 2191 ffff ffff ffff ffff ffff
30: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
40: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
50: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
60: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
70: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
80: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
90: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
a0: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
b0: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
c0: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
d0: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
e0: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
f0: ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff ffff
     */
}
#endif


/* Read the Ethernet Station Address
 */
PRIVATE
void
CsReadStationAddress(PCS8900 pCs, PUINT8 pSA)
{
    UINT16 aa[4]; /* align */
    IO_PORT         IoBase = pCs->IoBase;

    CsRegLock(pCs);
    CsGetPPMem(IoBase,CS_StationAddress+0,aa[0]);
    CsGetPPMem(IoBase,CS_StationAddress+2,aa[1]);
    CsGetPPMem(IoBase,CS_StationAddress+4,aa[2]);
    CsRegUnlock(pCs);

    memcpy(pSA,aa,6);
}


/* Process packets that have been completely received.
 * That means no errors, and "frame accepted into on-chip RAM"
 * as per specs pp 83 Figure 21 "Receive Frame Pre-Processing" flowchart.
 */
void
Cs8900Recv(PCS8900 pCs)
{
    IO_PORT         IoBase = pCs->IoBase;
    struct ifnet   *ifp = &pCs->is_if;
    INT32           Length;
    UINT16          Reg, Status;
    struct mbuf    *Pkt;
    struct ether_header eh;
    SCODE sc;

    DBGME(0,printf(_T("Enter Cs8900Recv\n")));

    /* Get Status.  In I/O mode this is just the top of the FIFO
     * off the DATA register, in Memory mode this is the start
     * of the packet.
     */
    Pkt = NULL;
    Status = CsGetRecvData(IoBase);

    /* Sometimes we get a status==0, along with length==0
     * This is handled below, dont assert.
     */
    LogIt(1,Status);

    /* Same for Length
     */
    Length = CsGetRecvData(IoBase);

    /* Sanity check the length
     * Sometimes we get I guess spurious recv interrupts, after reset.
     * BUGBUG find out why we get Length==0 so often
     */
    Length -= sizeof(struct ether_header);
    if ((Length <= 0) ||
        (Length > ETHERMTU)) {
        if (Length >= 0) {
            DBGME(1,printf(_T("Cs8900Recv: badlen x%x x%x\n"),Length,Status));
        }
        ifp->if_ierrors++;
        return;
    }

    assert(Length > 0 && Length < MCLBYTES);

    /* Get a buffer to copy to
     */
    sc = pCs->is_if.MBufFact->v->AllocateBuffer(pCs->is_if.MBufFact,
                                                Length,
                                                (PINETDRIVER) ifp,
                                                (PIBUFFER *) &Pkt);
    if (Pkt == 0 || FAILED(sc)) {
        /* No room, drop the packet
         */
        goto DropPacket;
    }

    /* Read&fix the ether header
     */
    CsReadRecvFrame(IoBase,(PUINT16)&eh,(sizeof eh) >> 1);

    eh.ether_type = ntohs(eh.ether_type);

    /* Either way..  and its time to copy now.
     */
    CsReadRecvFrame(IoBase,mtod(Pkt,PUINT16),Pkt->m_len >> 1);

    /* Possibly last byte
     */
    if (Length & 1) {
        Reg = CsGetRecvData(IoBase);
        (mtod(Pkt,PUINT8))[Pkt->m_len-1] = (UINT8)Reg;
    }

    LogIt(2,Length);

    /*
     * Hand the packet to the network module.
     */
    ifp->if_ipackets++;
    pCs->is_if.MBufFact->v->EtherInput(pCs->is_if.MBufFact,
                                       (PINETDRIVER) ifp,
                                       &eh, (PIBUFFER) Pkt);

    /* Nothing more to do
     */
    return;

DropPacket:
    LogIt(3,Length);
    if (Pkt)
        Pkt->m_hdr.v->Release((PIBUFFER) Pkt);
    Pkt = NULL;
    ifp->if_iqdrops++;
    ifp->if_ierrors++;
}

/*
 * Start output on interface.
 */
void MCT
Cs8900Xmit( PCS8900 pCs)
{
    struct mbuf     *Request, *m;
    UINT            Length;
    UINT16          Reg, Data16;
    UINT8           *pData8;
    int             i;
    struct ifnet    *ifp = &pCs->is_if;
    IO_PORT         IoBase = pCs->IoBase;

    DBGME(0,printf(_T("Enter Cs8900Xmit\n")));

    pCs->is_if.MBufFact->v->NetLock(pCs->is_if.MBufFact);

    /* If we came from interrupt, that must be the last xmit command
     * If we didnt, the flag is cleared anyways.
     */
    LogIt(4,ifp->if_flags);
    ifp->if_flags &= ~IFF_OACTIVE;

    /* Queue up as many frames as we can
     */
    while ((Request = pCs->is_if.if_snd.ifq_head) != NULL) {

        /* Did we already bid for this frame ?
         * [Thats how we'd know we came from interrupt]
         */
        if (pCs->TxBidding == FALSE) {

            /* See if we can transmit this (new!) frame now,
             * bidding the chip to see if it has room for it.
             * Length has already been checked to be less than ethermax.
             */
            Length = Request->m_pkthdr.len;
            LogIt(5,Length);
            CsSetTxReq(IoBase,
                       CS_TxCMD_TxStart_full,
                       Length);

            /* BusST tells us if there is room immediately
             * Gotta lock it..bummer.
             */
            CsRegLock(pCs);
            CsGetReg(IoBase,CS_BusST,Reg);
            CsRegUnlock(pCs);

            if ((Reg & (CS_BusST_Rdy4YxNOW|CS_BusST_TxBidErr)) !=
                CS_BusST_Rdy4YxNOW) {

                /* Cant do right this minute, or bogus
                 * We have done the bidding, remember and bail out.
                 */
                assert((Reg & CS_BusST_TxBidErr) == 0);

                pCs->TxBidding = TRUE;
                ifp->if_flags |= IFF_OACTIVE;
                LogIt(6,0);
                break;
            }
        }
        pCs->TxBidding = FALSE;

        /* Chip is (now) ready for the frame's data
         */
        IF_DEQUEUE(&pCs->is_if.if_snd, m);
        assert(m == Request);

        /* Copy the frame out to the chip's memory
         * Beware of odd chunks
         */
        Reg = 0;
        Data16 = 0;
        for (;m;m=m->m_next) {

            /* Did we endup odd ?
             */
            LogIt(7,m->m_len);

            pData8 = mtod(m,PUINT8);

            if (Reg) {
                /* Bummer.
                 * Complete the first word with the leftover
                 */
                Data16 = (Data16 | (*(pData8++) << 8));
                IOSpaceWriteInt16((IO_PORT)(IoBase+CS_IO_DATA),Data16);

                /* Loop by words
                 * Yes this is a bit tricky.
                 */
                for (i = 1; i < m->m_len - 1; i += 2) {
                    Data16 = pData8[0] | (pData8[1] << 8);
                    IOSpaceWriteInt16((IO_PORT)(IoBase+CS_IO_DATA),Data16);
                    pData8 += 2;
                }

            } else {
                /* Do pairs as much as we can
                 * Once in a while we get a bad pointer..
                 */
                if ((((ADDRESS)pData8) & 1) == 0) {
                    IOSpaceWriteFIFO16((IO_PORT)(IoBase+CS_IO_DATA),
                                       (UINT16 *) pData8,
                                       m->m_len >> 1);
                } else {
                    /* Hyperbummer.
                     */
                    for (i = 0; i < m->m_len - 1; i += 2) {
                        Data16 = pData8[0] | (pData8[1] << 8);
                        IOSpaceWriteInt16((IO_PORT)(IoBase+CS_IO_DATA),Data16);
                        pData8 += 2;
                    }
                }

            }

            /* Back to even ?
             * Ifnot, remember leftover byte
             */
            Reg = (UINT16)((m->m_len & 1) ^ Reg);
            if (Reg)
                Data16 = mtod(m,PUINT8)[m->m_len-1];
                
        }

        /* Write leftover if any
         */
        if (Reg)
            IOSpaceWriteInt16((IO_PORT)(IoBase+CS_IO_DATA),Data16);

        /* Off it goes.
         */
        LogIt(8,Request->m_pkthdr.len);
        pCs->is_if.if_opackets++;
        Request->m_hdr.v->Release((PIBUFFER) Request);
    }

    /* Note.
     * We do not need to worry about masking/unmasking xmit interrupts,
     * because regular tx-complete-ok do not generate any interrupts and
     * we only get OkToXmit when we actually need it, so that stays on.
     * Which is more efficient.
     */

    pCs->is_if.MBufFact->v->NetUnlock(pCs->is_if.MBufFact);
}

/*
 * Interrupt service routine.
 * Return TRUE if rescheduling needed.
 */
BOOL MCT
Cs8900Isr(PCDEVICE pDev, PBOOL pNotMyInterrupt)
{
    PCS8900 pCs = (PCS8900) pDev;
    IO_PORT IoBase = pCs->IoBase;

    /* Just pull off the first entry from the ISQ, which acks the chip.
     * The chip wont re-interrupt until (1.6us after) we empty the queue.
     * GIven that we do not empty the queue here then it wont.
     * This way the changes to the register state are minimized.
     */
    assert(pCs->ISQHead == 0);    /* make sure hw is sane */
    CsGetISQ(IoBase,pCs->ISQHead);

    /* Wakeup handler thread.
     */
    return Condition_InterruptSignal(&pCs->is_if.IntrPoint);
}

/* Thread-level interrupt handling
 */
BOOL MCT
Cs8900Interrupt(PCDEVICE pDev)
{
    PCS8900 pCs = (PCS8900) pDev;
    UINT16 Reg;

    DBGME(0,printf(_T("Enter Cs8900Interrupt\n")));

    /* NB: If the hw is broken we could use AtomicSwap to
     * synchronize with the ISR. Hopefully not.
     */
    Reg = pCs->ISQHead;
    assert(Reg != 0);
    pCs->ISQHead = 0;

    /* Loop pulling events off the queue
     */
    while (Reg) {

        DBGME(0,printf(_T("Cs8900Intr::Reg=%x\n"),Reg));

        switch (Reg & CS_ISQ_RegNum) {

        case CS_RxEvent_ID:
            /* Quick check for errors that dont commit a frame
             */
            if ((Reg &
                 (CS_RxEvent_Extradata|
                  CS_RxEvent_Runt|
                  CS_RxEvent_CRCError)) == 0) {

                /* A good frame, go get it
                 */
                Cs8900Recv(pCs);

            } else {
                /* Wish we could be more specific
                 */
                pCs->is_if.if_ierrors++;
            }
            break;

        case CS_TxEvent_ID:
            /* This is a xmit error
             * BUGBUG be more specific
             */
            LogIt(9,Reg);
            pCs->is_if.if_oerrors++;
            break;

        case CS_BufEvent_ID:
            /* Either a ready-for-xmit, or error.
             */
            LogIt(9,Reg);
            if (Reg & CS_BufEvent_Rdy4Tx) {

                /* We have enough room for that packet now.
                 */
                assert(pCs->is_if.if_flags & IFF_OACTIVE);
                Cs8900Xmit(pCs);

            } else if (Reg & CS_BufEvent_TxUnderrun) {

                /* This really shouldnt happen but we see it at start.
                 */
                pCs->MemoryErrors++;
                pCs->is_if.if_oerrors++;
            } else {

                /* You'd think we should handle coll/miss overflows.
                 * But they dont come in on this register, so see below.
                 */
                DBGME(3,printf(_T("Cs8900:Bad BufEvent %x\n"),Reg));
            }
            break;

        case CS_RxMISS_ID:
            /* The Rx MISS counter overflowed.
             * No prob, just bump our overflow counter.
             */
            pCs->MissOverflow++;
            break;

        case CS_TxCOL_ID:
            /* The Tx COLLISION counter overflowed.
             * No prob, just bump our overflow counter.
             */
            pCs->CollOverflow++;
            break;

        default:
            /* Something really strange that we cant handle happened
             */
            DBGME(3,printf(_T("Cs8900:Intr::Reg %x ?\n"),Reg));
        }
    
        /* Next ISQ entry, possibly enables interrupts.
         */
        CsGetISQ(pCs->IoBase,Reg);
    }

    /* Since we got out the top of the ISQ is zero and interrupts
     * are enabled now.  Nothing else we can/should do.
     */
    return TRUE;
}

/* Random other things..
 */
SCODE MCT
Cs8900Ioctl( PCDEVICE pDev, UINT Prop, caddr_t Data)
{
    PCS8900 pCs = (PCS8900) pDev;
    IO_PORT IoBase = pCs->IoBase;
    UINT16 Reg;

    DBGME(2,printf(_T("Enter Cs8900Ioctl(%x)\n"),Prop));

    /* Freshen up statistics, if we need. [yes we do]
     */
    if (Prop == SIOCGIFSTATS) {

        /* Compute missed packets count
         * BUGBUG Spec says a read zeroes the counter, so this isnt right
         */
        CsRegLock(pCs);
        CsGetReg(IoBase,CS_RxMISS,Reg);
        pCs->MissedPackets = pCs->MissOverflow << (16-CS_RxMISS_Shift) |
                             ((Reg >> CS_RxMISS_Shift) & CS_RxMISS_Mask);

        /* Compute collision count
         * BUGBUG Spec says a read zeroes the counter, so this isnt right
         */
        CsGetReg(IoBase,CS_TxCOL,Reg);
        pCs->CollisionCount = pCs->CollOverflow << (16-CS_TxCOL_Shift) |
                             ((Reg >> CS_TxCOL_Shift) & CS_TxCOL_Mask);

        CsRegUnlock(pCs);
    }

    /* We care for requests to turn the interface down/up.
     */
    if (Prop == SIOCSIFFLAGS) {

        if ((pCs->is_if.if_flags & IFF_UP) == 0) {
            /*
             * Down. Put it in low power mode.
             * HW Suspend is only 100uA, but we cant tickle the SLEEP pin
             * so the only thing we can do is SW Suspend (1mA).
             */
            CsRegLock(pCs);
            CsGetReg(IoBase,CS_SelfCTL,Reg);
            Reg |= CS_SelfCTL_SWSuspend;
            CsSetReg(IoBase,CS_SelfCTL,Reg);
            CsRegUnlock(pCs);

            /* NB: To take it out of Suspend mode it only needs a write
             * to the chip's assigned I/O space. The write is ignored.
             * Which means a call to Cs8900Reset() will do the trick.
             */
        } else {
            /*
             * Up. Turn power back on, this causes an internal reset.
             * See note above if this is the first time around.
             */
            pDev->v->Initialize(pDev);
        }
    }
    return NO_ERROR;
}


/* This doesnt belong here but we dont have a power mgmt interface yet
 */
void Cs8900Unsleep(void)
{
typedef struct {
    volatile UINT32 Level;
    volatile UINT32 Direction;
    volatile UINT32 Set;
    volatile UINT32 Clear;
    volatile UINT32 DetectRaising;
    volatile UINT32 DetectFalling;
    volatile UINT32 DetectStatus;
    volatile UINT32 Alternate;    
} GPIO;
#define gpio ((GPIO*)0x90040000)

    DBGME(1,printf("Unsleep cs8900 chip\n"));
    /* Take the chip out of sleep
     */
#define CS8900_SLEEP_PIN (1<<27)
    gpio->Direction |= CS8900_SLEEP_PIN; /* its an output */
    gpio->Set = CS8900_SLEEP_PIN | (gpio->Level & gpio->Direction);
    DBGME(1,_tprintf(_T("GPIO: l=%x d=%x 1=%x 0=%x r=%x f=%x s=%x a=%x\n"), *gpio));
}

/* Initialize/Reset/Reinitialize the card
 * NB: Assumes netlock held.
 */
INT MCT
Cs8900Reset(PCDEVICE pDev)
{
    PCS8900 pCs = (PCS8900) pDev;
    IO_PORT IoBase = pCs->IoBase;
    UINT16 Reg = 0;
    TIME Idle;
    int i;
    struct ifnet *ifp;

    DBGME(2,printf(_T("Enter Cs8900Reset\n")));
//    DBGME(0,CsDumpAllRegs(pCs,FALSE));

#if 0
    /* Stop network activity first.  A RESET will do it.
     * NB: See note in Ioctl viz power save, thats why duplicate read.
     */
    CsGetReg(IoBase,CS_SelfCTL,Reg);
    CsGetReg(IoBase,CS_SelfCTL,Reg);
    CsSetReg(IoBase,CS_SelfCTL,Reg|CS_SelfCTL_RESET);
    DBGME(3,printf(_T("CTL-> %x\n"),Reg|CS_SelfCTL_RESET));
#endif
    /* Say we are down and reinit sw state
     */
    ifp = &pCs->is_if;
    if (ifp->if_flags & IFF_RUNNING) {
        ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
    }
    pCs->CompleteResets++;
    pCs->TxBidding = FALSE;

    /* Wait as long as necessary: at least 10 msec, and until INITD
     */
    Int32ToInt64(Idle,TIME_RELATIVE(TIME_MILLIS(50)));
    SleepUntil(Idle);

#if 0
    /* Take it out of reset (BUGBUG makes no sense)
     */
    IOSpaceWriteInt16((IO_PORT)(IoBase+CS_IO_PP_PTR),0x0);

#endif

    /* Busy loop until done or timeout (1msec)
     */
    for (i = 0; i < 100; i++) {
        CsGetReg(IoBase,CS_SelfST,Reg);
        if ((Reg & (CS_SelfST_INITD|CS_SelfST_SIBUSY)) ==
            CS_SelfST_INITD)
            break;
        Delay(10);
    }
    if ((Reg & (CS_SelfST_INITD|CS_SelfST_SIBUSY)) !=
            CS_SelfST_INITD) {
        DBGME(3,printf(_T("Cs8900Reset failed %x\n"),Reg));
        DBGME(3,CsDumpAllRegs(pCs,FALSE));
        return 1;
    }

    /* Select INTRQ0 [BUGBUG system-specific]
     */
    CsSetReg(IoBase,CS_INTR,CS_INTR_0);

    /* Enable Memory mode if desired (reg17 pag 67)
     */
    Reg = CS_BusCTL_IOCHRDYE;
    CsSetReg(IoBase,CS_BusCTL,Reg);


    /* Configure the PHY: select BT10 and SQUELCH
     * This changes the threshold for a valid signal from 300mV to 180mV.
     * This is to fix problems we have seen with some (usually 8 port) hubs.
     */
    /* Before we change the LineCTL we must carefully stop the TX
     * Lets stop the RX just in case.
     */
    CsGetReg(IoBase,CS_LineCTL,Reg);
    Reg = Reg & ~(CS_LineCTL_SerRxON|CS_LineCTL_SerTxON);
    CsSetReg(IoBase,CS_LineCTL,Reg);
    Delay(50);

    if (Reg != (CS_LineCTL_10BaseT|CS_LineCTL_LoRxSquelch)) {
        Reg = CS_LineCTL_10BaseT|CS_LineCTL_LoRxSquelch;
        CsSetReg(IoBase,CS_LineCTL,Reg);
        Delay(50);
    }

    /* Now that we stopped the PHY, clearup any pending interrupts
     * They might be there un-acked because of the bootloader.
     */
    do {
        CsGetISQ(IoBase,Reg);
    } while (Reg != 0);
    pCs->ISQHead = 0;

    /* Configure the receiver: only good frames.
     */
    CsSetReg(IoBase,CS_RxCTL,
             CS_RxCTL_RxOKA|CS_RxCTL_IndividualA|CS_RxCTL_BroadcastA);

    /* Enable error and Tx bidding interrupts
     */
    CsSetReg(IoBase,CS_BufCFG,
             CS_BufCFG_TxUnderruniE|
             CS_BufCFG_TxColOvfliE|CS_BufCFG_MissOvfliE|
             CS_BufCFG_Rdy4TxiE);

    /* Enable Tx error interrupts
     * NB: Collisions are handled differently.
     */
    CsSetReg(IoBase,CS_TxCFG,
             CS_TxCFG_LostCRSiE|CS_TxCFG_SQEerroriE|
             CS_TxCFG_OutOfWindow|CS_TxCFG_JabberiE);

    /* Enable Receive interrupts
     * [also, clearup any pending recv frames]
     */
    CsSetReg(IoBase,CS_RxCFG, CS_RxCFG_Skip_1|
             CS_RxCFG_RxOKiE|CS_RxCFG_CRCerroriE|
             CS_RxCFG_RuntiE|CS_RxCFG_ExtradataiE);

    /* Enable the PHY at this point.
     */
    Reg = CS_LineCTL_10BaseT|CS_LineCTL_LoRxSquelch|
          CS_LineCTL_SerRxON|CS_LineCTL_SerTxON;
    CsSetReg(IoBase,CS_LineCTL,Reg);

    /* Enable Interrupts (reg17 pag 67)
     */
    Reg = CS_BusCTL_EnableIRQ|CS_BusCTL_IOCHRDYE;
    CsSetReg(IoBase,CS_BusCTL,Reg);

    /* Say we are up
     */
    ifp->if_flags |= (IFF_UP | IFF_RUNNING);

    /* We are rolling
     */
    DBGME(2,printf(_T("Cs8900:Reset done\n")));
    return 0;

}

struct INetDriverVtbl Cs8900DriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    (_INetDriver_Isr_METHOD_TYPE)  Cs8900Isr,
    (_INetDriver_HandleInterrupt_METHOD_TYPE) Cs8900Interrupt,
    (_INetDriver_Initialize_METHOD_TYPE) Cs8900Reset,
    NULL,
    NULL,
    (_INetDriver_Start_METHOD_TYPE) Cs8900Xmit,
    (_INetDriver_Ioctl_METHOD_TYPE) Cs8900Ioctl,
    NULL
};

/* Once-only initialization.
 * Most of the work is actually done by the Reset function.
 */
PCS8900
Cs8900Initialize(IO_PORT IoBase)
{
    PCS8900 pCs;
    UINT16 Reg;
    TIME ABit;
    static short unit=0;

    DBGME(2,printf(_T("Enter Cs8900Initialize %x\n"),IoBase));

#if LOGIT
    printf("Log @%x\n",(ADDRESS) &LogBuf[0]);
#endif

    Cs8900Unsleep();

    /* Look for its signature in I/O mode
     * NB: Typically gets 3128 if bootloader TFTP last xmit.
     * Otherwise needs a write to get out of power upreset.
     */
    Reg = IOSpaceReadInt16((IO_PORT)(IoBase+CS_IO_PP_PTR));
    if ((Reg & 0xF000) != 0x3000) {
        DBGME(3,printf(_T("No Cs8900 signature at %x (%x)\n"),IoBase,Reg));
        goto ErrorExit;
    }

    /* Make sure its functional
     * We do depend on the chip having an EEPROM and that been read.
     * Give it time to get out of reset.
     * ==> check CS_SelfST (see page 66)
     * NB: Typically gets x2d6 == EEPres INITD 3p3v
     */
    CsGetReg(IoBase,CS_SelfST,Reg);
    if (Reg & CS_SelfST_SIBUSY) {
        Int32ToInt64(ABit, TIME_RELATIVE(TIME_MILLIS(30)));
        SleepUntil(ABit);
        CsGetReg(IoBase,CS_SelfST,Reg);
    }
    if ((Reg &
         (CS_SelfST_INITD|CS_SelfST_SIBUSY|CS_SelfST_EEPresent))
        != (CS_SelfST_INITD|CS_SelfST_EEPresent)) {
        DBGME(3,printf(_T("Cs8900 at %x (%x) has no EEPROM, cant use.\n"),
                       IoBase,Reg));
        goto ErrorExit;
    }

    /* Ok we are doing it for real.
     */
    {
        PIHEAP pIHeap = CurrentHeap();
        pCs = (PCS8900) pIHeap->v->Alloc(pIHeap, HEAP_ZERO_MEMORY,
                                         sizeof(CS8900), 0);
    }
    if (pCs == NULL)
        goto ErrorExit;

    /* Set basic info.
     */
    pCs->IoBase = IoBase;
    pCs->ISQHead = 0;
    pCs->MissOverflow = 0;
    pCs->CollOverflow = 0;

    Mutex_Init(&pCs->Lock);

    /* Read station address off chip
     */
    CsReadStationAddress(pCs,&pCs->is_addr[0]);

    /* More init
     */
    pCs->is_if.v = &Cs8900DriverVtable;
    pCs->is_if.RefCnt = 1;
    pCs->is_if.if_unit = unit++;
    pCs->is_if.if_name = "cs";
    pCs->is_if.if_type = IFT_ETHER;
    pCs->is_if.if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
    pCs->is_if.if_baudrate = 10000000; /* 10Mbps */

    DBGME(2,printf(_T("Cs8900 CardAddr %02x:%02x:%02x:%02x:%02x:%02x Base %x\n"),
                   pCs->is_addr[0], pCs->is_addr[1],
                   pCs->is_addr[2], pCs->is_addr[3],
                   pCs->is_addr[4], pCs->is_addr[5],
                   pCs->IoBase));

//    DBGME(0,CsDumpEEPROM(pCs));
    return pCs;

ErrorExit:
    return NULL;
}

SCODE __cdecl Cs8900Constructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk, ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(Interrupt);

    DBGME(0,printf("Cs8900Constructor(%x %s %x %x %x)\n",
                   (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                   (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) Cs8900Initialize((IO_PORT)RegisterBase);

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,printf(_T("Cs8900Constructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl Cs8900DriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) Cs8900Constructor
};

static struct _DRIVERCOB Cs8900DriverCobObject = { &Cs8900DriverCobVtable, 1 };

PIUNKNOWN Cs8900CobMain(void)
{
    SCODE sc;
    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &Cs8900DriverCobVtable,
                          "CDeviceConstructor",
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &Cs8900DriverVtable, "CNetDriver",
                          &IID_INetDriver);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &Cs8900DriverCobObject;
}

/* DRIVER.DAT line for this one:
 *
 * Cs8900::Cs8900New:NetworkDriver 0x08000300 47
 *
 * It is registered in the name space under Cs8900.
 */
EXTERN_C PCDEVICE RTLCALLTYPE 
Cs8900New(PUINT pPortBase, PUINT pIrq)
{
    PCS8900 pCs;
    SCODE sc;

    sc = BaseInheritClass((ADDRESS *) &Cs8900DriverVtable, "CNetDriver",
                          &IID_INetDriver);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    pCs = Cs8900Initialize((IO_PORT)*pPortBase);
    if (pCs == NULL)
        return NULL;

    return (PCDEVICE) pCs;
}

