/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    dc21x4.c
 *
 * Purpose:
 *
 *    Device driver for the DEC 21x4x PCI Ethernet
 *
 * Author:
 *    A. Forin (sandrof), using information from:
 *      "DECchip 21040 Ethernet LAN Controller for PCI
 *       Hardware Reference Manual"
 *      EC-N0752-72, Digital Semiconductor. May 1994.
 *
 */

#define USE_INTRINSICS
#include <sock.h>
#include <if.h>
#include <if_types.h>
#include <ip/in.h>
#include <ip/arp.h>
#include <mmhal.h>
#include <drivers/drivers.h>
#include <stdio.h>
#include <mmmacros.h>

#include "dc21x4.h"

#ifdef _MSC_VER
#pragma warning(disable:4055)   /* cast from data pointer to function pointer  */
#endif

typedef BOOL (RTLCALLTYPE *SIGNAL)(PCONDITION);

/*
 * Implementation dependent constants and defines
 */
#define NRCV    64      /* Receive descriptors */
#define NXMT    16      /* Transmit descriptors */

#define HEAP_BUFFER_SIZE        (1520) /* 0x5F0 */
#define HEAP_BUFFERS            (HEAP_BUFFER_SIZE*(NRCV+NXMT))
#define HEAP_OVERHEAD           (sizeof(DC_LIST_ENTRY)*(NXMT+NRCV))
#define HEAP_MEMORY_SIZE        (HEAP_BUFFERS+HEAP_OVERHEAD)

#define DC_BUS_MODE_DEFAULT     (DC_BUS_CAL_16lw|(16<<DC_BUS_PBL_SHIFT))
#define DC_INTERRUPTS_ENABLED   (0x0001b1e9)
#define DC_OPER_BITS            (DC_OPER_ST | DC_OPER_CA /* | DC_OPER_SB */)

/*
 * SIA programming
 * [Add to this table for extensions]
 */
typedef const struct {
    UINT32 Csr13Value;
    UINT32 Csr14Value;
    UINT32 Csr15Value;
} MEDIA_PARAMS;

#define N_MEDIA 3
static MEDIA_PARAMS MediaTable[N_MEDIA] = {
    /*
     * Twisted pair (10BASE-T)
     */
    { 0x8f01, 0xffff, 0x0000 },

    /*
     * AUI
     */
    { 0x8f09, 0x0705, 0x0006 },

    /*
     * Twisted pair, full duplex  (!UNSUPPORTED!)
     */
    { 0x8f01, 0xfffd, 0x0000 }
};

/*
 * Dc21x4 descriptor
 */
typedef struct  _DC21x4 {
        struct arpcom   ac;             /* standard interface header    */
#define is_if ac.ac_if
#define is_addr ac.ac_enaddr
#define misscnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_MISS]
#define rstrtcnt ac.ac_if.if_data.ifi_errors[IFI_ERR_RESET]
#define bablcnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_XTIM]
#define merrcnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM]
#define sqercnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+1]
#if _DEBUG
#define numintr  ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+2]
#define spurious ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+3]
#define sawcsr   ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+4]
#endif

        /* This cache line alone used in Isr
         */
        SIGNAL          Signal;
        IO_PORT         PortBase;       /* Dc21x4 registers     */
        UINT            IsrStatus;      /* because spurious ?!BUGBUG? */
        UINT            pad;

        /* Data used in interrupt handler code
         */
        UINT            ReceiverHead;   /* Rcv buffer last read         */

        UINT            XmitterHead;    /* Xmt queue head (insert)      */
        UINT            XmitterTail;    /* Xmt queue tail (remove)      */

        PDC_LIST_ENTRY  RRing[NRCV];    /* Receive  ring desc. */
        volatile PTR    RBuf[NRCV];     /* Receive  buffers */
        PDC_LIST_ENTRY  TRing[NXMT];    /* Transmit ring desc. */
        volatile PTR    TBuf[NXMT];     /* Transmit buffers */

        /* Infrequently used data follows
         */
        UINT            RecvRingBase;   /* */
        UINT            XmitRingBase;   /* */

        UINT            MemBaseV;       /* Heap memory, Host virtual  */
        UINT            MemBaseP;       /* Heap memory, Host physical */
        UINT            TopAlloc;       /* Heap memory allocator */
        MEDIA_PARAMS    *MediaType;

} DC21x4, *PDC21x4;

#define pDC(_i_) ((PDC21x4)(_i_))
#define iDC(_p_) ((PINETDRIVER)(_p_))

/*
 * Functions we export to the upper layers (vtable)
 */
PRIVATE INT  MCT Dc21x4Initialize(PINETDRIVER This);
PRIVATE void MCT Dc21x4Start(PINETDRIVER This);
PRIVATE INT  MCT Dc21x4Ioctl(PINETDRIVER This, UINT32 Prop, PTR Data);
PRIVATE BOOL MCT Dc21x4Isr(PINETDRIVER This, PBOOL pNotMyInterrupt);
PRIVATE BOOL MCT Dc21x4HandleInterrupt(PINETDRIVER This);

struct INetDriverVtbl Dc21x4DriverVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    Dc21x4Isr,
    Dc21x4HandleInterrupt,
    Dc21x4Initialize,
    NULL,
    NULL,
    Dc21x4Start,
    Dc21x4Ioctl,
    NULL
};

/*
 * Access to registers
 */

#define Dc21x4ReadRegister(_d_,_r_) \
        IOSpaceReadInt32((IO_PORT)((_d_)->PortBase + (_r_)))

#define Dc21x4WriteRegister(_d_,_r_,_v_) \
        IOSpaceWriteInt32((IO_PORT)((_d_)->PortBase + (_r_)),(_v_))

/*
 * Stop the chip
 * Q: is it better to do a reset or a graceful stop ?
 */
#define Dc21x4Stop(_d_) \
        Dc21x4WriteRegister(_d_,DC_OPERATION_MODE,0)
#define Dc21x4Reset(_d_) \
{ \
        Dc21x4WriteRegister(_d_,DC_BUS_MODE,DC_BUS_SWR); \
        Delay(5000); \
        Dc21x4WriteRegister(_d_,DC_BUS_MODE,DC_BUS_MODE_DEFAULT); \
}

/*
 * Interrupt munging
 */
#define Dc21x4EnableInterrupts(_d_) \
        Dc21x4WriteRegister(_d_,DC_INTERRUPT_MASK,DC_INTERRUPTS_ENABLED)
#define Dc21x4DisableInterrupts(_d_) \
        Dc21x4WriteRegister(_d_,DC_INTERRUPT_MASK,0)


/*
 * Map a heap offset into an host-space virtual address
 * Map a heap offset into a chip-space physycal address
 */
#define Hmem(_d_,_o_)   (PTR)   ((BYTE*)(_o_) + (_d_)->MemBaseV)
#define Lmem(_d_,_o_)   (UINT32)((BYTE*)(_o_) + (_d_)->MemBaseP)

/*
 *
 * Internal functions & definitions
 *
 */

PRIVATE void Dc21x4SendSetupPacket( PDC21x4 sc);
PRIVATE void Dc21x4InitializeHeap( PDC21x4 sc);
PRIVATE volatile PTR Dc21x4DescAlloc (PDC21x4 sc, PDC_LIST_ENTRY pDesc);
PRIVATE PIBUFFER Dc21x4RecvPkt(PDC21x4 sc, volatile UINT8 *RBuf, int len,
                               struct ether_header *eh);
PRIVATE UINT Dc21x4SendPkt(struct mbuf *m, volatile UINT8 *SBuf);
PRIVATE void Dc21x4RecvInterrupt( PDC21x4 sc );
PRIVATE void Dc21x4XmitInterrupt( PINETDRIVER This );

PRIVATE void Dc21x4ReadPermanentAddress( PDC21x4 pThis, UINT8 *EthAddr );
PRIVATE UINT Dc21x4InstallEthernetAddress( PDC21x4 pThis, SETUP_PACKET_PERFECT *pPkt );
PRIVATE UINT DcMalloc();

#if _DEBUG
PRIVATE void DcDump( PDC21x4 pThis );
#endif

/*++
 *  Functions in this section are used exactly once
 *++
 */

/*
 * Driver initialization function, aka contructor for
 * a DC21x4 object.
 * This function assumes that hardware does exist at the given PortBase.
 * Returns a properly casted descriptor.
 */
PUBLIC PINETDRIVER RTLCALLTYPE Dc21x4New( ADDRESS * pPortBase, PUINT pIrq, PUINT pMedia)
{
    PDC21x4      pThis;
#ifdef VM
    PIADDRESSSPACE    pIAddr;
#else
    PIHEAP pIHeap;
#endif
    ADDRESS      Addr = 0;
    static short unit = 0;

#ifdef mmheavy
    SYSTEM_BUS_TYPE Bus;

    /* Sanity checks
     */
    {
        PIMACHINE pM = CurrentMachine();
        (void) pM->v->GetSystemBusType( pM, &Bus );
        if (Bus != PCI_BUS && Bus != PCI_BUS_M2) {
            printf(_T("DC21x4 only on PCI busses (got x%x)\n"), Bus);
            return NULL;
        }

        /* BUGBUG We should also check we have a 21040 that is not
         * a pass 0, looking at the PCI Config +0x8 Revision ID.
         * HOWEVER, one does not expect DEC to *sell* pass 0 chips.
         */
    }
#endif

    /*
     * Allocate a descriptor
     */
    pThis = (PDC21x4)malloc(sizeof(*pThis));
    if (pThis == NULL){
        DBG((_T("Could not allocate %d bytes\n"), sizeof *pThis));
        return NULL;
    }

    memset(pThis, 0, sizeof(*pThis));

    /*
     * Fill in the descriptor
     */
    pThis->PortBase = (IO_PORT) *pPortBase;

    if (*pMedia >= N_MEDIA)
        *pMedia = 0; /* Sanity */
    pThis->MediaType = &MediaTable[*pMedia];

    /*
     * Cache this pointer here, because at interrupt time
     * the Pcb might be NULL.
     */
    pThis->Signal = (SIGNAL) GetRtlFunctionAddress(Condition_InterruptSignal);

    /*
     * Read the address from the prom and save it. 
     */
    Dc21x4ReadPermanentAddress(pThis, pThis->is_addr);
#if _DEBUG
    printf(_T("dc: EthAddr %x %x %x %x %x %x\n"),
           pThis->is_addr[0], pThis->is_addr[1], pThis->is_addr[2], 
           pThis->is_addr[3], pThis->is_addr[4], pThis->is_addr[5]);
#endif

    /*
     * Reset chip (if warm reboot chip is in error)
     */
    Dc21x4DisableInterrupts(pThis);
    Dc21x4Stop(pThis);
    Delay(2000);
    Dc21x4Reset(pThis);

    /*
     * Allocate heap memory 
     */
#ifdef VM
    pIAddr = CurrentAddressSpace();

    Rc = pIAddr->v->Allocate(pIAddr, &Addr, HEAP_MEMORY_SIZE);
    if (FAILED(Rc)) {
        DBGME((_T("Could not allocate (%x)\n"), Rc));
        goto ErrorExit;
    }

    Rc = pIAddr->v->Commit(pIAddr, Addr, HEAP_MEMORY_SIZE, 
            ADDR_COMMIT_CONTIGUOUS | ADDR_COMMIT_LOCK);
    if (FAILED(Rc)) {
        DBGME((_T("Could not commit (%x)\n"), Rc));
        goto ErrorExit1;
    }

    pThis->MemBaseV = Addr;

    /*
     * Get phys address for it
     */
    Rc = GetPhysicalAddress( Addr, &pThis->MemBaseP );
    if (FAILED(Rc)) {
        DBGME((_T("GetPhys failed on %x (%x)\n"), Addr, Rc));
        goto ErrorExit1;
    }
#else
    pIHeap = CurrentHeap();
    Addr = (ADDRESS) pIHeap->v->Alloc(pIHeap, 0, HEAP_MEMORY_SIZE, 0);
    if (!Addr) {
        TRACE("Could not allocate\n");
        goto ErrorExit;
    }

    pThis->MemBaseV = Addr;
    pThis->MemBaseP = Addr;
#endif

    /*
     * Setup shared mem structures
     */
    Dc21x4InitializeHeap(pThis);

    /*
     * Initialize the standard interface descriptor 
     */
    pThis->is_if.v = &Dc21x4DriverVtbl;
    pThis->is_if.if_unit = unit++;
    pThis->is_if.if_name = "dc";
    pThis->is_if.if_type = IFT_ETHER;
    pThis->is_if.if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
    pThis->is_if.if_baudrate = 10*1000*1000; /* 10Mbps */

    /*
     * All is well
     */
    return (PINETDRIVER)pThis;

#ifdef VM
 ErrorExit1:
    (void) pIAddr->v->Deallocate(pIAddr, Addr, HEAP_MEMORY_SIZE);
#endif
 ErrorExit:
    free(pThis);
    return NULL;
}

/* Initialize shared memory structures.
 * Assumes chip is stopped.
 */
PRIVATE void
Dc21x4InitializeHeap( PDC21x4 pThis)
{
    int         HeapOffset;
    PDC_LIST_ENTRY      pDesc;
    int         i;

    /*
     * Reset memory allocator
     */
    pThis->TopAlloc = 0;
    memset((PTR)pThis->MemBaseV, 0, HEAP_MEMORY_SIZE);

    /*
     * Allocate both descriptor rings at once.
     * Note that whatever alignment requirements must be
     * guaranteed by the allocator.
     */
    HeapOffset = DcMalloc(pThis, sizeof (DC_LIST_ENTRY) * (NXMT + NRCV));

    /*
     * Initialize the buffer descriptors
     */
    pThis->RecvRingBase = Lmem(pThis,HeapOffset);

    for ( i = 0; i < NRCV ; i++, HeapOffset += sizeof(DC_LIST_ENTRY)) {
        pDesc = (PDC_LIST_ENTRY)Hmem(pThis,HeapOffset);
        pThis->RRing[i] = pDesc;
        pThis->RBuf[i] = Dc21x4DescAlloc (pThis, pDesc);
    }
    pThis->RRing[NRCV-1]->Control |= DC_RCTL_RER;


    pThis->XmitRingBase = Lmem(pThis,HeapOffset);

    for ( i = 0 ; i < NXMT ; i++, HeapOffset += sizeof(DC_LIST_ENTRY)) {
        pDesc = (PDC_LIST_ENTRY)Hmem(pThis,HeapOffset);
        pThis->TRing[i] = pDesc;
        pThis->TBuf[i]  = Dc21x4DescAlloc (pThis, pDesc);
    }
    pThis->TRing[NXMT-1]->Control |= DC_TCTL_TER;

}

/*
 * Allocate and initialize a ring descriptor.
 * Allocates a buffer from the heap and writes a descriptor
 * for that buffer to the host virtual address PDESC.
 */
PRIVATE volatile PTR
Dc21x4DescAlloc (PDC21x4 pThis, PDC_LIST_ENTRY pDesc)
{
    UINT                dp;

    /*
     * Allocate buffer from heap
     */
    dp = DcMalloc(pThis, HEAP_BUFFER_SIZE);

    /*
     * Build a descriptor pointing to it 
     */
    pDesc->FirstBuffer = Lmem(pThis,dp);
    pDesc->SecondBuffer = 0;
    pDesc->Status   = 0;
    pDesc->Control = HEAP_BUFFER_SIZE;

    /*
     * Flush the descriptor to heap
     */
    IOWriteBufferFlush();

    return (volatile PTR) Hmem(pThis,dp);
}

/*
 * Allocate a chunk of heap memory, permanently (no free).
 * The return value is an offset into the heap, which can be
 * passed to Hmem() and Lmem() to get valid host and chip addresses.
 */
PRIVATE UINT
DcMalloc(PDC21x4 pThis, int size)
{
    UINT    ret;

    /*
     * Round to double longword boundary
     */
    if (pThis->TopAlloc & 0xf)
        pThis->TopAlloc = (pThis->TopAlloc + 0x10) & ~0xf;

    ret = pThis->TopAlloc;
    pThis->TopAlloc += size;

    return ret;
}

/* Read the ethernet address that was burned into the prom.
 */
PRIVATE void Dc21x4ReadPermanentAddress( PDC21x4 pThis, UINT8 *EthAddr )
{
    int i;
    INT32 v;

    Dc21x4WriteRegister( pThis, DC_ROM, 0);

    /* BUGBUG Should really read all 32 bytes and verify the checksum.
     */

    for (i = 0; i < sizeof(pThis->is_addr); i++) {

        /* Try each byte at most 50 msecs
         */
        int Timeo = 50;
        do {
            Delay( 1000 ); /* 1 millisec */
            v = Dc21x4ReadRegister( pThis, DC_ROM );
        } while ( (v < 0) && Timeo--);

        if (v < 0) {
            printf(_T("dc: Could not read PROM byte %d\n"), i);
            EthAddr[i] = 0xff;
        } else
            EthAddr[i] = (UINT8) v; /* lowbyte */
    }
}

/*++
 *  Functions in this section are used rarely, on resets.
 *++
 */

/*
 * Start/Restart interface
 *
 * This is where interrupts are turned (back) on.
 */
PRIVATE INT MCT Dc21x4Initialize(PINETDRIVER This)
{
    PDC21x4 pThis = pDC(This);
    PIENDPOINTFACTORY mbf = pThis->is_if.MBufFact;
    PDC_LIST_ENTRY    *rp;
    UINT         i;
    struct ifnet *ifp;

    /*
     * Lock out transmission
     */
    mbf->v->NetLock(mbf);

    /*
     * stop the chip, disable interrupts
     */
    Dc21x4Stop(pThis);

    /*
     * stop network activity 
     */
    ifp = &pThis->is_if;
    if (ifp->if_flags & IFF_RUNNING) {
        ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
    }
    pThis->rstrtcnt++;

    DBG((_T("dc: %d restarts\n"), pThis->rstrtcnt));

    /*
     * Init the buffer descriptors and indexes for each of the rings. 
     */
    for (i = 0, rp = pThis->RRing; i < NRCV; i++, rp++)
        (*rp)->Status = DC_STATUS_CHIP_OWNS_IT;
    
    for (i = 0, rp = pThis->TRing; i < NXMT; i++, rp++)
        (*rp)->Status = 0; /* host owns it */

    IOWriteBufferFlush();

    pThis->XmitterTail = pThis->XmitterHead = pThis->ReceiverHead = 0;

    /*
     * Start the Dc21x4 and enable interrupts 
     */
    Dc21x4SendSetupPacket(pThis);

    ifp->if_flags |= (IFF_UP | IFF_RUNNING);

    /*
     * Let xmitters proceed
     */
    mbf->v->NetUnlock(mbf);

    return 0;
}

/* Send the setup packet to the chip,
 * and wait until its done digesting it.
 */
PRIVATE void Dc21x4SendSetupPacket( PDC21x4 pThis)
{
    UINT     i, v;
    PDC_LIST_ENTRY pDesc;

    Dc21x4DisableInterrupts(pThis);

    Dc21x4WriteRegister(pThis, DC_BUS_MODE, DC_BUS_MODE_DEFAULT);

    /*
     * Tell the chip where the rings are
     */
    Dc21x4WriteRegister(pThis, DC_RCV_LIST_BASE, pThis->RecvRingBase);
    Dc21x4WriteRegister(pThis, DC_XMT_LIST_BASE, pThis->XmitRingBase);
    
    /*
     * Initialize the SIA registers
     */
    Dc21x4WriteRegister(pThis, DC_SIA_RESET, 0);
    Delay(10000);
    Dc21x4WriteRegister(pThis, DC_SIA_GENERAL, pThis->MediaType->Csr15Value);
    Dc21x4WriteRegister(pThis, DC_SIA_XMT_RCV, pThis->MediaType->Csr14Value);
    Dc21x4WriteRegister(pThis, DC_SIA_CONNECTIVITY, pThis->MediaType->Csr13Value);


    /*
     * Start the INIT sequence.  Pick a buffer, any buffer.
     * NOTE: No interrupt expected.
     */
    i = (pThis->XmitterHead++) & (NXMT - 1);
    pThis->XmitterTail++;

    pDesc = pThis->TRing[i];

    /* Fill in the filtering info
     */
    i = Dc21x4InstallEthernetAddress(pThis, (SETUP_PACKET_PERFECT*) pThis->TBuf[i] );

    /* Give descriptor away
     * NOTE: Spec sez FS&LS should be zero. But it works if 1,
     * and it makes a hell of a lot more sense to me.
     */
    pDesc->Control = i | DC_TCTL_SETUP | DC_TCTL_FS | DC_TCTL_LS;
    IOWriteBufferFlush();

    pDesc->Status = DC_STATUS_CHIP_OWNS_IT;
    IOWriteBufferFlush();

    /*
     * Transmitter was stopped, start it.
     */
    Dc21x4WriteRegister(pThis, DC_OPERATION_MODE, DC_OPER_BITS);

    /* tell xmitter to poll, now!
     */
    Dc21x4WriteRegister(pThis, DC_XMT_POLL, DC_POLL_IT);

    /*
     * Give it plenty of time to settle
     * [This should take less than 1 msec]
     */
    v = 0; /* quiet bogus compiler warning */
    for (i = 0; i < 1000; i++) {
        Delay(1000);
        v = Dc21x4ReadRegister(pThis,DC_STATUS);
        if (v & DC_STATUS_TU)
            break;
    }

    /*
     * Did it work.
     */
    if ((v & DC_STATUS_TU) == 0) {
        printf(_T("dc: cannot initialize. Status x%x\n"), v);
#if _DEBUG
        DcDump(pThis);
#endif
        Dc21x4Stop(pThis);
        return;
    }

    /* Start receiver
     */
    Dc21x4WriteRegister(pThis, DC_OPERATION_MODE, DC_OPER_BITS | DC_OPER_SR);

    /*
     * Enable interrupts.
     */
    Dc21x4EnableInterrupts(pThis);
}

/* Initialize the address filter data
 */
PRIVATE UINT Dc21x4InstallEthernetAddress( PDC21x4 pThis, SETUP_PACKET_PERFECT *pPkt )
{
    UINT16 *EnAddr;
    int i;

    EnAddr = (UINT16*)pThis->is_addr;

    /* Broadcast address match
     */
    pPkt->EthernetAddresses[0][0] = 0xffff;
    pPkt->EthernetAddresses[0][1] = 0xffff;
    pPkt->EthernetAddresses[0][2] = 0xffff;

    /* Everywhere else, our own address
     */
    for (i = 1; i < DC_PERFECT_ENTRIES; i++) {

        pPkt->EthernetAddresses[i][0] = EnAddr[0];
        pPkt->EthernetAddresses[i][1] = EnAddr[1];
        pPkt->EthernetAddresses[i][2] = EnAddr[2];

    }

    return sizeof(*pPkt);
}

/*
 * Dynamic control over Dc21x4 behaviour
 */
PRIVATE INT MCT
Dc21x4Ioctl(PINETDRIVER This, UINT32 Prop, PTR Data)
{
    PDC21x4 pThis = pDC(This);
#if _DEBUG
    DcDump(pThis);
#endif
    /* We care for requests to turn the interface down/up.
     * When the system goes down we must be stopped, or
     * else we get the firmware in trouble.
     */
    if (Prop == SIOCSIFFLAGS) {

        if ((pThis->is_if.if_flags & IFF_UP) == 0)
            /*
             * Down. Take the easy way out: just disable interrupts.
             * If/when we get back up we'll just handle the errors.
             */
            Dc21x4DisableInterrupts(pThis);
        else
            /*
             * Up. Strictly speaking, needs protection.
             */
            Dc21x4EnableInterrupts(pThis);
    }
    return NO_ERROR;
}

/*++
 *  Functions in this section are high-runners.
 *++
 */

/*
 * Interrupt service routine.
 * Return TRUE if we do have an interrupt pending,
 * after telling the chip to keep quiet.
 */
PRIVATE BOOL MCT Dc21x4Isr(PINETDRIVER This, PBOOL pNotMyInterrupt)
{
    PDC21x4 pThis = pDC(This);
    UINT Status;

#if _DEBUG
    pThis->numintr++;
#endif

    Status = Dc21x4ReadRegister(pThis,DC_STATUS);

    if ((Status & (DC_STATUS_NIS | DC_STATUS_AIS)) == 0) {
#if _DEBUG
        pThis->spurious++;
#endif
        if (pNotMyInterrupt)
            *pNotMyInterrupt = TRUE;
        return FALSE;
    }

    pThis->IsrStatus |= Status;

    /* Ack */
    Dc21x4WriteRegister(pThis, DC_STATUS, Status);
    IOWriteBufferFlush();
    

    /* Wakeup handler thread and we are done.
     */
    return (*pThis->Signal)( &pThis->is_if.IntrPoint );
}

/*
 * Thread-level interrupt handling
 */
PRIVATE BOOL MCT Dc21x4HandleInterrupt(PINETDRIVER This)
{
    PDC21x4 pThis = pDC(This);
    UINT        Status;

 again:
    Status = pThis->IsrStatus;

#if _DEBUG
    pThis->sawcsr = Status;
#endif

    /*
     * Check for errors first
     */
    if ( Status & (0xb3a8) ) {
        DBG((_T("dc: ErrIntr x%x\n"), Status));

        /*
         * Missed packets (because no descs available)
         */
        if (Status & DC_STATUS_RU) {
            UINT32 nMisses = Dc21x4ReadRegister(pThis,DC_MISS_COUNTER);
            nMisses &= (DC_MISS_MASK | DC_MISS_OVF);
            pThis->misscnt += nMisses;
        }

        /*
         * Handle xmit/recv timeouts
         */
        if (Status & (DC_STATUS_TJT|DC_STATUS_RWT)) {
            pThis->bablcnt++;
            DBG((_T("dc: xmt/rcv timeout (%d)\n"), pThis->bablcnt));
        }

        /*
         * Handle DMA errors
         */
        if (Status & DC_STATUS_UNF) {
            pThis->merrcnt++;
            DBG((_T("dc: memory error (%d)\n"), pThis->merrcnt));
            
        }

        /*
         * Handle transceiver failures
         */
        if (Status & DC_STATUS_LNF) {
            pThis->sqercnt++;
            DBG((_T("dc: transceiver error (%d)"), pThis->sqercnt));
        }

        /*
         * Errors for which we must reset the chip
         */
        if ((Status & (DC_STATUS_SE | DC_STATUS_EB)) ==
                      (DC_STATUS_SE | DC_STATUS_PARITY)) {
            Dc21x4Initialize(iDC(pThis));
            return FALSE;
        }
    }

    if ( Status & DC_STATUS_RI )
        Dc21x4RecvInterrupt( pThis );

    if ( Status & DC_STATUS_TI )
        Dc21x4XmitInterrupt(iDC(pThis));

    /* See if another interrupt showed up
     */
    if (!AtomicCmpAndSwap( &pThis->IsrStatus, Status, 0))
        goto again;

    return TRUE;

}
 
/*
 * Handle a receiver complete interrupt.
 */
PRIVATE void Dc21x4RecvInterrupt(PDC21x4 pThis)
{
    int Current, Length;
    UINT Status;
    volatile PDC_LIST_ENTRY  pDesc;
    struct ether_header eh;
    PIBUFFER m;
    PIENDPOINTFACTORY mbf = pThis->is_if.MBufFact;

    mbf->v->NetLock(mbf);

    /*
     * Starting from where we left off, look around the receive ring and
     * pass on all complete packets. 
     */

    for (;;) {

        /*
         * Look at the current descriptor 
         */
CheckDescriptor:
        Current = pThis->ReceiverHead;
        pDesc = pThis->RRing[Current];
        Status = pDesc->Status;

        if (Status & DC_STATUS_CHIP_OWNS_IT)
            break;

/* BUGBUG error handling needs revision */
        /*
         * If not a whole packet, error 
         */
        if ((Status & (DC_RSTATUS_FS | DC_RSTATUS_LS)) !=
                      (DC_RSTATUS_FS | DC_RSTATUS_LS)) {
            DBG((_T("dc: Rring #%d, Status=%x !STP\n"), Current, Status));
            goto BadPacket;
        }
        /*
         * See if packet is chained (should not)
         */
        if (pDesc->Control & DC_RCTL_RCH) {
            DBG((_T("dc: Rring #%d, Status=%x CHAIN\n"), Current, Status));
            goto BadPacket;
        }
            
        Length = (Status & DC_RSTATUS_FL) >> DC_RSTATUS_FL_SHIFT;

        if (Length == 0) {
            /* Possible race between memory and interrupt lines.
             * Chip has signalled intr but the size update has not
             * made it to main memory yet.
             */
            goto CheckDescriptor;
        }

        /*
         * Drop header/trailing bytes from Length and copy packet
         */
        m = Dc21x4RecvPkt(pThis, pThis->RBuf[Current],
                          Length-4-sizeof(struct ether_header), &eh);

        /*
         * Return descriptor to chip, and move on to next
         * packet 
         */
    ReleaseIt:
        pDesc->Status = DC_STATUS_CHIP_OWNS_IT;
        IOWriteBufferFlush();

        pThis->ReceiverHead = ++Current & (NRCV - 1);

        if (m) {
            /*
             * Hand the packet to the network module.
             * NOTE: it is possible that the upper layer
             * will bite back and reset the interface.
             */
            pThis->is_if.if_ipackets++;
            mbf->v->EtherInput(mbf, (PINETDRIVER) &pThis->is_if, &eh, m);
        }
        continue;

        /*
         * Not a good packet, see what is wrong 
         */
    BadPacket:
        pThis->is_if.if_ierrors++;

        DBG((_T("dc: rcv error (x%x)\n"), Status));

        /*
         * Return descriptor to chip 
         */
        m = NULL;
        goto ReleaseIt;
    }

    mbf->v->NetUnlock(mbf);
}

/*
 * Pull a packet off the interface and
 * hand it up to the higher levels.
 */
PRIVATE PIBUFFER
Dc21x4RecvPkt(PDC21x4 pThis, volatile UINT8 *RBuf, int len, struct ether_header *eh)
{
    SCODE sc;
    struct ifnet *ifp = &pThis->is_if;
    PIBUFFER mh;
    PIENDPOINTFACTORY mbf = ifp->MBufFact;
    UINT16 ether_type;
    UINT wlen;

    if (len <= 0)
        return NULL;    /* sanity */

    sc = mbf->v->AllocateBuffer(mbf, 0, (PINETDRIVER) ifp, (PIBUFFER *) &mh);
    if (FAILED(sc)) {
        /*
         * No room, drop the packet
         */
        ifp->if_iqdrops++;
        return NULL;
    }

    /*
     * Read in ethernet header 
     */
    memcpy( (PTR)eh, (PTR)RBuf, sizeof(struct ether_header));
    RBuf += sizeof(struct ether_header);

    ether_type = ntohs(eh->ether_type);

    /* We dont do trailers
     */
    if (ether_type >= ETHERTYPE_TRAIL &&
        ether_type < ETHERTYPE_TRAIL+ETHERTYPE_NTRAILER)
        goto ErrorExit;

    eh->ether_type = ether_type;

    sc = mh->v->WriteAt(mh, UINT64_ZERO, (PBYTE) RBuf, len, &wlen);
    if (FAILED(sc))
        goto ErrorExit;

    ifp->if_ipackets++;
    return mh;

 ErrorExit:
    ifp->if_ierrors++;
    mh->v->Release(mh);
    return NULL;
}

/*
 * Handle a transmitter complete interrupt.
 */
PRIVATE void Dc21x4XmitInterrupt(PINETDRIVER This)
{
    PDC21x4 pThis = pDC(This);
    UINT Status;
    BOOL DidSome = FALSE;
    PDC_LIST_ENTRY pDesc;

    /*
     * Free up descriptors for all packets in queue for which
     * transmission is complete.  Start from queue tail, stop at first
     * descriptor we do not OWN, or which is in an inconsistent state
     * (chip still working). 
     */

    while (pThis->XmitterTail != pThis->XmitterHead) {

        pDesc = pThis->TRing[pThis->XmitterTail & (NXMT - 1)];
        Status = pDesc->Status;

        /*
         * Does the chip still own it ? 
         */
        if (Status & DC_STATUS_CHIP_OWNS_IT)
            break;

        pThis->XmitterTail++;
        DidSome = TRUE;

        pThis->is_if.if_opackets++;

        /*
         * Check for collisions
         */
        pThis->is_if.if_collisions += (Status & DC_TSTATUS_CC) >> DC_TSTATUS_CC_SHIFT;

        /*
         * Check for transmission errors. 
         */
        if (Status & DC_TSTATUS_ES) {

            if (Status & (DC_TSTATUS_EC | DC_TSTATUS_LC))
                pThis->is_if.if_collisions++;

            pThis->is_if.if_oerrors++;
            DBG((_T("dc: xmt error (x%x x%x)\n"), Status, pDesc->Control));

            /*
             * Restart chip on errors that disable the
             * transmit process (BUGBUG overkill?)
             */
            if (Status & DC_TSTATUS_TO) {
                Dc21x4Initialize(iDC(pThis));
                return;
            }
        }
    }
    /*
     * Dequeue next transmit request, if any. 
     */
    if (DidSome)
        Dc21x4Start(iDC(pThis));
}
 
/*
 * Move a packet into the shared heap.
 */
PRIVATE UINT
Dc21x4SendPkt(
        struct mbuf  *m,
        volatile UINT8 *SBuf)
{
    int    Length = 0;

    do {
        memcpy( (PTR)(SBuf + Length), mtod(m,PTR), m->m_len);

        Length += m->m_len;
        m = m->m_next;

    } while (m);

    /* Minimum/Maximum lengths
     */
    if (Length < 60)
        Length = 60;

    return Length & DC_TCTL_TBS1;
}

/*
 * Start output on interface.
 */
PRIVATE void MCT
Dc21x4Start(PINETDRIVER This)
{
    PDC21x4 pThis = pDC(This);
    struct mbuf     *Request;
    PDC_LIST_ENTRY     pDesc;
    UINT            PktLength, nQueued, Current, Intr;
    PIENDPOINTFACTORY mbf = pThis->is_if.MBufFact;

    mbf->v->NetLock(mbf);

    nQueued = 0;
    Intr = DC_TCTL_IC;

    while ((pThis->XmitterHead - pThis->XmitterTail) < NXMT)
    {
        /*
         * Dequeue the next transmit request, if any. 
         */
        IF_DEQUEUE(&pThis->is_if.if_snd, Request);
        if (Request == NULL)
            break;      /* Nothing (else) on the queue   */

        Current = (pThis->XmitterHead++) & (NXMT - 1);

        PktLength = Dc21x4SendPkt(Request, pThis->TBuf[Current]);

        /*
         * Give away buffer. We do not do chaining.
         */
        pDesc = pThis->TRing[Current];

        pDesc->Control = (pDesc->Control & DC_TCTL_TER) | Intr |
            PktLength | (DC_TCTL_FS | DC_TCTL_LS);

        pDesc->Status = DC_STATUS_CHIP_OWNS_IT;
        IOWriteBufferFlush();

        if (nQueued++ == 0) {
            /*
             * Tell the chip to rescan the descriptors _now_.
             * While sending full-blast biggies we might gain up to
             * 2 msec latency kicking the chip here instead of
             * doing it at the end of this loop.
             */
            Dc21x4WriteRegister(pThis, DC_XMT_POLL, DC_POLL_IT);
            Intr = 0;
        }

        Request->m_hdr.v->Release((PIBUFFER) Request);
    }

    mbf->v->NetUnlock(mbf);
}


#if _DEBUG

PRIVATE void DcDump( PDC21x4 pThis )
{
    IO_PORT          i;
    static char *PortName[16] = {
            "BUS MODE              ", /*        0 */
            "TXM POLL DEMAND       ", /*        1 */
            "RCV POLL DEMAND       ", /*        2 */
            "RCV DESC RING         ", /*        3 */
            "TXM DESC RING         ", /*        4 */
            "STATUS                ", /*        5 */
            "OPERATION MODE        ", /*        6 */
            "INTERRUPT MASK        ", /*        7 */
            "MISSED FRAME          ", /*        8 */
            "IDPROM                ", /*        9 */
            "RESERVED              ", /*       10 */
            "TIMER                 ", /*       11 */
            "SIA STATUS/GEN PURPOSE", /*       12 */
            "SIA MODE-0            ", /*       13 */
            "SIA MODE-1            ", /*       14 */
            "SIA MODE-2            "  /*       15 */
        };
#define NPRT 11
    static UINT8 Who[NPRT] = { 0, 3, 4, 5, 6, 7, 8, 12, 13, 14, 15 };

    printf(_T("DC21x4: register dump at PortBase %x\n"), pThis->PortBase);
    for (i = 0; i < NPRT; i++) {
        printf(_T("%hs %08x  "),
               PortName[Who[i]],
               Dc21x4ReadRegister( pThis, DC_REGISTER(Who[i]) ) );
        i++;
        printf(_T("%hs %08x\n"),
               PortName[Who[i]],
               Dc21x4ReadRegister( pThis, DC_REGISTER(Who[i]) ) );
    }

{
    UINT *p;
    p = (UINT*)pThis->RRing[0];
    if (p)
    for (i = 0; i < 4; i++) {
        printf(_T("%x %x %x %x   %x %x %x %x\n"),
               p[0], p[1], p[2], p[3],
               p[4], p[5], p[6], p[7]);
        p += 8;
    }
    p = (UINT*)pThis->TRing[0];
    if (p)
    for (i = 0; i < 4; i++) {
        printf(_T("%x %x %x %x   %x %x %x %x\n"),
               p[0], p[1], p[2], p[3],
               p[4], p[5], p[6], p[7]);
        p += 8;
    }
}
}

#endif

SCODE __cdecl Dc21x4Constructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        const _TCHAR * a1,
        const _TCHAR * a2,
        UINT Media,
        ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(a1);
    UnusedParameter(a2);

    DBG(("Dc21x4Constructor(%x %s %x %x %x media=%x)\n", pThis, Name, RegisterBase, Interrupt, ppUnk, Media));

    pUnk = (PIUNKNOWN) Dc21x4New(&RegisterBase, &Interrupt, &Media);

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBG(("Dc21x4Constructor failed\n"));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl Dc21x4DriverCobVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) Dc21x4Constructor
};

static struct _DRIVERCOB Dc21x4DriverCobObject = { &Dc21x4DriverCobVtbl, 1 };

PIUNKNOWN Dc21x4CobMain(void)
{
    SCODE sc;
    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &Dc21x4DriverCobVtbl,
                          "CDeviceConstructor",
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &Dc21x4DriverVtbl, "CNetDriver",
                          &IID_INetDriver);
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &Dc21x4DriverCobObject;
}
