/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    dc21x4.h
 *
 * Purpose:
 *
 *    Definitions for the DEC 21040 PCI Ethernet
 *
 * Author:
 *    A. Forin (sandrof), using information from:
 *      "DECchip 21040 Ethernet LAN Controller for PCI
 *       Hardware Reference Manual"
 *      EC-N0752-72, Digital Semiconductor. May 1994.
 *
 */

/*
 * CONFIGURATION SPACE
 */

/*
 * For documentation purposes.
 * Variations over PCI Specs 2.0 of standard configuration registers
 *
 * Command and Status, offset 0x04:
 *      Implements bits 0-2,6,8,23-26,28-31.
 *      DEVSEL is 01--medium, back-to-back is true.
 *      No additional bits.
 *
 * Device Identification, offset 0x8:
 *      Programming Interface is undefined (0)
 *
 * Miscellaneous, offset 0x0c:
 *      Implements the Latency Timer (at +1)
 *      Does not implement CacheLineSize, HeaderType, BIST
 *
 * Interrupt Configuration, offset 0x3c:
 *      Implements Interrupt Pin and Line.
 *      Does not implement Max_Lat, Min_Gnt.
 */

/*
 * Base Address registers
 */

#define DC_CNF_CBIO     0x10    /* Base I/O address (bit0 is 1) */
# define DC_CBIO_MASK   0xffffff80

#define DC_CNF_CBMA     0x14    /* Base memory address (bit0 is 0) */
# define DC_CBMA_MASK   0xffffff80

/*
 * Chip-specific configuration registers
 */
#define DC_CNF_CFDA     0x40    /* Driver scratch */
# define DC_CFDA_MASK   0x0000ff00


/*
 * I/O SPACE
 * [possibly memory-mapped, longword access only]
 *
 * Legend
 *      r       read access
 *      w       write access
 *      c       write-1-to-clear
 *      R       read-to-clear
 *      lw      longword (32 bits)
 */
#define DC_REGISTER(_n_)        ((_n_)*8)

/*
 * Bus Mode Register, defines the bus operating modes.
 */
#define DC_BUS_MODE             DC_REGISTER( 0)

#define DC_BUS_SWR      0x00000001   /* rw Software Reset */

#define DC_BUS_BAR      0x00000002   /* rw Bus Arbitration */
# define DC_BUS_BAR_RROBIN 0x00000002   /* else Rcv has priority */

#define DC_BUS_DSL      0x0000007c   /* rw Descrip Skip Length */
# define DC_BUS_DSL_SHIFT 2             /* # of lw between descriptors */

#define DC_BUS_BLE      0x00000080   /* rw Big/Little Endian data (1==Big) */

#define DC_BUS_PBL      0x00003f00   /* rw Max Burst Length */
# define DC_BUS_PBL_SHIFT 8             /* # of lw in max DMA len */
                                        /* if 0, min 16 max 64 (FIFO len) */

#define DC_BUS_CAL      0x0000c000   /* rw Cache Alignment */
# define DC_BUS_CAL_xxx  0x00000000     /* unused */
# define DC_BUS_CAL_8lw  0x00004000     /* align to 8 lw */
# define DC_BUS_CAL_16lw 0x00008000
# define DC_BUS_CAL_32lw 0x0000c000

#define DC_BUS_DAS      0x00010000   /* rw Diagnostic Address Space */

#define DC_BUS_TAP      0x00060000   /* rw Automatic Xmt Polling */
# define DC_BUS_TAP_NONE 0x00000000
# define DC_BUS_TAP_200u 0x00020000
# define DC_BUS_TAP_800u 0x00040000
# define DC_BUS_TAP_1600u 0x00060000

/*
 * Demand Transmit Poll
 * Asks chip to poll transmit descriptors.
 * Transmit process must be suspended, else ignored.
 */
#define DC_XMT_POLL             DC_REGISTER( 1)

/* Any value would do.. */
#define DC_POLL_IT      0x00000001   /*  w both xmt and recv */

/*
 * Demand Receive Poll
 * Asks chip to poll receive descriptors.
 * Receive process must be suspended, else ignored.
 */
#define DC_RCV_POLL             DC_REGISTER( 2)

/*
 * Receive Descriptor List Address
 * Must be lw aligned or else.
 * Only write when receive process stopped,
 * and then before start command.
 */
#define DC_RCV_LIST_BASE        DC_REGISTER( 3)
                                     /* rw Phys Address of Start of List */

/*
 * Transmit Descriptor List Address
 * Must be lw aligned or else.
 * Only write when transmit process stopped,
 * and then before start command.
 */
#define DC_XMT_LIST_BASE        DC_REGISTER( 4)
                                     /* rw Phys Address of Start of List */

/*
 * Interrupt (and) status register
 */
#define DC_STATUS               DC_REGISTER( 5)

#define DC_STATUS_TI    0x00000001   /* rc Transmit Interrupt */
#define DC_STATUS_TPS   0x00000002   /* rc Transmit Process Stopped */
#define DC_STATUS_TU    0x00000004   /* rc Xmt Buffer Unavailable */
#define DC_STATUS_TJT   0x00000008   /* rc Xmt Jabber Timeout */
#define DC_STATUS_UNF   0x00000020   /* rc Transmit Underflow (late DMA) */
#define DC_STATUS_RI    0x00000040   /* rc Receive Interrupt */
#define DC_STATUS_RU    0x00000080   /* rc Recv Buffer Unavailable */
#define DC_STATUS_RPS   0x00000100   /* rc Recv Process Stopped */
#define DC_STATUS_RWT   0x00000200   /* rc Rcv Watchdog Timeout (Babble) */
#define DC_STATUS_AT    0x00000400   /* rc AUI/TP pin toggled */
#define DC_STATUS_FD    0x00000800   /* rc Full-Duplex autoconf frame (1of2) */
#define DC_STATUS_LNF   0x00001000   /* rc Link Fail (see CSR12<2>)*/
#define DC_STATUS_SE    0x00002000   /* rc System Error (see EB field) */

#define DC_STATUS_AIS   0x00008000   /* rc Abnormal Interrupt Summary */
# define DC_INT_ABNORMAL 0x00003faa

#define DC_STATUS_NIS   0x00010000   /* rc Normal Interrupt Summary */
# define DC_INT_NORMAL   0x00000045

#define DC_STATUS_RS    0x000e0000   /* r  Receive Process State  */
# define DC_RS_STOPPED  0x00000000      /* reset or stop receive */
# define DC_RS_FETCH    0x00020000      /* ..a descriptor */
# define DC_RS_CHECK    0x00040000      /* ..for EOP before prefetch */
# define DC_RS_WAIT     0x00060000      /* ..for packet */
# define DC_RS_SUSPEND  0x00080000      /* no buffers */
# define DC_RS_CLOSING  0x000a0000      /* ..a recv desc */
# define DC_RS_FLUSH    0x000c0000      /* discarding packet fron FIFO */
# define DC_RS_WRITE    0x000e0000      /* ..data to memory */

#define DC_STATUS_TS    0x00700000   /* r  Transmit Process State */
# define DC_TS_STOPPED  0x00000000      /* reset or xmt jabber */
# define DC_TS_FETCH    0x00100000      /* ..a descriptor */
# define DC_TS_WAIT     0x00200000      /* ..for end of xmit */
# define DC_TS_READ     0x00300000      /* ..data from memory */
# define DC_TS_SETUP    0x00500000      /* got setup packet */
# define DC_TS_SUSPEND  0x00600000      /* FIFO underflow or no xmit descs */
# define DC_TS_CLOSING  0x00700000      /* ..a xmit desc */

#define DC_STATUS_EB    0x03800000   /* r  Error Bits */
# define DC_STATUS_PARITY  0x00000000   /* -> software reset */
# define DC_STATUS_M_ABORT 0x00800000   /* -> clear DC_STATUS_SE */
# define DC_STATUS_T_ABORT 0x01000000   /* -> clear DC_STATUS_SE */

/*
 * Operation Mode Register
 * Should be written last upon initialization
 */
#define DC_OPERATION_MODE       DC_REGISTER( 6)

#define DC_OPER_HP      0x00000001   /* r  Hash(1)/Perfect(0) Filtering */
#define DC_OPER_SR      0x00000002   /* rw Start Receiver process */
#define DC_OPER_HO      0x00000004   /* r  Hash-Only filtering */
#define DC_OPER_PB      0x00000008   /* rw Pass Bad packets */
#define DC_OPER_IF      0x00000010   /* r  Inverse Filtering mode */
#define DC_OPER_SB      0x00000020   /* rw Start Backoff Counter */
#define DC_OPER_PR      0x00000040   /* rw Promiscuous Mode */
#define DC_OPER_PM      0x00000080   /* rw Pass all Multicast */
#define DC_OPER_FKD     0x00000100   /* rw Flaky oscill Disabled */
#define DC_OPER_FD      0x00000200   /* rw Full-Duplex mode */

#define DC_OPER_OM      0x00000c00   /* rw Operating Mode (Loopbacks) */
# define DC_OPER_NORMAL  0x00000000
# define DC_OPER_TP_LB   0x00000800     /* see CSR14<1,8> */
# define DC_OPER_AUI_LB  0x00000800
# define DC_OPER_SIA_LB  0x00000800
# define DC_OPER_INTER_LB 0x00000400

#define DC_OPER_FC      0x00001000   /* rw Force Collision */
#define DC_OPER_ST      0x00002000   /* rw Start Transmit process */

#define DC_OPER_TR      0x0000c000   /* rw Transmit thReshold */
# define DC_TTHR_72      0x00000000     /* # bytes in FIFO before xmit */
# define DC_TTHR_96      0x00004000
# define DC_TTHR_128     0x00008000
# define DC_TTHR_160     0x0000c000

#define DC_OPER_BP      0x00010000   /* rw Back Pressure */
#define DC_OPER_CA      0x00020000   /* rw Capture Effect Avoidance */

/*
 * Interrupt Mask Register
 */
#define DC_INTERRUPT_MASK       DC_REGISTER( 7)
# define DC_INTERRUPTS  0x0001bfef      /* rw  See DC_STATUS_* bits */

/*
 * Missed Frame Counter
 */
#define DC_MISS_COUNTER         DC_REGISTER( 8)

#define DC_MISS_MASK    0x0000ffff   /* rR  missed since last read */
#define DC_MISS_OVF     0x00010000   /* rR  missed too many */

/*
 * Ethernet address ROM
 * Write any value to reset pointer
 */
#define DC_ROM                  DC_REGISTER( 9)

#define DC_ROM_DATA     0x000000ff   /* rw  byte from ROM */
#define DC_ROM_INVALID  0x80000000   /* rw  no data yet */

/*
 * Reserved
 */
#define DC_xxx                  DC_REGISTER(10)

/*
 * Full-Duplex autoconfiguration value
 */
#define DC_FULL_DUPLEX          DC_REGISTER(11)

/*
 * Serial Interface Attachment (SIA) Status register
 */
#define DC_SIA_STATUS           DC_REGISTER(12)

#define DC_SIA0_PAUI    0x00000001   /* r  AUI/TP Pin status */
#define DC_SIA0_NCR     0x00000002   /* r  Network Connection eRror */
#define DC_SIA0_LKF     0x00000004   /* r  Link Fail status */
#define DC_SIA0_APS     0x00000008   /* r  Auto Polarity State */
#define DC_SIA0_DSD     0x00000010   /* r  PLL Selftest Done */
#define DC_SIA0_DSP     0x00000020   /* r  PLL Selftest Pass */
#define DC_SIA0_DAZ     0x00000040   /* r  PLL synch on all-zero */
#define DC_SIA0_DAO     0x00000080   /* r  PLL synch on all-ones */


/*
 * SIA Connectivity and reset register
 * All bits are for diagnostics, except low 4.
 */
#define DC_SIA_CONNECTIVITY     DC_REGISTER(13)
#define DC_SIA_RESET            DC_REGISTER(13)

#define DC_SIA1_SRL     0x00000001   /* rw  SIA (not)reset */
#define DC_SIA1_PS      0x00000002   /* rw  Pin Select (of AUI/TP) */
#define DC_SIA1_CAC     0x00000004   /* rw  CSR AutoConfig */
#define DC_SIA1_AUI     0x00000008   /* rw  AUI vz 10base-t */
#define DC_SIA1_EDP     0x00000010   /* rw  SIA PLL External Input Enable */
#define DC_SIA1_ENI     0x00000020   /* rw  Encoder Input Multiplexer */
#define DC_SIA1_SIM     0x00000040   /* rw  Serial Input Mux */
#define DC_SIA1_ASE     0x00000080   /* rw  APLL Start Enable */
#define DC_SIA1_SEL     0x00000f00   /* rw  External Port Output Mux Select */
#define DC_SIA1_IE      0x00001000   /* rw  Input Enable */
#define DC_SIA1_OE13    0x00002000   /* rw  Output Enable Pins 1 3 */
#define DC_SIA1_OE24    0x00004000   /* rw  Output Enable Pins 2 4 */
#define DC_SIA1_OE57    0x00008000   /* rw  Output Enable Pins 5 6 7 */

/*
 * SIA Transmit and Receive register (diagnostics)
 */
#define DC_SIA_XMT_RCV          DC_REGISTER(14)

#define DC_SIA2_ECEN    0x00000001   /* rw  Encoder Enable */
#define DC_SIA2_LBK     0x00000002   /* rw  Loopback Enable */
#define DC_SIA2_DREN    0x00000004   /* rw  Driver Enable */
#define DC_SIA2_LSE     0x00000008   /* rw  Link Pulse Send Enable */

#define DC_SIA2_CPEN    0x00000030   /* rw  Compensation Enable */
# define DC_SIA2_CPEN_HIGH 0x00000020   /* high power mode */
# define DC_SIA2_CPEN_NORM 0x00000030   /* normal compensation */

#define DC_SIA2_RSQ     0x00000100   /* rw  Receive Squelch Enable */
#define DC_SIA2_CSQ     0x00000200   /* rw  Collision Squelch Enable */
#define DC_SIA2_CLD     0x00000400   /* rw  Collision Detect Enable */
#define DC_SIA2_SQE     0x00000800   /* rw  Signal Quality generate Enable */
#define DC_SIA2_LTE     0x00001000   /* rw  Link Test Enable */
#define DC_SIA2_APE     0x00002000   /* rw  Auto Polarity Enable */
#define DC_SIA2_SPP     0x00004000   /* rw  Set Polarity Plus */
#define DC_SIA2_xxx     0x00008000   /* rw  ?omission from manual? */

/*
 * SIA General purpose register
 */
#define DC_SIA_GENERAL          DC_REGISTER(15)

#define DC_SIA3_JBD     0x00000001   /* rw  Jabber Disable */
#define DC_SIA3_HUJ     0x00000002   /* rw  Host Unjab */
#define DC_SIA3_JCK     0x00000004   /* rw  Jabber Clock */
#define DC_SIA3_RWD     0x00000010   /* rw  Receive Watchdog Disable */
#define DC_SIA3_RWR     0x00000020   /* rw  Receive Watchdog Release */
#define DC_SIA3_TSCK    0x00000100   /* rw  Test Clock */
#define DC_SIA3_FUSQ    0x00000200   /* rw  Force Unsquelch */
#define DC_SIA3_FLF     0x00000400   /* rw  Force Link Fail */
#define DC_SIA3_DPST    0x00001000   /* rw  PLL Selftest Start */
#define DC_SIA3_FRL     0x00002000   /* rw  Force receiver low */

/*
 * MEMORY SPACE
 */

/*
 * Receive and Transmit descriptors look the same
 */
typedef struct {
    UINT32      Status;
    UINT32      Control;
    UINT32      FirstBuffer;
    UINT32      SecondBuffer;
} DC_LIST_ENTRY, *PDC_LIST_ENTRY;

/*
 * Status bits for a Receive Descriptor
 */
#define DC_RSTATUS_OF   0x00000001   /* FIFO Overflow (late DMA) */
#define DC_RSTATUS_CE   0x00000002   /* CRC error */
#define DC_RSTATUS_DB   0x00000004   /* Dribbling Bit */
#define DC_RSTATUS_RJ   0x00000010   /* Receive Watchdog */
#define DC_RSTATUS_FT   0x00000020   /* Trailer Packet */
#define DC_RSTATUS_CS   0x00000040   /* Late Collision Seen */
#define DC_RSTATUS_TL   0x00000080   /* Too Long */
#define DC_RSTATUS_LS   0x00000100   /* Last Segment */
#define DC_RSTATUS_FS   0x00000200   /* First Segment */
#define DC_RSTATUS_MF   0x00000400   /* Multicast Frame */
#define DC_RSTATUS_RF   0x00000800   /* Runt Frame */
#define DC_RSTATUS_DT   0x00003000   /* Loopback frame if nonzero */
#define DC_RSTATUS_LE   0x00004000   /* Length Error */

#define DC_RSTATUS_ES   0x00008000   /* Summary: OF CE CS TL RF LE */
# define DC_RSTATUS_ERRORS      0x000048c3

#define DC_RSTATUS_FL   0x7fff0000   /* */
# define DC_RSTATUS_FL_SHIFT    16

#define DC_STATUS_CHIP_OWNS_IT  0x80000000   /* either rcv or xmt */

/*
 * Control bits for a Receive Descriptor
 */
#define DC_RCTL_RBS1    0x000007ff   /* Length of 1st buffer */
#define DC_RCTL_RBS2    0x003ff800   /* Length of 2nd buffer */
#define DC_RCTL_RCH     0x01000000   /* Second address chained */
#define DC_RCTL_RER     0x02000000   /* Receive end of ring */

/*
 * Status bits for a transmit descriptor
 */
#define DC_TSTATUS_DE   0x00000001   /* Deferred xmission */
#define DC_TSTATUS_UF   0x00000002   /* Underflow Error (late DMA) */
#define DC_TSTATUS_LF   0x00000004   /* Link Fail */

#define DC_TSTATUS_CC   0x00000078   /* Collision Count */
# define DC_TSTATUS_CC_SHIFT 3

#define DC_TSTATUS_HF   0x00000080   /* Hartbeat Fail */
#define DC_TSTATUS_EC   0x00000100   /* Excessive Collisions */
#define DC_TSTATUS_LC   0x00000200   /* Late Collision */
#define DC_TSTATUS_NC   0x00000400   /* No Carrier */
#define DC_TSTATUS_LO   0x00000800   /* Lost Carrier */
#define DC_TSTATUS_TO   0x00004000   /* Jabber Timeout */
#define DC_TSTATUS_ES   0x00008000   /* Summary: UF LF EC LC NC LO TO */

/*
 * Control bits for a transmit descriptor
 */
#define DC_TCTL_TBS1    0x000007ff   /* 1st Buffer Length */
#define DC_TCTL_TBS2    0x003ff800   /* 2nd Buffer Length */

#define DC_TCTL_FT      0x10400000   /* Filtering type mask: */
# define DC_PERFECT      0x00000000
# define DC_HASH         0x00400000
# define DC_INVERSE      0x10000000
# define DC_HASH_ONLY    0x10400000

#define DC_TCTL_DPD     0x00800000   /* Disable Padding */
#define DC_TCTL_TCH     0x01000000   /* Second Address Chained */
#define DC_TCTL_TER     0x02000000   /* Transmit End of Ring */
#define DC_TCTL_AC      0x04000000   /* Add CRC Disable */
#define DC_TCTL_SETUP   0x08000000   /* Setup Packet */
#define DC_TCTL_FS      0x20000000   /* First Segment */
#define DC_TCTL_LS      0x40000000   /* Last Segment */
#define DC_TCTL_IC      0x80000000   /* Interrupt on Completion */

/* Setup packet, perfect filtering
 */
#define DC_PERFECT_ENTRIES 16

typedef struct {
    UINT32      EthernetAddresses[DC_PERFECT_ENTRIES][3];
} SETUP_PACKET_PERFECT;


/* Setup packet, imperfect filtering
 */

typedef struct {
    UINT32      Filter[32];
    UINT32      xxx1[7];
    UINT32      EthernetAddress[3];
    UINT32      xxx2[6];
} SETUP_PACKET_HASH;
