/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Driver for the 8255x Ethernet COntroller.
 * Applicable to the 82551/0/9/8/7/2 chips, according to Intel manuals.
 */
/* TODO:
 * - Untested codepath in Xmit when more than 4 frags
 * - If the PCI.CacheLineSize register is wrong should we fix it?
 */

/*#define _DEBUG 1*/
#if _DEBUG
#define DBGLVL 2 /* 0 noise, 1 - comment, 2 - major function (def DbgLvl == 3)*/
#endif

#define USE_INTRINSICS
#define DEVICE_ORDER LITTLE_ENDIAN
#include <mmlite.h>
#include <mmmacros.h>
#include <mmhal.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <sock.h>
#include <if.h>
#include <if_types.h>
#include <ip/in.h>
#include <ip/arp.h>

#include <drivers/drivers.h>
#include <busdriver.h>

#if (BYTE_ORDER==BIG_ENDIAN)
#define PCIDRIVER_NAME _T("pcibs")
#else
#define PCIDRIVER_NAME _T("pci")
#endif

#define STATS 0

#include <diagnostics.h>

#include <i8255x.h>

/* BUGBUG We sure could use a trivial VM system
 */
#define IoPhysicalToIoVirtual(x) ((PTR)x)
#define PhysicalToVirtual(x)     ((PTR)((ADDRESS)x))
#define VirtualToPhysical(x)     ((ADDRESS)x)

/*#define LOGIT 1*/
#if LOGIT
#define LOGSIZE 128
UINT LogP = 0;
struct {
    _TCHAR * Format;
    TIME When;
    UINT32 Info[1+4]; /* 2*16 total */
} LogBuf[LOGSIZE];

PRIVATE
void 
LogIt(_TCHAR *Format, UINT32 Info0, UINT32 Info1)
{
    TIME Now = CurrentTime();
    INT i;

    for (;;) {
        /* Disabled ?
         */
        if (LogP == ~0)
            return;
        /* Get a number, make sure its good.
         */
    Retry:
        i = AtomicInc(&LogP) - 1;
        if (i >= LOGSIZE) {
            /* Roundabout, with care
             */
            if (!AtomicCmpAndSwap(&LogP,
                                  i+1,
                                  0))
                goto Retry;
        } else
            /* Good one.
             */
            break;
    }
    LogBuf[i].Format  = Format;
    LogBuf[i].When = Now;
    LogBuf[i].Info[0] = Info0;
    LogBuf[i].Info[1] = Info1;
#if 0
    _tprintf(_T("LOG '%hs' %08x %08x %ld\n"),
           LogBuf[i].Format,
           LogBuf[i].Info[0],
           LogBuf[i].Info[1],
           LogBuf[i].When);

#endif
}

PRIVATE
void
PrintZeLog(void)
{
    UINT index, LogIs;
    _TCHAR *Format;
    TIME Now = CurrentTime(), t;

    /* Block the log while we print it
     */
    for (;;) {
        LogIs = LogP;
        if (AtomicCmpAndSwap(&LogP,
                             LogIs,
                             ~0))
            break;
    }
    /* LogIs the oldest entry when full
     * When non-full it points to a NULL entry.
     */
    index = LogIs;
    Int32ToInt64(t,0);
    do {

        /* Skip NULL entries (non-full and sanity)
         */
        Format = LogBuf[index].Format;
        if (Format == NULL)
            goto Next;

        _tprintf(Format,LogBuf[index].Info[0],LogBuf[index].Info[1]);
        _tprintf(_T("  %ld (+%ld)\n"),
                 LogBuf[index].When,
                 TimeSubtract(LogBuf[index].When,t));

        /* Keep track of delta-T
         */
        t = LogBuf[index].When;

        /* Next index, plus wraps
         */
    Next:
        if (++index >= LOGSIZE) index = 0;

    } while (index != LogIs);
    _tprintf(_T("Time now  %ld (+%ld)\n"), Now, TimeSubtract(Now, t) );
    /* Let the log go
     */
    LogP = LogIs;
}
#else
#define LogIt(x,y,z)
#define PrintZeLog()
#endif

typedef struct _IX_RX_BUFFER {
    PIBUFFER  m;
    IX_RFD Rfd;
    struct ether_header eh;
} IX_RX_BUFFER, *PIX_RX_BUFFER;

typedef struct _IX_TX_BUFFER {
    PIBUFFER  m;
    IX_TFD Tfd;
} IX_TX_BUFFER, *PIX_TX_BUFFER;

typedef struct _I8255X { /* Inherits from CDevice */

    struct arpcom ac;            /* standard ether interface header */
#define is_if ac.ac_if
#define is_addr ac.ac_enaddr
#define CollisionCount ac.ac_if.if_data.ifi_collisions
#define MissedPackets ac.ac_if.if_data.ifi_errors[IFI_ERR_MISS]
#define CompleteResets ac.ac_if.if_data.ifi_errors[IFI_ERR_RESET]
#define XmitTimeouts ac.ac_if.if_data.ifi_errors[IFI_ERR_XTIM]
#define MemoryErrors ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM]
#define FrameErrors  ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+1]
#define XmitDeferred ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+2]
#define XmitGood ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+3]
#define RecvGood ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+4]

    /* Driver's private data
     */
    PIX_CSR Csr;

    UINT32  PhyOui;
    UINT16  LinkStatus;
    UINT8   ChipRevision;
    UINT8   Pad;

    PIX_COMMAND CommandBuffer;
#define nRxBuffers 16
    PIX_RX_BUFFER RxHead, RxTail;

    PIX_TX_BUFFER TxBuffers;
    /* MUST be 2^n, else masking and wraps break
     */
#define nTxBuffers 8
#define nTxBuffersChopMask (nTxBuffers-1)
    /* Use UINT8 to test */
#define IXTYPE UINT32
    IXTYPE TxTail, TxHead;
#define TxIndex(n)  ((n) & nTxBuffersChopMask)
    
    PIX_STATISTICAL_COUNTERS StatisticalCounters;

    MUTEX   Lock;

} I8255X, *PI8255X;

/* Interrupt masking/unmasking
 */
#define INTERRUPTS_MASKED (IX_MSK_CNA|IX_MSK_FCP)
#define IxDisableAllInterrupts(_pIx_) { \
    IOSpaceWriteInt8((IO_PORT)&(_pIx_)->Csr->SCB.InterruptMask,IX_MSK_M); \
    IOWriteBufferFlush(); \
}
#define IxEnableInterrupts(_pIx_) { \
    IOSpaceWriteInt8((IO_PORT)&(_pIx_)->Csr->SCB.InterruptMask,INTERRUPTS_MASKED); \
    IOWriteBufferFlush(); \
}

/* Utils
 */
PRIVATE void
MilliSleep(int nMsecs)
{
    TIME Msecs;
    Int32ToInt64(Msecs,TIME_RELATIVE(TIME_MILLIS(nMsecs)));
    SleepUntil(Msecs);
}


/* Access to the PHY
 */
PRIVATE
BOOL 
IxMdiRead( PIX_CSR Csr, UINT RegNo, UINT16 *Value)
{
    UINT32 Data;
    UINT nRetries;

    Data = _bs32(IX_MC_READ | PHY_DEFAULT_ADDRESS | (RegNo << IX_MC_ADDR_SHIFT));

    IOSpaceWriteInt32((IO_PORT)&Csr->MdiControl,Data);
    IOWriteBufferFlush();

    nRetries = 1000;
    do {
        Delay(10);
        IOMemoryBarrier();
        Data = IOSpaceReadInt32((IO_PORT)&Csr->MdiControl);
        Data = _bs32(Data);
    } while (((Data & IX_MC_READY) == 0) && (--nRetries > 0));

    *Value = (UINT16) (Data & IX_MC_DATA_MASK);
    return (Data & IX_MC_READY) ? TRUE : FALSE;
}

#if _DEBUG /* unused */
BOOL 
IxMdiWrite( PIX_CSR Csr, UINT RegNo, UINT16 Value)
{
    UINT32 Data;
    UINT nRetries;

    /* NB: In principle, we should probe various PHY addresses
     *     and figure out (looking for a non-zero ID, for instance)
     *     which address was used to connect the PHY to the 8255x.
     *     In practice, #1 is used and is hardwired (but for the '57).
     */

    Data = _bs32(IX_MC_WRITE | PHY_DEFAULT_ADDRESS | (RegNo << IX_MC_ADDR_SHIFT) | Value);

    IOSpaceWriteInt32((IO_PORT)&Csr->MdiControl,Data);
    IOWriteBufferFlush();

    nRetries = 10000;
    do {
        Delay(1);
        Data = IOSpaceReadInt32((IO_PORT)&Csr->MdiControl);
        Data = _bs32(Data);
    } while (((Data & IX_MC_READY) == 0) && (--nRetries > 0));

    return (Data & IX_MC_READY) ? TRUE : FALSE;
}
#endif

/* Access to the EEPROM
 */

/* Wait for about 100us after toggling signals to the EEPROM.
 */
#define IxWaitForEeprom()    Delay(100)

PRIVATE
void
IxReadEepromUint16(PIX_CSR Csr, UINT WordIndex, PUINT16 Result)
{
    volatile UINT8 *EepromControl = (volatile UINT8 *)&Csr->EepromControl;
#if _DEBUG
    int DummyOut = 0;
#endif
    int iBit;
    UINT8 Reg;
    UINT16 Value16;

#define RaiseClock(_r_)                               \
    {                                                 \
        _r_ |= IX_EC_EESK;                            \
        IOSpaceWriteInt8((IO_PORT)EepromControl, _r_);\
        IOWriteBufferFlush();                         \
        IxWaitForEeprom();                            \
    }

#define LowerClock(_r_)                               \
    {                                                 \
        _r_ &= ~IX_EC_EESK;                           \
        IOSpaceWriteInt8((IO_PORT)EepromControl, _r_);\
        IOWriteBufferFlush();                         \
        IxWaitForEeprom();                            \
    }

    /*
     * We must shift out: <start bit> + <opcode> + <address>.
     */
    WordIndex = (1 << 8) | (IX_EEPROM_READ << 6) | (WordIndex & 0x3f);

    /*
     * Set the CS bit.
     */
    IOMemoryBarrier();
    Reg = IOSpaceReadInt8((IO_PORT)EepromControl);
    Reg &= ~(IX_EC_EESK | IX_EC_EEDI | IX_EC_EEDO);
    Reg |= IX_EC_EECS;
    IOSpaceWriteInt8((IO_PORT)EepromControl, Reg);
    IOWriteBufferFlush();

    /*
     * Shift out 9 bits - from MSB to LSB.
     */
    for (iBit = 8; iBit >= 0; iBit--)
    {
        if (WordIndex & (1 << iBit))
        {
            Reg |= IX_EC_EEDI;
        }
        else
        {
            Reg &= ~IX_EC_EEDI;
        }

        IOSpaceWriteInt8((IO_PORT)EepromControl, Reg);
        IOWriteBufferFlush();
        IxWaitForEeprom();

        /*
         * Pulse the clock.
         */
        RaiseClock(Reg);
        LowerClock(Reg);

#if _DEBUG
        if (iBit < 6) {
            if (DummyOut == 0) {
                Value16 = IOSpaceReadInt8((IO_PORT)EepromControl);
                if ((Value16 & IX_EC_EEDO) == 0)
                    DummyOut = iBit;
            }
        }
#endif
    }

#if _DEBUG
    /*
     * EEPROM DO bit must be 0 at this point (dummy-out).
     */
    if (DummyOut == 0) {
        Value16 = IOSpaceReadInt8((IO_PORT)EepromControl);
        if ((Value16 & IX_EC_EEDO) == 0)
            DummyOut = 666;
    }
#endif

    /*
     * Read out the 16 data bits.
     */
    Value16 = 0;
    Reg &= ~IX_EC_EEDI;

    for (iBit = 0; iBit < 16; iBit++)
    {
        /* Set clock valid
         */
        RaiseClock(Reg);

        /* Pick the bit up
         */
        Value16 = (UINT16)(Value16 << 1);
        IOMemoryBarrier();
        if (IOSpaceReadInt8((IO_PORT)EepromControl) & IX_EC_EEDO)
        {
            Value16 |= 1;
        }

        /* Flip clock
         */
        LowerClock(Reg);
    }

    /* De-select the chip
     */
    Reg &= ~IX_EC_EECS;
    IOSpaceWriteInt8((IO_PORT)EepromControl, Reg);
    IOWriteBufferFlush();

    /* JIC..
     */
    RaiseClock(Reg);
    LowerClock(Reg);

    /* Set endianess right
     */
    *Result = _bs16(Value16);
#undef RaiseClock
#undef LowerClock

#if _DEBUG
    if (DummyOut != 666)
        _tprintf(_T("IxReadEeprom: ERROR DummyOut at %d (!0)\n"),DummyOut);
#endif
}

/* Wait until the SCB is available
 */
PRIVATE
BOOL
IxIdle(PIX_CSR Csr)
{
    int nRetries = 70000; /* 700msec max */

    while (nRetries-- >= 0) {
        if (IOSpaceReadInt8((IO_PORT)&Csr->SCB.Command) == 0)
            return TRUE;
        Delay(10);
    }
    DBGME(3,_tprintf(_T("IxIdle timed out\n")));
    return FALSE;
}


#if _DEBUG
typedef _TCHAR *NAME;

PRIVATE const NAME RevisionNames[IX_MAXREV] = {
    _T("x?0"), _T("7a"), _T("7b"), _T("7c"),
    _T("8a"), _T("8b"), _T("9a"), _T("9b"),
    _T("9c"), _T("9ERa"), _T("x?a"), _T("x?b"),
    _T("0a"), _T("0b"), _T("0c"), _T("1a"),
    _T("1b")
};

#define RevisionName(r)  ((r < IX_MAXREV) ? RevisionNames[r] : _T("x??"))
#define PrintRev(r) _tprintf(_T("8255%s"),RevisionName(r))

typedef struct {
    NAME    Regname;
    UINT16  Offset;
    UINT8   Size;
    UINT8   Where;
#define R_CSR        2
#define R_MDI        3
} REGGIE;

PRIVATE
void 
IxDumpSomeRegs(ADDRESS IoBase, REGGIE *Regs, int nRegs)
{
    UINT RegNum;
    UINT32 Value = 0;
    UINT16 Value16;

    while (nRegs > 0)  {
        /* Print 4 per line */
        for (RegNum = 0; (nRegs > 0) && (RegNum < 4); RegNum++, nRegs--) {
            if (Regs[RegNum].Regname == NULL)
                goto NewlineAndReturn;
            switch (Regs[RegNum].Where) {
            case R_CSR:
                switch (Regs[RegNum].Size) {
                case 4:
                    Value = IOSpaceReadInt32((IO_PORT)(IoBase+Regs[RegNum].Offset));
                    Value = _bs32(Value);
                    break;
                case 2:
                    Value16 = IOSpaceReadInt16((IO_PORT)(IoBase+Regs[RegNum].Offset));
                    Value16 = _bs16(Value16);
                    Value   = (UINT32) Value16;
                    break;
                case 1:
                    Value = IOSpaceReadInt8((IO_PORT)(IoBase+Regs[RegNum].Offset));
                    break;
                }
                break;
            case R_MDI:
                if (IxMdiRead((PIX_CSR)IoBase,Regs[RegNum].Offset,&Value16))
                    Value = Value16;
                else {
                    _tprintf(_T("%s=cantread "),Regs[RegNum].Regname);
                    continue;
                }
                break;
            default:
                continue;
            }
#if 1
            if (Value)
                _tprintf(_T("%s(@%x)=%x "),
                         Regs[RegNum].Regname,
                         IoBase+Regs[RegNum].Offset,
                         Value);
#else
            _tprintf(_T("%s=%x "),Regs[RegNum].Regname,Value);
#endif
        }
        _tprintf(_T("\n "));
        Regs += RegNum;
    }
 NewlineAndReturn:
    _tprintf(_T("\n"));
}

PRIVATE
REGGIE 
CsrRegs[] = {
    { _T("Status"), 0x00, 1, R_CSR},
    { _T("StatAck"), 0x01, 1, R_CSR},
    { _T("Command"), 0x02, 1, R_CSR},
    { _T("InterruptMask"), 0x03, 1, R_CSR},
    { _T("GeneralPointer"), 0x04, 4, R_CSR},
    { _T("Port"), 0x08, 4, R_CSR},
    { _T("EepromControl"), 0x0e, 2, R_CSR},
    { _T("MdiControl"), 0x10, 4, R_CSR},
    { _T("RxDmaByteCount"), 0x14, 4, R_CSR},
    { _T("EarlyRxInt"), 0x18, 1, R_CSR},
    { _T("FcThreshold"), 0x19, 1, R_CSR},
    { _T("FcXonXoff"), 0x1a, 1, R_CSR},
    { _T("PMDR"), 0x1b, 1, R_CSR},
    { _T("GeneralControl"), 0x1c, 1, R_CSR},
    { _T("GeneralStatus"), 0x1d, 1, R_CSR},
    { _T("FunctionEvent"), 0x30, 4, R_CSR},
    { _T("FunctionEventMask"), 0x34, 4, R_CSR},
    { _T("FunctionEventState"), 0x38, 4, R_CSR},
    { _T("ForceEvent"), 0x3c, 4, R_CSR},
    { NULL, }
};
#define nCsrRegs ((sizeof CsrRegs) / (sizeof CsrRegs[0]))

PRIVATE
REGGIE
MdiRegs[] = {
    { _T("MDI_CONTROL"), IX_MDI_CONTROL, 2, R_MDI},
    { _T("MDI_STATUS"), IX_MDI_STATUS, 2, R_MDI},
    { _T("MDI_ID0"), IX_MDI_ID0, 2, R_MDI},
    { _T("MDI_ID1"), IX_MDI_ID1, 2, R_MDI},
    { _T("MDI_AN_ADVERTISE"), IX_MDI_AN_ADVERTISE, 2, R_MDI},
    { _T("MDI_AN_PARTNER"), IX_MDI_AN_PARTNER, 2, R_MDI},
    { _T("MDI_AN_EXPANSION"), IX_MDI_AN_EXPANSION, 2, R_MDI},
    { _T("PHY55_CSR"), IX_PHY55_CSR, 2, R_MDI},
    { _T("PHY55_CTL"), IX_PHY55_CTL, 2, R_MDI},
    { _T("PHY55_CLOCK"), IX_PHY55_CLOCK, 2, R_MDI},
    { _T("PHY55_FALSE_CARRIER"), IX_PHY55_FALSE_CARRIER, 2, R_MDI},
    { _T("PHY55_DISCONNECT"), IX_PHY55_DISCONNECT, 2, R_MDI},
    { _T("PHY55_FRAME"), IX_PHY55_FRAME, 2, R_MDI},
    { _T("PHY55_SYMBOL"), IX_PHY55_SYMBOL, 2, R_MDI},
    { _T("PHY55_100_EOF"), IX_PHY55_100_EOF, 2, R_MDI},
    { _T("PHY55_10_EOF"), IX_PHY55_10_EOF, 2, R_MDI},
    { _T("PHY55_JABBER"), IX_PHY55_JABBER, 2, R_MDI},
    { _T("PHY55_EQUALIZER"), IX_PHY55_EQUALIZER, 2, R_MDI},
    { _T("PHY55_CTL2"), IX_PHY55_CTL2, 2, R_MDI},
    { NULL, }
};
#define nMdiRegs ((sizeof MdiRegs) / (sizeof MdiRegs[0]))

PRIVATE
void
IxDumpAllRegs(PI8255X pIx, BOOL PhyRegsAlso)
{
    ADDRESS IoBase = (ADDRESS) pIx->Csr;

    _tprintf(_T("I8255X(%x) Base = %x regs:\n "),(ADDRESS)pIx,IoBase);
    IxDumpSomeRegs(IoBase,CsrRegs,nCsrRegs);
    if (PhyRegsAlso)
        IxDumpSomeRegs(IoBase,MdiRegs,nMdiRegs);
}

PRIVATE
void
IxDumpEeprom(PI8255X pIx)
{
    UINT16 Value; 
    UINT WordIndex;

    /* Minimal size is 64 words, with wraparound.
     * Intel specs document 256 words
     */
    for (WordIndex = 0; WordIndex < 0x100; WordIndex++) {
        IxReadEepromUint16(pIx->Csr,WordIndex,&Value);
        if ((WordIndex & 7) == 0)
            _tprintf(_T("\nEeprom[%2d]: "),WordIndex);
        _tprintf(_T("%4x "),Value);
    }
    _tprintf(_T("\n"));
    /* On my card I got:
     *   Eeprom[ 0]:    e  c07 a35d  302    0  102  147    0 
     *   Eeprom[ 8]: 1474 1062 8240  b00 8680    0    0    0 
     *   Eeprom[16]:    0    0    0    0    0    0    0    0 
     *   Eeprom[24]:    0    0    0    0    0    0    0    0 
     *   Eeprom[32]:    0    0    0    0    0    0    0    0 
     *   Eeprom[40]:    0    0    0    0    0    0    0    0 
     *   Eeprom[48]: 2801    0    0    0    0    0    0    0 
     *   Eeprom[56]:    0    0    0    0    0    0    0 a764 
     */
}

PRIVATE 
void
dump16(PUINT16 p, int nWords)
{
    int i;
    for (i = 0; i < nWords; i++)
        _tprintf(_T(" %x"), _bs16(p[i]));
}

/* Linearized Tfd is:
 * @00: UINT16 Tfd.TCB.h.Status     (IX_CBS_?)
 * @02: UINT16 Tfd.TCB.h.Command    (IX_CBC_?)
 * @04: UINT32 Tfd.TCB.h.LinkAddress
 * @08: UINT32 Tfd.TCB.TbdArrayAddress
 * @0c: UINT16 Tfd.TCB.TcbByteCount
 * @0e: UINT8  Tfd.TCB.Threshold
 * @0f: UINT8  Tfd.TCB.TbdNumber
 * @10: UINT32 Tfd.TBD[0].BufferAddress
 * @14: UINT16 Tfd.TBD[0].ActualCount
 * @16: UINT16 Tfd.TBD[0].Reserved
 * @18:            TBD[1]
 * @20:            TBD[2]
 * @28:            TBD[3]
 */
PRIVATE
void
IxDumpTfd(PI8255X pIx, int Index)
{
    int j;
    PIX_TX_BUFFER Tb;

    Tb = &pIx->TxBuffers[TxIndex(Index)];
    _tprintf(_T("Tb[%d]: %x @%x:"),
             Index, (ADDRESS)Tb->m, (ADDRESS)&Tb->Tfd.TCB.h.Status);
    dump16(&Tb->Tfd.TCB.h.Status,8);
    _tprintf(_T("\n  TBD:"));
    for (j = 0; j < IX_MAX_TBD; j++)
        _tprintf(_T(" %x.%x"),
                 _bs32(Tb->Tfd.TBD[j].BufferAddress),
                 _bs16(Tb->Tfd.TBD[j].ActualCount));
    _tprintf(_T("\n"));
}

PRIVATE
void
IxDumpTfds(PI8255X pIx, int nFirst, int nLast)
{
    int i, n;

    n = 0;
    for (i = nFirst; i < nLast; i++) {
        IxDumpTfd(pIx,i); n++;
    }
    if (n < nTxBuffers) {
        _tprintf(_T("------ Inactive TFDs\n"));
        while (n++ < nTxBuffers)
            IxDumpTfd(pIx,i++);
    }
}

PRIVATE
void
IxDumpRfds(PI8255X pIx)
{
    PIX_RX_BUFFER Rb, Rn;
    UINT32 Value32;
    int max = 20; /* We use 10 max, so put a stop on it */

    Rb = pIx->RxHead;
    for (;max-- > 0;Rb=Rn) {

        _tprintf(_T("Rb@%x: "), (ADDRESS)Rb);
        _tprintf(_T("%x {%x %x %x} %x %x %x\n"),
                 (ADDRESS)Rb->m,
                 _bs16(Rb->Rfd.h.Status),
                 _bs16(Rb->Rfd.h.Command),
                 _bs32(Rb->Rfd.h.LinkAddress),
                 _bs32(Rb->Rfd.RbdAddress),
                 _bs16(Rb->Rfd.ActualCount),
                 _bs16(Rb->Rfd.Size));

        /* Whats next, check for end of q
         */
        Value32 = Rb->Rfd.h.LinkAddress;
        Value32 = _bs32(Value32);
        if (Value32 == IX_LINK_ADDRESS_NULL)
            return;

        Value32 -= offsetof(IX_RX_BUFFER,Rfd);
        Rn = (PIX_RX_BUFFER) PhysicalToVirtual(Value32);

        /* Break out at tail, regardless
         */
        if (Rb == pIx->RxTail) {
            _tprintf(_T("Tail not terminated!\n"));
            return;
        }
    }
}

#endif


/* Read the Ethernet Station Address
 */
PRIVATE
void
IxReadStationAddress(PI8255X pIx, PUINT8 pSA)
{
    IxReadEepromUint16(pIx->Csr,0,((PUINT16)pSA)+0);
    IxReadEepromUint16(pIx->Csr,1,((PUINT16)pSA)+1);
    IxReadEepromUint16(pIx->Csr,2,((PUINT16)pSA)+2);
}

/* Issue an SCB-level command
 */
PRIVATE
BOOL
IxScbCommand(PIX_CSR Csr, UINT8 Command, UINT32 GpValue)
{
    if (!IxIdle(Csr))
        return FALSE;

    GpValue = _bs32(GpValue);
    IOSpaceWriteInt32((IO_PORT)&Csr->SCB.GeneralPointer,GpValue);
    IOWriteBufferFlush();
    IOSpaceWriteInt8((IO_PORT)&Csr->SCB.Command,Command);
    IOWriteBufferFlush();

    return TRUE;
}

/* Dump the statistical counters
 */
#define IxDumpCounters(_pIx_) IxScbCommand(_pIx_->Csr,IX_CMD_CU_DUMP_SC,0)
#define DumpCounterStatusOk IX_SCDUMP_COMPLETE

/* Synchronously execute a command
 * Only usable when the interface is not active.
 */
PRIVATE
BOOL
IxCommandAndWait(PI8255X pIx, UINT8 Command, PIX_COMMAND_HEADER Data, UINT DataSize)
{
    PIX_CSR Csr = pIx->Csr;
    UINT16 Value16;
    UINT32 Value32;
    UINT nRetries;

    /* Fill in header
     */
    Value16 = Command | IX_CBC_ELAST;
    Data->Status = 0;
    Data->Command = _bs16(Value16);
    Data->LinkAddress = _bs32(IX_LINK_ADDRESS_NULL);
    IOMemoryBarrier();

    /* Get it going
     */
    Value32 = VirtualToPhysical(Data);
    if (!IxScbCommand(Csr,IX_CMD_CU_START,Value32)) {
        DBGME(3,_tprintf(_T("IxCommand not started (%x)\n"),Value32));
        return FALSE;
    }

    /* Wait till its done
     * We cannot spin on the SCB because there is latency there,
     * and besides we must have a good cache coherency story anyways.
     * But yes, we could spin on StatAck&CNA
     */
    for (nRetries = 10000; nRetries >= 0; nRetries--) {
        Delay(100);
        Value16 = Data->Status;
        Value16 = _bs16(Value16);
        if (Value16 & IX_CBS_COMPLETE)
            break;
    }

    /* Clear the CNA bit
     */
    IOSpaceWriteInt8((IO_PORT)&Csr->SCB.StatAck,IX_SA_CNA);

    /* See what happened
     */
    if ((Value16 & IX_CBS_OK) == 0) {
        DBGME(3,_tprintf(_T("IxCommand failed %x\n"),Value16));
    }

    return ((Value16 & IX_CBS_OK) ? TRUE : FALSE);
}

/* Load the individual station address
 * NB: Should be a good one off the eeprom
 */
PRIVATE
BOOL
IxLoadStationAddress(PI8255X pIx)
{
    PIX_COMMAND Command = pIx->CommandBuffer;
    memcpy(Command->MacAddress.data,pIx->is_addr,6);
    return IxCommandAndWait(pIx, IX_CBC_MAC_ADDRESS,
                            &Command->MacAddress.h, sizeof(IX_COMMAND_HEADER)+6);
}

/* Default config, pagg. 61-80
 * Assume '58 or better
 * Values in {nn} come from a DUMP of a fresh card.
 */
static const UINT8 DefaultConfigData[24] = {
        22,   /* 0: Byte count */
        0x08, /* 1: Tx/Rx FIFO limits (0 tx, 32/64 rcv) */
        0x00, /* 2: Adaptive IFS (no) */
#define WI 1
#if WI
        0x01, /* 3: Terminate Write on Cache Line (no)
               *    Read Alignment Enable (no)
               *    MWI Enable (yes)  */
#else
        0,
#endif
        0,
              /* 4: Rx DMA maximum byte count div 4 */
        0,
              /* 5: Tx DMA max count div 4 
               *     DMBC Enable (no) */
        0x32, /* 6: Save bad frames (no)
               *    Discard overrun receive frames (yes, flipped)
               *    Extended stats (no, flipped)
               *    Extended TxCB (no, flipped)
               *    CI interrupt only (no, use CNA)
               *    TNO/TCO interrupt (no)
               *    Late SCB update (no) */
        0x03, /* 7: Dynamic TBD (no)
               *    Two frames in FIFO (no)
               *    Underrun retry (01b : whole frame)
               *    Discard short receive frames */
        0x01, /* 8: 503/MII (MII)                            {01}*/
        0x00, /* 9: MC Match wakeup '58 (no)                 {00}
               *    ARP wakeup '58 (no)
               *    Link wakep (no)
               *    VLAN ARP '58 or TCO '59 (no)
               *    UDP/TCP checksum (no, its useless) */
        0x2e, /* 10: Loopback (no)  [was x28 ok?]            {2e}
               *     Preamble length (10b 7 bytes)
               *     SA insertion (1 SA from memory) */
        0x00, /* 11: Linear priority (normal backoff)        {00} */
        0x61, /* 12: Interframe spacing (0110b 96 bit times) {60}
               *     Linear priority mode (1 reserved) */
        0x00, /* 13: IP Address low (dont)                   {ff} */
        0xf2, /* 14: IP Address high (dont)                  {ff} */
        0x48, /* 15: CRS|CDT (CRS only)                      {c8}
               *     1
               *     CRC16 (no)
               *     Ignore U/L (no)
               *     1
               *     Wait after win (no)
               *     Broadcast disable (no)
               *     Promiscuous mode (no) */
        0x00, /* 16: FC delay LSB (none)                     {00} */
        0x40, /* 17: FC delay MSB (none)                     {40} */
        0xf2, /* 18: Priority FC threshold (1111b none)      {f2}
               *     Long recv ok (no)
               *     Rx CRC transfer (no)
               *     Padding enable (yes)
               *     Stripping enable (no) */
        0x80, /* 19: Full duplex pin enable (yes)            {00}
               *     Force full duplex (no)
               *     Reject FC (no)
               *     Full duplex restart FC (no)
               *     Full duplex restop FC (no)
               *     Full duplex transmit FC disable (no)
               *     Magic packet wakeup disable (no)
               *     Address match wakeup (no) */
        0x3f, /* 20: Multiple IA (no)                        {3f}
               *     Priority FC location (1b byte 31) 
               *     11111b */
        0x0d, /* 21: Multicast all (yes)                     {05}
               *     101b */
        0,0   /* specs says so? */
    };

/* Load the configuration data
 */
PRIVATE
BOOL
IxLoadConfiguration(PI8255X pIx)
{
    PIX_COMMAND Command = pIx->CommandBuffer;
    memcpy(Command->Configure.data,DefaultConfigData,24);
    return IxCommandAndWait(pIx, IX_CBC_CONFIGURE,
                            &Command->Configure.h,
                            sizeof Command->Configure);
}

#if _DEBUG
PRIVATE
BOOL
IxDumpConfiguration(PI8255X pIx, _TCHAR *Where)
{
#define nDUMP (596/4)
    UINT32 *DumpData = malloc(nDUMP*4);
    UINT32 Value32;
    PIX_COMMAND Command = pIx->CommandBuffer;
    BOOL r = FALSE;
    int i;

    if (DumpData == NULL)
        return FALSE;
    memset(DumpData,0,nDUMP*4);

    Value32 = VirtualToPhysical(DumpData);
    Command->Dump.BufferAddress = _bs32(Value32);
    if (!IxCommandAndWait(pIx, IX_CBC_DUMP,
                          &Command->Configure.h,
                          sizeof Command->Dump))
        goto Out;

    MilliSleep(100);

    _tprintf(_T("i8255xDump %s:\n"),Where);
    for (i = 0; i < nDUMP; i++) {
        _tprintf(_T(" %08x"), DumpData[i]);
        if ((i & 7) == 7)
            _tprintf(_T("\n"));
    }
    _tprintf(_T("\n"));

    r = TRUE;
 Out:
    free(DumpData);
    return r;
}
#else
#define IxDumpConfiguration(x,y)
#endif

/* Start the receive unit, if it wasnt already
 */
PRIVATE
BOOL
IxStartRU(PI8255X pIx)
{
    PIX_CSR Csr = pIx->Csr;
    UINT8 Status;
    UINT32 Value32;
    PIX_RX_BUFFER Rb;

    /* If its ready already we are done
     */
    Status = IOSpaceReadInt8((IO_PORT)&Csr->SCB.Status);
    if ((Status & IX_ST_RUS_MASK) == IX_ST_RUS_READY)
        return TRUE;

    /* We must give it an RFD to fill up (head of list)
     * BUGBUG Might want to check its not done already
     */
    Rb = pIx->RxHead;
    Value32 = VirtualToPhysical(&Rb->Rfd);
    if (!IxScbCommand(Csr, IX_CMD_RU_START, Value32)) {
        DBGME(3,_tprintf(_T("IxStartRU wont start\n")));
        return FALSE;
    }

    /* Wait for the SCB to be done
     */
    if (!IxIdle(Csr)) {
        DBGME(3,_tprintf(_T("IxStartRU wont idle\n")));
        return FALSE;
    }

    /* Says the SCB wont reflect true status for a while. Sigh.
     */
    return TRUE;
}

/* Check the link status & speed
 */
PRIVATE
void
IxLinkStatus(PI8255X pIx)
{
    PIX_CSR Csr = pIx->Csr;
    UINT16 Status, Changes = 0;
    BOOL Is100Mbps = FALSE;

    /* After '59 its easy
     */
    if (pIx->ChipRevision >= IX_82559_A) {
        /* Link Status changes?
         */
        Status = IOSpaceReadInt8((IO_PORT)&Csr->PMDR);
        if ((Status & IX_PMDR_LINK_STATUS) == 0)
            return;

        /* Ack bit to clear
         */
        IOSpaceWriteInt8((IO_PORT)&Csr->PMDR,IX_PMDR_LINK_STATUS);

        /* So where are we
         */
        Status = IOSpaceReadInt8((IO_PORT)&Csr->GeneralStatus);
        Is100Mbps = (Status & IX_GS_100Mbps);
        if (Status & IX_GS_LINK_UP)
            goto LinkWentUp;
        else
            goto LinkWentDown;
    }

    /* Get the MDI status
     */
    if (!IxMdiRead(Csr,IX_MDI_STATUS,&Status)) {
        DBGME(3,_tprintf(_T("IxLinkStatus cant read\n")));
        return;
    }

    /* Did anything change?
     */
    Changes = pIx->LinkStatus ^ Status;
    if (!Changes)
        return;/* nope, done */

    /* What is it?
     * The only thing that matters, really, is the LINK_UP bit.
     */
    pIx->LinkStatus = Status;

    /* Did the link go down
     */
    if ((Changes & IX_MST_LINK_UP) && ((Status & IX_MST_LINK_UP) == 0)) {
    LinkWentDown:
        DBGME(2,_tprintf(_T("IxLinkStatus: down %x %x\n"),Changes,Status));
        pIx->is_if.if_flags &= ~IFF_RUNNING;
        pIx->is_if.if_baudrate = 0; /* until next time */
        return;
    }

    /* The link is up but is autonegotiation complete yet
     * If not, wait for later
     */
    if ((Status & IX_MST_AN_COMPLETE) == 0) {
        DBGME(1,_tprintf(_T("IxLinkStatus incomplete %x\n"),Status));
        return;
    }

    /* What speed are we running at
     * More like.. "what should we be running at"
     */
    if (!IxMdiRead(Csr,IX_MDI_AN_ADVERTISE,&Status)) {
        DBGME(3,_tprintf(_T("IxLinkStatus cant read adv\n")));
        return;
    }
    if (!IxMdiRead(Csr,IX_MDI_AN_PARTNER,&Changes)) {
        DBGME(3,_tprintf(_T("IxLinkStatus cant read par\n")));
        return;
    }

    Is100Mbps = (Status & Changes & (IX_AAA_TX_FULL_DUPLEX|IX_AAA_TX_HALF_DUPLEX));

 LinkWentUp:
    DBGME(0,_tprintf(_T("IxLinkStatus: up %x %x (10%sMbps)\n"),
                     Changes,Status,Is100Mbps ? _T("0") : _T("")));
    if (Is100Mbps)
        pIx->is_if.if_baudrate = 100000000; /* 100Mbps */
    else
        pIx->is_if.if_baudrate = 10000000; /* 10Mbps */


    /* We are up and about
     * BUGBUG do we have to restart the TX/RX units?
     */
    pIx->is_if.if_flags |=  IFF_RUNNING;
}

/* Allocate and initialize a receive buffer
 */
PRIVATE
BOOL
IxAllocateReceiveBuffer(PI8255X pIx, PIX_RX_BUFFER *Result)
{
    PIBUFFER m;
    PIX_RX_BUFFER Rb;
    UINT32 ActualSize;
    SCODE sc;
    struct ifnet *ifp = &pIx->is_if;

    /* Size our recv buffers on account of the UDP/TCP xsum that
     * is (possibly) appended by the interface (2 bytes).
     */
#define ETHERSIZE (ETHERMTU+sizeof(struct ether_header)+sizeof(UINT16))
#define bsize (UINT32)(sizeof(IX_RX_BUFFER)+ETHERSIZE)
    sc = pIx->ac.ac_if.MBufFact->v->AllocateBuffer(pIx->ac.ac_if.MBufFact,
                                                   bsize, (PINETDRIVER) ifp,
                                                   &m);
    if (FAILED(sc))
        return FALSE;

    /* Get pointer to data area, make sure its contig and big enmough
     */
    sc = m->v->GetDataPointer(m,0,(BYTE**)&Rb,&ActualSize);
    if (FAILED(sc) || (ActualSize != bsize)) {
        DBGME(3,_tprintf(_T("I8255xARB ?%d %d %x\n"),ActualSize,bsize,sc));
        goto DropPacket;
    }

    /* Adjust data area pointer on account of our prepended crud
     * Make sure the compiler does not play rounding tricks on us
     */
    ((PMBUF)m)->m_data += (sizeof(PTR)+sizeof(IX_RFD)+14);

    /* Remember selfpointer
     */
    Rb->m = m;

    Rb->Rfd.h.Status = 0;
    Rb->Rfd.h.Command = _bs16(IX_CBC_ELAST);
    Rb->Rfd.h.LinkAddress = _bs32(IX_LINK_ADDRESS_NULL);
    Rb->Rfd.RbdAddress = _bs32(IX_LINK_ADDRESS_NULL); /* I hope */
    Rb->Rfd.ActualCount = 0;
    Rb->Rfd.Size = _bs16(ETHERSIZE);

    *Result = Rb;
    return TRUE;

 DropPacket:
    m->v->Release(m);
    return FALSE;
#undef bsize
}

/* Process packets that have been completely received.
 */
PRIVATE
void
IxRecvComplete(PI8255X pIx)
{
    UINT32 Value32;
    UINT16 Value16, Length;
    PIX_RX_BUFFER Rb, Rn;
    PMBUF Pkt;
    struct ifnet *ifp = &pIx->is_if;

    DBGME(0,_tprintf(_T("Enter I8255xRecvComplete\n")));

    for (;;) {
        /* Look at the first in the queue, is it done yet.
         */
        if ((Rb = pIx->RxHead) == NULL)
            break;

        Value16 = Rb->Rfd.h.Status;
        if ((Value16 & _bs16(IX_CBS_COMPLETE)) == 0)
            break;

        /* Adjust head pointer
         */
        Value32 = Rb->Rfd.h.LinkAddress;
        if (Value32 == _bs32(IX_LINK_ADDRESS_NULL)) {
            pIx->RxHead = NULL;
            pIx->RxTail = NULL;
        } else {
            Value32 = _bs32(Value32);
            Value32 -= offsetof(IX_RX_BUFFER,Rfd);
            pIx->RxHead = (PIX_RX_BUFFER)PhysicalToVirtual(Value32);
        }

        /* Get 'nother one just like it
         */
        if (IxAllocateReceiveBuffer(pIx,&Rn)) {
            Value32 = VirtualToPhysical(&Rn->Rfd);
            if (pIx->RxTail != NULL) {
                /* BUGBUG what if ts completed already */
                pIx->RxTail->Rfd.h.LinkAddress = _bs32(Value32);
                pIx->RxTail->Rfd.h.Command = 0;
                pIx->RxTail = Rn;
                /* FlushCache() ? */
            } else {
                pIx->RxHead = Rn;
                pIx->RxTail = Rn;
            }
        }

        /* Error? drop ifso
         */
        Pkt = (PMBUF) Rb->m;

        /* Carefully compute length sans ether header
         * No trust no compilers me
         */
        Length = Rb->Rfd.ActualCount;
        Length = _bs16(Length) & IX_AC_MASK;
        if (Length > 14)
            Length -= 14;
        else
            Length = 0;

        /* Check status
         */
        if ((Value16 & _bs16(IX_CBS_OK)) == 0)
            goto DropPacket;

        /* Check length
         * NB: Account for possible xsum
         */
        if ((Length == 0) || (Length > (ETHERMTU+sizeof(UINT16))))
            goto DropPacket;

        /* A'right!  bswap the ethertype and ship
         * NB: m_data was adjusted before
         */
        Rb->eh.ether_type = ntohs(Rb->eh.ether_type);
        Pkt->m_pkthdr.len = Length;
        Pkt->m_len = Length;

        DBGME(0,_tprintf(_T("IxRecv got %d (%x) @%x!\n\r"),
                         Length,Rb->eh.ether_type,(ADDRESS)Rb));
//        LogIt(_T("Recv, length=%4d"),Length,0);
        ifp->if_ipackets++;
        ifp->MBufFact->v->EtherInput(ifp->MBufFact,
                                     (PINETDRIVER) ifp,
                                     &Rb->eh,
                                     (PIBUFFER) Pkt);
        continue;

    DropPacket:
        LogIt(_T("rcvdrop, length=%d status %x"),Length,_bs16(Value16));
        if (Pkt)
            Pkt->m_hdr.v->Release((PIBUFFER)Pkt);
        ifp->if_iqdrops++;
        ifp->if_ierrors++;
    }
}

/*
 * Start output on interface.
 */
void MCT
I8255xXmit( PI8255X pIx)
{
    int nChunks, FirstOne;
    IXTYPE i;
    SCODE sc;
    UINT32 Value32;
    PMBUF Request, m;
    PIX_TX_BUFFER Tb = NULL;
    struct ifnet *ifp = &pIx->is_if;
    UINT8 Status;

    DBGME(0,_tprintf(_T("Enter I8255xXmit\n")));

    pIx->is_if.MBufFact->v->NetLock(pIx->is_if.MBufFact);

    /* Remember where we started at
     */
    FirstOne = pIx->TxHead;

    /* If we came from interrupt, that must be the last xmit command
     * If we didnt, the flag is cleared anyways.
     */
    LogIt(_T("\nXmit, firstone %d flags=%x"),FirstOne,ifp->if_flags);
    ifp->if_flags &= ~IFF_OACTIVE;

    /* Queue up as many frames as we can
     */
    while ((Request = pIx->is_if.if_snd.ifq_head) != NULL) {

        /* Do we have room
         * NB: We cannot fill them *all* up, we must leave room for one.
         *     Because the CU can be suspended on that one, which we would
         *     rewrite but again with the S bit on. And lose sync.
         */
        i = pIx->TxHead - pIx->TxTail;
        if (i >= (nTxBuffers-1))
            break;/*no room */

        /* Dequeue the packet
         */
        IF_DEQUEUE(&pIx->is_if.if_snd, m);
        assert(m == Request);

        /* How many chunks? We can do at most 4(IX_MAX_TBD) directly
         */
        nChunks = 0;
        for (;m;m=m->m_next) {
            if (m->m_len)
                nChunks++;
        }

        DBGME(0,_tprintf(_T("I8255xXmit %d (%d).\n"),
                         Request->m_pkthdr.len, nChunks));

        if (nChunks > IX_MAX_TBD) {

            /* BUGBUG Untested codepath */
            DBGME(4,_tprintf(_T("i8255Xmit: %d chiunks?\n"),nChunks));
            sc = Request->m_hdr.v->MakeContiguous((PIBUFFER)Request,
                                                  0,
                                                  Request->m_pkthdr.len);

        }

        /* Remember where we started at
         */
        i = pIx->TxHead;
        Tb = pIx->TxBuffers + TxIndex(i);

        /* Increment head pointer, beware of wraparounds
         */
        if (++i == 0) {
            /* Signal XmitComplete to come back here,
             * as we might endup leaving buffers in the queue
             */
            ifp->if_flags |= IFF_OACTIVE;
            /* Notice we cannot just let it go to zero,
             * that would leave no room for TxTail to catch up.
             */
            i += nTxBuffers;
            LogIt(_T("WrapTx %x %x"),pIx->TxTail, pIx->TxHead);
        }
        pIx->TxHead = i;

        /* Fill in all segment pointers and sizes
         */
        Tb->m = (PIBUFFER) Request;
        i = 0;

        for (m=Request;m;m=m->m_next) {
            PTR p;
            if (m->m_len == 0)
                continue;

            p = mtod(m,PTR);
            Value32 = VirtualToPhysical(p);
            Tb->Tfd.TBD[i].BufferAddress = _bs32(Value32);
            Tb->Tfd.TBD[i].ActualCount   = _bs16((UINT16)m->m_len);
            i++;
        }
        Tb->Tfd.TCB.h.Status = 0;
        Tb->Tfd.TCB.TbdNumber = (UINT8)i;

        /* Off it goes.
         */
        Tb->Tfd.TCB.h.Command = _bs16(IX_CBC_TRANSMIT|IX_CBC_SF|
                                      IX_CBC_INTERRUPT|IX_CBC_SUSPEND);
        IOMemoryBarrier();
        LogIt(_T("Send, length=%4d Tb=%x"),Request->m_pkthdr.len,(ADDRESS)Tb);

        /* Now go back and fix the previous one to be non-suspend, non-intr
         * [This is harmless on the very first call to this function.]
         * NB: Do not reverse the order of these two operations.
         *     Otherwise the hw will keep on scanning the list while we
         *     might be interrupted away, and we lose sync with it.
         */
        if (--Tb < pIx->TxBuffers)
            Tb = pIx->TxBuffers + (nTxBuffers-1);
        Tb->Tfd.TCB.h.Command = Tb->Tfd.TCB.h.Command & 
                                   (~_bs16(IX_CBC_SUSPEND|IX_CBC_INTERRUPT));
        IOMemoryBarrier();
    }

    /* Are we (now) full?
     * If we are full, spare our callers a useless trip back here
     */
    i = pIx->TxHead - pIx->TxTail;
    if (i >= (nTxBuffers-1)) {
        /* We are full, spare our callers a useless trip back here
         */
        LogIt(_T("XmitBusy, head=%d flags=%x"),pIx->TxHead,ifp->if_flags);
        ifp->if_flags |= IFF_OACTIVE;
    }

    /* Did we queue up anything?
     * NB: TB does NOT point to the last one but rather the one before.
     */
    if (Tb) {

        /* Get the CU going (if its suspended or idle)
         */
        Status = IOSpaceReadInt8((IO_PORT)&pIx->Csr->SCB.Status);
        Status = Status & IX_ST_CUS_MASK;

        /* We hope to find it busy, somehow.
         */
        if (Status != IX_ST_CUS_IDLE) {

            /* We have already cleared the S bit in the last one, now resume.
             * NB: Resume is ok an Active (actually needed, it seems)
             */
            IOSpaceWriteInt8((IO_PORT)&pIx->Csr->SCB.Command,IX_CMD_CU_RESUME);
            IOWriteBufferFlush();
            LogIt(_T("CUResume fo=%d Tb=%x"),FirstOne,(ADDRESS)Tb);

        } else {
            /* Idle, restart it at the first buffer we just queued
             */
            Tb = pIx->TxBuffers + TxIndex(FirstOne);
            Value32 = VirtualToPhysical(&Tb->Tfd.TCB.h);
            if (!IxScbCommand(pIx->Csr,IX_CMD_CU_START,Value32)) {
                DBGME(3,_tprintf(_T("I8255xXmit CU wont!\n")));
            }
            LogIt(_T("CUStart fo=%d Tb=%x"),FirstOne,(ADDRESS)Tb);
        }
    }

    pIx->is_if.MBufFact->v->NetUnlock(pIx->is_if.MBufFact);
}

/* Handle Xmit complete interrupts
 */
PRIVATE
void
IxXmitComplete(PI8255X pIx)
{
    IXTYPE i;
    PIBUFFER Pkt;
    PIX_TX_BUFFER Tb;
    struct ifnet *ifp = &pIx->is_if;
    UINT16 Value16;

    DBGME(0,_tprintf(_T("Enter I8255xXmitComplete\n")));

 Again:
    for (i = pIx->TxTail; i < pIx->TxHead; i++) {

        Tb = pIx->TxBuffers + TxIndex(i);
        Value16 = Tb->Tfd.TCB.h.Status;
        if (Value16 & _bs16(IX_CBS_COMPLETE)) {

            /* Errors?
             */
            if ((Value16 & _bs16(IX_CBS_OK)) == 0)
                ifp->if_oerrors++;

            Pkt = Tb->m;
            Tb->m = NULL;
            LogIt(_T("XmCmpl pkt=%x Tfd=%x"),(ADDRESS)Pkt,(ADDRESS)Tb);
            if (Pkt) {
                /* Only count packets we actually sent
                 */
                ifp->if_opackets++;
                Pkt->v->Release(Pkt);
            }
            /* Should be AtomicInc but we are cute
             */
            pIx->TxTail++;
            continue;
        }

        LogIt(_T("XmCmpl ix=%d st=%x"),i,_bs16(Value16));
        break;
    }
    /* Check to see if the TxHead pointer wrapped-around
     * That would cause us to leave the loop too soon.
     */
    if (pIx->TxTail > pIx->TxHead) {
        LogIt(_T("TxWrap %x %x"),pIx->TxTail, pIx->TxHead);
        pIx->TxTail &= nTxBuffersChopMask;
        goto Again;
    }                

    /* Check if there might be something on the (upper) queue
     */
    if (ifp->if_flags & IFF_OACTIVE) {
        /* And that clears the flag
         */
        LogIt(_T("TxComp OACTIVE"),0,0);
        I8255xXmit(pIx);
    }
}

/*
 * Interrupt service routine.
 * Return TRUE if rescheduling needed.
 */
BOOL MCT
I8255xIsr(PCDEVICE pDev, PBOOL pNotMyInterrupt)
{
    PI8255X pIx = (PI8255X) pDev;

//    LogIt(_T("Isr!  StatAck= %2x"),IOSpaceReadInt8((IO_PORT)&pIx->Csr->SCB.StatAck),0);

    /* Disable interrupts
     */
    IxDisableAllInterrupts(pIx);

#if _DEBUG
    pIx->XmitDeferred++;
#endif

    /* Wakeup handler thread.
     */
    return Condition_InterruptSignal(&pIx->is_if.IntrPoint);
}

/* Thread-level interrupt handling
 */
BOOL MCT
I8255xInterrupt(PCDEVICE pDev)
{
    PI8255X pIx = (PI8255X) pDev;
    UINT8 Pending;

    //    DBGME(1,_tprintf(_T("Enter I8255xInterrupt(%x)\n"),pIx));

    for (;;) {

        Pending = IOSpaceReadInt8((IO_PORT)&pIx->Csr->SCB.StatAck);
        if (Pending == 0)
            break;

//        LogIt(_T("Intr, StatAck= %2x"),Pending,0);

        /* Acknowledge
         */
        IOSpaceWriteInt8((IO_PORT)&pIx->Csr->SCB.StatAck,Pending);

        /* Dispatch
         */
        if (Pending & IX_SA_CX) {
            IxXmitComplete(pIx);
        }
        if (Pending & IX_SA_FR) {
            IxRecvComplete(pIx);
        }
        if (Pending & IX_SA_RNR) {
            /* Kick it back into action ?*/
            LogIt(_T("RNR !!! %x"),Pending,0);
            IxRecvComplete(pIx);
            IxStartRU(pIx);
        }
#if _DEBUG
        if (Pending & (IX_SA_MDI|IX_SA_SWI))
            _tprintf(_T("InterruptPending %x not handled (%x)\n"),
                   Pending & (IX_SA_MDI|IX_SA_SWI), Pending);
#endif
    }

    /* Re-enable interrupts
     */
    IxEnableInterrupts(pIx);

    return TRUE;
}

/* Random other things..
 */
SCODE MCT
I8255xIoctl( PCDEVICE pDev, UINT Prop, caddr_t Data)
{
    PI8255X pIx = (PI8255X) pDev;
#if STATS
    UINT8 Status;
    int i;
    PIX_TX_BUFFER Tb;
    PIX_STATISTICAL_COUNTERS Stats = pIx->StatisticalCounters;
#endif

    DBGME(0,_tprintf(_T("Enter I8255xIoctl(%x)\n"),Prop));

#if STATS
    /* Freshen up statistics, if we need. [yes we do]
     */
    if (Prop == SIOCGIFSTATS) {

        /* Lock out senders, as we must idle the CU unit.
         */
        pIx->is_if.MBufFact->v->NetLock(pIx->is_if.MBufFact);

        /* The only way to idle it is to set the ELAST bit in a command.
         * Simply telling it to re-check the last one does not seem to work,
         * we really must issue a new one (like NOP, say). Sigh.
         */

        /* Wait until the CU goes into the suspended state (or idle, if we are lucky).
         * We post at most 7 packets of 1514*10 bits each, 105Kb max.
         * Ignoring IFS, that comes to about 10msecs at 10Mbps.
         * One tenth that at 100Mbps.
         */
        for (i = 0; i < 20; i++) {
            Status = IX_ST_CUS_MASK & IOSpaceReadInt8((IO_PORT)&pIx->Csr->SCB.Status);
            if ((Status == IX_ST_CUS_IDLE) || (Status == IX_ST_CUS_SUSPENDED))
                break;
            LogIt("Sleep st=%x i=%d",Status,i);
            MilliSleep(1);
        }

        if ((Status != IX_ST_CUS_IDLE) && (Status != IX_ST_CUS_SUSPENDED)) {
            LogIt("WontSusp! st=%x",Status,0);
            DBGME(3,_tprintf(_T("IxIoctl CU wont suspend %x\n"),Status));
            goto UnlockAndGo;
        }

        /* If its idle we can save time
         */
        if (Status == IX_ST_CUS_IDLE)
            goto InIdleState;

        /* Bummer. Prepare the NOP cmd
         */
        i = pIx->TxHead;
        Tb = pIx->TxBuffers + TxIndex(i);
        Tb->Tfd.TCB.h.Command = _bs16(IX_CBC_ELAST | IX_CBC_NOP);
        IOMemoryBarrier();

        LogIt("ELAST on %d Tb=%x",i,(ADDRESS)Tb);

        /* Tell it to recheck the previous one
         * NB: There *was* a previous one, it wouldnt be suspended otherwise.
         */
        i = pIx->TxHead - 1;  /* <0 ok */
        Tb = pIx->TxBuffers + TxIndex(i);
        Tb->Tfd.TCB.h.Command = _bs16(IX_CBC_NOP);
        IOMemoryBarrier();

        IOSpaceWriteInt8((IO_PORT)&pIx->Csr->SCB.Command,IX_CMD_CU_RESUME);
        IOWriteBufferFlush();

        /* Now wait until it goes idle. Shouldnt be long.
         */
        for (i = 0; i < 20; i ++) {
            Delay(100);
            Status = IX_ST_CUS_MASK & IOSpaceReadInt8((IO_PORT)&pIx->Csr->SCB.Status);
            if (Status == IX_ST_CUS_IDLE)
                break;
        }

        if (Status != IX_ST_CUS_IDLE) {
            LogIt("WontIdle! st=%x",Status,0);
            DBGME(3,_tprintf(_T("IxIoctl CU wont idle %x\n"),Status));
            goto UnlockAndGo;
        }

        /* A good time to dump stats
         * NB: We did NOT enable extended stats!
         */
    InIdleState:
        i = IX_SC_82557_STATUS;
        /* Has the dump stats command completed since the last time?
         */
        if (Stats->Counter[i] == _bs32(DumpCounterStatusOk)) {

            for (i = 0; i < IX_NUMBER_OF_COUNTERS_MAX; i++) {
                if (Stats->Counter[i]) {
                    Stats->Counter[i] = _bs32(Stats->Counter[i]);
#if _DEBUG
                    _tprintf(_T("CTR[%d] @%d = %x\n"),i,i*4,Stats->Counter[i]);
#endif
                }
            }
            LogIt("GotMilk",0,0);

            pIx->CollisionCount = Stats->Counter[IX_SC_COLLISION_T];
            pIx->MissedPackets  = Stats->Counter[IX_SC_RX_RESOURCES];
            pIx->MemoryErrors   = Stats->Counter[IX_SC_TX_UNDERRUN] +
                                  Stats->Counter[IX_SC_RX_OVERRUN];
            pIx->XmitTimeouts   = Stats->Counter[IX_SC_MAXCOL] +
                                  Stats->Counter[IX_SC_LATECOL];
            pIx->FrameErrors    = Stats->Counter[IX_SC_RX_CRC] +
                                  Stats->Counter[IX_SC_RX_ALIGNMENT] +
                                  Stats->Counter[IX_SC_RX_CDT] +
                                  Stats->Counter[IX_SC_RX_SHORT];
#if _DEBUG
#else
            pIx->XmitDeferred   = Stats->Counter[IX_SC_DEFERRED];
#endif
            pIx->XmitGood       = Stats->Counter[IX_SC_TGF];
            pIx->RecvGood       = Stats->Counter[IX_SC_RGF];

        }

        /* Issue a new dump command
         */
        Stats->Counter[i] = 0;
        Status = (IxDumpCounters(pIx)) ? 1 : 0;
        if (!Status) {
            LogIt("nocounters %x %d",Stats->Counter[i],Status);
            DBGME(3,_tprintf(_T("IxIoctl nocounters @%x\n"),(ADDRESS)Stats));
        }

        /* Let the lock go.  If anyone was waiting, they will find the CU idle
         * and they will restart it properly
         */
    UnlockAndGo:
        pIx->is_if.MBufFact->v->NetUnlock(pIx->is_if.MBufFact);

    }
#endif

    /* We care for requests to turn the interface down/up.
     */
    if (Prop == SIOCSIFFLAGS) {

        if ((pIx->is_if.if_flags & IFF_UP) == 0) {
            /*
             * Down. Put it in low power mode.
             */

        } else {
            /*
             * Up. Turn power back on, this causes an internal reset.
             */
            pDev->v->Initialize(pDev);
        }
    }

#if _DEBUG
    if (Prop == SIOCSIFDEBUG) {
        struct ifreq *rp = (struct ifreq *)Data;
        static NAME CuState[4] = {
            _T("Idle"), _T("Suspended"), _T("LPQ-active"), _T("HQP-active")
        };
        static NAME RuState[16] = {
            _T("Idle"), _T("Suspended"), _T("NoResources"), _T("3"),
            _T("Ready"), _T("5"), _T("6"), _T("7"),
            _T("8"), _T("9"), _T("A"), _T("B"),
            _T("C"), _T("D"), _T("E"), _T("F")
        };
        UINT8 iStatus = IOSpaceReadInt8((IO_PORT)&pIx->Csr->SCB.Status);

        _tprintf(_T("CU is %s RU is %s\n"),
               CuState[(iStatus&IX_ST_CUS_MASK)>>6],
               RuState[(iStatus&IX_ST_RUS_MASK)>>2]);

        IxDumpAllRegs(pIx,rp->ifr_debug.param > 2);
        IxDumpTfds(pIx,pIx->TxTail,pIx->TxHead);
        IxDumpRfds(pIx);
        if (rp->ifr_debug.param > 1)
            PrintZeLog();
    }
#endif

    return NO_ERROR;
}


/* Initialize/Reset/Reinitialize the card
 * NB: Assumes netlock held.
 */
INT MCT
I8255xReset(PCDEVICE pDev)
{
    PI8255X pIx = (PI8255X) pDev;
    struct ifnet *ifp;

    IxDisableAllInterrupts(pIx);

    LogIt(_T("Reset no=%8d"),pIx->CompleteResets,0);

    DBGME(2,_tprintf(_T("Enter I8255xReset\n")));
    DBGME(0,IxDumpAllRegs(pIx,TRUE));

    IxDumpConfiguration(pIx,_T("InReset"));

    /* Say we are down and reinit sw state
     */
    ifp = &pIx->is_if;
    if (ifp->if_flags & IFF_UP) {
        ifp->if_flags &= ~IFF_UP;
    }
    pIx->CompleteResets++;

    /* Can we allocate buffers here?
     * Problem is we dont have the factory pointer while in Initialize. Sigh.
     */
    if (pIx->RxHead == NULL) {
        int i;
        PIX_RX_BUFFER Rb;
        UINT32 Value32;

        /* Grab a few receive buffers
         */
        for (i = 0; i < nRxBuffers; i++) {
            if (!IxAllocateReceiveBuffer(pIx,&Rb)) {
                DBGME(3,_tprintf(_T("IxInit ARB nomem (%d)\n"),i));
                break;
            }
            if (pIx->RxHead == NULL) {
                pIx->RxHead = Rb;
                pIx->RxTail = Rb;
            } else {
                Value32 = VirtualToPhysical(&Rb->Rfd);
                pIx->RxTail->Rfd.h.Command = 0;
                pIx->RxTail->Rfd.h.LinkAddress = _bs32(Value32);
                pIx->RxTail = Rb;
            }
        }

        /* Did we allocate at least one buffer
         */
        if (pIx->RxHead == NULL) {
            DBGME(3,_tprintf(_T("IxReset nobufs\n")));
            return 1;
        }

        DBGME(0,IxDumpRfds(pIx));
    }

    /* See if the link is up already
     */
    IxLinkStatus(pIx);

    /* Start the RU
     */
    IxStartRU(pIx);

    /* Say we are up, but wait for AN complete to see if we are running
     */
    ifp->if_flags |= IFF_UP;

    /* Enable Interrupts
     */
    IxEnableInterrupts(pIx);

    /* We are rolling
     */
    DBGME(0,_tprintf(_T("I8255x:Reset done\n")));
    return 0;

}

/* Watchdog, looks for link up/down events
 */
INT MCT
I8255xWatchdog(PCDEVICE pDev)
{
    PI8255X pIx = (PI8255X) pDev;
    struct ifnet *ifp = &pIx->is_if;

    /* Check on the link status
     */
    IxLinkStatus(pIx);

    /* We want to be called back in if_slowtimo*N=3secs
     */
    ifp->if_timer = 3;
    return 0;
}

/* Stop&Unload method
 * UNTESTED
 */
SCODE MCT
I8255xUnload(PCDEVICE pDev)
{
    PI8255X pIx = (PI8255X) pDev;
    struct ifnet *ifp = &pIx->is_if;

    /* Kick the watch dog far out
     */
    ifp->if_timer = ~0;

    /* Stop brutally the hw by writing the PCIconfig
     */

    /* Detach the Isr from our interrupt
     */

    /* Tell ThreadLoop to terminate
     */

    /* Free all PMBUFS, both recv and xmit
     */

    /* Call ifdetach
     */

    /* Free pIx->CommandBuffer
     */

    /* Zero all fields, mark for recycling
     * Wait for the Release() method to free pIx itself.
     */

    /* Done
     */
    return S_OK;
}

struct INetDriverVtbl I8255xDriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    (_INetDriver_Isr_METHOD_TYPE)  I8255xIsr,
    (_INetDriver_HandleInterrupt_METHOD_TYPE) I8255xInterrupt,
    (_INetDriver_Initialize_METHOD_TYPE) I8255xReset,
    NULL,
    NULL,
    (_INetDriver_Start_METHOD_TYPE) I8255xXmit,
    (_INetDriver_Ioctl_METHOD_TYPE) I8255xIoctl,
    (_INetDriver_Watchdog_METHOD_TYPE) I8255xWatchdog
    // (_INetDriver_Unload_METHOD_TYPE) I8255xUnload
};

/* Once-only initializations.
 */
PRIVATE
void
IxInitializeTfds(PI8255X pIx)
{
    int i;
    PIX_TX_BUFFER Tb = pIx->TxBuffers;
    UINT32 Value32;

    DBGME(0,_tprintf(_T("Enter IxInitializeTfds\n")));

    /* Link them all in a circle
     */
    for (i = 0; i < nTxBuffers; i++) {
        /* Sanity zeroes
         */
        memset(Tb+i,0,sizeof(IX_TX_BUFFER));

        /* Link to th enext one
         */
        Value32 = VirtualToPhysical(&Tb[TxIndex(i+1)].Tfd.TCB.h);
        Value32 = _bs32(Value32);
        Tb[i].Tfd.TCB.h.LinkAddress = Value32;

        /* Array desc
         */
        Value32 = VirtualToPhysical(&Tb[i].Tfd.TBD[0]);
        Value32 = _bs32(Value32);
        Tb[i].Tfd.TCB.TbdArrayAddress = Value32;

        /* This is times-8
         */
        Tb[i].Tfd.TCB.Threshold = 1;
    }

    DBGME(0,IxDumpTfds(pIx,0,nTxBuffers));
}

PRIVATE
PI8255X
IxInitialize(UINT32 PciLocation)
{
    PIX_CSR Csr;
    PI8255X pIx = NULL;
    UINT Bus, Device, Function;
    UINT32 Value32;
    UINT16 Value16 = 0;
    PIHEAP pIHeap = CurrentHeap();
    PIBUSDRIVER OurPci = NULL;
    SCODE sc;
    static short unit=0;

    Bus = PciLocation >> 16;
    Device = (PciLocation >> 8) & 0xff;
    Function = PciLocation & 0xff;

    DBGME(1,_tprintf(_T("Enter IxInitialize(%d,%d,%d)\n"),
                   Bus,Device,Function));

#if LOGIT
    _tprintf(_T("I8255x Log @%x\n"),(ADDRESS) &LogBuf[0]);
#endif

    /* Make sure we know where to go
     */
    sc = BindToObject(CurrentNameSpace(),
                      PCIDRIVER_NAME,
                      NAME_SPACE_READ | NAME_SPACE_WRITE,
                      &IID_IBusDriver,
                      (void**)&OurPci);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("No PCI access routine!!\n")));
        goto ErrorExit;
    }

    /* Look at PCI space, check its there
     */
    Value32 = 0;
    OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                      0, &Value32, sizeof(Value32));
    if (Value32 != IX_8255x_PCI_SIGNATURE) {
        DBGME(3,_tprintf(_T("Signature at PCI %x (%x) not for I8255x (%x)\n"),
                       PciLocation,Value32,IX_8255x_PCI_SIGNATURE));
        goto ErrorExit;
    }
#if WI
    /* Verify that the cache line size matches
     */
    {
        UINT8 Value8;
        Value8 = 0;
        OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                          12, &Value8, sizeof(Value8));
        Value32 = Value8 * 4;
        if (Value32 != _DCACHELINESIZE) {
            DBGME(3,_tprintf(_T("CacheLineSize at PCI %x: %d should be %d\n"),
                             PciLocation,Value32,_DCACHELINESIZE));
        }
    }
#endif

    /* Grab our memory resources
     * Do this now, before enabling the device
     */
    pIx = (PI8255X) pIHeap->v->Alloc(pIHeap, HEAP_ZERO_MEMORY,
                                     sizeof(I8255X), 0);
    if (pIx == NULL)
        goto ErrorExit;

    pIx->CommandBuffer = pIHeap->v->Alloc(pIHeap, HEAP_CACHE_ALIGN,
                                          sizeof(IX_COMMAND) +
                                          sizeof(IX_STATISTICAL_COUNTERS) +
                                          nTxBuffers * sizeof(IX_TX_BUFFER),
                                          0);
    if (pIx->CommandBuffer == NULL)
        goto ErrorExit;

    /* Stats right after the command buffer
     */
    pIx->StatisticalCounters = (IX_STATISTICAL_COUNTERS*)(pIx->CommandBuffer + 1);
    memset(pIx->StatisticalCounters,0,sizeof(IX_STATISTICAL_COUNTERS));

    /* TFDs right after the stats
     */
    pIx->TxBuffers = (PIX_TX_BUFFER)(pIx->StatisticalCounters + 1);
    IxInitializeTfds(pIx);

    /* Basic initializations
     */
    Mutex_Init(&pIx->Lock);
    pIx->TxTail = pIx->TxHead = 0;

    /* Remember what revision of the chip we have
     */
    OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                      8, &pIx->ChipRevision, 1);

    /* Look at the assigned I/O location
     */
    Value32 = 0;
    OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                      16, &Value32, sizeof(Value32));
    Csr = (PIX_CSR) IoPhysicalToIoVirtual(Value32);

    /* Remember the address of the SCB
     */
    pIx->Csr = Csr;

    /* Enable it, by writing to the Command register
     * No: back-to-back,SERR#, wait cycle, parity, palette snoop, I/O
     * Yes: MWI, master, memory
     */
#if WI
    Value16 = 0x0016;
#else
    Value16 = 0x06;
#endif
    OurPci->v->WriteConfigurationSpace(OurPci,Bus,Device,Function,
                                      4, &Value16, sizeof(Value16));
    Delay(20);

    /* Do a selective before full reset (says them)
     * Wait ~20us on each one.
     */
    IOSpaceWriteInt32((IO_PORT)&Csr->Port,_bs32(IX_P_SELECTIVE_RESET));
    IOWriteBufferFlush();
    Delay(30);

    IOSpaceWriteInt32((IO_PORT)&Csr->Port,_bs32(IX_P_SOFT_RESET));
    IOWriteBufferFlush();
    Delay(30);

    /* Disable interrupts, JIC
     */
    IxDisableAllInterrupts(pIx);

    /* Zero the CU base 
     */
    (void) IxScbCommand(Csr,IX_CMD_CU_LOAD_BASE,0);
    MilliSleep(1);

    /* Same for the RU base
     */
    (void) IxScbCommand(Csr,IX_CMD_RU_LOAD_BASE,0);
    MilliSleep(1);

    /* Read the PHY's ID
     */
    IxMdiRead(Csr,IX_MDI_ID0,&Value16);
    pIx->PhyOui  = Value16 << 16;
    IxMdiRead(Csr,IX_MDI_ID1,&Value16);
    pIx->PhyOui |= Value16;

    /* Say what we believe in
     */
    DBGME(1,_tprintf(_T("Intel 8255%s @%x PCI bus %d, device %d, function %d"),
             RevisionName(pIx->ChipRevision), (ADDRESS)Csr, 
                     Bus, Device, Function));
    DBGME(1,_tprintf(_T(", PHY id %x\n"),pIx->PhyOui));

    DBGME(0,IxDumpAllRegs(pIx,TRUE));
    DBGME(0,IxDumpEeprom(pIx));

    /* Read the station address off the EEPROM
     */
    IxReadStationAddress(pIx,&pIx->is_addr[0]);

    IxDumpConfiguration(pIx,_T("BeforeConfig"));

    /* Load the configuration data
     */
    if (!IxLoadConfiguration(pIx)) {
        DBGME(3,_tprintf(_T("I8255x load config failed, ignoring\n")));
    }
    MilliSleep(1);

    /* Load the station address on chip
     */
    if (!IxLoadStationAddress(pIx)) {
        DBGME(3,_tprintf(_T("I8255x load MACaddr failed, ignoring\n")));
    }
    MilliSleep(1);
    
#if STATS
    /* Set the output buffer for stats.
     */
    Value32 = VirtualToPhysical(pIx->StatisticalCounters);
    if (!IxScbCommand(Csr,IX_CMD_CU_LOAD_DCA,Value32)) {
        DBGME(3,_tprintf(_T("I8255x load dca failed, ignoring\n")));
    }
    MilliSleep(10);
#endif

    /* More init
     */
    pIx->is_if.v = &I8255xDriverVtable;
    pIx->is_if.RefCnt = 1;
    pIx->is_if.if_unit = unit++;
    pIx->is_if.if_name = "ix";
    pIx->is_if.if_type = IFT_ETHER;
    pIx->is_if.if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST | IFF_ZEROCOPY;
    pIx->is_if.if_baudrate = 0; /* until we know better */

    DBGME(2,_tprintf(_T("I8255x CardAddr %02x:%02x:%02x:%02x:%02x:%02x Base %x\n"),
                   pIx->is_addr[0], pIx->is_addr[1],
                   pIx->is_addr[2], pIx->is_addr[3],
                   pIx->is_addr[4], pIx->is_addr[5],
                   (ADDRESS) pIx->Csr));

    return pIx;

ErrorExit:
    if (OurPci)
        OurPci->v->Release(OurPci);
    if (pIx) {
        /* NB: No paths can/should lead here with allocated receive buffers */
        assert(pIx->RxHead == NULL);

        if (pIx->CommandBuffer)
            pIHeap->v->Free(pIHeap,0,pIx->CommandBuffer);
        pIHeap->v->Free(pIHeap,0,pIx);
    }
    return NULL;
}

SCODE __cdecl I8255xConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk, ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf("I8255xConstructor(%x %s %x %x %x)\n",
                   (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                   (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) IxInitialize((UINT32)RegisterBase);

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("I8255xConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl I8255xDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) I8255xConstructor
};

static struct _DRIVERCOB I8255xDriverCobObject = { &I8255xDriverCobVtable, 1 };

PIUNKNOWN I8255xCobMain(void)
{
    SCODE sc;
    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &I8255xDriverCobVtable,
                          "CDeviceConstructor",
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &I8255xDriverVtable, "CNetDriver",
                          &IID_INetDriver);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &I8255xDriverCobObject;
}

EXTERN_C PCDEVICE RTLCALLTYPE 
I8255xNew(PUINT32 pPortBase, PUINT pIrq)
{
    PI8255X pIx;
    SCODE sc;

    sc = BaseInheritClass((ADDRESS *) &I8255xDriverVtable, "CNetDriver",
                          &IID_INetDriver);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    pIx = IxInitialize(*pPortBase);
    if (pIx == NULL)
        return NULL;

    return (PCDEVICE) pIx;
}

