/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Defines for Lan91c96 chip, according to SMSC specs.
 * Superset of previous chips (91c90/92/94/95/100)
 */

/* MMU data
 */

#define LAN_MMU_RAM     (6*1024)
#define LAN_MMU_NPAGES  24
#define LAN_MMU_PPP     6       /* max pages per packet */
#define LAN_MMU_PAGSIZ  256

/* Attribute space memory map (aka Bank4 registers)
 */

#define LAN_ATT_EPROM   0
#define LAN_ATT_CONFIG  0x8000  /* following are offsets from base */
# define LAN_ATT_ECOR   (0)
# define LAN_ATT_ECSR   (2)
/* NB: Could also look at ECOR+ECSR as a single 16bit register */

/* Ethernet configuration option register (ECOR)
 */
#define LAN_ECOR_SRESET         0x80
#define LAN_ECOR_LevIRQ         0x40
#define LAN_ECOR_WRATTRIB       0x04
#define LAN_ECOR_ENABLE         0x01

/* Ethernet configuration and status register (ECSR)
 */
#define LAN_ECSR_IOIs8          0x20
#define LAN_ECSR_PwrDown        0x04
#define LAN_ECSR_Intr           0x02

/* Packet format in buffer memory
 */
typedef struct {
    UINT16      Status;
    UINT16      ByteCount;      /* only 11 bits, always even */
    UINT16      Data[1];        /* varsize, max 1536-5 bytes */

    /* Last byte is a ControlByte: */
# define LAN_CTL_RECV           0x40
# define LAN_CTL_ODD            0x20
# define LAN_CTL_ODD_SHIFT      5
# define LAN_CTL_CRC            0x10

} LANPKT, *PLANPKT;

/* Receive status word, in front of pkts
 */
#define LAN_RS_ALGN_ERR         0x8000
#define LAN_RS_BROADCAST        0x4000
#define LAN_RS_BAD_CRC          0x2000
#define LAN_RS_ODDFRM           0x1000
#define LAN_RS_TOOLNG           0x0800
#define LAN_RS_TOOSHORT         0x0400
# define LAN_RS_ERRMASK         0xAC00
#define LAN_RS_HASHVALUE        0x007E
# define LAN_RS_HASHSHFT        1
#define LAN_RS_MULTCAST         0x0001

/* Register offsets
 */

#define LAN_BSR                 0xE     /* Common to all banks */

/* Bank0
 */
#define LAN_TCR                 0x0
#define LAN_EPHSR               0x2
#define LAN_RCR                 0x4
#define LAN_ECR                 0x6
#define LAN_MIR                 0x8
#define LAN_MCR                 0xA

/* Bank1
 */
#define LAN_CR                  0x0
#define LAN_BAR                 0x2
#define LAN_IA01                0x4
#define LAN_IA23                0x6
#define LAN_IA45                0x8
#define LAN_GPR                 0xA
#define LAN_CTR                 0xC

/* Bank2
 */
#define LAN_MMUCR               0x0
#define LAN_AUTOTX              0x1
#define LAN_PNR                 0x2
#define LAN_ARR                 0x3
#define LAN_FIFO                0x4
#define LAN_PTR                 0x6
#define LAN_DATA0               0x8
#define LAN_DATA1               0xA
#define LAN_INTERRUPT           0xC
#define LAN_IST                 0xC
#define LAN_ACK                 0xC
#define LAN_MSK                 0xD

/* Bank3
 */
#define LAN_MULTICAST_TABLE_0   0x0
#define LAN_MULTICAST_TABLE_1   0x2
#define LAN_MULTICAST_TABLE_2   0x4
#define LAN_MULTICAST_TABLE_3   0x6
#define LAN_MGMT                0x8
#define LAN_REV                 0xA
#define LAN_ERCV                0xC

/* Bank select register
 */
#define LAN_BSR_FIXED           0xFF00
#define LAN_BSR_SIGNATURE       0x3300
#define LAN_BSR_BANK            0x0003
#define LAN_BSR_DISABLE         0x0004

/* Transmit control register
 */
#define LAN_TCR_FDSE            0x8000
#define LAN_TCR_ETEN            0x4000
#define LAN_TCR_EPH_LOOP        0x2000
#define LAN_TCR_STP_SQET        0x1000
#define LAN_TCR_FDUPLX          0x0800
#define LAN_TCR_MON_CSN         0x0400
#define LAN_TCR_NOCRC           0x0100
#define LAN_TCR_PAD_EN          0x0080
#define LAN_TCR_TXP_EN          0x0008
#define LAN_TCR_FORCOL          0x0004
#define LAN_TCR_LOOP            0x0002
#define LAN_TCR_TXENA           0x0001

/* EPH Status register
 */
#define LAN_EPHSR_TXUNRN        0x8000
#define LAN_EPHSR_LINK_OK       0x4000
#define LAN_EPHSR_CTR_ROL       0x1000
#define LAN_EPHSR_EXC_DEF       0x0800
#define LAN_EPHSR_LOST_CARR     0x0400
#define LAN_EPHSR_LATCOL        0x0200
#define LAN_EPHSR_WAKEUP        0x0100
#define LAN_EPHSR_TX_DEFR       0x0080
#define LAN_EPHSR_LTX_BRD       0x0040
#define LAN_EPHSR_SQET          0x0020
#define LAN_EPHSR_16COL         0x0010
#define LAN_EPHSR_LTX_MULT      0x0008
#define LAN_EPHSR_MULCOL        0x0004
#define LAN_EPHSR_SNGLCOL       0x0002
#define LAN_EPHSR_TX_SUC        0x0001

/* Receive control register
 */
#define LAN_RCR_SOFT_RST        0x8000
#define LAN_RCR_FILT_CAR        0x4000
#define LAN_RCR_STRIP_CRC       0x0200
#define LAN_RCR_RXEN            0x0100
#define LAN_RCR_ALMUL           0x0004
#define LAN_RCR_PRMS            0x0002
#define LAN_RCR_RX_ABORT        0x0001

/* Counter register
 */
#define LAN_ECR_EXC_DEF_TX      0xF000
#define LAN_ECR_DEF_TX          0x0F00
#define LAN_ECR_MULTCOL         0x00F0
#define LAN_ECR_SNGLCOL         0x000F

/* Memory information register
 * NB: Multiply by 2 on the 91c100 (see below)
 */
#define LAN_MIR_FREE            0xFF00
#define LAN_MIR_SIZE            0x00FF

/* Memory configuration register
 */
#define LAN_MCR_FIXED           0xF100
#define LAN_MCR_SIGNATURE       0x3100
#define LAN_MCR_MULTIPLIER      0x0E00
#define LAN_MCR_MEM_RESERVE_TX  0x00FF

/* Configuration register
 */
#define LAN_CR_NO_WAIT          0x1000
#define LAN_CR_FULL_STEP        0x0400
#define LAN_CR_SET_SQLCH        0x0200
#define LAN_CR_AUI_SELECT       0x0100
#define LAN_CR_16BIT            0x0080
#define LAN_CR_DIS_LINK         0x0040
#define LAN_CR_INT_SEL1         0x0004
#define LAN_CR_INT_SEL0         0x0002

/* Base address register
 */
#define LAN_BAR_A15             0x8000  /* 0 at reset */
#define LAN_BAR_A14             0x4000  /* 0 at reset */
#define LAN_BAR_A13             0x2000  /* 0 at reset */
#define LAN_BAR_A9              0x1000  /* 1 at reset */
#define LAN_BAR_A8              0x0800  /* 1 at reset */
#define LAN_BAR_A7              0x0400  /* 0 at reset */
#define LAN_BAR_A6              0x0200  /* 0 at reset */
#define LAN_BAR_A5              0x0100  /* 0 at reset. A4,A10-12 mbz */
#define LAN_BAR_ROM_SIZE        0x00C0
#define LAN_BAR_RA18            0x0020
#define LAN_BAR_RA17            0x0010
#define LAN_BAR_RA16            0x0008
#define LAN_BAR_RA15            0x0004
#define LAN_BAR_RA14            0x0002

/* Individual address registers
 * Bit 0 of first register is first bit out on cable
 */

/* General purpose (address) register
 */
#define LAN_GPR_HIGH_BYTE       0xFF00 /* (you'd hope..) */
#define LAN_GPR_LOW_BYTE        0x00FF

/* Control register
 */
#define LAN_CTR_RCV_BAD         0x4000
#define LAN_CTR_PWRDN           0x2000
#define LAN_CTR_WAKEUP_EN       0x1000
#define LAN_CTR_AUTO_RELEASE    0x0800
#define LAN_CTR_FIXED           0x0100
#define LAN_CTR_LE_ENABLE       0x0080
#define LAN_CTR_CR_ENABLE       0x0040
#define LAN_CTR_TE_ENABLE       0x0020
#define LAN_CTR_EEPROM_SELECT   0x0004
#define LAN_CTR_RELOAD          0x0002
#define LAN_CTR_STORE           0x0001

/* MMU Command register
 */
#define LAN_MMUCR_COMMAND       0xF0
# define LAN_CMD_NOOP            0x00
# define LAN_CMD_TX_ALLOCATE     0x20
# define LAN_CMD_RESET_MMU       0x40
# define LAN_CMD_POP_RX_FIFO     0x60
# define LAN_CMD_POP_TX_FIFO     0x70
# define LAN_CMD_POPREL_RX_FIFO  0x80
# define LAN_CMD_REL_PACKET      0xA0
# define LAN_CMD_ENQ_PKT         0xC0
# define LAN_CMD_RESET_TX_FIFO   0xF0
#define LAN_MMUCR_N2            0x04
#define LAN_MMUCR_N1            0x02
#define LAN_MMUCR_N0            0x01
#define LAN_MMUCR_BUSY          0x01

/* Auto TX start register
 * Value is in 16x delay, 0 -> disable
 */

/* Packet number register
 */
#define LAN_PNR_MASK            0x1f

/* Allocation result register
 */
#define LAN_ARR_FAILED          0x80
#define LAN_ARR_MASK            LAN_PNR_MASK

/* FIFO ports register
 */
#define LAN_FIFO_REMPTY         0x8000
#define LAN_FIFO_RX_PKT_NUM     0x1F00
#define LAN_FIFO_TEMPTY         0x0080
#define LAN_FIFO_TX_PKT_NUM     0x001F

/* Pointer register
 */
#define LAN_PTR_RCV             0x8000
#define LAN_PTR_AUTO_INCR       0x4000
#define LAN_PTR_READ            0x2000
#define LAN_PTR_ETEN            0x1000
#define LAN_PTR_AutoTx          0x0800
#define LAN_PTR_POINTER         0x07FF

/* Interrupt status register
 * NB: joint work with MSK
 */
#define LAN_IST_TX_IDLE         0x80
#define LAN_IST_ERCV            0x40
#define LAN_IST_EPH             0x20
#define LAN_IST_RX_OVRN         0x10
#define LAN_IST_ALLOC           0x08
#define LAN_IST_TX_EMPTY        0x04
#define LAN_IST_TX              0x02
#define LAN_IST_RCV             0x01

/* Interrupt acknowledge register
 */
#define LAN_ACK_ERCV            0x40
#define LAN_ACK_RX_OVRN         0x10
#define LAN_ACK_TX_EMPTY        0x04
#define LAN_ACK_TX              0x02

/* Management interface register
 */
#define LAN_MGMT_nXDEC          0x0800
#define LAN_MGMT_IOS2           0x0400
#define LAN_MGMT_IOS1           0x0200
#define LAN_MGMT_IOS0           0x0100
#define LAN_MGMT_MDOE           0x0008
#define LAN_MGMT_MCLK           0x0004
#define LAN_MGMT_MDI            0x0002
#define LAN_MGMT_MDO            0x0001

/* Revision register
 */
#define LAN_REV_FIXED           0x3300
#define LAN_REV_CHIP            0x00F0
#define LAN_REV_STEP            0x000F

#define LAN_REV_90_92           0x0030
#define LAN_REV_94_96           0x0040 /* 96 starts at STEP=6 */
#define LAN_REV_95              0x0050
#define LAN_REV_100             0x0070
#define LAN_REV_100FD           0x0080
#define LAN_REV_110             0x0090

/* Early receive register
 */
#define LAN_ERCV_COUNTER        0xFF00
#define LAN_ERCV_DISCARD        0x0080
#define LAN_ERCV_THRESHOLD      0x001F
