/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    lance.c
 *
 * Purpose:
 *
 *    Device driver for the Allied Telesis PC-ISA Lance
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

/*
 * Code in here was derived from Mach sources:
 *
 * Mach Operating System
 * Copyright (c) 1993-1989 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 *      AMD Am7990 LANCE (Ethernet Interface)
 *      Author: Robert V. Baron & Alessandro Forin
 *      Date:   5/90
 */

#define USE_INTRINSICS
#include <mmlite.h>
#include <mmhal.h>
#include <sock.h>
#include <if.h>
#include <if_types.h>
#include <ip/in.h>
#include <ip/arp.h>
#include <stdio.h>
#include <drivers.h>
#ifdef _MSC_VER
#pragma warning(disable:4214)   /* nonstandard: bit field types other than int */
#pragma warning(disable:4055)   /* cast from data pointer to function pointer  */

#endif
#include <lance.h>

typedef BOOL (RTLCALLTYPE *SIGNAL)(PCONDITION);


#define _LANCE_DEBUG 0

#if _LANCE_DEBUG
#define DBGME(x) printf x
#else
#define DBGME(x)
#endif

#define RLOG    6               /* 2**6 = 64  receive descriptors */
#define TLOG    4               /* 2**4 = 16  transmit descriptors */
#define NRCV    (1<<RLOG)       /* Receive descriptors */
#define NXMT    (1<<TLOG)       /* Transmit descriptors */

#define LN_BUFFER_SIZE  (1520) /* 0x5F0 */
#define LN_BUFFERS (LN_BUFFER_SIZE*(NRCV+NXMT))
#define LN_OVERHEAD (sizeof(struct se_init_block)+12+(sizeof(struct se_desc)*(NXMT+NRCV)))
#define LN_MEMORY_SIZE  (LN_BUFFERS+LN_OVERHEAD)

/*
 * Lance descriptor
 */
typedef struct  _LANCE {
        struct arpcom   ac;             /* standard interface header    */
#define is_if ac.ac_if
#define is_addr ac.ac_enaddr
#define misscnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_MISS]
#define rstrtcnt ac.ac_if.if_data.ifi_errors[IFI_ERR_RESET]
#define bablcnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_XTIM]
#define merrcnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM]
#define sqercnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+1]
#if _LANCE_DEBUG
#define spurious  ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+2]
#define sawcsr   ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+3]
#endif

        SIGNAL          Signal;

        IO_PORT         rap;            /* Lance registers      */
        IO_PORT         rdp;            /* Lance registers      */
        UINT            MemBaseV;       /* Lance memory, Host virtual */
        UINT            MemBaseP;       /* Lance memory, Host physical */
        UINT            TopAlloc;       /* Lance memory allocator */
        UINT            InitBlock;      /* Init block address   */
        PIDMA           pDma;           /* Our DMA channel      */
        BOOL            DoStartDma;     /* Need StartXfer ?     */

        se_desc_t       RRing[NRCV];    /* Receive  ring desc. */
        volatile PTR    RBuf[NRCV];     /* Receive  buffers */
        se_desc_t       TRing[NXMT];    /* Transmit ring desc. */
        volatile PTR    TBuf[NXMT];     /* Transmit buffers */

        UINT    rcv_last;               /* Rcv buffer last read         */

        UINT    xmt_last;               /* Xmt queue head (insert)      */
        UINT    xmt_complete;           /* Xmt queue tail (remove)      */

} LANCE, *PLANCE;

/*
 * Functions we export to the upper layers (vtable)
 */
PRIVATE INT MCT LanceRestart( PLANCE sc);
PRIVATE void MCT LanceXmit( PLANCE sc);
PRIVATE SCODE MCT LanceIoctl( PLANCE sc, UINT Prop, PTR Data);
PRIVATE BOOL MCT LanceIsr( PLANCE sc, PBOOL pNotMyInterrupt);
PRIVATE BOOL MCT LanceInterrupt( PLANCE sc);

#pragma warning(disable:4028)   /* formal parameter 1 different.. */

PRIVATE struct INetDriverVtbl LanceVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    LanceIsr,       /* Isr          */
    LanceInterrupt, /* Intr         */
    LanceRestart,   /* Init         */
    NULL,
    NULL,    /* Output       */
    LanceXmit,      /* Start output */
    LanceIoctl,     /* Ioctl        */
    NULL            /* Watchdog     */
};

#pragma warning(default:4028)   /* reenable warning C4028 */

/*
 * Access to registers
 */

#define LanceReadRegister(_r_)    IOSpaceReadInt16(_r_)

PRIVATE void
LanceWriteRegister(IO_PORT reg, UINT16 Value, UINT16 Result, char *regname)
{
    int    i = 0;

    while ( IOSpaceReadInt16(reg) != (UINT16)Result) {
        IOSpaceWriteInt16(reg,Value);
        IOWriteBufferFlush();
        if (++i > 10000) {
            printf(_T("se: %hs did not settle (to x%x): x%x\n"),
                   regname, Result, IOSpaceReadInt16(reg));
            return;
        }
        Delay(100);
    }
}

/*
 * Interrupt munging
 */
#define LanceEnableInterrupts(_r_) IOSpaceWriteInt16(_r_, LN_CSR0_INEA )
#define LanceDisableInterrupts(_r_) IOSpaceWriteInt16(_r_,0)


/*
 Debug
 */
#if _LANCE_DEBUG
void LanceDump( PLANCE sc)
{
    UINT16 i, j, r112, r114;
    IO_PORT isacr;
    UINT8 RelevantPorts[] = {
        2, 1, 4, 12, 13, 14,
        6, 15, 17, 16, 21, 20,
        19, 18, 23, 22, 25, 24,
        29, 28, 27, 26, 31, 30,
        35, 34, 33, 32, 41, 40,
        45, 44, 43, 42, 63, 62,
        67, 66, 61, 60, 65, 64,
        46, 72, 74, 76, 78, 80,
        85, 84, 86, 89, 94
    };

    /* Pick up 112 and 114 before stop
     */
    LanceWriteRegister(sc->rap, 112, 112, "RAP");
    r112 = LanceReadRegister(sc->rdp);

    LanceWriteRegister(sc->rap, 114, 114, "RAP");
    r114 = LanceReadRegister(sc->rdp);

    /* Now stop and dump
     */
    LanceWriteRegister(sc->rap, 0, 0, "RAP");
    LanceWriteRegister(sc->rdp, LN_CSR0_STOP, LN_CSR0_STOP, "csr0");

    printf(_T("Lance register dump (after STOP):\n"));
    printf(_T("  112 %4x  114 %4x"), r112, r114);
    for (i = 0, j = 2; i < sizeof RelevantPorts; j = 0) {
        for (; j < 6 && i < sizeof RelevantPorts; j++) {
            IO_PORT p = (IO_PORT) RelevantPorts[i];
            LanceWriteRegister(sc->rap, p, p, "RAP");
            printf(_T("  %03.3d %4x"), p, LanceReadRegister(sc->rdp));
            i++;
        }
        printf(_T("\n"));
    }

    isacr = (UINT16)(sc->rdp + (0x16-0x10));
    printf(_T("ISACR register dump:\n"));
    for (i = 0; i < 8; ) {
        LanceWriteRegister(sc->rap, i, i, "RAP");
        printf(_T(" r%d %4x"), i, LanceReadRegister(isacr));
        i++;
        LanceWriteRegister(sc->rap, i, i, "RAP");
        printf(_T("    r%d %4x"), i, LanceReadRegister(isacr));
        i++;
        LanceWriteRegister(sc->rap, i, i, "RAP");
        printf(_T("    r%d %4x"), i, LanceReadRegister(isacr));
        i++;
        LanceWriteRegister(sc->rap, i, i, "RAP");
        printf(_T("    r%d %4x\n"), i, LanceReadRegister(isacr));
        i++;
    }
}
#endif

/*
 * Map a lance-space offset into an host-space virt address
 * Map a lance-space offset into a lance-space phys address
 */
#define Hmem(lna)       (PTR)((BYTE*)(lna) + sc->MemBaseV)
#define Lmem(lna)       (PTR)((BYTE*)(lna) + sc->MemBaseP)

/*
 * Memory addresses for LANCE are 24 bits wide.
 */
#define AddressLowPart(y)       ((UINT16)((ADDRESS)(y) & 0xffff))
#define AddressHighPart(y)      ((UINT8 )(((ADDRESS)(y)>>16) & 0xff))

/*
 *
 * Internal functions & definitions
 *
 */

PRIVATE BOOL LanceLoadIBlock( PLANCE sc);
PRIVATE void InitLanceSpace( PLANCE sc);
PRIVATE void LanceDescSet (se_desc_t lndesc, int val);
PRIVATE volatile PTR LanceDescAlloc (PLANCE sc, se_desc_t lndesc);
PRIVATE struct mbuf *LanceRecvPkt(PLANCE sc, volatile UINT8 *RBuf, int len, 
                                  struct ether_header *eh);
PRIVATE int LanceSendPkt(struct mbuf *m, volatile UINT8 *SBuf);
PRIVATE void LanceRecvInterrupt( PLANCE sc );
PRIVATE void LanceXmitInterrupt( PLANCE sc );

PRIVATE UINT se_malloc();

/* These might become messy, someday */
#define LanceDescCopyin(_f_,_t_,_n_) memcpy(_t_,_f_,_n_)
#define LanceDescCopyout(_f_,_t_,_n_) memcpy(_t_,_f_,_n_)
#define LanceDataCopyin(_f_,_t_,_n_) memcpy(_t_,_f_,_n_)
#define LanceDataCopyout(_f_,_t_,_n_) memcpy(_t_,_f_,_n_)
#define LanceDataOffset(_n_) _n_

/*
 * Driver initialization function, aka contructor for
 * a LANCE object.
 * This function assumes that a lance
 * hardware does exist at the given PortBase.
 * Returns a properly casted descriptor.
 */
PUBLIC PINETDRIVER RTLCALLTYPE LanceNew( PUINT pPortBase, PUINT pIrq, PUINT pDmaCh)
{
    PLANCE       sc;
    IO_PORT      rdp, rap;
    UINT8        *enaddr;
    int          i;
#ifdef VM
    PIADDRESSSPACE    pIAddr;
#else
    PIHEAP pIHeap;
#endif
    SCODE        Rc;
    ADDRESS      Addr = 0;
    static short unit = 0;
    static DMA_DESCRIPTION Dma = {
        0,              /* Version */
        TRUE,           /* Master */
        FALSE,          /* ScatterGather */
        FALSE,          /* DemandMode */
        FALSE,          /* AutoInitialize */
        FALSE,          /* Dma32BitAddresses */
        0,              /* BusNumber */
        0,              /* DmaChannel */
        Isa,            /* BusType */
        Width8Bits,     /* DmaWidth */
        Compatible,     /* DmaSpeed */
        0,              /* MaximumLength */
        0               /* DmaPort */
    };


    /*
     * Allocate a descriptor
     */
    sc = (PLANCE) Malloc(sizeof(*sc));
    if(sc == NULL){
        DBGME((_T("Could not allocate %d bytes\n"), sizeof *sc));
        return NULL;
    }

    /*
     * Fill in the descriptor
     */
    rdp = (IO_PORT) (*pPortBase + 0x10);   /* Amd79c960 pp46 */
    rap = (IO_PORT) (*pPortBase + 0x12);   /* Amd79c960 pp46 */
    sc->rap = rap;
    sc->rdp = rdp;

    /*
     * Cache this pointer here, because at interrupt time
     * the Pcb might be NULL.
     */
    sc->Signal = (SIGNAL) GetRtlFunctionAddress(Condition_InterruptSignal);

    /*
     * Reset the interface
     * NOTE: From now on csr0 is the selected one
     */
    LanceWriteRegister(rap, CSR0_SELECT, CSR0_SELECT, "RAP");
    LanceWriteRegister(rdp, LN_CSR0_STOP, LN_CSR0_STOP, "csr0");

    /*
     * Read the address from the prom and save it. 
     */
    enaddr = sc->is_addr;
    for (i = 0; i < 6; i++)
        enaddr[i] = (UINT8)LanceReadRegister((IO_PORT)(*pPortBase + 0 + i));

    DBGME((_T("Lance address %x-%x-%x-%x-%x-%x\n"),
         enaddr[0], enaddr[1], enaddr[2], 
         enaddr[3], enaddr[4], enaddr[5]));

    /*
     * Create our DMA channel
     * For Busmaster cards (PCI) we get 0xff
     */
    sc->DoStartDma = TRUE;
    Dma.DmaChannel = *pDmaCh;
    if (Dma.DmaChannel >= 0xff) {
        Dma.DmaWidth = Width32Bits;
        Dma.Dma32BitAddresses = TRUE;
        Dma.BusType = PCIBus;
        sc->DoStartDma = FALSE;
    }
    /* 16 bit dma ?
     */
    else if (Dma.DmaChannel < 4) {
        Dma.DmaWidth = Width8Bits;
    } else {
        Dma.DmaWidth = Width16Bits;
    }
    sc->pDma = CreateDmaChannel( &Dma, NULL );
    if (sc->pDma == DMA_INVALID_CHANNEL) {
        DBGME((_T("Could not get Dma channel %d\n"), Dma.DmaChannel));
        goto ErrorExit;
    }
    (void) sc->pDma->v->SeizeChannel( sc->pDma, NULL, 0, 0);

    /*
     * Allocate lance RAM buffer memory 
     */
#ifdef VM
    pIAddr = CurrentAddressSpace();

    Rc = pIAddr->v->Allocate(pIAddr, &Addr, LN_MEMORY_SIZE);
    if (FAILED(Rc)) {
        DBGME((_T("Could not allocate (%x)\n"), Rc));
        goto ErrorExit;
    }

    Rc = pIAddr->v->Commit(pIAddr, Addr, LN_MEMORY_SIZE, 
            ADDR_COMMIT_CONTIGUOUS | ADDR_COMMIT_LOCK);
    if (FAILED(Rc)) {
        DBGME((_T("Could not commit (%x)\n"), Rc));
        goto ErrorExit1;
    }

    sc->MemBaseV = Addr;

    /*
     * Get phys address for it
     */
    Rc = GetPhysicalAddress( Addr, &sc->MemBaseP );
    if (FAILED(Rc)) {
        DBGME((_T("GetPhys failed on %x (%x)\n"), Addr, Rc));
        goto ErrorExit1;
    }
#else
    pIHeap = ProcessHeap();
    Addr = (ADDRESS) pIHeap->v->Alloc(pIHeap, 0, LN_MEMORY_SIZE, 0);
    if (!Addr) {
        Rc = GetLastError();
        DBGME((_T("Could not allocate (%x)\n"), Rc));
        goto ErrorExit;
    }

    sc->MemBaseV = Addr;
    sc->MemBaseP = Addr;
#endif

#if 0
    /*
     * See what phys mem we got, we're in trouble if it's not contiguous.
     */
    {
      ADDRESS CheckV, CheckP;
      ADDR_SIZE PageSize, ResUnit;

      Rc = pIAddr->v->AllocationInformation(pIAddr, &PageSize, &ResUnit);
      if (FAILED(Rc)) {
        DBGME((_T("Could not get pagesize (%x)\n"), Rc));
        goto ErrorExit1;
      }

      CheckV = Addr;
      for(CheckV = Addr;CheckV < Addr + LN_MEMORY_SIZE;CheckV += PageSize){
          Rc = GetPhysicalAddress(CheckV, &CheckP );
          if (FAILED(Rc)) {
            DBGME((_T("GetPhys failed on %x (%x)\n"), CheckV, Rc));
            goto ErrorExit1;
          }
          DBGME((_T("virtual %x maps to physical %x\n"), CheckV, CheckP));
      }
    }
    /*
     * Verify we can touch all the memory
     */
    DBGME((_T("checking that we can touch all %d bytes\n"), LN_MEMORY_SIZE));
    (void) memset((PTR) Addr, 0, LN_MEMORY_SIZE);
#endif

    /*
     * Now give it to the lance
     */
    InitLanceSpace(sc);

    /*
     * Initialize the standard interface descriptor 
     */
    sc->is_if.v = &LanceVtable;
    sc->is_if.if_unit = unit++;
    sc->is_if.if_name = "se";
    sc->is_if.if_type = IFT_ETHER;
    sc->is_if.if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
    sc->is_if.if_baudrate = 10*1000*1000; /* 10Mbps */

    /*
     * All is well
     */
    return (PINETDRIVER)sc;

#ifdef VM
 ErrorExit1:
    (void) pIAddr->v->Deallocate(pIAddr, Addr, LN_MEMORY_SIZE);
#else
// unreachable    (void) pIHeap->v->Free(pIHeap, 0, (PTR) Addr);
#endif
 ErrorExit:
    Free(sc);
    return NULL;
}

/*
 * Start/Restart interface
 *
 * We use this internally on those errors that hang the chip.
 * The upper layers use it to get the lance going initially,
 * and for whatever other test/emergency reasons.
 *
 * After stopping the chip and effectively turning off the interface
 * we release all pending xmit packets and cause the chip to init
 * itself.  Then interrupts are turned (back) on.
 */
PRIVATE INT MCT LanceRestart( PLANCE sc)
{
    se_desc_t    *rp;
    struct ifnet *ifp;
    IO_PORT      rdp;
    UINT         i;
    struct se_init_block        init_block;

    rdp = sc->rdp;

    /*
     * Lock out transmission
     */
    ef->v->NetLock(ef);

    /*
     * stop the chip , disable interrupts
     */
    LanceWriteRegister(rdp, LN_CSR0_STOP, LN_CSR0_STOP, "csr0");

    /*
     * stop network activity 
     */
    ifp = &sc->is_if;
    if (ifp->if_flags & IFF_RUNNING) {
        ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
    }
    sc->rstrtcnt++;

    DBGME((_T("se: %d restarts\n"), sc->rstrtcnt));

    /*
     * Reprogram DMA
     * Since we are bus master, all this does is cascade mode.
     */
    i = LN_MEMORY_SIZE;
    if (sc->DoStartDma)
        sc->pDma->v->StartTransfer( sc->pDma, &sc->MemBaseP, 0, 0, &i, TRUE);

    /*
     * Station address or multicast might have been changed
     */
    LanceDescCopyin (Hmem(sc->InitBlock), (PTR)&init_block, sizeof(init_block));
    memcpy(&init_block.phys_addr_low, sc->is_addr, 6);
    LanceDescCopyout((PTR)&init_block, Hmem(sc->InitBlock), sizeof(init_block));

    /*
     * INIT the chip again.
     */
    (void) LanceLoadIBlock(sc);

    /*
     * Init the buffer descriptors and indexes for each of the rings. 
     */
    for (i = 0, rp = sc->RRing; i < NRCV; i++, rp++)
        LanceDescSet(*rp, LN_RSTATE_OWN);
    
    for (i = 0, rp = sc->TRing; i < NXMT; i++, rp++)
        LanceDescSet(*rp, 0);

    sc->xmt_complete = sc->xmt_last = sc->rcv_last = 0;

    ifp->if_flags |= (IFF_UP | IFF_RUNNING);

    /*
     * Start the Lance and enable interrupts 
     */
    IOSpaceWriteInt16(rdp, LN_CSR0_STRT | LN_CSR0_INEA);
    IOWriteBufferFlush();

    /*
     * Let xmitters proceed
     */
    ef->v->NetUnlock(ef);

    return 0;
}

PRIVATE BOOL LanceLoadIBlock( PLANCE sc)
{
    IO_PORT rdp = sc->rdp;
    IO_PORT rap = sc->rap;
    int     i;

    /*
     * Load LANCE control block. 
     */

#if BYTE_ORDER==BIG_ENDIAN
    /*
     * The lance sees data naturally as half word (16 bit) quantitites. 
     * Bit 2 (BSWP) in control register 3 (CSR3) controls byte swapping.
     * To quote the spec:
     *
     *    02    BSWP    BYTE SWAP allows the chip to 
     *    operate in systems that consider bits (15:08) of data pointers
     *    by an even address and bits (7:0) to be pointed by an 
     *    odd address.
     *
     *    When BSWP=1, the chip will swap the high and low bytes on DMA
     *    data transfers between the silo and bus memory. Only data from
     *    silo transfers is swapped; the Initialization Block data and 
     *    the Descriptor Ring entries are NOT swapped. (emphasis theirs)
     *
     * So on BIG_ENDIAN systems, the BSWP bit should be set. Note,
     * however, that all shorts in the descriptor ring and initialization
     * block need to be swapped. The BITFIELD macros in lance.h handle this.
     *
     * WARNING: The setting of the BCON bit might instead be system-specific.
     *
     */
    LanceWriteRegister(rap, CSR3_SELECT, CSR3_SELECT, "RAP");
    LanceWriteRegister(rdp, LN_CSR3_BSWP|LN_CSR3_BCON, 
                       LN_CSR3_BSWP|LN_CSR3_BCON, "csr3"); 
#endif
        
    LanceWriteRegister(rap, CSR1_SELECT, CSR1_SELECT, "RAP");
    LanceWriteRegister(rdp, AddressLowPart(Lmem(sc->InitBlock)),
                       AddressLowPart(Lmem(sc->InitBlock)), "csr1");

    LanceWriteRegister(rap, CSR2_SELECT, CSR2_SELECT, "RAP");
    LanceWriteRegister(rdp, AddressHighPart(Lmem(sc->InitBlock)),
                       AddressHighPart(Lmem(sc->InitBlock)), "csr2");

    /* ASEL might be off, we want it on.
     */
    if ((LanceReadRegister((IO_PORT)(rdp+(0x16-0x10))) & 0x2) == 0) {
        LanceWriteRegister((IO_PORT)(rdp+(0x16-0x10)), 0x2, 0x2, "BCR2");
    }

    LanceWriteRegister(rap, 4, 4, "RAP");
    IOSpaceWriteInt16(rdp, 0x133f); /* we care about DPOLL=1 */

    /*
     * Start the INIT sequence now
     */
    LanceWriteRegister(rap, CSR0_SELECT, CSR0_SELECT, "RAP");

    IOSpaceWriteInt16(rdp, LN_CSR0_IDON | LN_CSR0_INIT);
    IOWriteBufferFlush();

    /*
     * Give it plenty of time to settle
     */
    for (i = 0; i < 10000; i++) {
        Delay(100);
        if ((LanceReadRegister(rdp) & LN_CSR0_IDON) != 0)
            break;
    }

    /*
     * make sure got out okay
     */
    if ((LanceReadRegister(rdp) & LN_CSR0_IDON) == 0) {
        /* LanceWriteRegister(rdp, LN_CSR0_STOP, LN_CSR0_STOP, "csr0"); */
        printf(_T("se: cannot initialize,"));
        if (LanceReadRegister(rdp) & LN_CSR0_ERR)
            printf(_T("initialization error, csr = %04x\n"),
                   (LanceReadRegister(rdp) & 0xffff));
#if _LANCE_DEBUG
        LanceDump(sc);
#endif
        return TRUE;
    }
    /*
     * Do not enable interrupts just yet. 
     */
    return FALSE;
}
 
PRIVATE void
InitLanceSpace( PLANCE sc)
{
    int         lptr;
    se_desc_t   ringaddr;
    int         i;
    struct se_init_block        init_block;

    /*
     * Reset memory allocator
     */
    sc->TopAlloc = 0;
    memset((PTR)sc->MemBaseV, 0, LN_MEMORY_SIZE);
 /* printf(_T("Lance memory at %x   %x\n"), sc->MemBaseV, sc->MemBaseP); */

    /*
     * Allocate local RAM buffer memory for the init block,
     * fill in our local copy then copyout.
     */

    sc->InitBlock = se_malloc(sc, sizeof (struct se_init_block));

    /*
     * Set values on stack, then copyout en-masse
     */
    memset(&init_block, 0, sizeof(init_block));

    memcpy(&init_block.phys_addr_low, sc->is_addr, 6);

    /*
     * Allocate both descriptor rings at once.
     * Note that the quadword alignment requirement is
     * guaranteed by the allocator.
     */
    lptr = se_malloc(sc, sizeof (struct se_desc) * (NXMT + NRCV));

    /*
     * Initialize the buffer descriptors
     */
    init_block.recv_ring_pointer_lo = AddressLowPart(Lmem(lptr));
    init_block.recv_ring_pointer_hi = AddressHighPart(Lmem(lptr));
    init_block.recv_ring_len = RLOG;

    for ( i = 0; i < NRCV ; i++, lptr += sizeof(struct se_desc)) {
        ringaddr = (se_desc_t)Hmem(lptr);
        sc->RRing[i] = ringaddr;
        sc->RBuf[i] = LanceDescAlloc (sc, ringaddr);
    }

    init_block.xmit_ring_pointer_lo = AddressLowPart(Lmem(lptr));
    init_block.xmit_ring_pointer_hi = AddressHighPart(Lmem(lptr));
    init_block.xmit_ring_len = TLOG;

    for ( i = 0 ; i < NXMT ; i++, lptr += sizeof(struct se_desc)) {
        ringaddr = (se_desc_t)Hmem(lptr);
        sc->TRing[i] = ringaddr;
        sc->TBuf[i]  = LanceDescAlloc (sc, ringaddr);
    }

    /*
     * No logical address filtering
     */
#if 0
    init_block.logical_addr_filter0 = 0;
    init_block.logical_addr_filter1 = 0;
    init_block.logical_addr_filter2 = 0;
    init_block.logical_addr_filter3 = 0;
#endif

    /*
     * Move init block into lance space
     */
    LanceDescCopyout((PTR)&init_block, Hmem(sc->InitBlock), sizeof(init_block));
}

/*
 * Dynamic control over Lance behaviour
 */
PRIVATE SCODE MCT
LanceIoctl( PLANCE sc, UINT Prop, PTR Data)
{
#if _LANCE_DEBUG
    if (Prop == SIOCGIFSTATS) {
        UINT16 cur_csr = LanceReadRegister(sc->rdp);

        printf(_T("Lance: csr = %x  (%d %d %d)\n"),
               cur_csr, sc->rcv_last, sc->xmt_last, sc->xmt_complete);

    }
#endif
    /* We care for requests to turn the interface down/up.
     * When the system goes down we must be stopped, or
     * else we get the firmware in trouble.
     */
    if (Prop == SIOCSIFFLAGS) {

        IO_PORT          rdp = sc->rdp;

        if ((sc->is_if.if_flags & IFF_UP) == 0)
            /*
             * Down. Take the easy way out: just disable interrupts.
             * If/when we get back up we'll just handle the errors.
             */
            LanceDisableInterrupts(rdp);
        else
            /*
             * Up. Strictly speaking, needs protection.
             */
            LanceEnableInterrupts(rdp);
    }
    return NO_ERROR;
}

/*
 * Interrupt service routine.
 * Return TRUE if we do have an interrupt pending,
 * after telling the chip to keep quiet.
 */
PRIVATE BOOL MCT LanceIsr( PLANCE sc, PBOOL pNotMyInterrupt)
{
    IO_PORT rdp;

    /* 
     * Catch spurious interrupts
     */
    rdp = sc->rdp;
    if ((LanceReadRegister(rdp) & LN_CSR0_INTR) == 0) {
#if _LANCE_DEBUG
        sc->spurious++;
#endif
        if (pNotMyInterrupt)
            *pNotMyInterrupt = TRUE;
        return FALSE;
    }

    /* No more interrupts, until we say so.
     */
    LanceDisableInterrupts(rdp);

    /* Wakeup handler thread and we are done.
     */
    return (*sc->Signal)( &sc->is_if.IntrPoint );
}

/*
 * Thread-level interrupt handling
 */
PRIVATE BOOL MCT LanceInterrupt( PLANCE sc)
{
    IO_PORT      rdp = sc->rdp;
    UINT16       csr;

again:
    /* Pick up interrupt pending bits, and ack them.
     */
    csr = LanceReadRegister(rdp);
    IOSpaceWriteInt16(rdp, (UINT16) (csr & LN_CSR0_WTC) );
    IOWriteBufferFlush();

#if _LANCE_DEBUG
    sc->sawcsr = csr; /*debug*/
#endif

    /*
     * Check for errors first
     */
    if ( csr & LN_CSR0_ERR ) {
        /*
         * Missed packets (because no descs available)
         */
        if (csr & LN_CSR0_MISS) {
            /*
             * Stop the chip to prevent a corrupt packet from
             * being transmitted.  There is a known problem with
             * missed packet errors causing corrupted data to
             * be transmitted to the same host as was just
             * transmitted, with a valid crc appended to the
             * packet.  The only solution is to stop the chip,
             * which will clear the Lance silo, thus preventing
             * the corrupt data from being sent.
             */
            LanceWriteRegister(rdp, LN_CSR0_STOP, LN_CSR0_STOP, "csr0");

            sc->misscnt++;
#if DEBUG
            {
                int me = 0, lance = 0, index;
                struct se_desc r;
                for (index = 0; index < NRCV; index++) {
                    LanceDescCopyin((PTR)sc->RRing[index], (PTR)&r, sizeof(r));
                    if (r.status & LN_RSTATE_OWN)
                        lance++;
                    else
                        me++;
                }
                printf(_T("se: missed packet (%d) csr = %x, Lance %x, me %x\n"),
                       sc->misscnt, csr, lance, me);
            }
#endif
            goto ErrorExit;
        }

        /*
         * Handle xmit timeouts
         */
        if (csr & LN_CSR0_BABL) {
            sc->bablcnt++;
            DBGME((_T("se: xmt timeout (%d)\n"), sc->bablcnt));
        }

        /*
         * Handle DMA errors
         */
        if (csr & LN_CSR0_MERR) {
            sc->merrcnt++;
            DBGME((_T("se: memory error (%d)\n"), sc->merrcnt));
            
        }

        /*
         * Handle transceiver failures
         */
        if (csr & LN_CSR0_CERR) {
            sc->sqercnt++;
            DBGME((_T("se: transceiver error (%d)"), sc->sqercnt));
        }

        /*
         * If any of the errors that we dont count as fatal has
         * infact caused trouble, reset lance.
         */
        if (((csr & LN_CSR0_RXON) == 0)
            || ((csr & LN_CSR0_TXON) == 0)) {
            goto ErrorExit;
        }
    }

    if ( csr & LN_CSR0_RINT )
        LanceRecvInterrupt( sc );

    if ( csr & LN_CSR0_TINT )
        LanceXmitInterrupt( sc );


    /* See if something else showed up
     */
    csr = LanceReadRegister(rdp);
    if (csr & LN_CSR0_INTR)
        goto again;

    /* Turn interrupts on now
     */
    LanceEnableInterrupts(rdp);
    IOWriteBufferFlush();

    return TRUE;

 ErrorExit:
    LanceRestart(sc);
    return FALSE;
}
 
/*
 * Handle a transmitter complete interrupt.
 */
PRIVATE void LanceXmitInterrupt( PLANCE sc)
{
    int index, status, did = 0;
    struct se_desc  r;

    /*
     * Free up descriptors for all packets in queue for which
     * transmission is complete.  Start from queue tail, stop at first
     * descriptor we do not OWN, or which is in an inconsistent state
     * (lance still working). 
     */

    while (sc->xmt_complete != sc->xmt_last) {

        index = sc->xmt_complete & (NXMT - 1);
        LanceDescCopyin ((PTR)sc->TRing[index], (PTR)&r, sizeof(r));
        status = r.status;

        /*
         * Does lance still own it ? 
         */
        if (status & LN_TSTATE_OWN)
            break;

        sc->xmt_complete++; /* xxx Atomic... */
        did++;

        sc->is_if.if_opackets++;
        if (status & (LN_TSTATE_DEF|LN_TSTATE_ONE|LN_TSTATE_MORE))
            sc->is_if.if_collisions++;

        /*
         * Check for transmission errors. 
         */
        if (status & LN_TSTATE_ERR) {

            sc->is_if.if_oerrors++;
            DBGME((_T("se: xmt error (x%x)\n"), r.status2));

            if (r.status2 & (LN_TSTATE2_RTRY|LN_TSTATE2_LCOL))
                sc->is_if.if_collisions++;

            /*
             * Restart chip on errors that disable the
             * transmitter. 
             */
            if (r.status2 & LN_TSTATE2_DISABLE) {
                LanceRestart(sc);
                return;
            }
        }
    }
    /*
     * Dequeue next transmit request, if any. 
     */
    if (did)
        LanceXmit(sc);
}
 
/*
 * Handle a receiver complete interrupt.
 */
PRIVATE void LanceRecvInterrupt(PLANCE sc)
{
    int index, first, len;
    unsigned char   status, status1;
    int             ring_cnt;
    struct se_desc  r;
    struct ether_header eh;
    struct mbuf *m;

    ef->v->NetLock(ef);

    /*
     * Starting from where we left off, look around the receive ring and
     * pass on all complete packets. 
     */

    for (;;) {

        m = NULL;

        /*
         * Read in current descriptor 
         */
        LanceDescCopyin ((PTR)sc->RRing[sc->rcv_last], (PTR)&r, sizeof(r));
        status = r.status;
        if (status & LN_RSTATE_OWN)
            break;
        first = index = sc->rcv_last;

        /*
         * If not the start of a packet, error 
         */
        if (!(status & LN_RSTATE_STP)) {
            DBGME((_T("se: Rring #%d, status=%x !STP\n"), index, status));
            goto ReleaseIt;
        }

        /*
         * See if packet is chained (should not) by looking at
         * the last descriptor (OWN clear and ENP set).
         * Remember the status info in this last descriptor. 
         */
        ring_cnt = 1, status1 = status;
        while (((status1 & (LN_RSTATE_ERR | LN_RSTATE_OWN | LN_RSTATE_ENP)) == 0) &&
               (ring_cnt++ <= NRCV)) {
            struct se_desc  r1;
            index = (index + 1) & (NRCV - 1);
            LanceDescCopyin((PTR)sc->RRing[index], (PTR)&r1, sizeof(r1));
            status1 = r1.status;
        }

        /*
         * Chained packet (--> illegally sized!); re-init the
         * descriptors involved and ignore this bogus packet.  I
         * donno how, but it really happens that we get these
         * monsters. 
         */
        if (ring_cnt > 1) {
            /*
             * Return all descriptors to lance 
             */
            LanceDescSet(sc->RRing[first], LN_RSTATE_OWN);
            while (first != index) {
                first = (first + 1) & (NRCV - 1);
                LanceDescSet(sc->RRing[first], LN_RSTATE_OWN);
            }
            if (status1 & LN_RSTATE_ERR)
                DBGME((_T("se: rcv error %x (chained)\n"), status1));
            goto NextOne;
        }

        /*
         * Good packets must be owned by us and have the end of
         * packet flag.  And nothing else. 
         */
        if ((status & ~LN_RSTATE_STP) == LN_RSTATE_ENP) {

            if ((len = r.message_size) == 0)
                /* race seen on pmaxen: the lance
                 * has not updated the size yet ??
                 */
                continue;

            /*
             * Drop header/trailing bytes from len and copy packet
             */
            m = LanceRecvPkt(sc, sc->RBuf[first], len-4-sizeof(struct ether_header), &eh);
            /*
             * Return descriptor to lance, and move on to next
             * packet 
             */
        ReleaseIt:
            r.status = LN_RSTATE_OWN;
            LanceDescCopyout((PTR)&r, (PTR)sc->RRing[first], sizeof(r));

        NextOne:
            sc->rcv_last = ++index & (NRCV - 1);

            if (m) {
                /*
                 * Hand the packet to the network module.
                 */
                sc->is_if.if_ipackets++;
                ef->v->EtherInput(ef, &sc->is_if, &eh, m);
            }
            continue;
        }
        /*
         * Not a good packet, see what is wrong 
         */
        if (status & LN_RSTATE_ERR) {
            sc->is_if.if_ierrors++;

            DBGME((_T("se: rcv error (x%x)\n"), status));

            /*
             * Return descriptor to lance 
             */
            goto ReleaseIt;
        }
        /*
         * Race condition viz lance, Wait for the next
         * interrupt. 
         */
        break;
    }
    ef->v->NetUnlock(ef);
}

/*
 * Start output on interface.
 *
 */
PRIVATE void MCT
LanceXmit( PLANCE sc)
{
    struct mbuf     *request;
    struct se_desc  r;
    int             tlen, nqueued;
    int             index;
    PIENDPOINTFACTORY ef = sc->is_if->EndpFact;
    PIMBUFFACTORY mbf = sc->is_if->MBufFact;

    ef->v->NetLock(ef);

    nqueued = 0;
    while ((sc->xmt_last - sc->xmt_complete) < NXMT)
    {
        /*
         * Dequeue the next transmit request, if any. 
         */
        IF_DEQUEUE(&sc->is_if.if_snd, request);
        if (request == 0)
            break;      /* Nothing (else) on the queue   */

        index = (sc->xmt_last++) & (NXMT - 1);

        tlen = LanceSendPkt(request, sc->TBuf[index]);

        /*
         * Give away buffer.  Must copyin/out, set len,
         * and set the OWN flag.  We do not do chaining.
         */
        LanceDescCopyin((PTR)sc->TRing[index], (PTR)&r, sizeof(r));

        r.buffer_size = (UINT16)(-(tlen) | 0xf000);
        r.status = (LN_TSTATE_OWN | LN_TSTATE_STP | LN_TSTATE_ENP);

        LanceDescCopyout((PTR)&r, (PTR)sc->TRing[index], sizeof(r));
        IOWriteBufferFlush();

        if (nqueued++ == 0) {
            /*
             * Tell the chip to rescan the descriptors _now_.
             * While sending full-blast biggies we might gain up to
             * 2 msec latency kicking the lance here instead of
             * doing it at the end of this loop.
             */
            IOSpaceWriteInt16(sc->rdp, LN_CSR0_TDMD | LN_CSR0_INEA);
        }

        mbf->v->MBFree(mbf, request);

    }
#if 0
    if (nqueued) {
        sc->sc_if.if_flags |= IFF_OACTIVE;
    }
#endif
    ef->v->NetUnlock(ef);
}


/*
 * Pull a packet off the interface and
 * hand it up to the higher levels.
 */
PRIVATE struct mbuf *
LanceRecvPkt(PLANCE sc, volatile UINT8 *RBuf, int len, struct ether_header *eh)
{
    struct ifnet *ifp = &sc->is_if;
    struct mbuf  *mh, *m = NULL;
    UINT16 ether_type;
    PIMBUFFACTORY mbf = ifp->MBufFact;
    SCODE sc;

    if (len <= 0)
        return NULL;    /* sanity */

    sc = mbf->v->MBGetHdr(mbf, M_DONTWAIT, MT_DATA, &mh);
    if (FAILED(sc)) {
        /*
         * No room, drop the packet
         */
        ifp->if_iqdrops++;
        return NULL;
    }
    mh->m_pkthdr.rcvif = ifp;
    mh->m_pkthdr.len = len;
    mh->m_len = MHLEN;

    /*
     * Read in ethernet header 
     */
    LanceDataCopyin ((PTR)RBuf, (PTR)eh, sizeof(struct ether_header));
    RBuf += LanceDataOffset(sizeof(struct ether_header));

    ether_type = ntohs(eh->ether_type);
    if (ether_type >= ETHERTYPE_TRAIL &&
        ether_type < ETHERTYPE_TRAIL+ETHERTYPE_NTRAILER)
#if 0
    {
        /*
         * Deal with trailers..
         */
        offset = (ether_type - ETHERTYPE_TRAIL) << 9;
        copyin 2 bytes into nths(ether_type);
        copyin 2 bytes into nths(ll);
        len = offset + ll;
    } else
        offset = 0;
#else
        printf(_T("%x TRAILERS !!\n"), ether_type);
#endif
    eh->ether_type = ether_type;

    /* Small packet ?
     */
    if (len <= MHLEN) {
        LanceDataCopyin((PTR)RBuf, mtod(mh, PTR), len);
        mh->m_len = len;
        return mh;
    }

    /* Uhmm, needs two mbufs
     */
    m = m_get(M_DONTWAIT, MT_DATA);
    if (m == NULL)
        goto ErrorExit;

    /* A biggy ?
     */
    if (len >= MINCLSIZE) {
        sc = mbf->v->MBClGet(mbf, m, M_DONTWAIT);
        if ((m->m_flags & M_EXT) == 0)
            goto ErrorExit;
    }

    mh->m_next = m;
    m->m_len = len - MHLEN;

    LanceDataCopyin((PTR)RBuf, mtod(mh, PTR), MHLEN);
    RBuf += LanceDataOffset(MHLEN);

    LanceDataCopyin((PTR)RBuf, mtod(m, PTR), m->m_len);

    return mh;

 ErrorExit:
    ifp->if_ierrors++;
    mbf->v->MBFree(mbf, mh);
    if (m)
        mbf->v->MBFree(mbf, m);
    return NULL;
}


/*
 * Move a packet into Lance space
 */
PRIVATE int
LanceSendPkt(
        struct mbuf  *m,
        volatile UINT8 *SBuf)
{
    UINT16 *dp;
    int    len = 0;

    do {
        dp = mtod(m, UINT16 *);

        LanceDataCopyout((PTR)dp, (PTR)(SBuf + len), m->m_len);

        len += m->m_len;
        m = m->m_next;

    } while (m);

#if 0
{ int *p = (int *)SBuf;int i = len;
printf(_T("T:"));
 while (i > 0) { printf(" %x", *p++); i -= 4;}
printf(_T("\n"));
}
#endif

    if (len < LN_MINBUF_NOCH)
        /*
         * The lance needs at least this much data in a packet. Who
         * cares if I send some garbage that was left in the lance
         * buffer ?  If one can spoof packets then one can spoof
         * packets!
         */
        len = LN_MINBUF_NOCH;
    return len;
}

/*
 * Reset a descriptor's flags.
 * Optionally give the descriptor to the lance
 */
PRIVATE void
LanceDescSet (se_desc_t lndesc, int val)
{
    struct se_desc              desc;

    LanceDescCopyin((PTR)lndesc, (PTR)&desc, sizeof(desc));
    desc.desc4.bits = 0;
    desc.status     = (UINT8) val;
    LanceDescCopyout ((PTR)&desc, (PTR)lndesc, sizeof(desc));
    IOWriteBufferFlush();
}

/*
 * Allocate and initialize a ring descriptor.
 * Allocates a buffer from the lance memory and writes a descriptor
 * for that buffer to the host virtual address LNDESC.
 */
PRIVATE volatile PTR
LanceDescAlloc (PLANCE sc, se_desc_t lndesc)
{
    UINT                dp;
    struct se_desc      desc;

    /*
     * Allocate buffer in lance space 
     */
    dp = se_malloc(sc, LN_BUFFER_SIZE);

    /*
     * Build a descriptor pointing to it 
     */
    desc.addr_low = AddressLowPart(Lmem(dp));
    desc.addr_hi  = AddressHighPart(Lmem(dp));
    desc.status   = 0;
    desc.buffer_size = (UINT16)(0-LN_BUFFER_SIZE);
    desc.desc4.bits  = 0;

    /*
     * Copy the descriptor to lance space 
     */
    LanceDescCopyout ((PTR)&desc, (PTR)lndesc, sizeof(desc));
    IOWriteBufferFlush();

    return (volatile PTR) Hmem(dp);
}

/*
 * Allocate a chunk of lance RAM buffer. Since we never
 * give lance RAM buffer memory back, we'll just step up the
 * byte-count on a per-unit basis.
 *
 * The return value is an index into the lance memory, which can be
 * passed with Hmem() and Lmem() to get valid host and chip addresses.
 */
PRIVATE UINT
se_malloc(PLANCE sc, int size)
{
    UINT    ret;

    /*
     * Round to double longword boundary
     */
    if (sc->TopAlloc & 0xf)
        sc->TopAlloc = (sc->TopAlloc + 0x10) & ~0xf;

    ret = sc->TopAlloc;
    sc->TopAlloc += size;

#if _LANCE_DEBUG
    if (sc->TopAlloc > LN_MEMORY_SIZE) {
        printf(_T("se: se_malloc panic %x %x"), sc->TopAlloc, LN_MEMORY_SIZE);
        DebugBreak();
    }
#endif

    return ret;
}
