/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 *    Module: hostnic.c
 *
 *    Description: Routines for serial line nework object,
 *                 routine for dispatching thread (recv,filter,send),
 *                 and main entry point for this program.
 */
#include "hostnic.h"

/*
 *    Wrap an ethernet packet with length and checksum, serial format.
 */
void
SerialWrap(
           PPACKET Packet
          )
{
    UINT16     Sum;
    UINT8     *pData8;
    UINT32     i;

    /*
     *    Bit of an ugly hack: serial line expects packets to be
     *    encapsulated, length in front and xsum at the rear.
     */    

    Packet->Data -= 2;
    Packet->Data[1] = (UINT8)Packet->nBytesAvail;
    Packet->Data[0] = (UINT8)(Packet->nBytesAvail>>8);
    Packet->nBytesAvail += 2;

    /*
     *    Compute & add the xsum
     */

    pData8 = Packet->Data;
    Sum = 0;
    for (i = 0; i < Packet->nBytesAvail; i++)
        Sum += *pData8++;

    Packet->Data[Packet->nBytesAvail++] = (UINT8)(Sum >> 8);
    Packet->Data[Packet->nBytesAvail++] = (UINT8) Sum;
}

/*
 *    Filter function for serial line input packets.
 *    Must reassemble streaming data into full packets.
 */
PACKET_ACTION
SerialFilter(
            PNETWORK Input,
            PNETWORK Output,
            PPACKET Packet
)
{
    PPACKET Parking, Packet2;
    UINT32  Length;
    INT32   Missing;
    BOOL    Complete;

    /*
     *    Often will complete with a timeout, e.g. zero length
     */
    if (Packet->nBytesAvail == 0)
        return PacketActionDrop;

    /*
     *    Say what we got if we want to know
     */
    if (debug_level)
    {
        printf("s%d ",Packet->nBytesAvail);
        PacketDump(Packet,"SerialInput: ");
    }

    /*
     *    Do we have a half-packet, which this completes ?
     */
    Parking = (PPACKET) Input->FilterState;
    if (Parking != NULL)
    {
        /*
         *    Yes, lets see if we can complete it now
         *
         *    Packets from the serial cable are preceeded by their length,
         *    which we stick in Parking->Length once we have it.
         *    NB: PacketInit() resets the length to what it really is
         */
        if (Parking->nBytesAvail == 1)
        {
            /* 
             *    We know the length now.
             */
            Parking->Length = Packet->Data[0] | (Parking->Data[0] << 8);
            Parking->nBytesAvail = 0;
            Packet->Data += 1;
            Packet->nBytesAvail -= 1;

            if (debug_level)
                printf("L%d ",Parking->Length);

            /* 
             *    But is the length sane.
             */
            if ((Parking->Length == 0) ||
                (Parking->Length > MAX_ETHERNET_PACKET_SIZE))
            {
                printf("Outtasynch!!\n");

                PacketInit(Parking);
                PacketReadFile(Input->hFile,
                               Parking);

                Input->FilterState = NULL;
                return PacketActionDrop;
            }

        }

        /* 
         *    See how much data we are missing.
         * 
         */
        Missing = Parking->Length - Parking->nBytesAvail;

        /*
         *    NB: Another bitchy case: what if we are missing just
         *    the last byte(s) e.g. say the second part of the xsum.
         */
        if (Missing <= 0)
        {
            assert((Missing == 0) || (Missing == -1));
            Length = (2 + Missing);
            Complete = Packet->nBytesAvail >= Length;
            /* 
             *    no need to copy, really, wont send the xsum out.
             */
            if (!Complete)
            {
                assert(Packet->nBytesAvail == 1);
                Packet->nBytesAvail = 0;
            }
            Length = 0;
        } 
        else
        {
            Length = Missing;
            Complete = (Packet->nBytesAvail - 2) >= Length;
        }
            

        /* 
         */
        if (debug_level)
            printf("c%d ",Length);

        /* 
         *    Still not enough ?
         */
        if (!Complete)
        {
            /* 
             *    Copy this one more chunk and we are done for now
             */
            memcpy(Parking->Data + Parking->nBytesAvail,
                   Packet->Data,
                   Packet->nBytesAvail);
            Parking->nBytesAvail += Packet->nBytesAvail;

            if (debug_level)
                printf("p%d ",Parking->nBytesAvail);
            return PacketActionDrop;
        }

        /* 
         *    We got enough to complete one, and possibly more.
         */
        memcpy(Parking->Data + Parking->nBytesAvail,
               Packet->Data,
               Length);
        Parking->nBytesAvail += Length;

        /* 
         *    Post the parked, now-complete packet
         */
        if (debug_level)
            printf("ActuallyOut1: %d\n",Parking->nBytesAvail);
        PacketWriteFile(Output->hFile, Parking);
        Input->FilterState = NULL;

        /* 
         *    See if there is anyhting more
         */
        Packet->Data += Length + 2;
        Packet->nBytesAvail -= Length + 2;
        if (Packet->nBytesAvail == 0)
            return PacketActionDrop;

        /* 
         *    Fall through to the case of nothing pending
         */
        if (debug_level)
            printf("x%d ",Packet->nBytesAvail);
    }

    /* 
     *    We might receive multiple small packets at once
     */
 Again:
    assert(Packet->nBytesAvail > 0);

    /* 
     *    Starting by checking if we have at least the length.
     */
    if (Packet->nBytesAvail == 1)
    {
        /* Nope.  Keep it around for later (see code above).
         */
        Input->FilterState = Packet;
        if (debug_level)
            printf("y1 ");
        return PacketActionNone;
    }

    /* 
     *    Compute & check the length
     */
    Length = Packet->Data[1] | (Packet->Data[0] << 8);
    Packet->Length = Length;
    Packet->Data += 2;
    Packet->nBytesAvail -= 2;

    if (debug_level)
        printf("l%d ",Length);

    if ((Length == 0) ||
        (Length > MAX_ETHERNET_PACKET_SIZE))
    {
        printf("Outtasynch2!!\n");
        SerialEtherinfo(Output, Input);
        return PacketActionDrop;
    }

    /* 
     *    Do we have exactly the right amount of data ?
     */
    if ((Length + 2) == Packet->nBytesAvail)
    {
        /* 
         *    Yes, easy then.
         */
        Packet->nBytesAvail -= 2;
        if (debug_level)
        {
            printf("ActuallyOut: %d\n",Packet->nBytesAvail);
            PacketDump(Packet,"EtherOutput: ");
        }
        return PacketActionForward;
    }

    /* 
     *    If we dont have enough data..
     */
    if ((Length + 2) > Packet->nBytesAvail)
    {
        /*
         *    ..then put the packet in parking, for later.
         */
        Input->FilterState = Packet;
        if (debug_level)
            printf("f%d\n",Packet->nBytesAvail);
        return PacketActionNone;
    }

    /* 
     *    Mumble.. more than one packet in there.
     *    Also, must forward them in order.  
     *    Bummer.. lets do it with copies.. sigh.
     */
    PacketAllocate(&Packet2);
    PacketInit(Packet2);

    memcpy(Packet2->Data,
           Packet->Data,
           Length);
    Packet2->nBytesAvail = Length;

    /* 
     *    Post it, but remember this is an extra, over-budget one.
     */
    if (debug_level)
        printf("ActuallyOut2: %d\n",Packet2->nBytesAvail);
    PacketWriteFile(Output->hFile, Packet2);
    Packet2->Mode = PacketModeTransmittingExtra;

    /* 
     *    Skip over this data and restart from the beginning
     */
    Packet->Data += Length + 2;
    Packet->nBytesAvail -= Length + 2;
    goto Again;
}

/*
 *    Initialization packet for serial line driver
 */
BOOL
SerialEtherinfo(
                PNETWORK Ethernet,
                PNETWORK Serial
)
{
    PPACKET Packet;

    /* 
     *    Allocate & zero a packet
     */
  
    if (!PacketAllocate(&Packet))
    {
        printf("Outtamem66 %x\n", GetLastError());
        return FALSE;
    }

    PacketInit(Packet);

    /*
     *     Craft that special init packet
     */
    memcpy(Packet->Ethernet.DstAddr,Ethernet->EthernetAddress,6);
    memcpy(Packet->Ethernet.SrcAddr,Ethernet->EthernetAddress,6);
    Packet->Ethernet.Etype = 0x6666;
    memcpy(&Packet->Buffer[sizeof(ETHERNET_HEADER)],
           Ethernet->EthernetAddress,6);

    Packet->nBytesAvail = sizeof(ETHERNET_HEADER) + 6;

    /*
     *     Wrap length & xsum around it.
     */

    SerialWrap(Packet);

    /*
     *    Post the packet for xmit
     */

    PacketWriteFile(Serial->hFile, Packet);
    
    /*
     *    Done, packet should be recycled.
     */

    Packet->Mode = PacketModeTransmittingExtra;
    return TRUE;
}

/*
 *    Constructor for serial line network object.
 */
PNETWORK 
SerialCreate(
           char *Line,
           UINT BaudRate
           )
{
    UINT InterByteTime;
    DCB ControlBlock;
    HANDLE hLine;
    PNETWORK Network;
    COMMTIMEOUTS Timeouts;

    /* 
     *    Open & check the comm line
     */
    hLine = CreateFile( Line, GENERIC_READ | GENERIC_WRITE,
                        0, /* exclusive access */
                        NULL, /* no security attributes */
                        OPEN_EXISTING,
                        FILE_FLAG_OVERLAPPED,
                        NULL  /* hTemplate - must be NULL for comms */
                        );

    if (hLine == INVALID_HANDLE_VALUE) {
        printf("Could not open %s\n",Line);
        return NULL;
    }

    /*
     *    Allocate and initialize descriptor
     */
    Network = malloc(sizeof(NETWORK));
    if (Network == NULL) {
        printf("Outtamem?!? x%x\n", GetLastError);
        CloseHandle(hLine);
        return NULL;
    }

    memset(Network, 0, sizeof(NETWORK));

    Network->Filter = &SerialFilter;

    /* 
     *    Get what we wont change
     */
    ControlBlock.DCBlength = sizeof(ControlBlock);
    if (!GetCommState(hLine,&ControlBlock)) {
        printf("GetCommState() failed on %s (x%x)\n",Line,GetLastError());
        /* 
         *    might be a named pipe, ignore
         */
    }

    /* 
     *    No parity, no flow..
     */
    ControlBlock.BaudRate = BaudRate;
    ControlBlock.fBinary = TRUE;     /* no EOF check */
    ControlBlock.fParity = FALSE;
    ControlBlock.fOutxCtsFlow = FALSE;
    ControlBlock.fOutxDsrFlow = FALSE;
    ControlBlock.fDtrControl = DTR_CONTROL_ENABLE;
    ControlBlock.fDsrSensitivity = FALSE;
    ControlBlock.fTXContinueOnXoff = TRUE;
    ControlBlock.fOutX = FALSE;
    ControlBlock.fInX = FALSE;
    ControlBlock.fErrorChar = FALSE;
    ControlBlock.fNull = FALSE;
    ControlBlock.fRtsControl = RTS_CONTROL_ENABLE;
    ControlBlock.fAbortOnError = FALSE;
    ControlBlock.ByteSize = 8;
    ControlBlock.Parity = NOPARITY;
    ControlBlock.StopBits = ONESTOPBIT;
    if (!SetCommState(hLine,&ControlBlock)) {
        printf("SetCommState() failed on %s (x%x)\n",Line,GetLastError());
        /* 
         *    might be a named pipe, ignore
         */
    }

    /* 
     *    Doing overlapped I/O on the serial line under NT is funny.
     *    By default, it will wait to fillup completely the read buffer
     *    before signalling completion.  So we have to set a timeout.
     *    Timeouts are in milliseconds units (none on xmit)
     *    Luckily, we can do the optimal thing of setting a timeout that
     *    (a) only starts after the first byte is received
     *    (b) renews itself for as long as bytes keep coming in
     *    (c) expires after that much time elapses between bytes
     */

    InterByteTime = BaudRate / 10; /* approx byte rate */
    InterByteTime = 1000 / InterByteTime; /* in msecs, most likely 0 */
    InterByteTime++;  /* roundup, avoid zero */
    if (debug_level)
        printf("SerialRead inter-byte timeout is %d msecs\n",InterByteTime);
    Timeouts.ReadIntervalTimeout = InterByteTime;
    Timeouts.ReadTotalTimeoutConstant = 0;
    Timeouts.ReadTotalTimeoutMultiplier = 0;

    Timeouts.WriteTotalTimeoutMultiplier = 0;
    Timeouts.WriteTotalTimeoutConstant = 0;

    if (!SetCommTimeouts( hLine, &Timeouts)) {
        printf("SetCommTimeouts() failed (x%x), ignoring..\n",
               GetLastError());
    }

    Network->hFile = hLine;
    Network->BaudRate = BaudRate;
    /* 
     *    Leave address empty. 
     *    Could snoop on the line later but so what.
     */

    Network->bInitialized = TRUE;

    return Network;
}


/* 
 *    Global data
 */
char *Usage = "%s [-b<BAUDRATE>] [-verbose] [COM1|COM2|...]";
int debug_level = 0;

#define PKT_INTERVAL 500
UINT32 pkt_interval = PKT_INTERVAL;

BOOL Unfiltered = FALSE;

/* 
 *    Program's entry point
 */
int main(int argc, char **argv)
{
    char *Line, *pgmname = argv[0];
    UINT BaudRate;
    PNETWORK Ethernet = NULL, Serial = NULL;

    Line = "COM1";
    BaudRate = 38400;

    /* 
     *    Parse invocation arguments, if any
     */
    while (argc > 1) {
        if (argv[1][0] == '-') {
            switch (argv[1][1]) {
            case 'b':
            case 'B':
                BaudRate = atoi(argv[1]+2);
                printf("Baudrate will be %d baud\n", BaudRate);
                break;                
            case 'f':
            case 'F':
                Unfiltered = !Unfiltered;
                printf("Will %sfilter ether traffic.\n",
                       (Unfiltered ? "NOT " : ""));
                break;                
            case 'p':
            case 'P':
                pkt_interval = atoi(argv[1]+2);
                break;                
            case 'V': debug_level++;/*fallthrough*/
            case 'v': debug_level++;
                printf("debug_level = %d\n", debug_level);
                break;
            default:
                printf(Usage,pgmname);
                exit(-1);
            }
        } else {
            Line = argv[1];
        }
        argc--,argv++;
    }

    /* 
     *    Create the serial line first
     */

    Serial = SerialCreate(Line,BaudRate);
    if (Serial == NULL)
    {
        printf("SerialCreate FAILED(x%x)\n", GetLastError());
        goto Done;
    }

    /* 
     *    Create the ethernet packet filter next
     */

    Ethernet = EthernetCreate();
    if (Ethernet == NULL)
    {
        printf("EthernetCreate FAILED(x%x)\n", GetLastError());
        goto Done;
    }

    if (debug_level)
    printf(_T("HostNic CardAddr %02x:%02x:%02x:%02x:%02x:%02x\n"),
           Ethernet->EthernetAddress[0], Ethernet->EthernetAddress[1],
           Ethernet->EthernetAddress[2], Ethernet->EthernetAddress[3],
           Ethernet->EthernetAddress[4], Ethernet->EthernetAddress[5]);

    /*
     *    Create a single I/O completion port for send and
     *    receive operations to and from the two interfaces.
     */

    Ethernet->IoCompletionPort = CreateIoCompletionPort(
                                                       Ethernet->hFile,
                                                       NULL,
                                                       (ULONG) Ethernet, 
                                                       0
                                                       );

    assert(Ethernet->IoCompletionPort != NULL);

    Serial->IoCompletionPort = CreateIoCompletionPort(
                                                      Serial->hFile,
                                                      Ethernet->IoCompletionPort,
                                                      (ULONG) Serial,
                                                      0
                                                      );

    assert(Serial->IoCompletionPort != NULL);
    assert(Serial->IoCompletionPort == Ethernet->IoCompletionPort);


    /*
     *    Criss-cross the two interfaces
     */

    Ethernet->ForwardingTo = Serial;
    Serial->ForwardingTo = Ethernet;

    /*
     *     Post the MMLite serial driver's initialization packet
     */

    if (!SerialEtherinfo(Ethernet,Serial))
    {
        printf("Failed to post Serial Etherinfo, terminating.\n");
        goto Done;
    }

    /*
     *    Now become/create the thread that will
     *    handle the interactions with the network.
     */
#if 1
    NetworkThread( Ethernet );
#else
    Ethernet->hThread = CreateThread( 
                                    NULL,
                                    0,
                                    (LPTHREAD_START_ROUTINE) NetworkThread,
                                    Ethernet, /* for init handshake only */
                                    0,
                                    0
                                    );
    /*
     *    Give the NT thread time to get running.
     */

    while (!Ethernet->bInitialized)
        Sleep(100); /* msecs */

    
    /*
     *    Now wait until its done, we got nothing better to do
     */

    while (Ethernet->bInitialized)
        Sleep(10000); /* msecs */
#endif

Done:
    /* 
     *    Termination code.
     *    BUGBUG Should free all memory but.
     */
    if (Serial)
        CloseHandle(Serial->hFile);
    if (Ethernet)
        CloseHandle(Ethernet->hFile);

    return 0;
}


/*
 *    Function: NetworkThread()
 *
 *    Description: Send and receive thread.
 *
 */
void 
NetworkThread(
    IN void * ThreadContext
    )
{
    PNETWORK        Network;
    UINT32          Transferred;
    UINT32          BytesReceived, BytesForwarded;
    UINT32          PacketsReceived, PacketsDropped;
    PACKET *        Packet;
    SCODE           Result;
    BOOL            bResult;
    PACKET_ACTION   aResult;
    OVERLAPPED *    Overlapped;
    ULONG           Key;
    ULONG           CurrentTime, ElapsedTime;
    int             i;

    if (debug_level)
        printf("NetworkThread.\n");

    Network = (PNETWORK) ThreadContext;

    /*
     *    We have to run a little faster than normal 
     *    in order to handle a lot of network packets.
     */

    SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);

    /*
     *     Post all the receives, 50-50 on both interfaces
     */

    for (i = 0; i < MAX_RECEIVE_POOL_SIZE/2; i++)
    {
        bResult = PacketAllocate(&Packet);
        if (bResult)
        {
            assert(Packet != NULL);
            PacketInit(Packet);
            PacketReadFile(Network->hFile, Packet);
        }

        bResult = PacketAllocate(&Packet);
        if (bResult)
        {
            assert(Packet != NULL);
            PacketInit(Packet);
            PacketReadFile(Network->ForwardingTo->hFile, Packet);
        }
    }

    /*
     *     We are ready to roll
     */

    Network->bInitialized = TRUE;

    /*
     *    The loop below is the main processing loop for all
     *    Receives and transmits. 
     */

    if (debug_level)
        printf("Stats every %d packets received:\n", pkt_interval);

    ElapsedTime = GetTickCount();
    BytesForwarded = BytesReceived = 0;
    PacketsReceived = PacketsDropped = 0;
    for (i=0;;) 
    {
        /*
         *    Wait for an I/O to complete.
         */
        
        Packet = NULL;
        Overlapped = NULL;

        bResult = GetQueuedCompletionStatus(
                            Network->IoCompletionPort,
                            &Transferred,
                            &Key,
                            &Overlapped,
                            INFINITE
                            );

        /*
         *    If the completion was ok, move the packet to the 
         *    other interface for forwarding/receiving.
         */

        Packet = (PACKET *) Overlapped;
        Network = (PNETWORK) Key;

        if ( bResult != FALSE && Packet != NULL )
        {
            assert(Packet->Mode != PacketModeInvalid);

            /* 
             *    Is this a packet we just received ?
             */

            if ( Packet->Mode == PacketModeReceiving )
            {
                /*
                 *    Finish initializing pending packet. Keep stats.
                 */    
    
                Packet->nBytesAvail = Transferred;
                Packet->Result = S_OK;

                PacketsReceived++;
                BytesReceived += Transferred;

                /*
                 *    Check the filter to see if we should drop it or what
                 */

                aResult = Network->Filter(
                                          Network,
                                          Network->ForwardingTo,
                                          Packet);
                switch (aResult)
                {
                case PacketActionForward:
                    /*
                     *    Move it on to the other interface.
                     */

                    PacketWriteFile(Network->ForwardingTo->hFile,
                                    Packet);
                    break;

                case PacketActionDrop:
                    /*
                     *    Repost it, keep stats
                     */

                    PacketsDropped++;

                    PacketInit(Packet);

                    PacketReadFile(Network->hFile,
                                   Packet);

                case PacketActionNone:
                    /*
                     *    Filter already acted on it, ignore.
                     */

                    break;

                default:
                    /*
                     *    Shouldnt happen
                     */

                    assert(FALSE);

                }
            }
            /* 
             *    Or is this a packet we just sent
             */
            else
            {
                assert((Packet->Mode == PacketModeTransmitting) ||
                       (Packet->Mode == PacketModeTransmittingExtra));

                /*
                 *    Statistics
                 */

                i++;
                BytesForwarded += Transferred;

                /*
                 *    Move the packet back to receiving on the other iface,
                 *    unless its one of those extras over budget.
                 */

                if (Packet->Mode == PacketModeTransmitting)
                {
                    PacketInit(Packet);

                    PacketReadFile(Network->ForwardingTo->hFile,
                                   Packet);

                }
                else
                if (Packet->Mode == PacketModeTransmittingExtra)
                {
                    /*
                     *    Must be recycled.
                     */

                    Result = PacketFree(Packet);
                    assert(Result == S_OK);
                }
            }
        }

        /*
         *    Statistics: update every PKT_INTERVAL packets received
         */

        if ((debug_level) &&
            (PacketsReceived == pkt_interval))
        {
            /*
             *    See how long this interval was.
             */
            CurrentTime = GetTickCount();
            ElapsedTime = CurrentTime - ElapsedTime;

            /*
             *    Compute and print the stats
             */
            printf("RECV %.2f pps %.1f bps DROP %.2f pps FORWARD %.2f pps %.1f bps\n",
                   (double) (PacketsReceived * 1000) / (double) ElapsedTime,
                   (double) (BytesReceived * 8 * 1000) / (double) ElapsedTime,
                   (double) (PacketsDropped * 1000) / (double) ElapsedTime,
                   (double) (i * 1000) / (double) ElapsedTime,
                   (double) (BytesForwarded *8*1000) / (double) ElapsedTime);

            /*
             *    Get set for new round
             */
            i = 0;
            ElapsedTime = CurrentTime;
            BytesReceived = BytesForwarded = 0;
            PacketsReceived = PacketsDropped = 0;
        }
    }

    /*
     *     Should we ever terminate cleanly..
     */

    Network->bInitialized = TRUE;

}


