/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Intermediate network driver, for serial lines.
 * Should work on top of any IFile object, actually.
 */
#define USE_INTRINSICS
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <mmlite.h>
#include <sock.h>
#include <if.h>
#include <if_types.h>
#include <ip/in.h>
#include <ip/arp.h>
#include <net/dns/inet.h>

//#define _DEBUG 1
//#define DBGLVL DbgLvl
#include <mmmacros.h>
//#if _DEBUG
//INT DbgLvl = 1; /* 0 noise, 1 - comment, 2 - major function, 3 - error */
//#endif

/* Currently ETHER_MTU + overhead. Make smaller later? */
#define SERPLEX_MTU 1516

typedef struct _SERNET { /* Inherits from CDevice */

    struct arpcom ac;            /* standard ether interface header */
#define is_if ac.ac_if
#define is_addr ac.ac_enaddr
#define MissedPackets ac.ac_if.if_data.ifi_errors[IFI_ERR_MISS]
#define CompleteResets ac.ac_if.if_data.ifi_errors[IFI_ERR_RESET]
#define MemoryErrors ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM]
#define BadChecksum ac.ac_if.if_data.ifi_errors[IFI_ERR_XSUM]

    /* Driver's private data
     */
    PIFILE  Serial;
    MUTEX   Lock;

} SERNET, *PSERNET;

/* Set the Ethernet Station Address
 */
PRIVATE
void
CsSetStationAddress(PSERNET pCs, PUINT8 pSA)
{
    memcpy(&pCs->is_addr[0], pSA, 6);

    DBG((_T("SerNet: Interface name sr%d; Addr %02x:%02x:%02x:%02x:%02x:%02x\r\n"),
         pCs->is_if.if_unit,
         pCs->is_addr[0], pCs->is_addr[1],
         pCs->is_addr[2], pCs->is_addr[3],
         pCs->is_addr[4], pCs->is_addr[5]));
}

PRIVATE SCODE
WriteIt( PIFILE pOut,
         void *pData,
         UINT Len)
{
    static UINT64 Zero = Int64Initializer(0,0);
    
    if (pOut == NULL)
        return E_INVALID_PARAMETER;
    return pOut->v->WriteAt(pOut,
                            Zero,
                            pData,
                            Len,
                            &Len);
}

PRIVATE void
LongTimeout(PIFILE Link)
{
    UINT64 Param;
    UINT32 *p = (UINT32*)&Param;

    p[0] = 0;
    p[1] = TIME_SECONDS(10);
    (void) Link->v->SetSize(Link, Param);
    /* ignore errors, we tried */
}


/* Process packets that have been completely received.
 * That means no errors, and "frame accepted into on-chip RAM"
 * as per specs pp 83 Figure 21 "Receive Frame Pre-Processing" flowchart.
 */
void
SerNetRecv(PSERNET pCs)
{
    struct ifnet   *ifp = &pCs->is_if;
    INT32           Length = 0;
    UINT16          Says;
    PIBUFFER        Pkt = NULL;
    struct ether_header eh;
    SCODE sc;
    UINT len = 0;

    DBGME(0,printf(_T("Enter SerNetRecv\r\n")));

    /* Get the length, first two bytes
     * NB: does not include itself and trailing xsum bytes
     */

    sc = ReadWriteFile(pCs->Serial, FALSE, UINT64_ZERO,
                       (UINT8*)&Says, 2, &len);
    if (FAILED(sc) || (len != 2)) {
        /* Most likely a timeout, no prob
         */
        if (sc == E_TIMED_OUT)
            ifp->if_ierrors--; /* compensate for common path */
        DBGME(2,printf(_T("SerNetRecv: sc=%x len %d\r\n"),sc,Length));
        goto DropPacket;
    }

    /* Sanity check the length (out of sync)
     */
    Length = ntohs(Says);
    Length -= sizeof(struct ether_header);
    if ((Length <= 0) ||
        (Length > ETHERMTU)) {
        DBGME(2,printf(_T("SerNetRecv: badlen %d x%x\r\n"),Length,Says));
        pCs->MissedPackets++;
        goto DropPacket;
    }

    /* Get a buffer to copy to
     */
    sc = ifp->MBufFact->v->AllocateBuffer(ifp->MBufFact,
                                          Length, (PINETDRIVER) ifp,
                                          &Pkt);
    if (Pkt == 0 || FAILED(sc)) {
        /* No room, drop the packet
         */
        DBGME(3,printf("SerNetRecv: no mem sc=%x\r\n", sc));
        pCs->MemoryErrors++;
        goto DropPacket;
    }

    assert(Length > 0 && Length <= MCLBYTES);

    /* Read&fix the ether header */
    sc = ReadWriteFile(pCs->Serial, FALSE, UINT64_ZERO,
                       (UINT8*)&eh, sizeof(eh), &len);
    if (FAILED(sc) || (len != sizeof(eh))) {
        DBGME(2,printf("SerNetRecv: sc=%x small %d\r\n",sc,len));
        pCs->MissedPackets++;
        goto DropPacket;
    }

    eh.ether_type = ntohs(eh.ether_type);

    /* Get the rest of the data */
    assert(Length == ((struct mbuf *)Pkt)->m_pkthdr.len);
    sc = Pkt->v->ReadWriteFile(Pkt, pCs->Serial, FALSE, UINT64_ZERO,
                               0, Length, &len);
    if (FAILED(sc) || (Length != (INT32) len)) {
        DBGME(2,printf("SerNetRecv: sc=%x short %d != %d\r\n",sc,Length,len));
        pCs->MissedPackets++;
        goto DropPacket;
    }
    assert(len == (UINT) ((struct mbuf *)Pkt)->m_pkthdr.len);

    /* If this is the init packet then take it
     */
    if (eh.ether_type == 0x6666) {
        CsSetStationAddress(pCs,mtod((struct mbuf *)Pkt,UINT8*));
        DBGME(2,printf(_T("SerNet: init pkt L%d\r\n"),Length));
        goto DropPacket;
    }

    /*
     * Hand the packet to the network module.
     */
    DBGME(0,printf(_T("SerNet got %d (%x)!\n\r"),len,eh.ether_type));
    ifp->if_ipackets++;
    ifp->MBufFact->v->EtherInput(ifp->MBufFact, (PINETDRIVER) ifp, &eh, Pkt);

    /* Nothing more to do
     */
    return;

DropPacket:
    if (Pkt) {
        Pkt->v->Release(Pkt);
        ifp->if_iqdrops++;
    }
    ifp->if_ierrors++;
}

#define MAX_ETHERNET_PACKET_SIZE 1516

/*
 * Start output on interface.
 */
void MCT
SerNetXmit( PSERNET pCs)
{
    struct mbuf     *Request, *m;
    UINT8           *pData8;
    UINT8           *BigBuf; /* XXX */
    UINT            Len;
    UINT16          Length;
    PIFILE          pOut = pCs->Serial;
    
    DBGME(0,_tprintf(_T("Enter SerNetXmit\r\n")));

    Mutex_Lock(&pCs->Lock);
    BigBuf = malloc(MAX_ETHERNET_PACKET_SIZE);
    if (!BigBuf) {
        _tprintf(_T("Malloc failed!"));
        return;
    }
    /* Queue up as many frames as we can
     */
    while ((Request = pCs->is_if.if_snd.ifq_head) != NULL) {

        /* Dequeue the packet we'll xmit
         */
        IF_DEQUEUE(&pCs->is_if.if_snd, m);
        assert(m == Request);

        /* Send header and length out
         */
        Length = (UINT16) m->m_pkthdr.len;
        Length = ntohs(Length);
        pData8 = (UINT8*)&Length;
        memcpy(BigBuf, pData8, 2);
        Len = 2;

        /* Copy the frame out to the chip's memory
         * Beware of odd chunks
         */
        for (;m;m=m->m_next) {

            pData8 = mtod(m,PUINT8);

            memcpy(BigBuf + Len, pData8, m->m_len);//WriteIt(pOut, pData8, m->m_len);
            Len += m->m_len;
        }
        DBGME(0, printf("sernet send %d bytes.\n", Len));
        WriteIt(pOut, BigBuf, Len);

        /* Off it goes.
         */
        pCs->is_if.if_opackets++;
        Request->m_hdr.v->Release((PIBUFFER) Request);
    }

    Mutex_Unlock(&pCs->Lock);

    free(BigBuf);BigBuf = NULL;
}

/*
 * Interrupt service routine.
 * Return TRUE if rescheduling needed.
 */
BOOL MCT
SerNetIsr(PCDEVICE pDev, PBOOL pNotMyInterrupt)
{
    UnusedParameter(pDev);
    UnusedParameter(pNotMyInterrupt);
    return FALSE;
}

/* Thread loop.
 * This is where we can catch our thread and take it
 */
void MCT
SerNetThreadLoop(PSERNET pCs)
{
    DBGME(0,printf(_T("Enter SerNetLoop\r\n")));

    LongTimeout(pCs->Serial);

    /* bugbug unloading */
    for (;;) {
        SerNetRecv(pCs);
    }
}

/* Random other things..
 */
SCODE MCT
SerNetIoctl( PCDEVICE pDev, UINT Prop, caddr_t Data)
{
    PSERNET pCs = (PSERNET) pDev;

    DBGME(0,printf(_T("Enter SerNetIoctl(%x)\r\n"),Prop));

    /* Freshen up statistics, if we need. [no we dont]
     */
    if (Prop == SIOCGIFSTATS) {
    }

    /* We care for requests to turn the interface down/up.
     */
    if (Prop == SIOCSIFFLAGS) {

        if ((pCs->is_if.if_flags & IFF_UP) == 0) {
            /*
             * Down. Put it in low power mode.
             */
        } else {
            /*
             * Up. Turn power back on, this causes an internal reset.
             */
        }
    }
    return NO_ERROR;
}


/* Initialize/Reset/Reinitialize the card
 * NB: Assumes netlock held.
 */
INT MCT
SerNetReset(PCDEVICE pDev)
{
    PSERNET pCs = (PSERNET) pDev;
    struct ifnet *ifp;

    DBGME(1,printf(_T("Enter SerNetReset\r\n")));

    /* Say we are down and reinit sw state
     */
    ifp = &pCs->is_if;
    if (ifp->if_flags & IFF_RUNNING) {
        ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
    }

    pCs->CompleteResets++;

    /* Request the init packet
     */
    WriteIt(pCs->Serial, "\0\0", 2);
    

    /* Say we are up
     */
    ifp->if_flags |= (IFF_UP | IFF_RUNNING);

    /* We are rolling
     */

    DBGME(1,printf(_T("SerNet:Reset done\r\n")));
    return 0;

}

struct INetDriverVtbl SerNetDriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    (_INetDriver_Isr_METHOD_TYPE) SerNetIsr,
    NULL,
    (_INetDriver_Initialize_METHOD_TYPE) SerNetReset,
    (_INetDriver_ThreadLoop_METHOD_TYPE) SerNetThreadLoop,
    NULL,
    (_INetDriver_Start_METHOD_TYPE) SerNetXmit,
    (_INetDriver_Ioctl_METHOD_TYPE) SerNetIoctl,
    NULL
};

/* Once-only initialization.
 * Most of the work is actually done by the Reset function.
 */
PSERNET
SerNetInitialize(_TCHAR *Name, const _TCHAR *Subnet)
{
    PSERNET pCs;
    PIFILE Xport;
    static short unit=0;

    DBGME(1,printf(_T("Enter SerNetInitialize %s\r\n"),Name));

    BindToObject(CurrentNameSpace(), Name, NAME_SPACE_READ|NAME_SPACE_WRITE,
                 &IID_IFile,
                 (void**)&Xport);
    if (Xport == NULL)
        return NULL;

    /* Ok we are doing it for real.
     */
    {
        PIHEAP pIHeap = CurrentHeap();
        pCs = (PSERNET) pIHeap->v->Alloc(pIHeap,
                              HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("sernet"),
                              sizeof(SERNET), 0);
    }
    if (pCs == NULL)
        goto ErrorExit;

    /* Set basic info.
     */
    pCs->Serial = Xport;
    Mutex_Init(&pCs->Lock);

    /* We wont have an address until we got the init pkt,
     * so leave it zero for now.
     */

    /* More init
     */
    pCs->is_if.v = &SerNetDriverVtable;
    pCs->is_if.RefCnt = 1;
    pCs->is_if.if_unit = unit++;
    pCs->is_if.if_name = "sr";
    pCs->is_if.if_type = IFT_ETHER;
    pCs->is_if.if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
    pCs->is_if.if_baudrate = 38400; /* baud, probably */

#if 1
    /* Initialize the ethernet address as the contents of hostname.txt. It is locally unique. This is enough! */
    {
        PINAMESPACE ns = CurrentNameSpace();
        PIFILE defaultHostNameFile;
        SCODE sc = S_OK;
        UINT sizeRead;

        sc = BindToObject(ns, _T("hostname.txt"), NAME_SPACE_READ, &IID_IFile, (void *) &defaultHostNameFile);

        if (FAILED(sc))
            sc = BindToObject(ns, _T("fs/hostname.txt"), NAME_SPACE_READ, &IID_IFile, (void *) &defaultHostNameFile);

        if (FAILED(sc)) {
            DBG(("SerNetInitialize:Could not bind to file hostname.txt\n"));
            goto out;
        }
        
        sc = defaultHostNameFile->v->ReadAt(defaultHostNameFile, UINT64_ZERO, pCs->is_addr, 6,  &sizeRead);
        defaultHostNameFile->v->Release(defaultHostNameFile);
        if (FAILED(sc)) {
            DBG(("SerNetInitialize: failed in reading fs/hostname.txt\n"));
            goto out;
        }

        /* XXX Maybe MMLITE has multiple interfaces, so... 
         * Although the following algorithm still has pobrability 
         * to get two same ethernet address, it is quite small. Sign!
         */
        srand(Uint64ToUint32(CurrentTime()));
        pCs->is_addr[5] += unit;
        pCs->is_addr[4] += unit << 1;
        *(UINT*)pCs->is_addr += (UINT)rand(); /* XXX ntubig and ntubig2 use the same hostname file!!! */
        pCs->is_addr[0] &= 0xfe; /* make sure it is not a link-level mcast */
        if (Subnet) {
            pCs->ac.ac_ipaddr.s_addr = inet_addr(Subnet);
        }
    }
out:
#endif
#if 1
    DBGME(2,_tprintf(_T("SerNet: Interface name sr%d; Addr %02x:%02x:%02x:%02x:%02x:%02x IFile %s\r\n"),
                   pCs->is_if.if_unit,
                   pCs->is_addr[0], pCs->is_addr[1],
                   pCs->is_addr[2], pCs->is_addr[3],
                   pCs->is_addr[4], pCs->is_addr[5],
                   Name));
#endif
    return pCs;

ErrorExit:
    Xport->v->Release(Xport);
    return NULL;
}

SCODE __cdecl SerNetConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        _TCHAR *TransportName,
        /*optional*/ const _TCHAR *Subnet)
{
    PIUNKNOWN pUnk;
    
    UnusedParameter(Name);
    UnusedParameter(Interrupt);
    
    DBGME(0,printf(_T("SerNetConstructor(%x %s %x %x %x %s %s)\r\n"),
                   (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                   (ADDRESS) ppUnk, TransportName, Subnet ? Subnet : ""));

    pUnk = (PIUNKNOWN) SerNetInitialize(TransportName, Subnet);
    
    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,printf(_T("SerNetConstructor failed\r\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl SerNetDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) SerNetConstructor
};

static struct _DRIVERCOB SerNetDriverCobObject = { &SerNetDriverCobVtable, 1 };

PIUNKNOWN SerNetCobMain(void)
{
    SCODE sc;
    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &SerNetDriverCobVtable,
                          "CDeviceConstructor",
                          &IID_IDeviceConstructor);

    DBGME(0,printf(_T("SerNetCobMain() %x\r\n"),sc));

    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &SerNetDriverVtable, "CNetDriver",
                          &IID_INetDriver);
    assert(!FAILED(sc));
    DBGME(0,printf(_T("SerNetCobMain(1) %x\r\n"),sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &SerNetDriverCobObject;
}

#if 0
/* DRIVER.DAT line for this one:
 *
 * SerNet::SerNetNew:NetworkDriver com2
 *
 * It is registered in the name space under SerNet.
 */
EXTERN_C PCDEVICE RTLCALLTYPE 
SerNetNew(_TCHAR *TransportName)
{
    PSERNET pCs;
    SCODE sc;

    sc = BaseInheritClass((ADDRESS *) &SerNetDriverVtable, "CNetDriver",
                          &IID_INetDriver);
    assert(!FAILED(sc));
    DBGME(0,printf(_T("SerNetNew() %x\r\n"),sc));
    if (FAILED(sc))
        return NULL;

    pCs = SerNetInitialize(TransportName);
    if (pCs == NULL)
        return NULL;

    return (PCDEVICE) pCs;
}
#endif
