/* Copyright (c) Microsoft Corporation. All rights reserved. */
//=============================================================================
//  Module: packet..c
//
//  Description: This file contains packet routines and threads for the
//               Simulator NIC driver.
//  
//=============================================================================

#include "simnic.h"

//=============================================================================
//  Function: PacketInit()
//
//  Description: Initialize a new PACKET object.
//=============================================================================

void
PacketInit( 
    IN PACKET * Packet
    )
{
    memset(Packet, 0, sizeof(PACKET));

    Packet->Length = MAX_ETHERNET_PACKET_SIZE;
    Packet->Mode = PacketModeInvalid;
}

//=============================================================================
//  Function: PacketAllocate()
//
//  Description: Allocation a new packet.
//=============================================================================

BOOL
PacketAllocate( 
    IN PACKET ** Packet
    )
{
    //
    //  Allocate a packet. 
    //
        
#if 1
    /* Does not show up in memory stats */
    *Packet = (PACKET*) VirtualAlloc(0, sizeof(PACKET), MEM_COMMIT, PAGE_EXECUTE_READWRITE);
#else
    /* Shows up in memory total */
    *Packet = CurrentHeap()->v->Alloc(
                                CurrentHeap(), 
                                HEAP_DEBUG_NAME("simnic:packet") | HEAP_ZERO_MEMORY, 
                                sizeof(PACKET),
                                0
                                );
#endif

    if ( *Packet != NULL )
    {
        PacketInit(*Packet);

        return TRUE;
    }

    return FALSE;
}

//=============================================================================
//  Function: PacketWriteFile()
//
//  Description: Write a packet using NT WriteFile(). 
//=============================================================================

SCODE
PacketWriteFile(    
    IN HANDLE hFile,
    IN PACKET * Packet
    )
{
    SCODE   Result = S_OK;
    BOOL    bResult = FALSE;
    UINT32  nBytesTransferred = 0;

    //
    //  Call NT WriteFile().
    //

    Packet->Result = E_IO_PENDING;

    Packet->Mode = PacketModeTransmitting;

    bResult = WriteFile(
                    hFile, 
                    Packet->Buffer, 
                    Packet->nBytesAvail, 
                    &nBytesTransferred, 
                    &Packet->Overlapped
                    );
    
    //
    //  If the read request completed immediately,
    //  complete the pending request now.
    //

    if ( bResult != FALSE )
    {
        Packet->Result = S_OK;

        DebugBreak();

        return S_OK;
    }
    
    //
    //  The read request returned FALSE, check to see if its pending.
    //

    Result = GetLastError();

    if ( Result == ERROR_IO_PENDING ) 
    {
        return E_IO_PENDING;
    } 

    //
    //  Check for EOF. 
    //

    if ( Result == ERROR_HANDLE_EOF ) 
    {
        assert(nBytesTransferred == 0);

        DebugBreak();

        return S_FALSE;
    } 

    DebugBreak();

    return Result;
}

//=============================================================================
//  Function: PacketReadFile()
//
//  Description: Read (post) a packet using NT ReadFile(). 
//=============================================================================

SCODE
PacketReadFile( 
    IN HANDLE hFile,
    IN PACKET * Packet
    )
{
    SCODE   Result = S_OK;
    BOOL    bResult = FALSE;
    UINT32  nBytesTransferred = 0;

    //
    //  Call NT ReadFile().
    //

    Packet->nBytesAvail = 0;
    Packet->Result = E_IO_PENDING;

    Packet->Mode = PacketModeReceiving;

    bResult = ReadFile(
                    hFile, 
                    Packet->Buffer, 
                    Packet->Length, 
                    &nBytesTransferred, 
                    &Packet->Overlapped
                    );
    
    //
    //  If the read request completed immediately,
    //  complete the pending request now.
    //

    if ( bResult != FALSE )
    {
        Packet->nBytesAvail = nBytesTransferred;
        Packet->Result = S_OK;

        DebugBreak();

        return S_OK;
    }
    
    //
    //  The read request returned FALSE, check to see if its pending.
    //

    Result = GetLastError();

    if ( Result == ERROR_IO_PENDING ) 
    {
        return E_IO_PENDING;
    } 

    //
    //  Check for EOF. 
    //

    if ( Result == ERROR_HANDLE_EOF ) 
    {
        assert(nBytesTransferred == 0);

        DebugBreak();

        return S_FALSE;
    } 

    DebugBreak(); /* Hey! BUGBUG! What's this doing here!? */

    return Result;
}

//=============================================================================
//  Function: PacketQueueInit()
//
//  Description: Initialize a new PACKETQUEUE object.
//=============================================================================

void
PacketQueueInit(    
    IN PACKETQUEUE * PacketQueue
    )
{
#ifdef _DEBUG
//    printf("PacketQueueInit entered.\n");
#endif

    PacketQueue->Head   = NULL;
    PacketQueue->Tail   = NULL;
    PacketQueue->Length = 0;
    PacketQueue->Lock   = 0;
}

//=============================================================================
//  Function: AcquirePacketQueueLock()
//
//  Description: Acquire the queue lock.
//=============================================================================

void
AcquirePacketQueueLock( 
    IN PACKETQUEUE * PacketQueue
    )
{
    volatile UINT32 * LockPtr = &PacketQueue->Lock;

    _asm
    {
        ;
        ;   Wait for lock to become zero (i.e. unlocked).
        ;

        mov     edx, [LockPtr]

lock_test:

        cmp     [edx], 0
        jne     lock_test
        
        ;
        ;   The lock is currently set to zero (i.e. unlocked).
        ;

        mov     eax, 1
        xchg    [edx], eax
        cmp     eax, 0
        jne     lock_test

        ;
        ;   The lock has been aquired.
        ;
    }
}

//=============================================================================
//  Function: ReleasePacketQueueLock()
//
//  Description: Release the queue lock.
//=============================================================================

void
ReleasePacketQueueLock( 
    IN PACKETQUEUE * PacketQueue
    )
{
    PacketQueue->Lock = 0;
}

//=============================================================================
//  Function: EnqueuePacket()
//
//  Description: Enqueue a packet onto the queue.
//=============================================================================

void
EnqueuePacket(
    IN PACKETQUEUE * PacketQueue,
    IN PACKET * Packet
    )
{
    AcquirePacketQueueLock(PacketQueue);

    assert(Packet != NULL);

    Packet->Next = NULL;

    if ( PacketQueue->Tail != NULL )
    {
        PacketQueue->Tail->Next = Packet;
    }
    else
    {
        PacketQueue->Head = Packet;
    }

    PacketQueue->Tail = Packet;

    PacketQueue->Length++;

    ReleasePacketQueueLock(PacketQueue);
}

//=============================================================================
//  Function: DequeuePacket()
//
//  Description: Dequeue a packet from the queue.
//=============================================================================

PACKET * 
DequeuePacket(
    IN PACKETQUEUE * PacketQueue
    )
{
    PACKET * Packet = NULL;     

    AcquirePacketQueueLock(PacketQueue);

    if ( PacketQueue->Length != 0 )
    {
        Packet = PacketQueue->Head;

        assert(Packet != NULL);

        PacketQueue->Head = PacketQueue->Head->Next;

        PacketQueue->Length--;

        Packet->Next = NULL;

        if ( PacketQueue->Head == NULL )
        {
            PacketQueue->Tail = NULL;

            assert(PacketQueue->Length == 0);
        }
    }

    ReleasePacketQueueLock(PacketQueue);

    return Packet;
}

//=============================================================================
//  Function: GetPacketQueueLength()
//
//  Description: Return the length of the queue.
//=============================================================================

UINT32 
GetPacketQueueLength(
    IN PACKETQUEUE * PacketQueue
    )
{
    UINT32 Length;

    AcquirePacketQueueLock(PacketQueue);

    Length = PacketQueue->Length;

    ReleasePacketQueueLock(PacketQueue);

    return Length;
}

//=============================================================================
//  Function: GetMbufLength()
//
//  Description: Copy an MBUF chain to a piece of memory.
//=============================================================================

UINT32
GetMbufLength(
    IN struct mbuf * MBuf
    )
{
    UINT32 Length = 0;

    while( MBuf != NULL )
    {
        Length += MBuf->m_len;
        MBuf = MBuf->m_next;
    }

    return Length;
}

//=============================================================================
//  Function: CopyMbufToBuffer()
//
//  Description: Copy an MBUF chain to a piece of memory.
//=============================================================================

void
CopyMbufToBuffer(
    IN void * Buffer,
    IN struct mbuf * MBuf
    )
{
    UINT16 *    dp = NULL;
    char *      p = (char *) Buffer;

    while( MBuf != NULL )
    {
        dp = mtod(MBuf, UINT16 *);

        memcpy(p, dp, MBuf->m_len);

        p += MBuf->m_len;

        MBuf = MBuf->m_next;
    }
}

//=============================================================================
//  Function: AllocBufferFromMbuf()
//
//  Description: Allocate a piece of contiguous memory and copy the contents
//               of an MBUF into it. Fail if we're out of memory.
//=============================================================================

void *
AllocBufferFromMbuf(
    IN struct mbuf * MBuf
    )
{
    UINT32  Length = GetMbufLength(MBuf);
    void *  Buffer;

    //
    //  Allocate some memory.
    //

    Buffer = CurrentHeap()->v->Alloc(CurrentHeap(),
                                     HEAP_DEBUG_NAME("simnic:buffer"),
                                     Length, 0);

    if ( Buffer != NULL )
    {
        CopyMbufToBuffer(Buffer, MBuf);
    }

    return Buffer;
}


//=============================================================================
//    Section: Packet_Driver_Vtable::
//
//    Description: Interface to the NDIS filter driver implementation
//=============================================================================

typedef struct _PACKET_DRIVER_VTABLE {
    BOOLEAN (*GetSymbolicName)(
                               OUT UINT16 * SymbolicName
                               );
    HANDLE  (*OpenAdapter)    (
                               UINT16 * SymbolicName
                               );
    BOOL    (*Request)        (
                               IN HANDLE hFile,
                               IN BOOL Set,
                               IN PPACKET_OID_DATA OidData
                               );
    SCODE   (*SetFilter)      (
                               IN HANDLE hFile,
                               IN ULONG  Filter
                               );
    SCODE   (*GetAddress)     (
                               IN HANDLE hFile,
                               IN UINT8 *EthernetAddress
                               );
    SCODE   (*GetSpeed)       (
                               IN HANDLE hFile,
                               OUT ULONG *Speed
                               );
} PACKET_DRIVER_VTABLE, *PPACKET_DRIVER_VTABLE;

//=============================================================================
//    SubSection: PacketDriver::
//
//    Description: Old, homegrown, exclusive-use NDIS filter driver interface
//=============================================================================


//=============================================================================
//  Function: PacketDriverGetSymbolicName().
//
//  Description: Get the NT symbolic name for the packet driver.
//=============================================================================

BOOLEAN
PacketDriverGetSymbolicName(
    OUT UINT16 * SymbolicName
    )
{
    void *      hFile;
    UINT32      BytesReturned;
    UINT16      Buffer[1024];
    UINT16  *   pAdapterInfo;
    BOOL        bResult;

    memset(Buffer, 0, sizeof(Buffer));

    wsprintfW(
        SymbolicName,
        L"\\\\.\\Packet");
    
    hFile = CreateFileW(
                SymbolicName,
                GENERIC_WRITE | GENERIC_READ,
                0,
                NULL,
                OPEN_EXISTING,
                FILE_FLAG_OVERLAPPED,
                0
                );

    if ( hFile == INVALID_HANDLE_VALUE) 
    {
        return FALSE;
    }
    
    bResult = DeviceIoControl(
                        hFile,
                        IOCTL_ENUM_ADAPTERS,
                        NULL,
                        0,
                        Buffer,
                        sizeof(Buffer),
                        &BytesReturned,
                        NULL
                        );

    if ( bResult == FALSE )
    {
        return FALSE;
    }

    pAdapterInfo = (UINT16 *) Buffer;

    //
    // Skip the number of adapters, we'll just take the first one.
    // Making sure there is one..
    //

    if (*pAdapterInfo == 0)
    {
        return FALSE;
    }

    pAdapterInfo += 2; /* a one-wchar 'string' */

    //
    //  Skip the adapter name.
    //

    while( *pAdapterInfo != (UINT16) '\0' )
    {
        pAdapterInfo++;
    }

    pAdapterInfo++;     //... Skip the NULL.

    //
    //  Copy the symbolic name.
    //

    while( *pAdapterInfo != (UINT16) '\0' )
    {
        *SymbolicName++ = *pAdapterInfo++;
    }

    *SymbolicName++ = (UINT16) '\0';

    CloseHandle(hFile);

    return TRUE;
}

//=============================================================================
//  Function: PacketDriverOpenAdapter().
//
//  Description: Open the packet driver.
//=============================================================================

HANDLE
PacketDriverOpenAdapter(
    UINT16 * SymbolicName
    )
{
    UINT16  SymbolicLink[256];
    HANDLE  hFile;

    wsprintfW(
        SymbolicLink,
        L"\\\\.\\%s",
        &SymbolicName[12]
        );

    hFile = CreateFileW(
                SymbolicLink,
                GENERIC_WRITE | GENERIC_READ,
                FILE_SHARE_READ | FILE_SHARE_WRITE,
                NULL,
                OPEN_EXISTING,
                FILE_FLAG_OVERLAPPED,
                0
                );

    return hFile;
}

//=============================================================================
//  Function: PacketDriverRequest().
//
//  Description: Submit a request to the packet driver.
//=============================================================================

BOOL
PacketDriverRequest(
    IN HANDLE hFile,
    IN BOOL Set,
    IN PPACKET_OID_DATA OidData
    )
{
    UINT    BytesReturned;
    BOOL    bResult;
    DWORD   OidCommand;
    DWORD   OidLength;

    OidLength = sizeof(PACKET_OID_DATA) - 1 + OidData->Length;

    if ( Set != FALSE )
    {
        OidCommand = IOCTL_PROTOCOL_SET_OID;
    }
    else
    {
        OidCommand = IOCTL_PROTOCOL_QUERY_OID;
    }
    
    //
    //  Submit request.
    //

    bResult = DeviceIoControl(
                        hFile,
                        OidCommand,
                        OidData,
                        OidLength,
                        OidData,
                        OidLength,
                        &BytesReturned,
                        NULL
                        );

    return bResult;
}

//=============================================================================
//  Function: PacketDriverSetFilter().
//
//  Description: Set capture filters.
//=============================================================================

SCODE
PacketDriverSetFilter(
    IN HANDLE hFile,
    IN ULONG  Filter
    )
{
    ULONG               IoCtlBufferLength = sizeof(PACKET_OID_DATA)-1 + sizeof(ULONG);
    BYTE                IoCtlBuffer[sizeof(PACKET_OID_DATA)-1 + sizeof(ULONG)];
    PPACKET_OID_DATA    OidData = NULL;
    BOOL                bResult;

    memset(IoCtlBuffer, 0, IoCtlBufferLength);
    OidData = (PPACKET_OID_DATA) IoCtlBuffer;
    OidData->Oid = OID_GEN_CURRENT_PACKET_FILTER;
    OidData->Length = sizeof(ULONG);

    memcpy(OidData->Data, &Filter, sizeof Filter);

    bResult = PacketDriverRequest(hFile, TRUE, OidData);

    return (bResult ? S_OK : E_FAIL);
}

//=============================================================================
//  Function: PacketDriverGetAddress().
//
//  Description: Get ethernet address.
//=============================================================================

SCODE 
PacketDriverGetAddress(
    IN HANDLE hFile,
    IN UINT8 *EthernetAddress
    )
{
    ULONG               IoCtlBufferLength = sizeof(PACKET_OID_DATA)-1 + 6;
    BYTE                IoCtlBuffer[sizeof(PACKET_OID_DATA)-1 + 6];
    PPACKET_OID_DATA    OidData = NULL;
    BOOL                bResult;

    memset(IoCtlBuffer, 0, IoCtlBufferLength);
    OidData = (PPACKET_OID_DATA) IoCtlBuffer;
    OidData->Oid = OID_802_3_PERMANENT_ADDRESS;
    OidData->Length = 6;

    bResult = PacketDriverRequest(hFile, FALSE, OidData);

    memcpy(EthernetAddress, OidData->Data, 6);

    return (bResult ? S_OK : E_FAIL);
}


//=============================================================================
//  Function: PacketDriverGetSpeed().
//
//  Description: Get the link speed of the interface below.
//=============================================================================

SCODE 
PacketDriverGetSpeed(
    IN HANDLE hFile,
    OUT ULONG *Speed
    )
{
    ULONG               IoCtlBufferLength = sizeof(PACKET_OID_DATA)-1 + sizeof(ULONG);
    BYTE                IoCtlBuffer[sizeof(PACKET_OID_DATA)-1 + sizeof(ULONG)];
    PPACKET_OID_DATA    OidData = NULL;
    BOOL                bResult;

    memset(IoCtlBuffer, 0, IoCtlBufferLength);
    OidData = (PPACKET_OID_DATA) IoCtlBuffer;
    OidData->Oid = OID_GEN_LINK_SPEED;
    OidData->Length = sizeof(ULONG);

    bResult = PacketDriverRequest(hFile, FALSE, OidData);

    if (!bResult)
    {
#ifdef _DEBUG
//        printf("SPEED  didnt work.. sorry (%x)\n",GetLastError());
#endif
        *Speed = 0; // donno
    } else
        memcpy(Speed, OidData->Data, sizeof(ULONG));

    return (bResult ? S_OK : E_FAIL);
}

//=============================================================================
//    Vtable: PacketDriverVtable[].
//
//    Description: Interface to old, exclusive-use packet driver
//=============================================================================
static const PACKET_DRIVER_VTABLE PacketDriverVtable = {
    PacketDriverGetSymbolicName,
    PacketDriverOpenAdapter,
    PacketDriverRequest,
    PacketDriverSetFilter,
    PacketDriverGetAddress,
    PacketDriverGetSpeed
};

//=============================================================================
//    SubSection: VirtualpcDriver::
//
//    Description: VirtualPC's shared-use NDIS filter driver
//=============================================================================

#define NETSV_DRIVER_NAME					L"\\\\.\\VPCNetS2"

#define	kVPCNetSvVersionMajor	2
#define	kVPCNetSvVersionMinor	6
#define	kVPCNetSvVersion		((kVPCNetSvVersionMajor << 16) | kVPCNetSvVersionMinor)

#define FILE_DEVICE_PROTOCOL				0x8000

enum
{
	kIoctlFunction_SetOid					 = 0,
	kIoctlFunction_QueryOid,
	kIoctlFunction_Reset,
	kIoctlFunction_EnumAdapters,
	kIoctlFunction_GetStatistics,
	kIoctlFunction_GetVersion,
	kIoctlFunction_GetFeatures,
	kIoctlFunction_SendToHostOnly,
	kIoctlFunction_RegisterGuest,
	kIoctlFunction_DeregisterGuest,
	kIoctlFunction_CreateVirtualAdapter,
	kIoctlFunction_DestroyVirtualAdapter,
	kIoctlFunction_GetAdapterAttributes,
	kIoctlFunction_Bogus					 = 0xFFF
};

// These IOCTLs apply only to the control object
#if 0
#define IOCTL_ENUM_ADAPTERS					CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_EnumAdapters,				METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif
#define IOCTL_GET_VERSION					CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_GetVersion,				METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_GET_FEATURES					CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_GetFeatures,				METHOD_BUFFERED, FILE_ANY_ACCESS)

// These IOCTLs apply only to the adapter object
#if 0
#define IOCTL_PROTOCOL_RESET				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_Reset,					METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif
#define IOCTL_CREATE_VIRTUAL_ADAPTER		CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_CreateVirtualAdapter,		METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_DESTROY_VIRTUAL_ADAPTER		CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_DestroyVirtualAdapter,	METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_GET_ADAPTER_ATTRIBUTES		CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_GetAdapterAttributes,		METHOD_BUFFERED, FILE_ANY_ACCESS)

// These IOCTLs apply only to the virtual adapter object
#if 0
#define IOCTL_PROTOCOL_SET_OID				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_SetOid,					METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_PROTOCOL_QUERY_OID			CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_QueryOid,					METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif
#define IOCTL_SEND_TO_HOST_ONLY				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_SendToHostOnly,			METHOD_IN_DIRECT, FILE_ANY_ACCESS)

#define IOCTL_REGISTER_GUEST				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_RegisterGuest,			METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_DEREGISTER_GUEST				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_DeregisterGuest,			METHOD_BUFFERED, FILE_ANY_ACCESS)

// These IOCTLs apply to both the adapter and virtual adapter object
#define IOCTL_GET_STATISTICS				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_GetStatistics,			METHOD_BUFFERED, FILE_ANY_ACCESS)


// NETSV_REGISTER_GUEST.fFlags constants
#define kVirtualSwitchRegisterGuestFlag_AddressActionMask	0x00000003
#define	kVirtualSwitchRegisterGuestFlag_GenerateAddress		0x00000000	// The VPCNetSv driver should generate a unique MAC address
#define	kVirtualSwitchRegisterGuestFlag_AddressIsSuggested	0x00000001	// The VPCNetSv driver should use the specified address if not already in use - generate a new one otherwise
#define kVirtualSwitchRegisterGuestFlag_AddressIsExclusive	0x00000002	// The VPCNetSv driver must use the specified address - fail if not available
#define kVirtualSwitchRegisterGuestFlag_AddressMayDuplicate	0x00000003	// The VPCNetSv driver must use the specified address - may duplicate an existing address

#define kVirtualSwitchRegisterGuestFlag_AddressWasGenerated	0x80000000	// The VPCNetSv driver generated a new MAC address

#define OID_GEN_CURRENT_PACKET_FILTER			0x0001010E	// Set or Query
#define OID_802_3_CURRENT_ADDRESS           	0x01010102	// Query only

//#define NDIS_PACKET_TYPE_DIRECTED				0x00000001
//#define NDIS_PACKET_TYPE_MULTICAST				0x00000002
//#define NDIS_PACKET_TYPE_ALL_MULTICAST			0x00000004
//#define NDIS_PACKET_TYPE_BROADCAST				0x00000008
//#define NDIS_PACKET_TYPE_PROMISCUOUS			0x00000020

#define DEFAULT_PACKET_FILTER	(NDIS_PACKET_TYPE_DIRECTED + NDIS_PACKET_TYPE_MULTICAST + NDIS_PACKET_TYPE_BROADCAST)


#pragma pack(push)
#pragma pack(1)

typedef struct
{
	ULONG			fAdapterID;
}NETSV_CREATE_VIRTUAL_ADAPTER, *PNETSV_CREATE_VIRTUAL_ADAPTER;

#define ETHERNET_ADDRESS_LENGTH 6

typedef struct
{
	USHORT			fVersion;								// Version of this structure
	USHORT			fLength;								// Length of following data
	ULONG			fFlags;									// Flags
	UCHAR			fMACAddress[ETHERNET_ADDRESS_LENGTH];	// Guest VM MAC address
}NETSV_REGISTER_GUEST, *PNETSV_REGISTER_GUEST;

#define NETSV_REGISTER_GUEST_VERSION	0x0002
#define	NETSV_REGISTER_GUEST_LENGTH		(sizeof(NETSV_REGISTER_GUEST) - 2*sizeof(USHORT))

#define MAX_ADAPTERS						256
#define MAX_ADAPTER_BUFFER_SIZE				65536

#pragma pack(pop)


//=============================================================================
//    Function: VirtualpcDriverGetSymbolicName().
//
//    Description: Get the NT symbolic name for the packet driver we want.
//                 The driver supports multiple ones, we'll go for the first
//                 that has an ethernet address assigned and is not in use.
//=============================================================================

static
BOOLEAN
VirtualpcDriverGetSymbolicName(
    OUT UINT16 * SymbolicName
    )
{
    void *        hControl;
    UINT32        BytesReturned;
	UINT32        version;
    UINT16        Buffer[1024];
    UINT16       *pAdapterInfo;
    BOOL          bResult;
	
	hControl = CreateFileW(NETSV_DRIVER_NAME, 
		                  GENERIC_READ | GENERIC_WRITE, 
						  0, NULL, 
						  OPEN_EXISTING, 
						  FILE_FLAG_OVERLAPPED|FILE_FLAG_NO_BUFFERING, 
						  NULL);
						  
	if (hControl == INVALID_HANDLE_VALUE)
	{
//		printf("Failed to open VirtualPC Network driver - not installed?\n");
		return FALSE;
	}

	if (DeviceIoControl(hControl, 
						(DWORD)IOCTL_GET_VERSION,
						NULL, 0,
						&version, sizeof(version),
						&BytesReturned,
						NULL) == FALSE)
	{
//		printf("Bad VirtualPC Network driver - no version?!?\n");
		CloseHandle(hControl);
		return FALSE;
	}
	
	if (version < kVPCNetSvVersion)
	{
//		printf("Warning:  VirtualPC Network driver has unexpected version (Actual: %x, Expected: %x)\n", version, kVPCNetSvVersion);
	}

    bResult = DeviceIoControl(
                        hControl,
                        IOCTL_ENUM_ADAPTERS,
                        NULL,
                        0,
                        Buffer,
                        sizeof(Buffer),
                        &BytesReturned,
                        NULL
                        );

    if ( bResult == FALSE )
    {
        return FALSE;
    }

    pAdapterInfo = (UINT16 *) Buffer;

    //
    // Skip the number of adapters, we'll just take the first one.
    //

    pAdapterInfo += 2; /* a one-wchar 'string' */

    //
    //    Skip the adapter name.
    //

    while( *pAdapterInfo != (UINT16) '\0' )
    {
        pAdapterInfo++;
    }

    pAdapterInfo++;        //... Skip the NULL.

    //
    //    Copy the symbolic name.
    //

    while( *pAdapterInfo != (UINT16) '\0' )
    {
        *SymbolicName++ = *pAdapterInfo++;
    }

    *SymbolicName++ = (UINT16) '\0';

    CloseHandle(hControl)
;
    return TRUE;
}

//=============================================================================
//    Function: VirtualpcDriverOpenAdapter().
//
//    Description: Open the packet driver.
//=============================================================================

static char MicKey[] = "SOFTWARE\\Microsoft\\Invisible Computing";
static char MicKeyMac[] = "SerplexMAC";

static
SCODE
SetMyVirtualMac( IN UCHAR *Mac)
{
    SCODE sc = E_FAIL;
    HKEY Key;
    LONG Return;
    DWORD Length;
    UCHAR Data[128];

    Return = RegCreateKeyEx(HKEY_LOCAL_MACHINE,
                            MicKey,
                            0,
                            NULL, // or "" ?
                            REG_OPTION_NON_VOLATILE,
                            KEY_ALL_ACCESS,
                            NULL,
                            &Key,
                            NULL);
    if (Return != ERROR_SUCCESS) {
//        printf("Failed to create key '%s' (er=%d)\n", MicKey, Return);
        return E_FAIL;
    }

    sprintf((char*)Data,"%x:%x:%x:%x:%x:%x", Mac[0], Mac[1], Mac[2], Mac[3], Mac[4], Mac[5]);
    Length = strlen((char*)Data)+1;
    Return = RegSetValueExA(Key,
                            MicKeyMac,
                            0,
                            REG_SZ,
                            Data,
                            Length);

    if (Return != ERROR_SUCCESS) {
//        printf("Failed to write value '%s' in key '%s' (er=%d)\n", MicKeyMac, MicKey, Return);
        goto Done;
    }

    sc = S_OK;
 Done:
    RegCloseKey(Key);
    return sc;
}

static
SCODE
GetMyVirtualMac( OUT UCHAR *EthernetAddress)
{
    SCODE sc = E_FAIL;
    HKEY Key;
    LONG Return;
    DWORD Type, Length;
    UCHAR Data[128];
    ULONG Mac[6];

    Return = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                          MicKey,
                          0,
                          KEY_READ,
                          &Key);
    if (Return != ERROR_SUCCESS) {
//        printf("Cannot find the '%s' key (er=%d)\n", MicKey, Return);
        return E_FAIL;
    }

    Length = (sizeof Data) - 1;
    Return = RegQueryValueExA(Key,
                              MicKeyMac,
                              0,
                              &Type,
                              Data,
                              &Length);

    if (Return != ERROR_SUCCESS) {
//        printf("Missing value '%s' in key '%s' (er=%d)\n", MicKeyMac, MicKey, Return);
        goto Done;
    }

    if (Type != REG_SZ) {
//        printf("Value '%s\\%s' wrong type x%x (not REG_SZ)\n", MicKey, MicKeyMac, Type);
        goto Done;
    }

    Data[Length] = 0;
    /* JIC Try a few other formats in addition to what we write ourselves
     */
    Return = sscanf((char*)Data,"%x:%x:%x:%x:%x:%x", &Mac[0], &Mac[1], &Mac[2], &Mac[3], &Mac[4], &Mac[5]);
    if (Return != 6)
    Return = sscanf((char*)Data,"%x %x %x %x %x %x", &Mac[0], &Mac[1], &Mac[2], &Mac[3], &Mac[4], &Mac[5]);
    if (Return != 6)
    Return = sscanf((char*)Data,"%x-%x-%x-%x-%x-%x", &Mac[0], &Mac[1], &Mac[2], &Mac[3], &Mac[4], &Mac[5]);
    if (Return != 6) {
//        printf("MAC '%s' in bad taste (not xx:xx:xx:xx:xx:xx)\n", Data);
        goto Done;
    }

    EthernetAddress[0] = (UCHAR) Mac[0];
    EthernetAddress[1] = (UCHAR) Mac[1];
    EthernetAddress[2] = (UCHAR) Mac[2];
    EthernetAddress[3] = (UCHAR) Mac[3];
    EthernetAddress[4] = (UCHAR) Mac[4];
    EthernetAddress[5] = (UCHAR) Mac[5];

    sc = S_OK;
 Done:
    RegCloseKey(Key);
    return sc;
}

static
HANDLE
VirtualpcDriverInitializeInstance(
    UINT16 * SymbolicName,
    UINT16 * AdapterName,
    UINT32 AdapterID
    )
{
    HANDLE hVirtualAdapter;

    wsprintfW(
              AdapterName,
              L"\\\\.\\%s_%08lX", 
              &SymbolicName[12],
              AdapterID);
    AdapterName[1024-1]=L'\0';
	
    hVirtualAdapter = CreateFileW(AdapterName, 
                                  GENERIC_READ | GENERIC_WRITE, 
                                  0, 
                                  NULL, 
                                  OPEN_EXISTING, 
                                  FILE_FLAG_OVERLAPPED|FILE_FLAG_NO_BUFFERING, 
                                  NULL);


	// Register with the driver
    if (hVirtualAdapter != INVALID_HANDLE_VALUE)
	{
        ULONG  BytesReturned;
        SCODE sc;
        NETSV_REGISTER_GUEST GuestInfo;

        memset(&GuestInfo,0,sizeof GuestInfo);
        GuestInfo.fVersion	 = NETSV_REGISTER_GUEST_VERSION;
        GuestInfo.fLength	 = NETSV_REGISTER_GUEST_LENGTH;

        /* Do we have a MAC to use
         */
        sc = GetMyVirtualMac(GuestInfo.fMACAddress);

        /* If not, generate a new one
         */
        if (FAILED(sc))
     	{
            GuestInfo.fFlags = kVirtualSwitchRegisterGuestFlag_GenerateAddress;
        }
        else
     	{
            GuestInfo.fFlags = kVirtualSwitchRegisterGuestFlag_AddressIsExclusive;
        }

        if (DeviceIoControl(hVirtualAdapter,
					    IOCTL_REGISTER_GUEST,
						&GuestInfo, sizeof(GuestInfo),
						&GuestInfo, sizeof(GuestInfo),
						&BytesReturned,
						NULL) == FALSE) 
        {
//            printf("VirtualSwitchRegisterAdaptor failed (error %d)\n", GetLastError());
            CloseHandle(hVirtualAdapter);
            hVirtualAdapter = INVALID_HANDLE_VALUE;
        }
	
        /* Remember if we got a new one
         */
        if (GuestInfo.fFlags & kVirtualSwitchRegisterGuestFlag_AddressWasGenerated) 
        {
#if 0
            int i;
            printf("VirtualPCDriver::A new Guest MAC address was generated");
            for (i = 0; i < ETHERNET_ADDRESS_LENGTH; i++)
                printf(":%x",GuestInfo.fMACAddress[i]);
            printf("\n");
#endif
            SetMyVirtualMac(GuestInfo.fMACAddress);
        }
    }
	
    return hVirtualAdapter;
}


static
HANDLE
VirtualpcDriverOpenAdapter(
    UINT16 * SymbolicName
    )
{
    UINT16    AdapterName[1024];
    HANDLE    hHostAdapter = INVALID_HANDLE_VALUE;
    UINT32    NextAdapterID = 0;
    HANDLE    hVirtualAdapter = INVALID_HANDLE_VALUE;
    ULONG     bytesReturned;
    NETSV_CREATE_VIRTUAL_ADAPTER createAdapter;

    wsprintfW(
              AdapterName,
              L"\\\\.\\%s",
              &SymbolicName[12]
              );

    hHostAdapter = CreateFileW(
                               AdapterName,
                               GENERIC_WRITE | GENERIC_READ,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL,
                               OPEN_EXISTING,
                               FILE_FLAG_OVERLAPPED,
                               0
                               );

    if (hHostAdapter == INVALID_HANDLE_VALUE)
    {
//        printf("Failed to open VirtualPC Host adapter - not installed?\n");
        goto Bad;
    }

    // Create an adapter instance, starting low
    createAdapter.fAdapterID = 0;
	
    if (DeviceIoControl(hHostAdapter,
						(DWORD)IOCTL_CREATE_VIRTUAL_ADAPTER,
						&createAdapter, sizeof(createAdapter),
						&createAdapter, sizeof(createAdapter),
						&bytesReturned,
						NULL) == FALSE)
    {
//        printf("Failed to create VirtualPC Host adapter instance\n");
        goto Bad;
    }

    hVirtualAdapter = VirtualpcDriverInitializeInstance(
                                 SymbolicName,
                                 AdapterName,
                                 createAdapter.fAdapterID);

    // Done
    //
    sprintf(SymbolicName,"%ls",AdapterName);

 Bad:
    /* This piece of junk doesnt work if I close hHostAdapter */
    if (hVirtualAdapter == INVALID_HANDLE_VALUE)
        CloseHandle(hHostAdapter);
    return hVirtualAdapter;
}

//=============================================================================
//    Function: VirtualpcDriverRequest().
//
//    Description: Submit a control request to the packet driver.
//=============================================================================

#define VirtualpcDriverRequest PacketDriverRequest

//=============================================================================
//    Function: VirtualpcDriverSetFilter().
//
//    Description: Set capture filters.
//=============================================================================

#define VirtualpcDriverSetFilter PacketDriverSetFilter

//=============================================================================
//    Function: VirtualpcDriverGetAddress().
//
//    Description: Get the card's ethernet address.
//=============================================================================

static
SCODE 
VirtualpcDriverGetAddress(
    IN HANDLE hFile,
    IN UINT8 *EthernetAddress
    )
{
    UnusedParameter(hFile);

    return GetMyVirtualMac(EthernetAddress);
}


//=============================================================================
//    Function: VirtualpcDriverGetSpeed().
//
//    Description: Get the link speed of the interface below.
//=============================================================================

static
SCODE 
VirtualpcDriverGetSpeed(
    IN HANDLE hFile,
    OUT ULONG *Speed
    )
{
    /* This idiotic thing cannot do this.
     */
    *Speed = 10*1000*1000;
    return S_OK;
}

//=============================================================================
//    Vtable: VirtualpcDriverVtable[].
//
//    Description: Interface to old, exclusive-use packet driver
//=============================================================================
static const PACKET_DRIVER_VTABLE VirtualpcDriverVtable = {
    VirtualpcDriverGetSymbolicName,
    VirtualpcDriverOpenAdapter,
    VirtualpcDriverRequest,
    VirtualpcDriverSetFilter,
    VirtualpcDriverGetAddress,
    VirtualpcDriverGetSpeed
};

//=============================================================================
//  Function: NetworkInit()
//
//  Description: This is the NT receive thread that handles completed
//               packets from the NT network driver and queues them
//               for use by the MMLITE network stack.
//
//=============================================================================

SCODE 
NetworkInit(
    IN PNETWORK Network
    )
{
    BOOL        bResult, OldDriver = FALSE;
    SCODE       Result;
    UINT16      SymbolicName[MAX_LINK_NAME_LENGTH];
    UINT32      i;
    PACKET *    Packet;
    const PACKET_DRIVER_VTABLE *Driver;

#ifdef _DEBUG
//    printf("NetworkInit entered.\n");
#endif

    //
    //  Start from scratch.
    //

 TryAgain:
    Driver = (OldDriver) ? &PacketDriverVtable : &VirtualpcDriverVtable;
    memset(Network, 0, sizeof(NETWORK));

    //
    //  Make sure we have an interface, lest we crash and burn later
    //

    bResult = Driver->GetSymbolicName(SymbolicName);

    //
    //    Try again for the old one if the new one failed
    //

    if ((FALSE == bResult) && !OldDriver)
    {
        OldDriver = TRUE;
        goto TryAgain;
    }

    if (bResult == FALSE)
    {
#ifdef _DEBUG
//        printf("NetworkInit: no NIC.\n");
#endif
        return E_NO_INTERFACE;
    }

    //
    //  Open the NDIS packet driver so we can send and receive.
    //

    Network->hFile = Driver->OpenAdapter(SymbolicName);

    if (Network->hFile == INVALID_HANDLE_VALUE)
    {
#ifdef _DEBUG
//        printf("SimNic::NetworkInit: no suitable NIC.\n");
#endif
        return E_NO_INTERFACE;
    }

    //
    //  Initialize our descriptor
    //

    PacketQueueInit(&Network->FreePacketQueue);

    Network->FreePacketQueue.Signature = FREE_QUEUE_SIG;

    PacketQueueInit(&Network->ReceivePacketQueue);

    Network->ReceivePacketQueue.Signature = RECV_QUEUE_SIG;

    PacketQueueInit(&Network->TransmitPacketQueue);

    Network->TransmitPacketQueue.Signature = SEND_QUEUE_SIG;

    //
    //  Build up the transmit packet pool.
    //

    for(i = 0; i < MAX_TRANSMIT_POOL_SIZE; ++i)
    {
        bResult = PacketAllocate(&Packet);

        assert(Packet != NULL);

        if ( bResult != FALSE )
        {
            EnqueuePacket(
                    &Network->FreePacketQueue,
                    Packet
                    );
        }
    }

    //
    //  Build up the receive packet pool.
    //

    for(i = 0; i < MAX_RECEIVE_POOL_SIZE; ++i)
    {
        bResult = PacketAllocate(&Packet);

        assert(Packet != NULL);

        if ( bResult != FALSE )
        {
            EnqueuePacket(
                    &Network->ReceivePacketQueue,
                    Packet
                    );
        }
    }

    //
    //  Set up the capture filter.
    //

    Result = Driver->SetFilter(
                    Network->hFile,
                    NDIS_PACKET_TYPE_BROADCAST | 
                    NDIS_PACKET_TYPE_MULTICAST |
                    NDIS_PACKET_TYPE_ALL_MULTICAST |
                    NDIS_PACKET_TYPE_DIRECTED       
                    );

    //
    //  Get the ethernet address for this adapter.
    //

    Result = Driver->GetAddress(
                    Network->hFile, 
                    Network->EthernetAddress
                    );

    //
    //  Get the speed of this adapter.
    //

    Result = Driver->GetSpeed(
                                      Network->hFile, 
                                      &Network->BaudRate
                                      );

    //
    //  Create the I/O completion ports for send and
    //  receive operations to and from the packet driver.
    //

    Network->IoCompletionPort = CreateIoCompletionPort(
                                        Network->hFile,
                                        NULL,
                                        (ULONG) Network, 
                                        0
                                        );

    assert(Network->IoCompletionPort != NULL);

    //
    //  Now create the send and receive threads that will
    //  handle the interactions with the network.
    //

    Network->hThread = CreateThread( 
                            NULL,
                            0,
                            (LPTHREAD_START_ROUTINE) NetworkThread,
                            Network,
                            0,
                            0
                            );

    //
    //  Give the NT threads time to get running. This
    //  is completely arbitrary.
    //

    SleepUntil(TIME_RELATIVE(TIME_MILLIS(100)));

    Network->bInitialized = TRUE;

    return S_OK;
}

/* Guard CRT library from "interrupts" */
PRIVATE INT __cdecl NT_printf(const char *format, ...)
{
    va_list arg;
    INT ret;
    static int (*vpr)() = NULL;
    HANDLE h;
    //UINT32 State;

    if (vpr == NULL) {
        h = GetModuleHandle(_T("msvcrt.dll"));
        assert(h != INVALID_HANDLE_VALUE);
        vpr = (void*) GetProcAddress(h, "vprintf");
        assert(vpr != NULL);
    }

    //TURN_INTERRUPTS_OFF(State);
    va_start(arg,format);
    ret = (*vpr)( format, arg );
    va_end(arg);
    //RESTORE_INTERRUPTS(State);

    return ret;

}

void 
PacketDump(
    PPACKET Pkt,
    char *Headline
    )
{
    UINT32 i;

 //   if (debug_level < 3)
   //     return;
    //printf(Headline);
    for (i = 0;i < Pkt->nBytesAvail; i++)
        NT_printf(" %x",Pkt->Buffer[i]);
    NT_printf("\n");
}
//=============================================================================
//  Function: NetworkThread()
//
//  Description: Send and receive thread.
//
//=============================================================================

void 
NetworkThread(
    IN void * ThreadContext
    )
{
    PNETWORK        Network;
    UINT32          Transferred;
    PACKET *        Packet;
    BOOL            bResult;
    OVERLAPPED *    Overlapped;
    ULONG           Key;

#ifdef _DEBUG
//  printf("NetworkThread.\n");
#endif

    Network = (PNETWORK) ThreadContext;

    //
    //  We have to run a little faster than normal 
    //  in order to handle a lot of network packets.
    //

    SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);

    //
    //  The loop below is the main processing loop for all
    //  receives and transmits. 
    //

    for (;;) 
    {
        //
        //  Wait for a receive to complete.
        //
        
        Packet = NULL;
        Overlapped = NULL;

        bResult = GetQueuedCompletionStatus(
                            Network->IoCompletionPort,
                            &Transferred,
                            &Key,
                            &Overlapped,
                            INFINITE
                            );

        //
        //  If the completion was ok, move the packet to the 
        //  MMLITE network stack queue for further processing.
        //

        Packet = (PACKET *) Overlapped;

//PacketDump(Packet, "Simnic got: ");
        if ( bResult != FALSE && Packet != NULL )
        {
            assert(Packet->Mode != PacketModeInvalid);

            if ( Packet->Mode == PacketModeReceiving )
            {
                //
                //  Finish initializing pending packet.
                //  
    
                Packet->nBytesAvail = Transferred;

                Packet->Result = S_OK;

                //
                //  Move it on up to the application.
                //

                EnqueuePacket(&Network->ReceivePacketQueue, Packet);

#ifdef _DEBUG
//              printf("Thread: ReceivePacketQueue: Packet queue length = %u.\n", Network->ReceivePacketQueue.Length);
#endif
            }
            else
            {
                assert(Packet->Mode == PacketModeTransmitting);

                //
                //  Move the packet back to the free queue.
                //

                PacketInit(Packet);

                EnqueuePacket(&Network->FreePacketQueue, Packet);

#ifdef _DEBUG
//              printf("Thread: FreePacketQueue: Packet queue length = %u.\n", Network->FreePacketQueue.Length);
#endif

            }
        }
        else
        {
            // 
            // Unknown error 
            //

            Key = GetLastError();
            DebugBreak();
        }
    }
}

