/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include "simnic.h"
#include <mmmacros.h>

SCODE MCT SimNicQueryInterface(PINETDRIVER This, REFIID Iid, void** ppNew);
UINT MCT SimNicAddRef(PINETDRIVER This);
UINT MCT SimNicRelease(PINETDRIVER This);

/* INetDriver methods */
PRIVATE BOOL MCT SimNicIsr(PINETDRIVER iThis, PBOOL pNotMyInterrupt);
PRIVATE void MCT SimNicRecvInterrupt(PINETDRIVER iThis);
PRIVATE BOOL MCT SimNicInterrupt(PINETDRIVER iThis);
PRIVATE INT MCT SimNicOutput(struct ifnet *pThis, struct mbuf *Pkt, struct sockaddr *Addr, struct rtentry *Route);
PRIVATE void MCT SimNicXmit(PINETDRIVER iThis);
PRIVATE SCODE MCT SimNicIoctl(PINETDRIVER iThis, UINT Prop, PTR Data);
PRIVATE void MCT SimNicThreadLoop(PINETDRIVER iThis);
PRIVATE INT MCT SimNicInitialize(PINETDRIVER iThis);

PUBLIC DLL_EXPORT PIUNKNOWN RTLCALLTYPE SimNicNew(void);

typedef struct _DRIVER_INSTANCE 
{
    struct arpcom       ac;             /* standard interface header    */
    PNETWORK            Network;
} DRIVER_INSTANCE, SimNic;

#define SimNicDestroy NULL
#include <s_SimNic_v.c>

#define is_if ac.ac_if
#define is_addr ac.ac_enaddr
#define misscnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_MISS]
#define rstrtcnt ac.ac_if.if_data.ifi_errors[IFI_ERR_RESET]
#define bablcnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_XTIM]
#define merrcnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM]
#define sqercnt  ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+1]
#if _LANCE_DEBUG
#define spurious  ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+2]
#define sawcsr   ac.ac_if.if_data.ifi_errors[IFI_ERR_MEM+3]
#endif


SCODE MCT SimNicQueryInterface(PINETDRIVER This, REFIID Iid,
                                     void** ppNew)
{
    /* BUGBUG spooky */
    PDRIVER_INSTANCE Driver = (PDRIVER_INSTANCE) This;

    if (UuidCmp(&IID_INetDriver, Iid) || UuidCmp(&IID_IDevice,Iid) || 
        UuidCmp(&IID_IFile, Iid) || UuidCmp(&IID_IUnknown,Iid) ) 
    {
        ++(Driver->ac.ac_if.RefCnt);
        *ppNew = (PIUNKNOWN) Driver;
        return NO_ERROR;
    }

    *ppNew = NULL;
    return E_NO_INTERFACE;
}


UINT MCT SimNicAddRef(PINETDRIVER This)
{
    PDRIVER_INSTANCE Driver = (PDRIVER_INSTANCE) This;
    return ++(Driver->ac.ac_if.RefCnt);
}


#undef Free /* weird */

UINT MCT SimNicRelease(PINETDRIVER This)
{
    PDRIVER_INSTANCE Driver = (PDRIVER_INSTANCE) This;
    UINT Refs = --(Driver->ac.ac_if.RefCnt);
    if (Refs == 0)
        CurrentHeap()->v->Free(CurrentHeap(), 0, Driver);
    return Refs;
}


DRIVER_INSTANCE *CreateDriverInstance(PNETWORK Network)
{
    static short unit = 0;
    DRIVER_INSTANCE *   pThis;

    pThis = CurrentHeap()->v->Alloc(CurrentHeap(),
                            HEAP_DEBUG_NAME("simnic:inst") | HEAP_ZERO_MEMORY,
                            sizeof(DRIVER_INSTANCE), 0);

    /* Initialize the standard interface descriptor */
    pThis->is_if.v = &SimNicVtbl;
    pThis->is_if.RefCnt = 1;
    Condition_Init(&pThis->is_if.IntrPoint);

    pThis->is_if.if_unit = unit++;
    pThis->is_if.if_name = "sv";
    pThis->is_if.if_type = IFT_ETHER;
    pThis->is_if.if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST | IFF_UP | IFF_RUNNING;

    /* NT's speed is scaled by 100. Wisely, as we'll soon overflow to 10Gbps.
     */
    pThis->is_if.if_baudrate = Network->BaudRate * 100;

    memcpy(pThis->is_addr, Network->EthernetAddress, 6);

    pThis->is_if.if_addrlen = 6;

#ifdef _DEBUG
    printf(_T("SimNic CardAddr %02x:%02x:%02x:%02x:%02x:%02x\n"),
           pThis->is_addr[0], pThis->is_addr[1],
           pThis->is_addr[2], pThis->is_addr[3],
           pThis->is_addr[4], pThis->is_addr[5]);
#endif    
    pThis->Network = Network;
    return pThis;
}

/* Interrupt service routine. Return TRUE if we do have an interrupt 
 * pending, after telling the chip to keep quiet.
 */
PRIVATE BOOL MCT SimNicIsr(PINETDRIVER iThis, PBOOL pNotMyInterrupt)
{
    PDRIVER_INSTANCE di = (PDRIVER_INSTANCE) iThis;
    Condition_InterruptSignal(&di->is_if.IntrPoint);
    return TRUE;
}

/* Helper routine for creating an MBUF from a flat piece of memory. */
PRIVATE PIBUFFER SimNicRecvPkt(PINETDRIVER iThis, UINT8 * RBuf,
                               int len, OUT struct ether_header * eh)
{
    PDRIVER_INSTANCE di = (PDRIVER_INSTANCE) iThis;
    SCODE sc;
    struct ifnet *ifp = &di->is_if;
    PIBUFFER mh;
    UINT16 ether_type = 0;
    UINT wlen;

    /* Sanity check the length, we get garbage sometimes.
     */
    if (! (len >= 0 && len <= MAX_ETHERNET_PACKET_SIZE))
        return NULL;

    /* Check whether we have been correctly initialized,
     * and that the interface is actually up and running
     */
    if ((di->ac.ac_if.MBufFact == NULL) ||
        ((di->is_if.if_flags & IFF_UP) == 0))
        return NULL;

    sc = di->ac.ac_if.MBufFact->v->AllocateBuffer(di->ac.ac_if.MBufFact,
                                                  0, ifp, &mh);
    if (FAILED(sc))
        goto DropPacket;

    /* Read&fix ethernet header */
    assert(sizeof(struct ether_header) == MIN_ETHERNET_PACKET_SIZE);
    memcpy(eh, RBuf, sizeof(struct ether_header));
    RBuf += sizeof(struct ether_header);
    ether_type = ntohs(eh->ether_type);
    eh->ether_type = ether_type;

    sc = mh->v->WriteAt(mh, UINT64_ZERO, RBuf, len, &wlen);
    if (FAILED(sc)) {
        mh->v->Release(mh);
        goto DropPacket;
    }
    assert(wlen == (UINT) len);
    assert(((struct mbuf *)mh)->m_pkthdr.len == len);
    assert(((struct mbuf *)mh)->m_hdr.mh_len == len); /* XXX could be split into multiple mbufs */

/* OkPacket: */
    ifp->if_ipackets++;
    return mh;
 DropPacket:
    ifp->if_ierrors++;
    ifp->if_iqdrops++;
    return NULL;
}

PRIVATE void MCT SimNicRecvInterrupt(PINETDRIVER iThis)
{
    PDRIVER_INSTANCE di = (PDRIVER_INSTANCE) iThis;
    PACKET *Packet;
    NETWORK *Network = di->Network;
    struct ifnet *ifp = &di->is_if;
    PMBUF m = NULL;
    struct ether_header eh;
    UINT32 DataLength;

    assert(sizeof(ETHERNET_HEADER) == MIN_ETHERNET_PACKET_SIZE);

    /* Keep looping as long as we have packets to process. */
    while( GetPacketQueueLength(&Network->ReceivePacketQueue) != 0 ) {
        /* Dequeue the next packet from the receive queue. */
        Packet = DequeuePacket(&Network->ReceivePacketQueue);
        assert(Packet != NULL);

        if ( Packet != NULL ) {
            assert( Packet->Result == S_OK );
            assert(Packet->Length >= MIN_ETHERNET_PACKET_SIZE);
            assert(Packet->Length <= MAX_ETHERNET_PACKET_SIZE);

            /* Create and copy the packet into an MBUF
             * and move it up the network stack.
             */
            DataLength = Packet->Length - sizeof(ETHERNET_HEADER);
            assert(DataLength <= MAX_ETHERNET_DATA_SIZE);

            m = SimNicRecvPkt(di, 
                              Packet->Buffer,
                              DataLength,
                              &eh);
            /* The packet data has been copied to the MBUF,
             * therefore we can recycle this packet now.
             */
            PacketInit(Packet);
            PacketReadFile(Network->hFile, Packet);

            /* Process the packet up the stack - or not. */
            if (m != NULL) {
                assert(di->ac.ac_if.MBufFact != NULL);
                di->ac.ac_if.MBufFact->v->EtherInput(di->ac.ac_if.MBufFact,
                                                     ifp, &eh, m);
            }
        }
    }
}

PRIVATE BOOL MCT SimNicInterrupt(PINETDRIVER iThis)
{
    PDRIVER_INSTANCE di = (PDRIVER_INSTANCE) iThis;

    /* check if we can receive more packets */
    SimNicRecvInterrupt(iThis);

    /* if applicable, xmit some more stuff from the queue */
    if (di->is_if.if_flags & IFF_OACTIVE)
        SimNicXmit(iThis);

    return TRUE;
}


PRIVATE void MCT 
SimNicThreadLoop(PINETDRIVER iThis)
{
    PDRIVER_INSTANCE di = (PDRIVER_INSTANCE) iThis;
    TIME ThreadTimeout = TIME_RELATIVE(TIME_MILLIS(10));

    for (;;) {
        Condition_Wait(&di->ac.ac_if.IntrPoint, NULL, &ThreadTimeout);
        iThis->v->HandleInterrupt(iThis);
    }
}


PRIVATE INT MCT SimNicInitialize( PINETDRIVER iThis )
{
    PDRIVER_INSTANCE di = (PDRIVER_INSTANCE) iThis;
    NETWORK *   Network = di->Network;
    PACKET *    Packet;
    UINT32      i;
    static BOOL FirstTime = FALSE;

    /* stats.. */
    di->is_if.if_data.ifi_errors[IFI_ERR_RESET]++;

    if (FirstTime == FALSE) {
        FirstTime = TRUE;

        assert(Network->bInitialized);

        for (i = 0; i < MAX_RECEIVE_POOL_SIZE; i++) {
            Packet = DequeuePacket(&Network->ReceivePacketQueue);
            assert(Packet != NULL);
            PacketInit(Packet);
            PacketReadFile(Network->hFile, Packet);
        }
    }

    return 0;
}

PRIVATE BOOL SimNicSendPkt(NETWORK * Network, PMBUF MBuf)
{
    PACKET * Packet;
    SCODE sc;

    if (!Network->bInitialized)
        return FALSE;

    /* Allocate a new packet. */
    Packet = DequeuePacket(&Network->FreePacketQueue);
    if (Packet == NULL)
        return FALSE;

    /* Reset it. */
    PacketInit(Packet);

    /* Get the MBUF length. */
    Packet->nBytesAvail = GetMbufLength(MBuf);
    /* Length was checked to be less than ETHERMTU but lets make sure */
    assert(Packet->nBytesAvail <= Packet->Length);

    /* Copy the MBUF into the packet buffer. */
    CopyMbufToBuffer(Packet->Buffer, MBuf);

    /* Send the packet. */
    sc = PacketWriteFile(Network->hFile, Packet);
    return (sc != S_FALSE);
}

PRIVATE void MCT SimNicXmit(PINETDRIVER iThis)
{
    PDRIVER_INSTANCE di = (PDRIVER_INSTANCE) iThis;
    PMBUF request;

    /*
     * If we're called, there's stuff on the queue.  Keep oactive flag
     * high until the output queue is empty again.
     */
    di->is_if.if_flags |= IFF_OACTIVE;

    /* Lock queue while we work on it
     */
    di->is_if.MBufFact->v->NetLock(di->is_if.MBufFact);

    for (;;) {
        /* Take the next transmit request, if any.
         */
        request = di->is_if.if_snd.ifq_head;

        /* queue is empty, we're done */
        if (request == NULL)  {
            di->is_if.if_flags &= ~IFF_OACTIVE;
            break;
        }

        /* See if we can send it
         * If that works, dequeue, free and continue
         */
        if (SimNicSendPkt(di->Network, request)) {
            IF_DEQUEUE(&di->is_if.if_snd, request);
            di->is_if.if_opackets++;
            request->m_hdr.v->Release((PIBUFFER) request);
        }
        /* Otherwise we better stop now
         * NB: We *could* also drop all packets, after all the wire
         * is supposed to be lossy, right ?
         */
        else break;
    }
    /* Unlock queue
     */
    di->is_if.MBufFact->v->NetUnlock(di->is_if.MBufFact);
}

PRIVATE SCODE MCT SimNicIoctl(PINETDRIVER iThis, UINT Prop, PTR Data)
{
    PDRIVER_INSTANCE di = (PDRIVER_INSTANCE) iThis;
    /* We care for requests to turn the interface down/up.
     * But then again, we dont since the receiver checks the flags.
     */
#if 0
    if (Prop == SIOCSIFFLAGS) {
        if ((di->is_if.if_flags & IFF_UP) == 0) {
            /*
             * Interface going DOWN.
             */
            DisableInterrupts(di);
        else
            /*
             * Interface going UP.
             */
            EnableInterrupts(di);
    }
#endif
    return S_OK;
}

#if 1
SCODE SimNicConstructor(PIDEVICECONSTRUCTOR iThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk)
{
    PNETWORK    Network = NULL;
    SCODE       sc;

    UnusedParameter(Name);

    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);

    /* Allocate a NETWORK object. */
    Network = CurrentHeap()->v->Alloc(CurrentHeap(),
                                      HEAP_DEBUG_NAME("simnic:network"),
                                      sizeof(NETWORK), 0);

    if ( Network == NULL ) {
        sc = E_NOT_ENOUGH_MEMORY;
    } else {
        sc = NetworkInit(Network);
        if ( sc == S_OK && Network->bInitialized != FALSE ) {
            *ppUnk = (PIUNKNOWN) CreateDriverInstance(Network);
            sc = S_OK;
        }
    }
    return sc;
}


static struct IDeviceConstructorVtbl SimNicDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) SimNicConstructor
};

//#include <s_SimNicDriver_v.c>

static struct _DRIVERCOB SimNicDriverCobObject = { &SimNicDriverCobVtable, 1 };

PIUNKNOWN SimNicCobMain(void)   /* main() */
{
    SCODE sc;
    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &SimNicDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &SimNicVtbl, _T("CNetDriver"),
                          &IID_INetDriver);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &SimNicDriverCobObject;
}

#else
PUBLIC DLL_EXPORT PIUNKNOWN RTLCALLTYPE SimNicNew(void)
{
    PINETDRIVER DriverInstance = NULL;
    PNETWORK    Network = NULL;
    SCODE       sc;

//  printf("simnic.c:SimNicNew(PIBASERTL)\n");

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &SimNicDriverVtable, _T("CNetDriver"),
                          &IID_INetDriver);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    /* Allocate a NETWORK object. */
    Network = CurrentHeap()->v->Alloc(CurrentHeap(),
                                      HEAP_DEBUG_NAME("simnic:network2"),
                                      sizeof(NETWORK), 0);

    if ( Network != NULL ) {
        sc = NetworkInit(Network);
        if ( sc == S_OK && Network->bInitialized != FALSE )
            DriverInstance = CreateDriverInstance(Network);
    }

    return (PIUNKNOWN) DriverInstance;
}
#endif
