/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    fmradio.c
 *
 * Purpose:
 *
 *    Portable Device driver for FM-radio, over a serial line.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 *
 */

#include <stdio.h>
#include <mmlite.h>
#define _DEBUG 1
#include <diagnostics.h>
#undef DBGLVL
#define DBGLVL 1
#include <assert.h>
#include <mmhal.h>
#include <drivers/drivers.h>

/* Protocol defs
 */
#define T_SEND_STATUS 1
#define T_POWER_DOWN  2
#define T_POWER_UP    3
#define T_RADIO_SEND  4

#define R_STATUS 8
#define R_RADIO  9
#define R_ACCEL  10
#define R_GYRO   11
#define R_MAG    12
#define R_POW_UP 13
#define R_POW_D  14
#define R_SEND_R 15

#define TargetOf(_x_)  (((_x_)>>4)&0xf)
#define LengthOf(_x_) ((_x_)&0xf)
#define MakeHeader(_target_,_length_) ((UINT8)((_target_<<4)|((_length_)&0xf)))

#define MAX_RADIO_DATA 8

/* DOC:
 * Accelerometer packet is 4 bytes:
 *  LSB(x), MSB(x), LSB(y), MSB(y)
 */

/* Turn this on to get statistics about latencies
 */
#define _STATISTICS 0
#if _STATISTICS
#include <stdio.h>
#include <tchar.h>
#endif

/*
 * Concrete class definition for this device
 */

typedef struct CFMRadioDevice {
    struct CDevice CDevice;
    /* Our own state
     */
    _TCHAR *LinkName;
    PIFILE PicLink;
    UINT BaudRate;
    MUTEX ReadLock;
    MUTEX WriteLock;

} CFMRadioDevice, *PCFMRadioDevice;


/* Utilities
 */
#include <arm/at91x40.h>

#define PIN_ATTENTION PIO_SCLK0 /* p13 */
#define PIN_LISTENING PIO_TIOB1  /* p5  */
#define PinHigh(_p_)  ThePio->SetData = (_p_)
#define PinLow(_p_)   ThePio->ClearData = (_p_)
#define PinIs(_p_)    ((ThePio->PinDataStatus & (_p_)) != 0)
#define PinEnable(_p_) ThePio->Enable = (_p_)
#define PinIsIn(_p_)  ThePio->OutDisable = (_p_)
#define PinIsOut(_p_) ThePio->OutEnable = (_p_)

PRIVATE BOOL
InitializePic(void)
{
    DBGME(3,_tprintf(_T("InitializePic\r\n")));

    PinEnable(PIN_LISTENING);
    PinIsIn(PIN_LISTENING);

    PinEnable(PIN_ATTENTION);
    PinIsOut(PIN_ATTENTION);
    PinLow(PIN_ATTENTION);

    return TRUE;
}

PRIVATE SCODE
GetPicAttention(void)
{
    int i, nTries = 0;

 Again:
    /* Ask for attention
     */
    PinHigh(PIN_ATTENTION);
    /* Wait no more than 5 msecs
     */
    for (i = 0; i < 50;i++) {
        Delay(100);
        if (PinIs(PIN_LISTENING))
            break;
    }
    /* Try again if it didnt work
     */
    if (!PinIs(PIN_LISTENING)) {

        /* Make it flip
         */
        PinLow(PIN_ATTENTION);

        if (nTries++ >= 3)
            return E_TIMED_OUT;
        goto Again;
    }
    return S_OK;
}

PRIVATE SCODE
ReleasePic(void)
{
    int i;

    /* Release attention
     */
    PinLow(PIN_ATTENTION);

    /* Wait no more than 5 msecs for pin20 to go low
     */
    for (i = 0; i < 50;i++) {
        Delay(100);
        if (!PinIs(PIN_LISTENING))
            break;
    }

    return (!PinIs(PIN_LISTENING)) ? S_OK : E_TIMED_OUT;
}

PRIVATE SCODE
ReadIt( PIFILE pIn,
        UINT8 *pData,
        UINT *pLen)
{
    static UINT64 Zero = Int64Initializer(0,0);
    UINT Length = *pLen, nBytes;
    SCODE sc;

    if (pIn == NULL)
        return E_INVALID_PARAMETER;
    do {
        
        nBytes = 0;
        sc = pIn->v->ReadAt(pIn,
                          Zero,
                          pData,
                          Length,
                          &nBytes);
        if (FAILED(sc))
            break;

        Length -= nBytes;
        pData += nBytes;

    } while (Length > 0);

    *pLen -= Length;
    return sc;

}

/* Read SIZE bytes from device, block until.
 */
SCODE MCT FMRadioReadAt (PCFMRadioDevice this, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    UINT8 Scratch[16];
    PIFILE Link = this->PicLink;
    UINT nBytes;
    SCODE sc;

    DBGME(0,_tprintf(_T("FMRadioReadAt(%x .. %x %d %x)\r\n"),
                   (ADDRESS) this, (ADDRESS) pBuffer,
                   Size, (ADDRESS) pSizeRead));

    if (pSizeRead)
        *pSizeRead = 0;

    if (Link == NULL)
        return E_OBJECT_NOT_INITIALIZED;

    Mutex_Lock(&this->ReadLock);

    do {
        sc = Link->v->ReadAt(Link,
                             Position,
                             Scratch,
                             1,
                             &nBytes);

        if (FAILED(sc) || (nBytes != 1))
            goto Done;
        
        nBytes = LengthOf(Scratch[0]);

        /* Check where this is coming from
         * BUGBUG: demux properly
         */
#if 0
        if (TargetOf(Scratch[0]) == R_RADIO)
#else
        if (1)
#endif
        {
            if (nBytes <= Size) {
                sc = ReadIt(Link,pBuffer,&nBytes);
            } else {
                sc = ReadIt(Link,Scratch,&nBytes);
                memcpy(pBuffer,Scratch,Size);
                nBytes = Size;
            }
            if (pSizeRead)
                *pSizeRead = nBytes;
        } else {
            /* dontcarefornow
             */
            sc = ReadIt(Link,Scratch,&nBytes);
            nBytes = 0;
        }

    } while (nBytes == 0);

 Done:
    Mutex_Unlock(&this->ReadLock);

    DBGME(0,_tprintf(_T("FMRadioReadAt->%x\r\n"),sc));
    return sc;
}

/* Write SIZE bytes to device, block until.
 */
SCODE MCT FMRadioWriteAt (PCFMRadioDevice this, UINT64 Position, const BYTE *pBuffer,
                            UINT Size, PUINT pSizeWritten)
{
    SCODE sc;
    UINT8 Header;
    PIFILE Link = this->PicLink;

    DBGME(0,_tprintf(_T("FMRadioWriteAt(%x .. %x %d %x)\r\n"),
                   (ADDRESS) this, (ADDRESS) pBuffer,
                   Size, (ADDRESS) pSizeWritten));

    if (Link == NULL)
        return E_OBJECT_NOT_INITIALIZED;

    /* Max size */
    if (Size > MAX_RADIO_DATA)
        Size = MAX_RADIO_DATA;

    Header = MakeHeader(T_RADIO_SEND,Size);

    Mutex_Lock(&this->WriteLock);

    sc = GetPicAttention();
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("GetPicAttention FAILED x%x\r\n"), sc));
        /* Ignore */
    }

    if (pSizeWritten)
        *pSizeWritten = 0;

    sc = Link->v->WriteAt(Link,
                          Position,
                          &Header,
                          1,
                          pSizeWritten);
    if (FAILED(sc))
        goto Done;

    if (pSizeWritten)
        *pSizeWritten = 0;

    sc = Link->v->WriteAt(Link,
                          Position,
                          pBuffer,
                          Size,
                          pSizeWritten);

 Done:
    /* ignore errors on PIC release */
    (void) ReleasePic();

    Mutex_Unlock(&this->WriteLock);

    DBGME(0,_tprintf(_T("FMRadioWriteAt->%x (%d)\r\n"),
                   sc, (pSizeWritten) ? *pSizeWritten : 666));
    return sc;
}

#define THREAD_LOOP 0
#if THREAD_LOOP

/* Interrupt service routine.
 */
BOOL MCT FMRadioIsr(PCFMRadioDevice this, PBOOL pNotMyInterrupt)
{
    return FALSE;
}

/* Thread loop.
 * This is where we can catch our thread and take it
 */
void MCT
FMRadioThreadLoop(PCFMRadioDevice this)
{
    PIFILE Link;
    TIME OneSecondRelative;

    Int32ToInt64(OneSecondRelative, TIME_RELATIVE(TIME_SECONDS(1)));

printf("FMRadioLoop..\n\r");
    /* bugbug unloading */
    for (;;) {
        for (Link = this->PicLink;Link == NULL;Link = this->PicLink)
            SleepUntil(OneSecondRelative);

        sc = Link->v->ReadAt...
    }
}
#else
#define FMRadioIsr NULL
#define FMRadioThreadLoop NULL
#endif

/* SetSize method on serial line
 */
SCODE MCT FMRadioSetSize (PCFMRadioDevice this, UINT64 Size)
{
    DBGME(0,_tprintf(_T("FMRadioSetSize(%x %x.%x)\r\n"),(ADDRESS) this,Size));

    return E_NOT_IMPLEMENTED;
}


/* (Re)Initialization code.
 */
INT  MCT FMRadioInitialize(PCFMRadioDevice this)
{
    DBGME(0,_tprintf(_T("FMRadioInitialize(%x) %x\r\n"),
                   (ADDRESS) this, (ADDRESS) this->PicLink));

    /* Close it if we had it open
     */
    if (this->PicLink) {
        this->PicLink->v->Release(this->PicLink);
    }
    BindToObject(CurrentNameSpace(),
                 this->LinkName, NAME_SPACE_READ|NAME_SPACE_WRITE, &IID_IFile,
                 (void **) &this->PicLink);

    InitializePic();

    return 0;
}

struct CDeviceVtbl FMRadioDriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)) FMRadioReadAt,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Position, const BYTE *pBuffer, UINT Size, PUINT pSizeWritten)) FMRadioWriteAt,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Size)) FMRadioSetSize,
    NULL,
    (BOOL      (MCT *)(PCDEVICE pThis, PBOOL pWakeup)) FMRadioIsr,
    NULL,
    (INT       (MCT *)(PCDEVICE pThis)) FMRadioInitialize,
    (void      (MCT *)(PCDEVICE pThis)) FMRadioThreadLoop
};

/* Constructor
 */
void
CFMRadioDeviceInit (PCFMRadioDevice this, _TCHAR *LinkName, UINT Rate )
{
    this->LinkName = LinkName;
    this->BaudRate = Rate;
    Mutex_Init(&this->ReadLock);
    Mutex_Init(&this->WriteLock);

    this->CDevice.v = &FMRadioDriverVtable;
    this->CDevice.RefCnt = 1;
}


/* Exported constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE FMRadioNew( ADDRESS * pPortBase, PUINT pIrq, _TCHAR *LinkName, PUINT pBaudRate)
{
    PCDEVICE pSer;
    PIHEAP pH = CurrentHeap();

    UnusedParameter(pIrq);

    pSer = (PCDEVICE) pH->v->Alloc( pH, HEAP_ZERO_MEMORY, sizeof(CFMRadioDevice), 0 );
    if (pSer)
        CFMRadioDeviceInit( (PCFMRadioDevice)pSer, LinkName, *pBaudRate );

    return pSer;
}

/* ** 
*/
SCODE __cdecl FMRadioConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        UINT BaudRate,
        _TCHAR *LinkName,
        ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("FMRadioConstructor(%x %s %x %x %x %d %s)\n"),
                   (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                   (ADDRESS) ppUnk, BaudRate, LinkName));

    pUnk = (PIUNKNOWN) FMRadioNew(&RegisterBase,
                                 &Interrupt,
                                 LinkName,
                                 &BaudRate);

    DBGME(0,_tprintf(_T("FMRadioConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("FMRadioConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl FMRadioDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) FMRadioConstructor
};

static struct _DRIVERCOB FMRadioDriverCobObject = { &FMRadioDriverCobVtable, 1 };

PIUNKNOWN FMRadioCobMain(void)
{
    SCODE sc;

    DBGME(0,_tprintf(_T("FMRadioCobMain\r\n")));

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &FMRadioDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &FMRadioDriverVtable, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

#if THREAD_LOOP
#else
    /* Null the thread loop, to save a useless thread
     */
    FMRadioDriverVtable.ThreadLoop = NULL;
#endif

    DBGME(0,_tprintf(_T("FMRadioCobMain ok\r\n")));

    return (PIUNKNOWN) &FMRadioDriverCobObject;
}

