/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    serial.h
 *
 * Purpose:
 *
 *    Portable Device driver for serial lines, ARM AEB 8250 customization.
 *
 * Author:
 *    Scott Corbin (scottc)
 *
 * --------------------------------------------------------------------------
 */

#define TXRX(addr)          ((IO_PORT)(addr + 0*4))
#define BAUD_LSB(addr)      ((IO_PORT)(addr + 0*4))
#define BAUD_MSB(addr)      ((IO_PORT)(addr + 1*4))
#define INTR_ENAB(addr)     ((IO_PORT)(addr + 1*4))
#define INTR_ID(addr)       ((IO_PORT)(addr + 2*4))
#define FIFO_CTL(addr)      ((IO_PORT)(addr + 2*4))
#define LINE_CTL(addr)      ((IO_PORT)(addr + 3*4))
#define MODEM_CTL(addr)     ((IO_PORT)(addr + 4*4))
#define LINE_STAT(addr)     ((IO_PORT)(addr + 5*4))
#define MODEM_STAT(addr)    ((IO_PORT)(addr + 6*4))
#define SCR(addr)           ((IO_PORT)(addr + 7*4))

/* interrupt mask */
#define MODi        0      /* modem status change */
#define TRAi        2
#define RECi        4      /* receive char available */
#define LINi        6      /* line status interrupt (error) */
#define CTIi        0xc    /* character timeout indication (fifo ~= RECi) */
#define MASKi       0xf

/* line control register */
#define iWLS0       0x01    /*word length select bit 0 */        
#define iWLS1       0x02    /*word length select bit 2 */        
#define iSTB        0x04    /* number of stop bits */
#define iPEN        0x08    /* parity enable */
#define iEPS        0x10    /* even parity select */
#define iSP         0x20    /* stick parity */
#define iSETBREAK   0x40    /* break key */
#define iDLAB       0x80    /* divisor latch access bit */
#define i5BITS      0x00    /* 5 bits per char */
#define i6BITS      0x01    /* 6 bits per char */
#define i7BITS      0x02    /* 7 bits per char */
#define i8BITS      0x03    /* 8 bits per char */

/* line status register */
#define iDR         0x01    /* data ready */
#define iOR         0x02    /* overrun error */
#define iPE         0x04    /* parity error */
#define iFE         0x08    /* framing error */
#define iBRKINTR    0x10    /* a break has arrived */
#define iTHRE       0x20    /* tx hold reg is now empty */
#define iTSRE       0x40    /* tx shift reg is now empty */

/* interrupt id regisger */
#define iMODEM_INTR 0x01
#define iTX_INTR    0x02
#define iRX_INTR    0x04
#define iERROR_INTR 0x08

/* interrupt enable register */
#define iRX_ENAB    0x01
#define iTX_ENAB    0x02
#define iERROR_ENAB 0x04
#define iMODEM_ENAB 0x08

/* modem control register */
#define iDTR        0x01    /* data terminal ready */
#define iRTS        0x02    /* request to send */
#define iOUT1       0x04    /* COM aux line -not used */
#define iOUT2       0x08    /* turns intr to 386 on/off */        
#define iLOOP       0x10    /* loopback for diagnostics */

/* modem status register */
#define iDCTS       0x01    /* delta clear to send */
#define iDDSR       0x02    /* delta data set ready */
#define iTERI       0x04    /* trail edge ring indicator */
#define iDRLSD      0x08    /* delta rx line sig detect */
#define iCTS        0x10    /* clear to send */
#define iDSR        0x20    /* data set ready */
#define iRI         0x40    /* ring indicator */
#define iRLSD       0x80    /* rx line sig detect */

/* fifo control register (only in 16550) */
#define iFIFOENA    0x01    /* Enable fifos */
#define iCLRRCVRFIFO 0x02   /* Clear receive fifo */
#define iCLRXMITFIFO 0x04   /* Clear transmit fifo */
#define iDMAMODE    0x08    /* DMA transfer enable */
#define iFIFO1CH    0x00    /* Receive fifo trigger level 1 char */
#define iFIFO4CH    0x40    /* Receive fifo trigger level 4 chars*/
#define iFIFO8CH    0x80    /* Receive fifo trigger level 8 chars*/
#define iFIFO14CH   0xc0    /* Receive fifo trigger level 14 chars*/


/* Defines to customize the common code properly
 */
#define SER_READ_INTR_PORT(_i_,_p_) \
         _i_ = (IOSpaceReadInt8(INTR_ID(_p_)) & MASKi)

#define SER_SPURIOUS_INTERRUPT(_i_) \
        ((_i_) & 1)
#define SER_RECEIVE_INTERRUPT(_i_) \
         ((_i_ == RECi) || (_i_ == CTIi))
#define SER_TRANSMITTED_INTERRUPT(_i_) \
         (_i_ == TRAi)
#define SER_MODEM_INTERRUPT(_i_) \
         (_i_ == MODi)
#define SER_ERROR_INTERRUPT(_i_) \
         (_i_ == LINi)

#define SER_ERRORS(_i_,_p_) SER_ERROR_INTERRUPT(_i_)
#define SER_CLEAR_ERRORS(_i_,_p_)

#define SER_ISR_RECEIVED_DATA(_i_,_p_) \
         SER_RECV_BUFFER_READY(_p_)
#define SER_ISR_TRANSMITTED_DATA(_i_,_p_) \
         (SER_TRANSMITTED_INTERRUPT(_i_) && \
          SER_XMIT_BUFFER_READY(_p_))

#define SER_ENABLE_RECEIVE_INTERRUPTS(_p_) \
         IOSpaceWriteInt8(INTR_ENAB(_p_), iRX_ENAB)
#define SER_ENABLE_TRANSMIT_INTERRUPTS(_p_) \
         IOSpaceWriteInt8(INTR_ENAB(_p_), iRX_ENAB|iTX_ENAB)
#define SER_DISABLE_TRANSMIT_INTERRUPTS SER_ENABLE_RECEIVE_INTERRUPTS
#define SER_DISABLE_INTERRUPTS(_p_) \
         IOSpaceWriteInt8(INTR_ENAB(_p_), 0)

#define SER_ACK_INTERRUPTS(_p_)

#define SER_XMIT_BUFFER_READY(_p_) \
         (IOSpaceReadInt8(LINE_STAT(_p_)) & iTHRE)
#define SER_RECV_BUFFER_READY(_p_) \
         (IOSpaceReadInt8(LINE_STAT(_p_)) & iDR)

#define SER_READ_RECEIVE_BUFFER(_p_) IOSpaceReadInt8(TXRX(_p_))
#define SER_WRITE_XMIT_BUFFER(_p_,_c_) IOSpaceWriteInt8(TXRX(_p_),_c_)

#define SER_SET_BAUD_RATE(_p_,_b_) \
  {  \
         /* NB: UART clock rate = 24MHz */ \
         UINT16 Divisor = (UINT16)((24000000 / 16) / (_b_)); \
         \
         /* Set the divisor latch (DLAB) in the (LCR) */ \
         IOSpaceWriteInt8(LINE_CTL(_p_), \
                          (UINT8)(IOSpaceReadInt8(LINE_CTL(_p_)) | iDLAB)); \
         \
         /* Set the divisor latch value. */ \
         \
         IOSpaceWriteInt8(BAUD_MSB(_p_), (UINT8)((Divisor >> 8) & 0xff)); \
         IOSpaceWriteInt8(BAUD_LSB(_p_), (UINT8)(Divisor & 0xff)); \
  }

#define SER_SET_LINE_PARAMS(_p_,_l_) \
         IOSpaceWriteInt8(LINE_CTL(_p_), (UINT8)_l_)

/* Set LCR to 8-bit data, 1-stop bit, no parity.
 */
#define SER_DEFAULT_LINE_PARAMS i8BITS

#define SER_SET_MODEM_BITS(_p_,_b_) \
         IOSpaceWriteInt8(MODEM_CTL(_p_), _b_)

#define SER_DTR     iDTR
#define SER_RTS     iRTS
#define SER_INTGATE iOUT2

#define SER_INTERRUPTS_ON_XMIT_EMPTY 0

#define SER_SPECIAL_INIT(_p_) \
    { \
          /* Enable fifo and see if it sticks */ \
          IOSpaceWriteInt8(FIFO_CTL(_p_), iFIFOENA | iFIFO14CH); \
          if ((IOSpaceReadInt8(FIFO_CTL(_p_)) & iFIFO14CH) != iFIFO14CH) \
                /* iff both bits are set then 82500/16550 ok, else drop */ \
                IOSpaceWriteInt8(FIFO_CTL(_p_), 0x00); \
    }

#define SER_TEST_FOR_MORE_INTERRUPTS(_p_,_i_,_l_) \
   { \
         SER_READ_INTR_PORT(_i_,_p_); \
         if (((_i_) & 1) == 0) \
             goto _l_; \
   }

/* Device driver proper
 */
#include "serial.h"
