/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    dma_ser.h
 *
 * Purpose:
 *
 *    Portable Device driver for serial lines with DMA assists.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 *
 * NOTE: The naming of this file is topsy-turvy: this really is the "c" one..
 */

#include <mmlite.h>
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>

/* Turn this on to get statistics about latencies
 */
#define _STATISTICS 0
#define _PRINT_STATS 0
#if _STATISTICS
#include <stdio.h>
#include <tchar.h>
#endif

/*
 * Byte queue data type.
 *
 * The count of bytes in the queue is computed by using the Head and Tail
 * values.  Because there are Size+1 "states" (possible different counts
 * including empty), the maximum number of bytes that a queue can hold is
 * Size-1.
 *
 * The exception to this rule is that when a queue is initialized, Tail is
 * set to -1.  This means that a queue can be initialized "full", as is done
 * in WriteAt().
 */
typedef struct {
    CONDITION WaitPoint;
    MUTEX Lock;
    PUINT8 pBuffer;
    volatile INT Head, Tail;
    UINT Size;
    UINT Drops;
#if _STATISTICS
    TIME WakeupTime;
    volatile BOOL Signalled;
    UINT32 MaxWakeup, MinWakeup;
    UINT32 AvgWakeup, NWakeups;
    UINT32 NSleepups;
    UINT32 BytesMoved;
    UINT32 NInterrupts;
    UINT32 NRescheds;
    UINT32 PrintMe;
#endif
} BYTE_QUEUE;

/* Is the queue empty ?
 */
#define CQueueEmpty(_q_) \
    ((_q_).Head == (_q_).Tail)

/* How many bytes in the queue ?
 * Non-opaque, but interrupt-safe eval
 */
#define CQueueCountEval(_h_,_t_,_s_) \
    (((INT)(_h_) > (INT)(_t_)) ? \
     ((INT)(_h_) - (INT)(_t_)) : \
     ((_s_) - ((INT)(_t_) - (INT)(_h_))))

/* How many bytes in the queue ?
 * Opaque, but beware of multiple evals
 */
#define CQueueCount(_q_) \
    CQueueCountEval(_q_.Head,_q_.Tail,_q_.Size)

/* Initialize a byte queue
 */
#define CQueueInit(_q_) \
  { \
    Condition_Init(& (_q_).WaitPoint); \
    Mutex_Init(& (_q_).Lock); \
    (_q_).Tail = (_q_).Head = -1; \
    (_q_).Drops = 0; \
  }

/*
 * Concrete class definition for a serial line device
 */

typedef struct CSerialDevice {
    struct CDevice CDevice;
    /* Our own state
     */
    IO_PORT PortBase;
    UINT Reset;
    UINT Errors;
    UINT BaudRate, LineParams;
    TIME RecvTimeout, XmitTimeout;
    BYTE_QUEUE InQue, OutQue;
} CSerialDevice, *PCSerialDevice;


/* Forward refs
 */
INT  MCT SerialInitialize(PCSerialDevice this);


/* Read SIZE bytes from serial line.
 * Returns whats available, if anything.
 * Else waits until at least one char is in.
 * Code is safe against multiple readers (Mutex),
 * and against interrupts (one-party update rule).
 */
SCODE MCT SerialReadAt (PCSerialDevice this, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    UINT Avail;
    SCODE sc;
#if _STATISTICS
    BOOL Slept = FALSE;
#endif

    UnusedParameter(Position);

    DBGME(0,_tprintf(_T("SerialReadAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeRead));

    Mutex_Lock( &this->InQue.Lock );

    while (CQueueEmpty(this->InQue)) {
#if _STATISTICS
        this->InQue.Signalled = FALSE;
        sc = Condition_Wait( &this->InQue.WaitPoint,
                             &this->InQue.Lock,
                             &this->RecvTimeout);
        this->InQue.NSleepups++;
        Slept = TRUE;
#else
        sc = Condition_Wait( &this->InQue.WaitPoint,
                             &this->InQue.Lock,
                             &this->RecvTimeout);
#endif
        if (sc == E_TIMED_OUT)
            goto Done;
    }

    /* How much can we get ?
     */
    Avail = this->InQue.Head; /* cache against interrupts */
    Avail = CQueueCountEval(Avail,this->InQue.Tail,this->InQue.Size);

    Size = (Size > Avail) ? Avail : Size;

    *pSizeRead = Size;

#if _STATISTICS
    this->InQue.BytesMoved += Size;
    if (Slept)
    {
        TIME Now;
        BOOL chgd = FALSE;
        UINT32 delta;

        Now = CurrentTime();
        Now = TimeSubtract(Now,this->InQue.WakeupTime);
        delta = Uint64ToUint32(Now);
        if (delta > this->InQue.MaxWakeup) {
            this->InQue.MaxWakeup = delta;
            chgd = TRUE;
        }
        if (delta < this->InQue.MinWakeup) {
            this->InQue.MinWakeup = delta;
            chgd = TRUE;
        }
        this->InQue.AvgWakeup = 
            (this->InQue.AvgWakeup * this->InQue.NWakeups + delta) /
                (this->InQue.NWakeups);
#if _PRINT_STATS
        if (chgd || ((++this->InQue.PrintMe & 0x7ff) == 0)) {
            _tprintf(_T("{Wr %d: %d %d %d}"), 
                     this->InQue.NWakeups, this->InQue.MinWakeup,
                     this->InQue.MaxWakeup, this->InQue.AvgWakeup);
            _tprintf("[%d %d %d]",this->InQue.BytesMoved,
                     this->InQue.NInterrupts, this->InQue.NSleepups);
        }
#endif
    }
#endif

    /* copy data, first part (towards the end of buffer)
     */
    Avail = this->InQue.Size - 1 - this->InQue.Tail;
    if (Avail > Size) 
        Avail = Size;
    if (Avail) {
        memcpy( pBuffer,
                & this->InQue.pBuffer[ this->InQue.Tail + 1],
                Avail);
        pBuffer += Avail;
        this->InQue.Tail += Avail;
        Size -= Avail;
    }

    /* copy data, second part (start of buf)
     */
    if (Size) {
        memcpy( pBuffer,
                this->InQue.pBuffer,
                Size );
        this->InQue.Tail = Size - 1;
    }

    /* Tail ptr updated already, done.
     */
    sc = NO_ERROR;
 Done:
    Mutex_Unlock( &this->InQue.Lock );

    return sc;
}

/* Write SIZE bytes to serial line.
 * Does not return until all bytes are sent.
 * Code is safe against multiple writers (Mutex),
 * and against interrupts (one-party update rule).
 */
SCODE MCT SerialWriteAt (PCSerialDevice this, UINT64 Position, const BYTE *pBuffer,
                            UINT Size, PUINT pSizeWritten)
{
    UINT8 c = 0;
#if _STATISTICS
    BOOL Slept = FALSE;
#endif
    SCODE sc = NO_ERROR;

    UnusedParameter(Position);

    DBGME(0,_tprintf(_T("SerialWriteAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeWritten));

    /* Get ahold of send capability.  If someone else is
     * currently writing we wait here.
     */
    Mutex_Lock( &this->OutQue.Lock );

    /* Backoff if a reset has been requested
     */
    if (this->Reset) {
        Size = 0;
        sc = E_RESTART;
        goto Done;
    }

    /* Set data pointer. No need for copies
     */
    this->OutQue.pBuffer = (PUINT8) pBuffer;

    /* Set pointers into buffer
     */
    this->OutQue.Tail = -1;
    this->OutQue.Size = Size;
    this->OutQue.Head = Size - 1;

    /* Select which version of the out-data-pump is desired
     * NB: Usually the DMA version is a worse performer, it
     * cannot pipeline xfers at all.  Non-dma can post the 
     * first byte and leave without waiting.  Adding an extra
     * data copy to the DMA version (software FIFO-like) would
     * make them even, but.
     */
#if !defined(USE_XMIT_DMA)
    /* Non-DMA version.  
     */

    /* Get the transmission process started.
     *
     * There are two schemes used for xmit interrupts.
     * In one case, the chip interrupts any time its
     * xmit buffer is empty (including the idle state).
     * In the other case, an interrupt is generated only on
     * successful transmission of one (or the last) byte.
     */

#if SER_INTERRUPTS_ON_XMIT_EMPTY

    /* Send as many characters as we can now.  Enable interrupts if necessary
     * to finish the job.
     */
    while (SER_XMIT_BUFFER_READY(this->PortBase) && !CQueueEmpty(this->OutQue)) {
        c = this->OutQue.pBuffer[++this->OutQue.Tail];
        SER_WRITE_XMIT_BUFFER(this->PortBase, c);
    }

    if (!CQueueEmpty(this->OutQue)) {
        SER_ENABLE_TRANSMIT_INTERRUPTS(this->PortBase);
    }

#else /* SER_INTERRUPTS_ON_XMIT_EMPTY not */

    /* Must send the first byte out. Xmit interrupts are
     * always on.
     */
    if (SER_XMIT_BUFFER_READY(this->PortBase)) {
        c = this->OutQue.pBuffer[++this->OutQue.Tail];
        SER_WRITE_XMIT_BUFFER(this->PortBase, c);
    }

#endif /* SER_INTERRUPTS_ON_XMIT_EMPTY */
    /* Macros for common code below
     */
#define XMITTED_ALL_DATA \
    CQueueEmpty(this->OutQue)
#define ADJUST_XMIT_COUNT(Size) \
  { \
    Size = Size - 1 - this->InQue.Tail; \
  }
    /* End of non-DMA version
     */
#else /* USE_XMIT_DMA version */
    /* DMA version.
     */
    c++; /* quiet warning :-)) */

    /* Get the transmission process started.
     * Set data pointer. No need for copies
     */
    SER_START_XMIT_DMA(this->PortBase,pBuffer,Size);

    /* Macros for common code below
     */
#define XMITTED_ALL_DATA \
    SER_XMIT_DMA_COMPLETE(this->PortBase)
#define ADJUST_XMIT_COUNT(Size) \
  { \
      UINT LeftToXmit; \
      SER_STOP_XMIT_DMA(this->PortBase,LeftToXmit); \
      Size = Size - LeftToXmit; \
  }
    /* End of DMA verison
     */
#endif /* back to common code */

    /* Sleep with device held.
     *
     * An explicit wakeup on the condvar is required to break this.
     */
    while (!XMITTED_ALL_DATA) {
#if _STATISTICS
        this->OutQue.Signalled = FALSE;
        sc = Condition_Wait( &this->OutQue.WaitPoint,
                             NULL,
                             &this->XmitTimeout );
        this->OutQue.NSleepups++;
        Slept = TRUE;
#else
        sc = Condition_Wait( &this->OutQue.WaitPoint,
                             NULL,
                             &this->XmitTimeout );
#endif
        if ((sc == E_TIMED_OUT) || (this->Reset)) {
            /* Account for what was sent, if anything
             */
            ADJUST_XMIT_COUNT(Size);
            break;
        }
        sc = NO_ERROR;
    }

#if _STATISTICS
    this->OutQue.BytesMoved += Size;
    if (Slept)
    {
        TIME Now;
        BOOL chgd = FALSE;
        UINT32 delta;

        Now = CurrentTime();
        Now = TimeSubtract(Now,this->OutQue.WakeupTime);
        delta = Uint64ToUint32(Now);
        if (delta > this->OutQue.MaxWakeup) {
            this->OutQue.MaxWakeup = delta;
            chgd = TRUE;
        }
        if (delta < this->OutQue.MinWakeup) {
            this->OutQue.MinWakeup = delta;
            chgd = TRUE;
        }
        this->OutQue.AvgWakeup = 
            (this->OutQue.AvgWakeup * this->OutQue.NWakeups + delta) /
                (this->OutQue.NWakeups);
#if _PRINT_STATS
        if (chgd || ((++this->OutQue.PrintMe & 0x7ff) == 0)) {
            _tprintf(_T("{Ww %d: %d %d %d}"), 
                     this->OutQue.NWakeups, this->OutQue.MinWakeup,
                     this->OutQue.MaxWakeup, this->OutQue.AvgWakeup);
            _tprintf("[%d %d %d]",this->OutQue.BytesMoved,
                     this->OutQue.NInterrupts, this->OutQue.NSleepups);
        }
#endif
    }
#endif

 Done:
    Mutex_Unlock( &this->OutQue.Lock );

    if (pSizeWritten)
        *pSizeWritten = Size;
    return sc;
}

/* Interrupt service routine for serial line.
 * Acts as a DMA engine of sorts, pulling bytes
 * in and out of the queues as appropriate.
 */
BOOL MCT SerialIsr(PCSerialDevice this, PBOOL pNotMyInterrupt)
{
    INT Avail; /* signed !! */
    BOOL Wakeup = FALSE, Signal;
    UINT Intr;
    IO_PORT Addr = this->PortBase;

    /* Look at the cause of interrupt
     */
    SER_READ_INTR_PORT(Intr,Addr);

#ifdef SER_SPURIOUS_INTERRUPT
    if (SER_SPURIOUS_INTERRUPT(Intr)) {
        if (pNotMyInterrupt)
            *pNotMyInterrupt = TRUE;
        return FALSE;
    }
#else
    /* Default: yes its ours */
    UnusedParameter(pNotMyInterrupt);
#endif

    /* Check for errors
     */
    if (SER_ERRORS(Intr,Addr)) {
        this->Errors++;
        SER_CLEAR_ERRORS(Intr,Addr);
    }

    /* Always check if there's anything to receive.
     * Only signal the receiver if the queue was empty (optim).
     */
    Signal = CQueueEmpty(this->InQue);

#if defined(USE_RECV_DMA)
    /* DMA version
     */
    if (SER_RECV_DMA_INTERRUPT(Intr,Addr)) {

        /* Did we get a character (and then nothing --> timeout)
         */
        if (SER_RECV_DMA_TIMEOUT(Intr,Addr)) {

            /* Update sw Head pointer
             */
            SER_RECV_DMA_COUNT(Addr,Avail);

            /* Did we just fill in the last position in the buffer
             * and therefore just reloaded the DMA engine ?
             */
            if (Avail == (INT) this->InQue.Size)
                Avail = 0;

            this->InQue.Head = this->InQue.Size - Avail - 1;

            /* See if we have room for more.
               BUGBUG Must handle overruns!!
             */

            /* Restart the timeout thing
             * NB: Should trigger *after* recpt of next char
             */
            SER_START_RECV_DMA_TIMEOUT(Addr);
        }

        /* Watch out for overruns
         */
        if (SER_RECV_DMA_COMPLETE(Intr,Addr)) {
            SER_START_RECV_DMA(Addr,this->InQue.pBuffer,this->InQue.Size);
        }
#else /* dont USE_RECV_DMA */
    if (SER_ISR_RECEIVED_DATA(Intr,Addr)) {

        do {
            /* Pick the byte up
             */
            UINT8 c;
            c = SER_READ_RECEIVE_BUFFER(Addr);

            /* See if we have room for it.
             * [just a fast&funny !CQueueFull]
             */
            Avail = this->InQue.Tail - this->InQue.Head;
            if ((Avail != 1) && (Avail != (1 - (INT)this->InQue.Size))) {

                /* Where do we put it. Update ptr.
                 */
                Avail = this->InQue.Head + 1;
                if (Avail == (INT)this->InQue.Size)
                    Avail = 0;
                this->InQue.pBuffer[Avail] = c;
                this->InQue.Head = Avail;

            } else
                this->InQue.Drops++;

        } while (SER_RECV_BUFFER_READY(Addr));

#endif /* USE_RECV_DMA */
        /* Back to common code
         */
#if _STATISTICS
        this->InQue.NInterrupts++;
#endif

        if (Signal) {
#if _STATISTICS
            if (!this->InQue.Signalled) {
                this->InQue.Signalled = TRUE;
                this->InQue.WakeupTime = CurrentTime();
            }
            this->InQue.NWakeups++;
            Wakeup = Condition_InterruptSignal( & this->InQue.WaitPoint );
            if (Wakeup) this->InQue.NRescheds++;
#endif
            Wakeup = Condition_InterruptSignal( & this->InQue.WaitPoint );
        }

    }

#if !defined(USE_XMIT_DMA)
    /* Non-DMA version.  
     */
    if (SER_ISR_TRANSMITTED_DATA(Intr,Addr)) {

        Signal = FALSE;
        do {
            /* Have we got another char to send ?
             */
            UINT8 c;
            if (!CQueueEmpty(this->OutQue)) {

                if (this->Reset) {
                    /* Reset is pending.  Dump the rest of the buffer so
                     * the reset can proceed.
                     */
                    this->OutQue.Tail = this->OutQue.Head;
                    Signal = TRUE;

                } else {

                    /* Pick up next byte, update ptrs
                     */
                    Avail = this->OutQue.Tail + 1;
                    if (Avail == (INT)this->OutQue.Size)
                        Avail = 0;

                    c = this->OutQue.pBuffer[Avail];
                    this->OutQue.Tail = Avail;

                    /* Send char out
                     */
                    SER_WRITE_XMIT_BUFFER(Addr, c);
                }

            } else {

#if SER_INTERRUPTS_ON_XMIT_EMPTY
                /* Time to turn ints off
                 */
                SER_DISABLE_TRANSMIT_INTERRUPTS(Addr);        
#endif
                /* We just sent the last character, wakeup writer
                 */
                Signal = TRUE;
                break;
            }

        } while (SER_XMIT_BUFFER_READY(Addr));

        /* If the queue is now empty do signal regardless.
         * Meaning.. if the macro above indicates no more
         * room but we infact posted the last one we had.
         */
        //        Signal |= (CQueueEmpty(this->OutQue));

#else /* USE_XMIT_DMA */
    /* DMA version
     */
    if (SER_ISR_XMIT_COMPLETE(Intr,Addr)) {

        /* NB: This assumes the DMA engine can handle arbitrarily large
         * buffers. If that is not the case this code needs refininment,
         * which is why we set the OutQue.pBuffer and size at the start.
         */
        SER_STOP_XMIT_DMA(Addr,Avail);
        Signal = TRUE;

#endif /* USE_XMIT_DMA */
        /* Back to common code 
         */
#if _STATISTICS
        this->OutQue.NInterrupts++;
#endif
        if (Signal) {
#if _STATISTICS
            if (!this->OutQue.Signalled) {
                this->OutQue.WakeupTime = CurrentTime();
                this->OutQue.Signalled = TRUE;
            }
            this->OutQue.NWakeups++;
            Signal = Condition_InterruptSignal( & this->OutQue.WaitPoint );
            if (Signal) this->OutQue.NRescheds++;
            Wakeup |= Signal;
#else
            Wakeup |= Condition_InterruptSignal( & this->OutQue.WaitPoint );
#endif
        }
    }

    SER_ACK_INTERRUPTS(Addr);
    return Wakeup;
}


/* SetSize method on serial line
 */
SCODE MCT SerialSetSize (PCSerialDevice this, UINT64 Size)
{
    /* This is extremely wacky, but..
     */
    UINT32 *p = (UINT*)&Size;

    switch (*p) {
    case 0: /* set recv timeout */
        Int32ToInt64(this->RecvTimeout,TIME_RELATIVE((INT32)p[1]));
        break;
    case 1: /* set xmit timeout */
        Int32ToInt64(this->XmitTimeout,TIME_RELATIVE((INT32)p[1]));
        break;
    default: /* set baudrate and line parameters at once */
        this->BaudRate = p[0];
        this->LineParams = p[1];
        if (0 != SerialInitialize(this))
            return E_INVALID_PARAMETER;
    }
    return NO_ERROR;
}


/* (Re)Initialization code.
 * Resets both queues and hardware.
 * Leaves interrupts enabled, chip functional.
 */
INT  MCT SerialInitialize(PCSerialDevice this)
{
    IO_PORT Addr = this->PortBase;
    UINT8 c;

    DBGME(0,_tprintf(_T("SerialInit(%x)\n"), (ADDRESS) this));

    /* Flag that reset is pending.  This will abort any transmitting that is
     * in progress via WriteAt.
     */
    AtomicInc(&this->Reset);

    /* Wakeup any writer
     */
    Condition_Signal(&this->OutQue.WaitPoint);

    /* Wait for any transmitting threads to get out.  This gives us exclusive
     * access to the hardware.
     */
    Mutex_Lock(&this->OutQue.Lock);

    /* Turn off receiver, transmitter and interrupts.
     */
    SER_DISABLE_INTERRUPTS(Addr);

    /* Reset Queues.  If we can take the InQue lock, then reset that queue.
     * Otherwise we can't guarantee that we won't block forever.
     */
    if (Mutex_TryLock(&this->InQue.Lock)) {
        this->InQue.Tail = this->InQue.Head = -1;
        Mutex_Unlock(&this->InQue.Lock);
    }
    Condition_Signal(&this->InQue.WaitPoint);

    /* Declare out queue empty
     */
    this->OutQue.Tail = this->OutQue.Head = -1;

    /* Set baudrate
     */
    SER_SET_BAUD_RATE(Addr, this->BaudRate);

    /* Set line parameters
     */
    SER_SET_LINE_PARAMS(Addr,this->LineParams);

    /* Deassert DTR, RTS for a while
     */
    SER_SET_MODEM_BITS(Addr, 0 );

#if 0
    /* This turned out to be trouble (dont remember how though..). Umbug
     */
    Delay(500*1000);/* at least 150msec for MPlus mice */
#endif

    /* Read out any input leftovers
     */
    c = SER_READ_RECEIVE_BUFFER(Addr);

#if defined(USE_RECV_DMA)
    /* Start recv DMA, enable its overrun interrupt.
     */
    SER_START_RECV_DMA(Addr,this->InQue.pBuffer,this->InQue.Size);
#else
    /* Enable recv interrupts.
     */
    SER_ENABLE_RECEIVE_INTERRUPTS(Addr);
#endif

    /* Assert DTR, RTS.  On NS8250&co enable interrupt gate.
     */
    SER_SET_MODEM_BITS(Addr, SER_DTR | SER_RTS | SER_INTGATE);

#if !defined(USE_XMIT_DMA) && (SER_INTERRUPTS_ON_XMIT_EMPTY==0)
    /* Enable xmit-done interrupts.
     */
    SER_ENABLE_TRANSMIT_INTERRUPTS(Addr);

#endif /* !USE_XMIT_DMA && !SER_INTERRUPTS_ON_XMIT_EMPTY */

    /* We're done.
     */
    AtomicDec(&this->Reset);

    /* Let other threads back in.
     */
    Mutex_Unlock(&this->OutQue.Lock);

    return 0;
}

struct CDeviceVtbl SerialDriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)) SerialReadAt,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Position, const BYTE *pBuffer, UINT Size, PUINT pSizeWritten)) SerialWriteAt,
    (SCODE     (MCT *)(PCDEVICE pThis, UINT64 Size)) SerialSetSize,
    NULL,
    (BOOL      (MCT *)(PCDEVICE pThis, PBOOL pWakeup)) SerialIsr,
    NULL,
    (INT       (MCT *)(PCDEVICE pThis)) SerialInitialize,
    NULL
};

/* Constructor
 */
void
CSerialDeviceInit (PCSerialDevice this, IO_PORT Base, UINT Rate, UINT *pIrq )
{
    PIHEAP pH = CurrentHeap();

    this->Reset = 0;
    this->PortBase = Base;
    this->BaudRate = Rate;
    this->LineParams = SER_DEFAULT_LINE_PARAMS;
    this->RecvTimeout = TIME_FOREVER;
    this->XmitTimeout = TIME_FOREVER;
#ifndef SERIAL_INBUFFER_SIZE
#define SERIAL_INBUFFER_SIZE 512//128
#endif
    this->InQue.Size = this->OutQue.Size = SERIAL_INBUFFER_SIZE;
    CQueueInit( this->InQue );
    CQueueInit( this->OutQue );
    this->InQue.pBuffer = (PUINT8) pH->v->Alloc ( pH, HEAP_ZERO_MEMORY, this->InQue.Size, 0 );
    this->CDevice.v = &SerialDriverVtable;
    this->CDevice.RefCnt = 1;

    SER_SPECIAL_INIT(Base);

#if _STATISTICS
    this->InQue.MinWakeup = this->OutQue.MinWakeup = ~0;
#endif
}


/* Exported constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE SerialNew( PUINT pPortBase, PUINT pIrq, PUINT pBaudRate)
{
    PCDEVICE pSer;
    PIHEAP pH = CurrentHeap();

    pSer = (PCDEVICE) pH->v->Alloc( pH, HEAP_ZERO_MEMORY, sizeof(CSerialDevice), 0 );
    if (pSer)
        CSerialDeviceInit( (PCSerialDevice)pSer, (IO_PORT)*pPortBase, *pBaudRate, pIrq );

    return pSer;
}

/* ** 
*/
SCODE __cdecl SerialConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        PTR NoIdeaWhatsThisFor,
        PTR NoIdeaWhatsThisForNeither,
        UINT BaudRate,
        ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("SerialConstructor(%x %s %x %x %x %d)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk, BaudRate));

    pUnk = (PIUNKNOWN) SerialNew(&RegisterBase,
                                 &Interrupt,
                                 &BaudRate);

    DBGME(0,_tprintf(_T("SerialConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("SerialConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl SerialDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) SerialConstructor
};

static struct _DRIVERCOB SerialDriverCobObject = { &SerialDriverCobVtable, 1 };

PIUNKNOWN SerialCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &SerialDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &SerialDriverVtable, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    /* Null the thread loop, to save a useless thread
     */
    SerialDriverVtable.ThreadLoop = NULL;

    return (PIUNKNOWN) &SerialDriverCobObject;
}
