/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    eb63usart.h
 *
 * Purpose:
 *
 *    Portable Device Driver for serial lines, Atmel m63200 USART customization
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#define DEVICE_ORDER BIG_ENDIAN
#include <mmlite.h>
#include <mmhal.h>

/* Register Offsets
 */
#define USART_TAG(_p_)          ((IO_PORT)((_p_)+4*0))
#define USART_CONTROL(_p_)      ((IO_PORT)((_p_)+4*1))
#define USART_INTR_ENABLE(_p_)  ((IO_PORT)((_p_)+4*2))
#define USART_INTR_DISABLE(_p_) ((IO_PORT)((_p_)+4*3))
#define USART_INTR_MASK(_p_)    ((IO_PORT)((_p_)+4*4)) //10
#define USART_STATUS(_p_)       ((IO_PORT)((_p_)+4*5))
#define USART_RX_DATA(_p_)      ((IO_PORT)((_p_)+4*6))
#define USART_TX_DATA(_p_)      ((IO_PORT)((_p_)+4*7))
#define USART_BAUD(_p_)         ((IO_PORT)((_p_)+4*8)) //20

/*   Per-channel Registers    */

/* Control Register
 */
# define                USC_RESET       0x00000001
# define                USC_RSTRX       0x00000004
# define                USC_RSTTX       0x00000008
# define                USC_RXEN        0x00000010
# define                USC_RXDIS       0x00000020
# define                USC_TXEN        0x00000040
# define                USC_TXDIS       0x00000080
# define                USC_RSTSTA      0x00000100
# define                USC_STTBRK      0x00000200
# define                USC_STPBRK      0x00000400
# define                USC_STTO        0x00000800
# define                USC_CLK_SENDA   0x00010000
# define                USC_BPC_9       0x00020000
# define                USC_CLKO        0x00040000
# define                USC_EVEN        0x00000000
# define                USC_ODD         0x00080000
# define                USC_SPACE       0x00100000 /* forced 0 */
# define                USC_MARK        0x00180000 /* forced 1 */
# define                USC_NONE        0x00200000
# define                USC_MDROP       0x00300000
# define                USC_BPC_5       0x00000000
# define                USC_BPC_6       0x00400000
# define                USC_BPC_7       0x00800000
# define                USC_BPC_8       0x00c00000
# define                USC_CLKDIV      0x0f000000
# define                USC_CLKDIV_1    0x00000000
# define                USC_CLKDIV_2    0x01000000
# define                USC_CLKDIV_4    0x02000000
# define                USC_CLKDIV_8    0x03000000
# define                USC_CLKDIV_16   0x04000000
# define                USC_CLKDIV_32   0x05000000
# define                USC_CLKDIV_64   0x06000000
# define                USC_CLKDIV_128  0x07000000
# define                USC_CLKDIV_EXT  0x08000000
# define                USC_1STOP       0x00000000
# define                USC_1_5STOP     0x10000000
# define                USC_2STOP       0x20000000
# define                USC_ECHO        0x40000000 /* rx->tx, tx disabled */
# define                USC_LOOPBACK    0x80000000 /* tx->rx, rx/tx disabled */
# define                USC_ECHO2       0xc0000000 /* rx->tx, rx disabled */

/* Interrupt related registers
 */
# define                USI_RXRDY       0x00000001
# define                USI_TXRDY       0x00000002
# define                USI_RXBRK       0x00000004
# define                USI_ENDRX       0x00000008
# define                USI_ENDTX       0x00000010
# define                USI_OVRE        0x00000020
# define                USI_FRAME       0x00000040
# define                USI_PARE        0x00000080
# define                USI_TIMEOUT     0x00000100
# define                USI_TXEMPTY     0x00000200

/* Defines to customize the common code
 */
#define SER_READ_INTR_PORT(_i_,_p_) \
  { \
         _i_ = IOSpaceReadInt32(USART_STATUS(_p_)) & \
               IOSpaceReadInt32(USART_INTR_MASK(_p_)); \
         _i_ = _bs32(_i_); \
  }
#define SER_RECEIVE_INTERRUPT(_i_) \
         ((_i_) & (USI_RXRDY|USI_RXBRK|USI_OVRE|USI_FRAME|USI_PARE))
#define SER_TRANSMITTED_INTERRUPT(_i_) \
         ((_i_) & USI_TXRDY)

#define SER_ENABLE_RECEIVE_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_INTR_ENABLE(_p_), \
                _bs32(USI_RXRDY|USI_RXBRK|USI_OVRE|USI_FRAME|USI_PARE))
#define SER_ENABLE_TRANSMIT_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_INTR_ENABLE(_p_),_bs32(USI_TXRDY))
#define SER_DISABLE_TRANSMIT_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_INTR_DISABLE(_p_),_bs32(USI_TXRDY))
#define SER_DISABLE_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_INTR_DISABLE(_p_),~0)

#define SER_ACK_INTERRUPTS(_p_)

#define SER_ISR_RECEIVED_DATA(_i_,_p_) \
         SER_RECEIVE_INTERRUPT(_i_)
#define SER_ISR_TRANSMITTED_DATA(_i_,_p_) \
         (SER_TRANSMITTED_INTERRUPT(_i_))

#define SER_XMIT_BUFFER_READY(_p_) \
         (IOSpaceReadInt32(USART_STATUS(_p_)) & _bs32(USI_TXRDY))
#define SER_RECV_BUFFER_READY(_p_) \
         (IOSpaceReadInt32(USART_STATUS(_p_)) & _bs32(USI_RXRDY))

#define SER_READ_RECEIVE_BUFFER(_p_) \
         (UINT8)(IOSpaceReadInt32(USART_RX_DATA(_p_)))
#define SER_WRITE_XMIT_BUFFER(_p_,_c_) \
         IOSpaceWriteInt32(USART_TX_DATA(_p_),_c_)

/* BaudRate better be slower than faster, roundup 
 * NB: ASYNC, else no x16;
 */
#define SER_SET_BAUD_RATE(_p_,_b_) \
  { \
         struct MACHINE_INFO mi; \
         UINT _divisor_; \
         PIPROCESS pPrc = CurrentProcess(); \
         (void) pPrc->v->MachineInfo( pPrc, MACHINE_KIND_GENERIC, &mi ); \
         _divisor_ = Uint64ToUint32(mi.ProcessorSpeed); \
         _divisor_ = (_divisor_/8)/(_b_); \
         _divisor_ = (_divisor_/2) + (_divisor_&1); \
         _divisor_ = 0x29; /*divbugfix*/ \
         IOSpaceWriteInt32(USART_BAUD(_p_),_bs32(_divisor_)); \
  }

/* Line parameter encoding: stright from MODE register.
 * Default is 8-bit data, 1-stop bit, no parity.
 */
#define SER_DEFAULT_LINE_PARAMS \
  (USC_CLKDIV_4|USC_BPC_8|USC_NONE|USC_1STOP)

/* Logic: disable txrx, reset. Set mode.
 * Enable txrx and clear errors.
 */
#define SER_SET_LINE_PARAMS(_p_,_l_) \
  { \
         IOSpaceWriteInt32(USART_CONTROL(_p_),_bs32(USC_RXDIS|USC_TXDIS)); \
         IOSpaceWriteInt32(USART_CONTROL(_p_),_bs32(USC_RSTRX|USC_RSTTX)); \
         IOSpaceWriteInt32(USART_CONTROL(_p_),_bs32((_l_)|USC_RXEN|USC_TXEN|USC_RSTSTA|USC_STPBRK)); \
  }

/* We dont do modem control
 */
#define SER_SET_MODEM_BITS(_p_,_b_)

/* Transmitter interrupts when empty
 * [Or, thats the way we want to use it]
 */
#define SER_INTERRUPTS_ON_XMIT_EMPTY 1

/* Error handling, we got to
 */
#define SER_ERRORS(_i_,_p_) \
         ((_i_) & (USI_RXBRK|USI_OVRE|USI_FRAME|USI_PARE))
#define SER_CLEAR_ERRORS(_i_,_p_) \
  { \
         UINT32 _x_; \
         _x_ = IOSpaceReadInt32(USART_CONTROL(_p_)) | _bs32(USC_RSTSTA); \
         IOSpaceWriteInt32(USART_CONTROL(_p_),_x_); \
         _i_ = IOSpaceReadInt32(USART_STATUS(_p_)); \
         _i_ = _bs32(_i_); \
  }

/* Must enable clock to the part, and enable the PIO pins.
 * Find out which USART index it is, to select the right bits.
 * Also, clearup DMAs
 */
#define SER_SPECIAL_INIT(_p_) 

/* HACKALERT needs VM */
#if defined(mips)
#define _v2p(v) ((ADDRESS)(v) & ~0x80000000)
#else
#define _v2p(v) ((ADDRESS)(v))
#endif

/* BUGBUG later */
#define SER_TEST_FOR_MORE_INTERRUPTS(_p_,_i_,_l_)

/* Definitions for DMA usage
 */
#define SER_START_XMIT_DMA(_p_,_b_,_s_) error

#define SER_STOP_XMIT_DMA(_p_,_s_) error

#define SER_ISR_XMIT_COMPLETE(_i_,_p_) error

#define SER_XMIT_DMA_COMPLETE(_p_) error

/* Device driver proper
 * Dont use DMA at all
 */
#define SERIAL_INBUFFER_SIZE 512
#include "serial.h"
