/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    osart.c
 *
 * Purpose:
 *
 *    Portable Device Driver for serial lines, National USART customization
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */


/* Register Offsets
 */
#define USART_TBUF(_p_)         ((IO_PORT)((_p_)+4*0))
#define USART_RBUF(_p_)         ((IO_PORT)((_p_)+4*1))
#define USART_ICTRL(_p_)        ((IO_PORT)((_p_)+4*2))
#define USART_STAT(_p_)         ((IO_PORT)((_p_)+4*3))
#define USART_FRS(_p_)          ((IO_PORT)((_p_)+4*4))
#define USART_MDSL1(_p_)        ((IO_PORT)((_p_)+4*5))
#define USART_BAUD(_p_)         ((IO_PORT)((_p_)+4*6))
#define USART_PSR(_p_)          ((IO_PORT)((_p_)+4*7))
#define USART_OVR(_p_)          ((IO_PORT)((_p_)+4*8))
#define USART_MDSL2(_p_)        ((IO_PORT)((_p_)+4*9))
#define USART_SPOS(_p_)         ((IO_PORT)((_p_)+4*10))

#define AT_USART_0     ((ADDRESS)0xb8040000)
#define AT_USARTNO(_addr_) (((ADDRESS)_addr_ == AT_USART_0) ? 0 : 1)

/* Interrupt and Control Register
 */
# define                USI_TBE        0x01 /* ro tx buffer empty */
# define                USI_RBF        0x02 /* ro rx buffer full */
# define                USI_DCTS       0x04 /* ro delta on CTS_N */
# define                USI_CTS        0x08 /* ro ~CTS_N (==CTS) */
# define                USI_EFCI       0x10 /* rw enbable flow int */
# define                USI_ETI        0x20 /* rw enable tx int */
# define                USI_ERI        0x40 /* rw enable rx int */
# define                USI_EEI        0x80 /* rw enable err int */

/* Status Register
 */
# define                USS_PE         0x01 /* parity error */
# define                USS_FE         0x02 /* framing error */
# define                USS_DOE        0x04 /* rx overrun */
# define                USS_ERR        0x08 /* any 0x3 set */
# define                USS_BKD        0x10 /* break recvd */
# define                USS_BIT_9      0x20 /* recvd 9th bit */
# define                USS_XMIP       0x40 /* xmit in progress */
# define                USS_xx         0x80

/* Frame Select Register
 */
# define                USF_CHAR       0x03 /* mask for length */
# define                USF_CHAR_8     0x00
# define                USF_CHAR_7     0x01
# define                USF_CHAR_9     0x02
# define                USF_LOOP_9     0x03 /* loopback, 9 bits */
# define                USF_STOP_1     0x00
# define                USF_STOP_2     0x04
# define                USF_BIT_9      0x08 /* the 9th bit, xmt */
# define                USF_NO_PARITY  0x00
# define                USF_PARITY     0x40
# define                USF_ODD        0x00
# define                USF_EVEN       0x10
# define                USF_MARK_1     0x20
# define                USF_SPACE      0x30
# define                USF_xx         0x80

/* Mode Select Register 1
 */
# define                USM_ASYNC      0x00
# define                USM_SYNC       0x01
# define                USM_ATN        0x02 /* synch, attention mode */
# define                USM_BRK        0x04 /* force break */
# define                USM_CKS        0x08 /* synch, external clock */
# define                USM_ETD        0x10 /* enable tx DMA */
# define                USM_ERD        0x20 /* enable rx DMA */
# define                USM_FCE        0x40 /* flow-control enable */
# define                USM_RTS        0x80 /* RTS_N value */

/* Baud Rate Prescaler Register
 */
# define                USP_DIV_8_10   0x07 /* high part of baud */
# define                USP_PSC        0xf8 /* pre-scaler value */

/* Oversample Rate Register
 */
# define                USO_OVR_MASK   0x0f /* oversample rate, but.. */
# define                USO_OVR_16     0x00
# define                USO_OVR_7      0x07 /* and so on for 8-15 */
# define                USO_OVR_15     0x0f

/* Sample Position Register
 */
# define                USR_SAMP       0x0f /* values: 0..OVR-3 */


/* Defines to customize the common code
 */
#define SER_READ_INTR_PORT(_i_,_p_) \
  { \
         _i_ = IOSpaceReadInt32(USART_ICTRL(_p_)) | \
               (IOSpaceReadInt32(USART_STAT(_p_)) << 8); \
         /* Mask off disabled intrs */ \
         _i_ = (_i_ & ~(USI_TBE|USI_RBF)) | \
               (((_i_ & (USI_ETI|USI_ERI)) >> 5) & _i_); \
  }
#define SER_RECEIVE_INTERRUPT(_i_) \
         ((_i_) & (USI_RBF|((USS_ERR)<<8)))
#define SER_TRANSMITTED_INTERRUPT(_i_) \
         ((_i_) & USI_TBE)

/* Strictly speaking..cant write back bits 0..3 */
#define SER_ENABLE_RECEIVE_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_ICTRL(_p_), \
            IOSpaceReadInt32(USART_ICTRL(_p_)) | USI_ERI | USI_EEI)
#define SER_ENABLE_TRANSMIT_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_ICTRL(_p_), \
            IOSpaceReadInt32(USART_ICTRL(_p_)) | USI_ETI)
#define SER_DISABLE_TRANSMIT_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_ICTRL(_p_), \
            IOSpaceReadInt32(USART_ICTRL(_p_)) & ~USI_ETI)
#define SER_DISABLE_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(USART_ICTRL(_p_), 0)

#define SER_ACK_INTERRUPTS(_p_)

#define SER_ISR_RECEIVED_DATA(_i_,_p_) \
         SER_RECEIVE_INTERRUPT(_i_)
#define SER_ISR_TRANSMITTED_DATA(_i_,_p_) \
         (SER_TRANSMITTED_INTERRUPT(_i_))

#define SER_XMIT_BUFFER_READY(_p_) \
         (IOSpaceReadInt32(USART_ICTRL(_p_)) & USI_TBE)
#define SER_RECV_BUFFER_READY(_p_) \
         (IOSpaceReadInt32(USART_ICTRL(_p_)) & USI_RBF)

#define SER_READ_RECEIVE_BUFFER(_p_) \
         IOSpaceReadInt8(USART_RBUF(_p_))
#define SER_WRITE_XMIT_BUFFER(_p_,_c_) \
         IOSpaceWriteInt8(USART_TBUF(_p_),_c_)

/* BaudRate better be slower than faster, roundup 
 * BUGBUG Do it right..
 */
#define SER_SET_BAUD_RATE(_p_,_b_) \
  { \
     if (_b_ != 38400) DebugBreak(); \
     IOSpaceWriteInt32(USART_BAUD(_p_),4); \
     IOSpaceWriteInt32(USART_PSR(_p_),0xc<<3); \
  }

/* Line parameter encoding: stright from Frame Select Register.
 * Default is 8-bit data, 1-stop bit, no parity.
 */
#define SER_DEFAULT_LINE_PARAMS \
  (USF_CHAR_8|USF_STOP_1|USF_NO_PARITY)

/* Logic: disable txrx, reset. Set mode.
 * Enable txrx and clear errors. Or not.
 */
#define SER_SET_LINE_PARAMS(_p_,_l_) \
  { \
         IOSpaceWriteInt32(USART_FRS(_p_),(_l_)); \
  }

/* We dont do modem control
 */
#define SER_SET_MODEM_BITS(_p_,_b_)

/* Transmitter interrupts when empty
 * [Or, thats the way we want to use it]
 */
#define SER_INTERRUPTS_ON_XMIT_EMPTY 1

/* Error handling, we got to
 */
#define SER_ERRORS(_i_,_p_) \
         ((_i_) & ((USS_PE|USS_FE|USS_DOE|USS_BKD) << 8))
#define SER_CLEAR_ERRORS(_i_,_p_)

/* Must enable clock to the part, and enable the PIO pins.
 * Find out which USART index it is, to select the right bits.
 */
#ifdef _MSC_VER
#pragma warning(disable:4054)   /* cast from function ptr to data ptr (/Za) */
#endif
#define SER_SPECIAL_INIT(_p_) \
  { \
        AddDevice(this,(PTR)SerialIsr,0,*pIrq+1,NULL); \
  }

/* BUGBUG later */
#define SER_TEST_FOR_MORE_INTERRUPTS(_p_,_i_,_l_)

/* Definitions for DMA usage
 */
#define SER_START_XMIT_DMA(_p_,_b_,_s_) notyet

#define SER_STOP_XMIT_DMA(_p_,_s_) notyet

#define SER_ISR_XMIT_COMPLETE(_i_,_p_) notyet

#define SER_XMIT_DMA_COMPLETE(_p_) notyet

/* Device driver proper
 */
#include "serial.h"

