/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    sa.c
 *
 * Purpose:
 *
 *    Portable Device driver for serial lines, StrongARM 16550 customization.
 *
 * Author:
 *    Scott Corbin (scottc)
 *
 * --------------------------------------------------------------------------
 */

#define CONTROL0(addr)      ((IO_PORT)(addr + 0*4))
#define CONTROL1(addr)      ((IO_PORT)(addr + 1*4))
#define CONTROL2(addr)      ((IO_PORT)(addr + 2*4))
#define CONTROL3(addr)      ((IO_PORT)(addr + 3*4))
#define CONTROL4(addr)      ((IO_PORT)(addr + 4*4))
#define DATA(addr)          ((IO_PORT)(addr + 5*4))
#define STATUS0(addr)       ((IO_PORT)(addr + 7*4))
#define STATUS1(addr)       ((IO_PORT)(addr + 8*4))

/* Control register 0 */
#define PE          0x01    /* parity enable */
#define OES         0x02    /* odd/even parity select */
#define SBS         0x04    /* stop bit select */
#define DSS         0x08    /* data size select */
#define SCE         0x10    /* sample clock enable */
#define RCE         0x20    /* receive clock edge select */
#define TCE         0x40    /* transmit clock edge select */

/* Control registers 1 & 2 select the baud rate */

/* Control register 3 */
#define RXE         0x01    /* receiver enable */
#define TXE         0x02    /* transmitter enable */
#define BRK         0x04    /* break */
#define RIE         0x08    /* receive fifo interrupt enable */
#define TIE         0x10    /* transmit fifo interrupt enable */
#define LBM         0x20    /* loopback mode */

/* Data register, read access */
#define PRE         0x100   /* parity error */
#define FRE         0x200   /* framing error */
#define ROR         0x400   /* receiver overrun */

/* Status register 0 */
#define TFS         0x01    /* xmit fifo service request */
#define RFS         0x02    /* recv fifo service request */
#define RID         0x04    /* receiver idle */
#define RBB         0x08    /* receiver begin of break */
#define REB         0x10    /* receiver end of break */
#define EIF         0x20    /* error in fifo */
#define iMASK       0x3f

/* Status register 1 */
#define TBY         0x01    /* transmitter busy */
#define RNE         0x02    /* receive fifo not empty */
#define TNF         0x04    /* transmit fifo not full */
#define PER         0x08    /* parity error */
#define FER         0x10    /* framing error */
#define RER         0x20    /* receiver overrun */



/* Defines to customize the common code properly
 */
#define SER_READ_INTR_PORT(_i_,_p_) \
         _i_ = (IOSpaceReadInt32(STATUS0(_p_)) & iMASK)

#define SER_RECEIVE_INTERRUPT(_i_) \
         (_i_ & (RFS|RID))
#define SER_TRANSMITTED_INTERRUPT(_i_) \
         (_i_ & TFS)
#define SER_MODEM_INTERRUPT(_i_) \
         (0)
#define SER_ERROR_INTERRUPT(_i_) \
         (_i_ & EIF)

/* See pagg 11-119 for this
 * BUGBUG The EIF bit is set if ANY of the four bottom bytes are in error,
 * not just the bottom-est one.  Maybe they'll fix the chip someday :-))
 * For now, let the interrupt come again if it isnt on the bottom byte.
 * If they fix the hardware the code still works.
 */
#define SER_ERRORS(_i_,_p_) SER_ERROR_INTERRUPT(_i_)
#define SER_CLEAR_ERRORS(_i_,_p_) \
  {  \
         UINT _s1_; \
         do { \
                  _s1_ = IOSpaceReadInt32(STATUS1(_p_)); \
                  if (_s1_ & (PER|FER|RER)) \
                          (void) IOSpaceReadInt32(DATA(_p_)); \
         } while (_s1_ & (PER|FER|RER)); \
         /* Now refresh our idea of the status of things */ \
         SER_READ_INTR_PORT(_i_,_p_); \
  }

#define SER_ISR_RECEIVED_DATA(_i_,_p_) \
         SER_RECV_BUFFER_READY(_p_)
#define SER_ISR_TRANSMITTED_DATA(_i_,_p_) \
         (SER_TRANSMITTED_INTERRUPT(_i_) && \
          SER_XMIT_BUFFER_READY(_p_))

#define SER_ENABLE_RECEIVE_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(CONTROL3(_p_), RXE|TXE|RIE)
#define SER_ENABLE_TRANSMIT_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(CONTROL3(_p_), RXE|TXE|RIE|TIE)
#define SER_DISABLE_TRANSMIT_INTERRUPTS SER_ENABLE_RECEIVE_INTERRUPTS
#define SER_DISABLE_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(CONTROL3(_p_), 0)

#define SER_ACK_INTERRUPTS(_p_) \
         IOSpaceWriteInt32(STATUS0(_p_), RID|RBB|REB)

#define SER_XMIT_BUFFER_READY(_p_) \
         (IOSpaceReadInt32(STATUS1(_p_)) & TNF)
#define SER_RECV_BUFFER_READY(_p_) \
         (IOSpaceReadInt32(STATUS1(_p_)) & RNE)

#define SER_READ_RECEIVE_BUFFER(_p_) ((UINT8) IOSpaceReadInt32(DATA(_p_)))
#define SER_WRITE_XMIT_BUFFER(_p_,_c_) IOSpaceWriteInt32(DATA(_p_),_c_)

#define SER_SET_BAUD_RATE(_p_,_b_) \
  {  \
         UINT Divisor = (230400 / (_b_)) - 1; \
         \
         /* Set the divisor field */ \
         \
         IOSpaceWriteInt32(CONTROL1(_p_), ((Divisor >> 8) & 0xf)); \
         IOSpaceWriteInt32(CONTROL2(_p_), (Divisor & 0xff)); \
  }

#define SER_SET_LINE_PARAMS(_p_,_l_) \
         IOSpaceWriteInt32(CONTROL0(_p_), (UINT8)_l_)

/* Set LCR to 8-bit data, 1-stop bit, no parity.
 */
#define SER_DEFAULT_LINE_PARAMS DSS

/* BUGBUG These are implementation defined.
 */
#define SER_SET_MODEM_BITS(_p_,_b_)

#define SER_DTR
#define SER_RTS
#define SER_INTGATE

#define SER_INTERRUPTS_ON_XMIT_EMPTY 1


#define SER_SPECIAL_INIT(_p_)

/* Device driver proper
 */
#include "serial.h"
