/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    scn2681
 *
 * Purpose:
 *
 *    Portable Device Driver for serial lines, Signetics SCN2681 customization
 *    as used on IDT & Pulsar.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */


/* Register Offsets
 */
#define S2681_MODE_A(_p_)       ((IO_PORT)((_p_)+4*0))
#define S2681_STATUS_A(_p_)     ((IO_PORT)((_p_)+4*1))
#define S2681_BAUD_RATE_A(_p_)  ((IO_PORT)((_p_)+4*1))
#define S2681_CMD_A(_p_)        ((IO_PORT)((_p_)+4*2))
#define S2681_TXRX_A(_p_)       ((IO_PORT)((_p_)+4*3))
#define S2681_IPCR(_p_)         ((IO_PORT)((_p_)+4*4))
#define S2681_ACR(_p_)          ((IO_PORT)((_p_)+4*4))
#define S2681_ISR(_p_)          ((IO_PORT)((_p_)+4*5))
#define S2681_IMR(_p_)          ((IO_PORT)((_p_)+4*5))
#define S2681_CTU(_p_)          ((IO_PORT)((_p_)+4*6))
#define S2681_CTL(_p_)          ((IO_PORT)((_p_)+4*7))
#define S2681_MODE_B(_p_)       ((IO_PORT)((_p_)+4*8))
#define S2681_STATUS_B(_p_)     ((IO_PORT)((_p_)+4*9))
#define S2681_BAUD_RATE_B(_p_)  ((IO_PORT)((_p_)+4*9))
#define S2681_CMD_B(_p_)        ((IO_PORT)((_p_)+4*10))
#define S2681_TXRX_B(_p_)       ((IO_PORT)((_p_)+4*11))
#define S2681_IPR(_p_)          ((IO_PORT)((_p_)+4*13))
#define S2681_OPCR(_p_)         ((IO_PORT)((_p_)+4*13))
#define S2681_CNT_START(_p_)    ((IO_PORT)((_p_)+4*14)) /* read  */
#define S2681_OPR_SET(_p_)      ((IO_PORT)((_p_)+4*14)) /* write */
#define S2681_CNT_STOP(_p_)     ((IO_PORT)((_p_)+4*15)) /* read  */
#define S2681_OPR_CLEAR(_p_)    ((IO_PORT)((_p_)+4*15)) /* write */

/*   Per-channel Registers    */

/* Mode Register 1
 */
#define S2681_MR1_RxRTS_CONTROL 0x80
#define S2681_MR1_RxINT_FFULL   0x40
#define S2681_MR1_BLOCK_ERROR   0x20
#define S2681_MR1_PARITY_MASK   0x1C
# define S2681_EVEN_PARITY      0x00
# define S2681_ODD_PARITY       0x04
# define S2681_FCE_EVEN_PARITY  0x08
# define S2681_FCE_ODD_PARITY   0x0C
# define S2681_NO_PARITY        0x10
# define S2681_MULTI_DROP       0x18
#define S2681_MR1_BPW_MASK      0x03
# define S2681_5_BPW            0x00
# define S2681_6_BPW            0x01
# define S2681_7_BPW            0x02
# define S2681_8_BPW            0x03

/* Mode Register 2
 */
#define S2681_MR2_AUTO_ECHO     0x40
#define S2681_MR2_LOCAL_LOOP    0x80
#define S2681_MR2_REMOTE_LOOP   0xC0
#define S2681_MR2_TxRTS_CONTROL 0x20
#define S2681_MR2_CTS_ENABLE_Tx 0x10
#define S2681_MR2_STOP_BIT_MASK 0x0f
# define S2681_1_STOP_BIT       0x07
# define S2681_1_5_STOP_BIT     0x08
# define S2681_2_STOP_BIT       0x0F
/* The real story though is as follows
 *
 * 0  = 0.563    1 = 0.625    2 = 0.688    3 = 0.750
 * 4  = 0.563    5 = 0.625    6 = 0.688    7 = 0.750
 * 8  = 0.563    9 = 0.625   10 = 0.688   11 = 0.750
 * 12 = 0.563   13 = 0.625   14 = 0.688   15 = 0.750
 *
 * But when using 5 bpw add 0.5 to cases 0..7
 */

/* Clock Select Register
 * Baud rates depend on ACR[7]
 */
#define S2681_CSR_Rx_BAUD_MASK  0xF0
#define S2681_CSR_Tx_BAUD_MASK  0x0F
# define MakeS2681BaudRate(_x_)  ((_x_) | ((_x_) << 8))
/* ACR[7] == 0
 */
# define S2681_BR_50            0x00
# define S2681_BR_110           0x01
# define S2681_BR_134_5         0x02
# define S2681_BR_200           0x03
# define S2681_BR_300           0x04
# define S2681_BR_600           0x05
# define S2681_BR_1200          0x06
# define S2681_BR_1050          0x07
# define S2681_BR_2400          0x08
# define S2681_BR_4800          0x09
# define S2681_BR_7200          0x0A
# define S2681_BR_9600          0x0B
# define S2681_BR_38400         0x0C
# define S2681_BR_TIMER         0x0D
# define S2681_BR_EXT_x16       0x0E
# define S2681_BR_EXT_x1        0x0F
/* ACR[7] == 1
 * NOTE: missing cases same as above
 */
# define S2681_BR_75            0x00
# define S2681_BR_150           0x03
# define S2681_BR_2000          0x07
# define S2681_BR_1800          0x0A
# define S2681_BR_19200         0x0C

/* Command Register
 */
#define S2681_CR_Tx_DISABLE     0x08
#define S2681_CR_Tx_ENABLE      0x04
#define S2681_CR_Rx_DISABLE     0x02
#define S2681_CR_Rx_ENABLE      0x01
#define S2681_CR_MR1_SELECT     0x10
#define S2681_CR_Rx_RESET       0x20
#define S2681_CR_Tx_RESET       0x30
#define S2681_CR_ERR_RESET      0x40
#define S2681_CR_BREAK_RESET    0x50
#define S2681_CR_START_BREAK    0x60
#define S2681_CR_STOP_BREAK     0x70

/* Status Register
 */
#define S2681_SR_BREAK          0x80
#define S2681_SR_FE             0x40
#define S2681_SR_PE             0x20
#define S2681_SR_OE             0x10
#define S2681_SR_Tx_EMPTY       0x08
#define S2681_SR_Tx_READY       0x04
#define S2681_SR_FFULL          0x02
#define S2681_SR_Rx_READY       0x01

/*   Common registers   */

/* Output Port Configuration Register
 */
#define S2681_OPCR_TxRDYB       0x80    /* else OPR7 */
#define S2681_OPCR_TxRDYA       0x40    /* else OPR6 */
#define S2681_OPCR_RxRDYB       0x20    /* aka FFULLB. Else OPR5 */
#define S2681_OPCR_RxRDYA       0x10    /* aka FFULLA. Else OPR4 */
#define S2681_OPCR_OP3          0x00
#define S2681_OPCR_CT_OUT       0x04
#define S2681_OPCR_TxCB         0x08
#define S2681_OPCR_RxCB         0x0C
#define S2681_OPCR_OP2          0x00
#define S2681_OPCR_TxCA_x16     0x01
#define S2681_OPCR_TxCA         0x02
#define S2681_OPCR_RxCA         0x03

/* Auxillary Control Register
 */
#define S2681_ACR_BAUDRATE_1    0x80
#define S2681_ACR_CLOCK_SOURCE  0x70
# define S2681_CNT_IP2          0x00
# define S2681_CNT_TxCA         0x10
# define S2681_CNT_TxCB         0x20
# define S2681_CNT_EXT_d16      0x30
# define S2681_TIM_IP2          0x40
# define S2681_TIM_IP2_d16      0x50
# define S2681_TIM_EXT          0x60
# define S2681_TIM_EXT_d16      0x70
#define S2681_ACR_IP_MASK       0x0F
# define S2681_IP3_INT_ENABLE   0x08
# define S2681_IP2_INT_ENABLE   0x04
# define S2681_IP1_INT_ENABLE   0x02
# define S2681_IP0_INT_ENABLE   0x01

/* Input Port Change Register
 */
#define S2681_IPCR_DELTA_IP3    0x80
#define S2681_IPCR_DELTA_IP2    0x40
#define S2681_IPCR_DELTA_IP1    0x20
#define S2681_IPCR_DELTA_IP0    0x10
#define S2681_IPCR_IP3          0x08
#define S2681_IPCR_IP2          0x04
#define S2681_IPCR_IP1          0x02
#define S2681_IPCR_IP0          0x01

/* Interrupt Status Register
 */
#define S2681_ISR_IPCH          0x80
#define S2681_ISR_B_BREAK       0x40
#define S2681_ISR_B_RxRDY       0x20
#define S2681_ISR_B_TxRDY       0x10
#define S2681_ISR_COUNTER       0x08
#define S2681_ISR_A_BREAK       0x04
#define S2681_ISR_A_RxRDY       0x02
#define S2681_ISR_A_TxRDY       0x01

/* Interrupt Mask Register
 */
#define S2681_IMR_IPCH          0x80
#define S2681_IMR_B_BREAK       0x40
#define S2681_IMR_B_RxRDY       0x20
#define S2681_IMR_B_TxRDY       0x10
#define S2681_IMR_COUNTER       0x08
#define S2681_IMR_A_BREAK       0x04
#define S2681_IMR_A_RxRDY       0x02
#define S2681_IMR_A_TxRDY       0x01


/* Defines to customize the common code
 * NOTE we only handle channel B !!
 * Channel A must be used in polled mode.
 */
#define SER_READ_INTR_PORT(_i_,_p_) \
         _i_ = IOSpaceReadInt8(S2681_ISR(_p_))
#define SER_RECEIVE_INTERRUPT(_i_) \
         (_i_ & S2681_ISR_B_RxRDY)
#define SER_TRANSMITTED_INTERRUPT(_i_) \
         (_i_ & S2681_ISR_B_TxRDY)

#define SER_ISR_RECEIVED_DATA(_i_,_p_) \
         SER_RECV_BUFFER_READY(_p_)
#define SER_ISR_TRANSMITTED_DATA(_i_,_p_) \
         (SER_TRANSMITTED_INTERRUPT(_i_) && \
          SER_XMIT_BUFFER_READY(_p_))

#define SER_ENABLE_RECEIVE_INTERRUPTS(_p_) \
         IOSpaceWriteInt8(S2681_IMR(_p_),S2681_IMR_B_RxRDY)
#define SER_ENABLE_TRANSMIT_INTERRUPTS(_p_) \
         IOSpaceWriteInt8(S2681_IMR(_p_),S2681_IMR_B_RxRDY|S2681_IMR_B_TxRDY)
#define SER_DISABLE_TRANSMIT_INTERRUPTS SER_ENABLE_RECEIVE_INTERRUPTS
#define SER_DISABLE_INTERRUPTS(_p_) \
         IOSpaceWriteInt8(S2681_IMR(_p_),0)

#define SER_ACK_INTERRUPTS(_p_)

#define SER_XMIT_BUFFER_READY(_p_) \
         (IOSpaceReadInt8(S2681_ISR(_p_)) & S2681_ISR_B_TxRDY)
#define SER_RECV_BUFFER_READY(_p_) \
         (IOSpaceReadInt8(S2681_ISR(_p_)) &S2681_ISR_B_RxRDY)

#define SER_READ_RECEIVE_BUFFER(_p_) \
         IOSpaceReadInt8(S2681_TXRX_B(_p_))
#define SER_WRITE_XMIT_BUFFER(_p_,_c_) \
         IOSpaceWriteInt8(S2681_TXRX_B(_p_),_c_)

/* Logic: if it is one of the speeds we know, do it.
 * Else, round down to closest speed. Which is quite
 * possibly wrong.  [Alternate: "return E_INVALID_PARAMETER"]
 */
#define SER_SET_BAUD_RATE(_p_,_b_) \
  { \
         UINT _acr_ = 0, _br_, _bb_ = _b_; \
        _Retry_Baud_Rate_: \
         switch (_bb_) { \
         case 50: _br_ = MakeS2681BaudRate(S2681_BR_50); break; \
         case 75: _br_ = MakeS2681BaudRate(S2681_BR_75); _acr_ = S2681_ACR_BAUDRATE_1; break; \
         case 110: _br_ = MakeS2681BaudRate(S2681_BR_110); break; \
         case 134/*.5*/: _br_ = MakeS2681BaudRate(S2681_BR_134_5); break; \
         case 150: _br_ = MakeS2681BaudRate(S2681_BR_150); _acr_ = S2681_ACR_BAUDRATE_1; break; \
         case 200: _br_ = MakeS2681BaudRate(S2681_BR_200); break; \
         case 300: _br_ = MakeS2681BaudRate(S2681_BR_300); break; \
         case 600: _br_ = MakeS2681BaudRate(S2681_BR_600); break; \
         case 1050: _br_ = MakeS2681BaudRate(S2681_BR_1050); break; \
         case 1200: _br_ = MakeS2681BaudRate(S2681_BR_1200); break; \
         case 1800: _br_ = MakeS2681BaudRate(S2681_BR_1800); _acr_ = S2681_ACR_BAUDRATE_1; break; \
         case 2000: _br_ = MakeS2681BaudRate(S2681_BR_2000); _acr_ = S2681_ACR_BAUDRATE_1; break; \
         case 2400: _br_ = MakeS2681BaudRate(S2681_BR_2400); break; \
         case 4800: _br_ = MakeS2681BaudRate(S2681_BR_4800); break; \
         case 7200: _br_ = MakeS2681BaudRate(S2681_BR_7200); break; \
         case 9600: _br_ = MakeS2681BaudRate(S2681_BR_9600); break; \
         case 19200: _br_ = MakeS2681BaudRate(S2681_BR_19200); _acr_ = S2681_ACR_BAUDRATE_1; break; \
         case 38400: _br_ = MakeS2681BaudRate(S2681_BR_38400); break; \
         default: if ((_bb_) < 50) _bb_ = 51; \
                  _bb_ -= 1; \
                  goto _Retry_Baud_Rate_; \
         } \
         IOSpaceWriteInt8(S2681_ACR(_p_),(UINT8)(_acr_)); \
         IOSpaceWriteInt8(S2681_BAUD_RATE_B(_p_),(UINT8)(_br_)); \
  }

/* Line parameter encoding: MR1 on low byte, MR2 high
 * Default is 8-bit data, 1-stop bit, no parity.
 */
#define SER_DEFAULT_LINE_PARAMS \
  ((S2681_1_STOP_BIT << 8) | (S2681_NO_PARITY|S2681_8_BPW))

/* Logic: disable txrx, select MR1. Program MR1 and MR2.
 * Reset txrx. Enable txrx and clear errors.
 */
#define SER_SET_LINE_PARAMS(_p_,_l_) \
  { \
         IOSpaceWriteInt8(S2681_CMD_B(_p_),S2681_CR_Tx_DISABLE|S2681_CR_Rx_DISABLE|S2681_CR_MR1_SELECT); \
         IOSpaceWriteInt8(S2681_MODE_B(_p_),(UINT8)(((_l_)&0x1f)|S2681_MR1_BLOCK_ERROR)); \
         IOSpaceWriteInt8(S2681_MODE_B(_p_),(UINT8)(((_l_)>>8)&0x0f)); \
         IOSpaceWriteInt8(S2681_CMD_B(_p_),S2681_CR_Tx_RESET); \
         IOSpaceWriteInt8(S2681_CMD_B(_p_),S2681_CR_Rx_RESET); \
         IOSpaceWriteInt8(S2681_CMD_B(_p_),S2681_CR_Tx_ENABLE|S2681_CR_Rx_ENABLE|S2681_CR_ERR_RESET); \
  }

/* We dont do modem control
 */
#define SER_SET_MODEM_BITS(_p_,_b_)

/* Transmitter interrupts when empty
 */
#define SER_INTERRUPTS_ON_XMIT_EMPTY 1

/* Not yet BUGBUG */
#define SER_ERRORS(_i_,_p_) FALSE
#define SER_CLEAR_ERRORS(_i_,_p_)

/* Nothing special at init time
 */
#define SER_SPECIAL_INIT(_p_)

/* BUGBUG later */
#define SER_TEST_FOR_MORE_INTERRUPTS(_p_,_i_,_l_)

/* Device driver proper
 */
#include "serial.h"
