/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Microsoft MMLite Nov 2000 --sandrof
 * Remote IFile using serial lines.
 * 
 * Designed for an unreliable serial line link,
 * but can be used elsewhere. Only needs an IFile :-))
 */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

#define __HOSTFS_CLIENT 1
#include <drivers/hostfsp.h>
#include <drivers/serplex.h>

BOOL IsLongTimeout = FALSE;

#if INTERNAL_POLLING_TEST
#include "polling_serial.c"
#else

#define DBGLVL 3
/* Single-byte I/O 
 */
static const UINT64 Zero = Int64Initializer(0,0);

SCODE Getabyte(PIFILE Link, UINT8 *pByte)
{
    SCODE sc;
    UINT BytesRead;

    sc = Link->v->ReadAt(Link, Zero, pByte, 1, &BytesRead);

    return sc;
}

SCODE Petabyte(PIFILE Link, UINT8 Byte)
{
    SCODE sc;
    UINT BytesWritten;

    sc = Link->v->WriteAt(Link, Zero, &Byte, 1, &BytesWritten);

    return sc;
}

SCODE Petabytes(PIFILE Link, UINT8 *pBytes, UINT nBytes)
{
    SCODE sc;
    UINT BytesWritten;

    sc = Link->v->WriteAt(Link, Zero, pBytes, nBytes, &BytesWritten);

    return sc;
}

void LongTimeout(PIFILE Link)
{
    UINT64 Param;
    UINT32 *p = (UINT32*)&Param;

    p[0] = 0;
    p[1] = TIME_SECONDS(1);
    (void) Link->v->SetSize(Link, Param);
    /* ignore errors, we tried */

    IsLongTimeout = TRUE;
}

void ShortTimeout(PIFILE Link)
{
    UINT64 Param;
    UINT32 *p = (UINT32*)&Param;

    p[0] = 0;
    p[1] = TIME_MILLIS(100);
    (void) Link->v->SetSize(Link, Param);
    /* ignore errors, we tried */

    IsLongTimeout = FALSE;
}

#endif

#define LINK_HANDLE PIFILE
#define Pushout(_l_,_n_)

struct _LINK_ERRORS LinkErrors = {0,};
void PrintLinkErrors(struct _LINK_ERRORS *Errors)
{
#ifdef _DEBUG
    _tprintf(_T("{LE"));
    if (Errors->LongTimeout)
        _tprintf(_T(" timeo %d"), Errors->LongTimeout);
    if (Errors->BeforeSTX)
        _tprintf(_T(" stx %d"), Errors->BeforeSTX);
    if (Errors->Retransmit)
        _tprintf(_T(" rep %d"), Errors->Retransmit);
    if (Errors->FirstByteSum)
        _tprintf(_T(" 1st %d"), Errors->FirstByteSum);
    if (Errors->SecondByteSum)
        _tprintf(_T(" 2nd %d"), Errors->SecondByteSum);
    if (Errors->OutOfSequence)
        _tprintf(_T(" seq %d"), Errors->OutOfSequence);
    if (Errors->BadLength)
        _tprintf(_T(" len %d"), Errors->BadLength);
    if (Errors->Reset)
        _tprintf(_T(" rst %d"), Errors->Reset);
    _tprintf(_T("}\n"));
#endif
}
UINT PutPacketA(LINK_HANDLE hLine, struct packet *Pkt, UINT BufferSize)
{
    UINT8 *p = (UINT8*) Pkt;
    UINT Len = Pkt->packet_length;
    SCODE sc;
    UINT BytesWritten = BufferSize;
    
    sc = hLine->v->WriteAt(hLine, Zero, p, Len, &BytesWritten);

    if (FAILED(sc)) return 0;
    else return BytesWritten;
}
UINT GetPacketA(LINK_HANDLE hLine, struct packet *Pkt, UINT BufferSize)
{
    SCODE sc;
    UINT BytesRead = 0;
    
    sc = hLine->v->ReadAt(hLine, Zero, (BYTE *)Pkt, BufferSize, &BytesRead);

    if (FAILED(sc) || BytesRead < 2) 
        return 0;
    
    return BytesRead;
}

/* In-place request/reply. 
 */
SCODE LinkLevelRpc(PCONNECTION Connection, struct packet *Packet,
                   UINT PacketBufferSize)
{
    UINT nBytes, i;

    if (0 == PutPacketA( Connection->Link, Packet, PacketBufferSize)) {
        return E_UNEXPECTED_ERROR;
    }

Retry:
    nBytes = GetPacketA(Connection->Link, Packet, PacketBufferSize);
   
    if (nBytes == 0) {
        PrintLinkErrors(&LinkErrors);
//_tprintf(_T("<R>"));
        return E_RESTART;//E_TIMED_OUT;
    }

#if !defined(__HOSTFS_CLIENT)
    /* Client gets it in native byteorder, server maybe not
        */
    i = ByteSwap16(nBytes);
#else
    i = nBytes;
#endif
    if (i != Packet->packet_length) {
        /* BUGBUG Server should send a NACK ! */
        LinkErrors.BadLength++;
        goto NACK;
    }

    /* Verify packet flow logic, check reconnection/target restart
        */
    if ((Packet->seqno == Connection->SequenceNumber) ||
        (Packet->packet_type == PACKET_TYPE_RESET)) {
        /* Done */
        goto Done;
    }
    
    LinkErrors.OutOfSequence++;

    //_tprintf(_T("ORD: %d != %d\n"),Packet->seqno, Connection->SequenceNumber);

NACK:
    /* Send a NACK and repeat
        */
    Packet->packet_type = PACKET_TYPE_RESET;
    Packet->seqno = (UINT8) Connection->SequenceNumber;
    Packet->sc = E_RESTART;

    /* Send back, fix length */
    Packet->packet_length = MIN_PACKET_SIZE;
    (void) PutPacketA(Connection->Link, Packet, nBytes);

Done:
    /* Out of synch or initial handshake */
    if (Packet->packet_type == PACKET_TYPE_RESET) {
        Connection->SequenceNumber = Packet->seqno;
        LinkErrors.Reset++;
        PrintLinkErrors(&LinkErrors);

        /* There might be more than one packet queued, try again.
         * If there is something good we'll get it and all if well again.
         * If there is nothing we'll time out.
         * The problem here is that the link might have buffering,
         * if we send again we'll be forever out of synch.
         */
        goto Retry;
    }
    return NO_ERROR;
}


