/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <drivers/drivers.h>
#include <diagnostics.h>
#include <drivers/serp.h>

#define __HOSTFS_CLIENT 1
#include <drivers/hostfsp.h>

#include <drivers/serplex.h>
#include "getput.c"

PRIVATE PTRANSPORT SerplexTransportList = NULL;
MUTEX SerplexLock = MUTEX_INITIALIZER;

/*-------------------- IFile methods --------------------------*/

PRIVATE UINT MCT SerplexAddRef(PIFILE This)
{
    PSERPLEX File = (PSERPLEX) This;

    DBG(("SERPLEXADDREF "));
    return AtomicInc(&File->RefCnt);
}

PRIVATE SCODE MCT SerplexQueryInterface(PIFILE This, REFIID pIid, void* *ppNew)
{
    return GenericQueryInterface((PIUNKNOWN) This, pIid, ppNew, &IID_IFile);
}
 
PRIVATE UINT MCT SerplexRelease(PIFILE pThis)
{
    PSERPLEX File = (PSERPLEX) pThis;
    INT Refs;
    PIHEAP heap = CurrentHeap();

    Refs = AtomicDec(&File->RefCnt);
    DBG(("SerplexRELE%d ", Refs));
    if (Refs == 0) {
        INT tRefs;
        tRefs = AtomicDec(&File->transport->RefCnt);
        File->transport->MemberBitMap &= ~(1 << File->SerplexAddr); /* Remove from bitmap */
        if (tRefs == 0) {
            /* Remove File->Transport from list */
            PTRANSPORT pT, *pp = &SerplexTransportList;
            Mutex_Lock(&SerplexLock);
            for (pT = *pp; pT; pp = &pT->next, pT = pT->next) {
                if (pT == File->transport) {
                    *pp = pT->next; /* Make previous skip current */
                    pT->file->v->Release(pT->file);
                    heap->v->Free(heap, 0, pT);
                    break;
                }
            }
            Mutex_Unlock(&SerplexLock);
        }
        heap->v->Free(heap, 0, File);
        DBG(("Serplex transport RELE%d ", tRefs));
    }
    return Refs;
}

/*
 * Read from a file
 */
PRIVATE SCODE MCT
SerplexReadAt(PIFILE This, UINT64 Position, 
          BYTE *Buffer, UINT ByteCount, PUINT pSizeRead)
{
    PSERPLEX File =  (PSERPLEX)This;
    PTRANSPORT pT = File->transport;
    SCODE sc = S_OK;
    TIME WakeupTime;
    
    Int32ToInt64(WakeupTime, TIME_RELATIVE(TIME_MILLIS(500)));
    
    Mutex_Lock(&pT->ReadLock);

    pT->MemberBitMap |= (1 << File->SerplexAddr);

GetFromBuffer:
    /* Did we already read something? */
    if (File->SerplexAddr == CONSerplexAddr) {
        if (pT->ConAvail) {
            if (ByteCount > pT->ConAvail)
                ByteCount = pT->ConAvail;
            memcpy(Buffer, pT->ConReadBuffer, ByteCount);
            pT->ConAvail -= ByteCount;
            if(pT->ConAvail)
                memmove(pT->ConReadBuffer, &pT->ConReadBuffer[ByteCount], pT->ConAvail);
            *pSizeRead = ByteCount;
            // Look for a newline character to make scanf() happy
            {
                UINT i;
                for (i = 0; i < ByteCount; i++) {
                    if ((Buffer[i] == '\r') || (Buffer[i] == '\n')) {
                        sc = S_FALSE;
                        break;
                    }
                }
            }
        } else if (pT->Avail) {
            sc = Condition_Wait(&pT->ReadCondition, &pT->ReadLock, NULL);
            assert(sc == S_OK);
            goto GetFromBuffer;
        } else {
            goto ReadSerialLine;
        }
    } else if (pT->Avail) {
        if (pT->IntendedReceiver != File->SerplexAddr) {/* not me */
            /* The hostfs server just send the respond and then leave it alone. So if this response is lost,
             * The hostfs client will be waken up many times. We assume it is timed out! Then send restart 
             * request. If FS client missed a packet, it must inform the server to resend. 
             * Otherwise FS client would try to read it forever and the server never 
             * send the missed packet again.
             */
            if (File->SerplexAddr == FSSerplexAddr) {
                if (File->WakeupNum > 25) { 
                    File->WakeupNum = 0;
                    goto out;
                }
                File->WakeupNum++;
            }
Wait:
            sc = Condition_Wait(&pT->ReadCondition, &pT->ReadLock, &WakeupTime);
            if (SUCCEEDED(sc))
                goto GetFromBuffer;
            /* else TIMED OUT */

            /* 
             * Clean console buffer when timeout. Otherwise it will be deadlocked when the console is running 
             * an application on the foreground
             */
            if (pT->ConAvail)
                pT->ConAvail = 0;
        } else {
            if (ByteCount > pT->Avail)
                ByteCount = pT->Avail;
            memcpy(Buffer, &pT->ReadBuffer[pT->Consumed], ByteCount);
            pT->Avail -= ByteCount;
            pT->Consumed += ByteCount;
            *pSizeRead = ByteCount;
            File->WakeupNum = 0;
        }
    }  else if (pT->ConAvail) {
        /* Deal with the Console input first */
        goto Wait;
    } else {
        /* Read new Serplex frame */
        BOOL ok;
ReadSerialLine:
        for (;;) {
            assert(pT->Avail == 0);
            pT->IntendedReceiver = InvalidSerplexAddr;
            ok = GetPacket(pT);
#if 0
switch(File->SerplexAddr) {
case CONSerplexAddr:
    printf("<C");
    break;
case NICSerplexAddr:
    printf("<N");
    break;
case FSSerplexAddr:
    printf("<F");
    break;
default:
    printf("<I");
}
switch(pT->IntendedReceiver) {
case CONSerplexAddr:
    printf("C>");
    break;
case NICSerplexAddr:
    printf("N>");
    break;
case FSSerplexAddr:
    printf("F>");
    break;
default:
    printf("I>");
}
#endif
            if (ok == FALSE || pT->IntendedReceiver > 31 || (pT->MemberBitMap & (1 << pT->IntendedReceiver)) == 0) {
                /* Bad packet or Bad address */
                pT->Avail = 0;
                pT->IntendedReceiver = InvalidSerplexAddr;
                 
                if (File->SerplexAddr == FSSerplexAddr) {
                    /* If FS client missed a packet, it must inform the server to resend. 
                     * Otherwise if no valid packet comes in, FS client would always read 
                     * an invalid packet and the server never send the missed packet again.
                     * For NIC, the server will time out and resend the lost packet but a 
                     * long time loop will take so much time that raid64.exe test will fail.
                     * For Console, we must loop until we get a correct packet.
                     */
                    File->ContinueDropNum++;
                    if (File->ContinueDropNum > 25) {
                        File->ContinueDropNum = 0;
                        goto out;
                    }
                }
                continue;
            }
            File->ContinueDropNum = 0;
            if (pT->IntendedReceiver != File->SerplexAddr) {
                Condition_Broadcast(&pT->ReadCondition);
                if (File->SerplexAddr == NICSerplexAddr)
                    /* Sernet must wait otherwise Console would be blocked 
                     * for a long time. Otherwise the Console input cannot 
                     * be dealed with until Sernet gets a non-console pkt.
                     */
                    goto Wait;
            }
            goto GetFromBuffer;
        }
out:
        sc = E_TIMED_OUT;
    }
    Mutex_Unlock(&pT->ReadLock);
    Condition_Broadcast(&pT->ReadCondition);
    return sc;
}

/*
 * Write to a file
 */
PRIVATE SCODE MCT
SerplexWriteAt(PIFILE This, UINT64 Position, 
               const BYTE *Buffer, UINT ByteCount, PUINT pSizeWritten)
{
    PSERPLEX File = (PSERPLEX) This;
    SCODE sc = S_OK;
	UINT ret;

    Mutex_Lock(&File->transport->WriteLock);

    ret = PutPacket(File->transport->file, 
		            (BYTE *)Buffer,
					ByteCount, 
					ByteCount,
					File->SerplexAddr);
    *pSizeWritten = ByteCount;

    Mutex_Unlock(&File->transport->WriteLock);
    return sc;
}
 
/* Get/Set a file's size
 */
PRIVATE SCODE MCT SerplexGetSize(PIFILE This, PUINT64 pSize)
{
    PSERPLEX File = (PSERPLEX) This;
    return File->transport->file->v->GetSize(File->transport->file, pSize);
}
 
PRIVATE SCODE MCT SerplexSetSize(PIFILE This, UINT64 Size)
{
    PSERPLEX File = (PSERPLEX) This;
    return File->transport->file->v->SetSize(File->transport->file, Size);
}

const struct IFileVtbl SerplexVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    SerplexQueryInterface, 
    SerplexAddRef, 
    SerplexRelease, 
    SerplexReadAt, 
    SerplexWriteAt, 
    SerplexSetSize, 
    SerplexGetSize
};

/*--------------------- Constructor ----------------------*/

PIFILE SerplexNew(PTRANSPORT transport, BYTE SerplexAddr)
{
    PSERPLEX File;
    PIHEAP heap = CurrentHeap();

    File = heap->v->Alloc(heap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("serplex"), sizeof(*File), 0);
    if (!File) {
        return NULL;
    }

    File->v = &SerplexVtbl;
    File->RefCnt = 1;
    File->transport = transport;
    File->SerplexAddr = SerplexAddr;
    
    return (PIFILE) File;
}

SCODE __cdecl SerplexConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        _TCHAR *TransportName,
        _TCHAR *s2,
        UINT BaudRate,
        UINT Address)
{
    PIUNKNOWN pUnk;
    PIFILE file;
    SCODE sc = S_OK;
    UINT64 Size;
    UINT32 *p = (UINT32*)&Size;
    PTRANSPORT pT = SerplexTransportList;
    BOOL isFound = FALSE;
    PIHEAP heap = CurrentHeap();

    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);
    UnusedParameter(s2);
    
    DBG(("SerplexConstructor(%x %s %x %x %x)\n",
        (ADDRESS) pThis, Name, RegisterBase, Interrupt,
        (ADDRESS) ppUnk));
    if (Address > 31)
        return E_INVALID_PARAMETER;

    Mutex_Lock(&SerplexLock);
    if(pT) {
        do {
            if(_tcscmp(TransportName, pT->name) == 0) {
                isFound = TRUE;
                break;
            }
            pT=pT->next;
        } while(pT);
    }
    if(!isFound) {
        sc = BindToObject(CurrentNameSpace(), 
                          TransportName, 
                          NAME_SPACE_READ | NAME_SPACE_WRITE,
                          &IID_IFile,
                          (void **)&file);
        if (FAILED(sc)) {
            _tprintf(_T("SerplexConstructor: failed in opening %s, sc = %x\n"), TransportName, sc);
            goto out;
        }
        if(BaudRate) {
            p[0] = BaudRate;
            p[1] = 0;

            sc = file->v->SetSize(file, Size);
            if(FAILED(sc)) goto out;
        }
        pT = heap->v->Alloc(heap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("serplex:tr"),
                            sizeof(*pT)+(_tcslen(TransportName)+1)*sizeof(_TCHAR), 0); 
        if (!pT) {
            DBG(("SerplexConstructor: Alloc failed\n"));
            sc = E_NOT_ENOUGH_MEMORY;
            goto out;
        }
        
        pT->name = (_TCHAR*)(pT+1);
        _tcscpy(pT->name,TransportName);
        pT->RefCnt = 1;
        pT->file = file;
        Mutex_Init(&pT->WriteLock);
        Mutex_Init(&pT->ReadLock);
        Condition_Init(&pT->ReadCondition);
        pT->IntendedReceiver = InvalidSerplexAddr;
        pT->next = SerplexTransportList;
        SerplexTransportList = pT;
    } else {
        file = pT->file;
        AtomicInc(&pT->RefCnt);
    }

    pUnk = (PIUNKNOWN) SerplexNew(pT, (UINT8) Address);
        
    DBG(("SerplexConstructor returns %x\n", (ADDRESS) pUnk));
    if (pUnk) {
        *ppUnk = pUnk;
    } else {
        DBG(("SerplexConstructor failed\n"));
        file->v->Release(file);
        sc = E_FAIL;
    }
out:
    Mutex_Unlock(&SerplexLock);
    return sc;
}

static struct IDeviceConstructorVtbl SerplexDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) SerplexConstructor
};

static struct _DRIVERCOB SerplexDriverCobObject = { &SerplexDriverCobVtable, 1 };

PIUNKNOWN SerplexCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &SerplexDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &SerplexDriverCobObject;
}
