/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 *    Module: pnp.c
 *
 *    Description: Routines for PlugNPlay notifications
 */
 
#include "serplexd.h"

/*
 * The application will receive the WM_DEVICECHANGE message whenever 
 * a device interface event notification is sent. 
 * Currently, the only events that will be received for device interfaces are
 * DBT_DEVICEARRIVAL and DBT_DEVICEREMOVECOMPLETE.
 * 
 * The system broadcasts a set of default device change events to
 * all applications and services.
 * You do not need to register to receive these default events. 
 * See the Remarks section in RegisterDeviceNotification for details.
 *
 * So that's what the doc says, and so it works on W2003.
 * Under Win2k though we only get DBT_DEVNODES_CHANGED,
 * One for plug and two for unplug.  Oh well.
 */

/* Lots of unfortunate crap because our header files are NT4 compat
 */
#if (WINVER < 0x500)
#undef WINVER
#define WINVER 0x500
#define DEVICE_NOTIFY_WINDOW_HANDLE          0x00000000
#define DEVICE_NOTIFY_ALL_INTERFACE_CLASSES  0x00000004
#define WS_EX_NOACTIVATE        0x08000000L
#define HDEVNOTIFY PVOID
#undef _DBT_H
#include <dbt.h>
WINUSERAPI
HDEVNOTIFY
WINAPI
RegisterDeviceNotificationA(
    IN HANDLE hRecipient,
    IN LPVOID NotificationFilter,
    IN DWORD Flags
    );
WINUSERAPI
HDEVNOTIFY
WINAPI
RegisterDeviceNotificationW(
    IN HANDLE hRecipient,
    IN LPVOID NotificationFilter,
    IN DWORD Flags
    );
#ifdef UNICODE
#define RegisterDeviceNotification  RegisterDeviceNotificationW
#else
#define RegisterDeviceNotification  RegisterDeviceNotificationA
#endif // !UNICODE

#endif


GUID WceusbshGUID = { 0x25dbce51, 0x6c8f, 0x4a72, 
                      0x8a,0x6d,0xb5,0x4c,0x2b,0x4f,0xc8,0x35 };

BOOL DoRegisterDeviceInterface( 
    HWND hWnd,
    GUID InterfaceClassGuid, 
    HDEVNOTIFY *hDevNotify
)
/*
Routine Description:
    Registers for notification of changes in the device interfaces for
    the specified interface class GUID. 

Parameters:
    hWnd - The window handle to send notifications to.

    InterfaceClassGuid - The interface class GUID for the device 
        interfaces. 

    hDevNotify - Receives the device notification handle. On failure, 
        this value is NULL.

Return Value:
    If the function succeeds, the return value is TRUE.

    If the function fails, the return value is FALSE.
*/
{
    DEV_BROADCAST_DEVICEINTERFACE NotificationFilter;
    DWORD Err;

    ZeroMemory( &NotificationFilter, sizeof(NotificationFilter) );
    NotificationFilter.dbcc_size = 
        sizeof(DEV_BROADCAST_DEVICEINTERFACE);
    NotificationFilter.dbcc_devicetype = DBT_DEVTYP_DEVICEINTERFACE;
    NotificationFilter.dbcc_classguid = InterfaceClassGuid;

    *hDevNotify = RegisterDeviceNotification( hWnd, 
        &NotificationFilter,
        DEVICE_NOTIFY_WINDOW_HANDLE | DEVICE_NOTIFY_ALL_INTERFACE_CLASSES
    );

    if(!*hDevNotify) 
    {
        Err = GetLastError();
        printf( "RegisterDeviceNotification failed: %lx.\n", Err);
        return FALSE;
    }

    return TRUE;
}

/* Find out what we are running on
 */
static int _winver = 0;

/* Stolen from CRT
   */
void FindWindowsVersion(void)
{
    int _osplatform, _winmajor, _winminor, _osver;
    OSVERSIONINFOA osvi, *posvi = &osvi;

    /*
     * Get the full Win32 version
     */
    posvi->dwOSVersionInfoSize = sizeof(OSVERSIONINFOA);
    (void)GetVersionExA(posvi);

    _osplatform = posvi->dwPlatformId;
    _winmajor = posvi->dwMajorVersion;
    _winminor = posvi->dwMinorVersion;

    /*
     * The somewhat bizarre calculations of _osver and _winver are
     * required for backward compatibility (used to use GetVersion)
     */
    _osver = (posvi->dwBuildNumber) & 0x07fff;
    if ( _osplatform != VER_PLATFORM_WIN32_NT )
        _osver |= 0x08000;
    _winver = (_winmajor << 8) + _winminor;

#if 0
    printf("Windows: %x %x %x %x %x\n",
           _winver, _osplatform, _winmajor, _winminor, _osver);
#endif
}

/* main window procedure */
static LRESULT WINAPI FramebufferWndProc (
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
    HDEVNOTIFY Notify;
    LRESULT    lRet = 1;

    switch (uMsg)
    {
    case WM_CREATE:
        if (!DoRegisterDeviceInterface(hWnd,
                                       WceusbshGUID,
                                       &Notify))
            printf("FAILED to register for PnP notifications: %x\n",
                   GetLastError());

        break;

    case WM_COMMAND:
    case WM_SIZE:
    case WM_DESTROY:
        break;

    case WM_DEVICECHANGE:
    {
        PNETWORK Serial = (PNETWORK) GetWindowLongPtr( hWnd, GWLP_USERDATA);
        PDEV_BROADCAST_DEVICEINTERFACE b = (PDEV_BROADCAST_DEVICEINTERFACE) lParam;
        BOOL IsRemoval = TRUE;

        //printf("dc %x[%x,%x]\n",uMsg,wParam,lParam);

        switch (wParam)
        {
        case DBT_DEVICEARRIVAL:
            IsRemoval = FALSE;
            /* fall through */
        case DBT_DEVICEREMOVECOMPLETE:
            /* First check this is the right GUID.
             * We get two notifications, only need one
             */
            if (memcmp(&b->dbcc_classguid,
                       &WceusbshGUID,
                       sizeof WceusbshGUID)==0)
            {
                if (IsRemoval) {
                HandleRemoval:
                    printf("[USB cable unplugged, console suspended]\n");
                    SerialFinalize(Serial);
                } else {
                HandleInsertion:
                    if (SerialInitialize(Serial,TRUE) != NULL) {
                        printf("[USB cable was plugged in]\n");
                        SetCompletionPorts();
                        PostReads(Serial, MAX_PACKETS_PER_NETWORK);
                    }
                }
            }
            break;

        case DBT_DEVNODES_CHANGED:
            /* If this isnt W2k we dont care
             */
            if (_winver < 0x501) {
                /* Ok, the bozo one.
                 * Weirdest thing is that some other USB driver is 
                 * infact capable of generating the correct notifs.
                 * Thank you Intel.
                 */
                /* So we get ONE notif when unplugged, and TWO when plugged.
                 * See if we can open the driver.
                 */
                HANDLE hLine;
                hLine = CreateFile( Serial->Name, GENERIC_WRITE,
                                    FILE_SHARE_WRITE, /* exclusive access */
                                    NULL, /* no security attributes */
                                    OPEN_EXISTING,
                                    FILE_FLAG_OVERLAPPED,
                                    NULL
                                    );

                /* If we cannot open it its gone.
                 * But it might have been some *other* device.
                 */
                if ((hLine == INVALID_HANDLE_VALUE) &&
                    (Serial->bInitialized)) {
                    goto HandleRemoval;
                } else {
                    /* First off, close this handle
                     */
                    CloseHandle(hLine);
                    /* If we already refreshed our handle we are done
                     * NB: HandleInsertion might onyl work the second time around
                     */
                    if (!Serial->bInitialized) {
                        //printf("W2k replugged?\n");
                        goto HandleInsertion;
                    }
                }

            }

        default:
            /* ignore */
            break;
        }
        break;
    }
    default:
        /* pass all unhandled messages to DefWindowProc */
        lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }

    /* return 1 if handled message, 0 if not */
    return lRet;
}

BOOL PnpThread( PNETWORK Serial )
{
    WNDCLASS   wc;
    HWND       hWnd;
    HINSTANCE  hInstance;
    MSG       msg;
    static char szAppName[16]="mypnpevents";

    FindWindowsVersion();

    /* Find our root instance
     */
    hInstance = GetModuleHandleA(NULL);

    /* Register the frame class
     */
    wc.style         = CS_VREDRAW | CS_HREDRAW;
    wc.lpfnWndProc   = (WNDPROC) FramebufferWndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = NULL/*LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1))*/;
    wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
    wc.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName  = NULL/*szAppName*/;
    wc.lpszClassName = szAppName;

    if (!RegisterClass (&wc) ) {
        volatile DWORD err = GetLastError();
        printf("RegisterClass failed x%x\n",err);
        return FALSE;
    }

    /* Create the window, invisible to avoid clutter
     */
    hWnd = CreateWindowEx ( WS_EX_NOACTIVATE,
                            szAppName,
                            szAppName,
                            WS_ICONIC,
                            CW_USEDEFAULT,
                            CW_USEDEFAULT,
                            CW_USEDEFAULT,
                            CW_USEDEFAULT,
                            NULL,
                            NULL,
                            hInstance,
                            NULL);

    /* make sure window was created 
     */
    if (!hWnd) {
        volatile DWORD err = GetLastError();
        printf("CreateWindow failed x%x\n",err);
        return FALSE;
    }

    /* Associate the device pointer to it
     */
    (void) SetWindowLongPtr(hWnd,
                            GWLP_USERDATA,
                            (LONG_PTR) Serial );

    /* Pump messages till the end
     */

    ShowWindow(hWnd,SW_HIDE);
    UpdateWindow(hWnd);

    while (GetMessage (&msg, NULL, 0,0))
      {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }

    return TRUE;
}

