/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 *    Module: serplexd.c
 *
 *    Description: Routines for serial line multiplexier,
 *                 routine for dispatching thread (recv,filter,send),
 *                 and main entry point for this program.
 */
 
#define BYTESWAPEM 1
#include <drivers/serp.h>
#include "serplexd.h"
#define __HOSTFS_SERVER 1
#include <drivers/hostfsp.h>

BOOL NoEther = FALSE;
BOOL OldEther = FALSE;
BOOL doSendError = FALSE;
BOOL doByteswap = FALSE;
BOOL RawMode = FALSE;

#undef _bs32
#undef _bs16
#define _bs32(x) ((doByteswap) ? ByteSwap32(x) : x)
#define _bs16(x) ((doByteswap) ? ByteSwap16(x) : x)

void PacketDrop(PPACKET Packet);

void SendNACK(PPACKET Packet, PNETWORK Network) 
{
    struct packet *pPkt = (struct packet *)Packet->Data;
    /* Send a NACK and repeat
        */
    pPkt->packet_type = PACKET_TYPE_RESET;
    pPkt->seqno = 0;
    pPkt->sc = _bs32(E_RESTART);

    /* Send back, fix length */
    pPkt->packet_length = _bs16(MIN_PACKET_SIZE);
    Packet->SerplexAddr = FSSerplexAddr;
   
    SerplexWrap(Packet, Packet->SerplexAddr);
    PacketWriteFile(Network, Packet);
}

void SendAnError(PNETWORK in)
{
    UINT randomnum = rand(), rn, i;
    PPACKET ErrPkt;
    
    if (randomnum > 500)
        return;
    PacketAllocate(&ErrPkt);
    PacketInit(ErrPkt);
    ErrPkt->HomeNetwork = NULL; /* temporary packet, should not be reposted */

    ErrPkt->SerplexAddr = NICSerplexAddr;
    ErrPkt->nBytesAvail = randomnum/4;
    for(i=0; i<randomnum/sizeof(UINT)/4; i++) {
        rn = rand();
        memcpy(ErrPkt->Data + i * sizeof(UINT), &rn, sizeof(UINT));
    }
    
    if (in->ForwardingTo->Filter == SerialFilter) {
        PacketWriteFile(in->ForwardingTo, ErrPkt);
    } else if (in->Filter == SerialFilter) {
        PacketWriteFile(in, ErrPkt);
    }
}
void ForwardPacket(PNETWORK in, PPACKET Packet) 
{
    struct packet *pPkt = (struct packet *)Packet->Data;
    UINT nBytesTransferred;
    HANDLE hConOut;
 
    if(doSendError) 
        SendAnError(in);

    if (Packet->SerplexAddr == FSSerplexAddr) {
        /* Verify packet flow logic, check reconnection/target restart
            */
        if (Packet->nBytesAvail != (UINT32) _bs16(pPkt->packet_length)) {
            SendNACK(Packet, in);
            return;
        }
        
        if (doByteswap)
            Packet->nBytesAvail = ByteswappedHostFile_Server( NULL,
                                                   (void*) Packet->Data,
                                                   MAX_ETHERNET_PACKET_SIZE - SERIAL_OVERHEAD - (Packet->Data - Packet->Buffer));
        else
            Packet->nBytesAvail = HostFile_Server( NULL,
                                                              (void*) Packet->Data,
                                                              MAX_ETHERNET_PACKET_SIZE - SERIAL_OVERHEAD - (Packet->Data - Packet->Buffer));
        
        if (debug_level > 1 )
        {
            printf("ActuallyOut: %d\n",Packet->nBytesAvail);
            PacketDump(Packet,"SerialOutput: ");
        }
        SerplexWrap(Packet, Packet->SerplexAddr);
        PacketWriteFile(in, Packet);
    } else if ((!NoEther && Packet->SerplexAddr == NICSerplexAddr)
              || (in->Filter == ConsoleFilter && Packet->SerplexAddr == CONSerplexAddr))
    {
        if (debug_level > 1)
        {
            printf("ActuallYOut: %d\n",Packet->nBytesAvail);
            PacketDump(Packet,"EtherOutput: ");
        }

        if (in->ForwardingTo->Filter == SerialFilter) {
            /* Length is only used by the sernet layer */
            if (Packet->SerplexAddr == NICSerplexAddr) { /* Insert length */
                Packet->Data -= 2;
                Packet->Data[1] = (UINT8)Packet->nBytesAvail;
                Packet->Data[0] = (UINT8)(Packet->nBytesAvail>>8);
                Packet->nBytesAvail += 2;
            }
            SerplexWrap(Packet, Packet->SerplexAddr);
        }
        PacketWriteFile(in->ForwardingTo, Packet);
    } else if (Packet->SerplexAddr == CONSerplexAddr) {
        if (debug_level > 1)
        {
            printf("ActuallyOut: %d\n",Packet->nBytesAvail);
            PacketDump(Packet,"ConsoleOutput: ");
        }
        
        hConOut = GetStdHandle(STD_OUTPUT_HANDLE);
        WriteFile(hConOut, 
                  Packet->Data, 
                  Packet->nBytesAvail, 
                  &nBytesTransferred, 
                  0);
        if (RawMode)
            PacketDrop(Packet);
    } else if (RawMode) {
        if (debug_level > 1)
        {
            printf("sa %x %d bytes: %c..\n", 
                   Packet->SerplexAddr, Packet->nBytesAvail, Packet->Data[0]);
        }
        PacketWriteFile(in->ForwardingTo, Packet);
    }
}

UINT TotalActualLength = 0, TotalCompressedLength = 0;
/*
 *    Wrap an ethernet packet with length and checksum, serial format.
 */
BOOL SerplexWrap(PPACKET Packet, UINT8 SerplexAddr)
{
    UINT16      Sum;
    UINT        Size, i, nEscapes = 0;
    UINT8       *pSrc, *pNext, *Buf;
    UINT8       PrevByte, Byte, ByteNum;
    
    Sum = 0;
    TotalActualLength += Packet->nBytesAvail + 5;
    Size = Packet->nBytesAvail + 1;
    Buf = malloc(sizeof(UINT8) * Size);
    if (!Buf) {
        printf("SerplexWrap: memory allocation failed!\n");
    }
    
    memcpy(Buf, Packet->Data, Size-1);
    Buf[Size-1] = ~Buf[Size-2]; /* pad an extra char which is different from the last one */
    pNext = Packet->Data;
    pSrc = Buf;

    /* Wrap the packet with Serplex and compress*/
    
    *pNext++ = STH_CHAR;
    *pNext++ = SerplexAddr;
    Packet->nBytesAvail = 2;
    ByteNum = 0;
    PrevByte = *pSrc;
    for (;Size-- > 0;) {
        Byte = *pSrc++;
        if (Byte == PrevByte && ByteNum < 0xff) {
            ByteNum++;
        } else {
            if (ByteNum < 4) { /* no compressing if too small*/
                for(i=0;i<ByteNum;i++) {
                    if ((PrevByte == STH_CHAR) ||
                        (PrevByte == EOT_CHAR) || 
                        (PrevByte == ESC_CHAR) ||
                        (PrevByte == COMP_CHAR)) 
                    {
                        *pNext++ = ESC_CHAR;
                        Packet->nBytesAvail++;
                    }
                    *pNext++ = PrevByte;
                    Sum += PrevByte;
                    Packet->nBytesAvail++;
                }
            } else {
                *pNext++ = COMP_CHAR;
                Packet->nBytesAvail++;
                *pNext++ = ByteNum;
                Sum += ByteNum;
                Packet->nBytesAvail++;
                *pNext++ = PrevByte;
                Sum += PrevByte;
                Packet->nBytesAvail++;
            }
            PrevByte = Byte;
            ByteNum = 1;
        }
    }

    *pNext++ = EOT_CHAR;
    Packet->nBytesAvail++;
    
    /* Fix checksum on account of STX chars */
    if ((Sum & 0xff) == STH_CHAR)
        Sum++;
    if ((Sum >> 8) == STH_CHAR)
        Sum += (1 << 8);/* this must not rollover */

    *pNext++ = (UINT8)Sum;
    Packet->nBytesAvail++;
    *pNext++ = (UINT8)(Sum>>8);
    Packet->nBytesAvail++;

    TotalCompressedLength += Packet->nBytesAvail;

    free(Buf);
    return TRUE;
}

PACKET_ACTION SerplexFilter(PSERPLEXSTATE Input, PPACKET Packet, PPACKET Parking, BOOL JustAddr)
{
    UINT8 *pNext, Byte;
    UINT i, j, Length = 0;
    UINT DataLength = Packet->nBytesAvail;

    /* initial state should be zero */    

#define STX_WAIT            0
#define STX_ESCAPE          1
#define ADDR_WAIT           2
#define DATA_PULL           3
#define DATA_ESCAPE         4
#define CSUM0               5
#define CSUM1               6
#define COMP_NUM_WAIT       7
#define COMP_BODY_WAIT      8

    pNext = Parking->Data + Parking->nBytesAvail;

    for (i=0; i<DataLength; i++) {
        Byte = Packet->Data[i];

        /* Got a valid byte, see whatnot
         */
        if (debug_level > 3)
            printf("B%x ",Byte);

        switch (Input->State) {

        case STX_WAIT: /* Wait for STX */
            if (Byte == STH_CHAR)
                Input->State = ADDR_WAIT;/* get data now */
            else if (Byte == ESC_CHAR)
                Input->State = STX_ESCAPE;
            //else /* else ignore till we get it */
                //printf("SYN: x%x ",Byte);
            break;

        case STX_ESCAPE:
            Input->State = STX_WAIT;
            break;

        case ADDR_WAIT:
            Input->State = DATA_PULL;
            assert(Parking->SerplexAddr == InvalidSerplexAddr);
            Parking->SerplexAddr = Byte;
            if (JustAddr)
                return PacketActionNone;
            break;

        case DATA_PULL: /* Pull data in, end at EOT */
            if (Byte == COMP_CHAR) {
                Input->State = COMP_NUM_WAIT; /* compressed data coming up */
            } else if (Byte == EOT_CHAR) {
                Input->State = CSUM0; /* done with data */
                /* Fix expected checksum on account of STX chars */
                if ((Input->Sum & 0xff) == STH_CHAR)
                    Input->Sum++;
                if ((Input->Sum >> 8) == STH_CHAR)
                    Input->Sum += (1 << 8);/* this must not rollover */
            } else if (Byte == STH_CHAR) {
                //printf("Retransmission detected..\n");
                Input->State = ADDR_WAIT;
                //printf("!r");
                goto drop;
            } else if (Byte == ESC_CHAR) {
                Input->State = DATA_ESCAPE;
            } else {
                Length++;
                *pNext++ = Byte;
                Input->Sum += Byte;
            }
            break;

        case DATA_ESCAPE:
            Length++;
            *pNext++ = Byte;
            Input->Sum += Byte;
            Input->State = DATA_PULL;
            break;

        case COMP_NUM_WAIT: /* uncompress */
            Input->ByteNum = Byte;
            Input->State = COMP_BODY_WAIT;
            Input->Sum += Byte;
            break;

        case COMP_BODY_WAIT: /* uncompress */
            for(j=0; j<Input->ByteNum; j++) {
                *pNext++ = Byte;
                Length++;
            }
            Input->Sum += Byte;
            Input->State = DATA_PULL;
            Input->ByteNum = 0;
            break;
        
        case CSUM0: /* get and verify checksum, first byte */
            if ((Input->Sum & 0xff) == Byte)
                Input->State = CSUM1;
            else {
                //printf("Bad pkt checksum, 1st byte %x != %x [%x]\n",
                  //     Byte, Input->Sum & 0xff, Input->Sum>>8);

                /* wait for retransmit
                 */
                Input->State = STX_WAIT;
                //printf("!c");
                goto drop;

            }
            break;

        case CSUM1: /* get and verify checksum, second byte */
            if ((Input->Sum >> 8) != Byte) {
                //printf("Bad pkt checksum, 2nd byte %x != %x\n",
                  //     Byte, Input->Sum & 0xff);
                Input->State = STX_WAIT;
                //printf("!x");
                goto drop;
            }

            /* NB: We cannot detect missing zeroes with the check-sum,
             * NB: should have shifted also for proper polinomial sum.
             * We'll check the number of bytes instead.
             */
            
            /* Done */
            Input->State = STX_WAIT;
            Input->Sum = 0;
            Packet->nBytesAvail -= i + 1;
            Packet->Data += i + 1;
            Parking->nBytesAvail += Length;
            return PacketActionForward;
        }
    }
    /* Fully consumed */
    Packet->nBytesAvail = 0;
    Parking->nBytesAvail += Length;
    return PacketActionNone; /* incomplete */

drop:
    Input->Sum = 0;
    /* We processed part of the incoming packet. Remember that! */
    Packet->nBytesAvail -= i;
    Packet->Data += i;
    Input->Sum = 0;
    return PacketActionDrop;
}
/*
 * Console Filter
 */
PACKET_ACTION ConsoleFilter(PNETWORK Input, PPACKET Packet)
{
    if (Packet->Data[0] == 0x03) /* Ctrl-C */
        return PacketActionTerminate;
    Packet->SerplexAddr = CONSerplexAddr;
    return PacketActionForward;
}
/*
 * Null filter
 */
PACKET_ACTION NullFilter(PNETWORK Input, PPACKET Packet)
{
    Packet->SerplexAddr = CONSerplexAddr;
    return PacketActionForward;
}

/*
 *    Filter function for serial line input packets.
 *    Must reassemble streaming data into full packets.
 */
PACKET_ACTION SerialFilter(PNETWORK Input, PPACKET Packet)
{
    PPACKET Parking;
    UINT8   SerplexAddr;
    UINT len2;

    while (Packet->nBytesAvail) {
        /* Always copy data into parking. If there is no parking, then create one.
         * This simplifies SerplexFilter and allows for expanding packets and various
         * straddles. One assumption is that Serplex frame is one to one with hostfs
         * or hostnic frame, i.e. one Serplex frame carries one hostnic or hostfs message.
         * XXX This means the extra length field is unnecessary.
         * XXX Should cache temporary buffers so that the do not continuously
         * XXX get freed and allocated.
         */
        if (Input->FilterState == NULL) {
            PacketAllocate(&Parking);
            PacketInit(Parking);
            Parking->HomeNetwork = NULL; /* temporary packet, should not be reposted */
            Input->FilterState = Parking;
        }
        Parking = Input->FilterState;
        switch(SerplexFilter(&Input->SerplexState, Packet, Parking, FALSE)) {
        case PacketActionForward:
            if (Parking->SerplexAddr == NICSerplexAddr) {
                if (Parking->nBytesAvail < 2) {
                    //printf("!s");
                    goto drop;
                }
                /* Deal with length and init packet */
                len2 = (Parking->Data[0] << 8) | (Parking->Data[1]);
                if (Parking->SerplexAddr == NICSerplexAddr && len2 == 0 && Input->ForwardingTo)
                    goto drop;
                /* Drop length */
                Parking->Data += 2;
                Parking->nBytesAvail -= 2;

                /* Check for bad length */
                if (len2 != Parking->nBytesAvail) {
                    //printf("Bad length says %d != got %d\n", len2, Parking->nBytesAvail);
                    goto drop;
                }
            }
            /* The rest is for real */
            if (Parking->nBytesAvail == 0)
                goto drop; /* Be sure it is not a NULL packet */
            ForwardPacket(Input, Parking);
            Input->FilterState = NULL;
            continue;
        case PacketActionDrop:
drop:
            /* Drop and reuse bad packet */
            SerplexAddr = Parking->SerplexAddr;
            PacketInit(Parking);
            /* But first send a resync as needed */
            if (Input->ForwardingTo && SerplexAddr == NICSerplexAddr) {
                SerialEtherinfo(Input->ForwardingTo, Input);
            } else if (SerplexAddr == FSSerplexAddr) {
                Input->FilterState = NULL;
                SendNACK(Parking, Input);
            } else if (SerplexAddr == CONSerplexAddr) {
                ;
            } 
            continue;
        case PacketActionNone:
            if (debug_level > 1)
                printf("i%d", Parking->SerplexAddr);
            continue; /* Same as return PacketActionDrop; */
        }

    }
    return PacketActionDrop;
}

/*
 *    Initialization packet for serial line driver
 */
BOOL SerialEtherinfo(PNETWORK Ethernet, PNETWORK Serial)
{
    PPACKET Packet;
  
    /* 
     *    Allocate & zero a packet
     */
  
    if (!PacketAllocate(&Packet))
    {
        //printf("Outtamem66 %x\n", GetLastError());
        return FALSE;
    }

    PacketInit(Packet);

    /*
     *     Craft that special init packet
     */
    memcpy(Packet->Ethernet.DstAddr,Ethernet->EthernetAddress,6);
    memcpy(Packet->Ethernet.SrcAddr,Ethernet->EthernetAddress,6);
    Packet->Ethernet.Etype = 0x6666;
    memcpy(&Packet->Buffer[sizeof(ETHERNET_HEADER)],
           Ethernet->EthernetAddress,6);

    Packet->nBytesAvail = sizeof(ETHERNET_HEADER) + 6;

    /*
     *     Wrap length & xsum around it.
     */

    Packet->SerplexAddr = NICSerplexAddr;

    /*
     *    Post the packet for xmit
     */
    if (debug_level > 0)
        printf("-e");
    ForwardPacket(Ethernet, Packet);

    /*
     *    Done, packet should be recycled.
     */

    return TRUE;
}

/*
 * Create Console
 */
PNETWORK ConsoleCreate(void)
{
    PNETWORK    Network;
    
    Network = malloc(sizeof(NETWORK));
    if (Network == NULL)
        return NULL;

    memset(Network, 0, sizeof(NETWORK));
    
    /*
     *    Set the filter
     */

    Network->Name = "CON:";
    Network->Filter = ConsoleFilter;
    
    Network->bInitialized = TRUE;
    InitializeCriticalSection(&Network->Section);

    return Network;
}

/*
 *    Constructor for serial line network object.
 */
PNETWORK SerialInitialize(PNETWORK Serial, BOOL Quietly)
{
    UINT InterByteTime;
    DCB ControlBlock;
    COMMTIMEOUTS Timeouts;
    HANDLE hLine, IoCompletionPort;

    /* 
     *    Open & check the comm line
     */
    hLine = CreateFile( Serial->Name, GENERIC_READ | GENERIC_WRITE,
                        FILE_SHARE_WRITE, /* exclusive access (except PnP..) */
                        NULL, /* no security attributes */
                        OPEN_EXISTING,
                        FILE_FLAG_OVERLAPPED,
                        NULL  /* hTemplate - must be NULL for comms */
                        );

    if (hLine == INVALID_HANDLE_VALUE) {
        if (!Quietly)
            printf("Could not open %s\n",Serial->Name);
        return NULL;
    }

    Serial->hFile = hLine;

    /* 
     *    Get what we wont change
     */
    ControlBlock.DCBlength = sizeof(ControlBlock);
    if (!GetCommState(hLine,&ControlBlock)) {
        printf("GetCommState() failed on %s (x%x)\n",Serial->Name,GetLastError());
        /* 
         *    might be a named pipe, ignore
         */
    }

    /* 
     *    No parity, no flow..
     */
    ControlBlock.BaudRate = Serial->BaudRate;
    ControlBlock.fBinary = TRUE;     /* no EOF check */
    ControlBlock.fParity = FALSE;
    ControlBlock.fOutxCtsFlow = FALSE;
    ControlBlock.fOutxDsrFlow = FALSE;
    ControlBlock.fDtrControl = DTR_CONTROL_ENABLE;
    ControlBlock.fDsrSensitivity = FALSE;
    ControlBlock.fTXContinueOnXoff = TRUE;
    ControlBlock.fOutX = FALSE;
    ControlBlock.fInX = FALSE;
    ControlBlock.fErrorChar = FALSE;
    ControlBlock.fNull = FALSE;
    ControlBlock.fRtsControl = RTS_CONTROL_ENABLE;
    ControlBlock.fAbortOnError = FALSE;
    ControlBlock.ByteSize = 8;
    ControlBlock.Parity = NOPARITY;
    ControlBlock.StopBits = ONESTOPBIT;
    if (!SetCommState(hLine,&ControlBlock)) {
        printf("SetCommState() failed on %s (x%x)\n",Serial->Name,GetLastError());
        /* 
         *    might be a named pipe, ignore
         */
    }

    /* 
     *    Doing overlapped I/O on the serial line under NT is funny.
     *    By default, it will wait to fillup completely the read buffer
     *    before signalling completion.  So we have to set a timeout.
     *    Timeouts are in milliseconds units (none on xmit)
     *    Luckily, we can do the optimal thing of setting a timeout that
     *    (a) only starts after the first byte is received
     *    (b) renews itself for as long as bytes keep coming in
     *    (c) expires after that much time elapses between bytes
     */

    InterByteTime = Serial->BaudRate / 10; /* approx byte rate */
    InterByteTime = 1000 / InterByteTime; /* in msecs, most likely 0 */
    InterByteTime++;  /* roundup, avoid zero */
    if (debug_level)
        printf("SerialRead inter-byte timeout is %d msecs\n",InterByteTime);
    Timeouts.ReadIntervalTimeout = InterByteTime;
    Timeouts.ReadTotalTimeoutConstant = 0;
    Timeouts.ReadTotalTimeoutMultiplier = 0;

    Timeouts.WriteTotalTimeoutMultiplier = 0;
    Timeouts.WriteTotalTimeoutConstant = 0;

    if (!SetCommTimeouts( hLine, &Timeouts)) {
        printf("SetCommTimeouts() failed (x%x), ignoring..\n",
               GetLastError());
    }

    /*
     *    Create a single I/O completion port for send and
     *    receive operations to and from the three interfaces.
     */
    IoCompletionPort = Serial->IoCompletionPort;
    Serial->IoCompletionPort = CreateIoCompletionPort(Serial->hFile,
                                                      NULL,
                                                      (ULONG) Serial,
                                                      0);
    assert(Serial->IoCompletionPort != NULL);

    /*
     *    When this is called on a cable-plugged-in event we need to
     *    wakeup the thread waiting on the old IoCompletionPort.
     *    Send it a NULL completion.
     */
    if (IoCompletionPort) {
        PostQueuedCompletionStatus(IoCompletionPort,
                                   0,
                                   (ULONG_PTR) Serial,
                                   NULL);
    }

    /*
     *    Done initializing
     */
    Serial->bInitialized = TRUE;

    return Serial;
}

void SerialFinalize( PNETWORK Serial)
{
    CloseHandle(Serial->hFile);
    //CloseHandle(Serial->IoCompletionPort);
    Serial->hFile = INVALID_HANDLE_VALUE;
    Serial->bInitialized = FALSE;
}

PNETWORK SerialCreate(char *Line, UINT BaudRate)
{
    PNETWORK Network;

    /*
     *    Allocate and initialize descriptor
     */
    Network = malloc(sizeof(NETWORK));
    if (Network == NULL) {
        printf("Outtamem?!? x%x\n", GetLastError());
        return NULL;
    }

    memset(Network, 0, sizeof(NETWORK));
    
    Network->Name = Line;
    Network->Filter = (RawMode) ? NullFilter : SerialFilter;
    Network->BaudRate = BaudRate;

    if (!SerialInitialize(Network,FALSE)) {
        free(Network);
        return NULL;
    }

    /* 
     *    Leave address empty. 
     *    Could snoop on the line later but so what.
     */

    return Network;
}

BOOL PostReads(PNETWORK Network, UINT Count)
{
    UINT i;
    BOOL ok;
    PPACKET Packet;
    for (i = 0; i < Count; i++) {
        ok = PacketAllocate(&Packet);
        if (!ok) {
            printf("PostReads failed\n");
            return FALSE;
        }

        assert(Packet != NULL);
        PacketInit(Packet);
        Packet->HomeNetwork = Network;
        PacketReadFile(Network, Packet);
    }
    return TRUE;
}

/* 
 *    Global data
 */
char *Usage = "%s [-b<BAUDRATE>] [-verbose] [-filter] [-p<PORTNUMBER>] [-i<PACKET_INTERVAL>] [-no-ether] [-raw-mode] [-swap-endian] [-old-packet-driver] [COM1|COM2|...]";
int debug_level = 0;

#define PKT_INTERVAL 500
UINT32 pkt_interval = PKT_INTERVAL;

BOOL Unfiltered = FALSE;

PNETWORK Ethernet = NULL, Serial = NULL, Console = NULL;

void SetCompletionPorts(void)
{
    if (Ethernet) {
        HANDLE IoCompletionPort = Ethernet->IoCompletionPort;

        if (IoCompletionPort) {
            CloseHandle(IoCompletionPort);
        }

        Ethernet->IoCompletionPort = CreateIoCompletionPort(Ethernet->hFile,
                                                           Serial->IoCompletionPort,
                                                           (ULONG) Ethernet, 
                                                           0);

        if (Ethernet->IoCompletionPort == NULL) {
            printf("I did say it was shakey (%x)\n", GetLastError());
            exit(1);
        }

        assert(Ethernet->IoCompletionPort != NULL);

        assert(Serial->IoCompletionPort == Ethernet->IoCompletionPort);
    }

    /* CreateIoCompletionPort is broken for the console. Just assign it. */
    Console->IoCompletionPort = Serial->IoCompletionPort;

}


/* 
 *    Program's entry point
 */
int main(int argc, char **argv)
{
    char *Line, *pgmname = argv[0];
    UINT BaudRate, PortNumber;
    HANDLE hConThread = NULL;
    BOOL IsUsb = FALSE;

    Line = "COM1";
    BaudRate = 38400;
    PortNumber = 2;
    /* 
     *    Parse invocation arguments, if any
     */
    while (argc > 1) {
        if (argv[1][0] == '-') {
            switch (argv[1][1]) {
            case 'b':
            case 'B':
                BaudRate = atoi(argv[1]+2);
                printf("Baudrate will be %d baud\n", BaudRate);
                break;                
            case 'E':
            case 'e':
                doSendError = TRUE;
                break;
            case 'f':
            case 'F':
                Unfiltered = !Unfiltered;
                printf("Will %sfilter ether traffic.\n",
                       (Unfiltered ? "NOT " : ""));
                break;                
            case 'i':
            case 'I':
                pkt_interval = atoi(argv[1]+2);
                printf("Will print statistics every %d packets\n", pkt_interval);
                break;                
            case 'N':
            case 'n':
                printf("Will NOT attempt to use the NIC\n");
                NoEther = TRUE;
                break;
            case 'O':
            case 'o':
                printf("Will use the old packet filter\n");
                OldEther = TRUE;
                break;
            case 'p':
            case 'P':
                PortNumber = _ttoi(argv[1]+2);
                printf("Will use port no. %d\n", PortNumber);
                break;
            case 'r':
            case 'R':
                RawMode = TRUE;
                printf("Will exchange un-encoded data ['raw' mode].\n");
                break;
            case 'S':
            case 's':
                doByteswap = TRUE;
                printf("Will talk to a BigEndian client (mips,ppc,...)\n");
                break;
            case 'V': debug_level++;/*fallthrough*/
            case 'v': debug_level++;
                printf("debug_level = %d\n", debug_level);
                break;
            default:
                printf(Usage,pgmname);
                exit(-1);
            }
        } else {
            Line = argv[1];
        }
        argc--,argv++;
    }

    /*
     *    If we use a USB cable we must be ready for unplug/replug events.
     *    Infact, the cable might not be plugged in yet!
     *    BUGBUG Too many things depend on Serial->hFile at init time. Sigh.
     */
    if (_strnicmp(Line,"\\\\.\\wceusbsh",12) == 0)
        IsUsb = TRUE;

    /* 
     *    Create the serial line first
     */

    Serial = SerialCreate(Line,BaudRate);
    if (Serial == NULL)
    {
        printf("SerialCreate FAILED(x%x)\n", GetLastError());
        if (IsUsb)
            printf("Plug in the USB cable first, then start me.\n");
        goto Done;
    }

    /*
     *    Warn user of some unsafe practices :-)
     */
    if (IsUsb && !NoEther)
        printf("USB+Ethernet combo is unsafe, do not unplug the USB cable.\n");

    /*
     *    Start the Pnp thread
     */
    if (IsUsb) {
        HANDLE Thread;
        Thread = CreateThread(NULL,
                              0,
                              (LPTHREAD_START_ROUTINE) PnpThread,
                              Serial,
                              0,
                              0);
    }

    /* 
     *    Create the ethernet packet filter next
     */

    if (!NoEther) {
        Ethernet = EthernetCreate(OldEther);
        if (Ethernet == NULL)
        {
            printf("EthernetCreate(%sDriver) FAILED(x%x)\n", 
                   (OldEther) ? "Old" : "VirtualPC",
                   GetLastError());
            goto Done;
        }
    }

    if (debug_level && Ethernet)
    printf(_T("HostNic CardAddr %02x:%02x:%02x:%02x:%02x:%02x\n"),
           Ethernet->EthernetAddress[0], Ethernet->EthernetAddress[1],
           Ethernet->EthernetAddress[2], Ethernet->EthernetAddress[3],
           Ethernet->EthernetAddress[4], Ethernet->EthernetAddress[5]);

    /*
     *    Create console
     */
    Console = ConsoleCreate();
    if (Console == NULL)
    {
        printf("ConsoleCreate FAILED(x%x)\n", GetLastError());
        goto Done;
    }

    /*
     *    Criss-cross the three interfaces
     */

    if (Ethernet) {
        Ethernet->ForwardingTo = Serial;
    }
    Serial->ForwardingTo = Ethernet;
    Console->ForwardingTo = Serial;
    
    /*
     *     Point everyone to the same completion port
     */
    SetCompletionPorts();

    /*
     *     Post the MMLite serial driver's initialization packet
     */
    if (Ethernet) {
        if (!SerialEtherinfo(Ethernet,Serial))
        {
            printf("Failed to post Serial Etherinfo, terminating.\n");
            goto Done;
        }
    }

    /*
     *     Post all the receives, Equal number on each interface
     */

    PostReads(Serial, MAX_PACKETS_PER_NETWORK);
    if (Ethernet)
        PostReads(Ethernet, MAX_PACKETS_PER_NETWORK);
    PostReads(Console, MAX_PACKETS_PER_NETWORK);

    /*
     *    Create the console thread
     */
    hConThread = CreateThread(NULL,
                              0,
                              (LPTHREAD_START_ROUTINE) ConsoleReadThread,
                              Console,
                              0,
                              0);
    
    /*
     *    Now become the network thread
     */
    NetworkThread(Serial);

Done:
    /* 
     *    Termination code.
     *    BUGBUG Should free all memory but.
     */
    if (Serial)
        CloseHandle(Serial->hFile);
    if (Ethernet)
        CloseHandle(Ethernet->hFile);

    if (hConThread)
        CloseHandle(hConThread);

    return 0;
}

void PacketDrop(PPACKET Packet)
{
    if (Packet->HomeNetwork) {
        PacketInit(Packet);
        PacketReadFile(Packet->HomeNetwork, Packet);
    } else {
        PacketFree(Packet);
    }
}

/*
 *    Function: NetworkThread()
 *
 *    Description: Send and receive thread.
 *
 */
void NetworkThread(IN void * ThreadContext)
{
    PNETWORK        Network;
    HANDLE          IoCompletionPort;
    UINT32          Transferred;
    UINT32          BytesReceived, BytesForwarded;
    UINT32          PacketsReceived, PacketsDropped;
    PACKET *        Packet;
    BOOL            bResult;
    PACKET_ACTION   aResult;
    OVERLAPPED *    Overlapped;
    ULONG           Key;
    ULONG           CurrentTime, ElapsedTime;
    int             i;

    if (debug_level)
        printf("NetworkThread.\n");

    Network = (PNETWORK) ThreadContext;

    /*
     *    We have to run a little faster than normal 
     *    in order to handle a lot of network packets.
     */

    SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);

    /*
     *     We are ready to roll
     */

    Network->bInitialized = TRUE;

    /*
     *    The loop below is the main processing loop for all
     *    Receives and transmits. 
     */

    if (debug_level)
        printf("Stats every %d packets received:\n", pkt_interval);

    ElapsedTime = GetTickCount();
    BytesForwarded = BytesReceived = 0;
    PacketsReceived = PacketsDropped = 0;
    IoCompletionPort = Network->IoCompletionPort;
    for (i=0;;) 
    {
        /*
         *    Wait for an I/O to complete.
         */
        
        Packet = NULL;
        Overlapped = NULL;

        bResult = GetQueuedCompletionStatus(
                            IoCompletionPort,
                            &Transferred,
                            &Key,
                            &Overlapped,
                            INFINITE
                            );

        /*
         *    If the completion was ok, move the packet to the 
         *    other interface for forwarding/receiving.
         */

        Packet = (PACKET *) Overlapped;
        Network = (PNETWORK) Key;
        
        //rintf("C %s P%x[%d]\n",(Network) ? Network->Name : "boom",Packet,bResult);
        if ( bResult != FALSE && Packet != NULL)
        {
            assert(Packet->Mode != PacketModeInvalid);

            /* 
             *    Is this a packet we just received ?
             */

            if ( Packet->Mode == PacketModeReceiving )
            {
                /*
                 *    Finish initializing pending packet. Keep stats.
                 */    
    
                Packet->nBytesAvail = Transferred;
                Packet->Result = S_OK;

                PacketsReceived++;
                BytesReceived += Transferred;

                /*
                 *    Check the filter to see if we should drop it or what
                 */

                aResult = Network->Filter(Network, Packet);
                switch (aResult)
                { 
                case PacketActionForward:
                    /*
                     *    Move it on to the other interface.
                     */
                    ForwardPacket(Network, Packet);
                    
                    break;

                case PacketActionDrop:
                    /*
                     *    Repost it, keep stats
                     */

                    PacketsDropped++;

                    PacketInit(Packet);

                    PacketReadFile(Packet->HomeNetwork, Packet);
                    break;
                case PacketActionNone:
                    /*
                     *    Filter already acted on it, ignore.
                     */

                    break;
                case PacketActionTerminate:
                    /*
                     *  Ctrl-C is entered, so terminate.
                     */
                    return;

                default:
                    /*
                     *    Shouldnt happen
                     */

                    assert(FALSE);

                }
            }
            /* 
             *    Or is this a packet we just sent
             */
            else
            {
                assert(Packet->Mode == PacketModeTransmitting);
                /*
                 *    Statistics
                 */

                i++;
                BytesForwarded += Transferred;

                /*
                 *    Move the packet back to receiving on the other iface,
                 *    unless its one of those extras over budget.
                 */

                PacketDrop(Packet);
            }
        }

        /* 
         *    If the packet was completed in error it might be due to PnP
         */
        if ( bResult == FALSE && Packet != NULL)
        {
            if ( !Network->bInitialized )
                ; /* cable unplugged, waiting for replug */
            /*
             *    Either way, free up the packet
             */
            PacketFree(Packet);
        }        

        /*
         *    We do not have a packet; it might be a post-PnP wakeup
         */
        if ( (Packet == NULL) && (Network != NULL) )
        {
            printf("[You can resume typing now]\n");
            IoCompletionPort = Network->IoCompletionPort;
        }

        /*
         *    Othewise we are so confused.
         */
        if ( (Packet == NULL) && (Network == NULL) )
        {
            assert(FALSE);
        }

        /*
         *    Statistics: update every PKT_INTERVAL packets received
         */

        if ((debug_level) &&
            (PacketsReceived == pkt_interval))
        {
            /*
             *    See how long this interval was.
             */
            CurrentTime = GetTickCount();
            ElapsedTime = CurrentTime - ElapsedTime;

            /*
             *    Compute and print the stats
             */
            printf("RECV %.2f pps %.1f bps DROP %.2f pps FORWARD %.2f pps %.1f bps\n",
                   (double) (PacketsReceived * 1000) / (double) ElapsedTime,
                   (double) (BytesReceived * 8 * 1000) / (double) ElapsedTime,
                   (double) (PacketsDropped * 1000) / (double) ElapsedTime,
                   (double) (i * 1000) / (double) ElapsedTime,
                   (double) (BytesForwarded *8*1000) / (double) ElapsedTime);

            /*
             *    Get set for new round
             */
            i = 0;
            ElapsedTime = CurrentTime;
            BytesReceived = BytesForwarded = 0;
            PacketsReceived = PacketsDropped = 0;
        }
    }

    /*
     *     Should we ever terminate cleanly..
     */

    Network->bInitialized = TRUE;
}


void ConsoleReadThread(void *Context)
{
    HANDLE StdConIn;
    PNETWORK Console = (PNETWORK)Context;
    UINT nBytesTransferred;
    DWORD OldMode = 0, NewMode;
    
    printf("Console Thread ...\n");
    StdConIn = GetStdHandle(STD_INPUT_HANDLE);
 
    /* Modify console mode */
    if ( GetConsoleMode(StdConIn, &OldMode) == FALSE ) {
        assert(FALSE);
        return;              /* EOF */
    }

    NewMode = OldMode & ~(ENABLE_LINE_INPUT
                          | ENABLE_ECHO_INPUT
                          | ENABLE_WINDOW_INPUT
                          | ENABLE_MOUSE_INPUT);
    NewMode = 0;

    if ( SetConsoleMode(StdConIn, NewMode) == FALSE ) {
        assert(FALSE);
        return;
    }

    
    while (1) {
        BOOL ok;
        PPACKET Packet;
        INPUT_RECORD Input;
        DWORD count;
        int c;

        EnterCriticalSection(&Console->Section);
        Packet = Console->ReadQueue;
        if (!Packet) {
            LeaveCriticalSection(&Console->Section);
            Sleep(1);
            continue;
        }
        Console->ReadQueue = (PPACKET) Packet->Overlapped.Internal;
        LeaveCriticalSection(&Console->Section);

#if 1
        for (;;) {
            ok = ReadConsoleInput(StdConIn, &Input, 1, &count);
            if (!ok || count != 1) { /* EOF */
                printf("Console EOF\n");
                DebugBreak();
            }
            if (Input.EventType == KEY_EVENT
                && Input.Event.KeyEvent.bKeyDown == TRUE)
            {
                c = Input.Event.KeyEvent.uChar.AsciiChar;
                if (c != 0)
                    break;
            }
        }
        Packet->Data[0] = c;
        nBytesTransferred = 1;
#else
        ok = ReadFile(StdConIn, 
                      Packet->Data, 
                      Packet->Length, 
                      &nBytesTransferred, 
                      NULL /* ignored */
                      );
#endif

        ok = PostQueuedCompletionStatus(
                Console->IoCompletionPort,
                nBytesTransferred,
                (ULONG_PTR) Console,
                &Packet->Overlapped);
    }

    /* Restore mode. Makes control c work again. */
    //SetConsoleMode(StdConIn, OldMode);
}
