/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* 
 *    Simulator network driver definitions 
 */

#include <import_windows.h>
#include <mmlite.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <sock.h>
#include <if.h>
#include <if_types.h>
#include <ip/in.h>
#include <ip/arp.h>
#include "ntddpack.h"
#include <drivers/serplex.h>

/* 
 *    Definition of an 802.3 ethernet packet
 */
#pragma pack(1)
typedef struct _ETHERNET_HEADER {
    UINT8 DstAddr[6];
    UINT8 SrcAddr[6];
    union {
        UINT16 Etype;
        UINT16 Length;
    };
} ETHERNET_HEADER, *PETHERNET_HEADER;
#pragma pack()

#define MIN_ETHERNET_PACKET_SIZE 14
#define MAX_ETHERNET_DATA_SIZE 1500
#define SERIAL_OVERHEAD 2
#define MAX_ETHERNET_PACKET_SIZE (MIN_ETHERNET_PACKET_SIZE + MAX_ETHERNET_DATA_SIZE + SERIAL_OVERHEAD)

/* 
 *    Packet descriptor, internal format
 */
typedef enum _PACKET_MODE {
    PacketModeInvalid = 0,
    PacketModeReceiving = 1,
    PacketModeTransmitting = 2,
} PACKET_MODE;

typedef struct _PACKET PACKET, *PPACKET;

typedef struct _PACKET {
    OVERLAPPED  Overlapped;                     
    PACKET_MODE Mode;
    SCODE       Result;
    UINT32      Length;
    UINT32      nBytesAvail;
    UINT8       *Data;
    UINT8       SerplexAddr;
    struct _NETWORK *HomeNetwork;
    UINT32      SerialLength;
    union {
        UINT8   Buffer[MAX_ETHERNET_PACKET_SIZE];
        ETHERNET_HEADER Ethernet;
    };
};

/* 
 *    A constant to define our budget of packets
 */
#define MAX_PACKETS_PER_NETWORK           16

/* 
 *    Packet filtering functions
 */
typedef enum _PACKET_ACTION {
    PacketActionForward = 0,
    PacketActionDrop = 1,
    PacketActionNone = 2,
    PacketActionTerminate = 3
} PACKET_ACTION;

typedef struct _NETWORK *PNETWORK;
typedef PACKET_ACTION (*PACKET_FILTER)(PNETWORK Input, PPACKET Packet);

/* 
 *    Network interface
 */
typedef struct _NETWORK {
    BOOL bInitialized;

    /* NT handles for sending&receiving. */
    char *Name;
    HANDLE hFile;
    HANDLE IoCompletionPort;
    HANDLE hThread;
    /* The console doesn't do async so this queue simulates that */
    PPACKET ReadQueue;
    CRITICAL_SECTION Section;

    /* Where data is forwarded */
    PNETWORK ForwardingTo;

    /* Filter function and state */
    PACKET_FILTER Filter;
    void *FilterState;

    /* The network speed&address. */
    ULONG BaudRate;
    UINT8 EthernetAddress[6];

    SERPLEXSTATE SerplexState;

} NETWORK, * PNETWORK;

/* 
 *    Constructors
 */
extern PNETWORK EthernetCreate(BOOL OldDriver);
extern PNETWORK SerialCreate(char *Line,UINT BaudRate);
extern PNETWORK ConsoleCreate(void);

extern void NetworkThread(void * ThreadContext);
extern void ConsoleReadThread(void *Context);
extern BOOL PnpThread( PNETWORK Serial );

/*
 *    Destructors
 */
extern void SerialFinalize( PNETWORK Serial);

/* 
 *    Packet class methods
 */
extern void PacketInit(PACKET * Packet);
extern BOOL PacketAllocate(PACKET ** Packet);
extern BOOL PacketFree(PACKET * Packet);
extern void PacketDump(PACKET * Packet, char * Headline);
extern SCODE PacketWriteFile(PNETWORK Network, PACKET * Packet);
extern SCODE PacketReadFile(PNETWORK Network, PACKET * Packet);
extern BOOL PostReads(PNETWORK Network, UINT Count);

/*
 *    Serplex methods
 */
extern BOOL SerplexWrap(PPACKET Packet, UINT8);
extern PACKET_ACTION SerplexFilter(PSERPLEXSTATE Input, PPACKET Packet, PPACKET Parking, BOOL JustAddr);

/*
 *    Serial class methods
 */
extern PNETWORK SerialInitialize(PNETWORK Serial, BOOL Quietly);
extern BOOL SerialEtherinfo(PNETWORK Ethernet, PNETWORK Serial);
extern PACKET_ACTION SerialFilter(PNETWORK Input, PPACKET Packet);

/*
 *    Console method
 */
extern PACKET_ACTION ConsoleFilter(PNETWORK Input, PPACKET Packet);
extern void SetCompletionPorts(void);

/* 
 *    Options/debug/trace
 */
extern int debug_level;
extern BOOL Unfiltered;

/* Hostfile glue */
UINT HostFile_Server(void *handle, struct packet *p, UINT Size);
void ForwardPacket(PNETWORK in, PPACKET Packet);

#define MAX_CON_INPUT_BUFFER_SIZE 128
typedef struct _CON_INPUT_BUFFER {
    UINT8 Buf[MAX_CON_INPUT_BUFFER_SIZE];
    UINT Consumed;
    UINT Avail;
}CONINBUFF, *PCONINBUFF;
