/* Copyright (c) Microsoft Corporation. All rights reserved. */
//=============================================================================
//    Module: packet.c
//
//    Description: This file contains routines to use the packet filter driver.
//                 
//    
//=============================================================================

#include <drivers/serp.h>
#include "serplexd.h"

//=============================================================================
//    Section: Packet::
//
//    Description: Exported functions to operate on PACKET objects.
//=============================================================================


//=============================================================================
//    Function: PacketInit()
//
//    Description: (Re)Initialize a PACKET object.
//=============================================================================

void
PacketInit(    
    IN PACKET * Packet
    )
{
    //
    //    Point DATA back to the beginning, init other key fields
    //

    Packet->Data = Packet->Buffer;
    Packet->nBytesAvail = 0;
    Packet->Length = MAX_ETHERNET_PACKET_SIZE;
    Packet->Mode = PacketModeInvalid;
    Packet->SerplexAddr = InvalidSerplexAddr;
}

//=============================================================================
//    Function: PacketAllocate()
//
//    Description: Allocate and initialize a new packet.
//=============================================================================

BOOL
PacketAllocate(    
    IN PACKET ** Packet
    )
{
    //
    //    Allocate a packet. 
    //
        
    *Packet = malloc(sizeof(PACKET));

    if ( *Packet != NULL )
    {
        memset(*Packet, 0, sizeof(PACKET));
        PacketInit(*Packet);

        return TRUE;
    }

    return FALSE;
}

//=============================================================================
//    Function: PacketFree()
//
//    Description: Destroy a packet.
//=============================================================================

SCODE
PacketFree(
    IN PACKET * Packet
    )
{
    free(Packet);
    return S_OK;
}

//=============================================================================
//    Function: PacketDump()
//
//    Description: Print the full content of a packet
//=============================================================================

void 
PacketDump(
    IN PPACKET Pkt,
    IN char *Headline
    )
{
    UINT32 i;

 //   if (debug_level < 3)
   //     return;
    printf(Headline);
    for (i = 0;i < Pkt->nBytesAvail; i++)
        printf(" %x",Pkt->Data[i]);
    printf("\n");
}


//=============================================================================
//    Function: PacketWriteFile()
//
//    Description: Write a packet using NT WriteFile(). 
//=============================================================================

SCODE
PacketWriteFile(    
    PNETWORK Network,
    IN PACKET * Packet
    )
{
    SCODE    Result = S_OK;
    BOOL     bResult = FALSE;
    UINT32   nBytesTransferred = 0;
    
    //
    //    Call NT WriteFile().
    //

    Packet->Result = E_IO_PENDING;
 
    Packet->Mode = PacketModeTransmitting;

    /* Be sure it is not a NULL packet!!!! */
    if (Packet->nBytesAvail == 0)
        return E_INVALID_PARAMETER;

    bResult = WriteFile(
                    Network->hFile, 
                    Packet->Data, 
                    Packet->nBytesAvail, 
                    &nBytesTransferred, 
                    &Packet->Overlapped
                    );

    //
    //    The request cannot possibly complete immediately.
    //
    Result = GetLastError();
    if (( bResult == FALSE ) &&
        ( Result  == ERROR_IO_PENDING ) )
    {
        return E_IO_PENDING;
    }

    return Result;
}

//=============================================================================
//    Function: PacketReadFile()
//
//    Description: Read (post) a packet using NT ReadFile(). 
//=============================================================================

SCODE
PacketReadFile(    
    PNETWORK Network,
    IN PACKET * Packet
    )
{
    SCODE    Result = S_OK;
    BOOL    bResult = FALSE;
    UINT32    nBytesTransferred = 0;


    //
    //    Call NT ReadFile().
    //

    Packet->nBytesAvail = 0;
    Packet->Result = E_IO_PENDING;

    Packet->Mode = PacketModeReceiving;

    if (Network->hFile == NULL)
    {
        /* Post it on internal queue instead */
        EnterCriticalSection(&Network->Section);
        Packet->Overlapped.Internal = (ULONG_PTR) Network->ReadQueue;
        Network->ReadQueue = Packet;
        LeaveCriticalSection(&Network->Section);
        return S_OK;
    }

    bResult = ReadFile(
                    Network->hFile, 
                    Packet->Data, 
                    Packet->Length, 
                    &nBytesTransferred, 
                    &Packet->Overlapped
                    );
    
    //
    //    The request cannot possibly complete immediately.
    //
    Result = GetLastError();
    if (( bResult == FALSE ) &&
        ( Result  == ERROR_IO_PENDING ) )
    {
        return E_IO_PENDING;
    } 

    return Result;
}


//=============================================================================
//    Section: Packet_Driver_Vtable::
//
//    Description: Interface to the NDIS filter driver implementation
//=============================================================================

typedef struct _PACKET_DRIVER_VTABLE {
    BOOLEAN (*GetSymbolicName)(
                               OUT UINT16 * SymbolicName
                               );
    HANDLE  (*OpenAdapter)    (
                               UINT16 * SymbolicName
                               );
    BOOL    (*Request)        (
                               IN HANDLE hFile,
                               IN BOOL Set,
                               IN PPACKET_OID_DATA OidData
                               );
    SCODE   (*SetFilter)      (
                               IN HANDLE hFile,
                               IN ULONG  Filter
                               );
    SCODE   (*GetAddress)     (
                               IN HANDLE hFile,
                               IN UINT8 *EthernetAddress
                               );
    SCODE   (*GetSpeed)       (
                               IN HANDLE hFile,
                               OUT ULONG *Speed
                               );
} PACKET_DRIVER_VTABLE, *PPACKET_DRIVER_VTABLE;

//=============================================================================
//    SubSection: PacketDriver::
//
//    Description: Old, homegrown, exclusive-use NDIS filter driver interface
//=============================================================================


//=============================================================================
//    Function: PacketDriverGetSymbolicName().
//
//    Description: Get the NT symbolic name for the packet driver.
//=============================================================================

static
BOOLEAN
PacketDriverGetSymbolicName(
    OUT UINT16 * SymbolicName
    )
{
    void *        hFile;
    UINT32        BytesReturned;
    UINT16        Buffer[1024];
    UINT16       *pAdapterInfo;
    BOOL          bResult;

    memset(Buffer, 0, sizeof(Buffer));

    wsprintfW(
        SymbolicName,
        L"\\\\.\\Packet");
    
    hFile = CreateFileW(
                SymbolicName,
                GENERIC_WRITE | GENERIC_READ,
                0,
                NULL,
                OPEN_EXISTING,
                FILE_FLAG_OVERLAPPED,
                0
                );

    if ( hFile == INVALID_HANDLE_VALUE) 
    {
        return FALSE;
    }
    
    bResult = DeviceIoControl(
                        hFile,
                        IOCTL_ENUM_ADAPTERS,
                        NULL,
                        0,
                        Buffer,
                        sizeof(Buffer),
                        &BytesReturned,
                        NULL
                        );

    if ( bResult == FALSE )
    {
        return FALSE;
    }

    pAdapterInfo = (UINT16 *) Buffer;

    //
    // Skip the number of adapters, we'll just take the first one.
    // Making sure there is one..
    //

    if (*pAdapterInfo == 0)
    {
        return FALSE;
    }

    pAdapterInfo += 2; /* a one-wchar 'string' */

    //
    //    Skip the adapter name.
    //

    while( *pAdapterInfo != (UINT16) '\0' )
    {
        pAdapterInfo++;
    }

    pAdapterInfo++;        //... Skip the NULL.

    //
    //    Copy the symbolic name.
    //

    while( *pAdapterInfo != (UINT16) '\0' )
    {
        *SymbolicName++ = *pAdapterInfo++;
    }

    *SymbolicName++ = (UINT16) '\0';

    CloseHandle(hFile);

    return TRUE;
}

//=============================================================================
//    Function: PacketDriverOpenAdapter().
//
//    Description: Open the packet driver.
//=============================================================================

static
HANDLE
PacketDriverOpenAdapter(
    UINT16 * SymbolicName
    )
{
    UINT16    SymbolicLink[256];
    HANDLE    hFile;

    wsprintfW(
        SymbolicLink,
        L"\\\\.\\%s",
        &SymbolicName[12]
        );

    hFile = CreateFileW(
                SymbolicLink,
                GENERIC_WRITE | GENERIC_READ,
                FILE_SHARE_READ | FILE_SHARE_WRITE,
                NULL,
                OPEN_EXISTING,
                FILE_FLAG_OVERLAPPED,
                0
                );

    if (hFile != INVALID_HANDLE_VALUE)
        sprintf(SymbolicName,"%ls",SymbolicLink);
    return hFile;
}

//=============================================================================
//    Function: PacketDriverRequest().
//
//    Description: Submit a control request to the packet driver.
//=============================================================================

static
BOOL
PacketDriverRequest(
    IN HANDLE hFile,
    IN BOOL Set,
    IN PPACKET_OID_DATA OidData
    )
{
    UINT    BytesReturned;
    BOOL    bResult;
    DWORD    OidCommand;
    DWORD    OidLength;

    OidLength = sizeof(PACKET_OID_DATA) - 1 + OidData->Length;

    if ( Set != FALSE )
    {
        OidCommand = IOCTL_PROTOCOL_SET_OID;
    }
    else
    {
        OidCommand = IOCTL_PROTOCOL_QUERY_OID;
    }
    
    //
    //    Submit request.
    //

    bResult = DeviceIoControl(
                              hFile,
                              OidCommand,
                              OidData,
                              OidLength,
                              OidData,
                              OidLength,
                              &BytesReturned,
                              NULL
                              );

    return bResult;
}

//=============================================================================
//    Function: PacketDriverSetFilter().
//
//    Description: Set capture filters.
//=============================================================================

static
SCODE
PacketDriverSetFilter(
    IN HANDLE hFile,
    IN ULONG  Filter
    )
{
    ULONG               IoCtlBufferLength = sizeof(PACKET_OID_DATA)-1 + sizeof(ULONG);
    BYTE                IoCtlBuffer[sizeof(PACKET_OID_DATA)-1 + sizeof(ULONG)];
    PPACKET_OID_DATA    OidData = NULL;
    BOOL                bResult;

    memset(IoCtlBuffer, 0, IoCtlBufferLength);
    OidData = (PPACKET_OID_DATA) IoCtlBuffer;
    OidData->Oid = OID_GEN_CURRENT_PACKET_FILTER;
    OidData->Length = sizeof(ULONG);

    memcpy(OidData->Data, &Filter, sizeof Filter);

    bResult = PacketDriverRequest(hFile, TRUE, OidData);

    return (bResult ? S_OK : E_FAIL);
}

//=============================================================================
//    Function: PacketDriverGetAddress().
//
//    Description: Get the card's ethernet address.
//=============================================================================

static
SCODE 
PacketDriverGetAddress(
    IN HANDLE hFile,
    IN UINT8 *EthernetAddress
    )
{
    ULONG               IoCtlBufferLength = sizeof(PACKET_OID_DATA)-1 + 6;
    BYTE                IoCtlBuffer[sizeof(PACKET_OID_DATA)-1 + 6];
    PPACKET_OID_DATA    OidData = NULL;
    BOOL                bResult;

    memset(IoCtlBuffer, 0, IoCtlBufferLength);
    OidData = (PPACKET_OID_DATA) IoCtlBuffer;
    OidData->Oid = OID_802_3_PERMANENT_ADDRESS;
    OidData->Length = 6;

    bResult = PacketDriverRequest(hFile, FALSE, OidData);

    memcpy(EthernetAddress, OidData->Data, 6);

    return (bResult ? S_OK : E_FAIL);
}


//=============================================================================
//    Function: PacketDriverGetSpeed().
//
//    Description: Get the link speed of the interface below.
//=============================================================================

static
SCODE 
PacketDriverGetSpeed(
    IN HANDLE hFile,
    OUT ULONG *Speed
    )
{
    ULONG               IoCtlBufferLength = sizeof(PACKET_OID_DATA)-1 + sizeof(ULONG);
    BYTE                IoCtlBuffer[sizeof(PACKET_OID_DATA)-1 + sizeof(ULONG)];
    PPACKET_OID_DATA    OidData = NULL;
    BOOL                bResult;

    memset(IoCtlBuffer, 0, IoCtlBufferLength);
    OidData = (PPACKET_OID_DATA) IoCtlBuffer;
    OidData->Oid = OID_GEN_LINK_SPEED;
    OidData->Length = sizeof(ULONG);

    bResult = PacketDriverRequest(hFile, FALSE, OidData);

    if (!bResult)
    {
        printf("SPEED  didnt work.. sorry (%x)\n",GetLastError());
       *Speed = 0; // donno

    } else
        memcpy(Speed, OidData->Data, sizeof(ULONG));

    return (bResult ? S_OK : E_FAIL);
}

//=============================================================================
//    Vtable: PacketDriverVtable[].
//
//    Description: Interface to old, exclusive-use packet driver
//=============================================================================
static const PACKET_DRIVER_VTABLE PacketDriverVtable = {
    PacketDriverGetSymbolicName,
    PacketDriverOpenAdapter,
    PacketDriverRequest,
    PacketDriverSetFilter,
    PacketDriverGetAddress,
    PacketDriverGetSpeed
};

//=============================================================================
//    SubSection: VirtualpcDriver::
//
//    Description: VirtualPC's shared-use NDIS filter driver
//=============================================================================

#define NETSV_DRIVER_NAME					L"\\\\.\\VPCNetS2"

#define	kVPCNetSvVersionMajor	2
#define	kVPCNetSvVersionMinor	6
#define	kVPCNetSvVersion		((kVPCNetSvVersionMajor << 16) | kVPCNetSvVersionMinor)

#define FILE_DEVICE_PROTOCOL				0x8000

enum
{
	kIoctlFunction_SetOid					 = 0,
	kIoctlFunction_QueryOid,
	kIoctlFunction_Reset,
	kIoctlFunction_EnumAdapters,
	kIoctlFunction_GetStatistics,
	kIoctlFunction_GetVersion,
	kIoctlFunction_GetFeatures,
	kIoctlFunction_SendToHostOnly,
	kIoctlFunction_RegisterGuest,
	kIoctlFunction_DeregisterGuest,
	kIoctlFunction_CreateVirtualAdapter,
	kIoctlFunction_DestroyVirtualAdapter,
	kIoctlFunction_GetAdapterAttributes,
	kIoctlFunction_Bogus					 = 0xFFF
};

// These IOCTLs apply only to the control object
#if 0
#define IOCTL_ENUM_ADAPTERS					CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_EnumAdapters,				METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif
#define IOCTL_GET_VERSION					CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_GetVersion,				METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_GET_FEATURES					CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_GetFeatures,				METHOD_BUFFERED, FILE_ANY_ACCESS)

// These IOCTLs apply only to the adapter object
#if 0
#define IOCTL_PROTOCOL_RESET				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_Reset,					METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif
#define IOCTL_CREATE_VIRTUAL_ADAPTER		CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_CreateVirtualAdapter,		METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_DESTROY_VIRTUAL_ADAPTER		CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_DestroyVirtualAdapter,	METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_GET_ADAPTER_ATTRIBUTES		CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_GetAdapterAttributes,		METHOD_BUFFERED, FILE_ANY_ACCESS)

// These IOCTLs apply only to the virtual adapter object
#if 0
#define IOCTL_PROTOCOL_SET_OID				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_SetOid,					METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_PROTOCOL_QUERY_OID			CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_QueryOid,					METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif
#define IOCTL_SEND_TO_HOST_ONLY				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_SendToHostOnly,			METHOD_IN_DIRECT, FILE_ANY_ACCESS)

#define IOCTL_REGISTER_GUEST				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_RegisterGuest,			METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_DEREGISTER_GUEST				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_DeregisterGuest,			METHOD_BUFFERED, FILE_ANY_ACCESS)

// These IOCTLs apply to both the adapter and virtual adapter object
#define IOCTL_GET_STATISTICS				CTL_CODE(FILE_DEVICE_PROTOCOL, kIoctlFunction_GetStatistics,			METHOD_BUFFERED, FILE_ANY_ACCESS)


// NETSV_REGISTER_GUEST.fFlags constants
#define kVirtualSwitchRegisterGuestFlag_AddressActionMask	0x00000003
#define	kVirtualSwitchRegisterGuestFlag_GenerateAddress		0x00000000	// The VPCNetSv driver should generate a unique MAC address
#define	kVirtualSwitchRegisterGuestFlag_AddressIsSuggested	0x00000001	// The VPCNetSv driver should use the specified address if not already in use - generate a new one otherwise
#define kVirtualSwitchRegisterGuestFlag_AddressIsExclusive	0x00000002	// The VPCNetSv driver must use the specified address - fail if not available
#define kVirtualSwitchRegisterGuestFlag_AddressMayDuplicate	0x00000003	// The VPCNetSv driver must use the specified address - may duplicate an existing address

#define kVirtualSwitchRegisterGuestFlag_AddressWasGenerated	0x80000000	// The VPCNetSv driver generated a new MAC address

#define OID_GEN_CURRENT_PACKET_FILTER			0x0001010E	// Set or Query
#define OID_802_3_CURRENT_ADDRESS           	0x01010102	// Query only

#define NDIS_PACKET_TYPE_DIRECTED				0x00000001
#define NDIS_PACKET_TYPE_MULTICAST				0x00000002
#define NDIS_PACKET_TYPE_ALL_MULTICAST			0x00000004
#define NDIS_PACKET_TYPE_BROADCAST				0x00000008
#define NDIS_PACKET_TYPE_PROMISCUOUS			0x00000020

#define DEFAULT_PACKET_FILTER	(NDIS_PACKET_TYPE_DIRECTED + NDIS_PACKET_TYPE_MULTICAST + NDIS_PACKET_TYPE_BROADCAST)


#pragma pack(push)
#pragma pack(1)

typedef struct
{
	ULONG			fAdapterID;
}NETSV_CREATE_VIRTUAL_ADAPTER, *PNETSV_CREATE_VIRTUAL_ADAPTER;

#define ETHERNET_ADDRESS_LENGTH 6

typedef struct
{
	USHORT			fVersion;								// Version of this structure
	USHORT			fLength;								// Length of following data
	ULONG			fFlags;									// Flags
	UCHAR			fMACAddress[ETHERNET_ADDRESS_LENGTH];	// Guest VM MAC address
}NETSV_REGISTER_GUEST, *PNETSV_REGISTER_GUEST;

#define NETSV_REGISTER_GUEST_VERSION	0x0002
#define	NETSV_REGISTER_GUEST_LENGTH		(sizeof(NETSV_REGISTER_GUEST) - 2*sizeof(USHORT))

#define MAX_ADAPTERS						256
#define MAX_ADAPTER_BUFFER_SIZE				65536

#pragma pack(pop)


//=============================================================================
//    Function: VirtualpcDriverGetSymbolicName().
//
//    Description: Get the NT symbolic name for the packet driver we want.
//                 The driver supports multiple ones, we'll go for the first
//                 that has an ethernet address assigned and is not in use.
//=============================================================================

static
BOOLEAN
VirtualpcDriverGetSymbolicName(
    OUT UINT16 * SymbolicName
    )
{
    void *        hControl;
    UINT32        BytesReturned;
	UINT32        version;
    UINT16        Buffer[1024];
    UINT16       *pAdapterInfo;
    BOOL          bResult;
	
	hControl = CreateFileW(NETSV_DRIVER_NAME, 
		                  GENERIC_READ | GENERIC_WRITE, 
						  0, NULL, 
						  OPEN_EXISTING, 
						  FILE_FLAG_OVERLAPPED|FILE_FLAG_NO_BUFFERING, 
						  NULL);
						  
	if (hControl == INVALID_HANDLE_VALUE)
	{
		printf("Failed to open VirtualPC Network driver - not installed?\n");
		return FALSE;
	}

	if (DeviceIoControl(hControl, 
						(DWORD)IOCTL_GET_VERSION,
						NULL, 0,
						&version, sizeof(version),
						&BytesReturned,
						NULL) == FALSE)
	{
		printf("Bad VirtualPC Network driver - no version?!?\n");
		CloseHandle(hControl);
		return FALSE;
	}
	
	if (version < kVPCNetSvVersion)
	{
		printf("Warning:  VirtualPC Network driver has unexpected version (Actual: %x, Expected: %x)\n", version, kVPCNetSvVersion);
	}

    bResult = DeviceIoControl(
                        hControl,
                        IOCTL_ENUM_ADAPTERS,
                        NULL,
                        0,
                        Buffer,
                        sizeof(Buffer),
                        &BytesReturned,
                        NULL
                        );

    if ( bResult == FALSE )
    {
        return FALSE;
    }

    pAdapterInfo = (UINT16 *) Buffer;

    //
    // Skip the number of adapters, we'll just take the first one.
    //

    pAdapterInfo += 2; /* a one-wchar 'string' */

    //
    //    Skip the adapter name.
    //

    while( *pAdapterInfo != (UINT16) '\0' )
    {
        pAdapterInfo++;
    }

    pAdapterInfo++;        //... Skip the NULL.

    //
    //    Copy the symbolic name.
    //

    while( *pAdapterInfo != (UINT16) '\0' )
    {
        *SymbolicName++ = *pAdapterInfo++;
    }

    *SymbolicName++ = (UINT16) '\0';

    CloseHandle(hControl)
;
    return TRUE;
}

//=============================================================================
//    Function: VirtualpcDriverOpenAdapter().
//
//    Description: Open the packet driver.
//=============================================================================

static char MicKey[] = "SOFTWARE\\Microsoft\\Invisible Computing";
static char MicKeyMac[] = "SerplexMAC";

static
SCODE
SetMyVirtualMac( IN UCHAR *Mac)
{
    SCODE sc = E_FAIL;
    HKEY Key;
    LONG Return;
    DWORD Length;
    UCHAR Data[128];

    Return = RegCreateKeyEx(HKEY_LOCAL_MACHINE,
                            MicKey,
                            0,
                            NULL, // or "" ?
                            REG_OPTION_NON_VOLATILE,
                            KEY_ALL_ACCESS,
                            NULL,
                            &Key,
                            NULL);
    if (Return != ERROR_SUCCESS) {
        printf("Failed to create key '%s' (er=%d)\n", MicKey, Return);
        return E_FAIL;
    }

    sprintf((char*)Data,"%x:%x:%x:%x:%x:%x", Mac[0], Mac[1], Mac[2], Mac[3], Mac[4], Mac[5]);
    Length = strlen((char*)Data)+1;
    Return = RegSetValueExA(Key,
                            MicKeyMac,
                            0,
                            REG_SZ,
                            Data,
                            Length);

    if (Return != ERROR_SUCCESS) {
        printf("Failed to write value '%s' in key '%s' (er=%d)\n", MicKeyMac, MicKey, Return);
        goto Done;
    }

    sc = S_OK;
 Done:
    RegCloseKey(Key);
    return sc;
}

static
SCODE
GetMyVirtualMac( OUT UCHAR *EthernetAddress)
{
    SCODE sc = E_FAIL;
    HKEY Key;
    LONG Return;
    DWORD Type, Length;
    UCHAR Data[128];
    ULONG Mac[6];

    Return = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                          MicKey,
                          0,
                          KEY_READ,
                          &Key);
    if (Return != ERROR_SUCCESS) {
        printf("Cannot find the '%s' key (er=%d)\n", MicKey, Return);
        return E_FAIL;
    }

    Length = (sizeof Data) - 1;
    Return = RegQueryValueExA(Key,
                              MicKeyMac,
                              0,
                              &Type,
                              Data,
                              &Length);

    if (Return != ERROR_SUCCESS) {
        printf("Missing value '%s' in key '%s' (er=%d)\n", MicKeyMac, MicKey, Return);
        goto Done;
    }

    if (Type != REG_SZ) {
        printf("Value '%s\\%s' wrong type x%x (not REG_SZ)\n", MicKey, MicKeyMac, Type);
        goto Done;
    }

    Data[Length] = 0;
    /* JIC Try a few other formats in addition to what we write ourselves
     */
    Return = sscanf((char*)Data,"%x:%x:%x:%x:%x:%x", &Mac[0], &Mac[1], &Mac[2], &Mac[3], &Mac[4], &Mac[5]);
    if (Return != 6)
    Return = sscanf((char*)Data,"%x %x %x %x %x %x", &Mac[0], &Mac[1], &Mac[2], &Mac[3], &Mac[4], &Mac[5]);
    if (Return != 6)
    Return = sscanf((char*)Data,"%x-%x-%x-%x-%x-%x", &Mac[0], &Mac[1], &Mac[2], &Mac[3], &Mac[4], &Mac[5]);
    if (Return != 6) {
        printf("MAC '%s' in bad taste (not xx:xx:xx:xx:xx:xx)\n", Data);
        goto Done;
    }

    EthernetAddress[0] = (UCHAR) Mac[0];
    EthernetAddress[1] = (UCHAR) Mac[1];
    EthernetAddress[2] = (UCHAR) Mac[2];
    EthernetAddress[3] = (UCHAR) Mac[3];
    EthernetAddress[4] = (UCHAR) Mac[4];
    EthernetAddress[5] = (UCHAR) Mac[5];

    sc = S_OK;
 Done:
    RegCloseKey(Key);
    return sc;
}

static
HANDLE
VirtualpcDriverInitializeInstance(
    UINT16 * SymbolicName,
    UINT16 * AdapterName,
    UINT32 AdapterID
    )
{
    HANDLE hVirtualAdapter;

    wsprintfW(
              AdapterName,
              L"\\\\.\\%s_%08lX", 
              &SymbolicName[12],
              AdapterID);
    AdapterName[1024-1]=L'\0';
	
    hVirtualAdapter = CreateFileW(AdapterName, 
                                  GENERIC_READ | GENERIC_WRITE, 
                                  0, 
                                  NULL, 
                                  OPEN_EXISTING, 
                                  FILE_FLAG_OVERLAPPED|FILE_FLAG_NO_BUFFERING, 
                                  NULL);


	// Register with the driver
    if (hVirtualAdapter != INVALID_HANDLE_VALUE)
	{
        ULONG  BytesReturned;
        SCODE sc;
        NETSV_REGISTER_GUEST GuestInfo;

        memset(&GuestInfo,0,sizeof GuestInfo);
        GuestInfo.fVersion	 = NETSV_REGISTER_GUEST_VERSION;
        GuestInfo.fLength	 = NETSV_REGISTER_GUEST_LENGTH;

        /* Do we have a MAC to use
         */
        sc = GetMyVirtualMac(GuestInfo.fMACAddress);

        /* If not, generate a new one
         */
        if (FAILED(sc))
     	{
            GuestInfo.fFlags = kVirtualSwitchRegisterGuestFlag_GenerateAddress;
        }
        else
     	{
            GuestInfo.fFlags = kVirtualSwitchRegisterGuestFlag_AddressIsExclusive;
        }

        if (DeviceIoControl(hVirtualAdapter,
					    IOCTL_REGISTER_GUEST,
						&GuestInfo, sizeof(GuestInfo),
						&GuestInfo, sizeof(GuestInfo),
						&BytesReturned,
						NULL) == FALSE) 
        {
            printf("VirtualSwitchRegisterAdaptor failed (error %d)\n", GetLastError());
            CloseHandle(hVirtualAdapter);
            hVirtualAdapter = INVALID_HANDLE_VALUE;
        }
	
        /* Remember if we got a new one
         */
        if (GuestInfo.fFlags & kVirtualSwitchRegisterGuestFlag_AddressWasGenerated) 
        {
            int i;
            printf("VirtualPCDriver::A new Guest MAC address was generated");
            for (i = 0; i < ETHERNET_ADDRESS_LENGTH; i++)
                printf(":%x",GuestInfo.fMACAddress[i]);
            printf("\n");

            SetMyVirtualMac(GuestInfo.fMACAddress);
        }
    }
	
    return hVirtualAdapter;
}


static
HANDLE
VirtualpcDriverOpenAdapter(
    UINT16 * SymbolicName
    )
{
    UINT16    AdapterName[1024];
    HANDLE    hHostAdapter = INVALID_HANDLE_VALUE;
    UINT32    NextAdapterID = 0;
    HANDLE    hVirtualAdapter = INVALID_HANDLE_VALUE;
    ULONG     bytesReturned;
    NETSV_CREATE_VIRTUAL_ADAPTER createAdapter;

    wsprintfW(
              AdapterName,
              L"\\\\.\\%s",
              &SymbolicName[12]
              );

    hHostAdapter = CreateFileW(
                               AdapterName,
                               GENERIC_WRITE | GENERIC_READ,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL,
                               OPEN_EXISTING,
                               FILE_FLAG_OVERLAPPED,
                               0
                               );

    if (hHostAdapter == INVALID_HANDLE_VALUE)
    {
        printf("Failed to open VirtualPC Host adapter - not installed?\n");
        goto Bad;
    }

    // Create an adapter instance, starting low
    createAdapter.fAdapterID = 0;
	
    if (DeviceIoControl(hHostAdapter,
						(DWORD)IOCTL_CREATE_VIRTUAL_ADAPTER,
						&createAdapter, sizeof(createAdapter),
						&createAdapter, sizeof(createAdapter),
						&bytesReturned,
						NULL) == FALSE)
    {
        printf("Failed to create VirtualPC Host adapter instance\n");
        goto Bad;
    }

    hVirtualAdapter = VirtualpcDriverInitializeInstance(
                                 SymbolicName,
                                 AdapterName,
                                 createAdapter.fAdapterID);

    // Done
    //
    sprintf(SymbolicName,"%ls",AdapterName);

 Bad:
    /* This piece of junk doesnt work if I close hHostAdapter */
    if (hVirtualAdapter == INVALID_HANDLE_VALUE)
        CloseHandle(hHostAdapter);
    return hVirtualAdapter;
}

//=============================================================================
//    Function: VirtualpcDriverRequest().
//
//    Description: Submit a control request to the packet driver.
//=============================================================================

#define VirtualpcDriverRequest PacketDriverRequest

//=============================================================================
//    Function: VirtualpcDriverSetFilter().
//
//    Description: Set capture filters.
//=============================================================================

#define VirtualpcDriverSetFilter PacketDriverSetFilter

//=============================================================================
//    Function: VirtualpcDriverGetAddress().
//
//    Description: Get the card's ethernet address.
//=============================================================================

static
SCODE 
VirtualpcDriverGetAddress(
    IN HANDLE hFile,
    IN UINT8 *EthernetAddress
    )
{
    UnusedParameter(hFile);

    return GetMyVirtualMac(EthernetAddress);
}


//=============================================================================
//    Function: VirtualpcDriverGetSpeed().
//
//    Description: Get the link speed of the interface below.
//=============================================================================

static
SCODE 
VirtualpcDriverGetSpeed(
    IN HANDLE hFile,
    OUT ULONG *Speed
    )
{
    /* This idiotic thing cannot do this.
     */
    *Speed = 10*1000*1000;
    return S_OK;
}

//=============================================================================
//    Vtable: VirtualpcDriverVtable[].
//
//    Description: Interface to old, exclusive-use packet driver
//=============================================================================
static const PACKET_DRIVER_VTABLE VirtualpcDriverVtable = {
    VirtualpcDriverGetSymbolicName,
    VirtualpcDriverOpenAdapter,
    VirtualpcDriverRequest,
    VirtualpcDriverSetFilter,
    VirtualpcDriverGetAddress,
    VirtualpcDriverGetSpeed
};

//=============================================================================
//    Section: Ethernet::
//
//    Description: Exported functions to operate/create on NETWORK objects.
//=============================================================================


//=============================================================================
//    Function: PacketFilter().
//
//    Description: Drop non-interesting packets
//=============================================================================
#define ETHERTYPE_IP     0x0800
#define ETHERTYPE_ARP    0x0806
#define ETHERTYPE_REVARP 0x8035

#pragma pack(1)
typedef struct _UDP_PACKET {
// NB: no IP options!
    struct _IP {
        UINT8 v_hl;
        UINT8 tos;
        UINT16 len;
        UINT16 id;
        UINT16 offset;
        UINT8 ttl;
        UINT8 proto;
        UINT16 xsum;
        UINT32 src;
        UINT32 dst;
    } ip;
    struct _UDP {
        UINT16 src;
        UINT16 dst;
        UINT16 len;
        UINT16 xsum;
    } udp;
} UDP_PACKET, *PUDP_PACKET;
#pragma pack()

#define IP_PROTO_UDP     17
#define DHCP_SERVER_PORT 67

PACKET_ACTION EthernetFilter(PNETWORK Input, PPACKET Packet)
{
    UINT16 EtherType;
    PUDP_PACKET Udp;
    
    //
    //    Optionally we pass everything through
    //

    if (Unfiltered)
        return PacketActionForward;

    //
    //    Look at what ether type it is
    //

    EtherType = ntohs(Packet->Ethernet.Etype);
    switch (EtherType)
    {

        //
        //    To get an idea.. on my corporate trunk the noise is a mix
        //    of about 1 Kbps of ARP/REVARP traffic, 13 Kbps of IPv4,
        //    and about 36 Kbps of other packets, or about 50 Kbps total.
        //    This is averaged on a saturday afternoon, probably a low point.
        //    Keeping the ARP+IP traffic leaves about 24 Kbps of useful
        //    bandwidth on a serial cable at 38400 baud.
        //

    case ETHERTYPE_IP: 

        //
        //    There is enough background IP noise to warrant further checks.
        //    The most restrictive would be to only allow directed traffic,
        //    but this would cut out DHCP.
        //
        //    With the following filtering in effect the background noise
        //    (same saturday afternoon) is cut down to about 1 Kbps total,
        //    basically just the ARPs.
        //    Occasional DHCP traffic spikes would double the noise.
        //

        if (Packet->Ethernet.DstAddr[0] == 0xff)
        {
            if (debug_level > 1)
                printf("%x ", Packet->Ethernet.DstAddr[5]);

            //
            //    Check quickly for DHCP packets (from servers only!)
            //

            Udp = (PUDP_PACKET) (&Packet->Ethernet + 1);
            if ((Udp->ip.proto == IP_PROTO_UDP) &&
                (ntohs(Udp->udp.src) == DHCP_SERVER_PORT))
            {
if (debug_level > 0)
printf("#d");
                goto OkToSend;
            }

            //
            //    Anything else goes puff.
            //

            return PacketActionDrop;
        }

        //
        //    Not a broadcast, is it for us (well duh.. but promisc maybe)
        //

        if (memcmp(Packet->Ethernet.DstAddr,
                   Input->EthernetAddress,
                   6) == 0)
        {
if (debug_level > 0)
printf("#u");
            goto OkToSend;
        }

        //    Forward IP multicasts as well. (Could be more precise condition)
        if (Packet->Ethernet.DstAddr[0] == 0x01 
            && Packet->Ethernet.DstAddr[1] == 0x0 
            && Packet->Ethernet.DstAddr[2] == 0x5e
           )
        {
if (debug_level > 0)
printf("#m");
            goto OkToSend;
        }

        //
        //    Kill it.
        //

        return PacketActionDrop;

    case ETHERTYPE_REVARP:
    case ETHERTYPE_ARP: 
        //
        //    forward all ARP traffic
        //
if (debug_level > 0)
printf("#a");
        goto OkToSend;
        
    default:

        //
        //    Optionally say what and then drop.
        //

        if (debug_level > 1)
            printf("et%x ", EtherType);
    }
    return PacketActionDrop;

OkToSend:
    Packet->SerplexAddr = NICSerplexAddr;
    if (debug_level > 2)
    {
        printf("o%d ", Packet->nBytesAvail);
        PacketDump(Packet, "SerialOutput: ");
    }
    return PacketActionForward;
}

//=============================================================================
//    Function: EthernetCreate()
//
//    Description: Open and partially initialize a NETWORK object for
//                 the packet/vswitch driver.  Does not actually post packets.
//                 CompletionPort, sibling network(s) are also not done.
//
//=============================================================================

PNETWORK
EthernetCreate(BOOL OldDriver)
{
    PNETWORK    Network = NULL;
    BOOL        bResult;
    SCODE        Result;
    UINT16     *SymbolicName = NULL;
    const PACKET_DRIVER_VTABLE *Driver;

    SymbolicName = malloc(MAX_LINK_NAME_LENGTH * sizeof *SymbolicName);
    if (SymbolicName == NULL)
        goto Bad;

    //
    //    Allocate&check the descriptor.
    //

    Network = malloc(sizeof(NETWORK));
    if (Network == NULL)
        goto Bad;

 TryAgain:
    Driver = (OldDriver) ? &PacketDriverVtable : &VirtualpcDriverVtable;
    memset(Network, 0, sizeof(NETWORK));
    
    //
    //    Set the filter
    //

    Network->Filter = EthernetFilter;

    //
    //    Open the NDIS packet driver so we can send and receive.
    //

    memset(SymbolicName,0,sizeof SymbolicName);
    bResult = Driver->GetSymbolicName(SymbolicName);

    //
    //    Try again for the old one if the new one failed
    //

    if ((FALSE == bResult) && !OldDriver)
    {
        OldDriver = TRUE;
        goto TryAgain;
    }

    if (FALSE == bResult)
    {
        printf("No NDIS packet driver installed - unable to continue\n");
        goto Bad;
    }

    Network->Name = (char *)SymbolicName;
    Network->hFile = Driver->OpenAdapter(SymbolicName);

    if (Network->hFile == INVALID_HANDLE_VALUE)
    {
        printf("Failed to open packet driver '%ls' sc=%x\n",
               SymbolicName, GetLastError());
        goto Bad;
    }

    //
    //    Set up the capture filter.
    //

    Result = Driver->SetFilter(
                    Network->hFile,
                    NDIS_PACKET_TYPE_BROADCAST | 
                    NDIS_PACKET_TYPE_MULTICAST |
                    NDIS_PACKET_TYPE_ALL_MULTICAST |
                    NDIS_PACKET_TYPE_DIRECTED        
                    );
    if (Result != S_OK)
    {
        /* Shouldnt happen, but we can survive this */
        printf("Warning: Driver::SetFilter FAILED(%x)\n",Result);
    }

    //
    //    Get the ethernet address for this adapter.
    //

    Result = Driver->GetAddress(
                    Network->hFile, 
                    Network->EthernetAddress
                    );
    if (Result != S_OK)
    {
        printf("Fatal: Driver::GetAddress FAILED(%x)\n",Result);
        goto Bad;
    }

    //
    //    Get the speed of this adapter.
    //

    Result = Driver->GetSpeed(
                    Network->hFile, 
                    &Network->BaudRate
                    );
    if (Result != S_OK)
    {
        /* Shouldnt happen, but we can survive this too */
        printf("Warning: Driver::GetSpeed FAILED(%x)\n",Result);
        Network->BaudRate = 10*1000*1000;
    }


    //
    //    Needs more work to actually start functioning
    //

    Network->bInitialized = FALSE;

    return Network;

 Bad:
    if (SymbolicName != NULL)
        free(SymbolicName);

    if (Network != NULL)
    {
        if (Network->hFile != INVALID_HANDLE_VALUE)
            CloseHandle(Network->hFile);
        free(Network);
    }

    return NULL;
}

