/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    ehci.c
 *
 * Purpose:
 *
 *    Device driver for the Enhanced Host Controller (EHCI) USB 2.0 master.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <stddef.h> /* for offsetof() */
#include <tchar.h>
//#define _DEBUG 1
int EhciDebug = 2;
#define DBGLVL EhciDebug
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbmaster.h>
#include <busdriver.h>

/* BUGBUG We sure could use a trivial VM system
 */
#define IoPhysicalToIoVirtual(x) ((PTR)x)
#define PhysicalToVirtual(x)     ((PTR)((ADDRESS)x))
#define VirtualToPhysical(x)     ((ADDRESS)x)

#if (BYTE_ORDER==BIG_ENDIAN)
#define PCIDRIVER_NAME _T("pcibs")
#else
#define PCIDRIVER_NAME _T("pci")
#endif

#include "ehci.h"

/* This could be anything up to 1024, but save memory */
#define MAX_POLLING_FREQUENCY  16 /* msecs */

/* Define macros for inclusion of the interrupt schedule code
 */
#define __NEEDS_TWO_LEVEL_SCHEDULING 1
#define USK_LEVEL1_MAX     MAX_POLLING_FREQUENCY
#define USK_LEVEL2_MAX     8
#define USK_BW_PER_FRAME   (60000 * 8 / 10)
#define HC_NULL_POINTER    EHCI_POINTER_TYPE_TERMINATOR
#define HC_POINTER_TYPE    EHCI_FRAMELIST_LINK_POINTER

PRIVATE INLINE void EhciWriteRegister(volatile UINT32 *Address, UINT32 Value);
#define HcSetPointerValue(_pointer_,_value_) \
    EhciWriteRegister(_pointer_,_value_)

#include "usb_sched.c"

/* Add a back-pointer to (some of) our EDs so that we can find
 * the interrupt list they are on when they go away. Otherwise,
 * we would have to look all over the schedule for it.
 * NB: The same allocator is used for this and MYEHCI_QTD!
 */
typedef struct _MYEHCI_QH {
    /* goes first, for alignment reasons */
    EHCI_QH             Ed;   /* NB: at x30 this is 4 words bigger than a QTD */
    void                *TrashCan;
    void                *Itd;
    PUSKNODE            InterruptHead;
    UINT8               Pad[64-sizeof(EHCI_QH)-(2*sizeof(void*))-sizeof(PUSKNODE)];
} MYEHCI_QH, *PMYEHCI_QH;

/* Add a back-pointer to our TDs so that we can find the QH they are on
 * when one does not complete.  
 */
typedef struct _MYEHCI_QTD {
    /* goes first, for alignment reasons */
    EHCI_QTD            Td;   /* x20 */
    PMYEHCI_QH          Ed;
    UINT8               Pad[64-sizeof(EHCI_QTD)-sizeof(PMYEHCI_QH)];
} MYEHCI_QTD, *PMYEHCI_QTD;

/* Roundup the size of a split-iso TD so that we can allocate an array-of
 * that obeys the HC alignment requirements in each element.
 */
typedef struct _MYEHCI_SITD {
    EHCI_SITD           Td;   /* x1c */
    UINT8               Pad[32-sizeof(EHCI_SITD)];
} MYEHCI_SITD, *PMYEHCI_SITD;

/* Regular ITDs are sized correctly already
 */
typedef struct _MYEHCI_ITD {
    EHCI_ITD            Td;
} MYEHCI_ITD, *PMYEHCI_ITD;

/* Define the structure of our per-URB state,
 * which is an overlay of the two HCI-reserved pointers.
 */
typedef struct _HCI_STATE {
    PUSBM_REQUEST_BLOCK Next;   /* Needed for cancel to cleanup */
    union {
        PMYEHCI_QTD     Td;     /* Needed for cancel to cleanup */
        PMYEHCI_ITD     Itd;    /* Needed for cancel to cleanup */
        PMYEHCI_SITD    Sitd;   /* Needed for cancel to cleanup */
        ADDRESS         Tag;    /* 1 for Sitd, 0 otherwise: */
#define HCI_SITD_TAG 1
#define GetSitd(a)  ((PMYEHCI_SITD)((ADDRESS)(a) & ~HCI_SITD_TAG))
#define MakeSitd(s) ((PMYEHCI_SITD)((ADDRESS)(s) | HCI_SITD_TAG))
#define GetItd(a)   (a)
    } u;
} HCI_STATE, *PHCI_STATE;

/* Trivial hash table
 */
typedef struct _EHCI_HT {
#define InitialHashEntries 8  /* 2**n */
    int nEntries;
    struct _EHCI_HT_ENTRY {
        PTR    DataValue;
        UINT32 HashValue;
    } Entries[1]; /* varsized */
} EHCI_HT, *PEHCI_HT;

/* Our device objecy
 */
typedef struct _EHCI {
    struct CDevice CDevice;
    CONST_VTBL struct IUsbMasterControllerVtbl *EhciMasterv;

    PEHCI_CAPS Caps;
    PEHCI_OPS  Ops;
    UINT16     PortSw[EHCI_PR_PORTROUTE_MAX];
    UINT16     PortStatusChanges;

    UINT16 IsrStatus;
    UINT16 PeriodicListSize;
    EHCI_FRAMELIST_LINK_POINTER *PeriodicList;
    PUSKEDULE Schedule; 

    PMYEHCI_QH AsyncList;

    /* Pack our bools */
    UINT8 nPorts;
    UINT8 PortPowerControl;
    UINT8 EventPending;
    UINT8 UnusedForNow;

    PUSBM_REQUEST_BLOCK ActiveUrbs;
    PUSBM_REQUEST_BLOCK ActiveIsoUrbs;
    PUSBM_REQUEST_BLOCK EventsPacket;

    PEHCI_HT HashTable;
    MUTEX HashTableLock;

    PMYEHCI_QTD FreeTds;
    int FreeTdCount;
#define FreeTdMax 8

    void *TrashCan;

    struct _STATS {
        int IsrCnt;
        int IntrCnt;
        int RegCnt;
        int ChgCnt;
        int ErrCnt;
    } Stats;

} EHCI, *PEHCI;

/* Convert between interface and instance pointer and vice versa */
#define pEhciDevice(_i_) ((PEHCI)((_i_)))
#define pEhciMaster(_i_) ((PEHCI)((ADDRESS)(_i_) - offsetof(struct _EHCI, EhciMasterv)))
#define iEhciDevice(_p_) ((PIDEVICE)((_p_)))
#define iEhciMaster(_p_) ((PIUSBMASTERCONTROLLER)(&((PEHCI)(_p_))->EhciMasterv))

/*
 * Phys functions
 */
PRIVATE INLINE UINT32 EhciReadRegister(volatile UINT32 *Address)
{
    UINT32 Value = *Address;
    return _bs32(Value);
}

PRIVATE INLINE void EhciWriteRegister(volatile UINT32 *Address, UINT32 Value)
{
    Value = _bs32(Value);
    *Address = Value;
    IOWriteBufferFlush();
}

/*
 * General utilities
 */

/* Sleep for a few milliseconds
 */
PRIVATE void MilliSleep(int mSecs)
{
    TIME t;
    Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(mSecs)));
    SleepUntil(t);
}

/* Get the frame number for the currently executing frame
 */
PRIVATE INLINE UINT16 CurrentFrameNumber(PEHCI this)
{
    UINT32 Value32;
    Value32 = EhciReadRegister(&this->Ops->FrIndex);
    return (UINT16)Value32 & EHCI_FRI_MASK;
}

/* Compute the min/max bandwidth for xfers of a given type
 */
PRIVATE void
EhciComputeBandwidth(
    USBM_URB_FLAGS Flags,
    UINT16 MaximumPacketLength,
    UINT16 *MinimumStartBandwidth,
    UINT16 *MaximumStartBandwidth,
    UINT16 *MinimumEndBandwidth,
    UINT16 *MaximumEndBandwidth
    )
{
    UINT16 Overhead, EndOverhead;
    UINT16 Bandwidth, EndBandwidth;

    if (Flags & URB_FLAG_HIGH_SPEED) {
        /*
         * For an interrupt transfer, the minimum b/w is when a NAK is returned.
         * Call that the per-packet overhead.
         */
        Overhead     = (Flags & URB_FLAG_INTERRUPT) ? 55 : 38;
        Bandwidth    = MaximumPacketLength;
        EndOverhead  = 0;
        EndBandwidth = 0;

        goto Done;
    }

    /* Split-transfer, depends on direction and type
     */
    if (Flags & URB_FLAG_IN) {
        /*
         * An IN start transaction involves a start-split packet and a token.
         * The complete transaction involves a complete-split packet, a token,
         * and a data packet.  If it is *not* an isochronous IN then there is
         * an additional handshake packet.
         */
        Overhead     = 39;
        Bandwidth    = 0;
        EndOverhead  = (Flags & URB_FLAG_INTERRUPT) ? 75 : 58;
        EndBandwidth = MaximumPacketLength;

    } else {
        /*
         * An OUT start transaction involves a start-split packet, a token,
         * and a data packet.  For isochronous OUT nothing else happens.
         * The interrupt OUT complete transaction involves a complete-split
         * packet, a token, and a handshake packet.
         */
        Overhead     = 39;
        Bandwidth    = MaximumPacketLength;
        EndOverhead  = (Flags & URB_FLAG_INTERRUPT) ? 56 : 0;
        EndBandwidth = 0;
        
    }

 Done:
    /*
     * That is in bytes.  Add bitstuffing and turn it into high-speed clocks.
     */
    Bandwidth    = ((Bandwidth + Overhead) * 8 * 7) / 6;
    Overhead     = ((        0 + Overhead) * 8 * 7) / 6;
    EndBandwidth = ((EndBandwidth + EndOverhead) * 8 * 7) / 6;
    EndOverhead  = ((           0 + EndOverhead) * 8 * 7) / 6;

    /* Return the computed values
     */
    *MaximumStartBandwidth = Bandwidth;
    *MinimumStartBandwidth = Overhead;
    *MaximumEndBandwidth   = EndBandwidth;
    *MinimumEndBandwidth   = EndOverhead;
}

/*
 * Utilities for handling MYTDs
 */
PRIVATE PMYEHCI_QTD
TdAlloc(PEHCI this)
{
    PMYEHCI_QTD Td, Next;
    PIHEAP pHeap;

    /* Try to grab one from the free list
     */
    do {
        Td = this->FreeTds;
        if (Td == NULL)
            break;
        Next = (PMYEHCI_QTD) Td->Ed;
    } while (!AtomicCmpAndSwap((ADDRESS*) &this->FreeTds,
                               (ADDRESS) Td,
                               (ADDRESS) Next));

    /* If empty allocate a new one
     */
    if (Td != NULL) {
        this->FreeTdCount--; /* no prob if slightly off */
    } else {
        pHeap = CurrentHeap();

        assert(sizeof(MYEHCI_QH) == sizeof(MYEHCI_QTD));

        /* NB: NOT zeroed, like it comes from the freelist */
        Td = (PMYEHCI_QTD) pHeap->v->Alloc(pHeap, 0,
                                          sizeof *Td,
                                          EHCI_QH_ALIGNMENT+1);
        if (!Td) {
            DBGME(3,_tprintf(_T("EHCI(%x) Td outtamem.\n"), (ADDRESS) this));
            return NULL;
        }
    }

#if 1
    memset(Td,0,sizeof *Td);
#endif
    Td->Ed = NULL; /* sanity */
    return Td;
}

/* Free a list of TDs.
 * This is callable from an Isr,
 * in which case it wont go to the heap
 */
PRIVATE void
TdFree(PEHCI this, PMYEHCI_QTD Td, BOOL FromIsr)
{
    PMYEHCI_QTD Next;
    UINT32 Value32;
    PIHEAP pHeap = CurrentHeap();

    do {
        Value32 = EhciReadRegister(&Td->Td.Next) & EHCI_POINTER_MASK;

        /* Put it on free list?
         */
        if (FromIsr || (this->FreeTdCount < FreeTdMax)) {
            /* Normally-won race to the top
             */
            do {
                Next = this->FreeTds;
                Td->Ed = (void *) Next;
            } while (!AtomicCmpAndSwap((ADDRESS*) &this->FreeTds,
                                       (ADDRESS) Next,
                                       (ADDRESS) Td));

            this->FreeTdCount++; /* no prob if slightly off */

        } else {
            /* Nope, we got enough. Free back to heap
             */
            pHeap->v->Free(pHeap, 0, (void *)Td);
        }

        if (Value32)
            Td = (PMYEHCI_QTD)((ADDRESS)PhysicalToVirtual(Value32) - offsetof(MYEHCI_QTD,Td));
    } while (Value32);
}

/* Abuse: use the TDs for EDs also.
 */
#define EdAlloc(t)    (PMYEHCI_QH)TdAlloc(t)
#define EdFree(t,e,f) {((PMYEHCI_QTD)e)->Td.Next = 0;TdFree(t,(PMYEHCI_QTD)e,f);}

/* Utilities for iTDs and siTDs
 */
PRIVATE PMYEHCI_ITD
ItdAlloc(PEHCI this)
{
    PMYEHCI_ITD Itd;
    PIHEAP pHeap;

    pHeap = CurrentHeap();

    /* NB: NOT zeroed */
    Itd = (PMYEHCI_ITD) pHeap->v->Alloc(pHeap, 0,
                                        USK_LEVEL1_MAX * sizeof *Itd,
                                        EHCI_ITD_ALIGNMENT+1);
    if (!Itd) {
        DBGME(3,_tprintf(_T("EHCI(%x) Itd outtamem.\n"), (ADDRESS) this));
        return NULL;
    }

    return Itd;
}

/* Free a list of ITDs.
 * This is NOT callable from an Isr.
 */
PRIVATE void
ItdFree(PEHCI this, PMYEHCI_ITD Itd)
{
    PIHEAP pHeap = CurrentHeap();
    pHeap->v->Free(pHeap, 0, (void *)Itd);
}

PRIVATE PMYEHCI_SITD
SitdAlloc(PEHCI this)
{
    PMYEHCI_SITD Itd;
    PIHEAP pHeap;

    pHeap = CurrentHeap();

    /* NB: NOT zeroed */
    Itd = (PMYEHCI_SITD) pHeap->v->Alloc(pHeap, 0,
                                        USK_LEVEL1_MAX * sizeof *Itd,
                                        EHCI_SITD_ALIGNMENT+1);
    if (!Itd) {
        DBGME(3,_tprintf(_T("EHCI(%x) Sitd outtamem.\n"), (ADDRESS) this));
        return NULL;
    }

    return Itd;
}

#define SitdFree(t,s)  ItdFree(t,(PMYEHCI_ITD)s)

/* Utilities for handling EDs
 */
PRIVATE INLINE int
EdHash(UINT32 HashValue)
{
    /* Give two consecutive slots to each usb address.
     * One is for control, the other for EP 1.
     */
    return ((HashValue & EHCI_QH_DEVICE_MASK) << 1) |
            (HashValue & 0xe00/*EHCI_QH_ENDPT_MASK^lowbit*/) |
            ((HashValue & 0x100) >> EHCI_QH_ENDPT_SHIFT);
}

PRIVATE PMYEHCI_QH
LookupEd(PEHCI this,
         PUSBM_REQUEST_BLOCK Urb)
{
    int i, j, HashKey;
    UINT32 HashValue;
    PIHEAP pHeap;
    PEHCI_HT HashTable, NewHashTable;
    PMYEHCI_QH Ed = NULL;
    UINT8 DeviceAddress = Urb->DeviceAddress;
    UINT8 EndpointAddress = Urb->EndpointAddress;
    USBM_URB_FLAGS Flags = Urb->Flags;
    UINT8 HubDeviceAddress = 0;
    UINT8 HubPortNumber = 0;

    /* Configuring?
     */
    if (DeviceAddress == 0) {
        if (Flags & URB_FLAG_HIGH_SPEED)
            return this->AsyncList;
        /* Otherwise fall through and (possibly) create a new (0,0) endpoint
         */
        HubDeviceAddress = Urb->HubDeviceAddress;
        HubPortNumber = Urb->HubPortNumber;
    }

    /* Build the hash table key
     * The HashValue uniquely identifies an endpoint.
     * It also contains additional information used at removal time.
     */
    HashValue = (DeviceAddress & EHCI_QH_DEVICE_MASK) |
                ((EndpointAddress<<EHCI_QH_ENDPT_SHIFT) & EHCI_QH_ENDPT_MASK) |
                ((Flags & URB_FLAG_LOW_SPEED) ? EHCI_QH_1_5Mbps : 0) |
                ((Flags & URB_FLAG_HIGH_SPEED) ? EHCI_QH_480Mbps : 0) |
                ((Flags & URB_FLAG_ISOCHRONOUS) ? EHCI_QH_H : 0) |
                ((HubDeviceAddress & 0xff) << 16) |
                ((HubPortNumber & 0xff) << 24);

    assert(HashValue != 0);
    HashKey = EdHash(HashValue);

    /* BUGBUG Will need this lock-free if Isr plays with it
     */
    Mutex_Lock(&this->HashTableLock);

    HashTable = this->HashTable;
    if (HashTable == NULL) {
        pHeap = CurrentHeap();

        HashTable = (PEHCI_HT) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY,
                                               sizeof(*HashTable) +
                                               (InitialHashEntries-1) * sizeof(struct _EHCI_HT_ENTRY),
                                               0);
        if (!HashTable) {
            DBGME(3,_tprintf(_T("EHCI(%x) HT outtamem.\n"), (ADDRESS) this));
            goto UnlockAndDone;
        }
        HashTable->nEntries = InitialHashEntries;
        this->HashTable = HashTable;
    }

    /* Quick hit? */
    i = HashKey & (HashTable->nEntries-1);

    /* Scan forward the whole table to make sure we did not clash
     * or that this slot did not just free up
     */
    for (j = 0; j < HashTable->nEntries; j++) {
        if ((HashTable->Entries[i].HashValue | EHCI_QH_DTC) ==
            (HashValue | EHCI_QH_DTC)) {
            Ed = HashTable->Entries[i].DataValue;
            goto UnlockAndDone;
        }
        if (++i == HashTable->nEntries)
            i = 0;
    }

    /* Ok, we got to allocate one
     */
    Ed = EdAlloc(this);
    if (Ed == NULL) {
        DBGME(3,_tprintf(_T("EHCI(%x) ED outtamem.\n"), (ADDRESS) this));
        goto UnlockAndDone;
    }

    /* Mark it as freshly allocated
     */
    Ed->Ed.Next = 0;
    Ed->Ed.EndpointCaps = 0;

    /* Insert it in the hash table
     * Scan forward till we find a slot
     */
    if (Flags & URB_FLAG_IN)  /* remember direction for removal time */
        HashValue |= EHCI_QH_DTC;
 Again:
    i = HashKey & (HashTable->nEntries-1);
    for (HashKey = 0; HashKey < HashTable->nEntries; HashKey++) {
        if (HashTable->Entries[i].HashValue == 0) {
            HashTable->Entries[i].HashValue = HashValue;
            HashTable->Entries[i].DataValue = Ed;
            DBGME(0,_tprintf(_T("EdLookup(%d,%d,%d,%d)[%x] hash %x new @%x\n"),
                             DeviceAddress, EndpointAddress,
                             HubDeviceAddress, HubPortNumber,
                             Flags,
                             HashValue, (ADDRESS) Ed));
            goto UnlockAndDone;
        }
        if (++i == HashTable->nEntries)
            i = 0;
    }

    /* Bummer, the table is full;
     * Allocate a 2x sized one, copy, free old, set new, restart.
     */
    i = HashTable->nEntries * 2;

    pHeap = CurrentHeap();
    NewHashTable = (PEHCI_HT) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY,
                                              sizeof(*HashTable) +
                                              (i-1) * sizeof(struct _EHCI_HT_ENTRY),
                                              0);
    if (!NewHashTable) {
        DBGME(3,_tprintf(_T("EHCI(%x) HT outtamem2.\n"), (ADDRESS) this));
        goto UnlockAndDone;
    }

    memcpy(NewHashTable,
           HashTable,
           sizeof(*HashTable) + sizeof(struct _EHCI_HT_ENTRY) * (i/2 - 1));
    NewHashTable->nEntries = i;
    this->HashTable = NewHashTable;
    pHeap->v->Free(pHeap,0,HashTable);
    HashTable = NewHashTable;
    goto Again;

 UnlockAndDone:
    Mutex_Unlock(&this->HashTableLock);

    return Ed;
}

/* Remove an endpoint from the aperiodic list
 */
PRIVATE void
EhciDropEdFromAsync(PEHCI this,
                    PMYEHCI_QH Ed)
{
    PMYEHCI_QH NextEd;
    void *NextPtr;
    UINT32 Value32;

    DBGME(0,_tprintf(_T("Ehci(%x)DEFA %x\n"), (ADDRESS)this,(ADDRESS)Ed));

    /* NB: The first in the AsyncList is (0,0) and never goes away.
     */
    for (NextEd = this->AsyncList; NextEd != NULL; NextEd = NextPtr) {

        /* Read and sanity check the next pointer
         */
        Value32 = EhciReadRegister(&NextEd->Ed.Next);
        if ((Value32 & EHCI_POINTER_MASK) == 0)
            return;

        /* Got it?
         */
        NextPtr = PhysicalToVirtual(Value32 & EHCI_POINTER_MASK);
        if (NextPtr == Ed) {
            /* Yes, remove from the (circular) list
             */
            DBGME(0,_tprintf(_T("EhciDEFA %x was after %x\n"),
                             (ADDRESS)Ed, (ADDRESS)NextEd));
            NextEd->Ed.Next = Ed->Ed.Next;
            return;
        }
    }
}

/* Remove an endpoint from the periodic schedule
 */
PRIVATE void
EhciDropEdFromSchedule(PEHCI this,
                       PMYEHCI_QH Ed,
                       BOOL InputEndpoint)
{
    int Frequency, StartSlot;
    EHCI_FRAMELIST_LINK_POINTER *NextPtr;
    UINT32 EdPhysical = VirtualToPhysical(Ed) | EHCI_POINTER_TYPE_QH;
    UINT32 Value32;

    DBGME(0,_tprintf(_T("Ehci(%x)DEFS %x.  ephy %x\n"),
                     (ADDRESS)this,(ADDRESS)Ed, EdPhysical));

    /* Where do we start looking
     */
    Frequency = UskFrequency(Ed->InterruptHead);
    StartSlot = UskStart(Ed->InterruptHead);

    /* Its replicated a few times (at least 256/16==16)
     */
    for (;;) {

        /* Keep going until we find it
         */
        NextPtr = &this->PeriodicList[StartSlot];

        for (;;) {
            Value32 = EhciReadRegister(NextPtr);

            /* Stop at the end of the list (sanity)
             */
            if ((Value32 & EHCI_POINTER_TYPE_TERMINATOR) ||
                (Value32 == EdPhysical) ||
                (Value32 == 0))
                break;

            NextPtr = PhysicalToVirtual(Value32 & EHCI_POINTER_MASK);
        }

        /* Make sure the schedule is not broken
         */
        if (Value32 == EdPhysical) {

            /* If we are writing into the roots we must do it a few times,
             * otherwise its easy.
             */
            if (NextPtr != &this->PeriodicList[StartSlot]) {

                DBGME(0,_tprintf(_T("EhciDEFS: Node %x was after %x\n"),
                                 (ADDRESS) Ed, (ADDRESS) NextPtr));
                *NextPtr = Ed->Ed.Next;

            } else {

                int RootIndex;
                for (RootIndex = StartSlot;
                     RootIndex < this->PeriodicListSize;
                     RootIndex += MAX_POLLING_FREQUENCY) {

                    DBGME(0,_tprintf(_T("EhciDEFS: Node %x @root %d\n"),
                                     (ADDRESS) Ed, RootIndex));
                    this->PeriodicList[RootIndex] = Ed->Ed.Next;
                }
            }

        }

        /* Go look at the next slot, if any
         */
        StartSlot += Frequency;

        if (StartSlot >= USK_LEVEL1_MAX)
            break;
    }

    /* Unlink the next endpoint, if any
     */
    EhciWriteRegister(&Ed->Ed.Next,EHCI_POINTER_TYPE_TERMINATOR);

    /* Release the bandwidth reservation for the endpoint
     */
    {
        UINT16 MaxStartBw, MinStartBw;
        UINT16 MaxEndBw, MinEndBw;
        BOOL HighSpeed;

        Value32 = EhciReadRegister(&Ed->Ed.EndpointCaps);
        HighSpeed = ((Value32 & EHCI_QH_EPS) == EHCI_QH_480Mbps);
        Value32   = (Value32 & EHCI_QH_MAX_PACKET_SIZE) >> EHCI_QH_MPS_SHIFT;

        EhciComputeBandwidth((USBM_URB_FLAGS)
                             (URB_FLAG_INTERRUPT
                              | (HighSpeed ? URB_FLAG_HIGH_SPEED : 0)
                              | (InputEndpoint ? URB_FLAG_IN : URB_FLAG_OUT)),
                             (UINT16)Value32,
                             &MinStartBw,
                             &MaxStartBw,
                             &MinEndBw,
                             &MaxEndBw);

        if (HighSpeed)
            UskCancelReservation(this->Schedule,
                                 Ed->InterruptHead,
                                 MinStartBw, MaxStartBw);
        else
            UskCancelMultiReservation(this->Schedule,
                                      Ed->InterruptHead,
                                      MinStartBw, MaxStartBw,
                                      MinEndBw, MaxEndBw);
    }
}

/* Final cleanup phase for recycled EDs
 */
PRIVATE void
EhciEmptyTheTrashCan(PEHCI this)
{
    PMYEHCI_QH Ed;

    while ((Ed = this->TrashCan) != NULL) {
        this->TrashCan = Ed->TrashCan;

        DBGME(0,_tprintf(_T("EhciTrash %x (%x)\n"),
                         (ADDRESS)Ed, (ADDRESS)this->TrashCan));

        EdFree(this,Ed,TRUE);
    }
}

/* Enque a TD to an ED
 */
PRIVATE void
EdEnqueTd(PEHCI_QH Ed, PEHCI_QTD Td)
{
    UINT32 Value32;

    /* BUGBUG Only valid for 1st xfer!!!
     */
    Value32 = VirtualToPhysical(&Td->Next);
    EhciWriteRegister(&Ed->TransferOverlay.Next,Value32);
}

/* Debugging tools
 */
#if _DEBUG
typedef _TCHAR *MYSTRING;

typedef struct _REGFIELD {
    MYSTRING FieldName;
    UINT8 MsbNo, LsbNo; /* inclusive */
} REGFIELD, *PREGFIELD;

typedef struct _REGDESC {
    MYSTRING RegisterName;
    REGFIELD *Field; /* NULL terminated */
} REGDESC, *PREGDESC;

static const _TCHAR Spaces[16] = _T("               ");
#define Whitespace(n) (Spaces + (15 - ((n)&15)))

PRIVATE void
DumpReg(PREGDESC Format, UINT32 Value, int nSpaces)
{
    PREGFIELD Field;
    UINT32 Mask;
    UINT32 FieldValue;

    _tprintf(_T("%s%s %x"),Whitespace(nSpaces),Format->RegisterName,Value);
    if (Value) {
        _tprintf(_T(":"));
        for (Field = &Format->Field[0]; Field->FieldName != NULL; Field++) {
            Mask = ~0;
            Mask = Mask >> Field->LsbNo;
            Mask = Mask << Field->LsbNo;
            Mask = Mask << (31 - Field->MsbNo);
            Mask = Mask >> (31 - Field->MsbNo);

            FieldValue = (Value & Mask) >> Field->LsbNo;
            if (FieldValue!= 0)
                _tprintf(_T(" %s %x"),Field->FieldName, FieldValue);
        }
    }
    _tprintf(_T("\n"));
}

static REGFIELD PointerFields[] = {
        { _T("Phys"),  31, 0},
        { _T("Type"),  4, 1},
        { _T("T"),  0, 0},
        { NULL,}
};
static REGDESC Next = {
    _T("Next"), PointerFields
};



PRIVATE void
DumpTd(PEHCI_QTD Td, int nSpaces, BOOL IsOverlay)
{
    UINT32 Value32;
    int i;

    static REGFIELD TokenFields[] = {
        { _T("Ping"),  0, 0},
        { _T("EndSpl"), 1, 1},
        { _T("uFmiss"),  2, 2},
        { _T("XactErr"),  3, 3},
        { _T("Babble"),  4, 4},
        { _T("DataErr"),  5, 5},
        { _T("Halted"),  6, 6},
        { _T("Active"),  7, 7},
        { _T("PID"), 9, 8},
        { _T("Cerr"), 11, 10},
        { _T("CPage"),  14, 12},
        { _T("Ioc"), 15, 15},
        { _T("Len"), 30, 16},
        { _T("DT"), 31, 31},
        { NULL,}
    };
    static REGDESC Token = {
        _T("Token"), TokenFields
    };
    static REGDESC Alternate = {
        _T("Alternate"), PointerFields
    };

    _tprintf(_T("%sTransfer Descriptor @%x:\n"), Whitespace(nSpaces), (ADDRESS) Td);
    nSpaces++;

    Value32 = EhciReadRegister(&Td->Token);
    if (Value32)
        DumpReg(&Token,Value32,nSpaces);

    for (i = 0; i < 5; i++) {
        Value32 = EhciReadRegister(&Td->Page[i]);
        if (Value32)
            _tprintf(_T("%sPage[%d] %x\n"),Whitespace(nSpaces),i, Value32);
    }

    Value32 = EhciReadRegister(&Td->Alternate);
    if (Value32 != EHCI_POINTER_TYPE_TERMINATOR) {
        DumpReg(&Alternate,Value32,nSpaces);
        Value32 &= EHCI_POINTER_MASK;
        if (Value32 && IsOverlay) {
            PEHCI_QTD aTd;
            aTd = (PEHCI_QTD) PhysicalToVirtual(Value32);
            _tprintf(_T("%sAlternate"),Whitespace(nSpaces+1));
            DumpTd(aTd,nSpaces+1,FALSE);
        }
    }

    Value32 = EhciReadRegister(&Td->Next);
    if (Value32 != EHCI_POINTER_TYPE_TERMINATOR)
        DumpReg(&Next,Value32,nSpaces);

}

PRIVATE void
DumpItd(PEHCI_ITD Td, int nSpaces)
{
    int i;
    UINT32 Value32;

    static REGFIELD TransactionFields[] = {
        { _T("Off"), 11, 0},
        { _T("Pg"), 14, 12},
        { _T("Ioc"), 15, 15},
        { _T("Len"), 27, 16},
        { _T("Status"), 31, 28},
        { _T("XactErr"),  28, 28},
        { _T("Babble"),  29, 29},
        { _T("DataErr"),  30, 30},
        { _T("Active"),  31, 31},
        { NULL,}
    };
    static REGDESC Transaction = {
        _T("Transaction"), TransactionFields
    };

    static REGFIELD ItdPointerFields[] = {
        { _T("Page"), 31, 12},
        { _T("Pointer"), 31, 0},
        { NULL,}
    };
    static REGDESC Pointer = {
        _T("Pointer"), ItdPointerFields
    };

    static REGFIELD AddressFields[] = {
        { _T("Page"), 31, 12},
        { _T("Device"), 6, 0},
        { _T("EndPt"), 11, 8},
        { NULL,}
    };
    static REGDESC Address = {
        _T("[0]Pointer"), AddressFields
    };

    static REGFIELD MpsFields[] = {
        { _T("Page"), 31, 12},
        { _T("Mps"), 10, 0},
        { _T("IN"), 11, 11},
        { NULL,}
    };
    static REGDESC Mps = {
        _T("[1]Pointer"), MpsFields
    };

    static REGFIELD MultFields[] = {
        { _T("Page"), 31, 12},
        { _T("Mult"), 1, 0},
        { NULL,}
    };
    static REGDESC Mult = {
        _T("[3]Pointer"), MultFields
    };

    _tprintf(_T("%sI-Transfer Descriptor @%x:\n"),Whitespace(nSpaces), (ADDRESS) Td);
    nSpaces++;

    for (i = 0; i < 8; i++) {
        Value32 = EhciReadRegister(&Td->Transaction[i]);
        if (Value32) {
            _tprintf(_T("%s[%d]"),Whitespace(nSpaces),i);
            DumpReg(&Transaction,Value32,0);
        }
    }

    Value32 = EhciReadRegister(&Td->Pointer[0]);
    DumpReg(&Address,Value32,nSpaces);

    Value32 = EhciReadRegister(&Td->Pointer[1]);
    DumpReg(&Mps,Value32,nSpaces);

    Value32 = EhciReadRegister(&Td->Pointer[2]);
    if (Value32 & EHCI_ITD_MULT)
        DumpReg(&Mult,Value32,nSpaces);

    for (i = 3; i < 7; i++) {
        Value32 = EhciReadRegister(&Td->Pointer[i]);
        if (Value32) {
            _tprintf(_T("%s[%d]"),Whitespace(nSpaces),i);
            DumpReg(&Pointer,Value32,0);
        }
    }

    Value32 = EhciReadRegister(&Td->Next);
    if (Value32)
        DumpReg(&Next,Value32,nSpaces);

}

PRIVATE void
DumpSitd(PEHCI_SITD Td, int nSpaces)
{
    UINT32 Value32;

    static REGFIELD CapsFields[] = {
        { _T("Device"),  6, 0},
        { _T("r1"), 7, 7},
        { _T("EndPt"), 11, 8},
        { _T("r2"),  15, 12},
        { _T("HubDevice"),  22, 16},
        { _T("r3"),  23, 23},
        { _T("HubPort"),  30, 24},
        { _T("IN"), 31, 31},
        { NULL,}
    };
    static REGDESC Caps = {
        _T("Caps"), CapsFields
    };

    static REGFIELD ScheduleFields[] = {
        { _T("Smask"), 7, 0},
        { _T("Cmask"), 15, 8},
        { _T("mbz"), 31, 16},
        { NULL,}
    };
    static REGDESC Schedule = {
        _T("Schedule"), ScheduleFields
    };

    static REGFIELD TransferStateFields[] = {
        { _T("Status"), 7, 0},
        { _T("r0"), 0, 0},
        { _T("EndSpl"), 1, 1},
        { _T("uFmiss"),  2, 2},
        { _T("XactErr"),  3, 3},
        { _T("Babble"),  4, 4},
        { _T("DataErr"),  5, 5},
        { _T("ERR"),  6, 6},
        { _T("Active"),  7, 7},
        { _T("Pmask"), 15, 8},
        { _T("Len"), 25, 16},
        { _T("r1"), 29, 26},
        { _T("Page1"), 30, 30},
        { _T("IOC"), 31, 31},
        { NULL,}
    };
    static REGDESC TransferState = {
        _T("TransferState"), TransferStateFields
    };

    static REGFIELD Page0Fields[] = {
        { _T("Offset"), 11, 0},
        { _T("Page"), 31, 12},
        { _T("Pointer"), 31, 0},
        { NULL,}
    };
    static REGDESC Page0 = {
        _T("Page0"), Page0Fields
    };

    static REGFIELD Page1Fields[] = {
        { _T("Tcount"), 2, 0},
        { _T("TP"), 4, 3},
        { _T("r0"), 11, 5},
        { _T("Page"), 31, 12},
        { NULL,}
    };
    static REGDESC Page1 = {
        _T("Page1"), Page1Fields
    };

    static REGDESC Previous = {
        _T("Previous"), PointerFields
    };

    _tprintf(_T("%sSI-Transfer Descriptor @%x:\n"),Whitespace(nSpaces), (ADDRESS) Td);
    nSpaces++;

    Value32 = EhciReadRegister(&Td->EndpointCaps);
    DumpReg(&Caps,Value32,nSpaces);

    Value32 = EhciReadRegister(&Td->uFrameSchedule);
    DumpReg(&Schedule,Value32,nSpaces);

    Value32 = EhciReadRegister(&Td->TransferState);
    DumpReg(&TransferState,Value32,nSpaces);

    Value32 = EhciReadRegister(&Td->Page0);
    DumpReg(&Page0,Value32,nSpaces);

    Value32 = EhciReadRegister(&Td->Page1);
    DumpReg(&Page1,Value32,nSpaces);

    Value32 = EhciReadRegister(&Td->Previous);
    if (Value32)
        DumpReg(&Previous,Value32,nSpaces);

    Value32 = EhciReadRegister(&Td->Next);
    if (Value32)
        DumpReg(&Next,Value32,nSpaces);

}

PRIVATE void
DumpTdList(PEHCI_QTD Td, int nSpaces)
{
    UINT32 Value32;

    _tprintf(_T("%sTDList @%x:\n"), Whitespace(nSpaces), (ADDRESS) Td);
    nSpaces++;

    do {
        DumpTd(Td,nSpaces,FALSE);
        Value32 = EhciReadRegister(&Td->Next) & EHCI_POINTER_MASK;
        Td = (PEHCI_QTD) PhysicalToVirtual(Value32);
    } while (Value32);
}

PRIVATE void
DumpItdList(PMYEHCI_ITD Td, int nSpaces)
{
    int i;

    _tprintf(_T("%sITDList @%x:\n"), Whitespace(nSpaces), (ADDRESS) Td);
    nSpaces++;

    for (i = 0; i < USK_LEVEL1_MAX; i++) {
        DumpItd(&Td->Td,nSpaces);
        Td++;
    }
}

PRIVATE void
DumpSitdList(PMYEHCI_SITD Td, int nSpaces)
{
    int i;

    _tprintf(_T("%sSITDList @%x:\n"), Whitespace(nSpaces), (ADDRESS) Td);
    nSpaces++;

    for (i = 0; i < USK_LEVEL1_MAX; i++) {
        DumpSitd(&Td->Td,nSpaces);
        Td++;
    }
}

PRIVATE void
DumpEd(PEHCI_QH Ed, int nSpaces)
{
    UINT32 Value32;

    static REGFIELD CapsFields[] = {
        { _T("Device"),  6, 0},
        { _T("I"), 7, 7},
        { _T("EndPt"), 11, 8},
        { _T("EPS"),  13, 12},
        { _T("dtc"),  14, 14},
        { _T("H"),  15, 15},
        { _T("MPS"), 26, 16},
        { _T("C"), 27, 27},
        { _T("RL"), 31, 28},
        { NULL,}
    };
    static REGDESC Caps = {
        _T("Caps"), CapsFields
    };

    static REGFIELD CapsFields2[] = {
        { _T("Smask"),  7, 0},
        { _T("Cmask"), 15, 8},
        { _T("HubDevice"),  22, 16},
        { _T("HubPort"),  29, 23},
        { _T("Mult"), 31, 30},
        { NULL,}
    };
    static REGDESC Caps2 = {
        _T("Caps2"), CapsFields2
    };

    static REGDESC Current = {
        _T("CurrentTd"), PointerFields
    };

    _tprintf(_T("%sEndpoint @%x:\n"), Whitespace(nSpaces), (ADDRESS) Ed);
    nSpaces++;

    Value32 = EhciReadRegister(&Ed->Next);
    if (Value32)
        DumpReg(&Next,Value32,nSpaces);

    Value32 = EhciReadRegister(&Ed->EndpointCaps);
    if (Value32)
        DumpReg(&Caps,Value32,nSpaces);

    Value32 = EhciReadRegister(&Ed->EndpointCaps2);
    if (Value32)
        DumpReg(&Caps2,Value32,nSpaces);

    Value32 = EhciReadRegister(&Ed->Current);
    if (Value32)
        DumpReg(&Current,Value32,nSpaces);

    DumpTd(&Ed->TransferOverlay,nSpaces,TRUE);
}

PRIVATE void
DumpAsyncList(PEHCI_QH Head, int nSpaces)
{
    UINT32 Value32;
    PEHCI_QH Ed = Head;

    _tprintf(_T("%sAsyncList %x:\n"), Whitespace(nSpaces), (ADDRESS) Ed);
    nSpaces++;

    while (Ed) {
        DumpEd(Ed,nSpaces);
        Value32 = EhciReadRegister(&Ed->Next) & EHCI_POINTER_MASK;
        Ed = (PEHCI_QH)PhysicalToVirtual(Value32);
        if (Ed == Head)
            break;
    }

}

PRIVATE void
DumpPeriodicStride(EHCI_FRAMELIST_LINK_POINTER Value32, int nSpaces)
{
    PTR NextPtr;
    PEHCI_QH Ed;
    PEHCI_ITD Itd;
    PEHCI_SITD Sitd;

    while ((Value32 & EHCI_POINTER_MASK) &&
           ((Value32 & EHCI_POINTER_TYPE_TERMINATOR) == 0)) {

        NextPtr = PhysicalToVirtual(Value32 & EHCI_POINTER_MASK);

        if ((Value32 & EHCI_POINTER_TYPE_MASK) == EHCI_POINTER_TYPE_QH) {
            Ed = NextPtr;
            DumpEd(Ed,nSpaces++);
            Value32 = EhciReadRegister(&Ed->Next);
        }
        else if ((Value32 & EHCI_POINTER_TYPE_MASK) == EHCI_POINTER_TYPE_iTD) {
            Itd = NextPtr;
            DumpItd(Itd,nSpaces++);
            Value32 = EhciReadRegister(&Itd->Next);
        }
        else if ((Value32 & EHCI_POINTER_TYPE_MASK) == EHCI_POINTER_TYPE_siTD) {
            Sitd = NextPtr;
            DumpSitd(Sitd,nSpaces++);
            Value32 = EhciReadRegister(&Sitd->Next);
        }
        else
            return; /* UHU? */
    }
}
PRIVATE void
DumpPeriodicList(PEHCI this, int nSpaces)
{
    UINT i;
    UINT32 Value32;

    _tprintf(_T("%sPeriodicList %x:\n"), Whitespace(nSpaces), (ADDRESS) this->PeriodicList);
    nSpaces++;

    /* Full dump of the first rept of the schedule,
     * just sanity-check for the rest of it
     */
    for (i = 0;i < this->PeriodicListSize; i++) {
        Value32 = EhciReadRegister(&this->PeriodicList[i]);
        if ((Value32 & EHCI_POINTER_MASK) == 0)
            continue;

        _tprintf(_T("%s[%3d]  %x\n"),Whitespace(nSpaces),i, Value32);

        if (i < MAX_POLLING_FREQUENCY) {
            DumpPeriodicStride(Value32,nSpaces);
        }
    }
}

PRIVATE void
DumpCaps(PEHCI_CAPS Caps, int nSpaces)
{
    int i;
    UINT32 Value32;

    static REGFIELD VersionAndLengthFields[] = {
        { _T("Version"),  31, 16},
        { _T("CapLength"), 7,  0},
        { NULL,}
    };
    static REGDESC VersionAndLength = {
        _T("VersionAndLength"), VersionAndLengthFields
    };

    static REGFIELD HcsParamsFields[] = {
        { _T("DebugPort"), 23, 20},
        { _T("P_INDICATOR"), 16, 16},
        { _T("N_CC"), 15, 12},
        { _T("N_PCC"), 11, 8},
        { _T("PrRules"), 7, 7},
        { _T("PPC"), 4, 4},
        { _T("N_PORTS"), 3, 0},
        { NULL,}
    };
    static REGDESC HcsParams = {
        _T("HcsParams"), HcsParamsFields
    };

    static REGFIELD HccParamsFields[] = {
        { _T("EECP"), 15, 8},
        { _T("CachesWholeFrame"), 7, 7},
        { _T("Iso Threshold"), 6, 4},
        { _T("Park cap"), 2, 2},
        { _T("ProgFramelistLen"), 1, 1},
        { _T("64bits"), 0, 0},
        { NULL,}
    };
    static REGDESC HccParams = {
        _T("HccParams"), HccParamsFields
    };

    _tprintf(_T("%sCaps @%x:\n"), Whitespace(nSpaces), (ADDRESS) Caps);
    nSpaces++;

    Value32 = EhciReadRegister(&Caps->VersionAndLength);
    DumpReg(&VersionAndLength,Value32,nSpaces);

    Value32 = EhciReadRegister(&Caps->HcsParams);
    DumpReg(&HcsParams,Value32,nSpaces);

    Value32 = EhciReadRegister(&Caps->HccParams);
    DumpReg(&HccParams,Value32,nSpaces);

    _tprintf(_T("%sPortRoutes:"),Whitespace(nSpaces));
    for (i = 0; i < EHCI_PR_PORTROUTE_MAX; i++) {
        if ((i & 7) == 0) {
            Value32 = EhciReadRegister(&Caps->HcspPortRoute[i>>3]);
        }
        _tprintf(_T(" %2x"),Value32 & 0xf);
        Value32 >>= 4;
    }
    _tprintf(_T("\n"));
}

static REGFIELD UsbStsFields[] = {
        { _T("AsynSched"),  15, 15},
        { _T("PeriSched"),  14, 14},
        { _T("Reclamation"),  13, 13},
        { _T("HCHalted"),  12, 12},
        { _T("IntAA"),  5, 5},
        { _T("HostErr"),  4, 4},
        { _T("FramelRoll"),  3, 3},
        { _T("PortChange"),  2, 2},
        { _T("Err"),  1, 1},
        { _T("Int"),  0, 0},
        { NULL,}
};

PRIVATE void
DumpOps(PEHCI_OPS Ops, int nSpaces)
{
    int i;
    UINT32 Value32;
    static REGFIELD UsbCmdFields[] = {
        { _T("IntThreshold"),  23, 16},
        { _T("ParkEnable"), 11, 11},
        { _T("ParkCount"), 9, 8},
        { _T("LReset"), 7, 7},
        { _T("Doorbell"), 6, 6},
        { _T("AsynSchedE"), 5, 5},
        { _T("PeriSchedE"), 4, 4},
        { _T("FramelistSiz"), 3, 2},
        { _T("HCreset"), 1, 1},
        { _T("Run"), 0, 0},
        { NULL,}
    };
    static REGDESC UsbCmd = {
        _T("UsbCmd"), UsbCmdFields
    };
    static REGDESC UsbSts = {
        _T("UsbSts"), UsbStsFields
    };
    static REGFIELD UsbIntrFields[] = {
        { _T("IntAA"),  5, 5},
        { _T("HostErr"),  4, 4},
        { _T("FramelRoll"),  3, 3},
        { _T("PortChange"),  2, 2},
        { _T("Err"),  1, 1},
        { _T("Int"),  0, 0},
        { NULL,}
    };
    static REGDESC UsbIntr = {
        _T("UsbIntrEnable"), UsbIntrFields
    };

    static REGFIELD FrIndexFields[] = {
        { _T("Index"),  13, 0},
        { NULL,}
    };
    static REGDESC FrIndex = {
        _T("FrIndex"), FrIndexFields
    };

    static REGFIELD CtrlDsSegmentFields[] = {
        { _T("ADDR63..32"),  31, 0},
        { NULL,}
    };
    static REGDESC CtrlDsSegment = {
        _T("CtrlDsSegment"), CtrlDsSegmentFields
    };

    static REGFIELD PeriodicListBaseFields[] = {
        { _T("ADDR31..0"),  31, 0},
        { NULL,}
    };
    static REGDESC PeriodicListBase = {
        _T("PeriodicListBase"), PeriodicListBaseFields
    };

    static REGFIELD ConfigFlagFields[] = {
        { _T("CF"),  0, 0},
        { NULL,}
    };
    static REGDESC ConfigFlag = {
        _T("ConfigFlag"), ConfigFlagFields
    };

    static REGFIELD PortScFields[] = {
        { _T("WKOC_E"),  22, 22},
        { _T("WKDSCCT_E"),  21, 21},
        { _T("WKCNNT_E"),  20, 20},
        { _T("Test"),  19, 16},
        { _T("PIndicator"),  15, 14},
        { _T("PortOwner"),  13, 13},
        { _T("PP"),  12, 12},
        { _T("LineSt"),  11, 10},
        { _T("PortReset"),  8, 8},
        { _T("Suspend"),  7, 7},
        { _T("Resume"),  6, 6},
        { _T("OcC"),  5, 5},
        { _T("Oc"),  4, 4},
        { _T("EnC"),  3, 3},
        { _T("En"),  2, 2},
        { _T("CnC"),  1, 1},
        { _T("Cn"),  0, 0},
        { NULL,}
    };
    static REGDESC PortSc = {
        _T("PortSc"), PortScFields
    };

    _tprintf(_T("%sOps @%x:\n"), Whitespace(nSpaces), (ADDRESS) Ops);
    nSpaces++;

#define Dumpit(f) \
    Value32 = EhciReadRegister(&Ops->f); \
    DumpReg(&f,Value32,nSpaces);

    Dumpit(UsbCmd);  Dumpit(UsbSts);  Dumpit(UsbIntr); Dumpit(FrIndex);
    Dumpit(CtrlDsSegment);
    Dumpit(PeriodicListBase);
    Dumpit(ConfigFlag);
#undef Dumpit

    for (i = 0; i < 5/*N_PORTS*/; i++) {
        Value32 = EhciReadRegister(&Ops->PortSc[i]);
        _tprintf(_T("%s[%d]"),Whitespace(nSpaces),i);
        DumpReg(&PortSc,Value32,0);
    }

    Value32 = EhciReadRegister(&Ops->AsyncListAddr);
    DumpAsyncList(PhysicalToVirtual(Value32),nSpaces);

}

PRIVATE MYSTRING EhciErrors[16] = {
    _T("no"), _T("1"), _T("2"), _T("3"),
    _T("4"), _T("5"), _T("6"), _T("7"),
    _T("8"), _T("9"), _T("0xa"), _T("0xb"),
    _T("0xc"), _T("0xd"), _T("0xe"), _T("0xf")
};
#define EhciErrorName(n) EhciErrors[n&15]

PRIVATE void DumpPacket(PUSBM_REQUEST_BLOCK Packet)
{
    static REGFIELD FlagsFields[] = {
        { _T("OUT"),  0, 0},
        { _T("IN"),  1, 1},
        { _T("ISO"), 2, 2},
        { _T("CONTROL"), 3, 3},
        { _T("INTERRUPT"), 4, 4},
        { _T("LOW_SPEED"), 5, 5},
        { _T("HIGH_SPEED"), 6, 6},
        { _T("COMPLETE"), 14, 14},
        { _T("ERROR"), 15, 15},
        { NULL,}
    };
    static REGDESC Flags = {
        _T("Flags"), FlagsFields
    };

    int i, n;

    _tprintf(_T("Packet @x%x "), (ADDRESS)Packet);
    if (Packet->Flags & URB_FLAG_ERROR)
        _tprintf(_T("had a %s error\n"),EhciErrorName(Packet->FinalStatus));
    else
        _tprintf((Packet->Flags & URB_FLAG_COMPLETE) ? _T("successfully completed\n") :_T("is executing\n"));

#define indent 4
#define sp() Whitespace(indent)
    _tprintf(_T("%sDeviceAddress %x\n"), sp(), Packet->DeviceAddress);
    _tprintf(_T("%sEndpointAddress %x\n"), sp(), Packet->EndpointAddress);
    if (Packet->HubDeviceAddress) {
        _tprintf(_T("%sHUB route: %x %x\n"), sp(), 
                 Packet->HubDeviceAddress, Packet->HubPortNumber);
    }
    if (Packet->Flags & (URB_FLAG_INTERRUPT|URB_FLAG_ISOCHRONOUS))
        _tprintf(_T("%sInterval %d\n"), sp(), Packet->Interval);
    _tprintf(_T("%sPid %x\n"), sp(), Packet->Pid);
    DumpReg(&Flags,(UINT32)Packet->Flags,indent);
    _tprintf(_T("%sSetupLength %x\n"), sp(), Packet->SetupLength);
    n = Packet->SetupLength;
    if (n)
    {
        _tprintf(_T("%sSetupData[%d]@%x:"), sp(),n,(ADDRESS)Packet->SetupData);
        for (i = 0; i < n; i++)
            _tprintf(_T(" %2x"), Packet->SetupData[i]);
        _tprintf(_T("\n"));
    }
    _tprintf(_T("%sExpectedDataLength %x\n"), sp(), Packet->ExpectedDataLength);
    _tprintf(_T("%sMaximumPacketSize %d\n"), sp(), Packet->MaximumPacketSize);
    _tprintf(_T("%sActualDataLength %x\n"), sp(), Packet->ActualDataLength);
    n = Packet->ActualDataLength;
    if (n)
    {
        _tprintf(_T("%sResult[%d]@%x:"), sp(),n,(ADDRESS)Packet->Data);
        for (i = 0; i < n; i++)
            _tprintf(_T(" %2x"), Packet->Data[i]);
        _tprintf(_T("\n"));
    }
    if (Packet->Flags & URB_FLAG_ERROR)
        _tprintf(_T("%sFinalStatus %x\n"), sp(), Packet->FinalStatus);
#undef sp
#undef indent
}

PRIVATE void print_stats(struct _STATS *Stats)
{
    _tprintf(_T("--Counters:"));
    _tprintf(_T("  Isr-Interrupts = %d"), Stats->IsrCnt);
    _tprintf(_T("  Thd-Interrupts = %d\n"), Stats->IntrCnt);
    _tprintf(_T("\tUSBINT = %d"), Stats->RegCnt);
    _tprintf(_T("  Change = %d"), Stats->ChgCnt);
    _tprintf(_T("  Errors = %d\n"), Stats->ErrCnt);
}

PRIVATE void
DumpHt(PEHCI_HT Ht, int nSpaces)
{
    int i;

    _tprintf(_T("%sHashTable @%x:\n"),Whitespace(nSpaces), (ADDRESS) Ht);
    if (Ht == NULL)
        return;
    nSpaces++;

    _tprintf(_T("%s%d Slots%c\n"),Whitespace(nSpaces),
             Ht->nEntries, (Ht->nEntries > 0) ? _T(':') : _T(' '));
    for (i = 0; i < Ht->nEntries; i++) {
        /* Zero means free, dont print */
        if (Ht->Entries[i].HashValue == 0)
            continue;
        _tprintf(_T("%s[%d] Hash %x Data %x\n"),Whitespace(nSpaces+1), i,
                 Ht->Entries[i].HashValue, (ADDRESS) Ht->Entries[i].DataValue);
    }
}

PRIVATE void
DumpEhci(PEHCI this)
{
    PUSBM_REQUEST_BLOCK Urb;
    PHCI_STATE Hs;

    static REGDESC IsrSts = {
        _T("IsrStatus"), UsbStsFields
    };

    _tprintf(_T("Ehci @%x:\n"), (ADDRESS)this);

    print_stats(&this->Stats);

    DumpReg(&IsrSts,this->IsrStatus,1);
    _tprintf(_T(" PeriodicListSize %d"), this->PeriodicListSize);
    _tprintf(_T(" nPorts %d"), this->nPorts);
    _tprintf(_T(" Ppc %d"), this->PortPowerControl);
    _tprintf(_T(" EvtP %d"), this->EventPending);
    _tprintf(_T(" FreeTds %d"), this->FreeTdCount);

    DumpHt(this->HashTable,1);
    DumpCaps(this->Caps,1);
    DumpOps(this->Ops,1);
    DumpPeriodicList(this,1);

    for (Urb = this->ActiveUrbs; Urb != NULL; Urb = Hs->Next) {
        Hs = (PHCI_STATE)&Urb->HciReserved;

        _tprintf(_T(" ActiveUrb"));
        DumpPacket(Urb);

        _tprintf(_T(" TdForUrb"));
        DumpTdList(&Hs->u.Td->Td,1);
    }

    for (Urb = this->ActiveIsoUrbs; Urb != NULL; Urb = Hs->Next) {
        Hs = (PHCI_STATE)&Urb->HciReserved;

        _tprintf(_T(" ActiveIsoUrb"));
        DumpPacket(Urb);

        /* ITD or SITD ?
         */
        if (Hs->u.Tag & HCI_SITD_TAG)
            DumpSitd( & GetSitd(Hs->u.Sitd)->Td,1);
        else
            DumpItd( & GetItd(Hs->u.Itd)->Td,1);
    }

    if (this->TrashCan) {
        PMYEHCI_QH Trash = this->TrashCan;

        _tprintf(_T(" TrashCan"));
        while (Trash) {
            DumpEd(&Trash->Ed,1);
            Trash = Trash->TrashCan;
        }
    }
}
#else
#define DumpEhci(x)
#endif

/*
 * CDevice interface
 */

PRIVATE SCODE EhciDeviceQueryInterface (
    PIDEVICE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_IFile,iid)
        || UuidCmp(&IID_IDevice,iid))
    {
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }
    if (UuidCmp(&IID_IUsbMasterController, iid)) {
        /* Use a single refcount
         */
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (void*) iEhciMaster(pThis);
        return S_OK;
    }

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT EhciDeviceAddRef (
    PIDEVICE pThis)
{
    return AtomicInc(&pEhciDevice(pThis)->CDevice.RefCnt);
}

PRIVATE UINT EhciDeviceRelease (
    PIDEVICE pThis)
{
    PEHCI this = pEhciDevice(pThis);
    UINT Refs = AtomicDec(&this->CDevice.RefCnt);
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        pHeap->v->Free(pHeap, 0, (void *) this);
    }
    return Refs;
}

#if _DEBUG
PRIVATE SCODE MCT EhciDeviceReadAt (PIDEVICE pThis, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    PEHCI this = pEhciDevice(pThis);

    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);

    DumpEhci(this);

    if (pSizeRead)
        *pSizeRead = 0;
    return S_FALSE; /* EOF */

}
#else
#define EhciDeviceReadAt NULL
#endif

PRIVATE SCODE EhciDeviceGetSize (
    PIDEVICE pThis,
    /*out*/ UINT64 * Size)
{
    PEHCI this = pEhciDevice(pThis);
    UINT32 Ret = this->nPorts | (2 << 16);

    /* Number of ports on root hub, plus indicate we are 2.0
     */
    Uint32ToUint64(*Size,Ret);

    return S_OK;
}


/* Interrupt handling
 */
#define IMASK (EHCI_STS_HOST_SYSTEM_ERROR | EHCI_STS_PORT_CHANGE | \
               EHCI_STS_USBERRINT | EHCI_STS_USBINT)
#define BELLM (IMASK | EHCI_STS_ASYNC_INT)
#define EhciDisableInterrupts(Ops) EhciWriteRegister(&Ops->UsbIntr,0)
#define EhciEnableInterrupts(Ops)  EhciWriteRegister(&Ops->UsbIntr,IMASK)
#define EhciEnableDoorbellInterrupts(Ops)  EhciWriteRegister(&Ops->UsbIntr,BELLM)

PRIVATE BOOL EhciDeviceIsr (
    PIDEVICE pThis,
    /*out*/ BOOL * pNotOurInterrupt)
{
    PEHCI this = pEhciDevice(pThis);
    PEHCI_OPS Ops = this->Ops;
    UINT32 Status;

    this->Stats.IsrCnt++;

    /* This interrupt is shared, we must know which is ours
     */
    Status = EhciReadRegister(&Ops->UsbSts) & BELLM;
    if (Status == 0) {
        *pNotOurInterrupt = TRUE;
        return FALSE;
    }

    /* Disable further interrupts
     */
    EhciDisableInterrupts(Ops);

    /* The designer of the interrupt status register deserves to be shot
     */
    this->IsrStatus = (UINT16) Status;
    EhciWriteRegister(&Ops->UsbSts,Status);

    return Condition_InterruptSignal(&this->CDevice.IntrPoint);
}

/* An endpoint has been halted because of some error (STALL, most likely)
 * Return an error code to complete the transfer with and restart the endpoint.
 */
PRIVATE UINT8
EhciKickEd(PEHCI this, PEHCI_QH Ed, UINT32 LastTdPhysical)
{
    UINT32 Value32;
    PEHCI_QTD LastTdVirtual = PhysicalToVirtual(LastTdPhysical);
    UINT8 Status;

    /* Zero out the current TD pointer
     */
    Ed->Current = 0;

    /* Terminate the alternate
     */
    Ed->TransferOverlay.Alternate = _bs32(EHCI_POINTER_TYPE_TERMINATOR);

    /* Pick up status
     */
    Status = (UINT8)
        (EhciReadRegister(&Ed->TransferOverlay.Token) & EHCI_QTD_STATUS);

    /* Move the overlay on to the next TD
     */
    Ed->TransferOverlay.Next = LastTdVirtual->Next;

    /* Leave the endpoint halted unless it is a control one
     */
    Value32 = EhciReadRegister(&Ed->EndpointCaps) & EHCI_QH_ENDPT_MASK;
    Ed->TransferOverlay.Token = (Value32 == 0) ? 0 : _bs32(EHCI_QTD_HALTED);

    return Status;
}

/* One or more transfers have completed, with or without error.
 */
PRIVATE void
EhciHandleCompletions(PEHCI this, BOOL WithErrors)
{
    PUSBM_REQUEST_BLOCK *Previous = &this->ActiveUrbs;
    PUSBM_REQUEST_BLOCK Urb;
    PHCI_STATE Hs;
    PMYEHCI_QTD Td;
    PMYEHCI_QH Ed;
    UINT8 ErrorCode;
    UINT32 Value32, LastOne;
    
    for (Urb = *Previous; Urb != NULL; Urb = *Previous) {

        Hs = (PHCI_STATE)&Urb->HciReserved;
        Td = Hs->u.Td;
        LastOne = EhciReadRegister(&Td->Td.Alternate) & EHCI_POINTER_MASK;
        Value32 = VirtualToPhysical(&Td->Td);

        /* Walk the TD list to see that all is said and done.
         * Bail out as soon as we find one that is still active.
         */
        ErrorCode = 0;
        for (;;) {
            /* If we had errors we have to look at the QH also and see if its 
             * halted on this TD.
             */
            if (WithErrors) {
                Ed = Td->Ed;
                if (Ed->Ed.TransferOverlay.Token & _bs32(EHCI_QTD_HALTED)) {
                    ErrorCode = EhciKickEd(this,&Ed->Ed,LastOne);
                    goto IsComplete;
                }
            }

            if (Td->Td.Token & _bs32(EHCI_QTD_ACTIVE))
                goto NextUrb;

            /* If this is the last TD for this xfer make sure that 
             * we get out of this loop.  The overlay has been performed
             * and therefore it is now safe to clobber the next link
             */
            if (Value32 == LastOne)
                Td->Td.Next = _bs32(EHCI_POINTER_TYPE_TERMINATOR);

            /* Adjust the xfer count
             */
            Value32 = EhciReadRegister(&Td->Td.Token);
            if ((Value32 & EHCI_QTD_PID) != EHCI_QTD_PID_SETUP) {

#define CORRECTO 1
#if CORRECTO
                Urb->ActualDataLength -= (UINT16)
                    ((Value32 & EHCI_QTD_LEN) >> EHCI_QTD_LENSHIFT);
#endif
            }

            /* Check status
             */
            ErrorCode |= (UINT8)(Value32 & (EHCI_QTD_HALTED
                                            |EHCI_QTD_DATA_ERROR
                                            |EHCI_QTD_BABBLE
                                            |EHCI_QTD_XactErr
                                            |EHCI_QTD_uFrame_Miss));

            /* If halted the rest of it will never complete
             * Notice that this bit is not always written back.
             */
            if (Value32 & EHCI_QTD_HALTED) {
                break;
            }

            /* End of the list?
             */
            Value32 = EhciReadRegister(&Td->Td.Next) & EHCI_POINTER_MASK;
            if (Value32 == 0)
                break;
            Td = (PMYEHCI_QTD) PhysicalToVirtual(Value32);
        }

    IsComplete:
#if _DEBUG
        if (ErrorCode)
            DumpTdList(&Hs->u.Td->Td,0);
#endif        

        /* Free up the completed TD list
         */
        TdFree(this,Hs->u.Td,TRUE);

        /* Avoid racing CancelUrb by deleting the TD field
         */
        Hs->u.Td = NULL;

        /* Finalize the URB state
         */
        Urb->Flags |= (ErrorCode) ? (URB_FLAG_COMPLETE|URB_FLAG_ERROR)
                                     : URB_FLAG_COMPLETE;

        Urb->FinalStatus = (ErrorCode == EHCI_QTD_HALTED) ?
            URB_STATUS_STALL : ErrorCode;

        /* This URB has completed, unlink it
         */
        *Previous = Hs->Next;
#if _DEBUG
        if (Urb->Flags & URB_FLAG_ERROR) {
            DBGME(3,DumpPacket(Urb));
        } else {
            DBGME(0,DumpPacket(Urb));
        }
#endif
        /* Wakeup sleepers
         */
        Condition_InterruptSignal(Urb->Done);

        /* Done, next URB
         */
        continue;

        /* The URB has not completed, leave it in place and move on to the next
         */
    NextUrb:
        Previous = &Hs->Next;
    }
}

/* One or more isochronous transfers have completed, with or without errors.
 */
PRIVATE void
EhciHandleIsoCompletions(PEHCI this, BOOL WithErrors)
{
    PUSBM_REQUEST_BLOCK *Previous = &this->ActiveIsoUrbs;
    PUSBM_REQUEST_BLOCK Urb;
    PHCI_STATE Hs;
    PMYEHCI_ITD Itd;
    PMYEHCI_SITD Sitd;
    UINT8 ErrorCode;
    UINT32 Value32, Xvalue32;
    int t;
    
    for (Urb = *Previous; Urb != NULL; Urb = *Previous) {

        /* Check the status of this URB
         */
        Hs = (PHCI_STATE)&Urb->HciReserved;

        /* ITD or SITD ?  Set SITD to non-NULL IFF this is a split-xact.
         * Make sure it finished posting;
         */
        if ((Itd = GetItd(Hs->u.Itd)) == NULL)
            goto NextUrb;
        Sitd = NULL;
        if (Hs->u.Tag & HCI_SITD_TAG)
            Sitd = GetSitd(Hs->u.Sitd);

        /* Check if it has completed, skip this URB if not
         */
        if (Sitd) {
            Value32 = EhciReadRegister(&Sitd->Td.TransferState);
            if (Value32 & EHCI_SITD_ACTIVE)
                goto NextUrb;
        } else {
            for (t = 0; t < 8; t++) {
                Value32 = Itd->Td.Transaction[t];
                if (Value32 & _bs32(EHCI_ITD_ACTIVE))
                    goto NextUrb;
                if (Value32 & _bs32(EHCI_ITD_IOC))
                    break;
            }
        }

        /* Check to see if there are errors, what they are.
         */
        ErrorCode = 0;

        /* Which descriptor type is it
         */
        if (Sitd) {

            /* NB: We already read the TransferStatus above
             */

            /* Check status
             */
            ErrorCode |= (UINT8)(Value32 & (EHCI_SITD_ERR
                                            |EHCI_SITD_DATA_ERROR
                                            |EHCI_SITD_BABBLE
                                            |EHCI_SITD_XactErr
                                            |EHCI_SITD_uFrame_Miss));
            /* Adjust the xfer count
             */
            Value32 = (Value32 & EHCI_SITD_BYTELENGTH) >> EHCI_SITD_LSHIFT;

#if CORRECTO
            /* This is wrong */
            Urb->ActualDataLength -= (UINT16)Value32;
#endif

        } else {

            Value32 = 0;
            Xvalue32 = 0;
            /* Stop at the last transaction (it has the IOC bit set)
             */
            for (t = 0; (t < 8) && (0 == (Xvalue32 & EHCI_ITD_IOC)); t++) {

                Xvalue32 = EhciReadRegister(&Itd->Td.Transaction[t]);

                /* Check status, make compat with SITD
                 */
                ErrorCode |= (UINT8)((Xvalue32 & (EHCI_ITD_DATA_ERROR
                                                  |EHCI_ITD_BABBLE
                                                  |EHCI_ITD_XactErr))
                                     >> EHCI_ITD_STATUS_SHIFT);
                /* Adjust the xfer count
                 */
                Value32 += (Xvalue32 & EHCI_ITD_BYTELENGTH) >> EHCI_ITD_LSHIFT;
            }
#if CORRECTO
            /* Unclear */
            Urb->ActualDataLength = (UINT16)Value32;
#endif

        }

        /* Finalize the URB state
         */
        Urb->Flags |= URB_FLAG_COMPLETE;
        if (ErrorCode) {
            Urb->Flags |= URB_FLAG_ERROR;
#if _DEBUG
            if (Sitd) {
                DumpSitd(&Sitd->Td,0);
            } else {
                DumpItd(&Itd->Td,0);
            }
#endif
        }

        Urb->FinalStatus = ErrorCode;

        /* This URB has completed, unlink it
         */
        *Previous = Hs->Next;
#if _DEBUG
        if (Urb->Flags & URB_FLAG_ERROR) {
            DBGME(3,DumpPacket(Urb));
        } else {
            DBGME(0,DumpPacket(Urb));
        }
#endif

        /* Wakeup sleepers
         */
        Condition_InterruptSignal(Urb->Done);

        /* Done, next URB
         */
        continue;

        /* The URB has not completed, leave it in place and move on to the next
         */
    NextUrb:
        Previous = &Hs->Next;
    }
}

/* Thread-level interrupt handling
 */
PRIVATE BOOL MCT
EhciDeviceHandleInterrupt(PIDEVICE pThis)
{
    PEHCI this = pEhciDevice(pThis);
    PEHCI_OPS Ops = this->Ops;
    UINT32 Status;

    /* Read the status as it was at interrupt time
     */
    Status = this->IsrStatus;
 Again:
    /* Clear the interrupt bits that we'll be handling.
     * NB: This might or might not be a repeat of what the ISR did.
     */
    EhciWriteRegister(&Ops->UsbSts,Status);
    this->Stats.IntrCnt++;

    /* Check first for the most likely cases
     */
    if (Status & (EHCI_STS_USBINT | EHCI_STS_USBERRINT)) {
        this->Stats.RegCnt++;
        DBGME(0,_tprintf(_T("EhciIntr: int Sts %x Cnt %d\n"),
                         Status, this->Stats.IsrCnt));
        EhciHandleCompletions(this, (Status & EHCI_STS_USBERRINT) != 0);
        if (this->ActiveIsoUrbs)
            EhciHandleIsoCompletions(this, (Status & EHCI_STS_USBERRINT) != 0);
    }

    /* Check for doorbell interrupt
     */
    if (Status & EHCI_STS_ASYNC_INT) {
        DBGME(0,_tprintf(_T("EhciIntr: bell Sts %x Cnt %d\n"),
                         Status, this->Stats.IsrCnt));
        EhciEmptyTheTrashCan(this);
    }

    /* Check for port status changes
     */
#define EHCI_PS (EHCI_PS_OVERCURRENT|EHCI_PS_ENABLED|EHCI_PS_CONNECTED| \
                 EHCI_PS_PORT_SUSPEND|EHCI_PS_PORT_RESET|EHCI_PS_PP)
#define EHCI_PC (EHCI_PS_OVERCURRENT_CHANGE|EHCI_PS_ENABLED_CHANGE|EHCI_PS_CONNECTED_CHANGE)

    if (Status & EHCI_STS_PORT_CHANGE) {
        UINT8 i;
        UINT32 Value32;
        PUSBM_REQUEST_BLOCK Urb;

        this->Stats.ChgCnt++;

        this->PortStatusChanges = 0;
        for (i = 0; i < this->nPorts; i++) {
            Value32 = EhciReadRegister(&Ops->PortSc[i]);
            if (Value32 & EHCI_PC) {
                this->PortStatusChanges |= (2 << i);
            }
        }
        /* ??Disable STS_PORT_CHANGE interrupt?? */

        /* Notify any waiters
         */
        Urb = this->EventsPacket;
        if (Urb) {
            this->EventsPacket = NULL;
            Urb->ActualDataLength = 2;
            Set2(Urb->Data,this->PortStatusChanges);
            Urb->Flags |= URB_FLAG_COMPLETE;
            Condition_InterruptSignal(Urb->Done);
        }
    }

    /* Check for errors
     */
    if (Status & EHCI_STS_HOST_SYSTEM_ERROR) {
        this->Stats.ErrCnt++;
        _tprintf(_T("EhciIntr: err Sts %x Cnt %d\n"),
                 Status, this->Stats.IsrCnt);
        DumpEhci(this);
    }

    /* Check again for pending interrupts
     */
    Status = EhciReadRegister(&Ops->UsbSts) & IMASK;
    if (Status)
        goto Again;

    /* We are done this round, reenable and go back to sleep
     */
    EhciEnableInterrupts(Ops);
    return TRUE;
}


PRIVATE INT EhciDeviceInitialize (
    PIDEVICE pThis)
{
    PEHCI this = pEhciDevice(pThis);
    PEHCI_OPS Ops = this->Ops;
    UINT32 Value32;
    int i;

    DBGME(0,_tprintf(_T("EhciInitialize(%x)\n"), (ADDRESS)pThis));

    /* Issue HCReset and wait.
     */
    Value32 = EhciReadRegister(&Ops->UsbCmd);
    Value32 &= ~(EHCI_UC_RUN | 
                 EHCI_UC_PERIODIC_ENABLE | EHCI_UC_ASYNCH_ENABLE);
    Value32 |= EHCI_UC_HCRESET;
    EhciWriteRegister(&Ops->UsbCmd,Value32);

    for (i = 0; i < 1000; i++) {
        Delay(10);
        Value32 = EhciReadRegister(&Ops->UsbCmd);
        if ((Value32 & EHCI_UC_HCRESET) == 0)
            break;
    }
    if (Value32 & EHCI_UC_HCRESET) {
        DBGME(3,_tprintf(_T("EHCI @%x HCreset stuck.\n"),
                         (ADDRESS) this->Caps));
        return 1;
    }

    /* Disable interrupts
     */
    EhciDisableInterrupts(Ops);

    /* Reset pointers
     */
    EhciWriteRegister(&Ops->CtrlDsSegment,0);
    EhciWriteRegister(&Ops->AsyncListAddr,
                      VirtualToPhysical(this->AsyncList));
    EhciWriteRegister(&Ops->PeriodicListBase,
                      VirtualToPhysical(this->PeriodicList));

    /* Enable Async Park Mode if supported
     */
    Value32 = EhciReadRegister(&this->Caps->HccParams);
    if (Value32 & EHCI_HCS_ASYNCH_SCHEDULE_PARK) {
        Value32 = EHCI_UC_PARK_MODE | EHCI_UC_PARK_COUNT;
    } else
        Value32 = 0;
    Value32 |= EHCI_UC_8uf | EHCI_UC_RUN;

    /* Get it running
     */
    EhciWriteRegister(&Ops->UsbCmd,
                      Value32
                      | ((this->PeriodicListSize == 256) ?
                         EHCI_UC_FRAMELIST_256 : EHCI_UC_FRAMELIST_1024));
    Delay(20);

    /* Say we have configured it
     */
    EhciWriteRegister(&Ops->ConfigFlag,EHCI_CF);

    /* Powerup all ports, if we have to
     */
    if (this->PortPowerControl) {
        for (i = 0; i < this->nPorts; i++) {
            Value32 = EhciReadRegister(&Ops->PortSc[i]);
            Value32 |= EHCI_PS_PP;
            EhciWriteRegister(&Ops->PortSc[i],Value32);
        }
    }

    /* Assume a POTPG of 20msecs
     */
    MilliSleep(20);

    /* Now turn on interrupts and be ready for connect change events
     */
    EhciEnableInterrupts(Ops);

    return 0;
}

/*
 * IUsbMasterController interface
 */
PRIVATE SCODE EhciMasterQueryInterface (
    PIUSBMASTERCONTROLLER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return EhciDeviceQueryInterface(iEhciDevice(pEhciMaster(pThis)), iid, ppObject);
}

PRIVATE UINT EhciMasterAddRef (
    PIUSBMASTERCONTROLLER pThis)
{
    return EhciDeviceAddRef(iEhciDevice(pEhciMaster(pThis)));
}

PRIVATE UINT EhciMasterRelease (
    PIUSBMASTERCONTROLLER pThis)
{
    return EhciDeviceRelease(iEhciDevice(pEhciMaster(pThis)));
}

/* Insert an interrupt endpoint in the periodic schedule
 */
PRIVATE void
EhciInsertInterruptEd(PEHCI this,
                      PMYEHCI_QH Ed)
{
    int Frequency, StartSlot;
    EHCI_FRAMELIST_LINK_POINTER *NextPtr;
    PMYEHCI_QH NextEd;
    UINT32 Value32;

    /* See where it goes first
     */
    Frequency = UskFrequency(Ed->InterruptHead);
    StartSlot = UskStart(Ed->InterruptHead);

    /* Its replicated a few times (at least 256/16==16)
     */
    for (;;) {

        /* Skip all ISO TDs, they must-go-first
         */
        NextPtr = &this->PeriodicList[StartSlot];

        for (;;) {
            Value32 = EhciReadRegister(NextPtr);

            /* Stop at the first QH or at the end of the list
             */
            if ((Value32 & EHCI_POINTER_TYPE_TERMINATOR) ||
                ((Value32 & EHCI_POINTER_TYPE_MASK) == EHCI_POINTER_TYPE_QH))
                break;

            NextPtr = PhysicalToVirtual(Value32 & EHCI_POINTER_MASK);
        }

        /* Skip any endpoint with a lower polling frequency
         */
        for (NextEd = NULL; Value32 & EHCI_POINTER_MASK;) {

            NextEd = (PMYEHCI_QH) PhysicalToVirtual(Value32&EHCI_POINTER_MASK);

            if (UskIsLowerOrSame(NextEd->InterruptHead,Ed->InterruptHead))
                break;

            NextPtr = &NextEd->Ed.Next;
            Value32 = EhciReadRegister(NextPtr);
        }

        /* Make sure we dont do this twice
         */
        if (NextEd != Ed) {

            /* Link the endpoint to the next one, or null.
             */
            EhciWriteRegister(&Ed->Ed.Next,Value32);
            Value32 = VirtualToPhysical(Ed) | EHCI_POINTER_TYPE_QH;

            /* If we are writing into the roots we must do it a few times,
             * otherwise its easy.
             */
            if (NextPtr != &this->PeriodicList[StartSlot]) {

                DBGME(0,_tprintf(_T("EhciIIEd: Node %x after %x\n"),
                                 (ADDRESS) Ed, (ADDRESS) NextPtr));
                EhciWriteRegister(NextPtr,Value32);

            } else {

                int RootIndex;
                for (RootIndex = StartSlot;
                     RootIndex < this->PeriodicListSize;
                     RootIndex += MAX_POLLING_FREQUENCY) {

                    DBGME(0,_tprintf(_T("EhciIIEd: Node %x @root %d\n"),
                                     (ADDRESS) Ed, RootIndex));
                    EhciWriteRegister(&this->PeriodicList[RootIndex],Value32);
                }
            }
        }

        /* Go look at the next slot, if any
         */
        StartSlot += Frequency;

        if (StartSlot >= USK_LEVEL1_MAX)
            break;
    }
}

/* Insert a list of ISO ITDs in the schedule
 */
PRIVATE void
EhciInsertIsochronousUsb2Ed(PEHCI this,
                            PMYEHCI_QH Ed,
                            BOOL IsInput)
{
    int StartSlot;
    UINT32 Value32, page0, page1;
    PMYEHCI_ITD Itd = Ed->Itd;

    DBGME(0,_tprintf(_T("EhciIIU2Ed: Node %x %d\n"),
                     (ADDRESS) Ed, IsInput));

    /* Map from the ED info to the SITD info
     */
    Value32 = EhciReadRegister(&Ed->Ed.EndpointCaps);

    page0 = Value32 & (EHCI_QH_DEVICE_MASK | EHCI_QH_ENDPT_MASK);
    page1 = (Value32 & EHCI_QH_MAX_PACKET_SIZE) >> EHCI_QH_MPS_SHIFT;
    if (IsInput)
        page1 |= EHCI_ITD_INPUT;
    
    /* Swap if necessary */
    page0 = _bs32(page0);
    page1 = _bs32(page1);

    /* The schedule is replicated a few times (at least 256/16==16)
     */
    for (StartSlot = 0; StartSlot < MAX_POLLING_FREQUENCY; StartSlot++) {

        /* Init this Itd
         */
        Itd->Td.Transaction[0] = 0;
        Itd->Td.Transaction[1] = 0;
        Itd->Td.Transaction[2] = 0;
        Itd->Td.Transaction[3] = 0;
        Itd->Td.Transaction[4] = 0;
        Itd->Td.Transaction[5] = 0;
        Itd->Td.Transaction[6] = 0;
        Itd->Td.Transaction[7] = 0;

        Itd->Td.Pointer[0] = page0;
        Itd->Td.Pointer[1] = page1;
        Itd->Td.Pointer[2] = _bs32(EHCI_ITD_MULT_1xuf);
        Itd->Td.Pointer[3] = 0;
        Itd->Td.Pointer[4] = 0;
        Itd->Td.Pointer[5] = 0;
        Itd->Td.Pointer[6] = 0;

        /* The ISO TDs must-go-first
         */
        Itd->Td.Next = this->PeriodicList[StartSlot];

        Value32 = VirtualToPhysical(&Itd->Td);
        Value32 |= EHCI_POINTER_TYPE_iTD;

        /* Link it into all the repetitions of this slot
         */
        {
            int RootIndex;
            for (RootIndex = StartSlot;
                 RootIndex < this->PeriodicListSize;
                 RootIndex += MAX_POLLING_FREQUENCY) {

                DBGME(0,_tprintf(_T("EhciIsoIIEd: Node %x @root %d\n"),
                                 (ADDRESS) Itd, RootIndex));
                EhciWriteRegister(&this->PeriodicList[RootIndex],Value32);
            }
        }

        /* Go look at the next slot, if any
         */
        Itd++;
    }
}

/* Insert a list of ISO SITDs in the schedule
 */
PRIVATE void
EhciInsertIsochronousUsb1Ed(PEHCI this,
                            PMYEHCI_QH Ed,
                            BOOL IsInput)
{
    int StartSlot;
    UINT32 Value32, Caps2, EndpointCaps, uFrameSchedule;
    PMYEHCI_SITD Itd = Ed->Itd;

    DBGME(0,_tprintf(_T("EhciIIU1Ed: Node %x %d\n"),
                     (ADDRESS) Ed, IsInput));

    /* Map from the ED info to the SITD info
     */
    Value32 = EhciReadRegister(&Ed->Ed.EndpointCaps);
    Caps2   = EhciReadRegister(&Ed->Ed.EndpointCaps2);

    EndpointCaps = (Value32 & (EHCI_QH_DEVICE_MASK | EHCI_QH_ENDPT_MASK))
        | (Caps2 & EHCI_QH_HUB_MASK)
        | ((Caps2 & EHCI_QH_PORT_MASK) << 1);
    if (IsInput)
        EndpointCaps |= EHCI_SITD_INPUT;
    
    StartSlot = UskL2(Ed->InterruptHead);
    uFrameSchedule = (1 << StartSlot);
    if (IsInput)
        uFrameSchedule |= (7 << (StartSlot + 2)) << EHCI_SITD_Cmask_SHIFT;

    /* Swap if necessary */
    EndpointCaps = _bs32(EndpointCaps);
    uFrameSchedule = _bs32(uFrameSchedule);

    /* The schedule is replicated a few times (at least 256/16==16)
     */
    for (StartSlot = 0; StartSlot < MAX_POLLING_FREQUENCY; StartSlot++) {

        /* Init this Itd
         */
        Itd->Td.EndpointCaps = EndpointCaps;
        Itd->Td.uFrameSchedule = uFrameSchedule;
        Itd->Td.TransferState = 0;
        Itd->Td.Page0 = 0;
        Itd->Td.Page1 = 0;
        Itd->Td.Previous = _bs32(EHCI_POINTER_TYPE_TERMINATOR);

        /* The ISO TDs must-go-first
         */
        Itd->Td.Next = this->PeriodicList[StartSlot];

        Value32 = VirtualToPhysical(&Itd->Td);
        Value32 |= EHCI_POINTER_TYPE_siTD;

        /* Link it into all the repetitions of this slot
         */
        {
            int RootIndex;
            for (RootIndex = StartSlot;
                 RootIndex < this->PeriodicListSize;
                 RootIndex += MAX_POLLING_FREQUENCY) {

                DBGME(0,_tprintf(_T("EhciIsoIIEd: Node %x @root %d\n"),
                                 (ADDRESS) Itd, RootIndex));
                EhciWriteRegister(&this->PeriodicList[RootIndex],Value32);
            }
        }

        /* Go look at the next slot, if any
         */
        Itd++;
    }
}

/* Fillup one ITD with the proper buffer pointers
 */
PRIVATE PMYEHCI_ITD
EhciItdIsoTransfer(PEHCI this,
                    PUSBM_REQUEST_BLOCK Urb,
                    PMYEHCI_QH Ed)
{
    int Transaction, LastPage;
    PMYEHCI_ITD Itd = Ed->Itd;
    UINT32 Value32, LastPagePointer;
    UINT8 *BufferPointer;
    int nLeft, Length;

    BufferPointer = Urb->Data;
    nLeft = Urb->ExpectedDataLength;
    Itd += Urb->Interval & (USK_LEVEL1_MAX-1);

    /* Max 8 xactions per ITD
     */
    LastPage = -1;
    LastPagePointer = 1;/* make sure wont match at start */
    for (Transaction = 0; Transaction < 8; Transaction++) {

        /* Stop when we have covered the whole buffer
         */
        if (nLeft) {

          /* Set a new page pointer if we have to
           */
          Value32 = VirtualToPhysical(BufferPointer);
          if ((Value32 & EHCI_ITD_BUFFER) != LastPagePointer) {
              LastPagePointer = EhciReadRegister(&Itd->Td.Pointer[++LastPage]);
              LastPagePointer &= ~EHCI_ITD_BUFFER;
              EhciWriteRegister(&Itd->Td.Pointer[LastPage],
                                (Value32 & EHCI_ITD_BUFFER) | LastPagePointer);
              LastPagePointer = Value32 & EHCI_ITD_BUFFER;
          }

          /* Cap and adjust the length
           */
          Length = nLeft;
          if (Length > Urb->MaximumPacketSize)
              Length = Urb->MaximumPacketSize;
          if (Length > 1024)
              Length = 1024;
          if (LastPagePointer != ((Value32 + Length) & EHCI_ITD_BUFFER))
              Length = (LastPagePointer+4096) - Value32;

          /* Adjust pointers and counts for next round
           */
          nLeft -= Length;
          BufferPointer += Length;

          /* Activate this xaction in the ITD
           * BUGBUG This is risky, what if we get stopped?
           * => should dequeue the ITD, fill it, queue it back.
           */
          Value32 = EHCI_ITD_ACTIVE
              | (EHCI_ITD_BYTELENGTH & (Length << EHCI_ITD_LSHIFT))
              | (EHCI_ITD_PG & (LastPage << EHCI_ITD_PGSHIFT))
              | (EHCI_ITD_OFFSET & Value32);
          /* Last xfer?
           */
          if ((nLeft == 0) || (Transaction == 7)) {
              Value32 |= EHCI_ITD_IOC;
          }

          EhciWriteRegister(&Itd->Td.Transaction[Transaction],Value32);

        } else {
          Itd->Td.Transaction[Transaction] = 0;
        }
    }
    return Itd;
}

/* Fillup one SITD with the proper buffer pointers
 */
PRIVATE PMYEHCI_SITD
EhciSitdIsoTransfer(PEHCI this,
                    PUSBM_REQUEST_BLOCK Urb,
                    PMYEHCI_QH Ed)
{
    PMYEHCI_SITD Itd = Ed->Itd;
    UINT32 Value32;
    UINT8 *BufferPointer;
    int Length;

    BufferPointer = Urb->Data;
    Length = Urb->ExpectedDataLength;
    Itd += Urb->Interval & (USK_LEVEL1_MAX-1);

    /* Set the page pointers
     */
    Value32 = VirtualToPhysical(BufferPointer);
    EhciWriteRegister(&Itd->Td.Page0,Value32);

    Value32 = VirtualToPhysical((((ADDRESS)BufferPointer)+4096)& ~4095);
    if (Urb->Flags & URB_FLAG_OUT)
        Value32 |= 1; /* Tcount */
    /*Value32 |= EHCI_SITD_TP_ALL; */
    EhciWriteRegister(&Itd->Td.Page1,Value32);

    /* Cap and adjust the length
     */
    if (Length > Urb->MaximumPacketSize)
        Length = Urb->MaximumPacketSize;

    /* Activate this SITD
     */
    Value32 = EHCI_SITD_ACTIVE
        | (EHCI_SITD_BYTELENGTH & (Length << EHCI_SITD_LSHIFT))
        | EHCI_SITD_IOC;

    EhciWriteRegister(&Itd->Td.TransferState,Value32);

    return MakeSitd(Itd);
}

/* Submit an isochronous request.
 * Iso requests require ITDs.
 */
PRIVATE SCODE
EhciSubmitIso(PEHCI this,
              PUSBM_REQUEST_BLOCK Urb,
              UINT8 PortNumber)
{
    PMYEHCI_QH Ed = NULL;
    UINT32 Value32;
    PHCI_STATE Hs;
    SCODE sc = E_NOT_ENOUGH_MEMORY;

    /* Find the ED where this URB should be posted
     * This is a "virtual" ED, used only in software.
     * This might allocate a new ED if we aint seen it before
     */
    Ed = LookupEd(this,Urb);
    if (Ed == NULL) {
        DBGME(3,_tprintf(_T("Ehci@%x: ?Ed for %x.%x.%x (%x.%x)\n"),
                         (ADDRESS) this,
                         Urb->DeviceAddress,
                         Urb->EndpointAddress,
                         Urb->Flags,
                         Urb->HubDeviceAddress,
                         Urb->HubPortNumber));
        goto ErrorHandler;
    }

    /* Check if the Ed was ever used before.
     */
    if (Ed->Ed.EndpointCaps == 0) {

        UINT16 MaxStartBw, MinStartBw;
        UINT16 MaxEndBw, MinEndBw;
        PUSKNODE Node;

        /* Nope, initialize it
         */
        memset(Ed,0,sizeof *Ed);

        /* caps, both */
        Value32 =
            (Urb->MaximumPacketSize << EHCI_QH_MPS_SHIFT)
            | (Urb->DeviceAddress & EHCI_QH_DEVICE_MASK)
            | ((Urb->EndpointAddress<<EHCI_QH_ENDPT_SHIFT) & EHCI_QH_ENDPT_MASK)
            ;

        if (Urb->Flags & URB_FLAG_HIGH_SPEED) {
            /* caps */
            Value32 |= EHCI_QH_480Mbps;
            EhciWriteRegister(&Ed->Ed.EndpointCaps,Value32);

            /* descs */
            Ed->Itd = ItdAlloc(this);

            /* caps2 */
            Value32 = 0;
        } else {
            /* caps */
            if (Urb->Flags & URB_FLAG_LOW_SPEED)
                Value32 |= EHCI_QH_1_5Mbps;
            /* else 12Mbps which is encoded as 0 */
            EhciWriteRegister(&Ed->Ed.EndpointCaps,Value32);

            /* descs */
            Ed->Itd = SitdAlloc(this);

            /* caps2 */
            Value32 =
              ((Urb->HubDeviceAddress << EHCI_QH_HUB_SHIFT) & EHCI_QH_HUB_MASK)
              | ((Urb->HubPortNumber << EHCI_QH_PORT_SHIFT) & EHCI_QH_PORT_MASK);
        }
        EhciWriteRegister(&Ed->Ed.EndpointCaps2,Value32);

        /* Backtrack if we didnt get the memory
         */
        if (Ed->Itd == NULL) {
            Ed->Ed.EndpointCaps = 0;
            goto ErrorHandler;
        }

        /* For the isochronous case, the first time through make a
         * reservation for the necessary amount of b/w.
         * This reservation will only be cancelled when the device
         * is unplugged.  Notice that this happens only when the
         * endpoint is actually used for the first time, so we are
         * a bit conservative maybe but not wasteful.
         * Notice also that we'll overcommit the b/w if necessary.
         */
                
        /* See how many clocks this will take
         */
        EhciComputeBandwidth(Urb->Flags,
                             Urb->MaximumPacketSize,
                             &MinStartBw,
                             &MaxStartBw,
                             &MinEndBw,
                             &MaxEndBw);

        /* Make sure we have enough bandwidth
         */
        if (0 == (Urb->Flags & URB_FLAG_HIGH_SPEED)) {
            Node = UskMakeMultiReservation(this->Schedule,
                                           1,
                                           MinStartBw,
                                           MaxStartBw,
                                           MinEndBw,
                                           MaxEndBw);
        } else {
            Node = UskMakeReservation(this->Schedule,
                                      1,
                                      MinStartBw,
                                      MaxStartBw);
        }

        if (Node == USK_INVALID_NODE) {
            sc = E_NOT_RESERVED;
            goto ErrorHandler;
        }

        /* The Level2 index is the micro-frame slot
         */
        Value32 = EhciReadRegister(&Ed->Ed.EndpointCaps2);
        Value32 |= EHCI_QH_uFrame_Imask & (1 << UskL2(Node));

        /* For lower speed nodes set the completion mask as well
         */
        if (0 == (Urb->Flags & URB_FLAG_HIGH_SPEED)) {
            Value32 |= EHCI_QH_uFrame_Cmask 
                & (7 << (UskL2(Node) + 2 + EHCI_QH_uFC_SHIFT));
        }

        EhciWriteRegister(&Ed->Ed.EndpointCaps2,Value32);

        /* Insert the ITDs into the schedule
         */
        Ed->InterruptHead = Node;
        if (Urb->Flags & URB_FLAG_HIGH_SPEED)
            EhciInsertIsochronousUsb2Ed(this,Ed,
                                        (Urb->Flags & URB_FLAG_IN) ? TRUE:FALSE);
        else
            EhciInsertIsochronousUsb1Ed(this,Ed,
                                        (Urb->Flags & URB_FLAG_IN) ? TRUE : FALSE);

    }

    /* Where in the schedule ? */
    if (Urb->Interval == 255)
        Urb->Interval = (UINT8) ((CurrentFrameNumber(this) / 8) + 1);

    /* To be adjusted at the end of the transfer */
    Urb->ActualDataLength = Urb->ExpectedDataLength;

    /* Initialize the URB's HCI state, link it to the active list
     * NB: Itd being NULL marks it as non-fully posted yet.
     */
    Hs = (PHCI_STATE)(&Urb->HciReserved);
    Hs->u.Itd = NULL;
    Hs->Next = this->ActiveIsoUrbs;
    this->ActiveIsoUrbs = Urb;

    /* Fill up the xfer descriptors, according to the type
     */
    if (Urb->Flags & URB_FLAG_HIGH_SPEED) {
        Hs->u.Itd = EhciItdIsoTransfer(this,Urb,Ed);
#if _DEBUG
        /* Debug: dump it
         */
        DBGME(0,_tprintf(_T("We I-said this to ED %x[%x]: "),
                         (ADDRESS)Ed, EhciReadRegister(&Ed->Ed.EndpointCaps)));
        DBGME(0,DumpItd( & GetItd(Hs->u.Itd)->Td,0));
#endif
    } else {
        Hs->u.Sitd = EhciSitdIsoTransfer(this,Urb,Ed);
#if _DEBUG
        /* Debug: dump it
         */
        DBGME(0,_tprintf(_T("We SI-said this to ED %x[%x]: "),
                         (ADDRESS)Ed, EhciReadRegister(&Ed->Ed.EndpointCaps)));
        DBGME(0,DumpSitd( & GetSitd(Hs->u.Sitd)->Td,0));
#endif
    }

    /* Signal the HC to rescan the queue
     */
    Value32 = EhciReadRegister(&this->Ops->UsbCmd);
    Value32 |= EHCI_UC_PERIODIC_ENABLE;
    EhciWriteRegister(&this->Ops->UsbCmd,Value32);

    /* Done ok
     */
    return S_OK;

 ErrorHandler:
    if (Ed && Ed->Itd) {
        ItdFree(this,Ed->Itd);
        Ed->Itd = NULL;
    }
    return sc;
}

/* Submit a non-isochronous request.
 * Control/Bulk/Interrupt all use general TDs.
 * They only differ in where the ED goes.
 */
PRIVATE SCODE
EhciSubmitNonIso(PEHCI this,
                 PUSBM_REQUEST_BLOCK Urb,
                 UINT8 PortNumber)
{
    PMYEHCI_QTD HeadTd = NULL, Td = NULL, StatusTd = NULL;
    PMYEHCI_QH Ed = NULL;
    UINT32 Token, Value32, Toggle, QueueType;
    PHCI_STATE Hs;
    BOOL EdIsNew = FALSE;
    BOOL IsSetup = (Urb->Pid == USB_PID_SETUP);
    SCODE sc = E_FAIL;

    /* Quick reminder of where the ED will go
     */
#define ASYNC_Q   1
#define PERIODIC_Q 3
    if (Urb->Flags & URB_FLAG_INTERRUPT)
        QueueType = PERIODIC_Q;
    else
        QueueType = ASYNC_Q;

    /* Allocate a transfer descriptor for this xfer
     */
    HeadTd = TdAlloc(this);
    if (HeadTd == NULL) {
        /* we already groaned */
        return E_NOT_ENOUGH_MEMORY;
    }

    /* Find the ED where this TD should be posted
     * This might allocate a new ED if we aint seen it before
     */
    Ed = LookupEd(this,Urb);
    if (Ed == NULL) {
        DBGME(3,_tprintf(_T("Ehci@%x: ?Ed for %x.%x.%x (%x.%x)\n"),
                         (ADDRESS) this,
                         Urb->DeviceAddress,
                         Urb->EndpointAddress,
                         Urb->Flags,
                         Urb->HubDeviceAddress,
                         Urb->HubPortNumber));
        goto ErrorHandler;
    }

    /* Check if the Ed was ever used before.
     */
    if (Ed->Ed.EndpointCaps == 0) {

        /* Nope, initialize it
         */
        memset(Ed,0,sizeof *Ed);

        /* caps, both */
        Value32 =
            (1 << EHCI_QH_RL_SHIFT)
            | (Urb->MaximumPacketSize << EHCI_QH_MPS_SHIFT)
            | (IsSetup ? EHCI_QH_DTC : 0)
            | (Urb->DeviceAddress & EHCI_QH_DEVICE_MASK)
            | ((Urb->EndpointAddress<<EHCI_QH_ENDPT_SHIFT) & EHCI_QH_ENDPT_MASK)
            ;
        if (Urb->Flags & URB_FLAG_HIGH_SPEED) {
            /* caps */
            Value32 |= EHCI_QH_480Mbps;
            EhciWriteRegister(&Ed->Ed.EndpointCaps,Value32);

            /* caps2 */
            Value32 = EHCI_QH_MULT_1xuf;
        } else {
            /* caps */
            if (Urb->Flags & URB_FLAG_LOW_SPEED)
                Value32 |= EHCI_QH_1_5Mbps;
            /* else 12Mbps which is encoded as 0 */
            if (IsSetup)
                Value32 |= EHCI_QH_C;
            EhciWriteRegister(&Ed->Ed.EndpointCaps,Value32);

            /* caps2 */
            Value32 = EHCI_QH_MULT_1xuf
              | ((Urb->HubDeviceAddress << EHCI_QH_HUB_SHIFT) & EHCI_QH_HUB_MASK)
              | ((Urb->HubPortNumber << EHCI_QH_PORT_SHIFT) & EHCI_QH_PORT_MASK);
        }
        EhciWriteRegister(&Ed->Ed.EndpointCaps2,Value32);

        /* terminate links */
        EhciWriteRegister(&Ed->Ed.TransferOverlay.Next,
                          EHCI_POINTER_TYPE_TERMINATOR);
        EhciWriteRegister(&Ed->Ed.TransferOverlay.Alternate,
                          EHCI_POINTER_TYPE_TERMINATOR);

        /* First time we use this Ed, link it into the proper Q
         */
        assert(Ed != this->AsyncList);

        EdIsNew = TRUE;

        if (QueueType != PERIODIC_Q) {

            Ed->Ed.Next = this->AsyncList->Ed.Next;
            Value32 = VirtualToPhysical(Ed) | EHCI_POINTER_TYPE_QH;
            EhciWriteRegister(&this->AsyncList->Ed.Next,Value32);

        } else {
            /* For the interrupt case, the first time through make a
             * reservation for the necessary amount of b/w.
             * This reservation will only be cancelled when the device
             * is unplugged.  Notice that this happens only when the
             * endpoint is actually used for the first time, so we are
             * a bit conservative maybe but not wasteful.
             * Notice also that we'll overcommit the b/w if necessary.
             */
            UINT16 MaxStartBw, MinStartBw;
            UINT16 MaxEndBw, MinEndBw;
            UINT16 FrequencyInMsecs;
            PUSKNODE Node;
                
            /* See how many clocks this will take
             */
            EhciComputeBandwidth(Urb->Flags,
                                 Urb->MaximumPacketSize,
                                 &MinStartBw,
                                 &MaxStartBw,
                                 &MinEndBw,
                                 &MaxEndBw);

            /* Check the desired polling frequency
             * "0xff" means as slow as possible, which is 16msecs for us.
             */
            FrequencyInMsecs = Urb->Interval;
            if (FrequencyInMsecs >= 255)
                FrequencyInMsecs = MAX_POLLING_FREQUENCY;

            /* Find a node in the schedule with enough bandwidth
             */
            if (0 == (Urb->Flags & URB_FLAG_HIGH_SPEED)) {
                Node = UskMakeMultiReservation(this->Schedule,
                                               FrequencyInMsecs,
                                               MinStartBw,
                                               MaxStartBw,
                                               MinEndBw,
                                               MaxEndBw);
            } else {
                Node = UskMakeReservation(this->Schedule,
                                          FrequencyInMsecs,
                                          MinStartBw,
                                          MaxStartBw);
            }

            if (Node == USK_INVALID_NODE) {
                sc = E_NOT_RESERVED;
                goto ErrorHandler;
            }

            /* The Level2 index is the micro-frame slot
             */
            Value32 = EhciReadRegister(&Ed->Ed.EndpointCaps2);
            Value32 |= EHCI_QH_uFrame_Imask & (1 << UskL2(Node));

            /* For lower speed nodes set the completion mask as well
             */
            if (0 == (Urb->Flags & URB_FLAG_HIGH_SPEED)) {
                Value32 |= EHCI_QH_uFrame_Cmask 
                    & (7 << (UskL2(Node) + 2 + EHCI_QH_uFC_SHIFT));
            }

            EhciWriteRegister(&Ed->Ed.EndpointCaps2,Value32);

            /* Insert the node into the schedule
             */
            Ed->InterruptHead = Node;
            EhciInsertInterruptEd(this,Ed);
        }
    }

    /* Remember the QH so that we can unblock it on ernoderor.
     */
    HeadTd->Ed = Ed;
    HeadTd->Td.Next      = _bs32(EHCI_POINTER_TYPE_TERMINATOR);
    HeadTd->Td.Alternate = _bs32(EHCI_POINTER_TYPE_TERMINATOR);

    /* Set the Token info for the first TD
     */
    Toggle = 0;
    Token  = EHCI_QTD_ACTIVE | EHCI_QTD_CERR;

    /* If there is setup data involved complete a first descriptor
     */
    if (IsSetup) {
        EhciWriteRegister(&HeadTd->Td.Token,
                          Token
                          |Toggle
                          |((Urb->SetupLength<<EHCI_QTD_LENSHIFT)&EHCI_QTD_LEN)
                          |EHCI_QTD_PID_SETUP);

        Value32 = VirtualToPhysical(Urb->SetupData);
        EhciWriteRegister(&HeadTd->Td.Page[0],Value32);

        /* Adjust direction for the next one
         */
        Token |= (Urb->SetupData[0] & USB_RT_DTH) ?
                     EHCI_QTD_PID_IN : EHCI_QTD_PID_OUT;
        Toggle = EHCI_QTD_DT;

        /* Allocate another descriptor for the status final xfer,
         * or for the data portion, whichever comes first.
         * Link it onto the head of the xfer list.
         */
        Td = TdAlloc(this);
        if (Td == NULL) {
            /* we already groaned */
            goto ErrorHandler;
        }

        Td->Ed = Ed;
        Td->Td.Next      = _bs32(EHCI_POINTER_TYPE_TERMINATOR);
        Td->Td.Alternate = _bs32(EHCI_POINTER_TYPE_TERMINATOR);

        EhciWriteRegister(&HeadTd->Td.Next,VirtualToPhysical(&Td->Td));

    } else {
        if (Urb->Pid == USB_PID_IN)
            Token |= EHCI_QTD_PID_IN;
        else
            Token |= EHCI_QTD_PID_OUT; /* a zero */

        /* Otherwise there is no setup phase and the toggle?????
         */
        Td = HeadTd;
    }

    /* Data portion?
     */
    if (Urb->ExpectedDataLength) {

#if CORRECTO
#else
        memset(Urb->Data,0,Urb->ExpectedDataLength);
#endif

        EhciWriteRegister(&Td->Td.Token,
                          Token
                          | (IsSetup ? 0 : EHCI_QTD_IOC)
                          |Toggle
                          |((Urb->ExpectedDataLength<<EHCI_QTD_LENSHIFT)&EHCI_QTD_LEN)
                          );

        /* Fill in all 5 pointers, regardless */
        Value32 = VirtualToPhysical(Urb->Data);
        EhciWriteRegister(&Td->Td.Page[0],Value32);
        Value32 = VirtualToPhysical( (((ADDRESS)Urb->Data)+(4096*1)) & ~4095 );
        EhciWriteRegister(&Td->Td.Page[1],Value32);
        Value32 = VirtualToPhysical( (((ADDRESS)Urb->Data)+(4096*2)) & ~4095 );
        EhciWriteRegister(&Td->Td.Page[2],Value32);
        Value32 = VirtualToPhysical( (((ADDRESS)Urb->Data)+(4096*3)) & ~4095 );
        EhciWriteRegister(&Td->Td.Page[3],Value32);
        Value32 = VirtualToPhysical( (((ADDRESS)Urb->Data)+(4096*4)) & ~4095 );
        EhciWriteRegister(&Td->Td.Page[4],Value32);

        /* For bulk/interrupt xfers no status
         */
        if (!IsSetup)
            goto NoStatus;

        /* Get a Td for status and link it
         */
        StatusTd = TdAlloc(this);
        if (StatusTd == NULL) {
            goto ErrorHandler;
        }

        StatusTd->Ed = Ed;
        EhciWriteRegister(&Td->Td.Next, VirtualToPhysical(StatusTd));
        /* other pointers are next, down below */

        /* flip direction for the status phase
         */
        Token ^= EHCI_QTD_PID_IN;/*knows too much*/

    } else {
        assert (IsSetup);
        /* Its a setup with no data phase, that means Token currently says OUT
         */
        assert((Token & EHCI_QTD_PID) == EHCI_QTD_PID_OUT);
        Token |= EHCI_QTD_PID_IN;
        StatusTd = Td;
    }

    /* Make every possible alternate point to status
     */
    Value32 = VirtualToPhysical(StatusTd);
    EhciWriteRegister(&HeadTd->Td.Alternate, Value32);
    EhciWriteRegister(&Td->Td.Alternate, Value32);

    /* Fill in the status phase Td
     */
    EhciWriteRegister(&StatusTd->Td.Token,
                      Token
                      |EHCI_QTD_IOC
                      |EHCI_QTD_DT);
    StatusTd->Td.Next      = _bs32(EHCI_POINTER_TYPE_TERMINATOR);
    StatusTd->Td.Alternate = _bs32(EHCI_POINTER_TYPE_TERMINATOR);
    StatusTd->Td.Page[0] = 0;
    
  NoStatus:

    /* To be adjusted at the end of the transfer */
    Urb->ActualDataLength = Urb->ExpectedDataLength;

    /* Initialize the URB's HCI state, link it to the active list
     */
    Hs = (PHCI_STATE)(&Urb->HciReserved);
    Hs->u.Td = HeadTd;
    Hs->Next = this->ActiveUrbs;
    this->ActiveUrbs = Urb;

    /* Now queue the TD up onto the ED
     */
    EdEnqueTd(&Ed->Ed,&HeadTd->Td);

#if _DEBUG
    /* Debug: dump it
     */
    DBGME(0,_tprintf(_T("We said this to ED %x[%x]: "),
                     (ADDRESS)Ed, EhciReadRegister(&Ed->Ed.EndpointCaps)));
    DBGME(0,DumpTdList(&HeadTd->Td,0));
#endif

    /* Signal the HC to rescan the queue
     */
    Value32 = EhciReadRegister(&this->Ops->UsbCmd);
    if (QueueType == ASYNC_Q) {

        Value32 |= EHCI_UC_ASYNCH_ENABLE;

    } else if (QueueType == PERIODIC_Q) {

        Value32 |= EHCI_UC_PERIODIC_ENABLE;
    }
    EhciWriteRegister(&this->Ops->UsbCmd,Value32);

    /* Done ok
     */
    return S_OK;

 ErrorHandler:
    if (HeadTd)
        TdFree(this,HeadTd,FALSE);
    return sc;
}

/* Destroy all endpoints for a given USB address.
 * This is invoked on device removal, AFTER all pending URBs for 
 * these andpoints have been cancelled.
 */
PRIVATE void EhciCancelUrbsForFunction(
    PEHCI this, 
    UINT8 UsbAddress);

PRIVATE void
DestroyEdsForFunction(PEHCI this, 
                      UINT8 UsbAddress)
{
    PMYEHCI_QH Ed;
    int i;
    UINT32 Value32;
    BOOL IsInput;
    PEHCI_HT HashTable;

    DBGME(0,_tprintf(_T("Ehci(%x)DestroyEds(%x)\n"), (ADDRESS)this, UsbAddress));

    /* We do *not* free the config endpoint
     */
    if (UsbAddress == 0)
        return;

    /* Cancel out any active URBs for this USB address
     */
    EhciCancelUrbsForFunction(this,UsbAddress);

    /* Lock the hash table
     */
    Mutex_Lock(&this->HashTableLock);

    /* Sanity check the table is there (assert?)
     */
    HashTable = this->HashTable;
    if (HashTable != NULL) {

        /* Scan the whole table for EDs with this function address
         */
        for (i = 0; i < HashTable->nEntries; i++) {

            Value32 = HashTable->Entries[i].HashValue;
            if (UsbAddress == (Value32 & EHCI_QH_DEVICE_MASK)) {

                /* Got one, free slot
                 */
                DBGME(0,_tprintf(_T("Ehci(%x) Hash[%d] %x %x\n"), (ADDRESS)this,
                                 i, HashTable->Entries[i].HashValue,
                                 (ADDRESS) HashTable->Entries[i].DataValue));
                IsInput = (HashTable->Entries[i].HashValue & EHCI_QH_DTC) != 0;
                Ed = HashTable->Entries[i].DataValue;
                HashTable->Entries[i].DataValue = NULL;
                HashTable->Entries[i].HashValue = 0;

                /* We'll find it on the async list, unless its an interrupt
                 * It is an interrupt IFF the uframe schedule is set.
                 */
                Value32 = EhciReadRegister(&Ed->Ed.EndpointCaps2);
                if (Value32 & EHCI_QH_uFrame_Imask) {
                    EhciDropEdFromSchedule(this,Ed,IsInput);
                } else {
                    EhciDropEdFromAsync(this,Ed);
                }

                /* Now we would love to rid of it here but... the HC might
                 * still use it, or have it cashed. We must follow specs
                 * and ring the doorbell, wait for the interrupt and then
                 * we can be sure its gone.
                 * All we can do now is to set it aside for later.
                 * We'll do it for interrupt endpoints too, for simplicity.
                 */
                Ed->TrashCan = this->TrashCan;
                this->TrashCan = Ed;
            }
        }

        /* Ring the doorbell
         */
        Value32 = EhciReadRegister(&this->Ops->UsbCmd);
        Value32 |= EHCI_UC_ASYNCH_INT_DOORBELL;
        EhciWriteRegister(&this->Ops->UsbCmd,Value32);
        EhciEnableDoorbellInterrupts(this->Ops);
    }

    Mutex_Unlock(&this->HashTableLock);
}

PRIVATE UINT16
EhciMapPs( UINT16 Bits )
{
    UINT16 ret = 0;
    if (Bits & EHCI_PS_COMPANION_HC_OWNED)
        return ret;
    if (Bits & EHCI_PS_PP)
        ret |= USB_ST_PORT_POWER;
    if (Bits & EHCI_PS_PORT_RESET)
        ret |= USB_ST_PORT_RESET;
    if (Bits & EHCI_PS_PORT_SUSPEND)
        ret |= USB_ST_PORT_SUSPEND;
    if (Bits & (EHCI_PS_OVERCURRENT_CHANGE|EHCI_PS_OVERCURRENT))
        ret |= USB_ST_PORT_OVERCURRENT;
    if (Bits & (EHCI_PS_ENABLED_CHANGE|EHCI_PS_ENABLED))
        ret |= USB_ST_PORT_ENABLE;
    if (Bits & (EHCI_PS_CONNECTED_CHANGE|EHCI_PS_CONNECTED))
        ret |= USB_ST_PORT_CONNECTION;
    return ret;
}

/* A device was plugged in to a root port, reset and test the speed
 */
PRIVATE void
EhciPortReset(PEHCI this, UINT8 PortNo)
{
    int i;
    UINT32 Value32;
    PEHCI_OPS Ops = this->Ops;

    DBGME(0,_tprintf(_T("EhciConnection(%x)\n"), PortNo));

    /* A debounce time of >=100ms has already passed.
     * Check for low speed (D+ clear)
     */
    Value32 = EhciReadRegister(&Ops->PortSc[PortNo]);
    DBGME(0,_tprintf(_T("Port%d inserted(%x)\n"), PortNo, Value32));

    /* Give it to companions if slow
     */
    if ((Value32 & EHCI_PS_LINE_STATUS) == EHCI_PS_LINE_K_STATE) {
  ForgetIt:
        DBGME(0,_tprintf(_T("Port%d slow(%x)\n"), PortNo, Value32));
        Value32 |= EHCI_PS_COMPANION_HC_OWNED;
        EhciWriteRegister(&Ops->PortSc[PortNo],Value32);
        /*NB: will get another PORT_CHANGE interrupt, for disconnect */
        this->PortSw[PortNo] = 0;
        return;
    }

    /* Reset it
     */
    Value32 &= ~EHCI_PS_ENABLED;
    Value32 |= EHCI_PS_PORT_RESET;
    EhciWriteRegister(&Ops->PortSc[PortNo],Value32);

    /* Wait long enough
     */
    MilliSleep(50);

    /* Clear the reset
     */
    Value32 &= ~EHCI_PS_PORT_RESET;
    EhciWriteRegister(&Ops->PortSc[PortNo],Value32);

    /* Recovery time: 10ms
     */
    MilliSleep(10);

    /* Loop reading it until it clears...
     * NB: Should have taken no more than 2ms anyways.
     */
    for (i = 0; i < 100; i++) {
        Value32 = EhciReadRegister(&Ops->PortSc[PortNo]);
        if ((Value32 & EHCI_PS_PORT_RESET) == 0)
            break;
    }

    /* Speed check */
    DBGME(0,_tprintf(_T("Port%d check(%x)\n"), PortNo, Value32));
    if ((Value32 & EHCI_PS_ENABLED) == 0)
        goto ForgetIt;

    /* Yes, this is for us.
     */
    this->PortSw[PortNo] = USB_ST_PORT_RESET | USB_ST_PORT_HIGH_SPEED;
}

PRIVATE SCODE
EhciRootHub(PEHCI this,
  /*inout*/ PUSBM_REQUEST_BLOCK Urb)
{
    UINT16 Status;
    UINT16 Changes;
    UINT8 Port;
    BOOL Clear;
    UINT32 Value32, Bits;

    /* Is it an EVENT one
     */
    if (Urb->Pid == USB_PID_EVENT) {

        DBGME(0,_tprintf(_T("Ehci:Event %x\n"),Urb->Data[0]));
        if (Urb->Data[0] == USBM_EVENT_DISCONNECTED) {
            DestroyEdsForFunction(this,Urb->Data[1]);
        }
        goto DoneOk;
    }

    /* GetPortStatusChanges() ?
     */
    if ((Urb->Pid == USB_PID_IN) &&
        (Urb->Flags & URB_FLAG_INTERRUPT)) {
    
        this->EventsPacket = Urb;
        Changes = this->PortStatusChanges;
        Urb->ActualDataLength = 2;
        Set2(Urb->Data,Changes);
        if (Changes) {
            this->EventsPacket = NULL;
            Urb->Flags |= URB_FLAG_COMPLETE;
            Condition_InterruptSignal(Urb->Done);
        }
        return S_OK;
    }

    /* GetPortStatus() ?
     */
    if ((Urb->Pid == USB_PID_SETUP) &&
        (Urb->Flags & URB_FLAG_IN) &&
        (Urb->SetupData[1] == USB_R_GET_STATUS)) {

        /* Our ports start at 0
         */
        Port = Urb->SetupData[4] - 1;
        if (Port >= this->nPorts) {
            DBGME(3,_tprintf(_T("EHCI(%x) ?port%d?\n"), (ADDRESS)this, Port));
            return E_INVALID_PARAMETER;
        }

        /* Map our bit encodings to USB2 
         */
        Value32 = EhciReadRegister(&this->Ops->PortSc[Port]);
        Changes = EhciMapPs( (UINT16)(Value32 & EHCI_PC));
        Status  = EhciMapPs( (UINT16)(Value32 & EHCI_PS));

        /* sw emulation: LOW/HIGH_SPEED and sticky RESET */
        Changes |= this->PortSw[Port] & USB_ST_PORT_RESET;
        Status  |= this->PortSw[Port] & ~USB_ST_PORT_RESET;

        Urb->ActualDataLength = 4;
        Set2(Urb->Data,Status);
        Set2(Urb->Data+2,Changes);
        Condition_InterruptSignal(Urb->Done);
        goto DoneOk;
    }

    /* SetClearPortFeature() ?
     */
    if ((Urb->Pid == USB_PID_SETUP) &&
        (Urb->Flags & URB_FLAG_CONTROL)) {

        /* Our ports start at 0
         */
        Port = Urb->SetupData[4] - 1;
        if (Port >= this->nPorts) {
            DBGME(3,_tprintf(_T("EHCI(%x) ??port%d?\n"), (ADDRESS)this, Port));
            return E_INVALID_PARAMETER;
        }

        /* Which feature?
         */
        Changes = Get2(Urb->SetupData+2);
        Clear = (Urb->SetupData[1] == USB_R_CLEAR_FEATURE);
        switch (Changes) {
        case USB_FS_C_PORT_RESET:
            if (Clear) {
                this->PortSw[Port] &= ~USB_ST_PORT_RESET;
            }
            break;
        case USB_FS_PORT_RESET:
            if (!Clear) {
                EhciPortReset(this,Port);
            }
            break;
        case USB_FS_PORT_ENABLE:
            if (Clear) {
                /* Clear the enable bit
                 */
                Bits = EHCI_PS_ENABLED;
            /*ClearIt:*/
                Value32 = EhciReadRegister(&this->Ops->PortSc[Port]);
                Value32 &= ~EHCI_PC;
                Value32 &= ~Bits;
                EhciWriteRegister(&this->Ops->PortSc[Port],Value32);
                goto CheckForChanges;
            }
            break;
        case USB_FS_C_PORT_CONNECTION:
            if (Clear) {
                /* Clear the change bit
                 */
                Bits = EHCI_PS_CONNECTED_CHANGE;
            /*SetIt:*/
                Value32 = EhciReadRegister(&this->Ops->PortSc[Port]);
                Value32 &= ~EHCI_PC;
                Value32 |= Bits;
                EhciWriteRegister(&this->Ops->PortSc[Port],Value32);

                /* If there is nothing else..
                 */
            CheckForChanges:
                Value32 = EhciReadRegister(&this->Ops->PortSc[Port]);
                Value32 &= EHCI_PC;
                if (Value32 == 0)
                    this->PortStatusChanges &= ~(2 << Port);/*p0->p1*/
            }
            break;
        }

        Condition_InterruptSignal(Urb->Done);
        goto DoneOk;
    }

    /* Uhu?
     */
    return E_NOT_IMPLEMENTED;

 DoneOk:
    Urb->Flags |= URB_FLAG_COMPLETE;
    return S_OK;
}

PRIVATE SCODE EhciMasterSubmitUrb (
    PIUSBMASTERCONTROLLER pThis,
    /*inout*/ PUSBM_REQUEST_BLOCK Urb,
    UINT8 PortNumber)
{
    PEHCI this = pEhciMaster(pThis);

    DBGME(0,_tprintf(_T("Ehci:Submit(%x,%x,%x)\n"),
                     (ADDRESS)this, (ADDRESS)Urb, PortNumber));

    Urb->ActualDataLength = 0;
    DBGME(0,DumpPacket(Urb));
    
    /* Is this for the root hub?
     */
    if (Urb->DeviceAddress == 0xff)
        return EhciRootHub(this,Urb);

    /* Post a TD
     * Isochronous endpoints require different TDs.
     */
    if (Urb->Flags & URB_FLAG_ISOCHRONOUS)
        return EhciSubmitIso(this,Urb,PortNumber);
    else
        return EhciSubmitNonIso(this,Urb,PortNumber);

}

/* Inactivate a (list of) TD
 */
PRIVATE void EhciCancelTd(
    PEHCI this,
    PMYEHCI_QTD HeadTd)
{
    PMYEHCI_QH Ed;
    PMYEHCI_QTD Td;
    UINT32 Value32;

    /* Has the TD been completed already?
     * NB: The ISR zeroes an URB's TD field to avoid racing us here.
     */
    if (HeadTd == NULL)
        return;

    /* Walk the list now
     */
    Ed = HeadTd->Ed;
    for (Td = HeadTd;;) {

        /* The ISR cannot get to it, but the HC still can.
         * Mark the TD as non-active, jic it hasnt been overlayed.
         */
        Td->Td.Token = 0;

        /* Deactivate the overlay in case this TD is the current one.
         */
        Value32 = EhciReadRegister(&Ed->Ed.Current) & EHCI_POINTER_MASK;
        if (PhysicalToVirtual(Value32) != &Td->Td)
            /* Otherwise we are done with this TD */
            goto NextTd;

        Ed->Ed.TransferOverlay.Token = 0;

        /* That only works if the HC is currently doing something else.
         * Otherwise it can flush out the overlay from the on-chip area
         * and undo our write.  Spin for a little bit to make sure, redo it.
         */
        Delay(100);
        Ed->Ed.TransferOverlay.Token = 0;

        /* End of list?
         */
    NextTd:
        Value32 = EhciReadRegister(&Td->Td.Next) & EHCI_POINTER_MASK;
        if (Value32 == 0)
            break;

        Td = PhysicalToVirtual(Value32);
    }

    /* BUGBUG Only valid for 1st xfer!!!
     */
    TdFree(this,Td,FALSE);
    DBGME(3,_tprintf(_T("Ehci(%x)CancelTd %x ok.\n"),
                     (ADDRESS)this, (ADDRESS)HeadTd));
}

PRIVATE SCODE EhciMasterCancelUrb (
    PIUSBMASTERCONTROLLER pThis,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 PortNumber)
{
    PEHCI this = pEhciMaster(pThis);

    DBGME(0,_tprintf(_T("EhciCancel(%x,%x,%d)\n"),
                     (ADDRESS)this,(ADDRESS)Urb,PortNumber));

    if (Urb == this->EventsPacket)
        this->EventsPacket = NULL;
    else {
        PUSBM_REQUEST_BLOCK *pNextUrb;
        PHCI_STATE Hs;

        Hs = (PHCI_STATE)(&Urb->HciReserved);

        pNextUrb = &this->ActiveUrbs;
        while (pNextUrb) {
            if  (*pNextUrb == Urb) {

                /* Unlink it and the ISR wont find it.
                 */
                *pNextUrb = Hs->Next;

                /* Now cancel the TD for this URB
                 */
                EhciCancelTd(this,Hs->u.Td);
                goto GotIt;
            }
            pNextUrb = (PUSBM_REQUEST_BLOCK *) *pNextUrb;
        }

        pNextUrb = &this->ActiveIsoUrbs;
        while (pNextUrb) {
            if  (*pNextUrb == Urb) {
                *pNextUrb = Hs->Next;
                /* BUGBUG Abort active transactions in the schedule ...*/
                goto GotIt;
            }
            pNextUrb = (PUSBM_REQUEST_BLOCK *) *pNextUrb;
        }

        return E_OBJECT_NOT_FOUND;
    }

 GotIt:
    Urb->Flags |= URB_FLAG_ERROR;
    Condition_InterruptSignal(Urb->Done);
    return S_OK;
}

PRIVATE void EhciCancelUrbsForFunction(
    PEHCI this, 
    UINT8 UsbAddress)
{
    PUSBM_REQUEST_BLOCK *pNextUrb, Urb;
    PIUSBMASTERCONTROLLER pThis;

    DBGME(0,_tprintf(_T("Ehci(%x)CancelUrbsForFunction(%d)\n"),
                     (ADDRESS)this,UsbAddress));

    pThis = iEhciMaster(this);
 RescanNonIso:
    pNextUrb = &this->ActiveUrbs;
    while (pNextUrb) {
        Urb = *pNextUrb;

        if (Urb->DeviceAddress == UsbAddress) {
            EhciMasterCancelUrb(pThis,
                                Urb,
                                0xff /* BUGBUG useless? */);
            goto RescanNonIso;
        }

        pNextUrb = (PUSBM_REQUEST_BLOCK *) *pNextUrb;
    }

 RescanIso:
    pNextUrb = &this->ActiveIsoUrbs;
    while (pNextUrb) {
        Urb = *pNextUrb;

        if (Urb->DeviceAddress == UsbAddress) {
            EhciMasterCancelUrb(pThis,
                                Urb,
                                0xff /* BUGBUG useless? */);
            goto RescanIso;
        }

        pNextUrb = (PUSBM_REQUEST_BLOCK *) *pNextUrb;
    }
}

    

#include "s_Ehci_v.c"

/* Constructor */
PIDEVICE EhciNew( ADDRESS PciLocation )
{
    PIHEAP pHeap = CurrentHeap();
    UINT Bus, Device, Function;
    PIBUSDRIVER OurPci = NULL;
    SCODE sc;
    UINT32 Value32;
    UINT16 Value16;
    UINT8 Value8;
    PEHCI this = NULL;
    PEHCI_CAPS Caps = NULL;
    PEHCI_OPS Ops;
    PMYEHCI_QH Qhead;
    int i, ListSize;

    Bus = PciLocation >> 16;
    Device = (PciLocation >> 8) & 0xff;
    Function = PciLocation & 0xff;

    DBGME(1,_tprintf(_T("Enter EhciNew(%d,%d,%d)\n"),
                     Bus,Device,Function));

    /* Make sure we know where to go
     */
    sc = BindToObject(CurrentNameSpace(),
                      PCIDRIVER_NAME,
                      NAME_SPACE_READ | NAME_SPACE_WRITE,
                      &IID_IBusDriver,
                      (void**)&OurPci);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("No PCI access routine!!\n")));
        goto ErrorExit;
    }

    /* Look at PCI space, check its there
     */
    Value32 = 0;
    OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                      0, &Value32, sizeof(Value32));
    if (Value32 != PCIID_NEC_UPD720101) {
        /* Check if it is compatible.
         * The specs mandate CLASSC and SBRN.
         */
        OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                          8, &Value32, sizeof(Value32));
        OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                          0x60, &Value8, sizeof(Value8));

        if ((Value8 != EHCI_SBRN) ||
            ((Value32 & 0xfffff00) != EHCI_CLASSC)) {
            DBGME(3,_tprintf(_T("Signature at PCI %x not for EHCI.\n"),
                       PciLocation));
            goto ErrorExit;
        }
    }

    /* The UPD gets:
     * Command 4	Status 210	ClassC c032004	CacheLS/lat/headert/BIST 1010
     * Base[0] 80100000	Base[1] 0	SubID e01033	ROMa 0
     * Cap_ptr/.. 40	Intr 22100300	CapID/next 1	PMC fe02
     * PMCSR 0	PMCSR_BSE/Data 0	SBRN 20	FLADJ 20
     * WAKECAP 3f	EXT1 6cb03305	EXT2 0	LEGSUP 1	LEGCTL c0000000
    */

    /* Look at the assigned I/O location, make sure it was assigned
     */
    Value32 = 0;
    OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                      16, &Value32, sizeof(Value32));
    Caps = (PEHCI_CAPS) IoPhysicalToIoVirtual(Value32);
    if (Caps == NULL) {
        DBGME(3,_tprintf(_T("EHCI at PCI %x has no BaseAddress[0].\n"),
                       PciLocation));
        goto ErrorExit;
    }

    /* Enable it, by writing to the Command register
     * No: back-to-back,SERR#, wait cycle, parity, palette snoop, I/O
     * Yes: MWI, master, memory
     */
    Value16 = 0x0016;
    OurPci->v->WriteConfigurationSpace(OurPci,Bus,Device,Function,
                                      4, &Value16, sizeof(Value16));
    Delay(20);

    /* Check caps make sens...
     */
    Value32 = EhciReadRegister(&Caps->VersionAndLength);

    /* Offset to the Ops registers
     */
    Ops = (PEHCI_OPS) ((ADDRESS)Caps + (Value32 & EHCI_V_CAPLENGTH));

    /* How small can we make the periodic list
     */
    Value32 = EhciReadRegister(&Caps->HccParams);
    if (Value32 & EHCI_HCS_PROG_FRAME_LIST_LENGTH) {
        i = 256;
    } else {
        i = 1024;
    }

    /* Allocate all our resources
     */
    this = (PEHCI) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *this, 0);
    if (!this) {
        DBGME(3,_tprintf(_T("EHCI at PCI %x outtamem.\n"), PciLocation));
        goto ErrorExit;
    }

    this->PeriodicListSize = i;
    ListSize = i * sizeof(this->PeriodicList[0]);
#define bsize (ListSize + sizeof(MYEHCI_QH))

    /* Cache some values
     */
    Value32 = EhciReadRegister(&Caps->HcsParams);
    this->nPorts = (UINT8) (Value32 & EHCI_HCS_N_PORTS);
    this->PortPowerControl = (Value32 & EHCI_HCS_PPC) ? TRUE : FALSE;

    this->PeriodicList = pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY,/*qhead!*/
                                         bsize, EHCI_PLB_ALIGNMENT+1);
#undef bsize
    if (this->PeriodicList == NULL) {
        DBGME(3,_tprintf(_T("EHCI at PCI %x outtamem.\n"), PciLocation));
        goto ErrorExit;
    }

    Qhead = (PMYEHCI_QH)((ADDRESS)this->PeriodicList + ListSize);
    this->AsyncList = Qhead;

    /* Allocate the schedule for interrupt/isochronous endpoints
     */
    this->Schedule = UskCreateSchedule( this->PeriodicList,
                                        this->PeriodicListSize );
    if (this->Schedule == NULL)
        goto ErrorExit;

    /* Got resources, we are committed
     */
    this->CDevice.v = (struct CDeviceVtbl*)&EhciDeviceEhciVtbl;
    this->CDevice.RefCnt = 1;
    this->EhciMasterv = &EhciMasterEhciVtbl;
    this->Caps = Caps;
    this->Ops = Ops;

    /* Initialize the aperiodic list
     * NB: This is the endpoint #0 used during configuration, we only
     *     initialize here only the fields that will not change.
     */
    EhciWriteRegister(&Qhead->Ed.Next,
                      VirtualToPhysical(Qhead) | EHCI_POINTER_TYPE_QH);
    EhciWriteRegister(&Qhead->Ed.EndpointCaps,
                      (1 << EHCI_QH_RL_SHIFT)
                      | (8 << EHCI_QH_MPS_SHIFT)
                      | EHCI_QH_H
                      | EHCI_QH_DTC
                      | EHCI_QH_480Mbps
                      );
    EhciWriteRegister(&Qhead->Ed.TransferOverlay.Next,
                      EHCI_POINTER_TYPE_TERMINATOR);
    EhciWriteRegister(&Qhead->Ed.TransferOverlay.Alternate,
                      EHCI_POINTER_TYPE_TERMINATOR);

    /* Anything else in DeviceInitialize
     */
    return iEhciDevice(this);

 ErrorExit:
    if (Caps) {
        Value16 = 0x0000;
        OurPci->v->WriteConfigurationSpace(OurPci,Bus,Device,Function,
                                           4, &Value16, sizeof(Value16));
    }
    if (OurPci)
        OurPci->v->Release(OurPci);
    if (this) {
        if (this->PeriodicList)
            pHeap->v->Free(pHeap,0,this->PeriodicList);
        pHeap->v->Free(pHeap,0,this);
    }
    return NULL;

}

/* ** 
*/
SCODE __cdecl EhciConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("EhciConstructor(%x %s %x %x %d)\n"),
                     (ADDRESS) pThis, Name, (ADDRESS) ppUnk,
                     RegisterBase, Interrupt));

    pUnk = (PIUNKNOWN) EhciNew( RegisterBase );

    DBGME(0,_tprintf(_T("EhciConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("EhciConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl EhciDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) EhciConstructor
};

static struct _DRIVERCOB EhciDriverCobObject = { &EhciDriverCobVtable, 1 };

PIUNKNOWN EhciCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &EhciDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &EhciDeviceEhciVtbl, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &EhciDriverCobObject;
}
