/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    ohci.c
 *
 * Purpose:
 *
 *    Device driver for the Open Host Controller (OHCI) USB 1.0 master.
 *
 * Reference: "OpenHCI Open Host Controller Interface Specification for USB"
 *            Rev 1.0a, September 1999.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <stddef.h> /* for offsetof() */
#include <tchar.h>
//#define _DEBUG 1
int OhciDebug = 2;
#define DBGLVL OhciDebug
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbmaster.h>
#include <busdriver.h>

/* BUGBUG We sure could use a trivial VM system
 */
#define IoPhysicalToIoVirtual(x) ((PTR)x)
#define PhysicalToVirtual(x)     ((PTR)((ADDRESS)x))
#define VirtualToPhysical(x)     ((ADDRESS)x)

#if (BYTE_ORDER==BIG_ENDIAN)
#define PCIDRIVER_NAME _T("pcibs")
#else
#define PCIDRIVER_NAME _T("pci")
#endif

#include "ohci.h"

/* Define macros for inclusion of interrupt schedule code
 */
#define HC_NODE            OHCI_ED
#define HC_NODE_ALIGNMENT  (OHCI_ED_ALIGNMENT+1)
#define HC_ROOTS           OHCI_NINTED
#define HC_ROOTS_LOG2      5
#define HC_BUS_BANDWIDTH   USB_11_BUS_BANDWIDTH
PRIVATE INLINE void OhciWriteRegister(volatile UINT32 *Address, UINT32 Value);
#define HC_PHYS_LINK(_node_,_phys_ptr_) \
    OhciWriteRegister(&(_node_).Next,_phys_ptr_)
#define HC_INIT_NODE(_node_) \
    OhciWriteRegister(&(_node_).Caps,OHCI_ED_K)

#include "usb_sched.c"

/* Define the structure of our per-URB state,
 * which is an overlay of the two HCI-reserved pointers.
 * This is needed because TD are completed in random order.
 */
typedef struct _HCI_STATE {
    POHCI_ED Ed;     /* Needed to cleanup/suspend at the end of the xfer */
    UINT8    nFrags; /* To know when all TDs are done [4bits?]           */
    UINT8    cc;     /* Condition code "or" of all TDs [4 bits]          */
    UINT16   Pad;    /* Unused for now                                   */
} HCI_STATE, *PHCI_STATE;

/* Add a back-pointer to our TDs so that we can find the URB 
 * when one completes.  
 */
typedef struct _MYOHCI_TD {
    /* goes first, for alignment reasons */
    OHCI_TD             Td;
    PUSBM_REQUEST_BLOCK Urb;
} MYOHCI_TD, *PMYOHCI_TD;

/* Add a back-pointer to our iTDs so that we can find the URB 
 * when one completes.  iTDs are much less popular, it is not
 * worth it wasting four words to unify MYOHCI_TD=MY_OHCI_ITD
 */
typedef struct _MYOHCI_ITD {
    /* goes first, for alignment reasons */
    OHCI_ITD            Td;
    PUSBM_REQUEST_BLOCK Urb;
} MYOHCI_ITD, *PMYOHCI_ITD;

/* Add a back-pointer to (some of) our EDs so that we can find
 * the interrupt list they are on when they go away. Otherwise,
 * we would have to look all over the schedule for it.
 * NB: This struct is and must remain the same as an MYOHCI_TD.
 */
typedef struct _MYOHCI_ED {
    /* goes first, for alignment reasons */
    OHCI_ED             Ed;
    PUSKNODE            InterruptHead;
} MYOHCI_ED, *PMYOHCI_ED;

/* Trivial hash table
 */
typedef struct _OHCI_HT {
#define InitialHashEntries 8  /* 2**n */
    int nEntries;
    struct _OHCI_HT_ENTRY {
        PTR    DataValue;
        UINT32 HashValue;
    } Entries[1]; /* varsized */
} OHCI_HT, *POHCI_HT;

/* Our device object
 */
typedef struct _OHCI {
    struct CDevice CDevice;
    CONST_VTBL struct IUsbMasterControllerVtbl *OhciMasterv;

    POHCI_OPS Ops;
    POHCI_HCCA HCCA;

    PUSKEDULE Schedule; 

    POHCI_ED IsoHead;
    POHCI_ED ControlHead;
    POHCI_ED BulkHead;

    /* Pack our bools */
    UINT8  nPorts;
    UINT8  PortPowerControl;
    UINT16 PortStatusChanges;

    PUSBM_REQUEST_BLOCK EventsPacket;

    POHCI_HT HashTable;
    MUTEX HashTableLock;

    PMYOHCI_TD FreeTds;
    int FreeTdCount;
#define FreeTdMax 8

    PMYOHCI_ITD FreeItds;
    int FreeItdCount;
#define FreeItdMax 4

    struct _STATS {
        int IsrCnt;
        int IntrCnt;
        int RegCnt;
        int ChgCnt;
        int ErrCnt;
    } Stats;

} OHCI, *POHCI;

/* Convert between interface and instance pointer and vice versa */
#define pOhciDevice(_i_) ((POHCI)((_i_)))
#define pOhciMaster(_i_) ((POHCI)((ADDRESS)(_i_) - offsetof(struct _OHCI, OhciMasterv)))
#define iOhciDevice(_p_) ((PIDEVICE)((_p_)))
#define iOhciMaster(_p_) ((PIUSBMASTERCONTROLLER)(&((POHCI)(_p_))->OhciMasterv))

/*
 * Phys functions
 */
PRIVATE INLINE UINT32 OhciReadRegister(volatile UINT32 *Address)
{
    UINT32 Value = *Address;
    return _bs32(Value);
}

PRIVATE INLINE UINT16 OhciReadRegister16(volatile UINT16 *Address)
{
    UINT16 Value = *Address;
    return _bs16(Value);
}

PRIVATE INLINE void OhciWriteRegister(volatile UINT32 *Address, UINT32 Value)
{
    Value = _bs32(Value);
    *Address = Value;
    IOWriteBufferFlush();
}

PRIVATE INLINE void OhciWriteRegister16(volatile UINT16 *Address, UINT16 Value)
{
    Value = _bs16(Value);
    *Address = Value;
    IOWriteBufferFlush();
}

/* General utilities
 */

/* Sleep for a few milliseconds
 */
PRIVATE void MilliSleep(int mSecs)
{
    TIME t;
    Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(mSecs)));
    SleepUntil(t);
}

/* Get the frame number for the currently executing frame
 */
PRIVATE INLINE UINT16 CurrentFrameNumber(POHCI this)
{
    UINT32 Value32;
    Value32 = OhciReadRegister(&this->Ops->FmNumber);
    return (UINT16)Value32;
}

/* Compute the min/max bandwidth for xfers to a given node
 */
PRIVATE void
OhciComputeBandwidth(
    USBM_URB_FLAGS Flags,
    UINT16 MaximumPacketLength,
    UINT16 *MaxBw,
    UINT16 *MinBw
    )
{
    UINT16 Overhead;
    UINT16 Bandwidth;

    /*
     * For an interrupt transfer, the minimum b/w is when a NAK is returned.
     * Call that the per-packet overhead.
     */
    Overhead = (Flags & URB_FLAG_INTERRUPT) ? 13 : 9;

    /*
     * That is in bytes.  Add bitstuffing and turn it into full-speed clocks.
     */
    Bandwidth = ((MaximumPacketLength + Overhead) * 8 * 7) / 6;
    Overhead  = ((                  0 + Overhead) * 8 * 7) / 6;

    /*
     * A slow-speed node runs at 1/8th full speed.
     */
    if (Flags & URB_FLAG_LOW_SPEED) {
        Bandwidth *= 8;
        Overhead  *= 8;
    }

    /* Return the computed values
     */
    *MaxBw = Bandwidth;
    *MinBw = Overhead;
}

/*
 * Utilities for handling MYTDs
 */
PRIVATE PMYOHCI_TD
TdAlloc(POHCI this)
{
    PMYOHCI_TD Td, Next;
    PIHEAP pHeap;

    /* Try to grab one from the free list
     */
    do {
        Td = this->FreeTds;
        if (Td == NULL)
            break;
        Next = (PMYOHCI_TD) Td->Urb;
    } while (!AtomicCmpAndSwap((ADDRESS*) &this->FreeTds,
                               (ADDRESS) Td,
                               (ADDRESS) Next));

    /* If empty allocate a new one
     */
    if (Td != NULL) {
        this->FreeTdCount--; /* no prob if slightly off */
    } else {
        pHeap = CurrentHeap();

        /* NB: NOT zeroed, like it comes from the freelist */
        Td = (PMYOHCI_TD) pHeap->v->Alloc(pHeap, 0,
                                          sizeof *Td,
                                          OHCI_ED_ALIGNMENT+1);
        if (!Td) {
            DBGME(3,_tprintf(_T("OHCI(%x) Td outtamem.\n"), (ADDRESS) this));
            return NULL;
        }
    }

    Td->Urb = NULL; /* sanity */
    return Td;
}

/* Free a list of TDs.
 * This is callable from an Isr,
 * in which case it wont go to the heap
 */
PRIVATE void
TdFree(POHCI this, PMYOHCI_TD Td, BOOL FromIsr)
{
    PMYOHCI_TD Next;
    UINT32 Value32;
    PIHEAP pHeap = CurrentHeap();

    do {
        Value32 = OhciReadRegister(&Td->Td.Next);

        /* Put it on free list?
         */
        if (FromIsr || (this->FreeTdCount < FreeTdMax)) {
            /* Normally-won race to the top
             */
            do {
                Next = this->FreeTds;
                Td->Urb = (void *) Next;
            } while (!AtomicCmpAndSwap((ADDRESS*) &this->FreeTds,
                                       (ADDRESS) Next,
                                       (ADDRESS) Td));

            this->FreeTdCount++; /* no prob if slightly off */

        } else {
            /* Nope, we got enough. Free back to heap
             */
            pHeap->v->Free(pHeap, 0, (void *)Td);
        }

        if (Value32)
            Td = (PMYOHCI_TD)((ADDRESS)PhysicalToVirtual(Value32) - offsetof(MYOHCI_TD,Td));
    } while (Value32);
}

/* Abuse: use the TDs for EDs also.
 */
#define EdAlloc(t)    (POHCI_ED)TdAlloc(t)
#define EdFree(t,e,f) {((PMYOHCI_TD)e)->Td.Next = 0;TdFree(t,(PMYOHCI_TD)e,f);}

/* Utilities for iTDs
 */
PRIVATE PMYOHCI_ITD
ItdAlloc(POHCI this)
{
    PMYOHCI_ITD Itd, Next;
    PIHEAP pHeap;

    /* Try to grab one from the free list
     */
    do {
        Itd = this->FreeItds;
        if (Itd == NULL)
            break;
        Next = (PMYOHCI_ITD) Itd->Urb;
    } while (!AtomicCmpAndSwap((ADDRESS*) &this->FreeItds,
                               (ADDRESS) Itd,
                               (ADDRESS) Next));

    /* If empty allocate a new one
     */
    if (Itd != NULL) {
        this->FreeItdCount--; /* no prob if slightly off */
    } else {
        pHeap = CurrentHeap();

        /* NB: NOT zeroed, like it comes from the freelist */
        Itd = (PMYOHCI_ITD) pHeap->v->Alloc(pHeap, 0,
                                          sizeof *Itd,
                                          OHCI_ED_ALIGNMENT+1);
        if (!Itd) {
            DBGME(3,_tprintf(_T("OHCI(%x) Itd outtamem.\n"), (ADDRESS) this));
            return NULL;
        }
    }

    Itd->Urb = NULL; /* sanity */
    return Itd;
}

/* Free a list of ITDs.
 * This is callable from an Isr,
 * in which case it wont go to the heap
 */
PRIVATE void
ItdFree(POHCI this, PMYOHCI_ITD Itd, BOOL FromIsr)
{
    PMYOHCI_ITD Next;
    UINT32 Value32;
    PIHEAP pHeap = CurrentHeap();

    do {
        Value32 = OhciReadRegister(&Itd->Td.Next);

        /* Put it on free list?
         */
        if (FromIsr || (this->FreeItdCount < FreeItdMax)) {
            /* Normally-won race to the top
             */
            do {
                Next = this->FreeItds;
                Itd->Urb = (void *) Next;
            } while (!AtomicCmpAndSwap((ADDRESS*) &this->FreeItds,
                                       (ADDRESS) Next,
                                       (ADDRESS) Itd));

            this->FreeItdCount++; /* no prob if slightly off */

        } else {
            /* Nope, we got enough. Free back to heap
             */
            pHeap->v->Free(pHeap, 0, (void *)Itd);
        }

        if (Value32)
            Itd = (PMYOHCI_ITD)((ADDRESS)PhysicalToVirtual(Value32) - offsetof(MYOHCI_ITD,Td));
    } while (Value32);
}

/* Utilities for handling EDs
 */
PRIVATE INLINE int
EdHash(UINT32 HashValue)
{
    /* Give two consecutive slots to each usb address.
     * One is for control, the other for EP 1.
     */
    return ((HashValue & OHCI_ED_FA) << 1) |
        (HashValue & 0x700/*OHCI_ED_EN^lowbit*/) |
        ((HashValue & 0x080) >> OHCI_ED_EN_SHIFT);
}

PRIVATE POHCI_ED
LookupEd(POHCI this,
         UINT8 DeviceAddress,
         UINT8 EndpointAddress,
         USBM_URB_FLAGS Flags,
         UINT  Mps)
{
    int i, j, HashKey;
    UINT32 HashValue;
    PIHEAP pHeap;
    POHCI_HT HashTable, NewHashTable;
    POHCI_ED Ed = NULL;

    /* Configuring?
     */
    if (DeviceAddress == 0) {
        OhciWriteRegister(&this->ControlHead->Caps,
                          ((Flags & URB_FLAG_LOW_SPEED) ? OHCI_ED_S : 0)
                          | ((Mps << OHCI_ED_MPS_SHIFT) & OHCI_ED_MPS));
        return this->ControlHead;
    }

    /* Build the hash table key
     * The HashValue is the initial Ed.Caps field
     * Notice that it only depends on addresses and mps.
     */
    HashValue = (DeviceAddress & OHCI_ED_FA) |
                ((EndpointAddress << OHCI_ED_EN_SHIFT) & OHCI_ED_EN) |
                OHCI_ED_FROM_TD |
                ((Flags & URB_FLAG_LOW_SPEED) ? OHCI_ED_S : 0) |
                ((Flags & URB_FLAG_ISOCHRONOUS) ? OHCI_ED_F : 0) |
                ((Mps << OHCI_ED_MPS_SHIFT) & OHCI_ED_MPS);
    HashKey = EdHash(HashValue);

    /* BUGBUG Will need this lock-free if Isr plays with it
     */
    Mutex_Lock(&this->HashTableLock);

    HashTable = this->HashTable;
    if (HashTable == NULL) {
        pHeap = CurrentHeap();

        HashTable = (POHCI_HT) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY,
                                               sizeof(*HashTable) +
                                               (InitialHashEntries-1) * sizeof(struct _OHCI_HT_ENTRY),
                                               0);
        if (!HashTable) {
            DBGME(3,_tprintf(_T("OHCI(%x) HT outtamem.\n"), (ADDRESS) this));
            goto UnlockAndDone;
        }
        HashTable->nEntries = InitialHashEntries;
        this->HashTable = HashTable;
    }

    /* Quick hit? */
    i = HashKey & (HashTable->nEntries-1);

    /* Scan forward the whole table to make sure we did not clash
     * or that this slot did not just free up
     */
    for (j = 0; j < HashTable->nEntries; j++) {
        if (HashTable->Entries[i].HashValue == HashValue) {
            Ed = HashTable->Entries[i].DataValue;
            goto UnlockAndDone;
        }
        if (++i == HashTable->nEntries)
            i = 0;
    }

    /* Ok, we got to allocate one
     */
    Ed = EdAlloc(this);
    if (Ed == NULL)
        goto UnlockAndDone;


    /* Initialize it
     */
    OhciWriteRegister(&Ed->Caps,HashValue);
    Ed->Tail = 0;
    Ed->Head = 0;
    Ed->Next = 0;

    /* Insert it in the hash table
     * Scan forward till we find a slot
     */
 Again:
    i = HashKey & (HashTable->nEntries-1);
    for (HashKey = 0; HashKey < HashTable->nEntries; HashKey++) {
        if (HashTable->Entries[i].HashValue == 0) {
            HashTable->Entries[i].HashValue = HashValue;
            HashTable->Entries[i].DataValue = Ed;
            DBGME(1,_tprintf(_T("OdLookup(%d,%d,%d)[%x] hash %x new @%x\n"),
                             DeviceAddress, EndpointAddress, Mps,
                             Flags,
                             HashValue, (ADDRESS) Ed));
            goto UnlockAndDone;
        }
        if (++i == HashTable->nEntries)
            i = 0;
    }

    /* Bummer, the table is full;
     * Allocate a 2x sized one, copy, free old, set new, restart.
     */
    i = HashTable->nEntries * 2;

    pHeap = CurrentHeap();
    NewHashTable = (POHCI_HT) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY,
                                              sizeof(*HashTable) +
                                              (i-1) * sizeof(struct _OHCI_HT_ENTRY),
                                              0);
    if (!NewHashTable) {
        DBGME(3,_tprintf(_T("OHCI(%x) HT outtamem2.\n"), (ADDRESS) this));
        goto UnlockAndDone;
    }

    memcpy(NewHashTable,
           HashTable,
           sizeof(*HashTable) + sizeof(struct _OHCI_HT_ENTRY) * (i/2 - 1));
    NewHashTable->nEntries = i;
    this->HashTable = NewHashTable;
    pHeap->v->Free(pHeap,0,HashTable);
    HashTable = NewHashTable;
    goto Again;

 UnlockAndDone:
    Mutex_Unlock(&this->HashTableLock);

    return Ed;
}

PRIVATE BOOL
DropEdFromQueue(
    POHCI this,
    POHCI_ED Ed,
    POHCI_ED QueueHead,
    volatile UINT32 *QueueHeadPtr
    )
{
    ADDRESS QueueHeadPhys;
    ADDRESS EdPhys = VirtualToPhysical(Ed);

    /* Some Q heads are permanent
     */
    if (QueueHead == NULL) {
        /* Either the control or bulk q.  Head pointer is in the OPS registers.
         */
        QueueHeadPhys = OhciReadRegister(QueueHeadPtr);
        if (QueueHeadPhys == 0) {
            return FALSE;
        }

        /* Special case if its the first in the list
         */
        QueueHead = PhysicalToVirtual(QueueHeadPhys);
        if (QueueHead == Ed) {
            /* NB: NEXT pointer is physical (and byteswapped) already
             */
            *QueueHeadPtr = Ed->Next;
            IOWriteBufferFlush();
            goto FoundIt;
        }
    } else {
        assert(QueueHead != Ed );
    }

    /* Scan until we find it.
     */
    while (QueueHead != NULL) {

        /* Is it the next one in the list ?
         */
        QueueHeadPhys = OhciReadRegister(&QueueHead->Next);

        if (QueueHeadPhys == EdPhys) {

            /* Drop it from the list
             */
            QueueHead->Next = Ed->Next;
            IOMemoryBarrier();

            /* Strictly speaking... at this point we should just set it aside
             * and wait until the next frame before we can safely corrupt
             * the memory that Ed points to.  With a bit more code we could
             * queue it somewhere and turn on the SOF interrupt, which
             * the interrupt handler would turn back off after moving the
             * queue around and... etc etc.
             * Or we can just sleep for one frame time.
             */
        FoundIt:
            MilliSleep(1);

            EdFree(this,Ed,FALSE);
            return TRUE;
        }

        /* Nope, move on to the next
         */
        QueueHead = (POHCI_ED) PhysicalToVirtual(QueueHeadPhys);
    }         

    return FALSE;
}

PRIVATE BOOL
DropEdFromAnyQueue(
    POHCI this,
    POHCI_ED Ed,
    UINT32 Caps
    )
{
    PUSKNODE Node; 

    /* Now we got to figure out where its queued.
     * Endpoint number 0 is assigned to the control endpoint.
     */
    if ((Caps & OHCI_ED_EN) == 0) {
        /* Control endpoint, check ControlQueueED
         */
        if (!DropEdFromQueue(this, Ed, NULL, &this->Ops->ControlHeadED)) {
            DBGME(3,_tprintf(_T("Ohci: Ed %x not found on CQ!!\n"),
                             (ADDRESS) Ed));
            return FALSE;
        }
        return TRUE;
    }

    /* Isochronous endpoints must have the Format bit set
     */
    if (Caps & OHCI_ED_F) {
        /* Isochronous endpoint, check IsoHead
         */
        if (!DropEdFromQueue(this, Ed, this->IsoHead, 0)) {
            DBGME(3,_tprintf(_T("Ohci: Ed %x not found on IsoQ!!\n"),
                             (ADDRESS) Ed));
            return FALSE;
        }
        return TRUE;
    }

    /* Urgh. Check the bulk queue first.
     */
    if (DropEdFromQueue(this, Ed, NULL, &this->Ops->BulkHeadED)) {
        return TRUE;
    }

    /* Must be in the schedule then.
     */
    Node = ((PMYOHCI_ED)Ed)->InterruptHead;
    if ((Node != NULL) &&
        DropEdFromQueue(this, Ed, &Node->HcNode, NULL)) {
        /* Was in the schedule, release the b/w reservation.
         */
        UINT16 MinBw, MaxBw;

        Ed = NULL;/* has been freed */
        OhciComputeBandwidth((USBM_URB_FLAGS)
                              (URB_FLAG_INTERRUPT |
                               ((Caps & OHCI_ED_S) ? URB_FLAG_LOW_SPEED:0)),
                             (UINT16)((Caps&OHCI_ED_MPS) >> OHCI_ED_MPS_SHIFT),
                             &MinBw, &MaxBw);
        UskCancelReservation(this->Schedule,
                             Node,
                             MinBw, MaxBw);
        return TRUE;
    }

    /* Something rotten in Denmark..
     */
    return FALSE;
}

PRIVATE void
DestroyEdsForFunction(POHCI this, 
                      UINT8 UsbAddress)
{
    POHCI_ED Ed;
    int i;
    UINT32 Value32;
    POHCI_HT HashTable;

    /* We do *not* free the config endpoint
     */
    if (UsbAddress == 0)
        return;

    /* Lock the hash table
     */
    Mutex_Lock(&this->HashTableLock);

    /* Sanity check the table is there (assert?)
     */
    HashTable = this->HashTable;
    if (HashTable != NULL) {

        /* Scan the whole table for EDs with this function address
         */
        for (i = 0; i < HashTable->nEntries; i++) {

            Value32 = HashTable->Entries[i].HashValue;
            if (UsbAddress == (Value32 & OHCI_ED_FA)) {
                /* Got one, free slot
                 */
                Ed = HashTable->Entries[i].DataValue;
                HashTable->Entries[i].DataValue = NULL;
                HashTable->Entries[i].HashValue = 0;

                /* Drop it from whatever queue its on
                 */
                if (!DropEdFromAnyQueue(this,Ed,Value32)) {
                    DBGME(3,_tprintf(_T("Ohci: Ed %x not found on any Q!!\n"),
                                     (ADDRESS) Ed));
                    /* This is bad but.. at least dont leak memory
                     */
                    EdFree(this,Ed,FALSE);
                }
            }
        }
    }

    Mutex_Unlock(&this->HashTableLock);
}

/* Enque a TD to an ED
 */
PRIVATE void
EdEnqueTd(POHCI_ED Ed, PTR Head, PTR Tail)
{
    UINT32 Value32;
    UINT32 Toggle;

    /* Head first, tail second
     * NB: Must preserve the toggle bit!
     */
    Toggle = OhciReadRegister(&Ed->Head) & OHCI_POINTER_DATA_TOGGLE;
    Value32 = VirtualToPhysical(Head) | Toggle;
    OhciWriteRegister(&Ed->Head,Value32);

    Value32 = VirtualToPhysical(Tail);
    OhciWriteRegister(&Ed->Tail,Value32);

    /* If it was idle then un-idle it
     */
    Value32 = OhciReadRegister(&Ed->Caps);
    if (Value32 & OHCI_ED_K) {
        Value32 &= ~OHCI_ED_K;
        OhciWriteRegister(&Ed->Caps,Value32);
    }
}

/* Debugging tools
 */
#if _DEBUG
typedef _TCHAR *MYSTRING;

typedef struct _REGFIELD {
    MYSTRING FieldName;
    UINT8 MsbNo, LsbNo; /* inclusive */
} REGFIELD, *PREGFIELD;

typedef struct _REGDESC {
    MYSTRING RegisterName;
    REGFIELD *Field; /* NULL terminated */
} REGDESC, *PREGDESC;

static const _TCHAR Spaces[16] = _T("               ");
#define Whitespace(n) (Spaces + (15 - ((n)&15)))

PRIVATE void
DumpReg(PREGDESC Format, UINT32 Value, int nSpaces)
{
    PREGFIELD Field;
    UINT32 Mask;
    UINT32 FieldValue;

    _tprintf(_T("%s%s %x"),Whitespace(nSpaces),Format->RegisterName,Value);
    if (Value) {
        _tprintf(_T(":"));
        for (Field = &Format->Field[0]; Field->FieldName != NULL; Field++) {
            Mask = ~0;
            Mask = Mask >> Field->LsbNo;
            Mask = Mask << Field->LsbNo;
            Mask = Mask << (31 - Field->MsbNo);
            Mask = Mask >> (31 - Field->MsbNo);

            FieldValue = (Value & Mask) >> Field->LsbNo;
            if (FieldValue!= 0)
                _tprintf(_T(" %s %x"),Field->FieldName, FieldValue);
        }
    }
    _tprintf(_T("\n"));
}

PRIVATE void
DumpEd(POHCI_ED Ed, int nSpaces)
{
    UINT32 Value32;

    static REGFIELD CapsFields[] = {
        { _T("FA"),  6, 0},
        { _T("EN"), 10, 7},
        { _T("D"),  12, 11},
        { _T("S"),  13, 13},
        { _T("K"),  14, 14},
        { _T("F"),  15, 15},
        { _T("MPS"), 26, 16},
        { NULL,}
    };
    static REGDESC Caps = {
        _T("Caps"), CapsFields
    };

    static REGFIELD HeadFields[] = {
        { _T("HeadP"),  31, 0},
        { _T("Halted"), 0, 0},
        { _T("Toggle"),  1, 1},
        { NULL,}
    };
    static REGDESC Head = {
        _T("Head"), HeadFields
    };

    _tprintf(_T("%sEndpoint @%x:\n"), Whitespace(nSpaces), (ADDRESS) Ed);
    nSpaces++;

    Value32 = OhciReadRegister(&Ed->Caps);
    if (Value32)
    DumpReg(&Caps,Value32,nSpaces);

    Value32 = OhciReadRegister(&Ed->Tail);
    if (Value32)
    _tprintf(_T("%sTail %x\n"),Whitespace(nSpaces),Value32);

    Value32 = OhciReadRegister(&Ed->Head);
    if (Value32)
    DumpReg(&Head,Value32,nSpaces);

    Value32 = OhciReadRegister(&Ed->Next);
    if (Value32)
    _tprintf(_T("%sNext %x\n"),Whitespace(nSpaces),Value32);

}

PRIVATE void
DumpEdList(POHCI_ED Ed, int nSpaces)
{
    UINT32 Value32;

    _tprintf(_T("%sEndpointList %x:\n"), Whitespace(nSpaces), (ADDRESS) Ed);
    nSpaces++;

    while (Ed) {
        DumpEd(Ed,nSpaces);
        Value32 = OhciReadRegister(&Ed->Next);
        Ed = (POHCI_ED)PhysicalToVirtual(Value32);
    }

}

PRIVATE void
DumpTd(POHCI_TD Td, int nSpaces)
{
    UINT32 Value32;

    static REGFIELD CapsFields[] = {
        { _T("R"),  18, 18},
        { _T("DP"), 20, 19},
        { _T("DI"), 23, 21},
        { _T("T"),  25, 24},
        { _T("EC"), 27, 26},
        { _T("CC"), 31, 28},
        { NULL,}
    };
    static REGDESC Caps = {
        _T("Caps"), CapsFields
    };

    _tprintf(_T("%sTransfer Descriptor @%x:\n"), Whitespace(nSpaces), (ADDRESS) Td);
    nSpaces++;

    Value32 = OhciReadRegister(&Td->Caps);
    if (Value32)
    DumpReg(&Caps,Value32,nSpaces);

    Value32 = OhciReadRegister(&Td->CurrentBufferPointer);
    if (Value32)
    _tprintf(_T("%sCurrentBufferPointer %x\n"),Whitespace(nSpaces),Value32);

    Value32 = OhciReadRegister(&Td->BufferEnd);
    if (Value32)
    _tprintf(_T("%sBufferEnd %x\n"),Whitespace(nSpaces),Value32);

    Value32 = OhciReadRegister(&Td->Next);
    if (Value32)
    _tprintf(_T("%sNext %x\n"),Whitespace(nSpaces),Value32);

}

PRIVATE void
DumpItd(POHCI_ITD Td, int nSpaces)
{
    int i;
    UINT32 Value32;
    UINT16 Value16;

    static REGFIELD CapsFields[] = {
        { _T("SF"), 15, 0},
        { _T("DI"), 23, 21},
        { _T("FC"), 26, 24},
        { _T("CC"), 31, 18},
        { NULL,}
    };
    static REGDESC Caps = {
        _T("Caps"), CapsFields
    };

    _tprintf(_T("%sTransfer Descriptor @%x:\n"),Whitespace(nSpaces), (ADDRESS) Td);
    nSpaces++;

    Value32 = OhciReadRegister(&Td->Caps);
    if (Value32)
    DumpReg(&Caps,Value32,nSpaces);

    Value32 = OhciReadRegister(&Td->BufferPage0);
    if (Value32)
    _tprintf(_T("%sBufferPage0 %x\n"),Whitespace(nSpaces),Value32);

    Value32 = OhciReadRegister(&Td->BufferEnd);
    if (Value32)
    _tprintf(_T("%sBufferEnd %x\n"),Whitespace(nSpaces),Value32);

    Value32 = OhciReadRegister(&Td->Next);
    if (Value32)
    _tprintf(_T("%sNext %x\n"),Whitespace(nSpaces),Value32);

    _tprintf(_T("%sOffsetPsw:"),Whitespace(nSpaces));
    for (i = 0; i < 8; i++) {
        Value16 = OhciReadRegister16(&Td->OffsetPsw[i]);
        _tprintf(_T("%s%4x"),Whitespace(nSpaces),Value16);
        if ((i & 3) == 3)
            _tprintf(_T("\n"));
    }
}

PRIVATE void
DumpTdList(POHCI_TD Td, int nSpaces)
{
    UINT32 Value32;

    _tprintf(_T("%sTDList @%x:\n"), Whitespace(nSpaces), (ADDRESS) Td);
    nSpaces++;

    do {
        DumpTd(Td,nSpaces);
        Value32 = OhciReadRegister(&Td->Next);
        Td = (POHCI_TD) PhysicalToVirtual(Value32);
    } while (Value32);
}

PRIVATE void
DumpItdList(POHCI_ITD Td, int nSpaces)
{
    UINT32 Value32;

    _tprintf(_T("%sITDList @%x:\n"), Whitespace(nSpaces), (ADDRESS) Td);
    nSpaces++;

    do {
        DumpItd(Td,nSpaces);
        Value32 = OhciReadRegister(&Td->Next);
        Td = (POHCI_ITD) PhysicalToVirtual(Value32);
    } while (Value32);
}

PRIVATE void
DumpHcca(POHCI_HCCA Hcca, int nSpaces)
{
    int i, j;
    UINT32 Value32, Ed32;

    if (Hcca == NULL)
        return; /* sanity */

    _tprintf(_T("%sHCCA @%x:\n"),Whitespace(nSpaces), (ADDRESS) Hcca);
    nSpaces++;

    _tprintf(_T("%sInterruptTable:\n%s"),Whitespace(nSpaces),Whitespace(nSpaces));
    for (i = 0; i < OHCI_NINTED; i++) {
        Value32 = OhciReadRegister(&Hcca->InterruptTable[i]);
        _tprintf(_T("%s%8x"),Whitespace(nSpaces),Value32);
        if ((i & 3) == 3)
            _tprintf(_T("\n%s"),Whitespace(nSpaces));
    }

    /* Now dump (once) the descriptors in the IT */
    for (i = 0; i < OHCI_NINTED; i++) {
        Ed32 = OhciReadRegister(&Hcca->InterruptTable[i]);
        for (j = 0; j < i; j++) {
            Value32 = OhciReadRegister(&Hcca->InterruptTable[j]);
            if (Value32 == Ed32)
                goto NextOne;
        }
        if (Ed32) {
            _tprintf(_T("\n%sInterruptED[%d]"),Whitespace(nSpaces+1),i);
            DumpEdList((POHCI_ED)PhysicalToVirtual(Ed32),nSpaces+1);
        }
    NextOne:
        continue;
    }

    Value32 = OhciReadRegister(&Hcca->FrameNumber);
    _tprintf(_T("%sFrameNumber %x\n"),Whitespace(nSpaces),Value32);

    Value32 = OhciReadRegister(&Hcca->DoneHead);
    _tprintf(_T("%sDoneHead %x\n"),Whitespace(nSpaces),Value32);
    if (Value32) {
        _tprintf(_T("%sDoneHead"),Whitespace(nSpaces));
        DumpTd((POHCI_TD)PhysicalToVirtual(Value32),nSpaces+1);
    }

}

#define OPS_CONTROL 0x01
#define OPS_POINTER 0x02
#define OPS_FRAME   0x04
#define OPS_ROOTHUB 0x08
#define OPS_ALL     0xff

PRIVATE void
DumpOps(POHCI_OPS Ops, int nSpaces, UINT PartitionMask)
{
    int i, nPorts;
    UINT32 Value32;
    static REGFIELD RevisionFields[] = {
        { _T("REV"),  7, 0},
        { _T("LEGC"), 8, 8},
        { NULL,}
    };
    static REGDESC Revision = {
        _T("Revision"), RevisionFields
    };

    static REGFIELD ControlFields[] = {
        { _T("CBSR"),  1, 0},
        { _T("PLE"),  2, 2},
        { _T("IE"),  3, 3},
        { _T("CLE"),  4, 4},
        { _T("BLE"),  5, 5},
        { _T("HCFS"),  7, 6},
        { _T("IR"),  8, 8},
        { _T("RWC"),  9, 9},
        { _T("RWE"),  10, 10},
        { NULL,}
    };
    static REGDESC Control = {
        _T("Control"), ControlFields
    };

    static REGFIELD CommandStatusFields[] = {
        { _T("HCR"),  0, 0},
        { _T("CLF"),  1, 1},
        { _T("BLF"),  2, 2},
        { _T("OCR"),  3, 3},
        { _T("SOC"),  17, 16},
        { NULL,}
    };
    static REGDESC CommandStatus = {
        _T("CommandStatus"), CommandStatusFields
    };

    static REGFIELD InterruptStatusFields[] = {
        { _T("SO"),  0, 0},
        { _T("WDH"),  1, 1},
        { _T("SF"),  2, 2},
        { _T("RD"),  3, 3},
        { _T("UE"),  4, 4},
        { _T("FNO"),  5, 5},
        { _T("RHSC"),  6, 6},
        { _T("OC"),  30, 30},
        { _T("MIE"),  31, 31},
        { NULL,}
    };
    static REGDESC InterruptStatus = {
        _T("InterruptStatus"), InterruptStatusFields
    };
    static REGDESC InterruptEnable = {
        _T("InterruptEnable"), InterruptStatusFields
    };
    static REGDESC InterruptDisable = {
        _T("InterruptDisable"), InterruptStatusFields
    };

    static REGFIELD HCCAFields[] = {
        { _T("ADDR31..8"),  31, 0},
        { NULL,}
    };
    static REGDESC HCCA = {
        _T("HCCA"), HCCAFields
    };

    static REGFIELD EDFields[] = {
        { _T("ADDR31..4"),  31, 0},
        { NULL,}
    };
    static REGDESC PeriodCurrentED = {
        _T("PeriodCurrentED"), EDFields
    };
    static REGDESC ControlHeadED = {
        _T("ControlHeadED"), EDFields
    };
    static REGDESC ControlCurrentED = {
        _T("ControlCurrentED"), EDFields
    };
    static REGDESC BulkHeadED = {
        _T("BulkHeadED"), EDFields
    };
    static REGDESC BulkCurrentED = {
        _T("BulkCurrentED"), EDFields
    };
    static REGDESC DoneHead = {
        _T("DoneHead"), EDFields
    };

    static REGFIELD FmIntervalFields[] = {
        { _T("FI"),  13, 0},
        { _T("FSMPS"),  30, 16},
        { _T("FIT"),  31, 31},
        { NULL,}
    };
    static REGDESC FmInterval = {
        _T("FmInterval"), FmIntervalFields
    };

    static REGFIELD FmRemainingFields[] = {
        { _T("FR"),  13, 0},
        { _T("FRT"),  31, 31},
        { NULL,}
    };
    static REGDESC FmRemaining = {
        _T("FmRemaining"), FmRemainingFields
    };

    static REGFIELD FmNumberFields[] = {
        { _T("FN"),  15, 0},
        { NULL,}
    };
    static REGDESC FmNumber = {
        _T("FmNumber"), FmNumberFields
    };

    static REGFIELD PeriodicStartFields[] = {
        { _T("PS"),  13, 0},
        { NULL,}
    };
    static REGDESC PeriodicStart = {
        _T("PeriodicStart"), PeriodicStartFields
    };

    static REGFIELD LSThresholdFields[] = {
        { _T("LST"),  11, 0},
        { NULL,}
    };
    static REGDESC LSThreshold = {
        _T("LSThreshold"), LSThresholdFields
    };

    static REGFIELD RhDescriptorAFields[] = {
        { _T("NDP"),  7, 0},
        { _T("PSM"),  8, 8},
        { _T("NPS"),  9, 9},
        { _T("DT"),  10, 10},
        { _T("OCPM"),  11, 11},
        { _T("NOCP"),  12, 12},
        { _T("POTPG"),  31, 24},
        { NULL,}
    };
    static REGDESC RhDescriptorA = {
        _T("RhDescriptorA"), RhDescriptorAFields
    };

    static REGFIELD RhDescriptorBFields[] = {
        { _T("DR"),  3, 0},
        { _T("PPCM"), 19, 16},
        { NULL,}
    };
    static REGDESC RhDescriptorB = {
        _T("RhDescriptorB"), RhDescriptorBFields
    };

    static REGFIELD RhStatusFields[] = {
        { _T("LPS"),  0, 0},
        { _T("OCI"),  1, 1},
        { _T("DRWE"), 15, 15},
        { _T("LPSC"),  16, 16},
        { _T("OCIC"),  17, 17},
        { _T("CRWE"),  31, 31},
        { NULL,}
    };
    static REGDESC RhStatus = {
        _T("RhStatus"), RhStatusFields
    };

    static REGFIELD RhPortStatusFields[] = {
        { _T("CCS"),  0, 0},
        { _T("PES"),  1, 1},
        { _T("PSS"),  2, 2},
        { _T("POCI"), 3, 3},
        { _T("PRS"),  4, 4},
        { _T("PPS"),  8, 8},
        { _T("LSDA"),  9, 9},
        { _T("CSC"),  16, 16},
        { _T("PESC"), 17, 17},
        { _T("PSSC"), 18, 18},
        { _T("OCIC"), 19, 19},
        { _T("PRSC"), 20, 20},
        { NULL,}
    };
    static REGDESC RhPortStatus = {
        _T("RhPortStatus"), RhPortStatusFields
    };

    _tprintf(_T("%sOps @%x:\n"),Whitespace(nSpaces), (ADDRESS) Ops);
    nSpaces++;

#define Dumpit(f) \
    Value32 = OhciReadRegister(&Ops->f); \
    DumpReg(&f,Value32,nSpaces);

    if (PartitionMask & OPS_CONTROL) {
        Dumpit(Revision);  Dumpit(Control);  Dumpit(CommandStatus);
        Dumpit(InterruptStatus); Dumpit(InterruptEnable); Dumpit(InterruptDisable);
    }

    if (PartitionMask & OPS_POINTER) {
        Dumpit(HCCA);
        DumpHcca(PhysicalToVirtual(Value32),nSpaces+1);
        Dumpit(PeriodCurrentED);
        if (Value32) {
            _tprintf(_T("%sPeriodicCurrent"),Whitespace(nSpaces+1));
            DumpEd((POHCI_ED)PhysicalToVirtual(Value32),nSpaces+1);
        }
        Dumpit(ControlHeadED);
        if (Value32) {
            _tprintf(_T("%sControlHead"),Whitespace(nSpaces+1));
            DumpEdList((POHCI_ED)PhysicalToVirtual(Value32),nSpaces+1);
        }
        Dumpit(ControlCurrentED);
        if (Value32) {
            _tprintf(_T("%sControlCurrent"),Whitespace(nSpaces+1));
            DumpEd((POHCI_ED)PhysicalToVirtual(Value32),nSpaces+1);
        }
        Dumpit(BulkHeadED);
        if (Value32) {
            _tprintf(_T("%sBulkHead"),Whitespace(nSpaces+1));
            DumpEdList((POHCI_ED)PhysicalToVirtual(Value32),nSpaces+1);
        }
        Dumpit(BulkCurrentED);
        if (Value32) {
            POHCI_ED Ed;
            _tprintf(_T("%sBulkCurrent"),Whitespace(nSpaces+1));
            Ed = (POHCI_ED)PhysicalToVirtual(Value32);
            DumpEd(Ed,nSpaces+1);

            Value32 = OhciReadRegister(&Ed->Head) & OHCI_POINTER_MASK;
            if (Value32) {
                _tprintf(_T("%sBulk"),Whitespace(nSpaces+3));
                DumpTdList((POHCI_TD)PhysicalToVirtual(Value32),nSpaces+3);
            }
        }
        Dumpit(DoneHead); 
        if (Value32) {
            _tprintf(_T("%sDoneHead"),Whitespace(nSpaces+1));
            DumpEd((POHCI_ED)PhysicalToVirtual(Value32),nSpaces+1);
        }
    }

    if (PartitionMask & OPS_FRAME) {
        Dumpit(FmInterval); Dumpit(FmRemaining); Dumpit(FmNumber);
        Dumpit(PeriodicStart);
        Dumpit(LSThreshold);
    }

    if (PartitionMask & OPS_ROOTHUB) {
        Dumpit(RhDescriptorA);
        nPorts = Value32 & OHCI_RH_NDP;
        Dumpit(RhDescriptorB);
        Dumpit(RhStatus);

        for (i = 0; i < nPorts; i++) {
            Value32 = OhciReadRegister(&Ops->RhPortStatus[i]);
            _tprintf(_T("%s[%d]"),Whitespace(nSpaces),i);
            DumpReg(&RhPortStatus,Value32,0);
        }
    }
#undef Dumpit
}

PRIVATE MYSTRING OhciErrors[16] = {
    _T("no"), _T("CRC"), _T("BIT STUFFING"), _T("TOGGLE MISMATCH"),
    _T("STALL"), _T("NO RESPONSE"), _T("PID CHECK"), _T("PID"),
    _T("DATA OVERRUN"), _T("DATA UNDERRUN"), _T("0xa"), _T("0xb"),
    _T("BUFFER OVERRUN"), _T("BUFFER UNDERRUN"), _T("not accessed"), _T("not accessed 0xf")
};
#define OhciErrorName(n) OhciErrors[n&15]

PRIVATE void DumpPacket(PUSBM_REQUEST_BLOCK Packet)
{
    static REGFIELD FlagsFields[] = {
        { _T("OUT"),  0, 0},
        { _T("IN"),  1, 1},
        { _T("ISO"), 2, 2},
        { _T("CONTROL"), 3, 3},
        { _T("INTERRUPT"), 4, 4},
        { _T("LOW_SPEED"), 5, 5},
        { _T("HIGH_SPEED"), 6, 6},
        { _T("COMPLETE"), 14, 14},
        { _T("ERROR"), 15, 15},
        { NULL,}
    };
    static REGDESC Flags = {
        _T("Flags"), FlagsFields
    };

    int i, n;

    _tprintf(_T("Packet @x%x "), (ADDRESS)Packet);
    if (Packet->Flags & URB_FLAG_ERROR)
        _tprintf(_T("had a %s error\n"),OhciErrorName(Packet->FinalStatus));
    else
        _tprintf((Packet->Flags & URB_FLAG_COMPLETE) ? _T("successfully completed\n") :_T("is executing\n"));

#define indent 4
#define sp() Whitespace(indent)
    _tprintf(_T("%sDeviceAddress %x\n"), sp(), Packet->DeviceAddress);
    _tprintf(_T("%sEndpointAddress %x\n"), sp(), Packet->EndpointAddress);
    if (Packet->Flags & URB_FLAG_INTERRUPT)
        _tprintf(_T("%sPolling Interval %d\n"), sp(), Packet->Interval);
    _tprintf(_T("%sPid %x\n"), sp(), Packet->Pid);
    DumpReg(&Flags,(UINT32)Packet->Flags,indent);
    _tprintf(_T("%sSetupLength %x\n"), sp(), Packet->SetupLength);
    n = Packet->SetupLength;
    if (n)
    {
        _tprintf(_T("%sSetupData[%d]@%x:"), sp(),n,(ADDRESS)Packet->SetupData);
        for (i = 0; i < n; i++)
            _tprintf(_T(" %2x"), Packet->SetupData[i]);
        _tprintf(_T("\n"));
    }
    _tprintf(_T("%sExpectedDataLength %x\n"), sp(), Packet->ExpectedDataLength);
    _tprintf(_T("%sMaximumPacketSize %d\n"), sp(), Packet->MaximumPacketSize);
    _tprintf(_T("%sActualDataLength %x\n"), sp(), Packet->ActualDataLength);
    n = Packet->ActualDataLength;
    if (n)
    {
        _tprintf(_T("%sResult[%d]@%x:"), sp(),n,(ADDRESS)Packet->Data);
        for (i = 0; i < n; i++)
            _tprintf(_T(" %2x"), Packet->Data[i]);
        _tprintf(_T("\n"));
    }
    if (Packet->Flags & URB_FLAG_ERROR)
        _tprintf(_T("%sFinalStatus %x\n"), sp(), Packet->FinalStatus);
#undef sp
#undef indent
}

PRIVATE void print_stats(struct _STATS *Stats)
{
    _tprintf(_T(" Stats:"));
    _tprintf(_T(" Isr-Interrupts = %d"), Stats->IsrCnt);
    _tprintf(_T(" Thd-Interrupts = %d\n"), Stats->IntrCnt);
    _tprintf(_T("\tUSBINT = %d"), Stats->RegCnt);
    _tprintf(_T(" Change = %d"), Stats->ChgCnt);
    _tprintf(_T(" Errors = %d\n"), Stats->ErrCnt);
}

PRIVATE void
DumpHt(POHCI_HT Ht, int nSpaces)
{
    int i;

    _tprintf(_T("%sHashTable @%x:\n"),Whitespace(nSpaces), (ADDRESS) Ht);
    if (Ht == NULL)
        return;
    nSpaces++;

    _tprintf(_T("%s%d Slots%c\n"),Whitespace(nSpaces),
             Ht->nEntries, (Ht->nEntries > 0) ? _T(':') : _T(' '));
    for (i = 0; i < Ht->nEntries; i++) {
        /* Zero means free, dont print */
        if (Ht->Entries[i].HashValue == 0)
            continue;
        _tprintf(_T("%s[%d] Hash %x Data %x\n"),Whitespace(nSpaces+1), i,
                 Ht->Entries[i].HashValue, (ADDRESS) Ht->Entries[i].DataValue);
    }
}

PRIVATE void
DumpOhci(POHCI this)
{
    _tprintf(_T("Ohci @%x:\n"), (ADDRESS)this);

    print_stats(&this->Stats);

    _tprintf(_T(" HCCA %x\n"), (ADDRESS)this->HCCA);
    _tprintf(_T(" IsoHead %x"), (ADDRESS)this->IsoHead);
    _tprintf(_T(" ControlHead %x"), (ADDRESS)this->ControlHead);
    _tprintf(_T(" BulkHead %x\n"), (ADDRESS)this->BulkHead);
    _tprintf(_T(" nPorts %d"), this->nPorts);
    _tprintf(_T(" PortPowerControl %x"), (ADDRESS)this->PortPowerControl);
    _tprintf(_T(" EvPkt %x Psc %x\n"),
             (ADDRESS)this->EventsPacket, this->PortStatusChanges);
    DumpHt(this->HashTable,1);

    DumpOps(this->Ops,1,OPS_ALL);

}
#else
#define DumpOhci(x)
#define DumpPacket(x)
#define DumpTdList(x,y)
#define DumpItdList(x,y)
#endif

/*
 * CDevice interface
 */

PRIVATE SCODE OhciDeviceQueryInterface (
    PIDEVICE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_IFile,iid)
        || UuidCmp(&IID_IDevice,iid))
    {
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }
    if (UuidCmp(&IID_IUsbMasterController, iid)) {
        /* Use a single refcount
         */
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (void*) iOhciMaster(pThis);
        return S_OK;
    }

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT OhciDeviceAddRef (
    PIDEVICE pThis)
{
    return AtomicInc(&pOhciDevice(pThis)->CDevice.RefCnt);
}

PRIVATE UINT OhciDeviceRelease (
    PIDEVICE pThis)
{
    POHCI this = pOhciDevice(pThis);
    UINT Refs = AtomicDec(&this->CDevice.RefCnt);
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* BUGBUG Finalize object */

        pHeap->v->Free(pHeap, 0, (void *) this);
    }
    return Refs;
}

#if _DEBUG
PRIVATE SCODE MCT OhciDeviceReadAt (PIDEVICE pThis, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    POHCI this = pOhciDevice(pThis);

    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);

    DumpOhci(this);

    if (pSizeRead)
        *pSizeRead = 0;
    return S_FALSE; /* EOF */

}
#else
#define OhciDeviceReadAt NULL
#endif

/* Returns the number of ports attached
 */
PRIVATE SCODE OhciDeviceGetSize (
    PIDEVICE pThis,
    /*out*/ UINT64 * Size)
{
    POHCI this = pOhciDevice(pThis);

    /* Number of ports on root hub
     */
    Uint32ToUint64(*Size,this->nPorts);

    return S_OK;
}

/* Interrupt handling
 */
#define IMASK (OHCI_I_SO|OHCI_I_WDH|OHCI_I_RD|OHCI_I_UE|OHCI_I_RHSC|OHCI_I_MIE)
#define OhciDisableInterrupts(Ops) OhciWriteRegister(&Ops->InterruptDisable,OHCI_I_MIE)
#define OhciEnableInterrupts(Ops)  OhciWriteRegister(&Ops->InterruptEnable,IMASK)

PRIVATE BOOL OhciDeviceIsr (
    PIDEVICE pThis,
    /*out*/ BOOL * pNotOurInterrupt)
{
    POHCI this = pOhciDevice(pThis);
    POHCI_OPS Ops = this->Ops;
    UINT32 Status;

    this->Stats.IsrCnt++;

    /* This interrupt is shared, we must know which is ours
     */
    Status = OhciReadRegister(&Ops->InterruptStatus) & IMASK;
    if (Status == 0) {
        *pNotOurInterrupt = TRUE;
        return FALSE;
    }

    /* Disable further interrupts
     */
    OhciDisableInterrupts(Ops);

    return Condition_InterruptSignal(&this->CDevice.IntrPoint);
}

/* Check for changes in the root HUB status
 */
PRIVATE void
OhciRootHubStatus(POHCI this)
{
    POHCI_OPS Ops = this->Ops;
    UINT8 i;
    UINT32 Value32;
    PUSBM_REQUEST_BLOCK Urb;

#if 0
    DumpOps(this->Ops,1,OPS_ROOTHUB);
#endif

    this->PortStatusChanges = 0;
    for (i = 0; i < this->nPorts; i++) {
        Value32 = OhciReadRegister(&Ops->RhPortStatus[i]);
        if (Value32 & OHCI_PS_CHANGE) {
            this->PortStatusChanges |= (2 << i); /*p0->p1*/
        }
    }

    /* Notify any waiters
     */
    Urb = this->EventsPacket;
    if (Urb) {
        this->EventsPacket = NULL;
        Urb->ActualDataLength = 2;
        Set2(Urb->Data,this->PortStatusChanges);
        Urb->Flags |= URB_FLAG_COMPLETE;
        Condition_InterruptSignal(Urb->Done);
    }
}

/* Handle TD completion interrupts
 * TD lists are returned reversed, oh joy.
 */
PRIVATE void
HandleTdComplete(POHCI this, UINT32 DoneHeadValue)
{
    PMYOHCI_TD Td;
    POHCI_ED Ed;
    PUSBM_REQUEST_BLOCK Urb;
    PHCI_STATE Hs;
    UINT32 Value32;
    BOOL CompleteIt = TRUE;

    if (DoneHeadValue & 1) {
        DBGME(2,_tprintf(_T("OhciComplete: headache %x\n"), DoneHeadValue));
    }
    DoneHeadValue &= OHCI_POINTER_MASK;

    Value32 = DoneHeadValue;

    while (Value32) {

        /* Map back and grab the info
         */
        Td = (PMYOHCI_TD) PhysicalToVirtual(Value32);
        Urb = Td->Urb;
        Hs = (PHCI_STATE)(&Urb->HciReserved);
        Value32 = OhciReadRegister(&Td->Td.Caps);

        /* If the URB has been cancelled bail out
         */
        if (Hs->Ed == NULL)
            break;

        /* Remember the first error code, if any
         */
        if (Hs->cc == 0) {
            Hs->cc = (UINT8)((Value32 & OHCI_TD_CC) >> OHCI_TD_CC_SHIFT);

            /* Check bytecount (of leftover bytes).
             */
            if ((Hs->cc == 0) &&
                (Td->Td.BufferEnd && Td->Td.CurrentBufferPointer)) {
                Urb->ActualDataLength -= (UINT16)
                    (OhciReadRegister(&Td->Td.BufferEnd) -
                     OhciReadRegister(&Td->Td.CurrentBufferPointer) + 1);
            }
        }

        /* Are we done yet.
         * On error, most likely the ED has been halted and the remaining
         * TDs, if any, will not complete. So we stop on first error.
         */
        if ((--Hs->nFrags == 0) || (Hs->cc)) {
            /* Set completion flags and status
             */
            Urb->Flags |= (Hs->cc) ? (URB_FLAG_COMPLETE|URB_FLAG_ERROR)
                                     : URB_FLAG_COMPLETE;
            Urb->FinalStatus = Hs->cc;
            assert(OHCI_STALL_ERROR == (URB_STATUS_STALL << OHCI_TD_CC_SHIFT));

            /* Errors?  Check the ED and see if its halted.
             * Pick up by hand the rest of the TD list if so.
             * If it isnt halted we'll get another interrupt.
             */
            Ed = Hs->Ed;
            if (Hs->cc) {
                Value32 = OhciReadRegister(&Ed->Head);
                if (Value32 & OHCI_POINTER_HALTED) {
                    UINT32 TdPhys = Value32 & OHCI_POINTER_MASK;
                    PMYOHCI_TD TdBad;

                    /* Leave only the terminal node
                     */
                    Value32 = OhciReadRegister(&Ed->Tail);
                    Ed->Head = Ed->Tail;

                    /* Was there anything left.
                     */
                    while (Value32 != TdPhys) {
                        TdBad = (PMYOHCI_TD) PhysicalToVirtual(TdPhys);
                        
                        /* truncate list, free */
                        TdPhys = OhciReadRegister(&TdBad->Td.Next);
                        OhciWriteRegister(&TdBad->Td.Next,0);

                        DBGME(3,DumpTdList(&TdBad->Td,0));
                        TdFree(this,TdBad,FALSE);

                        /* Sanity */
                        if (TdPhys == 0)
                            break;
                    }
                    
                    CompleteIt = TRUE;
                } else {
                    CompleteIt = (Hs->nFrags == 0);
                }

                /* Say there is no valid data whatsoever
                 */
                Urb->ActualDataLength = 0;                
            }

            /* Mark the ED idle, if it is
             * This avoid hicups when we ping the queue the next time,
             * because the HC stops at the first non-skip one that has
             * Head==Tail.  Go figure.
             */
            if (Ed->Tail == (Ed->Head & _bs32(OHCI_POINTER_MASK))) {
                Value32 = OhciReadRegister(&Ed->Caps);
                Value32 |= OHCI_ED_K;
                OhciWriteRegister(&Ed->Caps,Value32);
            }

            /* Signal completion
             */
            if (CompleteIt) {
                /* Say what
                 */
#if _DEBUG
                if (Urb->Flags & URB_FLAG_ERROR) {
                    DBGME(3,DumpPacket(Urb));
                } else {
                    DBGME(0,DumpPacket(Urb));
                }
#endif
                Condition_InterruptSignal(Urb->Done);
            }
        }

        /* Next in line
         */
        Value32 = OhciReadRegister(&Td->Td.Next);
    }

    Td = (PMYOHCI_TD) PhysicalToVirtual(DoneHeadValue);
    DBGME(0,DumpTdList(&Td->Td,0));
    TdFree(this,Td,FALSE); /* BUGBUG fix to TRUE when move to isr context */
}

/* Thread-level interrupt handling
 */
PRIVATE BOOL MCT
OhciDeviceHandleInterrupt(PIDEVICE pThis)
{
    POHCI this = pOhciDevice(pThis);
    POHCI_OPS Ops = this->Ops;
    UINT32 Status;
    UINT32 Value32;

    /* Read but do not ack the interrupt yet, the queue head might advance
     */
    Status = OhciReadRegister(&Ops->InterruptStatus) & IMASK;
    this->Stats.IntrCnt++;

    /* Handle writeback of the DoneHead first
     */
    if (Status & OHCI_I_WDH) {

        Value32 = OhciReadRegister(&this->HCCA->DoneHead);
        OhciWriteRegister(&this->HCCA->DoneHead,0);

        /* Ack immediately */
        OhciWriteRegister(&Ops->InterruptStatus,OHCI_I_WDH);
        Status &= ~OHCI_I_WDH;

        if (Value32)
            HandleTdComplete(this,Value32);
    }

    /* HUB Changes
     */
    if (Status & OHCI_I_RHSC) {
        OhciRootHubStatus(this);
    }

    /* Resume signalling
     */
    if (Status & OHCI_I_RD) {
        _tprintf(_T("OhciIntr: resume Sts %x Cnt %d\n"),
                 Status, this->Stats.IsrCnt);
        DumpOhci(this);
    }
    
    /* Handle errors
     */
    if (Status & (OHCI_I_SO|OHCI_I_UE)) {
        _tprintf(_T("OhciIntr: ERROR Sts %x Cnt %d\n"),
                 Status, this->Stats.IsrCnt);
        DumpOhci(this);
    }

    /* We are done this round; ack, reenable and go back to sleep
     */
    OhciWriteRegister(&Ops->InterruptStatus,Status);
    OhciEnableInterrupts(Ops);
    return TRUE;
}


/* Initialization
 */
PRIVATE INT OhciDeviceInitialize (
    PIDEVICE pThis)
{
    POHCI this = pOhciDevice(pThis);
    POHCI_OPS Ops = this->Ops;
    UINT32 Value32, Potpg;
    UINT32 FmInterval;
    int i;

    DBGME(0,_tprintf(_T("OhciInitialize(%x)\n"), (ADDRESS)pThis));

    /* Disable interrupts
     */
    OhciDisableInterrupts(Ops);

    /* Save FmInterval across reset
     */
    FmInterval = OhciReadRegister(&Ops->FmInterval);

    /* Issue HCReset
     */
    Value32 = OHCI_CS_HCR;
    OhciWriteRegister(&Ops->CommandStatus,Value32);

    /* Initialize data structures while we wait for reset to complete
     * 1: Init qheads
     */
    memset(this->BulkHead,0,sizeof(OHCI_ED));
    OhciWriteRegister(&this->BulkHead->Caps,OHCI_ED_K);

    memset(this->ControlHead,0,sizeof(OHCI_ED));
    OhciWriteRegister(&this->ControlHead->Caps,(8<<OHCI_ED_MPS_SHIFT));

    memset(this->IsoHead,0,sizeof(OHCI_ED));
    OhciWriteRegister(&this->IsoHead->Caps,OHCI_ED_K|OHCI_ED_F);

    /* 3: Translate entries in the HCCA to physical addresses */
    for (i = 0; i < OHCI_NINTED; i++) {
        Value32 = VirtualToPhysical(this->HCCA->InterruptTable[i]);
        OhciWriteRegister(&this->HCCA->InterruptTable[i],Value32);
    }


    /* Wait for reset to complete
     */
    for (i = 0; i < 1000; i++) {
        Value32 = OhciReadRegister(&Ops->CommandStatus);
        if ((Value32 & OHCI_CS_HCR) == 0)
            break;
        Delay(10);
    }

    /* See if we reset ok
     * BUGBUG We really should also check that state is USB_SUSPEND
     */
    if (Value32 & OHCI_CS_HCR) {
        DBGME(3,_tprintf(_T("OHCI @%x HCreset stuck.\n"),
                         (ADDRESS) this->Ops));
        return 1;
    }

    /* Move it to the reset state
     */
    Value32 = (Value32 & ~(OHCI_C_HCFS)) | OHCI_C_USBRESET;
    OhciWriteRegister(&Ops->CommandStatus,Value32);

    /* Restore FmInterval
     */
    OhciWriteRegister(&Ops->FmInterval, FmInterval ^ OHCI_FI_FIT);

    /* Allocate 10% bw to control endpoints
     */
    FmInterval &= OHCI_FI_FI;
    Value32 = FmInterval - (FmInterval / 10);
    OhciWriteRegister(&Ops->PeriodicStart,Value32);

    /* Reset pointers
     */
    OhciWriteRegister(&Ops->ControlHeadED,
                      VirtualToPhysical(this->ControlHead));
    OhciWriteRegister(&Ops->BulkHeadED,
                      VirtualToPhysical(this->BulkHead));
    OhciWriteRegister(&Ops->HCCA,
                      VirtualToPhysical(this->HCCA));

    /* See how long we have to wait for the root HUB, in msecs
     */
    Potpg = OhciReadRegister(&Ops->RhDescriptorA);
    Potpg = (Potpg & OHCI_RH_POTPG) >> (OHCI_RH_POTPG_SHIFT - 1);

    /* Wait for USB bus to stabilize
     */
    MilliSleep(Potpg);

    /* Get it running
     */
    OhciWriteRegister(&Ops->Control,
                      OHCI_C_1to1 | OHCI_C_PLE | OHCI_C_IE | OHCI_C_CLE |
                      OHCI_C_BLE | OHCI_C_USBOPERATIONAL);
    Delay(20);

    /* If port power control is global then turn it on now
     */
    OhciWriteRegister(&Ops->RhStatus,OHCI_ST_SGP);

    MilliSleep(Potpg);

    /* Powerup all ports, if they are already up, this is a noop.
     * Also, set the reset bit in case they are powered up.
     */
    for (i = 0; i < this->nPorts; i++) {
        Value32 = OhciReadRegister(&Ops->RhPortStatus[i]);
        Value32 |= OHCI_PS_SPP | OHCI_PS_SPR;
        OhciWriteRegister(&Ops->RhPortStatus[i],Value32);
    }

#if 0
    /* Give powerup time to take
     */
    MilliSleep(500);
#endif

    /* Now turn on interrupts and be ready for connect change events
     */
    OhciEnableInterrupts(Ops);

    return 0;
}

/*
 * IUsbMasterController interface
 */
PRIVATE SCODE OhciMasterQueryInterface (
    PIUSBMASTERCONTROLLER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return OhciDeviceQueryInterface(iOhciDevice(pOhciMaster(pThis)), iid, ppObject);
}

PRIVATE UINT OhciMasterAddRef (
    PIUSBMASTERCONTROLLER pThis)
{
    return OhciDeviceAddRef(iOhciDevice(pOhciMaster(pThis)));
}

PRIVATE UINT OhciMasterRelease (
    PIUSBMASTERCONTROLLER pThis)
{
    return OhciDeviceRelease(iOhciDevice(pOhciMaster(pThis)));
}

/* Submit an isochronous request.
 * Iso requests require ITDs.
 */
PRIVATE SCODE
OhciSubmitIso(POHCI this,
              PUSBM_REQUEST_BLOCK Urb,
              UINT8 PortNumber)
{
    POHCI_ED Ed;
    UINT32 Value32;
    PMYOHCI_ITD Itd, terminal;
    PHCI_STATE Hs;
    UINT32 MsecsFromNow = Urb->Interval;

    /* We'll need a ITD, allocate it now before we commit the b/w
     */
    Itd = ItdAlloc(this);
    if (Itd == NULL)
        goto NoMemory;

    /* Optimistically, fill it in assuming it will fit in the schedule
     * Initialize the URB's HCI state
     * Leave the SF field empty for now
     */
    Hs = (PHCI_STATE)(&Urb->HciReserved);
    Urb->HciReserved2 = 0; /* one store does it all */
    /* To be adjusted at the end */
    Urb->ActualDataLength = Urb->ExpectedDataLength;

    Itd->Td.Next = 0;

    Value32 = VirtualToPhysical(Urb->Data);
    OhciWriteRegister(&Itd->Td.BufferPage0,Value32);
    OhciWriteRegister16(&Itd->Td.OffsetPsw[0],(UINT16)(Value32 & 0xfff));

    Value32 = VirtualToPhysical(Urb->Data + Urb->ExpectedDataLength - 1);
    OhciWriteRegister(&Itd->Td.BufferEnd,Value32);

    /* Lookup/allocate an endpoint
     */
    Ed = LookupEd(this,
                  Urb->DeviceAddress,
                  Urb->EndpointAddress,
                  Urb->Flags,
                  Urb->MaximumPacketSize);

    if (Ed == NULL) {
        goto NoMemory;
    }

    /* Remember which ED this Urb went to
     */
    Hs->Ed = Ed;

    /* Remember the Urb so that we can complete it.
     */
    Hs->nFrags++;
    Itd->Urb = Urb;

    /* Check if the Ed already has a terminal node, if so reuse it.
     */
    Value32 = OhciReadRegister(&Ed->Head) & OHCI_POINTER_MASK;
    if (Value32) {
        terminal = (PMYOHCI_ITD)((ADDRESS)PhysicalToVirtual(Value32) - offsetof(MYOHCI_ITD,Td));
    } else {
        terminal = ItdAlloc(this);
        if (terminal == NULL) {
            /* we already groaned */
            goto NoMemory;
        }

        /* This one does not get completed */
        memset(terminal,0,sizeof *terminal); /* sanity */
    }

    OhciWriteRegister(&Itd->Td.Next,VirtualToPhysical(&terminal->Td));

#if _DEBUG
    /* Debug: dump it
     */
    _tprintf(_T("We are i-saying this :"));
    DumpItdList(&Itd->Td,0);
#endif

    /* Set the SF field now
     */
    MsecsFromNow += CurrentFrameNumber(this);
    OhciWriteRegister(&Itd->Td.Caps,
                      OHCI_TDI_ISOCHRONOUS
                      | (MsecsFromNow & OHCI_TDI_SF)
                      | OHCI_NOT_ACCESSED );

    /* Insert the endpoint in the ISO list, if this is the first time
     * Assume that placing is irrelevant, the HC will skip it if necessary.
     */
    if (Ed->Head == 0) {
        Ed->Next = this->IsoHead->Next;
        OhciWriteRegister(&this->IsoHead->Next,VirtualToPhysical(Ed));
    }

    /* Insert the TD in the list
     */
    EdEnqueTd(Ed,&Itd->Td,&terminal->Td);

    return S_OK;
 NoMemory:
    if (Itd)
        ItdFree(this,Itd,FALSE);
    /* Ed is recycled later */
    return E_NOT_ENOUGH_MEMORY;
}

/* Submit a non-isochronous request.
 * Control/Bulk/Interrupt all use general TDs.
 * They only differ in where the ED goes.
 */
PRIVATE SCODE
OhciSubmitNonIso(POHCI this,
                 PUSBM_REQUEST_BLOCK Urb,
                 UINT8 PortNumber)
{
    PMYOHCI_TD HeadTd = NULL, Td = NULL;
    PMYOHCI_TD nTd;
    POHCI_ED Ed = NULL;
    UINT32 Caps, Value32, Toggle, QueueType;
    PHCI_STATE Hs;
    POHCI_ED HeadEd = NULL;
    BOOL EdIsNew = TRUE;

    /* Quick reminder of where the ED will go
     */
#define CONTROL_Q   1
#define BULK_Q      2
#define INTERRUPT_Q 3
    if (Urb->Flags & URB_FLAG_CONTROL)
        QueueType = CONTROL_Q;
    else if (Urb->Flags & URB_FLAG_INTERRUPT)
        QueueType = INTERRUPT_Q;
    else
        QueueType = BULK_Q;

    /* Find the ED where this TD should be posted
     * This might allocate a new ED if we aint seen it before
     */
    Ed = LookupEd(this,
                  Urb->DeviceAddress,
                  Urb->EndpointAddress,
                  Urb->Flags,
                  Urb->MaximumPacketSize);
    if (Ed == NULL) {
        DBGME(3,_tprintf(_T("Ohci@%x: ?Ed for %x.%x.%x\n"),
                         (ADDRESS) this,
                         Urb->DeviceAddress,
                         Urb->EndpointAddress,
                         Urb->Flags));
        goto ErrorHandler;
    }

    /* Initialize the URB's HCI state
     */
    Hs = (PHCI_STATE)(&Urb->HciReserved);
    Hs->Ed = Ed;
    Urb->HciReserved2 = 0; /* one store does it all */
    /* To be adjusted at the end of the transfer */
    Urb->ActualDataLength = Urb->ExpectedDataLength;

    /* Check if the Ed already has a terminal node, if so reuse it.
     * Note that this also tells us if the node was ever used before.
     */
    Value32 = OhciReadRegister(&Ed->Head) & OHCI_POINTER_MASK;

    if (Value32) {
        EdIsNew = FALSE;
        HeadTd = (PMYOHCI_TD)((ADDRESS)PhysicalToVirtual(Value32) - offsetof(MYOHCI_TD,Td));

    } else {

        /* First time we use this Ed, link it into the proper Q
         * BUGBUG Should we mark it as idle? it does have head=tail(=0)
         * Unless its there already [BUGBUG use this instead of init? ]
         */
        if (Ed != this->ControlHead) {
            if (QueueType != INTERRUPT_Q) {
                volatile UINT32 *QueuePtr;

                if (QueueType == CONTROL_Q) {
                    QueuePtr = &this->Ops->ControlHeadED;
                } else {
                    assert(QueueType == BULK_Q);
                    QueuePtr = &this->Ops->BulkHeadED;
                }
                    
                Value32 = OhciReadRegister(QueuePtr);
                OhciWriteRegister(&Ed->Next,Value32);
                OhciWriteRegister(QueuePtr,VirtualToPhysical(Ed));

            } else {
                /* For the interrupt case, the first time through make a
                 * reservation for the necessary amount of b/w.
                 * This reservation will only be cancelled when the device
                 * is unplugged.  Notice that this happens only when the
                 * endpoint is actually used for the first time, so we are
                 * a bit conservative maybe but not wasteful.
                 * Notice also that we'll overcommit the b/w if necessary.
                 */
                UINT16 MaxBw, MinBw;
                UINT16 FrequencyInMsecs;
                PUSKNODE Node;
                
                /* See how many clocks this will take
                 */
                OhciComputeBandwidth(Urb->Flags,
                                     Urb->MaximumPacketSize,
                                     &MaxBw,
                                     &MinBw);

                /* Check the desired polling frequency
                 * "0xff" means as slow as possible, which is 32msecs for us.
                 */
                FrequencyInMsecs = Urb->Interval;
                if (FrequencyInMsecs == 255)
                    FrequencyInMsecs = 32;
                /* Find a node in the schedule with enough bandwidth
                 */
                Node = UskMakeReservation(this->Schedule,
                                          FrequencyInMsecs,
                                          MinBw,
                                          MaxBw);

                if (Node == NULL) {
                    return E_NOT_RESERVED;
                }

                /* Handle linking issues: our ED goes into the list that
                 * starts at the reservation node, and we remember that node
                 * for when we'll have to cancel the reservation (on unplug).
                 */
                HeadEd = &Node->HcNode;
                ((PMYOHCI_ED)Ed)->InterruptHead = Node;
            }
        }

        /* Allocate a terminal node for it
         */
        HeadTd = TdAlloc(this);
        if (HeadTd == NULL) {
            /* we already groaned */
            return E_NOT_ENOUGH_MEMORY;
        }
    }

    /* Remember the Urb so that we can complete it.
     */
    Hs->nFrags++;
    HeadTd->Urb = Urb;
    HeadTd->Td.Next = 0;

    /* Set the Caps info
     */
    Toggle = OHCI_TD_T_VALID | OHCI_TD_DATA0;
    Caps = OHCI_NOT_ACCESSED;

    if (Urb->Pid == USB_PID_SETUP) {
        Caps |= OHCI_TD_SETUP; /* a zero */
    } else if (Urb->Pid == USB_PID_IN)
        Caps |= OHCI_TD_IN;
    else if (Urb->Pid == USB_PID_OUT)
        Caps |= OHCI_TD_OUT;

    OhciWriteRegister(&HeadTd->Td.Caps,Caps|Toggle);

    if (Urb->Pid == USB_PID_SETUP) {
        Caps |= (Urb->SetupData[0] & USB_RT_DTH) ? OHCI_TD_IN:OHCI_TD_OUT;
    }

    /* If there is setup data involved set their buffer pointers
     */
    if (Urb->SetupLength) {

        Value32 = VirtualToPhysical(Urb->SetupData);
        OhciWriteRegister(&HeadTd->Td.CurrentBufferPointer,Value32);
        Value32 = VirtualToPhysical(Urb->SetupData + Urb->SetupLength - 1);
        OhciWriteRegister(&HeadTd->Td.BufferEnd,Value32);

        /* Allocate another descriptor for the status final xfer,
         * or for the first data portion, whichever comes first.
         * Link it onto the head of the xfer list.
         */
        Td = TdAlloc(this);
        if (Td == NULL) {
            /* we already groaned */
            goto ErrorHandler;
        }

        Hs->nFrags++;
        Td->Urb = Urb;
        OhciWriteRegister(&HeadTd->Td.Next,VirtualToPhysical(&Td->Td));

    } else {
        /* Otherwise there is no setup phase and the toggle comes from the ED
         */
        Td = HeadTd;
        Toggle = 0;/* clearing the valid bit is enough */
    }

    /* Data portion?
     */
    if (Urb->ExpectedDataLength) {
        Toggle ^= OHCI_TD_DATA1;/* flip*/

        /* allow short for last portion */
        OhciWriteRegister(&Td->Td.Caps,
                          Caps
                          |(Toggle & OHCI_TD_T) | OHCI_TD_R/*last only!*/
#if 1
                          | (OHCI_TD_SF & Urb->ExpectedDataLength)
#endif
                          );
        Value32 = VirtualToPhysical(Urb->Data);
        OhciWriteRegister(&Td->Td.CurrentBufferPointer,Value32);
        Value32 = VirtualToPhysical(Urb->Data + Urb->ExpectedDataLength - 1);
        OhciWriteRegister(&Td->Td.BufferEnd,Value32);

        /* For bulk/interrupt xfers no status
         */
        if (QueueType != CONTROL_Q)
            goto AddTerminal;

        /* Get a Td for status
         */
        nTd = TdAlloc(this);
        if (nTd == NULL) {
            goto ErrorHandler;
        }

        Hs->nFrags++;
        nTd->Urb = Urb;
        OhciWriteRegister(&Td->Td.Next, VirtualToPhysical(nTd));

        Td = nTd;

        /* flip direction for the status phase
         */
        Caps ^= OHCI_TD_DP;

    } else {
        Caps &= ~OHCI_TD_DP;
        Caps |= OHCI_TD_IN;
    }

    /* Fill in the status phase Td (ctl only)
     */
    OhciWriteRegister(&Td->Td.Caps,
                      Caps | OHCI_TD_DATA1 | OHCI_TD_T_VALID);
    Td->Td.CurrentBufferPointer = 0;
    Td->Td.BufferEnd = 0;
    
    /* But we are not done yet.
     * The HC will stop processing TDs when the Tail==Head in the
     * endpoint descriptor, but without processing the last one!!
     * So we need to add one more terminal node to the list.
     * Yes, this is pathetic.
     */
  AddTerminal:
    nTd = TdAlloc(this);
    if (nTd == NULL) {
        /* we already groaned */
        goto ErrorHandler;
    }

    /* This one does not get completed */
    memset(nTd,0,sizeof *nTd); /* sanity */
    OhciWriteRegister(&Td->Td.Next,VirtualToPhysical(&nTd->Td));

#if _DEBUG
    /* Debug: dump it
     */
    DBGME(0,_tprintf(_T("We are saying this to ED %x[%x]: "),
                     (ADDRESS)Ed, OhciReadRegister(&Ed->Head)));
    DBGME(0,DumpTdList(&HeadTd->Td,0));
#endif

    /* Now queue it up onto the ED
     * This function also marks the ED not-idle
     */
    EdEnqueTd(Ed,&HeadTd->Td,&nTd->Td);

    /* Signal the HC to rescan the queue
     */
    if (QueueType == CONTROL_Q) {
        OhciWriteRegister(&this->Ops->CommandStatus,OHCI_CS_CLF);
    } else if (QueueType == BULK_Q) {
        OhciWriteRegister(&this->Ops->CommandStatus,OHCI_CS_BLF);
    } else if (QueueType == INTERRUPT_Q) {
        /* Link it in, unless its already there */
        if (EdIsNew && (HeadEd != NULL/*sanity*/)) {
            Ed->Next = HeadEd->Next;
            OhciWriteRegister(&HeadEd->Next,VirtualToPhysical(Ed));
        }
    }

    return S_OK;
 ErrorHandler:
    if (HeadTd)
        TdFree(this,HeadTd,FALSE);
    return E_FAIL;
}

PRIVATE SCODE
OhciRootHub(POHCI this,
  /*inout*/ PUSBM_REQUEST_BLOCK Urb)
{
    POHCI_OPS Ops = this->Ops;
    UINT16 Status;
    UINT16 Changes;
    UINT8 Port;
    BOOL Clear;
    UINT32 Value32;

    /* Is it an EVENT one
     */
    if (Urb->Pid == USB_PID_EVENT) {

        DBGME(0,_tprintf(_T("Ohci:Event %x\n"),Urb->Data[0]));
        if (Urb->Data[0] == USBM_EVENT_DISCONNECTED) {
            DestroyEdsForFunction(this,Urb->Data[1]);
        }
        goto DoneOk;
    }

    /* GetPortStatusChanges() ?
     */
    if ((Urb->Pid == USB_PID_IN) &&
        (Urb->Flags & URB_FLAG_INTERRUPT)) {
    
        this->EventsPacket = Urb;
        Changes = this->PortStatusChanges;
        Urb->ActualDataLength = 2;
        Set2(Urb->Data,Changes);
        if (Changes) {
            this->EventsPacket = NULL;
            Urb->Flags |= URB_FLAG_COMPLETE;
            Condition_InterruptSignal(Urb->Done);
        }
        return S_OK;
    }

    /* GetPortStatus() ?
     */
    if ((Urb->Pid == USB_PID_SETUP) &&
        (Urb->Flags & URB_FLAG_IN) &&
        (Urb->SetupData[1] == USB_R_GET_STATUS)) {

        /* Our ports start at 0
         */
        Port = Urb->SetupData[4] - 1;
        if (Port >= this->nPorts)
            return E_INVALID_PARAMETER;

        /* Map our bit encodings to USB2 (easy:-)
         */
#define OhciMapPc(v) (UINT16)((v&OHCI_PS_CHANGE) >> 16) /*->CONN,ENAB,SUSP,OVERC,RESET*/
#define OhciMapPs(v) (UINT16)(v & 0x31f) /*CCS,PES,PSS,POCI,PRS,PPS,LSDA*/
        Value32 = OhciReadRegister(&Ops->RhPortStatus[Port]);
        Changes = OhciMapPc( Value32 );
        Status  = OhciMapPs( Value32 );

        Urb->ActualDataLength = 4;
        Set2(Urb->Data,Status);
        Set2(Urb->Data+2,Changes);
        Condition_InterruptSignal(Urb->Done);
        goto DoneOk;
    }

    /* SetClearPortFeature() ?
     */
    if ((Urb->Pid == USB_PID_SETUP) &&
        (Urb->Flags & URB_FLAG_CONTROL)) {

        /* Our ports start at 0
         */
        Port = Urb->SetupData[4] - 1;
        if (Port >= this->nPorts)
            return E_INVALID_PARAMETER;

        /* Which feature?
         */
        Changes = Get2(Urb->SetupData+2);
        Clear = (Urb->SetupData[1] == USB_R_CLEAR_FEATURE);

        switch (Changes) {

        case USB_FS_C_PORT_RESET:
            if (Clear) {
                Value32 = OHCI_PS_PRSC;
                goto ClearIt;
            }
            break;

        case USB_FS_C_PORT_ENABLE:
            if (Clear) {
                Value32 = OHCI_PS_PESC;
                goto ClearIt;
            }
            break;

        case USB_FS_C_PORT_OVERCURRENT:
            if (Clear) {
                Value32 = OHCI_PS_OCIC;
                goto ClearIt;
            }
            break;

        case USB_FS_PORT_ENABLE:
            if (Clear) {
                Value32 = OHCI_PS_CPE;
                goto ClearIt;
            }
            break;

        case USB_FS_PORT_POWER:
            if (Clear) {
                Value32 = OHCI_PS_CPP;
                goto ClearIt;
            }
            break;

        case USB_FS_PORT_RESET:
            if (!Clear) {
                OhciWriteRegister(&Ops->RhPortStatus[Port],OHCI_PS_SPR);
            }
            break;

        case USB_FS_C_PORT_CONNECTION:
            if (Clear) {
                Value32 = OHCI_PS_CSC;
          ClearIt:
                /* Clear the change bit
                 */
                OhciWriteRegister(&Ops->RhPortStatus[Port],Value32);

                /* If there is nothing else..
                 */
                Value32 = OhciReadRegister(&Ops->RhPortStatus[Port]);
                Value32 &= OHCI_PS_CHANGE;
                if (Value32 == 0)
                    this->PortStatusChanges &= ~(2 << Port);/*p0->p1*/
            }
            break;
        }

        Condition_InterruptSignal(Urb->Done);
        goto DoneOk;
    }

    /* Uhu?
     */
    return E_NOT_IMPLEMENTED;

 DoneOk:
    Urb->Flags |= URB_FLAG_COMPLETE;
    return S_OK;
}

PRIVATE SCODE OhciMasterSubmitUrb (
    PIUSBMASTERCONTROLLER pThis,
    /*inout*/ PUSBM_REQUEST_BLOCK Urb,
    UINT8 PortNumber)
{
    POHCI this = pOhciMaster(pThis);

    DBGME(0,_tprintf(_T("Ohci:Submit(%x,%x,%x)\n"),
                     (ADDRESS)this, (ADDRESS)Urb, PortNumber));
    DBGME(0,DumpPacket(Urb));
    
    /* Is this for the root hub?
     */
    if (Urb->DeviceAddress == 0xff)
        return OhciRootHub(this,Urb);

    /* Post a TD
     * Isochronous endpoints require different TDs.
     */
    if (Urb->Flags & URB_FLAG_ISOCHRONOUS)
        return OhciSubmitIso(this,Urb,PortNumber);
    else
        return OhciSubmitNonIso(this,Urb,PortNumber);
}

PRIVATE SCODE OhciMasterCancelUrb (
    PIUSBMASTERCONTROLLER pThis,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 PortNumber)
{
    POHCI this = pOhciMaster(pThis);
    ADDRESS *Target;

    DBGME(0,_tprintf(_T("OhciCancel(%x,%x,%d)\n"),
                     (ADDRESS)this,(ADDRESS)Urb,PortNumber));

    if (Urb == this->EventsPacket)
        Target = (ADDRESS*) &this->EventsPacket;
    else {
        PHCI_STATE Hs;

        Hs = (PHCI_STATE)(&Urb->HciReserved);
        if (Hs->Ed == NULL)
            return E_OBJECT_NOT_FOUND;

        /* BUGBUG writemorecode: halt the ED, find (how?) the TD
         * that is for this URB, remove it, restart the ED.
         */
        Hs->Ed = NULL;
        goto Hack;
    }

    if (AtomicCmpAndSwap(Target,(ADDRESS)Urb,0)) {
    Hack:
        Urb->Flags |= URB_FLAG_ERROR;
        DBGME(1,DumpPacket(Urb));
        Condition_InterruptSignal(Urb->Done);
        return S_OK;
    }

    return E_OBJECT_NOT_FOUND;
}

#include "s_Ohci_v.c"

/* Constructor */
PIDEVICE OhciNew( ADDRESS PciLocation )
{
    PIHEAP pHeap = CurrentHeap();
    UINT Bus, Device, Function;
    PIBUSDRIVER OurPci = NULL;
    SCODE sc;
    UINT32 Value32;
    UINT16 Value16;
    POHCI this = NULL;
    POHCI_OPS Ops = NULL;

    Bus = PciLocation >> 16;
    Device = (PciLocation >> 8) & 0xff;
    Function = PciLocation & 0xff;

    DBGME(1,_tprintf(_T("Enter OhciNew(%d,%d,%d)\n"),
                     Bus,Device,Function));

    /* Make sure we know where to go
     */
    sc = BindToObject(CurrentNameSpace(),
                      PCIDRIVER_NAME,
                      NAME_SPACE_READ | NAME_SPACE_WRITE,
                      &IID_IBusDriver,
                      (void**)&OurPci);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("No PCI access routine!!\n")));
        goto ErrorExit;
    }

    /* Look at PCI space, check its there
     */
    Value32 = 0;
    OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                      0, &Value32, sizeof(Value32));
    if (Value32 != PCIID_NEC_UPD720101) {
        /* Check if it is compatible.
         * The specs mandate CLASSC
         */
        OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                          8, &Value32, sizeof(Value32));

        if (((Value32 & 0xffffff0) != OHCI_CLASSC)) {
            DBGME(3,_tprintf(_T("Signature at PCI %x not for OHCI.\n"),
                       PciLocation));
            goto ErrorExit;
        }
    }

    /* Look at the assigned I/O location, make sure it was assigned
     */
    Value32 = 0;
    OurPci->v->ReadConfigurationSpace(OurPci,Bus,Device,Function,
                                      16, &Value32, sizeof(Value32));
    Ops = (POHCI_OPS) IoPhysicalToIoVirtual(Value32);
    if (Ops == NULL) {
        DBGME(3,_tprintf(_T("OHCI at PCI %x has no BaseAddress[0].\n"),
                       PciLocation));
        goto ErrorExit;
    }

    /* Enable it, by writing to the Command register
     * No: back-to-back,SERR#, wait cycle, parity, palette snoop, I/O
     * Yes: MWI, master, memory
     */
    Value16 = 0x0016;
    OurPci->v->WriteConfigurationSpace(OurPci,Bus,Device,Function,
                                      4, &Value16, sizeof(Value16));
    Delay(40);

    /* Quick sanity check for revision register
     */
    Value32 = OhciReadRegister(&Ops->Revision);
    if ((Value32 & OHCI_REV_MASK) != OHCI_REV_1p0) {
        DBGME(3,_tprintf(_T("OHCI(%x) at PCI %x has bad revision(%x)\n"),
                         (ADDRESS)Ops, PciLocation, Value32));
        goto ErrorExit;
    }

    /* Allocate all our resources
     */
    this = (POHCI) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *this, 0);
    if (!this) {
        DBGME(3,_tprintf(_T("OHCI at PCI %x outtamem.\n"), PciLocation));
        goto ErrorExit;
    }

#define bsize (sizeof(OHCI_HCCA) + (3 * sizeof(OHCI_ED)))
    this->HCCA = pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY,
                                 bsize, OHCI_HCCA_ALIGNMENT+1);
#undef bsize
    if (this->HCCA == NULL) {
        DBGME(3,_tprintf(_T("OHCI at PCI %x outtamem.\n"), PciLocation));
        goto ErrorExit;
    }

    this->IsoHead = (POHCI_ED) (this->HCCA + 1);
    this->ControlHead = this->IsoHead + 1;
    this->BulkHead = this->ControlHead + 1;

    /* Allocate the schedule for interrupt endpoints
     */
    this->Schedule = UskCreateSchedule( (void **) this->HCCA);
    if (this->Schedule == NULL)
        goto ErrorExit;

    /* The isochronous list head must be...
     */
    {
        POHCI_ED OneMsInt = (POHCI_ED) UskMakeReservation(this->Schedule,
                                                          1,
                                                          0,
                                                          0);
        assert(OneMsInt != NULL);

        OhciWriteRegister(&OneMsInt->Next,VirtualToPhysical(this->IsoHead));
    }

    /* Got resources, we are committed
     */
    this->CDevice.v = (struct CDeviceVtbl*)&OhciDeviceOhciVtbl;
    this->CDevice.RefCnt = 1;
    this->OhciMasterv = &OhciMasterOhciVtbl;
    this->Ops = Ops;
    Mutex_Init(&this->HashTableLock);

    /* Cache some values
     */
    Value32 = OhciReadRegister(&Ops->RhDescriptorA);
    this->nPorts = (UINT8) (Value32 & OHCI_RH_NDP);
    this->PortPowerControl = (Value32 & OHCI_RH_NPS) ? FALSE : TRUE;

    //DumpOhci(this);

    /* Anything else in DeviceInitialize
     */
    return iOhciDevice(this);

 ErrorExit:
    if (Ops) {
        Value16 = 0x0000;
        OurPci->v->WriteConfigurationSpace(OurPci,Bus,Device,Function,
                                           4, &Value16, sizeof(Value16));
    }
    if (OurPci)
        OurPci->v->Release(OurPci);
    if (this) {
        if (this->Schedule)
            pHeap->v->Free(pHeap,0,this->Schedule);
        pHeap->v->Free(pHeap,0,this);
    }
    return NULL;

}

/* ** 
*/
SCODE __cdecl OhciConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("OhciConstructor(%x %s %x %x %d)\n"),
                     (ADDRESS) pThis, Name, (ADDRESS) ppUnk,
                     RegisterBase, Interrupt));

    pUnk = (PIUNKNOWN) OhciNew( RegisterBase );

    DBGME(0,_tprintf(_T("OhciConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("OhciConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl OhciDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) OhciConstructor
};

static struct _DRIVERCOB OhciDriverCobObject = { &OhciDriverCobVtable, 1 };

PIUNKNOWN OhciCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &OhciDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &OhciDeviceOhciVtbl, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &OhciDriverCobObject;
}
