/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usb_sched.c
 *
 * Purpose:
 *
 *    USB bus scheduling code, shared by various HCI masters.
 *
 * Reference: "OpenHCI Open Host Controller Interface Specification for USB"
 *            Rev 1.0a, September 1999.
 *
 * Reference: "Enhanced Host Controller Interface Specification
 *             for Universal Serial Bus"  Rev 1.0, March 2002.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

/* This file is included rather than separately compiled.
 * The idea is to be able to share it among HCIs, even though
 * they all require differently shaped nodes.
 *
 * Before including this file, the following must be defined:
 * 1- HC_NODE
 *    a typedef-ed structure defining the fields known to the HC
 * 2- HC_NODE_ALIGNMENT
 *    a 2^N constant defining the HC's alignment requirements for the above
 * 3- HC_ROOTS
 *    a 2^N constant defining the number of root pointers known to the hardware
 * 4- HC_ROOTS_LOG2
 *    log2(HC_ROOTS)
 * 5- HC_BUS_BANDWIDTH
 *    a constant, ether USB_11_BUS_BANDWIDTH or USB_20_BUS_BANDWIDTH
 * 6- HC_PHYS_LINK(HC_NODE _node_, UINT32 _physical_pointer_)
 *    a macro to set the forward link pointer known to the HC.
 * 7- HC_INIT_NODE(HC_NODE _node_)
 *    a macro to initialize any other non-zero field (except the link!)
 */

typedef struct _USKNODE {
    HC_NODE HcNode;
    UINT16 LeftParentIndex;
    UINT16 RightParentIndex;
    UINT16 MinimumBandwidth;
    UINT16 MaximumBandwidth;
    UINT8  Pad[(2*HC_NODE_ALIGNMENT)-sizeof(HC_NODE)-(4*sizeof(UINT16))];
} USKNODE, *PUSKNODE;

typedef struct _USKEDULE {
    UINT16 NodeCount;

    /* The nodes follow, after proper alignment
     */
} USKEDULE, *PUSKEDULE;

/* These are the exported functions
 */
PUSKNODE
UskMakeReservation(
    PUSKEDULE Schedule,
    UINT16 FrequencyInMsecs,
    UINT16 MinimumBandwidth,
    UINT16 MaximumBandwidth
    );

void
UskCancelReservation(
    PUSKEDULE Schedule,
    PUSKNODE Node,
    UINT16 MinimumBandwidth,
    UINT16 MaximumBandwidth
    );

PUSKEDULE
UskCreateSchedule(
    void **RootsList
    );

/*
 * Work function for propagating bandwidth changes, positive or negative.
 */
PRIVATE
void
UskPropagateChange(
    PUSKEDULE Schedule,
    int NodeIndex,
    int MinimumBandwidthChange,
    int MaximumBandwidthChange
    )
{
    PUSKNODE Node;

    /* Find the node's base, adjust and find our info
     */
    Node = ((PUSKNODE) Schedule) + NodeIndex;

    /* Adjust this one node
     */
    Node->MinimumBandwidth += MinimumBandwidthChange;
    Node->MaximumBandwidth += MaximumBandwidthChange;

    /* Recurse for parents
     */
    if (Node->RightParentIndex != 0)
        UskPropagateChange(Schedule,
                           Node->RightParentIndex,
                           MinimumBandwidthChange,
                           MaximumBandwidthChange);
    if (Node->LeftParentIndex != 0)
        UskPropagateChange(Schedule,
                           Node->LeftParentIndex,
                           MinimumBandwidthChange,
                           MaximumBandwidthChange);
}

/* 
 * Reserve a certain amounf of b/w in the schedule
 * The minimum b/w is when an interrupt transfer receives a NAK.
 * The maximum b/w is when one full packet is actually transferred.
 */
PUSKNODE
UskMakeReservation(
    PUSKEDULE Schedule,
    UINT16 FrequencyInMsecs,
    UINT16 MinimumBandwidth,
    UINT16 MaximumBandwidth
    )
{
    PUSKNODE Node;
    int FirstNodeIndex, LastSiblingIndex;
    int AvailableBandwidth, i;

    DBGME(0,_tprintf(_T("UskMakeReservation(%x,%d,%d,%d)\n"),
                     (ADDRESS) Schedule, FrequencyInMsecs,
                     MinimumBandwidth, MaximumBandwidth));

    /* Poor man's 2^x
     */
    i = FrequencyInMsecs;
    for (FirstNodeIndex = 1; i > 1; i >>= 1)
        FirstNodeIndex <<= 1;
    LastSiblingIndex = (FirstNodeIndex << 1) - 1;

    /* Sanity check, e.g. validate frequency
     */
    if (LastSiblingIndex > Schedule->NodeCount)
        goto Impossible;

    /* See if we can fit it with guaranted b/w
     */
    Node = ((PUSKNODE)Schedule) + FirstNodeIndex;

    for (i = FirstNodeIndex; i <= LastSiblingIndex; i++) {

        /* Do we have any guaranteed bandwidth left ?
         * And is it enough.
         */
        AvailableBandwidth = HC_BUS_BANDWIDTH - Node->MaximumBandwidth;
        if (AvailableBandwidth > MaximumBandwidth)
            goto FoundSpot;
        Node++;
    }

    /* Try again with a best-effort b/w
     */
    Node = ((PUSKNODE)Schedule) + FirstNodeIndex;

    for (i = FirstNodeIndex; i <= LastSiblingIndex; i++) {

        /* Do we have enough best-effort bandwidth left ?
         */
        AvailableBandwidth = HC_BUS_BANDWIDTH - Node->MinimumBandwidth;
        if (AvailableBandwidth > MinimumBandwidth)
            goto FoundSpot;
        Node++;
    }

    /* Ooops, doesnt fit at all
     */
 Impossible:
    DBGME(3,_tprintf(_T("UskMakeReservation: no b/w %d,%d,%d\n"),
                     FrequencyInMsecs, MinimumBandwidth, MaximumBandwidth));
    return NULL;

 FoundSpot:
    /* Make the bandwidth adjustments to this node and all its parents
     */
    UskPropagateChange(Schedule,
                       i,
                       0+MinimumBandwidth,
                       0+MaximumBandwidth);

    DBGME(0,_tprintf(_T("UskMakeReservation() -> %x @Schedule[%d]\n"),
                     (ADDRESS)Node, i));
    return Node;
}

/*
 * Drop a slot from the schedule, presumably at the end of a transfer
 */
void
UskCancelReservation(
    PUSKEDULE Schedule,
    PUSKNODE Node,
    UINT16 MinimumBandwidth,
    UINT16 MaximumBandwidth
    )
{
    DBGME(0,_tprintf(_T("UskCancelReservaion(%x,%x,%d,%d)\n"),
                     (ADDRESS) Schedule, (ADDRESS) Node,
                     MinimumBandwidth, MaximumBandwidth));

    UskPropagateChange(Schedule,
                       Node - ((PUSKNODE)Schedule),
                       0-MinimumBandwidth,
                       0-MaximumBandwidth);
}
    
/*
 * Work function for computing root pointer indexes
 */
PRIVATE INLINE
UINT
UskBitInvert(
    UINT n,
    UINT LogBits
)
{
    UINT r = 0;
    UINT i;

    for (i = 0; i < LogBits; i++)
        r |= ((n>>i)&1) << (LogBits-1-i);
    return r;
}

/*
 * Build the initial schedule
 */
PUSKEDULE
UskCreateSchedule(
    void **RootsList
    )
{
    UINT16 NumberOfNodes;
    PIHEAP pHeap = CurrentHeap();
    PUSKEDULE Schedule;
    PUSKNODE Node;
    int i, j;

    DBGME(0,_tprintf(_T("UskCreateSchedule(%d,%d,%d,%d,%x)\n"),
                     (int)sizeof(HC_NODE),HC_NODE_ALIGNMENT,
                     HC_ROOTS,HC_BUS_BANDWIDTH,(ADDRESS)RootsList));

    /* We *are* initializing the roots
     */
    assert( RootsList != NULL );

    /* Must be pointer-aligned
     */
    assert( (sizeof(HC_NODE) & (sizeof(ADDRESS)-1)) == 0);

    /* Must be 2^N and at least 2
     */
    assert(HC_ROOTS >= 2);
    assert( ((HC_ROOTS & (HC_ROOTS-1)) == 0) &&
            ((HC_ROOTS & ~(HC_ROOTS-1)) == HC_ROOTS) );

    /* The extra node at front contains the global schedule data.
     */
    assert( sizeof(USKEDULE) <= sizeof(USKNODE) );

    /* The node size must obey the alignment requirements
     */
    assert( ((sizeof(USKNODE) / HC_NODE_ALIGNMENT) * HC_NODE_ALIGNMENT) ==
            sizeof(USKNODE) );

    /* How many nodes in our schedule, which we allocate at once
     */
    NumberOfNodes = ((HC_ROOTS * 2) - 1) + 1;

    /* Allocate the schedule array
     */
    Schedule = (PUSKEDULE) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY,
                                           NumberOfNodes * sizeof(USKNODE),
                                           HC_NODE_ALIGNMENT);

    if (Schedule == NULL) {
        DBGME(3,_tprintf(_T("UskCreateSchedule: no mem %d,%d\n"),
                         (int)(NumberOfNodes * sizeof(USKNODE)),
                         HC_NODE_ALIGNMENT));
        return NULL;
    }

    /* Initialize global info
     */
    Schedule->NodeCount = NumberOfNodes - 1;

    /* Link all nodes into a tree, as per specs.
     * 
     * Use indeces for son-of relationships.
     * Roots have 0 (invalid) as parent indexes.
     * 
     * Use HC pointers for parent-of relationships.
     */
    j = 2;
    Node = (PUSKNODE)Schedule;
    for (i = 1; i < HC_ROOTS; i++) {
        UINT32 Value32 = VirtualToPhysical((Node + i));

        HC_INIT_NODE(Node[i].HcNode);

        Node[i].LeftParentIndex = j;
        HC_PHYS_LINK(Node[j++].HcNode,Value32);

        Node[i].RightParentIndex = j;
        HC_PHYS_LINK(Node[j++].HcNode,Value32);
    }

    /* Note that it is *not* a 1:1 mapping
     */
    Node += i;
    assert( ((PUINT8)Node) == ((PUINT8)Schedule) + (HC_ROOTS*sizeof(USKNODE)));
    for (i = 0; i < HC_ROOTS; i++) {
        HC_INIT_NODE(Node->HcNode);
        RootsList[UskBitInvert(i,HC_ROOTS_LOG2)] = Node;
        Node++;
    }

    /* We are done
     */
    DBGME(0,_tprintf(_T("UskCreateSchedule() -> %x\n"), (ADDRESS)Schedule));
    return Schedule;
}
