/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    sl811.c
 *
 * Purpose:
 *
 *    Device driver for Cypress's SL811HS USB master.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <stddef.h> /* for offsetof() */
#include <tchar.h>
//#define _DEBUG 1
#define DBGLVL 2
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbmaster.h>
#define URB_FLAG_PERIODIC (URB_FLAG_INTERRUPT|URB_FLAG_ISOCHRONOUS)

#include "sl811.h"

/* Support for isochronous and interrupt transactions requires a schedule.
 * Proper support for bulk transactions also require a schedule and a place
 * to keep the endpoint's state across DATA0/DATA1 transactions.
 */
#define MAX_POLLING_FREQUENCY 16
#define USK_LEVEL2_MAX 1
#define USK_LEVEL1_MAX MAX_POLLING_FREQUENCY
#define USK_BW_PER_FRAME   USB_11_BUS_BANDWIDTH
#define HC_NULL_POINTER    0
typedef ADDRESS HC_POINTER_TYPE;
#define HcSetPointerValue(_pointer_,_value_) *(_pointer_) = _value_

#include "usb_sched.c"


/* Endpoint descriptor
 */
typedef struct _SL_ED {
    ADDRESS             Next;  /* Same type as in PeriodicList */
#define SL_POINTER_TYPE_MASK 0x00000003
#define SL_POINTER_MASK      (~SL_POINTER_TYPE_MASK)
# define SL_POINTER_TYPE_NONP 0
# define SL_POINTER_TYPE_INT  1
# define SL_POINTER_TYPE_ISO  2
    PUSBM_REQUEST_BLOCK CurrentUrb;
    PUSBM_REQUEST_BLOCK NextUrb;
    UINT32              EndpointCaps;
#define SL_ED_DEVICE_MASK   0x0000007f /* We waste 4 bits to ease debug */
#define SL_ED_ENDPT_MASK    0x00000f00 /*1*/
# define SL_ED_ENDPT_MASK0  0x00000e00
# define SL_ED_ENDPT_MASK1  0x00000100
# define SL_ED_ENDPT_SHIFT  8
#define SL_ED_LOW_SPEED     0x00001000
#define SL_ED_FULL_SPEED    0x00002000 /*2*/
#define SL_ED_ISOCHRONOUS   0x00004000
#define SL_ED_INTERRUPT     0x00008000 /*3*/
#define SL_ED_MPS           0x07ff0000
# define SL_ED_MPS_SHIFT    16
#define SL_ED_IN            0x08000000
#define SL_ED_OUT           0x00000000
# define SL_ED_CAPS         0x0fffffff
# define SL_ED_STATE        0xf0000000
#define SL_ED_TOGGLE        0x10000000
#define SL_ED_ACTIVE        0x40000000 /*4*/
#define SL_ED_INVALID       0x80000000
    PUSKNODE            InterruptHead;
} SL_ED, *PSL_ED;

/* Trivial hash table
 */
typedef struct _SL_HT {
#define InitialHashEntries 8  /* 2**n */
    int nEntries;
    struct _SL_HT_ENTRY {
        PTR    DataValue;
        UINT32 HashValue;
    } Entries[1]; /* varsized */
} SL_HT, *PSL_HT;

/* State for our host controller
 */
typedef struct _PSL811 {
    struct CDevice CDevice;
    CONST_VTBL struct IUsbMasterControllerVtbl *Sl811Masterv;

    PSL811REGS Sl;
    UINT16 PortStatus;
    UINT16 PortStatusChanges;

    PUSBM_REQUEST_BLOCK CurrentPacket;
    PUSBM_REQUEST_BLOCK PendingPackets;
    PUSBM_REQUEST_BLOCK EventsPacket;

    UINT IsoPacketsActive;
    UINT BulkPacketsActive;
    PSL_ED CurrentEd;
    PSL_ED NonPeriodicList;
    PUSKEDULE Schedule;
    ADDRESS *PeriodicList;
#define SL_PERIODIC_LIST_SIZE 256
# define SlPeriodicIndex(n) (n & (SL_PERIODIC_LIST_SIZE-1))
    PSL_HT HashTable;
    MUTEX HashTableLock;

    struct _STATS {
        UINT IsrCnt;
        UINT IntrCnt;
        UINT SofCnt;
/* Strictly speaking this should be zeroed at each SlReset() */
#define FrameCounter Stats.SofCnt
        UINT InsCnt;
        UINT StepsCnt;
        UINT SuspCnt;
        UINT DoneCnt;
        UINT PktCnt;
        UINT IsoStarted, IsoDone;
    } Stats;

} *PSL811;

/* Convert between interface and instance pointer and vice versa */
#define pSl811Device(_i_) ((PSL811)((_i_)))
#define pSl811Master(_i_) ((PSL811)((ADDRESS)(_i_) - offsetof(struct _PSL811, Sl811Masterv)))
#define iSl811Device(_p_) ((PIDEVICE)((_p_)))
#define iSl811Master(_p_) ((PIUSBMASTERCONTROLLER)(&((PSL811)(_p_))->Sl811Masterv))

/*
 * Phys functions
 */
PRIVATE INLINE UINT8 SlReadReg(PSL811REGS Sl, UINT8 Regno)
{
    Sl->Address = Regno;
    Delay(1);
    return Sl->Data;
}

PRIVATE INLINE void SlWriteReg(PSL811REGS Sl, UINT8 Regno, UINT8 Value)
{
    Sl->Address = Regno;
    Delay(1);
    Sl->Data = Value;
    Delay(1);
}

PRIVATE INLINE void SlReadSeq(PSL811REGS Sl, UINT8 Regno, UINT Count, UINT8 *Data)
{
    Sl->Address = Regno;
    while (Count-- > 0)
        *Data++ = Sl->Data;
}

PRIVATE INLINE void SlWriteSeq(PSL811REGS Sl, UINT8 Regno, UINT Count, UINT8 *Data)
{
    Sl->Address = Regno;
    while (Count-- > 0)
        Sl->Data = *Data++;
}

/*
 * Scripts are sequential lists of Steps.
 * Each step consists of a start and of an end action.
 * The start action initiates the step, regardless of the chip's state.
 * It can be called multiple times, either as a followon to a previous step
 * or because the script was suspended (device returned a NAK, for instance)
 * or because the script repeats the step multiple times.
 * The end action checks that the command had the intended consequences,
 * that there are no errors and no NAKs, and possibly terminates the script.
 * All actions are invoked at interrupt time directly by the ISR.
 */
typedef int ISR_RESULT;
#define ISR_CONTINUE 0x01
#define ISR_REPEAT   0x02
#define ISR_DONE     0x04
#define ISR_ERROR    0x08
#define ISR_SUSPEND  0x10
typedef ISR_RESULT (*ISR_STARTACTION)(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet, BOOL RootSlow);
typedef ISR_RESULT (*ISR_ENDACTION)(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet);
typedef struct _ISR_STEP {
    ISR_STARTACTION StartAction;
    ISR_ENDACTION EndAction;
} ISR_STEP, *PISR_STEP;

/* Per-packet state 
 */
typedef struct _HCI_STATE {
    PUSBM_REQUEST_BLOCK NextPacket;
    UINT8               NextStep;
    UINT8               StartStep;
    UINT8               Toggle;
    UINT8               OtherInfo;
} HCI_STATE, *PHCI_STATE;

#define StepForPacket(_p_) \
        ((PHCI_STATE)&((_p_)->HciReserved))->NextStep
#define SetStepForPacket(_p_,_a_) \
        ((PHCI_STATE)&((_p_)->HciReserved))->NextStep = _a_
#define StartForPacket(_p_) \
        ((PHCI_STATE)&((_p_)->HciReserved))->StartStep
#define SetStartForPacket(_p_,_a_) \
        ((PHCI_STATE)&((_p_)->HciReserved))->StartStep = _a_

#define ActionStartForPacket(_p_) \
        AllScripts[((PHCI_STATE)&((_p_)->HciReserved))->NextStep].StartAction
#define ActionEndForPacket(_p_) \
        AllScripts[((PHCI_STATE)&((_p_)->HciReserved))->NextStep].EndAction

#define ToggleForPacket(_p_) \
        ((PHCI_STATE)&((_p_)->HciReserved))->Toggle
#define SetToggleForPacket(_p_,_n_) \
        ((PHCI_STATE)&((_p_)->HciReserved))->Toggle = _n_

#define OtherInfoForPacket(_p_) \
        ((PHCI_STATE)&((_p_)->HciReserved))->OtherInfo
#define SetOtherInfoForPacket(_p_,_n_) \
        ((PHCI_STATE)&((_p_)->HciReserved))->OtherInfo = _n_

#define NextForPacket(_p_) \
        ((PHCI_STATE)&((_p_)->HciReserved))->NextPacket
#define SetNextForPacket(_p_,_n_) \
        ((PHCI_STATE)&((_p_)->HciReserved))->NextPacket = _n_

/* Insertion-lookup of an ISO/INTERRUPT endpoint descriptor
 */

PRIVATE INLINE int
EdHash(UINT32 HashValue)
{
    /* Give two consecutive slots to each usb address.
     * One is for control, the other for EP 1.
     */
    return ((HashValue & SL_ED_DEVICE_MASK) << 1) |
            (HashValue & SL_ED_ENDPT_MASK0) |
            ((HashValue & SL_ED_ENDPT_MASK1) >> SL_ED_ENDPT_SHIFT);
}

PRIVATE PSL_ED
LookupEd(PSL811 this,
         PUSBM_REQUEST_BLOCK Urb)
{
    int i, j, HashKey;
    UINT32 HashValue;
    PIHEAP pHeap;
    PSL_HT HashTable, NewHashTable;
    PSL_ED Ed = NULL;
    UINT8 DeviceAddress = Urb->DeviceAddress;
    UINT8 EndpointAddress = Urb->EndpointAddress;
    USBM_URB_FLAGS Flags = Urb->Flags;

    /* Build the hash table key
     * The HashValue uniquely identifies an endpoint.
     * It also contains additional information used at removal time.
     */
    HashValue =
        (DeviceAddress & SL_ED_DEVICE_MASK) |
        ((EndpointAddress<<SL_ED_ENDPT_SHIFT) & SL_ED_ENDPT_MASK) |
        ((Urb->MaximumPacketSize<<SL_ED_MPS_SHIFT) & SL_ED_MPS) |
        ((Flags & URB_FLAG_LOW_SPEED) ? SL_ED_LOW_SPEED : SL_ED_FULL_SPEED) |
        ((Flags & URB_FLAG_ISOCHRONOUS) ? SL_ED_ISOCHRONOUS : 0)|
        ((Flags & URB_FLAG_INTERRUPT) ? SL_ED_INTERRUPT : 0)|
        ((Flags & URB_FLAG_IN) ? SL_ED_IN : SL_ED_OUT);       

    assert(HashValue != 0);
    HashKey = EdHash(HashValue);

    /* Lock the table
     */
    Mutex_Lock(&this->HashTableLock);

    HashTable = this->HashTable;
    if (HashTable == NULL)
    {
        pHeap = CurrentHeap();

        HashTable = (PSL_HT) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY,
                                               sizeof(*HashTable) +
                                               (InitialHashEntries-1) * sizeof(struct _SL_HT_ENTRY),
                                               0);
        if (!HashTable)
        {
            DBGME(3,_tprintf(_T("SL(%x) HT outtamem.\n"), (ADDRESS) this));
            goto UnlockAndDone;
        }
        HashTable->nEntries = InitialHashEntries;
        this->HashTable = HashTable;
    }

    /* Quick hit? */
    i = HashKey & (HashTable->nEntries-1);

    /* Scan forward the whole table to make sure we did not clash
     * or that this slot did not just free up
     */
    for (j = 0; j < HashTable->nEntries; j++)
    {
        if (HashTable->Entries[i].HashValue == HashValue)
        {
            Ed = HashTable->Entries[i].DataValue;
            goto UnlockAndDone;
        }
        if (++i == HashTable->nEntries)
            i = 0;
    }

    /* Ok, we got to allocate one
     */
#define EdAlloc(t) malloc(sizeof(SL_ED))
#define EdFree(t,e) free(e)
    Ed = EdAlloc(this);
    if (Ed == NULL)
    {
        DBGME(3,_tprintf(_T("SL(%x) ED outtamem.\n"), (ADDRESS) this));
        goto UnlockAndDone;
    }

    /* Mark it as freshly allocated
     */
    memset(Ed,0,sizeof *Ed);
    Ed->Next = HC_NULL_POINTER;
    Ed->EndpointCaps = SL_ED_INVALID | HashValue;

    /* Insert it in the hash table
     * Scan forward till we find a slot
     */
 Again:
    i = HashKey & (HashTable->nEntries-1);
    for (HashKey = 0; HashKey < HashTable->nEntries; HashKey++) {
        if (HashTable->Entries[i].HashValue == 0) {
            HashTable->Entries[i].HashValue = HashValue;
            HashTable->Entries[i].DataValue = Ed;
            DBGME(0,_tprintf(_T("EdLookup(%d,%d)[%x] hash %x new @%x\n"),
                             DeviceAddress, EndpointAddress,
                             Flags,
                             HashValue, (ADDRESS) Ed));
            goto UnlockAndDone;
        }
        if (++i == HashTable->nEntries)
            i = 0;
    }

    /* Bummer, the table is full;
     * Allocate a 2x sized one, copy, free old, set new, restart.
     */
    i = HashTable->nEntries * 2;

    pHeap = CurrentHeap();
    NewHashTable = (PSL_HT) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY,
                                              sizeof(*HashTable) +
                                              (i-1) * sizeof(struct _SL_HT_ENTRY),
                                              0);
    if (!NewHashTable) {
        DBGME(3,_tprintf(_T("SL(%x) HT outtamem2.\n"), (ADDRESS) this));
        goto UnlockAndDone;
    }

    memcpy(NewHashTable,
           HashTable,
           sizeof(*HashTable) + sizeof(struct _SL_HT_ENTRY) * (i/2 - 1));
    NewHashTable->nEntries = i;
    this->HashTable = NewHashTable;
    pHeap->v->Free(pHeap,0,HashTable);
    HashTable = NewHashTable;
    goto Again;

 UnlockAndDone:
    Mutex_Unlock(&this->HashTableLock);

    return Ed;
}

/* Lazy-evaluated initialization of the periodic elements
 */
PRIVATE BOOL
SlInitializePeriodic(PSL811 this)
{
    if (this->PeriodicList == NULL)
        this->PeriodicList = malloc(SL_PERIODIC_LIST_SIZE * sizeof(ADDRESS));
    if (this->PeriodicList == NULL)
        return FALSE;

    this->HashTable = NULL;
    Mutex_Init(&this->HashTableLock);
    this->IsoPacketsActive = 0;

    if (this->Schedule == NULL)
        this->Schedule = UskCreateSchedule(this->PeriodicList,
                                           SL_PERIODIC_LIST_SIZE);
    if (this->Schedule == NULL)
        return FALSE;

    return TRUE;
}

#define SlIsPeriodicOk(_this_) \
    (((_this_)->Schedule != NULL) || (SlInitializePeriodic(_this_)))

/* Forward
 */
PRIVATE BOOL SlEnoughTime(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet, UINT Length);

/* Start-actions
 */
PRIVATE ISR_RESULT SA_InitSetup(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet, BOOL IsSlow)
{
    UINT8 ctl;

    /* Sanity
     */
    assert(Packet->SetupLength == 8);

    /* Build the value to stick into the ctl register
     */
    ctl = SL_HCR_Out | SL_HCR_Arm | SL_HCR_Enable;

    /* preamble?
     */
    if ((Packet->Flags & URB_FLAG_LOW_SPEED) && !IsSlow)
        ctl |= SL_HCR_Preamble;

    /* Set buffer pointer to start of usable RAM
     */
    SlWriteReg(Sl,SL_HBA_A,SL_MEM);

    /* If we got something to say
     */
    SlWriteSeq(Sl,SL_MEM,8,Packet->SetupData);

    /* BUGBUG check for time
     */

    /* Length
     */
    SlWriteReg(Sl,SL_HBL_A,8);

    /* Set PID, endpoint address, device address
     */
    SlWriteReg(Sl,SL_HPID_A,
               SL_MAKE_HPID(Packet->Pid,Packet->EndpointAddress));
    SlWriteReg(Sl,SL_HDA_A, (UINT8)(Packet->DeviceAddress & SL_HDA_MASK));

    /* Go
     */
    SlWriteReg(Sl,SL_HCR_A,ctl);

    SetToggleForPacket(Packet, SL_HCR_DToggle);

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SA_InitIn(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet, BOOL IsSlow)
{
    UINT32 l = Packet->ExpectedDataLength - Packet->ActualDataLength;
    UINT8 ctl;

    /* Build the value to stick into the ctl register
     */
    ctl = ToggleForPacket(Packet);
    ctl |= SL_HCR_Arm | SL_HCR_Enable;

    /* isochronous?
     */
    if (Packet->Flags & URB_FLAG_ISOCHRONOUS)
        ctl |= SL_HCR_ISO;

    /* preamble?
     */
    if ((Packet->Flags & URB_FLAG_LOW_SPEED) && !IsSlow)
        ctl |= SL_HCR_Preamble;

    /* Length
     */
    if (l > Packet->MaximumPacketSize)
        l = Packet->MaximumPacketSize;
    if (l > SL_HBL_MAX)
        l = SL_HBL_MAX;
    SlWriteReg(Sl,SL_HBL_A,(UINT8)l);

    /* But do we have enough time
     */
    if (!SlEnoughTime(Sl,Packet,l))
        return ISR_SUSPEND;

    /* Set PID for IN phase, endpoint address
     */
    SlWriteReg(Sl,SL_HPID_A,
               SL_MAKE_HPID(SL_HPID_IN,Packet->EndpointAddress));
    SlWriteReg(Sl,SL_HDA_A, (UINT8)(Packet->DeviceAddress & SL_HDA_MASK));

    /* Go
     */
    SlWriteReg(Sl,SL_HCR_A,ctl);

    ctl = ToggleForPacket(Packet);
    ctl = ctl ^ SL_HCR_DToggle;
    SetToggleForPacket(Packet,ctl);

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SA_InitOut(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet, BOOL IsSlow)
{
    UINT32 l = Packet->ExpectedDataLength - Packet->ActualDataLength;
    UINT8 ctl;

    /* Sanity
     */
    if (l > SL_HBL_MAX)
        l = SL_HBL_MAX;

    /* Build the value to stick into the ctl register
     */
    ctl = ToggleForPacket(Packet);
    ctl |= SL_HCR_Arm | SL_HCR_Enable | SL_HCR_Out;

    /* isochronous?
     */
    if (Packet->Flags & URB_FLAG_ISOCHRONOUS)
        ctl |= SL_HCR_ISO;

    /* preamble?
     */
    if ((Packet->Flags & URB_FLAG_LOW_SPEED) && !IsSlow)
        ctl |= SL_HCR_Preamble;

    /* We got something to say
     */
    if (l)
    {
        SlWriteSeq(Sl,SL_MEM,l,Packet->Data+Packet->ActualDataLength);
    }

    /* Length
     */
    SlWriteReg(Sl,SL_HBL_A,(UINT8)l);

    /* Set PID for OUT phase, endpoint address
     */
    SlWriteReg(Sl,SL_HPID_A,
               SL_MAKE_HPID(SL_HPID_OUT,Packet->EndpointAddress));
    SlWriteReg(Sl,SL_HDA_A, (UINT8)(Packet->DeviceAddress & SL_HDA_MASK));

    /* Go
     */
    SlWriteReg(Sl,SL_HCR_A,ctl);

    ctl = ToggleForPacket(Packet);
    ctl = ctl ^ SL_HCR_DToggle;
    SetToggleForPacket(Packet,ctl);

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SA_InitStatusOut(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet, BOOL IsSlow)
{
    UINT8 ctl;

    /* Build the value to stick into the ctl register
     */
    ctl = SL_HCR_Arm | SL_HCR_Enable | SL_HCR_Out | SL_HCR_DToggle;

    /* preamble?
     */
    if ((Packet->Flags & URB_FLAG_LOW_SPEED) && !IsSlow)
        ctl |= SL_HCR_Preamble;

    /* Length
     */
    SlWriteReg(Sl,SL_HBL_A,0);

    /* Set PID for IN phase, endpoint address
     */
    SlWriteReg(Sl,SL_HPID_A,
               SL_MAKE_HPID(SL_HPID_OUT,Packet->EndpointAddress));
    SlWriteReg(Sl,SL_HDA_A, (UINT8)(Packet->DeviceAddress & SL_HDA_MASK));

    /* Go
     */
    SlWriteReg(Sl,SL_HCR_A,ctl);

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SA_InitStatusIn(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet, BOOL IsSlow)
{
    UINT8 ctl;

    /* Build the value to stick into the ctl register
       */
    ctl = SL_HCR_Arm | SL_HCR_Enable | SL_HCR_DToggle;

    /* preamble?
     */
    if ((Packet->Flags & URB_FLAG_LOW_SPEED) && !IsSlow)
        ctl |= SL_HCR_Preamble;

    /* Length
     */
    SlWriteReg(Sl,SL_HBL_A,0);

    /* Set PID for IN phase, endpoint address
     */
    SlWriteReg(Sl,SL_HPID_A,
               SL_MAKE_HPID(SL_HPID_IN,Packet->EndpointAddress));
    SlWriteReg(Sl,SL_HDA_A, (UINT8)(Packet->DeviceAddress & SL_HDA_MASK));

    /* Go
     */
    SlWriteReg(Sl,SL_HCR_A,ctl);

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SA_InitOut2(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet, BOOL IsSlow)
{
    /* No idea why this should be necessary.
     * Minimum value seems to be around 86-90us.
     */
    Delay(100);
    return SA_InitOut(Sl,Packet,IsSlow);
}

/* End-actions
 */
PRIVATE ISR_RESULT SA_CheckErrors(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet)
{
    UINT8 Status;;

    Status = SlReadReg(Sl,SL_STATUS_A);

    /* Everything ok?
     */
    if (0 == (Status & (SL_STATUS_NAK |
                        SL_STATUS_Error |
                        SL_STATUS_Timeout |
                        SL_STATUS_STALL |
                        SL_STATUS_Overflow)))
    {
        return ISR_CONTINUE;
    }

    /* Just a glitch?
     */
    if (Status & SL_STATUS_NAK)
    {
        return ISR_SUSPEND;
    }

    /* Something serious, bail out
     * Swap STALL and Timeout to make STALL same as others
     */
#define MiStatus(s) ((USBM_URB_STATUS)((s & ~0x84) | ((s&0x80)>>5) | ((s&4)<<5)))
    Packet->FinalStatus = MiStatus(Status);
    return ISR_ERROR;
}

PRIVATE ISR_RESULT SA_StoreBytes(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet)
{
    UINT Length;
    UINT Bytes2Go = 0;
    ISR_RESULT r;

    r = SA_CheckErrors(Sl,Packet);
    if (r != ISR_CONTINUE)
        return r;

    Length = SlReadReg(Sl,SL_HBL_A) - SlReadReg(Sl,SL_TC_A);
    /* BUGBUG Shakey h/w ...
     */
    Length &= 0xff;

    if (Length)
    {
        SlReadSeq(Sl,SL_MEM,Length,Packet->Data+Packet->ActualDataLength);
        Packet->ActualDataLength += (UINT16)Length;

        Bytes2Go = Packet->ExpectedDataLength - Packet->ActualDataLength;
        /* But not if we did not get a full packet */
        if (Length < Packet->MaximumPacketSize)
            Bytes2Go = 0;
    }

    /* Ask for more iff...
     */
    if (Bytes2Go &&
        Length)
    {
        return ISR_REPEAT;
    }

    /* Only go to status for SETUPs
     */
    if (Packet->Pid == USB_PID_IN)
        return ISR_DONE;

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SA_LoadBytes(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet)
{
    UINT Length;
    UINT Bytes2Go;
    ISR_RESULT r;

    r = SA_CheckErrors(Sl,Packet);
    if (r != ISR_CONTINUE)
        return r;

    Length = SlReadReg(Sl,SL_HBL_A) - SlReadReg(Sl,SL_TC_A);

    Packet->ActualDataLength += (UINT16)Length;
    Bytes2Go = Packet->ExpectedDataLength - Packet->ActualDataLength;

    if (Bytes2Go)
    {
        return ISR_REPEAT;
    }

    /* Only go to status for SETUPs
     */
    if (Packet->Pid == USB_PID_OUT)
        return ISR_DONE;

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SA_ControlDone(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet)
{
    ISR_RESULT r;

    r = SA_CheckErrors(Sl,Packet);
    if (r != ISR_CONTINUE)
        return r;

    return ISR_DONE;
}

/*
 * Actual scripts are linearized in this array
 */
PRIVATE const ISR_STEP AllScripts[] = {
#define NoDataControlScript 0
    { SA_InitSetup, SA_CheckErrors},
    { SA_InitStatusIn, SA_ControlDone},
#define ControlReadScript   2
    { SA_InitSetup, SA_CheckErrors},
    { SA_InitIn, SA_StoreBytes},
    { SA_InitStatusOut, SA_ControlDone},
#define ControlWriteScript  5
    { SA_InitSetup, SA_CheckErrors},
    { SA_InitOut2, SA_LoadBytes},
    { SA_InitStatusIn, SA_ControlDone},
#define ReadScript          8
    { SA_InitIn, SA_StoreBytes},
#define WriteScript         9
    { SA_InitOut, SA_LoadBytes}
};

#if _DEBUG
#define SizeofAllScripts ((sizeof AllScripts) / sizeof AllScripts[0])
typedef _TCHAR *NAME;
NAME ActionStartNames[SizeofAllScripts] = {
    _T("InitSetup"), _T("InitStatusIn"),
    _T("InitSetup"), _T("InitIn"), _T("InitStatusOut"),
    _T("InitSetup"), _T("InitOut2"), _T("InitStatusIn"),
    _T("InitIn"),
    _T("InitOut"),
};

NAME ActionEndNames[SizeofAllScripts] = {
    _T("CheckErrors"), _T("ControlDone"),
    _T("CheckErrors"), _T("StoreBytes"), _T("ControlDone"),
    _T("CheckErrors"), _T("LoadBytes"), _T("ControlDone"),
    _T("StoreBytes"),
    _T("LoadBytes"),
};

/*
 * LOGGING
 */
#define MAXLOG 64
#define DUMPSIZE (16+16)
typedef UINT8 MINIDUMP[DUMPSIZE];

PRIVATE struct {
    INT16 Step;
    UINT16 Fetched;
    UINT16 Frame;
    UINT8 Toggle;
    UINT8 Other;
    MINIDUMP Regs;
} Log[MAXLOG];
PRIVATE int CurL = 0;

#define SetLog(_n_) CurL = _n_

#else
#define SetLog(_n_)
#endif

/* Compute the min/max bandwidth for xfers to a given node
 * BUGBUG Revise for ISO
 */
PRIVATE void
SlComputeBandwidth(
    USBM_URB_FLAGS Flags,
    UINT16 MaximumPacketLength,
    UINT16 *MaxBw,
    UINT16 *MinBw
    )
{
    UINT16 Overhead;
    UINT16 Bandwidth;

    /*
     * For an interrupt transfer, the minimum b/w is when a NAK is returned.
     * Call that the per-packet overhead.
     */
    Overhead = (Flags & URB_FLAG_INTERRUPT) ? 13 : 9;

    /*
     * That is in bytes.  Add bitstuffing and turn it into full-speed clocks.
     */
    Bandwidth = ((MaximumPacketLength + Overhead) * 8 * 7) / 6;
    Overhead  = ((                  0 + Overhead) * 8 * 7) / 6;

    /*
     * A slow-speed node runs at 1/8th full speed.
     */
    if (Flags & URB_FLAG_LOW_SPEED) {
        Bandwidth *= 8;
        Overhead  *= 8;
    }

    /* Return the computed values
     */
    *MaxBw = Bandwidth;
    *MinBw = Overhead;
}

PRIVATE BOOL SlEnoughTime(PSL811REGS Sl, PUSBM_REQUEST_BLOCK Packet, UINT Length)
{
    UINT TicksLeft;
    UINT16 MaxTicks, MinTicks;

    SlComputeBandwidth(Packet->Flags,
                       (UINT16) Length,
                       &MaxTicks,
                       &MinTicks);
    TicksLeft = SlReadReg(Sl,SL_SOF_HI) * 64;
    return (TicksLeft > MaxTicks);
}

/* Rolling index for the current frame
 */
#define CurrentFrameNumber(_this_) ((_this_)->FrameCounter & 0x7ff)

/* Insert an interrupt endpoint in the periodic schedule
 */
PRIVATE void
SlInsertInterruptEd(PSL811 this,
                    PSL_ED Ed)
{
    int Frequency, StartSlot;
    ADDRESS *NextPtr;
    PSL_ED NextEd;
    UINT32 Value32;

    /* See where it goes first
     */
    Frequency = UskFrequency(Ed->InterruptHead);
    StartSlot = UskStart(Ed->InterruptHead);

    /* Its replicated a few times (at least 256/16==16)
     */
    for (;;) {

        /* Skip all ISO EDs, they must-go-first
         */
        NextPtr = &this->PeriodicList[StartSlot];

        for (;;) {
            Value32 = *NextPtr;

            /* Stop at the first INT or at the end of the list
             */
            if ((Value32 == HC_NULL_POINTER) ||
                ((Value32 & SL_POINTER_TYPE_MASK) == SL_POINTER_TYPE_INT))
                break;

            NextPtr = (ADDRESS*) (Value32 & SL_POINTER_MASK);
        }

        /* Skip any endpoint with a lower polling frequency
         */
        for (NextEd = NULL; Value32 & SL_POINTER_MASK;) {

            NextEd = (PSL_ED) (Value32&SL_POINTER_MASK);

            if (UskIsLowerOrSame(NextEd->InterruptHead,Ed->InterruptHead))
                break;

            NextPtr = &NextEd->Next;
            Value32 = *NextPtr;
        }

        /* Make sure we dont do this twice
         */
        if (NextEd != Ed) {

            /* Link the endpoint to the next one, or null.
             */
            Ed->Next = Value32;
            Value32 = (ADDRESS)(Ed) | SL_POINTER_TYPE_INT;

            /* If we are writing into the roots we must do it a few times,
             * otherwise its easy.
             */
            if (NextPtr != &this->PeriodicList[StartSlot]) {

                DBGME(0,_tprintf(_T("SlIIEd: Node %x after %x\n"),
                                 (ADDRESS) Ed, (ADDRESS) NextPtr));
                *NextPtr = Value32;

            } else {

                int RootIndex;
                for (RootIndex = StartSlot;
                     RootIndex < SL_PERIODIC_LIST_SIZE;
                     RootIndex += MAX_POLLING_FREQUENCY) {

                    DBGME(0,_tprintf(_T("SlIIEd: Node %x @root %d\n"),
                                     (ADDRESS) Ed, RootIndex));
                    this->PeriodicList[RootIndex] = Value32;
                }
            }
        }

        /* Go look at the next slot, if any
         */
        StartSlot += Frequency;

        if (StartSlot >= USK_LEVEL1_MAX)
            break;
    }
}

/* Insert a ISO ED in the schedule
 */
PRIVATE void
SlInsertIsoEd(PSL811 this,
              PSL_ED Ed)
{
    int StartSlot;
    UINT32 Value32;

    DBGME(0,_tprintf(_T("SlIIU2Ed: Node %x\n"),
                     (ADDRESS) Ed));

    /* The schedule is replicated a few times (at least 256/16==16)
     */
    for (StartSlot = 0; StartSlot < MAX_POLLING_FREQUENCY; StartSlot++) {

        /* The ISO EDs must-go-first
         */
        Ed->Next = this->PeriodicList[StartSlot];

        Value32 = (ADDRESS)Ed;
        Value32 |= SL_POINTER_TYPE_ISO;

        /* Link it into all the repetitions of this slot
         */
        {
            int RootIndex;
            for (RootIndex = StartSlot;
                 RootIndex < SL_PERIODIC_LIST_SIZE;
                 RootIndex += MAX_POLLING_FREQUENCY) {

                DBGME(0,_tprintf(_T("SlIsoIIEd: Node %x @root %d\n"),
                                 (ADDRESS) Ed, RootIndex));
                this->PeriodicList[RootIndex] = Value32;
            }
        }

        /* Go look at the next slot, if any
         */
    }
}


/* Remove an endpoint from the periodic schedule
 */
PRIVATE void
SlDropEdFromSchedule(PSL811 this,
                     PSL_ED Ed)
{
    int Frequency, StartSlot;
    ADDRESS *NextPtr;
    ADDRESS EdPointerValue;
    UINT32 Value32;

    /* Where do we start looking
     */
    EdPointerValue = (ADDRESS)Ed;
    EdPointerValue |= (Ed->EndpointCaps & SL_ED_ISOCHRONOUS) ? 
        SL_POINTER_TYPE_ISO : SL_POINTER_TYPE_INT;

    Frequency = UskFrequency(Ed->InterruptHead);
    StartSlot = UskStart(Ed->InterruptHead);

    DBGME(1,_tprintf(_T("Sl(%x)DEFS %x.  ephy %x f=%d s=%d\n"),
                     (ADDRESS)this,(ADDRESS)Ed, EdPointerValue,
                     Frequency, StartSlot));

    /* Its replicated a few times (at least 256/16==16)
     */
    for (;;) {

        /* Keep going until we find it
         */
        NextPtr = &this->PeriodicList[StartSlot];

        for (;;) {
            Value32 = *NextPtr;

            /* Stop at the end of the list (sanity)
             */
            if ((Value32 == EdPointerValue) ||
                (Value32 == HC_NULL_POINTER))
                break;

            NextPtr = (ADDRESS*) (Value32 & SL_POINTER_MASK);
        }

        /* Make sure the schedule is not broken
         */
        if (Value32 == EdPointerValue) {

            /* If we are writing into the roots we must do it a few times,
             * otherwise its easy.
             */
            if (NextPtr != &this->PeriodicList[StartSlot]) {

                DBGME(0,_tprintf(_T("SlDEFS: Node %x was after %x\n"),
                                 (ADDRESS) Ed, (ADDRESS) NextPtr));
                *NextPtr = Ed->Next;

            } else {

                int RootIndex;
                for (RootIndex = StartSlot;
                     RootIndex < SL_PERIODIC_LIST_SIZE;
                     RootIndex += MAX_POLLING_FREQUENCY) {

                    DBGME(0,_tprintf(_T("SlDEFS: Node %x @root %d\n"),
                                     (ADDRESS) Ed, RootIndex));
                    this->PeriodicList[RootIndex] = Ed->Next;
                }
            }

        }

        /* Go look at the next slot, if any
         */
        StartSlot += Frequency;

        if (StartSlot >= USK_LEVEL1_MAX)
            break;
    }

    /* Unlink the next endpoint, if any
     */
    Ed->Next = HC_NULL_POINTER;

    /* Release the bandwidth reservation for the endpoint
     */
    {
        UINT16 MaxStartBw, MinStartBw;
        USBM_URB_FLAGS Flags;

        Value32 = Ed->EndpointCaps;
        Flags   = (Value32 & SL_ED_LOW_SPEED) ? URB_FLAG_LOW_SPEED : 0;
        Flags  |= (Value32 & SL_ED_INTERRUPT) ? URB_FLAG_INTERRUPT : URB_FLAG_ISOCHRONOUS;
        Flags  |= (Value32 & SL_ED_IN) ? URB_FLAG_IN : URB_FLAG_OUT;
        Value32 = (Value32 & SL_ED_MPS) >> SL_ED_MPS_SHIFT;

        SlComputeBandwidth(Flags,
                           (UINT16)Value32,
                           &MinStartBw,
                           &MaxStartBw);

        UskCancelReservation(this->Schedule,
                             Ed->InterruptHead,
                             MinStartBw, MaxStartBw);
    }
}

PRIVATE void
SlDropEdFromAsync(PSL811 this,
                  PSL_ED Ed)
{
    PSL_ED *Previous;

    assert(offsetof(SL_ED,Next) == 0);

    Previous = & this->NonPeriodicList;

    while (*Previous) {
        if (*Previous == Ed) {
            *Previous = (PSL_ED)Ed->Next;
        }
        Previous = (PSL_ED*) *Previous;
    }
}

/* Destroy all endpoints for a given USB address.
 */
PRIVATE void SlCancelUrbsForFunction(
    PSL811 this, 
    UINT8 UsbAddress)
{
    /* NYI */
}

PRIVATE void
DestroyEdsForFunction(PSL811 this, 
                      UINT8 UsbAddress)
{
    PSL_ED Ed;
    int i;
    UINT32 Value32;
    PSL_HT HashTable;

    DBGME(0,_tprintf(_T("Sl(%x)DestroyEds(%x)\n"), (ADDRESS)this, UsbAddress));

    /* We do *not* free the config endpoint
     */
    if (UsbAddress == 0)
        return;

    /* Cancel out any active URBs for this USB address
     */
    SlCancelUrbsForFunction(this,UsbAddress);

    /* Lock the hash table
     */
    Mutex_Lock(&this->HashTableLock);

    /* Sanity check the table is there (assert?)
     */
    HashTable = this->HashTable;
    if (HashTable != NULL) {

        /* Scan the whole table for EDs with this function address
         */
        for (i = 0; i < HashTable->nEntries; i++) {

            Value32 = HashTable->Entries[i].HashValue;
            if (UsbAddress == (Value32 & SL_ED_DEVICE_MASK)) {

                /* Got one, free slot
                 */
                DBGME(1,_tprintf(_T("Sl(%x) Hash[%d] %x %x\n"), (ADDRESS)this,
                                 i, HashTable->Entries[i].HashValue,
                                 (ADDRESS)HashTable->Entries[i].DataValue));
                Ed = HashTable->Entries[i].DataValue;
                HashTable->Entries[i].DataValue = NULL;
                HashTable->Entries[i].HashValue = 0;

                /* We'll find it on the async list, unless its an interrupt/iso
                 */
                Value32 = Ed->EndpointCaps;
                if (Value32 & (SL_ED_ISOCHRONOUS | SL_ED_INTERRUPT))
                {
                    SlDropEdFromSchedule(this,Ed);
                } else
                {
                    SlDropEdFromAsync(this,Ed);
                }

                /* BUGBUG Revise for races
                 */
                DBGME(3,_tprintf(_T("Sl: EdFree(%x)\n"), (ADDRESS)Ed));
                EdFree(this,Ed);
            }
        }

    }

    Mutex_Unlock(&this->HashTableLock);
}

/* Queue a packet to an endpoint
 */
PRIVATE void SlQueueUrb(PUSBM_REQUEST_BLOCK Urb, PSL_ED Ed)
{
#if 0
    /* Link the Urb into the active list for this endpoint
     * BUGBUG revise for races
     */
    SetNextForPacket(Urb,Ed->NextUrb);
    Ed->NextUrb = Urb;
#else
    PUSBM_REQUEST_BLOCK Next = Ed->NextUrb;
    SetNextForPacket(Urb,NULL);
    if (Next == NULL)
    {
        Ed->NextUrb = Urb;
    } else {
        while (NextForPacket(Next) != NULL)
            Next = NextForPacket(Next);
        SetNextForPacket(Next,Urb);
    }
#endif

    /* Signal the HC to rescan the queue for this ED
     * BUGBUG revise for races
     */
    Ed->EndpointCaps |= SL_ED_ACTIVE;
}

/*
 * Initiator for ISO & INT xfers
 */
PRIVATE SCODE SlStartIso(PSL811 this, PUSBM_REQUEST_BLOCK Urb)
{
    PSL_ED Ed = NULL;
    SCODE sc = E_INVALID_PARAMETER;

    /* The hardware has limitations in ISO mode, say so upfront.
     */
    if (Urb->MaximumPacketSize > SL_HBL_MAX)
        return E_INVALID_PARAMETER;

    /* Dynamic-init can only fail because of memory
     */
    if (!SlIsPeriodicOk(this))
        return E_NOT_ENOUGH_MEMORY;

    /* Find the ED where this URB should be posted
     * This is a "virtual" ED, used only in software.
     * This might allocate a new ED if we aint seen it before
     */
    Ed = LookupEd(this,Urb);
    if (Ed == NULL) {
        DBGME(3,_tprintf(_T("Sl@%x: ?i-Ed for %x.%x.%x\n"),
                         (ADDRESS) this,
                         Urb->DeviceAddress,
                         Urb->EndpointAddress,
                         Urb->Flags));
        goto ErrorHandler;
    }

    /* Check if the Ed was ever used before.
     */
    if (Ed->EndpointCaps & SL_ED_INVALID) {

        UINT16 MaxStartBw, MinStartBw, Frequency;
        PUSKNODE Node;

        /* Nope, initialize it
         */
        Ed->EndpointCaps &= ~SL_ED_INVALID;

        /* Make a reservation for the necessary amount of b/w.
         * This reservation will only be cancelled when the device
         * is unplugged.  Notice that this happens only when the
         * endpoint is actually used for the first time, so we are
         * a bit conservative maybe but not wasteful.
         * Notice also that we'll overcommit the b/w if necessary.
         */
                
        /* See how many clocks this will take
         */
        SlComputeBandwidth(Urb->Flags,
                           Urb->MaximumPacketSize,
                           &MinStartBw,
                           &MaxStartBw);

        /* Make sure we have enough bandwidth
         */
        Frequency = (Urb->Flags & URB_FLAG_ISOCHRONOUS) ? 1 : Urb->Interval;
        if (Frequency >= USK_LEVEL1_MAX)
            Frequency = USK_LEVEL1_MAX;
        Node = UskMakeReservation(this->Schedule,
                                  Frequency,
                                  MinStartBw,
                                  MaxStartBw);

        if (Node == USK_INVALID_NODE) {
            sc = E_NOT_RESERVED;
            goto ErrorHandler;
        }

        /* Insert the Endpoint into the schedule
         */
        Ed->InterruptHead = Node;
        if (Urb->Flags & URB_FLAG_ISOCHRONOUS)
            SlInsertIsoEd(this,Ed);
        else
            SlInsertInterruptEd(this,Ed);
    }

#if 0
    /* BUGBUG  BUGBUG  BUGBUG  BUGBUG  BUGBUG  BUGBUG */
    /* Where in the schedule ? */
    if (Urb->Interval == 255)
        Urb->Interval = (UINT8) (CurrentFrameNumber(this) + 1);
#endif

    this->Stats.IsoStarted++;

    /* Enqueue and we are done
     */
    SlQueueUrb(Urb,Ed);
    AtomicInc(&this->IsoPacketsActive);

    /* Done ok
     */
    return S_OK;

 ErrorHandler:
    return sc;

}

/*
 * Initiator for CONTROL and BULK xfers
 */
PRIVATE SCODE SlStartNonIso(PSL811 this, PUSBM_REQUEST_BLOCK Urb)
{
    PSL_ED Ed = NULL;
    SCODE sc = E_INVALID_PARAMETER;

    /* Find the ED where this URB should be posted
     * This is a "virtual" ED, used only in software.
     * This might allocate a new ED if we aint seen it before
     * NB: The *only* reason we need endpoints for non-iso xfers is 
     * that we must keep track of the toggle bit (a.k.a. DATA0/DATA1) 
     * across subsequent bulk transactions to the same endpoint.  Sigh.
     */
    Ed = LookupEd(this,Urb);
    if (Ed == NULL) {
        DBGME(3,_tprintf(_T("Sl@%x: ?Ed for %x.%x.%x\n"),
                         (ADDRESS) this,
                         Urb->DeviceAddress,
                         Urb->EndpointAddress,
                         Urb->Flags));
        goto ErrorHandler;
    }

    /* Check if the Ed was ever used before.
     */
    if (Ed->EndpointCaps & SL_ED_INVALID) {

        /* Nope, initialize it
         */
        Ed->EndpointCaps &= ~SL_ED_INVALID;

        /* Insert the Endpoint in the non-periodic list
         */
        do {
            Ed->Next = (ADDRESS)this->NonPeriodicList;
        } while (!AtomicCmpAndSwap( (ADDRESS*)&this->NonPeriodicList,
                                    Ed->Next,
                                    (ADDRESS)Ed));
    }

    /* Enqueue and we are done
     */
    SlQueueUrb(Urb,Ed);
    AtomicInc(&this->BulkPacketsActive);

    /* Done ok
     */
    return S_OK;

 ErrorHandler:
    return sc;

}

/* Packet scheduler.
 */

/* Walk an endpoint list looking for the first active one
 */
PRIVATE PUSBM_REQUEST_BLOCK SlWalkEndpointList(PSL811 this, PSL_ED Ed)

{
    PUSBM_REQUEST_BLOCK Packet;
    ADDRESS Value32;

    for (;;)
    {
        /* Only if this endpoint has something going
         */
        if (Ed->EndpointCaps & SL_ED_ACTIVE)
        {
            /* Pick either the current or the next packet
             */
            if (((Packet = Ed->CurrentUrb) != NULL) ||
                ((Packet = Ed->NextUrb) != NULL))
            {
                /* If we picked the next then pop the queue
                 */
                if (Packet != Ed->CurrentUrb)
                {
                    Ed->CurrentUrb = Packet;
                    Ed->NextUrb = NextForPacket(Packet);

                    /* Copy the toggle over
                     */
                    SetToggleForPacket(Packet,
                                       (Ed->EndpointCaps & SL_ED_TOGGLE) ? 
                                       SL_HCR_DToggle : 0);

                }

                /* This is the endpoint we are working with today
                 */
                this->CurrentEd = Ed;

                /* Restart this packet's script from the beginning
                 */
                if (Packet->Flags & URB_FLAG_PERIODIC)
                    SetStepForPacket(Packet, StartForPacket(Packet));

                /* ..and return it
                 */
                return Packet;
            } else
            {
                /* Spare us trouble later */
                Ed->EndpointCaps &= ~SL_ED_ACTIVE;
            }
        }

        /* Next endpoint
         */
        Value32 = Ed->Next;

        /* End of list?
         */
        Ed = (PSL_ED) (Value32 & SL_POINTER_MASK);
        if (Ed == NULL)
            break;

    }
    return NULL;
}

PRIVATE PUSBM_REQUEST_BLOCK SlNextPacket(PSL811 this, BOOL IsStartOfFrame)
{
    PSL_ED Ed = this->CurrentEd;
    UINT32 Value32;
    PUSBM_REQUEST_BLOCK Packet = NULL;

    /* If this is not the start of frame we must continue with the next endpoint.
     */
    if (!IsStartOfFrame && Ed/*sanity*/)
    {
        /* Is the current (previous!) endpoint on the periodic list?
         */
        Value32 = Ed->Next;
        if (Ed->EndpointCaps & (SL_ED_ISOCHRONOUS | SL_ED_INTERRUPT))
            /* Yes, we must exhaust the ISO list before we look at the non-periodic
             */
            goto WalkIsoList;

        /* No, look at the next element in the non-periodic list
         */
        Ed = (PSL_ED) (Value32 & SL_POINTER_MASK);
        goto WalkBulkList;
    }

    /* Priority goes to the ISO xfers
     */
    if (this->IsoPacketsActive)
    {
        /* Start from the correct root
         */
        Value32 = CurrentFrameNumber(this);
        Value32 = SlPeriodicIndex(Value32);
        Value32 = this->PeriodicList[Value32];

    WalkIsoList:
        Ed = (PSL_ED) (Value32 & SL_POINTER_MASK);
        if (Ed != NULL)
            Packet = SlWalkEndpointList(this,Ed);
                                    
        if (Packet)
            return Packet;
    }

    /* Look at the non-periodic list
     * BUGBUG round-robin schedule this list
     */
    Ed = this->NonPeriodicList;
 WalkBulkList:
    if (Ed != NULL)
        Packet = SlWalkEndpointList(this, Ed);
    return Packet;
}

/*
 * Initiator
 */
PRIVATE SCODE SlStart(PSL811 this, PUSBM_REQUEST_BLOCK Urb)
{
    DBGME(0,_tprintf(_T("SlStart(%x,%x)\n"),(ADDRESS)this,(ADDRESS)Urb));

    Urb->Flags &= ~URB_FLAG_STATUS;
    Urb->ActualDataLength = 0;
    Urb->FinalStatus = 0;
    SetStepForPacket(Urb,StartForPacket(Urb));
    SetNextForPacket(Urb,NULL);
    SetToggleForPacket(Urb,0);

    /* ISO and INTERRUPT are handled separately
     * NB: Commenting out this line gives you some sort of polling-mode
     *     behaviour that might perhaps suffice in certain cases.
     */
    if (Urb->Flags & URB_FLAG_PERIODIC)
        return SlStartIso(this,Urb);
    else
        return SlStartNonIso(this,Urb);
}

#if _DEBUG
typedef _TCHAR *MYSTRING;

PRIVATE void fancyreg(int i, UINT8 c)
{
    int j;
    static struct {
        MYSTRING Regname;
        MYSTRING Bitname[8];
    } Regs[16] = {
    {_T("HCR_A"),
     { _T("Arm"), _T("Enable"), _T("Out"), _T("r3"),
       _T("ISO"), _T("SOF"), _T("DToggle"), _T("Preamble") }},
    {_T("HBA_A"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("HBL_A"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("STATUS_A"),
     { _T("ACK"), _T("Error"), _T("Timeout"), _T("Sequence"),
       _T("Setup"), _T("Overflow"), _T("NAK"), _T("STALL")}},
    {_T("TC_A"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("CTRL1"),
     {_T("SOF_en"), _T("r1"), _T("r2"), _T("Reset"),
      _T("force_J"), _T("LowSpeed"), _T("Suspend"),_T("r7")}},
    {_T("INTEN"),
     {_T("DONE_A"), _T("DONE_B"), _T("r2"), _T("r3"),
      _T("SOF"),_T("Insert"),_T("Detect"),_T("D_PLUS"),}},
    {_T("SAR"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("HCR_B"),
     { _T("Arm"), _T("Enable"), _T("Out"), _T("r3"),
       _T("ISO"), _T("SOF"), _T("DToggle"), _T("Preamble") }},
    {_T("HBA_B"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("HBL_B"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("STATUS_B"),
     { _T("ACK"), _T("Error"), _T("Timeout"), _T("Sequence"),
       _T("Setup"), _T("Overflow"), _T("NAK"), _T("STALL")}},
    {_T("TC_B"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("ISTATUS"),
     {_T("DONE_A"), _T("DONE_B"),_T("r2"), _T("r3"),
      _T("SOF"),_T("Insert"),_T("Detect"),_T("D_PLUS"),}},
    {_T("REV"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("SOF_HI"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7")}}
    };

    _tprintf(_T("%8s %2x = "), Regs[i].Regname, c);
    for (j = 0; j < 8; j++)
        if (c & (1<<j))
            _tprintf(_T(" %s"), Regs[i].Bitname[j]);
    _tprintf(_T("\n"));
}

PRIVATE void fancyprint(UINT8 *Regvals)
{
    int i;
    for (i = 0; i < 16; i++)
        fancyreg(i,Regvals[i]);
}

PRIVATE _TCHAR *ActionStartName( UINT ActionIndex )
{
    if (ActionIndex <= SizeofAllScripts)
        return ActionStartNames[ActionIndex];
    return _T("UnknownStartAction");
}

PRIVATE _TCHAR *ActionEndName( UINT ActionIndex )
{
    if (ActionIndex <= SizeofAllScripts)
        return ActionEndNames[ActionIndex];
    return _T("UnknownEndAction");
}

PRIVATE _TCHAR *ScriptName( UINT ActionIndex )
{
    if (ActionIndex >= WriteScript)
        return _T("Write Data");
    if (ActionIndex >= ReadScript)
        return _T("Read Data");
    if (ActionIndex >= ControlWriteScript)
        return _T("Control Write");
    if (ActionIndex >= ControlReadScript)
        return _T("Control Read");
    return _T("No Data Control");
}

PRIVATE int StepCountForPacket( PUSBM_REQUEST_BLOCK Packet)
{
    int ActionIndex = StepForPacket(Packet);

    if (ActionIndex >= WriteScript)
        return ActionIndex - WriteScript;
    if (ActionIndex >= ReadScript)
        return ActionIndex - ReadScript;
    if (ActionIndex >= ControlWriteScript)
        return ActionIndex - ControlWriteScript;
    if (ActionIndex >= ControlReadScript)
        return ActionIndex - ControlReadScript;
    return ActionIndex;
}

PRIVATE void DumpPacket(PUSBM_REQUEST_BLOCK Packet, _TCHAR *Banner)
{
    int i, n;

    _tprintf(_T("%sPacket @x%x %s"), Banner, (ADDRESS)Packet,
           ((Packet->Flags & URB_FLAG_ERROR) ? _T("had error in") : 
            ((Packet->Flags & URB_FLAG_COMPLETE) ? _T("successfully completed") :_T("is executing"))));

    _tprintf(_T(" script %s at step %d\n"),
             ScriptName(StepForPacket(Packet)),StepCountForPacket(Packet));

    _tprintf(_T("\tNextPacket %x\n"), (ADDRESS)NextForPacket(Packet));
    _tprintf(_T("\tDeviceAddress %x\n"), Packet->DeviceAddress);
    _tprintf(_T("\tEndpointAddress %x\n"), Packet->EndpointAddress);
    _tprintf(_T("\tPid %x\n"), Packet->Pid);
    _tprintf(_T("\tFlags %x\n"), Packet->Flags);
    _tprintf(_T("\tSetupLength %x\n"), Packet->SetupLength);
    n = Packet->SetupLength;
    if (n)
    {
        _tprintf(_T("\tSetupData[%d]:"),n);
        for (i = 0; i < n; i++)
            _tprintf(_T("%2x "), Packet->SetupData[i]);
        _tprintf(_T("\n"));
    }
    _tprintf(_T("\tExpectedDataLength %x\n"), Packet->ExpectedDataLength);
    _tprintf(_T("\tMaximumPacketSize %d\n"), Packet->MaximumPacketSize);
    _tprintf(_T("\tToggle %x OtherInfo %x\n"),
             ToggleForPacket(Packet), OtherInfoForPacket(Packet));
    _tprintf(_T("\tActualDataLength %x\n"), Packet->ActualDataLength);
    n = Packet->ActualDataLength;
    if (n)
    {
        _tprintf(_T("\tResult[%d]:"),n);
        for (i = 0; (i < 64) && (i < n); i++)
            _tprintf(_T("%2x "), Packet->Data[i]);
        if (i < n)
            _tprintf(_T("..."));
        _tprintf(_T("\n"));
    }
    if (Packet->Flags & URB_FLAG_ERROR)
        _tprintf(_T("\tFinalStatus %x\n"), Packet->FinalStatus);
}

PRIVATE void DumpEd(PSL_ED Ed, _TCHAR *Banner)
{
    _tprintf(_T("%sEndpoint @x%x:\n"), Banner, (ADDRESS)Ed);
    _tprintf(_T("\tCaps %08x CurrentUrb %x  NextUrb %x NextEndpoint %x\n"),
             Ed->EndpointCaps, Ed->CurrentUrb, Ed->NextUrb, Ed->Next);
}

PRIVATE void print_stats(struct _STATS *Stats)
{
    _tprintf(_T("--Counters:\n"));
    _tprintf(_T("\tInterrupts = %d\n"), Stats->IntrCnt);
    _tprintf(_T("\tStartOfFrame = %d\n"), Stats->SofCnt);
    _tprintf(_T("\t(diff = %d)\n"), Stats->IsrCnt);
    _tprintf(_T("\tInsertions = %d\n"), Stats->InsCnt);
    _tprintf(_T("\tA-done = %d\n"), Stats->DoneCnt);
    _tprintf(_T("\tActions = %d\n"), Stats->StepsCnt);
    _tprintf(_T("\tSuspensions = %d\n"), Stats->SuspCnt);
    _tprintf(_T("\tPkts = %d\n"), Stats->PktCnt);
    _tprintf(_T("\tIso = %ds %dd\n"), Stats->IsoStarted, Stats->IsoDone);
}

PRIVATE void Sl811Dump(PSL811 this, BOOL DumpRegisters)
{
    PUSBM_REQUEST_BLOCK Packet;
    int i;
    static UINT8 Regs[DUMPSIZE];
    UINT8 *Dump;

    print_stats(&this->Stats);

    Packet = this->CurrentPacket;
    if (Packet != NULL)
    {
        DumpPacket(Packet,_T("--Current "));
    }

    if (this->EventsPacket || this->PendingPackets)
        _tprintf(_T("--EventsPacket %x PendingPackets %x\n"),
                 (ADDRESS)this->EventsPacket, (ADDRESS)this->PendingPackets);

    if (this->IsoPacketsActive | this->BulkPacketsActive)
    {
        PSL_ED Ed = this->CurrentEd;

        _tprintf(_T("--PacketsActive %d iso %d bulk\n"),
                 this->IsoPacketsActive, this->BulkPacketsActive);
        if (Ed != NULL)
        {
            DumpEd(Ed,_T("--Current"));
        }
    }

    if (CurL)
    {
        _tprintf(_T("Logdump[%d] for %s:\n"), CurL, ScriptName(Log[0].Step));
        for (i = 0; i < CurL; i++)
        {
            int j;

            Dump = Log[i].Regs;
            _tprintf(_T("Step %d %s [Frame %d Offset %d Tog %x Oth %x] => %s with:\n"), i,
                     ActionStartName(Log[i].Step),
                     Log[i].Frame, Log[i].Fetched, Log[i].Toggle, Log[i].Other,
                     ActionEndName(Log[i].Step));
        Dumpit:
            fancyprint(Dump);
            for (j = 16; j < DUMPSIZE; j++)
            {
                if ((j & 15) == 0)
                    _tprintf(_T("Mem[%03d]: "), j);
                _tprintf(_T("%2x "), Dump[j]);
                if ((j & 15) == 15)
                    _tprintf(_T("\n"));
            }
            _tprintf(_T("\n"));
        }
    }

    if (DumpRegisters) {

        DumpRegisters = FALSE;
        SlReadSeq(this->Sl,0,DUMPSIZE,Regs);
        Dump = Regs;
        _tprintf(_T("Current chip status:\n"));
        goto Dumpit; /* brrrrrrrr!!! */
    }
}

#else
#define Sl811Dump(t,d)
#endif

/*
 * CDevice interface
 */

PRIVATE SCODE Sl811DeviceQueryInterface (
    PIDEVICE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_IFile,iid)
        || UuidCmp(&IID_IDevice,iid))
    {
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }
    if (UuidCmp(&IID_IUsbMasterController, iid))
    {
        /* Use a single refcount
         */
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (void*) iSl811Master(pThis);
        return S_OK;
    }

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT Sl811DeviceAddRef (
    PIDEVICE pThis)
{
    return AtomicInc(&pSl811Device(pThis)->CDevice.RefCnt);
}

PRIVATE UINT Sl811DeviceRelease (
    PIDEVICE pThis)
{
    PSL811 this = pSl811Device(pThis);
    UINT Refs = AtomicDec(&this->CDevice.RefCnt);
    if (Refs == 0)
    {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        pHeap->v->Free(pHeap, 0, (void *) this);
    }
    return Refs;
}

PRIVATE SCODE Sl811DeviceReadAt (
    PIDEVICE pThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PSL811 this = pSl811Device(pThis);

    /* In debug, look at state from shell with a "type sl811"
     */
    Sl811Dump(this,TRUE);

    UnusedParameter(this);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);
    UnusedParameter(pSizeRead);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE Sl811DeviceWriteAt (
    PIDEVICE pThis,
    UINT64 Position,
    const BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeWritten)
{
    PSL811 this = pSl811Device(pThis);

    UnusedParameter(this);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);
    UnusedParameter(pSizeWritten);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE Sl811DeviceSetSize (
    PIDEVICE pThis,
    UINT64 Size)
{
    PSL811 this = pSl811Device(pThis);

    UnusedParameter(this);
    UnusedParameter(Size);

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE Sl811DeviceGetSize (
    PIDEVICE pThis,
    /*out*/ UINT64 * Size)
{
    /* Number of ports on root hub
     */
    Uint32ToUint64(*Size,1);

    return S_OK;
}

PRIVATE BOOL Sl811DeviceIsr (
    PIDEVICE pThis,
    /*out*/ BOOL * pNotOurInterrupt)
{
    PSL811 this = pSl811Device(pThis);
    PSL811REGS Sl = this->Sl;
    PUSBM_REQUEST_BLOCK CurrentPacket;
    UINT8 iStatus;
    BOOL Reschedule = FALSE, DoScript = FALSE;
    ISR_RESULT r;

    this->Stats.IntrCnt++;

    /* See whats up
     */
    iStatus = SlReadReg(Sl,SL_ISTATUS);

    /* Insertion detected?
     */
    if (iStatus & SL_INTEN_Insert)
    {
        this->Stats.InsCnt++;
        this->Stats.IsrCnt++;

        CurrentPacket = this->EventsPacket;
        this->PortStatusChanges = USB_ST_PORT_CONNECTION;
        if (CurrentPacket)
        {
            this->EventsPacket = NULL;

            CurrentPacket->Flags |= URB_FLAG_COMPLETE;
            Set2(CurrentPacket->Data,(1<<1)); /* bitmap, portno=1 */
            CurrentPacket->ActualDataLength = 2;
            Reschedule = Condition_InterruptSignal(CurrentPacket->Done);
        }
    }
    
    /* Command complete?
     */
    CurrentPacket = this->CurrentPacket;
    if (iStatus & SL_INTEN_DONE_A)
    {
        this->Stats.DoneCnt++;
        this->Stats.IsrCnt++;

        if (CurrentPacket)
        {
#if _DEBUG
            if (CurL < MAXLOG)
            {
                Log[CurL].Step = StepForPacket(CurrentPacket);
                Log[CurL].Frame = (UINT16) this->FrameCounter;
                Log[CurL].Fetched = CurrentPacket->ActualDataLength;
                Log[CurL].Toggle = ToggleForPacket(CurrentPacket);
                Log[CurL].Other = OtherInfoForPacket(CurrentPacket);
                SlReadSeq(Sl,0,DUMPSIZE,Log[CurL++].Regs);
            }
#endif
            /* Invoke the action-termination procedure
             */
            r = ActionEndForPacket(CurrentPacket)(Sl,CurrentPacket);

            /*
             * See what we do next with this packet
             */
            if (r == ISR_CONTINUE)
            {
                /* Ok, this step was completed normally, move on
                 */
                SetOtherInfoForPacket(CurrentPacket,1);
                StepForPacket(CurrentPacket)++;
                DoScript = TRUE;
            } else
            if (r == ISR_SUSPEND)
            {
                /* Wait until the next SOF, the device is not ready (NAK)
                 * The toggle bit stays with the packet.
                 */
                this->Stats.SuspCnt++;
                SetOtherInfoForPacket(CurrentPacket,2);
                this->CurrentPacket = NULL;

            } else
            if (r != ISR_REPEAT)
            {
                /* The script is complete
                 */
                PSL_ED CurrentEd = this->CurrentEd;

                assert((r == ISR_DONE) || (r == ISR_ERROR));

                this->Stats.PktCnt++;

                /* Move the toggle bit back into the endpoint
                 */
                CurrentEd->EndpointCaps = (CurrentEd->EndpointCaps & ~SL_ED_TOGGLE) |
                    (ToggleForPacket(CurrentPacket) ? SL_ED_TOGGLE : 0);

                /* BUGBUG Let the action decide if the packet is complete */
                if (CurrentPacket->ActualDataLength ||
                    (0 == (CurrentPacket->Flags & URB_FLAG_PERIODIC)))
                {
                    if (CurrentPacket->Flags & URB_FLAG_PERIODIC)
                    {
                        this->Stats.IsoDone++;
                        this->IsoPacketsActive--;
                    } else
                    {
                        this->BulkPacketsActive--;
                    }
                    CurrentEd->CurrentUrb = NULL;
                    CurrentPacket->Flags |= URB_FLAG_COMPLETE;
                }
                /* If there was an error the packet *is* complete
                 */
                if (r != ISR_DONE)
                    CurrentPacket->Flags |= URB_FLAG_ERROR | URB_FLAG_COMPLETE;

                SetOtherInfoForPacket(CurrentPacket,0x99);
                this->CurrentPacket = NULL;
                if (CurrentPacket->Flags & URB_FLAG_COMPLETE)
                {
                    Reschedule |= Condition_InterruptSignal(CurrentPacket->Done);
                }
            } else
            {
                DoScript = TRUE;
            }
#if 1
            /* is this ready */
            if (!DoScript)
                goto Pickapacket;
#endif
        }
    }

    /* Start-of-frame?
     */
    if (iStatus & SL_INTEN_SOF)
    {
        this->FrameCounter++;

Pickapacket:
        if (this->CurrentPacket == NULL)
        {
            /* Select one from either the async or periodic
             */
            if (this->IsoPacketsActive | this->BulkPacketsActive)
            {
                CurrentPacket = SlNextPacket(this,
                                             (iStatus & SL_INTEN_SOF) != 0);
                if (CurrentPacket)
                {
                    this->CurrentPacket = CurrentPacket;
                    SetOtherInfoForPacket(CurrentPacket,3);
                    DoScript = TRUE;
                }
            }
        }
    }

    /* Ack the interrrupt
     */
    SlWriteReg(Sl,SL_ISTATUS,iStatus);

    /* Initiate the next step in the packet's script.
     */
    if (DoScript)
    {
        /* Execute the start action.  It might tell us that there is not enough
         * time left in the current frame, or that all is fine (ISR_CONTINUE).
         */
        r = ActionStartForPacket(CurrentPacket)(Sl,
                                                CurrentPacket,
                                                (this->PortStatus & USB_ST_PORT_LOW_SPEED) != 0);

        this->Stats.StepsCnt++;

        if (r == ISR_CONTINUE)
        {
            SetOtherInfoForPacket(CurrentPacket,5);
        }
        else
        {
            /* Wait until the next SOF, not enough time left in this frame.
             */
            assert(r == ISR_SUSPEND);

            this->Stats.SuspCnt++;
            SetOtherInfoForPacket(CurrentPacket,4);
            this->CurrentPacket = NULL;
        }
    }

    return Reschedule;
}

PRIVATE void MilliSleep(int mSecs)
{
    TIME t;
    Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(mSecs)));
    SleepUntil(t);
}

PRIVATE UINT16 SlReset(PSL811 this)
{
    PSL811REGS Sl = this->Sl;
    _TCHAR *m = NULL;
    UINT8 c;
    UINT16 ret = USB_ST_PORT_POWER;

    /* Follow the recommended reset procedure
     */

    /* Disable interrupts
     */
    SlWriteReg(Sl,SL_INTEN, 0);

    /* Dis-Arm and reset bus
     */
    SlWriteReg(Sl,SL_HCR_A, 0);
    SlWriteReg(Sl,SL_HCR_B, 0);
    SlWriteReg(Sl,SL_CTRL1, 0);

    /* Wait 
     */
    MilliSleep(500);

    /* Clear any interrupts we might have generated
     */
    SlWriteReg(Sl, SL_ISTATUS, 0xFF);

    /* Check on the bus' connection status
     */
    c = SlReadReg(Sl,SL_ISTATUS);

    /* Nothing there?
     */
    if (c & SL_INTEN_Detect) 
    {
        m = _T("No device or no power");
    }
    else 
    {
        this->PortStatusChanges |= USB_ST_PORT_RESET;
        ret |= USB_ST_PORT_CONNECTION | USB_ST_PORT_ENABLE;
        if (!(c & SL_INTEN_D_PLUS ))
        {	
            /* Slow speed
             */
            ret |= USB_ST_PORT_LOW_SPEED;

            SlWriteReg(Sl,SL_SOF_HI,SL_SOFH_1ms|SL_SOFH_Master|SL_SOFH_Polarity);
            SlWriteReg(Sl,SL_SOF_LOW, SL_SOFL_1ms);
            SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_Reset|SL_CTRL1_LowSpeed);
            MilliSleep(10);
            SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_LowSpeed);

            m = _T("Low speed device connected");
        }
        else
        {
            /* Full speed
             */

            SlWriteReg(Sl,SL_SOF_HI,SL_SOFH_1ms|SL_SOFH_Master);
            SlWriteReg(Sl,SL_SOF_LOW, SL_SOFL_1ms);
            SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_Reset|SL_CTRL1_EOF2);
            MilliSleep(10);
            SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_EOF2);

            m = _T("Full speed device connected");
        }

        /* start the SOF or EOP */
        SlWriteReg(Sl,SL_HPID_A, SL_MAKE_HPID(SL_HPID_SOF,0));
        SlWriteReg(Sl,SL_HDA_A, 0);
        SlWriteReg(Sl,SL_HCR_A, SL_HCR_Arm);
    }

    MilliSleep(25);

    /* Clear any pending interrupts
     */
    SlWriteReg(Sl,SL_ISTATUS, 0xFF);

    /* Enable them interrupts
     */
    SlWriteReg(Sl,SL_INTEN,
               SL_INTEN_DONE_A |
               SL_INTEN_SOF |
               SL_INTEN_Insert );

    /* Clear the port connection event
     */
    this->PortStatusChanges &= ~USB_ST_PORT_CONNECTION;

    /* Say what happened
     */
    DBGME(3,_tprintf(_T("%s\n"), m));

    return ret;
}

PRIVATE INT Sl811DeviceInitialize (
    PIDEVICE pThis)
{
    PSL811 this = pSl811Device(pThis);

    UnusedParameter(this);
    return 0;
}

/*
 * IUsbMasterController interface
 */
PRIVATE SCODE Sl811MasterQueryInterface (
    PIUSBMASTERCONTROLLER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return Sl811DeviceQueryInterface(iSl811Device(pSl811Master(pThis)), iid, ppObject);
}

PRIVATE UINT Sl811MasterAddRef (
    PIUSBMASTERCONTROLLER pThis)
{
    return Sl811DeviceAddRef(iSl811Device(pSl811Master(pThis)));
}

PRIVATE UINT Sl811MasterRelease (
    PIUSBMASTERCONTROLLER pThis)
{
    return Sl811DeviceRelease(iSl811Device(pSl811Master(pThis)));
}

PRIVATE SCODE
Sl811RootHub(PSL811 this,
   /*inout*/ PUSBM_REQUEST_BLOCK Urb)
{
    UINT16 Status;
    UINT16 Changes;

    if (Urb->Pid == USB_PID_EVENT)
    {
        DBGME(0,_tprintf(_T("Sl811:Event %x\n"),Urb->Data[0]));
        if (Urb->Data[0] == USBM_EVENT_DISCONNECTED)
        {
            DestroyEdsForFunction(this,Urb->Data[1]);
        }
        goto DoneOk;
    }

    /* GetPortStatusChanges() ?
     */
    if ((Urb->Pid == USB_PID_IN) &&
        (Urb->Flags & URB_FLAG_INTERRUPT))
    {
    
        this->EventsPacket = Urb;
        Changes = this->PortStatusChanges;
        Urb->ActualDataLength = 2;
        Set2(Urb->Data,2); /* port1 */
        if (Changes)
        {
            this->EventsPacket = NULL;
            Urb->Flags |= URB_FLAG_COMPLETE;
            Condition_InterruptSignal(Urb->Done);
        }
        DBGME(0,_tprintf(_T("Sl811:gpsc %x\n"),Changes));
        return S_OK;
    }

    /* GetPortStatus() ?
     */
    if ((Urb->Pid == USB_PID_SETUP) &&
        (Urb->Flags & URB_FLAG_IN) &&
        (Urb->SetupData[1] == USB_R_GET_STATUS))
    {

        Changes = this->PortStatusChanges;
        /* If we had a change in connection status see what the current status is.
         */
        if (Changes & USB_ST_PORT_CONNECTION)
            this->PortStatus = SlReset(this);
        Status = this->PortStatus;

        Urb->ActualDataLength = 4;
        Set2(Urb->Data,Status);
        Set2(Urb->Data+2,Changes);
        Condition_InterruptSignal(Urb->Done);
        DBGME(0,_tprintf(_T("Sl811:gps %x %x\n"),Status, Changes));
        goto DoneOk;
    }

    /* SetClearPortFeature() ?
     */
    if ((Urb->Pid == USB_PID_SETUP) &&
        (Urb->Flags & URB_FLAG_CONTROL))
    {

        /* Which feature? And we only got one port.
         */
        Changes = Get2(Urb->SetupData+2);
        DBGME(0,_tprintf(_T("Sl811:scpf %x %x\n"), Changes,Urb->SetupData[1]));

        /* We only care for clearing of C_RESET
         */
        if ((Urb->SetupData[1] == USB_R_CLEAR_FEATURE) &&
            (Changes == USB_FS_C_PORT_RESET)) {
            this->PortStatusChanges &= ~USB_ST_PORT_RESET;
        }

        Condition_InterruptSignal(Urb->Done);
        goto DoneOk;
    }

    /* Uhu?
     */
    DBGME(3,_tprintf(_T("Sl811RootHub noimp\n")));
    return E_NOT_IMPLEMENTED;

 DoneOk:
    Urb->Flags |= URB_FLAG_COMPLETE;
    return S_OK;
}

BOOL Previously = FALSE;

PRIVATE SCODE Sl811MasterSubmitUrb (
    PIUSBMASTERCONTROLLER pThis,
    /*inout*/ PUSBM_REQUEST_BLOCK Urb,
    UINT8 PortNumber)
{
    PSL811 this = pSl811Master(pThis);

    if (Urb == NULL) {
        Sl811Dump(this,FALSE);
        return E_INVALID_PARAMETER;
    }

    DBGME(0,_tprintf(_T("Sl811:Submit(%x,%x,%x)\n"),
                     (ADDRESS)pThis, (ADDRESS)Urb, PortNumber));
#if _DEBUG
if (Previously) {
_tprintf(_T("previously..."));
Sl811Dump(this,FALSE);
}
#endif

    SetLog(0);
    DBGME(1,_tprintf(_T("Sl811:Urb %x,%x,%x,%x\n"), Urb->DeviceAddress,
                     Urb->Pid, Urb->SetupData[1], Urb->SetupData[3]));
    

    UnusedParameter(PortNumber);

    /* Is this for the root hub?
     */
    if (Urb->DeviceAddress == 0xff)
        return Sl811RootHub(this,Urb);

    /* Find the script for it
     */
    if (Urb->Pid == USB_PID_SETUP)
    {
        assert(Urb->Flags & URB_FLAG_CONTROL);
        if (Urb->Flags & URB_FLAG_IN)
        {
            SetStartForPacket(Urb,ControlReadScript);
        } else
        if (Urb->Flags & URB_FLAG_OUT)
        {
            SetStartForPacket(Urb,ControlWriteScript);
        } else
        {
            SetStartForPacket(Urb,NoDataControlScript);
        }
    }
    else if (Urb->Pid == USB_PID_IN)
    {
        SetStartForPacket(Urb,ReadScript);
    }
    else if (Urb->Pid == USB_PID_OUT)
    {
        SetStartForPacket(Urb,WriteScript);
    }

    return SlStart(this,Urb);
}

PRIVATE SCODE Sl811MasterCancelUrb (
    PIUSBMASTERCONTROLLER pThis,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 PortNumber)
{
    PSL811 this = pSl811Master(pThis);
    ADDRESS *Target;

    if (Urb == this->CurrentPacket)
        Target = (ADDRESS*) &this->CurrentPacket;
    else
    if (Urb == this->EventsPacket)
        Target = (ADDRESS*) &this->EventsPacket;
    else
    {
        /* BUGBUG Walk the PendingPackets list */
        return E_OBJECT_NOT_FOUND;
    }

    if (AtomicCmpAndSwap(Target,(ADDRESS)Urb,0))
    {
        Urb->Flags |= URB_FLAG_ERROR;
        Condition_InterruptSignal(Urb->Done);
        return S_OK;
    }

    return E_OBJECT_NOT_FOUND;
}

#include "s_Sl811_v.c"

/* Constructor */
PIDEVICE Sl811New( ADDRESS RegisterBase )
{
    PIHEAP pHeap = CurrentHeap();

    PSL811 this = (PSL811) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *this, 0);
    if (!this)
        return NULL;

    this->CDevice.v = (struct CDeviceVtbl*)&Sl811DeviceSl811Vtbl;
    this->CDevice.RefCnt = 1;
    this->Sl811Masterv = &Sl811MasterSl811Vtbl;
    this->Sl = (PSL811REGS)RegisterBase;

#if 1
    /* HACK ALERT
     * The TAMU external proto board has interfacing issues, too much noise.
     * By experimentation it was found that the minimal error rate is for a
     * specific setting of the EBI's CS register.
     */
#if 0//before
    if (RegisterBase == 0x20000000)
        *(volatile UINT32 *)0xffe00008 = 0x20002126;
#else
    if (RegisterBase == 0x20000000)
        *(volatile UINT32 *)0xffe00008 = 0x20002326;
#endif
#endif

    /* Say we must poll first to know if a device is attached
     */
    this->PortStatusChanges = USB_ST_PORT_CONNECTION;

    return iSl811Device(this);
}

/* ** 
*/
SCODE __cdecl Sl811Constructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("Sl811Constructor(%x %s %x %x %d)\n"),
                     (ADDRESS) pThis, Name, (ADDRESS) ppUnk,
                     RegisterBase, Interrupt));

    pUnk = (PIUNKNOWN) Sl811New( RegisterBase );

    DBGME(0,_tprintf(_T("Sl811Constructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk)
    {
        *ppUnk = pUnk;
        return S_OK;
    } else
    {
        DBGME(3,_tprintf(_T("Sl811Constructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl Sl811DriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) Sl811Constructor
};

static struct _DRIVERCOB Sl811DriverCobObject = { &Sl811DriverCobVtable, 1 };

PIUNKNOWN Sl811CobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &Sl811DriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &Sl811DeviceSl811Vtbl, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &Sl811DriverCobObject;
}
