/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    sl811test.c
 *
 * Purpose:
 *
 *    Simple test for Cypress's SL811HS USB master.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */



#include <mmlite.h>
#include <tchar.h>
#include <stdio.h>
#include <drivers\usb.h>
#include <usbmaster.h>
#include <arm/at91m63200.h>
#include "sl811.h"

#define TheSl ((PSL811REGS)0x20000000)

PRIVATE INLINE UINT8 SlReadReg(PSL811REGS Sl, UINT8 Regno)
{
    Sl->Address = Regno;
    Delay(10);
    return Sl->Data;
}

PRIVATE INLINE void SlWriteReg(PSL811REGS Sl, UINT8 Regno, UINT8 Value)
{
    Sl->Address = Regno;
    Delay(10);
    Sl->Data = Value;
    Delay(10);
}

PRIVATE INLINE void SlReadSeq(PSL811REGS Sl, UINT8 Regno, UINT Count, UINT8 *Data)
{
    Sl->Address = Regno;
    while (Count-- > 0)
        *Data++ = Sl->Data;
}

PRIVATE INLINE void SlWriteSeq(PSL811REGS Sl, UINT8 Regno, UINT Count, UINT8 *Data)
{
    Sl->Address = Regno;
    while (Count-- > 0)
        Sl->Data = *Data++;
}

typedef _TCHAR *MYSTRING;

PRIVATE void fancyreg(int i, UINT8 c)
{
    int j;
    static struct {
        MYSTRING Regname;
        MYSTRING Bitname[8];
    } Regs[16] = {
    {_T("HCR_A"),
     { _T("Arm"), _T("Enable"), _T("Direction"), _T("r3"),
       _T("ISO"), _T("SOF"), _T("DToggle"), _T("Preamble") }},
    {_T("HBA_A"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("HBL_A"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("STATUS_A"),
     { _T("ACK"), _T("Error"), _T("Timeout"), _T("Sequence"),
       _T("Setup"), _T("Overflow"), _T("NAK"), _T("STALL")}},
    {_T("TC_A"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("CTRL1"),
     {_T("SOF_en"), _T("r1"), _T("r2"), _T("Reset"),
      _T("force_J"), _T("LowSpeed"), _T("Suspend"),_T("r7")}},
    {_T("INTEN"),
     {_T("DONE_A"), _T("DONE_B"), _T("r2"), _T("r3"),
      _T("SOF"),_T("Insert"),_T("Detect"),_T("D_PLUS"),}},
    {_T("SAR"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("HCR_B"),
     { _T("Arm"), _T("Enable"), _T("Direction"), _T("r3"),
       _T("ISO"), _T("SOF"), _T("DToggle"), _T("Preamble") }},
    {_T("HBA_B"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("HBL_B"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("STATUS_B"),
     { _T("ACK"), _T("Error"), _T("Timeout"), _T("Sequence"),
       _T("Setup"), _T("Overflow"), _T("NAK"), _T("STALL")}},
    {_T("TC_B"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("ISTATUS"),
     {_T("DONE_A"), _T("DONE_B"),_T("r2"), _T("r3"),
      _T("SOF"),_T("Insert"),_T("Detect"),_T("D_PLUS"),}},
    {_T("REV"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7") }},
    {_T("SOF_HI"),
     { _T("b0"), _T("b1"), _T("b2"), _T("b3"), _T("b4"), _T("b5"), _T("b6"), _T("b7")}}
    };

    _tprintf(_T("%8s %2x = "), Regs[i].Regname, c);
    for (j = 0; j < 8; j++)
        if (c & (1<<j))
            _tprintf(_T(" %s"), Regs[i].Bitname[j]);
    _tprintf(_T("\n"));
}

PRIVATE void fancyprint(UINT8 *Regvals)
{
    int i;
    for (i = 0; i < 16; i++)
        fancyreg(i,Regvals[i]);
}

PRIVATE int isdiff(UINT8 *a, UINT8 *b)
{
    int i;
    for (i = 0; i < 256; i++)
        if (a[i] != b[i])
            return i;
    return -1;
}

PRIVATE void MilliSleep(int mSecs)
{
    TIME t;
    Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(mSecs)));
    SleepUntil(t);
}

PRIVATE struct {
    int Flips;
    int IntrCnt;
    int SofCnt;
    int InsCnt;
    int DoneCnt;
    int Loopy;
} Stats = {-1,0,};

PRIVATE void print_stats(void)
{
    _tprintf(_T("--Counters:\n"));
    _tprintf(_T("\tFlips = %d\n"), Stats.Flips);
    _tprintf(_T("\tInterrupts = %d\n"), Stats.IntrCnt);
    _tprintf(_T("\tStartOfFrame = %d\n"), Stats.SofCnt);
    _tprintf(_T("\tInsertions = %d\n"), Stats.InsCnt);
    _tprintf(_T("\tA-done = %d\n"), Stats.DoneCnt);
    _tprintf(_T("\tLoopy = %d\n"), Stats.Loopy);
}

typedef struct _USB_PKT *PUSB_PKT;
typedef int ISR_RESULT;
#define ISR_CONTINUE 1
#define ISR_REPEAT   2
#define ISR_DONE     4
#define ISR_ERROR    8
typedef ISR_RESULT (*ISR_ACTION)(PSL811REGS Sl, PUSB_PKT Packet);

#define MAXSCRIPT 4

typedef const struct _SCRIPT {
    _TCHAR *Name;
    ISR_ACTION Action[MAXSCRIPT];
} SCRIPT, *PSCRIPT;

/* This wants to be a USBM_REQUEST_BLOCK, but C cannot initialize unions so.
 */
typedef struct _USB_PKT {
    PSCRIPT Script;
    int Step;
    UINT8 DeviceAddress;
    UINT8 EndpointAddress;
    UINT8 Pid;
    USBM_URB_FLAGS Flags;
    UINT8 SetupData[8];
    UINT8 SetupLength;
    UINT8 ActualDataLength;
    UINT8 ExpectedDataLength;
    UINT8 MaximumDataLength;
    UINT8 Data[256];
    UINT8 FinalStatus;
    UINT8 NAKs;
} USB_PKT;

PRIVATE void DumpPacket(PUSB_PKT Packet)
{
    int i, n;

    _tprintf(_T("Packet @x%x %s"), (ADDRESS)Packet,
           ((Packet->Flags & URB_FLAG_ERROR) ? _T("had error in") : 
            ((Packet->Flags & URB_FLAG_COMPLETE) ? _T("successfully completed") :_T("is executing"))));

    _tprintf(_T(" script @%x"), (ADDRESS) Packet->Script);
    if (Packet->Script)
        _tprintf(_T(" %s"), Packet->Script->Name);
    _tprintf(_T(" at step %d\n"), Packet->Step);

    _tprintf(_T("\tDeviceAddress %x\n"), Packet->DeviceAddress);
    _tprintf(_T("\tEndpointAddress %x\n"), Packet->EndpointAddress);
    _tprintf(_T("\tPid %x\n"), Packet->Pid);
    _tprintf(_T("\tFlags %x\n"), Packet->Flags);
    _tprintf(_T("\tSetupLength %x\n"), Packet->SetupLength);
    n = Packet->SetupLength;
    if (n)
    {
        _tprintf(_T("\tSetupData[%d]:"),n);
        for (i = 0; i < n; i++)
            _tprintf(_T("%2x "), Packet->SetupData[i]);
        _tprintf(_T("\n"));
    }
    _tprintf(_T("\tExpectedDataLength %x\n"), Packet->ExpectedDataLength);
    _tprintf(_T("\tMaximumDataLength %d\n"), Packet->MaximumDataLength);
    _tprintf(_T("\tActualDataLength %x\n"), Packet->ActualDataLength);
    n = Packet->ActualDataLength;
    if (n)
    {
        _tprintf(_T("\tResult[%d]:"),n);
        for (i = 0; i < n; i++)
            _tprintf(_T("%2x "), Packet->Data[i]);
        _tprintf(_T("\n"));
    }
    _tprintf(_T("\tNAKs %d\n"), Packet->NAKs);
    if (Packet->Flags & URB_FLAG_ERROR)
        _tprintf(_T("\tFinalStatus %x\n"), Packet->FinalStatus);
}

PRIVATE void DumpPacketResults(PUSB_PKT Packet, _TCHAR *Banner)
{
    int i;

    _tprintf(Banner);
    for (i = 0; i < Packet->ActualDataLength; i++)
        _tprintf(_T("%2x "), Packet->Data[i]);
    _tprintf(_T("\n"));
}


#define MAXLOG 32
#define DUMPSIZE (16+64)
typedef UINT8 MINIDUMP[DUMPSIZE];

PRIVATE struct {
    int Step;
    MINIDUMP Regs;
} Log[MAXLOG];
PRIVATE int CurL = 0;

typedef struct _USBD {
    PSL811REGS Sl;

    CONDITION SleepPoint;

    BOOL HasNewPacket;
    PUSB_PKT CurrentPacket;
    PUSB_PKT LastPacket;
    
    CONDITION InsertRemove;
    BOOL IsAttached;

    /* work in progress */
    struct {
        UINT8 Address;
        UINT8 ReportDescriptorLength;
        UINT8 InputReportLength;
        UINT8 OutputReportLength;
        BOOL  IsHid;
        USB_DEVICE_struct DeviceDescriptor;
    } Client;

} USBD, *PUSBD;


PUSBD UsbdNew(PSL811REGS Sl)
{
    PUSBD Usbd = malloc(sizeof(USBD));
    if (Usbd)
    {
        memset(Usbd,0,sizeof(USBD));
        Usbd->Sl = Sl;
        Condition_Init(&Usbd->SleepPoint);
        Condition_Init(&Usbd->InsertRemove);
    }
    return Usbd;
}

void UsbdDestroy(PUSBD Usbd)
{
    if (Usbd)
    {
        Condition_Broadcast(&Usbd->InsertRemove);
        Condition_Broadcast(&Usbd->SleepPoint);
        Condition_Destroy(&Usbd->SleepPoint);
        Condition_Destroy(&Usbd->InsertRemove);
        free(Usbd);
    }
}

PRIVATE void UsbdDump(PUSBD Usbd)
{
    PUSB_PKT Packet;

    Packet = Usbd->LastPacket;
    if (Packet != NULL)
    {
        _tprintf(_T("Last pkt used script '%s' for %d steps\n"), 
               Packet->Script->Name, Packet->Step+1);
        DumpPacket(Packet);
    }

    Packet = Usbd->CurrentPacket;
    if (Packet != NULL)
    {
        _tprintf(_T("Current pkt used script '%s' and executed %d steps\n"), 
               Packet->Script->Name, Packet->Step+1);
        DumpPacket(Packet);
    }

}

PRIVATE BOOL UsbdWait(PUSBD Usbd)
{
    SCODE sc;
    TIME Timeout;
    for (;;) {
        Int32ToInt64(Timeout, TIME_RELATIVE(TIME_SECONDS(3)));
        sc = Condition_Wait(&Usbd->SleepPoint, NULL, &Timeout);
        if (sc == E_TIMED_OUT)
            return FALSE;
        /* done ? */
        if (Usbd->CurrentPacket == NULL)
        {
            /* success? */
            return ((Usbd->LastPacket->Flags & URB_FLAG_ERROR) == 0);
        }
    }
}

BOOL SlIsr(void *This, PBOOL pNotMyInterrupt)
{
    PUSBD Usbd = (PUSBD)This;
    PSL811REGS Sl = Usbd->Sl;
    PUSB_PKT CurP = Usbd->CurrentPacket;
    UINT8 iStatus;
    BOOL Reschedule = FALSE, DoScript = FALSE;

    Stats.IntrCnt++;

    /* See whats up
     */
    iStatus = SlReadReg(Sl,SL_ISTATUS);

    /* Start-of-frame? Just ACK
     */
    if (iStatus & SL_INTEN_SOF)
    {
        DoScript = Usbd->HasNewPacket;
        Usbd->HasNewPacket = FALSE;
        Stats.SofCnt++;
    }

    /* Insertion detected?
     */
    if (iStatus & SL_INTEN_Insert)
    {
        Stats.InsCnt++;
        Reschedule = Condition_InterruptSignal(&Usbd->InsertRemove);
    }
    
    /* Command complete?
     */
    if (iStatus & SL_INTEN_DONE_A)
    {
        Stats.DoneCnt++;

        if (CurP && CurP->Script)
        {
            if (CurL < MAXLOG)
            {
                Log[CurL].Step = CurP->Step;
                SlReadSeq(Sl,0,DUMPSIZE,Log[CurL++].Regs);

                DoScript = TRUE;
            }
            else
            {
                CurP->Flags |= URB_FLAG_ERROR;
                CurP->FinalStatus = SlReadReg(Sl,SL_STATUS_A);

                Usbd->LastPacket = CurP;
                Usbd->CurrentPacket = NULL;
                Reschedule |= Condition_InterruptSignal(&Usbd->SleepPoint);
            }

        }
    }

    SlWriteReg(Sl,SL_ISTATUS,iStatus);

    if (DoScript)
    {
        ISR_RESULT r = CurP->Script->Action[CurP->Step](Sl,CurP);
        if (r & ISR_CONTINUE)
            CurP->Step++;
        else
        if ((r & ISR_REPEAT) == 0)
        {
            if (r == ISR_DONE)
                CurP->Flags |= URB_FLAG_COMPLETE;
            else
                CurP->Flags |= URB_FLAG_ERROR;

            Usbd->LastPacket = CurP;
            Usbd->CurrentPacket = NULL;
            Reschedule |= Condition_InterruptSignal(&Usbd->SleepPoint);
        }
    }

    return Reschedule;
}

PRIVATE int ResetLoop(PSL811REGS Sl)
{
    UINT8 c;
    int ret = 0;

    for (;;)
    {

    /* Disable interrupts
     */
    SlWriteReg(Sl,SL_INTEN, 0);

    /* Dis-Arm
     */
    SlWriteReg(Sl,SL_CTRL1, 0);

    MilliSleep(500);

    /* Clear any interrupts we might have generated
     */
    SlWriteReg(Sl, SL_ISTATUS, 0xFF);

    /* Check on the bus' connection status
     */
    c = SlReadReg(Sl,SL_ISTATUS);

    /* Nothing there?
     */
    if (c & SL_INTEN_Detect) 
    {
        ret = 0;
        _tprintf(_T("No device or no power\n"));
        return 0;
    }

    if ((c & SL_INTEN_D_PLUS ))
    {
        /* Full speed
         */
        SlWriteReg(Sl,SL_SOF_HI,SL_SOFH_1ms|SL_SOFH_Master);
        SlWriteReg(Sl,SL_SOF_LOW, SL_SOFL_1ms);
        SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_Reset|SL_CTRL1_EOF2);
        MilliSleep(10);
        SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_EOF2);
    } else 
    {

        /* Slow speed
         */
        SlWriteReg(Sl,SL_SOF_HI,SL_SOFH_1ms|SL_SOFH_Master|SL_SOFH_Polarity);
        SlWriteReg(Sl,SL_SOF_LOW, SL_SOFL_1ms);
        SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_Reset|SL_CTRL1_LowSpeed);
        MilliSleep(10);
        SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_LowSpeed);
    }
}

}

PRIVATE void InstallTheIsr(PUSBD Usbd)
{
    PioA->Disable = PIOA_IRQ1;
    PioA->FilterDisable = PIOA_IRQ1;
    (void) AddDevice(Usbd,(PTR) SlIsr,
                     DEVICE_FLAGS_LEVEL_TRIGGER |
                     DEVICE_FLAGS_POSITIVE_LEVEL,
                     30, NULL);
}

PRIVATE int SlReset(PSL811REGS Sl, PUSBD Usbd, BOOL InstallIsr)
{
    _TCHAR *m = NULL;
    UINT8 c;
    int ret = 0;

    /* Follow the recommended reset procedure
     */

    /* Disable interrupts
     */
    SlWriteReg(Sl,SL_INTEN, 0);

    /* Dis-Arm
     */
    SlWriteReg(Sl,SL_HCR_A, 0);
    SlWriteReg(Sl,SL_HCR_B, 0);
    SlWriteReg(Sl,SL_CTRL1, 0);

    MilliSleep(500);

    /* Clear any interrupts we might have generated
     */
    SlWriteReg(Sl, SL_ISTATUS, 0xFF);

    /* Check on the bus' connection status
     */
    c = SlReadReg(Sl,SL_ISTATUS);

    /* Nothing there?
     */
    if (c & SL_INTEN_Detect) 
    {
        ret = 0;
        m = _T("No device or no power");
        if (Usbd)
        {
            Usbd->Client.Address = 0;
            Usbd->IsAttached = FALSE;
        }
    }
    else 
    {
        if (Usbd)
        {
            Usbd->Client.Address = 0;
            Usbd->IsAttached = TRUE;
        }

        if (!(c & SL_INTEN_D_PLUS ))
        {	
            /* Slow speed
             */
            SlWriteReg(Sl,SL_SOF_HI,SL_SOFH_1ms|SL_SOFH_Master|SL_SOFH_Polarity);
            SlWriteReg(Sl,SL_SOF_LOW, SL_SOFL_1ms);
            SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_Reset|SL_CTRL1_LowSpeed);
            MilliSleep(10);
            SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_LowSpeed);

            ret = 1;
            m = _T("Low speed device connected");
        }
        else
        {
            /* Full speed
             */
            SlWriteReg(Sl,SL_SOF_HI,SL_SOFH_1ms|SL_SOFH_Master);
            SlWriteReg(Sl,SL_SOF_LOW, SL_SOFL_1ms);
            SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_Reset|SL_CTRL1_EOF2);
            MilliSleep(10);
            SlWriteReg(Sl,SL_CTRL1,SL_CTRL1_SOF_en|SL_CTRL1_EOF2);

            ret = 2;
            m = _T("Full speed device connected");
        }

        /* start the SOF or EOP */
        SlWriteReg(Sl,SL_HPID_A, SL_MAKE_HPID(SL_HPID_SOF,0));
        SlWriteReg(Sl,SL_HDA_A, 0);
        SlWriteReg(Sl,SL_HCR_A, SL_HCR_Arm);
    }

    MilliSleep(25);

    /* Clear any pending interrupts
     */
    SlWriteReg(Sl,SL_ISTATUS, 0xFF);

    /* Connect our ISR
     */
    if (InstallIsr)
        InstallTheIsr(Usbd);

    /* Enable them interrupts
     */
    SlWriteReg(Sl,SL_INTEN,
               SL_INTEN_DONE_A |
               SL_INTEN_SOF |
               SL_INTEN_Insert );

    /* Say what happened
     */
    fancyreg(SL_ISTATUS,c);
    _tprintf(_T("%s\n"), m);

    return ret;
}

PRIVATE void SlDestroy(PSL811REGS Sl)
{
    SlWriteReg(Sl,SL_INTEN,0);
    SlWriteReg(Sl, SL_ISTATUS, 0xFF);
    RemoveDevice((PTR)Sl,30);
}

PRIVATE UINT8 b1[256], b2[256];

PRIVATE void fancydump(PSL811REGS Sl)
{
    SlReadSeq(Sl,0,256,b1);
    fancyprint(b1);
}

PRIVATE MUTEX SlLock;
PRIVATE BOOL blimp = FALSE;

PRIVATE int loopy(PSL811REGS Sl)
{
    UINT8 iStatus, iPrevious, iMask;
    int n = 0;

    SlReset(Sl, NULL, FALSE);

    iMask = SlReadReg(Sl,SL_INTEN);
    iPrevious = 0x66;
    while (!blimp)
    {
        Mutex_Lock(&SlLock);

        Stats.Loopy++;
        iStatus = SlReadReg(Sl,SL_ISTATUS);

        if (iStatus & 0x33/*& iMask*/)
        {
            if (iStatus & SL_INTEN_DONE_A)
                fancydump(Sl);
            SlIsr(Sl,NULL);
            Delay(1);

            if ((++n & 0xfff) == 0)
                print_stats();
        }

        if (iStatus != iPrevious)
        {
            Stats.Flips++;
            iPrevious = iStatus;
        }

        Mutex_Unlock(&SlLock);

    }

    SlDestroy(Sl);

    return 0;
}

PRIVATE BOOL SlStart(PUSBD Usbd, PUSB_PKT Packet)
{
    /* Not if we have something going on
     */
    if (Usbd->CurrentPacket)
        return FALSE;

    Usbd->CurrentPacket = Packet;

    CurL = 0;

    Packet->DeviceAddress = Usbd->Client.Address;

    Packet->Flags &= ~URB_FLAG_STATUS;
    Packet->ActualDataLength = 0;
    Packet->MaximumDataLength = Packet->ExpectedDataLength;
    Packet->Step = 0;
    Packet->NAKs = 0;
    Packet->FinalStatus = 0;

    /* We cannot start a new packet on top of the SOF one, which
     * the chip automatically generates for us every 1ms.
     * There is a bit in the HCR to handle this, but unfortunately
     * it does not really work (see erratas).
     */

    /* Start the script (from the ISR) at the next SOF interrupt
     * NB: Conceivably, we might test how much time is left in the
     *     ~~1ms slot between SOF packets and see if we can slip
     *     the SOP in there.
     */
    Usbd->HasNewPacket = TRUE;

    return TRUE;
}

/* Script helpers
 */
PRIVATE ISR_RESULT SH_InitIn(PSL811REGS Sl, PUSB_PKT Packet)
{
    UINT8 l = Packet->MaximumDataLength;
    UINT8 ctl;

    /* Build the value to stick into the ctl register
       */
    ctl = (SlReadReg(Sl,SL_HCR_A) & SL_HCR_DToggle) ^ SL_HCR_DToggle;
    ctl |= SL_HCR_Arm | SL_HCR_Enable;

    /* Keep buffer pointer to start of usable RAM
     */

    /* Length
     */
    SlWriteReg(Sl,SL_HBL_A,l);

    /* Set PID for IN phase, endpoint address
     */
    SlWriteReg(Sl,SL_HPID_A,
               SL_MAKE_HPID(SL_HPID_IN,Packet->EndpointAddress));

    /* Go
     */
    SlWriteReg(Sl,SL_HCR_A,ctl);

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SH_InitOut(PSL811REGS Sl, PUSB_PKT Packet)
{
    UINT8 l = Packet->MaximumDataLength;
    UINT8 ctl;

    /* Sanity
     */
    if (l > SL_HBL_MAX)
        l = SL_HBL_MAX;

    /* Build the value to stick into the ctl register
       */
    ctl = SlReadReg(Sl,SL_HCR_A);
    ctl |= SL_HCR_Arm | SL_HCR_Enable | SL_HCR_Out;

    /* Keep buffer pointer to start of usable RAM
     */

    /* We got something to say
     */
    if (l)
    {
        UINT8 bufaddr = 0x10; 
        SlWriteSeq(Sl,bufaddr,l,Packet->Data+Packet->ActualDataLength);
    }

    /* Length
     */
    SlWriteReg(Sl,SL_HBL_A,l);

    /* Set PID for OUT phase, endpoint address
     */
    SlWriteReg(Sl,SL_HPID_A,
               SL_MAKE_HPID(SL_HPID_OUT,Packet->EndpointAddress));

    /* Go
     */
    SlWriteReg(Sl,SL_HCR_A,ctl);

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SH_InitStatusOut(PSL811REGS Sl, PUSB_PKT Packet)
{
    UINT8 ctl;

    /* Build the value to stick into the ctl register
       */
    ctl = SL_HCR_Arm | SL_HCR_Enable | SL_HCR_Out | SL_HCR_DToggle;

    /* Keep buffer pointer to start of usable RAM
     */

    /* Length
     */
    SlWriteReg(Sl,SL_HBL_A,0);

    /* Set PID for IN phase, endpoint address
     */
    SlWriteReg(Sl,SL_HPID_A,
               SL_MAKE_HPID(SL_HPID_OUT,Packet->EndpointAddress));

    /* Go
     */
    SlWriteReg(Sl,SL_HCR_A,ctl);

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SH_CheckErrors(PSL811REGS Sl, PUSB_PKT Packet, ISR_ACTION RetryFunction)
{
    UINT8 Status;;

    Status = SlReadReg(Sl,SL_STATUS_A);

    /* Everything ok?
     */
    if (0 == (Status & (SL_STATUS_NAK |
                        SL_STATUS_Error |
                        SL_STATUS_Timeout |
                        SL_STATUS_STALL |
                        SL_STATUS_Overflow)))
    {
        return ISR_CONTINUE;
    }

    /* Just a glitch?
     */
    if (Status & SL_STATUS_NAK)
    {
        Packet->NAKs++;
        (void) RetryFunction(Sl,Packet);
        return ISR_REPEAT;
    }

    /* Something serious, bail out
     */
    Packet->FinalStatus = Status;
    return ISR_ERROR;
}

PRIVATE ISR_RESULT SH_InitStatusIn(PSL811REGS Sl, PUSB_PKT Packet)
{
    UINT8 ctl;

    /* Build the value to stick into the ctl register
       */
    ctl = SL_HCR_Arm | SL_HCR_Enable | SL_HCR_DToggle;

    /* Keep buffer pointer to start of usable RAM
     */

    /* Length
     */
    SlWriteReg(Sl,SL_HBL_A,0);

    /* Set PID for IN phase, endpoint address
     */
    SlWriteReg(Sl,SL_HPID_A,
               SL_MAKE_HPID(SL_HPID_IN,Packet->EndpointAddress));

    /* Go
     */
    SlWriteReg(Sl,SL_HCR_A,ctl);

    return ISR_CONTINUE;
}


/* Script actions
 */
PRIVATE ISR_RESULT SA_InitSetup(PSL811REGS Sl, PUSB_PKT Packet)
{
    UINT8 l = Packet->SetupLength;
    UINT8 ctl, bufaddr;

    /* Sanity
     */
    if (l > SL_HBL_MAX)
        l = SL_HBL_MAX;

    /* Build the value to stick into the ctl register
       */
    ctl = (l) ? SL_HCR_Out : SL_HCR_DToggle;
    ctl |= SL_HCR_Arm | SL_HCR_Enable;

    /* isochronous?
     */
    if (Packet->Flags & URB_FLAG_ISOCHRONOUS)
        ctl |= SL_HCR_ISO;

    /* preamble?
     */
    if ((Packet->Flags & URB_FLAG_LOW_SPEED) && 0)//first hub is low speed
        ctl |= SL_HCR_Preamble;

    /* Toggle (?!?!?)
       */
    if (0)
        ctl |= SL_HCR_DToggle;

    /* Set buffer pointer to start of usable RAM
     * BUGBUG Fix to use both A&B at once
     */
    bufaddr = 0x10; 
    SlWriteReg(Sl,SL_HBA_A,bufaddr);

    /* If we got something to say
     */
    if (l)
    {
        SlWriteSeq(Sl,bufaddr,l,Packet->SetupData);
    }

    /* Length
     */
    SlWriteReg(Sl,SL_HBL_A,l);

    /* Set PID, endpoint address, device address
     */
    SlWriteReg(Sl,SL_HPID_A,
               SL_MAKE_HPID(Packet->Pid,Packet->EndpointAddress));
    SlWriteReg(Sl,SL_HDA_A, (UINT8)(Packet->DeviceAddress & SL_HDA_MASK));

    /* Go
     */
    SlWriteReg(Sl,SL_HCR_A,ctl);

    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SA_InitIn(PSL811REGS Sl, PUSB_PKT Packet)
{
    ISR_RESULT r;

    // BUGBUG is the Retry right?
    r = SH_CheckErrors(Sl,Packet,SH_InitIn);
    if (r != ISR_CONTINUE)
        return r;

    return SH_InitIn(Sl,Packet);
}

PRIVATE ISR_RESULT SA_InitOut(PSL811REGS Sl, PUSB_PKT Packet)
{
    ISR_RESULT r;

    // BUGBUG is the Retry right?
    r = SH_CheckErrors(Sl,Packet,SH_InitOut);
    if (r != ISR_CONTINUE)
        return r;

    return SH_InitOut(Sl,Packet);
}

PRIVATE ISR_RESULT SA_InitStatusIn(PSL811REGS Sl, PUSB_PKT Packet)
{
    ISR_RESULT r;

    // BUGBUG is the Retry right?
    r = SH_CheckErrors(Sl,Packet,SA_InitSetup);
    if (r != ISR_CONTINUE)
        return r;

    return SH_InitStatusIn(Sl,Packet);
}

PRIVATE ISR_RESULT SA_StoreBytes(PSL811REGS Sl, PUSB_PKT Packet)
{
    UINT8 Status, Address, Length;

    Address = SlReadReg(Sl,SL_HBA_A);//bugbug readseq
    Length = SlReadReg(Sl,SL_HBL_A) - SlReadReg(Sl,SL_TC_A);

    if (Length)
    {
        SlReadSeq(Sl,Address,Length,Packet->Data+Packet->ActualDataLength);
        Packet->ActualDataLength += Length;

        Packet->MaximumDataLength -= Length;
    }

    /* Ask for more iff...
     */
    Status = SlReadReg(Sl,SL_STATUS_A);
    if ((0 == (Status & (SL_STATUS_NAK |
                         SL_STATUS_Error |
                         SL_STATUS_Timeout |
                         SL_STATUS_STALL |
                         SL_STATUS_Overflow))) &&
        (Packet->MaximumDataLength) &&
        Length)
    {
        (void) SH_InitIn(Sl,Packet);
        return ISR_REPEAT;
    }

    return ISR_DONE;
}

PRIVATE ISR_RESULT SA_StoreControlBytes(PSL811REGS Sl, PUSB_PKT Packet)
{
    ISR_RESULT r;

    r = SH_CheckErrors(Sl,Packet,SH_InitIn);
    if (r != ISR_CONTINUE)
        return r;

    r = SA_StoreBytes(Sl,Packet);
    if (r != ISR_DONE)
        return r;

    SH_InitStatusOut(Sl,Packet);
    return ISR_CONTINUE;
}

/* Special case for varsized descriptors, with length as the first byte in
 * Some devices STALL if we ask for more, and that we normally think of as
 * an error.
 */
PRIVATE ISR_RESULT SA_GDStoreControlBytes(PSL811REGS Sl, PUSB_PKT Packet)
{
    UINT8 Address, Length;
    ISR_RESULT r;

    r = SH_CheckErrors(Sl,Packet,SH_InitIn);
    if (r != ISR_CONTINUE)
        return r;

    Address = SlReadReg(Sl,SL_HBA_A);//bugbug readseq
    Length  = SlReadReg(Sl,SL_HBL_A) - SlReadReg(Sl,SL_TC_A);

    if (Length)
    {
        SlReadSeq(Sl,Address,Length,Packet->Data+Packet->ActualDataLength);
        Packet->ActualDataLength += Length;

        //NOTE: This is where we do the length adjustment
        if (Packet->Data[0] < Packet->ExpectedDataLength)
        {
            /* Only happens on first burst in
             */
            Packet->ExpectedDataLength = Packet->Data[0];
            Packet->MaximumDataLength = Packet->Data[0];
        }

        Packet->MaximumDataLength -= Length;

        if (Packet->MaximumDataLength)
        {
            (void) SH_InitIn(Sl,Packet);
            return ISR_REPEAT;
        }
    }

    SH_InitStatusOut(Sl,Packet);
    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SA_LoadBytes(PSL811REGS Sl, PUSB_PKT Packet)
{
    UINT8 Address, Length;
    ISR_RESULT r;

    r = SH_CheckErrors(Sl,Packet,SH_InitOut);
    if (r != ISR_CONTINUE)
        return r;

    Address = SlReadReg(Sl,SL_HBA_A);//bugbug readseq
    Length = SlReadReg(Sl,SL_HBL_A) - SlReadReg(Sl,SL_TC_A);

    Packet->ActualDataLength += Length;
    Packet->MaximumDataLength -= Length;

    if (Packet->MaximumDataLength)
    {
        UINT8 ctl;

        /* Toggle, do NOT go yet */
        ctl = (SlReadReg(Sl,SL_HCR_A) & SL_HCR_DToggle) ^ SL_HCR_DToggle;
        SlWriteReg(Sl,SL_HCR_A,ctl);

        (void) SH_InitOut(Sl,Packet);
        return ISR_REPEAT;
    }

    SH_InitStatusIn(Sl,Packet);
    return ISR_CONTINUE;
}

PRIVATE ISR_RESULT SA_ReadDone(PSL811REGS Sl, PUSB_PKT Packet)
{
    ISR_RESULT r;

    r = SH_CheckErrors(Sl,Packet,SH_InitStatusOut);
    if (r != ISR_CONTINUE)
        return r;

    return ISR_DONE;
}

PRIVATE ISR_RESULT SA_WriteDone(PSL811REGS Sl, PUSB_PKT Packet)
{
    ISR_RESULT r;

    r = SH_CheckErrors(Sl,Packet,SH_InitStatusIn);
    if (r != ISR_CONTINUE)
        return r;

    return ISR_DONE;
}

PRIVATE SCRIPT NoDataControlScript = {
    _T("No Data Control"),
    { SA_InitSetup, SA_InitStatusIn, SA_WriteDone, NULL }
};
PRIVATE SCRIPT ControlReadScript = {
    _T("Control Read"),
    { SA_InitSetup, SA_InitIn, SA_StoreControlBytes, SA_ReadDone }
};
/* special case of the above */
PRIVATE SCRIPT GetDescriptorScript = {
    _T("Get Descriptor"),
    { SA_InitSetup, SA_InitIn, SA_GDStoreControlBytes, SA_ReadDone }
};
PRIVATE SCRIPT ControlWriteScript = {
    _T("Control Write"),
    { SA_InitSetup, SA_InitOut, SA_LoadBytes, SA_WriteDone }
};
PRIVATE SCRIPT ReadScript = {
    _T("Read Data"),
    { SH_InitIn, SA_StoreBytes, NULL, NULL }
};

PRIVATE USB_PKT p1 = {
    &GetDescriptorScript, 0,
    0,0,USB_PID_SETUP, URB_FLAG_IN | URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
    {0x80,USB_R_GET_DESCRIPTOR/*6*/,  0,USB_DT_DEVICE/*1*/,  0,0, 18, 0},8, 0, 
    18, 0, {0,},
};

PRIVATE USB_PKT p2 = {
    &NoDataControlScript, 0,
    0,0,USB_PID_SETUP, URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
    {0,USB_R_SET_ADDRESS/*5*/,  1,0,  0,0,  0,0}, 8, 0, 
    0, 0, {0,}
};

PRIVATE USB_PKT p3 = {
    &ControlReadScript, 0,
    0,0,USB_PID_SETUP, URB_FLAG_IN | URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
#define xx 0x22 //9
    {0x80,USB_R_GET_DESCRIPTOR/*6*/,  0,USB_DT_CONFIG/*2*/,  0,0, xx, 0}, 8, 0, 
    xx, 0, {0,},
#undef xx
};

PRIVATE USB_PKT p4 = {
    &GetDescriptorScript, 0,
    0,0,USB_PID_SETUP, URB_FLAG_IN | URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
    {0x80,USB_R_GET_DESCRIPTOR/*6*/,  0,USB_DT_STRING/*3*/,  0,0, 255, 0}, 8, 0, 
    255, 0, {0,},
};

PRIVATE USB_PKT p5 = {
    &ControlReadScript, 0,
    0,0,USB_PID_SETUP, URB_FLAG_IN | URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
    {0x80,USB_R_GET_STATUS/*0*/,  0,0,  0,0, 2, 0}, 8, 0, 
    2, 0, {0,},
};

PRIVATE USB_PKT p6 = {
    &NoDataControlScript, 0,
    0,0,USB_PID_SETUP, URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
    {0,USB_R_SET_CONFIG/*9*/,  1,0,  0,0,  0,0}, 8, 0, 
    0, 0, {0,}
};

PRIVATE USB_PKT p7 = {
    &NoDataControlScript, 0,
    0,0x81/*EndpointAddress 0?*/,USB_PID_SETUP, URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
    {0x21,USB_R_SET_IDLE/*a*/,  0,0,  0,0,  0,0}, 8, 0, 
    0, 0, {0,}
};

PRIVATE USB_PKT p8 = {
    &ControlReadScript, 0,
    0,0,USB_PID_SETUP, URB_FLAG_IN | URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
#define zz 0x4a //9
    {0x81,USB_R_GET_DESCRIPTOR/*6*/,  0,USB_DT_REPORT/*x22*/,  0,0, zz, 0}, 8, 0, 
    zz, 0, {0,},
};

PRIVATE USB_PKT p9 = {
    &ControlReadScript, 0,
    0,0,USB_PID_SETUP, URB_FLAG_IN | URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
    {0xa1,USB_R_GET_REPORT/*1*/,  0,USB_RP_INPUT/*1*/,  0,0, 4, 0}, 8, 0, 
    4, 0, {0,},
};

PRIVATE USB_PKT p10 = {
    &ControlWriteScript, 0,
    0,0,USB_PID_SETUP, URB_FLAG_OUT | URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
#define pp 1
    {0x21,USB_R_SET_REPORT/*9*/,  0,USB_RP_OUTPUT/*2*/,  0,0, pp, 0}, 8, 0, 
    pp, 0, {0,},
#undef pp
};

PRIVATE USB_PKT p11 = {
    &ReadScript, 0,
    0,0x81,USB_PID_IN, URB_FLAG_IN | URB_FLAG_LOW_SPEED,
    {0,}, 0, 0, 
    4, 0, {0,},
};

PRIVATE USB_PKT p12 = {
    &GetDescriptorScript, 0,
    0,0,USB_PID_SETUP, URB_FLAG_IN | URB_FLAG_CONTROL | URB_FLAG_LOW_SPEED,
    {0x80,USB_R_GET_DESCRIPTOR/*6*/,  0,USB_DT_QUALIF/*1*/,  0,0, 10, 0}, 8, 0, 
    10, 0, {0,},
};

/* Avoid code duplications between command loop at PnP thread
 */
PRIVATE BOOL SlGetDeviceDescriptor(PUSBD Usbd)
{
    SlStart(Usbd,&p1);
    if (UsbdWait(Usbd))
    {
        memcpy((void *)&Usbd->Client.DeviceDescriptor,
               p1.Data,
               sizeof Usbd->Client.DeviceDescriptor);
        DumpPacketResults(&p1,_T("DeviceDescriptor:"));
        return TRUE;
    }

    _tprintf(_T("Failed to GET_DESCRIPTOR(device)\n"));
    return FALSE;
}

PRIVATE BOOL SlGetDeviceQualifier(PUSBD Usbd)
{
    SlStart(Usbd,&p12);
    if (UsbdWait(Usbd))
    {
        DumpPacketResults(&p1,_T("DeviceQualifier:"));
        return TRUE;
    }

    _tprintf(_T("Failed to GET_DESCRIPTOR(USB2.0 device qualifier)\n"));
    return FALSE;
}

PRIVATE BOOL SlSetDeviceAddress(PUSBD Usbd)
{
    UINT8 c;

    SlStart(Usbd,&p2);
    if (UsbdWait(Usbd))
    {
        c = Usbd->Client.Address;
        Usbd->Client.Address = p2.SetupData[2];
        _tprintf(_T("Client address is now x%x (was x%x)\n"),
               Usbd->Client.Address, c);
        return TRUE;
    }

    _tprintf(_T("Failed to SET_ADDRESS\n"));
    return FALSE;
}

PRIVATE BOOL SlGetConfigurationDescriptor(PUSBD Usbd)
{
    SlStart(Usbd,&p3);
    if (UsbdWait(Usbd))
    {
        int i;
        DumpPacketResults(&p3,_T("ConfigDescriptor:"));

        /* Look for the HID report descriptor length, if any */
        for (i = 0; i < p3.ActualDataLength; i++) {
            if (p3.Data[i+1] == USB_DT_HID) {
                Usbd->Client.IsHid = TRUE;
                Usbd->Client.ReportDescriptorLength = p3.Data[i+7];
                _tprintf(_T("ReportDescLength: %d\n"),Usbd->Client.ReportDescriptorLength);
                break;
            }
            i += p3.Data[i];
        }
        _tprintf(_T("MaxPower: %d mA\n"),p3.Data[8]*2);
        return TRUE;
    }

    _tprintf(_T("Failed to GET_CONFIG\n"));
    return FALSE;
}

PRIVATE BOOL SlGetStringDescriptors(PUSBD Usbd)
{
    BOOL GotSome = FALSE;
    int i;

    for (i = 0; i < 128; i++)
    {
        p4.SetupData[2] = i;
        /* LANGID? */
        if (i == 0)
        {
            p4.SetupData[4] = 0;
            p4.SetupData[5] = 0;
            p4.SetupData[6] = 4;
            p4.ExpectedDataLength = 4;
        }
        else
        {
            p4.SetupData[4] = 9;
            p4.SetupData[5] = 4;
            p4.SetupData[6] = 255;
            p4.ExpectedDataLength = 255;
        }

        SlStart(Usbd,&p4);
        if (UsbdWait(Usbd))
        {
            _tprintf(_T("[%d]"), i);
            DumpPacketResults(&p4,_T("GetString:"));
            if (i)
            {
                int j, n = p4.Data[0];
                _tprintf(_T("\t\""));
                for (j = 2; j < n; j += 2)
                    _puttchar(p4.Data[j]);
                _tprintf(_T("\"\n"));
            }
            GotSome = TRUE;
        }
        else
            break;
    }

    if (GotSome)
    {
        /* Try once more for MSFT's "OS" string thing
         */
        p4.SetupData[2] = 0xee;
        p4.SetupData[4] = 0;
        p4.SetupData[5] = 0;
        p4.SetupData[6] = 18;
        p4.ExpectedDataLength = 18;

        SlStart(Usbd,&p4);
        if (UsbdWait(Usbd))
        {
            _tprintf(_T("[%d]"), 0xee);
            DumpPacketResults(&p4,_T("GetString OS:"));
            {
                int j, n = p4.Data[0];
                _tprintf(_T("\t\""));
                for (j = 2; j < n; j += 2)
                    _puttchar(p4.Data[j]);
                _tprintf(_T("\"\n"));
            }
        }

        return TRUE;
    }
    _tprintf(_T("Failed to GET any STRING_DESC\n"));
    return FALSE;
}

PRIVATE BOOL SlGetStatus(PUSBD Usbd)
{
    SlStart(Usbd,&p5);
    if (UsbdWait(Usbd))
    {
        DumpPacketResults(&p5,_T("GetStatus:"));
        _tprintf(_T("Status: x%x\n"),
               (p5.Data[1] << 8) | p5.Data[0]);
        return TRUE;
    }

    _tprintf(_T("Failed to GET_STATUS\n"));
    return FALSE;
}

PRIVATE BOOL SlSetConfiguration(PUSBD Usbd)
{
    SlStart(Usbd,&p6);
    if (UsbdWait(Usbd))
    {
        DumpPacketResults(&p6,_T("SetConfig."));
        return TRUE;
    }

    _tprintf(_T("Failed to SET_CONFIG\n"));
    return FALSE;
}

PRIVATE BOOL SlSetIdle(PUSBD Usbd)
{
    SlStart(Usbd,&p7);
    if (UsbdWait(Usbd))
    {
        DumpPacketResults(&p7,_T("SetIdle:"));
        return TRUE;
    }

    _tprintf(_T("Failed to SET_IDLE\n"));
    return FALSE;
}

PRIVATE BOOL ParseHidReportDescriptor(PUSBD Usbd, UINT8 *Descriptor, UINT8 Length);

PRIVATE BOOL SlGetReportDescriptor(PUSBD Usbd)
{
    UINT8 Length = zz;

    if (Usbd->Client.ReportDescriptorLength)
        Length = Usbd->Client.ReportDescriptorLength;

    p8.SetupData[6] = Length;
    p8.ExpectedDataLength = Length;

    SlStart(Usbd,&p8);
    if (UsbdWait(Usbd))
    {
        DumpPacketResults(&p8,_T("ReportDescriptor:"));
        ParseHidReportDescriptor(Usbd,p8.Data,p8.ActualDataLength);
        return TRUE;
    }

    _tprintf(_T("Failed to GET_REPORT_DESCRIPTOR\n"));
    return FALSE;
}

PRIVATE BOOL SlGetReport(PUSBD Usbd)
{
    SlStart(Usbd,&p9);
    if (UsbdWait(Usbd))
    {
        DumpPacketResults(&p9,_T("Report:"));
        return TRUE;
    }

    _tprintf(_T("Failed to GET_REPORT\n"));
    return FALSE;
}

PRIVATE BOOL SlSetReport(PUSBD Usbd)
{
    if (Usbd->Client.OutputReportLength)
        p11.ExpectedDataLength = Usbd->Client.OutputReportLength;

    /* Try all combos of NUM(1) CAPS(2) SCROLL(4)
     */
    if (++(p10.Data[0]) == 8)
        p10.Data[0] = 0;

    SlStart(Usbd,&p10);
    if (UsbdWait(Usbd))
    {
        DumpPacketResults(&p10,_T("SetReport."));
        return TRUE;
    }

    _tprintf(_T("Failed to SET_REPORT\n"));
    return FALSE;
}

PRIVATE BOOL SlReadReport(PUSBD Usbd, BOOL Verbose)
{
    if (Usbd->Client.InputReportLength)
        p11.ExpectedDataLength = Usbd->Client.InputReportLength;

    SlStart(Usbd,&p11);
    if (UsbdWait(Usbd))
    {
        if (Verbose || p11.ActualDataLength)
            DumpPacketResults(&p11,_T("ReadReport:"));
        return TRUE;
    }

    if (Verbose)
        _tprintf(_T("Failed to INput REPORT\n"));
    return FALSE;
}

/*
 */
PRIVATE BOOL ParseHidReportDescriptor(PUSBD Usbd, UINT8 *Descriptor, UINT8 Length)
{
    int i;
    int InSize, OutSize, FeaSize;
    int Rc, Rs;
    UINT8 c;

    FeaSize = InSize = OutSize = 0;
    Rc = Rs = 0;
    for (i = 0; i < Length; i++)
    {
        c = Descriptor[i];
        i += c & 3;
        if (c == 0x75)
            Rs = Descriptor[i];
        if (c == 0x95)
            Rc = Descriptor[i];
        if (c == 0x81)
            InSize += (Rs * Rc);
        if (c == 0x91)
            OutSize += (Rs * Rc);
        if (c == 0xb1)
            FeaSize += (Rs * Rc);
    }
    Usbd->Client.InputReportLength = InSize / 8;
    Usbd->Client.OutputReportLength = OutSize / 8;
    _tprintf(_T("ReportLengths: InSize %d OutSize %d Feature %d\n"),
           InSize,OutSize, FeaSize);
    return TRUE;
}

/*
 */
PRIVATE int HidReadThread(PUSBD Usbd)
{
    _tprintf(_T("Listener thread started.\n"));

    while (!blimp && Usbd->IsAttached)
    {
        SlReadReport(Usbd,FALSE);

        MilliSleep(100);
    }

    _tprintf(_T("Listener thread done.\n"));
    return 0;
}

PRIVATE int PnpThread(PUSBD Usbd)
{
    SCODE sc;

    while (!blimp)
    {
        sc = Condition_Wait(&Usbd->InsertRemove,NULL,NULL);
        if (blimp)
            break;

        SlReset(Usbd->Sl,Usbd,FALSE);

        if (!Usbd->IsAttached)
        {
            _tprintf(_T("Device removal detected\n"));
            memset(&Usbd->Client,0,sizeof Usbd->Client);
            /* Abort any active packets
             */
            if (Usbd->CurrentPacket) {
                Condition_Signal(&Usbd->SleepPoint);
            }
            continue;
        }
        _tprintf(_T("Device attachment detected\n"));

        /* Get the device descriptor
         */
        if (!SlGetDeviceDescriptor(Usbd))
            continue;

        /* Set the device's address
         */
        if (!SlSetDeviceAddress(Usbd))
            continue;

        /* Get the configuration descriptor
         */
        if (!SlGetConfigurationDescriptor(Usbd))
            continue;

        /* Is it v2.0 ?
         */
        if (Usbd->Client.DeviceDescriptor.BcdUSB[1] == 2)
            (void) SlGetDeviceQualifier(Usbd);

        /* Rest is HID-specific 
         */
        if (!Usbd->Client.IsHid) {
            _tprintf(_T("%x.%x,%x.%x is not a HID.\n"),
                   Usbd->Client.DeviceDescriptor.IdVendor[0],
                   Usbd->Client.DeviceDescriptor.IdVendor[1],
                   Usbd->Client.DeviceDescriptor.IdProduct[0],
                   Usbd->Client.DeviceDescriptor.IdProduct[1]);
            continue;
        }

        /* Select the default configuration
         */
        if (!SlSetConfiguration(Usbd))
            continue;

        /* Get the report descriptor
         */
        if (!SlGetReportDescriptor(Usbd))
            continue;

        /* Just for aesthetic, get the strings
         * No prob if this fails.
         */
        (void) SlGetStringDescriptors(Usbd);

#if 0
        /* Set the idle
         * No idea why this always fails?
         */
        if (!SlSetIdle(Usbd))
            continue;

#endif

    }

    return 0;
}

/* Print and optionally save the content of the chip's RAM
 */
PRIVATE void DumpRam(PSL811REGS Sl, UINT8 *Where)
{
    int i;
    UINT8 c;
    BOOL Dumped = FALSE;

    for (i = 0; i < 0x100; i++)
    {
        c = SlReadReg(Sl,(UINT8)i);
        if ((i & 15) == 0)
            _tprintf(_T("\nMem[%03d]: "), i);
        _tprintf(_T("%2x "), c);
        if (Where)
            Where[i] = c;
    }
    _tprintf(_T("\n"));

    /* Repeat the dump a few times, sequentially, to check
     * (a) that we get the same results (b) with the faster read speeds
     */
    if (Where)
    for (c = 0; c < 10; c++)
    {
        SlReadSeq(Sl,0,256,b2);
        i = isdiff(Where,b2);
        if (i >= 0)
        {
            _tprintf(_T("%d-th dump repetition failed (at byte %d)\n"),c,i);
            if (!Dumped)
            {
                Dumped = TRUE;

                for (i = 0; i < 0x100; i++)
                {
                    c = b2[i];
                    if ((i & 15) == 0)
                        _tprintf(_T("\nReg%3d: "), i);
                    _tprintf(_T("%2x "), c);
                }
                _tprintf(_T("\n"));
            }
        }
    }

}

/* Destructively tests if we can write/verify access to the chip's RAM
 * Returns TRUE if the RAM looks ok to us
 */
PRIVATE BOOL RamTest(PSL811REGS Sl, int Verbosely)
{
    int i, j, n, fail = 0;
    UINT8 c;

    /* Bytes 13-15 dont "take" on a working chip
     */
#define RamTestFirstByte 0x10
    if (Verbosely)
        _tprintf(_T("\nTesting RAM (starting at byte %d)\n"),RamTestFirstByte);

    for (n = 0; n < 3; n++)
    {
        for (j = 0; j < 256; j++)
        {
            for (i = RamTestFirstByte; i < 0x100; i++)
            {
                SlWriteReg(TheSl,(UINT8)i,(UINT8)(i+j));
            }

            for (i = RamTestFirstByte; i < 0x100; i++)
            {
                c = SlReadReg(TheSl,(UINT8)i);
                if (c != (UINT8)(i+j))
                {
                    if (((fail++ == 0) && Verbosely) ||
                        (Verbosely > 1))
                        _tprintf(_T("RamTest err @%d: x%x != x%x\n"), i,
                                 c, (UINT8)(i+j));
                }
            }
        }
        if (Verbosely)
            _tprintf(_T("\rFailures: %3d "),fail);
    }

    if (fail)
    {
        n = n * j * i;
        j = (fail * 100);
        i = j / n; /* int part */
        j -= (i * n);
        j = (j * 100) / n; /* frac part */
        _tprintf(_T("RamTest FAILED %d times (of %d total, %d.%02d%%)\n"),
                 fail, n, i, j);
    }
    else
        _tprintf(_T("RamTest PASS\n"));
    if (Verbosely)
        DumpRam(Sl,NULL);
    return (fail == 0);
}

PRIVATE BOOL RamTest2(PSL811REGS Sl, int Verbosely)
{
    int i, j, n, fail = 0;
    UINT8 c;

    /* Bytes 13-15 dont "take" on a working chip
     */
//#define RamTestFirstByte 0x10
    if (Verbosely)
        _tprintf(_T("\nTesting RAM (starting at byte %d)\n"),RamTestFirstByte);

    for (n = 0; n < 3; n++)
    {
        for (j = 0; j < 256; j++)
        {
            for (i = RamTestFirstByte; i < 0x100; i++)
            {
                SlWriteReg(TheSl,(UINT8)i,(UINT8)(j));
            }

            for (i = RamTestFirstByte; i < 0x100; i++)
            {
                c = SlReadReg(TheSl,(UINT8)i);
                if (c != (UINT8)(j))
                {
                    if (((fail++ == 0) && Verbosely) ||
                        (Verbosely > 1))
                        _tprintf(_T("RamTest err @%d: x%x != x%x\n"), i,
                                 c, (UINT8)(j));
                }
            }
        }
        if (Verbosely)
            _tprintf(_T("\rFailures: %3d "),fail);
    }

    if (fail)
    {
        n = n * j * i;
        j = (fail * 100);
        i = j / n; /* int part */
        j -= (i * n);
        j = (j * 100) / n; /* frac part */
        _tprintf(_T("RamTest FAILED %d times (of %d total, %d.%02d%%)\n"),
                 fail, n, i, j);
    }
    else
        _tprintf(_T("RamTest PASS\n"));
    if (Verbosely)
        DumpRam(Sl,NULL);
    return (fail == 0);
}

int _tmain(int argc, _TCHAR **argv)
{
    int i;
    _TCHAR c;
    UINT8 r;
    PUSBD Usbd = NULL;

    ParseCmdLine(&argc,&argv);

    Mutex_Init(&SlLock);

    _tprintf(_T("Current RAM contents:\n"));
    if (argc > 1)
    {
        /* Read and print internal RAM to see if it makes sense
         */
        DumpRam(TheSl,b1);

        /* Do it again, readably
         */
        fancyprint(b1);
    }
    else
    {
        fancydump(TheSl);
    }

    if (argc > 1)
    {
        c = argv[1][0];

        if ((c == _T('p')) || (c == _T('P')))
        {
            SCODE sc;
            PIPROCESS pPrc = CurrentProcess();

            _tprintf(_T("\nStarting interactive testing (with%s real ISR)\n"),
                   (c == _T('p')) ? _T("out") : _T(""));
            _tprintf(_T("\nType '?' for help.\n"));

            Usbd = UsbdNew(TheSl);

            if (c == _T('p'))
                sc = pPrc->v->CreateThread(pPrc, 0,
                                           (THREAD_FUNCTION) loopy,
                                           (THREAD_ARGUMENT) TheSl,
                                           4*1024,
                                           NULL, NULL);
            else
            {
                Condition_InterruptSignal(&Usbd->InsertRemove);
                InstallTheIsr(Usbd);

                sc = pPrc->v->CreateThread(pPrc, 0,
                                           (THREAD_FUNCTION) PnpThread,
                                           (THREAD_ARGUMENT) Usbd,
                                           4*1024,
                                           NULL, NULL);
            }

            for (;;)
            {
                _tprintf(_T("Command [?|c|d|D|r|i|p|s|l|t|w|z|1..0|e]:"));
                c = (UINT8) _gettchar();
                _puttchar(_T('\n'));
                if (c == _T('?'))
                    _tprintf(_T("?: this message\n")
                             _T("c: clear ISTATUS register\n")
                             _T("d: dump regstate\n")
                             _T("D: dump RAMstate\n")
                             _T("r: reset chip\n")
                             _T("i: print ISTATUS register\n")
                             _T("p: print stats\n")
                             _T("s: last script results\n")
                             _T("l: dump ISR log\n")
                             _T("t: start listener thread.\n")
                             _T("u: test r/w of chip's RAM, verbose.\n")
                             _T("U: test#2 r/w of chip's RAM, verbose.\n")
                             _T("w: test r/w of chip's RAM.\n")
                             _T("W: test r/w of chip's RAM, various CS values\n")
                             _T("z: zero all registers\n")
                             _T("1|2|...0: send n-th SETUP pkt out\n")
                             _T("e: end test\n")
                           );
                if (c == _T('e'))
                    break;
                if (c == _T('p'))
                    print_stats();
                if (c == _T('i'))
                {
                    Mutex_Lock(&SlLock);
                    r = SlReadReg(TheSl,SL_ISTATUS);
                    fancyreg(SL_ISTATUS,r);
                    Mutex_Unlock(&SlLock);
                }
                if (c == _T('c'))
                {
                    Mutex_Lock(&SlLock);
                    r = SlReadReg(TheSl,SL_ISTATUS);
                    SlWriteReg(TheSl,SL_ISTATUS,r);
                    Mutex_Unlock(&SlLock);
                }
                if (c == _T('d'))
                {
                    Mutex_Lock(&SlLock);
                    fancydump(TheSl);
                    Mutex_Unlock(&SlLock);
                }
                if (c == _T('D'))
                {
                    Mutex_Lock(&SlLock);
                    DumpRam(TheSl,NULL);
                    Mutex_Unlock(&SlLock);
                    if (Usbd->CurrentPacket)
                        _tprintf(_T("Script: %d\n"),Usbd->CurrentPacket->Step);
                    print_stats();
                }
                if (c == _T('l'))
                {
                    _tprintf(_T("Logdump[%d]:\n"), CurL);
                    for (i = 0; i < CurL; i++)
                    {
                        UINT8 *Dump = Log[i].Regs;
                        int j;
                        _tprintf(_T("Step %d:\n"), Log[i].Step);
                        fancyprint(Dump);
                        for (j = 16; j < DUMPSIZE; j++)
                        {
                            if ((j & 15) == 0)
                                _tprintf(_T("\nMem[%03d]: "), j);
                            _tprintf(_T("%2x "), Dump[j]);
                        }
                        _tprintf(_T("\n"));
                    }
                }
                if (c == _T('r'))
                {
                    Mutex_Lock(&SlLock);
                    SlReset(TheSl,NULL,FALSE);
                    Usbd->Client.Address = 0;
                    Mutex_Unlock(&SlLock);
                }
                if (c == _T('t'))
                {
                    /* Create a thread that spins reading
                     */
                    _tprintf(_T("\nStarting listening thread\n"));
            
                    sc = pPrc->v->CreateThread(pPrc, 0,
                                               (THREAD_FUNCTION) HidReadThread,
                                               (THREAD_ARGUMENT) Usbd,
                                               4*1024,
                                               NULL, NULL);
                }
                if (c == _T('u'))
                {
                    RamTest(TheSl,2);
                    
                }
                if (c == _T('U'))
                {
                    RamTest2(TheSl,2);
                    
                }
                if (c == _T('w'))
                {
                    RamTest(TheSl,1);
                    
                }
                if (c == _T('W'))
                {
                    static UINT32 CsValues[] = {
                        0x20002122, 0x20002126, 0x2000212a, 0x2000212e,
                        0x20002132, 0x20002136, 0x2000213a, 0x2000213e,

                        0x20002322, 0x20002326, 0x2000232a, 0x2000232e,
                        0x20002332, 0x20002336, 0x2000233a, 0x2000233e,

                        0x20002f22, 0x20002f26, 0x20002f2a, 0x20002f2e,
                        0x20002f32, 0x20002f36, 0x20002f3a, 0x20002f3e,
                    };
#define nCS_VALUES ((sizeof CsValues) / sizeof CsValues[0])
#undef EBICS_NWS
# define                EBICS_NWS(_n_)  (((_n_)>>2)&0x7)
#undef EBICS_TDF
# define                EBICS_TDF(_n_)  (((_n_)>>9)&0x7)
                    for (i = 0; i < nCS_VALUES; i++) {
                        *(volatile UINT32 *)0xffe00008 = CsValues[i];
                        _tprintf(_T("CS2=%08x WS=%d TDF=%d: "),
                                    CsValues[i], EBICS_NWS(CsValues[i]),
                                    EBICS_TDF(CsValues[i]));
                        RamTest(TheSl,0);
                    }                    
                }
                if (c == _T('z'))
                {
                    int n;
                    _tprintf(_T("Zeroing RAM contents.\n"));
                    Mutex_Lock(&SlLock);
                    for (n = 0; n < 256; n++)
                        SlWriteReg(TheSl,(UINT8)n,0);
                    Mutex_Unlock(&SlLock);
                }
                if (c == _T('s'))
                {
                    UsbdDump(Usbd);
                }

                if (c == _T('1'))
                {
                    SlGetDeviceDescriptor(Usbd);
                }

                if (c == _T('2'))
                {
                    SlSetDeviceAddress(Usbd);
                }

                if (c == _T('3'))
                {
                    SlGetConfigurationDescriptor(Usbd);
                }

                if (c == _T('4'))
                {
                    SlReadReport(Usbd,TRUE);
                }

                if (c == _T('5'))
                {
                    SlGetStatus(Usbd);
                }

                if (c == _T('6'))
                {
                    SlSetConfiguration(Usbd);
                }

                if (c == _T('7'))
                {
                    SlSetIdle(Usbd);
                }

                if (c == _T('8'))
                {
                    SlGetReportDescriptor(Usbd);
                }

                if (c == _T('9'))
                {
                    SlGetReport(Usbd);
                }

                if (c == _T('0'))
                {
                    SlSetReport(Usbd);
                }

                if (c == _T('S'))
                {
                    SlGetStringDescriptors(Usbd);
                }

            }


            blimp = TRUE;

            SlDestroy(TheSl);

        }
        else if (c == _T('z'))
        {
            int n;
            _tprintf(_T("Zeroing RAM contents.\n"));
            for (n = 0; n < 256; n++)
                SlWriteReg(TheSl,(UINT8)n,0);
        }
        else if (c == _T('f'))
        {
            _tprintf(_T("Looping forever, resetting the device.\n"));
            _tprintf(_T("Remove the device to end this test.\n"));
            ResetLoop(TheSl);
        }
        else if (c == _T('t'))
        {
            _tprintf(_T("Sanity-testing the SL811 at x%x.\n"), (ADDRESS)TheSl);
            RamTest(TheSl,1);
        }
        else if (c == _T('T'))
        {
            _tprintf(_T("Operating the device for 5 seconds, reset+interrupts.\n"));

            Usbd = UsbdNew(TheSl);
            SlReset(TheSl,Usbd,TRUE);
            MilliSleep(5000);
            SlDestroy(TheSl);
            fancydump(TheSl);
        }
        else
            _tprintf(_T("\nUsage: %s [T|t|p|P|z|f]\n"), argv[0]);

        print_stats();
    }

    if (Usbd)
        UsbdDestroy(Usbd);

    FreeCmdLine(&argc,&argv);
    return 0;
}

