/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usb_sched.c
 *
 * Purpose:
 *
 *    USB bus scheduling code
 *
 * Reference: "OpenHCI Open Host Controller Interface Specification for USB"
 *            Rev 1.0a, September 1999.
 *
 * Reference: "Enhanced Host Controller Interface Specification
 *             for Universal Serial Bus"  Rev 1.0, March 2002.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

/* Needed defines and their meanings
 *
 * USK_LEVEL2_MAX
 *      Set this to 1 for USB1 single-level scheduling and to the number of
 *      u-frames per msec otherwise (e.g. 8 for USB2).
 * USK_LEVEL1_MAX
 *      Set this to no more than the number of hardware-supported roots.
 *      It could be the same, half of that, ...
 * USK_BW_PER_FRAME
 *      Either the USB1 bandwidth, or the USB2 b/w adjusted for uframe splits.
 * HC_NULL_POINTER
 *      Whatever the hardware understands to be an invalid (root) pointer
 * HcSetPointerValue(ADDRESS* Pointer, ADDRESS Value)
 *      Either an assignment, a byteswap or a funcall to do a "*Pointer=Value"
 */

/* We will do some gimmicks to reference our nodes
 */
typedef UINT32 PUSKNODE;
#define USK_INVALID_NODE         0 /* freq>1 */
#define UskL2(_n_)               ((_n_)&(USK_LEVEL2_MAX-1))
#define UskFrequency(_n_)        (((_n_)>>(8+12)))
#define UskStart(_n_)            (((_n_)>>8)&(USK_LEVEL1_MAX-1))
#define UskEncode(_l2_,_f_,_s_)  ((PUSKNODE)(UskL2(_l2_) \
                                             | ((_s_)<<8) \
                                             | ((_f_)<<(8+12))))

/* Notice that frequency must dominate */
#define UskIsLowerOrSame(_a_,_b_)  ((_a_) <= (_b_))

#define CheckL2(a,b,c) assert( UskL2(UskEncode(a,b,c)) == a)
#define CheckFr(a,b,c) assert( UskFrequency(UskEncode(a,b,c)) == b)
#define CheckSt(a,b,c) assert( UskStart(UskEncode(a,b,c)) == c)

/* The schedule is basically a table of in-use bandwidths.
 */
typedef struct _USKBW {
    UINT16  MinimumBandwidth;
    UINT16  MaximumBandwidth;
} USKBW, *PUSKBW;

typedef USKBW USKBW_ARRAY[USK_LEVEL1_MAX];
#define UskBwRef(_a_,_l2_,_l1_) &((_a_)[_l1_])

typedef struct _USKEDULE {
    HC_POINTER_TYPE *Roots;
    UINT16           RootsCount;
    UINT16           Pad;
    USKBW_ARRAY      InUse;
} USKEDULE, *PUSKEDULE;

/* These are the exported functions
 */
PRIVATE PUSKNODE
UskMakeReservation(
    PUSKEDULE Schedule,
    UINT16 FrequencyInMsecs,
    UINT16 MinimumBandwidth,
    UINT16 MaximumBandwidth
    );

PRIVATE void
UskCancelReservation(
    PUSKEDULE Schedule,
    PUSKNODE Node,
    UINT16 MinimumBandwidth,
    UINT16 MaximumBandwidth
    );

PRIVATE PUSKEDULE
UskCreateSchedule(
    HC_POINTER_TYPE *RootsList,
    UINT16 RootsCount
    );

/*
 * Work function for propagating bandwidth changes, positive or negative.
 */
PRIVATE
void
UskPropagateChange(
    PUSKEDULE Schedule,
    int Level2StartIndex,
    int Level1StartIndex,
    int Frequency,
    int MinimumBandwidthChange,
    int MaximumBandwidthChange
    )
{
    PUSKBW Node;

    for (;;) {
        /* Dereference the indeces to get to the desired node
         */
        Node = UskBwRef(Schedule->InUse,Level2StartIndex,Level1StartIndex);

        /* Adjust this one node
         */
        Node->MinimumBandwidth += MinimumBandwidthChange;
        Node->MaximumBandwidth += MaximumBandwidthChange;

        /* Move on to next slot, dont overshoot
         */
        Level1StartIndex += Frequency;

        if (Level1StartIndex >= USK_LEVEL1_MAX)
            break;
    }
}

/* 
 * Reserve a certain amount of b/w in the schedule
 * The minimum b/w is when an interrupt transfer receives a NAK.
 * The maximum b/w is when one full packet is actually transferred.
 */
PRIVATE PUSKNODE
UskMakeReservation(
    PUSKEDULE Schedule,
    UINT16 FrequencyInMsecs,
    UINT16 MinimumBandwidth,
    UINT16 MaximumBandwidth
    )
{
    PUSKBW Node;
    PUSKNODE Reservation;
    int Level2Index, Level1Index, StartIndex;
    int AvailableBandwidth;

    DBGME(0,_tprintf(_T("UskMakeReservation(%x,%d,%d,%d)\n"),
                     (ADDRESS) Schedule, FrequencyInMsecs,
                     MinimumBandwidth, MaximumBandwidth));

    /* Note that the loops are structured such that we will automatically
     * trim any excessive frequency down to the maximum allowed.
     */
    if (FrequencyInMsecs == 0)
        goto Impossible;

    /* Unfortunately, our indexing tricks only work on frequencies that are 2^n
     * Roundup the proposed frequency to the closest 2^n
     */
    for (StartIndex = 1; StartIndex < FrequencyInMsecs; StartIndex <<= 1)
        ;
    FrequencyInMsecs = (UINT16) StartIndex;

    /* See if we can fit it with guaranteed b/w
     */
    for (Level2Index = 0; Level2Index < USK_LEVEL2_MAX; Level2Index++) {

        for (StartIndex = 0; StartIndex < FrequencyInMsecs; StartIndex++) {

            for (Level1Index = StartIndex;
                 Level1Index < USK_LEVEL1_MAX;
                 Level1Index += FrequencyInMsecs) {

                /* Dereference the indeces to get to the desired node
                 */
                Node = UskBwRef(Schedule->InUse,Level2Index,Level1Index);

                /* Do we have any guaranteed bandwidth left ?
                 * And is it enough.
                 */
                AvailableBandwidth = USK_BW_PER_FRAME - Node->MaximumBandwidth;
                if (AvailableBandwidth > MaximumBandwidth)
                    goto FoundSpot;
            }
        }
    }

    /* Try again with a best-effort b/w
     */
    for (Level2Index = 0; Level2Index < USK_LEVEL2_MAX; Level2Index++) {

        for (StartIndex = 0; StartIndex < FrequencyInMsecs; StartIndex++) {

            for (Level1Index = StartIndex;
                 Level1Index < USK_LEVEL1_MAX;
                 Level1Index += FrequencyInMsecs) {

                /* Dereference the indeces to get to the desired node
                 */
                Node = UskBwRef(Schedule->InUse,Level2Index,Level1Index);

                /* Do we have enough best-effort bandwidth left ?
                 */
                AvailableBandwidth = USK_BW_PER_FRAME - Node->MinimumBandwidth;
                if (AvailableBandwidth > MinimumBandwidth)
                    goto FoundSpot;
            }
        }
    }

    /* Ooops, doesnt fit at all
     */
 Impossible:
    DBGME(3,_tprintf(_T("UskMakeReservation: no b/w for (%d,%d,%d)\n"),
                     FrequencyInMsecs, MinimumBandwidth, MaximumBandwidth));
    return USK_INVALID_NODE;

 FoundSpot:
    /* Make the bandwidth adjustments
     */
    UskPropagateChange(Schedule,
                       Level2Index,
                       StartIndex,
                       FrequencyInMsecs,
                       0+MinimumBandwidth,
                       0+MaximumBandwidth);

    /* Encode our triplet
     */
    Reservation = UskEncode(Level2Index,FrequencyInMsecs,StartIndex);

    DBGME(0,_tprintf(_T("UskMakeReservation() %x @Schedule[%d,%d,%d] => %x\n"),
                     (ADDRESS)Node,
                     Level2Index, Level1Index, StartIndex, Reservation));
    return Reservation;
}

/*
 * Drop a slot from the schedule, presumably at device removal
 */
PRIVATE void
UskCancelReservation(
    PUSKEDULE Schedule,
    PUSKNODE Node,
    UINT16 MinimumBandwidth,
    UINT16 MaximumBandwidth
    )
{
    DBGME(0,_tprintf(_T("UskCancelReservaion(%x,%x,%d,%d)\n"),
                     (ADDRESS) Schedule, (ADDRESS) Node,
                     MinimumBandwidth, MaximumBandwidth));

    UskPropagateChange(Schedule,
                       UskL2(Node),
                       UskStart(Node),
                       UskFrequency(Node),
                       0-MinimumBandwidth,
                       0-MaximumBandwidth);
}
    

/*
 * Build the initial schedule
 */
PRIVATE PUSKEDULE
UskCreateSchedule(
    HC_POINTER_TYPE *RootsList,
    UINT16 RootsCount
    )
{
    PIHEAP pHeap = CurrentHeap();
    PUSKEDULE Schedule;
    int i;

    DBGME(0,_tprintf(_T("UskCreateSchedule(%x,%d,%d,%d,%d)\n"),
                     (ADDRESS)RootsList,RootsCount,
                     USK_BW_PER_FRAME, USK_LEVEL2_MAX, USK_LEVEL1_MAX));

    /* We *are* initializing the roots
     */
    assert( RootsList != NULL );

    /* Ensure the various constants will work
     */
    CheckL2(USK_LEVEL2_MAX-1,USK_LEVEL1_MAX-1,USK_LEVEL1_MAX-1);
    CheckFr(USK_LEVEL2_MAX-1,USK_LEVEL1_MAX-1,USK_LEVEL1_MAX-1);
    CheckSt(USK_LEVEL2_MAX-1,USK_LEVEL1_MAX-1,USK_LEVEL1_MAX-1);
#if USK_LEVEL2_MAX > 1
    CheckL2(USK_LEVEL2_MAX-2,USK_LEVEL1_MAX-2,USK_LEVEL1_MAX-2);
    CheckFr(USK_LEVEL2_MAX-2,USK_LEVEL1_MAX-2,USK_LEVEL1_MAX-2);
    CheckSt(USK_LEVEL2_MAX-2,USK_LEVEL1_MAX-2,USK_LEVEL1_MAX-2);
#endif

    /* Some constant must be 2^N and at least 2
     */
#define Is2xn(_n_) \
    assert( (_n_) >= 2); \
    assert( (((_n_) & ((_n_)-1)) == 0) && \
            (((_n_) & ~((_n_)-1)) == (_n_)) )

    Is2xn( USK_LEVEL1_MAX );
    Is2xn( RootsCount );

    /* Allocate the schedule 
     */
    Schedule = (PUSKEDULE) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY,
                                           sizeof(USKEDULE), 0);

    if (Schedule == NULL) {
        DBGME(3,_tprintf(_T("UskCreateSchedule: no mem %u\n"),
                         (int)sizeof(USKEDULE)));
        return NULL;
    }

    /* Initialize global info
     */
    Schedule->RootsCount = RootsCount;
    Schedule->Roots      = RootsList;

    /* Terminate all roots
     */
    for (i = 0; i < RootsCount; i++) {
        HcSetPointerValue(&RootsList[i],HC_NULL_POINTER);
    }

    /* We are done
     */
    DBGME(0,_tprintf(_T("UskCreateSchedule() -> %x\n"), (ADDRESS)Schedule));
    return Schedule;
}
