/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbhub.c
 *
 * Purpose:
 *
 *    Device driver for USB HUBs, roots or not.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#define DEVICE_ORDER LITTLE_ENDIAN
#include <mmlite.h>
#include <mmhal.h>
#include <stddef.h> /* for offsetof() */
//#define _DEBUG 1
int UsbHubDebug = 3;
#define DBGLVL UsbHubDebug
#include <diagnostics.h>
#include <assert.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbmaster.h>

const _TCHAR *PnpName = _T("pnp");

/* Per-port status information
 * Dynamically sized according to the HUB's supported port number
 */
struct PORT_STATUS {
    PIUSBDRIVER Driver;
    UINT16 Vid, Pid, Dev;
    UINT8 Pad;
    UINT8 HubVersion; /* 0:non-hub 1:USB1x 2:USB2x */
    USBM_EVENT Status; /* CONNECTING/CONNECTED/.../DISCONNECTED */
    UINT8 UsbAddress;
    USBM_URB_FLAGS SpeedFlags;
};

/* The HUB object
 */
typedef struct _USBHUB {
    struct CDevice CDevice;
    CONST_VTBL struct IUsbHubVtbl *UsbHubv;
    CONST_VTBL struct IUsbDriverVtbl *UsbHubDriverv;
    /* Fields common to RootHubs and HUBs
     */
    CONDITION SleepPoint;
    UINT8 MyUsbAddress;
    UINT8 UpstreamPort;
    UINT8 ThreadStarted, ThreadShouldExit;
    PIUSBMASTERCONTROLLER Hci;
    PIUSBHUB Hub;
#define IAmRoot(this) (this->Hub == NULL)
    PINAMESPACE NameSpace;
    MUTEX  ConfigurationLock;
#define nUsbAddressesInUse (128/32)
    UINT32 UsbAddressesInUse[nUsbAddressesInUse];
    struct PORT_STATUS *Ports;
    UINT8  NumPorts;
    /* Fields specific to a HUB device
     */
    UINT8  InterruptEndpoint;
    UINT8  PollingRate;
    UINT8  MaximumControlPacketSize;
    UINT16 MaximumPacketSize;
    UINT16 IsUsb20;
} USBHUB, *PUSBHUB;

/* Convert between interface and instance pointer and vice versa */
#define pUsbHubDevice(_i_) ((PUSBHUB)((_i_)))
#define iUsbHubDevice(_p_) ((PIDEVICE)((_p_)))
#define pUsbHub(_i_) ((PUSBHUB)((ADDRESS)(_i_) - offsetof(struct _USBHUB, UsbHubv)))
#define iUsbHub(_p_) ((PIUSBHUB)(&((PUSBHUB)(_p_))->UsbHubv))
#define pUsbHubDriver(_i_) ((PUSBHUB)((ADDRESS)(_i_) - offsetof(struct _USBHUB, UsbHubDriverv)))
#define iUsbHubDriver(_p_) ((PIUSBDRIVER)(&((PUSBHUB)(_p_))->UsbHubDriverv))

/* UsbHubDevice IUnknown
 */
PRIVATE SCODE UsbHubDeviceQueryInterface (
    PIDEVICE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbHubDevQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));

    /* IDevice
     */
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_IFile,iid)
        || UuidCmp(&IID_IDevice,iid))
    {
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }

    /* IUsbHub::IUsbMasterController
     */
    if (UuidCmp(&IID_IUsbHub, iid) || UuidCmp(&IID_IUsbMasterController, iid)) {
        /* Use a single refcount
         */
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (void*) iUsbHub(pThis);
        return S_OK;
    }

    /* IUsbDriver
     */
    if (UuidCmp(&IID_IUsbDriver, iid)) {
        /* Use a single refcount
         */
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (void*) iUsbHubDriver(pThis);
        return S_OK;
    }


    /* Delegate INameSpace
     */
    if (UuidCmp(iid,&IID_INameSpace)) {
        PINAMESPACE pNs = pUsbHubDevice(pThis)->NameSpace;
        if (pNs)
            return pNs->v->QueryInterface(pNs,
                                          iid,
                                          ppObject);
        /* otherwise fail through */
    }

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT UsbHubDeviceAddRef (
    PIDEVICE pThis)
{
    DBGME(0,_tprintf(_T("UsbHubDevAR(%x)\n"),(ADDRESS)pThis));
    return AtomicInc(&pUsbHubDevice(pThis)->CDevice.RefCnt);
}

PRIVATE UINT UsbHubDeviceRelease (
    PIDEVICE pThis)
{
    PUSBHUB this = pUsbHubDevice(pThis);
    UINT Refs = AtomicDec(&this->CDevice.RefCnt);
    DBGME(0,_tprintf(_T("UsbHubDevRel(%x,%d)\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        pHeap->v->Free(pHeap, 0, (void *) this);
    }
    return Refs;
}

/* One wonders why we inherit from CDevice at all?
 */

PRIVATE SCODE UsbHubDeviceGetSize (
    PIDEVICE pThis,
    /*out*/ UINT64 * Size)
{
    PUSBHUB this = pUsbHubDevice(pThis);

    Uint32ToUint64(*Size,this->NumPorts);
    return S_OK;
}

/* Utilities for ThreadLoop
 */

/* Allocate the array of port descriptors, according to the number
 * of ports supported by the HUB/RootHub
 */
PRIVATE BOOL AllocatePorts(PUSBHUB this)
{
    this->Ports = malloc(this->NumPorts * sizeof(struct PORT_STATUS));
    if (this->Ports == NULL) {
        DBGME(3,_tprintf(_T("UsbHub(%x): cantalloc %d ports\n"),(ADDRESS)this,this->NumPorts));
        return FALSE;
    }
    memset(this->Ports,0,this->NumPorts * sizeof(struct PORT_STATUS));
    return TRUE;
}

/* Acquire the configuration lock on this bus, wait if necessary
 */
PRIVATE SCODE UsbHubTakeConfigurationLock(PUSBHUB this)
{
    SCODE sc;
    struct USBM_REQUEST_BLOCK Urb;
    UINT8 Data;

    DBGME(0,_tprintf(_T("UsbHub(%x):TakeConfLock(%d)\n"), 
                     (ADDRESS) this, this->MyUsbAddress));
    if (IAmRoot(this)) {
        Mutex_Lock(&this->ConfigurationLock);
        sc = S_OK;
        goto Out;
    }

    memset(&Urb,0,sizeof Urb);
    Urb.DeviceAddress = 0xff; /* sanity */
    Urb.Pid = USB_PID_EVENT;
    Urb.Flags = URB_FLAG_CONTROL;
    Urb.Data = &Data;
    Data = USBM_EVENT_TAKE_CONFIGURATION_LOCK;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 &Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):TakeConfLock SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        return sc;
    }
 Out:
    DBGME(0,_tprintf(_T("UsbHub(%x):TakeConfLock(%d) sc=%x\n"),
                     (ADDRESS) this, this->MyUsbAddress, sc));
    return sc;
}

/* Release the configuration lock on this bus
 */
PRIVATE SCODE UsbHubDropConfigurationLock(PUSBHUB this)
{
    SCODE sc;
    struct USBM_REQUEST_BLOCK Urb;
    UINT8 Data;

    DBGME(0,_tprintf(_T("UsbHub(%x):DropConfLock(%d)\n"),
                     (ADDRESS) this, this->MyUsbAddress));
    if (IAmRoot(this)) {
        Mutex_Unlock(&this->ConfigurationLock);
        sc = S_OK;
        goto Out;
    }

    memset(&Urb,0,sizeof Urb);
    Urb.DeviceAddress = 0xff; /* sanity */
    Urb.Pid = USB_PID_EVENT;
    Urb.Flags = URB_FLAG_CONTROL;
    Urb.Data = &Data;
    Data = USBM_EVENT_DROP_CONFIGURATION_LOCK;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 &Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):DropConfLock SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        return sc;
    }
 Out:
    DBGME(0,_tprintf(_T("UsbHub(%x):DropConfLock(%d) sc=%x\n"),
                     (ADDRESS) this, this->MyUsbAddress, sc));
    return sc;
}

/* Allocate a UsbAddress out of our in-use bitmap.
 * Returns zero on error.
 */
PRIVATE UINT8 UsbHubAllocateAddress(
    PUSBHUB this
    )
{
    /* Find the next available address
     */
    UINT32 w;
    int i;
    UINT8 UsbAddress;

    /* Look for a promising bucket
     */
 Retry:
    for (i = 0; i < nUsbAddressesInUse; i++) {
        w = this->UsbAddressesInUse[i];
        if (w != (~0))
            break;
    }
    if (i == nUsbAddressesInUse) {
        DBGME(3,_tprintf(_T("UsbHub(%x) too many addresses ?!?!?\n"),this));
        return 0;
    }

    /* Find a slot
     */
    UsbAddress = i;/*remember the bucket*/
    for (i = 0; i < 32; i++)
        if ((w & (1 << i)) == 0)
            break;
    /* wont fail */
    assert(i < 32);
    
    /* This function might be called by two threads, conceivably.
     * One of them would have been blocked for a looooong time but...
     */
    if (!AtomicCmpAndSwap(&this->UsbAddressesInUse[UsbAddress],
                          w,
                          w | (1 << i)))
        goto Retry;

    UsbAddress = (32 * UsbAddress) + i;
    return UsbAddress;
}

/* Release a UsbAddress from our in-use bitmap.
 * Address zero cannot be released.
 */
PRIVATE BOOL UsbHubReleaseAddress(
    PUSBHUB this,
    UINT8 UsbAddress
    )
{
    int i = UsbAddress / 32;
    UINT32 w;

    if (UsbAddress && (i < nUsbAddressesInUse)) {
        UsbAddress = UsbAddress - (i * 32);
        do {
            w = this->UsbAddressesInUse[i];
        } while (!AtomicCmpAndSwap(&this->UsbAddressesInUse[i],
                                   w,
                                   w & ~(1 << UsbAddress)));
        return TRUE;
    }
    return FALSE;
}

/* Assign a USB address to a device
 */
PRIVATE SCODE SetAddress(PUSBHUB this, UINT Port)
{
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbHub(%x):SetAddr(%d,%d)\n"),(ADDRESS)this,
                     Port,this->Ports[Port].UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &this->SleepPoint;
    Urb.DeviceAddress = 0;
    Urb.HubDeviceAddress = this->MyUsbAddress;
    Urb.HubPortNumber = (UINT8)(Port+1);/* we are zero-based */
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_CONTROL | this->Ports[Port].SpeedFlags;
    Urb.SetupData[1] = USB_R_SET_ADDRESS;
    Urb.SetupData[2] = this->Ports[Port].UsbAddress;
    Urb.SetupLength = 8;
    Urb.MaximumPacketSize = 8;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 &Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):SetAdd SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);
    sc = (Urb.Flags & URB_FLAG_ERROR) ? E_FAIL : S_OK;
    DBGME(0,_tprintf(_T("UsbHub(%x):SetAddr(%d,%d) ok. sc=x%x\n"),(ADDRESS)this,
                     Port,this->Ports[Port].UsbAddress,sc));
    return sc;
}

/* Query for the device's control endpoint maximum packet size
 * NB: This should only be used on address 0, before SetAddress().
 * The reason is that certain HCI's use the MPS as part of the key
 * used to find an endpoint descriptor. If we get there once with
 * one MPS and then with another they'll get quite confused.
 */
PRIVATE SCODE GetMaximumPacketSize(PUSBHUB this,
                                   UINT Port,
                                   UINT8 *pMps,
                                   UINT8 *pHub)
{
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
    const UINT8 UsbAddress = 0;
    UINT8 Data[8];

    DBGME(0,_tprintf(_T("UsbHub(%x):GetMps(%d,%d)\n"),(ADDRESS)this,Port,UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &this->SleepPoint;
    Urb.DeviceAddress = UsbAddress;
    Urb.HubDeviceAddress = this->MyUsbAddress;
    Urb.HubPortNumber = (UINT8)(Port+1);/* we are zero-based */
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_IN|URB_FLAG_CONTROL | this->Ports[Port].SpeedFlags;
    Urb.SetupData[0] = USB_RT_DTH;
    Urb.SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb.SetupData[3] = USB_DT_DEVICE;
    Urb.SetupData[6] = 8;
    Urb.SetupLength = 8;
    Urb.MaximumPacketSize = 8;
    Urb.ExpectedDataLength = 8;
    Urb.Data = Data;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 &Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):GetMps SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);
    sc = (Urb.Flags & URB_FLAG_ERROR) ? E_FAIL : S_OK;
    DBGME(0,_tprintf(_T("UsbHub(%x):GetMps(%d)->%x ok. sc=x%x\n"),(ADDRESS)this,
                     UsbAddress,(ADDRESS)Data,sc));

    *pMps = Data[7];
    /* Is it a HUB? */
    *pHub = (Data[4] == USB_DC_HUB) ? Data[3] : 0;
    return sc;
}

/* Query for the device descriptor information
 * Caller must free() the returned descriptor data
 */
PRIVATE SCODE GetDeviceDescriptor(PUSBHUB this,
                                  UINT Port,
                                  UINT8 Mps,
                                  PUSB_DEVICE_struct *pDescriptor)
{
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
#define maxdesc 64
    UINT8 *Data = malloc(maxdesc);

    DBGME(0,_tprintf(_T("UsbHub(%x):GetDe(%d,%d)\n"),(ADDRESS)this,Port,this->Ports[Port].UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &this->SleepPoint;
    Urb.DeviceAddress = this->Ports[Port].UsbAddress;
    Urb.HubDeviceAddress = this->MyUsbAddress;
    Urb.HubPortNumber = (UINT8)(Port+1);/* we are zero-based */
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_IN|URB_FLAG_CONTROL | this->Ports[Port].SpeedFlags;
    Urb.SetupData[0] = USB_RT_DTH;
    Urb.SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb.SetupData[3] = USB_DT_DEVICE;
    Urb.SetupData[6] = maxdesc;
    Urb.SetupLength = 8;
    Urb.MaximumPacketSize = Mps;
    Urb.ExpectedDataLength = maxdesc;
    Urb.Data = Data;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 &Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):GetDe SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);
    sc = (Urb.Flags & URB_FLAG_ERROR) ? E_FAIL : S_OK;
    DBGME(0,_tprintf(_T("UsbHub(%x):GetDe(%d)->%x ok. sc=x%x\n"),(ADDRESS)this,
                     this->Ports[Port].UsbAddress,(ADDRESS)Data,sc));

    *pDescriptor = (PUSB_DEVICE_struct) Data;
    return sc;
#undef maxdesc
}

/* Parse the HUB descriptor, pick up NumPorts
 */
PRIVATE SCODE ParseHubDescriptor(
    PUSBHUB this,
    PUSB_HUB_struct Descriptor
    )
{
    if (Descriptor->DescriptorType != USB_DT_HUB)
        return E_FAIL;

    this->NumPorts = Descriptor->NumberOfPorts;
    DBGME(0,_tprintf(_T("Hub: np %d ch %x pow %d cur %d\n"),
                     Descriptor->NumberOfPorts,
                     Get2(Descriptor->HubCharacteristics),
                     Descriptor->PowerOnGood * 2,
                     Descriptor->ControlCurrent));

    return S_OK;
}

/* Parse the configuration information returned by a HUB device.
 * Fields of interest are assigned into the HUB object.
 * These include the number of ports and the interrupt endpoint.
 */
PRIVATE SCODE
ParseConfigurationDescriptor(
    PUSBHUB this,
    UINT8 *Descriptor,
    UINT  Length)
{
    UINT i;
    UINT8 Dt, Ds, EpInt = 0/*invalid*/, Poll = 0;
    UINT16 Mps = 0;
    PUSB_ENDPOINT_struct Ed;

    /* Hunt for the interface and endpoint descriptors
     */
    for (i = 0; i < Length; i += Ds)
    {
        /* Get and check the size
         */
        Ds = Descriptor[i];
        if (Ds == 0) /* sanity */
            break;

        /* Look at the type
         */
        Dt = Descriptor[i+1];

        /* HUB?
         */
        if (Dt == USB_DT_HUB) {
            (void) ParseHubDescriptor(this,(PUSB_HUB_struct)(Descriptor + i));
            continue;
        }

        /* If not an endpoint skip
         */
        if (Dt != USB_DT_ENDP)
            continue;

        /* We expect only one input, interrupt endpoint
         */
        Ed = (PUSB_ENDPOINT_struct)(Descriptor + i);
        if (Ed->EndpointAddress & USB_EA_IN) {
            if ((Ed->Attributes & USB_AE_MASK) == USB_AE_INTERRUPT)
                EpInt = Ed->EndpointAddress & ~USB_EA_IN;
            else
                continue; /* say what? */
            Mps = Ed->MaxPacketSize[0] | (Ed->MaxPacketSize[1] << 8);
            Poll = Ed->Interval;
        }
    }

    DBGME(1,_tprintf(_T("HubConfig: Ports %d Mps %d Intr %d Poll %d\n"),
                     this->NumPorts, Mps, EpInt, Poll));

    /* Sanity check what we got
     */
    if (Mps == 0)
        return E_FAIL;

    this->InterruptEndpoint = EpInt;
    this->MaximumPacketSize = Mps;
    this->PollingRate = Poll ? Poll : 255;

    return S_OK;
}

PRIVATE SCODE SetConfiguration(
    PUSBHUB this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 Index)
{
    SCODE sc;

    memset(Urb,0,sizeof *Urb);
    Urb->Done = &this->SleepPoint;
    Urb->DeviceAddress = this->MyUsbAddress;
#if 0
    Urb->HubDeviceAddress = this->MyUsbAddress;
    Urb->HubPortNumber = (UINT8)(Port+1);/* we are zero-based */
#endif
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->SetupData[1] = USB_R_SET_CONFIG;
    Urb->SetupData[2] = Index;
    Urb->SetupLength = 8;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):SetCo SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        return sc;
    }

    Condition_Wait(Urb->Done,NULL,NULL);
    sc = (Urb->Flags & URB_FLAG_ERROR) ? E_FAIL : S_OK;
    DBGME(0,_tprintf(_T("UsbHub(%x):SetCo(%d) ok. sc=x%x\n"),(ADDRESS)this,
                     this->MyUsbAddress,sc));

    return sc;
}
    
PRIVATE SCODE GetConfiguration(
    PUSBHUB this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 *Data,
    UINT8 DataSize,
    UINT8 UsbAddress,
    UINT Port,
    UINT8 Mps
    )
{
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbHub(%x):GetCo(%d,%d)\n"),(ADDRESS)this,
                     this->UpstreamPort,UsbAddress));

    memset(Urb,0,sizeof *Urb);
    Urb->Done = &this->SleepPoint;
    Urb->DeviceAddress = UsbAddress;
    Urb->HubDeviceAddress = this->MyUsbAddress;
    Urb->HubPortNumber = (UINT8)(Port+1);/* we are zero-based */
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    if (this->Ports)
        Urb->Flags |= this->Ports[Port].SpeedFlags;
    Urb->SetupData[0] = USB_RT_DTH;
    Urb->SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb->SetupData[3] = USB_DT_CONFIG;
    Urb->SetupData[6] = DataSize;
    Urb->SetupLength = 8;
    Urb->MaximumPacketSize = Mps;
    Urb->ExpectedDataLength = DataSize;
    Urb->Data = Data;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):GetCo SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        goto Out;
    }

    Condition_Wait(Urb->Done,NULL,NULL);
    sc = (Urb->Flags & URB_FLAG_ERROR) ? E_FAIL : S_OK;
    DBGME(0,_tprintf(_T("UsbHub(%x):GetCo(%d)->%x ok. sc=x%x\n"),(ADDRESS)this,
                     UsbAddress,(ADDRESS)Data,sc));

 Out:
    return sc;
}

/* Specifically ask for the HUB descriptor, then parse it.
 */
PRIVATE SCODE GetHubDescriptor(
    PUSBHUB this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 *Data,
    UINT8 DataSize
    )
{
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbHub(%x):GetHub(%d,%d)\n"),(ADDRESS)this,
                     this->UpstreamPort,this->MyUsbAddress));

    memset(Urb,0,sizeof *Urb);
    Urb->Done = &this->SleepPoint;
    Urb->DeviceAddress = this->MyUsbAddress;
#if 0
    Urb->HubDeviceAddress = this->MyUsbAddress;
    Urb->HubPortNumber = (UINT8)(Port+1);/* we are zero-based */
#endif
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_DTH|USB_RT_CLASS;
    Urb->SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb->SetupData[3] = USB_DT_HUB;
    Urb->SetupData[6] = DataSize;
    Urb->SetupLength = 8;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;
    Urb->ExpectedDataLength = DataSize;
    Urb->Data = Data;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):GetHub SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        return sc;
    }

    Condition_Wait(Urb->Done,NULL,NULL);
    sc = (Urb->Flags & URB_FLAG_ERROR) ? E_FAIL : S_OK;
    DBGME(0,_tprintf(_T("UsbHub(%x):GetHub(%d)->%x ok. sc=x%x\n"),(ADDRESS)this,
                     this->MyUsbAddress,(ADDRESS)Data,sc));

    /* Look at what we got
     */
    return ParseHubDescriptor(this,(PUSB_HUB_struct) Data);
}

/* Change a port's feature, either setting it or clearing it
 */
PRIVATE SCODE SetOrClearPortFeature(
    PUSBHUB this,
    UINT8 PortNumber,
    UINT16 Feature,
    BOOL   fSetIt
    )
{
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbHub(%x):%cetFe(%d,%x,%d,%d)\n"),(ADDRESS)this,
                     (fSetIt) ? _T('S'):_T('C'), 
                     PortNumber, Feature,
                     this->UpstreamPort,this->MyUsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &this->SleepPoint;
    Urb.DeviceAddress = this->MyUsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_CONTROL;
    Urb.SetupData[0] = USB_RT_CLASS | USB_RT_OTHER; //x23, 20 for hub
    Urb.SetupData[1] = (fSetIt) ? USB_R_SET_FEATURE : USB_R_CLEAR_FEATURE;
    Set2(&Urb.SetupData[2],Feature);
    Urb.SetupData[4] = PortNumber;
    Urb.SetupLength = 8;
    Urb.MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 &Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):?etFe SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);
    sc = (Urb.Flags & URB_FLAG_ERROR) ? E_FAIL : S_OK;
    DBGME(0,_tprintf(_T("UsbHub(%x):?etFe(%d) ok. sc=x%x\n"),(ADDRESS)this,
                     this->MyUsbAddress,sc));

    return sc;
}
#define SetPortFeature(t,p,f)   SetOrClearPortFeature(t,(UINT8)(p+1),f,TRUE)
#define ClearPortFeature(t,p,f) SetOrClearPortFeature(t,(UINT8)(p+1),f,FALSE)
#define SetHubFeature(t,p,f)    SetOrClearPortFeature(t,0,f,TRUE)
#define ClearHubFeature(t,p,f)  SetOrClearPortFeature(t,0,f,FALSE)

/* Turn power on/off to one or all ports
 * Return value is only meaningful on a per-port basis.
 * BUGBUG HUB power?
 */
#define ALL_PORTS 0xff
PRIVATE SCODE HubPortPower(
    PUSBHUB this,
    UINT8 PortNumber,
    BOOL PowerOn
    )
{
    if (PortNumber != ALL_PORTS)
        return SetPortFeature(this,PortNumber,USB_FS_PORT_POWER);

    for (PortNumber = 0; PortNumber < this->NumPorts; PortNumber++)
        (void) SetPortFeature(this,PortNumber,USB_FS_PORT_POWER);
    return S_OK;
}

/* Configure a HUB.
 * Query the HUB for its configuration information, then parse it.
 * Assign to the HUB object the relevant information.
 * Do a SET_CONFIG that gets the HUB running.
 */
PRIVATE SCODE UsbHubConfigureHub(
    PUSBHUB this)
{
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
#define maxdesc 64
    UINT8 *Data = malloc(maxdesc);

    if (Data == NULL)
        return E_NOT_ENOUGH_MEMORY;

    DBGME(0,_tprintf(_T("UsbHub(%x):Config(%d,%d)\n"),(ADDRESS)this,
                     this->UpstreamPort,this->MyUsbAddress));

    sc = GetConfiguration(this,&Urb,Data,maxdesc,
                          this->MyUsbAddress, 0,
                          this->MaximumControlPacketSize);
    if (FAILED(sc))
        goto Out;

    /* Find out how many ports we should support
     * and which one is the interrupt endpoint
     */
    sc = ParseConfigurationDescriptor(this,
                                      Data,
                                      Urb.ActualDataLength);
    if (FAILED(sc))
        goto Out;

    /* Some HUBs do not provide the HUB descriptor in the config. Sigh.
     */
    if (this->NumPorts == 0) {
        sc = GetHubDescriptor(this,&Urb,Data,maxdesc);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x):no HUB desc?? (sc=%x)\n"),(ADDRESS)this, sc));
            goto Out;
        }
    }

    /* Ok, get it running
     */
    sc = SetConfiguration(this,&Urb,1);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):SetConfig failed?? (sc=%x)\n"),(ADDRESS)this, sc));
        goto Out;
    }

    /* Turn on power to all ports.
     * Beware that many HUBs lie on their number of ports.
     */
    sc = HubPortPower(this,ALL_PORTS,TRUE);

 Out:
    free(Data);
    return sc;
#undef maxdesc
}

/* Query a HUB for a given Port's status and change bitmaps.
 * Returned info is clear of bytesex issues.
 * Also usable for queryng the HUB status.
 */
typedef struct _USB_PORT_STATUS {
    /* See USB_ST_PORT_<foo> for bit definitions */
    UINT16 PortStatus;
    UINT16 PortChange;
} USB_PORT_STATUS, *PUSB_PORT_STATUS;

#define GetHubStatus(t,s)    GetHubOrPortStatus(t,0,s)
#define GetPortStatus(t,p,s) GetHubOrPortStatus(t,(UINT8)(p+1),s)

PRIVATE SCODE GetHubOrPortStatus(
    PUSBHUB this,
    UINT8 PortNumber,
    PUSB_PORT_STATUS Status
    )
{
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbHub(%x):GetPs(%d,%d,%d)\n"),(ADDRESS)this, PortNumber,
                     this->UpstreamPort,this->MyUsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &this->SleepPoint;
    Urb.DeviceAddress = this->MyUsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb.SetupData[0] = (PortNumber) ? (USB_RT_DTH|USB_RT_CLASS|USB_RT_OTHER)
                          :(USB_RT_DTH|USB_RT_CLASS);
    Urb.SetupData[1] = USB_R_GET_STATUS;
    Urb.SetupData[4] = PortNumber;
    Urb.SetupData[6] = 4;
    Urb.SetupLength = 8;
    Urb.MaximumPacketSize = this->MaximumControlPacketSize;
    Urb.ExpectedDataLength = 4; /* no trust no compiler me */
    Urb.Data = (UINT8*) Status;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 &Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):GetPs SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);
    sc = (Urb.Flags & URB_FLAG_ERROR) ? E_FAIL : S_OK;
    DBGME(0,_tprintf(_T("UsbHub(%x):GetPs(%d)->%x ok. sc=x%x\n"),(ADDRESS)this,
                     this->MyUsbAddress,(ADDRESS)Status,sc));

    Status->PortStatus = _bs16(Status->PortStatus);
    Status->PortChange = _bs16(Status->PortChange);
    return sc;
}


/* Post a request for status change indications to the interrupt endpoint
 * of a HUB device.
 */
PRIVATE SCODE GetPortStatusChanges(
    PUSBHUB this,
    UINT8 *ChangeBitmap,
    UINT8 BitmapLength
    )
{
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbHub(%x):GetSc(%d,%d)\n"),(ADDRESS)this,
                     this->UpstreamPort,this->MyUsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &this->SleepPoint;
    Urb.DeviceAddress = this->MyUsbAddress;
    Urb.EndpointAddress = this->InterruptEndpoint;
    Urb.Pid = USB_PID_IN;
    Urb.Flags = URB_FLAG_IN|URB_FLAG_INTERRUPT;
    Urb.SetupLength = 0;
    Urb.Interval = this->PollingRate;
    Urb.ExpectedDataLength = BitmapLength;
    Urb.Data = ChangeBitmap; 
    Urb.MaximumPacketSize = this->MaximumPacketSize;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 &Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):GetSc SubmitUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);

    /* We might be terminated, and the Urb will still be at the HCI.
     */
    if (this->ThreadShouldExit) {
        sc = this->Hci->v->CancelUrb(this->Hci,
                                     &Urb,
                                     this->UpstreamPort);
        DBGME(0,_tprintf(_T("UsbHub(%x):Canc %x sc=x%x\n"),(ADDRESS)this,
                         (ADDRESS) &Urb, sc));
        sc = E_FAIL;
    } else {
        sc = (Urb.Flags & URB_FLAG_ERROR) ? E_FAIL : S_OK;
    }
    DBGME(0,_tprintf(_T("UsbHub(%x):GetSc(%d)->%x ok. sc=x%x\n"),(ADDRESS)this,
                     this->MyUsbAddress,(ADDRESS)ChangeBitmap,sc));

    return sc;
}


/* Ask the PnP manager to locate a driver for a device.
 * We will try to load the PnP manager if it is not yet present.
 * The Ids string must be formatted by "Vid%04x_Pid%04x_Dev%04x".
 * The returned DriverName always includes the ".cob" suffix.
 * If DriverNameByteLength is not enough it is the PnP info that suffers.
 * The function can fail either because the PnP manager cannot be located,
 * or started, or because the device is indeed not in the PnP tables.
 */
PRIVATE SCODE PnpLookupDriver(_TCHAR *Ids,
                              _TCHAR *DriverName,
                              UINT DriverNameByteLength)
{
    PINAMESPACE Ns = NULL;
    PIUNKNOWN pUnk = NULL;
    PIFILE File = NULL;
    UINT ByteLength = 0;
    SCODE sc;

    /* Default fail return
     */
    DriverName[0] = 0;

    /* See if the PnP manager is already loaded
     */
    sc = BindToObject(CurrentNameSpace(),
                      PnpName,
                      NAME_SPACE_READ,
                      &IID_INameSpace,
                      (void **) &Ns);

    if (FAILED(sc)) {
        /* Nope, try to load it now then
         */
        sc = BindToObject(CurrentNameSpace(),
                          _T("COB\\pnp.cob"),
                          NAME_SPACE_READ,
                          &IID_IUnknown,
                          (void **) &pUnk);
        if (FAILED(sc)) {
            /* Urgh...
             */
            DBGME(3,_tprintf(_T("UsbHub no 'pnp' manager sc=%x\n"), sc));
            return sc;
        }

        /* Try again
         */
        sc = BindToObject(CurrentNameSpace(),
                          PnpName,
                          NAME_SPACE_READ,
                          &IID_INameSpace,
                          (void **) &Ns);

        pUnk->v->Release(pUnk);
        pUnk = NULL;

        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub bad 'pnp' manager sc=%x\n"), sc));
            return sc;
        }
    }

    /* Lookup the ID
     */
    sc = BindToObject(Ns,
                      Ids,
                      NAME_SPACE_READ,
                      &IID_IFile,
                      (void **) &File);

    if (FAILED(sc)) {
        DBGME(2,_tprintf(_T("UsbHub no pnp entry for %s\n"), Ids));
        goto Out;
    }

    /* Good, see what we got now
     */
    sc = File->v->ReadAt(File,
                         UINT64_ZERO,
                         (BYTE*) DriverName,
                         DriverNameByteLength - (4*sizeof(_TCHAR)), /* ".cob" ! */
                         &ByteLength);

    /* Shouldnt fail, really */
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub cant read pnp entry sc=%x\n"), sc));
        goto Out;
    }

    /* Put the finishing touches and we are done
     */
    _tcscpy((_TCHAR*)(((BYTE*)DriverName)+ByteLength),_T(".cob"));

    DBGME(1,_tprintf(_T("UsbHub:PnpLook(%s)->%s\n"),Ids,DriverName));
    sc = S_OK;
 Out:
    if (File)
        File->v->Release(File);
    if (Ns)
        Ns->v->Release(Ns);
    return sc;
}

/* Sleep for a few milliseconds
 */
PRIVATE void MilliSleep(int mSecs)
{
    TIME t;
    Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(mSecs)));
    SleepUntil(t);
}

BOOL hack = 0;
int s0 = 100;
int s1 = 50;
int s2 = 10;

/* Wait for a HUB to signal status changes.
 * Handle here some events, return only connect/disconnect events
 * Returns info packaged into the pseudo-URB that is used in RootHub
 * to talk to the HCI underneath.
 * BUGBUG Should be the other way around!!!
 */
PRIVATE SCODE UsbHubWaitForEvent(
    PUSBHUB this,
    PUSBM_REQUEST_BLOCK Urb
    )
{
    SCODE sc;
    UINT8 Port;
    UINT8 BitmapLength;
    /* This accounts for 15 ports max, though 8 is more likely */
    UINT8 Bitmap[2];
    UINT16 Changes;
    BOOL LockHeld = FALSE;
    int nLoops = 0;
    USB_PORT_STATUS Status;

    BitmapLength = 1 + (this->NumPorts / 8);
    if (BitmapLength > 2) {
        DBGME(3,_tprintf(_T("UsbHub(%x):Wait for too many, %d > 2\n"),(ADDRESS)this,BitmapLength));
        return E_FAIL;
    }

    /* ON powerup, all disconnected ports say st=100 (PORT_POWER) ch=0
     * ON port1-connect says Changes=2, st=301 (SPEED,POWER,CONN) ch=1 (CONN)
     * after reset says st=303 (SPEED,POWER,ENABLE,CONN) ch=10 (RESET)
     * ON port1-disconnect says st=100 ch=1
     */
 WaitAgain:
    /* Wait for some
     */
    sc = GetPortStatusChanges(this,Bitmap,BitmapLength);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):gpsc failed sc=%x\n"),(ADDRESS)this,sc));
        return sc;
    }

    Changes = Bitmap[0];
    if (BitmapLength > 1)
        Changes |= (Bitmap[1] << 8);

    DBGME(0,_tprintf(_T("Hub(%x)Event: %x\n"),(ADDRESS)this,Changes));

    /* Handle them one port at a time
     *
     * NB: Whatever we do not handle must persist and we will find it
     * the next time we get back here.  Meaning, until we explicitly
     * clear the changed bit it keeps coming back to us.
     *
     * NB: Bit0 is for HUB, 1 for Port1 etc.  But we number ports at 0.
     *
     * NB: I did try to handle more than one port at a time. It works, except..
     *     you endup with more than one device answering to (0,0). Urgh.
     */

    for (Port = 0; Port < this->NumPorts; Port++)
        if (Changes & (1 << (Port+1)))
            goto HandlePort;

    /* This does happen. When unplugging it sometimes manages to speak..
     */
    DBGME(3,_tprintf(_T("Hub(%x)Event: Bad changes %x\n"),(ADDRESS)this,Changes));
    return E_FAIL;

 HandlePort:
    sc = GetPortStatus(this,Port,&Status);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):gps failed sc=%x\n"),(ADDRESS)this,sc));
        goto Bad;
    }
    DBGME(1,_tprintf(_T("Hub(%x)Event: port%d st=%x ch=%x\n"),(ADDRESS)this,
                     Port,
                     Status.PortStatus,
                     Status.PortChange));

    if (Status.PortChange == 0) {
        /* Dont loop forever !!
         */
        if (++nLoops >= 100) {
            sc = E_RESTART;
            goto Bad;
        }
        MilliSleep(10);
        goto HandlePort;
    }

    /* Connection/disconnection event?
     */
    if (Status.PortChange & USB_ST_PORT_CONNECTION) {

        /* Ack */
        sc = ClearPortFeature(this,Port,USB_FS_C_PORT_CONNECTION);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x):cpf1 failed sc=%x\n"),(ADDRESS)this,sc));
            goto Bad;
        }

        /* Disconnection?  Easy.
         */
        if ((Status.PortStatus & USB_ST_PORT_CONNECTION) == 0) {

            Urb->ActualDataLength = 2;
            Urb->Data[0] = USBM_EVENT_DISCONNECTING;
            Urb->Data[1] = Port;
            sc = S_OK;
            if (LockHeld)
                goto Bad; /* uhu? */
            goto Out;
        }

        /* Connection. Apply Tattdb==100msec before starting reset.
         */
        MilliSleep((hack) ? s0 : 100);

        /* Take the lock
         */
        sc = UsbHubTakeConfigurationLock(this);
        assert(sc == S_OK);
        LockHeld = TRUE;

        /* Start driving reset
         */
        sc = SetPortFeature(this,Port,USB_FS_PORT_RESET);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x):spf1 failed sc=%x\n"),(ADDRESS)this,sc));
            goto Bad;
        }

        /* Wait for the required amount
         */
        MilliSleep((hack) ? s1 : 50);

        /* No go back and check again, we should have a reset complete waiting.
         */
        goto HandlePort;
    }

    /* Reset complete?
     */
    if (Status.PortChange & USB_ST_PORT_RESET) {

        assert(LockHeld);

        /* Ack */
        sc = ClearPortFeature(this,Port,USB_FS_C_PORT_RESET);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x):cpf2 failed sc=%x\n"),(ADDRESS)this,sc));
            goto Bad;
        }

        /* Reset recovery time
         */
        MilliSleep((hack) ? s2 : 10);

        /* Recheck now
         */
        sc = GetPortStatus(this,Port,&Status);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x):gps1 failed sc=%x\n"),(ADDRESS)this,sc));
            goto Bad;
        }
        DBGME(1,_tprintf(_T("Hub(%x)Event: Nport%d st=%x ch=%x\n"),(ADDRESS)this,
                         Port,
                         Status.PortStatus,
                         Status.PortChange));

        assert(Status.PortStatus & USB_ST_PORT_ENABLE);

        Urb->ActualDataLength = 4;
        Urb->Data[0] = USBM_EVENT_CONNECTING;
        Urb->Data[1] = Port;
        /* Low/High speed?
         */
        Set2(Urb->Data+2,0); /* FULL_SPEED */
        if (Status.PortStatus & USB_ST_PORT_LOW_SPEED) {
            Set2(Urb->Data+2,URB_FLAG_LOW_SPEED);
        }
        else if (Status.PortStatus & USB_ST_PORT_HIGH_SPEED) {
            Set2(Urb->Data+2,URB_FLAG_HIGH_SPEED);
        }
        goto Out;
    }

    /* OHCI on top of EHCI gets this
     */
    if (Status.PortChange & USB_ST_PORT_ENABLE) {
        /* Ack */
        sc = ClearPortFeature(this,Port,USB_FS_C_PORT_ENABLE);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x):cpf3 failed sc=%x\n"),(ADDRESS)this,sc));
            /* BUGBUG ignore? */
            goto Bad;
        }
        goto WaitAgain;
    }

    /* House on fire?
     */
    if (Status.PortChange & USB_ST_PORT_OVERCURRENT) {
        /* Turn it off asap
         */
        sc = ClearPortFeature(this,Port,USB_FS_PORT_POWER);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x):cpf4 failed sc=%x\n"),(ADDRESS)this,sc));
            /* ignore error */
        }

        /* Ack */
        sc = ClearPortFeature(this,Port,USB_FS_C_PORT_OVERCURRENT);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x):cpf5 failed sc=%x\n"),(ADDRESS)this,sc));
            /* ignore error */
        }

        goto WaitAgain;
    }

    /* We dont handle this yet, shouldnt happen.
     */
    assert (Status.PortChange & USB_ST_PORT_SUSPEND);
    sc = E_RESTART;
 Bad:
    if (LockHeld)
        (void) UsbHubDropConfigurationLock(this);
    return sc;

    /* We got something to say
     */
 Out:
    Urb->Flags |= URB_FLAG_COMPLETE;
    return S_OK;
}

/* Handle a disconnection event
 */
PRIVATE void UsbHubHandleDisconnection(
    PUSBHUB this,
    UINT Port,
    _TCHAR *DriverName,
    PUSBM_REQUEST_BLOCK Urb
    )
{
    SCODE sc = 0;
    UINT8 UsbAddress;
    PIUSBDRIVER Driver;

    Driver = this->Ports[Port].Driver;
    this->Ports[Port].Status = USBM_EVENT_DISCONNECTED;
    if (Driver) {
        /* Unregister it, ignore errors
         */
        _stprintf(DriverName,_T("Vid%04x_Pid%04x_Dev%04x"),
                  this->Ports[Port].Vid,
                  this->Ports[Port].Pid,
                  this->Ports[Port].Dev);

        sc = this->NameSpace->v->Unregister(this->NameSpace,
                                            DriverName);

        if (FAILED(sc)) {
            DBGME(2,_tprintf(_T("UsbHub(%x) Unregister %s failed sc=x%x\n"),(ADDRESS)this,
                             DriverName, sc));
        }
        /* And terminate it
         */
        DBGME(0,_tprintf(_T("UsbHub(%x):harnold %x\n"),(ADDRESS)this,(ADDRESS)Driver));
        sc = Driver->v->YouAreTerminated(Driver,
                                         iUsbHub(this));
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x) YouAreTerminated failed sc=x%x\n"),(ADDRESS)this,sc));
        }
        (void) Driver->v->Release(Driver);
        this->Ports[Port].Driver = NULL;

    }
    DBGME(0,_tprintf(_T("UsbHub(%x):EvDisc sc=x%x\n"),(ADDRESS)this,sc));

    /* Free up the UsbAddress it was using
     */
    UsbAddress = this->Ports[Port].UsbAddress;
    if (UsbAddress == 0)
        return;
    this->Ports[Port].UsbAddress = 0;

    /* Notify the HCI that we are reclaiming this address
     */
    Urb->DeviceAddress = 0xff;
    Urb->EndpointAddress = 0;
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->Pid = USB_PID_EVENT;
    Urb->Data[0] = USBM_EVENT_DISCONNECTED;
    Urb->Data[1] = UsbAddress;

    sc = this->Hci->v->SubmitUrb(this->Hci,
                                 Urb,
                                 this->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x):DiscUrb failed sc=x%x\n"),(ADDRESS)this,sc));
    }

    /* Reclaim it
     * For a non-root HUB this was done in the Submit above.
     */
    if (IAmRoot(this)) {
        UsbHubReleaseAddress(this,UsbAddress);
    }
}

/* Main control thread for the driver.
 * Loops waiting for attach/detach events and starts/terminates 
 * the proper device driver for the attached USB device.
 */
PRIVATE void UsbHubDeviceThreadLoop (
    PIDEVICE pThis)
{
    PUSBHUB this = pUsbHubDevice(pThis);
    PUSBM_REQUEST_BLOCK Urb = NULL;
    UINT Event, Port;
    USBM_URB_FLAGS Speed;
    UINT8 UsbAddress;
    UINT8 MaximumControlPacketSize;
    SCODE sc;
    PIUSBDRIVER Driver;
    _TCHAR *DriverName = NULL;
    _TCHAR *DriverCobName = NULL;
    PUSB_DEVICE_struct Descriptor;
    BOOL IsaHub;
    PIUNKNOWN pUnk;
    PIDEVICECONSTRUCTOR pDevCons;
    int n;

#define nDriverName (64*sizeof(_TCHAR))
    DriverName = malloc(nDriverName);
    if (DriverName == NULL)
        goto Out;

    DriverCobName = malloc(nDriverName);
    if (DriverCobName == NULL)
        goto Out;

#define payload 16
    Urb = (PUSBM_REQUEST_BLOCK) malloc(payload+sizeof *Urb);
    if (Urb == NULL)
        goto Out;
    memset(Urb,0,sizeof *Urb);

    /* This wont change */
    Urb->Done = &this->SleepPoint;

    DBGME(0,_tprintf(_T("UsbHub(%x):TL(%x,%x,%x) %d\n"),(ADDRESS)this,
                     (ADDRESS)DriverName,(ADDRESS)DriverCobName,
                     (ADDRESS)Urb,this->ThreadShouldExit));


    while (!this->ThreadShouldExit) {

        DBGME(0,_tprintf(_T("UsbHub@%x: looptop %x\n"),
                         (ADDRESS) this, (ADDRESS) Urb));

        /* Wait for an event
         */
        Urb->Data = (PUINT8) (Urb+1);
        Urb->DeviceAddress = this->MyUsbAddress;
        Urb->EndpointAddress = 0;
        Urb->Data[0] = USBM_EVENT_STATUS_WAIT;
        Urb->Flags = URB_FLAG_CONTROL;
        Urb->Pid = USB_PID_EVENT;
        Urb->ExpectedDataLength = payload;

        sc = UsbHubWaitForEvent(this,Urb);

        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x) WaitEvent failed sc=x%x\n"),(ADDRESS)this,sc));
            break;
        }

        /* Check we havent been terminatd
         */
        if (this->ThreadShouldExit)
            break;

        /* We got an event 
         */
        Event = Urb->Data[0];
        Port  = Urb->Data[1];
        Speed = Get2(Urb->Data+2);

        DBGME(1,_tprintf(_T("UsbHub(%x):Ev %d p%d\n"),(ADDRESS)this,Event,Port));

        /* Did a device get unplugged ?
         */
        if (Event == USBM_EVENT_DISCONNECTING) {
            /* Perform our cleanup
             */
            UsbHubHandleDisconnection(this,Port,DriverName,Urb);
            continue;
        }
            
        /* The only other event of interest here is a connection
         */
        if (Event != USBM_EVENT_CONNECTING)
            continue;

        /* Got a CONNECT
         * NB: Data[2,3] is a URB_FLAG for the speed, 0 means a FULL_SPEED one
         */
        DBGME(0,_tprintf(_T("UsbHub(%x):EvCon %x %x\n"),(ADDRESS)this,
                         Port, Speed));

        /* Mark the port in the connecting state
         */
        this->Ports[Port].Status     = USBM_EVENT_CONNECTING;
        this->Ports[Port].UsbAddress = 0;
        this->Ports[Port].SpeedFlags = Speed;
        this->Ports[Port].HubVersion = 0;
        this->Ports[Port].Driver     = NULL;

        /* There was some thought of doing power admission control.
         * Unfortunately, all devices lie and say they need "100 mA",
         * [brutally copying the example from the specs I suppose].
         * So the only thing we can trust is h/w overcurrent protection.
         * No point wasting code then.
         */
	
        /* The device needs an address
         */
        if (IAmRoot(this)) {

            /* Allocate one
             */
            UsbAddress = UsbHubAllocateAddress(this);

            /* Did we run out of addresses?
             */
            if (UsbAddress == 0) {
                /* These are the error handlers for the configuration sequence.
                 * They mimic the sequence itself.
                 */
            GotNoAddress:
                /* We did not allocate a USB address
                 */
                DBGME(3,_tprintf(_T("UsbHub(%x) no address!\n"),(ADDRESS)this));
                UsbAddress = 0;

            FailedToConfigure:
                /* We are still holding the configuraiton lock, drop it.
                 */
                sc = UsbHubDropConfigurationLock(this);
                assert(sc == S_OK);

            TurnItOff:
                /* Disable this port. If/when the user unplugs the device
                 * we will get a disconnection and all goes back to normal.
                 */
                sc = ClearPortFeature(this,Port,USB_FS_PORT_ENABLE);
                if (FAILED(sc)) {
                    DBGME(3,_tprintf(_T("UsbHub(%x):poweroff failed sc=%x\n"),(ADDRESS)this,sc));
                    /* ignore error */
                }

                /* Go back to listening for changes.
                 */
                continue;
            }

        } else {

            /* Ask the Root HUB on this bus for an address
             */
            Urb->DeviceAddress = 0xff;
            Urb->EndpointAddress = 0;
            Urb->Flags = URB_FLAG_CONTROL;
            Urb->Pid = USB_PID_EVENT;
            Urb->Data[0] = USBM_EVENT_ADDRESS_ALLOCATION;
            sc = this->Hci->v->SubmitUrb(this->Hci,
                                         Urb,
                                         this->UpstreamPort);

            if (FAILED(sc))
                goto GotNoAddress;

            UsbAddress = Urb->Data[1];
        }

        /* Remember that we need to give it back
         */
        this->Ports[Port].UsbAddress = UsbAddress;

        /* Before we lose address0 we need to find the control endpoint's MPS
         * This can be bigger than 8 for USB2.0
         */
        for (n = 1; n < 4; n++) {
            sc = GetMaximumPacketSize(this,Port,
                                      &MaximumControlPacketSize,
                                      &this->Ports[Port].HubVersion);
            if (!FAILED(sc))
                break;
            MilliSleep(500*n);
        }

        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x) GetMPS failed, sc=%x.\n"),
                             (ADDRESS)this,sc));
            goto FailedToConfigure;
        }

        /* Assign the address to the device
         */
        DBGME(2,_tprintf(_T("UsbHub(%x) new UsbAddress=%d\n"),(ADDRESS)this,UsbAddress));
        this->Ports[Port].UsbAddress = UsbAddress;

        sc = SetAddress(this,Port);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x) SetA failed, sc=%x.\n"),
                             (ADDRESS)this,sc));
            goto FailedToConfigure;
        }

        /* Release the lock
         */
        sc = UsbHubDropConfigurationLock(this);
        assert(sc == S_OK);

        /* The device is now connected, but we have no driver for it yet
         */
        this->Ports[Port].Status = USBM_EVENT_CONNECTED;

        /* Let us find a driver for it
         */
        sc = GetDeviceDescriptor(this,Port,MaximumControlPacketSize,
                                 &Descriptor);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x) GetDD failed, sc=%x\n"),
                             (ADDRESS)this, sc));
            goto TurnItOff;
        }

        /* What have we got here?
         */
        this->Ports[Port].Vid = Get2(Descriptor->IdVendor);
        this->Ports[Port].Pid = Get2(Descriptor->IdProduct);
        this->Ports[Port].Dev = Get2(Descriptor->BcdDevice);

        IsaHub = (Descriptor->DeviceClass == USB_DC_HUB);
        _stprintf(DriverName,_T("Vid%04x_Pid%04x_Dev%04x"),
                  this->Ports[Port].Vid,
                  this->Ports[Port].Pid,
                  this->Ports[Port].Dev);
        free((void*)Descriptor);

        /* Ask the PnP manager about it
         */
        _tcscpy(DriverCobName,_T("COB\\"));
        sc = PnpLookupDriver(/*in*/DriverName,/*out*/DriverCobName+4,nDriverName-4);

        /* If that didnt work, maybe we can still do something
         */
        if (FAILED(sc)) {
            struct USBM_REQUEST_BLOCK Urb2;
            UINT8 *Data2;
            SCODE sc2;

            /* We can handle a generic HUB, unless PnP says otherwise.
             */
            if (IsaHub) {
                _tcscpy(DriverCobName+4,_T("usbhub.cob"));
                sc = S_OK;
                goto GotSome;
            }
            /* We can also try our generic USBSTOR
             */
#define maxdesc 64
            Data2 = malloc(maxdesc);
            if (Data2 != NULL) {
                int i, ds;
                    
                sc2 = GetConfiguration(this,&Urb2,Data2,maxdesc,
                                       UsbAddress, Port,
                                       MaximumControlPacketSize);
                if (!FAILED(sc2)) {
                    for (i = 0; i < Urb2.ActualDataLength; i += ds) {
                        ds = Data2[i];
                        if (ds == 0) break;
                        if ((Data2[i+1] == USB_DT_IFACE) &&
                            (Data2[i+5] == USB_DC_STORAGE)) {
                            _tcscpy(DriverCobName+4,_T("usbstor.cob"));
                            sc = S_OK;
                            break;
                        }
                    }
                }
                free(Data2);
#undef maxdesc
            }
#ifndef _WANTS_USBNULL
#define _WANTS_USBNULL 1
#endif
#if _WANTS_USBNULL
            if (FAILED(sc)) {
                /* Starting the usbnull driver makes usbdir.exe work
                 */
                _tcscpy(DriverCobName+4,_T("usbnull.cob"));
                sc = S_OK;
            }
#endif
        }

        /* Do we know?
         */
    GotSome:
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x) no driver for %s\n"),(ADDRESS)this,
                             DriverName));
            continue;
        }

        /* Load the driver & find the constructor
         */
        DBGME(2,_tprintf(_T("Loading driver %s for %s\n"),DriverCobName,DriverName));
        sc = BindToObject(CurrentNameSpace(),
                          DriverCobName,
                          NAME_SPACE_READ | NAME_SPACE_WRITE,
                          &IID_IDeviceConstructor,
                          (void **) &pDevCons);

        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x) COB %s did not load, sc=%x\n"),(ADDRESS)this,
                             DriverCobName, sc));
            continue;
        }

        /* Invoke it with appropriate arguments
         */
        sc = pDevCons->v->Constructor(pDevCons,
                                      DriverName,
                                      0,0,
                                      &pUnk,
                                      NULL,
                                      (PTR) iUsbHub(this),
                                      (ADDRESS) UsbAddress,
                                      (ADDRESS) MaximumControlPacketSize,
                                      0,0,0);
        if (FAILED(sc)) {    /* constructor failed */
            DBGME(3,_tprintf(_T("UsbHub(%x): %s constructor failed sc=%x\n"),(ADDRESS)this,
                             DriverCobName, sc));
            continue;
        }
        pDevCons->v->Release(pDevCons);

        DBGME(0,_tprintf(_T("UsbHub(%x): %s constructor got %x. sc=%x\n"),(ADDRESS)this,
                         DriverCobName, (ADDRESS)pUnk, sc));

        /* Get the correct interface
         */
        sc = pUnk->v->QueryInterface(pUnk,
                                     &IID_IUsbDriver,
                                     (void **)&Driver);
        if (FAILED(sc)) {    /* wrong driver */
            DBGME(3,_tprintf(_T("UsbHub(%x): %s not an IUsbDriver?!? sc=%x\n"),(ADDRESS)this,
                             DriverCobName, sc));
            (void) pUnk->v->Release(pUnk);
            continue;
        }

        /* The device is now operational
         */
        this->Ports[Port].Driver = Driver;

        /* Namespace stuff, *before* we fully enable it.
         * NB: A HUB has done this already.
         * BTW: If we didnt already know IsaHub.. QI(IDriver,&IID_IUsbHub)
         */
        if (!IsaHub) {
            /* Regular guy. Stomp on old name jic.
             */
            sc = this->NameSpace->v->Register(this->NameSpace,
                                              DriverName,
                                              pUnk,
                                              NAME_SPACE_READ | 
                                              NAME_SPACE_WRITE |
                                              NAME_SPACE_CREATE,
                                              NULL);
            DBGME(0,_tprintf(_T("UsbHub(%x): Registered %s in NS %x. sc=%x\n"),(ADDRESS)this,
                             DriverName, (ADDRESS)this->NameSpace, sc));
            /* Ignore failure, even though it makes it useless??
             * But maybe it made a name for itself, oh well.
             */
            if (FAILED(sc)) {
                DBGME(3,_tprintf(_T("UsbHub(%x): register %s failed?!? sc=%x\n"),(ADDRESS)this,
                                 DriverName, sc));
            }
        }

        /* Let the IUnknown go, we'll keep the Driver
         */
        (void) pUnk->v->Release(pUnk);

        /* Enable the driver now
         */
        sc = Driver->v->IamYourHub(Driver,iUsbHub(this),Port);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x): IamYourHub %s failed?!? sc=%x\n"),(ADDRESS)this,
                             DriverName, sc));
        }

        /* Notify the HCI that we have put this address to good use
         * Ignore any error returns, this is optional.
         */
        Urb->DeviceAddress = 0xff;
        Urb->EndpointAddress = 0;
        Urb->Flags = URB_FLAG_CONTROL;
        Urb->Pid = USB_PID_EVENT;
        Urb->Data[0] = USBM_EVENT_CONNECTED;
        Urb->Data[1] = UsbAddress;

        sc = this->Hci->v->SubmitUrb(this->Hci,
                                     Urb,
                                     this->UpstreamPort);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x) ConnUrb failed sc=x%x\n"),(ADDRESS)this,sc));
        }

        /* Done!
         */
    }

    /* Cleanup before we go away
     * First off, the namespace so that we can become invisible.
     */
    this->NameSpace->v->Release(this->NameSpace);

    /* Then each of the active drivers.
     * A user will not be gracious to unplug all devices before the HUB.
     */
    for (Port = 0; Port < this->NumPorts; Port++)
        if (this->Ports[Port].Status == USBM_EVENT_CONNECTED) {
            this->Ports[Port].Status = USBM_EVENT_DISCONNECTED;
            Driver = this->Ports[Port].Driver;
            if (Driver) {
                sc = Driver->v->YouAreTerminated(Driver,
                                                 iUsbHub(this));
                if (FAILED(sc)) {
                    DBGME(3,_tprintf(_T("UsbHub(%x) YouAreTerminated2 failed sc=x%x\n"),(ADDRESS)this,sc));
                }
                (void) Driver->v->Release(Driver);
                this->Ports[Port].Driver = NULL;
            }
            /* Notify the HCI that we have reclaimed this address
             */
            Urb->DeviceAddress = 0xff;
            Urb->EndpointAddress = 0;
            Urb->Flags = URB_FLAG_CONTROL;
            Urb->Pid = USB_PID_EVENT;
            Urb->Data[0] = USBM_EVENT_DISCONNECTED;
            Urb->Data[1] = this->Ports[Port].UsbAddress;

            sc = this->Hci->v->SubmitUrb(this->Hci,
                                         Urb,
                                         this->UpstreamPort);
            if (FAILED(sc)) {
                DBGME(3,_tprintf(_T("UsbHub(%x) DiscUrb failed sc=x%x\n"),(ADDRESS)this,sc));
            }
            
        }

    /* No packets should have gone through during cleanup,
     * but they sure cannot go through now.  Besides, we are dead.
     */
    this->MyUsbAddress = 0;

    /* Release the HUB we are attached to, if any
     */
    if (this->Hub)
        this->Hub->v->Release(this->Hub);

    /* Relase the HCI
     */
    this->Hci->v->Release(this->Hci);

    Condition_Destroy(&this->SleepPoint);

    /* We are really done now
     */
 Out:
    if (Urb)
        free(Urb);
    if (DriverName)
        free(DriverName);
    if (DriverCobName)
        free(DriverCobName);
    /* Dont do this, we cant really have another one try
     */
    //this->ThreadStarted = FALSE;
}

/* UsbHub IUnknown
 */
PRIVATE SCODE UsbHubQueryInterface (
    PIUSBHUB pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbHubQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));
    return UsbHubDeviceQueryInterface(iUsbHubDevice(pUsbHub(pThis)), iid, ppObject);
}

PRIVATE UINT UsbHubAddRef (
    PIUSBHUB pThis)
{
    DBGME(0,_tprintf(_T("UsbHubAR(%x)\n"),(ADDRESS)pThis));
    return UsbHubDeviceAddRef(iUsbHubDevice(pUsbHub(pThis)));
}

PRIVATE UINT UsbHubRelease (
    PIUSBHUB pThis)
{
    DBGME(0,_tprintf(_T("UsbHubRel(%x)\n"),(ADDRESS)pThis));
    return UsbHubDeviceRelease(iUsbHubDevice(pUsbHub(pThis)));
}

/* Handle some requests in a RootHub
 */
PRIVATE SCODE UsbHubHandleRequest (
    PUSBHUB this,
    /*inout*/ PUSBM_REQUEST_BLOCK Urb)
{
    const UINT8 *Data;
    UINT16 Length;

    static const UINT8 DeviceDescriptor[] = {
        sizeof(USB_DEVICE_DESCRIPTOR),
        USB_DT_DEVICE,
        Con2(USB_v2_0),
        USB_DC_HUB, 0, 0, /* class,subclass,protocol */
        64,               /* max packet size */
        Con2(0x45e), Con2(0), Con2(0),   /* vendor,product,device*/
        1, 2, 0,          /* ix of manuf,product,serial number */
        1                 /* num configurations */
    };
    static const UINT8 ConfigDescriptor[] = {
        sizeof(USB_CONFIGURATION_DESCRIPTOR),
        USB_DT_CONFIG,
        Con2(sizeof(USB_CONFIGURATION_DESCRIPTOR)),   /* total length of all descs returned */
        1,               /* num interfaces */
        1,               /* config value */
        0,               /* ix of config */
        USB_AT_SELF_POWERED,  /* attributes */
        1                     /* max power */
    };
#if (BYTE_ORDER == BIG_ENDIAN)
#define _t(x) ByteSwap16BitConstant(_T(x))
#define _e(x) ByteSwap16BitConstant(x)
#else
#define _t(x) _T(x)
#define _e(x) (x)
#endif
    static const UINT16 String0[] = {
        _e((USB_DT_STRING<<8) | (2*2)),
        _e(0x409)
    };
    static const UINT16 String1[] = {
        _e((USB_DT_STRING<<8) | (10*2)),
        /* "Microsoft" */
        _t('M'),_t('i'),_t('c'),_t('r'),_t('o'),_t('s'),_t('o'),_t('f'),_t('t'),_t(' ')
    };

    static const UINT16 String2[] = {
        _e((USB_DT_STRING<<8) | (18*2)),
        /* "Software Root Hub" */
        _t('S'),_t('o'),_t('f'),_t('t'),_t('w'),_t('a'),_t('r'),_t('e'),_t(' '),
        _t('R'),_t('o'),_t('o'),_t('t'),_t(' '),
        _t('H'),_t('u'),_t('b')
    };
#undef _t

    if (Urb->SetupData[1] != USB_R_GET_DESCRIPTOR)
        return E_NOT_IMPLEMENTED;

    switch (Urb->SetupData[3]) {
    case USB_DT_DEVICE:
        Data = DeviceDescriptor;
        Length = sizeof DeviceDescriptor;
        break;
    case USB_DT_CONFIG:
        Data = ConfigDescriptor;
        Length = sizeof ConfigDescriptor;
        break;
    case USB_DT_STRING:
        if (Urb->SetupData[2] == 0) {
            Data = (UINT8*) String0;
            Length = sizeof String0;
        }
        else
        if (Urb->SetupData[2] == 1) {
            Data = (UINT8*) String1;
            Length = sizeof String1;
        }
        else
        if (Urb->SetupData[2] == 2) {
            Data = (UINT8*) String2;
            Length = sizeof String2;
        }
        else {
            /* BUGBUG should STALL instead. Urgh */
            return E_NOT_IMPLEMENTED;
        }
        break;
    default:
        return E_NOT_IMPLEMENTED;
    }

    if (Urb->ExpectedDataLength < Length)
        Length = Urb->ExpectedDataLength;

    memcpy(Urb->Data,Data,Length);
    if (this->IsUsb20 &&
        (Urb->SetupData[3] == USB_DT_DEVICE) &&
        (Length > 6)) {
        Urb->Data[6] = 1; /* Single TT */
    }
    Urb->ActualDataLength = Length;
    Urb->Flags |= URB_FLAG_COMPLETE;
    Condition_InterruptSignal(Urb->Done);

    return S_OK;
}

/* Handle a USB Request Block submission.
 * If it is addressed to a RootHub it is handled here,
 * otherwise it is forwarded to the underlying HCI
 */
PRIVATE SCODE UsbHubSubmitUrb (
    PIUSBHUB pThis,
    /*inout*/ PUSBM_REQUEST_BLOCK Urb,
    UINT8 PortNumber)
{
    PUSBHUB this = pUsbHub(pThis);
    UINT8 UsbAddress;
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbHub(%x):%sSubU(%x)\n"),(ADDRESS)this,
                     (IAmRoot(this) ? _T("Root") :_T("")),
                     (ADDRESS)Urb));

    /* Is it for me-a-root-hub ?
     */
    if (IAmRoot(this)) {
        /* Filter any EVENT URBs to our HC
         */
        if (Urb->Pid == USB_PID_EVENT) {
            /* Fully handle address allocation requests
             */
            if (Urb->Data[0] == USBM_EVENT_ADDRESS_ALLOCATION) {
                UsbAddress = UsbHubAllocateAddress(this);
                Urb->Data[1] = UsbAddress;
                Urb->Flags |= URB_FLAG_COMPLETE |
                    ((UsbAddress == 0) ? URB_FLAG_ERROR : 0);
                return S_OK;
            }

            /* Ditto for the configuration lock
             */
            if (Urb->Data[0] == USBM_EVENT_TAKE_CONFIGURATION_LOCK) {
                sc = UsbHubTakeConfigurationLock(this);
                Urb->Flags |= URB_FLAG_COMPLETE;
                return sc;
            }

            if (Urb->Data[0] == USBM_EVENT_DROP_CONFIGURATION_LOCK) {
                sc = UsbHubDropConfigurationLock(this);
                Urb->Flags |= URB_FLAG_COMPLETE;
                return sc;
            }

            /* Handle deallocations, forward them to the HC as well
             */
            if (Urb->Data[0] == USBM_EVENT_DISCONNECTED) {
                UsbAddress = Urb->Data[1];
                if (!UsbHubReleaseAddress(this,UsbAddress)) {
                    DBGME(3,_tprintf(_T("UsbHub(%x): ?rel%d?\n"),
                                     (ADDRESS)this,UsbAddress));
                    return E_INVALID_PARAMETER;
                }
                goto SubmitIt;
            }

            /* Forward fully-connected notifications, jic.
             */
            if (Urb->Data[0] == USBM_EVENT_CONNECTED) {
                goto SubmitIt;
            }

            /* Nothing else should have come here
             */
            DBGME(3,_tprintf(_T("UsbHub(%x): ?req%d?\n"),(ADDRESS)this,Urb->Data[0]));
            return E_INVALID_PARAMETER;
        }

        /* Also capture URBs that are infact addressed to us
         */
        if (Urb->DeviceAddress == this->MyUsbAddress)
            return UsbHubHandleRequest(this,Urb);
    }

    /* Nope, does that thing exist?
     * NB: To avoid leaking USB addresses we must allow some URBs even though
     * the port has been disconnected.  This is the case, for instance, of a
     * user disconnecting a non-root HUB that has active devices attached.
     */
    if ((PortNumber >= this->NumPorts) ||
        ((this->Ports[PortNumber].Status != USBM_EVENT_CONNECTED) &&
         (Urb->DeviceAddress != 0xff))) {
        DBGME(3,_tprintf(_T("UsbHub(%x): ?Sub %d\n"),(ADDRESS)this,PortNumber));
        return E_INVALID_PARAMETER;
    }

    /* Speed issues
     * Only fix the flags if it is not a hub,
     * or if the hub is speaking to itself.
     */
    if ((this->Ports[PortNumber].HubVersion == 0) ||
        (Urb->DeviceAddress == this->Ports[PortNumber].UsbAddress))
        Urb->Flags |= this->Ports[PortNumber].SpeedFlags;

    /* Routing issues
     */
    if (0 == (this->Ports[PortNumber].SpeedFlags & URB_FLAG_HIGH_SPEED)) {
        Urb->HubDeviceAddress = this->MyUsbAddress;
        Urb->HubPortNumber = PortNumber+1;/* we are zero-based */
    }

 SubmitIt:
    return this->Hci->v->SubmitUrb(this->Hci,
                                Urb,
                                this->UpstreamPort);
}

/* Handle a USB Request Block cancellation request.
 */
PRIVATE SCODE UsbHubCancelUrb (
    PIUSBHUB pThis,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 PortNumber)
{
    PUSBHUB this = pUsbHub(pThis);

    DBGME(0,_tprintf(_T("UsbHub(%x):CanU(%x)\n"),(ADDRESS)this,(ADDRESS)Urb));

    /* Check this->Ports[PortNo].Status */
    return this->Hci->v->CancelUrb(this->Hci,
                                Urb,
                                this->UpstreamPort);
}

/* Return the number of attached ports
 */
PRIVATE SCODE UsbHubQueryPorts (
    PIUSBHUB pThis,
    /*out*/ UINT * nPorts)
{
    PUSBHUB this = pUsbHub(pThis);

    DBGME(0,_tprintf(_T("UsbHub(%x):QP(%x)\n"),(ADDRESS)this,(ADDRESS)nPorts));

    *nPorts = this->NumPorts;

    return S_OK;
}

/* UsbHubDriver IUnknown
 */
PRIVATE SCODE UsbHubDriverQueryInterface (
    PIUSBDRIVER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbHubDrivQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));
    return UsbHubDeviceQueryInterface(iUsbHubDevice(pUsbHubDriver(pThis)), iid, ppObject);
}

PRIVATE UINT UsbHubDriverAddRef (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbHubDrivAR(%x)\n"),(ADDRESS)pThis));
    return UsbHubDeviceAddRef(iUsbHubDevice(pUsbHubDriver(pThis)));
}

PRIVATE UINT UsbHubDriverRelease (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbHubDrivRel(%x)\n"),(ADDRESS)pThis));
    return UsbHubDeviceRelease(iUsbHubDevice(pUsbHubDriver(pThis)));
}

/* A HUB driver tells us it is the one we should talk to.
 * For uniformity, also called on RootHubs with HUB==NULL.
 * This function creates our worker thread.
 */
PRIVATE SCODE UsbHubDriverIamYourHub (
    PIUSBDRIVER pThis,
    PIUSBHUB Hub,
    UINT PortNumber)
{
    PUSBHUB this = pUsbHubDriver(pThis);

    DBGME(0,_tprintf(_T("UsbHub(%x):IYH(%x,%d) %d\n"),
                     (ADDRESS)this,(ADDRESS)Hub,PortNumber,this->ThreadStarted));

    /* NB: this is not called on a root HUB
     * Well, in that case we are self-calling and Hub==NULL; Sigh.
     */
    this->UpstreamPort = (UINT8) PortNumber;

    /* For a non-root hub, start the ThreadLoop now.
     */
    if (!this->ThreadStarted) {
        /* NB: Do *not* use CurrentProcess(), we are on a borrowed thread
         */
        PIPROCESS pPrc = NULL;
        PIMODULE pMod;
        PITHREAD pThd;
        SCODE sc;

        /* Now that we know our port we can interrogate the hardware
         * We need to get an upper bound on the number of supported ports.
         * This also initializes the HUB and makes it ready for events.
         */
        if (Hub != NULL) {
            UsbHubConfigureHub(this);
            AllocatePorts(this);
        }

        sc = GetModuleContainsAddr( (ADDRESS) UsbHubDeviceThreadLoop, &pMod);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x): !gmca sc=%x\n"),(ADDRESS)this,sc));
            return sc;
        }

        sc = pMod->v->QueryInterface(pMod,&IID_IProcess,(void **)&pPrc);
        pMod->v->Release(pMod);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x): !qi sc=%x\n"),(ADDRESS)this,sc));
            return sc;
        }

        sc = pPrc->v->CreateThread(pPrc, CREATE_THREAD_STATIC_SCOPING,
                                   (THREAD_FUNCTION) UsbHubDeviceThreadLoop,
                                   (THREAD_ARGUMENT) iUsbHubDevice(this),
                                   0, NULL, &pThd);
        pPrc->v->Release(pPrc);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x): !created ThreadLoop sc=%x\n"),(ADDRESS)this,sc));
            return sc;
        }
        this->ThreadStarted = TRUE;
    }
    return S_OK;
}

/* Our HUB tells us to terminate, ordinarily because the HUB device
 * has been unplugged from the USB hub, or it died somehow.
 */
PRIVATE SCODE UsbHubDriverYouAreTerminated (
    PIUSBDRIVER pThis,
    PIUSBHUB SaysWho)
{
    PUSBHUB this = pUsbHubDriver(pThis);

    DBGME(0,_tprintf(_T("UsbHub(%x):YAT(%x)\n"),(ADDRESS)this,(ADDRESS)SaysWho));

    /* Wakeup the ThreadLoop, let it do the cleaning
     */
    this->ThreadShouldExit = TRUE;
    Condition_InterruptSignal(&this->SleepPoint);
    return S_OK;
}

/* Return our own USB address and the upstream port where we are connected.
 */
PRIVATE SCODE UsbHubDriverGetUsbAddress (
    PIUSBDRIVER pThis,
    PUINT UsbAddress,
    PUINT PortNumber)
{
    PUSBHUB this = pUsbHubDriver(pThis);

    DBGME(0,_tprintf(_T("UsbHub(%x):GUA(%x)\n"),(ADDRESS)this,(ADDRESS)UsbAddress));

    if (UsbAddress == NULL) {
        return E_INVALID_PARAMETER;
    }
    if (this->MyUsbAddress == 0)
        return E_OBJECT_NOT_INITIALIZED;

    *UsbAddress = this->MyUsbAddress;
    *PortNumber = this->UpstreamPort;

    return S_OK;
}

/* Vtables
 */
#include "s_UsbHub_v.c"

/* Constructor
 */
PIDEVICE UsbHubNew(
    const _TCHAR *MyName,
    _TCHAR *HciName,
    UINT8 UsbAddress,
    UINT8 MaximumControlPacketSize,
    PIUNKNOWN /*PIUSBMASTERCONTROLLER*/ UsbTreeLocation)
{
    PINAMESPACE NameSpace = CurrentNameSpace();
    PIHEAP pHeap = CurrentHeap();
    SCODE sc;
    PUSBHUB this = (PUSBHUB) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *this, 0);
    if (!this)
        return NULL;

    this->CDevice.v = (struct CDeviceVtbl*)&UsbHubDeviceUsbHubVtbl;
    this->CDevice.RefCnt = 1;
    this->UsbHubv = &UsbHubUsbHubVtbl;
    this->UsbHubDriverv = &UsbHubDriverUsbHubVtbl;

    this->MyUsbAddress = UsbAddress;
    this->MaximumControlPacketSize = MaximumControlPacketSize;
    this->UsbAddressesInUse[0] = 1; /* 0 is invalid, regardless */
    Condition_Init(&this->SleepPoint);
    Mutex_Init(&this->ConfigurationLock);

    /* Have we been started by the autoconfig code?
     * We migth start ourselves when we find an attached HUB.
     * Make sure there are only two ways this can work.
     */
    if ((UsbTreeLocation == NULL) && (HciName == NULL))
        return NULL;

    if (UsbTreeLocation == NULL) {	
        /* Autoconfig. We must at least have an HCI
         */
        sc = BindToObject(NameSpace,
                          HciName,
                          NAME_SPACE_READ | NAME_SPACE_WRITE,
                          &IID_IUsbMasterController,
                          (void **) &this->Hci);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x): HCI %s not found sc=%x\n"),(ADDRESS)this,HciName,sc));
            goto Nope;
        }
    } else {
        /* We are a regular HUB driver.
         * Get ahold of our upstream (root?) HUB
         */
        if (HciName == NULL) /* sanity */
            HciName = _T("NonameHci");
        sc = UsbTreeLocation->v->QueryInterface(UsbTreeLocation,
                                                &IID_IUsbMasterController,
                                                (void **)&this->Hci);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x): %s not an HCI sc=%x\n"),(ADDRESS)this,HciName,sc));
            goto Nope;
        }

        /* Find its namespace. NB: this is not optional
         */
        sc = UsbTreeLocation->v->QueryInterface(UsbTreeLocation,
                                                &IID_INameSpace,
                                                (void **)&NameSpace);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x): no namespace for %s sc=%x\n"),(ADDRESS)this,HciName,sc));
            goto Nope;
        }

    }

    /* See if we are going to be a root HUB
     * [Usually its because we are started by autoconfig, but.]
     * An actual HCI chip driver does NOT implement the IUsbHub interface.
     */
    sc = this->Hci->v->QueryInterface(this->Hci,
                                      &IID_IUsbHub,
                                      (void **)&this->Hub);
    if (FAILED(sc)) {
        DBGME(2,_tprintf(_T("UsbHub(%x): %s @%x is a RootHub for %s (sc=%x)\n"),(ADDRESS)this,
                         MyName,(ADDRESS)this, HciName,sc));
        this->Hub = NULL;
    }

    /* For a real hub, find out number and state of ports
     * BUGBUG : Add methods to IUsbMasterController!
     */
    if (IAmRoot(this)) {
        PIFILE HciFile;
        UINT64 NumPorts;

        sc = this->Hci->v->QueryInterface(this->Hci,
                                          &IID_IFile,
                                          (void **)&HciFile);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x): %s no IFile (sc=%x)\n"),(ADDRESS)this,HciName,sc));
            goto Nope;
        }

        sc = HciFile->v->GetSize(HciFile, &NumPorts);
        (void) HciFile->v->Release(HciFile);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbHub(%x): Hci::GetSize failed? sc=%x\n"),(ADDRESS)this,sc));
            goto Nope;
        }
        this->NumPorts = (UINT8) Uint64ToUint32(NumPorts);
        this->IsUsb20  = (UINT16) (Uint64ToUint32(NumPorts) >> 16);
        this->MyUsbAddress = 0xff;

        if (!AllocatePorts(this))
            goto Nope;

    }  else {
        /* Query the HUB for the number of supported ports
         * Unfortunately, we do not yet know our UpstreamPort and we cannot.
         */
        this->NumPorts = 0; /* fornow */
    }

    /* Namespace stuff
     * We need to do this to create the hybrid namespace
     * NB: Preempt the config code ifso, enter our name now.
     */
    sc = BindToObject(NameSpace,
                      MyName,
                      NAME_SPACE_READ | 
                      NAME_SPACE_WRITE |
                      NAME_SPACE_CREATE,
                      &IID_INameSpace,
                      (void **) &this->NameSpace);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x): cant Bind %s, sc=%x\n"),(ADDRESS)this,MyName,sc));
        /* ignore?? */
    }

    DBGME(0,_tprintf(_T("UsbHub(%x):New NameSpace %x, nPorts=%d sc=%x\n"),(ADDRESS)this,
                     (ADDRESS)this->NameSpace,this->NumPorts,sc));

    sc = NameSpace->v->Register(NameSpace,
                                MyName,
                                (PIUNKNOWN) this,
                                NAME_SPACE_READ | 
                                NAME_SPACE_WRITE |
                                NAME_SPACE_CREATE,
                                NULL);

    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbHub(%x): cant Register %s, sc=%x\n"),(ADDRESS)this,MyName,sc));
        /* ignore?? */
    }

    /* For a root hub, start the ThreadLoop now.
     * A non-root hub must wait for the UpstreamPort instead.
     */
    if (IAmRoot(this)) {
        (void) UsbHubDriverIamYourHub(iUsbHubDriver(this),NULL,0);
    }

    return iUsbHubDevice(this);
 Nope:
    if (this->Ports)
        free(this->Ports);
    if (this->Hub)
        this->Hub->v->Release(this->Hub);
    if (this->Hci)
        this->Hci->v->Release(this->Hci);
    if (NameSpace != CurrentNameSpace())
        NameSpace->v->Release(NameSpace);

    free(this);
    return NULL;
}

/* ** 
*/
SCODE __cdecl UsbHubConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * MyName,
        ADDRESS UnusedRegisterBase,
        UINT UnusedInterrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        _TCHAR *HciName,
        PIUNKNOWN /*PIUSBMASTERCONTROLLER*/ UsbTreeLocation,
        ADDRESS UsbAddress,
        ADDRESS MaximumControlPacketSize)
{
    PIUNKNOWN pUnk;

    UnusedParameter(UnusedRegisterBase);
    UnusedParameter(UnusedInterrupt);

    DBGME(0,_tprintf(_T("UsbHubConstructor(%x %s %x %s %d %x)\n"),
                     (ADDRESS) pThis, MyName, (ADDRESS) ppUnk,
                     (HciName) ? HciName : _T("NULL"),
                     UsbAddress, (ADDRESS) UsbTreeLocation));

    pUnk = (PIUNKNOWN) UsbHubNew(MyName,
                                 HciName,
                                 (UINT8) UsbAddress,
                                 (UINT8) MaximumControlPacketSize,
                                 UsbTreeLocation);

    DBGME(0,_tprintf(_T("UsbHubConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UsbHubConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UsbHubDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UsbHubConstructor
};

static struct _DRIVERCOB UsbHubDriverCobObject = { &UsbHubDriverCobVtable, 1 };

PIUNKNOWN UsbHubCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UsbHubDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UsbHubDeviceUsbHubVtbl, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &UsbHubDriverCobObject;
}
