/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbkeybd.c
 *
 * Purpose:
 *
 *    Device driver for USB keyboards.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <stddef.h> /* for offsetof() */
//#define _DEBUG 1
#define DBGLVL 2
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers/drivers.h>
#include <drivers/usb.h>
#include <drivers/keybd_hw.h>
#include <usbmaster.h>

#define ST_RAW_MODE    0x8000
#define ST_AUTOREPEAT2 0x4000
#define ST_AUTOREPEAT1 0x2000
#define ST_AUTOREPEAT  0x6000
#define ST_NUMLOCK     0x0001
#define ST_CAPSLOCK    0x0002
#define ST_SCRLLOCK    0x0004
# define ST_LEDMASK    0x0007
#include "usbkbtable.c"

#define K_META     0x44
#define K_CONTROL  0x11
#define K_SHIFT    0x22
#define K_WIN      0x88

typedef struct _USBKEYBD {
    struct CDevice CDevice;
    CONST_VTBL struct IUsbDriverVtbl *UsbKeybdDriverv;
    PIUSBHUB Hub;
    PUSBM_REQUEST_BLOCK AvailUrb;
    UINT8 MyUsbAddress;
    UINT8 PortNo;
    UINT8 InEndpoint;
    UINT8 PollingInterval;
    UINT8 MaximumControlPacketSize;
    UINT8 InReportSize;
    UINT8 OutReportSize;
#if defined(m68k)
    UINT NextKeyIndex;
    UINT Interface;
    UINT ReportDescriptorSize;
#else
    UINT8 NextKeyIndex;
    UINT8 Interface;
    UINT8 ReportDescriptorSize;
#endif
    UINT16 MaximumPacketSize;
    UINT16 KeyboardNo;
#define MYNAMELENGTH 12 /* bignuf */
#define MakeMyName(_this_,_name_)  _stprintf(_name_,_T("keyboard%d"),(_this_)->KeyboardNo)
    UINT16 State; /* see above */
    UINT8 *LastReport;
    UINT8 *LastRawReport;
    const UINT8 *RegularTable;
    const UINT8 *ShiftTable;
    const UINT8 *ControlTable;
    const UINT8 *MetaTable;
} USBKEYBD, *PUSBKEYBD;

/* Convert between interface and instance pointer and vice versa */
#define pUsbKeybdDevice(_i_) ((PUSBKEYBD)((_i_)))
#define iUsbKeybdDevice(_p_) ((PIDEVICE)((_p_)))
#define pUsbKeybdDriver(_i_) ((PUSBKEYBD)((ADDRESS)(_i_) - offsetof(struct _USBKEYBD, UsbKeybdDriverv)))
#define iUsbKeybdDriver(_p_) ((PIUSBDRIVER)(&((PUSBKEYBD)(_p_))->UsbKeybdDriverv))

PRIVATE SCODE UsbKeybdDeviceQueryInterface (
    PIDEVICE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbKeybdDevQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));

    /* IDevice
     */
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_IFile,iid)
        || UuidCmp(&IID_IDevice,iid))
    {
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }

    /* IUsbDriver
     */
    if (UuidCmp(&IID_IUsbDriver, iid)) {
        /* Use a single refcount
         */
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (void*) iUsbKeybdDriver(pThis);
        return S_OK;
    }

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT UsbKeybdDeviceAddRef (
    PIDEVICE pThis)
{
    DBGME(0,_tprintf(_T("UsbKeybdDevAR(%x)\n"),(ADDRESS)pThis));
    return AtomicInc(&pUsbKeybdDevice(pThis)->CDevice.RefCnt);
}

PRIVATE UINT UsbKeybdDeviceRelease (
    PIDEVICE pThis)
{
    PUSBKEYBD this = pUsbKeybdDevice(pThis);
    UINT Refs = AtomicDec(&this->CDevice.RefCnt);
    DBGME(0,_tprintf(_T("UsbKeybdDevRel(%x,%d)\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        if (this->AvailUrb)
            pHeap->v->Free(pHeap, 0, (void *) this->AvailUrb);

        pHeap->v->Free(pHeap, 0, (void *) this);
    }
    return Refs;
}

/* Issue: When two/three keys are pressed down almost at the same time 
 * we get reports with sequences like [0] [a] [ab] [bc] [c] [0].
 * We would like to map that to [0] [a] [b] [c] [0].
 */
PRIVATE void FilterCopy(
    UINT8 *Destination,
    UINT8 *PreviousInput,
    UINT8 *CurrentInput,
    int MaxLen)
{
#if 0
    memcpy(Destination,CurrentInput,MaxLen);
#else
    UINT8 c1, c2;
    int i, j, n;

    Destination[0] = CurrentInput[0]; /* flags */
    Destination[1] = CurrentInput[1]; /* whatever this is */

    /* Remove all duplicates.  Example input:
     * gr8 0: 0 19 0 0 0 0 0..   'p'
     * gr8 0: 0 19 18 0 0 0 0..  'po'
     * gr8 0: 0 19 18 12 0 0 0.. 'poi'
     * gr8 0: 0 18 12 0 0 0 0..  'oi'
     * gr8 0: 0 12 0 0 0 0 0..   'i'
     * gr8 0: 0 0 0 0 0 0 0..    ==> "poi"
     */
    DBGME(2,_tprintf(_T("[%d %d:"),Destination[0],Destination[1]));
    n = 2;
    for (i = 2; i < MaxLen;) {
        c1 = CurrentInput[i];
        if (c1 == 0) break;

        for (j = 2; j < MaxLen; j++) {
            c2 = PreviousInput[j];
            if (c2 == 0) break;
            /* drop ? */
            if (c1 == c2) 
                goto DropIt;
        }
        DBGME(2,_tprintf(_T(" %d"), c1));
        Destination[n++] = c1;
    DropIt:
        i++;
    }
    /* Zero out whatever room is left */
    for (;n < MaxLen; n++)
        Destination[n] = 0;
    DBGME(2,_tprintf(_T(" 0..]\n")));

    /* Preserve the raw data for the next time */
    memcpy(PreviousInput,CurrentInput,MaxLen);
#endif
}


/* Read from the IN endpoint one reportfull
 */
PRIVATE SCODE GetReport(
    PUSBKEYBD this)
{
    SCODE sc;
    TIME Timeout, *pTimeout;
    PUSBM_REQUEST_BLOCK Urb = this->AvailUrb;

    if ((Urb == NULL || (this->Hub == NULL))) {
        /* Must have been terminated
         */
        DBGME(3,_tprintf(_T("UsbKeybd:GetReport no Urb\n")));
        return E_OBJECT_NOT_FOUND;
    }

    /* The URB might still be active because we timed out last time,
     * in which case we are in AUTOREPEAT2.
     * Or it might have been completed in the mean time.
     * Either way, we use *our* field ExpectedDataLength as a flag,
     * to indicate that the URB is "own by the HC".
     */
    if (Urb->ExpectedDataLength == 0) {

        Urb->Done = &this->CDevice.IntrPoint;
        Urb->DeviceAddress = this->MyUsbAddress;
        Urb->EndpointAddress = this->InEndpoint;
        Urb->Pid = USB_PID_IN;
        Urb->Flags = URB_FLAG_IN|URB_FLAG_INTERRUPT;
        Urb->SetupLength = 0;
        Urb->Interval = this->PollingInterval;
        Urb->ExpectedDataLength = (UINT8) this->InReportSize;
        Urb->MaximumPacketSize = this->MaximumPacketSize;

        sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbKeybd:GetReport sc=%x\n"),sc));
            return sc;
        }
    }

    /* Autorepeat?
     */
    pTimeout = &Timeout;
    if (this->State & ST_AUTOREPEAT2) {
        Int32ToInt64(Timeout,TIME_RELATIVE(TIME_MILLIS(16)));
    } else if (this->State & ST_AUTOREPEAT1) {
        Int32ToInt64(Timeout,TIME_RELATIVE(TIME_MILLIS(500)));
    } else {
        pTimeout = NULL;
    }

    sc = Condition_Wait(Urb->Done,NULL,pTimeout);

    /* Check results: might have been terminated, or no data available
     */
    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbKeybd:GetReport er=%x.%x\n"),
                         Urb->Flags, Urb->FinalStatus));

        /* The URB must be reposted */
        Urb->ExpectedDataLength = 0;

        return E_MEDIA_CHANGED;
    }

    /* Also check it did not timeout (autorpt)
     */
    if ((Urb->Flags & URB_FLAG_COMPLETE) && Urb->ActualDataLength) {
        DBGME(2,_tprintf(_T("gr%d %x: %d %d %d %d %d %d %d..\n"),
                         Urb->ActualDataLength,
                         Urb->Data[0], Urb->Data[1],
                         Urb->Data[2], Urb->Data[3],
                         Urb->Data[4], Urb->Data[5],
                         Urb->Data[6], Urb->Data[7]));

        /* Filter and validate the new report
         */
        this->NextKeyIndex = 2;
        FilterCopy(this->LastReport,
                   this->LastRawReport,
                   Urb->Data,
                   this->InReportSize);

        /* All-keys-up means we are no longer in autorepeat mode.
         * Otherwise we restart from the beginning
         */
        this->State &= ~ST_AUTOREPEAT;
        if (Urb->Data[2] != 0)
            this->State |= ST_AUTOREPEAT1;

        /* The URB must be reposted */
        Urb->ExpectedDataLength = 0;

    } else {
        if (!(Urb->Flags & URB_FLAG_COMPLETE)) {
            /* Bump up the autorept state, with saturation
             */
            UINT16 Astate = this->State + ST_AUTOREPEAT1;
            if ((Astate & ST_AUTOREPEAT) != ST_AUTOREPEAT)
                this->State = Astate;
            /* else we stay in AST_AUTOREPEAT2 */

            /* And the previous report stays
             */
            this->NextKeyIndex = 2;

        } else {
            /* The URB must be reposted */
            Urb->ExpectedDataLength = 0;
        }
    }

    return S_OK;
}

PRIVATE SCODE SetReport(
    PUSBKEYBD this,
    PUSBM_REQUEST_BLOCK Urb)
{
    SCODE sc;

    if ((Urb == NULL || (this->Hub == NULL))) {
        /* Must have been terminated
         */
        DBGME(3,_tprintf(_T("UsbKeybd:SetReport no Urb\n")));
        return E_OBJECT_NOT_FOUND;
    }
    Urb->Done = &this->CDevice.IntrPoint;
    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->EndpointAddress = 0;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_OUT | URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_CLASS|USB_RT_IFACE;
    Urb->SetupData[1] = USB_R_SET_REPORT;
    Urb->SetupData[2] = 0;
    Urb->SetupData[3] = USB_RP_OUTPUT;
    Urb->SetupData[4] = this->Interface;
    Urb->SetupData[5] = 0;
    Set2(Urb->SetupData+6,((UINT16)this->OutReportSize));
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = (UINT8) this->OutReportSize;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbKeybd:SetReport sc=%x\n"),sc));
        return sc;
    }

    Condition_Wait(Urb->Done,NULL,NULL);

    /* Check results: migth have been terminated
     */
    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbKeybd:SetReport er=%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}

PRIVATE SCODE SetLeds(
    PUSBKEYBD this,
    PUSBM_REQUEST_BLOCK Urb)
{
    Urb->Data[0] = this->State & ST_LEDMASK;
    return SetReport(this,Urb);
}

/* The ethernal q: raw or translated?
 * Use the offset to tell.
 */
PRIVATE SCODE ReadAtTranslated(
    PUSBKEYBD this,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    BYTE Flags;
    BYTE Key;
    BOOL ChangeLeds = FALSE;
    SCODE sc;
    const UINT8 *Table = NULL;
    PUSBM_REQUEST_BLOCK Urb = this->AvailUrb;

    DBGME(0,_tprintf(_T("UsbKeybd:RAted(%x,%x,%d)\n"),
                     (ADDRESS)this,(ADDRESS)pBuffer,Size));

    *pSizeRead = 0;

    /* Do we have some.
     */
 TryAgain:
    if (ChangeLeds) {
        SetLeds(this,Urb);
        ChangeLeds = FALSE;
    }

    while (this->NextKeyIndex >= this->InReportSize) {
        sc = GetReport(this);
        if (FAILED(sc))
            return sc;
    }

    Flags = this->LastReport[0];
    if (Flags & K_META)
        Table = this->MetaTable;
    else if (Flags & K_CONTROL)
        Table = this->ControlTable;
    else if ((Flags & K_SHIFT) || (this->State & ST_CAPSLOCK))
        Table = this->ShiftTable;
    else
        Table = this->RegularTable;

    do {
        Key = this->LastReport[this->NextKeyIndex++];

        /* Translate the keycode
         */
        Key = Table[Key];
        DBGME(0,_tprintf(_T(":%x "),Key));

        if ((Key == CAPS_LOCK) || (Key == NUM_LOCK) || (Key == SCROLL_LOCK)) {
            /* Flip it
             */
            this->State = this->State ^ KeyToState(Key);

            /* Turn the LED on/off (now!)
             */
            ChangeLeds = TRUE;

            /* Keep processing keys. Dubious what we should do with them but.
             */
        } else
        if (Key) {
            *pBuffer++ = Key;
            *pSizeRead = *pSizeRead + 1;
            Size--;
        }

    } while (Size && (this->NextKeyIndex < this->InReportSize));

    /* If we got nothing try again
     */
    if (*pSizeRead == 0)
        goto TryAgain;

    /* Dont forget to change the LEDs, jic
     */
    if (ChangeLeds) {
        SetLeds(this,Urb);
    }

    DBGME(1,_tprintf(_T("UsbKeybd:RAted %d %c\n"),*pSizeRead,pBuffer[0-(*pSizeRead)]));
    return S_OK;
}
     
PRIVATE SCODE UsbKeybdDeviceReadAt (
    PIDEVICE pThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PUSBKEYBD this = pUsbKeybdDevice(pThis);
    PKEYBOARD_INPUT_DATA Input = (PKEYBOARD_INPUT_DATA)pBuffer;
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbKeybd:RAt(%x,%x,%d)\n"),
                     (ADDRESS)this,(ADDRESS)pBuffer,Size));

    if ((this->State & ST_RAW_MODE) == 0)
        return ReadAtTranslated(this,pBuffer,Size,pSizeRead);

    if (Size != sizeof(KEYBOARD_INPUT_DATA))
        return E_BUFFER_TOO_SMALL;

    /* Do we have some.
     */
    if (this->NextKeyIndex >= this->InReportSize) {
        sc = GetReport(this);
        if (FAILED(sc))
            return sc;
    }

    /* Pick em up then
     */
    Input->Flags = this->LastReport[0];
    Input->MakeCode = this->LastReport[this->NextKeyIndex++];
    /* BUGBUG...drop error codes though;
     */

    Input->ReportTime = CurrentTime();

    return S_OK;
}

PRIVATE SCODE UsbKeybdDeviceGetSize (
    PIDEVICE pThis,
    /*out*/ UINT64 * Size)
{
    PUSBKEYBD this = pUsbKeybdDevice(pThis);

    Uint32ToUint64(*Size,this->InReportSize);

    return S_OK;
}

PRIVATE SCODE UsbKeybdDriverQueryInterface (
    PIUSBDRIVER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbKeybdDrivQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));
    return UsbKeybdDeviceQueryInterface(iUsbKeybdDevice(pUsbKeybdDriver(pThis)), iid, ppObject);
}

PRIVATE UINT UsbKeybdDriverAddRef (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbKeybdDrivAR(%x)\n"),(ADDRESS)pThis));
    return UsbKeybdDeviceAddRef(iUsbKeybdDevice(pUsbKeybdDriver(pThis)));
}

PRIVATE UINT UsbKeybdDriverRelease (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbKeybdDrivRel(%x)\n"),(ADDRESS)pThis));
    return UsbKeybdDeviceRelease(iUsbKeybdDevice(pUsbKeybdDriver(pThis)));
}

/* Utils
 */
PRIVATE void ParseHidReportDescriptor(
    UINT8 *Descriptor,
    UINT16 Length,
    UINT8 *pInputReportLength,
    UINT8 *pOutputReportLength)
{
    int i;
    int InSize, OutSize, FeaSize;
    int Rc, Rs;
    UINT8 c;

    FeaSize = InSize = OutSize = 0;
    Rc = Rs = 0;
    for (i = 0; i < Length; i++)
    {
        c = Descriptor[i];
        i += c & 3;
        if (c == 0x75)
            Rs = Descriptor[i];
        if (c == 0x95)
            Rc = Descriptor[i];
        if (c == 0x81)
            InSize += (Rs * Rc);
        if (c == 0x91)
            OutSize += (Rs * Rc);
        if (c == 0xb1)
            FeaSize += (Rs * Rc);
    }
    DBGME(1,_tprintf(_T("ReportLengths: InSize %d (%d bytes) OutSize %d (%d bytes) Feature %d (%d bytes)\n"),
                     InSize,InSize/8,
                     OutSize,OutSize/8,
                     FeaSize,FeaSize/8));

    *pInputReportLength = (UINT8) (InSize / 8);
    *pOutputReportLength = (UINT8) (OutSize / 8);
}

PRIVATE void ParseConfigurationDescriptor(
    PUSBKEYBD this,
    UINT8 *Descriptor,
    UINT16 Length)
{
    int i;
    UINT8 Dt, Ds, Ep = 0x81, Interval = 10, Rs = 0, Iface = 255;
    UINT16 Mps = 4;
    PUSB_ENDPOINT_struct Ed = NULL;
    PUSB_HID_struct Hd = NULL;
    PUSB_INTERFACE_struct If;

    /* Hunt for "the" input endpoint descriptor in the first interface
     */
    for (i = 0; i < Length; i += Ds)
    {
        /* Get and check the size
         */
        Ds = Descriptor[i];
        if (Ds == 0) /* sanity */
            break;

        /* Look at the descriptor type now
         */
        Dt = Descriptor[i+1];

        /* Are we looking at the right interface yet
         */
        if ((Iface == 255) && (Dt != USB_DT_IFACE))
            continue;
        if (Dt == USB_DT_IFACE) {
            If = (PUSB_INTERFACE_struct)(Descriptor + i);
            if ((If->InterfaceClass == USB_DC_HID) &&
                (If->InterfaceSubClass == USB_DS_BOOT) &&
                (If->InterfaceProtocol == USB_DP_KEYBOARD))
                Iface = If->InterfaceNumber;
            else
                continue;
        }

        /* Is this the HID descriptor? We need the report size
         */
        if ((Dt == USB_DT_HID) && (Hd == NULL)) {
            Hd = (PUSB_HID_struct)(Descriptor + i);
            Rs = Get2(Hd->ReportDescriptorLength);
            continue;
        }

        /* If not an endpoint skip
         */
        if (Dt != USB_DT_ENDP)
            continue;

        /* Is it the one we want (input)
         */
        Ed = (PUSB_ENDPOINT_struct)(Descriptor + i);
        if (Ed->EndpointAddress & USB_EA_IN) {
            Ep = (Ed->EndpointAddress & ~USB_EA_IN);
            Mps = Ed->MaxPacketSize[0] | (Ed->MaxPacketSize[1] << 8);
            Interval = Ed->Interval;
            /* We really want an interrupt endpoint, but..
             */
            if ((Ed->Attributes & USB_AE_MASK) == USB_AE_INTERRUPT)
                break;
        }
    }
    DBGME(1,_tprintf(_T("InEndpoint: Address %x Mps %d Rs %d Interval %d Iface %d\n"),
                     Ep, Mps, Rs, Interval, Iface));

    this->ReportDescriptorSize = Rs;
    this->InEndpoint = Ep;
    this->MaximumPacketSize = Mps;
    this->PollingInterval = Interval;
    this->Interface = Iface;
}

PRIVATE SCODE GetConfigurationDescriptor(
    PUSBKEYBD this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT16 Length)
{
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_DTH;
    Urb->SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb->SetupData[2] = 0;
    Urb->SetupData[3] = USB_DT_CONFIG;
    Urb->SetupData[4] = 0;
    Urb->SetupData[5] = 0;
    Set2(Urb->SetupData+6,Length);
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = (UINT8)Length;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbKeybd:GCD SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbKeybd:GCD SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}
    
PRIVATE SCODE SetConfiguration(
    PUSBKEYBD this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 Index)
{
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->SetupData[0] = 0;
    Urb->SetupData[1] = USB_R_SET_CONFIG;
    Urb->SetupData[2] = Index;
    Urb->SetupData[3] = 0;
    Urb->SetupData[4] = 0;
    Urb->SetupData[5] = 0;
    Urb->SetupData[6] = 0;
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = 0;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbKeybd:SC SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbKeybd:SC SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}
    
PRIVATE SCODE SetIdle(
    PUSBKEYBD this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 Period)
{
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->EndpointAddress = 0;
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_CLASS|USB_RT_IFACE;
    Urb->SetupData[1] = USB_R_SET_IDLE;
    Urb->SetupData[2] = 0;
    Urb->SetupData[3] = Period;
    Urb->SetupData[4] = this->Interface;
    Urb->SetupData[5] = 0;
    Urb->SetupData[6] = 0;
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = 0;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbKeybd:SI SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbKeybd:SI SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }
    return S_OK;
}
    
PRIVATE SCODE GetReportDescriptor(
    PUSBKEYBD this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT16 MaxSize
    )
{
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_DTH|USB_RT_IFACE;
    Urb->SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb->SetupData[2] = 0;
    Urb->SetupData[3] = USB_DT_REPORT;
    Urb->SetupData[4] = this->Interface;
    Urb->SetupData[5] = 0;
    Set2(&Urb->SetupData[6],MaxSize);
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = MaxSize;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbKeybd:GRD SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    /* Special check */
    if ((Urb->ActualDataLength == 0) && (Urb->Flags & URB_FLAG_ERROR)) {
        DBGME(3,_tprintf(_T("UsbKeybd:GRD SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}
    

PRIVATE SCODE UsbKeybdDriverIamYourHub (
    PIUSBDRIVER pThis,
    PIUSBHUB Hub,
    UINT PortNumber)
{
    PUSBKEYBD this = pUsbKeybdDriver(pThis);
    PUSBM_REQUEST_BLOCK Urb = NULL;
    SCODE sc;
    UINT16 Length;
    PINAMESPACE Ns = CurrentNameSpace();
    _TCHAR MyName[MYNAMELENGTH];

    DBGME(0,_tprintf(_T("UsbKeybd:IYH(%x,%x,%d)\n"),
                     (ADDRESS)this,(ADDRESS)Hub,PortNumber));

    /* Check not already owned
     */
    if (this->Hub) {
        DBGME(3,_tprintf(_T("UsbKeybd: already owned.\n")));
        return E_ALREADY_EXISTS;
    }

    /* QI to make sure and to take a ref
     */
    sc = Hub->v->QueryInterface(Hub,
                                &IID_IUsbHub,
                                (void **)&this->Hub);
    if (FAILED(sc)) {
        DBGME(1,_tprintf(_T("UsbKeybd bad Hub, sc=%x\n"),sc));
        return sc;
    }

    this->PortNo = (UINT8) PortNumber;
    //done above: this->Hub = Hub;

    /* Get an Urb and an extra buffer
     */
    Urb = malloc((64*3) + sizeof *Urb);
    if (Urb == NULL) {
        sc = E_NOT_ENOUGH_MEMORY;
        goto Error;
    }
    memset(Urb,0,(64*3) + sizeof *Urb);
    Urb->Done = &this->CDevice.IntrPoint;
    Urb->Data = (UINT8*)(Urb+1);
    this->LastReport = Urb->Data + 64;
    this->LastRawReport = Urb->Data + (64*2);

    this->AvailUrb = Urb;

    /* Get and set the config
     * Get it twice, first time just to find length.
     */
    sc = GetConfigurationDescriptor(this,Urb,9);
    if (FAILED(sc))
        goto Error;

    Length = Get2(Urb->Data+2);
    sc = GetConfigurationDescriptor(this,Urb,Length);
    if (FAILED(sc))
        goto Error;

    /* Find out about the endpoint(s)
     */
    ParseConfigurationDescriptor(this,Urb->Data,Urb->ActualDataLength);

    sc = SetConfiguration(this,Urb,1);
    if (FAILED(sc))
        goto Error;

    /* Find out what the report looks like
     */
    if (this->ReportDescriptorSize != 0) {
        sc = GetReportDescriptor(this,Urb,this->ReportDescriptorSize);
        if (FAILED(sc))
            goto Error;

        ParseHidReportDescriptor(Urb->Data,Urb->ActualDataLength,
                                 &this->InReportSize,
                                 &this->OutReportSize);
    } else {
        /* Oh boy.
         */
        this->OutReportSize = 0;
        this->InReportSize = (UINT8) this->MaximumPacketSize;
    }

    /* Disable autorepeat.
     */
    sc = SetIdle(this,Urb,0);
    if (FAILED(sc)) {
        DBGME(1,_tprintf(_T("UsbKeybd cantidle, sc=%x\n"),sc));
        //goto Error;
    }

    /* Add top-level namespace entry, for user convenience
     * HUB only puts us in the USB tree.
     * Try again to avoid clashes if multiple keyboards are attached
     * Put a limit to it, though, just in case.
     */
    for (;;) {
        MakeMyName(this,MyName);
        sc = Ns->v->Register(Ns,
                             MyName,
                             (PIUNKNOWN) this,
                             NAME_SPACE_READ | 
                             NAME_SPACE_WRITE |
                             NAME_SPACE_CREATE |
                             NAME_SPACE_FAILIFEXIST,
                             NULL);
        if ((sc == E_ALREADY_EXISTS) && (this->KeyboardNo < 1000))
            this->KeyboardNo++;
        else
            break;
    }

    return S_OK;
 Error:
    if (this->Hub) {
        this->Hub->v->Release(this->Hub);
        this->Hub = NULL;
    }
    if (this->AvailUrb) {
        free(this->AvailUrb);
        this->AvailUrb = NULL;
    }
    DBGME(1,_tprintf(_T("UsbKeybd:IYH fails, sc=%x\n"),sc));
    return sc;
}

PRIVATE SCODE UsbKeybdDriverYouAreTerminated (
    PIUSBDRIVER pThis,
    PIUSBHUB SaysWho)
{
    PUSBKEYBD this = pUsbKeybdDriver(pThis);
    SCODE sc;
    PINAMESPACE Ns = CurrentNameSpace();
    _TCHAR MyName[MYNAMELENGTH];

    DBGME(0,_tprintf(_T("UsbKeybd:YAT(%x)\n"),(ADDRESS)SaysWho));

    /* Check args, and that we had started ok
     */
    if (this->Hub == NULL) {
        return E_INVALID_PARAMETER;
    }

    /* Remove from namespace
     */
    MakeMyName(this,MyName);
    sc = Ns->v->Unregister(Ns,MyName);

    /* Cancel anything pending
     */
    if (this->AvailUrb)
        sc = this->Hub->v->CancelUrb(this->Hub,this->AvailUrb,this->PortNo);

    /* LEt HUB go
     */
    this->Hub->v->Release(this->Hub);
    this->Hub = NULL;

    /* BUGBUG make sure refcount works out.
     */
    if (this->CDevice.RefCnt > 1)
        _tprintf(_T("UsbKeybd::KillYourself too many refs %d\n"),
               this->CDevice.RefCnt);

    return S_OK;
}

PRIVATE SCODE UsbKeybdDriverGetUsbAddress(
    PIUSBDRIVER pThis,
    PUINT UsbAddress,
    PUINT PortNumber)
{
    PUSBKEYBD this = pUsbKeybdDriver(pThis);

    if (UsbAddress == NULL)
        return E_INVALID_PARAMETER;
    if (this->MyUsbAddress == 0)
        return E_OBJECT_NOT_INITIALIZED;

    *UsbAddress = this->MyUsbAddress;
    *PortNumber = this->PortNo;

    return S_OK;
}

#include "s_UsbKeybd_v.c"

/* Constructor */
PIDEVICE UsbKeybdNew(
    const _TCHAR *MyName,
    _TCHAR *HciName,
    UINT8 UsbAddress,
    UINT8 MaximumControlPacketSize,
    PIUNKNOWN /*IUSBMASTERCONTROLLER*/ UsbTreeLocation)
{
    PIHEAP pHeap = CurrentHeap();
    PUSBKEYBD this;

    if (UsbAddress == 0) {
        DBGME(3,_tprintf(_T("UsbKeybdNew: bad UsbAddress 0\n")));
        return NULL;
    }

    this = (PUSBKEYBD) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *this, 0);
    if (!this)
        return NULL;

    this->CDevice.v = (struct CDeviceVtbl*)&UsbKeybdDeviceUsbKeybdVtbl;
    this->CDevice.RefCnt = 1;
    this->UsbKeybdDriverv = &UsbKeybdDriverUsbKeybdVtbl;

    /* Do this because it aint config that starts us
     */
    Condition_Init(&this->CDevice.IntrPoint);

    this->MyUsbAddress = UsbAddress;
    this->MaximumControlPacketSize = MaximumControlPacketSize;
    this->NextKeyIndex = 0xff;/*bigger than any conceivable report size (8bits) */
    this->RegularTable = UsbRegularTable;
    this->ShiftTable = UsbShiftTable;
    this->ControlTable = UsbControlTable;
    this->MetaTable = UsbMetaTable;

    return iUsbKeybdDevice(this);
}
/* ** 
*/
SCODE __cdecl UsbKeybdConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * MyName,
        ADDRESS UnusedRegisterBase,
        UINT UnusedInterrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        _TCHAR *HciName,
        PIUNKNOWN /*PIUSBMASTERCONTROLLER*/ UsbTreeLocation,
        ADDRESS UsbAddress,
        ADDRESS MaximumControlPacketSize)
{
    PIUNKNOWN pUnk;

    UnusedParameter(UnusedRegisterBase);
    UnusedParameter(UnusedInterrupt);

    DBGME(0,_tprintf(_T("UsbKeybdConstructor(%x %s %x %s %d %x)\n"),
                     (ADDRESS) pThis, MyName, (ADDRESS) ppUnk,
                     (HciName) ? HciName : _T("NULL"),
                     UsbAddress, (ADDRESS) UsbTreeLocation));

    pUnk = (PIUNKNOWN) UsbKeybdNew(MyName,
                                   HciName,
                                   (UINT8) UsbAddress,
                                   (UINT8) MaximumControlPacketSize,
                                   UsbTreeLocation);

    DBGME(0,_tprintf(_T("UsbKeybdConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UsbKeybdConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UsbKeybdDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UsbKeybdConstructor
};

static struct _DRIVERCOB UsbKeybdDriverCobObject = { &UsbKeybdDriverCobVtable, 1 };

PIUNKNOWN UsbKeybdCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UsbKeybdDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UsbKeybdDeviceUsbKeybdVtbl, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &UsbKeybdDriverCobObject;
}
