/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbkvm.c
 *
 * Purpose:
 *
 *    Device driver for the IoGear (ATEN) KVM USB switch.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <string.h>
#include <stddef.h> /* for offsetof() */
//#define _DEBUG 1
//#define DBGLVL 3
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbmaster.h>

typedef struct _USBKVM {
    struct CDevice CDevice;
    CONST_VTBL struct IUsbDriverVtbl *UsbKvmDriverv;
    PIUSBHUB Hub;
    PIUSBDRIVER Mouse;
    PIUSBDRIVER Keyboard;
    UINT8 MyUsbAddress;
    UINT8 PortNo;
    UINT8 MaximumControlPacketSize;
} USBKVM, *PUSBKVM;

/* Convert between interface and instance pointer and vice versa */
#define pUsbKvmDevice(_i_) ((PUSBKVM)((_i_)))
#define iUsbKvmDevice(_p_) ((PIDEVICE)((_p_)))
#define pUsbKvmDriver(_i_) ((PUSBKVM)((ADDRESS)(_i_) - offsetof(struct _USBKVM, UsbKvmDriverv)))
#define iUsbKvmDriver(_p_) ((PIUSBDRIVER)(&((PUSBKVM)(_p_))->UsbKvmDriverv))

/* IDevice interface
 */
PRIVATE SCODE UsbKvmDeviceQueryInterface (
    PIDEVICE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbKvmDevQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));

    /* IDevice
     */
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_IFile,iid)
        || UuidCmp(&IID_IDevice,iid))
    {
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }

    /* IUsbDriver
     */
    if (UuidCmp(&IID_IUsbDriver, iid)) {
        /* Use a single refcount
         */
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (void*) iUsbKvmDriver(pThis);
        return S_OK;
    }

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT UsbKvmDeviceAddRef (
    PIDEVICE pThis)
{
    DBGME(0,_tprintf(_T("UsbKvmDevAR(%x)\n"),(ADDRESS)pThis));
    return AtomicInc(&pUsbKvmDevice(pThis)->CDevice.RefCnt);
}

PRIVATE UINT UsbKvmDeviceRelease (
    PIDEVICE pThis)
{
    PUSBKVM this = pUsbKvmDevice(pThis);
    UINT Refs = AtomicDec(&this->CDevice.RefCnt);
    DBGME(0,_tprintf(_T("UsbKvmDevRel(%x,%d)\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        pHeap->v->Free(pHeap, 0, (void *) this);
    }
    return Refs;
}

/* IUsbDriver interface
 */
PRIVATE SCODE UsbKvmDriverQueryInterface (
    PIUSBDRIVER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbKvmDrivQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));
    return UsbKvmDeviceQueryInterface(iUsbKvmDevice(pUsbKvmDriver(pThis)), iid, ppObject);
}

PRIVATE UINT UsbKvmDriverAddRef (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbKvmDrivAR(%x)\n"),(ADDRESS)pThis));
    return UsbKvmDeviceAddRef(iUsbKvmDevice(pUsbKvmDriver(pThis)));
}

PRIVATE UINT UsbKvmDriverRelease (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbKvmDrivRel(%x)\n"),(ADDRESS)pThis));
    return UsbKvmDeviceRelease(iUsbKvmDevice(pUsbKvmDriver(pThis)));
}

/* Utilities
 */
PRIVATE SCODE StartDriver (
    PUSBKVM this,
    _TCHAR *Suffix,
    PIUSBDRIVER *ppDriver)
{
    SCODE sc;
    PIDEVICECONSTRUCTOR pDevCons;
    PIUNKNOWN pUnk;
    PIUSBDRIVER Driver;
    _TCHAR DriverCobName[12+8];

    /* Default nothing
     */
    *ppDriver = NULL;

    /* Build the driver's COB name
     */
    _sntprintf(DriverCobName,sizeof DriverCobName / sizeof(_TCHAR),
               _T("COB\\usb%s.cob"),
               Suffix);

    /* Load the driver & find the constructor
     */
    DBGME(2,_tprintf(_T("Kvm(%x) Loading driver %s\n"),(ADDRESS)this,DriverCobName));
    sc = BindToObject(CurrentNameSpace(),
                      DriverCobName,
                      NAME_SPACE_READ | NAME_SPACE_WRITE,
                      &IID_IDeviceConstructor,
                      (void **) &pDevCons);

    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("Kvm(%x) COB %s did not load, sc=%x\n"),(ADDRESS)this,
                         DriverCobName, sc));
        return sc;
    }

    /* Invoke it with appropriate arguments
     */
    sc = pDevCons->v->Constructor(pDevCons,
                                  Suffix,
                                  0,0,
                                  &pUnk,
                                  NULL,
                                  (PTR) this->Hub,
                                  (ADDRESS) this->MyUsbAddress,
                                  (ADDRESS) this->MaximumControlPacketSize,
                                  0,0,0);
    (void) pDevCons->v->Release(pDevCons);

    if (FAILED(sc)) {    /* constructor failed */
        DBGME(3,_tprintf(_T("Kvm(%x): %s constructor failed sc=%x\n"),(ADDRESS)this,
                         DriverCobName, sc));
        return sc;
    }

    DBGME(0,_tprintf(_T("Kvm(%x): %s constructor got %x. sc=%x\n"),(ADDRESS)this,
                     DriverCobName, (ADDRESS)pUnk, sc));

    /* Get the correct interface
     */
    Driver = NULL;
    sc = pUnk->v->QueryInterface(pUnk,
                                 &IID_IUsbDriver,
                                 (void **)&Driver);
    (void) pUnk->v->Release(pUnk);
    if (FAILED(sc)) {    /* wrong driver */
        DBGME(3,_tprintf(_T("Kvm(%x): %s not an IUsbDriver?!? sc=%x\n"),(ADDRESS)this,
                         DriverCobName, sc));
        return sc;
    }

    /* Forward the IamYourHub call
     */
    sc = Driver->v->IamYourHub(Driver,this->Hub,this->PortNo);
    if (FAILED(sc)) {    /* bad driver */
        DBGME(3,_tprintf(_T("Kvm(%x): %s fails IamYH?!? sc=%x\n"),(ADDRESS)this,
                         DriverCobName, sc));
        Driver->v->Release(Driver);
        return sc;
    }

    *ppDriver = Driver;
    return sc;
}


PRIVATE SCODE UsbKvmDriverIamYourHub (
    PIUSBDRIVER pThis,
    PIUSBHUB Hub,
    UINT PortNumber)
{
    PUSBKVM this = pUsbKvmDriver(pThis);
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbKvm:IYH(%x,%x,%d)\n"),
                     (ADDRESS)this,(ADDRESS)Hub,PortNumber));

    /* Check not already owned
     */
    if (this->Hub) {
        DBGME(3,_tprintf(_T("UsbKvm: already owned.\n")));
        return E_ALREADY_EXISTS;
    }

    /* QI to make sure and to take a ref
     */
    sc = Hub->v->QueryInterface(Hub,
                                &IID_IUsbHub,
                                (void **)&this->Hub);
    if (FAILED(sc)) {
        DBGME(1,_tprintf(_T("UsbKvm bad Hub, sc=%x\n"),sc));
        return sc;
    }

    this->PortNo = (UINT8) PortNumber;
    //done above: this->Hub = Hub;

    /* Start the mouse and keyboard drivers, ignore failures
     */
    sc = StartDriver(this,_T("keybd"),&this->Keyboard);
    if (FAILED(sc)) {
        DBGME(2,_tprintf(_T("UsbKvm missing keybd driver, sc=%x\n"),sc));
    }

    sc = StartDriver(this,_T("mouse"),&this->Mouse);
    if (FAILED(sc)) {
        DBGME(2,_tprintf(_T("UsbKvm missing mouse driver, sc=%x\n"),sc));
    }

    return S_OK;
}

PRIVATE void NotifyOfTermination(PIUSBDRIVER *ppDriver, PIUSBHUB SaysWho)
{
    PIUSBDRIVER Driver = *ppDriver;
    if (Driver) {
        Driver->v->Release(Driver);
        Driver->v->YouAreTerminated(Driver,SaysWho);
    }
    *ppDriver = NULL;
}

PRIVATE SCODE UsbKvmDriverYouAreTerminated (
    PIUSBDRIVER pThis,
    PIUSBHUB SaysWho)
{
    PUSBKVM this = pUsbKvmDriver(pThis);

    DBGME(0,_tprintf(_T("UsbKvm:YAT(%x)\n"),(ADDRESS)SaysWho));

    /* Check args, and that we had started ok
     */
    if (this->Hub == NULL) {
        return E_INVALID_PARAMETER;
    }

    /* Forward to the drivers
     */
    NotifyOfTermination(&this->Mouse,SaysWho);
    NotifyOfTermination(&this->Keyboard,SaysWho);

    /* Let the HUB go
     */
    this->Hub->v->Release(this->Hub);
    this->Hub = NULL;

    return S_OK;
}

PRIVATE SCODE UsbKvmDriverGetUsbAddress (
    PIUSBDRIVER pThis,
    PUINT UsbAddress,
    PUINT PortNumber
                      )
{
    PUSBKVM this = pUsbKvmDriver(pThis);

    if (UsbAddress == NULL)
        return E_INVALID_PARAMETER;
    if (this->MyUsbAddress == 0)
        return E_OBJECT_NOT_INITIALIZED;

    *UsbAddress = this->MyUsbAddress;
    *PortNumber = this->PortNo;
    return S_OK;
}

#include "s_UsbKvm_v.c"

/* Constructor */
PIDEVICE UsbKvmNew(
    const _TCHAR *MyName,
    _TCHAR *HciName,
    UINT8 UsbAddress,
    UINT8 MaximumControlPacketSize,
    PIUNKNOWN /*IUSBMASTERCONTROLLER*/ UsbTreeLocation)
{
    PIHEAP pHeap = CurrentHeap();
    PUSBKVM this;

    if (UsbAddress == 0) {
        DBGME(3,_tprintf(_T("UsbKvmNew: bad UsbAddress 0\n")));
        return NULL;
    }

    this = (PUSBKVM) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *this, 0);
    if (!this)
        return NULL;

    this->CDevice.v = (struct CDeviceVtbl*)&UsbKvmDeviceUsbKvmVtbl;
    this->CDevice.RefCnt = 1;
    this->UsbKvmDriverv = &UsbKvmDriverUsbKvmVtbl;

    /* Do this because it aint config that starts us
     */
    Condition_Init(&this->CDevice.IntrPoint);

    this->MyUsbAddress = UsbAddress;
    this->MaximumControlPacketSize = MaximumControlPacketSize;

    return iUsbKvmDevice(this);
}

/* ** 
*/
SCODE __cdecl UsbKvmConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * MyName,
        ADDRESS UnusedRegisterBase,
        UINT UnusedInterrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        _TCHAR *HciName,
        PIUNKNOWN /*PIUSBMASTERCONTROLLER*/ UsbTreeLocation,
        ADDRESS UsbAddress,
        ADDRESS MaximumControlPacketSize)
{
    PIUNKNOWN pUnk;

    UnusedParameter(UnusedRegisterBase);
    UnusedParameter(UnusedInterrupt);

    DBGME(0,_tprintf(_T("UsbKvmConstructor(%x %s %x %s %d %x)\n"),
                     (ADDRESS) pThis, MyName, (ADDRESS) ppUnk,
                     (HciName) ? HciName : _T("KVM"),
                     UsbAddress, (ADDRESS) UsbTreeLocation));

    pUnk = (PIUNKNOWN) UsbKvmNew(MyName,
                                   HciName,
                                   (UINT8) UsbAddress,
                                   (UINT8) MaximumControlPacketSize,
                                   UsbTreeLocation);

    DBGME(0,_tprintf(_T("UsbKvmConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UsbKvmConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UsbKvmDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UsbKvmConstructor
};

static struct _DRIVERCOB UsbKvmDriverCobObject = { &UsbKvmDriverCobVtable, 1 };

PIUNKNOWN UsbKvmCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UsbKvmDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UsbKvmDeviceUsbKvmVtbl, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &UsbKvmDriverCobObject;
}
