/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbmouse.c
 *
 * Purpose:
 *
 *    Device driver for USB mice.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <string.h>
#include <stddef.h> /* for offsetof() */
//#define _DEBUG 1
//#define DBGLVL 0
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbmaster.h>

typedef struct _USBMOUSE {
    struct CDevice CDevice;
    CONST_VTBL struct IUsbDriverVtbl *UsbMouseDriverv;
    UINT8 MyUsbAddress;
    PIUSBHUB Hub;
    UINT8 PortNo;
    UINT8 InEndpoint;
    UINT16 MaximumPacketSize;
    UINT8 MaximumControlPacketSize;
    UINT8 PollingInterval;
    UINT16 MiceNo;
#define MYNAMELENGTH 12 /* bignuf */
#define MakeMyName(_this_,_name_)  _stprintf(_name_,_T("mouse%d"),(_this_)->MiceNo)
    PUSBM_REQUEST_BLOCK AvailUrb;
    UINT16 ReportSize;
    UINT8  Interface;
} USBMOUSE, *PUSBMOUSE;

/* Convert between interface and instance pointer and vice versa */
#define pUsbMouseDevice(_i_) ((PUSBMOUSE)((_i_)))
#define iUsbMouseDevice(_p_) ((PIDEVICE)((_p_)))
#define pUsbMouseDriver(_i_) ((PUSBMOUSE)((ADDRESS)(_i_) - offsetof(struct _USBMOUSE, UsbMouseDriverv)))
#define iUsbMouseDriver(_p_) ((PIUSBDRIVER)(&((PUSBMOUSE)(_p_))->UsbMouseDriverv))

PRIVATE SCODE UsbMouseDeviceQueryInterface (
    PIDEVICE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbMouseDevQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));

    /* IDevice
     */
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_IFile,iid)
        || UuidCmp(&IID_IDevice,iid))
    {
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }

    /* IUsbDriver
     */
    if (UuidCmp(&IID_IUsbDriver, iid)) {
        /* Use a single refcount
         */
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (void*) iUsbMouseDriver(pThis);
        return S_OK;
    }

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT UsbMouseDeviceAddRef (
    PIDEVICE pThis)
{
    DBGME(0,_tprintf(_T("UsbMouseDevAR(%x)\n"),(ADDRESS)pThis));
    return AtomicInc(&pUsbMouseDevice(pThis)->CDevice.RefCnt);
}

PRIVATE UINT UsbMouseDeviceRelease (
    PIDEVICE pThis)
{
    PUSBMOUSE this = pUsbMouseDevice(pThis);
    UINT Refs = AtomicDec(&this->CDevice.RefCnt);
    DBGME(0,_tprintf(_T("UsbMouseDevRel(%x,%d)\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        if (this->AvailUrb)
            pHeap->v->Free(pHeap, 0, (void *) this->AvailUrb);

        pHeap->v->Free(pHeap, 0, (void *) this);
    }
    return Refs;
}

/* Read from the IN endpoint one reportfull
 */
PRIVATE SCODE GetReport(
    PUSBMOUSE this,
    PUSBM_REQUEST_BLOCK Urb)
{
    SCODE sc;

    if ((Urb == NULL || (this->Hub == NULL))) {
        /* Must have been terminated
         */
        DBGME(3,_tprintf(_T("UsbMouse:GetReport no Urb\n")));
        return E_OBJECT_NOT_FOUND;
    }
    Urb->Done = &this->CDevice.IntrPoint;
    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->EndpointAddress = this->InEndpoint;
    Urb->Pid = USB_PID_IN;
    Urb->Flags = URB_FLAG_IN|URB_FLAG_INTERRUPT;
    Urb->SetupLength = 0;
    Urb->Interval = this->PollingInterval;
    Urb->ExpectedDataLength = (UINT8) this->ReportSize;
    Urb->MaximumPacketSize = this->MaximumPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbMouse:GetReport sc=%x\n"),sc));
        return sc;
    }

    Condition_Wait(Urb->Done,NULL,NULL);

    /* Check results: migth have been terminated, or no data available
     */
    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbMouse:GetReport er=%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}

PRIVATE SCODE UsbMouseDeviceReadAt (
    PIDEVICE pThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PUSBMOUSE this = pUsbMouseDevice(pThis);
    PUSBM_REQUEST_BLOCK Urb = this->AvailUrb;
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbMouse:RAt(%x,%x,%d)\n"),
                     (ADDRESS)this,(ADDRESS)pBuffer,Size));

    *pSizeRead = 0;
    if (Size < this->ReportSize)
        return E_BUFFER_TOO_SMALL;

    sc = GetReport(this,Urb);
    if (FAILED(sc))
        return sc;

    *pSizeRead = this->ReportSize;
    memcpy(pBuffer,Urb->Data,*pSizeRead);
    DBGME(0,_tprintf(_T("UsbMouse:RAt -> %d\n"),*pSizeRead));

    return S_OK;
}

PRIVATE SCODE UsbMouseDeviceGetSize (
    PIDEVICE pThis,
    /*out*/ UINT64 * Size)
{
    PUSBMOUSE this = pUsbMouseDevice(pThis);

    Uint32ToUint64(*Size,this->ReportSize);
    return S_OK;
}

PRIVATE SCODE UsbMouseDriverQueryInterface (
    PIUSBDRIVER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbMouseDrivQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));
    return UsbMouseDeviceQueryInterface(iUsbMouseDevice(pUsbMouseDriver(pThis)), iid, ppObject);
}

PRIVATE UINT UsbMouseDriverAddRef (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbMouseDrivAR(%x)\n"),(ADDRESS)pThis));
    return UsbMouseDeviceAddRef(iUsbMouseDevice(pUsbMouseDriver(pThis)));
}

PRIVATE UINT UsbMouseDriverRelease (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbMouseDrivRel(%x)\n"),(ADDRESS)pThis));
    return UsbMouseDeviceRelease(iUsbMouseDevice(pUsbMouseDriver(pThis)));
}

/* Utils
 */
PRIVATE void ParseHidReportDescriptor(
    UINT8 *Descriptor,
    UINT16 Length,
    UINT16 *pInputReportLength)
{
    int i;
    int InSize, OutSize, FeaSize;
    int Rc, Rs;
    UINT8 c;

    FeaSize = InSize = OutSize = 0;
    Rc = Rs = 0;
    for (i = 0; i < Length; i++)
    {
        c = Descriptor[i];
        i += c & 3;
        if (c == 0x75)
            Rs = Descriptor[i];
        if (c == 0x95)
            Rc = Descriptor[i];
        if (c == 0x81)
            InSize += (Rs * Rc);
        if (c == 0x91)
            OutSize += (Rs * Rc);
        if (c == 0xb1)
            FeaSize += (Rs * Rc);
    }
    DBGME(1,_tprintf(_T("ReportLengths: InSize %d (%d bytes) OutSize %d (%d bytes) Feature %d (%d bytes)\n"),
                     InSize,InSize/8,
                     OutSize,OutSize/8,
                     FeaSize,FeaSize/8));

    *pInputReportLength = InSize / 8;
}

PRIVATE void ParseConfigurationDescriptor(
    PUSBMOUSE this,
    UINT8 *Descriptor,
    UINT16 Length)
{
    int i;
    UINT8 Dt, Ds, Ep = 0x81, Interval = 10, Iface = 255;
    UINT16 Mps = 4;
    PUSB_ENDPOINT_struct Ed;
    PUSB_INTERFACE_struct If;

    /* Hunt for "the" input endpoint descriptor
     */
    for (i = 0; i < Length; i += Ds)
    {
        /* Get and check the size
         */
        Ds = Descriptor[i];
        if (Ds == 0) /* sanity */
            break;

        /* Which descriptor is it
         */
        Dt = Descriptor[i+1];

        /* Are we looking at the right interface yet
         */
        if ((Iface == 255) && (Dt != USB_DT_IFACE))
            continue;
        if (Dt == USB_DT_IFACE) {
            If = (PUSB_INTERFACE_struct)(Descriptor + i);
            if ((If->InterfaceClass == USB_DC_HID) &&
                (If->InterfaceSubClass == USB_DS_BOOT) &&
                (If->InterfaceProtocol == USB_DP_MOUSE))
                Iface = If->InterfaceNumber;
            else
                continue;
        }

        /* If not an endpoint skip
         */
        if (Dt != USB_DT_ENDP)
            continue;

        /* Is it the one we want (input)
         */
        Ed = (PUSB_ENDPOINT_struct)(Descriptor + i);
        if (Ed->EndpointAddress & USB_EA_IN) {
            Ep = Ed->EndpointAddress;
            Mps = Ed->MaxPacketSize[0] | (Ed->MaxPacketSize[1] << 8);
            Interval = Ed->Interval;
            /* We really want an interrupt endpoint, but..
             */
            if ((Ed->Attributes & USB_AE_MASK) == USB_AE_INTERRUPT)
                break;
        }
    }
    DBGME(1,_tprintf(_T("InEndpoint: Address %x Mps %d Interval %d Iface %d\n"),
                     Ep, Mps, Interval, Iface));

    this->InEndpoint = Ep;
    this->MaximumPacketSize = Mps;
    this->PollingInterval = Interval;
    this->Interface = Iface;
}

PRIVATE SCODE GetConfigurationDescriptor(
    PUSBMOUSE this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT16 Length)
{
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_DTH;
    Urb->SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb->SetupData[2] = 0;
    Urb->SetupData[3] = USB_DT_CONFIG;
    Urb->SetupData[4] = 0;
    Urb->SetupData[5] = 0;
    Set2(Urb->SetupData+6,Length);
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = (UINT8)Length;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbMouse:GCD SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbMouse:GCD SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}
    
PRIVATE SCODE SetConfiguration(
    PUSBMOUSE this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 Index)
{
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->SetupData[0] = 0;
    Urb->SetupData[1] = USB_R_SET_CONFIG;
    Urb->SetupData[2] = Index;
    Urb->SetupData[3] = 0;
    Urb->SetupData[4] = 0;
    Urb->SetupData[5] = 0;
    Urb->SetupData[6] = 0;
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = 0;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbMouse:SC SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbMouse:SC SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}
    
PRIVATE SCODE SetIdle(
    PUSBMOUSE this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 Period)
{
#if notyet
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->EndpointAddress = 0x81; /* BUGBUG from parsing reports */
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_CLASS|USB_RT_IFACE;
    Urb->SetupData[1] = USB_R_SET_IDLE;
    Urb->SetupData[2] = 0;
    Urb->SetupData[3] = Period;
    Urb->SetupData[4] = 0;
    Urb->SetupData[5] = 0;
    Urb->SetupData[6] = 0;
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = 0;
    Urb->MaximumPacketSize = this->MaximumPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbMouse:SI SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbMouse:SI SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }
#endif
    return S_OK;
}
    
PRIVATE SCODE GetReportDescriptor(
    PUSBMOUSE this,
    PUSBM_REQUEST_BLOCK Urb)
{
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_DTH|USB_RT_IFACE;
    Urb->SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb->SetupData[2] = 0;
    Urb->SetupData[3] = USB_DT_REPORT;
    Urb->SetupData[4] = this->Interface;
    Urb->SetupData[5] = 0;
    Urb->SetupData[6] = 255;
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = 255;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbMouse:GRD SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    /* Special check */
    if ((Urb->ActualDataLength == 0) && (Urb->Flags & URB_FLAG_ERROR)) {
        DBGME(3,_tprintf(_T("UsbMouse:GRD SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}
    

PRIVATE SCODE UsbMouseDriverIamYourHub (
    PIUSBDRIVER pThis,
    PIUSBHUB Hub,
    UINT PortNumber)
{
    PUSBMOUSE this = pUsbMouseDriver(pThis);
    PUSBM_REQUEST_BLOCK Urb = NULL;
    SCODE sc;
    UINT16 Length;
    PINAMESPACE Ns = CurrentNameSpace();
    _TCHAR MyName[MYNAMELENGTH];

    DBGME(0,_tprintf(_T("UsbMouse:IYH(%x,%x,%d)\n"),
                     (ADDRESS)this,(ADDRESS)Hub,PortNumber));

    /* Check not already owned
     */
    if (this->Hub) {
        DBGME(3,_tprintf(_T("UsbMouse: already owned.\n")));
        return E_ALREADY_EXISTS;
    }

    /* QI to make sure and to take a ref
     */
    sc = Hub->v->QueryInterface(Hub,
                                &IID_IUsbHub,
                                (void **)&this->Hub);
    if (FAILED(sc)) {
        DBGME(1,_tprintf(_T("UsbMouse bad Hub, sc=%x\n"),sc));
        return sc;
    }

    this->PortNo = (UINT8) PortNumber;
    //done above: this->Hub = Hub;

    /* Get an Urb
     */
    Urb = malloc(256 + sizeof *Urb);
    if (Urb == NULL) {
        sc = E_NOT_ENOUGH_MEMORY;
        goto Error;
    }
    memset(Urb,0,sizeof *Urb);
    Urb->Done = &this->CDevice.IntrPoint;
    Urb->Data = (UINT8*)(Urb+1);

    this->AvailUrb = Urb;

    /* Get and set the config
     * Get it twice, first time just to find length.
     */
    sc = GetConfigurationDescriptor(this,Urb,9);
    if (FAILED(sc))
        goto Error;

    Length = Get2(Urb->Data+2);
    sc = GetConfigurationDescriptor(this,Urb,Length);
    if (FAILED(sc))
        goto Error;

    /* Find out about the endpoint(s)
     */
    ParseConfigurationDescriptor(this,Urb->Data,Urb->ActualDataLength);

    sc = SetConfiguration(this,Urb,1);
    if (FAILED(sc))
        goto Error;

    /* Find out what the report looks like
     */
    sc = GetReportDescriptor(this,Urb);
    if (FAILED(sc))
        goto Error;

    ParseHidReportDescriptor(Urb->Data,Urb->ActualDataLength,
                             &this->ReportSize);

    /* Set it idling
     */
    sc = SetIdle(this,Urb,0);
    if (FAILED(sc))
        goto Error;

    /* Add top-level namespace entry, for user convenience
     * HUB only puts us in the USB tree.
     * Try again to avoid clashes if multiple mice are attached
     * Put a limit to it, though, just in case.
     */
    for (;;) {
        MakeMyName(this,MyName);
        sc = Ns->v->Register(Ns,
                             MyName,
                             (PIUNKNOWN) this,
                             NAME_SPACE_READ | 
                             NAME_SPACE_WRITE |
                             NAME_SPACE_CREATE |
                             NAME_SPACE_FAILIFEXIST,
                             NULL);
        if ((sc == E_ALREADY_EXISTS) && (this->MiceNo < 1000))
            this->MiceNo++;
        else
            break;
    }

#if notyet
    /* Start generic mouse on top
     */
    StartDriver(_T("mouse"),_T("COB\mouse.cob"),this->MyName,MouseProtocol);
#endif

    return S_OK;
 Error:
    if (this->Hub) {
        this->Hub->v->Release(this->Hub);
        this->Hub = NULL;
    }
    if (this->AvailUrb) {
        free(this->AvailUrb);
        this->AvailUrb = NULL;
    }
    return sc;
}

PRIVATE SCODE UsbMouseDriverYouAreTerminated (
    PIUSBDRIVER pThis,
    PIUSBHUB SaysWho)
{
    PUSBMOUSE this = pUsbMouseDriver(pThis);
    SCODE sc;
    PINAMESPACE Ns = CurrentNameSpace();
    _TCHAR MyName[MYNAMELENGTH];

    DBGME(0,_tprintf(_T("UsbMouse:YAT(%x)\n"),(ADDRESS)SaysWho));

    /* Check args, and that we had started ok
     */
    if (this->Hub == NULL) {
        return E_INVALID_PARAMETER;
    }

    /* Remove from namespace
     */
    MakeMyName(this,MyName);
    sc = Ns->v->Unregister(Ns,MyName);

#if notyet
    /* Stop generic mouse on top
     */
    ...how?? IQ(IDriver++,&ms) ms->KillYourself();
#endif

    /* Cancel anything pending
     */
    if (this->AvailUrb)
        sc = this->Hub->v->CancelUrb(this->Hub,this->AvailUrb,this->PortNo);

    /* LEt HUB go
     */
    this->Hub->v->Release(this->Hub);
    this->Hub = NULL;

    /* BUGBUG make sure refcount works out.
     */
    if (this->CDevice.RefCnt > 1)
        _tprintf(_T("UsbMouse::KillYourself too many refs %d\n"),
               this->CDevice.RefCnt);

    return S_OK;
}

PRIVATE SCODE UsbMouseDriverGetUsbAddress (
    PIUSBDRIVER pThis,
    PUINT UsbAddress,
    PUINT PortNumber)
{
    PUSBMOUSE this = pUsbMouseDriver(pThis);

    if (UsbAddress == NULL)
        return E_INVALID_PARAMETER;
    if (this->MyUsbAddress == 0)
        return E_OBJECT_NOT_INITIALIZED;

    *UsbAddress = this->MyUsbAddress;
    *PortNumber = this->PortNo;

    return S_OK;
}

#include "s_UsbMouse_v.c"

/* Constructor */
PIDEVICE UsbMouseNew(
    const _TCHAR *MyName,
    _TCHAR *HciName,
    UINT8 UsbAddress,
    UINT8 MaximumControlPacketSize,
    PIUNKNOWN /*IUSBMASTERCONTROLLER*/ UsbTreeLocation)
{
    PIHEAP pHeap = CurrentHeap();
    PUSBMOUSE this;

    if (UsbAddress == 0) {
        DBGME(3,_tprintf(_T("UsbMouseNew: bad UsbAddress 0\n")));
        return NULL;
    }

    this = (PUSBMOUSE) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *this, 0);
    if (!this)
        return NULL;

    this->CDevice.v = (struct CDeviceVtbl*)&UsbMouseDeviceUsbMouseVtbl;
    this->CDevice.RefCnt = 1;
    this->UsbMouseDriverv = &UsbMouseDriverUsbMouseVtbl;

    /* Do this because it aint config that starts us
     */
    Condition_Init(&this->CDevice.IntrPoint);

    this->MyUsbAddress = UsbAddress;
    this->MaximumControlPacketSize = MaximumControlPacketSize;

    return iUsbMouseDevice(this);
}

/* ** 
*/
SCODE __cdecl UsbMouseConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * MyName,
        ADDRESS UnusedRegisterBase,
        UINT UnusedInterrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        _TCHAR *HciName,
        PIUNKNOWN /*PIUSBMASTERCONTROLLER*/ UsbTreeLocation,
        ADDRESS UsbAddress,
        ADDRESS MaximumControlPacketSize)
{
    PIUNKNOWN pUnk;

    UnusedParameter(UnusedRegisterBase);
    UnusedParameter(UnusedInterrupt);

    DBGME(0,_tprintf(_T("UsbMouseConstructor(%x %s %x %s %d %x)\n"),
                     (ADDRESS) pThis, MyName, (ADDRESS) ppUnk,
                     (HciName) ? HciName : _T("NULL"),
                     UsbAddress, (ADDRESS) UsbTreeLocation));

    pUnk = (PIUNKNOWN) UsbMouseNew(MyName,
                                   HciName,
                                   (UINT8) UsbAddress,
                                   (UINT8) MaximumControlPacketSize,
                                   UsbTreeLocation);

    DBGME(0,_tprintf(_T("UsbMouseConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UsbMouseConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UsbMouseDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UsbMouseConstructor
};

static struct _DRIVERCOB UsbMouseDriverCobObject = { &UsbMouseDriverCobVtable, 1 };

PIUNKNOWN UsbMouseCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UsbMouseDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UsbMouseDeviceUsbMouseVtbl, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &UsbMouseDriverCobObject;
}
