/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbnull.c
 *
 * Purpose:
 *
 *    Empty device driver for USB devices.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <string.h>
#include <stddef.h> /* for offsetof() */
//#define _DEBUG 1
//#define DBGLVL 0
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbmaster.h>

typedef struct _USBNULL {
    struct CDevice CDevice;
    CONST_VTBL struct IUsbDriverVtbl *UsbNullDriverv;
    UINT8 MyUsbAddress;
    PIUSBHUB Hub;
    UINT8 PortNo;
    UINT8 MaximumControlPacketSize;
    int NullNo;
#define MYNAMELENGTH 12 /* bignuf */
#define MakeMyName(_this_,_name_)  _stprintf(_name_,_T("usbnull%d"),(_this_)->NullNo)
} USBNULL, *PUSBNULL;

/* Convert between interface and instance pointer and vice versa */
#define pUsbNullDevice(_i_) ((PUSBNULL)((_i_)))
#define iUsbNullDevice(_p_) ((PIDEVICE)((_p_)))
#define pUsbNullDriver(_i_) ((PUSBNULL)((ADDRESS)(_i_) - offsetof(struct _USBNULL, UsbNullDriverv)))
#define iUsbNullDriver(_p_) ((PIUSBDRIVER)(&((PUSBNULL)(_p_))->UsbNullDriverv))

/* IDevice interface
 */
PRIVATE SCODE UsbNullDeviceQueryInterface (
    PIDEVICE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbNullDevQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));

    /* IDevice
     */
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_IFile,iid)
        || UuidCmp(&IID_IDevice,iid))
    {
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }

    /* IUsbDriver
     */
    if (UuidCmp(&IID_IUsbDriver, iid)) {
        /* Use a single refcount
         */
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (void*) iUsbNullDriver(pThis);
        return S_OK;
    }

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT UsbNullDeviceAddRef (
    PIDEVICE pThis)
{
    DBGME(0,_tprintf(_T("UsbNullDevAR(%x)\n"),(ADDRESS)pThis));
    return AtomicInc(&pUsbNullDevice(pThis)->CDevice.RefCnt);
}

PRIVATE UINT UsbNullDeviceRelease (
    PIDEVICE pThis)
{
    PUSBNULL this = pUsbNullDevice(pThis);
    UINT Refs = AtomicDec(&this->CDevice.RefCnt);
    DBGME(0,_tprintf(_T("UsbNullDevRel(%x,%d)\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        pHeap->v->Free(pHeap, 0, (void *) this);
    }
    return Refs;
}

/* IUsbDriver interface
 */
PRIVATE SCODE UsbNullDriverQueryInterface (
    PIUSBDRIVER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbNullDrivQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));
    return UsbNullDeviceQueryInterface(iUsbNullDevice(pUsbNullDriver(pThis)), iid, ppObject);
}

PRIVATE UINT UsbNullDriverAddRef (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbNullDrivAR(%x)\n"),(ADDRESS)pThis));
    return UsbNullDeviceAddRef(iUsbNullDevice(pUsbNullDriver(pThis)));
}

PRIVATE UINT UsbNullDriverRelease (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbNullDrivRel(%x)\n"),(ADDRESS)pThis));
    return UsbNullDeviceRelease(iUsbNullDevice(pUsbNullDriver(pThis)));
}

/* Utilities
 */

/* Assign a configuration index to the device,
 * such that it enters the configured state.
 */
PRIVATE SCODE SetConfiguration(
    PUSBNULL this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 Index)
{
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->SetupData[0] = 0;
    Urb->SetupData[1] = USB_R_SET_CONFIG;
    Urb->SetupData[2] = Index;
    Urb->SetupData[3] = 0;
    Urb->SetupData[4] = 0;
    Urb->SetupData[5] = 0;
    Urb->SetupData[6] = 0;
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = 0;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,Urb,this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbNull:SC SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbNull:SC SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}
    
PRIVATE SCODE UsbNullDriverIamYourHub (
    PIUSBDRIVER pThis,
    PIUSBHUB Hub,
    UINT PortNumber)
{
    PUSBNULL this = pUsbNullDriver(pThis);
    PUSBM_REQUEST_BLOCK Urb = NULL;
    SCODE sc;
    PINAMESPACE Ns = CurrentNameSpace();
    _TCHAR MyName[MYNAMELENGTH];

    DBGME(0,_tprintf(_T("UsbNull:IYH(%x,%x,%d)\n"),
                     (ADDRESS)this,(ADDRESS)Hub,PortNumber));

    /* Check not already owned
     */
    if (this->Hub) {
        DBGME(3,_tprintf(_T("UsbNull: already owned.\n")));
        return E_ALREADY_EXISTS;
    }

    /* QI to make sure and to take a ref
     */
    sc = Hub->v->QueryInterface(Hub,
                                &IID_IUsbHub,
                                (void **)&this->Hub);
    if (FAILED(sc)) {
        DBGME(1,_tprintf(_T("UsbNull bad Hub, sc=%x\n"),sc));
        return sc;
    }

    this->PortNo = (UINT8) PortNumber;
    //done above: this->Hub = Hub;

    /* Get an Urb
     */
    Urb = malloc(256 + sizeof *Urb);
    if (Urb == NULL) {
        sc = E_NOT_ENOUGH_MEMORY;
        goto Error;
    }
    memset(Urb,0,sizeof *Urb);
    Urb->Done = &this->CDevice.IntrPoint;
    Urb->Data = (UINT8*)(Urb+1);

    /* Set the default configuration idex
     */
    sc = SetConfiguration(this,Urb,1);
    if (FAILED(sc))
        goto Error;

    /* Add top-level namespace entry, for user convenience (warning, really)
     * HUB only puts us in the USB tree.
     * Try again to avoid clashes if multiple null are attached
     * Put a limit to it, though, just in case.
     */
    for (;;) {
        MakeMyName(this,MyName);
        sc = Ns->v->Register(Ns,
                             MyName,
                             (PIUNKNOWN) this,
                             NAME_SPACE_READ | 
                             NAME_SPACE_WRITE |
                             NAME_SPACE_CREATE |
                             NAME_SPACE_FAILIFEXIST,
                             NULL);
        if ((sc == E_ALREADY_EXISTS) && (this->NullNo < 1000))
            this->NullNo++;
        else
            break;
    }

    return S_OK;
 Error:
    if (this->Hub) {
        this->Hub->v->Release(this->Hub);
        this->Hub = NULL;
    }
    return sc;
}

PRIVATE SCODE UsbNullDriverYouAreTerminated (
    PIUSBDRIVER pThis,
    PIUSBHUB SaysWho)
{
    PUSBNULL this = pUsbNullDriver(pThis);
    SCODE sc;
    PINAMESPACE Ns = CurrentNameSpace();
    _TCHAR MyName[MYNAMELENGTH];

    DBGME(0,_tprintf(_T("UsbNull:YAT(%x)\n"),(ADDRESS)SaysWho));

    /* Check args, and that we had started ok
     */
    if (this->Hub == NULL) {
        return E_INVALID_PARAMETER;
    }

    /* Remove from namespace
     */
    MakeMyName(this,MyName);
    sc = Ns->v->Unregister(Ns,MyName);

    /* Cancel anything pending
     */

    /* Let the HUB go
     */
    this->Hub->v->Release(this->Hub);
    this->Hub = NULL;

    return S_OK;
}

PRIVATE SCODE UsbNullDriverGetUsbAddress (
    PIUSBDRIVER pThis,
    PUINT UsbAddress,
    PUINT PortNumber
                      )
{
    PUSBNULL this = pUsbNullDriver(pThis);

    if (UsbAddress == NULL)
        return E_INVALID_PARAMETER;
    if (this->MyUsbAddress == 0)
        return E_OBJECT_NOT_INITIALIZED;

    *UsbAddress = this->MyUsbAddress;
    *PortNumber = this->PortNo;
    return S_OK;
}

#include "s_UsbNull_v.c"

/* Constructor */
PIDEVICE UsbNullNew(
    const _TCHAR *MyName,
    _TCHAR *HciName,
    UINT8 UsbAddress,
    UINT8 MaximumControlPacketSize,
    PIUNKNOWN /*IUSBMASTERCONTROLLER*/ UsbTreeLocation)
{
    PIHEAP pHeap = CurrentHeap();
    PUSBNULL this;

    if (UsbAddress == 0) {
        DBGME(3,_tprintf(_T("UsbNullNew: bad UsbAddress 0\n")));
        return NULL;
    }

    this = (PUSBNULL) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *this, 0);
    if (!this)
        return NULL;

    this->CDevice.v = (struct CDeviceVtbl*)&UsbNullDeviceUsbNullVtbl;
    this->CDevice.RefCnt = 1;
    this->UsbNullDriverv = &UsbNullDriverUsbNullVtbl;

    /* Do this because it aint config that starts us
     */
    Condition_Init(&this->CDevice.IntrPoint);

    this->MyUsbAddress = UsbAddress;
    this->MaximumControlPacketSize = MaximumControlPacketSize;

    return iUsbNullDevice(this);
}

/* ** 
*/
SCODE __cdecl UsbNullConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * MyName,
        ADDRESS UnusedRegisterBase,
        UINT UnusedInterrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        _TCHAR *HciName,
        PIUNKNOWN /*PIUSBMASTERCONTROLLER*/ UsbTreeLocation,
        ADDRESS UsbAddress,
        ADDRESS MaximumControlPacketSize)
{
    PIUNKNOWN pUnk;

    UnusedParameter(UnusedRegisterBase);
    UnusedParameter(UnusedInterrupt);

    DBGME(0,_tprintf(_T("UsbNullConstructor(%x %s %x %s %d %x)\n"),
                     (ADDRESS) pThis, MyName, (ADDRESS) ppUnk,
                     (HciName) ? HciName : _T("NULL"),
                     UsbAddress, (ADDRESS) UsbTreeLocation));

    pUnk = (PIUNKNOWN) UsbNullNew(MyName,
                                   HciName,
                                   (UINT8) UsbAddress,
                                   (UINT8) MaximumControlPacketSize,
                                   UsbTreeLocation);

    DBGME(0,_tprintf(_T("UsbNullConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UsbNullConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UsbNullDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UsbNullConstructor
};

static struct _DRIVERCOB UsbNullDriverCobObject = { &UsbNullDriverCobVtable, 1 };

PIUNKNOWN UsbNullCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UsbNullDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UsbNullDeviceUsbNullVtbl, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &UsbNullDriverCobObject;
}
