/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    protocol.c
 *
 * Purpose:
 *
 *    Transport protocols for USBSTOR.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <string.h>
#include <stddef.h> /* for offsetof() */
#define _DEBUG 1
int UsbStorClassesDebug = 2;
#define DBGLVL UsbStorClassesDebug
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbmaster.h>

#include "transport.h"
#include "class.h"

/* SCSI Class implementation
 */
#include "scsi.h"
#include "scsilib.h"

typedef struct _SCSI_DEVICE {
    const struct IFileVtbl *v;
    UINT RefCnt;
    PTRANSPORT Transport;
    UINT32 BlockSize;
    UINT32 DiskSize;
    UINT8  UniformCommandSize;
    UINT8  RemovableMedia;
#define EASYBUFFER_SIZE 128
    UINT8 EasyBuffer[EASYBUFFER_SIZE];
} SCSI_DEVICE, *PSCSI_DEVICE;

/* Utils
 */
/* Sleep for a few milliseconds
 */
PRIVATE void MilliSleep(int mSecs)
{
    TIME t;
    Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(mSecs)));
    SleepUntil(t);
}

/* Generic command execution routine
 */
#define FLOW_NO_DATA 0
#define FLOW_DATA_IN 1
#define FLOW_DATA_OUT 2

PRIVATE SCODE ScsiDoCommand(
    PSCSI_DEVICE this,
    PCDB Cdb,
    int Flow
    )
{
    SCODE sc;
    PTRANSPORT Transport = this->Transport;
    PCDB rCdb;

    DBGME(0,_tprintf(_T("SCSI: SendCmd\n")));
    sc = Transport->v->SendCommand(Transport,
                                   Cdb,
                                   TRUE);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbScsi: DoCommand Submit failed sc=x%x\n"),sc));
        goto CheckResetRecovery;
    }

    switch (Flow) {

    case FLOW_DATA_IN:
        DBGME(0,_tprintf(_T("SCSI: ReadData\n")));
        sc = Transport->v->ReadData(Transport,
                                    Cdb,
                                    TRUE);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbScsi: DoCommand read failed sc=x%x\n"),sc));
            goto CheckResetRecovery;
        }
        break;

    case FLOW_DATA_OUT:
        DBGME(0,_tprintf(_T("SCSI: WriteData\n")));
        sc = Transport->v->SendData(Transport,
                                    Cdb,
                                    TRUE);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbScsi: DoCommand writ failed sc=x%x\n"),sc));
            goto CheckResetRecovery;
        }
        break;

    case FLOW_NO_DATA:
    default:
        break;
    }

    /* Some transports are capable of command queueing, some are not.
     * Pass as argument what we expect to see completed.
     */
    DBGME(0,_tprintf(_T("SCSI: ReadStatus\n")));
    rCdb = Cdb;
    sc = Transport->v->ReadStatus(Transport,
                                  &rCdb);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbScsi: DoCommand status failed sc=x%x\n"),sc));
        goto CheckResetRecovery;
    }

    /* If the transport returns a tag, make sure its the right one.
     */
    if (rCdb && (rCdb != Cdb)) {
        DBGME(3,_tprintf(_T("UsbScsi: DoCommand cdb %x != %x !!\n"),
                         (ADDRESS)rCdb, (ADDRESS)Cdb));
    }

    return sc;

 CheckResetRecovery:
    if (sc == E_RESTART) {
        /* Needs reset recovery
         * BUGBUG SCSI doesnt need to know, does it??
         */
        SCODE sc1 = Transport->v->ResetRecovery(Transport);
        DBGME(3,_tprintf(_T("UsbScsi: RESET gets sc=x%x\n"),sc1));
        if (FAILED(sc1))
            return sc1;
    }
    return sc;
}

/* Compose, issue and check an INQUIRY
 */
PRIVATE SCODE ScsiDoInquiry(
    PSCSI_DEVICE this
    )
{
    CDB cdb;
    SCODE sc;
    SCSI2_INQUIRY_DATA *Data;

    memset(&cdb,0,sizeof cdb);
    cdb.DataPointer = this->EasyBuffer;
    cdb.DataLength  = sizeof(SCSI2_INQUIRY_DATA);
    cdb.Flags = CDB_F_IN;

    ScsiComposeInquiry(&cdb,SCSI_INQ_STD_DATA,sizeof(SCSI2_INQUIRY_DATA));
    if (this->UniformCommandSize) 
        cdb.CbLength = this->UniformCommandSize;

    DBGME(1,_tprintf(_T("SCSI: INQUIRY\n")));
    sc = ScsiDoCommand(this, &cdb, FLOW_DATA_IN);

    /* Check to see if this is a removable media (floppy)
     */
    if (!FAILED(sc)) {
        Data = (SCSI2_INQUIRY_DATA *)cdb.DataPointer;
        ScsiDumpInquiryData(Data,(DBGLVL==0),this->Transport->InquiryName);

        this->RemovableMedia = (Data->rmb != 0);        
    }

    return sc;
}

/* Compose, issue and check an READ CAPACITY
 */
PRIVATE SCODE ScsiDoReadCapacity(
    PSCSI_DEVICE this
    )
{
    CDB cdb;
    SCODE sc;
    SCSI_RCAP_DATA *Data;

    memset(&cdb,0,sizeof cdb);
    cdb.DataPointer = this->EasyBuffer;
    cdb.DataLength  = sizeof(SCSI_RCAP_DATA);
    cdb.Flags = CDB_F_IN;

    DBGME(1,_tprintf(_T("SCSI: READCAP\n")));
    ScsiComposeReadCapacity(&cdb,0);
    if (this->UniformCommandSize) 
        cdb.CbLength = this->UniformCommandSize;

    sc = ScsiDoCommand(this,&cdb, FLOW_DATA_IN);
    if (FAILED(sc))
        return sc;

    Data = (SCSI_RCAP_DATA *)cdb.DataPointer;
    DBGME(2,ScsiDumpCapacityData(Data));

    /* Returns the last valid block address */
    this->DiskSize  = 1;
    this->DiskSize += (Data->lba1<<24) |
                      (Data->lba2<<16) |
                      (Data->lba3<< 8) |
                       Data->lba4;
    this->BlockSize = (Data->blen1<<24) |
                      (Data->blen2<<16) |
                      (Data->blen3<<8 ) |
                       Data->blen4;
    return sc;
}

/* Compose, issue and check a REQUEST SENSE
 */
PRIVATE SCODE ScsiDoRequestSense(
    PSCSI_DEVICE this
    )
{
    CDB cdb;
    SCODE sc;
    SCSI_SENSE_DATA *Data;

    memset(&cdb,0,sizeof cdb);
    cdb.DataPointer = this->EasyBuffer;
    cdb.DataLength  = sizeof(SCSI_SENSE_DATA);
#if 1
    memset(cdb.DataPointer,0,cdb.DataLength);
#endif
    cdb.Flags = CDB_F_IN;

    ScsiComposeRequestSense(&cdb,sizeof(SCSI_SENSE_DATA));
    if (this->UniformCommandSize) 
        cdb.CbLength = this->UniformCommandSize;

    DBGME(1,_tprintf(_T("SCSI: REQUEST SENSE\n")));
    sc = ScsiDoCommand(this, &cdb, FLOW_DATA_IN);

    if (!FAILED(sc)) {
        Data = (SCSI_SENSE_DATA *)cdb.DataPointer;
        ScsiDumpSenseData(Data,cdb.DataLength);
    }

    return sc;
}

/* Compose, issue and check a READ
 */
PRIVATE SCODE ScsiDoRead(
    PSCSI_DEVICE this,
    UINT32 StartBlock,
    UINT32 nBlocks,
    BYTE *Buffer,
    UINT *pSizeRead
    )
{
    CDB cdb;
    SCODE sc;
    UINT BlocksRemaining = nBlocks;
    UINT MaxBlocks;

    *pSizeRead = 0;

    memset(&cdb,0,sizeof cdb);

    /* Break it down in pieces no larger than 4KB.
     * The OHCI driver can only handle one 4kb page crossing.
     * BUGBUG There should be a better way to handle this.
     */
    MaxBlocks = 0x1000 / this->BlockSize;

    do {
        nBlocks = BlocksRemaining;
        if (nBlocks > MaxBlocks)
            nBlocks = MaxBlocks;

        cdb.DataPointer = Buffer;
        cdb.DataLength  = (UINT16)(nBlocks * this->BlockSize);
        cdb.Flags = CDB_F_IN;

        /* Apparently, it only understands the long form
         */
        ScsiComposeLongRead(&cdb,StartBlock,(UINT16)nBlocks);
        if (this->UniformCommandSize) 
            cdb.CbLength = this->UniformCommandSize;

        DBGME(1,_tprintf(_T("SCSI: READ %x\n"),StartBlock));
        sc = ScsiDoCommand(this,&cdb, FLOW_DATA_IN);
        if (FAILED(sc)) {
            /* force a reqsns */
            if (sc == E_RESTART) {
                ScsiDoRequestSense(this);
                continue;
            } else{
                /* force a full restart */
                this->BlockSize = 0;
                return sc;
            }
        }

        *pSizeRead += cdb.DataLength;
        Buffer     += cdb.DataLength;

        StartBlock += nBlocks;
        BlocksRemaining -= nBlocks;
    } while (BlocksRemaining);

    return sc;
}

/* Compose, issue and check a START/STOP UNIT
 */
PRIVATE SCODE ScsiDoStartUnit(
    PSCSI_DEVICE this,
    BOOL TurnItOn
    )
{
    CDB cdb;
    SCODE sc;

    /* no data */
    memset(&cdb,0,sizeof cdb);

    ScsiComposeStartUnit(&cdb,SCSI_CMD_SS_START);
    if (this->UniformCommandSize) 
        cdb.CbLength = this->UniformCommandSize;

    DBGME(0,_tprintf(_T("SCSI: START/STOP\n")));
    sc = ScsiDoCommand(this, &cdb, FLOW_NO_DATA);

    DBGME(1,_tprintf(_T("SCSI: START/STOP->%x.%x\n"), sc, cdb.FinalStatus));
    if (!FAILED(sc))
        sc = cdb.FinalStatus;

    return sc;
}

/* Compose, issue and check a TEST UNIT READY
 */
PRIVATE SCODE ScsiDoTestUnitReady(
    PSCSI_DEVICE this
    )
{
    CDB cdb;
    SCODE sc;

    /* no data */
    memset(&cdb,0,sizeof cdb);

    ScsiComposeTestUnitReady(&cdb);
    if (this->UniformCommandSize) 
        cdb.CbLength = this->UniformCommandSize;

    DBGME(0,_tprintf(_T("SCSI: TEST U READY\n")));
    sc = ScsiDoCommand(this, &cdb, FLOW_NO_DATA);

    DBGME(1,_tprintf(_T("SCSI: TEST U READY->%x.%x\n"), sc, cdb.FinalStatus));
    if (!FAILED(sc))
        sc = cdb.FinalStatus;

    return sc;
}

/* Start a floppy or spinup the disk.
 * Return E_NO_MEDIA_IN_DRIVE if media absent.
 */
PRIVATE SCODE ScsiStartDisk(
    PSCSI_DEVICE this
    )
{
    SCODE sc;
    int i;

    /* See whassup
     */
    sc = ScsiDoRequestSense(this);
    if (FAILED(sc))
        goto Error;

    /* Wait until it spins up
     * Try a few times to work around not-ready-to-ready-transitions.
     */
    for (i = 0; i < 4; i++) {
        sc = ScsiDoTestUnitReady(this);
        if (!FAILED(sc))
            break;
        sc = ScsiDoRequestSense(this);
        if (FAILED(sc))
            goto Error;
    }

#if 0
    /* This gets the Iomega Zip drives very confused. So dont. */
    /* We must start a floppy to know its capacity, it depends on the media.
     * Then again, there might not be a media in the unit right now.
     * Try a few times, clearing up the sense data.
     */
    for (i = 0; i < 4; i++) {
        sc = ScsiDoStartUnit(this,TRUE);
        if (!FAILED(sc))
            break;
        sc = ScsiDoRequestSense(this);
        if (FAILED(sc))
            goto Error;
        MilliSleep(100);
    }
#endif

    /* At this point this must not fail.
     */
    sc = ScsiDoReadCapacity(this);
    if (FAILED(sc))
        goto Error;

    return S_OK;
 Error:
    ScsiDoRequestSense(this);

    /* If its not removable then its broken
     */
    if (this->RemovableMedia)
        return E_NO_MEDIA_IN_DRIVE;
    else
        return E_DISK_RESET_FAILED;
}


/* IFile interface
 */
PRIVATE SCODE MCT ScsiQueryInterface (PIFILE pThis, REFIID Iid, void* *ppNew)
{
    PSCSI_DEVICE this = (PSCSI_DEVICE)pThis;

    if (UuidCmp(&IID_IFile,Iid) || UuidCmp(&IID_IUnknown,Iid))
    {
        AtomicInc(&this->RefCnt);
        *ppNew = (PIUNKNOWN) pThis;
        return NO_ERROR;
    }
    *ppNew = NULL;
    return E_NO_INTERFACE;
}

/* Add a reference
 */
PRIVATE UINT MCT ScsiAddRef ( PIFILE pThis )
{
    PSCSI_DEVICE this = (PSCSI_DEVICE)pThis;
    return AtomicInc(&this->RefCnt);
}

/* Release a reference
 */
PRIVATE UINT MCT ScsiRelease ( PIFILE pThis )
{
    PSCSI_DEVICE this = (PSCSI_DEVICE)pThis;
    UINT Refs;

    Refs = AtomicDec(&this->RefCnt);
    if (Refs == 0)
        free(this);
    return Refs;
}

/* 
 */
PRIVATE SCODE MCT ScsiReadAt ( PIFILE pThis, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    UINT64 Temp;
    SCODE sc;
    PSCSI_DEVICE this = (PSCSI_DEVICE)pThis;
    UINT32 StartingBlock;
    UINT32 NumberOfBlocks;

    DBGME(1,_tprintf(_T("UsbStor:ScsiReadAt(%x,%x,%x)\n"),
                     (ADDRESS)pThis,Uint64ToUint32(Position),Size));

    /* Check if there is media in the unit
     */
    if (this->BlockSize == 0) {
        sc = ScsiStartDisk(this);
        if (FAILED(sc))
            return sc;
    }

    /* Where? */
    Temp = Uint64DividedByUint32(Position,this->BlockSize);
    StartingBlock = Uint64ToUint32(Temp);

    /* Howmany? */
    NumberOfBlocks = Size / this->BlockSize;
    if (NumberOfBlocks == 0)
        return E_INVALID_PARAMETER;

    /* EOF or past? */
    if (StartingBlock >= this->DiskSize) {
        if (StartingBlock != this->DiskSize)
            return E_INVALID_PARAMETER;
        /* EOF */
        *pSizeRead = 0;
        return S_FALSE;
    }

    /* Beyond? */
    if ((StartingBlock + NumberOfBlocks) > this->DiskSize) {
        NumberOfBlocks = this->DiskSize - StartingBlock;
    }

    sc = ScsiDoRead(this,StartingBlock,NumberOfBlocks,pBuffer,pSizeRead);

    return sc;
}

PRIVATE SCODE MCT ScsiWriteAt ( PIFILE pThis, UINT64 Position, const BYTE *pBuffer,
                               UINT Size, PUINT pSizeWritten)
{
    DBGME(1,_tprintf(_T("UsbStor:ScsiWriteAt(%x,%x)\n"),
                     (ADDRESS)pThis,Uint64ToUint32(Position)));
    UnusedParameter(pThis);
    UnusedParameter(Position);
    UnusedParameter(pBuffer);
    UnusedParameter(Size);
    UnusedParameter(pSizeWritten);

    return E_ACCESS_DENIED; /* ? */
}

/* This one is used by FATFS to poll for media changes.
 */
PRIVATE SCODE MCT ScsiSetSize(  PIFILE pThis, UINT64 Size)
{
    PSCSI_DEVICE this = (PSCSI_DEVICE)pThis;
    static const UINT64 MediaPoll = Int64Initializer(~0,~0);

    DBGME(0,_tprintf(_T("UsbStor:ScsiSetSize(%x,%x)\n"),
                     (ADDRESS)pThis,Uint64ToUint32(Size)));

    /* Return E_MEDIA_CHANGED if the floppy was swapped/ejected/not started.
     */
    if (Int64Equal(Size,MediaPoll)) {
        if (this->BlockSize == 0)
            return E_MEDIA_CHANGED;
        return S_OK;
    }

    return E_ACCESS_DENIED; /* ? */
}

PRIVATE SCODE MCT ScsiGetSize(  PIFILE pThis, PUINT64 pSize)
{
    PSCSI_DEVICE this = (PSCSI_DEVICE)pThis;

    DBGME(0,_tprintf(_T("UsbStor:ScsiGetSize(%x,%x)\n"),
                     (ADDRESS)pThis,(ADDRESS)pSize));

    if (this->DiskSize == 0)
        return E_OBJECT_NOT_INITIALIZED;

    Uint32ToUint64(*pSize,this->DiskSize);
    *pSize = Uint64TimesUint32(*pSize,this->BlockSize);
    return S_OK;
}

const struct IFileVtbl ScsiVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    ScsiQueryInterface,
    ScsiAddRef,
    ScsiRelease,
    ScsiReadAt,
    ScsiWriteAt,
    ScsiSetSize,
    ScsiGetSize
};    

/* Contructor for a class that uses some form of SCSI commands
 */
PRIVATE SCODE 
CreateGenericScsiClass(
    PTRANSPORT Transport,
    PIFILE *ppFile,
    UINT8 FixedSize
    )
{
    SCODE sc;
    PSCSI_DEVICE Device;
    int i;

    /* Allocate and initialize our object
     */
    Device = malloc(sizeof(SCSI_DEVICE));
    if (Device == NULL)
        return E_NOT_ENOUGH_MEMORY;

    Device->v = &ScsiVtbl;
    Device->RefCnt = 1;
    Device->Transport = Transport;
    Device->BlockSize = 0;
    Device->DiskSize = 0;
    Device->UniformCommandSize = FixedSize;

    /* Unless we find otherwise..
     */
    *ppFile = (PIFILE) Device;

    /* Do a REQUEST_SENSE first thing first
     * Repeat a few times to give it time to initialize.
     */
    for (i = 0; i < 10; i++) {
        sc = ScsiDoRequestSense(Device);
        if (!FAILED(sc))
            break;
        MilliSleep(100);
    }

    /* Issue an INQUIRY
     * This tells us if this is a removable media device.
     */
    sc = ScsiDoInquiry(Device);
    if (FAILED(sc)) {
        goto Error;
    }

    /* Find out how big the disk is, unless its a floppy.
     */
    if (!Device->RemovableMedia) {
        sc = ScsiDoReadCapacity(Device);
        if (FAILED(sc))
            goto Error;
    }

    return S_OK;

 Error:
    free(Device);
    return sc;
}

/* Contructor for the SCSI class
 */
SCODE CreateScsiClass(PTRANSPORT Transport, PIFILE *ppFile)
{
    return CreateGenericScsiClass(Transport,ppFile,0);
}

/* Other constructors
 */
SCODE CreateAtapiClass(PTRANSPORT Transport, PIFILE *ppFile)
{
    UnusedParameter(Transport);
    return E_NOT_IMPLEMENTED;
}

SCODE CreateQicClass(PTRANSPORT Transport, PIFILE *ppFile)
{
    UnusedParameter(Transport);
    return E_NOT_IMPLEMENTED;
}

SCODE CreateUfiClass(PTRANSPORT Transport, PIFILE *ppFile)
{
    return CreateGenericScsiClass(Transport,ppFile,12);
}

SCODE CreateSffClass(PTRANSPORT Transport, PIFILE *ppFile)
{
    UnusedParameter(Transport);
    return E_NOT_IMPLEMENTED;
}

