/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    scsi.h
 *
 * Purpose:
 *
 *    SCSI Class definitions
 *
 * References:
 *    "Small Computer Systems Interface (SCSI)", ANSI Draft
 *     X3T9.2/82-2 - Rev 17B December 1985   (SCSI-1)
 *
 *    "Small Computer System Interface - 2 (SCSI-II)", ANSI Draft
 *     X3T9.2/86-109 -  Rev 10C March 1990
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

/*
 *    Definitions for the SCSI-1 Standard
 */

/*
 * Bus phases
 */

#define    SCSI_IO                   0x01        /* Input/Output */
#define    SCSI_CD                   0x02        /* Command/Data */
#define    SCSI_MSG                  0x04        /* Message */

#define    SCSI_PHASE_MASK           0x07
#define    SCSI_PHASE(x)             ((x)&SCSI_PHASE_MASK)

#define    SCSI_PHASE_DATAO          0x00                        /* 0 */
#define    SCSI_PHASE_DATAI          SCSI_IO                     /* 1 */
#define    SCSI_PHASE_CMD            SCSI_CD                     /* 2 */
#define    SCSI_PHASE_STATUS         (SCSI_CD|SCSI_IO)           /* 3 */
                                  /* 4..5 ANSI reserved */
#define    SCSI_PHASE_MSG_OUT        (SCSI_MSG|SCSI_CD)          /* 6 */
#define    SCSI_PHASE_MSG_IN         (SCSI_MSG|SCSI_CD|SCSI_IO)  /* 7 */

/*
 * Single byte messages
 *
 * originator:   I-nitiator T-arget
 * T-support:    M-andatory O-ptional
 */

#define SCSI_COMMAND_COMPLETE        0x00    /* M T   */
#define SCSI_EXTENDED_MESSAGE        0x01    /*   IT  */
#define SCSI_SAVE_DATA_POINTER       0x02    /* O T   */
#define SCSI_RESTORE_POINTERS        0x03    /* O T   */
#define SCSI_DISCONNECT              0x04    /* O T   */
#define SCSI_I_DETECTED_ERROR        0x05    /* M I   */
#define SCSI_ABORT                   0x06    /* M I   */
#define SCSI_MESSAGE_REJECT          0x07    /* M IT  */
#define SCSI_NOP                     0x08    /* M I   */
#define SCSI_MSG_PARITY_ERROR        0x09    /* M I   */
#define SCSI_LNKD_CMD_COMPLETE       0x0a    /* O T   */
#define SCSI_LNKD_CMD_COMPLETE_F     0x0b    /* O T   */
#define SCSI_BUS_DEVICE_RESET        0x0c    /* M I   */
                    /* 0x0d..0x11 scsi2 */
                    /* 0x12..0x1f reserved */
#define SCSI_IDENTIFY                0x80    /* IT */
# define SCSI_IFY_ENABLE_DISCONNECT  0x40    /* I  */
# define SCSI_IFY_LUNTAR             0x20    /* IT */
# define SCSI_IFY_LUN_MASK           0x07    /* IT */


/* Message codes 0x30..0x7f are reserved */

/*
 * Extended messages, codes and formats
 */

#define SCSI_MODIFY_DATA_PTR         0x00    /* T  */
typedef struct _SCSI_MOD_PTR {
    UINT8    xtn_msg_tag;            /* const 0x01 */
    UINT8    xtn_msg_len;            /* const 0x05 */
    UINT8    xtn_msg_code;           /* const 0x00 */
    UINT8    xtn_msg_arg_1000;       /* MSB, signed 2cmpl */
    UINT8    xtn_msg_arg_0200;
    UINT8    xtn_msg_arg_0030;
    UINT8    xtn_msg_arg_0004;       /* LSB */
} SCSI_MOD_PTR, *PSCSI_MOD_PTR;

#define SCSI_SYNC_XFER_REQUEST       0x01    /* IT */
typedef struct _SCSI_SYNCH_XFER_REQ {
    UINT8    xtn_msg_tag;            /* const 0x01 */
    UINT8    xtn_msg_len;            /* const 0x03 */
    UINT8    xtn_msg_code;           /* const 0x01 */
    UINT8    xtn_msg_xfer_period;    /* times 4nsecs */
    UINT8    xtn_msg_xfer_offset;    /* pending ack window */
#define SCSI_SYNCH_XFER_OFFANY       0xff    /* T  unlimited */
} SCSI_SYNCH_XFER_REQ, *PSCSI_SYNCH_XFER_REQ;

#define SCSI_EXTN_IDENTIFY           0x02    /* IT -2 */
typedef struct _SCSI_XTN_IDENTIFY {
    UINT8    xtn_msg_tag;            /* const 0x01 */
    UINT8    xtn_msg_len;            /* const 0x02 */
    UINT8    xtn_msg_code;           /* const 0x02 */
    UINT8    xtn_msg_sublun;
} SCSI_XTN_IDENTIFY, *PSCSI_XTN_IDENTIFY;

                    /* 0x03..0x7f reserved */

#define SCSI_XTN_VENDOR_UQE          0x80    /* vendor unique bit */
typedef struct _SCSI_XTN_VEDOR_UNIQUE {
    UINT8    xtn_msg_tag;            /* const 0x01 */
    UINT8    xtn_msg_len;            /* args' len+1 (0-->256)*/
    UINT8    xtn_msg_code;           /* const 0x80..0xff */
    UINT8    xtn_msg_args[1];        /* 0..255 bytes */
} SCSI_XTN_VEDOR_UNIQUE, *PSCSI_XTN_VEDOR_UNIQUE;


/*
 * Commands, generic structures
 */

/* SIX byte commands
 */
typedef struct _SCSI_COMMAND_GROUP_0 {
    UINT8    scsi_cmd_code;          /* group(7..5) and command(4..1) */
#define SCSI_CODE_GROUP              0xe0
#define SCSI_CODE_CMD                0x1f
    UINT8    scsi_cmd_lun_and_lba1;  /* lun(7..5) and block# msb[20..16] */
#define SCSI_LUN_MASK                0xe0
#define SCSI_LBA_MASK                0x1f
# define SCSI_LUN_SHIFT              5
    UINT8    scsi_cmd_lba2;          /* block#[15.. 8] */
    UINT8    scsi_cmd_lba3;          /* block#[ 7.. 0] */
    UINT8    scsi_cmd_xfer_len;      /* if required */
    UINT8    scsi_cmd_ctrl_byte;     /* contains: */
#define SCSI_CTRL_VUQ                0xc0    /* vendor unique bits */
#define SCSI_CTRL_RESVD              0x3c    /* reserved, mbz */
#define SCSI_CTRL_FLAG               0x02    /* snd complete_with_flag at end */
#define SCSI_CTRL_LINK               0x01    /* link this command with next */
} SCSI_COMMAND_GROUP_0, *PSCSI_COMMAND_GROUP_0;

/* TEN byte commands
 */
typedef struct _SCSI_COMMAND_GROUP_1 {
    UINT8    scsi_cmd_code;          /* group(7..5) and command(4..1) */
    UINT8    scsi_cmd_lun_and_relbit;/* lun(7..5) and RelAdr(0) */
#define SCSI_RELADR                  0x01
    UINT8    scsi_cmd_lba1;          /* block#[31..24] */
    UINT8    scsi_cmd_lba2;          /* block#[23..16] */
    UINT8    scsi_cmd_lba3;          /* block#[15.. 8] */
    UINT8    scsi_cmd_lba4;          /* block#[ 7.. 0] */
    UINT8    scsi_cmd_xxx;           /* reserved, mbz */
    UINT8    scsi_cmd_xfer_len_1;    /* if required */
    UINT8    scsi_cmd_xfer_len_2;    /* if required */
    UINT8    scsi_cmd_ctrl_byte;     /* see above */
} SCSI_COMMAND_GROUP_1, *PSCSI_COMMAND_GROUP_1,
  SCSI_COMMAND_GROUP_2, *PSCSI_COMMAND_GROUP_2;

/* TWELVE byte commands
 */
typedef struct _SCSI_COMMAND_GROUP_5 {
    UINT8    scsi_cmd_code;          /* group(7..5) and command(4..1) */
    UINT8    scsi_cmd_lun_and_relbit;/* lun(7..5) and RelAdr(0) */
    UINT8    scsi_cmd_lba1;          /* block#[31..24] */
    UINT8    scsi_cmd_lba2;          /* block#[23..16] */
    UINT8    scsi_cmd_lba3;          /* block#[15.. 8] */
    UINT8    scsi_cmd_lba4;          /* block#[ 7.. 0] */
    UINT8    scsi_cmd_xfer_len_1;    /* if required */
    UINT8    scsi_cmd_xfer_len_2;    /* if required */
    UINT8    scsi_cmd_xfer_len_3;    /* if required */
    UINT8    scsi_cmd_xfer_len_4;    /* if required */
    UINT8    scsi_cmd_xxx1;          /* reserved, mbz */
    UINT8    scsi_cmd_ctrl_byte;     /* see above */
} SCSI_COMMAND_GROUP_5, *PSCSI_COMMAND_GROUP_5;


/*
 * Commands, codes and aliases
 */

                /* GROUP 0 */
typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_TEST_UNIT_READY;   
#define SCSI_CMD_TEST_UNIT_READY     0x00    /* O all 2M all */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_REZERO;   
#define SCSI_CMD_REZERO_UNIT         0x01    /* O disk worm rom */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_REWIND;   
#define SCSI_CMD_REWIND              0x01    /* M tape */
# define SCSI_CMD_REW_IMMED          0x01

                    /* 0x02 vendor unique */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_REQUEST_SENSE;   
#define SCSI_CMD_REQUEST_SENSE       0x03    /* M all */
# define scsi_cmd_allocation_length  scsi_cmd_xfer_len

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_FORMAT;   
#define SCSI_CMD_FORMAT_UNIT         0x04    /* M disk O prin */
# define SCSI_CMD_FMT_FMTDATA        0x10
# define SCSI_CMD_FMT_CMPLIST        0x08
# define SCSI_CMD_FMT_LIST_TYPE      0x07
# define scsi_cmd_intleave1          scsi_cmd_lba3
# define scsi_cmd_intleave2          scsi_cmd_xfer_len

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_BLOCK_LIMITS;   
#define SCSI_CMD_READ_BLOCK_LIMITS   0x05    /* E tape */

                    /* 0x06 vendor unique */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_REASSIGN_BLOCKS;   
#define SCSI_CMD_REASSIGN_BLOCKS     0x07    /* O disk worm */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_READ;   
#define SCSI_CMD_READ                0x08    /* M disk tape O worm rom */
#define    SCSI_CMD_RECEIVE          0x08    /* O proc */
#define    SCSI_CMD_GET_MESSAGE      0x08    /* M comm */
# define SCSI_CMD_TP_FIXED           0x01        /* tape */
# define scsi_cmd_tp_len1            scsi_cmd_lba2
# define scsi_cmd_tp_len2            scsi_cmd_lba3
# define scsi_cmd_tp_len3            scsi_cmd_xfer_len
                    /* largest addressable blockno */
#define    SCSI_CMD_READ_MAX_LBA     ((1 << 21) - 1)

                    /* 0x09 vendor unique */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_WRITE;   
#define SCSI_CMD_WRITE               0x0a    /* M disk tape O worm */
#define SCSI_CMD_PRINT               0x0a    /* M prin */
#define SCSI_CMD_SEND                0x0a    /* M proc */
#define SCSI_CMD_SEND_MESSAGE        0x0a    /* M comm */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_SEEK;   
#define SCSI_CMD_SEEK                0x0b    /* O disk worm rom */
#define SCSI_CMD_TRACK_SELECT        0x0b    /* O tape */
#define SCSI_CMD_SLEW_AND_PRINT      0x0b    /* O prin */
# define SCSI_CMD_SLW_CHANNEL        0x01
# define scsi_cmd_tp_trackno         scsi_cmd_xfer_len
# define scsi_cmd_slew_value         scsi_cmd_lba2

                    /* 0x0c..0x0e vendor unique */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_REV_READ;   
#define SCSI_CMD_READ_REVERSE        0x0f    /* O tape */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_WRITE_FILEMARKS;
#define SCSI_CMD_WRITE_FILEMARKS     0x10    /* M tape */
#define SCSI_CMD_FLUSH_BUFFER        0x10    /* M prin */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_SPACE;   
#define SCSI_CMD_SPACE               0x11    /* O tape */
# define SCSI_CMD_SP_BLOCKS          0x00
# define SCSI_CMD_SP_FIL             0x01
# define SCSI_CMD_SP_SEQ_FIL         0x02
# define SCSI_CMD_SP_EOT             0x03

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_INQUIRY;   
#define SCSI_CMD_INQUIRY             0x12    /* E all (2M all) */
# define SCSI_CMD_INQ_EVPD           0x01    /* 2 */
# define scsi_cmd_page_code          scsi_cmd_lba2    /* 2 */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_VERIFY;   
#define SCSI_CMD_VERIFY_0            0x13    /* O tape */
# define SCSI_CMD_VFY_BYTCMP         0x02

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_RECOVER_BUFFER;   
#define SCSI_CMD_RECOVER_BUFFERED_DATA 0x14    /* O tape prin */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_MODE_SELECT;   
#define SCSI_CMD_MODE_SELECT         0x15    /* O disk tape prin worm rom */
# define SCSI_CMD_MSL_PF             0x10
# define SCSI_CMD_MSL_SP             0x01

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_RESERVE;   
#define SCSI_CMD_RESERVE             0x16    /* O disk tape prin worm rom */
# define SCSI_CMD_RES_3RDPTY         0x10
# define SCSI_CMD_RES_3RDPTY_DEV     0x0e
# define SCSI_CMD_RES_EXTENT         0x01
# define scsi_cmd_reserve_id         scsi_cmd_lba2
# define scsi_cmd_extent_llen1       scsi_cmd_lba3
# define scsi_cmd_extent_llen2       scsi_cmd_xfer_len

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_RELEASE;   
#define SCSI_CMD_RELEASE             0x17    /* O disk tape prin worm rom */

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_COPY;   
#define SCSI_CMD_COPY                0x18    /* O all */
# define SCSI_CMD_CPY_PAD            0x01    /* 2 */
# define scsi_cmd_paraml_len0        scsi_cmd_lba2
# define scsi_cmd_paraml_len1        scsi_cmd_lba3
# define scsi_cmd_paraml_len2        scsi_cmd_xfer_len

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_ERASE;   
#define SCSI_CMD_ERASE               0x19    /* O tape */
# define SCSI_CMD_ER_LONG            0x01

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_MODE_SENSE;   
#define SCSI_CMD_MODE_SENSE          0x1a    /* O disk tape prin worm rom */
# define scsi_cmd_ms_pagecode        scsi_cmd_lba2

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_START;   
#define SCSI_CMD_START_STOP_UNIT     0x1b    /* O disk prin worm rom */
#define SCSI_CMD_LOAD_UNLOAD         0x1b    /* O tape */
# define SCSI_CMD_SS_IMMED           0x01
# define scsi_cmd_ss_flags           scsi_cmd_xfer_len
# define SCSI_CMD_SS_START           0x01
# define SCSI_CMD_SS_RETEN           0x02
# define SCSI_CMD_SS_RETAIN          0x01
# define SCSI_CMD_SS_EJECT           0x02

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_RECEIVE_DIAG;   
#define SCSI_CMD_RECEIVE_DIAG_RESULTS 0x1c    /* O all */
# define scsi_cmd_allocation_length1 scsi_cmd_lba3
# define scsi_cmd_allocation_length2 scsi_cmd_xfer_len

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_SEND_DIAG;   
#define SCSI_CMD_SEND_DIAGNOSTICS    0x1d    /* O all */
# define SCSI_CMD_DIAG_SELFTEST      0x04
# define SCSI_CMD_DIAG_DEVOFFL       0x02
# define SCSI_CMD_DIAG_UNITOFFL      0x01

typedef SCSI_COMMAND_GROUP_0         SCSI_COMMAND_MEDIUM_REMOVAL;   
#define SCSI_CMD_PREVENT_ALLOW_REMOVAL 0x1e    /* O disk tape worm rom */
# define scsi_cmd_pa_prevent         scsi_cmd_xfer_len /* 0x1 */

                    /* 0x1f reserved */

                /* GROUP 1 */
                    /* 0x20..0x24 vendor unique */

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_READ_CAPACITY; 
#define SCSI_CMD_READ_CAPACITY       0x25    /* E disk worm rom */
# define scsi_cmd_rcap_flags         scsi_cmd_xfer_len_2
# define SCSI_CMD_RCAP_PMI           0x01

                    /* 0x26..0x27 vendor unique */

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_LONG_READ;     
#define SCSI_CMD_LONG_READ           0x28    /* E disk M worm rom */

                    /* 0x29 vendor unique */

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_LONG_WRITE;   
#define SCSI_CMD_LONG_WRITE          0x2a    /* E disk M worm */

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_LONG_SEEK;   
#define SCSI_CMD_LONG_SEEK           0x2b    /* O disk worm rom */

                    /* 0x2c..0x2d vendor unique */

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_WRITE_VFY;   
#define SCSI_CMD_WRITE_AND_VERIFY    0x2e    /* O disk worm */
# define SCSI_CMD_VFY_BYTCHK         0x02

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_VERIFY_LONG;   
#define SCSI_CMD_VERIFY_1            0x2f    /* O disk worm rom */
# define SCSI_CMD_VFY_BLKVFY         0x04

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_SEARCH;   
#define SCSI_CMD_SEARCH_HIGH         0x30    /* O disk worm rom */
# define SCSI_CMD_SRCH_INVERT        0x10
# define SCSI_CMD_SRCH_SPNDAT        0x02
#define SCSI_CMD_SEARCH_EQUAL        0x31    /* O disk worm rom */
#define SCSI_CMD_SEARCH_LOW          0x32    /* O disk worm rom */

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_SET_LIMITS;   
#define SCSI_CMD_SET_LIMITS          0x33    /* O disk worm rom */
# define SCSI_CMD_SL_RDINH           0x02
# define SCSI_CMD_SL_WRINH           0x01

                    /* 0x34..0x38 reserved */

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_COMPARE;     
#define SCSI_CMD_COMPARE             0x39    /* O all */
# define scsi_cmd_1_paraml1          scsi_cmd_lba2
# define scsi_cmd_1_paraml2          scsi_cmd_lba3
# define scsi_cmd_1_paraml3          scsi_cmd_lba4

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_COPY_VFY;   
#define SCSI_CMD_COPY_AND_VERIFY     0x3a    /* O all */
# define SCSI_CMD_CPY_BYTCHK         0x02

                    /* 0x3b..0x3f reserved */

                /* GROUP 2 */
                    /* 0x40..0x5f reserved */

                /* GROUP 3 */
                    /* 0x60..0x7f reserved */

                /* GROUP 4 */
                    /* 0x80..0x9f reserved */

                /* GROUP 5 */
                    /* 0xa0..0xaf vendor unique */
                    /* 0xb0..0xbf reserved */

                /* GROUP 6 */
                    /* 0xc0..0xdf vendor unique */

                /* GROUP 7 */
                    /* 0xe0..0xff vendor unique */


/*
 * Command-specific results and definitions
 */

/* inquiry data */
typedef struct _SCSI_INQUIRY_DATA {
    UINT8    periph_type;
#define SCSI_DISK                    0x00
#define SCSI_TAPE                    0x01
#define SCSI_PRINTER                 0x02
#define SCSI_CPU                     0x03
#define SCSI_WORM                    0x04
#define SCSI_CDROM                   0x05

    BITFIELD_2( UINT8,
            device_type : 7,
            rmb : 1);
    BITFIELD_3( UINT8,
            ansi : 3,
            ecma : 3,
            iso : 2);
    UINT8    reserved;
    UINT8    length;
    UINT8    param[1];
} SCSI_INQUIRY_DATA, *PSCSI_INQUIRY_DATA;

#define SCSI_INQ_STD_DATA            0xff

/* request sense data */
#define SCSI_SNS_NOSENSE             0x0
#define SCSI_SNS_RECOVERED           0x1
#define SCSI_SNS_NOTREADY            0x2
#define SCSI_SNS_MEDIUM_ERR          0x3
#define SCSI_SNS_HW_ERR              0x4
#define SCSI_SNS_ILLEGAL_REQ         0x5
#define SCSI_SNS_UNIT_ATN            0x6
#define SCSI_SNS_PROTECT             0x7
#define SCSI_SNS_BLANK_CHK           0x8
#define SCSI_SNS_VUQE                0x9
#define SCSI_SNS_COPY_ABRT           0xa
#define SCSI_SNS_ABORTED             0xb
#define SCSI_SNS_EQUAL               0xc
#define SCSI_SNS_VOLUME_OVFL         0xd
#define SCSI_SNS_MISCOMPARE          0xe
#define SCSI_SNS_RESERVED            0xf

typedef struct _SCSI_SENSE_DATA {
    BITFIELD_3( UINT8,
        error_code : 4,
        error_class : 3,
        addr_valid : 1);
# define SCSI_SNS_XTENDED_SENSE_DATA 0x7    /* e.g. error_class=7 */
    union {
        struct {
            BITFIELD_2(UINT8,
                       lba_msb : 5,
                       vuqe : 3);
            UINT8    lba;
            UINT8    lba_lsb;
        } non_xtnded;
        struct {
            UINT8    segment_number;
            BITFIELD_5(UINT8,
                       sense_key : 4,
                       res : 1,
                       ili : 1,
                       eom : 1,
                       fm : 1);
            UINT8    info0;
            UINT8    info1;
            UINT8    info2;
            UINT8    info3;
            UINT8    add_len;
            UINT8    cmd_specific[4];
            UINT8    asc;
            UINT8    ascq;
            UINT8    fru_code;
            UINT8    sk_spec[3];
        } xtended;
    } u;
} SCSI_SENSE_DATA, *PSCSI_SENSE_DATA;


/* mode select params */
typedef struct _SCSI_MODE_SELECT_PARAM {
    UINT8    reserved1;
    UINT8    medium_type;
    BITFIELD_3(UINT8,
            speed:4,    /* special for tapes, reserved in SCSI-1 */
            buffer_mode:3,
            reserved2:1);
    UINT8    desc_len;
    struct scsi_mode_parm_blockdesc {
        UINT8    density_code;
        UINT8    nblocks1;
        UINT8    nblocks2;
        UINT8    nblocks3;
        UINT8    reserved;
        UINT8    reclen1;
        UINT8    reclen2;
        UINT8    reclen3;
    } descs[1];        /* VARSIZE */
} SCSI_MODE_SELECT_PARAM, *PSCSI_MODE_SELECT_PARAM;

/* mode sense data (TAPE) */
typedef struct _SCSI_MODE_SENSE_DATA {
    UINT8    data_len;
    UINT8    medium_type;
    BITFIELD_3(UINT8,
            speed : 4,
            buffered_mode : 3,
            wp : 1);
    UINT8    bdesc_len;
    struct {
        UINT8    density_code;
        UINT8    no_blks_msb;
        UINT8    no_blks;
        UINT8    no_blks_lsb;
        UINT8    reserved;
        UINT8    blen_msb;
        UINT8    blen;
        UINT8    blen_lsb;
    } bdesc[1];    /* VARSIZE */
    /* vuqe data might follow */
} SCSI_MODE_SENSE_DATA, *PSCSI_MODE_SENSE_DATA;

/* read capacity data */
typedef struct _SCSI_RCAP_DATA {
    UINT8    lba1;
    UINT8    lba2;
    UINT8    lba3;
    UINT8    lba4;
    UINT8    blen1;
    UINT8    blen2;
    UINT8    blen3;
    UINT8    blen4;
} SCSI_RCAP_DATA, *PSCSI_RCAP_DATA;

/* defect list(s) */
typedef struct _SCSI_DEFECT_DATA {
    UINT8    res1;
    UINT8    res2;
    UINT8    list_len_msb;
    UINT8    list_len_lsb;
    struct {
        UINT8    blockno_msb;
        UINT8    blockno_sb1;
        UINT8    blockno_sb2;
        UINT8    blockno_lsb;
    } defects[1];    /* VARSIZE */
} SCSI_DEFECT_DATA, *PSCSI_DEFECT_DATA;

/* block limits (tape) */
typedef struct _SCSI_BLIMITS_DATA {
    UINT8    res1;
    UINT8    maxlen_msb;
    UINT8    maxlen_sb;
    UINT8    maxlen_lsb;
    UINT8    minlen_msb;
    UINT8    minlen_lsb;
} SCSI_BLIMITS_DATA, *PSCSI_BLIMITS_DATA;

/*
 * Status byte (a-la scsi1)
 */

typedef union _SCSI_STATUS_BYTE {
    struct {
    BITFIELD_4( UINT8,
            scsi_status_vendor_uqe1:1,
            scsi_status_code:4,
            scsi_status_vendor_uqe:2,
            scsi_status_reserved:1);
# define SCSI_ST_GOOD                0x00        /* scsi_status_code values */
# define SCSI_ST_CHECK_CONDITION     0x01
# define SCSI_ST_CONDITION_MET       0x02
# define SCSI_ST_BUSY                0x04
# define SCSI_ST_INT_GOOD            0x08
# define SCSI_ST_INT_MET             0x0a
# define SCSI_ST_RES_CONFLICT        0x0c
                    /* anything else is reserved */
    } st;
    UINT8 bits;
} SCSI_STATUS_BYTE, *PSCSI_STATUS_BYTE;

/*
 * Definitions for SCSI-2
 */

/*
 * Single byte messages
 *
 * originator:    I-nitiator T-arget
 * T-support:    M-andatory O-ptional
 */

#define SCSI_ABORT_TAG               0x0d    /* O I 2 */
#define SCSI_CLEAR_QUEUE             0x0e    /* O I 2 */
#define SCSI_INITIATE_RECOVERY       0x0f    /* O IT2 */
#define SCSI_RELEASE_RECOVERY        0x10    /* O I 2 */
#define SCSI_TERMINATE_IO_PROCESS    0x11    /* O I 2 */

/*
 * Two byte messages
 */
#define SCSI_SIMPLE_QUEUE_TAG        0x20    /* O IT2 */
#define SCSI_HEADOF_QUEUE_TAG        0x21    /* O I 2 */
#define SCSI_ORDERED_QUEUE_TAG       0x22    /* O I 2 */
#define SCSI_IGNORE_WIDE_RESIDUE     0x23    /* O  T2 */
                    /* 0x24..0x2f reserved */

/*
 * Extended messages, codes and formats
 */

#define SCSI_WIDE_XFER_REQUEST       0x03    /* IT 2 */
typedef struct _SCSI_WIDE_XFER {
    UINT8    xtn_msg_tag;            /* const 0x01 */
    UINT8    xtn_msg_len;            /* const 0x02 */
    UINT8    xtn_msg_code;           /* const 0x03 */
    UINT8    xtn_msg_xfer_width;
} SCSI_WIDE_XFER, *PSCSI_WIDE_XFER;

/*
 * Commands, codes and aliases
 */

                /* GROUP 1 */

typedef SCSI_COMMAND_GROUP_1         SCSI_CMD_READ_DEFECT;
#define SCSI_CMD_READ_DEFECT_DATA    0x37    /* O2 disk opti */
# define SCSI_CMD_RDD_LIST_TYPE      0x07
# define SCSI_CMD_RDD_GLIST          0x08
# define SCSI_CMD_RDD_PLIST          0x10

typedef SCSI_COMMAND_GROUP_1         SCSI_CMD_WRITE_BUFFER;
#define SCSI_CMD_WRITE_BUFFER        0x3b    /* O2 all */
# define SCSI_CMD_BUF_MODE           0x07
# define scsi_cmd_buf_id             scs_cmd_lba1
# define scsi_cmd_buf_offset1        scs_cmd_lba2
# define scsi_cmd_buf_offset2        scs_cmd_lba3
# define scsi_cmd_buf_offset3        scs_cmd_lba4
# define scsi_cmd_buf_alloc1         scs_cmd_xxx
# define scsi_cmd_buf_alloc2         scs_cmd_xfer_len_1
# define scsi_cmd_buf_alloc3         scs_cmd_xfer_len_2

typedef SCSI_COMMAND_GROUP_1         SCSI_COMMAND_READ_BUFFER;   
#define SCSI_CMD_READ_BUFFER         0x3c    /* O2 all */

                /* GROUP 2 */

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_CHANGE_DEFINITION;
#define SCSI_CMD_CHANGE_DEFINITION   0x40    /* O2 all */
# define scsi_cmd_chg_save           scsi_cmd_lba1
# define scsi_cmd_chg_definition     scsi_cmd_lba2
# define SCSI_CMD_CHG_CURRENT        0x00
# define SCSI_CMD_CHG_SCSI_1         0x01
# define SCSI_CMD_CHG_CCS            0x02
# define SCSI_CMD_CHG_SCSI_2         0x03

                    /* 0x41 reserved */

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_READ_SUBCHANNEL;       
#define SCSI_CMD_READ_SUBCH          0x42    /* O2 rom */
# define SCSI_CMD_CD_MSF             0x02
# define SCSI_CMD_RS_SUBQ            0x40
# define scsi_cmd_rs_format          scsi_cmd_lba2
# define SCSI_CMD_RS_FMT_SUBQ        0x00
# define SCSI_CMD_RS_FMT_CURPOS      0x01
# define SCSI_CMD_RS_FMT_CATALOG     0x02
# define SCSI_CMD_RS_FMT_ISRC        0x03
# define scsi_cmd_rs_trackno         scsi_cmd_xxx


typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_READ_TOC;         
#define SCSI_CMD_READ_TOC            0x43    /* O2 rom */
# define scsi_cmd_rtoc_startT        scsi_cmd_xxx

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_READ_HEADER;   
#define SCSI_CMD_READ_HEADER         0x44    /* O2 rom */

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_PLAY_AUDIO;   
#define SCSI_CMD_PLAY_AUDIO          0x45    /* O2 rom */

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_PLAY_AUDIO_MSF; 
#define SCSI_CMD_PLAY_AUDIO_MSF      0x47    /* O2 rom */
# define scsi_cmd_pamsf_startM       scsi_cmd_lba2
# define scsi_cmd_pamsf_startS       scsi_cmd_lba3
# define scsi_cmd_pamsf_startF       scsi_cmd_lba4
# define scsi_cmd_pamsf_endM         scsi_cmd_xxx
# define scsi_cmd_pamsf_endS         scsi_cmd_xfer_len_1
# define scsi_cmd_pamsf_endF         scsi_cmd_xfer_len_2

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_PLAY_AUDIO_TI; 
#define SCSI_CMD_PLAY_AUDIO_TI       0x48    /* O2 rom */
# define scsi_cmd_pati_startT        scsi_cmd_lba3
# define scsi_cmd_pati_startI        scsi_cmd_lba4
# define scsi_cmd_pati_endT          scsi_cmd_xfer_len_1
# define scsi_cmd_pati_endI          scsi_cmd_xfer_len_2

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_PLAY_AUDIO_TR; 
#define SCSI_CMD_PLAY_AUDIO_TR       0x49    /* O2 rom */
# define scsi_cmd_patr_startT        scsi_cmd_xxx


typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_PAUSE_RESUME;    
#define SCSI_CMD_PAUSE_RESUME        0x4b    /* O2 rom */
# define SCSI_CMD_PAUSRES_RESUME     0x01
# define scsi_cmd_pausres_res        scsi_cmd_xfer_len_2

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_LOGSEL;    
#define SCSI_CMD_LOG_SELECT          0x4c    /* O2 all */
# define SCSI_CMD_LOG_SP             0x01
# define SCSI_CMD_LOG_PCR            0x02
# define scsi_cmd_log_page_control   scsi_cmd_lba1

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_LOGSENSE;    
#define SCSI_CMD_LOG_SENSE           0x4d    /* O2 all */
# define SCSI_CMD_LOG_PPC            0x02
# define scsi_cmd_log_page_code      scsi_cmd_lba1
# define scsi_cmd_log_param_ptr1     scsi_cmd_lba4
# define scsi_cmd_log_param_ptr2     scsi_cmd_xxx


                    /* 0x4e..0x54 reserved */

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_MODE_SELECT_LONG;    
#define SCSI_CMD_MODE_SELECT_2       0x55    /* Z2 */
# define SCSI_CMD_MSL2_PF            0x10
# define SCSI_CMD_MSL2_SP            0x01

                    /* 0x56..0x59 reserved */

typedef SCSI_COMMAND_GROUP_2         SCSI_COMMAND_MODE_SENSE_LONG;    
#define SCSI_CMD_MODE_SENSE_2        0x5a    /* Z2 */
# define SCSI_CMD_MSS2_DBD           0x08

                    /* 0x5b..0x5f reserved */

                /* GROUP 5 */

typedef SCSI_COMMAND_GROUP_5         SCSI_COMMAND_PLAY_AUDIO_LONG;        
#define SCSI_CMD_PLAY_AUDIO_LONG     0xa5    /* O2 rom */

typedef SCSI_COMMAND_GROUP_5         SCSI_COMMAND_PLAY_AUDIO_TR_LONG;    
#define SCSI_CMD_PLAY_AUDIO_TR_LONG  0xa9    /* O2 rom */
# define scsi_cmd_patrl_startT       scsi_cmd_xxx1


/*
 * Command specific defines
 */
typedef struct _SCSI2_INQUIRY_DATA {
    BITFIELD_2(UINT8,
            periph_type : 5,
            periph_qual : 3);
#define SCSI_SCANNER                 0x06    /* periph_type values */
#define SCSI_MEMORY                  0x07
#define SCSI_J_BOX                   0x08
#define SCSI_COMM                    0x09
#define SCSI_PREPRESS1               0x0a
#define SCSI_PREPRESS2               0x0b

#define SCSI_PERIPH_CONNECTED        0x00    /* periph_qual values */
#define SCSI_PERIPH_DISCONN          0x20
#define SCSI_PERIPH_NOTHERE          0x30

    BITFIELD_2(UINT8,
            device_type : 7,
            rmb : 1);

    BITFIELD_3( UINT8,
            ansi : 3,
            ecma : 3,
            iso : 2);

    BITFIELD_4( UINT8,
            response_fmt : 4,
            res1 : 2,
            trmIOP : 1,
            aenc : 1);
    UINT8    length;
    UINT8    res2;
    UINT8    res3;

    BITFIELD_8(UINT8,
            SftRe : 1,
            CmdQue : 1,
            res4 : 1,
            Linked : 1,
            Sync : 1,
            Wbus16 : 1,
            Wbus32 : 1,
            RelAdr : 1);

    UINT8    vendor_id[8];
    UINT8    product_id[16];
    UINT8    product_rev[4];
    /* Do *not* define vendor-unique data fields.
     */
#if 0
    UINT8    vendor_uqe[20];
    UINT8    reserved[40];
    UINT8    vendor_uqe1[1];    /* varsize */
#endif
} SCSI2_INQUIRY_DATA, *PSCSI2_INQUIRY_DATA;

#define SCSI_INQ_SUPP_PAGES          0x00
#define SCSI_INQ_A_INFO              0x01    /* 0x01..0x1f, really */
#define SCSI_INQ_SERIALNO            0x80
#define SCSI_INQ_IMPL_OPDEF          0x81
#define SCSI_INQ_A_IMPL_OPDEF        0x82

/* mode_select */
typedef struct _SCSI2_MODE_PARAM {
    UINT8    data_len;
    UINT8    medium_type;
    UINT8    device_specific;
    UINT8    desc_len;
    /* block descriptors are optional, same struct as scsi1 */
    /* page info then follows, see individual pages */
} SCSI2_MODE_PARAM, *PSCSI2_MODE_PARAM;

/*
 * CDROM definitions
 */
typedef union _CDROM_ADDRESS {
    struct {
        UINT8    xxx;
        UINT8    minute;
        UINT8    second;
        UINT8    frame;
    } msf;
    struct {
        UINT8    lba1;
        UINT8    lba2;
        UINT8    lba3;
        UINT8    lba4;
    } lba;
} CDROM_ADDRESS, *PCDROM_ADDRESS;

typedef struct _CDROM_TOC {
    UINT8    len1;        /* MSB */
    UINT8    len2;        /* LSB */
    UINT8    first_track;
    UINT8    last_track;
    struct cdrom_toc_desc {

        UINT8    xxx;

        BITFIELD_2(UINT8,
            control : 4,
            adr : 4);

        UINT8    trackno;
        UINT8    xxx1;
        CDROM_ADDRESS    absolute_address;
    } descs[1];            /* varsize */
} CDROM_TOC, *PCDROM_TOC;

typedef struct _CDROM_CHANNEL_DATA {
    UINT8    xxx;

    UINT8    audio_status;
#define SCSI_CDST_INVALID            0x00
#define SCSI_CDST_PLAYING            0x11
#define SCSI_CDST_PAUSED             0x12
#define SCSI_CDST_COMPLETED          0x13
#define SCSI_CDST_ERROR              0x14
#define SCSI_CDST_NO_STATUS          0x15

    UINT8    len1;
    UINT8    len2;
    struct cdrom_chanQ {
        UINT8    format;
        BITFIELD_2(UINT8,
            control : 4,
            adr : 4);
        UINT8    trackno;
        UINT8    indexno;
        CDROM_ADDRESS    absolute_address;
        CDROM_ADDRESS    relative_address;
        BITFIELD_2(UINT8,
            xxx : 7,
            mcv : 1);
        UINT8    catalog[15];
        BITFIELD_2(UINT8,
            xxx1 : 7,
            tcv  : 1);
        UINT8    isrc[15];
    } subQ;
} CDROM_CHANNEL_DATA, *PCDROM_CHANNEL_DATA;

/* subsets */
typedef struct _CDROM_CHANNEL_CURPOS {
    UINT8    xxx;
    UINT8    audio_status;
    UINT8    len1;
    UINT8    len2;
    struct {
        UINT8    format;
        BITFIELD_2(UINT8,
            control : 4,
            adr : 4);
        UINT8    trackno;
        UINT8    indexno;
        CDROM_ADDRESS    absolute_address;
        CDROM_ADDRESS    relative_address;
    } subQ;
} CDROM_CHANNEL_CURPOS, *PCDROM_CHANNEL_CURPOS;

typedef struct _CDROM_CHANNEL_CATALOG {
    UINT8    xxx;
    UINT8    audio_status;
    UINT8    len1;
    UINT8    len2;
    struct {
        UINT8    format;
        UINT8    xxx1[3];
        BITFIELD_2(UINT8,
            xxx : 7,
            mcv : 1);
        UINT8    catalog[15];
    } subQ;
} CDROM_CHANNEL_CATALOG, *PCDROM_CHANNEL_CATALOG;

typedef struct _CDROM_CHANNEL_ISRC {
    UINT8    xxx;
    UINT8    audio_status;
    UINT8    len1;
    UINT8    len2;
    struct {
        UINT8    format;
        BITFIELD_2(UINT8,
            control : 4,
            adr : 4);
        UINT8    trackno;
        UINT8    xxx0;
        BITFIELD_2(UINT8,
            xxx1 : 7,
            tcv  : 1);
        UINT8    isrc[15];
    } subQ;
} CDROM_CHANNEL_ISRC, *PCDROM_CHANNEL_ISRC;

/* Audio page */
typedef struct _CDROM_AUDIO_PAGE {
    SCSI_MODE_SENSE_DATA    h;    /* includes bdescs */
    UINT8    page_code;
#define SCSI_CD_AUDIO_PAGE           0x0e
    UINT8    page_len;
    BITFIELD_4(UINT8,
        xxx1 : 1,
        sotc : 1,
        imm  : 1,
        xxx2 : 5);
    UINT8    xxx3[2];
    BITFIELD_3(UINT8,
        fmt : 4,
        xxx4 : 3,
        aprv : 1);
    UINT8    bps1;
    UINT8    bps2;
    BITFIELD_2(UINT8,
        sel0 : 4,
        xxx5 : 4);
    UINT8    vol0;
    BITFIELD_2(UINT8,
        sel1 : 4,
        xxx6 : 4);
    UINT8    vol1;
    BITFIELD_2(UINT8,
        sel2 : 4,
        xxx7 : 4);
    UINT8    vol2;
    BITFIELD_2(UINT8,
        sel3 : 4,
        xxx8 : 4);
    UINT8    vol3;
} CDROM_AUDIO_PAGE, *PCDROM_AUDIO_PAGE;

/*
 * Status byte (a-la scsi2)
 */

typedef union _SCSI2_STATUS_BYTE {
    struct {
    BITFIELD_3( UINT8,
            scsi_status_reserved1:1,
            scsi_status_code:5,
            scsi_status_reserved2:2);
                            /* more scsi_status_code values */
                    /* 00..0c as in SCSI-I */
# define SCSI_ST2_CMD_TERMINATED     0x11    /* 2 */
# define SCSI_ST2_QUEUE_FULL         0x14    /* 2 */
                    /* anything else is reserved */
    } st;
    UINT8 bits;
} SCSI2_STATUS_BYTE, *PSCSI2_STATUS_BYTE;

