 /* Copyright (c) Microsoft Corporation. All rights reserved. */
 /* --------------------------------------------------------------------------
  *
  * Module:
  *
  *    scsilib.c
  *
  * Purpose:
  *
  *    Utilities to build SCSI command blocks
  *
  * References:
  *    "Small Computer Systems Interface (SCSI)", ANSI Draft
  *     X3T9.2/82-2 - Rev 17B December 1985   (SCSI-1)
  *
  *    "Small Computer System Interface - 2 (SCSI-II)", ANSI Draft
  *     X3T9.2/86-109 -  Rev 10C March 1990
  *
  * Author:
  *    A. Forin (sandrof)
  *
  * --------------------------------------------------------------------------
  */

 #include <mmlite.h>
 #include <diagnostics.h>
 #include <mmhal.h>
 #include "scsi.h"
 #include "cdb.h"
 #include "scsilib.h"

 /* BUGBUG Split in multiple files to help non-MSFT linkers
  */

 /* Command for all devices
  */
 void ScsiComposeInquiry(
     PCDB  Cdb,
     UINT8 CodePage,
     UINT8 DataLength
    )
{
    SCSI_COMMAND_INQUIRY      *Command;

    Command = (SCSI_COMMAND_INQUIRY*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

    Command->scsi_cmd_code = SCSI_CMD_INQUIRY;
    Command->scsi_cmd_lun_and_lba1 = 0;
    Command->scsi_cmd_lba3      = 0;
    Command->scsi_cmd_xfer_len  = DataLength;
    Command->scsi_cmd_ctrl_byte = 0;    /* not linked */

    if (CodePage != SCSI_INQ_STD_DATA) {
        Command->scsi_cmd_lun_and_lba1 |= SCSI_CMD_INQ_EVPD;
        Command->scsi_cmd_page_code     = CodePage;
    } else
        Command->scsi_cmd_page_code = 0;
}

/*
 * REQUESTE SENSE (Mandatory, All)
 */
void ScsiComposeRequestSense(
    PCDB Cdb,
    UINT8 DataLength
    )
{
    SCSI_COMMAND_REQUEST_SENSE *Command;

    Command = (SCSI_COMMAND_REQUEST_SENSE*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

    Command->scsi_cmd_code = SCSI_CMD_REQUEST_SENSE;
    Command->scsi_cmd_lun_and_lba1 = 0;
    Command->scsi_cmd_lba2              = 0;
    Command->scsi_cmd_lba3              = 0;
    Command->scsi_cmd_allocation_length = DataLength;
    Command->scsi_cmd_ctrl_byte         = 0;    /* not linked */
}

void ScsiComposeModeSense(
    PCDB Cdb,
	UINT8 Page,
    UINT8 Length
    )
{
	SCSI_COMMAND_MODE_SENSE	*Command;

    Command = (SCSI_COMMAND_MODE_SENSE*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

	Command->scsi_cmd_code = SCSI_CMD_MODE_SENSE;
	Command->scsi_cmd_lun_and_lba1 = 0;
	Command->scsi_cmd_ms_pagecode  = Page;
	Command->scsi_cmd_lba3         = 0;
	Command->scsi_cmd_xfer_len     = Length;
	Command->scsi_cmd_ctrl_byte    = 0;	/* not linked */
}

/*
 * START STOP UNIT
 */
void ScsiComposeStartUnit(
    PCDB  Cdb,
    UINT8 StartFlags
    )
{
    SCSI_COMMAND_START *Command;

    Command = (SCSI_COMMAND_START*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

    Command->scsi_cmd_code = SCSI_CMD_START_STOP_UNIT;
    Command->scsi_cmd_lun_and_lba1 = SCSI_CMD_SS_IMMED;/* 0 won't work ? */
    Command->scsi_cmd_lba2      = 0;
    Command->scsi_cmd_lba3      = 0;
    Command->scsi_cmd_ss_flags  = StartFlags;
    Command->scsi_cmd_ctrl_byte = 0;    /* not linked */
}

/*
 * TEST UNIT READY
 */
void ScsiComposeTestUnitReady(
    PCDB  Cdb
    )
{
    SCSI_COMMAND_TEST_UNIT_READY *Command;

    Command = (SCSI_COMMAND_TEST_UNIT_READY*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

    Command->scsi_cmd_code = SCSI_CMD_TEST_UNIT_READY;
    Command->scsi_cmd_lun_and_lba1 = 0;
    Command->scsi_cmd_lba2      = 0;
    Command->scsi_cmd_lba3      = 0;
    Command->scsi_cmd_ss_flags  = 0;
    Command->scsi_cmd_ctrl_byte = 0;    /* not linked */
}

/*
 * SCSI commands (partially) specific to disks
 */
void ScsiComposeRead(
    PCDB   Cdb,
    UINT32 SectorNumber,
    UINT8  NumberOfBlocks
    )
{
    SCSI_COMMAND_READ *Command;

    Command = (SCSI_COMMAND_READ*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

    Command->scsi_cmd_code = SCSI_CMD_READ;
    Command->scsi_cmd_lun_and_lba1 = (UINT8)((SectorNumber>>16)&SCSI_LBA_MASK);
    Command->scsi_cmd_lba2         = (UINT8)((SectorNumber>> 8)&0xff);
    Command->scsi_cmd_lba3         = (UINT8)((SectorNumber    )&0xff);
    Command->scsi_cmd_xfer_len     = NumberOfBlocks;
    Command->scsi_cmd_ctrl_byte    = 0;    /* not linked */
}

void ScsiComposeWrite(
    PCDB   Cdb,
    UINT32 SectorNumber,
    UINT8  NumberOfBlocks
    )
{
    SCSI_COMMAND_WRITE *Command;

    Command = (SCSI_COMMAND_WRITE*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

    Command->scsi_cmd_code = SCSI_CMD_WRITE;
    Command->scsi_cmd_lun_and_lba1 = (UINT8)((SectorNumber>>16)&SCSI_LBA_MASK);
    Command->scsi_cmd_lba2         = (UINT8)((SectorNumber>> 8)&0xff);
    Command->scsi_cmd_lba3         = (UINT8)((SectorNumber    )&0xff);
    Command->scsi_cmd_xfer_len     = NumberOfBlocks;
    Command->scsi_cmd_ctrl_byte    = 0;    /* not linked */
}

void ScsiComposeReadCapacity(
    PCDB   Cdb,
    UINT32 LogicalBlockNumber
    )
{
    SCSI_COMMAND_READ_CAPACITY    *Command;

    Command = (SCSI_COMMAND_READ_CAPACITY*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

    memset(Command, 0, sizeof(*Command));
    Command->scsi_cmd_code = SCSI_CMD_READ_CAPACITY;
    /* all zeroes, unless... */
    if (LogicalBlockNumber) {
        Command->scsi_cmd_rcap_flags = SCSI_CMD_RCAP_PMI;
        Command->scsi_cmd_lba1 = (UINT8)((LogicalBlockNumber>>24));
        Command->scsi_cmd_lba2 = (UINT8)((LogicalBlockNumber>>16)&0xff);
        Command->scsi_cmd_lba3 = (UINT8)((LogicalBlockNumber>> 8)&0xff);
        Command->scsi_cmd_lba4 = (UINT8)((LogicalBlockNumber    )&0xff);
    }
}

void ScsiComposeMediumRemoval(
    PCDB Cdb,
    BOOL AllowRemoval
    )
{
    SCSI_COMMAND_MEDIUM_REMOVAL *Command;

    Command = (SCSI_COMMAND_MEDIUM_REMOVAL*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

    Command->scsi_cmd_code = SCSI_CMD_PREVENT_ALLOW_REMOVAL;
    Command->scsi_cmd_lun_and_lba1 = 0;
    Command->scsi_cmd_lba2         = 0;
    Command->scsi_cmd_lba3         = 0;
    Command->scsi_cmd_pa_prevent   = AllowRemoval ? 0 : 1;
    Command->scsi_cmd_ctrl_byte    = 0;    /* not linked */
}

/* Group 1 Commands */

void ScsiComposeLongRead(
    PCDB   Cdb,
    UINT32 SectorNumber,
    UINT16 NumberOfBlocks
    )
{
    SCSI_COMMAND_LONG_READ *Command;

    Command = (SCSI_COMMAND_LONG_READ*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

    Command->scsi_cmd_code = SCSI_CMD_LONG_READ;
    Command->scsi_cmd_lun_and_relbit = 0;
    Command->scsi_cmd_lba1         = (UINT8)(SectorNumber >> 24);
    Command->scsi_cmd_lba2         = (UINT8)(SectorNumber >> 16);
    Command->scsi_cmd_lba3         = (UINT8)(SectorNumber >>  8);
    Command->scsi_cmd_lba4         = (UINT8)(SectorNumber);
    Command->scsi_cmd_xxx          = 0;
    Command->scsi_cmd_xfer_len_1   = (UINT8)(NumberOfBlocks >> 8);
    Command->scsi_cmd_xfer_len_2   = (UINT8)(NumberOfBlocks);
    Command->scsi_cmd_ctrl_byte    = 0;    /* not linked */
}

void ScsiComposeLongWrite(
    PCDB   Cdb,
    UINT32 SectorNumber,
    UINT16 NumberOfBlocks
    )
{
    SCSI_COMMAND_LONG_WRITE *Command;

    Command = (SCSI_COMMAND_LONG_WRITE*)(Cdb->Cb);
    Cdb->CbLength = sizeof(*Command);

    Command->scsi_cmd_code = SCSI_CMD_LONG_WRITE;
    Command->scsi_cmd_lun_and_relbit = 0;
    Command->scsi_cmd_lba1         = (UINT8)(SectorNumber >> 24);
    Command->scsi_cmd_lba2         = (UINT8)(SectorNumber >> 16);
    Command->scsi_cmd_lba3         = (UINT8)(SectorNumber >>  8);
    Command->scsi_cmd_lba4         = (UINT8)(SectorNumber);
    Command->scsi_cmd_xxx          = 0;
    Command->scsi_cmd_xfer_len_1   = (UINT8)(NumberOfBlocks >> 8);
    Command->scsi_cmd_xfer_len_2   = (UINT8)(NumberOfBlocks);
    Command->scsi_cmd_ctrl_byte    = 0;    /* not linked */
}


/* Utilities to look at data
 */
typedef _TCHAR *MYSTRING;

void ScsiDumpInquiryData(
    SCSI2_INQUIRY_DATA *Data,
    BOOL Verbose,
    UINT8 *DeviceName
    )
{
    static MYSTRING periph_names[10] = {
        _T("disk"), _T("tape"), _T("printer"), _T("processor"), _T("WORM-disk"),
        _T("CD-ROM"), _T("scanner"), _T("memory"), _T("jukebox"), _T("communication")
    };
    static MYSTRING periph_state[4] = {
        _T("online"), _T("offline"), _T("?"), _T("absent")
    };

    int i, j = 0;
    UINT8 *devname = DeviceName;

    /* Drop all (mandatory) spaces
     */
    for (i = 0; i < 8; i++)
        if (Data->vendor_id[i] != ' ')
            devname[j++] = Data->vendor_id[i];
    if (j) devname[j++] = ' ';
    for (i = 0; i < 16; i++)
        if (Data->product_id[i] != ' ')
            devname[j++] = Data->product_id[i];
    if (j) devname[j++] = ' ';
    for (i = 0; i < 4; i++)
        if (Data->product_rev[i] != ' ')
            devname[j++] = Data->product_rev[i];
    devname[j] = 0;

    if (!Verbose)
        return;

    /* Say some
     */
    _tprintf(_T("(%s) is a %s%s %s (%s %x)"), devname,
           (Data->rmb) ? _T("") : _T("non-"), _T("removable SCSI"),
           (Data->periph_type > 10) ?
           _T("?device?") : periph_names[Data->periph_type],
           periph_state[Data->periph_qual & 0x3],
           Data->device_type);
    _tprintf(_T("\n\t%s%s%s"),
           Data->iso ? _T("ISO-compliant, ") : _T(""),
           Data->ecma ? _T("ECMA-compliant, ") : _T(""),
           Data->ansi ? _T("ANSI-compliant, ") : _T(""));
    if (Data->ansi)
        _tprintf(_T("%s%d, "), _T("SCSI-"), Data->ansi);
    if (Data->response_fmt == 2)
        _tprintf(_T("%s%s%s%s%s%s%s%s%s%s%s\n"), _T("Supports: "),
               Data->aenc ? _T("AENC, ") : _T(""),
               Data->trmIOP ? _T("TrmIOP, ") : _T(""),
               Data->RelAdr ? _T("RelAdr, ") : _T(""),
               Data->Wbus32 ? _T("32 bit xfers, ") : _T(""),
               Data->Wbus16 ? _T("16 bis xfers, ") : _T(""),
               Data->Sync ? _T("Sync xfers, ") : _T(""),
               Data->Linked ? _T("Linked cmds, ") : _T(""),
               Data->CmdQue ? _T("Tagged cmd queues, ") : _T(""),
               Data->SftRe ? _T("Soft") : _T("Hard"), _T("-RESET"));
}

void ScsiDumpCapacityData(
    SCSI_RCAP_DATA *Data
    )
{
    UINT32 DiskSize, BlockLength;

    DiskSize = (Data->lba1<<24) |
               (Data->lba2<<16) |
               (Data->lba3<< 8) |
                Data->lba4;
    BlockLength = (Data->blen1<<24) |
                  (Data->blen2<<16) |
                  (Data->blen3<<8 ) |
                   Data->blen4;
    _tprintf(_T("Capacity: %u blocks of %u bytes\n"),
             DiskSize+1, BlockLength);
}

static MYSTRING SenseMessage[SCSI_SNS_RESERVED+1] = {
    _T("No Sense Data"),/* shouldn't happen */
    _T("Recovered"),
    _T("Unit not ready"),
    _T("Medium"),
    _T("Hardware failure"),
    _T("Illegal request"),
    _T("Unit Attention Condition"),
    _T("Protection"),
    _T("Blank Check"),
    _T("Vendor Unique"),
    _T("Copy Operation Aborted"),
    _T("Aborted Command"),
    _T("Equal Comparison"),
    _T("Volume Overflow"),
    _T("Miscompare"),
    _T("Reserved")
};

void ScsiDumpSenseData(
    SCSI_SENSE_DATA *Data,
    UINT SenseDataLength
    )
{
    /* Quick silencing check */
    if ((SenseDataLength >= 14) &&
        (Data->error_code == 0) &&
        (Data->u.xtended.sense_key == 0) &&
        (Data->u.xtended.asc == 0) &&
        (Data->u.xtended.ascq == 0))
        return;

    _tprintf(_T("Sense data: %s%s, segment %d"), 
             SenseMessage[Data->u.xtended.sense_key], _T(" error"),
             Data->u.xtended.segment_number);
#if 0
    {
        UINT i;
        UINT8 *p = (UINT8*)Data;
        for (i = 0; i < SenseDataLength; i++) {
            if ((i & 31) == 0)
                printf("\n[%2d]",i);
            printf(" %2x",p[i]);
        }
        printf("\n");
    }
#endif

    if (Data->u.xtended.ili) 
        _tprintf(_T(", IncorrectLengthIndicator"));
    if (Data->u.xtended.eom)
        _tprintf(_T(", EndOfMedium"));
    if (Data->u.xtended.fm) 
        _tprintf(_T(", FileMark"));

    if (Data->addr_valid) {
        UINT32 AdditionalInformation;
        AdditionalInformation = (Data->u.xtended.info0 << 24) |
                                (Data->u.xtended.info1 << 16) |
                                (Data->u.xtended.info2 <<  8) |
                                 Data->u.xtended.info3;
        _tprintf(_T(", AdditionalInformation x%x"), AdditionalInformation);
    }

    if ((SenseDataLength >= 14) && (Data->u.xtended.add_len >= 6))
        ScsiDumpAdditionalSenseData(Data->u.xtended.asc,
                                    Data->u.xtended.ascq);
    _tprintf(_T("\n"));
}

/*
 * Table of the official SCSI-2 error messages
 * Last update:
 *    X3T9.2/86-109, Revision 10c, March 9, 1990
 */
static struct ADDITIONAL_SENSE_KEYS_MAP {
    UINT8    byte12;
    UINT8    byte13;
    MYSTRING Meaning;
} AdditionalSenseMessages[] = {
    { 0x00, 0x00, _T("No additional sense information") },
    { 0x00, 0x01, _T("Filemark detected") },
    { 0x00, 0x02, _T("End-of-partition/medium detected") },
    { 0x00, 0x03, _T("Setmark detected") },
    { 0x00, 0x04, _T("Beginning of partition/medium detected") },
    { 0x00, 0x05, _T("End-of-data detected") },
    { 0x00, 0x06, _T("I/O process terminated") },
    { 0x00, 0x11, _T("Audio play operation in progress") },
    { 0x00, 0x12, _T("Audio play operation paused") },
    { 0x00, 0x13, _T("Audio play operation successfully completed") },
    { 0x00, 0x14, _T("Audio play operation stopped due to error") },
    { 0x00, 0x15, _T("No current audio status to return") },
    { 0x01, 0x00, _T("No index/sector signal") },
    { 0x02, 0x00, _T("No seek complete") },
    { 0x03, 0x00, _T("Peripheral device write fault") },
    { 0x03, 0x01, _T("No write current") },
    { 0x03, 0x02, _T("Excessive write errors") },
    { 0x04, 0x00, _T("Logical unit not ready, cause not reportable") },
    { 0x04, 0x01, _T("Logical unit is in process of becoming ready") },
    { 0x04, 0x02, _T("Logical unit not ready, initializing command required") },
    { 0x04, 0x03, _T("Logical unit not ready, manual intervention required") },
    { 0x04, 0x04, _T("Logical unit not ready, format in progress") },
    { 0x05, 0x00, _T("Logical unit does not respond to selection") },
    { 0x06, 0x00, _T("No reference position found") },
    { 0x07, 0x00, _T("Multiple peripheral devices selected") },
    { 0x08, 0x00, _T("Logical unit communication failure") },
    { 0x08, 0x01, _T("Logical unit communication time-out") },
    { 0x08, 0x02, _T("Logical unit communication parity error") },
    { 0x09, 0x00, _T("Track following error") },
    { 0x09, 0x01, _T("Tracking servo failure") },
    { 0x09, 0x02, _T("Focus servo failure") },
    { 0x09, 0x03, _T("Spindle servo failure") },
    { 0x0a, 0x00, _T("Error log overflow") },
    { 0x0c, 0x00, _T("Write error") },
    { 0x0c, 0x01, _T("Write error recovered with auto-reallocation") },
    { 0x0c, 0x02, _T("Write error - auto-reallocation failed") },
    { 0x10, 0x00, _T("Id CRC or ECC error") },
    { 0x10, 0x04, _T("Recovered data with LEC") },
    { 0x11, 0x00, _T("Unrecovered read error") },
    { 0x11, 0x01, _T("Read retries exhausted") },
    { 0x11, 0x02, _T("Error too long to correct") },
    { 0x11, 0x03, _T("Multiple read errors") },
    { 0x11, 0x04, _T("Unrecovered read error - auto-reallocate failed") },
    { 0x11, 0x05, _T("L-EC uncorrectable error") },
    { 0x11, 0x06, _T("CIRC unrecovered error") },
    { 0x11, 0x07, _T("Data resynchronization error") },
    { 0x11, 0x08, _T("Incomplete block read") },
    { 0x11, 0x09, _T("No gap found") },
    { 0x11, 0x0a, _T("Miscorrected error") },
    { 0x11, 0x0b, _T("Unrecovered read error - recommend reassignment") },
    { 0x11, 0x0c, _T("Unrecovered read error - recommend rewrite the data") },
    { 0x12, 0x00, _T("Address mark not found for id field") },
    { 0x13, 0x00, _T("Address mark not found for data field") },
    { 0x14, 0x00, _T("Recorded entity not found") },
    { 0x14, 0x01, _T("Record not found") },
    { 0x14, 0x02, _T("Filemark or setmark not found") },
    { 0x14, 0x03, _T("End-of-data not found") },
    { 0x14, 0x04, _T("Block sequence error") },
    { 0x15, 0x00, _T("Random positioning error") },
    { 0x15, 0x01, _T("Mechanical positioning error") },
    { 0x15, 0x02, _T("Positioning error detected by read of medium") },
    { 0x16, 0x00, _T("Data synchronization mark error") },
    { 0x17, 0x00, _T("Recovered data with no error correction applied") },
    { 0x17, 0x01, _T("Recovered data with retries") },
    { 0x17, 0x02, _T("Recovered data with positive head offset") },
    { 0x17, 0x03, _T("Recovered data with negative head offset") },
    { 0x17, 0x04, _T("Recovered data with retries and/or CIRC applied") },
    { 0x17, 0x05, _T("Recovered data using previous sector id") },
    { 0x17, 0x06, _T("Recovered data without ECC - data auto-reallocated") },
    { 0x17, 0x07, _T("Recovered data without ECC - recommend reassignment") },
    { 0x18, 0x00, _T("Recovered data with error correction applied") },
    { 0x18, 0x01, _T("Recovered data with error correction and retries applied") },
    { 0x18, 0x02, _T("Recovered data - data auto-reallocated") },
    { 0x18, 0x03, _T("Recovered data with CIRC") },
    { 0x18, 0x05, _T("Recovered data - recommended reassignment") },
    { 0x19, 0x00, _T("Defect list error") },
    { 0x19, 0x01, _T("Defect list not available") },
    { 0x19, 0x02, _T("Defect list error in primary list") },
    { 0x19, 0x03, _T("Defect list error in grown list") },
    { 0x1a, 0x00, _T("Parameter list length error") },
    { 0x1b, 0x00, _T("Synchronous data transfer error") },
    { 0x1c, 0x00, _T("Defect list not found") },
    { 0x1c, 0x01, _T("Primary defect list not found") },
    { 0x1c, 0x02, _T("Grown defect list not found") },
    { 0x1d, 0x00, _T("Miscompare during verify operation") },
    { 0x1e, 0x00, _T("Recovered id with ECC correction") },
    { 0x20, 0x00, _T("Invalid command operation code") },
    { 0x21, 0x00, _T("Logical block address out of range") },
    { 0x21, 0x01, _T("Invalid element address") },
    { 0x22, 0x00, _T("Illegal function") },
    { 0x24, 0x00, _T("Invalid field in CDB") },
    { 0x24, 0x02, _T("Log parameters changed") },
    { 0x25, 0x00, _T("Logical unit not supported") },
    { 0x26, 0x00, _T("Invalid field in parameter list") },
    { 0x26, 0x01, _T("Parameter not supported") },
    { 0x26, 0x02, _T("Parameter value invalid") },
    { 0x26, 0x03, _T("Threshold parameters not supported") },
    { 0x27, 0x00, _T("Write protected") },
    { 0x28, 0x00, _T("Not ready to ready transition (medium may have changed)") },
    { 0x28, 0x01, _T("Import or export element accessed") },
    { 0x29, 0x00, _T("Power on, reset, or bus device reset occurred") },
    { 0x2a, 0x00, _T("Parameters changed") },
    { 0x2a, 0x01, _T("Mode parameters changed") },
    { 0x2b, 0x00, _T("Copy cannot execute since host cannot disconnect") },
    { 0x2c, 0x00, _T("Command sequence error") },
    { 0x2c, 0x01, _T("Too many windows specified") },
    { 0x2c, 0x02, _T("Invalid combination of windows specified") },
    { 0x2d, 0x00, _T("Overwrite error on update in place") },
    { 0x2f, 0x00, _T("Commands cleared by another initiator") },
    { 0x30, 0x00, _T("Incompatible medium installed") },
    { 0x30, 0x01, _T("Cannot read medium - unknown format") },
    { 0x30, 0x02, _T("Cannot read medium - incompatible format") },
    { 0x30, 0x03, _T("Cleaning cartridge installed") },
    { 0x31, 0x00, _T("Medium format corrupted") },
    { 0x31, 0x01, _T("Format command failed") },
    { 0x32, 0x00, _T("No defect spare location available") },
    { 0x32, 0x01, _T("Defect list update failure") },
    { 0x33, 0x00, _T("Tape length error") },
    { 0x36, 0x00, _T("Ribbon, ink, or toner failure") },
    { 0x37, 0x00, _T("Rounded parameter") },
    { 0x39, 0x00, _T("Saving parameters not supported") },
    { 0x3a, 0x00, _T("Medium not present") },
    { 0x3b, 0x00, _T("Sequential positioning error") },
    { 0x3b, 0x01, _T("Tape position error at beginning of medium") },
    { 0x3b, 0x02, _T("Tape position error at end of medium") },
    { 0x3b, 0x03, _T("Tape or electronic vertical forms unit not ready") },
    { 0x3b, 0x04, _T("Slew failure") },
    { 0x3b, 0x05, _T("Paper jam") },
    { 0x3b, 0x06, _T("Failed to sense top-of-form") },
    { 0x3b, 0x07, _T("Failed to sense bottom-of-form") },
    { 0x3b, 0x08, _T("Reposition error") },
    { 0x3b, 0x09, _T("Read past end of medium") },
    { 0x3b, 0x0a, _T("Read past beginning of medium") },
    { 0x3b, 0x0b, _T("Position past end of medium") },
    { 0x3b, 0x0c, _T("Position past beginning of medium") },
    { 0x3b, 0x0d, _T("Medium destination element full") },
    { 0x3b, 0x0e, _T("Medium source element empty") },
    { 0x3d, 0x00, _T("Invalid bits in identify message") },
    { 0x3e, 0x00, _T("Logical unit has not self-configured yet") },
    { 0x3f, 0x00, _T("Target operating conditions have changed") },
    { 0x3f, 0x01, _T("Microcode has been changed") },
    { 0x3f, 0x02, _T("Changed operating definition") },
    { 0x3f, 0x03, _T("Inquiry data has changed") },
    { 0x40, 0x00, _T("RAM failure") },
    { 0x40, 0xff, _T("Diagnostic failure on component <NN>") },
    { 0x41, 0x00, _T("Data path failure") },
    { 0x42, 0x00, _T("Power on or self-test failure") },
    { 0x43, 0x00, _T("Message error") },
    { 0x44, 0x00, _T("Internal target failure") },
    { 0x45, 0x00, _T("Select or reselect failure") },
    { 0x46, 0x00, _T("Unsuccessful soft reset") },
    { 0x47, 0x00, _T("SCSI parity error") },
    { 0x48, 0x00, _T("Initiator detected message received") },
    { 0x49, 0x00, _T("Invalid message error") },
    { 0x4a, 0x00, _T("Command phase error") },
    { 0x4b, 0x00, _T("Data phase error") },
    { 0x4c, 0x00, _T("Logical unit failed self-configuration") },
    { 0x4e, 0x00, _T("Overlapped commands attempted") },
    { 0x50, 0x00, _T("Write append error") },
    { 0x50, 0x01, _T("Write append position error") },
    { 0x50, 0x02, _T("Position error related to timing") },
    { 0x51, 0x00, _T("Erase failure") },
    { 0x52, 0x00, _T("Cartridge fault") },
    { 0x53, 0x00, _T("Media load or eject failed") },
    { 0x53, 0x01, _T("Unload tape failure") },
    { 0x53, 0x02, _T("Medium removal prevented") },
    { 0x54, 0x00, _T("SCSI to host system interface failure") },
    { 0x55, 0x00, _T("System resource failure") },
    { 0x57, 0x00, _T("Unable to recover table-of-contents") },
    { 0x58, 0x00, _T("Generation does not exist") },
    { 0x59, 0x00, _T("Updated block read") },
    { 0x5a, 0x00, _T("Operator request or state change input (unspecified)") },
    { 0x5a, 0x01, _T("Operator medium removal request") },
    { 0x5a, 0x02, _T("Operator selected write protect") },
    { 0x5a, 0x03, _T("Operator selected write permit") },
    { 0x5b, 0x00, _T("Log exception") },
    { 0x5b, 0x01, _T("Threshold condition met") },
    { 0x5b, 0x02, _T("Log counter at maximum") },
    { 0x5b, 0x03, _T("Log list codes exhausted") },
    { 0x5c, 0x00, _T("RPL status change") },
    { 0x5c, 0x01, _T("Spindles synchronized") },
    { 0x5c, 0x02, _T("Spindles not synchronized") },
    { 0x60, 0x00, _T("Lamp failure") },
    { 0x61, 0x00, _T("Video acquisition error") },
    { 0x61, 0x01, _T("Unable to acquire video") },
    { 0x61, 0x02, _T("Out of focus") },
    { 0x62, 0x00, _T("Scan head positioning error") },
    { 0x63, 0x00, _T("End of user area encountered on this track") },
    { 0x64, 0x00, _T("Illegal mode for this track") },
    { 0, 0, 0}
};

void ScsiDumpAdditionalSenseData(
     UINT8 Key,
     UINT8 Qualifier
     )
{
    register struct ADDITIONAL_SENSE_KEYS_MAP *Entry;

    for (Entry = AdditionalSenseMessages; Entry->Meaning; Entry++) {
        if (Entry->byte12 != Key) continue;
        if ((Entry->byte12 == 0x40) && Qualifier) {
            _tprintf(_T(", %s, NN=x%x"), Entry->Meaning, Qualifier);
            return;
        }
        if (Entry->byte13 == Qualifier) {
            _tprintf(_T(" %s"), Entry->Meaning);
            return;
        }
    };
    _tprintf(_T(", Unknown additional sense keys: 0x%x 0x%x\n"),
             Key, Qualifier);
}

/* Error handling and messaging
 */
void ScsiPrintError(
    UINT8 Flags,
    UINT8  ErrorCode,
    UINT8  *AdditionalData
    )
{
    MYSTRING Message1, Message2;
    SCSI2_STATUS_BYTE  Status;

    switch (SCSI_ERR_CLASS(Flags)) {

    case SCSI_ERR_STATUS:
        Message1 = _T("Bad status return");
        Status.bits = ErrorCode;

        switch (Status.st.scsi_status_code) {

        case SCSI_ST_GOOD:
        case SCSI_ST_CONDITION_MET:
        case SCSI_ST_INT_GOOD:
        case SCSI_ST_INT_MET:
            return;    /* all is fine */
        case SCSI_ST_CHECK_CONDITION:
            Message2 = _T("Check condition"); break;
        case SCSI_ST_RES_CONFLICT:
            Message2 = _T("Reservation conflict"); break;
        case SCSI_ST_BUSY:
            Message2 = _T("Target busy"); break;
        case SCSI_ST2_QUEUE_FULL:
            Message2 = _T("Queue full"); break;
        case SCSI_ST2_CMD_TERMINATED:
            Message2 = _T("Command terminated"); break;
        default:
            Message2 = _T("Strange"); break;
        }
        break;
    case SCSI_ERR_SENSE:
        Message1 = _T("Sensed a");
        Message2 = SenseMessage[ErrorCode & 0xf];
        break;
    case SCSI_ERR_MSEL:
        Message1 = _T("Mode select broken"); Message2 = _T(""); break;
    default:
        Message1 = _T("Generic"); Message2 = _T("");
    }

    if (SCSI_ERR_GRAVITY(Flags)) {
        _tprintf(_T("\n%s: %s %s %sx%x"), _T("ScsiError"), Message1, Message2,
            _T("error, code "), ErrorCode);

        /* Dump the 10 additional bytes
         */
        if (AdditionalData) {
            int i;
            _tprintf(_T(", additional info:"));
            for (i = 0; i < 10; i++)
                _tprintf(_T(" %x"),AdditionalData[i]);
        }
        _tprintf(_T("\n"));
    }
}

SCODE ScsiIsSenseErrorCorrectable(
     SCSI_SENSE_DATA *SenseData
     )
{
    UINT8 ErrorCode;

    if (SenseData->error_class != SCSI_SNS_XTENDED_SENSE_DATA) {
        DBGME(4,_tprintf(_T("Scsi:Bad sense data, vuqe class x%x code x%x\n"),
                         SenseData->error_class, SenseData->error_code));
        return E_FAIL;
    }

    ErrorCode = SenseData->u.xtended.sense_key;

    switch (ErrorCode) {
    case SCSI_SNS_NOSENSE:
    case SCSI_SNS_EQUAL:
        return S_OK;

    case SCSI_SNS_RECOVERED:
        ScsiPrintError(SCSI_ERR(SCSI_ERR_BAD,SCSI_ERR_SENSE),
                       ErrorCode, SenseData->u.xtended.cmd_specific);
        return S_OK;

    case SCSI_SNS_UNIT_ATN:
        ScsiPrintError(SCSI_ERR(SCSI_ERR_IGNORE,SCSI_ERR_SENSE),
                       ErrorCode, SenseData->u.xtended.cmd_specific);
        return S_OK;

    case SCSI_SNS_NOTREADY:
        return E_RESTART;

    case SCSI_SNS_ILLEGAL_REQ:
        return E_INVALID_PARAMETER;

    default:
        /* e.g.
           case SCSI_SNS_MEDIUM_ERR:
           case SCSI_SNS_HW_ERR:
           case SCSI_SNS_PROTECT:
           case SCSI_SNS_BLANK_CHK:
           case SCSI_SNS_VUQE:
           case SCSI_SNS_COPY_ABRT:
           case SCSI_SNS_ABORTED:
           case SCSI_SNS_VOLUME_OVFL:
           case SCSI_SNS_MISCOMPARE:
           case SCSI_SNS_RESERVED:
        */
        ScsiPrintError(SCSI_ERR(SCSI_ERR_FATAL,SCSI_ERR_SENSE),
                       ErrorCode, SenseData->u.xtended.cmd_specific);
        return E_FAIL;
    }
}

/* Hackywhacy for UFI use. Urgh.
 */
SCODE ScsiQuickTranslateSenseCodes(
    UINT8 SenseKey,
    UINT8 Asc,
    UINT8 Ascq
    )
{
    /* This makes no sense ;-)
     */
    if (SenseKey == SCSI_SNS_NOSENSE)
        switch (Asc) {
        case 0x28:
            return E_MEDIA_CHANGED;
        case 0x3a:
            return E_NO_MEDIA_IN_DRIVE;
        default:
            break;
        }
    return E_INVALID_PARAMETER;
}
