/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    protocol.c
 *
 * Purpose:
 *
 *    Transport protocols for USBSTOR.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <string.h>
#include <stddef.h> /* for offsetof() */
#define _DEBUG 1
int UsbStorTransportDebug = 2;
#define DBGLVL UsbStorTransportDebug
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbmaster.h>

#include "transport.h"

/* Bulk Transport
 */
#include "bulk.h"

/* Utilities
 */
PRIVATE SCODE BulkReset(
    PTRANSPORT this,
    PUSBM_REQUEST_BLOCK Urb)
{
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->EndpointAddress = 0;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_CLASS|USB_RT_IFACE;
    Urb->SetupData[1] = USB_R_STORAGE_RESET;
    Urb->SetupData[2] = 0;
    Urb->SetupData[3] = 0;
    Urb->SetupData[4] = 0; /* Interface */
    Urb->SetupData[5] = 0;
    Urb->SetupData[6] = 0;
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = 0;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,
                                 Urb,
                                 this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbBulk:R SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbBulk:R SubmitUrb failed er=x%x\n"),
                         Urb->FinalStatus));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}

/* Clear the HALT condition on an endpoint
 */
PRIVATE SCODE PipeReset(
    PTRANSPORT this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 Endpoint)
{
    SCODE sc;

    DBGME(0,_tprintf(_T("UsbTransport:PipeReset(%d)\n"), Endpoint));

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->EndpointAddress = 0;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_ENDP;
    Urb->SetupData[1] = USB_R_CLEAR_FEATURE;
    Set2(&Urb->SetupData[2],USB_FS_ENDP_STALL);
    Urb->SetupData[4] = Endpoint;
    Urb->SetupData[5] = 0;
    Urb->SetupData[6] = 0;
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = 0;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,
                                 Urb,
                                 this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbTransport:P SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbTransport:P SubmitUrb failed er=x%x\n"),
                         Urb->FinalStatus));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}

/* Send a tagged command block to the device
 */
PRIVATE SCODE BulkSendCommand(
    PTRANSPORT this,
    PCDB Cdb,
    BOOL WaitForCompletion)
{
    SCODE sc;
    PUSBM_REQUEST_BLOCK Urb = this->ControlUrb;
    PUSB_CBW_struct Cbw;

    /* Use the allocated data array
     */
    Cbw = (PUSB_CBW_struct)Urb->Data;

    /* Fillup the CBW
     */
    Set4(Cbw->Signature,USB_CBW_SIGNATURE);
    Set4(Cbw->Tag, (ADDRESS)Cdb);
    Set4(Cbw->DataTransferLength,(UINT32)Cdb->DataLength);
    Cbw->LUN = 0;
    Cbw->Flags = (Cdb->Flags & CDB_F_IN) ? USB_CBW_F_IN : USB_CBW_F_OUT;
    Cbw->CommandBlockLength = Cdb->CbLength;
#if 0
    memcpy((PTR)Cbw->CommandBlock,Cdb->Cb,Cdb->CbLength);
#else
    {
        int i;
        for (i = 0; i < Cdb->CbLength; i++)
            Cbw->CommandBlock[i] = Cdb->Cb[i];
        for (; i < sizeof(USB_CBW); i++)
            Cbw->CommandBlock[i] = 0;
    }
#endif

    /* Fill the remaider of the URB
     */
    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->EndpointAddress = this->OutEndpoint;
    Urb->Pid = USB_PID_OUT;
    Urb->Flags = URB_FLAG_OUT; /* bulk */
    Urb->SetupLength = 0;
    Urb->ExpectedDataLength = sizeof(USB_CBW);
    Urb->MaximumPacketSize = this->MaximumPacketSizeOut;

    sc = this->Hub->v->SubmitUrb(this->Hub,
                                 Urb,
                                 this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbBulk:S SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    /* That might be all
     */
    if (!WaitForCompletion)
        return S_OK;

    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbBulk:S SubmitUrb failed er=x%x\n"),
                         Urb->FinalStatus));
        /* STALL ? */
        return (Urb->FinalStatus == URB_STATUS_STALL) ? E_RESTART : E_MEDIA_CHANGED;
    }

    return S_OK;
}
    
/* Send data to the device
 * NB: Clobbers the Urb->Data pointer
 */
PRIVATE SCODE BulkSendData(
    PTRANSPORT this,
    PCDB Cdb,
    BOOL WaitForCompletion)
{
    SCODE sc;
    PUSBM_REQUEST_BLOCK Urb = this->DataUrb;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->EndpointAddress = this->OutEndpoint;
    Urb->Pid = USB_PID_OUT;
    Urb->Flags = URB_FLAG_OUT; /* bulk */
    Urb->SetupLength = 0;
    Urb->ExpectedDataLength = Cdb->DataLength;
    Urb->Data = Cdb->DataPointer;
    Urb->MaximumPacketSize = this->MaximumPacketSizeOut;

    sc = this->Hub->v->SubmitUrb(this->Hub,
                                 Urb,
                                 this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbBulk:DO SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    /* That might be all
     */
    if (!WaitForCompletion)
        return S_OK;

    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbBulk:DO SubmitUrb failed er=x%x\n"),
                         Urb->FinalStatus));
        /* STALL ? */
        return (Urb->FinalStatus == URB_STATUS_STALL) ? E_RESTART : E_MEDIA_CHANGED;
    }

    return S_OK;
}
    
/* Read data from the device
 * NB: Clobbers the Urb->Data pointer
 */
PRIVATE SCODE BulkOrCbiReadData(
    PTRANSPORT this,
    PCDB Cdb,
    BOOL WaitForCompletion,
    BOOL SetReturnedLength
    )
{
    PUSBM_REQUEST_BLOCK Urb = this->DataUrb;
    SCODE sc;

    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->EndpointAddress = this->InEndpoint;
    Urb->Pid = USB_PID_IN;
    Urb->Flags = URB_FLAG_IN; /* bulk */
    Urb->SetupLength = 0;
    Urb->ExpectedDataLength = Cdb->DataLength;
    Urb->Data = Cdb->DataPointer;
    Urb->MaximumPacketSize = this->MaximumPacketSizeIn;

    sc = this->Hub->v->SubmitUrb(this->Hub,
                                 Urb,
                                 this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbBulk:DI SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    /* That might be all
     */
    if (!WaitForCompletion)
        return S_OK;

    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbBulk:DI SubmitUrb failed er=x%x\n"),
                         Urb->FinalStatus));
        /* STALL ? */
        return (Urb->FinalStatus == URB_STATUS_STALL) ? E_RESTART : E_MEDIA_CHANGED;
    }

    if (SetReturnedLength)
        Cdb->DataLength = Urb->ActualDataLength;

    return S_OK;
}
    
PRIVATE SCODE BulkReadData(
    PTRANSPORT this,
    PCDB Cdb,
    BOOL WaitForCompletion)
{
    return BulkOrCbiReadData(this,Cdb,WaitForCompletion,FALSE);
}

/* Read the command completion status
 * Returns E_RESTART when reset recovery is necessary
 */
PRIVATE SCODE BulkReadStatus(
    PTRANSPORT this,
    PCDB *pCdb)
{
    SCODE sc;
    PUSBM_REQUEST_BLOCK Urb = this->StatusUrb;
    PUSB_CSW_struct Csw;
    PCDB Cdb;
    int nTries = 0;

    /* Use the allocated data array
     */
    Csw = (PUSB_CSW_struct)Urb->Data;

    /* Fill the remaider of the URB
     */
 Retry:
    nTries++;
    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->EndpointAddress = this->InEndpoint;
    Urb->Pid = USB_PID_IN;
    Urb->Flags = URB_FLAG_IN; /* bulk */
    Urb->SetupLength = 0;
    Urb->ExpectedDataLength = sizeof(USB_CSW);
    Urb->MaximumPacketSize = this->MaximumPacketSizeIn;

    sc = this->Hub->v->SubmitUrb(this->Hub,
                                 Urb,
                                 this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbBulk:ST SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    /* Did we get STALLed ?
     */
    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbBulk:ST SubmitUrb failed er=x%x\n"),
                         Urb->FinalStatus));
        /* Needs a reset recovery?
         */
        if ((Urb->FinalStatus != URB_STATUS_STALL) || (nTries > 1))
            return E_RESTART;

        /* Must clear the Bulk-in pipe
         */
        sc = PipeReset(this,Urb,(UINT8)(this->InEndpoint|USB_EA_IN));
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbBulk:ST PipeReset failed sc=x%x\n"),sc));
            return sc;
        }

        /* Now try again, one more time
         */
        goto Retry;
    }

    /* Check and look at the CSW
     */
    if ((Urb->ActualDataLength != sizeof(USB_CSW)) ||
        (Get4(Csw->Signature) != USB_CSW_SIGNATURE) ||
        (Get4(Csw->Tag) == 0)) {
        DBGME(3,_tprintf(_T("UsbBulk:ST bad%d sig=%x tag=%x\n"),
                         Urb->ActualDataLength,
                         Get4(Csw->Signature),
                         Get4(Csw->Tag)));
        /* Needs reset recovery
         */
        return E_RESTART;
    }

    /* Pick up the tag, adjust xfer info
     */
    Cdb = (PCDB) Get4(Csw->Tag); /* BUGBUG 64bit unsafe! */
    Cdb->DataLength -= Get4(Csw->DataResidue);

    /* Translate the final status code
     */
    Cdb->FinalStatus = (Csw->Status == USB_CSW_S_OK) ? S_OK :
        ((Csw->Status == USB_CSW_S_PHASE_ERROR) ? E_RESTART : E_FAIL);

    /* Return
     */
    *pCdb = Cdb;
    return S_OK;
}
    
/* Perform RESET recovery
 */
PRIVATE SCODE BulkResetRecovery(
    PTRANSPORT this)
{
    SCODE sc;
    PUSBM_REQUEST_BLOCK Urb = this->ControlUrb;

    DBGME(0,_tprintf(_T("UsbBulk:ResetRecovery\n")));

    sc = BulkReset(this,Urb);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbBulk:RR BulkReset failed sc=%x\n"),sc));
        return sc;
    }

    sc = PipeReset(this,Urb,(UINT8)(this->InEndpoint|USB_EA_IN));
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbBulk:RR PipeReset-In failed sc=%x\n"),sc));
        return sc;
    }

    sc = PipeReset(this,Urb,this->OutEndpoint);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbBulk:RR PipeReset-Out failed sc=%x\n"),sc));
        return sc;
    }

    return sc;
}

/* Constructor for the Bulk transport
 */
SCODE CreateBulkTransport(PTRANSPORT this)
{
    SCODE sc;
    static struct ITransportVtbl ITransportVtable = {
        BulkResetRecovery,
        BulkSendCommand,
        BulkSendData,
        BulkReadData,
        BulkReadStatus
    };

    /* Set our vtable
     */
    this->v = &ITransportVtable;

    /* The specs rather clearly state that this is *not* an optional command,
     * but nonetheless we have seen someone failing it.  I suppose they
     * expect us to do an INQUIRY and look at the HARD-RESET option and
     * swallow the fact you cant reset it.
     * Bad news is the device gets into a really messy state afterwards.
     */
#if 0
    /* Reset the device
     */
    DBGME(0,_tprintf(_T("UsbBulk:Resetting..\n")));
    sc = BulkResetRecovery(this);
    DBGME(0,_tprintf(_T("UsbBulk:Reset->%x\n"), sc));
#else
    sc = S_OK;
#endif

    return sc;
}

/* CBI transport
 */
#include "cbi.h"

/* Send a command down the control pipe
 */
PRIVATE SCODE CbiSendCmd(
    PTRANSPORT this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT16 Length,   // BUGBUG its always 12!!
    const UINT8 *Command,
    BOOL WaitForCompletion
    )
{
    SCODE sc;

    assert(Length == 12);

    /* Build the ADSC
     */
    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->EndpointAddress = 0;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_CONTROL | URB_FLAG_OUT;
    Urb->SetupData[0] = USB_RT_CLASS|USB_RT_IFACE;
    Urb->SetupData[1] = 0;
    Urb->SetupData[2] = 0;
    Urb->SetupData[3] = 0;
    Urb->SetupData[4] = 0;
    Urb->SetupData[5] = 0;
    Set2(&Urb->SetupData[6],Length);
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = Length;
    Urb->Data = (UINT8*)Command;
    Urb->MaximumPacketSize = this->MaximumControlPacketSize;

    sc = this->Hub->v->SubmitUrb(this->Hub,
                                 Urb,
                                 this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbCbi:Snd SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    /* That might be all
     */
    if (!WaitForCompletion)
        return S_OK;

    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbCbi:Snd SubmitUrb failed er=x%x\n"),
                         Urb->FinalStatus));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}

/* Send a Command Block Reset
 */
PRIVATE SCODE CbiReset(
    PTRANSPORT this,
    PUSBM_REQUEST_BLOCK Urb)
{
    SCODE sc;
    static const UINT8 Reset[12] =
        {0x1d,0x4,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff};

    DBGME(0,_tprintf(_T("UsbCbi:Reset\n")));

    sc = CbiSendCmd(this,Urb,12,Reset,TRUE);
    return sc;
}

/* Send a tagged command block to the device
 */
PRIVATE SCODE CbiSendCb(
    PTRANSPORT this,
    PCDB Cdb,
    BOOL WaitForCompletion)
{
    SCODE sc;
    PUSBM_REQUEST_BLOCK Urb = this->DataUrb;

    sc = CbiSendCmd(this,Urb,Cdb->CbLength,Cdb->Cb,WaitForCompletion);
    return sc;
}

/* Send data to the device
 * NB: Clobbers the Urb->Data pointer
 */
#define CbiSendData BulkSendData


/* Read data from the device
 * NB: Clobbers the Urb->Data pointer
 */
PRIVATE SCODE CbiReadData(
    PTRANSPORT this,
    PCDB Cdb,
    BOOL WaitForCompletion)
{
    return BulkOrCbiReadData(this,Cdb,WaitForCompletion,TRUE);
}


/* Read the command completion status
 * Returns E_RESTART when reset recovery is necessary
 */
extern SCODE ScsiQuickTranslateSenseCodes(UINT8,UINT8,UINT8);

PRIVATE SCODE CbiReadStatus(
    PTRANSPORT this,
    PCDB *pCdb)
{
    SCODE sc;
    PUSBM_REQUEST_BLOCK Urb = this->StatusUrb;
    PUSB_CBI_STATUS_struct Csw;
    PCDB Cdb;
    int nTries = 0;

    /* Use the allocated data array
     */
    Csw = (PUSB_CBI_STATUS_struct)Urb->Data;

    /* Fill the remaider of the URB
     */
 Retry:
    nTries++;
    Urb->DeviceAddress = this->MyUsbAddress;
    Urb->EndpointAddress = this->InterruptEndpoint;
    Urb->Pid = USB_PID_IN;
    Urb->Flags = URB_FLAG_IN | URB_FLAG_INTERRUPT;
    Urb->SetupLength = 0;
    Urb->Interval = 1;
    Urb->ExpectedDataLength = sizeof(USB_CBI_STATUS);
    Urb->MaximumPacketSize = this->MaximumPacketSizeInt;

    sc = this->Hub->v->SubmitUrb(this->Hub,
                                 Urb,
                                 this->PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbCbi:ST SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    /* Did we get STALLed ?
     */
    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbCbi:ST SubmitUrb failed er=x%x\n"),
                         Urb->FinalStatus));
        /* Needs a reset recovery?
         */
        if ((Urb->FinalStatus != URB_STATUS_STALL) || (nTries > 1))
            return E_RESTART;

#if 0
        /* Must clear the Cbi-in pipe
         */
        sc = PipeReset(this,Urb,(UINT8)(this->InEndpoint|USB_EA_IN));
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("UsbCbi:ST PipeReset failed sc=x%x\n"),sc));
            return sc;
        }
#endif
        /* Now try again, one more time
         */
        goto Retry;
    }

    /* Check and look at the STATUS
     */
    if (Urb->ActualDataLength != sizeof(USB_CBI_STATUS)) {

        DBGME(3,_tprintf(_T("UsbCbi:ST bad l=%d t=%x v=%x\n"),
                         Urb->ActualDataLength,
                         Csw->Type, Csw->Value));
        /* Needs reset recovery
         */
        return E_RESTART;
    }

    DBGME(0,_tprintf(_T("UsbCbi:Status %x %x\n"),
                     Csw->Type, Csw->Value));

    /* Pick up the status
     */
    Cdb = *pCdb;/*Urgh*/

    /* Translate the final status code
     * If this is the UFI speaking...
     */
    if (Csw->Type != USB_CBI_STATUS_T_COMPLETION) {
        /* But they forgot where the sense key is. Sigh.
         */
        Cdb->FinalStatus = ScsiQuickTranslateSenseCodes(0,Csw->Type,Csw->Value);
    } else {

        Cdb->FinalStatus = (Csw->Value == USB_CBI_STATUS_PASS) ? S_OK :
            ((Csw->Value == USB_CBI_STATUS_PHASE_ERROR) ? E_RESTART : E_FAIL);
    }

    /* Return
     */
    return S_OK;
}

/* Perform RESET recovery
 */
PRIVATE SCODE CbiResetRecovery(
    PTRANSPORT this)
{
    SCODE sc;
    PUSBM_REQUEST_BLOCK Urb = this->DataUrb;

    DBGME(2,_tprintf(_T("UsbCbi:ResetRecovery\n")));

    sc = CbiReset(this,Urb);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbCbi:RR CbiReset failed sc=%x\n"),sc));
        return sc;
    }

    sc = PipeReset(this,Urb,(UINT8)(this->InEndpoint|USB_EA_IN));
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbCbi:RR PipeReset-In failed sc=%x\n"),sc));
        return sc;
    }

    sc = PipeReset(this,Urb,this->OutEndpoint);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbCbi:RR PipeReset-Out failed sc=%x\n"),sc));
        return sc;
    }

    return sc;
}


/* Constructor for the CBI transport
 */
SCODE CreateCbiTransport(PTRANSPORT this)
{
    SCODE sc;
    static struct ITransportVtbl ITransportVtable = {
        CbiResetRecovery, CbiSendCb, CbiSendData, CbiReadData, CbiReadStatus
    };

    /* Set out vtable
     */
    this->v = &ITransportVtable;

    /* Reset the device
     */
    DBGME(0,_tprintf(_T("UsbCbi:Resetting..\n")));
    sc = CbiResetRecovery(this);
    DBGME(0,_tprintf(_T("UsbCbi:Reset->%x\n"), sc));

    return sc;
}

/* CB transport NYI
 */

SCODE CreateCbTransport(PTRANSPORT this)
{
    UnusedParameter(this);
    return E_NOT_IMPLEMENTED;
}

