/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbstor.c
 *
 * Purpose:
 *
 *    Device driver for USB Mass Storage Devices.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <string.h>
#include <stddef.h> /* for offsetof() */
#define _DEBUG 1
int UsbStorDebug = 2;
#define DBGLVL UsbStorDebug
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbmaster.h>

#include "transport.h"
#include "class.h"

typedef struct _USBSTOR {
    struct CDevice CDevice;
    CONST_VTBL struct IUsbDriverVtbl *UsbStorDriverv;

    /* The transport and class drivers use most of the dynamic state 
     */
    TRANSPORT  Transport;

    /* Interface to the class driver is an IFile
     */
    PIFILE Class;

    UINT8 TransportType;
    UINT8 Subclass;
    UINT16 DiskNo;
#define MYNAMELENGTH 12 /* bignuf */
#define MakeMyName(_this_,_name_)  _stprintf(_name_,_T("disk%d"),(_this_)->DiskNo)
} USBSTOR, *PUSBSTOR;

/* Convert between interface and instance pointer and vice versa */
#define pUsbStorDevice(_i_) ((PUSBSTOR)((_i_)))
#define iUsbStorDevice(_p_) ((PIDEVICE)((_p_)))
#define pUsbStorDriver(_i_) ((PUSBSTOR)((ADDRESS)(_i_) - offsetof(struct _USBSTOR, UsbStorDriverv)))
#define iUsbStorDriver(_p_) ((PIUSBDRIVER)(&((PUSBSTOR)(_p_))->UsbStorDriverv))

PRIVATE SCODE UsbStorDeviceQueryInterface (
    PIDEVICE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbStorDevQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));

    /* IDevice
     */
    if (UuidCmp(&IID_IUnknown,iid) || UuidCmp(&IID_IFile,iid)
        || UuidCmp(&IID_IDevice,iid))
    {
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (PIUNKNOWN) pThis;
        return S_OK;
    }

    /* IUsbDriver
     */
    if (UuidCmp(&IID_IUsbDriver, iid)) {
        /* Use a single refcount
         */
        AtomicInc(&((PCDEVICE)pThis)->RefCnt);
        *ppObject = (void*) iUsbStorDriver(pThis);
        return S_OK;
    }

    *ppObject = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT UsbStorDeviceAddRef (
    PIDEVICE pThis)
{
    DBGME(0,_tprintf(_T("UsbStorDevAR(%x)\n"),(ADDRESS)pThis));
    return AtomicInc(&pUsbStorDevice(pThis)->CDevice.RefCnt);
}

PRIVATE UINT UsbStorDeviceRelease (
    PIDEVICE pThis)
{
    PUSBSTOR this = pUsbStorDevice(pThis);
    UINT Refs = AtomicDec(&this->CDevice.RefCnt);
    DBGME(0,_tprintf(_T("UsbStorDevRel(%x,%d)\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        if (this->Transport.ControlUrb)
            pHeap->v->Free(pHeap, 0, (void *) this->Transport.ControlUrb);
        if (this->Transport.DataUrb)
            pHeap->v->Free(pHeap, 0, (void *) this->Transport.DataUrb);
        if (this->Transport.StatusUrb)
            pHeap->v->Free(pHeap, 0, (void *) this->Transport.StatusUrb);
        if (this->Transport.InterruptUrb)
            pHeap->v->Free(pHeap, 0, (void *) this->Transport.InterruptUrb);

        pHeap->v->Free(pHeap, 0, (void *) this);
    }
    return Refs;
}

PRIVATE SCODE MCT UsbStorDeviceReadAt ( PIDEVICE pThis, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    PUSBSTOR this = pUsbStorDevice(pThis);

    DBGME(0,_tprintf(_T("UsbStorDevReadAt(%x,%x)\n"),
                     (ADDRESS)pThis,Uint64ToUint32(Position)));
    if (this->Class)
        return this->Class->v->ReadAt( this->Class,
                                       Position,
                                       pBuffer,
                                       Size,
                                       pSizeRead);

    return E_ACCESS_DENIED;
}

PRIVATE SCODE MCT UsbStorDeviceWriteAt ( PIDEVICE pThis, UINT64 Position, const BYTE *pBuffer,
                               UINT Size, PUINT pSizeWritten)
{
    PUSBSTOR this = pUsbStorDevice(pThis);

    DBGME(0,_tprintf(_T("UsbStorDevWriteAt(%x,%x)\n"),
                     (ADDRESS)pThis,Uint64ToUint32(Position)));

    if (this->Class)
        return this->Class->v->WriteAt( this->Class,
                                        Position,
                                        pBuffer,
                                        Size,
                                        pSizeWritten);

    return E_ACCESS_DENIED;
}

PRIVATE SCODE MCT UsbStorDeviceSetSize(  PIDEVICE pThis, UINT64 Size)
{
    PUSBSTOR this = pUsbStorDevice(pThis);

    DBGME(0,_tprintf(_T("UsbStorDevSetSize(%x,%x)\n"),
                     (ADDRESS)pThis,Uint64ToUint32(Size)));
    if (this->Class)
        return this->Class->v->SetSize( this->Class,
                                        Size);
    return E_ACCESS_DENIED;
}

PRIVATE SCODE MCT UsbStorDeviceGetSize(  PIDEVICE pThis, PUINT64 pSize)
{
    PUSBSTOR this = pUsbStorDevice(pThis);

    DBGME(0,_tprintf(_T("UsbStorDevGetSize(%x,%x)\n"),
                     (ADDRESS)pThis,(ADDRESS)pSize));
    if (this->Class)
        return this->Class->v->GetSize( this->Class,
                                        pSize);
    return E_ACCESS_DENIED;
}

PRIVATE SCODE UsbStorDriverQueryInterface (
    PIUSBDRIVER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    DBGME(0,_tprintf(_T("UsbStorDrivQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));
    return UsbStorDeviceQueryInterface(iUsbStorDevice(pUsbStorDriver(pThis)), iid, ppObject);
}

PRIVATE UINT UsbStorDriverAddRef (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbStorDrivAR(%x)\n"),(ADDRESS)pThis));
    return UsbStorDeviceAddRef(iUsbStorDevice(pUsbStorDriver(pThis)));
}

PRIVATE UINT UsbStorDriverRelease (
    PIUSBDRIVER pThis)
{
    DBGME(0,_tprintf(_T("UsbStorDrivRel(%x)\n"),(ADDRESS)pThis));
    return UsbStorDeviceRelease(iUsbStorDevice(pUsbStorDriver(pThis)));
}

/* Utils
 */
PUSBM_REQUEST_BLOCK UrbAlloc(
    PUSBSTOR this,
    UINT DataSize)
{
    PUSBM_REQUEST_BLOCK Urb;

    Urb = malloc(DataSize + sizeof *Urb);
    if (Urb != NULL) {
        memset(Urb,0,sizeof *Urb);
        Urb->Done = &this->CDevice.IntrPoint;
        if (DataSize)
            Urb->Data = (UINT8*)(Urb+1);
    }
    return Urb;
}

PRIVATE void ParseConfigurationDescriptor(
    PUSBSTOR this,
    UINT8 *Descriptor,
    UINT16 Length)
{
    int i;
    UINT8 Dt, Ds, EpIn = 0x81, EpOut = 2, EpInt = 0/*invalid*/;
    UINT16 Mps = 64;
    PUSB_ENDPOINT_struct Ed;
    PUSB_INTERFACE_struct If;

    /* Hunt for the interface and endpoint descriptors
     */
    for (i = 0; i < Length; i += Ds)
    {
        /* Get and check the size
         */
        Ds = Descriptor[i];
        if (Ds == 0) /* sanity */
            break;

        /* Look at the type
         */
        Dt = Descriptor[i+1];

        /* Interface?
         */
        if (Dt == USB_DT_IFACE) {
            If = (PUSB_INTERFACE_struct)(Descriptor + i);

            /* Per specs, the interface we want must be first
             */
            if (If->InterfaceClass != USB_DC_STORAGE)
                break;

            this->Subclass = If->InterfaceSubClass;
            this->TransportType = If->InterfaceProtocol;

            continue;
        }

        /* If not an endpoint skip
         */
        if (Dt != USB_DT_ENDP)
            continue;

        /* We can have at most 3 endpoints, depending on transport.
         */
        Ed = (PUSB_ENDPOINT_struct)(Descriptor + i);
        if (Ed->EndpointAddress & USB_EA_IN) {

            Mps = Ed->MaxPacketSize[0] | (Ed->MaxPacketSize[1] << 8);

            /* The CBI transport wants an IN interrupt endpoint
             * All the others a bulk IN.
             */
            if ((Ed->Attributes & USB_AE_MASK) == USB_AE_INTERRUPT) {
                EpInt = Ed->EndpointAddress & ~USB_EA_IN;
                this->Transport.MaximumPacketSizeInt = Mps;
            } else if ((Ed->Attributes & USB_AE_MASK) == USB_AE_BULK) {
                EpIn = Ed->EndpointAddress & ~USB_EA_IN;
                this->Transport.MaximumPacketSizeIn = Mps;
            } else
                continue; /* say what? */
        } else {
            /* "the" output endpoint. Better be bulk.
             */
            if ((Ed->Attributes & USB_AE_MASK) == USB_AE_BULK) {
                EpOut = Ed->EndpointAddress;
                this->Transport.MaximumPacketSizeOut = Mps;
            }
        }
    }

    DBGME(1,_tprintf(_T("StorEndpoints: In %d (%d) Out %d (%d) Intr %d (%d)\n"),
                     EpIn, this->Transport.MaximumPacketSizeIn,
                     EpOut, this->Transport.MaximumPacketSizeOut,
                     EpInt, this->Transport.MaximumPacketSizeInt));

    this->Transport.InEndpoint = EpIn;
    this->Transport.OutEndpoint = EpOut;
    this->Transport.InterruptEndpoint = EpInt;
}

#if _DEBUG
_TCHAR *SubClassName(UINT8 SubclassCode)
{
    switch (SubclassCode) {
    case USB_DS_RBC:
        return _T("RBC");
#if 1 /* MSFT memory Units */
    case 66:
        return _T("Msft-RBC");
#endif
    case USB_DS_ATAPI:
        return _T("ATAPI");
    case USB_DS_QIC:
        return _T("QIC");
    case USB_DS_UFI:
        return _T("UFI");
    case USB_DS_SFF_8070i:
        return _T("SFF_8070i");
    case USB_DS_SCSI:
        return _T("SCSI");
    default:
        break;
    }
    return _T("ILLEGAL");
}

_TCHAR *TransportName(UINT8 TransportCode)
{
    switch (TransportCode) {
    case USB_DP_CBI:
        return _T("CBI with command completion");
    case USB_DP_CBI_NO_CC:
        return _T("CBI without command completion");
    case USB_DS_BULK_ONLY:
        return _T("Bulk-Only");
    default:
        break;
    }
    return _T("ILLEGAL");
}
#endif

PRIVATE SCODE GetConfigurationDescriptor(
    PUSBSTOR this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT16 Length)
{
    SCODE sc;

    Urb->DeviceAddress = this->Transport.MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_DTH;
    Urb->SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb->SetupData[2] = 0;
    Urb->SetupData[3] = USB_DT_CONFIG;
    Urb->SetupData[4] = 0;
    Urb->SetupData[5] = 0;
    Set2(Urb->SetupData+6,Length);
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = (UINT8)Length;
    Urb->MaximumPacketSize = this->Transport.MaximumControlPacketSize;

    sc = this->Transport.Hub->v->SubmitUrb(this->Transport.Hub,
                                           Urb,
                                           this->Transport.PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbStor:GCD SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbStor:GCD SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}
    
PRIVATE SCODE SetConfiguration(
    PUSBSTOR this,
    PUSBM_REQUEST_BLOCK Urb,
    UINT8 Index)
{
    SCODE sc;

    Urb->DeviceAddress = this->Transport.MyUsbAddress;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->SetupData[0] = 0;
    Urb->SetupData[1] = USB_R_SET_CONFIG;
    Urb->SetupData[2] = Index;
    Urb->SetupData[3] = 0;
    Urb->SetupData[4] = 0;
    Urb->SetupData[5] = 0;
    Urb->SetupData[6] = 0;
    Urb->SetupData[7] = 0;
    Urb->SetupLength = 8;
    Urb->ExpectedDataLength = 0;
    Urb->MaximumPacketSize = this->Transport.MaximumControlPacketSize;

    sc = this->Transport.Hub->v->SubmitUrb(this->Transport.Hub,
                                           Urb,
                                           this->Transport.PortNo);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbStor:SC SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }
    Condition_Wait(Urb->Done,NULL,NULL);

    if (Urb->Flags & URB_FLAG_ERROR) {
        DBGME(3,_tprintf(_T("UsbStor:SC SubmitUrb failed er=x%x\n"),Urb->Flags));
        return E_MEDIA_CHANGED;
    }

    return S_OK;
}
    
PRIVATE SCODE UsbStorDriverIamYourHub (
    PIUSBDRIVER pThis,
    PIUSBHUB Hub,
    UINT PortNumber)
{
    PUSBSTOR this = pUsbStorDriver(pThis);
    PUSBM_REQUEST_BLOCK Urb = NULL;
    SCODE sc;
    UINT16 Length;
    PINAMESPACE Ns = CurrentNameSpace();
    _TCHAR MyName[MYNAMELENGTH];

    DBGME(0,_tprintf(_T("UsbStor:IYH(%x,%x,%d)\n"),
                     (ADDRESS)this,(ADDRESS)Hub,PortNumber));

    /* Check not already owned
     */
    if (this->Transport.Hub) {
        DBGME(3,_tprintf(_T("UsbStor: already owned.\n")));
        return E_ALREADY_EXISTS;
    }

    /* QI to make sure and to take a ref
     */
    sc = Hub->v->QueryInterface(Hub,
                                &IID_IUsbHub,
                                (void **)&this->Transport.Hub);
    if (FAILED(sc)) {
        DBGME(1,_tprintf(_T("UsbStor bad Hub, sc=%x\n"),sc));
        return sc;
    }

    this->Transport.PortNo = (UINT8) PortNumber;
    //done above: this->Transport.Hub = Hub;

    /* We need all URBs
     */
    Urb = UrbAlloc(this,256);
    this->Transport.ControlUrb = Urb;
    this->Transport.DataUrb = UrbAlloc(this,0);
    this->Transport.StatusUrb = UrbAlloc( this,128 );
    this->Transport.InterruptUrb = UrbAlloc( this,4 );
    if ((this->Transport.ControlUrb == NULL) ||
        (this->Transport.DataUrb == NULL) ||
        (this->Transport.InterruptUrb == NULL) ||
        (this->Transport.StatusUrb == NULL))
    {
        sc = E_NOT_ENOUGH_MEMORY;
        goto Error;
    }

    /* Get and set the config
     * Get it twice, first time just to find length.
     */
    sc = GetConfigurationDescriptor(this,Urb,9);
    if (FAILED(sc))
        goto Error;

    Length = Get2(Urb->Data+2);
    sc = GetConfigurationDescriptor(this,Urb,Length);
    if (FAILED(sc))
        goto Error;

    /* Find out about the transport and the endpoint(s)
     */
    ParseConfigurationDescriptor(this,Urb->Data,Urb->ActualDataLength);

    /* Say what we think
     */
    DBGME(1,_tprintf(_T("UsbStor: class %s, transport %s.\n"),
                     SubClassName(this->Subclass),
                     TransportName(this->TransportType)));

    sc = SetConfiguration(this,Urb,1);
    if (FAILED(sc))
        goto Error;

    /* Start the transport driver
     * Do it before we declare ourselves an official "disk<n>"
     */
    switch (this->TransportType) {
    case USB_DS_BULK_ONLY:
        sc = CreateBulkTransport(&this->Transport);
        break;
    case USB_DP_CBI:
        sc = CreateCbiTransport(&this->Transport);
        break;
    case USB_DP_CBI_NO_CC:
        sc = CreateCbTransport(&this->Transport);
        break;
    default:
        DBGME(3,_tprintf(_T("UsbStor: transport %s unsupported\n"),
                         TransportName(this->TransportType)));
        sc = E_NOT_IMPLEMENTED;
        goto Error;
    }

    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbStor: transport %s failed to start, sc=%x.\n"),
                         TransportName(this->TransportType), sc));
        goto Error;
    }

    /* Now the class driver
     */
    switch (this->Subclass) {
    case USB_DS_ATAPI:
        sc = CreateAtapiClass(&this->Transport, &this->Class);
        break;
    case USB_DS_QIC:
        sc = CreateQicClass(&this->Transport, &this->Class);
        break;
    case USB_DS_UFI:
        sc = CreateUfiClass(&this->Transport, &this->Class);
        break;
    case USB_DS_SFF_8070i:
        sc = CreateSffClass(&this->Transport, &this->Class);
        break;
#if 1
    case 66:
#endif
    case USB_DS_RBC:
    case USB_DS_SCSI:
        sc = CreateScsiClass(&this->Transport, &this->Class);
        break;
    default:
        DBGME(3,_tprintf(_T("UsbStor: class %s unsupported\n"),
                         SubClassName(this->Subclass)));
        sc = E_NOT_IMPLEMENTED;
        goto Error;
    }

    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbStor: class %s failed to start, sc=%x.\n"),
                         SubClassName(this->Subclass), sc));
        goto Error;
    }

    /* Add top-level namespace entry, for user convenience
     * HUB only puts us in the USB tree.
     * Try again to avoid clashes if multiple disks are attached
     * Put a limit to it, though, just in case.
     */
    for (;;) {
        MakeMyName(this,MyName);
        sc = Ns->v->Register(Ns,
                             MyName,
                             (PIUNKNOWN) this,
                             NAME_SPACE_READ | 
                             NAME_SPACE_WRITE |
                             NAME_SPACE_CREATE |
                             NAME_SPACE_FAILIFEXIST,
                             NULL);
        if ((sc == E_ALREADY_EXISTS) && (this->DiskNo < 1000))
            this->DiskNo++;
        else
            break;
    }

    DBGME(2,_tprintf(_T("%s is a '%s'\n"),
                     MyName, this->Transport.InquiryName));

#if notyet
    /* Start the filesystem on top of us.
     * BUGBUG: needs an intermediate driver that discovers the filesys format.
     */
    _TCHAR FsName[12];
    _stprintf(FsName,_T("fs%d"),this->DiskNo);
    StartDriver(FsName,_T("COB\fatfs.cob"),this->MyName);
#endif

    return S_OK;
 Error:
    if (this->Transport.Hub) {
        this->Transport.Hub->v->Release(this->Transport.Hub);
        this->Transport.Hub = NULL;
    }
    if (this->Transport.ControlUrb) {
        free(this->Transport.ControlUrb);
        this->Transport.ControlUrb = NULL;
    }
    if (this->Transport.DataUrb) {
        free(this->Transport.DataUrb);
        this->Transport.DataUrb = NULL;
    }
    if (this->Transport.StatusUrb) {
        free(this->Transport.StatusUrb);
        this->Transport.StatusUrb = NULL;
    }
    if (this->Transport.InterruptUrb) {
        free(this->Transport.InterruptUrb);
        this->Transport.InterruptUrb = NULL;
    }
    return sc;
}

PRIVATE SCODE UsbStorDriverYouAreTerminated (
    PIUSBDRIVER pThis,
    PIUSBHUB SaysWho)
{
    PUSBSTOR this = pUsbStorDriver(pThis);
    SCODE sc;
    PINAMESPACE Ns = CurrentNameSpace();
    _TCHAR MyName[MYNAMELENGTH];

    DBGME(0,_tprintf(_T("UsbStor:YAT(%x)\n"),(ADDRESS)SaysWho));

    /* Check args, and that we had started ok
     */
    if (this->Transport.Hub == NULL) {
        return E_INVALID_PARAMETER;
    }

    /* Remove from namespace
     */
    MakeMyName(this,MyName);
    sc = Ns->v->Unregister(Ns,MyName);

#if notyet
    /* Stop generic stor on top
     */
    ...how?? IQ(IDriver++,&ms) ms->KillYourself();
#endif

    /* Cancel anything pending
     */
    if (this->Transport.ControlUrb)
        sc = this->Transport.Hub->v->CancelUrb(this->Transport.Hub,
                                               this->Transport.ControlUrb,
                                               this->Transport.PortNo);

    /* LEt HUB go
     */
    this->Transport.Hub->v->Release(this->Transport.Hub);
    this->Transport.Hub = NULL;

    /* BUGBUG make sure refcount works out.
     */
    if (this->CDevice.RefCnt > 1)
        _tprintf(_T("UsbStor::KillYourself too many refs %d\n"),
               this->CDevice.RefCnt);

    return S_OK;
}

PRIVATE SCODE UsbStorDriverGetUsbAddress (
    PIUSBDRIVER pThis,
    PUINT UsbAddress,
    PUINT PortNumber)
{
    PUSBSTOR this = pUsbStorDriver(pThis);

    if (UsbAddress == NULL)
        return E_INVALID_PARAMETER;
    if (this->Transport.MyUsbAddress == 0)
        return E_OBJECT_NOT_INITIALIZED;

    *UsbAddress = this->Transport.MyUsbAddress;
    *PortNumber = this->Transport.PortNo;

    return S_OK;
}

#include "s_UsbStor_v.c"

/* Constructor */
PIDEVICE UsbStorNew(
    const _TCHAR *MyName,
    _TCHAR *HciName,
    UINT8 UsbAddress,
    UINT8 MaximumControlPacketSize,
    PIUNKNOWN /*IUSBMASTERCONTROLLER*/ UsbTreeLocation)
{
    PIHEAP pHeap = CurrentHeap();
    PUSBSTOR this;

    if (UsbAddress == 0) {
        DBGME(3,_tprintf(_T("UsbStorNew: bad UsbAddress 0\n")));
        return NULL;
    }

    this = (PUSBSTOR) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *this, 0);
    if (!this)
        return NULL;

    this->CDevice.v = (struct CDeviceVtbl*)&UsbStorDeviceUsbStorVtbl;
    this->CDevice.RefCnt = 1;
    this->UsbStorDriverv = &UsbStorDriverUsbStorVtbl;

    /* Do this because it aint config that starts us
     */
    Condition_Init(&this->CDevice.IntrPoint);

    this->Transport.MyUsbAddress = UsbAddress;
    this->Transport.MaximumControlPacketSize = MaximumControlPacketSize;

    return iUsbStorDevice(this);
}

/* ** 
*/
SCODE __cdecl UsbStorConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * MyName,
        ADDRESS UnusedRegisterBase,
        UINT UnusedInterrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        _TCHAR *HciName,
        PIUNKNOWN /*PIUSBMASTERCONTROLLER*/ UsbTreeLocation,
        ADDRESS UsbAddress,
        ADDRESS MaximumControlPacketSize)
{
    PIUNKNOWN pUnk;

    UnusedParameter(UnusedRegisterBase);
    UnusedParameter(UnusedInterrupt);

    DBGME(0,_tprintf(_T("UsbStorConstructor(%x %s %x %s %d %x)\n"),
                     (ADDRESS) pThis, MyName, (ADDRESS) ppUnk,
                     (HciName) ? HciName : _T("NULL"),
                     UsbAddress, (ADDRESS) UsbTreeLocation));

    pUnk = (PIUNKNOWN) UsbStorNew(MyName,
                                   HciName,
                                   (UINT8) UsbAddress,
                                   (UINT8) MaximumControlPacketSize,
                                   UsbTreeLocation);

    DBGME(0,_tprintf(_T("UsbStorConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UsbStorConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UsbStorDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UsbStorConstructor
};

static struct _DRIVERCOB UsbStorDriverCobObject = { &UsbStorDriverCobVtable, 1 };

PIUNKNOWN UsbStorCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UsbStorDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UsbStorDeviceUsbStorVtbl, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &UsbStorDriverCobObject;
}
