/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    udc.c
 *
 * Purpose:
 *
 *    Device driver for Intel's UDC USB slave.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbslave.h>

#include "udc.h"
#define MAXIN  16  /* Specs says depth of xmit FIFO is 16 */
#define MAXOUT 20  /* Specs says depth of recv FIFO is 20 */

/* Event logging and printing
 */
#define LOGGING 0
#if LOGGING
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <ctype.h>
#endif

/* Byte queue (for recv)
 */
typedef struct {
    CONDITION WaitPoint;
    TIME TimeOut;
    volatile INT Head, Tail;
    UINT Size;
    UINT Drops;
#define MAXOUTBUF 512
    UINT8 Buffer[MAXOUTBUF+8+8];
} BYTE_QUEUE;

/* Is the queue empty ?
 */
#define CQueueEmpty(_q_) \
    ((_q_).Head == (_q_).Tail)

/* How many bytes in the queue ?
 * Non-opaque, but interrupt-safe eval
 */
#define CQueueCountEval(_h_,_t_,_s_) \
    (((INT)(_h_) > (INT)(_t_)) ? \
     ((INT)(_h_) - (INT)(_t_)) : \
     ((_s_) - ((INT)(_t_) - (INT)(_h_))))

/* How many bytes in the queue ?
 * Opaque, but beware of multiple evals
 */
#define CQueueCount(_q_) \
    CQueueCountEval(_q_.Head,_q_.Tail,_q_.Size)

/* How much room left in the queue ?
 * Non-opaque, but interrupt-safe eval
 */
#define CQueueFreeEval(_h_,_t_,_s_) \
    (((INT)(_h_) > (INT)(_t_)) ? \
     ((_s_)-1 - ((INT)(_h_) - (INT)(_t_))) : \
     (((INT)(_t_) - (INT)(_h_))-1))

/* How much room left in the queue ?
 * Opaque, but beware of multiple evals
 */
#define CQueueFree(_q_) \
    CQueueFreeEval(_q_.Head,_q_.Tail,_q_.Size)

/* Initialize a byte queue
 */
#define CQueueInit(_q_) \
  { \
    (_q_).TimeOut = TIME_FOREVER; \
    Condition_Init(& (_q_).WaitPoint); \
    (_q_).Tail = (_q_).Head = -1; \
    (_q_).Size = MAXOUTBUF; \
    (_q_).Drops = 0; \
  }

/* UDC device object
 */
typedef struct _UDC {
    struct CDevice CDevice;
    const struct IUsbSlaveControllerVtbl *v;

    /* Our own state
     */
    PUdc Registers;
    PIUSBSLAVEDRIVER Driver;

    /* State per USB specs
     */
    UINT16 DeviceStatus;
    UINT16 ConfigurationSelected;
    UINT16 IfaceStatus;
    UINT16 EndpStatus[2];

    /* Control endpoint state
     */
    UINT8 *XmitZeroDataPtr;
    UINT XmitZeroCount;
    UINT8 *RecvZeroDataPtr;
    UINT RecvZeroCount;
    UINT RecvZeroMore;
    USB_REQUEST_PACKET LastCmd;

    UINT8 FunnySize;
    UINT8 XactPhase;
#define UDC_CTL_IDLE 0
#define UDC_CTL_XMIT 1
#define UDC_CTL_XMIT_END0 2  /* gets multiple intrpts at end.. umbug */
#define UDC_CTL_XMIT_END1 3
#define UDC_CTL_XMIT_END2 4
#define UDC_CTL_RECV 8

    UINT8 ConnectionState;
#define UDC_CON_DISCONNECTED 0  /* cable unplugged */
#define UDC_CON_CONNECTING 1    /* cable plugged, but host hasnt spoken */
#define UDC_CON_CONNECTED 2     /* all is dandy */

    /* Endpoint 1 (OUT aka recv) state
     */
    BYTE_QUEUE OneQue;

    /* Endpoint 2 (IN aka xmit) state
     */
    UINT8 *XmitTwoDataPtr;
    volatile int XmitTwoCount;
    CONDITION XmitTwoWaitPoint;
    TIME XmitTwoTimeOut;

    struct {
        volatile BOOL HwActive;
        volatile BOOL Loading;
        volatile UINT Count;
        UINT8 Data[MAXIN];
    } SwFifo;

    /* Stats and misc
     */
    UINT nReset;
    UINT nInts;
#if LOGGING
    UINT nRecvd;
    UINT nSent;
    UINT nReqs;
    UINT nReps;
#endif

} UDC, *PUDC;

#define pUDC(_t_) ((PUDC)(_t_))
#define pUDCfromSLAVE(_t_) ((PUDC)(((BYTE *)_t_) - offsetof(UDC,v)))
#define pSLAVEfromUDC(_t_) ((PIUSBSLAVECONTROLLER)(&(_t_)->v))


/* Debug stuff
 */
#if LOGGING
PRIVATE void Log(UINT n);
PRIVATE void PrintLog(void);
PRIVATE void UdcState(PUDC this, UINT n);
#include "log.c"
#else
#define Log(x)
#define PrintLog()
#define UdcState(x,y)
#endif

#define DONT_SPLIT_THE_UDC 0
#if DONT_SPLIT_THE_UDC
/* IUsbSlaveDriver interface (before split)
 */
#include "irene.c"
#endif

/* Device was reset, either at startup or by the host
 */
BOOL UdcReset(PUDC this)
{
    UINT x;
    PUdc Regs = this->Registers;

    this->nReset++;

    /* Says wait till this goes off */
    while (Regs->Control & UDC_C_UDA)
        Delay(1);

    /* Clear the reset bit
     */
    while (Regs->Status & UDC_ST_RSTIR)
        Regs->Status |= UDC_ST_RSTIR;

    /* Try to clear the SRM bit
     */
    for (x = 0; x < 20; x++) {
        if (Regs->Control & UDC_C_SRM)
            Regs->Control &= ~UDC_C_SRM;
        else
            break;
    }

    /* SRM bit only goes away if host reset or cable inserted
     */
    if (x < 20) {
        /* Say we might receive
         */
        while (Regs->MaxOut != (MAXOUT-1))
            Regs->MaxOut = MAXOUT-1;
        /* Say we might xmit
         */
        while (Regs->MaxIn != (MAXIN-1))
            Regs->MaxIn = MAXIN-1;
        /* Errata #29: set bit 7 to avoid FIFO starving (stepping B5)
         * NB: Not what we have on the Isyc boards (B2) but..
         */
        while ((Regs->Control & UDC_C_REM) == 0)
            Regs->Control |= UDC_C_REM;
    } else {
        /* cable disconnected ? */
    }

    this->XactPhase = UDC_CTL_IDLE;

    return TRUE;
}

/* Stop the UDC
 */
void UdcStop(PUDC this)
{
    PUdc Regs = this->Registers;
    UINT x;

    /* BUGBUG be nice and check UDA
     */
    Regs->Control = UDC_C_UDD;
    x = Regs->Control;
}

/* Initialize the UDC
 */
INT UdcInit(PCDEVICE pThis)
{
    PUDC this = pUDC(pThis);
    PUdc Regs = this->Registers;
    UINT x;

    /* We cannot initialize until we have an owner.
     */
    if (this->Driver == NULL)
        return 0;

    UdcState(this,555);
    UdcStop(this);
    UdcState(this,666);

    /* clear sticky bits
     */
    Regs->Status = ~0;
    x = Regs->Status;

    /* Enable interface
     */
    Regs->Control = UDC_C_SRM;
    x = Regs->Control;

    /* Say its a reset
     */
    UdcReset(this);

    UdcState(this,999);
    _tprintf(_T("Plug in the USB cable now..\n\r"));

    return 0;
}

/* Read from the control fifo
 */
UINT UdcReadFifo0(PUDC this, UINT8 *pData)
{
    PUdc Regs = this->Registers;
    UINT Count, Wc, x;
    /* Errata #32 says FIFO logic a bit shakey.. */

    /* Here and later on: spec say it might take a couple cycles
     * for the registers to be consistent, but by the next read
     * they are definitely ok.  So we readback or read twice.
     */
    Wc = Regs->WriteCountZero;
    Wc = Regs->WriteCountZero;

    /* Repeat while we got data
     */
    for (Count = Wc; Count > 0; Count--) {
        /* Pull one byte out
         */
        *pData++ = (UINT8) Regs->DataZero;

        /* Now wait till the UDC notices
         */
        for (x = 0;Regs->WriteCountZero == Count;x++) {
            /* Dont wait too long though
             */
            if (x > 10000) {
                DBGME(3,_tprintf(_T("ReadFifo stuck(%x %x)\r\n"),Wc,Count));
                goto Out;
            }
            Delay(1);
        }
    }

    /* Return whatever much data we got
     */
 Out:
    return Wc - Count;
}

/* Errata #32 says FIFO logic a lot shakey..
 */
UINT UdcWriteFifo0(PUDC this, UINT8 *pData, UINT Count)
{
    PUdc Regs = this->Registers;
    UINT Wc, Sent, x, nRetry;
#if LOGGING
    UINT8 *l;
#endif

    /* Trim to our max size
     */
    Sent = 0;
    if (Count > 8)
        Count = 8;

#if LOGGING
    /* Get a log buffer
     */
    l = &FromMe[this->nReps][1];
#endif

    /* For as long as we got some
     */
    for (;Count-- > 0;) {

        /* Take note of what was there
         */
        Wc = Regs->WriteCountZero;
        Wc = Regs->WriteCountZero;

        /* We'll have to try a few times, possibly
         */
        nRetry = 0;
    Again:
#if LOGGING
        /* Note what we say
         */
        *l = *pData;
#endif
        /* Push the byte out
         */
        Regs->DataZero = *pData;
        
        /* Make sure the UDC notices
         */
        for (x = 0; Regs->WriteCountZero == Wc; x++) {
            /* Spin a while
             */
            if (x > 100) {
                /* Retry a few times
                 */
                if (nRetry++ < 10)
                    goto Again;

                /* Dont get stuck here forever
                 */
                DBGME(3,_tprintf(_T("WriteFifo stuck(%x %x)\r\n"),Wc,Count));
                UdcState(this,99);
                goto Out;
            }
            Delay(1);
        }

        /* Ok, this one got over the wall, next please
         */
#if LOGGING
        l++;
#endif
        Sent++;
        pData++;
    }

    /* Say how much we actually sent
     */
 Out:
#if LOGGING
    FromMe[this->nReps++][0] = (UINT8)Sent;
    /* Wrap buffers if necessary (spooler is slow/non existant)
     */
    if (this->nReps >= MAXREP) this->nReps--;
#endif

    return Sent;
}

/* Read from the data fifo
 */
UINT UdcReadFifo1(PUDC this, UINT8 *pData)
{
    PUdc Regs = this->Registers;
    UINT Count, x;

    /* Say we got none for starters
     */
    Count = 0;

    /* Keep going while we got stuff
     */
    for (;;) {
        /* Have we pulled the last one out
         */
        x = Regs->EndpointOneControl;
        x = Regs->EndpointOneControl;
        if ((x & UDC_C1_RNE) == 0)
            break;

        /* Nope, pick one
         */
        pData[Count++] = (UINT8) Regs->Data;
    }

    /* Receipt is complete
     */
    Regs->EndpointOneControl = UDC_C1_RPC;
    x = Regs->EndpointOneControl;

    /* Return how much we got
     */
    return Count;
}

/* Write to the data fifo
 */
UINT UdcWriteFifo1(PUDC this, UINT8 *pData, UINT Count)
{
    PUdc Regs = this->Registers;
    UINT Sent, x;

    /* Say we did nothing, and trim for sanity
     */
    Sent = 0;
    if (Count > MAXIN)
        Count = MAXIN;

    /* Say how much we'll send, make it stick
     */
    while (Regs->MaxIn != (Count-1))
        Regs->MaxIn = Count-1;

    /* One at a time..
     */
    for (;Count-- > 0;) {
        /* Make sure all is still ok
         */
        x = Regs->EndpointTwoControl;
        x = Regs->EndpointTwoControl;
        if (x & (UDC_C2_TPE|UDC_C2_TPC|UDC_C2_TUR))
            break;

        /* Push it out
         */
        Regs->Data = *pData;
        
        /* Next pls
         */
        Sent++;
        pData++;
    }

    /* Return how much we wrote
     */
    return Sent;
}


/* Code for the Software Fifo.
 * This is necessary for performance reasons,
 * we dont want to be sending one byte per USB packet.
 */
void LoadFifo1(PUDC this); /* forward */

void LoadSwFifo1(PUDC this)
{
    int n, i, j;

    /* Lock out the ISR, gently.
     * When an interrupt occurs while we are in here the ISR
     * notices and backs out but notices that HwActive==FALSE.
     * So we'll pick it up at the end of this routine.
     * The xmit ISR can only trigger once, when its done sending.
     */
    this->SwFifo.Loading = TRUE;

    Log( ('s' | 0x900) | (this->SwFifo.Count << 16) );

    /* Sanity. This was happening without a volatile or two.
     */
    if (this->XmitTwoCount < 0) {
        Log( ('o' | 0x900) | (this->XmitTwoCount << 16) );
        this->XmitTwoCount = 0;
    }

    /* See how much we can possibly push this turn
     */
    n = this->XmitTwoCount;
    Log( (1 | 0x600) | (n << 16) );

    /* There is some problem if we use the full fifo.
     * Cant say without a scope but.. since this was built for use
     * with a DMA engine (but of course they screwed it up and the
     * erratas say dont even think about using DMA..sigh) I assume
     * it expects a full load to be followed by another and only knows
     * that its done if it sees a partial count. 
     * So we always use a partial count.
     */
#define __MAXIN (MAXIN-1)
    /* Trim, accounting for whats already in there
     */
    if (n > __MAXIN) n = __MAXIN;
    i = this->SwFifo.Count;
    if (n > (__MAXIN-i)) n = __MAXIN-i;
#undef __MAXIN
    Log( (0 | 0x600) | (n << 16) );
    /* Likely just one byte
     */
    for (j = 0; j < n; j++)
        this->SwFifo.Data[i+j] = this->XmitTwoDataPtr[j];
    /* Adjust counters and pointers
     */
    this->XmitTwoDataPtr += n;
    this->XmitTwoCount -= n;
    this->SwFifo.Count += n;

    /* And we are done
     */
    this->SwFifo.Loading = FALSE;

    /* Get thing started if the hw was quiescent
     */
    if (!this->SwFifo.HwActive)
        LoadFifo1(this);
}

/* Load the actual hardware fifo.
 */
void LoadFifo1(PUDC this)
{
    UINT Count;

    /* If we got here it must be that..
     */
    this->SwFifo.HwActive = FALSE;

    /* If this is from the ISR back off if necessary
     */
    if (this->SwFifo.Loading)
        return;

    /* Is there some left
     */
    Log( ('h' | 0x900) | (this->SwFifo.Count << 16) );
    if (this->SwFifo.Count == 0) {
        /* Sure but.. no need to wakeup for something we can do ourselves.
         */
        if (this->XmitTwoCount > 0) {
            /* Push onto the sw fifo, which then comes back here
             */
            Log( (0 | 0x400) | (this->XmitTwoCount << 16) );
            LoadSwFifo1(this);
        }
        /* Done.
         */
        return;
    }

    /* Copy from sw to hw fifo
     */
    Count = UdcWriteFifo1(this,this->SwFifo.Data,this->SwFifo.Count);
    Log( (0 | 0x500) | (Count << 16) );

    /* Say how much we managed to get out (all of it unless error)
     */
    this->SwFifo.Count -= Count;
    Log( (0 | 0x500) | (this->SwFifo.Count << 16) );

    /* Now the hw is churning
     */
    this->SwFifo.HwActive = TRUE;

    /* Since we emptied the sw fifo see if we can reload it.
     * This improves performance because the writer thread wakes up
     * earlier, and can come back with more data before we are done
     * sending this packet out.  Which makes for a bigger packet next.
     */
    if ((this->SwFifo.Count == 0) &&
        (this->XmitTwoCount > 0))
        LoadSwFifo1(this);
}

/* Send one packet on the control fifo
 */
void UdcSendControlPacket(PUDC this)
{
    PUdc Regs = this->Registers;
    UINT Count, x;

    /* Write it out, but not all of it might fit.
     */
    Count = UdcWriteFifo0(this,this->XmitZeroDataPtr,this->XmitZeroCount);

    /* Adjust counter and pointer
     */
    this->XmitZeroCount -= Count;
    this->XmitZeroDataPtr += Count;
#if LOGGING
    this->nSent += Count;
#endif

    /* Say we got a packet ready
     */
    Regs->EndpointZeroControl = UDC_C0_IPR;
    x = Regs->EndpointZeroControl;
    
    /* what next ? */
    if (this->XmitZeroCount) {
        /* There is still something to send, stay in this phase
         */
        this->XactPhase = UDC_CTL_XMIT;
    } else {
        /* If we closed with a full size packet the hw expects us to send more
         * and we'll need to send a zero-sized packet.
         */
        if (Count == 8)
            this->FunnySize = 0;
        /* That might be necessary also in the case we send
         * precisely as much data as the host asked for
         */
        if (this->FunnySize)
            /* No need for more
             */
            this->XactPhase = UDC_CTL_XMIT_END1;
        else
            /* Will need the zero-sized pkt
             */
            this->XactPhase = UDC_CTL_XMIT_END0;
    }
}

/* Interrupt Service Routine for the control endpoint
 */
BOOL MCT UdcE0Isr(PUDC this)
{
    UINT Status;
    UINT Count, x, i;
    SCODE sc;
    PUSB_REQUEST_struct pReq;
    PUdc Regs = this->Registers;

    /* Whassup ?
     */
    Status = Regs->EndpointZeroControl;
    Status = Regs->EndpointZeroControl;

    Log( (Status | 0x100) | (this->XactPhase << 16) | (this->nReqs << 24));

    /* SE bit says we got shutup
     */
    while (Status & UDC_C0_SE) {
        /* Clear it
         */
        Regs->EndpointZeroControl = UDC_C0_SSE;
        Status = Regs->EndpointZeroControl;
        Status = Regs->EndpointZeroControl;

        /* Back to not doing much
         */
        this->XactPhase = UDC_CTL_IDLE;
        Log( (Status | 0x100) | (this->XactPhase << 16) | (this->nReqs << 24));
    }

    /* Depending on where we are..
     */
    switch (this->XactPhase) {

    case UDC_CTL_IDLE:

        /* We should have a new request ready for us
         */
        if (Status & UDC_C0_OPR) {

            /* Copy to the current request buffer
             */
            pReq = (PUSB_REQUEST_struct)&this->LastCmd;
            i = UdcReadFifo0(this,(UINT8*)pReq);
#if LOGGING
            /* Get a buffer to log it into, wrap if necessary
             */
            x = this->nReqs++;
            if (this->nReqs >= MAXREQ) this->nReqs--;
            pReq = (PUSB_REQUEST_struct)(ToMe[x]);

            /* Copy and account for stats
             */
            memcpy((UINT8*)pReq,&this->LastCmd,i);
            this->nRecvd += i;
#endif
            /* act on (some) device-level request
             */
            if (pReq->RequestType & (USB_RT_CLASS|USB_RT_VENDOR))
                goto ClassRequest;

            switch (pReq->Request) {

            case USB_R_GET_DESCRIPTOR:
                /* Expected length is?
                 */
                Count = Get2(pReq->Length);

                /* ack */
                if (Count) {
                    /* Make sure we clear the bit
                     */
                    while (Regs->EndpointZeroControl & UDC_C0_OPR) {
                        Regs->EndpointZeroControl = UDC_C0_SO;
                        x = Regs->EndpointZeroControl;
                    }
                }

                /* Which descriptor does it want ?
                 * NB: We cannot do v2.0 speeds, so dont even try
                */
                if (pReq->Value[1] == USB_DT_QUALIF)
                    goto NoReply;

                /* Ask class driver, telling it what we expect to get max
                 */
                this->XmitZeroCount = Count;
                sc = this->Driver->v->GetDescriptor(
                              this->Driver,
                              pReq->Value[1],
                              pReq->Value[0],
                              &this->XmitZeroDataPtr,
                              &this->XmitZeroCount);
                if (FAILED(sc)) {
                    DBGME(3,_tprintf(_T("BadDesc: x%x\r\n"),Get2(pReq->Value)));
                    goto NoReply;
                }

            NeedsInputPhase:
                /* Trim to what the host asked for
                 */
                if (this->XmitZeroCount > Count)
                    this->XmitZeroCount = Count;

                /* If its an exact match.. zerothing at the end
                 */
                this->FunnySize = (Count == this->XmitZeroCount);
                
                /* Start sending our reply
                 */
                UdcSendControlPacket(this);
                break;

            case USB_R_SET_ADDRESS:
                /* Just do it
                 */
                Regs->Address = Get2(pReq->Value) & UDC_A_MASK;
                goto NoReply;
                break;

            case USB_R_GET_STATUS:
                /* ack */
                Regs->EndpointZeroControl = UDC_C0_SO;
                Count = Regs->EndpointZeroControl;

                /* What status is desired ?
                 * Set the appropriate pointer
                 */
                switch (pReq->RequestType & USB_RT_RECIPIENT) {

                case USB_RT_DEV: 
                    this->XmitZeroDataPtr = (UINT8*)&this->DeviceStatus; break;

                case USB_RT_IFACE:
                    this->XmitZeroDataPtr = (UINT8*)&this->IfaceStatus; break;

                case USB_RT_ENDP:
                    this->XmitZeroDataPtr =
                        (UINT8*)&this->EndpStatus[Get2(pReq->Index)];
                    break;

                default:
                    /* oops
                     */
                    goto Error;
                }
                /* Status is defined to be two bytes
                 * BUGBUG byteorder not right if BigEndian
                 */
                this->XmitZeroCount = sizeof(UINT16); /* caveat emptor */

                /* Send reply
                 */
                UdcSendControlPacket(this);
                break;

            case USB_R_SET_CONFIG:
                /* This is the last command we get when we connect to
                 * the host. Dont know if its in the specs but.
                 */
                this->ConnectionState = UDC_CON_CONNECTED;
                this->ConfigurationSelected = Get2(pReq->Value);
                goto NoReply;
                break;

            case USB_R_CLEAR_FEATURE:
                /* we only handle clear stall on our data endpoints
                 */
                if ((pReq->RequestType != USB_RT_ENDP) ||
                    (pReq->Value[0] != USB_FS_ENDP_STALL))
                    goto Error;

                /* Same thing needs to be done, but bits and registers...
                 */
                if (pReq->Index[0] == (USB_EA_IN | 2)) {
                    /* force stall */
                    Regs->EndpointTwoControl = UDC_C2_FST;
                    x = Regs->EndpointTwoControl;
                    x = Regs->EndpointTwoControl;

                    /* clear it */
                    Regs->EndpointTwoControl = 0;
                    x = Regs->EndpointTwoControl;
                    x = Regs->EndpointTwoControl;

                    /* clear sent stall */
                    Regs->EndpointTwoControl = UDC_C2_SST;
                    x = Regs->EndpointTwoControl;
                    x = Regs->EndpointTwoControl;

                }

                if (pReq->Index[0] == (USB_EA_OUT | 1)) {
                    /* force stall */
                    Regs->EndpointOneControl = UDC_C1_FST;
                    x = Regs->EndpointOneControl;
                    x = Regs->EndpointOneControl;

                    /* clear it */
                    Regs->EndpointOneControl = 0;
                    x = Regs->EndpointOneControl;
                    x = Regs->EndpointOneControl;

                    /* clear sent stall */
                    Regs->EndpointOneControl = UDC_C1_SST;
                    x = Regs->EndpointOneControl;
                    x = Regs->EndpointOneControl;
                }
                goto NoReply;

            default:
            ClassRequest:
                /* Hand off to class driver
                 */
                this->RecvZeroMore = this->RecvZeroCount;
                sc = this->Driver->v->HandleRequest(
                              this->Driver,
                              pReq,
                              &this->RecvZeroDataPtr,
                              &this->RecvZeroMore);
                this->RecvZeroCount = 0;

                /* Are we done ?
                 */
                if (sc == S_OK)
                    goto NoReply;

                /* Can we not handle it at all?
                 */
                if (FAILED(sc))
                    goto Error;

                /* The request is followed by a data phase
                 * Which one?
                 */
                if (sc == S_BUFFER_TOO_SMALL) {
                    /* OUT
                     */
                    Regs->EndpointZeroControl = UDC_C0_SO;
                    x = Regs->EndpointZeroControl;
                    x = Regs->EndpointZeroControl;

                    this->XactPhase = UDC_CTL_RECV;
                    goto Done;
                }

                if (sc == S_MUST_REPLY) {
                    /* IN
                     * Urgh, swap them buffers
                     */
                    this->XmitZeroDataPtr = this->RecvZeroDataPtr;
                    this->XmitZeroCount   = this->RecvZeroMore;
                    goto NeedsInputPhase;
                }

                /* Anything else.. we are soo confused.
                 */

            Error:
                DBGME(3,_tprintf(_T("BadReq: x%x %d\r\n"),pReq->Request,i));

                /* BUGBUG Shouldnt we STALL instead ??
                 */

            NoReply:
                /* We got nothing to say, or we are confused
                 */
                Regs->EndpointZeroControl = UDC_C0_SO | UDC_C0_DE;
                x = Regs->EndpointZeroControl;
                x = Regs->EndpointZeroControl;
                break;
            }
        }
        break;

    case UDC_CTL_XMIT:
        /* In the middle of sending a longish reply, send another chunk
         */
        UdcSendControlPacket(this);
        break;

    case UDC_CTL_XMIT_END0:
        /* Do it one more time
         */
        Regs->EndpointZeroControl = UDC_C0_IPR;
        x = Regs->EndpointZeroControl;
        this->XactPhase = UDC_CTL_XMIT_END1;
        break;

    case UDC_CTL_XMIT_END1:
        /* We are really done
         */
        Regs->EndpointZeroControl = UDC_C0_DE;
        x = Regs->EndpointZeroControl;
        this->XactPhase = UDC_CTL_XMIT_END2;
        break;

    case UDC_CTL_XMIT_END2:
        /* I mean.. really!
         */
        this->XactPhase = UDC_CTL_IDLE;
        break;

    case UDC_CTL_RECV:
        /* We should have gotten more data
         */
        if (Status & UDC_C0_OPR) {

            /* Copy it where it should go
             */
            i = UdcReadFifo0(this,this->RecvZeroDataPtr+this->RecvZeroCount);

#if LOGGING
            /* Account for stats
             */
            this->nRecvd += i;
#endif

            this->RecvZeroCount += i;
            this->RecvZeroMore  -= i;

        }

        /* Got all of it ?
         */
        if (this->RecvZeroMore > 0) {
            /* Nope, ack and stay
             */
            goto NeedsOutputPhase;
        }

        /* Now act on the command
         */
        pReq = (PUSB_REQUEST_struct)&this->LastCmd;
        Log( (':' | 0x900) | (pReq->Request << 16));
        Log( (this->RecvZeroDataPtr[1] | 0x200) |
             (this->RecvZeroDataPtr[2] << 16) |
             (this->RecvZeroDataPtr[0] << 24) );

        /* Hand off to class driver
         */
        this->RecvZeroMore = this->RecvZeroCount;
        sc = this->Driver->v->HandleRequest(
                              this->Driver,
                              pReq,
                              &this->RecvZeroDataPtr,
                              &this->RecvZeroMore);
        this->RecvZeroCount = 0;

        /* Are we done ?
         */
        if (sc == S_OK) {
            /* Yep, ack
             */
            Regs->EndpointZeroControl = UDC_C0_SO | UDC_C0_DE;
            x = Regs->EndpointZeroControl;
            x = Regs->EndpointZeroControl;

            this->RecvZeroDataPtr = NULL;
            this->RecvZeroMore = 0;
            this->XactPhase = UDC_CTL_XMIT_END2;

            goto Done;
        }

        /* Can we not handle it at all?
         */
        if (FAILED(sc))
            goto SendStall;

        /* Needs even more data, get more and stay in this phase
         */
    NeedsOutputPhase:
        Regs->EndpointZeroControl = UDC_C0_SO;
        x = Regs->EndpointZeroControl;
        x = Regs->EndpointZeroControl;
        break;

    default:
        /* Should not happen, try to regain sanity
         */
        DBGME(3,_tprintf(_T("BadPhase: x%x\r\n"),this->XactPhase));
    SendStall:
        Regs->EndpointZeroControl = UDC_C0_FST;
        x = Regs->EndpointZeroControl;
        x = Regs->EndpointZeroControl;
        this->XactPhase = UDC_CTL_IDLE;
        break;
    }

    /* No threads to wakeup here
     */
    Done:
    return FALSE;
}

/* ISR for endpoint 1, the OUT endpoint (recv data)
 */
BOOL MCT UdcE1Isr(PUDC this)
{
    PUdc Regs = this->Registers;
    UINT i, x, n;
    INT Avail; /* signed !! */
    UINT8 c = 0;

    /* for log */
    i = Regs->EndpointOneControl;

    /* Empty the fifo
     */
    n = 0;
    for (;;) {
        /* Are we there yet
         */
        x = Regs->EndpointOneControl;
        x = Regs->EndpointOneControl;
        if ((x & UDC_C1_RNE) == 0)
            break;

        /* Pick the byte up
         */
        n++;
        c = (UINT8) Regs->Data;

        /* See if we have room for it.
         * [just a fast&funny !CQueueFull]
         */
        Avail = this->OneQue.Tail - this->OneQue.Head;
        if ((Avail != 1) && (Avail != (1 - (INT)this->OneQue.Size))) {

            /* Where do we put it. Update ptr.
             */
            Avail = this->OneQue.Head + 1;
            if (Avail == (INT)this->OneQue.Size)
                Avail = 0;
            this->OneQue.Buffer[Avail] = c;
            this->OneQue.Head = Avail;
            
        } else
            /* Oh well..
             */
            this->OneQue.Drops++;
    }

    /* ack */
    Regs->EndpointOneControl = UDC_C1_RPC;
    x = Regs->EndpointOneControl;

    Log( (i | 0x200) | (n << 16) | (c << 24));

    /* Wakeup reader
     */
    return Condition_InterruptSignal(&this->OneQue.WaitPoint);
}

/* ISR for endpoint 2, the IN endpoint (xmit data)
 */
BOOL MCT UdcE2Isr(PUDC this)
{
    UINT x;
    BOOL r;
    PUdc Regs = this->Registers;

    /* Better have a complete packet else we donno
     */
    x = Regs->EndpointTwoControl;
    if (x & UDC_C2_TPC) {
        /* NB: This clears error bits as well */
        Regs->EndpointTwoControl = x;
    }

    /* Move from sw to hw fifo
     */
    Log( (x | 0x300) | (this->SwFifo.Count<<16) );
    LoadFifo1(this);

    /* Wakeup iff we are done sending
     */
    r = FALSE;
    if (this->XmitTwoCount == 0)
        r = Condition_InterruptSignal(&this->XmitTwoWaitPoint);

    Log( (x | 0x300) | (r << 16) );
    return r;
}

/* ISR for the whole thing. Handles errors, dispatches as appropriate
 */
BOOL MCT UdcIsr(PCDEVICE pThis, BOOL *pNotOurs)
{
    PUDC this = pUDC(pThis);
    PUdc Regs = this->Registers;
    UINT Status;
    BOOL Resched = FALSE;

    /* Manual says always read twice.. umbug.
     */
    Status = Regs->Status;
    Status = Regs->Status;

    /* Clear all requests at once
     */
    Regs->Status = Status;

    /* Log and stats
     */
    Log(Status);
    this->nInts++;

    /* Suspend request interrupt (cable unplugged)
     */
    if (Status & UDC_ST_SUSIR) {
        /* Mask this interrupt
         */
        Regs->Control |= UDC_C_SRM;

        /* Notify our clients
         */
        this->ConnectionState = UDC_CON_DISCONNECTED;
        this->SwFifo.Count = 0;
        this->SwFifo.HwActive = FALSE;
        Resched |= Condition_InterruptSignal(&this->OneQue.WaitPoint);
        Resched |= Condition_InterruptSignal(&this->XmitTwoWaitPoint);
    }

    /* Resume
     */
    if (Status & UDC_ST_RESIR) {
        /* Toggle the RES bit to acknowledge
         */
        while ((Regs->Control & UDC_C_RES) == 0)
            Regs->Control |= UDC_C_RES;
        while (Regs->Control & UDC_C_RES)
            Regs->Control &= ~UDC_C_RES;

        /* Unmask the resume interrupt
         */
        while (Regs->Control & UDC_C_SRM)
            Regs->Control &= ~UDC_C_SRM;

        /* But wait until the host gets back to us
         */
        this->ConnectionState = UDC_CON_CONNECTING;
    }

    /* Reset
     */
    if (Status & UDC_ST_RSTIR) {
        UdcReset(this);
    }

    /* Control endpoint
     */
    if (Status & UDC_ST_EIR) {
        Resched |= UdcE0Isr(this);
    }

    /* Receive endpoint
     */
    if (Status & UDC_ST_RIR) {
        Resched |= UdcE1Isr(this);
    }

    /* Transmit endpoint
     */
    if (Status & UDC_ST_TIR) {
        Resched |= UdcE2Isr(this);
    }

    return Resched;
}

/* IUsbSlaveController interface
 */
SCODE UdcSlaveQueryInterface (
        /*in*/ PIUSBSLAVECONTROLLER pThis,
        /*in*/ REFIID iid,
        /*out*/ void** ppObject)
{
    PUDC this = pUDCfromSLAVE(pThis);
    return this->CDevice.v->QueryInterface((PCDEVICE)this,
                                           iid,
                                           ppObject);
}

/* Use a single refcount
 */
UINT UdcSlaveAddRef (
        /*in*/ PIUSBSLAVECONTROLLER pThis)
{
    PUDC this = pUDCfromSLAVE(pThis);
    return this->CDevice.v->AddRef((PCDEVICE)this);
}

/* Use a single refcount
 */
UINT UdcSlaveRelease (
        /*in*/ PIUSBSLAVECONTROLLER pThis)
{
    PUDC this = pUDCfromSLAVE(pThis);
    return this->CDevice.v->Release((PCDEVICE)this);
}

SCODE UdcSlaveTakeOwnership (
        /*in*/ PIUSBSLAVECONTROLLER pThis,
        /*in*/ PIUSBSLAVEDRIVER Owner)
{
    PUDC this = pUDCfromSLAVE(pThis);
    SCODE sc;

    /* Enforce single-use rule
     */
    if (this->Driver) {
        DBGME(3,_tprintf(_T("TakeOwnership already %x.\n"),this->Driver));
        return E_ALREADY_EXISTS;
    }

    /* Check argument
     * NB: Do this secondly, this way one can just check if its owned already
     */
    if (Owner == NULL) {
        DBGME(3,_tprintf(_T("TakeOwnership NULL.\n")));
        return E_INVALID_PARAMETER;
    }

    /* Make sure its the right thing
     */
    sc = Owner->v->QueryInterface(Owner,
                                  &IID_IUsbSlaveDriver,
                                  (void **) &Owner);
    if (sc == S_OK) {
        /* That took a ref also
         */
        this->Driver = Owner;
        UdcInit( (PCDEVICE) this);
    }
    DBGME(3,_tprintf(_T("TakeOwnership QI %x.\n"),sc));
    return sc;
}

SCODE UdcSlaveReleaseOwnership (
        /*in*/ PIUSBSLAVECONTROLLER pThis,
        /*in*/ PIUSBSLAVEDRIVER Owner)
{
    PUDC this = pUDCfromSLAVE(pThis);

    /* Enforce single-use rule
     */
    if (this->Driver == NULL)
        return E_ALREADY_EXISTS;

    /* Check argument
     * NB: Do this secondly, this way one can just check if its free
     */
    if (Owner == NULL)
        return E_INVALID_PARAMETER;

    /* Make sure its the right thing
     */
    if (Owner != this->Driver)
        return E_INVALID_PARAMETER;

    /* Stop it then
     */
    UdcStop(this);
    this->Driver = NULL;

    /* That took a ref also
     */
    Owner->v->Release(Owner);

    return S_OK;
}

SCODE UdcSlaveQueryEndpoints (
        /*in*/ PIUSBSLAVECONTROLLER pThis,
        /*out*/ UINT * nControl,
        /*out*/ UINT * nInput,
        /*out*/ UINT * nOutput)
{
    UnusedParameter(pThis);

    /* We got one of each
     */
    if (nControl)
        *nControl = 1;
    if (nInput)
        *nInput = 1;
    if (nOutput)
        *nOutput = 1;

    return S_OK;
}

SCODE UdcSlaveQueryEndpointProperties (
        /*in*/ PIUSBSLAVECONTROLLER pThis,
        /*in*/ USBD_ENDPOINT_TYPE EndpointType,
        /*in*/ UINT EndpointIndex,
        /*out*/ PUSBD_ENDPOINT_PROPERTIES Properties)
{
    UnusedParameter(pThis);

    /* We got one of each
     */
    if (EndpointIndex > 0)
        return E_OBJECT_NOT_FOUND;

    switch (EndpointType) {

    case USBD_ENDPOINT_CONTROL:
        Properties->MaximumPacketSize = 8;
        Properties->PhysicalIndex = 0;
        Properties->IsIsochronous = FALSE;
        break;

    case USBD_ENDPOINT_INPUT:
        Properties->MaximumPacketSize = MAXIN;
        Properties->PhysicalIndex = 1;
        Properties->IsIsochronous = FALSE;
        break;

    case USBD_ENDPOINT_OUPUT:
        Properties->MaximumPacketSize = MAXOUT;
        Properties->PhysicalIndex = 2;
        Properties->IsIsochronous = FALSE;
        break;

    default:
        return E_INVALID_PARAMETER;
    }
    return S_OK;
}

const struct IUsbSlaveControllerVtbl UdcSlaveVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    UdcSlaveQueryInterface,
    UdcSlaveAddRef,
    UdcSlaveRelease,
    UdcSlaveTakeOwnership,
    UdcSlaveReleaseOwnership,
    UdcSlaveQueryEndpoints,
    UdcSlaveQueryEndpointProperties
};

/* IFile interface
 */

/*
 * Read from a file
 */
SCODE MCT
UdcReadAt(PCDEVICE pThis, UINT64 Position, 
          BYTE *Buffer, UINT ByteCount, PUINT pSizeRead)
{
    PUDC this = pUDC(pThis);
    UINT Avail;
    SCODE sc;

    UnusedParameter(Position);

    DBGME(0,_tprintf(_T("UdcReadAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) Buffer,
                     ByteCount, (ADDRESS) pSizeRead));

    /* Default ret
     */
    if (pSizeRead)
        *pSizeRead = 0;

    /* Backoff if not connected
     */
    if (this->ConnectionState != UDC_CON_CONNECTED) {
        return E_NOT_READY;
    }

    /* Wait till we got some
     */
    while (CQueueEmpty(this->OneQue)) {

        /* Was the cable unplugged ?
         */
        if (this->ConnectionState != UDC_CON_CONNECTED) {
            sc = E_NOT_READY;
            goto Done;
        }

        /* Nope, wait for data
         */
        sc = Condition_Wait( &this->OneQue.WaitPoint,
                             NULL,
                             &this->OneQue.TimeOut);

        /* If timeout expired we are done
         */
        if (sc == E_TIMED_OUT) {
            Log( ('t' | 0x900) | ((CQueueEmpty(this->OneQue)) << 16) );
            goto Done;
        }
    }

    /* How much can we get ?
     */
    Avail = this->OneQue.Head; /* cache against interrupts */
    Avail = CQueueCountEval(Avail,this->OneQue.Tail,this->OneQue.Size);

    ByteCount = (ByteCount > Avail) ? Avail : ByteCount;

    /* say so
     */
    if (pSizeRead)
        *pSizeRead = ByteCount;

    /* copy data, first part (towards the end of buffer)
     */
    Avail = this->OneQue.Size - 1 - this->OneQue.Tail;
    if (Avail > ByteCount) 
        Avail = ByteCount;
    if (Avail) {
        memcpy( Buffer,
                & this->OneQue.Buffer[ this->OneQue.Tail + 1],
                Avail);
        Buffer += Avail;
        this->OneQue.Tail += Avail;
        ByteCount -= Avail;
    }

    /* copy data, second part (start of buf)
     */
    if (ByteCount) {
        memcpy( Buffer,
                this->OneQue.Buffer,
                ByteCount );
        this->OneQue.Tail = ByteCount - 1;
    }

    /* Tail ptr updated already, done.
     */
    sc = NO_ERROR;
 Done:

    return sc;
}


/*
 * Write to a file
 */
SCODE MCT
UdcWriteAt(PCDEVICE pThis, UINT64 Position, 
           const BYTE *Buffer, UINT ByteCount, PUINT pSizeWritten)
{
    PUDC this = pUDC(pThis);
    SCODE sc = S_OK;

    UnusedParameter(Position);

    /* Default ret
     */
    if (pSizeWritten)
      *pSizeWritten = 0;

    /* Backoff if not connected
     */
    if (this->ConnectionState != UDC_CON_CONNECTED) {
        return E_NOT_READY;
    }

    Log( ('w' | 0x900) | (this->SwFifo.Count << 16) );

    /* Set the buffer pointer first, then the counter.
     * This way of the ISR triggets it gets it immediately
     */
    this->XmitTwoDataPtr = (UINT8*) Buffer;
    this->XmitTwoCount = ByteCount;

    /* Move it itno the sw fifo, if it fits
     */
    LoadSwFifo1(this);

    /* Well it didnt.  Wait here until its done (by the ISR)
     */
    while (this->XmitTwoCount) {
        /* Check for an unplugged cable
         */
        if (this->ConnectionState != UDC_CON_CONNECTED) {
            this->XmitTwoCount = 0;
            sc = E_NOT_READY;
            goto Done;
        }

        /* Wait till its [on its way] out.
         */
        sc = Condition_Wait(&this->XmitTwoWaitPoint,
                            NULL,
                            &this->XmitTwoTimeOut);

        /* Shouldnt use xmit timeouts really, but.
         */
        if (sc == E_TIMED_OUT) {
            ByteCount -= this->XmitTwoCount;
            break;
        }
    }

    /* Say what we did if anything
     */
    if (pSizeWritten)
      *pSizeWritten = ByteCount;

 Done:
    return sc;
}

/* Get/Set a file's size
 */
#define UdcGetSize NULL

SCODE MCT UdcSetSize(PCDEVICE pThis, UINT64 Size)
{
    PUDC this = pUDC(pThis);

    /* This is extremely wacky, but..
     */
    UINT32 *p = (UINT32*)&Size;

    switch (*p) {
    case 0: /* set recv timeout */
        Int32ToInt64(this->OneQue.TimeOut,TIME_RELATIVE((INT32)p[1]));
        break;
    case 1: /* set xmit timeout */
        Int32ToInt64(this->XmitTwoTimeOut,TIME_RELATIVE((INT32)p[1]));
        break;
    default:
        return E_INVALID_PARAMETER;
    }
    return NO_ERROR;
}

#if LOGGING

void MCT
UdcThreadLoop( PCDEVICE pThis )
{
    PUDC this = pUDC(pThis);
    TIME t;
    int i = 0;
    UINT IntCo = this->nInts;

    for (;;) {
        Int32ToInt64(t, TIME_RELATIVE(TIME_SECONDS(1)));
        SleepUntil(Int64TimesInt32(t, 1));
        if (IntCo != this->nInts) {
            IntCo = this->nInts;
            UdcState(this,i++);
        } else
            i = 0;
    }    
}
#else
#define UdcThreadLoop NULL
#endif

SCODE UdcQueryInterface (struct CDevice *pThis, REFIID Iid, void* *ppNew)
{
    if (UuidCmp(&IID_IFile,Iid) || UuidCmp(&IID_IUnknown,Iid)
        || UuidCmp(&IID_IDevice,Iid))
    {
        AtomicInc(&pThis->RefCnt);
        *ppNew = (PIUNKNOWN) pThis;
        return NO_ERROR;
    }
    if (UuidCmp(&IID_IUsbSlaveController,Iid)) {
        /* Use a single refcount
         */
        AtomicInc(&pThis->RefCnt);
        *ppNew = (PIUNKNOWN) pSLAVEfromUDC( pUDC(pThis) );
        return NO_ERROR;
    }
    *ppNew = NULL;
    return E_NO_INTERFACE;
}

static struct CDeviceVtbl UdcDriverVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    UdcQueryInterface,
    NULL,
    NULL,
    UdcReadAt, 
    UdcWriteAt, 
    UdcSetSize, 
    UdcGetSize,
    UdcIsr,
    NULL, /* HandleInterrupt, but we dont need a thread */
    UdcInit,
    UdcThreadLoop  /* ThreadLoop, no need unless LOGGING is on */
};

/* Exported constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE UdcNew(IO_PORT PortBase, PUINT pIrq)
{
    PUDC this;
    PIHEAP pH = CurrentHeap();

    UnusedParameter(pIrq);

    /* Make sure the compiler is not causing trouble
     */
    assert(sizeof(USB_DEVICE_DESCRIPTOR) == 18);
    assert(sizeof(USB_CONFIGURATION_DESCRIPTOR) == 9);
    assert(sizeof(USB_INTERFACE_DESCRIPTOR) == 9);
    assert(sizeof(USB_ENDPOINT_DESCRIPTOR) == 7);

    /* Allocate our object
     */
    this = (PUDC) pH->v->Alloc(pH,
                               HEAP_ZERO_MEMORY,
                               sizeof(*this),
                               0);
    if (!this)
      return NULL;

    /* Standard part, make IUnknown code happy
     */
    this->CDevice.v = &UdcDriverVtable;
    this->CDevice.RefCnt = 1;
    this->v = &UdcSlaveVtbl;

    /* Initialize what of our state is not zero
     */
    this->Registers = (PUdc) PortBase;
    Condition_Init(&this->XmitTwoWaitPoint);
    CQueueInit(this->OneQue);

    this->XmitTwoTimeOut = TIME_FOREVER;

    this->DeviceStatus = USB_ST_SELF_POWERED;

#if DONT_SPLIT_THE_UDC
    this->Driver = IreneHack();
#endif

#if LOGGING
    DBGME(3,_tprintf(_T("   UDC log at %x %x.%x\n"),
                     (ADDRESS)TheLog, (ADDRESS)&LogHead, (ADDRESS)&LogTail));
#endif

    return (PCDEVICE) this;
}

/* ** 
*/
SCODE __cdecl UdcConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("UdcConstructor(%x %s %x %x %x)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) UdcNew((IO_PORT) RegisterBase, &Interrupt);

    DBGME(0,_tprintf(_T("UdcConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UdcConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UdcDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UdcConstructor
};

static struct _DRIVERCOB UdcDriverCobObject = { &UdcDriverCobVtable, 1 };

PIUNKNOWN UdcCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UdcDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UdcDriverVtable, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    /* Null the thread loop, to save a useless thread
     */
    UdcDriverVtable.ThreadLoop = UdcThreadLoop;

    return (PIUNKNOWN) &UdcDriverCobObject;
}
