/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbdkeybd.c
 *
 * Purpose:
 *
 *    Device driver for a slave USB keybd (meaning, the keybd itself).
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbslave.h>

/* Our device object
 */
typedef struct _USBDKEYBD {
    struct CDevice CDevice;
    const struct IUsbSlaveDriverVtbl *v;
    PIUSBSLAVECONTROLLER Controller;
    PIFILE ControllerIfile;
    MUTEX Lock;
    BYTE RequestData[64];
} USBDKEYBD, *PUSBDKEYBD;

#define pUSBDKEYBD(_t_) ((PUSBDKEYBD)(_t_))
#define pUSBDKEYBDfromSLAVEDRIVER(_t_) ((PUSBDKEYBD)(((BYTE *)_t_) - offsetof(USBDKEYBD,v)))
#define pSLAVEDRIVERfromUSBDKEYBD(_t_) ((PIUSBSLAVEDRIVER)(&(_t_)->v))

/* Descriptors per USB.
 */
#define HAS_STRINGS 1
PRIVATE USB_STRING String1[] = 
    { (USB_DT_STRING<<8) | (66),
      /* "Microsoft Natural Keyboard Elite" */
      0x4d, 0x69, 0x63, 0x72, 0x6f, 0x73, 0x6f, 0x66, 0x74, 0x20,
      0x4e, 0x61, 0x74, 0x75, 0x72, 0x61, 0x6c, 0x20, 0x4b, 0x65,
      0x79, 0x62, 0x6f, 0x61, 0x72, 0x64, 0x20, 0x45, 0x6c, 0x69,
      0x74, 0x65
    };

PRIVATE USB_STRING_DESCRIPTOR Strings[] =
{ String1
};
#define nStrings (sizeof Strings / sizeof Strings[0] )

#if HAS_STRINGS

/* String descriptors
 * Index 0 is the array of LANGIDs
 * We only do English, regardless.
 */
PRIVATE UINT16 String0[] = { (USB_DT_STRING<<8) | (4), 0x409 };

#endif

PRIVATE USB_DEVICE_DESCRIPTOR MyDeviceDescriptor = {
    sizeof(USB_DEVICE_DESCRIPTOR),
    USB_DT_DEVICE,
    Con2(USB_v2_0),
    0, 0, 0,          /* class,subclass,protocol */
    8,                /* max packet size */
    Con2(0x45e), Con2(0xb), Con2(0x207),   /* vendor,product,device*/
    0, 1, 0,          /* ix of manuf,product,serial number */
    1                 /* num configurations */
};

#define sizeof_MyConfigurationDescriptors 34

PRIVATE _PACKED struct {
    USB_CONFIGURATION_DESCRIPTOR Config;
    USB_INTERFACE_DESCRIPTOR Iface;
    USB_HID_DESCRIPTOR Hid;
    USB_ENDPOINT_DESCRIPTOR Ep1;
} MyConfigurationDescriptors = 
    {
      {
        sizeof(USB_CONFIGURATION_DESCRIPTOR),
        USB_DT_CONFIG,
        Con2(sizeof_MyConfigurationDescriptors),   /* total length of all descs returned */
        1,               /* num interfaces */
        1,               /* config value */
        0,               /* ix of config */
        USB_AT_BUS_POWERED|USB_AT_REMOTE_WAKEUP,  /* attributes */
        100/2            /* max power */
      },{
        sizeof(USB_INTERFACE_DESCRIPTOR),
        USB_DT_IFACE,
        0,               /* iface no */
        0,               /* alternate */
        1,               /* num endp */
        USB_DC_HID, USB_DS_BOOT, USB_DP_KEYBOARD, /* class,subclass,protocol */
        0                /* ix of iface */
      },{
        sizeof(USB_HID_DESCRIPTOR),
        USB_DT_HID,
        Con2(0x110),    /* BcdHID */
        0,              /* CountryCode */
        1,              /* NumDescriptors to follow */
        USB_DT_REPORT,  /* ReportDescriptorType */
        Con2(63)        /* ReportDescriptorLength */
      },{
        sizeof(USB_ENDPOINT_DESCRIPTOR),
        USB_DT_ENDP,
        USB_EA_IN | 2,   /* address */
        USB_AE_INTERRUPT,/* attrib */
        Con2(8),         /* max pkt size */
        10               /* interval () */
      }
};

#define MyInputReportSize 8

PRIVATE UINT8 MyReportDescriptor[63] = {
    0x05, 0x01,  // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,  // USAGE (Keyboard)
    0xa1, 0x01,  // COLLECTION (Application)
      0x05, 0x07,  // USAGE_PAGE (Keyboard/Keypad)
      0x19, 0xe0,  // USAGE_MINIMUM (224)
      0x29, 0xe7,  // USAGE_MAXIMUM (231)
      0x15, 0x00,  // LOGICAL_MINIMUM (0)
      0x25, 0x01,  // LOGICAL_MAXIMUM (1)
      0x95, 0x08,  // REPORT_COUNT (8)
      0x75, 0x01,  // REPORT_SIZE (1)
      0x81, 0x02,  // INPUT (Data, Var, Abs)
      0x95, 0x01,  // REPORT_COUNT (1)
      0x75, 0x08,  // REPORT_SIZE (8)
      0x81, 0x01,  // INPUT (Constant)
      0x95, 0x05,  // REPORT_COUNT (5)
      0x75, 0x01,  // REPORT_SIZE (1)
      0x05, 0x08,  // USAGE_PAGE (LEDs)
      0x19, 0x01,  // USAGE_MINIMUM (1)
      0x29, 0x05,  // USAGE_MAXIMUM (5)
      0x91, 0x02,  // OUTPUT (Data, Var, Abs)
      0x95, 0x01,  // REPORT_COUNT (1)
      0x75, 0x03,  // REPORT_SIZE (3)
      0x91, 0x01,  // OUTPUT (Constant)
      0x95, 0x06,  // REPORT_COUNT (6)
      0x75, 0x08,  // REPORT_SIZE (8)
      0x15, 0x00,  // LOGICAL_MINIMUM (0)
      0x25, 0x65,  // LOGICAL_MAXIMUM (101)
      0x05, 0x07,  // USAGE_PAGE (Keyboard/Keypad)
      0x19, 0x00,  // USAGE_MINIMUM (0)
      0x29, 0x65,  // USAGE_MAXIMUM (101)
      0x81, 0x00,  // INPUT (Data, Array, Abs)
    0xc0         // END_COLLECTION
};

SCODE UsbdkeybdDriverQueryInterface (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ REFIID iid,
        /*out*/ void** ppObject)
{
    PUSBDKEYBD this = pUSBDKEYBDfromSLAVEDRIVER(pThis);
    return this->CDevice.v->QueryInterface((PCDEVICE)this,
                                           iid,
                                           ppObject);
}

UINT UsbdkeybdDriverAddRef (
        /*in*/ PIUSBSLAVEDRIVER pThis)
{
    PUSBDKEYBD this = pUSBDKEYBDfromSLAVEDRIVER(pThis);
    return this->CDevice.v->AddRef((PCDEVICE)this);
}

UINT UsbdkeybdDriverRelease (
        /*in*/ PIUSBSLAVEDRIVER pThis)
{
    PUSBDKEYBD this = pUSBDKEYBDfromSLAVEDRIVER(pThis);
    return this->CDevice.v->Release((PCDEVICE)this);
}

SCODE UsbdkeybdDriverGetDescriptor (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ USBD_DESCRIPTOR_TYPE DescriptorType,
        /*in*/ UINT Index,
        /*out*/ BYTE ** ppDescriptor,
        /*out*/ UINT * pSize)
{
    PUSBDKEYBD this = pUSBDKEYBDfromSLAVEDRIVER(pThis);
    UnusedParameter(this);

    /* Validate arguments
     */
    if ((ppDescriptor == NULL) || (pSize == NULL))
        return E_INVALID_PARAMETER;

    switch (DescriptorType) {
    case USBD_DESCRIPTOR_DEVICE:
        *ppDescriptor = (BYTE*) &MyDeviceDescriptor;
        *pSize        = sizeof MyDeviceDescriptor;
        break;
    case USBD_DESCRIPTOR_CONFIGURATION:
        *ppDescriptor = (BYTE*) &MyConfigurationDescriptors;
        *pSize        = sizeof_MyConfigurationDescriptors;
        break;
    case USBD_DESCRIPTOR_REPORT:
        *ppDescriptor = (UINT8*) &MyReportDescriptor;
        *pSize        = sizeof MyReportDescriptor;
        break;
    case USBD_DESCRIPTOR_STRING:
#if HAS_STRINGS
        if (Index == 0) {
            *ppDescriptor = (UINT8*) &String0;
        } else if (Index <= nStrings) {
            *ppDescriptor = (UINT8*) Strings[Index-1];
        } else
            goto error;
        *pSize = (*ppDescriptor)[0];
        break;
#else
        goto error;
#endif
    case USBD_DESCRIPTOR_INTERFACE:
    case USBD_DESCRIPTOR_ENDPOINT:
    case USBD_DESCRIPTOR_DEVICE_QUALIFIER:
    case USBD_DESCRIPTOR_HID:
    case USBD_DESCRIPTOR_PHYSICAL:
    default:
    error:
        return E_NOT_IMPLEMENTED;
    }

    return S_OK;
}

SCODE UsbdkeybdDriverSetDescriptor (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ USBD_DESCRIPTOR_TYPE DescriptorType,
        /*in*/ UINT Index,
        /*in*/ const BYTE * Descriptor,
        /*in*/ UINT Size)
{
    PUSBDKEYBD this = pUSBDKEYBDfromSLAVEDRIVER(pThis);
    UnusedParameter(this);
    return E_NOT_IMPLEMENTED;
}

SCODE UsbdkeybdDriverHandleRequest (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ const PUSB_REQUEST_struct RequestPacket,
        /*inout*/ BYTE ** ppRequestData,
        /*inout*/ UINT * pSize)
{
    PUSBDKEYBD this = pUSBDKEYBDfromSLAVEDRIVER(pThis);
    BYTE *pData = *ppRequestData;

    switch (RequestPacket->Request) {

    case USB_R_GET_REPORT:      // 0x01
    case USB_R_GET_IDLE:        // 0x02
    case USB_R_GET_PROTOCOL:    // 0x03
        goto error;//NYI

    case USB_R_SET_REPORT:      // 0x09
        /* Did we get all we needed ?
         */
        if (pData == NULL) {
            /* Uhmm, definitely not
             * See how much the host wants to send us, trim if necessary
             */
            UINT Count = Get2(RequestPacket->Length);
            if (Count > sizeof this->RequestData)
                Count = sizeof this->RequestData;

            *pSize = Count;
            *ppRequestData = this->RequestData;
            return S_BUFFER_TOO_SMALL;
        }

        /* Got some data, all is well
         * BUGBUG Handle fragmented case for large xfers
         */
        break;

    case USB_R_SET_IDLE:        // 0x0a
    case USB_R_SET_PROTOCOL:    // 0x0b
        break;

    default:
    error:
        return E_NOT_IMPLEMENTED;
    }

    return S_OK;
}

const struct IUsbSlaveDriverVtbl UsbdkeybdSlaveDriverVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    UsbdkeybdDriverQueryInterface,
    UsbdkeybdDriverAddRef,
    UsbdkeybdDriverRelease,
    UsbdkeybdDriverGetDescriptor,
    UsbdkeybdDriverSetDescriptor,
    UsbdkeybdDriverHandleRequest
};

/* CDevice interface
 */

/*
 * Write to a file (send to host)
 */
SCODE MCT
UsbdkeybdWriteAt(PCDEVICE pThis, UINT64 Position, 
           const BYTE *Buffer, UINT ByteCount, PUINT pSizeWritten)
{
    PUSBDKEYBD this = pUSBDKEYBD(pThis);
    SCODE sc;

    UnusedParameter(Position);

    Mutex_Lock(&this->Lock);

    // BUGBUG split into keybd report pieces?

    sc = this->ControllerIfile->v->WriteAt(this->ControllerIfile,
                                           UINT64_ZERO,
                                           Buffer,
                                           ByteCount,
                                           pSizeWritten);

    Mutex_Unlock(&this->Lock);

    return sc;
}

/* Get/Set a file's size
 */
SCODE MCT UsbdkeybdGetSize( PCDEVICE pThis, PUINT64 pSize)
{
    UnusedParameter(pThis);

    Uint32ToUint64(*pSize,MyInputReportSize);
    return S_OK;
}

SCODE MCT UsbdkeybdSetSize(PCDEVICE pThis, UINT64 Size)
{
    PUSBDKEYBD this = pUSBDKEYBD(pThis);

    return this->ControllerIfile->v->SetSize(this->ControllerIfile,
                                             Size);

}

SCODE UsbdkeybdQueryInterface (struct CDevice *pThis, REFIID Iid, void* *ppNew)
{
    if (UuidCmp(&IID_IFile,Iid) || UuidCmp(&IID_IUnknown,Iid)
        || UuidCmp(&IID_IDevice,Iid))
    {
        AtomicInc(&pThis->RefCnt);
        *ppNew = (PIUNKNOWN) pThis;
        return NO_ERROR;
    }
    if (UuidCmp(&IID_IUsbSlaveDriver,Iid)) {
        /* Use a single refcount
         */
        AtomicInc(&pThis->RefCnt);
        *ppNew = (PIUNKNOWN) pSLAVEDRIVERfromUSBDKEYBD( pUSBDKEYBD(pThis) );
        return NO_ERROR;
    }
    *ppNew = NULL;
    return E_NO_INTERFACE;
}

static struct CDeviceVtbl UsbdkeybdCDeviceVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    UsbdkeybdQueryInterface,
    NULL,
    NULL, //BUGBUG leaks two refs to the controller!!
    NULL,
    UsbdkeybdWriteAt, 
    UsbdkeybdSetSize, 
    UsbdkeybdGetSize,
    NULL,
    NULL,
    NULL,
    NULL
};

/* Exported constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE UsbdkeybdNew(const _TCHAR *ControllerName)
{
    PUSBDKEYBD this;
    PIUSBSLAVECONTROLLER Controller = NULL;
    PIFILE ControllerIfile = NULL;
    SCODE sc;
    PIHEAP pH = CurrentHeap();

    /* Make sure the controller exists, and is of the right type
     */
    sc = BindToObject(CurrentNameSpace(),
                      ControllerName,
                      0,
                      &IID_IUsbSlaveController,
                      (void **)&Controller);

    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbdkeybdConstructor %s not found\n"),ControllerName));
        return NULL;
    }

    /* Get the IFile as well
     */
    sc = Controller->v->QueryInterface(Controller,
                                       &IID_IFile,
                                       (void **)&ControllerIfile);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbdkeybdConstructor %s not an IFILE\n"),ControllerName));
        goto Error;
    }

    /* Allocate our object
     */
    this = (PUSBDKEYBD) pH->v->Alloc(pH,
                               HEAP_ZERO_MEMORY,
                               sizeof(*this),
                               0);
    if (!this) {
        DBGME(3,_tprintf(_T("UsbdkeybdConstructor no mem\n")));
        goto Error;
    }

    /* Standard part, make IUnknown code happy
     */
    this->CDevice.v = &UsbdkeybdCDeviceVtable;
    this->CDevice.RefCnt = 1;

    /* Initialize what of our state is not zero
     */
    this->v = &UsbdkeybdSlaveDriverVtbl;
    this->Controller = Controller;
    this->ControllerIfile = ControllerIfile;

    Mutex_Init(&this->Lock);

    /* Take ownership of the controller
     * NB: This also gets it started
     */
    sc = Controller->v->TakeOwnership(Controller,
                                      pSLAVEDRIVERfromUSBDKEYBD(this));
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbdkeybdConstructor cant own %s sc=%x\n"),
                         ControllerName,sc));
        pH->v->Free(pH,0,this);
        this = NULL;
        goto Error;
    }

    return (PCDEVICE) this;

 Error:
    if (Controller)
        Controller->v->Release(Controller);
    if (ControllerIfile)
        ControllerIfile->v->Release(ControllerIfile);

    return NULL;
}

/* ** 
*/
SCODE __cdecl UsbdkeybdConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        const _TCHAR * ControllerName,
        ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("UsbdkeybdConstructor(%x %s %x %x %x %s)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk, ControllerName));

    pUnk = (PIUNKNOWN) UsbdkeybdNew(ControllerName);

    DBGME(0,_tprintf(_T("UsbdkeybdConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UsbdkeybdConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UsbdkeybdCDeviceCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UsbdkeybdConstructor
};

static struct _DRIVERCOB UsbdkeybdDriverCobObject = { &UsbdkeybdCDeviceCobVtable, 1 };

PIUNKNOWN UsbdkeybdCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UsbdkeybdCDeviceCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UsbdkeybdCDeviceVtable, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    /* Null the thread loop, to save a useless thread
     */
    UsbdkeybdCDeviceVtable.ThreadLoop = NULL;

    return (PIUNKNOWN) &UsbdkeybdDriverCobObject;
}
