/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbdmouse.c
 *
 * Purpose:
 *
 *    Device driver for a slave USB mouse (meaning, the mouse itself).
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbslave.h>

/* Our device object
 */
typedef struct _USBDMOUSE {
    struct CDevice CDevice;
    const struct IUsbSlaveDriverVtbl *v;
    PIUSBSLAVECONTROLLER Controller;
    PIFILE ControllerIfile;
    MUTEX Lock;
    BYTE RequestData[64];
} USBDMOUSE, *PUSBDMOUSE;

#define pUSBDMOUSE(_t_) ((PUSBDMOUSE)(_t_))
#define pUSBDMOUSEfromSLAVEDRIVER(_t_) ((PUSBDMOUSE)(((BYTE *)_t_) - offsetof(USBDMOUSE,v)))
#define pSLAVEDRIVERfromUSBDMOUSE(_t_) ((PIUSBSLAVEDRIVER)(&(_t_)->v))

/* Descriptors per USB.
 */
#define HAS_STRINGS 0

PRIVATE USB_DEVICE_DESCRIPTOR MyDeviceDescriptor = {
    sizeof(USB_DEVICE_DESCRIPTOR),
    USB_DT_DEVICE,
    Con2(USB_v1_0),
    0, 0, 0, /* class,subclass,protocol */
    8,                /* max packet size */
    Con2(0x45e), Con2(0x9), Con2(0x105),   /* vendor,product,device*/
    0, 0, 0,          /* ix of manuf,product,serial number */
    1                 /* num configurations */
};

#define sizeof_MyConfigurationDescriptors 34

PRIVATE _PACKED struct {
    USB_CONFIGURATION_DESCRIPTOR Config;
    USB_INTERFACE_DESCRIPTOR Iface;
    USB_HID_DESCRIPTOR Hid;
    USB_ENDPOINT_DESCRIPTOR Ep1;
} MyConfigurationDescriptors = 
    {
      {
        sizeof(USB_CONFIGURATION_DESCRIPTOR),
        USB_DT_CONFIG,
        Con2(sizeof_MyConfigurationDescriptors),   /* total length of all descs returned */
        1,               /* num interfaces */
        1,               /* config value */
        0,               /* ix of config */
        USB_AT_SELF_POWERED|USB_AT_REMOTE_WAKEUP,  /* attributes */
        1                     /* max power */
      },{
        sizeof(USB_INTERFACE_DESCRIPTOR),
        USB_DT_IFACE,
        0,               /* iface no */
        0,               /* alternate */
        1,               /* num endp */
        USB_DC_HID, USB_DS_BOOT, USB_DP_MOUSE, /* class,subclass,protocol */
        0                /* ix of iface */
      },{
        sizeof(USB_HID_DESCRIPTOR),
        USB_DT_HID,
        Con2(0x100),    /* BcdHID */
        0,              /* CountryCode */
        1,              /* NumDescriptors to follow */
        USB_DT_REPORT,  /* ReportDescriptorType */
        Con2(74)        /* ReportDescriptorLength */
      },{
        sizeof(USB_ENDPOINT_DESCRIPTOR),
        USB_DT_ENDP,
        USB_EA_IN | 2,   /* address */
        USB_AE_INTERRUPT,/* attrib */
        Con2(4),         /* max pkt size */
        10               /* interval () */
      }
};

#define MyInputReportSize 4

PRIVATE UINT8 MyReportDescriptor[74] = {
    5,0x1,     // USAGE_PAGE (Generic Desktop)
    0x9,0x2,   // USAGE (Mouse)
    0xa1,0x1,  // COLLECTION (Application)
     0x9,0x1,   // USAGE (Pointer)
     0xa1,0x0,  // COLLECTION (Physical)
      0x5,0x9,   // USAGE_PAGE (Button)
      0x19,0x1,  // USAGE_MIN (Button 1)
      0x29,0x3,  // USAGE_MAX (Button 3)
      0x15,0x0,  // LOGICAL_MINIMUM (0)
      0x25,0x1,  // LOGICAL_MAXIMUM (1)
      0x75,0x1,  // REPORT_SIZE (1)
      0x95,0x3,  // REPORT_COUNT (3)
      0x81,0x2,  // INPUT (Data, Var, Abs)
      0x75,0x5,  // REPORT_SIZE (5)
      0x95,0x1,  // REPORT_COUNT (1)
      0x81,0x1,  // INPUT (Constant)
      0x5,0x1,   // USAGE_PAGE (Generic Desktop)
      0x9,0x30,  // USAGE (X)
      0x9,0x31,  // USAGE (Y)
      0x9,0x38,  // USAGE (Wheel)
      0x15,0x81, // LOGICAL_MINIMUM (-127)
      0x25,0x7f, // LOGICAL_MAXIMUM (127)
      0x75,0x8,  // REPORT_SIZE (8)
      0x95,0x3,  // REPORT_COUNT (3)
      0x81,0x6,  // INPUT (Data, Var, Rel)
     0xc0,      // END_COLLECTION
     0x9,0x3c,  // USAGE (Motion Wakeup)
     0x5,0xff,  // USAGE_PAGE (Vendor Defined)
     0x9,0x1,   // USAGE (??)
     0x15,0x0,  // LOGICAL_MINIMUM (0)
     0x25,0x1,  // LOGICAL_MAXIMUM (1)
     0x75,0x1,  // REPORT_SIZE (1)
     0x95,0x2,  // REPORT_COUNT (2)
     0xb1,0x22, // FEATURE (Data, Var, Abs, No Preferred)
     0x75,0x6,  // REPORT_SIZE (6)
     0x95,0x1,  // REPORT_COUNT (1)
     0xb1,0x1,  // FEATURE (Constant)
    0xc0       // END_COLLECTION
};

SCODE UsbdmouseDriverQueryInterface (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ REFIID iid,
        /*out*/ void** ppObject)
{
    PUSBDMOUSE this = pUSBDMOUSEfromSLAVEDRIVER(pThis);
    return this->CDevice.v->QueryInterface((PCDEVICE)this,
                                           iid,
                                           ppObject);
}

UINT UsbdmouseDriverAddRef (
        /*in*/ PIUSBSLAVEDRIVER pThis)
{
    PUSBDMOUSE this = pUSBDMOUSEfromSLAVEDRIVER(pThis);
    return this->CDevice.v->AddRef((PCDEVICE)this);
}

UINT UsbdmouseDriverRelease (
        /*in*/ PIUSBSLAVEDRIVER pThis)
{
    PUSBDMOUSE this = pUSBDMOUSEfromSLAVEDRIVER(pThis);
    return this->CDevice.v->Release((PCDEVICE)this);
}

SCODE UsbdmouseDriverGetDescriptor (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ USBD_DESCRIPTOR_TYPE DescriptorType,
        /*in*/ UINT Index,
        /*out*/ BYTE ** ppDescriptor,
        /*out*/ UINT * pSize)
{
    PUSBDMOUSE this = pUSBDMOUSEfromSLAVEDRIVER(pThis);
    UnusedParameter(this);

    /* Validate arguments
     */
    if ((ppDescriptor == NULL) || (pSize == NULL))
        return E_INVALID_PARAMETER;

    switch (DescriptorType) {
    case USBD_DESCRIPTOR_DEVICE:
        *ppDescriptor = (BYTE*) &MyDeviceDescriptor;
        *pSize        = sizeof MyDeviceDescriptor;
        break;
    case USBD_DESCRIPTOR_CONFIGURATION:
        *ppDescriptor = (BYTE*) &MyConfigurationDescriptors;
        *pSize        = sizeof_MyConfigurationDescriptors;
        break;
    case USBD_DESCRIPTOR_REPORT:
        *ppDescriptor = (UINT8*) &MyReportDescriptor;
        *pSize        = sizeof MyReportDescriptor;
        break;
    case USBD_DESCRIPTOR_STRING:
#if HAS_STRINGS
        if (Index == 0) {
            *ppDescriptor = (UINT8*) &String0;
        } else if (Index <= nStrings) {
            *ppDescriptor = (UINT8*) Strings[Index-1];
        } else
            goto error;
        *pSize = (*ppDescriptor)[0];
        break;
#else
        goto error;
#endif
    case USBD_DESCRIPTOR_INTERFACE:
    case USBD_DESCRIPTOR_ENDPOINT:
    case USBD_DESCRIPTOR_DEVICE_QUALIFIER:
    case USBD_DESCRIPTOR_HID:
    case USBD_DESCRIPTOR_PHYSICAL:
    default:
    error:
        return E_NOT_IMPLEMENTED;
    }

    return S_OK;
}

SCODE UsbdmouseDriverSetDescriptor (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ USBD_DESCRIPTOR_TYPE DescriptorType,
        /*in*/ UINT Index,
        /*in*/ const BYTE * Descriptor,
        /*in*/ UINT Size)
{
    PUSBDMOUSE this = pUSBDMOUSEfromSLAVEDRIVER(pThis);
    UnusedParameter(this);
    return E_NOT_IMPLEMENTED;
}

SCODE UsbdmouseDriverHandleRequest (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ const PUSB_REQUEST_struct RequestPacket,
        /*inout*/ BYTE ** ppRequestData,
        /*inout*/ UINT * pSize)
{
    PUSBDMOUSE this = pUSBDMOUSEfromSLAVEDRIVER(pThis);
    BYTE *pData = *ppRequestData;

    switch (RequestPacket->Request) {

    case USB_R_GET_REPORT:      // 0x01
    case USB_R_GET_IDLE:        // 0x02
    case USB_R_GET_PROTOCOL:    // 0x03
        goto error;//NYI

    case USB_R_SET_REPORT:      // 0x09
        /* Did we get all we needed ?
         */
        if (pData == NULL) {
            /* Uhmm, definitely not
             * See how much the host wants to send us, trim if necessary
             */
            UINT Count = Get2(RequestPacket->Length);
            if (Count > sizeof this->RequestData)
                Count = sizeof this->RequestData;

            *pSize = Count;
            *ppRequestData = this->RequestData;
            return S_BUFFER_TOO_SMALL;
        }

        /* Got some data, all is well
         * BUGBUG Handle fragmented case for large xfers
         */
        break;

    case USB_R_SET_IDLE:        // 0x0a
    case USB_R_SET_PROTOCOL:    // 0x0b
        break;

    default:
    error:
        return E_NOT_IMPLEMENTED;
    }

    return S_OK;
}

const struct IUsbSlaveDriverVtbl UsbdmouseSlaveDriverVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    UsbdmouseDriverQueryInterface,
    UsbdmouseDriverAddRef,
    UsbdmouseDriverRelease,
    UsbdmouseDriverGetDescriptor,
    UsbdmouseDriverSetDescriptor,
    UsbdmouseDriverHandleRequest
};

/* CDevice interface
 */

/*
 * Write to a file (send to host)
 */
SCODE MCT
UsbdmouseWriteAt(PCDEVICE pThis, UINT64 Position, 
           const BYTE *Buffer, UINT ByteCount, PUINT pSizeWritten)
{
    PUSBDMOUSE this = pUSBDMOUSE(pThis);
    SCODE sc;

    UnusedParameter(Position);

    Mutex_Lock(&this->Lock);

    // BUGBUG split into mouse report pieces?

    sc = this->ControllerIfile->v->WriteAt(this->ControllerIfile,
                                           UINT64_ZERO,
                                           Buffer,
                                           ByteCount,
                                           pSizeWritten);

    Mutex_Unlock(&this->Lock);

    return sc;
}

/* Get/Set a file's size
 */
SCODE MCT UsbdmouseGetSize( PCDEVICE pThis, PUINT64 pSize)
{
    UnusedParameter(pThis);

    Uint32ToUint64(*pSize,MyInputReportSize);
    return S_OK;
}

SCODE MCT UsbdmouseSetSize(PCDEVICE pThis, UINT64 Size)
{
    PUSBDMOUSE this = pUSBDMOUSE(pThis);

    return this->ControllerIfile->v->SetSize(this->ControllerIfile,
                                             Size);

}

SCODE UsbdmouseQueryInterface (struct CDevice *pThis, REFIID Iid, void* *ppNew)
{
    if (UuidCmp(&IID_IFile,Iid) || UuidCmp(&IID_IUnknown,Iid)
        || UuidCmp(&IID_IDevice,Iid))
    {
        AtomicInc(&pThis->RefCnt);
        *ppNew = (PIUNKNOWN) pThis;
        return NO_ERROR;
    }
    if (UuidCmp(&IID_IUsbSlaveDriver,Iid)) {
        /* Use a single refcount
         */
        AtomicInc(&pThis->RefCnt);
        *ppNew = (PIUNKNOWN) pSLAVEDRIVERfromUSBDMOUSE( pUSBDMOUSE(pThis) );
        return NO_ERROR;
    }
    *ppNew = NULL;
    return E_NO_INTERFACE;
}

static struct CDeviceVtbl UsbdmouseCDeviceVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    UsbdmouseQueryInterface,
    NULL,
    NULL, //BUGBUG leaks two refs to the controller!!
    NULL,
    UsbdmouseWriteAt, 
    UsbdmouseSetSize, 
    UsbdmouseGetSize,
    NULL,
    NULL,
    NULL,
    NULL
};

/* Exported constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE UsbdmouseNew(const _TCHAR *ControllerName)
{
    PUSBDMOUSE this;
    PIUSBSLAVECONTROLLER Controller = NULL;
    PIFILE ControllerIfile = NULL;
    SCODE sc;
    PIHEAP pH = CurrentHeap();

    /* Make sure the controller exists, and is of the right type
     */
    sc = BindToObject(CurrentNameSpace(),
                      ControllerName,
                      0,
                      &IID_IUsbSlaveController,
                      (void **)&Controller);

    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbdmouseConstructor %s not found\n"),ControllerName));
        return NULL;
    }

    /* Get the IFile as well
     */
    sc = Controller->v->QueryInterface(Controller,
                                       &IID_IFile,
                                       (void **)&ControllerIfile);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbdmouseConstructor %s not an IFILE\n"),ControllerName));
        goto Error;
    }

    /* Allocate our object
     */
    this = (PUSBDMOUSE) pH->v->Alloc(pH,
                               HEAP_ZERO_MEMORY,
                               sizeof(*this),
                               0);
    if (!this) {
        DBGME(3,_tprintf(_T("UsbdmouseConstructor no mem\n")));
        goto Error;
    }

    /* Standard part, make IUnknown code happy
     */
    this->CDevice.v = &UsbdmouseCDeviceVtable;
    this->CDevice.RefCnt = 1;

    /* Initialize what of our state is not zero
     */
    this->v = &UsbdmouseSlaveDriverVtbl;
    this->Controller = Controller;
    this->ControllerIfile = ControllerIfile;

    Mutex_Init(&this->Lock);

    /* Take ownership of the controller
     * NB: This also gets it started
     */
    sc = Controller->v->TakeOwnership(Controller,
                                      pSLAVEDRIVERfromUSBDMOUSE(this));
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbdmouseConstructor cant own %s sc=%x\n"),
                         ControllerName,sc));
        pH->v->Free(pH,0,this);
        this = NULL;
        goto Error;
    }

    return (PCDEVICE) this;

 Error:
    if (Controller)
        Controller->v->Release(Controller);
    if (ControllerIfile)
        ControllerIfile->v->Release(ControllerIfile);

    return NULL;
}

/* ** 
*/
SCODE __cdecl UsbdmouseConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        const _TCHAR * ControllerName,
        ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("UsbdmouseConstructor(%x %s %x %x %x %s)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk, ControllerName));

    pUnk = (PIUNKNOWN) UsbdmouseNew(ControllerName);

    DBGME(0,_tprintf(_T("UsbdmouseConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UsbdmouseConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UsbdmouseCDeviceCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UsbdmouseConstructor
};

static struct _DRIVERCOB UsbdmouseDriverCobObject = { &UsbdmouseCDeviceCobVtable, 1 };

PIUNKNOWN UsbdmouseCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UsbdmouseCDeviceCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UsbdmouseCDeviceVtable, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    /* Null the thread loop, to save a useless thread
     */
    UsbdmouseCDeviceVtable.ThreadLoop = NULL;

    return (PIUNKNOWN) &UsbdmouseDriverCobObject;
}
