/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbdserial.c
 *
 * Purpose:
 *
 *    Device driver for a slave USB serial line (meaning, the serial line itself).
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <diagnostics.h>
#include <assert.h>
#include <mmhal.h>
#include <drivers\drivers.h>
#include <drivers\usb.h>
#include <usbslave.h>

// BUGBUG get from UDC !!!
#define MAXIN  16  /* Specs says depth of xmit FIFO is 16 */
#define MAXOUT 20  /* Specs says depth of recv FIFO is 20 */

/* Our device object
 */
typedef struct _USBDSERIAL {
    struct CDevice CDevice;
    const struct IUsbSlaveDriverVtbl *v;
    PIUSBSLAVECONTROLLER Controller;
    PIFILE ControllerIfile;
    MUTEX Lock;
    UINT SerialStatus;
    BYTE RequestData[64];
} USBDSERIAL, *PUSBDSERIAL;

#define pUSBDSERIAL(_t_) ((PUSBDSERIAL)(_t_))
#define pUSBDSERIALfromSLAVEDRIVER(_t_) ((PUSBDSERIAL)(((BYTE *)_t_) - offsetof(USBDSERIAL,v)))
#define pSLAVEDRIVERfromUSBDSERIAL(_t_) ((PIUSBSLAVEDRIVER)(&(_t_)->v))

/* Descriptors per USB.
 */
#define HAS_STRINGS 0

#define USB_R_SET_CONTROL_LINE_STATE 0x22
#define USB_CD_ACTIVATE_CARRIER 0x02
#define USB_CD_DTR_PRESENT 0x01

PRIVATE USB_DEVICE_DESCRIPTOR MyDeviceDescriptor = {
    sizeof(USB_DEVICE_DESCRIPTOR),
    USB_DT_DEVICE,
    Con2(USB_v1_0),
    0xff, 0xff, 0xff, /* class,subclass,protocol */
    8,                /* max packet size */
    Con2(0x45e), Con2(0xce), Con2(0),   /* vendor,product,device*/
    0, 0, 0,          /* ix of manuf,product,serial number */
    1                 /* num configurations */
};

#define sizeof_MyConfigurationDescriptors 32

PRIVATE _PACKED struct {
    USB_CONFIGURATION_DESCRIPTOR Config;
    USB_INTERFACE_DESCRIPTOR Iface;
    USB_ENDPOINT_DESCRIPTOR Ep1;
    USB_ENDPOINT_DESCRIPTOR Ep2;
} MyConfigurationDescriptors = 
    {
      {
        sizeof(USB_CONFIGURATION_DESCRIPTOR),
        USB_DT_CONFIG,
        Con2(sizeof_MyConfigurationDescriptors),   /* total length of all descs returned */
        1,               /* num interfaces */
        1,               /* config value */
        0,               /* ix of config */
        USB_AT_SELF_POWERED,  /* attributes */
        1                     /* max power */
      },{
        sizeof(USB_INTERFACE_DESCRIPTOR),
        USB_DT_IFACE,
        0,               /* iface no */
        0,               /* alternate */
        2,               /* num endp */
        0xff, 0xff, 0xff,/* class,subclass,protocol */
        0                /* ix of iface */
      },{
        sizeof(USB_ENDPOINT_DESCRIPTOR),
        USB_DT_ENDP,
        USB_EA_OUT | 1, /* address */
        USB_AE_BULK,    /* attrib */
        Con2(MAXOUT),   /* max pkt size */
        0               /* interval (n/a) */
      },{
        sizeof(USB_ENDPOINT_DESCRIPTOR),
        USB_DT_ENDP,
        USB_EA_IN | 2,  /* address */
        USB_AE_BULK,    /* attrib */
        Con2(MAXIN),    /* max pkt size */
        0               /* interval (n/a) */
      }
};

#define MyInputReportSize 1

PRIVATE UINT8 MyReportDescriptor[1] = {0};

SCODE UsbdserialDriverQueryInterface (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ REFIID iid,
        /*out*/ void** ppObject)
{
    PUSBDSERIAL this = pUSBDSERIALfromSLAVEDRIVER(pThis);
    return this->CDevice.v->QueryInterface((PCDEVICE)this,
                                           iid,
                                           ppObject);
}

UINT UsbdserialDriverAddRef (
        /*in*/ PIUSBSLAVEDRIVER pThis)
{
    PUSBDSERIAL this = pUSBDSERIALfromSLAVEDRIVER(pThis);
    return this->CDevice.v->AddRef((PCDEVICE)this);
}

UINT UsbdserialDriverRelease (
        /*in*/ PIUSBSLAVEDRIVER pThis)
{
    PUSBDSERIAL this = pUSBDSERIALfromSLAVEDRIVER(pThis);
    return this->CDevice.v->Release((PCDEVICE)this);
}

SCODE UsbdserialDriverGetDescriptor (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ USBD_DESCRIPTOR_TYPE DescriptorType,
        /*in*/ UINT Index,
        /*out*/ BYTE ** ppDescriptor,
        /*out*/ UINT * pSize)
{
    PUSBDSERIAL this = pUSBDSERIALfromSLAVEDRIVER(pThis);
    UnusedParameter(this);

    /* Validate arguments
     */
    if ((ppDescriptor == NULL) || (pSize == NULL))
        return E_INVALID_PARAMETER;

    switch (DescriptorType) {
    case USBD_DESCRIPTOR_DEVICE:
        *ppDescriptor = (BYTE*) &MyDeviceDescriptor;
        *pSize        = sizeof MyDeviceDescriptor;
        break;
    case USBD_DESCRIPTOR_CONFIGURATION:
        *ppDescriptor = (BYTE*) &MyConfigurationDescriptors;
        *pSize        = sizeof_MyConfigurationDescriptors;
        break;
    case USBD_DESCRIPTOR_REPORT:
        *ppDescriptor = (UINT8*) &MyReportDescriptor;
        *pSize        = sizeof MyReportDescriptor;
        break;
    case USBD_DESCRIPTOR_STRING:
#if HAS_STRINGS
        if (Index == 0) {
            *ppDescriptor = (UINT8*) &String0;
        } else if (Index <= nStrings) {
            *ppDescriptor = (UINT8*) Strings[Index-1];
        } else
            goto error;
        *pSize = (*ppDescriptor)[0];
        break;
#else
        goto error;
#endif
    case USBD_DESCRIPTOR_INTERFACE:
    case USBD_DESCRIPTOR_ENDPOINT:
    case USBD_DESCRIPTOR_DEVICE_QUALIFIER:
    case USBD_DESCRIPTOR_HID:
    case USBD_DESCRIPTOR_PHYSICAL:
    default:
    error:
        return E_NOT_IMPLEMENTED;
    }

    return S_OK;
}

SCODE UsbdserialDriverSetDescriptor (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ USBD_DESCRIPTOR_TYPE DescriptorType,
        /*in*/ UINT Index,
        /*in*/ const BYTE * Descriptor,
        /*in*/ UINT Size)
{
    PUSBDSERIAL this = pUSBDSERIALfromSLAVEDRIVER(pThis);
    UnusedParameter(this);
    return E_NOT_IMPLEMENTED;
}

SCODE UsbdserialDriverHandleRequest (
        /*in*/ PIUSBSLAVEDRIVER pThis,
        /*in*/ const PUSB_REQUEST_struct RequestPacket,
        /*inout*/ BYTE ** ppRequestData,
        /*inout*/ UINT * pSize)
{
    PUSBDSERIAL this = pUSBDSERIALfromSLAVEDRIVER(pThis);

    switch (RequestPacket->Request) {

    case USB_R_SET_CONTROL_LINE_STATE: // 0x22
        this->SerialStatus = Get2(RequestPacket->Value);
        break;

    default:
        return E_NOT_IMPLEMENTED;
    }

    return S_OK;
}

const struct IUsbSlaveDriverVtbl UsbdserialSlaveDriverVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    UsbdserialDriverQueryInterface,
    UsbdserialDriverAddRef,
    UsbdserialDriverRelease,
    UsbdserialDriverGetDescriptor,
    UsbdserialDriverSetDescriptor,
    UsbdserialDriverHandleRequest
};

/* CDevice interface
 */

/*
 * Read from a file (recv from host)
 */
SCODE MCT
UsbdserialReadAt(PCDEVICE pThis, UINT64 Position, 
                 BYTE *Buffer, UINT ByteCount, PUINT pSizeRead)
{
    PUSBDSERIAL this = pUSBDSERIAL(pThis);
    SCODE sc;

    UnusedParameter(Position);

    Mutex_Lock(&this->Lock);

    sc = this->ControllerIfile->v->ReadAt(this->ControllerIfile,
                                           UINT64_ZERO,
                                           Buffer,
                                           ByteCount,
                                           pSizeRead);

    Mutex_Unlock(&this->Lock);

    return sc;
}

/*
 * Write to a file (send to host)
 */
SCODE MCT
UsbdserialWriteAt(PCDEVICE pThis, UINT64 Position, 
           const BYTE *Buffer, UINT ByteCount, PUINT pSizeWritten)
{
    PUSBDSERIAL this = pUSBDSERIAL(pThis);
    SCODE sc;

    UnusedParameter(Position);

    Mutex_Lock(&this->Lock);

    // BUGBUG split into serial report pieces?

    sc = this->ControllerIfile->v->WriteAt(this->ControllerIfile,
                                           UINT64_ZERO,
                                           Buffer,
                                           ByteCount,
                                           pSizeWritten);

    Mutex_Unlock(&this->Lock);

    return sc;
}

/* Get/Set a file's size
 */
SCODE MCT UsbdserialGetSize( PCDEVICE pThis, PUINT64 pSize)
{
    UnusedParameter(pThis);

    Uint32ToUint64(*pSize,MyInputReportSize);
    return S_OK;
}

SCODE MCT UsbdserialSetSize(PCDEVICE pThis, UINT64 Size)
{
    PUSBDSERIAL this = pUSBDSERIAL(pThis);

    return this->ControllerIfile->v->SetSize(this->ControllerIfile,
                                             Size);

}

SCODE UsbdserialQueryInterface (struct CDevice *pThis, REFIID Iid, void* *ppNew)
{
    if (UuidCmp(&IID_IFile,Iid) || UuidCmp(&IID_IUnknown,Iid)
        || UuidCmp(&IID_IDevice,Iid))
    {
        AtomicInc(&pThis->RefCnt);
        *ppNew = (PIUNKNOWN) pThis;
        return NO_ERROR;
    }
    if (UuidCmp(&IID_IUsbSlaveDriver,Iid)) {
        /* Use a single refcount
         */
        AtomicInc(&pThis->RefCnt);
        *ppNew = (PIUNKNOWN) pSLAVEDRIVERfromUSBDSERIAL( pUSBDSERIAL(pThis) );
        return NO_ERROR;
    }
    *ppNew = NULL;
    return E_NO_INTERFACE;
}

static struct CDeviceVtbl UsbdserialCDeviceVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    UsbdserialQueryInterface,
    NULL,
    NULL, //BUGBUG leaks two refs to the controller!!
    UsbdserialReadAt, 
    UsbdserialWriteAt, 
    UsbdserialSetSize, 
    UsbdserialGetSize,
    NULL,
    NULL,
    NULL,
    NULL
};

/* Exported constructor
 */
EXTERN_C PCDEVICE RTLCALLTYPE UsbdserialNew(const _TCHAR *ControllerName)
{
    PUSBDSERIAL this;
    PIUSBSLAVECONTROLLER Controller = NULL;
    PIFILE ControllerIfile = NULL;
    SCODE sc;
    PIHEAP pH = CurrentHeap();

    /* Make sure the controller exists, and is of the right type
     */
    sc = BindToObject(CurrentNameSpace(),
                      ControllerName,
                      0,
                      &IID_IUsbSlaveController,
                      (void **)&Controller);

    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbdserialConstructor %s not found\n"),ControllerName));
        return NULL;
    }

    /* Get the IFile as well
     */
    sc = Controller->v->QueryInterface(Controller,
                                       &IID_IFile,
                                       (void **)&ControllerIfile);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbdserialConstructor %s not an IFILE\n"),ControllerName));
        goto Error;
    }

    /* Allocate our object
     */
    this = (PUSBDSERIAL) pH->v->Alloc(pH,
                               HEAP_ZERO_MEMORY,
                               sizeof(*this),
                               0);
    if (!this) {
        DBGME(3,_tprintf(_T("UsbdserialConstructor no mem\n")));
        goto Error;
    }

    /* Standard part, make IUnknown code happy
     */
    this->CDevice.v = &UsbdserialCDeviceVtable;
    this->CDevice.RefCnt = 1;

    /* Initialize what of our state is not zero
     */
    this->v = &UsbdserialSlaveDriverVtbl;
    this->Controller = Controller;
    this->ControllerIfile = ControllerIfile;

    Mutex_Init(&this->Lock);

    /* Take ownership of the controller
     * NB: This also gets it started
     */
    sc = Controller->v->TakeOwnership(Controller,
                                      pSLAVEDRIVERfromUSBDSERIAL(this));
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("UsbdserialConstructor cant own %s sc=%x\n"),
                         ControllerName,sc));
        pH->v->Free(pH,0,this);
        this = NULL;
        goto Error;
    }

    return (PCDEVICE) this;

 Error:
    if (Controller)
        Controller->v->Release(Controller);
    if (ControllerIfile)
        ControllerIfile->v->Release(ControllerIfile);

    return NULL;
}

/* ** 
*/
SCODE __cdecl UsbdserialConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        const _TCHAR * ControllerName,
        ...)
{
    PIUNKNOWN pUnk;
    UnusedParameter(Name);
    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("UsbdserialConstructor(%x %s %x %x %x %s)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk, ControllerName));

    pUnk = (PIUNKNOWN) UsbdserialNew(ControllerName);

    DBGME(0,_tprintf(_T("UsbdserialConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("UsbdserialConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl UsbdserialCDeviceCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) UsbdserialConstructor
};

static struct _DRIVERCOB UsbdserialDriverCobObject = { &UsbdserialCDeviceCobVtable, 1 };

PIUNKNOWN UsbdserialCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &UsbdserialCDeviceCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    sc = BaseInheritClass((ADDRESS *) &UsbdserialCDeviceVtable, _T("CDevice"),
                          &IID_IDevice);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    /* Null the thread loop, to save a useless thread
     */
    UsbdserialCDeviceVtable.ThreadLoop = NULL;

    return (PIUNKNOWN) &UsbdserialDriverCobObject;
}
