/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implements NullFramebuffer concrete class */
#include <mmlite.h>
//#define _DEBUG 1
#define DBGLVL 3
#include <diagnostics.h>
#include <assert.h>
#include <drivers\drivers.h>
#include <framebuffer.h>

typedef struct _PNULL_FRAMEBUFFER {
    const struct IFrameBufferVtbl *v;
    UINT RefCnt;
    FRAMEBUFFER_MODE Mode;
    UINT16 Hlen, Vlen;
    UINT16 Hvis, Vvis;
    UINT16 Hpch, Vpch;
    UINT8 Bpp, bpp;
    UINT  PaletteLength;
    UINT8 *Palette;
    UINT8 *BackingStore;
} *PNULL_FRAMEBUFFER;

/* Convert between interface and instance pointer and vice versa */
#define pNullFramebuffer(_i_) ((PNULL_FRAMEBUFFER)((_i_)))
#define iNullFramebuffer(_p_) ((PIFRAMEBUFFER)((_p_)))

PRIVATE SCODE NullFramebufferQueryInterface (
    PIFRAMEBUFFER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    SCODE sc;
    DBGME(0,_tprintf(_T("NullFramebufferQI(%x,%x.%x.%x.%x)\n"), (ADDRESS)pThis,
             ((PUINT)iid)[0],((PUINT)iid)[1],((PUINT)iid)[2],((PUINT)iid)[3]));

    sc = GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                               &IID_IFrameBuffer);
    return sc;
}

PRIVATE UINT NullFramebufferAddRef (
    PIFRAMEBUFFER pThis)
{
    DBGME(0,_tprintf(_T("NullFramebufferAR(%x)\n"),(ADDRESS)pThis));
    return AtomicInc(&pNullFramebuffer(pThis)->RefCnt);
}

PRIVATE UINT NullFramebufferRelease (
    PIFRAMEBUFFER pThis)
{
    PNULL_FRAMEBUFFER This = pNullFramebuffer(pThis);
    UINT Refs = AtomicDec(&This->RefCnt);

    DBGME(0,_tprintf(_T("NullFramebufferRel(%x,%d)\n"),(ADDRESS)pThis,Refs));
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        if (This->BackingStore)
            pHeap->v->Free(pHeap, 0, (void *) This->BackingStore);
        pHeap->v->Free(pHeap, 0, (void *) This);
    }
    return Refs;
}

PRIVATE SCODE NullFramebufferMap (
    PIFRAMEBUFFER pThis,
    /*inout*/ PTR * Where,
    /*inout*/ UINT * Length)
{
    PNULL_FRAMEBUFFER This = pNullFramebuffer(pThis);

    DBGME(1,_tprintf(_T("NullFramebufferMap(%x %x[%x] %x[%x])\n"), (ADDRESS)This,
                     (ADDRESS) Where, (ADDRESS) *Where,
                     (ADDRESS) Length, (ADDRESS) *Length));

    *Length = This->Hlen * This->Vlen;

    if (*Where != NULL) {

        if (This->BackingStore == NULL) {
            This->BackingStore = *Where;
            return S_OK;
        }
        return E_INVALID_PARAMETER;
    }

    if (This->BackingStore == NULL) {
        This->BackingStore = malloc(*Length);
        if (This->BackingStore == NULL)
            return E_NOT_ENOUGH_MEMORY;
    }

    *Where = This->BackingStore;
    return S_OK;
}

PRIVATE SCODE NullFramebufferUnmap (
    PIFRAMEBUFFER pThis,
    /*inout*/ PTR Where)
{
    PNULL_FRAMEBUFFER This = pNullFramebuffer(pThis);

    DBGME(1,_tprintf(_T("NullFramebufferUnmap(%x %x)\n"), (ADDRESS)This,
                     (ADDRESS) Where));

    if (Where != This->BackingStore)
        return E_INVALID_PARAMETER;

    free(This->BackingStore);
    This->BackingStore = NULL;
    return S_OK;
}

PRIVATE SCODE NullFramebufferGetProperty (
    PIFRAMEBUFFER pThis,
    FRAMEBUFFER_PROPERTY Property,
    /*inout*/ UINT32 * Value)
{
    PNULL_FRAMEBUFFER This = pNullFramebuffer(pThis);
    UINT32 v;

    DBGME(1,_tprintf(_T("NullFramebufferGetProp(%x %x %x)\n"), (ADDRESS)This,
                     Property, (ADDRESS)Value));

    switch (Property) {
    case FB_PROP_TOTAL_SIZE:
        v = This->Hlen * This->Hlen; break;
    case FB_PROP_HORIZONTAL_LENGTH:
        v = This->Hlen; break;
    case FB_PROP_HORIZONTAL_VISIBLE_LENGTH:
        v = This->Hvis; break;
    case FB_PROP_HORIZONTAL_PORCH:
        v = 0; break;
    case FB_PROP_VERTICAL_LENGTH:
        v = This->Vlen; break;
    case FB_PROP_VERTICAL_VISIBLE_LENGTH:
        v = This->Vvis; break;
    case FB_PROP_VERTICAL_PORCH:
        v = 0; break;
    case FB_PROP_PIXEL_BITSIZE:
        v = This->bpp; break;
    case FB_PROP_PIXEL_BYTESIZE:
        v = This->Bpp; break;
    case FB_PROP_PIXEL_ENCODING:
        v = PIXEL_TYPE_PALETTE_INDEX; break;
    case FB_PROP_MODE:
        v = This->Mode; break;
    case FB_PROP_PALETTE_LENGTH:
        v = This->PaletteLength; break;
    case FB_PROP_PALETTE_BITSIZE:
        v = 24; break;
    case FB_PROP_PALETTE_BYTESIZE:
        v = 4; break;
    case FB_PROP_PALETTE_ENCODING:
        v = PIXEL_TYPE_R8G8B8; break;
    default:
        return E_NOT_IMPLEMENTED;
    }

    *Value = v;
    return S_OK;
}

PRIVATE SCODE NullFramebufferSetMode (
    PIFRAMEBUFFER pThis,
    FRAMEBUFFER_MODE Mode)
{
    PNULL_FRAMEBUFFER This = pNullFramebuffer(pThis);

    DBGME(1,_tprintf(_T("NullFramebufferSetM(%x %x) was %x\n"), (ADDRESS)This,
                     Mode, This->Mode));

    /* Not if we have already committed 
     */
    if ((This->Mode != Mode) &&
        (This->BackingStore != NULL))
        return E_NOT_IMPLEMENTED;

    switch (Mode) {
    case FB_MODE_320x200_in_320x400_8bpp_PALETTE:
        This->Hvis = 320;
        This->Vvis = 200;
        This->Hlen = 320;
        This->Vlen = 400;
        This->Hpch = 0;
        This->Vpch = 0;
        This->bpp = 8;
        This->Bpp = 1;
        break;
    case FB_MODE_640x480_in_768x480_8bpp_PALETTE:
        This->Hvis = 640;
        This->Vvis = 480;
        This->Hlen = 768;
        This->Vlen = 480;
        This->Hpch = 0;
        This->Vpch = 0;
        This->bpp = 8;
        This->Bpp = 1;
        break;
    }

    return S_OK;
}

PRIVATE SCODE NullFramebufferGetPalette (
    PIFRAMEBUFFER pThis,
    /*out*/ UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING ExpectedPixelFormat)
{
    PNULL_FRAMEBUFFER This = pNullFramebuffer(pThis);

    DBGME(1,_tprintf(_T("NullFramebufferGetP(%x %x %x)\n"), (ADDRESS)This,
                     (ADDRESS) Buffer, Length));

    UnusedParameter(ExpectedPixelFormat);

    if ((Buffer == NULL) ||
        (Length < (4 * This->PaletteLength)))
        return E_INVALID_PARAMETER;

    memcpy(Buffer, This->Palette, 4 * This->PaletteLength);

    return S_OK;
}

PRIVATE SCODE NullFramebufferGetPaletteEntry (
    PIFRAMEBUFFER pThis,
    UINT Index,
    /*out*/ UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING ExpectedPixelFormat)
{
    PNULL_FRAMEBUFFER This = pNullFramebuffer(pThis);

    DBGME(1,_tprintf(_T("NullFramebufferGetPE(%x %x %x %x)\n"), (ADDRESS)This,
                     Index, (ADDRESS) Buffer, Length));

    UnusedParameter(ExpectedPixelFormat);

    if ((Index >= This->PaletteLength) ||
        (Buffer == NULL) ||
        (Length < 4))
        return E_INVALID_PARAMETER;

    memcpy(Buffer, This->Palette + (Index * 4), 4);

    return S_OK;
}

PRIVATE SCODE NullFramebufferSetPalette (
    PIFRAMEBUFFER pThis,
    const UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING PixelFormat)
{
    PNULL_FRAMEBUFFER This = pNullFramebuffer(pThis);

    DBGME(1,_tprintf(_T("NullFramebufferSetP(%x %x %x)\n"), (ADDRESS)This,
                     (ADDRESS) Buffer, Length));

    UnusedParameter(PixelFormat);

    if ((Buffer == NULL) ||
        (Length < (4 * This->PaletteLength)))
        return E_INVALID_PARAMETER;

    memcpy(This->Palette, Buffer, 4 * This->PaletteLength);

    return S_OK;
}

PRIVATE SCODE NullFramebufferSetPaletteEntry (
    PIFRAMEBUFFER pThis,
    UINT Index,
    const UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING PixelFormat)
{
    PNULL_FRAMEBUFFER This = pNullFramebuffer(pThis);

    DBGME(1,_tprintf(_T("NullFramebufferSetPE(%x %x %x %x)\n"), (ADDRESS)This,
                     Index, (ADDRESS) Buffer, Length));

    UnusedParameter(PixelFormat);

    if ((Index >= This->PaletteLength) ||
        (Buffer == NULL) ||
        (Length < 4))
        return E_INVALID_PARAMETER;

    memcpy(This->Palette + (Index * 4), Buffer, 4);

    return S_OK;
}

#include "s_NullFramebuffer_v.c"

/* Constructor */
PIFRAMEBUFFER NullFramebufferNew(
    ADDRESS FramebufferBase,
    ADDRESS PaletteBase)
{
    PIHEAP pHeap = CurrentHeap();
    PNULL_FRAMEBUFFER This;

    DBGME(1,_tprintf(_T("NullFramebufferNew(%x %x)\n"),
                     FramebufferBase, PaletteBase));

    UnusedParameter(FramebufferBase);
    UnusedParameter(PaletteBase);

    This = (PNULL_FRAMEBUFFER) pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, sizeof *This + (256*4), 0);
    if (!This)
        return NULL;

    /* Initialize non-zero fields
     */
    This->v = &NullFramebufferVtbl;
    This->RefCnt = 1;
    This->PaletteLength = 256;
    This->Palette = (UINT8*)(This + 1);

    /* Default initial state, can be changed (once) later
     */
    (void) NullFramebufferSetMode(iNullFramebuffer(This),
                                  FB_MODE_640x480_in_768x480_8bpp_PALETTE);

    return iNullFramebuffer(This);
}

/* ** 
*/
SCODE __cdecl NullFramebufferConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * MyName,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        ADDRESS PaletteBase,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Interrupt);

    DBGME(0,_tprintf(_T("NullFramebufferConstructor(%x %s %x)\n"),
                     (ADDRESS) pThis, MyName, (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) NullFramebufferNew(RegisterBase,PaletteBase);

    DBGME(1,_tprintf(_T("NullFramebufferConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("NullFramebufferConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl NullFramebufferDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) NullFramebufferConstructor
};

static struct _DRIVERCOB NullFramebufferDriverCobObject = { &NullFramebufferDriverCobVtable, 1 };

PIUNKNOWN noframebufferCobMain(void)
{
    SCODE sc;

    DBGME(0,_tprintf(_T("NullFramebufferCobMain()\n")));

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &NullFramebufferDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    assert(!FAILED(sc));
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &NullFramebufferDriverCobObject;
}
