/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Under Windows, implements a simulated window I/O.
 * Implements an IFrameBuffer for painting, an IFile for keyboard input,
 * and an IFile for mouse input.
 */
#include <mmlite.h>
#include <stddef.h> /* for offsetof() */
//#define _DEBUG 1
#define DBGLVL 3
#include <diagnostics.h>
#include <assert.h>
#include <drivers\drivers.h>

#include "WinFb.h"

typedef struct _SIMFB {
    const struct IFrameBufferVtbl *fv;
    const struct IFileVtbl *kv;
    const struct IFileVtbl *mv;
    /* Single-instanced object, no refcount */
    /*UINT RefCnt;*/
    void *WinFbState;
} SIMFB, *PSIMFB;

/* Convert between interface and instance pointer and vice versa */
#define pSimFb(_i_)  ((PSIMFB)((_i_)))
#define iSimFb(_p_)  ((PIFRAMEBUFFER)((_p_)))
#define pSimKbd(_i_) ((PSIMFB)((ADDRESS)(_i_) - offsetof(struct _SIMFB, kv)))
#define iSimKbd(_p_) ((PIFILE)(&((PSIMFB)(_p_))->kv))
#define pSimMs(_i_)  ((PSIMFB)((ADDRESS)(_i_) - offsetof(struct _SIMFB, mv)))
#define iSimMs(_p_)  ((PIFILE)(&((PSIMFB)(_p_))->mv))

/* IFrameBuffer implementation
 */
PRIVATE SCODE SimFbQueryInterface (
    PIFRAMEBUFFER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    SCODE sc;
    sc = GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                               &IID_IFrameBuffer);
    return sc;
}

PRIVATE SCODE SimFbMap (
    PIFRAMEBUFFER pThis,
    /*inout*/ PTR * Where,
    /*inout*/ UINT * Length)
{
    PSIMFB This = pSimFb(pThis);
    WinFbInit(&This->WinFbState);

    UnusedParameter(Length);

    return WinFbGetFramebufferAddress(This->WinFbState,Where);
}

PRIVATE SCODE SimFbGetProperty (
    PIFRAMEBUFFER pThis,
    FRAMEBUFFER_PROPERTY Property,
    /*inout*/ UINT32 * Value)
{
    PSIMFB This = pSimFb(pThis);
    WinFbInit(&This->WinFbState);

    DBGME(1,_tprintf(_T("SimFbGetProp(%x %x %x)\n"), (ADDRESS)This,
                     Property, (ADDRESS)Value));

    return WinFbGetProp(This->WinFbState,
                        Property,
                        Value);
}

PRIVATE SCODE SimFbSetMode (
    PIFRAMEBUFFER pThis,
    FRAMEBUFFER_MODE Mode)
{
    PSIMFB This = pSimFb(pThis);
    WinFbInit(&This->WinFbState);

    return WinFbSetMode(This->WinFbState,
                        Mode);
}

PRIVATE SCODE SimFbGetPalette (
    PIFRAMEBUFFER pThis,
    /*out*/ UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING ExpectedPixelFormat)
{
    PSIMFB This = pSimFb(pThis);
    WinFbInit(&This->WinFbState);

    return WinFbGetPalette(This->WinFbState,Buffer,Length,ExpectedPixelFormat);
}

PRIVATE SCODE SimFbGetPaletteEntry (
    PIFRAMEBUFFER pThis,
    UINT Index,
    /*out*/ UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING ExpectedPixelFormat)
{
    PSIMFB This = pSimFb(pThis);
    WinFbInit(&This->WinFbState);

    return WinFbGetPaletteEntry(This->WinFbState,Index,
                                Buffer,Length,ExpectedPixelFormat);
}

PRIVATE SCODE SimFbSetPalette (
    PIFRAMEBUFFER pThis,
    const UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING PixelFormat)
{
    PSIMFB This = pSimFb(pThis);
    WinFbInit(&This->WinFbState);

    return WinFbSetPalette(This->WinFbState,Buffer,Length,PixelFormat);
}

PRIVATE SCODE SimFbSetPaletteEntry (
    PIFRAMEBUFFER pThis,
    UINT Index,
    const UINT8* Buffer,
    UINT Length,
    PIXEL_ENCODING PixelFormat)
{
    PSIMFB This = pSimFb(pThis);
    WinFbInit(&This->WinFbState);

    return WinFbSetPaletteEntry(This->WinFbState,Index,
                                Buffer,Length,PixelFormat);
}

/* IFile implementation for the keyboard
 */
PRIVATE SCODE SimKbdQueryInterface (
    PIFILE pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    SCODE sc;
    sc = GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                               &IID_IFile);
    return sc;
}

/* Pick up any pending events from the keyboard
 */
PRIVATE SCODE SimKbdReadAt (
    PIFILE pThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PSIMFB This = pSimKbd(pThis);
    WinFbInit(&This->WinFbState);

    if (Size < sizeof(KEYBOARD_INPUT_DATA))
        return E_INVALID_PARAMETER;

    return WinKbdRead(This->WinFbState,
                      (PKEYBOARD_INPUT_DATA)pBuffer,
                      pSizeRead);
}

/* IFile implementation for the mouse
 */

#define SimMsQueryInterface SimKbdQueryInterface

/* Pick up any pending events from the mouse
 */
PRIVATE SCODE SimMsReadAt (
    PIFILE pThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PSIMFB This = pSimMs(pThis);
    WinFbInit(&This->WinFbState);

    if (Size < sizeof(MOUSE_INPUT_DATA))
        return E_INVALID_PARAMETER;

    return WinMsRead(This->WinFbState,
                     (PMOUSE_INPUT_DATA)pBuffer,
                     pSizeRead);
}

/* Auto-generated Vtables */
#include "s_SimFb_v.c"

/* Constructor */
PIUNKNOWN SimFbNew(
    ADDRESS BaseAddress,
    ADDRESS PaletteAddress)
{
    /* We can only do one (although..)
     */
    static struct _SIMFB TheFb = {
        &SimFbSimFbVtbl,
        &SimKbdSimFbVtbl,
        &SimMsSimFbVtbl,
        NULL
    };

    UnusedParameter(PaletteAddress);

    /* First argument selects between framebuffer, keyboard and mouse
     */
    if (BaseAddress == 0)
        return (PIUNKNOWN) iSimFb(&TheFb);
    if (BaseAddress == 1)
        return (PIUNKNOWN) iSimKbd(&TheFb);
    if (BaseAddress == 2)
        return (PIUNKNOWN) iSimMs(&TheFb);
    return NULL;
}
