/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <import_windows.h>
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>

#include "WinFb.h"

#include "_palette.c"

/* If we are also emulating a keyboard..
 */
typedef struct _KK_STATE {
    UINT nHead;             /* Remove chars from here */
    UINT nTail;             /* Insert chars here */
    UINT Keys[1];           /* variable-size ring of pending chars */
} KK_STATE, *PKK_STATE;

#define KBD_QUEUE_SIZE 64

/* ...and a mouse...
 */
typedef struct _MM_STATE {
    UINT nHead;             /* Remove chars from here */
    UINT nTail;             /* Insert chars here */
    struct _MMOVE {
        UINT Buttons;
        UINT Coordinates;
    } Moves[1];             /* variable-size ring of mouse moves&presses */
} MM_STATE, *PMM_STATE;

#define MS_QUEUE_SIZE 64

/* Our device's state
 */
typedef struct _FB320_STATE {
    FRAMEBUFFER_MODE CurrentMode;
    int ScreenWidth;
    int ScreenHeight;
    int ScreenVisibleWidth;
    int ScreenVisibleHeight;
    void *pDispBits;        /* Screen's pixels */
    ATOM hClass;            /* Window class */
    HWND hPaintWnd;         /* Handle on paint window */
    BOOL WndIsActive;       /* Used for messing up the system palette */
    HDC hDC;                /* Window's own DC */
    BITMAPINFO *pBMI;       /* Used to create the DIB Section */
    HDC hDCDIBSection;      /* Handle to compatible DIBSection */
    HBITMAP hBitmap;        /* Handle to DIBSection bitmap */
    volatile BOOL PaletteChanged; /* Palette out of date */
    BOOL PaletteEverChanged; /* Did the app ever try to set the palette */
    UINT Cycles;            /* Cycle count, e.g. internal timer */
    UINT NextRepaint;       /* Cycle count at which screen refreshes */
    volatile BOOL bFbModified;  /* But did we ever write to it */
    UINT nRepaints;         /* How many times was it refreshed */

    PKK_STATE pKbd;         /* Non NULL if we should emulate a kbd */
    BOOL  KbdBlocks;        /* Indicate if signalling is desired */
    CONDITION KbdWait;      /* Wait here for blocking calls */

    PMM_STATE pMs;          /* Non NULL if we should emulate a mouse */
    BOOL  MsBlocks;         /* Indicate if signalling is desired */
    CONDITION MsWait;      /* Wait here for blocking calls */

} FB320_STATE, *PFB320_STATE;


typedef unsigned char Pix;  /* one pixel is one byte for us */


/* ***********************  Windows Framebuffer Emulation *************** */


static _TCHAR szAppName[32];

/* forward declarations of functions in this module
 */
static BOOL    InitializeWindow (PFB320_STATE);
static LONG    CommandHandler (PFB320_STATE, WPARAM, LPARAM);
static void    InitDisp(PFB320_STATE pState);
static void    Repaint(HWND hWnd, HDC hDC, BOOL bRefresh, BOOL bBumpFrame);
static void    ReinstallPalette(PFB320_STATE pState);
static void    AppActivate(PFB320_STATE pState, BOOL fActive);
static LONG    WINAPI Fb320WndProc (HWND, UINT, WPARAM, LPARAM);/*?*/

/* Constructor
 */
static void StartMeUp(PFB320_STATE pState)
{
    /* Create my window
     */
    if (!InitializeWindow (pState) ) {
        return;
    }

    /* Create a paintable object
     */
    InitDisp(pState);

    /* Done
     */
}

static BOOL InitializeWindow ( PFB320_STATE pState )
{
    WNDCLASS   wc;
    HWND       hWnd;
    HINSTANCE  hInstance;
    RECT       rc;

    /* Find our instance
     */
    hInstance = GetModuleHandleA(NULL);

    /* Register the frame class
     */
    _stprintf(szAppName,_T("Fb%dx%d"),
              pState->ScreenVisibleWidth,
              pState->ScreenVisibleHeight);

    wc.style         = CS_VREDRAW | CS_HREDRAW | CS_OWNDC;
    wc.lpfnWndProc   = (WNDPROC) Fb320WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = NULL/*LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1))*/;
    wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
    wc.hbrBackground = GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName  = NULL/*szAppName*/;
    wc.lpszClassName = szAppName;

    pState->hClass = RegisterClass (&wc);
    if (!pState->hClass ) {
        volatile DWORD err = GetLastError();
        return FALSE;
    }

    /* Create the frame
     */
    rc.left = rc.top = 32;
    rc.right = pState->ScreenVisibleWidth-1;
    rc.bottom = pState->ScreenVisibleHeight+18;
    AdjustWindowRect(&rc, WS_OVERLAPPEDWINDOW, FALSE);

    hWnd = CreateWindow (szAppName,
             szAppName,
             WS_POPUP | WS_CAPTION | WS_SYSMENU | WS_BORDER
                      | WS_MAXIMIZEBOX | WS_MINIMIZEBOX,
             rc.left,
             rc.top,
             rc.right,
             rc.bottom,
             NULL,
             NULL,
             hInstance,
             NULL);

    /* make sure window was created 
     */
    if (!hWnd) {
        volatile DWORD err = GetLastError();
        return FALSE;
    }

    /* Tell the window to remember me (and viceversa)
     */
    (void) SetWindowLong(hWnd,
                         GWL_USERDATA,
                         (LONG) pState);
    pState->hPaintWnd = hWnd;

    /* Get the device context for the new window
     */
    pState->hDC = GetDC(hWnd);

    ShowWindow(hWnd, /*SW_SHOWNORMAL*/SW_SHOWDEFAULT);

    UpdateWindow(hWnd);

    return TRUE;
}

/* main window procedure */
static LONG WINAPI Fb320WndProc (
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
    LONG    lRet = 1;
    PFB320_STATE pState = (PFB320_STATE) GetWindowLong( hWnd, GWL_USERDATA);

    switch (uMsg)
    {
        case WM_CLOSE:
            if (pState->hDC)
                ReleaseDC(pState->hPaintWnd, pState->hDC);

            /* call destroy window to cleanup and go away */
            DestroyWindow (hWnd);
            if (pState && pState->PaletteEverChanged)
                AppActivate(pState, FALSE);
            break;

        case WM_DESTROY:
            if (pState->hDC)
                ReleaseDC(pState->hPaintWnd, pState->hDC);

            if (pState && pState->PaletteEverChanged)
                AppActivate(pState, FALSE);

            PostQuitMessage (0);
            break;

        case WM_PAINT:
            {
                PAINTSTRUCT ps;
                HDC hDC = BeginPaint(hWnd, &ps);
                Repaint(hWnd, hDC, TRUE, FALSE);
                EndPaint(hWnd, &ps);
            }
            break;

        case WM_ACTIVATE:
        case WM_ACTIVATEAPP:
            /* (in)activation: must switch back and forth our palette
             */
            if (pState && pState->PaletteEverChanged)
            {
                AppActivate(pState, (BOOL)wParam);
            }
            break;

        case WM_TIMER:
            pState->bFbModified = TRUE;
            break;

        default:
            /* pass all unhandled messages to DefWindowProc */
            lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }

    /* return 1 if handled message, 0 if not */
    return lRet;
}

static void InitDisp(PFB320_STATE pState)
{
    HDC hdc;
    BITMAPINFO *pBMI = pState->pBMI;

    hdc = GetDC(pState->hPaintWnd);

    if (!pBMI)
        DebugBreak();

    pBMI->bmiHeader.biSize = sizeof (BITMAPINFOHEADER);
    pBMI->bmiHeader.biWidth = pState->ScreenWidth;
    pBMI->bmiHeader.biHeight = -pState->ScreenHeight;
    pBMI->bmiHeader.biPlanes = 1;
    pBMI->bmiHeader.biBitCount = 8;
    pBMI->bmiHeader.biCompression = BI_RGB;
    pBMI->bmiHeader.biSizeImage = pState->ScreenWidth * pState->ScreenHeight;
    pBMI->bmiHeader.biXPelsPerMeter = 0;
    pBMI->bmiHeader.biYPelsPerMeter = 0;
    pBMI->bmiHeader.biClrUsed = 0;
    pBMI->bmiHeader.biClrImportant = 0;

    memcpy(pBMI->bmiColors, DefaultPalette, sizeof(DefaultPalette));

    pState->hDCDIBSection = CreateCompatibleDC(hdc);
    pState->hBitmap = CreateDIBSection (pState->hDCDIBSection,
                                        pBMI,
                                        DIB_RGB_COLORS,
                                        &pState->pDispBits,
                                        0,
                                        0);

    if (pState->hBitmap == NULL) {
        _TCHAR Error[256];
        SCODE sc = GetLastError();
        _stprintf(Error, _T("CreateDIBSection failed. sc=%x\n"), sc);
        MessageBox (pState->hPaintWnd, Error,
                    _T("Application Error"), MB_OK | MB_ICONEXCLAMATION);
        return;
    }


    SelectObject(pState->hDCDIBSection, pState->hBitmap);

    ReleaseDC(pState->hPaintWnd, hdc);
}

#define CHANGEMODES 1
#if CHANGEMODES
/* Destructor
 */
void TearMeDown(PFB320_STATE pState)
{
    HINSTANCE  hInstance;

    /* -- DestroyDisp() -- */

    DeleteObject(pState->hBitmap);
    pState->hBitmap = NULL;
    pState->pDispBits = NULL;

    ReleaseDC(pState->hPaintWnd,pState->hDCDIBSection);

    /* -- FinalizeWindow() -- */

    /* Find our instance
     */
    hInstance = GetModuleHandleA(NULL);

    /* Release the DC
     */
    ReleaseDC(pState->hPaintWnd,pState->hDC);
    pState->hDC = NULL;

    /* Close the window down
     */
    CloseHandle(pState->hPaintWnd);
    pState->hPaintWnd = 0;

    /* And finally, deregister the class
     */
    UnregisterClass((LPCTSTR)pState->hClass,hInstance);
}

#endif


static const int SysPalIndex[] = {
        COLOR_ACTIVEBORDER,             COLOR_ACTIVECAPTION,
        COLOR_APPWORKSPACE,             COLOR_BACKGROUND,
        COLOR_BTNFACE,                  COLOR_BTNSHADOW,
        COLOR_BTNTEXT,                  COLOR_CAPTIONTEXT,
        COLOR_GRAYTEXT,                 COLOR_HIGHLIGHT,
        COLOR_HIGHLIGHTTEXT,            COLOR_INACTIVEBORDER,
        COLOR_INACTIVECAPTION,          COLOR_MENU,
        COLOR_MENUTEXT,                 COLOR_SCROLLBAR,
        COLOR_WINDOW,                   COLOR_WINDOWFRAME,
        COLOR_WINDOWTEXT,               COLOR_INACTIVECAPTIONTEXT,
        COLOR_BTNHIGHLIGHT,

        COLOR_3DDKSHADOW, COLOR_3DLIGHT, COLOR_INFOTEXT, COLOR_INFOBK
};

#define rgbBlack RGB(0,0,0)
#define rgbWhite RGB(255,255,255)
static COLORREF MonoColors[] = {
        rgbBlack,                               rgbWhite,
        rgbWhite,                               rgbWhite,
        rgbWhite,                               rgbBlack,
        rgbBlack,                               rgbBlack,
        rgbBlack,                               rgbBlack,
        rgbWhite,                               rgbWhite,
        rgbWhite,                               rgbWhite,
        rgbBlack,                               rgbWhite,
        rgbWhite,                               rgbBlack,
        rgbBlack,                               rgbBlack,
        rgbBlack,

        rgbBlack, rgbWhite, rgbWhite, rgbBlack
};

#define NumSysColors (sizeof(SysPalIndex)/sizeof(SysPalIndex[1]))
static COLORREF OldSysColors[NumSysColors];

static void AppActivate(PFB320_STATE pState, BOOL fActive)
{
    HDC hdc;
    int i;

    hdc = GetDC(NULL);

    if (fActive && !pState->WndIsActive) {
        /* Becoming Active Window
         */
        if ((GetSystemPaletteUse(hdc) == SYSPAL_STATIC)) {

            /* Save off the current system colors
             */
            for (i=0; i<NumSysColors; i++) {
                OldSysColors[i] = GetSysColor(SysPalIndex[i]);
            }
            /* Assign new colors
             */
            //SetSystemPaletteUse(hdc, SYSPAL_NOSTATIC);
            //SetSysColors(NumSysColors, SysPalIndex, MonoColors);

        }

        ReinstallPalette(pState);

    } else if (!fActive && pState->WndIsActive) {
        /* going to background
         */
        if (GetSystemPaletteUse(hdc) == SYSPAL_NOSTATIC) {
            //SetSystemPaletteUse(hdc, SYSPAL_STATIC);
        }
        //SetSysColors(NumSysColors, SysPalIndex, OldSysColors);

    }

    pState->WndIsActive = fActive;


    ReleaseDC(NULL,hdc);
}

static void Repaint(HWND hWnd, HDC hDC, BOOL bRefresh, BOOL bBumpFrame)
{
    _TCHAR szFrame[100];
    PFB320_STATE pState = (PFB320_STATE) GetWindowLong( hWnd, GWL_USERDATA);

    if (pState) {

        if (bRefresh) {
            BitBlt(hDC, 
                   0, 0, pState->ScreenVisibleWidth, pState->ScreenVisibleHeight,
                   pState->hDCDIBSection, 0, 0,
                   SRCCOPY);
            GdiFlush();
        }

        if (bBumpFrame) {
            pState->nRepaints++;
            if ((pState->nRepaints & 0xff) == 0)
            {
                _stprintf(szFrame, _T("%s Frame = %d"),
                          szAppName, pState->nRepaints);
                SetWindowText(hWnd, szFrame);
            }
        }
    }
}

#define IOUNDEFINED 0xffffffff

/* ************************ Keyboard Device ***************** */

SCODE WinKbdRead(PFB320_STATE pState,
                 PKEYBOARD_INPUT_DATA Report,
                 UINT *pSizeRead)
{
    PKK_STATE pKbd = pState->pKbd;
    UINT Key;

    /* First time in, initialize.
     * NB: RACE ! not MT safe
     */
    if (pKbd == NULL) {
        pKbd = (PKK_STATE) malloc(sizeof(*pKbd) + (KBD_QUEUE_SIZE-1) * sizeof(UINT));
        if (pKbd) 
            memset(pKbd, 0, sizeof(*pKbd) + (KBD_QUEUE_SIZE-1) * sizeof(UINT));
        Condition_Init(&pState->KbdWait);
        pState->KbdBlocks = FALSE;
        pState->pKbd = pKbd;
        if (pKbd == NULL) {
            /* try again later, maybe */
            *pSizeRead = 0;
            return S_OK;
        }
    }

    /* Wait until something is queued.
     */
    for (;;) {
        pState->KbdBlocks = TRUE;
        if (pKbd->nHead < pKbd->nTail)
            break;
        Condition_Wait(&pState->KbdWait,NULL,NULL);
    }

    Key = (pKbd->Keys[(pKbd->nHead++) & (KBD_QUEUE_SIZE-1)]);

    Report->ReportTime = CurrentTime(); /* XXX should be recorded earlier */
    Report->MakeCode = (UINT16) (Key & 0xffff); /* scancode */
    if (Key & 0x10000)
        Report->Flags = KEYBOARD_FLAGS_UP_STROKE;
    else
        Report->Flags = KEYBOARD_FLAGS_DOWN_STROKE;

    *pSizeRead = sizeof(KEYBOARD_INPUT_DATA);
    return S_OK;
}

/* ************************ Mouse Device ***************** */

SCODE WinMsRead(PFB320_STATE pState,
                PMOUSE_INPUT_DATA Report,
                UINT *pSizeRead)
{
    PMM_STATE pMs = pState->pMs;
    struct _MMOVE Move;

    /* First time in, initialize.
     * NB: RACE ! not MT safe
     */
    if (pMs == NULL) {
        pMs = (PMM_STATE) malloc(sizeof(*pMs) + (MS_QUEUE_SIZE-1) * sizeof(UINT));
        if (pMs) 
            memset(pMs, 0, sizeof(*pMs) + (MS_QUEUE_SIZE-1) * sizeof(UINT));
        Condition_Init(&pState->MsWait);
        pState->MsBlocks = FALSE;
        pState->pMs = pMs;
        if (pMs == NULL) {
            /* try again later, maybe */
            *pSizeRead = 0;
            return S_OK;
        }
    }

    /* Wait until something is queued.
     */
    for (;;) {
        pState->MsBlocks = TRUE;
        if (pMs->nHead < pMs->nTail)
            break;
        Condition_Wait(&pState->MsWait,NULL,NULL);
    }

    Move = (pMs->Moves[(pMs->nHead++) & (MS_QUEUE_SIZE-1)]);

    Report->ReportTime = CurrentTime(); /* XXX should be recorded earlier */
    Report->LastX = (UINT16)(Move.Coordinates);
    Report->LastY = (UINT16)(Move.Coordinates >> 16);
    Report->LastZ = 0;
    Report->RawButtons = 0;
    if (Move.Buttons & MK_LBUTTON) Report->RawButtons |= MOUSE_LEFT_BUTTON_DOWN;
    if (Move.Buttons & MK_MBUTTON) Report->RawButtons |= MOUSE_MIDDLE_BUTTON_DOWN;
    if (Move.Buttons & MK_RBUTTON) Report->RawButtons |= MOUSE_RIGHT_BUTTON_DOWN;

    *pSizeRead = sizeof(MOUSE_INPUT_DATA);
    return S_OK;
}

/* Windows message handler function
 */
void Fb320MsgLoop(PFB320_STATE pState)
{
    MSG Message;

    while (GetMessage(&Message, pState->hPaintWnd, 0, 0)) {

        if (Message.message == WM_QUIT)
            break;

        if (Message.message == WM_KEYDOWN || Message.message == WM_KEYUP) {
            /* updown stroke */
            UINT k = (Message.message == WM_KEYDOWN) ? 0 : 0x10000;

            /* key value. Pass both scancode and Vkey */
            k |= ((Message.lParam >> 16) & 0xff) /* scancode */
                | (Message.wParam << 17); /* vkey */

            if (pState->pKbd) {
                pState->pKbd->Keys[ (pState->pKbd->nTail++)
                                    & (KBD_QUEUE_SIZE-1)] = k;
                if (pState->KbdBlocks)
                    Condition_InterruptSignal(&pState->KbdWait);
            }

        } else if (Message.message == WM_MOUSEMOVE) {
            struct _MMOVE m;

            m.Coordinates = Message.lParam;
            m.Buttons     = Message.wParam;

            if (pState->pMs) {
                pState->pMs->Moves[ (pState->pMs->nTail++)
                                    & (MS_QUEUE_SIZE-1)] = m;
                if (pState->MsBlocks)
                    Condition_InterruptSignal(&pState->MsWait);
            }


        } else {
            TranslateMessage(&Message);
            DispatchMessage(&Message);
        }

        if (pState->bFbModified) {
            pState->bFbModified = FALSE;
            Repaint( pState->hPaintWnd, pState->hDC, TRUE, TRUE );
        }
        if (pState->PaletteChanged) {
            pState->PaletteChanged = FALSE;
            ReinstallPalette(pState);
        }
    }

}

/* This one needs up/down keycodes
 */
void Fb320Winops(PFB320_STATE pState)
{
    PKK_STATE pKbd;
    MSG Message;
    UINT k;

    if (pState && ((pKbd = pState->pKbd) != NULL)) {
        while (PeekMessage(&Message, pState->hPaintWnd,
                           WM_KEYDOWN, WM_KEYUP,
                           PM_REMOVE)) {
            /* updown stroke */
            k = (Message.message == WM_KEYDOWN) ? 0 : 0x10000;

            /* key value. Pass both scancode and Vkey */
            k |= ((Message.lParam >> 16) & 0xff) /* scancode */
                | (Message.wParam << 17); /* vkey */

            pKbd->Keys[ (pKbd->nTail++) & (KBD_QUEUE_SIZE-1)] = k;
        }
    }
}

/* ***********************  Framebuffer Device *************** */

/* Change the entire color palette 
 * PALETTE points to triples R,G,B of bytes, in emulated memory
 */
#define DOGAMMA 1
#if DOGAMMA
static const unsigned char gamma[256] = 
{8,12,16,19,22,24,27,29,31,34,36,38,40,41,43,45,47,49,50,52,
53,55,57,58,60,61,63,64,65,67,68,70,71,72,74,75,76,77,79,80,
81,82,84,85,86,87,88,90,91,92,93,94,95,96,98,99,100,101,102,103,
104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,
124,125,126,127,128,129,130,131,132,133,134,135,135,136,137,138,139,140,141,142,
143,143,144,145,146,147,148,149,150,150,151,152,153,154,155,155,156,157,158,159,
160,160,161,162,163,164,165,165,166,167,168,169,169,170,171,172,173,173,174,175,
176,176,177,178,179,180,180,181,182,183,183,184,185,186,186,187,188,189,189,190,
191,192,192,193,194,195,195,196,197,197,198,199,200,200,201,202,202,203,204,205,
205,206,207,207,208,209,210,210,211,212,212,213,214,214,215,216,216,217,218,219,
219,220,221,221,222,223,223,224,225,225,226,227,227,228,229,229,230,231,231,232,
233,233,234,235,235,236,237,237,238,238,239,240,240,241,242,242,243,244,244,245,
246,246,247,247,248,249,249,250,251,251,252,253,253,254,254,255};

#endif


SCODE WinFbGetPalette(PFB320_STATE pState,
                      /*out*/ UINT8* Palette,
                      UINT Length,
                      PIXEL_ENCODING ExpectedPixelFormat)
{
    RGBQUAD *      pRgb;
    UINT i;

    if (Length < (256*4))
        return E_INVALID_PARAMETER;

    pRgb = pState->pBMI->bmiColors;

    if (ExpectedPixelFormat == PIXEL_TYPE_R8G8B8X8) {

        memcpy(Palette,pRgb,Length);

    } else if (ExpectedPixelFormat == PIXEL_TYPE_R8G8B8) {

        Length = Length / 3;
        for (i = 0; i < Length; i++) {
        
            *Palette++ = pRgb->rgbRed;
            *Palette++ = pRgb->rgbGreen;
            *Palette++ = pRgb->rgbBlue;

            pRgb++;
        }
    } else
        return E_INVALID_PARAMETER;

    return S_OK;
}

SCODE WinFbSetPalette(PFB320_STATE pState,
                      const UINT8* Palette,
                      UINT Length,
                      PIXEL_ENCODING PixelFormat)
{
    RGBQUAD *      pRgb;
    UINT i;

    if (Length > (256*4))
        return E_INVALID_PARAMETER;

    pRgb = pState->pBMI->bmiColors;

    if (PixelFormat == PIXEL_TYPE_R8G8B8X8) {

        memcpy(pRgb,Palette,Length);

    } else if (PixelFormat == PIXEL_TYPE_R8G8B8) {

        Length = Length / 3;
        for (i = 0; i < Length; i++) {
        
#if DOGAMMA
            pRgb->rgbRed = gamma[*Palette++];
            pRgb->rgbGreen = gamma[*Palette++];
            pRgb->rgbBlue = gamma[*Palette++];
#else
            pRgb->rgbRed = *Palette++;
            pRgb->rgbGreen = *Palette++;
            pRgb->rgbBlue = *Palette++;
#endif

            pRgb++;
        }
    } else
        return E_INVALID_PARAMETER;

    pState->PaletteChanged = TRUE;
    pState->PaletteEverChanged = TRUE;

    return S_OK;
}

SCODE WinFbGetPaletteEntry(PFB320_STATE pState,
                           UINT Index,
                           /*out*/ UINT8* Buffer,
                           UINT Length,
                           PIXEL_ENCODING ExpectedPixelFormat)
{
    RGBQUAD *pRgb = pState->pBMI->bmiColors;

    if ((Length < 4) ||
        (Index > 255))
        return E_INVALID_PARAMETER;

    pRgb += Index;

    if (ExpectedPixelFormat == PIXEL_TYPE_R8G8B8X8) {
        *(RGBQUAD*)Buffer = *pRgb;
    } else if (ExpectedPixelFormat == PIXEL_TYPE_R8G8B8) {

        Buffer[0] = pRgb->rgbRed;
        Buffer[1] = pRgb->rgbGreen;
        Buffer[2] = pRgb->rgbBlue;

    } else
        return E_INVALID_PARAMETER;

    return S_OK;
}


SCODE WinFbSetPaletteEntry(PFB320_STATE pState,
                           UINT Index,
                           const UINT8* Buffer,
                           UINT Length,
                           PIXEL_ENCODING PixelFormat)
{
    RGBQUAD *pRgb = pState->pBMI->bmiColors;

    if ((Length > 4) ||
        (Index > 255))
        return E_INVALID_PARAMETER;

    pRgb += Index;

    if (PixelFormat == PIXEL_TYPE_R8G8B8X8) {
        *pRgb = *(RGBQUAD*)Buffer;
    } else if (PixelFormat == PIXEL_TYPE_R8G8B8) {

#if DOGAMMA
        pRgb->rgbRed = gamma[Buffer[0]];
        pRgb->rgbGreen = gamma[Buffer[1]];
        pRgb->rgbBlue = gamma[Buffer[2]];
#else
        pRgb->rgbRed = Buffer[0];
        pRgb->rgbGreen = Buffer[1];
        pRgb->rgbBlue = Buffer[2];
#endif

    } else
        return E_INVALID_PARAMETER;

    pState->PaletteChanged = TRUE;
    pState->PaletteEverChanged = TRUE;

    return S_OK;
}

static void ReinstallPalette(PFB320_STATE pState)
{
    int i;

    i = SetDIBColorTable(pState->hDCDIBSection,0,256,pState->pBMI->bmiColors);
    if (i != 256) {
        _TCHAR Error[256];
        SCODE sc = GetLastError();
        _stprintf(Error, _T("Could only set %d colors out of 256? sc=%x\n"),
                  i, sc);
        MessageBox (pState->hPaintWnd, Error,
                    _T("Application Error"), MB_OK | MB_ICONEXCLAMATION);
    }

#if 0
    no help;
    PostMessage(HWND_BROADCAST, WM_SYSCOLORCHANGE, 0, 0);
#endif
}

SCODE WinFbGetFramebufferAddress(PFB320_STATE pState, PTR *pAddress)
{
    *pAddress = pState->pDispBits;
    return S_OK;
}

extern SCODE WinFbGetProp(PFB320_STATE pState,
                          FRAMEBUFFER_PROPERTY Property,
                          /*inout*/ UINT32 * Value)
{
    UINT32 v;

    /* We must have a mode
     */
    if (pState->CurrentMode == FB_MODE_UNDEFINED)
        return E_OBJECT_NOT_INITIALIZED;

    switch (Property) {
    case FB_PROP_TOTAL_SIZE:
        v = pState->ScreenWidth * pState->ScreenHeight; break;
    case FB_PROP_HORIZONTAL_LENGTH:
        v = pState->ScreenWidth; break;
    case FB_PROP_HORIZONTAL_VISIBLE_LENGTH:
        v = pState->ScreenVisibleWidth; break;
    case FB_PROP_HORIZONTAL_PORCH:
        v = 0; break;
    case FB_PROP_VERTICAL_LENGTH:
        v = pState->ScreenHeight; break;
    case FB_PROP_VERTICAL_VISIBLE_LENGTH:
        v = pState->ScreenVisibleHeight; break;
    case FB_PROP_VERTICAL_PORCH:
        v = 0; break;
    case FB_PROP_PIXEL_BITSIZE:
        v = 8; break;
    case FB_PROP_PIXEL_BYTESIZE:
        v = 1; break;
    case FB_PROP_PIXEL_ENCODING:
        v = PIXEL_TYPE_PALETTE_INDEX; break;
    case FB_PROP_MODE:
        v = pState->CurrentMode; break;
    case FB_PROP_PALETTE_LENGTH:
        v = 256; break;
    case FB_PROP_PALETTE_BITSIZE:
        v = 24; break;
    case FB_PROP_PALETTE_BYTESIZE:
        v = 4; break;
    case FB_PROP_PALETTE_ENCODING:
        v = PIXEL_TYPE_R8G8B8X8; break;
    default:
        return E_NOT_IMPLEMENTED;
    }

    *Value = v;
    return S_OK;
}

HANDLE Fb320Thread;
void _stdcall Fb320ThreadRoutine(void *p);

SCODE WinFbSetMode (PFB320_STATE pState, FRAMEBUFFER_MODE Mode)
{
    /* We cant change mode
     */
    if ((pState->CurrentMode != (FB_MODE_UNDEFINED)) &&
        (Mode != pState->CurrentMode)) {
#if CHANGEMODES
        /* Tell thread to terminate */
        PostMessage(pState->hPaintWnd, WM_QUIT, 0, 0);

        /* Wait till its dead */
        while (pState->CurrentMode != FB_MODE_UNDEFINED)
            Sleep(100); /* msecs */

        CloseHandle(Fb320Thread);
        pState->CurrentMode = 0;
        Fb320Thread = CreateThread(NULL,
                               0,
                               (LPTHREAD_START_ROUTINE)Fb320ThreadRoutine,
                               pState,
                               0,
                               0);

        while (pState->CurrentMode != FB_MODE_UNDEFINED)
            Sleep(100); /* msecs */
#else
        return E_NOT_IMPLEMENTED;
#endif
    }

    /* We know how to do these
     */
    switch (Mode) {
    case FB_MODE_320x200_in_320x400_8bpp_PALETTE:
        pState->ScreenVisibleWidth = 320;
        pState->ScreenVisibleHeight = 200;
        pState->ScreenWidth = 320;
        pState->ScreenHeight = 400;
        break;
    case FB_MODE_640x480_in_768x480_8bpp_PALETTE:
        pState->ScreenVisibleWidth = 640;
        pState->ScreenVisibleHeight = 480;
        pState->ScreenWidth = 768;
        pState->ScreenHeight = 512;
        break;
    default:
        return E_NOT_IMPLEMENTED;
    }

    pState->CurrentMode = Mode;

    /* Now synchronize with the window thread */
    while (pState->pDispBits == NULL)
        Sleep(100); /* msecs */

    return S_OK;
}

PFB320_STATE Fb320New(void)
{
    PFB320_STATE pState;

    pState = (PFB320_STATE) malloc(sizeof(*pState));
    memset(pState,0,sizeof(*pState));

    pState->pBMI = (BITMAPINFO *) malloc (sizeof(BITMAPINFOHEADER)
                                          + (256 * sizeof(RGBQUAD)));
    return pState;
}

void _stdcall Fb320ThreadRoutine(void *p)
{
    PFB320_STATE pState = p;
    int i;

    /* Wait until we know how to create it
     */
    pState->CurrentMode = FB_MODE_UNDEFINED;
    while (pState->CurrentMode == FB_MODE_UNDEFINED)
        Sleep(100); /* msecs */

    //printf("Fb320ThreadRoutine\n");
    StartMeUp(pState);

    if (pState->pDispBits == 0)
        DebugBreak();

    i = SetTimer(pState->hPaintWnd, 12345, 40, NULL);
    if (i == 0) {
        //printf("SetTimer failed sc=%x\n", GetLastError());
    }
    Fb320MsgLoop(pState);

#if CHANGEMODES
    TearMeDown(pState);
    pState->CurrentMode = FB_MODE_UNDEFINED;
#endif
}

/* NB: This could be called by multiple threads in parallel. Sigh.
 */
void WinFbInit(void **ppState)
{
    PFB320_STATE pState;
    volatile ADDRESS *pInitialized;
    TIME_CONSTRAINT MyConstraint;

#define FB_UNINITED 0
#define FB_INITING 1

    /* Done already ? Most popular
     */
    if ((UINT)(*ppState) > FB_INITING)
        return;

    /* Could be called by real time threads, make sure
     * we dont create a priority-inversion problem
     */
    Int32ToInt64(MyConstraint.Start,0);
    Int32ToInt64(MyConstraint.Estimate,1);
    Int32ToInt64(MyConstraint.Deadline,TIME_RELATIVE(100*10));
    MyConstraint.Criticality = _CRITICAL;
    
    BeginConstraint( FALSE, &MyConstraint, NULL);/* nested! */

    if (AtomicCmpAndSwap((PUINT)ppState,
                          FB_UNINITED,
                          FB_INITING)) {

        pState = Fb320New();
        pInitialized = (volatile ADDRESS *) &pState->CurrentMode;

        Fb320Thread = CreateThread(NULL,
                               0,
                               (LPTHREAD_START_ROUTINE)Fb320ThreadRoutine,
                               pState,
                               0,
                               0);

        /* Wait until other thread got far enough */
        while (*pInitialized == 0)
            ;

        *ppState = pState;

        EndConstraint(NULL);
        return;
    }

    EndConstraint(NULL);

    /* Itsa race, make sure we wait
     */
    pInitialized = (volatile ADDRESS *) ppState;
    
    while (*pInitialized == FB_INITING)
        ;


}
