/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Test program for the Atmel AT45DB161 Serial FLASH memory device
 * [as used in the EB63 dev board for the AT91M63200 ARM chip]
 */
#include <stdio.h>
#include <mmlite.h>
#include <stdlib.h>
#include <tchar.h>
#include <ctype.h>
#include <assert.h>
#include <arm\at91m63200.h>

/* The board connects the uC to the device as follows:
 * nCS == PA26 aka NPCS0
 *  SO == PA24 aka MOSI
 *  SI == PA25 aka MISO
 * SCK == PA23 aka SPCK
 * nRESET = NRST
 * RDY == Vcc
 * nWP == Vcc
 * Additionally, a strap E12 can pull nCS high. It must be
 * installed in the 2-3 position to enable the device.
 * This is *not* always the factory setting.
 *
 * There are two ways to talk to the device, using the SPI interface
 * or directly manipulating the I/O pins.  
 * The SPI interface is limited to 1.2 Mbps, the device to 13 Mbps max.
 * Depending on the uC' performance direct manipulation could be faster.
 */

int Verbose = 0;
#define _TEST 0

/* Opcodes
 */
#define SF_PAGE_READ            0x52
#define SF_BUFFER1_READ         0x54
#define SF_BUFFER2_READ         0x56
#define SF_PAGE_CACHE1          0x53
#define SF_PAGE_CACHE2          0x55
#define SF_PAGE_COMPARE1        0x60
#define SF_PAGE_COMPARE2        0x61
#define SF_BUFFER1_WRITE        0x84
#define SF_BUFFER2_WRITE        0x87
#define SF_BUFFER1_FLUSH_PERASE 0x83
#define SF_BUFFER2_FLUSH_PERASE 0x86
#define SF_BUFFER1_FLUSH        0x88
#define SF_BUFFER2_FLUSH        0x89
#define SF_PAGE_ERASE           0x81
#define SF_BLOCK_ERASE          0x50
#define SF_PAGE_PROGRAM1        0x82
#define SF_PAGE_PROGRAM2        0x85
#define SF_PAGE_REWRITE1        0x58
#define SF_PAGE_REWRITE2        0x59
#define SF_READ_STATUS          0x57

/* Status bits (MSB order)
 */
#define SF_ST_READY 0x80
#define SF_ST_COMP  0x40
#define SF_ST_SIZE  0x38  /* should be 0x28 for db161 */
#define SF_ST_xx    0x07

/* Other defines
 */
#define SF_PAGESIZE 528 /* bytes */

/* PIO Interface
 */
#define ThePio PioA
#define nCS PIOA_NPCS0
#define SO  PIOA_MOSI
#define SI  PIOA_MISO
#define SCK PIOA_SPCK
#define WireIsIn(_p_)   ThePio->OutDisable = (_p_)
#define WireIsOut(_p_)  ThePio->OutEnable = (_p_)
#define WireLow(_p_)    ThePio->ClearData = (_p_)
#define WireHigh(_p_)   ThePio->SetData = (_p_)
#define WireIs(_p_)     ((ThePio->PinDataStatus & (_p_)) != 0)
#define WireEnable(_p_) ThePio->Enable = (_p_)

/* Initialization
 */
void InitWires()
{
    /* Before we enable them pins, set safe values.
     * All we really need is to keep the device disabled.
     */
    WireHigh(nCS);

    /* We will use the clock in SPI mode 3
     */
    WireHigh(SCK);

    /* Set proper directions
     */
    WireIsIn(SI);
    WireIsOut(nCS|SO|SCK);
 
    /* Make sure the Pio owns them
     */
    WireEnable(nCS|SI|SCK|SO);

    /* Done
     */
}

/* De/Select the device
 */
void FlashSelect(BOOL Enabled)
{
    if (Enabled) {
        WireLow(nCS);
    } else {
        WireHigh(nCS);
    }
}
#define FlashEnable() FlashSelect(TRUE)
#define FlashDisable() FlashSelect(FALSE)

/* Write a given number of bits to the device
 */
void FlashOutput(PUINT8 Bits, UINT nBits)
{
    UINT i;
    UINT8 c = 0;

    if (Verbose > 1) _tprintf(_T(" >"));

    /* Enable the device.
     * NB: If already enabled this is a no-op.
     */
    FlashEnable();

    /* Should wait 250 ns before lowering SCK.
     * Its a given I suppose, but..
     */
    Delay(0);

    /* Ok, do it one bit at a time
     */
    for (i = 0; i < nBits; i++) {
        /* Fetch next byte if necessary
         */
        if ((i % 8) == 0)
            c = *Bits++;
        
        /* Set the value before we toggle the clock
         * NB: Must be done MSB first, remember
         */
        if (Verbose > 1) _tprintf(_T(" %d"), (c & 0x80) ? 1 : 0);

        if (c & 0x80) {
            WireHigh(SO);
        } else {
            WireLow(SO);
        }

        /* Toggle
         * SCK should stay down a minimum of 35 ns, again a given.
         */
        WireLow(SCK);
        WireHigh(SCK);

        /* The hold time is zero, we dont have to wait.
         * Before we can toggle SCK again we need 35 ns, again..
         */

        /* Next bit
         */
        c <<= 1;
    }    
}

/* Read a given number of bits from the device
 */
void FlashInput(PUINT8 Bits, UINT nBits)
{
    UINT i;
    UINT8 c = 0;

    /* Device should already be selected or else
     * [it only talks to us if we send it a command]
     */
    assert(WireIs(nCS) == 0);

    if (Verbose > 1) _tprintf(_T(" <"));
    /* Ok, do it one bit at a time
     */
    for (i = 0; i < nBits; i++) {

        /* Next bit
         */
        c <<= 1;

        /* Toggle
         * SCK should stay down a minimum of 10 ns, again a given.
         */
        WireLow(SCK);

        /* Get the value before we toggle the clock
         * NB: Must be done MSB first, remember
         */
        c |= (WireIs(SI)) ? 1 : 0;
        WireHigh(SCK);

        if (Verbose > 1) _tprintf(_T(" %d"), (c & 1));

        /* Before we can toggle SCK again we need 20 ns, again..
         */

        /* Store next byte if we are done
         */
        if ((i % 8) == 7)
            *Bits++ = c;
    }    

    /* In the unlikely case we want a weird number of bits..
     */
    if ((i % 8) != 0)
        *Bits++ = c;
    if (Verbose > 1) _tprintf(_T("\r\n"));
}


/* Read the device's status
 */
UINT8 FlashStatus(void)
{
    UINT8 Status;

    Status = SF_READ_STATUS;
    FlashOutput(&Status, 8);
    FlashInput(&Status,8);
    FlashDisable();
    return Status;
}

/* Decode the device's size in bits
 */
UINT32 FlashSize(UINT8 Status)
{
    Status = (Status & SF_ST_SIZE) >> 3;
    return (512*1024) << Status;
}

/* Write one page of data
 */
BOOL FlashPageWrite(UINT PageNumber, UINT8 *Data)
{
    UINT8 CommandBits[4];
    int i, c = 0;

    /* First xfer the bits to Buffer 1
     */
    CommandBits[0] = SF_BUFFER1_WRITE;
    CommandBits[1] = 0;
    CommandBits[2] = 0;
    CommandBits[3] = 0;
    FlashOutput(CommandBits,4*8);
    FlashOutput(Data,SF_PAGESIZE*8);
    FlashDisable();

    /* Next issue the erase&program command
     */
    CommandBits[0] = SF_BUFFER1_FLUSH_PERASE;
    CommandBits[1] = (UINT8)(PageNumber >> 6);
    CommandBits[2] = (UINT8)(PageNumber << 2);
    CommandBits[3] = 0;
    FlashOutput(CommandBits,4*8);
    FlashDisable();

    /* Then wait till its done. Says it should take about 10-20 msecs
     * Dont wait forever
     */
    for (i = 0; i < 10; i++) {
        TIME t;

        /* Is it done yet ?
         */
        c = FlashStatus();
        if (c & SF_ST_READY)
            break;

        /* Nope, sleep a bit
         */
        Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(10)));
        SleepUntil(t);
    }
    return (c & SF_ST_READY);
}

UINT8 Buffer[SF_PAGESIZE];

#if _TEST

/* Read one page of data from memory
 */
BOOL FlashPageRead(UINT PageNumber, UINT8 *Data)
{
    UINT8 CommandBits[8];

    CommandBits[0] = SF_PAGE_READ;
    CommandBits[1] = (UINT8)(PageNumber >> 6);
    CommandBits[2] = (UINT8)(PageNumber << 2);
    CommandBits[3] = 0;
    /* the last 32 bits are dont-care so we dont care */

    FlashOutput(CommandBits,8*8);
    FlashInput(Data,SF_PAGESIZE*8);
    FlashDisable();
    return TRUE;
}

BOOL DumpPage(UINT PageNumber)
{
    int i;
    int c, prev;

    if (!FlashPageRead(PageNumber,Buffer)) {
        _tprintf(_T(" Could not read page no. %d\n\r"), PageNumber);
        return FALSE;
    }
    _tprintf(_T(" Dump of page no. %d"), PageNumber);
    
    prev = -1;
    for (i = 0; i < SF_PAGESIZE; i++) {
        c = Buffer[i];
        if (c == prev) { _tprintf(_T(".")); continue; }
        prev = c;
        _tprintf(_T("\r\n%d: %x "), i, c);
        if (isprint(c))
            _tprintf(_T("'%c'"),c);
    }

    _tprintf(_T("\r\n"));

    return TRUE;
}

UINT8 Buffer2[SF_PAGESIZE];

BOOL WriteTest(void)
{
    _stprintf((_TCHAR*)Buffer,_T("This is a test, it is only a test.\r\n"));
    FlashPageWrite(3,Buffer);
    FlashPageRead(3,Buffer2);
    return (memcmp(Buffer,Buffer2,SF_PAGESIZE) == 0);
}

#endif


/* Write a file to FLASH
 */
BOOL WriteFile(_TCHAR *FileName, ADDRESS Offset)
{
    FILE *f;
    size_t n, i;
    UINT CurPage;
    UINT32 CheckSum;

    /* Make sure offset is a page multiple, warn if not (and good luck)
     */
    CheckSum = Offset;
    CurPage  = Offset / SF_PAGESIZE;
    Offset   = Offset - (CurPage * SF_PAGESIZE);
    if (Offset != 0) {
        _tprintf(_T("WARNING: offset 0x%x is not a page multiple !!\r\n"),
                 CheckSum);
        _tprintf(_T("WARNING: %d==%d*%d+%d is not going to work..\r\n"),
                 CheckSum,SF_PAGESIZE,CurPage,Offset);
    }

    /* Print the warning, we got burned by this.
     */
    _tprintf(_T("Are you sure you want to do this ? "));
    i = _gettchar();
    _tprintf(_T("\r\n"));
    if ((i != _T('y')) && (i != _T('Y'))) {
        _tprintf(_T("No harm done.\r\n"));
        return FALSE;
    }

    /* Open the file
     */
    f = _tfopen(FileName, _T("rb"));
    if (f == NULL) {
        _tprintf(_T("File %s not found.\r\n"),FileName);
        return FALSE;
    }

    /* Ok say we gonna do it
     */
    _tprintf(_T("FLASHing file %s at offset %x [page %u+%u].\r\n"),
             FileName,CheckSum,CurPage,Offset);

    CheckSum = 0;
    for (;;) {
        n = fread(Buffer, 1, sizeof Buffer, f);
        if (n == 0) break;

        for (i = 0; i < n; i++)
            CheckSum += Buffer[i];

        _tprintf(_T("#"));
        for (i = 0; i < 5; i++)
            if (FlashPageWrite(CurPage,Buffer))
                break;
        Offset = 0;
        CurPage++;
    }
    _tprintf(_T(" done.\r\nCheckSum=x%x (%d).\r\n"), CheckSum, CheckSum);

    fclose(f);
    return TRUE;
}

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    int i;
    UINT8 st = 0;
    _TCHAR *FileName;
    ADDRESS Offset;

    InitWires();

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc))
        goto Done;

    /* Make sure the device makes sense and it is not confused
     * Might use the size info too.. sometimes.
     */
    _tprintf(_T("Checking Serial Flash status:"));
    for (i = 0; i < 8; i++) {
        st = FlashStatus();
        _tprintf(_T(" %x"),st);
        if (st & SF_ST_READY)
            break;
        Delay(2000);
    }
    if ((st & SF_ST_READY) == 0) {
        _tprintf(_T(" Something wrong.\n\r"));
        goto Done;
    }
    _tprintf(_T(" (ok)\n\r"));

    _tprintf(_T("Flash size is %d bits\n\r"), FlashSize(st));

#if _TEST
    FileName = NULL;
    Offset = 0;
    for (i = 0; i < 4; i++)
        DumpPage(i);

    if (argc > 1) {
        sc = WriteTest();
        _tprintf(_T("Write test %s.\n\r"), (sc) ? _T("ok") : _T("FAILED"));
    }
#else
    /* Use the same interface of eb_flash, so skip the first arg
     */
    if (argc > 1)
        argv++; argc--;

    if (argc > 1) {
        FileName = argv[1];
        argv++; argc--;
    } else {
        FileName = NULL;
    }

    if (argc > 1) {
        Offset = _tcstoul(argv[1],NULL,16);
        argv++; argc--;
    } else {
        Offset = 0x100000;
    }

    /* Ignore the no-erase, it is not an option here
     */
    if (FileName != NULL) {
        if (!WriteFile(FileName,Offset))
            _tprintf(_T("NOT "));
    }
#endif

    _tprintf(_T("Done.\n\r"));

 Done:
    (void) FreeCmdLine(&argc,&argv);
    return 0;
}

