/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * List the contents of a namespace, recursively
 */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#ifndef MAX_PATH
#define MAX_PATH 127  /* _TCHARs ! */
#endif

/* Return values.
 */
#define rOK 0
#define rQUIT (-1)

/* Lexicon
 */
#define CR '\r'
#define LF '\n'
#define TAB '\t'
#define SP ' '
#define DOT '.'
#define SLASH PATH_SEPARATOR1
#define REM '!'
#define MINUS '-'
#define EOS ((_TCHAR)0)

/* Globals
 */
BOOL fRecurse = FALSE;

/* Strings: error messages, help messages, punctuation
 */
static _TCHAR eUsage[] = _TEXT("Usage: ll.exe [-r] namespace [namespace]*\r\n");
static _TCHAR eMemory[] = _TEXT("Out of memory.\r\n");
static _TCHAR EMPTY[] = _TEXT("");

/* Allocate one longer on account of SLASHes
 */
#define MALLOC(s) malloc((s)+sizeof(_TCHAR))
#define FREE(p) free(p)

/* Free up at the end of evaluation.
 * Its a list and we free it in reverse
 */
void arfree(_TCHAR **argv, int argc)
{
    while (argc-- > 0)
        if (argv[argc] != NULL)
            FREE(argv[argc]);
    FREE(argv);
}

/* Make a copy of a singleton token.
 */
_TCHAR *tokcpy(_TCHAR *tok, int len)
{
    _TCHAR *ntok = MALLOC((len+1)*sizeof(_TCHAR));
    if (ntok != NULL) {
        memcpy(ntok,tok,len*sizeof(_TCHAR));
        ntok[len] = 0;
    }
    return ntok;
}

/* Cleanup whitespaces from a command line.
 * Returns an array of pointers to the individual strings,
 * and a count of them.  Array is NOT NULL-terminated.
 * Caller responsible for freeing up *ptokens memory (with arfree).
 * Also returns the number of characters left unparsed, in case we
 * encounter a CR/LF/EOS.
 */
int tokenize(_TCHAR *line, int *plen, _TCHAR ***ptokens)
{
    int i, n, ll, nchars;
    _TCHAR ch, *p, *firstp;
    BOOL t = FALSE;

    *ptokens = NULL;
    nchars = *plen;
    if ((nchars == 0) || (line[0] == 0))
        return 0;

    /* Count how many tokens, cleanup
     */
    n = 0;
    firstp = NULL;
    for (i = 0; i < nchars; i++) {
        /* Just any joe char ?
         */
        ch = line[i];
        if ((ch == SP) || (ch == TAB) || (ch == CR) ||
            (ch == LF) || (ch == EOS)) {

            /* Change into standard sep for quicker testing later
             */
            line[i] = SP;

            /* Nope, blank or end of parse
             * Do we terminate a token ?
             */
            if (t) {
                n++;
                t = FALSE;
            }

            /* If just blank continue
             */
            if ((ch == TAB) || (ch == SP) || (ch == CR)) {
                continue;
            }
            /* LF/EOS: Terminate parsing here
             * Return count of unparsed
             */
            *plen = nchars - i - 1;
            nchars = i + 1;
            break;
        }
        /* good char
         */
        if (!t) {
            t = TRUE;
            if (firstp == NULL)
                firstp = line + i;
        }
    }
    /* Tokens end here no matter what
     */
    if (t)
        n++;

    /* So does parsing
     */
    if (i == nchars)
        *plen = 0;

    /* Got anything ?  Drop comments.
     */
    if ((n == 0) || (*firstp == REM))
        return 0;

    /* Allocate array for pointers
     */
    *ptokens = MALLOC(n*sizeof(_TCHAR*));
    if (*ptokens == NULL) {
        _putts(eMemory);
        return 0;
    }

    /* Make a second pass, copying tokens
     */
    t = FALSE;
    ll = n = 0;
    p = line;
    for (i = 0; i < nchars; i++) {

        /* Blank ?
         */
        ch = line[i];
        if (ch == SP) {

            /* End of token ?
             */
            if (t) {
 lastok:
                /* Yes, and possibly the last one
                 * Make a copy.
                 */
                p = tokcpy(p,ll);
                if (p == NULL) {
                    /* outtamem, return what we got
                     */
                    return n;
                }
                (*ptokens)[n++] = p;
                t = FALSE;

                /* Last token ?
                 */
                if (i >= nchars)
                    return n;
            }
            /* Done with blank, keep going
             */
            continue;
        }
        /* A good char.  If start of token, remember the start
         * and start counting chars.
         */
        if (!t) {
            p = line+i;
            t = TRUE;
            ll = 1;
        } else {
            /* More chars to this token
             */
            ll++;
        }
    }
    /* If we didnt end clean, go finish up.
     */
    if (t)
        goto lastok;

    /* Clean end
     */
    return n;
}

/*+++++++++++++++++
 * Commands
 */
typedef struct _MOREWORK {
    struct _MOREWORK *Next;
    _TCHAR *Name;
    PINAMESPACE Ns;
} MOREWORK, *PMOREWORK;

PMOREWORK Head, Tail;

void doMore(PINAMESPACE pNs, _TCHAR *Name, int NameLen)
{
    PMOREWORK Work;

    /* Simplistic loop detection
     */
    for (Work = Head; Work; Work = Work->Next) {
        if ((Work->Ns == pNs) &&
            (_tcscmp(Work->Name,Name) == 0))
            return;
    }

    Work = MALLOC(sizeof *Work);
    if (Work == NULL) {
        goto NoMem;
    }

    Work->Name = tokcpy(Name,NameLen);
    Work->Ns = pNs;

    if (Work->Name == NULL) {
        FREE(Work);
        goto NoMem;
    }
    pNs->v->AddRef(pNs);
    
    Work->Next = NULL;
    if (Head == NULL) {
        Head = Work;
    } else
        Tail->Next = Work;
    Tail = Work;

    return;
 NoMem:
    _putts(eMemory);
}


/* ENTRY
 * List one or more entries in a namespace
 */
int doEntry(PINAMESPACE pNs, int argc, _TCHAR **argv)
{
    SCODE sc;
    PIFILE File;
    PINAMESPACE Name;
    UINT NameLen;
    UINT64 FileSize;
    BOOL SizeOk;
    BOOL IsaNs;
#define nDigs 21
    _TCHAR Size[nDigs]; /* 64bits in decimal is 20 digits */

    while (argc > 0) {

        /* Try for a file
         */
        SizeOk = FALSE;
        sc = BindToObject(pNs,
                          argv[0],
                          NAME_SPACE_READ,
                          &IID_IFile,
                          (void**)&File);

        if (!FAILED(sc)) {
            sc = File->v->GetSize(File,&FileSize);
            SizeOk = !FAILED(sc);
            File->v->Release(File);
        }

        /* Also try for a namespace, with and without the slash (sigh).
         */
        NameLen = _tcslen(argv[0]);
        IsaNs = FALSE;
        sc = BindToObject(pNs,
                          argv[0],
                          NAME_SPACE_READ,
                          &IID_INameSpace,
                          (void**)&Name);
        if (FAILED(sc)) {
            argv[0][NameLen] = SLASH;
            argv[0][NameLen+1] = EOS;
            sc = BindToObject(pNs,
                              argv[0],
                              NAME_SPACE_READ,
                              &IID_INameSpace,
                              (void**)&Name);

        }

        if (!FAILED(sc)) {
            /* Remember this for later, if we'll recurse
             */
            IsaNs = TRUE;
            if (fRecurse) {
                /* Stay away from "." and ".."
                 */
                if ((argv[0][0] == DOT) &&
                    ((argv[0][1] == EOS) ||
                     (argv[0][1] == SLASH) ||
                     (argv[0][1] == DOT))) {
                } else {
                    doMore(pNs,argv[0],NameLen+1);
                }
            }

            /* Let it go, for now
             */
            Name->v->Release(Name);
        } else {
            /* Drop the trailing separator
             */
            argv[0][NameLen] = EOS;
        }

        /* Build the size string, either a number or whitespace
         */
        if (SizeOk)
            _stprintf(Size,_T("%20lu"),FileSize);
        else {
            int i;
            for (i = 0; i < nDigs; i++)
                Size[i] = _T(' ');
            Size[nDigs-1] = EOS;
        }

        /* Say what
         */
        _tprintf(_T("%s %s%s\n"), Size, argv[0],
                 IsaNs ? _T("*") : _T(""));

        /* Next one
         */
        argc--;
        argv++;
    }

    return rOK;
}

/* DIR
 * List the content of a namespace
 */
int doDir(PINAMESPACE pRoot, int argc, _TCHAR **argv)
{
    PINAMESPACE pNs;
    _TCHAR *buffer;
    UINT length;
    SCODE sc;

    /* Get a buffer to work with
     */
    buffer = MALLOC(MAX_PATH * sizeof(_TCHAR));
    if (buffer == NULL) {
        _putts(eMemory);
        return rOK;
    }

    while (argc > 0) {

        /* Root itself?
         */
        length = _tcslen(argv[0]);
        if ((length == 1) && (IS_PATH_SEPARATOR(argv[0][0]))) {
            pNs = pRoot;
            pNs->v->AddRef(pNs);
            goto JustDoIt;
        }

        /* List one directory
         */
        sc = BindToObject(pRoot,
                          argv[0],
                          NAME_SPACE_READ,
                          &IID_INameSpace,
                          (void**)&pNs);
        if (FAILED(sc)) {
            if (sc != E_NO_INTERFACE) {
                _tprintf(_T("Namespace %s not found\n"),argv[0]);
            } else {
                doEntry(pRoot,1,argv);
            }
            goto Next;
        }

        /* Drop the trailing slash, if any, before we print it
         */
        if (argv[0][length-1] == SLASH)
            argv[0][length-1] = EOS;

        JustDoIt:
        _tprintf(_T("\nNamespace %s\n\n"),argv[0]);

        /* Get started with the listing
         */
        length = 0;
        sc = pNs->v->FindNext(pNs,
                              EMPTY,
                              NULL,
                              buffer,
                              MAX_PATH-sizeof(_TCHAR),
                              &length);

        /* Loop while we find some
         */
        while (!FAILED(sc) && (length > 0)) {

            /* Say what, one per line
             * Ignore errors.
             */
            buffer[length] = EOS;
            (void) doEntry(pNs,1,&buffer);

            /* Next ?
             */
            sc = pNs->v->FindNext(pNs,
                                  EMPTY,
                                  buffer,
                                  buffer,
                                  MAX_PATH-sizeof(_TCHAR),
                                  &length);
        }
    

        /* Release NS ref
         */
        (void) pNs->v->Release(pNs);

        /* Move on to the next one
         */
    Next:
        argc--;
        argv++;
    }

    /* Free resources and go home
     */
    FREE(buffer);
    return rOK;
}

/* Parse and evaluate a command line
 */
int command(_TCHAR **cmd, int *pcmdlen)
{
    int argc, r;
    _TCHAR **argv, **argv_save;

    /* Split it in pieces
     */
    argc = tokenize(*cmd,pcmdlen,&argv);

    /* Make sure something is left
     */
    argv_save = argv;
 Recheck:
    if (argc == 0)
        return rOK;

    /* Recurse?
     */
    if ((argv[0][0] == MINUS) &&
        ((argv[0][1] == _T('r')) ||
         (argv[0][1] == _T('R')))) {
        fRecurse = TRUE;
        argc--;
        argv++;
        goto Recheck;
    }

    /* Some to do, evaluate it
     */
    r = doDir(CurrentNameSpace(),argc,argv);

    /* Ok, that was the first pass.
     * Now we got to recurse?
     */
    {
        PMOREWORK Current = Head, Temp;
        while (Current != NULL) {

            r = doDir(Current->Ns,1,&Current->Name);

            /* Next in line
             */
            Current = Current->Next;
        }

        /* Now make a second pass releasing all resources.
         * This is needed for loop-avoidance 
         */
        Current = Head;
        while (Current != NULL) {
            /* Release resources
             */
            Temp = Current;
            Current = Current->Next;
            
            Temp->Ns->v->Release(Temp->Ns);
            FREE(Temp->Name);
            FREE(Temp);
        }
    }

    /* Free up resources and go home
     */
    arfree(argv_save,argc);
    return r;
}

/*+++++++++++++++++
 * Main loop
 */

int _tmain(void)
{
    int r = rOK;
    _TCHAR *cmd;
    int cmdlen;

    /* See if we got any command line stuff
     */
    cmd = (_TCHAR*) ProcessArgs();
    if ((cmd == NULL) || (cmd[0] == EOS)) {
        _putts(eUsage);
        return rQUIT;
    }
    cmdlen = _tcslen(cmd);
    r = command(&cmd,&cmdlen);

    return r;
}

