/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>

#ifdef arm
/* CerfBoard specific */
typedef struct {
    volatile UINT32 Level;
    volatile UINT32 Direction;
    volatile UINT32 Set;
    volatile UINT32 Clear;
    volatile UINT32 DetectRaising;
    volatile UINT32 DetectFalling;
    volatile UINT32 DetectStatus;
    volatile UINT32 Alternate;
} GPIO;
#define gpio ((GPIO*)0x90040000)

#define WM_HWBUTTONDOWN  0
#define WM_HWBUTTONUP    1
#define WM_HWBUTTONCLICK 2
#define WM_HWBUTTONQUIT  3
void ButtonHandler(UINT ButtonNo, UINT Event);

#define nBUTTONS 6
BOOL bQuit = FALSE;
int numInts = 0;
static struct _BUTTONS {
        CONDITION ButtonPress;
        UINT ButtonsChanged;
        struct _BUTTON_STATE {
                TIME LastOnAt;
                TIME LastOffAt;
                BOOL On;
                int nCount;
        } ButtonState[nBUTTONS];
} TheButtons = {CONDITION_INITIALIZER, 0,};

static BOOL MCT ButtonIsr(void *This, PBOOL pNotMyInterrupt)
{
        UINT ButtonNo = (UINT)This;
        TIME Now = CurrentTime(), Last;
        static const TIME DEBOUNCE_TIME = Int64Initializer(0,TIME_MILLIS(10));

        // Debounce the hw
        Last = (TheButtons.ButtonState[ButtonNo].On) ?
            TheButtons.ButtonState[ButtonNo].LastOnAt :
            TheButtons.ButtonState[ButtonNo].LastOffAt;
        Last = TimeSubtract(Now,Last);
        if (TimeLess(Last,DEBOUNCE_TIME)) {
            return FALSE;
        }

        //NOTE: We are the Isr, so we are atomic
        TheButtons.ButtonsChanged |= 1<<ButtonNo;

        //NOTE: Record the event time
        if (TheButtons.ButtonState[ButtonNo].On) {
                TheButtons.ButtonState[ButtonNo].LastOffAt = Now;
        }else{
                TheButtons.ButtonState[ButtonNo].LastOnAt = Now;
        }

        TheButtons.ButtonState[ButtonNo].nCount ++;
        numInts++;
        
        //NOTE: Flip the state bit
        TheButtons.ButtonState[ButtonNo].On = !TheButtons.ButtonState[ButtonNo].On;

        //NOTE: Wake up the thread, if necessary
        return Condition_InterruptSignal(&TheButtons.ButtonPress);

}

static void THREAD_LINKAGE ButtonThread(THREAD_ARGUMENT arg)
/*---------------------------------------------------------------------------*\
 * 
\*---------------------------------------------------------------------------*/
{
    struct _BUTTONS *Butts = (struct _BUTTONS *)arg;
    UINT LastSeen = 0, ButtonNo;
    TIME_CONSTRAINT TimeConstraint;
    SCODE sc;

    printf("ButtonThread started (%x)\r\n",(ADDRESS)Butts);

    // Construct a very urgent TimeConstraint and do an
    // initial BeginConstraint
    Int32ToInt64(TimeConstraint.Start, 0);/* can start immediately after wait*/
    Int32ToInt64(TimeConstraint.Estimate, 0); /* does not really matter */
    Int32ToInt64(TimeConstraint.Deadline,0); /* immediately if not sooner */
    TimeConstraint.Criticality = TRUE;
    (void) BeginConstraint(FALSE, &TimeConstraint, NULL);

    /* Power pullups */
    gpio->Set = 1 << 5;         /* make pin J2/19 +5V */
    gpio->Direction |= (1 << 5);   /* output */

    /* Set up ISRs etc */
    printf("Initializing Buttons 1,2,3,4\r\n");
    sc = AddDevice((PTR)0, (PTR)ButtonIsr, (ADDRESS)0, 8, 0);
    sc = AddDevice((PTR)1, (PTR)ButtonIsr, (ADDRESS)0, 9, 0);
    sc = AddDevice((PTR)2, (PTR)ButtonIsr, (ADDRESS)0, 33, 0);
    sc = AddDevice((PTR)3, (PTR)ButtonIsr, (ADDRESS)0, 34, 0);

    for (;!bQuit;) {
        //NOTE: Wait for a button press
        Int32ToInt64(TimeConstraint.Start, 0);
        Int32ToInt64(TimeConstraint.Estimate, 0);
        Int32ToInt64(TimeConstraint.Deadline,0);
        TimeConstraint.Criticality = TRUE;
                
        (void) Condition_WaitAndBeginConstraint(&Butts->ButtonPress,
                                                NULL,
                                                NULL,
                                                TRUE,
                                                NULL/*&TimeConstraint*/,
                                                NULL);
                
        //NOTE: Pick up what happened
        do {
            LastSeen = Butts->ButtonsChanged;
        } while (!AtomicCmpAndSwap(&Butts->ButtonsChanged, LastSeen, 0));
                
       /* Treat 1 + 2 as quit */
        if (Butts->ButtonState[1].On && Butts->ButtonState[2].On) {
            ButtonHandler(0, WM_HWBUTTONQUIT);
            break;
        }
        
        for (ButtonNo=0; ButtonNo<nBUTTONS; ButtonNo++) {
            if (LastSeen & (1<<ButtonNo)) {
                ButtonHandler(ButtonNo, (Butts->ButtonState[ButtonNo].On)
                              ? WM_HWBUTTONDOWN: WM_HWBUTTONUP);
                
                if (TheButtons.ButtonState[ButtonNo].nCount >= 2) {
                    TheButtons.ButtonState[ButtonNo].nCount = 0;

                    ButtonHandler(ButtonNo, WM_HWBUTTONCLICK);
                }
            }
        }
 
    }
        
    //NOTE: Disable button interrupts
    RemoveDevice((PTR)0,8);
    RemoveDevice((PTR)1,9);
    RemoveDevice((PTR)2,33);
    RemoveDevice((PTR)3,34);

    /* Disable pullup current */
    gpio->Clear = 1 << 5;
    gpio->Direction &= ~(1 << 5);

    printf("ButtonThread terminated (%x)\r\n",(ADDRESS)Butts);

}

SCODE InitButtons(void)
{
        PIPROCESS pProcess = CurrentProcess();

        Condition_Init(&TheButtons.ButtonPress);
        return pProcess->v->CreateThread(pProcess, 0, ButtonThread,
                                         (THREAD_ARGUMENT)&TheButtons,
                                         0, NULL, NULL);

}
#else
SCODE InitButtons(void)
{
    printf("InitButtons: No buttons on this device\n");
    return E_NOT_IMPLEMENTED;
}
#endif
