/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <tchar.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <winsock.h>

#include <diagnostics.h>

#define LINK_AS_EXE 1

/* Make an IFile for a PERSistent SOCKet connection
 */
typedef struct _CPersock {
  const struct IFileVtbl *v;
  UINT RefCnt;

  int State;
#define UNCONNECTED 0
#define CONNECTED 1

  SOCKET Socket;
  _TCHAR *HisName;
  SOCKADDR_IN MyPort;
  SOCKADDR_IN HisPort;
#if LINK_AS_EXE
  PIMODULE MyModule;
#endif

} CPersock, *PCPersock;

/* forwards */
PRIVATE BOOL MakeConnected(PCPersock this);
PRIVATE BOOL MakeDisconnected(PCPersock this);
PRIVATE BOOL ResolveName(PCPersock this);



/* IFILE implementation
 */

/* Look for some interface
 */
PRIVATE SCODE MCT PersockQueryInterface (PIFILE pThis, REFIID Iid, void* *ppNew)
{
    PCPersock this = (PCPersock)pThis;
    return GenericQueryInterface((PIUNKNOWN) this, Iid, ppNew,
                                 &IID_IFile);
}

/* Add a reference
 */
PRIVATE UINT MCT PersockAddRef ( PIFILE pThis )
{
    PCPersock this = (PCPersock)pThis;
    return AtomicInc(&this->RefCnt);
}

/* Release a reference
 */
PRIVATE UINT MCT PersockRelease ( PIFILE pThis )
{
    PCPersock this = (PCPersock)pThis;
    UINT Refs;

    Refs = AtomicDec(&this->RefCnt);
    if (Refs == 0) {
#if LINK_AS_EXE
        PIMODULE pMod = this->MyModule;
#endif
        if (this->Socket != INVALID_SOCKET)
            closesocket(this->Socket);
        _tprintf(_T("Persock(%s:%d) deleted\r\n"),
               this->HisName,htons(this->HisPort.sin_port));
        free(this);
#if LINK_AS_EXE
        return pMod->v->Release(pMod);
#endif
    }
    return Refs;
}


PRIVATE SCODE MCT PersockReadAt (PIFILE pThis, UINT64 Position, BYTE *pBuffer, UINT Size, PUINT pSizeRead)
{
    PCPersock this = (PCPersock)pThis;
    int Count;

    DBGME(0,_tprintf(_T("PersockReadAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeRead));

    while (this->State != CONNECTED)
        MakeConnected(this);

    Count = recv(this->Socket,(char*)pBuffer,Size,0);
    if (Count < 0) {
        Count = WSAGetLastError();
        /* Handle timeouts silently
         */
        if (Count == WSAETIMEDOUT)
            return E_TIMED_OUT;

        _tprintf(_T("recv failed (%d) disconnecting.\n"),Count);
        MakeDisconnected(this);
        return E_RESTART;
    }

    if (pSizeRead)
        *pSizeRead = Count;

    return NO_ERROR;
}

PRIVATE SCODE MCT PersockWriteAt (PIFILE pThis, UINT64 Position, const BYTE *pBuffer, UINT Size, PUINT pSizeWritten)
{
    PCPersock this = (PCPersock)pThis;
    int Count;

    DBGME(0,_tprintf(_T("PersockWriteAt(%x %ld %x %d %x)\n"),
                     (ADDRESS) this, Position, (ADDRESS) pBuffer,
                     Size, (ADDRESS) pSizeWritten));

    if (pSizeWritten)
        *pSizeWritten = 0;

    /* Make one attempt to connect if not yet
     */
 Again:
    if ((this->State != CONNECTED) &&
        !MakeConnected(this)) {
        return E_RESTART;
    }

    /* Ok, send it out
     */
    /* Assume NAGLE */
    Count = send(this->Socket,(const char *)pBuffer,Size,0);

    /* It might have been broken or somehting
     */
    if (Count < 0) {

        /* did our partner leave?
         */
        Count = WSAGetLastError();
        MakeDisconnected(this);

        if ((Count == WSAECONNRESET) || (Count == WSAESHUTDOWN)) {
            _tprintf(_T("lost TCP connection to %s, retrying..\n"),
                     this->HisName);
            goto Again;
        } else
            _tprintf(_T("send to %s failed (%d)\n"),
                           this->HisName, Count);
        return E_RESTART;
    }

    if (pSizeWritten)
        *pSizeWritten = Count;
    return NO_ERROR;
}

PRIVATE SCODE MCT PersockSetSize (PIFILE pThis, UINT64 Size)
{
    PCPersock this = (PCPersock)pThis;
    UINT32 *p = (UINT32*)&Size;
    int millis;/* the NT way */

    if ((this->State != CONNECTED) || (*p > 1))
        return E_INVALID_PARAMETER;

    millis = p[1] / TIME_MILLIS(1);
    millis = setsockopt(this->Socket, SOL_SOCKET,
                       (*p == 0) ? SO_RCVTIMEO : SO_SNDTIMEO,
                        (const char *)&millis, sizeof millis);

    return (millis == 0) ? NO_ERROR : E_FAIL;
}

PRIVATE SCODE MCT PersockGetSize (PIFILE pThis, PUINT64 pSize)
{
    /*PCPersock this = (PCPersock)pThis;*/
    return E_NOT_IMPLEMENTED;
}

const struct IFileVtbl PersockVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    PersockQueryInterface,
    PersockAddRef,
    PersockRelease,
    PersockReadAt,
    PersockWriteAt,
    PersockSetSize,
    PersockGetSize
};

/* Constructor
 */
PIFILE PersockNew(_TCHAR *HisName, UINT HisPort)
{
    PCPersock this;
    int len = _tcslen(HisName);

    /* allocate, check and zero memory
     */
    this = malloc(len + sizeof(_TCHAR) + sizeof *this);
    if (this == NULL)
        return NULL;
    memset(this,0,sizeof *this);

    /* Initialize object
     */
    this->v = &PersockVtbl;
    this->RefCnt = 1;
    this->Socket = INVALID_SOCKET;
    this->HisName = (_TCHAR *)(this+1);
    _tcscpy(this->HisName,HisName);
    this->HisPort.sin_port = htons((UINT16)HisPort);

#if LINK_AS_EXE
    /* Take a reference on this image, it is not a COB
     */
    {
        SCODE sc;
        sc = GetModuleContainsAddr((ADDRESS)PersockNew, &this->MyModule);
        /* Ignore errors, no idea how to cope */
    }

#endif

    /* Rest is done dynamically 
     */

    return (PIFILE)this;
}

/* Worker functions
 */
PRIVATE BOOL ResolveName(PCPersock this)
{
    SOCKADDR_IN *to;
    struct hostent *hp;

    to = (SOCKADDR_IN *)&this->HisPort;
    to->sin_family = AF_INET;
    to->sin_len = 16;
    to->sin_addr.s_addr = inet_addr((char*)this->HisName); /* IP address ? */

    if (to->sin_addr.s_addr == (UINT)-1) {
        to->sin_addr.s_addr = 0;
        hp = gethostbyname((char*)this->HisName);          /* nope */
        if (!hp) {
            _tprintf(_T("Unknown host %s\n"), this->HisName);
            return FALSE;
        }

        to->sin_family = (UINT8) hp->h_addrtype;
        to->sin_len = (UINT8) hp->h_length;
        memcpy((PTR)&to->sin_addr,hp->h_addr,hp->h_length);
    }
    return TRUE;
}

PRIVATE BOOL MakeConnected(PCPersock this)
{
    int len = sizeof this->HisPort;
    int error;

    /* Who do we talk to ?
     */
    if ((this->HisPort.sin_addr.s_addr == 0) && !ResolveName(this)) {
        /* ...noise */
        return FALSE;
    }

    /* Get us a socket
     */
    this->Socket = socket( AF_INET, SOCK_STREAM, 0);
    if (this->Socket == INVALID_SOCKET) {
        /* ...noise */
        return FALSE;
    }

    /* Bind it (to ANY)
     */
    memset(&this->MyPort,0,sizeof this->MyPort);
    this->MyPort.sin_family = AF_INET;
    this->MyPort.sin_len = 16;
    this->MyPort.sin_port = 0;

    error = bind(this->Socket,
                 (const struct sockaddr *)&this->MyPort,
                 sizeof(this->MyPort));
        
    if (error) {
        /* warn but ignore */
        _tprintf(_T("Warning: bind failed err=%d\r\n"), WSAGetLastError());
    }

    /* Attempt to connect to the indicated server
     * BUGBUG Make this non blocking 
     */
    error = connect(this->Socket,
                   (struct sockaddr *)&this->HisPort,
                    len);

    /* If error bail out
     */
    if (error != 0) {
        error = WSAGetLastError();
        MakeDisconnected(this);
        _tprintf(_T("%s: connect failed (%d)\n"),
                           this->HisName, error);
        return FALSE;
    }

    /* Say who we are talking to
     */
    _tprintf(_T("[connected with %s port %d]\r\n"),
                       inet_ntoa(this->HisPort.sin_addr),
                       htons(this->HisPort.sin_port));

    /* Lets dance
     */
    this->State = CONNECTED;
    return TRUE;
}

PRIVATE BOOL MakeDisconnected(PCPersock this)
{
    closesocket(this->Socket);
    this->Socket = INVALID_SOCKET;

    this->State = UNCONNECTED;
    return TRUE;
}


/* User interface (sic;-)
 */
int _tmain(INT argc, _TCHAR *argv[])
{
    _TCHAR *why, *Ip, *Persname;
    UINT Port;
    PIFILE this = NULL;
    PINAMESPACE pNs = CurrentNameSpace();
    SCODE sc;

    /* Parse and verify args
     */
    ParseCmdLine(&argc,&argv);

    if (argc != 4) {
        why = _T("Usage: <hostname/IP address> <portnumber> <IFile name>");
        sc = E_INVALID_PARAMETER;
        goto ErrorExit;
    }

    Ip = argv[1];
    Port = _tcstoul(argv[2],NULL,10);
    Persname = argv[3];

    /* Build our object
     */
    this = PersockNew(Ip,Port);
    if (this == NULL) {
        why = _T("PersockNew failed");
        sc = E_NOT_ENOUGH_MEMORY;
        goto ErrorExit;
    }

    /* Put it in the namespace
     */
    sc = pNs->v->Register(pNs, Persname, (PIUNKNOWN) this, 0, NULL);

    /* Release this regardless
     */
    this->v->Release(this);
    this = NULL;

    /* Now check if the registration worked
     */
    if (FAILED(sc)) {
        why =_T("Failed to register");
        goto ErrorExit;
    }

    /* And we are done
     */
    goto Exit;

ErrorExit:
    _tprintf(_T("%s, sc=%x\r\n"), why, sc);
    _tprintf(_T("This test FAILED\r\n"));
Exit:
    FreeCmdLine(&argc,&argv);
    return 0;
}
