/* Copyright (c) Microsoft Corporation. All rights reserved. */
    
/* SNTPv4 client implementation

   The client is implemented as a COB with all functionality 
   accessible through ISntpClient. 

   It connects to a remote SNTP server to set the local clock
   to match the server's clock.

   The client can be invoked directly to set the local clock, 
   compute the offset from the server, etc., or it can run as 
   a persistent service using a worker thread to periodically 
   synchronize the local clock with the server. 

   The SNTP time format is 64 bits, the first 32 bits denote the
   seconds and the second 32 bits denote the fraction of second.

   The MMLite time format is also 64 bits but has less fidelity. 
   Hence not all clock values that can be expressed in the 
   MMLite format can be expressed in the SNTP format.
   
   We check for this when the SNTP client is initialized, and
   if the local clock is outside of the SNTP range then we reset 
   it using the function SanityCheckLocalClock, to a value that 
   can be represented using the SNTP time format.
   
   The synchronization thread and start/stop functions use a 
   mutex to protect themselves; all the other functions just 
   ignore multi-threading. So don't use them if the sync thread
   is running... */

/* How to set Windows XP up as a time server:
        http://support.microsoft.com/kb/314054/EN-US/ 
        
   But! Why bother, since Microsoft runs a Stratum 1 NTP/SNTP
   server we can use instead? It's at 172.31.40.19. See
   
        http://msrinfo/msrsupp/corpnet/timeserver.htm 
        
   It's a dedicated box sitting on the 1st floor of 113, receives
   a time feed from a satellite, and sometimes gets wedged and 
   needs a reset - I lost a half-hour debugging the receive-reply
   code because the server had crashed. 

   Since the server is behind the MS firewall this code needs 
   to be recompiled with a different value for DEFAULT_TIME_SERVER
   to work on the outside... */

#include "sntp.h"

#include <soapif.h>

/* ISntpClient implementation */
    
PRIVATE SCODE MCT SntpClientQueryInterface(PISNTPCLIENT iThis, REFIID Iid, void** Unk)
{
    SCODE sc;
    SntpClient *This = pSntpClient(iThis);

    sc = GenericQueryInterface((IUnknown *) iThis, Iid, Unk, &IID_ISntpClient);
    DBGME(0, printf("SntpClientQI sc=%x this=%x unk=%x\n", 
        sc, (ADDRESS) This, (ADDRESS) *Unk));

    return sc;
}

PRIVATE UINT MCT SntpClientAddRef(PISNTPCLIENT iThis)
{
    UINT Refs;
    SntpClient *This = pSntpClient(iThis);

    Refs = AtomicInc(&This->Refs);

    DBGME(0, printf("SntpClientAddRef: Refs=%d\n", Refs));
    return (Refs);
}

PRIVATE UINT MCT SntpClientRelease(PISNTPCLIENT iThis)
{
    UINT Refs;
    SntpClient *This = pSntpClient(iThis);

    Refs = AtomicDec(&This->Refs);

    if (Refs)
        return Refs;

    DBGME(0, printf("SntpClientRelease: Refs=%d\n", Refs));

    if (INVALID_SOCKET != This->Socket) {
        closesocket(This->Socket);
        This->Socket = INVALID_SOCKET;
    }

    if (NULL != This->ServerName) {
        CurrentHeap()->v->Free(CurrentHeap(), 0, This->ServerName);
        This->ServerName = NULL;
    }

    Mutex_Destroy(&This->Mx);

    return (Refs);
}
    
/* Just to let mutex-holding functions look up the adjusted current time */
PRIVATE SCODE MCT SntpClientCurrentTimeNoLock(PISNTPCLIENT iThis, PINT64 Time)
{
    METHOD_DECLARE(SntpClient, CurrentTimeNoLock);
    TIME adjusted;

    CHECKSETSC(!Int64Equal(TIME_ORIGIN, This->LocalClockOffset), E_FAIL,
               "look up time; local clock offset not known");

    CHECKSETSC(This->ThreadIsRunning, E_FAIL,
               "look up time; client not started");

    adjusted = Int64Add(CurrentTime(), This->RangeAdjustment);

    /* The LocalClockOffset is stored as a negative number, because it 
       is a relative time. Hence Int64Add if the offset should be 
       subtracted from the adjusted time, and Int64Subtract if the 
       offset should be added... */

//    DBGME(0, 
        printf("adjusted=0x%llx offset=0x%llx subtract=0x%llx add=0x%llx bool=%d\n", 
               adjusted, This->LocalClockOffset, 
               Int64Subtract(adjusted, This->LocalClockOffset), 
               Int64Add(adjusted, This->LocalClockOffset), 
               This->SubtractOffset);//);

    *Time = (This->SubtractOffset) 
        ? Int64Add(adjusted, This->LocalClockOffset)
        : Int64Subtract(adjusted, This->LocalClockOffset);
    
    DBG(("sntp.c:SntpClientCurrentTime: \n\ttime=0x%llx\n", *Time));

    sc = S_OK;

METHOD_CLEANUP;
    return sc;
}
    
PRIVATE SCODE MCT SntpClientCurrentTime(PISNTPCLIENT iThis, PINT64 Time)
{
    METHOD_DECLARE(SntpClient, CurrentTime);

    Mutex_Lock(&This->Mx);
    sc = SntpClientCurrentTimeNoLock(iThis, Time);
    goto Out;

METHOD_CLEANUP;
    Mutex_Unlock(&This->Mx);
    return sc;
}
    
/* Translate from the MMLite time format to the SNTP time format. 

   The MMLite time format is the Windows time format, a 64-bit signed 
   integer with 0x0 defined as 1601 January 1 0:00:00 and a time unit of
   100 ns. Negative values are considered relative, but can be ignored
   here. 

   The SNTP time format is described in RFC-1305. It is a 64-bit unsigned 
   fixed-point number with the first 32 bits denoting seconds and the 
   rest denoting the fraction of second. This is a much higher granularity, 
   but the SNTP format rolls over much sooner. The SNTP format defines
   0x0 as 1900 January 1 0:00:00. 

   How to compute the conversion?

   Let SNTP(t) be the SNTP representation of the time t, and MML(t) the 
   MMLite representation. Each is 64 bits, unsigned. Then 

       MML(t)   = a * SNTP(t) + b
       SNTP(t)  = (MML(t) - b)/a

    These values are calculated by hand and used to derive the scaling 
    and offset constants:

    SNTP("1900 January 1  0:00:00") = 0x 0
     MML("1900 January 1  0:00:00") = 0x 014F373BFDE04000

    SNTP("1968 January 18 3:14:08") = 0x 8000000000000000
     MML("1968 January 18 3:14:08") = 0x 019B9918FE2CA1DD

    SNTP("2005 January 1 0:00:00") = 0x C5806680 00000000
     MML("2005 January 1 0:00:00") = 0x 01c4ef94 d6228000

    The offset is

        b = 0x 014F373BFDE04000

    with the scaling factor 

      1/a' = 0x8000000000000000 / 0x004c4b4000000000
          = 0x1ad (hex) 
          = 429 (dec)

    This has insufficient precision, with rounding error in 
    excess of 1 second. The correct scaling factor is 

        a = a' + e
    
    We improve a' by left-shifting the scaling constant by 16 
    bits, and approximate the error constant e with a series of 
    negative powers of two (derived by hand... slowly...). 
    
    The truncated scaling factor is 0x01AD; too much rounding error.
    Left-shifted 16 bits it becomes 0x01AD7F29.

    Note that the MMLite times must be absolute, e.g. they're
    actually signed integers but the high-order bit must not 
    be set. 

    Note also that the Windows SNTP client control utility can be 
    used to convert Windows and SNTP time epochs to human readable
    strings; invoke w32time/? in a command prompt. (Of course I 
    didn't discover this until after I was done...)
*/
UINT64 OFFSET = Int64Initializer(0x014F373B, 0xFDE04000);

#define MML_SCALING     0x01AD7F29 
UINT32 MML_CORRECTIONS[] 
    = { 17, 19, 21, 23, 24, 25, 26, 29, 31, 33, 34, 35, 36, 38, 41, 45 };
#define NUM_MML_CORRECTIONS sizeof(MML_CORRECTIONS)/sizeof(UINT32)
    
/* XXX This should return an SCODE, because errors are possible */
UINT64 MCT TranslateAbsoluteMMLiteTimeToSntp(TIME mmlTime)
{
    UINT64 distance, scaled, shifted;
    UINT64 corrected, error = Int64Initializer(0x0, 0x0);
    INT i;

    /* Check whether the input can be translated to SNTP */
    if (TimeIsRelative(mmlTime)) {
        DBGME(3, printf("sntp.c:Translate...ToSntp: bad input=0x%llx (negative)\n", 
            mmlTime));
        return UINT64_ZERO;
    }
        
    if (Uint64Less(*(UINT64*) &mmlTime, OFFSET)) {
        /* Oops, can't */
        DBGME(3, printf("sntp.c:Translate...ToSntp: input=0x%llx outside SNTP time format range\n", 
             mmlTime));
        return UINT64_ZERO;
    }

    /* SNTP turned the high-order bit (bit 64) on in 1968. 
       We right-shift 16 bits and lose the less significant half
       of the fraction, to make up for left-shifting the scaling
       factor */

    /* Convert using SNTP(t)  = (MML(t) - b)/a */
    distance = Uint64Subtract(*(UINT64*) &mmlTime, OFFSET);
    shifted = Uint64RShift(distance, 16);
    scaled = Uint64TimesUint32(shifted, MML_SCALING);

    /* Approximate the round-off error as a fraction of MML(t) - b */
     for (i = 0; i < NUM_MML_CORRECTIONS; ++i)
        error = Uint64Add(error, Uint64RShift(distance, MML_CORRECTIONS[i]));

    /* Correct the error, SNTP(t) = (MML(t) - b)/a + ERROR */
    corrected = Uint64Add(scaled, error);

    DBGME(0, printf("distance=0x%llx scaled=0x%llx error=0x%llx corrected=0x%llx\n", 
        distance, scaled, error, corrected));

    return corrected;
}
    
/* Translate from the SNTP time format to the MMLite time format. 
   See comments about TranslateMMLiteTimeToSntp */

#define SNTP_SCALING     MML_SCALING 

UINT32 SNTP_CORRECTIONS[] = { 35, 36, 37, 38, 40, 45 };
#define NUM_SNTP_CORRECTIONS sizeof(SNTP_CORRECTIONS)/sizeof(UINT32)

TIME MCT TranslateSntpTimeToAbsoluteMMLite(UINT64 sntpTime)
{
    UINT64 scaled, shifted, moved; 
    UINT64 corrected, error = Int64Initializer(0x0, 0x0);
    INT i;

    /* All input values are allowed */

    /* Convert using MML(t) = a * SNTP(t) + b */
    scaled = Uint64DividedByUint32(sntpTime, SNTP_SCALING);
    shifted = Uint64LShift(scaled, 16);
    moved = Uint64Add(shifted, OFFSET);

    /* Approximate the rounding error */
    for (i = 0; i < NUM_SNTP_CORRECTIONS; ++i) 
        error = Uint64Add(error, Uint64RShift(sntpTime, SNTP_CORRECTIONS[i]));

    /* Correct for the rounding error */
    corrected = Uint64Subtract(moved, error);

    return *(TIME*) &corrected;
}
    
/* Prepare a sockaddr structure for the given server */
PRIVATE SCODE MCT CreateSockAddr(const _TCHAR *Server, struct sockaddr_in *sockaddr)
{
    METHOD_DECLARE_NOTHIS(sntp.c, CreateSockAddr);
    UINT32 ipaddr = _tinet_addr(Server);
    if (INADDR_NONE != ipaddr) {
        sockaddr->sin_addr.s_addr = ipaddr;
#if 0 /* XXX does not link with winshim, stdcall linkage issue */
    } else {
        struct hostent *hostentry;
        if (hostentry = gethostbyname(Server)) {  
            memcpy(&sockaddr->sin_addr,
                   hostentry->h_addr, 
                   hostentry->h_length);
        } else {
            CHECKSETSC(FALSE, E_FAIL, "gethostbyname");
        }
#endif
    } else {
        CHECKSETSC(FALSE, E_FAIL, "look up SNTP server");
    }
    sockaddr->sin_family = AF_INET;
    sockaddr->sin_port = htons(123);

    sc = S_OK;

METHOD_CLEANUP;
    return sc;
}
    
/* Prepare the NTP request buffer (see RFC 2030) */
SCODE MCT CreateSntpRequest(PISNTPCLIENT iThis, UINT32 *Buffer, UINT Len) 
{
    METHOD_DECLARE(SntpClient, CreateSntpRequest);
    UINT32 *fields, tmp;
    TIME adjusted;

    assert(SNTP_REQUEST_BYTES == Len);

    /* Pack the bits */
    memset(Buffer, 0, SNTP_REQUEST_BYTES);
    fields = Buffer;

    /* LI:          0                               2 bits
       VN:          4   (for SNTPv4)                3 bits
       Mode:        3   (client request)            3 bits
       Stratum:     15  (lowest-level client)       8 bits
       Poll:        10  (max poll interval 1024 s)  8 bits 
       Precision:   -6  (local clock, ~10 ms)       8 bits 
       
       |0010|0011|0000|1111|0000|1010|1000|0110|
         2    3    0    f    0    a    8    6
       */
    *fields++ = 0x230f0a86;

    /* Root delay:  0 */
    /* Root dispersion: 0 */
    /* Reference identifier: 0 */
    fields += 3;

    /* Reference timestamp: 0 */
    /* Originate timestamp: 0 - this is correct */
    /* Receive timestamp: 0 */
    fields = (void*) (((UINT64 *) fields) + 3);

    /* Get the transmit timestamp */
    adjusted = Int64Add(CurrentTime(), This->RangeAdjustment);
    *((UINT64 *) fields) = TranslateAbsoluteMMLiteTimeToSntp(adjusted);
    /* 64-bit swap to network byte order */
    if (BYTE_ORDER != BIG_ENDIAN) {
        tmp = *fields;
        *fields = htonl(*(fields+1));
        *(fields+1) = htonl(tmp);
    }
    fields += 2;

    /* Done - the caller has to release the buffer! */
    assert(Buffer + SNTP_REQUEST_WORDS == fields);

    sc = S_OK;

METHOD_CLEANUP;
    return sc;
}
    
/* Send the SNTP request to the server */
SCODE MCT SntpClientSendRequest(
    PISNTPCLIENT iThis,
    const _TCHAR *Server)
{
    METHOD_DECLARE(SntpClient, SendRequest);
    UINT32 Request[SNTP_REQUEST_WORDS];
    int i; 

    CHECKSETSC(iThis, E_INVALID_PARAMETER, "send; client is null");
    CHECKSETSC(Server, E_INVALID_PARAMETER, "send; no time server");

    /* Lazy address and socket initialization */
    if (INVALID_SOCKET == This->Socket) {
        sc = CreateSockAddr(Server, &This->SockAddr);
        CHECKSC("create sockaddr");
        This->Socket = socket(AF_INET, SOCK_DGRAM, 0);
        CHECKSETSC(INVALID_SOCKET != This->Socket, E_FAIL, "no socket");
    }       

    /* This function must deallocate the request buffer */
    sc = CreateSntpRequest(iThis, Request, SNTP_REQUEST_BYTES);
    CHECKSC("create SNTP request");

    DBGME(0, printf("Request buffer:\n"));
    for (i = 0; i < SNTP_REQUEST_WORDS; ++i)
        DBGME(0, printf("\t0x%x\n", ntohl(Request[i])));

    /* Send the request */
    DBGME(1, printf("sntp.c:SendRequest: Sending request to %s:%d ...\n",
         inet_ntoa(This->SockAddr.sin_addr),
         ntohs(This->SockAddr.sin_port)));

    if (0 > sendto(This->Socket, (const char *) Request, SNTP_REQUEST_BYTES, 0, 
                   (PSOCKADDR) &This->SockAddr, sizeof(This->SockAddr)))
        CHECKSETSC(FALSE, E_FAIL, "send request");

    DBGME(1, printf("sntp.c:SendRequest: success\n"));

METHOD_CLEANUP;
    /* Request buffer is stack-allocated */
    // if (Request)
    //     CurrentHeap()->v->Free(CurrentHeap(), 0, (PTR) Request);
    /* If something goes wrong we discard the cached socket */
    if (FAILED(sc)) { 
        if (This && INVALID_SOCKET != This->Socket) {
            closesocket(This->Socket);
            This->Socket = INVALID_SOCKET;
        }
    }
    return sc;
}
    
/* Receive the SMTP reply from the server */
SCODE MCT SntpClientReceiveReply(
    PISNTPCLIENT iThis, 
    UINT64 *ot, UINT64 *rt, UINT64 *tt, UINT64 *dt, 
    PTIME timestamp)
{
    METHOD_DECLARE(SntpClient, ReceiveReply);
    UINT32 Buffer[SNTP_REQUEST_WORDS];
    INT Len = -1, i, SockAddrSz = sizeof(This->SockAddr);
    TIME adjusted;

    CHECKSETSC(INVALID_SOCKET != This->Socket, E_FAIL, "no socket");

    /* Wait for the reply */
    DBGME(1, printf("Waiting for reply...\n"));
    memset(Buffer, 0, SNTP_REQUEST_BYTES);
    Len = recvfrom(This->Socket, 
        (void *) Buffer, SNTP_REQUEST_BYTES, 0, 
        (PSOCKADDR) &This->SockAddr, &SockAddrSz);
    CHECKSETSC(SNTP_REQUEST_BYTES == Len, E_FAIL, "failed to receive");

    /* Create the destination timestamp as soon as the reply arrives */
    *timestamp = CurrentTime();
    adjusted = Int64Add(*timestamp, This->RangeAdjustment);
    *dt = TranslateAbsoluteMMLiteTimeToSntp(adjusted);

    DBGME(0, printf("Reply buffer:\n"));
    for (i = 0; i < SNTP_REQUEST_WORDS; ++i)
        DBGME(0, printf("\t0x%x\n", ntohl(Buffer[i])));

    /* Extract the timestamps */
    if (BYTE_ORDER == BIG_ENDIAN) {
        /* No byteswap necessary */
        *ot = *((UINT64 *) (Buffer + SNTP_ORIGINATE_TIMESTAMP_OFFSET));
        *rt = *((UINT64 *) (Buffer + SNTP_RECEIVE_TIMESTAMP_OFFSET));
        *tt = *((UINT64 *) (Buffer + SNTP_TRANSMIT_TIMESTAMP_OFFSET));
    } else {
        /* Have to do full 64-bit byteswap */
        UINT32 high, low, *next = Buffer + SNTP_ORIGINATE_TIMESTAMP_OFFSET;
        high = ntohl(*next); low = ntohl(*(next + 1));
        *((UINT32*) ot) = low; *(((UINT32*) ot) + 1) = high;
        next = Buffer + SNTP_RECEIVE_TIMESTAMP_OFFSET;
        high = ntohl(*next); low = ntohl(*(next + 1));
        *((UINT32*) rt) = low; *(((UINT32*) rt) + 1) = high;
        next = Buffer + SNTP_TRANSMIT_TIMESTAMP_OFFSET;
        high = ntohl(*next); low = ntohl(*(next + 1));
        *((UINT32*) tt) = low; *(((UINT32*) tt) + 1) = high;

        /* Better version but doesn't compile with ARM/gcc
        UINT32 *next = Buffer + SNTP_ORIGINATE_TIMESTAMP_OFFSET;
        *ot = Int64Initializer(ntohl(*next), ntohl(*(next + 1)));
        next = Buffer + SNTP_RECEIVE_TIMESTAMP_OFFSET;
        *rt = Int64Initializer(ntohl(*next), ntohl(*(next + 1)));
        next = Buffer + SNTP_TRANSMIT_TIMESTAMP_OFFSET;
        *tt = Int64Initializer(ntohl(*next), ntohl(*(next + 1)));
        */
    }

    DBGME(0, printf("After byteswapping:\n"));
    DBGME(0, printf("\tOriginate timestamp=0x%llx\n", *ot));
    DBGME(0, printf("\tReceive timestamp=0x%llx\n", *rt));
    DBGME(0, printf("\tTransmit timestamp=0x%llx\n", *tt));
    DBGME(0, printf("\tDestination timestamp=0x%llx\n", *dt));

    DBGME(1, printf("sntp.c:ReceiveReply: success\n"));

METHOD_CLEANUP;
    /* Discard the cached socket if we failed */
    if (FAILED(sc) && This && INVALID_SOCKET != This->Socket) {
        closesocket(This->Socket);
        This->Socket = INVALID_SOCKET;
    }
    return sc;
}
    
/* Compute the offset in SNTP time format. 
   ot: Originate timestamp
   rt: Receive timestamp
   tt: Transmit timestamp
   dt: Destination timestamp
   Returns ((rt - ot) + (tt - dt))/2 */
SCODE MCT ComputeSntpFormatOffset(
    UINT64 ot, UINT64 rt, UINT64 tt, UINT64 dt, 
    PUINT64 result, PBOOL negative)
{
    /* Instead of dividing by two at the end we do it up front, 
       which shaves off a bit and sets the high-order bit to zero, 
       which is handy because the arithmetic is much easier with 
       signed integers... */

    INT64 ot2, rt2, tt2, dt2, term1, term2, tmp;
    BOOL term1neg, term2neg;

    ot = Uint64RShift(ot, 1);
    rt = Uint64RShift(rt, 1);
    tt = Uint64RShift(tt, 1);
    dt = Uint64RShift(dt, 1);

    ot2 = *(INT64*) &ot; 
    rt2 = *(INT64*) &rt;
    tt2 = *(INT64*) &tt;
    dt2 = *(INT64*) &dt;

    term1 = Int64Subtract(rt2, ot2);
    term2 = Int64Subtract(tt2,  dt2);

    /* Figure out the result and the sign bit. Yay, 65 bit arithmetic */
    term1neg = Int64IsNegative(term1);
    term2neg = Int64IsNegative(term2);
    *negative = FALSE;
    switch (term1neg + term2neg) {
        case 2: /* Both negative */
            *negative = TRUE;
            term1 = Int64TimesInt32(term1, -1);
            term2 = Int64TimesInt32(term2, -1);
            /* Fall through */
        case 0: /* Both positive */
            /* Adding two positive numbers that may have all but the highest 
               bit set can result in a number that requires all 64 bits, so 
               switch to UINT64 now */
            *result = Uint64Add(*(UINT64*) &term1, *(UINT64*) &term2);
            /* Already dealt with *negative */
            break;
        case 1: /* Signs differ  */
            /* INT64 arithmetic is sufficient since the highest bit can't be set */
            tmp = Int64Add(term1, term2);
            *negative = Int64IsNegative(tmp);
            if (*negative)
                tmp = Int64TimesInt32(tmp, -1);
            /* Result is positive and top bit not set */
            *result = *(UINT64*) &tmp;
            break;
    }
 
    return S_OK;
}
    
/* Compute the offset of the local clock from the server clock */
PRIVATE SCODE MCT SntpClientComputeLocalClockOffset(
    PISNTPCLIENT iThis, 
    const _TCHAR *ServerName,
    PTIME LocalClockOffset, 
    PBOOL SubtractOffset, 
    PTIME TimeStamp)
{
    METHOD_DECLARE(SntpClient, ComputeLocalClockOffset);
    UINT64 ot, rt, tt, dt, offset;
    
    sc = SntpClientSendRequest(iThis, ServerName);
    if (FAILED(sc))
        return sc;

    /* Wait for the reply and get a timestamp with it */
    sc = SntpClientReceiveReply(iThis, &ot, &rt, &tt, &dt, TimeStamp);

    sc = ComputeSntpFormatOffset(ot, rt, tt, dt, &offset, SubtractOffset);

    /* Translate the offset magnitude */
    *LocalClockOffset = TranslateSntpTimeToAbsoluteMMLite(*(UINT64*) &offset);

    /* Strip off the OFFSET, which is added during the translation */
    *LocalClockOffset = Int64Subtract(*LocalClockOffset, *(INT64*) &OFFSET); 
    
    /* Flip the sign bit, because MMLite format relative times are negative */
    *LocalClockOffset = Int64TimesInt32(*LocalClockOffset, -1);
    
    DBGME(0, printf("SNTP format offset=0x%llx\n", offset));
    DBGME(0, printf("MMLi format offset=0x%llx\n", *LocalClockOffset));

METHOD_CLEANUP;
    return sc;
}
    
PRIVATE SCODE MCT SntpClientAdjustLocalClock(
    PISNTPCLIENT iThis,
    TIME LocalClockOffset,
    BOOL SubtractOffset, 
    TIME TimeStamp)
{
    METHOD_DECLARE(SntpClient, AdjustLocalClock);

    assert(TimeIsRelative(LocalClockOffset));

    /* Store the offset internally */
    This->LocalClockOffset = LocalClockOffset;
    This->SubtractOffset = SubtractOffset;
    This->LastSynchronized = TimeStamp;

    /* Also store the offset in the SOAP engine */

    /* The SOAP engine maintains its own time offset; 
       This may or may not be a good idea, but if we 
       set it here then the SOAP code is not dependent 
       on this SNTP code for time look-up.
       
       Note that offset is a relative 
       time, which is a negative 64-bit integer. This 
       encodes a non-negative duration. The accompanying
       SubtractOffset boolean indicates whether the 
       positive duration (in the form of a negative number...)
       should be added or subtracted from the system time. 
       Right, convoluted... but in keeping with the 
       convention of absolute/relative time encoding
       troughout the rest of the system
       
       The SOAP engine currently does the wrong thing with 
       its time offsets; it stores durations, but uses the 
       sign bit to indicate whether the clock is ahead or 
       behind... feels more natural, breaks the convention. */
    if (SNTP_SOAP_INTEGRATE) {
        PICONVERT convert;
        sc = BindToObject(
            CurrentNameSpace(), _T("COB/soap.cob"), 
            NAME_SPACE_READ,
            &IID_ISoapFactory, (void **) &convert);
        if (! FAILED(sc)) {
            /* This code takes into account the SOAP time 
               offset bug described above. Otherwise it would
               be
               sc = convert->v->SetTimeOffset(convert, 
                        LocalClockOffset, SubtractOffset) */
            TIME SoapOffset = (SubtractOffset)
                ? LocalClockOffset
                : Int64TimesInt32(LocalClockOffset, -1);
            SoapOffset = Int64Add(SoapOffset, This->RangeAdjustment);
            sc = convert->v->SetTimeOffset(convert, SoapOffset, FALSE);
            /* There is nothing to do if this fails... */
            convert->v->Release(convert);
        }
    }

    sc = S_OK;
    return sc;
}
    
PRIVATE SCODE MCT SntpClientSynchronizeWithServer(
    PISNTPCLIENT iThis, 
    const _TCHAR *ServerName)
{
    METHOD_DECLARE_NOTHIS(SntpClient, SynchronizeWithServer);
    TIME LocalClockOffset;
    BOOL SubtractOffset;
    TIME TimeStamp;

    sc = SntpClientComputeLocalClockOffset(
        iThis, ServerName, &LocalClockOffset, &SubtractOffset, &TimeStamp);
    CHECKSC("compute local clock offset");

    DBGME(0, printf("sntp.c:SntpClientSynchronizeWithServer: \n\toffset=0x%llx negative=%d timestamp=0x%llx\n", 
         LocalClockOffset, SubtractOffset, TimeStamp));

    sc = SntpClientAdjustLocalClock(iThis, LocalClockOffset, SubtractOffset, TimeStamp);
    CHECKSC("adjust local clock");

    /* Produce nice print-out... */
    {
        TIME adjusted;
        sc = SntpClientCurrentTimeNoLock(iThis, &adjusted);
        CHECKSC("look up wall time");

        printf("Boot time: 0x%llx\n", CurrentTime());
        printf("Wall time: 0x%llx\n", adjusted);
    }

METHOD_CLEANUP;
    return sc;
}
    
PRIVATE void THREAD_LINKAGE SntpClientSynchronizationThread(THREAD_ARGUMENT arg)
{
    SCODE sc;
    PISNTPCLIENT iThis = (PISNTPCLIENT) arg;
    SntpClient *This = pSntpClient(iThis);
    BOOL Stop = FALSE;
    UINT ConsecutiveFailures = 0;

    printf("sntp.c:SntpClientSynchronizationThread: \n\tSNTP client is running\n");

    do {
        Mutex_Lock(&This->Mx);

        if (!This->ThreadIsRunning)
            This->ThreadIsRunning = TRUE;

        /* Attempt to synchronize, but stop the thread if we've failed to 
           often in a row */
        sc = This->v->SynchronizeWithServer(iThis, This->ServerName);
        if (FAILED(sc)) {
            DBGME(4, printf("sntp.c:SntpClientSynthronizationThread: Failed to synchronize, sc=%x", sc));
            if (SNTP_MAX_FAILURES < ++ConsecutiveFailures) {
                Stop = TRUE;
                DBGME(4, printf("sntp.c:SntpClientSynchronizationThread: %d subsequent failures\n", 
                    ConsecutiveFailures));
            }
        } else {
            ConsecutiveFailures = 0;
            Stop = This->StopThread;
        }

        Mutex_Unlock(&This->Mx);

        if (! Stop) {
            TIME later;
            Int32ToInt64(later, TIME_RELATIVE(TIME_SECONDS((INT) This->Period)));
            SleepUntil(later);
        }
        else break;

    } while (TRUE);

    printf("sntp.c:SntpClientSynchronizationThread: \n\tSNTP client stopped\n");

    iThis->v->Release(iThis);
}
    
/* Start the time synchronization worker thread with the given refresh period */
PRIVATE SCODE MCT SntpClientStart(
    PISNTPCLIENT iThis, 
    const _TCHAR *ServerName, 
    UINT Period)
{
    METHOD_DECLARE(SntpClient, Start);

    printf("sntp.c:SntpClientStart: \n\tStarting SNTP client\n");

    if (SNTP_CLIENT_MIN_PERIOD > Period ||
        SNTP_CLIENT_MAX_PERIOD < Period)
        CHECKSETSC(FALSE, E_INVALID_PARAMETER, "start; period outside range");

    if (NULL == ServerName)
        ServerName = DEFAULT_TIME_SERVER;

    Mutex_Lock(&This->Mx);

    CHECKSETSC(!This->ThreadIsRunning, E_OBJECT_NOT_INITIALIZED,
               "start; call stop first");

    DBGME(3, printf("sntp.c:SntpClientStart: \n"));
    DBGME(3, printf("\tSetting period to %d\n", Period));
    DBGME(3, printf("\tSynchronizing with %s\n", ServerName));

    if (NULL != This->ServerName)
        CurrentHeap()->v->Free(CurrentHeap(), 0, This->ServerName);
    This->ServerName = _tcsdup(ServerName);

    This->Period = Period;

    iThis->v->AddRef(iThis);

    /* Ensure the SNTP slave thread runs within the SNTP COB process */
    {
        PIPROCESS prc;
        PITHREAD pthd;
        PIMODULE mod;

        sc = GetModuleContainsAddr((ADDRESS) SntpClientSynchronizationThread, &mod);
        CHECKSC("get SNTP COB module");

        sc = mod->v->QueryInterface(mod, &IID_IProcess, (void **) &prc);
        mod->v->Release(mod);
        CHECKSC("get process from SNTP COB module");

        sc = prc->v->CreateThread(prc, CREATE_THREAD_STATIC_SCOPING,
            (THREAD_FUNCTION) SntpClientSynchronizationThread,
            (THREAD_ARGUMENT) iThis, 0, NULL, &pthd);
        prc->v->Release(prc);
        CHECKSC("create SNTP slave thread");
        pthd->v->Release(pthd);
    }

    DBGME(3, printf("sntp.c:SntpClientStart: \n\tSNTP client started\n"));

    This->StopThread = FALSE;

    sc = S_OK;

METHOD_CLEANUP;
    Mutex_Unlock(&This->Mx);
    return sc;
}
    
/* Stop the time synchronization thread */
PRIVATE SCODE MCT SntpClientStop(PISNTPCLIENT iThis)
{
    SntpClient *This = pSntpClient(iThis);

    DBGME(3, printf("sntp.c:SntpClientStop: \n\tStopping SNTP client...\n"));

    Mutex_Lock(&This->Mx);
    This->StopThread = TRUE;
    Mutex_Unlock(&This->Mx);

    return S_OK;
}
    
/* If the system returns a current time of earlier than the 
   beginning of the 20th century then negotiation with a time 
   server will fail because the SNTP range is more limited 
   than the Windows range - so we adjust the range by the 
   difference between the SNTP and Windows zeros */
PRIVATE void SetRangeAdjustment(SntpClient *This)
{
    TIME systemTime = CurrentTime();
    UINT64 sntpzero = OFFSET, system = *(UINT64 *) &systemTime;
    if (Uint64Less(system, sntpzero))
        This->RangeAdjustment = *(PTIME) &sntpzero;
    /* Else the range adjustment is left at zero */
}
    
/* Component */

const struct ISntpClientVtbl SntpClientVtbl;

PIUNKNOWN CreateComponent(void)
{
    SntpClient *This;
    PIHEAP pHeap = CurrentHeap();

    This = (SntpClient *) pHeap->v->
        Alloc(pHeap, 
            HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME(_T("client.c:CreateComponent")), 
            sizeof(SntpClient), 0);
    if (! This)
        return NULL;

    This->v = &SntpClientVtbl;
    This->Socket = INVALID_SOCKET; 
    This->ServerName = _tcsdup(DEFAULT_TIME_SERVER);
    This->StopThread = FALSE;
    This->ThreadIsRunning = FALSE;
    This->Period = SNTP_CLIENT_MIN_PERIOD;
    This->RangeAdjustment = TIME_ORIGIN;
    This->LocalClockOffset = TIME_ORIGIN;
    This->SubtractOffset = FALSE;
    This->LastSynchronized = TIME_ORIGIN;
    This->Refs = 1;

    Mutex_Init(&This->Mx);

    SetRangeAdjustment(This);

    return (PIUNKNOWN) This;
}
    
PIUNKNOWN sntpCobMain(void)
{
    return (PIUNKNOWN) CreateComponent();
}

#include "s_SntpClient_v.c"

