/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Utility program for the Intel 28F128 (J3A150) StrataFlash memory device
 * [as used in the Intrinsyc CErf dev board for the SA-1110 StrongARM chip]
 * Untested on the other 28Fxxx devices.
 */
#include <stdio.h>
#include <mmlite.h>
#include <stdlib.h>
#include <tchar.h>

/* Turn these on for debugging */
#if 0
BOOL Verbose = TRUE;
#define paranoia 1
#else
BOOL Verbose = FALSE;
#define paranoia 0
#endif

/* Default as visible on the CERF board (MMLite virtual mapping, uncached)
 */
#define SF_DEFAULT_ADDRESS ((ADDRESS)0x51000000)

/* Product ID codes
 */
#define MANUF_INTEL  0x89
#define DEVICE_320   0x16
#define DEVICE_640   0x17
#define DEVICE_128   0x18

/* Sector size maps
 */
#define MAX_SECTORS 1024 /* used by the Unlock command */
#define nDELTAS 3
typedef /*NOT! const*/ struct _SMAP {
    struct {
        UINT32 nSectors;
        UINT32 nKB;
    } Deltas[nDELTAS];
    UINT8 DeviceCode;
    UINT8 TotalSectors;
    _TCHAR *DeviceName;
} SMAP, *PSMAP;

#define nMAPS 3
/*NOT! const */ struct _SMAP SectorMaps[nMAPS] = {

    {
     {{32,128},{0,0},},
     DEVICE_320, 32,
     _TEXT("28F320")
    },
    {
     {{64,128},{0,0},},
     DEVICE_640, 64,
     _TEXT("28F640")
    },
    {
     {{128,128},{0,0},},
     DEVICE_128, 128,
     _TEXT("28F128")
    }
};

/* Returns the size in KBytes of a given sector,
 * or -1 for bad arguments.
 */
int KBinSector(PSMAP SecMap, UINT SecNo)
{
    int i, s = SecNo;

    for (i = 0; i < nDELTAS; i++) {
        if (SecNo < SecMap->Deltas[i].nSectors)
            return SecMap->Deltas[i].nKB;
        SecNo -= SecMap->Deltas[i].nSectors;
    }

#if paranoia
    _tprintf(_TEXT("KS: Bad SecNo x%x\r\n"), s);
#endif
    s = 0;
    return -1;
}

#define SectorSize(_map_,_sector_) (1024 * KBinSector(_map_,_sector_))

/* Whats the starting offset of sector N
 */
ADDRESS SectorStart(PSMAP SecMap, int SecNo)
{
    int i, s = SecNo;
    ADDRESS Offset = 0;

    for (i = 0; i < nDELTAS; i++) {
        if ((UINT)SecNo < SecMap->Deltas[i].nSectors)
            return 1024 * (Offset + (SecMap->Deltas[i].nKB * SecNo));
        SecNo -= SecMap->Deltas[i].nSectors;
        Offset += SecMap->Deltas[i].nSectors * SecMap->Deltas[i].nKB;
    }

#if paranoia
    _tprintf(_TEXT("SS: Bad SecNo x%x\r\n"), s);
#endif
    s = 0;
    return ~0;
}

/* Whats the ending offset (exclusive) of sector N
 */
ADDRESS SectorEnd(PSMAP SecMap, int SecNo)
{
    ADDRESS Offset = SectorStart(SecMap,SecNo);
    if (Offset != ~0)
        return Offset + (1024 * KBinSector(SecMap,SecNo));
#if paranoia
    _tprintf(_TEXT("SE: Bad SecNo x%x\r\n"), SecNo);
#endif
    return ~0;
}

/* What sector number corresponds to a given offset
 */
UINT SectorNumber(PSMAP SecMap, ADDRESS Offset)
{
    UINT i;
    UINT SecNo = 0;
    ADDRESS s = Offset;

    Offset /= 1024;
    for (i = 0; i < nDELTAS; i++) {
        if (Offset < (UINT)
            ((SecMap->Deltas[i].nSectors * SecMap->Deltas[i].nKB)))
            return SecNo + (Offset / SecMap->Deltas[i].nKB);
        SecNo += SecMap->Deltas[i].nSectors;
        Offset -= SecMap->Deltas[i].nSectors * SecMap->Deltas[i].nKB;
    }

#if paranoia
    _tprintf(_TEXT("SN: Bad Offset x%x\r\n"), s);
#endif
    s = 0;
    return ~0;
}

#if paranoia
/* Print the whole sector map, and the reverse mapping too
 * [Internal consistency checks]
 */
void SaySectors(PSMAP Map)
{
    UINT SecNo;
    ADDRESS Offset;

    _tprintf(_TEXT("SectorMap for DeviceCode x%x\r\n"), Map->DeviceCode);
    for (SecNo = 0; SecNo <= Map->TotalSectors; SecNo++) {
        _tprintf(_TEXT("SecNo %2d: %2d KB at x%06x..x%06x\r\n"), SecNo,
                 KBinSector(Map,SecNo), 
                 SectorStart(Map,SecNo), 
                 SectorEnd(Map,SecNo)
                 );
    }

    _tprintf(_TEXT("Reverse SectorMap for DeviceCode x%x\r\n"), Map->DeviceCode);
    Offset = 0;
    SecNo = ~0;
    do {
        UINT Sec;
        Sec = SectorNumber(Map,Offset);
        if (Sec != SecNo) {
            SecNo = Sec;
            _tprintf(_TEXT("Offset x%x => Sector %d\r\n"), Offset, SecNo);
        }
        Offset += 1024;
    } while (SecNo != ~0);
}
#endif



/*****************************************
 * Semi-generic operations
 */
typedef struct _FLASH FLASH;
typedef struct _FlashVtable {
    void (*WriteUint8)    (FLASH *This,UINT Offset,UINT8 Value);
    void (*ReadUint8)     (FLASH *This,UINT Offset,UINT8 *Value);
    void (*WriteUint16)   (FLASH *This,UINT Offset,UINT16 Value);
    void (*ReadUint16)    (FLASH *This,UINT Offset,UINT16 *Value);
    void (*WriteUint32)   (FLASH *This,UINT Offset,UINT32 Value);
    void (*ReadUint32)    (FLASH *This,UINT Offset,UINT32 *Value);
    BOOL (*WordProgram)   (FLASH *This, ADDRESS Offset, UINT16 *pValues,
                           BOOL EraseCheck, int *nWritten);
    BOOL (*BufferProgram) (FLASH *This, ADDRESS Offset, UINT16 *pValues,
                           BOOL EraseCheck, int *nWritten);
    
} FlashVtable;

struct _FLASH {
    FlashVtable *v;
    UINT nChips;
    ADDRESS BaseAddress;
    UINT BufferedBytes;
    PSMAP Map;
};


/*****************************************
 * Hardware access proper, single-chip
 */
void SingleWriteUint8  (FLASH *This,UINT Offset,UINT8 Value)
{
    volatile UINT8 * Where = (volatile UINT8 *) (This->BaseAddress + Offset);
    *Where = Value;
}

void SingleReadUint8   (FLASH *This,UINT Offset,UINT8 *Value)
{
    volatile UINT8 * Where = (volatile UINT8 *) (This->BaseAddress + Offset);
    *Value = *Where;
}

void SingleWriteUint16 (FLASH *This,UINT Offset,UINT16 Value)
{
    volatile UINT16 * Where = (volatile UINT16 *) (This->BaseAddress + Offset);
    *Where = Value;
}

void SingleReadUint16  (FLASH *This,UINT Offset,UINT16 *Value)
{
    volatile UINT16 * Where = (volatile UINT16 *) (This->BaseAddress + Offset);
    *Value = *Where;
}

/* This one should not be used, probably */
void SingleWriteUint32 (FLASH *This,UINT Offset,UINT32 Value)
{
#if 0
    /* The chip cannot take back-to-back writes */
    volatile UINT32 * Where = (volatile UINT32 *) (This->BaseAddress + Offset);
    *Where = Value;
#else
    UINT16 v0, v1;

    /* Unfortunately, this is bytesex dependent */
#if (BYTE_ORDER == BIG_ENDIAN)
    v1 = (UINT16) Value;
    v0 = (UINT16) (Value >> 16);
#else
    v0 = (UINT16) Value;
    v1 = (UINT16) (Value >> 16);
#endif
    SingleWriteUint16(This,Offset,v0);
    SingleWriteUint16(This,Offset+2,v1);
#endif
}

void SingleReadUint32  (FLASH *This,UINT Offset,UINT32 *Value)
{
    /* back-to-back reads must be ok */
    volatile UINT32 * Where = (volatile UINT32 *) (This->BaseAddress + Offset);
    *Value = *Where;
}

/******************************************************************
 * Hardware access proper, paired-chips
 * NB: This set of ops assumes two chips in parallel on a 32bit bus,
 *     each operaiton is repeated in parallel to both chips
 */

void TwinWriteUint8  (FLASH *This,UINT Offset,UINT8 Value)
{
    volatile UINT32 * Where = (volatile UINT32 *) (This->BaseAddress + Offset);
    UINT32 v = Value | ((UINT32)Value << 16);

    v = _bs32(v);
    if (Verbose) _tprintf(_T("tw8 @%x <= %x"),(ADDRESS) Where,v);
    *Where = v;
    if (Verbose) _tprintf(_T("\n"));
}

void TwinReadUint8   (FLASH *This,UINT Offset,UINT8 *Value)
{
    volatile UINT32 * Where = (volatile UINT32 *) (This->BaseAddress + Offset);
    UINT32 v;
    v = *Where;
    v = _bs32(v);
    if (Verbose) _tprintf(_T("tr8 @%x -> %x"),(ADDRESS) Where,v);
    *Value = (UINT8) v;
    if (Verbose) _tprintf(_T("\n"));
}

/* This one should *not* be used, error-prone */
void TwinWriteUint16 (FLASH *This,UINT Offset,UINT16 Value)
{
    volatile UINT16 * Where = (volatile UINT16 *) (This->BaseAddress + Offset);
    if (Verbose) _tprintf(_T("tw16 @%x <= %x"),(ADDRESS) Where,Value);
    *Where = Value;
    if (Verbose) _tprintf(_T("\n"));
}

void TwinReadUint16  (FLASH *This,UINT Offset,UINT16 *Value)
{
    volatile UINT16 * Where = (volatile UINT16 *) (This->BaseAddress + Offset);
    *Value = *Where;
    if (Verbose) _tprintf(_T("tr16 @%x -> %x"),(ADDRESS) Where,*Value);
    if (Verbose) _tprintf(_T("\n"));
}

void TwinWriteUint32 (FLASH *This,UINT Offset,UINT32 Value)
{
    volatile UINT32 * Where = (volatile UINT32 *) (This->BaseAddress + Offset);
    Value = _bs32(Value);
    if (Verbose) _tprintf(_T("tw32 @%x <= %x"),(ADDRESS) Where,Value);
    *Where = Value;
    if (Verbose) _tprintf(_T("\n"));
}

void TwinReadUint32  (FLASH *This,UINT Offset,UINT32 *Value)
{
    volatile UINT32 * Where = (volatile UINT32 *) (This->BaseAddress + Offset);
    UINT32 v;
    v = *Where;
    v = _bs32(v);
    *Value = v;
    if (Verbose) _tprintf(_T("tr32 @%x -> %x"),(ADDRESS) Where,*Value);
    if (Verbose) _tprintf(_T("\n"));
}

/*****************************************
 * Operations on the chip(s)
 */

/* Defines for the STATUS register
 */
#define ST_reserved          0x01
#define ST_BLOCK_LOCKED      0x02
#define ST_PROGRAM_SUSPENDED 0x04
#define ST_LOW_VOLTAGE       0x08
#define ST_LOCK_BIT_ERROR    0x10
#define ST_ERASE_ERROR       0x20
#define ST_ERASE_SUSPENDED   0x40
#define ST_READY             0x80
#define ST_ERASE_MASK        0xee  /* bits to check after erase command */
#define ST_MASK              0xfe  /* ignore reserved */

/* Command set (what we use of it)
 */
#define CMD_CONFIRM       0xd0
#define CMD_READ_ARRAY    0xff
#define CMD_READ_ID       0x90
#define CMD_READ_STATUS   0x70
#define CMD_CLEAR_STATUS  0x50
#define CMD_WRITE_WORD    0x40
#define CMD_WRITE_BUFFER  0xe8
#define CMD_ERASE_SETUP   0x20
#define CMD_ERASE_CONFIRM CMD_CONFIRM
#define CMD_SET_PREFIX    0x60  /* set read config, lock bits */
#define CMD_LOCK          0x01
#define CMD_UNLOCK        CMD_CONFIRM
/* What we dont use of it
 */
#define CMD_READ_QUERY    0x98
# define BUFFER_BYTES          32
#define CMD_ERASE_SUSPEND 0xb0
#define CMD_ERASE_RESUME  CMD_CONFIRM
#define CMD_CONFIGURATION 0xb8
#define CMD_PROTECT       0xc0

/* Enter the Product ID mode (Read Identifier Codes)
 */
void ProductIdEnter(FLASH *This)
{
    This->v->WriteUint8(This,0,CMD_READ_ID);
}

/* Exit the Product ID mode (enter Read Array mode)
 */
void ProductIdExit(FLASH *This)
{
    This->v->WriteUint8(This,0,CMD_READ_ARRAY);
}

/* Read the status register
 */
UINT8 ReadStatusRegister(FLASH *This)
{
    UINT8 Status;

    This->v->WriteUint8(This,0,CMD_READ_STATUS);
    This->v->ReadUint8(This,0,&Status);
    This->v->WriteUint8(This,0,CMD_READ_ARRAY);
    return Status;
}

/* Clear error bits in status
 */
void ClearStatusRegister(FLASH *This)
{
    This->v->WriteUint8(This,0,CMD_CLEAR_STATUS);
}

/* Decode status bits
 */
typedef _TCHAR *NAME;

void PrintStatus(UINT8 Status)
{
    NAME BitNames[8] = {
        _T("reserved"), _T("BLOCK_LOCKED"),
        _T("PROGRAM_SUSPENDED"), _T("LOW_VOLTAGE"),
        _T("LOCK_BIT_ERROR"), _T("ERASE_ERROR"),
        _T("ERASE_SUSPENDED"), _T("READY")
    };
    int i;
    BOOL OneSet = FALSE;

    _tprintf(_T("[status %x ="),Status);
    for (i = 0; i < 8; i++) {
        if (Status & (1<<i)) {
            _tprintf(_T("%c%s"), 
                     (OneSet) ? _T('|') : _T(' '),
                     BitNames[i]);
            OneSet = TRUE;
        }
    }
    _tprintf(_T("]\r\n"));
}

/* 
 * The device can lock up under certain conditions.
 * There is no software workaround [must toggle RP# to GND]
 * Check if it seems that we are in that state.
 */
BOOL IsIrresponsive(FLASH *This)
{
    UINT8 Status = ReadStatusRegister(This);

    if (Status & ST_READY)
        return FALSE;

    if ((Status & ST_ERASE_MASK) == 
        (ST_LOCK_BIT_ERROR|ST_ERASE_SUSPENDED|ST_ERASE_ERROR)) {
        /* yes, looks that way */
        return TRUE;
    }

    /* Something is indeed amiss, but we dont really know for sure */
    PrintStatus(ReadStatusRegister(This));
    ClearStatusRegister(This);
    PrintStatus(ReadStatusRegister(This));

    if ((Status & ST_MASK) == 
        (ST_LOCK_BIT_ERROR|ST_ERASE_SUSPENDED|ST_ERASE_ERROR)) {
        /* yes, looks that way */
        return TRUE;
    }

    return FALSE;
}


/* Write one 16bit word
 */
BOOL
SingleWordProgram(FLASH *This, ADDRESS Offset, UINT16 *Values,
                  BOOL EraseCheck, int *nWritten)
{
#ifndef m68k /* XXX Generates invalid and.w instruction */
    UINT8 Status;
    UINT16 i, Data16, Value;

    *nWritten = 0;

    Value = Values[0];
    Value = _bs16(Value);
    This->v->ReadUint16(This,Offset,&Data16);
    if (Verbose) {
        _tprintf(_TEXT("Location x%x was x%x\r\n"), 
                 This->BaseAddress+Offset, Data16);
    }

    if ((EraseCheck) && (Data16 != 0xffff))
        _tprintf(_TEXT("Offset %x not ERASED, wont take.\r\n"),Offset);

    This->v->WriteUint8(This,0,CMD_WRITE_WORD);
    This->v->WriteUint16(This,Offset,Value);

    /* Wait until the operation is completed
     * Specs say it takes between 210 and 630 us
     * Errata says 360 TYP and Max=TBD (sic)
     */
    Delay(400);

    for (i = 0; i < 10; i++) {
        This->v->ReadUint8(This,Offset,&Status);
        if ((Status & ST_READY)) break;
        Delay(100);
    }

    ProductIdExit(This);

    This->v->ReadUint16(This,Offset,&Data16);
    if (Verbose) {
        _tprintf(_TEXT("Location x%x is now x%x\r\n"), 
                 This->BaseAddress+Offset, Data16);
    }
             
    if ((Data16 != Value)) {
        PrintStatus(Status);
        _tprintf(_TEXT(". That didnt work, try again.. [%x != %x]\r\n"), 
                 Data16, Value);
        ClearStatusRegister(This);
        return FALSE;
    }
#endif /* !m68k */
    *nWritten = 2;
    return TRUE;
}

/* Write one buffer, 16bit words at a time
 */
BOOL
SingleBufferProgram(FLASH *This, ADDRESS Offset, UINT16 *Values,
                  BOOL EraseCheck, int *nWritten)
{
#ifndef m68k /* XXX Generates invalid and.w instruction */
    UINT8 Status;
    UINT16 i, Data16, Value = 0;

    *nWritten = 0;

    for (i = 0; i < This->BufferedBytes; i+= 2) {
        This->v->ReadUint16(This,Offset+i,&Data16);
        if (Verbose) {
            _tprintf(_TEXT("Location x%x was x%x\r\n"), 
                     This->BaseAddress+Offset+i, Data16);
        }

        if ((EraseCheck) && (Data16 != 0xffff))
            _tprintf(_TEXT("Offset %x not ERASED, wont take.\r\n"),Offset+i);
    }

    /* Specs say to retry if necessary */
    for (i = 0; i < 5; i++) {
        This->v->WriteUint8(This,Offset,CMD_WRITE_BUFFER);
        Delay(10);
        This->v->ReadUint8(This,Offset,&Status);
        if ((Status & ST_READY)) break;
    }
    if (0 == (Status & ST_READY)) {
        _tprintf(_TEXT("FAILED BufferProgram at Location x%x, Status= x%x\r\n"), 
                 This->BaseAddress+Offset, Status);
        return FALSE;
    }

    /* Say how many words we'll be sending */
    This->v->WriteUint8(This,Offset,(UINT8)(This->BufferedBytes/2));

    /* Send the data */
    for (i = 0; i < This->BufferedBytes; i+= 2) {
        Value = Values[i/2];
        Value = _bs16(Value);
        This->v->WriteUint16(This,Offset+i,Value);
        Delay(10);//jic
    }

    /* Write confirmation */
    This->v->WriteUint8(This,Offset,CMD_CONFIRM);

    /* Wait until the operation is completed
     * Specs say it takes between 800 and 2400 us
     * Errata says 1600 TYP and Max=TBD (sic), but fixed in stepping A3 and above.
     */
    Delay(800);

    for (i = 0; i < 20; i++) {
        This->v->WriteUint8(This,Offset,CMD_READ_STATUS);
        This->v->ReadUint8(This,Offset,&Status);
        if ((Status & ST_READY)) break;
        Delay(200);
    }

    ProductIdExit(This);

    /* Verify */
    for (i = 0; i < This->BufferedBytes; i+= 2) {
        This->v->ReadUint16(This,Offset+i,&Data16);
        if (Verbose) {
            _tprintf(_TEXT("Location x%x is now x%x\r\n"), 
                     This->BaseAddress+Offset+i, Data16);
        }
             
        Value = Values[i/2];
        Value = _bs16(Value);

        if ((Data16 != Value)) {
            PrintStatus(Status);
            _tprintf(_TEXT(". That didnt work, try again.. [%x != %x]\r\n"), 
                     Data16, Value);
            ClearStatusRegister(This);
            return FALSE;
        }
    }
#endif /* !m68k */
    *nWritten = This->BufferedBytes;
    return TRUE;
}

/* Write one 32bit word
 */
BOOL
TwinWordProgram(FLASH *This, ADDRESS Offset, UINT16 *Values,
                BOOL EraseCheck, int *nWritten)
{
#ifndef m68k /* XXX Generates invalid and.w instruction */
    UINT8 Status;
    UINT32 i, Data32, Value;
    UINT16 v0, v1;

    *nWritten = 0;

    v0 = Values[0];
    v0 = _bs16(v0);
    v1 = Values[1];
    v1 = _bs16(v1);
    Value = v0 | ((UINT32)v1 << 16);
    This->v->ReadUint32(This,Offset,&Data32);
    if (Verbose) {
        _tprintf(_TEXT("Location x%x was x%x\r\n"), 
                 This->BaseAddress+Offset, Data32);
    }

    if ((EraseCheck) && (Data32 != 0xffffffff))
        _tprintf(_TEXT("Offset %x not ERASED, wont take.\r\n"),Offset);

    This->v->WriteUint8(This,0,CMD_WRITE_WORD);
    This->v->WriteUint32(This,Offset,Value);

    /* Wait until the operation is completed
     * Specs say it takes between 210 and 630 us
     * Errata says 360 TYP and Max=TBD (sic)
     */
    Delay(400);

    for (i = 0; i < 10; i++) {
        This->v->ReadUint8(This,Offset,&Status);
        if ((Status & ST_READY)) break;
        Delay(100);
    }

    ProductIdExit(This);

    This->v->ReadUint32(This,Offset,&Data32);
    if (Verbose) {
        _tprintf(_TEXT("Location x%x is now x%x\r\n"), 
                 This->BaseAddress+Offset, Data32);
    }
             
    if ((Data32 != Value)) {
        PrintStatus(Status);
        _tprintf(_TEXT(". That didnt work, try again.. [%x != %x]\r\n"), 
                 Data32, Value);
        ClearStatusRegister(This);
        return FALSE;
    }
#endif /* !m68k */
    *nWritten = 4;
    return TRUE;
}

/* Write one buffer, 32bit words at a time
 */
BOOL
TwinBufferProgram(FLASH *This, ADDRESS Offset, UINT16 *Values,
                BOOL EraseCheck, int *nWritten)
{
#ifndef m68k /* XXX Generates invalid and.w instruction */
    UINT8 Status;
    UINT32 i, Data32, Value;
    UINT16 v0 = 0, v1;

    *nWritten = 0;

    for (i = 0; i < This->BufferedBytes; i+= 4) {
        This->v->ReadUint32(This,Offset+i,&Data32);
        if (Verbose) {
            _tprintf(_TEXT("Location x%x was x%x\r\n"), 
                     This->BaseAddress+Offset+i, Data32);
        }

        if ((EraseCheck) && (Data32 != 0xffffffff))
            _tprintf(_TEXT("Offset %x not ERASED, wont take.\r\n"),Offset+i);
    }

    /* Specs say to retry if necessary */
    for (i = 0; i < 5; i++) {
        This->v->WriteUint8(This,Offset,CMD_WRITE_BUFFER);
        Delay(10);
        This->v->ReadUint8(This,Offset,&Status);
        if ((Status & ST_READY)) break;
    }
    if (0 == (Status & ST_READY)) {
        _tprintf(_TEXT("FAILED BufferProgram at Location x%x, Status= x%x\r\n"), 
                 This->BaseAddress+Offset, Status);
        return FALSE;
    }

    /* Say how many words we'll be sending */
    This->v->WriteUint8(This,Offset,(UINT8)(This->BufferedBytes/4)); /* to each twin! */

    /* Send the data */
    for (i = 0; i < This->BufferedBytes; i+= 4) {
        v0 = Values[i/2];
        v0 = _bs16(v0);
        v1 = Values[1+(i/2)];
        v1 = _bs16(v1);
        Value = v0 | ((UINT32)v1 << 16);
        This->v->WriteUint32(This,Offset+i,Value);
        Delay(10);//jic
    }

    /* Write confirmation */
    This->v->WriteUint8(This,Offset,CMD_CONFIRM);

    /* Wait until the operation is completed
     * Specs say it takes between 800 and 2400 us
     * Errata says 1600 TYP and Max=TBD (sic), but fixed in stepping A3 and above.
     */
    Delay(800);

    for (i = 0; i < 20; i++) {
        This->v->WriteUint8(This,Offset,CMD_READ_STATUS);
        This->v->ReadUint8(This,Offset,&Status);
        if ((Status & ST_READY)) break;
        Delay(200);
    }

    ProductIdExit(This);

    /* Verify */
    for (i = 0; i < This->BufferedBytes; i+= 4) {
        This->v->ReadUint32(This,Offset+i,&Data32);
        if (Verbose) {
            _tprintf(_TEXT("Location x%x is now x%x\r\n"), 
                     This->BaseAddress+Offset+i, Data32);
        }
             
        v0 = Values[i/2];
        v0 = _bs16(v0);
        v1 = Values[1+(i/2)];
        v1 = _bs16(v1);
        Value = v0 | ((UINT32)v1 << 16);

        if ((Data32 != Value)) {
            PrintStatus(Status);
            _tprintf(_TEXT(". That didnt work, try again.. [%x != %x]\r\n"), 
                     Data32, Value);
            ClearStatusRegister(This);
            return FALSE;
        }
    }

#endif /* !m68k */
    *nWritten = This->BufferedBytes;
    return TRUE;
}

/* Erase one sector
 */
BOOL SectorErase(FLASH *This, int SecNo)
{
    ADDRESS Offset;
    UINT8 Status = 0;
    UINT16 i;
    TIME t;

    Offset = SectorStart(This->Map,SecNo);

    _tprintf(_TEXT("Erasing sector %d [offset x%x] ..."), SecNo, Offset);

    This->v->WriteUint8(This,Offset,CMD_ERASE_SETUP);
    This->v->WriteUint8(This,Offset,CMD_ERASE_CONFIRM);

    /* Wait until the erase is actually completed
     * Specs say it will take between 1 and 5 seconds.
     * Errata says it takes 2 sec min and 25 sec max.
     * Double that before giving up.
     */
    Int32ToInt64(t,TIME_RELATIVE(TIME_SECONDS(2)));
    for (i = 0; i < 20; i++) {
        /* Sleep for at least 2 seconds
         */
        SleepUntil(t);

        This->v->ReadUint8(This,Offset,&Status);
        if ((Status & ST_READY)) break;
        PrintStatus(Status);
    }

    ProductIdExit(This);

    if ((Status & ST_ERASE_MASK) == ST_READY) {
        _tprintf(_TEXT("done.\r\n"));
        return TRUE;
    }

    PrintStatus(Status);
    _tprintf(_TEXT(" NOT completed.\r\n"));
    ClearStatusRegister(This);
    return FALSE;
}

/* Apply the write-lock to a sector
 */
void SectorLock(FLASH *This, int SecNo)
{
    ADDRESS Offset;
    UINT8 Status;
    UINT16 i;

    Offset = SectorStart(This->Map,SecNo);

    _tprintf(_TEXT("Locking sector %d [offset x%x] ..."), SecNo, Offset);

    This->v->WriteUint8(This,0,CMD_SET_PREFIX);
    This->v->WriteUint8(This,Offset,CMD_LOCK);

    /* Wait until the lock is complete.
     * Specs say this takes between 64 and 75 usecs.
     */
    Delay(100);

    for (i = 0; i < 10; i++) {
        This->v->ReadUint8(This,Offset,&Status);
        if ((Status & ST_READY)) break;
        Delay(100);
    }

    ProductIdExit(This);

    if ((Status & ST_MASK) == ST_READY) {
        _tprintf(_TEXT("done.\r\n"));
        return;
    }

    PrintStatus(Status);
    _tprintf(_TEXT(" NOT completed.\r\n"));
    ClearStatusRegister(This);
}


/* Is there a lock on a given sector
 */
BOOL IsSectorLocked(FLASH *This, int SecNo)
{
    ADDRESS Offset;
    UINT8 Data, Data1;

    Offset = SectorStart(This->Map,SecNo);
    if (Offset == ~0)
        return FALSE; /* non existant is not locked ;-) */

    ProductIdEnter(This);
    /* Lockout info is at address 2 of the given sector, meaning A0=0 A1=1.
     */
    This->v->ReadUint8(This,Offset+(0x0002*2*This->nChips),&Data);
    This->v->ReadUint8(This,Offset+(0x0003*2*This->nChips),&Data1);

    ProductIdExit(This);

#if paranoia
    if ((Data & 1) == 0)
        _tprintf(_TEXT("%d@%x: %x %x\r\n"), 
                 SecNo, Offset, Data1, Data);
#endif

    return (Data & 1);
}

/* Remove the write-lock from a sector
 * Unfortunately, we can only remove all locks at once. Sigh.
 * So to 
 */
UINT32 LockMap[MAX_SECTORS/32];
void SectorUnlock(FLASH *This, int SecNo)
{
    UINT NumLocked = 0;
    UINT8 Status = 0;
    UINT i;
    TIME t;

    /* Sanity
     */
    if (!IsSectorLocked(This,SecNo)) {
        _tprintf(_TEXT("Sector %d is not locked.\r\n"),SecNo);
        return;
    }

    /* Build the locked sectors map
     */
    for (i = 0; i < MAX_SECTORS; i++) {
        if (IsSectorLocked(This,i)) {
            NumLocked++;
            LockMap[ i/32 ] = LockMap[ i/32 ] | (1<<(i%32));
        } else {
            LockMap[ i/32 ] = LockMap[ i/32 ] & ~ (1<<(i%32));
        }
    }
    /* Check we actually got something to do
     */
    if (NumLocked == 0) {
        _tprintf(_TEXT("Sectors are all unlocked already.\r\n"));
        return;
    }

    /* Clear the one 
     */
    LockMap[ SecNo/32 ] = LockMap[ SecNo/32 ] & ~ (1<<(SecNo%32));
    NumLocked--;
    
    _tprintf(_TEXT("Unlocking all sectors ..."));

    This->v->WriteUint8(This,0,CMD_SET_PREFIX);
    This->v->WriteUint8(This,0,CMD_LOCK);

    /* Wait until the lock is complete.
     * Specs say this takes between 0.5 and 0.7 secs.
     */
    Int32ToInt64(t,TIME_RELATIVE(TIME_SECONDS(1)));
    for (i = 0; i < 5; i++) {
        /* Sleep for at least 1 second
         */
        SleepUntil(t);

        This->v->ReadUint8(This,0,&Status);
        PrintStatus(Status);
        if ((Status & ST_READY)) break;
    }

    ProductIdExit(This);

    if ((Status & ST_MASK) == ST_READY) {
        _tprintf(_TEXT("done.\r\n"));
    } else {
        PrintStatus(Status);
        _tprintf(_TEXT(" NOT completed.\r\n"));
        ClearStatusRegister(This);
    }

    /* We might be done already
     */
    if (NumLocked == 0)
        return;

    _tprintf(_TEXT("Relocking all sectors but %d ..."), SecNo);
    for (i = 0; i < MAX_SECTORS; i++) {
        if (SectorStart(This->Map,i) == ~0)
            return; /* done */
        if (LockMap[ i/32 ] & (1<<(i%32)))
            SectorLock(This,i);
    }
}


/* Print the list of locked sectors (if any)
 */
void PrintLockedSectors(FLASH *This)
{
    UINT SecNo;
    ADDRESS Offset;
    BOOL SomeLocked = FALSE;

    for (SecNo = 0; ; SecNo++) {
        Offset = SectorStart(This->Map,SecNo);
        if (Offset == ~0)
            break;

        if (IsSectorLocked(This,SecNo)) {
            _tprintf(_TEXT("SecNo x%d at Offset x%x is locked.\r\n"),
                     SecNo, Offset);
            SomeLocked = TRUE;
        }
    }
    if (!SomeLocked)
        _tprintf(_TEXT("None of the sectors are locked.\r\n"));
}

/* Put a file at OFFSET from the start of the FLASH
 */
UINT16 Buffer[512]; /* do it 1KB at a time */
void WriteFile(FLASH *This, _TCHAR *FileName, ADDRESS Offset, BOOL Erase)
{
    FILE *f;
    size_t n, i;
    UINT CurSector;
    UINT32 CheckSum;
    UINT8 *pbBuffer;

    /* Make sure the device is not screwed up
     */
    if (Erase && IsIrresponsive(This)) {
        _tprintf(_TEXT("FLASH is locked-up (or mapped cacheable?), wont work. "));
    }

    /* Print the warning, we got burned by this.
     */
    _tprintf(_TEXT("Are you sure you want to do this ? "));
    i = _gettchar();
    _tprintf(_TEXT("\r\n"));
    if ((i != _T('y')) && (i != _T('Y'))) {
        _tprintf(_TEXT("No harm done.\r\n"));
        return;
    }

    f = _tfopen(FileName, _T("rb"));
    if (f == NULL) {
        _tprintf(_TEXT("File %s not found.\r\n"),FileName);
        return;
    }

    _tprintf(_TEXT("FLASHing file %s at offset %x.\r\n"),
             FileName,Offset);
    CurSector = ~0;
    CheckSum = 0;
    pbBuffer = (UINT8*) Buffer;
    for (;;) {
        n = fread(Buffer, 1, sizeof Buffer, f);
        if (n == 0) break;

        for (i = 0; i < n; i++)
            CheckSum += pbBuffer[i];

        i = SectorNumber(This->Map,Offset);
        if (i != CurSector) {
            _tprintf(_TEXT("\r\n"));
            CurSector = i;
            if (Erase) 
                SectorErase(This,CurSector);
        }

        _tprintf(_TEXT("#"));
        for (i = 0; i < n/2;) {
            int nTries;
            int nWritten = 0;/*we expect 2 or 4 */

#if 0
            if (This->BufferedBytes && ((n-(2*i)) >= This->BufferedBytes)) {
                for (nTries = 0; nTries < 5; nTries++)
                    if (This->v->BufferProgram(This,Offset,Buffer+i,Erase,&nWritten))
                        break;
            } else
#endif
            {
                for (nTries = 0; nTries < 5; nTries++)
                    if (This->v->WordProgram(This,Offset,Buffer+i,Erase,&nWritten))
                        break;
            }
            Offset += nWritten;
            i += (nWritten/2);
        }
    }
    _tprintf(_TEXT(" done.\r\nByteSum=x%x (%d).\r\n"), CheckSum, CheckSum);

    fclose(f);
}

/* Identify type and the sector map of the FLASH.
 * Argument is the base address of the device and the count of chips on the bus (1/2)
 * Returns FALSE if failed
 */
FlashVtable SingleOps = {
    SingleWriteUint8,
    SingleReadUint8,
    SingleWriteUint16,
    SingleReadUint16,
    SingleWriteUint32,
    SingleReadUint32,
    SingleWordProgram,
    SingleBufferProgram
};

FlashVtable TwinOps = {
    TwinWriteUint8,
    TwinReadUint8,
    TwinWriteUint16,
    TwinReadUint16,
    TwinWriteUint32,
    TwinReadUint32,
    TwinWordProgram,
    TwinBufferProgram
};

BOOL FlashIdentify(FLASH *This)
{
    UINT8 Mid, Did;
    int i;

    if (This->nChips == 1)
        This->v = &SingleOps;
    else
        This->v = &TwinOps;

    ProductIdEnter(This);
    This->v->ReadUint8(This,0x0000*2*This->nChips,&Mid);
    This->v->ReadUint8(This,0x0001*2*This->nChips,&Did);
    ProductIdExit(This);

    _tprintf(_TEXT("ManufacturerId: x%x  DeviceId: x%x\r\n"), Mid, Did);

    for (i = 0; i < nMAPS; i++)
        if (SectorMaps[i].DeviceCode == Did) {
            This->Map = &SectorMaps[i];
            /* double the sector sizes if twin-chips */
            This->Map->Deltas[0].nKB *= This->nChips;
            return TRUE;
        }

    return FALSE;
}

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    ADDRESS FlashAddress;
    _TCHAR *FileName;
    FLASH Flash;
    ADDRESS Offset;
    BOOL Erase = TRUE;
    BOOL Twins = FALSE;
    BOOL BufferedWrites = FALSE;

#if 0/* code selftest */
    SaySectors(&SectorMaps[0]);
    SaySectors(&SectorMaps[1]);
#endif

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_TEXT("Usage: FlashBaseAddress [filename [offset [NOERASE | TWINCHIPS | BUFFERED]]]\r\n"));
        goto Done;
    }

    if (argc > 1)
        FlashAddress = _tcstoul(argv[1],NULL,16);
    else
        FlashAddress = SF_DEFAULT_ADDRESS;

    if (argc > 2)
        FileName = argv[2];
    else
        FileName = NULL;

    if (argc > 3)
        Offset = _tcstoul(argv[3],NULL,16);
    else
        Offset = 0x20000; /* leave sector 0 alone */

    while (argc > 4) {
        if (0 == _tcsicmp(argv[4], _TEXT("NOERASE")))
            Erase = FALSE;
        if (0 == _tcsicmp(argv[4], _TEXT("TWINCHIPS")))
            Twins = TRUE;
        if (0 == _tcsicmp(argv[4], _TEXT("BUFFERED")))
            BufferedWrites = TRUE;
        argc--, argv++;
    }

    _tprintf(_TEXT("Checking FLASH @%x\r\n"), FlashAddress);
    memset(&Flash,0,sizeof Flash);
    Flash.nChips = (Twins) ? 2 : 1;
    Flash.BaseAddress = FlashAddress;
    Flash.BufferedBytes = (BufferedWrites) ? ((Twins) ? BUFFER_BYTES*2 : BUFFER_BYTES) : 0;

    if (FlashIdentify(&Flash)) {
        _tprintf(_TEXT("Flash was recognized, type %s\r\n"), Flash.Map->DeviceName);

#if paranoia
        SaySectors(Flash.Map);
#endif

        PrintLockedSectors(&Flash);

        if (FileName != NULL) {
            WriteFile(&Flash,FileName,Offset,Erase);
        }
    } else
        _tprintf(_TEXT("Flash chip not recognized, test FAILED\r\n"));

 Done:
    (void) FreeCmdLine(&argc,&argv);
    return 0;
}

