/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * ttcp
 *
 * Test TCP/UDP performance.
 * Makes a connection on port 5001 and transfers buffers.
 *
 * This version for MMLITE (#define MMLITE) or NT systems.
 * Based on the original Unix version written by numerous people including
 * (among unknown others) Terry Slattery, Mike Muuss, and Robert S. Miles.
 */
#if defined(MMLITE)
#include <mmlite.h>

unsigned long GetMillisecondClock(void)
{
    TIME Now = CurrentTime();
    Now = Int64DividedByInt32(Now,TIME_MILLIS(1));
    return Int64ToInt32(Now);
}
/* XXX Delay() spins, relevant ? umbug */

#define SET_SIN_LENGTH(_s_,_l_) { (_s_).sin_len = _l_; }

#else

#define _TCHAR TCHAR
#define TSTR char *
#define _TEXT(t) TEXT(t)

#define Delay(_n_) Sleep(_n_)
#define GetMillisecondClock GetTickCount

#define UINT16 unsigned short
#define UINT8  unsigned char

#define SET_SIN_LENGTH(_s_,_l_)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <winsock.h>
#include <ctype.h>

#define TEST 0

/*
 * Debugging support.
 */
#if 1
#define DBGOUT(lvl,s) if (DbgLvl <= lvl) s;
#else
#define DBGOUT(lvl,s) ;
#endif

/*
 * Whining support.
 */
void PrintError(char *s)
{
    _tprintf(_TEXT("%hs, errno x%x\n"), s, WSAGetLastError());
}

/*
 * Defaults; override via command-line arguments.
 */
INT BufLen = 8192;
INT NumBufs = 2048;
UINT16 Port = 5001;
INT Udp = 0;             /* 0 = tcp, !0 = udp */
INT Trans;               /* 0 = receive, !0 = transmit mode */
INT SinkMode = 1;        /* 0 = normal I/O, !0 = sink/source mode */
INT PreserveBlocks = 0;  /* 1 = preserve buffer blocking across connection */
INT VerifyData = 0;      /* 1 = check data received is same as what was sent */
INT DbgLvl = 2;          /* 0 = all debugging, 2 = normal */

/*
 * These routines are for writing a specific data pattern out to the
 * transmitted buffer (send side) and checking that the pattern is what
 * it is supposed to be (receive side).  A simple data verification test.
 */
void Pattern(char *pC, INT Count)
{
    unsigned char c;

    c = 0;
    while(Count-- > 0) {
        while(!isprint((c & 0x7F)))
            c++;
        *pC++ = (char)(c++ & 0x7F);
    }
}

INT CheckPattern(char *pC, INT Count)
{
    unsigned char c;

    c = 0;
    while(Count-- > 0)  {
        while(!isprint((c & 0x7F)))
            c++;
        if (*pC++ != (char)(c++ & 0x7F))
            return Count + 1;
    }
    return 0;
}

/*
 * This function performs the function of a recv but will call recv multiple
 * times until it gets the requested number of characters.  This can be
 * necessary because stream connections don't always deliver data with the
 * same grouping as it is written with.
 */
INT FullRead(SOCKET Sock, char *pBuf, UINT Wanted)
{
    UINT Total = 0;
    INT Got;

    DBGOUT(0,_tprintf(_TEXT("In FullRead()\n")));

    do {
        Got = recv(Sock, pBuf, Wanted - Total, 0);
        if (Got == SOCKET_ERROR) {
            PrintError("recv failed");
            return(-1);
        }
        if (Got == 0)
            break;
        Total += (unsigned)Got;
        pBuf += Got;
    } while(Total < Wanted);

    return((INT)Total);
}

/*
 * Receive (up to) desired amount of data from peer.
 */
INT DoRead(SOCKET Sock, char *pBuf, UINT Wanted)
{
    INT Got;

    DBGOUT(0,_tprintf(_TEXT("In DoRead()\n")));

    if (Udp) {
        Got = recvfrom(Sock, pBuf, Wanted, 0, 0, 0);
    } else {
        if (PreserveBlocks)
            Got = FullRead(Sock, pBuf, Wanted);
        else
            Got = recv(Sock, pBuf, Wanted, 0);
    }

    return(Got);
}

/*
 * Send desired amount of data to peer.
 */
INT DoWrite(SOCKET Sock, char *pBuf, UINT Send)
{
    INT Sent;

    DBGOUT(0,_tprintf(_TEXT("In DoWrite()\n")));

    if (Udp) {
again:
        Sent = send(Sock, pBuf, Send, 0);
        if (Sent == SOCKET_ERROR && WSAGetLastError() == WSAENOBUFS) {
            _tprintf(_TEXT("Out of network buffer space, pausing briefly...\n"));
            Delay(100);
            goto again;
        }
    } else {
        Sent = send(Sock, pBuf, Send, 0);
    }

    return(Sent);
}

/*
 * Be friendly and helpful, then kill the process.
 */
void Usage(void)
{
    _tprintf(_TEXT("Usage: ttcp -t [-options] host \n"));
    _tprintf(_TEXT("  -l##  length of bufs to send (default %d)\n"), BufLen);
    _tprintf(_TEXT("  -n##  number of bufs to send (default %d)\n"), NumBufs);
    _tprintf(_TEXT("  -p##  port number to send to (default %d)\n"), Port);
    _tprintf(_TEXT("  -u    use UDP instead of TCP\n"));
    _tprintf(_TEXT("Usage: ttcp -r [-options] \n"));
    _tprintf(_TEXT("  -l##  length of network read buf (default %d)\n"), BufLen);
    _tprintf(_TEXT("  -p##  port number to listen at (default %d)\n"), Port);
    _tprintf(_TEXT("  -V    verify data received is same as that sent\n"));
    _tprintf(_TEXT("  -B    output only full blocks, as specified in -l##\n"));
    _tprintf(_TEXT("  -u    use UDP instead of TCP\n"));

    exit(1);
}

int _tmain(INT ArgC, _TCHAR *ArgV[])
{
    WSADATA WsaData;
    SOCKET Zowie;
    char *pBuffer = NULL;
    _TCHAR *Host;
    char *host;
    struct sockaddr_in TheirSockAddr, OurSockAddr;
    UINT32 Time0, TimeElapsed;
    INT Got, SockBufSize, TotalTransferred = 0;

#if defined(MMLITE)
    ParseCmdLine(&ArgC,&ArgV);
#endif

    if (ArgC < 2)
        Usage();  /* never returns */

    /*
     * Parse options.
     */
    ArgV++; ArgC--;
    while(ArgC > 0 && ArgV[0][0] == '-') {
        switch (ArgV[0][1]) {

        case 't':
            Trans = 1;
            break;
        case 'r':
            Trans = 0;
            break;
        case 'n':
            NumBufs = _ttoi(&ArgV[0][2]);
            break;
        case 'l':
            BufLen = _ttoi(&ArgV[0][2]);
            break;
        case 'p':
            Port = (short) _ttoi(&ArgV[0][2]);
            break;
        case 'u':
            Udp = 1;
            break;
        case 'V':
            VerifyData = 1;
            break;
        case 'B':
            PreserveBlocks = 1;
            break;
        case 'x':
            DbgLvl = (short) _ttoi(&ArgV[0][2]);
            break;
        default:
            Usage();  /* never returns */
        }
        ArgV++; ArgC--;
    }

    /*
     * Fire up winsock.
     */
    if (WSAStartup(0x0101, &WsaData) == SOCKET_ERROR) {
        PrintError("WSAStartup failed");
        exit(1);
    }
    if (Udp && BufLen > WsaData.iMaxUdpDg) {
        _tprintf(_TEXT("Buffer len %d is larger than max datagram size (%d)\n"),
               BufLen, WsaData.iMaxUdpDg);
        exit(1);
    }

    if (Trans) {
        /*
         * As the transmit side, we should have one argument left:
         * the remote host name (or address).
         */
        if (ArgC != 1)
            Usage();  /* never returns */

        /*
         * Convert remote hostname to an internet socket address.
         */
        memset(&TheirSockAddr, 0, sizeof(TheirSockAddr));
        Host = ArgV[0];
#if defined(_UNICODE)
        /*
         * Winsock's inet_addr and gethostbyname only take ASCII.
         * Convert unicode hostname.
         */
        {
        int Index;

        host = malloc(_tcslen(Host) + 1);
        for (Index = 0; Host[Index]; Index++)
            host[Index] = (char) Host[Index];
        host[Index] = 0;
        }
#else
        host = Host;
#endif
        if (_ttoi(Host) > 0 ) {
            /* was given host in numeric format */
            TheirSockAddr.sin_family = AF_INET;
            SET_SIN_LENGTH(TheirSockAddr, 16);
            TheirSockAddr.sin_addr.s_addr = inet_addr(host);
        } else {
            /* was given host in name format */
            struct hostent *pHostInfo;

            if ((pHostInfo = gethostbyname(host)) == NULL) {
                _tprintf(_TEXT("bad hostname\n"));
                exit(1);
            }
            TheirSockAddr.sin_family = (UINT8) pHostInfo->h_addrtype;
            SET_SIN_LENGTH(TheirSockAddr, (UINT8)pHostInfo->h_length);
            memcpy(&TheirSockAddr.sin_addr, pHostInfo->h_addr,
                   pHostInfo->h_length);
        }
        TheirSockAddr.sin_port = htons(Port);
        OurSockAddr.sin_port = 0;  /* let system pick one */
    } else {
        /*
         * As the receive side, we should have no arguments left.
         */
        if (ArgC != 0)
            Usage();  /* never returns */

        /*
         * Just need to set our port number.
         */
        OurSockAddr.sin_port = htons(Port);
    }

    /*
     * Grab some memory to hold our data buffer.
     */
    if ((pBuffer = (char *)malloc(BufLen)) == (char *)NULL) {
        PrintError("malloc");
        exit(1);
    }

    _tprintf(_TEXT("ttcp%hs: NumBufs=%d, BufLen=%d, Port=%d %hs\n"),
           Trans ? "-t" : "-r", NumBufs, BufLen, Port, Udp ? "udp" : "tcp");

    /*
     * Create our socket.
     */
    if ((Zowie = socket(AF_INET, Udp ? SOCK_DGRAM : SOCK_STREAM, 0))
            == INVALID_SOCKET) {
        PrintError("socket");
        exit(1);
    }

    /*
     * Bind to our address (really only necessary for receive side).
     */
    OurSockAddr.sin_family = AF_INET;
    SET_SIN_LENGTH(OurSockAddr, 16);
    OurSockAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(Zowie, (struct sockaddr *)&OurSockAddr, sizeof(OurSockAddr))
            == SOCKET_ERROR) {
        PrintError("bind");
        exit(1);
    }

    /*
     * Set socket buffer size appropriately (four times the user buffer size).
     */
    SockBufSize = BufLen * 4;
    if (setsockopt(Zowie, SOL_SOCKET, Trans? SO_SNDBUF : SO_RCVBUF,
            (char *)&SockBufSize, sizeof(SockBufSize)) == SOCKET_ERROR) {
        PrintError("setsockopt");
        exit(1);
    }

#if TEST
    {
        INT o, i;

        i = sizeof(INT);
        if (getsockopt(Zowie, SOL_SOCKET, Trans? SO_SNDBUF : SO_RCVBUF,
                (char*)&o, &i) < 0)
            PrintError("getsockopt");
        _tprintf(_TEXT("socket buffer size is %d\n"), o);
    }
#endif

    /*
     * Establish a connection to our peer.
     * Note that in TCP case, the timer esentially starts running once
     * the connection is established.  So we don't do any informational
     * printing until the connection closes.
     */
    if (Trans) {
            /*
             * Transmitter is the client.
             * Need to connect to the receiver (server), for UDP this
             * is just a psuedo-connection.
             */
            if (connect(Zowie, (struct sockaddr *)&TheirSockAddr,
                        sizeof(TheirSockAddr)) == SOCKET_ERROR) {
                PrintError("connect");
                exit(1);
            }
    } else {
        /*
         * Receiver is the server.
         * UDP is connectionless, do nothing special.
         */
        if (!Udp) {
            /*
             * Wait for a (single) connection.
             */
            SOCKET New;
            int TheirLen;

            listen(Zowie, 1);
            TheirLen = sizeof(TheirSockAddr);
            if ((New = accept(Zowie, (struct sockaddr *)&TheirSockAddr,
                              &TheirLen)) == SOCKET_ERROR) {
                PrintError("connect");
                exit(1);
            }
            closesocket(Zowie);
            Zowie = New;
#if TEST
        {
            INT o, i;

            i = sizeof(INT);
            if (getsockopt(Zowie, SOL_SOCKET, Trans? SO_SNDBUF : SO_RCVBUF,
                    (char*)&o, &i) < 0)
                PrintError("getsockopt");
            _tprintf(_TEXT("socket buffer size is %d\n"), o);
        }
#endif
        }

        DBGOUT(1,_tprintf(_TEXT("ttcp%hs: connected to %hs\n"),
               Trans ? "-t" : "-r", inet_ntoa(TheirSockAddr.sin_addr)));
    }

    /*
     * Everything's ready, start the time trials.
     */
    if (SinkMode) {
        /*
         * Sink/source artificial data.
         */
        if (Trans) {
            Pattern(pBuffer, BufLen);
            if (Udp) {
                /*
                 * For UDP, we send an initial packet to tell
                 * the receiver that we're starting.
                 */
                (void)DoWrite(Zowie, pBuffer, 4);
                DBGOUT(1,_tprintf(_TEXT("Sent start marker\n")));
            }
            Time0 = GetMillisecondClock();
            while (NumBufs-- && DoWrite(Zowie, pBuffer, BufLen) == BufLen) {
                TotalTransferred += BufLen;
                DBGOUT(1,_tprintf(_TEXT("Sent %d bytes, total = %d\n"),
                    BufLen, TotalTransferred));
            }
            TimeElapsed = GetMillisecondClock() - Time0;
            if (Udp) {
                /*
                 * For UDP, we send a final packet to tell the receiver
                 * that we're done.  Because UDP is not reliable, packets
                 * may get lost, the receiver may overrun, etc.  So we send
                 * several of these in the hope that one will get through.
                 */
                (void)DoWrite(Zowie, pBuffer, 4);
                DBGOUT(1,_tprintf(_TEXT("Sent end marker\n")));
                Delay(1000);
                (void)DoWrite(Zowie, pBuffer, 4);
                Delay(1000);
                (void)DoWrite(Zowie, pBuffer, 4);
            }
        } else {
            Time0 = GetMillisecondClock();
            while ((Got = DoRead(Zowie, pBuffer, BufLen)) > 0) {
                if (Udp && Got <= 4) {
                    /*
                     * This is a UDP start/end marker packet.
                     */
                    static INT Going = 0;
                    if (Going) {
                        /* "EOF" */
                        DBGOUT(1,_tprintf(_TEXT("Got end marker\n")));
                        break;
                    }
                    Going = 1;
                    Time0 = GetMillisecondClock();
                    DBGOUT(1,_tprintf(_TEXT("Got start marker\n")));
                } else {
                    TotalTransferred += Got;
                    DBGOUT(1,_tprintf(_TEXT("Read %d bytes, total = %d\n"),
                           Got, TotalTransferred));
                    if (VerifyData) {
                        INT Bad = CheckPattern(pBuffer, Got);

                        if (Bad)
                            _tprintf(_TEXT("Data corrupt at byte %d of %d!\n"),
                                   Got - Bad, Got);
                    }
                }
            }
            TimeElapsed = GetMillisecondClock() - Time0;
        }
    } else {
        /*
         * Read/write data from stdin/stdout.
         * Timing values here are kinda bogus because they include
         * the time to do the stdin/stdout I/O, which could block.
         */
        Time0 = GetMillisecondClock();
#if !defined(MMLITE)  /* MMLITE doesn't have I/O redirection (yet) */
        if (Trans) {
            while((Got = _read(0, pBuffer, BufLen)) > 0 &&
                  DoWrite(Zowie, pBuffer, Got) == Got)
                TotalTransferred += Got;
        } else {
            while((Got = DoRead(Zowie, pBuffer, BufLen)) > 0 &&
                  _write(1, pBuffer, Got) == Got)
                TotalTransferred += Got;
        }
#else
        _tprintf(_TEXT("Sorry, but MMLITE can't do I/O redirection yet.\n"));
        Usage();  /* never returns */
#endif
        TimeElapsed = GetMillisecondClock() - Time0;
    }

    /*
     * Print out timing info.
     */
    _tprintf(_TEXT("ttcp%hs: %d bytes processed\n"), Trans ? "-t" : "-r",
           TotalTransferred);

    if (TimeElapsed > 0)
        _tprintf(_TEXT("ttcp%hs: took %9d ms = %9d KB/sec, %9d Kbits/sec\n"),
               Trans ? "-t" : "-r", TimeElapsed,
               ((TotalTransferred / TimeElapsed) * 1000) / 1024,
               ((TotalTransferred * 8 / TimeElapsed) * 1000) / 1024);
    else
        _tprintf(_TEXT("Elapsed time is too small/bogus (0 seconds)\n"));

    /*
     * All done.
     */
    if (pBuffer)
        free(pBuffer);
    closesocket(Zowie);
    WSACleanup();
    _tprintf(_TEXT("Exiting...\n"));
#if defined(MMLITE)
    FreeCmdLine(&ArgC,&ArgV);
#endif
    return 0;
}
