/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <tchar.h>

/* Heap memory block headers
 */
typedef struct _USEDBLK USEDBLK, *PUSEDBLK;

struct _USEDBLK { /* Allocated heap memory (multiple of 8 bytes)  */
    UINT Size;    /* Includes header. flags in lsbits (see below) */
    UINT Sign;    /* Signature                                    */
};

typedef struct _FREEBLK FREEBLK, *PFREEBLK;

struct _FREEBLK {    /* Free heap memory                 */
    UINT Size;       /* Must match USEDBLK               */
    FREEBLK *pNext;  /* Next free block on this list     */
    FREEBLK *pPrev;  /* Previous free block on this list */
    PUINT pThis;     /* Actually, last word in block     */
};


/* Flags stored in lower bits of size field in the above structures
 */
#define HB_FREE      0x00000001  /* Block is free                        */
#define HB_PREVFREE  0x00000002  /* Previous block is free               */
#define HB_DECOMMIT  0x00000004  /* Decommited memory in this free block */
#define HB_LOCKCOUNT 0xF0000000  /* Mask for getting Lock Count          */
#define HB_FLAGS     0xF0000007  /* Mask for all the flags and lockcount */
#define HB_SIZE      (~HB_FLAGS) /* Mask for getting the size            */

#if _UINTSIZE == 16
#define HB_ALIGN    4            /* Allocation alignment                 */
#else
#define HB_ALIGN    8            /* Allocation alignment                 */
#endif
#define HB_ROUNDOFF (HB_ALIGN-1) /* Used to round up allocation size     */
/*
 * Note that the least significant bits used for flags (ie lsb of HB_FLAGS)
 * must be smaller or equal to HB_ROUNDOFF
 */

typedef struct {
    ADDR_SIZE Reserve;  /* amount of address space reserved for the heap */
    ADDR_SIZE Commit;   /* of that, how much is actually committed */
    ADDR_SIZE Used;     /* of that, how much is actually allocated */
    ADDR_SIZE MaxUsed;
    UINT Allocs;        /* total number of alloc operations */
} HEAP_STATUS;

/* Free block start with the above header and ends with a pointer
 * to the this header.  Used block could be smaller but when freed
 * is converted to free block.  Thus the minimum size of a block is:
 */
#define MINBLKSIZE ((sizeof(FREEBLK) + HB_ROUNDOFF) & HB_SIZE)

/* Note that a structure in helper.c assumes one can store a LISTNODE and
 * a PTR in the smallest allocated block.
 */

typedef struct _FLH {     /* Free list header                     */
        UINT MaxSize;     /* Size of largest block in this list   */
        FREEBLK Header;   /* Pseudo heap header used as list head */
    } FLH, * PFLH;

/* Heap header
 */
typedef struct _HEAPINFO {  /* Per-heap information                          */
    const struct IHeapVtbl *Vtbl; /* Reserved for IHEAP vtbl                 */
    UINT Sign;              /* Signature                                     */
    UINT Refs;              /* Number of references                          */
    UINT Flags;             /* Passed to CreatHeap                           */
    MUTEX Mx;               /* Serialize access to the heap.                 */
    HEAP_STATUS Status;     /* Reserve, Commit, etc fields                   */
    UINT FreeListCnt;       /* Number of free lists                          */
    FLH FreeList[1];        /* Free list heads, actual number == FreeListCnt */
} HEAPINFO, *PHEAPINFO;

/* BUGBUG - Allow heap creator specify number of free lists, sizes ?
 */
#define FREELISTCNT 4

int _tmain(int argc, _TCHAR **argv)
{
    PIHEAP pHeap = CurrentHeap();
    PHEAPINFO pFF;
    UINT Size;
    PUSEDBLK pfirst;
    ADDR_SIZE TotUsed, TotFree;
    int i;

    pFF = (PHEAPINFO)pHeap;
    _tprintf(_T("Heap at x%x\r\n"), (ADDRESS) pFF);

    for (i = 0; i < FREELISTCNT; i++)
        _tprintf(_T("FL[%d]: %d %d %x %x \r\n"),i,
               pFF->FreeList[i].MaxSize,
               pFF->FreeList[i].Header.Size,
               (ADDRESS)pFF->FreeList[i].Header.pNext,
               (ADDRESS)pFF->FreeList[i].Header.pPrev);

    Size = sizeof(HEAPINFO) + sizeof(FLH) * (FREELISTCNT - 1);
    Size = (Size + HB_ROUNDOFF) & HB_SIZE;
    pfirst = (PUSEDBLK) ((ADDRESS)pFF + Size);
    
    _tprintf(_T("pfirst at x%x\r\n"), (ADDRESS) pfirst);

    TotFree = 0;
    TotUsed = 0;
    for (;;) {
        Size = pfirst->Size & HB_SIZE;
        if (pfirst->Sign)
            _tprintf(_T("Sign %x "), pfirst->Sign);
        _tprintf(_T("@%x: %s %d\r\n"), (ADDRESS) pfirst,
               (pfirst->Size&HB_FREE) ? _T("free") : _T("used"),
               Size);
        if (pfirst->Size&HB_FREE)
            TotFree += Size;
        else
            TotUsed += Size;
        if (Size == 0)
            break;
        pfirst = (PUSEDBLK) ((ADDRESS)pfirst + Size);
    }

    _tprintf(_T("Summary: %d used %d free\n"), TotUsed, TotFree);
    return 0;
}
