/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* A subheap that allocates memory from another heap in relatively large blocks
 * and deallocates them all at once when the subheap is deleted.
 */

#include <mmlite.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "heap.h"

#if defined(_DEBUG)
#define DPRINT(x) printf x
#define EPRINT(x) printf x
#else
#define DPRINT(x) 
#define EPRINT(x) 
#endif

typedef struct {
    ADDR_SIZE Reserve;  /* amount of address space reserved for the heap */
    ADDR_SIZE Used;     /* of that, how much is actually allocated       */
    ADDR_SIZE Avail;    /* ..and how much is left free.                  */
} HEAP_STATUS;


typedef struct HEAPBLOCK {
    ADDRESS End;                /* [b+1 ... (b+1)+Size[ */
    struct HEAPBLOCK *Next;
} *PHEAPBLOCK;

/* Heap header
 */
typedef struct _TmpHeap {  /* Per-heap information                      */
    const struct IHeapVtbl *Vtbl; /* Reserved for IHEAP vtbl             */
    UINT Refs;              /* Number of references                      */
    UINT Flags;             /* Passed to CreatHeap                       */
    MUTEX Mx;               /* Serialize access to the heap.             */
    HEAP_STATUS Status;     /* Reserve, Commit, etc fields               */
    PIHEAP pRealHeap;       /* Underlying heap. Big blocks alloced from here */
    PHEAPBLOCK CurrentBlock;
    ADDRESS FirstFree;
    ADDRESS LastAlloc;      /* Remember last alloc for simple Realloc/Free */
    struct HEAPBLOCK Block;
} HEAPINFO, *PHEAPINFO;

#include <s_TmpHeap_v.c>
#include <s_TmpHeapFactory_v.c>

/* Minimum default alignment is defines by C */
#if _UINTSIZE == 16
#define HB_ALIGN    4            /* Allocation alignment                 */
#else
#define HB_ALIGN    8            /* Allocation alignment                 */
#endif
#define HB_ROUNDOFF (HB_ALIGN-1) /* Used to round up allocation size     */


#define pHP(pHeap) ((PHEAPINFO)(pHeap))

#ifdef _MSC_VER
//#pragma warning(disable:4127)  /* conditional expression is constant */
//#pragma warning(disable:4702)  /* unreachable code */
#endif

#define IncUsedCnt(_phi_,_size_) \
{ \
      (_phi_)->Status.Used += (_size_); \
      (_phi_)->Status.Avail -= (_size_); \
}

/* Lock/Unlock functions
 */
#define LockHeap(_phi_,_Flags_) \
{ \
    if (!((_Flags_) & HEAP_NO_SERIALIZE)) \
        Mutex_Lock(&(_phi_)->Mx); \
}

#define UnlockHeap(_phi_,_Flags_) \
{ \
    if (!((_Flags_) & HEAP_NO_SERIALIZE)) \
        Mutex_Unlock(&(_phi_)->Mx); \
}

/* *** HPAlloc
 *
 * Allocates a heap block of at least Size bytes from the heap pointed
 * by pHeap.  If Flags contains HEAP_ZERO_MEMORY also zero fills the new block.
 * Returns pointer to the block of memory, NULL if error.
 */
PRIVATE PTR MCT THAlloc(IHeap *pHeap, UINT Flags, UINT Size, UINT Alignment)
{
    PHEAPINFO x = pHP(pHeap);
    ADDRESS Mem;
    PHEAPBLOCK pBlock;
    UINT More;

    if (Alignment == 0)         /* default */
        Alignment = HB_ALIGN;

    DPRINT(("THAlloc: pHeap %x, Flags %x, Size %x Align %x\n",
            (UINT) pHeap, Flags, Size, Alignment));

    /* Lock heap if necessary
     */
    LockHeap(x, Flags);

 Again:
   /* Proposed pointer */
    Mem = __ROUND(x->FirstFree, Alignment);

    /* Do we have room for it? */
    if (Mem + Size <= x->CurrentBlock->End) {
        /* OK, we're done */
        x->FirstFree = Mem+Size;
        x->LastAlloc = Mem;
        UnlockHeap(x, Flags);
        if (Flags & HEAP_ZERO_MEMORY)
            memset((PTR) Mem, 0, Size);
        return (PTR) Mem;
    } 

    if (!x->pRealHeap)
        return NULL;

    /* Allocate more from x->pRealHeap and retry */
    More = Size;
    if (More < 0x1000)
        More = 0x1000;          /* XXX Get some better number from somewhere */
    pBlock = x->pRealHeap->v->Alloc(x->pRealHeap, 0,
                                    sizeof(*pBlock)+More, sizeof(UINT));
    if (!pBlock) {
        EPRINT(("Error extending TmpHeap\n"));
        UnlockHeap(x, Flags);
        return NULL;
    }
    x->FirstFree = (ADDRESS)(pBlock+1);
    pBlock->End = x->FirstFree + More;
    /* Link at the end of the chain off x->Block (see Release) */
    pBlock->Next = NULL;
    x->CurrentBlock->Next = pBlock;
    x->CurrentBlock = pBlock;
    goto Again;
}

PRIVATE PTR THRealloc(IHeap *pHeap, UINT Flags, PTR Mem, UINT NewSize, UINT Alignment)
{
    PHEAPINFO x = pHP(pHeap);

    LockHeap(x, Flags);

    if ((ADDRESS) Mem != x->LastAlloc) {
        UnlockHeap(x, Flags);
        return NULL;
    }

    if ((ADDRESS) Mem + NewSize > x->CurrentBlock->End) {
        UnlockHeap(x, Flags);
        return NULL;
    }

    x->FirstFree = (ADDRESS) Mem + NewSize;
    x->LastAlloc = (ADDRESS) Mem;

    UnlockHeap(x, Flags);
    return Mem;
}

/* *** HPStatus
 *
 * Returns status information about this heap.
 */
PRIVATE SCODE MCT THStatus(IHeap *pHeap, ADDR_SIZE *pReserve, ADDR_SIZE *pCommit, ADDR_SIZE *pUsed, ADDR_SIZE *pMaxUsed)
{
    PHEAPINFO phi;

    phi = pHP(pHeap);

    if (pReserve)
        *pReserve = phi->Status.Reserve;
    if (pCommit)
        *pCommit = phi->Status.Reserve;
    if (pUsed)
        *pUsed = phi->Status.Used;
    if (pMaxUsed) {
        *pMaxUsed = phi->Status.Used;
    }

    return NO_ERROR;
}


/* *** Constructor: HeapCreate
 *
 * Creates heap from a block of memory.  pmem points to a block of memory
 * in the appropriate address space to be used.  InitSize is the number of
 * bytes already commited at the beginning of memory block.  MaxSize is
 * total number of bytes in memory block.  Flags passes HEAP_ZERO_MEMORY etc.
 * Returns pointer to the heap, NULL if failed.
 *
 * InitSize is also used as a per-allocation increment (==ChunkSize).
 * After reaching MaxSize allocations will fail no matter what.
 *
 */
SCODE THFCreate(
    PIHEAPFACTORY iThis,
    ADDRESS Mem,
    UINT Flags,
    UINT InitSize,
    UINT MaxSize,
    /*out*/ PIHEAP *ppHeap)
{
    PHEAPINFO x;
    PIHEAP curheap = CurrentHeap();

    /* Establish minimum allocation unit */
    if (InitSize < 256)
        InitSize = 256;

    DPRINT(("Enter THFCreate, Size %08x, Flags %08x\n",MaxSize,Flags));

    /* Initialize HEAPINFO structure (per-heap information).
     */
    x = curheap->v->Alloc(curheap, 0, sizeof(*x), sizeof(UINT));
    if (!x) {
        EPRINT(("Error exit CreateTmpHeap\n"));
        return E_NOT_ENOUGH_MEMORY;
    }

    x->Vtbl = &TmpHeapVtbl;
    x->Refs = 1;
    x->Flags = Flags;
    Mutex_Init(&x->Mx);

    x->Status.Reserve = MaxSize;
    x->Status.Used    = 0;
    x->Status.Avail   = MaxSize;

    x->pRealHeap = NULL;
    x->CurrentBlock = &x->Block;
    x->FirstFree = Mem;
    x->LastAlloc = 0;
    x->Block.End = x->FirstFree + MaxSize;
    x->Block.Next = NULL;

    *ppHeap = (PIHEAP) x;

    return S_OK;
}

SCODE THFCreateNested(
    PIHEAPFACTORY iThis,
    PIHEAP pRealHeap,
    UINT Flags,
    UINT InitSize,
    UINT MaxSize,
    /*out*/ PIHEAP *ppHeap)
{
    PHEAPINFO x;

    /* Establish minimum allocation unit */
    if (InitSize < 256)
        InitSize = 256;

    DPRINT(("Enter CreateTmpHeap Real %08x, InitSize %08x, MaxSize %08x, Flags %08x\n",(UINT)pRealHeap,InitSize,MaxSize,Flags));

    /* Initialize HEAPINFO structure (per-heap information).
     */
    x = pRealHeap->v->Alloc(pRealHeap, 0,
                            sizeof(*x)+InitSize, sizeof(UINT));
    if (!x) {
        EPRINT(("Error exit CreateTmpHeap\n"));
        return E_NOT_ENOUGH_MEMORY;
    }

    x->Vtbl = &TmpHeapVtbl;
    x->Refs = 1;
    x->Flags = Flags;
    Mutex_Init(&x->Mx);

    x->Status.Reserve = MaxSize;
    x->Status.Used    = sizeof(HEAPINFO);
    x->Status.Avail   = MaxSize - sizeof(HEAPINFO);

    x->pRealHeap = pRealHeap;
    x->CurrentBlock = &x->Block;
    x->FirstFree = (ADDRESS)(x+1); /* start allocating right after header */
    x->LastAlloc = 0;
    x->Block.End = x->FirstFree + InitSize;
    x->Block.Next = NULL;

    DPRINT(("Done CreateTmpHeap\n"));
    *ppHeap = (PIHEAP) x;

    return S_OK;
}

/*
 * Exported methods
 */
PRIVATE UINT MCT THAddRef(PIHEAP pThis)
{
    PHEAPINFO x = pHP(pThis);

    return AtomicInc(&x->Refs);
}

PRIVATE UINT MCT THRelease(PIHEAP pThis)
{
    PHEAPINFO x = pHP(pThis);
    UINT newrefs = AtomicDec(&x->Refs);

    if (newrefs == 0) {
        PHEAPBLOCK bnext, b = x->Block.Next;
        PIHEAP heap = x->pRealHeap;
        if (!heap) {
            heap = CurrentHeap();
            assert(b == NULL);
        }
        /* Delete the first block + header */
        heap->v->Free(heap, 0, pThis);
        while (b) {
            /* Delete all subsequent blocks.
             * Be careful not to access deleted data.
             */
            bnext = b->Next;
            heap->v->Free(heap, 0, b);
            b = bnext;
        }
    }

    return newrefs;
}

PRIVATE SCODE MCT THQueryInterface(PIHEAP pThis, REFIID pIid, void* *ppUnk)
{
    return GenericQueryInterface((IUnknown *)pThis,pIid,ppUnk,&IID_IHeap);
}

SCODE THFQueryInterface (
                         /*in*/ PIHEAPFACTORY pThis,
                         /*in*/ REFIID iid,
                         /*out*/ void** ppObject
)
{
    return GenericQueryInterface((PIUNKNOWN) pThis, iid, ppObject, &IID_IHeapFactory);
}

UINT THFAddRef (
        /*in*/ PIHEAPFACTORY pThis
)
{
    PHEAPCOB DbgHeapFactory = (PHEAPCOB) pThis;
    return AtomicInc(&DbgHeapFactory->RefCnt);
}

UINT THFRelease (
                 /*in*/ PIHEAPFACTORY pThis
)
{
    PHEAPCOB DbgHeapFactory = (PHEAPCOB) pThis;
    UINT Refs;

    Refs = AtomicDec(&DbgHeapFactory->RefCnt);
    /* Static object, no free */
    return Refs;
}

PRIVATE struct HEAPCOB TmpHeapCobObject = {
    &TmpHeapFactoryVtbl,
    1
};

PIUNKNOWN TmpHeapCobMain(void)
{
    return (PIUNKNOWN) &TmpHeapCobObject;
}
