/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define OPTIMIZE        /* enable Base.h optimizations, since we're the "base" */

/* Byteorder issues
 * The image format is litte-endian, regardless.
 * But if we are running on a big-endian processor we need swaps
 */
#define DEVICE_ORDER LITTLE_ENDIAN

#include <mmlite.h>
#include <mmhal.h>
#include <machdep.h>   /* FlushCache() and the need of it */
#include <base/loader.h>
#include <base/cons.h>

#include <stdio.h>
#include <assert.h>
#include <string.h>

#if !defined(ppc)
#define DEBUGGER
#endif

#include <base/debugger.h>
#include <diagnostics.h>

extern PIBASERTL pTheBaseRtl;

#define USES_COMPLUS20
#if !defined(__INCLUDE_NT)
#include <wchar.h>
#include <loaders/pe_image.h>
#endif

#define EXPORT_VA(x)     ((x)->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress)
#define IMPORT_VA(x)     ((x)->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress)
#define RELOC_VA(x)      ((x)->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].VirtualAddress)
#define RESOURCE_VA(x)   ((x)->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].VirtualAddress)

#define EXPORT_SIZE(x)   ((x)->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].Size)
#define IMPORT_SIZE(x)   ((x)->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].Size)
#define RELOC_SIZE(x)    ((x)->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_BASERELOC].Size)
#define RESOURCE_SIZE(x) ((x)->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_RESOURCE].Size)
#define DEBUGDIR_VA(x)   ((x)->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].VirtualAddress)
#define DEBUGDIR_SIZE(x) ((x)->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size)

typedef char IMAGE_CHAR;

#ifdef _DEBUG
#define LDRDBG(x) DBGME(0,printf x)
#define LDROUT(x) /*DBGME(1,printf x)*/
#define LDRERR(x) /*DBGME(3,printf x)*/
#else
#define LDRDBG(x)
#define LDROUT(x)
#define LDRERR(x)
#endif

typedef struct SECTION_INFO {
    ADDRESS PrelinkAddress;
    ADDRESS ActualAddress;
    UINT Size;
    UINT Flags;
#define SECTION_INFO_FLAG_NEEDS_DEALLOC 1
} *PSECTION_INFO, SECTION_INFO;

/* This struct is passed around while loading.  It is deleted afterwards. */
typedef struct _LDR {
#ifdef DEBUGGER
    PTR HeaderFluffForDebugger;
#endif
    IMAGE_NT_HEADERS *PeHeader;
    IMAGE_SECTION_HEADER *Sections;
    ADDRESS Discardable;
    ADDRESS Mem;        /* non-discardable */
    BOOL idataNeeded;
    INT idataSectionNumber;
#define MAX_IMPORTED_MODULES 32
    PIMODULE ImportedModules[MAX_IMPORTED_MODULES];
    UINT nImportedModules;
    ADDRESS MetaData;
    MODULEENTRY CorEntryPoint;
    UINT32 EntryPointToken;
#define MAX_RELOCATED_SECTIONS 16
    SECTION_INFO RelocatedSections[MAX_RELOCATED_SECTIONS];
    UINT nRelocatedSections;
#ifdef USES_COMPLUS20
    IMAGE_COR20_HEADER *pCorHdr;
#endif
} *PLDR;

/* Map address to unrelocated section and get the relocated address */
PRIVATE ADDRESS RelocatePtr(PLDR ldr, ADDRESS a)
{
    IMAGE_SECTION_HEADER *Section;
    UINT Count, i;
    ADDRESS aa;

    Count = _bs16(ldr->PeHeader->FileHeader.NumberOfSections);
    Section = ldr->Sections;

    for (i = 0; i < Count; i++, Section++) {
        /* Hack: We stored the old address in PointerToRawData */
        if (a >= Section->PointerToRawData
            && a < Section->PointerToRawData + Section->Misc.VirtualSize)
          {
              /* This is it so map it */
          RelocateIt:
              aa = a - Section->PointerToRawData;
              aa += Section->VirtualAddress;
              LDRDBG(("RelocatePtr %x -> %x (%s)\n", a, aa, Section->Name));
              return aa;
          }
    }
 
    Section = ldr->Sections;
    for (i = 0; i < Count; i++, Section++) {
       /* Work around references just past the section end that really
         * belong to the previous section.  The compiler turned a loop
         * into one.  The problem is that the relocation info is lossy.
         * Anyways, this is pretty safe as correct matches were sorted
         * out in the previous loop.
         */
        LDRDBG(("RelocatePtr RETRY a=%x s=%d p=%x sz=%x\n",
                a, i, Section->PointerToRawData, Section->Misc.VirtualSize));
        if (a == Section->PointerToRawData + Section->Misc.VirtualSize) {
            aa = a - Section->PointerToRawData;
            aa += Section->VirtualAddress;
            LDRDBG(("RelocatePtr IMMEDIATELY PAST END %x -> %x (%s)\n",
                    a, aa, Section->Name));
            return aa;
        }

    }

    /* Work around for ApplyFixup()'s inappropriate optimization.
     * It should use RVA+offset every time instead of RVA once.
     * It can happen that RVA==0 is below the first section.
     */
    if (a == _bs32(ldr->PeHeader->OptionalHeader.ImageBase)) {
        /* Find the first section's virtual address
         */
        int idx = 0;
        Section = ldr->Sections;
        aa = ~0;
        for (i = 0; i < Count; i++, Section++) {
            if (Section->PointerToRawData < aa) {
                aa = Section->PointerToRawData;
                idx = i;
            }
        }
        Section = ldr->Sections + idx;
        LDRDBG(("ImageBase!! "));
        goto RelocateIt;
    }

    LDRERR(("RelocatePtr %x failed\n", a));
    return 0;
}

/* Map address that was RVA in original image into relocated address */
PRIVATE ADDRESS RelocateRva(PLDR ldr, ADDRESS a)
{
    return RelocatePtr(ldr, a + _bs32(ldr->PeHeader->OptionalHeader.ImageBase));
}

PRIVATE SCODE LdrReadImage(PIFILE pFile, PTR pBuf, UINT Size, INT Offset)
{
    UINT CntRead;
    SCODE  sc;
    UINT64 Offset64;

 Again:
    Uint32ToUint64(Offset64,Offset);
    sc = pFile->v->ReadAt(pFile, Offset64, pBuf, Size, &CntRead);
    if (FAILED(sc))
      goto Error;

    if (CntRead && (CntRead < Size)) {
        pBuf = (PTR)((PBYTE)pBuf + CntRead);
        Size -= CntRead;
        Offset += CntRead;
        goto Again;
    }

    if (CntRead != Size) {
      LDRDBG(("LdrReadImage: ReadFromFile: Cnt %x CntRead %x\n",
              Size, CntRead));
      sc = E_BAD_EXE_FORMAT;
      goto Error;
    }

    return NO_ERROR;

 Error:
    return sc;
}

PRIVATE SCODE ReadHeaders(PLDR ldr, PIFILE pFile)
{
    UINT PeHeaderOffset;
    IMAGE_DOS_HEADER *DosHeader;
    IMAGE_NT_HEADERS *PeHeader;
    BYTE *pMem;
    PIHEAP pIHeap = CurrentHeap();
    SCODE sc;

    /* Read in the header.
     */
#ifdef DEBUGGER
    pMem = (BYTE *)pIHeap->v->Alloc(pIHeap, HEAP_DEBUG_NAME("pe_ldr1"),
                                    4096, 4096);
    ldr->HeaderFluffForDebugger = pMem;
#else
    pMem = (BYTE *)pIHeap->v->Alloc(pIHeap, HEAP_DEBUG_NAME("pe_ldr2"),
                                    sizeof(IMAGE_NT_HEADERS), 0);
#endif
    if (pMem == NULL)
      return E_NOT_ENOUGH_MEMORY;

    ldr->PeHeader = (IMAGE_NT_HEADERS *) pMem;  /* ->LDR */

    /* First read DOS header, then NT Header into the same area */
    assert(sizeof(IMAGE_NT_HEADERS) >= sizeof(IMAGE_DOS_HEADER));

    sc = LdrReadImage(pFile, pMem, sizeof(IMAGE_DOS_HEADER), 0);
    if (FAILED(sc))
      return sc;

    DosHeader = (IMAGE_DOS_HEADER *) pMem;
    if (DosHeader->e_magic != _bs16(0x5a4d)) {
        LDRERR(("Error -- Not a valid PE file!\n"));
        return E_INVALID_EXE_SIGNATURE;
    }

    PeHeaderOffset = _bs32(DosHeader->e_lfanew);

#if defined(DEBUGGER) && 0
    assert(PeHeaderOffset + sizeof(IMAGE_NT_HEADERS) <= 1024);
    pMem += PeHeaderOffset;
#endif
    sc = LdrReadImage(pFile, pMem, sizeof(IMAGE_NT_HEADERS), PeHeaderOffset);
    if (FAILED(sc))
      return sc;

    /* Is this a valid PE file?
     */
    PeHeader = (IMAGE_NT_HEADERS *) pMem;
    if (PeHeader->Signature != _bs32(IMAGE_NT_SIGNATURE)) {
        LDRERR(("Error -- Not a valid PE file!\n"));
        return E_INVALID_EXE_SIGNATURE;
    }

#if defined(MMLITE_VERSION)
    /* Is the version compatible with our build ?
     */
    if (_bs16(PeHeader->OptionalHeader.MajorImageVersion) > MMLITE_VERSION) {
        LDRERR(("Error -- Incompatible version!\n"));
        return E_INCOMPATIBLE_VERSION;
    }
#endif

    /* Reserve space for section headers.  */
#if defined(DEBUGGER)
    pMem += sizeof(IMAGE_NT_HEADERS);
    assert(_bs32(PeHeader->OptionalHeader.SizeOfHeaders) <= 4096);
#else
    pMem = pIHeap->v->Alloc(pIHeap, HEAP_DEBUG_NAME("pe_ldr3"),
        _bs32(PeHeader->OptionalHeader.SizeOfHeaders) - sizeof(IMAGE_NT_HEADERS), 0);
#endif

    if (!pMem)
      return E_NOT_ENOUGH_MEMORY;

    ldr->Sections = (IMAGE_SECTION_HEADER *) pMem;      /* ->LDR */

    /* Read remainder of headers */
    sc = LdrReadImage(pFile, pMem,
             _bs32(PeHeader->OptionalHeader.SizeOfHeaders) - sizeof(IMAGE_NT_HEADERS),
             PeHeaderOffset + sizeof(IMAGE_NT_HEADERS));

    return sc;
}

PRIVATE void CalculateSizes(PLDR ldr, UINT *pSize, UINT *pDiscardableSize)
{
    IMAGE_SECTION_HEADER *Section;
    UINT  Count, i, Size;

    Count = _bs16(ldr->PeHeader->FileHeader.NumberOfSections);
    Section = ldr->Sections;

    for (i = 0; i < Count; i++, Section++) {
#if defined(ppc)
        /* With this compiler, the RawSize is the true one
         */
        Size = _bs32(Section->SizeOfRawData);
#else
        Size = _bs32(Section->Misc.VirtualSize);
#endif
#if defined(DEBUGGER)
        /* Round up to page boundary */
        Size = (Size + 4095) & ~4095;
#else
        /* Round up to 8 byte boundary */
        Size = (Size + 7) & ~7;
#endif
        LDRDBG(("Section %s size = x%x\n",Section->Name,Size));
        Section->Misc.VirtualSize = Size;

        if (_bs32(Section->Characteristics) & IMAGE_SCN_MEM_DISCARDABLE)
          *pDiscardableSize += Size;
        else
          *pSize += Size;
    }
}

PRIVATE SCODE ReadAndFixSections(PLDR ldr, PIFILE pFile)
{
    IMAGE_SECTION_HEADER *Section;
    PSECTION_INFO MySection;
    UINT Count, i, ZeroSize, RawSize;
    SCODE sc;
    ADDRESS va, OldVa, PreLinkedAt;
    ADDRESS DiscardableVa, PersistentVa;
    UINT DiskOff;
    BOOL Discard;

    Count = _bs16(ldr->PeHeader->FileHeader.NumberOfSections);
    Section = ldr->Sections;
    PreLinkedAt = _bs32(ldr->PeHeader->OptionalHeader.ImageBase);

    LDRDBG(("ReadAndFixSections Mem=%x Count=%x Section=%x pe=%x\n",
            ldr->Mem, Count, (ADDRESS) Section, (ADDRESS) ldr->PeHeader));

    PersistentVa = ldr->Mem;
    DiscardableVa = ldr->Discardable;

    for (i = 0; i < Count; i++, Section++) {
        OldVa = _bs32(Section->VirtualAddress);
        RawSize = _bs32(Section->SizeOfRawData);
        DiskOff = _bs32(Section->PointerToRawData);
        Discard = _bs32(Section->Characteristics) & IMAGE_SCN_MEM_DISCARDABLE;

        if (Discard)
            va = DiscardableVa;
        else
            va = PersistentVa;

//        Section->Misc.VirtualSize = _bs32(Section->Misc.VirtualSize);
        LDRDBG(("Read %s sect %s: va %x -> %x, size %x raw %x dsk %x\n",
                Discard ? "discardable" : "",
                Section->Name, OldVa, va, Section->Misc.VirtualSize,
                RawSize, DiskOff));

        if (RawSize > 0) {
            if (RawSize > Section->Misc.VirtualSize)
                RawSize = Section->Misc.VirtualSize;
            sc = LdrReadImage(pFile, (PTR) va, RawSize, DiskOff);
            if (FAILED(sc)) {
                LDRERR(("ReadSections %s failed %x\n",Section->Name,sc));
                return sc;
            }
        }
        /* Store old va into PointerToRawData for later relocation use */
        Section->PointerToRawData = OldVa + PreLinkedAt;

        /* Make VirtualAddress point to the in core one */
        Section->VirtualAddress = va;

        ZeroSize = Section->Misc.VirtualSize - RawSize;
        LDRDBG(("Zero from %x for %x\n", va + RawSize, ZeroSize));
        if (ZeroSize > 0)
          memset((PTR) (va + RawSize), 0, ZeroSize);

        /* Special processing for .idata */
        if (_bs32(OldVa) == IMPORT_VA(ldr->PeHeader)) {
            ldr->idataSectionNumber = i; /* ->LDR */
        }

        if (!Discard) {
            /* Record section in our relocation map */
            if (ldr->nRelocatedSections > MAX_RELOCATED_SECTIONS) {
                LDRERR(("Too many sections\n"));
                return E_TOO_MANY_CONTIGUOUS;
            }
            MySection = &ldr->RelocatedSections[ldr->nRelocatedSections++];
            MySection->PrelinkAddress = OldVa + PreLinkedAt;
            MySection->ActualAddress = va;
            MySection->Size = Section->Misc.VirtualSize;
            MySection->Flags = 0;
        }

        /* Advance to next section load point */
        if (Discard)
          DiscardableVa += Section->Misc.VirtualSize;
        else
          PersistentVa += Section->Misc.VirtualSize;
    }

    return S_OK;
}

PRIVATE void FixupHeader(PLDR ldr)
{
    int i;
    ADDRESS a, aa;
    IMAGE_DATA_DIRECTORY *pEntry;

    /* Try to find (and remember) the COM+ 2.0 header first */
    pEntry = &ldr->PeHeader->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_COMHEADER];
    if (pEntry->VirtualAddress != 0 && pEntry->Size != 0) {
        pEntry->VirtualAddress = _bs32(pEntry->VirtualAddress);
        pEntry->Size = _bs32(pEntry->Size);
        if (pEntry->Size >= offsetof(IMAGE_COR20_HEADER, StrongNameSignature)) {
            ldr->pCorHdr = (IMAGE_COR20_HEADER *)RelocateRva(ldr, pEntry->VirtualAddress);
            if (ldr->pCorHdr != NULL) {
                pEntry = &ldr->pCorHdr->MetaData;
                if (pEntry->VirtualAddress != 0 && pEntry->Size != 0) {
                    pEntry->VirtualAddress = _bs32(pEntry->VirtualAddress);
                    pEntry->Size = _bs32(pEntry->Size);
                    ldr->MetaData = RelocateRva(ldr, pEntry->VirtualAddress);
                    ldr->EntryPointToken = _bs32(ldr->pCorHdr->EntryPointToken);
                }
            }
        }
    }

    /* Go through all the data directories and fixup their virtual addresses */
    for (i = 0; i < IMAGE_NUMBEROF_DIRECTORY_ENTRIES; i++) {
        a = ldr->PeHeader->OptionalHeader.DataDirectory[i].VirtualAddress;
        if (a == 0)
          continue;

        a = _bs32(a);
        aa = RelocateRva(ldr, a);
        LDRDBG(("DataDir[%d].VirtualAddress %x -> %x\n", i, a, aa));
        if (aa == 0)
            aa = a;
        ldr->PeHeader->OptionalHeader.DataDirectory[i].VirtualAddress = aa;
    }
}


PRIVATE UINT ResolveImportEntry(PLDR ldr, PIMODULE pModule, UINT iat)
{
    ADDRESS Val = 0;
    SCODE sc;

    if (iat & IMAGE_ORDINAL_FLAG) {
        iat &= ~IMAGE_ORDINAL_FLAG;
        sc = pModule->v->GetFunctionAddress(pModule, NULL, iat, &Val);
        if (FAILED(sc))
            LDRERR(("LOADER: Ordinal %u not found in module %x\n",
                    iat, (ADDRESS) pModule));
    } else {
        PIMAGE_IMPORT_BY_NAME pImp = (PIMAGE_IMPORT_BY_NAME) RelocateRva(ldr,
                                                                         iat);
        UINT Hint = (UINT) pImp->Hint;
        IMAGE_CHAR *pName = (IMAGE_CHAR *) pImp->Name;
#ifdef _UNICODE
        _TCHAR *_tname = _strtow(pName);
#else
#define _tname pName
#endif
        if (_tname == NULL)
            sc = E_NOT_ENOUGH_MEMORY;
        else
            sc = pModule->v->GetFunctionAddress(pModule, _tname, Hint, &Val);
#ifdef _UNICODE
        free(_tname);
#else
#undef _tname
#endif
        if (ldr->pCorHdr != NULL && ldr->CorEntryPoint == NULL && strcmp(pName, "_CorExeMain") == 0)
            ldr->CorEntryPoint = (MODULEENTRY)Val;
        if (FAILED(sc))
            LDRERR(("LOADER: Name %hs not found in module %x\n",
                    pName, (ADDRESS) pModule));
    }
    return Val;
}

PRIVATE SCODE ResolveImports(PLDR ldr)
{
    PIMODULE pModule;
    IMAGE_IMPORT_DESCRIPTOR *pidt;
    UINT *     piat;
    UINT *     piatPreSnap;
    UINT       Tmp;

    if (IMPORT_VA(ldr->PeHeader) == 0)
        return NO_ERROR; /* No imports */

    LDRDBG(("Resolving imports...\n"));

    pidt = (IMAGE_IMPORT_DESCRIPTOR *) _bs32(IMPORT_VA(ldr->PeHeader));

    for ( ; pidt->Name ; pidt++ ) {
        IMAGE_CHAR *pModName = (IMAGE_CHAR *) RelocateRva(ldr, _bs32(pidt->Name));

        /* Find the module header for the imported module
        */
#ifdef _UNICODE
        pModule = NULL;
        LDRERR(("PeLdr needs Unicode conversion for module lookup\n"));
        return E_MOD_NOT_FOUND;
#else
        pModule = LookupModule(pModName, 0);
        if (pModule == NULL) {
            PIPROCESS pPrc = CurrentProcess();
            pPrc->v->LoadImage(pPrc, pModName, NULL, LOAD_IMAGE_CALL_ENTRY_POINT, &pModule);
            if (pModule == NULL) {
#ifdef SIM
                DBGME(3,printf("WARNING: module \"%s\" not loaded and not loadable, trying NT loader\n", pModName));
                pModule = NT_LookupModule(pModName, 0);
                if (pModule == NULL)
#endif
                {
                    LDRERR(("ResolveImports: No Module for %hs\n",pModName));
                    return E_MOD_NOT_FOUND;
                }
            }
        }
#endif

        /*
         * Resolve all of the import address table entries for this module.
         * FirstThunk has the pre-snapped table, Characteristics
         * (as the name implies?) has the relocateable values
         */
        piatPreSnap = (UINT *) RelocateRva(ldr, _bs32((UINT) pidt->FirstThunk));

        if (pidt->Characteristics)
            piat = (UINT *) RelocateRva(ldr, _bs32((UINT) pidt->Characteristics));
        else
            piat = piatPreSnap;

        while (*piat != 0) {
            Tmp = ResolveImportEntry(ldr, pModule, _bs32(*piat));
            LDRDBG(("ResolveImports (%x=%hs %x)-> %x; snap=%x piat=%x\n",
                    (ADDRESS)pModule, pModName, _bs32(*piat), Tmp, 
                    (ADDRESS)piatPreSnap, (ADDRESS)piat));
            if (Tmp == 0) {
                /* Unable to resolve import. */
                LDRERR(("Unresolved external %x not found in module %x %hs\n",
                        (ADDRESS) piatPreSnap, (ADDRESS) pModule, pModName));
                pModule->v->Release(pModule);
                return E_PROC_NOT_FOUND;
            }

            *piatPreSnap++ = Tmp;
            piat++;
        }
        if (ldr->nImportedModules >= MAX_IMPORTED_MODULES) {
            pModule->v->Release(pModule);
            return E_TOO_MANY_CONTIGUOUS; /* XXX too many dependents */
        }
        ldr->ImportedModules[ldr->nImportedModules++] = pModule;
    }

    return NO_ERROR;
}

/* This function converts a pointer into the import section into the pointer
 * the import table points at.
 */
PRIVATE ADDRESS ReResolveImportEntry(PLDR ldr, ADDRESS addr)
{
    IMAGE_IMPORT_DESCRIPTOR *pidt;
    UINT *     piat;
    ADDRESS Result = 0;

    if (IMPORT_VA(ldr->PeHeader) == 0)
        return 0; /* No imports */

    LDRDBG(("Shortcutting import entry %x\n", addr));

    pidt = (IMAGE_IMPORT_DESCRIPTOR *) _bs32(IMPORT_VA(ldr->PeHeader));
    for ( ; pidt->Name ; pidt++ ) {
        piat = (UINT *) RelocateRva(ldr, (UINT) pidt->FirstThunk);

        for ( ; *piat != 0; piat++)
          if (addr == (ADDRESS) piat)
            Result = *piat;
    }

    return Result;
}

PRIVATE BOOL PointsToSection(PLDR ldr, INT SectionNumber, ADDRESS a)
{
    BOOL rv;
    IMAGE_SECTION_HEADER *Section;

    if (SectionNumber < 0
        || SectionNumber >= ldr->PeHeader->FileHeader.NumberOfSections)
      return FALSE;

    Section = &ldr->Sections[SectionNumber];

    rv = (a >= Section->VirtualAddress
          && a < Section->VirtualAddress + Section->Misc.VirtualSize);

    LDRDBG(("PointsToSection sn=%d a=%x -> %s\n",
            SectionNumber, a, rv ? "TRUE" : "FALSE"));
    return rv;
}

#ifdef i386
PRIVATE BOOL IsJmpIndirect(ADDRESS ReloPtr)
{
    UINT8 *instr = (UINT8*) (ReloPtr - 2);
    BOOL rv = instr[0] == 0xff && instr[1] == 0x25;

    LDRDBG(("IsJmpIndirect [%02x %02x] -> %s\n",
            instr[0], instr[1], rv ? "TRUE" : "FALSE"));
    return rv;
}

PRIVATE BOOL IsCallIndirect(ADDRESS ReloPtr)
{
    UINT8 *instr = (UINT8*) (ReloPtr - 2);
    BOOL rv = instr[0] == 0xff && instr[1] == 0x15;

    LDRDBG(("IsCallIndirect [%02x %02x] -> %s\n",
            instr[0], instr[1], rv ? "TRUE" : "FALSE"));
    return rv;
}

PRIVATE void IndirectToRel(ADDRESS ReloPtr, ADDRESS Tgt, UINT8 OpCode)
{
    UINT8 *instr = (UINT8*) (ReloPtr - 2);

    LDRDBG(("IndirectToRel: ReloPtr=%x Tgt=%x ", ReloPtr, Tgt));

    /* Tgt is absolute.  Make it PC-relative */
    Tgt = Tgt - (ADDRESS) &instr[5];

    LDRDBG(("PC-target=%x (-%x) ", Tgt, -(INT)Tgt));
    LDRDBG(("%02x %02x %02x %02x %02x %02x ",
            instr[0],  instr[1], instr[2], instr[3],  instr[4],  instr[5]));
    instr[0] = OpCode;
    *(UINT32 *)&instr[1] = Tgt; /* PC-relative immediate */
    instr[5] = 0x90;    /* Pad with NOP */

    LDRDBG(("-> %02x %02x %02x %02x %02x %02x\n",
            instr[0],  instr[1], instr[2], instr[3],  instr[4],  instr[5]));
}
#endif /* i386 */

PRIVATE ADDRESS RelocateEntry(PLDR ldr, ADDRESS Tgt, ADDRESS *ReloPtr)
{
    ADDRESS a;

    /* In a real memory system we want to call server functions directly
     * and not through the import table.  What we do is to first init
     * the import table, then do the fixups.  Now when we are relocating
     * a fixup entry we check if this was indeed a call through the
     * import table.  If it was we replace it with a direct call if we
     * know how.
     * Note that if all accesses to the import section could be resolved
     * the .idata section becomes dead and can be thrown away.
     *
     * In other cases we just relocate based on RelocDelta.
     */
    LDRDBG(("#%x ",*ReloPtr));

    /* First relocate.  This way we point into the relocated section. */
    a = RelocatePtr(ldr, Tgt);

    if (PointsToSection(ldr, ldr->idataSectionNumber, a)) {
        ADDRESS aa;

        aa = ReResolveImportEntry(ldr, a);
        LDRDBG(("RelocateEntry %x -> %x -> %x\n", Tgt, a, aa));

#ifdef i386
        if (aa && IsJmpIndirect((ADDRESS) ReloPtr)) {
            IndirectToRel((ADDRESS) ReloPtr, aa, 0xe9);
            return 0;
        }
        if (aa && IsCallIndirect((ADDRESS) ReloPtr)) {
            IndirectToRel((ADDRESS) ReloPtr, aa, 0xe8);
            return 0;
        }
#endif
        /* We had a reference to idata that could not be removed.
         * Thus .idata can't be discarded.
         */
        ldr->idataNeeded++; /* ->LDR */
    }
    return a;
}

PRIVATE SCODE FixupEntry(PLDR ldr, UINT Type, ADDRESS *pTgt, UINT16 Adjust, ADDRESS Diff)
{
    INT      Temp;
#if (BYTE_ORDER==BIG_ENDIAN)
    const int low16 = 1;
#else
    const int low16 = 0;
#endif
    /* BUGUBG - Compiler may generate jump table for this switch
     * statement, which breaks when code run at mapped addresses.
     * Use if else instead.
     */
    if (Type == IMAGE_REL_BASED_HIGHLOW) {
        /* HighLow - (32bits) relocate the high and low halfs */
        LDRDBG(("HIGHLOW "));
        Temp = RelocateEntry(ldr, *pTgt, pTgt);
        if (Temp)
          *pTgt = Temp;
    }
    else if (Type == IMAGE_REL_BASED_ABSOLUTE) {
        LDRDBG(("ABS\n"));
        ; /* Absolute : no fixup required */
    }
    else if (Type == IMAGE_REL_BASED_HIGH) {
        /* High - (16-bits) relocate the high half of an address */
        Temp = (((UINT16*) pTgt)[low16]) << 16;
        LDRDBG(("HIGH "));
        Temp = RelocateEntry(ldr, Temp, pTgt);
        assert(Temp);
        ((UINT16*) pTgt)[low16] = (UINT16) (Temp >> 16);
    }
    else if (Type == IMAGE_REL_BASED_LOW) {
        /* Low - (16-bit) relocate the low half of an address */
        Temp = ((INT16*) pTgt)[low16];
#if defined(ppc)
        /* IMHO this should be for all platforms but what do I know
         */
        Temp = Temp + Diff;
        LDRDBG(("LOW #%x Diff=%x Temp=%x\n",*pTgt,Diff,Temp));
#else
        LDRDBG(("%x LOW ",Adjust));
        UnusedParameter(Diff);
        Temp = RelocateEntry(ldr, Temp, pTgt);
#endif
        assert(Temp);
        ((UINT16*) pTgt)[low16] = (UINT16) Temp;
    }
    else if (Type == IMAGE_REL_BASED_HIGHADJ) {
        /* Adjust high - (16-bits) relocate the high half of an
         * address and adjust for sign extension of low half.
         */
        Temp = (((UINT16*) pTgt)[low16]) << 16;
        Temp += (INT16) Adjust;
        LDRDBG(("HIGHADJ "));
        Temp = RelocateEntry(ldr, Temp, pTgt);
        assert(Temp);
        Temp += 0x8000;
        ((UINT16*) pTgt)[low16] = (UINT16) (Temp >> 16);
        /* Tell caller that we used up next relocation entry */
        return S_FALSE;
    }
    else if (Type == IMAGE_REL_BASED_MIPS_JMPADDR) {
        /* JumpAddress - (32-bits) relocate a MIPS jump address */
        Temp = (*((PUINT) pTgt) & 0x3ffffff) << 2;
        LDRDBG(("JMPADDR "));
        Temp = RelocateEntry(ldr, Temp, pTgt);
        assert(Temp);
        *((PUINT) pTgt) = (*((PUINT) pTgt) & ~0x3ffffff) |
          ((Temp >> 2) & 0x3ffffff);
    }
    else {
        DBGME(3,printf("FixupEntry: Unrecognized fixup Type=%x\n", Type));
        return E_BAD_EXE_FORMAT;
    }
    return S_OK;
}

PRIVATE SCODE ApplyFixups(PLDR ldr)
{
    IMAGE_BASE_RELOCATION     *pfxrCur, *pfxrEnd;
    UINT     FixupBase;
    UINT16   *pReloc, *pRelocEnd;
    ADDRESS  RelocationDifference;
    SCODE sc;

    if (RELOC_SIZE(ldr->PeHeader) == 0) {
        LDRERR(("WARNING: module contains no relocations!\n"));
        return NO_ERROR; /* This module contains no fixups */
    }

    LDRDBG(("ApplyFixups begins, %d relocs\n",_bs32(RELOC_SIZE(ldr->PeHeader))));

    /* Find beginning, end of relocation information */
    pfxrCur = (IMAGE_BASE_RELOCATION *) (RELOC_VA(ldr->PeHeader));
    pfxrEnd = (IMAGE_BASE_RELOCATION *) ((UINT) pfxrCur + _bs32(RELOC_SIZE(ldr->PeHeader)));

    while (pfxrCur < pfxrEnd) {
        /* Find end of relocation block */
        pReloc = (UINT16*) pfxrCur;
        pRelocEnd = (UINT16*) ((UINT) pReloc + _bs32(pfxrCur->SizeOfBlock));

        /* empty bundle indicates the end of the fixup section
         * even if smaller than the indicated section size.
         */
        if (pRelocEnd == pReloc)
            break;

        /* BUGBUG! check if pRelocEnd is within relocation section
         * and not less than block start.
         */

        /* find page virtual address */
        LDRDBG(("RVA %x ",_bs32(pfxrCur->VirtualAddress)));
        FixupBase = RelocateRva(ldr, _bs32(pfxrCur->VirtualAddress));
        assert(FixupBase != 0);
        RelocationDifference = FixupBase - 
            (_bs32(ldr->PeHeader->OptionalHeader.ImageBase) + _bs32(pfxrCur->VirtualAddress));

        /* Relocation info assumes page size <= 4K
         */
        /* Do page relocation:
         *
         * skip page RVA and size and point to relocation information
         * for each iteration: advance to next type/offset
         */
        for (pReloc = (UINT16*) ++pfxrCur; pReloc < pRelocEnd ; pReloc++) {
            UINT16   TypeInfo;
            ADDRESS  Offset, *pTgt;

            TypeInfo = _bs16(*pReloc);
            Offset = TypeInfo & 0xfff;
            pTgt = (ADDRESS*) (FixupBase + Offset);

            LDRDBG(("Rel %x (%x) ",Offset,pTgt));
            sc = FixupEntry(ldr, TypeInfo >> 12, pTgt, _bs16(*(pReloc+1)),
                            RelocationDifference);
            if (FAILED(sc))
              goto Out;
            if (sc == S_FALSE)
              pReloc++;         /* skip one */
        }
        pfxrCur = (IMAGE_BASE_RELOCATION *) pReloc;
    }
    sc = NO_ERROR;
 Out:
    LDRDBG(("ApplyFixups ends sc=%x\n",sc));
    return sc;
}

PRIVATE ADDRESS GetEntryPoint(PLDR ldr)
{
    ADDRESS Entry;

    if (ldr->pCorHdr != NULL)
        Entry = (ADDRESS)ldr->CorEntryPoint;
    else {
        /* Fix up entry point */
        Entry = _bs32(ldr->PeHeader->OptionalHeader.AddressOfEntryPoint);
        if (Entry != 0) { /* Fixup entry point */
            Entry = RelocateRva(ldr, Entry);;
        }
        LDROUT(("Entrypoint: %x -> %x\n",
                _bs32(ldr->PeHeader->OptionalHeader.AddressOfEntryPoint), Entry));
    }
    return Entry;
}

/* See if we have an export table and if so relocate the entries */
PRIVATE SCODE PrepareExports(PLDR ldr, PTR *pExportData, UINT *pExportSize)
{
    IMAGE_EXPORT_DIRECTORY *pedt;
    PUINT   peat;
    PUINT  pName;   /* pointer to current name in name table */
    UINT    i;
    ADDRESS a;

    if (EXPORT_SIZE(ldr->PeHeader) == 0) {
        LDROUT(("No exports\n"));
        return NO_ERROR;
    }
    LDROUT(("RelocateExports: found exports\n"));

    pedt = (IMAGE_EXPORT_DIRECTORY *) EXPORT_VA(ldr->PeHeader);

    peat = (PUINT) RelocateRva(ldr, (UINT) pedt->AddressOfFunctions);
    pedt->AddressOfFunctions = (PTR) peat; /* update it */

    pName = (PUINT) RelocateRva(ldr, (UINT) pedt->AddressOfNames);
    pedt->AddressOfNames = (PTR) pName; /* ditto */

    pedt->AddressOfNameOrdinals = (PTR) RelocateRva(ldr,
                                          (UINT) pedt->AddressOfNameOrdinals);

    /* Fixup function pointers */
    for (i = 0; i < pedt->NumberOfFunctions; i++) {
        if (peat[i] == 0) {     /* Unused slot */
            LDROUT(("ExportFromOrdinal i=%d skipped\n",i));
            continue;
        }
        a = RelocateRva(ldr, peat[i]);
        LDROUT(("ExportFromOrdinal i %d Addr %x -> %x\n",i,peat[i],a));
        peat[i] = a;
    }

    /* Fixup name pointers */
    for (i = 0; i < pedt->NumberOfNames; i++) {
        pName[i] = RelocateRva(ldr, pName[i]);
        LDROUT(("ExportFromName %s\n", (char*)pName[i]));
    }

    *pExportData = (PTR) EXPORT_VA(ldr->PeHeader);
    *pExportSize = EXPORT_SIZE(ldr->PeHeader);

    LDROUT(("Exports processed, data is %x size is %x\n",
            (ADDRESS)*pExportData, *pExportSize));

    return NO_ERROR;
}

/* Called from Module->GetFunctionAddress in process.c */
SCODE MCT PeGetFunctionAddress (
        PTR ExportData,
        UINT ExportSize,
        const _TCHAR * ModName,
        UINT Ordinal,
        /* [out] */ ADDRESS *pAddress)
{
    IMAGE_EXPORT_DIRECTORY *pedt;
    PUINT   peat;
    PUINT  pName;   /* pointer to current name in name table */
    UINT16 *pord;      /* pointer to beginning of ordinal table */
    UINT   i;
    IMAGE_CHAR *pModStr;
    ADDRESS Addr;
    INT    HighLimit;
    INT    LowLimit;
    INT    Direction;

    LDROUT(("PeGetFunctionAddress(%x %x \"%s\" %d %x)\n",
            (ADDRESS)ExportData, ExportSize, ModName ? ModName : "(nil)",
            Ordinal, (ADDRESS)pAddress));

    if (ExportSize == 0) {
        LDROUT(("PeGetFunctionAddress: No exports\n"));
        return E_PROC_NOT_FOUND;
    }

    /* We did the relocations above so everything is now relocated. */
    pedt = (IMAGE_EXPORT_DIRECTORY *) ExportData;
    peat = (PUINT) pedt->AddressOfFunctions;
    pName = (PUINT) pedt->AddressOfNames;
    pord = (UINT16*) pedt->AddressOfNameOrdinals;

    /* If we have a name, map that to ordinal first */
    if (ModName) {
        /* Search the names table for the specified function name. */
        HighLimit = pedt->NumberOfNames - 1;
        LowLimit = 0;

        for (;;) {
            i = (HighLimit + LowLimit) >> 1;
            pModStr = (IMAGE_CHAR *) pName[i];

#ifdef _UNICODE
            LDRERR(("PeLdr needs Unicode conversion for GetFunctionAddr\n"));
            goto Error;
#else
            Direction = strcmp(pModStr,ModName);
#endif
            if (Direction == 0)
              break;

            if (HighLimit <= LowLimit)
              goto Error;

            if (Direction > 0) {
                if (i > 1)
                  HighLimit = i - 1;
                else
                  HighLimit = 0;
            }
            else
              LowLimit = i + 1;
        }
        Ordinal = (UINT) pord[i] + pedt->Base;
        LDROUT(("Mapped name %s to ordinal %d\n", ModName, Ordinal));
    }

    /* Map ordinal to address */
    Ordinal -= pedt->Base;
    if (Ordinal >= pedt->NumberOfFunctions) {
        goto Error;
    }
    Addr = peat[Ordinal];
    LDROUT(("Ordinal %d (+%d) is %x\n", Ordinal, pedt->Base, Addr));

    if (pAddress)
      *pAddress = Addr;
    return NO_ERROR;

  Error:
    LDRERR(("PeGetFunctionAddress failed\n"));
    return E_INVALID_ADDRESS;
}

PRIVATE void ReleaseLdr(PLDR ldr)
{
    PIHEAP pIHeap = CurrentHeap();
    PIMODULE pMod;
    UINT i;

#ifdef DEBUGGER
    if (ldr->HeaderFluffForDebugger)
        pIHeap->v->Free(pIHeap, 0, ldr->HeaderFluffForDebugger);
#else
    if (ldr->PeHeader)
        pIHeap->v->Free(pIHeap, 0, ldr->PeHeader);
    if (ldr->Sections)
        pIHeap->v->Free(pIHeap, 0, ldr->Sections);
    if (ldr->Mem)
        pIHeap->v->Free(pIHeap, 0, (PTR) ldr->Mem);
#endif
    if (ldr->Discardable)
        pIHeap->v->Free(pIHeap, 0, (PTR) ldr->Discardable);

    for (i = 0; i < ldr->nImportedModules; i++) {
        pMod = ldr->ImportedModules[i];
        LDROUT(("Dropping ref for module x%x in ReleaseLdr\n", (ADDRESS)pMod));
        pMod->v->Release(pMod);
    }
    ldr->nImportedModules = 0;
}

/*
 * Load a PE-formatted program image into memory.
 */

SCODE LdrLoadImage(
           PINAMESPACE pNameSpace,
           const _TCHAR * pImage,
           const _TCHAR * pArgs,
           UINT32 Flags,
           IModule **ppIModule)
{
    SCODE sc;
    UINT64 FileSize, FileSize64;
    UINT ActualSize = 0, DiscardableSize = 0;

    MODULEENTRY EntryPoint;
    PIMODULE pMod;
    PIPROCESS pPrc;
    PIHEAP pIHeap = CurrentHeap();
    PIFILE pFile = NULL;
    PIUNKNOWN pUnk;
    struct _LDR _ldr, *ldr = NULL;
    PTR ExportData;
    UINT ExportSize;

    LDRDBG(("LoadImage(%s)\n", pImage));

    if (pNameSpace == NULL)
        return E_INVALID_PARAMETER;

    sc = pNameSpace->v->Bind(pNameSpace, pImage, NAME_SPACE_READ, &pUnk);
    if (FAILED(sc))
        goto Error;

    sc = pUnk->v->QueryInterface(pUnk, &IID_IFile, (void**) &pFile);
    pUnk->v->Release(pUnk);     /* Done with pUnk */
    if (FAILED(sc))
        goto Error;

    sc = pFile->v->GetSize(pFile, &FileSize);
    if (FAILED(sc))
        goto Error;

    Int64FromHighAndLow(FileSize64,0,sizeof(IMAGE_NT_HEADERS));
    if (Uint64Less(FileSize,FileSize64)) {
        sc = E_BAD_EXE_FORMAT;
        goto Error;
    }

    ldr = &_ldr;
    memset(ldr, 0, sizeof(*ldr));
    ldr->idataSectionNumber = -1;

    sc = ReadHeaders(ldr, pFile);
    if (FAILED(sc))
      goto Error;

    CalculateSizes(ldr, &ActualSize, &DiscardableSize);

    /* Allocate memory for persistent program data, e.g. code */
    if (ActualSize > 0) {
#ifdef DEBUGGER
        /* The debugger expects contiguity and it "knows" too much */
        PTR tmp = ldr->HeaderFluffForDebugger;
        ActualSize += 4096;
        tmp = pIHeap->v->Realloc(pIHeap, HEAP_DEBUG_NAME("pe_ldr4"),
                                 tmp, ActualSize, 4096);
        if (!tmp) {
            /* failed, old memory intact */
            sc = E_NOT_ENOUGH_MEMORY;
            goto Error;
        }
        ldr->Mem = (ADDRESS) tmp + 4096;
        if (tmp != ldr->HeaderFluffForDebugger) {
            /* Realloc moved memory. Readjust some pointers. */
            ldr->HeaderFluffForDebugger = tmp;
            ldr->PeHeader = tmp;
            ldr->Sections = (PTR) ((ADDRESS) tmp + sizeof(IMAGE_NT_HEADERS));
        }
#else
        ldr->Mem = (ADDRESS) pIHeap->v->Alloc(pIHeap,
                                              HEAP_DEBUG_NAME("pe_ldr5"),
                                              ActualSize, 0);
#endif
        if (!ldr->Mem) {
            sc = E_NOT_ENOUGH_MEMORY;
            goto Error;
        }
    }

    /* Allocate memory for discardable data, e.g. relocations */
    if (DiscardableSize > 0) {
        ldr->Discardable = (ADDRESS)pIHeap->v->Alloc(pIHeap,
                                                    HEAP_DEBUG_NAME("pe_ldr6"),
                                                     DiscardableSize, 0);
        if (!ldr->Discardable) {
            sc = E_NOT_ENOUGH_MEMORY;
            goto Error;
        }
    }

    LDRDBG(("LoadImage %s: ActualSize=%x DiscardableSize=%x ldr=%x\n",
            pImage, ActualSize, DiscardableSize, (ADDRESS)ldr));

    LDRDBG(("Mem=%x Discardable=%x Size=%x Prelink=%x\n",
            ldr->Mem, ldr->Discardable,
            _bs32(ldr->PeHeader->OptionalHeader.SizeOfImage),
            _bs32(ldr->PeHeader->OptionalHeader.ImageBase)));

    sc = ReadAndFixSections(ldr, pFile);
    if (FAILED(sc))
        goto Error;

    /* Done with file */
    pFile->v->Release(pFile);
    pFile = NULL;

    FixupHeader(ldr);

    sc = ResolveImports(ldr);
    if (FAILED(sc))
        goto Error;

    sc = ApplyFixups(ldr);
    if (FAILED(sc))
        goto Error;

    EntryPoint = (MODULEENTRY) GetEntryPoint(ldr);

    ExportData = NULL;
    ExportSize = 0;
    sc = PrepareExports(ldr, &ExportData, &ExportSize);
    if (FAILED(sc))
      goto Error;

#if defined(__NEED_FLUSH_CACHE)
    FlushCache((ADDRESS)ldr->Mem,ActualSize);
#endif

#ifdef DEBUGGER
    /* We need to propagate the actual allocation including prepended junk */
    ldr->Mem = (ADDRESS) ldr->HeaderFluffForDebugger;
    ldr->HeaderFluffForDebugger = 0;

#endif

     /* Create a descriptor for this module we just loaded. */
    sc = ModuleCreate(pImage, pArgs,
                      (PTR) ldr->Mem, ActualSize,
                      EntryPoint, ExportData, ExportSize,
                      ldr->ImportedModules,
                      ldr->nImportedModules,
                      ldr->MetaData,
                      ldr->EntryPointToken,
                      Flags,
                      &pMod);
    if (FAILED(sc))
        goto Error;

    ldr->nImportedModules = 0;  /* Don't drop refs */

    LDROUT(("Image %s is x%x bytes at x%x. EntryPoint=%x\n",
            pImage, ActualSize, ldr->Mem, (ADDRESS)EntryPoint));

    /* Done with ldr, but keep non-discardable memory around */
    ldr->Mem = 0;
    ReleaseLdr(ldr);
    ldr = NULL;

    /* Hand caller one ref iff ppIModule is valid pointer. */
    if (ppIModule) {
        pMod->v->AddRef(pMod);
        *ppIModule = pMod;
    }

    if (EntryPoint == 0) {
        LDROUT(("%s: entry point not called (found to be zero)\n", pImage));
    } else if (Flags & LOAD_IMAGE_CALL_ENTRY_POINT) {
        if (Flags & LOAD_IMAGE_CREATE_THREAD) {
            /* Find an IProcess descriptor for the module. */
            pMod->v->QueryInterface(pMod, &IID_IProcess, (void**) &pPrc);

            /* Create a new thread to invoke image's entry point.
             */
            sc = pPrc->v->CreateThread(pPrc, Flags,
                                       (THREAD_FUNCTION) EntryPoint,
                                       pTheBaseRtl,
                                       0,/*default stacksize*/
                                       NULL, NULL);

            /* Release all our references (new thread still holds one).
             */
            pPrc->v->Release(pPrc);
        } else {
            /* DLL: Call the entry point directly.
             */
            LDROUT(("Calling entry point at %x, args = %x\n",
                    (ADDRESS)EntryPoint, (ADDRESS)pArgs));
            EntryPoint(pTheBaseRtl);
            LDROUT(("User returned\n"));
        }
    }
    pMod->v->Release(pMod);

    return NO_ERROR;

Error:
    if (ldr)
      ReleaseLdr(ldr);
    if (pFile)
        pFile->v->Release(pFile);
    return sc;
}
