/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * MMLite loader for SE image format.
 */

/* Always included by someone else, who picks the Sw/HwLdrLoadImage functions
 */
#if !defined(SwLdrLoadImage) || !defined(HwLdrLoadImage)
#error Sw/HwLdrLoadImage not all defined
#endif

/* Definitions for our security header
 */
typedef struct _SE_HEADER {
    UINT32 SwImageOffset;
    UINT32 SwImageSize;
    UINT32 SwProperties; /* nothing yet */
    UINT32 HwImageOffset;
    UINT32 HwImageSize;
    UINT32 HwProperties;
#define SE_HW_PRIVILEGED        0x00000001
#define SE_HW_PERIPHERAL        0x00000002
#define SE_HW_PRIORITY_MASK     0x0000000c
#define SE_HW_PRIORITY_1        0x00000004
#define SE_HW_PRIORITY_2        0x00000008
#define SE_HW_HIGH_PRIORITY     0x0000000c
    UINT32 DigestOffset;
    UINT32 DigestSize;
    UINT32 DigestProperties; /* nothing yet */
    UINT32 MySize;
    UINT32 MagicNumber;
#define SE_MAGIC_BE  0x53453031  /* "SE01" */
#define SE_MAGIC_LE  0x31304553
} SE_HEADER, *PSE_HEADER;

#define SECTION_ROUNDUP 32

/* Utility
 */
PRIVATE
BOOL
ReadIt(FILE_HANDLE f,
       void *Data,
       UINT32 nBytes)
{
    INT i;

    for (;nBytes > 0;) {
        i = _Read(f,Data,nBytes);
        if ((i > 0) && ((UINT32)i <= nBytes)) {
            Data = (char *)Data + i;
            nBytes -= i;
            continue;
        }
        DBGME(3,_tprintf(_T("Ace::ReadIt? (%d %d)\n"), i, nBytes));
        return FALSE;
    }
    return TRUE;
}

PRIVATE
UINT32 
Checksum32(UINT32 *Start, UINT32 nWords, UINT32 Seed)
{
    while (nWords-- > 0)
        Seed += *Start++;
    return Seed;
}

PRIVATE
BOOL
SeReadAndSum(FILE_HANDLE f,
             UINT32 StartOffset, UINT32 AmountInBytes,
             UINT32 *Buffer, UINT32 BufferSize,
             UINT32 *Sum)
{
    INT i;

    DBGME(0,printf("SeReadAndSum(%x,%x,%x,%x,%x,%x ",
                   (ADDRESS)f,StartOffset,AmountInBytes,(ADDRESS)Buffer,BufferSize,*Sum));

    /* Start from the beginning
     */
    _Seek(f, StartOffset);

    /* One buffer at a time
     */
    while (AmountInBytes > 0) {

        if (BufferSize > AmountInBytes)
            BufferSize = AmountInBytes;

        i = _Read( f, (BYTE*)Buffer, BufferSize);
        if (i != (INT)BufferSize) {
            DBGME(3,printf("SeReadAndSum: short read %d\n", i));
            return FALSE;
        }

        AmountInBytes -= BufferSize;

        *Sum = Checksum32( Buffer, BufferSize / 4, *Sum);

    }
    DBGME(0,printf(")\n"));
    return TRUE;
}
/*
 * Load an SE-formatted program image into memory.
 */
SCODE LdrLoadImage(PINAMESPACE pNameSpace,
                const _TCHAR * pImage,
                const _TCHAR * pArgs,
                UINT Flags,
                IModule **ppIModule)
{
    SCODE sc = E_INVALID_PARAMETER; /* most likely */
    FILE_HANDLE f;
    UINT32 FileSize, BufferSize = 0, i;
    SE_HEADER SeHeader, SeHeaderCopy;
    BOOL MustSwap;
    UINT32 Sum32, *Buffer = NULL;

    /* Check that the file exists, see how big it is
     */
    f = _Open( pImage, 0x1 );
    if (f == INVALID_FILE_HANDLE) {
        DBGME(3,printf("File %s not found\n",pImage));
        return E_FILE_NOT_FOUND;
    }

    FileSize = _Stat( f );
    if (FileSize < sizeof(SE_HEADER)) {
        DBGME(3,printf("File too small %d\n",FileSize));
        goto Exit;
    }

    /* Get the header
     */
    _Seek(f, FileSize - sizeof(SE_HEADER));
    i = _Read( f, (BYTE*)&SeHeader, sizeof(SE_HEADER));
    if (i != sizeof(SE_HEADER)) {
        DBGME(3,printf("Cant read File header %d\n",i));
        goto Exit;
    }

    /* Is this a valid SE file? If not just hand it over to sw
     */
#if (BYTE_ORDER==BIG_ENDIAN)
#define MYSEMAGIC SE_MAGIC_BE
#else
#define MYSEMAGIC SE_MAGIC_LE
#endif
    if ((SeHeader.MagicNumber == SE_MAGIC_BE) ||
        (SeHeader.MagicNumber == SE_MAGIC_LE)) {
        MustSwap = SeHeader.MagicNumber != MYSEMAGIC;
    } else {
        DBGME(3,printf("SeHeader: Bad MagicNumber x%x\n", SeHeader.MagicNumber));
        goto LoadSoftwareImage;
    }

    /* Byteswap if necessary
     */
    if (MustSwap) {
        SeHeaderCopy.SwImageOffset = SwapWord(SeHeader.SwImageOffset);
        SeHeaderCopy.SwImageSize = SwapWord(SeHeader.SwImageSize);
        SeHeaderCopy.SwProperties = SwapWord(SeHeader.SwProperties);
        SeHeaderCopy.HwImageOffset = SwapWord(SeHeader.HwImageOffset);
        SeHeaderCopy.HwImageSize = SwapWord(SeHeader.HwImageSize);
        SeHeaderCopy.HwProperties = SwapWord(SeHeader.HwProperties);
        SeHeaderCopy.DigestOffset = SwapWord(SeHeader.DigestOffset);
        SeHeaderCopy.DigestSize = SwapWord(SeHeader.DigestSize);
        SeHeaderCopy.DigestProperties = SwapWord(SeHeader.DigestProperties);
        SeHeaderCopy.MySize = SwapWord(SeHeader.MySize);
        SeHeaderCopy.MagicNumber = SwapWord(SeHeader.MagicNumber);
    } else {
        SeHeaderCopy = SeHeader;
    }

    /* Sanity check the header, drop empty files
     */
    if ((SeHeaderCopy.MySize != sizeof SeHeader) ||
        (SeHeaderCopy.SwImageOffset > FileSize) ||
        (SeHeaderCopy.SwImageSize   > FileSize) ||
        ((SeHeaderCopy.SwImageOffset + SeHeaderCopy.SwImageSize) > FileSize) ||
        (SeHeaderCopy.HwImageOffset > FileSize) ||
        (SeHeaderCopy.HwImageSize   > FileSize) ||
        ((SeHeaderCopy.HwImageOffset + SeHeaderCopy.HwImageSize) > FileSize) ||
        (SeHeaderCopy.DigestOffset > FileSize) ||
        (SeHeaderCopy.DigestSize   > FileSize) ||
        ((SeHeaderCopy.DigestOffset + SeHeaderCopy.DigestSize) > FileSize) ||
        (SeHeaderCopy.DigestSize == 0) || /* required */
        ((SeHeaderCopy.SwImageSize | SeHeaderCopy.HwImageSize) == 0) /* empty */
        )
    {
        DBGME(3,printf("SeHeader: Bad content.\n"));
        goto Exit;
    }

    /* Verify the file's integrity
     * BUGBUG Fix for real digests
     */
    Sum32 = 0;

    /* Checksum the sw side, if present
     */
    if (SeHeaderCopy.SwImageSize) {
#define DGCHUNK 512 /* Bytes! */
        Buffer = (UINT32*) malloc(DGCHUNK);
        if (Buffer == NULL) {
            DBGME(3,printf("SeLdr: outtamem(sw)!\n"));
            goto Exit;
        }

        if (! SeReadAndSum(f,
                           SeHeaderCopy.SwImageOffset, SeHeaderCopy.SwImageSize, 
                           Buffer,DGCHUNK,&Sum32)) {
            DBGME(3,printf("SeReadAndSum(sw) failed.\n"));
            goto Exit;
        }
    
        /* The hw part will be bigger, not worth trying to reuse buffer
         */
        free(Buffer); Buffer = NULL;
    }

    /* Read in and checksum the hw side, if present
     * NB: Section roundup to benefit the ACE loader
     */
    if (SeHeaderCopy.HwImageSize) {
        BufferSize = (SeHeaderCopy.HwImageSize + SECTION_ROUNDUP - 1) & ~(SECTION_ROUNDUP - 1);
        Buffer = (UINT32*) malloc(BufferSize);
        if (Buffer == NULL) {
            DBGME(3,printf("SeLdr: outtamem(hw) %d!\n", BufferSize));
            goto Exit;
        }

        _Seek(f,SeHeaderCopy.HwImageOffset);
        if (!ReadIt(f,Buffer,SeHeaderCopy.HwImageSize)) {
            DBGME(3,_tprintf(_T("SeLdr: short read.\n")));
            goto Exit;
        }

        /* Zero the excess
         */
        memset(((char *)Buffer + SeHeaderCopy.HwImageSize),
               0,
               BufferSize - SeHeaderCopy.HwImageSize);

        /* Sum now
         */
        Sum32 = Checksum32( Buffer, SeHeaderCopy.HwImageSize / 4, Sum32);
    }

    /* Header too */
    Sum32 = Checksum32( (UINT32*)&SeHeader, sizeof(SeHeader) / 4, Sum32);

    /* Read digest in, check it
     */
    {
        UINT32 Digest;
        _Seek(f, SeHeaderCopy.DigestOffset);
        i = _Read( f, (BYTE*)&Digest, 4);
        if (i != 4) {
            DBGME(3,printf("Cant read File digest %d\n",i));
            goto Exit;
        }

        /* Check it now
         */
        if (0 != memcmp(&Digest,&Sum32,4)) {
            DBGME(3,printf("File %s: wrong signature %x (%x).\n", pImage,Sum32,Digest));
            goto Exit;
        }
    }

    /* Load the FPGA extension, if there is one
     */
    if (SeHeaderCopy.HwImageSize != 0) {
        sc = HwLdrLoadImage(Buffer,BufferSize,SeHeaderCopy.HwProperties);
        if (FAILED(sc)) {
            DBGME(3,printf("HwLdrLoadImage failed, sc=x%x\n", sc));
            goto Exit;
        }
        /* Notify machdep code of it, if not already
         */
        Flags |= LOAD_IMAGE_ENABLE_EXTENSION0;
        if (SeHeaderCopy.HwProperties)
            Flags |= SeHeaderCopy.HwProperties << 28;/*hackattack*/
    }

    /* Done with buffer */
    free(Buffer); Buffer = NULL;

    /* Done with the HW side.
     */
 LoadSoftwareImage:
    _Close(f); f = INVALID_FILE_HANDLE;

    /* Load the software program now, if there is one
     */
    if (SeHeader/*Copy*/.SwImageSize != 0)
        sc = SwLdrLoadImage(pNameSpace, pImage, pArgs, Flags, ppIModule);
    return sc;

Exit:
    /* Close files
     */
    if (Buffer) free(Buffer);
    if (f != INVALID_FILE_HANDLE)
        _Close(f);
    DBGME(2,printf("se_ldr loaded(%s) sc=%x\n", pImage, sc));
    return sc;
}
