/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmmacros.h>
#include <sock.h>
#include <diagnostics.h>
#include <assert.h>
#include <string.h>

#include <if.h>
#include <if_dl.h>
#include <dhcp.h>
#include <dhcp_fsm.h>
#include <ip/in.h>
#include <dhcp/dhcp_var.h>
#include <in_var.h>

/*
 * Handle incoming DHCP packets.
 */
void DhcpInput(struct mbuf *m){
    struct Dhcp *Dhcp;
    struct DhcpInfo NewIn, *IfInfo;
    struct ifnet *pNetInterface;
    UINT8 MsgType;
    INT PktLen;

    PktLen = m->m_pkthdr.len;
    pNetInterface = m->m_pkthdr.rcvif;

    /*
     * IfInfo points to DHCP information kept on this interface.
     * NewIn contains DHCP information extracted from this packet
     * that we are not committed to keeping yet.
     */
    IfInfo = pNetInterface->pDhcpInfo;
    assert(IfInfo != NULL); /* make sure DhcpClearStored info was called */
    memset(&NewIn, 0, sizeof(struct DhcpInfo));

    /*
     * Make sure packet is big enough to hold a minimal DHCP header.
     */
    if (PktLen < (INT) sizeof(struct Dhcp)) {
            /*
             * Too small: bogus packet.
             */
            m_freem(m);
            return;
    }

    /*
     * Ugly hack: the DHCP header is too huge for m_pullup.
     * Rather than fix m_pullup (although it is broken), or be cautious
     * about stepping through mbuf chains as we parse the packet, we
     * instead elect to alloc a chunk of memory and plop the packet in it.
     */
    Dhcp = bsd_malloc(PktLen, 0, M_NOWAIT);
    if (Dhcp == NULL) {
        DBGME(3,printf(_T("DhcpInput: out of memory?!?\n")));
        m_freem(m);
        return;
    }
    m_copydata(m, 0, PktLen, (PBYTE) Dhcp);
    m_freem(m);
    m = NULL;

    /*
     * Parse DHCP header.
     */

    if (Dhcp->Op != DHCP_OP_REPLY) {
        goto ErrorExit;
    }

    if (Dhcp->HdwrAddrType != IfInfo->HdwrAddrType ||
          Dhcp->HdwrAddrLen != IfInfo->HdwrAddrLen) {
        goto ErrorExit;
    }
    if (memcmp(Dhcp->ClntHdwrAddr, IfInfo->HdwrAddr,
               IfInfo->HdwrAddrLen) != 0) {
        goto ErrorExit;
    }

    /*
     * Verify that this reply is in response to our request by
     * checking that the transaction ids match.
     */
    if (Dhcp->XId != IfInfo->XIdInUse) {
        goto ErrorExit;
    }

    /*
     * Pull out boot (TFTP) server IP address and/or boot image path
     * if we were given them.  Note that File is a null-terminated
     * ASCII string.
     */
    if (Dhcp->SrvrIpAddr != 0) {
        NewIn.BootSrvrIpAddr = Dhcp->SrvrIpAddr;
    }
    if (*Dhcp->File != (char) 0) {
        INT Len;

        Len = strlen(Dhcp->File);
        NewIn.pBootFilename = bsd_malloc(Len, 0, M_NOWAIT);
        if (NewIn.pBootFilename == NULL)
            panic(_T("DhcpParseOptions: out of memory?!?\n"));
        memcpy(NewIn.pBootFilename, Dhcp->File, Len);
    }

    /*
     * Parse options.
     */
    MsgType = DhcpParseOptions((UINT8 *)Dhcp + sizeof(struct Dhcp),
                               PktLen - sizeof(struct Dhcp), &NewIn);
    if (MsgType == 0)
        goto ErrorExit;

    /*
     * What we do next depends upon what we got
     * and what state we were in when we got it.
     * Note: we ignore all packets not explicitly checked for.
     */
    switch(MsgType) {

    case DHCP_TYPE_ACK:
        switch(IfInfo->State) {

        case DHCPS_REBOOTING:
        case DHCPS_REQUESTING:
        case DHCPS_RENEWING:
        case DHCPS_REBINDING:
#if 1
            /* XXX
             * Check if the current IP is auto IP. If so, release it.
             */
            {
                struct in_aliasreq IfReq;
        
                in_control(NULL, SIOCGIFADDR, &IfReq, pNetInterface);
                if ((IfReq.ifra_addr.sin_addr.s_addr & 0xffff) == 0xfea9) {
                    in_control(NULL, SIOCDIFADDR, &IfReq, pNetInterface);
                    pNetInterface->if_flags |= IFF_NOADDR;
                }
            }
#endif
            /* We have a lease. Cancel auto IP recheck timeout */
            IfInfo->AutoIPRecheckTimer = 0;
            /*
             * We have a lease. Cancel outstanding retransmission
             * timeout and reset the backoff value.
             */
            IfInfo->RexmitTimer = 0;
            IfInfo->RexmitInterval = DHCP_REXMIT_DEFAULT
                + (tv_sec(CurrentTime()) & 0x3);
            /*
             * Save away relevant information from input packet, reparse
             * the options for keeps (could copy over from NewIn instead,
             * but this saves code space).  Apply pertinent info,
             * record the lease, and start timers.
             */
            IfInfo->BootSrvrIpAddr = NewIn.BootSrvrIpAddr;
            if (IfInfo->pBootFilename)
                bsd_free(IfInfo->pBootFilename, 0);
            IfInfo->pBootFilename = NewIn.pBootFilename;
            NewIn.pBootFilename = NULL;
            (void)DhcpParseOptions((UINT8 *)Dhcp + sizeof(struct Dhcp),
                                   PktLen - sizeof(struct Dhcp), IfInfo);
            IfInfo->LeasedIpAddr = IfInfo->OfferedIpAddr;
            IfInfo->OfferedIpAddr = 0;
            DhcpRecordLease(pNetInterface);
            /*
             * Synchronize with DhcpWaitTillBound.
             */
            IfInfo->State = DHCPS_BOUND;
            WakeUp((PTR)&IfInfo->State);
        }
        break;

    case DHCP_TYPE_NAK:
        switch(IfInfo->State) {
            UINT8 Tmp;

        case DHCPS_RENEWING:
        case DHCPS_REBINDING:
            /*
             * We lost our lease (everything must go!).
             */
            if_down(pNetInterface);
            /* fall into ... */

        case DHCPS_REBOOTING:
        case DHCPS_REQUESTING:
            /*
             * Back to square one.
             */
            Tmp = IfInfo->Flags;
            DhcpInitStoredInfo(pNetInterface);
            DhcpStart(pNetInterface, Tmp);
        }
        break;

    case DHCP_TYPE_OFFER:
        /* We could be selective here in case we get multiple replies */
        if (IfInfo->State == DHCPS_SELECTING) {
                /*
                 * Take it.  Cancel outstanding retransmission
                 * timeout and reset the backoff value.
                 */
                IfInfo->RexmitTimer = 0;
                IfInfo->RexmitInterval = DHCP_REXMIT_DEFAULT
                    + (tv_sec(CurrentTime()) & 0x3);
                IfInfo->OfferedIpAddr = Dhcp->YourIpAddr;
                IfInfo->DhcpSrvrIpAddr = NewIn.DhcpSrvrIpAddr;

                IfInfo->State = DHCPS_REQUESTING;
                DhcpOutput(pNetInterface);  /* send DHCPREQUEST */
        }
    }

ErrorExit:
    /*
     * Release the packet buffer and incoming data areas we didn't want.
     */
    bsd_free(Dhcp, 0);
    DhcpFreeInfoAreas(&NewIn);
}

#include <arp.h>
extern void DhcpAutoIPSendProbe(struct ifnet *pNetInterface);

void DhcpAutoIPInput(struct ifnet *pNetInterface, struct ether_arp *ea)
{
    struct DhcpInfo *Info;
    int op = ntohs(ea->arp_op);

    Info = pNetInterface->pDhcpInfo;
    
    if (Info == NULL)
        return;

    /* If it is neither ARP request nor reply or tried 10 times, then return */
    if ((op != ARPOP_REQUEST && op != ARPOP_REPLY) || Info->AutoIpRetryCount >= 10)
        return;

    /* If it is our own request forget about it */
    if (op == ARPOP_REQUEST && !memcmp(Info->HdwrAddr, ea->arp_sha, sizeof(*ea->arp_sha)))
        return;

    /* If reply with different IP just drop it */
    if (op == ARPOP_REPLY && Info->AutoIpCandidate != *(UINT32*)ea->arp_spa)
        return;

    /* If probe for different IP, drop it */
    if (op == ARPOP_REQUEST /*&& *(UINT32*)ea->arp_spa != 0*/ && Info->AutoIpCandidate != *(UINT32*)ea->arp_tpa)
        return;

    /* If another host with different IP send a request, just drop it */
    if (op == ARPOP_REQUEST && Info->AutoIpCandidate != *(UINT32*)ea->arp_spa)
        return;

    /* Otherwise trouble, must pick new address */
    DBGME(3,printf(_T("This auto IP is in use by another host. Reselect one!\n")));
    Info->AutoIpRetryCount++;
    Info->AutoIpCandidate = 0;
    DhcpAutoIPSendProbe(pNetInterface);
}
