/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <sock.h>
#include <in_systm.h>
#include <if.h>
#include <route.h>
#include <in.h>
#include <ip.h>
#include <in_pcb.h>
#include <ip_var.h>
#include <udp.h>
#include <udp_var.h>
#include <dhcp.h>
#include <dhcp_fsm.h>
#include <dhcp/dhcp_var.h>
#include <arp.h>

/*
 * Send a DHCP packet.
 *
 * What to send is determined by our per-interface state.  Note that we
 * switch to the new state BEFORE sending the packet corresponding to
 * that state transition.  We can do this since only one type of packet
 * is sent in the process of reaching any particular state.  This makes
 * handling retransmission simpler as it is the same as the normal case.
 */
void DhcpOutput(struct ifnet *pNetInterface){
    struct mbuf *m;
    struct Dhcp *Dhcp;
    struct DhcpInfo *Info = pNetInterface->pDhcpInfo;
    BOOL fBroadcast = TRUE;
    UINT8 MsgType;
    UINT8 *Opts;
    INT OptLen = 0;

#if 0
    /* Do this every time to avoid serplexd race */
    Info->HdwrAddrLen = pNetInterface->if_addrlen;
    memcpy(Info->HdwrAddr, ((struct arpcom *)pNetInterface)->ac_enaddr,
           Info->HdwrAddrLen);
#endif

    /*
     * Grab an mbuf cluster for this packet.
     * BUGBUG: This code assumes entire DHCP packet is less than MCLBYTES.
     */
    MGETHDR(m, M_DONTWAIT, MT_DATA);
    if (m == NULL)
        return;
    MCLGET(m, M_DONTWAIT);
    if ((m->m_flags & M_EXT) == 0){
        m_free(m);
        return;
    }
    m->m_data += max_linkhdr + sizeof(struct udpiphdr);  /* effective */
    m->m_pkthdr.rcvif = (struct ifnet *)0;

    /*
     * Fill in fixed header fields (most are zero).
     */
    Dhcp = mtod(m, struct Dhcp *);
    memset(Dhcp, 0, sizeof(struct Dhcp));  /* zero it all first */
    Dhcp->Op = DHCP_OP_REQUEST;
    Dhcp->HdwrAddrType = Info->HdwrAddrType;
    Dhcp->HdwrAddrLen = Info->HdwrAddrLen;
    memcpy(Dhcp->ClntHdwrAddr, Info->HdwrAddr, Info->HdwrAddrLen);

    /*
     * Pick a new Xid (when applicable).
     * According to the spec, we should pick a new Xid when we send
     * a DHCPDISCOVER, DHCPINFORM, or initial DHCPREQUEST (the latter
     * applies ONLY when booting with a known network address).  The spec
     * leaves it up to the implementation whether or not to pick a new Xid
     * for each retransmission if retranmission is required.
     * Note this has an effect upon time calculations re TimeStarted.
     *
     * This Xid must be returned by the server in any DHCPOFFER (responding
     * to a DHCPDISCOVER) or DHCPACK/DHCPNAK (responding to a DHCPINFORM or
     * initial DHCPREQUEST).  Similarly, we must send the same Xid that was
     * in the DHCPOFFER back to the server in any subsequent DHCPREQUEST.
     * (presumably, this applies to DHCPREQUEST's sent when renewing or
     * rebinding as well)
     *
     * It appears to be up to us whether or not to pick a new Xid on
     * DHCPDECLINE or DHCPRELEASE.  We choose to stick with the same one.
     */
    if (!Info->XIdInUse)  /* BUGBUG? Or check State == SELECTING/REBOOTING ? */
        Info->XIdInUse = (UINT32) Int64ToInt32(CurrentTime());
    Dhcp->XId = Info->XIdInUse;

    /*
     * If we're in a state where we have a valid lease on an address,
     * use that address in the header.
     */
    if (Info->LeasedIpAddr)
        Dhcp->ClntIpAddr = Info->LeasedIpAddr;

    /*
     * Figure out what we need to send and how we need to send it
     * by looking at our current state.
     */
    switch(Info->State) {
    case DHCPS_SELECTING:
        MsgType = DHCP_TYPE_DISCOVER;
        break;

    case DHCPS_RENEWING:
        fBroadcast = FALSE;
        /*  fall into ... */
    case DHCPS_REBOOTING:
    case DHCPS_REQUESTING:
    case DHCPS_REBINDING:
        MsgType = DHCP_TYPE_REQUEST;
        break;

    case DHCPS_INIT:
        MsgType = DHCP_TYPE_DECLINE;
        Info->RexmitInterval = 0;
        break;

    case DHCPS_CLOSED:
        MsgType = DHCP_TYPE_RELEASE;
        Info->RexmitInterval = 0;
        break;

    default:
        MsgType = 0;
        panic(_T("DhcpOutput: in bogus state\n"));
    }

    /*
     * Append various options and set packet length.
     * Options to think about:
     * Class Identifier -- SHOULD -- we'd have to make something up
     */
    Opts = (UINT8 *)Dhcp + sizeof(struct Dhcp);

    /* Magic Cookie */
    *Opts++ = 99;*Opts++ = 130;*Opts++ = 83;*Opts++ = 99;OptLen += 4;

    /* Message Type */
    *Opts++ = DHCP_OPTION_MESSAGE_TYPE;
    *Opts++ = 1;*Opts++ = MsgType;OptLen += 3;

    /*
     * Client Identifier: for now use the network media address.
     * BUGBUG?  Think about switching to STB serial number.
     */
    *Opts++ = DHCP_OPTION_CLIENT_ID;*Opts++ = (UINT8)(Info->HdwrAddrLen + 1);
    *Opts++ = Info->HdwrAddrType;
    memcpy(Opts, Info->HdwrAddr, Info->HdwrAddrLen);Opts += Info->HdwrAddrLen;
    OptLen += Info->HdwrAddrLen + 3;

    /*
     * Server Identifier: only sent when picking a particular server's
     * offer and when declining or releasing a lease.  NOT sent when
     * rebooting with a previously allocated address, or when renewing
     * or rebinding.
     */
    if (Info->State == DHCPS_REQUESTING || Info->State == DHCPS_CLOSED) {
        *Opts++ = DHCP_OPTION_SERVER_IDENTIFIER;
        *Opts++ = 4;memcpy(Opts, &Info->DhcpSrvrIpAddr, 4);
        Opts += 4;OptLen += 6;
    }

    /*
     * Requested IP Address: only sent when picking a particular offered
     * address, declining a lease, or rebooting with a previously allocated
     * address.  NOT sent when renewing, rebinding or releasing a lease.
     */
    if (Info->OfferedIpAddr){
        *Opts++ = DHCP_OPTION_REQUESTED_ADDRESS;
        *Opts++ = 4;memcpy(Opts, &Info->OfferedIpAddr, 4);
        Opts += 4;OptLen += 6;
    }

    /* Parameter List: ask for various options to be supplied */
    *Opts++ = DHCP_OPTION_PARAMETER_REQUEST_LIST;
    *Opts++ = 6;
    *Opts++ = DHCP_OPTION_ROUTER_ADDRESS;
    *Opts++ = DHCP_OPTION_SUBNET_MASK;
    *Opts++ = DHCP_OPTION_DOMAIN_NAME_SERVERS;
    *Opts++ = DHCP_OPTION_HOST_NAME;
    *Opts++ = DHCP_OPTION_TIME_SERVERS;
    *Opts++ = DHCP_OPTION_DOMAIN_NAME;
    OptLen += 2 + 6;

    /* End Options */
    *Opts++ = DHCP_OPTION_END;OptLen += 1;

    /*
     * BOOTP compat: The BOOTP spec has a 64 byte "vendor information" field
     * following the boot file name field.  For DHCP, this has been replaced
     * with a variable length options field.  So while a valid DHCP packet
     * can be less than 300 bytes (in UDP payload), some BOOTP relay agents
     * require a packet be at least 300 bytes long in order to be forwarded.
     * Therefore we pad out our packets to be at least 300 bytes.
     */
    while(sizeof(struct Dhcp) + OptLen < 300){
        *Opts++ = DHCP_OPTION_PAD;
        OptLen++;
    }

    m->m_len = sizeof(struct Dhcp) + OptLen;
    m->m_pkthdr.len = m->m_len;

    /*
     * Set local (source) and foreign (destination) addresses.
     */
    if (Info->LeasedIpAddr)
        Info->pInPcb->inp_laddr.s_addr = Info->LeasedIpAddr;
    else
        Info->pInPcb->inp_laddr.s_addr = 0;

    if (fBroadcast)
        Info->pInPcb->inp_faddr.s_addr = htonl(INADDR_BROADCAST);
    else
        Info->pInPcb->inp_faddr.s_addr = Info->DhcpSrvrIpAddr;

    /*
     * Patch our phoney route to point to our interface and destination.
     */
    Info->pInPcb->inp_route.rt_valid = 1;
    Info->pInPcb->inp_route.rt_ifa = pNetInterface->if_addrlist; /* XXX */
    Info->pInPcb->inp_route.rt_gw.sin_addr = Info->pInPcb->inp_faddr;

    /*
     * Send the packet.
     */
    udp_output(Info->pInPcb, m, 0, 0);

    /*
     * Start the retransmission timer.
     */
    Info->RexmitTimer = Info->RexmitInterval;
}
