/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmmacros.h>
#include <stdlib.h>
#include <sock.h>
#include <string.h>
#include <stdio.h>
#include <diagnostics.h>
#include <assert.h>

#include <in_systm.h>
#include <if.h>
#include <route.h>
#include <in.h>
#include <in_var.h>
#include <ip.h>
#include <in_pcb.h>
#include <ip_var.h>
#include <arp.h>
#include <dhcp.h>
#include <dhcp_fsm.h>
#include <if_types.h>
#include <dhcp/dhcp_var.h>
#include <dns/nameser.h>
#include <dns/inet.h>

void DhcpJitterInit(UINT8 *pHdwrAddr);
void DhcpJitterLargeDelay(void);
void DhcpJitterSmallDelay(void);
static void FillSin(struct sockaddr_in *sin, UINT32 Addr, UINT16 Port);
PRIVATE void DhcpSetDefaultHostName(struct DhcpInfo *Info);
PRIVATE void DhcpSetMulticastDNSInfo(struct DhcpInfo *Info);

extern void arpwhohas(struct arpcom *ac, struct in_addr *addr);

/*
 * This protocol control block and routing table entry are set up to
 * allow us to pass DHCP packets through the existing UDP/IP stack with
 * as little special-casing of the existing code as possible.
 */
struct inpcb DhcpInPcb;

UINT DhcpJitterSeed;
UINT DhcpJitterRandom(void);

/*
 * Randomly select a auto IP address within a given subnet
 * (the first and last 256 are reserved for future use). 
 * Then issue an ARP request for the suggested address to check if this address
 * is already in use or not. While waiting for a possible response for this ARP
 * request, the client must also listen for other ARP probes for the same address
 * (but not from its own hardware address). If the client receives a response to 
 * the ARP request, or sees another ARP probe for the same address, it consider 
 * the address as being in use, and try another address.
 * See "draft-ietf-dhc-ipv4-autoconfig-04.txt".
 */
PRIVATE UINT32 ChooseAutoIP(UINT32 val)
{
    UINT32  r;
    BYTE    b;

    if (val == INADDR_ANY || val == 0)
        val = htonl( (169 << 24) | (254 << 16) ); // subnet "169.254/16",
    else if (val & htonl(0xffff))
        goto Done; // Fixed IP
    val &= htonl(0xffff0000);
    /* The first and last 256 are reserved for future use,
     * and should not be a network or broadcast IP */
    do {
        r = (UINT32)rand();
        b = (BYTE)(r >> 8);
    } while (b == 0 || b == 0xff || r == 0 || r == 0xffff);

    val |= htonl( (r & 0xff) | (b << 8) );
    //r = htonl(0xa9fe9e70); /* 169.254.158.112 fixed for testing */
 Done:
    return val;
}

/*
 */
extern PTR arplookup(u_long addr, int create, int proxy);
BOOL isInArpTable(UINT32 ip)
{
#if 1
    PTR la;
    
    /* Lookup autoip in the ARP table first, do not allocate */
    /* BUGBUG methinks this should be arpwhohas() ? --af */
    la = arplookup(ip, 0, 0);
    if (la != NULL) 
        return TRUE; /* Got it in the ARP cache, so reselect one */
#endif
    return FALSE;
}

/*
 * IPv4 Auto-configuration
 */
PRIVATE void SetAutoIP(struct ifnet *pNetInterface, UINT32 autoip) 
{
    struct DhcpInfo *Info = pNetInterface->pDhcpInfo;
    struct in_aliasreq IfReq;

#if 1
    /* XXX
     * Release the current IP.
     */
    in_control(NULL, SIOCGIFADDR, &IfReq, pNetInterface);
    in_control(NULL, SIOCDIFADDR, &IfReq, pNetInterface);
#endif

#if 1
    /*
     * Leases are relative to time of our DHCPDISCOVER or DHCPREQUEST message.
     * If not explicitly set by their seperate options, T1 and T2
     * default to .5 and .875 of the lease duration, respectively.
     *
     * Note that since our DHCP timer runs continously, these timers may
     * already have lost a few seconds.  This is o.k.
     */
    Info->RenewalTimer = 0;
    Info->LeaseTimer = 0;
    Info->RebindTimer = 0;
#endif

    /* Recheck every 5 minutes */
    Info->AutoIPRecheckTimer = 0; //XXX keep it alive forever
    /* Set the interface address, subnet mask and broadcast address. */
    pNetInterface->if_flags &= ~IFF_NOADDR;

    memset(&IfReq, 0, sizeof IfReq);
    IfReq.ifra_addr.sin_len = (UINT8) sizeof(struct sockaddr_in);
    IfReq.ifra_addr.sin_family = AF_INET;
    IfReq.ifra_addr.sin_addr.s_addr = autoip;

    /* Set the subnet mask */
    IfReq.ifra_mask.sin_len = (UINT8) sizeof(struct sockaddr_in);
    IfReq.ifra_mask.sin_family = AF_INET;
    IfReq.ifra_mask.sin_addr.s_addr = htonl(0xffff0000);

    /* Set the broadcast address */
    IfReq.ifra_broadaddr.sin_len = (UINT8) sizeof(struct sockaddr_in);
    IfReq.ifra_broadaddr.sin_family = AF_INET;
    IfReq.ifra_broadaddr.sin_addr.s_addr = htonl(0x0000ffff) | autoip;

    in_control(NULL /* dangerous */, SIOCAIFADDR, &IfReq, pNetInterface);
    
#if 1
    {
        UINT8 *p;
        p = (UINT8 *) &autoip;
        DBGME(2,printf(_T("Auto IP Addr %d.%d.%d.%d\n"),
               p[0], p[1], p[2], p[3]));
        p = (UINT8 *) &(IfReq.ifra_mask.sin_addr.s_addr);
        DBGME(2,printf(_T("Subnet Mask %d.%d.%d.%d\n"),
               p[0], p[1], p[2], p[3]));
        p = (UINT8 *) &(IfReq.ifra_broadaddr.sin_addr.s_addr);
        DBGME(2,printf(_T("Broadcast Addr %d.%d.%d.%d\n"),
               p[0], p[1], p[2], p[3]));
    }
#endif
}

/*
 * Various DHCP subroutines live here.
 */

/*
 * Initialize the DHCP subsystem.
 */
void DhcpInit(void)
{
    /*
     * Set up fields in our Internet Protocol Control Block
     * that are constant throughout our usage.
     */
    DhcpInPcb.inp_lport = htons(DHCP_CLIENT_PORT);
    DhcpInPcb.inp_fport = htons(DHCP_SERVER_PORT);
    DhcpInPcb.inp_options = 0;  /* no IP options */
    DhcpInPcb.inp_moptions = 0;  /* no multicast options */
}

/*
 * Start the DHCP process for a particular interface.
 */
void DhcpStart(struct ifnet *pNetInterface, UINT8 Flags)
{
    struct DhcpInfo *Info = pNetInterface->pDhcpInfo;

    if (Flags)
        Info->Flags |= Flags;

    Info->TimeStarted = CurrentTime();  /* leases relative to */

    switch (pNetInterface->if_type) {
    case IFT_ETHER:
    case IFT_ISO88023:
    case IFT_ISO88024:
    case IFT_ISO88025:
        Info->HdwrAddrType = DHCP_HARDWARE_TYPE_10MB_EITHERNET;
        break;
    case IFT_ATM:
        Info->HdwrAddrType = DHCP_HARDWARE_TYPE_ATM;
        break;
    default:
        Info->HdwrAddrType = 0; /* donno */
        break;
    }
    Info->HdwrAddrLen = pNetInterface->if_addrlen;
    assert(Info->HdwrAddrLen != 0);
    memcpy(Info->HdwrAddr, ((struct arpcom *)pNetInterface)->ac_enaddr,
           Info->HdwrAddrLen);

    pNetInterface->if_flags |= IFF_NOADDR;

    Info->State = DHCPS_SELECTING;

    DhcpJitterInit(Info->HdwrAddr);
    DhcpJitterLargeDelay();
    DhcpJitterSmallDelay();

    DhcpSetDefaultHostName(Info);  // added to get default host name from file on disk
    DhcpSetMulticastDNSInfo(Info); // add entry for querying multicast dns first
    DhcpOutput(pNetInterface);
}

/*
 * Initializes with default host name from file "fs\hostname.txt"
 */

PRIVATE void DhcpSetDefaultHostName(struct DhcpInfo *Info)
{
 
    PIFILE defaultHostNameFile;
    char hostName[MAXDNAME + 1];    // MAXDNAME is max size of domain name -> nameser.h
    char *ptr;
    UINT len = 0;

    SCODE sc = BindToObject(CurrentNameSpace(), _T("hostname.txt"), NAME_SPACE_READ, &IID_IFile, (void *) &defaultHostNameFile);

    if (FAILED(sc))
        sc = BindToObject(CurrentNameSpace(), _T("fs/hostname.txt"), NAME_SPACE_READ, &IID_IFile, (void *) &defaultHostNameFile);

    if (FAILED(sc)) {
        printf("dhcp_sub.c:Could not bind to file hostname.txt\n");
        return;
    }

    sc= ReadWriteFile(defaultHostNameFile, FALSE, UINT64_ZERO,
                      (PBYTE)hostName, MAXDNAME + 1, &len);
    hostName[len] = '\0';
    /* Check for EOF character so that hostname can be entered into serial flash manually Control-D or Z */
#define CZ '\032'
#define CD '\004'
    if ((ptr = strchr(hostName, CZ)) != NULL)
        *ptr = '\0';
    if ((ptr = strchr(hostName, CD)) != NULL)
        *ptr = '\0';
    len = strlen(hostName);
    defaultHostNameFile->v->Release(defaultHostNameFile);
    if (FAILED(sc)) {
        printf("dhcp_sub.c:Could not get default host name\n");
        return;
    }

    if (Info->HostName)
        bsd_free(Info->HostName, 0);
    Info->HostName = bsd_malloc(len + 1, M_DHCPHOST, M_NOWAIT);
    if (Info->HostName == NULL)
        panic(_T("DhcpSetDefaultHostName: out of memory?!? or couldn't read default host name\n"));

    memcpy(Info->HostName, hostName, len);
    Info->HostName[len] = 0;
    DBGME(3,printf(_T("Got default host name \"%s\" from file\n"), Info->HostName));
}

/*
 * Entry for querying multicast DNS first
 */

PRIVATE void DhcpSetMulticastDNSInfo(struct DhcpInfo *Info)
{
    struct in_addr inaddr;

    inet_aton(NAMESERVER_MULTICAST_ADDRESS, &inaddr);
    FillSin(&Info->nsaddr_list[0], inaddr.s_addr, NAMESERVER_PORT);
    if (Info->nscount < 1)
        Info->nscount = 1;
}

void DhcpAutoIPSendProbe(struct ifnet *pNetInterface)
{
    struct DhcpInfo *Info = pNetInterface->pDhcpInfo;
    struct in_addr addr;
    
retry:
    /* re-probe previous address if any */
    if (Info->AutoIpCandidate == 0)
        Info->AutoIpCandidate = ChooseAutoIP(((struct arpcom *)pNetInterface)->ac_ipaddr.s_addr);

    if (isInArpTable(Info->AutoIpCandidate)) {
        Info->AutoIpCandidate = 0;
        goto retry;
    }

    addr.s_addr = Info->AutoIpCandidate;
    //((struct arpcom *)pNetInterface)->ac_ipaddr.s_addr = 0;
    arpwhohas((struct arpcom *)pNetInterface, &addr);}

/*
 * Wait for the interface to be bound to an IP address.
 */
INT DhcpWaitTillBound(struct ifnet *pNetInterface)
{
    INT error = 0;
    TIME TimeOut = CurrentTime();
    struct DhcpInfo *Info = pNetInterface->pDhcpInfo;
#if 1
    UINT RetryCount = 0;
    struct in_aliasreq IfReq;

    /* Get the current IP */
    in_control(NULL, SIOCGIFADDR, &IfReq, pNetInterface);
    /* If the target IP is auto ip, then... */
    if ((ntohl(IfReq.ifra_addr.sin_addr.s_addr) & IN_CLASSB_NET) == 0xa9fe0000)
        assert(Info->AutoIpCandidate == IfReq.ifra_addr.sin_addr.s_addr);
    
    srand(Uint64ToUint32(TimeOut));
    Info->AutoIpRetryCount = 0;
    DhcpAutoIPSendProbe(pNetInterface);
#endif

    /* Try DHCP */
    /* We wait a short time and boot with AutoIP if necessary.
     * DHCP will keep retrying a few times afterwards. We assume 
     * AutoIP retrying is fast enough to be done 10 times within 
     * 2 seconds!
     */
    Int32ToInt64(TimeOut, TIME_RELATIVE(TIME_SECONDS(60)));
    while (Info->State != DHCPS_BOUND) {
        error = XXXWaitFor((PTR)&Info->State, "DhcpWaitTillBound", TimeOut);
        if (error != 0) { /* timeout */
#if 1
            if (RetryCount != Info->AutoIpRetryCount && Info->AutoIpRetryCount >= 10) {
                RetryCount = Info->AutoIpRetryCount;
                continue; /* wait again */
            }
#else
            DBGME(3,printf(_T("DHCP bind timed out sc=%x\n"), error));
#endif
            break;
        }
    }
#if 1
    if (error == 0)
        return 0; /* Got DHCP offer! */

    if (Info->AutoIpRetryCount >= 10) {
        DBGME(3,_tprintf(_T("Auto IP retry count (%d) exceeded. Start without IP!\n"), Info->AutoIpRetryCount));
        return 1;/* We tried ... Then start without IP */
    }
    /* Otherwise the candidate survived.
     * This assumes we had time to try 10 addresses in 28 seconds. The AutoIP
     * protocol is inherently unreliable: if any ARP probes get lost we
     * might get duplicate addresses. This 28 second assumption should be the
     * least of the worries :).
     */
    SetAutoIP(pNetInterface, Info->AutoIpCandidate);

    /* Otherwise we have already had an auto IP so don't do it. */
    return 0;
#else
    return error;
#endif
}

/*
 * Retrieve results of DHCP transactions.
 */
INT DhcpGetInfo(struct ifnet *pNetInterface, struct DhcpInfo *pInfo)
{
    struct DhcpInfo *Info = pNetInterface->pDhcpInfo;

    /*
     * Just return whatever we have.
     */
    *pInfo = *Info;
    return 0;
}

/*
 * Retrieve stored option values.
 */
INT DhcpGetOption(struct ifnet *pNetInterface, UINT8 Opt,
                  struct DhcpOption *pOpt)
{
//    struct DhcpInfo *pInfo = pNetInterface->pDhcpInfo;
//    PBYTE pStuff;

    /*
     * We only store certain site-specific options.
     * Complain if asked about others.
     */
    switch(Opt){
#if 0
    case DHCP_OPTION_MITV_CLASS_STORE:
        pStuff = pInfo->pCSOption;
        break;
    case DHCP_OPTION_MITV_DIRECTORY_SERVICE:
        pStuff = pInfo->pDSOption;
        break;
    case DHCP_OPTION_MITV_SUBSCRIBER_SERVICE:
        pStuff = pInfo->pISSOption;
        break;
    case DHCP_OPTION_MITV_POLICY_FILE:
        pStuff = pInfo->pPolicyOption;
        break;
    case DHCP_OPTION_MITV_VERSION:
        pStuff = pInfo->pVersionOption;
        break;
#endif
    default:
        return EINVAL;
    }

#if 0 /* unreachable... */
    if (pStuff == NULL)
        return EFAULT;  /* for lack of better error */

    pOpt->Option = Opt;
    pOpt->Len = *pStuff++;
    memcpy(pOpt->Data, pStuff, pOpt->Len);

    return 0;
#endif
}

/*
 * Retrieve boot filename.
 */
INT DhcpGetFile(struct ifnet *pNetInterface, char *pFile)
{
    struct DhcpInfo *pInfo = pNetInterface->pDhcpInfo;

    if (pInfo->pBootFilename == NULL)
        return EFAULT;  /* for lack of better error */

    strcpy(pFile, pInfo->pBootFilename);

    return 0;
}

static void FillSin(struct sockaddr_in *sin, UINT32 Addr, UINT16 Port)
{
    memset(sin, 0, sizeof(*sin));
    sin->sin_len = sizeof(struct sockaddr_in);
    sin->sin_family = AF_INET;
    sin->sin_port = htons(Port);
    sin->sin_addr.s_addr = Addr;
}

/*
 * Parse options.
 *
 * Return 0 if error, else return value of "Message Type" option.
 *
 * BUGBUG: need to verify server id option in states > 1 and < 6?
 */
UINT8 DhcpParseOptions(UINT8 *Opts, INT OptLen, struct DhcpInfo *Info)
{
    UINT8 *OptEnd;
    UINT8 MsgType;

    if (OptLen < 8) {
        /*
         * Minimal option field is 4 byte "magic cookie" + 3 byte
         * "DHCP Message Type" option + 1 byte "End" Option.
         */
        return 0;
    }
    OptEnd = Opts + OptLen;

    /*
     * Verify "magic cookie" exists at start of option field.
     */
    if (*Opts++ != 99)
        return 0;
    if (*Opts++ != 130)
        return 0;
    if (*Opts++ != 83)
        return 0;
    if (*Opts++ != 99)
        return 0;

    /*
     * BUGBUG: this doesn't handle bogus option lengths well.  Fix later.
     * With the exception of the "Message Type" option (which it insists upon
     * getting), this code assumes that server sends the options the RFC says
     * it is required to send.
     */
    MsgType = 0;
    while(Opts < OptEnd) {
        if (*Opts == DHCP_OPTION_PAD) {
            Opts++;
            continue;
        }
        if (*Opts == DHCP_OPTION_END)
            break;
        if (*Opts == DHCP_OPTION_MESSAGE_TYPE) {
            Opts += 2;
            MsgType = *Opts++;
            continue;
        }
        if (*Opts == DHCP_OPTION_SERVER_IDENTIFIER) {
            Opts += 2;
            memcpy(&Info->DhcpSrvrIpAddr, Opts, 4);
            Opts += 4;
            continue;
        }
        if (*Opts == DHCP_OPTION_LEASE_TIME) {
            Opts +=2;
            memcpy(&Info->LeaseTimer, Opts, 4);
            Opts += 4;
            Info->LeaseTimer = ntohl(Info->LeaseTimer);
            continue;
        }
        if (*Opts == DHCP_OPTION_RENEWAL_TIME) {
            Opts +=2;
            memcpy(&Info->RenewalTimer, Opts, 4);
            Opts += 4;
            Info->RenewalTimer = ntohl(Info->RenewalTimer);
            continue;
        }
        if (*Opts == DHCP_OPTION_REBIND_TIME) {
            Opts +=2;
            memcpy(&Info->RebindTimer, Opts, 4);
            Opts += 4;
            Info->RebindTimer = ntohl(Info->RebindTimer);
            continue;
        }
        if (*Opts == DHCP_OPTION_ROUTER_ADDRESS) {
            /* could be multiple, only take first one */
            Opts += 2;
            memcpy(&Info->DefaultRouter, Opts, 4);
            Opts--;
            Opts += *Opts + 1;
            continue;
        }
        if (*Opts == DHCP_OPTION_DOMAIN_NAME_SERVERS) {
            UINT32 Count, i, Addr;

            Opts++;
            Count = *Opts;
            Opts++;
            assert((Count & 3) == 0);
            Count /= 4;
            /* could be multiple, only take up to MAXNS */
/* OLD CODE
            if (Count > MAXNS)
                Count = MAXNS;

            Info->nscount = Count;
            for (i = 0; i < Count; i++) {
                memcpy(&Addr, Opts + i * 4, 4);
                FillSin(&Info->nsaddr_list[i], Addr, NAMESERVER_PORT);
            }
*/
            // New code -> reserves first slot for multicast DNS server
            if (Count > MAXNS - 1)
                Count = MAXNS - 1;

            Info->nscount = Count + 1;

            DhcpSetMulticastDNSInfo(Info);
            for (i = 1; i < Count + 1; i++) {
                memcpy(&Addr, Opts + i * 4, 4);
                FillSin(&Info->nsaddr_list[i], Addr, NAMESERVER_PORT);
            }
            /* Fill the rest with a null address (0.0.0.0) */
            for (; i < MAXNS; i++)
                FillSin(&Info->nsaddr_list[i], 0, NAMESERVER_PORT);
            Opts--;
            Opts += *Opts + 1;
#if _DEBUG
            {
                UINT8 *p0, *p1, *p2;
                p0 = (UINT8*) &Info->nsaddr_list[0].sin_addr.s_addr;
                p1 = (UINT8*) &Info->nsaddr_list[1].sin_addr.s_addr;
                p2 = (UINT8*) &Info->nsaddr_list[2].sin_addr.s_addr;
                DBGME(2,printf(_T("DHCP got %d DNS servers %d.%d.%d.%d  %d.%d.%d.%d  %d.%d.%d.%d\n"),
                       Info->nscount,
                       p0[0],p0[1],p0[2],p0[3], p1[0],p1[1],p1[2],p1[3], p2[0],p2[1],p2[2],p2[3]));
            }
#endif
            continue;
        }
        if (*Opts == DHCP_OPTION_SUBNET_MASK) {
            Opts += 2;
            memcpy(&Info->SubnetMask, Opts, 4);
            Opts += 4;
            continue;
        }
        if (*Opts == DHCP_OPTION_HOST_NAME) {
            UINT Len = Opts[1];
            Opts += 2;
            if (Info->HostName)
                bsd_free(Info->HostName, 0);
            Info->HostName = bsd_malloc(Len + 1, M_DHCPHOST, M_NOWAIT);
            if (Info->HostName == NULL)
                panic(_T("DhcpParseOptions: out of memory?!?\n"));
            memcpy(Info->HostName, Opts, Len);
            Info->HostName[Len] = 0;
            Opts += Len;
            DBGME(3,printf(_T("Dhcp got host name \"%s\"\n"), Info->HostName));
            continue;
        }
        if (*Opts == DHCP_OPTION_DOMAIN_NAME) {
            UINT Len = Opts[1];
            Opts += 2;
            if (Info->DomainName)
                bsd_free(Info->DomainName, 0);
            Info->DomainName = bsd_malloc(Len + 1, M_DHCPOPT, M_NOWAIT);
            if (Info->DomainName == NULL)
                panic(_T("DhcpParseOptions: out of memory?!?\n"));
            memcpy(Info->DomainName, Opts, Len);
            Info->DomainName[Len] = 0;
            Opts += Len;
            DBGME(3,printf(_T("Dhcp got domain name \"%s\"\n"), Info->DomainName));
            continue;
        }
        if (*Opts == DHCP_OPTION_TIME_SERVERS) {
            /* could be multiple, only take first one */
            Opts += 2;
            memcpy(&Info->TimeServer, Opts, 4);
            Opts += 4;
            continue;
        }
        /*
         * An option type we don't care about.  Skip it.
         */
        Opts++;
        Opts += *Opts + 1;
    }
    if (Opts >= OptEnd)
        return 0;  /* ran out of buffer before we ran out of options */

    return MsgType;
}

/*
 * Free dynamically allocated areas connected to an 'Info' structure.
 */
void DhcpFreeInfoAreas(struct DhcpInfo *Info)
{
    if (Info->pBootFilename)
        bsd_free(Info->pBootFilename, 0);
}

/*
 * Zero/reinitialize the contents of our 'Info' structure.
 * We always use the same inpcb, so insert it here.
 */
void DhcpInitStoredInfo(struct ifnet *pNetInterface)
{
    struct DhcpInfo *Info = pNetInterface->pDhcpInfo;

    if (Info == NULL) {
        Info = bsd_malloc(sizeof(*Info), M_DHCPINFO, M_NOWAIT);
        if (Info == NULL)
          panic(_T("DhcpStart out of mem!\n"));
        memset(Info, 0, sizeof(*Info));
        pNetInterface->pDhcpInfo = Info;
        /* XXX Never freed, should do it when deleting our inpcb */
    } else {
        DhcpFreeInfoAreas(Info);
    }

    memset(Info, 0, sizeof(struct DhcpInfo));
    Info->RexmitInterval = DHCP_REXMIT_DEFAULT + (tv_sec(CurrentTime()) & 0x3);
    Info->pInPcb = &DhcpInPcb;

    /* Default nameserver to UPnP multicast port */
    DhcpSetMulticastDNSInfo(Info);
}

/*
 * Apply information from server to configure our interface.
 */
void DhcpRecordLease(struct ifnet *pNetInterface)
{
    struct DhcpInfo *Info = pNetInterface->pDhcpInfo;
    struct in_aliasreq IfReq;
    UINT32 Adjustment;

    /*
     * Leases are relative to time of our DHCPDISCOVER or DHCPREQUEST message.
     * If not explicitly set by their seperate options, T1 and T2
     * default to .5 and .875 of the lease duration, respectively.
     *
     * Note that since our DHCP timer runs continously, these timers may
     * already have lost a few seconds.  This is o.k.
     */
    Adjustment = tv_sec(TimeSubtract(CurrentTime(),Info->TimeStarted));
    Info->LeaseTimer -= Adjustment;
    if (Info->RenewalTimer == 0)
        Info->RenewalTimer = Info->LeaseTimer >> 1;  /* 1/2 lease */
    else
        Info->RenewalTimer -= Adjustment;
    if (Info->RebindTimer == 0)
        Info->RebindTimer = (Info->LeaseTimer >> 3) * 7;  /* 7/8 lease */
    else
        Info->RebindTimer -= Adjustment;

    /*
     * Set the interface address, subnet mask and broadcast address.
     */
    pNetInterface->if_flags &= ~IFF_NOADDR;

    memset(&IfReq, 0, sizeof IfReq);
    IfReq.ifra_addr.sin_len = (UINT8) sizeof(struct sockaddr_in);
    IfReq.ifra_addr.sin_family = AF_INET;
    IfReq.ifra_addr.sin_addr.s_addr = Info->LeasedIpAddr;

    if (Info->SubnetMask) {
        UINT32 BcastAddr;

        IfReq.ifra_mask.sin_len = (UINT8) sizeof(struct sockaddr_in);
        IfReq.ifra_mask.sin_family = AF_INET;
        IfReq.ifra_mask.sin_addr.s_addr = Info->SubnetMask;

        BcastAddr = ~Info->SubnetMask | Info->LeasedIpAddr;
        IfReq.ifra_broadaddr.sin_len = (UINT8) sizeof(struct sockaddr_in);
        IfReq.ifra_broadaddr.sin_family = AF_INET;
        IfReq.ifra_broadaddr.sin_addr.s_addr = BcastAddr;
    }

    in_control(NULL /* dangerous */, SIOCAIFADDR, &IfReq, pNetInterface);

    /*
     * If we were given routing info, apply it.
     */
    if (Info->DefaultRouter) {
        struct in_addr ina;

        ina.s_addr = Info->DefaultRouter;
        rt_addgw(ina);
#if 1
        {
            UINT8 *p;
            p = (UINT8 *) &Info->DefaultRouter;
            DBGME(2,printf(_T("Default route through gateway %d.%d.%d.%d\n"),
                   p[0], p[1], p[2], p[3]));
        }
#endif
    }

#if 1
    {
        UINT8 *p;
        p = (UINT8 *) &Info->LeasedIpAddr;
        printf(_T("Have lease for IP Addr %d.%d.%d.%d\n"),
               p[0], p[1], p[2], p[3]);
        p = (UINT8 *) &(IfReq.ifra_mask.sin_addr.s_addr);
        DBGME(2,printf(_T("Subnet Mask %d.%d.%d.%d\n"),
               p[0], p[1], p[2], p[3]));
        p = (UINT8 *) &(IfReq.ifra_broadaddr.sin_addr.s_addr);
        DBGME(2,printf(_T("Broadcast Addr %d.%d.%d.%d\n"),
               p[0], p[1], p[2], p[3]));
    }
#endif
}

/*
 * Run through our interfaces, performing retransmissions and
 * expiring address leases as neccessary.
 *
 * All timers are non-zero if running, and contain the number of seconds
 * (minus one) remaining until expiration.
 *
 * BUGBUG? Since RenewalTimer < RebindTimer < LeaseTimer ALWAYS, we might
 * be able to stack them and save some code/data space.  Not clear.
 */
void DhcpTimer(void)
{
    struct ifnet *pIf;
    INT NeedOutput;
    struct DhcpInfo *Info;

    NetLock();
    for (pIf = ifnet; pIf; pIf = pIf->if_next) {
        NeedOutput = 0;
        Info = pIf->pDhcpInfo;
        if (Info->RexmitTimer && --Info->RexmitTimer == 0) {
            /*
             * Retransmission timer expired.  Backoff retransmission interval
             * and resend.  Note that we're much less aggressive about
             * retransmitting when we currently have a valid lease.
             *
             * BUGBUG: What to do if retransmission value gets big?
             * We're supposed to give up if it gets > 64.  What then?
             * Presumeably we'll want the STB to go into "no network" mode.
             */
            Info->RexmitInterval <<= 1;  /* double it */
            if (Info->RexmitInterval > 64)
                Info->RexmitInterval = 64; /* cap it */
            NeedOutput++;
        }

        if (Info->RenewalTimer && --Info->RenewalTimer == 0) {
            /*
             * Renewal timer (T1) expired.  Change to renewing state, set
             * retransmission interval to one half the time until the rebind
             * timer expires, and send a DHCPREQUEST to our server.
             */
            Info->State = DHCPS_RENEWING;
            Info->RexmitInterval = Info->RebindTimer >> 1;
            if (Info->RexmitInterval < 60)
                Info->RexmitInterval = 60;
            Info->TimeStarted = CurrentTime();
            NeedOutput++;
        }

        if (Info->RebindTimer && --Info->RebindTimer == 0) {
            /*
             * Rebinding timer (T2) expired.  Change to rebinding state, set
             * retransmission interval to one half the time until the lease
             * timer expires, and broadcast a DHCPREQUEST.
             */
            Info->State = DHCPS_REBINDING;
            Info->RexmitInterval = Info->LeaseTimer >> 1;
            if (Info->RexmitInterval < 60)
                Info->RexmitInterval = 60;
            Info->TimeStarted = CurrentTime();
            NeedOutput++;
        }

        if (Info->LeaseTimer && --Info->LeaseTimer == 0) {
            /*
             * Lease timer expired.
             * This is really bad news, so immediately try for a new lease.
             * BUGBUG? or should we initiate a complete reboot?
             */
            if_down(pIf);
            DhcpInitStoredInfo(pIf);
            DhcpStart(pIf, 0);
            NeedOutput = 0;  /* true paranoia */
        }

        if (Info->AutoIPRecheckTimer && --Info->AutoIPRecheckTimer == 0) {
            /*
             * AutoIPRecheck timer (T1) expired.  Change to renewing state, set
             * retransmission interval to 1 second because we don't want to wait 
             * for a long time in DhcpWaitTillBound(), and send a DHCPREQUEST to
             * our server.
             */
            Info->State = DHCPS_RENEWING;
            Info->RexmitInterval = 1;
            DhcpStart(pIf, 0); /* Auto IP will recheck DHCP per 5 minutes */
            NeedOutput = 0;
            //NeedOutput++;
        }

        if (NeedOutput) {
            DhcpJitterSmallDelay();
            DhcpOutput(pIf);
        }
    }
    NetUnlock();
}

UINT DhcpJitterSeed;

#define Multiplier ((UINT)(0x80000000ul - 19)) // 2**31 - 19
#define Increment  ((UINT)(0x80000000ul - 61)) // 2**31 - 61
#define Modulus    ((UINT)(0x80000000ul - 1))  // 2**31 - 1

UINT DhcpJitterRandom(void)
{
    DhcpJitterSeed = ((Multiplier * DhcpJitterSeed) + Increment) % Modulus;
    return DhcpJitterSeed;
}

/* Initializes the jitter seed from a 16-byte hardware address.
 * Some of those bytes might not be initialized; this is OK.
 *
 * The seed is a global variable instead of a member of DhcpInfo
 * structure because I don't want to change an exported structure
 * at this time.
 *
 * Note that there is no locking of DhcpJitterSeed; this is OK.
 */
void DhcpJitterInit(UINT8 *pHdwrAddr)
{
    DhcpJitterSeed = ((UINT *)pHdwrAddr)[0] ^ ((UINT *)pHdwrAddr)[1];
    (void) DhcpJitterRandom();
    (void) DhcpJitterRandom();
    (void) DhcpJitterRandom();
}

/* Does a large random delay, roughly up to a second.
 */
void DhcpJitterLargeDelay(void)
{
    TIME delay;

    /* Take 23 low bits and convert them to a TIME value.
     * TIME uses 100ns units so this gives us a delay
     * of 0 to ~840 millis.
     */
    Int32ToInt64(delay, (DhcpJitterRandom() & ((1 << 23) - 1)));

    SleepUntil(TimeSubtract(TIME_ORIGIN, delay)); /* TIME_RELATIVE(Delay) */
}

/* Does a small random delay. This function is called from the timer thread
 * and hence it can't delay for long. The idea here is that if there
 * are many STBs booting simultaneously, then DhcpJitterLargeDelay
 * will still leave many of them acting in concert. (Whenever two STBs
 * hash to the same random delay modulo the kernel's 10ms sleep granularity.)
 * Putting a small random delay before each retransmission will continue to
 * spread these STBs apart in time.
 */
void DhcpJitterSmallDelay(void)
{
    UINT Micros;

    /* Take 7 low bits and convert them to number of microseconds.
     * This gives us a delay of 0 to 127 micros.
     */
    Micros = DhcpJitterRandom() & ((1 << 7) - 1);

    Delay(Micros);
}
