/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1988 Stephen Deering.
 * Copyright (c) 1992, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Stephen Deering of Stanford University.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      @(#)igmp.c      8.1 (Berkeley) 7/19/93
 */

/* Internet Group Management Protocol (IGMP) routines. */


#include <sock.h>
#ifdef __WANTS_IGMP

#include <if.h>
#include <route.h>

#include <ip/in.h>
#include <ip/in_var.h>
#include <ip/in_systm.h>
#include <ip/ip.h>
#include <ip/ip_var.h>
#include <igmp.h>
#include <igmp_var.h>

extern struct ifnet loif;

static int igmp_timers_are_running = 0;
static u_long igmp_all_hosts_group;

static void igmp_sendreport (struct in_multi *);

void
igmp_init(void)
{
        /*
         * To avoid byte-swapping the same value over and over again.
         */
        igmp_all_hosts_group = htonl(INADDR_ALLHOSTS_GROUP);
}

#ifdef _MSC_VER
#pragma warning(disable: 4702) /* unreachable code (IN_*_MULTI macros) */
#endif
void
igmp_input(struct mbuf *m, INT iphlen)
{
        struct igmp *igmp;
        struct ip *ip;
        int igmplen;
        struct ifnet *ifp = m->m_pkthdr.rcvif;
        int minlen;
        struct in_multi *inm;
        struct in_ifaddr *ia;
        struct in_multistep step;

        _COLLECT_STAT(++igmpstat.igps_rcv_total);

        ip = mtod(m, struct ip *);
        igmplen = ip->ip_len;

        /*
         * Validate lengths
         */
        if (igmplen < IGMP_MINLEN) {
                _COLLECT_STAT(++igmpstat.igps_rcv_tooshort);
                m_freem(m);
                return;
        }
        minlen = iphlen + IGMP_MINLEN;
        if ((m->m_flags & M_EXT || m->m_len < minlen) &&
            (m = m_pullup(m, minlen)) == 0) {
                _COLLECT_STAT(++igmpstat.igps_rcv_tooshort);
                return;
        }

        /*
         * Validate checksum
         */
        m->m_data += iphlen;
        m->m_len -= iphlen;
        igmp = mtod(m, struct igmp *);
        if (in_cksum(m, igmplen)) {
                _COLLECT_STAT(++igmpstat.igps_rcv_badsum);
                m_freem(m);
                return;
        }
        m->m_data -= iphlen;
        m->m_len += iphlen;
        ip = mtod(m, struct ip *);

        switch (igmp->igmp_type) {

        case IGMP_HOST_MEMBERSHIP_QUERY:
                _COLLECT_STAT(++igmpstat.igps_rcv_queries);

                if (ifp == &loif)
                        break;

                if (ip->ip_dst.s_addr != igmp_all_hosts_group) {
                        _COLLECT_STAT(++igmpstat.igps_rcv_badqueries);
                        m_freem(m);
                        return;
                }

                /*
                 * Start the timers in all of our membership records for
                 * the interface on which the query arrived, except those
                 * that are already running and those that belong to the
                 * "all-hosts" group.
                 */
                IN_FIRST_MULTI(step, inm);
                while (inm != NULL) {
                        if (inm->inm_ifp == ifp && inm->inm_timer == 0 &&
                            inm->inm_addr.s_addr != igmp_all_hosts_group) {
                                inm->inm_timer =
                                    IGMP_RANDOM_DELAY(inm->inm_addr);
                                igmp_timers_are_running = 1;
                        }
                        IN_NEXT_MULTI(step, inm);
                }

                break;

        case IGMP_HOST_MEMBERSHIP_REPORT:
                _COLLECT_STAT(++igmpstat.igps_rcv_reports);

                if (ifp == &loif)
                        break;

                if (!IN_MULTICAST(ntohl(igmp->igmp_group.s_addr)) ||
                    igmp->igmp_group.s_addr != ip->ip_dst.s_addr) {
                        _COLLECT_STAT(++igmpstat.igps_rcv_badreports);
                        m_freem(m);
                        return;
                }

                /*
                 * KLUDGE: if the IP source address of the report has an
                 * unspecified (i.e., zero) subnet number, as is allowed for
                 * a booting host, replace it with the correct subnet number
                 * so that a process-level multicast routing demon can
                 * determine which subnet it arrived from.  This is necessary
                 * to compensate for the lack of any way for a process to
                 * determine the arrival interface of an incoming packet.
                 */
                if ((ntohl(ip->ip_src.s_addr) & IN_CLASSA_NET) == 0) {
                        IFP_TO_IA(ifp, ia);
                        if (ia) ip->ip_src.s_addr = htonl(ia->ia_subnet);
                }

                /*
                 * If we belong to the group being reported, stop
                 * our timer for that group.
                 */
                IN_LOOKUP_MULTI(igmp->igmp_group, ifp, inm);
                if (inm != NULL) {
                        inm->inm_timer = 0;
                        _COLLECT_STAT(++igmpstat.igps_rcv_ourreports);
                }

                break;
        }

        /*
         * Pass all valid IGMP packets up to any process(es) listening
         * on a raw IGMP socket.
         */
        rip_input(m);
}
#ifdef _MSC_VER
#pragma warning(default: 4702)
#endif

void
igmp_joingroup(struct in_multi *inm)
{
        NetLock();

        if (inm->inm_addr.s_addr == igmp_all_hosts_group ||
            inm->inm_ifp == &loif)
                inm->inm_timer = 0;
        else {
                igmp_sendreport(inm);
                inm->inm_timer = IGMP_RANDOM_DELAY(inm->inm_addr);
                igmp_timers_are_running = 1;
        }
        NetUnlock();
}

void
igmp_leavegroup(struct in_multi *inm)
{
        /*
         * No action required on leaving a group.
         */
}

#ifdef _MSC_VER
#pragma warning(disable: 4702) /* unreachable code (IN_*_MULTI macros) */
#endif
void
igmp_fasttimo(void)
{
        struct in_multi *inm;
        struct in_multistep step;

        /*
         * Quick check to see if any work needs to be done, in order
         * to minimize the overhead of fasttimo processing.
         */
        if (!igmp_timers_are_running)
                return;

        NetLock();
        igmp_timers_are_running = 0;
        IN_FIRST_MULTI(step, inm);
        while (inm != NULL) {
                if (inm->inm_timer == 0) {
                        /* do nothing */
                } else if (--inm->inm_timer == 0) {
                        igmp_sendreport(inm);
                } else {
                        igmp_timers_are_running = 1;
                }
                IN_NEXT_MULTI(step, inm);
        }
        NetUnlock();
}
#ifdef _MSC_VER
#pragma warning(default: 4702)
#endif

static void
igmp_sendreport(struct in_multi *inm)
{
        struct mbuf *m;
        struct igmp *igmp;
        struct ip *ip;
        struct ip_moptions *imo;
        struct ip_moptions simo;
        struct route_in rti;

        MGETHDR(m, M_DONTWAIT, MT_HEADER);
        if (m == NULL)
                return;
        /*
         * Assume max_linkhdr + sizeof(struct ip) + IGMP_MINLEN
         * is smaller than mbuf size returned by MGETHDR.
         */
        m->m_data += max_linkhdr;
        m->m_len = sizeof(struct ip) + IGMP_MINLEN;
        m->m_pkthdr.len = sizeof(struct ip) + IGMP_MINLEN;

        ip = mtod(m, struct ip *);
        ip->ip_tos = 0;
        ip->ip_len = sizeof(struct ip) + IGMP_MINLEN;
        ip->ip_off = 0;
        ip->ip_p = IPPROTO_IGMP;
        ip->ip_src.s_addr = INADDR_ANY;
        ip->ip_dst = inm->inm_addr;

        igmp = (struct igmp *)(ip + 1);
        igmp->igmp_type = IGMP_HOST_MEMBERSHIP_REPORT;
        igmp->igmp_code = 0;
        igmp->igmp_group = inm->inm_addr;
        igmp->igmp_cksum = 0;
        igmp->igmp_cksum = (u_short) in_cksum(m, IGMP_MINLEN);

        imo = &simo;
        bzero((caddr_t)imo, sizeof(*imo));
        imo->imo_multicast_ifp = inm->inm_ifp;
        imo->imo_multicast_ttl = 1;

        memset(&rti, 0, sizeof(rti));
        rti.rt_valid = 1;
        rti.rt_gw.sin_family = AF_INET;
        rti.rt_gw.sin_len = sizeof(struct sockaddr_in);
        rti.rt_gw.sin_addr = inm->inm_addr;
        rti.rt_ifa = &inm->inm_ia->ia_ifa;

        ip_output(m, NULL, &rti, 0, imo);

        _COLLECT_STAT(++igmpstat.igps_snd_reports);
}

#endif /* __WANTS_IGMP */
