/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1988, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)if_ether.c    8.1 (Berkeley) 6/10/93
 */

/*
 * Ethernet address resolution protocol.
 * TODO:
 *    add "inuse/lock" bit (or ref. count) along with valid bit
 */

#include <sock.h>

#include <if.h>
#include <if_dl.h>
#include <route.h>

#include <in.h>
#include <in_systm.h>
#include <in_var.h>
#include <ip.h>
#include <arp.h>

static struct llinfo_arp *arphead;

#define SIN(s) ((struct sockaddr_in *)s)
#define SDL(s) ((struct sockaddr_dl *)s)
#define SRP(s) ((struct sockaddr_inarp *)s)

#define ETHERTYPE_IPTRAILERS ETHERTYPE_TRAIL

/* timer values */
#define ARPT_PRUNE (5*60)    /* walk list every 5 minutes */
#define ARPT_KEEP (20*60)    /* once resolved, good for 20 more minutes */
#define ARPT_DOWN (20)       /* once declared down, don't send for 20 secs */

static void arprequest (struct arpcom *, u_long *, u_long *, u_char *);
static void arptfree (struct llinfo_arp *);
struct llinfo_arp *arplookup (u_long, INT, INT);
static void in_arpinput (struct mbuf *);
static INT doarptimer(PTR, PTR);
static INT arpcmp(PTR, PTR);

extern struct ifnet loif;
INT arp_inuse, arp_allocated, arp_intimer;

#define ARP_MAXTRIES 5
#define USELOOPBACK 1

static INT
doarptimer(PTR pItem, PTR pDummy)
{
    struct llinfo_arp *la = pItem;

    if (la->la_expire && la->la_expire <= tv_sec(time())) {
        assert(la->la_hold == 0); /* XXX: unsure about this one */
        return 0;
    }

    return 1;
}

/*
 * Timeout routine.  Age arp_tab entries periodically.
 */
/* ARGSUSED */
void arptimer(void)
{

    NetLock();
    ListWalk(&arphead, llarpoff, LIST_WALK_REMOVE, doarptimer, NULL);
    NetUnlock();
}


/*
 * Add an entry to the ARP table
 * 
 * To be explicitly called from the "user" level.
 */
void
arp_addentry(struct ifnet *ifp, INT flags,
    struct sockaddr_in *sin, struct sockaddr_dl *dl)
{
    struct llinfo_arp *la;

    la = bsd_malloc(sizeof(struct llinfo_arp), M_RTABLE/*XXX*/,
      M_WAITOK | M_ZERO);
    ListNodeInit(llarpoff, la);
    memcpy(&la->la_inaddr, sin, sizeof(struct sockaddr_in));
    memcpy(&la->la_dladdr, sin, sizeof(struct sockaddr_dl));
    la->la_flags = flags;

    NetLock();
    ListAddFirst(&arphead, llarpoff, la);
    NetUnlock();
}

/*
 * Nuke an entry from the ARP table.
 *
 * To be explicitly called from the "user" level.
 */
void
arp_delentry(struct ifnet *ifp, struct sockaddr_in *sin)
{

    NetLock();
    ListWalk(&arphead, llarpoff, LIST_WALK_REMOVE, arpcmp, sin);
    NetUnlock();
}

/*
 * Broadcast an ARP packet, asking who has addr on interface ac.
 */
void
arpwhohas(struct arpcom *ac, struct in_addr *addr)
{
    arprequest(ac, &ac->ac_ipaddr.s_addr, &addr->s_addr, ac->ac_enaddr);
}

/*
 * Broadcast an ARP request. Caller specifies:
 *   - arp header source ip address
 *   - arp header target ip address
 *   - arp header source ethernet address
 */
static void
arprequest(struct arpcom *ac, u_long *sip, u_long *tip, u_char *enaddr)
{
    struct mbuf *m;
    struct ether_header *eh;
    struct ether_arp *ea;
    struct sockaddr sa;

    if ((m = m_gethdr(M_DONTWAIT, MT_DATA)) == NULL)
        return;
    m->m_len = sizeof(*ea);
    m->m_pkthdr.len = sizeof(*ea);
    m->m_flags |= M_BCAST;
    MH_ALIGN(m, sizeof(*ea));
    ea = mtod(m, struct ether_arp *);
    eh = (struct ether_header *)sa.sa_data;
    bzero((caddr_t)ea, sizeof (*ea));
    bcopy((caddr_t)etherbroadcastaddr, (caddr_t)eh->ether_dhost,
        sizeof(eh->ether_dhost));
    eh->ether_type = ETHERTYPE_ARP;     /* if_output will swap */
    ea->arp_hrd = htons(ARPHRD_ETHER);
    ea->arp_pro = htons(ETHERTYPE_IP);
    ea->arp_hln = sizeof(ea->arp_sha);  /* hardware address length */
    ea->arp_pln = sizeof(ea->arp_spa);  /* protocol address length */
    ea->arp_op = htons(ARPOP_REQUEST);
    bcopy((caddr_t)enaddr, (caddr_t)ea->arp_sha, sizeof(ea->arp_sha));
    bcopy((caddr_t)sip, (caddr_t)ea->arp_spa, sizeof(ea->arp_spa));
    bcopy((caddr_t)tip, (caddr_t)ea->arp_tpa, sizeof(ea->arp_tpa));
    sa.sa_family = AF_UNSPEC;
    sa.sa_len = sizeof(sa);
    (*ac->ac_if.v->Output)(iIF(&ac->ac_if), m, &sa, NULL);
}

/*
 * Resolve an IP address into an ethernet address.  If success,
 * desten is filled in.  If there is no entry in arptab,
 * set one up and broadcast a request for the IP address.
 * Hold onto this mbuf and resend it once the address
 * is finally resolved.  A return value of 1 indicates
 * that desten has been filled in and the packet should be sent
 * normally; a 0 return indicates that the packet has been
 * taken over here, either now or for later transmission.
 */
INT
arpresolve(struct arpcom *ac, struct route_in *rti, struct mbuf *m,
    u_char *desten)
{
    struct llinfo_arp *la;
    struct sockaddr_dl *sdl;

    if (m->m_flags & M_BCAST) {  /* broadcast */
        bcopy((caddr_t)etherbroadcastaddr, (caddr_t)desten,
            sizeof(etherbroadcastaddr));
        return (1);
    }
    if (m->m_flags & M_MCAST) {  /* multicast */
        ETHER_MAP_IP_MULTICAST(&rti->rt_gw.sin_addr, desten);
        return(1);
    }

#if 0
    arpdump();
#endif
    
    la = arplookup(rti->rt_gw.sin_addr.s_addr, 1, 0);
    if (la == 0) {
        errlog(LOG_DEBUG, _T("arpresolve: can't allocate llinfo"), 0, 0);
        m_freem(m);
        return (0);
    }

    sdl = &la->la_dladdr;

    /*
     * Check for resolved & valid entry
     */
    if ((la->la_flags & (LAF_EMBRYO | LAF_DOWN)) == 0
      && ((la->la_expire == 0) || (la->la_expire > tv_sec(time())))) {
        bcopy(LLADDR(sdl), desten, sdl->sdl_alen);
        return 1;
    }

    /*
     * There is an arptab entry, but no ethernet address
     * response yet.  Replace the held mbuf with this
     * latest one.
     */
    if (la->la_hold)
        m_freem(la->la_hold);
    la->la_hold = m;
    if (la->la_expire) {
        if (la->la_asked == 0 || la->la_expire != tv_sec(time())) {
            la->la_expire = tv_sec(time());
            if (la->la_asked++ < ARP_MAXTRIES)
                arpwhohas(ac, &rti->rt_gw.sin_addr);
            else {
                la->la_flags |= LAF_DOWN;
                la->la_expire += ARPT_DOWN;
                la->la_asked = 0;
            }
        }
    }
    return (0);
}

/*
 * Common length and type checks are done here,
 * then the protocol-specific routine is called.
 */
void
arp_input(struct mbuf *m)
{
    struct arphdr *ar;

    NetLock();

    if (m == 0 || (m->m_flags & M_PKTHDR) == 0)
        panic(_T("arp_input: buffer chain has no packet header!\n"));
    if (m->m_len >= sizeof(struct arphdr) &&
            ((ar = mtod(m, struct arphdr *)) != NULL) &&
            ntohs(ar->ar_hrd) == ARPHRD_ETHER && m->m_len >=
            (INT)(sizeof(struct arphdr) + 2 * ar->ar_hln + 2 * ar->ar_pln))
        switch (ntohs(ar->ar_pro)) {

        case ETHERTYPE_IP:
        case ETHERTYPE_IPTRAILERS:
            in_arpinput(m);
        }
    else
        m_freem(m);

    NetUnlock();
}

extern void DhcpAutoIPInput(struct ifnet *pNetInterface, struct ether_arp *ea);

/*
 * ARP for Internet protocols on 10 Mb/s Ethernet.
 * Algorithm is that given in RFC 826.
 * In addition, a sanity check is performed on the sender
 * protocol address, to catch impersonators.
 * We no longer handle negotiations for use of trailer protocol:
 * Formerly, ARP replied for protocol type ETHERTYPE_TRAIL sent
 * along with IP replies if we wanted trailers sent to us,
 * and also sent them in response to IP replies.
 * This allowed either end to announce the desire to receive
 * trailer packets.
 * We no longer reply to requests for ETHERTYPE_TRAIL protocol either,
 * but formerly didn't normally send requests.
 */
static void
in_arpinput(struct mbuf *m)
{
    struct ether_arp *ea;
    struct arpcom *ac = (struct arpcom *)m->m_pkthdr.rcvif;
    struct ether_header *eh;
    struct llinfo_arp *la = 0;
    struct in_ifaddr *ia, *maybe_ia = 0;
    struct sockaddr_dl *sdl;
    struct sockaddr sa;
    struct in_addr isaddr, itaddr, myaddr;
    INT op;

    ea = mtod(m, struct ether_arp *);
    op = ntohs(ea->arp_op);
    bcopy((caddr_t)ea->arp_spa, (caddr_t)&isaddr, sizeof (isaddr));
    bcopy((caddr_t)ea->arp_tpa, (caddr_t)&itaddr, sizeof (itaddr));

    /*
     * In case of an Auto IP address, pass it onto the DHCP code
     * for evaluation.
     */
    if ((ntohl(itaddr.s_addr) & IN_CLASSB_NET) == 0xa9fe0000
      || (ntohl(isaddr.s_addr) & IN_CLASSB_NET) == 0xa9fe0000)
        DhcpAutoIPInput(m->m_pkthdr.rcvif, ea);

    /* XXX: best address selection */
    for (ia = _in_ifaddr; ia; ia = ia->ia_next) {
        if (ia->ia_ifp == &ac->ac_if) {
            if ((itaddr.s_addr == ia->ia_addr.sin_addr.s_addr) ||
              (isaddr.s_addr == ia->ia_addr.sin_addr.s_addr)) {
                maybe_ia = ia;
                break;
            }
        }
    }
    if (maybe_ia == 0)
        goto out;
    myaddr = ia ? ia->ia_addr.sin_addr : maybe_ia->ia_addr.sin_addr;

    if (!bcmp((caddr_t)ea->arp_sha, (caddr_t)ac->ac_enaddr,
            sizeof(ea->arp_sha)))
        goto out;  /* it's from me, ignore it. */
    if (!bcmp((caddr_t)ea->arp_sha, (caddr_t)etherbroadcastaddr,
            sizeof(ea->arp_sha))) {
        errlog(LOG_ERR,
            _T("arp: ether address is broadcast for IP address %x!\n"),
            ntohl(isaddr.s_addr), 0);
        goto out;
    }
    if (isaddr.s_addr == myaddr.s_addr) {
        errlog(LOG_ERR,
           _T("duplicate IP address %x!! sent from ethernet address: %hs\n"),
           ntohl(isaddr.s_addr), (UINT)ether_sprintf(ea->arp_sha));
        itaddr = myaddr;
        goto reply;
    }
    la = arplookup(isaddr.s_addr, itaddr.s_addr == myaddr.s_addr, 0);
    if (la) {
        sdl = &la->la_dladdr;
        if (sdl->sdl_alen &&
                bcmp((caddr_t)ea->arp_sha, LLADDR(sdl), sdl->sdl_alen))
            errlog(LOG_INFO, _T("arp info overwritten for %x by %hs\n"),
                isaddr.s_addr, (UINT)ether_sprintf(ea->arp_sha));
        bcopy((caddr_t)ea->arp_sha, LLADDR(sdl),
                sdl->sdl_alen = sizeof(ea->arp_sha));
        if (la->la_expire)
            la->la_expire = tv_sec(time()) + ARPT_KEEP;
        la->la_asked = 0;
        la->la_flags &= ~(LAF_EMBRYO | LAF_DOWN);
        if (la->la_hold) {
            (*ac->ac_if.v->Output)(iIF(&ac->ac_if),
                                   la->la_hold,
                                   (struct sockaddr *)&la->la_inaddr, 0);
            la->la_hold = 0;
        }
    }
reply:
    if (op != ARPOP_REQUEST) {
    out:
        m_freem(m);
        return;
    }
    if (itaddr.s_addr == myaddr.s_addr) {
        /* I am the target */
        bcopy((caddr_t)ea->arp_sha, (caddr_t)ea->arp_tha, sizeof(ea->arp_sha));
        bcopy((caddr_t)ac->ac_enaddr, (caddr_t)ea->arp_sha,
            sizeof(ea->arp_sha));
    } else {
        la = arplookup(itaddr.s_addr, 0, 1 /*SIN_PROXY*/);
        if (la == NULL)
            goto out;
        sdl = &la->la_dladdr;
        bcopy((caddr_t)ea->arp_sha, (caddr_t)ea->arp_tha, sizeof(ea->arp_sha));
        bcopy(LLADDR(sdl), (caddr_t)ea->arp_sha, sizeof(ea->arp_sha));
    }

    bcopy((caddr_t)ea->arp_spa, (caddr_t)ea->arp_tpa, sizeof(ea->arp_spa));
    bcopy((caddr_t)&itaddr, (caddr_t)ea->arp_spa, sizeof(ea->arp_spa));
    ea->arp_op = htons(ARPOP_REPLY);
    ea->arp_pro = htons(ETHERTYPE_IP);  /* let's be sure! */
    eh = (struct ether_header *)sa.sa_data;
    bcopy((caddr_t)ea->arp_tha, (caddr_t)eh->ether_dhost,
        sizeof(eh->ether_dhost));
    m->m_flags &= ~(M_BCAST|M_MCAST);
    eh->ether_type = ETHERTYPE_ARP;
    sa.sa_family = AF_UNSPEC;
    sa.sa_len = sizeof(sa);
    (*ac->ac_if.v->Output)(iIF(&ac->ac_if), m, &sa, NULL);
    return;
}

static INT
arpcmp(PTR pItem, PTR pArg)
{
     struct sockaddr_in *s;
     UINT32 *addr = pArg;

     s = (struct sockaddr_in *)&((struct llinfo_arp *)pItem)->la_inaddr;
     return !(s->sin_addr.s_addr == *addr);
}

#if 0
#define LIST_DOUBLE2_SIMPLECMP(funcname, argtype, argmember, cmpoper)   \
static INT funcname(PTR, PTR);                                          \
static INT funcname(PTR p1, PTR p2)                                     \
{                                                                       \
    argtype *s1 = (void *)p1, *s2 = (void *p2);                         \
    return !(s1->argmember cmpoper s2->argmember);                      \
}

LIST_DOUBLE2_SIMPLECMP(arpcmp, struct llinfo_arp, la_inaddr.sin_addr.s_addr, ==)
#endif

/*
 * Lookup or enter a new address in arptab.
 */
struct llinfo_arp *
arplookup(u_long addr, INT create, INT proxy)
{
    struct llinfo_arp *la;

    la = ListWalk(&arphead, llarpoff, LIST_WALK_SEARCH, arpcmp, &addr);
    if (la)
        return la;
    
    /* request to add new entry */
    if (create) {
        la = bsd_malloc(sizeof(struct llinfo_arp), M_RTABLE/*XXX*/,
          M_NOWAIT|M_ZERO);
        if (!la)
            return NULL;

        ListNodeInit(llarpoff, la);
        la->la_inaddr.sin_len = sizeof(struct sockaddr_in);
        la->la_inaddr.sin_family = AF_INET;
        la->la_expire = 1;
        la->la_flags |= LAF_EMBRYO;
        memcpy(&la->la_inaddr.sin_addr, &addr, sizeof(struct in_addr));

        ListAddFirst(&arphead, llarpoff, la);

        return la;
    }

    return NULL;
}

#undef DEBUG
#ifdef DEBUG
static INT entrydump(PTR, PTR);
static INT
entrydump(PTR p1, PTR p2)
{
    struct llinfo_arp *la = p1;
    UINT8 *ina = (UINT8 *)&la->la_inaddr.sin_addr;
    UINT8 *dla = (UINT8 *)LLADDR(&la->la_dladdr);

    printf("%02d.%02d.%02d.%02d\t%s\t0x%x\t%d\t%u\t0x%x\n",
      ina[0], ina[1], ina[2], ina[3], ether_sprintf(dla),
      (UINT8 *)la->la_hold, la->la_asked, la->la_expire, la->la_flags);

    return S_OK;
}

void
arpdump()
{

    printf("Begin ARP Table\n");
    printf("inaddr\t\tdladdr\t\t\tmbuf\tasked\texpire\t\tflags\n");
    ListWalk(&arphead, llarpoff, LIST_WALK_ITERATE, entrydump, NULL);
    printf("End ARP Table\n");
}
#endif /* DEBUG */
