/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1991, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)in.c    8.1 (Berkeley) 6/10/93
 */

#include <sock.h>
#include <stdio.h>

#include <if.h>
#include <if_types.h>
#include <route.h>

#include <in_systm.h>
#include <in.h>
#include <in_var.h>
#include <igmp_var.h>
#include <arp.h>

INT ether_ioctl(struct ifnet *ifp, INT cmd, PTR data); /* in ethernet.c */

#ifdef INET
/*
 * Return the network number from an internet address.
 */
UINT32 in_netof(struct in_addr in)
{
    UINT32 i = ntohl(in.s_addr);
    UINT32 net;
    struct in_ifaddr *ia;

    if (IN_CLASSA(i))
        net = i & IN_CLASSA_NET;
    else if (IN_CLASSB(i))
        net = i & IN_CLASSB_NET;
    else if (IN_CLASSC(i))
        net = i & IN_CLASSC_NET;
    else if (IN_CLASSD(i))
        net = i & IN_CLASSD_NET;
    else
        return (0);

    /*
     * Check whether network is a subnet;
     * if so, return subnet number.
     */
    for (ia = _in_ifaddr; ia; ia = ia->ia_next)
        if (net == ia->ia_net)
            return (i & ia->ia_subnetmask);
    return (net);
}

#ifndef SUBNETSARELOCAL
#define SUBNETSARELOCAL 1
#endif
INT subnetsarelocal = SUBNETSARELOCAL;
/*
 * Return 1 if an internet address is for a ``local'' host
 * (one to which we have a connection).  If subnetsarelocal
 * is true, this includes other subnets of the local net.
 * Otherwise, it includes only the directly-connected (sub)nets.
 */
INT in_localaddr(struct in_addr in)
{
    UINT32 i = ntohl(in.s_addr);
    struct in_ifaddr *ia;

    if (subnetsarelocal) {
        for (ia = _in_ifaddr; ia; ia = ia->ia_next)
            if ((i & ia->ia_netmask) == ia->ia_net)
                return (1);
    } else {
        for (ia = _in_ifaddr; ia; ia = ia->ia_next)
            if ((i & ia->ia_subnetmask) == ia->ia_subnet)
                return (1);
    }
    return (0);
}

/*
 * Determine whether an IP address is in a reserved set of addresses
 * that may not be forwarded, or whether datagrams to that destination
 * may be forwarded.
 */
INT in_canforward(struct in_addr in)
{
    UINT32 i = ntohl(in.s_addr);
    UINT32 net;

    if (IN_EXPERIMENTAL(i) || IN_MULTICAST(i))
        return (0);
    if (IN_CLASSA(i)) {
        net = i & IN_CLASSA_NET;
        if (net == 0 || net == (IN_LOOPBACKNET << IN_CLASSA_NSHIFT))
            return (0);
    }
    return (1);
}

/*
 * Trim a mask in a sockaddr
 */
void in_socktrim(struct sockaddr_in *ap)
{
    char *cplim = (char *) &ap->sin_addr;
    char *cp = (char *) (&ap->sin_addr + 1);

    ap->sin_len = 0;
    while (--cp >= cplim)
        if (*cp) {
            (ap)->sin_len = (UINT8)((cp - (char *)ap) + (UINT8)1);
            break;
        }
}

INT in_interfaces;  /* number of external internet interfaces */
extern struct ifnet loif;

#if 0
void pa(struct sockaddr_in *s)
{
    printf("%x %x %x %x (%x %x)\n", 
       s->sin_len, s->sin_family, s->sin_port, s->sin_addr.s_addr,
       *(int*)&s->sin_zero, *(int*)&s->sin_zero[4]);
}
#endif

/*
 * Generic internet control operations (ioctl's).
 * Ifp is 0 if not an interface-specific ioctl.
 */
INT in_control(struct socket *so, INT cmd, PTR data, struct ifnet *ifp)
{
    struct sockaddr_in oldaddr;
    struct ifreq *ifr = (struct ifreq *)data;
    struct in_ifaddr *ia = 0;
    struct ifaddr *ifa;
    struct in_ifaddr *oia;
    struct in_aliasreq *ifra = (struct in_aliasreq *)data;
    INT error, hostIsNew, maskIsNew;
    UINT32 i;

    /*
     * Find address for this interface, if it exists.
     */
    if (ifp)
        for (ia = _in_ifaddr; ia; ia = ia->ia_next)
            if (ia->ia_ifp == ifp)
                break;

    switch (cmd) {

    case SIOCAIFADDR:
    case SIOCDIFADDR:
        if (ifra->ifra_addr.sin_family == AF_INET)
            for (oia = ia; ia; ia = ia->ia_next) {
            if (ia->ia_ifp == ifp  &&
                ia->ia_addr.sin_addr.s_addr ==
                ifra->ifra_addr.sin_addr.s_addr)
                break;
        }
        if (cmd == SIOCDIFADDR && ia == 0)
            return (EADDRNOTAVAIL);
        /* FALLTHROUGH */
    case SIOCSIFADDR:
    case SIOCSIFNETMASK:
    case SIOCSIFDSTADDR:
#ifdef SECURITY
        if ((so->so_state & SS_PRIV) == 0)
            return (EACCES); /* af: was EPERM */
#endif

        if (ifp == 0)
            panic(_T("in_control"));
        if (ia == (struct in_ifaddr *)0) {
            oia = (struct in_ifaddr *)
                bsd_malloc(sizeof *oia, M_IFADDR, M_WAITOK);
            if (oia == (struct in_ifaddr *)NULL)
                return (ENOBUFS);
            bzero((caddr_t)oia, sizeof *oia);
            ia = _in_ifaddr;
            if (ia != NULL) {
                for ( ; ia->ia_next; ia = ia->ia_next)
                    continue;
                ia->ia_next = oia;
            } else
                _in_ifaddr = oia;
            ia = oia;
            ifa = ifp->if_addrlist;
            if (ifa != NULL) {
                for ( ; ifa->ifa_next; ifa = ifa->ifa_next)
                    continue;
                ifa->ifa_next = (struct ifaddr *) ia;
            } else
                ifp->if_addrlist = (struct ifaddr *) ia;
            ia->ia_ifa.ifa_addr = (struct sockaddr *)&ia->ia_addr;
            ia->ia_ifa.ifa_dstaddr
                    = (struct sockaddr *)&ia->ia_dstaddr;
            ia->ia_ifa.ifa_netmask
                    = (struct sockaddr *)&ia->ia_sockmask;
            ia->ia_sockmask.sin_len = 8;
            if (ifp->if_flags & IFF_BROADCAST) {
                ia->ia_broadaddr.sin_len = sizeof(ia->ia_addr);
                ia->ia_broadaddr.sin_family = AF_INET;
            }
            ia->ia_ifp = ifp;
            if (ifp != &loif)
                in_interfaces++;
        }
        break;

    case SIOCSIFBRDADDR:
#ifdef SECURITY
        if ((so->so_state & SS_PRIV) == 0)
            return (EACCES); /* af: was EPERM */
#endif
        /* FALLTHROUGH */

    case SIOCGIFADDR:
    case SIOCGIFNETMASK:
    case SIOCGIFDSTADDR:
    case SIOCGIFBRDADDR:
        if (ia == (struct in_ifaddr *)0)
            return (EADDRNOTAVAIL);
        break;
    }
    switch (cmd) {

    case SIOCGIFADDR:
        *((struct sockaddr_in *)&ifr->ifr_addr) = ia->ia_addr;
        break;

    case SIOCGIFBRDADDR:
        if ((ifp->if_flags & IFF_BROADCAST) == 0)
            return (EINVAL);
        *((struct sockaddr_in *)&ifr->ifr_dstaddr) = ia->ia_broadaddr;
        break;

    case SIOCGIFDSTADDR:
        if ((ifp->if_flags & IFF_POINTOPOINT) == 0)
            return (EINVAL);
        *((struct sockaddr_in *)&ifr->ifr_dstaddr) = ia->ia_dstaddr;
        break;

    case SIOCGIFNETMASK:
        *((struct sockaddr_in *)&ifr->ifr_addr) = ia->ia_sockmask;
        break;

    case SIOCSIFDSTADDR:
        if ((ifp->if_flags & IFF_POINTOPOINT) == 0)
            return (EINVAL);
        oldaddr = ia->ia_dstaddr;
        ia->ia_dstaddr = *(struct sockaddr_in *)&ifr->ifr_dstaddr;
        if (ifp->v->Ioctl &&
            ((error = ifp->v->Ioctl(iIF(ifp), SIOCSIFDSTADDR, (PTR)ia)) != 0)) {
            ia->ia_dstaddr = oldaddr;
            return (error);
        }
        break;

    case SIOCSIFBRDADDR:
        if ((ifp->if_flags & IFF_BROADCAST) == 0)
            return (EINVAL);
        ia->ia_broadaddr = *(struct sockaddr_in *)&ifr->ifr_broadaddr;
        break;

    case SIOCSIFADDR:
        return (in_ifinit(ifp, ia,
            (struct sockaddr_in *) &ifr->ifr_addr, 1));

    case SIOCSIFNETMASK:
        i = ifra->ifra_addr.sin_addr.s_addr;
        ia->ia_subnetmask = ntohl(ia->ia_sockmask.sin_addr.s_addr = i);
        break;

    case SIOCAIFADDR:
        maskIsNew = 0;
        hostIsNew = 1;
        error = 0;
        if (ia->ia_addr.sin_family == AF_INET) {
            if (ifra->ifra_addr.sin_len == 0) {
                ifra->ifra_addr = ia->ia_addr;
                hostIsNew = 0;
            } else if (ifra->ifra_addr.sin_addr.s_addr ==
                           ia->ia_addr.sin_addr.s_addr)
                hostIsNew = 0;
        }
        if (ifra->ifra_mask.sin_len) {
            in_ifscrub(ifp, ia);
            ia->ia_sockmask = ifra->ifra_mask;
            ia->ia_subnetmask =
                 ntohl(ia->ia_sockmask.sin_addr.s_addr);
            maskIsNew = 1;
        }
        if ((ifp->if_flags & IFF_POINTOPOINT) &&
            (ifra->ifra_dstaddr.sin_family == AF_INET)) {
            in_ifscrub(ifp, ia);
            ia->ia_dstaddr = ifra->ifra_dstaddr;
            maskIsNew  = 1; /* We lie; but the effect's the same */
        }
#if 0
printf("Addrs are:\n");
pa(&ifra->ifra_addr);
pa(&ifra->ifra_broadaddr);
pa(&ifra->ifra_mask);
#endif
        if (ifra->ifra_addr.sin_family == AF_INET &&
            (hostIsNew || maskIsNew))
            error = in_ifinit(ifp, ia, &ifra->ifra_addr, 0);
        if ((ifp->if_flags & IFF_BROADCAST) &&
            (ifra->ifra_broadaddr.sin_family == AF_INET))
            ia->ia_broadaddr = ifra->ifra_broadaddr;
        return (error);

    case SIOCDIFADDR:
        in_ifscrub(ifp, ia);
        if ((ifa = ifp->if_addrlist) == (struct ifaddr *)ia)
            ifp->if_addrlist = ifa->ifa_next;
        else {
            while (ifa->ifa_next &&
                   (ifa->ifa_next != (struct ifaddr *)ia))
                    ifa = ifa->ifa_next;
            if (ifa->ifa_next)
                ifa->ifa_next = ((struct ifaddr *)ia)->ifa_next;
            else
                printf("Couldn't unlink inifaddr from ifp\n");
        }
        oia = ia;
        if (oia == (ia = _in_ifaddr))
            _in_ifaddr = ia->ia_next;
        else {
            while (ia->ia_next && (ia->ia_next != oia))
                ia = ia->ia_next;
            if (ia->ia_next)
                ia->ia_next = oia->ia_next;
            else
                printf("Didn't unlink inifaddr from list\n");
        }
        IFAFREE((&oia->ia_ifa));
        break;

    default:
        if (ifp == 0 || ifp->v->Ioctl == 0)
            return (EOPNOTSUPP);
        return ((*ifp->v->Ioctl)(iIF(ifp), cmd, data));
    }
    return (0);
}

/*
 * Delete any existing route for an interface.
 */
void in_ifscrub(struct ifnet *ifp, struct in_ifaddr *ia)
{

    if (&ia->ia_ifa == rt_getgw())
        rt_delgw();

    arp_delentry(ifp, &ia->ia_addr);
}

/*
 * Initialize an interface's internet address
 * and routing table entry.
 */
INT in_ifinit(struct ifnet *ifp, struct in_ifaddr *ia,
              struct sockaddr_in *sin, INT scrub)
{
    UINT32 i = ntohl(sin->sin_addr.s_addr);
    struct sockaddr_in oldaddr;
    struct ifaddr *ifa;
    INT error = 0;

    NetLock();

#ifdef mmheavy
    {
        PIMACHINE pMach = CurrentMachine();
        struct arpcom *arp = (struct arpcom *)ifp;

        /* Provide seed material to the kernel.
         * If ifp really isn't an arpcom, no harm done
         * as long as we don't fault.
         */
        pMach->v->SeedIpids(pMach, (unsigned char *)sin, sizeof sin);
        pMach->v->SeedIpids(pMach, (unsigned char *)&arp->ac_enaddr, sizeof arp->ac_enaddr);
    }
#endif

    oldaddr = ia->ia_addr;
    ia->ia_addr = *sin;
    /*
     * Give the interface a chance to initialize
     * if this is its first address,
     * and to validate the address if necessary.
     */
    if (ifp->if_type == IFT_ETHER)
        error = ether_ioctl(ifp, SIOCSIFADDR, (caddr_t)ia);
    else if (ifp->v->Ioctl)
        error = (*ifp->v->Ioctl)(iIF(ifp), SIOCSIFADDR, (PTR)ia);

    if (error) {
        NetUnlock();
        ia->ia_addr = oldaddr;
        return (error);
    }

    NetUnlock();
    if (scrub) {
        ia->ia_ifa.ifa_addr = (struct sockaddr *)&oldaddr;
        in_ifscrub(ifp, ia);
        ia->ia_ifa.ifa_addr = (struct sockaddr *)&ia->ia_addr;
    }
    if (IN_CLASSA(i))
        ia->ia_netmask = IN_CLASSA_NET;
    else if (IN_CLASSB(i))
        ia->ia_netmask = IN_CLASSB_NET;
    else
        ia->ia_netmask = IN_CLASSC_NET;

    /*
     * The subnet mask usually includes at least the standard network part,
     * but may may be smaller in the case of supernetting.
     * If it is set, we believe it.
     */
    if (ia->ia_subnetmask == 0) {
        ia->ia_subnetmask = ia->ia_netmask;
        ia->ia_sockmask.sin_addr.s_addr = htonl(ia->ia_subnetmask);
    } else
        ia->ia_netmask &= ia->ia_subnetmask;
#if 0
printf("Sockmask is:\n");
pa(&ia->ia_sockmask);
#endif
#if 0
    {
    UINT8 *p;
    p = (UINT8 *) &ia->ia_subnetmask;
    printf("subnetmask is %d.%d.%d.%d\n",
            p[3], p[2], p[1], p[0]);
    }
#endif
    ia->ia_net = i & ia->ia_netmask;
    ia->ia_subnet = i & ia->ia_subnetmask;
    in_socktrim(&ia->ia_sockmask);

    /*
     * Add arp entry
     */
    for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
        if (ifa->ifa_addr->sa_family == AF_LINK) {
            arp_addentry(ifp, 0, sin, (struct sockaddr_dl *)ifa->ifa_addr);
            break;
        }
    }

#ifdef MULTICAST
    /*
     * If the interface supports multicast, join the "all hosts"
     * multicast group on that interface.
     */
    if (ifp->if_flags & IFF_MULTICAST) {
        struct in_addr addr;

        addr.s_addr = htonl(INADDR_ALLHOSTS_GROUP);
        in_addmulti(&addr, ifp);
    }
#endif
    return (error);
}

#if MONARCH_ADHOC
int
in_broadcast_any_ifp(struct in_addr dst)
{
    struct ifnet *ifp;
    
    for (ifp = ifnet; ifp; ifp = ifp->if_next) {
    //TAILQ_FOREACH(ifp, &ifnet, if_link) {
        if(in_broadcast(dst, ifp))
            return 1;   /* found address */
    }
    return 0;
}
#endif

/*
 * Return 1 if the address might be a local broadcast address.
 */
INT in_broadcast(struct in_addr in, struct ifnet *ifp)
{
    struct ifaddr *ifa;
    UINT32 t;

    if (in.s_addr == INADDR_BROADCAST ||
        in.s_addr == INADDR_ANY)
        return 1;
    if ((ifp->if_flags & IFF_BROADCAST) == 0)
        return 0;
    t = ntohl(in.s_addr);
    /*
     * Look through the list of addresses for a match
     * with a broadcast address.
     */
#define ia ((struct in_ifaddr *)ifa)
    for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next)
        if (ifa->ifa_addr->sa_family == AF_INET &&
            (in.s_addr == ia->ia_broadaddr.sin_addr.s_addr ||
             in.s_addr == ia->ia_netbroadcast.s_addr ||
             /*
              * Check for old-style (host 0) broadcast.
              */
             t == ia->ia_subnet || t == ia->ia_net))
                return 1;
    return (0);
#undef ia
}
/*
 * Add an address to the list of IP multicast addresses for a given interface.
 */
struct in_multi *in_addmulti(struct in_addr *ap, struct ifnet *ifp)
{
    struct in_multi *inm;
    struct ifreq ifr;
    struct in_ifaddr *ia;
    INT error;

    NetLock();
    /*
     * See if address already in list.
     */
    IN_LOOKUP_MULTI(*ap, ifp, inm);
    if (inm != NULL) {
        /*
         * Found it; just increment the reference count.
         */
        ++inm->inm_refcount;
    }
    else {
        /*
         * New address; allocate a new multicast record
         * and link it into the interface's multicast list.
         */
        inm = (struct in_multi *)bsd_malloc(sizeof(*inm),
            M_IPMADDR, M_NOWAIT);
        if (inm == NULL) {
            NetUnlock();
            return (NULL);
        }
        inm->inm_addr = *ap;
        inm->inm_ifp = ifp;
        inm->inm_refcount = 1;
        IFP_TO_IA(ifp, ia);
        if (ia == NULL) {
            bsd_free(inm, M_IPMADDR);
            NetUnlock();
            return (NULL);
        }
        inm->inm_ia = ia;
        inm->inm_next = ia->ia_multiaddrs;
        ia->ia_multiaddrs = inm;
        /*
         * Ask the network driver to update its multicast reception
         * filter appropriately for the new address.
         */
        ((struct sockaddr_in *)&ifr.ifr_addr)->sin_family = AF_INET;
        ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr = *ap;
        if (ifp->if_type == IFT_ETHER)
            error = ether_ioctl(ifp, SIOCADDMULTI, (caddr_t)&ifr);
        else if (ifp->v->Ioctl)
            error = (*ifp->v->Ioctl)(iIF(ifp), SIOCADDMULTI, (PTR)&ifr);
        else
            error = EINVAL;
        if (error) {
            ia->ia_multiaddrs = inm->inm_next;
            bsd_free(inm, M_IPMADDR);
            NetUnlock();
            return (NULL);
        }
        /*
         * Let IGMP know that we have joined a new IP multicast group.
         */
#ifdef __WANTS_IGMP
        igmp_joingroup(inm);
#endif
    }
    NetUnlock();
    return (inm);
}

/*
 * Delete a multicast address record.
 */
void in_delmulti(struct in_multi *inm)
{
    struct in_multi **p;
    struct ifreq ifr;
    struct ifnet *ifp;

    NetLock();
    if (--inm->inm_refcount == 0) {
        /*
         * No remaining claims to this record; let IGMP know that
         * we are leaving the multicast group.
         */
#ifdef __WANTS_IGMP
        igmp_leavegroup(inm);
#endif
        /*
         * Unlink from list.
         */
        for (p = &inm->inm_ia->ia_multiaddrs;
             *p != inm;
             p = &(*p)->inm_next)
             continue;
        *p = (*p)->inm_next;
        /*
         * Notify the network driver to update its multicast reception
         * filter.
         */
        ifp = inm->inm_ifp;
        ((struct sockaddr_in *)&(ifr.ifr_addr))->sin_family = AF_INET;
        ((struct sockaddr_in *)&(ifr.ifr_addr))->sin_addr =
                                inm->inm_addr;
        if (ifp->if_type == IFT_ETHER)
            (void) ether_ioctl(ifp, SIOCDELMULTI, (caddr_t)&ifr);
        else
            (void) (*ifp->v->Ioctl)(iIF(ifp), SIOCDELMULTI, (PTR) &ifr);
        bsd_free(inm, M_IPMADDR);
    }
    NetUnlock();
}
#endif
