/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1988, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)ip_input.c    8.1 (Berkeley) 6/10/93
 */

#include <sock.h>
#include <domain.h>

#include <if.h>
#include <route.h>

#include <in.h>
#include <in_systm.h>
#include <ip.h>
#include <in_pcb.h>
#include <in_var.h>
#include <ip_var.h>
#include <ip_icmp.h>
#include <assert.h>

#if MONARCH_ADHOC
#include <dsr/dsr_queue.h>
#include <dsr/ip6_opts.h>
#include <dsr/ip6.h>
#include <dsr/dsr_var.h>
#include <dsr/dsr_pktinfo.h>
#include <dsr/dsr_output.h>
#include <dsr/dsr_mbuf.h>
#endif

#if MONARCH_ADHOC

/* The sysctl's have been commented out.
 * If needed, will also have to define IPCTL_HOMEAGENT,
 * IPCTL_FOREIGNAGENT, IPCTL_LIMITED_BROADCAST_TO_ALL_IFNETS in in.h
 * -- Amit
 */
//int ip_homeagent = 0;
//SYSCTL_INT(_net_inet_ip, IPCTL_HOMEAGENT, homeagent,
//     CTLFLAG_RW, &ip_homeagent, 0, "");

int ip_foreignagent = 0;
//SYSCTL_INT(_net_inet_ip, IPCTL_FOREIGNAGENT, foreignagent,
//     CTLFLAG_RW, &ip_foreignagent, 0, "");

int ip_limited_broadcast_to_all_ifnets = 0;
//SYSCTL_INT(_net_inet_ip, IPCTL_LIMITED_BROADCAST_TO_ALL_IFNETS,
//     limited_broadcast_to_all_ifnets,
//     CTLFLAG_RW, &ip_limited_broadcast_to_all_ifnets, 0, "");

#endif /* MONARCH_ADHOC */

#ifdef DIAGNOSTIC
int ipprintfs = 0;
#endif

extern struct domain inetdomain;
extern const struct protosw inetsw[];
u_char ip_protox[IPPROTO_MAX];
struct in_ifaddr *_in_ifaddr;  /* first inet address */

#if MONARCH_ADHOC || !_MINIMIZE_NET
/*
 * We need to save the IP options in case a protocol wants to respond
 * to an incoming packet over the same route if the packet got here
 * using IP source routing.  This allows connection establishment and
 * maintenance when the remote end is on a network that is not known
 * to us.
 */
int ip_nhops = 0;
static struct ip_srcrt {
    struct in_addr dst;             /* final destination */
    char nop;                       /* one NOP to align */
    char srcopt[IPOPT_OFFSET + 1];  /* OPTVAL, OLEN and OFFSET */
    struct in_addr route[MAX_IPOPTLEN/sizeof(struct in_addr)];
} ip_srcrt;
#endif

static void save_rte(u_char *, struct in_addr);
/*
 * IP initialization: fill in IP protocol switch table.
 * All protocols not implemented in kernel go to raw IP protocol handler.
 */
void
ip_init(void)
{
    const struct protosw *pr;
    int i;

    pr = pffindproto(PF_INET, IPPROTO_RAW, SOCK_RAW);
    if (pr == 0)
        panic(_T("ip_init"));
    for (i = 0; i < IPPROTO_MAX; i++)
        ip_protox[i] = (u_char)(pr - inetsw);
    for (pr = inetdomain.dom_protosw;
        pr < inetdomain.dom_protoswNPROTOSW; pr++)
        if (pr->pr_domain->dom_family == PF_INET &&
            pr->pr_protocol && pr->pr_protocol != IPPROTO_RAW)
            ip_protox[pr->pr_protocol] = (u_char)(pr - inetsw);
    ipq.next = ipq.prev = &ipq;
    ip_id = (u_short)(tv_sec(time()) & 0xffff);
#ifdef GATEWAY
    i = (if_index + 1) * (if_index + 1) * sizeof (u_long);
    ip_ifmatrix = (u_long *) malloc(i, M_RTABLE, M_WAITOK);
    bzero((char *)ip_ifmatrix, i);
#endif
}

struct sockaddr_in ipaddr = { sizeof(ipaddr), AF_INET };

/*
 * Ip input routine.  Checksum and byte swap header.  If fragmented
 * try to reassemble.  Process options.  Pass to next level.
 */
void
ip_input(struct mbuf *m)
{
    struct ip *ip;
    struct ipq *fp;
    struct in_ifaddr *ia;
    INT32 hlen;

    NetLock();

#ifdef DIAGNOSTIC
    if ((m->m_flags & M_PKTHDR) == 0)
        panic(_T("ip_input: buffer chain has no packet header!\n"));
#endif

    _COLLECT_STAT(ipstat.ips_total++);

    if (m->m_len < sizeof (struct ip) &&
        (m = m_pullup(m, sizeof (struct ip))) == 0) {
        _COLLECT_STAT(ipstat.ips_toosmall++);
        goto done;
    }
    ip = mtod(m, struct ip *);
    if (ip->ip_v != IPVERSION) {
        _COLLECT_STAT(ipstat.ips_badvers++);
        goto bad;
    }
    hlen = ip->ip_hl << 2;
    if (hlen < sizeof(struct ip)) {    /* minimum header length */
        _COLLECT_STAT(ipstat.ips_badhlen++);
        goto bad;
    }
    if (hlen > m->m_len) {
        if ((m = m_pullup(m, hlen)) == 0) {
            _COLLECT_STAT(ipstat.ips_badhlen++);
            goto done;
        }
        ip = mtod(m, struct ip *);
    }
    ip->ip_sum = (u_short) in_cksum(m, hlen);
    if (ip->ip_sum) {
        _COLLECT_STAT(ipstat.ips_badsum++);
        goto bad;
    }
    /*
     * Convert fields to host representation.
     */
    NTOHS(ip->ip_len);
    if (ip->ip_len < hlen) {
        _COLLECT_STAT(ipstat.ips_badlen++);
        goto bad;
    }
    NTOHS(ip->ip_id);
    NTOHS(ip->ip_off);

    /*
     * Check that the amount of data in the buffers
     * is as at least much as the IP header would have us expect.
     * Trim mbufs if longer than we expect.
     * Drop packet if shorter than we expect.
     */
    if (m->m_pkthdr.len < ip->ip_len) {
        _COLLECT_STAT(ipstat.ips_tooshort++);
        goto bad;
    }
    if (m->m_pkthdr.len > ip->ip_len) {
        if (m->m_len == m->m_pkthdr.len) {
            m->m_len = ip->ip_len;
            m->m_pkthdr.len = ip->ip_len;
        } else
            m_adj(m, ip->ip_len - m->m_pkthdr.len);
    }

    /*
     * Process options and, if not destined for us,
     * ship it on.  ip_dooptions returns 1 when an
     * error was detected (causing an icmp message
     * to be sent and the original packet to be freed).
     */
#if MONARCH_ADHOC || !_MINIMIZE_NET
    ip_nhops = 0;  /* for source routed packets */
#endif
    if (hlen > sizeof (struct ip) && ip_dooptions(m))
        goto done;

    /*
     * Check for DHCP case where an interface is receiving even
     * though it doesn't have an IP address yet.  If so, we're waiting
     * for DHCP packets to tell us our IP address.  We can't do anything
     * with any other packets, so drop anything that isn't UDP here
     * (UDP's input routine will drop anything that isn't DHCP).
     */
    if (m->m_pkthdr.rcvif->if_flags & IFF_NOADDR) {
        if (ip->ip_p == IPPROTO_UDP) {
            goto ours;
        } else {
            goto bad;
        }
    }

    /*
     * If no IP addresses have been set yet but the interfaces
     * are receiving, can't do anything with incoming packets yet.
     */
    if (_in_ifaddr == NULL)
        goto bad;

    /*
     * Check our list of addresses, to see if the packet is for us.
     */
    for (ia = _in_ifaddr; ia; ia = ia->ia_next) {
#define satosin(sa) ((struct sockaddr_in *)(sa))

        if (IA_SIN(ia)->sin_addr.s_addr == ip->ip_dst.s_addr)
            goto ours;
        if (
#ifdef DIRECTED_BROADCAST
                ia->ia_ifp == m->m_pkthdr.rcvif &&
#endif
                (ia->ia_ifp->if_flags & IFF_BROADCAST)) {
            u_long t;
            if (satosin(&ia->ia_broadaddr)->sin_addr.s_addr ==
                    ip->ip_dst.s_addr)
                goto ours;
            if (ip->ip_dst.s_addr == ia->ia_netbroadcast.s_addr)
                goto ours;
            /*
             * Look for all-0's host part (old broadcast addr),
             * either for subnet or net.
             */
            t = ntohl(ip->ip_dst.s_addr);
            if (t == ia->ia_subnet)
                goto ours;
            if (t == ia->ia_net)
                goto ours;
        }
    }

    if (IN_MULTICAST(ntohl(ip->ip_dst.s_addr))) {
        struct in_multi *inm;

        /*
         * See if we belong to the destination multicast group on the
         * arrival interface.
         */
#if MONARCH_ADHOC
        if (m->m_pkthdr.rcvif->if_flags & IFF_ADHOC) {
            IN_LOOKUP_MULTI(ip->ip_dst, &dsrif[0], inm);
        } else {
            IN_LOOKUP_MULTI(ip->ip_dst, m->m_pkthdr.rcvif, inm);
        }
#else
        IN_LOOKUP_MULTI(ip->ip_dst, m->m_pkthdr.rcvif, inm);
#endif
        if (inm == NULL) {
            _COLLECT_STAT(ipstat.ips_cantforward++);
            m_freem(m);
            goto done;
        }
        goto ours;
    }
    if (ip->ip_dst.s_addr == (u_long)INADDR_BROADCAST)
        goto ours;
    if (ip->ip_dst.s_addr == INADDR_ANY)
        goto ours;

#if MONARCH_ADHOC
    /*
     *  If we've overheard a packet on a DSR interface and
     *  we are operating in DSR mode, do some snooping...
     */
    if((dsrif[0].if_flags & IFF_UP) &&
       (m->m_pkthdr.rcvif->if_flags & IFF_ADHOC)) {

        dsr_promisc_input(m);
        /* frees the packet */

        goto done;
    }
#endif /* MONARCH_ADHOC */

    /*
     * Not for us; forward if possible and desirable.
     */
    _COLLECT_STAT(ipstat.ips_cantforward++);
    m_freem(m);
    goto done;

ours:
    /*
     * If offset or IP_MF are set, must reassemble.
     * Otherwise, nothing need be done.
     * (We could look in the reassembly queue to see
     * if the packet was previously fragmented,
     * but it's not worth the time; just let them time out.)
     */

#if MONARCH_ADHOC
    /* For DSR packets we never try reassembly,
     * because the DSR headers need to be removed
     * before that is done and also DSR wants to send
     * ACKs for each fragment (atleast at the last hop,
     * when ACKs are explicitly requested)
     * -- Amit
     */
    if ((check_for_dsr(&m) == 0) &&
        (ip->ip_off &~ IP_DF)) {
#else
    if (ip->ip_off &~ IP_DF) {
#endif /* MONARCH_ADHOC */
        if (m->m_flags & M_EXT) {        /* XXX */
            if ((m = m_pullup(m, sizeof (struct ip))) == 0) {
                _COLLECT_STAT(ipstat.ips_toosmall++);
                goto done;
            }
            ip = mtod(m, struct ip *);
        }
        /*
         * Look for queue of fragments
         * of this datagram.
         */
        for (fp = ipq.next; fp != &ipq; fp = fp->next)
            if (ip->ip_id == fp->ipq_id &&
                ip->ip_src.s_addr == fp->ipq_src.s_addr &&
                ip->ip_dst.s_addr == fp->ipq_dst.s_addr &&
                ip->ip_p == fp->ipq_p)
                goto found;
        fp = 0;
found:

        /*
         * Adjust ip_len to not reflect header,
         * set low bit of ip_tos if more fragments are expected,
         * convert offset of this to bytes.
         */
        ip->ip_len = (u_short)(ip->ip_len - hlen);
        ip->ip_tos &= ~1;
        if (ip->ip_off & IP_MF)
            ip->ip_tos |= 1;
        ip->ip_off <<= 3;

        /*
         * If datagram marked as having more fragments
         * or if this is not the first fragment,
         * attempt reassembly; if it succeeds, proceed.
         */
        if ((ip->ip_tos & 1) || ip->ip_off) {
            _COLLECT_STAT(ipstat.ips_fragments++);
            m = ip_reass(m, fp);
            if (m == 0)
                goto done;
            _COLLECT_STAT(ipstat.ips_reassembled++);
            ip = mtod(m, struct ip *);
        } else
            if (fp)
                ip_freef(fp);
    } else
        ip->ip_len = (u_short)(ip->ip_len - hlen);

#if MONARCH_ADHOC
    /* XXX this is a temporary hack only until we have the 
       rest of the DSR code working -dam 10/27/98 */
    if(ip->ip_dst.s_addr != INADDR_BROADCAST) {
        m->m_flags &= ~M_BCAST;
    }
#endif /* MONARCH_ADHOC */
    /*
     * Switch out to protocol's input routine.
     */
    _COLLECT_STAT(ipstat.ips_delivered++);

#if MONARCH_ADHOC
    switch(ip->ip_p) {
    case IP6_NHDR_HOP:
    case IP6_NHDR_RT:
    case IP6_NHDR_DOPT:
    case IP6_NHDR_NONH:
        dsrstat.dsrs_received++;
        /*
         *  We've got a DSR packet.  DSR must be "on" and
         *  the packet must have been received over an
         *  IFF_ADHOC interface.
         */
        if ((dsrif[0].if_flags & IFF_UP) == 0)
            goto bad;
        if ((m->m_pkthdr.rcvif->if_flags & IFF_ADHOC) == 0)
            goto bad;

        (*inetsw[ip_protox[ip->ip_p]].pr_input)(m, 0);

        break;
    default:
        /*
         *  I'm the last hop, send an explicit acknowledgment.
         */
        if((dsrif[0].if_flags & IFF_UP) &&
           (m->m_pkthdr.rcvif->if_flags & IFF_ADHOC))
            dsr_send_ack(m);

        (*inetsw[ip_protox[ip->ip_p]].pr_input)(m, hlen);
    }
#else
    (*inetsw[ip_protox[ip->ip_p]].pr_input)(m, hlen);
#endif /* MONARCH_ADHOC */

done:
    NetUnlock();
    return;
bad:
    m_freem(m);
    NetUnlock();
    return;
}

/*
 * Take incoming datagram fragment and try to
 * reassemble it into whole datagram.  If a chain for
 * reassembly of this datagram already exists, then it
 * is given as fp; otherwise have to make a chain.
 */
struct mbuf *
ip_reass(struct mbuf *m, struct ipq *fp)
{
    struct ip *ip = mtod(m, struct ip*);
    struct ip *q;
    struct mbuf *n, *prev, *nxt;
    INT32 i;
    UINT32 next_offset;

    /* Terminate chain
     */
    m->m_nextpkt = NULL;

    /*
     * If first fragment to arrive, create a reassembly queue.
     */
    if (fp == 0) {
        if ((fp = bsd_malloc(sizeof(*fp), M_FTABLE, M_NOWAIT)) == NULL)
            goto dropfrag;
        fp->ipq_ttl = IP_FRAGTTL;
        fp->ipq_p = ip->ip_p;
        fp->ipq_id = ip->ip_id;
        fp->ipq_src = ((struct ip *)ip)->ip_src;
        fp->ipq_dst = ((struct ip *)ip)->ip_dst;
        insque((MYQUEUE*)fp, (MYQUEUE*)&ipq);
        fp->ipq_first = fp->ipq_last = m;
        /* recheck: should we check we are done ? */
        return NULL;
    }

    assert(fp->ipq_first != NULL);
    assert(fp->ipq_last != NULL);

    /* A new fragment for this queue.
     */
    nxt = NULL;
    prev = fp->ipq_last;
    q = mtod(prev,struct ip *);
    if ((q->ip_off + q->ip_len > ip->ip_off)) {

        /* Find the first segment 'nxt' that starts after this one.
         * Also, the last segment 'prev' which starts not-after this one.
         * If they exist.
         */
        prev = nxt = NULL;
        n = fp->ipq_first;
        for (;n != NULL; ) {
            q = mtod(n,struct ip*);
            if (q->ip_off > ip->ip_off) {
                nxt = n;
                break;
            }
            prev = n;
            n = n->m_nextpkt;
        }

        /* If there is a preceding segment, it may provide some of
         * our data already.  If so, drop the excess data from that
         * segment.  If it provides all of our data, drop us.
         */
        if (prev != NULL) {
            q = mtod(prev,struct ip*);
            i = (INT32)(q->ip_off + q->ip_len)
                 - ip->ip_off;
            if (i > 0) {
                if (i >= ip->ip_len)
                    goto dropfrag;
                m_adj(prev,-i);
                q->ip_len -= (UINT16) i;
            }
            /* There is a prev, ipq_first wont change
             */
            prev->m_nextpkt = m;
        } else {
            /* There is no prev, so we are first (and we cant be last) */
            fp->ipq_first = m;
        }

        /* While we overlap succeeding segments trim us or,
         * if they are completely covered, dequeue them.
         */
        if (nxt != NULL) {
            q = mtod(nxt,struct ip*);
            i = (INT32)(ip->ip_off + ip->ip_len) - q->ip_off;
            while (i > 0) {
                /* partial overlap ? trim & stop
                 */
                if (i < q->ip_len) {
                    m_adj(m,-i);
                    ip->ip_len -= (UINT16) i;
                    break;
                }
                /* full overlap ? drop & continue
                 */
                n = nxt;
                nxt = n->m_nextpkt;
                if (fp->ipq_last == n) fp->ipq_last = nxt;
                m_freem(n);
                if (nxt == NULL)
                    break;
                q = mtod(nxt,struct ip*);
                i = (INT32)(ip->ip_off + ip->ip_len) - q->ip_off;
            }
        }
        m->m_nextpkt = nxt;
        if (nxt == NULL) {
            /* there is no next, so we are last */
            fp->ipq_last = m;
        }

    } else {
        /* common case: next frag in order */
        /* also, the one after a missed one */
        prev->m_nextpkt = m;
    }

    /*
     * Stick new segment in its place;
     * check for complete reassembly.
     * 'prev' points to the frag that preceeds m, if any.
     */
    if (fp->ipq_last == prev)
        fp->ipq_last = m;

    /* Have we got all the fragments yet?
     */
    q = mtod(fp->ipq_last,struct ip *);
    if (q->ip_tos & 1)
        return (0);
    next_offset = 0;
    for (n = fp->ipq_first; n != NULL; n = n->m_nextpkt) {
        q = mtod(n,struct ip *);
        if (q->ip_off != next_offset)
            return (0);
        next_offset += q->ip_len;
    }

    /* Reassembly is complete; concatenate mbuf chains into one.
     */
    m = fp->ipq_first;
    n = m->m_next;
    m->m_next = 0;
    m_cat(m, n); /* a close friend of m_pullup */
    for (n = m->m_nextpkt; n != NULL; n = nxt) {
        /* drop ip header now */
        q = mtod(n,struct ip *);
        i = q->ip_hl << 2;
        n->m_data += i;
        n->m_len -= i;
        nxt = n->m_nextpkt;     /* get next before freeing n. */
        m_cat(m, n);
    }
    m->m_nextpkt = NULL;

    /* Create header for new ip packet by
     * modifying header of first packet;
     * dequeue and discard fragment reassembly header.
     */
    ip = mtod(m,struct ip *);
    ip->ip_len = (u_short)next_offset;
    ip->ip_tos &= ~1;
    ip->ip_src = fp->ipq_src;
    ip->ip_dst = fp->ipq_dst;

    remque((MYQUEUE*)fp);
    (void) bsd_free(fp, M_FTABLE);

    m->m_flags |= M_PKTHDR;
    m->m_pkthdr.len = next_offset + (ip->ip_hl << 2);
    return m;

dropfrag:
    _COLLECT_STAT(ipstat.ips_fragdropped++);
    m_freem(m);
    return (0);
}

/*
 * Free a fragment reassembly header and all
 * associated datagrams.
 */
void
ip_freef(struct ipq *fp)
{
    struct mbuf *q, *p;

    for (q = fp->ipq_first; q != NULL; q = p) {
        p = q->m_nextpkt;
        m_freem(q);
    }

    remque((MYQUEUE*)fp);
    (void) bsd_free(fp, M_FTABLE);
}

/*
 * IP timer processing;
 * if a timer expires on a reassembly
 * queue, discard it.
 */
void
ip_slowtimo(void)
{
    struct ipq *fp;

    NetLock();
    fp = ipq.next;
    if (fp == 0) {
        NetUnlock();
        return;
    }
    while (fp != &ipq) {
        --fp->ipq_ttl;
        fp = fp->next;
        if (fp->prev->ipq_ttl == 0) {
            _COLLECT_STAT(ipstat.ips_fragtimeout++);
            ip_freef(fp->prev);
        }
    }
    NetUnlock();
}

/*
 * System running low on mbufs; try to free some up by dropping all the
 * fragments belonging to the oldest packet on the reassembly list.
 */
void
ip_drain(void)
{
    struct ipq *pFrag, *pOldest;

    pFrag = ipq.next;
    if (pFrag == 0)
        return;

    pOldest = pFrag;
    while (pFrag != &ipq) {
        if (pOldest->ipq_ttl > pFrag->ipq_ttl)
            pOldest = pFrag;
        pFrag = pFrag->next;
    }

    if (pOldest != &ipq) {
        _COLLECT_STAT(ipstat.ips_fragdropped++);
        ip_freef(pOldest);
    }
}

/*
 * Do option processing on a datagram,
 * possibly discarding it if bad options are encountered,
 * or forwarding it if source-routed.
 * Returns 1 if packet has been forwarded/freed,
 * 0 if the packet should be processed further.
 */
INT
ip_dooptions(struct mbuf *m)
{
    struct ip *ip = mtod(m, struct ip *);
    u_char *cp;
    struct ip_timestamp *ipt;
    struct in_ifaddr *ia;
    INT opt, optlen, cnt, code, type = ICMP_PARAMPROB, forward = 0;
    struct in_addr *sin, dst;
    n_time ntime;

    dst = ip->ip_dst;
    cp = (u_char *)(ip + 1);
    cnt = (ip->ip_hl << 2) - sizeof (struct ip);
    for (; cnt > 0; cnt -= optlen, cp += optlen) {
        opt = cp[IPOPT_OPTVAL];
        if (opt == IPOPT_EOL)
            break;
        if (opt == IPOPT_NOP)
            optlen = 1;
        else {
            optlen = cp[IPOPT_OLEN];
            if (optlen <= 0 || optlen > cnt) {
                code = &cp[IPOPT_OLEN] - (u_char *)ip;
                goto bad;
            }
        }
        switch (opt) {

        default:
            break;

#if 0
#if MONARCH_ADHOC || !_MINIMIZE_NET
        /*
         * Source routing with record.
         * Find interface with current destination address.
         * If none on this machine then drop if strictly routed,
         * or do nothing if loosely routed.
         * Record interface address and bring up next address
         * component.  If strictly routed make sure next
         * address is on directly accessible net.
         */
        case IPOPT_LSRR:
        case IPOPT_SSRR:
            if ((off = cp[IPOPT_OFFSET]) < IPOPT_MINOFF) {
                code = &cp[IPOPT_OFFSET] - (u_char *)ip;
                goto bad;
            }
            ipaddr.sin_addr = ip->ip_dst;
            ia = (struct in_ifaddr *)
                  ifa_ifwithaddr((struct sockaddr *)&ipaddr);
            if (ia == 0) {
                if (opt == IPOPT_SSRR) {
                    type = ICMP_UNREACH;
                    code = ICMP_UNREACH_SRCFAIL;
                    goto bad;
                }
                /*
                 * Loose routing, and not at next destination
                 * yet; nothing to do except forward.
                 */
                break;
            }
            off--;  /* 0 origin */
            if (off > (int)(optlen - sizeof(struct in_addr))) {
                /*
                 * End of source route.  Should be for us.
                 */
                save_rte(cp, ip->ip_src);
                break;
            }
            /*
             * locate outgoing interface
             */
            bcopy((caddr_t)(cp + off), (caddr_t)&ipaddr.sin_addr,
                sizeof(ipaddr.sin_addr));
            if (opt == IPOPT_SSRR) {
                if ((ia = (struct in_ifaddr *)
                        ifa_ifwithdstaddr((struct sockaddr *)&ipaddr)) == 0)
                    ia = (struct in_ifaddr *)
                          ifa_ifwithnet((struct sockaddr *)&ipaddr);
            } else
                ia = ip_rtaddr(ipaddr.sin_addr);
            if (ia == 0) {
                type = ICMP_UNREACH;
                code = ICMP_UNREACH_SRCFAIL;
                goto bad;
            }
            ip->ip_dst = ipaddr.sin_addr;
            bcopy((caddr_t)&(IA_SIN(ia)->sin_addr),
                (caddr_t)(cp + off), sizeof(struct in_addr));
            cp[IPOPT_OFFSET] += sizeof(struct in_addr);
            /*
             * Let ip_input's mcast routing check handle mcast pkts
             */
            forward = !IN_MULTICAST(ntohl(ip->ip_dst.s_addr));
            break;
#endif
        case IPOPT_RR:
            if ((off = cp[IPOPT_OFFSET]) < IPOPT_MINOFF) {
                code = &cp[IPOPT_OFFSET] - (u_char *)ip;
                goto bad;
            }
            /*
             * If no space remains, ignore.
             */
            off--;            /* 0 origin */
            if (off > (INT)(optlen - sizeof(struct in_addr)))
                break;
            bcopy((caddr_t)(&ip->ip_dst), (caddr_t)&ipaddr.sin_addr,
                sizeof(ipaddr.sin_addr));
            /*
             * locate outgoing interface; if we're the destination,
             * use the incoming interface (should be same).
             */
            if ((ia = (struct in_ifaddr *)
                    ifa_ifwithaddr((struct sockaddr *)&ipaddr)) == 0 &&
                    (ia = ip_rtaddr(ipaddr.sin_addr)) == 0) {
                type = ICMP_UNREACH;
                code = ICMP_UNREACH_HOST;
                goto bad;
            }
            bcopy((caddr_t)&(IA_SIN(ia)->sin_addr),
                (caddr_t)(cp + off), sizeof(struct in_addr));
            cp[IPOPT_OFFSET] += sizeof(struct in_addr);
            break;
#endif

        case IPOPT_TS:
            code = cp - (u_char *)ip;
            ipt = (struct ip_timestamp *)cp;
            if (ipt->ipt_len < 5)
                goto bad;
            if (ipt->ipt_ptr > ipt->ipt_len - sizeof (long)) {
                if (++ipt->ipt_oflw == 0)
                    goto bad;
                break;
            }
            sin = (struct in_addr *)(cp + ipt->ipt_ptr - 1);
            switch (ipt->ipt_flg) {

            case IPOPT_TS_TSONLY:
                break;

            case IPOPT_TS_TSANDADDR:
                if (ipt->ipt_ptr + sizeof(n_time) +
                    sizeof(struct in_addr) > ipt->ipt_len)
                    goto bad;
                ipaddr.sin_addr = dst;
                ia = (struct in_ifaddr *)
                        ifaof_ifpforaddr((struct sockaddr *)&ipaddr,
                        m->m_pkthdr.rcvif);
                if (ia == 0)
                    continue;
                bcopy((caddr_t)&IA_SIN(ia)->sin_addr,
                    (caddr_t)sin, sizeof(struct in_addr));
                ipt->ipt_ptr += sizeof(struct in_addr);
                break;

            case IPOPT_TS_PRESPEC:
                if (ipt->ipt_ptr + sizeof(n_time) +
                    sizeof(struct in_addr) > ipt->ipt_len)
                    goto bad;
                bcopy((caddr_t)sin, (caddr_t)&ipaddr.sin_addr,
                    sizeof(struct in_addr));
                if (ifa_ifwithaddr((struct sockaddr *)&ipaddr) == 0)
                    continue;
                ipt->ipt_ptr += sizeof(struct in_addr);
                break;

            default:
                goto bad;
            }
            ntime = iptime();
            bcopy((caddr_t)&ntime, (caddr_t)cp + ipt->ipt_ptr - 1,
                sizeof(n_time));
            ipt->ipt_ptr += sizeof(n_time);
        }
    }
    return (0);
bad:
    /* XXX icmp_error adds in hdr length */
    ip->ip_len = (u_short)(ip->ip_len - (ip->ip_hl << 2));
    icmp_error(m, type, code, 0, 0);

    _COLLECT_STAT(ipstat.ips_badoptions++);
    return (1);
}

/* used only by source routing and record routing */
#if 0
/*
 * Given address of next destination (final or next hop),
 * return internet address info of interface to be used to get there.
 */
struct in_ifaddr *
ip_rtaddr(struct in_addr dst)
{
    struct ifaddr *ifa;

    if (rt_ifaout(dst, &ifa, NULL) != S_OK)
        return (struct in_ifaddr *)0;

    return (struct in_ifaddr *)ifa;
}
#endif

#if MONARCH_ADHOC || !_MINIMIZE_NET
/*
 * Save incoming source route for use in replies,
 * to be picked up later by ip_srcroute if the receiver is interested.
 */
static void
save_rte(u_char *option, struct in_addr dst)
{
    unsigned olen;

    olen = option[IPOPT_OLEN];
#ifdef DIAGNOSTIC
    if (ipprintfs)
        printf(_T("save_rte: olen %d\n"), olen);
#endif
    if (olen > sizeof(ip_srcrt) - (1 + sizeof(dst)))
        return;
    bcopy((caddr_t)option, (caddr_t)ip_srcrt.srcopt, olen);
    ip_nhops = (olen - IPOPT_OFFSET - 1) / sizeof(struct in_addr);
    ip_srcrt.dst = dst;
}

/*
 * Retrieve incoming source route for use in replies,
 * in the same form used by setsockopt.
 * The first hop is placed before the options, will be removed later.
 */
struct mbuf *
ip_srcroute(void)
{
    struct in_addr *p, *q;
    struct mbuf *m;

    if (ip_nhops == 0)
        return ((struct mbuf *)0);
    m = m_get(M_DONTWAIT, MT_SOOPTS);
    if (m == 0)
        return ((struct mbuf *)0);

#define OPTSIZ    (sizeof(ip_srcrt.nop) + sizeof(ip_srcrt.srcopt))

    /* length is (nhops+1)*sizeof(addr) + sizeof(nop + srcrt header) */
    m->m_len = ip_nhops * sizeof(struct in_addr) + sizeof(struct in_addr) +
        OPTSIZ;
#ifdef DIAGNOSTIC
    if (ipprintfs)
        printf(_T("ip_srcroute: nhops %d mlen %d"), ip_nhops, m->m_len);
#endif

    /*
     * First save first hop for return route
     */
    p = &ip_srcrt.route[ip_nhops - 1];
    *(mtod(m, struct in_addr *)) = *p--;
#ifdef DIAGNOSTIC
    if (ipprintfs)
        printf(_T(" hops %lx"), ntohl(mtod(m, struct in_addr *)->s_addr));
#endif

    /*
     * Copy option fields and padding (nop) to mbuf.
     */
    ip_srcrt.nop = IPOPT_NOP;
    ip_srcrt.srcopt[IPOPT_OFFSET] = IPOPT_MINOFF;
    bcopy((caddr_t)&ip_srcrt.nop,
        mtod(m, caddr_t) + sizeof(struct in_addr), OPTSIZ);
    q = (struct in_addr *)(mtod(m, caddr_t) + sizeof(struct in_addr) + OPTSIZ);
#undef OPTSIZ
    /*
     * Record return path as an IP source route,
     * reversing the path (pointers are now aligned).
     */
    while (p >= ip_srcrt.route) {
#ifdef DIAGNOSTIC
        if (ipprintfs)
            printf(_T(" %lx"), ntohl(q->s_addr));
#endif
        *q++ = *p--;
    }
    /*
     * Last hop goes to final destination.
     */
    *q = ip_srcrt.dst;
#ifdef DIAGNOSTIC
    if (ipprintfs)
        printf(_T(" %lx\n"), ntohl(q->s_addr));
#endif
    return (m);
}
#endif

/*
 * Strip out IP options, at higher
 * level protocol in the kernel.
 * Second argument is buffer to which options
 * will be moved, and return value is their length.
 * XXX should be deleted; last arg currently ignored.
 */
void
ip_stripoptions(struct mbuf *m, struct mbuf *mopt)
{
    int i;
    struct ip *ip = mtod(m, struct ip *);
    caddr_t opts;
    INT olen;

    olen = (ip->ip_hl<<2) - sizeof (struct ip);
    opts = (caddr_t)(ip + 1);
    i = m->m_len - (sizeof (struct ip) + olen);
    bcopy(opts  + olen, opts, (unsigned)i);
    m->m_len -= olen;
    if (m->m_flags & M_PKTHDR)
        m->m_pkthdr.len -= olen;
    ip->ip_hl = sizeof(struct ip) >> 2;
}

u_short inetctlerrmap[PRC_NCMDS] = {
    0,            0,            0,            0,
    0,            EMSGSIZE,     EHOSTDOWN,    EHOSTUNREACH,
    EHOSTUNREACH, EHOSTUNREACH, ECONNREFUSED, ECONNREFUSED,
    EMSGSIZE,     EHOSTUNREACH, 0,            0,
    0,            0,            0,            0,
    ENOPROTOOPT
};
