/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1988, 1990, 1993
 *      Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      @(#)ip_output.c 8.1 (Berkeley) 6/10/93
 */

#include <sock.h>

#include <if.h>
#include <route.h>

#include <in.h>
#include <in_systm.h>
#include <ip.h>
#include <in_pcb.h>
#include <in_var.h>
#include <ip_var.h>
#include <if_types.h>

#if MONARCH_ADHOC
#include <diagnostics.h>
#include <net/dsr/dsr_mbuf.h>
#ifndef senderr
#define senderr(e) { error = (e); goto bad;}
#endif

#endif

static struct mbuf *ip_insertoptions (struct mbuf *, struct mbuf *, INT *);
static void ip_mloopback
        (struct ifnet *, struct mbuf *, struct sockaddr_in *);

#if MONARCH_ADHOC
/*
 * RFC 1122, Section 3.3.6
 *      There has been discussion on whether a datagram addressed
 * to the Limited Broadcast address ought to be sent from all
 * the interfaces of a multihomed host.  This specification
 * takes no stand on the issue.
 *
 * Well, we have decided to take a stand on this issue.  This
 * stack now sends packets addressed to the limited broadcast
 * out all network interfaces.
 */
extern int ip_limited_broadcast_to_all_ifnets;

int
ip_output_limited_broadcast_to_all_ifnets(struct mbuf *m,
                      const struct route *ro,
                      const int flags,
                      const struct sockaddr_in *dst);
#endif /* MONARCH_ADHOC */

/*
 * IP output.  The packet in mbuf chain m contains a skeletal IP
 * header (with len, off, ttl, proto, tos, src, dst).
 * The mbuf chain containing the packet will be freed.
 * The mbuf opt, if present, will not be freed.
 */
INT
ip_output(
        struct mbuf *m0,
        struct mbuf *opt,
        struct route_in *rti,
        INT flags,
        struct ip_moptions *imo)
{
        struct ip *ip, *mhip;
        struct ifnet *ifp;
        struct mbuf *m = m0;
        INT hlen = sizeof (struct ip);
        INT len, off, error = 0;
        struct sockaddr_in *dst;
	struct sockaddr_in fakedst;
        struct in_ifaddr *ia;

#ifdef  DIAGNOSTIC
        if ((m->m_flags & M_PKTHDR) == 0)
                panic("ip_output no HDR");
#endif
        if (opt) {
                m = ip_insertoptions(m, opt, &len);
                hlen = len;
        }
        ip = mtod(m, struct ip *);
        /*
         * Fill in IP header.
         */
        if ((flags & (IP_RAWOUTPUT)) == 0) {
                ip->ip_v = IPVERSION;
                ip->ip_off &= IP_DF;
                ip->ip_id = htons(ip_id++);
                ip->ip_hl = (u_char)(hlen >> 2);
                _COLLECT_STAT(ipstat.ips_localout++);
        } else {
                hlen = ip->ip_hl << 2;
        }

        if (rti->rt_valid == 0)
                goto bad;
        dst = &rti->rt_gw;

        ia = (struct in_ifaddr *)rti->rt_ifa->ifa_ifp->if_addrlist; /* XXX */
        ifp = rti->rt_ifa->ifa_ifp;

        /*
         * Check for IP multicast.  XXX: this is bad form
         */
        memset(&fakedst, 0, sizeof(fakedst));
        fakedst.sin_family = AF_INET;
        fakedst.sin_len = sizeof(struct sockaddr_in);
        fakedst.sin_addr = ip->ip_dst;

        if (IN_MULTICAST(ntohl(fakedst.sin_addr.s_addr))) {
                struct in_multi *inm;
                extern struct ifnet loif;

                m->m_flags |= M_MCAST;

                dst = &fakedst;
                /*
                 * See if the caller provided any multicast options
                 */
                if (imo != NULL) {
                        ip->ip_ttl = imo->imo_multicast_ttl;
                        if (imo->imo_multicast_ifp != NULL)
                                ifp = imo->imo_multicast_ifp;
                } else
                        ip->ip_ttl = IP_DEFAULT_MULTICAST_TTL;
                /*
                 * Confirm that the outgoing interface supports multicast.
                 */
                if ((ifp->if_flags & IFF_MULTICAST) == 0) {
                        _COLLECT_STAT(ipstat.ips_noroute++);
                        error = ENETUNREACH;
                        goto bad;
                }
                /*
                 * If source address not specified yet, use address
                 * of outgoing interface.
                 */
                if (ip->ip_src.s_addr == INADDR_ANY) {
                        struct in_ifaddr *ia;

                        for (ia = _in_ifaddr; ia; ia = ia->ia_next)
                                if (ia->ia_ifp == ifp) {
                                        ip->ip_src = IA_SIN(ia)->sin_addr;
                                        break;
                                }
                }

                IN_LOOKUP_MULTI(ip->ip_dst, ifp, inm);
                if (inm != NULL &&
                   (imo == NULL || imo->imo_multicast_loop)) {
                        /*
                         * If we belong to the destination multicast group
                         * on the outgoing interface, and the caller did not
                         * forbid loopback, loop back a copy.
                         */
                        ip_mloopback(ifp, m, dst);
                }
                /*
                 * Multicasts with a time-to-live of zero may be looped-
                 * back, above, but must not be transmitted on a network.
                 * Also, multicasts addressed to the loopback interface
                 * are not sent -- the above call to ip_mloopback() will
                 * loop back a copy if this host actually belongs to the
                 * destination group on the loopback interface.
                 */
                if (ip->ip_ttl == 0 || ifp == &loif) {
                        m_freem(m);
                        goto done;
                }

                goto sendit;
        }

        /*
         * If source address not specified yet, use address
         * of outgoing interface.
         */
#if 0
        if (ip->ip_src.s_addr == INADDR_ANY)
                ip->ip_src = IA_SIN(ia)->sin_addr;
#endif

#if MONARCH_ADHOC
        if(ip_limited_broadcast_to_all_ifnets &&
           ip->ip_dst.s_addr == INADDR_BROADCAST) {
            error = ip_output_limited_broadcast_to_all_ifnets(m, ro,
                                      flags, dst);
            goto done;
        }
#endif /* MONARCH_ADHOC */

        /*
         * Look for broadcast address and
         * and verify user is allowed to send
         * such a packet.
         */
        if (in_broadcast(dst->sin_addr, ifp)) {
#if MONARCH_ADHOC
        /*
         *  Let broadcast packets through to the DSR interface.
         *  Can't set IFF_BROADCAST on that interface, because
         *  ARPs are then sent to that interface.  This screws
         *  up routers that connect the adhoc network to the
         *  rest of the Internet. - jgb 11/27/98
         */
        if (ifp->if_type != IFT_DSR &&
            (ifp->if_flags & IFF_BROADCAST) == 0) {
#else
        if ((ifp->if_flags & IFF_BROADCAST) == 0) {
#endif /* MONARCH_ADHOC */
                        error = EADDRNOTAVAIL;
                        goto bad;
                }
                if ((flags & IP_ALLOWBROADCAST) == 0) {
                        error = EACCES;
                        goto bad;
                }
                /* don't allow broadcast messages to be fragmented */
                if (ip->ip_len > ifp->if_mtu) {
                        error = EMSGSIZE;
                        goto bad;
                }
                m->m_flags |= M_BCAST;
        } else
                m->m_flags &= ~M_BCAST;

 sendit:
        /*
         * If small enough for interface, can just send directly.
         */
        if (ip->ip_len <= ifp->if_mtu) {
                ip->ip_len = htons(ip->ip_len);
                ip->ip_off = htons(ip->ip_off);
                ip->ip_sum = 0;
                ip->ip_sum = (u_short) in_cksum(m, hlen);

                error = (*ifp->v->Output)(iIF(ifp), m,
                                (struct sockaddr *)0, rti);
                goto done;
        }
        /*
         * Too large for interface; fragment if possible.
         * Must be able to put at least 8 bytes per fragment.
         */
        if (ip->ip_off & IP_DF) {
                error = EMSGSIZE;
                _COLLECT_STAT(ipstat.ips_cantfrag++);
                goto bad;
        }
        len = (ifp->if_mtu - hlen) &~ 7;
        if (len < 8) {
                error = EMSGSIZE;
                goto bad;
        }

    {
        INT mhlen, firstlen = len;
        struct mbuf **mnext = &m->m_nextpkt;

        /*
         * Loop through length of segment after first fragment,
         * make new header and copy data of each part and link onto chain.
         */
        m0 = m;
        mhlen = sizeof (struct ip);
        for (off = hlen + len; off < ip->ip_len; off += len) {
                MGETHDR(m, M_DONTWAIT, MT_HEADER);
                if (m == 0) {
                        error = ENOBUFS;
                        _COLLECT_STAT(ipstat.ips_odropped++);
                        goto sendorfree;
                }
                m->m_data += max_linkhdr;
                mhip = mtod(m, struct ip *);
                *mhip = *ip;
                if (hlen > sizeof (struct ip)) {
                        mhlen = ip_optcopy(ip, mhip) + sizeof (struct ip);
                        mhip->ip_hl = (u_char)(mhlen >> 2);
                }
                m->m_len = mhlen;
                mhip->ip_off = (u_short)(((off - hlen) >> 3) + (ip->ip_off & ~IP_MF));
                if (ip->ip_off & IP_MF)
                        mhip->ip_off |= IP_MF;
                if (off + len >= ip->ip_len)
                        len = ip->ip_len - off;
                else
                        mhip->ip_off |= IP_MF;
                mhip->ip_len = htons((u_short)(len + mhlen));
                m->m_next = m_copy(m0, off, len);
                if (m->m_next == 0) {
                        (void) m_free(m);
                        error = ENOBUFS;        /* ??? */
                        _COLLECT_STAT(ipstat.ips_odropped++);
                        goto sendorfree;
                }
                m->m_pkthdr.len = mhlen + len;
                m->m_pkthdr.rcvif = (struct ifnet *)0;
                mhip->ip_off = htons((u_short)mhip->ip_off);
                mhip->ip_sum = 0;
                mhip->ip_sum = (u_short) in_cksum(m, mhlen);
                *mnext = m;
                mnext = &m->m_nextpkt;
                _COLLECT_STAT(ipstat.ips_ofragments++);
        }
        /*
         * Update first fragment by trimming what's been copied out
         * and updating header, then send each fragment (in order).
         */
        m = m0;
        m_adj(m, hlen + firstlen - ip->ip_len);
        m->m_pkthdr.len = hlen + firstlen;
        ip->ip_len = htons((u_short)m->m_pkthdr.len);
        ip->ip_off = htons((u_short)(ip->ip_off | IP_MF));
        ip->ip_sum = 0;
        ip->ip_sum = (u_short) in_cksum(m, hlen);
sendorfree:
        for (m = m0; m; m = m0) {
                m0 = m->m_nextpkt;
                m->m_nextpkt = 0;
                if (error == 0)
                        error = (*ifp->v->Output)(iIF(ifp), m,
                            (struct sockaddr *)dst, rti);
                else
                        m_freem(m);
        }

        if (error == 0)
                _COLLECT_STAT(ipstat.ips_fragmented++);
    }
done:
        return (error);
bad:
        m_freem(m0);
        goto done;
}

/*
 * Insert IP options into preformed packet.
 * Adjust IP destination as required for IP source routing,
 * as indicated by a non-zero in_addr at the start of the options.
 */
static struct mbuf *
ip_insertoptions(
        struct mbuf *m,
        struct mbuf *opt,
        INT *phlen)
{
        struct ipoption *p = mtod(opt, struct ipoption *);
        struct mbuf *n;
        struct ip *ip = mtod(m, struct ip *);
        unsigned optlen;

        optlen = opt->m_len - sizeof(p->ipopt_dst);
        if (optlen + ip->ip_len > IP_MAXPACKET)
                return (m);             /* XXX should fail */
        if (p->ipopt_dst.s_addr)
                ip->ip_dst = p->ipopt_dst;
        if (m->m_flags & M_EXT || (ADDRESS) m->m_data - optlen < (ADDRESS) m->m_pktdat) {
                MGETHDR(n, M_DONTWAIT, MT_HEADER);
                if (n == 0)
                        return (m);
                n->m_pkthdr.len = m->m_pkthdr.len + optlen;
                m->m_len -= sizeof(struct ip);
                m->m_data += sizeof(struct ip);
                n->m_next = m;
                m = n;
                m->m_len = optlen + sizeof(struct ip);
                m->m_data += max_linkhdr;
                bcopy((caddr_t)ip, mtod(m, caddr_t), sizeof(struct ip));
        } else {
                m->m_data -= optlen;
                m->m_len += optlen;
                m->m_pkthdr.len += optlen;
                ovbcopy((caddr_t)ip, mtod(m, caddr_t), sizeof(struct ip));
        }
        ip = mtod(m, struct ip *);
        bcopy((caddr_t)p->ipopt_list, (caddr_t)(ip + 1), (unsigned)optlen);
        *phlen = sizeof(struct ip) + optlen;
        ip->ip_len = (u_short)(ip->ip_len + optlen);
        return (m);
}

/*
 * Copy options from ip to jp,
 * omitting those not copied during fragmentation.
 */
INT
ip_optcopy(
        struct ip *ip,
        struct ip *jp)
{
        u_char *cp, *dp;
        INT opt, optlen, cnt;

        cp = (u_char *)(ip + 1);
        dp = (u_char *)(jp + 1);
        cnt = (ip->ip_hl << 2) - sizeof (struct ip);
        for (; cnt > 0; cnt -= optlen, cp += optlen) {
                opt = cp[0];
                if (opt == IPOPT_EOL)
                        break;
                if (opt == IPOPT_NOP) {
                        /* Preserve for IP mcast tunnel's LSRR alignment. */
                        *dp++ = IPOPT_NOP;
                        optlen = 1;
                        continue;
                } else
                        optlen = cp[IPOPT_OLEN];
                /* bogus lengths should have been caught by ip_dooptions */
                if (optlen > cnt)
                        optlen = cnt;
                if (IPOPT_COPIED(opt)) {
                        bcopy((caddr_t)cp, (caddr_t)dp, (unsigned)optlen);
                        dp += optlen;
                }
        }
        for (optlen = dp - (u_char *)(jp+1); optlen & 0x3; optlen++)
                *dp++ = IPOPT_EOL;
        return (optlen);
}

/*
 * IP socket option processing.
 */
INT
ip_ctloutput(
        INT op,
        struct socket *so,
        INT level,
        INT optname,
        struct mbuf **mp)
{
        struct inpcb *inp = sotoinpcb(so);
        struct mbuf *m = *mp;
        INT optval = 0;
        int error = 0;

        if (level != IPPROTO_IP)
                goto freeit;
        else switch (op) {

        case PRCO_SETOPT:
                switch (optname) {
                case IP_OPTIONS:
                        return (ip_pcbopts(&inp->inp_options, m));

                case IP_TOS:
                case IP_TTL:
                        if (m->m_len != sizeof(INT))
                                error = EINVAL;
                        else {
                                optval = *mtod(m, INT *);
                                switch (optname) {

#ifdef SUPPORT_IP_SOCKOPTS
                                case IP_TOS:
                                        inp->inp_ip.ip_tos = (u_char) optval;
                                        break;

                                case IP_TTL:
                                        inp->inp_ip.ip_ttl = (u_char) optval;
                                        break;
#endif
                                }
                        }
                        break;
#undef OPTSET

                case IP_MULTICAST_IF:
                case IP_MULTICAST_TTL:
                case IP_MULTICAST_LOOP:
                case IP_ADD_MEMBERSHIP:
                case IP_DROP_MEMBERSHIP:
                        error = ip_setmoptions(optname, &inp->inp_moptions, m);
                        break;

                freeit:
                default:
                        error = EINVAL;
                        break;
                }
                if (m)
                        (void)m_free(m);
                break;

        case PRCO_GETOPT:
                switch (optname) {
                case IP_OPTIONS:
                        *mp = m = m_get(M_WAIT, MT_SOOPTS);
                        if (inp->inp_options) {
                                m->m_len = inp->inp_options->m_len;
                                bcopy(mtod(inp->inp_options, caddr_t),
                                    mtod(m, caddr_t), (unsigned)m->m_len);
                        } else
                                m->m_len = 0;
                        break;

                case IP_TOS:
                case IP_TTL:
                        *mp = m = m_get(M_WAIT, MT_SOOPTS);
                        m->m_len = sizeof(INT);
                        switch (optname) {
#ifdef SUPPORT_IP_SOCKOPTS
                        case IP_TOS:
                                optval = inp->inp_ip.ip_tos;
                                break;

                        case IP_TTL:
                                optval = inp->inp_ip.ip_ttl;
                                break;
#endif
                        }
                        *mtod(m, INT *) = optval;
                        break;

                case IP_MULTICAST_IF:
                case IP_MULTICAST_TTL:
                case IP_MULTICAST_LOOP:
                case IP_ADD_MEMBERSHIP:
                case IP_DROP_MEMBERSHIP:
                        error = ip_getmoptions(optname, inp->inp_moptions, mp);
                        break;

                default:
                        error = EINVAL;
                        break;
                }
                break;
        }
        return (error);
}

/*
 * Set up IP options in pcb for insertion in output packets.
 * Store in mbuf with pointer in pcbopt, adding pseudo-option
 * with destination address if source routed.
 */
INT
#ifdef notyet
ip_pcbopts(INT optname, struct mbuf **pcbopt, struct mbuf *m)
#else
ip_pcbopts(struct mbuf **pcbopt, struct mbuf *m)
#endif
{
        INT cnt, optlen;
        u_char *cp;
        u_char opt;

        /* turn off any old options */
        if (*pcbopt)
                (void)m_free(*pcbopt);
        *pcbopt = 0;
        if (m == (struct mbuf *)0 || m->m_len == 0) {
                /*
                 * Only turning off any previous options.
                 */
                if (m)
                        (void)m_free(m);
                return (0);
        }

#if ! (defined(vax) || defined(i386))
        if (m->m_len % sizeof(long))
                goto bad;
#endif
        /*
         * IP first-hop destination address will be stored before
         * actual options; move other options back
         * and clear it when none present.
         */
        if ((ADDRESS) m->m_data + m->m_len + sizeof(struct in_addr)
            >= (ADDRESS) &m->m_dat[MLEN])
                goto bad;
        cnt = m->m_len;
        m->m_len += sizeof(struct in_addr);
        cp = mtod(m, u_char *) + sizeof(struct in_addr);
        ovbcopy(mtod(m, caddr_t), (caddr_t)cp, (unsigned)cnt);
        bzero(mtod(m, caddr_t), sizeof(struct in_addr));

        for (; cnt > 0; cnt -= optlen, cp += optlen) {
                opt = cp[IPOPT_OPTVAL];
                if (opt == IPOPT_EOL)
                        break;
                if (opt == IPOPT_NOP)
                        optlen = 1;
                else {
                        optlen = cp[IPOPT_OLEN];
                        if (optlen <= IPOPT_OLEN || optlen > cnt)
                                goto bad;
                }
                switch (opt) {

                default:
                        break;

                case IPOPT_LSRR:
                case IPOPT_SSRR:
                        /*
                         * user process specifies route as:
                         *      ->A->B->C->D
                         * D must be our final destination (but we can't
                         * check that since we may not have connected yet).
                         * A is first hop destination, which doesn't appear in
                         * actual IP option, but is stored before the options.
                         */
                        if (optlen < IPOPT_MINOFF - 1 + sizeof(struct in_addr))
                                goto bad;
                        m->m_len -= sizeof(struct in_addr);
                        cnt -= sizeof(struct in_addr);
                        optlen -= sizeof(struct in_addr);
                        cp[IPOPT_OLEN] = (u_char) optlen;
                        /*
                         * Move first hop before start of options.
                         */
                        bcopy((caddr_t)&cp[IPOPT_OFFSET+1], mtod(m, caddr_t),
                            sizeof(struct in_addr));
                        /*
                         * Then copy rest of options back
                         * to close up the deleted entry.
                         */
                        ovbcopy((caddr_t)(&cp[IPOPT_OFFSET+1] +
                            sizeof(struct in_addr)),
                            (caddr_t)&cp[IPOPT_OFFSET+1],
                            (unsigned)cnt + sizeof(struct in_addr));
                        break;
                }
        }
        if (m->m_len > MAX_IPOPTLEN + sizeof(struct in_addr))
                goto bad;
        *pcbopt = m;
        return (0);

bad:
        (void)m_free(m);
        return (EINVAL);
}

/*
 * Set the IP multicast options in response to user setsockopt().
 */
INT
ip_setmoptions(
        INT optname,
        struct ip_moptions **imop,
        struct mbuf *m)
{
        int error = 0;
        u_char loop;
        int i;
        struct in_addr addr;
        struct ip_mreq *mreq;
        struct ifnet *ifp;
        struct ip_moptions *imo = *imop;

        if (imo == NULL) {
                /*
                 * No multicast option buffer attached to the pcb;
                 * allocate one and initialize to default values.
                 */
                imo = (struct ip_moptions*)bsd_malloc(sizeof(*imo), M_IPMOPTS,
                    M_WAITOK);

                if (imo == NULL)
                        return (ENOBUFS);
                *imop = imo;
                imo->imo_multicast_ifp = NULL;
                imo->imo_multicast_ttl = IP_DEFAULT_MULTICAST_TTL;
                imo->imo_multicast_loop = IP_DEFAULT_MULTICAST_LOOP;
                imo->imo_num_memberships = 0;
        }

        switch (optname) {

        case IP_MULTICAST_IF:
                /*
                 * Select the interface for outgoing multicast packets.
                 */
                if (m == NULL || m->m_len != sizeof(struct in_addr)) {
                        error = EINVAL;
                        break;
                }
                addr = *(mtod(m, struct in_addr *));
                /*
                 * INADDR_ANY is used to remove a previous selection.
                 * When no interface is selected, a default one is
                 * chosen every time a multicast packet is sent.
                 */
                if (addr.s_addr == INADDR_ANY) {
                        imo->imo_multicast_ifp = NULL;
                        break;
                }
                /*
                 * The selected interface is identified by its local
                 * IP address.  Find the interface and confirm that
                 * it supports multicasting.
                 */
                INADDR_TO_IFP(addr, ifp);
                if (ifp == NULL || (ifp->if_flags & IFF_MULTICAST) == 0) {
                        error = EADDRNOTAVAIL;
                        break;
                }
                imo->imo_multicast_ifp = ifp;
                break;

        case IP_MULTICAST_TTL:
                /*
                 * Set the IP time-to-live for outgoing multicast packets.
                 */
                if (m == NULL || m->m_len != 1) {
                        error = EINVAL;
                        break;
                }
                imo->imo_multicast_ttl = *(mtod(m, u_char *));
                break;

        case IP_MULTICAST_LOOP:
                /*
                 * Set the loopback flag for outgoing multicast packets.
                 * Must be zero or one.
                 */
                if (m == NULL || m->m_len != 1 ||
                   (loop = *(mtod(m, u_char *))) > 1) {
                        error = EINVAL;
                        break;
                }
                imo->imo_multicast_loop = loop;
                break;

        case IP_ADD_MEMBERSHIP:
                /*
                 * Add a multicast group membership.
                 * Group must be a valid IP multicast address.
                 */
                if (m == NULL || m->m_len != sizeof(struct ip_mreq)) {
                        error = EINVAL;
                        break;
                }
                mreq = mtod(m, struct ip_mreq *);
                if (!IN_MULTICAST(ntohl(mreq->imr_multiaddr.s_addr))) {
                        error = EINVAL;
                        break;
                }
                /*
                 * If no interface address was provided, use the interface of
                 * the route to the given multicast address.
                 */
                if (mreq->imr_interface.s_addr == INADDR_ANY) {
			struct ifaddr *ifa;

                        rt_ifaout(mreq->imr_multiaddr, &ifa, 0);
			ifp = ifa->ifa_ifp;
                } else {
                        INADDR_TO_IFP(mreq->imr_interface, ifp);
		}

                /*
                 * See if we found an interface, and confirm that it
                 * supports multicast.
                 */
                if (ifp == NULL || (ifp->if_flags & IFF_MULTICAST) == 0) {
                        error = EADDRNOTAVAIL;
                        break;
                }
                /*
                 * See if the membership already exists or if all the
                 * membership slots are full.
                 */
                for (i = 0; i < imo->imo_num_memberships; ++i) {
                        if (imo->imo_membership[i]->inm_ifp == ifp &&
                            imo->imo_membership[i]->inm_addr.s_addr
                                                == mreq->imr_multiaddr.s_addr)
                                break;
                }
                if (i < imo->imo_num_memberships) {
                        error = EADDRINUSE;
                        break;
                }
                if (i == IP_MAX_MEMBERSHIPS) {
                        error = ETOOMANYREFS;
                        break;
                }
                /*
                 * Everything looks good; add a new record to the multicast
                 * address list for the given interface.
                 */
                if ((imo->imo_membership[i] =
                    in_addmulti(&mreq->imr_multiaddr, ifp)) == NULL) {
                        error = ENOBUFS;
                        break;
                }
                ++imo->imo_num_memberships;
                break;

        case IP_DROP_MEMBERSHIP:
                /*
                 * Drop a multicast group membership.
                 * Group must be a valid IP multicast address.
                 */
                if (m == NULL || m->m_len != sizeof(struct ip_mreq)) {
                        error = EINVAL;
                        break;
                }
                mreq = mtod(m, struct ip_mreq *);
                if (!IN_MULTICAST(ntohl(mreq->imr_multiaddr.s_addr))) {
                        error = EINVAL;
                        break;
                }
                /*
                 * If an interface address was specified, get a pointer
                 * to its ifnet structure.
                 */
                if (mreq->imr_interface.s_addr == INADDR_ANY)
                        ifp = NULL;
                else {
                        INADDR_TO_IFP(mreq->imr_interface, ifp);
                        if (ifp == NULL) {
                                error = EADDRNOTAVAIL;
                                break;
                        }
                }
                /*
                 * Find the membership in the membership array.
                 */
                for (i = 0; i < imo->imo_num_memberships; ++i) {
                        if ((ifp == NULL ||
                             imo->imo_membership[i]->inm_ifp == ifp) &&
                             imo->imo_membership[i]->inm_addr.s_addr ==
                             mreq->imr_multiaddr.s_addr)
                                break;
                }
                if (i == imo->imo_num_memberships) {
                        error = EADDRNOTAVAIL;
                        break;
                }
                /*
                 * Give up the multicast address record to which the
                 * membership points.
                 */
                in_delmulti(imo->imo_membership[i]);
                /*
                 * Remove the gap in the membership array.
                 */
                for (++i; i < imo->imo_num_memberships; ++i)
                        imo->imo_membership[i-1] = imo->imo_membership[i];
                --imo->imo_num_memberships;
                break;

        default:
                error = EOPNOTSUPP;
                break;
        }

        /*
         * If all options have default values, no need to keep the mbuf.
         */
        if (imo->imo_multicast_ifp == NULL &&
            imo->imo_multicast_ttl == IP_DEFAULT_MULTICAST_TTL &&
            imo->imo_multicast_loop == IP_DEFAULT_MULTICAST_LOOP &&
            imo->imo_num_memberships == 0) {
                bsd_free(*imop, M_IPMOPTS);
                *imop = NULL;
        }

        return (error);
}

/*
 * Return the IP multicast options in response to user getsockopt().
 */
INT
ip_getmoptions(
        INT optname,
        struct ip_moptions *imo,
        struct mbuf **mp)
{
        u_char *ttl;
        u_char *loop;
        struct in_addr *addr;
        struct in_ifaddr *ia;

        *mp = m_get(M_WAIT, MT_SOOPTS);

        switch (optname) {

        case IP_MULTICAST_IF:
                addr = mtod(*mp, struct in_addr *);
                (*mp)->m_len = sizeof(struct in_addr);
                if (imo == NULL || imo->imo_multicast_ifp == NULL)
                        addr->s_addr = INADDR_ANY;
                else {
                        IFP_TO_IA(imo->imo_multicast_ifp, ia);
                        addr->s_addr = (ia == NULL) ? INADDR_ANY
                                        : IA_SIN(ia)->sin_addr.s_addr;
                }
                return (0);

        case IP_MULTICAST_TTL:
                ttl = mtod(*mp, u_char *);
                (*mp)->m_len = 1;
                *ttl = (u_char) ((imo == NULL) ? IP_DEFAULT_MULTICAST_TTL
                                               : imo->imo_multicast_ttl);
                return (0);

        case IP_MULTICAST_LOOP:
                loop = mtod(*mp, u_char *);
                (*mp)->m_len = 1;
                *loop = (u_char) ((imo == NULL) ? (u_char)IP_DEFAULT_MULTICAST_LOOP
                                                : imo->imo_multicast_loop);
                return (0);

        default:
                return (EOPNOTSUPP);
        }
}

/*
 * Discard the IP multicast options.
 */
void
ip_freemoptions(
        struct ip_moptions *imo)
{
        int i;

        if (imo != NULL) {
                for (i = 0; i < imo->imo_num_memberships; ++i)
                        in_delmulti(imo->imo_membership[i]);
                bsd_free(imo, M_IPMOPTS);
        }
}

/*
 * Routine called from ip_output() to loop back a copy of an IP multicast
 * packet to the input queue of a specified interface.  Note that this
 * calls the output routine of the loopback "driver", but with an interface
 * pointer that might NOT be &loif -- easier than replicating that code here.
 */
static void
ip_mloopback(
        struct ifnet *ifp,
        struct mbuf *m,
        struct sockaddr_in *dst)
{
        struct ip *ip;
        struct mbuf *copym;

        copym = m_copy(m, 0, M_COPYALL);
        if (!copym)
            return;

        /* If the ip header is in an M_EXT then it is shared with
         * that in ip_output and the byteswapping would happen twice.
         * Actualize the copy of the header as needed and that
         * is incidentally just what m_pullup does.
         */
        m = m_pullup(copym, sizeof(struct ip));
        if (!m) {
            m_free(copym);
            return;
        }

        /*
         * We don't bother to fragment if the IP length is greater
         * than the interface's MTU.  Can this possibly matter?
         */
        ip = mtod(m, struct ip *);
        ip->ip_len = htons(ip->ip_len);
        ip->ip_off = htons(ip->ip_off);
        ip->ip_sum = 0;
        ip->ip_sum = (u_short) in_cksum(m, ip->ip_hl << 2);
        (void) looutput(iIF(ifp), m, (struct sockaddr *)dst, NULL);
}
