/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <assert.h>
#include <diagnostics.h>
#include <sock.h>
#include <stdio.h>
#include <string.h>

/* Globals collected here from random header files */
#include <if.h>
#include <route.h>
#include <ip/in.h>
#include <ip/in_systm.h>
#include <ip/ip.h>
#include <ip/in_pcb.h>
#include <ip/in_var.h>
#include <ip/ip_var.h>
#include <ip/ip_icmp.h>
#include <raw_cb.h>
#include <udp/udp.h>
#include <udp/udp_var.h>
#include <ip/arp.h>
#include <igmp/igmp.h>
#include <igmp/igmp_var.h>
#include <tcp/tcp.h>
#include <tcp/tcp_fsm.h>
#include <tcp/tcp_seq.h>
#include <tcp/tcp_tim.h>
#include <tcp/tcp_var.h>
#include <dhcp/dhcp_var.h>

#if MONARCH_ADHOC
#include <dsr/dsr_var.h>
#endif /* MONARCH_ADHOC */

INT max_linkhdr;            /* largest link-level header */
INT max_protohdr;           /* largest protocol header */
INT max_hdr;                /* largest link+protocol header */
INT max_datalen;            /* MHLEN - max_hdr */
#ifdef _WANTS_STATS
struct mbstat mbstat;
struct udpstat udpstat;
struct igmpstat igmpstat;
struct ipstat ipstat;
struct rtstat    rtstat;
struct tcpstat tcpstat;  /* tcp statistics */
#endif
struct ipq ipq;        /* ip reass. queue */
UINT16 ip_id;          /* ip packet ctr, for ids */
struct rawcb rawcb;  /* head of list */
struct ifnet *ifnet;
struct domain *domains;
struct inpcb udb;
struct inpcb tcb;        /* head of queue of active tcpcb's */
UINT32 tcp_now;          /* for RFC 1323 timestamps */
tcp_seq tcp_iss; /* tcp initial send seq # */

struct ifnet *lo0ifp;
struct ifaddr *ifa_multiaddr;

/* end globals */

#ifdef __TCS__
struct _NETFACTORY {
    const struct IEndpointFactoryVtbl *v;
    UINT RefCnt;
    BOOL Started;
};
#endif
extern struct _NETFACTORY TheNetFactory;

#undef CHECKSC
#define CHECKSC(_s_) assert(SUCCEEDED(sc)) 

#include <ip/in.h>
#include <dhcp/dhcp_var.h>

#ifdef i386
#define PASCAL __stdcall
#else
#define PASCAL
#endif
char * PASCAL inet_ntoa(struct in_addr in);

#if 0
char * PASCAL inet_ntoa(struct in_addr in)
{
    static char b[18];  /* perfectly MP safe Berkeley junk, tzk tzk */
    UINT8 *ptr;

    ptr = (UINT8 *) &in.s_addr;
    sprintf(b, _T("%d.%d.%d.%d"), ptr[0], ptr[1], ptr[2], ptr[3]);

    return b;
}
#endif

/* NOTE: CobMain() entry point is now in endptfac.c */


/* Basic initialization: do when necessary (when first socket is created) */
SCODE ProtocolStart(UINT what)
{
    SCODE sc = S_OK;
    PIUNKNOWN pUnk;
    struct ifnet *ifp;
    PIENDPOINT ep = NULL;
    PIENDPOINTFACTORY ef;
    struct ifreq _ifr, *rp = &_ifr;
#if !MONARCH_ADHOC
    struct DhcpInfo info;
#endif
        
    NetLock();

    ef = (PIENDPOINTFACTORY) &TheNetFactory;

    /* sanity check: current system depends on lo0 */
    if ((lo0ifp == 0) && ((what & PROTOCOL_START_LO) == 0)) {
        DBGME(1,printf("Attempting to load network without loopback.....\n"));
        sc = EINVAL;
        goto out;
    }

    if (what & PROTOCOL_START_DRIVERS) {
        DBGME(1,printf("Initializing drivers.....\n"));
        sc = CurrentNameSpace()->v->Bind(CurrentNameSpace(), _T("COB/drivers.cob"), 0, &pUnk);
        CHECKSC("bind COB/drivers.cob");

        /* LEAK REF. The drivers cannot be unloaded for now. */
        sc = CurrentNameSpace()->v->Bind(CurrentNameSpace(), _T("COB/drivercfg.cob"), 0, &pUnk);
        CHECKSC("drivercfg");
        pUnk->v->Release(pUnk);
    }

    if (what & (PROTOCOL_START_DHCP | PROTOCOL_START_LO)) {
        sc = ef->v->CreateEndpoint(ef, AF_INET, SOCK_DGRAM, 0, &ep);
        CHECKSC("CreateEndpoint");
    }

    /* Set an IP for each interface */
    for (ifp = ifnet; ifp; ifp = ifp->if_next) {

        /* loopback gets a static address */
        if (!memcmp(ifp->if_name, "lo", strlen(ifp->if_name))) {
            struct ifaliasreq ifra;
            struct sockaddr_in *sin;

            if ((what & PROTOCOL_START_LO) == 0)
                continue;

#define LONAME "lo0"
            memset(&ifra, 0, sizeof(struct ifaliasreq)); 
            strcpy(ifra.ifra_name, LONAME);

            sin = (struct sockaddr_in *)&ifra.ifra_addr;
            sin->sin_family = AF_INET;
            sin->sin_len = sizeof(struct sockaddr_in);
            sin->sin_addr.s_addr = htonl(0x7f000001);

            sin = (struct sockaddr_in *)&ifra.ifra_broadaddr;
            sin->sin_family = AF_INET;
            sin->sin_len = sizeof(struct sockaddr_in);
            sin->sin_addr.s_addr = htonl(0x7fffffff);

            sin = (struct sockaddr_in *)&ifra.ifra_mask;
            sin->sin_family = AF_INET;
            sin->sin_len = sizeof(struct sockaddr_in);
            sin->sin_addr.s_addr = htonl(0xff000000);

            sc = in_control(NULL /* XXX not used */, SIOCAIFADDR, &ifra, ifp);
            if (sc)
                goto out;

            lo0ifp = ifp;
            continue;
        }

        if (what & PROTOCOL_START_DHCP) {
            sprintf(rp->ifr_name, "%s%d", ifp->if_name, (UINT) ifp->if_unit);

            rp->ifr_dhcpflags = 0;
#if !MONARCH_ADHOC
            /* Initiate DHCP transaction, and wait for it to finish. */
            sc = ep->v->Ioctl(ep, SIOCIDHCP, (BYTE*)rp);
            CHECKSC("SIOCIDHCP");

            sc = ep->v->Ioctl(ep, SIOCGIFADDR, (BYTE*)rp);
            if (FAILED(sc)) {
                DBGME(3,printf("Ioctl SIOGGIFADDR for %s failed: sc = %x.\n",rp->ifr_name,sc));
                continue;
            }

            DBGME(3,printf("Enabled IP address %hs for interface %hs.\n",
                inet_ntoa(((struct sockaddr_in *)&rp->ifr_addr)->sin_addr),
                rp->ifr_name));

            *((struct ifreq *)&info) = *rp;
            sc = ep->v->Ioctl(ep, SIOCGDHCPINFO, (BYTE*)&info);
            if (FAILED(sc)) {
                DBGME(3,printf("Ioctl SIOCGDHCPINFO for %s failed: sc = %x.\n",rp->ifr_name,sc));
                continue;
            } else {
                struct in_addr a;

                DBGME(2,_tprintf(_T("The DHCP information is %svalid, our lease "),
                                 (info.LeaseTimer) ? _T("") : _T("in")));
                if (info.LeaseTimer) {
                    DBGME(2,_tprintf(_T("will expire in %d seconds.\r\n"),
                                     info.LeaseTimer));
                } else {
                    DBGME(2,_tprintf(_T("has expired!\r\n")));
                }
                a.s_addr = info.DhcpSrvrIpAddr;
                DBGME(2,printf("[%hs %hs]\n", 
                               "Our IP address was assigned by the DHCP Server",
                               inet_ntoa(a)));
            }
        }
#endif
    }
    /* XXX if some interfaces worked, should return S_OK */
    if (ep)
        ep->v->Release(ep);

 out:
    NetUnlock();
    return sc;
}

void insque( MYQUEUE *What, MYQUEUE *Where)
{
    What->Next = Where->Next;
    Where->Next = What;
    What->Prev = Where;
    What->Next->Prev = What;
}
void remque( MYQUEUE *What)
{
    What->Next->Prev = What->Prev;
    What->Prev->Next = What->Next;
    What->Prev = NULL;
}

TIME time(void)
{
    return CurrentTime();
}

void panic( char *Msg )
{
    printf("IPROTOCOL panic: %s\n", Msg);
    DebugBreak();
}

void errlog( UINT Level, char *Fmt, UINT a1, UINT a2)
{
    char *sl[4] = { "debug", "info", "warning", "ERROR" };
    printf("IPROTOCOL %s: ", sl[Level & 3]);
    printf(Fmt, a1, a2);
}

void
bsd_free( PTR Ptr, UINT Tag)
{
    PIHEAP h = CurrentHeap();
    h->v->Free(h,0,Ptr);
}

PTR bsd_malloc(UINT NBytes, UINT Tag, UINT Options)
{
    PIHEAP h = CurrentHeap();
    UINT flags = HEAP_DEBUG_NAME(Tag);

    if (Options & M_ZERO)
        flags |= HEAP_ZERO_MEMORY;
    return (PTR)h->v->Alloc(h, flags, NBytes, 0);
}
