/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright Microfrost etc etc
 * Gulty writer: Alessandro Forin (sandrof)
 */

#include <assert.h>
#include <sock.h>
#include <diagnostics.h>

#ifdef INET
#include <if.h>
#include <in.h>
#include <arp.h>
#include <dhcp/dhcp_var.h>
#endif

#if MONARCH_ADHOC
#include <dsr/dsr_timer.h>
#endif

#include <stdio.h>

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Spl package
 * +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 */

/*
 * This is pretty ugly, should rip spls out of the code and do it right.
 */
struct {
    MUTEX Mutex;
    PITHREAD Owner;
    INT Nested;
}NetGlobalLock;

void RTLCALLTYPE NetLock(void)
{
    PITHREAD Who;

    Who = CurrentThread();

    if (Who != NetGlobalLock.Owner){
        Mutex_Lock(&NetGlobalLock.Mutex);
        NetGlobalLock.Owner = Who;
    }

    NetGlobalLock.Nested++;
}

void RTLCALLTYPE NetUnlock(void)
{
    if (CurrentThread() != NetGlobalLock.Owner)
        panic(_T("Mismatched NetLock/NetUnlock calls\n"));

    NetGlobalLock.Nested--;
    if (NetGlobalLock.Nested == 0){
        NetGlobalLock.Owner = NULL;
        Mutex_Unlock(&NetGlobalLock.Mutex);
    }
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Sleep/Wakeup package
 * +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 */

/* WARNING: NO_EVENT is not sleepable
 */
#define NO_EVENT NULL

typedef struct _WREC {
    struct _WREC *Next;
    PTR Event;
    UINT Result;
#define EV_TIMEDOUT 0
#define EV_SIGNALLED 1
    const char *Debug;
    MUTEX Lock;
    CONDITION WaitList;
} WREC;

#define WREC_HASH 4
#define SleepHash(x) &SleepList[ (((UINT)x) >> 2) & (WREC_HASH-1) ]

WREC SleepList[ WREC_HASH ];

/*
 * Wait until a given event is signalled
 * Return ETIMEDOUT if non-signalled and timed out.
 * Returns non-zero if wait was interrupted (xxx)
 *
 * Note TimeOut is expected to be positive and relative to the present.
 */
INT XXXWaitFor(PTR Event, const char *Msg, TIME TimeOut)
{
    WREC *p;
    INT depth, retval;

    assert(Event != NO_EVENT)

    if (NetGlobalLock.Nested == 0){
        panic(_T("Yow! Sleeping without holding the global lock!\n"));
    }

    /*
     * Look for empty slot in list.
     */
    p = SleepHash(Event);
    while (p->Event != NO_EVENT && p->Next)
        p = p->Next;

    if (p->Event == NO_EVENT) {
        /*
         * Found free WREC, reuse it.
         */
         p->Event = Event;
    } else {
        /*
         * Allocate a new one.
         */
        WREC *q;
        q = (WREC *)bsd_malloc(sizeof(WREC), M_WAITFOR, 0);
        q->Event = Event;
        q->Debug = Msg;
        q->Next = NULL;
        Mutex_Init(&q->Lock);
        Condition_Init(&q->WaitList);

        p->Next = q;
        p = q;
    }

    NetGlobalLock.Owner = NULL;
    depth = NetGlobalLock.Nested;
    NetGlobalLock.Nested = 0;

    /*
     * If no TimeOut value is specified (i.e. it is zero), wait forever.
     * Otherwise, wait only as long as specified (relative to the present).
     * Avoid nonsensical behavior caused by negative TimeOut values by
     * treating them as zero.
     */
    p->Result = EV_TIMEDOUT;
    if (TimeEqual(TimeOut, TIME_ORIGIN)) {
        Condition_Wait(&p->WaitList, &NetGlobalLock.Mutex, NULL);
    } else {
        if (!TimeLess(TimeOut, TIME_ORIGIN))
            TimeOut = TimeSubtract(TIME_ORIGIN,TimeOut);
        Condition_Wait(&p->WaitList, &NetGlobalLock.Mutex, &TimeOut);
    }

    NetGlobalLock.Owner = CurrentThread();
    NetGlobalLock.Nested = depth;

    retval = ((p->Result == EV_SIGNALLED) ? 0 : ETIMEDOUT);

    /*
     * Release WREC for reuse.
     * BUGBUG?: WREC pool will grow on demand, but never shrink.
     */
    p->Event = NO_EVENT;

    return(retval);
}

void WakeUp(PTR Event)
{
    WREC *p;

    assert(Event != NO_EVENT)

    if (NetGlobalLock.Nested == 0){
        panic(_T("Yow! Signalling without holding the global lock!\n"));
    }

    /*
     * Loop through list, signal every waiter for this event.
     */
    for (p = SleepHash(Event);;p = p->Next) {
        if (p->Event == Event) {
            p->Result = EV_SIGNALLED;
            Condition_Signal(&p->WaitList);
        }
        if (p->Next == NULL)
            break;
    }
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Timer package
 * +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 */

/*
 * This is simple, small and effective
 */

#define MSEC TIME_RELATIVE(TIME_MILLIS(1L))

typedef struct {
    int ReptPrevious;
    INT32 Delay;
    void (*Func)(void);
} MARUTI;/* :-) */

MARUTI NetTimers[] = {
     {   0, 200L*MSEC, pffasttimo  }     /* every 200 msecs */
#if MONARCH_ADHOC
    ,{   1,   0*MSEC, dsr_discovery_timer } /* every 200 msecs */
    ,{   1,   0*MSEC, dsr_rxmt_timer }      /* every 200 msecs */
#endif /* MONARCH_ADHOC */
    ,{   2, 100L*MSEC, pfslowtimo  }     /* every 500 msecs */
    ,{   2,   0*MSEC, if_slowtimo }     /* every second */
    ,{   1,   0*MSEC, DhcpTimer   }     /* every second */
#if MONARCH_ADHOC

    ,{  30,   0*MSEC, dsr_sendbuffer_timer  }   /* every 30 seconds */
#ifdef INET
    ,{  10,   0*MSEC, arptimer    }     /* every 5 minutes (arpt_prune) */
#endif

#else

#ifdef INET
    ,{ 300,   0*MSEC, arptimer    }          /* every 5 minutes (arpt_prune) */
#endif

#endif /* MONARCH_ADHOC */
};
#define NNETTIMERS (sizeof NetTimers / sizeof NetTimers[0])

static void Nest( int ix )
{
    MARUTI *p = &NetTimers[ix];
    TIME t;
    int i;

    /* Do anything that must be done before
     */
    for (i = 0; i < p->ReptPrevious; i++)
        Nest(ix - 1);

    /* If we have to wait a bit longer, wait
     */
    if (p->Delay) {
        Int32ToInt64(t,p->Delay);
        SleepUntil(t);
    }

    /* Waited long enough, do it
     */
    (p->Func)();
}

static void THREAD_LINKAGE TimerThread(THREAD_ARGUMENT xxx)
{
    for (;;) {
        Nest( NNETTIMERS - 1);
    }
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Initialization routines
 * +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 */

void SyncPrimInit()
{
    int i;

    Mutex_Init(&NetGlobalLock.Mutex);
    NetGlobalLock.Owner = NULL;
    NetGlobalLock.Nested = 0;

    for (i = 0; i < WREC_HASH; i++) {
        Mutex_Init(&SleepList[i].Lock);
        Condition_Init(&SleepList[i].WaitList);
        SleepList[i].Event = NO_EVENT;
    }
}

static PITHREAD pTimersThd;

PIPROCESS GetCurrentProcessStaticScope()
{
    PIMODULE pIMod;
    PIPROCESS pIPrc;

    /* Use current module rather than CurrentProcess() so as to
     * account the new thread against the protocol stack rather than
     * against the calling process.  This makes the callling process
     * go away properly.  Explanation: CurrentProcess() returns the
     * current thread's process, i.e. dynamic scope.  The current
     * module, however is statically scoped.
     */
#ifdef h8                       /* there is only one process so far... */
    pIPrc = CurrentProcess();
    pIPrc->v->AddRef(pIPrc); /* released by caller */
#else
    SCODE sc = GetModuleContainsAddr(GetPC(), &pIMod);
    if (FAILED(sc)) {
        DBGME(3,printf("protocol: GetModuleContainsAddr failed, using CurrentProcess\n"));
        pIPrc = CurrentProcess();
        pIPrc->v->AddRef(pIPrc); /* released by caller */
    } else {
      sc = pIMod->v->QueryInterface(pIMod, &IID_IProcess, (void**) &pIPrc);
      pIMod->v->Release(pIMod);

      if (FAILED(sc))
        panic(_T("GetCPSS"));
    }
#endif

    return pIPrc;
}

void StartTimers()
{
    PIPROCESS pIPrc;
    SCODE sc;

    pIPrc = GetCurrentProcessStaticScope();

    /* Measured stack use on NTU debug is 944 bytes */
    sc = pIPrc->v->CreateThread(pIPrc, CREATE_THREAD_STATIC_SCOPING,
                                TimerThread, NULL, 0/*1200*/,
                                NULL,
                                &pTimersThd);
    pIPrc->v->Release(pIPrc);
    if (FAILED(sc))
        panic(_T("Could not create the Timer thread\n"));

    DBGME(2,printf("protocol: Created timers thread %x\n", (UINT) pTimersThd));
    DBGME(2,printf("protocol: &NetGlobalLock = %x\n", (UINT) &NetGlobalLock));
    DBGME(2,printf("protocol: &SleepList = %x\n", (UINT) &SleepList));
}
