/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implements CBuffer concrete class --jvh 2003 */
/* Allows treating mbufs as objects, including IFile */
#include <mmlite.h>
#include <sock.h>
#include <limits.h>
#include <assert.h>
#include <crypto.h>
#include <mmmacros.h>

extern SCODE CEndpointAllocateBuffer (
    /*in*/ PIENDPOINT pThis,
    /*in*/ UINT32 DataSize,
    /*out*/ PIBUFFER * NewBuffer);

/* Convert between interface and instance pointer and vice versa */
#define pMB(_i_) ((PMBUF)((_i_)))
#define iMB(_p_) ((PIBUFFER)((_p_)))

typedef struct mbuf CBuffer;
#define CBufferDestroy NULL
#include "s_CBuffer_v.c"

/* IUnknown: implements IUnknown, IFile, and IBuffer. */
PRIVATE SCODE MBQueryInterface (
    PIBUFFER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    SCODE sc = GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                                     &IID_IBuffer);
    if (sc == E_NO_INTERFACE)
        sc = GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                                   &IID_IFile);
    return sc;
}

PRIVATE UINT MBAddRef (
    PIBUFFER pThis)
{
    PMBUF m = pMB(pThis);
    UINT RefCnt = AtomicInc(&m->m_hdr.mh_RefCnt);
    assert(RefCnt < _UI8_MAX);
    return RefCnt;
}

extern void MBDelete(struct mbuf *); /* in u_mbuf.c */

PRIVATE UINT MBRelease (PIBUFFER pThis)
{
    struct mbuf *m = (struct mbuf *) pThis;
    UINT RefCnt = AtomicDec(&m->m_hdr.mh_RefCnt);
    if (RefCnt > 0)
        return RefCnt;

    MBDelete(m);
    return 0;
}

/* Helper: Count length of mbuf */
PRIVATE UINT m_length(struct mbuf *m)
{
    UINT len;
    if (m->m_hdr.mh_flags & M_PKTHDR)
        return m->m_pkthdr.len; /* assume it is correct */
    for (len = 0; m; m = m->m_hdr.mh_next)
        len += m->m_hdr.mh_len;
    return len;
}

/* IFile: Copy data out of buffer */
PRIVATE SCODE MBReadAt (
    PIBUFFER pThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PMBUF m = pMB(pThis);
    UINT32 Off;
    SCODE sc = S_OK;
    UINT len;

    Off = Uint64ToUint32(Position);
    len = m_length(m);

    if (Off > (UINT32) len)
        Size = 0;
    if (Size > len - Off) {
        Size = len - Off;
        sc = S_FALSE;
    }
    if (Size > 0)
        m_copydata(m, Off, Size, pBuffer);

    *pSizeRead = Size;

    return sc;
}

/* IFile: Copy data into buffer. Extends buffer as needed. */
PRIVATE SCODE MBWriteAt (
    PIBUFFER pThis,
    UINT64 Position,
    const BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeWritten)
{
    PMBUF top = pMB(pThis);
    struct mbuf *m, *n;
    INT len, off;
    BOOL atend;

    if (Int64Equal(Position, UINT64_MAX)) {
        atend = TRUE;
        off = 0;
    } else if (Uint64High32Bits(Position) > 0) {
        /* Enforce size limit on mbufs */
        return E_INVALID_PARAMETER;
    } else {
        atend = FALSE;
        off = Uint64Low32Bits(Position);
        /* Make sure there are no overflows */
        if (off < 0 || (INT)Size < 0 || (INT)Size + off < 0 )
            return E_INVALID_PARAMETER;
    }

    *pSizeWritten = Size;
    if (Size == 0)
        return S_OK; /* nothing written, nothing asked */

    for (m = top; TRUE; m = m->m_hdr.mh_next) {
        if (!atend) {
            /* See if some data should go into the middle of existing mbuf */
            if (off < m->m_hdr.mh_len) {
                len = m->m_hdr.mh_len - off;
                len = min(len, Size);
                memcpy(mtod(m, caddr_t) + off, pBuffer, len);
                Size -= len;
                pBuffer += len;
                if (Size == 0)  /* all written */
                    return S_OK;
                off = 0; /* continue at beginning of next mbuf */
            } else
                off -= m->m_hdr.mh_len;
        }

        /* for loop must be executed at least once so test here */
        if (!m->m_hdr.mh_next)
            break;
    }

    /* Offset must be either at end or within data. No holes allowed. */
    if (!atend && off)
        return E_INVALID_PARAMETER;

    /* Dump all remaining new data at end of existing data */
    while (Size > 0) {
        /* Make it a cluster if there is no previous data and big needed */
        if (!(m->m_hdr.mh_flags & M_EXT)
            &&  (m->m_hdr.mh_len == 0)
            && (Size >= MINCLSIZE))
        {
            MCLGET(m, M_WAIT);
            if (!(m->m_hdr.mh_flags & M_EXT))
                goto Error;
        }

        len = M_TRAILINGSPACE(m);
        len = min(len, Size);

        if (len) {
            memcpy(mtod(m, caddr_t) + m->m_hdr.mh_len, pBuffer, len);
            Size -= len;
            m->m_hdr.mh_len += len;
            pBuffer += len;
            if (top->m_hdr.mh_flags & M_PKTHDR)
                top->m_pkthdr.len += len;
        }

        /* If it didn't fit in the last mbuf, grab another one */
        if (Size) {
            MGET(n, M_WAIT, MT_DATA);
            if (!n)
                goto Error;
            m->m_hdr.mh_next = n;
            m = n;
        }
    }

    return S_OK;

 Error:
    /* Return partial write result. Adjust as to how much was done. */
    if (*pSizeWritten - Size == 0)
        return E_NOT_ENOUGH_MEMORY; /* nothing written, oops */
    /* At least something written, consider it a success */
    *pSizeWritten -= Size;
    return S_OK;
}

/* IFile: Set size of buffer. Truncate or extend as appropriate.
 * NOTE: data is not zeroed. It is the callers responsibility to make
 * sure the data is overwritten before transmission!
 */
PRIVATE SCODE MBSetSize (
    PIBUFFER pThis,
    UINT64 Size64)
{
    PMBUF m = pMB(pThis), n;
    INT32 Size = (INT32) Uint64Low32Bits(Size64), len;

    if (Uint64High32Bits(Size64) > 0 || Size < 0) {
        /* Enforce size limit on mbufs */
        return E_INVALID_PARAMETER;
    }

    len = m_length(m);

    /* Already correct size? */
    if (len == Size)
        return S_OK;

    /* We will set it or kill the whole chain */
    if (m->m_hdr.mh_flags & M_PKTHDR)
        m->m_pkthdr.len = Size;

    /* Skip unaffected head */
    while (m->m_hdr.mh_next && Size > m->m_hdr.mh_len) {
        Size -= m->m_hdr.mh_len;
        m = m->m_hdr.mh_next;
    }

    if (m->m_hdr.mh_len >= Size) {
        /* TRUNCATE */
        m_freem(m->m_hdr.mh_next); /* Free all trailing stuff */
        m->m_hdr.mh_next = NULL;
        m->m_hdr.mh_len = Size; 
        if (Size == 0) {
            /* Make it point to the beginning of the mbuf */
            if (m->m_hdr.mh_flags & M_EXT)
                m->m_hdr.mh_data = m->M_dat.MH.MH_dat.MH_ext.ext_buf->m_hdr.mh_data;
            else if (m->m_hdr.mh_flags & M_PKTHDR)
                m->m_hdr.mh_data = (PBYTE)m->M_dat.MH.MH_dat.MH_databuf;
            else
                m->m_hdr.mh_data = (PBYTE)m->M_dat.M_databuf;
        }
        return S_OK;
    }

    /* EXTEND */
    while (Size > 0) {
        /* Make it a cluster if there is no previous data and big needed */
        if (!(m->m_hdr.mh_flags & M_EXT)
            &&  (m->m_hdr.mh_len == 0)
            && (Size >= MINCLSIZE))
        {
            MCLGET(m, M_WAIT); 
            if (!(m->m_hdr.mh_flags & M_EXT))
                goto Error;
        }

        len = M_TRAILINGSPACE(m);
        len = min(len, Size);

        if (len) {
            Size -= len;
            m->m_hdr.mh_len += len;
        }

        /* If it didn't fit in the last mbuf, grab another one */
        if (Size) {
            MGET(n, M_WAIT, MT_DATA);
            if (!n)
                goto Error;
            m->m_hdr.mh_next = n;
            m = n;
        }
    }
    return S_OK;

 Error:
    /* Extension failed: Kill the chain and return error. */
    MBSetSize(pThis, UINT64_ZERO);
    return E_NOT_ENOUGH_MEMORY;
}

/* IFile: Get size of buffer */
PRIVATE SCODE MBGetSize (
    PIBUFFER pThis,
    /*out*/ UINT64 * Size)
{
    PMBUF m = pMB(pThis);

    UINT len = m_length(m);

    Uint32ToUint64(*Size, len);
    return S_OK;
}

/* IBuffer: Join two buffers together */
PRIVATE SCODE MBAppend (
    PIBUFFER pThis,
    PIBUFFER buf)
{
    struct mbuf *m = pMB(pThis);
    struct mbuf *n = pMB(buf);

    NetLock();
    buf->v->AddRef(buf);        /* external ref + ptr from pThis makes two */
    /* Keep total length accurate when present */
    if (m->m_hdr.mh_flags & M_PKTHDR)
        m->m_pkthdr.len += m_length(n);
    m_cat(m, n);
    NetUnlock();
    return S_OK;
}

/* IBuffer: Split a buffer into two pieces and return tail */
PRIVATE PIBUFFER MBSplit (
    PIBUFFER pThis,
    UINT Length)
{
    struct mbuf *m = pMB(pThis);

    return (PIBUFFER) m_split(m, Length, M_WAIT);
    /* XXX Produces bogus header M_PKTHDR for second buffer */
}

/* IBuffer: Drop data from head (positive) or tail (negative) of buffer */
PRIVATE void MBTruncate (
    /*in*/ PIBUFFER pThis,
    /*in*/ INT Adjustment)
{
    PMBUF m = pMB(pThis);

    NetLock(); /* XXX should be unnecessary. Why did I add it? */
    /* XXX ownership rules: external user and stack are not allowed to mess
     * XXX with mbuf at once. So no lock needed here. ... or in more places.
     */
    m_adj(m, Adjustment);

    NetUnlock();
}

/* IBuffer: Make a chunk of the data in a buffer continuous in memory.
 * This could be for example a struct that needs direct access.
 * Extends the buffer if needed.
 */
PRIVATE SCODE MBMakeContiguous (
    PIBUFFER pThis,
    UINT Offset,
    UINT Length)
{
    PMBUF top = pMB(pThis);
    struct mbuf *m, *n, *q;
    UINT a, b;

    if (Length == 0) /* nothing to do */
        return S_OK;

    /* Size larger than a standard cluster currently not supported */
    assert(Length <= MCLBYTES);
    if (Length > MCLBYTES)
        return E_INVALID_PARAMETER;

    /* Skip head of chain */
    for (m = top;
         m && (INT) Offset >= m->m_hdr.mh_len;
         Offset -= m->m_hdr.mh_len, m = m->m_hdr.mh_next)
        ;
    if (!m)
        return E_INVALID_PARAMETER; /* past EOF */

    assert(m->m_hdr.mh_len > (INT) Offset);

    /* It is already contiguous? */
    if (m->m_hdr.mh_len >= (INT) (Offset + Length))
        return S_OK;

    a = Offset + Length - m->m_hdr.mh_len; /* short this much */
    b = M_TRAILINGSPACE(m);
    n = m->m_hdr.mh_next;
    
    /* If the data does not fit in m it needs to go after m */
    if (a > b) {
        assert(Length > a);     /* XXX obvious assertion */
        a = Length - a;  /* need to move this much */
        if (n && M_LEADINGSPACE(n) >= (INT) a) {
            n->m_hdr.mh_data -= a;
            /* len gets adjusted below */
        } else {
            /* The following buffer head is insufficient, insert new buf */
            q = n;
            n = m_get(M_WAIT, m->m_hdr.mh_type);
            if (!n)
                return E_NOT_ENOUGH_MEMORY;
            n->m_hdr.mh_next = q;
        }
        m->m_hdr.mh_next = n;   /* link it up (same or new) */

        if (Length > MHLEN && !(n->m_hdr.mh_flags & M_EXT)) {
            assert(n->m_hdr.mh_len == 0);
            MCLGET(n, M_WAIT);
            if (!(n->m_hdr.mh_flags & M_EXT))
                return E_NOT_ENOUGH_MEMORY;
        }

        /* Copy straddling data to next (usually new) mbuf */
        m->m_hdr.mh_len -= a;
        n->m_hdr.mh_len += a;
        memcpy(mtod(n, caddr_t), mtod(m, caddr_t) + Offset, a);
        m = n;

        a = Length - m->m_hdr.mh_len; /* still short this much */
        b = M_TRAILINGSPACE(m);
        n = m->m_hdr.mh_next;
    }
    assert (a <= b); /* Data fits at end of m */

    if (a > 0) {
        while (n) {
            /* Copy straddling data, but there might not be enough...*/
            b = n->m_hdr.mh_len;
            b = min(a,b);
            memcpy(mtod(m, caddr_t) + m->m_hdr.mh_len, mtod(n, caddr_t), b);
            n->m_hdr.mh_len -= b;
            n->m_hdr.mh_data += b;
            m->m_hdr.mh_len += b;
            a -= b;
            if (a <= 0)
                break;
            n = n->m_hdr.mh_next;
        }
    }

    /* Not enough data? Zero rest. It fits (we made space above). */
    if (a > 0) {
        memset(mtod(m, caddr_t) + m->m_hdr.mh_len, 0, a);
        m->m_hdr.mh_len += a;
        if (top->m_hdr.mh_flags & M_PKTHDR)
            top->m_pkthdr.len += a;
    }

    return S_OK;
}

PRIVATE SCODE MBInsert (
    PIBUFFER pThis,
    UINT Offset,
    UINT Size)
{
    PMBUF top = pMB(pThis);
    struct mbuf *m, *m2, *prev = NULL;
    UINT a, Off;
    UINT64 size64;
    SCODE sc;

    while (Size > 0) {
        /* Skip any partial elements */
        for (m = top, Off = Offset, prev = NULL;
             Off >= (UINT) m->m_hdr.mh_len;
             m = m->m_hdr.mh_next)
        {
            if (!m->m_hdr.mh_next)
                return E_INVALID_PARAMETER;
            Off -= m->m_hdr.mh_len;
            prev = m;
        }

        if (Off == 0) {
            /* Insert at beginning */
            if (!prev) {
                /* Special case: very beginning */
                if (m->m_flags & M_EXT) {
                    /* Move cluster over to new mbuf */
                    m2 = m_get(M_WAIT, MT_DATA);
                    assert(m2);
                    m2->M_dat.MH.MH_dat.MH_ext.ext_buf = m->M_dat.MH.MH_dat.MH_ext.ext_buf;
                    m2->M_dat.MH.MH_dat.MH_ext.ext_free = m->M_dat.MH.MH_dat.MH_ext.ext_free;
                    m2->m_hdr.mh_data = m->m_hdr.mh_data;
                    m2->m_hdr.mh_len = m->m_hdr.mh_len;
                    m2->m_hdr.mh_flags |= M_EXT;

                    /* Turn old head into non-cluster */
                    m->m_hdr.mh_flags &=~ M_EXT;
                    if (m->m_hdr.mh_flags & M_PKTHDR)
                        m->m_hdr.mh_data = (PBYTE)m->M_dat.MH.MH_dat.MH_databuf;
                    else
                        m->m_hdr.mh_data = (PBYTE)m->M_dat.M_databuf;
                    m->m_hdr.mh_len = 0;

                    m2->m_hdr.mh_next = m->m_hdr.mh_next;
                    m->m_hdr.mh_next = m2;

                    continue;
                } else {
                    /* Move the data to another buffer that follows */
                    m2 = m_get(M_WAIT, m->m_hdr.mh_type);
                    assert(m2);
                    memcpy(mtod(m2, PBYTE), mtod(m, PBYTE), m->m_hdr.mh_len);
                    m2->m_hdr.mh_len = m->m_hdr.mh_len;
                    m->m_hdr.mh_len = 0;

                    m2->m_hdr.mh_next = m->m_hdr.mh_next;
                    m->m_hdr.mh_next = m2;
                }
            }

            /* Try to stuff the next one */
            a = M_LEADINGSPACE(m);
            if (a) {
                a = min(a, Size);
                prev->m_hdr.mh_data -= a;
                prev->m_hdr.mh_len += a;
                Size -= a;
                continue;
            }

            if (prev) {
                /* Turn it into cluster when appropriate */
                if (Size > MINCLSIZE
                    && prev->m_hdr.mh_len == 0
                    && (prev->m_hdr.mh_flags & M_EXT) == 0)
                {
                    MCLGET(prev, M_WAIT);
                }
                /* Stuff previous mbuf to the extent possible */
                a = M_TRAILINGSPACE(prev);
                if (a) {
                    a = min(a, Size);
                    prev->m_hdr.mh_len += a;
                    Size -= a;
                    continue;
                }

                /* Insert a new block here */
                Uint32ToUint64(size64, Size);
                sc = CEndpointAllocateBuffer(NULL, Size, (PIBUFFER*) &m2);
                if (FAILED(sc))
                    return sc;
                sc = iMB(m2)->v->SetSize(iMB(m2), size64);
                if (FAILED(sc))
                    return sc;

                /* Attach old tail at end of new buffer */
                prev->m_hdr.mh_next = m2;
                for ( ; m2->m_hdr.mh_next; m2 = m2->m_hdr.mh_next)
                    ;
                m2->m_hdr.mh_next = m;
                Size = 0;
                continue;
            }
        } else if (Off == (UINT) m->m_hdr.mh_len) {
            /* Insert at end */
        }
    }

    assert(Size == 0);
    return S_OK;;
}

/* IBuffer: Return usable direct pointer to data in buffer */
PRIVATE SCODE MBGetDataPointer (
    PIBUFFER pThis,
    UINT Offset,
    /*out*/ BYTE ** DataPointer,
    /*out*/ PUINT32 Size)
{
    PMBUF top = pMB(pThis);
    struct mbuf *m;
    INT off = (INT) Offset;

    if (off < 0)
        return E_INVALID_PARAMETER;

    /* Skip any partial elements */
    for (m = top; off >= m->m_hdr.mh_len; m = m->m_hdr.mh_next) {
        if (!m->m_hdr.mh_next)
            return E_INVALID_PARAMETER;
        Offset -= m->m_hdr.mh_len;
    }

    *DataPointer = mtod(m, BYTE*) + Offset;
    *Size = m->m_hdr.mh_len - Offset;
    return S_OK;
}

/* IBuffer: Read from file into mbuf or write from mbuf into file */
PRIVATE SCODE MBReadWriteFile (
    PIBUFFER pThis,
    PIFILE File,
    BOOL WriteIt,       /* TRUE: mbuf --> file; FALSE: file --> mbuf  */
    UINT64 Position,    /* UINT64_MAX = append to file, WriteIt */
    UINT Offset,        /* UINT_MAX = append to mbuf, !WriteIt */
    UINT Size,          /* UINT_MAX = as much as possible */
    /*out*/ UINT * pSizeRead)
{
    PMBUF top = pMB(pThis);
    struct mbuf *m, *n;
    INT len, off = (INT) Offset;
    UINT nRead;
    SCODE sc = S_OK;
    BOOL atend = FALSE;

    *pSizeRead = 0;

    /* Make sure there are no overflows (part 1) */
    if ((INT) Size < 0 && Size != UINT_MAX)
        return E_INVALID_PARAMETER;

    if (WriteIt) {
        /* Copying from mbuf to file. Crop to mbuf size. */
        len = m_length(top);
        if (off > len)
            Size = 0;
        if ((INT) Size > len - off || Size == UINT_MAX)
            Size = len - off;
    } else {
        /* Infinite read of file. Get size from file. */
        if (Size == UINT_MAX) {
            UINT64 Size64;
            sc = File->v->GetSize(File, &Size64);
            if (FAILED(sc))
                return sc;
            Size = Uint64Low32Bits(Size64);
            Size -= Offset;
            /* Size limit! */
            if ((INT) Size < 0 || Uint64High32Bits(Size64) != 0)
                return E_INVALID_PARAMETER;
        }
        if (Offset == UINT_MAX) {
            atend = TRUE;
            off = 0;
        }
    }

    /* Make sure there are no overflows (part 2) */
    if (off < 0 || (INT)Size + off < 0 )
        return E_INVALID_PARAMETER;

    if (Size == 0)
        return S_OK; /* nothing written, nothing asked */

    for (m = top; TRUE; m = m->m_hdr.mh_next) {
        if (!atend) {
            /* See if some data is in the middle of existing mbuf */
            if (off < m->m_hdr.mh_len) {
                len = m->m_hdr.mh_len - off;
                len = min(len, Size);
                sc = ReadWriteFile(File, WriteIt, Position,
                                   (PUINT8)(mtod(m, caddr_t) + off), len,
                                   &nRead);
                if (FAILED(sc) || nRead == 0) {
                    assert(sc != (SCODE) E_BUFFER_TOO_SMALL); /* NYI */
                    return sc;
                }
                assert(nRead == (UINT) len);

                *pSizeRead += nRead;
                if (!Int64Equal(Position, UINT64_MAX))
                    Position = Uint64AddUint32(Position, nRead);
                off += nRead;
                Size -= nRead;
                if (Size == 0)  /* all written */
                    return S_OK;
                off = 0; /* continue at beginning of next mbuf */
            } else
                off -= m->m_hdr.mh_len;
        }

        /* for loop must be executed at least once so test here */
        if (!m->m_hdr.mh_next)
            break;
    }

    /* Reading into outside of mbuf extends */
    assert (WriteIt == FALSE);
	if (WriteIt)
		return S_OK; // XXX

    nRead = 1; /* any non-zero initially */
    while (Size > 0 && nRead > 0) {
        /* Make it a cluster if there is no previous data and big needed */
        if (!(m->m_hdr.mh_flags & M_EXT)
            &&  (m->m_hdr.mh_len == 0)
            && (Size >= MINCLSIZE))
        {
            MCLGET(m, M_WAIT);
        }

        len = M_TRAILINGSPACE(m);
        len = min(len, Size);

        if (len) {
            sc = File->v->ReadAt(File,
                                 Position,
                                 (PUINT8) (mtod(m, caddr_t) + m->m_hdr.mh_len),
                                 len,
                                 &nRead);
            if (FAILED(sc)) {
                if (sc == E_BUFFER_TOO_SMALL) {
                    /* if remaining size was too small, give up */
                    if (len == (INT) Size)
                        return S_BUFFER_TOO_SMALL;
                    /* If the namespace could not operate on the remainder,
                     * skip to next mbuf
                     */
                    nRead = len; /* hack: non-zero, equal (for tests) */
                } else {
                    return sc;  /* nothing more to be done here */
                }
            } else {
                assert(nRead <= (UINT) len);
                Size -= nRead;
                m->m_hdr.mh_len += nRead;
                if (top->m_hdr.mh_flags & M_PKTHDR)
                    top->m_pkthdr.len += nRead;
                Position = Uint64AddUint32(Position, nRead);
                *pSizeRead += nRead;
            }
        }

        /* If it didn't fit in the last mbuf, grab another one */
        if (Size && (INT) nRead == len) {
            MGET(n, M_WAIT, MT_DATA);
            m->m_hdr.mh_next = n;
            m = n;
        }
    }

    return sc;
}

/* IBuffer: En/decrypt buffer */
PRIVATE SCODE MBApplyCypher(
    PIBUFFER iThis,
    PICYPHER Cypher,
    CYPHER_OPS Operation)
{
    PMBUF m = pMB(iThis);
    SCODE sc;
    UINT rlen, wlen, bsize = 0, bsize2 = 0, len, len2;
    UINT Consumed = 0, Produced = 0, Inserted = 0;
    INT Off = 0, Off2 = 0, Size;

    /* First figure out correct block size to use */
    bsize = MHLEN;
    sc = Cypher->v->GetBlockLengths(Cypher, &Operation, &bsize, &bsize2);
    if (FAILED(sc))
        return sc;

    Size = m_length(m);
    DBG(("MBApplyCypher Size=%d bsize=%d\n", Size, bsize));
    if (Size == 0)
        return S_OK;

    /* Split code into three cases so as to keep it simpler */
    if (Operation & CYPHER_SIZE_FIXED) while (Size > 0) {
        /* Straddle or end */
        if (m->m_hdr.mh_len < Off + (INT) bsize) {
            sc = MBMakeContiguous(iThis, Consumed, bsize);
            if (FAILED(sc))
                return sc;
        }
        /* Exactly at end of m, goto next */
        if (m->m_hdr.mh_len == Off) {
            m = m->m_hdr.mh_next;
            Off = 0;
            continue;
        }
        assert(m->m_hdr.mh_len > Off);
        assert(m->m_hdr.mh_len >= (INT) bsize);

        sc = Cypher->v->Cypher(Cypher, Operation,
                               mtod(m, PBYTE) + Off, bsize,
                               NULL,
                               mtod(m, PBYTE) + Off, bsize,
                               &rlen);
        if (FAILED(sc))
            return sc;
        assert(rlen == bsize);
        Off += bsize;
        Consumed += bsize;
        Size -= bsize;
    } else if (Operation & (CYPHER_SIZE_GROWS|CYPHER_SIZE_SHRINKS)) {
        struct mbuf *m2;
        UINT64 size64;

        m2 = m;
        while (Size > 0) {
            assert((Operation & CYPHER_SIZE_GROWS) == 0
                   || Produced <= Consumed + Inserted);

            /* Exactly at end of m, goto next */
            if (m->m_hdr.mh_len == Off) {
                m = m->m_hdr.mh_next;
                Off = 0;
                continue;
            }
            if (m2->m_hdr.mh_len == Off2) {
                m2 = m2->m_hdr.mh_next;
                Off2 = 0;
                continue;
            }
            assert(m->m_hdr.mh_len > Off);
            assert(m2->m_hdr.mh_len > Off2);

            if (Operation & CYPHER_SIZE_GROWS) {
                wlen = Consumed+Inserted-Produced;
                if (m == m2 || wlen < bsize+bsize2) {
                    sc = MBInsert(iMB(m), Off2, MCLBYTES-48/*MCLOVH*/);
                    assert(SUCCEEDED(sc));
                    Off += MCLBYTES-48;
                    Inserted += MCLBYTES-48;
                    for (; Off >= m->m_hdr.mh_len; m = m->m_hdr.mh_next) {
                        if (!m->m_hdr.mh_next)
                            return E_INVALID_PARAMETER;
                        Off -= m->m_hdr.mh_len;
                    }
                    continue;
                }
            } else {
                wlen = Size;
            }
            len2 = m2->m_hdr.mh_len - Off2;
            len2 = min(len2, wlen);

            len = m->m_hdr.mh_len - Off;
            len = min(Size, len);

            rlen = wlen = 0;
            sc = Cypher->v->Cypher(Cypher, Operation,
                                   mtod(m, PBYTE) + Off, len,
                                   &rlen,
                                   mtod(m2, PBYTE) + Off2, len2,
                                   &wlen);
            if (FAILED(sc) && sc != E_BUFFER_TOO_SMALL)
                return sc;

            assert(rlen <= (UINT) m->m_hdr.mh_len - Off);
            assert(wlen <= (UINT) m2->m_hdr.mh_len - Off2);
            Off += rlen;
            Off2 += wlen;
            Consumed += rlen;
            Produced += wlen;
            Size -= rlen;
            if (sc == E_BUFFER_TOO_SMALL) {
                assert(m2->m_hdr.mh_len > Off2);
                assert((Operation & CYPHER_SIZE_GROWS) == 0
                       || Inserted >= (UINT) m2->m_hdr.mh_len - Off2);
                Inserted -= m2->m_hdr.mh_len - Off2;
                m2->m_hdr.mh_len = Off2; /* Adjust end */
            } else if (wlen == 0) {
                if ((UINT) Size < bsize) {
                    Operation |= CYPHER_LAST;
                    continue;
                }
                /* Nothing produced, make sure we have enough data */
                /* Move trailing data back to earlier XXX better */
                assert(bsize > (UINT) m->m_hdr.mh_len - Off);
                assert(Produced + Inserted - Consumed > bsize + bsize2);
                sc = MBMakeContiguous(iMB(m), Off, bsize);
                if (FAILED(sc))
                    return sc;
            }
        }
        Uint32ToUint64(size64, Produced);
        sc = iThis->v->SetSize(iThis, size64);
        DBG(("MBApplyCypher VARIABLE SIZE produced %d bytes from %d\n",
             Produced, Consumed));
    } else {
        assert(FALSE);
        return E_INVALID_PARAMETER;
    }

    return sc;
}
