/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implements EndPair concrete class in component protocol */

#include <mmlite.h>
#include <mmmacros.h>
struct sockaddr;
#include <netif.h>

typedef struct _EndPair {
    CONST_VTBL struct IEndpoint1Vtbl *v;
    UINT RefCnt;
    CONST_VTBL struct IEndpoint1Vtbl *P1v;

    PIBUFFER in;
    PIBUFFER out;

    PCONDITION cd0;
    PCONDITION cd1;

    MUTEX mx;
} EndPair, *PENDPAIR;

#define EndPairDestroy NULL
#include "s_EndPair_v.c"

/* Convert between interface and instance pointer and vice versa */
#define pP0(_i_) ((PENDPAIR)((_i_)))
#define iP0(_p_) ((PIENDPOINT1)((_p_)))
#define pP1(_i_) ((PENDPAIR)((_i_) ? (ADDRESS)(_i_) - offsetof(struct _EndPair, P1v) : 0))
#define iP1(_p_) ((PIENDPOINT1)((_p_) ? &(_p_)->P1v : NULL))

/* Store count of P1 in high half of RefCnt */
#define P1_SHIFT (_UINTSIZE/2)
#define P1_MASK ((1<<P1_SHIFT)-1)

PRIVATE SCODE P0QueryInterface (
    PIENDPOINT1 iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    SCODE sc;
    sc = GenericQueryInterface((IUnknown *)iThis, iid, ppObject,
                               &IID_IEndpoint1);
    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IFile, iid)) {
        iThis->v->AddRef(iThis);
        *ppObject = (void*) iP0(pP0(iThis));
        sc = S_OK;
    }

    return sc;
}

PRIVATE SCODE P1QueryInterface (
    PIENDPOINT1 iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    SCODE sc;
    sc = GenericQueryInterface((IUnknown *)iThis, iid, ppObject,
                               &IID_IEndpoint1);
    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IFile, iid)) {
        iThis->v->AddRef(iThis);
        *ppObject = (void*) iP1(pP1(iThis));
        sc = S_OK;
    }

    return sc;
}

PRIVATE UINT P0AddRef (
    PIENDPOINT1 iThis)
{
    return AtomicInc(&pP0(iThis)->RefCnt);
}

PRIVATE UINT P1AddRef (
    PIENDPOINT1 iThis)
{
    return AtomicAdd(&pP1(iThis)->RefCnt, 1 << P1_SHIFT);
}

PRIVATE UINT P0Release (
    PIENDPOINT1 iThis)
{
    PENDPAIR This = pP0(iThis);
    UINT Refs = AtomicDec(&This->RefCnt);
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        RELEASE(This->in);
        RELEASE(This->out);

        pHeap->v->Free(pHeap, 0, (void *) This);
    }
    return Refs;
}

PRIVATE UINT P1Release (
    PIENDPOINT1 iThis)
{
    PENDPAIR This = pP1(iThis);
    UINT Refs = AtomicSub(&This->RefCnt, 1 << P1_SHIFT);
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */

        RELEASE(This->in);
        RELEASE(This->out);

        pHeap->v->Free(pHeap, 0, (void *) This);
    }
    return Refs;
}

SCODE CEndpointAllocateBuffer (
    PIENDPOINT pThis,
    UINT32 DataSize,
    PIBUFFER * NewBuffer);

PRIVATE SCODE P0ReadAt (
    PIENDPOINT1 iThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PENDPAIR This = pP0(iThis);
    PIBUFFER im = This->in;

    return im->v->ReadAt(im, Position, pBuffer, Size, pSizeRead);
}

PRIVATE SCODE P1ReadAt (
    PIENDPOINT1 iThis,
    UINT64 Position,
    /*out*/ BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeRead)
{
    PENDPAIR This = pP1(iThis);
    PIBUFFER im = This->out;

    return im->v->ReadAt(im, Position, pBuffer, Size, pSizeRead);
}

PRIVATE SCODE P0WriteAt (
    PIENDPOINT1 iThis,
    UINT64 Position,
    const BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeWritten)
{
    METHOD_DECLARE_NOTHIS(p0, WriteAt);
    PENDPAIR This = pP0(iThis);
    PIBUFFER im;

    Mutex_Lock(&This->mx);
    im = This->in;
    if (!im) {
        sc = CEndpointAllocateBuffer(NULL, Size, &im);
        CHECKSC("AllocateBuffer");
        This->in = im;
    }
    sc = im->v->WriteAt(im, Position, pBuffer, Size, pSizeWritten);
    CHECKSC("im->WriteAt");
    Condition_Broadcast(This->cd1);

METHOD_CLEANUP;
    Mutex_Unlock(&This->mx);
    return sc;
}

PRIVATE SCODE P1WriteAt (
    PIENDPOINT1 iThis,
    UINT64 Position,
    const BYTE * pBuffer,
    UINT Size,
    /*out*/ UINT * pSizeWritten)
{
    METHOD_DECLARE_NOTHIS(p1, WriteAt);
    PENDPAIR This = pP1(iThis);
    PIBUFFER im;

    Mutex_Lock(&This->mx);
    im = This->out;
    if (!im) {
        sc = CEndpointAllocateBuffer(NULL, Size, &im);
        CHECKSC("AllocateBuffer");
        This->out = im;
    }
    sc = im->v->WriteAt(im, Position, pBuffer, Size, pSizeWritten);
    Condition_Broadcast(This->cd0);
    CHECKSC("im->WriteAt");

METHOD_CLEANUP;
    Mutex_Unlock(&This->mx);
    return sc;
}

PRIVATE SCODE P0SetSize (
    PIENDPOINT1 iThis,
    UINT64 Size)
{
    PENDPAIR This = pP0(iThis);
    PIBUFFER im = This->in;

    return im->v->SetSize(im, Size);
}

PRIVATE SCODE P1SetSize (
    PIENDPOINT1 iThis,
    UINT64 Size)
{
    PENDPAIR This = pP1(iThis);
    PIBUFFER im = This->out;

    return im->v->SetSize(im, Size);
}

PRIVATE SCODE P0GetSize (
    PIENDPOINT1 iThis,
    /*out*/ UINT64 * Size)
{
    PENDPAIR This = pP0(iThis);
    PIBUFFER im = This->in;

    return im->v->GetSize(im, Size);
}

PRIVATE SCODE P1GetSize (
    PIENDPOINT1 iThis,
    /*out*/ UINT64 * Size)
{
    PENDPAIR This = pP1(iThis);
    PIBUFFER im = This->out;

    return im->v->GetSize(im, Size);
}

PRIVATE SCODE P0AllocateBuffer (
    PIENDPOINT1 iThis,
    UINT32 DataSize,
    /*out*/ PIBUFFER * NewBuffer)
{
    return CEndpointAllocateBuffer(NULL, DataSize, NewBuffer);
}

PRIVATE SCODE P1AllocateBuffer (
    PIENDPOINT1 iThis,
    UINT32 DataSize,
    /*out*/ PIBUFFER * NewBuffer)
{
    return CEndpointAllocateBuffer(NULL, DataSize, NewBuffer);
}

PRIVATE SCODE P0SendBuffer (
    PIENDPOINT1 iThis,
    PIBUFFER iBuffer,
    UINT32 Flags,
    const struct sockaddr * pTo,
    UINT32 ToSize)
{
    METHOD_DECLARE_NOTHIS(p0, SendBuffer);
    PENDPAIR This = pP0(iThis);
    PIBUFFER im;

    UnusedParameter(Flags);
    UnusedParameter(pTo);
    UnusedParameter(ToSize);

    sc = S_OK;
    Mutex_Lock(&This->mx);
    im = This->in;
    if (!im)
        This->in = iBuffer;
    else
        sc = im->v->Append(im, iBuffer);

METHOD_CLEANUP;
    Mutex_Unlock(&This->mx);

    Condition_Broadcast(This->cd1);

    return sc;
}

PRIVATE SCODE P1SendBuffer (
    PIENDPOINT1 iThis,
    PIBUFFER iBuffer,
    UINT32 Flags,
    const struct sockaddr * pTo,
    UINT32 ToSize)
{
    METHOD_DECLARE_NOTHIS(p1, SendBuffer);
    PENDPAIR This = pP1(iThis);
    PIBUFFER im;

    UnusedParameter(Flags);
    UnusedParameter(pTo);
    UnusedParameter(ToSize);

    sc = S_OK;
    Mutex_Lock(&This->mx);
    im = This->out;
    if (!im)
        This->out = iBuffer;
    else
        sc = im->v->Append(im, iBuffer);

METHOD_CLEANUP;
    Mutex_Unlock(&This->mx);

    Condition_Broadcast(This->cd0);

    return sc;
}

PRIVATE SCODE P0RecvBuffer (
    PIENDPOINT1 iThis,
    /*out*/ PIBUFFER* piBuffer,
    UINT32 Size,
    UINT32 Flags,
    /*out*/ struct sockaddr * pFrom,
    UINT32 FromSize,
    /*out*/ UINT32* pFromLength)
{
    METHOD_DECLARE_NOTHIS(p0, RecvBuffer);
    PENDPAIR This = pP0(iThis);
    PIBUFFER im;

    UnusedParameter(Flags);
    UnusedParameter(pFrom);
    UnusedParameter(FromSize);
    UnusedParameter(pFromLength);

    sc = S_OK;

    Mutex_Lock(&This->mx);
    im = This->out;
    *piBuffer = im;

    if (!im) {
        if (This->RefCnt & P1_MASK)
            CHECKSETSC(FALSE, E_WOULD_BLOCK, "no data, still open");
        else
            CHECKSETSC(FALSE, E_END_OF_DATA, "no data, EOF");
    }

    /* Grab the requested amount, leave the rest here */
    This->out = im->v->Split(im, Size);

METHOD_CLEANUP;
    Mutex_Unlock(&This->mx);

    if (FAILED(sc))
        return sc;
    else if (*piBuffer)
        return S_OK;
    else
        return S_FALSE;
}

PRIVATE SCODE P1RecvBuffer (
    PIENDPOINT1 iThis,
    /*out*/ PIBUFFER* piBuffer,
    UINT32 Size,
    UINT32 Flags,
    /*out*/ struct sockaddr * pFrom,
    UINT32 FromSize,
    /*out*/ UINT32* pFromLength)
{
    METHOD_DECLARE_NOTHIS(p1, RecvBuffer);
    PENDPAIR This = pP1(iThis);
    PIBUFFER im;

    UnusedParameter(Flags);
    UnusedParameter(pFrom);
    UnusedParameter(FromSize);
    UnusedParameter(pFromLength);

    sc = S_OK;

    Mutex_Lock(&This->mx);
    im = This->in;
    *piBuffer = im;

    if (!im) {
        if (This->RefCnt & ~P1_MASK)
            CHECKSETSC(FALSE, E_WOULD_BLOCK, "no data, still open");
        else
            CHECKSETSC(FALSE, E_END_OF_DATA, "no data, EOF");
    }

    /* Grab the requested amount, leave the rest here */
    This->in = im->v->Split(im, Size);

METHOD_CLEANUP;
    Mutex_Unlock(&This->mx);

    if (FAILED(sc))
        return sc;
    else if (*piBuffer)
        return S_OK;
    else
        return S_FALSE;
}

PRIVATE SCODE P0SetCondition (
    PIENDPOINT1 iThis,
    PCONDITION pCond)
{
    PENDPAIR This = pP0(iThis);

    if (This->cd0)
        return E_OBJECT_NOT_INITIALIZED; /* already... */
    
    This->cd0 = pCond;

    return S_OK;
}

PRIVATE SCODE P1SetCondition (
    PIENDPOINT1 iThis,
    PCONDITION pCond)
{
    PENDPAIR This = pP1(iThis);

    if (This->cd1)
        return E_OBJECT_NOT_INITIALIZED; /* already... */
    
    This->cd1 = pCond;

    return S_OK;
}

SCODE MCT CEndpointFactoryCreateEndpointPair(PIENDPOINTFACTORY pThis,
                                     PIENDPOINT1 *ppNew1, PIENDPOINT1 *ppNew2)
{
    PENDPAIR This = malloc(sizeof(*This));
    if (!This)
        return E_NOT_ENOUGH_MEMORY;

    memset(This, 0, sizeof(*This));
    This->v = &P0EndPairVtbl.v;
    This->P1v = &P1EndPairVtbl.v;
    This->RefCnt = 1 | (1 << P1_SHIFT);
    Mutex_Init(&This->mx);

    *ppNew1 = iP0(This);
    *ppNew2 = iP1(This);

    return S_OK;
}
