/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1988, 1990, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)uipc_socket.c    8.1 (Berkeley) 6/10/93
 */

#include <sock.h>
#include <if.h>
#include <route.h>
#include <assert.h>

/* instead of sys/time.h
 */
struct timeval {
    long tv_sec;   /* seconds */
    long tv_usec;  /* and microseconds */
};

/*
 * Things found in u_sock.c
 * BUGBUG: these prototypes should be elsewhere.
 */
void sorflush(struct socket *so);
INT sodisconnect(struct socket *so);
INT sockargs(struct mbuf **mp, caddr_t buf, UINT buflen, INT type);
INT sendit(struct socket *so, struct msghdr *mp, UINT flags);
INT sendit2(struct socket *so, struct msghdr *mp, struct mbuf *, UINT flags);
INT recvit2(struct socket *so, struct msghdr *mp, UINT32 *namelenp,
            struct mbuf **mp0);

/*
 * The IEndpoint object
 */

/* Destructor */
PRIVATE SCODE CEndpointDestroy(PIUNKNOWN iThis)
{
    struct socket *so = (struct socket *)iThis;

    /*
     * Close down (just in case we aren't already).
     */
    (void) soclose(so);

    /*
     * Mark us as having no references.
     */
    if (so->so_state & SS_NOFDREF)
        panic(_T("trying to release an unreferenced endpoint"));
    so->so_state |= SS_NOFDREF;

    /*
     * Free ourselves.
     */
    sofree(so);

    return S_OK;
}

#if 0
/*
 * Add a reference.
 */
PRIVATE UINT MCT CEndpointAddRef(PIENDPOINT pThis)
{
    INT Refs;
    struct socket *so = (struct socket *)pThis;

    Refs = AtomicInc(&((PISOCKET)pThis)->RefCnt);
    return Refs;
}

/*
 * Release a reference.
 * When last ref is gone, destructor.
 */
PRIVATE UINT MCT CEndpointRelease(PIENDPOINT pThis)
{
    INT Refs;
    struct socket *so = (struct socket *)pThis;

    Refs = AtomicDec(&((PISOCKET)pThis)->RefCnt);

    if (Refs == 0)
        CEndpointDestroy((PIUNKNOWN) so);

    return Refs;
}
        
PRIVATE SCODE MCT CEndpointQueryInterface(PIENDPOINT pThis, REFIID pIid, void* *ppUnk)
{
    if (UuidCmp(&IID_IEndpoint, pIid)
        || UuidCmp(&IID_IFile, pIid)
        || UuidCmp(&IID_IUnknown, pIid)) {
        CEndpointAddRef(pThis);
        *ppUnk = (PIUNKNOWN) pThis;
        return NO_ERROR;
    }
    *ppUnk = NULL;
    return E_NO_INTERFACE;
}
#endif

/*
 * Receive data from peer.
 */
PRIVATE SCODE MCT CEndpointReadAt(PIENDPOINT pThis, UINT64 Flags, BYTE *Buffer,
                          UINT Len, PUINT pRetSize)
{
    struct msghdr msg;
    INT ret;

    msg.msg_name = 0;
    msg.msg_namelen = 0;
    msg.req.uio_pointer = (caddr_t)Buffer;
    msg.req.uio_resid = Len;
    msg.msg_flags = Int64ToInt32(Flags);

    NetLock();
    ret = recvit2((struct socket *)pThis, &msg, NULL, NULL);
    NetUnlock();
    if (ret == -1) {
        PITHREAD th = CurrentThread();
        return th->v->GetValue(th, THREAD_VALUE_LAST_ERROR);
    }
    *pRetSize = ret;
    return NO_ERROR;
}

/*
 * Send data to peer.
 */
PRIVATE SCODE MCT CEndpointWriteAt(PIENDPOINT pThis, UINT64 Flags,
                                  const BYTE *Buffer, UINT Len, PUINT pRetSize)
{
    struct msghdr msg;
    INT ret;

    msg.req.uio_pointer = (caddr_t)Buffer;
    msg.req.uio_resid = Len;
    msg.msg_name = 0;
    msg.msg_namelen = 0;
    msg.msg_flags = 0;

    NetLock();
    ret = sendit2((struct socket *)pThis, &msg, NULL, Int64ToInt32(Flags));
    NetUnlock();
    if (ret == -1) {
        PITHREAD th = CurrentThread();
        return th->v->GetValue(th, THREAD_VALUE_LAST_ERROR);
    }

    *pRetSize = ret;
    return NO_ERROR;
}

PRIVATE SCODE MCT CEndpointSetSize(IEndpoint *This, UINT64 Size)
{
    UnusedParameter(This);
    UnusedParameter(Size);
    return E_NOT_IMPLEMENTED;
}
        
PRIVATE SCODE MCT CEndpointGetSize(IEndpoint *This, UINT64 *pSize)
{
    UnusedParameter(This);
    UnusedParameter(pSize);
    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE MCT CEndpointAccept(PIENDPOINT pThis, PIENDPOINT *ppThat,
                    struct sockaddr *pAddr, UINT32 AddrSize, UINT32 *pAddrSize)
{
    struct mbuf *nam;
    struct socket *so = (struct socket *)pThis;
    SCODE error = 0;

    NetLock();

    if ((so->so_options & SO_ACCEPTCONN) == 0) {
        error = (EINVAL);
        goto ErrExit;
    }
    if ((so->so_state & SS_NBIO) && so->so_qlen == 0) {
        error = (EWOULDBLOCK);
        goto ErrExit;
    }
    while (so->so_qlen == 0 && so->so_error == 0) {
        if (so->so_state & SS_CANTRCVMORE) {
            so->so_error = ECONNABORTED;
            break;
        }
        /*
         * Sleep waiting for connection attempts.
         */
        error = XXXWaitFor((PTR)&so->so_timeo, netcon, TIME_ORIGIN);
        if (error)
            goto ErrExit;
    }
    if (so->so_error) {
        error = so->so_error;
        so->so_error = 0;
        goto ErrExit;
    }

    /*
     * At least one connection now, dequeue it.
     */
    { struct socket *aso = so->so_q;
      if (soqremque(aso, 1) == 0)
        panic(_T("accept"));
      so = aso;
    }

    if ((so->so_state & SS_NOFDREF) == 0)
        panic(_T("Accept: Unaccepted Endpoint already has user ref?!?\n"));
    so->so_state &= ~SS_NOFDREF;

    nam = m_get(M_WAIT, MT_SONAME);
    if (nam == NULL) {
        /* Out of memory, abort
         */
        (void) soabort(so);
        FREE(so, M_SOCKET);
        error = ENOBUFS;
        goto ErrExit;
    }
    (void) (*so->so_proto->pr_usrreq)(so, PRU_ACCEPT, (struct mbuf *)0,
                                      nam, (struct mbuf *)0);

    if (pAddr && AddrSize) {
#ifdef COMPAT_43
        mtod(nam, struct osockaddr *)->sa_family =
             mtod(nam, struct sockaddr *)->sa_family;
#endif
        if (AddrSize > (UINT32)nam->m_len)
            AddrSize = nam->m_len;
        /* SHOULD COPY OUT A CHAIN HERE */
        memcpy(pAddr, mtod(nam, caddr_t), AddrSize);
        *pAddrSize = AddrSize;
    }
    m_freem(nam);
ErrExit:
    NetUnlock();
    if (error)
        so = NULL;

    *ppThat = (PIENDPOINT)so;

    if (error)
        return STATUS_WIN32_ERROR(error);
    return NO_ERROR;
}

SCODE MCT CEndpointBind(PIENDPOINT pThis, const struct sockaddr *Name, UINT32 NameLen)
{
    struct socket *so = (struct socket *)pThis;
    struct mbuf *nam;
    SCODE error;

    error = sockargs(&nam, (caddr_t)Name, NameLen, MT_SONAME);
    if (error == 0) {
        NetLock();
        error = (*so->so_proto->pr_usrreq)(so, PRU_BIND, (struct mbuf *)0,
                                           nam, (struct mbuf *)0);
        NetUnlock();
        m_freem(nam);
    }

    if (error)
        return STATUS_WIN32_ERROR(error);
    return NO_ERROR;
}

PRIVATE SCODE MCT CEndpointClose(PIENDPOINT pThis)
{
    INT error;

    error = soclose((struct socket *)pThis);

    if(error)
        return STATUS_WIN32_ERROR(error);

    return NO_ERROR;
}

PRIVATE SCODE MCT CEndpointConnect(PIENDPOINT pThis, const struct sockaddr *Name, UINT32 NameLen)
{
    struct socket *so = (struct socket *)pThis;
    struct mbuf *nam;
    SCODE error;

    if ((so->so_state & SS_NBIO) && (so->so_state & SS_ISCONNECTING))
        return STATUS_WIN32_ERROR(EALREADY);
    error = sockargs(&nam, (caddr_t)Name, NameLen, MT_SONAME);
    if (error) return STATUS_WIN32_ERROR(error);

    if (so->so_options & SO_ACCEPTCONN) {
        error = (EOPNOTSUPP);
        goto bad;
    }

    /*
     * If protocol is connection-based, can only connect once.
     * Otherwise, if connected, try to disconnect first.
     * This allows user to disconnect by connecting to, e.g.,
     * a null address.
     */
    NetLock();
    if (so->so_state & (SS_ISCONNECTED|SS_ISCONNECTING) &&
        ((so->so_proto->pr_flags & PR_CONNREQUIRED) ||
         ((error = sodisconnect(so)) != 0)))
        error = EISCONN;
    else
        error = (*so->so_proto->pr_usrreq)(so, PRU_CONNECT,
            (struct mbuf *)0, nam, (struct mbuf *)0);
    NetUnlock();

    if (error)
        goto bad;
    if ((so->so_state & SS_NBIO) && (so->so_state & SS_ISCONNECTING)) {
        m_freem(nam);
        return STATUS_WIN32_ERROR(EINPROGRESS);
    }

    NetLock();
    while ((so->so_state & SS_ISCONNECTING) && so->so_error == 0) {
        error = XXXWaitFor((PTR)&so->so_timeo, netcon, TIME_ORIGIN);
        if (error)
            break;
    }
    if (error == 0) {
        error = so->so_error;
        so->so_error = 0;
    }
    NetUnlock();
bad:
    so->so_state &= ~SS_ISCONNECTING;
    m_freem(nam);

    if (error)
        return STATUS_WIN32_ERROR(error);
    return NO_ERROR;
}

PRIVATE SCODE MCT CEndpointGetName(PIENDPOINT pThis, struct sockaddr *Name,
                          UINT32 NameSize, PUINT32 pNameLen, 
                          ENDPOINT_NAME_TYPE WhichName)
{
    struct socket *so = (struct socket *)pThis;
    struct mbuf *m;
    SCODE error;

    switch(WhichName) {

    case ENDPOINT_ADDR_MINE:
        WhichName = PRU_SOCKADDR;
        break;

    case ENDPOINT_ADDR_PEER:
        WhichName = PRU_PEERADDR;
        if ((so->so_state & (SS_ISCONNECTED|SS_ISCONFIRMING)) == 0)
            return STATUS_WIN32_ERROR(ENOTCONN);
        break;

    default:
        return STATUS_WIN32_ERROR(EINVAL);
    }

    NetLock();

    m = m_getclr(M_WAIT, MT_SONAME);
    if (m == NULL) {
        NetUnlock();
        return STATUS_WIN32_ERROR(ENOBUFS);
    }
    error = (*so->so_proto->pr_usrreq)(so, WhichName, 0, m, 0);
    if (error)
        goto bad;
    if (NameSize > (UINT) m->m_len)
        NameSize = m->m_len;
#ifdef COMPAT_43
    mtod(m, struct osockaddr *)->sa_family =
            mtod(m, struct sockaddr *)->sa_family;
#endif
    memcpy( Name, mtod(m, caddr_t), NameSize);
    *pNameLen = NameSize;
bad:
    m_freem(m);

    NetUnlock();

    if(error)
        return STATUS_WIN32_ERROR(error);
    else
        return NO_ERROR;
}

PRIVATE SCODE MCT CEndpointGetSockOpt(PIENDPOINT pThis, INT32 Level,
                    UINT32 OptName, BYTE *OptVal, UINT32 OptValSize, PUINT32 pOptLen)
{
    struct socket *so = (struct socket *)pThis;
    struct mbuf *m = NULL;
    UINT valsize;
    SCODE error;

    NetLock();

    valsize = (OptVal) ? OptValSize : 0;

    if (Level != SOL_SOCKET) {
        if (so->so_proto && so->so_proto->pr_ctloutput) {
            error = ((*so->so_proto->pr_ctloutput)
                  (PRCO_GETOPT, so, Level, OptName, &m));
        } else {
            NetUnlock();
            return STATUS_WIN32_ERROR(ENOPROTOOPT);
        }
    } else {
        m = m_get(M_WAIT, MT_SOOPTS);
        m->m_len = sizeof(INT);

        switch (OptName) {

        case SO_LINGER:
            m->m_len = sizeof (struct linger);
            mtod(m, struct linger *)->l_onoff = so->so_options & SO_LINGER;
            mtod(m, struct linger *)->l_linger = so->so_linger;
            break;

        case SO_USELOOPBACK:
        case SO_DONTROUTE:
        case SO_DEBUG:
        case SO_KEEPALIVE:
        case SO_REUSEADDR:
        case SO_REUSEPORT:
        case SO_BROADCAST:
        case SO_OOBINLINE:
            *mtod(m, INT *) = so->so_options & OptName;
            break;

        case SO_TYPE:
            *mtod(m, INT *) = so->so_type;
            break;

        case SO_ERROR:
            *mtod(m, INT *) = so->so_error;
            so->so_error = 0;
            break;

        case SO_SNDBUF:
            *mtod(m, UINT32 *) = so->so_snd.sb_hiwat;
            break;

        case SO_RCVBUF:
            *mtod(m, UINT32 *) = so->so_rcv.sb_hiwat;
            break;

        case SO_SNDLOWAT:
            *mtod(m, UINT32 *) = so->so_snd.sb_lowat;
            break;

        case SO_RCVLOWAT:
            *mtod(m, UINT32 *) = so->so_rcv.sb_lowat;
            break;

        case SO_SNDTIMEO:
        case SO_RCVTIMEO:
            {
            TIME val = (OptName == SO_SNDTIMEO ?
                 so->so_snd.sb_timeo : so->so_rcv.sb_timeo);
#ifdef RIGHT_WAY
            /* BSD code returns a pointer to a struct timeval */
            struct timeval *tv;

            m->m_len = sizeof(*tv);
            tv = mtod(m, struct timeval *);
            tv->tv_sec = (long)(val / 10000000);
            val -= (tv->tv_sec * 10000000);
            tv->tv_usec = (long)val;
#else /* NT way */
            /* NT code takes a pointer to an int */
            INT *bogus;

            bogus = mtod(m, INT *);
            *bogus = Int64ToInt32(Int64DividedByInt32(val,TIME_MILLIS(1)));
#endif
            break;
            }

        default:
            (void)m_free(m);
            NetUnlock();
            return STATUS_WIN32_ERROR(ENOPROTOOPT);
        }
        error = 0;
    }

    if (error == 0 && OptVal && valsize && m != NULL) {
        if (valsize > (UINT)m->m_len)
            valsize = m->m_len;
        memcpy(OptVal, mtod(m, caddr_t), valsize);
        *pOptLen = valsize;
    }
    if (m != NULL)
        (void) m_free(m);

    NetUnlock();

    if (error)
        return STATUS_WIN32_ERROR(error);
    else
        return NO_ERROR;
}

PRIVATE SCODE MCT CEndpointIoctl(PIENDPOINT pThis, UINT32 cmd, UINT8 *data)
{
    struct socket *so = (struct socket *)pThis;
    INT error = 0;

    switch (cmd) {

    case FIONBIO:
        if (*(INT *)data)
            so->so_state |= SS_NBIO;
        else
            so->so_state &= ~SS_NBIO;
        break;

    case FIOASYNC:
        if (*(INT *)data) {
            so->so_state |= SS_ASYNC;
            so->so_rcv.sb_flags |= SB_ASYNC;
            so->so_snd.sb_flags |= SB_ASYNC;
        } else {
            so->so_state &= ~SS_ASYNC;
            so->so_rcv.sb_flags &= ~SB_ASYNC;
            so->so_snd.sb_flags &= ~SB_ASYNC;
        }
        break;

    case FIONREAD:
        *(UINT32 *)data = so->so_rcv.sb_cc;
        break;

    case SIOCSPGRP:
        so->so_pgid = *(INT *)data;
        break;

    case SIOCGPGRP:
        *(INT *)data = so->so_pgid;
        break;

    case SIOCATMARK:
        *(INT *)data = (so->so_state & SS_RCVATMARK) != 0;
        break;

    default:
        /*
         * Interface/routing/protocol specific ioctls:
         * interface and routing ioctls should have a
         * different entry since a socket's unnecessary
         */
        NetLock();
        if (IOCGROUP(cmd) == 'i')
            error = ifioctl(so, cmd, (UINT8 *) data);
        else if (IOCGROUP(cmd) == 'r')
            error = rtioctl(cmd, (caddr_t)data);
        else
            error = (*so->so_proto->pr_usrreq)(so, PRU_CONTROL, 
                    (struct mbuf *)cmd, (struct mbuf *)data, (struct mbuf *)0);
        NetUnlock();
    }

    if (error)
        return STATUS_WIN32_ERROR(error);

    return NO_ERROR;
}

PRIVATE SCODE MCT CEndpointListen(PIENDPOINT pThis, INT32 BackLog)
{
    struct socket *so = (struct socket *)pThis;
    SCODE error;

    NetLock();
    error = (*so->so_proto->pr_usrreq)(so, PRU_LISTEN, (struct mbuf *)0,
                                       (struct mbuf *)0, (struct mbuf *)0);
    if (error) {
        NetUnlock();
        return STATUS_WIN32_ERROR(error);
    }
    if (so->so_q == 0)
        so->so_options |= SO_ACCEPTCONN;
    if (BackLog < 0)
        BackLog = 0;
    so->so_qlimit = (short)min(BackLog, SOMAXCONN);
    NetUnlock();
    return (NO_ERROR);
}

/*
 * Receive from and return the origin
 */
PRIVATE SCODE MCT CEndpointRecvFrom(PIENDPOINT pThis,
                   BYTE *pBuffer, UINT32 Size,
                   UINT32 *pRecvSize, UINT32 Flags, struct sockaddr *pFrom,
                   UINT32 FromSize, UINT32 *pFromSize)
{
    struct msghdr msg;
    INT retval;

    msg.msg_namelen = FromSize;
    msg.msg_name = pFrom;
    msg.req.uio_pointer = (caddr_t) pBuffer;
    msg.req.uio_resid = Size;
    msg.msg_flags = Flags;
#ifdef COMPAT_43
    msg.msg_flags |= MSG_COMPAT;
#endif

    NetLock();
    retval = recvit2((struct socket *)pThis, &msg, pFromSize, NULL);
    NetUnlock();
    if (retval == -1) {
        PITHREAD th = CurrentThread();
        return th->v->GetValue(th, THREAD_VALUE_LAST_ERROR);
    }

    *pRecvSize = retval;
    return S_OK;
}

/*
 * Send to a given destination
 */
PRIVATE SCODE MCT CEndpointSendTo(PIENDPOINT pThis,
                 const BYTE *pBuffer, UINT32 Size,
                 UINT32 *pSizeSent, UINT32 Flags, const struct sockaddr *pTo,
                 UINT32 ToSize)
{
    struct msghdr msg;
    INT retval;

    msg.req.uio_pointer = (caddr_t)pBuffer;
    msg.req.uio_resid = Size;
    msg.msg_name = (struct sockaddr *) pTo;
    msg.msg_namelen = ToSize;
    msg.msg_flags = 0;

    NetLock();
    retval = sendit2((struct socket *)pThis, &msg, NULL, Flags);
    NetUnlock();
    if (retval == -1) {
        PITHREAD th = CurrentThread();
        return th->v->GetValue(th, THREAD_VALUE_LAST_ERROR);
    }

    *pSizeSent = retval;
    return S_OK;
}

PRIVATE SCODE MCT CEndpointSetSockOpt(PIENDPOINT pThis,
             UINT32 Level, UINT32 OptName, const BYTE *OptVal, UINT32 OptLen)
{
    struct socket *so = (struct socket *)pThis;
    struct mbuf *m0 = NULL, *m;
    SCODE error;

    NetLock();
    error = (OptVal) ? sockargs(&m0, (caddr_t)OptVal, OptLen, MT_SOOPTS) : 0;
    NetUnlock();
    if (error)
        return STATUS_WIN32_ERROR(error);

    m = m0;

    if (Level != SOL_SOCKET) {
        if (so->so_proto && so->so_proto->pr_ctloutput){
            NetLock();
            error = ((*so->so_proto->pr_ctloutput)
                  (PRCO_SETOPT, so, Level, OptName, &m0));
            NetUnlock();
            if (error)
                return STATUS_WIN32_ERROR(error);
            else
                return NO_ERROR;
        }
        error = ENOPROTOOPT;
    } else {
        switch (OptName) {

        case SO_LINGER:
            if (m == NULL || m->m_len != sizeof (struct linger)) {
                error = EINVAL;
                goto bad;
            }
            so->so_linger = (short) mtod(m, struct linger *)->l_linger;
            /* fall thru... */

        case SO_DEBUG:
        case SO_KEEPALIVE:
        case SO_DONTROUTE:
        case SO_USELOOPBACK:
        case SO_BROADCAST:
        case SO_REUSEADDR:
        case SO_REUSEPORT:
        case SO_OOBINLINE:
            if (m == NULL || m->m_len < sizeof(INT)) {
                error = EINVAL;
                goto bad;
            }
            if (*mtod(m, INT *))
                so->so_options |= OptName;
            else
                so->so_options &= ~OptName;
            break;

        case SO_SNDBUF:
        case SO_RCVBUF:
        case SO_SNDLOWAT:
        case SO_RCVLOWAT:
            if (m == NULL || m->m_len < sizeof(INT)) {
                error = EINVAL;
                goto bad;
            }
            switch (OptName) {

            case SO_SNDBUF:
            case SO_RCVBUF:
                NetLock();
                if (sbreserve(OptName == SO_SNDBUF ? &so->so_snd : &so->so_rcv,
                        (UINT32) *mtod(m, INT *)) == 0) {
                    NetUnlock();
                    error = ENOBUFS;
                    goto bad;
                }
                NetUnlock();
                break;

            case SO_SNDLOWAT:
                so->so_snd.sb_lowat = *mtod(m, UINT32 *);
                break;
            case SO_RCVLOWAT:
                so->so_rcv.sb_lowat = *mtod(m, UINT32 *);
                break;
            }
            break;

        case SO_SNDTIMEO:
        case SO_RCVTIMEO:
            {
            TIME t;
#ifdef RIGHT_WAY
            /* BSD code takes a pointer to struct timeval */
            struct timeval *tv;

            if (m == NULL || m->m_len < sizeof (*tv)) {
                error = EINVAL;
                goto bad;
            }
            tv = mtod(m, struct timeval *);

            t = TIME_MICROS(tv->tv_usec) + TIME_SECONDS(tv->tv_sec);
#else /* NT way */
            /* NT code takes a pointer to an int */
            INT *bogus;

            if (m == NULL || m->m_len < sizeof(INT)) {
                error = EINVAL;
                goto bad;
            }
            bogus = mtod(m, INT *);
            Int32ToInt64(t, TIME_MILLIS(*bogus));
#endif
            switch (OptName) {

            case SO_SNDTIMEO:
                so->so_snd.sb_timeo = t;
                break;
            case SO_RCVTIMEO:
                so->so_rcv.sb_timeo = t;
                break;
            }
            break;
            }

        default:
            error = ENOPROTOOPT;
            break;
        }
        if (error == 0 && so->so_proto && so->so_proto->pr_ctloutput) {
            NetLock();
            (void) ((*so->so_proto->pr_ctloutput)
                     (PRCO_SETOPT, so, Level, OptName, &m0));
            NetUnlock();
            m = NULL; /* freed by protocol */
        }
    }
bad:
    if (m)
        (void) m_free(m);

    if (error)
        return STATUS_WIN32_ERROR(error);
    else
        return NO_ERROR;
}

/*
 * Disable sends and/or receives on an Endpoint.
 *
 * Valid values for "How":
 *    0 - subsequent receives on the Endpoint will be disallowed.
 *    1 - subsequent sends are disallowed.
 *    2 - both 0&1
 */
PRIVATE SCODE MCT CEndpointShutdown(PIENDPOINT pThis, UINT32 How)
{
    struct socket *so = (struct socket *) pThis;
    const struct protosw *pr = so->so_proto;
    SCODE error = 0;

    NetLock();  /* paranoia */

    How++;
    if (How & 1)
        sorflush(so);
    if (How & 2)
        error = ((*pr->pr_usrreq)(so, PRU_SHUTDOWN, (struct mbuf *)0,
                 (struct mbuf *)0, (struct mbuf *)0));

    NetUnlock();

    if (error)
        return STATUS_WIN32_ERROR(error);
    else
        return NO_ERROR;
}

SCODE CEndpointAllocateBuffer (
    /*in*/ PIENDPOINT pThis,
    /*in*/ UINT32 DataSize,
    /*out*/ PIBUFFER * NewBuffer)
{
    struct socket *so = (struct socket *) pThis;
    struct mbuf *m;

    UnusedParameter(so);

    m = m_gethdr(M_WAIT, MT_DATA);
    if (m == NULL)
        return E_NOT_ENOUGH_MEMORY;
    m->m_pkthdr.len = 0;
    m->m_pkthdr.rcvif = (struct ifnet *)0;    

    if (DataSize >= MINCLSIZE) {
        MCLGET(m, M_WAIT);
        if (!(m->m_flags & M_EXT)) {
            m_free(m);
            return E_NOT_ENOUGH_MEMORY;
        }
    }

    /* XXX if datagram to be sent, reserve space for headers */

    *NewBuffer = (PIBUFFER) m;
    return S_OK;
}

PRIVATE SCODE CEndpointSendBuffer (
    /*in*/ PIENDPOINT pThis,
    /*in*/ PIBUFFER iBuffer,
    /*in*/ UINT32 Flags,
    /*in*/ const struct sockaddr * pTo,
    /*in*/ UINT32 ToSize)
{
    SCODE sc = S_OK;
    struct msghdr msg;
    INT retval;
    UINT64 Size64;
    iBuffer->v->GetSize(iBuffer, &Size64);
    iBuffer->v->AddRef(iBuffer); /* the rest of the stack consumes... */

    msg.req.uio_pointer = NULL;
    msg.req.uio_resid = Uint64ToUint32(Size64);
    msg.msg_name = (struct sockaddr *) pTo;
    msg.msg_namelen = ToSize;
    msg.msg_flags = 0;

    NetLock();
    retval = sendit2((struct socket *)pThis, &msg,
                     (struct mbuf *) iBuffer, Flags);
    if (retval == -1) {
        PITHREAD th = CurrentThread();
        struct socket *so = (struct socket *)pThis;
        if (so->so_type == SOCK_STREAM)
            iBuffer->v->Release(iBuffer); /* ...except when TCP failed */
        sc = th->v->GetValue(th, THREAD_VALUE_LAST_ERROR);
    }
    NetUnlock();

    /**pSizeSent = retval;*/
    return sc;
}

PRIVATE SCODE CEndpointRecvBuffer (
    PIENDPOINT pThis,
    /*out*/ PIBUFFER* piBuffer,
    UINT32 Size,
    UINT32 Flags,
    /*out*/ struct sockaddr * pFrom,
    UINT32 FromSize,
    /*out*/ UINT32 *pFromLength)
{
    struct socket *This = (struct socket *) pThis;
    struct msghdr msg;
    struct mbuf *m;
    INT retval;

    msg.msg_namelen = FromSize;
    msg.msg_name = (struct sockaddr *) pFrom;
    msg.req.uio_pointer = NULL;
    msg.req.uio_resid = Size;
    msg.msg_flags = Flags;

    NetLock();
    retval = recvit2(This, &msg, pFromLength, &m);
    NetUnlock();
    if (retval == -1) {
        PITHREAD th = CurrentThread();
        return th->v->GetValue(th, THREAD_VALUE_LAST_ERROR);
    }

    /* Adjust length to reflect actual data present */
    if (m && (m->m_hdr.mh_flags & M_PKTHDR))
        m->m_pkthdr.len = retval;

    *piBuffer = (PIBUFFER) m;
    // *pRecvSize = retval;
    if (retval == 0)
        return S_FALSE;
    return S_OK;
}

PRIVATE SCODE CEndpointSetCondition (
    PIENDPOINT pThis,
    PCONDITION pCond)
{
    struct socket *This = (struct socket *) pThis;

    PCONDITION old;

    old = (PCONDITION) AtomicSwap((PUINT) &This->pCond, (UINT) pCond);

    /* Signal old condition as the status changed for this socket */
    if (old)
        Condition_Broadcast(old);

    return S_OK;
}

typedef struct socket CEndpoint;

#include <s_CEndpoint_v.c>
