/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * The IEndpointFactory object
 */
#include <mmlite.h>
#include <diagnostics.h>
#include <assert.h>
#include <sock.h>
#include <if.h>
#include <route.h>
#include <base/cobdesc.h>

extern SCODE MCT CEndpointBind(PIENDPOINT, const struct sockaddr *, UINT32);
extern void mbinit(void);
extern void SyncPrimInit(void);
extern void domaininit(void);

/*
 * bogus pull-in of NetGlobalLock for WaitForEndpoints
 */
extern struct {
    MUTEX Mutex;
    PITHREAD Owner;
    INT Nested;
} NetGlobalLock;

typedef struct _NETFACTORY {
    struct _DefCob super;
    BOOL Started;
} CEndpointFactory, *PNETFACTORY;

#define pNF(_i_) ((PNETFACTORY) (_i_))

#define _MANUAL_COBMAIN 1
#define CEndpointFactoryDestroy NULL
#include <s_CEndpointFactory_v.c>

#if 0
/*
 * Add a reference.
 */
PRIVATE UINT MCT CEndpointFactoryAddRef(PIENDPOINTFACTORY pThis)
{
    PNETFACTORY This = pNF(pThis);
    return AtomicInc(&This->RefCnt);
}

/*
 * Release a reference.
 * When last ref is gone, destructor.
 */
PRIVATE UINT MCT CEndpointFactoryRelease(PIENDPOINTFACTORY pThis)
{
    PNETFACTORY This = pNF(pThis);
    return AtomicDec(&This->RefCnt);
}

/*
 * Query support.
 * We inherit solely from IUnknown.
 */
PRIVATE SCODE MCT CEndpointFactoryQueryInterface(PIENDPOINTFACTORY pThis,
                                                 REFIID pIid,
                                                 void* *ppUnk)
{
    SCODE sc;
    sc = GenericQueryInterface((PIUNKNOWN) pThis, pIid, ppUnk,
                               &IID_IProgram);
    if (sc == E_NO_INTERFACE) {
        sc = GenericQueryInterface((PIUNKNOWN) pThis, pIid, ppUnk,
                                   &IID_IEndpointFactory);
    }
    return sc;
}
#endif


/* Declare v-table type. XXX should be in edpoint.c */
extern const struct CUnknownVtbl CEndpointVtbl;
/*
 * Build a new Endpoint object.
 */

PRIVATE SCODE MCT EndpointObjNew(PIENDPOINTFACTORY pThis,
                                 UINT32 Family, UINT32 Type, UINT32 Protocol,
                                 PIENDPOINT *ppNew)
{
    const struct protosw *prp;
    struct socket *so;
    SCODE error;
    PNETFACTORY This = pNF(pThis);
#if 0
    SCODE sc;
    /* If protocol stack is not yet started do it now */
    if (!This->Started) {
        sc = S_OK;
        /* Guard against other threads (but this one re-enters) */
        NetLock();
        if (!This->Started) {
            This->Started = TRUE;
            sc = ProtocolStart(PROTOCOL_START_DRIVERS | PROTOCOL_START_LO);
        }
        NetUnlock();
        if (FAILED(sc)) {
            DBG(("endpointfactory: ProtocolStart failed sc=%x\n"));
            return sc;
        }
    }
#endif

    if (Protocol)
        prp = pffindproto(Family, Protocol, Type);
    else
        prp = pffindtype(Family, Type);
    if (prp == 0 || prp->pr_usrreq == 0) {
        error = (EPROTONOSUPPORT);
        goto ErrExit;
    }
    if (prp->pr_type != Type) {
        error = (EPROTOTYPE);
        goto ErrExit;
    }

    MALLOC(so, struct socket *, sizeof(*so), M_SOCKET, M_WAIT);
    bzero((caddr_t)so, sizeof(*so));
    so->v = (void *) &CEndpointVtbl.v;
    so->RefCnt = 1;
    so->so_type = (short)Type;
#ifdef SECURITY
    if (CurrentThread()->p_ucred->cr_uid == 0)
        so->so_state = SS_PRIV;
#else
    /* everyone is priv */
    so->so_state = SS_PRIV;
#endif
    so->so_proto = prp; 
    NetLock();
    error = (*prp->pr_usrreq)(so, PRU_ATTACH, (struct mbuf *)0,
                              (struct mbuf *)Protocol, (struct mbuf *)0);
    NetUnlock();
    if (error) {
        so->so_state |= SS_NOFDREF;
        sofree(so);
        goto ErrExit;
    }

    *ppNew = (PIENDPOINT)so;

ErrExit:
    if (error)
        return STATUS_WIN32_ERROR(error);
    return S_OK;
}

SCODE MCT CEndpointFactoryCreateEndpointPair(PIENDPOINTFACTORY pThis,
                                             PIENDPOINT1 *ppNew1, PIENDPOINT1 *ppNew2);

/*
 * Condition for threads waiting for Endpoints in WaitForEndpoints to wait on.
 */
CONDITION WaitHere;

/*
 * Wait for one or more Endpoints previously built by this factory to
 * undergo some notable event (like becoming readable, writeable, etc).
 *
 * Since we want to be able to wait on multiple endpoints simultaneously,
 * but can only wait on one condition at once, we use the same condition
 * for all waiting and re-check event occurrence on all wakeups.
 */
PRIVATE SCODE MCT WaitForEndpoints(PIENDPOINTFACTORY pThis, PIENDPOINT *Endpts,
                                   ENDPOINT_REQ_TYPES *EventTypes, UINT NumEndpts,
                                   TIME Timeout)
{
    UINT Loop, Satisfied = 0;
    SCODE sc = S_OK;

    NetLock();

    /*
     * If given a relative timeout, make it absolute.
     */
    if (TimeLess(Timeout,TIME_ORIGIN))
        Timeout = TimeSubtract(CurrentTime(),Timeout);

    do {
        /*
         * For each Endpoint in the array, first verify that it came from
         * this endpoint factory.  Next, check to see if event type we're
         * interested in has already been satisfied.  If it has not, mark
         * that we're waiting on it and wait (at least until timeout expires)
         * for the condition to be satisfied.
         */
        for (Loop = 0;Loop < NumEndpts;Loop++) {
            if (!(EventTypes[Loop]
                  & (ENDPOINT_READ|ENDPOINT_WRITE|ENDPOINT_EXCEPTION)))
                continue;
            /*
             * Verify Endpoint came from this factory.
             */
            if (Endpts[Loop]->v->Bind != CEndpointBind) {
                sc = E_INVALID_PARAMETER;
                goto Leave;
            }
            /*
             * We use the upper bits of the entries in the EventTypes array
             * to mark temporarilly that an event has been satisfied.  We
             * overwrite the request bits with the result bits on exit.
             *
             * Clear all the OUT bits.
             */
            EventTypes[Loop] &= ~(ENDPOINT_ISREADABLE|ENDPOINT_ISWRITEABLE|ENDPOINT_ISEXCEPTIONAL);

            if (EventTypes[Loop] & ENDPOINT_READ) {
                if (soreadable((struct socket *)Endpts[Loop])) {
                    EventTypes[Loop] |= ENDPOINT_ISREADABLE;
                    Satisfied++;
                } else
                    ((struct socket *)Endpts[Loop])->so_rcv.sb_flags |= SB_SEL;
            }
            if (EventTypes[Loop] & ENDPOINT_WRITE) {
                if (sowriteable((struct socket *)Endpts[Loop])) {
                    EventTypes[Loop] |= ENDPOINT_ISWRITEABLE;
                    Satisfied++;
                } else
                    ((struct socket *)Endpts[Loop])->so_snd.sb_flags |= SB_SEL;
            }
            if (EventTypes[Loop] & ENDPOINT_EXCEPTION) {
                if (((struct socket *)Endpts[Loop])->so_oobmark || 
                    ((struct socket *)Endpts[Loop])->so_state & SS_RCVATMARK) {
                    EventTypes[Loop] |= ENDPOINT_ISEXCEPTIONAL;
                    Satisfied++;
                } else
                    ((struct socket *)Endpts[Loop])->so_rcv.sb_flags |= SB_SEL;
            }
        }

        if (Satisfied)
            goto Leave;

        /*
         * None of the events we are interested in are currently satisfied.
         * Wait around for something to happen (or for timeout to expire).
         *
         * Note we are awakened when *ANY* Endpoint being waited for is
         * modified, so we need to go back and re-check each of our Endpoints
         * if that happens.
         */
        if (!TimeEqual(Timeout,TIME_ORIGIN)) {
            UINT Depth;

            NetGlobalLock.Owner = NULL;
            Depth = NetGlobalLock.Nested;
            NetGlobalLock.Nested = 0;

            sc = Condition_Wait(&WaitHere, &NetGlobalLock.Mutex, &Timeout);

            NetGlobalLock.Owner = CurrentThread();
            NetGlobalLock.Nested = Depth;
        } else
            sc = E_TIMED_OUT;

    } while (sc != E_TIMED_OUT);

Leave:
    NetUnlock();
    return sc;
}

SCODE CEndpointAllocateBuffer (
    PIENDPOINT pThis,
    UINT32 DataSize,
    PIBUFFER * NewBuffer);

PRIVATE SCODE CEndpointFactoryAllocateBuffer (
    PIENDPOINTFACTORY pThis,
    UINT32 DataSize,
    /*optional*/ PINETDRIVER recvif,
    /*out*/ PIBUFFER * NewBuffer)
{
    SCODE sc;
    PIBUFFER im;
    UINT64 Size64;

    sc = CEndpointAllocateBuffer(NULL, DataSize, &im);
    if (FAILED(sc))
        return sc;

    /* If called from a driver, set the actual size rather than just treating
     * the size as a hint.
     */
    if (recvif) {
        ((struct mbuf *)im)->m_pkthdr.rcvif = (struct ifnet *) recvif;
        Uint32ToUint64(Size64, DataSize);
        sc = im->v->SetSize(im, Size64);
        if (FAILED(sc)) {
            im->v->Release(im);
            return sc;
        }
    }
    *NewBuffer = im;
    return sc;
}

PRIVATE SCODE MCT 
CEndpointFactoryEtherInput(PIENDPOINTFACTORY pThis, PINETDRIVER ifp,
                           struct ether_header *eh, PIBUFFER m)
{
    ether_input((struct ifnet *) ifp, eh, (struct mbuf *) m);
    return S_OK;
}


PRIVATE SCODE MCT
CEndpointFactoryEtherOutput(PIENDPOINTFACTORY pThis, PINETDRIVER ifp,
                            PIBUFFER m, struct sockaddr *dst,
                            struct route_in *rt0)
{
    return ether_output((struct ifnet *) ifp, (struct mbuf *) m, dst, rt0);
}


PRIVATE void MCT 
CEndpointFactoryNetLock(PIENDPOINTFACTORY pThis)
{
    NetLock();
}


PRIVATE void MCT 
CEndpointFactoryNetUnlock(PIENDPOINTFACTORY pThis)
{
    NetUnlock();
}

PRIVATE SCODE MCT 
CEndpointFactoryIfAttach(PIENDPOINTFACTORY pThis, PINETDRIVER ifp)
{
    if_attach((struct ifnet *) ifp);
    return S_OK;
}

PRIVATE void MCT
CEndpointFactorySignalCondition(PIENDPOINTFACTORY pThis, PCONDITION pCond)
{
    /* Barrier */
    Mutex_Lock(&NetGlobalLock.Mutex);
    Mutex_Unlock(&NetGlobalLock.Mutex);

    Condition_Signal(pCond);
}

PRIVATE SCODE MCT
CEndpointFactoryWaitCondition(PIENDPOINTFACTORY pThis,
                              PCONDITION pCond, PMUTEX pMutex,
                              PTIME pTimeOut)
{
    SCODE sc;

    /* Same locking order prevents race with WaitForEndpoints probe:
     * i.e. hold pMutex while calling WaitForEndpoints(TIME_ORIGIN);
     */
    Mutex_Lock(&NetGlobalLock.Mutex);

    if (pMutex)
        Mutex_Unlock(pMutex);

    sc = Condition_Wait(pCond, &NetGlobalLock.Mutex, pTimeOut);

    Mutex_Unlock(&NetGlobalLock.Mutex);

    if (pMutex)
        Mutex_Lock(pMutex);

    return sc;
}

#include <in.h>
#include <dhcp/dhcp_var.h>
#include <dns/inet.h>
#include <net/if.h>
#include <net/ip/in_var.h>

PRIVATE SCODE MCT parseip(const _TCHAR *cmdline, struct in_addr *inap);

/* XXX: eats arbitrary amount of input, not very general-purpose */
PRIVATE SCODE MCT
parseip(const _TCHAR *cmdline, struct in_addr *inap)
{
    UINT len;
#define MAXIP 16
    char ip[MAXIP+1];

    if (!*cmdline)
        return !S_OK;

    len = _tcscspn(cmdline, _T(" \t"));
    if (len == 0 || len > MAXIP)
        return !S_OK;

    _tcsncpy(ip, cmdline, len);
    ip[len] = _EOS;
    inap->s_addr = inet_addr(ip);

    return S_OK;
}

PRIVATE SCODE MCT
CEndpointFactoryMain(PIENDPOINTFACTORY pThis, const _TCHAR *cmdline)
{
    SCODE sc = S_OK;
    UINT len;
    const _TCHAR *arg1 = NULL;
#define MAXIFNAME 10
    _TCHAR ifname[MAXIFNAME+1];
    struct in_addr ina;
    struct ifaliasreq   addreq;
    struct ifnet *ifp;
    PNETFACTORY This = pNF(pThis);

    if (!cmdline || !*cmdline)
        goto Usage;

    len = _tcscspn(cmdline, _T(" \t"));
    if (len == 0 || len > MAXIFNAME)
        goto Usage;
    _tcsncpy(ifname, cmdline, len);
    ifname[len] = _EOS;

    cmdline += len;
    if (*cmdline)
        cmdline++;

    /* XXX: this adhoc parser is getting overcrowded */

    if (!_tcscmp(ifname, _T("start"))) {
        This->Started = TRUE;
        sc = ProtocolStart(PROTOCOL_START_DRIVERS | PROTOCOL_START_LO);
        goto out;
    }

    if (!_tcscmp(ifname, _T("dhcp"))) {
        UINT Flags = PROTOCOL_START_DHCP;
        if (!This->Started)
            Flags |= PROTOCOL_START_DRIVERS | PROTOCOL_START_LO;
        This->Started = TRUE;
        sc = ProtocolStart(Flags);
        goto out;
    }

    if (!_tcscmp(ifname, _T("gateway"))) {
        if (parseip(cmdline, &ina) != S_OK)
            goto Usage;

        sc = rt_addgw(ina);
        goto out;
    }

    if (!_tcscmp(ifname, _T("multi-if"))) {
        extern struct ifaddr *ifa_multiaddr;
        struct sockaddr_in ifsin;
        struct ifaddr *ifa;

        memset(&ifsin, 0, sizeof(ifsin));
        if (parseip(cmdline, &ifsin.sin_addr) != S_OK)
            goto Usage;
        ifsin.sin_family = AF_INET;
        ifsin.sin_len = sizeof(struct sockaddr_in);

        ifa = ifa_ifwithaddr((struct sockaddr *)&ifsin);
        if (!ifa) {
            _tprintf(_T("no interface with such address\n"));
            sc = E_INVALID_PARAMETER;
            goto out;
        }

        ifa_multiaddr = ifa;
        goto out;
    }

#define CMNDCMP(a,b) _tcsncmp(a, b, sizeof(b)-1)

    if (!CMNDCMP(cmdline, "up")) {
    } else if (!CMNDCMP(cmdline, "up")) {
    } else if (!CMNDCMP(cmdline, "down")) {
    } else if (!CMNDCMP(cmdline, "dhcp")) {
    } else if (!CMNDCMP(cmdline, "autoip")) {
    } else if (!CMNDCMP(cmdline, "inet")) {
         /* Macro for accessing sockaddr as sockaddr_in */
#define sin(_sa_) (*(struct sockaddr_in *)&(_sa_))

        /* XXX */
        cmdline += sizeof("inet")-1;
        if (*cmdline)
            cmdline++;

        if (parseip(cmdline, &ina) != S_OK)
            goto Usage;

        ifp = ifunit(ifname);
        if (!ifp)
            goto Usage;

        memset(&addreq, 0, sizeof(addreq));
        _tcsncpy(addreq.ifra_name, ifname, sizeof addreq.ifra_name);

        sin(addreq.ifra_addr).sin_len = sizeof(struct sockaddr_in);
        sin(addreq.ifra_addr).sin_family = AF_INET;
        sin(addreq.ifra_addr).sin_addr = ina;

        sin(addreq.ifra_mask).sin_len = sizeof(struct sockaddr_in);
        sin(addreq.ifra_mask).sin_family = AF_INET;
        sin(addreq.ifra_mask).sin_addr.s_addr = ntohl(0xffff0000); /* XXXXXX */

        sin(addreq.ifra_broadaddr).sin_len = sizeof(struct sockaddr_in);
        sin(addreq.ifra_broadaddr).sin_family = AF_INET;
        sin(addreq.ifra_broadaddr).sin_addr.s_addr =
          sin(addreq.ifra_addr).sin_addr.s_addr
          & sin(addreq.ifra_mask).sin_addr.s_addr;

        sc = in_control(NULL, SIOCAIFADDR, &addreq, ifp);
        _tprintf(_T("ip = \"%hs\", a = 0x%08x b = 0x%08x m = 0x%08x\r\n"),
               inet_ntoa(ina),
               sin(addreq.ifra_addr).sin_addr.s_addr,
               sin(addreq.ifra_broadaddr).sin_addr.s_addr,
               sin(addreq.ifra_mask).sin_addr.s_addr);

    }
 out:
    return sc;

 Usage:
    printf("Usage: start | dhcp | gateway IP | multi-if IP | IF inet IP\n");
    return E_INVALID_PARAMETER;
}

struct _NETFACTORY TheNetFactory;
#if 0
 = {
    {&CEndpointFactoryVtbl,},
    1,
    FALSE
};
#endif

/*
 * Entry point: Create an Endpoint factory and return it to cobmgr
 */

/* XXX: kludged beyond all recognition */
int domaininitdone;

PIUNKNOWN protocolCobMain(void)
{

    if (!domaininitdone) {
        domaininitdone = 1;

        mbinit();
        SyncPrimInit();
        domaininit();
        loopattach();

#if MONARCH_ADHOC
        ifdsr_attach(NULL);
#endif /* MONARCH_ADHOC */

        TRACE("ProtocolCobMain");
    }

    return GenericCobInit((void *) &protocol_CobDesc, (void *) &TheNetFactory); 
    //return (PIUNKNOWN) &TheNetFactory;
}
