/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1989, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)if_ethersubr.c    8.1 (Berkeley) 6/10/93
 */

#include <mmmacros.h>

#include <sock.h>
#include <diagnostics.h>

#include <if.h>
#include <if_dl.h>
#include <route.h>
#include <if_llc.h>
#include <if_types.h>

#ifdef INET
#include <in.h>
#include <in_var.h>
#include <ip_var.h>
#endif
#include <arp.h>

const UINT8 etherbroadcastaddr[6] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
extern struct ifnet loif;
#define senderr(e) { error = (e); goto bad;}

#if MONARCH_ADHOC
#include <dsr/dsr_include.h>
#endif

/*
 * Ethernet output routine.
 * Encapsulate a packet of type family for the local net.
 * Use trailer local net encapsulation if enough data in first
 * packet leaves a multiple of 512 bytes of data in remainder.
 * Assumes that ifp is actually pointer to arpcom structure.
 */
INT RTLCALLTYPE
ether_output(struct ifnet *ifp, struct mbuf *m0, struct sockaddr *dst,
    struct route_in *rti)
{
    struct route_in rti_local;
    short type;
    int error = 0;
    UINT8 edst[6];
    struct mbuf *m = m0;
    struct mbuf *mcopy = (struct mbuf *)0;
    struct ether_header *eh;
    INT off, len = m->m_pkthdr.len;
    struct arpcom *ac = (struct arpcom *)ifp;
    INT copyflags = (INT)M_COPYALL | ((ifp->if_flags & IFF_ZEROCOPY) ? M_COPYDEEP : 0);
    INT family;

    assert(dst || rti);

    if (len > ETHERMTU)
        panic(_T("ether_output: > MTU !"));

    if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING))
        senderr(ENETDOWN);

#ifdef __WANTS_SNMP
    ifp->if_lastchange = time();
#endif

    if (dst)
        family = dst->sa_family;
    else
        family = AF_INET;

    switch (family) {

#ifdef INET
    case AF_INET:
        /* XXX: artifact of lack of design */
        if (!rti) {
            rti_local.rt_valid = 1;
            rti_local.rt_gw = *(struct sockaddr_in *)dst;
            rti = &rti_local;
        }

        if (!arpresolve(ac, rti, m, edst))
            return (0);  /* if not yet resolved */
        /* If broadcasting on a simplex interface, loopback a copy */
        if ((m->m_flags & (M_BCAST | M_MCAST)) &&
            (ifp->if_flags & IFF_SIMPLEX))
            mcopy = m_copy(m, 0, copyflags);
        off = m->m_pkthdr.len - m->m_len;
        type = ETHERTYPE_IP;
        break;
#endif

    case AF_UNSPEC:
        eh = (struct ether_header *)dst->sa_data;
        bcopy((caddr_t)eh->ether_dhost, (caddr_t)edst, sizeof (edst));
        type = eh->ether_type;
        break;

    default:
        DBGME(3,printf(_T("%hs%d: can't handle af%d\n"), ifp->if_name, ifp->if_unit,
               dst->sa_family));
        senderr(EAFNOSUPPORT);
    }


    if (mcopy)
        (void) looutput(iIF(ifp), mcopy, dst, rti);

    /*
     * Add local net header.  If no space in first mbuf, allocate another.
     */
    M_PREPEND(m, sizeof(struct ether_header), M_DONTWAIT);
    if (m == 0)
        senderr(ENOBUFS);
    eh = mtod(m, struct ether_header *);
    type = htons((UINT16)type);
    bcopy((caddr_t)&type,(caddr_t)&eh->ether_type, sizeof(eh->ether_type));
    bcopy((caddr_t)edst, (caddr_t)eh->ether_dhost, sizeof (edst));
    bcopy((caddr_t)ac->ac_enaddr, (caddr_t)eh->ether_shost,
          sizeof(eh->ether_shost));

#if 0
	{
	UINT8 *dh, *sh;

	dh = (UINT8 *)&eh->ether_dhost;
	sh = (UINT8 *)&eh->ether_shost;
	printf("sending ethernet frame to %x:%x:%x:%x:%x:%x from "
	  "%x:%x:%x:%x:%x:%x\n", dh[0], dh[1], dh[2], dh[3], dh[4], dh[5],
	  sh[0], sh[1], sh[2], sh[3], sh[4], sh[5]);
	}
#endif
         
    NetLock();
    /*
     * Queue message on interface, and start output if interface
     * not yet active.
     */
    if (IF_QFULL(&ifp->if_snd)) {
        IF_DROP(&ifp->if_snd);
        NetUnlock();
        senderr(ENOBUFS);
    }
    IF_ENQUEUE(&ifp->if_snd, m);
    if ((ifp->if_flags & IFF_OACTIVE) == 0)
        (*ifp->v->Start)(iIF(ifp));
    NetUnlock();
    ifp->if_obytes += len + sizeof (struct ether_header);
    if (m->m_flags & (M_MCAST|M_BCAST))
        ifp->if_omcasts++;
    return (error);

bad:
    if (m)
        m_freem(m);
    return (error);
}

/*
 * Process a received Ethernet packet;
 * the packet is in the mbuf chain m without
 * the ether header, which is provided separately.
 */
void RTLCALLTYPE
ether_input(struct ifnet *ifp, struct ether_header *eh, struct mbuf *m)
{
    struct arpcom *ac = (struct arpcom *)ifp;
    UnusedParameter(ac);

#if defined (ISO) || defined (LLC)
    struct llc *l;
#endif

#if MONARCH_ADHOC
    if ((ifp->if_flags & IFF_UP) == 0 ||
        check_topology(ac, eh, m, "ether_input") == 0) {
        m_freem(m);
        return;
    }
#else
    if ((ifp->if_flags & IFF_UP) == 0) {
        m_freem(m);
        return;
    }
#endif

#ifdef __WANTS_SNMP
    ifp->if_lastchange = time();
#endif
    ifp->if_ibytes += m->m_pkthdr.len + sizeof (*eh);
    if (bcmp((caddr_t)etherbroadcastaddr, (caddr_t)eh->ether_dhost,
        sizeof(etherbroadcastaddr)) == 0)
        m->m_flags |= M_BCAST;
    else if (eh->ether_dhost[0] & 1)
        m->m_flags |= M_MCAST;
    if (m->m_flags & (M_BCAST|M_MCAST))
        ifp->if_imcasts++;

    switch (eh->ether_type) {
#ifdef INET
    case ETHERTYPE_IP:
        ip_input(m);
        break;

    case ETHERTYPE_ARP:
        arp_input(m);
        break;
#endif
    default:
        m_freem(m);
        return;
    }
}

/*
 * Convert Ethernet address to printable (loggable) representation.
 */
static const char digits[] = "0123456789abcdef";
char *
ether_sprintf(UINT8 *ap)
{
    int i;
    static char etherbuf[18];
    char *cp = etherbuf;

    for (i = 0; i < 6; i++) {
        *cp++ = digits[*ap >> 4];
        *cp++ = digits[*ap++ & 0xf];
        *cp++ = ':';
    }
    *--cp = 0;
    return (etherbuf);
}

/*
 * Perform common duties while attaching to interface list
 */
void
ether_ifattach(struct ifnet *ifp)
{
    struct ifaddr *ifa;
    struct sockaddr_dl *sdl;

    ifp->if_type = IFT_ETHER;
    ifp->if_addrlen = 6;
    ifp->if_hdrlen = 14;
    ifp->if_mtu = ETHERMTU;
    for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
        sdl = (struct sockaddr_dl *)ifa->ifa_addr;
        if (sdl != NULL && sdl->sdl_family == AF_LINK) {
            sdl->sdl_type = IFT_ETHER;
            sdl->sdl_alen = ifp->if_addrlen;
            bcopy((caddr_t)((struct arpcom *)ifp)->ac_enaddr,
                  LLADDR(sdl), ifp->if_addrlen);
            break;
        }
    }
}

const UINT8 ether_ipmulticast_min[6] = { 0x01, 0x00, 0x5e, 0x00, 0x00, 0x00 };
const UINT8 ether_ipmulticast_max[6] = { 0x01, 0x00, 0x5e, 0x7f, 0xff, 0xff };

/*
 * Add an Ethernet multicast address or range of addresses to the list for a
 * given interface.
 */
INT
ether_addmulti(struct ifreq *ifr, struct arpcom *ac)
{
    struct ether_multi *enm;
    struct sockaddr_in *sin;
    UINT8 addrlo[6];
    UINT8 addrhi[6];

    NetLock();
    switch (ifr->ifr_addr.sa_family) {

    case AF_UNSPEC:
        bcopy(ifr->ifr_addr.sa_data, addrlo, 6);
        bcopy(addrlo, addrhi, 6);
        break;

#ifdef INET
    case AF_INET:
        sin = (struct sockaddr_in *)&(ifr->ifr_addr);
        if (sin->sin_addr.s_addr == INADDR_ANY) {
            /*
             * An IP address of INADDR_ANY means listen to all
             * of the Ethernet multicast addresses used for IP.
             * (This is for the sake of IP multicast routers.)
             */
            bcopy(ether_ipmulticast_min, addrlo, 6);
            bcopy(ether_ipmulticast_max, addrhi, 6);
        } else {
            ETHER_MAP_IP_MULTICAST(&sin->sin_addr, addrlo);
            bcopy(addrlo, addrhi, 6);
        }
        break;
#endif

    default:
        NetUnlock();
        return (EAFNOSUPPORT);
    }

    /*
     * Verify that we have valid Ethernet multicast addresses.
     */
    if ((addrlo[0] & 0x01) != 1 || (addrhi[0] & 0x01) != 1) {
        NetUnlock();
        return (EINVAL);
    }
    /*
     * See if the address range is already in the list.
     */
    ETHER_LOOKUP_MULTI(addrlo, addrhi, ac, enm);
    if (enm != NULL) {
        /*
         * Found it; just increment the reference count.
         */
        ++enm->enm_refcount;
        NetUnlock();
        return (0);
    }
    /*
     * New address or range; malloc a new multicast record
     * and link it into the interface's multicast list.
     */
    enm = (struct ether_multi *)bsd_malloc(sizeof(*enm), M_IFMADDR, M_NOWAIT);
    if (enm == NULL) {
        NetUnlock();
        return (ENOBUFS);
    }
    bcopy(addrlo, enm->enm_addrlo, 6);
    bcopy(addrhi, enm->enm_addrhi, 6);
    enm->enm_ac = ac;
    enm->enm_refcount = 1;
    enm->enm_next = ac->ac_multiaddrs;
    ac->ac_multiaddrs = enm;
    ac->ac_multicnt++;
    NetUnlock();
    /*
     * Return ENETRESET to inform the driver that the list has changed
     * and its reception filter should be adjusted accordingly.
     */
    return (ENETRESET);
}

/*
 * Delete a multicast address record.
 */
INT
ether_delmulti(struct ifreq *ifr, struct arpcom *ac)
{
    struct ether_multi *enm;
    struct ether_multi **p;
    struct sockaddr_in *sin;
    UINT8 addrlo[6];
    UINT8 addrhi[6];

    NetLock();
    switch (ifr->ifr_addr.sa_family) {

    case AF_UNSPEC:
        bcopy(ifr->ifr_addr.sa_data, addrlo, 6);
        bcopy(addrlo, addrhi, 6);
        break;

#ifdef INET
    case AF_INET:
        sin = (struct sockaddr_in *)&(ifr->ifr_addr);
        if (sin->sin_addr.s_addr == INADDR_ANY) {
            /*
             * An IP address of INADDR_ANY means stop listening
             * to the range of Ethernet multicast addresses used
             * for IP.
             */
            bcopy(ether_ipmulticast_min, addrlo, 6);
            bcopy(ether_ipmulticast_max, addrhi, 6);
        } else {
            ETHER_MAP_IP_MULTICAST(&sin->sin_addr, addrlo);
            bcopy(addrlo, addrhi, 6);
        }
        break;
#endif

    default:
        NetUnlock();
        return (EAFNOSUPPORT);
    }

    /*
     * Look up the address in our list.
     */
    ETHER_LOOKUP_MULTI(addrlo, addrhi, ac, enm);
    if (enm == NULL) {
        NetUnlock();
        return (EINVAL); /* af: was ENXIO */
    }
    if (--enm->enm_refcount != 0) {
        /*
         * Still some claims to this record.
         */
        NetUnlock();
        return (0);
    }
    /*
     * No remaining claims to this record; unlink and free it.
     */
    for (p = &enm->enm_ac->ac_multiaddrs;*p != enm;p = &(*p)->enm_next)
        continue;
    *p = (*p)->enm_next;
    bsd_free(enm, M_IFMADDR);
    ac->ac_multicnt--;
    NetUnlock();
    /*
     * Return ENETRESET to inform the driver that the list has changed
     * and its reception filter should be adjusted accordingly.
     */
    return (ENETRESET);
}

/*
 * Ioctl code common to all ethernet interfaces
 */
int ether_ioctl(struct ifnet *ifp, INT cmd, caddr_t data)
{
    struct ifaddr *ifa = (struct ifaddr *)data;
    struct arpcom *ac = (struct arpcom *)ifp;
    int error = 0;

    NetLock();
    switch (cmd) {

    case SIOCSIFADDR:
        ifp->if_flags |= IFF_UP;
        switch (ifa->ifa_addr->sa_family) {
#ifdef INET
        case AF_INET:
            error = (*ifp->v->Initialize)(iIF(ifp));  /* before arpwhohas */
            if (error)
                goto out;
            ac->ac_ipaddr =    IA_SIN(ifa)->sin_addr;
            arpwhohas(ac, &IA_SIN(ifa)->sin_addr);
            break;
#endif
        default:
            error = (*ifp->v->Initialize)(iIF(ifp));
            break;
        }
        break;

#ifdef MULTICAST
    case SIOCADDMULTI:
    case SIOCDELMULTI:
        /* Update our multicast list  */
        error = (cmd == SIOCADDMULTI) ?
            ether_addmulti((struct ifreq *)data, ac) :
            ether_delmulti((struct ifreq *)data, ac);

        if (error == ENETRESET) {
            /*
             * Multicast list has changed; set the hardware
             * filter accordingly.
             */
            error = (*ifp->v->Initialize)(iIF(ifp));
        }
        break;
#endif

    default:
        error = EINVAL;
    }

out:
    NetUnlock();
    return (error);
}
