/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1980, 1986, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)if.c    8.1 (Berkeley) 6/29/93
 */

#include <sock.h>
#include <string.h>
#include <route.h>

#include <if.h>
#include <if_dl.h>
#include <if_types.h>
#include <in.h>
#include <dhcp_var.h>

/*
 * Network interface utility routines.
 *
 * Routines with ifa_ifwith* names take sockaddr *'s as
 * parameters.
 */
UINT16 if_index = 0;
struct ifaddr **ifnet_addrs;
static char *sprint_d (UINT, char *, INT);

/*
 * Attach an interface to the
 * list of "active" interfaces.
 */
void RTLCALLTYPE
if_attach(struct ifnet *ifp)
{
    UINT socksize, ifasize, masklen;
    INT namelen, unitlen;
    char workbuf[12], *unitname;
    struct ifnet **p = &ifnet;
    struct sockaddr_dl *sdl;
    struct ifaddr *ifa;
    static INT if_indexlim = 8;

    if (ifp->if_snd.ifq_maxlen == 0)
        ifp->if_snd.ifq_maxlen = IFQ_MAXLEN;

    /*
     * Initialize the DHCP info now (before we link into interface list)
     * so that SIOCGDHCPINFO and the Dhcp timers work correctly at all times.
     */
    DhcpInitStoredInfo(ifp);

    ifp->if_next = NULL;  /* paranoia */
    while (*p)
        p = &((*p)->if_next);
    *p = ifp;
    ifp->if_index = ++if_index;
    if (ifnet_addrs == 0 || if_index >= if_indexlim) {
        unsigned n = (if_indexlim <<= 1) * sizeof(ifa);
        struct ifaddr **q = (struct ifaddr **)
                    bsd_malloc(n, M_IFADDR, M_WAITOK);
        if (ifnet_addrs) {
            bcopy((char *)ifnet_addrs, (char *)q, n/2);
            bsd_free((char *)ifnet_addrs, M_IFADDR);
        }
        ifnet_addrs = q;
    }
    /*
     * create a Link Level name for this device
     */
    unitname = sprint_d((UINT)ifp->if_unit, workbuf, sizeof(workbuf));
    namelen = strlen(ifp->if_name);
    unitlen = strlen(unitname);
#define _offsetof(t, m) ((INT)((char *)&((t *)0)->m))
    masklen = _offsetof(struct sockaddr_dl, sdl_data[0]) +
                   unitlen + namelen;
    socksize = masklen + ifp->if_addrlen;
#define ROUNDUP(a) (1 + (((a) - 1) | (sizeof(long) - 1)))
    socksize = ROUNDUP(socksize);
    if (socksize < sizeof(*sdl))
        socksize = sizeof(*sdl);
    ifasize = sizeof(*ifa) + 2 * socksize;
    ifa = (struct ifaddr *)bsd_malloc(ifasize, M_IFADDR, M_WAITOK);
    if (ifa) {
        bzero((char *)ifa, ifasize);
        sdl = (struct sockaddr_dl *)(ifa + 1);
        sdl->sdl_len = (UINT8)socksize;
        sdl->sdl_family = AF_LINK;
        bcopy(ifp->if_name, sdl->sdl_data, namelen);
        bcopy(unitname, namelen + (char *)sdl->sdl_data, unitlen);
        sdl->sdl_nlen = (UINT8)(namelen += unitlen);
        sdl->sdl_index = ifp->if_index;
        sdl->sdl_type = ifp->if_type;
        ifnet_addrs[if_index - 1] = ifa;
        ifa->ifa_ifp = ifp;
        ifa->ifa_next = ifp->if_addrlist;
        ifp->if_addrlist = ifa;
        ifa->ifa_addr = (struct sockaddr *)sdl;
        sdl = (struct sockaddr_dl *)(socksize + (char *)sdl);
        ifa->ifa_netmask = (struct sockaddr *)sdl;
        sdl->sdl_len = (UINT8)masklen;
        while (namelen != 0)
            sdl->sdl_data[--namelen] = -1;
    }

    if (ifp->if_type == IFT_ETHER)
        ether_ifattach(ifp);

#if 0
    /* This call is redundant. The device handling thread
     * makes this same call, right after we do it. And then
     * another call is made when the interface is given
     * its protocol-specific address.
     */
    if (ifp->v->Initialize)
        (*ifp->v->Initialize)(ifp);
#endif
}

void
ifafree(struct ifaddr *ifa)
{

    if (ifa == NULL)
        panic(_T("ifafree"));
    if (ifa->ifa_refcnt == 0)
        bsd_free(ifa, M_IFADDR);
    else
        ifa->ifa_refcnt--;
}

/*
 * Locate an interface based on a complete address.
 */
struct ifaddr *
ifa_ifwithaddr(struct sockaddr *addr)
{
    struct ifnet *ifp;
    struct ifaddr *ifa;

#define    equal(a1, a2) \
  (bcmp((char *)(a1), (char *)(a2), ((struct sockaddr *)(a1))->sa_len) == 0)
    for (ifp = ifnet; ifp; ifp = ifp->if_next)
        for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
        if (ifa->ifa_addr->sa_family != addr->sa_family)
            continue;
        if (equal(addr, ifa->ifa_addr))
            return (ifa);
        if ((ifp->if_flags & IFF_BROADCAST) && ifa->ifa_broadaddr &&
            equal(ifa->ifa_broadaddr, addr))
            return (ifa);
    }
    return ((struct ifaddr *)0);
}
/*
 * Locate the point to point interface with a given destination address.
 */
struct ifaddr *
ifa_ifwithdstaddr(struct sockaddr *addr)
{
    struct ifnet *ifp;
    struct ifaddr *ifa;

    for (ifp = ifnet; ifp; ifp = ifp->if_next) 
        if (ifp->if_flags & IFF_POINTOPOINT)
        for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
            if (ifa->ifa_addr->sa_family != addr->sa_family)
                continue;
            if (equal(addr, ifa->ifa_dstaddr))
                return (ifa);
    }
    return ((struct ifaddr *)0);
}

/*
 * Find an interface on a specific network.  XXX: should return
 * most specific found, but does not do that right now.
 */
struct ifaddr *
ifa_ifwithnet(struct sockaddr *addr)
{
    struct ifnet *ifp;
    struct ifaddr *ifa;
    struct ifaddr *ifa_maybe = (struct ifaddr *) 0;
    UINT af = addr->sa_family;
    UINT8 *addr_data = addr->sa_data, *cplim;

    if (af == AF_LINK) {
        struct sockaddr_dl *sdl = (struct sockaddr_dl *)addr;
        if (sdl->sdl_index && sdl->sdl_index <= if_index)
        return (ifnet_addrs[sdl->sdl_index - 1]);
    }
    for (ifp = ifnet; ifp; ifp = ifp->if_next)
        for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
        UINT8 *cp, *cp2, *cp3;

        if (ifa->ifa_addr->sa_family != af || ifa->ifa_netmask == 0)
            next: continue;
        cp = addr_data;
        cp2 = ifa->ifa_addr->sa_data;
        cp3 = ifa->ifa_netmask->sa_data;
        cplim = ifa->ifa_netmask->sa_len + (UINT8 *)ifa->ifa_netmask;
        while (cp3 < cplim)
            if ((*cp++ ^ *cp2++) & *cp3++)
                goto next;
        return (ifa);
    }

    return NULL;
}

#if 0
/*
 * Find an interface using a specific address family
 */
struct ifaddr *
ifa_ifwithaf(INT af)
{
    struct ifnet *ifp;
    struct ifaddr *ifa;

    for (ifp = ifnet; ifp; ifp = ifp->if_next)
        for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next)
        if (ifa->ifa_addr->sa_family == af)
            return (ifa);
    return ((struct ifaddr *)0);
}
#endif

/*
 * Find an interface address specific to an interface best matching
 * a given address.
 */
struct ifaddr *
ifaof_ifpforaddr(struct sockaddr *addr, struct ifnet *ifp)
{
    struct ifaddr *ifa;
    UINT8 *cp, *cp2, *cp3;
    UINT8 *cplim;
    struct ifaddr *ifa_maybe = 0;
    UINT af = addr->sa_family;

    if (af >= AF_MAX)
        return (0);
    for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
        if (ifa->ifa_addr->sa_family != af)
            continue;
        ifa_maybe = ifa;
        if (ifa->ifa_netmask == 0) {
            if (equal(addr, ifa->ifa_addr) ||
                (ifa->ifa_dstaddr && equal(addr, ifa->ifa_dstaddr)))
                return (ifa);
            continue;
        }
        cp = addr->sa_data;
        cp2 = ifa->ifa_addr->sa_data;
        cp3 = ifa->ifa_netmask->sa_data;
        cplim = ifa->ifa_netmask->sa_len + (UINT8 *)ifa->ifa_netmask;
        for (; cp3 < cplim; cp3++)
            if ((*cp++ ^ *cp2++) & *cp3)
                break;
        if (cp3 == cplim)
            return (ifa);
    }
    return (ifa_maybe);
}

/*
 * Mark an interface down and notify protocols of
 * the transition.
 * NOTE: must be called at splnet or equivalent.
 */
void
if_down(struct ifnet *ifp)
{
    struct ifaddr *ifa;

    ifp->if_flags &= ~IFF_UP;
    for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next)
        pfctlinput(PRC_IFDOWN, ifa->ifa_addr);
    if_qflush(&ifp->if_snd);
}

/*
 * Flush an interface queue.
 */
void
if_qflush(struct ifqueue *ifq)
{
    struct mbuf *m, *n;

    n = ifq->ifq_head;
    while ((m = n) != NULL) {
        n = m->m_nextpkt;
        m_freem(m);
    }
    ifq->ifq_head = 0;
    ifq->ifq_tail = 0;
    ifq->ifq_len = 0;
}

/*
 * Handle interface watchdog timer routines.  Called
 * from softclock, we decrement timers (if set) and
 * call the appropriate interface routine on expiration.
 */
void
if_slowtimo( void )
{
    struct ifnet *ifp;

    NetLock();
    for (ifp = ifnet; ifp; ifp = ifp->if_next) {
        if (ifp->if_timer != 0) {
            ifp->if_timer--;
            if (ifp->if_timer)
                continue;
        }
        if (ifp->v->Watchdog)
            (*ifp->v->Watchdog)(iIF(ifp));
    }
    NetUnlock();
#if 0
    timeout(if_slowtimo, (void *)0, hz / IFNET_SLOWHZ);
#endif
}

/*
 * Map interface name to
 * interface structure pointer.
 */
struct ifnet *
ifunit(char *name)
{
    char *cp;
    struct ifnet *ifp;
    INT unit;
    unsigned len;
    char *ep, c;

    for (cp = name; cp < name + IFNAMSIZ && *cp; cp++)
        if (*cp >= '0' && *cp <= '9')
            break;
    if (*cp == '\0' || cp == name + IFNAMSIZ)
        return ((struct ifnet *)0);
    /*
     * Check for wildcard case.  If first character of
     * interface name is '*', match first interface that
     * isn't the loopback.
     */
    if (*name == '*') {
        for (ifp = ifnet; ifp; ifp = ifp->if_next) {
            if (bcmp(ifp->if_name, "lo", 2))
                break;
        }
        return (ifp);
    }
    /*
     * Save first char of unit, and pointer to it,
     * so we can put a null there to avoid matching
     * initial substrings of interface names.
     */
    len = cp - name + 1;
    c = *cp;
    ep = cp;
    for (unit = 0; *cp >= '0' && *cp <= '9'; )
        unit = unit * 10 + *cp++ - '0';
    *ep = 0;
    for (ifp = ifnet; ifp; ifp = ifp->if_next) {
        if (bcmp(ifp->if_name, name, len))
            continue;
        if (unit == ifp->if_unit)
            break;
    }
    *ep = c;
    return (ifp);
}

/*
 * Interface ioctls.
 */
INT ifioctl(struct socket *so, INT cmd, UINT8 * data)
{
    struct ifnet *ifp;
    struct ifreq *ifr;
    INT error;

    switch (cmd) {

    case SIOCGIFCONF:
    case OSIOCGIFCONF:
        return (ifconf(cmd, data));
    }
    ifr = (struct ifreq *)data;
    ifp = ifunit(ifr->ifr_name);

    if (ifp == 0)
        return (EINVAL); /* af: was ENXIO */

    switch (cmd) {

    case SIOCGIFFLAGS:
        ifr->ifr_flags = ifp->if_flags;
        break;

    case SIOCGIFMETRIC:
        ifr->ifr_metric = ifp->if_metric;
        break;

    case SIOCSIFFLAGS:
#ifdef SECURITY
        if (error = suser(p->p_ucred, &p->p_acflag))
            return (error);
#endif
        NetLock();
        if (ifp->if_flags & IFF_UP && (ifr->ifr_flags & IFF_UP) == 0) {
            if_down(ifp);
        }
        if (ifr->ifr_flags & IFF_UP && (ifp->if_flags & IFF_UP) == 0) {
            ifp->if_flags |= IFF_UP;
        }
        ifp->if_flags = (UINT16)((ifp->if_flags & IFF_CANTCHANGE) |
            (ifr->ifr_flags &~ IFF_CANTCHANGE));
        if (ifp->v->Ioctl)
            (void) (*ifp->v->Ioctl)(iIF(ifp), cmd, data);
        NetUnlock();
        break;

    case SIOCSIFMETRIC:
#ifdef SECURITY
        if (error = suser(p->p_ucred, &p->p_acflag))
            return (error);
#endif
        ifp->if_metric = ifr->ifr_metric;
        break;

    case SIOCADDMULTI:
    case SIOCDELMULTI:
#ifdef SECURITY
        if (error = suser(p->p_ucred, &p->p_acflag))
            return (error);
#endif
        if (ifp->v->Ioctl == NULL)
            return (EOPNOTSUPP);
        return ((*ifp->v->Ioctl)(iIF(ifp), cmd, data));

#ifdef MMLITE_EXT
    case SIOCGIFSTATS:
        /* Possibly refresh stats
         */
        if (ifp->v->Ioctl)
            (void) (*ifp->v->Ioctl)(iIF(ifp), cmd, data);
        *((struct if_data *)ifr) = ifp->if_data;
        break;

    case SIOCIDHCP:
        /*
         * Start DHCP protocol exchange on this interface.
         */
        if ((ifp->if_flags & (IFF_UP | IFF_BROADCAST)) !=
            (IFF_UP | IFF_BROADCAST))
            return ENETDOWN;
#if 0
        ifp->if_flags = (UINT16)(ifp->if_flags | IFF_UP | IFF_BROADCAST);
        (void)(*ifp->v->Ioctl)(iIF(ifp), SIOCSIFFLAGS, 0);
#endif
        DhcpStart(ifp, ifr->ifr_dhcpflags);
        return DhcpWaitTillBound(ifp);

    case SIOCGDHCPINFO:
        /*
         * Retrieve information from DHCP records.
         */
        error = DhcpGetInfo(ifp, (struct DhcpInfo *) data);
        return error;

    case SIOCGDHCPINFOPTR:
        /*
         * Return pointer to DHCP records.
         */
        *(PTR*) data = ifp->pDhcpInfo;
        return 0;

    case SIOCGDHCPOPTION:
        /*
         * Retrieve option information from DHCP records.
         */
        error = DhcpGetOption(ifp, ifr->ifr_dhcpflags,
                              (struct DhcpOption *) data);
        return error;

    case SIOCGDHCPFILE:
        /*
         * Retrieve the boot filename from DHCP records.
         */
        error = DhcpGetFile(ifp, (char *) data);
        return error;

#endif

    default:
        if (so->so_proto == 0)
            return (EOPNOTSUPP);
#ifndef COMPAT_43
        return ((*so->so_proto->pr_usrreq)(so, PRU_CONTROL,
            cmd, data, ifp));
#else
        {
        INT ocmd = cmd;

        switch (cmd) {

        case SIOCSIFDSTADDR:
        case SIOCSIFADDR:
        case SIOCSIFBRDADDR:
        case SIOCSIFNETMASK:
#if BYTE_ORDER != BIG_ENDIAN
            if (ifr->ifr_addr.sa_family == 0 &&
                ifr->ifr_addr.sa_len < 16) {
                ifr->ifr_addr.sa_family = ifr->ifr_addr.sa_len;
                ifr->ifr_addr.sa_len = 16;
            }
#else
            if (ifr->ifr_addr.sa_len == 0)
                ifr->ifr_addr.sa_len = 16;
#endif
            break;

        case OSIOCGIFADDR:
            cmd = SIOCGIFADDR;
            break;

        case OSIOCGIFDSTADDR:
            cmd = SIOCGIFDSTADDR;
            break;

        case OSIOCGIFBRDADDR:
            cmd = SIOCGIFBRDADDR;
            break;

        case OSIOCGIFNETMASK:
            cmd = SIOCGIFNETMASK;
        }
        /* XXX Args below are suspicious.  Check PTR casting validity */
        error =  ((*so->so_proto->pr_usrreq)(so, PRU_CONTROL,
                                (PTR) cmd, (PTR) data, (PTR) ifp));
        switch (ocmd) {

        case OSIOCGIFADDR:
        case OSIOCGIFDSTADDR:
        case OSIOCGIFBRDADDR:
        case OSIOCGIFNETMASK:
            *(UINT16 *)&ifr->ifr_addr = ifr->ifr_addr.sa_family;
        }
        return (error);

        }
#endif
    }
    return (0);
}

/*
 * Return interface configuration
 * of system.  List may be used
 * in later ioctl's (above) to get
 * other information.
 */
INT ifconf(INT cmd, UINT8 * data)
{
    struct ifconf *ifc = (struct ifconf *)data;
    struct ifnet *ifp = ifnet;
    struct ifaddr *ifa;
    char *cp, *ep;
    struct ifreq ifr, *ifrp;
    INT space = ifc->ifc_len, skip;

    ifrp = ifc->ifc_reql;
    ep = ifr.ifr_name + sizeof (ifr.ifr_name) - 2;
    skip = ifc->ifc_tok;
    for (; space > sizeof (ifr) && ifp; ifp = ifp->if_next) {
        if (skip-- > 0)
            continue;
        strncpy(ifr.ifr_name, ifp->if_name, sizeof (ifr.ifr_name) - 2);
        for (cp = ifr.ifr_name; cp < ep && *cp; cp++)
            continue;
        *cp++ = (char)('0' + ifp->if_unit);
        *cp = '\0';
        if ((ifa = ifp->if_addrlist) == 0) {
            bzero((char *)&ifr.ifr_addr, sizeof(ifr.ifr_addr));
            memcpy(ifrp, &ifr, sizeof (ifr));
            space -= sizeof (ifr), ifrp++;
        } else 
            for ( ; space > sizeof (ifr) && ifa; ifa = ifa->ifa_next) {
            struct sockaddr *sa = ifa->ifa_addr;
#ifdef COMPAT_43
            if (cmd == OSIOCGIFCONF) {
                struct osockaddr *osa =
                     (struct osockaddr *)&ifr.ifr_addr;
                ifr.ifr_addr = *sa;
                osa->sa_family = sa->sa_family;
                memcpy(ifrp, &ifr, sizeof (ifr));
                ifrp++;
            } else
#endif
            if (sa->sa_len <= sizeof(*sa)) {
                ifr.ifr_addr = *sa;
                memcpy(ifrp, &ifr, sizeof (ifr));
                ifrp++;
            } else {
                space -= sa->sa_len - sizeof(*sa);
                if (space < sizeof (ifr))
                    break;
                memcpy(ifrp, &ifr, sizeof (ifr.ifr_name));
                memcpy(&ifrp->ifr_addr, sa, sa->sa_len);
                ifrp = (struct ifreq *)
                    (sa->sa_len + (char *)&ifrp->ifr_addr);
            }
            space -= sizeof (ifr);
        }
    }
    ifc->ifc_len -= space;
    return (0);
}

static char *
sprint_d(UINT n, char *buf, INT buflen)
{
    char *cp = buf + buflen - 1;

    *cp = 0;
    do {
        cp--;
        *cp = "0123456789"[n % 10];
        n /= 10;
    } while (n != 0);
    return (cp);
}
