/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)if_loop.c    8.1 (Berkeley) 6/10/93
 */

/*
 * Loopback interface driver.
 */

#include <sock.h>
#include <diagnostics.h>

#include <if.h>
#include <if_types.h>
#include <route.h>

#ifdef INET
#include <in.h>
#include <in_systm.h>
#include <in_var.h>
#include <ip.h>
#include <ip_var.h>
#endif

#define LOMTU (1024+512)

struct ifnet loif;

void THREAD_LINKAGE loIntr(THREAD_ARGUMENT arg);

const struct INetDriverVtbl LoopVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL, NULL, NULL,           /* IUnknown */
    NULL, NULL, NULL, NULL,     /* IFile */
    NULL,      /* Isr */
    NULL,      /* Intr */
    NULL,      /* Init */
    NULL,      /* ThreadLoop */
    looutput,  /* Output */
    NULL,      /* Start output */
    loioctl,   /* Ioctl */
    NULL       /* Watchdog */
};

PITHREAD pLoopBackThd;

PIPROCESS GetCurrentProcessStaticScope(void);

MUTEX loGuard;  /* BUGBUG -- shouldn't be global.  Rework if queue someday */

void
loopattach( void )
{
    struct ifnet *ifp = &loif;
    PIPROCESS pIPrc;
    SCODE sc;

    Mutex_Init(&loGuard);

    ifp->if_name = "lo";
    ifp->if_mtu = LOMTU;
    ifp->if_flags = (UINT16)(IFF_LOOPBACK | IFF_MULTICAST | IFF_RUNNING);
    ifp->v = &LoopVtable;
    ifp->RefCnt = 1;
    ifp->if_type = IFT_LOOP;
    ifp->if_hdrlen = 0;
    ifp->if_addrlen = 0;
    Condition_Init(&ifp->IntrPoint);

    if_attach(ifp);

    pIPrc = GetCurrentProcessStaticScope();

    /* start input thread */
    /* Measured stack use on NTU debug is 488 bytes */
    sc = pIPrc->v->CreateThread(pIPrc, CREATE_THREAD_STATIC_SCOPING,
                                loIntr, (THREAD_ARGUMENT)ifp,
                                0/* 700 too small for DHCP output */,
                                NULL,
                                &pLoopBackThd);
    pIPrc->v->Release(pIPrc);
    if (FAILED(sc))
        panic(_T("Couldn't start loopback interface thread!\n"));

    DBGME(2,printf("protocol: Created loop-back thread %x\n", (ADDRESS) pLoopBackThd));
}

void THREAD_LINKAGE loIntr(THREAD_ARGUMENT arg)
{
    struct ifnet *ifp = (struct ifnet *)arg;
    struct ifqueue *ifq = 0;
    struct mbuf *m;
        UnusedParameter(ifq);

    for (;;) {
        /* wait for looutput to signal us */
        Mutex_Lock(&loGuard);
        while((ifp->if_flags & IFF_OACTIVE) == 0) {
            Condition_Wait(&ifp->IntrPoint, &loGuard, NULL);
        }
        ifp->if_flags &= ~IFF_OACTIVE;
        Mutex_Unlock(&loGuard);

    another:
        /* pull off a packet */
        NetLock();
        IF_DEQUEUE(&ifp->if_snd, m);
        NetUnlock();
        if (m == 0) {
            continue;  /* quit when queue exhausted */
        }

        /* record where packet came from */
        m->m_pkthdr.rcvif = ifp;

        ifp->if_ipackets++;
        ifp->if_ibytes += m->m_pkthdr.len;

#ifdef INET
        /* send it to IP (only protocol we currently support) */
        NetLock();
        ip_input(m);
        NetUnlock();
#endif
        goto another;
    }
}

/*
 * NOTE that looutput is sometimes called with some other interface
 * (not &loif) as This. For example, see ip_mloopback or ether_output.
 *
 * So we mostly ignore This and use &loif.
 *
 * However we record the original interface in m->m_pkthdr.rcvif
 * in case it should be needed someday. (At the moment loIntr
 * overwrites this field; see above.)
 */
INT MCT
looutput(PINETDRIVER This, struct mbuf *m, struct sockaddr *dst,
         struct route_in *rti)
{
    struct ifnet *ifp = &loif;

    if ((m->m_flags & M_PKTHDR) == 0)
        panic(_T("looutput no HDR"));
#ifdef __WANTS_SNMP
    ifp->if_lastchange = time();
#endif

    ifp->if_opackets++;
    ifp->if_obytes += m->m_pkthdr.len;

    /*
     * Queue message on interface and signal input thread.
     */
    NetLock();
    if (IF_QFULL(&ifp->if_snd)) {
        IF_DROP(&ifp->if_snd);
        NetUnlock();
        m_freem(m);
        return(ENOBUFS);
    }
    /* record where packet originally came from */
    m->m_pkthdr.rcvif = (struct ifnet *)This;
    IF_ENQUEUE(&ifp->if_snd, m);
    NetUnlock();
    Mutex_Lock(&loGuard);
    ifp->if_flags |= IFF_OACTIVE;
    Condition_Signal(&ifp->IntrPoint);
    Mutex_Unlock(&loGuard);
    return (0);
}

/*
 * Process an ioctl request.
 */
INT MCT
loioctl(PINETDRIVER This, UINT32 cmd, PTR data)
{
    struct ifnet *ifp = (struct ifnet *)This;
    struct ifaddr *ifa;
    struct ifreq *ifr;
    int error = 0;

    switch (cmd) {

    case SIOCSIFADDR:
        ifp->if_flags |= IFF_UP;
        ifa = (struct ifaddr *)data;
        /*
         * Everything else is done at a higher level.
         */
        break;

    case SIOCADDMULTI:
    case SIOCDELMULTI:
        ifr = (struct ifreq *)data;
        if (ifr == 0) {
            error = EAFNOSUPPORT;  /* XXX */
            break;
        }
        switch (ifr->ifr_addr.sa_family) {

#ifdef INET
        case AF_INET:
            break;
#endif

        default:
            error = EAFNOSUPPORT;
            break;
        }
        break;

    default:
        error = EINVAL;
    }
    return (error);
}
