/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * The code in here is a heavily modified derivative of BSD' MBUFs.
 * The copyright below should stay, even if the code keeps changing.
 * Perpetrator: Alessandro Forin (sandrof) Jan 2000.
 */
/*
 * Copyright (c) 1982, 1986, 1988, 1991, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)uipc_mbuf.c    8.1 (Berkeley) 7/20/93
 */

#include <sock.h>
#include <domain.h>
#include <diagnostics.h>
#include <assert.h>

/* Our buffer stats dont need to be 100% precise,
 * should the need arise just make these two macros AtomicFoo()s.
 */
#define __WANTS_STATS 0
#if __WANTS_STATS
#define STATS_INC(_x_) AtomicInc(&(_x_))
#define STATS_DEC(_x_) AtomicDec(&(_x_))
static struct mbstat mbstat;
#else
#define STATS_INC(_x_)
#define STATS_DEC(_x_)
#endif

/* Disable macros for the sake of "visual" debugging
 */
#undef m_next     /*m_hdr.mh_next*/
#undef m_len      /*m_hdr.mh_len*/
#undef m_data     /*m_hdr.mh_data*/
#undef m_type     /*m_hdr.mh_type*/
#undef m_flags    /*m_hdr.mh_flags*/
#undef m_nextpkt  /*m_hdr.mh_nextpkt*/
#undef m_pkthdr   /*M_dat.MH.MH_pkthdr*/
#undef m_ext      /*M_dat.MH.MH_dat.MH_ext*/
#undef m_pktdat   /*M_dat.MH.MH_dat.MH_databuf*/
#undef m_dat      /*M_dat.M_databuf*/

#undef mtod
#define mtod(m,t) ((t)((m)->m_hdr.mh_data))
#undef MH_ALIGN
#define MH_ALIGN(m, len) \
{ (m)->m_hdr.mh_data += (MHLEN - (len)) &~ (sizeof(long) - 1); }

#undef M_ALIGN
#define M_ALIGN(m, len) \
{ (m)->m_hdr.mh_data += (MLEN - (len)) &~ (sizeof(long) - 1); }

#define MCLOVH (offsetof(struct mbuf,M_dat.MH.MH_dat.MH_ext)+sizeof(struct m_ext))

/*
 * Section1: Memory management routines for buffers
 */

/* Declare v-table type */
#include <base/cobdesc.h>
extern const struct CUnknownVtbl CBufferVtbl;

/* Allocate an mbuf of the proper size
 */
static struct mbuf * RTLCALLTYPE
m_mballoc(INT size, INT how, INT type)
{
    struct mbuf *m;
    PIHEAP Heap = CurrentHeap();

    /*
     * get it from the heap
     */
    m = (struct mbuf *) Heap->v->Alloc(Heap, HEAP_DEBUG_NAME(M_MBUF), size, 0);
    if (m)
        goto GotIt;

    /* we might be done
     */
    if (how == M_DONTWAIT) {
        STATS_INC(mbstat.m_drops);
        return NULL;
    }

    STATS_INC(mbstat.m_wait);
 Again:
    /* ask protocols to free up some
     */
    NetLock();
    m_reclaim();
    NetUnlock();

    /*
     * retry now
     */
    m = (struct mbuf *) Heap->v->Alloc(Heap, HEAP_DEBUG_NAME(M_MBUF), size, 0);
    if (m)
        goto GotIt;

    TRACE("mballoc no mem!\n");
#if notyet
    /* sleep if necessary
     */
    Sleepers = 1;
    XXXWaitFor(&m_mballoc,"mballoc",TIME_ORIGIN);
#endif

    /* repeat
     */
    goto Again;

 GotIt:
    STATS_INC(mbstat.m_mtypes[type]);
    /* NB: Stream these assignments to memory
     */
    m->m_hdr.mh_next = (struct mbuf *)NULL;
    m->m_hdr.mh_nextpkt = (struct mbuf *)NULL;
    m->m_hdr.mh_data = (PBYTE)m->M_dat.M_databuf; /* assume small, fix later if not */
    m->m_hdr.mh_len = 0; /* empty */
    m->m_hdr.mh_type = (UINT16) (type); /* XXX should be UINT16 all the way */
    m->m_hdr.mh_flags = 0;

    m->m_hdr.v = (const void *) &CBufferVtbl.v;
    m->m_hdr.mh_RefCnt = 1;
    
    return m;
}

/* Allocates and initializes a small buffer
 */
struct mbuf * RTLCALLTYPE
m_get(INT how, INT type)
{
    return m_mballoc(MSIZE, how, type);
}

/* Same, and zeroes its content
 */
struct mbuf *
m_getclr(INT how, INT type)
{
    struct mbuf *m;

    m = m_get(how,type);
    if (m != NULL)
        memset(mtod(m,PBYTE),0,MLEN);
    return m;
}

/* Turns a small buffer into a large one
 * Or just allocate a large one if M0 is NULL.
 */
struct mbuf * RTLCALLTYPE
m_clget(struct mbuf *m0, INT how)
{
    struct mbuf *m;

    /* Allocate buffer, possibly wait
     */
    m = m_mballoc(MCLBYTES+MCLOVH,
                  how, MT_DATA);
    if (m) {
        /* Fixup from defaults
         */
        m->m_hdr.mh_data = ((PBYTE)m) + MCLOVH;
        m->M_dat.MH.MH_dat.MH_ext.ext_buf = m;
        m->M_dat.MH.MH_dat.MH_ext.ext_count = 1;
        m->M_dat.MH.MH_dat.MH_ext.ext_free = NULL;
        m->m_hdr.mh_flags |= M_EXT;

        /* Adjust M0
         */
        if (m0 == NULL) {
            /* Not a separate entity */
            m->m_hdr.mh_RefCnt = 0;
        } else {
            /* Funny non-bug: Noone knows of m so it doesnt count as ref */
            /* m->M_dat.MH.MH_dat.MH_ext.ext_count = 2; dont */

            m0->M_dat.MH.MH_dat.MH_ext.ext_buf = m;
            m0->M_dat.MH.MH_dat.MH_ext.ext_free = NULL;

            m0->m_hdr.mh_data = m->m_hdr.mh_data;
            m0->m_hdr.mh_flags |= M_EXT;
        }
    }
    return m0;
}

/* allocates an mbuf and initializes it to contain a packet header
 * and internal data.
 */
struct mbuf * RTLCALLTYPE
m_gethdr(INT how, INT type)
{
    struct mbuf *m;

    m = m_get(how, type);
    if (m) {
        m->m_hdr.mh_data = (PBYTE)m->M_dat.MH.MH_dat.MH_databuf;
        m->m_hdr.mh_flags |= M_PKTHDR;
    }
    return m;
}

/* Frees the head of a buffer chain.
 * If EXT, releases a reference down below
 * Returns the next buffer in the chain.
 */
struct mbuf *
m_free(struct mbuf *m)
{
    struct mbuf *n, *t;

    UINT RefCnt = AtomicDec(&m->m_hdr.mh_RefCnt);
    if (RefCnt > 0) {
        /* Deferred deletion of this one */
        n = m->m_hdr.mh_next;
        m->m_hdr.mh_next = NULL;
        return n;
    }

    if (m->m_hdr.mh_flags & M_EXT) {
        n = m->M_dat.MH.MH_dat.MH_ext.ext_buf;
        if (n->M_dat.MH.MH_dat.MH_ext.ext_free)
            (*(n->M_dat.MH.MH_dat.MH_ext.ext_free))(n);
        else {
            if (AtomicDec(&n->M_dat.MH.MH_dat.MH_ext.ext_count) == 0) {
                STATS_DEC(mbstat.m_mtypes[n->m_hdr.mh_type]);
                t = n->m_hdr.mh_next;
                free(n);
                if (m == n) {
                    return t;
                }
            }
        }
    }

    n = m->m_hdr.mh_next;
    STATS_DEC(mbstat.m_mtypes[m->m_hdr.mh_type]);
    free(m);

    return n;
}

/* Frees an entire buffer chain
 */
void RTLCALLTYPE
m_freem(struct mbuf *m)
{
    for (;m;m = m_free(m))
        ;
}

/* Called from MBRelease() */
void MBDelete (struct mbuf *m)
{
    NetLock();
    m->m_hdr.mh_RefCnt = 1;     /* m_free decreases it again */
    assert(!m->m_hdr.mh_nextpkt);
    m_freem(m);
    NetUnlock(); 
}

/*
 * Section2: Operations on buffers
 */

/*
 * Compute the amount of space available
 * before the current start of data in an mbuf.
 * Cant do to a large one cuz refcounts (but its fixable..)
 */
UINT m_leadingspace(struct mbuf *m)
{
    UINT len = 0;
    struct mbuf *n;

    if (m->m_hdr.mh_flags & M_EXT) {
        n = m->M_dat.MH.MH_dat.MH_ext.ext_buf;
        /* But only if no other references */
        if (n->M_dat.MH.MH_dat.MH_ext.ext_count == 1)
            len = (ADDRESS) m->m_hdr.mh_data - (ADDRESS) n->m_hdr.mh_data;
    } 
    else if (m->m_hdr.mh_flags & M_PKTHDR)
        len = (ADDRESS) m->m_hdr.mh_data - (ADDRESS) m->M_dat.MH.MH_dat.MH_databuf;
    else
        len = (ADDRESS) m->m_hdr.mh_data - (ADDRESS) m->M_dat.M_databuf;
    return len;
}

/*
 * Arrange to prepend space of size plen to mbuf m.
 * If a new mbuf must be allocated, how specifies whether to wait.
 * If how is M_DONTWAIT and allocation fails, the original mbuf chain
 * is freed and m is set to NULL.
 */
struct mbuf *
m_prepend(struct mbuf *m, UINT len, INT how, BOOL AtEnd)
{
    struct mbuf *n;

    if (m_leadingspace(m) >= len) {
        m->m_hdr.mh_data -= len;
        m->m_hdr.mh_len += len;
        if (m->m_hdr.mh_flags & M_PKTHDR)
            m->M_dat.MH.MH_pkthdr.len += len;
        return m;
    }

    n = m_get(how, m->m_hdr.mh_type);
    if (n == (struct mbuf *)NULL) {
        m_freem(m);
        return NULL;
    }
    if (m->m_hdr.mh_flags & M_PKTHDR) {
        M_COPY_PKTHDR(n, m);
        n->M_dat.MH.MH_pkthdr.len += len;
        m->m_hdr.mh_flags &= ~M_PKTHDR;
    }
    n->m_hdr.mh_next = m;
    if (len < MHLEN && AtEnd)
        MH_ALIGN(n, len);
    n->m_hdr.mh_len = len;

    return n;
}

/*
 * Make a copy of an mbuf chain starting "off0" bytes from the beginning,
 * continuing for "len" bytes.  If len is M_COPYALL, copy to end of mbuf.
 * If len is M_COPYDEEP make a full copy of large mbufs also.
 * The wait parameter is a choice of M_WAIT/M_DONTWAIT from caller.
 */
struct mbuf *
m_copym(struct mbuf *m, INT off0, INT len, INT wait)
{
    struct mbuf *n, **np;
    INT off = off0;
    struct mbuf *top;
    INT copyhdr = 0;
    static char me[] = "m_copym";

    assert (!(off < 0 || len < 0));

    if (off == 0 && m->m_hdr.mh_flags & M_PKTHDR)
        copyhdr = 1;
    while (off > 0) {
        assert (m != 0);
        if (off < m->m_hdr.mh_len)
            break;
        off -= m->m_hdr.mh_len;
        m = m->m_hdr.mh_next;
    }
    np = &top;
    top = 0;
    while (len > 0) {
        if (m == 0) {
            assert (len & M_COPYALL);
            break;
        }
        n = m_get(wait, m->m_hdr.mh_type);
        *np = n;
        if (n == 0)
            goto nospace;
        if (copyhdr) {
            M_COPY_PKTHDR(n, m);
            if (len == M_COPYALL)
                n->M_dat.MH.MH_pkthdr.len -= off0;
            else
                n->M_dat.MH.MH_pkthdr.len = len;
            copyhdr = 0;
        }
        n->m_hdr.mh_len = min(len, m->m_hdr.mh_len - off);
        if (m->m_hdr.mh_flags & M_EXT) {
            if ((len & M_COPYDEEP) == 0) {
                n->m_hdr.mh_data = m->m_hdr.mh_data + off;

                assert (m->M_dat.MH.MH_dat.MH_ext.ext_buf != NULL);

                AtomicInc(&m->M_dat.MH.MH_dat.MH_ext.ext_buf->M_dat.MH.MH_dat.MH_ext.ext_count);

                n->M_dat.MH.MH_dat.MH_ext = m->M_dat.MH.MH_dat.MH_ext;
                n->m_hdr.mh_flags |= M_EXT;
            } else {
                (void) m_clget(n,wait);
                if ((n->m_hdr.mh_flags & M_EXT) == 0)
                    goto nospace;
                bcopy(mtod(m, caddr_t)+off, mtod(n, caddr_t),
                      (unsigned)n->m_hdr.mh_len);
            }
        } else
            bcopy(mtod(m, caddr_t)+off, mtod(n, caddr_t),
                (unsigned)n->m_hdr.mh_len);
        if ((len & M_COPYALL) == 0)
            len -= n->m_hdr.mh_len;
        off = 0;
        m = m->m_hdr.mh_next;
        np = &n->m_hdr.mh_next;
    }
    return top;
nospace:
    m_freem(top);
    return NULL;
}

/*
 * Copy data from an mbuf chain starting "off" bytes from the beginning,
 * continuing for "len" bytes, into the indicated buffer.
 */
void m_copydata(struct mbuf *m, INT off, INT len, PBYTE cp)
{
    unsigned count;
    static char me[] = "m_copydata";

    assert (!(off < 0 || len < 0));

    while (off > 0) {
        assert (m != 0);

        if (off < m->m_hdr.mh_len)
            break;
        off -= m->m_hdr.mh_len;
        m = m->m_hdr.mh_next;
    }
    while (len > 0) {
        assert (m != 0);

        count = min(m->m_hdr.mh_len - off, len);
        bcopy(mtod(m, caddr_t) + off, cp, count);
        len -= count;
        cp = cp + count;
        off = 0;
        m = m->m_hdr.mh_next;
    }
}

/*
 * Copy data from a buffer back into the indicated mbuf chain,
 * starting "off" bytes from the beginning, extending the mbuf
 * chain if necessary.
 */
void
m_copyback(
        struct  mbuf *m0,
        INT off,
        INT len,
        PTR cp)
{
        INT mlen;
        struct mbuf *m = m0, *n;
        INT totlen = 0;

        if (m0 == 0)
                return;
        while (off > (mlen = m->m_hdr.mh_len)) {
                off -= mlen;
                totlen += mlen;
                if (m->m_hdr.mh_next == 0) {
                        n = m_getclr(M_DONTWAIT, m->m_hdr.mh_type);
                        if (n == 0)
                                goto out;
                        n->m_hdr.mh_len = min(MLEN, len + off);
                        m->m_hdr.mh_next = n;
                }
                m = m->m_hdr.mh_next;
        }
        while (len > 0) {
                mlen = min (m->m_hdr.mh_len - off, len);
                bcopy(cp, off + mtod(m, caddr_t), (unsigned)mlen);
                cp = (char*)cp + mlen;
                len -= mlen;
                mlen += off;
                off = 0;
                totlen += mlen;
                if (len == 0)
                        break;
                if (m->m_hdr.mh_next == 0) {
                        n = m_get(M_DONTWAIT, m->m_hdr.mh_type);
                        if (n == 0)
                                break;
                        n->m_hdr.mh_len = min(MLEN, len);
                        m->m_hdr.mh_next = n;
                }
                m = m->m_hdr.mh_next;
        }
out:    if (((m = m0)->m_hdr.mh_flags & M_PKTHDR)
	    && (m->M_dat.MH.MH_pkthdr.len < totlen))
                m->M_dat.MH.MH_pkthdr.len = totlen;
}

/*
 * Concatenate mbuf chain n to m.
 * Both chains must be of the same type (e.g. MT_DATA).
 * Any m_pkthdr is not updated.
 */
void m_cat(struct mbuf *m, struct mbuf *n)
{
    while (m->m_hdr.mh_next)
        m = m->m_hdr.mh_next;
    while (n) {
        if (m->m_hdr.mh_flags & M_EXT ||
            (ADDRESS) m->m_hdr.mh_data + m->m_hdr.mh_len + n->m_hdr.mh_len >= (ADDRESS) &m->M_dat.M_databuf[MLEN]) {
            /* just join the two chains */
            m->m_hdr.mh_next = n;
            return;
        }
        /* splat the data from one into the other */
        bcopy(mtod(n, caddr_t), mtod(m, caddr_t) + m->m_hdr.mh_len, (u_int)n->m_hdr.mh_len);
        m->m_hdr.mh_len += n->m_hdr.mh_len;
        n = m_free(n);
    }
}

/* Adjust the length of a buffer chain,
 * meaning drop data from either head or tail.
 * If REQ_LEN is negative do it at the tail.
 */
void m_adj(struct mbuf *mp, INT req_len)
{
    INT len = req_len;
    struct mbuf *m;
    INT count;

    if ((m = mp) == NULL)
        return;
    if (len >= 0) {
        /*
         * Trim from head.
         */
        while (m != NULL && len > 0) {
            if (m->m_hdr.mh_len <= len) {
                len -= m->m_hdr.mh_len;
                m->m_hdr.mh_len = 0;
                m = m->m_hdr.mh_next;
            } else {
                m->m_hdr.mh_len -= len;
                m->m_hdr.mh_data += len;
                len = 0;
            }
        }
        m = mp;
        if (mp->m_hdr.mh_flags & M_PKTHDR)
            m->M_dat.MH.MH_pkthdr.len -= (req_len - len);
    } else {
        /*
         * Trim from tail.  Scan the mbuf chain,
         * calculating its length and finding the last mbuf.
         * If the adjustment only affects this mbuf, then just
         * adjust and return.  Otherwise, rescan and truncate
         * after the remaining size.
         */
        len = -len;
        count = 0;
        for (;;) {
            count += m->m_hdr.mh_len;
            if (m->m_hdr.mh_next == (struct mbuf *)0)
                break;
            m = m->m_hdr.mh_next;
        }
        if (m->m_hdr.mh_len >= len) {
            m->m_hdr.mh_len -= len;
            if (mp->m_hdr.mh_flags & M_PKTHDR)
                mp->M_dat.MH.MH_pkthdr.len -= len;
            return;
        }
        count -= len;
        if (count < 0)
            count = 0;
        /*
         * Correct length for chain is "count".
         * Find the mbuf with last data, adjust its length,
         * and toss data from remaining mbufs on chain.
         */
        m = mp;
        if (m->m_hdr.mh_flags & M_PKTHDR)
            m->M_dat.MH.MH_pkthdr.len = count;
        for (; m; m = m->m_hdr.mh_next) {
            if (m->m_hdr.mh_len >= count) {
                m->m_hdr.mh_len = count;
                break;
            }
            count -= m->m_hdr.mh_len;
        }
        while ((m = m->m_hdr.mh_next) != NULL)
            m->m_hdr.mh_len = 0;
    }
}

/*
 * Rearange an mbuf chain so that len bytes are contiguous
 * and in the data area of an mbuf (so that mtod and dtom
 * will work for a structure of size len).  Returns the resulting
 * mbuf chain on success, frees it and returns null on failure.
 * If there is room, it will add up to max_protohdr-len extra bytes to the
 * contiguous region in an attempt to avoid being called next time.
 */
INT MPFail;

struct mbuf *
m_pullup(struct mbuf *n, INT len)
{
    struct mbuf *m;
    INT count;
    INT space;

    /*
     * If first mbuf has no cluster, and has room for len bytes
     * without shifting current data, pullup into it,
     * otherwise allocate a new mbuf to prepend to the chain.
     */
    if ((n->m_hdr.mh_flags & M_EXT) == 0 &&
        (ADDRESS) n->m_hdr.mh_data + len < (ADDRESS) &n->M_dat.M_databuf[MLEN] && n->m_hdr.mh_next) {
        if (n->m_hdr.mh_len >= len)
            return n;
        m = n;
        n = n->m_hdr.mh_next;
        len -= m->m_hdr.mh_len;
    } else {
        if (len > MHLEN)
            goto bad;
        m = m_get(M_DONTWAIT, n->m_hdr.mh_type);
        if (m == 0)
            goto bad;
        m->m_hdr.mh_len = 0;
        if (n->m_hdr.mh_flags & M_PKTHDR) {
            M_COPY_PKTHDR(m, n);
            n->m_hdr.mh_flags &= ~M_PKTHDR;
        }
    }
    space = (ADDRESS) &m->M_dat.M_databuf[MLEN] - (ADDRESS) (m->m_hdr.mh_data + m->m_hdr.mh_len);
    do {
        count = min(min(max(len, max_protohdr), space), n->m_hdr.mh_len);
        bcopy(mtod(n, caddr_t), mtod(m, caddr_t) + m->m_hdr.mh_len, (unsigned)count);
        len -= count;
        m->m_hdr.mh_len += count;
        n->m_hdr.mh_len -= count;
        space -= count;
        if (n->m_hdr.mh_len)
            n->m_hdr.mh_data += count;
        else
            n = m_free(n);
    } while (len > 0 && n);
    if (len > 0) {
        (void) m_free(m);
        goto bad;
    }
    m->m_hdr.mh_next = n;
    return m;
bad:
    m_freem(n);
    MPFail++;
    return NULL;
}


/*
 * Partition an mbuf chain in two pieces, returning the tail --
 * all but the first len0 bytes.  In case of failure, it returns NULL and
 * attempts to restore the chain to its original state.
 */
struct mbuf *
m_split(struct mbuf *m0, INT len0, INT wait)
{
    struct mbuf *m, *n;
    UINT len = len0, remain;

    for (m = m0; m && len > (UINT)m->m_hdr.mh_len; m = m->m_hdr.mh_next)
        len -= m->m_hdr.mh_len;
    if (m == 0)
        return NULL;
    remain = m->m_hdr.mh_len - len;
    if (m0->m_hdr.mh_flags & M_PKTHDR) {
        MGETHDR(n, wait, m0->m_hdr.mh_type);
        if (n == 0)
            return NULL;
        n->M_dat.MH.MH_pkthdr.rcvif = m0->M_dat.MH.MH_pkthdr.rcvif;
        n->M_dat.MH.MH_pkthdr.len = m0->M_dat.MH.MH_pkthdr.len - len0;
        m0->M_dat.MH.MH_pkthdr.len = len0;
        if (m->m_hdr.mh_flags & M_EXT)
            goto extpacket;
        if (remain > MHLEN) {
            /* m can't be the lead packet */
            MH_ALIGN(n, 0);
            n->m_hdr.mh_next = m_split(m, len, wait);
            if (n->m_hdr.mh_next == 0) {
                (void) m_free(n);
                return NULL;
            } else
                return n;
        } else
            MH_ALIGN(n, remain);
    } else if (remain == 0) {
        n = m->m_hdr.mh_next;
        m->m_hdr.mh_next = 0;
        return n;
    } else {
        MGET(n, wait, m->m_hdr.mh_type);
        if (n == 0)
            return NULL;
        M_ALIGN(n, remain);
    }
extpacket:
    if (m->m_hdr.mh_flags & M_EXT) {
        n->m_hdr.mh_flags |= M_EXT;
        n->M_dat.MH.MH_dat.MH_ext = m->M_dat.MH.MH_dat.MH_ext;
        AtomicInc(&m->M_dat.MH.MH_dat.MH_ext.ext_buf->M_dat.MH.MH_dat.MH_ext.ext_count);
        //mclrefcnt[mtocl(m->M_dat.MH.MH_dat.MH_ext.ext_buf)]++;
        //m->M_dat.MH.MH_dat.MH_ext.ext_size = 0; /* For Accounting XXXXXX danger */
        n->m_hdr.mh_data = m->m_hdr.mh_data + len;
    } else {
        bcopy(mtod(m, caddr_t) + len, mtod(n, caddr_t), remain);
    }
    n->m_hdr.mh_len = remain;
    m->m_hdr.mh_len = len;
    n->m_hdr.mh_next = m->m_hdr.mh_next;
    m->m_hdr.mh_next = 0;
    return n;
}

#if 0
/*
 * Routine to copy from device local memory into mbufs.
 */
struct mbuf *
m_devget(char *buf, int totlen, struct ifnet *ifp,
         void (*mymemcpy)(void *to, const void *from, size_t count))
{
    struct mbuf *m;
    struct mbuf *top = 0, **mp = &top;
    int len;
    char *cp;
    char *epkt;

    cp = buf;
    epkt = cp + totlen;

    MGETHDR(m, M_DONTWAIT, MT_DATA);
    if (m == 0)
        return (0);
    m->M_dat.MH.MH_pkthdr.rcvif = ifp;
    m->M_dat.MH.MH_pkthdr.len = totlen;
    m->m_hdr.mh_len = MHLEN;

    while (totlen > 0) {
        if (top) {
            MGET(m, M_DONTWAIT, MT_DATA);
            if (m == 0) {
                m_freem(top);
                return (0);
            }
            m->m_hdr.mh_len = MLEN;
        }
        len = min(totlen, epkt - cp);
        if (len >= MINCLSIZE) {
            MCLGET(m, M_DONTWAIT);
            if (m->m_hdr.mh_flags & M_EXT)
                m->m_hdr.mh_len = len = min(len, MCLBYTES);
            else
                len = m->m_hdr.mh_len;
        } else {
            /*
             * Place initial small packet/header at end of mbuf.
             */
            if (len < m->m_hdr.mh_len) {
                if (top == 0 && len + max_linkhdr <= m->m_hdr.mh_len)
                    m->m_hdr.mh_data += max_linkhdr;
                m->m_hdr.mh_len = len;
            } else
                len = m->m_hdr.mh_len;
        }
        if (mymemcpy)
            mymemcpy(mtod(m, caddr_t), cp, (size_t) len);
        else
            memcpy(mtod(m, caddr_t), cp, (size_t) len);
        cp += len;
        *mp = m;
        mp = &m->m_hdr.mh_next;
        totlen -= len;
        if (cp == epkt)
            cp = buf;
    }
    return (top);
}
#endif

/*
 * Module initialization
 */
void mbinit(void)
{
    /* nothing yet */
}
