/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1988, 1990, 1993
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)uipc_socket.c    8.1 (Berkeley) 6/10/93
 */

#include <sock.h>
#include <domain.h>
#include <if.h>
#include <route.h>
#include <assert.h>

extern CONDITION WaitHere;

void sorflush(struct socket *so);
INT sodisconnect(struct socket *so);
INT sockargs(struct mbuf **mp, caddr_t buf, UINT buflen, INT type);

/*
 * Free a no longer needed socket structure.
 * But first, make sure nobody is still using it and that
 * it isn't holding onto anything.
 */
void sofree(struct socket *so)
{
    /*
     * If either a protocol or a user still has a
     * reference to us, then don't go away just yet.
     */
    if (so->so_pcb || (so->so_state & SS_NOFDREF) == 0)
        return;

    /*
     * If we're on one of another socket's accept queues,
     * (i.e. the queues of incoming incomplete or ready connections)
     * then remove ourselves from that queue.
     */
    if (so->so_head) {
        if (!soqremque(so, 0) && !soqremque(so, 1))
            panic(_T("sofree dq"));  /* we weren't on a queue ?!? */
    }

    /*
     * Free any buffers still held by the socket (throw away the data)
     * and wake up any waiting readers.
     */
    sbrelease(&so->so_snd);
    sorflush(so);

    /*
     * Free the actual structure.
     */
    FREE(so, M_SOCKET);
}

/*
 * Close a socket on user request.
 * Initiate disconnect if presently connected.
 *
 * Note: We don't free the actual socket structure here,
 * that must wait until last reference has been dropped.
 */
INT soclose(struct socket *so)
{
    INT error = 0;

    NetLock();  /* conservative */
    /*
     * If we were listening for incoming connections,
     * we must clean up any partial or unaccepted connects.
     */
    if (so->so_options & SO_ACCEPTCONN) {
        while (so->so_q0)
            (void) soabort(so->so_q0);
        while (so->so_q)
            (void) soabort(so->so_q);
    }

    /*
     * If we no longer have a protocol control block,
     * then we have already detached from the protocol.
     */
    if (so->so_pcb == 0)
        goto leave;  /* that's all we need to do */

    /*
     * If we're currently connected, we must first disconnect with our peer.
     */
    if (so->so_state & SS_ISCONNECTED) {
        /*
         * If we haven't already started disconnecting, do so.
         */
        if ((so->so_state & SS_ISDISCONNECTING) == 0) {
            error = sodisconnect(so);
            if (error)
                goto detach;
        }
        /*
         * If the user wants us to wait for the disconnect to complete,
         * (and we're not non-blocking) then wait for it to do so (or timeout).
         */
        if (so->so_options & SO_LINGER) {
            if ((so->so_state & SS_ISDISCONNECTING) &&
                    (so->so_state & SS_NBIO))
                goto detach;
            while (so->so_state & SS_ISCONNECTED) {
                TIME linger;
                Int32ToInt64(linger,so->so_linger);
                error = XXXWaitFor((PTR)&so->so_timeo, netcls,
                                   Int64TimesInt32(linger,TIME_MILLIS(1)));
                if (error)
                    break;
            }
        }
    }

    /*
     * If we're still attached to a protocol, let it cut itself loose.
     */
detach:
    if (so->so_pcb) {
        INT error2 = (*so->so_proto->pr_usrreq)(so, PRU_DETACH,
                      (struct mbuf *)0, (struct mbuf *)0, (struct mbuf *)0);
        if (error == 0)
            error = error2;
    }

leave:
    /* If some other thread is blocked wake it */
    so->so_error = ECONNABORTED;
    sorwakeup(so);
    sowwakeup(so);

    NetUnlock();
    return (error);
}

/*
 * Must be called at splnet...
 */
INT soabort(struct socket *so)
{
    return ((*so->so_proto->pr_usrreq)(so, PRU_ABORT, (struct mbuf *)0,
             (struct mbuf *)0, (struct mbuf *)0));
}

INT sodisconnect(struct socket *so)
{
    INT error;

    NetLock();
    if ((so->so_state & SS_ISCONNECTED) == 0) {
        error = ENOTCONN;
        goto bad;
    }
    if (so->so_state & SS_ISDISCONNECTING) {
        error = EALREADY;
        goto bad;
    }
    error = (*so->so_proto->pr_usrreq)(so, PRU_DISCONNECT, (struct mbuf *)0,
                                       (struct mbuf *)0, (struct mbuf *)0);
bad:
    NetUnlock();
    return (error);
}

INT sendit2(struct socket *so, struct msghdr *mp, struct mbuf *iBuffer, UINT flags)
{
    struct mbuf *to;
    INT len = 0, error;
    
    if (mp->msg_name) {
        error = sockargs(&to, (caddr_t) mp->msg_name, mp->msg_namelen, MT_SONAME);
        if (error)
            goto bad;
    } else
        to = NULL;

    len = mp->req.uio_resid;
    error = sosend2(so, to, &mp->req, iBuffer, flags);
    if (error) {
        if (mp->req.uio_resid != len &&
                (error == EINTR || error == EWOULDBLOCK))
            error = 0;
#if notyet
        if (error == ESHUTDOWN)
            psignal(p, SIGPIPE);
#endif
    }

    if (to)
        m_freem(to);
    bad:
    if (error) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(error));
        return -1;
    }
    return (len - mp->req.uio_resid);
}

INT recvit2(struct socket *so, struct msghdr *mp, UINT32 *namelenp,
            struct mbuf **mp0)
{
    INT len, error, retval = 0;
    struct mbuf *from = 0;
    
    len = mp->req.uio_resid;
    error = soreceive(so, &from, &mp->req, mp0,
              (struct mbuf **)0, &mp->msg_flags);
    if (error) {
        if (mp->req.uio_resid != len &&
                (error == EINTR || error == EWOULDBLOCK))
            error = 0;
    }
    if (error)
        goto out;
    retval = len - mp->req.uio_resid;
    if (mp->msg_name) {
        len = mp->msg_namelen;
        if (len <= 0 || from == 0)
            len = 0;
        else {
#ifdef COMPAT_43
            if (mp->msg_flags & MSG_COMPAT)
                mtod(from, struct osockaddr *)->sa_family =
                    mtod(from, struct sockaddr *)->sa_family;
#endif
            if (len > from->m_len)
                len = from->m_len;
            memcpy(mp->msg_name, mtod(from, caddr_t), len);
        }
        mp->msg_namelen = len;
        if (namelenp)
            *namelenp = len;
    }
out:
    if (from)
        m_freem(from);
    if (error) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(error));
        return -1;
    }
    return (retval);
}


/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#define    SBLOCKWAIT(f)    (((f) & MSG_DONTWAIT) ? M_NOWAIT : M_WAITOK)
/*
 * Send on a socket.
 * If send must go all at once and message is larger than
 * send buffering, then hard error.
 * Lock against other senders.
 * If must go all at once and not enough room now, then
 * inform user that this would block and do nothing.
 * Otherwise, if nonblocking, send as much as possible.
 * The data to be sent is described by "uio".
 *
 * Returns nonzero on error, timeout or signal; callers
 * must check for short counts if EINTR is returned.
 * Data and control buffers are freed on return.
 */
INT sosend2(struct socket *so, struct mbuf *addr, struct uio *uio,
            struct mbuf *send_m, UINT flags)
{
    struct mbuf **mp;
    struct mbuf *m, *top = send_m;
    long space, len, resid;
    INT error, dontroute, mlen;
    INT atomic = sosendallatonce(so);

    resid = uio->uio_resid;

    dontroute =
        (flags & MSG_DONTROUTE) && (so->so_options & SO_DONTROUTE) == 0 &&
        (so->so_proto->pr_flags & PR_ATOMIC);
#if 0
    p->p_stats->p_ru.ru_msgsnd++;
#endif

#define snderr(errno) { error = errno; NetUnlock(); goto release; }

restart:
    error = sblock(&so->so_snd, SBLOCKWAIT(flags));
    if (error) goto out;
    do {
        NetLock();
        if (so->so_state & SS_CANTSENDMORE)
            snderr(ESHUTDOWN);
        if (so->so_error)
            snderr(so->so_error);
        if ((so->so_state & SS_ISCONNECTED) == 0) {
            if (so->so_proto->pr_flags & PR_CONNREQUIRED) {
                if ((so->so_state & SS_ISCONFIRMING) == 0)
                    snderr(ENOTCONN);
            } else if (addr == 0)
                snderr(EDESTADDRREQ);
        }
        space = sbspace(&so->so_snd);
        if (flags & MSG_OOB)
            space += 1024;
        if (atomic && resid > (long)so->so_snd.sb_hiwat)
            snderr(EMSGSIZE);
        if (space < resid &&
                (atomic || space < (long)so->so_snd.sb_lowat || space < 0)) {
            if (so->so_state & SS_NBIO)
                snderr(EWOULDBLOCK);
            sbunlock(&so->so_snd);
            error = sbwait(&so->so_snd);
            NetUnlock();
            if (error)
                goto out;
            goto restart;
        }
        NetUnlock();
        mp = &top;
        do {
            if (send_m) {
                uio->uio_resid = resid = 0; /* XXX all in one shot */
            } else do {
                if (top == NULL) {
                    MGETHDR(m, M_WAIT, MT_DATA);
                    mlen = MHLEN;
                    m->m_pkthdr.len = 0;
                    m->m_pkthdr.rcvif = (struct ifnet *)0;
                } else {
                    MGET(m, M_WAIT, MT_DATA);
                    mlen = MLEN;
                }
                if (resid >= MINCLSIZE && space >= MCLBYTES) {
                    MCLGET(m, M_WAIT);
                    if ((m->m_flags & M_EXT) == 0)
                        goto nopages;
                    mlen = MCLBYTES;
#ifdef    MAPPED_MBUFS
                    len = min(MCLBYTES, resid);
#else
                    if (atomic && top == 0) {
                        len = min(MCLBYTES - max_hdr, resid);
                        m->m_data += max_hdr;
                    } else
                        len = min(MCLBYTES, resid);
#endif
                    space -= len;  /* was MCLBYTES */
                } else {
nopages:
                    len = min(min(mlen, resid), space);
                    space -= len;
                    /*
                     * For datagram protocols, leave room
                     * for protocol headers in first mbuf.
                     */
                    if (atomic && top == 0 && len < mlen)
                        MH_ALIGN(m, len);
                }
                uiomovein(mtod(m, caddr_t), (INT)len, uio);
                resid = uio->uio_resid;
                m->m_len = len;
                *mp = m;
                top->m_pkthdr.len += len;
                if (error)
                    goto release;
                mp = &m->m_next;
                if (resid <= 0) {
                    if (flags & MSG_EOR)
                        top->m_flags |= M_EOR;
                    break;
                }
            } while (space > 0 && atomic);
            if (dontroute)
                so->so_options |= SO_DONTROUTE;
            NetLock();  /* XXX */
            error = (*so->so_proto->pr_usrreq)(so, (flags & MSG_OOB) ?
                     PRU_SENDOOB : PRU_SEND, top, addr, NULL);
            NetUnlock();
            if (send_m) assert(resid == 0);
            if (dontroute)
                so->so_options &= ~SO_DONTROUTE;
            top = 0;
            mp = &top;
            if (error)
                goto release;
        } while (resid && space > 0);
    } while (resid);

release:
    sbunlock(&so->so_snd);
out:
    return (error);
}

/*
 * Implement receive operations on a socket.
 * We depend on the way that records are added to the sockbuf
 * by sbappend*.  In particular, each record (mbufs linked through m_next)
 * must begin with an address if the protocol so specifies,
 * followed by an optional mbuf or mbufs containing ancillary data,
 * and then zero or more mbufs of data.
 * In order to avoid blocking network interrupts for the entire time here,
 * we splx() while doing the actual copy to user space.
 * Although the sockbuf is locked, new data may still be appended,
 * and thus we must maintain consistency of the sockbuf during that time.
 *
 * The caller may receive the data as a single mbuf chain by supplying
 * an mbuf **mp0 for use in returning the chain.  The uio is then used
 * only for the count in uio_resid.
 */
INT soreceive(struct socket *so, struct mbuf **paddr, struct uio *uio,
              struct mbuf **mp0, struct mbuf **controlp, UINT32 *flagsp)
{
    struct mbuf *m, **mp;
    UINT flags;
    INT len, error, offset;
    const struct protosw *pr = so->so_proto;
    struct mbuf *nextrecord;
    INT moff, type = -1;
    INT orig_resid = uio->uio_resid;

    mp = mp0;
    if (paddr)
        *paddr = 0;
    if (controlp)
        *controlp = 0;
    if (flagsp)
        flags = *flagsp &~ MSG_EOR;
    else
        flags = 0;
    if (flags & MSG_OOB) {
        m = m_get(M_WAIT, MT_DATA);
        error = (*pr->pr_usrreq)(so, PRU_RCVOOB,
                 m, (struct mbuf *)(flags & MSG_PEEK), (struct mbuf *)0);
        if (error)
            goto bad;
        do {
            uiomoveout(mtod(m, caddr_t),
                       (INT) min(uio->uio_resid, m->m_len), uio);
            m = m_free(m);
        } while (uio->uio_resid && error == 0 && m);
bad:
        if (m)
            m_freem(m);
        return (error);
    }
    if (mp)
        *mp = (struct mbuf *)0;
    if (so->so_state & SS_ISCONFIRMING && uio->uio_resid)
        (*pr->pr_usrreq)(so, PRU_RCVD, (struct mbuf *)0,
            (struct mbuf *)0, (struct mbuf *)0);

restart:
    error = sblock(&so->so_rcv, SBLOCKWAIT(flags));
    if (error)
        return (error);
    NetLock();

    m = so->so_rcv.sb_mb;
    /*
     * If we have less data than requested, block awaiting more
     * (subject to any timeout) if:
     *   1. the current count is less than the low water mark, or
     *   2. MSG_WAITALL is set, and it is possible to do the entire
     *    receive operation at once if we block (resid <= hiwat).
     *   3. MSG_DONTWAIT is not set
     * If MSG_WAITALL is set but resid is larger than the receive buffer,
     * we have to do the receive in sections, and thus risk returning
     * a short count if a timeout or signal occurs after we start.
     */
    if (m == 0 || (((flags & MSG_DONTWAIT) == 0 &&
        (long)so->so_rcv.sb_cc < uio->uio_resid) &&
        (so->so_rcv.sb_cc < so->so_rcv.sb_lowat ||
        ((flags & MSG_WAITALL) && uio->uio_resid <= (long)so->so_rcv.sb_hiwat)) &&
        m->m_nextpkt == 0 && (pr->pr_flags & PR_ATOMIC) == 0)) {
#ifdef DIAGNOSTIC
        if (m == 0 && so->so_rcv.sb_cc)
            panic(_T("receive 1"));
#endif
        if (so->so_error) {
            if (m)
                goto dontblock;
            error = so->so_error;
            if ((flags & MSG_PEEK) == 0)
                so->so_error = 0;
            goto release;
        }
        if (so->so_state & SS_CANTRCVMORE) {
            if (m)
                goto dontblock;
            else
                goto release;
        }
        for (; m; m = m->m_next)
            if (m->m_type == MT_OOBDATA  || (m->m_flags & M_EOR)) {
                m = so->so_rcv.sb_mb;
                goto dontblock;
            }
        if ((so->so_state & (SS_ISCONNECTED|SS_ISCONNECTING)) == 0 &&
            (so->so_proto->pr_flags & PR_CONNREQUIRED)) {
            error = ENOTCONN;
            goto release;
        }
        if (uio->uio_resid == 0)
            goto release;
        if ((so->so_state & SS_NBIO) || (flags & MSG_DONTWAIT)) {
            error = EWOULDBLOCK;
            goto release;
        }
        sbunlock(&so->so_rcv);
        error = sbwait(&so->so_rcv);
        NetUnlock();
        if (error)
            return (error);
        goto restart;
    }
dontblock:
#if 0
    if (uio->uio_procp)
        uio->uio_procp->p_stats->p_ru.ru_msgrcv++;
#endif
    nextrecord = m->m_nextpkt;
    if (pr->pr_flags & PR_ADDR) {
#ifdef DIAGNOSTIC
        if (m->m_type != MT_SONAME)
            panic(_T("receive 1a"));
#endif
        orig_resid = 0;
        if (flags & MSG_PEEK) {
            if (paddr)
                *paddr = m_copy(m, 0, m->m_len);
            m = m->m_next;
        } else {
            sbfree(&so->so_rcv, m);
            if (paddr) {
                *paddr = m;
                so->so_rcv.sb_mb = m->m_next;
                m->m_next = 0;
                m = so->so_rcv.sb_mb;
            } else {
                MFREE(m, so->so_rcv.sb_mb);
                m = so->so_rcv.sb_mb;
            }
        }
    }
    while (m && m->m_type == MT_CONTROL && error == 0) {
        if (flags & MSG_PEEK) {
            if (controlp)
                *controlp = m_copy(m, 0, m->m_len);
            m = m->m_next;
        } else {
            sbfree(&so->so_rcv, m);
            if (controlp) {
                if (pr->pr_domain->dom_externalize &&
                        mtod(m, struct cmsghdr *)->cmsg_type == SCM_RIGHTS)
                    error = (*pr->pr_domain->dom_externalize)(m);
                *controlp = m;
                so->so_rcv.sb_mb = m->m_next;
                m->m_next = 0;
                m = so->so_rcv.sb_mb;
            } else {
                MFREE(m, so->so_rcv.sb_mb);
                m = so->so_rcv.sb_mb;
            }
        }
        if (controlp) {
            orig_resid = 0;
            controlp = &(*controlp)->m_next;
        }
    }
    if (m) {
        if ((flags & MSG_PEEK) == 0)
            m->m_nextpkt = nextrecord;
        type = m->m_type;
        if (type == MT_OOBDATA)
            flags |= MSG_OOB;
    }
    moff = 0;
    offset = 0;
    while (m && uio->uio_resid > 0 && error == 0) {
        if (m->m_type == MT_OOBDATA) {
            if (type != MT_OOBDATA)
                break;
        } else if (type == MT_OOBDATA)
            break;
#ifdef DIAGNOSTIC
        else if (m->m_type != MT_DATA && m->m_type != MT_HEADER)
            panic(_T("receive 3"));
#endif
        so->so_state &= ~SS_RCVATMARK;
        len = uio->uio_resid;
        if (so->so_oobmark && len > (INT)so->so_oobmark - offset)
            len = so->so_oobmark - offset;
        if (len > m->m_len - moff)
            len = m->m_len - moff;
        /*
         * If mp is set, just pass back the mbufs.
         * Otherwise copy them out via the uio, then free.
         * Sockbuf must be consistent here (points to current mbuf,
         * it points to next record) when we drop priority;
         * we must note any additions to the sockbuf when we
         * block interrupts again.
         */
        if (mp == 0) {
            NetUnlock();
            uiomoveout(mtod(m, caddr_t) + moff, (INT)len, uio);
            NetLock();
        } else
            uio->uio_resid -= len;
        if (len == m->m_len - moff) {
            if (m->m_flags & M_EOR)
                flags |= MSG_EOR;
            if (flags & MSG_PEEK) {
                m = m->m_next;
                moff = 0;
            } else {
                nextrecord = m->m_nextpkt;
                sbfree(&so->so_rcv, m);
                if (mp) {
                    *mp = m;
                    mp = &m->m_next;
                    so->so_rcv.sb_mb = m = m->m_next;
                    *mp = (struct mbuf *)0;
                } else {
                    MFREE(m, so->so_rcv.sb_mb);
                    m = so->so_rcv.sb_mb;
                }
                if (m)
                    m->m_nextpkt = nextrecord;
            }
        } else {
            if (flags & MSG_PEEK)
                moff += len;
            else {
                if (mp)
                    *mp = m_copym(m, 0, len, M_WAIT);
                m->m_data += len;
                m->m_len -= len;
                so->so_rcv.sb_cc -= len;
            }
        }
        if (so->so_oobmark) {
            if ((flags & MSG_PEEK) == 0) {
                so->so_oobmark -= len;
                if (so->so_oobmark == 0) {
                    so->so_state |= SS_RCVATMARK;
                    break;
                }
            } else {
                offset += len;
                if (offset == (INT)so->so_oobmark)
                    break;
            }
        }
        if (flags & MSG_EOR)
            break;
        /*
         * If the MSG_WAITALL flag is set (for non-atomic socket),
         * we must not quit until "uio->uio_resid == 0" or an error
         * termination.  If a signal/timeout occurs, return
         * with a short count but without error.
         * Keep sockbuf locked against other readers.
         */
        while (flags & MSG_WAITALL && m == 0 && uio->uio_resid > 0 &&
                !sosendallatonce(so) && !nextrecord) {
            if (so->so_error || so->so_state & SS_CANTRCVMORE)
                break;
            error = sbwait(&so->so_rcv);
            if (error) {
                sbunlock(&so->so_rcv);
                NetUnlock();
                return (0);
            }
            if ((m = so->so_rcv.sb_mb) != NULL)
                nextrecord = m->m_nextpkt;
        }
    }

    if (m && pr->pr_flags & PR_ATOMIC) {
        flags |= MSG_TRUNC;
        if ((flags & MSG_PEEK) == 0)
            (void) sbdroprecord(&so->so_rcv);
    }
    if ((flags & MSG_PEEK) == 0) {
        if (m == 0)
            so->so_rcv.sb_mb = nextrecord;
        if (pr->pr_flags & PR_WANTRCVD && so->so_pcb)
            (*pr->pr_usrreq)(so, PRU_RCVD, (struct mbuf *)0,
                (struct mbuf *)flags, (struct mbuf *)0);
    }
    if (orig_resid == uio->uio_resid && orig_resid &&
            (flags & MSG_EOR) == 0 && (so->so_state & SS_CANTRCVMORE) == 0) {
        sbunlock(&so->so_rcv);
        NetUnlock();
        goto restart;
    }
        
    if (flagsp)
        *flagsp |= flags;
release:
    sbunlock(&so->so_rcv);
    NetUnlock();
    return (error);
}

void sorflush(struct socket *so)
{
    struct sockbuf *sb = &so->so_rcv;
    const struct protosw *pr = so->so_proto;
    struct sockbuf asb;

    sb->sb_flags |= SB_NOINTR;
    (void) sblock(sb, M_WAITOK);
    NetLock();
    socantrcvmore(so);
    sbunlock(sb);
    asb = *sb;
    bzero((caddr_t)sb, sizeof (*sb));
    NetUnlock();
    if (pr->pr_flags & PR_RIGHTS && pr->pr_domain->dom_dispose)
        (*pr->pr_domain->dom_dispose)(asb.sb_mb);
    sbrelease(&asb);
}


void sohasoutofband(struct socket *so)
{
#if notyet
    struct proc *p;

    if (so->so_pgid < 0)
        gsignal(-so->so_pgid, SIGURG);
    else if (so->so_pgid > 0 && (p = pfind(so->so_pgid)) != 0)
#endif
        psignal(p, SIGURG);

/* was:    selwakeup(&so->so_rcv.sb_sel); */
    if (so->so_rcv.sb_flags & SB_SEL) {
        so->so_rcv.sb_flags &= ~SB_SEL;
        Condition_Signal(&WaitHere);
    }
}

/*
 * Copy argument(s) into an mbuf
 */
INT sockargs(struct mbuf **mp, caddr_t buf, UINT buflen, INT type)
{
    struct sockaddr *sa;
    struct mbuf *m;

    if (buflen > MLEN)
        return (EINVAL);

    m = m_get(M_WAIT, type);
    if (m == NULL)
        return (ENOBUFS);
    m->m_len = buflen;
    memcpy(mtod(m, caddr_t), buf, buflen);
    *mp = m;
    if (type == MT_SONAME) {
        sa = mtod(m, struct sockaddr *);

#if defined(COMPAT_43) && BYTE_ORDER != BIG_ENDIAN
        if (sa->sa_family == 0 && sa->sa_len < AF_MAX)
            sa->sa_family = sa->sa_len;
#endif
        sa->sa_len = (UINT8)buflen;
    }
    return (0);
}
