/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1988, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      @(#)tcp_usrreq.c        8.1 (Berkeley) 6/10/93
 */

#include <sock.h>
#include <netproto.h>

#ifdef __WANTS_TCP

#if 0
#include <sys/stat.h>
#endif

#include <if.h>
#include <route.h>

#include <in.h>
#include <in_systm.h>
#include <ip.h>
#include <in_pcb.h>
#include <in_var.h>
#include <ip_var.h>
#include <tcp.h>
#include <tcp_fsm.h>
#include <tcp_seq.h>
#include <tcp_tim.h>
#include <tcp_var.h>
#include <tcpip.h>
#include <tcp_dbg.h>

INT in_pcbbindNEW(struct inpcb *inp, struct sockaddr_in *Name);
INT in_pcbconnectNEW(struct inpcb *inp, struct sockaddr_in *Name);
void in_setpeeraddrNEW(struct inpcb *inp, struct sockaddr_in *Name);
void in_setsockaddrNEW(struct inpcb *inp, struct sockaddr_in *Name);

/*
 * TCP protocol interface to socket abstraction.
 */
extern  char *tcpstates[];

/*
 * Process a TCP user request for TCP tb.  If this is a send request
 * then m is the mbuf chain of send data.  If this is a timer expiration
 * (called from the software clock routine), then timertype tells which timer.
 */
PRIVATE SCODE TcpControl(PINETPROTOCOL This, struct socket *so, INT Cmd, PTR Data, PTR ifp)
{
    return in_control(so, Cmd, Data, ifp);
}

PRIVATE SCODE TcpAttach(PINETPROTOCOL This, struct socket *so)
{
    struct inpcb *inp;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    /*
     * When a TCP is attached to a socket, then there will be
     * a (struct inpcb) pointed at by the socket, and this
     * structure will point at a subsidary (struct tcpcb).
     */

    /*
     * TCP attaches to socket via PRU_ATTACH, reserving space,
     * and an internet control block.
     */
    if (inp) {
        error = EISCONN;
        goto Out;
    }
    error = tcp_attach(so);
    if (error)
        goto Out;
    if ((so->so_options & SO_LINGER) && so->so_linger == 0)
        so->so_linger = TCP_LINGERTIME;

Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpDetach(PINETPROTOCOL This, struct socket *so)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    /*
     * PRU_DETACH detaches the TCP protocol from the socket.
     * If the protocol state is non-embryonic, then can't
     * do this directly: have to initiate a PRU_DISCONNECT,
     * which may finish later; embryonic TCB's can just
     * be discarded here.
     */
    if (tp->t_state > TCPS_LISTEN)
        tp = tcp_disconnect(tp);
    else
        tp = tcp_close(tp);


Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpBind(PINETPROTOCOL This, struct socket *so, struct sockaddr_in *Name)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    /*
     * Give the socket an address.
     */
    error = in_pcbbindNEW(inp, Name);


Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpListen(PINETPROTOCOL This, struct socket *so)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    /*
     * Prepare to accept connections.
     */
    if (inp->inp_lport == 0)
        error = in_pcbbind(inp, (struct mbuf *)0);
    if (error == 0)
        tp->t_state = TCPS_LISTEN;



Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpConnect(PINETPROTOCOL This, struct socket *so, struct sockaddr_in *Name)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    /*
     * Initiate connection to peer.
     * Create a template for use in transmissions on this connection.
     * Enter SYN_SENT state, and mark socket as connecting.
     * Start keep-alive timer, and seed output sequence space.
     * Send initial segment on connection.
     */
    if (inp->inp_lport == 0) {
        error = in_pcbbindNEW(inp, NULL);
        if (error)
            goto Out;
    }
    error = in_pcbconnectNEW(inp, Name);
    if (error)
        goto Out;
    tp->t_template = tcp_template(tp);
    if (tp->t_template == 0) {
        in_pcbdisconnect(inp);
        error = ENOBUFS;
        goto Out;
    }
    /* Compute window scaling to request.  */
    while (tp->request_r_scale < TCP_MAX_WINSHIFT &&
           (u_long)(TCP_MAXWIN << tp->request_r_scale) < so->so_rcv.sb_hiwat)
        tp->request_r_scale++;
    soisconnecting(so);
    tcpstat.tcps_connattempt++;
    tp->t_state = TCPS_SYN_SENT;
    tp->t_timer[TCPT_KEEP] = TCPTV_KEEP_INIT;
    tp->iss = tcp_iss; tcp_iss += TCP_ISSINCR/2;
    tcp_sendseqinit(tp);
    error = tcp_output(tp);



Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpAccept(PINETPROTOCOL This, struct socket *so, struct sockaddr_in *Name)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    /*
     * Accept a connection.  Essentially all the work is
     * done at higher levels; just return the address
     * of the peer, storing through addr.
     */
    in_setpeeraddrNEW(inp, Name);



Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpDisconnect(PINETPROTOCOL This, struct socket *so)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    /*
     * Initiate disconnect from peer.
     * If connection never passed embryonic stage, just drop;
     * else if don't need to let data drain, then can just drop anyways,
     * else have to begin TCP shutdown process: mark socket disconnecting,
     * drain unread data, state switch to reflect user close, and
     * send segment (e.g. FIN) to peer.  Socket will be really disconnected
     * when peer sends FIN and acks ours.
     *
     * SHOULD IMPLEMENT LATER PRU_CONNECT VIA REALLOC TCPCB.
     */
    tp = tcp_disconnect(tp);



Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpShutdown(PINETPROTOCOL This, struct socket *so)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    /*
     * Mark the connection as being incapable of further output.
     */
    socantsendmore(so);
    tp = tcp_usrclosed(tp);
    if (tp)
        error = tcp_output(tp);



Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpRcvd(PINETPROTOCOL This, struct socket *so)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    /*
     * After a receive, possibly send window update to peer.
     */
    (void) tcp_output(tp);



Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpSend(PINETPROTOCOL This, struct socket *so, void *Buf, UINT Len)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


#if 1
    error = This->v->Output(This, tp, Buf, Len);
#else
    /*
     * Do a send by putting data in output queue and updating urgent
     * marker if URG set.  Possibly send more data.
     */
    sbappend(&so->so_snd, m);
    error = tcp_output(tp);
#endif



Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpAbort(PINETPROTOCOL This, struct socket *so)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    /*
     * Abort the TCP.
     */
    tp = tcp_drop(tp, ECONNABORTED);



Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpSense(PINETPROTOCOL This, struct socket *so)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


#if notyet
    ((struct stat *) m)->st_blksize = so->so_snd.sb_hiwat;
#else
    error = EOPNOTSUPP;
#endif


Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpRcvOob(PINETPROTOCOL This, struct socket *so, void *Buf, UINT Flags)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    if ((so->so_oobmark == 0 &&
         (so->so_state & SS_RCVATMARK) == 0) ||
        so->so_options & SO_OOBINLINE ||
        tp->t_oobflags & TCPOOB_HADDATA) {
        error = EINVAL;
        goto Out;
    }
    if ((tp->t_oobflags & TCPOOB_HAVEDATA) == 0) {
        error = EWOULDBLOCK;
        goto Out;
    }
    *(caddr_t)Buf = tp->t_iobc;
    if ((Flags & MSG_PEEK) == 0)
        tp->t_oobflags ^= (TCPOOB_HAVEDATA | TCPOOB_HADDATA);


Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpSendOob(PINETPROTOCOL This, struct socket *so, void *Buf, UINT Len)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    if (sbspace(&so->so_snd) < -512) {
        error = ENOBUFS;
        goto Out;
    }
    /*
     * According to RFC961 (Assigned Protocols),
     * the urgent pointer points to the last octet
     * of urgent data.  We continue, however,
     * to consider it to indicate the first octet
     * of data past the urgent section.
     * Otherwise, snd_up should be one lower.
     */
    tp->t_force = 1;
    tp->snd_up = tp->snd_una + so->so_snd.sb_cc;
    error = This->v->Output(This, tp, Buf, Len);
    tp->t_force = 0;


Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpSockAddr(PINETPROTOCOL This, struct socket *so, struct sockaddr_in *Name)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    in_setsockaddrNEW(inp, Name);



Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpPeerAddr(PINETPROTOCOL This, struct socket *so, struct sockaddr_in *Name)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    in_setpeeraddrNEW(inp, Name);



Out:
    NetUnlock();
    return (error);
}

PRIVATE SCODE TcpPruSlowTimo(PINETPROTOCOL This, struct socket *so, INT Timer)
{
    struct inpcb *inp;
    struct tcpcb *tp = NULL;
    int error = 0;

    NetLock();
    inp = sotoinpcb(so);
    if (!inp) {
        error = EINVAL;
        goto Out;
    }
    tp = intotcpcb(inp);


    tcp_timers(tp, Timer);



Out:
    NetUnlock();
    return (error);
}

/*ARGSUSED*/
INT
tcp_usrreq(struct socket *so, INT req,
           struct mbuf *m, struct mbuf *nam, struct mbuf *control)
{
    PINETPROTOCOL Proto = NULL; /* XXX */

    switch (req) {
    case PRU_CONTROL:
        return Proto->v->Control(Proto, so, (int) m, (caddr_t) nam,
                                 (struct ifnet *) control);
    case PRU_ATTACH:
        return Proto->v->Attach(Proto, so);
    case PRU_DETACH:
        return Proto->v->Detach(Proto, so);
    case PRU_BIND:
        return Proto->v->Bind(Proto, so, mtod(nam, struct sockaddr_in *));
    case PRU_LISTEN:
        return Proto->v->Listen(Proto, so);
    case PRU_CONNECT:
        return Proto->v->Connect(Proto, so, mtod(nam, struct sockaddr_in *));
    case PRU_CONNECT2:
        return EOPNOTSUPP;
    case PRU_DISCONNECT:
        return Proto->v->Disconnect(Proto, so);
    case PRU_ACCEPT:
        return Proto->v->Accept(Proto, so, mtod(nam, struct sockaddr_in *));
    case PRU_SHUTDOWN:
        return Proto->v->Shutdown(Proto, so);
    case PRU_RCVD:
        return Proto->v->Rcvd(Proto, so);
    case PRU_SEND:
        /* XXX assume all data is in immediate m */
        return Proto->v->Send(Proto, so, m->m_data, m->m_len);
    case PRU_ABORT:
        return Proto->v->Abort(Proto, so);
    case PRU_SENSE:
        return Proto->v->Sense(Proto, so);
    case PRU_RCVOOB:
        /* XXX assume all data is in immediate m */
        return Proto->v->RcvOob(Proto, so, m->m_data, (int)nam);
    case PRU_SENDOOB:
        /* XXX assume all data is in immediate m */
        return Proto->v->SendOob(Proto, so, m->m_data, m->m_len);
    case PRU_SOCKADDR:
        return Proto->v->SockAddr(Proto, so, mtod(nam, struct sockaddr_in *));
    case PRU_PEERADDR:
        return Proto->v->PeerAddr(Proto, so, mtod(nam, struct sockaddr_in *));
    case PRU_SLOWTIMO:
        return Proto->v->PruSlowtimo(Proto, so, (int)nam);
    default:
        panic(_T("tcp_usrreq"));
    }
    return EINVAL;
}

INT
tcp_ctloutput(INT op, struct socket *so, INT level, INT optname, struct mbuf **mp)
{
        int error = 0;
        struct inpcb *inp;
        struct tcpcb *tp;
        struct mbuf *m;
        int i;

        NetLock();
        inp = sotoinpcb(so);
        if (inp == NULL) {
                NetUnlock();
                return (ECONNRESET);
        }
        if (level != IPPROTO_TCP) {
                error = ip_ctloutput(op, so, level, optname, mp);
                NetUnlock();
                return (error);
        }
        tp = intotcpcb(inp);

        switch (op) {

        case PRCO_SETOPT:
                m = *mp;
                switch (optname) {

                case TCP_NODELAY:
                        if (m == NULL || m->m_len < sizeof (int))
                                error = EINVAL;
                        else if (*mtod(m, int *))
                                tp->t_flags |= TF_NODELAY;
                        else
                                tp->t_flags &= ~TF_NODELAY;
                        break;

                case TCP_MAXSEG:
                        if (m && (i = *mtod(m, int *)) > 0 && i <= tp->t_maxseg)
                                tp->t_maxseg = (u_short) i;
                        else
                                error = EINVAL;
                        break;

                default:
                        error = EINVAL;
                        break;
                }
                if (m)
                        (void) m_free(m);
                break;

        case PRCO_GETOPT:
                *mp = m = m_get(M_WAIT, MT_SOOPTS);
                m->m_len = sizeof(int);

                switch (optname) {
                case TCP_NODELAY:
                        *mtod(m, int *) = tp->t_flags & TF_NODELAY;
                        break;
                case TCP_MAXSEG:
                        *mtod(m, int *) = tp->t_maxseg;
                        break;
                default:
                        error = EINVAL;
                        break;
                }
                break;
        }
        NetUnlock();
        return (error);
}

u_long  tcp_sendspace = 1024*8;
u_long  tcp_recvspace = 1024*8;

/*
 * Attach TCP protocol to socket, allocating
 * internet protocol control block, tcp control block,
 * bufer space, and entering LISTEN state if to accept connections.
 */
INT
tcp_attach(struct socket *so)
{
        struct tcpcb *tp;
        struct inpcb *inp;
        int error;

        if (so->so_snd.sb_hiwat == 0 || so->so_rcv.sb_hiwat == 0) {
                error = soreserve(so, tcp_sendspace, tcp_recvspace);
                if (error)
                        return (error);
        }
        error = in_pcballoc(so, &tcb);
        if (error)
                return (error);
        inp = sotoinpcb(so);
        tp = tcp_newtcpcb(inp);
        if (tp == 0) {
                int nofd = so->so_state & SS_NOFDREF;   /* XXX */

                so->so_state &= ~SS_NOFDREF;    /* don't free the socket yet */
                in_pcbdetach(inp);
                so->so_state |= nofd;
                return (ENOBUFS);
        }
        tp->t_state = TCPS_CLOSED;
        return (0);
}

/*
 * Initiate (or continue) disconnect.
 * If embryonic state, just send reset (once).
 * If in ``let data drain'' option and linger null, just drop.
 * Otherwise (hard), mark socket disconnecting and drop
 * current input data; switch states based on user close, and
 * send segment to peer (with FIN).
 */
struct tcpcb *
tcp_disconnect(struct tcpcb *tp)
{
        struct socket *so = tp->t_inpcb->inp_socket;

        if (tp->t_state < TCPS_ESTABLISHED)
                tp = tcp_close(tp);
        else if ((so->so_options & SO_LINGER) && so->so_linger == 0)
                tp = tcp_drop(tp, 0);
        else {
                soisdisconnecting(so);
                sbflush(&so->so_rcv);
                tp = tcp_usrclosed(tp);
                if (tp)
                        (void) tcp_output(tp);
        }
        return (tp);
}

/*
 * User issued close, and wish to trail through shutdown states:
 * if never received SYN, just forget it.  If got a SYN from peer,
 * but haven't sent FIN, then go to FIN_WAIT_1 state to send peer a FIN.
 * If already got a FIN from peer, then almost done; go to LAST_ACK
 * state.  In all other cases, have already sent FIN to peer (e.g.
 * after PRU_SHUTDOWN), and just have to play tedious game waiting
 * for peer to send FIN or not respond to keep-alives, etc.
 * We can let the user exit from the close as soon as the FIN is acked.
 */
struct tcpcb *
tcp_usrclosed(struct tcpcb *tp)
{

        switch (tp->t_state) {

        case TCPS_CLOSED:
        case TCPS_LISTEN:
        case TCPS_SYN_SENT:
                tp->t_state = TCPS_CLOSED;
                tp = tcp_close(tp);
                break;

        case TCPS_SYN_RECEIVED:
        case TCPS_ESTABLISHED:
                tp->t_state = TCPS_FIN_WAIT_1;
                break;

        case TCPS_CLOSE_WAIT:
                tp->t_state = TCPS_LAST_ACK;
                break;
        }
        if (tp && tp->t_state >= TCPS_FIN_WAIT_2)
                soisdisconnected(tp->t_inpcb->inp_socket);
        return (tp);
}

struct NetProtocolVtbl TcpNetProtocolVtbl = {
    NULL,
    NULL,
    NULL,

/* protocol-protocol hooks */
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,

/* user-protocol hook (pr_usrreq) */
    TcpAttach,
    TcpDetach,
    TcpBind,
    TcpListen,
    TcpConnect,
    TcpAccept,
    TcpDisconnect,
    TcpShutdown,
    TcpRcvd,
    TcpSend,
    TcpAbort,
    TcpControl,
    TcpSense,
    TcpRcvOob,
    TcpSendOob,
    TcpSockAddr,
    TcpPeerAddr,
    TcpPruSlowTimo,

/* utility hooks */
    NULL,
    NULL,
    NULL,
    NULL,
    NULL
};

#endif /* __WANTS_TCP */
