/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Copyright (c) 1982, 1986, 1988, 1990, 1993
 *    Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by the University of
 *    California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *    @(#)udp_usrreq.c    8.1 (Berkeley) 6/10/93
 */

#include <sock.h>
#ifdef __WANTS_UDP

#include <stdio.h>

#include <if.h>
#include <route.h>

#include <in.h>
#include <in_systm.h>
#include <ip.h>
#include <in_pcb.h>
#include <in_var.h>
#include <ip_var.h>
#include <ip_icmp.h>
#include <udp.h>
#include <udp_var.h>
#include <dhcp/dhcp.h>
#include <dhcp/dhcp_var.h>

/*
 * UDP protocol implementation.
 * Per RFC 768, August, 1980.
 */
#ifndef COMPAT_42
#define UDPCKSUM 1
#else
#define UDPCKSUM 0        /* XXX */
#endif

struct sockaddr_in udp_in = { sizeof(udp_in), AF_INET };
struct inpcb *udp_last_inpcb = &udb;

static void udp_detach(struct inpcb *);
static void udp_notify(struct inpcb *, INT);
static struct mbuf *udp_saveopt(caddr_t, INT, INT);

void
udp_init(void)
{
    udb.inp_next = udb.inp_prev = &udb;
    DhcpInit();
}

void
udp_input(struct mbuf *m, INT iphlen)
{
    struct ip *ip;
    struct udphdr *uh;
    struct inpcb *inp;
    struct mbuf *opts = 0;
    INT len;
    struct ip save_ip;

    _COLLECT_STAT(udpstat.udps_ipackets++);

    /*
     * Strip IP options, if any; should skip this,
     * make available to user, and use on returned packets,
     * but we don't yet have a way to check the checksum
     * with options still present.
     */
    if (iphlen > sizeof(struct ip)) {
        ip_stripoptions(m, (struct mbuf *)0);
        iphlen = sizeof(struct ip);
    }

    /*
     * Get IP and UDP header together in first mbuf.
     */
    ip = mtod(m, struct ip *);
    if (m->m_len < (INT)(iphlen + sizeof(struct udphdr))) {
        if ((m = m_pullup(m, iphlen + sizeof(struct udphdr))) == 0) {
            _COLLECT_STAT(udpstat.udps_hdrops++);
            return;
        }
        ip = mtod(m, struct ip *);
    }
    uh = (struct udphdr *)((caddr_t)ip + iphlen);

    /*
     * Make mbuf data length reflect UDP length.
     * If not enough data to reflect UDP length, drop.
     */
    len = ntohs(uh->uh_ulen);
    if (ip->ip_len != len) {
        if (len > ip->ip_len || len < sizeof(struct udphdr)) {
            _COLLECT_STAT(udpstat.udps_badlen++);
            goto bad;
        }
        m_adj(m, len - ip->ip_len);
        /* ip->ip_len = len; */
    }
    /*
     * Save a copy of the IP header in case we want restore it
     * for sending an ICMP error message in response.
     */
    save_ip = *ip;

    /*
     * Checksum extended UDP header and data.
     */
    if (UDPCKSUM && uh->uh_sum) {
        ((struct ipovly *)ip)->ih_next = 0;
        ((struct ipovly *)ip)->ih_prev = 0;
        ((struct ipovly *)ip)->ih_x1 = 0;
        ((struct ipovly *)ip)->ih_len = uh->uh_ulen;
        uh->uh_sum = (UINT16) in_cksum(m, len + sizeof(struct ip));
        if (uh->uh_sum) {
            _COLLECT_STAT(udpstat.udps_badsum++);
            m_freem(m);
            return;
        }
    }

    /*
     * Check for DHCP packets.
     */
    if (uh->uh_sport == htons(DHCP_SERVER_PORT) &&
        uh->uh_dport == htons(DHCP_CLIENT_PORT)){
        /*
         * This is a DHCP packet.  Hand it off.
         */
        iphlen += sizeof(struct udphdr);
        m->m_len -= iphlen;
        m->m_pkthdr.len -= iphlen;
        m->m_data += iphlen;
        DhcpInput(m);
        return;
    }
    if (m->m_pkthdr.rcvif->if_flags & IFF_NOADDR) {
        /*
         * This packet was received by an interface without an IP
         * address.  We've allowed it to proceed this far as DHCP relies
         * on this capability (see above).  But since it's not
         * for DHCP, drop it now.
         */
        goto bad;
    }

    if (IN_MULTICAST(ntohl(ip->ip_dst.s_addr)) ||
#if MONARCH_ADHOC
        /*
         * Check all interfaces to see if the broadcast address
         * matches - jgb 12/06/98
         */
        in_broadcast_any_ifp(ip->ip_dst)
#else
        in_broadcast(ip->ip_dst, m->m_pkthdr.rcvif)
#endif
        ) {
        struct socket *last;
        /*
         * Deliver a multicast or broadcast datagram to *all* sockets
         * for which the local and remote addresses and ports match
         * those of the incoming datagram.  This allows more than
         * one process to receive multi/broadcasts on the same port.
         * (This really ought to be done for unicast datagrams as
         * well, but that would cause problems with existing
         * applications that open both address-specific sockets and
         * a wildcard socket listening to the same port -- they would
         * end up receiving duplicates of every unicast datagram.
         * Those applications open the multiple sockets to overcome an
         * inadequacy of the UDP socket interface, but for backwards
         * compatibility we avoid the problem here rather than
         * fixing the interface.  Maybe 4.5BSD will remedy this?)
         */

        /*
         * Construct sockaddr format source address.
         */
        udp_in.sin_port = uh->uh_sport;
        udp_in.sin_addr = ip->ip_src;
        m->m_len -= sizeof(struct udpiphdr);
        m->m_data += sizeof(struct udpiphdr);
        /*
         * Locate pcb(s) for datagram.
         * (Algorithm copied from raw_intr().)
         */
        last = NULL;
        for (inp = udb.inp_next; inp != &udb; inp = inp->inp_next) {
            if (inp->inp_lport != uh->uh_dport)
                continue;
            if (inp->inp_laddr.s_addr != INADDR_ANY) {
                if (inp->inp_laddr.s_addr !=
                    ip->ip_dst.s_addr)
                    continue;
            }
            if (inp->inp_faddr.s_addr != INADDR_ANY) {
                if (inp->inp_faddr.s_addr !=
                    ip->ip_src.s_addr ||
                    inp->inp_fport != uh->uh_sport)
                    continue;
            }

            if (last != NULL) {
                struct mbuf *n;

                if ((n = m_copy(m, 0, M_COPYALL)) != NULL) {
                    if (sbappendaddr(&last->so_rcv, (struct sockaddr *)&udp_in,
                         n, (struct mbuf *)0) == 0) {
                        m_freem(n);
                        _COLLECT_STAT(udpstat.udps_fullsock++);
                    } else
                        sorwakeup(last);
                }
            }
            last = inp->inp_socket;
            /*
             * Don't look for additional matches if this one does
             * not have either the SO_REUSEPORT or SO_REUSEADDR
             * socket options set.  This heuristic avoids searching
             * through all pcbs in the common case of a non-shared
             * port.  It * assumes that an application will never
             * clear these options after setting them.
             */
            if (((last->so_options&(SO_REUSEPORT|SO_REUSEADDR)) == 0))
                break;
        }

        if (last == NULL) {
            /*
             * No matching pcb found; discard datagram.
             * (No need to send an ICMP Port Unreachable
             * for a broadcast or multicast datgram.)
             */
            _COLLECT_STAT(udpstat.udps_noportbcast++);
            goto bad;
        }
        if (sbappendaddr(&last->so_rcv, (struct sockaddr *)&udp_in,
             m, (struct mbuf *)0) == 0) {
            _COLLECT_STAT(udpstat.udps_fullsock++);
            goto bad;
        }
        sorwakeup(last);
        return;
    }

    /*
     * Locate pcb for datagram.
     * We cache the last pcb, if not matched, do lookup.
     */
    inp = udp_last_inpcb;
    if (inp->inp_lport != uh->uh_dport ||
        inp->inp_fport != uh->uh_sport ||
        inp->inp_faddr.s_addr != ip->ip_src.s_addr ||
        inp->inp_laddr.s_addr != ip->ip_dst.s_addr) {
        inp = in_pcblookup(&udb, ip->ip_src, uh->uh_sport, ip->ip_dst,
                           uh->uh_dport, INPLOOKUP_WILDCARD);
        if (inp)
            udp_last_inpcb = inp;
        _COLLECT_STAT(udpstat.udpps_pcbcachemiss++);
    }
    if (inp == 0) {
        /*
         * Not for anybody we know.
         */
        _COLLECT_STAT(udpstat.udps_noport++);
        *ip = save_ip;
        ip->ip_len = (UINT16)(ip->ip_len + iphlen);
        icmp_error(m, ICMP_UNREACH, ICMP_UNREACH_PORT, 0, 0);
        return;
    }

    /*
     * Construct sockaddr format source address.
     * Stuff source address and datagram in user buffer.
     */
    udp_in.sin_port = uh->uh_sport;
    udp_in.sin_addr = ip->ip_src;

#if 0
    if (inp->inp_flags & INP_CONTROLOPTS) {
        struct mbuf **mp = &opts;
        if (inp->inp_flags & INP_RECVDSTADDR) {
            *mp = udp_saveopt((caddr_t) &ip->ip_dst,
                sizeof(struct in_addr), IP_RECVDSTADDR);
            if (*mp)
                mp = &(*mp)->m_next;
        }

        /* options were tossed above */
        if (inp->inp_flags & INP_RECVOPTS) {
            *mp = udp_saveopt((caddr_t) opts_deleted_above,
                sizeof(struct in_addr), IP_RECVOPTS);
            if (*mp)
                mp = &(*mp)->m_next;
        }
        /* ip_srcroute doesn't do what we want here, need to fix */
        if (inp->inp_flags & INP_RECVRETOPTS) {
            *mp = udp_saveopt((caddr_t) ip_srcroute(),
                sizeof(struct in_addr), IP_RECVRETOPTS);
            if (*mp)
                mp = &(*mp)->m_next;
        }
    }
#endif

    iphlen += sizeof(struct udphdr);
    m->m_len -= iphlen;
    m->m_pkthdr.len -= iphlen;
    m->m_data += iphlen;
    if (sbappendaddr(&inp->inp_socket->so_rcv, (struct sockaddr *)&udp_in,
        m, opts) == 0) {
        _COLLECT_STAT(udpstat.udps_fullsock++);
        goto bad;
    }
    sorwakeup(inp->inp_socket);
    return;
bad:
    m_freem(m);
    if (opts)
        m_freem(opts);
}

/*
 * Create a "control" mbuf containing the specified data
 * with the specified type for presentation with a datagram.
 */
static
struct mbuf *
udp_saveopt(caddr_t p, INT size, INT type)
{
    struct cmsghdr *cp;
    struct mbuf *m;

    if ((m = m_get(M_DONTWAIT, MT_CONTROL)) == NULL)
        return ((struct mbuf *) NULL);
    cp = (struct cmsghdr *) mtod(m, struct cmsghdr *);
    bcopy(p, CMSG_DATA(cp), size);
    size += sizeof(*cp);
    m->m_len = size;
    cp->cmsg_len = size;
    cp->cmsg_level = IPPROTO_IP;
    cp->cmsg_type = type;
    return (m);
}

/*
 * Notify a udp user of an asynchronous error;
 * just wake up so that he can collect error status.
 */
static void
udp_notify(struct inpcb *inp, INT errno)
{
    inp->inp_socket->so_error = (UINT16) errno;
    sorwakeup(inp->inp_socket);
    sowwakeup(inp->inp_socket);
}

void
udp_ctlinput(INT cmd, struct sockaddr *sa, caddr_t arg)
{
    struct ip *ip = (struct ip *)arg;
    struct udphdr *uh;
    extern struct in_addr zeroin_addr;
    extern UINT8 inetctlerrmap[];

    if (!PRC_IS_REDIRECT(cmd) &&
        ((unsigned)cmd >= PRC_NCMDS || inetctlerrmap[cmd] == 0))
        return;
    if (ip) {
        uh = (struct udphdr *)((caddr_t)ip + (ip->ip_hl << 2));
        in_pcbnotify(&udb, sa, uh->uh_dport, ip->ip_src, uh->uh_sport,
            cmd, udp_notify);
    } else
        in_pcbnotify(&udb, sa, 0, zeroin_addr, 0, cmd, udp_notify);
}

INT udp_output(struct inpcb *inp, struct mbuf *m, struct mbuf *addr,
               struct mbuf *control)
{
    struct udpiphdr *ui;
    INT len = m->m_pkthdr.len;
    struct in_addr laddr = {0};
    INT IpFlags, error = 0;

    if (control)
        m_freem(control);        /* XXX */

    if (addr) {
        laddr = inp->inp_laddr;
        if (inp->inp_faddr.s_addr != INADDR_ANY) {
            error = EISCONN;
            goto release;
        }
        /*
         * Must block input while temporarily connected.
         */
        NetLock();
        error = in_pcbconnect(inp, addr);
        if (error) {
            NetUnlock();
            goto release;
        }
    } else {
        if (inp->inp_faddr.s_addr == INADDR_ANY) {
            error = ENOTCONN;
            goto release;
        }
    }
    /*
     * Calculate data length and get a mbuf
     * for UDP and IP headers.
     */
    M_PREPEND(m, sizeof(struct udpiphdr), M_DONTWAIT);
    if (m == 0) {
        error = ENOBUFS;
        goto release;
    }

    /*
     * Fill in mbuf with extended UDP header
     * and addresses and length put into network format.
     */
    ui = mtod(m, struct udpiphdr *);
    ui->ui_next = ui->ui_prev = 0;
    ui->ui_x1 = 0;
    ui->ui_pr = IPPROTO_UDP;
    ui->ui_len = htons((UINT16)(len + sizeof(struct udphdr)));
    ui->ui_src = inp->inp_laddr;
    ui->ui_dst = inp->inp_faddr;
    ui->ui_sport = inp->inp_lport;
    ui->ui_dport = inp->inp_fport;
    ui->ui_ulen = ui->ui_len;

    /*
     * Stuff checksum and output datagram.
     */
    ui->ui_sum = 0;
    if (UDPCKSUM) {
        if ((ui->ui_sum = (UINT16) in_cksum(m, sizeof(struct udpiphdr) + len)) == 0)
        ui->ui_sum = 0xffff;
    }
    ((struct ip *)ui)->ip_len = (UINT16)(sizeof(struct udpiphdr) + len);
#ifdef SUPPORT_IP_SOCKOPTS
    ((struct ip *)ui)->ip_ttl = inp->inp_ip.ip_ttl;    /* XXX */
    ((struct ip *)ui)->ip_tos = inp->inp_ip.ip_tos;    /* XXX */
#else
    ((struct ip *)ui)->ip_ttl = IP_DEFTTL;
    ((struct ip *)ui)->ip_tos = 0;
#endif
    _COLLECT_STAT(udpstat.udps_opackets++);

    if (inp->inp_socket)
        IpFlags = inp->inp_socket->so_options & (SO_DONTROUTE | SO_BROADCAST);
    else
        IpFlags = IP_ALLOWBROADCAST;  /* DHCP case */

    error = ip_output(m, inp->inp_options, &inp->inp_route, IpFlags,
                      inp->inp_moptions);

    if (addr) {
        in_pcbdisconnect(inp);
        inp->inp_laddr = laddr;
        NetUnlock();
    }
    return (error);

release:
    m_freem(m);
    return (error);
}

UINT32 udp_sendspace = 65508;  /* really max datagram size */
UINT32 udp_recvspace = 40 * (1024 + sizeof(struct sockaddr_in));
                       /* 40 1K datagrams */

INT udp_usrreq(struct socket *so, INT req, struct mbuf *m, struct mbuf *addr,
               struct mbuf *control)
{
    struct inpcb *inp = sotoinpcb(so);
    INT error = 0;

    if (req == PRU_CONTROL)
        return (in_control(so, (INT)m, (caddr_t)addr,
            (struct ifnet *)control));
    if (inp == NULL && req != PRU_ATTACH) {
        error = EINVAL;
        goto release;
    }
    /*
     * Note: need to block udp_input while changing
     * the udp pcb queue and/or pcb addresses.
     */
    switch (req) {

    case PRU_ATTACH:
        if (inp != NULL) {
            error = EINVAL;
            break;
        }
        NetLock();
        error = in_pcballoc(so, &udb);
        NetUnlock();
        if (error)
            break;
        error = soreserve(so, udp_sendspace, udp_recvspace);
        if (error)
            break;
#ifdef SUPPORT_IP_SOCKOPTS
        ((struct inpcb *) so->so_pcb)->inp_ip.ip_ttl = IP_DEFTTL;
#endif
        break;

    case PRU_DETACH:
        udp_detach(inp);
        break;

    case PRU_BIND:
        NetLock();
        error = in_pcbbind(inp, addr);
        NetUnlock();
        break;

    case PRU_LISTEN:
        error = EOPNOTSUPP;
        break;

    case PRU_CONNECT:
        if (inp->inp_faddr.s_addr != INADDR_ANY) {
            error = EISCONN;
            break;
        }
        NetLock();
        error = in_pcbconnect(inp, addr);
        NetUnlock();
        if (error == 0)
            soisconnected(so);
        break;

    case PRU_CONNECT2:
        error = EOPNOTSUPP;
        break;

    case PRU_ACCEPT:
        error = EOPNOTSUPP;
        break;

    case PRU_DISCONNECT:
        if (inp->inp_faddr.s_addr == INADDR_ANY) {
            error = ENOTCONN;
            break;
        }
        NetLock();
        in_pcbdisconnect(inp);
        inp->inp_laddr.s_addr = INADDR_ANY;
        NetUnlock();
        so->so_state &= ~SS_ISCONNECTED;        /* XXX */
        break;

    case PRU_SHUTDOWN:
        socantsendmore(so);
        break;

    case PRU_SEND:
        return (udp_output(inp, m, addr, control));

    case PRU_ABORT:
        soisdisconnected(so);
        udp_detach(inp);
        break;

    case PRU_SOCKADDR:
        in_setsockaddr(inp, addr);
        break;

    case PRU_PEERADDR:
        in_setpeeraddr(inp, addr);
        break;

    case PRU_SENSE:
        /*
         * stat: don't bother with a blocksize.
         */
        return (0);

    case PRU_SENDOOB:
    case PRU_FASTTIMO:
    case PRU_SLOWTIMO:
    case PRU_PROTORCV:
    case PRU_PROTOSEND:
        error =  EOPNOTSUPP;
        break;

    case PRU_RCVD:
    case PRU_RCVOOB:
        return (EOPNOTSUPP);    /* do not free mbuf's */

    default:
        panic(_T("udp_usrreq"));
    }

release:
    if (control) {
        printf(_T("udp control data unexpectedly retained\n"));
        m_freem(control);
    }
    if (m)
        m_freem(m);
    return (error);
}

static void
udp_detach(struct inpcb *inp)
{
    struct socket *so = inp->inp_socket;

    NetLock();

    soisdisconnected(so);
    if (inp == udp_last_inpcb)
        udp_last_inpcb = &udb;
    in_pcbdetach(inp);
    NetUnlock();
}

#if 0
/*
 * Sysctl for udp variables.
 */
udp_sysctl(INT *name, UINT namelen, void *oldp, size_t *oldlenp, void *newp,
           size_t newlen)
{
    extern int ip_ttl;

    /* all sysctl names at this level are terminal */
    if (namelen != 1)
        return (EINVAL); /* af: was ENOTDIR */

    switch (name[0]) {
    case UDPCTL_CHECKSUM:
        return (sysctl_int(oldp, oldlenp, newp, newlen, &udpcksum));
    default:
        return (ENOPROTOOPT);
    }
    /* NOTREACHED */
}
#endif

#endif /* __WANTS_UDP */
