/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* -------------------------------------------------------------------------
 *
 * Module:
 *
 *    winsock.c
 *
 * Purpose:
 *
 *    WINSOCK.DLL, socket compliant interface to IEndpoint.
 *
 * Authors:
 *    A. Forin (sandrof)
 *    Brian Zill (bzill)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <netif.h>

#include <winsock.h>
#include <stdio.h>
#include <string.h>

struct ifnet;
struct mbuf;

#include <dhcp_var.h>

//#undef WINAPI
//#define WINAPI
#define DEBUG 0
#undef DBGOUT
#define DBGOUT(Class, Level, x) ;


/*
 * System call interface to the socket abstraction.
 */
#define MAKE_SOCKET(x)  ((SOCKET) x)
#define GET_ENDPOINT(x) ((PIENDPOINT)x)

#define PROTOCOL                _T("COB/protocol.cob");

PIENDPOINTFACTORY pSocketBldr = NULL;

SCODE BindSocketBuilder(PIENDPOINTFACTORY *ppSocketBldr)
{
    PINAMESPACE pINameSpace = CurrentNameSpace();
    PIUNKNOWN pUnk;
    SCODE sc;
    const _TCHAR *Name = PROTOCOL;

    /*
     * Find our Endpoint Factory, it should be in the namespace.
     */
    sc = pINameSpace->v->Bind(pINameSpace, Name,
                              NAME_SPACE_READ, &pUnk);
    DBGOUT(DbgWinsock, 0, dprintf(_T("Bind got 0x%08x 0x%08x\n"), sc, pUnk));
    if (FAILED(sc)){
#ifdef _DEBUG
        printf("Bind %s Failed 0x%08x\n", Name, sc);
#endif
        *ppSocketBldr = NULL;
        return sc;
    }

    /*
     * Make sure it's the right thing.
     */
    sc = pUnk->v->QueryInterface(pUnk, &IID_IEndpointFactory, (void**)ppSocketBldr);
    DBGOUT(DbgWinsock, 0, dprintf(_T("QI got 0x%08x 0x%08x\n"), sc,
           *ppSocketBldr));
    if (FAILED(sc))
        DBGOUT(DbgWinsock, 1, dprintf(_T("QI Failed 0x%08x\n"), sc));

    pUnk->v->Release(pUnk);
    return sc;
}

/*
 * Socket Routines
 */

/*
 * Wait for an incoming connection.
 */
SOCKET WINAPI accept(SOCKET s, struct sockaddr *pName, int *pNameLen)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    PIENDPOINT pNew;
    struct sockaddr DummyAddr;
    UINT32 DummySize = sizeof(DummyAddr);
    SCODE sc;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

   /*
    * BUGBUG: Accept doesn't allow the pointers to the returned address area
    * to be optional (i.e. NULL).  This was a problem in an old IPC
    * implementation, actually, and this was a workaround for that.
    * For now, plug in a dummy area if we're given a null name or name length.
    */
    if (pName == 0 || pNameLen == 0 || *pNameLen == 0)
        sc = pIEnd->v->Accept(pIEnd, &pNew, &DummyAddr, (UINT32)DummySize,
                              (UINT32 *)&DummySize);
    else
        sc = pIEnd->v->Accept(pIEnd, &pNew, pName,
                              (UINT32)((pNameLen) ? *pNameLen : 0),
                              (UINT32 *)pNameLen);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return INVALID_SOCKET;
    }

    return MAKE_SOCKET(pNew);
}

/*
 * Bind a socket to a particular local address.
 */
int WINAPI bind(SOCKET s, const struct sockaddr *name, int namelen)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->Bind(pIEnd, name, namelen);

    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }
    return 0;
}

/*
 * Explicitly close down a socket and release our reference to it.
 */
int WINAPI closesocket(SOCKET s)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;

    DBGOUT(DbgWinsock, 0, dprintf(_T("In closesocket for 0x%08x\n"), s));

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->Close(pIEnd);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }

    (void) pIEnd->v->Release(pIEnd);

    return 0;
}

/*
 * Initiate a connection from this socket to the specified remote address.
 */
int WINAPI connect(SOCKET s, const struct sockaddr *name, int namelen)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->Connect(pIEnd, name, namelen);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }

    return 0;
}

/*
 * Get name of our socket's peer (if we're connected).
 */
int WINAPI getpeername(SOCKET s, struct sockaddr *asa, int *alen)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;
    UINT32 Len = 0, *pLen = &Len;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->GetName(pIEnd, asa, (alen) ? *alen : 0, pLen,
                           ENDPOINT_ADDR_PEER);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }
    if (alen)
        *alen = (int)Len;

    return 0;
}

/*
 * Get name of our socket.
 */
int WINAPI getsockname(SOCKET s, struct sockaddr *asa, int *alen)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;
    UINT32 Len = 0, *pLen = &Len;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->GetName(pIEnd, asa, (alen) ? *alen : 0, pLen,
                           ENDPOINT_ADDR_MINE);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }
    if (alen)
        *alen = (int)Len;

    return 0;
}

/*
 * Get our socket's option settings.
 */
int WINAPI getsockopt(SOCKET s, int level, int name, char *val, int *avalsize)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;
    UINT32 Len = 0, *pLen = &Len;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->GetSockOpt(pIEnd, level, name, (BYTE *)val,
                              (avalsize) ? *avalsize : 0, pLen);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }
    if (avalsize)
        *avalsize = (int)Len;

    return 0;
}

/*
 * Perform an ioctl on a socket.
 */
int WINAPI ioctlsocket(SOCKET s, long cmd, u_long *argp)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->Ioctl(pIEnd, (UINT32)cmd, (BYTE*)argp);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }

    return 0;
}

/*
 * Ready to accept connections.
 */
int WINAPI listen(SOCKET s, int backlog)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->Listen(pIEnd, backlog);

    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }
    return 0;
}

/*
 * Receive data from our peer.
 */
int WINAPI recv(SOCKET s, char *buf, int len, int flags)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    UINT64 Flags64;
    SCODE sc;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    Int32ToInt64(Flags64,flags);
    sc = pIEnd->v->ReadAt(pIEnd, Flags64, (BYTE *)buf, (UINT) len,
                          (PUINT)&len);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }

    return len;
}

/*
 * Receive data from somewhere and return the origin.
 */
int WINAPI recvfrom(SOCKET s, char *buf, int len, int flags,
                    struct sockaddr *from, int *fromlen)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    struct sockaddr DummyAddr;
    UINT32 DummySize = sizeof(DummyAddr);
    SCODE sc;

    DBGOUT(DbgWinsock, 0,
           dprintf(_T("recvfrom: s = 0x%08x, buf = 0x%x, len = %d, "),
                   s, buf, len));
    DBGOUT(DbgWinsock, 0,
           dprintf(_T("flags = 0x%x, from = 0x%x, fromlen = 0x%x\n"),
                   flags, from, fromlen));

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    /*
     * BUGBUG: RecvFrom doesn't allow the pointers to the returned address area
     * to be optional (i.e. NULL).  This is a MMLITE IPC problem, actually.
     * For now, plug in a dummy area if we're given a null from or from length.
     */
    if (from == 0 || fromlen == 0 || *fromlen == 0)
        sc = pIEnd->v->RecvFrom(pIEnd, (BYTE *)buf, len, (UINT32*)&len, flags,
                                &DummyAddr, (UINT32)DummySize,
                                (UINT32 *)&DummySize);
    else
        sc = pIEnd->v->RecvFrom(pIEnd, (BYTE *)buf, len, (UINT32*)&len, flags, from,
                                (UINT32)((fromlen) ? *fromlen : 0),
                                (UINT32 *)fromlen);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }

    return len;
}

/*
 * Check on socket availability.
 */
int WINAPI select(int nfds, fd_set *readfds, fd_set *writefds,
                  fd_set *exceptfds, const struct timeval *tv)
{
    PIENDPOINT SoleEndpt, *EndptArray;
    BYTE SoleEventType, *EventTypeArray;
    UINT Count, TotalCount, Loop, Marker;
    TIME TimeToWait;
    PIHEAP pHeap = NULL;
    SCODE sc;

    UnusedParameter(nfds);

    /*
     * Convert from three of winsock's 'fd_set's to a single array of
     * MMLITE 'IEndpoint's.  Another array of bytes holds the bitfield
     * specifying what event we're interested in for each Endpoint.
     * Optimize the common case of only one Endpoint waited on.
     */
    TotalCount = 0;
    if (readfds)
        TotalCount = (UINT)readfds->fd_count;
    if (writefds)
        TotalCount += (UINT)writefds->fd_count;
    if (exceptfds)
        TotalCount += (UINT)exceptfds->fd_count;

    if (TotalCount == 1 || TotalCount == 0) {
        /*
         * One or no Endpoints, use local variables and save Alloc cost.
         */
        EventTypeArray = &SoleEventType;
        EndptArray = &SoleEndpt;
    } else {
        /*
         * Allocate space for both arrays in one chunk and then split it.
         */
        pHeap = CurrentHeap();
        EventTypeArray = (BYTE *)pHeap->v->Alloc(pHeap,
                                             HEAP_DEBUG_NAME("wsock32:select"),
                           (TotalCount * sizeof(PIENDPOINT)) + TotalCount, 0);
        if (EventTypeArray == NULL) {
            sc = E_NOT_ENOUGH_MEMORY;
            goto ErrorExit;
        }
        EndptArray = (PIENDPOINT *)(EventTypeArray + TotalCount);
    }

    /*
     * Now load up arrays from the fd_sets.
     */
    Loop = 0;
    if (readfds)
         for (;Loop < readfds->fd_count;Loop++) {
             EndptArray[Loop] = GET_ENDPOINT(readfds->fd_array[Loop]);
             EventTypeArray[Loop] = ENDPOINT_READ;
         }
    Marker = Loop;
    if (writefds)
         for (Loop = 0;Loop < writefds->fd_count;Loop++) {
             EndptArray[Marker + Loop] =
                 GET_ENDPOINT(writefds->fd_array[Loop]);
             EventTypeArray[Loop] = ENDPOINT_WRITE;
         }
    Marker += Loop;
    if (exceptfds)
         for (Loop = 0;Loop < exceptfds->fd_count;Loop++) {
             EndptArray[Marker + Loop] =
                 GET_ENDPOINT(exceptfds->fd_array[Loop]);
             EventTypeArray[Loop] = ENDPOINT_EXCEPTION;
         }

    /*
     * Convert winsock 'struct timeval' to 'TIME'.
     * Note select() takes a relative amount of time to wait.
     */
    if (tv) {
        INT64 us;

        /* Convert-relative the microseconds part */
        Int32ToInt64(us,TIME_RELATIVE(tv->tv_usec));
        us = Int64TimesInt32(us,TIME_MICROS(1));

        /* Convert-relative the seconds part */
        Int32ToInt64(TimeToWait,TIME_RELATIVE(tv->tv_sec));
        TimeToWait = Int64TimesInt32(TimeToWait,TIME_SECONDS(1));

        TimeToWait = Int64Add(TimeToWait,us);
    } else
        TimeToWait = TIME_FOREVER;

    /*
     * Find the socket builder (if we haven't already).
     */
    if (pSocketBldr == NULL) {
#ifdef mmheavy
        sc = RegisterEndpointInterfaces();
        if (FAILED(sc))
            goto ErrorExit;
#endif
        sc = BindSocketBuilder(&pSocketBldr);
        if (FAILED(sc))
            goto ErrorExit;
    }

    /*
     * Make the actual call.  EventTypeArray is in/out.
     */
    sc = pSocketBldr->v->WaitForEndpoints(pSocketBldr, EndptArray,
             EventTypeArray, TotalCount, TimeToWait);
    if (FAILED(sc))
        goto ErrorExit;

    /*
     * Note those Endpoints which have had their events occur,
     * delete all other sockets from the fd_sets.
     */
    TotalCount = 0;
    Loop = 0;
    if (readfds) {
         Count = 0;
         for (;Loop < readfds->fd_count;Loop++) {
             if (EventTypeArray[Loop] & ENDPOINT_ISREADABLE)
                 readfds->fd_array[Count++] = MAKE_SOCKET(EndptArray[Loop]);
         }
         readfds->fd_count = (unsigned short)Count;
         TotalCount += Count;
    }
    Marker = Loop;
    if (writefds) {
         Count = 0;
         for (Loop = 0;Loop < writefds->fd_count;Loop++) {
             if (EventTypeArray[Marker + Loop] & ENDPOINT_ISWRITEABLE)
                 writefds->fd_array[Count++] =
                       MAKE_SOCKET(EndptArray[Marker + Loop]);
         }
         writefds->fd_count = (unsigned short)Count;
         TotalCount += Count;
    }
    Marker += Loop;
    if (exceptfds) {
         Count = 0;
         for (Loop = 0;Loop < exceptfds->fd_count;Loop++) {
             if (EventTypeArray[Marker + Loop] & ENDPOINT_ISEXCEPTIONAL)
                 exceptfds->fd_array[Count++] =
                       MAKE_SOCKET(EndptArray[Marker + Loop]);
         }
         exceptfds->fd_count = (unsigned short)Count;
         TotalCount += Count;
    }

ErrorExit:
    /*
     * Free any buffers we allocated above.
     */
    if (EventTypeArray && EventTypeArray != &SoleEventType)
        pHeap->v->Free(pHeap, 0, (void *)EventTypeArray);

    if (FAILED(sc)) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }
    return TotalCount;
}

/*
 * Send data to our peer.
 */
int WINAPI send(SOCKET s, const char *buf, int len, int flags)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    UINT64 Flags64;
    SCODE sc;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    Int32ToInt64(Flags64,flags);
    sc = pIEnd->v->WriteAt(pIEnd, Flags64, (BYTE *)buf, (UINT) len,
                           (PUINT)&len);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }

    return len;
}

/*
 * Send to a given destination.
 */
int WINAPI sendto(SOCKET s, const char *buf, int len, int flags,
                  const struct sockaddr *to, int tolen)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->SendTo(pIEnd, (BYTE *)buf, len, (UINT32*)&len, flags, to, tolen);

    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }

    return len;
}

/*
 * Set our socket's option settings.
 */
int WINAPI setsockopt(SOCKET s, int level, int name, const char *val,
                      int valsize)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->SetSockOpt(pIEnd, level, name, (BYTE *)val, valsize);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }

    return 0;
}

/*
 * Break a connection.
 */
int WINAPI shutdown(SOCKET s, int how)
{
    PIENDPOINT pIEnd = GET_ENDPOINT(s);
    SCODE sc;

   /*
    * BUGBUG: Should do better verification that socket is legit.
    */
    if (s == INVALID_SOCKET || pIEnd == NULL) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, WSAENOTSOCK);
        return INVALID_SOCKET;
    }

    sc = pIEnd->v->Shutdown(pIEnd, how);
    if (FAILED(sc)){
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
        return SOCKET_ERROR;
    }

    return 0;
}

/*
 * Create a brand new socket.
 */
SOCKET WINAPI socket(int domain, int type, int protocol)
{
    PIENDPOINT pNew;
    SCODE sc;

    /*
     * See if we already have a pointer to the socket builder cached.
     */
    if (pSocketBldr == NULL) {
#ifdef mmheavy
         sc = RegisterEndpointInterfaces();
         if (FAILED(sc))
             goto ErrorExit;
#endif
         sc = BindSocketBuilder(&pSocketBldr);
         if (FAILED(sc))
             goto ErrorExit;
    }

    /*
     * Ask it to create a new socket, please.
     */
    sc = pSocketBldr->v->CreateEndpoint(pSocketBldr, domain, type,
             protocol, &pNew);
    if (FAILED(sc))
        goto ErrorExit;

    return MAKE_SOCKET(pNew);

ErrorExit:
    {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, sc);
    }
    return INVALID_SOCKET;
}


/*--------------------------------------------------------------------*\

                      Host name database handling

\*--------------------------------------------------------------------*/
#ifdef mmheavy
#ifdef UNICODE
void RTLCALLTYPE AsciiStringer(char *pDst, const TCHAR *pSrc)
{
    while ((*pDst++ = (char) *pSrc++) != 0)
        continue;
}
#endif

#ifdef _MSC_VER
#pragma warning(disable:4055)  /* from data pointer to function pointer */
#endif

static void sprintf(char *Where, const _TCHAR *Fmt, ...)
{
    va_list args;
    STRINGER Stringer;

#ifdef UNICODE
    Stringer = (STRINGER) &AsciiStringer;
#else
    Stringer = (STRINGER) GetRtlFunctionAddress(StrCpy);
#endif

    va_start(args, Fmt);
    StrFormat(Stringer, Where, Fmt, args);
    va_end(args);
}
#endif

#if 0
/*
 * Database functions
 */
struct hostent * WINAPI gethostbyaddr(const char *addr, int len, int type)
{
    static struct hostent TempEnt;  /* BUGBUG: not multi-caller safe */
    static char TempName[18];  /* ditto */
    static char *TempAlias[1];  /* ditto */
    static char TempAddr[4];  /* ditto */
    static char *TempList[2];  /* ditto */

    if (len != 4 || type != PF_INET) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAHOST_NOT_FOUND));
        return NULL;
    }

    /*
     * BUGBUG: We currently don't have any database support.
     * BUGBUG: For now, just stringify IP addresses.
     */
    (void) strcpy(TempName, inet_ntoa(*(struct in_addr *)addr));
    TempAlias[0] = NULL;
    (void) memcpy(TempAddr, (char *)addr, 4);
    TempList[0] = TempAddr;
    TempList[1] = NULL;
    
    TempEnt.h_name = TempName;
    TempEnt.h_aliases = TempAlias;
    TempEnt.h_addrtype = PF_INET;
    TempEnt.h_length = 4;
    TempEnt.h_addr_list = TempList;

    return &TempEnt;
}

struct hostent * WINAPI gethostbyname(const char *name)
{
     UINT32 hostaddr;

    /*
     * BUGBUG: We currently don't have any database support.
     * BUGBUG: For now, just resolve IP address strings.  Note this
     * BUGBUG: this is expressly forbidden in the winsock 1.1 doc!
     */
    hostaddr = inet_addr(name);

    if (hostaddr == INADDR_NONE) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAHOST_NOT_FOUND));
        return NULL;
    }

    return gethostbyaddr((char *)&hostaddr, 4, PF_INET);
}

int WINAPI gethostname(char *name, int namelen)
{
    SOCKET Skt;
    char *HName;
    struct ifreq {  /* BUGBUG: reduced form of ifreq just for this hack */
        char ifr_name[16];
        struct sockaddr ifr_addr;
    } NetifInfo;

    DBGOUT(DbgWinsock, 0, dprintf(
           _T("In gethostname with name = 0x%08x, namelen = %d\n"),

           name, namelen));

    /*
     * BUGBUG: Should MMLITE even support the concept of a hostname?
     * BUGBUG: For now, our hostname is an ASCII stringified version
     * BUGBUG: of one of our interface's IP address.  This is an ugly hack.
     */
#define SIOCGIFADDR _IOWR('i', 33, struct ifreq) /* get interface address */

    /*
     * Find out what one of our IP addresses is.
     */
    Skt = socket(AF_INET, SOCK_DGRAM, 0);
    if (Skt == INVALID_SOCKET) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSANOTINITIALISED));
        return SOCKET_ERROR;
    }

    strcpy(NetifInfo.ifr_name,"*0");
    if (ioctlsocket(Skt, SIOCGIFADDR, (u_long *)&NetifInfo) < 0) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAENETDOWN));
        goto ErrorExit2;
    }

    /*
     * BUGBUG: more hackery -- winsock.h uses old-style sockaddr.
     * New sockaddr (returned by SIOCGIFADDR) splits old 16-bit sa_family
     * field into an 8-bit sa_len and a new 8-bit sa_family.
     */
    if (NetifInfo.ifr_addr.sa_family & 0xff != sizeof(struct sockaddr_in)) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAENETDOWN));
        goto ErrorExit2;
    }
    if (NetifInfo.ifr_addr.sa_family & 0xff00 != AF_INET) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAENETDOWN));
        goto ErrorExit2;
    }

    /*
     * Stringify it and hand it back
     */
    HName = inet_ntoa(((struct sockaddr_in *)&(NetifInfo.ifr_addr))->sin_addr);
    if (namelen <= (int)strlen(HName)) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAEFAULT));
        goto ErrorExit2;
    }
    strcpy(name, HName);

    (void) closesocket(Skt);
    return 0;

ErrorExit2:
    (void) closesocket(Skt);
    return SOCKET_ERROR;
}

#else

static INT ipstringify(char *name, int namelen, char *addr)
{
    return _snprintf(name, namelen, "%u.%u.%u.%u",
                     ((unsigned)addr[0] & 0xff),
                     ((unsigned)addr[1] & 0xff),
                     ((unsigned)addr[2] & 0xff),
                     ((unsigned)addr[3] & 0xff));
}

int WINAPI gethostname(char *name, int namelen)
{
    SOCKET Skt;
    struct DhcpInfo DhcpInfo;
    struct ifreq {  /* BUGBUG: reduced form of ifreq just for this hack */
        char ifr_name[16];
        struct sockaddr ifr_addr;
    } *pNetifInfo = (PTR) &DhcpInfo;

    DBGOUT(DbgWinsock, 0, dprintf(
           _T("In gethostname with name = 0x%08x, namelen = %d\n"),

           name, namelen));

    /*
     * Find out what one of our IP addresses is.
     */
    Skt = socket(AF_INET, SOCK_DGRAM, 0);
    if (Skt == INVALID_SOCKET) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSANOTINITIALISED));
        return SOCKET_ERROR;
    }

    strcpy(pNetifInfo->ifr_name,"*0");
    if (ioctlsocket(Skt, SIOCGDHCPINFO, (PTR) &DhcpInfo) < 0) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAENETDOWN));
        goto ErrorExit2;
    }

    if (DhcpInfo.HostName) {
        strncpy(name, DhcpInfo.HostName, namelen);
        if (namelen > 0)
            name[namelen-1] = '\0';     /* grr, whoever specified strncpy */
    } else {
        /* DHCP did not provide host name.  Use IP address and DNS */
        struct hostent *hp;
        strcpy(pNetifInfo->ifr_name,"*0");
        if (ioctlsocket(Skt, SIOCGIFADDR, (u_long *)pNetifInfo) < 0) {
            PITHREAD th = CurrentThread();
            th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAENETDOWN));
            goto ErrorExit2;
        }

        /*
         * BUGBUG: more hackery -- winsock.h uses old-style sockaddr.
         * New sockaddr (returned by SIOCGIFADDR) splits old 16-bit sa_family
         * field into an 8-bit sa_len and a new 8-bit sa_family.
         * BUGBUG: Handle it either way. We seem to have a newer winsock.h now
         */
        if ((sizeof(pNetifInfo->ifr_addr.sa_family) == 2)
            ? ((pNetifInfo->ifr_addr.sa_family & 0xff) /* old style */
               != sizeof(struct sockaddr_in))
            : ((pNetifInfo->ifr_addr.sa_len) /* new style */
               != sizeof(struct sockaddr_in)))
        {
            PITHREAD th = CurrentThread();
            th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAENETDOWN));
            goto ErrorExit2;
        }
        if ((sizeof(pNetifInfo->ifr_addr.sa_family) == 2)
            ? ((pNetifInfo->ifr_addr.sa_family & 0xff00) != AF_INET)
            : (pNetifInfo->ifr_addr.sa_family != AF_INET))
        {
            PITHREAD th = CurrentThread();
            th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAENETDOWN));
            goto ErrorExit2;
        }

        hp = gethostbyaddr((PTR) &((struct sockaddr_in *)&pNetifInfo->ifr_addr)->sin_addr, sizeof(struct in_addr), AF_INET);

        if (hp) {
            strncpy(name, hp->h_name, namelen);
        } else {
            /* Ok, so what if we dont have a name ?
             * A: use the stringified IP address as such,
             * no matter who might complain about it.
             */
            ipstringify(name,namelen,
                        (char *) &((struct sockaddr_in *)&pNetifInfo->ifr_addr)->sin_addr);
        }
        if (namelen > 0)
            name[namelen-1] = '\0';     /* grr, whoever specified strncpy */
    }

    (void) closesocket(Skt);

    return 0;

ErrorExit2:
    (void) closesocket(Skt);
    return SOCKET_ERROR;
}

UINT32 gethostaddr(void)
{
    SOCKET Skt;
    struct DhcpInfo DhcpInfo;
    struct ifreq {  /* BUGBUG: reduced form of ifreq just for this hack */
        char ifr_name[16];
        struct sockaddr ifr_addr;
    } *pNetifInfo = (PTR) &DhcpInfo;

    /*
     * Find out what one of our IP addresses is.
     */
    Skt = socket(AF_INET, SOCK_DGRAM, 0);
    if (Skt == INVALID_SOCKET) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSANOTINITIALISED));
        return INADDR_NONE;
    }

    strcpy(pNetifInfo->ifr_name,"*0");
    if (ioctlsocket(Skt, SIOCGIFADDR, (u_long *)pNetifInfo) < 0) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAENETDOWN));
        goto ErrorExit2;
    }

    /*
     * BUGBUG: more hackery -- winsock.h uses old-style sockaddr.
     * New sockaddr (returned by SIOCGIFADDR) splits old 16-bit sa_family
     * field into an 8-bit sa_len and a new 8-bit sa_family.
     * BUGBUG: Handle it either way. We seem to have a newer winsock.h now
     */
    if ((sizeof(pNetifInfo->ifr_addr.sa_family) == 2)
        ? ((pNetifInfo->ifr_addr.sa_family & 0xff) /* old style */
           != sizeof(struct sockaddr_in))
        : ((pNetifInfo->ifr_addr.sa_len) /* new style */
           != sizeof(struct sockaddr_in)))
    {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAENETDOWN));
        goto ErrorExit2;
    }
    if ((sizeof(pNetifInfo->ifr_addr.sa_family) == 2)
        ? ((pNetifInfo->ifr_addr.sa_family & 0xff00) != AF_INET)
        : (pNetifInfo->ifr_addr.sa_family != AF_INET))
    {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAENETDOWN));
        goto ErrorExit2;
    }

    (void) closesocket(Skt);

    return (((struct sockaddr_in *)&pNetifInfo->ifr_addr)->sin_addr.s_addr);

ErrorExit2:
    (void) closesocket(Skt);
    return INADDR_NONE;
}
#endif

struct protoent * WINAPI getprotobyname(const char *name)
{
    PITHREAD th = CurrentThread();
    th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAEOPNOTSUPP));
    UnusedParameter(name);
    return NULL;
}

struct protoent * WINAPI getprotobynumber(int proto)
{
    PITHREAD th = CurrentThread();
    th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAEOPNOTSUPP));
    UnusedParameter(proto);
    return NULL;
}

struct servent * WINAPI getservbyname(const char * name, const char * proto)
{
    PITHREAD th = CurrentThread();
    th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAEOPNOTSUPP));
    UnusedParameter(name);
    UnusedParameter(proto);
    return NULL;
}

struct servent * WINAPI getservbyport(int port, const char * proto)
{
    PITHREAD th = CurrentThread();
    th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAEOPNOTSUPP));
    UnusedParameter(port);
    UnusedParameter(proto);
    return NULL;
}

/*
 * The following section contains Microsoft Windows-specific
 * extensions to the standard Berkeley socket interface.
 */

/*
 * Return last socket error
 */
int WINAPI WSAGetLastError(void)
{
    PITHREAD th = CurrentThread();
    return th->v->GetValue(th, THREAD_VALUE_LAST_ERROR) & 0xffff; /* strip off excess error bits */
}

/*
 * Initialization and version control.
 * Note that WSAStartup returns error codes directly,
 * and not via WSASet/BaseGetLastError.
 */
#define MY_VERSION 0x0101 /* high byte is minor version, low byte is major */

int WINAPI WSAStartup(UINT16 wVersionRequested, LPWSADATA lpWSAData)
{
    SCODE sc;
    
    /*
     * See if we already have a pointer to the socket builder cached.
     * If not, get one.  This will fail if the network subsystem is not
     * loaded and thus provides a check that it is.
     */
    if (pSocketBldr == NULL) {
#ifdef mmheavy
         sc = RegisterEndpointInterfaces();
         if (FAILED(sc))
             return WSASYSNOTREADY;
#endif
         sc = BindSocketBuilder(&pSocketBldr);
         if (FAILED(sc))
             return WSASYSNOTREADY;
    }

    /*
     * We only support one version, but winsock spec says to only return
     * WSAVERNOTSUPPORTED if the version requested by the application is
     * less than the lowest one we support.  If the app requests a higher
     * version, return the WSADATA structure and let the app decide whether
     * our version is acceptable.
     */

    if ((wVersionRequested & 0x00ff) < (MY_VERSION & 0x00ff) ||
        ((wVersionRequested & 0x00ff) == (MY_VERSION & 0x00ff) &&
        (wVersionRequested & 0xff00) < (MY_VERSION & 0xff00))) {
        PITHREAD th = CurrentThread();
        th->v->SetValue(th, THREAD_VALUE_LAST_ERROR, STATUS_WIN32_ERROR(WSAVERNOTSUPPORTED));
        return WSAVERNOTSUPPORTED;
    }

    /*
     * Meaning of fields in WSAData:
     *
     * wVersion: version we expect caller to use
     * wHighVersion: highest version we support
     * szDescription: description of implementation 
     * szSystemStatus: relevant status or config info
     * iMaxSockets: maximum number of sockets we provide
     * iMaxUdpDg: largest UDP datagram we support (in bytes)
     * lpVendorInfo: a pointer to vendor-specific data
     */

        //
        //      BUGBUG: The following non-standard NULL parameter was added
        //      do to an early managed C++ compiler bug. 
        //      
        //

        if ( lpWSAData != NULL )
        {
                memset(lpWSAData, 0, sizeof(*lpWSAData));

                lpWSAData->wVersion = MY_VERSION;
                lpWSAData->wHighVersion = MY_VERSION;

                strncpy(
                        lpWSAData->szDescription, 
                        "Microsoft MMLITE WinSock DLL",
                        WSADESCRIPTION_LEN
                        );

                lpWSAData->iMaxSockets = FD_SETSIZE;
                lpWSAData->iMaxUdpDg = 65507;  /* 64KB minus minimal IP & UDP headers */
        }

    return 0;
}

/*
 * User-mandatory finalization function.
 * Release the ref we got.
 */
int WINAPI WSACleanup(void)
{
    if (pSocketBldr) {
        pSocketBldr->v->Release(pSocketBldr);
        pSocketBldr = NULL;
    }
    return 0;
}

/*
 * Support function for FD_ISSET macro.
 */
extern int WINAPI __WSAFDIsSet(SOCKET s, fd_set *f)
{
    int i = f->fd_count;

    while (i--)
        if (f->fd_array[i] == s)
            return 1;

    return 0;
}
