#include <mmlite.h>
#include <stdio.h>
#include <queue.h>

#if defined(_DEBUG)
#define DBG(_x_) printf _x_
#else
#define DBG(_x_)
#endif

/* 
Lock-free queueing code. This code allows adding and removing nodes
from a queue without locking or unlocking the queue. The major
advantages are that the code can be called directly from interrupt
handlers, and there is no impact on the scheduler.

The code is based on the paper:
        Simple, Fast, and Practical Non-Blocking and Blocking 
        Concurrent Queue Algorithms \Lambda 
        Maged M. Michael Michael L. Scott 
        Department of Computer Science 
        University of Rochester 
        Rochester, NY 14627-0226 
        fmichael,scottg@cs.rochester.edu 

Hacks:

1. The algorithm requires that there be at least one node on the list
at all times. If we didn't do things this way, we would need to deal
with the the empty list case, which makes the code much more
complex. The case where there's a single node in the list is
interpreted as the empty queue condition.

The original algorithm managed its own node allocation and only
allowed applications to queue up values which were placed in the
nodes. On a dequeue operation, it just returned the value in the node
after the head and then deallocated the head, leaving the second node
(which had contained the value to be returned) in the queue.  We want
to allow applications to allocate and queue up the nodes themselves,
not just values. To get around this, we initialize the queue to point
to a 'dummy' node.  When dequeuing a node, we check to see if it's the
dummy, and if it is, we just put it back at the tail of the list and
try again.

2. To avoid the ABA problem, the original algorithm kept a version
count on each pointer and required each list to manage its own node
allocation/deallocation so that a node from one list could never end
up in another list. We want to let apps queue up their own nodes, and
would prefer to avoid the extra version count associated with each
pointer (which requires a special CAS2 instruction which isn't
generally available, or using a 16-bit node index and 16-bit count
merged into one 32-bit word)

The most likely problem would be when enqueuing a node by linking it
to the last node in the list, which in the original algorithm would be
pointing to NULL. Suppose, while enqueueing a new node, we get
preempted and the node we were linking to is dequeued and then
re-enqueued on another list. In this case, the CAS of our enqueue
might still succeed (the node we're linking to still points to NULL),
but we would get enqueued on the wrong list! To avoid this, we
represent the NULL pointer as a pointer to the list header itself.
Thus, if a node is dequeued and then enqueued on another list, it's
next pointer will be different, and anyone trying to queue up behind
it on the first list will fail their CAS. Also, if the node is
dequeued and then free'd, it's pretty unlikely that it's next field
will by happenstance be set to point to the list header, so this
covers that case also. Finally, remember that when a node is dequeued,
it's next pointer will never be left pointing to the list header,
since we can never dequeue the last node from the list.  (there must
be at least one node in the list at all times, remember)
*/

typedef struct _MINQUEUENODE
{
    QUEUENODE *pNext;
} MINQUEUENODE;

typedef struct _QUEUE
{
    const struct IQueueVtbl *Vtbl;
    UINT RefCount;
    QUEUENODE * volatile pHead;
    QUEUENODE * volatile pTail;
    MINQUEUENODE Dummy; /* Should be 'QUEUENODE Dummy', but we only need first field */
    PCONDITION pCond;
} QUEUE, *PQUEUE;

#define PQ(_Queue_) ((PQUEUE)_Queue_)

extern const IQueueVtbl QueueVtbl;
SCODE QueueCreate(PCONDITION pCond, IQueue **ppIQueue)
{
    IHeap *pIHeap;
    QUEUE *pNewQueue;

    DBG(("QueueCreate: pCond=0x%x\n",pCond));
    pIHeap = CurrentHeap();
    pNewQueue = (PQUEUE)pIHeap->v->Alloc(pIHeap,0,sizeof(QUEUE),0);
    pNewQueue->Vtbl = &QueueVtbl;
    pNewQueue->RefCount = 1;
    pNewQueue->pCond = pCond;
    pNewQueue->pHead = pNewQueue->pTail = (QUEUENODE *)&pNewQueue->Dummy;
    pNewQueue->Dummy.pNext = (QUEUENODE *)pNewQueue;
    *ppIQueue = (IQueue *)pNewQueue;
    DBG(("QueueCreate: returning IQueue=0x%x\n",pNewQueue));
    return S_OK;
}

SCODE MCT QueueQueryInterface(IQueue *pThis, REFIID Iid, void **ppObject)
{
    return GenericQueryInterface((IUnknown *)pThis,Iid,ppObject,&IID_IQueue);
}

/* Lock a socket structure so that it can't be deleted until we unlock it*/
UINT MCT QueueAddRef(IQueue *pThis)
{
    return AtomicInc(&PQ(pThis)->RefCount);
}

/* Unlock a socket structure & maybe delete*/
UINT MCT QueueRelease(IQueue *pThis)
{
    QUEUE *pQueue;
    int RefCnt;

    pQueue = PQ(pThis);

    RefCnt=AtomicDec(&pQueue->RefCount);
    if (RefCnt == 0)
    {
        PIHEAP pIHeap;
        pIHeap = CurrentHeap();
        pIHeap->v->Free(pIHeap,0,pQueue);
    }

    return RefCnt;
}

#define CAS(_ptr_,_old_,_new_) AtomicCmpAndSwap((PUINT)(_ptr_),(UINT)(_old_),(UINT)(_new_))

void QueuePutNode(QUEUE *pQueue, QUEUENODE *pNode)
{
    QUEUENODE *pTail;
    
    /* Set next ptr of node to point to list header (this is null) */
    pNode->pNext=(QUEUENODE *)pQueue;
    for (;;)                    /* Keep trying until successful */
    {
        pTail = pQueue->pTail;  /* Read tail ptr */
        if (CAS(&pTail->pNext,pQueue,pNode)) /* Try to link in new node */
        {
            /* New node linked in, update tail (maybe not needed?) */
            CAS(&pQueue->pTail,pTail,pNode);
            return;
        }

        /* Try to move tail ptr & try again */
        CAS(&pQueue->pTail,pTail,pTail->pNext);
    }
}

void QueuePut(IQueue *pThis, QUEUENODE *pNode)
{
    QUEUE *pQueue;
    pQueue = PQ(pThis);

    DBG(("QueuePut: IQueue=0x%x, pNode=0x%x\n",pThis,pNode));
    QueuePutNode(pQueue,pNode);
    if (pQueue->pCond)
        Condition_Signal(pQueue->pCond);

    return;
}

BOOL QueuePutFromInterrupt(IQueue *pThis, QUEUENODE *pNode)
{
    QUEUE *pQueue;
    pQueue = PQ(pThis);

    DBG(("QueuePutFromInterrupt: IQueue=0x%x, pNode=0x%x\n",pThis,pNode));
    QueuePutNode(pQueue,pNode);
    if (pQueue->pCond)
        return Condition_InterruptSignal(pQueue->pCond);

    return FALSE;
}

/* Get a node from the head of the queue. 
 * This does double-duty: it's used for both Get and GetFromInterrupt
 */
QUEUENODE *QueueGet(IQueue *pThis)
{
    QUEUENODE *pHead;
    QUEUENODE *pTail;
    QUEUENODE *pNext;
    QUEUE *pQueue;

    pQueue = PQ(pThis);
    DBG(("QueueGet: IQueue=0x%x\n",pThis));

    for (;;)
    {
        pHead = pQueue->pHead;  /* Read head ptr */
        pTail = pQueue->pTail;  /* Read tail ptr */
        pNext = pHead->pNext;   /* Read head->next ptr */
        if (pHead != pQueue->pHead) /* validate pHead */
        {
            continue;           /* invalid - try again */
        }

        if (pHead == pTail)     /* Is queue empty or tail failing behind */
        {
            if (pNext==(QUEUENODE *)pQueue) /* Is queue empty? */
            {
                DBG(("QueueGet: IQueue=0x%x, returning NULL\n"));
                return NULL;    /* Yes, nothing to return */
            }
            /* Tail falling behind, try to advance */
            CAS(&pQueue->pTail,pTail,pNext);
            continue;           /* and try again */
        }

        if (CAS(&pQueue->pHead,pHead,pNext)) /* Try to swing head to next node */
        {                       /* pHead is now ours! */
            if (pHead!=(QUEUENODE *)&pQueue->Dummy) /* Is it a 'real' node? */
            {
                DBG(("QueueGet: IQueue=0x%x, returning pNode=0x%x\n",pThis,pHead));
                return pHead;   /* Yes- return it */
            }
            /* It's the dummy. Queue it back up & try again */
            QueuePutNode(pQueue,pHead);
        }
    }
}

PCONDITION QueueGetConditionVariable(IQueue *pThis)
{
    QUEUE *pQueue;

    pQueue = PQ(pThis);

    DBG(("QueueGetEventHandle: IQueue=0x%x, pCond=0x%x\n",
         pThis,pQueue->pCond));
    return pQueue->pCond;
}

CONST_VTBL struct IQueueVtbl QueueVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    QueueQueryInterface,
    QueueAddRef,
    QueueRelease,
    QueuePut,
    QueuePutFromInterrupt,
    QueueGet,
    QueueGet,
    QueueGetConditionVariable
};
