/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define OPTIMIZE        /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <fred.h>
#include <base/thread.h>
#include <base/schedul.h>
#include <mmmacros.h>

#define _STATISTICS 0

/* MMLite Constraint-based scheduler object (Cbs)
 */
typedef struct _CBS {
    CONST_VTBL struct ISchedulerVtbl *v;
    UINT PreemptionDisableCount;    
    BOOL ContextSwitchNeeded;
#if NEWLIST
    PTHDINFO ReadyQueue;
#else
    PLISTNODE ReadyQueue;
#endif
#if SLEEPLIST
    PTHDINFO SleepQueue;
#else
    PLISTNODE SleepQueue;
#endif
    UINT nThreads;
    TIME IdleTime;
    BOOL FirstThreadBooted;
#if _STATISTICS
    UINT32 nIdle;
#endif
} CBS, *PCBS;

#define pCBS(_x_) ((PCBS)_x_)

#define RescheduleNeeded() Cbs.ContextSwitchNeeded = TRUE
#define RescheduleNotNeeded() Cbs.ContextSwitchNeeded = FALSE

/* Forward decls & macros
 */
#define DBGOUT(_x_,_y_,_z_)
PRIVATE PKCONSTRAINT IThreadAllocConstraint( PTHDINFO pThd );
PRIVATE void IThreadFreeConstraint( PTHDINFO pThd, PKCONSTRAINT pConstr );
PRIVATE void IThreadFreeConstraints(PTHDINFO pThd);
PRIVATE BOOL IThreadCheckFreeConstraint( PTHDINFO pThd );
#if NEWLIST
PRIVATE PTHDINFO MostUrgentThreadOnQueue(PTHDINFO pThdHead);
#else
PRIVATE PTHDINFO MostUrgentThreadOnQueue(PLISTNODE plnThdHead);
#endif
#define MostUrgentReadyThread() MostUrgentThreadOnQueue(Cbs.ReadyQueue)
PRIVATE void IReadyThread(PTHDINFO pThd, BOOL CanReschedule);
PRIVATE TIME TimerInterrupt(void);
PRIVATE PTHDINFO RescheduleInterrupt(PTHDINFO pThdOld, PTHDINFO pThdNew);
PRIVATE PKCONSTRAINT AllocConstraint(void);
PRIVATE void InitializeConstraint(PKCONSTRAINT Constraint, PTHDINFO pThd);
PRIVATE void ThreadQueueMutexWaiter(PTHDINFO pThd, PTHDINFO pThdWaiter, PMUTEX pMutex);
PRIVATE void ThreadWakeBlockedThreads(PTHDINFO pThdOwner, PMUTEX pMutex);
PRIVATE PTHDINFO ThreadQueueCondWaiter(PTHDINFO pThdFirst, PTHDINFO pThdNew, PCONDITION pcd);
PRIVATE void NormalizeTimeout(PTIME Timeout);
PRIVATE void PutThreadOnSleepQueue(PTHDINFO pThd);
PRIVATE SCODE IThreadBeginConstraintBeforeWaitValidate(PTHDINFO pThd,
                      BOOL EndPrevious, PTIME_CONSTRAINT TimeConstraint);
PRIVATE void IThreadBeginConstraintBeforeWait(PTHDINFO pThd, BOOL EndPrevious,
                            PTIME_CONSTRAINT TimeConstraint, PTIME pTimeTaken);
PRIVATE SCODE IThreadBeginConstraintAfterWait(PTHDINFO pThd, SCODE sc);
PRIVATE PTHDINFO ThreadWakeWaitingThread(PTHDINFO pThd, SCODE sc, BOOL fAll);
PRIVATE PTHDINFO PreemptIfNecessary( PTIME pNextTimeSlice );
PRIVATE void IThreadRecalculateInheritance(PTHDINFO pThd);
PRIVATE void IConstraintGetInheritedUrgency(PKCONSTRAINT Constraint,
                                           PKURGENCY Urgency);
PRIVATE void IThreadSleep(PTHDINFO pThd, TIME Time);


/* Scheduler object methods. 
 */
PRIVATE IThread * MCT CbsReschedule( IScheduler *This, IThread *pCurrent);
PRIVATE BOOL MCT CbsDisablePreemption( IScheduler *This, BOOL Recursive);
PRIVATE BOOL MCT CbsEnablePreemption( IScheduler *This, BOOL Recursive);
PRIVATE BOOL MCT CbsPreemptionIsEnabled( IScheduler *This);
PRIVATE SCODE MCT CbsAddThread( IScheduler *This, IThread *Thread,PTIME_CONSTRAINT Constraint);
PRIVATE void MCT CbsRemoveThread( IScheduler *This, IThread *Thread);
PRIVATE void MCT CbsWakeupOnMutex(IScheduler *This,PMUTEX pMutex,IThread *Owner);
PRIVATE void MCT CbsWaitForMutex(IScheduler *This,IThread *Thread,PMUTEX pMutex,IThread *Owner);
PRIVATE PITHREAD MCT CbsWaitForCondition(IScheduler *This,IThread *Thread,
                                 PCONDITION pCondition,IThread *pFirstThread,
                                 PTIME pTimeOut);
PRIVATE PITHREAD MCT CbsWakeWaitingThreads(IScheduler *This,IThread *pFirstThread,
                                   SCODE Result,BOOL fAll);
PRIVATE SCODE MCT CbsConditionTimedWaitAndBeginConstraint(IScheduler *This,
                                                  PCONDITION pCondition,
                                                  PMUTEX pMutex,
                                                  TIME *pTimeout,
                                                  BOOL EndPrevious,
                                                  PTIME_CONSTRAINT pTimeConstraint,
                                                  PTIME pTimeTaken,
                                                  WAITCOND_CALLBACK WaitCallback);
PRIVATE TIME MCT CbsGetIdleTime(IScheduler *This);


PRIVATE CONST_VTBL struct ISchedulerVtbl CbsVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    CbsReschedule,
    CbsAddThread,
    CbsRemoveThread,
    CbsDisablePreemption,
    CbsEnablePreemption,
    CbsPreemptionIsEnabled,
    CbsWaitForMutex,
    CbsWakeupOnMutex,
    CbsWaitForCondition,
    CbsWakeWaitingThreads,
    CbsConditionTimedWaitAndBeginConstraint,
    CbsGetIdleTime
};

/* Scheduler object
 */
PRIVATE CBS Cbs = {
    &CbsVtbl,
    0,
    FALSE,
    NULL,
    NULL,
    0,
};

IScheduler *SchedulerCreate(void)
{
    Cbs.v = &CbsVtbl;
    Cbs.PreemptionDisableCount = 0;
    Cbs.ContextSwitchNeeded = FALSE;
    Cbs.ReadyQueue = NULL;
    Cbs.SleepQueue = NULL;
    Cbs.nThreads = 0;
    Int32ToInt64(Cbs.IdleTime,0);
    Cbs.FirstThreadBooted = FALSE;
#if _STATISTICS
    Cbs.nIdle = 0;
#endif
    return (IScheduler *) &Cbs;
}

/* DEBUG SUPPORT
 */

#ifndef LOGIT
#define LOGIT 0
#endif

#if LOGIT

#define ON_EB63 0
#if ON_EB63 /* optim for EB63 */
#include <arm\at91m63200.h>
#define ThePIT (&Tc0->Channel0)
#define TimeStamp() ThePIT->Counter
#define TIMETYPE UINT pad, 
#define TimeStampSet(_t_,_v_) _t_ = _v_
#else
#define TimeStamp() CurrentTime()
#define TIMETYPE TIME
#define TimeStampSet(_t_,_v_) Int32ToInt64(_t_,_v_)
#endif

#define LOGSIZE 128*4
UINT LogP = 0;
struct {
    UINT Key;
    UINT Info;
    TIMETYPE When;
} LogBuf[LOGSIZE];
char *LogNames[] = {
    "unused",
    "rit Cur=", /* 1 - Resch. Intr. Taken -- CbsReschedule called */
    "rin Cxt=", /* 2 - Resch. Intr. Nottaken -- _Reschedule found locked */
    "rl  Cur=", /* 3 - RescheduleInterrupt main loop */
    "idl Cxt=", /* 4 - IdleThread loop */
    "~id Cxt=", /* 5 - IdleThread loop ended */
    "d   Thd=", /* 6 - Dead thread */
    "s       ", /* 7 - SleepUntil called */
    "out Cur=", /* 8 - Out to previous mode -- CbsReschedule returns */
    "cwi Cur=", /* 9 - Condition_Wait entered */
    "dr  Cur=", /*10 - Calling DrainDeferredConditions */
    "prq que=", /*11 - Put..ReadyQueue called */
    "    que=", /*12 - IdleLoop finds ready queue == .. */
    "cis  st=", /*13 - ConditionInterruptSignal called */
    "lt  cnt=", /*14 - ConditionInterruptSignal while inside CbsReschedule */
    "w       ", /*15 - Out of SleepUntil */
    "orq que=", /*16 - Pull..ReadyQueue called */
    "psq que=", /*17 - Put..SleepQueue called */
    "osq que=", /*18 - Pull..SleepQueue called */
};

void LogIt(UINT Key, UINT Info)
{
    TIMETYPE Now;
    INT i;

    for (;;) {
        /* Disabled ?
         */
        if (LogP == ~0)
            return;
        /* Get a number, make sure its good.
         */
        i = AtomicInc(&LogP) - 1;
        if (i >= LOGSIZE) {
            /* Roundabout, with care
             */
            (void) AtomicCmpAndSwap(&LogP,
                                    i+1,
                                    0);
        } else
            /* Good one.
             */
            break;
    }
    /* Take time now, after we got a slot.
     * Otherwise the printing will show time going backward
     */
    Now = TimeStamp();
    LogBuf[i].When = Now;
    LogBuf[i].Key = Key;
    LogBuf[i].Info  = Info;
#if 0
    printf("LOG %hs %08x %ld\n",
           LogNames[Key],
           LogBuf[i].Info,
           LogBuf[i].When);

#endif
}

void PrintSchedulerLog(void)
{
    UINT index, LogIs;
    UINT key;
    TIMETYPE Now = TimeStamp(), t;

    /* Block scheduler
     */
    Cbs.PreemptionDisableCount++;

    /* Block the log while we print it
     */
    for (;;) {
        LogIs = LogP;
        if (AtomicCmpAndSwap(&LogP,
                             LogIs,
                             ~0))
            break;
    }
    index = LogIs;
    TimeStampSet(t,0);
    do {

        if (++index >= LOGSIZE) index = 0;

        key = LogBuf[index].Key;
        if (key == 0)
            continue;

#if ON_EB63
        printf("%s %08x  %05d (+%d)\n",
               LogNames[key],
               LogBuf[index].Info,
               LogBuf[index].When,
               LogBuf[index].When-t);
#else
        printf("%s %08x  %ld (+%ld)\n",
               LogNames[key],
               LogBuf[index].Info,
               LogBuf[index].When,
               TimeSubtract(LogBuf[index].When,t));
#endif
        t = LogBuf[index].When;

    } while (index != LogIs);

#if ON_EB63
    printf(" now  %d (+%d)\n", Now, Now - t );
#else
    printf(" now  %ld (+%ld)\n", Now, TimeSubtract(Now, t) );
#endif
    /* Let the log go
     */
    LogP = LogIs;

    /* Unlock scheduler
     */
    Cbs.PreemptionDisableCount--;
}
#else
#define LogIt(x,y)
#endif /* LOGIT */

/* Machdep ops&defines
 */
#include "_schedul.c"

/*
 * THE GUTS OF THE THING START HERE
 */
PRIVATE IThread * MCT CbsReschedule( IScheduler *This, IThread *pCurrent)
{
    TIME time;
    UINT StateWas = pTH(pCurrent)->State;
    PITHREAD pThdMost;

    UnusedParameter(This);

    LogIt(1,(UINT) pCurrent);

 Again:
    if (DeferredConditionSignals) {
        LogIt(10,(UINT) pCurrent);
        DrainDeferredConditions();
    }

    /* Timer has expired, check the sleep queue
     */
    time = TimerInterrupt();

    /* Preempt the current thread if necessary due to a ready thread
     * now being more urgent than the current thread. This can happen
     * either because a more urgent thread woke up, or because
     * ConditionSignal changed the relative urgencies of the
     * current and some previously ready thread.
     */
    pThdMost = iTH(PreemptIfNecessary(&time));

    /* Dont preempt the IdleThread, but do set the timer.
     */
    if (SetNextInterrupt(time, &pCBS(This)->ContextSwitchNeeded) &&
        (StateWas != THD_IDLE))
    {
        if (DeferredConditionSignals) {
            LogIt(14, 0);
            /* set this back the way it was */
            pCBS(This)->ContextSwitchNeeded = TRUE;
            goto Again;
        }

        pCurrent = iTH(RescheduleInterrupt(pTH(pCurrent), pTH(pThdMost)));
    }

    LogIt(8,(UINT)pCurrent);
    return pCurrent;
}

/* Disable preemption, return TRUE if preemption was already disabled.
 */
PRIVATE BOOL MCT CbsDisablePreemption( IScheduler *This, BOOL Recursive)
{
    UINT WhatItIsNow;

    WhatItIsNow = AtomicInc( &pCBS(This)->PreemptionDisableCount );

    assert( Recursive || (WhatItIsNow == 1) );
    //assert(WhatItIsNow == 1);  YYY also happens from Condition_Wait/WaitCond/Mutex_Unlock /* XXX only can happen from debug.c:fflush() */
    UnusedParameter(Recursive);

    return WhatItIsNow != 1;
}

/* Enable preemption, return TRUE if preemption is now enabled
 */
PRIVATE BOOL MCT CbsEnablePreemption( IScheduler *This, BOOL Recursive)
{
    UINT WhatItIsNow;

    WhatItIsNow = AtomicDec( &pCBS(This)->PreemptionDisableCount );

    assert( Recursive || (WhatItIsNow == 0) );
    //assert(WhatItIsNow == 0);  YYY also happens from Condition_Wait/WaitCond/Mutex_Unlock /* XXX only can happen from debug.c:fflush() */
    UnusedParameter(Recursive);

    if (WhatItIsNow == 0 &&
        pCBS(This)->ContextSwitchNeeded)
    {
        /* pCBS(This)->ContextSwitchNeeded = FALSE; DONT */
        Reschedule();
    }

    return WhatItIsNow == 0;
}


PRIVATE TIME MCT CbsGetIdleTime(IScheduler *This)
{
    return pCBS(This)->IdleTime;
}

PRIVATE BOOL MCT CbsPreemptionIsEnabled( IScheduler *This)
{
    return pCBS(This)->PreemptionDisableCount == 0;
}


/* Schedule new thread. If tc is set, also do BeginConstraint on it */
PRIVATE SCODE MCT CbsAddThread( IScheduler *This, IThread *Thread,
                               PTIME_CONSTRAINT tc)
{
    PKCONSTRAINT Constraint = &pTH(Thread)->InitialConstraint;
    SCODE sc = S_OK;

    pTH(Thread)->Constraint = Constraint;
    InitializeConstraint(Constraint, pTH(Thread));

    if (tc) {
        if (IThreadCheckFreeConstraint(pTH(Thread)))
            return E_NOT_ENOUGH_MEMORY;
    
        sc = IThreadBeginConstraintBeforeWaitValidate(pTH(Thread), FALSE, tc);
        if (FAILED(sc))
            return sc;
    }

    /* Boot problem. Make sure we reschedule later, not now
     */
    CbsDisablePreemption(This,FALSE);
    pCBS(This)->nThreads++;
    /* Should be THD_READY, but this cures an assertion */
    ThreadSetState(Thread,THD_WAITING);
    if (tc) {
        TIME start;
        IThreadBeginConstraintBeforeWait(pTH(Thread), FALSE, tc, NULL);
        start = pTH(Thread)->PendingConstraint->Start;
        if (TimeLess(CurrentTime(), start)) {
            ThreadSetState(Thread,THD_READY); /* another assertion... */
            IThreadSleep(pTH(Thread), start);
        } else {
            IReadyThread(pTH(Thread),FALSE);
        }
    } else {
        IReadyThread(pTH(Thread),FALSE);
    }

    CbsEnablePreemption(This,FALSE);
    RescheduleNeeded();

    return sc;
}

PRIVATE void MCT CbsRemoveThread( IScheduler *This, IThread *Thread)
{
    UINT IntrState;
    PITHREAD NewThread;
    PCXTINFO pProperContext;

    /* We can only deal with suicide
     * [would just need a bunch more code otherwise]
     */
    assert(CurrentThread() == Thread);

    CbsDisablePreemption(This,FALSE);

    /* If this is the last thread then the game is over
     */
    if (--pCBS(This)->nThreads == 0)
        BaseDelete();

    LogIt(6,(UINT)Thread);

    /* Be nice to know if Thread is holding any mutexes, but how ?
     * We can only tell if someone is already waiting for them.
     * User will have to take care of the rest.
     */
    ThreadWakeBlockedThreads(pTH(Thread), NULL);

    /* Say that thread is going away
     */
    pTH(Thread)->State = THD_DEAD;

    /* For suicide, must switch to some other thread in order
     * to be able to reclaim stack.
     */
    NewThread = iTH(RescheduleInterrupt(pTH(Thread), NULL));
#ifdef not_suicide
    /* Otherwise, among other things, make sure thread wasnt sleeping or what.
     */
    {
        UINT wstate = IThreadGetWaitType(Thread);
        if (wstate == WAIT_CONDITION)
            ConditionTimeoutThread((PCONDITION)pTH(Thread)->WaitObject,
                                           Thread,
                                           ThreadConditionCancel);
        else if (wstate == WAIT_MUTEX) {
            /* asleep, waiting on mutex.. */
        }
    }
#endif

    /* Either way, we are now running in a 'safe' thread context.
     * [possibly on the wrong stack though].
     */

    /* We cannot free memory while holding the scheduler lock.
     * If we stole a thread, make sure it returns where it should.
     * We might get preempted before we are done with this code.
     */
    pProperContext = pTH(NewThread)->CxtInfo;
    CbsEnablePreemption(This,FALSE);

    /* Release the dead thread's resources.
     * This must be guaranteed not to take e.g. the heap mutex,
     * because it might be held by the thread we stole --> deadlock.
     */
    IThreadFreeConstraints(pTH(Thread));
    Thread->v->Release(Thread); /* BUGBUG stack! */

    /* If this was a suicide, special handling.
     */
    TURN_INTERRUPTS_OFF(IntrState);
    pTH(NewThread)->CxtInfo = pProperContext;
    LoadThreadContext(NewThread);
    /* NOTREACHED */
    assert(FALSE);
#ifdef _DEBUG
    DebugBreak();
#endif

    /* Otherwise just get back to caller
     */
}

PRIVATE void MCT CbsWakeupOnMutex(
                IScheduler *This,
                PMUTEX pMutex,
                IThread *Owner)
{
    UnusedParameter(This);
    ThreadWakeBlockedThreads(pTH(Owner),pMutex);
}

PRIVATE void MCT CbsWaitForMutex(
                IScheduler *This,
                IThread *Thread,
                PMUTEX pMutex,
                IThread *Owner)
{
    UnusedParameter(This);
    ThreadQueueMutexWaiter(pTH(Owner),pTH(Thread),pMutex);
}

PRIVATE PITHREAD MCT CbsWaitForCondition(
                    IScheduler *This,
                    IThread *Thread,
                    PCONDITION pCondition,
                    IThread *pFirstThread,
                    PTIME pTimeOut)
{
    PTHDINFO pNewThread;
    UnusedParameter(This);

    LogIt(9, (UINT) Thread);
    pNewThread = ThreadQueueCondWaiter(pTH(pFirstThread), pTH(Thread),
                                       pCondition);

    if (pTimeOut != NULL) { /* Place in sleep queue */
        NormalizeTimeout(pTimeOut);
        pTH(Thread)->StartTime = *pTimeOut;
        PutThreadOnSleepQueue(pTH(Thread));
    }

    return iTH(pNewThread);
}

PRIVATE PITHREAD MCT CbsWakeWaitingThreads(
                    IScheduler *This,
                    IThread *pFirstThread,
                    SCODE Result,
                    BOOL fAll)
{
    UnusedParameter(This);
    return iTH(ThreadWakeWaitingThread(pTH(pFirstThread),Result,fAll));
}

PRIVATE SCODE MCT CbsConditionTimedWaitAndBeginConstraint(
                IScheduler *This,
                PCONDITION pCondition,
                PMUTEX pMutex,
                TIME *pTimeout,
                BOOL EndPrevious,
                PTIME_CONSTRAINT pTimeConstraint,
                PTIME pTimeTaken,
                WAITCOND_CALLBACK WaitCallback)
{
    SCODE sc, sc2;
    PITHREAD pThd;

    pThd = CurrentThread();

    if (IThreadCheckFreeConstraint(pTH(pThd)))
        return E_NOT_ENOUGH_MEMORY;

    CbsDisablePreemption(This,FALSE);

    /* Do pre-wait BeginConstraint validation */
    sc = IThreadBeginConstraintBeforeWaitValidate(pTH(pThd),EndPrevious,
                                                  pTimeConstraint);
    if (sc != S_OK) {
        /* We have not yet released the mutex, so dont retake it
         */
        pMutex = NULL;
        goto Exit;
    }

    (*WaitCallback)(pCondition,pMutex,pTimeout);

    /* Do pre-wait BeginConstraint processing */
    IThreadBeginConstraintBeforeWait(pTH(pThd),EndPrevious,
                                     pTimeConstraint,pTimeTaken);

    /* Now enable pre-emption to make us block. When we get back, redisable */
    CbsEnablePreemption(This,FALSE);
    CbsDisablePreemption(This,FALSE);

    sc = pTH(pThd)->LastError;

    /* Do after-wait BeginConstraint processing */
    sc2 = IThreadBeginConstraintAfterWait(pTH(pThd),S_FALSE);

    /* Return status from cond wait,
     * unless constraint failure is predicted
     */
    assert(sc2 == S_OK || sc2 == S_OUTATIME);
    if (sc2 == S_OUTATIME)
        sc = S_OUTATIME;

Exit:
    CbsEnablePreemption(This,FALSE);

    /* Must return the mutex locked in *all* cases
     */
    if (pMutex != NULL)
        Mutex_Lock(pMutex);

    return sc;
}


/* INNER GUTS OF THE BEAST FROM HERE ON
 */


/* Macros
 */
#define ThreadGetQueue(_t_) &(pTH(_t_)->Queue)
#define ThreadGetSleepQueue(_t_) &(pTH(_t_)->SleepQueue)

#define GetThreadFromQueue(_pln_)  ((_pln_) ? GetThreadFrom(_pln_,Queue) :NULL)

#define GetThreadFromSleepQueue(_pln_)   GetThreadFrom(_pln_,SleepQueue)

#define IThreadGetWaitType(_thd_) (pTH(_thd_)->WaitType)

#define ThreadDied(_pTh_) (pTH(_pTh_)->State == THD_DEAD)


#if 1
/* Lotsa unresolved issues here
 */

PRIVATE INLINE
void StartInheritance(PKCONSTRAINT Waiter, PKCONSTRAINT Owner)
{
    /* Both parameters must be non-null.
     */
    assert(Waiter != NULL);
    assert(Owner != NULL);

    /* The waiter must not already be inheriting to an owner. Thus it must
     * neither point directly to an owner, nor be on a list of constraints
     * inheriting to an owner. (The last condition is represented by having
     * the constraints NextWaiter field point back to the constraint itself).
     */
    assert(Waiter->Owner == NULL);
    assert(Waiter->NextWaiter == Waiter);

    /* A constraint cannot inherit to itself.
     */
    assert(Owner != Waiter);

    /* Set up inheritance by pointing the waiter to the owner...
     */
    Waiter->Owner = Owner;

    /* And putting the waiter on the list of threads inheriting to the owner.
     */
    Waiter->NextWaiter = Owner->FirstWaiter;
    Owner->FirstWaiter = Waiter;

    /* The waiter now appears to be inheriting to an owner.
     */
    assert(Waiter->Owner != NULL);
    assert(Waiter->NextWaiter != Waiter);
}

PRIVATE void IThreadSetStateWaiting(PTHDINFO pThd, PTHDINFO pThdOwner,
                                   INT WaitType, PTR WaitObject, UINT WaitData)
{
    /* A thread about to wait must not already be waiting.
     */
    assert(pTH(pThd)->State != THD_WAITING);

    /* If this wait specifies a resource owner, then point the waiting thread
     * and the owner thread at each other.
     */
    if (pThdOwner != NULL)
        StartInheritance(pTH(pThd)->Constraint, pTH(pThdOwner)->Constraint);

    /* Record the remainder of the wait parameters in the waiting thread.
     */
    pTH(pThd)->WaitType = WaitType;
    pTH(pThd)->WaitObject = WaitObject;
    pTH(pThd)->WaitData = WaitData;
    pTH(pThd)->State = THD_WAITING;
}

#if !NEWLIST
/* Remove this thread from a queue.  Return a pointer to the next thread
 * on the queue, NULL if this thread not on any queue or if queue is empty
 * (this was the only thread in the queue).
 */
PRIVATE PLISTNODE ThreadDequeue(PTHDINFO pThd)
{
    PLISTNODE plnNext;

    plnNext = ListRemove(&pTH(pThd)->Queue);
    pTH(pThd)->QueueType = QUEUE_NONE;
    return plnNext;
}
#endif

#define IThreadGetQueueType(_th_) (pTH(_th_)->QueueType)

#define IThreadSetQueueType(_th_,_qt_) (pTH(_th_)->QueueType = _qt_)

/* Add this thread to the tail of a queue specified by plnThdHead.
 * Assumes thread not already on a queue.
 * Returns new head of queue (which will be the same as the old head,
 * unless the queue was empty, in which case it will point to pThd).
 */
#if !NEWLIST
PRIVATE PLISTNODE ThreadQueueFifo(PTHDINFO pThd, PLISTNODE plnThdHead)
{
    if (plnThdHead != NULL)
        ListInsert(plnThdHead->pPrev,&pTH(pThd)->Queue);
    else
        plnThdHead = &pTH(pThd)->Queue;
    return plnThdHead;
}
#endif

/* Mark thread as ready and insert in ready queue, sorted by constraint */
PRIVATE void PutThreadOnReadyQueue(PTHDINFO pThd)
{
    DBGOUT(DBGCONTEXT,0,printf("PutThreadOnReadyQueue %08x\n", (uint) pThd));
    assert(IThreadGetQueueType(pThd) == QUEUE_NONE);
    assert(!pTheScheduler->v->PreemptionIsEnabled(pTheScheduler));

#if NEWLIST
    ListAddLast(&Cbs.ReadyQueue, TQ_off, pThd);
#else
    Cbs.ReadyQueue = ThreadQueueFifo(pThd,Cbs.ReadyQueue);
#endif
    IThreadSetQueueType(pThd,QUEUE_READY);
    LogIt(11,(UINT)Cbs.ReadyQueue);
}

/* Remove given thread from ready queue */
PRIVATE void PullThreadOffReadyQueue(PTHDINFO pThd)
{
#if NEWLIST
    pTH(pThd)->QueueType = QUEUE_NONE;
    ListRemoveItem(&Cbs.ReadyQueue, TQ_off, pThd);
#else
    PLISTNODE plnThd, plnNewHead;

    DBGOUT(DBGCONTEXT,0,printf("PullThreadOffReadyQueue %08x\n", (uint)pThd));

    plnThd = ThreadGetQueue(pThd);
    plnNewHead = ThreadDequeue(pThd);

    /* If pThd is the only one on queue ListRemove returns NULL, plnReady
     * point to pThd's queue.  If it wasn't on the queue to begin with,
     * plnReady points to something else.
     */
    if (Cbs.ReadyQueue == plnThd)
        Cbs.ReadyQueue = plnNewHead;
#endif

    LogIt(16,(UINT)Cbs.ReadyQueue);
    assert(IThreadGetQueueType(pThd) == QUEUE_NONE);
}

PRIVATE INT ThreadStartTimeCmp(PTR t1, PTR t2)
{
    PTHDINFO pThd1 = t1, pThd2 = t2;
    TIME d = Int64Subtract(pThd1->StartTime, pThd2->StartTime);
    if (Int64IsNegative(d))
        return -1;
    else if (Int64IsZero(d))
        return 0;
    else
        return 1;
}

/* Insert thread in sleep queue, sorted by time.
 * Thread's state must be THD_WAITING before this function is called.
 */
PRIVATE void PutThreadOnSleepQueue(PTHDINFO pThd)
{
#if !SLEEPLIST
    PTHDINFO pThdHead, pThdTail, pThdNode;
    PLISTNODE plnThdHead, plnHead;
#endif

    DBGOUT(DBGCONTEXT,0,printf("PutThreadOnSleepQueue %08x\n", (UINT) pThd));
    assert(ThreadGetState(pThd) == THD_WAITING);
    assert(!pTheScheduler->v->PreemptionIsEnabled(pTheScheduler));

#if SLEEPLIST
    /* XXX Ought to retain the old optimization of testing against the last
       XXX separately */
    ListWalk(&Cbs.SleepQueue, TSQ_off, LIST_WALK_ADD | LIST_WALK_SORTED,
             ThreadStartTimeCmp, pThd);
#else
    plnThdHead = Cbs.SleepQueue;

    if (plnThdHead == NULL) { /* Queue empty */
        assert(pTH(pThd)->SleepQueue.pPrev == &pTH(pThd)->SleepQueue);
        assert(pTH(pThd)->SleepQueue.pNext == &pTH(pThd)->SleepQueue);
        plnHead = &pTH(pThd)->SleepQueue;
        goto Found;
    }

    pThdHead = GetThreadFromSleepQueue(plnThdHead);
    pThdTail = GetThreadFromSleepQueue(pTH(pThdHead)->SleepQueue.pPrev);

    /* If not earlier than latest on the list, then put it at the tail */
    if (!TimeLess(pTH(pThd)->StartTime, pTH(pThdTail)->StartTime)) {
        ListInsert(&pTH(pThdTail)->SleepQueue, &pTH(pThd)->SleepQueue);
        plnHead = &pTH(pThdHead)->SleepQueue;
        goto Found;
    }

    /* If earlier than earliest on the list, then put it at the head */
    if (TimeLess(pTH(pThd)->StartTime, pTH(pThdHead)->StartTime)) {
        ListInsert(&pTH(pThdTail)->SleepQueue,&pTH(pThd)->SleepQueue);
        plnHead = &pTH(pThd)->SleepQueue;
        goto Found;
    }

    /* Head and tail must be different, else one of the above would be true */
    assert(pThdHead != pThdTail);

    /* Otherwise place before first thread with later start time */
    pThdNode = GetThreadFromSleepQueue(pTH(pThdHead)->SleepQueue.pNext);
    for (;;) {
        assert(pThdNode != pThdHead);
        if (TimeLess(pTH(pThd)->StartTime, pTH(pThdNode)->StartTime)) {
            ListInsert(pTH(pThdNode)->SleepQueue.pPrev,&pTH(pThd)->SleepQueue);
            break;
        }

        pThdNode = GetThreadFromSleepQueue(pTH(pThdNode)->SleepQueue.pNext);
    }

    plnHead = &pTH(pThdHead)->SleepQueue;
 Found:

    Cbs.SleepQueue = plnHead;
#endif
    LogIt(17,(UINT)Cbs.SleepQueue);
}

/* Find and remove thread from sleep queue, do nothing if not on sleep queue.
 */
PRIVATE void PullThreadOffSleepQueue(PTHDINFO pThd)
{
#if !SLEEPLIST
    PLISTNODE plnThd, plnNewHead;
#endif
    DBGOUT(DBGCONTEXT,0,printf("PullThreadOffSleepQueue %08x\n", (UINT)pThd));
    assert(!pTheScheduler->v->PreemptionIsEnabled(pTheScheduler));

#if SLEEPLIST
    ListRemoveItem(&Cbs.SleepQueue, TSQ_off, pThd);
#else
    plnThd = ThreadGetSleepQueue(pThd);
    plnNewHead = ListRemove(plnThd);

    /* If pThd is the only one on sleep queue ListRemove returns NULL, plnSleep
     * point to pThd's sleep queue.  If it wasn't on the sleep queue to begin
     * with, plnSleep points to something else.
     */
    if (Cbs.SleepQueue == plnThd)
        Cbs.SleepQueue = plnNewHead;
#endif
    LogIt(18,(UINT)Cbs.SleepQueue);
}

PUBLIC TIME SleepOverhead = Int64Initializer(0, 0);

PRIVATE void NormalizeTimeout(PTIME Timeout)
{
    TIME CurTime;

    CurTime = CurrentTime();

    /* if Timeout is relative to current time, make it absolute.
     * This guarantees a greater-or-equal time than now.
     */
    if (TimeIsNotPositive(*Timeout))
    {
        INT64SUB(*Timeout, CurTime, *Timeout);

    /* if the Timeout time has already occurred, use current time
     */
    }
    else if (TimeLess(*Timeout,CurTime))
    {
        *Timeout = CurTime;
    }
    /* Account for overheads, subtracting them from *Timeout
     */
    INT64SUB(*Timeout, *Timeout, SleepOverhead);
}

PRIVATE void IThreadSleep(PTHDINFO pThd, TIME Time)
{
    assert(!pTheScheduler->v->PreemptionIsEnabled(pTheScheduler));

    NormalizeTimeout(&Time);
    pTH(pThd)->StartTime = Time;
    IThreadSetStateWaiting(pThd,NULL,WAIT_TIMER,NULL,0);
    PutThreadOnSleepQueue(pThd);
    RescheduleNeeded();
}



#ifdef _MSC_VER
#pragma warning(disable:4127)  /* conditional expression is constant */
#pragma warning(disable:4702)  /* unreachable code */
#endif

/* end unresolved issues */
#endif

#if defined(__NO_BUILTIN_INT64)
const TIME TIME_SHARING_TIMESLICE = Int64Initializer(0,400000/*Low*/);
#else
#define TIME_SHARING_TIMESLICE   400000 /* 40 millisecs */
#endif

#if SLEEPLIST
/* Compare two urgencies (like strcmp).
 * 0 = equal, -1 = u1 is more urgent, 1 = u is.
 */
PRIVATE INT UrgencyCmp(void *_u1, void *_u2)
{
    PKURGENCY u1 = _u1, u2 = _u2;
    INT c1, c2;

    assert(!TimeLess(u1->RestartBy,TIME_ORIGIN) );
    assert(!TimeLess(u2->RestartBy,TIME_ORIGIN) );

    /* Note: time-share implies noncritical */
    c1 = u1->Criticality;
    c2 = u2->Criticality;
    if (c1 == c2) {
        TIME d = Int64Subtract(u1->RestartBy,u2->RestartBy);
        if (Int64IsNegative(d))
            return -1;
        else if (Int64IsZero(d))
            return 0;
        else
            return 1;
    }                            

    if (c2 & TIME_SHARING)
        return -1;

    return ((c1 & CRITICALITY_MASK) - c2);
}
#endif

/* Return true if urgencies are equal.
 */
PRIVATE BOOL IUrgencyEQ(PKURGENCY u1, PKURGENCY u2)
{
    assert(!TimeLess(u1->RestartBy,TIME_ORIGIN) );
    assert(!TimeLess(u2->RestartBy,TIME_ORIGIN) );

    if (u1->Criticality != u2->Criticality)
        return FALSE;

    return TimeEqual(u1->RestartBy,u2->RestartBy);
}

/* Return true if u1 is more urgent than u2.
 */
PRIVATE BOOL IUrgencyGT(PKURGENCY u1, PKURGENCY u2)
{
    INT c1, c2;

    assert(!TimeLess(u1->RestartBy,TIME_ORIGIN) );
    assert(!TimeLess(u2->RestartBy,TIME_ORIGIN) );

    /* Note: time-share implies noncritical */
    c1 = u1->Criticality;
    c2 = u2->Criticality;
    if (c1 == c2)
        return TimeLess(u1->RestartBy,u2->RestartBy);

    if (c2 & TIME_SHARING)
        return TRUE;

    return ((c1 & CRITICALITY_MASK) > c2);
}

/* Return true if u1 is no less urgent than u2.
 */
PRIVATE BOOL IUrgencyGTEQ(PKURGENCY u1, PKURGENCY u2)
{
    INT c1, c2;

    assert(!TimeLess(u1->RestartBy,TIME_ORIGIN) );
    assert(!TimeLess(u2->RestartBy,TIME_ORIGIN) );

    /* Note: time-share implies noncritical */
    c1 = u1->Criticality;
    c2 = u2->Criticality;
    if (c1 == c2)
        return !TimeLess(u2->RestartBy,u1->RestartBy);

    /* If they are not the same criticality they cant be EQ,
     * ==> same as above
     */
    if (c2 & TIME_SHARING)
        return TRUE;

    return ((c1 & CRITICALITY_MASK) > c2);
}

#if 0/*notused*/
/* Return true if thread t1 is exactly as urgent as thread t2.
 */
PRIVATE BOOL IThreadUrgencyEQ(PITHREAD t1, PITHREAD t2)
{
    KURGENCY u1, u2;

    assert(!ThreadDied(t1));
    assert(!ThreadDied(t2));

    IConstraintGetInheritedUrgency(pTH(t1)->Constraint, &u1);
    IConstraintGetInheritedUrgency(pTH(t2)->Constraint, &u2);

    return IUrgencyEQ(&u1, &u2);
}
#endif

/* Return true if thread t1 is more urgent than thread t2.
 */
PRIVATE BOOL IThreadUrgencyGT(PTHDINFO t1, PTHDINFO t2)
{
    KURGENCY u1, u2;

    assert(!ThreadDied(t1));
    assert(!ThreadDied(t2));

    IConstraintGetInheritedUrgency(pTH(t1)->Constraint, &u1);
    IConstraintGetInheritedUrgency(pTH(t2)->Constraint, &u2);

    return IUrgencyGT(&u1, &u2);
}

#if SLEEPLIST
PRIVATE INT ThreadUrgencyCmp(void *_t1, void *_t2)
{
    PITHREAD t1 = _t1, t2 = _t2;
    KURGENCY u1, u2;

    if (t2)
        return -1;

    assert(!ThreadDied(t1));
    assert(!ThreadDied(t2));

    IConstraintGetInheritedUrgency(pTH(t1)->Constraint, &u1);
    IConstraintGetInheritedUrgency(pTH(t2)->Constraint, &u2);

    return UrgencyCmp(&u1, &u2);
}
#endif

/* Return true if thread t1 is no less urgent than thread t2.
 */
PRIVATE BOOL IThreadUrgencyGTEQ(PTHDINFO t1, PTHDINFO t2, PTIME pTimeDiff)
{
    KURGENCY u1, u2;
    BOOL RetVal;

    assert(!ThreadDied(t1));
    assert(!ThreadDied(t2));

    IConstraintGetInheritedUrgency(pTH(t1)->Constraint, &u1);
    IConstraintGetInheritedUrgency(pTH(t2)->Constraint, &u2);

    RetVal = IUrgencyGTEQ(&u1, &u2);

    /* See how long can T1 run
     */
    if (RetVal && pTimeDiff) {
        TIME kt = CurrentTime();
        INT64SUB(kt,u2.RestartBy,kt);
        if (!TimeIsRelative(kt) && TimeLess(kt,*pTimeDiff))
            *pTimeDiff = kt;
    }

    return RetVal;
}

PRIVATE void InitializeConstraint(PKCONSTRAINT Constraint, PTHDINFO pThd)
{
    TIME kt = CurrentTime();


    assert(Constraint);
    assert(pThd);

    Constraint->TimeOfLastRefresh = kt;

#if 0
    if (pcval) {
        Constraint->LocalUrgency.Criticality = pcval->Criticality;;
        if (TimeIsRelative(pcval->Deadline)) {
            /* Convert from relative to absolute kernel time. */
            INT64SUB(pcval->Deadline,kt,pcval->Deadline);
            /* If we have overflowed, use TIME_FOREVER. */
            if (TimeLess(pcval->Deadline,kt))
                pcval->Deadline = TIME_FOREVER;
        }
        /* Convert deadline into restartby time (deadline - est runtime). */
        INT64SUB(Constraint->LocalUrgency.RestartBy,
                 pcval->Deadline, pcval->Estimate);
        if (TimeLess(Constraint->LocalUrgency.RestartBy,TIME_ORIGIN))
            Constraint->LocalUrgency.RestartBy = TIME_ORIGIN;

        /* Save the initial restartby value. */
        Constraint->InitialRestartBy = Constraint->LocalUrgency.RestartBy;
        /* Copy remaining constraint values into current constraint struct. */
        Constraint->LocalUrgency.Criticality = pcval->Criticality;

    } else {
        Constraint->LocalUrgency.Criticality = _NONCRITICAL | TIME_SHARING;
        Constraint->InitialRestartBy = kt; /* shouldnt matter */
    }
#else
    Constraint->LocalUrgency.Criticality = _NONCRITICAL | TIME_SHARING;
    Constraint->InitialRestartBy = kt; /* shouldnt matter */
#endif
    Constraint->LocalUrgency.RestartBy = kt;
    Constraint->SurroundingConstraint = NULL;
    Constraint->Thread = iTH(pThd);
    Constraint->Owner = NULL;
    Constraint->FirstWaiter = NULL;
    Constraint->NextWaiter = Constraint; /* Means: it is on any list */
/*  Constraint->CachedInheritedUrgency; */
    Constraint->CachedInheritedUrgencyValid = FALSE;
}

PRIVATE void UpdateRestartBy(PKCONSTRAINT Constraint)
{
    TIME TimeSinceLastRefresh, Delta, kt;
    /*BOOL PreemptionWasDisabled;*/

    /* Sanity check */
    if (Constraint == NULL)
        return;

    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);
    /*PreemptionWasDisabled = CbsDisablePreemption(pTheScheduler,TRUE);*/
    kt = CurrentTime();

    /* It is only valid to refresh the RestartBy of the current thread.
     * (The RestartBy of a non-running thread remains constant.)
     */
    assert(Constraint->Thread == CurrentThread());

    /* Calculate how long we have been running since this thread's
     * RestartBy was last refreshed.
     */
    //assert(!TimeLess(kt,Constraint->TimeOfLastRefresh) );
    INT64SUB(TimeSinceLastRefresh,kt, Constraint->TimeOfLastRefresh);
    //assert(!TimeLess(TimeSinceLastRefresh,TIME_ORIGIN) );

    /* (If no time has passed, then there is nothing to do.)
     */
    if (TimeEqual(TimeSinceLastRefresh,TIME_ORIGIN))
        goto Exit;

    /* For time-sharing threads, stretch the run time by the
     * number of existing threads.
     */
    if (Constraint->LocalUrgency.Criticality & TIME_SHARING) {
        INT t32;
#if 0
        /* There should be NO WAY that we run a thread without
         * a timer interrupt for so long as to need more than
         * 32 bits to compute the following.
         * Therefore I am leaving this fix out, so that noone else
         * misunderstands and fixes the wrong bug.
         */
        if (!TimeNeeds32Bits(TimeSinceLastRefresh)) {
            INT t;
            t32 = Int64ToInt32(TimeSinceLastRefresh);
            t = t32 * Cbs.nThreads;
            if ((t32 < 0) || (t < t32)) /* overflow */
                goto DoIt64;
            Int32ToInt64(Delta,t);
        } else {
        DoIt64:
            Delta = Int64TimesInt32 (TimeSinceLastRefresh,
                                     Cbs.nThreads);
        }
#else
        t32 = Int64ToInt32(TimeSinceLastRefresh);
        t32 = t32 * Cbs.nThreads;
        Int32ToInt64(Delta,t32);
#endif
    } else
        Delta = TimeSinceLastRefresh;

    /* Delta is not smaller than TimeSinceLastRefresh.
     * And since TimeSinceLastRefresh is nonnegative, so is Delta.
     */
#if !defined(SIM) && 0
    assert(!TimeLess(Delta,TimeSinceLastRefresh) );
    assert(!TimeLess(Delta,TIME_ORIGIN) );
#endif

    /* Add Delta and correct for any overflow.
     *
     * The following assertions imply that the sum will lie
     * between 0 and 2*TIME_FOREVER, inclusive, not counting for
     * round-off, or between [0..TIME_FOREVER] or [-TIME_FOREVER-1, -2],
     * counting for roundoff. Thus any negative value will indicate
     * that an overflow has occurred, and any nonnegative value will
     * indicate that an overflow has not occurred.
     */
    assert(!TimeLess(Constraint->LocalUrgency.RestartBy,TIME_ORIGIN) );

    INT64ADD(Constraint->LocalUrgency.RestartBy,Constraint->LocalUrgency.RestartBy,Delta);

    /* As discussed above, negative value implies overflow.
     * Represent overflow with TIME_FOREVER.
     */
    if (TimeLess(Constraint->LocalUrgency.RestartBy,TIME_ORIGIN))
        Constraint->LocalUrgency.RestartBy = TIME_FOREVER;

    /* And so the result is a valid absolute time.
     */
    assert(!TimeLess(Constraint->LocalUrgency.RestartBy,TIME_ORIGIN) );

    /* Update TimeOfLastRefresh.
     */
    Constraint->TimeOfLastRefresh = kt;
Exit:
    /*CbsEnablePreemption(pTheScheduler,PreemptionWasDisabled)*/
    ;
}

PRIVATE void IThreadResolveConstraint(PTHDINFO pThd, BOOL IsRunning)
{
    PKCONSTRAINT Constraint = pTH(pThd)->Constraint;
    PKCONSTRAINT NewConstraint;
    PKPENDING_CONSTRAINT Pending;

    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);
    assert(Constraint);

    /* There must be a pending constraint */
    Pending = pTH(pThd)->PendingConstraint;
    assert(Pending);

    /* There must be a free constraint structure. */
    NewConstraint = IThreadAllocConstraint(pThd);
    assert(NewConstraint);
    InitializeConstraint(NewConstraint, pThd);

    /* Make sure RestartBy is up-to-date */
    if (IsRunning)
        UpdateRestartBy(Constraint);

    /* Either replace the current constraint, or push a new constraint. */
    if (Pending->EndPrevious) {

        assert(Constraint);
        assert(Constraint->SurroundingConstraint);

        /* We are replacing the current constraint on the constraint stack. */
        NewConstraint->SurroundingConstraint =
            Constraint->SurroundingConstraint;

        /* Record the time taken during the current constraint if desired. */
        if (Pending->TimeTaken) {
            INT64SUB(*Pending->TimeTaken,Constraint->LocalUrgency.RestartBy,
                                   Constraint->InitialRestartBy);
        }

        /* Free the current constraint. */
        IThreadFreeConstraint(pThd, Constraint);

    } else {

        /* We are pushing a new constraint on top of the current one. */
        NewConstraint->SurroundingConstraint = Constraint;
    }

    /* The new constraint must have a surrounding constraint. */
    assert(NewConstraint->SurroundingConstraint);

    /* Convert the deadline to be absolute in kernel time. */
    if (TimeIsRelative(Pending->Deadline)) {

        TIME kt = CurrentTime();

        /*
         * Convert from relative to absolute kernel time.
         */
        INT64SUB(Pending->Deadline,kt,Pending->Deadline);

        /* If we have overflowed, use TIME_FOREVER.
         */
        if (TimeLess(Pending->Deadline,kt))
            Pending->Deadline = TIME_FOREVER;

    }

    /* The deadline must now be absolute (non-negative) */
    assert(!TimeLess(Pending->Deadline,TIME_ORIGIN) );

    /* Convert deadline into restartby time (deadline - estimated runtime). */
    INT64SUB(NewConstraint->LocalUrgency.RestartBy,
        Pending->Deadline, Pending->Estimate);
    if (TimeLess(NewConstraint->LocalUrgency.RestartBy,TIME_ORIGIN))
        NewConstraint->LocalUrgency.RestartBy = TIME_ORIGIN;

    assert(!TimeLess(NewConstraint->LocalUrgency.RestartBy,TIME_ORIGIN) );

    /* Save the initial restartby value. */
    NewConstraint->InitialRestartBy = NewConstraint->LocalUrgency.RestartBy;

    /* Copy remaining constraint values into current constraint structure. */
    NewConstraint->LocalUrgency.Criticality = Pending->Criticality;

    /* Mark thread as no longer using its pending constraint. */
    assert(pTH(pThd)->PendingConstraint == &pTH(pThd)->PConst);
    pTH(pThd)->PendingConstraint = NULL;

    /* Recalculate inheritance, recursively if necessary. */
    pTH(pThd)->Constraint = NewConstraint;
    IThreadRecalculateInheritance(pThd);

    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);
}

PRIVATE SCODE IThreadBeginConstraintBeforeWaitValidate(PTHDINFO pThd,
                      BOOL EndPrevious, PTIME_CONSTRAINT TimeConstraint)
{
    /* We must have a free constraint structure at this point. */
    assert(pTH(pThd)->FreeConstraint != NULL);

    /* Cannot end a constraint if it does not have a surrounding constraint. */
    assert(pTH(pThd)->Constraint);
    if (EndPrevious && pTH(pThd)->Constraint->SurroundingConstraint == NULL)
        return E_INVALID_PARAMETER;

    /* The only valid criticalities are 0 and 1 */
    if (TimeConstraint->Criticality & ~CRITICALITY_MASK)
        return E_INVALID_PARAMETER;

    /* The estimated time must be nonnegative */
    if (TimeIsRelative(TimeConstraint->Estimate))
        return E_INVALID_PARAMETER;

    return S_OK;
}

PRIVATE void IThreadBeginConstraintBeforeWait(PTHDINFO pThd,
           BOOL EndPrevious, PTIME_CONSTRAINT TimeConstraint, PTIME pTimeTaken)
{
    PKPENDING_CONSTRAINT PendingConstraint;

    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);

    /* The thread's one PendingConstraint must not already be in use.
     */
    assert(pTH(pThd)->PendingConstraint == NULL);
    PendingConstraint = &pTH(pThd)->PConst;

    /* Make StartTime absolute if it is relative.
     */
    PendingConstraint->Start = TimeConstraint->Start;

    if (TimeIsRelative(PendingConstraint->Start)) {

        TIME ct = CurrentTime();

        INT64SUB(PendingConstraint->Start,ct,PendingConstraint->Start);

        /* Use TIME_ORIGIN upon underflow.
         */
        if (TimeLess(PendingConstraint->Start,ct))
            PendingConstraint->Start = TIME_ORIGIN;
    }

    /* Save parameters in PendingConstraint structure.
     */
    PendingConstraint->TimeTaken = pTimeTaken;
    PendingConstraint->EndPrevious = EndPrevious;
    PendingConstraint->Criticality = TimeConstraint->Criticality;
    PendingConstraint->Estimate = TimeConstraint->Estimate;
    PendingConstraint->Deadline = TimeConstraint->Deadline;

    /* Mark constraint in use now.
     */
    pTH(pThd)->PendingConstraint = PendingConstraint;
}

PRIVATE SCODE IThreadBeginConstraintAfterWait(PTHDINFO pThd, SCODE sc)
{
    PKCONSTRAINT Constraint;

    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);
    assert(ThreadGetState(pThd) == THD_RUNNING);
    assert(iTH(pThd) == CurrentThread());


    /* If the thread still has a pending constraint and the start time
     * has not already occurred, then wait until it has
     */
    if (sc == S_OK && pTH(pThd)->PendingConstraint) {
        TIME t1, t2;
        t1 = CurrentTime();
        t2 = pTH(pThd)->PendingConstraint->Start;
        if (TimeLess(t1,t2)) {

            /* Note that the call below returns immediately since the kernel
             * is locked.  However this thread will be preempted
             * as soon as the kernel is unlocked.
             */
            IThreadSleep(pTH(CurrentThread()), t2);
            return S_FALSE;
        }
    }

    /* The thread must have either a surrounded constraint or a
     * pending constraint.
     */
    assert(pTH(pThd)->Constraint);
    assert(pTH(pThd)->Constraint->SurroundingConstraint ||
           pTH(pThd)->PendingConstraint);

    /* If the thread still has a pending constraint, then resolve it.
     * This only happens if we never waited or slept
     */
    if (pTH(pThd)->PendingConstraint)
        IThreadResolveConstraint(pThd,TRUE);

    /* The thread is running with a constraint (not pending).
     */
    assert(ThreadGetState(pThd) == THD_RUNNING);
    assert(pTH(pThd)->PendingConstraint == NULL);
    Constraint = pTH(pThd)->Constraint;
    assert(Constraint);
    assert(!TimeEqual(Constraint->TimeOfLastRefresh,TIME_ORIGIN));
    assert(Constraint->SurroundingConstraint);

    /* If our restart by time has passed, then return probable failure */
    UpdateRestartBy(Constraint);
    {
        TIME Now = CurrentTime();

        if (TimeLess(Constraint->LocalUrgency.RestartBy,Now))
            return S_OUTATIME;
    }

    return S_OK;
}

PRIVATE SCODE PostBeginConstraint(SCODE sc)
{
    PITHREAD pThd;

    pThd = CurrentThread();

    CbsDisablePreemption(pTheScheduler,FALSE);

    sc = IThreadBeginConstraintAfterWait(pTH(pThd), sc);

    CbsEnablePreemption(pTheScheduler,FALSE);

    return sc;
}


SCODE RTLCALLTYPE EndConstraint(PTIME pTimeTaken)
{
    PITHREAD pThd;
    PKCONSTRAINT Constraint;
    SCODE sc;

    pThd = CurrentThread();

    CbsDisablePreemption(pTheScheduler,FALSE);

    /* The current constraint must have a surrounding constraint. */
    Constraint = pTH(pThd)->Constraint;
    assert(Constraint);
    if (Constraint->SurroundingConstraint == NULL) {
        sc = E_ACCESS_DENIED;
        goto Exit;
    }

    /* The thread must be running. */
    assert(ThreadGetState(pThd) == THD_RUNNING);
    assert(!TimeEqual(Constraint->TimeOfLastRefresh,TIME_ORIGIN));
    UpdateRestartBy(Constraint);

    /* Return time taken if so desired. */
    if (pTimeTaken != NULL) {
        /* RestartBy time advances when the thread runs, so use difference */
        INT64SUB(*pTimeTaken,Constraint->LocalUrgency.RestartBy,
                       pTH(pThd)->Constraint->InitialRestartBy);
    }

    /* Pop constraint off of constraint stack. */
    pTH(pThd)->Constraint = Constraint->SurroundingConstraint;

    /* Recalculate thread urgency */
    IThreadRecalculateInheritance(pTH(pThd));

    /* Free ended constraint, and return status */

    IThreadFreeConstraint(pTH(pThd), Constraint);
    sc = S_OK;

Exit:
    CbsEnablePreemption(pTheScheduler,FALSE);
    return sc;
}


SCODE 
RTLCALLTYPE 
BeginConstraint(
        BOOL EndPrevious,
        PTIME_CONSTRAINT TimeConstraint,
        PTIME pTimeTaken
        )
{
    PITHREAD pThd;
    SCODE sc;

    pThd = CurrentThread();

    if (IThreadCheckFreeConstraint(pTH(pThd)))
        return E_NOT_ENOUGH_MEMORY;

    CbsDisablePreemption(pTheScheduler,FALSE);

    /* Do pre-wait processing. */
    sc = IThreadBeginConstraintBeforeWaitValidate(pTH(pThd), EndPrevious,
                                                  TimeConstraint);
    if (sc == S_OK)
        IThreadBeginConstraintBeforeWait(pTH(pThd),
                                         EndPrevious, TimeConstraint,
                                         pTimeTaken);

    CbsEnablePreemption(pTheScheduler,FALSE);

    if (sc == S_OK)
        sc = PostBeginConstraint(sc);
    if (sc == S_FALSE)
        sc = PostBeginConstraint(sc);

    return sc;

}

/* Returns most urgent thread on queue (without removing it from queue).
 */
#if NEWLIST
PRIVATE PTHDINFO MostUrgentThreadOnQueue(PTHDINFO pThdHead)
{
    //return ListWalk(pHead, TQ_off, LIST_WALK_LARGEST, ThreadUrgencyCmp, NULL);
    PTHDINFO pThd, pThdMost;

    if (pThdHead == NULL)
        return NULL;

    pThdMost = pThd = pThdHead;
    for (;;) {
        pThd = ListNodeNext(TQ_off, pThd);
        if (pThd == pThdHead)
            break;
        if (IThreadUrgencyGT(pThd, pThdMost)) {
            pThdMost = pThd;
        }
    }

    return pThdMost;
}
#else
PRIVATE PTHDINFO MostUrgentThreadOnQueue(PLISTNODE plnThdHead)
{
    PTHDINFO pThd, pThdHead, pThdMost;

    if (plnThdHead == NULL)
        return NULL;

    pThdHead = pThdMost = pThd = GetThreadFromQueue(plnThdHead);
    for (;;) {
        pThd = GetThreadFromQueue(pTH(pThd)->Queue.pNext);
        if (pThd == pThdHead)
            break;
        if (IThreadUrgencyGT(pThd, pThdMost)) {
            pThdMost = pThd;
        }
    }

    return pThdMost;
}
#endif

#if 0 /* compiler says unused */
/* Removes and returns most urgent thread from queue.
 */
PRIVATE PITHREAD PullMostUrgentThreadOffQueue(PLISTNODE *pplnThdHead)
{
    PITHREAD pThd;
    PLISTNODE plnThd, plnNewHead;

    pThd = MostUrgentThreadOnQueue(*pplnThdHead);
    plnThd = ThreadGetQueue(pThd);
    plnNewHead = ThreadDequeue(pThd);

    /* If pThd is the only one on queue ListRemove returns NULL, *pplnThdHead
     * point to pThd's queue.  If it wasn't on the queue to begin with,
     * *pplnThdHead points to something else.
     */
    if (*pplnThdHead == plnThd)
        *pplnThdHead = plnNewHead;

    assert(IThreadGetQueueType(pThd) == QUEUE_NONE);

    return pThd;
}
#endif

/* Remove next runnable thread from its queue.
 * This rescheduling path is 'synchronous', e.g.
 * we get here IFF RescheduleNeeded() is called.
 */
PRIVATE PTHDINFO RescheduleInterrupt(PTHDINFO pThdOld, PTHDINFO pThdNew)
{
    BOOL RuntimeWasUpdated = FALSE;
ReEval:
    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);
    LogIt(3,(UINT)pThdOld);

    if (pThdNew == NULL)
        pThdNew = MostUrgentReadyThread();

    /* The old thread exists, might be dieing or could be running.
     */
    assert(pThdOld);

    /* Do we have anyone to switch to ?
     * This includes ourselves, since we have no idle thread.
     */
    if (pThdNew == NULL) {

        PCXTINFO pSavedCxt;
        UINT StateWas;

        /* If we did not have to stop then dont.
         */
        if (ThreadGetState(pThdOld) == THD_RUNNING)
            return pThdOld;

/*printf("{nil}\n");*/
        /* About to play IdleThread().
         * Make sure this idle time is charged properly.
         * Compute run time now, but claim we did it later.
         */
        if (!RuntimeWasUpdated) {
            UpdateRestartBy(pTH(pThdOld)->Constraint);
            RuntimeWasUpdated = TRUE;
        }

        /* We are the last runnable thread.
         * Two things can change this: sleep time expires for the
         * head of the sleep queue, or an interrupt routine signals
         * and wakes up a thread.
         * We have already taken care of the first by setting the
         * interval timer.
         * For the second, we just need to release the scheduler lock.
         * One catch tho: if/when we get preempted we do not really
         * want to overwrite our saved-state and be stuck here.
         */
        StateWas = pTH(pThdOld)->State;
        pTH(pThdOld)->State = THD_IDLE;
        pSavedCxt = pTH(pThdOld)->CxtInfo;
        LogIt(4,(UINT)pSavedCxt);

        /* Notice that we wont be preempted (see CbsReschedule && THD_IDLE)
         */
        CbsEnablePreemption(pTheScheduler, FALSE);

        /* Spin while the ready queue is empty.
         */
#if NEWLIST
        while ((*((volatile PTHDINFO *)&Cbs.ReadyQueue)) == NULL) {
#else
        while ((*((volatile PLISTNODE*)&Cbs.ReadyQueue)) == NULL) {
#endif
#ifdef MACHINE_IDLE
            MACHINE_IDLE(Cbs.ReadyQueue);
#endif
            LogIt(12,(UINT)Cbs.ReadyQueue);
        }

        CbsDisablePreemption(pTheScheduler, FALSE);

        LogIt(5,(UINT)pSavedCxt);
        pTH(pThdOld)->CxtInfo = pSavedCxt;
        if (pTH(pThdOld)->State == THD_IDLE)
           pTH(pThdOld)->State = StateWas;

        /* Idle time, second part
         */
        {
            TIME t, Now = CurrentTime();
            INT64SUB(t,Now,pTH(pThdOld)->Constraint->TimeOfLastRefresh);
            INT64ADD(Cbs.IdleTime,Cbs.IdleTime,t);
#if _STATISTICS
            Cbs.nIdle++;
#endif
                
            pTH(pThdOld)->Constraint->TimeOfLastRefresh = Now;
        }
        goto ReEval;
    }

    /* The new thread exists, has not died, and is ready to run.
     */
    assert(pThdNew);
    assert(!ThreadDied(pThdNew));
    assert(ThreadGetState(pThdNew) == THD_READY);

    /* Pull the new thread off the ready queue and mark it running.
     */
    PullThreadOffReadyQueue(pThdNew);
    ThreadSetState(pThdNew,THD_RUNNING);

    /* If the old and new threads are different, then do the work
     * that must be done when switching threads.
     */
    if (pThdOld != pThdNew) {

        /* Update the RestartBy of the old thread.
         */
        if (!RuntimeWasUpdated)
            UpdateRestartBy(pTH(pThdOld)->Constraint);

        /* Then put it back on the ready queue
         */
        if (ThreadGetState(pThdOld) == THD_RUNNING) {
            ThreadSetState(pThdOld, THD_READY);
            PutThreadOnReadyQueue(pThdOld);
        }

        /* Save current time for future UpdateRestartBy calls on new thread.
         */
        pTH(pThdNew)->Constraint->TimeOfLastRefresh =
            pTH(pThdOld)->Constraint->TimeOfLastRefresh;

        /* Make the new thread be the current thread, switching context
         * (user regs, TLB, floating point state, etc.) as necessary.
         */
        ThreadSetCurrent(iTH(pThdNew));
    }

    return pThdNew;
}

/* Reschedule if less urgent than other threads
 */
PRIVATE PTHDINFO PreemptIfNecessary( PTIME pNextTimeSlice )
{
    PTHDINFO pThdCur, pThdNxt = NULL;

    pThdCur = pTH(CurrentThread());

    DBGOUT(DBGCONTEXT,0,printf("PreemptIfNecessary CurThd %08x\n",(UINT)pThdCur));

    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);

    /* Watch out for zombies
     */
    assert(pThdCur != NULL);
    if (ThreadDied(pThdCur)) {
        /* If the current thread is dead then we should reschedule.
         * However, when death occurs the thread will look for
         * an undertaker, spinning IdleThread while none is found.
         * Which is the case we catch here.
         * Therefore we say the opposite.  Obviously ;-))
         */
        RescheduleNotNeeded();
        goto Exit;
    }

    /* If the current thread is already rescheduling, then we are done.
     */
    if (ThreadGetState(pThdCur) != THD_RUNNING)
        goto Exit;

    /* Fetch the most urgent ready (runnable but not running) thread.
     */
    pThdNxt = MostUrgentReadyThread();

    /* If there are no other runnable threads, then we are done.
     */
    if (pThdNxt == NULL)
        goto Exit;

    /* Refresh the constraint of the currently running thread before
     * we start comparing it to the constraints of the ready threads.
     */
    UpdateRestartBy(pTH(pThdCur)->Constraint);

    /* If current thread remains the most urgent runnable thread,
     * then do not preempt.
     * If the current thread is not less urgent than the most urgent
     * runnable thread, then we let it keep running.
     */
    if (IThreadUrgencyGTEQ(pThdCur, pThdNxt,pNextTimeSlice)) {
        goto Exit;
    }

    /* We have decided to preempt the current thread.
     */
    pTH(pThdCur)->WaitObject = NULL;
    RescheduleNeeded();

    /* All cases must end here, to turn off the timer.
     */
Exit:
    return pThdNxt;
}

/* The programmable timer interrupt went off.
 * [This is on the interrupt path from the timer.]
 *
 * Move all threads whose start time has arrived from sleep queue to
 * ready queue.  Return start time of earliest thread, Large number if none.
 *
 */
PRIVATE TIME TimerInterrupt(void)
{
    TIME ThdTime;
    TIME CurTime;
    PTHDINFO pThd;

    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);
    assert(CheckStackOverflow(CurrentThread()));

    if (Cbs.SleepQueue!=NULL) {

        CurTime = CurrentTime();

        do {

            /* Get earliest start time thread
             */
#if SLEEPLIST
            pThd = Cbs.SleepQueue;
#else
            pThd = (PTHDINFO) GetThreadFromSleepQueue(Cbs.SleepQueue);
#endif
            assert(pTH(pThd)->State == THD_WAITING
                   || pTH(pThd)->State == THD_IDLE);

            /* Thread's start time has arrived yet ?
             * Compute time distance from now.
             */
            INT64SUB(ThdTime,pTH(pThd)->StartTime,CurTime);

            if (TimeIsNotPositive(ThdTime))  {

                /* Yes, make runnable.
                 */
                PullThreadOffSleepQueue(pThd);
                
                if (IThreadGetWaitType(pThd) == WAIT_CONDITION) {

                    ConditionTimeoutThread(
                                           (PCONDITION)pTH(pThd)->WaitObject,
                                           iTH(pThd),
                                           ThreadConditionCancel);
                } else {
                    IReadyThread(pThd,FALSE);/*preempt..later*/
                }

            } else {
                /* Threads are queued by timeout; if thread's timeout hasn't
                 * arrived timeouts of threads after it havn't arrived either,
                 * no need to scan the rest of the queue.
                 */
                
                /* If noone is runnable then TheCurrentThread can keep
                 * on going until the wakeup time of PTHD.  Otherwise
                 * we must cover for the time-sharing case: if the most
                 * urgent thread on the run queue is not a RT thread,
                 * IThreadUrgencyGTEQ() will not trim down ThdTime.
                 * Therefore we have to do it here (conservatively).
                 */
                if ((Cbs.ReadyQueue != NULL) &&
                    (TimeLess(TIME_SHARING_TIMESLICE,ThdTime)))
                    ThdTime = TIME_SHARING_TIMESLICE;
                return ThdTime;
            }
        } while (Cbs.SleepQueue != NULL);
    }

    return TIME_SHARING_TIMESLICE;
}

/* make ready from waiting, sleeping... anything besides ready or running */
PRIVATE void IReadyThread(PTHDINFO pThd, BOOL CanReschedule)
{
    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);

    /* The thread might have died while we were playing with it.
     * For example, the helper thread is preemptible
     * while it does things for other threads, so it can be preempted
     * and then they can be terminated.
     */
    if (ThreadDied(pThd))
        return;

    /* if pThd is running, use PutThreadOnReadyQueue */
    /* if pThd is ready, we have already done all this */
    assert(ThreadGetState(pThd) != THD_RUNNING);
    assert(ThreadGetState(pThd) != THD_READY);

    /* pThd must not be on an object queue or on the sleep queue */
    assert(IThreadGetQueueType(pThd) == QUEUE_NONE);

    /* if thread was waiting, may resolve constraint with relative wait */
    if (pTH(pThd)->PendingConstraint
        && (ThreadGetState(pThd) == THD_WAITING
           || ThreadGetState(pThd) == THD_IDLE)) /* BUGBUG yes/no for IDLE? */
      {
          IThreadResolveConstraint(pThd,FALSE);
      }

    /* make thread ready */
    ThreadSetState(pThd,THD_READY);
    pTH(pThd)->WaitObject = NULL;
    PutThreadOnReadyQueue(pThd);

    /* preempt current thread if appropriate */
    if (CanReschedule)
        (void) PreemptIfNecessary(NULL);

}

PRIVATE void IConstraintGetInheritedUrgency(PKCONSTRAINT Constraint,
                                           PKURGENCY Urgency)
{
    PKCONSTRAINT Waiter;
    KURGENCY WaiterUrgency;

    assert(Constraint);
    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);

#define TRUST_CACHED_INHERITED_URGENCY  0
#if     TRUST_CACHED_INHERITED_URGENCY
    /* If the cached inherited urgency is still valid, then just return it. */
    if (Constraint->CachedInheritedUrgencyValid) {
#ifdef _ASSERT
        KURGENCY OldUrgency, NewUrgency;

        OldUrgency = Constraint->CachedInheritedUrgency;

        /* See what happens if we force a reevaluation */
        Constraint->CachedInheritedUrgencyValid = FALSE;
        IConstraintGetInheritedUrgency(Constraint, &NewUrgency);

        /* It had better be the same! */
        assert(IUrgencyEQ(&OldUrgency, &NewUrgency));
#endif
        *Urgency = Constraint->CachedInheritedUrgency;
        return;
    }
#endif

    /* InheritedUrgency starts at uninherited (base) urgency. */
    Constraint->CachedInheritedUrgency = Constraint->LocalUrgency;

    /* Recalculate inherited as max of local and all waiter urgencies. */
    for (Waiter = Constraint->FirstWaiter;
         Waiter != NULL;
         Waiter = Waiter->NextWaiter) {

        assert(Waiter->Thread);
        assert(pTH(Waiter->Thread)->Constraint);

        IConstraintGetInheritedUrgency(Waiter, &WaiterUrgency);

        if (IUrgencyGT(&WaiterUrgency, &Constraint->CachedInheritedUrgency)) {
            Constraint->CachedInheritedUrgency = WaiterUrgency;
        }
    }

    /* Mark InheritedUrgency as valid, and return it. */
    Constraint->CachedInheritedUrgencyValid = TRUE;
    *Urgency = Constraint->CachedInheritedUrgency;
}

PRIVATE void IThreadRecalculateInheritance(PTHDINFO pThd)
{
    PKCONSTRAINT Constraint = pTH(pThd)->Constraint;
    BOOL OldUrgencyValid;
    KURGENCY OldUrgency;
    KURGENCY NewUrgency;
    UINT State;

    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);

    /* Save the cached inherited urgency if it is valid */
    OldUrgencyValid = Constraint->CachedInheritedUrgencyValid;
    if (OldUrgencyValid)
        OldUrgency = Constraint->CachedInheritedUrgency;

    /* Force a reevaluation of the inherited urgency. */
    Constraint->CachedInheritedUrgencyValid = FALSE;
    IConstraintGetInheritedUrgency(Constraint, &NewUrgency);

    /* If the cached urgency was valid, and it has not changed, then return */
    if (OldUrgencyValid && IUrgencyEQ(&OldUrgency, &NewUrgency))
        return;

    /* If thread is ready or running, then reschedule if necessary, and return.
     * Otherwise, the thread must be waiting, in which case we propagate
     * the change to the thread that it is waiting on.
     */
    State = ThreadGetState(pThd);
    if (State == THD_READY || State == THD_RUNNING) {
        (void) PreemptIfNecessary(NULL);
        return;
    }

    /* Otherwise, the thread must be waiting, in which case we propagate
     * the change to the thread that it is waiting on.
     */
    assert(State == THD_WAITING || State == THD_IDLE);
    if (Constraint->Owner) {
        PTHDINFO pThdOwner = pTH(Constraint->Owner->Thread);

        /* make sure that pThdOwner points back to the right constraint */
        assert(pTH(pThdOwner)->Constraint == Constraint->Owner);

        /* propagate inheritance */
        IThreadRecalculateInheritance(pThdOwner);
    }
}

/* For use by waiter, to take inheritance back from owner.
 */
PRIVATE void StopInheritance(PKCONSTRAINT Owner, PKCONSTRAINT Waiter)
{
    PKCONSTRAINT *WaiterPtr;

    /* The owner and the waiter must be distinct and both non-null,
     * and the owner must be who the waiter thinks it is.
     */
    assert(Owner != NULL);
    assert(Waiter != NULL);
    assert(Waiter != Owner);
    assert(Waiter->Owner == Owner);

    /* The waiter must be on a owner's waiter list.
     */
    assert(Waiter->NextWaiter != Waiter);

    /* Find and remove constraint from the owner's waiter list
     */
    for (WaiterPtr = &Owner->FirstWaiter; ;
         WaiterPtr = &(*WaiterPtr)->NextWaiter) {

        /* The waiter MUST be on the list somewhere.
         * Thus we should never reach the end of the list in this loop.
         */
        assert(*WaiterPtr != NULL); /* end of list marker */
        assert(*WaiterPtr != (*WaiterPtr)->NextWaiter); /* just in case... */

        /* If we have found the waiter, then remove it and exit the loop.
         */
        if (*WaiterPtr == Waiter) {
            *WaiterPtr = (*WaiterPtr)->NextWaiter;
            break;
        }
    }

    /* Mark the constraint as not having an owner and
     * not being on an owner's waiter list.
     */
    Waiter->Owner = NULL;
    Waiter->NextWaiter = Waiter;
}

/* Exported form of StopInheritance (uses threads instead of constraints).
 */
#define IThreadStopInheritance(_own_,_wtr_) \
     StopInheritance(pTH(_own_)->Constraint,pTH(_wtr_)->Constraint)

/* Sleeping interface
 */

SCODE SleepUntil(TIME Until)
{
    BOOL PreemptionWasDisabled;

    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler));

    PreemptionWasDisabled = CbsDisablePreemption(pTheScheduler,TRUE);
    LogIt(7,0-Uint64Low32Bits(Until));
    IThreadSleep(pTH(CurrentThread()), Until);
    CbsEnablePreemption(pTheScheduler,PreemptionWasDisabled);
    LogIt(15,0);
    return S_OK;
}

/* Mutex interface
 */

/* Add pThdWaiter to the list of threads blocked by pThd.
 * pThd owns mutex pmx.
 */
PRIVATE void ThreadQueueMutexWaiter(PTHDINFO pThd, PTHDINFO pThdWaiter,
                                    PMUTEX pMutex)
{
    DBGOUT(DBGCONTEXT,1,printf("Block Thd %08x on pmx %08x owner Thd %08x\n",(UINT)pThdWaiter,(UINT)pMutex,(UINT)pThd));

#if NEWLIST
    ListAddLast(&pThd->ThdMxWaiters, TQ_off, pThdWaiter);
 #else
    pTH(pThd)->plnThdMxWaiters = ThreadQueueFifo(pThdWaiter,
                                                 pTH(pThd)->plnThdMxWaiters);
#endif
    IThreadSetQueueType(pThdWaiter,QUEUE_MUTEX);
    IThreadSetStateWaiting(pThdWaiter,pThd,WAIT_MUTEX,(PTR)pMutex,(UINT)pThd);

    RescheduleNeeded();
}

/* Wake all threads blocked on mutex pmx which is owned by pThd.  Note that the
 * list hanging off of pThd may include threads waiting on different mutexes.
 * Only remove those who are waiting on pmx; if pmx == NULL remove all threads.
 */
#if NEWLIST
PRIVATE void ThreadWakeBlockedThreads(PTHDINFO pThdOwner, PMUTEX pMutex)
{
    PTHDINFO Waiter, First, Next;
    BOOL MightReschedule = FALSE;

    DBGOUT(DBGCONTEXT,1,printf("Wake up threads blocked on pmx %08x owner Thd %08x\n",(UINT)pMutex,(UINT)pThdOwner));

    Next = NULL; /* Quiet compiler warning */

    First = pThdOwner->ThdMxWaiters; /* Get an anchor */
    if (First == NULL)
        return;

    for (Waiter = First; Waiter != NULL; Waiter = Next) {
        assert(Waiter->State == THD_WAITING);
        assert(Waiter->WaitType == WAIT_MUTEX);

        /* Remove threads waiting on this wait object,
         * all threads if wait object not specified.
         */
        if (pMutex == NULL || Waiter->WaitObject == (PTR)pMutex) {

            DBGOUT(DBGCONTEXT,1,printf("Making Thd %08x ready\n",(UINT)Waiter));

            /* Termination case:
             */
            if (pMutex == NULL) {
                PMUTEX pMx = (PMUTEX) Waiter->WaitObject;
                pMx->_mutex_state[0] = 0; /* BUGBUGBUG hack */
            }

            Waiter->WaitData = 0; /* See ThreadQueueMutexWaiter() */
            Waiter->QueueType = QUEUE_NONE;
            Next = ListNodeRemove(TQ_off, Waiter); /* Remove from queue */

            /* assert(IThreadInheritsFrom(pThdOwner, pThdWaiter)); */
            IThreadStopInheritance(pThdOwner, Waiter);
            IReadyThread(Waiter,FALSE);            /* Sets state to THD_READY */
            MightReschedule = TRUE;

            if (Waiter == First) { /* Dequeued anchor    */
                First = Next;   /* New anchor         */
                continue;             /* Process new anchor */
            }
        }
        else /* Skip thread waiting on some other object */
            Next = ListNodeNext(TQ_off, Waiter);

        if (Next == First)
            break;
    }

    pThdOwner->ThdMxWaiters = Next;


    if (MightReschedule && pMutex != NULL)
        (void) PreemptIfNecessary(NULL);
}
#else
PRIVATE void ThreadWakeBlockedThreads(PTHDINFO pThdOwner, PMUTEX pMutex)
{
    PLISTNODE plnFirst, plnWaiter, plnNext;
    PTHDINFO pThdWaiter;
    BOOL MightReschedule = FALSE;

    DBGOUT(DBGCONTEXT,1,printf("Wake up threads blocked on pmx %08x owner Thd %08x\n",(UINT)pMutex,(UINT)pThdOwner));

    plnNext = NULL; /* Quiet compiler warning */

    plnFirst = pTH(pThdOwner)->plnThdMxWaiters; /* Get an anchor */
    if (plnFirst == NULL)
        return;

    for (plnWaiter = plnFirst; plnWaiter != NULL; plnWaiter = plnNext) {

        pThdWaiter = GetThreadFromQueue(plnWaiter);

        assert(pTH(pThdWaiter)->State == THD_WAITING);
        assert(pTH(pThdWaiter)->WaitType == WAIT_MUTEX);

        /* Remove threads waiting on this wait object,
         * all threads if wait object not specified.
         */
        if (pMutex == NULL || pTH(pThdWaiter)->WaitObject == (PTR)pMutex) {

            DBGOUT(DBGCONTEXT,1,printf("Making Thd %08x ready\n",(UINT)pThdWaiter));

            /* Termination case:
             */
            if (pMutex == NULL) {
                PMUTEX pMx = (PMUTEX)pTH(pThdWaiter)->WaitObject;
                pMx->_mutex_state[0] = 0; /* BUGBUGBUG hack */
            }

            pTH(pThdWaiter)->WaitData = 0; /* See ThreadQueueMutexWaiter() */
            plnNext = ThreadDequeue(pThdWaiter); /* Remove from queue       */

            /* assert(IThreadInheritsFrom(pThdOwner, pThdWaiter)); */
            IThreadStopInheritance(pThdOwner, pThdWaiter);
            IReadyThread(pThdWaiter,FALSE);            /* Sets state to THD_READY */
            MightReschedule = TRUE;

            if (plnWaiter == plnFirst) { /* Dequeued anchor    */
                plnFirst = plnNext;   /* New anchor         */
                continue;             /* Process new anchor */
            }
        }
        else /* Skip thread waiting on some other object */
            plnNext = plnWaiter->pNext;

        if (plnNext == plnFirst)
            break;
    }

    pTH(pThdOwner)->plnThdMxWaiters = plnNext;


    if (MightReschedule && pMutex != NULL)
        (void) PreemptIfNecessary(NULL);
}
#endif

/* Condition interface
 */

/* Add pThdNew to list of threads waiting on a condition waited by pThdFirst.
 * pThdFirst already waits on the condition, is the 'first waiter'.  Returns
 * pointer to a new 'first waiter'.  If pThdFirst == NULL there isn't 'first
 * waiter', and pThdNew becomes the 'first waiter'.
 *
 * Note that both the 'first waiter' and the new thread should not be on any
 * queue, or list (that is Queue should be unused) when calling this function
 * though Queue field can be used before, after this function is called.
 */
PRIVATE PTHDINFO ThreadQueueCondWaiter(PTHDINFO pThdFirst, PTHDINFO pThdNew, PCONDITION pcd)
{
    DBGOUT(DBGCONTEXT,1,printf("Wait Thd %08x on pcd %08x first waiter Thd %08x\n",(UINT)pThdNew,(UINT)pcd,(UINT)pThdFirst));

    /* Mark the new thread as waiting on the condition.
     */
    IThreadSetQueueType(pThdNew,QUEUE_CONDITION);
    IThreadSetStateWaiting(pThdNew,NULL,WAIT_CONDITION,pcd,0);

    /* Make sure we do not get confused later
     */
    RescheduleNeeded();

    /* If there is no first waiter, then the new thread becomes the
     * first waiter.
     */
    if (pThdFirst == NULL)
        return pThdNew;

    /* Otherwise, just add the new thread to the list of waiters, keeping
     * the first waiter the same. We don't order by constraint here; instead,
     * we look for the most urgent thread when we dequeue (if we do a signal
     * instead of a broadcast).
     */
#if NEWLIST
    ListAddLast(&pThdFirst->ThdCdWaiters, TQ_off, pThdNew);                
#else
    pTH(pThdFirst)->plnThdCdWaiters =
        ThreadQueueFifo(pThdNew, pTH(pThdFirst)->plnThdCdWaiters);
#endif

    return pThdFirst;
}


/* Remove 'first waiter' thread off of condition's queue.  The next thread on
 * the queue, if any, becomes 'first waiter'.  Returns pointer to the new
 * 'first waiter', NULL if none.
 *
 * Note that the 'first waiter' should not be on any queue, or list (that is
 * Queue should be unused) when calling this function though Queue field can
 * be used before, after this function is called.
 */
#if NEWLIST
PRIVATE PTHDINFO ThreadRemoveFirstWaitingThread(PTHDINFO pThdFirst)
{
    PTHDINFO pNewFirst;

    assert(pThdFirst != NULL);

    pNewFirst = pThdFirst->ThdCdWaiters;
    if (pNewFirst == NULL)
        return NULL;

    pNewFirst->QueueType = QUEUE_NONE;
    pNewFirst->ThdCdWaiters = ListNodeRemove(TQ_off, pNewFirst);
    assert(pNewFirst->ThdCdWaiters != pNewFirst);

    return pNewFirst;
}
#else
PRIVATE PTHDINFO ThreadRemoveFirstWaitingThread(PTHDINFO pThdFirst)
{
    PLISTNODE plnNewFirst;
    PTHDINFO pThdNewFirst;

    assert(pThdFirst != NULL);

    plnNewFirst = pTH(pThdFirst)->plnThdCdWaiters;
    if (plnNewFirst == NULL)
        return NULL;

    pThdNewFirst = GetThreadFromQueue(plnNewFirst); /* Next on list */
    pTH(pThdNewFirst)->plnThdCdWaiters = ThreadDequeue(pThdNewFirst);
    assert(GetThreadFromQueue(pTH(pThdNewFirst)->plnThdCdWaiters) !=
           pThdNewFirst);

    return pThdNewFirst;
}
#endif



/* Wake a thread waiting on a condition.
 * sc = S_OK : Condition signalled
 * sc = S_FALSE: Thread waiting on the condition terminated.
 */
PRIVATE void WakeOneWaitingThread(PTHDINFO pThd, SCODE sc, BOOL Preempt)
{
    DBGOUT(DBGCONTEXT,1,printf("Making Thd %08x ready (pcd %08x)\n",(UINT)pThd,(UINT)pTH(pThd)->WaitObject));
    DBGOUT(DBGCONTEXT,3,if (pTH(pThd)->State != THD_WAITING || pTH(pThd)->WaitType != WAIT_CONDITION) printf(
           "WakeOneWaitingThread: pThd %08x State %d != THD_WAITING ||\n"
           "WaitType %d != WAIT_CONDITION\n",(UINT)pThd,pTH(pThd)->State,pTH(pThd)->WaitType));

    assert(pTH(pThd)->State == THD_WAITING || pTH(pThd)->State == THD_IDLE);
    assert(pTH(pThd)->WaitType == WAIT_CONDITION);

#if SLEEPLIST
    /* Pull off sleep queue if it is on it */
    if (!ListNodeIsAlone(TSQ_off, pThd) || (Cbs.SleepQueue == pThd))
#endif
        PullThreadOffSleepQueue(pThd);
    pTH(pThd)->LastError = sc;
    IReadyThread(pThd,Preempt);            /* Sets state to THD_READY */
}


/* Wake up all threads waiting on a condition whose 'first waiter' is pThd,
 * including pThd.  sc values specified in WakeOneWaitingThread.
 *
 * Note that the 'first waiter' should not be on any queue, or list (that is
 * Queue should be unused) when calling this function; however, the Queue
 * field can be used before and after this function is called.
 */
#if NEWLIST
PRIVATE void ThreadWakeAllWaitingThreads(PTHDINFO pThd, SCODE sc)
{
    PTHDINFO Next;

    assert(pThd != NULL);
    pThd->QueueType = QUEUE_NONE;

    DBGOUT(DBGCONTEXT,1,printf("Wake up all waiters on pcd %08x first Thd %08x sc %08x\n",pTH(pThd)->WaitObject,pThd,sc));

    /* Start by waking the 'first waiter' and saving its list of threads */
    Next = pThd->ThdCdWaiters;  /* Thds (exc 'first waiter') */
    pThd->ThdCdWaiters = NULL;  /* This one is removed       */
    WakeOneWaitingThread(pThd,sc,FALSE);        /* Wake it up                */

    /* Now wake any remaining threads */
    while (Next) {
        pThd = Next;
        pThd->QueueType = QUEUE_NONE;
        Next = ListNodeRemove(TQ_off, pThd);
        assert(Next != pThd);
        WakeOneWaitingThread(pThd,sc,FALSE);
    }
    (void) PreemptIfNecessary(NULL);
}
#else
PRIVATE void ThreadWakeAllWaitingThreads(PTHDINFO pThd, SCODE sc)
{
    PLISTNODE plnNext;

    assert(pThd != NULL);
    pTH(pThd)->QueueType = QUEUE_NONE; /* Quiet assertions */

    DBGOUT(DBGCONTEXT,1,printf("Wake up all waiters on pcd %08x first Thd %08x sc %08x\n",pTH(pThd)->WaitObject,pThd,sc));

    /* Start by waking the 'first waiter' and saving its list of threads */
    plnNext = pTH(pThd)->plnThdCdWaiters; /* Thds (exc 'first waiter') */
    pTH(pThd)->plnThdCdWaiters = NULL;    /* This one is removed       */
    WakeOneWaitingThread(pThd,sc,FALSE);        /* Wake it up                */

    /* Now wake any remaining threads */
    while (plnNext) {
        pThd = GetThreadFromQueue(plnNext);
        plnNext = ThreadDequeue(pThd);
        assert(GetThreadFromQueue(plnNext) != pThd);
        WakeOneWaitingThread(pThd,sc,FALSE);
    }
    (void) PreemptIfNecessary(NULL);
}
#endif

/* Wake up the most urgent thread waiting on the condition whose
 * 'first waiter' is pThd.  sc values specified in WakeOneWaitingThread.
 * Return pointer to the new first waiter, or NULL if none.
 *
 * Note that the 'first waiter' should not be on any queue, or list (that is
 * Queue should be unused) when calling this function; however, the Queue
 * field can be used before and after this function is called.
 */
#if NEWLIST
PRIVATE PTHDINFO ThreadWakeMostUrgentWaitingThread(PTHDINFO pThd, SCODE sc)
{
#if 0
    pThdWake = ListWalk(&pThd->ThdCdWaiters, TQ_off, LIST_WALK_LARGEST,
                        ThreadUrgencyCmp, NULL);
#else
    PTHDINFO pThdWake, Tmp;

    assert(pThd != NULL);
    pTH(pThd)->QueueType = QUEUE_NONE; /* Quiet assertions */

    DBGOUT(DBGCONTEXT,1,printf("Wake up most urgent waiter on pcd %08x first Thd %08x sc %08x\n",pTH(pThd)->WaitObject,pThd,sc));

    /* Find the most urgent thread by saving the list, detaching the list
     * from the 'first waiter', adding the 'first waiter' to the list, and
     * then removing the most urgent thread from the list.
     *
     * But first, handle the special case of only one waiter.
     */
    Tmp = pThd->ThdCdWaiters;                   /* save list */
    if (Tmp == NULL) {                          /* handle special case */
        WakeOneWaitingThread(pThd,sc,TRUE);     /* wake the only thread */
        return NULL;                            /* no threads left */
    }

    pThd->ThdCdWaiters = NULL;                  /* detach list from waiter */
    ListAddLast(&Tmp, TQ_off, pThd);            /* add waiter to list */
    pThdWake = MostUrgentThreadOnQueue(Tmp);    /* get most urgent thread */

    pThd->QueueType = QUEUE_NONE;               /* Quiet assertions */
    Tmp = ListNodeRemove(TQ_off, pThdWake);     /* remove it from list */

    /* Pull new 'first waiter' from the remaining list, and point it at the
     * new remaining list.
     */
    pThd = Tmp;
    if (pThd) {
        pThd->QueueType = QUEUE_NONE;
        pThd->ThdCdWaiters = ListNodeRemove(TQ_off, pThd);
    }
#endif

    /* Now wake the most urgent thread, and return the new 'first waiter'.
     */
    WakeOneWaitingThread(pThdWake,sc,TRUE);
    return pThd; /* New 'first waiter' */
}
#else
PRIVATE PTHDINFO ThreadWakeMostUrgentWaitingThread(PTHDINFO pThd, SCODE sc)
{
    PLISTNODE plnTmp;
    PTHDINFO pThdWake;

    assert(pThd != NULL);
    pTH(pThd)->QueueType = QUEUE_NONE; /* Quiet assertions */

    DBGOUT(DBGCONTEXT,1,printf("Wake up most urgent waiter on pcd %08x first Thd %08x sc %08x\n",pTH(pThd)->WaitObject,pThd,sc));

    /* Find the most urgent thread by saving the list, detaching the list
     * from the 'first waiter', adding the 'first waiter' to the list, and
     * then removing the most urgent thread from the list.
     *
     * But first, handle the special case of only one waiter.
     */
    plnTmp = pTH(pThd)->plnThdCdWaiters;        /* save list */
    if (plnTmp == NULL) {                       /* handle special case */
        WakeOneWaitingThread(pThd,sc,TRUE);     /* wake the only thread */
        return NULL;                            /* no threads left */
    }

    pTH(pThd)->plnThdCdWaiters = NULL;          /* detach list from waiter */
    plnTmp = ThreadQueueFifo(pThd,plnTmp);      /* add waiter to list */
    pThdWake = MostUrgentThreadOnQueue(plnTmp); /* get most urgent thread */

    plnTmp = ThreadDequeue(pThdWake);           /* remove it from list */

    /* Pull new 'first waiter' from the remaining list, and point it at the
     * new remaining list.
     */
    if (plnTmp) {
        pThd = GetThreadFromQueue(plnTmp);
        pTH(pThd)->plnThdCdWaiters = ThreadDequeue(pThd);
    } else {
        pThd = NULL;
    }

    /* Now wake the most urgent thread, and return the new 'first waiter'.
     */
    WakeOneWaitingThread(pThdWake,sc,TRUE);
    return pThd; /* New 'first waiter' */
}
#endif

/* If fAll is set, then wake up all threads waiting on the condition whose
 * 'first waiter' is pThd; otherwise, just wake the most urgent such thread.
 *
 * Note that the 'first waiter' should not be on any queue, or list (that is
 * Queue should be unused) when calling this function; however, the Queue
 * field can be used before and after this function is called.
 */
PRIVATE PTHDINFO ThreadWakeWaitingThread(PTHDINFO pThd, SCODE sc, BOOL fAll)
{
    if (fAll) {
        ThreadWakeAllWaitingThreads(pThd, sc);
        return NULL;
    } else
        return ThreadWakeMostUrgentWaitingThread(pThd, sc);
}

/* This function is called to cancel a wait on a condition, either because the
 * wait has timed out or because the thread is being terminated. pThd is the
 * 'first waiter' on the condition; pThdCancel is the thread whose wait is
 * being cancelled.  it may be the same as pThd. Returns a pointer to new
 * 'first waiter', NULL if none.
 */
#if NEWLIST
PITHREAD ThreadConditionCancel(PITHREAD iThd, PITHREAD iThdCancel)
{
    PTHDINFO Cur, First, Next;
    PTHDINFO pThd = pTH(iThd);
    PTHDINFO pThdCancel = pTH(iThdCancel);

    DBGOUT(DBGCONTEXT,1,printf("Cancel wait of Thd %08x on Cond %08x (first Thd %08x)\n",pThdCancel,pTH(pThd)->WaitObject,pThd));

    if (pThd == pThdCancel) {
        if (ThreadDied(pThdCancel))
            return iTH(ThreadRemoveFirstWaitingThread(pTH(pThdCancel)));
        else
            return iTH(ThreadWakeWaitingThread(pTH(pThdCancel),
                                               E_TIMED_OUT, FALSE));
    }

    /* Not 'first waiter' */

    First = pThd->ThdCdWaiters; /* Get an anchor */

    if (First != NULL) {
        Cur = First;

        do { /* Search for pThdCancel */
            if (Cur == pThdCancel) { /* Found */

                Cur->QueueType = QUEUE_NONE;
                Next = ListNodeRemove(TQ_off, Cur);
                if (Cur == First)              /* Dequeued anchor  */
                    pThd->ThdCdWaiters = Next; /* New anchor       */

                if (!ThreadDied(pThdCancel)) /* Wake it up */
                    WakeOneWaitingThread(pThdCancel, E_TIMED_OUT, TRUE);
                break;
            }
            Cur = ListNodeNext(TQ_off, Cur);
        } while (Cur != First);
    }

    return iTH(pThd);
}
#else
PITHREAD ThreadConditionCancel(PITHREAD pThd, PITHREAD pThdCancel)
{
    PLISTNODE plnFirst, plnCur, plnNext;
    PITHREAD pThdCur;

    DBGOUT(DBGCONTEXT,1,printf("Cancel wait of Thd %08x on Cond %08x (first Thd %08x)\n",pThdCancel,pTH(pThd)->WaitObject,pThd));

    if (pThd == pThdCancel) {
        if (ThreadDied(pThdCancel))
            return iTH(ThreadRemoveFirstWaitingThread(pTH(pThdCancel)));
        else
            return iTH(ThreadWakeWaitingThread(pTH(pThdCancel),
                                               E_TIMED_OUT, FALSE));
    }

    /* Not 'first waiter' */

    plnFirst = pTH(pThd)->plnThdCdWaiters; /* Get an anchor */

    if (plnFirst != NULL) {
        plnCur = plnFirst;

        do { /* Search for pThdCancel */
            pThdCur = iTH(GetThreadFromQueue(plnCur));
            if (pThdCur == pThdCancel) { /* Found */

                plnNext = ThreadDequeue(pTH(pThdCur));
                if (plnCur == plnFirst)              /* Dequeued anchor  */
                    pTH(pThd)->plnThdCdWaiters = plnNext; /* New anchor       */

                if (!ThreadDied(pTH(pThdCancel))) /* Wake it up */
                    WakeOneWaitingThread(pTH(pThdCancel), E_TIMED_OUT, TRUE);
                break;
            }
            plnCur = plnCur->pNext;
        } while (plnCur != plnFirst);
    }

    return pThd;
}
#endif


/* MMgr for constraint structures
 */


/* Free list of constraints.
 */
PRIVATE PTR pConstraintFree = NULL;

/* Callable without restrictions. Allocates a constraint
 * from the global free list.  Returns NULL if none available.
 */
#define QuickAllocConstraint() \
    (PKCONSTRAINT) AtomicLIFORemove( &pConstraintFree )

/* Callable without restrictions. Frees a constraint
 * to the global free list.
 */
#define QuickFreeConstraint(_c_) \
    AtomicLIFOInsert( &pConstraintFree, (PTR)_c_)

/* Garbage-collect unused constraints.
 */
#if 0
PRIVATE void CollectConstraints(void)
{
    PKCONSTRAINT Constraint;

    /* Pick up the head of the free list.  If someone else
     * races us to it (this same code, some other thread)
     * repeat until we either get a non-null list head or
     * the list was (or has become) empty.
     */
    do {
        Constraint = pConstraintFree;
    } while (!AtomicCmpAndSwap( (PUINT)&pConstraintFree,
                                (UINT) Constraint,
                                (UINT) NULL));


    while (Constraint != NULL) {
        PKCONSTRAINT Next = Constraint->Owner;

        FreeConstraint(Constraint);
        Constraint = Next;
    }
}
#endif

/* Allocate a constraint.  Try for the free list first, then the heap.
 */
PRIVATE PKCONSTRAINT AllocConstraint(void)
{
    PKCONSTRAINT Constraint;

    /* To prevent the free list from growing unnecessarily,
     * first try to allocate from it. In the common case,
     * we are allocating the constraint for a new thread,
     * and when user threads die their constraints get
     * into the free list, so in steady state we will
     * recycle constraints from the free list.
     */
    Constraint = QuickAllocConstraint();

    if (Constraint == NULL) {
        PIHEAP pHeap = CurrentHeap();

        Constraint = (PKCONSTRAINT)
                pHeap->v->Alloc(pHeap, HEAP_DEBUG_NAME("constraint"),
                                sizeof(KCONSTRAINT), 0);

#ifdef _DEBUG
        if (Constraint == NULL)
            printf("AllocConstraint -> %x\n", (UINT) Constraint);
#endif
    }

    return Constraint;
}

/* If this thread doesn't have a free constraint reserved
 * for its use, then allocate one.
 */
PRIVATE BOOL IThreadCheckFreeConstraint(PTHDINFO pThd)
{
    if (pTH(pThd)->FreeConstraint == NULL)
        pTH(pThd)->FreeConstraint = AllocConstraint();
    if (pTH(pThd)->FreeConstraint != NULL) {
        return FALSE;
    }
    return TRUE;
}

/* Get a free constraint for this thread to use.
 * One of the two CheckFreeConstraint variants must have
 * been used already, to reserve a free constraint
 * for this thread's use.
 */
PRIVATE PKCONSTRAINT IThreadAllocConstraint(PTHDINFO pThd)
{
    PKCONSTRAINT Constraint = pTH(pThd)->FreeConstraint;

    assert(Constraint != NULL);
    pTH(pThd)->FreeConstraint = NULL;
    return Constraint;
}

/* This thread is finished using this constraint structure.
 * Save it for later it we don't already have a free one saved;
 * otherwise put on global free list.
 */
PRIVATE void IThreadFreeConstraint(PTHDINFO pThd, PKCONSTRAINT Constraint)
{
    assert(pTheScheduler->v->PreemptionIsEnabled(pTheScheduler) == FALSE);

    if (pTH(pThd)->FreeConstraint == NULL) {

        pTH(pThd)->FreeConstraint = Constraint;

    } else {

        QuickFreeConstraint(Constraint);
    }
}

/* Free all the constraints associated with a thread.
 * This is only called by a dead thread, cannot block.
 * Used by thread cleanup.
 */
PRIVATE void IThreadFreeConstraints(PTHDINFO pThd)
{
    PKCONSTRAINT Constraint;

    Constraint = pTH(pThd)->Constraint;
    if ((Constraint != NULL) &&
        (Constraint != &pTH(pThd)->InitialConstraint)) {

        /* BUGBUG What if the thread died while inside a nested constraint?
         * Can there be more than one constraint to free here?
         */
        assert(Constraint->Thread == iTH(pThd));
        CurrentHeapDelayedFree( Constraint );
    }

    if ((Constraint = pTH(pThd)->FreeConstraint) != NULL)
        CurrentHeapDelayedFree( Constraint );
}
