/* cb_sched2.c
 * Rewrite of constraint-based scheduler.
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#define OPTIMIZE      /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <fred.h>
#include <diagnostics.h>
#include <base/schedul.h>
#include <schedulif.h>
#include <base/thread2.h>
#include <base/tconstraint2.h>
#include <base/mutex2.h>
#include <util/list-single.h>

/* Enable/disable logging by uncommenting/commenting the following line */
/* #define LOGGING_ENABLED 1 */

#include <util/logger.h>

#ifdef LOGGING_ENABLED
#define SCHEDULER_LOG_SIZE 128
LOG_ENTRY SchedulerLog[SCHEDULER_LOG_SIZE];

LOG SchedLog;
PLOG pSchedLog = &SchedLog;

DECLARE_LOG_EVENT_TYPE(RESCHEDULE             , "Resched ", TRUE );
DECLARE_LOG_EVENT_TYPE(RESCHEDULE_MAIN        , "ReschInt", TRUE );
DECLARE_LOG_EVENT_TYPE(IDLE_THREAD            , "enIdle  ", TRUE );
DECLARE_LOG_EVENT_TYPE(END_IDLE_THREAD        , "exIdle  ", TRUE );
DECLARE_LOG_EVENT_TYPE(DEAD_THREAD            , "dead    ", TRUE );
DECLARE_LOG_EVENT_TYPE(SLEEP_UNTIL            , "enSleep ", TRUE );
DECLARE_LOG_EVENT_TYPE(OUT                    , "OUT     ", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_CONDITION_WAIT   , "enCdWait", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_CONDITION_WAIT    , "exCdWait", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_CONDITION_SIGNAL , "enCdSign", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_CONDITION_SIGNAL  , "exCdSign", TRUE );
DECLARE_LOG_EVENT_TYPE(DRAIN_DEFERRED         , "drainDef", TRUE );
DECLARE_LOG_EVENT_TYPE(PUT_READY_QUEUE        , "addReady", TRUE );
DECLARE_LOG_EVENT_TYPE(POP_READY_QUEUE        , "delReady", TRUE );
DECLARE_LOG_EVENT_TYPE(PUT_SLEEP_QUEUE        , "addSleep", TRUE );
DECLARE_LOG_EVENT_TYPE(POP_SLEEP_QUEUE        , "delSleep", TRUE );
DECLARE_LOG_EVENT_TYPE(THREAD_STATE           , "ThdState", TRUE );
DECLARE_LOG_EVENT_TYPE(ADD_CONSTRAINT         , "AddKC   ", TRUE );
DECLARE_LOG_EVENT_TYPE(ADD_THREAD             , "AddThd  ", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT                , "Preempt ", TRUE );
DECLARE_LOG_EVENT_TYPE(REMOVE_CONSTRAINT      , "RemoveKC", TRUE );
DECLARE_LOG_EVENT_TYPE(CONSTRAINT_STACK_FREE  , "KCStackF", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_WAKE_THREADS     , "enWake  ", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_WAKE_REALTIME     , "exWakeRT", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_WAKE_TIMESHARE    , "exWakeTS", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_UPDATE_RESTARTBY , "enUpRsBy", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_UPDATE_RESTARTBY  , "exUpRsBy", TRUE );
DECLARE_LOG_EVENT_TYPE(NO_UPDATE_RESTARTBY    , "noUpRsBy", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT_DEAD           , "PreDead ", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT_RESCHEDULING   , "PreResch", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT_NO_OTHERS      , "PreNoOth", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT_STILL_URGENT   , "PreStill", TRUE );
DECLARE_LOG_EVENT_TYPE(MAKE_THREAD_READY      , "MakReady", TRUE );

#undef LOGGING_ENABLED

#endif

#if defined(__NO_BUILTIN_INT64)
const TIME TIME_SHARING_TIMESLICE = Int64Initializer(0,400000/*Low*/);
#else
#define TIME_SHARING_TIMESLICE   400000 /* 40 millisecs */
#endif

#include <base/kconstraint2.h>

/* MMLite Constraint-based scheduler object (Cbs) */
typedef struct _CBS {
  const struct ISchedulerVtbl *v;
  UINT PreemptionDisableCount;    
  BOOL ContextSwitchNeeded;
  LIST lReadyQueue;
  LIST lSleepQueue;
  PKCONSTRAINT pkcFreeStackHead;
  UINT RefCount;
  UINT nThreads;
  TIME IdleTime;
  BOOL FirstThreadBooted;
} CBS, *PCBS;

#define pCBS(_x_) ((PCBS)_x_)

#define RescheduleNeeded() Cbs.ContextSwitchNeeded = TRUE
#define RescheduleNotNeeded() Cbs.ContextSwitchNeeded = FALSE

/*  assert(CurrentHeap()->v->Validate(CurrentHeap(),0,0) */

/* ===========================================================================
 * Ready Queue Macros
 */

#define /* PLIST */                                                           \
CbsGetReadyQueue()                                                            \
     &(Cbs.lReadyQueue)

#define /* void */                                                            \
CbsReadyQueueInitialize()                                                     \
     List_Init(CbsGetReadyQueue())

#define /* BOOL */                                                            \
CbsReadyQueueIsEmpty()                                                        \
     List_IsEmpty(CbsGetReadyQueue())

#define /* PKCONSTRAINT */ /* Must call CbsReadyQueueIsEmpty first! */        \
CbsReadyQueueFindMostUrgentConstraint()                                       \
     ((PKCONSTRAINT)(Node_GetData(List_GetHead(CbsGetReadyQueue()))))

#define /* PKCONSTRAINT */ /* Must call CbsReadyQueueIsEmpty first! */        \
CbsReadyQueueRemoveMostUrgentConstraint()                                     \
     ((PKCONSTRAINT)(Node_GetData(List_HeadRemove(CbsGetReadyQueue()))))

#define /* void */                                                            \
CbsReadyQueueAddConstraint(/* PKCONSTRAINT */ _pkc_)                          \
     List_InsertSortedDescending(CbsGetReadyQueue(),                         \
                                KConstraint_UrgencyGT,                        \
                                KConstraint_GetSchedQueueNode(_pkc_))

/* ===========================================================================
 * Sleep Queue Macros
 */

#define /* PLIST */                                                           \
CbsGetSleepQueue()                                                            \
     &(Cbs.lSleepQueue)

#define /* void */                                                            \
CbsSleepQueueInitialize()                                                     \
     List_Init(CbsGetSleepQueue())

#define /* PNODE */                                                           \
CbsSleepQueueGetFirstNode()                                                   \
     List_GetHead(CbsGetSleepQueue())

#define /* void */                                                            \
CbsSleepQueueIsEmpty()                                                        \
     List_IsEmpty(CbsGetSleepQueue())

#define /* PKCONSTRAINT */ /* Must call CbsReadyQueueIsEmpty first! */        \
CbsSleepQueueFindMostUrgentConstraint()                                     \
     ((PKCONSTRAINT)(Node_GetData(List_GetHead(CbsGetSleepQueue()))))

#define /* PKCONSTRAINT */ /* Must call CbsSleepQueueIsEmpty first! */        \
CbsSleepQueueRemoveMostUrgentConstraint()                                     \
     ((PKCONSTRAINT)(Node_GetData(List_HeadRemove(CbsGetSleepQueue()))))

#define /* void */                                                            \
CbsRecycleConstraint(/* PKCONSTRAINT */ _pkc_)                                \
     Cbs.pkcFreeStackHead = KConstraint_Push(Cbs.pkcFreeStackHead, (_pkc_))

#define /* void */                                                            \
CbsFreeConstraint(/* PKCONSTRAINT */ _pkc_)                                   \
   CurrentHeapDelayedFree(_pkc_);

/* ===========================================================================
 * Scheduler functions that implement exported interface methods.
 */

PRIVATE SCODE MCT CbsQueryInterface( 
    PISCHEDULER pThis,
    /* [in] */ REFIID pIid,
    /* [out] */ void **ppUnk
    );

PRIVATE UINT MCT CbsAddRef(
    PISCHEDULER pThis
    );

PRIVATE UINT MCT CbsRelease (
    PISCHEDULER pThis
    );

PRIVATE PITHREAD MCT CbsReschedule(
    PISCHEDULER This,
    PITHREAD pThdCurrent
    );

PRIVATE BOOL MCT CbsDisablePreemption(
    PISCHEDULER This,
    BOOL Recursive
    );

PRIVATE BOOL MCT CbsEnablePreemption(
    PISCHEDULER This,
    BOOL Recursive
    );

PRIVATE BOOL MCT CbsPreemptionIsEnabled(
    PISCHEDULER This
    );

PRIVATE void MCT CbsAddThread(
    PISCHEDULER This,
    PITHREAD Thread,
    PTIME_CONSTRAINT pcval);

PRIVATE void MCT CbsRemoveThread(
    PISCHEDULER This,
    PITHREAD Thread
    );

PRIVATE void MCT CbsWakeUpOnMutex(
    PISCHEDULER This,
    PMUTEX pMutex
    );

PRIVATE void MCT CbsWaitForMutex(
    PISCHEDULER This,
    PMUTEX pMutex
    );

PRIVATE void MCT CbsWaitForCondition(
    PISCHEDULER This,
    PCONDITION pCondition,
    PTIME pTimeOut
    );

PRIVATE void MCT CbsWakeUpOnCondition(
    PISCHEDULER This,
    PCONDITION pCond,
    BOOL fAll
    );

PRIVATE SCODE MCT CbsConditionTimedWaitAndBeginConstraint(
    PISCHEDULER This,
    PCONDITION pCondition,
    PMUTEX pMutex,
    PTIME pTimeout,
    BOOL EndPrevious,
    PTIME_CONSTRAINT pTimeConstraint,
    PTIME pTimeTaken,
    WAITCOND_CALLBACK WaitCallback
    );

PRIVATE TIME MCT CbsGetIdleTime(
    PISCHEDULER This
    );

/* ===========================================================================
 * Constraint-based scheduler vtable.
 */

PRIVATE CONST_VTBL struct ISchedulerVtbl CbsVtbl = {
  __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
  CbsQueryInterface,
  CbsAddRef,
  CbsRelease,
  CbsAddThread,
  CbsRemoveThread,
  CbsDisablePreemption,
  CbsEnablePreemption,
  CbsPreemptionIsEnabled,
  CbsWaitForMutex,
  CbsWakeUpOnMutex,
  CbsWaitForCondition,
  CbsWakeUpOnCondition,
  CbsConditionTimedWaitAndBeginConstraint,
  CbsGetIdleTime,
};

/* Scheduler object */
PRIVATE CBS Cbs;

/* Include machine dependent operations & defines */
#include "_schedul.c"

IScheduler* SchedulerCreate(void)
{
  Cbs.v = &CbsVtbl;
  Cbs.PreemptionDisableCount = 0;
  Cbs.ContextSwitchNeeded = FALSE;
  CbsReadyQueueInitialize();
  CbsSleepQueueInitialize();
  Cbs.pkcFreeStackHead = NULL;
  Cbs.RefCount = 0;
  Cbs.nThreads = 0;
  Int32ToInt64(Cbs.IdleTime, 0);
  Cbs.FirstThreadBooted = FALSE;
  Log_Init(pSchedLog, SCHEDULER_LOG_SIZE, SchedulerLog);
  return (PISCHEDULER)&Cbs;
}

/* ===========================================================================
 * Private scheduler functions
 */

PUBLIC TIME SleepOverhead = Int64Initializer(0, 0);

/* Insert thread in sleep queue, sorted by urgency.
 * If the Thread's state is not waiting, it is set to Sleeping.
 */
PRIVATE void
CbsMakeThreadReady(
    PTHDINFO pThdToMakeReady,
    BOOL PreemptCurrentThread
    );

/* SCODE */
/* SleepUntil( */
/*     TIME tSleep */
/*     ); */

PRIVATE void
CbsUpdateRestartBy(
    PKCONSTRAINT pkcUpdate
    );

PRIVATE void
CbsConditionWakeupTransform(
    PTR pkcWaiter
    );

PRIVATE void
CbsResolvePendingConstraint(
    PKCONSTRAINT pkcCurrent,
    BOOL IsRunning
    );

PRIVATE PKCONSTRAINT
CbsRescheduleInterrupt(
    PKCONSTRAINT pkcOld,
    PKCONSTRAINT pkcNew
    );

PRIVATE PKCONSTRAINT 
CbsPreemptIfNecessary(
    PTIME pNextTimeSlice
    );

PRIVATE INLINE void 
CbsSleepQueueAddConstraint(
    PKCONSTRAINT pkcNew
    )
{
  PTHDINFO pThdNew;
  assert(pkcNew != NULL);
  pThdNew = (PTHDINFO)KConstraint_GetSchedObject(pkcNew);
  assert(pThdNew != NULL);
  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);

  /* Make sure we are not already in sleep queue */
  assert(List_FindData(CbsGetSleepQueue(), pkcNew) == NULL);

  /* Insert pThdNew before the first thread with a later start time. */
  List_InsertSortedDescending(CbsGetSleepQueue(), KConstraint_UrgencyGT,
                              KConstraint_GetSchedQueueNode(pkcNew));
  Thread_SetState(pThdNew, THREAD_SLEEPING);
  Log_WriteEntry(pSchedLog, LOG_PUT_SLEEP_QUEUE, (UINT)pkcNew);
}

/* Changes the given timeout to be absolute and in the present or future. */
PRIVATE INLINE void
CbsNormalizeTimeout(
    PTIME tTimeout
    )
{
  TIME tNow = CurrentTime();

  /* if Timeout is relative to current time, make it absolute.
   * This guarantees a greater-or-equal time than now. */
  if (TimeIsNotPositive(*tTimeout)) {
    INT64SUB(*tTimeout, tNow, *tTimeout);
  }
  /* if the Timeout time has already occurred, use current time */
  else if (TimeLess(*tTimeout, tNow)) {
    *tTimeout = tNow;
  }
 /* Account for overheads, subtracting them from *Timeout  */
  INT64SUB(*tTimeout, *tTimeout, SleepOverhead);
}

/* Move all threads whose start time has arrived from sleep queue to
 * ready queue.  Return start time of earliest thread.
 * If there is no earliest thread, return the timesharing timeslice so that
 * non-real-time threads can still run.
 */
PRIVATE INLINE TIME
CbsWakeStartingThreads(void)
{
  TIME tUntilStart;
  TIME tNow;
  PTHDINFO pThdMostUrgent;
  PKCONSTRAINT pkcMostUrgent;

  Log_WriteEntry(pSchedLog, LOG_ENTER_WAKE_THREADS, (UINT)NULL);

  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);
  assert(CheckStackOverflow(CurrentThread()));

  if (!CbsSleepQueueIsEmpty()) {

    tNow = CurrentTime();

    do {
      /* Get most urgent sleeping thread */
      pkcMostUrgent = CbsSleepQueueFindMostUrgentConstraint();
      pThdMostUrgent = (PTHDINFO)KConstraint_GetSchedObject(pkcMostUrgent);
      /* We are waking up a waiting thread, which may or may not be idling */

      assert(Thread_IsIdle(pThdMostUrgent) ||
             Thread_IsSleeping(pThdMostUrgent));

      /* Compute time until thread must start. */
      INT64SUB(tUntilStart, KConstraint_GetRestartBy(pkcMostUrgent), tNow);

      if (TimeIsNotPositive(tUntilStart))  {
        (void) CbsSleepQueueRemoveMostUrgentConstraint();
        Log_WriteEntry(pSchedLog, LOG_POP_SLEEP_QUEUE, (UINT)pkcMostUrgent);

        if (Thread_IsWaitingForCondition(pThdMostUrgent)) {
          Condition_WaiterQueueRemoveConstraint
            (Thread_GetWaitObject(pThdMostUrgent), pkcMostUrgent);
          Thread_SetLastError(pThdMostUrgent, E_TIMED_OUT);
          Condition_TimedOut(Thread_GetWaitObject(pThdMostUrgent));

          /*        KConstraintStopInheritance(pkcMostUrgent); */
        }
        assert(Node_IsAlone(KConstraint_GetWaitQueueNode(pkcMostUrgent)));
        CbsMakeThreadReady((PTHDINFO)pThdMostUrgent, FALSE);
        assert(Thread_IsReady(pThdMostUrgent));
      } else {
        /* Thread queued is sorted by increasing starting time, so we only
         * need to check if the first thread's time has arrived.
         *
         * If no one is runnable then TheCurrentThread can keep
         * on going until the wakeup time of pThdMostUrgent.  Otherwise
         * we must cover for the time-sharing case: if the most
         * urgent thread on the run queue is not a RT thread,
         * IThreadUrgencyGTEQ() will not trim down TimeUntilStart.
         * Therefore we have to do it here (conservatively).
         */
        /*        if (!CbsReadyQueueIsEmpty() && */
        /*            (TimeLess(TIME_SHARING_TIMESLICE, tUntilStart))) { */
        /*          tUntilStart = TIME_SHARING_TIMESLICE; */
        /*        } */
        /*        printf("Waking\n"); */
        /*        ListTransformData(CbsGetSleepQueue(), Transform); */
        Log_WriteEntry(pSchedLog, LOG_EXIT_WAKE_REALTIME, Int64ToInt32(tUntilStart));

        return tUntilStart;
      }
    } while (!CbsSleepQueueIsEmpty());
  }

  Log_WriteEntry(pSchedLog, LOG_EXIT_WAKE_TIMESHARE,
                 Int64ToInt32(TIME_SHARING_TIMESLICE));
  return TIME_SHARING_TIMESLICE;
}

/* PRIVATE PKCONSTRAINT CbsGetFreeConstraint(); */

PRIVATE INLINE SCODE
CbsBeginConstraintBeforeWaitValidate(
    PKCONSTRAINT pkc,
    PTIME_CONSTRAINT ptc
    )
{
  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);
  assert(pkc != NULL);

  /* Cannot end a constraint if it does not have a surrounding constraint. */

  if (TConstraint_GetEndPrevious(ptc) && !KConstraint_HasStackNext(pkc)) {
    return E_INVALID_PARAMETER;
  }

  /* The only valid criticalities are 0 and 1 */
  /*      if (TConstraint_GetCriticality(ptc) & ~CRITICAL) { */
  /*        return E_INVALID_PARAMETER; */
  /*      } */

  /* The estimated time must be nonnegative */
  if (TimeIsRelative(TConstraint_GetEstimateTime(ptc))) {
    return E_INVALID_PARAMETER;
  }

  return S_OK;
}

PRIVATE INLINE void
CbsBeginConstraintBeforeWait(
    PKCONSTRAINT pkc,
    PTIME_CONSTRAINT ptcNew
    )
{
  TIME tStart;

  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);
  assert(pkc != NULL);

  tStart = TConstraint_GetStartTime(ptcNew);

  if (TimeIsRelative(tStart)) {

    TIME tCurrent = CurrentTime();

    INT64SUB(tStart, tCurrent, tStart);

    /* Use TIME_ORIGIN upon underflow. */
    if (TimeLess(tStart, tCurrent)) {
      TConstraint_SetStartTime(ptcNew, TIME_ORIGIN);
    }
  }

  KConstraint_SetPending(pkc, ptcNew);
}

PRIVATE INLINE SCODE
CbsBeginConstraintAfterWait(
    PKCONSTRAINT pkcCurrent
    )
{
  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);

  assert(pkcCurrent != NULL);

  /* If the thread still has a pending constraint, then resolve it.
   * This only happens if we never waited or slept
   */
  if (KConstraint_HasPending(pkcCurrent)) {
    CbsResolvePendingConstraint(pkcCurrent, TRUE);
    pkcCurrent =
      Thread_GetCurrentConstraint(KConstraint_GetSchedObject(pkcCurrent));
  }

  /* The thread is running with a constraint (not pending).
   */
  assert(Thread_IsRunning(KConstraint_GetSchedObject(pkcCurrent)));
  assert(KConstraint_HasPending(pkcCurrent) == FALSE);
  assert(!TimeEqual(KConstraint_GetLastRefresh(pkcCurrent), TIME_ORIGIN));

  /* If our restart by time has passed, then return probable failure */

  if (TimeLess(KConstraint_GetRestartBy(pkcCurrent), CurrentTime())) {
    return S_OUTATIME;
  }

  return S_OK;
}

/* Allocate a constraint.  Try for the free list first, then the heap.
 */
PRIVATE INLINE PKCONSTRAINT
CbsAllocConstraint(void)
{
  PKCONSTRAINT pkcNew;

  /* To prevent the free list from growing unnecessarily,
   * first try to allocate from it. In the common case,
   * we are allocating the constraint for a new thread,
   * and when user threads die their constraints get
   * into the free list, so in steady state we will
   * recycle constraints from the free list.
   */
  pkcNew = Cbs.pkcFreeStackHead;
    
  /*     pkcNew = (PKCONSTRAINT) AtomicLIFORemove((PTR*)&Cbs.pkcFreeStackHead); */
  if (pkcNew == NULL) {
    PIHEAP pHeap = CurrentHeap();

    pkcNew = (PKCONSTRAINT)pHeap->v->Alloc(pHeap, HEAP_DEBUG_NAME("cbs"), 
                                           sizeof(KCONSTRAINT), 0);
  }
  else {
    Cbs.pkcFreeStackHead = KConstraint_Pop(Cbs.pkcFreeStackHead);
  }
  assert(pkcNew != Cbs.pkcFreeStackHead);

  return pkcNew;
}

PRIVATE INLINE void
CbsRecycleConstraintStack(
    PKCONSTRAINT pkcStackHead
    )
{
  PKCONSTRAINT pkcNext;

  Log_WriteEntry(pSchedLog, LOG_CONSTRAINT_STACK_FREE, (UINT)CurrentThread());
  while (pkcStackHead != NULL) {
    pkcNext = KConstraint_GetStackNext(pkcStackHead);
    KConstraint_SetStackNext(pkcStackHead, NULL);
    CbsRecycleConstraint(pkcStackHead);
    pkcStackHead = pkcNext;
  }
}

PRIVATE INLINE void 
CbsFreeConstraintStack(
    PKCONSTRAINT pkcStackHead
    )
{
  PKCONSTRAINT pkcNext;

  Log_WriteEntry(pSchedLog, LOG_CONSTRAINT_STACK_FREE, (UINT)CurrentThread());
  while (pkcStackHead != NULL) {
    pkcNext = KConstraint_GetStackNext(pkcStackHead);
    KConstraint_SetStackNext(pkcStackHead, NULL);
    CbsFreeConstraint(pkcStackHead);
    pkcStackHead = pkcNext;
  }
}

/* ===========================================================================
 * Interface method implementations
 */

PRIVATE SCODE MCT
CbsQueryInterface( 
    PISCHEDULER pThis,
    /* [in] */ REFIID pIid,
    /* [out] */ void **ppUnk
    )
{
  return GenericQueryInterface((IUnknown *)pThis, pIid, ppUnk, &IID_IThread);
}
        
PRIVATE UINT MCT 
CbsAddRef(
    PISCHEDULER pThis
    )
{
  UINT RefCnt;

  RefCnt = AtomicInc(&Cbs.RefCount);

  return RefCnt;
}
        
PRIVATE UINT MCT
CbsRelease (
    PISCHEDULER pThis
    )
{
  UINT RefCnt;

  RefCnt = AtomicDec(&Cbs.RefCount);

  return RefCnt;
}

PRIVATE PITHREAD MCT
CbsReschedule(
    IScheduler *Scheduler,
    PITHREAD pThdCurrent
    )
{
  TIME tEarliestStart;
  PKCONSTRAINT pkcMostUrgent;
  PKCONSTRAINT pkcReschedule;
  /*    UINT OldState; */

  /*     PTHDINFO pThdCurrent = (PTHDINFO)CurrentThread(); */

  assert(pThdCurrent != NULL);
  /*    OldState = Thread_GetState(pThdCurrent); */

  pkcReschedule = Thread_GetCurrentConstraint(pThdCurrent);
  Log_WriteEntry(pSchedLog, LOG_RESCHEDULE, (UINT)pkcReschedule);

 Again:
  /* If there are any deferred condition signals, process them all. */
  if (DeferredConditionSignals) {
    Log_WriteEntry(pSchedLog, LOG_DRAIN_DEFERRED, (UINT)pThdCurrent);
    Condition_DrainDeferred();
  }

  /* Move all threads whose start times have passed from the sleep queue
   * to the ready queue and return time until earliest starting thread */
  tEarliestStart = CbsWakeStartingThreads();

  /* Preempt the current thread if necessary due to a ready thread
   * now being more urgent than the current thread. This can happen
   * either because a more urgent thread woke up, or because a
   * ConditionSignal changed the relative urgencies of the
   * current and some previously ready thread. */
  pkcMostUrgent = CbsPreemptIfNecessary(&tEarliestStart);

  /* Don't preempt the IdleThread, but do set the timer. */

  if (SetNextInterrupt(tEarliestStart,
                       &pCBS(Scheduler)->ContextSwitchNeeded) &&
      !Thread_IsIdle(pThdCurrent)) {
    pkcReschedule = CbsRescheduleInterrupt(pkcReschedule,
                                           pkcMostUrgent);
    if (DeferredConditionSignals) {
       /* set this back the way it was */
            pCBS(Scheduler)->ContextSwitchNeeded = TRUE;
           goto Again;
    }

  }
  /* We emerge from the scheduler as the current thread */
  Log_WriteEntry(pSchedLog, LOG_OUT, (UINT)pkcReschedule);
  return (PITHREAD)KConstraint_GetSchedObject(pkcReschedule);
}

/* Disables preemption, or if recursion (nested disables) is TRUE,
 * increments the preemption disable count.
 * Returns TRUE if preemption was already disabled, FALSE otherwise. */
PRIVATE BOOL MCT CbsDisablePreemption(
    IScheduler *This,
    BOOL Recursive
    )
{
  UINT currentDisableCount;
    
  currentDisableCount = AtomicInc(&(pCBS(This)->PreemptionDisableCount));

  assert(Recursive || (currentDisableCount == 1));

  return (currentDisableCount != 1);
}

/* Enables preemption, or if recursion (nested disables) is TRUE,
 * decrements the preemption disable count.
 * Returns TRUE if preemption is now enabled, FALSE otherwise.
 */
PRIVATE BOOL MCT CbsEnablePreemption(
    IScheduler *This,
    BOOL Recursive
    )
{
  UINT currentDisableCount;

  currentDisableCount = AtomicDec(&(pCBS(This)->PreemptionDisableCount));

  assert(Recursive || (currentDisableCount == 0));

  if ((currentDisableCount == 0) && pCBS(This)->ContextSwitchNeeded) {
    Reschedule();
  }

  return (currentDisableCount == 0);
}

PRIVATE TIME MCT CbsGetIdleTime(
    IScheduler *This
    )
{
  return (pCBS(This)->IdleTime);
}

PRIVATE BOOL MCT CbsPreemptionIsEnabled(
    IScheduler *This
    )
{
  return (pCBS(This)->PreemptionDisableCount == 0);
}

/* Adds the given thread to the given scheduler */
void MCT CbsAddThread(
    IScheduler *This,
    IThread *pThd,
    PTIME_CONSTRAINT pcval)
{
  PKCONSTRAINT pkcNew;
    
  Thread_Prologue();

  CbsDisablePreemption(This, FALSE);
  pkcNew = CbsAllocConstraint();
  KConstraint_Init(pkcNew, pcval);
  KConstraint_SetSchedObject(pkcNew, pThdArg);
  Thread_SetCurrentConstraint(pThdArg, pkcNew);

  /* Boot problem. Make sure we reschedule later, not now */
  pCBS(This)->nThreads++;
  CbsMakeThreadReady((PTHDINFO)pThdArg, FALSE);

  CbsEnablePreemption(This, FALSE);
  RescheduleNeeded();           /* late for boot */
  Log_WriteEntry(pSchedLog, LOG_ADD_THREAD, (UINT)pThdArg);
  Log_WriteEntry(pSchedLog, LOG_ADD_CONSTRAINT, (UINT)pkcNew);
}

/* Removes the given thread from the given scheduler */
void MCT CbsRemoveThread(
    IScheduler *This,
    IThread *pThd
    )
{
  UINT IntrState;
  PKCONSTRAINT pkcNew;
  PKCONSTRAINT pkcRemove;
  PTHDINFO pThdNew;
  PCXTINFO pProperContext;
  PTR pWaitObject;

  Thread_Prologue();

  pkcRemove = Thread_GetCurrentConstraint(pThdArg);

  /* We can only deal with suicide
   * [would just need a bunch more code otherwise]
   */
  assert(CurrentThread() == (PITHREAD)pThdArg);

  CbsDisablePreemption(This, FALSE);

  /* If this is the last thread then the game is over */
  if (--pCBS(This)->nThreads == 0) {
    CbsFreeConstraintStack(Cbs.pkcFreeStackHead);
    BaseDelete();
  }

  Log_WriteEntry(pSchedLog, LOG_DEAD_THREAD, (UINT)pThdArg);

  Thread_SetState(pThdArg, THREAD_DEAD);

  /* For suicide, must switch to another thread to reclaim stack. */
  pkcNew = CbsRescheduleInterrupt(Thread_GetCurrentConstraint(pThdArg),
                                  NULL);
  pThdNew = (PTHDINFO)KConstraint_GetSchedObject(pkcNew);

  /* Remove the dying threads from any waiter queues. */
  pWaitObject = Thread_GetWaitObject(pThdArg);

  if (pWaitObject != NULL) {
    /*     if (ThreadIsWaitingForCondition(pThdArg)) { */
    /* This will always be a condition, since the current thread is never
     * waiting on a mutex */
    Condition_WaiterQueueRemoveConstraint(pWaitObject, pkcRemove);
    /*        KConstraintStopInheritance(pkcRemove); */
  }

  /* Either way, we are now running in a 'safe' thread context.
   * [possibly on the wrong stack though]. */

  /* We cannot free memory while holding the scheduler lock.
   * If we stole a thread, make sure it returns where it should.
   * We might get preempted before we are done with this code. */
  pProperContext = Thread_GetContext(pThdNew);

  CbsEnablePreemption(This, FALSE);

  /* Release the dead thread's resources.
   * This must be guaranteed not to take e.g. the heap mutex,
   * because it might be held by the thread we stole --> deadlock.
   * Note that pkcRemove no longer equals what it did above due to
   * a context switch, so we have to explicitly get a valid copy
   * from the dying thread.
   */
  Thread_UnlockMutexes((PTHDINFO)pThdArg);
  CbsFreeConstraintStack(Thread_GetCurrentConstraint(pThdArg));

  (PITHREAD)pThdArg->v->Release((PITHREAD)pThdArg); /* BUGBUG stack! */

  /* If this was a suicide, special handling.  */
  TURN_INTERRUPTS_OFF(IntrState);
  Thread_SetContext(pThdNew, pProperContext);
  LoadThreadContext((PITHREAD)pThdNew);
  /* Execution should never reach this point */
  assert(FALSE);
#ifdef _DEBUG
  DebugBreak();
#endif
}

/* Wake all threads blocked on mutex pMutex. This mutex must have an
 * owner (caught by assertion in StopConstraintInheritance).
 */
void MCT CbsWakeUpOnMutex(
    IScheduler *This,
    PMUTEX pMutex
    )
{
  PTHDINFO pThdMostUrgent;
  PTHDINFO pThdCurrent;
  PKCONSTRAINT pkcMostUrgent;
  PKCONSTRAINT pkcCurrent;
  Mutex_Prologue();

  UnusedParameter(This);

  pThdCurrent = (PTHDINFO)CurrentThread();
  pkcCurrent = Thread_GetCurrentConstraint(pThdCurrent);

  assert(pMutexArg != NULL);
  /* If waiter queue is empty, Mutex_Lock should not have called us */
  assert(!Mutex_WaiterQueueIsEmpty(pMutexArg));

  /* Remove most urgent thread on the mutex's waiter queue,
   * set its owner to NULL, and recalculcate inheritance. */
  pkcMostUrgent = Mutex_WaiterQueueRemoveMostUrgentConstraint(pMutexArg);
  assert(pkcMostUrgent != NULL);

  /* The old mutex owner is now free from any inherited constraints
   * since it no longer owns this mutex.
   */
  Urgency_Copy(KConstraint_GetCurrentUrgency(pkcCurrent),
               KConstraint_GetLocalUrgency(pkcCurrent));

  pThdMostUrgent = (PTHDINFO)KConstraint_GetSchedObject(pkcMostUrgent);
  assert(pThdMostUrgent != NULL);

  assert(Thread_IsWaitingForMutex(pThdMostUrgent));
  Mutex_SetSchedObject(pMutexArg, pThdMostUrgent);
  if (!Mutex_WaiterQueueIsEmpty(pMutexArg)) {
    Mutex_SetTaken(pMutexArg);
  }
  Thread_MutexListAdd(pThdMostUrgent, pMutexArg);
  KConstraint_RecalculateInheritance(pkcMostUrgent);

  CbsMakeThreadReady((PTHDINFO)pThdMostUrgent, TRUE);
}

void MCT CbsWakeUpOnCondition(
    IScheduler *This,
    PCONDITION pCond,
    BOOL fAll
    )
{
  Condition_Prologue();

  assert(pCond != NULL);
  /* If waiter queue is empty, UnwaitCond should not have called us */
  assert(!Condition_WaiterQueueIsEmpty(pCondArg));

  Log_WriteEntry(pSchedLog, LOG_ENTER_CONDITION_SIGNAL, (UINT)pCondArg);

  if (fAll == TRUE) {
    List_TransformData(Condition_GetWaiterQueue(pCondArg),
                              CbsConditionWakeupTransform);
    List_Init(Condition_GetWaiterQueue(pCondArg));
  }
  else {
    PKCONSTRAINT pkcMostUrgent;

    pkcMostUrgent =
      Condition_WaiterQueueRemoveMostUrgentConstraint(pCondArg);
    assert(List_FindData(Condition_GetWaiterQueue(pCondArg),
                         pkcMostUrgent) == NULL);
    CbsConditionWakeupTransform(pkcMostUrgent);
    Log_WriteEntry(pSchedLog, LOG_ENTER_CONDITION_SIGNAL,
                   (UINT)pkcMostUrgent);
  }      
  CbsPreemptIfNecessary(NULL);
}

void MCT
CbsWaitForMutex(
    PISCHEDULER This,
    PMUTEX pMutex
    )
{
  PKCONSTRAINT pkcWaiter;
  PITHREAD pThdWaiter = CurrentThread();
  Mutex_Prologue();

  UnusedParameter(This);

  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);

  pkcWaiter = Thread_GetCurrentConstraint(pThdWaiter);

  assert(pMutexArg != NULL);
  assert(Thread_IsRunning(pThdWaiter));

  /* Add constraint to mutex's waiter queue and start inheritance */
  Mutex_WaiterQueueAddConstraint(pMutexArg, pkcWaiter);
  KConstraint_StartInheritance(Thread_GetCurrentConstraint(
                                   Mutex_GetSchedObject(pMutexArg)),
                               KConstraint_GetCurrentUrgency(pkcWaiter));
  Thread_SetWaiting(pThdWaiter, THREAD_WAIT_MUTEX, pMutexArg);
  RescheduleNeeded();
}

void MCT
CbsWaitForCondition(
    PISCHEDULER This,
    PCONDITION pCond,
    PTIME pTimeOut
    )
{
  PKCONSTRAINT pkcWaiter;
  PITHREAD pThdWaiter;

  Condition_Prologue();

  UnusedParameter(This);
  pThdWaiter = CurrentThread();

  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);
  assert(pCondArg != NULL);
  assert(Thread_IsRunning(pThdWaiter));

  pkcWaiter = Thread_GetCurrentConstraint(pThdWaiter);

  Log_WriteEntry(pSchedLog, LOG_ENTER_CONDITION_WAIT, (UINT)pkcWaiter);
  assert(Node_IsAlone(KConstraint_GetWaitQueueNode(pkcWaiter)));

  CbsUpdateRestartBy(pkcWaiter);
  Condition_WaiterQueueAddConstraint(pCondArg, pkcWaiter);
  Thread_SetWaiting(pThdWaiter, THREAD_WAIT_CONDITION, pCondArg);

  if (pTimeOut != NULL) { /* Place in sleep queue */
            if (!TimeEqual(*pTimeOut, TIME_FOREVER)) { 
    CbsNormalizeTimeout(pTimeOut);
    KConstraint_SetRestartBy(pkcWaiter,*pTimeOut);
    Urgency_Copy(KConstraint_GetCurrentUrgency(pkcWaiter),
                 KConstraint_GetLocalUrgency(pkcWaiter)); 
    CbsSleepQueueAddConstraint(pkcWaiter);
    assert(List_FindData(CbsGetSleepQueue(), pkcWaiter) == KConstraint_GetSchedQueueNode(pkcWaiter));
    assert(Thread_IsWaitingForTimeout(pThdWaiter));
         } 
  }
  assert(!Thread_IsRunning(pThdWaiter));
  RescheduleNeeded();
  Log_WriteEntry(pSchedLog, LOG_EXIT_CONDITION_WAIT, (UINT)pkcWaiter);
}

SCODE MCT
CbsConditionTimedWaitAndBeginConstraint(
    IScheduler *This,
    PCONDITION pCond,
    PMUTEX pMutex,
    TIME *pTimeout,
    BOOL bEndPrevious,
    PTIME_CONSTRAINT ptcNew,
    PTIME ptTimeTaken,
    WAITCOND_CALLBACK WaitCallback
    )
{
  SCODE sc;
  SCODE sc2;
  PKCONSTRAINT pkcCurrent;
  PTHDINFO pThdCurrent;

  Condition_Prologue();
  Mutex_Prologue();

  if (bEndPrevious) {
    TConstraint_SetEndPrevious(ptcNew);
  }
  else {
    TConstraint_ClearEndPrevious(ptcNew);
  }

  pThdCurrent = (PTHDINFO)CurrentThread();
  pkcCurrent = Thread_GetCurrentConstraint(pThdCurrent);

  /*      if (CbsCheckFreeConstraint(pThd)) { */
  /*        return pTH(pThd)->LastError; */
  /*      } */

  CbsDisablePreemption(This, FALSE);

  /* Do pre-wait BeginConstraint validation */
  sc = CbsBeginConstraintBeforeWaitValidate(pkcCurrent, ptcNew);

  if (sc != S_OK) {
    /* we have not released the mutex, so don't retake it */
    goto Exit;
  }

  pMutexArg = NULL;
  (*WaitCallback)((PCONDITION)pCondArg, (PMUTEX)pMutexArg, pTimeout);
  
  /* Validate user-level time constraint before the yield. */
  CbsBeginConstraintBeforeWait(pkcCurrent, ptcNew);
  
  /* Now enable pre-emption to block, then redisable */
  CbsEnablePreemption(This, FALSE);
  /* This is the yield */
  CbsDisablePreemption(This, FALSE);
  
  sc = Thread_GetLastError(pThdCurrent);
  
  /* Do after-yield processing */
  
  sc2 = CbsBeginConstraintAfterWait(pkcCurrent);
  pkcCurrent = Thread_GetCurrentConstraint(pThdCurrent);
  CbsUpdateRestartBy(pkcCurrent);
  
  /* Return time taken if so desired. */
  if (ptTimeTaken != NULL) {
    /* RestartBy time advances when the thread runs, so use difference */
    INT64SUB(*ptTimeTaken,
             CurrentTime(),
             /*               KConstraintGetRestartBy(pkcCurrent), */
             KConstraint_GetInitialRestartBy(pkcCurrent));
  }
  
    /* Return cond wait status, unless constraint failure is predicted */
  assert(sc2 == S_OK || sc2 == S_OUTATIME);
  if (sc2 == S_OUTATIME) {
    sc = S_OUTATIME;
  }
  
 Exit:

  CbsEnablePreemption(This, FALSE);

  /* Must return with the mutex locked in *all* cases */
  if (pMutexArg != NULL) {
    Mutex_Lock((PMUTEX)pMutexArg);
  }

  return sc;
}


/* ===========================================================================
 * Implementation of private functions.
 */


#ifdef _MSC_VER
#pragma warning(disable:4127)  /* conditional expression is constant */
#pragma warning(disable:4702)  /* unreachable code */
#endif

/* If (pkcNew != NULL) add pkcOld to sleep queue, make pkcNew ready to run.
 * Else if (pkcNew == NULL && pkcOld still running) do nothing.
 * Else if (pkcNew == NULL && pkcOld not running) make most urgent constraint
 *   ready to run, or spin idle thread while ready queue is empty.
 * Returns the next running constraint; never returns NULL because of
 *   spinning idle thread.
 * This rescheduling path is 'synchronous',
 * i.e. we get here if and only if RescheduleNeeded() is called.
 */
PRIVATE PKCONSTRAINT CbsRescheduleInterrupt(
    PKCONSTRAINT pkcOld,
    PKCONSTRAINT pkcNew
    )
{
  PTHDINFO pThdOld;
  PTHDINFO pThdNew;

  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);
  Log_WriteEntry(pSchedLog, LOG_RESCHEDULE_MAIN, (UINT)pkcOld);

  /* The old thread exists (either running or about to die) */
  assert(pkcOld != NULL);
  pThdOld = (PTHDINFO)KConstraint_GetSchedObject(pkcOld);

  /*      printf("ReadyQueue\n"); */
  /*      ListTransformData(CbsGetReadyQueue(), Transform); */

  if (pkcNew == NULL) {
    if (CbsReadyQueueIsEmpty()) {
      /* there is no other ready thread to switch to */
      PCXTINFO pSavedCxt;
      /*UINT OldState;*/
      TIME tStartIdle;

      /* If current thread can keep running, then let it */
      if (Thread_IsRunning(pThdOld)) {
        return pkcOld;
      }

      /* Begin playing idle thread and set time to come out of it. */
      /*      CbsUpdateRestartBy(pkcOld); */
      /*      RuntimeWasUpdated = TRUE; */
      tStartIdle = CurrentTime();

      /* Save state in case we get preempted by a new ready thread */
      Thread_SetIdle(pThdOld);
      /*OldState = Thread_GetState(pThdOld);*/
      /*Thread_SetState(pThdOld, THREAD_IDLE);*/
      pSavedCxt = Thread_GetContext(pThdOld);
      Log_WriteEntry(pSchedLog, LOG_IDLE_THREAD, (UINT)pkcOld);

      CbsEnablePreemption(pTheScheduler, FALSE);
      /* CbsReschedule won't preempt here, just add new ready threads to
       * the queue */

      /* Spin while the ready queue is empty. */
      while (CbsReadyQueueIsEmpty()) {
#ifdef MACHINE_IDLE
        MACHINE_IDLE(CbsReadyQueueIsEmpty());
#endif
      }

      CbsDisablePreemption(pTheScheduler, FALSE);
      Log_WriteEntry(pSchedLog, LOG_END_IDLE_THREAD, (UINT)pkcOld);

      Thread_SetContext(pThdOld, pSavedCxt);
      Thread_ClearIdle(pThdOld);
      /*        if (Thread_IsIdle(pThdOld)) {
                Thread_SetState(pThdOld, OldState);
                }
      */
      /* Update scheduler idle time and LastRefresh time of old constraint */
      {
        TIME tNow = CurrentTime();
        INT64SUB(tStartIdle, tNow, tStartIdle);
        INT64ADD(Cbs.IdleTime, Cbs.IdleTime, tStartIdle);
        KConstraint_SetLastRefresh(pkcOld, tNow);
      }
    }
    pkcNew = CbsReadyQueueRemoveMostUrgentConstraint();
    Log_WriteEntry(pSchedLog, LOG_POP_READY_QUEUE, (UINT)pkcNew);
  }
  else {
    List_Remove(CbsGetReadyQueue(), KConstraint_GetSchedQueueNode(pkcNew));
  }

  assert(Node_IsAlone(KConstraint_GetSchedQueueNode(pkcNew)));
  /*     assert(NodeIsAlone(KConstraintGetSchedQueueNode(pkcOld))); */

  pThdNew = (PTHDINFO)KConstraint_GetSchedObject(pkcNew);

  Log_WriteEntry(pSchedLog, LOG_END_IDLE_THREAD, (UINT)pkcNew);

  /* The new thread exists, has not died, and is ready to run.  */
  assert(pThdNew != NULL);
  assert(!Thread_IsDead(pThdNew));
  assert(Thread_IsReady(pThdNew));
  assert(Thread_GetWaitObject(pThdNew) == NULL);
  assert(Node_IsAlone(KConstraint_GetSchedQueueNode(pkcNew)));
  assert(Node_IsAlone(KConstraint_GetWaitQueueNode(pkcNew)));

  Thread_SetState(pThdNew, THREAD_RUNNING);

  /* If the old and new threads are different, then do context switching */
  if (pThdNew != pThdOld) {

    /* Then put it back on the ready queue
     */
    if (Thread_IsRunning(pThdOld)) {

      /* Update the RestartBy of the old thread. */

      CbsUpdateRestartBy(pkcOld);
      CbsReadyQueueAddConstraint(pkcOld);
      Thread_SetState(pThdOld, THREAD_READY);
    }

    /* Save current time for future UpdateRestartBy calls on new thread. */
    KConstraint_SetLastRefresh(pkcNew, CurrentTime());

    /*                                KConstraintGetLastRefresh(pkcOld)); */

    /* Make the new thread current, switching context if necessary */
    ThreadSetCurrent((PITHREAD)pThdNew);
  }

  return pkcNew;
}

/* Returns the constraint belonging to the most urgent ready, non-running
 * thread. If there are no ready threads, no ready threads
 * more urgent than the currently running thread, or the
 * currently running thread is already rescheduling, returns NULL.
 */
PRIVATE PKCONSTRAINT CbsPreemptIfNecessary(
    PTIME pNextTimeSlice
    )
{
  PTHDINFO pThdCurrent;
  PKCONSTRAINT pkcCurrent;
  PKCONSTRAINT pkcNext = NULL;

  pThdCurrent = (PTHDINFO)CurrentThread();
  pkcCurrent = Thread_GetCurrentConstraint(pThdCurrent);

  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);

  /* Watch out for threads that are already dead. */
  assert(pThdCurrent != NULL);

  if (Thread_IsDead(pThdCurrent)) {
    /* If the current thread is dead then we should reschedule.
     * However, when death occurs the thread will look for
     * an undertaker, spinning IdleThread while none is found.
     * Which is the case we catch here.
     * Therefore we say the opposite.  Obviously ;-))
     */
    RescheduleNotNeeded();
    Log_WriteEntry(pSchedLog, LOG_PREEMPT_DEAD, (UINT)pkcCurrent);
    return NULL;
  }

  /* If the current thread is already rescheduling, then we are done. */
  if (!Thread_IsRunning(pThdCurrent)) {
    Log_WriteEntry(pSchedLog, LOG_PREEMPT_RESCHEDULING, (UINT)pkcCurrent);
    return NULL;
  }

  /* If there are no other runnable threads, then we are done. */
  if (CbsReadyQueueIsEmpty()) {
    Log_WriteEntry(pSchedLog, LOG_PREEMPT_NO_OTHERS, (UINT)pkcCurrent);
    return NULL;
  }

  /* Fetch the most urgent ready (runnable but not running) thread. */
  pkcNext = CbsReadyQueueFindMostUrgentConstraint();
  assert(pkcNext != NULL);

  Log_WriteEntry(pSchedLog, LOG_PREEMPT, (UINT)pkcNext);
    
  /* Refresh the constraint of the currently running thread before
   * we start comparing it to the constraints of the ready threads.
   */
  pkcCurrent = Thread_GetCurrentConstraint(pThdCurrent);
  CbsUpdateRestartBy(pkcCurrent);

  /* If current thread remains the most urgent runnable thread,
   * then do not preempt.
   * If the current thread is not less urgent than the most urgent
   * runnable thread, then we let it keep running.
   */
  if (KConstraint_UrgencyGTEQ(pkcCurrent, pkcNext)) {
    Log_WriteEntry(pSchedLog, LOG_PREEMPT_STILL_URGENT, (UINT)pkcCurrent);
    return NULL;
  }

  /* We have decided to preempt the current thread. */
  /*      List_Remove(CbsGetReadyQueue(), KConstraintGetSchedQueueNode(pkcNext)); */
  /*      ThreadClearWaiting(pThdCurrent); */
  RescheduleNeeded();

  /* All cases must end here, to turn off the timer. */
  return pkcNext;
}

/* make ready from waiting, sleeping... anything besides ready or running */
PRIVATE void CbsMakeThreadReady(
    PTHDINFO pThd,
    BOOL bPreemptCurrentThread)
{
  PKCONSTRAINT pkcReady;
  PNODE pnTemp;

  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);

  if (pThd == NULL) {
    return;
  }

  /* The thread might have died while we were playing with it.
   * For example, the helper thread is preemptible
   * while it does things for other threads, so it can be preempted
   * and then they can be terminated.
   */
  if (Thread_IsDead(pThd)) {
    return;
  }

  /* the thread should not be running
   * if it is, use PutThreadOnReadyQueue instead */
  /* if the thread is ready, we have already done all this */
  assert(!Thread_IsRunning(pThd));
  assert(!Thread_IsReady(pThd));

  /* the thread must not be on an object queue or on the sleep queue */
  /*      assert(ThreadIsNotOnWaitQueue(pThdArg)); */

  pkcReady = Thread_GetCurrentConstraint(pThd);
  assert(Node_IsAlone(KConstraint_GetSchedQueueNode(pkcReady)));
  assert(Node_IsAlone(KConstraint_GetWaitQueueNode(pkcReady)));
  Log_WriteEntry(pSchedLog, LOG_MAKE_THREAD_READY, (UINT)pkcReady);

  /* if thread was waiting, may resolve constraint with relative wait */

  if (KConstraint_HasPending(pkcReady) &&
      (Thread_IsWaiting(pThd) || Thread_IsSleeping(pThd) || Thread_IsIdle(pThd))) {
    /* BUGBUG yes/no for IDLE? */
    CbsResolvePendingConstraint(pkcReady, FALSE);
  }

  assert(List_FindData(CbsGetSleepQueue(), pkcReady) == NULL);

  /* make thread ready */
  assert(Node_IsAlone(KConstraint_GetSchedQueueNode(pkcReady)));
  Thread_ClearWaiting(pThd);
  CbsReadyQueueAddConstraint(pkcReady);
  pnTemp = List_FindData(CbsGetSleepQueue(), pkcReady);
  assert(pnTemp == NULL);

  assert(Thread_IsReady(pThd));

  /* preempt current thread if appropriate */
  if (bPreemptCurrentThread) {
    CbsPreemptIfNecessary(NULL);
  }

}

SCODE SleepUntil(TIME tSleep)
{
  BOOL PreemptionWasDisabled;
  PTHDINFO pThdCurrent;
  PKCONSTRAINT pkcCurrent;

  PreemptionWasDisabled = CbsDisablePreemption(pTheScheduler, TRUE);
  Log_WriteEntry(pSchedLog, LOG_SLEEP_UNTIL, 0-Uint64Low32Bits(tSleep));

  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);

  CbsNormalizeTimeout(&tSleep);
  pThdCurrent = (PTHDINFO)CurrentThread();

  assert(Thread_IsRunning(pThdCurrent));
  pkcCurrent = Thread_GetCurrentConstraint(pThdCurrent);
  KConstraint_SetRestartBy(pkcCurrent, tSleep);
  KConstraint_RecalculateInheritance(pkcCurrent);
  /*     Thread_SetWaiting(pThdCurrent, THREAD_WAITING, NULL); */
  /*     CbsReadyQueueAddConstraint(pkcCurrent); */
  CbsSleepQueueAddConstraint(pkcCurrent);
  assert(!Thread_IsRunning(pThdCurrent));
  RescheduleNeeded();

  CbsEnablePreemption(pTheScheduler, PreemptionWasDisabled);
  return S_OK;
}

SCODE RTLCALLTYPE EndConstraint(
    PTIME ptTimeTaken
    )
{
  PTHDINFO pThdCurrent;
  PKCONSTRAINT pkcCurrent;
  PKCONSTRAINT pkcNew;
  SCODE sc;

  pThdCurrent = (PTHDINFO)CurrentThread();

  CbsDisablePreemption(pTheScheduler, FALSE);

  /* The current constraint must have a surrounding constraint. */
  pkcCurrent = Thread_GetCurrentConstraint(pThdCurrent);

  assert(pkcCurrent != NULL);
  /* If the constraint has a wait object owner, it wouldn't be running
   * and couldn't end its own constraint.
   */
  /*      assert(KConstraintGetOwner(pkcCurrent) == NULL); */

  if (!KConstraint_HasStackNext(pkcCurrent)) {
    sc = E_ACCESS_DENIED;
    goto Exit;
  }

  /* The thread must be running. */
  assert(Thread_IsRunning(pThdCurrent));
  assert(!TimeEqual(KConstraint_GetLastRefresh(pkcCurrent), TIME_ORIGIN));

  /*      CbsUpdateRestartBy(pkcCurrent); */

  /* Return time taken if so desired. */
  if (ptTimeTaken != NULL) {
    /* RestartBy time advances when the thread runs, so use difference */
    INT64SUB(*ptTimeTaken,
             CurrentTime(),
             /*                KConstraintGetRestartBy(pkcCurrent), */
             KConstraint_GetInitialRestartBy(pkcCurrent));
  }

  /* Pop constraint off of constraint stack. */
  pkcNew = KConstraint_Pop(pkcCurrent);
  Thread_SetCurrentConstraint(pThdCurrent, pkcNew);

  /* Recalculate thread urgency */
  /*      KConstraintStopInheritance(pkcCurrent); */
  KConstraint_RecalculateInheritance(pkcNew);

  /* Free ended constraint, and return status */

  CbsFreeConstraint(pkcCurrent);
  sc = S_OK;

 Exit:
  CbsEnablePreemption(pTheScheduler, FALSE);
  return sc;
}

SCODE RTLCALLTYPE
BeginConstraint(
    BOOL bEndPrevious,
    PTIME_CONSTRAINT ptcNew,
    PTIME ptTimeTaken
    )
{
  PKCONSTRAINT pkcCurrent;
  PTHDINFO pThdCurrent;
  TIME tRestartBy;
  SCODE sc;

  assert(ptcNew != NULL);

  if (bEndPrevious) {
    TConstraint_SetEndPrevious(ptcNew);
  }
  else {
    TConstraint_ClearEndPrevious(ptcNew);
  }

  pThdCurrent = (PTHDINFO)CurrentThread();
  pkcCurrent = Thread_GetCurrentConstraint(pThdCurrent);

  CbsDisablePreemption(pTheScheduler, FALSE);

  /* Do pre-wait processing. */
  sc = CbsBeginConstraintBeforeWaitValidate(pkcCurrent, ptcNew); 
  if (sc != S_OK) {
    goto Exit;
  }

  CbsBeginConstraintBeforeWait(pkcCurrent, ptcNew);
  TConstraint_SetTimeTaken(ptcNew, ptTimeTaken);

  /* Current constraint may change if pkcCurrent has a pending constraint */
  sc = CbsBeginConstraintAfterWait(pkcCurrent);
  if (sc != S_OK) {
    goto Exit;
  }

  pkcCurrent = Thread_GetCurrentConstraint(pThdCurrent);

  tRestartBy = KConstraint_GetRestartBy(pkcCurrent);
  if (TimeLess(CurrentTime(), tRestartBy)) {
    SleepUntil(tRestartBy);
  }

 Exit:
  CbsEnablePreemption(pTheScheduler, FALSE);

  return sc;
}

/* Called by CbsWakeUpOnCondition on every waiter for a condition.
 * If the waiter is also sleeping on a timeout, remove it from sleep queue.
 * Make the waiter's wait queue node alone, move the waiter into ready queue,
 * then mark it ready */
PRIVATE void 
CbsConditionWakeupTransform(
    PTR pkcArg
    )
{
  PTHDINFO pThdWaiter;
  PKCONSTRAINT pkcWaiter = (PKCONSTRAINT)pkcArg;

  assert(pkcWaiter != NULL);

  pThdWaiter = (PTHDINFO)KConstraint_GetSchedObject(pkcWaiter);
  assert(Thread_IsWaitingForCondition(pThdWaiter));

  if (Thread_IsWaitingForTimeout(pThdWaiter)) {
    Log_WriteEntry(pSchedLog, LOG_THREAD_STATE,
                   (UINT)Thread_GetState(pThdWaiter));
    assert(List_FindData(CbsGetSleepQueue(), pkcWaiter) != NULL);
    List_Remove(CbsGetSleepQueue(), 
                KConstraint_GetSchedQueueNode(pkcWaiter));
    assert(List_FindData(CbsGetSleepQueue(), pkcWaiter) == NULL);
  }
  /*     if (List_FindData(CbsGetSleepQueue(), pkcWaiter) != NULL) { */
  /*       Log_Print(pSchedLog); */
  /*       assert(FALSE); */
  /*     } */

  /* We have to make each node alone first b/c MakeThreadReady expects it */
  Node_SetNext(KConstraint_GetWaitQueueNode(pkcWaiter), NULL);
  
  Log_WriteEntry(pSchedLog, LOG_EXIT_CONDITION_SIGNAL, (UINT)pkcWaiter);
  assert(Node_IsAlone(KConstraint_GetSchedQueueNode(pkcWaiter)));
  /*      KConstraintStopInheritance(pkcWaiter); */
  CbsMakeThreadReady(pThdWaiter, FALSE);
}

PRIVATE void
CbsResolvePendingConstraint(
    PKCONSTRAINT pkcCurrent,
    BOOL IsRunning
    )
{
  PKCONSTRAINT pkcNew;
  PTIME_CONSTRAINT ptcPending;
  TIME tDeadline;
  TIME tRestartBy;

  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);
  assert(pkcCurrent != NULL);

  /* There must be a pending constraint */
  assert(KConstraint_HasPending(pkcCurrent));

  ptcPending = KConstraint_GetPending(pkcCurrent);

  /* Make sure RestartBy is up-to-date */
  if (IsRunning) {
    CbsUpdateRestartBy(pkcCurrent);
  }

  /* Mark thread as no longer using its pending constraint.
   * The calling thread must free the memory allocated to this constraint.
   */
  KConstraint_SetPending(pkcCurrent, NULL);

  if (TConstraint_GetEndPrevious(ptcPending) == TRUE) {
    /* Replace the current constraint on the stack */

    assert(KConstraint_HasStackNext(pkcCurrent));

    /* We are replacing the current constraint on the constraint stack. */

    /* Record the time taken during the current constraint if desired. */
    if (TConstraint_GetTimeTaken(ptcPending) != NULL) {
      INT64SUB(*TConstraint_GetTimeTaken(ptcPending),
               KConstraint_GetRestartBy(pkcCurrent),
               KConstraint_GetInitialRestartBy(pkcCurrent));
    }

    pkcNew = pkcCurrent;

  }
  else {
    /* Push a new constraint on the stack. */
    PITHREAD pThdOwner = (PITHREAD)KConstraint_GetSchedObject(pkcCurrent);
    PKCONSTRAINT pkcStackHead;
    pkcNew = CbsAllocConstraint();
    assert(pkcNew != NULL);
      
    KConstraint_Init(pkcNew, NULL);
    KConstraint_SetSchedObject(pkcNew, pThdOwner);
      
    Thread_SetCurrentConstraint(pThdOwner, KConstraint_Push(pkcCurrent,
                                                            pkcNew));
    pkcStackHead = pkcNew;

    while (pkcStackHead != NULL) {
      pkcStackHead = KConstraint_GetStackNext(pkcStackHead);
    }

  }

  tDeadline = TConstraint_GetDeadlineTime(ptcPending);
    
  /* Convert the deadline to be absolute in kernel time. */
  if (TimeIsRelative(tDeadline)) {

    TIME tNow = CurrentTime();

    /*
     * Convert from relative to absolute kernel time.
     */
    INT64SUB(tDeadline, tNow, tDeadline);

    /* If we have overflowed, use TIME_FOREVER.
     */
    if (TimeLess(tDeadline, tNow)) {
      tDeadline = TIME_FOREVER;
    }

  }

  /* The deadline must now be absolute (non-negative) */
  assert(!TimeLess(tDeadline, TIME_ORIGIN));

  /* Convert deadline into restartby time (deadline - estimated runtime). */
  INT64SUB(tRestartBy, tDeadline,
           TConstraint_GetEstimateTime(ptcPending));

  if (TimeLess(tRestartBy, TIME_ORIGIN)) {
    tRestartBy = TIME_ORIGIN;
  }

  /* Copy current restart by to local restart by */
  KConstraint_SetRestartBy(pkcNew, tRestartBy);
  assert(!TimeLess(KConstraint_GetRestartBy(pkcNew), TIME_ORIGIN) );

  /* Save the initial restartby value. */
  KConstraint_SetInitialRestartBy(pkcNew, tRestartBy);

  /* Copy remaining constraint values into current constraint structure. */
  KConstraint_SetCriticality(pkcNew,
                             TConstraint_GetCriticality(ptcPending));

  KConstraint_RecalculateInheritance(pkcNew);

  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);
}

PRIVATE void
CbsUpdateRestartBy(
    PKCONSTRAINT pkcUpdate
    )
{
  TIME tLastRefresh;
  TIME tDelta;
  TIME tNow;
  TIME tRestartBy;
  BOOL PreemptionWasDisabled;

  assert(pkcUpdate != NULL);
  assert(Node_IsAlone(KConstraint_GetSchedQueueNode(pkcUpdate)));
  assert(CbsPreemptionIsEnabled(pTheScheduler) == FALSE);

  Log_WriteEntry(pSchedLog, LOG_ENTER_UPDATE_RESTARTBY, (UINT)pkcUpdate);
  PreemptionWasDisabled = CbsDisablePreemption(pTheScheduler, TRUE);

  tNow = CurrentTime();

  /* It is only valid to refresh the RestartBy of the current thread.
   * (The RestartBy of a non-running thread remains constant.)
   */
  assert(KConstraint_GetSchedObject(pkcUpdate) == (PTR)CurrentThread());

  /* Calculate how long we have been running since this thread's
   * RestartBy was last refreshed.
   */
  assert(!TimeLess(tNow, KConstraint_GetLastRefresh(pkcUpdate)));
  INT64SUB(tLastRefresh, tNow, KConstraint_GetLastRefresh(pkcUpdate));
  assert(!TimeLess(tLastRefresh, TIME_ORIGIN));

  /* If no time has passed, then there is nothing to do. */
  if (TimeEqual(tLastRefresh, TIME_ORIGIN)) {
    Log_WriteEntry(pSchedLog, LOG_NO_UPDATE_RESTARTBY,
                   Int64ToInt32(tLastRefresh));
    goto Exit;
  }

  /* For time-sharing threads, stretch the run time by the
   * number of existing threads. */
  if (Urgency_IsTimeSharing(KConstraint_GetCurrentUrgency(pkcUpdate))) {
    INT t32; 
    t32 = Int64ToInt32(tLastRefresh);  
    t32 = t32 * Cbs.nThreads;  
    Int32ToInt64(tDelta, t32);
     /*      tDelta = TIME_SHARING_TIMESLICE;*/
  }
  else {
    tDelta = tLastRefresh;
  }

  /* Delta is not smaller than TimeSinceLastRefresh.
   * And since TimeSinceLastRefresh is nonnegative, so is Delta.
   */
#if !defined(SIM)
  assert(!TimeLess(tDelta, tLastRefresh));
  assert(!TimeLess(tDelta, TIME_ORIGIN) );
#endif

  /* Add Delta and correct for any overflow.
   *
   * The following assertions imply that the sum will lie
   * between 0 and 2*TIME_FOREVER, inclusive, not counting for
   * round-off, or between [0..TIME_FOREVER] or [-TIME_FOREVER-1, -2],
   * counting for roundoff. Thus any negative value will indicate
   * that an overflow has occurred, and any nonnegative value will
   * indicate that an overflow has not occurred.
   */
  assert(!TimeLess(KConstraint_GetRestartBy(pkcUpdate), TIME_ORIGIN));

  tRestartBy = KConstraint_GetRestartBy(pkcUpdate);
  INT64ADD(tRestartBy, tRestartBy, tDelta);

  /* Negative value implies overflow, represent with TIME_FOREVER. */
  if (TimeLess(tRestartBy, TIME_ORIGIN)) {
    tRestartBy = TIME_FOREVER;
  }

  KConstraint_SetRestartBy(pkcUpdate, tRestartBy);

  /* tRestartBy is a valid absolute time. */
  assert(!TimeLess(tRestartBy, TIME_ORIGIN));

  /* Update TimeOfLastRefresh. */
  KConstraint_SetLastRefresh(pkcUpdate, tNow);
  KConstraint_RecalculateInheritance(pkcUpdate);
  Log_WriteEntry(pSchedLog, LOG_EXIT_UPDATE_RESTARTBY, Int64ToInt32(tDelta));
 Exit:
  CbsEnablePreemption(pTheScheduler, PreemptionWasDisabled);
}
