/* cb_sched3.c
 * Constraint-based scheduler policy.
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <mmlite.h>
#include <fred.h>
#include <diagnostics.h>
#include <base/schedul.h>
#include <schedulif.h>
#include <base/thread2.h>
#include <base/tconstraint2.h>
#include <base/mutex2.h>
#include <util/list-single.h>
#include <activif.h>

#include <base/urgency2.h>
#include <base/kconstraint2.h>
#include <base/workitem.h>
#include "sched-multi.h"

/* Enable/disable logging by uncommenting/commenting the following line */
/* #define LOGGING_ENABLED 1 */

#include <util/logger.h>

#ifdef LOGGING_ENABLED
#define SCHEDULER_LOG_SIZE 128
LOG_ENTRY SchedulerLog[SCHEDULER_LOG_SIZE];

LOG SchedLog;
PLOG pSchedLog = &SchedLog;

DECLARE_LOG_EVENT_TYPE(RESCHEDULE             , "Resched ", TRUE );
DECLARE_LOG_EVENT_TYPE(RESCHEDULE_MAIN        , "ReschInt", TRUE );
DECLARE_LOG_EVENT_TYPE(IDLE_THREAD            , "enIdle  ", TRUE );
DECLARE_LOG_EVENT_TYPE(END_IDLE_THREAD        , "exIdle  ", TRUE );
DECLARE_LOG_EVENT_TYPE(DEL_ACTIVITY           , "DEAD    ", TRUE );
DECLARE_LOG_EVENT_TYPE(SLEEP_UNTIL            , "enSleep ", TRUE );
DECLARE_LOG_EVENT_TYPE(OUT                    , "OUT     ", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_CONDITION_WAIT   , "enCdWait", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_CONDITION_WAIT    , "exCdWait", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_CONDITION_SIGNAL , "enCdSign", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_CONDITION_SIGNAL  , "exCdSign", TRUE );
DECLARE_LOG_EVENT_TYPE(DRAIN_DEFERRED         , "drainDef", TRUE );
DECLARE_LOG_EVENT_TYPE(PUT_READY_QUEUE        , "addReady", TRUE );
DECLARE_LOG_EVENT_TYPE(POP_READY_QUEUE        , "delReady", TRUE );
DECLARE_LOG_EVENT_TYPE(PUT_SLEEP_QUEUE        , "addSleep", TRUE );
DECLARE_LOG_EVENT_TYPE(POP_SLEEP_QUEUE        , "delSleep", TRUE );
DECLARE_LOG_EVENT_TYPE(THREAD_STATE           , "ThdState", TRUE );
DECLARE_LOG_EVENT_TYPE(ADD_THREAD             , "AddThd  ", TRUE );
DECLARE_LOG_EVENT_TYPE(PROCESS_CONSTRAINT     , "Process ", TRUE );
DECLARE_LOG_EVENT_TYPE(PROCESS_INVALID        , "PrcInval", TRUE );
DECLARE_LOG_EVENT_TYPE(PROCESS_OUTATIME       , "PrcTime ", TRUE );
DECLARE_LOG_EVENT_TYPE(ADD_FAILED             , "AddFail ", TRUE );
DECLARE_LOG_EVENT_TYPE(ADD_WORKITEM           , "AddWork ", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT                , "Preempt ", TRUE );
DECLARE_LOG_EVENT_TYPE(DEL_WORKITEM           , "DelWork  ", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_WAKE_THREADS     , "enWake  ", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_WAKE_REALTIME     , "exWakeRT", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_WAKE_TIMESHARE    , "exWakeTS", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_UPDATE_RESTARTBY , "enUpRsBy", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_UPDATE_RESTARTBY  , "exUpRsBy", TRUE );
DECLARE_LOG_EVENT_TYPE(NO_UPDATE_RESTARTBY    , "noUpRsBy", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT_DEAD           , "PreDead ", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT_RESCHEDULING   , "PreResch", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT_NO_OTHERS      , "PreNoOth", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT_STILL_URGENT   , "PreStill", TRUE );
DECLARE_LOG_EVENT_TYPE(MAKE_THREAD_READY      , "MakReady", TRUE );

#undef LOGGING_ENABLED

#endif

#if !defined(_INT64)
const TIME TIME_SHARING_TIMESLICE = Int64Initializer(0,400000/*Low*/);
#else
#define TIME_SHARING_TIMESLICE   400000 /* 40 millisecs */
#endif

/* MMLite Constraint-based scheduler object (Cbs) */
typedef struct _SCHED {
  SCHED_MULTI_STRUCT;
  LIST lReadyQueue;
  LIST lSleepQueue;
} SCHED, *PSCHED;

#define pSCHED(_x_) ((PSCHED)_x_)

/* ===========================================================================
 * Ready Queue Macros
 */

#define /* PLIST */                                                           \
CbsSched_GetReadyQueue()                                                      \
     &(TheScheduler.lReadyQueue)

#define /* void */                                                            \
CbsSched_ReadyQueueInitialize()                                               \
     List_Init(CbsSched_GetReadyQueue())

#define /* BOOL */                                                            \
CbsSched_ReadyQueueIsEmpty()                                                  \
     List_IsEmpty(CbsSched_GetReadyQueue())

#define /* PACINFO */ /* Must call Sched_ReadyQueueIsEmpty first! */          \
CbsSched_ReadyQueueFindMostUrgentActivity()                                   \
     ((PACTINFO)(Node_GetData(List_GetHead(CbsSched_GetReadyQueue()))))

#define /* PACTINFO */ /* Must call Sched_ReadyQueueIsEmpty first! */         \
CbsSched_ReadyQueueRemoveMostUrgentActivity()                                 \
     ((PACTINFO)(Node_GetData(List_HeadRemove(CbsSched_GetReadyQueue()))))

#define /* void */                                                            \
CbsSched_ReadyQueueAddActivity(/* PIACTIVITY */ _pAct_)                       \
     List_TailAdd(CbsSched_GetReadyQueue(), Activity_GetNode(_pAct_))

#define /* void */                                                            \
CbsSched_ReadyQueueRemoveActivity(/* PIACTIVITY */ _pAct_)                    \
     List_Remove(CbsSched_GetReadyQueue(), Activity_GetNode(_pAct_))

/* ===========================================================================
 * Sleep Queue Macros
 */

#define /* PLIST */                                                           \
CbsSched_GetSleepQueue()                                                      \
     &(TheScheduler.lSleepQueue)

#define /* void */                                                            \
CbsSched_SleepQueueInitialize()                                               \
     List_Init(CbsSched_GetSleepQueue())

#define /* PNODE */                                                           \
CbsSched_SleepQueueGetFirstNode()                                             \
     List_GetHead(CbsSched_GetSleepQueue())

#define /* void */                                                            \
CbsSched_SleepQueueIsEmpty()                                                  \
     List_IsEmpty(CbsSched_GetSleepQueue())

#define /* PACTINFO */ /* Must call Sched_ReadyQueueIsEmpty first! */         \
CbsSched_SleepQueueFindMostUrgentActivity()                                   \
     ((PACTINFO)(Node_GetData(List_GetHead(CbsSched_GetSleepQueue()))))

#define /* PACTINFO */ /* Must call Sched_SleepQueueIsEmpty first! */         \
CbsSched_SleepQueueRemoveMostUrgentActivity()                                 \
     ((PACTINFO)(Node_GetData(List_HeadRemove(CbsSched_GetSleepQueue()))))

/* ===========================================================================
 * Scheduler functions that implement exported interface methods.
 */

PRIVATE SCODE MCT
CbsSched_CreateConstraint(
    PISCHEDULER This,
    TIME tStartTime,
    TIME tDeadline,
    TIME tEstimate,
    UINT Recurrence,
    BOOL Hardness,
    BOOL Adjust,
    BOOL Infinite,
    PIUNKNOWN* ppNewConstraint
    );

PRIVATE SCODE MCT 
CbsSched_SetConstraint(
    PISCHEDULER This,
    PTIME_CONSTRAINT pConstraint,
    PTIME pTimeTaken
    );

/* ===========================================================================
 * Constraint-based scheduler vtable.
 */

PRIVATE CONST_VTBL struct ISchedulerVtbl CbsSchedVtbl = {
  __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
  Sched_QueryInterface,
  Sched_AddRef,
  Sched_Release,
  Sched_AddActivity,
  Sched_RemoveActivity,
  Sched_DisablePreemption,
  Sched_EnablePreemption,
  Sched_PreemptionIsEnabled,
  Sched_WaitForMutex,
  Sched_WakeUpOnMutex,
  Sched_WaitForCondition,
  Sched_WakeUpOnCondition,
  Sched_WaitForConditionWithConstraint,
  CbsSched_SetConstraint,
  Sched_GetIdleTime,
};

/* Scheduler object */
PRIVATE SCHED TheScheduler;

/* Include machine dependent operations & defines */
#include "_schedul.c"

/* Constructor called from base */

PISCHEDULER 
SchedulerCreate(void)
{
  TheScheduler.v = &CbsSchedVtbl;
  Log_Init(pSchedLog, SCHEDULER_LOG_SIZE, SchedulerLog);
  SCHED_CREATE();
  return (PISCHEDULER)&TheScheduler;
}

/* ===========================================================================
 * Private scheduler function prototypes
 */

PACTINFO
CbsSched_FindMostUrgentMutexWaiter(
    PACTINFO pAct
    );

PRIVATE void 
CbsSched_RecalculateInheritance(
    PACTINFO pActRecalc
    );

PRIVATE void 
CbsSched_UpdateRestartBy(
    PACTINFO pActUpdate
    );

PRIVATE PACTINFO 
CbsSched_RescheduleInterrupt(
    PACTINFO pActOld,
    PACTINFO pActNew
    );

PRIVATE PACTINFO
CbsSched_PreemptIfNecessary(
    PTIME pNextTimeSlice
    );

PRIVATE BOOL
CbsSched_MutexWaiterCompareGT(
    PTR pMX1,
    PTR pMX2
    );

PRIVATE BOOL
CbsSched_RestartByGT(
    PTR pAct1,
    PTR pAct2
    );

PRIVATE SCODE
Sched_ProcessConstraint(
    PACTINFO pActCurrent,
    PTIME_CONSTRAINT ptcNew,
    PTIME pTimeTaken
    );

/* ===========================================================================
 * Private scheduler function implementations
 */

PRIVATE BOOL
CbsSched_RestartByGT(
    PTR pAct1,
    PTR pAct2
    )
{
  PKCONSTRAINT pConstraint1 = Activity_GetConstraint(pAct1);
  PKCONSTRAINT pConstraint2 = Activity_GetConstraint(pAct2);
  return TimeLess(KConstraint_GetRestartBy(pConstraint1),
                  KConstraint_GetRestartBy(pConstraint2));
}

/* Insert activity into sleep queue, sorted by urgency and set sleeping state.
 */
PRIVATE INLINE void
Sched_SleepActivity(
    PACTINFO pActSleep,
    PTIME ptSleep
    )
{
  BOOL PreemptionWasDisabled;
  PKCONSTRAINT pkcCurrent;

  assert(pActSleep != NULL);
  PreemptionWasDisabled = Sched_DisablePreemption(pTheScheduler, TRUE);
  Log_WriteEntry(pSchedLog, LOG_SLEEP_UNTIL, 0-Uint64Low32Bits(*ptSleep));
  assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE);
  
  Sched_NormalizeTimeout(ptSleep);
/*   pActCurrent = (PACTINFO)CurrentActivity(); */
/*   if (Activity_GetThread(pActCurrent) != NULL) { */
/*     pActCurrent = Activity_GetThread(pActCurrent); */
/*   } */
  
  assert(!Activity_IsSleeping(pActSleep));
  assert(!Activity_IsIdle(pActSleep));
  pkcCurrent = Activity_GetConstraint(pActSleep);
  KConstraint_SetRestartBy(pkcCurrent, *ptSleep);
  CbsSched_RecalculateInheritance(pActSleep);
  List_InsertSortedDescending(CbsSched_GetSleepQueue(), CbsSched_RestartByGT,
                              Activity_GetNode(pActSleep));
  Activity_SetState(pActSleep, ACTIVITY_SLEEPING);
  Log_WriteEntry(pSchedLog, LOG_PUT_SLEEP_QUEUE, (UINT)pActSleep);
  /*     Thread_SetWaiting(pThdCurrent, THREAD_WAITING, NULL); */
  /*     CbsReadyQueueAddConstraint(pkcCurrent); */

  assert(!Activity_IsRunning(pActSleep));
  RescheduleNeeded();
  
  Sched_EnablePreemption(pTheScheduler, PreemptionWasDisabled);


/*   if (Activity_GetThread(pActSleep) != NULL) { */
/*     pActSleep = Activity_GetThread(pActSleep); */
/*   } */

/*   assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE); */

  /* Insert pActSleep before the first thread with a later start time. */
}

/* #ifdef _MSC_VER */
/* #pragma warning(disable:4127)  /\* conditional expression is constant *\/ */
/* #pragma warning(disable:4702)  /\* unreachable code *\/ */
/* #endif */

PRIVATE void 
CbsSched_RecalculateInheritance(
    PACTINFO pActRecalc
    )
{
  PACTINFO pActMostUrgent;
  PKCONSTRAINT pkcRecalc;
  PURGENCY puRecalc;

  pkcRecalc = Activity_GetConstraint(pActRecalc);
  puRecalc = KConstraint_GetCurrentUrgency(pkcRecalc);
  Urgency_Copy(puRecalc, KConstraint_GetLocalUrgency(pkcRecalc));

  pActMostUrgent =
    CbsSched_FindMostUrgentMutexWaiter(pActRecalc);
  
    
  if (pActMostUrgent != NULL) {
    PURGENCY puMostUrgent;
    puMostUrgent = 
      KConstraint_GetCurrentUrgency(Activity_GetConstraint(pActMostUrgent));
    if (!Urgency_IsTimeSharing(puMostUrgent) &&
	Urgency_GT(puMostUrgent, puRecalc)) {
      Urgency_Copy(puRecalc, puMostUrgent);
    }
  }
}

PRIVATE void 
CbsSched_UpdateRestartBy(
    PACTINFO pActUpdate
    )
{
  PKCONSTRAINT pkcUpdate;
  TIME tLastRefresh;
  TIME tDelta;
  TIME tNow;
  TIME tRestartBy;
  BOOL PreemptionWasDisabled;

  pkcUpdate = Activity_GetConstraint(pActUpdate);
  assert(pkcUpdate != NULL);
  assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE);

  Log_WriteEntry(pSchedLog, LOG_ENTER_UPDATE_RESTARTBY, (UINT)pActUpdate);
  PreemptionWasDisabled = Sched_DisablePreemption(pTheScheduler, TRUE);

  tNow = CurrentTime();

  /* It is only valid to refresh the RestartBy of the current thread.
   * (The RestartBy of a non-running thread remains constant.)
   */

  /* Calculate how long we have been running since this thread's
   * RestartBy was last refreshed.
   */
  assert(!TimeLess(tNow, KConstraint_GetLastRefresh(pkcUpdate)));
  INT64SUB(tLastRefresh, tNow, KConstraint_GetLastRefresh(pkcUpdate));
  assert(!TimeLess(tLastRefresh, TIME_ORIGIN));

  /* If no time has passed, then there is nothing to do. */
  if (TimeEqual(tLastRefresh, TIME_ORIGIN)) {
    Log_WriteEntry(pSchedLog, LOG_NO_UPDATE_RESTARTBY,
                   Int64ToInt32(tLastRefresh));
    goto Exit;
  }

  /* For time-sharing threads, stretch the run time by the
   * number of existing threads. */
  if (Urgency_IsTimeSharing(KConstraint_GetCurrentUrgency(pkcUpdate))) {
    INT t32;
    t32 = Int64ToInt32(tLastRefresh);
    t32 = t32 * Sched_GetActCount();
    Int32ToInt64(tDelta, t32);
    /*     tDelta = TIME_SHARING_TIMESLICE; */
  }
  else {
    tDelta = tLastRefresh;
  }

  /* Delta is not smaller than TimeSinceLastRefresh.
   * And since TimeSinceLastRefresh is nonnegative, so is Delta.
   */
#if !defined(SIM)
  assert(!TimeLess(tDelta, tLastRefresh) );
  assert(!TimeLess(tDelta, TIME_ORIGIN) );
#endif

  /* Add Delta and correct for any overflow.
   *
   * The following assertions imply that the sum will lie
   * between 0 and 2*TIME_FOREVER, inclusive, not counting for
   * round-off, or between [0..TIME_FOREVER] or [-TIME_FOREVER-1, -2],
   * counting for roundoff. Thus any negative value will indicate
   * that an overflow has occurred, and any nonnegative value will
   * indicate that an overflow has not occurred.
   */
  assert(!TimeLess(KConstraint_GetRestartBy(pkcUpdate), TIME_ORIGIN));

  tRestartBy = KConstraint_GetRestartBy(pkcUpdate);
  INT64ADD(tRestartBy, tRestartBy, tDelta);

  /* Negative value implies overflow, represent with TIME_FOREVER. */
  if (TimeLess(tRestartBy, TIME_ORIGIN)) {
    tRestartBy = TIME_FOREVER;
  }
  
  KConstraint_SetRestartBy(pkcUpdate, tRestartBy);

  /* tRestartBy is a valid absolute time. */
  assert(!TimeLess(tRestartBy, TIME_ORIGIN));

  /* Update TimeOfLastRefresh. */
  KConstraint_SetLastRefresh(pkcUpdate, tNow);
  CbsSched_RecalculateInheritance(pActUpdate);
  Log_WriteEntry(pSchedLog, LOG_EXIT_UPDATE_RESTARTBY, Int64ToInt32(tDelta));
 Exit:
  Sched_EnablePreemption(pTheScheduler, PreemptionWasDisabled);
}

/* If (pConstraintNew != NULL) add pConstraintOld to sleep queue, make pConstraintNew ready to run.
 * Else if (pConstraintNew == NULL && pConstraintOld still running) do nothing.
 * Else if (pConstraintNew == NULL && pConstraintOld not running) make most urgent constraint
 *   ready to run, or spin idle thread while ready queue is empty.
 * Returns the next running constraint; never returns NULL because of
 *   spinning idle thread.
 * This rescheduling path is 'synchronous',
 * i.e. we get here if and only if RescheduleNeeded() is called.
 */
PRIVATE PACTINFO 
CbsSched_RescheduleInterrupt(
    PACTINFO pActOld,
    PACTINFO pActNew
    )
{

  assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE);
  Log_WriteEntry(pSchedLog, LOG_RESCHEDULE_MAIN, (UINT)pActOld);

/*   assert(Activity_GetThread(pActOld) == NULL); */
  /* The old thread exists (either running or about to die) */
  assert(pActOld != NULL);

  if (pActNew == NULL) {
    if (CbsSched_ReadyQueueIsEmpty()) {
      /* there is no other ready thread to switch to */
      PCXTINFO pSavedCxt;
      TIME tStartIdle;

      /* If current thread can keep running, then let it */
      if (Activity_IsRunning(pActOld)) {
        return pActOld;
      }

      /* Begin playing idle thread and set time to come out of it. */
      tStartIdle = CurrentTime();

      /* Save state in case we get preempted by a new ready thread */
      Activity_SetIdle(pActOld);
      pSavedCxt = Thread_GetContext(pActOld);
      Log_WriteEntry(pSchedLog, LOG_IDLE_THREAD, (UINT)pActOld);

      Sched_EnablePreemption(pTheScheduler, FALSE);
      /* Sched_Reschedule won't preempt here, just add new ready threads to
       * the queue */

      /* Spin while the ready queue is empty. */
      while (CbsSched_ReadyQueueIsEmpty()) {
#ifdef MACHINE_IDLE
        MACHINE_IDLE(CbsSched_ReadyQueueIsEmpty());
#endif
      }

      Sched_DisablePreemption(pTheScheduler, FALSE);
      Log_WriteEntry(pSchedLog, LOG_END_IDLE_THREAD, (UINT)pActOld);

      Thread_SetContext(pActOld, pSavedCxt);
      Activity_ClearIdle(pActOld);

      /* Update scheduler idle time and LastRefresh time of old constraint */
      {
        TIME tNow = CurrentTime();
        INT64SUB(tStartIdle, tNow, tStartIdle);
        INT64ADD(TheScheduler.IdleTime, TheScheduler.IdleTime, tStartIdle);
        KConstraint_SetLastRefresh(Activity_GetConstraint(pActOld), tNow);
      }
    }
    pActNew = CbsSched_ReadyQueueRemoveMostUrgentActivity();
    Log_WriteEntry(pSchedLog, LOG_POP_READY_QUEUE, (UINT)pActNew);
  }
  else {
    List_Remove(CbsSched_GetReadyQueue(),
                Activity_GetNode(pActNew));
  }

  assert(Node_IsAlone(Activity_GetNode(pActNew)));
  Log_WriteEntry(pSchedLog, LOG_END_IDLE_THREAD, (UINT)pActNew);

  /* The new thread exists, has not died, and is ready to run.  */
  assert(pActNew != NULL);
  assert(Activity_IsReady(pActNew));
  assert(Activity_GetWaitObject(pActNew) == NULL);
  assert(Node_IsAlone(Activity_GetNode(pActNew)));
  /* Work items can be on the thread pool queue */
/*   assert(Node_IsAlone(Activity_GetWaitNode(pActNew))); */

  Activity_SetState(pActNew, ACTIVITY_RUNNING);

  /* If the old and new threads are different, then do context switching */
  if (pActNew != pActOld) {

    /* Then put it back on the ready queue
     */
    Log_WriteEntry(pSchedLog, LOG_THREAD_STATE, (UINT)Activity_GetState(pActOld));
    /* The only old threads that should be readied again are the first threads
     * (bootstrapping case, after the first reschedule), a currently running
     * thread, or a non-work-item.
     */
    if ((Activity_IsStarting(pActOld) || Activity_IsRunning(pActOld))) {
      /* Update the RestartBy of the old thread. */
      CbsSched_UpdateRestartBy(pActOld);
      Log_WriteEntry(pSchedLog, LOG_PUT_READY_QUEUE, (UINT)pActOld);
      CbsSched_ReadyQueueAddActivity(pActOld);
      assert(!Activity_IsWaiting(pActOld));
      Activity_SetState(pActOld, ACTIVITY_READY);
    }

    /* Save current time for future UpdateRestartBy calls on new thread. */
    KConstraint_SetLastRefresh(Activity_GetConstraint(pActNew),
                               CurrentTime());


    if (WorkItem_HasPseudoThread(pActNew)) {
      /* For a work item, call its switch function. */
      WORK_SWITCH_FUNCTION SwitchFunc;
      PWORKINFO pWorkOld = NULL;
      SwitchFunc = WorkItem_GetSwitchFunction(pActNew);
      assert(SwitchFunc != NULL);
      pActNew = SwitchFunc((PWORKINFO)pActNew, &pWorkOld);
      if (pWorkOld != NULL) {
        CbsSched_UpdateRestartBy(pActOld);
        Log_WriteEntry(pSchedLog, LOG_PUT_READY_QUEUE, (UINT)pWorkOld);
        CbsSched_ReadyQueueAddActivity(pWorkOld);
        assert(!Activity_IsWaiting(pWorkOld));
        Activity_SetState(pWorkOld, ACTIVITY_READY);
      }
      assert(pActNew != NULL);
/*       assert(!WorkItem_HasPseudoThread(pActNew)); */
/*       do { */
/*         pActNew = CbsSched_RescheduleInterrupt(pActNew, NULL); */
/*       } while (WorkItem_HasPseudoThread(pActNew)); */
/*       WorkItem_Dereference(pActNew); */
   }

    /* Make sure we are returning a thread */
/*     assert(Activity_GetThread(pActNew) == NULL); */

    /* Make the new thread current */
    ActivitySetCurrent((PIACTIVITY)pActNew);

  }

  return pActNew;
}

/* Returns the constraint belonging to the most urgent ready, non-running
 * thread. If there are no ready threads, no ready threads
 * more urgent than the currently running thread, or the
 * currently running thread is already rescheduling, returns NULL.
 */
PRIVATE PACTINFO
CbsSched_PreemptIfNecessary(
    PTIME pNextTimeSlice
    )
{
  PACTINFO pActCurrent;
  PKCONSTRAINT pkcCurrent;
  PACTINFO pActNext;
  PKCONSTRAINT pkcNext = NULL;

  pActCurrent = (PACTINFO)CurrentActivity();
  pkcCurrent = Activity_GetConstraint(pActCurrent);
  assert(pkcCurrent != NULL);

  assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE);

  /* Watch out for threads that are already dead. */
  assert(pActCurrent != NULL);

  if (Activity_IsDead(pActCurrent)) {
    /* If the current thread is dead then we should reschedule.
     * However, when death occurs the thread will look for
     * an undertaker, spinning IdleThread while none is found.
     * Which is the case we catch here.
     * Therefore we say the opposite.  Obviously ;-))
     */
    RescheduleNotNeeded();
    Log_WriteEntry(pSchedLog, LOG_PREEMPT_DEAD, (UINT)pActCurrent);
    return NULL;
  }

  /* If the current thread is already rescheduling, then we are done. */
  if (!Activity_IsRunning(pActCurrent)) {
    Log_WriteEntry(pSchedLog, LOG_PREEMPT_RESCHEDULING,
                   (UINT)pActCurrent);
    return NULL;
  }

  /* If there are no other runnable threads, then we are done. */
  if (CbsSched_ReadyQueueIsEmpty()) {
    Log_WriteEntry(pSchedLog, LOG_PREEMPT_NO_OTHERS, (UINT)pActCurrent);
    return NULL;
  }

  /* Fetch the most urgent ready (runnable but not running) thread. */
  pActNext = CbsSched_ReadyQueueFindMostUrgentActivity();
  assert(Activity_IsReady(pActNext));
  assert(!Activity_IsWaiting(pActNext));
  assert(Activity_GetWaitObject(pActNext) == NULL);
  assert(pActNext != NULL);

  pkcNext = Activity_GetConstraint(pActNext);
  assert(pkcNext != NULL);

  Log_WriteEntry(pSchedLog, LOG_PREEMPT, (UINT)pActNext);
    
  /* Refresh the constraint of the currently running thread before
   * we start comparing it to the constraints of the ready threads.
   */
  CbsSched_UpdateRestartBy(pActCurrent);

  /* If current thread remains the most urgent runnable thread,
   * then do not preempt.
   * If the current thread is not less urgent than the most urgent
   * runnable thread, then we let it keep running.
   */
  if (KConstraint_UrgencyGTEQ(pkcCurrent, pkcNext)) {
    Log_WriteEntry(pSchedLog, LOG_PREEMPT_STILL_URGENT,
                   (UINT)pkcCurrent);
    return NULL;
  }
  
  /* We have decided to preempt the current thread. */
  RescheduleNeeded();

  /* All cases must end here, to turn off the timer. */
  return pActNext;
}

PRIVATE INLINE void
Sched_MutexRemoveWaiter(
    PMX pMutex,
    PACTINFO pActWaiter
    )
{
  PKCONSTRAINT pConstraintCurrent = Activity_GetConstraint(pActWaiter);
  /* The old mutex owner is now free from any inherited constraints
   * since it no longer owns this mutex.
   */
  Urgency_Copy(KConstraint_GetCurrentUrgency(pConstraintCurrent),
               KConstraint_GetLocalUrgency(pConstraintCurrent));
}

PRIVATE BOOL
CbsSched_MutexWaiterCompareGT(
    PTR pMX1,
    PTR pMX2
    )
{
  PNODE pn1 = List_GetHead(Mutex_GetWaiterQueue(pMX1));
  PNODE pn2 = List_GetHead(Mutex_GetWaiterQueue(pMX2));
  if (pn2 == NULL) {
    return TRUE;
  }
  if (pn1 == NULL) {
    return FALSE;
  }

  return CbsSched_RestartByGT((PACTINFO)Node_GetData(pn1),
                              (PACTINFO)Node_GetData(pn2));
}

PRIVATE INLINE PACTINFO
CbsSched_FindMostUrgentMutexWaiter(
    PACTINFO pAct
    )
{
  PMX pMutexMostUrgent;
  
  assert(pAct != NULL);
  /* Make sure we are not a work item */
  /*   assert(Activity_GetWard(pThd) == NULL); */
  
  pMutexMostUrgent = List_FindDataMax(Activity_GetMutexList(pAct),
                                      CbsSched_MutexWaiterCompareGT);
  
  if (pMutexMostUrgent != NULL) {
    PLIST plWaiters = Mutex_GetWaiterQueue(pMutexMostUrgent);
    if (!List_IsEmpty(plWaiters))
      return (PACTINFO)Node_GetData(List_GetHead(plWaiters));
  }
  
  return NULL;
}

/* Hook function called by Sched_WakeupOnCondition to handle policy-specific
 * mutex relocking (i.e. unlocking and waking up the most urgent waiter).
 * Note that pActWaiter has already been added to pMutex's
 * waiter queue before this function is called. */
PRIVATE INLINE void
Sched_MutexRelock(
    PMX pMutex,
    PACTINFO pActWaiter
    )
{
  CbsSched_RecalculateInheritance(pActWaiter);
}

/* Propagates up urgencies that may be more or equally urgent than
 * owners, but not less than.
 */
PRIVATE INLINE void
CbsSched_StartInheritance(
    PACTINFO pActOwner,
    PURGENCY puNew
    )
{
  PURGENCY puCurrent;
  PKCONSTRAINT pkcOwner = Activity_GetConstraint(pActOwner);

  assert(pkcOwner != NULL);

  puCurrent = KConstraint_GetCurrentUrgency(pkcOwner);

  if (Urgency_IsTimeSharing(puCurrent)) {
    return;
  }

  if (Urgency_GT(puNew, puCurrent)) {
    /*       PACTINFO pActOwner = (PACTINFO)KConstraint_GetSchedObject(pkcOwner); */

    Urgency_Copy(puCurrent, puNew);

    if (Activity_IsWaitingForMutex(pActOwner)) {
      PMUTEX pMutex = Activity_GetWaitObject(pActOwner);
      PACTINFO pActWaitee = Mutex_GetSchedObject(pMutex);
      PKCONSTRAINT pkcOwnerOwner = Activity_GetConstraint(pActWaitee);
      CbsSched_StartInheritance(pActWaitee, puNew);
    }
  }
}	    

/* Hook function called by Sched_WaitForMutex to handle policy-specific
 * mutex waiting. Note that pActWaiter has already been added to pMutex's
 * waiter queue before this function is called. */
PRIVATE INLINE void
Sched_MutexAddWaiter(
    PMX pMutex,          /* Mutex to wait on */
    PACTINFO pActWaiter  /* Activity doing the waiting */
    )
{
  PKCONSTRAINT pConstraintWaiter = Activity_GetConstraint(pActWaiter);
  CbsSched_UpdateRestartBy(pActWaiter);
  List_InsertSortedDescending(Mutex_GetWaiterQueue(pMutex),
                              CbsSched_RestartByGT,
                              Activity_GetWaitNode(pActWaiter));
  CbsSched_StartInheritance(
      (PACTINFO)Mutex_GetSchedObject(pMutex),
      KConstraint_GetCurrentUrgency(pConstraintWaiter));
}

PRIVATE INLINE void
Sched_ConditionRemoveWaiter(
    PCOND pCond,
    PACTINFO pActWaiter
    )
{
  CbsSched_PreemptIfNecessary(NULL);
}

PRIVATE INLINE void
CbsSched_ResolveConstraint(
    PACTINFO pActCurrent,
    PTIME_CONSTRAINT ptcPending
    )
{
  PKCONSTRAINT pkcCurrent;
/*   PKCONSTRAINT pkcNew; */
/*   PTIME_CONSTRAINT ptcPending; */
  TIME tDeadline;
  TIME tRestartBy;
  
  assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE);
  pkcCurrent = Activity_GetConstraint(pActCurrent);
  assert(pkcCurrent != NULL);

  assert(ptcPending != NULL);

  /* Make sure RestartBy is up-to-date */
/*   if (IsRunning) { */
  CbsSched_UpdateRestartBy(pActCurrent);
/*   } */

  tDeadline = TConstraint_GetDeadlineTime(ptcPending);
  
  /* Convert the deadline to be absolute in kernel time. */
  if (TimeIsRelative(tDeadline)) {
    TIME tNow = CurrentTime();
    /* Convert from relative to absolute kernel time. */
    INT64SUB(tDeadline, tNow, tDeadline);
    
    /* If we have overflowed, use TIME_FOREVER. */
    if (TimeLess(tDeadline, tNow)) {
      tDeadline = TIME_FOREVER;
    }
  }
  
  /* The deadline must now be absolute (non-negative) */
  assert(!TimeLess(tDeadline, TIME_ORIGIN));
  
  /* Convert deadline into restartby time (deadline - estimated runtime). */
  INT64SUB(tRestartBy, tDeadline,
           TConstraint_GetEstimateTime(ptcPending));
  
  if (TimeLess(tRestartBy, TIME_ORIGIN)) {
    tRestartBy = TIME_ORIGIN;
  }
  
  /* Copy current restart by to local restart by */
  KConstraint_SetRestartBy(pkcCurrent, tRestartBy);
  
  assert(!TimeLess(KConstraint_GetRestartBy(pkcCurrent), TIME_ORIGIN) );
  
  /* Save the initial restartby value. */
  KConstraint_SetInitialRestartBy(pkcCurrent, tRestartBy);
  
  /* Copy remaining constraint values into current constraint structure. */
  KConstraint_SetCriticality(pkcCurrent,
                             TConstraint_GetCriticality(ptcPending));

  /* Propagate up any waiters to adjust the newly-resolve constraint */  
  CbsSched_RecalculateInheritance(pActCurrent);
  
  assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE);
}

/* Hook function called by Sched_WaitForCondition to handle policy-specific
 * condition waiting. Note that pActWaiter has already been added to pCond's
 * waiter queue before this function is called. */
PRIVATE INLINE void
Sched_ConditionAddWaiter(
    PCOND pCond,         /* The condition being waited on */
    PACTINFO pActWaiter, /* Activity that waits on this condition */
    PTIME pTimeOut       /* Optional timeout for wait; may be NULL */
    )
{
  CbsSched_UpdateRestartBy(pActWaiter);
  List_InsertSortedDescending(Condition_GetWaiterQueue(pCond),
                              CbsSched_RestartByGT,
                              Activity_GetWaitNode(pActWaiter));
  if (pTimeOut != NULL) { /* Place in sleep queue */
    /* BUG The arm board may depend on infinite timeouts being trimmed down
     * to the maximum time, which is not the same as a NULL timeout */
    if (!TimeEqual(*pTimeOut, TIME_FOREVER)) {
/*       PKCONSTRAINT pkcWaiter = Activity_GetConstraint(pActWaiter); */
/*       Sched_NormalizeTimeout(pTimeOut); */
/*       KConstraint_SetRestartBy(pkcWaiter, *pTimeOut); */
/*       Urgency_Copy(KConstraint_GetCurrentUrgency(pkcWaiter), */
/*                    KConstraint_GetLocalUrgency(pkcWaiter)); */
      Sched_SleepActivity(pActWaiter, pTimeOut);
      assert(List_FindData(CbsSched_GetSleepQueue(), pActWaiter) ==
             Activity_GetNode(pActWaiter));
    }
  }
}

/* make ready from waiting, sleeping... anything besides ready or running */
PRIVATE INLINE void
Sched_ReadyActivity(
    PACTINFO pAct,
    BOOL PreemptNow
    )
{
  assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE);
    
  if (pAct == NULL) {
    return;
  }

  /* The thread might have died by now, if the readying (helper) thread is
   * preemptible. */
  if (Activity_IsDead(pAct)) {
    return;
  }
  
  /* the thread should not be running, ready.
  /* if the thread is ready, we have already done all this */
  assert(!Activity_IsRunning(pAct));
  assert(!Activity_IsReady(pAct));

  Log_WriteEntry(pSchedLog, LOG_MAKE_THREAD_READY, (UINT)pAct);
  
  if (Activity_IsSleeping(pAct)) {
    /* if this was a timed waiter, remove it from sleep queue */
    List_Remove(CbsSched_GetSleepQueue(), Activity_GetNode(pAct));
  }
  /* Make sure we are not on the sleep queue */
  assert(List_FindData(CbsSched_GetSleepQueue(), pAct) == NULL);
  assert(Node_IsAlone(Activity_GetWaitNode(pAct)));

  /* make thread ready */
  Activity_ClearWaiting(pAct);
  assert(Activity_GetWaitObject(pAct) == NULL);
  CbsSched_ReadyQueueAddActivity(pAct);
  
  /* preempt current thread if appropriate */
  if (PreemptNow) {
    CbsSched_PreemptIfNecessary(NULL);
  }

}

PRIVATE SCODE
Sched_ProcessConstraint(
    PACTINFO pActCurrent,
    PTIME_CONSTRAINT ptcNew,
    PTIME pTimeTaken
    )
{
  TIME tRestartBy;
  PKCONSTRAINT pkcCurrent;
  TIME tNow;

  Log_WriteEntry(pSchedLog, LOG_PROCESS_CONSTRAINT, (UINT)pActCurrent);
  pkcCurrent = Activity_GetConstraint(pActCurrent);

  tNow = CurrentTime();

  if (pTimeTaken != NULL) {
    /* RestartBy time advances when the thread runs, so use difference */
    INT64SUB(*pTimeTaken,
             tNow,
             /*  	       KConstraintGetRestartBy(pkcCurrent), */
             KConstraint_GetInitialRestartBy(pkcCurrent));
  }

  if (ptcNew == NULL) {
    /* we are going back to a timesharing constraint */
    KConstraint_Init(pkcCurrent);
    KConstraint_SetRestartBy(pkcCurrent,
                             Int64Add(tNow, TIME_SHARING_TIMESLICE));
  }
  else {
    /* Do pre-wait BeginConstraint validation */
    if (TimeIsRelative(TConstraint_GetEstimateTime(ptcNew))) {
      Log_WriteEntry(pSchedLog, LOG_PROCESS_INVALID, (UINT)pActCurrent);
      return E_INVALID_PARAMETER;
    }
    
    Sched_NormalizeTimeout(&(ptcNew->Start));

    CbsSched_ResolveConstraint(pActCurrent, ptcNew);

    tRestartBy = KConstraint_GetRestartBy(pkcCurrent);
    if (TimeLess(tRestartBy, CurrentTime())) {
      Log_WriteEntry(pSchedLog, LOG_PROCESS_OUTATIME, (UINT)pActCurrent);
      DBG(("Out of time %ld\n", Int64Subtract(tRestartBy, CurrentTime())));
      return S_OUTATIME;
    }
  }

  return S_OK;
}

/* Move all threads whose start time has arrived from sleep queue to
 * ready queue.  Return start time of earliest thread.
 * If there is no earliest thread, return the timesharing timeslice so that
 * non-real-time threads can still run.
 */
PRIVATE INLINE TIME 
CbsSched_WakeStartingActivities(void)
{
  TIME tUntilStart;
  TIME tNow;
  PACTINFO pActMostUrgent;
  /*   PCONSTRAINT pConstraintMostUrgent; */

  Log_WriteEntry(pSchedLog, LOG_ENTER_WAKE_THREADS, (UINT)NULL);

  assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE);
  /*   assert(CheckStackOverflow(CurrentActivity())); */

  if (!CbsSched_SleepQueueIsEmpty()) {

    tNow = CurrentTime();

    do {
      /* Get most urgent sleeping thread */
      pActMostUrgent = CbsSched_SleepQueueFindMostUrgentActivity();
      /*       pThdMostUrgent = (PTHDINFO)KConstraint_GetSchedObject(pkcMostUrgent); */
      /* We are waking up a waiting thread, which may or may not be idling */
      /* Get most urgent sleeping thread */
      /*       pConstraintMostUrgent = Sched_SleepQueueFindMostUrgentConstraint(); */
      /*       pThdMostUrgent = (PTHDINFO)KConstraint_GetSchedObject(pConstraintMostUrgent); */

      assert(Activity_IsIdle(pActMostUrgent) ||
             Activity_IsSleeping(pActMostUrgent));

      /* Compute time until thread must start. */
      INT64SUB(tUntilStart,
               KConstraint_GetRestartBy(Activity_GetConstraint(pActMostUrgent)),
               tNow);

      if (TimeIsNotPositive(tUntilStart))  {
/*         CbsSched_SleepQueueRemoveMostUrgentActivity(); */
        Log_WriteEntry(pSchedLog, LOG_POP_SLEEP_QUEUE, (UINT)pActMostUrgent);
/*         assert(Node_IsAlone(Activity_GetNode(pActMostUrgent))); */

        if (Activity_IsWaitingForCondition(pActMostUrgent)) {
          /* ReadyActivity takes care of clearing wait state and object */
          Condition_WaiterQueueRemoveActivity
            (Activity_GetWaitObject(pActMostUrgent), pActMostUrgent);
          Activity_SetLastError(pActMostUrgent, E_TIMED_OUT);
          Condition_TimedOut(Activity_GetWaitObject(pActMostUrgent));
          /*  	    KConstraintStopInheritance(pConstraintMostUrgent); */
        }
        assert(Node_IsAlone(Activity_GetWaitNode(pActMostUrgent)));
        Sched_ReadyActivity(pActMostUrgent, FALSE);
      } 
      else {
        /* Thread queued is sorted by increasing starting time, so we only
         * need to check if the first thread's time has arrived.
         *
         * If no one is runnable then TheCurrentThread can keep
         * on going until the wakeup time of pThdMostUrgent.  Otherwise
         * we must cover for the time-sharing case: if the most
         * urgent thread on the run queue is not a RT thread,
         * IThreadUrgencyGTEQ() will not trim down TimeUntilStart.
         * Therefore we have to do it here (conservatively).
         */
        /*  	  if (!Sched_ReadyQueueIsEmpty() && */
        /*  	      (TimeLess(TIME_SHARING_TIMESLICE, tUntilStart))) { */
        /*  	    tUntilStart = TIME_SHARING_TIMESLICE; */
        /*  	  } */
        /*  	  printf("Waking\n"); */
        /*  	  ListTransformData(Sched_GetSleepQueue(), Transform); */
        Log_WriteEntry(pSchedLog, LOG_EXIT_WAKE_REALTIME,
                       Int64ToInt32(tUntilStart));

        return tUntilStart;
      }
    } while (!CbsSched_SleepQueueIsEmpty());
  }

  Log_WriteEntry(pSchedLog, LOG_EXIT_WAKE_TIMESHARE,
                 Int64ToInt32(TIME_SHARING_TIMESLICE));
  return TIME_SHARING_TIMESLICE;
}

/* ===========================================================================
 * Include generic scheduling mechanism here
 */

#include "sched-multi.c"

/* ===========================================================================
 * Interface method implementations
 */

PRIVATE PIACTIVITY MCT
Sched_Reschedule(
    PISCHEDULER Scheduler,
    PACTINFO pActCurrent
    )
{
  TIME tEarliestStart;
  PACTINFO pActMostUrgent;
  BOOL OldContextSwitchNeeded;

/*   assert(Activity_GetThread(pActCurrent) == NULL); */
  assert(pActCurrent != NULL);

  Log_WriteEntry(pSchedLog, LOG_RESCHEDULE, (UINT)pActCurrent);

  /* If there are any deferred condition signals, process them all. */
  if (DeferredConditionSignals) {
    Log_WriteEntry(pSchedLog, LOG_DRAIN_DEFERRED, (UINT)pActCurrent);
    Condition_DrainDeferred();
  }

  /* Move all activities whose start times have passed from the sleep queue
   * to the ready queue and return time until earliest starting thread */
  tEarliestStart = CbsSched_WakeStartingActivities();

  /* Preempt the current activity if necessary due to a ready thread
   * now being more urgent than the current thread. This can happen
   * either because a more urgent thread woke up, or because a
   * ConditionSignal changed the relative urgencies of the
   * current and some previously ready thread. */
  pActMostUrgent = CbsSched_PreemptIfNecessary(&tEarliestStart);

  /* Set the timer for the next start time. */
  OldContextSwitchNeeded = 
    SetNextInterrupt(tEarliestStart, &pSCHED(Scheduler)->ContextSwitchNeeded);

  /*  */
  if (OldContextSwitchNeeded && !Activity_IsIdle(pActCurrent)) {
    pActCurrent = CbsSched_RescheduleInterrupt(pActCurrent,
                                               pActMostUrgent);
  }

/*   WorkItem_Dereference(pActCurrent); */

  Log_WriteEntry(pSchedLog, LOG_OUT, (UINT)pActCurrent);
  return (PIACTIVITY)pActCurrent;
}

/* Removes the given activity from the given scheduler */
void MCT
Sched_RemoveActivity(
    PISCHEDULER This,
    PIACTIVITY pAct
    )
{
  UINT IntrState;
  PACTINFO  pActNew;
  PCXTINFO pProperContext;

  Activity_Prologue();

  /* We can only deal with thread suicide or a surrogate thread killing
   * its work item. Threads can be so cruel. */
  assert((CurrentActivity() == (PIACTIVITY)pActArg) ||
         WorkItem_HasPseudoThread(pActArg));

  Sched_DisablePreemption(This, FALSE);

  /* If this is the last thread then the game is over */
  Sched_DecActCount();
  if (Sched_GetActCount() == 0) {
    BaseDelete();
  }

  Log_WriteEntry(pSchedLog, LOG_DEL_ACTIVITY, (UINT)pActArg);

  Activity_SetState(pActArg, ACTIVITY_DEAD);

  if (Activity_GetThread(pActArg) != NULL) {
    /* if we are not a native thread, fall back through to the scheduler */
    Sched_EnablePreemption(This, FALSE);
    pActArg->v->Release((PIACTIVITY)pActArg);
    Log_WriteEntry(pSchedLog, LOG_DEL_WORKITEM, (UINT)pActArg);
    return;
  }
  /* For suicide, must switch to another thread to reclaim stack. */
/*   do { */
    pActNew = CbsSched_RescheduleInterrupt((PACTINFO)pActArg, NULL);
/*   } while (WorkItem_HasPseudoThread(pActNew)); */

    assert(Activity_GetThread(pActNew) == NULL);
/*   WorkItem_Dereference(pActNew); */

  /* Remove the dying activity from any condition waiter queues.
   * The currently running activity is never waiting on a mutex */

  if (Activity_IsWaitingForCondition(pActArg)) {
    /* This will always be a condition, since the current thread is never
     * waiting on a mutex */
    Condition_WaiterQueueRemoveActivity(Activity_GetWaitObject(pActArg),
                                        pActArg);
    /*        KConstraintStopInheritance(pConstraintRemove); */
  }

  /* Either way, we are now running in a 'safe' thread context.
   * [possibly on the wrong stack though]. */

  /* We cannot free memory while holding the scheduler lock.
   * If we stole a thread, make sure it returns where it should.
   * We might get preempted before we are done with this code. */
  pProperContext = Thread_GetContext(pActNew);

  Sched_EnablePreemption(This, FALSE);

  /* Release the dead thread's resources.
   * This must be guaranteed not to take e.g. the heap mutex,
   * because it might be held by the thread we stole --> deadlock.
   * Note that pConstraintRemove no longer equals what it did above due to
   * a context switch, so we have to explicitly get a valid copy
   * from the dying thread.
   */
  Activity_UnlockMutexes((PACTINFO)pActArg);

  pActArg->v->Release((PIACTIVITY)pActArg); /* BUGBUG stack! */

  /* If this was a suicide, special handling.  */
  TURN_INTERRUPTS_OFF(IntrState);
  Thread_SetContext(pActNew, pProperContext);
  LoadThreadContext((PIACTIVITY)pActNew);
  /* Execution should never reach this point */
  assert(FALSE);
#ifdef _DEBUG
  DebugBreak();
#endif
}

void
SleepUntil(
    TIME tSleep
    )
{
  PACTINFO pActCurrent = (PACTINFO)CurrentActivity();
  assert(pActCurrent != NULL);
  Sched_SleepActivity(pActCurrent, &tSleep);
  /*   return S_OK; */
}

PRIVATE SCODE
CbsSched_SetConstraint(
    PISCHEDULER This,
    PTIME_CONSTRAINT ptcNew,
    PTIME pTimeTaken
    )
{
  PACTINFO pActCurrent;
  SCODE sc;

  Sched_DisablePreemption(pTheScheduler, FALSE);
  pActCurrent = (PACTINFO)CurrentActivity();
  assert(pActCurrent != NULL);
  sc = Sched_ProcessConstraint(pActCurrent, ptcNew, pTimeTaken);
  {
    TIME tRestartBy = 
      KConstraint_GetRestartBy(Activity_GetConstraint(pActCurrent));
    if (TimeLess(CurrentTime(), tRestartBy)) {
      Sched_SleepActivity((PACTINFO)pActCurrent, &tRestartBy);
    }
    else {
      CbsSched_PreemptIfNecessary(NULL);
    }
  }
  Sched_EnablePreemption(pTheScheduler, FALSE);
  return sc;
}
