/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <vmemif.h>

extern PIVMVIEW TheCurrentVmView;	/* XXX */

#pragma warning(disable:4701)	/* vc60 doesnt realize pState is initialized */
/*
 * Low-level context-switch code.
 * Called with interrupt register state on stack (via pushad) == CXTINFO
 * Stack state is (top-down):
 *
 * (high)
 *                saved EFLAGS     ;; 32bits
 *                saved CS         ;; 32bits
 *                saved eip        ;; 32bits
 *                <garbage0>;       ;; 32bits
 *                <garbage1>       ;; 32bits
 *                <garbage2>       ;; 32bits
 *                <garbage3>       ;; 32bits
 *                saved state;     ;; N*32bits
 * PCXTINFO esp-> <garbage4>
 */
__declspec(naked) void _Reschedule(void)
{
    PITHREAD pThread;
    PCXTINFO pState;

    __asm {
#ifdef SIM
	/* A complete frame is needed for the debugger to make stack traces */
	push ebp;
	mov ebp, esp;
	mov	eax,esp;
	sub	esp,__LOCAL_SIZE;
	add eax, 8;
	mov	pState,eax;
#else
	/* Setup pointer to our saved state
	 * We dont have to worry about registers, they are saved already.
	 */
	mov	eax,esp;
	/* xxx adapt isr entry code to fit into CXTINFO struct
	 */
	mov	ebp, esp;
	sub	esp,__LOCAL_SIZE;

	mov	pState,eax;
#endif
    }

    /* Return immediately if we cannot reschedule.
     */
    if (Cbs.PreemptionDisableCount)
    {
	RescheduleNeeded(); /* later */
	/*printf("Hit upon preemption disabled..\n");*/
	goto GetOut;
    }

    /* Remember where we stored our state
     * BUGBUG can this be done with interrupts on ?
     */
    pThread = CurrentThread();

    if (Cbs.FirstThreadBooted)
	pTH(pThread)->CxtInfo = pState;
    Cbs.FirstThreadBooted = TRUE;
#ifdef SIM
Again:
#endif
    /* Rescheduling path cannot be reentrant.
     * Interrupts ok then.
     */
    Cbs.PreemptionDisableCount++;
    ENABLE_INTERRUPTS();

    /* Invoke scheduler to find who should run next
     */
    pThread = CbsReschedule(pTheScheduler, pThread);

    /* Decided who should run next. No more interrupts.
     */
    DISABLE_INTERRUPTS();
    Cbs.PreemptionDisableCount--;
#ifdef SIM
//    assert(Cbs.PreemptionDisableCount == 0);
    /* This shouldn't happen on real hardware but on the simulator it does */
    if (Cbs.ContextSwitchNeeded)
	goto Again;
#endif
    //assert(Cbs.ContextSwitchNeeded == FALSE);
    /* Switch address spaces if necessary */
    if (TheCurrentVmView != pTH(pThread)->pVmView) {
	register PIVMVIEW vv = pTH(pThread)->pVmView;
	if (vv) {
	    TheCurrentVmView = vv;
	    vv->v->SwitchTo(vv);
	}
    }

    pState = pTH(pThread)->CxtInfo;
GetOut:
    HWLoadThreadContext(pState);
}
#pragma warning(default:4701)

/*
 * Voluntary context-switch code.
 * Only must save enough state for a function return.
 * But reload might be from interrupt code above, hence frame.
 */
__declspec(naked) void Reschedule(void)
{
    __asm {
	/* Get return PC
	 */
	pop	eax;

	/* Build IRETD frame (no ring switch, no SS+ESP)
	 */
	pushfd;
	push	cs;
	push	eax;

	/* Build rest of frame. EAX is dont care.
	 */
	push ds;
	push es;
	push gs;
	push fs;
	pushad;
	push 0;			/*  pFpa */

	/* _Reschedule expects interrupts off
	 */
	DISABLE_INTERRUPTS();
#ifdef SIM
	call	_Reschedule;
	int 3;
#else
	jmp	_Reschedule;
#endif
	/* no return */
    }
}

#pragma warning(disable:4100)   /* disable warning C4100: unreferenced param */
#ifdef SIM
void SimInt_CallPending(void);

__declspec(naked) void HWLoadThreadContext(PCXTINFO pCxt)
{
    __asm {
	mov	esp, [esp+4];	/* fetch argument pCxt */

	/* Enable interrupts and check for pending. */
	/* x = AtomicCmpAndSwap(&Simint_State, SIGINT_DISABLED, 0);
	 * if (!x) {
	 *     SimInt_state = SIGINT_DISABLED;
	 *     fixup state;
	 *     goto NextInterrupt;
	 * }
	 */
	mov	eax, SIMINT_DISABLED;
	mov	ecx, OFFSET SimInt_State;
	mov	edx, 0;
#if 0 /* compilerbug */
        lock cmpxchg [ecx],edx; /* ZF==1 IFF exchanged ok */
#else
        __emit 0xf0;   /* lock prefix */
        __emit 0x0f;   /* opcode.1 */
        __emit 0xb1;   /* opcode.2 */
        __emit 0x11;   /* /r: [ecx],edx */
#endif
	/* If it failed, then SIGINT_PENDING must have been set */
	jnz	OutPending;

	pop	ecx;		/* throw away pFpa */
	popad;
	pop	fs;
	pop	gs;
	pop	es;
	pop	ds;
	iretd;

	/* An interrupt is pending.  It is one of the two we handle */
    OutPending:
#if 1
	/* ok, this is not so faithful,
	 * we leave an extra frame here and pop it later
	 */
	call OFFSET SimInt_CallPending;
#else
	mov	SimInt_State, SIMINT_DISABLED;
	/* The state is now the same as at end of SchedulingInterrupt
	 * which jumps to the beginning of this function.
	 */
	call _Reschedule;
#endif

	pop	ecx;		/* throw away pFpa */
	popad;
	pop	fs;
	pop	gs;
	pop	es;
	pop	ds;
	iretd;
    }
}
#else
/* Load state from pCxt (aka longjmp).  Change processor modes if necessary */
__declspec(naked) void HWLoadThreadContext(PCXTINFO pCxt)
{
    __asm {
	mov	esp, [esp+4];	/* fetch argument pCxt */

	pop	ecx;		/* throw away pFpa */
	popad;
	pop	fs;
	pop	gs;
	pop	es;
	pop	ds;
	iretd;
    }
}
#endif
#pragma warning(default:4100)

/*
 * Entry point for direct switch to a thread.
 * Called on thread termination, should just load state and go.
 */
#define LoadThreadContext(_th_) \
    HWLoadThreadContext(pTH(_th_)->CxtInfo)
