/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define OPTIMIZE        /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <fred.h>
#include <base/thread.h>
#include <base/schedul.h>
#include <schedulif.h>

#if defined(_DEBUG)
#define DBG(_x_) printf _x_
#else
#define DBG(_x_)
#endif

/* Null scheduler.  Can be used with one thread only.
 */
#ifdef _MSC_VER
#pragma warning(disable:4100)  /* unreferenced formal parameter */
#endif

typedef struct _CBS {
    CONST_VTBL struct ISchedulerVtbl *v;
    UINT PreemptionDisableCount;    
    BOOL ContextSwitchNeeded;
    BOOL FirstThreadBooted;
} CBS, *PCBS;

CBS Cbs; /* forward */

#define pCBS(_x_) ((PCBS)_x_)

#define RescheduleNeeded() Cbs.ContextSwitchNeeded = TRUE
#define RescheduleNotNeeded() Cbs.ContextSwitchNeeded = FALSE

/* Scheduler object methods. 
 */
PRIVATE IThread * MCT CbsReschedule( IScheduler *This, IThread *pCurrent);
BOOL MCT CbsDisablePreemption( IScheduler *This, BOOL Recursive);
BOOL MCT CbsEnablePreemption( IScheduler *This, BOOL Recursive);
BOOL MCT CbsPreemptionIsEnabled( IScheduler *This);
SCODE MCT CbsAddThread( IScheduler *This, IThread *Thread, PTIME_CONSTRAINT Constraint);
void MCT CbsRemoveThread( IScheduler *This, IThread *Thread);
void MCT CbsWakeupOnMutex(IScheduler *This,PMUTEX pMutex,IThread *Owner);
void MCT CbsWaitForMutex(IScheduler *This,IThread *Thread,PMUTEX pMutex,IThread *Owner);
PITHREAD MCT CbsWaitForCondition(IScheduler *This,IThread *Thread,
                                 PCONDITION pCondition,IThread *FirstThread,
                                 PTIME pTimeOut);
PITHREAD MCT CbsWakeWaitingThreads(IScheduler *This,IThread *FirstThread,
                                   SCODE Result,BOOL fAll);
SCODE MCT CbsConditionTimedWaitAndBeginConstraint(IScheduler *This,
                                                  PCONDITION pCondition,
                                                  PMUTEX pMutex,
                                                  TIME *pTimeout,
                                                  BOOL EndPrevious,
                                                  PTIME_CONSTRAINT pTimeConstraint,
                                                  PTIME pTimeTaken,
                                                  WAITCOND_CALLBACK WaitCallback);
TIME MCT CbsGetIdleTime(IScheduler *This);


static CONST_VTBL struct ISchedulerVtbl CbsVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    CbsReschedule,
    CbsAddThread,
    CbsRemoveThread,
    CbsDisablePreemption,
    CbsEnablePreemption,
    CbsPreemptionIsEnabled,
    CbsWaitForMutex,
    CbsWakeupOnMutex,
    CbsWaitForCondition,
    CbsWakeWaitingThreads,
    CbsConditionTimedWaitAndBeginConstraint,
    CbsGetIdleTime
};

/* Scheduler object
 */
CBS Cbs = {
    &CbsVtbl,
    0,
    FALSE,
    FALSE
};

IScheduler *SchedulerCreate(void)
{
    Cbs.v = &CbsVtbl;
    Cbs.PreemptionDisableCount = 0;
    Cbs.ContextSwitchNeeded = FALSE;
    Cbs.FirstThreadBooted = FALSE;
    return (IScheduler *) &Cbs;
}

/* DEBUG SUPPORT
 */

#define LogIt(x,y)

/*
 * THE GUTS OF THE THING START HERE
 */

#if defined(__NO_BUILTIN_INT64)
const TIME TIME_SHARING_TIMESLICE = Int64Initializer(0,400000 /*Low*/);
#else
#define TIME_SHARING_TIMESLICE   TIME_MILLIS(40)
#endif
PRIVATE IThread * MCT CbsReschedule( IScheduler *This, IThread *pCurrent)
{
    BOOL x;

    if (DeferredConditionSignals)
        DrainDeferredConditions();

    SetNextInterrupt(TIME_SHARING_TIMESLICE, &x);
    return pCurrent;
}

/* Disable preemption, return TRUE if preemption was already disabled.
 */
BOOL MCT CbsDisablePreemption( IScheduler *This, BOOL Recursive)
{
    UINT WhatItIsNow;

    WhatItIsNow = AtomicInc( &pCBS(This)->PreemptionDisableCount );

    assert( Recursive || (WhatItIsNow == 1) );
    UnusedParameter(Recursive);

    return WhatItIsNow != 1;
}

/* Enable preemption, return TRUE if preemption is now enabled
 */
BOOL MCT CbsEnablePreemption( IScheduler *This, BOOL Recursive)
{
    UINT WhatItIsNow;

    WhatItIsNow = AtomicDec( &pCBS(This)->PreemptionDisableCount );

    assert( Recursive || (WhatItIsNow == 0) );
    UnusedParameter(Recursive);

    if (WhatItIsNow == 0 &&
        pCBS(This)->ContextSwitchNeeded) {
        /* pCBS(This)->ContextSwitchNeeded = FALSE; DONT */
        Reschedule();
    }

    return WhatItIsNow == 0;
}


TIME MCT CbsGetIdleTime(IScheduler *This)
{
    return TIME_ORIGIN;
}

BOOL MCT CbsPreemptionIsEnabled( IScheduler *This)
{
    return pCBS(This)->PreemptionDisableCount == 0;
}


SCODE MCT CbsAddThread( IScheduler *This, IThread *Thread,
                       PTIME_CONSTRAINT Constraint)
{
    return E_NOT_IMPLEMENTED;
}

void MCT CbsRemoveThread( IScheduler *This, IThread *Thread)
{
}

void MCT CbsWakeupOnMutex(
                IScheduler *This,
                PMUTEX pMutex,
                IThread *Owner)
{
    UnusedParameter(This);
}

void MCT CbsWaitForMutex(
                IScheduler *This,
                IThread *Thread,
                PMUTEX pMutex,
                IThread *Owner)
{
    UnusedParameter(This);
    DBG(("Waiting for mutex!\n"));
}

PITHREAD MCT CbsWaitForCondition(
                    IScheduler *This,
                    IThread *Thread,
                    PCONDITION pCondition,
                    IThread *FirstThread,
                    PTIME pTimeOut)
{
    return NULL;
}

PITHREAD MCT CbsWakeWaitingThreads(
                    IScheduler *This,
                    IThread *FirstThread,
                    SCODE Result,
                    BOOL fAll)
{
    UnusedParameter(This);
    return NULL;
    //return ThreadWakeWaitingThread(FirstThread,Result,fAll);
}

SCODE MCT CbsConditionTimedWaitAndBeginConstraint(
                IScheduler *This,
                PCONDITION pCondition,
                PMUTEX pMutex,
                TIME *pTimeout,
                BOOL EndPrevious,
                PTIME_CONSTRAINT pTimeConstraint,
                PTIME pTimeTaken,
                WAITCOND_CALLBACK WaitCallback)
{
    return E_NOT_IMPLEMENTED;
}


PRIVATE void NormalizeTimeout(PTIME Timeout)
{
    TIME CurTime;

    CurTime = CurrentTime();

    /* if Timeout is relative to current time, make it absolute.
     * This guarantees a greater-or-equal time than now.
     */
    if (TimeIsNotPositive(*Timeout))
    {
        INT64SUB(*Timeout, CurTime, *Timeout);

    /* if the Timeout time has already occurred, use current time
     */
    }
    else if (TimeLess(*Timeout,CurTime))
    {
        *Timeout = CurTime;
    }
    return;
    DebugBreak();
}

#ifdef _MSC_VER
#pragma warning(disable:4127)  /* conditional expression is constant */
#pragma warning(disable:4702)  /* unreachable code */
#endif


SCODE EndConstraint(PTIME pTimeTaken)
{
    return E_NOT_IMPLEMENTED;
}


SCODE BeginConstraint(BOOL EndPrevious,
                             PTIME_CONSTRAINT TimeConstraint,
                             PTIME pTimeTaken)
{
    return E_NOT_IMPLEMENTED;
}

/* Sleeping interface
 */

SCODE SleepUntil(TIME Until)
{
    PITHREAD pThd;
    TIME t;

    BOOL PreemptionWasDisabled;

    PreemptionWasDisabled = CbsDisablePreemption(pTheScheduler,TRUE);
    LogIt(7,0);

    NormalizeTimeout(&Until);
    pThd = CurrentThread();
    pTH(pThd)->StartTime = Until;
    CbsEnablePreemption(pTheScheduler,PreemptionWasDisabled);
    /* Spin */
    for (;;) {
        t = CurrentTime();
        if (TimeLess(Until,t))
            break;
    }
    return S_OK;
    DebugBreak();
}

/* This function is called to cancel a wait on a condition, either because the
 * wait has timed out or because the thread is being terminated. pThd is the
 * 'first waiter' on the condition; pThdCancel is the thread whose wait is
 * being cancelled.  it may be the same as pThd. Returns a pointer to new
 * 'first waiter', NULL if none.
 */
PITHREAD ThreadConditionCancel(PITHREAD pThd, PITHREAD pThdCancel)
{
    DBG(("Cancel wait of Thd %08x on Cond %08x (first Thd %08x)\n",
         (UINT) pThdCancel, (UINT) pTH(pThd)->WaitObject, (UINT) pThd));

    return NULL;
}

#define SLOWER_LOWSCHED 0
#include "_schedul.c"

/* XXX Stuff this here so that mmsmall builds without cobmgr */
BOOL CobInit(void)
{
    return FALSE;
}

PUBLIC SCODE MCT BaseInitializeClassRegistry (
    void)
{
    return S_OK;
}

void BaseDeleteClassRegistry (void) {}
void CobMgrExit(void) {}
