/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define OPTIMIZE        /* enable Base.h optimizations, since we're the "base" */

#include <mmlite.h>
#include <fred.h>
#include <base\thread2.h>
#include <base\condition2.h>
#include <base\schedul.h>
#include <schedulif.h>
#include <util/list-single.h>

#if defined(_DEBUG)
#define DBG(_x_) printf _x_
#else
#define DBG(_x_)
#endif

/* Null scheduler.  Can be used with one thread only.
 */
#ifdef _MSC_VER
#pragma warning(disable:4100)  /* unreferenced formal parameter */
#endif

typedef struct _CBS {
  CONST_VTBL struct ISchedulerVtbl *v;
  UINT PreemptionDisableCount;    
  BOOL ContextSwitchNeeded;
  PNODE ReadyQueue;
  PNODE SleepQueue;
  INT RefCount;
  UINT nThreads;
  TIME IdleTime;
  BOOL FirstThreadBooted;
} CBS, *PCBS;

PRIVATE CBS Cbs; /* forward */

#define pCBS(_x_) ((PCBS)_x_)

#define RescheduleNeeded() Cbs.ContextSwitchNeeded = TRUE
#define RescheduleNotNeeded() Cbs.ContextSwitchNeeded = FALSE

/* Scheduler object methods. 
 */

PRIVATE SCODE MCT CbsQueryInterface( 
    PISCHEDULER pThis,
    /* [in] */ REFIID pIid,
    /* [out] */ void **ppUnk
    );

PRIVATE UINT MCT CbsAddRef(
    PISCHEDULER pThis
    );

PRIVATE UINT MCT CbsRelease (
     PISCHEDULER pThis
     );

PRIVATE PITHREAD MCT CbsReschedule(
    PISCHEDULER This,
    PITHREAD pThd
);

PRIVATE BOOL MCT CbsDisablePreemption(
     PISCHEDULER This,
     BOOL Recursive
     );

PRIVATE BOOL MCT CbsEnablePreemption(
     PISCHEDULER This,
     BOOL Recursive
     );

PRIVATE BOOL MCT CbsPreemptionIsEnabled(
     PISCHEDULER This
     );

PRIVATE void MCT CbsAddThread(
     PISCHEDULER This,
     PITHREAD Thread,
     PTIME_CONSTRAINT pcval
     );

PRIVATE void MCT CbsRemoveThread(
     PISCHEDULER This,
     PITHREAD Thread
     );

PRIVATE void MCT CbsWakeUpOnMutex(
     PISCHEDULER This,
     PMUTEX pMutex
     );

PRIVATE void MCT CbsWaitForMutex(
     PISCHEDULER This,
     PMUTEX pMutex
     );

PRIVATE void MCT CbsWaitForCondition(
     PISCHEDULER This,
     PCONDITION pCondition,
     PTIME pTimeOut
     );

PRIVATE void MCT CbsWakeUpOnCondition(
     PISCHEDULER This,
     PCONDITION pCondition,
     BOOL fAll
     );

PRIVATE SCODE MCT CbsConditionTimedWaitAndBeginConstraint(
     PISCHEDULER This,
     PCONDITION pCondition,
     PMUTEX pMutex,
     TIME *pTimeout,
     BOOL EndPrevious,
     PTIME_CONSTRAINT pTimeConstraint,
     PTIME pTimeTaken,
     WAITCOND_CALLBACK WaitCallback
     );

PRIVATE TIME MCT CbsGetIdleTime(
     PISCHEDULER This
     );

static CONST_VTBL struct ISchedulerVtbl CbsVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    CbsQueryInterface,
    CbsAddRef,
    CbsRelease,
/*     CbsReschedule, */
    CbsAddThread,
    CbsRemoveThread,
    CbsDisablePreemption,
    CbsEnablePreemption,
    CbsPreemptionIsEnabled,
    CbsWaitForMutex,
    CbsWakeUpOnMutex,
    CbsWaitForCondition,
    CbsWakeUpOnCondition,
    CbsConditionTimedWaitAndBeginConstraint,
    CbsGetIdleTime
};

/* Scheduler object
 */
PRIVATE CBS Cbs = {
    &CbsVtbl,
    0,
    FALSE,
    NULL,
    NULL,
    0,
};

IScheduler *SchedulerCreate(void)
{
    Cbs.v = &CbsVtbl;
    Cbs.PreemptionDisableCount = 0;
    Cbs.ContextSwitchNeeded = FALSE;
    Cbs.ReadyQueue = NULL;
    Cbs.SleepQueue = NULL;
    Cbs.RefCount = 0;
    Cbs.nThreads = 0;
    Int32ToInt64(Cbs.IdleTime,0);
    Cbs.FirstThreadBooted = FALSE;
    return (IScheduler *) &Cbs;
}

/* DEBUG SUPPORT
 */

#define LogIt(x,y)

/*
 * THE GUTS OF THE THING START HERE
 */

#if !defined(_INT64)
const TIME TIME_SHARING_TIMESLICE = Int64Initializer(0,100000/*Low*/);
#else
#define TIME_SHARING_TIMESLICE   100000 /* 40 millisecs */
#endif

PRIVATE SCODE MCT CbsQueryInterface( 
    PISCHEDULER pThis,
    /* [in] */ REFIID pIid,
    /* [out] */ void **ppUnk
    )
{
    return GenericQueryInterface((IUnknown *)pThis, pIid, ppUnk, &IID_IThread);
}
        
PRIVATE UINT MCT CbsAddRef(
    PISCHEDULER pThis
    )
{
    UINT RefCnt;

    RefCnt = AtomicInc((PTR)&Cbs.RefCount);

    return RefCnt;
}
        
PRIVATE UINT MCT CbsRelease (
     PISCHEDULER pThis
     )
{
    UINT RefCnt;

    RefCnt = AtomicDec((PTR)&Cbs.RefCount);

    return RefCnt;
}

PRIVATE PITHREAD MCT CbsReschedule(
    PISCHEDULER This,
    PITHREAD pThd
     )
{
    BOOL x;

    if (DeferredConditionSignals)
        Condition_DrainDeferred();

    SetNextInterrupt(TIME_SHARING_TIMESLICE, &x);
    return CurrentThread();
}

/* Disable preemption, return TRUE if preemption was already disabled.
 */
PRIVATE BOOL MCT CbsDisablePreemption(
     PISCHEDULER This,
     BOOL Recursive
     )
{
    UINT WhatItIsNow;

    WhatItIsNow = AtomicInc( &pCBS(This)->PreemptionDisableCount );

    assert( Recursive || (WhatItIsNow == 1) );
    UnusedParameter(Recursive);

    return WhatItIsNow != 1;
}

/* Enable preemption, return TRUE if preemption is now enabled
 */
PRIVATE BOOL MCT CbsEnablePreemption(
     PISCHEDULER This,
     BOOL Recursive
     )
{
    UINT WhatItIsNow;

    WhatItIsNow = AtomicDec( &pCBS(This)->PreemptionDisableCount );

    assert( Recursive || (WhatItIsNow == 0) );
    UnusedParameter(Recursive);

    if (WhatItIsNow == 0 &&
	pCBS(This)->ContextSwitchNeeded) {
        /* pCBS(This)->ContextSwitchNeeded = FALSE; DONT */
	Reschedule();
    }

    return WhatItIsNow == 0;
}


PRIVATE TIME MCT CbsGetIdleTime(
     PISCHEDULER This
     )
{
    return pCBS(This)->IdleTime;
}

PRIVATE BOOL MCT CbsPreemptionIsEnabled(
     PISCHEDULER This
     )
{
    return pCBS(This)->PreemptionDisableCount == 0;
}


PRIVATE void MCT CbsAddThread(
     PISCHEDULER This,
     PITHREAD Thread,
     PTIME_CONSTRAINT pcval
     )
{
}

PRIVATE void MCT CbsRemoveThread(
     PISCHEDULER This,
     PITHREAD Thread
     )
{
}

PRIVATE void MCT CbsWakeUpOnMutex(
     PISCHEDULER This,
     PMUTEX pMutex
     )
{
    UnusedParameter(This);
}

PRIVATE void MCT CbsWaitForMutex(
     PISCHEDULER This,
     PMUTEX pMutex
     )
{
    UnusedParameter(This);
    DBG(("Waiting for mutex!\n"));
}

PRIVATE void MCT CbsWaitForCondition(
     PISCHEDULER This,
     PCONDITION pCondition,
     PTIME pTimeOut
     )
{
    return;
}

PRIVATE void MCT CbsWakeUpOnCondition(
     PISCHEDULER This,
     PCONDITION pCondition,
     BOOL fAll
     )
{
    UnusedParameter(This);
    return;
}

PRIVATE SCODE MCT CbsConditionTimedWaitAndBeginConstraint(
     PISCHEDULER This,
     PCONDITION pCondition,
     PMUTEX pMutex,
     PTIME pTimeout,
     BOOL EndPrevious,
     PTIME_CONSTRAINT pTimeConstraint,
     PTIME pTimeTaken,
     WAITCOND_CALLBACK WaitCallback)
{
    return E_NOT_IMPLEMENTED;
}


PRIVATE void NormalizeTimeout(PTIME Timeout)
{
    TIME CurTime;

    CurTime = CurrentTime();

    /* if Timeout is relative to current time, make it absolute.
     * This guarantees a greater-or-equal time than now.
     */
    if (TimeIsNotPositive(*Timeout))
    {
        INT64SUB(*Timeout, CurTime, *Timeout);

    /* if the Timeout time has already occurred, use current time
     */
    }
    else if (TimeLess(*Timeout,CurTime))
    {
        *Timeout = CurTime;
    }
    return;
    DebugBreak();
}

#ifdef _MSC_VER
#pragma warning(disable:4127)  /* conditional expression is constant */
#pragma warning(disable:4702)  /* unreachable code */
#endif


SCODE EndConstraint(PTIME pTimeTaken)
{
    return E_NOT_IMPLEMENTED;
}


SCODE BeginConstraint(BOOL EndPrevious,
                             PTIME_CONSTRAINT TimeConstraint,
			     PTIME pTimeTaken)
{
    return E_NOT_IMPLEMENTED;
}

/* Sleeping interface
 */

SCODE SleepUntil(TIME Until)
{
    PITHREAD pThd;
    TIME t;

    BOOL PreemptionWasDisabled;

    PreemptionWasDisabled = CbsDisablePreemption(pTheScheduler,TRUE);
    LogIt(7,0);

    NormalizeTimeout(&Until);
    pThd = CurrentThread();
/*      pTH(pThd)->StartTime = Until; */
    CbsEnablePreemption(pTheScheduler,PreemptionWasDisabled);
    /* Spin */
    for (;;) {
	t = CurrentTime();
	if (TimeLess(Until,t))
	    break;
    }
    return S_OK;
    DebugBreak();
}

/* This function is called to cancel a wait on a condition, either because the
 * wait has timed out or because the thread is being terminated. pThd is the
 * 'first waiter' on the condition; pThdCancel is the thread whose wait is
 * being cancelled.  it may be the same as pThd. Returns a pointer to new
 * 'first waiter', NULL if none.
 */
PITHREAD ThreadConditionCancel(PITHREAD pThd, PITHREAD pThdCancel)
{
    DBG(("Cancel wait of Thd %08x on Cond %08x (first Thd %08x)\n",
	 (UINT) pThdCancel, (UINT) pTH(pThd)->pWaitObject, (UINT) pThd));

    return NULL;
}

#define SLOWER_LOWSCHED 0
#include "_schedul.c"

/* XXX Stuff this here so that mmsmall builds without cobmgr */
BOOL CobInit(void)
{
    return FALSE;
}

PUBLIC SCODE MCT BaseInitializeClassRegistry (
    void)
{
    return S_OK;
}

void BaseDeleteClassRegistry (void) {}
void CobMgrExit(void) {}
