/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include "sched.c"

/* Adds the given activity to the given scheduler */
PRIVATE SCODE MCT 
Sched_AddActivity(
    PISCHEDULER This,
    PIACTIVITY pAct,
    PTIME_CONSTRAINT ptcNew
    )
{
  SCODE sc;

  Activity_Prologue();
    
  Sched_DisablePreemption(This, FALSE);
  sc = Sched_ProcessConstraint((PACTINFO)pAct, ptcNew, NULL);
  if (sc != S_OK) {
    /* If processing the constraint failed, fall back to timesharing */
    Log_WriteEntry(pSchedLog, LOG_ADD_FAILED, (UINT)pAct);
    ptcNew = NULL;
  }

  if (WorkItem_HasPseudoThread(pAct)) {
    /* For work items, copy time constraint for later transfer to surrogate */
/*     if (ptcNew != NULL) { */
/*       WorkItem_SetThreadConstraint(pAct, *ptcNew); */
/*     } */
    Log_WriteEntry(pSchedLog, LOG_ADD_WORKITEM, (UINT)pActArg);
  }

  /* Increment the thread count */
  Sched_IncActCount();
  {
    TIME tRestartBy = 
      KConstraint_GetRestartBy(Activity_GetConstraint(pActArg));
    if (TimeLess(CurrentTime(), tRestartBy)) {
      Sched_SleepActivity((PACTINFO)pActArg, &tRestartBy);
    }
    else {
      Sched_ReadyActivity((PACTINFO)pActArg, TRUE);
    }
  }

  Sched_EnablePreemption(This, FALSE);
/*   RescheduleNeeded(); */
  Log_WriteEntry(pSchedLog, LOG_ADD_THREAD, (UINT)pActArg);
/*   Log_Print(pSchedLog); */
  return sc;
}

/* Wake all threads blocked on mutex pMutex. This mutex must have an
 * owner (caught by assertion in StopConstraintInheritance).
 */
PRIVATE void MCT
Sched_WakeUpOnMutex(
    PISCHEDULER This,
    PMUTEX pMutex
    )
{
  PACTINFO pActMostUrgent;
  PACTINFO pActCurrent;

  Mutex_Prologue();

/*   UnusedParameter(This); */

  pActCurrent = (PACTINFO)CurrentActivity();

  assert(pMutexArg != NULL);
  /* If waiter queue is empty, Mutex_Lock should not have called us */
  assert(!Mutex_WaiterQueueIsEmpty(pMutexArg));

  /* Remove most urgent thread on the mutex's waiter queue,
   * set its owner to NULL, and recalculcate inheritance. */
  pActMostUrgent = Mutex_WaiterQueueRemoveMostUrgent(pMutexArg);
  assert(pActMostUrgent != NULL);

  /* This doesn't actually remove the waiter, but takes care of scheduler
   * policy-specific behaviour */
  Sched_MutexRemoveWaiter((PMX)pMutexArg, pActCurrent);

  assert(Activity_IsWaitingForMutex(pActMostUrgent));
  Mutex_SetSchedObject(pMutexArg, pActMostUrgent);
  if (!Mutex_WaiterQueueIsEmpty(pMutexArg)) {
    Mutex_SetTaken(pMutexArg);
  }
  Activity_MutexListAdd(pActMostUrgent, pMutexArg);
  /* This doesn't actually lock the mutex, but takes care of scheduler
   * policy-specific behaviour */
  Sched_MutexRelock((PMX)pMutexArg, pActMostUrgent);
  
  Sched_ReadyActivity(pActMostUrgent, TRUE);
  assert(!Activity_IsWaitingForMutex(pActCurrent));
}

PRIVATE void MCT
Sched_WakeUpOnCondition(
    PISCHEDULER This,
    PCONDITION pCond,
    BOOL fAll,
    PTIME_CONSTRAINT ptcNew
    )
{
  Condition_Prologue();

  assert(pCond != NULL);
  /* If waiter queue is empty, UnwaitCond should not have called us */
  assert(!Condition_WaiterQueueIsEmpty(pCondArg));

  Log_WriteEntry(pSchedLog, LOG_ENTER_CONDITION_SIGNAL, (UINT)pCondArg);

  if (fAll == TRUE) {
    /* Go through the list of waiters, remove them, and wake them up */
    List_TransformDataWithArg(Condition_GetWaiterQueue(pCondArg),
                              Sched_ConditionWakeupTransform, ptcNew);
    /* Clear the list */
    List_Init(Condition_GetWaiterQueue(pCondArg));
    Log_WriteEntry(pSchedLog, LOG_EXIT_CONDITION_SIGNAL, (UINT)0);
  }
  else {
    PACTINFO pActMostUrgent;

    pActMostUrgent =
      Condition_WaiterQueueRemoveMostUrgent(pCondArg);
    assert(pActMostUrgent != NULL);
    assert(List_FindData(Condition_GetWaiterQueue(pCondArg),
                         pActMostUrgent) == NULL);

    Sched_ConditionWakeupTransform(pActMostUrgent, ptcNew);
    Log_WriteEntry(pSchedLog, LOG_EXIT_CONDITION_SIGNAL,
                   (UINT)pActMostUrgent);
    assert(Activity_GetWaitObject(pActMostUrgent) == NULL);
  }      
  /* This doesn't actually remove the waiter, it just takes care of scheduler
   * policy-specific behaviour */
  Sched_ConditionRemoveWaiter((PCOND)pCond, NULL);
}

PRIVATE void MCT
Sched_WaitForMutex(
    PISCHEDULER This,
    PMUTEX pMutex
    )
{
  PACTINFO pActWaiter = (PACTINFO)CurrentActivity();

  Mutex_Prologue();

/*   UnusedParameter(This); */

  assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE);
  assert(pMutexArg != NULL);
  assert(Activity_IsRunning(pActWaiter));

  /* Add constraint to mutex's waiter queue and start inheritance */
  /* This adds the waiter in an order specified by the scheduler policy. */
  Sched_MutexAddWaiter((PMX)pMutexArg, pActWaiter);
  Activity_SetWaiting(pActWaiter, ACTIVITY_WAIT_MUTEX, pMutexArg);
  RescheduleNeeded();
}

PRIVATE void MCT
Sched_WaitForCondition(
    PISCHEDULER This,
    PCONDITION pCond,
    PTIME pTimeOut
    )
{
  PACTINFO pActWaiter;
/*   PKCONSTRAINT pConstraintWaiter; */

  Condition_Prologue();

/*   UnusedParameter(This); */
  pActWaiter = (PACTINFO)CurrentActivity();

  assert(Sched_PreemptionIsEnabled(pTheScheduler) == FALSE);
  assert(pCondArg != NULL);
  assert(Activity_IsRunning(pActWaiter));

  Log_WriteEntry(pSchedLog, LOG_ENTER_CONDITION_WAIT, (UINT)pActWaiter);
  assert(Node_IsAlone(Activity_GetNode(pActWaiter)));

  if (Activity_GetThread(pActWaiter) != NULL) {
    /* if this is a work item, dereference to get the surrogate thread */
    pActWaiter = Activity_GetThread(pActWaiter);
  }
  /* pActWaiter is now a thread, and we can call thread methods on it */

/*   pConstraintWaiter = Activity_GetConstraint(pActWaiter); */

  /* This doesn't actually add the waiter, it just takes care of scheduler
   * policy-specific behaviour. */
  Activity_SetWaiting(pActWaiter, ACTIVITY_WAIT_CONDITION, pCondArg);
  Sched_ConditionAddWaiter((PCOND)pCondArg, pActWaiter, pTimeOut);

  assert(!Activity_IsRunning(pActWaiter));
  RescheduleNeeded();
  Log_WriteEntry(pSchedLog, LOG_EXIT_CONDITION_WAIT, (UINT)pActWaiter);
}

PRIVATE void
Sched_ConditionWakeupTransform(
    PTR pActArg,
    PTR ptcNewArg
    )
{
  PACTINFO pActWaiter = (PACTINFO)pActArg;
  PTIME_CONSTRAINT ptcNew = (PTIME_CONSTRAINT)ptcNewArg;
  
  assert(pActWaiter != NULL);
  assert(Activity_IsWaitingForCondition(pActWaiter));

  /* Make it alone */
  Node_SetNext(Activity_GetWaitNode(pActWaiter), NULL);

  /*      KConstraintStopInheritance(pkcWaiter); */
  Sched_ReadyActivity(pActWaiter, FALSE);
  if (ptcNew != NULL) {
    SCODE sc;
    sc = Sched_ProcessConstraint((PACTINFO)pActWaiter, ptcNew, NULL);
    Activity_SetLastError(pActWaiter, sc);
  }

  assert(Activity_GetWaitObject(pActWaiter) == NULL);
  assert(!Activity_IsWaitingForCondition(pActWaiter));
}

PRIVATE SCODE MCT
Sched_WaitForConditionWithConstraint(
    PISCHEDULER This,
    PCONDITION pCond,
    PMUTEX pMutex,
    PTIME pTimeout,
    PTIME_CONSTRAINT ptcNew,
    PTIME pTimeTaken
    )
{
  SCODE sc;
  SCODE sc2;
  PKCONSTRAINT pkcCurrent;
  PACTINFO pActCurrent;

  Condition_Prologue();
  Mutex_Prologue();

  pActCurrent = (PACTINFO)CurrentActivity();
  pkcCurrent = Activity_GetConstraint(pActCurrent);

  Sched_DisablePreemption(This, FALSE);

  
  /* Do pre-wait BeginConstraint validation */
  if (!TimeIsRelative(TConstraint_GetEstimateTime(ptcNew))) {
    Sched_NormalizeTimeout(&(ptcNew->Start));
    /* Validate user-level time constraint before the yield. */
/*     Sched_ConditionWaitWithConstraintBefore(pActCurrent, ptcNew); */

    /* Now enable pre-emption to block, then redisable */
    Sched_EnablePreemption(This, FALSE);
    /* This is the yield */
    Sched_DisablePreemption(This, FALSE);
      
    sc = Activity_GetLastError(pActCurrent);
      
    /* Do after-yield processing */
    sc2 = Sched_SetConstraint(This, ptcNew, pTimeTaken); 
/*     sc2 = Sched_ConditionWaitWithConstraintAfter((PCOND)pCond, pActCurrent, */
/*                                                  pTimeTaken); */
    CbsSched_UpdateRestartBy(pActCurrent);

    /* Return cond wait status, unless constraint failure is predicted */
    assert(sc2 == S_OK || sc2 == S_OUTATIME);
    if (sc2 == S_OUTATIME) {
      sc = S_OUTATIME;
    }

  }

  Sched_EnablePreemption(This, FALSE);

  /* Must return with the mutex locked in *all* cases */
  if (pMutexArg != NULL) {
    Mutex_Lock((PMUTEX)pMutexArg);
  }

  return sc;
}

