/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
   TODO
   delayed writes
   large I/Os
   read-ahead

   BUGS
   all buffers on list #3 (hash on clusterno!!)
*/
#define VERBOSE 0

#include <mmlite.h>
#include <stdio.h>
#include "fatfs.h"
#include "listfunc.h"

#include <bufcache.h>

/* These two constants/variables can be tuned
 * by machdep defines, possibly to dynamically
 * match the amount of available main memory.
 */
#ifndef BCACHE_MAX_BUFFERS
#define BCACHE_MAX_BUFFERS 16
#endif

/* Must be 2^N */
#ifndef BCACHE_NLISTS
#define BCACHE_NLISTS 4
#endif

#define HashBlockNo(_b_)  ((_b_) & (BCACHE_NLISTS-1))

typedef struct {
    LISTNODE            Link;
    UINT                Tag;
    DISK_ADDRESS        BlockNo;
    UINT                RefCnt;
    UINT16              BlkSize;
    UINT16              State;   /* One of BLK_DIRTY&co */
    BYTE                *Data;
} CACHED_BLOCK, *PCACHED_BLOCK;

typedef struct {
    MUTEX       Lock;
    CONDITION   WaitPoint;
    LISTNODE    Buffers;
} BUFFER_LIST;


typedef struct {
    MUTEX       Lock;   /* synchronize writers */
    CONDITION   WaitPoint;
    UINT        BuffersToAllocate;
    BOOL        Inited;
    UINT        Hits;
    UINT        Misses;
    BUFFER_LIST Heads[BCACHE_NLISTS];
} BUFCACHE;

/* Initialize the buffer cache
 * BUGBUG should be system-wide, and
 * accessible to user-level filesystems.
 */
PRIVATE BUFCACHE TheBufferCache = {MUTEX_INITIALIZER,CONDITION_INITIALIZER,0, }; /* bad initializer but .. */

void BcacheInit(void)
{
    int i;

    if (TheBufferCache.Inited)
        return;
    TheBufferCache.Inited = TRUE;

    Mutex_Init( &TheBufferCache.Lock);
    Condition_Init( &TheBufferCache.WaitPoint);
    TheBufferCache.BuffersToAllocate = BCACHE_MAX_BUFFERS;
    TheBufferCache.Hits = TheBufferCache.Misses = 0;

    for (i = 0; i < BCACHE_NLISTS; i++) {
        Mutex_Init(&TheBufferCache.Heads[i].Lock);
        Condition_Init(&TheBufferCache.Heads[i].WaitPoint);
        ListInitialize(TheBufferCache.Heads[i].Buffers);
    }
}

/* Finalize the buffer cache, release all memory
 */
void BcacheDestroy(void)
{
    int i;
    BUFFER_LIST *pList;
    LISTNODE    *pBuf;
    CACHED_BLOCK *pBlk;
    PIHEAP pIHeap = CurrentHeap();

    if (!TheBufferCache.Inited)
        return;

    Mutex_Destroy( &TheBufferCache.Lock);
    Condition_Destroy( &TheBufferCache.WaitPoint);

    for (i = 0; i < BCACHE_NLISTS; i++) {
        Mutex_Destroy(&TheBufferCache.Heads[i].Lock);
        Condition_Destroy(&TheBufferCache.Heads[i].WaitPoint);

        pList = &TheBufferCache.Heads[ i ];
        pBuf = pList->Buffers.pNext;

        while (pBuf != &pList->Buffers) {
            pBlk = (CACHED_BLOCK *)pBuf;
            pBuf = pBuf->pNext;

            if (pBlk->Data)
                pIHeap->v->Free(pIHeap,0,pBlk->Data);
            pIHeap->v->Free(pIHeap,0,pBlk);
        }

        ListInitialize(TheBufferCache.Heads[i].Buffers);
    }

    TheBufferCache.Inited = FALSE;
}

/* Return a pointer to the actual data cached
 * for a block
 */
BYTE *BcacheGetBufferData( PTR pBuf)
{
    PCACHED_BLOCK pBlk = (PCACHED_BLOCK)pBuf;

    return pBlk->Data;
}

/* Return the amount of data cached in a block
 */
UINT BcacheGetBufferSize( PTR pBuf)
{
    PCACHED_BLOCK pBlk = (PCACHED_BLOCK)pBuf;

    return pBlk->BlkSize * MIN_BCACHE_BLOCK_SIZE;
}

/* Return a the block number for a buffer
 */
DISK_ADDRESS BcacheGetBufferBlockNo( PTR pBuf)
{
    PCACHED_BLOCK pBlk = (PCACHED_BLOCK)pBuf;

    return pBlk->BlockNo;
}

/* Find a block for reading. If not found,
 * read it in and cache it.
 */
SCODE BcacheLookupRead( PIFILE Device, DISK_ADDRESS Block, UINT BlkSize, PTR *ppRet)
{
    BUFFER_LIST *pList;
    LISTNODE    *pBuf;
    BOOL        fFound;
    CACHED_BLOCK *pBlk, *pTmp;
    SCODE       sc;
    PIHEAP pIHeap;

    /* Roundup to block size
     */
    BlkSize = (BlkSize + MIN_BCACHE_BLOCK_SIZE - 1) / MIN_BCACHE_BLOCK_SIZE;

    /* Lock and scan the proper list
     */
    pList = &TheBufferCache.Heads[ HashBlockNo(Block) ];
    pBuf = &pList->Buffers;
    fFound = FALSE;
    pBlk = NULL; /* compilerbrain */
    pTmp = NULL; /* compilerbrain */

    DBGOUT(DBGFS,0,_tprintf(_T("BcacheLRead %x %x %d %d\n"), 
                          (ADDRESS)Device, Block, BlkSize, HashBlockNo(Block)));

    Mutex_Lock(&pList->Lock);

    while (pBuf->pNext != &pList->Buffers) {

        pBuf = pBuf->pNext;
        pBlk = (CACHED_BLOCK *)pBuf;

        if (pBlk->Tag == (UINT)Device &&
            pBlk->BlockNo <= Block &&
            (pBlk->BlockNo + pBlk->BlkSize) > Block) {

            /* Got it, or something overlapping anyways
             */
            fFound = TRUE;
            break;
        }
    }

    if (fFound) {
        /* Bump refcount so that it does not go away
         */
        AtomicAdd(&pBlk->RefCnt, 1);
    }

    Mutex_Unlock(&pList->Lock);


#define mediacheck 1
#if mediacheck
    if (fFound) {
        /* Before we trust this block, check if the device
         * has removable media, and if so if the media changed.
         */
        UINT64 Inf;
        Int64FromHighAndLow(Inf,~0,~0);
        sc = Device->v->SetSize( Device, Inf );
        if (sc == E_MEDIA_CHANGED) {
            /* Urgh.
             * Leave it up to the filesystem to deal with
             * any dirty block left in the cache. E.g. the
             * only sensible thing to do might be to ask
             * the user pls put it back in...
             */
            AtomicSub(&pBlk->RefCnt, 1);
            return sc;
        }
    }
    /* If we did not find the block, and the media has changed,
     * we will find out in the read below and handle it normally.
     */
#endif

    if (fFound) {
        AtomicAdd(&TheBufferCache.Hits, 1);
        *ppRet = (PTR) pBlk;
        return NO_ERROR;
    }

    AtomicAdd(&TheBufferCache.Misses, 1);
#if VERBOSE
    _tprintf(_T("{Miss %x [%d/%d]}"), 
             Block,
             TheBufferCache.Hits,
             TheBufferCache.Misses);
#endif

    /* Mumble, we need a new block.
     * Can we just get a new one.
     */
    pIHeap = CurrentHeap();

    Mutex_Lock(&TheBufferCache.Lock);

    if (TheBufferCache.BuffersToAllocate > 0) {

        /* Get a fresh block
         */

        TheBufferCache.BuffersToAllocate--;
        Mutex_Unlock(&TheBufferCache.Lock);

        pBlk = (CACHED_BLOCK *) pIHeap->v->Alloc(pIHeap, HEAP_ZERO_MEMORY, sizeof(CACHED_BLOCK), 0);
        DBGOUT(DBGFS,0,_tprintf(_T("blr %x(%d)\n"),(ADDRESS)pBlk,sizeof(CACHED_BLOCK)));
        if (pBlk == NULL)
            return E_OUT_OF_STRUCTURES;

        ListInitialize(pBlk->Link);

    } else {
        /* Must wait for a free one
         */
        UINT i;

        pBlk = NULL;

    retry:
        /* This is where we implement the replacement strategy.
         * An easy and effective strategy is as follows.
         *
         * Look in all free lists, starting at the bottom of
         * each list (active blocks are moved to front).
         * Start looking from the list where the block will go.
         *
         * The optimal approach is LRU with second chance. To
         * implement that we could use a separate list of free
         * blocks, and move non-referenced blocks to the list
         * when its found empty.  This is *exactly* like what
         * an OS that we wont name did for its VM policy...
         * ..but we know how much code that took.
         */
#define FOLLOW pPrev

        for (i = HashBlockNo(Block); pBlk == NULL; ) {

            pList = &TheBufferCache.Heads[ i ];
            pBuf = &pList->Buffers;

            Mutex_Lock(&pList->Lock);

            pBuf = pBuf->FOLLOW;
            while (pBuf != &pList->Buffers) {

                pTmp = (CACHED_BLOCK *)pBuf;
                pBuf = pBuf->FOLLOW;

                /* xxx perhaps two separate passes would be
                 * better. But I bet tag==0 is infrequent.
                 */
                if ((pTmp->Tag == 0) || (pTmp->RefCnt == 0)) {
                    pBlk = pTmp;
                    (void) ListRemove(&pBlk->Link);
#if VERBOSE
                    _tprintf(_T("(%x %x -> %x)"), (ADDRESS)pBlk, pBlk->BlockNo, Block);
#endif
                    break;
                }
            }

            Mutex_Unlock(&pList->Lock);

            if (++i == BCACHE_NLISTS)
                i = 0;
            if (i == HashBlockNo(Block))
                break;
        }
#undef FOLLOW

        if (pBlk == NULL) {
            Condition_Wait( &TheBufferCache.WaitPoint,
                            &TheBufferCache.Lock,
                            NULL);
            goto retry;
        }

        Mutex_Unlock(&TheBufferCache.Lock);
    }

    /* A delayed-write block?
     */
    if (pBlk->State & BLK_DIRTY)
    {
        pBlk->RefCnt = 2; /* do not signal */
        BcacheSpillBuffer( pBlk, TRUE);
    }

    /* Got a free block, fill it in and add it to the list
     * Make sure we make no duplicates.
     */
    pBlk->BlockNo = Block;
    pBlk->Tag = (UINT)Device;
    pBlk->RefCnt = 1;
    pBlk->State = BLK_INVALID_DATA;

    /* While we hold no locks, commit memory for block
     */
    if (pBlk->BlkSize < (UINT16) BlkSize)
    {
        if (pBlk->Data != NULL)
            pIHeap->v->Free(pIHeap,0,pBlk->Data);

        pBlk->Data = pIHeap->v->Alloc(pIHeap,0,BlkSize * MIN_BCACHE_BLOCK_SIZE, 0);
        DBGOUT(DBGFS,0,_tprintf(_T("blr2 %x(%d)\n"),(ADDRESS)pBlk->Data,BlkSize*MIN_BCACHE_BLOCK_SIZE));
        if (pBlk->Data == NULL) {
            pIHeap->v->Free(pIHeap,0,pBlk);
            return E_OUT_OF_STRUCTURES;
        }
    }
    pBlk->BlkSize = (UINT16) BlkSize;


    pList = &TheBufferCache.Heads[ HashBlockNo(Block) ];
    pBuf = &pList->Buffers;

    Mutex_Lock(&pList->Lock);

    while (pBuf->pNext != &pList->Buffers) {

        pBuf = pBuf->pNext;
        pTmp = (CACHED_BLOCK *)pBuf;
        if (pTmp->Tag == (UINT)Device &&
            pTmp->BlockNo <= Block &&
            (pTmp->BlockNo + pTmp->BlkSize) > Block) {

            /* Someone raced us to it, 
             * take a ref for later and invalidate duplicate
             */
            AtomicAdd(&pTmp->RefCnt, 1);
            pBlk->Tag = 0;
            pBlk->RefCnt = 0;
            break;
        }
    }

    /* Add buffer to list, wether valid or not.
     * If invalid, it becomes a free one.
     * xxx invalid should go at end
     */
    fFound = (pBlk->Tag == 0);
    ListInsert(&pList->Buffers, &pBlk->Link);

    /* See if we have to wait for read complete
     */
    while (fFound && (pTmp->State == BLK_INVALID_DATA))
        Condition_Wait(&pList->WaitPoint, &pList->Lock, NULL);


    Mutex_Unlock(&pList->Lock);

    /* We might not have to do the read
     */
    if (fFound) {
        *ppRet = (PTR) pTmp;
        return NO_ERROR;
    }

    /* Nope, we gotta read it in
     */
    {
        UINT ReadSize;
        UINT64 Bo;
        Uint32ToUint64(Bo,Block * MIN_BCACHE_BLOCK_SIZE);
        sc = Device->v->ReadAt( Device, Bo, pBlk->Data, BlkSize * MIN_BCACHE_BLOCK_SIZE,
                               &ReadSize);
    }

    /* Anyone was waiting, wake em up
     */
    pBlk->State = 0;
    if (pBlk->RefCnt > 1)
        Condition_Signal(&pList->WaitPoint);

    /* Done, at the long last.
     */
    *ppRet = (PTR) pBlk;
    return sc;
}


/* Find a block for writing. If not found,
 * read it in and cache it.
 */
SCODE BcacheLookupWrite( PIFILE Device, DISK_ADDRESS Block, UINT BlkSize, PTR *ppRet)
{
    SCODE sc;
    DBGOUT(DBGFS,1,_tprintf(_T("BcacheLWrite %x %x %d\n"), 
                          (ADDRESS)Device, Block, BlkSize));
    sc = BcacheLookupRead( Device, Block, BlkSize, ppRet);
#if 0
    /* It might not get written afterall
     */
    if (!FAILED(sc))
        ((PCACHED_BLOCK)*ppRet)->State = BLK_DIRTY;
#endif
    return sc;
}

/* Release a buffer. If modified, ignore.
 */
void BcacheFlushBuffer( PTR pBuf, BOOL RidOfIt )
{
    CACHED_BLOCK *pBlk = (CACHED_BLOCK *)pBuf;
    INT Refs;

    DBGOUT(DBGFS,0,_tprintf(_T("BcacheFBuffer %x %x %d %d\n"), 
                          (ADDRESS)pBuf, pBlk->BlockNo, pBlk->BlkSize, RidOfIt));

    /* Take one ref away
     */
    Refs = AtomicSub(&pBlk->RefCnt, 1);

    /* If this is not the last reference, done.
     */
    if (Refs > 0)
        return;

    assert(Refs == 0);

    /* Someone might be waiting for a free block
     */
    Condition_Signal(&TheBufferCache.WaitPoint);
}

/* Release a buffer. If modified, write it back.
 */
void BcacheSpillBuffer( PTR pBuf, BOOL RidOfIt  )
{
    PCACHED_BLOCK pBlk = (PCACHED_BLOCK)pBuf;
    UINT          WriteSize, rb;
    PIFILE        Device;
    SCODE         sc;
    UINT64        Offset64;

    DBGOUT(DBGFS,0,_tprintf(_T("BcacheSBuffer %x %x %d %d (%x)\n"), 
                          (ADDRESS)pBuf, pBlk->BlockNo, pBlk->BlkSize,
                          RidOfIt, pBlk->State));

    /* Write back if necessary
     */
    if (RidOfIt && (pBlk->State & BLK_DIRTY))
    {
        Device = (PIFILE) pBlk->Tag;
        WriteSize = pBlk->BlkSize * MIN_BCACHE_BLOCK_SIZE;

        Uint32ToUint64(Offset64,pBlk->BlockNo * MIN_BCACHE_BLOCK_SIZE);
        sc = Device->v->WriteAt( Device, Offset64,
                                pBlk->Data, WriteSize, &rb);
        
        if (!FAILED(sc) && rb == WriteSize)
            pBlk->State &= ~BLK_DIRTY;
        else {
            _tprintf(_T("Bcache write error = %x\n"), sc);
            /* Release buffer as dirty. Will retry write.
             */
        }
    }

    /* Buffer is clean, release it
     */
    BcacheFlushBuffer(pBuf,RidOfIt);
}

/* Release all buffers with a given tag.
 * Caller picks buffer disposition
 */
PRIVATE void BcacheScanTag( UINT Tag, BOOL IsInvalid )
{
    BUFFER_LIST *pList;
    LISTNODE    *pBuf;
    int         i;
    BOOL        Found = FALSE;

    DBGOUT(DBGFS,1,_tprintf(_T("BcacheSTag %x %d\n"), Tag, IsInvalid));

    for (i = 0; i < BCACHE_NLISTS; i++) {


        pList = &TheBufferCache.Heads[ i ];
        pBuf = &pList->Buffers;

        Mutex_Lock(&pList->Lock);

        pBuf = pBuf->pNext;
        while (pBuf != &pList->Buffers) {
            CACHED_BLOCK *pBlk;

            pBlk = (CACHED_BLOCK *)pBuf;
            pBuf = pBuf->pNext;

            if (pBlk->Tag == Tag) {

                Found = TRUE;

                if (IsInvalid) {

                    /* BUGBUG what if in use 
                     */
                    pBlk->RefCnt = 0;
                    pBlk->Tag = 0;
                    pBlk->State = 0;

                } else {
                
                    /* Just clean the block, no invalidates
                     */
                    if (pBlk->State & BLK_DIRTY) {
                        AtomicAdd( &pBlk->RefCnt, 1);
                        BcacheSpillBuffer( pBlk, TRUE );
                    }

                    /* else its ok
                     */
                }
            }
        }

        Mutex_Unlock(&pList->Lock);

    }

    if (Found)
        Condition_Signal(&TheBufferCache.WaitPoint);
}

/* Release all buffers with a given tag.
 * Throw away the modified ones.
 */
void BcacheFlushTag( UINT Tag )
{
    DBGOUT(DBGFS,0,_tprintf(_T("BcacheFlushTag %x\n"), Tag));
    BcacheScanTag( Tag, TRUE );
}

/* Release all buffers with a given tag.
 * Write back the modified ones.
 */
void BcacheSpillTag( UINT Tag )
{
    DBGOUT(DBGFS,0,_tprintf(_T("BcacheSpillTag %x\n"), Tag));
    BcacheScanTag( Tag, FALSE );
}


/* Get buffer status bits
 */
UINT16 BcacheGetBufferState( PTR pBuf)
{
    PCACHED_BLOCK pBlk = (PCACHED_BLOCK)pBuf;

    return pBlk->State;
}

/* Modify buffer status bits, add to them
 */
SCODE BcacheAddToBufferState( PTR pBuf, UINT16 Bits)
{
    PCACHED_BLOCK pBlk = (PCACHED_BLOCK)pBuf;

    pBlk->State |= Bits; /* BUGBUG atomicity */
    return NO_ERROR;
}

/* Modify buffer status bits, clear some of them
 */
SCODE BcacheRemoveFromBufferState( PTR pBuf, UINT16 Bits)
{
    PCACHED_BLOCK pBlk = (PCACHED_BLOCK)pBuf;

    pBlk->State &= ~Bits; /* BUGBUG atomicity */
    return NO_ERROR;
}

#if _DEBUG
void BcacheStats(void)
{
    int i;
    BUFFER_LIST *pList;
    LISTNODE    *pBuf;
    CACHED_BLOCK *pBlk;
    int nTot, nDirty, nBusy;

    _tprintf(_T(" BcacheStats: %d/%d  hits/misses"), 
        TheBufferCache.Hits,
        TheBufferCache.Misses);

    nTot = nBusy = nDirty = 0;
    for (i = 0; i < BCACHE_NLISTS; i++) {
        pList = &TheBufferCache.Heads[ i ];
        pBuf = &pList->Buffers;

        _tprintf(_T("\n%d: "), i);

/*      Mutex_Lock(&pList->Lock);*/
        while (pBuf->pNext != &pList->Buffers) {
            pBuf = pBuf->pNext;
            pBlk = (CACHED_BLOCK *)pBuf;

            nTot++;
            if (pBlk->Tag) {
                _tprintf(_T(" %x %x %d %x %d\n"),
                        pBlk->Tag, pBlk->BlockNo,
                        pBlk->RefCnt, pBlk->State,
                        pBlk->BlkSize);
                if (pBlk->RefCnt)
                    nBusy++;
                if (pBlk->State & BLK_DIRTY)
                    nDirty++;
            }
        }
/*      Mutex_Unlock(&pList->Lock);*/

    }    

    _tprintf(_T("%d buffers %d dirty %d busy\n"),
            nTot, nDirty, nBusy);
}
#endif
