/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Microsoft MMLite June 1995 --jvh
 * Remote IFile using sockets.
 *
 * The server could reside eg. on an NT box and the file could be a disk image.
 * => Remote FS, remote devices.
 */

#undef _DEBUG /* silence! */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mmlite.h>
#include <tchar.h>
#include <drivers/drivers.h>
//#define DBGLVL 3
#include <mmmacros.h>

#define DBGIT(l,x) DBGME(l,printf x)
#define TDBGIT(l,x) DBGME(l,_tprintf x)

#define __HOSTFS_CLIENT 1
#include <drivers/hostfsp.h>
#include <drivers/serplex.h>
 
#if defined(_UNICODE)
PRIVATE void StrNCpyWcharToTchar(_TCHAR *pDst, const wchar_t *pSrc,UINT Count);
PRIVATE UINT StrNCpyTcharToWchar(wchar_t *pDst, const _TCHAR *pSrc,UINT Count);

PRIVATE SCODE ComposePathToWchar(wchar_t *NewPath, UINT MaxLen,
                                 const _TCHAR *Prefix, const _TCHAR *Name,
                                 UINT *pLengthInBytes);
#endif
PRIVATE SCODE ComposePathToTchar(_TCHAR *NewPath, UINT MaxLen,
                                 const _TCHAR *Prefix, const _TCHAR *Name,
                                 UINT *pLengthInBytes);

/* Used for debugging the protocol: use a serial.c
 * that directly polls the serial line.  Run something
 * else in the background and this generates occasional
 * losses of (random) characters.  
 * Other tests seem less effective.
 */
#define INTERNAL_POLLING_TEST 0

#if 0
void *MALLOC(UINT s, char *w)
{ void *p = malloc(s);printf("m%x (%d) %s\n",p,s,w);return p;}
#define FREE(p) {printf("f%x\n",p);free(p);}
#else
#define MALLOC(_s_,_w_) mymalloc(_s_)
#define FREE(_p_) free(_p_)
#endif

PRIVATE void * __cdecl mymalloc(size_t size)
{
    PIHEAP pIHeap = CurrentHeap();

    return pIHeap->v->Alloc(pIHeap, HEAP_DEBUG_NAME("hostfs"), size, 0);
}

/*------------------ Memory mgmt, not on stack -------------------------*/
#if 1
/*static*/ struct packet *freepkt = NULL;

/*static*/ struct packet *
getafix(void)
{
    struct packet *p;
    do {
        p = freepkt;
    } while ((p != NULL) &&
             !AtomicCmpAndSwap((PUINT)&freepkt,
                               (ADDRESS)p,
                               0));
    if (p == NULL) {
        p = MALLOC(MAX_PACKET_SIZE,"getafix");
    }
    return p;
}

/*static*/ void
putafix(struct packet *p)
{
    if (!AtomicCmpAndSwap((PUINT)&freepkt,
                          0,
                          (ADDRESS)p))
        FREE(p);
}

/*static*/ void
nomorefix(void)
{
    PTR p = freepkt;
    if (AtomicCmpAndSwap((PUINT)&freepkt,
                         (ADDRESS)p,
                         0))
        FREE(p);
}

#else
#define getafix() ((struct packet *)malloc(sizeof(struct packet)))
#define putafix(p) free(p)
#endif

/*------------------ Low level  network rpc -------------------------*/
/* 
 * Send a packet and receive into the same buffer.
 * The send buffer is never guaranteed to be valid on return.
 * Access to the connection is being serialized here.
 */
PRIVATE SCODE packet_rpc(PCONNECTION Conn, struct packet *p,  PACKETTYPE expect_type)
{
    SCODE sc;

    if (Conn == NULL)
        return E_INVALID_PARAMETER;

    Mutex_Lock(&Conn->Mutex);

    p->seqno = (UINT8)Conn->SequenceNumber++;

    if (Conn->Link == NULL) { /* local sim case */
        UINT length = HostFile_Server((void *) Conn, p, MAX_PACKET_SIZE);
        if (length == 0)
            sc = E_FAIL;
        else
            sc = S_OK;
    } else {
        DBGIT(0,("hostfs waiting for seqno: %x\n", Conn->SequenceNumber));
        sc = LinkLevelRpc(Conn,p,MAX_PACKET_SIZE);
    }
 
    if (!FAILED(sc))
        sc = packet_check(p, expect_type);

    Mutex_Unlock(&Conn->Mutex);
    return sc;
}

/* Post-process RPC.  Returns TRUE if the call should be retried */
#define packet_rpc_check(_connection_,_scode_) \
                        ((_scode_) == E_RESTART)

/*-------------------- CONNECTION mgmt --------------------------*/

/* Build a connection over an IFILE Link.
 * Link can be null, for debug (polling serial line)
 */
PRIVATE SCODE MakeConnection(PIFILE Link, PCONNECTION *pConn)
{
    PCONNECTION Conn;

    DBGIT(1,("MakeConnection(x%x, x%x)\n", (ADDRESS)Link,(ADDRESS)pConn));

    Conn = (PCONNECTION) MALLOC(sizeof(*Conn),"MakeConnection");
    if (!Conn)
        return E_NOT_ENOUGH_MEMORY;

    memset(Conn, 0, sizeof(*Conn));
    Conn->RefCnt = 1;
    Mutex_Init(&Conn->Mutex);
    Conn->Link = Link;
    if (Link)
        Link->v->AddRef(Link);
    Conn->SequenceNumber = 0;

    *pConn = Conn;
    return NO_ERROR;
}

PRIVATE UINT MCT ConnectionAddRef(PCONNECTION Conn)
{
    INT Refs;

    DBGIT(0,("ConnectionADDREF "));
    Refs = AtomicInc(&Conn->RefCnt);
    return Refs;
}

PRIVATE UINT MCT ConnectionRelease(PCONNECTION Conn)
{
    INT Refs;

    Refs = AtomicDec(&Conn->RefCnt);
    DBGIT(0,("ConnectionRELEASE%d ",Refs));
    if (Refs == 0) {
        if (Conn->Link)
            Conn->Link->v->Release(Conn->Link);
        FREE((PTR) Conn);
        nomorefix();          /* free cached buffer */
    }
    return Refs;
}

/*-------------------- IFile stuff --------------------------*/
extern const struct IFileVtbl NetFileVtbl; /* forward */
extern const struct INameSpaceVtbl NetFileNSVtbl; /* ditto */

typedef struct _NETFILENS {
    const struct INameSpaceVtbl *v;
    UINT RefCnt;
    PCONNECTION Conn;
    _TCHAR Prefix[MAX_PATH];    /* relative path */
} *PNETFILENS;

typedef struct _NETFILE {
    const struct IFileVtbl *v;
    UINT RefCnt;
    PCONNECTION Conn;
    UINT32 Handle;
} *PNETFILE;

PRIVATE SCODE AllocNetFile(PCONNECTION Conn, PNETFILE *pFile)
{
    PNETFILE File;

    File = (PNETFILE) MALLOC(sizeof(struct _NETFILE),"AllocNetFile");
    if (!File) {
        *pFile = NULL;
        return E_NOT_ENOUGH_MEMORY;
    }

    File->v = &NetFileVtbl;
    File->RefCnt = 1;
    File->Handle = 0;
    if (Conn)
        ConnectionAddRef(Conn);
    File->Conn = Conn;

    *pFile = File;
    return NO_ERROR;
}

PRIVATE SCODE AllocNetFileNS(PCONNECTION Conn, PNETFILENS *pNs)
{
    PNETFILENS Ns;

    Ns = (PNETFILENS) MALLOC(sizeof(struct _NETFILENS),"AllocNetFileNS");
    if (!Ns) {
        *pNs = NULL;
        return E_NOT_ENOUGH_MEMORY;
    }

    Ns->v = &NetFileNSVtbl;
    Ns->RefCnt = 1;
    if (Conn)
        ConnectionAddRef(Conn);
    Ns->Conn = Conn;
    _tcscpy(Ns->Prefix, _T(""));

    *pNs = Ns;
    return NO_ERROR;
}

PRIVATE SCODE OpenNetFile(PCONNECTION Conn, const _TCHAR *Prefix,
                          const _TCHAR *name, PNETFILE *File_out, UINT Flags)
{
    SCODE sc;
    PNETFILE File;
    UINT len;

    struct packet *p;
    struct open_req *req;
    struct open_rep *rep;

    DBGIT(1,("OpenNetFile(%x \"%s\" \"%s\" %x %x\n",
             (ADDRESS) Conn, Prefix, name, (ADDRESS)File_out, Flags));

    if ((p = getafix()) == NULL) return E_NOT_ENOUGH_MEMORY;

    /* Remove undefined flags */
    Flags &= (NAME_SPACE_READ|NAME_SPACE_WRITE|NAME_SPACE_CREATE|NAME_SPACE_FAILIFEXIST);

    sc = AllocNetFile(Conn,&File);
    if (FAILED(sc)) {
        goto Error;
    }

    /* Got connection.  Now get a handle */
    req = &p->u.open_req;
    rep = &p->u.open_rep;

 reinit_and_resend:
    packet_init(p, PACKET_TYPE_OPEN_REQ);
    req->flags = Flags;

#if defined(_UNICODE)
    sc = ComposePathToWchar((wchar_t*)req->path, MAX_PATH, Prefix, name, &len);
#else
    {
        STRING_ESCAPE *As = (STRING_ESCAPE *)req->path;
        As->Escape = UNICODE_NOT_A_CHARACTER;
        sc = ComposePathToTchar((_TCHAR*)As->String, MAX_PATH, Prefix, name, &len);
        len += offsetof(STRING_ESCAPE,String);
    }
#endif
    if (FAILED(sc))
        goto Error;
    len += offsetof(struct open_req,path);
    packet_prepare(p, len);

    sc = packet_rpc(Conn, p, PACKET_TYPE_OPEN_REP);
    if (packet_rpc_check(Conn, sc))
        goto reinit_and_resend;
    if (FAILED(sc))
        goto Error;

    File->Handle = rep->handle;
 Done:
    *File_out = File;
    putafix(p);
    return sc;
 Error:
    DBGIT(3,("OpenNetFile: failed \"%s\" sc=x%x\n",name,sc));
    if (File)
        File->v->Release((PIFILE)File);
    File = NULL;
    goto Done;
}

/* Similar to OpenNetFile */
PRIVATE SCODE NetFileMkDir(PNETFILENS Ns)
{
    SCODE sc;
    PCONNECTION Conn = Ns->Conn;
    struct packet *p;
    struct open_req *req;
    UINT Flags, len;

    if ((p = getafix()) == NULL) return E_NOT_ENOUGH_MEMORY;

    Flags = NAME_SPACE_READ|NAME_SPACE_WRITE|
        NAME_SPACE_CREATE|NAME_SPACE_FAILIFEXIST|NAME_SPACE_EXTENSION;

    /* Got connection.  Now get a handle */
    req = &p->u.open_req;
 reinit_and_resend:
    packet_init(p, PACKET_TYPE_OPEN_REQ);

#if defined(_UNICODE)
    len = StrNCpyTcharToWchar((wchar_t*)req->path, Ns->Prefix, MAX_PATH);
#else
    {
        STRING_ESCAPE *As = (STRING_ESCAPE *)req->path;
        len = strlen(Ns->Prefix) + 1;
        if (len > MAX_PATH) len = MAX_PATH;
        As->Escape = UNICODE_NOT_A_CHARACTER;
        memcpy((PTR)As->String, Ns->Prefix, len);
        As->String[MAX_PATH-1] = 0;
    }
    
#endif
    req->flags = Flags;
    len += offsetof(struct open_req,path);
    packet_prepare(p, len);

    sc = packet_rpc(Conn, p, PACKET_TYPE_OPEN_REP);
    if (packet_rpc_check(Conn, sc))
        goto reinit_and_resend;

    putafix(p);
    return sc;
}

PRIVATE SCODE NetFileDestroy(PNETFILE File)
{
    SCODE sc;
    PCONNECTION Conn = File->Conn;
    struct packet *p;
    struct close_req *req;

    /* Close file iff */
    if (File->Handle != 0)
    {
        if ((p = getafix()) == NULL) return E_NOT_ENOUGH_MEMORY;
        req = &p->u.close_req;
    reinit_and_resend:
        packet_init(p, PACKET_TYPE_CLOSE_REQ);
        req->handle = File->Handle;
        packet_prepare(p, sizeof(*req));

        sc = packet_rpc(Conn, p, PACKET_TYPE_CLOSE_REP);
        if (packet_rpc_check(Conn, sc))
            goto reinit_and_resend;
        if (FAILED(sc)) {
            DBGIT(3,("NetFileDestroy: packet_rpc failed sc=x%x (ignored)\n",
                sc));
        }
        putafix(p);
        /* ignore errors */
    }
    /* Destroy socket */
    ConnectionRelease(File->Conn);

    FREE((PTR) File);
    return NO_ERROR;
}

/* Delete a file/directory
 */
PRIVATE SCODE DeleteNetFile(PCONNECTION Conn, const _TCHAR *Prefix,
                            const _TCHAR *name)
{
    SCODE sc;
    UINT len;

    struct packet *p;
    struct delete_req *req;

    DBGIT(1,("DeleteNetFile(%x \"%s\" \"%s\"\n",
             (ADDRESS) Conn, Prefix, name));

    if ((p = getafix()) == NULL) return E_NOT_ENOUGH_MEMORY;

    req = &p->u.delete_req;

 reinit_and_resend:
    packet_init(p, PACKET_TYPE_DELETE_REQ);
    req->flags = 0;

#if defined(_UNICODE)
    sc = ComposePathToWchar((wchar_t*)req->path, MAX_PATH, Prefix, name, &len);
#else
    {
        STRING_ESCAPE *As = (STRING_ESCAPE *)req->path;
        As->Escape = UNICODE_NOT_A_CHARACTER;
        sc = ComposePathToTchar((_TCHAR*)As->String, MAX_PATH, Prefix, name, &len);
        len += offsetof(STRING_ESCAPE,String);
    }
#endif
    if (FAILED(sc))
        goto Error;
    len += offsetof(struct delete_req,path);
    packet_prepare(p, len);

    sc = packet_rpc(Conn, p, PACKET_TYPE_DELETE_REP);
    if (packet_rpc_check(Conn, sc))
        goto reinit_and_resend;
    if (FAILED(sc))
        goto Error;

 Done:
    putafix(p);
    return sc;
 Error:
    DBGIT(3,("DeleteNetFile: failed sc=x%x\n",sc));
    goto Done;
}


/*-------------------- IFile methods --------------------------*/

PRIVATE UINT MCT NetFileAddRef(PIFILE This)
{
    PNETFILE File = (PNETFILE) This;

    DBGIT(0,("NetFileADDREF "));
    return AtomicInc(&File->RefCnt);
}

PRIVATE UINT MCT NetFileNSAddRef(PINAMESPACE This)
{
    PNETFILENS Ns = (PNETFILENS) This;

    DBGIT(0,("NSaddref "));
    return AtomicInc(&Ns->RefCnt);
}

PRIVATE SCODE MCT NetFileQueryInterface(PIFILE This, REFIID pIid, void* *ppNew)
{
    if (UuidCmp(&IID_IFile, pIid) || UuidCmp(&IID_IUnknown, pIid)) {
        DBGIT(0,("NetFileQIok "));
        NetFileAddRef(This);
        *ppNew = This;
        return NO_ERROR;
    }
    DBGIT(0,("NetFileQIfail "));
    *ppNew = NULL;
    return E_NO_INTERFACE;
}

PRIVATE SCODE MCT NetFileNSQueryInterface(PINAMESPACE This, REFIID pIid, void* *ppNew)
{
    if (UuidCmp(&IID_INameSpace, pIid) || UuidCmp(&IID_IUnknown, pIid)) {
        DBGIT(0,("NSqiok "));
        NetFileNSAddRef((PINAMESPACE) This);
        *ppNew = This;
        return NO_ERROR;
    }
    DBGIT(0,("NSqifail "));
    *ppNew = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT MCT NetFileRelease(PIFILE pThis)
{
    PNETFILE File = (PNETFILE) pThis;
    INT Refs;

    Refs = AtomicDec(&File->RefCnt);
    DBGIT(0,("NetFileRELE%d ", Refs));
    if (Refs == 0)
        (void) NetFileDestroy(File);

    return Refs;
}

/*
 * Read from a file
 */
PRIVATE SCODE MCT
NetFileReadAt(PIFILE This, UINT64 Position, 
          BYTE *Buffer, UINT ByteCount, PUINT pSizeRead)
{
    SCODE sc;
    PNETFILE File = (PNETFILE) This;
    PCONNECTION Conn = File->Conn;
    UINT nread, ActualCount;
    struct packet *p;
    struct read_req *req;
    struct read_rep *rep;

    if (ByteCount > MAX_READ_SIZE)
        ActualCount = MAX_READ_SIZE;
    else
        ActualCount = ByteCount;

    if ((p = getafix()) == NULL) return E_NOT_ENOUGH_MEMORY;
    req = &p->u.read_req;
 reinit_and_resend:
    packet_init(p, PACKET_TYPE_READ_REQ);
    req->handle = File->Handle;
    req->position = Position;
    req->count = ActualCount;
    packet_prepare(p, sizeof(*req));

    sc = packet_rpc(Conn, p, PACKET_TYPE_READ_REP);
    if (packet_rpc_check(Conn, sc))
        goto reinit_and_resend;
    if (FAILED(sc)) {
        DBGIT(0,("NetFileReadAt failed sc=%x p.sc=%x len=%x\n",
                sc, p->sc, p->packet_length));
        goto Done;
    }

    rep = &p->u.read_rep;
    nread = rep->nread > ActualCount ? ActualCount : rep->nread;

    DBGIT(0,("NetFileReadAt(T=x%x P=%llx C=x%x (asked %x) PR=x%x nRead=%x sc=%x\n",
            (ADDRESS) This, Position, ActualCount, ByteCount, (ADDRESS) pSizeRead, nread, sc));

    memcpy(Buffer, rep->data, nread);
    *pSizeRead = nread;
 Done:
    putafix(p);
    return sc;
}

/*
 * Write to a file
 */
PRIVATE SCODE MCT
NetFileWriteAt(PIFILE This, UINT64 Position, 
               const BYTE *Buffer, UINT ByteCount, PUINT pSizeWritten)
{
    SCODE sc;
    PNETFILE File = (PNETFILE) This;
    PCONNECTION Conn = File->Conn;

    struct packet *p;
    struct write_req *req;
    struct write_rep *rep;

    DBGIT(0,("NetFileWriteAt(T=x%x P=%llx C=x%x PW=x%x\n",
            (ADDRESS) This, Position, ByteCount, (ADDRESS) pSizeWritten));
    if (ByteCount > MAX_READ_SIZE)
        ByteCount = MAX_READ_SIZE;

    if ((p = getafix()) == NULL) return E_NOT_ENOUGH_MEMORY;
    req = &p->u.write_req;
    rep = &p->u.write_rep;
 reinit_and_resend:
    packet_init(p, PACKET_TYPE_WRITE_REQ);
    req->handle = File->Handle;
    req->position = Position;
    req->count = ByteCount;
    memcpy(req->data, Buffer, ByteCount);
    packet_prepare(p, sizeof(*req) - (MAX_READ_SIZE - ByteCount));

    sc = packet_rpc(Conn, p, PACKET_TYPE_WRITE_REP);

    DBGIT(0,("NetFileWriteAt_END(T=x%x P=%lx sc=x%x N=x%x\n",
            (ADDRESS) This, Position, sc, rep->nwritten));
    if (packet_rpc_check(Conn, sc))
        goto reinit_and_resend;
    if (!FAILED(sc))
        *pSizeWritten = rep->nwritten;
    putafix(p);
    return sc;
}

/* Get/Set a file's size
 */
PRIVATE SCODE MCT NetFileGetSize(PIFILE This, PUINT64 pSize)
{
    SCODE sc;
    PNETFILE File = (PNETFILE) This;
    PCONNECTION Conn = File->Conn;

    struct packet *p;
    struct getsize_req *req;
    struct getsize_rep *rep;

    if ((p = getafix()) == NULL) return E_NOT_ENOUGH_MEMORY;
    req = &p->u.getsize_req;
    rep = &p->u.getsize_rep;
 reinit_and_resend:
    packet_init(p, PACKET_TYPE_GETSIZE_REQ);
    req->handle = File->Handle;
    packet_prepare(p, sizeof(*req));

    sc = packet_rpc(Conn, p, PACKET_TYPE_GETSIZE_REP);
    if (packet_rpc_check(Conn, sc))
        goto reinit_and_resend;
    if (!FAILED(sc)) {
        DBGIT(1,("NetFileGetSize -> %llx\n", rep->size));
        *pSize = rep->size;
    }
    putafix(p);
    return sc;
}

PRIVATE SCODE MCT NetFileSetSize(PIFILE This, UINT64 Size)
{
    SCODE sc;
    PNETFILE File = (PNETFILE) This;
    PCONNECTION Conn = File->Conn;
    struct packet *p;
    struct setsize_req *req;

    DBGIT(0,("NetFileSetSize "));

    if ((p = getafix()) == NULL) return E_NOT_ENOUGH_MEMORY;
    req = &p->u.setsize_req;
 reinit_and_resend:
    packet_init(p, PACKET_TYPE_SETSIZE_REQ);
    req->handle = File->Handle;
    req->size = Size;
    packet_prepare(p, sizeof(*req));

    sc = packet_rpc(Conn, p, PACKET_TYPE_SETSIZE_REP);
    if (packet_rpc_check(Conn, sc))
        goto reinit_and_resend;

    putafix(p);
    return sc;
}

const struct IFileVtbl NetFileVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    NetFileQueryInterface, 
    NetFileAddRef, 
    NetFileRelease, 
    NetFileReadAt, 
    NetFileWriteAt, 
    NetFileSetSize, 
    NetFileGetSize
};


PRIVATE UINT MCT NetFileNSRelease(PINAMESPACE This)
{
    PNETFILENS Ns = (PNETFILENS) This;
    INT Refs;

    Refs = AtomicDec(&Ns->RefCnt);
    DBGIT(0,("NSrele%x ", Refs));
    if (Refs == 0) {
        if (Ns->Conn)
            ConnectionRelease(Ns->Conn);
        FREE((PTR) Ns);
    }
    return Refs;
}

PRIVATE SCODE ParseAndLookupNS(const _TCHAR *pName, PNETFILENS *ppNs)
{
    PCONNECTION Conn;
    PNETFILENS Ns;
    PIFILE Link = NULL;
    SCODE sc = S_OK;

#if INTERNAL_POLLING_TEST
    sc = MakeConnection(Link,&Conn);
    if (FAILED(sc)) {
        DBGIT(3,("ParseAndLookupNS: MakeConnection failed x%x\n",sc));
        return sc;
    }
#else
    if (pName) {
        /* Lookup the link driver */
        sc = BindToObject(CurrentNameSpace(), pName,
                          NAME_SPACE_READ|NAME_SPACE_WRITE,
                          &IID_IFile, (void *) &Link);
        if (FAILED(sc)) {
            DBGIT(3,("ParseAndLookupNS: Bind %s failed sc=%x\n",pName,sc));
            return E_INVALID_PARAMETER;
        }
    }

    /* Open adds ref to Link */
    sc = MakeConnection(Link,&Conn);
    if (FAILED(sc)) {
        DBGIT(3,("ParseAndLookupNS: MakeConnection failed x%x\n",sc));
        return sc;
    }
    if (Link)
        Link->v->Release(Link);
#endif

    Ns = NULL;
    sc = AllocNetFileNS(Conn,&Ns);
    if (FAILED(sc)) {
        DBGIT(3,("AllocNetFileNS failed %x\n",sc));
    }

    /* Do this regardless, since AllocNetFileNS() took one ref */
    ConnectionRelease(Conn);
    *ppNs = Ns;
    return sc;
}

/* Give "inetaddr:inetport/remotename", get IFile.  :inetport is optional */
PRIVATE SCODE MCT NetFileNSBind(PINAMESPACE This, const _TCHAR *Name, NAME_SPACE_FLAGS Flags, PIUNKNOWN *Object)
{
    PNETFILENS Ns = (PNETFILENS) This, NewNs;
    UINT len;
    SCODE sc = NO_ERROR;
    PNETFILE File;

    DBGIT(1,("NetFileNSBind(x%x \"%s\" x%x)\n", (ADDRESS) This, Name, Flags));
    /* Check flags. Parse name. Open remote file. */
    if (Object == NULL)
        return E_INVALID_PARAMETER;
    *Object = NULL;
    if (Name == NULL)
        return E_INVALID_PARAMETER;

    if (Ns->Conn == NULL) {
        /* This is the initial Name Space. Get a connected one and use it. */
        sc = ParseAndLookupNS(Name,&Ns);
        if (FAILED(sc)) {
            *Object = NULL;
            return sc;
        }
    } else {
        Ns->v->AddRef((PINAMESPACE) Ns);
    }
    
    len = _tcslen(Name);
    if (Name[0] == _EOS) {      /* DIR start directory or this directory */
        *Object = (PIUNKNOWN) Ns;       /* May be NULL */
        DBGIT(1,("NetFileNSBind Ns: Name=\"%s\" Ns=x%x sc=x%x\n",
                 Name, (ADDRESS) Ns, sc));
    } else if (IS_PATH_SEPARATOR(Name[len - 1])) { /* DIR */
        /* 
         * BUGBUG: mind "/" root directory and "//" UNC quotes etc.
         * BUT, they are not seen in practice.
         */
        /* Make sure the remembered prefix is big enough.  Make dynamic! */
        if (_tcslen(Ns->Prefix) + len + 2 > MAX_PATH) {
            sc = E_INVALID_PARAMETER;
            goto ErrorA;
        }

        sc = AllocNetFileNS(Ns->Conn,&NewNs);
        if (FAILED(sc)) {
            goto ErrorA;
        }

        sc = ComposePathToTchar(NewNs->Prefix, MAX_PATH, Ns->Prefix, Name, &len);
        Ns->v->Release((PINAMESPACE) Ns);
        if (FAILED(sc))
            goto ErrorB;
        DBGIT(1,("NetFileNSBind subdir: NewNs=%x Prefix=\"%s\"\n",
                 (ADDRESS) NewNs, NewNs->Prefix));

        /* Check if it really exists on remote or create it. */
        if (Flags & NAME_SPACE_CREATE) {
            /* Create it! */
            sc = NetFileMkDir(NewNs);
            if (FAILED(sc)) {
                DBGIT(3,("NetFileNSBind: create subdir \"%s\" failed %x\n",
                         NewNs->Prefix, sc));
                goto ErrorB;
            }
        } else {
            /* Does it exist then? */
            _TCHAR tmpbuf[2];
            UINT tmp;
            sc = NewNs->v->FindNext((PINAMESPACE) NewNs, _T(""), _T(""),
                                    tmpbuf, sizeof(tmpbuf), &tmp);
            if (FAILED(sc)) {
                DBGIT(2,("NetFileNSBind: nonexisting subdir \"%s\" rejected %x\n",
                         NewNs->Prefix, sc));
                goto ErrorB;
            }
        }

        *Object = (PIUNKNOWN) NewNs;
        return NO_ERROR;
    ErrorA:
        Ns->v->Release((PINAMESPACE) Ns);
        *Object = NULL;
        return sc;
    ErrorB:
        NewNs->v->Release((PINAMESPACE) NewNs);
        *Object = NULL;
        return sc;
    } else {                    /* FILE */
        sc = OpenNetFile(Ns->Conn, Ns->Prefix, Name, &File, Flags);
        DBGIT(1,("NetFileNSBind File: Name=%s Ns=x%x file=x%x fla=%x sc=x%x\n",
                 Name, (ADDRESS) Ns, (ADDRESS) File, Flags, sc));
        *Object = (PIUNKNOWN) File;     /* may be NULL */
        Ns->v->Release((PINAMESPACE) Ns);               /* not returned */
    }
    return sc;
}

PRIVATE SCODE MCT NetFileNSFindNext(PINAMESPACE This, const _TCHAR *Prefix,
         const _TCHAR *Previous,_TCHAR *Buffer, UINT BufSize, PUINT StrLen_out)
{
    SCODE sc;
    PNETFILENS Ns = (PNETFILENS) This;
    PCONNECTION Conn = Ns->Conn;
    UINT len;

    struct packet *p;
    struct findnext_req *req;
    struct findnext_rep *rep;

    if (Previous == NULL)
      Previous = _T("");

    TDBGIT(1,(_T("NetFileNSFindNext(Ns=x%x Ns->Prefix=\"%s\" Prefix=\"%s\" Previ=\"%s\" Bs=%d) Conn=x%x\n"),
             (ADDRESS) Ns, Ns->Prefix, Prefix ? Prefix : _T("NULL"), Previous,
             BufSize, (ADDRESS)Conn));

    /* at least the _EOS must fit in Buffer */
    if (BufSize < sizeof(_TCHAR) || Buffer == NULL) {
        TDBGIT(3,(_T("bad NetFileNSFindNext, args=x%x,\"%s\",\"%s\",x%x,%d,x%x\n"),
                 (ADDRESS) This, Prefix ? Prefix : _T("NULL"), Previous,
                 (ADDRESS)Buffer, BufSize, (ADDRESS)StrLen_out));
        return E_INVALID_PARAMETER;
    }

    if ((p = getafix()) == NULL) return E_NOT_ENOUGH_MEMORY;
    req = &p->u.findnext_req;
    rep = &p->u.findnext_rep;
 reinit_and_resend:
    packet_init(p, PACKET_TYPE_FINDNEXT_REQ);
#if defined(_UNICODE)
    {
        wchar_t *wstring;
        wstring = (wchar_t*)req->prefix;
        sc = ComposePathToWchar(wstring, MAX_PATH, Ns->Prefix, Prefix, &len);
        req->offset_of_previous = (UINT16) len;
        wstring = (wchar_t*)(req->prefix+len);
        len += StrNCpyTcharToWchar(wstring, Previous, MAX_PATH);
    }
#else
    {
        STRING_ESCAPE *As = (STRING_ESCAPE *)req->prefix;
        /* Server xlates */
        As->Escape = UNICODE_NOT_A_CHARACTER;
        sc = ComposePathToTchar((_TCHAR*)As->String, MAX_PATH, Ns->Prefix, Prefix, &len);
        if (!FAILED(sc)) {
            UINT len1;
            req->offset_of_previous = (UINT16) len;
            sc = ComposePathToTchar((_TCHAR*)(As->String+len), MAX_PATH,
                                    _T(""), Previous,
                                    &len1);
            len += offsetof(STRING_ESCAPE,String);
            len += len1;
        }
    }
    
#endif
    if (FAILED(sc))
        goto Done;

    len += offsetof(struct findnext_req,prefix);
    packet_prepare(p, len);

    sc = packet_rpc(Conn, p, PACKET_TYPE_FINDNEXT_REP);
    if (packet_rpc_check(Conn, sc))
        goto reinit_and_resend;
    if (FAILED(sc))
        goto Done;

    /* BUGBUG needs something like strnlen(s,MAX_PATH) */
#if defined(_UNICODE)
    len = wcslen((wchar_t*) rep->name);
    if (len < BufSize/sizeof(_TCHAR)) {
        StrNCpyWcharToTchar(Buffer, (wchar_t*) rep->name, len);
    }
    else {
        sc = S_BUFFER_TOO_SMALL;
        len = 0;
    }
#else
    {
        STRING_ESCAPE *As = (STRING_ESCAPE *)rep->name;
        if (As->Escape != UNICODE_NOT_A_CHARACTER) {
            /* Bad server !! */
            sc = E_INVALID_METHOD; /* BUGBUG need a more precise one */
            len = 0;
        }
        else {
            len = strlen(As->String);
            if (len < BufSize/sizeof(_TCHAR)) {
                strncpy(Buffer,As->String,len);
            }
            else {
                sc = S_BUFFER_TOO_SMALL;
                len = 0;
            }
        }
    }
#endif

    Buffer[len] = _EOS;
    if (StrLen_out)
      *StrLen_out = len;
 Done:
    putafix(p);
    return sc;
}

PRIVATE SCODE MCT NetFileNSUnregister(PINAMESPACE This, const _TCHAR * Name)
{
    PNETFILENS Ns = (PNETFILENS) This;
    UINT len;
    SCODE sc;

    DBGIT(1,("NetFileNSUnregister(x%x \"%s\"+\"%s\")\n",
             (ADDRESS) This, Ns->Prefix, Name));

    /* Make sure it fits
     */
    len = _tcslen(Name);
    if (len + _tcslen(Ns->Prefix) + 2 > MAX_PATH)
        return E_INVALID_PARAMETER;

    sc = DeleteNetFile(Ns->Conn, Ns->Prefix, Name);

    return sc;
}


/* We DO NOT need separate implementations for each unimplemented method 
 * because WE GOT RID of __stdcall. Argh. Ha!
 */
PRIVATE SCODE MCT Unimp()
{
    return E_NOT_IMPLEMENTED;
}

const struct INameSpaceVtbl NetFileNSVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NetFileNSQueryInterface,
    NetFileNSAddRef,
    NetFileNSRelease,
    (_INameSpace_Register_METHOD_TYPE) Unimp,
    NetFileNSUnregister,
    NetFileNSBind,
    NetFileNSFindNext,
    (_INameSpace_GetCapabilities_METHOD_TYPE) Unimp
};

/*--------------------- Constructor ----------------------*/

PINAMESPACE InitHostFileSystem(const _TCHAR *LinkName)
{
    PNETFILENS Ns = NULL;

#if INTERNAL_POLLING_TEST
    (void) AllocNetFileNS(NULL,&Ns);
    Ns->v->Bind((PINAMESPACE)Ns,"release/",0,(PIUNKNOWN*)&Ns);
#else
    (void) ParseAndLookupNS(LinkName,&Ns);
#endif
    
    return (PINAMESPACE) Ns;
}

/* ------------ Loose ends from netmisc.c ----------------- */
PRIVATE SCODE ComposePathToTchar(_TCHAR *NewPath, UINT MaxLen,
                                 const _TCHAR *Prefix, const _TCHAR *Name,
                                 UINT *pLengthInBytes)
{
    UINT len1, len2;

    len1 = _tcslen(Prefix);
    len2 = Name ? _tcslen(Name) : 0;
    if ((len1+len2+2) > MaxLen) {
        /* So what should we do? */
        TDBGIT(3,(_T("ComposePath \"%s\" \"%s\" too long %d+%d+2 > %d\n"),
                  Prefix, Name ? Name : _T("NULL"), len1, len2, MaxLen));
        *pLengthInBytes = 0;
        return E_INVALID_PARAMETER; /* reject it altogether */
    }

    /* Combine Prefix and Name  "Prefix/Name" or "Prefix" or "Name" or "" */
    NewPath[0] = _EOS;
    if (len1) {
        _tcscpy(NewPath, Prefix);
        /* add NT separator */
        if (len2) {
            NewPath[len1++] = _T('\\');
            /* beware: no EOS */
        }
    }            
    if (len2)
        _tcscpy(NewPath+len1, Name);

    /* Remove trailing slash */
    len1 += len2;
    if (len1 > 0 && IS_PATH_SEPARATOR(NewPath[len1 - 1]))
        NewPath[--len1] = _EOS; /* NB: account for shorter strlen */

    TDBGIT(1,(_T("ComposePath \"%s\" \"%s\" -> \"%s\" sizeof(*p)=%d => %d\n"),
              Prefix, Name ? Name : _T("NULL"), NewPath,
              (UINT) sizeof(struct packet), len1+1));
    *pLengthInBytes = len1 + 1; /* _EOS counts! */
    return NO_ERROR;
}

#if defined(_UNICODE)
/* Count should be non-zero and indicates the size of the destination buffer.
 * Always null-terminates the destination string.
 */
PRIVATE void StrNCpyWcharToTchar(_TCHAR *pDst, const wchar_t *pSrc, UINT Count)
{
    _TCHAR c;

    while (--Count != 0) {
        c = (_TCHAR) *pSrc++;
        if (c == _EOS)
            break;
        *pDst++ = c;
    }
    *pDst = _EOS;
}

/* Count should be non-zero and indicates the size of the destination buffer.
 * Always null-terminates the destination string.
 * Returns the number of BYTES copied.
 */
PRIVATE UINT StrNCpyTcharToWchar(wchar_t *pDst, const _TCHAR *pSrc, UINT Count)
{
    wchar_t c, *pDstSave = pDst;

    while (--Count != 0) {
        c = (wchar_t) *pSrc++;
        if (c == _EOS)
            break;
        *pDst++ = c;
    }
    *pDst++ = _EOS;
    return (pDst-pDstSave)*sizeof(wchar_t);
}

#if _UNICODE
PRIVATE SCODE ComposePathToWchar(wchar_t *NewPath, UINT MaxLen,
                                 const _TCHAR *Prefix, const _TCHAR *Name,
                                 UINT *pLengthInBytes)
{
    return ComposePathToTchar(NewPath, MaxLen, Prefix, Name, pLengthInBytes);
}
#else
PRIVATE SCODE ComposePathToWchar(wchar_t *NewPath, UINT MaxLen,
                                 const _TCHAR *Prefix, const _TCHAR *Name,
                                 UINT *pLengthInBytes)
{
    _TCHAR Buffer[MAX_PATH * sizeof(_TCHAR)];
    SCODE sc;

    sc = ComposePathToTchar(Buffer, MAX_PATH, Prefix, Name, pLengthInBytes);
    if (FAILED(sc))
        return sc;

    StrNCpyTcharToWchar(NewPath, Buffer, MaxLen);
    *pLengthInBytes *= 2;
    return sc;
}
#endif

#endif

/* ** 
*/
SCODE __cdecl HostfsConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS RegisterBase,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        _TCHAR *TransportName)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Name);
    UnusedParameter(RegisterBase);
    UnusedParameter(Interrupt);
    
    DBGME(0,_tprintf(_T("HostfsConstructor(%x %s %x %x %x)\n"),
                     (ADDRESS) pThis, Name, RegisterBase, Interrupt,
                     (ADDRESS) ppUnk));

    pUnk = (PIUNKNOWN) InitHostFileSystem(TransportName);

    DBGME(0,_tprintf(_T("HostfsConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("HostfsConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl HostfsDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) HostfsConstructor
};

static struct _DRIVERCOB HostfsDriverCobObject = { &HostfsDriverCobVtable, 1 };

PIUNKNOWN HostfsCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &HostfsDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &HostfsDriverCobObject;
}
