/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    fat16.c
 *
 * Purpose:
 *
 *    Constructor for a FAT12/16 filesystem driver
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */


#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <drivers/drivers.h>
//#define _DEBUG 1
#define DBGLVL 2
#include <diagnostics.h>
#include <assert.h>
#include <winsock.h>

/* Forward decls
 */
extern const struct IFileVtbl ResilientTcpVtbl;

/* Our object
 */
typedef struct _RESILIENTTCP {
    const struct IFileVtbl *v;
    UINT RefCnt;
    SOCKET Socket;
    _TCHAR *HostName;
    struct sockaddr_in HostAddress;
    UINT PortNumber;
} *PRESILIENTTCP;

/* Creator/destructor
 */
PRIVATE PRESILIENTTCP ResilientTcpCreate(const _TCHAR *HostName, UINT PortNumber)
{
    PRESILIENTTCP File;
    UINT HostNameLength = _tcslen(HostName) + 1;

    DBGME(0,_tprintf(_T("RestcpCreate %s %d\n"), HostName, PortNumber));

    File = (PRESILIENTTCP) malloc(sizeof(struct _RESILIENTTCP)+HostNameLength);
    if (!File) {
        return NULL;
    }

    memset(File,0,sizeof *File);
    File->v = &ResilientTcpVtbl;
    File->RefCnt = 1;
    File->HostName = (_TCHAR*)(File+1);
    _tcsncpy(File->HostName,HostName,HostNameLength);
    File->PortNumber = PortNumber;
    File->Socket = INVALID_SOCKET;

    return File;
}

PRIVATE SCODE ResilientTcpDestroy(PRESILIENTTCP File)
{
    DBGME(0,_tprintf(_T("RestcpDestroy %x\n"), (ADDRESS)File));

    if (File->Socket != INVALID_SOCKET)
        closesocket(File->Socket);
    free(File);

    return S_OK;
}

/* Make a resilient TCP connection to the given host.
 * This only ever returns when the connection is established.
 * It will forever retry, caveat user.
 */
PRIVATE void ResilientTcpConnect(PRESILIENTTCP File)
{
    struct hostent *hp;
    struct sockaddr_in *to;
    int r;
    TIME RetryTime;

    for (;;) {

        /* A simple check?
         */
        if (File->Socket != INVALID_SOCKET)
            return;

        /* Do we have an IP address yet
         */
        to = &File->HostAddress;
        if (to->sin_addr.s_addr == 0) {

            /* Nope, find it
             */
            DBGME(0,_tprintf(_T("RestcpConnect addr\n")));

            to->sin_family = AF_INET;
            to->sin_len = 16;
            to->sin_port = htons((UINT16)File->PortNumber);

            /* Check first for an IP address as argument
             */
            to->sin_addr.s_addr = _tinet_addr(File->HostName);
            if (to->sin_addr.s_addr == (u_int)-1) {

                DBGME(0,_tprintf(_T("RestcpConnect name\n")));

                /* Is it a name then
                 */
                to->sin_addr.s_addr = 0;
                hp = gethostbyname((char*)File->HostName);
                if (!hp) {
                    DBGME(3,_tprintf(_T("ResTcp: Cant resolve %s yet\n"),
                                     File->HostName));
                    goto Failure;
                }

                to->sin_family = (UINT8) hp->h_addrtype;
                memcpy((PTR)&to->sin_addr, hp->h_addr, hp->h_length);
            }

            DBGME(0,_tprintf(_T("ResTcp addr for %s is %s\n"),
                             File->HostName, inet_ntoa(to->sin_addr)));
        }

        /* Create a socket for the connection
         */
        File->Socket = socket(to->sin_family, SOCK_STREAM, 0);
        if (File->Socket == INVALID_SOCKET) {
            DBGME(3,_tprintf(_T("ResTcp: Cant socket(%d,%d,0) yet\n"),
                             to->sin_family, SOCK_STREAM));
            goto Failure;
        }
        DBGME(0,_tprintf(_T("ResTcp socket @%x\n"), File->Socket));

        /* Set a recv timeout
         */
        r = 3000;
        r = setsockopt(File->Socket,SOL_SOCKET,SO_RCVTIMEO,(PTR)&r,sizeof r);
        if (r != 0) {
            r = WSAGetLastError();
            DBGME(3,_tprintf(_T("ResTcp: Cant rcvtimeo, err=%d, ignoring\n"),
                             r));
        }

        /* Try to connect it
         */
        r = connect(File->Socket, (const struct sockaddr*)to, sizeof(*to));
        if (r == 0)
            return;

        r = WSAGetLastError();
        DBGME(3,_tprintf(_T("ResTcp: Cant connect to %s yet, err=%d\n"),
                         File->HostName, r));

        /* If it fails for some reasons I get EINVAL afterwards.
         */
        closesocket(File->Socket);
        File->Socket = INVALID_SOCKET;

    Failure:
        Int32ToInt64(RetryTime,TIME_RELATIVE(TIME_SECONDS(5)));
        SleepUntil(RetryTime);
    }
}

#if _DEBUG
PRIVATE void DumpBytes(const UINT8 *p, UINT c)
{
    _tprintf(_T("[%d]"), c);
    while (c-- > 0)
        _tprintf(_T(" %x"), *p++);
    _tprintf(_T("\n"));
}
#else
#define DumpBytes(p,c)
#endif

/* IUnknown interface
 */
PRIVATE UINT MCT ResilientTcpAddRef(PIFILE This)
{
    PRESILIENTTCP File = (PRESILIENTTCP) This;

    DBGME(0,_tprintf(_T("ResilientTcpADDREF %d\n"),File->RefCnt+1));
    return AtomicInc(&File->RefCnt);
}

PRIVATE UINT MCT ResilientTcpRelease(PIFILE pThis)
{
    PRESILIENTTCP File = (PRESILIENTTCP) pThis;
    INT Refs;

    Refs = AtomicDec(&File->RefCnt);
    DBGME(0,_tprintf(_T("ResilientTcpRELE %d\n"), Refs));
    if (Refs == 0)
        (void) ResilientTcpDestroy(File);

    return Refs;
}

PRIVATE SCODE MCT ResilientTcpQueryInterface(PIFILE This, REFIID pIid, void* *ppNew)
{
    return GenericQueryInterface((PIUNKNOWN)This,pIid,ppNew,&IID_IFile);
}

/* IFile addtnl methods
 */

/*
 * Read from a file
 */
PRIVATE SCODE MCT
ResilientTcpReadAt(PIFILE This, UINT64 Position, 
          BYTE *Buffer, UINT ByteCount, PUINT pSizeRead)
{
    int Count;
    PRESILIENTTCP File = (PRESILIENTTCP) This;

    UnusedParameter(Position);

 Retry:
    DBGME(0,_tprintf(_T("ResilientTcpReadAt(T=x%x B=x%x C=x%x PR=x%x\n"),
            (ADDRESS) This, (ADDRESS)Buffer, ByteCount, (ADDRESS) pSizeRead));
 
    ResilientTcpConnect(File);

    Count = recv(File->Socket,(char*)Buffer,ByteCount,0);
    if (Count >= 0) {
        if (pSizeRead)
            *pSizeRead = (UINT) Count;
        DBGME(0,_tprintf(_T("ResilientTcpReadAt(T=%x) ->%d "),
                         (ADDRESS) This, Count));
        DBGME(0,DumpBytes(Buffer,Count));
        return S_OK;
    }
    
    Count = WSAGetLastError();
    DBGME(3,_tprintf(_T("ResilientTcpReadAt(T=%x) ->-1 er=%d\n"),
                         (ADDRESS) This, Count));

    if (Count == WSAETIMEDOUT)
        return E_RESTART;

    closesocket(File->Socket);
    File->Socket = INVALID_SOCKET;
    goto Retry;
}

/*
 * Write to a file
 */
PRIVATE SCODE MCT
ResilientTcpWriteAt(PIFILE This, UINT64 Position, 
               const BYTE *Buffer, UINT ByteCount, PUINT pSizeWritten)
{
    int Count;
    PRESILIENTTCP File = (PRESILIENTTCP) This;

    UnusedParameter(Position);

 Retry:
    DBGME(0,_tprintf(_T("ResilientTcpWriteAt(T=x%x B=x%x C=x%x PR=x%x\n"),
            (ADDRESS)This, (ADDRESS)Buffer, ByteCount, (ADDRESS)pSizeWritten));
 
    ResilientTcpConnect(File);

    Count = send(File->Socket,(char*)Buffer,ByteCount,0);
    if (Count >= 0) {
        if (pSizeWritten)
            *pSizeWritten = (UINT) Count;
        DBGME(0,_tprintf(_T("ResilientTcpWriteAt(T=%x) ->"),
                         (ADDRESS) This));
        DBGME(0,DumpBytes(Buffer,Count));
        return S_OK;
    }

    DBGME(3,_tprintf(_T("ResilientTcpWriteAt(T=%x) ->%d er=%d\n"),
                         (ADDRESS) This, Count, WSAGetLastError()));

    closesocket(File->Socket);
    File->Socket = INVALID_SOCKET;
    goto Retry;
}

/* Get/Set a file's size
 */
PRIVATE SCODE MCT ResilientTcpGetSize(PIFILE This, PUINT64 pSize)
{
    UnusedParameter(This);
    DBGME(0,_tprintf(_T("ResilientTcpGetSize @%x\n"), (ADDRESS)pSize));

    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE MCT ResilientTcpSetSize(PIFILE This, UINT64 Size)
{
    UnusedParameter(This);
    DBGME(0,_tprintf(_T("ResilientTcpSetSize %lld\n"), Size));

    return E_NOT_IMPLEMENTED;
}

const struct IFileVtbl ResilientTcpVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    ResilientTcpQueryInterface, 
    ResilientTcpAddRef, 
    ResilientTcpRelease, 
    ResilientTcpReadAt, 
    ResilientTcpWriteAt, 
    ResilientTcpSetSize, 
    ResilientTcpGetSize
};

/* Exported constructor (COB)
 */

PIFILE CreateResilientTcpSocket(const _TCHAR *HostName, UINT PortNumber)
{
    return (PIFILE) ResilientTcpCreate(HostName,PortNumber);
}


/* ** 
*/
SCODE __cdecl RestcpConstructor(PIDEVICECONSTRUCTOR pThis,
        const _TCHAR * Name,
        ADDRESS PortNumber,
        UINT Interrupt,
        /*out*/ PIUNKNOWN* ppUnk,
        _TCHAR *HostName,
        ...)
{
    PIUNKNOWN pUnk;

    UnusedParameter(Name);
    UnusedParameter(Interrupt);
    
    DBGME(0,_tprintf(_T("RestcpConstructor(%x %s %x %x %x %s)\n"),
                     (ADDRESS) pThis, Name, PortNumber, Interrupt,
                     (ADDRESS) ppUnk, HostName));

    pUnk = (PIUNKNOWN) CreateResilientTcpSocket(HostName,PortNumber);

    DBGME(0,_tprintf(_T("RestcpConstructor returns %x\n"), (ADDRESS) pUnk));

    if (pUnk) {
        *ppUnk = pUnk;
        return S_OK;
    } else {
        DBGME(3,_tprintf(_T("RestcpConstructor failed\n")));
        return E_FAIL;
    }
}

static struct IDeviceConstructorVtbl RestcpDriverCobVtable = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NULL,
    NULL,
    NULL,
    (_IDeviceConstructor_Constructor_METHOD_TYPE) RestcpConstructor
};

static struct _DRIVERCOB RestcpDriverCobObject = { &RestcpDriverCobVtable, 1 };

PIUNKNOWN RestcpCobMain(void)
{
    SCODE sc;

    /* Import default functions */
    sc = BaseInheritClass((ADDRESS *) &RestcpDriverCobVtable,
                          _T("CDeviceConstructor"),
                          &IID_IDeviceConstructor);
    if (FAILED(sc))
        return NULL;

    return (PIUNKNOWN) &RestcpDriverCobObject;
}
