/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>

/* RAM/ROM based, memory-mapped filesystem
 *
 * The structure is trivial: the first 'page' contains a header and the
 * one and only directory in the filesystem.  The header has a signature,
 * the total size, the count and the offset of the files themselves (in
 * case the directory grows to more than one 'page'), the alignment, and
 * the offset of the string table.
 * Each directory entry contains offset (from header) and size of the file.
 * The MS bit in the file size is set if it is read only.
 * The string table follows.
 * The Nth directory entry corresponds to the Nth string table entry.
 * Filenames are UNICODE strings
 * Each file is 'page'-aligned and concatenated in the rest of the FS.
 *
 * There are three views into the filesystem.  First, it is a namespace.
 * A trivial one, with no extensions or recursion.  Second, it creates
 * IFILE objects.  Third, it creates IMEMORYOBJECTs to map
 * executables directly, without any copies.
 */

#define ROMFSACCESS  (ADDR_PROT_READ | ADDR_PROT_EXECUTE)
#if 1
#define DBGOUT(_x_,_y_,_z_)
#else
#include <stdio.h>
#define DBGOUT(_x_,_y_,_z_) _z_
#endif

typedef struct {
    UINT32      Signature;
#define ROM_SIGNATURE 0xabaddeed
    UINT32      TotalSize;
    UINT32      BlockSize; /* must be 1page for VM mapping */
    UINT32      FilesOffset;
    UINT32      StringsOffset;
    UINT32      NumberOfFiles;
} ROM_HEADER;

typedef struct {
    UINT32      FileOffset;
    UINT32      FileSize;

} ROM_DIRENTRY;

#define PointerToStrings(_h_) ((const _TCHAR *)((UINT8*)(_h_) + ((_h_)->StringsOffset)))
#define PointerToDirents(_h_) ((ROM_DIRENTRY*)((_h_) + 1))
#define SizeOfFile(File) (*(File)->pFileSize & 0x7fffffff)
#define IsReadOnly(File) (*(File)->pFileSize & 0x80000000)

/* Namespace Object
 */
typedef struct {
    const struct INameSpaceVtbl *v;
    UINT    RefCnt;
    PIHEAP  pIHeap;
#ifdef __WANTS_MEMORY_OBJECT
    PIMEMORYOBJECT pMemoryObject;
#endif

    ROM_HEADER *pHeader;

} ROM_NAME;

/* File Object
 *
 * Note that the size is a pointer to the ROM_DIRENTRY.FileSize,
 * so that when the size changes it is permanently updated.
 */
typedef struct {
    const struct IFileVtbl *v;
    UINT    RefCnt;
    PIHEAP  pIHeap;
    ROM_NAME *pDir;

    UINT8   *pData;
    UINT32  FileMaxSize;
    UINT32* pFileSize;
    UINT    FileIndex;

} ROM_FILE;


/* Auxillary search function
 * Looks for a filename in string space, starting from pSTRINGS.
 * Returns the filename index (starting at 0) or -1 if not found.
 * Also returns a pointer to the next filename following pNAME.
 * Search is case-insensitive.
 */
static _TCHAR ToLower(_TCHAR c)
{
    if (c >= 'A' && c <= 'Z')
        return (_TCHAR)(c - 'A' + 'a');
    return c;
}

static INT StrSearch( ROM_HEADER *pHdr, const _TCHAR * pName,
                      const _TCHAR * *pStrings)
{
    const _TCHAR * cp;
    INT i, m;

    cp = *pStrings;

    /* cp points to the begginnig of an array of strings (not string pointers)
     * ie  string1\0string2\0string3\0\0
     */
    m = (INT) pHdr->NumberOfFiles;

    for (i = 0; i < m; i++) {
        _TCHAR c;
        const _TCHAR * p;

        /* Stop on a mismatch, or _EOS (either string)
         */
        c = ToLower(*cp++);
        for (p = pName; *p != _EOS; p++) {
            if (c != ToLower(*p) || c == _EOS)
                break;
            c = ToLower(*cp++);
        }

        /* If we stopped with a match it must be an _EOS
         */
        if (c == ToLower(*p) /* == _EOS */) {
            *pStrings = cp;
            return i;
        }

        /* keep going till _EOS on this filename
         */
        for (; c != _EOS; cp++)
            c = *cp;
    }
    return -1;
}

/* Auxillary copy function (ffirst/fnext)
 * Copies the string and sets the length.
 */
SCODE RomNameToStr(const _TCHAR * Name, _TCHAR * Buf, UINT BufSize,
                   PUINT pStrLen)
{
    UINT NameLen = _tcslen(Name);

    if ((NameLen + 1) * sizeof(_TCHAR) > BufSize) {
        Buf[0] = _EOS;
        *pStrLen = 0;
        return S_BUFFER_TOO_SMALL;
    }

    _tcscpy(Buf,Name);
    if (pStrLen)
      *pStrLen = NameLen;
    return (NameLen > 0) ? S_OK : S_NO_MORE_ENTRIES;
}


/* IFile methods
 * Hopefully, users will use IMemoryObject instead.
 */
SCODE MCT RomFileQueryInterface(PIFILE pThis, REFIID pIid, void* *ppNew)
{
    ROM_FILE * File = (ROM_FILE *) pThis;
#ifdef __WANTS_MEMORY_OBJECT
    ADDR_FLAGS Prot = ROMFSACCESS;

    if (IsReadOnly(File))
        Prot |= ADDR_PROT_COPY_ON_WRITE;
    else
        Prot |= ADDR_PROT_WRITE;

    /* Mapped file ?
     */
    if (UuidCmp(&IID_IMemoryObject, pIid)) {
        PIMEMORYOBJECT pObj;
        ADDR_SIZE Offset, FileSize;
        ROM_DIRENTRY *pDiren;

        pObj = File->pDir->pMemoryObject;
        pDiren = PointerToDirents(File->pDir->pHeader);
        Offset = pDiren[File->FileIndex].FileOffset;
        FileSize = (ADDR_SIZE) SizeOfFile(File);

        return pObj->v->RestrictedAccess(pObj,Offset,FileSize,
                                                 Prot,(PIMEMORYOBJECT*) ppNew);

    }
#endif

    if (UuidCmp(&IID_IFile, pIid) || UuidCmp(&IID_IUnknown, pIid)) {
        File->RefCnt++;
        *ppNew = pThis;
        return NO_ERROR;
    }

    *ppNew = NULL;
    return E_NO_INTERFACE;
}

UINT MCT RomFileAddRef(PIFILE pThis)
{
    ROM_FILE * File = (ROM_FILE *) pThis;

    return AtomicInc(&File->RefCnt);
}

UINT MCT RomFileRelease(PIFILE pThis)
{
    ROM_FILE * File = (ROM_FILE *) pThis;
    UINT Refs;

    Refs = AtomicDec(&File->RefCnt);
    if (Refs == 0) {
        if (File->pDir)
            File->pDir->v->Release((PINAMESPACE)File->pDir);
        File->pIHeap->v->Free( File->pIHeap, 0, File);
    }

    return Refs;
}

/*
 * Read from a file
 */
SCODE MCT
RomFileReadAt(PIFILE pThis, UINT64 Position, 
                                 BYTE *Buffer, UINT ByteCount, PUINT pSizeRead)
{
    UINT BytesRead, CurOffset;
    ROM_FILE * File = (ROM_FILE *)pThis;
    UINT FileSize = (UINT) SizeOfFile(File);

    CurOffset = (UINT) Int64ToInt32(Position);

    DBGOUT(DBGFS, 0, printf("RomFileReadAt(%x, %x, %x, %x)\n", 
           (ADDRESS) File, CurOffset, (ADDRESS) Buffer, ByteCount));

    if (CurOffset >= FileSize) {
        if (CurOffset != FileSize)
            return E_INVALID_PARAMETER;

        BytesRead = 0;  /* Read at EOF is not an error */
        goto Exit;
    }

    if (CurOffset + ByteCount > FileSize)
        ByteCount = FileSize - CurOffset;

    memcpy(Buffer, File->pData + CurOffset, ByteCount);

    BytesRead = ByteCount;
Exit:
    if (pSizeRead != NULL)
        *pSizeRead = BytesRead;
    return (CurOffset == FileSize) ? S_FALSE : S_OK;
}


/*
 * Write to a file
 */
SCODE MCT
RomFileWriteAt(PIFILE pThis, UINT64 Position, 
               const BYTE *Buffer, UINT ByteCount, PUINT pSizeWritten)
{
    ROM_FILE * File = (ROM_FILE *)pThis;
    UINT32 FileOffset;

    if (IsReadOnly(File))
        return E_ACCESS_DENIED;

    FileOffset = (UINT32) Int64ToInt32(Position);

    if ((FileOffset + ByteCount) > File->FileMaxSize)  {
        if (FileOffset >= File->FileMaxSize)
            return E_OUT_OF_STRUCTURES;
        ByteCount = File->FileMaxSize - FileOffset;
    }

    memcpy(File->pData + FileOffset, (PTR) Buffer, ByteCount);

    if ((FileOffset + ByteCount) > *File->pFileSize)
        *File->pFileSize = FileOffset + ByteCount;

    if (pSizeWritten)
        *pSizeWritten = ByteCount;

    return NO_ERROR;
}

/* Get/Set a file's size
 */
SCODE MCT RomFileGetSize(PIFILE pThis, PUINT64 pSize)
{
    ROM_FILE * File = (ROM_FILE *)pThis;
    Int64FromHighAndLow(*pSize, 0, SizeOfFile(File));
    return NO_ERROR;
}

SCODE MCT RomFileSetSize(PIFILE pThis, UINT64 Size)
{
    ROM_FILE * File = (ROM_FILE *)pThis;
    UINT32 size = (UINT32) Int64ToInt32(Size);

    if (IsReadOnly(File))
        return E_ACCESS_DENIED;

    if (size > File->FileMaxSize)
        return E_ACCESS_DENIED;
    *File->pFileSize = size;
    return NO_ERROR;
}

static const struct IFileVtbl RomFileVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    RomFileQueryInterface,
    RomFileAddRef,
    RomFileRelease,
    RomFileReadAt,
    RomFileWriteAt,
    RomFileSetSize,
    RomFileGetSize
};


/*
 * Constructor for IFILE objects
 * (IMEMORYOBJECTs come stright from QueryInterface)
 */
PRIVATE SCODE RomFileNew( ROM_NAME *pRoot, const _TCHAR * pName,
                         ROM_FILE **ppNew)
{
    const _TCHAR * cp;
    ROM_FILE * File;
    PIHEAP pIHeap;
    ROM_DIRENTRY *pDir;
    ROM_HEADER *pHdr;
    INT i;

    DBGOUT(DBGFS, 0, printf("RomFileNew %x %s\n", (ADDRESS) pRoot, pName));

    /*
     * Allocate a File descriptor
     */
    pIHeap = pRoot->pIHeap;
    File = (ROM_FILE *) pIHeap->v->Alloc(pIHeap, HEAP_ZERO_MEMORY, sizeof(ROM_FILE), 0);
    if (File == NULL)
        return E_OUT_OF_STRUCTURES;

    File->v = &RomFileVtbl;
    File->RefCnt = 1;
    File->pIHeap = pIHeap;

    /*
     * Search, from right after the header
     */
    pHdr = pRoot->pHeader;
    pDir = PointerToDirents(pHdr);
    cp = PointerToStrings(pHdr);

    i = StrSearch( pHdr, pName, &cp );

    if (i >= 0) {
        File->pData = ((UINT8*)pHdr) + pDir[i].FileOffset;
        File->pFileSize = &pDir[i].FileSize;
        File->FileMaxSize = (SizeOfFile(File) + pHdr->BlockSize - 1)  &
                            ~(pHdr->BlockSize - 1);
        File->FileIndex = (UINT) i;
        File->pDir = pRoot;
        pRoot->v->AddRef((PINAMESPACE)pRoot);
        *ppNew = File;
        return S_OK;
    }

    /* No such numbah
     */
    File->v->Release((PIFILE)File);
    return E_FILE_NOT_FOUND;
}

/*
 * INameSpace methods
 */
SCODE MCT RomNameQueryInterface(PINAMESPACE pThis, REFIID pIid, void* *ppUnk)
{
    ROM_NAME * pIName = (ROM_NAME *)pThis;

    if (UuidCmp(&IID_INameSpace, pIid) || UuidCmp(&IID_IUnknown, pIid)) {
        pIName->RefCnt++;
        *ppUnk = (PIUNKNOWN) pIName;
        return NO_ERROR;
    }
    *ppUnk = NULL;
    return E_NO_INTERFACE;
}

/*
 * Add a reference to a Name object
 */
UINT MCT RomNameAddRef(PINAMESPACE pThis)
{
    ROM_NAME * pIName = (ROM_NAME *)pThis;

    DBGOUT(DBGFS, 0, printf("RomAddRef %x(%d)\n", (ADDRESS) pIName, pIName->RefCnt));

    return AtomicInc(&pIName->RefCnt);
}


/*
 * Release a reference to a Name object
 * When last ref is gone, destructor.
 */
UINT MCT RomNameRelease(PINAMESPACE pThis)
{
    ROM_NAME * pIName = (ROM_NAME *)pThis;
    UINT Refs;

    DBGOUT(DBGFS, 0, printf("RomRelease %x(%d)\n", (ADDRESS) pIName, pIName->RefCnt));

    Refs = AtomicDec(&pIName->RefCnt);
    if (Refs == 0) {
#ifdef __WANTS_MEMORY_OBJECT
        pIName->pMemoryObject->v->Release(pIName->pMemoryObject);
#endif
        pIName->pIHeap->v->Free(pIName->pIHeap, 0, pIName);
    }
    return Refs;
}

/* Register/Unregister not supported
 */
SCODE MCT RomNameRegister(PINAMESPACE pThis, const _TCHAR * pName, 
                          PIUNKNOWN pObj, UINT Flags, PINAMESPACE pSNS)
{
    UnusedParameter(pThis);
    UnusedParameter(pName);
    UnusedParameter(pObj);
    UnusedParameter(Flags);
    UnusedParameter(pSNS);


    DBGOUT(DBGFS, 2, printf("RomNameRegister?!? %s\n", pName));
    return E_ACCESS_DENIED;
}

SCODE MCT RomNameUnregister(PINAMESPACE pThis, const _TCHAR * pName)
{
    UnusedParameter(pThis);
    UnusedParameter(pName);

    DBGOUT(DBGFS, 2, printf("RomNameUnregister?!? %s\n", pName));
    return E_ACCESS_DENIED;
}


/* Bind to a named object
 */
SCODE MCT RomNameBind(PINAMESPACE pThis, const _TCHAR * pName, 
                      NAME_SPACE_FLAGS Flags, PIUNKNOWN *ppObject)
{
    ROM_FILE * File;
    SCODE sc;

    UnusedParameter(Flags);

    DBGOUT(DBGFS, 0, printf("RomNameBind %s Flags %08x\n", pName, Flags));

    /* Special case
     */
    if (*pName == _EOS) {
        pThis->v->AddRef(pThis);
        *ppObject = (PIUNKNOWN)pThis;
        return NO_ERROR;
    }

    /* Lookup the pathname
     */
    sc = RomFileNew( (ROM_NAME*) pThis, pName, &File );

    if (FAILED(sc))
        return sc;

    *ppObject = (PIUNKNOWN) File;
    return S_OK;

}

/*
 * Find entry that follows pPrevious
 */
SCODE MCT RomNameFindNext(PINAMESPACE pThis, const _TCHAR * pPrefix, 
       const _TCHAR * pPrevious, _TCHAR * pBuffer, UINT BufSize, PUINT pStrLen)
{
    const _TCHAR * cp;
    ROM_DIRENTRY *pDir;
    ROM_HEADER *pHdr;
    INT i, last;
    SCODE sc;

    ROM_NAME *pRoot = (ROM_NAME *)pThis;

    UnusedParameter(pPrefix);

    DBGOUT(DBGFS, 0, printf("RomNameFindNext %s %s %x\n",
           pPrefix ? pPrefix : "NULL", pPrevious ? pPrevious : "NULL", BufSize));

    /* Check previous still around
     */
    pHdr = pRoot->pHeader;
    pDir = PointerToDirents(pHdr);
    cp = PointerToStrings(pHdr);

    if (pHdr->NumberOfFiles == 0)
        return E_FILE_NOT_FOUND;

    /* If asking for first we got it
     * If pPrevious is NULL or an empty string, return the first entity!
     */
    if (pPrevious == NULL || *pPrevious == _EOS) {
        i = -1;
        if (pHdr->NumberOfFiles == 0)
            return E_FILE_NOT_FOUND;
        last = (INT) pHdr->NumberOfFiles - 1;
        goto CopyOut;
    }

    i = StrSearch( pHdr, pPrevious, &cp );

    /* Check for bogus or last file
     */
    last = (INT) pHdr->NumberOfFiles - 1;

    if ((i < 0) || (i == last)) {
        pBuffer[0] = _EOS;
        if (pStrLen)
          *pStrLen = 0;
        if (i < 0)
            return E_FILE_NOT_FOUND;
        else
            return S_NO_MORE_ENTRIES;
    }

    /* Copy out, drop any padding
     */
 CopyOut:
    sc = RomNameToStr(cp,pBuffer,BufSize,pStrLen);
    if (i == (last - 1))
        sc = S_NO_MORE_ENTRIES;
    DBGOUT(DBGFS, 0, printf("Exit RomNameFindNext %s\n", pBuffer));

    return sc;
}

/*
 * Find first entry in directory
 */
SCODE MCT RomNameFindFirst(PINAMESPACE pThis, const _TCHAR * pPrefix,
                           _TCHAR * pBuffer, UINT BufSize, PUINT pStrLen)
{
#if 1
    /* same as code below ... but untested */
    if (pPrefix == NULL)
        return E_INVALID_PARAMETER;
    return RomNameFindNext(pThis, pPrefix, NULL, pBuffer, BufSize, pStrLen);
#else
    ROM_NAME * pIName = (ROM_NAME *)pThis;
    ROM_HEADER *pHdr;
    SCODE sc;

    DBGOUT(DBGFS, 0, printf("RomNameFindFirst %s %x\n", pPrefix, BufSize));

    if (pPrefix == NULL)
        return E_INVALID_PARAMETER;

    pHdr = pIName->pHeader;

    /* Sanity
     */
    if (pHdr->NumberOfFiles == 0)
        goto ErrorExit;

    /* Copy out, drop any padding
     */
    sc = RomNameToStr(PointerToStrings(pHdr),pBuffer,BufSize,pStrLen);

    DBGOUT(DBGFS, 0, printf("Exit RomNameFindFirst %s (sc %x\n",pBuffer,sc));

    return sc;

ErrorExit:
    DBGOUT(DBGFS, 1, printf("Error exit RomNameFindFirst\n"));
    return E_FILE_NOT_FOUND;
#endif
}

SCODE MCT RomNameGetCapabilities(PINAMESPACE pThis, const _TCHAR * pName, 
                                 NAME_SPACE_FLAGS *pFlags)
{
    ROM_FILE * File;
    SCODE sc;

    DBGOUT(DBGNAME, 2, printf("RomNameGetCapabilities %s\n",pName));

    /* Lookup the pathname
     */
    sc = RomFileNew( (ROM_NAME*)pThis, pName, &File );

    if (FAILED(sc))
        return sc;

    /* Return result
     */
    if (pFlags != NULL)
        *pFlags = NAME_SPACE_READ | NAME_SPACE_SHARE_READ;

    ((PIFILE) File)->v->Release((PIFILE)File);

    return S_OK;
}



static const struct INameSpaceVtbl RomNameVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    RomNameQueryInterface, 
    RomNameAddRef, 
    RomNameRelease, 
    RomNameRegister, 
    RomNameUnregister, 
    RomNameBind, 
    RomNameFindNext, 
    RomNameGetCapabilities
};


/*
 * Constructor for this Name Space
 */
PINAMESPACE RomFsNew( PTR BaseOfMemory, UINT Size)
{
    PIHEAP pIHeap;
    ROM_HEADER *pHdr;
    ROM_NAME *pfn;
#ifdef __WANTS_MEMORY_OBJECT
    PIMEMORYOBJECT pObj, pRomObj;
    PIADDRESSSPACE pAs;
    SCODE sc;
#endif

    /* Sanity checks
     */
    pHdr = (ROM_HEADER *)BaseOfMemory;
    if ((pHdr->Signature != ROM_SIGNATURE) || (pHdr->TotalSize > Size)) {
        DBGOUT(DBGFS,2,printf("Bad ROMFS x%x x%x ? x%x\n",
                               pHdr->Signature, pHdr->TotalSize, Size));
        return NULL;
    }

    pIHeap = CurrentHeap();
#ifdef __WANTS_MEMORY_OBJECT
    pAs = GetKernelAddressSpace();

    sc = pAs->v->QueryInterface(pAs,&IID_IMemoryObject,&pObj);
    if (FAILED(sc))
        return NULL;

    sc = pObj->v->RestrictedAccess(pObj,(ADDRESS) BaseOfMemory,
                      (ADDR_SIZE) Size,ROMFSACCESS | ADDR_PROT_WRITE,&pRomObj);

    pObj->v->Release(pObj);
    if (FAILED(sc))
        return NULL;
#endif

    pfn = (ROM_NAME *) pIHeap->v->Alloc(pIHeap, 0, sizeof(ROM_NAME), 0);
    if (pfn == NULL) {
#ifdef __WANTS_MEMORY_OBJECT
        pRomObj->v->Release(pRomObj);
#endif
        return NULL;
    }

    pfn->v = &RomNameVtbl;
    pfn->RefCnt = 1;
    pfn->pIHeap = pIHeap;
#ifdef __WANTS_MEMORY_OBJECT
    pfn->pMemoryObject = pRomObj;
#endif

    pfn->pHeader = pHdr;

    DBGOUT(DBGFS,1,printf("And returned %08x\n",(ADDRESS) pfn));

    return (PINAMESPACE) pfn;
}
