/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>

/* Serial Flash/ROM based filesystem
 *
 * The structure is trivial: the first 'page' contains a header and the
 * one and only directory in the filesystem.  The header has a signature,
 * the total size, the count and the offset of the files themselves (in
 * case the directory grows to more than one 'page'), the alignment, and
 * the offset of the string table.
 * Each directory entry contains offset (from header) and size of the file.
 * The MS bit in the file size is set if it is read only.
 * The string table follows.
 * The Nth directory entry corresponds to the Nth string table entry.
 * Filenames are UNICODE strings
 * Each file is 'page'-aligned and concatenated in the rest of the FS.
 *
 * There are two views into the filesystem.  First, it is a namespace.
 * A trivial one, with no extensions or recursion.  Second, it creates
 * IFILE objects.
 */

#include <diagnostics.h>

typedef struct {
    UINT32      Signature;
#define ROM_SIGNATURE 0xabaddeed
    UINT32      TotalSize;
    UINT32      BlockSize;
    UINT32      FilesOffset;
    UINT32      StringsOffset;
    UINT32      NumberOfFiles;
} ROM_HEADER;

typedef struct {
    UINT32      FileOffset;
    UINT32      FileSize;

} ROM_DIRENTRY;

#define PointerToStrings(_h_) ((const _TCHAR *)((UINT8*)(_h_) + ((_h_)->StringsOffset)))
#define PointerToDirents(_h_) ((ROM_DIRENTRY*)((_h_) + 1))
#define SizeOfFile(File) ((File)->FileSize & 0x7fffffff)
#define IsReadOnly(File) ((File)->FileSize & 0x80000000)

/* Namespace Object
 * Only one of these, for the root. But maybe someday...
 */
typedef struct {
    const struct INameSpaceVtbl *v;
    UINT    RefCnt;
    PIFILE  Storage;

    /* Cached
     * BUGBUG With some more code we could spare the extra RAM burden but.
     */
    ROM_HEADER *pHeader;

} ROM_NAME;

/* File Object
 */
typedef struct {
    const struct IFileVtbl *v;
    UINT    RefCnt;
    ROM_NAME *pDir;

    UINT32  DataOffset;
    UINT    FileSize;

} ROM_FILE;


/* Auxillary search function
 * Looks for a filename in string space, starting from pSTRINGS.
 * Returns the filename index (starting at 0) or -1 if not found.
 * Also returns a pointer to the next filename following pNAME.
 * Search is case-insensitive.
 */
PRIVATE _TCHAR ToLower(_TCHAR c)
{
    if (c >= 'A' && c <= 'Z')
        return (_TCHAR)(c - 'A' + 'a');
    return c;
}

PRIVATE INT StrSearch( ROM_HEADER *pHdr, const _TCHAR * pName,
                      const _TCHAR * *pStrings)
{
    const _TCHAR * cp;
    INT i, m;

    cp = *pStrings;

    /* cp points to the beginning of an array of strings (not string pointers)
     * ie  string1\0string2\0string3\0\0
     */
    m = (INT) pHdr->NumberOfFiles;
    for (i = 0; i < m; i++) {
        _TCHAR c;
        const _TCHAR * p;

        /* Stop on a mismatch, or _EOS (either string)
         */
        c = ToLower(*cp++);
        for (p = pName; *p != _EOS; p++) {
            if (c != ToLower(*p) || c == _EOS)
                break;
            c = ToLower(*cp++);
        }

        /* If we stopped with a match it must be an _EOS
         */
        if (c == ToLower(*p) /* == _EOS */) {
            *pStrings = cp;
            return i;
        }

        /* keep going till _EOS on this filename
         */
        for (; c != _EOS; cp++)
            c = *cp;
    }
    return -1;
}

/* Auxillary copy function (ffirst/fnext)
 * Copies the string and sets the length.
 */
PRIVATE SCODE SromNameToStr(const _TCHAR * Name, _TCHAR * Buf, UINT BufSize,
                           PUINT pStrLen)
{
    UINT NameLen = _tcslen(Name);

    if ((NameLen + 1) * sizeof(_TCHAR) > BufSize) {
        Buf[0] = _EOS;
        *pStrLen = 0;
        return S_BUFFER_TOO_SMALL;
    }

    _tcscpy(Buf,Name);
    if (pStrLen)
      *pStrLen = NameLen;
    return (NameLen > 0) ? NO_ERROR : S_NO_MORE_ENTRIES;
}

/* Aux storage access function
 * Returns all or nothing, crosses blocks
 */
PRIVATE SCODE ReadFromStorage( PIFILE File,
                               UINT Offset,
                               UINT8 *Destination,
                               UINT nBytes)
{
    UINT64 Offset64;
    SCODE sc;
    UINT nRead;

    /* Loop till we get it all or fail
     */
    for (;;) {
        Uint32ToUint64(Offset64,Offset);
        sc = File->v->ReadAt(File,
                             Offset64,
                             Destination,
                             nBytes,
                             &nRead);
        /* Check for errors and eof
         */
        if ((FAILED(sc)) || (nRead == 0))
            return E_INVALID_PARAMETER;

        /* Are we done yet ?
         */
        if (nRead >= nBytes)
            break;

        /* Nope, keep reading
         */
        Destination += nRead;
        Offset += nRead;
        nBytes -= nRead;
    }
    return NO_ERROR;
}

/* IFile methods
 */
PRIVATE SCODE MCT 
SromFileQueryInterface(PIFILE pThis, REFIID pIid, void* *ppNew)
{
    ROM_FILE * File = (ROM_FILE *) pThis;

    if (UuidCmp(&IID_IFile, pIid) || UuidCmp(&IID_IUnknown, pIid)) {
        File->RefCnt++;
        *ppNew = pThis;
        return NO_ERROR;
    }

    *ppNew = NULL;
    return E_NO_INTERFACE;
}

PRIVATE UINT MCT SromFileAddRef(PIFILE pThis)
{
    ROM_FILE * File = (ROM_FILE *) pThis;

    return AtomicInc(&File->RefCnt);
}

PRIVATE UINT MCT SromFileRelease(PIFILE pThis)
{
    ROM_FILE * File = (ROM_FILE *) pThis;
    UINT Refs;

    Refs = AtomicDec(&File->RefCnt);
    if (Refs == 0) {
        PIHEAP pIHeap = CurrentHeap();
        if (File->pDir)
            File->pDir->v->Release((PINAMESPACE)File->pDir);
        pIHeap->v->Free( pIHeap, 0, File);
    }

    return Refs;
}

/*
 * Read from a file
 */
PRIVATE SCODE MCT
SromFileReadAt(PIFILE pThis, UINT64 Position, 
               BYTE *Buffer, UINT ByteCount, PUINT pSizeRead)
{
    ROM_FILE * File = (ROM_FILE *)pThis;
    UINT CurOffset, FileSize = (UINT) SizeOfFile(File);
    SCODE sc = NO_ERROR;

    CurOffset = (UINT) Int64ToInt32(Position);

    DBGME(0, _tprintf(_T("SromFileReadAt(%x, %x, %x, %x)\n"), 
           (ADDRESS) File, CurOffset, (ADDRESS) Buffer, ByteCount));

    if (CurOffset >= FileSize) {
        if (CurOffset != FileSize)
            return E_INVALID_PARAMETER;

        ByteCount = 0;  /* Read at EOF is not an error */
        sc = S_FALSE;
        goto Exit;
    }

    if (CurOffset + ByteCount > FileSize)
        ByteCount = FileSize - CurOffset;

    sc = ReadFromStorage(File->pDir->Storage,
                         File->DataOffset + CurOffset,
                         Buffer,
                         ByteCount);
    if (FAILED(sc))
        ByteCount = 0;
Exit:
    if (pSizeRead != NULL)
        *pSizeRead = ByteCount;
    return sc;
}


/*
 * Write to a file
 */
PRIVATE SCODE MCT
SromFileWriteAt(PIFILE pThis, UINT64 Position, 
               const BYTE *Buffer, UINT ByteCount, PUINT pSizeWritten)
{
    /* This filesystem is read-only */
    UnusedParameter(pThis);
    UnusedParameter(Position);
    UnusedParameter(Buffer);
    UnusedParameter(ByteCount);

    if (pSizeWritten)
        *pSizeWritten = 0;
    return E_ACCESS_DENIED;
}

/* Get/Set a file's size
 */
PRIVATE SCODE MCT SromFileGetSize(PIFILE pThis, PUINT64 pSize)
{
    ROM_FILE * File = (ROM_FILE *)pThis;
    Int64FromHighAndLow(*pSize, 0, SizeOfFile(File));
    return NO_ERROR;
}

PRIVATE SCODE MCT SromFileSetSize(PIFILE pThis, UINT64 Size)
{
    UnusedParameter(pThis);
    UnusedParameter(Size);

    return E_ACCESS_DENIED;
}

PRIVATE const struct IFileVtbl SromFileVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    SromFileQueryInterface,
    SromFileAddRef,
    SromFileRelease,
    SromFileReadAt,
    SromFileWriteAt,
    SromFileSetSize,
    SromFileGetSize
};


/*
 * Constructor for IFILE objects
 */
PRIVATE SCODE SromFileNew( ROM_NAME *pRoot, const _TCHAR * pName,
                         ROM_FILE **ppNewFile)
{
    const _TCHAR * cp;
    ROM_FILE * File;
    PIHEAP pIHeap;
    ROM_DIRENTRY *pDir;
    ROM_HEADER *pHdr;
    INT i;

    DBGME(0, _tprintf(_T("SromFileNew %x %s\n"), (ADDRESS)pRoot,pName));

    /*
     * Allocate a File descriptor
     */
    pIHeap = CurrentHeap();
    File = (ROM_FILE *) pIHeap->v->Alloc(pIHeap, HEAP_ZERO_MEMORY, sizeof(ROM_FILE), 0);
    if (File == NULL) {
        return (E_OUT_OF_STRUCTURES);
    }

    File->v = &SromFileVtbl;
    File->RefCnt = 1;

    /*
     * Search, from right after the header
     */
    pHdr = pRoot->pHeader;
    pDir = PointerToDirents(pHdr);
    cp = PointerToStrings(pHdr);

    i = StrSearch( pHdr, pName, &cp );

    if (i >= 0) {
        File->DataOffset = pDir[i].FileOffset;
        File->FileSize = pDir[i].FileSize;
        File->pDir = pRoot;
        pRoot->v->AddRef((PINAMESPACE)pRoot);
        *ppNewFile = File;
        return NO_ERROR;
    }

    /* No such numbah
     */
    pIHeap->v->Free(pIHeap,0,File);
    return (E_FILE_NOT_FOUND);
}

/*
 * INameSpace methods
 */
PRIVATE SCODE MCT SromNameQueryInterface(PINAMESPACE pThis, REFIID pIid, void* *ppUnk)
{
    ROM_NAME * pIName = (ROM_NAME *)pThis;

    if (UuidCmp(&IID_INameSpace, pIid) || UuidCmp(&IID_IUnknown, pIid)) {
        pIName->RefCnt++;
        *ppUnk = (PIUNKNOWN) pIName;
        return NO_ERROR;
    }
    *ppUnk = NULL;
    return E_NO_INTERFACE;
}

/*
 * Add a reference to a Name object
 */
PRIVATE UINT MCT SromNameAddRef(PINAMESPACE pThis)
{
    ROM_NAME * pIName = (ROM_NAME *)pThis;

    DBGME(0, _tprintf(_T("SromAddRef %x(%d)\n"), (ADDRESS) pIName, pIName->RefCnt));

    return AtomicInc(&pIName->RefCnt);
}


/*
 * Release a reference to a Name object
 * When last ref is gone, destructor.
 */
PRIVATE UINT MCT SromNameRelease(PINAMESPACE pThis)
{
    ROM_NAME * pIName = (ROM_NAME *)pThis;
    UINT Refs;

    DBGME(0, _tprintf(_T("SromRelease %x(%d)\n"), (ADDRESS) pIName, pIName->RefCnt));

    Refs = AtomicDec(&pIName->RefCnt);
    if (Refs == 0) {
        PIHEAP  pIHeap = CurrentHeap();
        pIName->Storage->v->Release(pIName->Storage);
        /* NB: One big alloc for ROM_NAME and root copy, so.. */
        pIHeap->v->Free(pIHeap, 0, pIName);
    }
    return Refs;
}

/* Register/Unregister not supported
 */
PRIVATE SCODE MCT SromNameRegister(PINAMESPACE pThis, const _TCHAR * pName, 
                          PIUNKNOWN pObj, UINT Flags, PINAMESPACE pSNS)
{
    UnusedParameter(pThis);
    UnusedParameter(pName);
    UnusedParameter(pObj);
    UnusedParameter(Flags);
    UnusedParameter(pSNS);


    DBGME(2, _tprintf(_T("SromNameRegister?!? %s\n"), pName));
    return E_ACCESS_DENIED;
}

PRIVATE SCODE MCT SromNameUnregister(PINAMESPACE pThis, const _TCHAR * pName)
{
    UnusedParameter(pThis);
    UnusedParameter(pName);

    DBGME(2, _tprintf(_T("SromNameUnregister?!? %s\n"), pName));
    return E_ACCESS_DENIED;
}


/* Bind to a named object
 */
PRIVATE SCODE MCT SromNameBind(PINAMESPACE pThis, const _TCHAR * pName, 
                      NAME_SPACE_FLAGS Flags, PIUNKNOWN *ppObject)
{
    UnusedParameter(Flags);

    DBGME(0, _tprintf(_T("SromNameBind %s Flags %08x\n"), pName, Flags));

    /* Special case
     */
    if (*pName == _EOS) {
        pThis->v->AddRef(pThis);
        *ppObject = (PIUNKNOWN)pThis;
        return NO_ERROR;
    }

    /* Lookup the pathname, return new file if found, error otherwise.
     */
    return SromFileNew( (ROM_NAME*) pThis, pName, (ROM_FILE **)ppObject);
}

/*
 * Find entry that follows pPrevious
 */
PRIVATE SCODE MCT 
SromNameFindNext(PINAMESPACE pThis, const _TCHAR * pPrefix, 
                const _TCHAR * pPrevious, _TCHAR * pBuffer, UINT BufSize,
                PUINT pStrLen)
{
    const _TCHAR * cp;
    ROM_DIRENTRY *pDir;
    ROM_HEADER *pHdr;
    INT i, last;
    SCODE sc;

    ROM_NAME *pRoot = (ROM_NAME *)pThis;

    UnusedParameter(pPrefix);

    DBGME(0, _tprintf(_T("SromNameFindNext %s %s %x\n"),
           pPrefix ? pPrefix : _T("NULL"), 
           pPrevious ? pPrevious : _T("NULL"), BufSize));

    /* Check previous still around
     */
    pHdr = pRoot->pHeader;
    pDir = PointerToDirents(pHdr);
    cp = PointerToStrings(pHdr);

    if (pHdr->NumberOfFiles == 0)
        return E_FILE_NOT_FOUND;

    /* If asking for first we got it
     * If pPrevious is NULL or an empty string, return the first entity!
     */
    if (*pPrevious == _EOS || pPrevious == NULL) {
        i = -1;
        if (pHdr->NumberOfFiles == 0)
            return E_FILE_NOT_FOUND;
        last = (INT) pHdr->NumberOfFiles - 1;
        goto CopyOut;
    }

    i = StrSearch( pHdr, pPrevious, &cp );

    /* Check for bogus or last file
     */
    last = (INT) pHdr->NumberOfFiles - 1;

    if ((i < 0) || (i == last)) {
        pBuffer[0] = _EOS;
        if (pStrLen)
          *pStrLen = 0;
        if (i < 0)
            return E_FILE_NOT_FOUND;
        else
            return S_NO_MORE_ENTRIES;
    }

    /* Copy out, drop any padding
     */
 CopyOut:
    sc = SromNameToStr(cp,pBuffer,BufSize,pStrLen);
    if (i == (last - 1))
        sc = S_NO_MORE_ENTRIES;
    DBGME(0, _tprintf(_T("Exit SromNameFindNext %s\n"), pBuffer));

    return sc;
}

PRIVATE SCODE MCT 
SromNameGetCapabilities(PINAMESPACE pThis, const _TCHAR * pName, 
                       NAME_SPACE_FLAGS *pFlags)
{
    ROM_FILE * File;
    SCODE sc;

    DBGME(2, _tprintf(_T("SromNameGetCapabilities %s\n"),pName));

    /* Lookup the pathname
     */
    sc = SromFileNew( (ROM_NAME*)pThis, pName, &File );

    if (FAILED(sc))
        return sc;

    /* Return result
     */
    if (pFlags != NULL)
        *pFlags = NAME_SPACE_READ | NAME_SPACE_SHARE_READ;

    /* No longer needs this
     */
    ((PIFILE) File)->v->Release((PIFILE)File);

    return NO_ERROR;
}



PRIVATE const struct INameSpaceVtbl SromNameVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    SromNameQueryInterface, 
    SromNameAddRef, 
    SromNameRelease, 
    SromNameRegister, 
    SromNameUnregister, 
    SromNameBind, 
    SromNameFindNext, 
    SromNameGetCapabilities
};


/*
 * Constructor for this Name Space
 */
EXTERN_C PINAMESPACE SromFsNew( PIFILE Storage)
{
    PIHEAP pIHeap;
    ROM_HEADER *pHdr, Header;
    ROM_NAME *pfn;
    SCODE sc;

    /* Read in and check the header
     */
    sc = ReadFromStorage(Storage, 0, (UINT8*)&Header, sizeof Header);

    /* Sanity check
     */
    if ((FAILED(sc)) || (Header.Signature != ROM_SIGNATURE)) {
        DBGME(2,_tprintf(_T("Bad ROMFS sc=x%x sig=x%x\n"),
                               sc, Header.Signature));
        return NULL;
    }

    /* Allocate room for object and copy of root info
     */
    pIHeap = CurrentHeap();

    pfn = (ROM_NAME *) pIHeap->v->Alloc(pIHeap, 0,
                                        sizeof(ROM_NAME) + Header.FilesOffset,
                                        0);
    if (pfn == NULL) {
        return NULL;
    }

    pHdr = (ROM_HEADER *) (pfn + 1);

    sc = ReadFromStorage(Storage, 0, (UINT8*)pHdr, Header.FilesOffset);
    if (FAILED(sc))
        goto ErrorExit;

    pfn->v = &SromNameVtbl;
    pfn->RefCnt = 1;
    pfn->Storage = Storage;
    Storage->v->AddRef(Storage);
    pfn->pHeader = pHdr;

    DBGME(1,_tprintf(_T("SROMFS returns %x\n"),(ADDRESS) pfn));

    return (PINAMESPACE) pfn;

 ErrorExit:
    pIHeap->v->Free(pIHeap,0,pfn);
    return NULL;
}
