/* Implementation of doubly-linked lists and nodes for queues and stacks.
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <util/list-double.h>
#include <assert.h>

/* Optionally let an architecture provide specially optimized
 * versions of these functions.
 */
#if !defined(__MACHDEP_LISTFUNC)

/* Pushes psnNew onto stack rooted at psnStackHead.
 * Returns new stack head, which is psnNew.
 * psnNew must be non-NULL and alone.
 */
PDNODE DNode_Push(
    PDNODE pdnStackHead,
    PDNODE pdnNew
    )
{
  PDNODE pdnTail;

  assert(pdnNew != NULL);
  assert(DNode_IsAlone(pdnNew));

  if (pdnStackHead != NULL) {
    /* If list was non-empty */
    pdnTail = DNode_GetPrev(pdnStackHead);
    DNode_Link(pdnTail, pdnNew);
  }
  
  
  /* Handle the case where the stack is initially empty */
  if (pdnStackHead == NULL) {
    pdnStackHead = pdnNew;
  }

  DNode_Link(pdnNew, pdnStackHead);
  return pdnNew;
}

/* Pops the top node on a stack rooted at pdnStackHead and stores a pointer
 * to the popped node in pdnPopped. Returns the new stack head.
 * All arguments must be non-NULL.
 */
PDNODE DNode_Pop(
    PDNODE pdnStackHead
    )
{
  PDNODE pdnNewStackHead;
  PDNODE pdnPrev;
  PDNODE pdnNext;

  assert(pdnStackHead != NULL);
/*   assert(ppdnPopped != NULL); */

  pdnNewStackHead = DNode_GetNext(pdnStackHead);
  pdnPrev = DNode_GetPrev(pdnStackHead);
  pdnNext = DNode_GetNext(pdnStackHead);

  if (pdnNewStackHead == pdnStackHead) {
    /* if stack contained one node, popping it makes it empty */
    pdnNewStackHead = NULL;
  }

  /* Make the old stack head alone */
  DNode_Link(pdnStackHead, pdnStackHead);

  /* Close the gap around it */
  DNode_Link(pdnPrev, pdnNext);

/*   *ppdnPopped = pdnStackHead; */

  return pdnNewStackHead;
}

/* None of the arguments should be null */
void DNode_InsertAfter(
    PDNODE pdnInsertAfter,
    PDNODE pdnNew
    )
{
  PDNODE pdnOldAfter;

  assert(pdnNew != NULL);
  /* pdnNew must not already be on a list or stack;
   * if so, call DList_Remove first */
  assert(DNode_IsAlone(pdnNew));
  assert(pdnInsertAfter != NULL);

  pdnOldAfter = DNode_GetNext(pdnInsertAfter);

  DNode_Link(pdnNew, pdnOldAfter);
  DNode_Link(pdnInsertAfter, pdnNew);
}

/* pdnPrevious should not be alone; you can't remove it from itself */
void DNode_RemoveAfter(
    PDNODE pdnPrevious
    )
{
  PDNODE pdnNewNext;
  PDNODE pdnNext;

  assert(pdnPrevious != NULL);
  assert(!DNode_IsAlone(pdnPrevious));

  pdnNext = DNode_GetNext(pdnPrevious);
  pdnNewNext = DNode_GetNext(pdnNext);

  /* Make the removed node alone */
  DNode_Link(pdnNext, pdnNext);
  /* Link up the remaining nodes */
  DNode_Link(pdnPrevious, pdnNewNext);
}

/* Removes pdnRemove from the circular stack rooted at pdnStackHead.
 * Returns the new stack head, which is the same as the old stack head
 * unless the old stack head was the one removed, in which case NULL is
 * returned. All arguments must be non-NULL.
 */
PDNODE DNode_StackRemove(
    PDNODE pdnStackHead,
    PDNODE pdnRemove
    )
{
  PDNODE pdnCurrent;
  PDNODE pdnPrev;
  PDNODE pdnNext;

  assert(pdnRemove != NULL);
  /* If pdnRemove is alone, it doesn't need to be removed */
  assert(!DNode_IsAlone(pdnRemove));

  pdnCurrent = DNode_GetNext(pdnStackHead);

  while ((pdnCurrent != pdnStackHead) && (pdnCurrent != pdnRemove)) {
    pdnCurrent = DNode_GetNext(pdnCurrent);
  }

  if (pdnCurrent != pdnRemove) {
    /* pdnRemove not found */
    return pdnStackHead;
  }

  /* Close the gap around it */
  pdnPrev = DNode_GetPrev(pdnRemove);
  pdnNext = DNode_GetNext(pdnRemove);
  DNode_Link(pdnPrev, pdnNext);
  /* Make it alone */
  DNode_Link(pdnRemove, pdnRemove);

  if (pdnRemove == pdnStackHead) {
    return pdnNext;
  }

  return pdnStackHead;
}

/* Makes every node in this stack alone.
 * Returns the new stack head, which is always NULL.
 * pdnStackHead should not be null.
 */
PDNODE DNode_StackClear(
    PDNODE pdnStackHead
    )
{
  PDNODE pdnCurrent;
  PDNODE pdnNext;

  assert(pdnStackHead != NULL);

  pdnCurrent = pdnStackHead;

  do {
    pdnNext = DNode_GetNext(pdnCurrent);
    DNode_Link(pdnCurrent, pdnCurrent);
    pdnCurrent = pdnNext;
  } while (pdnCurrent != pdnStackHead);

  return NULL;
}

/* Traverses list plClear and removes all nodes from head to tail.
 * i.e. sets their pdnNext to NULL, but does not change their data.
 * Then sets head and tail to NULL.
 * plClear must not be NULL and must have already been initialized.
 */
void DList_Clear(
    PDLIST pdlThis
    )
{
  assert(pdlThis != NULL);
  DNode_StackClear(DList_GetHead(pdlThis));
  DList_SetHead(pdlThis, NULL);
}    

/* Inserts the node pdnNew at the head of the NULL-terminated list pdlThis.
 * All of the arguments should be non-NULL.
 * pdnNew cannot already be on a list (pdnNew->pdnNext == NULL),
 * otherwise we corrupt both the old list and the new list.
 */
void DList_HeadAdd(
    PDLIST pdlThis,
    PDNODE pdnNew
    )
{
  assert(pdlThis != NULL);

  DList_SetHead(pdlThis, DNode_Push(DList_GetHead(pdlThis), pdnNew));
}

/* Inserts the node pdnNew at the tail of the NULL-terminated list pdlThis.
 * All of the arguments should be non-NULL.
 * pdnNew cannot already be on a list (pdnNew->pdnNext == NULL),
 * otherwise we corrupt both the old list and the new list.
 */
void DList_TailAdd(
    PDLIST pdlThis,
    PDNODE pdnNew
    )
{
  assert(pdlThis != NULL);

  if (DList_IsEmpty(pdlThis)) {
    /* List is empty */
    DList_SetHead(pdlThis, pdnNew);
  }

  DNode_Push(DList_GetHead(pdlThis), pdnNew);
}

/* Removes the node at the head of pdlThis and returns it.
 * If list is empty, returns NULL.
 * Updates head and tail pointers of plRemoveFrom accordingly.
 * All the arguments should be non-NULL.
 */
PDNODE DList_HeadRemove(
    PDLIST pdlThis
    )
{
  PDNODE pdnOldHead;

  assert(pdlThis != NULL);

  pdnOldHead = DList_GetHead(pdlThis);
  DList_SetHead(pdlThis, DNode_Pop(pdnOldHead));

  if (pdnOldHead == NULL) {
    /* List is already empty */
    return NULL;
  }

  return pdnOldHead;

}

/* Removes the node at the tail of plRemoveFrom and returns it.
 * If list is empty, returns NULL.
 * Updates head and tail pointers of plRemoveFrom accordingly.
 * All the arguments should be non-NULL.
 */
PDNODE DList_TailRemove(
    PDLIST plRemoveFrom
    )
{
  PDNODE pdnOldTail;

  assert(plRemoveFrom != NULL);

  pdnOldTail = DList_GetTail(plRemoveFrom);

  if (pdnOldTail == NULL) {
    return NULL;
  }
  else {
    PDNODE pdnNewTail = DList_GetHead(plRemoveFrom);

    if (pdnNewTail == pdnOldTail) {
      DList_Init(plRemoveFrom);
      return pdnOldTail;
    }

    while (DNode_GetNext(pdnNewTail) != pdnOldTail) {
      pdnNewTail = DNode_GetNext(pdnNewTail);
    }
    DNode_RemoveAfter(pdnNewTail);
/*     DList_SetTail(plRemoveFrom, pdnNewTail); */

    return pdnOldTail;
  }
}

void DList_Remove(
    PDLIST pdlThis,
    PDNODE pdnRemove
    )
{
  PDNODE pdnCurrent;
  PDNODE pdnPrevious = NULL;

  pdnCurrent = DList_GetHead(pdlThis);

  while (pdnCurrent != NULL) {
    if (pdnCurrent == pdnRemove) {
      break;
    }
    pdnPrevious = pdnCurrent;
    pdnCurrent = DNode_GetNext(pdnCurrent);
  }

  /* node must be in list */
  assert(pdnCurrent != NULL);

  if (pdnCurrent == DList_GetHead(pdlThis)) {
    DList_HeadRemove(pdlThis);
  }
  else {
    DNode_RemoveAfter(pdnPrevious);
  }
}

/* Returns a pointer to the node containing pFindData if it exists.
 * Otherwise returns NULL.
 * plFindList must be non-NULL.
 */
PDNODE DList_FindData(
    PDLIST pdlThis,
    PTR pFindData
    )
{
  PDNODE pdnCurrent;
  PDNODE pdnHead;

  assert(pdlThis != NULL);

  if (DList_IsEmpty(pdlThis)) {
    return NULL;
  }

  pdnCurrent = pdnHead = DList_GetHead(pdlThis);

  do {
    if (DNode_GetData(pdnCurrent) == pFindData) {
      return pdnCurrent;
    }
    pdnCurrent = DNode_GetNext(pdnCurrent);
  } while (pdnCurrent != pdnHead);
    
  return NULL;
}

/* Finds the node in the given list matching the data pFindData and
 * removes it from the list, updating list head/tail pointers accordingly.
 * Otherwise returns NULL.
 * pdlThisList must be non-NULL.
 */
PDNODE DList_FindDataAndRemove(
    PDLIST pdlThis,
    PTR pFindData
    )
{
  PDNODE pdnCurrent;
  PDNODE pdnHead;
  PDNODE pdnPrevious = NULL;

  assert(pdlThis != NULL);

  if (DList_IsEmpty(pdlThis)) {
    return NULL;
  }

  pdnCurrent = pdnHead = DList_GetHead(pdlThis);

  do {
    if (DNode_GetData(pdnCurrent) == pFindData) {
      break;
    }
    pdnPrevious = pdnCurrent;
    pdnCurrent = DNode_GetNext(pdnCurrent);
  } while (pdnCurrent != pdnHead);

  if (pdnCurrent == DList_GetHead(pdlThis)) {
    DList_HeadRemove(pdlThis);
  }
  else {
    DNode_RemoveAfter(pdnPrevious);
/*     if (pdnCurrent == DList_GetTail(pdlThis)) { */
/*       DList_SetTail(pdlThis, pdnPrevious); */
/*     } */
  }
  return pdnCurrent;
}

PTR DList_FindDataMax(
    PDLIST pdlThis,
    BOOL (*CompareGT)(PTR, PTR)
    )
{
  PTR pCurrent;
  PTR pMax;
  PDNODE pdnMax;
  PDNODE pdnCurrent;
  PDNODE pdnHead;

  assert(pdlThis != NULL);

  if (DList_IsEmpty(pdlThis)) {
    return NULL;
  }

  pdnHead = DList_GetHead(pdlThis);
  pMax = DNode_GetData(pdnHead);
  pdnMax = pdnHead;

  /* If the list only has one node, it is the maximum */
  if (DNode_IsAlone(pdnHead)) {
    return pMax;
  }

  pdnCurrent = DNode_GetNext(pdnHead);

  do {
    pCurrent = DNode_GetData(pdnCurrent);
    if (CompareGT(pCurrent, pMax)) {
      pMax = pCurrent;
      pdnMax = pdnCurrent;
    }
    pdnCurrent = DNode_GetNext(pdnCurrent);
  } while (pdnCurrent != pdnHead);

  return pMax;
}    

PTR DList_RemoveDataMax(
    PDLIST pdlThis,
    BOOL (*CompareGT)(PTR, PTR)
    )
{
  PTR pCurrent;
  PTR pMax;
  PDNODE pdnMax;
  PDNODE pdnPrevMax = NULL;
  PDNODE pdnCurrent;
  PDNODE pdnPrevious = NULL;
  PDNODE pdnHead;

  assert(pdlThis != NULL);
  assert(CompareGT != NULL);

  if (DList_IsEmpty(pdlThis)) {
    return NULL;
  }

  pdnHead = DList_GetHead(pdlThis);

  assert(pdnHead != NULL);

  pMax = DNode_GetData(pdnHead);
  pdnMax = pdnHead;

  /* If the list only has one node, it is the maximum */
  if (DNode_IsAlone(pdnHead)) {
    DList_HeadRemove(pdlThis);
    return pMax;
  }

  pdnCurrent = DNode_GetNext(pdnHead);

  do {
    pCurrent = DNode_GetData(pdnCurrent);
    if (CompareGT(pCurrent, pMax)) {
      pMax = pCurrent;
      pdnMax = pdnCurrent;
      pdnPrevMax = pdnPrevious;
    }
    pdnPrevious = pdnCurrent;
    pdnCurrent = DNode_GetNext(pdnCurrent);
  } while (pdnCurrent != pdnHead);

  assert(pdnPrevious != NULL);

  if (pdnMax == pdnHead) {
    /* Discard the head */
    DList_HeadRemove(pdlThis);
  }
  else {
/*     if (pdnMax == DList_GetTail(pdlThis)) { */
/*       DList_SetTail(pdlThis, pdnPrevMax); */
/*     } */

    DNode_RemoveAfter(pdnPrevMax);
  }

  return pMax;
}    

void DList_TransformData(
    PDLIST pdlThis,
    void (*Transform)(PTR)
    )
{
  PDNODE pdnCurrent;
  PDNODE pdnHead;

  assert(pdlThis != NULL);
  assert(Transform != NULL);

  if (DList_IsEmpty(pdlThis)) {
    return;
  }

  pdnCurrent = pdnHead = DList_GetHead(pdlThis);

  do {
    Transform(DNode_GetData(pdnCurrent));
    pdnCurrent = DNode_GetNext(pdnCurrent);
  } while (pdnCurrent != pdnHead);
}

void DList_TransformDataWithArg(
    PDLIST pdlThis,
    void (* Transform)(PTR, PTR),
    PTR pArg
    )
{
  PDNODE pdnCurrent;
  PDNODE pdnHead;

  assert(pdlThis != NULL);
  assert(Transform != NULL);
  assert(pArg != NULL);

  if (DList_IsEmpty(pdlThis)) {
    return;
  }

  pdnCurrent = pdnHead = DList_GetHead(pdlThis);

  do {
    Transform(DNode_GetData(pdnCurrent), pArg);
    pdnCurrent = DNode_GetNext(pdnCurrent);
  } while (pdnCurrent != pdnHead);
}

/* pdlThis must already be sorted in descending order using
 * CompareGT. All arguments must be non-NULL.
 */
void DList_InsertSortedDescending(
    PDLIST pdlThis,
    BOOL (*CompareGT)(PTR, PTR),
    PDNODE pdnNew
    )
{
  PTR pNewData;
  PDNODE pdnCurrent;
  PDNODE pdnPrevious;
  PDNODE pdnHead;

  assert(pdnNew != NULL);
  assert(pdlThis != NULL);
  assert(CompareGT != NULL);

  pNewData = DNode_GetData(pdnNew);
  pdnCurrent = pdnHead = DList_GetHead(pdlThis);
  pdnPrevious = NULL;
  
  if (!DList_IsEmpty(pdlThis)) {

    do {
      if (!CompareGT(DNode_GetData(pdnCurrent), pNewData)) {
	break;
      }
      pdnPrevious = pdnCurrent;
      pdnCurrent = DNode_GetNext(pdnCurrent);
    } while (pdnCurrent != pdnHead);
  }

  if (pdnPrevious == NULL) {
    /* List was initially empty, or new node has maximal data */
    DList_HeadAdd(pdlThis, pdnNew);
  }
  else {
    DNode_InsertAfter(pdnPrevious, pdnNew);
  }
}

#endif /* __MACHDEP_LISTFUNC */
