/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <assert.h>

#ifdef _DEBUG
PRIVATE INT NodeEq(PTR p1, PTR p2)
{
    return p1 != p2;
}
#endif

/* Insert new node pNew after an existing node pNode.
 * Returns pointer to the new node.
 */
PRIVATE void ListNodeInsert(UINT Offset, PTR pNode, PTR pNew)
{
    ListNodeNext(Offset, pNew) = ListNodeNext(Offset, pNode);
    ListNodePrev(Offset, pNew) = pNode;

    ListNodePrev(Offset, ListNodeNext(Offset, pNode)) = pNew;
    ListNodeNext(Offset, pNode) = pNew;
}

/* Remove a node from a list.
 * Returns pointer to the next node; NULL if this was the only node.
 */
PTR ListNodeRemove(UINT Offset, PTR pNode)
{
    PLIST_NODE_LINK pNodeLink = ListNodeGetLink(Offset, pNode);
    PTR pNext;

    if ((pNext = pNodeLink->pNext) == pNode)
        return NULL;

    ListNodePrev(Offset, pNext) = pNodeLink->pPrev;
    ListNodeNext(Offset, pNodeLink->pPrev) = pNext;
    pNodeLink->pPrev = pNodeLink->pNext = pNode;
    return pNext;
}

void ListAddFirst(PTR pHead, UINT Offset, PTR pNew)
{
    PLIST_HEAD Head = pHead;
    PTR pItem = *Head;

    if (pItem) {
        ListNodeInsert(Offset, ListNodePrev(Offset, pItem), pNew);
    } else {
        ListNodeInit(Offset, pNew);
    }
    *Head = pNew;
}

void ListAddBefore(PTR pHead, UINT Offset, PTR pItem, PTR pNew)
{
    PLIST_HEAD Head = pHead;
    /* Check that item is on list. */
    assert(pItem&&ListWalk(pHead,Offset,LIST_WALK_SEARCH,NodeEq,pItem)==pItem);

    ListNodeInsert(Offset, ListNodePrev(Offset, pItem), pNew);
    if (*Head == pItem)
        *Head = pNew;
}

void ListAddLast(PTR pHead, UINT Offset, PTR pNew)
{
    PLIST_HEAD Head = pHead;
    PTR pItem = *Head;

    if (pItem) {
        ListNodeInsert(Offset, ListNodePrev(Offset, pItem), pNew);
    } else {
        ListNodeInit(Offset, pNew);
        *Head = pNew;
    }
}

void ListAddAfter(PTR pHead, UINT Offset, PTR pItem, PTR pNew)
{
    PLIST_HEAD Head = pHead;
    /* Check that item is on list */
    assert(pItem&&ListWalk(pHead,Offset,LIST_WALK_SEARCH,NodeEq,pItem)==pItem);

    ListNodeInsert(Offset, pItem, pNew);
}

PTR ListRemoveFirst(PTR pHead, UINT Offset)
{
    PLIST_HEAD Head = pHead;
    PTR pItem = *Head;

    if (pItem)
        *Head = ListNodeRemove(Offset, pItem);
    return pItem;
}

//PTR ListRemoveLast(PTR pHead, UINT Offset);
void ListRemoveItem(PTR pHead, UINT Offset, PTR pItem)
{
    PLIST_HEAD Head = pHead;
    PTR Next;
    assert(pItem&&ListWalk(pHead,Offset,LIST_WALK_SEARCH,NodeEq,pItem)==pItem);

    Next = ListNodeRemove(Offset, pItem);
    /* If this was the first one in the list, make head point to next */
    if (*Head == pItem)
        *Head = Next;
}

PTR ListWalk(PTR pHead, UINT Offset, LIST_WALK Function,
             INT (*Transform)(PTR pItem, PTR pArg), PTR pArg)
{
    PLIST_HEAD Head;
    PTR pItem, Next, bestItem;
    INT ret, bestret;
    BOOL sorted = Function & LIST_WALK_SORTED;
    BOOL backwards = Function & LIST_WALK_BACKWARDS;

    bestItem = NULL;
    bestret = 0;

    Function &= ~(LIST_WALK_SORTED | LIST_WALK_BACKWARDS);

    if (Function == LIST_WALK_LARGEST) {
        bestItem = pArg;
        sorted = TRUE;
    } else if (Function == LIST_WALK_BESTFIT) {
        sorted = TRUE;
    }

    /*
     * If we don't have a list to walk, punt.
     * For LIST_WALK_SCODE this is implicit success.
     */
    if (!pHead)
        return NULL;
    Head = pHead;

    pItem = *Head;
    if (backwards)
        pItem = ListNodePrev(Offset, pItem);

    if (pItem) do {
        if (backwards)
            Next = ListNodePrev(Offset, pItem);
        else
            Next = ListNodeNext(Offset, pItem);

        ret = Transform(pItem, pArg);

        if (Function) {
            if (Function == LIST_WALK_SCODE) {
                if (FAILED((SCODE) ret))
                    return (PTR) ret;
            } else if (sorted ? ret>=0 : ret==0) switch (Function) {
            case LIST_WALK_REMOVE:
                ListRemoveItem(pHead, Offset, pItem);
                return pItem;
            case LIST_WALK_ADD_AFTER:
                ListAddAfter(pHead, Offset, pItem, pArg);
                return NULL;
            case LIST_WALK_ADD:
                ListAddBefore(pHead, Offset, pItem, pArg);
                return NULL;
            case LIST_WALK_SEARCH:
                return pItem;
            case LIST_WALK_BESTFIT:
                if (ret == 0)
                    return pItem;

                if (bestret == 0 || ret < bestret) {
                    bestret = ret;
                    bestItem = pItem;
                }
                break;
            case LIST_WALK_LARGEST:
                if (ret != 0) /* ret == 0 case could go either way... */
                    pArg = bestItem = pItem;
                break;
            default:
                assert(FALSE);
            }
        }
        pItem = Next;
    } while (*Head && pItem != *Head);

    /* If not inserted anywhere else, insert last */
    if (Function == LIST_WALK_ADD || Function == LIST_WALK_ADD_AFTER)
        ListAddLast(pHead, Offset, pArg);

    assert((PTR) S_OK == NULL); /* when LIST_WALK_SCODE, returns S_OK */

    return bestItem;            /* NULL in most cases */
}

/* Copy list using Copier and return new head.
 * Upon failure calls Deleter for nodes already copied and returns NULL.
 * pArg might be heap etc.
 */
PTR ListCopy(const void *pHead, UINT Offset,
             PTR (*Copier)(PTR pItem, PTR pArg),
             INT (*Deleter)(PTR __pItem, PTR __pArg),
             PTR pArg)
{
    PTR nOld, nNew, hNew = NULL, hOld = *(PLIST_HEAD) pHead;

    nOld = hOld;
    if (nOld) do {
        nNew = Copier(nOld, pArg);
        if (!nNew) {
            ListWalk(&hNew, Offset, 0, Deleter, pArg);
            return NULL;
        }
        ListAddLast(&hNew, Offset, nNew);
        nOld = ListNodeNext(Offset, nOld);
    } while (hOld != nOld);
    return hNew;
}

#if 0
/*************** test case **********************/
typedef struct _TEST {
    PTR v;
    UINT RefCnt;
    ListNodeDeclare(_TEST, Link);
#define Link_off offsetof(TEST, Link)
    INT zzz;
} TEST, *PTEST;


TEST t1;
TEST t2 = {0, 1, ListNodeInitializer(t2), 42};
PTEST tHead = NULL;

INT NodeCmp(PTR p1, PTR p2)
{
    PTEST n1 = p1, n2 = p2;
    return n2->zzz - n1->zzz;
}

INT NodeDel(PTR p1, PTR p2)
{
    PTEST n1 = p1;
    return 0;
}

void ListTest(void)
{
    ListNodeInit(Link_off, &t1);
    t1.zzz = 7;

    ListAddLast(&tHead, Link_off, &t1);
    ListWalk(&tHead, Link_off, LIST_WALK_ADD, NodeCmp, &t2);
    ListWalk(&tHead, Link_off, LIST_WALK_REMOVE, NodeDel, &t2);
}
#endif
