/* Implementation of singly-linked lists and nodes for queues and stacks.
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <util/list-single.h>
#include <assert.h>

/* Optionally let an architecture provide specially optimized
 * versions of these functions.
 */
#if !defined(__MACHDEP_LISTFUNC)

/* Pushes pnNew onto stack rooted at pnStackHead.
 * Returns new stack head, which is pnNew.
 * pnNew must be non-NULL and alone.
 */
PNODE Node_Push(
    PNODE pnStackHead,
    PNODE pnNew
    )
{
  assert(pnNew != NULL);
  assert(Node_IsAlone(pnNew));

  Node_SetNext(pnNew, pnStackHead);
  return pnNew;
}

/* Pops the top node on a stack rooted at pnStackHead and stores a pointer
 * to the popped node in pnPopped. Returns the new stack head.
 * All arguments must be non-NULL.
 */
PNODE Node_Pop(
    PNODE pnStackHead,
    PPNODE ppnPopped
    )
{
  PNODE pnNewStackHead;

  if (pnStackHead == NULL) {
    return NULL;
  }

  pnNewStackHead = Node_GetNext(pnStackHead);
  Node_SetNext(pnStackHead, NULL);

  if (ppnPopped != NULL) {
    *ppnPopped = pnStackHead;
  }

  return pnNewStackHead;
}

void Node_InsertAfter(
    PNODE pnInsertAfter,
    PNODE pnNew
    )
{
  assert(pnNew != NULL);
  /* pnNew must not already be on a list or stack;
   * if so, call Node_Remove first */
  assert(Node_GetNext(pnNew) == NULL);

  if (pnInsertAfter != NULL) {
    Node_SetNext(pnNew, Node_GetNext(pnInsertAfter));
    Node_SetNext(pnInsertAfter, pnNew);
  }
}

void Node_RemoveAfter(
    PNODE pnPrevious
    )
{
  PNODE pnTemp;

  assert(pnPrevious != NULL);
  assert(!Node_IsAlone(pnPrevious));

  pnTemp = Node_GetNext(pnPrevious);
  if (pnTemp != NULL) {
    Node_SetNext(pnPrevious, Node_GetNext(pnTemp));
    Node_SetNext(pnTemp, NULL);
  }
}

/* Removes pnRemove from the NULL-terminated stack rooted at pnStackHead.
 * Returns the new stack head, which is the same as the old stack head
 * unless the old stack head was the one removed.
 * All arguments must be non-NULL.
 */
PNODE Node_StackRemove(
    PNODE pnStackHead,
    PNODE pnRemove
    )
{
  PNODE pnCurrent;
  PNODE pnPrevious;

  assert(pnRemove != NULL);

  pnCurrent = pnStackHead;
  pnPrevious = NULL;

  while ((pnCurrent != NULL) && (pnCurrent != pnRemove)) {
    pnPrevious = pnCurrent;
    pnCurrent = Node_GetNext(pnCurrent);
  }

  if (pnCurrent == NULL) {
    /* Stack is empty or pnRemove not found */
    return pnStackHead;
  }
  else if (pnPrevious == NULL) {
    /* pnRemove = pnStackHead */
    PNODE pnNewStackHead = Node_GetNext(pnStackHead);
    Node_SetNext(pnRemove, NULL);
    return pnNewStackHead;
  }
  else {
    /* pnRemove was found, is not the stack head,
     * and pnPrevious precedes it */
    Node_RemoveAfter(pnPrevious);
    return pnStackHead;
  }
}

/* Makes every node in this stack alone (next == NULL).
 * Returns the new stack head, which is always NULL.
 */
PNODE Node_StackClear(
    PNODE pnStackHead
    )
{
  PNODE pnCurrent;
  PNODE pnNext;

  pnCurrent = pnStackHead;

  while (pnCurrent != NULL) {
    pnNext = Node_GetNext(pnCurrent);
    Node_SetNext(pnCurrent, NULL);
    pnCurrent = pnNext;
  }

  return NULL;
}

/* Traverses list pslThis and removes all nodes from head to tail.
 * i.e. sets their pnNext to NULL, but does not change their data.
 * Then sets head and tail to NULL.
 * pslThis must not be NULL and must have already been initialized.
 */
void List_Clear(
    PLIST pslThis
    )
{
  assert(pslThis != NULL);

  Node_StackClear(List_GetHead(pslThis));
  List_Init(pslThis);
}    

/* Inserts the node pnNew at the head of the NULL-terminated list pslThis.
 * All of the arguments should be non-NULL.
 * pnNew cannot already be on a list (pnNew->pnNext == NULL),
 * otherwise we corrupt both the old list and the new list.
 */
void List_HeadAdd(
    PLIST pslThis,
    PNODE pnNew
    )
{
  PNODE pnHead;

  assert(pslThis != NULL);

  pnHead = List_GetHead(pslThis);
  if (pnHead == NULL) {
    /* List was initially empty */
    List_SetTail(pslThis, pnNew);
  }

  List_SetHead(pslThis, Node_Push(pnHead, pnNew));
}

/* Inserts the node pnNew at the tail of the NULL-terminated list pslThis.
 * All of the arguments should be non-NULL.
 * pnNew cannot already be on a list (pnNew->pnNext == NULL),
 * otherwise we corrupt both the old list and the new list.
 */
void List_TailAdd(
    PLIST pslThis,
    PNODE pnNew
    )
{
  PNODE pnOldTail;

  assert(pslThis != NULL);
  assert(Node_IsAlone(pnNew));

  pnOldTail = List_GetTail(pslThis);

  if (pnOldTail == NULL) {
    /* list was initially empty */
    List_SetHead(pslThis, pnNew);
  }
  else {
    Node_SetNext(pnOldTail, pnNew);
  }
  
  List_SetTail(pslThis, pnNew);
}

/* Removes the node at the head of pslRemoveFrom and returns it.
 * If list is empty, returns NULL.
 * Updates head and tail pointers of pslThis accordingly.
 * All the arguments should be non-NULL.
 */
PNODE List_HeadRemove(
    PLIST pslThis
    )
{
  PNODE pnOldHead;

  assert(pslThis != NULL);

  List_SetHead(pslThis, Node_Pop(List_GetHead(pslThis), &pnOldHead));

  if (pnOldHead == NULL) {
    /* List is already empty */
    return NULL;
  }
  else {
    if (pnOldHead == List_GetTail(pslThis)) {
      /* Removing only node make list empty. */
      List_SetTail(pslThis, NULL);
    }
      
    return pnOldHead;
  }
}

/* Removes the node at the tail of pslRemoveFrom and returns it.
 * If list is empty, returns NULL.
 * Updates head and tail pointers of pslRemoveFrom accordingly.
 * All the arguments should be non-NULL.
 */
PNODE List_TailRemove(
    PLIST pslThis
    )
{
  PNODE pnOldTail = List_GetTail(pslThis);
  if (pnOldTail != NULL) {
    /* list is not empty */
    List_Remove(pslThis, pnOldTail);
  }
  return pnOldTail;
}

/* Removes pnRemove from pslThis (it MUST be present, or assertion fails). */
void List_Remove(
    PLIST pslThis,
    PNODE pnRemove
    )
{
  PNODE pnCurrent;
  PNODE pnPrevious = NULL;

  assert(pslThis != NULL);
  assert(pnRemove != NULL);

  pnCurrent = List_GetHead(pslThis);

  while (pnCurrent != NULL) {
    if (pnCurrent == pnRemove) {
      break;
    }
    pnPrevious = pnCurrent;
    pnCurrent = Node_GetNext(pnCurrent);
  }

  /* pnRemove must be in the list */
  assert(pnCurrent != NULL);

  if (pnCurrent == List_GetHead(pslThis)) {
    List_HeadRemove(pslThis);
  }
  else {
    Node_RemoveAfter(pnPrevious);
    if (pnCurrent == List_GetTail(pslThis)) {
      List_SetTail(pslThis, pnPrevious);
    }
  }
}

/* Frees all nodes in this list.
 * Only call this function if all nodes have been dynamically allocated.
 */
/*  void List_Free(PLIST pslFree) */
/*  { */
/*      PNODE pnCurrent; */
/*      PNODE pnLeader; */
/*      PNODE pnHead; */

/*      assert(pslFree != NULL); */

/*      pnHead = List_GetHead(pslFree); */

/*      if (pnHead == NULL) { */
/*        return; */
/*      } */

/*      pnCurrent = pnHead; */
/*      do { */
/*        pnLeader = Node_GetNext(pnCurrent); */
/*        pIHeap->v->Free(0, pnCurrent); */
/*        pnCurrent = pnLeader; */
/*      } while (pnCurrent != pnHead); */
/*  } */

/* Returns a pointer to the node containing pFindData if it exists.
 * Otherwise returns NULL.
 * pslFindList must be non-NULL.
 */
PNODE List_FindData(
    PLIST pslFind,
    PTR pFindData
    )
{
  PNODE pnCurrent;

  assert(pslFind != NULL);

  pnCurrent = List_GetHead(pslFind);

  while (pnCurrent != NULL) {
    if (Node_GetData(pnCurrent) == pFindData) {
      return pnCurrent;
    }
    pnCurrent = Node_GetNext(pnCurrent);
  }
    
  return NULL;
}

/* Finds the node in the given list matching the data pFindData and
 * removes it from the list, updating list head/tail pointers accordingly.
 * Otherwise returns NULL.
 * pslFindList must be non-NULL.
 */
PNODE List_FindDataAndRemove(
    PLIST pslFind,
    PTR pFindData
    )
{
  PNODE pnCurrent;
  PNODE pnPrevious = NULL;

  assert(pslFind != NULL);

  pnCurrent = List_GetHead(pslFind);

  while (pnCurrent != NULL) {
    if (Node_GetData(pnCurrent) == pFindData) {
      break;
    }
    pnPrevious = pnCurrent;
    pnCurrent = Node_GetNext(pnCurrent);
  }

  if (pnCurrent == List_GetHead(pslFind)) {
    List_HeadRemove(pslFind);
  }
  else {
    Node_RemoveAfter(pnPrevious);
    if (pnCurrent == List_GetTail(pslFind)) {
      List_SetTail(pslFind, pnPrevious);
    }
  }
  return pnCurrent;
}

PTR List_FindDataMax(
    PLIST pslSearch,
    CompareFunc CompareGT
    )
{
  PTR pCurrent;
  PTR pMax;
  PNODE pnMax;
  PNODE pnCurrent;
  PNODE pnHead;

  assert(pslSearch != NULL);

  if (List_IsEmpty(pslSearch)) {
    return NULL;
  }

  pnHead = List_GetHead(pslSearch);
  pMax = Node_GetData(pnHead);
  pnMax = pnHead;

  /* If the list only has one node, it is the maximum */
  if (Node_IsAlone(pnHead)) {
    return pMax;
  }

  pnCurrent = Node_GetNext(pnHead);

  while (pnCurrent != NULL) {
    pCurrent = Node_GetData(pnCurrent);
    if (CompareGT(pCurrent, pMax)) {
      pMax = pCurrent;
      pnMax = pnCurrent;
    }
    pnCurrent = Node_GetNext(pnCurrent);
  }

  return pMax;
}    

PTR List_RemoveDataMax(
    PLIST pslSearch,
    CompareFunc CompareGT
    )
{
  PTR pCurrent;
  PTR pMax;
  PNODE pnMax;
  PNODE pnPrevMax = NULL;
  PNODE pnCurrent;
  PNODE pnPrevious = NULL;
  PNODE pnHead;

  assert(pslSearch != NULL);
  assert(CompareGT != NULL);

  if (List_IsEmpty(pslSearch)) {
    return NULL;
  }

  pnHead = List_GetHead(pslSearch);

  assert(pnHead != NULL);

  pMax = Node_GetData(pnHead);
  pnMax = pnHead;

  /* If the list only has one node, it is the maximum */
  if (Node_IsAlone(pnHead)) {
    List_HeadRemove(pslSearch);
    return pMax;
  }

  pnCurrent = Node_GetNext(pnHead);

  while (pnCurrent != NULL) {
    pCurrent = Node_GetData(pnCurrent);
    if (CompareGT(pCurrent, pMax)) {
      pMax = pCurrent;
      pnMax = pnCurrent;
      pnPrevMax = pnPrevious;
    }
    pnPrevious = pnCurrent;
    pnCurrent = Node_GetNext(pnCurrent);
  }

  assert(pnPrevious != NULL);

  if (pnMax == pnHead) {
    /* Discard the head */
    List_HeadRemove(pslSearch);
  }
  else {
    if (pnMax == List_GetTail(pslSearch)) {
      List_SetTail(pslSearch, pnPrevMax);
    }

    Node_RemoveAfter(pnPrevMax);
  }

  return pMax;
}    

void List_TransformData(
    PLIST pslThis,
    TransformFunc Transform
    )
{
  PNODE pnCurrent;
  PNODE pnNext;

  assert(pslThis != NULL);
  assert(Transform != NULL);

  pnCurrent = List_GetHead(pslThis);

  while (pnCurrent != NULL) {
    /* We get the next pointer first, b/c transforming may make it null */
    pnNext = Node_GetNext(pnCurrent);
    Transform(Node_GetData(pnCurrent));
    pnCurrent = pnNext;
  }
}

void List_TransformDataWithArg(
    PLIST pslThis,
    TransformArgFunc Transform,
    PTR pArg
    )
{
  PNODE pnCurrent;
  PNODE pnNext;

  assert(pslThis != NULL);
  assert(Transform != NULL);
/*   assert(pArg != NULL); */

  pnCurrent = List_GetHead(pslThis);

  while (pnCurrent != NULL) {
    /* We get the next pointer first, b/c transforming may make it null */
    pnNext = Node_GetNext(pnCurrent);
    Transform(Node_GetData(pnCurrent), pArg);
    pnCurrent = pnNext;
  }
}

/* pslThis must already be sorted in descending order using
 * CompareGT. All arguments must be non-NULL.
 */
void List_InsertSortedDescending(
    PLIST pslThis,
    CompareFunc CompareGT,
    PNODE pnNew
    )
{
  PTR pNewData;
  PNODE pnCurrent;
  PNODE pnPrevious = NULL;
  PNODE pnHead;

  assert(pnNew != NULL);
  assert(pslThis != NULL);
  assert(CompareGT != NULL);

  pNewData = Node_GetData(pnNew);

  pnCurrent = pnHead = List_GetHead(pslThis);

  while (pnCurrent != NULL) {
    if (!CompareGT(Node_GetData(pnCurrent), pNewData)) {
      break;
    }
    pnPrevious = pnCurrent;
    pnCurrent = Node_GetNext(pnCurrent);
  }

  if (pnCurrent == pnHead) {
    /* List was initially empty, or new node has maximal data */
    List_HeadAdd(pslThis, pnNew);
  }
  else if (pnCurrent == NULL) {
    /* New node has minimal data; we fell off the end of the list */
    List_TailAdd(pslThis, pnNew);
  }
  else {
    Node_InsertAfter(pnPrevious, pnNew);
  }
}

#endif /* __MACHDEP_LISTFUNC */
