/* Implementation of generalized logging facility.
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <stdio.h>
#include <mmlite.h>
#define LOGGING_ENABLED 1
#include <util/logger.h>
#include <assert.h>

/* Uncomment the following line to print log entries */
/* #define PRINTF_DEBUG */

/* Type index for an initialized log entry */
#define LOG_UNUSED 0

void Log_Init(
    PLOG plThis,
    UINT BufferSize,
    PLOG_ENTRY pleBuffer
    )
{
  UINT i;
  PLOG_ENTRY pleLocal;

  assert(plThis != NULL);
  assert(pleBuffer);

  // Allocate buffer of log entries
  plThis->pleBuffer = pleBuffer;
  plThis->BufferSize = BufferSize;
  /* What's the point of a zero-size log? */
  assert(BufferSize != 0);

  pleLocal = pleBuffer;

  // initialize all entries by making the types null
  for (i = 0; i < BufferSize; i++) {
    pleLocal[i].pletType = NULL;
/*       LogBuffer[i].Time = t; */
  }

  plThis->EntryIndex = 0;
}

/* Writes a log event */
void Log_WriteEntry(PLOG plThis, PLOG_EVENT_TYPE pletType, UINT Data)
{
    TIME tNow = CurrentTime();
    PLOG_ENTRY pleBuffer;
    UINT BufferSize;
    UINT SyncIndex;
    PUINT pEntryIndexAddr;

    assert(plThis);
    assert(pletType != NULL);

    if (!(pletType->bIsEnabled)) {
      return;
    }
    
    BufferSize = plThis->BufferSize;
    /* Make sure Log_Init has already been called with non-zero buffer size */
    assert(BufferSize != 0);
    pleBuffer = plThis->pleBuffer;

    /* Wraparound log index if it reaches the end. */
    if (plThis->EntryIndex >= BufferSize) {
      plThis->EntryIndex = 0;
    }

    pEntryIndexAddr = &(plThis->EntryIndex);

    while (TRUE) {

      /* Check to see if logging is disabled because schedule printing
       * is in progress (see PrintSchedulerLog below) */
      if (*pEntryIndexAddr == ~0) {
	return;
      }

      /* Set temp variable for synchronization check */
      SyncIndex = AtomicInc(pEntryIndexAddr);
      /* Check to see if we need the index to wraparound */
      if (SyncIndex >= BufferSize) {
	/* Wraparound the log index if PrintSchedulerLog hasn't already
	 * signaled for a block
	 */
	(void) AtomicCmpAndSwap(pEntryIndexAddr, SyncIndex, 0);
      } else {
	/* Logging is not blocked, go ahead and record a new event. */
	break;
      }
    }

    SyncIndex--;
    pleBuffer[SyncIndex].tTime = tNow;
    pleBuffer[SyncIndex].pletType = pletType;
    pleBuffer[SyncIndex].Data = Data;
#ifdef PRINTF_DEBUG
    printf("LOG %d %08x %ld\n",
	   pleBuffer[SyncIndex].Type->szName,
	   pleBuffer[SyncIndex].Data,
	   pleBuffer[SyncIndex].Time);
#endif
}

/* You must disable preemption in the scheduler before calling this function
 * if you do not wish log printing to be interrupted.
 */
void Log_Print(PLOG plThis)
{
    UINT EntryIndex;
    UINT LogIndexMarker;
    UINT BufferSize;
    PLOG_ENTRY pleBuffer;
    PLOG_ENTRY pleCurrent; /* iterates through event types */
/*     TIME tNow = CurrentTime(); */
    TIME tMarker; /* iterates through event times */
    TIME tEntryTime;
    PLOG_EVENT_TYPE pletType;
    UINT* pEntryIndexAddr = &(plThis->EntryIndex);
    UINT StopIndex;

    /* Block the log if it's currently being printed via Log_Print */
    while (TRUE) {
      LogIndexMarker = *pEntryIndexAddr;
      if (AtomicCmpAndSwap((UINT*)pEntryIndexAddr, LogIndexMarker, ~0)) {
	break;
      }
    }
    EntryIndex = LogIndexMarker;
    BufferSize = plThis->BufferSize;
    pleBuffer = plThis->pleBuffer;

    if (LogIndexMarker == 0) {
      StopIndex = BufferSize - 1;
    }
    else {
      StopIndex = LogIndexMarker - 1;
    }

    /* Initialize time marker to 0, because we are only interested in
     * the relative time of events in the scheduler log */
    Int32ToInt64(tMarker, 0);

    _tprintf(_TEXT("===LOG OUTPUT===\n"));

    do {
      EntryIndex++;
      if (EntryIndex >= BufferSize) {
	EntryIndex = 0;
      }

      pleCurrent = &pleBuffer[EntryIndex];
      pletType = pleCurrent->pletType;
      if (pletType == NULL) {
	continue;
      }

      tEntryTime = pleCurrent->tTime;

      _tprintf(_TEXT("%s %08x  %d (+%d)\n"),
	       pletType->szName,
	       pleCurrent->Data,
	       Int64ToInt32(tEntryTime),
	       Int64ToInt32(TimeSubtract(tEntryTime, tMarker)));

      tMarker = tEntryTime;
      
    } while (EntryIndex != StopIndex);

/*     _tprintf(_TEXT(" now  %ld (+%ld)\n"), */
/* 	     Now, Int64ToInt32(TimeSubtract(Now, tMarker))); */

    /* Let the log go */
    plThis->EntryIndex = LogIndexMarker;

}
