/* tree-avl.c
 * Implementation of an AVL tree.
 * References:
 *   Introduction to Algorithms, Cormen, Leiserson, Rivest. MIT Press
 *   http://www.vbwm.com/art_2001/avltree08/
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <util/tree-avl.h>
#include <diagnostics.h>

/******************************************************************************
 * PRIVATE MACROS
 */

#define /* PATNODE */                                                         \
AVLTree_GetSentinel(/* PAVLTREE */ _pat_)                                     \
    &(pAT(_pat_)->atnSentinel)


/******************************************************************************
 * PRIVATE FUNCTION PROTOTYPES
 */

PRIVATE TREE_RESULT_ENUM
AVLNode_Insert(
    PPATNODE ppParent,
    CompareFunc cfSort,
    PATNODE pNew
    );

PRIVATE TREE_RESULT_ENUM
AVLNode_RemoveHelper(
    PPATNODE ppParent,
    CompareFunc cfSort,
    PATNODE pRemove
    );

/******************************************************************************
 * PRIVATE FUNCTION DEFINITIONS
 */

/* Recursive helper to insert nodes into the left subtree with AVLNode_Insert
 * Handles CASES B1 or B3
 * Rebalances with rotations and updates balance factors accordingly.
 */
PRIVATE INLINE TREE_RESULT_ENUM
AVLNode_LeftInsert(
    PPATNODE ppParent,
    CompareFunc cfSort,
    PATNODE pNew
    )
{
  TREE_RESULT_ENUM resultInsert;
  PATNODE pParent = *ppParent;
  PPATNODE ppLeft = ATNode_GetLeftPointer(pParent);

  resultInsert = AVLNode_Insert(ppLeft, cfSort, pNew);
  switch(resultInsert) {
  case HEIGHT_GROWN: {
    /* if left subtree height has grown, adjust parent balance factor */
    ATNode_AdjustLeft(pParent);
    switch(ATNode_GetBalance(pParent)) {
    case LEFT_HEAVY:
      /* if parent is now left heavy, it was previously balanced */
      return HEIGHT_GROWN;
    case DOUBLE_LEFT_HEAVY: {
      PATNODE pLeft = ATNode_GetLeft(pParent);
      /* if parent is now double-left heavy, we need to rebalance */
      if (ATNode_GetBalance(pLeft) == RIGHT_HEAVY) {
	/* Reduce CASE B3 to CASE B1 */
	ATNode_RotateLeft(ppLeft, TRUE);
      }
      ATNode_RotateRight(ppParent, FALSE);
      return HEIGHT_UNCHANGED;
    }
    default:
      ;
    }
  }
  default:
    return HEIGHT_UNCHANGED;
  }
}

/* Recursive helper to insert nodes into the right subtree with AVLNode_Insert
 * Handles CASES B2 or B4
 * Rebalances with rotations and updates balance factors accordingly.
 */
PRIVATE INLINE TREE_RESULT_ENUM
AVLNode_RightInsert(
    PPATNODE ppParent,
    CompareFunc cfSort,
    PATNODE pNew
    )
{
  TREE_RESULT_ENUM resultInsert;
  PATNODE pParent = *ppParent;
  PPATNODE ppRight = ATNode_GetRightPointer(pParent);

  resultInsert = AVLNode_Insert(ppRight, cfSort, pNew);
  switch(resultInsert) {
  case HEIGHT_GROWN: {
    /* if right subtree height has grown, adjust parent balance factor */
    ATNode_AdjustRight(pParent);
    switch(ATNode_GetBalance(pParent)) {
    case RIGHT_HEAVY:
      /* if parent is now right heavy, it was previously balanced */
      return HEIGHT_GROWN;
    case DOUBLE_RIGHT_HEAVY: {
      /* if parent is now double-right heavy, we need to rebalance */
      PATNODE pRight = ATNode_GetRight(pParent);
      if (ATNode_GetBalance(pRight) == LEFT_HEAVY) {
	/* Reduce CASES B4 to B2 */
	ATNode_RotateRight(ppRight, TRUE);
      }
      ATNode_RotateLeft(ppParent, FALSE);
      return HEIGHT_UNCHANGED;
    }
    default:
      ;
    }
  }
  default:
    return HEIGHT_UNCHANGED;
  }
}

/* Inserts a node into a sorted AVL tree rooted at ppParent,
 * preserving the order of cfSort. All nodes must have been inserted with the
 * same cfSort and must be retrieved or removed using the same cfSort.
 * For an AVL tree, if the insertion results in a double-heavy balance while
 * unwinding the recursive call stack, we must rebalance.
 * There are 4 balancing cases that are the same for insertion and removal:
 * CASE B1: Parent is double-left heavy and left child is left-heavy
 *          Rebalance by right-rotating about the parent. Afterwards, both
 *          parent and left child are balanced.
 * CASE B2: Parent is double-right heavy and right child is right-heavy
 *          Rebalance by left-rotating about the parent. Afterwards, both
 *          parent and right child are balanced.
 * CASE B3: Parent is double-left heavy and left child is right-heavy
 *          Rebalance by left-rotating the left child, reducing to CASE B1.
 * CASE B4: Parent is double-right heavy and right child is left-heavy
 *          Rebalance by right-rotating the right child, reducing to CASE B2.
 */
PRIVATE TREE_RESULT_ENUM
AVLNode_Insert(
    PPATNODE ppParent,
    CompareFunc cfSort,
    PATNODE pNew
    )
{
  PATNODE pParent;
  PTR pNewData;
  int compareResult;

  assert(ppParent != NULL);
  assert(pNew != NULL);
  assert(cfSort != NULL);
  assert(ATNode_IsLeaf(pNew));

  pParent = *ppParent;

  if (pParent == NULL) {
    /* special case: we are inserting into an empty list */
    *ppParent = pNew;
    return SUCCESS;
  }

  pNewData = TNode_GetData(pNew);
  compareResult = cfSort(pNewData, TNode_GetData(pParent));
  
  if (compareResult < 0) {
    /* insert to the left */
    if (TNode_HasLeft(pParent)) {
      /* Recursive case: insert and balance to the left. CASES B1 or B3 */
      return AVLNode_LeftInsert(ppParent, cfSort, pNew);
    }
    else {
      /* Base case: add a new left leaf */
      ATNode_SetLeft(pParent, pNew);
      ATNode_AdjustLeft(pParent);
      if (TNode_HasRight(pParent)) {
	return HEIGHT_UNCHANGED;
      }
      return HEIGHT_GROWN;
    }
  }
  else if (compareResult > 0) {
    /* insert to the right */
    if (TNode_HasRight(pParent)) {
      /* Recursive case: insert and balance to the right CASES B2 or B4 */
      return AVLNode_RightInsert(ppParent, cfSort, pNew);
    }
    else {
      /* Base case: add a new right leaf */
      ATNode_SetRight(pParent, pNew);
      ATNode_AdjustRight(pParent);
      if (TNode_HasLeft(pParent)) {
	return HEIGHT_UNCHANGED;
      }
      return HEIGHT_GROWN;
    }
  }

  return DUPLICATE_KEY;
}

/* Recursive helper to remove nodes from the left subtree with AVLNode_Remove
 * Handles CASES B2 or B4
 * Rebalances with rotations and updates balance factors accordingly.
 */
PRIVATE INLINE TREE_RESULT_ENUM
AVLNode_RemoveLeft(
    PPATNODE ppParent,
    CompareFunc cfSort,
    PATNODE pRemove
    )
{
  TREE_RESULT_ENUM resultRemove;
  PATNODE pParent = *ppParent;
  PPATNODE ppLeft = ATNode_GetLeftPointer(pParent);

  DBG(("RemoveLeft %x \n", pParent));

  resultRemove = AVLNode_RemoveHelper(ppLeft, cfSort, pRemove);
  switch(resultRemove) {
  case KEY_NOT_FOUND: return KEY_NOT_FOUND;
  case HEIGHT_SHRUNK: {
    /* if left subtree height has shrunk, adjust parent balance factor */
    ATNode_AdjustRight(pParent);
    switch(ATNode_GetBalance(pParent)) {
    case RIGHT_HEAVY:
      /* if parent is now right heavy, it was previously balanced */
      return HEIGHT_UNCHANGED;
    case DOUBLE_RIGHT_HEAVY: {
      PPATNODE ppRight = ATNode_GetRightPointer(pParent);
      /* if parent is now double-right heavy, we need to rebalance */
      if (ATNode_GetBalance(*ppRight) == LEFT_HEAVY) {
	/* Reduce CASES B4 to B2 */
	ATNode_RotateRight(ppRight, TRUE);
      }
      ATNode_RotateLeft(ppParent, FALSE);
      return HEIGHT_SHRUNK;
    }
    default:
      /* if parent is now single-left-heavy or balanced, fall through  */  ;
    }
  }
  default:
    /* if left subtree height hasn't changed, do nothing */
    return HEIGHT_UNCHANGED;
  }
}

/* Recursive helper to remove nodes from the right subtree with AVLNode_Remove
 * Handles CASES B1 or B3
 * Rebalances with rotations and updates balance factors accordingly.
 */
PRIVATE INLINE TREE_RESULT_ENUM
AVLNode_RemoveRight(
    PPATNODE ppParent,
    CompareFunc cfSort,
    PATNODE pRemove
    )
{
  TREE_RESULT_ENUM resultRemove;
  PATNODE pParent = *ppParent;
  PPATNODE ppRight = ATNode_GetRightPointer(pParent);

  DBG(("RemoveRight %x \n", pParent));

  resultRemove = AVLNode_RemoveHelper(ppRight, cfSort, pRemove);
  switch(resultRemove) {
  case KEY_NOT_FOUND: return KEY_NOT_FOUND;
  case HEIGHT_SHRUNK: {
    /* if right subtree height has shrunk, adjust parent balance factor */
    ATNode_AdjustLeft(pParent);
    switch(ATNode_GetBalance(pParent)) {
    case LEFT_HEAVY:
      /* if parent is now left heavy, it was previously balanced */
      return HEIGHT_UNCHANGED;
    case DOUBLE_LEFT_HEAVY: {
      /* if parent is now double-left heavy, we need to rebalance */
      PPATNODE ppLeft = ATNode_GetLeftPointer(pParent);
      if (ATNode_GetBalance(*ppLeft) == RIGHT_HEAVY) {
	/* Reduce CASE B3 to B1 */
	ATNode_RotateLeft(ppLeft, TRUE);
      }
      ATNode_RotateRight(ppParent, FALSE);
      return HEIGHT_SHRUNK;
    }
    default:
      /* if parent is now single-right-heavy or balanced, fall through */ ;
    }
  }
  default:
    /* if left subtree height hasn't changed, do nothing */
    return HEIGHT_UNCHANGED;
  }
}


/* Removes a node from a sorted AVL tree rooted at ppParent, preserving the
 * total ordering of cfSort, which must be the same used during insertion.
 * For an AVL tree, if deletion results in a double-heavy balance while
 * unwinding the recursive call stack, we must rebalance.
 * Complement of AVL_Insert, in that subtrees are shrunk instead of grown.
 * The 4 cases are the same as for AVL_Insert, but now LeftRemove handles
 * the right-heavy cases and RightRemove handles the left-heavy cases.
 */  
PRIVATE TREE_RESULT_ENUM
AVLNode_RemoveHelper(
    PPATNODE ppParent,  /* Root of subtree to remove from; must be non-null */
    CompareFunc cfSort, /* Sorting function that imposes total ordering     */
    PATNODE pRemove     /* Node to remove; must be non-null                 */
    )
{
  PATNODE pParent;
  PTR pRemoveData;
  int compareResult;

  pParent = *ppParent;

  pRemoveData = TNode_GetData(pRemove);
  compareResult = cfSort(pRemoveData, TNode_GetData(pParent));

  DBG(("RemoveHelper %x %d\n", pParent, compareResult));
  
  if (compareResult < 0) {
    /* insert to the left */
    if (TNode_HasLeft(pParent)) {
      /* Recursive case: remove and balance to the left. */
      return AVLNode_RemoveLeft(ppParent, cfSort, pRemove);
    }
    else {
      return KEY_NOT_FOUND;
    }
  }
  else if (compareResult > 0) {
    /* insert to the right */
    if (TNode_HasRight(pParent)) {
      /* Recursive case: remove and balance to the right */
      return AVLNode_RemoveRight(ppParent, cfSort, pRemove);
    }
    else {
      return KEY_NOT_FOUND;
    }
  }
  /* At this point, (pParent == pRemove) and is the key to remove. */
  DBG(("RemoveHelper removing %x %x %x\n", pRemove, ATNode_GetLeft(pRemove),
	 ATNode_GetRight(pRemove)));

  if (ATNode_IsSentinel(pRemove)) {
    /* we have found the replace node to splice out */
    PATNODE pLeft;
    PATNODE pRight;
    PATNODE pReplaceChild;
    PPATNODE ppRemove;
    /*       PATNODE pRemoveOriginal; */
    /* Recover pSentinel and restore right child (may be null) */

    pLeft = ATNode_GetLeft(pRemove);
    pRight = ATNode_GetRight(pRemove);

    if (pLeft != pRemove) {
      /* pLeft is our sentinel value */
      pReplaceChild = ATNode_GetLeft(pLeft);
      ATNode_SetRight(pRemove, pReplaceChild);
      ppRemove = ATNode_GetData(pLeft);
    }
    else {
      /* pRight is our sentinel value */
      assert(pRight != pRemove);
      pReplaceChild = ATNode_GetLeft(pRight);
      ATNode_SetLeft(pRemove, pReplaceChild);
      ppRemove = ATNode_GetData(pRight);
    }
    
    {
      PATNODE pRemoveOriginal;
      PATNODE pReplaceOriginal;
      PPATNODE ppReplaceLeft;
      PPATNODE ppReplaceRight;
      /* Resample the original pRemove before we swap */
      pRemoveOriginal = *ppRemove;
      pReplaceOriginal = pRemove;
      ppReplaceLeft = ATNode_GetLeftPointer(pRemove);
      ppReplaceRight = ATNode_GetRightPointer(pRemove);
    
      /*       pRemoveOriginal = *ppRemove; */
      ATNode_Swap(ppRemove, ppParent);
    
      if (ATNode_GetLeft(pRemoveOriginal) == pReplaceOriginal) {
	*ppReplaceLeft = pReplaceChild;
/* 	ATNode_SetLeft(pReplaceOriginal, pReplaceChild); */
      }
      else if (ATNode_GetRight(pRemoveOriginal) == pReplaceOriginal) {
	*ppReplaceRight = pReplaceChild;
/* 	ATNode_SetRight(pReplaceOriginal, pReplaceChild); */
      }
      else {
	*ppParent = pReplaceChild;
      }

      ATNode_SetLeft(pRemoveOriginal, NULL);
      ATNode_SetRight(pRemoveOriginal, NULL);

    }

    return HEIGHT_SHRUNK;
  }
  else {
    PATNODE pLeft = ATNode_GetLeft(pRemove);
/*     PATNODE pRight = ATNode_GetRight(pRemove); */
    PATNODE pReplace;
    PATNODE pSentinel;

    assert(ATNode_IsSentinel(pLeft));
    /* we have reached the original pRemove and must swap it out */
    pSentinel = pLeft;
    
    /* Restore pRemove's left child from pSentinel */
    ATNode_SetLeft(pRemove, ATNode_GetData(pSentinel));
    
    if (ATNode_IsLeaf(pRemove)) {
      /* if removing a node, don't need to splice out a replacement */
      ATNode_SetLeft(pRemove, NULL);
      *ppParent = NULL;
      return HEIGHT_SHRUNK;
    }
    else if (ATNode_GetBalance(pRemove) == LEFT_HEAVY) {
      /* swapping with predecessor will shift a left-heavy tree to the right */
      pReplace = ATNode_ChildPredecessor(pRemove);
      /* we are now looking for the predecessor child */
      /* 	pRemoveSave = ATNode_GetLeft(pReplace); */
      /* predecessor doesn't have a right child, use it to hold pSentinel */
      ATNode_SetRight(pReplace, pSentinel);
      ATNode_SetLeft(pSentinel, ATNode_GetLeft(pReplace));
      ATNode_SetLeft(pReplace, pReplace);
    }
    else {
      /* swapping with successor will shift a right-heavy tree to the right */
      pReplace = ATNode_ChildSuccessor(pRemove);
      /* we are now looking for the successor child */
      /* 	pRemoveSave = ATNode_GetRight(pReplace); */
      /* successor doesn't have left child, use it to hold pSentinel */
      ATNode_SetLeft(pReplace, pSentinel); 
      ATNode_SetLeft(pSentinel, ATNode_GetRight(pReplace));
      ATNode_SetRight(pReplace, pReplace);
    }
    
    assert(pReplace != NULL);
    
    /* Save the parent pointer */
    ATNode_SetData(pSentinel, ppParent);
    
    /* Now go look for the replacement */
    return AVLNode_RemoveHelper(ppParent, cfSort, pReplace);
  }
  
}


/******************************************************************************
 * PUBLIC FUNCTION DEFINITIONS
 */

TREE_RESULT_ENUM
AVLTree_Insert(
    PAVLTREE pThis,
    PATNODE pNew
    )
{
  return AVLNode_Insert(&AVLTree_GetRoot(pThis),
			AVLTree_GetSortFunc(pThis), pNew);
}

TREE_RESULT_ENUM
AVLTree_Remove(
    PAVLTREE pThis,
    PATNODE pRemove
    )
{
  assert(pThis != NULL);
  assert(pRemove != NULL);

  /* Only go through this pSentinel crap if pRemove is not a leaf */
/*   if (!ATNode_IsLeaf(pRemove)) { */
  {
    PATNODE pSentinel = AVLTree_GetSentinel(pThis);
    assert(ATNode_IsSentinel(pSentinel));

    /* Copy pRemove's left child to pSentinel for safekeeping */
    ATNode_SetData(pSentinel, ATNode_GetLeft(pRemove));
    /*   ATNode_SetRight(pSentinel, ATNode_GetRight(pRemove)); */
    
    /* Point pRemove's left child to pSentinel to mark it for RemoveHelper */
    ATNode_SetLeft(pRemove, pSentinel);
    /*   ATNode_SetRight(pRemove, pSentinel); */
/*   } */
  } 
  /* Do the actual recursive removal.
   */
  if (AVLNode_RemoveHelper(&AVLTree_GetRoot(pThis), AVLTree_GetSortFunc(pThis),
			   pRemove) != KEY_NOT_FOUND) {
    return SUCCESS;
  }
  return KEY_NOT_FOUND;
}

