/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <util/tree.h>
#include <assert.h>

/* Rotate the subtree to the left
 */
void
TNode_RotateLeft(
    PPTNODE ppRoot /* Node to rotate about; must not be null. */
    )
{
  PTNODE pOldRoot;
  PTNODE pOldRight;

  assert(ppRoot != NULL);

  pOldRoot = *ppRoot;
  pOldRight = TNode_GetRight(pOldRoot);
  TNode_SetRight(pOldRoot, TNode_GetLeft(pOldRight));
  TNode_SetLeft(pOldRight, pOldRoot);
  *ppRoot = pOldRight;
}

/* Rotate the subtree to the right.
 */
void
TNode_RotateRight(
    PPTNODE ppRoot /* Node to rotate about; must not be null. */
    )
{
  PTNODE pOldRoot;
  PTNODE pOldLeft;

  assert(ppRoot != NULL);

  pOldRoot = *ppRoot;
  pOldLeft = TNode_GetLeft(pOldRoot);
  TNode_SetLeft(pOldRoot, TNode_GetRight(pOldLeft));
  TNode_SetRight(pOldLeft, pOldRoot);
  *ppRoot = pOldLeft;
}

/* Searchs the subtree for the given key value and returns the pointer to
 * the corresponding node, or NULL if it cannot be found.
 */
PTNODE
TNode_Search(
    PTNODE pRoot,
    CompareFunc cfSort,
    PTR pKey
    )
{
  PTNODE pCurrent = pRoot;
  
  while (pCurrent != NULL) {
    int cmp = cfSort(TNode_GetData(pCurrent), pKey );

    if (cmp == 0) {
      break;
    }
    
    if (cmp > 0) {
      pCurrent = TNode_GetLeft(pCurrent);
    }
    else {
      pCurrent = TNode_GetRight(pCurrent);
    }
  }
  
  return pCurrent;
}

/* Returns a pointer to the child successor of the root,
 * which must not be a leaf. */
PTNODE
TNode_ChildSuccessor(
    PTNODE pRoot     /* Pointer to subtree root to search for successor */
    )
{
  PTNODE pRight;
  PTNODE pPrev;
  PTNODE pPrevPrev;
/*   PTNODE pRoot; */

/*   assert(ppRoot != NULL); */

/*   pRoot = *ppRoot; */

  assert(pRoot != NULL);
  assert(!TNode_IsLeaf(pRoot))

  pPrevPrev = NULL;
  pPrev = pRoot;
  pRight = TNode_GetRight(pRoot);

  while (pRight != NULL) {
    pPrevPrev = pPrev;
    pPrev = pRight;
    pRight = TNode_GetLeft(pRight);
  }

  if (pPrevPrev == pRoot) {
    /* successor is right child */
    return TNode_GetRight(pRoot);
  }

  return TNode_GetLeft(pPrevPrev);
}  

/* Returns a pointer to the child predecessor of the root,
 * which must not be a leaf. */
PTNODE
TNode_ChildPredecessor(
    PTNODE pRoot     /* Root of subtree in which to search for a successor*/
    )
{
/*   PTNODE pRoot; */
  PTNODE pLeft;
  PTNODE pPrev;
  PTNODE pPrevPrev;

/*   assert(ppRoot != NULL); */

/*   pRoot = *ppRoot; */
  assert(pRoot != NULL);
  assert(!TNode_IsLeaf(pRoot))

  pPrevPrev = NULL;
  pPrev = pRoot;
  pLeft = TNode_GetLeft(pRoot);

  while (pLeft != NULL) {
    pPrevPrev = pPrev;
    pPrev = pLeft;
    pLeft = TNode_GetRight(pLeft);
  }

  if (pPrevPrev == pRoot) {
    /* predecessor is left child */
    return TNode_GetLeft(pRoot);
  }

  return TNode_GetRight(pPrevPrev);
}  

/* Swaps the two nodes indirectly referred to below.
 * There are 5 cases:
 * CASE S1: *pptn2 is the direct-LEFT-child of *pptn1
 * CASE S2: *pptn2 is the direct-RIGHT-child of *pptn1
 * CASE S3: *pptn1 is the direct-LEFT-child of *pptn2
 * CASE S4: *pptn1 is the direct-RIGHT-child of *pptn2..
 * CASE S5: *pptn1 and *pptn2 are NOT direct children of each other.
 */
void
TNode_Swap(
    PPTNODE pptn1, /* pointer to first node to swap, must be non-null  */
    PPTNODE pptn2  /* pointer to second node to swap, must be non-null */
    )
{
  PTNODE ptn1;
  PTNODE ptn2;
  PTNODE ptn1Left;
  PTNODE ptn1Right;
  PTNODE ptn2Left;
  PTNODE ptn2Right;

  assert(pptn1 != NULL);
  assert(pptn2 != NULL);

  ptn1 = *pptn1;
  assert(ptn1 != NULL);
  ptn1Left = TNode_GetLeft(ptn1);
  ptn1Right = TNode_GetRight(ptn1);

  ptn2 = *pptn2;
  assert(ptn2 != NULL);
  ptn2Left = TNode_GetLeft(ptn2);
  ptn2Right = TNode_GetRight(ptn2);
  
  if (ptn1Left == ptn2) {
    /* CASE S1: ptn2 direct-left-child-of ptn1 */
    ptn1Left = ptn1; /* Otherwise ptn2's left would be set to itself below.
		      * Also takes care of setting *pptn2 b/c == ptn1Left */
    *pptn1 = ptn2;   /* set ptn1's parent pointer to ptn2 now */
  }
  else if (ptn1Right == ptn2) {
    /* CASE S2: ptn2 direct-right-child-of ptn1 */
    ptn1Right = ptn1; /* Otherwise ptn2's right would be set to itself below.
		       * Also takes care of setting *pptn2 b/c == ptn1Right */
    *pptn1 = ptn2;   /* set ptn1's parent pointer to ptn2 now */
  }
  else if (ptn2Left == ptn1) {
    /* CASE S3: ptn1 direct-left-child-of ptn2 */
    ptn2Left = ptn2; /* Otherwise ptn2's left would be set to itself below.
		      * Also takes care of setting *pptn2 b/c == ptn1Left */
    *pptn2 = ptn1;   /* set ptn1's parent pointer to ptn2 now */
  }
  else if (ptn2Right == ptn1) {
    /* CASE S4: ptn1 direct-right-child-of ptn2 */
    ptn2Right = ptn2; /* Otherwise ptn2's right would be set to itself below.
		       * Also takes care of setting *pptn2 b/c == ptn1Right */
    *pptn2 = ptn1;   /* set ptn1's parent pointer to ptn2 now */
  }
  else {
    /* CASE S5: no direct parent-child relationships */
    *pptn1 = ptn2;
    *pptn2 = ptn1;
  }

  TNode_SetLeft(ptn1, ptn2Left);
  TNode_SetRight(ptn1, ptn2Right);
  TNode_SetLeft(ptn2, ptn1Left);
  TNode_SetRight(ptn2, ptn1Right);
}
  
