/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <string.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <netif.h>
#include <net/if.h>
#include <winsock.h>
#include <web/proxy.h>
#include <web/udpsoapserv.h>
#include <arm/at91m63200.h> /* EB63 specific */
#include "calc.h"

#define TIME_MEASURE 0

#if !_MINIMIZE
void MorseString(char* s);

/* MEMORY
 * memory
 */
SCODE doMemory(BOOL DumpAll)
{
    struct MACHINE_INFO mi;
    SCODE sc;
    PIPROCESS pPrc = CurrentProcess();
    PIHEAP pHeap = CurrentHeap();
    ADDR_SIZE Reserve,Commit,Used,MaxUsed;

    /* See how much memory we have
     */
    sc = pPrc->v->MachineInfo( pPrc, MACHINE_KIND_GENERIC, &mi );
    if (FAILED(sc)) {
        return sc;
    }

    /* The heap now
     */
    sc = pHeap->v->Status(pHeap, &Reserve, &Commit, &Used, &MaxUsed);
    if (FAILED(sc)) {
        return sc;
    }

    _tprintf(_T("   Total     Free Reserved   Commit     Used  MaxUsed\r\n"));
    _tprintf(_T("%8u %8u %8u %8u %8u %8u\n\r"),
             mi.TotalPhysicalMemory, mi.AvailablePhysicalMemory,
             Reserve, Commit, Used, MaxUsed);

    /* If we are using a debug heap, make extra arg dump all allocations */
    if (DumpAll)
        pHeap->v->Status(pHeap, NULL, NULL, NULL, NULL);

    return (SCODE) Used;
}
#endif

const struct ISampleCalculatorVtbl SAMPLECalculatorVtbl;

/* Read a Self struct from a file */
SCODE ReadSelfInfo(/*in*/const char *FileName, /*out*/PSELFINFO Self)
{
    PIFILE file = NULL;
    UINT32 size, rsize;
    UINT64 pos = UINT64_ZERO;
    SCODE sc = S_OK;
    PIHEAP Heap = CurrentHeap();
    
    //DBG(("ReadSelfInfo from %s...\n",FileName));
    DBG(("Read key pair and self certificate from cache ..."));
    sc = BindToObjectA(CurrentNameSpace(),
                    FileName,
                    NAME_SPACE_READ,
                    &IID_IFile,
                    (void **)&file);
    if (FAILED(sc)) {
        DBG(("Binding to %s failed! sc=x%x\n", FileName, sc));
        return sc;
    }
    
    /* Read the certificate and signed certificate */
    sc = file->v->ReadAt(file, pos, (PBYTE)Self, sizeof(CERTIFICATE) + CERTSIZE, &size);
    if (FAILED(sc) || size != sizeof(CERTIFICATE) + CERTSIZE) {
        DBG(("Reading %s failed! sc=x%x size=%d\n", FileName, sc, size));
        sc = E_FAIL;
        goto Out;
    }
    pos = Uint64AddUint32(pos, size);

    /* Read the PriKeySize */
    size = 0;
    sc = file->v->ReadAt(file, pos, (PBYTE)&Self->PriKeySize, sizeof(UINT), &size);
    if (FAILED(sc) || size != sizeof(UINT)) {
        DBG(("Reading %s failed! sc=x%x size=%d\n", FileName, sc, size));
        sc = E_FAIL;
        goto Out;
    }
    pos = Uint64AddUint32(pos, size);

    assert(Self->PriKeySize == 0x2bc);

    /* Alloc memory for the private key */
    Self->PriKey = (PBSAFE_KEY) Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("sec:pri"), Self->PriKeySize, 0);
    if (!Self->PriKey) {
        DBG(("malloc failed!\n"));
        sc = E_NOT_ENOUGH_MEMORY;
        goto Out;
    }

    /* Read the private key */
    size = 0x200; //0x200 is the max buffer size of ReadAt
    rsize = 0;
    do {
        sc = file->v->ReadAt(file, pos, (PBYTE)Self->PriKey + rsize, size, &size);
        if (sc != S_OK) {//S_FALSE means getting the EOF
            DBG(("Reading %s failed! sc=x%x size=%d\n", FileName, sc, size));
            sc = E_FAIL;
            goto Out;
        }
        rsize += size;
        pos = Uint64AddUint32(pos, size);
        size = Self->PriKeySize - rsize;
    } while (size > 0);
    assert(size == 0);
    
    /* Read the PubKeySize */
    size = 0;
    sc = file->v->ReadAt(file, pos, (PBYTE)&Self->PubKeySize, sizeof(UINT), &size);
    if (FAILED(sc) || size != sizeof(UINT)) {
        DBG(("Reading %s failed! sc=x%x size=%d\n", FileName, sc, size));
        sc = E_FAIL;
        goto Out;
    }
    pos = Uint64AddUint32(pos, size);

    assert(Self->PubKeySize == 0x9c);

    /* Alloc memory for the public key */
    Self->PubKey = (PBSAFE_KEY)Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("sec:pub"), Self->PubKeySize, 0);
    if (!Self->PubKey) {
        DBG(("malloc failed!\n"));
        sc =E_NOT_ENOUGH_MEMORY;
        goto Out;
    }
    
    /* Read the public key */
    size = 0;
    sc = file->v->ReadAt(file, pos, (PBYTE)Self->PubKey, Self->PubKeySize, &size);
    if (FAILED(sc) || size != Self->PubKeySize) {
        DBG(("Reading %s failed! sc=x%x size=%d\n", FileName, sc, size));
        sc = E_FAIL;
        goto Out;
    }
    pos = Uint64AddUint32(pos, size);

    /* Alloc memory for the owner's public key */
    Self->OwnerPubKey = (PBSAFE_KEY)Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("sec:owner"), Self->PubKeySize, 0);
    if (!Self->OwnerPubKey) {
        DBG(("malloc failed!\n"));
        sc =E_NOT_ENOUGH_MEMORY;
        goto Done;
    }
    
    /* Read the owner's public key */
    size = 0;
    sc = file->v->ReadAt(file, pos, (PBYTE)Self->OwnerPubKey, Self->PubKeySize, &size);
    if (FAILED(sc) || size != Self->PubKeySize) {
        DBG(("Have no owner yet(in %s)! sc=x%x size=%d\n", FileName, sc, size));
        goto Out1;
    }
    assert(*(UINT*)Self->OwnerPubKey == RSA1); // assert it is a public key
    //DBG(("Done reading %s!\n", FileName));
    DBG(("Done.\n"));
    goto Done;
Out:
    if (Self->PriKey) {
        free(Self->PriKey); Self->PriKey = NULL;
    }
    if (Self->PubKey) {
        free(Self->PubKey); Self->PubKey = NULL;
    }
Out1:
    if (Self->OwnerPubKey) {
        free(Self->OwnerPubKey); Self->OwnerPubKey = NULL;
    }
Done:
    if (file)
        file->v->Release(file);
    return sc;
}

/* Write a Self struct into a file */
SCODE WriteSelfInfo(/*in*/const char *FileName, /*out*/PSELFINFO Self)
{
    PIFILE file = NULL;
    UINT32 wsize, size = 0;
    UINT64 pos = UINT64_ZERO;
    SCODE sc = S_OK;

    //DBG(("writeSelfInfo into %s...\n",FileName));
    DBG(("Cache the key pair and self certificate ...\n"));
    sc = BindToObjectA(CurrentNameSpace(),
                    FileName,
                    NAME_SPACE_WRITE | NAME_SPACE_CREATE,
                    &IID_IFile,
                    (void **)&file);
    if (FAILED(sc)) {
        DBG(("WriteSelfInfo: Binding to %s failed! sc=x%x\n", FileName, sc));
        return sc;
    }
    
    /* clear the old stuff in the file */
    file->v->SetSize(file, UINT64_ZERO);

    /* Write the certificate and signed certificate */
    sc = file->v->WriteAt(file, pos, (PBYTE)Self, sizeof(CERTIFICATE) + CERTSIZE, &size);
    if (FAILED(sc) || size != sizeof(CERTIFICATE) + CERTSIZE) {
        DBG(("Writing %s failed! sc=x%x size=%d\n", FileName, sc, size));
        sc = E_FAIL;
        goto Out;
    }
    pos = Uint64AddUint32(pos, size);

    /* Write the PriKeySize */
    size = 0;
    sc = file->v->WriteAt(file, pos, (PBYTE)&Self->PriKeySize, sizeof(UINT), &size);
    if (FAILED(sc) || size != sizeof(UINT)) {
        DBG(("Writing %s failed! sc=x%x size=%d\n", FileName, sc, size));
        sc = E_FAIL;
        goto Out;
    }
    pos = Uint64AddUint32(pos, size);

    /* Write the private key */
    size = 0x200;
    wsize = 0;
    do {
        sc = file->v->WriteAt(file, pos, (PBYTE)Self->PriKey + wsize, size, &size);
        if (FAILED(sc)) {
            DBG(("Writing %s failed! sc=x%x size=%d\n", FileName, sc, size));
            sc = E_FAIL;
            goto Out;
        }
        wsize += size;
        pos = Uint64AddUint32(pos, size);
        size = Self->PriKeySize - wsize;
    } while (size > 0);
    assert(size == 0);

    /* Write the PubKeySize */
    size = 0;
    sc = file->v->WriteAt(file, pos, (PBYTE)&Self->PubKeySize, sizeof(UINT), &size);
    if (FAILED(sc) || size != sizeof(UINT)) {
        DBG(("Writing %s failed! sc=x%x size=%d\n", FileName, sc, size));
        sc = E_FAIL;
        goto Out;
    }
    pos = Uint64AddUint32(pos, size);

    /* Write the public key */
    size = 0;
    sc = file->v->WriteAt(file, pos, (PBYTE)Self->PubKey, Self->PubKeySize, &size);
    if (FAILED(sc) || size != Self->PubKeySize) {
        DBG(("Writing %s failed! sc=x%x size=%d\n", FileName, sc, size));
        sc = E_FAIL;
        goto Out;
    }
    pos = Uint64AddUint32(pos, size);

    /* Write the owner's public key */
    if (Self->OwnerPubKey) {
        size = 0;
        sc = file->v->WriteAt(file, pos, (PBYTE)Self->OwnerPubKey, Self->PubKeySize, &size);
        if (FAILED(sc) || size != Self->PubKeySize) {
            DBG(("Writing %s failed! sc=x%x size=%d\n", FileName, sc, size));
            sc = E_FAIL;
            goto Out;
        }
    }
    //DBG(("Done writing %s!\n",FileName));
    DBG(("Done caching the key pair and self certificate.\n"));

Out:
    if (file)
        file->v->Release(file);
    
    return sc;
}

PSAMPLECALCULATOR CreateCalc(void)
{
    PSAMPLECALCULATOR calc;
    PIHEAP Heap = CurrentHeap();
#if TIME_MEASURE
    TIME t0, t1;
#endif
    SCODE sc;
        
    /* Allocate memory for the server object */
    calc = (PSAMPLECALCULATOR) Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("calc"),
                                              sizeof(SAMPLECALCULATOR), 0);
    if (!calc) {
        DBG(("CreateCalc: Alloc failed!"));
        return NULL;
    }

    /* Initialize the calc object */
    calc->v = &SAMPLECalculatorVtbl;
    calc->Refs = 1;
    Mutex_Init(&calc->InitOwnerLock);

    calc->isSKeyValid = FALSE;

    /* Allocate memory for the Self struct */
    calc->Self = (PSELFINFO) Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("calc:self"),
                                            sizeof(SELFINFO), 0);
    if (!calc->Self) {
        DBG(("CreateCalc: Alloc failed!"));
        goto Out;
    }
    {
#ifdef i386
        char *name;
        PIFILE file;
        /* Initialize the certificate */
        sc = BindToObject(CurrentNameSpace(),
                            _T("serplex6e"),
                            NAME_SPACE_READ,
                            &IID_IFile,
                            (void **)&file);
        if (FAILED(sc)) {
#if 1
            calc->InfoFileName = "fs/TV.slf";
            name = "TV";
#else
            calc->InfoFileName = "fs/fridge.slf";
            name = "fridge";
#endif
        } else {
            /* This is ntubig2 */
            file->v->Release(file);
#if 1
            calc->InfoFileName = "fs/fridge.slf";
            name = "fridge";
#else
            calc->InfoFileName = "fs/seal.slf";
            name = "seal";
#endif
        }
#else
#define MAXDNAME 80
        char name[MAXDNAME+1], *ptr;
        PIFILE file;
        UINT len;
        sc = BindToObject(CurrentNameSpace(), _T("hostname.txt"), NAME_SPACE_READ, &IID_IFile, (void *) &file);

        if (FAILED(sc))
            sc = BindToObject(CurrentNameSpace(), _T("fs/hostname.txt"), NAME_SPACE_READ, &IID_IFile, (void *) &file);

        if (FAILED(sc)) {
            DBG(("CreateCalc:Could not bind to file hostname.txt\n"));
            goto Out;
        }

        sc= ReadWriteFile(file, FALSE, UINT64_ZERO, (PBYTE)name, MAXDNAME, &len);
        file->v->Release(file);
        if (FAILED(sc)) {
            DBG(("CreateCalc.c: Could not get default host name\n"));
            goto Out;
        }
        name[len] = '\0';
        /* Check for EOF character so that hostname can be entered into serial flash manually Control-D or Z */
#define CZ '\032'
#define CD '\004'
        if ((ptr = strchr(name, CZ)) != NULL)
            *ptr = '\0';
        if ((ptr = strchr(name, CD)) != NULL)
            *ptr = '\0';
        calc->InfoFileName = (char*)malloc(MAXDNAME+1);
        strcpy(calc->InfoFileName, "srfs/");
        strcat(calc->InfoFileName, name);
        strcat(calc->InfoFileName, ".slf");
        //printf("InforFileName=%s, name=%s\n", calc->InfoFileName, name);
#endif
        sc = ReadSelfInfo(calc->InfoFileName, calc->Self);
        if (FAILED(sc)) {
            strcpy((char*)calc->Self->Cert.ID, name);
            strcpy((char*)calc->Self->Cert.Name, name);
        }
    }

    /* Initialize the rsa cypher */
    calc->Cypher = GetCypher(_T("COB/rsa.cob"),RSAKEYSTRENGTH);
    if (calc->Cypher == NULL) {
        DBG(("Failed in creating cypher COB/rsa.cob\n"));
        goto Out;
    }

    /* Key generation takes forever and has nothing to do with 
     * the rest of the calculator code (particularly time synchro)
     * so I'm stripping it out for now - Stebbi
     */
#if 0
    /* Generate the key pair. The memory is allocated in GenKey for prikey and pubkey */
    if (calc->Self->PriKey == NULL || calc->Self->PubKey == NULL ) {
        DBG(("Generating keys ...\n"));
#if TIME_MEASURE
        t0 = CurrentTime();
#endif
        sc = calc->Cypher->v->GenKey(calc->Cypher, RSAKEYSTRENGTH, 
                                     (void*)&calc->Self->PriKey, &calc->Self->PriKeySize,
                                     (void*)&calc->Self->PubKey, &calc->Self->PubKeySize);
#if TIME_MEASURE
        t1 = CurrentTime();
#endif
        if (FAILED(sc)) {
            DBG(("GenRsaKeyPair failed\n"));
            goto Out;
        }
#if TIME_MEASURE
        t0 = Int64Subtract(t1, t0);
        t1 = Int64DividedByInt32(t0, 10000000);
        DBG(("GenRsaKeyPair took %lld 00 ns (%lld s)\n", t0, t1));
#endif    
        memcpy(&calc->Self->Cert.PubKey, calc->Self->PubKey, sizeof(PUB_KEY));

        sc = WriteSelfInfo(calc->InfoFileName, calc->Self);
        //assert(SUCCEEDED(sc));
    }
    DBG(("Done initializing keys and self certificate!\n"));
#endif
    goto Done;

Out: /* Without security stuff */
    if (calc->Self) {
        if (calc->Self->PriKey) {
            Heap->v->Free(Heap, 0, calc->Self->PriKey);
            calc->Self->PriKey = NULL;
        }
        if (calc->Self->PubKey) {
            Heap->v->Free(Heap, 0, calc->Self->PubKey);
            calc->Self->PubKey = NULL;
        }
        Heap->v->Free(Heap, 0, calc->Self);
        calc->Self = NULL;
    }
    if (calc->OwnerURL) {
        Heap->v->Free(Heap, 0, calc->OwnerURL);
        calc->OwnerURL = NULL;
    }
Done:
    DBG(("CreateCalc %x\n", (UINT) calc));
    return calc;
}

PIUNKNOWN calcCobMain(void)
{
    //TRACE("CalcCobMain");
    return (PIUNKNOWN) CreateCalc();
}

PRIVATE SCODE MCT SAMPLECalculatorQueryInterface(PISAMPLECALCULATOR Calculator, REFIID Iid, void** Unk)
{

    //TRACE("SAMPLECalculatorInterface");
    return GenericQueryInterface((PIUNKNOWN) Calculator, Iid, Unk, &IID_ISampleCalculator);
}

PRIVATE UINT MCT SAMPLECalculatorAddRef(PISAMPLECALCULATOR Calculator)
{
    UINT Refs;
    PSAMPLECALCULATOR This = (PSAMPLECALCULATOR) Calculator;

    //TRACE("SAMPLECalculatorAddRef");
    
    Refs = AtomicInc(&This->Refs);
    return (Refs);
}

PRIVATE UINT MCT SAMPLECalculatorRelease(PISAMPLECALCULATOR Calculator)
{
    UINT Refs;
    PSAMPLECALCULATOR This = (PSAMPLECALCULATOR) Calculator;
    PIHEAP Heap = CurrentHeap();

    //TRACE("SAMPLECalculatorRelease");
    
    Refs = AtomicDec(&This->Refs);

    if (Refs)
        return Refs;

    RELEASE(This->Cypher);

    /* Free memory. Self should be stored in the flash memory */
    if (This->Self) {
        if (This->Self->PriKey)
            Heap->v->Free(Heap, 0, This->Self->PriKey);
        if (This->Self->PubKey)
            Heap->v->Free(Heap, 0, This->Self->PubKey);
        if (This->Self->OwnerPubKey)
            Heap->v->Free(Heap, 0, This->Self->PubKey);
        Heap->v->Free(Heap, 0, This->Self);
    }
    if (This->OwnerURL)
        Heap->v->Free(Heap, 0, This->OwnerURL);
    Heap->v->Free(Heap, 0, This);

    return (Refs);
}

PRIVATE UINT32 MCT SAMPLECalculatorAdd(PISAMPLECALCULATOR Calculator, UINT32 argOne, UINT32 argTwo)
{
    DBG(("Coming into the ADD function with arguments %d and %d calc=%x\n", argOne, argTwo, (ADDRESS) Calculator));
    return (UINT32) (argOne + argTwo);
}

PRIVATE SCODE MCT SAMPLECalculatorAsyncTest (
    PISAMPLECALCULATOR pThis,
    const _TCHAR *foo)
{
    assert(foo);
    return S_NO_RESPONSE;
}

#if !_MINIMIZE
PRIVATE UINT32 MCT SAMPLECalculatorSubtract(PISAMPLECALCULATOR Calculator, UINT8 argOne, UINT16 argTwo)
{
    DBG(("Coming into the SUBTRACT function with arguments %d and %d \n", argOne, argTwo));
    return (UINT32) (argOne - argTwo);
}

PRIVATE UINT32 MCT SAMPLECalculatorMultiply(PISAMPLECALCULATOR Calculator, UINT32 argOne, UINT32 argTwo)
{
    DBG(("Coming into the Multiply function with arguments %d and %d with result %d\n",
        argOne, argTwo, argOne *argTwo));
    return (UINT32) (argOne * argTwo);
}

PRIVATE UINT32 MCT SAMPLECalculatorDivide(PISAMPLECALCULATOR Calculator, UINT32 argOne, UINT32 argTwo)
{
    return (UINT32) (argOne / argTwo);
}

PRIVATE UINT32 MCT SAMPLECalculatorIdentify(PISAMPLECALCULATOR Calculator, UINT32 argOne, UINT32 argTwo)
{
    SCODE sc = E_FAIL;
    PIFILE piFile = NULL;
    _TCHAR * Server_Ip_File = _T("fs/NT_Server_Ip.txt");
    PISAMPLECALCULATOR piscs = NULL;
    PINAMESPACE NameSpace = CurrentNameSpace();
    PIUNKNOWN Unk;
    char * resultSent = NULL;

    if (argOne == 999)
        return (UINT32) doMemory(argTwo);

    //DBG(("Opening file named %s for echoing\n", Server_Ip_File));
    sc = NameSpace->v->Bind(NameSpace, Server_Ip_File, NAME_SPACE_READ, &Unk);
    if (FAILED(sc)) return sc;

    //DBG(("Bound input file\n"));
    sc = Unk->v->QueryInterface(Unk, &IID_IFile, (void**) &piFile);
    Unk->v->Release(Unk);
    if (FAILED(sc)) return sc;
    //DBG(("Queried for input file interface\n"));

    sc = NameSpace->v->Bind(NameSpace,
                            _T("COB/soapclientstub.cob"),
                            0,
                            (PIUNKNOWN*) &Unk);
    if (FAILED(sc)) {
        //DBG(("calc.c: Identify: Failed to load SOAPCLIENTSTUB COB for SOAPCLIENSTUB\n"));
        goto Out;
    }

    sc = Unk->v->QueryInterface(Unk, &IID_ISampleCalculator, (void**) &piscs);
    Unk->v->Release(Unk);
    if (FAILED(sc)) {
        //DBG(("soapclientstubtest.c: main: SOAPCLIENTSTUB COB doesn't support ISoapClientStub(?)\n"));
        goto Out;
    }   

    //piscs->v->Initialize(piscs, piFile);
 
    sc = E_NOT_IMPLEMENTED;
    //sc = piscs->v->Identify(piscs, argOne, 1);

    if (FAILED(sc)) {
        //printf("Could not identify %d sc=%x\n", argOne, sc);
    } else {
        /* 1 identifies to use the NT request file */
        //resultSent = piscs->v->GetResult(piscs);

//        printf("Identified %d as %s \n", argOne, resultSent);
//        MorseString(resultSent);

        if(strcmp(resultSent, "NO") == 0 )
            sc = S_FALSE;
        else
            sc = S_OK;
    }
     
Out:
    if (piFile)
        piFile->v->Release(piFile);
    if (piscs)
        piscs->v->Release(piscs);

    return sc;
}

PRIVATE UINT32 MCT SAMPLECalculatorSetLed(PISAMPLECALCULATOR Calculator, UINT32 LedNumber, UINT32 OnOff)
{
    DBG(("Coming into the SETLED function with arguments %d and %d \n",
         LedNumber, OnOff));
#ifdef arm
    /* Silently check its in range */
    if (LedNumber < 8) {
        if (OnOff) {
            PioB->ClearData = PIOB_PB8 << LedNumber;
        } else {
            PioB->SetData = PIOB_PB8 << LedNumber;
        }
    }
    return (~PioB->DataStatus >> 8) & 0xff;
#else
{
    static UINT32 Mask = 0;
    if (OnOff)
        Mask |= 1 << LedNumber;
    else
        Mask &= ~(1 << LedNumber);
    return Mask;
}
#endif
}

PRIVATE UINT32 MCT SAMPLECalculatorStringTest (
    PISAMPLECALCULATOR pThis,
    /*out*/ _TCHAR * outstr,
    /*in*/ UINT outstrsize)
{
    _tcscpy(outstr, _T("foo-return a string!"));
    return 0;
}

PRIVATE UINT32 MCT SAMPLECalculatorOutArgTest(PISAMPLECALCULATOR Calculator, 
                                  UINT In1, 
                                  UINT *Out1, 
                                  UINT *Out2, 
                                  UINT *Out3)
{
    *Out1 = In1;
    *Out2 = In1 + 1;
    *Out3 = In1 + 2;
    return 0;
}

PRIVATE UINT SAMPLECalculatorMarshalTest (
    PISAMPLECALCULATOR pThis,
    /*in*/ const _TCHAR * instr,
    /*out*/ PUINT outint,
    /*out*/ PINT64 outint64,
    /*out*/ _TCHAR * outstr,
    /*in*/ UINT outstrsize,
    /*in*/ PTR inblob,
    /*in*/ UINT inblobsize,
    /*out*/ PTR outblob,
    /*in*/ UINT outblobsize,
    /*out*/ PUINT outbloblength,
    /*inout*/ PTR inoutblob,
    /*in*/ UINT inoutblobsize,
    /*inout*/ PUINT inoutbloblength)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR) pThis;
    UnusedParameter(x);

#if 0
    printf("SAMPLECalculatorMarshalTest: this=%x instr=%x &instr=%x outint=%x &outint=%x\n",
           (UINT)pThis,(UINT)instr,(UINT)&instr,(UINT)outint,(UINT)&outint);
    printf("SAMPLECalculatorMarshalTest: '%s' '%.*s'\n",
           instr, (int) *inoutbloblength, (char*) inoutblob);
#endif

    *outint = 987;
    *outint64 = TIME_FOREVER;
    _tcscpy(outstr, _T("zappa"));
    _tcscpy(outblob, _T("bar"));
    *outbloblength = 3;

    _tcsncpy(inoutblob, inblob, inblobsize);
    *inoutbloblength = inblobsize;
    
    return 42;
}

PRIVATE SCODE MCT SAMPLECalculatorListTest (
    PISAMPLECALCULATOR pThis,
    PListTestStruct ListArg,
    /*inout*/ PListTestStruct *ListArg2)
{
    assert(ListArg);
    return 0;
}

PRIVATE void MCT SAMPLECalculatorechoVoid (
    PISAMPLECALCULATOR pThis)
{
    //TRACE("SAMPLECalculatorechoVoid");
}

PRIVATE UINT32 MCT SAMPLECalculatorechoInteger (
    PISAMPLECALCULATOR pThis,
    UINT inputInteger)
{
    //TRACE("SAMPLECalculatorechoInteger");
    return inputInteger;
}

PRIVATE SCODE SAMPLECalculatorechoBase64 (
    /*in*/ PISAMPLECALCULATOR pThis,
    /*in*/ void* inputData,
    /*in*/ UINT inputDataSize,
    /*out*/ void* outputData,
    /*out*/ UINT outputDataSize,
    /*out*/ UINT* outputDataLength)
{
    PSAMPLECALCULATOR x = pSAMPLECalculator(pThis);

    UnusedParameter(x);

    if (inputDataSize > outputDataSize)
        inputDataSize = outputDataSize; /* truncate on overflow */
    memcpy(outputData, inputData, inputDataSize);
    *outputDataLength = inputDataSize;
    return S_OK;
}

#endif /* !_MINIMIZE */

PRIVATE void SAMPLECalculatorechoString (
    PISAMPLECALCULATOR pThis,
    /*in*/ const _TCHAR * instr,
    /*out*/ _TCHAR * outstr,
    UINT outstrsize)
{
    _tcscpy(outstr, instr);
}

#if !_MINIMIZE

PRIVATE const _TCHAR* MCT SAMPLECalculatorechoString1 (
    PISAMPLECALCULATOR pThis,
    const _TCHAR * instr)
{
    //TRACE("SAMPLECalculatorechoString");
    return instr;
}

PRIVATE void SAMPLECalculatorechoDate (
    PISAMPLECALCULATOR pThis,
    /*in*/ TIME t,
    /*out*/ TIME *Return)
{
    //TRACE("SAMPLECalculatorechoString");
    *Return = t;
}

PRIVATE SCODE SAMPLECalculatorQueryTime (
    PISAMPLECALCULATOR iThis,
    TIME time,
    TIME *current,
    TIME *offset)
{
    DBG(("SAMPLECalculatorQueryTime this=%x\n", (ADDRESS) iThis));

    *current = CurrentTime();
    //*current = Int64AddInt32(*current, TIME_SECONDS(180)); /* XXX test */
    *offset = Int64Subtract(*current, time);

    return S_OK;
}


PRIVATE void SAMPLECalculatorechoBoolean (
    /*in*/ PISAMPLECALCULATOR pThis,
    /*in*/ BOOL inputBoolean,
    /*out*/ BOOL* outputBoolean)
{
    UnusedParameter(pThis);

    *outputBoolean = inputBoolean;
}

PRIVATE void MCT SAMPLECalculatorechoIntegerArray (
    PISAMPLECALCULATOR pThis,
    /*in*/ UINT * inputIntegerArray,
    /*in*/ UINT inputIntegerArraySize,
    /*out*/ UINT * outputIntegerArray,
    /*in*/ UINT outputIntegerArraySize,
    /*out*/ PUINT outputIntegerArrayLength)
{
    UINT i;
    //TRACE("SAMPLECalculatorechoIntegerArray");
    for (i=0;i<inputIntegerArraySize;i++)
        outputIntegerArray[i]=inputIntegerArray[i];
    *outputIntegerArrayLength = inputIntegerArraySize;
}

PRIVATE void SAMPLECalculatorechoFloat (
    /*in*/ PISAMPLECALCULATOR pThis,
    /*in*/ UINT inputFloat,
    /*out*/ UINT* outputFloat)
{
    UnusedParameter(pThis);

    *outputFloat = inputFloat;
}

PRIVATE void MCT SAMPLECalculatorechoFloatArray (
    PISAMPLECALCULATOR pThis,
    /*in*/ UINT * inputIntegerArray,
    /*in*/ UINT inputIntegerArraySize,
    /*out*/ UINT * outputIntegerArray,
    /*in*/ UINT outputIntegerArraySize,
    /*out*/ PUINT outputIntegerArrayLength)
{
    UINT i;
    //TRACE("SAMPLECalculatorechoFloatArray (pseudo-float)");
    for (i=0;i<inputIntegerArraySize;i++)
        outputIntegerArray[i]=inputIntegerArray[i];
    *outputIntegerArrayLength = inputIntegerArraySize;
}

PRIVATE UINT* MCT SAMPLECalculatorechoIntegerArray1 (
    PISAMPLECALCULATOR pThis,
    /*in*/ UINT * inputIntegerArray)
{
    return inputIntegerArray;
}

PRIVATE void MCT SAMPLECalculatorechoStringArray (
    PISAMPLECALCULATOR pThis,
    /*in*/ _TCHAR** inputStringArray,
    /*in*/ UINT inputStringArraySize,
    /*out*/ _TCHAR** outputStringArray,
    /*in*/ UINT outputStringArraySize,
    /*out*/ PUINT outputStringArrayLength)
{
    UINT i;
//    TRACE("SAMPLECalculatorechoStringArray");
    for (i=0;i<inputStringArraySize;i++)
        _tcscpy(outputStringArray[i], inputStringArray[i]);
    *outputStringArrayLength = inputStringArraySize;
}

PRIVATE _TCHAR** MCT SAMPLECalculatorechoStringArray1 (
    PISAMPLECALCULATOR pThis,
    /*in*/ _TCHAR** inputStringArray)
{
//    TRACE("SAMPLECalculatorechoStringArray1");
    return inputStringArray;
}

PRIVATE void MCT SAMPLECalculatorechoStruct (
    PISAMPLECALCULATOR pThis,
    /*in*/ PSOAPStruct inputStruct,
    /*out*/ PSOAPStruct outputStruct)
{
//    TRACE("SAMPLECalculatorechoStruct");
    outputStruct->varInt = inputStruct->varInt;
    _tcscpy(outputStruct->varString, inputStruct->varString);
    outputStruct->varFloat = inputStruct->varFloat;
}

PRIVATE PSOAPStruct MCT SAMPLECalculatorechoStruct1 (
    PISAMPLECALCULATOR pThis,
    /*in*/ PSOAPStruct inputStruct)
{
    //TRACE("SAMPLECalculatorechoStruct1");
    return inputStruct;
}

PRIVATE void MCT SAMPLECalculatorechoStructArray (
    PISAMPLECALCULATOR pThis,
    /*in*/ PSOAPStruct inputStructArray,
    /*in*/ UINT inputStructArraySize,
    /*out*/ PSOAPStruct outputStructArray,
    /*in*/ UINT outputStructArraySize,
    /*out*/ PUINT outputStructArrayLength)
{
    UINT i;
    //TRACE("SAMPLECalculatorechoStructArray");

    for (i=0;i<inputStructArraySize;i++)
        memcpy(&outputStructArray[i], &inputStructArray[i], sizeof(inputStructArray[i]));
    *outputStructArrayLength = inputStructArraySize;
}

PRIVATE void SAMPLECalculatorechoDecimal (
    /*in*/ PISAMPLECALCULATOR pThis,
    /*in*/ INT64 inputDecimal,
    /*out*/ INT64* outputDecimal)
{
    UnusedParameter(pThis);

    *outputDecimal = inputDecimal;
}

PRIVATE void SAMPLECalculatorechoSimpleTypesAsStruct (
    /*in*/ PISAMPLECALCULATOR pThis,
    const _TCHAR * inputString,
    /*in*/ UINT inputInteger,
    /*in*/ UINT inputFloat,
    /*out*/ PSOAPStruct outputStruct)
{
    UnusedParameter(pThis);

    outputStruct->varString = (_TCHAR *) inputString;
    outputStruct->varInt = inputInteger;
    outputStruct->varFloat = inputFloat;
}

PRIVATE void MCT SAMPLECalculatorechoNestedArray (
    PISAMPLECALCULATOR pThis,
    /*in*/ PSOAPArrayStruct inputStruct,
    /*out*/ PSOAPArrayStruct outputStruct)
{
    //TRACE("SAMPLECalculatorechoNestedArray");
    *outputStruct = *inputStruct; /* XXX copy array separately */
}

PRIVATE void SAMPLECalculatorechoNestedStruct (
    /*in*/ PISAMPLECALCULATOR pThis,
    /*in*/ PSOAPStructStruct inputStruct,
    /*out*/ PSOAPStructStruct outputStruct)
{
    UnusedParameter(pThis);

    outputStruct->varString = inputStruct->varString;
    outputStruct->varInt = inputStruct->varInt;
    outputStruct->varFloat = inputStruct->varFloat;
    outputStruct->varStruct.varInt = inputStruct->varStruct.varInt;
    outputStruct->varStruct.varString = inputStruct->varStruct.varString;
    outputStruct->varStruct.varFloat = inputStruct->varStruct.varFloat;
}

PRIVATE void SAMPLECalculatorechoStructAsSimpleTypes (
    /*in*/ PISAMPLECALCULATOR pThis,
    /*in*/ PSOAPStruct inputStruct,
    /*out*/ _TCHAR * outputString,
    /*in*/ UINT outputStringSize,
    /*out*/ PUINT outputInteger,
    /*out*/ PUINT outputFloat)
{
    UnusedParameter(pThis);

    _tcscpy(outputString, inputStruct->varString);
    UnusedParameter(outputStringSize);

    *outputInteger = inputStruct->varInt;
    *outputFloat = inputStruct->varFloat;
}

PRIVATE void MCT SAMPLECalculatorecho2DStringArray (
    PISAMPLECALCULATOR pThis,
    /*in*/ _TCHAR** inputStringArray,
    /*in*/ UINT inputStringArraySize,
    /*out*/ _TCHAR** outputStringArray,
    /*in*/ UINT outputStringArraySize,
    /*out*/ PUINT outputStringArrayLength)
{
    UINT i;
//    TRACE("SAMPLECalculatorechoStringArray");
    for (i=0;i<inputStringArraySize;i++)
        _tcscpy(outputStringArray[i], inputStringArray[i]);
    *outputStringArrayLength = inputStringArraySize;
}
#endif

PRIVATE PISAMPLECALCULATOR GetProxy(PIHTTPCONNECTOR pConn, const _TCHAR *URL)
{
    PISOAPFACTORY soapfactory = NULL;
    PISAMPLECALCULATOR pxy = NULL;
    SCODE sc;

    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &soapfactory);
    if (FAILED(sc)) {
        DBG(("SAMPLECalculatorInit: Could not bind to soap.cob sc=%x\n", sc));
        return NULL;
    }

    sc = soapfactory->v->CreateProxy(soapfactory, pConn,
                                     URL, _T("tns"),
                                     0,
                                     (PIUNKNOWN*) &pxy);
    soapfactory->v->Release(soapfactory);
    if (FAILED(sc)) {
        DBG(("SAMPLECalculatorInit: CreateProxy %s failed sc=%x\n", URL, sc));
        return NULL;
    }
    return pxy;
}

PRIVATE UINT SAMPLECalculatorGetPubKey(PISAMPLECALCULATOR pThis,
                               /*out*/ void *PubKey)
{
    PSAMPLECALCULATOR x = pSAMPLECalculator(pThis);
    
    //TRACE("SAMPLECalculatorGetPubKey...\n");
    assert(37*4==sizeof(PUB_KEY));
    
    DBG(("Send the pub key.\n"));
    memcpy(PubKey, x->Self->PubKey, sizeof(PUB_KEY));

#if 0
    {
        UINT i, z;

        for(i=0;i<37;i++) {
            z = ((UINT32*)&x->Self->Cert.PubKey)[i];
            DBG((" x%x",z));
        }
        DBG(("\n"));
    }
#endif

    return S_OK;
}

PRIVATE UINT SAMPLECalculatorGetCertificate(PISAMPLECALCULATOR pThis,
                                   /*out*/ void *signedCert)
{
    PSAMPLECALCULATOR x = pSAMPLECalculator(pThis);
    
    TRACE("SAMPLECalculatorGetCertficate...\n");
    
    memcpy(signedCert, x->Self->SignedCert, CERTSIZE);

    return S_OK;
}

SCODE CryptABlock(/*in*/PICYPHER Cypher,
            /*in*/void *InText,
            /*in*/UINT InTextLen,
            /*in*/const BYTE *Key,
            /*in*/UINT Flag,
            /*out*/void *OutText)
{
    BYTE InBuffer[RSAKEYSIZE_1024];
    BYTE OutBuffer[RSAKEYSIZE_1024];
    UINT nBytes;
    SCODE sc = S_OK;
    
    //TRACE("Crypt...\n");

    if (InTextLen > RSAKEYSIZE_1024 - 8)
        return E_INVALID_PARAMETER;

    memset(OutBuffer,0,RSAKEYSIZE_1024);
    memcpy(InBuffer, InText, InTextLen);
    memset(InBuffer + InTextLen, 0, RSAKEYSIZE_1024 - InTextLen);

    sc = Cypher->v->SetKey(Cypher, Key, RSAKEYSTRENGTH/8, 
                           Flag ? CYPHER_ENCRYPT : CYPHER_DECRYPT);
    if (FAILED(sc)) {
        DBG(("SetKey failed sc=%x\n", sc));
        goto Out;
    }
    nBytes = 0;
    sc = Cypher->v->Cypher(Cypher,
                           Flag ? CYPHER_ENCRYPT : CYPHER_DECRYPT,
                           InBuffer,
                           RSAKEYSIZE_1024,
                           NULL,
                           OutBuffer,
                           RSAKEYSIZE_1024,
                           &nBytes);
    if (FAILED(sc)) {
        DBG(("Cypher failure (sc=%x nBytes=%d).\n", sc, nBytes));
        goto Out;
    }
    memcpy(OutText, OutBuffer, RSAKEYSIZE_1024 - 8);
    sc = S_OK;
Out:
    return sc;
}

PRIVATE UINT SAMPLECalculatorCrypt(PISAMPLECALCULATOR pThis,
                             /*in*/void *InText,
                             /*out*/void *OutText,
                             /*in*/BOOL Flag)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)pThis;
    PICYPHER Cypher = x->Cypher;
    SCODE sc = S_OK;
    
    //TRACE("SAMPLECalculatorCrypt...\n");

    sc = CryptABlock(Cypher, InText, RSAKEYSIZE_1024 - 8,
                     (Flag ? (const BYTE *)x->Self->PubKey : (const BYTE *)x->Self->PriKey),
                     Flag, OutText);
    return sc;
}

PRIVATE UINT SAMPLECalculatorOwnerAuthenticate(PISAMPLECALCULATOR pThis,
                                       /*in*/void *PubKey,
                                       /*in*/void *InText,
                                       /*out*/void *OutText)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)pThis;
    PICYPHER Cypher = x->Cypher;
    BYTE Buf[128];
    struct {
        PUB_KEY Key;
        UINT32 pad[2];
    } PK;
    SCODE sc = S_OK;
#if TIME_MEASURE
    TIME t0, t1;
#endif
    
    //TRACE("SAMPLECalculatorOwnerAuthenticate...\n");
    DBG(("Decrypt with the seal's private key..."));
#if TIME_MEASURE
    t0 = CurrentTime();
#endif
    sc = CryptABlock(Cypher, InText, RSAKEYSIZE_1024 - 8,
                     (const BYTE *)x->Self->PriKey,
                     FALSE, Buf);
#if TIME_MEASURE
    t1 = CurrentTime();
    t0 = Int64Subtract(t1, t0);
    t1 = Int64DividedByInt32(t0, 10000000);
    DBG(("took %lld 00 ns (%lld s)...Done!\n", t0, t1));
#endif

    memcpy(&PK.Key, PubKey, sizeof(PUB_KEY));
    PK.pad[0] = 0;PK.pad[1] = 0;
    DBG(("Encrypt with the device's public key..."));
#if TIME_MEASURE
    t0 = CurrentTime();
#endif
    sc = CryptABlock(Cypher, Buf, RSAKEYSIZE_1024 - 8,
                     (const BYTE *)&PK,
                     TRUE, OutText);
#if TIME_MEASURE 
    t1 = CurrentTime();
    t0 = Int64Subtract(t1, t0);
    t1 = Int64DividedByInt32(t0, 10000000);
    DBG(("took %lld 00 ns (%lld s)...Done!\n", t0, t1));
#endif
    return sc;
}

SCODE SignCertificate(PICYPHER Cypher,
                    /*in*/BYTE *cert, 
                    /*out*/BYTE *signedCert,
                    /*in*/PBSAFE_KEY PriKey)
{
    /* The datalen of BSAFE_KEY which defines the maxium bytes of the block. */
    UINT blocksize = ((UINT*)PriKey)[3];
    SCODE sc = S_OK;
#if TIME_MEASURE
    TIME t0, t1;
#endif

    TRACE("SAMPLECalculatorSignCertficate...\n");

    assert(2*(RSAKEYSIZE_1024 - 8)==CERTSIZE);
#if TIME_MEASURE
    t0 = CurrentTime();
#endif
    /* Sign the first half (126 bytes) of the certificate */
    sc = CryptABlock(Cypher, cert, blocksize,
                     (const BYTE*)PriKey,
                     FALSE, signedCert);
    if (FAILED(sc))
        goto Out;

    /* Sign the second half (128 bytes) of the certificate */
    sc = CryptABlock(Cypher, cert + blocksize, blocksize,
                     (const BYTE*)PriKey,
                     FALSE, signedCert + RSAKEYSIZE_1024 - 8);
#if TIME_MEASURE
    t1 = CurrentTime();
    t0 = Int64Subtract(t1, t0);
    t1 = Int64DividedByInt32(t0, 10000000);
    DBG(("took %lld 00 ns (%lld s)...", t0, t1));
#endif
    DBG(("SignCertificate Done!\n"));
Out:
    assert(sc==S_OK);
    return sc;
}

SCODE VerifyCertificate(PICYPHER Cypher, 
                       /*in*/BYTE *SignedCert,
                       /*out*/BYTE *UnsignedCert,
                       /*in*/PBSAFE_KEY PubKey)
{
    BYTE OutBuffer[RSAKEYSIZE_1024 - 8];
    /* The datalen of BSAFE_KEY which defines the maxium bytes of the block. */
    UINT blocksize = ((UINT*)PubKey)[3];
    SCODE sc = S_OK;
#if TIME_MEASURE
    TIME t0, t1;
#endif
    
    DBG(("Verifying the device's certificate..."));
    /* Unsign the first 128 bytes of the certificate */
    memset(OutBuffer, 0, RSAKEYSIZE_1024 - 8);
#if TIME_MEASURE
    t0 = CurrentTime();
#endif
    sc = CryptABlock(Cypher, SignedCert, RSAKEYSIZE_1024 - 8,
                     (const BYTE*)PubKey,
                     TRUE, OutBuffer);
    if (FAILED(sc))
        goto UnsignError;
    
    memcpy(UnsignedCert, OutBuffer, blocksize);

    /* The second 128 bytes of the certificate */
    memset(OutBuffer, 0, RSAKEYSIZE_1024 - 8);
    sc = CryptABlock(Cypher,
                     SignedCert + RSAKEYSIZE_1024 - 8, RSAKEYSIZE_1024 - 8,
                     (const BYTE*)PubKey, TRUE, OutBuffer);
#if TIME_MEASURE
    t1 = CurrentTime();
#endif
    if (FAILED(sc))
        goto UnsignError;
    
    memcpy(UnsignedCert + blocksize, OutBuffer, blocksize);
#if TIME_MEASURE    
    t0 = Int64Subtract(t1, t0);
    t1 = Int64DividedByInt32(t0, 10000000);
    DBG(("took %lld 00 ns (%lld s)...", t0, t1));
#endif
    DBG(("VerifyCertificate Done!\n"));
UnsignError:
    return sc;
}

PRIVATE UINT SAMPLECalculatorSignCertificate(PISAMPLECALCULATOR pThis,
                                    /*in*/void *cert, 
                                    /*out*/void *signedCert)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)pThis;
    PICYPHER Cypher = x->Cypher;
    
    TRACE("SAMPLECalculatorSignCertficate...\n");

    return SignCertificate(Cypher, cert, signedCert, x->Self->PriKey);
}

/* Generate random UINT32 */
PRIVATE UINT32 RandUint32(void)
{
    static UINT32 RandomData = 0x81fa590b;
    RandomData = RandomData*43 + Uint64ToUint32(CurrentTime()) + (RandomData >> 16);
    return RandomData;
}

/*
 * To revoke its owner, the device need authenticate the seal first. 
 */
PRIVATE void Revoke(THREAD_ARGUMENT Arg)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)Arg;
    PICYPHER Cypher = x->Cypher;
    PISAMPLECALCULATOR pxy = NULL;
    BYTE *large_rand = NULL;
    BYTE OutBuffer[RSAKEYSIZE_1024];
    BYTE back_rand[RSAKEYSIZE_1024];
    /* The datalen of BSAFE_PUB_KEY which defines the maxium bytes of the block. */
    UINT blocksize = ((UINT*)x->Self->OwnerPubKey)[3]; 
    UINT nBytes;
    SCODE sc = S_OK;

    Mutex_Lock(&x->InitOwnerLock);

    DBG(("Revoke ...\n"));
    if (x->Self->OwnerPubKey == NULL)
        goto Out;

    pxy = GetProxy(NULL, x->OwnerURL);
    if (pxy == NULL) {
        DBG(("Revoke: GetProxy %s failed\n", x->OwnerURL));
        goto Out;
    }

    large_rand = (BYTE*)malloc((blocksize/sizeof(UINT)+1)*sizeof(UINT));
    if (large_rand == NULL) {
        DBG(("malloc failed\n"));
        goto Out;
    }
    /* Generate a large real random number.*/
    for (nBytes = 0; nBytes < blocksize; nBytes += sizeof(UINT32))
        *(UINT32*)&large_rand[nBytes] = RandUint32();

    sc = CryptABlock(Cypher, large_rand, blocksize,
                     (const BYTE*)x->Self->OwnerPubKey,
                     TRUE, OutBuffer);
    if (FAILED(sc))
        goto Out;

    /* Owner authentication */
    DBG(("Authenticate the seal..."));
#if 1
    memset(back_rand, 0, RSAKEYSIZE_1024);
    sc = pxy->v->OwnerAuthenticate(pxy, (void*)x->Self->PubKey, OutBuffer, back_rand);
    if (FAILED(sc)) {
        DBG(("Revoke: OwnerAuthenticate failed 0x%x\n", sc));
        goto Out;
    }
    sc = CryptABlock(Cypher, back_rand, RSAKEYSIZE_1024 - 8,
                     (const BYTE*)x->Self->PriKey,
                     FALSE, OutBuffer);
    if (FAILED(sc))
        goto Out;
    if (memcmp(large_rand, OutBuffer, blocksize)) {
        DBG(("Revoke: You are not my owner!\n"));
        goto Out;
    }
#else
    sc = pxy->v->Crypt(pxy, OutBuffer, back_rand, FALSE);
    if (FAILED(sc)) {
        DBG(("Revoke: Crypt failed 0x%x\n", sc));
        goto Out;
    }
    
    if (memcmp(large_rand, back_rand, blocksize)) {
        DBG(("Revoke: You are not my owner!\n"));
        goto Out;
    }
#endif
    DBG(("Done.\n"));
    assert(x->Self->OwnerPubKey);
    free(x->Self->OwnerPubKey); x->Self->OwnerPubKey = NULL;
    sc = WriteSelfInfo(x->InfoFileName, x->Self);
    if (FAILED(sc)) {
        DBG(("WriteSelfInfo failed 0x%x\n", sc));
        goto Out;
    }
    DBG(("Done revoking!\n"));
Out:
    if (large_rand)
        free(large_rand);
    if (pxy)
        pxy->v->Release(pxy);
    Mutex_Unlock(&x->InitOwnerLock);
    return;
}

PRIVATE UINT SAMPLECalculatorRevoke(PISAMPLECALCULATOR pThis, const _TCHAR * SealURL)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)pThis;
    PIPROCESS prc;
    PITHREAD pthd;
    PIMODULE mod;
    SCODE sc = S_OK;

    //Mutex_Lock(&x->InitOwnerLock);
    if(x->OwnerURL != NULL)
        free(x->OwnerURL);
    x->OwnerURL = _tcsdup(SealURL);
    if (x->Self->OwnerPubKey == NULL) {
        sc = E_OBJECT_NOT_INITIALIZED; /* Have no owner yet */
        goto Out;
    }
    
    /* Create a thread to init the owner */
    sc = GetModuleContainsAddr((ADDRESS)Revoke, &mod);
    if (FAILED(sc)) goto Out;
    sc = mod->v->QueryInterface(mod, &IID_IProcess, (void **) &prc);
    mod->v->Release(mod);
    if (FAILED(sc)) goto Out;

    sc = prc->v->CreateThread(prc, CREATE_THREAD_STATIC_SCOPING,
                              (THREAD_FUNCTION)Revoke, (THREAD_ARGUMENT)x, 0,
                              NULL, &pthd);
    prc->v->Release(prc);
    if (FAILED(sc)) goto Out;
    pthd->v->Release(pthd);
Out:
    //Mutex_Unlock(&x->InitOwnerLock);
    return sc;
}

PRIVATE void InitOwner(THREAD_ARGUMENT Arg)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)Arg;
    PISAMPLECALCULATOR pxy = NULL;
    CERTIFICATE unsignedcert;
    struct {
        PUB_KEY Key;
        UINT32 pad[2];
    } PubKey;
    SCODE sc;
    
    Mutex_Lock(&x->InitOwnerLock);
    DBG(("InitOwner ...\n"));
    if (x->Self->OwnerPubKey != NULL) {
        DBG(("Already have an owner\n"));
        goto Out;
    }

    pxy = GetProxy(NULL, x->OwnerURL);
    if (pxy == NULL) {
        DBG(("SAMPLECalculatorInit: GetProxy %s failed\n", x->OwnerURL));
        goto Out;
    }

    DBG(("Getting the seal's public key ..."));
    sc = pxy->v->GetPubKey(pxy, (void*)&PubKey.Key);
    if (FAILED(sc)) {
        DBG(("GetPubKey failed sc=0x%x\n", sc));
        goto Out;
    }
    PubKey.pad[0] = 0;PubKey.pad[1] = 0;
    DBG(("InitOwner GetPubKey Done!\n"));
    
    /* Call the proxy */
    DBG(("Asking owner to sign the cerificate ...\n"));
    sc = pxy->v->SignCertificate(pxy, &x->Self->Cert, x->Self->SignedCert);
    if (FAILED(sc)) {
        DBG(("SAMPLECalculatorInit: SignCertificate failed 0x%x\n", sc));
        goto Out;
    }
    DBG(("InitOwner SignCertificate Done!\n"));

    /* verify the signed certificate */
    sc = VerifyCertificate(x->Cypher,
                           (BYTE*)x->Self->SignedCert,
                           (BYTE*)&unsignedcert,
                           (PBSAFE_KEY)&PubKey);
    if (FAILED(sc))
        goto Out;
    
    if (!memcmp(&x->Self->Cert, &unsignedcert, sizeof(CERTIFICATE))) {
        x->Self->OwnerPubKey = (PBSAFE_KEY)malloc(sizeof(PubKey));
        if (!x->Self->OwnerPubKey) {
            DBG(("SAMPLECalculatorInitOwner: malloc failed!\n"));
            goto Out;
        }

        memcpy(x->Self->OwnerPubKey, &PubKey, sizeof(PubKey));

        sc = WriteSelfInfo(x->InfoFileName, x->Self);
        //assert(SUCCEEDED(sc));
    } else {
        DBG(("Failed in verifying the signed certificate!\n"));
    }
    DBG(("Done initowner!\n"));
Out:
    if (pxy)
        pxy->v->Release(pxy);
    Mutex_Unlock(&x->InitOwnerLock);
    return;
}

PRIVATE UINT SAMPLECalculatorInitOwner(PISAMPLECALCULATOR pThis, 
                               const _TCHAR * SealURL)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)pThis;
    PIPROCESS prc;
    PITHREAD pthd;
    PIMODULE mod;
    SCODE sc = S_OK;
    
    //Mutex_Lock(&x->InitOwnerLock);
    if(x->OwnerURL != NULL)
        free(x->OwnerURL);
    x->OwnerURL = _tcsdup(SealURL);
    if (x->Self->OwnerPubKey != NULL) {
        sc = E_ALREADY_EXISTS; /* Already have a owner */
        goto Out;
    }
    
    /* Create a thread to init the owner */
    sc = GetModuleContainsAddr((ADDRESS) InitOwner, &mod);
    if (FAILED(sc)) goto Out;
    sc = mod->v->QueryInterface(mod, &IID_IProcess, (void **) &prc);
    mod->v->Release(mod);
    if (FAILED(sc)) goto Out;

    sc = prc->v->CreateThread(prc, CREATE_THREAD_STATIC_SCOPING,
                              (THREAD_FUNCTION)InitOwner,
                              (THREAD_ARGUMENT)x, 0,
                              NULL, &pthd);
    prc->v->Release(prc);
    if (FAILED(sc)) goto Out;
    pthd->v->Release(pthd);
Out:
    //Mutex_Unlock(&x->InitOwnerLock);
    return sc;
}

/*
 * Digital signiture alone do not provide authentication. It cannot prevent replay attacks.
 * But for session key exchanging, even if a replay attacker record a signed certificate of 
 * a printer and resend it later to pretend to be the printer, he cannot get the session key
 * because it is encrypted with the printer's public key. 
 */
PRIVATE UINT SAMPLECalculatorSetSessionKey(PISAMPLECALCULATOR pThis,
                                   /*in*/void* SignedClientCert,
                                   /*in*/void* SKey)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)pThis;
    BYTE OutBuffer[RSAKEYSIZE_1024 - 8], OutBuffer1[RSAKEYSIZE_1024 - 8];
    struct {
        PUB_KEY Key;
        UINT32 pad[2];
    } PartnerPubKey;
    SCODE sc;
#if TIME_MEASURE
    TIME t0, t1;
#endif    
    DBG(("Setting up session key ...\n"));
    DBG(("Authenticate the client ...\n"));
    sc = VerifyCertificate(x->Cypher, SignedClientCert, 
                           (BYTE*)&x->ClientCert, x->Self->OwnerPubKey);
    if (FAILED(sc) || x->ClientCert.PubKey.magic != RSA1) {
        DBG(("SAMPLECalculatorSetSessionKey: VerifyCertificate failed\n"));
        goto Out;
    }
    
    memcpy(&PartnerPubKey, &x->ClientCert.PubKey, sizeof(PUB_KEY));
    PartnerPubKey.pad[0] = 0;PartnerPubKey.pad[1] = 0;
#if TIME_MEASURE
    t0= CurrentTime();
#endif
    sc = CryptABlock(x->Cypher, SKey, RSAKEYSIZE_1024 - 8,
                     (const BYTE*)x->Self->PriKey,
                     FALSE, OutBuffer);
#if TIME_MEASURE
    t1 = CurrentTime();
#endif
    if (FAILED(sc))
        goto Out;
#if TIME_MEASURE    
    t0 = Int64Subtract(t1, t0);
    t1 = Int64DividedByInt32(t0, 10000000);
    DBG(("Decryting a block(128 bytes) w/ the private key took %lld 00 ns (%lld s).\n", t0, t1));
    t0 = CurrentTime();
#endif    
    sc = CryptABlock(x->Cypher, OutBuffer, RSAKEYSIZE_1024 - 8,
                     (const BYTE*)&PartnerPubKey,
                     TRUE, OutBuffer1);
#if TIME_MEASURE
    t1 = CurrentTime();
#endif
    if (FAILED(sc))
        goto Out;
#if TIME_MEASURE    
    t0 = Int64Subtract(t1, t0);
    t1 = Int64DividedByInt32(t0, 10000000);
    DBG(("Encryting a block(128 bytes) w/ the public key took %lld 00 ns (%lld s).\n", t0, t1));
#endif
    memcpy(&x->SKey, OutBuffer1, sizeof(SESSIONKEY));
    assert(x->SKey.magic == SKEY);
    x->isSKeyValid = TRUE;
    DBG(("Done client authentication!\n"));
#ifdef _DEBUG
    {
        UINT i;
        printf("Got a session key: ");
        for (i = 0; i < SKEY_STRENGTH/8; i++)
            printf(" x%x", x->SKey.Key[i]);
        printf("\n");
    }
#endif
    DBG(("Done setting up session key!\n"));
Out:
    return sc;
}

PRIVATE UINT SAMPLECalculatorClearSessionKey(PISAMPLECALCULATOR pThis)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)pThis;
    
    TRACE("SAMPLECalculatorClearSessionKey");
    x->isSKeyValid = FALSE;
    DBG(("SAMPLECalculatorClearSessionKey Done!\n"));
    return S_OK;
}

PRIVATE UINT32 GetInterfaceIp(const char *iname)
{
    PINAMESPACE ns = CurrentNameSpace();
    PIENDPOINTFACTORY ef;
    PIENDPOINT ep;
    struct ifreq ifr;
    UINT32 LocalIPAddr = INADDR_ANY;
    SCODE sc;
    
    /* Prepare address and port to listen at */
    memset((void*) &ifr, 0, sizeof(struct ifreq));

    /* Get an endpoint factory to create an endpoint */
    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void**) &ef);
    if (FAILED(sc)) goto Out;

    /* Create an endpoint to query for the interface address */
    sc = ef->v->CreateEndpoint(ef, AF_INET, SOCK_DGRAM, 0, &ep);
    ef->v->Release(ef);
    if (FAILED(sc)) goto Out;

    /* This is where we specify which interface to use */
    strncpy(ifr.ifr_name, iname, sizeof(ifr.ifr_name));
    
    /* Query for the interface address */
    sc = ep->v->Ioctl(ep, SIOCGIFADDR, (BYTE*) &ifr);
    ep->v->Release(ep);
    if (FAILED(sc)) goto Out;

    LocalIPAddr = ((struct sockaddr_in *) &(ifr.ifr_addr))->sin_addr.s_addr;
Out:
    return LocalIPAddr;
}

PRIVATE SCODE SAMPLECalculatorDiscoverService(PISAMPLECALCULATOR pThis,
                                     const _TCHAR *Service,
                                     /*out*/_TCHAR *ServiceURL,
                                     /*in*/UINT ServiceURLSize)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)pThis;
    UINT32 PubChannelIPAddr;
    _TCHAR URL[80];

    if (_tcscmp((_TCHAR*)x->Self->Cert.ID, Service)) {
        DBG(("I am %s, not %s!\n",(char*)x->Self->Cert.ID, Service));
        return S_NO_RESPONSE;
    }
    
    PubChannelIPAddr = GetInterfaceIp(PubChannel);//public channel
//printf(PubChannel);
    _tcscpy(URL, _tinet_ntoa(*(struct in_addr *)&PubChannelIPAddr));
    _tcscat(URL, _T("/COB/calc.cob"));
//printf(URL);
    if (_tcslen(URL) < ServiceURLSize) {
        _tcscpy(ServiceURL, URL);
        DBG(("Response to service discovery %s\n", URL));
        return S_OK;
    } else {
        return E_INVALID_PARAMETER;
    }
}

/* If pConn is not NULL, then use it as the ISOAPCONNECTOR
 * instead of creating a new ISOAPCONNECTOR.
 * To clear session key, just set its strength as zero.
 */
SCODE ClearSessionKey(/*in*/PIHTTPCONNECTOR pConn,
                     /*in*/const _TCHAR *PartnerURL,
                     /*out*/PSESSIONKEY SKey)
{
    PISAMPLECALCULATOR pxy = NULL;
    SCODE sc;
    
    DBG(("Clear session key ...\n"));
    pxy = GetProxy(pConn, PartnerURL);
    if (pxy == NULL) {
        DBG(("ResetSessionKey: GetProxy %s failed\n", PartnerURL));
        return E_FAIL;
    }

    sc = pxy->v->ClearSessionKey(pxy);
    
    pxy->v->Release(pxy);

    DBG(("ClearSessionKey Done!\n"));
    return sc;
}

/* If pConn is not NULL, then use it as the ISOAPCONNECTOR
 * instead of creating a new ISOAPCONNECTOR.
 */
SCODE SetSessionKey(/*in*/PISAMPLECALCULATOR calc,
                    /*in*/PIHTTPCONNECTOR pConn,
                    /*in*/const _TCHAR *PartnerURL,
                    /*in*/CERTIFICATE PartnerCert,
                    /*out*/PSESSIONKEY SKey)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)calc;
    struct {
        PUB_KEY Key;
        UINT32 pad[2];
    } PartnerPubKey;
    PISAMPLECALCULATOR pxy = NULL;
    BYTE OutBuffer[RSAKEYSIZE_1024 - 8], OutBuffer1[RSAKEYSIZE_1024 - 8];
    SCODE sc;
    UINT i;
#if TIME_MEASURE
    TIME t0, t1;
#endif
    
    DBG(("Set up session key ...\n"));
    pxy = GetProxy(pConn, PartnerURL);
    if (pxy == NULL) {
        DBG(("SetSessionKey: GetProxy %s failed\n", PartnerURL));
        return E_FAIL;
    }
    
    memcpy(&PartnerPubKey, &PartnerCert.PubKey, sizeof(PUB_KEY));
    PartnerPubKey.pad[0] = 0;PartnerPubKey.pad[1] = 0;

    /* Generate a 128-bit AES key.*/
    SKey->magic = SKEY;
    for (i = 0; i < SKEY_STRENGTH/8; i += sizeof(UINT32))
        *(UINT32*)&SKey->Key[i] = RandUint32();
    SKey->KeyStrength = SKEY_STRENGTH;
    //SKey->TimeStamp = CurrentTime();

#ifdef _DEBUG
    printf("Generate a session key: ");
    for (i = 0; i < SKEY_STRENGTH/8; i++)
        printf(" x%x", SKey->Key[i]);
    printf("\n");
#endif
#if TIME_MEASURE
    t0 = CurrentTime();
#endif
    sc = CryptABlock(x->Cypher, (BYTE*)SKey, sizeof(SESSIONKEY),
                     (const BYTE*)x->Self->PriKey,
                     FALSE, OutBuffer);
    if (FAILED(sc))
        goto Out;

    sc = CryptABlock(x->Cypher, OutBuffer, RSAKEYSIZE_1024 - 8,
                     (const BYTE*)&PartnerPubKey,
                     TRUE, OutBuffer1);
#if TIME_MEASURE
    t1 = CurrentTime();
#endif
    if (FAILED(sc))
        goto Out;
    sc = pxy->v->SetSessionKey(pxy, x->Self->SignedCert, OutBuffer1);
    assert(sc==S_OK);
#if TIME_MEASURE
    t0 = Int64Subtract(t1, t0);
    t1 = Int64DividedByInt32(t0, 10000000);
    DBG(("Seting session key took %lld 00 ns (%lld s)!\n", t0, t1));
#endif
    DBG(("Done setting up session key!\n"));
Out:
    pxy->v->Release(pxy);
    return sc;
}

/*
 * Get the signed certificate of the partner and then unsign it with 
 * the owner's public key.
 */
SCODE ParnerAuthenticate(PISAMPLECALCULATOR calc,
                             /*in*/PIHTTPCONNECTOR pConn,
                             /*in*/const _TCHAR *PartnerURL,
                             /*out*/PCERTIFICATE PartnerCert)
{
    PSAMPLECALCULATOR x = (PSAMPLECALCULATOR)calc;
    PISAMPLECALCULATOR pxy = NULL;
    BYTE SignedCert[CERTSIZE];
    SCODE sc;
    
    if (x->Self->OwnerPubKey == NULL) {
        DBG(("Have no owner yet!\n"));
        return E_NOT_READY;
    }
    pxy = GetProxy(pConn, PartnerURL);
    if (pxy == NULL) {
        DBG(("SAMPLECalculatorInit: GetProxy %s failed\n", PartnerURL));
        return E_FAIL;
    }
    sc = pxy->v->GetCertificate(pxy, SignedCert);
    if (FAILED(sc)) {
        DBG(("SAMPLECalculatorInit: GetCertificate failed\n"));
        goto Out;
    }
    
    sc = VerifyCertificate(x->Cypher, SignedCert, (BYTE*)PartnerCert, x->Self->OwnerPubKey);
    if (FAILED(sc) || PartnerCert->PubKey.magic != RSA1) {
        DBG(("SAMPLECalculatorInit: VerifyCertificate failed\n"));
        goto Out;
    }
Out:
    pxy->v->Release(pxy);
    return sc;
}
#include "s_SAMPLECalculator_v.c"
