/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Sample component that acts as a CIM resource in WS-Management.
 * - The sample is included in the WS_DEMO profile (conf/profiles.xml)
 * - The factory object (returned by CobMain) implements IProgram that allows
 * a simple command line interface and ICimResourceFactory that allows creating
 * and accessing CIM objects as resources.
 * - An enumerator object allows enumerating resources this component provides.
 * - The enumerator implements IEnum that in turn is used by the provider
 * for WS-Enumerate.
 * - The enumerator and factory interfaces are defined in conf/wsifhelp.xml
 * - The concrete classes and components are specified in cimres.xml in the
 * the current directory. <component name="cimdb"> is the one.
 */
#include "demoserviceinfotype.h"
#include <wsif.h>
#include <sampleif.h>
#include <web/cont.h>
#include <httpif.h>
#include <stdlib.h>
#include <stdio.h>

/* Make TheCimFactory "global" */
PRIVATE CimFactory *TheCimFactory = 0;
#define _MANUAL_COBMAIN

/*** Generated v-tables ***/
#include <s_CimResource_v.c>
#include <s_CimEnum_v.c>
#include <s_CimFactory_v.c>

#define rloff offsetof(struct RecordList, ListNode)

#define MEASURE 1
#define MEASUREMENT_COUNT 1000

#if MEASURE
PRIVATE BOOL doMeasure = FALSE;
#endif


/*****Temporary Functions******/


/** This is a temporary function which generates the client.*/
PRIVATE PICIMRESOURCE ReturnClient()
{
    PICIMRESOURCE root = (PICIMRESOURCE)TheCimFactory->Root;
        return CimDirectoryLookup(root, _T("EVIDENCE/GuestClient"));
}



/*** Helpers ***/


/** Struct is used for returning the result of Access Checks*/
#if __WANTS_TRUST 
typedef struct AccessTemplate
{
    UINT Access;
    _TCHAR *evidenceUrl;
}AccessTemplate;
#endif

#if __WANTS_TRUST 
/** Meta Policy for combining the result of Access Checks of in and out policies*/
PRIVATE UINT MetaPolicyAccessCheck( UINT Access_in, UINT Access_out)
{
    if(Access_in == 1 && Access_out == 1)
        return 1;
    else
        return 0;
}
#endif

#if __WANTS_TRUST 
/*Perfroms access check for a specific current request and name of top policy*/
PRIVATE SCODE PerformAccessCheckSub(CurrentRequest *cr, _TCHAR *PolicyName, AccessTemplate *at)
{
         

    CimResource *Policy;
    struct Selector policySelector;
    at->Access = 1;
    at->evidenceUrl = NULL;
    
    policySelector.Name = _T("Name");
    policySelector.Value = PolicyName;
    policySelector.Type = NULL;
    policySelector.ListNode_Selector.pNext = &policySelector;
    policySelector.ListNode_Selector.pPrev = &policySelector;
    
    if(cr->Context)
    {
        Policy = (CimResource *) CimResourceShadowLookup(cr->Context,_T("TrustScript"), &policySelector);
        if(Policy)
        {
             const PropertyClass *policyClass = Policy->PropertyClass;
             at->Access = policyClass->GetValue(policyClass, Policy->Value, cr,_T("Value"));
             if(!at->Access)
             {
                at->evidenceUrl = ((struct TrustScript *)Policy->Value)->RequiredEvidenceUrl;
                return E_ACCESS_DENIED;
             }

        }

    }
  
    return S_OK;

}
#endif

#if __WANTS_TRUST 
/* Performs the Access Check for both in and out policies*/
PRIVATE SCODE PerformAccessCheck(PICIMRESOURCE iThis, const _TCHAR *ResourceURL, UINT mode, AccessTemplate *at)
{
    METHOD_DECLARE_NOTHIS(cimdb, PerformAccessCheck);
    AccessTemplate at_in;
    AccessTemplate at_out;
    CurrentRequest cr_in;

    CurrentRequest cr_out;
    
    _TCHAR *evidenceUrl = NULL;
    int length = 0;
    
    _TCHAR *policyUrl = malloc((_tcslen(ResourceURL)+9+1) * sizeof(_TCHAR));
    CHECKSETSC(policyUrl, E_NOT_ENOUGH_MEMORY, "malloc");

    _tcscpy(policyUrl, _T("EVIDENCE/"));
    _tcscat(policyUrl, ResourceURL);
    cr_in.Context = CimDirectoryLookup(iThis, policyUrl);

    cr_in.Client = ReturnClient();
    cr_in.Mode = mode;
    at_in.Access = 1;
    at_out.Access = 1;
    at->Access = 1;

    sc = PerformAccessCheckSub(&cr_in, _T("TopPolicy_in"), &at_in);
    
#if MEASURE
    if (doMeasure) {
        TIME t0,t1,timediff;
        int i;
        t0 = CurrentTime();
       
        for(i = 0; i < MEASUREMENT_COUNT; i++) {
            sc = PerformAccessCheckSub(&cr_in, _T("TopPolicy_in"), &at_in);
        }
        t1=CurrentTime();
        timediff = Int64Subtract(t1, t0);
//#ifndef _MINIMIZE 
        printf("Time Taken for Perform Access Check: %d\n", timediff);
//#endif
    }
#endif

    cr_out.Context = cr_in.Client;
    cr_out.Client = cr_in.Context;
    cr_out.Mode = cr_in.Mode;

    sc = PerformAccessCheckSub(&cr_out, _T("TopPolicy_out"), &at_out);

    if(at_in.evidenceUrl && at_out.evidenceUrl)
    {
#if 0
        evidenceUrl = malloc ( (strlen(at_in.evidenceUrl) + strlen(at_out.evidenceUrl) + 1)*sizeof(_TCHAR));
        _tcscpy(evidenceUrl,at_in.evidenceUrl); 
        _tcscat(evidenceUrl,_T("\n"));
        _tcscat(evidenceUrl, at_out.evidenceUrl);
#else
        /* just return one of them */
        evidenceUrl = at_in.evidenceUrl; 
#endif
    } else if(at_in.evidenceUrl)
    { 
        evidenceUrl = at_in.evidenceUrl;
    } else if(at_out.evidenceUrl)
    {     
        evidenceUrl = at_out.evidenceUrl;
    }
    
    at->Access = MetaPolicyAccessCheck(at_in.Access, at_out.Access);
    at->evidenceUrl = evidenceUrl;

METHOD_CLEANUP;
    if(policyUrl)
        free(policyUrl);
    return at->Access ? S_OK : E_ACCESS_DENIED;
}
#endif


#if __WANTS_TRUST 
PRIVATE SCODE PerformAccessCheckReturn(PICIMRESOURCE iThis, const _TCHAR *ResourceURL, UINT mode, AccessTemplate *at)
{
    METHOD_DECLARE_NOTHIS(cimdb, PerformAccessCheck);
    AccessTemplate at_in;
    AccessTemplate at_out;
    CurrentRequest cr_in;

    CurrentRequest cr_out;
    
    _TCHAR *evidenceUrl = NULL;
    int length = 0;
    
    _TCHAR *policyUrl = malloc((_tcslen(ResourceURL)+9+1)*sizeof(_TCHAR));
    CHECKSETSC(policyUrl, E_NOT_ENOUGH_MEMORY, "malloc");

    _tcscpy(policyUrl, _T("EVIDENCE/"));
    _tcscat(policyUrl, ResourceURL);
    cr_in.Client = CimDirectoryLookup(iThis, policyUrl);

    cr_in.Context = ReturnClient();
    cr_in.Mode = mode;
    at_in.Access = 1;
    at_out.Access = 1;
    at->Access = 1;

    
    sc = PerformAccessCheckSub(&cr_in, _T("TopPolicy_in"), &at_in);

    cr_out.Context = cr_in.Client;
    cr_out.Client = cr_in.Context;
    cr_out.Mode = cr_in.Mode;

    sc = PerformAccessCheckSub(&cr_out, _T("TopPolicy_out"), &at_out);

    if(at_in.evidenceUrl && at_out.evidenceUrl)
    {
#if 0
        evidenceUrl = malloc ( (strlen(at_in.evidenceUrl) + strlen(at_out.evidenceUrl) + 1)*sizeof(_TCHAR));
        _tcscpy(evidenceUrl,at_in.evidenceUrl); 
        _tcscat(evidenceUrl,_T("\n"));
        _tcscat(evidenceUrl, at_out.evidenceUrl);
#else
        evidenceUrl = at_in.evidenceUrl;
#endif

    }else if(at_in.evidenceUrl)
    { 
        evidenceUrl = at_in.evidenceUrl;
    }else if(at_out.evidenceUrl)
    {     
        evidenceUrl = at_out.evidenceUrl;
    }
    
    at->Access = MetaPolicyAccessCheck(at_in.Access, at_out.Access);
    at->evidenceUrl = evidenceUrl;

METHOD_CLEANUP;
    if(policyUrl)
        free(policyUrl);
    return at->Access ? S_OK : E_ACCESS_DENIED;

}
#endif

PRIVATE INT CimEnumBump(CimEnum *pEnum, CimResource *pResource);


#ifdef _DEBUG
PRIVATE INT NodeEq(PTR p1, PTR p2)
{
    return p1 != p2;
}
#endif

/* XXX fixup EPR to fit specific use of CIM_NumericSensor.
 * XXX BETTER CONSTRUCT ON DEMAND (caps attack in the right place :)
 */
PRIVATE PEndpointReference_QUIET UpdateEpr(CimResource *This)
{
    PSelector sel = This->Epr->ReferenceProperties->Value;
    CPPropertyClass Class = This->PropertyClass;

    for ( ; Class; Class = Class->super) {
        if (Class->UpdateSelector)
            Class->UpdateSelector(Class, This->Value, sel, NULL, NULL);
    }

    return This->Epr;
}

#define selOff offsetof(struct Selector, ListNode_Selector)
#define AnyItemOff offsetof(struct AnyItem, ListNode)
#define itemsoff offsetof(struct Items, ListNode)
#define EprOff offsetof(struct EndpointReference_QUIET, ListNode)



/*** Resource ***/

PRIVATE void ResourceInit(CimResource *This, CimResource *Parent,
                          const PropertyClass *Class)
{
    PIUNKNOWN Factory = (PIUNKNOWN) TheCimFactory;

    This->super.v = CimResource_ClassDesc.iface[0].v;
    This->super.RefCnt = 1;

    ListNodeInit(croff, This);
  //  ListNodeInit(shadowoff, This);
    This->PropertyClass = Class;

    This->Parent = Parent;
    if (Parent) /* If Root then avoid circular ref */
        Factory->v->AddRef(Factory);
}

PRIVATE CimResource *ResourceNew(PIHEAP Heap, CimResource *Parent,
                                 const PropertyClass *Class, BOOL NeedsCopy)
{
    UINT Size = 0, selCount = 0;
    CimResource *This;
    PEndpointReference_QUIET epr;
    PAnyItem item;
    PSelector sel;

    if (Class) {
        const PropertyClass *c;
        for (c = Class; c; c = c->super) {
            if (c->UpdateSelector)
                c->UpdateSelector(c, NULL, NULL, &selCount, &Size);
        }
    }

    Size += sizeof(*This);
    if (NeedsCopy)
        Size += Class->DataSize;

    /* Construct EPR here so as to be done with it */
    Size += sizeof(*epr) + sizeof(*item) * selCount
                         + sizeof(*sel) * selCount;

    This = (void *) Heap->v->
        Alloc(Heap, HEAP_DEBUG_NAME("cimdb:resource") | HEAP_ZERO_MEMORY,
              Size, 0);
    if (!This)
        return NULL;

    ResourceInit(This, Parent, Class);

    This->Heap = Heap;
    Heap->v->AddRef(Heap);

    epr = (void*) (This + 1);
    item = (void *) (epr + 1);
    /* Note that UpdateSelector knows the selectors are consecutive */
    sel = (void *) (item + selCount); 

    if (NeedsCopy)
        This->Value = sel + selCount;

    ListNodeInit(EprOff, epr);
    This->Epr = epr;
    epr->Address = Parent ? Parent->Epr->Address : TheCimFactory->Epr.Address;

    for ( ; selCount; selCount--) {
        /* Make a list of items, each a selector */
        ListAddLast(&epr->ReferenceProperties, AnyItemOff, item);
        ListNodeInit(selOff, sel + selCount - 1);
        item->Value = sel + selCount - 1;
        item->ValueType = _T("wsman:Selector");
    }

    return This;
}

/* Generic copier, each subclass deals with the fields it requires.
 * Those that are not included in Src, Init instead.
 */
PRIVATE SCODE ResourceCopy(const PropertyClass *Class, PTR Dst,
                           const _TCHAR *EncodingType, const UINT32 *FieldSel,
                           const void *Src,
                           BOOL IsSrcSub, BOOL NeedsCopy, PIHEAP Heap, 
                           PICIMRESOURCE Resource)
{
    SCODE sc = S_OK;

    if (!IsSrcSub && EncodingType && Class->EncodingType_in) {
        if (!_tcscmp(Class->EncodingType_in, EncodingType))
            IsSrcSub = TRUE;
    }

    if (Class->super) {
        sc = ResourceCopy(Class->super, Dst, EncodingType, FieldSel,
                          Src, IsSrcSub, NeedsCopy, Heap,Resource);
        if (FAILED(sc))
            return sc;
    } else if (EncodingType && !IsSrcSub) {
        /* Not a subclass at all */
        return E_INVALID_METHOD; /* XXX */
    }
    if (IsSrcSub) {
        if (Class->Copy && NeedsCopy)
            sc = Class->Copy(Class, Dst, Src, Heap);

        /* Init function needs to be called everytime because of intitializing the Shadow Directory*/
        if (Class->Init)
            sc = Class->Init(Class, Dst, Heap,Resource);
    } else {
        if (Class->Init)
            sc = Class->Init(Class, Dst, Heap,Resource);
    }
    return sc;
}


PRIVATE INT ResourceDelete(PTR pItem, /*unused*/ PTR pArg) 
{
    CimResource *This = pItem;
    PIHEAP Heap = This->Heap;
    CimResource *Parent = This->Parent;
    //CimResource *ShadowParent = This->ShadowParent;
    CPPropertyClass Class = This->PropertyClass;

    if (Parent) {
        ListWalkT(CimEnum, cenoff, CimResource,
                  &Parent->Enumerators, LIST_WALK_ITERATE,
                  CimEnumBump, This);
        ListRemoveItem(&Parent->Resources, croff, This);
        /* Resources keep ref to factory, except root */
        RELEASE((PIUNKNOWN) TheCimFactory);
    }


    /* Call each finalizer */
    for ( ; Class; Class = Class->super)
        if (Class->Finalize)
            Class->Finalize(Class, This->Value, Heap);


    if (Heap) {
        Heap->v->Free(Heap, 0, This);
        Heap->v->Release(Heap);
    }
    return 0;
}


PRIVATE SCODE MCT CimResourceDestroy(PIUNKNOWN iThis)
{
    METHOD_DECLARE(CimResource, Destroy);
    CimFactory *Factory = TheCimFactory;

    ((PIUNKNOWN)Factory)->v->AddRef((PIUNKNOWN) Factory); /* XXX hmm. ensure last ref for unlock */
    Mutex_Lock(&Factory->mx);

    ResourceDelete(This, NULL);

    Mutex_Unlock(&Factory->mx);
    RELEASE((PIUNKNOWN) Factory);

    sc = S_OK;
METHOD_CLEANUP;
    return sc;
}

PRIVATE INT ResourceMatchSelector(struct Selector *sel, CimResource *Resource)
{
    const PropertyClass *Class;
    SCODE sc;

    for (Class = Resource->PropertyClass; Class; Class = Class->super) {
        if (Class->Match) {
            sc = Class->Match(Class, Resource->Value, sel);
            if (sc == E_NAME_NOT_FOUND)
                continue;
            return sc;
        }        
    }
    return E_WSMAN_INVALID_SELECTORS;
}

typedef struct ResourceMatchTemplate {
    const _TCHAR *ResourceURI;
    /*optional*/ PSelector SelectorSet;
    /*optional*/ const void *Value;
    /*optional*/ const UINT32 *FieldSel;
} ResourceMatchTemplate;

/* Compare a resource against given args. Return 0 for match (strcmp style) */
PRIVATE INT ResourceMatch(CimResource *Resource, ResourceMatchTemplate *rmt)
{
    const PropertyClass *Class;
    INT ret;
    SCODE sc = S_OK;

    if (!rmt->ResourceURI || !rmt->ResourceURI)
        return 1;

    /* First match ResourceURI, if provided */
    for (Class = Resource->PropertyClass; Class; Class = Class->super) {
        ret = _tcscmp(Class->ResourceName, rmt->ResourceURI);
        if (ret)
            ret = _tcscmp(Class->ResourceURI, rmt->ResourceURI);
        if (!ret) {
            /* Now that the URI matches, check the selectors */
            if (rmt->SelectorSet) {
                sc = (SCODE) ListWalkT(struct Selector, selOff, CimResource,
                                       &rmt->SelectorSet, LIST_WALK_SCODE,
                                       ResourceMatchSelector, Resource);
            } else {
                for (Class = Resource->PropertyClass; Class; Class = Class->super) {
                    if (Class->Match2) {
                        sc = Class->Match2(Class, Resource->Value, rmt->Value, rmt->FieldSel);
                        if (sc == E_NAME_NOT_FOUND)
                            continue;
                        return sc;
                    }        
                }
            }
            return SUCCEEDED(sc) ? 0 : 1;
        }
    }

    /* No matching ResourceURI */
    return 1;
}

/*** Enum ***/

CimEnum *CimEnumNew(PIHEAP Heap, RESOURCE_FLAGS Flags, CimResource *Owner, PDataItem Query)
{
    CimEnum *This = (void *) Heap->v->
        Alloc(Heap, HEAP_DEBUG_NAME("cimdb:enum") | HEAP_ZERO_MEMORY,
              sizeof(*This), 0);
    if (!This)
        return NULL;

    This->Owner = Owner;
    ((PICIMFACTORY) Owner)->v->AddRef((PICIMFACTORY) Owner);

    This->super.v = CimEnum_ClassDesc.iface[0].v;
    This->super.RefCnt = 1;

    This->Flags = Flags;
    This->Query = Query;

    return This;
}

INT CimEnumDelete(PTR pItem, PTR pArg) 
{
    PIHEAP Heap = pArg;
    CimEnum *This = pItem;
    CimResource *Factory = This->Owner;

    ListRemoveItem(&Factory->Enumerators, cenoff, This);
    RELEASE((PIUNKNOWN) This->Owner);

    Heap->v->Free(Heap, 0, This);
    return 0;
}

PRIVATE SCODE MCT CimEnumDestroy(PIUNKNOWN iThis)
{
    METHOD_DECLARE(CimEnum, Destroy);
    PIHEAP Heap = CurrentHeap(); 
    CimFactory *Factory = TheCimFactory;

    ((PIUNKNOWN)Factory)->v->AddRef((PIUNKNOWN) Factory); /* XXX hmm. ensure last ref for unlock */
    Mutex_Lock(&Factory->mx);
    CimEnumDelete(This, Heap);
    Mutex_Unlock(&Factory->mx);
    RELEASE((PIUNKNOWN) Factory);

    sc = S_OK;
METHOD_CLEANUP;
    return sc;
}

/* A resource was deleted, bump given enumerator off of it */
PRIVATE INT CimEnumBump(CimEnum *pEnum, CimResource *pResource)
{
    /* Lock is held */
    CimResource *pFactory = pEnum->Owner;

    if (pEnum->Current == pResource) {
        /* The resource must exist */
        assert(pResource&&ListWalk(&pFactory->Resources,croff,LIST_WALK_SEARCH,
                                   NodeEq,pResource)==pResource);
        /* Fix list to point to previous (or none) */
        if (pResource == pFactory->Resources)
            pEnum->Current = NULL;
        else
            pEnum->Current = ListNodePrev(croff, pResource);
    }

    return 0;
}

PRIVATE SCODE MCT CimEnumPull(
    PIENUM iThis,
    /*optional*/ PIHEAP Heap,
    /*optional*/ PTIME pTimeout,
    /*out*/ PTR *pEntry,
    /*out*/ _TCHAR *pType)
{
    METHOD_DECLARE(CimEnum, Pull);
    CimResource *Factory = This->Owner;
    CimResource *Resource;
    const PropertyClass *Class;
    BOOL Last = FALSE;
    
    Mutex_Lock(&TheCimFactory->mx);

    CHECKSETSC(pEntry && pType, E_INVALID_PARAMETER, "param sanity");

    for (;;) {
        Resource = NULL;
        if (This->Current)
            Resource = ListNodeNext(croff, This->Current);
        /* last or none */
        CHECKSETSC(Resource != Factory->Resources,
                   E_END_OF_DATA, "enum exhausted");

        if (!Resource)      /* first */
            Resource = Factory->Resources;

        This->Current = Resource;
        if (ListNodeNext(croff, Resource) == Factory->Resources)
            Last = TRUE;
        
        if (This->Query) {
            ResourceMatchTemplate rmt;
            PDataItem Query = This->Query;
            rmt.ResourceURI = Query->ValueType;
            rmt.SelectorSet = NULL;
            rmt.Value = Query->Value;
            rmt.FieldSel = Query->ValueSelector;
            if (ResourceMatch(Resource, &rmt) == 0)
                break;              /* match! */
            else
                continue;           /* skip this entry */
        } else {
            break;
        }
    }

    CHECKSETSC(Resource, E_ENTRY_NOT_FOUND, "find matching entry");

    if (This->Flags & RESOURCE_RETURN_EPR) {
        *(PEndpointReference_QUIET)pEntry = *UpdateEpr(Resource);
        _tcscpy(pType, _T("wsa:EndpointReference"));
        sc = S_OK;
        /* XXX */
    } else {

        // XXX This current request is nothing to do with Access Check. It is just created to give as argument to get method.
        CurrentRequest cr;
        cr.Mode = 1;
        cr.Context = (PICIMRESOURCE) This->Owner;
        cr.Client = ReturnClient();
       
        Class = Resource->PropertyClass;
        sc = Class->Get(Class, Resource->Value, &cr);
        CHECKSC("get update of resource");

        *pEntry = Resource->Value;
        if (Resource && pType) {
            for ( ; Class; Class = Class->super)
                if (Class->EncodingType_out) {
                    _tcscpy(pType, Class->EncodingType_out);
                    goto Out;
                }
        }
        CHECKSETSC(FALSE, E_OBJECT_NOT_INITIALIZED,
                   "find EncodingType resource class or super class");
    }

METHOD_CLEANUP;
    Mutex_Unlock(&TheCimFactory->mx);
    if (Last && sc == S_OK)
        sc = S_FALSE;
    return sc;
}

PRIVATE SCODE MCT CimEnumReleasePullResult(PIENUM iThis,
                                           PIHEAP Heap,
                                           PTR Entry,
                                           const _TCHAR *Type)
{
    METHOD_DECLARE(CimEnum, ReleasePullResult);

    sc = S_OK;

    /* No need to do anything here */

METHOD_CLEANUP;
    return sc;
}

/*** Factory ***/

PRIVATE SCODE MCT CimFactoryDestroy(PIUNKNOWN iThis)
{
    METHOD_DECLARE(CimFactory, Destroy);
    PIHEAP Heap = CurrentHeap();
    sc = S_OK;

    DBG(("CimFactoryDestroy %x\n", (UINT) This));

    RELEASE((PICIMRESOURCE) This->Root);
    Condition_Destroy(&This->cd);
    Mutex_Destroy(&This->mx);
    Heap->v->Free(Heap, 0, This);

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE MCT CimResourceEnumerate (
    /*in*/ PICIMRESOURCE iThis,
    /*optional*/ PIHEAP Heap,
    /*in*/ const _TCHAR *To,
    /*in*/ const _TCHAR *name,
    RESOURCE_FLAGS Flags,
    PDataItem Query, /* XXX Note that callee does not make copy */
    /*out*/ PIENUM *pEnum,
    /*out*/ PFaultTemplate Fault )
{
    METHOD_DECLARE(CimResource, Enumerate);
    CimEnum *Enum;
    CimResource *ParentResource;
#if __WANTS_TRUST 
    AccessTemplate atreturn;
    AccessTemplate at;
#endif
    /* XXX Lock needs to be held during access check XXX */
    
#if __WANTS_TRUST 
    sc = PerformAccessCheck(iThis,To,0, &at);

    if(sc == E_ACCESS_DENIED) {
        ((FaultTemplate *)Fault)->FaultMessage = at.evidenceUrl;     
    }
    CHECKSC("PerformAccessCheck");
#endif
    ParentResource = (CimResource *)CimDirectoryLookup(iThis,To);

    if(ParentResource)
        Enum = CimEnumNew(CurrentHeap(), Flags, ParentResource, Query);
    else
        Enum = CimEnumNew(CurrentHeap(), Flags, This, Query);
    
    CHECKSETSC(Enum, E_NOT_ENOUGH_MEMORY, "CimEnumNew");

    Mutex_Lock(&TheCimFactory->mx);
    /* NOTE: Possibly enforce limit on number of enumerators */
    if(ParentResource)
        ListAddLast(&ParentResource->Enumerators, cenoff, Enum);
    else
        ListAddLast(&This->Enumerators, cenoff, Enum);

    Mutex_Unlock(&TheCimFactory->mx);
#if __WANTS_TRUST 
    /* This access check is for the the return message */
    sc = PerformAccessCheckReturn(iThis,To,0, &atreturn);
    if(sc == E_ACCESS_DENIED) {
       ((FaultTemplate *)Fault)->FaultMessage = atreturn.evidenceUrl;
    }
    CHECKSC("PerformAccessCheckReturn");
#endif
    *pEnum = (PIENUM) Enum;
    sc = S_OK;
    
METHOD_CLEANUP;
    return sc;
}

extern const PropertyClass demo_sensors_Prop;
extern const PropertyClass TrustScript_Prop;
extern const PropertyClass DemoTrustScript_Prop;
extern const PropertyClass TrustEvidence_Prop;
extern const PropertyClass DemoTrustEvidence_Prop;
extern const PropertyClass ServiceInfoType_Prop;
extern const PropertyClass DemoServiceInfoType_Prop;
extern const PropertyClass StickState_Prop;
extern const PropertyClass Demo_StickState_Prop;
extern const PropertyClass Subscription_Prop;
extern const PropertyClass Demo_Subscription_Prop;
extern const PropertyClass GyroPolicy_Prop;
extern const PropertyClass Demo_GyroPolicy_Prop;

const struct _ClassATable {
    const _TCHAR *Name;
    const PropertyClass *Class;
} PropertyClassTable[] = {
#ifndef _MINIMIZE
    { _T("demo_sensors"), &demo_sensors_Prop },
#endif
#if __WANTS_TRUST
    { _T("TrustScript"), &TrustScript_Prop },
    { _T("http://tempuri.org/trust/TrustScript"), &TrustScript_Prop },
    { _T("DemoTrustScript"), &DemoTrustScript_Prop },
    { _T("http://tempuri.org/trust/DemoTrustScript"), &DemoTrustScript_Prop },
    { _T("TrustEvidence"), &TrustEvidence_Prop },
    { _T("http://tempuri.org/trust/TrustEvidence"), &TrustEvidence_Prop },
    { _T("DemoTrustEvidence"), &DemoTrustEvidence_Prop },
    { _T("http://tempuri.org/trust/DemoTrustEvidence"), &DemoTrustEvidence_Prop },
#endif
    { _T("ServiceInfoType"), &ServiceInfoType_Prop },
    { _T("dssp:ServiceInfoType"), &ServiceInfoType_Prop },
    { _T("StickState"), &Demo_StickState_Prop },
    { _T("dssp:Subscription"), &Subscription_Prop },
    { _T("Subscription"), &Demo_Subscription_Prop },
    { _T("st1:GyroPolicy"), &GyroPolicy_Prop },
    { _T("GyroPolicy"), &Demo_GyroPolicy_Prop },
    { _T("http://schemas.tempuri.org/2007/07/stick.html"), &StickState_Prop },
    { _T("DemoServiceInfoType"), &DemoServiceInfoType_Prop },
    { _T("dssp:DemoServiceInfoType"), &DemoServiceInfoType_Prop },
    {NULL, NULL}
};


PRIVATE SCODE LookupClass(
    const _TCHAR *ResourceURI,
    CPPropertyClass *ppClass)
{
    const struct _ClassATable *entry;
    for (entry = PropertyClassTable; entry->Name; entry++) {
        if (!_tcscmp(ResourceURI, entry->Name)) {
            *ppClass = entry->Class;
            return S_OK;
        }
    }
    return E_ENTRY_NOT_FOUND;
}

PRIVATE SCODE StartClasses(
    /*in*/ PICIMFACTORY iThis)
{
#ifndef _MINIMIZE
    CPPropertyClass class;
    /* XXX finesse this */    
    class = &demo_sensors_Prop;
    if (class->Start)
        class->Start(class, iThis);
#endif
    return S_OK;
}

/* Lock is held. */

PICIMRESOURCE CimDirectoryLookup(
    PICIMRESOURCE iThis, const _TCHAR *To)
{    
    METHOD_DECLARE(CimResource, Lookup);
    ResourceMatchTemplate rmt;    
    CimResource *Resource;
    PICIMRESOURCE ret = NULL;
    struct Selector  selector;
    _TCHAR *prefix = _tcsreturnprefix(To);
    const _TCHAR *suffix = _tcsreturnsuffix(To);

    CHECKSETSC(prefix, E_NOT_ENOUGH_MEMORY, "_tcsreturnprefix");

    sc = S_OK;
    rmt.ResourceURI = _T("ServiceInfoType");
    selector.Name = _T("Name");
    selector.Value = prefix;
    selector.Type = NULL;
    selector.ListNode_Selector.pNext = &selector;
    selector.ListNode_Selector.pPrev = &selector;

    rmt.SelectorSet =  &selector;
    rmt.FieldSel = NULL;

    Resource = ListWalkT(CimResource, croff, ResourceMatchTemplate,
                         &This->Resources, LIST_WALK_SEARCH,
                         ResourceMatch, &rmt);
    CHECKSETSC(Resource, E_NAME_NOT_FOUND, "CimDirectoryLookup: can't find resource");

    if(!_tcscmp(To,suffix))
        ret = (PICIMRESOURCE) Resource;
    else
        ret = CimDirectoryLookup((PICIMRESOURCE)Resource,suffix);

METHOD_CLEANUP;
    if (prefix)
        free(prefix);
    return ret;
}

PICIMRESOURCE CimResourceShadowLookup(
    PICIMRESOURCE iThis,
    const _TCHAR *ResourceURI,
    PSelector SelectorSet)
{
    //Works for DemoServiceInfoType Only. 
    //Appropriate type checking has been done.
    
    PICIMRESOURCE Resource = NULL;
    PICIMRESOURCE Client = iThis;

    
    while(!Resource && Client)
    {      
            Resource = CimResourceLookup(Client, ResourceURI, SelectorSet);
            if (!_tcscmp(    (((CimResource *)Client)->PropertyClass)->ResourceName , _T("DemoServiceInfoType") )  
                || !_tcscmp(    (((CimResource *)Client)->PropertyClass)->ResourceURI , _T("DemoServiceInfoType") ) ) 
            {
                Client = ReturnShadowParent(((CimResource *)Client)->Value);
            }
            else 
                Client = NULL;
    }
    return Resource;
}

PICIMRESOURCE CimResourceLookup(
    PICIMRESOURCE iThis,
    const _TCHAR *ResourceURI,
    PSelector SelectorSet)
{
    METHOD_DECLARE(CimResource, Lookup);
    ResourceMatchTemplate rmt;    
    CimResource *Resource;

    sc = S_OK; /* macro side effect ... */

    rmt.ResourceURI = ResourceURI;
    rmt.SelectorSet = SelectorSet;
    rmt.FieldSel = NULL;

    Resource = ListWalkT(CimResource, croff, ResourceMatchTemplate,
                         &This->Resources, LIST_WALK_SEARCH,
                         ResourceMatch, &rmt);
    return (PICIMRESOURCE) Resource;
}

/* Lock is held. */

UINT CimResourceGetValue(
    PICIMRESOURCE iThis, CurrentRequest *currentRequest, const _TCHAR *Field)
{
    METHOD_DECLARE(CimResource, Lookup);
    const PropertyClass *Class;
    Class = This->PropertyClass;
    sc = S_OK;

    return Class->GetValue(Class, This->Value, currentRequest,Field);
   
}


PRIVATE SCODE CimResourceGet (
    PICIMRESOURCE iThis,const _TCHAR *To,
    const _TCHAR *ResourceURI, PSelector SelectorSet, TIME OperationTimeout,
    void *Value, _TCHAR *pValueType, UINT ValueSize)
{
    return S_OK;
}
    
PRIVATE SCODE CimResourceGetResource (
    PICIMRESOURCE iThis,const _TCHAR *To,
    const _TCHAR *ResourceURI, PSelector SelectorSet, TIME OperationTimeout,
    void *Value, _TCHAR *pValueType, UINT ValueSize, /*out*/ PFaultTemplate Fault )
{
    METHOD_DECLARE(CimResource, GetResource);
    //ResourceMatchTemplate rmt;    
    CimResource *Resource;
    CimResource *ParentResource = NULL;
    const PropertyClass *Class;
#if __WANTS_TRUST 
    AccessTemplate at;
    AccessTemplate atreturn;
#endif
    // XXX This current request is nothing to do with Access Check. It is just created to give as argument to get method.
    CurrentRequest cr;
    //_TCHAR *policyUrl = malloc(strlen(To)+1);
    //strcpy(policyUrl,To);
    cr.Context = CimDirectoryLookup(iThis,To);
    cr.Client = ReturnClient();
    cr.Mode = 1;
#if __WANTS_TRUST 
    sc = PerformAccessCheck(iThis,To,0, &at);
       
    //If the access check fails, the method returns
    if(sc == E_ACCESS_DENIED) {
        ((FaultTemplate *)Fault)->FaultMessage = at.evidenceUrl;    
    }
    CHECKSC("PerformAccessCheck");
#endif
    sc = S_OK;
    Mutex_Lock(&TheCimFactory->mx);
    ParentResource = (CimResource *) CimDirectoryLookup(iThis, To);
    if(ParentResource)
    {
       Resource = (CimResource *) CimResourceLookup((PICIMRESOURCE)ParentResource, ResourceURI, SelectorSet); 
    }
    else
    {
         Resource = (CimResource *) CimResourceLookup(iThis, ResourceURI, SelectorSet); 

    }
        
#if 0
    rmt.ResourceURI = ResourceURI;
    rmt.SelectorSet = SelectorSet;
    rmt.FieldSel = NULL;

    Resource = ListWalkT(CimResource, croff, ResourceMatchTemplate,
                         &This->Resources, LIST_WALK_SEARCH,
                         ResourceMatch, &rmt);
#endif
    CHECKSETSC(Resource, E_WSMAN_INVALID_SELECTORS, "find matching resource");

    Class = Resource->PropertyClass;
    sc = Class->Get(Class, Resource->Value, &cr);
    CHECKSC("get update of resource");

    CHECKSETSC(sizeof(Class->DataSize <= ValueSize), E_WSMAN_ENCODING_LIMIT,
               "make sure data fits into buffer");

    if (Resource && pValueType) {
        for ( ; Class; Class = Class->super)
            if (Class->EncodingType_out) {
                CHECKSETSC(ValueSize >= Class->DataSize, E_BUFFER_TOO_SMALL,
                           "verify data fits in output buffer");
                memcpy(Value, Resource->Value, Class->DataSize);
                _tcscpy(pValueType, Class->EncodingType_out);
#if __WANTS_TRUST 
                /* This access check is for the the return message */
                sc = PerformAccessCheckReturn(iThis,To,0, &atreturn);
                if(sc == E_ACCESS_DENIED) {
                    ((FaultTemplate *)Fault)->FaultMessage = atreturn.evidenceUrl;
                }
                CHECKSC("PerformAccessCheckReturn");
#endif
                goto Out;
            }
    }
    CHECKSETSC(FALSE, E_OBJECT_NOT_INITIALIZED,
               "find EncodingType resource class or super class");

METHOD_CLEANUP;
    Mutex_Unlock(&TheCimFactory->mx);
    return sc;
}

PRIVATE SCODE CimResourceChangeResource (
    PICIMRESOURCE iThis,
    RESOURCE_FLAGS Flags,
    const _TCHAR *To,
    const _TCHAR *ResourceURI, PSelector SelectorSet,
    const void *NewValue, const _TCHAR *NewValueType, const UINT32 *FieldSel,
    void *Value, _TCHAR *pValueType, UINT ValueSize, /*out*/ PFaultTemplate Fault )
{

    METHOD_DECLARE(CimResource, ChangeResource);
    ResourceMatchTemplate rmt;    
    CimResource *Resource;
    CimResource *ParentResource = NULL;
    const PropertyClass *Class;
    PIHEAP Heap = CurrentHeap();
#if __WANTS_TRUST     
    AccessTemplate atreturn;
    AccessTemplate at;  
#endif
    /* vars with "sub" relate to subscriptions
     * vars with "evi" relate to evidence, in this case, gyro policy
     */
    CimResource *Resource_sub = NULL;
    CimResource *Resource_evi = NULL;
    const PropertyClass *Class_sub;
    const PropertyClass *Class_evi;
    void *pullvalue = NULL;
    void *pullvalue_evi = NULL;
    CimResource *ParentResourceSub = NULL;
    CimResource *ParentResourceEvi = NULL;
#if 1    
    _TCHAR ToSub[60];
    _TCHAR ToEvi[60];
    SCODE scpull, scpull_evi, sc_evidence = S_OK;
    _TCHAR subresulttype[100];
    _TCHAR eviresulttype[100];
    PIENUM enmSub, enmEvi;
    const _TCHAR *temp_storage = NULL;
    _TCHAR *new_subscriber = NULL;
#endif
    ResourceMatchTemplate rmt_sub, rmt_evi;
    const _TCHAR *baseSub = _T("Subscription");
    const _TCHAR *baseEvi = _T("GyroPolicy");
    
    struct Subscription *subresult2 = NULL;
    struct GyroPolicy *eviresult2 = NULL;
#if 1
#if __WANTS_TRUST 
    sc = PerformAccessCheck(iThis,To,1, &at);
    if(sc == E_ACCESS_DENIED) {
       ((FaultTemplate *)Fault)->FaultMessage = at.evidenceUrl;
    }
    CHECKSC("PerformAccessCheck");
#endif
#endif
    sc = S_OK;
    rmt.ResourceURI = ResourceURI;
    rmt.SelectorSet = SelectorSet;
    rmt.FieldSel = FieldSel;

    if (!ResourceURI || !*ResourceURI)
        rmt.ResourceURI = NewValueType;

    if (!SelectorSet) {
        rmt.Value = NewValue;
        rmt.ResourceURI = NewValueType;
    }
    
    Mutex_Lock(&TheCimFactory->mx);
    ParentResource = (CimResource *)CimDirectoryLookup(iThis,To);
    if(ParentResource)
        Resource = ListWalkT(CimResource, croff, ResourceMatchTemplate,
                         &ParentResource->Resources, LIST_WALK_SEARCH,
                         ResourceMatch, &rmt);
    else
        Resource = ListWalkT(CimResource, croff, ResourceMatchTemplate,
                         &This->Resources, LIST_WALK_SEARCH,
                         ResourceMatch, &rmt);

    if (Resource) {
        CHECKSETSC(!(Flags & NAME_SPACE_FAILIFEXIST), E_ALREADY_EXISTS,
                   "resource already exists");
        Class = Resource->PropertyClass;
    } else {
        CHECKSETSC(Flags & NAME_SPACE_CREATE, E_WSMAN_INVALID_SELECTORS,
                   "find matching resource");
        sc = LookupClass(ResourceURI, &Class); 
        
        CHECKSC("LookupClass");
        if(!ParentResource){
            Resource = ResourceNew(Heap, This, Class, TRUE);
            CHECKSETSC(Resource, E_NOT_ENOUGH_MEMORY, "create new resource");
            ListAddLast(&This->Resources, croff, Resource);

        }
        else{
                Resource = ResourceNew(Heap, ParentResource, Class, TRUE);
                CHECKSETSC(Resource, E_NOT_ENOUGH_MEMORY, "create new resource");
                ListAddLast(&ParentResource->Resources, croff, Resource);
        }

    }

    CHECKSETSC(sizeof(Class->DataSize <= ValueSize), E_WSMAN_ENCODING_LIMIT,
               "make sure data fits into buffer");

    
    if (Flags & RESOURCE_DELETE) {
        ResourceDelete(Resource, NULL);
        /* XXX take extra ref on factory untile Mutex_Unlock */
    } else {
        /* XXX If this is a subscription, perform 'gethostbyname' here,
         * which is for now just the EPR address inserted in the command
         * line, e.g. cimdb.cob -a 
         * Eventually this would need to be fixed to work with non-localhost nodes. 
         */
        if(!_tcscmp(NewValueType, _T("dssp:Subscription"))){
            PSubscription tempsub = (PSubscription) NewValue;
            new_subscriber = malloc(_tcslen(tempsub->Subscriber) + 10*sizeof(_TCHAR));//
            temp_storage = _tcsreturnsuffix(tempsub->Subscriber);
            temp_storage = _tcsreturnsuffix(++temp_storage);            
            memcpy(new_subscriber, tempsub->Subscriber, 7 * sizeof(_TCHAR));
            new_subscriber[7] = '\0';
            _tcscat(new_subscriber, TheCimFactory->EprAddress);
            _tcscat(new_subscriber, _T(":50000/"));
            _tcscat(new_subscriber, temp_storage);
            _tcscpy(tempsub->Subscriber, new_subscriber);

            if(new_subscriber)
                free(new_subscriber);
        }
        
        sc = ResourceCopy(Class, Resource->Value,
                          NewValueType, FieldSel,
                          NewValue, FALSE, TRUE, Heap,(PICIMRESOURCE)Resource);
        CHECKSC("ResourceCopy");

        sc = Class->Put(Class, Resource->Value);
        CHECKSC("put to update resource");

        Mutex_Unlock(&TheCimFactory->mx);

        /* Now that the state has been updated, check for subscribers. */             
        if((_tcscmp(NewValueType,_T("dssp:Subscription")))&&(_tcscmp(NewValueType,_T("dssp:ServiceInfoType")))) {
            _tcscpy(ToSub, _T("SUBSCRIPTIONS/"));
            _tcscat(ToSub, To);
            ParentResourceSub = (CimResource *)CimDirectoryLookup(iThis,ToSub);
            if(ParentResourceSub) {
                /* Reason to believe there are subscriptions present, so check for policies first. */
                _tcscpy(ToEvi, ToSub+5);
                memcpy(ToEvi, _T("EVIDENCE"), 8*sizeof(_TCHAR));

                ParentResourceEvi = (CimResource *)CimDirectoryLookup(iThis,ToEvi);
                if(ParentResourceEvi) {
                    /* Perform a get on the states, which will enumerate the policies. */
                    CimResourceEnumerate((ICimResource*)ParentResourceEvi,CurrentHeap(),ToEvi,baseEvi,0, NULL,
                                         (void*) &enmEvi, NULL);
                    if(enmEvi){
                        do 
                        {
                            scpull_evi = CimEnumPull(enmEvi,
                                /*optional*/ CurrentHeap(),
                                /*optional*/ NULL,
                                /*out*/ &pullvalue_evi,
                                /*out*/ eviresulttype);

                            /* Get value in GyroPolicy format */
                            eviresult2 = (PGyroPolicy) pullvalue_evi;
                            rmt_evi.Value = pullvalue_evi;
                            rmt_evi.ResourceURI = _T("GyroPolicy");
                            rmt_evi.SelectorSet = NULL;
                            rmt_evi.FieldSel = 0;

                            Resource_evi = ListWalkT(CimResource, croff, ResourceMatchTemplate,
                                                     &ParentResourceEvi->Resources, LIST_WALK_SEARCH,
                                                     ResourceMatch, &rmt_evi);

                            if(Resource_evi){
                                Class_evi = Resource_evi->PropertyClass;
                                sc_evidence = Class_evi->Apply(Class_evi, (PICIMRESOURCE) Resource_evi,(PICIMRESOURCE) Resource);
                            }
                        }while(scpull_evi != 0x00000001);
                    }//endif
                }
                                
                /* Perform a get on the states, which will enumerate the subscriptions. */
                CimResourceEnumerate((ICimResource*)ParentResourceSub,CurrentHeap(),ToSub,baseSub,0, NULL,
                                     (void*) &enmSub, NULL);
                if(enmSub){
                    
                    do 
                    {
                        sc = CimEnumPull(enmSub,
                            /*optional*/ CurrentHeap(),
                            /*optional*/ NULL,
                            /*out*/ &pullvalue,
                            /*out*/ subresulttype);
                        scpull = sc;

                        /* Get value in Subscription format*/
                        subresult2 = (PSubscription) pullvalue;
                        rmt_sub.Value = pullvalue;
                        rmt_sub.ResourceURI = _T("Subscription");
                        rmt_sub.FieldSel = 0;
                        rmt_sub.SelectorSet = NULL;

                        Resource_sub = ListWalkT(CimResource, croff, ResourceMatchTemplate,
                                                 &ParentResourceSub->Resources, LIST_WALK_SEARCH,
                                                 ResourceMatch, &rmt_sub);

                        if(Resource_sub){
                            if(sc_evidence == S_OK) {
                                Class_sub = Resource_sub->PropertyClass;
                                Class_sub->Apply(Class_sub, (PICIMRESOURCE) Resource_sub,(PICIMRESOURCE) Resource);
                            }
                        }

                    }while(scpull != 0x00000001);
                }//endif
            }//endifParentResourceSub
        }//endif

        Mutex_Lock(&TheCimFactory->mx);

        if (Flags & RESOURCE_RETURN_EPR) {
            *(PEndpointReference_QUIET)Value = *UpdateEpr(Resource);
            if (pValueType)
                _tcscpy(pValueType, _T("wsa:EndpointReference"));
        } else if (Resource && pValueType) {
            for ( ; Class; Class = Class->super)
                if (Class->EncodingType_out) {
                    CHECKSETSC(ValueSize >= Class->DataSize, E_BUFFER_TOO_SMALL,
                               "verify data fits in output buffer");
                    memcpy(Value, Resource->Value, Class->DataSize);
                    _tcscpy(pValueType, Class->EncodingType_out);
#if __WANTS_TRUST 
                    /* This access check is for the the return message */
                    sc = PerformAccessCheckReturn(iThis,To,1, &atreturn);
                    if(sc == E_ACCESS_DENIED) {
                       ((FaultTemplate *)Fault)->FaultMessage = atreturn.evidenceUrl;
                    }
                    CHECKSC("PerformAccessCheckReturn");
#endif
                    goto Out;
                }
            CHECKSETSC(FALSE, E_OBJECT_NOT_INITIALIZED,
                       "find EncodingType resource class or super class");
        }
    }

METHOD_CLEANUP;
    
    Mutex_Unlock(&TheCimFactory->mx);
    
    return sc;
}

#if 0
PRIVATE SCODE CimResourceQueryResource (
    PICIMRESOURCE iThis,
    const _TCHAR *ResourceURI, PSelector SelectorSet,
    const void *NewValue, const _TCHAR *ValueType, const UINT32 *FieldSel,
    void *Value, _TCHAR *pValueType, UINT ValueSize)
{
    METHOD_DECLARE(CimResource, QueryResource);
    ResourceMatchTemplate rmt;    
    CimResource *Resource;
    const PropertyClass *Class;
    CurrentRequest currentRequest;
    //PIHEAP Heap = CurrentHeap();

    sc = S_OK;

    rmt.ResourceURI = ResourceURI;
    rmt.SelectorSet = SelectorSet;
    if (!SelectorSet) {
        rmt.Value = NewValue;
        rmt.ResourceURI = ValueType;
    }

    Mutex_Lock(&TheCimFactory->mx);
    Resource = ListWalkT(CimResource, croff, ResourceMatchTemplate,
                         &This->Resources, LIST_WALK_SEARCH,
                         ResourceMatch, &rmt);
    if (Resource) {
        Class = Resource->PropertyClass;

        currentRequest.Context = iThis;
        currentRequest.Client = ReturnClient();
       // currentRequest.Client =CimDirectoryLookup(iThis,_T("root"));


        sc = Class->Get(Class, Resource->Value,&currentRequest);

        /*Connect output values and type*/
        memcpy(Value, Resource->Value, Class->DataSize);
        _tcscpy((void*) pValueType, Resource->PropertyClass->super->ResourceURI);
    } else {
        //Error message here...no resource exists
    }

METHOD_CLEANUP;
    Mutex_Unlock(&TheCimFactory->mx);
    return sc;
}
#endif

#if 0
PRIVATE SCODE CimResourceDelete (
    PICIMRESOURCE iThis,const _TCHAR *To,
    const _TCHAR *ResourceURI, PSelector SelectorSet,
    const void *NewValue, const _TCHAR *NewValueType, const UINT32 *FieldSel,
    void *Value, _TCHAR *pValueType, UINT ValueSize)
{
    METHOD_DECLARE(CimResource, Delete);
    ResourceMatchTemplate rmt;  
    CimResource *Resource;
    const PropertyClass *Class;
    PIHEAP Heap = CurrentHeap();

    CimResource *Parent;// = This->Parent;

    sc = S_OK;

    rmt.ResourceURI = ResourceURI;
    rmt.SelectorSet = SelectorSet;
    if (!SelectorSet) {
        rmt.Value = NewValue;
        rmt.ResourceURI = NewValueType;
    }
     
    Mutex_Lock(&TheCimFactory->mx);
    Resource = ListWalkT(CimResource, croff, ResourceMatchTemplate,
                         &This->Resources, LIST_WALK_SEARCH,
                         ResourceMatch, &rmt);
    
    if (Resource) {
        Parent = Resource->Parent;

        if (Parent) {
        ListWalkT(CimEnum, cenoff, CimResource,
                  &Parent->Enumerators, LIST_WALK_ITERATE,
                  CimEnumBump, This);
        ListRemoveItem(&Parent->Resources, croff, This);
        RELEASE((PIUNKNOWN) Parent);
        }

        /* Call each finalizer */           
        for ( ; Class; Class = Class->super)
            if (Class->Finalize)
                Class->Finalize(Class, This->Value, Heap);

        if (Heap) {
            Heap->v->Free(Heap, 0, This);
            Heap->v->Release(Heap);
        }
    }

    METHOD_CLEANUP;

    Mutex_Unlock(&TheCimFactory->mx);
    return sc;
}
#endif

#if 0
#include <stdlib.h>
#include <netif.h>
#include <winsock.h>
#include <net/ip/in.h>          /* struct ip_mreq */
#include <net/if.h>             /* for struct ifreq */

PRIVATE UINT32 GetInterfaceIp(const _TCHAR *iname, UINT inamelen)
{
    PINAMESPACE ns = CurrentNameSpace();
    PIENDPOINTFACTORY ef;
    PIENDPOINT ep;
    struct ifreq ifr;
    UINT32 LocalIPAddr = INADDR_ANY;
    SCODE sc;
    
    /* Prepare address and port to listen at */
    memset((void*) &ifr, 0, sizeof(struct ifreq));

    /* Get an endpoint factory to create an endpoint */
    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void**) &ef);
    if (FAILED(sc)) goto Out;

    /* Create an endpoint to query for the interface address */
    sc = ef->v->CreateEndpoint(ef, AF_INET, SOCK_DGRAM, 0, &ep);
    ef->v->Release(ef);
    if (FAILED(sc)) goto Out;

    /* This is where we specify which interface to use */
#ifdef _UNICODE
{
    char *s = _wcstos(iname);
    if (!s)
        return E_NOT_ENOUGH_MEMORY;
    strncpy(ifr.ifr_name, s, __min(inamelen, sizeof(ifr.ifr_name)));
    free(s);
}
#else
    strncpy(ifr.ifr_name, iname, __min(inamelen, sizeof(ifr.ifr_name)));
#endif
    
    /* Query for the interface address */
    sc = ep->v->Ioctl(ep, SIOCGIFADDR, (BYTE*) &ifr);
    ep->v->Release(ep);
    if (FAILED(sc)) goto Out;

    LocalIPAddr = ((struct sockaddr_in *) &(ifr.ifr_addr))->sin_addr.s_addr;
Out:
    return LocalIPAddr;
}
#endif

PRIVATE SCODE CimFactoryMain (
    PICIMFACTORY iThis,
    const _TCHAR * cmdline)
{
    METHOD_DECLARE(CimFactory, Main);
    PICONVERT convert = NULL;
    PIFILE init = NULL;
    PIHEAPFACTORY heapfactory = NULL;
    PICIMRESOURCE cimdir = NULL;
    PIHEAP Heap = NULL;
    PTR val;
    EndpointReference_QUIET Epr = {0,};
    //struct SelectorSet SelectorSet = {0,};
    _TCHAR typeis[100], *arg;
    FaultTemplate Fault;
    
    sc = S_OK;

    if (!cmdline || *cmdline == _EOS)
        cmdline = _T("fs/demosens.xml");

    arg = _tcspbrk(cmdline, _T(" \t"));
    if (!_tcsncmp(cmdline, _T("-a"), 2)) {
        /* cmd -a http://1.2.3.4/wsman */
        Mutex_Lock(&This->mx);
        if (This->EprAddress) {
            Mutex_Unlock(&This->mx);
            CHECKSETSC(FALSE, E_ALREADY_EXISTS,
                       "set EprAddress, already set");
        }
        if (arg)
            This->EprAddress = _tcsdup(++arg);
        else
            This->EprAddress = _tcsdup(_T("http://127.0.0.1/wsman"));
        This->Epr.Address = This->EprAddress;
        Mutex_Unlock(&This->mx);
        cmdline = _T("");
#if MEASURE
    } else if (_tcslen(cmdline) > 0 && !_tcsncmp(cmdline, _T("-m"), 2)) {
        doMeasure = !doMeasure;
        cmdline += 2;
        return S_OK;
#endif
    } else if (!This->Started) {
        if (AtomicCmpAndSwap(&This->Started, 0, TRUE)) {
#if 0
            if (!This->EprAddress) {
                UINT ip = GetInterfaceIp("*0", 2);
                This->EprAddress = malloc((13+15+1) * sizeof(_TCHAR));
                CHECKSETSC(This->EprAddress, E_NOT_ENOUGH_MEMORY,"malloc EPR");
                _stprintf(This->EprAddress, _T("http://%s/wsman"),
                          inet_ntoa(*(struct in_addr *)&ip));
            }
#endif
            StartClasses(iThis);
            This->Started = TRUE;
        }
    }
   
    if (!_tcsncmp(cmdline, _T("-f"), 2)) {
        sc = BindToObject(CurrentNameSpace(),
                          _T("COB/convert.cob"), NAME_SPACE_READ,
                          &IID_IConvert, (void **) &convert);
        CHECKSC("get soap.cob");

        sc = BindToObject(CurrentNameSpace(), _T("COB/tmp_heap.cob"),
                          NAME_SPACE_READ,
                          &IID_IHeapFactory, (void **) &heapfactory);
        CHECKSC("get tmp heap factory");

        sc = heapfactory->v->CreateNested(heapfactory, CurrentHeap(),
                                          0, 0, 10000, &Heap);
        CHECKSC("create tmp heap");

        sc = BindToObject(CurrentNameSpace(), cmdline, NAME_SPACE_READ,
                          &IID_IFile, (void **) &init);
        CHECKSC("get init file");

        sc = convert->v->ProcessFragmentFile(convert, init, Heap,
                                             &val, typeis);
        CHECKSC("Process init file");

        sc = iThis->v->QueryInterface(iThis, &IID_ICimResource, (void **) &cimdir);
        CHECKSC("QI for wxf");

        /* Note: need to copy resource since subclass is larger */
        sc = cimdir->v->ChangeResource(cimdir, NAME_SPACE_CREATE,_T("root"),
                                    _T("demo_sensors"), NULL,
                                    val, typeis, NULL,
                                    &Epr, NULL, sizeof(Epr),&Fault);
        CHECKSC("CreateResource");
    }
METHOD_CLEANUP;
    RELEASE(init);
    RELEASE(cimdir);
    RELEASE(Heap);
    RELEASE(convert);
    RELEASE(heapfactory);
    return sc;
}

PRIVATE SCODE CimFactoryQueryInterface(PICIMFACTORY iThis, REFIID IidQuery,
                                       void **ppObject)
{
    DemoServiceInfoType* sit;
    METHOD_DECLARE(CimFactory, QueryInterface);
    if (UuidCmp(&IID_ICimResource, IidQuery)
        || UuidCmp(&IID_IWsTransfer, IidQuery))
    {
        PICIMRESOURCE root = (PICIMRESOURCE) This->Root;
        if (!root) {
            extern const PropertyClass DemoServiceInfoType_Prop;
            root = (PICIMRESOURCE) ResourceNew(CurrentHeap(), NULL, &DemoServiceInfoType_Prop, FALSE);
            CHECKSETSC(root, E_NOT_ENOUGH_MEMORY, "create root resource");
            sit = (DemoServiceInfoType *)root;
            This->Root = (CimResource *) sit;
            
            sc = root->v->QueryInterface(root, IidQuery, ppObject);
            /* Once the root directory becomes empty and no user refs it goes away */
            //RELEASE(root); 
        } else {

            sc = root->v->QueryInterface(root, IidQuery, ppObject);

        }
        CHECKSC("QI root");
    } else {
        sc = GenericQueryInterface((PIUNKNOWN) iThis, IidQuery,
                                   ppObject, &IID_IProgram);
        CHECKSC("GenericQI");
    }

METHOD_CLEANUP;
    return sc;
}

PIUNKNOWN cimdbCobMain(void)
{
    PIUNKNOWN cob = GenericCobInit((void*) &cimdb_CobDesc, NULL);
    TheCimFactory = (CimFactory *) cob;
    return cob;
}
