/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include "demosens.h"
#include <ctype.h>
#include <limits.h>
#include <stdlib.h>

/*********** Demo sensor interface as specified ******************/
#define DEMO_SENSOR_OK 0
#define DEMO_SENSOR_INVALID_ID 1
#define DEMO_SENSOR_TIMEOUT 2
#define DEMO_SENSOR_BUS_ERROR 3
#define DEMO_SENSOR_NOT_INITIALIZED 4
#define DEMO_SENSOR_UNKNOWN 255

int GetSensor(UINT8 ID, UINT8 * val);
int SetSensor(UINT8 ID, UINT8 * val);

/*********** Fake implementation ******************/
int GetSensor(UINT8 ID, UINT8 * val)
{
    static UINT8 _val = 1;
    
    if (ID > 35)
        return DEMO_SENSOR_INVALID_ID;
        
    if (ID <= 11)
        *val = 60;
    else if (ID <= 16)
        *val = 40;
    else
        *val = _val++;
    DBG(("GetSensor %d seq=%d\n", ID, *val));
    return DEMO_SENSOR_OK;
}

int SetSensor(UINT8 ID, UINT8 * val)
{
    if (ID > 35)
        return DEMO_SENSOR_INVALID_ID;
    DBG(("SetSensor %d=%d\n", ID, *val));
    return DEMO_SENSOR_OK;
}

/*********** map names ******************/

PRIVATE const _TCHAR * const id_names[] = {
    _T("CPU0_Temp"),
    _T("CPU1_Temp"),
    _T("CPU2_Temp"),
    _T("CPU3_Temp"),
    _T("MB0_Temp"),
    _T("MB1_Temp"),
    _T("MB2_Temp"),
    _T("MB3_Temp"),
    _T("VRM0_Temp"),
    _T("VRM1_Temp"),
    _T("VRM2_Temp"),
    _T("VRM3_Temp"),
    _T("Volt1"),
    _T("Volt2"),
    _T("Volt3"),
    _T("Volt4"),
    _T("Volt5"),
    _T("Power GPIO"),
    _T("Reset GPIO"),
    _T("NMI GPIO"),
    _T("SMI GPIO"),
    _T("PCI_Reset GPIO"),
    _T("mCard_ID0 GPIO"),
    _T("mCard_ID1 GPIO"),
    _T("mCard_ID2 GPIO"),
    _T("Fault_LED GPIO"),
    _T("Chassis_ID GPIO"),
    _T("Local_Lockout GPIO"),
    _T("SCI_Int GPIO"),
    _T("Clear_CMOS GPIO"),
    _T("Therm_Trip GPIO"),
    _T("ACPI0 GPIO"),
    _T("ACPI1 GPIO"),
    _T("ACPI2 GPIO"),
    _T("Power_OK GPIO"),
    _T("AC_OK GPIO")
};
#define ID_NAMES_COUNT sizeof(id_names)/sizeof(_TCHAR*)

PRIVATE UINT FindId(const _TCHAR *Name)
{
    UINT id;
    const _TCHAR *entry;
    for (id = 0; id < ID_NAMES_COUNT; id++) {
        entry = id_names[id];
        if (!_tcscmp(entry, Name))
            return id;
    }
    /* Nothing found, return invalid id */
    return ~0;
}

PRIVATE const _TCHAR *FindName(UINT id)
{
    if (id >= ID_NAMES_COUNT)
        return NULL;

    return id_names[id];
}

/*********** demo_sensor class ******************/

PRIVATE SCODE demo_sensors_Get(const PropertyClass *Class, PTR Value, CurrentRequest *currentRequest)
{
    DemoSensor *sen = Value;
    UINT8 val;
    int ret;

    ret = GetSensor((UINT8) sen->UnitId, &val);

    /* For bitmaps restrict value to a single bit */
    if (sen->super.RateUnits == 60)
        val &= 1;

    switch (ret) {
    case DEMO_SENSOR_OK:
        sen->super.CurrentReading = val;
        return S_OK;
    case DEMO_SENSOR_INVALID_ID:
        return E_INVALID_PARAMETER;
    case DEMO_SENSOR_TIMEOUT:
        return E_TIMED_OUT;
    case DEMO_SENSOR_BUS_ERROR:
        return E_FAIL;
    case DEMO_SENSOR_NOT_INITIALIZED:
        return E_OBJECT_NOT_INITIALIZED;
    case DEMO_SENSOR_UNKNOWN:
    default:
        return E_FAIL;
    }
}

PRIVATE SCODE demo_sensors_Put(const PropertyClass *Class, PTR Value)
{
    DemoSensor *sen = Value;
    UINT8 val;
    int ret;

    val = (UINT8) sen->super.CurrentReading;
    
    ret = SetSensor((UINT8) sen->UnitId, &val);

    switch (ret) {
    case DEMO_SENSOR_OK:
        sen->super.CurrentReading = val;
        return S_OK;
    case DEMO_SENSOR_INVALID_ID:
        return E_INVALID_PARAMETER;
    case DEMO_SENSOR_TIMEOUT:
        return E_TIMED_OUT;
    case DEMO_SENSOR_BUS_ERROR:
        return E_FAIL;
    case DEMO_SENSOR_NOT_INITIALIZED:
        return E_OBJECT_NOT_INITIALIZED;
    case DEMO_SENSOR_UNKNOWN:
    default:
        return E_FAIL;
    }
}

/* Initialize private fields */
PRIVATE SCODE demo_sensors_Init(const PropertyClass *Class, PTR Value,
                                /*unused*/ PIHEAP Heap,PICIMRESOURCE Resource)
{
    DemoSensor *sen = Value;
    UINT id;
 
    if (!sen->super.DeviceID || !sen->super.ElementName)
        return E_INVALID_PARAMETER;

    if (_istdigit(*sen->super.DeviceID)) {
        id = _ttoi(sen->super.DeviceID);
        if (id > _UI8_MAX)
            return E_INVALID_PARAMETER;
    } else {
        id = FindId(sen->super.DeviceID);
        if (id > _UI8_MAX)
            return E_INVALID_PARAMETER;
    }
    if (_tcscmp(sen->super.ElementName, sen->super.DeviceID))
        return E_INVALID_PARAMETER;

    sen->UnitId = id;
    return S_OK;
}

PRIVATE SCODE demo_sensors_Copy(const PropertyClass *Class,
                                PTR _Dst, const void *_Src, PIHEAP Heap)
{
    DemoSensor *sen = _Dst;
    const DemoSensor *src = _Src;

    sen->UnitId = src->UnitId;

    return S_OK;
}

#include "demodata.h"

SCODE demo_sensors_Start(const PropertyClass *Class, PICIMFACTORY Factory)
{
    METHOD_DECLARE_NOTHIS(demo_sensors, Start);
    UINT i;
    const DemoSensor *sen;
    PICIMRESOURCE cimdir = NULL;
    FaultTemplate Fault;
    EndpointReference_QUIET Epr = {0,};

    sc = Factory->v->QueryInterface(Factory, &IID_ICimResource, (void **) &cimdir);
    CHECKSC("QI for cimdir");

    for (i = 0; i < DEMO_INIT_COUNT; i++) {
        sen = &_demo_init_data[i];

        sc = cimdir->v->ChangeResource(cimdir,
                //NAME_SPACE_CREATE, _T("demo_sensors"), NULL,
                NAME_SPACE_FAILIFEXIST | NAME_SPACE_CREATE |RESOURCE_RETURN_EPR,_T("wsman"), _T("demo_sensors"), NULL,
                sen, _T("cimns:CIM_NumericSensor"), NULL,
                &Epr, NULL, sizeof(Epr), &Fault);
        CHECKSC("CreateResource");
    }

METHOD_CLEANUP;
    RELEASE(cimdir);
    return sc;
}

extern const PropertyClass CIM_NumericSensor_Prop;

const PropertyClass demo_sensors_Prop = {
    _T("wsman:demo_sensors"), _T("demo_sensors"),
    sizeof(DemoSensor),
    _T("cimns:CIM_NumericSensor"), 
    _T("cimns:CIM_NumericSensor"),
    &CIM_NumericSensor_Prop,
    demo_sensors_Get, demo_sensors_Put,
    demo_sensors_Init, NULL, demo_sensors_Copy,
    NULL, NULL, NULL,NULL, demo_sensors_Start
};
