/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <string.h>
#include <ctype.h>
#include <mmlite.h>
#include <crypto.h>

PRIVATE const char Map[] = {
    /* 0 */ 'A',
    /* 1 */ 'B',
    /* 2 */ 'C',
    /* 3 */ 'D',
    /* 4 */ 'E',
    /* 5 */ 'F',
    /* 6 */ 'G',
    /* 7 */ 'H',
    /* 8 */ 'I',
    /* 9 */ 'J',
    /* 10 */ 'K',
    /* 11 */ 'L',
    /* 12 */ 'M',
    /* 13 */ 'N',
    /* 14 */ 'O',
    /* 15 */ 'P',
    /* 16 */ 'Q',
    /* 17 */ 'R',
    /* 18 */ 'S',
    /* 19 */ 'T',
    /* 20 */ 'U',
    /* 21 */ 'V',
    /* 22 */ 'W',
    /* 23 */ 'X',
    /* 24 */ 'Y',
    /* 25 */ 'Z',
    /* 26 */ 'a',
    /* 27 */ 'b',
    /* 28 */ 'c',
    /* 29 */ 'd',
    /* 30 */ 'e',
    /* 31 */ 'f',
    /* 32 */ 'g',
    /* 33 */ 'h',
    /* 34 */ 'i',
    /* 35 */ 'j',
    /* 36 */ 'k',
    /* 37 */ 'l',
    /* 38 */ 'm',
    /* 39 */ 'n',
    /* 40 */ 'o',
    /* 41 */ 'p',
    /* 42 */ 'q',
    /* 43 */ 'r',
    /* 44 */ 's',
    /* 45 */ 't',
    /* 46 */ 'u',
    /* 47 */ 'v',
    /* 48 */ 'w',
    /* 49 */ 'x',
    /* 50 */ 'y',
    /* 51 */ 'z',
    /* 52 */ '0',
    /* 53 */ '1',
    /* 54 */ '2',
    /* 55 */ '3',
    /* 56 */ '4',
    /* 57 */ '5',
    /* 58 */ '6',
    /* 59 */ '7',
    /* 60 */ '8',
    /* 61 */ '9',
    /* 62 */ '+',
    /* 63 */ '/'
};

PRIVATE char PadChar = '=';

/* Return the position in array above or a special token as appropriate */
#define SKIP 0x100
#define PAD  0x200
PRIVATE UINT Decode(const _TCHAR c)
{
    if (c >= 'A' && c <= 'Z')
        return c - 'A';
    if (c >= 'a' && c <= 'z')
        return c - 'a' + 26;
    if (c >= '0' && c <= '9')
        return c - '0' + 52;
    if (c == '+')
        return 62;
    if (c == '/')
        return 63;
    if (isspace(c))
        return SKIP;
    if (c == PadChar)
        return PAD;
    /* else might signal error... */
    return PAD;
}

/* Read one char, decode it, and update count */
PRIVATE UINT GetChar(const _TCHAR **ppData, UINT *pRemains)
{
    UINT Count = *pRemains, val = PAD;
    const _TCHAR *pData = *ppData;
    while (Count > 0) {
        val = Decode(*pData++);
        if (val == PAD) {
            /* Reached end. Stop. */
            Count = 0;
            break;
        }

        Count--;
        if (val != SKIP)
            break;
        val = PAD;
    }
    *pRemains = Count;
    *ppData = pData;
    return val;
}

SCODE base64Coder(PTR pThis,
    /*in*/ CYPHER_OPS Operation,
    /*in*/ const BYTE * pInBuffer,
    /*in*/ UINT InSize,
    /*out*/ UINT *pSizeConsumed,
    /*out*/ BYTE * pOutBuffer,
    /*in*/ UINT OutSize,
    /*out*/ UINT *pSizeProduced)
{
    UINT d1,d2,d3,d4, count = 0, total = 0;

    if (Operation == CYPHER_ENCRYPT) { /* encode */
        _TCHAR *pOut = (_TCHAR*) pOutBuffer;
        /* Check that it fits up front.
         * 1/3 overhead rounded up to 4 bytes (1/4 overhead seen from output).
         * More overhead from \r\n (64/2)
         */
        if (__TRUNC(OutSize,4) - OutSize/4 - OutSize/32 < InSize)
            return E_BUFFER_TOO_SMALL;

        while (InSize-- > 0) {
            /* Max 76 chars per line. But 64 is simpler and ok. */
            if (count++ >= 16) { /* 4 at a time */
                *pOut++ = '\r';
                *pOut++ = '\n';
                count = 0;
                total += 66;
            }

            d1 = *pInBuffer++;
            /* 6 bits of first input */
            *pOut++ = Map[d1 >> 2];

            d1 &= 0x03;
            d1 <<= 4;
            if (InSize-- == 0) {
                *pOut++ = Map[d1];
                *pOut++ = PadChar;
                *pOut++ = PadChar;
                break;
            }

            d2 = *pInBuffer++;
            /* 2 bits of first input, 4 of second */
            *pOut++ = Map[d1 | ((d2 & 0xf0) >> 4)];

            d2 &= 0x0f;
            d2 <<= 2;
            if (InSize-- == 0) {
                *pOut++ = Map[d2];
                *pOut++ = PadChar;
                break;
            }

            d3 = *pInBuffer++;
            /* 4 bits of second input, 2 of third */
            *pOut++ = Map[d2 | ((d3 & 0xc0) >> 6)];
            /* 6 bits of third input */
            *pOut++ = Map[d3 & 0x3f];
        }

        /* Put in newline at end if chunk size was right. */
        if (count >= 16) { /* 4 at a time */
            *pOut++ = '\r';
            *pOut++ = '\n';
            total += 2;
        }

        *pSizeProduced = total + count * 4;
        if (pSizeConsumed)
            *pSizeConsumed = InSize;

    } else { /* decode */
        const _TCHAR *pIn = (const _TCHAR*) pInBuffer;
        for (;;) {
            d1 = GetChar(&pIn, &InSize);
            d2 = GetChar(&pIn, &InSize);
            if (d1 == PAD || d2 == PAD)
                break;

            if (total++ > OutSize)
                return E_BUFFER_TOO_SMALL;
            /* 6,2 */
            *pOutBuffer++ = (UINT8) ((d1 << 2) | (d2 >> 4));

            d3 = GetChar(&pIn, &InSize);
            if (d3 == PAD)
                break;

            if (total++ > OutSize)
                return E_BUFFER_TOO_SMALL;
            /* 4,4 */
            *pOutBuffer++ = (UINT8) (((d2 & 0xf) << 4) | (d3 >> 2));

            d4 = GetChar(&pIn, &InSize);
            if (d4 == PAD)
                break;

            if (total++ > OutSize)
                return E_BUFFER_TOO_SMALL;
            /* 2,6 */
            *pOutBuffer++ = (UINT8) (((d3 & 0x3) << 6) | d4);
        }

        *pSizeProduced = total;
        if (pSizeConsumed)
            *pSizeConsumed = InSize;
    }

    return S_OK;
}

