/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <util/list-double2.h>
#include <base/cobdesc.h>
#include <wsif.h>
#include <soapif.h>
#include <httpif.h>
#include <netif.h>
#include <tokenizerif.h>
#include "convert.h"
#include "deserializer.h"

#include <web/udpsoapserv.h>

typedef struct _ConverterFactory {
    struct _DefCob super;

    /* The is the offset from the local system clock to the
       authoritative wall clock (in UTC). A forward offset is 
       a positive number and indicates that the local clock is
       behind the wall clock */
    INT64 TimeOffset; 
} ConverterFactory, *PSOAPFACTORY;

/* XXX should be generated and may be wrong size XXX */
const CLASSDESC(CSoapSerializer, 3);
const CLASSDESC(CSoapSerializer2, 3);
const CLASSDESC(Deserializer, 3);
#define _MANUAL_COBMAIN

#include <s_ConverterFactory_v.c>

extern const struct XMLNAMESPACE IWsAddressing_ifd_namespace_wsa;

PRIVATE SCODE ConverterFactoryCreateSerializer (
    PICONVERT Factory,
    PIFILE Output_XML_File,
    PARSE_URL_FLAGS Flags,
    /*out*/ PISOAPSERIALIZER* serializer)
{
    return CreateSerializer(Output_XML_File, Flags, serializer);
}

PRIVATE SCODE ConverterFactoryCreateDeserializer(PICONVERT Factory,
                                        const _TCHAR * reqUri, UINT reqUriLength,
                                        PICONTINUATION pCont, /* NULL for server */
                                        PISAXHANDLER *ppHandler)
{
    struct TypeInfo *ptypeinfo;
    SCODE sc = GetTypeInfo(&ptypeinfo);
    PIHEAP Heap = _GetHeap(_T("Deserializer"), 4000);
    if (FAILED(sc)) {
        DBG(("ConverterFactoryCreateDeserializ: InitTypeInfo fail sc=%x\n",sc));
        return sc;
    }

    return CreateDeserializer(ppHandler, ptypeinfo,
                              reqUri, Heap, pCont);
}

PRIVATE SCODE ConverterFactorySetTimeOffset(PICONVERT Factory, 
                                       TIME Off,
                                       BOOL AddWhenSerializing)
{
    PSOAPFACTORY This = (PSOAPFACTORY) Factory;

#if 0 /* Fixing this will likely break any callers */
    if (!AddWhenSerializing)
        Off = Int64Subtract(TIME_ORIGIN, Off);
    if (!TimeIsNotPositive(Off))
        return E_INVALID_PARAMETER;
#endif

    This->TimeOffset = Off;
    return S_OK;
}

#define use_bax 0

#if use_bax
SCODE XBaxParse(PITOKENIZER Tokenizer, PISAXHANDLER Handler);
#else
#define XBaxParse(Tokenizer, Handler) E_NOT_IMPLEMENTED
#endif

PRIVATE SCODE ConverterFactoryProcessRequest (
    PICONVERT iThis,
    PARSE_URL_FLAGS Flags,
    PITOKENIZER Tokenizer,
    const _TCHAR *LocalUrl,
    PISAXHANDLER *pDeserializer, /* inout for restartability */
    PICONTINUATION *pCont)
{
    METHOD_DECLARE(ConverterFactory, ProcessRequest);
    PISAXPARSER SAXParser = NULL;
    PIHEAP Heap = _GetHeap(_T("Handler"), 4000);
    PICONTINUATION cont = NULL;
    struct TypeInfo *pTypeInfo = NULL;
    PDESERIALIZER Handler = NULL;

    TRACE("SOAPHTTPProcessPost");
    assert(Tokenizer != NULL);

    if (!*pDeserializer) {
        /* Fetch the Typeinfo from the input interface definition file */
        sc = GetTypeInfo(&pTypeInfo);
        CHECKSC("GetTypeInfo");

        sc = CreateDeserializer(pDeserializer, pTypeInfo,
                                NULL, Heap, NULL);
        CHECKSC("CreateDeserializer");
    }

    if (Flags & PARSE_URL_FLAGS_XTE) {
        sc = XBaxParse(Tokenizer, *pDeserializer);
    } else {
        /* Construct the SAX parser */
        sc = BindToObject(CurrentNameSpace(), _T("COB/sax.cob"),NAME_SPACE_READ,
                          &IID_ISaxParser, (void**) &SAXParser);
        CHECKSC("bind to sax.cob");

        sc = SAXParser->v->Parse(SAXParser, Tokenizer, *pDeserializer,
                                 &((PDESERIALIZER)*pDeserializer)->SaxState);
    }

    CHECKSC("Request message parsing");

    Handler = (PDESERIALIZER) *pDeserializer;
    if (Handler) {
        cont = Handler->prpcinfo;
        CHECKSETSC(cont, E_UNEXPECTED_ERROR, "exists(cont)");

        assert(pCONT(cont)->Method && pCONT(cont)->Frame
               && !pCONT(cont)->Request && !pCONT(cont)->namespaces);

        /* Save into cont what we need later */
        /* Now done here instead of deserializer.c after CreateAppStack */
        pCONT(cont)->namespaces = Handler->namespaces;
        Handler->namespaces = NULL;

        if (Handler->envelopeNS)
            pCONT(cont)->envelopeNS = Handler->envelopeNS;

        assert (Handler->ultimateReceiver == TRUE); /* routing disabled */
        *pCont = cont;
        cont->v->AddRef(cont); /* XXX might as well zap from Handler? */
    }

METHOD_CLEANUP;
    CHECK_HEAP();
    RELEASE(SAXParser);
    return sc;
}

PRIVATE SCODE ConverterFactoryProcessRequestFile (
    PICONVERT iThis,
    PARSE_URL_FLAGS Flags,
    PIFILE InputFile,
    PIFILE OutputFile,
    const _TCHAR *LocalUrl)
{
    METHOD_DECLARE(ConverterFactory, ProcessRequestFile);
    PITOKENIZER Tokenizer = NULL;
    PITOKENIZERFACTORY TokenizerFactory = NULL;
    PIUNKNOWN Deserializer = NULL;
    PICONTINUATION cont = NULL;

#if 0
    if (Flags & PARSE_URL_FLAGS_AES) {
        DBG(("soapreq: Decrypt the file...\n"));
        sc = AESBuffer(pSKey->KeyStrength, pSKey->Key,
                       Input, CYPHER_DECRYPT);
        CHECKSC("AESBuffer");
    }
#endif

    sc = BindToObject(CurrentNameSpace(), _T("COB/tokenizer.cob"), 0,
                      &IID_ITokenizerFactory, (void**) &TokenizerFactory);
    CHECKSC("bind to tokenizer.cob");

    sc = TokenizerFactory->v->Create(TokenizerFactory, (PIUNKNOWN) InputFile,
                                     &Tokenizer);
    CHECKSC("Create tokenizer");

    sc = ConverterFactoryProcessRequest(iThis, Flags, Tokenizer, LocalUrl,
                                        (PISAXHANDLER *) &Deserializer,
                                        &cont);
    CHECKSC("ProcessRequest");

    sc = cont->v->Execute(cont, FALSE, NULL);
    assert(sc == S_NO_RESPONSE || FAILED(sc));
    CHECKSC("cont->Execute");

    /* XXX If response is requested, create request obj, prepare to execute, then wait */

METHOD_CLEANUP;
    RELEASE(Tokenizer);
    RELEASE(TokenizerFactory);
    RELEASE(Deserializer);
    return sc;
}

PRIVATE SCODE ConverterFactoryProcessRequestString (
    PICONVERT iThis,
    PARSE_URL_FLAGS Flags,
    const _TCHAR *InputData,
    const _TCHAR *LocalUrl)
{
    METHOD_DECLARE(ConverterFactory, ProcessRequestString);
    PIENDPOINTFACTORY netfactory = NULL;
    PIBUFFER Input = NULL;
    PINAMESPACE ns = CurrentNameSpace();
    UINT InputDataSize = _tcslen(InputData) * sizeof(_TCHAR);

    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void **) &netfactory);
    CHECKSC("load protocol.cob");

    /* Ask network stack to create "temp file" */
    sc = netfactory->v->AllocateBuffer(netfactory, InputDataSize, NULL,
                                       &Input);
    CHECKSC("AllocateBuffer");

    /* Copy plain buffer into it */
    sc = Input->v->WriteAt(Input, UINT64_ZERO,
                           (BYTE*) InputData, InputDataSize,
                           &InputDataSize);
    CHECKSC("copy plain buffer to IBuffer");

    sc = ConverterFactoryProcessRequestFile(iThis, Flags, (PIFILE) Input,
                                       NULL, LocalUrl);
    CHECKSC("ProcessRequestFile");

METHOD_CLEANUP; 
    RELEASE(netfactory);
    RELEASE(Input);
    return sc;
}

PRIVATE SCODE ConverterFactoryProcessData(
    PICONVERT iThis,
    PARSE_URL_FLAGS Flags,
    PITOKENIZER Tokenizer,
    PISAXHANDLER *pDeserializer,
    PIHEAP Heap,
    PAnyItem Item)
{
    METHOD_DECLARE(ConverterFactory, ProcessData);
    struct TypeInfo *pTypeInfo = NULL;
    PDESERIALIZER Handler = NULL;
    PISAXPARSER SAXParser = NULL;
    CPMETHOD_DESCR metd;
    CPARGUMENT_DESCR argd;
    UINT i;

    if (!*pDeserializer) {
        /* Fetch the Typeinfo from the input interface definition file */
        sc = GetTypeInfo(&pTypeInfo);
        CHECKSC("GetTypeInfo");

        sc = CreateDeserializer(pDeserializer, pTypeInfo,
                                NULL, Heap, NULL);
        CHECKSC("CreateDeserializer");

        Handler = (PDESERIALIZER) *pDeserializer;
        Handler->current->State = DS_BODY;

        metd = FindMethod(&IWsAddressing_ifd_namespace_wsa,
                          _T("AnyItem"), FALSE);
        Handler->current->SeenItems = Handler->current->_SeenItems;

        for (i = 0; i < metd->table_size; i++) {
            argd = &metd->ptable[i];
            if (argd->descr_type == ARG_ANYTYPE)
                break;
        }

        Handler->current->href = MakeHref(&Handler->hreflist, Handler->Heap,
                                          _T("1"), 1,
                                          Item, NULL, Handler->xmldepth, NULL,
                                          argd, NULL, 0, 0);

#define STARTED     0x40000000
        Handler->SaxState = STARTED;

        /* Also make the current list href point to the same */
        Handler->current->listhref.ID = _T("#list7");
        Handler->current->listhref.addr = Item;
        Handler->current->listhref.argd = argd;

        /* XXX Undo temporary heap that was created and replace with given */
        {
            PCONTINFO cont = pCONT(Handler->prpcinfo);
            RELEASE(cont->TmpHeap);
            cont->TmpHeap = Heap;
            if (Heap)
                Heap->v->AddRef(Heap);
        }
        
    }

    if (Flags & PARSE_URL_FLAGS_XTE) {
        sc = XBaxParse(Tokenizer, *pDeserializer);
    } else {
        /* Construct the SAX parser */
        sc = BindToObject(CurrentNameSpace(), _T("COB/sax.cob"),NAME_SPACE_READ,
                          &IID_ISaxParser, (void**) &SAXParser);
        CHECKSC("bind to sax.cob");

        sc = SAXParser->v->Parse(SAXParser, Tokenizer, *pDeserializer,
                                 &((PDESERIALIZER)*pDeserializer)->SaxState);
    }

    CHECKSC("Parse");

METHOD_CLEANUP;
    CHECK_HEAP();
    RELEASE(SAXParser);
    return sc;
}

PRIVATE SCODE ConverterFactoryProcessFragmentFile (
    PICONVERT iThis,
    PIFILE InputFile,

    PIHEAP Heap,
    PTR *pData,
    _TCHAR *pType)
{
    METHOD_DECLARE(ConverterFactory, ProcessFragmentFile);
    PITOKENIZER Tokenizer = NULL;
    PITOKENIZERFACTORY TokenizerFactory = NULL;
    PINAMESPACE ns = CurrentNameSpace();
    PISAXHANDLER Deserializer = NULL;
    struct AnyItem Item;

    memset(&Item, 0, sizeof(Item));

    sc = BindToObject(CurrentNameSpace(), _T("COB/tokenizer.cob"), 0,
                      &IID_ITokenizerFactory, (void**) &TokenizerFactory);
    CHECKSC("bind to tokenizer.cob");

    sc = TokenizerFactory->v->Create(TokenizerFactory, (PIUNKNOWN) InputFile,
                                     &Tokenizer);
    CHECKSC("Create tokenizer");

    sc = ConverterFactoryProcessData(iThis, 0, Tokenizer, &Deserializer,
                                Heap, &Item);
    CHECKSC("ProcessData");

    *pData = Item.Value;
    if (Item.ValueType && pType)
        _tcscpy(pType, Item.ValueType);

    /* Do not leak ValueType string if this was a permanent heap */
    if (Item.ValueType && Heap)
        Heap->v->Free(Heap, 0, Item.ValueType);

METHOD_CLEANUP; 
    CHECK_HEAP();
    RELEASE(Deserializer);
    RELEASE(Tokenizer);
    RELEASE(TokenizerFactory);
    return sc;
}

PRIVATE SCODE ConverterFactorySerializeFragmentFile (
    PICONVERT iThis,
    PTR Data,
    const _TCHAR *Type,
    PIFILE OutputFile)
{
    METHOD_DECLARE(ConverterFactory, SerializeFragmentFile);
    PINAMESPACE ns = CurrentNameSpace();
    struct AnyItem Item;

    memset(&Item, 0, sizeof(Item));
    Item.Value = Data;
    Item.ValueType = (_TCHAR *) Type;
    ListNodeInit(offsetof(struct AnyItem, ListNode), &Item);

    sc = E_NOT_IMPLEMENTED;

METHOD_CLEANUP; 
    CHECK_HEAP();
    return sc;
}

PRIVATE SCODE ConverterFactoryProcessFragment (
    PICONVERT iThis,
    const _TCHAR *InputData,

    PIHEAP Heap,
    PTR *pData,
    _TCHAR *pType)
{
    METHOD_DECLARE(ConverterFactory, ProcessFragment);
    PIENDPOINTFACTORY netfactory = NULL;
    PIBUFFER Input = NULL;
    PINAMESPACE ns = CurrentNameSpace();
    UINT InputDataSize = _tcslen(InputData) * sizeof(_TCHAR);

    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void **) &netfactory);
    CHECKSC("load protocol.cob");

    /* Ask network stack to create "temp file" */
    sc = netfactory->v->AllocateBuffer(netfactory, InputDataSize, NULL,
                                       &Input);
    CHECKSC("AllocateBuffer");

    /* Copy plain buffer into it */
    sc = Input->v->WriteAt(Input, UINT64_ZERO,
                           (BYTE*) InputData, InputDataSize,
                           &InputDataSize);
    CHECKSC("copy plain buffer to IBuffer");

    sc = ConverterFactoryProcessFragmentFile(iThis, (PIFILE) Input, Heap,
                                        pData, pType);
    CHECKSC("ProcessFragmentFile");

METHOD_CLEANUP; 
    CHECK_HEAP();
    RELEASE(netfactory);
    RELEASE(Input);
    return sc;
}


PRIVATE SCODE ConverterFactoryGetNamespace (
    PICONVERT iThis,
    const _TCHAR * prefix,
    UINT len,
    /*out*/ CPXMLNAMESPACE* pXns)
{
    METHOD_DECLARE(ConverterFactory, GetNamespace);
    CPXMLNAMESPACE Xns;

    sc = S_OK;

    Xns = GetNamespace(prefix, len);
    CHECKSETSC(Xns, E_ENTRY_NOT_FOUND, "GetNamespace");

    *pXns = Xns;

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE ConverterFactoryFindMethod (
    PICONVERT iThis,
    CPXMLNAMESPACE pXns,
    const _TCHAR * Name,
    BOOL IsMethod,
    /*out*/ void** pMetd)
{
    METHOD_DECLARE(ConverterFactory, FindMethod);
    CPMETHOD_DESCR metd;

    sc = S_OK;

    metd = FindMethod(pXns, Name, IsMethod);
    CHECKSETSC(metd, E_ENTRY_NOT_FOUND, "FindMethod");

    *pMetd = (void*) metd;

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE ConverterFactoryDestroyNamespaces (
    PICONVERT iThis,
    PXMLNAMESPACE* pXns,
    PIHEAP Heap)
{
    METHOD_DECLARE(ConverterFactory, DestroyNamespaces);

    sc = S_OK;

    DestroyNameSpaces(pXns, Heap);

METHOD_CLEANUP;
    return sc;
}

/* SOAP box construction/destruction */
PRIVATE SCODE ConverterFactoryDestroy(PIUNKNOWN Factory)
{
    TRACE("ConverterFactoryDestroy");
    /* Nothing to do here */
    return S_OK;
}

PRIVATE ConverterFactory *pTheConverterFactory = NULL;

TIME GetSoapTimeOffset(void)
{
    return pTheConverterFactory->TimeOffset;
}

PIUNKNOWN convertCobMain(void)
{
    PIUNKNOWN cob = GenericCobInit((void*) &convert_CobDesc, NULL);
    pTheConverterFactory = (ConverterFactory *) cob;
    return cob;
}
