/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Deserialize SAX events into stack frames -jvh 2003 */
//#define _DEBUG 1
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <netif.h>

#include "convert.h"
#include "deserializer.h"
#include "tokens.h"
#include "xns.h"

#include <s_soapenc_t.h>
#include <s_soap11_t.h>
#include <s_soap12_t.h>

#ifdef _DEBUG
#define NOISE(_a_) _tprintf _a_
#else
#define NOISE(_a_) 
#endif

PRIVATE SCODE DeserializeItem(PDESERIALIZER This,
                              BOOL IsMethod, BOOL NewHref /* when applicable*/,
                              CPARGUMENT_DESCR argd,
                              const _TCHAR *Value, UINT Len, PIBUFFER Content);

/* Record current element's namespace so that empty tags and attributes are
 * resolved against it.
 */
PRIVATE SCODE SetCurrentNs(PDESERIALIZER This, CPXMLNAMESPACE Xns)
{
    struct _XMLNAMESPACEREF *Active, *Current;
    UINT i;

    /* First get rid of obsolete ones */
    for (i = This->ActiveNsCount; i > 0; i--) {
        if (This->ActiveNamespaces[i].xmldepth > This->xmldepth)
            This->ActiveNsCount--;
        else
            break;
    }

    NOISE((_T("SetCurrentNs(%s.%d)\n"), Xns->name, This->ActiveNsCount));

    if (This->ActiveNsCount >= MAXNSCOUNT)
        return E_OUT_OF_STRUCTURES;

    /* When this one is already current, never mind */
    if (This->ActiveNsCount > 0) {
        Current = &This->ActiveNamespaces[This->ActiveNsCount-1];
        if (Current->Xns == Xns)
            return S_OK;
    }

    Active = &This->ActiveNamespaces[This->ActiveNsCount++];
    
    Active->Xns = Xns;
    Active->xmldepth = This->xmldepth;
    return S_OK;
}

PRIVATE SCODE AddNamespace(PDESERIALIZER This,
                           const _TCHAR* Prefix, const _TCHAR* QName)
{
    PXMLNAMESPACE ns;
    CPXMLNAMESPACE cns;
    PIHEAP heap = This->Heap;
    UINT PrefixLength = _tcslen(Prefix);
    UINT Size = (sizeof(*ns) + PrefixLength + _tcslen(QName) +2)*sizeof(_TCHAR);
    struct _XMLNAMESPACEREF *Active;
    UINT i;

    /* First get rid of obsolete ones */
    for (i = This->ActiveNsCount; i > 0; i--) {
        if (This->ActiveNamespaces[i].xmldepth > This->xmldepth)
            This->ActiveNsCount--;
        else
            break;
    }

    if (This->ActiveNsCount >= MAXNSCOUNT)
        return E_OUT_OF_STRUCTURES;
    Active = &This->ActiveNamespaces[This->ActiveNsCount++];

    Active->any.Type = ENTRY_TYPE_XNS_REF;
    Active->any.Flags = ENTRY_FLAGS_CONST;
    Active->any.RefCnt = 1;

    /* Check for duplicates and discard this one if any */
    for (cns = This->namespaces; cns; cns = cns->next)
        if (!_tcscmp(cns->name, Prefix) && !_tcscmp(cns->QName, QName)) {
            Active->Xns = cns;
            Active->xmldepth = This->xmldepth;
            /* XXX save prefix */
            return S_FALSE;
        }

    ns = heap->v->Alloc(heap,
                        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("deser:ns"),
                        Size, 0);
    if (ns == NULL)  return E_NOT_ENOUGH_MEMORY;

    /* Stuff name and qname after header. NULs to terminate. */
    ns->name = (void*)(ns + 1);
    _tcscpy((void*) ns->name, Prefix);
    ns->QName = ns->name + PrefixLength + 1;
    _tcscpy((void*) ns->QName, QName);

    /* LookupNamespace resolves */
    ns->nstype = NSTYPE_UNRESOLVED;

    /* Link up */
    ns->next = This->namespaces;
    This->namespaces = ns;

    Active->Xns = ns;
    Active->xmldepth = This->xmldepth;

    return S_OK;
}

CPXMLNAMESPACE FindXnsByURL(CPNS_TABLE nst, const _TCHAR *value)
{
    UINT16 i;
    CPXMLNAMESPACE xe;

    assert(nst && nst->any.Type == ENTRY_TYPE_NS_TABLE);

    for (i = 0; i < nst->Count; i++) {
        /* XXX binary search! */
        xe = (CPXMLNAMESPACE) nst->Table[i];
        if (xe->any.Type != ENTRY_TYPE_XNS)
            continue;

        if (!_tcscmp(value, xe->QName)) {
            return xe;
        }
    }
    return NULL;
}

CPXMLNAMESPACE FindXnsByPrefix(CPNS_TABLE nst, const _TCHAR *value, UINT len)
{
    UINT16 i;
    CPXMLNAMESPACE xe;

    assert(nst && nst->any.Type == ENTRY_TYPE_NS_TABLE);

    for (i = 0; i < nst->Count; i++) {
        /* XXX binary search! */
        xe = (CPXMLNAMESPACE) nst->Table[i];
        if (xe->any.Type != ENTRY_TYPE_XNS)
            continue;

        if (_tcsncmp(value, xe->name, len))
            continue;

        /* Require complete match of predefined prefix */
        if (_tcslen(xe->name) != len)
            continue;

        return xe;
    }
    return NULL;
}

PRIVATE CPXMLNAMESPACE LookupNamespace(PDESERIALIZER This, const _TCHAR* Prefix)
{
    CPXMLNAMESPACE cns, ns2;
    PXMLNAMESPACE ns1;

    /* First resolve local name to QName */
    for (cns = This->namespaces; cns; cns = cns->next)
        if (!_tcscmp(cns->name, Prefix))
            goto Found;
    /* Default default namespace */
    if (_tcslen(Prefix) == 0)
        return This->ptypeinfo->defaultns;

#if 1
    return FindXnsByPrefix(pTheNamespaceNamespace, Prefix, _tcslen(Prefix));
#else
    /* Not found. Look in predefined as second chance (XTE skips when appr.) */
    for (ns2 = This->ptypeinfo->namespaces; ns2; ns2 = ns2->next)
        if (!_tcscmp(ns2->name, Prefix)) {
            if (ns2->nstype == NSTYPE_UNRESOLVED)
                return NULL;
            return ns2;
        }
    return NULL;
#endif

 Found:
    /* Next look for QName in predefined namespaces */
    if (cns->nstype == NSTYPE_UNRESOLVED) {
#if 1
        ns2 = FindXnsByURL(pTheNamespaceNamespace, cns->QName);
            if (ns2)
#else
        for (ns2 = This->ptypeinfo->namespaces; ns2; ns2 = ns2->next)
            if (!_tcscmp(ns2->QName, cns->QName))
#endif
            {
                /* Make ns resolved */
                ns1 = (PXMLNAMESPACE) cns;/* get rid of 'const' -> ok UNRES */
                ns1->nstype = ns2->nstype;
                ns1->id = ns2->id;
                ns1->uri_length = ns2->uri_length;
                ns1->ifd = ns2->ifd;
                ns1->entries = ns2->entries;
                return cns;
            }
        DBG(("Note: unrecognized namespace QName %s", cns->QName));
    }
    return cns;
}

CPXMLNAMESPACE LookupNsQName(PISAXHANDLER iThis, PSTR_ENTRY QName)
{
#if 1
    return FindXnsByURL(pTheNamespaceNamespace, QName->Str);
#else
    PDESERIALIZER This = pDeserializer(iThis);
    CPXMLNAMESPACE ns;

    /* Look for QName in predefined namespaces */
    for (ns = This->ptypeinfo->namespaces; ns; ns = ns->next)
            if (!_tcscmp(ns->QName, QName->Str))
                return ns;
    return ns;
#endif
}

void DestroyNameSpaces(PXMLNAMESPACE *ppns, PIHEAP Heap)
{
    PXMLNAMESPACE ns, next;

    for (ns = *ppns; ns; ns = next) {
        next = (PXMLNAMESPACE) ns->next;
        Heap->v->Free(Heap, 0, ns);
    }

    *ppns = NULL;
}

#ifdef __ADSPBLACKFIN__
/* bf compiler chokes when optimizing... */
#pragma optimize_off
#endif
PRIVATE SCODE ResolveName(PDESERIALIZER This,
                          /* inout */ PSAXATTRIBUTE Entry)
{
    UINT i, j;
    CPXMLNAMESPACE Xns;
    CPNS_TABLE nst;
    CPSTR_ENTRY te;

    assert(Entry->Token == TOKEN_invalid);

    Xns = LookupNamespace(This, Entry->Prefix);
    if (!Xns) {
        DBG(("ResolveName: ns '%s' invalid\n", Entry->Prefix));
        return E_NAME_NOT_FOUND;
    }

    Entry->Xns = Xns;

    nst = (CPNS_TABLE) Xns->entries;
    if (!nst || nst->any.Type != ENTRY_TYPE_NS_TABLE)
        return E_NAME_NOT_FOUND;

    for (i = 0; i < nst->Count; i++) {
        /* XXX binary search! */
        te = (CPSTR_ENTRY) nst->Table[i];
        if (te->any.Type != ENTRY_TYPE_STRING)
            continue;

        if (!_tcscmp(Entry->Name, te->Str)) {
            Entry->Token = i;
            return S_OK;
        }
    }

    if (_tcslen(Entry->Prefix) != 0) {
        DBG(("ResolveName: '%s' not in ns '%s'\n", Entry->Name, Entry->Prefix));
        return E_NAME_NOT_FOUND;
    }

    /* Try in current namespace for unqualified names (as per content model) */
    if (This->ActiveNsCount == 0) {
        DBG(("ResolveName: '%s', with no default ns\n", Entry->Name));
        return E_NAME_NOT_FOUND;
    }

    for (j = This->ActiveNsCount; j > 0; j--) {
        CPXMLNAMESPACE Xns1 = This->ActiveNamespaces[j-1].Xns;
        assert(Xns1);
        Xns = LookupNamespace(This, Xns1->name); /* argh, resolve NS */
        assert(Xns);

        nst = (CPNS_TABLE) Xns->entries;
        assert(nst && nst->any.Type == ENTRY_TYPE_NS_TABLE);

        for (i = 0; i < nst->Count; i++) {
            /* XXX binary search! */
            te = (CPSTR_ENTRY) nst->Table[i];
            if (te->any.Type != ENTRY_TYPE_STRING)
                continue;

            if (!_tcscmp(Entry->Name, te->Str)) {
                Entry->Token = i;
                Entry->Xns = Xns;
                /* XXX should check for content model, unqualified,
                 * XXX but they are currently lumped together 
                 * XXX and furthermore it might be ok
                 * XXX to go both ways -> check XML spec.
                 */
                return S_OK;
            }
        }
    }

//assert(FALSE);
#if 0
    for (i = 0; i < Xns->StringCount; i++)
        if (!_tcscmp(Entry->Name, Xns->Strings[i])) {
            Entry->Token = i;
            return S_OK;
        }

    /* XXX Look in MoreStrings */

    if (Xns->nstype != NSTYPE_ANY) {
        DBGT((_T("deserializer: Did not find name '%s' in namespace %s=%s\n"),
             Entry->Name, Xns->name, Xns->QName));
        return E_NAME_NOT_FOUND;
    }

    for (j = 0; j < This->ptypeinfo->DefaultCount; j++) {
        Xns = This->ptypeinfo->DefaultXns[j];
        for (i = 0; i < Xns->StringCount; i++)
            if (!_tcscmp(Entry->Name, Xns->Strings[i])) {
                DBGT((_T("Found non-qualified token %s in default xns %s\n",
                     Entry->Name, Xns->name));
                Entry->Xns = Xns;
                Entry->Token = i;
                return S_OK;
            }
    }

    if (spare) {
        Xns = spare;

        Xns = &pre_ns_3;
        for (i = 0; i < Xns->StringCount; i++)
            if (!_tcscmp(Name, Xns->Strings[i])) {
                DBGT((_T("Found non-qualified token %s in soapenc xns %s\n"),
                     Name, Xns->name));
                return i;
            }
    }

    /* Try to match empty prefixes with the next defined namespace */
    for (Xns = This->namespaces;
         Xns && Xns->xmldepth > This->xmldepth;
         Xns = Xns->next)
        ;
    if (!Xns)
        return NO_TOKEN;

    /* Next look for QName in predefined namespaces */
    for (ns2 = This->ptypeinfo->namespaces; ns2; ns2 = ns2->next)
        if (!_tcscmp(ns2->QName, Xns->QName)) {
            Xns = ns2;
            break;
        }

    for (i = 0; i < Xns->StringCount; i++)
        if (!_tcscmp(Name, Xns->Strings[i])) {
            DBGT((_T("Found non-qualified token %s in namespace %s\n"),
                 Name, Xns->name));
            return i;
        }

    /* XXX Look in MoreStrings */

#endif
    DBGT((_T("Could not resolve non-qualified name %s ns=%s\n"),
         Entry->Name, Xns->name));

    return E_NAME_NOT_FOUND;
}


CPMETHOD_DESCR FindMethod(CPXMLNAMESPACE Xns, const _TCHAR *Name, BOOL IsM)
{
    CPMETHOD_DESCR metd;
    CPINTERFACE_DESCR ifd;
    UINT    methodcount;
    UINT    counter;

    //DBGT((_T("FindMethod: %s\n"), Name));
    for (ifd = Xns->ifd; ifd; ifd = ifd->iface_extras->SuperClass) {
        methodcount = ifd->table_size;
        metd = ifd->ptable;
        for (counter = 0; counter < methodcount; counter++, metd++) {
            /* Skip entries of wrong type */
            if (IsM) {
                if (metd->descr_type != DESCR_TYPE_METHOD
                    && metd->descr_type != DESCR_TYPE_METHOD_RESPONSE)
                    continue;
            } else {
                if (metd->descr_type != DESCR_TYPE_STRUCT)
                    continue;
            }
            if (!_tcscmp(Name, metd->name))
                return metd;
        }
    }

    return NULL;
}

PRIVATE PHREF FindHref(PHREF phref, const _TCHAR *value, INT valueSize)
{
    while (phref) {
        if (valueSize >= 0 && _tcsnequal(value,valueSize, phref->ID+1, 0))
            return phref;
        else if (_ttoi(phref->ID) == -valueSize)
            return phref;
        phref = phref->next;
    }
    return NULL;
}

#if 0 /* not used */
/* Find matching href and zap it */
PRIVATE void DeleteHref(PHREF *pphref, const _TCHAR *value, UINT valueSize, PIHEAP Heap)
{
    PHREF phref;
    for (phref = *pphref; phref; pphref = &phref->next, phref = *pphref) {
        if (_tcsnequal(value,valueSize, phref->ID+1, 0)) {
            *pphref = phref->next;
            Heap->v->Free(Heap, 0, phref);
            return;
        }
    }
}
#endif

PHREF MakeHref(PHREF *pphref, PIHEAP Heap,
               const _TCHAR *value, UINT valueLength,
               PTR addr, PTR data, UINT xmldepth,
               PUINT pStoredSize,
               CPARGUMENT_DESCR argd, PHREF parent,
               ARGUMENT_DESCR_FLAGS Flags, UINT id)
{
    PHREF phref;
    UINT slen = valueLength+1;

    if (!value)
        slen = 20;       /* for "id%d" */

    phref = Heap->v->Alloc(Heap,
                           HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("deser:href"),
                           sizeof(**pphref) + slen * sizeof(_TCHAR),
                           0);
    if (phref == NULL)  return NULL;

    if (value) {
        phref->ID = (_TCHAR *) (phref + 1);
        memcpy(phref->ID, value, slen * sizeof(_TCHAR));
        phref->ID[valueLength] = _EOS;
    } else if (id != 0) {
        phref->ID = (_TCHAR *) (phref + 1);
        _stprintf(phref->ID, _T("id%d"), id);
    }
    phref->addr = addr;
    phref->data = data;
    phref->pStoredSize = pStoredSize;
    phref->argd = argd;
    phref->xmldepth = xmldepth;
    if (argd) {
        phref->structd = argd->_structd;
        /* store flags but get in/out from original */
        phref->flags = ((Flags&(ARGUMENT_DESCR_FLAGS_IN|ARGUMENT_DESCR_FLAGS_OUT|ARGUMENT_DESCR_FLAGS_ISRESPONSE))
                        | argd->_Flags);
        phref->type = argd->descr_type;
        phref->ListNodeOffset = argd->_ListNodeOffset;
    } else if (parent) {
        phref->structd = parent->structd;
        /* store flags but get in/out from original + remember if response */
        phref->flags = ((Flags&(ARGUMENT_DESCR_FLAGS_IN|ARGUMENT_DESCR_FLAGS_OUT|ARGUMENT_DESCR_FLAGS_ISRESPONSE))
                        | parent->flags);
        phref->type = parent->type;
        phref->ListNodeOffset = parent->ListNodeOffset;
        phref->Xns = parent->Xns;
    }

    if (!value)
        phref->ArrayIndex = valueLength;

    if (pphref) {
        if (value && valueLength == 5 && !_tcscmp(value, _T("#list"))) {
            /* Insert first */
            phref->next = *pphref;
            *pphref = phref;
        } else {
            /* last */
            while (*pphref)
                pphref = &(*pphref)->next;
            *pphref = phref;
        }
    }

    return phref;
}

void DestroyHRef(PHREF phref, PIHEAP Heap)
{
    Heap->v->Free(Heap, 0, phref);
}

void DestroyHReflist(PHREF *pphref, PIHEAP Heap)
{
    PHREF phref, tmp;
    for (phref  = *pphref; phref; phref = tmp) {
        tmp = phref->next;
        Heap->v->Free(Heap, 0, phref);
    }
    *pphref = NULL;
}

PRIVATE SCODE AllocComplex(PHREF phref, UINT ArraySize, PICONTINUATION cont)
{
    /* Allocate space for struct, perhaps array. On stack. */
    CPARGUMENT_DESCR argd = phref->argd;
    PTR p;
    UINT Size;

    if (phref->pStoredSize)
        *phref->pStoredSize = ArraySize;

    /* is it already allocated? */
    /* Also allocate if return path with dynamic typing & size
     * See also ActivationSpaceSetBuffer.
     */
    if (phref->data && (void *) pCONT(cont)->ArgSpace != phref->data) {
        if (phref->addr) /* is there a dangling pointer to it? */
            *(char**)phref->addr = phref->data;
    } else {
        /* One for struct, many for array */
        if (argd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY) {
            assert(ArraySize > 0);
            switch (argd->descr_type) {
            case ARG_STRUCT:
                Size = argd->_structd->_FrameSize * ArraySize;
                break;
            case ARG_INT:
            case ARG_FLOAT:
                Size = sizeof(INT) * ArraySize;
                break;
            case ARG_STRING:
                Size = (sizeof(ADDRESS) * ArraySize) * sizeof(_TCHAR);
                break;
            default:
                Size = 0;
                DebugBreak();
                return E_INVALID_PARAMETER;
            }
        } else { /* struct */
            assert(ArraySize == 0);
            if (argd->_structd)
                Size = argd->_structd->_FrameSize;
            else
                Size = phref->structd->_FrameSize;
        }

        p = cont->v->Alloc(cont, HEAP_ZERO_MEMORY, Size, 0);
        assert(p);
        if (!p)
            return E_NOT_ENOUGH_MEMORY;

        if (phref->addr) {
            if (argd->descr_type == ARG_LISTHEAD) {
                ListAddLast(phref->addr, argd->_ListNodeOffset, p);
                NOISE((_T("AllocComplex(List a=%x off=%x p=%x)\n"),
                       phref->addr, argd->_ListNodeOffset, p));
            } else {
                *(char**)phref->addr = p;
            }
        }
        phref->data = p;
    }
    return S_OK;
}

SCODE base64(PTR pThis,
    /*in*/ CYPHER_OPS Operation,
    /*in*/ const BYTE * pInBuffer,
    /*in*/ UINT InSize,
    /*out*/ UINT *pSizeConsumed,
    /*out*/ BYTE * pOutBuffer,
    /*in*/ UINT OutSize,
    /*out*/ UINT *pSizeProduced)
{
    METHOD_DECLARE_NOTHIS(deserializer.c, base64);
    PICYPHERCONSTRUCTOR b64f = NULL;
    PICYPHER b64 = NULL;
    PIUNKNOWN unk;
    PINAMESPACE ns = CurrentNameSpace();

    sc = BindToObject(ns, _T("COB/base64.cob"), NAME_SPACE_READ,
                      &IID_ICypherConstructor, (void **) &b64f);
    CHECKSC("load base64.cob");

    sc = b64f->v->Constructor(b64f, &unk, NULL, 0);
    CHECKSC("Construct");

    sc = unk->v->QueryInterface(unk, &IID_ICypher, (void **) &b64);
    unk->v->Release(unk);
    CHECKSC("QueryInterface");

    sc = b64->v->Cypher(b64, Operation,
                        pInBuffer, InSize, pSizeConsumed,
                        pOutBuffer, OutSize, pSizeProduced);
    CHECKSC("Cypher (base64 decode)");

METHOD_CLEANUP;
    RELEASE(b64);
    RELEASE(b64f);

    return sc;
}

PRIVATE UINT DeserializeSimpleType(DESCR_TYPE argtype, UINT FLags,
                                   PTR where,
                                   const _TCHAR *value, UINT valueLength,
                                   UINT MaxSize)
{
    switch (argtype) {
    case ARG_INT: {
        UINT* addr = where;
        assert(MaxSize >= sizeof(INT));
        if (MaxSize >= sizeof(INT)) {
            if (!value)
                *addr = 0;
            else
                *addr = _ttoin(value, valueLength);
        }
        return sizeof(INT); 
    }
    case ARG_FLOAT: {
        UINT* addr = where;
        assert(MaxSize >= sizeof(INT));
        if (MaxSize >= sizeof(INT)) {
            if (!value || _tcsnequal(_T("NaN"),3, value,valueLength))
                *addr = INT_MAX;   /* "NaN" */
            else if (_tcsnequal(_T("INF"),3, value,valueLength))
                *addr = INT_MAX-1;   /* one below the NaN */
            else if (_tcsnequal(_T("-INF"),4, value,valueLength))
                *addr = INT_MIN;   /* oh well, this is at best a joke */
            else
                *addr = _ttoin(value, valueLength);
        }
        return sizeof(INT); 
    }
    case ARG_INT64: {
        INT64* addr = where;
        _TCHAR tmp[24]; /* There are 20 digits in a 64 bit decimal number */

        assert(MaxSize >= sizeof(UINT64));
        if (MaxSize >= sizeof(UINT64)) {
            if (!value) {
                *addr = INT64_ZERO;
            } else if (valueLength > 20) {
                *addr = TIME_FOREVER; /* XXX UINT64_MAX */
            } else {
                _tcsncpy(tmp, value, valueLength);
                tmp[valueLength] = 0;
                *addr = _ttoi64(tmp);
            }
        }
        return sizeof(UINT64); 
    }
    case ARG_BOOL: {
        BOOL* addr = where;
        assert(MaxSize >= sizeof(BOOL));
        if (MaxSize >= sizeof(BOOL)) {
            /* Treat all unknown values (along with "true" and "1") as TRUE */
            /* "false", "0", and "" are treated as false */
            if (!value || valueLength == 0)
                *addr = FALSE;
            else if (_tcsnequal(_T("false"),5, value,valueLength)
                     || _tcsnequal(_T("0"),1, value,valueLength))
                *addr = FALSE;
            else
                *addr = TRUE;
        }
        return sizeof(BOOL);
    }

    case ARG_DATETIME: {
        SCODE sc;
        TIME *addr = where;
        TIME t;

        sc = TimeParse(value, &t);
        assert(SUCCEEDED(sc));
        if (FAILED(sc))
            return sc;

        if (TimeIsPositive(t)) {
            TIME t2 = GetSoapTimeOffset();
            t = Int64Subtract(t, t2);
            /* Make sure we don't under/overflow */
            if (Int64IsNegative(t)) {
                if (Int64IsNegative(t2))
                    t = TIME_FOREVER;
                else
                    t = TIME_ORIGIN;
            }
        }
        //PrintTime("Deserialized time", t);
        assert(MaxSize >= sizeof(TIME));
        if (MaxSize >= sizeof(TIME)) {
            *addr = t;
        }
        return sizeof(TIME);
    }

    case ARG_STRING: /* given as pointer */
    case ARG_QNAME:
    case ARG_URI: {
        _TCHAR *addr = where;
        /* Check MaxSize. */
        assert(MaxSize >= sizeof(_TCHAR));
        assert(MaxSize >= sizeof(_TCHAR) * valueLength + sizeof(_TCHAR));

        if (MaxSize < sizeof(_TCHAR))
            ;
        else if (valueLength * sizeof(_TCHAR) + sizeof(_TCHAR) > MaxSize)
            MaxSize = MaxSize/sizeof(_TCHAR) - sizeof(_TCHAR);
        else
            MaxSize = valueLength;

        if (MaxSize >= sizeof(_TCHAR)) {
            if (value)
                _tcsncpy(addr, value, MaxSize);
            addr[MaxSize] = _T('\0');
        }
        return (valueLength + 1) * sizeof(_TCHAR);
    }
    case ARG_BLOB:
        if (value) {
            base64(NULL, CYPHER_DECRYPT,
                   (PTR) value, valueLength,
                   NULL,
                   where, MaxSize,
                   &valueLength);
        }
        return valueLength;

    default: /* given as pointers */
        DebugBreak();
        return sizeof(ADDRESS);
    }
}

PRIVATE SCODE DeserializeHref(PDESERIALIZER This, CPARGUMENT_DESCR itemd,
                              const _TCHAR * value, UINT valueLength,
                              PHREF phref, BOOL IsNewHRef, PIBUFFER Content)
{
    PICONTINUATION cont = This->prpcinfo;
    DESCR_TYPE type = phref->argd->descr_type;
    UINT offset = This->current->arrayindex;
    char *data = phref->data;

    /* See if this is a struct item */
    if (itemd) {
        assert(data);
        /* Calculate correct spot to put data in
         * Nullifies scaling below;
         */
        if (phref->argd->_structd)
            data += phref->argd->_structd->_FrameSize * offset
                + itemd->FrameOffset;
        else
            data += phref->structd->_FrameSize * offset
                + itemd->FrameOffset;

        offset = 0;
        type = itemd->descr_type;
    }

    if (IsNewHRef && itemd) {
        if (itemd->descr_type == ARG_LISTHEAD
            || (itemd->descr_type == ARG_STRUCT && !value))
        {
            SCODE sc = S_OK;
            phref = &This->current->listhref;
            /* List head must be zeroed before use */
#if 0 /* tmp allocs are zeroed. User must zero presented heads -> i.e.
       * not necessary here so don't. Nameless listheads continue on
       * existing ones.
       */
            *(PLIST_HEAD) data = NULL; //Where = NULL;
#endif
            assert(!phref->ID);
            phref->ID = _T("#list2");
            phref->addr = data;
            phref->argd = itemd;
            phref->xmldepth = This->xmldepth;
            This->current->href = phref;

            /* Structs need to be allocated, do it now (there iss one) */
            if (itemd->descr_type != ARG_LISTHEAD) {
                assert ((itemd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY) == 0);
                sc = AllocComplex(phref, 0, cont);
            }
            return sc;
        }

        /* Nested array or struct inside struct. or struct pointer... */
        if (itemd->_Flags & ARGUMENT_DESCR_FLAGS_BYREF) {
            assert(value);
            if (!MakeHref(&This->hreflist, This->Heap,
                          value, valueLength,
                          data, NULL, This->xmldepth, NULL,
                          itemd, phref, phref->flags, 0))
                return E_NOT_ENOUGH_MEMORY;
        } else {
            if (!MakeHref(&This->hreflist, This->Heap,
                          value, valueLength,
                          NULL, data, This->xmldepth, NULL,
                          itemd, phref, phref->flags, 0))
                return E_NOT_ENOUGH_MEMORY;
        }
        return S_OK;
    }

    switch(type) {
    case ARG_DATETIME: {
        TIME* addr = (TIME*)data + offset;
        DeserializeSimpleType(type, 0, addr,
                              value, valueLength, sizeof(TIME));
        break;
    }
    case ARG_FLOAT:
    case ARG_BOOL:
    case ARG_INT: {
        UINT* addr = (UINT*)data + offset;
        DeserializeSimpleType(type, 0, addr,
                              value, valueLength, sizeof(INT));
        break;
    }
    case ARG_STRING:
    case ARG_QNAME:
    case ARG_URI: {
        _TCHAR** addr = (_TCHAR**)data + offset;
        if (!(This->current->SavedState & DSS_HEADER)
            && (phref->flags & ARGUMENT_DESCR_FLAGS_ISRESPONSE)
            && *addr != NULL) {
            /* XXX Check size against size-is */
            _tcsncpy(*addr, value, valueLength);
            (*addr)[valueLength] = '\0';
        } else {
            /* Allocate space on stack and return value */
            /* XXX Check for return path where caller forgot to init struct */
            _TCHAR *p = cont->v->Alloc(cont, HEAP_ZERO_MEMORY,
                    __ROUND((valueLength+1)*sizeof(_TCHAR), 4), 0);
            if (!p)
                return E_NOT_ENOUGH_MEMORY;
            _tcsncpy(p, value, valueLength);
            p[valueLength] = '\0';
            *addr = p;
        }
        break;
    }
    case ARG_IBUFFER: {
        PIBUFFER* addr = (PIBUFFER*)data + offset;
        if (Content) {
            Content->v->AddRef(Content);
        } else {
            /* Got the argument delivered as XML content.
             * Convert it to IBuffer.
             */
            PIENDPOINTFACTORY epf = NULL;
            SCODE sc;
            sc = BindToObject(CurrentNameSpace(), _T("COB/protocol.cob"),
                              NAME_SPACE_READ, &IID_IEndpointFactory,
                              (void **) &epf);
            assert(SUCCEEDED(sc));

            sc = epf->v->AllocateBuffer(epf, valueLength, NULL, &Content);
            assert(SUCCEEDED(sc));

            sc = ReadWriteFile((PIFILE) Content, TRUE,
                               UINT64_MAX, (void*) value, valueLength, NULL);
            assert(SUCCEEDED(sc));

            RELEASE(epf);
        }
        *addr = Content;

        break;
    }

    case ARG_STRUCT: {
        CPARGUMENT_DESCR argd = phref->argd;
        PTR addr = (PTR) ((char*)data + offset * argd->_structd->_FrameSize);

        if (!MakeHref(&This->hreflist, This->Heap,
                      value, valueLength,
                      NULL, addr, This->xmldepth, NULL, argd, phref, phref->flags, 0))
            return E_NOT_ENOUGH_MEMORY;
        break;
    }
    case ARG_LISTHEAD: {
        SCODE sc;
        CPARGUMENT_DESCR structd = itemd->_structd;

        This->current++;
        if (This->current >= &This->_current[MAX_STRUCT_NESTING])
            return E_OUT_OF_STRUCTURES;
        memset(This->current, 0, sizeof(*This->current));
        This->current->SeenItems = This->current->_SeenItems;

        This->current->href = phref;
        This->current->State = DS_STRUCT;
        This->current->listhref.argd = itemd;

        sc = DeserializeItem(This, FALSE, TRUE, itemd, NULL, 0,
                             NULL);

        This->current->State = DS_LIST;
        This->current->href = NULL;

        sc = This->v->Text((PISAXHANDLER) This, value, valueLength, NULL, 0); 

        This->current--;

        return sc;
    }
    default:
        DebugBreak();
    }

    return S_OK;
}

/* Allocate and fill in space for server side "out" arguments */
PRIVATE SCODE FillOutArgument(PDESERIALIZER This, CPARGUMENT_DESCR argd,
                             PUINT pStoredSize)
{
    PICONTINUATION cont = This->prpcinfo;
    UINT Flags = argd->_Flags;
    UINT sizeis = 0; /* The allocation size for a string or similar */
    char *Frame = (char *) pCONT(cont)->Frame;
    char *MyFrameSlot = Frame + argd->FrameOffset;

    assert(Flags & ARGUMENT_DESCR_FLAGS_OUT);
    assert(!(Flags & ARGUMENT_DESCR_FLAGS_IN) || (argd->descr_type == ARG_LISTHEAD));

    if (Flags & ARGUMENT_DESCR_FLAGS_ARRAY) {
        ADDRESS *array;
        if (argd->ArraySize) 
            sizeis = argd->ArraySize;
        else
            sizeis = 50;       /* XXX part of interface */

        /* Store array length if indicated */
        if (pStoredSize)
            *pStoredSize = sizeis;

        /* XXX populate strings for output struct arrays */
        if (argd->descr_type == ARG_INT || argd->descr_type == ARG_FLOAT) {
            array = cont->v->Alloc(cont, HEAP_ZERO_MEMORY,
                                   sizeof(INT)*sizeis, 0);
        } else if (argd->descr_type == ARG_STRING
                   || argd->descr_type == ARG_ANYTYPE
                   || argd->descr_type == ARG_QNAME) {
            UINT j;
            UINT strsize = 100; // XXX the max size of a sting cannot exceed 100;
            array = cont->v->Alloc(cont, HEAP_ZERO_MEMORY,
                                   sizeof(ADDRESS)*sizeis, 0);

            for (j = 0; j < sizeis; j++) {
                array[j] = (ADDRESS) cont->v->Alloc(cont,
                        HEAP_ZERO_MEMORY, __ROUND(strsize+1, 4), 0);
            }
        } else {
            assert(argd->descr_type == ARG_STRUCT);
            array = cont->v->Alloc(cont, HEAP_ZERO_MEMORY,
                                   argd->_structd->_FrameSize*sizeis, 0);
        }

        /* Push pointer to array into frame */
        *(ADDRESS*) MyFrameSlot = (ADDRESS)array;

        /* Push size-of and length-is unless the array is fixed size */
        if (argd->SizeIsArg)
            *(UINT*) (Frame + (argd + argd->SizeIsArg)->FrameOffset) = sizeis;
        if (argd->LengthIsArg) {
            if (!pStoredSize)
                pStoredSize = cont->v->Alloc(cont,
                                            HEAP_ZERO_MEMORY, sizeof(UINT), 0);
            *pStoredSize = sizeis;
            
            *(PTR*) (Frame + (argd + argd->LengthIsArg)->FrameOffset) = pStoredSize;
        }
    } else if (argd->descr_type == ARG_LISTHEAD) {
        /* Give it some space and push pointer */
        ADDRESS *addr = cont->v->Alloc(cont, HEAP_ZERO_MEMORY, sizeof(PTR), 0);
        *(ADDRESS**) MyFrameSlot = addr;
    } else if (argd->_structd) { /* struct */
        /* out straight struct, no arrays, no hrefs */
        CPMETHOD_DESCR structd;
        ADDRESS *addr;
        UINT i;

        structd = argd->_structd;
        assert(structd);

        /* Give it some space and push pointer */
        addr = cont->v->Alloc(cont, HEAP_ZERO_MEMORY, structd->_FrameSize, 0);
        *(ADDRESS**) MyFrameSlot = addr;

        /* Allocate space for any strings */
        for (i = 0; i < structd->table_size; i++) {
            CPARGUMENT_DESCR itemd = &structd->ptable[i];
            if (itemd->descr_type == ARG_STRING
                || itemd->descr_type == ARG_URI)
            {
                UINT strsize = 100 * sizeof(_TCHAR); // XXX the max size of a string cannot exceed 100;
                addr[i] = (ADDRESS) cont->v->Alloc(cont,
                                   HEAP_ZERO_MEMORY, __ROUND(strsize+1, 4), 0);
            }
        }
    } else {
        /* simple type */
        void *array;

        if (argd->descr_type == ARG_INT
            || argd->descr_type == ARG_BOOL
            || argd->descr_type == ARG_FLOAT) {
            array = cont->v->Alloc(cont, HEAP_ZERO_MEMORY, sizeof(INT), 0);
            *(INT*) array = 66667777;
        } else if (argd->descr_type == ARG_INT64
                   || argd->descr_type == ARG_DATETIME) {
            array = cont->v->Alloc(cont, HEAP_ZERO_MEMORY, sizeof(INT64), 0);
            *(INT64*) array = TIME_FOREVER;
        } else if (argd->descr_type == ARG_BLOB) {
            sizeis = 500;
            array = cont->v->Alloc(cont, HEAP_ZERO_MEMORY, sizeis, 0);
        } else if (argd->descr_type == ARG_STRING
                   || argd->descr_type == ARG_URI) {
            sizeis = 100;       /* XXX check for @size-is */
            array = cont->v->Alloc(cont, HEAP_ZERO_MEMORY,
                                   __ROUND(sizeis * sizeof(_TCHAR), 4), 0);
        } else if (argd->descr_type == ARG_ANYTYPE) {
            sizeis = 500;
            array = cont->v->Alloc(cont, HEAP_ZERO_MEMORY, sizeis, 0);
            if (argd->TypeIsArg) {
                _TCHAR **ti = (_TCHAR**)(Frame+(argd+argd->TypeIsArg)->FrameOffset);
                UINT typesizeis = 100;       /* XXX check for @size-is */
                *ti = cont->v->Alloc(cont, HEAP_ZERO_MEMORY,
                                   __ROUND(typesizeis * sizeof(_TCHAR), 4), 0);
            }
        } else {
            assert(FALSE);
            array = NULL;
        }

        /* Push pointer to array into frame */
        *(ADDRESS*) MyFrameSlot = (ADDRESS)array;

        /* Push size-of and length-is unless the array is fixed size */
        if (argd->SizeIsArg)
            *(UINT*) (Frame + (argd + argd->SizeIsArg)->FrameOffset) = sizeis;
        if (argd->LengthIsArg) {
            if (!pStoredSize)
                pStoredSize = cont->v->Alloc(cont, HEAP_ZERO_MEMORY,
                                             sizeof(UINT), 0);
            if (!pStoredSize)
                return E_NOT_ENOUGH_MEMORY;
            *pStoredSize = 0;
            *(PTR*) (Frame + (argd + argd->LengthIsArg)->FrameOffset) = pStoredSize;
        }
    }
    return S_OK;
}

PRIVATE SCODE DeserializeArgument(PDESERIALIZER This,
                                  CPARGUMENT_DESCR argd, BOOL IsResponse,
                                  const _TCHAR * value, UINT valueLength,
                                  PIBUFFER Content)
{
    PICONTINUATION cont = This->prpcinfo;
    char *Frame = (char *) pCONT(cont)->Frame;
    char *MyFrameSlot = Frame + argd->FrameOffset;
    PTR Where;
    UINT Flags = argd->_Flags;
    UINT sizeis = 0, actual; /* The allocation size for a string or similar */
    PUINT pStoredSize = NULL; 

    /* Disallow client from overriding size-of/length-of arguments.
     * This is for (1) allocation purposes (2) to avoid allocating conflicting
     * length-is slots (#2 could be taken care of).
     */
    if (Flags & ARGUMENT_DESCR_FLAGS_MODIFIER_OF)
        return S_FALSE;

     /* Figure out where the actual argument data is */
    if (IsResponse) {
        assert(Flags & ARGUMENT_DESCR_FLAGS_OUT);
        if (Flags & ARGUMENT_DESCR_FLAGS_RETURN)
            Where = &pCONT(cont)->RetVal;
        else
            Where = *(PTR*) MyFrameSlot;
    } else {
        if ((Flags & ARGUMENT_DESCR_FLAGS_OUT)
            && (argd->LengthIsArg || (Flags & ARGUMENT_DESCR_FLAGS_ARRAY)))
        {
            /* Store actual size in save area after argument pointer */
            //pStoredSize = (ADDRESS*) &cont->pSaveArea[argd->FrameOffset];
            /* Allocate space for size information */
            pStoredSize = cont->v->Alloc(cont,
                                         HEAP_ZERO_MEMORY, sizeof(UINT), 0);
        }

#if 0
        /* out, inout: push pointer to value. in: push value */        
        if (Flags & (ARGUMENT_DESCR_FLAGS_OUT|ARGUMENT_DESCR_FLAGS_BYREF))
            Where = NULL; //ActivationSpaceReserve(cont);
        else
#endif
            /* Data allocated later when needed */
            Where = MyFrameSlot;

        /* Store pointer into save area as needed.
         * This is done for the benefit of caller saved arguments
         * and the need to pick up the pointer after the call.
         */
        //if (argd->SaveIndex)
        //    *(ADDRESS*) &cont->pSaveArea[argd->FrameOffset] = (ADDRESS) Where;
    }

    if (argd->descr_type == ARG_LISTHEAD) {
        PHREF phref = &This->current->listhref;
        memset(phref, 0, sizeof(*phref));
        /* List head must be zeroed before use */
        *(PLIST_HEAD) Where = NULL;
        assert(!phref->ID);
        phref->ID = _T("#list");
        phref->addr = Where;
        assert(!pStoredSize);
        phref->xmldepth = This->xmldepth;
        phref->argd = argd;
    } else if (argd->descr_type == ARG_CONTINUATION) {
        PICONTINUATION *addr = Where;
        assert(Where);
        *addr = cont;
    } else if (argd->descr_type == ARG_STRUCT
               || argd->descr_type == ARG_ANYTYPE
               || (Flags & ARGUMENT_DESCR_FLAGS_ARRAY))
    {
        if (IsResponse) {
            /* Fetch sizeof if provided */
            if (argd->SizeIsArg && argd->descr_type == ARG_ANYTYPE) {
                /* Take the buffer to be a mem pool for variable allocation */
                SCODE sc;
                UINT *si = (UINT*)(Frame+(argd+argd->SizeIsArg)->FrameOffset);
                sizeis = *si;
                sc = cont->v->SetBuffer(cont, Where, sizeis);
                if (FAILED(sc))
                    return sc;
                /* If this was just a buffer setting arg, do not make href */
                if (argd->_Flags & ARGUMENT_DESCR_FLAGS_NONE)
                    return S_OK;
            }

            if (valueLength == 5 && !_tcscmp(value, _T("!LIST"))) {
                PHREF phref = &This->current->listhref;
                /* List head must be zeroed before use */
                assert(!phref->ID);
                memset(phref, 0, sizeof(*phref));
                phref->ID = _T("!LIST");
                phref->addr = MyFrameSlot;
                phref->data = Where;
                phref->pStoredSize = pStoredSize;
                phref->argd = argd;
                phref->xmldepth = This->xmldepth;
                phref->type = argd->descr_type;
                phref->flags = (argd->_Flags|ARGUMENT_DESCR_FLAGS_ISRESPONSE) & (ARGUMENT_DESCR_FLAGS_IN|ARGUMENT_DESCR_FLAGS_OUT|ARGUMENT_DESCR_FLAGS_ISRESPONSE);
                // ETC XXX;
            } else if (!MakeHref(&This->hreflist, This->Heap,
                                 value, valueLength,
                                 MyFrameSlot, Where,
                                 This->xmldepth, pStoredSize,
                                 argd, NULL,
                                 (ARGUMENT_DESCR_FLAGS) (argd->_Flags
                                       | ARGUMENT_DESCR_FLAGS_ISRESPONSE), 0))
                return E_NOT_ENOUGH_MEMORY;
        } else if (Flags & ARGUMENT_DESCR_FLAGS_IN) {
            if (valueLength == 5 && !_tcscmp(value, _T("!LIST"))) {
                PHREF phref = &This->current->listhref;
                /* List head must be zeroed before use */
                assert(!phref->ID);
                memset(phref, 0, sizeof(*phref));
                phref->ID = _T("!LIST");
                phref->addr = MyFrameSlot;
                phref->data = NULL;
                phref->pStoredSize = pStoredSize;
                phref->argd = argd;
                phref->xmldepth = This->xmldepth;
                phref->type = argd->descr_type;
                phref->flags = ((argd->_Flags&(ARGUMENT_DESCR_FLAGS_IN|ARGUMENT_DESCR_FLAGS_OUT|ARGUMENT_DESCR_FLAGS_ISRESPONSE)));
                // ETC XXX;
            } else {
                PHREF phref = MakeHref(&This->hreflist, This->Heap,
                                       value, valueLength,
                                       MyFrameSlot, NULL, This->xmldepth, pStoredSize,
                                       argd, NULL, argd->_Flags, 0);
                if (value && !phref)
                    return E_NOT_ENOUGH_MEMORY;
                if (argd->SelectorIsArg) {
                    PTR *zi = (PTR*)(Frame+(argd+argd->SelectorIsArg)->FrameOffset);
                    /* XXX Should lazy evaluate until correct size is known */
                    UINT32 *z;
                    z = cont->v->Alloc(cont, HEAP_ZERO_MEMORY,
                                       sizeof(UINT32)*MAX_SELECTOR, 0);
                    *zi = z;
                    phref->FieldSelector = z;
                }
            }
        } else {
            assert(!IsResponse);
            FillOutArgument(This, argd, pStoredSize);
        }
    } else {
        DESCR_TYPE Type = argd->descr_type;
        if (Type == ARG_ANYTYPE) {
            DebugBreak();
            Type = ARG_STRING;  /* default */
        }

        /* Default size for item is in the typeinfo */
        sizeis = argd->ArraySize;
        if (Type == ARG_BLOB && !IsResponse && sizeis == 0)
            sizeis = 1000;      /* XXX depends on stack space */
        if (sizeis == 0)
            sizeis = 8;

        if (IsResponse) {
            /* Fetch sizeof if provided */
            if (argd->SizeIsArg) {
                UINT *si = (UINT*)(Frame+(argd+argd->SizeIsArg)->FrameOffset);
                sizeis = *si;
            }
        }

        if (!IsResponse
            && (Flags & (ARGUMENT_DESCR_FLAGS_OUT|ARGUMENT_DESCR_FLAGS_BYREF)))
        {
            /* Pre-allocate an estimated size (should be max possible) */
            sizeis = valueLength * sizeof(_TCHAR) + sizeof(UINT64);
            Where = cont->v->Alloc(cont,
                                   HEAP_ZERO_MEMORY, __ROUND(sizeis, 4), 0);
        }
        actual = DeserializeSimpleType(Type, 0, Where,
                                       value,valueLength, sizeis);
        if (!IsResponse) {
            /* max size in out or inout */
            if (sizeis < actual || !(Flags & ARGUMENT_DESCR_FLAGS_OUT))
                sizeis = actual;

            if (Flags & (ARGUMENT_DESCR_FLAGS_OUT|ARGUMENT_DESCR_FLAGS_BYREF)){
                /* Push pointer to value. Shrink buffer to actual size. */
                *(ADDRESS*) MyFrameSlot = (ADDRESS) cont->v->Realloc(cont,
                     HEAP_REALLOC_IN_PLACE_ONLY, Where, __ROUND(sizeis, 4), 0);
            }
            /* Push sizeis/lengthis/typeis/selector arguments if requested. */
            if (argd->SizeIsArg) {
                UINT *si = (UINT*)(Frame+(argd+argd->SizeIsArg)->FrameOffset);
                *si = sizeis; /* max size when OUT */
            }
            if (argd->LengthIsArg && (Flags & ARGUMENT_DESCR_FLAGS_OUT)) {
                PTR* li = (PTR*)(Frame+(argd+argd->LengthIsArg)->FrameOffset);
                assert(pStoredSize);
                *li = pStoredSize;
            }
            if (argd->TypeIsArg) {
                PTR *ti = (PTR*)(Frame+(argd+argd->TypeIsArg)->FrameOffset);
                *ti = NULL;
            }
            if (argd->SelectorIsArg) {
               PTR *zi = (PTR*)(Frame+(argd+argd->SelectorIsArg)->FrameOffset);
                /* XXX Should figure out correct size (fieldcount or 1) */
                UINT32 *z;
                assert(Flags & ARGUMENT_DESCR_FLAGS_OUT); /* in untested */
                z = cont->v->Alloc(cont, HEAP_ZERO_MEMORY,
                                   sizeof(UINT32)*MAX_SELECTOR, 0);
                *zi = z;
            }
            if (pStoredSize)
                *pStoredSize = actual;
        } else {
            /* Return length if requested */
            if (argd->LengthIsArg && (Flags & ARGUMENT_DESCR_FLAGS_OUT)) {
                PUINT *li = (PUINT*)(Frame+(argd+argd->LengthIsArg)->FrameOffset);
                **li = actual;
            }
            /* Return type if requested */
            if (argd->TypeIsArg && (Flags & ARGUMENT_DESCR_FLAGS_OUT)) {
                PTR **ti = (PTR**)(Frame+(argd+argd->TypeIsArg)->FrameOffset);
                **ti = NULL;
            }
            /* Return selector as BOOL when requested */
            if (argd->SelectorIsArg && (Flags & ARGUMENT_DESCR_FLAGS_OUT)) {
                UINT32 **zi;
                assert(Flags & ARGUMENT_DESCR_FLAGS_OUT); /* in untested */
                zi = (UINT32**)(Frame+(argd+argd->SelectorIsArg)->FrameOffset);
                /* If this was an empty arg the value is "", if missing it is
                 * NULL. Use that to determine whether present.
                 */
                if (value)
                    **zi = TRUE;
                else
                    **zi = FALSE;
            }
        }
    }
    return S_OK;
}

/* Helper for dealing with common code */
PRIVATE SCODE DeserializeItem(PDESERIALIZER This,
                              BOOL IsMethod, BOOL NewHref /* when applicable*/,
                              CPARGUMENT_DESCR argd,
                              const _TCHAR *Value, UINT Len, PIBUFFER Content)
{
    if (IsMethod)
        return DeserializeArgument(This, argd,
            (This->current->SavedState & DSS_HEADER) ? FALSE: This->isResponse,
            Value, Len, Content);
    else
        return DeserializeHref(This, argd, Value, Len,
                               This->current->href, NewHref, Content);
}


/* Go through all arguments that have not been seen and fill them in */
PRIVATE SCODE FillMissingItems(PDESERIALIZER This, BOOL IsResponse)
{
    METHOD_DECLARE_NOTHIS(Deserializer, FillMissingItems);
    CPMETHOD_DESCR metd = This->current->metd;
    UINT argix;
    CPARGUMENT_DESCR argd;

    sc = S_OK;
    assert(metd->table_size/8 < sizeof(This->current->_SeenItems));

    for (argix = 0; argix < metd->table_size; argix++) {
        argd = &metd->ptable[argix];

        /* Has it been seen yet? */
        if (This->current->SeenItems[argix/32] & (1 << argix%32)) {
            if (argd->descr_type == ARG_CONTINUATION)
                return E_SCHEMA_CHECK;
            continue;
        }

        /* See if there are @wsa:To args in the method.
         * transmitted after the '?' in the wsa:To value.
         * Both ATTRIBUTE and TEXT flags are set.
         */
        if ((This->isResponse ? (argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT)
                              : (argd->_Flags & ARGUMENT_DESCR_FLAGS_IN))
            && (argd->_Flags & ARGUMENT_DESCR_FLAGS_ATTRIBUTE)
            && (argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT))
        {
            UINT len;

            CHECKSETSC(This->UriQuery, E_PARSE_ERROR,"find arg after URI");
            len = _tcscspn(This->UriQuery, _T("="));
            CHECKSETSC(This->UriQuery[len], E_PARSE_ERROR,"find URI arg  '=' ");
            /* XXX Only handle one arg in URI for now */
            if (_tcsncmp(argd->name, This->UriQuery, len)
                || _tcslen(argd->name) != len)
                CHECKSETSC(FALSE, E_PARSE_ERROR,"match URI arg name");
            sc = DeserializeItem(This, TRUE, FALSE, argd,
                                 This->UriQuery + len + 1,
                                 _tcslen(This->UriQuery) - len - 1,
                                 NULL);
            /* Mark the arg as received */
            This->current->SeenItems[argix/32] |= (1 << argix%32);
            continue;
        }

        /* No filling for the return value, etc */
        if (argd->_Flags & (ARGUMENT_DESCR_FLAGS_RETURN
                            |ARGUMENT_DESCR_FLAGS_MODIFIER_OF
                            |ARGUMENT_DESCR_FLAGS_NONE))
            continue;

        /* Skip "in" and "inout" arguments on client response path */
        if (IsResponse && (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT)
                           || (argd->_Flags & ARGUMENT_DESCR_FLAGS_IN)))
            continue;

        /* Skip lists */
        if (argd->descr_type == ARG_LISTNODE)
            continue;

        /* Deal with @serialize-as="!HEADER" */
        if (argd->_Flags & ARGUMENT_DESCR_FLAGS_HEADER) {
            PICONTINUATION cont = This->prpcinfo;
            char *Frame = (char *) pCONT(cont)->Frame;
            char *MyFrameSlot = Frame + argd->FrameOffset;
            PWSCONTEXT pContext = pCONT(cont)->pContext;

            assert ((argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT) == 0);
            assert(!IsResponse);

            /* XXX So far only deal with a few specific well-known headers */
            /* XXX ugly & checks needed */
            /* XXX OPTIONAL */
            if (!_tcscmp(argd->name, _T("ResourceURI"))) {
                _TCHAR **addr = (void *) MyFrameSlot;
                *addr = pContext->ResourceURI;
            } else if (!_tcscmp(argd->name, _T("SelectorSet"))) {
                PSelector_wsa *addr = (void *) MyFrameSlot;
                *addr = pContext->SelectorSet;
            } else if (!_tcscmp(argd->name, _T("OperationTimeout"))) {
                TIME *addr = (void *) MyFrameSlot;
                *addr = pContext->OperationTimeout;
            } else if (!_tcscmp(argd->name, _T("To"))) {
                _TCHAR **addr = (void *) MyFrameSlot;
                PEndpointReference To = pContext->To;
                if (!To)
                    return E_SCHEMA_CHECK;
                *addr = To->Address;
            } else {
                DBGT((_T("deserializer: Unsupport !HEADER arg '%s' in '%s'\n"),
                      argd->name, metd->name));
                return E_NOT_IMPLEMENTED;
            }
            continue;
        }

        /* Arg is actually missing */
        if ((argd->descr_type != ARG_CONTINUATION)
            && (argd->_Flags & (IsResponse ? ARGUMENT_DESCR_FLAGS_OUT
                                           : ARGUMENT_DESCR_FLAGS_IN)))
        {
            /* Check for invalid omission (BOOL text false is non-existant) */
            if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_OPTIONAL)
                && !(argd->descr_type == ARG_BOOL
                     && (argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT)))
            {
                DBGT((_T("deserializer: Missing mandatory arg '%s' in '%s'\n"),
                     argd->name, metd->name));
                assert(argd->_Flags & ARGUMENT_DESCR_FLAGS_OPTIONAL);
                return E_SCHEMA_CHECK;
            }
//            if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT))
//                continue;       /* optional in: OK.  (init on server?) */
        }

        /* Missing optional "out" or "inout" argument */
        if (This->current->State == DS_METHOD) {
            DeserializeArgument(This, argd, IsResponse, NULL, 0, NULL);
            if (!IsResponse && (argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT))
                FillOutArgument(This, argd, NULL);
        }
    }

    /* Reset SeenItems for next round */
METHOD_CLEANUP;
    memset(This->current->_SeenItems, 0, sizeof(This->current->_SeenItems));
    This->current->SeenItems = This->current->_SeenItems;
    return sc;
}

PRIVATE DESCR_TYPE ResolveXsdType(const _TCHAR *Name)
{
    if (!_tcscmp(Name, _T("int")))
        return ARG_INT;
    else if (!_tcscmp(Name, _T("string")))
        return ARG_STRING;
    else if (!_tcscmp(Name, _T("dateTime")) || !_tcscmp(Name, _T("duration")))
        return ARG_DATETIME;
    else if (!_tcscmp(Name, _T("boolean")))
        return ARG_BOOL;
    else if (!_tcscmp(Name, _T("float")))
        return ARG_FLOAT;
    else if (!_tcscmp(Name, _T("decimal")))
        return ARG_INT64;
    else if (!_tcscmp(Name, _T("IID")))
        return ARG_GUID;
    else if (!_tcscmp(Name, _T("anyURI")))
        return ARG_URI;
    else if (!_tcscmp(Name, _T("base64Binary")))
        return ARG_BLOB;
    else 
        return ARG_STRUCT;
}

PRIVATE SCODE DeserializeArray(PDESERIALIZER This,
                               CPSAXATTRIBUTE Attr, UINT AttrCount)
{
    CPXMLNAMESPACE Xns;
    CPARGUMENT_DESCR argd;
    PHREF phref = NULL;

    /* Deal with mandatory @id and @ArrayType */
    This->current->arraytype = DESCR_TYPE_ANY;
    for ( ; Attr && AttrCount > 0; Attr++, AttrCount--)
        if (!_tcscmp(Attr->Name, _T("id"))) {
            phref = FindHref(This->hreflist,
                             Attr->Value, _tcslen(Attr->Value));
            if (!phref)
                return E_SCHEMA_CHECK;
            This->current->href = phref;
        } else if ((Attr->Xns->id == NS_soapenc_ifd_namespace_soapenc)
                   && !_tcscmp(Attr->Name, _T("arrayType")))
        {
            _TCHAR *type, *size, *xsd;

            /* XXX should use tmp buffer XXX */
            xsd = (_TCHAR*) Attr->Value;
            type = _tcssplit(xsd, _T(":"));
            if (!type)
                return E_SCHEMA_CHECK;
            /* type must be xsd qualified type */
            Xns = LookupNamespace(This, xsd);
            if (!Xns)
                return E_SCHEMA_CHECK;
            /* Either xsd or xsi (really this interface) */
            if (Xns->nstype != NSTYPE_STRUCT
                && Xns->nstype != NSTYPE_INTERFACE)
                return E_SCHEMA_CHECK;

            size = _tcssplit(type, _T("["));
            if  (!size)
                return E_PARSE_ERROR;
            if (_tcssplit(size, _T("]")) == NULL)
                return E_PARSE_ERROR;

            This->current->arraytype = ResolveXsdType(type);
            This->current->arraysize = _ttoi(size);
            This->current->arrayindex = -1; /* preincremented */
            if (This->current->arraytype == ARG_STRUCT)
                This->current->metd = Xns->ifd;
        }
    /* Make sure @id and @ArrayType were indeed present */
    if (!phref || (This->current->arraytype == DESCR_TYPE_ANY))
        return E_SCHEMA_CHECK;
    argd = phref->argd;

    /* XXX inout arrays broken. Need to consolidate in+out codes */
    assert(This->isResponse || !(argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT));
    /* An array must be an array */
    assert(argd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY);

    /* Check that the xsd type matches the typeinfo.
     * If the size is fixed, also check if it matches.
     * For nonarray, argd->ArraySize is zero.
     */
    if (This->current->arraytype != argd->descr_type)
        return E_SCHEMA_CHECK; // XXX should define a new SCODE;
    if (!argd->SizeIsArg && This->current->arraysize != argd->ArraySize)
        return E_SCHEMA_CHECK;

    /* Allocate stack space for array
     * and initialize size-is and length-is values
     */
    if (phref->addr) {
        PICONTINUATION cont = This->prpcinfo;
        SCODE sc = AllocComplex(phref, This->current->arraysize, cont);
        if (FAILED(sc))
            return sc;
        if (phref->pStoredSize)
            *phref->pStoredSize = This->current->arraysize;
        if (argd->SizeIsArg) {/* the array size is variable */
            /* Figure out where the size-is arg is.
             * Start from MyFrameSlot (...->addr)
             */
            ADDRESS *si;
            si = (ADDRESS*)((char*)phref->addr - argd->FrameOffset
                            + (argd+argd->SizeIsArg)->FrameOffset);
            assert(((ADDRESS*)phref->addr + 1) == si);
            if (!(This->current->SavedState & DSS_HEADER)
                && (phref->flags & ARGUMENT_DESCR_FLAGS_ISRESPONSE)) {
                if (*si < This->current->arraysize) {
                    TRACE("Response array too large");
                    return E_INVALID_PARAMETER;
                }
            } else {
                *si = This->current->arraysize;
            }
        }
        if (!(This->current->SavedState & DSS_HEADER)
            && This->isResponse && argd->LengthIsArg)
        {
            PUINT *li;
            li =(PUINT*)((char*)phref->addr - argd->FrameOffset
                         +(argd+argd->LengthIsArg)->FrameOffset);
            **li = This->current->arraysize;
        }
    } else {
        /* Array nested within struct. Must be fixed size. [XXX last]*/
        assert(!argd->SizeIsArg && argd->ArraySize > 0);
    }
    return S_OK;
}

/* Post-process wsa:To and related headers. Called at beginning of <Body>. */
PRIVATE SCODE ResolveWsaHeaders(PDESERIALIZER This)
{
    METHOD_DECLARE_NOTHIS(Deserializer, ResolveWsaHeaders);
    PICONTINUATION tmpCont = NULL;
    PWSCONTEXT pCtx;

    sc = S_OK;

    /* Check for a WS-Addressing derived destination. 
     * The wsa:To header overrides the HTTP destination
     */
    pCtx = pCONT(This->prpcinfo)->pContext;
    if (pCtx) {
        if (pCtx->To && pCtx->To->Address) {
            _TCHAR *Local;
            UINT LocalLen;
            UINT UriQueryLength;
            const _TCHAR *UriQuery;
            PISOAPFACTORY soapfactory;
            sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"),
                              NAME_SPACE_READ, &IID_ISoapFactory,
                              (void**) &soapfactory);
            CHECKSC("bind to soap.cob");

            sc = soapfactory->v->ParseURL(soapfactory, pCtx->To->Address,
                                          NULL, NULL, NULL, NULL,
                                          (void *) &Local, &LocalLen,
                                          &UriQuery, &UriQueryLength,
                                          NULL);
            soapfactory->v->Release(soapfactory);
            CHECKSC("ParseURL wsa:To");

            /* Overwrite the HTTP-derived destination (if any)
             * with the local URL from the addressing header
             */
            if (This->ShouldFreeURL) 
                free((void*) This->reqLocalURL);
            This->reqLocalURL = _tcsndup(Local, LocalLen);
            This->ShouldFreeURL = TRUE;
            if (This->UriQuery)
                free((void*) This->UriQuery);
            This->UriQuery = _tcsndup(UriQuery, UriQueryLength);
        }

        if (pCtx->Action) {
            /* This can be of various formats. The only one of consequence
             * is "methodsNS/methodName" where the method is nameless.
             * The method wrapper is omitted in the Body, so special
             * parsing is required.
             * We also accept "methodsNS:methodName" for the benefit of DSSP.
             */
            CPXMLNAMESPACE ns;
            UINT len, i, xnsix, namelessindex = 1;
            _TCHAR *s;
            CPMETHOD_DESCR metd;
            CPARGUMENT_DESCR argd;

            /* Look for QName in predefined namespaces */
#if 1
            for (xnsix = 0; xnsix < pTheNamespaceNamespace->Count; xnsix++) {
                ns = (CPXMLNAMESPACE) pTheNamespaceNamespace->Table[xnsix];
                if (ns->any.Type != ENTRY_TYPE_XNS)
                    continue;
#else
            for (ns = This->ptypeinfo->namespaces; ns; ns = ns->next) {
#endif
                len = _tcslen(ns->QName);
                s = pCtx->Action;
                if (!_tcsncmp(ns->QName, s, len)) {
                    s += len;
                    if (*s != _T('/') && *s != _T(':'))
                        break;
                    s++;
                    metd = FindMethod(ns, s, TRUE);
                    /* If the Action is anyway malformed, ignore it */
                    if (!metd)
                        break;
                    if ((metd->any.Flags & ENTRY_FLAGS_NAMELESS) == 0)
                        break;
                    assert (metd->descr_type == DESCR_TYPE_METHOD
                            || metd->descr_type == DESCR_TYPE_METHOD_RESPONSE);

                    /* wsa:To is mandatory here */
                    if (!This->reqLocalURL)
                        return E_SCHEMA_CHECK;

                    if (This->current->SavedState & DSS_HEADER) {
                        DBG(("nameless method in header not allowed\n"));
                        DebugBreak();
                        CHECKSETSC(FALSE, E_SCHEMA_CHECK,
                                   "nameless method in header not allowed");
                    }

                    /*** This is an implicit method call ***/

                    /* Reply path: not much to do here */
                    if (This->isResponse){
                        /* If we allocated a temporary cont for header
                         * processing, get rid of it.Delay until we process
                         * the URI though.
                         */
                        tmpCont = This->prpcinfo;
                        This->prpcinfo = iCONT(This->pCont); /* use saved cont */
                        if (This->pCont) /* will be released later */
                            This->pCont->v->AddRef(iCONT(This->pCont));

                        /* Check that this is the reply that was expected
                         * by the continuation.
                         */
                        if (metd->NameToken != pCONT(This->prpcinfo)->metd->ResponseNameToken)
                        {
                            DBG(("deserializer: reply mismatch %s (for %s)\n",
                                 metd->name,
                                 pCONT(This->prpcinfo)->metd->name));
                            assert(!_tcscmp(metd->name, _T("Fault")));
                            CHECKSETSC(FALSE, E_SCHEMA_CHECK, "match reply");
                        }
                    } else {
                        sc = This->prpcinfo->v->SetMethod(This->prpcinfo,
                             ns, metd, This->reqLocalURL, NULL);
                        CHECKSC("SetMethod");
                    }

                    /* See if there are nameless args inside the nameless
                     * method.
                     */
                    for (i = 0; i < metd->table_size; i++) {
                        argd = &metd->ptable[i];

                        if (This->isResponse) {
                            if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT))
                                continue;
                        } else {
                            if (!(argd->_Flags & ARGUMENT_DESCR_FLAGS_IN))
                                continue;
                        }

                        /* Nameless args must be untyped structs */
                        /* Also none so as to process any client buffer */
                        if (argd->descr_type != ARG_BOOL &&
                            (argd->_Flags
                            & (ARGUMENT_DESCR_FLAGS_TEXT
                               |ARGUMENT_DESCR_FLAGS_NONE)) &&
                            !(argd->_Flags & (ARGUMENT_DESCR_FLAGS_ATTRIBUTE
                                              | ARGUMENT_DESCR_FLAGS_HEADER)))
                        {
                            _TCHAR buf[10];
                            /* Mark arg as received XXX check we get struct */
                            This->current->SeenItems[i/32] |= (1 << i%32);
                            This->current->argd = argd;

                            /* "nameless" ref */
                            _stprintf(buf, _T("%d"), namelessindex);
                            namelessindex++;
                            sc = DeserializeItem(This, TRUE, TRUE, argd,
                                                 buf, _tcslen(buf), NULL);
                            CHECKSC("deserialize nameless arg pseudo-ref");
                        }
                    }
                    /* Mess with state machine: go directly to DS_METHOD */
                    assert(This->current->State == DS_BODY);
                    This->current->metd = metd;
                    pCONT(This->prpcinfo)->methodNS = ns;
                    This->current->State = DS_METHOD;
                    break;
                }
            }
        }

        /* Save sendTime from received context when present */
        NOISE((_T("*********************************** set sendTime\n")));
        pCONT(This->prpcinfo)->sendTime = pCtx->sendTime;
    }

METHOD_CLEANUP;
    /* Release tmp cont if one was consumed */
    RELEASE(tmpCont);
    return sc;
}

PRIVATE SCODE SetTypeIs(PDESERIALIZER This, PHREF phref,
                        CPXMLNAMESPACE Xns, CPMETHOD_DESCR structd)
{
    CPARGUMENT_DESCR argd = phref->argd;
    PICONTINUATION cont = This->prpcinfo;
    char *pArg = phref->addr;
    _TCHAR **ti;
    UINT len;
    CPXMLNAMESPACE ns2;

    /* Canonicalize name to standard prefix */
    ns2 = FindXnsByURL(pTheNamespaceNamespace, Xns->QName);
    if (ns2)
        Xns = ns2; /* Replace with builtin */

    len = _tcslen(Xns->name) + _tcslen(structd->name) + 2;

    /* Deliver "xnsPrefix:structName" to type-is arg */
    assert(argd->TypeIsArg);
    ti = (void*) (pArg + (argd+argd->TypeIsArg)->FrameOffset - argd->FrameOffset);
    if ((phref->flags & ARGUMENT_DESCR_FLAGS_ISRESPONSE) == 0)
        *ti = cont->v->Alloc(cont, HEAP_ZERO_MEMORY, len * sizeof(_TCHAR), 0);
    /* else check for fit when @size-is */

    if (!*ti)
        return E_NOT_ENOUGH_MEMORY;

    _stprintf(*ti, _T("%s:%s"), Xns->name, structd->name);
#if 0
    /* The content is allocated from the activation heap */
    phref->flags &= ~ARGUMENT_DESCR_FLAGS_ISRESPONSE;
    /* Now that we know what the type is let's subtract the size
     * from the buffer, which got set as cont->ArgSpace.
     */
    ActivationSpaceExtract(cont,);
#endif
    return S_OK;
}

/* Process resolved entries */
SCODE RStartElement(PISAXHANDLER iThis, CPSAXATTRIBUTE Elem,
                    CPSAXATTRIBUTE Attr, UINT AttrCount)
{
    PDESERIALIZER This = pDeserializer(iThis);
    SCODE sc = S_OK;
    CPXMLNAMESPACE Xns = Elem->Xns;
    const _TCHAR *Name = Elem->Name;
    CPMETHOD_DESCR metd, metd2;
    CPARGUMENT_DESCR argd;
    PHREF phref = NULL;
    BOOL IsNamelessArg = FALSE; /* for assertion only */

    switch (This->current->State) {
    case DS_EMPTY:
        if (Xns->id != NS_soap11_ifd_namespace_soap11
            && Xns->id != NS_soap12_ifd_namespace_soap)
            return E_SCHEMA_CHECK;
        if (!_tcscmp(Name, _T("Envelope")))
            This->current->State = DS_ENVELOPE;
        else
            return E_SCHEMA_CHECK;
        This->envelopeNS = Xns; /* remember SOAP version for reply */
        return S_OK;
    case DS_ENVELOPE:
        if (Xns->id != NS_soap11_ifd_namespace_soap11
            && Xns->id != NS_soap12_ifd_namespace_soap)
            return E_SCHEMA_CHECK;
        if (!_tcscmp(Name, _T("Header"))) {
            This->current->State = DS_HEADER;
        } else if (!_tcscmp(Name, _T("Body"))) {
            /* ROLE ultimateReceiver: handle it. Else pass through */
            This->current->State = DS_BODY;
            if (This->ultimateReceiver) {
                /* Post-process wsa:To and wsa:Action and related headers */
                sc = ResolveWsaHeaders(This);
                if (FAILED(sc))
                    return sc;
            } else {
                DBGT((_T("deserializer: ROLE pass through Body\n")));
                /* This subtree will now be delegated. */
                This->RoleAction = RA_forward;
                This->skiptoxmldepth = This->xmldepth;
            }
        } else
            return E_SCHEMA_CHECK;
        return S_OK;
    case DS_HEADER:
        /* Record that we need to get back to header mode */
        This->current->SavedState |= DSS_HEADER;
        /* XXX Role check */
        goto DS_BODY;
    DS_BODY:
    case DS_BODY:
    case DS_LIST:
        /* An array? */
        if (Xns->id == NS_soapenc_ifd_namespace_soapenc) {
            if (_tcscmp(Name, _T("Array")))
                return E_SCHEMA_CHECK;

            This->current->State = DS_ARRAY;
            return DeserializeArray(This, Attr, AttrCount);
        }

        /* A method? */
        if (!Xns->ifd)
          return E_INVALID_METHOD;

        assert (Xns->nstype==NSTYPE_INTERFACE || Xns->nstype==NSTYPE_STRUCT);
        if (Xns->nstype != NSTYPE_INTERFACE && Xns->nstype != NSTYPE_STRUCT)
            return E_SCHEMA_CHECK;

        if (This->current->State == DS_LIST
            || This->current->State == DS_STRUCT
            || IsNamelessArg)
            metd = NULL;
        else
            metd = FindMethod(Xns, Name, TRUE);

        if (metd && (metd->descr_type == DESCR_TYPE_METHOD
                     || metd->descr_type == DESCR_TYPE_METHOD_RESPONSE))
        {
            This->current->State = DS_METHOD;
            This->current->metd = metd;

            /* This is the real method: not much more to do in reply path */
            if (!(This->current->SavedState & DSS_HEADER) && This->isResponse){
                UINT i;
                /* If we allocated a temporary cont for header processing,
                 * get rid of it now.
                 */
                RELEASE(This->prpcinfo);
                This->prpcinfo = iCONT(This->pCont); /* use saved continuation */

                /* Check that this is the reply that was expected by the
                 * continuation.
                 */
                if (metd->NameToken
                    != pCONT(This->prpcinfo)->metd->ResponseNameToken)
                {
                    DBG(("deserializer: reply mismatch %s (for %s)\n",
                         metd->name, pCONT(This->prpcinfo)->metd->name));
                    assert(!_tcscmp(metd->name, _T("Fault")));
                    return E_SCHEMA_CHECK;
                }

                /* See if there is an argument for setting temp
                 * allocation buffer.
                 */
                for (i = 0; i < metd->table_size; i++) {
                    argd = &metd->ptable[i];
                    if (argd->descr_type == ARG_ANYTYPE
                        && argd->SizeIsArg
                        && (argd->_Flags & ARGUMENT_DESCR_FLAGS_NONE))
                        DeserializeArgument(This, argd, TRUE, NULL, 0, NULL);
                }

                return S_OK; 
            }

            /* Ok, this is a request.
             * We need to allocate a stack and initialize it
             */
            if (This->current->SavedState & DSS_HEADER) {
                /* In header, use deserializer object, synch execution */
                sc = This->prpcinfo->v->SetMethod(This->prpcinfo,
                                            Xns, metd, NULL, (PIUNKNOWN) This);
            } else {
                /* If no URI was provided through http or routing/addressing, abort */
                if (!This->reqLocalURL)
                    return E_NAME_NOT_FOUND;

                //assert(This->reqLocalURL[0]);
                sc = This->prpcinfo->v->SetMethod(This->prpcinfo,
                                                  Xns, metd,
                                                  This->reqLocalURL, NULL);
            }
            if (FAILED(sc))
                return sc;

            pCONT(This->prpcinfo)->methodNS = Xns;

            /* See if there are attributes that match input arguments
             * Ignore any excess attributes
             * XXX should match namespace too!
             * XXX should we require ARGUMENT_DESCR_FLAGS_ATTRIBUTE ?
             */
            for ( ; Attr && AttrCount > 0; Attr++, AttrCount--) {
                UINT i;
                for (i = 0; i < metd->table_size; i++) {
                    if (!_tcscmp(metd->ptable[i].name, Attr->Name)) {
                        argd = &metd->ptable[i];

                        /* Mark the arg as received */
                        assert(i/8 < sizeof(This->current->_SeenItems));
                        /* XXX Check for duplicates? */
                        assert(!(This->current->SeenItems[i/32] & (1 << i%32)));
                        This->current->SeenItems[i/32] |= (1 << i%32);

                        sc = DeserializeItem(This, TRUE, FALSE, argd,
                                             Attr->Value, _tcslen(Attr->Value),
                                             NULL);
                    }
                }
            }
            return sc;
        }

        /* Struct then? */
        metd = FindMethod(Xns, Name, FALSE);
        if (!metd)
            return E_SCHEMA_CHECK;
StructFound:
        assert(metd->descr_type == DESCR_TYPE_STRUCT);
        metd2 = metd;

        assert(!phref || (metd->any.Flags & ENTRY_FLAGS_NAMELESS) || (argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT)); /* for now, must be inline */
        if (This->current->State != DS_LIST &&
            (metd->any.Flags & ENTRY_FLAGS_NAMELESS) == 0)
        {
            /* Deal with mandatory @id and XXX @type */
            UINT i;
            CPSAXATTRIBUTE a = Attr;
            for (i = AttrCount; a && i > 0; a++, i--) {
                if (!_tcscmp(a->Name, _T("id"))) {
                    PHREF h  = FindHref(This->hreflist,
                                        a->Value, _tcslen(a->Value));
                    /* If there is no matching href, pretend @id isn't there */
                    if (h)
                        phref = h;
                    break;
                }
            }
            /* If there is no ID but there was a nameless ref, use it */
            if (!phref) {
                /* XXX Some inline structs get here incorrectly.
                 * They need their own href. Works for lists though...
                 */
                //assert(IsNamelessArg);
                phref = FindHref(This->hreflist, NULL, -1);
                /* If there was a selector-is arg use it */
                if (phref && phref->FieldSelector)
                    This->current->SeenItems = phref->FieldSelector;
            }
        }
        if (This->current->State == DS_LIST || !phref) {
            /* Struct within list or inline struct */
            phref = &This->current->listhref;
            if (This->current->State == DS_LIST) {
                This->current->SavedState |= DSS_LIST; /* Allocate new next */
                assert(phref->argd->descr_type == ARG_LISTHEAD);
            }
            metd2 = phref->argd->_structd;
            assert (metd2);

            /* Special treatment for nameless list heads */
            if (metd2->any.Flags & ENTRY_FLAGS_NAMELESS) {
                /* Search for suitable argument. This is either ANYTYPE
                 * or struct pointer or whatever else could match.
                 * XXX Assume first arg is ANYTYPE and nameless.
                 */
                CPARGUMENT_DESCR itemd = &metd2->ptable[0];
                assert(itemd->descr_type == ARG_ANYTYPE);
                assert(itemd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT);
                assert(itemd->TypeIsArg);
                sc = AllocComplex(phref, 0, This->prpcinfo);
                if (FAILED(sc))
                    return sc;
                phref = MakeHref(&This->hreflist, This->Heap,
                                 _T("INLINE"), 1,
                                 phref->data, NULL, This->xmldepth, NULL,
                                 itemd, NULL, phref->flags, 0);
            } else if (This->current->State != DS_LIST) {
                /* Make sure @id (and XXX @type) were indeed present
                 * or LISTNODE or inline
                 */
                return E_SCHEMA_CHECK;
            }
        }

        assert (phref);

        if (phref->argd && phref->argd->descr_type == ARG_ANYTYPE) {
            sc = SetTypeIs(This, phref, Xns, metd);
            if (FAILED(sc))
                return sc;
            /* Remember what we are waiting for */
            phref->structd = metd;
        }

        This->current->href = phref;
        This->current->arraytype = DESCR_TYPE_ANY;

        /* XXX MISSING xsi:type type check */
        sc = AllocComplex(phref, This->current->arraysize, This->prpcinfo);
        if (FAILED(sc))
            return sc;

        This->current->State = DS_STRUCT;
        This->current->metd = metd;

        /* See if there are attributes that match input arguments
         * Ignore any excess attributes
         * XXX should match namespace too?
         * XXX should we require ARGUMENT_DESCR_FLAGS_ATTRIBUTE ?
         */
        for ( ; Attr && AttrCount > 0; Attr++, AttrCount--) {
            UINT i;
            for (i = 0; i < metd->table_size; i++) {
                if (!_tcscmp(metd->ptable[i].name, Attr->Name)) {
                    argd = &metd->ptable[i];

                    /* Mark the arg as received */
                    assert(i/8 < sizeof(This->current->_SeenItems));
                    /* XXX Check for duplicates? */
                    assert(!(This->current->SeenItems[i/32] & (1 << i%32)));
                    This->current->SeenItems[i/32] |= (1 << i%32);

                    sc = DeserializeItem(This, FALSE, FALSE, argd,
                                         Attr->Value, _tcslen(Attr->Value),
                                         NULL);
                }
            }
        }
        return S_OK;

    case DS_METHOD:
    case DS_STRUCT:
    DS_STRUCT: {
        BOOL IsMethod = (This->current->State == DS_METHOD);
        CPARGUMENT_DESCR textargd = NULL;
        UINT i;
        assert(Xns->nstype != NSTYPE_UNRESOLVED);

        metd = This->current->metd;

        /* Search for the correct arg in the method */
        for (i = 0; i < metd->table_size; i++) {
            argd = &metd->ptable[i];
            if (!_tcscmp(argd->name, Name)) {
                This->current->argd = argd;

                assert(i/8 < sizeof(This->current->_SeenItems));
                /* Check for duplicates (except if this is a nameless list) */
                if (argd->descr_type != ARG_LISTHEAD
                    && (argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT) == 0)
                {
                    assert(!(This->current->SeenItems[i/32] & (1 << i%32)));
                    //if (This->current->SeenItems[i/32] & (1 << i%32))
                    //    return E_SCHEMA_CHECK;
                }
                /* Mark the arg as received */
                This->current->SeenItems[i/32] |= (1 << i%32);
                goto  FoundItem;
            }

            if ((argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT)
                && ((argd->descr_type == ARG_LISTHEAD)
                    || (argd->descr_type == ARG_STRUCT))
                && (!_tcscmp(Name, argd->_structd->name)
                    || ((argd->_structd)->any.Flags & ENTRY_FLAGS_NAMELESS))
                && ((This->isResponse && (argd->_Flags & ARGUMENT_DESCR_FLAGS_OUT))
                     || (!This->isResponse && (argd->_Flags & ARGUMENT_DESCR_FLAGS_IN))))
            {
                /* Nested struct within list */
                UINT Saved = This->current->SavedState;
                PHREF h = This->current->href;

              //DBGT((_T("Found struct '%s' in nameless list '%s' in '%s'\n"),
              //     Name, argd->name, metd->name));

                if (!This->current->argd)
                    This->current->argd = argd;

                if ((Saved & DSS_NAMELESSARG)
                    && (h->argd == argd) && (h->xmldepth == This->xmldepth))
                {
                    assert(argd->descr_type == ARG_LISTHEAD);
                    /* Continuing on a list already started */
                    h->data = NULL; /* allocate new listnode next */
                } else {
                    /* Mark listhead as received */
                    This->current->SeenItems[i/32] |= (1 << i%32);

                    This->current->SavedState |= DSS_TO_NAMELESS;
                    This->current++;
                    if (This->current >= &This->_current[MAX_STRUCT_NESTING])
                        return E_OUT_OF_STRUCTURES;
                    memset(This->current, 0, sizeof(*This->current));
                    This->current->SeenItems = This->current->_SeenItems;
                    This->current->SavedState = DSS_NESTED|DSS_NAMELESSARG
                        |((argd->descr_type == ARG_LISTHEAD) ? DSS_LIST : 0)
                        |(Saved&DSS_HEADER);
                }

                This->current->State = DS_LIST;
                This->current->href = h;
                sc = DeserializeItem(This,IsMethod, TRUE, argd, NULL,0,
                                     NULL);
                metd = argd->_structd;
                /* Nameless within nameless */
                if (argd->_structd->any.Flags & ENTRY_FLAGS_NAMELESS) {
                    if (argd->_structd->ptable[0].descr_type == ARG_ANYTYPE) {
                        This->current->State = DS_LIST;
                        goto DS_BODY;
                    } else {
                        /* It looks like we are actually looking at a field */
                        This->current->State = DS_STRUCT;
                        This->current->metd = metd;
                        goto DS_STRUCT;
                    }
                }
                goto StructFound;
            }
#ifdef NAMELESS_STRUCTS
            for(j = 0; j < argd->_structd->table_size; j++)
                if (!_tcscmp(argd->_structd->ptable[j].name, Name)) {
                    DBGT((_T("Found '%s' in '%s' in nameless item '%s' in '%s'\n"),
                          Name, argd->_structd->name, argd->name, metd->name));
                    DebugBreak();
                }
#endif
        }
        /* If this was not a valid argument it might have been a struct
         * in the rest of the body. With nameless methods it is ambigous
         * where the arglist ends. This can only happen at Body level, i.e.
         * xmldepth == 3.
         */
        if ((metd->any.Flags & ENTRY_FLAGS_NAMELESS) && (This->xmldepth == 3)) {
            sc = FillMissingItems(This,
                This->isResponse && !(This->current->SavedState & DSS_HEADER));
            if (FAILED(sc))
                return sc;
            IsNamelessArg = TRUE;
            goto DS_BODY;
        }

        DBGT((_T("RStartElement method/struct %s does not contain item %s\n"),
             metd->name, Name));
        return E_SCHEMA_CHECK;
    FoundItem:
        /* See if there is an href attribute */
        {
            CPSAXATTRIBUTE a = Attr;
            UINT i = AttrCount;

            for ( ; a && i > 0; a++, i--)
                if (!_tcscmp(a->Name, _T("href"))) {
                    /* href is only allowed for struct/array */
                    if ((argd->descr_type != ARG_STRUCT)
                        && (argd->descr_type != ARG_ANYTYPE)
                        && !(argd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY))
                        return E_SCHEMA_CHECK;

                    /* Remember where to get back after arg received */
                    if (This->current->State == DS_STRUCT) {
                        assert(!(This->current->SavedState & DSS_STRUCT));
                        This->current->SavedState |= DSS_STRUCT;
                    }
                    This->current->State = DS_ARG_RECEIVED;

                    return DeserializeItem(This, IsMethod, TRUE, argd,
                                           a->Value, _tcslen(a->Value),
                                           NULL);
                }
        }
        /* No href, data follows inline */

        if (argd->descr_type == ARG_LISTHEAD) {        
            /* Nested struct within list */
            UINT Saved = This->current->SavedState;
            CPMETHOD_DESCR structd = argd->_structd;
            PHREF h = This->current->href;

            This->current++;
            if (This->current >= &This->_current[MAX_STRUCT_NESTING])
                return E_OUT_OF_STRUCTURES;
            memset(This->current, 0, sizeof(*This->current));
            This->current->SeenItems = This->current->_SeenItems;
            This->current->SavedState = DSS_NESTED|DSS_LIST|(Saved&DSS_HEADER);

            This->current->State = DS_LIST;
            This->current->href = h;
            sc = DeserializeItem(This, IsMethod, TRUE, argd, NULL,0,
                                 NULL);

            if (argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT) {
                if (!This->current->href)
                    This->current->href = &This->current->listhref; /* XXX added for CDATA test */
                This->current->SavedState &= ~DSS_LIST; /* see REnd */
                This->current->State = DS_STRUCT;
                assert(argd->_structd);
                This->current->metd = argd->_structd;
                sc = AllocComplex(This->current->href,
                                  This->current->arraysize, This->prpcinfo);
            }
            return sc;
        }

        /* Remember where to get back after arg received */
        if (This->current->State == DS_STRUCT) {
            assert(!(This->current->SavedState & DSS_STRUCT));
            This->current->SavedState |= DSS_STRUCT;
        }
        This->current->State = DS_ARG;

        /* inline array/struct */
        if ((argd->descr_type == ARG_STRUCT)
            || (argd->_Flags & ARGUMENT_DESCR_FLAGS_ARRAY))
        {
            /* Nested struct */
            UINT Saved = This->current->SavedState;
            CPMETHOD_DESCR structd = argd->_structd;
            PHREF h = This->current->href;

            if ((This->current->SavedState & DSS_STRUCT) == 0)
                IsNamelessArg = TRUE;

            This->current->State = DS_ARG_RECEIVED;

            This->current++;
            if (This->current >= &This->_current[MAX_STRUCT_NESTING])
                return E_OUT_OF_STRUCTURES;
            memset(This->current, 0, sizeof(*This->current));
            This->current->SeenItems = This->current->_SeenItems;
            This->current->SavedState = DSS_NESTED|(Saved&DSS_HEADER);

            This->current->State = DS_STRUCT;

            /* If the argument is a nameless struct, use its name for struct
             * search. Similar to nameless arg.
             */
            if (structd && ((structd->any.Flags & ENTRY_FLAGS_NAMELESS)
                            || (argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT)))
            {
                Name = structd->name;
                if (argd->_structXns)
                    Xns = (void*) argd->_structXns; /* else same */

                This->current->href = h;
                sc = DeserializeItem(This, IsMethod, TRUE, argd,
                                     _T("!LIST"), 5,
                                     NULL);

                if (!This->current->href)
                    This->current->href = &This->current->listhref;
                phref = This->current->href;

                This->current->SavedState &= ~DSS_LIST; /* see REnd */
                This->current->State = DS_STRUCT;
                This->current->metd = structd;
                sc = AllocComplex(phref,
                                  This->current->arraysize, This->prpcinfo);
                if (FAILED(sc))
                    return sc;
            }

            goto DS_BODY;
        }

        if (argd->descr_type == ARG_BOOL
            && (argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT))
        {
            /* Special use for BOOL serialize-as="text()"
             * empty elem (TRUE) or nothing (FALSE)
             */
            sc = DeserializeItem(This,
                                 !(This->current->SavedState & DSS_STRUCT),
                                 FALSE, This->current->argd,
                                 _T("true"), 4, NULL);
            This->current->State = DS_ARG_RECEIVED;
        }
        return S_OK;
    }
    case DS_ARRAY:
        /* Accept any name for array item */

        /* Check if the array size is correct */
        if (++This->current->arrayindex >= This->current->arraysize)
            return E_INVALID_PARAMETER;

        /* See if there is an href attribute */
        for ( ; Attr && AttrCount > 0; Attr++, AttrCount--)
            if (!_tcscmp(Attr->Name, _T("href"))) {
                This->current->State = DS_ARRAY_ITEM_RECEIVED;
                return DeserializeHref(This, NULL,
                                       Attr->Value, _tcslen(Attr->Value),
                                       This->current->href, TRUE, NULL);
            }
        /* Otherwise value follows inline */
        This->current->State = DS_ARRAY_ITEM;
        return S_OK;
    case DS_ARRAY_ITEM:
    case DS_ARG:
        /* Inline struct */
        /* Should not get here if href="#id1" style is used.
         * Simple values are delivered through Text().
         */
        DebugBreak();
        return E_INTERNAL_ERROR;
    default:
        DebugBreak();
        return E_INTERNAL_ERROR;
    };
}

SCODE MCT REndElement(PISAXHANDLER iThis, CPSAXATTRIBUTE Elem)
{
    METHOD_DECLARE(Deserializer, REndElement);
    sc = S_OK;

 Again:
    switch (This->current->State) {
    case DS_ENVELOPE:
        assert(!_tcscmp(Elem->Name, _T("Envelope")));
        This->current->State = DS_EMPTY;
        break;
    case DS_HEADER:
        assert(!_tcscmp(Elem->Name, _T("Header")));
        This->current->State = DS_ENVELOPE;
        break;
    case DS_BODY:
        assert(!_tcscmp(Elem->Name, _T("Body")));
        This->current->State = DS_ENVELOPE;
        break;
    case DS_METHOD:
        sc = FillMissingItems(This,
                This->isResponse && !(This->current->SavedState & DSS_HEADER));
        CHECKSC("FillMissingItems");

        /* The method could have been within a header block */
        if (This->current->SavedState & DSS_HEADER) {
            PICONTINUATION cont = This->prpcinfo;
            This->current->State = DS_HEADER;
            This->current->SavedState &= ~DSS_HEADER;

            /* Process the header in-thread using the app stack */
            sc = cont->v->Execute(cont, TRUE, NULL);
            CHECKSC("Execute");
        } else {
            This->current->State = DS_BODY;
            /* If this was a nameless method/struct, there will be no separate
             * closing element so handle it now.
             */
            if (This->current->metd
                && (This->current->metd->any.Flags & ENTRY_FLAGS_NAMELESS))
                goto Again;
        }
        break;
    case DS_ARRAY:
        This->current->arraysize = 0;
        This->current->arrayindex = 0;
        goto struct_array_common;
    case DS_STRUCT:
        assert(!_tcscmp(Elem->Name, This->current->metd->name)
               || (This->current->metd->any.Flags & ENTRY_FLAGS_NAMELESS));
        sc = FillMissingItems(This,
                This->isResponse && !(This->current->SavedState & DSS_HEADER));
        CHECKSC("FillMissingItems struct");
    struct_array_common:
        if (This->current->SavedState & DSS_LIST) {
            This->current->State = DS_LIST;
            This->current->SavedState &= ~DSS_LIST;
            This->current->href->data = NULL;
        } else if (This->current->SavedState & DSS_NESTED) {
            assert(!(This->current->SavedState & DSS_LIST));
            This->current--;
            assert(This->current >= &This->_current[0]);
#if 1
            /* If this was an empty head, wait for more args, else end */
            if ((This->current->argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT) == 0
                || This->current->argd->descr_type != ARG_LISTHEAD)
                goto Again;
#endif
            /* In case this closed an inline list it also closes the arg */
            if (This->current->State == DS_ARG_RECEIVED) {
                if (This->current->SavedState & DSS_STRUCT) {
                    This->current->State = DS_STRUCT;
                    This->current->SavedState &= ~DSS_STRUCT;
                } else {
                    This->current->State = DS_METHOD;
                }
            }
        } else {
            This->current->State = DS_BODY;
        }
        break;
    case DS_ARG:
        /* Deal with empty argument */
        if (This->current->SavedState & DSS_STRUCT) {
            This->current->State = DS_STRUCT;
            This->current->SavedState &= ~DSS_STRUCT;
            sc = DeserializeHref(This, This->current->argd, _T(""), 0,
                                 This->current->href, FALSE, NULL);
            CHECKSC("DeserializeHref");
        } else {
            This->current->State = DS_METHOD;
            sc = DeserializeArgument(This, This->current->argd,
                   This->isResponse && !(This->current->SavedState&DSS_HEADER),
                   _T(""), 0, NULL);
            CHECKSC("DeserializeArgument");
        }

        break;
    case DS_ARG_RECEIVED:
        if (This->current->SavedState & DSS_STRUCT) {
            This->current->State = DS_STRUCT;
            This->current->SavedState &= ~DSS_STRUCT;
        } else {
            This->current->State = DS_METHOD;
        }
        /* After a nameless arg there is one less EndElement (/x BOOL)
         * unless this is also a nameless method in which case there
         * might be more args within the method.
         * But if we came from a named struct, this extra transition
         * was already done.
         */
        if ((This->current->argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT)
            && This->current->argd->descr_type != ARG_BOOL
            && !(This->current->metd->any.Flags & ENTRY_FLAGS_NAMELESS)
            && (This->current->argd->descr_type != ARG_STRUCT
                || (This->current->argd->ptable->any.Flags & ENTRY_FLAGS_NAMELESS)))
            goto Again;
        /* Otherwise this must match the arg name */
        assert(!_tcscmp(Elem->Name, This->current->argd->name));

        break;
    case DS_ARRAY_ITEM:
        /* Empty array item */
        DebugBreak();           /* XXX. Make sure it is marshaled right */
        goto DS_ARRAY_ITEM_RECEIVED;
    DS_ARRAY_ITEM_RECEIVED:
    case DS_ARRAY_ITEM_RECEIVED:
        This->current->State = DS_ARRAY;
        break;
    case DS_LIST:
        assert(This->current->SavedState & DSS_NESTED);
        This->current--;
        assert(This->current >= &This->_current[0]);
        /* Returning for unnamed listhead */
        if (This->current->SavedState & DSS_TO_NAMELESS) {
            This->current->SavedState &=~ DSS_TO_NAMELESS;
            goto Again;
        }
        break;
    case DS_EMPTY:
        DebugBreak();
        CHECKSETSC(FALSE, E_PARSE_ERROR, "DS_EMPTY");
        break;
    default:
        DebugBreak();
        CHECKSETSC(FALSE, E_PARSE_ERROR, "default label");
    }

METHOD_CLEANUP;
    return sc;
}

/******************* ISaxHandler methods **************************/

PRIVATE SCODE MCT
DeserializerStartDocument(PISAXHANDLER iThis)
{
    PDESERIALIZER This = pDeserializer(iThis);
    TRACE("SOAPDeserializerStartDocument");

    /* Maybe some reset cases might allow a different state? */
    assert(This->current->State == DS_EMPTY);

    This->current = &This->_current[0];
    memset(This->current, 0, sizeof(*This->current));
    This->current->SeenItems = This->current->_SeenItems;
    This->current->State = DS_EMPTY;
    return S_OK;
}

PRIVATE SCODE MCT
DeserializerEndDocument(PISAXHANDLER iThis)
{
    PDESERIALIZER This = pDeserializer(iThis);
    TRACE("SOAPDeserializerEndDocument");

    /* Verify that the document was terminated properly
     * Either a soap message (DS_EMPTY) or an XML fragment (DS_BODY).
     */
    if (This->current->State != DS_EMPTY
        && This->current->State != DS_BODY)
    {
        assert(FALSE);          /* XXX debug only */
        return E_SCHEMA_CHECK;
    }

    return S_OK;
}

PRIVATE SCODE MCT
DeserializerStartElement(PISAXHANDLER iThis, PSAXATTRIBUTE Elem,
                         PSAXATTRIBUTE Attributes, UINT Count)
{
    METHOD_DECLARE(Deserializer, StartElement);
    UINT i, RAttrCount;
    PSAXATTRIBUTE Attr;
#define MAXATTRIBUTES 10        /* non-ns */
    SAXATTRIBUTE RAttributes[MAXATTRIBUTES], *RAttr;

    This->xmldepth++;

    NOISE((_T("DSTART(%s:%s)"), Elem->Prefix, Elem->Name));

    /* SUBTREE DELEGATION: delegate or skip current XML subtree */
    if (This->RoleAction != RA_doit) {
        sc = S_OK;
        if (This->RoleAction == RA_forward && This->Serializer) {
            sc = This->Serializer->v->StartElement(This->Serializer, Elem,
                                                   Attributes, Count);
            CHECKSC("forward");
        }
        /* Do nothing when RA_skip or no msg needed */
        goto Out;
    }

    /* Start by declaring namespaces */
    for (i = 0, Attr = Attributes; i < Count; i++, Attr++) {
        if (!_tcscmp(Attr->Prefix, _T("xmlns"))) {
            AddNamespace(This, Attr->Name, Attr->Value);
        } else if (!_tcscmp(Attr->Prefix, _T(""))
                   && (!_tcscmp(Attr->Name, _T("xmlns")))) {
            /* Default NS */
            AddNamespace(This, _T(""), Attr->Value);
        }
    }

    /* Then resolve the prefixes against the namespaces */
    sc = ResolveName(This, Elem);
	if (FAILED(sc)) {
		DBG(("Resolvename failed, IGNORED\n"));
		This->RoleAction = RA_skip;
		This->skiptoxmldepth = This->xmldepth;
		sc = S_OK;
		goto Out;

	}
    CHECKSC("ResolveName");

    sc = SetCurrentNs(This, Elem->Xns);
    CHECKSC("SetCurrentNs");

    RAttr = RAttributes;
    RAttrCount = 0;
    for (i = 0, Attr = Attributes; i < Count; i++, Attr++) {
        if (!_tcscmp(Attr->Prefix, _T("xmlns")))
            continue;
        if (!_tcscmp(Attr->Prefix, _T(""))
            && !_tcscmp(Attr->Name, _T("xmlns")))
            continue;

        RAttr->Prefix = Attr->Prefix;
        RAttr->Name = Attr->Name;
        RAttr->Token = TOKEN_invalid;
        RAttr->Value = Attr->Value;
        sc = ResolveName(This, RAttr);
        CHECKSC("ResolveName");

        RAttr++; RAttrCount++;
    }

    sc = RStartElement(iThis, Elem, RAttributes, RAttrCount);
#ifdef _DEBUG
    if (FAILED(sc)) {
        DBG(("%s: <%s:%s", _method_name, Elem->Prefix, Elem->Name));
        for (i = 0, RAttr = RAttributes; i < RAttrCount; i++, RAttr++)
            DBG((" %s:%s=%s", RAttr->Prefix, RAttr->Name, RAttr->Value));
        DBG(("\n"));
    }
#endif
    CHECKSC("RStartElement");

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE MCT
DeserializerEndElement(PISAXHANDLER iThis, PSAXATTRIBUTE Elem)
{
    METHOD_DECLARE(Deserializer, EndElement);
    UINT i;

    NOISE((_T("DEND(%s:%s.%d)"), Elem->Prefix, Elem->Name, This->xmldepth));

    /* SUBTREE DELEGATION: delegate or skip current XML subtree */
    if (This->RoleAction != RA_doit) {
        /* Check to see if we have reached the end of the subtree */
        if (This->xmldepth != This->skiptoxmldepth) {
            sc = S_OK;
            if (This->RoleAction == RA_forward && This->Serializer)
                sc = This->Serializer->v->EndElement(This->Serializer, Elem);
            /* Do nothing when RA_skip or no msg needed */
            goto Out;
        }
        /* Resume normal processing after this */
        This->skiptoxmldepth = 0;
        This->RoleAction = RA_doit;
		sc = S_OK;
		goto Out;
    }

    sc = ResolveName(This, Elem);
    CHECKSC("ResolveName");

    sc = REndElement(iThis, Elem);
    CHECKSC("REndElement");

METHOD_CLEANUP;
    This->xmldepth--;
    /* Delete namespaces from scope when needed */
    for (i = This->ActiveNsCount; i > 0; i--) {
        if (This->ActiveNamespaces[i].xmldepth > This->xmldepth)
            This->ActiveNsCount--;
        else
            break;
    }
    return sc;
}

PRIVATE SCODE MCT   
DeserializerText(PISAXHANDLER iThis,
                 const _TCHAR * Text, UINT TextLength,
                 PIBUFFER Content, UINT Flags)
{
    METHOD_DECLARE(Deserializer, Text);

    NOISE((_T("DTEXT(%.*s)"), (int) TextLength, Text));

    /* SUBTREE DELEGATION: delegate or skip current XML subtree */
    if (This->RoleAction != RA_doit) {
        sc = S_OK;
        if (This->RoleAction == RA_forward && This->Serializer) {
            sc = This->Serializer->v->Text(This->Serializer,
                                           Text, TextLength,
                                           Content, Flags);
            CHECKSC("forward");
        }
        /* Do nothing when RA_skip or no msg needed */
        goto Out;
    }

    switch (This->current->State) {
    case DS_ARG:
        This->current->State = DS_ARG_RECEIVED;
        sc = DeserializeItem(This,
                             !(This->current->SavedState & DSS_STRUCT),
                             FALSE, This->current->argd,
                             Text, TextLength, Content);
        CHECKSC("DS_ARG");
        break;
    case DS_ARRAY_ITEM:
        This->current->State = DS_ARRAY_ITEM_RECEIVED;
        sc = DeserializeHref(This, NULL, Text, TextLength,
                             This->current->href, FALSE, NULL);
        CHECKSC("DS_ARRAY_ITEM");
        break;
    case DS_EMPTY:
        /* Text outside Envelope not allowed */
        DBG(("DS_EMPTY: Garbage outside XML ignored\n"));
        //CHECKSETSC(FALSE, E_SCHEMA_CHECK, "DS_EMPTY");
        sc = S_OK;
        break;
    case DS_METHOD:
    case DS_STRUCT: {
        /* One of the args/fields may be marked @serialize-as="text()" */
        CPMETHOD_DESCR metd;
        CPARGUMENT_DESCR argd;
        UINT i;
        /* The arguments do not have to be ns qualified */

        metd = This->current->metd;

        for (i = 0; i < metd->table_size; i++) {
            argd = &metd->ptable[i];
            if (argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT) {
                assert(i/8 < sizeof(This->current->_SeenItems));
                /* Check for duplicates */
                assert(!(This->current->SeenItems[i/32] & (1 << i%32)));
                if (This->current->SeenItems[i/32] & (1 << i%32))
                    return E_SCHEMA_CHECK;
                /* Mark the arg as received */
                This->current->SeenItems[i/32] |= (1 << i%32);
                sc = DeserializeItem(This,
                                     (This->current->State == DS_METHOD),FALSE,
                                     argd, Text, TextLength, Content);
                if (This->current->SavedState & DSS_LIST)
                    This->current->href->data = NULL; /* Allocate a new one next */
                CHECKSC("item");
                goto Out;
            }
        }
        DBGT((_T("DeserializerText m/struct %s doesn't contain text() item\n"),
             metd->name));
        CHECKSETSC(FALSE, E_SCHEMA_CHECK, "unexpected text item");
        break;
    }
    case DS_LIST: {
        /* nameless list head within struct points to struct of same name */
        PHREF phref = This->current->href;
        /* One of the args/fields may be marked @serialize-as="text()" */
        CPARGUMENT_DESCR argd, itemd;
        CPMETHOD_DESCR structd;
        UINT i;

        /* Inlined simple list 1) from method. 2) from struct (phref) */
        if (!phref
            || (phref->argd->descr_type == ARG_LISTHEAD
                && (This->current->SavedState == (DSS_NESTED|DSS_LIST))))
        {
            UINT len;
            _TCHAR *p, *p2;
            /* This is an inlined simpletype list */
            phref = &This->current->listhref;
            argd = phref->argd;
            structd = argd->_structd;

            assert(structd->any.Flags & ENTRY_FLAGS_NAMELESS);
            /* Assume first item is where this goes */
            itemd = &structd->ptable[0];
            assert(itemd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT);
            assert(!Content && Text);

            /* Parse string into tokens */
            assert(Text[TextLength] == _EOS);
            /* XXX ignores const */
            /* XXX error checking here */
            p = (_TCHAR*) Text;
            while(p) {
                len = _tcsspn(Text, _T(" \t\r\n"));
                p += len;
                if (!*p)
                    break;
                p2 = _tcssplit(p, _T(" \t\r\n"));
                DBG(("SIMPLE_LIST(%s)\n", p));

                phref->data = NULL;
                sc = AllocComplex(phref, 0, This->prpcinfo);
                if (FAILED(sc))
                    return sc;
                This->current->href = phref;
                sc = DeserializeItem(This, FALSE, FALSE, itemd,
                                     p, _tcslen(p), NULL);
                if (FAILED(sc))
                    return sc;
                p = p2;
            };
            
            return S_OK;
        }

        if (!phref)
            This->current->href;
        assert(phref);
        argd = phref->argd;
        assert(argd && argd->_structd);
        assert(argd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT);
        structd = argd->_structd;
        assert(argd->descr_type == ARG_LISTHEAD);

        /* Search for text field in struct */
        for (i = 0; i < structd->table_size; i++) {
            itemd = &structd->ptable[i];
            if ((itemd->_Flags & ARGUMENT_DESCR_FLAGS_TEXT) == 0)
                continue;
            assert (itemd->descr_type == ARG_STRING);
            /* Found */
            sc = AllocComplex(phref, This->current->arraysize, This->prpcinfo);
            if (FAILED(sc))
                return sc;

            /* Mark the arg as received */
            assert(i/8 < sizeof(This->current->_SeenItems));
            /* XXX Check for duplicates? */
            assert(!(This->current->SeenItems[i/32] & (1 << i%32)));
            This->current->SeenItems[i/32] |= (1 << i%32);

            sc = DeserializeItem(This, FALSE, FALSE, itemd,
                                 Text, TextLength, Content);
            return sc;
        }
        return E_SCHEMA_CHECK;
    }
    default:
        DebugBreak();
        CHECKSETSC(FALSE, E_INTERNAL_ERROR, "bad state");
    }

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE MCT
DeserializerStartInstruction(PISAXHANDLER Sax, PSAXATTRIBUTE Elem,
                             PSAXATTRIBUTE Attr, UINT Count)

{
    return S_OK;
}

PRIVATE SCODE MCT
DeserializerEndInstruction(PISAXHANDLER Sax, PSAXATTRIBUTE Elem)
{
    return S_OK;
}

PRIVATE SCODE MCT DeserializerReset(PISAXHANDLER piscs, PIFILE file)
{
    TRACE("DeserializeReset");
    return E_NOT_IMPLEMENTED;
}

PRIVATE SCODE MCT DeserializerGetContent(PISAXHANDLER piscs, PIFILE *pFile)
{
    TRACE("DeserializeGetContent");
    return E_NOT_IMPLEMENTED;
}

/******************** IUnknown ***********************/
PRIVATE SCODE MCT
DeserializerQueryInterface(PISAXHANDLER iThis, REFIID IidQuery, void** Object)
{
    SCODE sc;
    const _TCHAR *cob;

    TRACE("DeserializerQueryInterface");
    sc = GenericQueryInterface((PIUNKNOWN)iThis, IidQuery, Object,
                               &IID_ISaxHandler);
    if (sc != E_NO_INTERFACE)
        return sc;

    cob = NULL;
    if (UuidCmp(&IID_IWsAddressing, IidQuery)
             || UuidCmp(&IID_IWsEventing, IidQuery)
             || UuidCmp(&IID_IWsManagement, IidQuery)
             || UuidCmp(&IID_IWsDeviceProfile, IidQuery))
        cob = _T("COB/wsaddr.cob");
    else if (UuidCmp(&IID_IWsReservation, IidQuery))
        cob = _T("COB/wsres.cob");
    else if (UuidCmp(&IID_IWsDiscovery, IidQuery))
        cob = _T("COB/wsdiscov.cob");
    else if (UuidCmp(&IID_IDssp, IidQuery))
        cob = _T("COB/dssp.cob");
   else
        return sc;

    return BindToObject(CurrentNameSpace(), cob, 
                        NAME_SPACE_READ, IidQuery, Object);
}

PRIVATE UINT MCT DeserializerAddRef(PISAXHANDLER iThis)
{
    UINT Refs;
    PDESERIALIZER This = (PDESERIALIZER) iThis;

    Refs = AtomicInc(&This->Refs);
    //DBG(("DeserializerAddRef Refs=%d\n", Refs));
    return Refs;
}

PRIVATE UINT MCT DeserializerRelease(PISAXHANDLER iThis)
{
    PDESERIALIZER This = (PDESERIALIZER) iThis;
    PIHEAP Heap = This->Heap;
    UINT Refs = AtomicDec(&This->Refs);

    //DBG(("DeserializerRelease Refs=%d\n", Refs));

    if (Refs > 0)
        return Refs;

    /* Release any memory associated with the structure pointers;
     * Sometimes done in FinishSoapActivation but then prpcinfo == NULL.
     */
    RELEASE(This->prpcinfo);
    This->prpcinfo = NULL;

    RELEASE(This->Serializer);

    DestroyNameSpaces(&This->namespaces, Heap);
    DestroyHReflist(&This->hreflist, Heap);

    if (This->ShouldFreeURL)    /* When allocated through routing */
        free((void*) This->reqLocalURL);
    This->reqLocalURL = NULL;/* else DeAllocated by the HTTP object */
                             /* Pre-requisite to its Usage: HTTPREQUEST object 
                                should be alive and valid */
    if (This->UriQuery)
        free((void *) This->UriQuery);

#if 0 /* if always FALSE */
    /* If the rpcinfo was not delegated to worker thread then zap it */
    if (This->prpcinfo && !This->isResponse) {
        //if (This->prpcinfo->pContext)
        //    WsContextDestroy(This->prpcinfo->pContext);

        CurrentThread()->v->
            SetValue(CurrentThread(), THREAD_VALUE_RPC_CONTINUATION, 0);
        Heap->v->Free(Heap, 0, This->prpcinfo);
    }
#endif

    Heap->v->Free(Heap, 0, (PTR) This);
    _PutHeap(Heap);
    return (Refs);
}
/*** v-table ***/
#include <s_deserializer_v.c>

/**** Constructor to create the Sax Handler object ****/
SCODE MCT CreateDeserializer(PISAXHANDLER *SaxHandler,
                             struct TypeInfo *ptypeinfo,
                             const _TCHAR * reqLocalURL,
                             PIHEAP Heap,
                             PICONTINUATION iCont) /* NULL for server */
{
    METHOD_DECLARE_NOTHIS(deserializer.c, CreateDeserializer);
    PDESERIALIZER   This = NULL;
    PISOAPFACTORY soapfactory = NULL;
           
    TRACE("SAX CallBacks Create");

    sc = BindToObject(CurrentNameSpace(),
                      _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &soapfactory); 
    CHECKSC("bind to soap factory");

    This = Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("deser"),
                          sizeof(*This), 0);
    CHECKSETSC(This, E_NOT_ENOUGH_MEMORY, "alloc deserializer");

    This->v     = &DeserializerVtbl;
    This->Refs  = 1;
    This->Heap = Heap;

    This->current = &This->_current[0];
    This->current->State = DS_EMPTY;

    if (iCont) {
        iCont->v->AddRef(iCont);
        This->pCont = pCONT(iCont);
        This->isResponse = TRUE;
    }

    This->ptypeinfo = ptypeinfo;
    This->ultimateReceiver = TRUE; /* unless overridden by routing */
    This->RoleAction = RA_doit;

    This->current->arraytype = -1;

    if (reqLocalURL)
        This->reqLocalURL = reqLocalURL + 1;

    /* Allocate the RPCINFO structure, but leave it zeroed out. 
     * It will be partially filled in during header processing. 
     * It is fully initialized in appstack.c:CreateAppStack.
     */
    sc = soapfactory->v->CreateContinuation(soapfactory, 0, &This->prpcinfo);
    CHECKSC("CreateContinuation");

    /* Make the RPCINFO structure available to in-thread header 
     * processing code using IThread::GetValue. The RPCINFO is 
     * transferred to the invocation thread in CreateAppStack
     */
    CurrentThread()->v->
        SetValue(CurrentThread(), THREAD_VALUE_RPC_CONTINUATION, (ADDRESS) This->prpcinfo);
    *SaxHandler = (PISAXHANDLER) This;

METHOD_CLEANUP;
    if (FAILED(sc) && This)
        Heap->v->Free(Heap, 0, This);
    RELEASE(soapfactory);
    return sc;
}
