/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>
#include <mmlite.h>
#include "convert.h"
#include "deserializer.h"
#include <diagnostics.h>

#include "xns.h"
//#include "bax.h"

UINT AddRefEntry(PANY_ENTRY Entry)
{
    if (Entry->Flags & ENTRY_FLAGS_CONST)
        return 1;

    return AtomicInc((PUINT) &Entry->RefCnt);
}

UINT ReleaseEntry(PANY_ENTRY Entry)
{
    UINT RefCnt;

    if (Entry->Flags & ENTRY_FLAGS_CONST)
        return 1;

    RefCnt = AtomicDec((PUINT) &Entry->RefCnt);
    if (RefCnt == 0) {
        PIHEAP Heap = CurrentHeap();
        if (Entry->Type == ENTRY_TYPE_ATTR)
        {
            PATTR_ENTRY e = (PATTR_ENTRY) Entry;
            ReleaseEntry(&e->val->any);
        }
        Heap->v->Free(Heap, 0, Entry);
    }
    return RefCnt;
}

/* Allocate string entry and initialize all but data */
PANY_ENTRY AllocNsTable(PIHEAP Heap)
{
    PNS_TABLE nst;
    nst = (PNS_TABLE) Heap->v->Alloc(Heap,
                              HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME(_T("bax:nst")),
                              sizeof(*nst) + sizeof(PANY_ENTRY) * MAXNSSIZE,
                              0);
    if (!nst) {
        DBG(("bax/AllocNamespaceTable: alloc failed.\n"));
        return NULL;
    }

    return &nst->any;
}

/* Allocate string entry and initialize all but data */
PATTR_ENTRY AllocAttrEntry(PIHEAP Heap, UINT ns, UINT id,
                           PSTR_ENTRY val)
{
    PATTR_ENTRY Entry;
    Entry = (PATTR_ENTRY) Heap->v->Alloc(Heap,
                                         HEAP_DEBUG_NAME(_T("bax:attrentry")),
                                         sizeof(*Entry), 0);
    if (!Entry) {
        DBG(("bax/AllocAttrEntry: alloc failed.\n"));
        return NULL;
    }
    Entry->any.Type = ENTRY_TYPE_ATTR;
    Entry->any.RefCnt = 1;

    Entry->ns = ns;
    Entry->id = id;
    assert(val);
    AddRefEntry(&val->any);
    Entry->val = val;

    return Entry;
}

/* Allocate string entry and initialize all but data */
PXNS_REF_ENTRY AllocXnsRefEntry(PIHEAP Heap, PSTR_ENTRY Prefix,
                                CPXMLNAMESPACE Xns, UINT XmlDepth)
{
    PXNS_REF_ENTRY Entry;
    Entry = (PXNS_REF_ENTRY) Heap->v->Alloc(Heap,
                                            HEAP_DEBUG_NAME(_T("bax:xnsref")),
                                            sizeof(*Entry), 0);
    if (!Entry) {
        DBG(("bax/AllocAttrEntry: alloc failed.\n"));
        return NULL;
    }
    Entry->any.Type = ENTRY_TYPE_XNS_REF;
    Entry->any.RefCnt = 1;

    AddRefEntry(&Prefix->any);
    Entry->Prefix = Prefix;
    Entry->xmldepth = XmlDepth;

    Entry->Xns = Xns;

    return Entry;
}

/* Allocate string entry and initialize (data optionally) */
PSTR_ENTRY AllocStrEntry(PIHEAP Heap, UINT StrLen, CBCHAR *Str)
{
    PSTR_ENTRY Entry;
    Entry = (PSTR_ENTRY) Heap->v->Alloc(Heap,
                                        HEAP_DEBUG_NAME(_T("bax:strentry")),
                             sizeof(*Entry) + StrLen*(sizeof(BCHAR)+1), 0);
    if (!Entry) {
        DBG(("bax/AllocStrEntry: alloc failed.\n"));
        return NULL;
    }
    Entry->any.Type = ENTRY_TYPE_STRING;
    Entry->any.RefCnt = 1;
    Entry->any.Flags = 0;

    //Entry->Len = StrLen;
    *(BCHAR*)&Entry->Str[StrLen] = 0;

    if (Str)
        memcpy((BCHAR*) Entry->Str, Str, StrLen*sizeof(BCHAR));

    return Entry;
}

/* Check that a value can be pushed */
BOOL CheckPush(PNS_TABLE This)
{
    if (This->Count+1 > MAXNSSIZE)
        return FALSE;
    return TRUE;
}

void PushIt(PNS_TABLE This, CPANY_ENTRY Entry)
{
    assert(This->Table[This->Count] == NULL);
    assert(Entry != NULL);
    This->Table[This->Count++] = (PANY_ENTRY) Entry;
}

BOOL PushEntry(PNS_TABLE This, CPANY_ENTRY Entry)
{
    if (!CheckPush(This))
        return FALSE;
    PushIt(This, Entry);
    return TRUE;
}

BOOL AssignEntry(PNS_TABLE This, UINT Which, CPANY_ENTRY Entry)
{
    if (Which > MAXNSSIZE)
        return FALSE;

    assert(This->Table[Which] == NULL);
    assert(Entry != NULL);
    This->Table[Which] = (PANY_ENTRY) Entry;

    /* Adjust stack top */
    if (Which > This->Count)
        This->Count = Which;

    return TRUE;
}

/* Check that Where is within stack (Where is 0-based) */
BOOL StrCheckContains(PNS_TABLE This, UINT Where)
{
    INT i = This->Count - Where;
    if (i-1 < 0)
        return FALSE;
    return TRUE;
}

PNS_TABLE GetTab(CPANY_ENTRY e)
{
    if (!e)
        return NULL;
    if (e->Type == ENTRY_TYPE_XNS_REF) {
        PXNS_REF_ENTRY xnr = (PXNS_REF_ENTRY) e;
        e = (CPANY_ENTRY) xnr->Xns;
    }
    if (e->Type == ENTRY_TYPE_XNS) {
        CPXMLNAMESPACE Xns = (CPXMLNAMESPACE) e;
        e = (CPANY_ENTRY) Xns->entries;
    }
    if (e->Type == ENTRY_TYPE_NS_TABLE) {
        return (PNS_TABLE) e;
    }
    return NULL;
}

/* Return copy of top relative entry */
PANY_ENTRY GetTop(PANY_ENTRY any, UINT Which, BOOL AddRef, ENTRY_TYPE Type)
{
    if (any->Type == ENTRY_TYPE_NS_TABLE) {
        PNS_TABLE This = (PNS_TABLE) any;
        PANY_ENTRY Entry;
        INT i;

        i = (INT) This->Count - Which - 1;
        if (i < 0)
            return NULL;
        Entry = This->Table[i];
        /* Type check */
        if (Type != ENTRY_TYPE_ANY && Type != Entry->Type)
            return NULL;
        if (AddRef)
            AddRefEntry(Entry);
        return Entry;
    }
    return NULL;
}

/* Return copy of absolute (bottom relative) entry */
PANY_ENTRY GetBottom(CPANY_ENTRY any, UINT Which, BOOL AddRef, ENTRY_TYPE Type)
{
    PANY_ENTRY Entry = NULL;
    PNS_TABLE tab = GetTab(any);

    if (!tab)
        return NULL;

    if (Which >= (UINT) tab->Count)
        return NULL;

    Entry = tab->Table[Which];

    if (!Entry)
        return NULL;

    /* Type check */
    if (Type != ENTRY_TYPE_ANY && Type != Entry->Type)
        return NULL;
    if (AddRef)
        AddRefEntry(Entry);
    return Entry;
}

/* push attr */
SCODE AttrPush(PNS_TABLE This, UINT ns, UINT id, PSTR_ENTRY val)
{
    PATTR_ENTRY Entry;

    if (!CheckPush(This))
        return E_OUT_OF_STRUCTURES;

    Entry = AllocAttrEntry(CurrentHeap(), ns, id, val);
    if (!Entry)
        return E_NOT_ENOUGH_MEMORY;

    PushIt(This, &Entry->any);

    return S_OK;
}

/* push xmlns */
SCODE XnsRefPush(PNS_TABLE This, PSTR_ENTRY Prefix, PSTR_ENTRY QName,
                 PISAXHANDLER Deserializer, UINT XmlDepth)
{
    PXNS_REF_ENTRY Entry;
    CPXMLNAMESPACE Xns;

    if (!CheckPush(This))
        return E_OUT_OF_STRUCTURES;

    Xns = LookupNsQName(Deserializer, QName);
    Entry = AllocXnsRefEntry(CurrentHeap(), Prefix, Xns, XmlDepth);
    if (!Entry)
        return E_NOT_ENOUGH_MEMORY;

    PushIt(This, &Entry->any);

    return S_OK;
}

/* push literal */
SCODE StrPushLiteral(PNS_TABLE This, CBCHAR *str, UINT Len)
{
    PSTR_ENTRY Entry;

    if (Len == 0)
        Len = _tcslen(str);

    if (!CheckPush(This))
        return E_OUT_OF_STRUCTURES;

    Entry = AllocStrEntry(CurrentHeap(), Len, str);
    if (!Entry)
        return E_NOT_ENOUGH_MEMORY;

    PushIt(This, &Entry->any);

    return S_OK;
}

/* push copy of entry n */
SCODE DupN(PNS_TABLE This, UINT Which)
{
    PANY_ENTRY Entry;

    if (!CheckPush(This))
        return E_OUT_OF_STRUCTURES;

    Entry = GetTop(&This->any, Which, TRUE, ENTRY_TYPE_ANY);
    if (!Entry)
        return E_FAIL;
    PushIt(This, Entry);

    return S_OK;
}

/* push partial copy of top entry, dropped */
SCODE StrSubstr(PNS_TABLE This, UINT Start, UINT Len)
{
    PSTR_ENTRY e, Entry;
    INT i;
    UINT eLen;

    i = (INT) This->Count - 1;
    if (i < 0)
        return E_INVALID_PARAMETER;

    e = (PSTR_ENTRY) This->Table[i];
    eLen = _tcslen(e->Str);

    /* The substring cannot be larger than the original */
    if (Start + Len > eLen)
        return E_INVALID_PARAMETER;

    /* XXX optimize by copying in place when applicable */
    Entry = AllocStrEntry(CurrentHeap(), Len, e->Str + Start);
    if (!Entry)
        return E_NOT_ENOUGH_MEMORY;

    ReleaseEntry(&e->any);
    This->Table[i] = &Entry->any;

    //printf("ss{%*s}", Entry->Len, Entry->Str);
    return S_OK;
}

/* push glued together copy of N top entries, dropped */
SCODE StrConcat(PNS_TABLE This, UINT n)
{
    PSTR_ENTRY e, Entry;
    INT i;
    UINT j, Len, eLen;

    assert(n >= 1);

    /* Make sure we've got enough entries. i is the index of the new one. */
    i = (INT) This->Count - n - 1 + 1; /* +1 for new entry */
    if (i < 0)
        return E_INVALID_PARAMETER;

    /* Calculate total length */
    Len = 0;
    for (j = i; j < i+n; j++) {
        e = (PSTR_ENTRY) This->Table[j];
        Len += _tcslen(e->Str);
    }

    Entry = AllocStrEntry(CurrentHeap(), Len, NULL);
    if (!Entry)
        return E_NOT_ENOUGH_MEMORY;

    Len = 0;
    for (j = i; j < i+n; j++) {
        /* Copy each segment */
        e = (PSTR_ENTRY) This->Table[j];
        eLen = _tcslen(e->Str);
        memcpy((BCHAR*) Entry->Str + Len, e->Str, eLen * sizeof(BCHAR));
        Len += eLen;
        /* Drop them */
        ReleaseEntry(&e->any);
        This->Table[j] = NULL;
    }

    /* Push new entry and adjust stack size */
    This->Table[i] = &Entry->any;
    This->Count = i + 1;

    //printf("cc{%*s}", Entry->Len, Entry->Str);
    return S_OK;
}

/* drop top entry */
SCODE Drop(PNS_TABLE This, UINT Which)
{
    PANY_ENTRY Entry;
    INT i;

    i = (INT) This->Count - Which - 1;
    if (i < 0)
        return E_INVALID_PARAMETER;
    Entry = This->Table[i];

    ReleaseEntry(Entry);

    /* Move remaining entries down */
    for ( ; (UINT) i < This->Count; i++)
        This->Table[i] = This->Table[i+1];

    /* Zap the last one */
    This->Table[i] = NULL;   
    This->Count--;

    return S_OK;
}

/* Various utilities that use CPNS_TABLE */
CBCHAR *GetString(CPNS_TABLE nst, UINT Which)
{
    CPSTR_ENTRY te = (CPSTR_ENTRY) GetBottom(&nst->any, Which,
                                             FALSE, ENTRY_TYPE_STRING);
    if (te)
        return te->Str;
    else
        return _T("");
}

UINT16 FindString(CPNS_TABLE nst, CBCHAR *value)
{
    UINT16 i;
    CPSTR_ENTRY te;

    assert(nst && nst->any.Type == ENTRY_TYPE_NS_TABLE);

    for (i = 0; i < nst->Count; i++) {
        /* XXX binary search! */
        te = (CPSTR_ENTRY) nst->Table[i];
        if (te->any.Type != ENTRY_TYPE_STRING)
            continue;

        if (!_tcscmp(value, te->Str)) {
            return i;
        }
    }
    return TOKEN_invalid;
}



SCODE BResolveName(/* inout */ PSAXATTRIBUTE Entry)
{
    CPXMLNAMESPACE Xns = Entry->Xns;
    CPNS_TABLE nst;
    PSTR_ENTRY str;
    UINT i;

    nst = (CPNS_TABLE) Xns->entries;
    assert(nst->any.Type == ENTRY_TYPE_NS_TABLE);

    if (Entry->Token != TOKEN_invalid) {
        assert(Entry->Name == NULL);
        str = (PSTR_ENTRY) GetBottom((CPANY_ENTRY) nst, Entry->Token,
                                     TRUE, ENTRY_TYPE_STRING);
        if (!str)
            return E_SCHEMA_CHECK;
        Entry->Name = str->Str;
        return S_OK;
    }

    for (i = 0; i < nst->Count; i++) {
        if (nst->Table[i]->Type != ENTRY_TYPE_STRING)
            continue;
        str = (PSTR_ENTRY) nst->Table[i];
        if (_tcscmp(str->Str, Entry->Name))
            continue;

        Entry->Token = i;
        return S_OK;
    }

    assert(FALSE);
    Entry->Token = 254;
    return S_OK;
}
