/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <netif.h>
#include <tokenizerif.h>
#include <httputils.h>
#include <tokenizerutils.h>
#include <heaps/dbg_heap.h>
#include <winsock.h>
#include "../soap/soap.h"
#include "httpserver.h"
#include "http.h"
#include "httpconstants.h"

/* Populate the interface */
#include <s_httpconnector_v.c>

SCODE MCT HttpConnectorConnect(PIHTTPCONNECTOR iThis);

PRIVATE SCODE MCT HttpConnectorAllocateBuffer(PIHTTPCONNECTOR pihc,
                                              PIBUFFER *pmi)
{
    HttpConnector * This = (HttpConnector *)pihc;
    PIENDPOINT ep;
    SCODE sc;

    assert (This->ProtocolState != HPS_DEAD);
    ep = This->Endpt;

    if (!ep) {
        assert (!This->isServer);

        sc = HttpConnectorConnect(pihc);
        if (FAILED(sc)) return sc;
        ep = This->Endpt;
    }

    if (!ep)
        return E_OBJECT_NOT_INITIALIZED;

    return ep->v->AllocateBuffer(ep, 0, pmi);
}

PRIVATE SCODE MCT HttpConnectorSendMessage(PIHTTPCONNECTOR iThis,
    UINT StatusCode,
    const _TCHAR *SoapAction,
    /*optional*/ PIFILE ContentFile,
    /*optional*/ PICONTINUATION iCont) /* NULL when async or reply */
{
    METHOD_DECLARE(HttpConnector, SendMessage);
    PCONTINFO cont = (PCONTINFO) iCont;
    _TCHAR *Url;
    PIENDPOINT ep;

    Mutex_Lock(&This->mx);

    Url = This->RequestURI;
    ep = This->Endpt;

    if (!This->isServer) {
        if (!Url && cont && cont->WsaTo)
            Url = cont->WsaTo;
        if (!Url && cont && cont->pContext && cont->pContext->To)
            Url = cont->pContext->To->Address;

        CHECKSETSC(Url, E_INVALID_PARAMETER, "current Url");

        if (This->cont) {
            assert(FALSE);
            CHECKSETSC(FALSE, E_OBJECT_NOT_INITIALIZED, "-wrong state-");
        }

        if (cont) {
            This->cont = iCont;
            cont->v->AddRef(iCont);
        }

        Url = _tcschr(Url, '/');
        CHECKSETSC(Url, E_INVALID_PARAMETER, "process URL 1");

        Url = _tcschr(Url+1, '/');
        CHECKSETSC(Url, E_INVALID_PARAMETER, "process URL 2");

        Url = _tcschr(Url+1, '/');
        CHECKSETSC(Url, E_INVALID_PARAMETER, "process URL 3");

        if (!ep) {
            sc = HttpConnectorConnect((PIHTTPCONNECTOR) This);
            CHECKSC("HttpConnectorConnect reconnect");
            ep = This->Endpt;
            assert(ep);
        }

        sc = HttpSend(ContentFile, ep,
                      This->Flags, This->Hostname, Url, This->BasicKey,
                      (struct sockaddr *) &This->sockaddr,
                      sizeof(This->sockaddr),
                      SoapAction, TOKEN_HTTP_1_1, StatusCode, FALSE);
        CHECKSC("HttpSend");
    } else {
        BOOL ConnectionClose = (!HTTPSessionIsKeepAlive(This)
                                || (StatusCode != TOKEN_OK));

        assert(!cont);

        sc = HttpSend(ContentFile, This->Endpt,
                      This->Flags, NULL, NULL, NULL,
                      (This->Flags & PARSE_URL_FLAGS_UDP) ? (struct sockaddr *) &This->sockaddr : NULL,
                      (This->Flags & PARSE_URL_FLAGS_UDP) ? sizeof(This->sockaddr) : 0,
                      SoapAction, This->HTTPVersion, StatusCode,
                      ConnectionClose);
                  
        TRACE("HTTPResponseSendMessage");
        CHECK_HEAP();
    }

METHOD_CLEANUP;
    Mutex_Unlock(&This->mx);
    return sc;
}

/* (re)-Initialize connector to receive a new message */
PRIVATE SCODE HttpConnectorStartReceive(PIHTTPCONNECTOR iThis, BOOL isResponse)
{
    METHOD_DECLARE(HttpConnector, StartReceive);
    PINAMESPACE ns = CurrentNameSpace();
    PITOKENIZER Tokenizer = NULL;
    PIUNKNOWN Input = NULL;
    PIENDPOINT ep = This->Endpt;

    sc = S_OK;

    /* Initialize to defaults */
    This->Chunked = FALSE;
    This->KeepAlive = TRUE;
    This->ContentLength = 0;
    This->StatusCode = TOKEN_OK;

    if ((This->Flags & PARSE_URL_FLAGS_HTTP) == 0) {
        if (isResponse)
            This->Method = TOKEN_OK;
        else
            This->Method = TOKEN_POST;
    }

    if (This->Flags & PARSE_URL_FLAGS_UDP) {
#define MAX_MSG_SIZE 0x1000
        RELEASE_NULL(This->Tokenizer);
        sc = ep->v->RecvBuffer(ep,
                               (PIBUFFER *) &Input,
                               MAX_MSG_SIZE, 0,
                               (struct sockaddr*)&This->sockaddr,
                               sizeof(This->sockaddr), &This->sa_len);
        if (sc == STATUS_WIN32_ERROR(WSAETIMEDOUT)) /* messy error code */
            sc = E_TIMED_OUT;
        if (sc == STATUS_WIN32_ERROR(WSAEWOULDBLOCK)) /* messy error code */
            sc = E_WOULD_BLOCK;
        if (sc == S_FALSE || !Input)
            sc = E_END_OF_DATA;

        CHECKSC("RecvBuffer UDP message");
        This->KeepAlive = FALSE;
    } else {
        if (!This->isServer) {
            /* Set TCP_NODELAY */
            /* BUGBUG do we really want to do this? */
            UINT NoDelay = 1;

            sc = ep->v->SetSockOpt(ep,
                                   IPPROTO_TCP, 
                                   TCP_NODELAY, 
                                   (PBYTE) &NoDelay, 
                                   sizeof(UINT));
            CHECKSC("set TCP_NODELAY");
        }

        ep->v->AddRef(ep);      /* symmetry with UDP case */
        Input = (PIUNKNOWN) ep;
    }

    Tokenizer = This->Tokenizer;
    if (Tokenizer) {
        assert ((This->Flags & PARSE_URL_FLAGS_UDP) == 0);
        assert(This->KeepAlive == TRUE || isResponse);
    } else {
        PITOKENIZERFACTORY tokenizerfactory;
        sc = BindToObject(ns, _T("COB/tokenizer.cob"), NAME_SPACE_READ,
                          &IID_ITokenizerFactory, (void **) &tokenizerfactory);
        CHECKSC("BIND tokenizer factory");

        sc = tokenizerfactory->v->Create(tokenizerfactory, 
                                         Input,
                                         &Tokenizer);
        RELEASE(tokenizerfactory);
        CHECKSC("Create tokenizer");
        This->Tokenizer = Tokenizer;
        assert(sc != S_FALSE);
    }

METHOD_CLEANUP;
    RELEASE(Input);
    return sc;
}


PRIVATE SCODE HttpConnectorGetPut(PIHTTPCONNECTOR iThis,
                                  BOOL isPut, PIFILE File)
{
    METHOD_DECLARE(HttpConnector, GetPut);

#ifndef _MINIMIZE
    if (isPut) {
        sc = iThis->v->SendMessage(iThis, TOKEN_PUT, NULL, File, NULL);
    } else {
        File->v->AddRef(File);
        This->ResponseFile = File;
        sc = iThis->v->SendMessage(iThis, TOKEN_GET, NULL, NULL, NULL);
    }
    CHECKSC("SendMessage");

    do {
        /* Keep receiving until enough data received */
        sc = iThis->v->ReceiveMessage(iThis);
    } while (sc == E_RESTART || sc == E_WOULD_BLOCK);
    CHECKSC("ReceiveMessage");

    sc = iThis->v->Done(iThis);
    CHECKSC("Done");
#else
    sc = E_NOT_IMPLEMENTED;
#endif

METHOD_CLEANUP;
    return sc;
}

/* Negotiate session key */
PRIVATE SCODE HttpConnectorSetKey (PIHTTPCONNECTOR iThis)
{
    METHOD_DECLARE(HttpConnector, SetKey);
#if 1
    sc = E_NOT_IMPLEMENTED;
#else
    PISAMPLECALCULATOR calc = NULL;
    
    if (!(This->Flags & PARSE_URL_FLAGS_AES))
        return S_OK;

    sc = BindToObject(CurrentNameSpace(), _T("COB/calc.cob"),NAME_SPACE_READ,
                        &IID_ISampleCalculator, (void**) &calc);
    CHECKSC("bind to calc.cob");

    /* Server Authentication */
    DBG(("Server Authentication ..."));
    sc = ParnerAuthenticate(calc, iThis, This->Url, &This->ServerCert);
    CHECKSC("PartnerAuthenticate");

    DBG(("Done Authenticating!\n"));
    /* Randomly generate and set the session key */
    sc = SetSessionKey(calc, iThis, This->Url, This->ServerCert, &This->SKey);
    CHECKSC("SetSessionKey");
    This->isSKeyValid = TRUE;
#endif
METHOD_CLEANUP;
#if 0
    RELEASE(calc);
#endif
    return sc;
}

/* Unset session key */
PRIVATE SCODE HttpConnectorClearKey(PIHTTPCONNECTOR iThis)
{
    METHOD_DECLARE(HttpConnector, Clearkey);
#if 1
    sc = E_NOT_IMPLEMENTED;
#else
    if (!(This->Flags & PARSE_URL_FLAGS_AES))
        return S_OK;

    sc = ClearSessionKey(iThis, This->Url, &This->SKey);
    CHECKSC("ClearSessionKey");
    This->isSKeyValid = FALSE;
#endif

METHOD_CLEANUP;
    return sc;
}

/* Negotiate session key */
SCODE HttpConnectorSetBasicKey (PIHTTPCONNECTOR iThis,
                                        const _TCHAR *User,
                                        const _TCHAR *Pass)
{
    METHOD_DECLARE(HttpConnector, SetBasicKey);
    const _TCHAR *t;
    BYTE *InBuf, *OutBuf, *p;
    UINT InLen = _tcslen(User) + _tcslen(Pass) + 2;
    UINT OutLen = InLen * 2;

    InBuf = malloc(InLen);
    OutBuf = malloc(OutLen);

    CHECKSETSC(InBuf && OutBuf, E_NOT_ENOUGH_MEMORY, "alloc key buffers");

    for (p = InBuf, t = User; *t; p++, t++)
        *p = (BYTE) *t;
    *p++ = ':';
    for (t = Pass; *t; p++, t++)
        *p = (BYTE) *t;
    *p = '\0';

    sc = base64Coder(NULL, CYPHER_ENCRYPT,
                     InBuf, InLen-1, NULL,
                     OutBuf, OutLen, &OutLen);
    CHECKSC("base64 encode");

    OutBuf = realloc(OutBuf, OutLen+1);

    OutBuf[OutLen] = '\0';

    This->BasicKey = (const char *) OutBuf;
    OutBuf = NULL;

METHOD_CLEANUP;
    if (InBuf)
        free(InBuf);
    if (OutBuf)
        free(OutBuf);
    return sc;
}

SCODE HttpConnectorListen(PIHTTPCONNECTOR iThis,
                          PTIME timeo)
{
    METHOD_DECLARE(HttpConnector, Listen);

    Mutex_Lock(&This->mx);

    CHECKSETSC(This->Server, E_OBJECT_NOT_INITIALIZED, "Server not started");

    sc = This->Server->v->SetEndpoint(This->Server, This->Endpt,
                                      This->Flags,// |PARSE_URL_FLAGS_ANONYMOUS,  
                                      (PIUNKNOWN) iThis,
                                      timeo);
    CHECKSC("SetEndpoint");

METHOD_CLEANUP;
    Mutex_Unlock(&This->mx);
    return sc;
}

SCODE MCT HttpConnectorConnect(PIHTTPCONNECTOR iThis)
{
    METHOD_DECLARE(HttpConnector, Connect);
    PIENDPOINTFACTORY ef = NULL;
    UINT32 RcvTimeo;

    Mutex_Lock(&This->mx);

    /* Get an endpoint factory to create an endpoint */
    sc = BindToObject(CurrentNameSpace(), _T("COB/protocol.cob"),
                      NAME_SPACE_READ, &IID_IEndpointFactory,
                      (void**) &ef);
    CHECKSC("Bind(COB/protocol.cob)");
        
    /* Create an endpoint to query for the interface address */
    sc = ef->v->CreateEndpoint(ef, AF_INET,
       (This->Flags & PARSE_URL_FLAGS_UDP) ? SOCK_DGRAM : SOCK_STREAM,
       0, &This->Endpt);
    ef->v->Release(ef);
    CHECKSC("CreateEndpoint");

    /* Set recv timeout to 60 seconds, the NT way */
    RcvTimeo = 60000;
    sc = This->Endpt->v->SetSockOpt(This->Endpt,
                                    SOL_SOCKET,SO_RCVTIMEO,
                                    (PBYTE)&RcvTimeo,
                                    sizeof RcvTimeo);
    CHECKSC("set receive timeout");

    if ((This->Flags & PARSE_URL_FLAGS_UDP) == 0) {
#if defined(_DEBUG)
    {
        DBG(("httpconnector: Connecting to %s:%d\n",
             inet_ntoa(This->sockaddr.sin_addr),
             ntohs(This->sockaddr.sin_port)));
    }
#endif
        sc = This->Endpt->v->Connect(This->Endpt, 
                                     (struct sockaddr * )&This->sockaddr, 
                                     sizeof(This->sockaddr));
        CHECKSC("connect");

        /* Must set tcp-soap socket to non-block */
        if ((This->Flags & PARSE_URL_FLAGS_HTTP) == 0
            && (This->Flags & PARSE_URL_FLAGS_SOAP))
        {
            int val = 1;
            sc = This->Endpt->v->Ioctl(This->Endpt, FIONBIO, (PTR) &val);
            CHECKSC("Set tcp-soap socket to non-block");
        }
    }

METHOD_CLEANUP;
    Mutex_Unlock(&This->mx);
    RELEASE(ef); 

    return sc;
}

/* Server HttpConnector constructor */
SCODE HTTPSessionCreate(PILISTENER Server, 
                        PIENDPOINT ep,
                        PARSE_URL_FLAGS Flags,
                        HttpConnector ** Session)
{
    METHOD_DECLARE_NOTHIS(HttpSession, Create);
    HttpConnector *conn;
    PIHEAP Heap = _GetHeap(_T("HttpSession"), 10000);

    sc = S_OK;

    /* Allocation */
    conn = Heap->v->Alloc(Heap,
                          HEAP_DEBUG_NAME("httpconnnector")|HEAP_ZERO_MEMORY,
                          sizeof(*conn), 0);
    CHECKSETSC(conn, E_NOT_ENOUGH_MEMORY, "alloc session");

    conn->Refs = 1;
    conn->v = &HttpConnectorVtbl.v;
    conn->Heap = Heap;
    conn->Flags = Flags;

    /* Initialization */
    if (Server) {
        Server->v->AddRef(Server);
        conn->Server = Server;
        conn->isServer = TRUE;
    }

    if (ep)
        ep->v->AddRef(ep);
    conn->Endpt = ep;
    conn->Chunked = FALSE;
    conn->ContentLength = 0;

    *Session = conn;

METHOD_CLEANUP; 
    if (FAILED(sc) && conn) { 
        RELEASE(conn->Server);
        Heap->v->Free(Heap, 0, (PTR) conn);
    }
    _PutHeap(Heap);

    return sc;
}

/* Client HttpConnector constructor */
SCODE MCT HttpFactoryCreateConnector (
    PIWEBPROTOCOLFACTORY iThis,
    const _TCHAR *Url,
    PARSE_URL_FLAGS ExtraFlags,
    /*out*/ PIHTTPCONNECTOR *pISoapConnector)
{
    METHOD_DECLARE(HttpFactory, CreateConnector);
    PIHEAP Heap = _GetHeap(_T("HttpConnector"), 1000);
    const _TCHAR *HostName;
    UINT HostNameLen;
    PISOAPFACTORY SoapFactory = NULL;
    UINT32 ipaddr;
    UINT16 Port;
    PARSE_URL_FLAGS Flags = 0;
    HttpConnector * conn = NULL;
    PILISTENER Server = This->Server;

    /* Set Flags */
    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &SoapFactory);
    CHECKSC("bind to soap.cob");

    sc = SoapFactory->v->ParseURL(SoapFactory, Url, &Flags, 
                                  &HostName, &HostNameLen, 
                                  &Port, NULL,NULL,NULL,NULL,NULL);
    CHECKSC("ParseURL");

    /* Set flags that cannpt be derived from URL */
    Flags |= ExtraFlags;

    /* Create connector (common with server side) */
    sc = HTTPSessionCreate(NULL, NULL, Flags, &conn);
    CHECKSC("HTTPSessionCreate");

    /* Set client specific fields */

    /* Remember the server that is associated so that clients can register
     * e.g. UDP multicast sockets for replies.
     */
    if (Server) {
        Server->v->AddRef(Server);
        assert(!conn->Server);
        conn->Server = Server;
    }

    /* Remember that we expect replies and send requests, not the opposite */
    conn->isServer = FALSE;

    conn->RequestURI = _tcsdup(Url);
    CHECKSETSC(conn->RequestURI, E_NOT_ENOUGH_MEMORY, "strdup URL");

    conn->Hostname = malloc((HostNameLen+1) * sizeof(_TCHAR));
    CHECKSETSC(conn->Hostname, E_NOT_ENOUGH_MEMORY, "strdup hostname");

    _tcsncpy(conn->Hostname, HostName, HostNameLen);
    conn->Hostname[HostNameLen] = '\0';

   /* Prepare address and port to listen at */
    /* Set sockaddr */
    ipaddr = _tinet_addr(conn->Hostname);
    memset(&conn->sockaddr, 0, sizeof(conn->sockaddr));

    if (ipaddr != INADDR_NONE) {
        conn->sockaddr.sin_addr.s_addr = ipaddr;
    } else {
        char *host;
#if 1 /* XXX does not link with winshim, stdcall linkage issue */
        struct hostent *hostentry;
#ifdef _UNICODE
        host = _wcstos(conn->Hostname);
        CHECKSETSC(host, E_NOT_ENOUGH_MEMORY, "convert unicode hostname");
#else
        host = conn->Hostname;
#endif
        if ( (hostentry = gethostbyname(host) )) {  
            memcpy(&conn->sockaddr.sin_addr,
                   hostentry->h_addr, 
                   hostentry->h_length);
            DBG(("udpconnector: Successful gethostbyname %s\n",
                 inet_ntoa(conn->sockaddr.sin_addr)));
        } else
#endif
            CHECKSETSC(FALSE, E_FAIL, "gethostbyname");
#if defined(_UNICODE) && 1
        free(host);
#endif
    }

    conn->sockaddr.sin_family = AF_INET;
    conn->sockaddr.sin_port = htons(Port);

    sc = HttpConnectorConnect((PIHTTPCONNECTOR) conn);
    CHECKSC("HttpConnectorConnect");

METHOD_CLEANUP;
    RELEASE(SoapFactory); 
    if (FAILED(sc))
        RELEASE((PIHTTPCONNECTOR) conn);
    else
        *pISoapConnector = (PIHTTPCONNECTOR) conn;
    if (FAILED(sc))
        DBG(("URL=%s\n", Url));
    return sc;
}

PRIVATE UINT ScToHttpStatus(SCODE sc)
{
    switch (sc) {
    case E_NOT_IMPLEMENTED:
        return TOKEN_METHOD_NOT_ALLOWED;
    case E_FILE_NOT_FOUND:
    case E_PATH_NOT_FOUND:
        return TOKEN_NOT_FOUND;
    case E_TIMED_OUT:
        return TOKEN_REQUEST_TIME_OUT;
    default:
        return TOKEN_INTERNAL_SERVER_ERROR;
    }
}

PRIVATE void LogMsg(PIBUFFER hdr)
{
#if defined(_DEBUG) && defined(i386)
    /* Spit out the serialized data into file for examination */
    PIFILE file;
    UINT wrote;
    SCODE sc = BindToObject(CurrentNameSpace(), _T("fs/httpconn-out.txt"),
                            NAME_SPACE_CREATE | NAME_SPACE_WRITE,
                            &IID_IFile, (void **) &file);
    if (SUCCEEDED(sc)) {
        hdr->v->ReadWriteFile(hdr, file, TRUE,
                              UINT64_MAX, 0, UINT_MAX, &wrote);
        file->v->Release(file);
    }
#endif
}

/* XXX duplicate in createtokenizer.c */
PRIVATE SCODE MCT Create_Tokenizer(PIUNKNOWN tok, PITOKENIZER *Tokenizer)
{
    SCODE sc;
    PITOKENIZERFACTORY Factory;

    /* Prepare a tokenizer */
    //DBG(("Preparing file tokenizer\n"));
    if (*Tokenizer) {
        (*Tokenizer)->v->Release(*Tokenizer);
        *Tokenizer = NULL;
    }

    sc = BindToObject(CurrentNameSpace(), _T("COB/tokenizer.cob"), 0,
                      &IID_ITokenizerFactory, (void**) &Factory);

    if (FAILED(sc))
        return sc;
    //DBG(("Queried for tokenizer factory interface\n"));

    sc = Factory->v->Create(Factory,
                            (PIUNKNOWN) tok, 
                            Tokenizer);
    Factory->v->Release(Factory);
    //DBG(("Constructed tokenizer instance sc=%x\n", sc));
    return sc;
}

PRIVATE void HTTPRequestDeleteHeaders(PIHTTPCONNECTOR Request)
{
    HttpConnector * This = (HttpConnector *) Request;
    PHTTPHEADER Headers;
    PIHEAP Heap = This->Heap;

    /* Free the request headers */
    Headers = &This->Headers;
    do {
        PHTTPHEADER More;
        UINT i;
        for (i = 0; i < NUM_FIELDS_IN_BLOCK; ++i) {
            if (Headers->Fields[i].StringLength > 0)
                Heap->v->Free(Heap, 0, (PTR) Headers->Fields[i].Name);
            Headers->Fields[i].Name = 0;
            Headers->Fields[i].StringLength = 0;
            if (Headers->Fields[i].Value)
                Heap->v->Free(Heap, 0, (PTR) Headers->Fields[i].Value);
            Headers->Fields[i].Value = 0;
            Headers->Fields[i].ValueLength = 0;
        }
        More = Headers->More;
        Headers->More = 0;
        if (Headers != &This->Headers) /* first one is part of This */
            Heap->v->Free(Heap, 0, (PTR) Headers);
        Headers = More;
    }
    while (Headers != NULL);

}

PRIVATE SCODE MCT
HttpConnectorDone(PIHTTPCONNECTOR Request)
{
    HttpConnector * This = (HttpConnector *) Request;
    PIHEAP Heap = CurrentHeap();

    TRACE("HTTPRequestDone");

    Mutex_Lock(&This->mx);

    if (This->isServer) {
        /* Reset per request flags (Content-Type etc. determine the rest) */
        if (This->Flags & PARSE_URL_FLAGS_HTTP)
            This->Flags &= (PARSE_URL_FLAGS_HTTP | PARSE_URL_FLAGS_UDP
                            | PARSE_URL_FLAGS_FILE);

        if (This->RequestURI) {
            Heap->v->Free(Heap, 0, This->RequestURI);
            This->RequestURI = NULL;
        }
        This->RequestURILength = 0;
        This->HTTPVersion = 0;
    }

    HTTPRequestDeleteHeaders(Request);

    RELEASE_NULL(This->Target);
    This->StatusCode = 0;
    This->SaxState = 0;
    This->Method = 0;           /* HTTP GET/POST/... */

    This->State = HS_START;

    /* Should be cacheable but would require partial cleanup */
    RELEASE_NULL(This->Deserializer);

    if (This->isServer && This->ProtocolState != HPS_DEAD) {
        assert(This->Deserializer == NULL);

        if (HTTPSessionIsKeepAlive(This)
            || (This->Flags & PARSE_URL_FLAGS_UDP)
            //&& (This->Flags & PARSE_URL_FLAGS_UDP) == 0
            )
        {
            /* Re-enable receiving as appropriate */
            This->Server->v->SetEndpoint(This->Server, This->Endpt,
                                         0, (PIUNKNOWN) Request, NULL);
        }
    } else {
        PICONTINUATION cont = This->cont;
        This->cont = NULL;

        /* When necessary, create a new socket later */
        if (This->ProtocolState == HPS_DEAD) {
            /* No more receiving on this connector */
            RELEASE_NULL(This->Tokenizer);
            RELEASE_NULL(This->Endpt);
            This->ProtocolState = HPS_START;
        }

        /* All done, wake up client */
        if (cont && !This->isServer) {
            /* Client keeps circular ref during RPC. Break it here.
             * Must release before client readied as cont might be on stack.
             * The Complete method figures it out.
             */

            cont->v->Complete(cont); /* Consumes ref */
        }

    }
    Condition_Signal(&This->cd);

    Mutex_Unlock(&This->mx);
    return S_OK;
}

PRIVATE SCODE MCT HttpConnectorDestroy(PIUNKNOWN Request)
{
    HttpConnector * This = (HttpConnector *) Request;
    PIHEAP Heap = This->Heap;
    PIENDPOINT ep = This->Endpt;

    TRACE("HttpConnectorDestroy");

    if (ep) {
        if (!This->isServer && (This->Flags & PARSE_URL_FLAGS_UDP) == 0)
            ep->v->Close(ep);
        ep->v->Release(ep);
    }

    if (This->RequestURI)
        Heap->v->Free(Heap, 0, (PTR) This->RequestURI);
    if (This->Hostname) 
        Heap->v->Free(Heap, 0, (PTR) This->Hostname);

    RELEASE(This->Server);
    RELEASE(This->Tokenizer);
    Mutex_Destroy(&This->mx);
    Condition_Destroy(&This->cd);
    Heap->v->Free(Heap, 0, (PTR) This);

    _PutHeap(Heap);
    return S_OK;
}

/*******XXXXXXXXXXXXXXXXXXXXXX****************/
#define pCUnknown(_i_) ((struct CUnknown *)(_i_))
#define iCUnknown(_p_) ((PIUNKNOWN)(_p_))
#define dCUnknown(_i_) (((struct CUnknownVtbl *)(((BYTE*)(_i_)->v) - offsetof(struct CUnknownVtbl, v)))->classd);

static UINT MCT HttpConnectorAddRef(PIHTTPCONNECTOR iThis)
{
    struct CUnknown *This = pCUnknown(iThis);
    UINT RefCnt = AtomicInc(&This->RefCnt);

    return RefCnt;
}

static UINT MCT HttpConnectorRelease(PIHTTPCONNECTOR iThis)
{
    struct CUnknown *This = pCUnknown(iThis);
    UINT RefCnt = AtomicDec(&This->RefCnt);

    if ((RefCnt & ~REFCNT_STATIC) == 0) {
        const struct _ClassDesc *classd = dCUnknown(iThis);
        if (classd->Destroy)
            classd->Destroy((PIUNKNOWN) iThis);
    }
    return RefCnt;
}
/*******XXXXXXXXXXXXXXXXXXXXXX****************/

const HTTPTOKENSTRING 
RequestMethodTokens[NUM_REQUEST_METHOD_TOKENS] = { 
#ifndef _MINIMIZE
    { "GET",    3,  TOKEN_GET },
    { "HEAD",   4,  TOKEN_HEAD },
#endif
    { "POST",   4,  TOKEN_POST },
#ifndef _MINIMIZE
    { "PUT",    3,  TOKEN_PUT },
    { "DELETE", 6,  TOKEN_DELETE },
    { "OPTIONS", 7, TOKEN_OPTIONS },
    { "TRACE",  5,  TOKEN_TRACE } 
#endif
};

PRIVATE SCODE MCT
HTTPRequestParseRequestLine(PIHTTPCONNECTOR Request, BOOL isResponse)
{
    SCODE sc;
    HttpConnector *  This        = (HttpConnector *) Request;
    PITOKENIZER     Tokenizer   = This->Tokenizer;
    PBYTE*          Content     = Tokenizer->v->GetContent(Tokenizer);
    PBYTE*          LookAhead   = Tokenizer->v->GetLookAhead(Tokenizer);
    PBYTE*          LookAhead2   = Tokenizer->v->GetLookAhead2(Tokenizer);
    UINT i, cmp;
    char numbuf[12];

    TRACE("HTTPRequestParseRequestLine");

    Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead);
    Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead2);

    /* Get rid of any trailing newlines that might be left from previous */
    EatWHITESPACE(Content);

    /* Parse the Method */
    if (!isResponse) {
        /* Match with the right method */
        for (i = 0; i < NUM_REQUEST_METHOD_TOKENS; ++i) {
            sc = Tokenizer->v->CmpToken(Tokenizer, 
                                        LookAhead, 
                                        (const PBYTE) RequestMethodTokens[i].String, 
                                        RequestMethodTokens[i].Length, 
                                        &cmp);
            if (sc == S_OK) {
                This->Method = RequestMethodTokens[i].Token;
                break; 
            } else if (FAILED(sc))
                goto err_ncmp;
        }

        /* Advance to the request URI */
        if (i < NUM_REQUEST_METHOD_TOKENS) {
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, 
                                             LookAhead,
                                             RequestMethodTokens[i].Length, NULL); 
            if (sc != S_OK)
                goto err_next; 
        } else {
            /* BUGBUG Extension methods need special treatment. We leave 
             * Request->Method with the value 0. To support extensions, 
             * derive a new HTTP server. 
             *
             * A different approach might use something like 
             *      union {
             *          HTTPREQUESTTOKEN Method;
             *          char*            ExtensionMethod;
             *      };
             * in the HTTPSERVER structure, but for now I don't want that.
             */

            if (!*LookAhead)
                return S_FALSE;
        
            DBG(("HTTPRequestParseRequestLine trouble with extension"));
            DebugBreak();
            return E_NOT_IMPLEMENTED;
        }
        EatSP(LookAhead); 

        /* Parse the Request-URI */
        assert(!This->RequestURI && This->RequestURILength == 0);
        sc = Tokenizer->v->AllocateAndCopyNextToken(Tokenizer, 
                                                    (const PBYTE*) LookAhead, 
                                                    LookAhead2, 
                                                    (const PBYTE) " ", 
                                                    1, 
                                                    &(This->RequestURI), 
                                                    &(This->RequestURILength));

        if (FAILED(sc)) goto err_next;
        if (sc == S_FALSE) sc = E_FAIL;
        if (FAILED(sc)) goto err_next; 
        EatSP(LookAhead2);
        Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);
    }

    /* Parse the HTTP-Version */
    sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 7, NULL); 
    if (sc == S_FALSE) sc = E_FAIL;
    if (FAILED(sc)) goto err_next; 
    if (**LookAhead == '1') {
        This->HTTPVersion = TOKEN_HTTP_1_1;
    } else {
        /* Fall back to HTTP 1.0 processing */
        This->HTTPVersion = TOKEN_HTTP_1_0;
        This->StatusCode = TOKEN_HTTP_VERSION_UNSUPP;
    }
    sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL); 
    if (sc == S_FALSE) sc = E_FAIL;
    if (FAILED(sc)) goto err_next; 

    if (isResponse) {
        UINT TokenLength;
        EatSP(LookAhead);
        Tokenizer->v->SetIterator(Tokenizer, LookAhead, LookAhead2);
        sc = Tokenizer->v->NextToken(Tokenizer, LookAhead2, (PBYTE) " ", 1, &TokenLength);
        if (FAILED(sc)) goto err_next; 
        if (TokenLength != 3) {
            return E_NOT_IMPLEMENTED;
        }
        sc = Tokenizer->v->CopyTokenA(Tokenizer, LookAhead,
                                      LookAhead2, numbuf);
        cmp = atoi(numbuf);
        if (cmp < TOKEN_CONTINUE) /* 100 */
            return E_NOT_IMPLEMENTED;
        This->Method = cmp;
        /* skip rest of line */
        sc = Tokenizer->v->NextToken(Tokenizer, LookAhead2, (PBYTE) "\r\n", 2, &TokenLength);
        if (FAILED(sc)) goto err_next; 
        Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);
    }

    /* Finish the line */
    EatOneCRLF(LookAhead); 

    TRACE("HTTPRequestParseRequestLine ok");
    Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);
    return S_OK;

err_eat_macro:
    TRACE("HTTPRequestParseRequestLine eat");
err_ncmp:
    TRACE("HTTPRequestParseRequestLine ncmp");
err_next:
    TRACE("HTTPRequestParseRequestLine next");

    /* Will reparse when E_WOULD_BLOCK */
    if (This->RequestURI && !isResponse) {
        PIHEAP Heap = CurrentHeap();
        Heap->v->Free(Heap, 0, This->RequestURI);
        This->RequestURI = NULL;
        This->RequestURILength = 0;
        This->HTTPVersion = 0;
    }

    return sc;
}

PRIVATE SCODE MCT
HTTPRequestParseHeaderFields(PIHTTPCONNECTOR Request)
{
    SCODE sc = S_OK;
    HttpConnector *      This        = (HttpConnector *) Request;
    PHTTPHEADER         Header      = &This->Headers;
    UINT                FieldNum    = 0;
    PHTTPHEADERFIELD    Field       = &Header->Fields[FieldNum];
    PIHEAP              Heap        = This->Heap;
    PITOKENIZER         Tokenizer   = This->Tokenizer;
    PBYTE*              Content     = Tokenizer->v->GetContent(Tokenizer);
    PBYTE*              LookAhead   = Tokenizer->v->GetLookAhead(Tokenizer);
    PBYTE*              LookAhead2   = Tokenizer->v->GetLookAhead2(Tokenizer);
    BOOL SaveIt;

    TRACE("HTTPRequestParseHeaderField");

    Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead);
    Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead2);

    /* Parse header fields until we encounter the CRLF which 
     * separates the header from the content
     */
    while (!IsCR(*LookAhead) && !IsLF(*LookAhead)) {

        UINT                i, cmp;
        const HTTPTOKENSTRING* Tokens;
        BYTE                Found = 0;

#ifndef _MINIMIZE
        /* Find the right field type */
        Tokens = &(RequestFieldTokens[0]);
        for (i = 0; i < NUM_REQUEST_FIELD_TOKENS; ++i) {
            sc = Tokenizer->v->CmpToken(Tokenizer, 
                                        LookAhead, 
                                        (PBYTE) Tokens[i].String, 
                                        Tokens[i].Length, 
                                        &cmp);
            if (sc == S_OK) {
                Found = 1;
                break; 
            }
            else if (FAILED(sc))
                goto err_ncmp;
        }
#endif

        if (! Found) {
            Tokens = &(EntityFieldTokens[0]);
            for (i = 0; i < NUM_ENTITY_FIELD_TOKENS; ++i) {
                sc = Tokenizer->v->CmpToken(Tokenizer, 
                                            LookAhead, 
                                            (PBYTE) Tokens[i].String, 
                                            Tokens[i].Length, 
                                            &cmp);
                if (sc == S_OK) {
                        Found = 1;
                        break; 
                }
                else if (FAILED(sc)) goto err_ncmp;
            }
        }

        if (Found) {
            SaveIt = TRUE;
            Field->StringLength = 0;
            assert(Field->Name == 0);
            Field->Name = Tokens[i].Token;
            
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, 
                                             LookAhead,
                                             Tokens[i].Length, NULL); 
            if (sc == S_FALSE) sc = E_FAIL;
            if (FAILED(sc)) goto err_next; 
        }
        else {
            SaveIt = FALSE;
            /* Didn't recognize the field */
            assert(Field->Name == 0);
            sc = Tokenizer->v->AllocateAndCopyNextTokenA(Tokenizer, 
                                                        LookAhead, 
                                                        LookAhead2, 
                                                        (PBYTE) ":", 
                                                        1, 
                                                        (char**)&(Field->Name),
                                                        &(Field->StringLength));
            if (sc == S_FALSE) sc = E_FAIL;
            if (FAILED(sc)) goto err_store_tkn;
            EatSP(LookAhead2);
            Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);

            if (!SaveIt) {
                /* XXX. Avoid allocation above instead XXX */
                Heap->v->Free(Heap,0, (PBYTE*) Field->Name);
                Field->Name = 0;
                Field->StringLength = 0;
            }
        }

        EatSP(LookAhead); /* Deviates from RFC 2068 */
        assert(IsCOLON(*LookAhead));
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL); 
        if (sc == S_FALSE) sc = E_FAIL;
        if (FAILED(sc)) goto err_next; 
        EatSP(LookAhead); 

        /* Store the field value. Failure means the field name string, if 
         * allocated, has to be deallocated. */
        Tokenizer->v->SetIterator(Tokenizer, LookAhead, LookAhead2);
        while (!IsCR(*LookAhead2) && !IsLF(*LookAhead2)) {
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
            if (sc == S_FALSE) sc = E_FAIL;
            if (FAILED(sc)) {
                DBG(("HTTPRequestParseHeaderFields: AdvanceSymbol failed\n"));
                goto err_next;
            }
        }

        {
            sc = Tokenizer->v->Distance(Tokenizer, LookAhead, LookAhead2,
                                        &Field->ValueLength);
            if (FAILED(sc)) return sc;
            assert(Field->Value == NULL);
            if (SaveIt) {
                Field->Value = Heap->v->Alloc(Heap,
                                              HEAP_DEBUG_NAME("http:field:value"),
                                              Field->ValueLength + 1, 0);
                if (Field->Value == NULL) goto err_alloc_fval;
                sc = Tokenizer->v->CopyTokenA(Tokenizer, LookAhead, LookAhead2, Field->Value);
                if (FAILED(sc)) {
                    DBG(("HTTPRequestParseHeaderFields: CopyToken failed\n"));
                    return sc;
                }

                Field->Value[Field->ValueLength] = '\0';
            }
            if (Field->Name == TOKEN_CONTENT_LENGTH) {
                if (!This->Chunked) /* Ignore this: chunking overrides */
                    This->ContentLength = atoi((char*) Field->Value);
            } else if (Field->Name == TOKEN_CONNECTION) {
                if (!_stricmp(Field->Value, "Keep-Alive")) {
                    //DBG(("Connection KeepAlive\n"));
                    This->KeepAlive = TRUE;
                } else {
                    DBG(("Connection KeepAlive disabled\n"));
                    This->KeepAlive = FALSE;
                }
            } else if (Field->Name == TOKEN_TRANSFER_ENCODING) {
                if (!_stricmp(Field->Value, "chunked")) {
                    DBG(("Chunked content\n"));
                    This->Chunked = TRUE;
                    This->ContentLength = 0; /* Ignore any Content-Length */
                } else {
                    This->Chunked = FALSE;
                }
            }
        }

        /* Advance to the next field */
        Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);
        assert(IsCR(*LookAhead) || IsLF(*LookAhead));
        EatOneCRLF(LookAhead);

        /* Prepare for more fields */
        if (SaveIt)
            ++FieldNum;
        if (FieldNum == NUM_FIELDS_IN_BLOCK) {
            assert(Header->More == NULL);
            Header->More = Heap->v->Alloc(Heap,
                        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("http:header:more"),
                        sizeof(HTTPHEADER), 0);
            if (Header->More == NULL) goto err_more_alloc;
            Header = Header->More;
            FieldNum = 0;
        }
        Field = &(Header->Fields[FieldNum]);
    }
    
    if (sc == E_TIMED_OUT)
        goto err_hopeless;
    /* else fall through to error handling */

    if (SUCCEEDED(sc)) {
        Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);
    } else {
err_more_alloc:
err_alloc_fval:
err_eat_macro:
err_store_tkn:
err_next:
err_ncmp:
err_hopeless:
        /* If not all headers were processed, clean the table
         * for potential restart.
         */
        HTTPRequestDeleteHeaders(Request);
    }
 
    return sc;
}

PRIVATE SCODE MCT
HTTPRequestStartBody(PIHTTPCONNECTOR iThis)
{
    METHOD_DECLARE(HttpConnector, StartBody);
    PITOKENIZER         Tokenizer   = This->Tokenizer;
    PBYTE*              Content     = Tokenizer->v->GetContent(Tokenizer);
    PBYTE*              LookAhead   = Tokenizer->v->GetLookAhead(Tokenizer);
    PBYTE*              LookAhead2   = Tokenizer->v->GetLookAhead2(Tokenizer);
    PBYTE*              LookAhead3   = Tokenizer->v->GetLookAhead3(Tokenizer);
    UINT                chunksize;

    sc = S_OK;

    if (This->State == HS_END_HEAD) {
        if (This->HTTPVersion != TOKEN_HTTP_1_1) {
            DBG(("Connection KeepAlive disabled for non 1.1 HTTP version for the benefit of proxies\n"));
            This->KeepAlive = FALSE;
        }

        sc = Tokenizer->v->SetContentLength(Tokenizer, 2);
        CHECKSC("SetContentLength(2)");
        This->State = HS_END_HEAD2;
    }

    if (This->State == HS_END_HEAD2) {
        EatOneCRLF(Content);

        This->State = HS_BODY;

        if (This->Chunked) {
            Tokenizer->v->SetContentLength(Tokenizer, INT_MAX);
            Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead);
        }
    }

    /* Restartable chunking: LookAhead always points to next chunk head
     * at entry to this loop.
     */
    if (This->Chunked && This->State == HS_BODY) do {
        char *chunk;
        UINT distance;

        Tokenizer->v->SetIterator(Tokenizer, LookAhead, LookAhead2);
        Tokenizer->v->SetIterator(Tokenizer, LookAhead, LookAhead3);

        /* Remove any spaces in before chunk header but get them zapped */
        EatWHITESPACE(LookAhead2);

        sc = Tokenizer->v->AllocateAndCopyNextTokenA(Tokenizer, 
                                                     LookAhead2,
                                                     LookAhead3, 
                                                     (PBYTE) "\r\n", 2, 
                                                     &chunk, &distance);
        CHECKSC("NextToken");
        if (sc != S_OK)
            goto Out;
        assert(chunk);

        chunksize = atoh(chunk);
        free(chunk);

        if (chunksize) {
            EatOneCRLF(LookAhead3);

            sc = Tokenizer->v->Distance(Tokenizer, LookAhead, LookAhead3,
                                        &distance);
            CHECKSC("Distance of chunk head");

            /* Pull in trailing data before split */
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead3,
                                             chunksize, NULL);
            CHECKSC("Pre-advance through chunk");

            /* Now next chunk is in, zap head */
            sc = Tokenizer->v->CarveBuffer(Tokenizer, LookAhead,
                                           distance, NULL);
            assert(sc != E_WOULD_BLOCK);
            CHECKSC("CarveBuffer");

            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead,
                                             chunksize, NULL);
            assert(sc != E_WOULD_BLOCK);
            CHECKSC("Advance through chunk");
        } else {
            /* Last null chunk, truncate end by 3 */
            sc = Tokenizer->v->CarveBuffer(Tokenizer, NULL,
                                           distance + 2, NULL);
            CHECKSC("Carve last");
        }

        This->ContentLength += chunksize;
    } while (chunksize > 0);
    This->State = HS_DISPATCH;

    /* All header fields have been parsed - advance to the body */
    Tokenizer->v->SetContentLength(Tokenizer, This->ContentLength);

    if (sc == E_END_OF_DATA)
        sc = S_FALSE;

    METHOD_CLEANUP;
err_eat_macro:
    return sc;
}

/* Data has arrived for session or connector. Process it! */
PRIVATE SCODE MCT HttpConnectorReceiveMessage(PIHTTPCONNECTOR iThis)
{
    METHOD_DECLARE(HttpConnector, ReceiveMessage);
    PITOKENIZER Tokenizer;
    PBYTE *Content;
    BOOL isResponse = !This->isServer;

    DBGME(0,printf("HTTPSessionProcess: Processing request state=%d\n",
                   This->State));

    //Mutex_Lock(&This->mx); // not necessary: socket is disabled or, when udp,
                             // we are processing a separate buffer
    Tokenizer = This->Tokenizer;
    if (Tokenizer) {
        Content = Tokenizer->v->GetContent(Tokenizer);
    } else {
        Content = NULL;
        assert(This->State == HS_START);
    }

    for (;;) switch (This->State) {
    case HS_START:
        /* (re-) initialize connector/request */
        sc = HttpConnectorStartReceive(iThis, isResponse);
        CHECKSC("HttpConnectorStartReceive");

        /* A new tokenizer was potentially created */
        Tokenizer = This->Tokenizer;
        assert(Tokenizer);
        Content = Tokenizer->v->GetContent(Tokenizer);

        //assert(This->Refs == 1 || isResponse);
        //iThis->v->AddRef(iThis);  /* ??? */

        This->State = HS_START2;
        DBGME(0,printf("HTTPServerService: Parsing request\n"));

    case HS_START2:
        /* Reinitialize tokenizer to read more data */
        sc = Tokenizer->v->SetContentLength(Tokenizer, INT_MAX);
        CHECKSC("SetContentLength");
        if (sc == S_FALSE)
            goto Out;               /* EOF */

        if ((This->Flags & PARSE_URL_FLAGS_HTTP) == 0) {
            This->State = HS_DISPATCH;
            continue;
        }
        This->State = HS_LINE1;

    case HS_LINE1:
        /* Parse the request line */
        sc = HTTPRequestParseRequestLine(iThis, isResponse);
        CHECKSC("ParseRequestLine");
        if (sc == S_FALSE)
            goto Out;               /* EOF */
        This->State = HS_HEADERS;

    case HS_HEADERS:
        /* In case there was a packet boundary and E_WOULD_BLOCK at end of req line
         * make sure Content does not point past end of buffer
         */
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, Content, 0, NULL);
        CHECKSC("Resync Content 1");

        /* Parse the request header, one field at a time */
        sc = HTTPRequestParseHeaderFields(iThis);
        CHECKSC("ParseHeaderFields");
        if (sc == S_FALSE)
            goto Out;               /* EOF */
        This->State = HS_END_HEAD;

    case HS_END_HEAD:
    case HS_END_HEAD2:
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, Content, 0, NULL);
        CHECKSC("Resync Content 2");

        sc = HTTPRequestStartBody(iThis);
        CHECKSC("HTTPRequestStartBody");

    case HS_DISPATCH:
    case HS_XML:
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, Content, 0, NULL);
        CHECKSC("Resync Content 3");

        /* We don't know what to do with any message content, 
         * so the responsibility for finding out gets shifted 
         * over to whatever object eventually processes the 
         * request. It will use the parsed headers to find out 
         * about possible content. This responsibly explicitly 
         * belongs with the IHTTPPROCESS implementation... */
        DBGME(0,printf("HTTPServerService: Dispatching request\n"));
        sc = HTTPRequestDispatch(iThis, isResponse);
        /* Note that another thread may have been started that accesses This */
        if (sc != E_WOULD_BLOCK) {
            //Mutex_Unlock(&This->mx);
            //RELEASE_NULL(iThis); /* ??? */
            //Mutex_Lock(&This->mx);
        }
        CHECKSC("HttpRequestDispatch");
        goto Out;

    default:
    case HS_RUNNING:
        assert(This->State == HS_RUNNING);
        CHECKSETSC(FALSE, E_OBJECT_BUSY, "process state");
        /* Restart later */
    }

METHOD_CLEANUP;
    //Mutex_Unlock(&This->mx);
    if ((FAILED(sc) && sc != E_WOULD_BLOCK
         && sc != E_OBJECT_BUSY && sc != E_RESTART)
       || sc == S_FALSE)
    {
        //        RELEASE(iThis);         /* ??? */
//        assert(This->State == HS_START);
        This->ProtocolState = HPS_DEAD;
    }
    return sc;
}


/* HTTP message content parsing */
/* HTTP request data access methods */
SCODE MCT
HttpConnectorGetRequestURI(PIHTTPCONNECTOR Request,
                           PBYTE* URI,
                           INT*   Length)
{
    HttpConnector * This = (HttpConnector *) Request;
    *URI = (PBYTE) This->RequestURI;
    *Length = This->RequestURILength;

    return S_OK;
}

SCODE MCT
HttpConnectorGetHeaderField(PIHTTPCONNECTOR Request, 
                            INT /*HTTPTOKEN*/ FieldToken, 
                            PHTTPHEADERFIELD* Field)
{
    HttpConnector *      This        = (HttpConnector *) Request;
    PHTTPHEADER         Header      = &This->Headers;
    PHTTPHEADERFIELD    Current;

    TRACE("HTTPRequestGetHeaderField");

    *Field = NULL;
    do {
        UINT i;
        Current = &Header->Fields[0];
        for (i = 0; i < NUM_FIELDS_IN_BLOCK; ++i) {
            if (Current->Name == FieldToken) {
                /* Paydirt */
                *Field = Current;
                return S_OK;
            }
            else if (Current->Name == 0) {
                /* Done looking through all fields 
                 * (but there are more fields in this block) */
                return E_NAME_NOT_FOUND;
            }
            ++Current;
        }

        /* We've exhausted the header block but 
         * need to check if there are more */
        Header = Header->More;
    }
    while (Header != NULL);

    /* Getting down here means that the last block of fields was fully 
     * utilized, but the requested field is nowhere to be found */
    return E_NAME_NOT_FOUND;
}

PITOKENIZER MCT
HttpConnectorGetContent(PIHTTPCONNECTOR Request)
{
    return ((HttpConnector *) Request)->Tokenizer;
}

extern SCODE MCT
Create_Tokenizer(PIUNKNOWN tok, PITOKENIZER *Tokenizer);
/* HTTP request dispatching */
SCODE MCT
HTTPRequestDispatch(PIHTTPCONNECTOR iThis, BOOL isResponse)
{
    METHOD_DECLARE(HttpConnector, Dispatch);
    PIHTTPPROCESS   Process     = NULL;
    PITOKENIZER SubTokenizer = NULL;
    _TCHAR *URI = This->RequestURI;
    const _TCHAR *ContentType = NULL;
    PIBUFFER TmpBuffer = NULL;
    char *s1, *s2;
    _TCHAR *ts1, *ts2;
    PIUNKNOWN Target = NULL;
    PICONVERT convert = NULL;
    PISAXHANDLER SAXCallbacks = NULL;
    PISAXPARSER SaxParser = NULL;
    PITOKENIZER Tokenizer = iThis->v->GetContent(iThis);
    BOOL isDone = FALSE;

    if (This->State == HS_DISPATCH) {
        TcpServer *Server = (TcpServer *) ((HttpConnector *) This)->Server;
        PHTTPHEADERFIELD Field;
//        PIBUFFER im;

        BOOL isSSOAP = FALSE;

#if 0
        sc = iThis->v->AllocateBuffer(iThis, &im);
        assert(!FAILED(sc));
        assert(!This->ResponseFile);
        sc = im->v->QueryInterface(im, &IID_IFile, (void**) &This->ResponseFile);
        im->v->Release(im);
        assert(!FAILED(sc));
#endif

        if ((This->Flags & PARSE_URL_FLAGS_HTTP)
            && (This->isServer || (This->Flags & PARSE_URL_FLAGS_SOAP)))
        {
            /* First look at Content-type and decide if SOAP */
            This->Flags = PARSE_URL_FLAGS_HTTP;
            sc = iThis->v->GetHeaderField(iThis, TOKEN_CONTENT_TYPE, &Field);
            if (SUCCEEDED(sc)) {
                s1 = Field->Value;
                s2 = _strsplit(s1, ";");
                //DBG(("Content-type: [%s]", s1));
#ifdef _UNICODE
                ContentType = _strtow(s1);
#else
                ContentType = s1;
#endif
                if (!strcmp(s1, "xml/soap")
                    || !strcmp(s1, "application/xml")
                    || !strcmp(s1, "application/soap+xml"))
                {
                    This->Flags |= PARSE_URL_FLAGS_SOAP;
                } else if (!strcmp(s1, "text/xml")) {
                    This->Flags |= PARSE_URL_FLAGS_SOAP
                                | PARSE_URL_FLAGS_OLD_CONTENT_TYPE;
                } else if (!strcmp(s1, "x-aes-soap")) {
                    //DBG((" -> http/EncryptedSOAP\n"));
                    isSSOAP = TRUE;
                    This->Flags |= PARSE_URL_FLAGS_SOAP | PARSE_URL_FLAGS_AES;
                } else {
                    //DBG(("\n"));
                }
#if 0
                /* Put the semicolon back */
                if (s2)
                    *(s2-1) = '\0';
#endif
            }
        }

        /* Lookup the requested dispatch target in the system namespace */
        if ((This->Flags & PARSE_URL_FLAGS_SOAP) == 0
            && !isSSOAP && URI && URI[1] == _T('\0'))
        {
            /* System namespace requested */
            Server->NameSpace->v->AddRef(Server->NameSpace); /* account for copy */
            This->Target = (PIUNKNOWN) Server->NameSpace;
        } else if (isSSOAP) {
            /* secured soap */
            UINT Size, adSize;
            PBYTE*  Content = NULL;
            PBYTE*  LookAhead = NULL;

            Tokenizer = iThis->v->GetContent(iThis);
            Content = Tokenizer->v->GetContent(Tokenizer);
            LookAhead = Tokenizer->v->GetLookAhead(Tokenizer);

            Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead);

            sc = iThis->v->AllocateBuffer(iThis, &TmpBuffer);
            /* Copy tokenizer content into new buffer */
            /* XXX Could often use existing buffer */
            while(1) {
                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, &adSize);
                if (FAILED(sc) && sc != E_END_OF_DATA) goto error;
                if (sc == E_END_OF_DATA) adSize++;

                TmpBuffer->v->WriteAt(TmpBuffer, UINT64_MAX, 
                                      (PBYTE)*Content, adSize, &Size);
                Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);
                if (adSize < 1 || sc == E_END_OF_DATA) break;
            }
            
            sc = AESBuffer(SKEY_STRENGTH, INIT_SKEY,
                           TmpBuffer, CYPHER_DECRYPT);
            if (FAILED(sc)) {
                DBG(("httprequest.c: Decrypt failed\n"));
                goto error;
            }

            sc = Create_Tokenizer((PIUNKNOWN)TmpBuffer, &SubTokenizer);
            RELEASE_NULL(TmpBuffer);
            if (sc != S_OK) {
                DBG(("The tokenizer timeout occurred\n"));
                sc = E_NO_INTERFACE;
                goto error;
            }

            Content = Tokenizer->v->GetContent(Tokenizer);
            LookAhead = Tokenizer->v->GetLookAhead(Tokenizer);
        }

        if (!isResponse) {
            if ((This->Flags & PARSE_URL_FLAGS_SOAP) || isSSOAP) {
                ts2 = NULL;
#if 0
                ts1 = _T("COB/soap.cob");
#else
                ts1 = NULL;
                This->Target = (PIUNKNOWN) &HttpProcessSoap;
                This->Target->v->AddRef(This->Target); /* symmetry */
#endif
            } else {
                NAME_SPACE_FLAGS nsflags = NAME_SPACE_READ;
            
                ts1 = URI + 1; /* w/o leading slash */
                ts2 = _tcssplit(ts1, _T("?")); /* trailing args */

#ifndef _MINIMIZE
                if (This->Method == TOKEN_PUT)
                    nsflags |= NAME_SPACE_WRITE | NAME_SPACE_CREATE;
#endif
                sc = BindToObject(Server->NameSpace, ts1, nsflags,
                                  &IID_IUnknown, (void**)&This->Target);

                if (ts2)             /* put question mark back */
                    *(ts2-1) = _T('?');

                if (FAILED(sc) || sc == S_FALSE) {
                    /* Ouch, target nowhere to be found */
                    This->StatusCode = TOKEN_NO_CONTENT;
                    goto error;
                }
            }
        }

        This->SaxState = 0;
        This->State = HS_XML;
    }

    if (!isResponse) {
        /* Target found - query for the appropriate interface */
        sc = This->Target->v->QueryInterface(This->Target, &IID_IHttpProcess, 
                                             (void**) &Process);
#ifndef _MINIMIZE
        if (sc == E_NO_INTERFACE) {
            sc = This->Target->v->QueryInterface(This->Target, &IID_INameSpace,
                                                 (void**) &Target);
            if (SUCCEEDED(sc)) {
                Process = (PIHTTPPROCESS) &HttpProcessNS;
                /* httpserver.h has a funny idea of multiple inheritance
                 * and has some NULL methods. Clean up some day.
                 */
                if (Process->v->AddRef)
                    Process->v->AddRef(Process); /* symmetry with code above: we release later */
            }
        }
        if (sc == E_NO_INTERFACE) {
            sc = This->Target->v->QueryInterface(This->Target, &IID_IFile, 
                                                 (void**) &Target);
            if (! FAILED(sc)) {
                Process = (PIHTTPPROCESS) &HttpProcessFile;
                if (Process->v->AddRef)
                    Process->v->AddRef(Process);
            }
        }
#endif /*_MINIMIZE*/
        if (sc == E_NO_INTERFACE) {
            /* Ouch, target doesn't support any useful interface */
            This->StatusCode = TOKEN_METHOD_NOT_ALLOWED;
        }
        if (FAILED(sc)) goto error;

        if (This->Target && Target) { /* don't leak the previous ref */
            This->Target->v->Release(This->Target);
            This->Target = Target;
        }
    }
    This->StatusCode = TOKEN_OK;

    /* Dispatch the request to the appropriate handler */
    switch (This->Method) {
#ifndef _MINIMIZE
    case TOKEN_HEAD: /* Fall-through */
    case TOKEN_GET:
        RELEASE_NULL(This->ResponseFile); /* XXX was unnecessarily created above */
        sc = Process->v->Get(Process, iThis, &This->ResponseFile);
        break;
#endif
    case TOKEN_POST:
        RELEASE_NULL(This->ResponseFile); /* XXX was unnecessarily created above */
        sc = Process->v->Post(Process,
                              This->Flags,
                              SubTokenizer ? SubTokenizer : This->Tokenizer,
                              URI,
                              (PIBUFFER*) &This->ResponseFile,
                              (PIUNKNOWN *) &This->Deserializer,
                              (UINT*) &This->State,
                              iThis);
        if (FAILED(sc)) goto error;

        break;
#ifndef _MINIMIZE
    case TOKEN_PUT:
        sc = Process->v->Put(Process, iThis);
        break;
    case TOKEN_DELETE:
        sc = Process->v->Delete(Process, iThis);
        break;
    case TOKEN_OPTIONS:
        sc = Process->v->Options(Process, iThis);
        break;
    case TOKEN_TRACE:
        sc = E_NOT_IMPLEMENTED;
        break;
#endif
    case TOKEN_OK:
    case TOKEN_INTERNAL_SERVER_ERROR:
        if (This->Flags & PARSE_URL_FLAGS_SOAP) {
            assert(This->cont && isResponse);

            if (!This->Deserializer) {
                sc = BindToObject(CurrentNameSpace(),
                                  _T("COB/convert.cob"), NAME_SPACE_READ,
                                  &IID_IConvert, (void **) &convert); 
                CHECKSC("BIND soap factory");
                sc = convert->v->CreateDeserializer(convert, _T(""), 0,
                                                    This->cont,
                                                    &This->Deserializer);
                CHECKSC("CreateDeserializer");
            }

            sc = BindToObject(CurrentNameSpace(),
                              _T("COB/sax.cob"), NAME_SPACE_READ,
                              &IID_ISaxParser, (void **) &SaxParser);
            CHECKSC("Bind(COB/sax.cob");

            sc = SaxParser->v->Parse(SaxParser,
                                 SubTokenizer ? SubTokenizer : This->Tokenizer,
                                     This->Deserializer,
                                     &This->SaxState);
        } else if (This->Method == TOKEN_OK) {
            sc = S_OK;
            if (This->ResponseFile) { /* HTTP GET */
                /* Extract content while leaving the tokenizer ready for
                 * next packet
                 */
                PIBUFFER buf = NULL;
                PBYTE* Content = Tokenizer->v->GetContent(Tokenizer);
                UINT Written;

                sc = Tokenizer->v->CarveBuffer(Tokenizer, Content, UINT_MAX,
                                               &buf);
                CHECKSC("CarveBuffer extract response payload");

                sc = buf->v->ReadWriteFile(buf, This->ResponseFile, TRUE,
                                           UINT64_ZERO, 0, UINT_MAX,
                                           &Written);
                buf->v->Release(buf);
                CHECKSC("write response payload to file");
            }
        } else {
            sc = E_SCHEMA_CHECK;
        }
#if USE_SERVER_FOR_RESPONSE
        //isDone = TRUE;
#else
        /* Partial RequestDone */
        if (!This->isServer && This->HTTPVersion == TOKEN_HTTP_1_0)
            RELEASE_NULL(This->Endpt); /* Create new socket next */
        This->Method = 0;           /* HTTP GET/POST/... */
        This->HTTPVersion = 0;
        HTTPRequestDeleteHeaders(iThis);
        RELEASE_NULL(This->Deserializer);
        This->SaxState = 0;
        This->State = HS_START;
#endif
        CHECKSC("Parse");

        break;
    case TOKEN_CONTINUE:
        assert(This->ContentLength == 0);
        /* Partial RequestDone */
        This->Method = 0;           /* HTTP GET/POST/... */
        This->HTTPVersion = 0;
        HTTPRequestDeleteHeaders(iThis);
        This->State = HS_START;
        sc = E_RESTART;
        break;
        
    default:
        sc = E_INVALID_METHOD;
        break;
    }

METHOD_CLEANUP;
error:
#ifdef _UNICODE
    if (ContentType)
        free((PTR) ContentType);
#endif
    if (FAILED(sc)) {
      /* Oops, let's try to mollify the client and process something else */
        if (/*This->StatusCode == TOKEN_NOT_SET ||*/ This->StatusCode == TOKEN_OK)
            This->StatusCode = ScToHttpStatus(sc);
    }
    assert(sc != S_FALSE);
    if (!isResponse && This->isServer
        && sc != S_NO_RESPONSE && sc != E_WOULD_BLOCK && sc != E_RESTART)
    {
        if (!This->ResponseFile) {
            PIBUFFER im;
            sc = iThis->v->AllocateBuffer(iThis, &im);
            assert(!FAILED(sc));
            sc = im->v->QueryInterface(im, &IID_IFile,
                                       (void**) &This->ResponseFile);
            im->v->Release(im);
            assert(!FAILED(sc));
        }

        if (This->ResponseFile) {
            sc = This->v->SendMessage((PIHTTPCONNECTOR) This,
                                      This->StatusCode,
                                      NULL,
                                      This->ResponseFile,
                                      NULL);
        }

        /* Since we already sent the reply we are now ready to receive more */
        isDone = TRUE;
    }
    /* XXX Should we return this error or the previous error? Check caller.*/

    if (Process && Process->v->Release)
        Process->v->Release(Process);

    if (sc != E_WOULD_BLOCK)
        RELEASE_NULL(This->ResponseFile);

    RELEASE(TmpBuffer);
    RELEASE(SubTokenizer);
    RELEASE(SaxParser);
    RELEASE(SAXCallbacks);
    RELEASE(convert);
    //assert(!isResponse == !This->cont); // SOAP only
    //RELEASE_NULL(This->cont);

    if (isDone)
        iThis->v->Done(iThis);

    return sc;
}

BOOL HTTPSessionIsKeepAlive(HttpConnector * Session)
{
    TRACE("HTTPSessionIsKeepAlive");

    return Session->KeepAlive;
}
