/* Copyright (c) Microsoft Corporation. All rights reserved. */

//#define _DEBUG 1
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <netif.h>
#include "http.h"
#include <tcpserver.h>
#include "httpconstants.h"
#include "httpserver.h"
#include "tokenizerutils.h"
#include "../soap/soap.h"

PRIVATE void LogMsg(PIBUFFER hdr)
{
#if defined(_DEBUG) && defined(i386)
    /* Spit out the serialized data into file for examination */
    PIFILE file;
    UINT wrote;
    SCODE sc = BindToObject(CurrentNameSpace(), _T("fs/httpconn-out.txt"),
                            NAME_SPACE_CREATE | NAME_SPACE_WRITE,
                            &IID_IFile, (void **) &file);
    if (SUCCEEDED(sc)) {
        hdr->v->ReadWriteFile(hdr, file, TRUE,
                              UINT64_MAX, 0, UINT_MAX, &wrote);
        file->v->Release(file);
    }
#endif
}

PRIVATE const char OK[]                       = "200 ok";
PRIVATE const char NO_CONTENT[]               = "204 no content";

PRIVATE const char BAD_REQUEST[]              = "400 bad request";
PRIVATE const char NOT_FOUND[]                = "404 not found";
PRIVATE const char METHOD_NOT_ALLOWED[]       = "405 method not allowed";
PRIVATE const char UNSUPP_MEDIA_TYPE[]        = "415 unsupported media type";

PRIVATE const char INTERNAL_SERVER_ERROR[]    = "500 internal server error";
PRIVATE const char NOT_IMPLEMENTED[]          = "501 not implemented";
PRIVATE const char HTTP_VERSION_UNSUPP[]    = "505 http version not supported";

const char *HttpTokenToString(UINT Token)
{
    BYTE const *    MessageSegment;

    switch (Token) {
    case TOKEN_OK:
#ifndef _MINIMIZE
    case TOKEN_OPTIONS_RESPONSE:
    case TOKEN_GET_RESPONSE:
    case TOKEN_HEAD_RESPONSE:
    case TOKEN_PUT_RESPONSE:
    case TOKEN_DELETE_RESPONSE:
    case TOKEN_TRACE_RESPONSE:
    case TOKEN_EXTMETH_RESPONSE:
#endif
    case TOKEN_POST_RESPONSE:
        MessageSegment = (BYTE*) OK;
        break;
    case TOKEN_NO_CONTENT:
        MessageSegment      = (BYTE*) NO_CONTENT;
        break;
    case TOKEN_METHOD_NOT_ALLOWED:
        MessageSegment      = (BYTE*) METHOD_NOT_ALLOWED;
        break;
#ifndef _MINIMIZE
    case TOKEN_UNSUPP_MEDIA_TYPE:
        MessageSegment      = (BYTE*) UNSUPP_MEDIA_TYPE;
        break;
    case TOKEN_BAD_REQUEST:
        MessageSegment      = (BYTE*) BAD_REQUEST;
        break;
    case TOKEN_GET:
        MessageSegment = (BYTE*) "GET";
        break;
    case TOKEN_HEAD:
        MessageSegment = (BYTE*) "HEAD";
        break;
    case TOKEN_PUT:
        MessageSegment = (BYTE*) "PUT";
        break;
    case TOKEN_DELETE:
        MessageSegment = (BYTE*) "DELETE";
        break;
    case TOKEN_OPTIONS:
        MessageSegment = (BYTE*) "OPTIONS";
        break;
    case TOKEN_TRACE:
        MessageSegment = (BYTE*) "TRACE";
        break;
#endif
    case TOKEN_POST:
        MessageSegment = (BYTE*) "POST";
        break;
    case TOKEN_NOT_FOUND:
        MessageSegment      = (BYTE*) NOT_FOUND;
        break;
    case TOKEN_HTTP_VERSION_UNSUPP:
        MessageSegment      = (BYTE*) HTTP_VERSION_UNSUPP;
        break;
    case TOKEN_SOAP_FAULT:
        MessageSegment      = (BYTE*) INTERNAL_SERVER_ERROR;
        break;

    default: /* fall-thru */
    case TOKEN_INTERNAL_SERVER_ERROR:
        MessageSegment      = (BYTE*) INTERNAL_SERVER_ERROR;
        break;
    }

    return (char*) MessageSegment;
}


#define w(_s_) \
    MACRO_BEGIN \
    const char *s=(_s_);\
    sc = ReadWriteFile((PIFILE)hdr, TRUE, UINT64_MAX, (PBYTE) s, strlen(s), NULL); \
    CHECKSC("Write"); \
    MACRO_END

#define wt(_s_) \
    MACRO_BEGIN \
    const _TCHAR *s=(_s_);\
    BYTE c; \
    for (; *s; s++) { \
      c = (BYTE) *s; /* XXX UTF8 */ \
      sc = ReadWriteFile((PIFILE) hdr, TRUE, UINT64_MAX, &c, 1, NULL); \
      CHECKSC("Write"); \
    } \
    MACRO_END

SCODE MCT HttpSend(
    /*optional*/ PIFILE ContentFile,
    PIENDPOINT ep,

    UINT Flags,
    const _TCHAR *HostName,
    _TCHAR *LocalUrl,
    const char *BasicKey,       /* RFC 2617 basic authentication. UNSECURE. */
    struct sockaddr *sockaddr,
    UINT sockaddr_size,

    const _TCHAR *SoapAction,
    INT HTTPVersion,
    UINT HttpToken,             /* POST/GET/... or "200"/... */
    BOOL ConnectionClose)       /* Generate Connection: close */
{
    METHOD_DECLARE_NOTHIS(HttpConnector, Send);
    UINT size, amount;
    UINT64 size64, Position = UINT64_ZERO;
    PIBUFFER data = NULL, hdr = NULL, next = NULL;
    char num[20];
    BOOL toBeChunked = FALSE;

    const char *HttpMethod = HttpTokenToString(HttpToken);
    if (!HttpMethod)
        return E_INVALID_PARAMETER;

    /* Is this a buffer or a vanilla file? */
    if (ContentFile) {
        sc = ContentFile->v->QueryInterface(ContentFile, &IID_IBuffer,
                                            (void**) &data);
        if (SUCCEEDED(sc))
            ContentFile = NULL;     /* use buffer from now on */
    }

    /* If there is data to send, figure out size */
    if (HttpToken == TOKEN_OK
        || HttpToken == TOKEN_SOAP_FAULT
        || HttpToken == TOKEN_POST
#ifndef _MINIMIZE
        || HttpToken == TOKEN_PUT
        || HttpToken == TOKEN_GET_RESPONSE
#endif
        || HttpToken == TOKEN_POST_RESPONSE)
    {
        if (data)
            sc = data->v->GetSize(data, &size64);
        else if (ContentFile)
            sc = ContentFile->v->GetSize(ContentFile, &size64);
        else
            size64 = UINT64_ZERO;

        if (FAILED(sc))
            HttpToken = TOKEN_INTERNAL_SERVER_ERROR;

        /* First make sure we _can_ read the content */
        /* BUGBUG... don't want to hang on synchronous read... */
        if (!data && ContentFile && HttpToken != TOKEN_POST) {
            sc = ContentFile->v->ReadAt(ContentFile,
                                        UINT64_ZERO, NULL, 0, &amount);
            if (FAILED(sc)) {
                HttpToken = TOKEN_METHOD_NOT_ALLOWED;
            } else if (Int64IsZero(size64)) {
                /* Unknown size: chunk it */
                toBeChunked = TRUE;
            }
        }

    } else {
        size64 = UINT64_ZERO;
    }

    if (Uint64High32Bits(size64) != 0) {
#ifndef _MINIMIZE
        HttpToken = TOKEN_REQ_ENTITY_TOO_LARGE;
#endif
        CHECKSETSC(FALSE, E_INVALID_PARAMETER, "bogus request size");
        size = 0;
    } else {
        size = Uint64Low32Bits(size64);
    }

    /* Encrypt the buffer now if we have all the content */
    if (Flags & PARSE_URL_FLAGS_AES) {
        /* Pad if necessary up to AES block size */
        size = __ROUND(size, CHUNKSIZE);
    }

    if ((Flags & PARSE_URL_FLAGS_HTTP) == 0) { 
        /* No headers, just data. Just raw TCP or UDP message. */
        /* If this wasn't in a buffer yet, we need to read it now */
        if (ContentFile) {
            assert(!data);
            sc = ep->v->AllocateBuffer(ep, 0, &data);
            CHECKSC("Alloc IBuffer non-HTTP");
#define UDP_MAX 32000           /* XXX */
            sc = data->v->ReadWriteFile(data, ContentFile, FALSE,
                                        UINT64_ZERO, UINT_MAX, UDP_MAX,
                                        &size);
            CHECKSC("ReadWriteFile non-HTTP");
            ContentFile = NULL;
        }
        hdr = data;
        data = NULL;
        /* Some sanity checking: */
        if (!sockaddr || (sockaddr_size == 0 && (Flags & PARSE_URL_FLAGS_UDP)))
            CHECKSETSC(FALSE, E_INVALID_PARAMETER,
                       "UDP send - requires sockaddr");
    } else {
        /* Create the http stuff */
        sc = ep->v->AllocateBuffer(ep, 0, &hdr);
        CHECKSC("AllocHeader IBuffer");

        /* create http header */
        if (HostName) { /* CLIENT */
            /* POST /COB/calc.cob HTTP/1.1
             * Content-Type: application/soap+xml;charset=UTF-8
             * HOST: 127.0.0.1
             * Content-Length: 579
             */

            w(HttpMethod);
            w(" ");
            wt(LocalUrl);
            if (HTTPVersion == TOKEN_HTTP_1_0) {
                w(" HTTP/1.0\r\n");
            } else {
                w(" HTTP/1.1\r\n");
            }
        } else { /* SERVER */
            /* HTTP/1.1 200 ok
             * Content-Type: application/soap+xml; charset=UTF-8
             * Content-Length: 545
             */
            if (HTTPVersion == TOKEN_HTTP_1_0) {
                w("HTTP/1.0 ");
            } else {
                w("HTTP/1.1 ");
            }
            w(HttpMethod);
            w("\r\n");
        }

        if (Flags & PARSE_URL_FLAGS_AES) {
            w("Content-Type: x-aes-soap;charset=UTF-8\r\n");
        } else if (Flags & PARSE_URL_FLAGS_SOAP) {
            if (Flags & (CONT_FLAGS_SOAP_1_1 | PARSE_URL_FLAGS_OLD_CONTENT_TYPE))
                w("Content-Type: text/xml\r\n");
            else
                w("Content-Type: application/soap+xml;charset=UTF-8\r\n");
        }

        if (HostName) {
            w("HOST: ");
            wt(HostName);
            w("\r\n");
        } else if (ConnectionClose) {
            w("Connection: close\r\n");
        }
#if 1
        if (SoapAction) {
            w("SOAPAction: \"");
            wt(SoapAction);
            w("\"\r\n");
        }
#endif
        /* If the buffer is too big we'll chunk it. Also if size is unknown. */
        if (toBeChunked || size > MAX_CHUNK_SIZE) {
            toBeChunked = TRUE;
            w("transfer-encoding: chunked\r\n");
        } else {
            w("Content-Length: ");
            sprintf(num, "%d", size);
            w(num);
            w("\r\n");
        }

        if (BasicKey) {
            w("Authorization: Basic ");
            w(BasicKey);
            w("\r\n");
        }

        /* Finish off the message header */
        w("\r\n");
    }

    do {
        if (ContentFile) {
            assert(!data);
            sc = ep->v->AllocateBuffer(ep, 0, &data);
            CHECKSC("Alloc IBuffer HTTP");
            sc = data->v->ReadWriteFile(data, ContentFile, FALSE,
                                        Position, 0, MAX_CHUNK_SIZE,
                                        &size);
            CHECKSC("ReadWriteFile HTTP");
        } else if (data) {
            next = data->v->Split(data, MAX_CHUNK_SIZE);
            sc = data->v->GetSize(data, &size64);
            CHECKSC("IBuffer GetSize");
            size = Uint64Low32Bits(size64);
        } else
            size = 0;

        if (toBeChunked) {
            /* Write variable size chunk length field into header */
            sprintf(num, "%x\r\n", size);
            w(num);
        }

        if (data && (Flags & PARSE_URL_FLAGS_AES)) {
            sc = AESBuffer(SKEY_STRENGTH, INIT_SKEY, data, CYPHER_ENCRYPT);
            CHECKSC("AESBuffer HTTP");
        }

        if (data) {
            sc = hdr->v->Append(hdr, data);
            CHECKSC("Append mbuf");
            RELEASE_NULL(data);
        }

        //if (toBeChunked)
        //    w("\r\n");

        /* XXX Homogenize new line + length production later... */
        if ((Flags & PARSE_URL_FLAGS_SOAP) == 0)
            w("\r\n");

        DBG(("Sending %s message to socket %x %s:%d\n",
             (Flags & PARSE_URL_FLAGS_UDP) ? "UDP"
             : (Flags & PARSE_URL_FLAGS_HTTP) ? "HTTP" : "TCP",
             ep,
             sockaddr ? inet_ntoa(((struct sockaddr_in*) sockaddr)->sin_addr) : "",
             sockaddr ? ntohs(((struct sockaddr_in*)sockaddr)->sin_port) : 0));

        /* Send off the header (with data if not chunked) */
        LogMsg(hdr);
        sc = ep->v->SendBuffer(ep, hdr, 0, sockaddr, sockaddr_size);
        DBG(("message sent sc=%x...", sc));
        CHECKSC("SendBuffer");

        RELEASE_NULL(hdr);
        RELEASE(data);
        data = next;
        if (size == 0)
            break;

        Position = Uint64AddUint32(Position, size);
        sc = ep->v->AllocateBuffer(ep, 0, &hdr);
        CHECKSC("Alloc IBuffer chunk header");

    } while (TRUE);

METHOD_CLEANUP;
    RELEASE(hdr);
    RELEASE(data);

    return sc;
}
