/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <netif.h>
#include <heaps/dbg_heap.h>
#include "http.h"
#include "httpserver.h"
#include <web/cont.h>

#ifndef _MINIMIZE
/* HTTP request processing - namespace targets */
typedef struct _HttpNSContent {
    const struct IFileVtbl*     v;
    UINT                        Refs;
    PIHTTPCONNECTOR             Request;
    PHTTPHEADERFIELD            HostField;  /* The host field reference points 
                                             * into the request structure, so 
                                             * the request reference count is 
                                             * enough to make sure the host 
                                             * field doesn't go away */
    _TCHAR*                     Previous;
} HttpNSContent, * PHTTPNSCONTENT;

#include <s_HttpNSContent_v.c>
#include <s_HttpProcessNS_v.c>
#include <s_HttpProcessFile_v.c>
#else /*_MINIMIZE*/
#define HttpProcessFile_ClassDesc HttpProcessSoap_ClassDesc
#define HttpProcessNS_ClassDesc HttpProcessSoap_ClassDesc
#define HttpNSContent_ClassDesc HttpProcessSoap_ClassDesc
#endif /* _MINIMIZE */

#include <s_HttpProcessSoap_v.c>
#include <s_HttpFactory_v.c>

PRIVATE SCODE MCT
HttpFactoryDestroy(PIUNKNOWN iThis)
{
    METHOD_DECLARE(HttpFactory, Destroy);
    PIHEAP Heap = CurrentHeap();

    sc = Heap->v->Free(Heap, 0, (PTR) This);

METHOD_CLEANUP;
    return S_OK;
}

/* HTTP server factory implementation */
#ifndef _MINIMIZE
const struct _HttpProcessNS HttpProcessNS =  {&HttpProcessNSVtbl.v, 1};
const struct _HttpProcessFile HttpProcessFile = {&HttpProcessFileVtbl.v, 1};
#endif
const struct _HttpProcessSoap HttpProcessSoap = {&HttpProcessSoapVtbl.v, 1};

PRIVATE SCODE HttpFactoryMain(PIWEBPROTOCOLFACTORY iThis,
                   const _TCHAR *cmdline)
{
    METHOD_DECLARE(HttpFactory, Main);
    PIFILE file = NULL;
    PIHTTPCONNECTOR conn = NULL;

    if (_tcslen(cmdline) == 0)
        cmdline = _T("http");

    /* If there is just an URL this is a server */
    if (!_tcschr(cmdline, _T(' '))) {
        if (!This->Server) {
            PILISTENER Server;
            sc = HttpFactoryCreateServer(iThis, CurrentNameSpace(), 0,
                                         &Server);
            CHECKSC("CreateServer");

            if (!AtomicCmpAndSwap((PUINT) &This->Server, 0, (UINT) Server)) {
                RELEASE_NULL(Server);
            }
        }

        sc = This->Server->v->Listen(This->Server, cmdline);
        CHECKSC("Listen");

        /* Create one thread but no more */
        if (This->ThreadCount < 1) {
            PIPROCESS prc;
            PITHREAD pthd;
            PIMODULE mod;

            AtomicInc(&This->ThreadCount);

            if (This->ThreadCount > 1) {
                AtomicDec(&This->ThreadCount);
            } else {
                /* Create a thread inside the Server module */
                sc = GetModuleContainsAddr((ADDRESS) This->Server->v->Loop,
                                           &mod);
                CHECKSC("GetModule");

                sc = mod->v->QueryInterface(mod, &IID_IProcess, (void **)&prc);
                RELEASE(mod);
                CHECKSC("QI to proc");

                sc = prc->v->CreateThread(prc, CREATE_THREAD_STATIC_SCOPING,
                                       (THREAD_FUNCTION) This->Server->v->Loop,
                                          This->Server, 0, NULL, &pthd);
                RELEASE(prc);
                CHECKSC("CreateThread");
                RELEASE(pthd);
            }
        }
    } else {
        /* http.cob g fs/foo http://127.0.0.1/fs/hostname.txt
         * http.cob p fs/hostname.txt http://127.0.0.1/fs/bar
         */
        const _TCHAR *fname;
        UINT len;
        fname = cmdline + 2;
        len = _tcscspn(fname, _T(" "));
        CHECKSETSC(len && fname[len] == _T(' ') && fname[len+1]
                   &&  fname[len+1] != _T(' '),
                   E_PARSE_ERROR, "usage: get file URL");
        fname = _tcsndup(fname, len);
        CHECKSETSC(fname, E_NOT_ENOUGH_MEMORY, "dup file name");
        sc = BindToObject(CurrentNameSpace(), fname,
                          NAME_SPACE_READ|NAME_SPACE_WRITE|NAME_SPACE_CREATE,
                          &IID_IFile, (void **) &file);
        free((void *) fname);
        CHECKSC("bind to file");

        sc = iThis->v->CreateConnector(iThis, cmdline + len + 3, 0, &conn);
        CHECKSC("CreateConnector");

        sc = conn->v->GetPut(conn,
                             cmdline[0] == _T('p'),
                             file);
        CHECKSC("http client GetPut");
    }

METHOD_CLEANUP;
    RELEASE(file);
    RELEASE(conn);
    return sc;
}

/* HTTP server */

SCODE MCT
HttpServerService(PILISTENER iThis, PIENDPOINT ep,
                  PARSE_URL_FLAGS UrlFlags, PIHTTPCONNECTOR iSession)
{
    METHOD_DECLARE(TcpServer, Service);
    HttpConnector *Session = (HttpConnector *) iSession;

    if (!Session) {
        sc = HTTPSessionCreate(iThis, ep, UrlFlags, &Session);
        CHECKSC("HttpSessionCreate");
        iSession = (PIHTTPCONNECTOR) Session;
    }

    /* Process HTTP requests until the session is over.
     * In case of failure, the loop apologizes to the 
     * client and does all request-oriented cleanup (except see below) */

    do {
        sc = Session->v->ReceiveMessage((PIHTTPCONNECTOR) Session);
    } while (sc == E_RESTART);
    CHECKSC("ReceiveMessage");

METHOD_CLEANUP;
     if (Session) { 
         if (sc == E_WOULD_BLOCK) {
             sc = iThis->v->SetEndpoint(iThis, Session->Endpt,
                                        Session->Flags,
                                        (PIUNKNOWN) Session,
                                        NULL); //Session->timeo);
             /* Keep going regardless of outcome. If not registered, the
              * connection will simply be closed below.
              */
         } else if (!Session->isServer) {
             if ((UrlFlags & PARSE_URL_FLAGS_UDP) == 0
                 && (FAILED(sc)
                     || sc == S_FALSE
                     || !HTTPSessionIsKeepAlive(Session)))
             {
                 /* Close socket upon failure / single use socket */
                 Session->ProtocolState = HPS_DEAD;
             }

             if (Session->cont && FAILED(sc)) {
                 /* The client processing failed. Update return val w/ error */
                 PCONTINFO cont = (PCONTINFO) Session->cont;
                 if (FAILED(sc) && SUCCEEDED(cont->RetVal))
                     cont->RetVal = sc;
             }
             iSession->v->Done(iSession);
         } else if (FAILED(sc) || sc == S_FALSE) {
             iSession->v->Done(iSession);
         }

         RELEASE(iSession);
     }
    RELEASE(ep);

    return sc;
}

#ifndef _MINIMIZE
PRIVATE SCODE HttpProcessNSDestroy(PIUNKNOWN Content)
{
    return S_OK;
}

PRIVATE SCODE MCT
HttpProcessNSGet(PIHTTPPROCESS iThis, PIHTTPCONNECTOR Request,
                 /*out*/ PIFILE *pResponseFile)
{
    PHTTPNSCONTENT  Content     = NULL;
    PIHEAP          Heap        = ((HttpConnector *)Request)->Heap;
    
    TRACE("HTTPProcessNSGet");

    Content = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("httpprocessns"),
                             sizeof(HttpNSContent), 0);
    if (Content == NULL) return E_NOT_ENOUGH_MEMORY;

    Content->v          = &HttpNSContentVtbl.v;
    Content->Refs       = 1;
    Content->Request    = Request;
    Content->Previous   = NULL;

    *pResponseFile = (PIFILE) Content;

    return S_OK;
}

PRIVATE SCODE HttpNSContentDestroy(PIUNKNOWN Content)
{
    PHTTPNSCONTENT  This = (PHTTPNSCONTENT) Content;
    PIHEAP Heap = ((HttpConnector *)This->Request)->Heap;

    TRACE("HTTPNSContentRelease");

    if (This->Previous != NULL) 
        Heap->v->Free(Heap, 0, (PTR) This->Previous);

    Heap->v->Free(Heap, 0, This);

    return S_OK;
}

PRIVATE SCODE MCT
HttpNSContentReadAt(PIFILE Content, UINT64 Position, 
                    BYTE* Buffer, UINT BufferSize, UINT* BufferSizeUsed)
{
    SCODE           sc;
    PHTTPNSCONTENT  This                = (PHTTPNSCONTENT) Content;
    HttpConnector *  Request             = (HttpConnector *) This->Request;
    PINAMESPACE     NameSpace           = (PINAMESPACE) Request->Target;
    BYTE*           Current             = Buffer;
    UINT            TmpSize             = 0x100;
    UINT            TmpSizeUsed         = 0;
    PIHEAP Heap;
    _TCHAR *Tmp = NULL;

    TRACE("HTTPNSContentReadAt");

    *BufferSizeUsed = 0;

    if (BufferSize == 0) return S_OK;

    /* Strings used to generate the page.
     * Each is referenced only once so we inline them. */
#define HTML_PREFIX                     "<html>\r\n<head>\r\n<title>MMLite</title>\r\n</head>\r\n<body>\r\n<p>\r\n"
#define HTML_PREFIX_LENGTH              (sizeof(HTML_PREFIX) - 1)
#define NSPACE_LINE_PREFIX              "    <a href=\"http://"
#define NSPACE_LINE_PREFIX_LENGTH       (sizeof(NSPACE_LINE_PREFIX) - 1)
#define NSPACE_CLOSE_HREF               "\">"
#define NSPACE_CLOSE_HREF_LENGTH        (sizeof(NSPACE_CLOSE_HREF) - 1)
#define NSPACE_LINE_POSTFIX             "</a><br>\r\n"
#define NSPACE_LINE_POSTFIX_LENGTH      (sizeof(NSPACE_LINE_POSTFIX) - 1)
#define HTML_POSTFIX                    "</body>\r\n</html>\r\n"
#define HTML_POSTFIX_LENGTH             (sizeof(HTML_POSTFIX) - 1)

    /* One time initialization of a fresh content structure */
    if (This->Previous == NULL) {
        /* Or we went past the end... */
        if (!Int64IsZero(Position))
            return S_FALSE;

        /* No previous entry - the content structure is fresh */
        Heap = ((HttpConnector *)This->Request)->Heap;

        /* This memory will be freed when the namespace content 
         * structure is eventually released (HTTPNSContentRelease). */
        /* We need two buffers so just make it twice as big... */
        This->Previous = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("httpcontenttmp"),
                                        TmpSize * sizeof(_TCHAR) * 2, 0);
        if (This->Previous == NULL) return E_NOT_ENOUGH_MEMORY;
        This->Previous[0] = _EOS;

        /* Get the host request header field */
        sc = Request->v->GetHeaderField((PIHTTPCONNECTOR) Request, 
                                        TOKEN_HOST, 
                                        &This->HostField);
        if (FAILED(sc) || sc == S_FALSE) {
            Request->StatusCode = TOKEN_BAD_REQUEST;
            return (sc != S_FALSE ? sc : E_ENTRY_NOT_FOUND);
        }

        /* Write the beginning of the generated page */
        *BufferSizeUsed = HTML_PREFIX_LENGTH;
        if (*BufferSizeUsed > BufferSize) return E_BUFFER_TOO_SMALL;
        strncpy((char*) Current, HTML_PREFIX, sizeof(HTML_PREFIX) - 1);
        Current += HTML_PREFIX_LENGTH;

        /* We already made some progress, so we can actually quit here if 
         * the buffer is very small */
    }
    /* Use the second half of the preallocated buffer for temporary storage */
    Tmp = This->Previous + TmpSize;

    /* Add namespace entries to the buffer until we run out of space.
     * Each time the method gets call we must progress through at least 
     * one name. The first call must progress through the page header and 
     * the first namespace entry. The last call must cover both the last 
     * namespace entry and the page footer.
     */

    do {
        UINT URLSize;

        /* Get the next namespace element */
        sc = NameSpace->v->FindNext(NameSpace, 
                                    NULL, 
                                    (_TCHAR*) This->Previous, 
                                    (_TCHAR*) Tmp, 
                                    TmpSize, 
                                    &TmpSizeUsed);

        /* BUGBUG We don't support any namespace entries of more than TmpSize bytes.
         * It may be a good idea to simply put in an error indication, right into 
         * the page, and continue as if nothing had happened. That would help if 
         * people start adding namespace entries that may be long and therefore 
         * could kill the HTTP server */
        if (sc == S_BUFFER_TOO_SMALL) 
            return E_BUFFER_TOO_SMALL;
        if (FAILED(sc))
            return sc;

        if (TmpSizeUsed == 0) {
            /* We're trying to read past the end of the entry list. This is 
             * guaranteed not to happen except when the read is requested after 
             * all content has previously been successfully read. In that case 
             * we should return S_OK but zero bytes. */
            *BufferSizeUsed = 0;
            return S_OK;
        }

        /* Calculate the length of the namespace entry string */
        URLSize = 
            NSPACE_LINE_PREFIX_LENGTH
            + This->HostField->ValueLength
            + Request->RequestURILength
            + TmpSizeUsed
            + NSPACE_CLOSE_HREF_LENGTH
            + TmpSizeUsed
            + NSPACE_LINE_POSTFIX_LENGTH;

        /* Check whether the supplied buffer can contain the entry string */
        if (BufferSize - *BufferSizeUsed < URLSize + HTML_POSTFIX_LENGTH) {
            /* Oops, don't have space to add the element */
            if (*BufferSizeUsed > 0)
                /* Oh goody, we already made some progress */
                return S_OK;
            else
                return E_BUFFER_TOO_SMALL;
        }

        _tcsncpy(This->Previous, Tmp, TmpSizeUsed);
        This->Previous[TmpSizeUsed] = _EOS;

        /* Copy the the namespace entry string into the supplied buffer */
        strncpy((char*) Current, NSPACE_LINE_PREFIX, NSPACE_LINE_PREFIX_LENGTH);
        Current += NSPACE_LINE_PREFIX_LENGTH;
        strncpy((char*) Current, 
                (const char*) This->HostField->Value, 
                This->HostField->ValueLength);
        Current += This->HostField->ValueLength;
        strncpy((char*) Current, 
                (const char*) Request->RequestURI, 
                Request->RequestURILength);
        Current += Request->RequestURILength;
 
        /* ... Short interlude... sometimes we need an extra slash... */
        if (Request->RequestURILength > 1 && *(Current - 1)!= '/') {
            /* Something more than the root asked for */
            ++URLSize; 
            (Current++)[0] = '/'; 
        }

#ifdef _UNICODE
        /* the web page is still produced as ASCII */
        sprintf((char*) Current, "%.*ls", TmpSizeUsed, (const char*) Tmp);
#else
        strncpy((char*) Current, (const char*) Tmp, TmpSizeUsed);
#endif
        Current += TmpSizeUsed;
        strncpy((char*) Current, NSPACE_CLOSE_HREF, NSPACE_CLOSE_HREF_LENGTH);
        Current += sizeof(NSPACE_CLOSE_HREF) - 1;
#ifdef _UNICODE
        sprintf((char*) Current, "%.*ls", TmpSizeUsed, (const char*) Tmp);
#else
        strncpy((char*) Current, (const char*) Tmp, TmpSizeUsed);
#endif
        Current += TmpSizeUsed;
        strncpy((char*) Current, NSPACE_LINE_POSTFIX, NSPACE_LINE_POSTFIX_LENGTH);
        Current += NSPACE_LINE_POSTFIX_LENGTH;

        /* Phew, got that sucker =8) */
        *BufferSizeUsed += URLSize;
    }
    while (sc != S_NO_MORE_ENTRIES);

    /* All the namespace entries have been processed. Put a lid on it... 
     * We already made sure that there is enough space in the buffer. */
    strncpy((char*) Current, HTML_POSTFIX, HTML_POSTFIX_LENGTH);
    Current += HTML_POSTFIX_LENGTH;
    *BufferSizeUsed += HTML_POSTFIX_LENGTH;

    /* Throw away tmp content. It is no longer needed and we
     * use it to recognize we are either at the beginning or at the end.
     */
    Heap = ((HttpConnector *)This->Request)->Heap;
    Heap->v->Free(Heap, 0, (PTR) This->Previous);
    This->Previous = NULL;

    /* All message content read */
    return S_FALSE;
}

PRIVATE SCODE MCT 
HttpNSContentGetSize(PIFILE File, UINT64* Size) 
{
    TRACE("HTTPNSContentGetSize");

    UnusedParameter(File);
    *Size = UINT64_ZERO;

    return S_OK;
}

PRIVATE SCODE HttpProcessFileDestroy(PIUNKNOWN Content)
{
    return S_OK;
}

/* HTTP request processing - file targets */
PRIVATE SCODE MCT
HttpProcessFileGet(PIHTTPPROCESS Process, PIHTTPCONNECTOR RequestIf,
                   /*out*/ PIFILE *pResponseFile)
{
    HttpConnector *    Request     = (HttpConnector *) RequestIf;
    PIFILE Content;

    TRACE("HTTPProcessFileGet");

    Content = (PIFILE) Request->Target;
    Content->v->AddRef(Content);
    *pResponseFile = Content;

    return S_OK;
}

PRIVATE SCODE MCT
HttpProcessFilePut(PIHTTPPROCESS iThis, PIHTTPCONNECTOR RequestIf)
{
    HttpConnector    *Request     = (HttpConnector *) RequestIf;
    PITOKENIZER       Tokenizer   = Request->Tokenizer;
    PBYTE*            Content     = Tokenizer->v->GetContent(Tokenizer);
    PIBUFFER buf = NULL;
    UINT Written;
    METHOD_DECLARE_NOTHIS(HttpProcessFile, Put);

    /* Extract content while leaving the tokenizer ready for next packet */
    sc = Tokenizer->v->CarveBuffer(Tokenizer, Content, UINT_MAX,
                                   &buf);
    CHECKSC("CarveBuffer extract payload");

    sc = buf->v->ReadWriteFile(buf, (PIFILE) Request->Target, TRUE,
                               UINT64_ZERO, 0, UINT_MAX,
                               &Written);

    CHECKSC("write payload to file");

METHOD_CLEANUP;
    RELEASE(buf); 
    return sc;
}

#endif /* _MINIMIZE */

#include <web/cont.h>

PRIVATE SCODE HttpProcessSoapDestroy(PIUNKNOWN Content)
{
    return S_OK;
}

PRIVATE SCODE HttpProcessSoapPost(PIHTTPPROCESS iThis, 
                                  PARSE_URL_FLAGS Flags,
                                  PITOKENIZER Tokenizer,
                                  const _TCHAR *LocalUrl,
                                  PIBUFFER *pContent,
                                  PIUNKNOWN *pDeserializer,
                                  UINT *ParsingState,
                                  PIHTTPCONNECTOR Request)
{
    METHOD_DECLARE_NOTHIS(HttpProcessSoap, Post);
    PCONTINFO cont = NULL;
    PISOAPSERIALIZER Serializer = NULL;
    PICONVERT convert = NULL;
    PIHEAP Heap = _GetHeap(_T("Handler"), 4000);
    
    TRACE("SOAPHTTPProcessPost");
    assert(Tokenizer != NULL);

    sc = BindToObject(CurrentNameSpace(), _T("COB/convert.cob"),
                      NAME_SPACE_READ, &IID_IConvert,
                      (void**) &convert);
    CHECKSC("bind to soap.cob");

    sc = convert->v->ProcessRequest(convert, Flags, Tokenizer,
                                    LocalUrl,
                                    (PISAXHANDLER *) pDeserializer,
                                    (PICONTINUATION*) &cont);
    CHECKSC("ProcessRequest");

    CHECKSC2("Request message parsing");
    *ParsingState = HS_RUNNING;

    CHECKSETSC(cont, E_UNEXPECTED_ERROR, "exists(cont)");
    assert(cont->Method && cont->Frame
           && !cont->Serializer && !cont->Request);

    assert(!cont->Request);
    cont->Request = Request;
    if (Request)
        Request->v->AddRef(Request);

    CHECK_HEAP();
    sc = cont->v->Execute(iCONT(cont), FALSE, NULL);
    assert(sc == S_NO_RESPONSE || FAILED(sc));
    CHECK_HEAP();
    if (Request && FAILED(sc)) {
        /* Drop ref to request we took above */
        RELEASE(Request);
        cont->Request = NULL;
    }
    CHECKSC("cont->Execute");

    if (cont && cont->metd
        && cont->metd->ResponseNameToken == TOKEN_invalid)
        sc = S_NO_RESPONSE;
        
METHOD_CLEANUP;
    assert(sc != S_FALSE);
    /* When UDP, don't send fault message unless we know at least the method
     * and that that method wanted a reply.
     */
    if (FAILED(sc) && (Flags & PARSE_URL_FLAGS_UDP)
        && (!cont || !cont->metd || (cont->metd->ResponseNameToken
                                     == TOKEN_invalid)))
    {
        sc = S_NO_RESPONSE;
    }

    /* Only generate Fault message when continuation present.
     * In certain failure cases there will only be an HTTP error or
     * connection closure / drop.
     */
    if (FAILED(sc) && sc != E_WOULD_BLOCK && cont && convert) {
        SCODE sc2 = sc;
        CHECK_HEAP();
        if (!Serializer)
            Serializer = cont->Serializer;

        if (!Serializer) {
            sc = convert->v->CreateSerializer(convert, (PIFILE) *pContent,
                                              Flags, &Serializer);
            CHECKSC2("CreateSerializer");
        }

        Serializer->v->Reset(Serializer, NULL);
        CHECKSC2("truncate response file");

        sc = Serializer->v->FaultMessage(Serializer, iCONT(cont), sc2,
                                         NULL, NULL);
        CHECKSC2("generate FaultMessage");
    }

METHOD_CLEANUP2;
    CHECK_HEAP();

    if (Request && sc != S_NO_RESPONSE && sc != E_WOULD_BLOCK && Serializer) {
        PIFILE File = NULL;
        SCODE sc2 = sc;
        sc = Serializer->v->GetContent(Serializer, &File);
        if (SUCCEEDED(sc)) {
            sc = Request->v->SendMessage(Request,
                                         (SUCCEEDED(sc2)
                                          ? TOKEN_OK : TOKEN_SOAP_PARSING_ERROR),
                                         NULL, /* SoapAction */
                                         File,
                                         NULL);
            File->v->Release(File);
            sc = S_NO_RESPONSE;
        }
    }

    //if (sc != E_WOULD_BLOCK) // XXX restart case!
        RELEASE_NULL(*pDeserializer); /* releases cont */
    RELEASE(iCONT(cont));
    RELEASE(Serializer);
    RELEASE(convert);
    return sc;
}
