/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <mmlite.h>
#include <netif.h>
#include <tcpserver.h>
#include <httpif.h>
#include <mmmacros.h>
#include <net/ip/in.h>          /* struct ip_mreq */
#include <net/if.h>             /* for struct ifreq */

#include "http.h"
#include <s_tcpserver_v.c>

#define BLOCKING 0

/* IUnknown implementation */

PRIVATE SCODE MCT 
TcpServerQueryInterface(PILISTENER This, REFIID Iid, void **Unk)
{
    //DBG(("tcpserver.c::TCPServerQueryInterface\n"));
    
    return GenericQueryInterface((IUnknown *) This, Iid, Unk, &IID_IListener);
}


PRIVATE UINT MCT 
TcpServerAddRef(PILISTENER This)
{
    UINT RefCnt;
    TcpServer *TCPServer = (TcpServer *) This;

    //DBG(("tcpserver.c::TCPServerAddRef\n"));
    
    RefCnt = AtomicInc(&TCPServer->Refs);
    return (RefCnt);
}

PRIVATE SCODE MCT 
TcpServerDestroy(PIUNKNOWN iThis)
{
    TcpServer *This = (TcpServer *) iThis;


    RELEASE(This->NameSpace);

    Mutex_Destroy(&This->mx);
    Condition_Destroy(&This->cd);

//    if (This->ServerEndpoint)
//        Server->ServerEndpoint->v->Release(Server->ServerEndpoint);

    /* Deallocate the memory for the TCP server object */
    {
        IHeap *Heap = CurrentHeap();
        (void) Heap->v->Free(Heap, 0, This);
    }

    return S_OK;
}

PRIVATE UINT MCT 
TcpServerRelease(PILISTENER iThis)
{
    UINT RefCnt;
    TcpServer *This = (TcpServer *) iThis;

    //DBG(("tcpserver.c::TCPServerRelease\n"));

    RefCnt = AtomicDec(&This->Refs);
    if (RefCnt == 0)
        TcpServerDestroy((PIUNKNOWN) This);

    return RefCnt;
}


PRIVATE UINT32 GetInterfaceIp(const _TCHAR *iname, UINT inamelen)
{
    PINAMESPACE ns = CurrentNameSpace();
    PIENDPOINTFACTORY ef;
    PIENDPOINT ep;
    struct ifreq ifr;
    UINT32 LocalIPAddr = INADDR_ANY;
    SCODE sc;
    
    /* Prepare address and port to listen at */
    memset((void*) &ifr, 0, sizeof(struct ifreq));

    /* Get an endpoint factory to create an endpoint */
    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void**) &ef);
    if (FAILED(sc)) goto Out;

    /* Create an endpoint to query for the interface address */
    sc = ef->v->CreateEndpoint(ef, AF_INET, SOCK_DGRAM, 0, &ep);
    ef->v->Release(ef);
    if (FAILED(sc)) goto Out;

    /* This is where we specify which interface to use */
#ifdef _UNICODE
{
    char *s = _wcstos(iname);
    if (!s)
        return E_NOT_ENOUGH_MEMORY;
    strncpy(ifr.ifr_name, s, __min(inamelen, sizeof(ifr.ifr_name)));
    free(s);
}
#else
    strncpy(ifr.ifr_name, iname, __min(inamelen, sizeof(ifr.ifr_name)));
#endif
    
    /* Query for the interface address */
    sc = ep->v->Ioctl(ep, SIOCGIFADDR, (BYTE*) &ifr);
    ep->v->Release(ep);
    if (FAILED(sc)) goto Out;

    LocalIPAddr = ((struct sockaddr_in *) &(ifr.ifr_addr))->sin_addr.s_addr;
Out:
    DBG(("GetInterfaceIp got %x %s\n", LocalIPAddr, inet_ntoa(*(struct in_addr *)&LocalIPAddr)));
    return LocalIPAddr;
}

PRIVATE SCODE AddEp(TcpServer *This, PIENDPOINT ep, BOOL Accept, PARSE_URL_FLAGS Flags)
{
    UINT j;
    /* Try to stuff into unused slot */
    if (This->inuse < This->Count) {
        for (j = 0; j < This->Count; j++)
            if (!This->eps[j]) {
                This->eps[j] = ep;
                This->flags[j] = Flags;
                This->reqs[j] = ENDPOINT_READ;
                This->timeouts[j] = TIME_FOREVER;
                assert(!This->keys[j]);
                This->inuse++;
                DBG(("tcpserver: new socket %x in slot %d\n", ep, j));
                break;
            }
        assert(j < This->Count);
    } else if (This->Count >= MAXENDP) {
        return E_OUT_OF_STRUCTURES; /* enforce limit */
    } else {
        /* extend table */
        This->eps[This->Count] = ep;
        This->flags[This->Count] = Flags;
        This->timeouts[This->Count] = TIME_FOREVER;
        if (Accept)
            This->reqs[This->Count] = ENDPOINT_USER1 | ENDPOINT_READ;
        else
            This->reqs[This->Count] = ENDPOINT_READ;

        assert(!This->keys[This->Count]);
        DBG(("tcpserver: new socket %x in end slot %d\n", ep, This->Count));
        This->Count++;
        This->inuse++;
    }
    assert(This->Count <= MAXENDP);
    assert(This->inuse <= MAXENDP);
    
    /* too many? */
    if (This->inuse >= MAXENDP) {
        /* temporarily disable new sockets -- limit reached */
        for (j = 0; j < This->Count; j++)
            if (This->reqs[j] & ENDPOINT_USER1)
                This->reqs[j] &= ~(ENDPOINT_READ
                                   | ENDPOINT_ISREADABLE);
    }

#if !BLOCKING
    /* Request wakeups */
    return ep->v->SetCondition(ep, &This->cd);
#else
    return S_OK;
#endif
}

PRIVATE SCODE RemoveEp(TcpServer *This, PIENDPOINT ep, BOOL dorelease)
{
    UINT i;
    for (i = 0; i < This->Count; i++) {
        if (This->eps[i] == ep) {
            if (dorelease) {
                RELEASE(This->keys[i]);
                RELEASE(This->eps[i]);
            }
            This->keys[i] = NULL;
            This->eps[i] = NULL;
            This->reqs[i] = 0;

            if (This->Count-1 == i)
                This->Count--;
            This->inuse--;

            DBG(("tcpserver.c: removed socket %x in slot %d count=%d\n",
                 ep, i, This->Count));

            /* Restart disabled accept sockets */
            for (i = 0; i < This->Count; i++)
                if (This->reqs[i] & ENDPOINT_USER1)
                    This->reqs[i] |= ENDPOINT_READ;
            assert(This->Count <= MAXENDP && This->Count >= 0);
            assert(This->inuse <= This->Count && This->inuse >= 0);
            
            return S_OK;
        }
    }
    return E_ENTRY_NOT_FOUND;
}

/* IListener implementation */

PRIVATE SCODE MCT
TcpServerListen(PILISTENER iThis, const _TCHAR *AddressURL)
{
    METHOD_DECLARE(TcpServer, Listen);
    PIENDPOINT ep = NULL, ep2 = NULL;
    PINAMESPACE ns = CurrentNameSpace();
    PIENDPOINTFACTORY ef = NULL;
    struct sockaddr_in Addr;
    PISOAPFACTORY   SoapFactory;
    PARSE_URL_FLAGS Flags;
    const _TCHAR *host;
    UINT hostlen;
    UINT32 LocalIPAddr;
    UINT16 LocalPort;
        
    sc = BindToObject(ns, _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &SoapFactory);
    CHECKSC("bind to soap.cob");

    sc = SoapFactory->v->ParseURL(SoapFactory, AddressURL,
                                  &Flags, &host, &hostlen, &LocalPort,
                                  NULL,NULL,NULL,NULL,NULL);
    RELEASE(SoapFactory);
    CHECKSC("ParseURL");
    
    /* Get an endpoint factory interface */
    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void**) &ef);
    CHECKSC("bind to protocol.cob");

    /* Construct the server endpoint */
    sc = ef->v->CreateEndpoint(ef, AF_INET,
           (Flags & PARSE_URL_FLAGS_UDP) ? SOCK_DGRAM : SOCK_STREAM,
           0, &ep);
    CHECKSC("CreateEndpoint");

    /* Decide on local address and port to listen at */
    if (hostlen == 0) {
        LocalIPAddr = INADDR_ANY;
    } else if ((LocalIPAddr = _tinet_addr(host)) == INADDR_NONE) {
        /* If it isn't an IP address then it might be an interface name */
        LocalIPAddr = GetInterfaceIp(host, hostlen);
    }

    /* Bind the endpoint */

    /* Build the server endpoint address structure */
    /* Precondition: LocalIPAddr and LocalPort are *already* in network byte order */
    memset((char*) &Addr, 0, sizeof(Addr));
    Addr.sin_family         = AF_INET;
    Addr.sin_port           = htons(LocalPort);
    Addr.sin_addr.s_addr    = (UINT32) LocalIPAddr;
        
    /* Bind the endpoint to the address */
    sc = ep->v->Bind(ep,(struct sockaddr*)&Addr,sizeof(Addr));
    CHECKSC("ep.Bind");

    DBG(("tcpserver.c::TCPServerListen: Address %x, port %x, network byte order\n", 
         LocalIPAddr, LocalPort));

    DBG(("tcpserver.c: Listen at %s:%d %s\n",
         inet_ntoa(*(struct in_addr *)&LocalIPAddr), LocalPort,
         (Flags & PARSE_URL_FLAGS_UDP) ? "SOCK_DGRAM" : "SOCK_STREAM"));

    if (Flags & PARSE_URL_FLAGS_UDP) {
        if (LocalIPAddr == INADDR_ANY && hostlen != 0) {
            _TCHAR *tmp = _tcsndup(host, hostlen);
            if (tmp) {
                LocalIPAddr = _tinet_addr(tmp);
                free (tmp);
            }
        }

        /* Join the multicast group if it is multicast IP. */
        if (IN_MULTICAST(ntohl(LocalIPAddr))) {
            struct ip_mreq mreq;
#if 1
            struct sockaddr_in sin;
            UINT slen;

            /* Find the right interface for the multicast route */
            sin.sin_family = AF_INET;
            sin.sin_len = sizeof(struct sockaddr_in);
            sin.sin_port = htons(LocalPort);
            sin.sin_addr.s_addr = LocalIPAddr;

            sc = ef->v->CreateEndpoint(ef, AF_INET, SOCK_DGRAM, 0, &ep2);
            CHECKSC("CreateEndpoint2");
    
            sc = ep2->v->Connect(ep2, (struct sockaddr *) &sin, sizeof(sin));
            CHECKSC("Connect to multicast addr for iface discovery purposes");

            slen = sizeof(struct sockaddr_in);
            sc = ep2->v->GetName(ep2, (struct sockaddr *) &sin,
                                 slen, &slen, ENDPOINT_ADDR_MINE);
            CHECKSC("GetName for iface discovery purposes");

            DBG(("tcpserver.c: multicast interface appears to be %s\n",
                inet_ntoa(sin.sin_addr)));

            mreq.imr_interface.s_addr = sin.sin_addr.s_addr;
#else
            mreq.imr_interface.s_addr = INADDR_ANY;
#endif
            mreq.imr_multiaddr.s_addr = LocalIPAddr;

            sc = ep->v->SetSockOpt (ep,
                                    IPPROTO_IP, 
                                    IP_ADD_MEMBERSHIP, 
                                    (const BYTE *)&mreq, 
                                    sizeof (mreq));
            CHECKSC("join multicast group");
        }
        Mutex_Lock(&This->mx);
        sc = AddEp(This, ep, FALSE, Flags);
        Mutex_Unlock(&This->mx);
        CHECKSC("AddEp");
    } else {
        /* Listen on the endpoint */
        sc = ep->v->Listen(ep, 5);
        CHECKSC("ep.listen");

        Mutex_Lock(&This->mx);
        sc = AddEp(This, ep, TRUE, Flags); /* Need to "Accept" this socket */
        Mutex_Unlock(&This->mx);
        CHECKSC("AddEp/accept");
    }

    /* In case the server was already running, make it refresh now */
    Condition_Broadcast(&This->cd);

METHOD_CLEANUP;
    if (FAILED(sc) && ep) {
        ep->v->Close(ep);
        ep->v->Release(ep);
    }
    RELEASE(ef);
    RELEASE(ep2);

    return sc;
}

PRIVATE SCODE MCT TcpServerLoop(PILISTENER iThis)
{
    METHOD_DECLARE(TcpServer, Loop);
    UINT i;
    PIENDPOINTFACTORY epf;
    PARSE_URL_FLAGS Flags;
    PIUNKNOWN Key;
    PIENDPOINT ep = NULL;

    Mutex_Lock(&This->mx);

    /* Listen not called or failed */
    CHECKSETSC(This->Count != 0, E_OBJECT_NOT_INITIALIZED,
               "template socket missing");

    /* Get an endpoint factory interface */
    sc = BindToObject(CurrentNameSpace(),
                      _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void**) &epf);
    CHECKSC("get socket factory");

    /* Loop, accepting incoming service requests */
    for (;;) {
        /* Prepare the client endpoint address structure */
        struct sockaddr_in Addr;
        UINT32 AddrLen = sizeof(Addr);

        //DBG(("tcpserver.c: TCPServerRun: Waiting for service request\n"));
#if !BLOCKING
        /* Probe network select, then block on condition until something
         * happens. We registered the condition with the various sockets.
         */
        sc = epf->v->WaitForEndpoints(epf,
                                      This->eps, This->reqs, This->Count,
                                      TIME_ORIGIN);
        if (sc == E_TIMED_OUT) {
            sc = epf->v->WaitCondition(epf, &This->cd, &This->mx,
                                       &This->NextTimeout);
        }
#else
        Mutex_Unlock(&This->mx); /* XXX Wa... uses values: RACE */
        sc = epf->v->WaitForEndpoints(epf, This->eps, This->reqs, This->Count,
                                      This->NextTimeout);
        Mutex_Lock(&This->mx);
#endif
        if (sc == E_TIMED_OUT) {
            TIME next = This->NextTimeout, newnext = TIME_FOREVER, t;

            /* NextTimeout expired. Thus some socket needs to be closed. */
            for (i = 0; i < This->Count; i++) {
                if (!This->eps[i])
                    continue;
                t = This->timeouts[i];
                if (TimeLess(t, next)) {
                    /* Just close the socket, it will now show up
                     * in readable set and get cleaned up normally.
                     */
                    This->eps[i]->v->Close(This->eps[i]);
                } else if (TimeLess(t, newnext)) {
                    newnext = t;
                }
                This->NextTimeout = newnext;
            }
            sc = S_OK;
        }
        CHECKSC("Wait");

        for (i = 0; i < This->Count; i++) {
            if (!(This->reqs[i] & ENDPOINT_ISREADABLE))
                continue;
            assert(This->eps[i]);

            if (This->reqs[i] & ENDPOINT_USER1) {
                /* Accept an incoming connection */
                sc = This->eps[i]->v->Accept(This->eps[i], &ep,
                           (struct sockaddr *) &Addr, AddrLen, &AddrLen);
                if (sc == STATUS_WIN32_ERROR(WSAEWOULDBLOCK))
                    continue;
                CHECKSC("Accept");

                /* Apply an access policy to the incoming connection */
                sc = (*This->v->ApplyAccessPolicy)(iThis, &Addr);
                if (sc == S_FALSE || FAILED(sc)) {
                    ep->v->Release(ep);
                    continue;
                }

                DBG(("tcpserver: ACCEPT from %d\n", i));
                sc = AddEp(This, ep, FALSE, This->flags[i]);
                if (FAILED(sc))
                    ep->v->Release(ep); /* no can do, just drop it */
                CHECKSC("AddEp");

                break; /* accept more before doing others... */
            }

            DBG(("tcpserver.c: Servicing request slot %d count=%d\n",
                 i, This->Count));

            ep = This->eps[i];
            Flags = This->flags[i];
            Key = This->keys[i];

            /* Remove from set, will be re-entered later as needed */
            RemoveEp(This, ep, FALSE);

            /* Service the accepted request */
            (void) This->v->Service(iThis, ep, Flags, (PIHTTPCONNECTOR) Key);
            /* Errors ignored here as they are handled below */
        }
    }

    METHOD_CLEANUP;
    DBG(("tcpserver.c: server shutting down count=%d\n", This->Count));
    for (i = 0; i < This->Count; i++) {
        if (This->eps[i]) {
            This->eps[i]->v->Close(This->eps[i]);
            This->eps[i]->v->Release(This->eps[i]);
            This->eps[i] = NULL;
        }
        RELEASE_NULL(This->keys[i]);
    }
    Mutex_Unlock(&This->mx);

    return sc;
}


PRIVATE SCODE MCT
TcpServerApplyAccessPolicy(PILISTENER iThis, PTR ClientAddr)
{
    DBG(("tcpserver.c::TCPServerApplyAccessPolicy\n"));
    
    return S_OK;
}


PRIVATE SCODE MCT
TcpServerService(PILISTENER iThis, PIENDPOINT ep,
                 PARSE_URL_FLAGS Flags, PIHTTPCONNECTOR Conn)
{
    METHOD_DECLARE(TcpServer, Service);

#if !BLOCKING
    Mutex_Unlock(&This->mx);
#endif

    sc = HttpServerService(iThis, ep, Flags, Conn); 

#if !BLOCKING
    Mutex_Lock(&This->mx);
#endif

    return sc;
}

PRIVATE SCODE MCT TcpServerSetEndpoint(PILISTENER iThis,
                               PIENDPOINT ep, UINT Flags, PIUNKNOWN Key,
                               PTIME pTimeout)
{
#if BLOCKING
    return S_OK;
#else
    UINT j;
    METHOD_DECLARE(TcpServer, SetEndpoint);
    TIME timeo = TIME_FOREVER;

    sc = S_OK;

    /* Normalize timeout */
    if (pTimeout) {
        TIME now = CurrentTime();
        if (TimeIsRelative(timeo))
            timeo = Int64Subtract(now, timeo);
        /* Do not accept timeouts in the past. The caller cleans up. */
        if (TimeLess(timeo, now))
            return E_TIMED_OUT;
    }

    /* When called from HTTPRequestDone the socket is already there */
    Mutex_Lock(&This->mx);
    if (Key == NULL) {
        sc = RemoveEp(This, ep, TRUE);
        CHECKSC("RemoveEp");
        goto Out;
    }

    /* Set next "interrupt" */
    if (TimeLess(timeo, This->NextTimeout))
        This->NextTimeout = timeo;

    for (j = 0; j < This->Count; j++) {
        if (This->eps[j] == ep) {
            ENDPOINT_REQ_TYPES r = This->reqs[j];
            This->reqs[j] |= ENDPOINT_READ;
            assert(This->keys[j] == Key || This->keys[j] == NULL);
            This->keys[j] = Key;
            //if (pTimeout)
                This->timeouts[j] = timeo;

            if ((r & ENDPOINT_READ) == 0)
                DBG(("tcpserver: re-enabled socket %x in slot %d count=%d\n",
                     ep, j, This->Count));
            goto Out; /* found */
        }
    }
    /* New then. Use AddEp but fix up the entry afterwards. */
    sc = AddEp(This, ep, FALSE, Flags);
    CHECKSC("AddEp");

    /* Account for the additional reference */
    ep->v->AddRef(ep);

    for (j = 0; j < This->Count; j++)
        if (This->eps[j] == ep) {
            This->keys[j] = Key;
            if (Key)
                Key->v->AddRef(Key);
            This->timeouts[j] = timeo;
            goto Out;
        }

METHOD_CLEANUP; /* just for print */
    Mutex_Unlock(&This->mx);
    /* Since we added a new socket, the waiter thread needs to refresh
     * its wait state.
     */
    CHECK_HEAP();
    Condition_Broadcast(&This->cd);
    return sc;
#endif
}

SCODE MCT HttpFactoryCreateServer (
    PIWEBPROTOCOLFACTORY Factory, 
    PINAMESPACE NameSpace,
    UINT Flags,
    /*out*/ PILISTENER* Server)
{
    TcpServer *This = NULL;
    PIHEAP Heap = CurrentHeap();

    TRACE("HTTPFactoryCreateServer");

    /* Allocate memory for the server instance */
    This = (TcpServer *) Heap->v->Alloc(Heap,
                                HEAP_ZERO_MEMORY|HEAP_DEBUG_NAME("httpserver"),
                                sizeof(TcpServer), 0);
    if (! This) return E_NOT_ENOUGH_MEMORY;

    This->v                 = &TcpServerVtbl.v;
    This->Refs              = 1;
    This->NameSpace         = NameSpace;
    This->NextTimeout = TIME_FOREVER;

    Factory->v->AddRef(Factory);
    NameSpace->v->AddRef(NameSpace);

    *Server = (PILISTENER) This;

    return S_OK;
}
