/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implements ReservationBuilder concrete class */

#include <stdio.h>
#include "builder.h"
#include "planner.h"
#include "../sensor/instigator.h" /* XXX tmp for PRODUCER_RESERVATION_NAME */

extern void TimeFormat(TIME t, _TCHAR *buf, UINT Len);
extern UINT32 WsResRandUint32(void);
extern _TCHAR *CreateUniqueTaskName(void);
extern _TCHAR *CreateFullReservationName(
    const _TCHAR * taskName, 
    const _TCHAR *resName, 
    const _TCHAR *Suffix);

/* Reservation builder implementation */
    
PRIVATE SCODE ReservationBuilderQueryInterface (
    PIRESERVATIONBUILDER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                                 &IID_IReservationBuilder);
}
    
PRIVATE UINT ReservationBuilderAddRef (
    PIRESERVATIONBUILDER pThis)
{
    return AtomicInc(&pReservationBuilder(pThis)->RefCnt);
}
    
#define trloff offsetof(struct trigger, ListNode)

extern INT TriggerDelete(PTR pItem, PTR pArg);

PRIVATE UINT ReservationBuilderRelease (
    PIRESERVATIONBUILDER pThis)
{
    PRESERVATIONBUILDER This = pReservationBuilder(pThis);
    UINT Refs = AtomicDec(&This->RefCnt);
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        DBGME(1, printf("builder.c:ReservationBuilderRelease: this=0x%x refs=%d host=%s name=%s samples=%s\n", 
                        (ADDRESS) This, Refs,
                        This->Host, This->FullName, This->DataCollection));
        /* Finalize object */
        if (This->Host) {
            free(This->Host);
            This->Host = NULL;
        }
        if (This->FullName) {
            free(This->FullName);
            This->FullName = NULL;
        }
        if (This->Reservation) {
            if (This->Reservation->TriggerList)
                /* Only one trigger is allowed, so we don't have to 
                   walk the list to free */
                pHeap->v->Free(pHeap, 0, This->Reservation->TriggerList);
            pHeap->v->Free(pHeap, 0, This->Reservation);
            This->Reservation = NULL;
        }
        if (This->DataCollection) {
            free(This->DataCollection);
            This->DataCollection = NULL;
        }
        This->Task = NULL;
		if (This->DispatchCallback) {
			This->DispatchCallback = NULL;
			RELEASE_NULL(This->DispatchModule);
		}
        This->DispatchData = NULL;
		if (This->AdaptationCallback) {
			This->AdaptationCallback = NULL;
			RELEASE_NULL(This->DispatchModule);
		}
        This->AdaptationData = NULL;
        pHeap->v->Free(pHeap, 0, (void *) This);
    }
    return Refs;
}
    
PRIVATE INT ReservationBuilderDelete(PTR pItem, PTR pArg)
{
    PIRESERVATIONBUILDER ResB = (PIRESERVATIONBUILDER) pItem;
    PIHEAP pHeap = (PIHEAP) pArg;
    UnusedParameter(pHeap);
    return ResB->v->Release(ResB);
}
    
PRIVATE SCODE ReservationBuilderSetParameters (
    PIRESERVATIONBUILDER iThis,
    TIME Deadline, INT Duration, INT Tolerance)
{
    METHOD_DECLARE(ReservationBuilder, SetParameters);

    /* XXX Add sanity checks here */

    This->Reservation->deadline = Deadline;

    assert(Duration < 0);
    Int32ToInt64(This->Reservation->duration, Duration);
    Int32ToInt64(This->Reservation->tolerance, Tolerance);
    
    sc = S_OK;

METHOD_CLEANUP;
    return sc;
}
    
/* Changes a reservation from single-shot to periodical.

   The Period parameter is a duration, hence relative time, 
   hence a negative 64 bit number. 
   
   The Period also must be large enough to allow the planner 
   to renew the reservation */
PRIVATE SCODE ReservationBuilderSetRecurrence (
    PIRESERVATIONBUILDER iThis,
    TIME Period,
    UINT Repeats)
{
    METHOD_DECLARE(ReservationBuilder, SetRecurrence);
    PIHEAP heap = CurrentHeap();
    struct trigger *trigger = This->Reservation->TriggerList;
    INT64 RenewalSlack;
    

    sc = S_OK;

    CHECKSETSC(Int64Less(Period, INT64_ZERO), E_INVALID_PARAMETER,
               "set period: must be a relative time");

    Int32ToInt64(RenewalSlack, PLANNER_RENEWAL_SLACK);
    CHECKSETSC(Int64Less(Period, RenewalSlack), E_INVALID_PARAMETER,
               "set period: too small for renewal");

    CHECKSETSC(Repeats <= DEFAULT_TRIGGER_MAX_COUNT, E_INVALID_PARAMETER,
               "set repeats: maximum exceeded");

    /* Initialize a new trigger if first time invocation */
    if (!trigger) {
        trigger = (struct trigger *) heap->v->Alloc(heap, 
            HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("builder.c:SetRecurrence"), 
            sizeof(struct trigger), 0);
        CHECKSETSC(trigger, E_NOT_ENOUGH_MEMORY, "create trigger");

        /* Sequence number 1 means first occurrence of periodic reservation */
        This->Reservation->seqno = 1;
    }

    trigger->maxCount = Repeats;
    trigger->offset = Period;

    ListNodeInit(trloff, trigger);

    /* Trigger deallocation (see wsres.c:TriggerDelete) does  
       not free the 'what' field of the trigger structure */
    trigger->what = This->Reservation->name; 
    CHECKSETSC(trigger->what, E_NOT_ENOUGH_MEMORY, "create trigger subject");

    This->Reservation->TriggerList = trigger;

METHOD_CLEANUP;
    if (FAILED(sc) && trigger)
        heap->v->Free(heap, 0, trigger);
    return sc;
}
    
PRIVATE SCODE ReservationBuilderSetAdaptivity(
    PIRESERVATIONBUILDER iThis, 
    const _TCHAR *DataCollection,
    UINT Confidence, 
    PTR Callback, PTR Data)
{
    PRESERVATIONBUILDER This = pReservationBuilder(iThis);

    This->DataCollection = _tcsdup(DataCollection);
    This->Confidence = Confidence;

    if (Callback) {

		/* The adaptation callback can only be set once. 
		If we allow resetting it then we run into trouble 
		figuring out when to release the module that contains
		the callback code */
		assert(!This->AdaptationCallback);

		/* Ensure we hang onto the module that contains the adaptation 
		code - grep for RELEASE_NULL to find the release code */
        GetModuleContainsAddr((ADDRESS) Callback, &This->DispatchModule);

		/* See also SetDispatch for module reference count code */

        This->AdaptationCallback = 
            (SCODE (*)(PTR, PIPLANNER, PRESERVATIONBUILDER, PTR, PISTATISTICS, PBOOL)) 
            Callback;

        This->AdaptationData = Data;

    } else {

        /* Providing adaptation data without a callback to 
           process it makes no sense */
        assert(!Data);
    }

    return S_OK;
}
    
PRIVATE SCODE ReservationBuilderSetDispatch(
    PIRESERVATIONBUILDER iThis, 
    PTR Callback, PTR Data)
{
    PRESERVATIONBUILDER This = pReservationBuilder(iThis);

	/* The dispatch callback can only be set once. 
	   If we allow resetting it then we run into trouble 
	   figuring out when to release the module that contains
	   the callback code */
	assert(!This->DispatchModule);

	/* Ensure we hang onto the module that contains the dispatch 
	   code - grep for RELEASE_NULL to find the release code */
    GetModuleContainsAddr((ADDRESS) Callback, &This->DispatchModule);

	/* See also SetDispatch for module reference count code */

    This->DispatchCallback = (SCODE (*)(PTR, PTR)) Callback;
    This->DispatchData = Data;

    return S_OK;
}
    
PRIVATE SCODE ReservationBuilderCreateContext (
    PIRESERVATIONBUILDER iThis,
    const _TCHAR *FeedbackURL,
    /* out */ PWSCONTEXT *context)
{
    METHOD_DECLARE(ReservationBuilder, CreateContext);
    PIHEAP heap = CurrentHeap();
    PEndpointReference endpoint;

    sc = S_OK;
    *context = (PWSCONTEXT) heap->v->Alloc(heap, 
        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("planner.c:CreateContext"), 
        sizeof(WSCONTEXT), 0);
    CHECKSETSC(*context, E_NOT_ENOUGH_MEMORY, "create context object");

    endpoint = (PEndpointReference) heap->v->Alloc(heap, 
        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("builder.c:CreateContext/ep"), 
        sizeof(EndpointReference) + sizeof(_TCHAR) * (_tcslen(This->Host)+1), 0);
    CHECKSETSC(endpoint, E_NOT_ENOUGH_MEMORY, "create endpoint");

    (*context)->To = endpoint;
    (*context)->Reservation = This->Reservation;
    (*context)->taskName = _tcsdup(This->FullName);
    (*context)->TaskName = _tcsdup(pTaskBuilder(This->Task)->TaskName);
    (*context)->FeedbackURL = _tcsdup(FeedbackURL);

    /* endpoint is single allocation, see ws.c */
    endpoint->Address = (_TCHAR *) (endpoint + 1);
    _tcscpy(endpoint->Address, This->Host);

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE ReservationBuilderResetContext(
    PIRESERVATIONBUILDER iThis,
    PWSCONTEXT context)
{
    METHOD_DECLARE(ReservationBuilder, ResetContext);

    context->Reservation = This->Reservation;

    sc = S_OK;

METHOD_CLEANUP;
    return sc;
}
    
/* Returns S_OK if the reservation has an effective trigger, 
   or E_FAIL if the reservation has already been fully dispatched, 
   e.g. if there are no further repeats that need to be established 
   with the remote scheduler the reservation is intended for.

   Treats the TriggerList as a list of 0 or 1 triggers, no more. 
   We have no use for multiple triggers and I don't know of any 
   motivation either */
PRIVATE SCODE ReservationBuilderIsRepeating(
    PIRESERVATIONBUILDER iThis)
{
    PRESERVATIONBUILDER This = pReservationBuilder(iThis);
    struct reservation *Res = This->Reservation;
    struct trigger *Trigger = Res->TriggerList;
    return (Trigger 
            && ! Int64Equal(Trigger->offset, INT64_ZERO) 
            && 0 < Trigger->maxCount)
        ? S_OK 
        : E_FAIL;
}

PRIVATE SCODE ReservationBuilderIsAdaptive(
    PIRESERVATIONBUILDER iThis)
{
    PRESERVATIONBUILDER This = pReservationBuilder(iThis);
    return 0 != This->Confidence ? S_OK : E_FAIL;
}


PRIVATE SCODE ReservationBuilderDispatch(
    PIRESERVATIONBUILDER iThis)
{
    PRESERVATIONBUILDER This = pReservationBuilder(iThis);
    return (*This->DispatchCallback)(This->DispatchData, This);
}
    
PRIVATE SCODE ReservationBuilderAdapt(
    PIRESERVATIONBUILDER iThis, 
    PTR DataPtr, 
    PISTATISTICS Statistics, 
    PBOOL NothingAdapts)
{
    PRESERVATIONBUILDER This = pReservationBuilder(iThis);
    PMAINTAINERDATA Data = (PMAINTAINERDATA) DataPtr;

    /* Decline to self-adapt if the instigator did not provide 
       an adaptation callback */
    if (!This->AdaptationCallback)
        return S_NO_RESPONSE;

    return (*This->AdaptationCallback)
                (This->AdaptationData, 
                pTaskBuilder(This->Task)->Planner, 
                This, 
                Data, 
                Statistics, 
                NothingAdapts);
}
    
PRIVATE const _TCHAR * ReservationBuilderGetFullName(
    PIRESERVATIONBUILDER iThis)
{
    PRESERVATIONBUILDER This = pReservationBuilder(iThis);
    return _tcsdup(This->FullName);
}
    
#include "s_ReservationBuilder_v.c"

/* Task builder implementation */

PRIVATE SCODE TaskBuilderQueryInterface (
    PITASKBUILDER pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                                 &IID_ITaskBuilder);
}
    
PRIVATE UINT TaskBuilderAddRef (
    PITASKBUILDER pThis)
{
    return AtomicInc(&pTaskBuilder(pThis)->RefCnt);
}
    
PRIVATE UINT TaskBuilderRelease (
    PITASKBUILDER pThis)
{
    PTASKBUILDER This = pTaskBuilder(pThis);
    UINT Refs = AtomicDec(&This->RefCnt);
    DBGME(1,printf("builder.c:TaskBuilderRelease this=0x%x\n",(ADDRESS)pThis));
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        DBGME(1, printf("builder.c:TaskBuilderRelease: this=0x%x refs=%d\n", 
                        (ADDRESS) This, Refs));
        /* Finalize object */
        if (This->TaskName) {
            free(This->TaskName);
            This->TaskName = NULL;
        }
        if (This->Planner) {
            This->Planner->v->Release(This->Planner);
            This->Planner = NULL;
        }
        if (This->Instigator) {
            This->Instigator->v->Release(This->Instigator);
            This->Instigator = NULL;
        }
        ListWalk(&This->Reservations, rbloff, 
            LIST_WALK_ITERATE, ReservationBuilderDelete, pHeap);
        This->Reservations = NULL;
        pHeap->v->Free(pHeap, 0, (void *) This);
    }
    return Refs;
}
    
INT TaskBuilderDelete(PTR pItem, PTR pArg)
{
    PITASKBUILDER Task = (PITASKBUILDER) pItem;
    PIHEAP pHeap = (PIHEAP) pArg;
    UnusedParameter(pHeap);
    Task->v->Release(Task);
    return 0;
}
    
PRIVATE SCODE MCT TaskBuilderCreateReservation(
    PITASKBUILDER iThis, 
    const _TCHAR *hostUri, 
    const _TCHAR *fbakUri, 
    const _TCHAR *rsrvName, 
    PIRESERVATIONBUILDER *out)
{
    METHOD_DECLARE(TaskBuilder, CreateReservation);
    PRESERVATIONBUILDER builder = NULL;
    PIHEAP heap = CurrentHeap();
    _TCHAR *fullName;

    CHECKSETSC(rsrvName, E_INVALID_PARAMETER, "check name parameter");

    builder = heap->v->Alloc(heap, 
        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME(_T("builder.c:CreateReservation")), 
        sizeof(ReservationBuilder), 0);
    CHECKSETSC(builder, E_NOT_ENOUGH_MEMORY, "create builder");

    builder->v = &ReservationBuilderVtbl;
    builder->RefCnt = 1;

    builder->Reservation = heap->v->Alloc(heap, 
        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME(_T("builder.c:CreateReservation")), 
        sizeof(struct reservation), 0);
    CHECKSETSC(builder->Reservation, E_NOT_ENOUGH_MEMORY,"create reservation");

    builder->Host = _tcsdup(hostUri);

    /* XXX The wildcarding needs to be a parameter to this func --jvh */
    if (!_tcscmp(rsrvName, PRODUCER_RESERVATION_NAME))
        fullName = CreateFullReservationName(This->TaskName, rsrvName, _T("/*"));
    else
        fullName = CreateFullReservationName(This->TaskName, rsrvName, _T(""));
    CHECKSETSC(fullName, E_NOT_ENOUGH_MEMORY, "create full name");

    builder->Reservation->name = _tcsdup(rsrvName);

    builder->FullName = fullName;

    builder->Reservation->contact = _tcsdup(fbakUri); 
    CHECKSETSC(builder->Reservation->contact, E_NOT_ENOUGH_MEMORY,
               "duplicate contact info");

    ListNodeInit(rloff, builder->Reservation);
    ListNodeInit(rbloff, builder);

    builder->Task = iTaskBuilder(This);

    /* Default is not adaptive */
    builder->Confidence = 0; 

    if (!This->Reservations)
        This->Reservations = builder;
    else ListAddLast(&This->Reservations, rbloff, builder);

    *out = (PIRESERVATIONBUILDER) builder;

    sc = S_OK;

METHOD_CLEANUP;
    if (FAILED(sc)) {
        if (builder) {
            if (builder->Host) {
                free(builder->Host);
                builder->Host = NULL;
            }
            if (builder->Reservation) {
                free(builder->Reservation->name);
                free(builder->Reservation->contact);
                heap->v->Free(heap, 0, builder->Reservation);
                builder->Reservation = NULL;
            }
            heap->v->Free(heap, 0, builder);
        }
    }
    return sc;
}
    
PRIVATE SCODE TaskBuilderActivate(
    PITASKBUILDER iThis, PIINSTIGATOR Instigator)
{
    METHOD_DECLARE(TaskBuilder, Activate);

    /* Register the task for maintenance */
    sc = This->Planner->v->ActivateTask(This->Planner, iThis);
    CHECKSC("activate task");

    sc = S_OK;

METHOD_CLEANUP;
    return sc;
}
    
#include "s_TaskBuilder_v.c"

SCODE TaskBuilderCreate(
    PIPLANNER Planner, PIINSTIGATOR Instigator, PITASKBUILDER *Task)
{
    METHOD_DECLARE_NOTHIS(TaskBuilder, Create);
    PTASKBUILDER New = NULL;
    PIHEAP Heap = CurrentHeap();

    *Task = NULL;

    CHECKSETSC(Planner, E_INVALID_PARAMETER, "check task parameters");
    //if (!Instigator)
    //    CHECKSETSC(E_INVALID_PARAMETER, "check task parameters");

    New = Heap->v->Alloc(Heap, 
        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME(_T("planner.c:CreateTask")), 
        sizeof(TaskBuilder), 0);
    CHECKSETSC(New, E_NOT_ENOUGH_MEMORY, "create task");

    New->v = &TaskBuilderVtbl;
    New->RefCnt = 1;

    New->Planner = Planner;
    Planner->v->AddRef(Planner);

    New->Instigator = Instigator;
    if (Instigator)
        Instigator->v->AddRef(Instigator);

    New->TaskName = CreateUniqueTaskName();
    CHECKSETSC(New->TaskName, E_NOT_ENOUGH_MEMORY, "create task name");

    ListNodeInit(tbloff, New);

    *Task = iTaskBuilder(New);
    sc = S_OK;

METHOD_CLEANUP;
    if (FAILED(sc)) {
        if (New) {
            if (New->TaskName)
                free(New->TaskName);
            Heap->v->Free(Heap, 0, New);
        }
    }
    return sc;
}
    
