/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implements Planner concrete class */
#include <mmlite.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <stdio.h>

extern void TimeFormat(TIME t, _TCHAR *buf, UINT Len);

#include "builder.h"
#include "planner.h"

//#define EXTRA 5                 /* create some overlap */
#define EXTRA 0

/* OVERVIEW:

   The Planner is a run-time library that provides an application 
   framework for a client application called the Instigator. The 
   remaining components are the remote Real-time Thread Scheduler, 
   and the Sampler.
   
   The Instigator specifies and sets up real-time interactions 
   between remote hosts, using the Planner. The Instigator's role 
   is to specify configuration data, including reservation parameters
   and the code that claims the reservations. The configuration data
   forms a Task structure, containing one or more reservations.

   The Planner's role is to format the configuration data, send 
   reservation requests to the remote real-time thread schedulers, 
   maintain reservation leases, and collect statistics about 
   reservations to adapt reservation parameters over time. 

   The Sampler receives samples of actual execution times for claimed 
   reservations, taken by the remote Schedulers. The Sampler computes 
   statistics over the collected data. 
*/

/* DATA STRUCTURES:

   The planner maintains a list of Tasks. 

   Each Task has a list of Reservations.

   The planner maintains a Live List of Reservations that require 
   lease renewal and adaptation.
*/

/* THREADING:

   The planner has an interface that is entered by client threads. 

   It has an internal thread that owns the list of live reservations
   (maintainer data structures; live list) and does renewal. 

   It has an internal thread that periodically collects feedback 
   for live reservations, and updates reservation parameters as 
   necessary. 

   It has a house-keeping thread that discards tasks with no living
   reservations from the task list. 

   Only the client and house-keeping threads modify the task list. 
   Only the client and maintainer threads modify the live list. 

   See also lifetime notes below. 
*/

/* FOOTPRINT:

   Three stacks make for a big footprint...

   Both the housekeeper and feedback threads are periodic, and are 
   therefore trivial to combine. 
   
   The maintainer thread wakes up on a time-out or on a signal. If 
   the live list of objects the maintainer thread needs to observe 
   is modified by inserting a virtual object that represents a feedback 
   or house-keeping time-out, then the three threads have been combined 
   into one. 

   Since the instigator thread terminates when the planner has been 
   configured, that means only one thread can take care of everything. 

   Were not going to have time to implement this  but the story is 
   there, so you can feel free to play with it. 

   Caveat  single-threading the planning library would be risky if 
   there were a lot of reservations, because we would start missing 
   renewal deadlines. But, I use a single lock for synchronization as 
   it is, so that has the same problem. This whole thing would need more 
   fine-grained, probably per-object (reservation) locking to scale up. 
*/

/* LIFETIME:

   The planner is destroyed when the reference count reaches 0. 

   Each task in the task list has a reference. The reference 
   count is zeroed when either

    (a) there are no tasks and the last client releases, or
    (b) there are no clients and the house-keeper discards 
        the last task. 

   When the reference count goes to zero both the task list 
   and the live list must be empty. 

   When the reference count goes to zero, all internal threads
   are instructed to stop. 

   The feedback and maintainer thread set their *IsRunning variables
   to FALSE as soon as they see StopAllThreads set, and then exit.

   The house-keeper ignores StopAllThreads until the other threads 
   have quit, then deletes the planner and exits
*/

/* SYNCHRONIZATION:

   One global mutex for the planner protects the planner, the task 
   list and the live list, and is taken by the interface, maintainer, 
   feedback and housekeeper threads. 

   There is a risk of the other threads getting in the way of the 
   maintainer. The other threads must not make remote calls - the 
   feedback thread in particular has to request stats locally. 

   Fixing requires fine-grained locking, most likely at the level 
   of per-reservation and live list entry.
*/

/* RELIABILITY:

   Reservations should be established/renewed/adapted in transactions, 
   but aren't. A single failure halts the system - OK as long as we're 
   just collecting numbers. 

*/

#define LOCK_PLANNER_MUTEX() \
    do { \
    Mutex_Lock(&This->Mx); \
    DBGME(1, printf(_method_name)); \
    DBGME(1, printf(": locked planner mutex=0x%x\n", \
        (ADDRESS) &This->Mx)); \
    } while (0); \

#define UNLOCK_PLANNER_MUTEX() \
    do { \
    DBGME(3, printf(_method_name)); \
    DBGME(3, printf(": unlocking planner mutex=0x%x\n", \
        (ADDRESS) &This->Mx)); \
    Mutex_Unlock(&This->Mx); \
    } while (0); \

#define WAIT_FOR_PLANNER_CONDITION_1() \
    do { \
    DBGME(3, printf(_method_name)); \
    DBGME(3, printf(": waiting for planner condition(1)=0x%x", \
        (ADDRESS) &This->Cnd1)); \
    if (Timeout) \
        DBGME(1, printf(" timeout=0x%llx\n", *Timeout)); \
    else DBGME(1, printf(" timeout=NULL\n")); \
    Condition_Wait(&This->Cnd1, &This->Mx, Timeout); \
    DBGME(3, printf(_method_name)); \
    DBGME(3, printf(": woke on planner condition(1)=0x%x mx=%x\n", \
        (ADDRESS) &This->Cnd1, &This->Mx)); \
    } while (0); \

#define SIGNAL_PLANNER_CONDITION_1() \
    do { \
    DBGME(3, printf(_method_name)); \
    DBGME(3, printf(": signaling planner condition(1)=0x%x\n", \
        (ADDRESS) &This->Cnd1)); \
    Condition_Signal(&This->Cnd1); \
    } while (0); \

#define WAIT_FOR_PLANNER_CONDITION_2() \
    do { \
    DBGME(3, printf(_method_name)); \
    DBGME(3, printf(": waiting for planner condition(2)=0x%x", \
        (ADDRESS) &This->Cnd2)); \
    if (Timeout) \
        DBGME(1, printf(" timeout=0x%llx\n", *Timeout)); \
    else DBGME(1, printf(" timeout=NULL\n")); \
    Condition_Wait(&This->Cnd2, &This->Mx, Timeout); \
    DBGME(3, printf(_method_name)); \
    DBGME(3, printf(": woke on planner condition(2)=0x%x mx=%x\n", \
        (ADDRESS) &This->Cnd2, &This->Mx)); \
    } while (0); \

#define BROADCAST_PLANNER_CONDITION_2() \
    do { \
    DBGME(3, printf(_method_name)); \
    DBGME(3, printf(": broadcasting planner condition(2)=0x%x\n", \
        (ADDRESS) &This->Cnd2)); \
    Condition_Broadcast(&This->Cnd2); \
    } while (0); \

/* Planner interface */

PRIVATE SCODE PlannerQueryInterface(
    PIPLANNER iThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return GenericQueryInterface((IUnknown *)iThis, iid, ppObject,
                                 &IID_IPlanner);
}
    
PRIVATE UINT PlannerAddRef(PIPLANNER iThis)
{
    return AtomicInc(&pPlanner(iThis)->RefCnt);
}
    
PRIVATE UINT PlannerRelease(PIPLANNER iThis)
{
    METHOD_DECLARE(Planner, Release);
    UINT Refs = AtomicDec(&This->RefCnt);
    sc = S_OK;

    if (Refs == 0) {
        LOCK_PLANNER_MUTEX();
        This->StopAllThreads = TRUE;
        SIGNAL_PLANNER_CONDITION_1();
        BROADCAST_PLANNER_CONDITION_2();
        UNLOCK_PLANNER_MUTEX();
    }

    METHOD_CLEANUP;
    return Refs;
}
    
PRIVATE void PlannerDelete(PIPLANNER iThis)
{
    PPLANNER This = pPlanner(iThis);
    PIHEAP pHeap = CurrentHeap();

    assert(This->StopAllThreads);
    assert(0 == This->RefCnt);
    assert(NULL == This->Tasks);
    assert(NULL == This->MaintainerData);

    /* Finalize object */
    Mutex_Destroy(&This->Mx);
    pHeap->v->Free(pHeap, 0, (void *) This);
}
    
/* Allocates and initializes a task builder object. 
   The new task builder contains no reservations. 
   The planner retains a reference to the new task 
   builder, but the task is not active. 
   The task builder also takes a reference to the 
   planner */
PRIVATE SCODE PlannerCreateTask (
    PIPLANNER iThis,
    PIINSTIGATOR Instigator,
    /*out*/ PITASKBUILDER* Task)
{
    METHOD_DECLARE(Planner, CreateTask);
    PITASKBUILDER New;

    sc = TaskBuilderCreate(iPlanner(This), Instigator, &New);
    CHECKSC("create task builder");

    *Task = New;
    
METHOD_CLEANUP;
    if (FAILED(sc)) {
        if (New)
            New->v->Release(New);
    }
    return sc;
}

PRIVATE INT MdCmp(PTR a, PTR b)
{
    PMAINTAINERDATA aa = a;
    PMAINTAINERDATA bb = b;

    INT64 r = Int64Subtract(aa->RenewReservationBy, bb->RenewReservationBy);
    if (Int64IsNegative(r))
        return -1;
    else if (Int64IsZero(r))
        return 0;
    else
        return 1;
}
    
/* Move the parameter maintainer data structure to the right place
   in the list, to ensure the list is correctly sorted */
PRIVATE void PlannerResortMaintainerData(
    PIPLANNER iThis, PMAINTAINERDATA Data)
{
    PPLANNER This = pPlanner(iThis);
    //PMAINTAINERDATA Prev = Data->ListNode.pPrev; 
    //PMAINTAINERDATA Next = Data->ListNode.pNext;

    /* Remove the node from the list */
    ListRemoveItem(&This->MaintainerData, mdloff, Data);

    /* Re-insert it in the right place */
    ListWalk(&This->MaintainerData, mdloff, 
             LIST_WALK_SORTED | LIST_WALK_ADD,
             MdCmp, Data);
#if 0
    Prev->ListNode.pNext = Next;
    Next->ListNode.pPrev = Prev;
    Data->ListNode.pPrev = Data;
    Data->ListNode.pNext = Data;

    if (This->MaintainerData == Data)
        This->MaintainerData = Next;

    /* Re-insert it in the right place */
    if (Int64Less(Data->RenewReservationBy, Prev->RenewReservationBy)) {

        /* The reservation should move towards the head of the list...
           but this never happens, because renewal n+1 never occurs
           before renewal n, so cry foul */
        assert(FALSE);

    } else {

        PMAINTAINERDATA Follower = Next;
        BOOL AtEnd = FALSE;

        while (Int64Less(Follower->RenewReservationBy, Data->RenewReservationBy)) {
            /* Did we reach the end of the list? */
            if (Follower == Follower->ListNode.pNext)
                AtEnd = TRUE;
            else Follower = Follower->ListNode.pNext;
        }

        if (AtEnd) {
            PMAINTAINERDATA Tail = Follower;
            Tail->ListNode.pNext = Data;
            Data->ListNode.pPrev = Tail;
            Data->ListNode.pNext = Data;
        } else {
            Data->ListNode.pNext = Follower;
            Data->ListNode.pPrev = Follower->ListNode.pPrev;
            Data->ListNode.pPrev->ListNode.pNext = Data;
            Follower->ListNode.pPrev = Data;

            /* Did we insert at the front? */
            if (This->MaintainerData == Follower)
                This->MaintainerData = Data;
        }
    }
#endif
}
    
INT MaintainerDataDelete(PTR pItem, PTR pArg)
{
    PMAINTAINERDATA Data = (PMAINTAINERDATA) pItem;
    PIHEAP pHeap = (PIHEAP) pArg;

    assert(ListNodeIsAlone(mdloff, Data));
    pHeap->v->Free(pHeap, 0, Data);

    return 0;
}
    
/* Refreshes the internal maintainer data structures, 
   to account for modifications in a freshly dispatched reservation */
PRIVATE SCODE PlannerRefreshMaintainer(
    PIPLANNER iThis, PIRESERVATIONBUILDER iRes, TIME DispatchTime)
{
    PPLANNER This = pPlanner(iThis);
    PMAINTAINERDATA Data = NULL, Current;
    PRESERVATIONBUILDER Res = pReservationBuilder(iRes);
    struct reservation *Reservation = Res->Reservation;
    struct trigger *Trigger = Reservation->TriggerList;

    assert(This->MaintainerData);

    /* Locate the maintainer data entry for the reservation */
    Current = This->MaintainerData;
    do {
        if (Current->Reservation == Res) 
            Data = Current;
        Current = ListNodeNext(mdloff, Current);
    } while (!Data && This->MaintainerData != Current);

    assert(Data);

    /* Store the dispatch time, just for debugging */
    Data->LastDispatchedAt = DispatchTime;

    /* Adjust counts to account for the latest dispatch */
    {
        /* Adjust the remaining count downwards */
        INT before = Data->RemainingCount;
        Data->RemainingCount -= DEFAULT_TRIGGER_LEASE_COUNT;
        if (0 > Data->RemainingCount)
            Data->RemainingCount = 0;

        /* Adjust the trigger max count upwards */
        Trigger->maxCount += DEFAULT_TRIGGER_LEASE_COUNT;
        if (Trigger->maxCount > Data->OriginalMaxCount)
            Trigger->maxCount = Data->OriginalMaxCount;

        /* Adjust the sequence number upwards */
        Reservation->seqno += DEFAULT_TRIGGER_LEASE_COUNT;

        DBGME(1, printf(
            "planner.c:RefreshMaintainer: live=0x%x res=0x%x original=%d before=%d after=%d\n", 
            (ADDRESS) Data, (ADDRESS) Data->Reservation, 
            Data->OriginalMaxCount, before, Data->RemainingCount));
    }

    /* Does this reservation still need to be maintained? */
    if (0 >= Data->RemainingCount) {

        /* No - the maintainer data can be discarded */
        ListRemoveItem(&This->MaintainerData, mdloff, Data);
        if (Data == This->MaintainerData)
            This->MaintainerData = NULL;
        MaintainerDataDelete((PTR) Data, (PTR) CurrentHeap()); 
        DBGME(2, printf("planner.c:RefreshMaintainer: died=0x%x head=0x%x\n", 
            (ADDRESS) Data, (ADDRESS) This->MaintainerData));

        /* And the reservation trigger is spent */
        Trigger->maxCount = 0;

    } else { 

        /* Set up to renew the reservation later */
        TIME TriggerInterval, MaintainerInterval;

        PrintTime("####Planned timeout now", Reservation->deadline);

        /* TriggerInterval is the length of the entire lease being sent */
        TriggerInterval = Int64TimesInt32(Trigger->offset, DEFAULT_TRIGGER_LEASE_COUNT);

        /* Allow time for transmitting the renewal message */
        MaintainerInterval = 
            Int64SubtractInt32(TriggerInterval, PLANNER_RENEWAL_SLACK);

        /* Reset the reservation renewal time */
        Data->RenewReservationBy = 
            Int64Subtract(Reservation->deadline, MaintainerInterval);

        /* Reset the reservation deadline. */
        Reservation->deadline = 
            Int64Subtract(Reservation->deadline, TriggerInterval);

        PrintTime("####Planned timeout next", Reservation->deadline);

        /* Ensure the maintainer data list of reservations is sorted correctly */
        PlannerResortMaintainerData(iThis, Data);
    }

    return S_OK;
}
    
/* Establishes or renews the reservation at the remote node,  
   and refreshes the internal maintainer data structures */
PRIVATE SCODE PlannerDispatchReservation(PTR pItem, PTR pArg)
{
    SCODE sc = S_OK;
    PIRESERVATIONBUILDER iRes = (PIRESERVATIONBUILDER) pItem;
    PRESERVATIONBUILDER Res = pReservationBuilder(iRes);
    PIPLANNER Planner = pTaskBuilder(Res->Task)->Planner;
    INT DispatchAnyReservation = (INT) pArg;
    BOOL Repeats = SUCCEEDED(iRes->v->IsRepeating(iRes));

    if (DispatchAnyReservation || Repeats) {

        TIME DispatchTime = CurrentTime();

        DBGME(1, 
            printf("planner.c:DispatchReservation: task=0x%x res=0x%x any=%d repeats=%d\n", 
                (ADDRESS) Res->Task, (ADDRESS) Res, 
                DispatchAnyReservation, Repeats));

        sc = iRes->v->Dispatch(iRes);

        if (!SUCCEEDED(sc)) {
            /* XXX No graceful recovery possible
               - this has to run to completion, or else we halt */
            printf("!!!Failed to dispatch to reservation - HALTING\n");
            assert(FALSE); 
        }

        if (Repeats) {
            sc = PlannerRefreshMaintainer(Planner, iRes, DispatchTime);
        }
    }
    return sc;
}
    
/* Move to task builder */
PRIVATE SCODE PlannerDispatchTask(
    PTASKBUILDER Task, INT DispatchAllReservations)
{
    DBGME(1, printf("planner.c:DispatchTask: task=0x%x all=%d\n", 
        (ADDRESS) Task, DispatchAllReservations));

	/* HACK XXX Traversing the list backwards to make the 
	   sensor demo work for the tech fest...
	   
	   The problem is that the producer dispatch does not return, 
	   so the instigator thread is blocked for a long time before 
	   executing the consumer dispatch. The consumer dispatch 
	   misses its deadline and all is not well...
	
	   Since the consumer dispatch is asynchronous we do the backwards
	   walk hack to prevent the crash... and then Johannes fixes the 
	   oh-about-fifteen different bugs that went into this nonsense
	   (after the tech fest...) */
    return (SCODE) ListWalk(&Task->Reservations, rbloff, 
        LIST_WALK_SCODE | LIST_WALK_BACKWARDS, 
		PlannerDispatchReservation, 
        (PTR) DispatchAllReservations);
}
    
/* Creates a maintainer data structure for the parameter 
   reservation, initializes the maintainer data */ 
PRIVATE SCODE MaintainerDataCreate(
    PPLANNER This, PRESERVATIONBUILDER Res, PMAINTAINERDATA *Data)
{
    PIHEAP Heap = CurrentHeap();
    PMAINTAINERDATA New;
    struct trigger *Trigger = Res->Reservation->TriggerList;

    /* Guaranteed to always assert because this function is only 
       called from ActivateReservation and that function checked 
       that the trigger is alive */
    assert(Trigger);

    New = Heap->v->Alloc(Heap, 
        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("planner.c:MaintainerDataCreate"), 
        sizeof(MAINTAINERDATA), 0);
    if (!New) 
        return E_NOT_ENOUGH_MEMORY;

    /* Initial values indicating there has not been a dispatch */
    New->Reservation = Res;
    New->OriginalMaxCount = Trigger->maxCount;
    New->RemainingCount = Trigger->maxCount;
    if (DEFAULT_TRIGGER_LEASE_COUNT < Trigger->maxCount-EXTRA) 
        Trigger->maxCount = DEFAULT_TRIGGER_LEASE_COUNT+EXTRA; 

    /* This is immediately reset, before insertion into the list */
    New->RenewReservationBy = TIME_ORIGIN;

    /* Sufficient for now */
    New->StatisticalPrediction = Res->Reservation->duration;

    ListNodeInit(mdloff, New);

    /* Done */
    *Data = New;
    return S_OK;
}

/* Creates a live list entry for a repeating reservation, 
   or does nothing */
PRIVATE SCODE PlannerActivateReservation(PTR pItem, PTR pArg)
{
    PRESERVATIONBUILDER Res = (PRESERVATIONBUILDER) pItem;
    PIRESERVATIONBUILDER iRes = iReservationBuilder(Res);
    PIPLANNER iThis = pTaskBuilder(Res->Task)->Planner;
    METHOD_DECLARE(Planner, ActivateReservation);
    PBOOL TaskHasNoTriggers = (PBOOL) pArg;
    PMAINTAINERDATA Data;

    /* Check whether the reservation is triggered */
    if (FAILED(iRes->v->IsRepeating(iRes))) {
        /* Nothing to do */
        sc = S_OK;
        goto Out;
    }

    /* The reservation does have a trigger... */
    *TaskHasNoTriggers &= FALSE; 

    /* Does the maintainer thread already know about this reservation? */
    if (This->MaintainerData) {
        PMAINTAINERDATA Current = This->MaintainerData;
        do {
            if (Current->Reservation == Res) {
                /* Yes it does */
                sc = S_OK;
                goto Out;
            }
            Current = ListNodeNext(mdloff, Current);
        } while (This->MaintainerData != Current);
    }

    /* No, maintainer doesn't know about reservation, needs to be told */
    sc = MaintainerDataCreate(This, Res, &Data);
    CHECKSC("create maintainer data");

    /* Insert the new entry at the front of the list, with a renewal 
       time in the distant past. The renewal time must be reset, and 
       the list re-sorted, before the lock is released! This is done 
       in the dispatch code (PlannerDispatchReservation) */
    Data->RenewReservationBy = TIME_ORIGIN;
    ListAddFirst(&This->MaintainerData, mdloff, Data);

    /* Adjust the reservation trigger maxCount */
    if (DEFAULT_TRIGGER_MAX_COUNT < Res->Reservation->TriggerList->maxCount)
        Res->Reservation->TriggerList->maxCount = DEFAULT_TRIGGER_MAX_COUNT+EXTRA;

METHOD_CLEANUP;
    return sc;
}
    
/* Utility for ActivateTask */
PRIVATE INT NodeEq(PTR p1, PTR p2)
{
    return p1 != p2;
}
    
/* Only invoked by planner library clients (e.g. TaskBuilderActivate) */
PRIVATE SCODE PlannerActivateTask(
    PIPLANNER iThis, PITASKBUILDER iTask)
{
    METHOD_DECLARE(Planner, ActivateTask);
    PTASKBUILDER Task = pTaskBuilder(iTask);
    BOOL TaskHasNoTriggers = TRUE;

    if (!iThis || !iTask)
        return E_INVALID_PARAMETER;

    LOCK_PLANNER_MUTEX();
    
    /* Add the task to the task list */
    if (!ListWalk(&This->Tasks, tbloff, LIST_WALK_SEARCH, NodeEq, Task))
        ListAddLast(&This->Tasks, tbloff, Task);
    else CHECKSETSC(FALSE, E_FAIL, "maintain task - task already activated?");

    /* Add the reservations to the maintainer thread's reservation list. 
       The reservations are added to the front of the list, with a 
       renewal time in the distant past. The renewal time is reset and 
       the list sorted properly, by the dispatch code invoked below */
    sc = (SCODE) ListWalk(&Task->Reservations, rbloff, 
        LIST_WALK_SCODE, PlannerActivateReservation, &TaskHasNoTriggers);
    CHECKSC("maintain reservation");

    /* Cycle through the reservations and dispatch for the first time.

       XXX Reservation dispatch does not fail gracefully
       This code simply dispatches until all have succeeded, 
       or one fails - if one fails then no further dispatches are 
       attempted, and no roll-back is attempted either. The result
       is undefined behavior. The fix is changing the scheduler 
       interface to use a two-phase commit protocol. Not trivial. 

       XXX Inserted a DebugBreak() in DispatchReservation, halt on fail.
       
       Note that the maintainer thread dispatches just before updating 
       the maintainer data structures - here we did the opposite, 
       because that way the trigger repeat counts are adjusted before 
       the first dispatch */
    sc = PlannerDispatchTask(Task, 1);
    CHECKSC("dispatch all reservations");

    /* The task's reservations have been dispatched, refreshed and 
       correctly added into the live list - now instruct the maintainer
       thread to look after them, and possibly wake up the others if 
       the live list was empty */
    SIGNAL_PLANNER_CONDITION_1();

METHOD_CLEANUP;
    UNLOCK_PLANNER_MUTEX();
    return sc;
}
    
/* Provides access to the internal planner data structures that maintain
   a reservation. 

   The query name is matched against the paths (full names) of the 
   data entries, so should consist of task name, reservation name and 
   sequence number/wildcard.
   
   This is only intended for use by instigators that 
   implement custom reservation parameter adaptation (e.g. provide an 
   adaptation callback function to RESERVATIONBUILDER::SetAdaptive). 
   
   AND DO NOT INVOKE THIS UNLESS YOU ARE HOLDING THE RIGHT LOCK */
PRIVATE SCODE
PlannerGetPlannerDataByFullName(
    PIPLANNER iThis, 
    const _TCHAR *QueryName, 
    PIRESERVATIONBUILDER *Reservation,
    PTR *MaintainerData)
{
    METHOD_DECLARE(Planner, GetPlannerDataByFullName);
    PMAINTAINERDATA Data = This->MaintainerData;
    PMAINTAINERDATA Found = NULL;
    
    sc = E_NAME_NOT_FOUND;

    /* Check for trivial case of empty maintainer list */
    if (!Data)
        goto Out;

    do {
        /* Check whether this entry matches */

        _TCHAR *CurrentName = Data->Reservation->FullName;
        if (0 == _tcscmp(CurrentName, QueryName)) {
            Found = Data;
            break;
        }

        /* No match, move on to the next one */

        Data = Data->ListNode.pNext;

    } while (NULL != Data && NULL == Found && Data != This->MaintainerData);

    if (Found) {
        *MaintainerData = (PTR) Found;
        *Reservation = iReservationBuilder(Found->Reservation);
        sc = S_OK;
    }

METHOD_CLEANUP;
    return sc;
}

/* Maintenance thread */ 

PRIVATE INT 
MaintainerStopRepeats(PTR pItem, PTR pArg)
{
    PMAINTAINERDATA Data = (PMAINTAINERDATA) pItem;
    struct trigger *Trigger = 
        Data->Reservation->Reservation->TriggerList;
    assert(Trigger);

    Data->RemainingCount = 0;
    Trigger->maxCount = 0;

    return 0;
}
    
void THREAD_LINKAGE MaintainerThread(THREAD_ARGUMENT arg)
{
    PIPLANNER iThis = (PIPLANNER) arg;
    METHOD_DECLARE(Planner, MaintainerThread);
    sc = S_OK;

    LOCK_PLANNER_MUTEX();

    while (TRUE) {
        TIME Now = CurrentTime(), *Timeout;
        PMAINTAINERDATA Pending = This->MaintainerData;

        /* Dispatch all reservations that needed to be renewed in the past.
           Since we timed out on the live list head renewal time it is 
           guaranteed to be in the past */
        while (Pending && Int64Less(Pending->RenewReservationBy, Now)) {

            DBGME(1, 
                printf("planner.c:MaintainerThread: pending=0x%x now=0x%llx renew=0x%llx remain=%d\n", 
                    (ADDRESS) Pending, 
                    Now, Pending->RenewReservationBy, 
                    Pending->RemainingCount));

            PlannerDispatchReservation((PTR) Pending->Reservation, (PTR) 0);

            if (Pending == This->MaintainerData
                && !Int64Less(Now, Pending->RenewReservationBy))
                DBG(("PLANNER.c: RUNNING BEHIND\n"));

            Pending = This->MaintainerData;
            Now = CurrentTime();
        }

        CHECK_HEAP();

        /* Wake up in time for the next reservation, or just sleep if nobody alive... */
        Timeout = Pending ? &Pending->RenewReservationBy : NULL;

        /* Kick the other guys in case they're asleep */
        if (This->Tasks) 
            BROADCAST_PLANNER_CONDITION_2(); 

        /* Wait until time-out and then process the list head, 
           or wait until signal to terminate or re-examine the list */
        WAIT_FOR_PLANNER_CONDITION_1();

        /* Is the planner terminating? */
        if (This->StopAllThreads) break;
    }

    METHOD_CLEANUP;
    /* We're terminating for some reason - mark all reservations dead */
    sc = (SCODE) ListWalk(&This->MaintainerData, tbloff, 
        LIST_WALK_SCODE, MaintainerStopRepeats, This);
    assert(SUCCEEDED(sc));

    This->MaintainerThreadIsRunning = FALSE;
    UNLOCK_PLANNER_MUTEX();
}
    
/* Feedback thread */

/* Adapts reservation parameters based on statistical feedback.

   This method is only invoked in the feedback thread. It is 
   included in the planner interface so instigators that do custom 
   adaptation can use it as a helper. 
   
   Adaptive reservations that do not have an associated adaptation 
   callback are passed to this method by the planner */
PRIVATE SCODE 
PlannerAdaptReservation(
    PIPLANNER iThis, 
    PTR DataPtr, 
    PIRESERVATIONBUILDER iRes, 
    PBOOL NothingAdapts)
{
    METHOD_DECLARE(Planner, AdaptReservation);
    PRESERVATIONBUILDER Res = pReservationBuilder(iRes);
    PMAINTAINERDATA Data = (PMAINTAINERDATA) DataPtr;

    *NothingAdapts &= FALSE;

    /* Compute feedback statistics... */
    sc = This->StatisticsCob->v->GetPredictionFromNormalDistribution(
        This->StatisticsCob, 
        Res->DataCollection, 
        Res->Confidence, 
        &Data->StatisticalPrediction);
    if (FAILED(sc)) {
        printf("planner.c:PlannerAdaptReservation: no feedback, sc=%x\n", sc);
        goto Out;
    }

    /* Adapt the reservation... */
    {
        /* Blend the prediction with the previous estimate */
        TIME StatFraction, EstFraction, tmp;
        StatFraction = Int64TimesInt32(Data->StatisticalPrediction, 4);
        StatFraction = Int64DividedByInt32(StatFraction, 5);
        EstFraction = Int64DividedByInt32(Res->Reservation->duration, 5);

        /* Set the new estimate */
        tmp = Int64Add(StatFraction, EstFraction);
        assert(TimeIsNotPositive(tmp));
        assert(TimeIsNotPositive(StatFraction));
        DBGME(1, printf(
            "planner.c:PlannerAdaptReservation: adjusting from=0x%llx to=0x%llx\n", 
            Res->Reservation->duration, tmp));
        Res->Reservation->duration = tmp;
    }

Out:
    return sc; 
}


/* Adapts reservation parameters based on statistical feedback */
PRIVATE SCODE 
FeedbackVisitReservation(PTR pItem, PTR pArg)
{
    SCODE sc;
    PMAINTAINERDATA Data = (PMAINTAINERDATA) pItem;
    PRESERVATIONBUILDER Res = pReservationBuilder(Data->Reservation);
    PPLANNER This = pPlanner(pTaskBuilder(Res->Task)->Planner);
    PBOOL NothingAdapts = (PBOOL) pArg;

    if (FAILED(Res->v->IsAdaptive(iReservationBuilder(Res)))) {
        sc = S_OK;
        goto Out;
    }

    /* Offer the instigator to run custom adaptation code */

    sc = Res->v->Adapt(
        iReservationBuilder(Res), (PTR) Data, This->StatisticsCob, NothingAdapts);

    /* Three types of return values are possible:

         - S_OK means the reservation self-adapted and the planner
           doesn't need to take any further action;
    
         - FAILED(sc) means the reservation attempted to self-adapt
           but somehow failed;
      
         - S_NO_RESPONSE means the reservation declined to attempt 
           self-adaptation. The planner should apply the default 
           adaptation behavior instead */

    if (SUCCEEDED(sc)) 
        if (S_NO_RESPONSE == sc) 
            sc = This->v->AdaptReservation(
                iPlanner(This), (PTR) Data, iReservationBuilder(Res), NothingAdapts);

Out:
    return sc; 
}
    
/* Updates all live reservation estimates using sampled data */
void THREAD_LINKAGE FeedbackThread(THREAD_ARGUMENT arg)
{
    PIPLANNER iThis = (PIPLANNER) arg;
    METHOD_DECLARE(Planner, FeedbackThread);
    sc = S_OK;

    while (TRUE) {
        BOOL Stop;

        LOCK_PLANNER_MUTEX();

        /* Is the planner terminating? */
        Stop = This->StopAllThreads;

        if (! Stop) {
            BOOL NothingAdapts = TRUE;

            ListWalk(&This->MaintainerData, mdloff, 
                LIST_WALK_ITERATE, FeedbackVisitReservation, (PTR) &NothingAdapts);

            if (!This->MaintainerData || NothingAdapts) {
                /* Wait until a task is added or the planner is stopped */
                TIME *Timeout = NULL;
                WAIT_FOR_PLANNER_CONDITION_2();
            }
        } else {
            This->FeedbackThreadIsRunning = FALSE;
        }
    
        UNLOCK_PLANNER_MUTEX();

        CHECK_HEAP();

        if (Stop) break;

        /* Sleep for a period */
        {
            TIME Timeout;
            Int32ToInt64(Timeout, TIME_RELATIVE(TIME_SECONDS(FEEDBACK_PERIOD)));
            SleepUntil(Timeout);
        }
    }
    METHOD_CLEANUP;
}
    
/* House-keeping thread */

/* Determines whether a reservation is alive */
PRIVATE SCODE HouseKeeperVisitReservation(PTR pItem, PTR pArg)
{
    PIRESERVATIONBUILDER iRes = (PIRESERVATIONBUILDER) pItem;
    PBOOL TaskIsDead = (PBOOL) pArg;

    /* Check whether the reservation still needs to be kept around */
    *TaskIsDead &= FAILED(iRes->v->IsRepeating(iRes));

    return S_OK;
}
    
/* Removes a task from the task list if it has no reservations 
   that need to be renewed */
PRIVATE SCODE HouseKeeperVisitTask(PTR pItem, PTR pArg)
{
    PTASKBUILDER Task = (PTASKBUILDER) pItem;
    PIPLANNER iThis = Task->Planner;
    METHOD_DECLARE(Planner, HouseKeeperVisitTask);
    BOOL TaskIsDead = TRUE;

    UnusedParameter(pArg);

    /* Check each reservation individually. 
       XXX Reservation maintenance does not fail gracefully. 
       If one list item fails then processing stops there. 
       See also reservation dispatch which has the same problem */ 
    sc = (SCODE) ListWalk(&Task->Reservations, rbloff, 
        LIST_WALK_SCODE, HouseKeeperVisitReservation, &TaskIsDead);
    CHECKSC("visit reservation");

    /* Throw out the dead task if appropriate. 
       The task must not be pointed to anywhere outside the task 
       list, which means the reservations already have to be cleared
       out of the maintainer data */
    if (TaskIsDead) {

        DBGME(1,
            printf("planner.c:HouseKeeperVisitTask: dead=0x%x\n", 
                (ADDRESS) Task));

        /* Removing current node during list walk is OK ;) */
        ListRemoveItem(&This->Tasks, tbloff, Task);
        if (Task == This->Tasks)
            This->Tasks = NULL;
        Task->v->Release(iTaskBuilder(Task));
    }

    sc = S_OK;

METHOD_CLEANUP;
    return sc;
}
    
/* Walks the task list and discards dead task objects */
void THREAD_LINKAGE HouseKeeperThread(THREAD_ARGUMENT arg) 
{
    PIPLANNER iThis = (PIPLANNER) arg;
    METHOD_DECLARE(Planner, HouseKeeperThread);

    while (TRUE) {
        BOOL Stop;

        LOCK_PLANNER_MUTEX();

        sc = (SCODE) ListWalk(&This->Tasks, tbloff, 
            LIST_WALK_SCODE, HouseKeeperVisitTask, NULL);
        if (FAILED(sc))
            /* XXX Internal error results in memory leak here 
               - but never happened so far */
            printf("Failed to visit task for housekeeping\n");
    
        if (!This->Tasks)
            DBGME(1, printf("planner.c:HouseKeeperThread: no tasks left\n"));

        /* Is the planner terminating? */
        Stop = This->StopAllThreads 
            && ! This->FeedbackThreadIsRunning
            && ! This->MaintainerThreadIsRunning;

        if (!Stop) {
            /* Wait until a task is added or the planner is stopped */
            TIME *Timeout = NULL;
            WAIT_FOR_PLANNER_CONDITION_2();
        }

        UNLOCK_PLANNER_MUTEX();

        CHECK_HEAP();

        if (Stop) break;
        else {
            /* Sleep for a period */
            TIME Timeout;
            Int32ToInt64(Timeout, TIME_RELATIVE(TIME_SECONDS(HOUSE_KEEPER_PERIOD)));
            SleepUntil(Timeout);
        }
    }
    METHOD_CLEANUP;
}
    
/* Planner component */

#include "s_Planner_v.c"
    
const struct IPlannerVtbl PlannerVtbl;
    
PRIVATE SCODE StartAThread(THREAD_FUNCTION func, char *name,
                           THREAD_ARGUMENT arg, BOOL *pIsRunning)
{
    METHOD_DECLARE_NOTHIS(planner.c, StartAThread);

    /* Ensure the thread runs within the planner COB process */
    PIPROCESS prc;
    PITHREAD pthd;
    PIMODULE mod;

    assert(!pIsRunning || !*pIsRunning);

    sc = GetModuleContainsAddr((ADDRESS) func, &mod);
    CHECKSC("get planner COB module");

    sc = mod->v->QueryInterface(mod, &IID_IProcess, (void **) &prc);
    mod->v->Release(mod);
    CHECKSC("get process from planner COB module");

    sc = prc->v->CreateThread(prc, CREATE_THREAD_STATIC_SCOPING,
                              func, arg, 0, NULL, &pthd);
    prc->v->Release(prc);
    CHECKSC("create thread");
    pthd->v->Release(pthd);

    DBGME(2, printf("planner.c:StartAThread: thd=0x%x %s\n", pthd, name));

    if (pIsRunning)
        *pIsRunning = TRUE;

METHOD_CLEANUP;
    return sc;
}
    
PIUNKNOWN CreatePlannerComponent(void)
{
    METHOD_DECLARE_NOTHIS(planner.c, CreatePlannerComponent);
    PPLANNER This;
    PIHEAP heap = CurrentHeap();
    PISTATISTICS StatisticsCob;

    sc = S_OK;

    This = (PPLANNER) heap->v->Alloc(heap, HEAP_ZERO_MEMORY
                   | HEAP_DEBUG_NAME(_T("planner.c:CreatePlannerComponent")), 
               sizeof(Planner), 0);
    if (! This)
        return NULL;

    This->v = &PlannerVtbl;
    This->RefCnt = 1;

    This->Tasks = NULL;
    This->MaintainerData = NULL;

    Mutex_Init(&This->Mx);
    Condition_Init(&This->Cnd1);
    Condition_Init(&This->Cnd2);
    This->StopAllThreads = FALSE;

    LOCK_PLANNER_MUTEX();

    /* Load the local feedback collector */
    sc = BindToObject(CurrentNameSpace(), _T("COB/statistics.cob"),
                      NAME_SPACE_READ,
                      &IID_IStatistics, (void **) &StatisticsCob);
    CHECKSC("load statistics COB");
    This->StatisticsCob = StatisticsCob;

    StartAThread(MaintainerThread, "MaintainerThread",
                 iPlanner(This), &This->MaintainerThreadIsRunning);
    StartAThread(FeedbackThread, "FeedbackThread",
                 iPlanner(This), &This->FeedbackThreadIsRunning);
    StartAThread(HouseKeeperThread, "HouseKeeperThread",
                 iPlanner(This), &This->HouseKeeperThreadIsRunning);

    METHOD_CLEANUP;
    UNLOCK_PLANNER_MUTEX();

    return (PIUNKNOWN) This;
}
    
PIUNKNOWN plannerCobMain(void)
{
    return (PIUNKNOWN) CreatePlannerComponent();
}
    
