/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <assert.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <wsifhelp.h>
#include "sampling.h"

#define _MANUAL_COBMAIN 1
#include "s_Sampling_v.c"

PRIVATE SCODE MCT SamplingQueryInterface(PISAMPLING pSmpl, REFIID Iid, void** Unk)
{
    SCODE sc;
    PSAMPLING pThis = pSampling(pSmpl);

    sc = GenericQueryInterface((IUnknown *) pSmpl, Iid, Unk, &IID_ISampling);
    DBG(("SamplingQI sc=%x stats=%x unk=%x\n", 
        sc, (ADDRESS) pThis, (ADDRESS) *Unk));

    return sc;
}
        
PRIVATE UINT MCT SamplingAddRef(PISAMPLING pSmpl)
{
    UINT Refs;
    PSAMPLING pThis = pSampling(pSmpl);

    Refs = AtomicInc(&pThis->Refs);

    DBG(("SamplingAddRef: Refs=%d\n", Refs));
    return (Refs);
}

PRIVATE SCODE SamplingDestroy(IUnknown *iThis)
{
    PSAMPLING pThis = pSampling(iThis);
    return S_OK;
}

PRIVATE UINT MCT SamplingRelease(PISAMPLING pSmpl)
{
    UINT Refs;
    PSAMPLING This = pSampling(pSmpl);

    Refs = AtomicDec(&This->Refs);

    if (Refs == 0)
        SamplingDestroy((IUnknown *) This);

    return (Refs);
}

PRIVATE SCODE MCT RewriteUrl(const _TCHAR *Url, _TCHAR *Buf, UINT BufLen)
{
#define STATS_SUFFIX _T("/COB/statistics.cob")
#define STATS_SX_LEN 18
    UINT UrlLen = _tcslen(Url);
    _TCHAR *Suffix = Buf;

    if (UrlLen + STATS_SX_LEN >= BufLen)
        goto LongUrl;

    _tcscpy(Buf, Url);

    /* Pass the http://a.b.c.d/ prefix */
    { 
        UINT i = 0;
        while ('\0' != *Suffix) {
            if ('/' == *Suffix)
                if (++i >= 3)
                    break;
            ++Suffix;
        }
    }

    _tcscpy(Suffix, STATS_SUFFIX);

    DBG(("sampling.c:RewriteUrl: Rewrote url=%s to url=%s\n", Url, Buf));
    return S_OK;

LongUrl:
    DBG(("sampling.c:RewriteUrl: Too long: url=%s\n", Url));
    return E_FAIL;
}

PRIVATE SCODE MCT UpdateProxy(PSAMPLING pThis, const _TCHAR *Url)
{
    SCODE sc;
    char *ErrorMessage = NULL;
    PISTATISTICS Pxy = NULL;
    PISOAPFACTORY Soap = NULL;
    _TCHAR Buf[PROXY_URL_MAXLEN];

    DBG(("sampling.c:UpdateProxy: url=%s\n", Url));

    /* Rewrite the parameter URL with the COB/statistics.cob suffix */
    sc = RewriteUrl(Url, Buf, PROXY_URL_MAXLEN);
    if (FAILED(sc)) {
        ErrorMessage = "sampling.c:UpdateProxy: Bad URL\n";
        goto BadUrl;
    }

    /* Was a proxy for this URL already cached? */
    if (pThis->Proxy && 0 == _tcscmp(pThis->ProxyUrl, Buf))
        goto Out;

    /* Load the SOAP COB */
    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &Soap);
    if (FAILED(sc)) {
        ErrorMessage = "sampling.c:SamplingAddSample: No SOAP COB\n";
        goto NoSoap;
    }

    /* Create a proxy for the remote statistics computation module */
    sc = Soap->v->
        CreateProxy(Soap, NULL,
                    Buf, _T("stats"),
                    0,
                    (PIUNKNOWN*) &Pxy);
    if (FAILED(sc)) {
        ErrorMessage = "sampling.c:SamplingAddSample: No collector proxy\n";
        goto NoProxy;
    }

    /* Get rid of the old cached proxy and cache the new one with its URL */
    if (pThis->Proxy) 
        pThis->Proxy->v->Release(pThis->Proxy);
    pThis->Proxy = Pxy;
    _tcscpy(pThis->ProxyUrl, Buf);

    DBG(("sampling.c:UpdateProxy: Created proxy for url=%s\n", pThis->ProxyUrl));

    sc = S_OK;
    goto Out;

NoProxy: 
    /* Oops - screwed up the stored URL for the cached proxy; wipe the slate */
    *pThis->ProxyUrl = '\0';
    /* Fall-thru */
NoSoap: 
    /* Fall-thru */
BadUrl: 
    /* Fall-thru */
Out:
    if (ErrorMessage) 
        DBG((ErrorMessage));
    return sc;
}

PRIVATE SCODE MCT SamplingAddSample(
    PISAMPLING pSmpl, 
    const _TCHAR *Name, UINT Number,
    TIME ReceiveTime, TIME SendTime,
    TIME Requested, TIME Consumed, 
    const _TCHAR *Url)
{
    SCODE sc;
    PSAMPLING pThis = pSampling(pSmpl);

    DBG(("sampling.c:SamplingAddSample name=%s number=%d url=%s\n", 
        Name, Number, Url));

    Mutex_Lock(&pThis->Mx);

    /* Check whether the right proxy has been cached already, or create new */
    sc = UpdateProxy(pThis, Url);
    if (FAILED(sc)) goto Out;

    /* Simply deliver the data to the collector.
     * We may later inline this code at the invocation point and lose this COB, 
     * or decide to do something more clever - i.e. real sampling. 
     * But this is all we need for now
     */
    sc = pThis->Proxy->v->
        AddSample(pThis->Proxy, Name, Number,
                  ReceiveTime, SendTime, Requested, Consumed);
    if (FAILED(sc)) {
        DBG(("sampling.c:SamplingAddSample: No delivery for name=%s number=%d sc=%x\n", 
            Name, Number, sc));
        goto Out;
    }

    DBG(("sampling.c:SamplingAddSample: success name=%s number=%d\n", Name, Number));
    sc = S_OK;

Out:
    Mutex_Unlock(&pThis->Mx);
    return S_OK;
}

PRIVATE SCODE MCT SamplingReportError(
    PISAMPLING pSmpl, 
    const _TCHAR *Name, UINT Seqno, TIME When,
    SCODE err, const _TCHAR *Message,
    const _TCHAR *Url)
{
    SCODE sc;
    PSAMPLING pThis = pSampling(pSmpl);

    DBGT((_T("sampling.c:SamplingReportError name=%s seqno=%d sc=%x msg=%s url=%s\n"), Name, Seqno, err, Message, Url));

    sc = E_NOT_IMPLEMENTED;
    assert(FALSE);

    return sc;
}

/* Component */

PIUNKNOWN CreateSamplingComponent(void)
{
    PSAMPLING pSa;
    PIHEAP pHeap = CurrentHeap();

    pSa = (PSAMPLING) pHeap->v->
        Alloc(pHeap, HEAP_DEBUG_NAME(_T("sampling.c:CreateSamplingComponent")), 
        sizeof(Sampling) + PROXY_URL_BUFSZ, 0);
    if (! pSa)
        return NULL;

    pSa->v = &SamplingVtbl;
    pSa->Proxy = NULL;
    pSa->ProxyUrl = (_TCHAR *) (pSa + 1);
    pSa->Refs = 1;

    Mutex_Init(&pSa->Mx);

    return (PIUNKNOWN) pSa;
}

PIUNKNOWN samplingCobMain(void)
{
    return (PIUNKNOWN) CreateSamplingComponent();
}



