/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <assert.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <wsifhelp.h>
#include "statistics.h"

#include "s_Statistics_v.c"

#if 0
PRIVATE SCODE MCT StatisticsQueryInterface(PISTATISTICS pStat, REFIID Iid, void** Unk)
{
    SCODE sc;
    PSTATISTICS pThis = pStatistics(pStat);

    UnusedParameter(pThis);

    sc = GenericQueryInterface((IUnknown *) pStat, Iid, Unk, &IID_IStatistics);
    DBGME(0, printf("StatisticsQI sc=%x stats=%x unk=%x\n", 
        sc, (ADDRESS) pThis, (ADDRESS) *Unk));

    return sc;
}
          
PRIVATE UINT MCT StatisticsAddRef(PISTATISTICS pStat)
{
    UINT Refs;
    PSTATISTICS pThis = pStatistics(pStat);

    Refs = AtomicInc(&pThis->RefCnt);

    DBGME(0, printf("StatisticsAddRef: Refs=%d\n", Refs));
    return (Refs);
}

PRIVATE UINT MCT StatisticsRelease(PISTATISTICS pStat)
{
    UINT Refs;
    PSTATISTICS This = pStatistics(pStat);

    Refs = AtomicDec(&This->Refs);

    if (0 >= Refs)
        StatisticsDestroy((IUnknown *) This);

    DBGME(0, printf("StatisticsRelease: Refs=%d\n", Refs));
    return (Refs);
}
#endif

PRIVATE SCODE StatisticsDestroy(IUnknown *iThis)
{
    PSTATISTICS This = pStatistics(iThis);


    /* Free the sample pool names */
    if (This->Names) {
        INT i; 
        for (i = 0; NUM_SAMPLE_POOLS > i; ++i)
            if (This->Names[i])
                free(This->Names[i]);
    }

    /* Free the sample pools */
    if (This->Data) {
        /* Doesn't need to be freed separately, except names in samples */
#ifdef _DEBUG
        UINT i, j;
        for (i = 0; NUM_SAMPLE_POOLS > i; ++i) {
            PSAMPLE w = This->Data[i].Window;
            for (j = 0; j < WINDOW_SIZE; j++)
                if (w[j].Name)
                    free(w[j].Name);
        }
#endif
    }

    if (This->Log)
        This->Log->v->Release(This->Log);

//        Mutex_Destroy(&This->Mx);

    CurrentHeap()->v->Free(CurrentHeap(), 0, This);

    return S_OK;
}
    
PRIVATE SCODE MCT StatisticsGetSamplePool(
    PISTATISTICS pStat, 
    const _TCHAR *Name, 
    PDATA *ppData)
{
    SCODE sc;
    PSTATISTICS pThis = pStatistics(pStat);
    _TCHAR *Pool;
    UINT i;

    *ppData = NULL;

    /* Assume the pool doesn't exist */
    sc = E_NAME_NOT_FOUND;

    /* Look for a pool name that matches the input parameter */
    for (i = 0; (! *ppData) && (i < NUM_SAMPLE_POOLS); ++i) {
        Pool = pThis->Names[i];
        /* Is there a data collection in this slot? 
           If so, does it match the parameter name? */
        if (Pool && 0 == _tcscmp(Name, Pool)) {
            /* Yes - return this pool then */
            *ppData = &(pThis->Data[i]);
            sc = S_OK;
            break;
        }
    }

    return sc;
}
    
PRIVATE SCODE MCT StatisticsCreateSamplePool(
    PISTATISTICS pStat, 
    const _TCHAR *Name, 
    PDATA *ppData)
{
    SCODE sc;
    PSTATISTICS pThis = pStatistics(pStat);
    UINT i, free = -1;

    /* Find an empty slot */
    for (i = 0; -1 == free && i < NUM_SAMPLE_POOLS; ++i) 
        if (! pThis->Names[i]) 
            free = i;

    do {
        /* Give up if there is no empty slot */
        if (-1 == free) {
            sc = E_NOT_ENOUGH_MEMORY;
            break;
        }

        /* Create a new pool using an empty slot and return that */
        pThis->Names[free] = _tcsdup(Name);
        *ppData = &(pThis->Data[free]);
        sc = S_OK;

    } while (0);

    return sc;
}
    
PRIVATE SCODE StatisticsWriteToLog(
    PISTATISTICS iThis, 
    const _TCHAR *Name, UINT Number,
    PDATA Data, 
    PSAMPLE NewSample);
    
PRIVATE SCODE MCT StatisticsAddSample(
    PISTATISTICS pStat, 
    const _TCHAR *Name, UINT Number,
    TIME ReceiveTime, TIME SendTime,
    TIME Requested, TIME Consumed)
{
    SCODE sc;
    PSTATISTICS pThis = pStatistics(pStat);
    PDATA pData;
    PSAMPLE pNewSample;

    DBGME(1, printf(
        "statistics.c:AddSample: name=%s number=%d\n", Name, Number));

    Mutex_Lock(&pThis->Mx);

    /* Demultiplex to the right sample pool */
    sc = StatisticsGetSamplePool(pStat, Name, &pData);
    if (! pData) {
        sc = StatisticsCreateSamplePool(pStat, Name, &pData);
        if (FAILED(sc)) goto Out; 
    }
    
    /* Locate the next empty slot in the window */
    if (! pData->LatestSample) {
        pNewSample = &(pData->Window[0]);
    }
    else { 
        pNewSample = pData->LatestSample + 1;
        if (pData->Window + WINDOW_SIZE <= pNewSample + 1)
            pData->IsWindowFull = TRUE;
        if (pData->Window + WINDOW_SIZE <= pNewSample) 
            pNewSample = &(pData->Window[0]);
    }

    /* Write the new sample */
#if _DEBUG
    if (pNewSample->Name)
        free(pNewSample->Name);
    pNewSample->Name = _tcsdup(Name);
    pNewSample->Number = Number;
#endif
    pNewSample->Requested = Requested;
    pNewSample->Consumed = Consumed;

    /* Done adding the sample */
    pData->LatestSample = pNewSample;
    sc = S_OK;

Out:
    Mutex_Unlock(&pThis->Mx);

    if (SUCCEEDED(sc))
        StatisticsWriteToLog(pStat, Name, Number, pData, pNewSample);

    return sc;
}
    
#define PRECISION_SHIFT 4
    
PRIVATE SCODE MCT StatisticsGetNormalDistribution(
    PISTATISTICS pStat, 
    const _TCHAR *Name, 
    TIME *pMean, 
    TIME *pVariance)
{
    SCODE sc;
    PSTATISTICS pThis = pStatistics(pStat);
    char *ErrorMessage = NULL;
    PDATA pData;
    INT64 Mean, Variance;
    UINT i, NeedsUnlock = FALSE;

    /* Get the sample pool */
    DBGME(1, printf(
        "statistics.c:GetNormalDistribution: Sample pool name=%s\n", 
        Name));
    sc = StatisticsGetSamplePool(pStat, Name, &pData);
    if (! pData) {
        ErrorMessage = 
            "statistics.c:GetNormalDistribution: Sample pool not found\n";
        sc = E_NAME_NOT_FOUND;
        goto Out;
    }

    Mutex_Lock(&pThis->Mx);
    NeedsUnlock = TRUE;

    /* Check whether enough samples are available to predict from */
    if (! pData->IsWindowFull) {
        /* Not enough data - zero the out parameters and quit */
        Int32ToInt64(*pMean, 0);
        Int32ToInt64(*pVariance, 0);
        ErrorMessage = 
            "statistics.c:GetNormalDistribution: Insufficient data\n";
        sc = E_FAIL;
        goto Out;
    }

    /* Compute the mean over the window of samples */
    Int32ToInt64(Mean, 0);
    for (i = 0; i < WINDOW_SIZE; ++i) 
        Mean = Int64Add(Mean, pData->Window[i].Consumed);
    Mean = Int64LShift(Mean, PRECISION_SHIFT); /* Avoid truncation */
    Mean = Int64AddInt32(Mean, WINDOW_SIZE/2); /* Correct for truncation */
    Mean = Int64DividedByInt32(Mean, WINDOW_SIZE);

    /* Compute the variance over the window of samples */
    Int32ToInt64(Variance, 0);
    for (i = 0; i < WINDOW_SIZE; ++i) {
        INT64 Sample = Int64LShift(pData->Window[i].Consumed, 4); 
        INT64 Distance = Int64Subtract(Sample, Mean);
        INT64 Square = Int64TimesInt64(Distance, Distance);
        Variance = Int64Add(Variance, Square);
    }
    Variance = Int64AddInt32(Variance, (WINDOW_SIZE - 1) >> 1); /* Truncation */
    Variance = Int64DividedByInt32(Variance, WINDOW_SIZE - 1); /* -1 b/c sampling */

    /* Done */
    *pMean = Int64RShift(Mean, 4); 
    *pVariance = Int64RShift(Variance, PRECISION_SHIFT << 1); /* Double shift b/c square */

    sc = S_OK;

Out:
    if (NeedsUnlock)
        Mutex_Unlock(&pThis->Mx);
    if (ErrorMessage)
        DBGME(1, printf(ErrorMessage));
    return sc;
}
    
/* If m and s be the mean and standard deviation defining
   a normal distribution then the predicted estimate p is 
   defined as p = m + k * s, where k is a multiplication 
   constant. This is a table t of 49 values t[i] = 2^8 * k, 
   chosen such that the multiplication constant t[i] will 
   cover (i + 50)% of the distribution of resource consumption 
   samples. 
   
   The table was computed with the Excel formula
     =FLOOR(((NORMINV(0.01 * C50, 40, 1.5) - 40)/1.5) * 2^8)
   with the C column holding the values i = 50...99 at indices 
   50...99. Note the left shift by 8 bits at the end. */
INT ConfidenceTable[] = { 
    0, 6, 12, 19, 25, 32, 38, 45, 51, 58, 
    64, 71, 78, 84, 91, 98, 105, 112, 119, 126, 
    134, 141, 149, 156, 164, 172, 180, 189, 197, 206,   
    215, 224, 234, 244, 254, 265, 276, 288, 300, 313,  
    328, 343, 359, 377, 398, 421, 448, 481, 525, 595 };

extern INT64 Int64SquareRoot(INT64 x);
    
PRIVATE SCODE MCT StatisticsGetPredictionFromNormalDistribution(
    PISTATISTICS pStat, 
    const _TCHAR *Name, 
    UINT Confidence,
    PTIME pPrediction)
{
    /* No locking is necessary because internal data structures 
       are not touched directly */

    SCODE sc;
    char *ErrorMessage = NULL;
    INT64 Multiplier, Mean, Variance, StdDev;
    
    Int32ToInt64(*((INT64*) pPrediction), 0);

    if (50 > Confidence || 99 < Confidence) {
        ErrorMessage = "statistics.c:GetPrediction: Invalid confidence parameter\n";
        sc = E_INVALID_PARAMETER;
        goto Out;
    }

    Int32ToInt64(Multiplier, ConfidenceTable[Confidence - 50]);

    sc = StatisticsGetNormalDistribution(pStat, Name, &Mean, &Variance);
    if (FAILED(sc)) {
        ErrorMessage = "statistics.c:GetPrediction: Distribution not computed\n";
        goto Out;
    }

    StdDev = Int64SquareRoot(Variance);

    /* Subtract the confidence interval from the mean, 
       because the values are relative time intervals, 
       which means they are negative 64-bit integers */
    *pPrediction = Int64Subtract(Mean, Int64RShift(Int64TimesInt64(Multiplier, StdDev), 8));
    DBGME(1, printf("statistics.c:GetPrediction: name=%s confidence=%d prediction=%lld\n", 
        Name, Confidence, *pPrediction));
    
    sc = S_OK;

Out:
    if (ErrorMessage)
        DBG((ErrorMessage));
    return sc;
}
    
PRIVATE SCODE StatisticsWriteToLog(
    PISTATISTICS iThis, 
    const _TCHAR *Name, UINT Number,
    PDATA Data, 
    PSAMPLE NewSample)
{
    METHOD_DECLARE(Statistics, WriteToLog);

    UINT Confidences[] = { 90, 95, 99 };
#define NUM_PREDICTS sizeof(Confidences)/sizeof(UINT)
    TIME Predictions[NUM_PREDICTS];
#define BufSz 1024
    _TCHAR Buffer[BufSz];
    UINT i;
    INT NeedsUnlock = FALSE;

    CHECKSETSC(Buffer, E_FAIL, "create buffer");

    /* This->Mx must not be held */

    /* Compute the predictions */
    for (i = 0; NUM_PREDICTS > i; ++i) {
        sc = StatisticsGetPredictionFromNormalDistribution(
            iThis, Name, Confidences[i], &(Predictions[i]));
        if (FAILED(sc))
            Predictions[i] = INT64_ZERO;
    }

    Mutex_Lock(&This->Mx);
    NeedsUnlock = TRUE;

    /* Lazy open the log */
    if (!This->Log) {
        _stprintf(Buffer, _T("fs/sampling.log"));
        sc = BindToObject(
            CurrentNameSpace(), Buffer,
            NAME_SPACE_CREATE | NAME_SPACE_WRITE,
            &IID_IFile, (void **) &This->Log);
        CHECKSC("open log");
    }

    /* Fill the buffer 
       - don't change this without growing it!
       And don't count the bytes wrong ;) 
       The whole buffer will be written, 
       so use it all or make a bug... */
    {
        INT Offset = 0;
        INT64 ForPrinting1, ForPrinting2;
        ForPrinting1 = Int64TimesInt32(NewSample->Requested, -1);
        ForPrinting2 = Int64TimesInt32(NewSample->Consumed, -1);
        _stprintf(Buffer, 
            _T("0x%x,\t%lld,\t%lld,\t"), Number, ForPrinting1, ForPrinting2);
        Offset = _tcslen(Buffer);
        for (i = 0; NUM_PREDICTS > i; ++i) {
            ForPrinting1 = Int64TimesInt32(Predictions[i], -1);
            _stprintf(Buffer + Offset, _T("%lld,\t"), ForPrinting1);
            Offset = _tcslen(Buffer);
        }
        _stprintf(Buffer + Offset, Name);
        Offset = _tcslen(Buffer);
        _stprintf(Buffer + Offset, _T("\n"));
    }

    /* Write the buffer to the log file*/
    {
        UINT Wrote = 0;
        UINT64 LogSz = Int64ToUint64(INT64_ZERO);

        sc = This->Log->v->GetSize(This->Log, &LogSz);
        CHECKSC("get log size");

        sc = This->Log->v->WriteAt(
            This->Log, LogSz, (const PBYTE) Buffer, _tcslen(Buffer), &Wrote);
        CHECKSC("write to log");
    }

METHOD_CLEANUP;
    if (This->Log) {
        /* Must release to get the file written to disk, on ntubig */
        This->Log->v->Release(This->Log);
        This->Log = NULL;
    }

    if (NeedsUnlock)
        Mutex_Unlock(&This->Mx);

    return sc;
}

#if 0    
/* Component */

const struct IStatisticsVtbl StatisticsVtbl; 

PIUNKNOWN CreateStatisticsComponent(void)
{
    PSTATISTICS pSa;
    PIHEAP pHeap = CurrentHeap();

    pSa = (PSTATISTICS) pHeap->v->Alloc(pHeap, 
            HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME(_T("statistics.c:CreateStatisticsComponent")), 
            sizeof(Statistics), 0);
    if (! pSa)
        return NULL;

    pSa->v = &StatisticsVtbl;
    pSa->Refs = 1;
    
    Mutex_Init(&pSa->Mx);

    return (PIUNKNOWN) pSa;
}
    
PIUNKNOWN StatisticsCobMain(void)
{
    return (PIUNKNOWN) CreateStatisticsComponent();
}
#endif




