/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* XML parser implementation */

#include <string.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <tokenizerif.h>
#include <tokenizerutils.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <netif.h>
#include <stdlib.h>
#include "sax.h"

#define TOKEN_invalid (-1)

/* SAX parser structure. Cob object. */
typedef struct _SAXPARSER {
    struct ISaxParserVtbl const *v;
    UINT                Refs;
} *PSAXPARSER;

/* IUnknown implementation */
PRIVATE SCODE MCT
SaxParserQueryInterface(PISAXPARSER Parser, REFIID Iid, void** Unk)
{
    TRACE("SaxParserQueryInterface");

    return GenericQueryInterface((PIUNKNOWN) Parser, Iid, Unk,
                                 &IID_ISaxParser);
}

PRIVATE UINT MCT
SaxParserAddRef(PISAXPARSER Parser)
{
    UINT Refs;
    PSAXPARSER This = (PSAXPARSER) Parser;

    TRACE("SaxParserAddRef");

    Refs = AtomicInc(&This->Refs);
    return (Refs);
}

PRIVATE UINT MCT
SaxParserRelease(PISAXPARSER Parser)
{
    UINT Refs;
    PSAXPARSER This = (PSAXPARSER) Parser;

    TRACE("SaxParserRelease");

    Refs = AtomicDec(&This->Refs);
    /* This object is static variable so no need to free */
    
    return (Refs);
}

/* Use AllocateAndCopyNextToken instead of NextToken to solve the problem
 * of splitting element or attr names.
 * XXX Should not allocate unnecessarily. Use m_pullup?
 */

#define STARTED     0x40000000
#define ROOTFOLLOWS 0x20000000

PRIVATE SCODE SaxParserParse(PISAXPARSER Parser, PITOKENIZER Tokenizer,
                             PISAXHANDLER Handler, UINT *pDepth)
{
    METHOD_DECLARE_NOTHIS(SaxParser, Parse);
    PBYTE*  Content   = Tokenizer->v->GetContent(Tokenizer);
    PBYTE*  LookAhead = Tokenizer->v->GetLookAhead(Tokenizer);
    PBYTE*  LookAhead2 = Tokenizer->v->GetLookAhead2(Tokenizer);
    PBYTE*  LookAhead3 = Tokenizer->v->GetLookAhead3(Tokenizer);

    PIHEAP  Heap                = Tokenizer->v->GetHeap(Tokenizer);

    _TCHAR *   AttributeName      = NULL;
    UINT    AttributeNameLength = 0;

    _TCHAR *  Token               = NULL; 
    UINT    TokenLength         = 0;
    _TCHAR *  ElementName         = NULL;
    UINT    ElementNameLength   = 0;

    _TCHAR *tmp;
    UINT i, AttrCount;
#define MAXATTRIBUTES 15
    SAXATTRIBUTE Attributes[MAXATTRIBUTES], *Attr, _elem, *Elem = &_elem;

    if (*pDepth == 0) {
        sc = Handler->v->StartDocument(Handler);
        CHECKSC("StartDocument");
            *pDepth |= STARTED;
    }

    memset(Elem, 0, sizeof(*Elem));
    Elem->Name = NULL;
    Elem->Token = TOKEN_invalid;
    Elem->Prefix = _T("");
    AttrCount = 0;

    /* Parse the elements */
    do {
        UINT        StartElement        = 0;
        UINT        EndElement          = 0;
        UINT        isAttribute         = 0;
        UINT        StartInstruction    = 0;
        UINT        EndInstruction      = 0;

        if (Elem->Name) {
            /* Clear element and attribute table from previous round */
            Heap->v->Free(Heap, 0, _tcslen(Elem->Prefix) ? (PTR)Elem->Prefix
                                                        : (PTR)Elem->Name);
            for (i = 0, Attr = &Attributes[0]; i < AttrCount; i++, Attr++) {
                if (_tcslen(Attr->Prefix))
                    Heap->v->Free(Heap, 0, (PTR) Attr->Prefix);
                else if (Attr->Name)
                    Heap->v->Free(Heap, 0, (PTR) Attr->Name);
                if (_tcslen(Attr->Value))
                    Heap->v->Free(Heap, 0, (PTR) Attr->Value);
            }
            AttrCount = 0;
            ElementName = NULL;
            ElementNameLength = 0;
            Elem->Name = NULL;
            Elem->Token = TOKEN_invalid;
            Elem->Xns = NULL;
        }

        EatWHITESPACE(Content);

        /* Deal with characters between the current location and the next element */
        Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead);
        sc = Tokenizer->v->AllocateAndCopyNextToken(Tokenizer, 
                                                    Content,
                                                    LookAhead, 
                                                    (PBYTE) "<", 1, 
                                                    &Token,
                                                    &TokenLength);
//assert(sc == S_OK || sc == E_WOULD_BLOCK); 
        CHECKSC("next <");
 
        EatWHITESPACE(LookAhead);

        if (TokenLength > 0) {
            isAttribute = 1;
            /* Strip trailing whitespace */
            while (TokenLength > 0 && IsWhitespace(&Token[TokenLength-1])) {
                Token[TokenLength-1] = _EOS; /* For DeserializeText simple list */
                TokenLength--;
            }
            sc = Handler->v->Text(Handler, Token, TokenLength, NULL, 0);
        }
        if (Token) {
            Heap->v->Free(Heap, 0, (PTR) Token); 
            Token = NULL; 
            TokenLength = 0;
        }
        CHECKSC("Text");

        /* Locate the left bracket */ 
        Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);

        /* Advance past the left bracket */
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
        CHECKSC("Advance past <");

        /* Closing or opening? */
        if (IsASCII(*LookAhead, '/')) {
            /* Closing */
            /* Advance past the slash */
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
            CHECKSC("advance past /");
            
            EndElement = 1;
           
        } else if( IsASCII(*LookAhead, '?')) {
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
            CHECKSC("advance past ?");

            StartElement = 1; //This is to allow the attribute processing to take place.
            StartInstruction = 1; //This is to distinguish between attributes of an instruction and an element

            *pDepth |= ROOTFOLLOWS;
        } else if (IsASCII(*LookAhead, '!')) {
            UINT cmp, Length = 0;
            _TCHAR numbuf[12];
            BOOL IsBin = FALSE;

            /* Special command */
            /* <![CDATA[...]]>  <!DOCTYPE...> or
             * experimental <!BIN nnnnn[...]>
             */
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
            CHECKSC("Advance1");

            EatWHITESPACE(LookAhead);
            /* Parse the element name */

            CHECKSETSC(IsASCII(*LookAhead, '['), E_PARSE_ERROR,
                       "<! w/o [ unsupported");
            
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
            CHECKSC("Advance1");

            Tokenizer->v->SetIterator(Tokenizer, LookAhead, LookAhead2);
            
            sc = Tokenizer->v->NextToken(Tokenizer, LookAhead2, (PBYTE) " [", 2, &TokenLength);
            CHECKSC("next space or [");

            if (TokenLength == 3
                && (Tokenizer->v->CmpToken(Tokenizer, LookAhead,
                                           (PBYTE) "BIN", 3, &cmp) == S_OK))
                IsBin = TRUE;

            if (IsBin
                || (TokenLength == 5
                    && (Tokenizer->v->CmpToken(Tokenizer, LookAhead,
                                               (PBYTE)"CDATA", 5, &cmp) == S_OK)))
            {
                EatWHITESPACE(LookAhead2);
                Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);

                sc = Tokenizer->v->NextToken(Tokenizer, LookAhead2, (PBYTE) "[", 1, &TokenLength);
                CHECKSC("next [");
                if (IsBin) {
                    /* BIN has numeric arg, CDATA does not */
                    CHECKSETSC(TokenLength <= 10, E_PARSE_ERROR,
                               "number too long");
                
                    sc = Tokenizer->v->CopyToken(Tokenizer,
                                                 LookAhead, LookAhead2,
                                                 numbuf);
                    CHECKSC("copy numbuf");

                    Length = _ttoi(numbuf);
                    DBG(("SaxBIN Length=%d\n", Length));
                }

                EatWHITESPACE(LookAhead2);
                CHECKSETSC(IsASCII(*LookAhead2, '['), E_PARSE_ERROR,
                           "!BIN/!CDATA requires [");

                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
                CHECKSC("Advance1");

                Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);

                if (IsBin) {
                    sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, Length, NULL);
                    CHECKSC("AdvanceN");
                } else {
                    sc = Tokenizer->v->NextTokenSequence(Tokenizer,
                                                         LookAhead2, LookAhead3,
                                                         (PBYTE) "]]>", 3,
                                                         &TokenLength);
                    CHECKSC("find ]]>");
                    Length = TokenLength - 3;
                }
                /* Attachment here */
                Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);
                {
                    PIBUFFER b;
                    /* Carve out mbuf, call Attachment. */
                    sc = Tokenizer->v->CarveBuffer(Tokenizer, Content, Length,
                                                   &b);
                    CHECKSC("CarveBuffer");
                    assert(b);
                    sc = Handler->v->Text(Handler, NULL, 0, b, 0);
                    RELEASE(b);
                    CHECKSC("Handler->Attachment");
                }
#ifdef _UNICODE
                if (IsBin && Length % 1) {
                    /* Length is odd and this is Unicode, advance 1 padding */
                    sc = Tokenizer->v->AdvanceSymbol(Tokenizer,Content,1,NULL);
                    assert(sc != E_WOULD_BLOCK); /* XXX can happen fixit */
                    CHECKSC("Drop odd byte");
                }
#endif
                sc = Tokenizer->v->NextTokenSequence(Tokenizer,
                                                     Content, LookAhead3,
                                                     (PBYTE) "]]>", 3,
                                                     &TokenLength);
                assert(sc != E_WOULD_BLOCK); /* can't happen */
                CHECKSC("find ]]>");

                Tokenizer->v->SetIterator(Tokenizer, LookAhead3, Content);
                continue;
            } else
                CHECKSETSC(FALSE, E_PARSE_ERROR, "unsupported !extension");
        } else if( IsASCII(*LookAhead, '-')) {
            /* Looks like a comment */
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
            CHECKSC("advance past -");
            CHECKSETSC(IsASCII(*Content, '-'), E_PARSE_ERROR, "expected dash");
            /* Valid comment */
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
            CHECKSC("advance comment");

            /* Search for --> */
            do {
                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead,
                                                 1, NULL);
                CHECKSC("advance comment2");
                if (!IsASCII(*LookAhead, '-'))
                    continue;

                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead,
                                                 1, NULL);
                CHECKSC("advance comment3");
                if (!IsASCII(*LookAhead, '-'))
                    continue;

                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead,
                                                 1, NULL);
                CHECKSC("advance comment4");
                if (!IsASCII(*LookAhead, '>'))
                    return E_PARSE_ERROR;
            } while (TRUE);
            /* Comment processed, drop it */
            Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);
            continue;
        } else {
            /* Opening */
            StartElement = 1;
            *pDepth &= ~ROOTFOLLOWS;
        }

        EatWHITESPACE(LookAhead);
        /* Parse the element name */
        Tokenizer->v->SetIterator(Tokenizer, LookAhead, LookAhead2);

        sc = Tokenizer->v->AllocateAndCopyNextToken(Tokenizer, 
                                                    LookAhead,
                                                    LookAhead2, 
                                                    (const PBYTE )" \t\r\n/>",
                                                    6,
                                                    &ElementName,
                                                    &ElementNameLength);
        if (FAILED(sc) || sc == S_FALSE) goto err_eat_macro;

        EatWHITESPACE(LookAhead2);
        Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);

        tmp = _tcssplit(ElementName, _T(":"));
        if (tmp) {
            /* Empty prefix is not allowed */
            CHECKSETSC(_tcslen(ElementName), E_PARSE_ERROR,
                       "empty elem prefix w/ colon");
            Elem->Prefix = ElementName;
            Elem->Name = tmp;
        } else {
            Elem->Prefix = _T("");
            Elem->Name = ElementName;
        }
        Elem->Xns = NULL;

        /* Parse attributes if appropriate */
        if (StartElement) {
            do {
                sc = Tokenizer->v->AllocateAndCopyNextToken(Tokenizer, 
                                                            LookAhead,
                                                            LookAhead2, 
                                                            (const PBYTE) " =/>", 4,
                                                            &AttributeName,
                                                            &AttributeNameLength);
                CHECKSC("next  =/>");

                EatWHITESPACE(LookAhead2);
                Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);

                if (IsASCII(*LookAhead2, '=')) {
                    CHECKSETSC(AttrCount < MAXATTRIBUTES, E_OUT_OF_STRUCTURES,
                               "too many attributes");
                    Attr = &Attributes[AttrCount];

                    sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
                    CHECKSC("advance attr");

                    EatWHITESPACE(LookAhead2);

                    /* Be sure '"' follows */
                    CHECKSETSC(IsASCII(*LookAhead2, '"'), E_PARSE_ERROR,
                               "missing closing quote");
                    sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
                    
                    CHECKSC("advance quote");
                    Tokenizer->v->SetIterator(Tokenizer,
                                              LookAhead2, LookAhead);
                    
                    sc = Tokenizer->v->AllocateAndCopyNextToken(Tokenizer, 
                                                                LookAhead,
                                                                LookAhead2, 
                                                                (const PBYTE) "\"", 
                                                                1, 
                                                                &Token,
                                                                &TokenLength);
                    CHECKSC("next quote");

                    /* Advance past the attribute */
                    sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
                    CHECKSC("advance attr");

                    tmp = _tcssplit(AttributeName, _T(":"));
                    if (tmp) {
                        /* Empty prefix is not allowed */
                        CHECKSETSC(_tcslen(AttributeName), E_PARSE_ERROR,
                                   "empty attr prefix w/ colon");
                        Attr->Prefix = AttributeName;
                        Attr->Name = tmp;
                    } else {
                        Attr->Prefix = _T("");
                        Attr->Name = AttributeName;
                    }
                    Attr->Token = TOKEN_invalid;
                    Attr->Xns = NULL;

                    if (TokenLength > 0) {
                        Attr->Value = Token;
                    } else {
                        Attr->Value = _T("");
                        if (Token)
                            Heap->v->Free(Heap, 0, (PTR)Token);
                    }
                    Token = NULL;
                    AttrCount++;

                    AttributeName = NULL;
                    AttributeNameLength = TokenLength = 0;

                    EatWHITESPACE(LookAhead2);
                
                    Tokenizer->v->SetIterator(Tokenizer,
                                              LookAhead2, LookAhead);
                }
                if (AttributeName) {
                    Heap->v->Free(Heap, 0, (PTR) AttributeName);
                    AttributeName = NULL;
                    AttributeNameLength = 0;
                }
            } while (! IsASCII(*LookAhead2, '/')
                     && ! IsASCII(*LookAhead2, '>'));
            /* Pull in two bytes for /> so as to make sure we don't get
             * inconvenient E_WOULD_BLOCK after calling Start
             */
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 2, NULL);
            CHECKSC("advance close 2");

            /* We are committed regardless of restarts */
            if (!EndElement && !EndInstruction)
                ++*pDepth;

            if (StartInstruction)
                sc = Handler->v->StartInstruction(Handler, Elem,
                                                  Attributes, AttrCount);
            else
                sc = Handler->v->StartElement(Handler, Elem,
                                              Attributes, AttrCount);
            CHECKSC("Start");
        } else {
            /* Pull in one byte for > so as to make sure we don't get
             * inconvenient E_WOULD_BLOCK before calling End
             */
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
            CHECKSC("advance close 1");

            /* We are committed regardless of restarts */
            --*pDepth;
        }

        /* The content pointer points at either the element right bracket, or a slash */

        if (IsASCII(*LookAhead, '/')) {
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
            assert(sc != E_WOULD_BLOCK); /* can't happen, see above */
            CHECKSC("advance / 2");
            CHECKSETSC(IsASCII(*LookAhead, '>'), E_PARSE_ERROR, "missing >");
            CHECKSETSC(EndElement == 0, E_PARSE_ERROR, "unexpected slash");
            EndElement = 1;
            Elem->Token = TOKEN_invalid;
            --*pDepth;
        } else if(StartInstruction) {
            if (IsASCII(*Content, '?')) {
                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead,
                                                 1, NULL);
                assert(sc != E_WOULD_BLOCK); /* can't happen, see above */
                CHECKSC("advance ? 2");
    
                CHECKSETSC(IsASCII(*LookAhead, '>'), E_PARSE_ERROR,
                           "expected >");
    
                StartInstruction = 0;
                EndInstruction = 1;
                StartElement = 0;
                EndElement = 0;
            } else {
                StartInstruction = 0;
                EndInstruction = 1;
                StartElement = 0;
                EndElement = 0;
            }
            --*pDepth;
        }

        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
        assert(sc != E_WOULD_BLOCK); /* can't happen, see above */
        if (FAILED(sc) && !(sc == E_END_OF_DATA && *pDepth == STARTED))
            goto err_eat_macro;

        if (EndElement) {
            sc = Handler->v->EndElement(Handler, Elem);
            CHECKSC("EndElement");
        }
        if (EndInstruction) {
            sc = Handler->v->EndInstruction(Handler, Elem);
            CHECKSC("EndInstruction");
        }
        if (ElementName) { 
            //Heap->v->Free(Heap, 0, (PTR)ElementName);
            ElementName = NULL;
            ElementNameLength = 0;
        }
        /* Done with this element, drop data */
        Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);
    } while (*pDepth != STARTED);

    /* be sure ElementDepth == 0 */
    CHECKSETSC(*pDepth == STARTED, E_PARSE_ERROR, "incomplete document");

    sc = Handler->v->EndDocument(Handler);
    *pDepth = 0;

#define err_eat_macro zzz
    /* Get rid of any trailing newlines that might be present in the buffer */
    EatWHITESPACE(Content);

    /* Get rid of any trailing newlines that might be present in the buffer */
    EatWHITESPACE(Content);
err_eat_macro:
    sc = S_OK;                  /* ignore errors at tail */
#undef err_eat_macro

err_eat_macro:
METHOD_CLEANUP;
//assert(sc == S_OK || sc == E_WOULD_BLOCK); 
    if (Elem->Name) {
        /* Clear element and attribute table from previous round */
        Heap->v->Free(Heap, 0, _tcslen(Elem->Prefix) ? (PTR)Elem->Prefix
                                                    : (PTR)Elem->Name);
        for (i = 0, Attr = &Attributes[0]; i < AttrCount; i++, Attr++) {
            if (_tcslen(Attr->Prefix))
                Heap->v->Free(Heap, 0, (PTR) Attr->Prefix);
            else if (Attr->Name)
                Heap->v->Free(Heap, 0, (PTR) Attr->Name);
            if (_tcslen(Attr->Value))
                Heap->v->Free(Heap, 0, (PTR) Attr->Value);
        }
        AttrCount = 0;
        ElementName = NULL;
        ElementNameLength = 0;
        Elem->Name = NULL;
        Elem->Token = TOKEN_invalid;
        Elem->Xns = NULL;
    }
    if (ElementName) { 
        Heap->v->Free(Heap, 0, (PTR)ElementName);
        ElementName = NULL;
    }
    if (Token) { 
        Heap->v->Free(Heap, 0, (PTR)Token);
        Token = NULL;
    }
    if (AttributeName) {
        Heap->v->Free(Heap, 0, (PTR)AttributeName);
        AttributeName = NULL;
    }

    return sc;
}

#ifdef _MINIMIZE

PRIVATE SCODE SaxParserRead(PISAXPARSER Parser, PITOKENIZER Tokenizer,
                            UINT* Type, PSAXATTRIBUTE* Elem, PSAXATTRIBUTE* Attributes, 
                            UINT* AttributeCount, _TCHAR** Text, UINT* Length,
                            PIBUFFER* Buffer)
{
    return S_OK;
}

#else

PRIVATE SCODE SaxParserRead(PISAXPARSER Parser, PITOKENIZER Tokenizer,
                            UINT* Type, PSAXATTRIBUTE* Elem, PSAXATTRIBUTE* Attributes, 
                            UINT* AttributeCount, _TCHAR** Text, UINT* Length,
                            PIBUFFER* Buffer)
{
    METHOD_DECLARE_NOTHIS(SaxParser, Read);
    PBYTE*  Content   = Tokenizer->v->GetContent(Tokenizer);
    PBYTE*  LookAhead = Tokenizer->v->GetLookAhead(Tokenizer);
    PBYTE*  LookAhead2 = Tokenizer->v->GetLookAhead2(Tokenizer);
    PBYTE*  LookAhead3 = Tokenizer->v->GetLookAhead3(Tokenizer);

    _TCHAR *Token = NULL;
    UINT TokenLength = 0;
    _TCHAR *ElementName = NULL;
    UINT ElementNameLength = 0;
    _TCHAR *AttributeName = NULL;
    UINT AttributeNameLength = 0;
    _TCHAR *tmp;
    UINT i, AttrCount = 0;
    UINT StartElement = 0;
    UINT EndElement = 0;
    UINT StartInstruction = 0;
    UINT EndInstruction = 0;
    PIHEAP  Heap = Tokenizer->v->GetHeap(Tokenizer);
#define MAXATTRIBUTES 15
    SAXATTRIBUTE AttributesTmp[MAXATTRIBUTES], *Attr;

    sc = S_OK;

    EatWHITESPACE(Content);

    /* Deal with characters between the current location and the next element */
    Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead);
    sc = Tokenizer->v->AllocateAndCopyNextToken(Tokenizer, 
                                                Content,
                                                LookAhead, 
                                                (PBYTE) "<", 1, 
                                                &Token,
                                                &TokenLength);
    CHECKSC("next <");

    EatWHITESPACE(LookAhead);

    if (TokenLength > 0) {
        /* Strip trailing whitespace */
        while (TokenLength > 0 && IsWhitespace(&Token[TokenLength-1])) {
            Token[TokenLength-1] = _EOS; /* For DeserializeText simple list */
            TokenLength--;
        }
        *Type = TYPE_TEXT;
        *Text = Token;
        *Length = TokenLength;
        Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);
        goto Out;
    }
    if (Token) {
        Heap->v->Free(Heap, 0, (PTR) Token);
        Token = NULL;
        TokenLength = 0;
    }

    /* Locate the left bracket */ 
    Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);

    /* Advance past the left bracket */
    sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
    CHECKSC("Advance past <");

    /* Closing or opening? */
    if (IsASCII(*LookAhead, '/')) {
        /* Closing */
        /* Advance past the slash */
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
        CHECKSC("advance past /");
         
        EndElement = 1;
    } else if( IsASCII(*LookAhead, '?')) {
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
        CHECKSC("advance past ?");

        StartElement = 1; //This is to allow the attribute processing to take place.
        StartInstruction = 1; //This is to distinguish between attributes of an instruction and an element
    } else if (IsASCII(*LookAhead, '!')) {
        UINT cmp, Len = 0;
        _TCHAR numbuf[12];
        BOOL IsBin = FALSE;

        /* Special command */
        /* <![CDATA[...]]>  <!DOCTYPE...> or
         * experimental <!BIN nnnnn[...]>
         */
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
        CHECKSC("Advance1");

        EatWHITESPACE(LookAhead);
        /* Parse the element name */

        CHECKSETSC(IsASCII(*LookAhead, '['), E_PARSE_ERROR, "<! w/o [ unsupported");
            
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
        CHECKSC("Advance1");

        Tokenizer->v->SetIterator(Tokenizer, LookAhead, LookAhead2);
            
        sc = Tokenizer->v->NextToken(Tokenizer, LookAhead2, (PBYTE) " [", 2, &TokenLength);
        CHECKSC("next space or [");

        if (TokenLength == 3
            && (Tokenizer->v->CmpToken(Tokenizer, LookAhead,
                                       (PBYTE) "BIN", 3, &cmp) == S_OK))
            IsBin = TRUE;

        if (IsBin
            || (TokenLength == 5
                && (Tokenizer->v->CmpToken(Tokenizer, LookAhead,
                                           (PBYTE)"CDATA", 5, &cmp) == S_OK)))
        {
            EatWHITESPACE(LookAhead2);
            Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);

            sc = Tokenizer->v->NextToken(Tokenizer, LookAhead2, (PBYTE) "[", 1, &TokenLength);
            CHECKSC("next [");
            if (IsBin) {
                /* BIN has numeric arg, CDATA does not */
                CHECKSETSC(TokenLength <= 10, E_PARSE_ERROR,"number too long");
                
                sc = Tokenizer->v->CopyToken(Tokenizer,
                                             LookAhead, LookAhead2,
                                             numbuf);
                CHECKSC("copy numbuf");

                Len = _ttoi(numbuf);
                DBG(("SaxBIN Length=%d\n", Len));
            }

            EatWHITESPACE(LookAhead2);
            CHECKSETSC(IsASCII(*LookAhead2, '['), E_PARSE_ERROR,
                       "!BIN/!CDATA requires [");

            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
            CHECKSC("Advance1");

            Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);

            if (IsBin) {
                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, Len, NULL);
                CHECKSC("AdvanceN");
            } else {
                sc = Tokenizer->v->NextTokenSequence(Tokenizer,
                                                     LookAhead2, LookAhead3,
                                                     (PBYTE) "]]>", 3,
                                                     &TokenLength);
                CHECKSC("find ]]>");
                Len = TokenLength - 3;
            }
            /* Attachment here */
            Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);
            {
                //PIBUFFER b;
                /* Carve out mbuf, call Attachment. */
                sc = Tokenizer->v->CarveBuffer(Tokenizer, Content, Len, Buffer); //&b);
                CHECKSC("CarveBuffer");
                assert(*Buffer);//assert(b);
                *Type = TYPE_TEXT_CDATA; // Text in IBuffer
            }
#ifdef _UNICODE
            if (IsBin && Len % 1) {
                /* Length is odd and this is Unicode, advance 1 padding */
                sc = Tokenizer->v->AdvanceSymbol(Tokenizer,Content,1,NULL);
                assert(sc != E_WOULD_BLOCK); /* XXX can happen fixit */
                CHECKSC("Drop odd byte");
            }
#endif
            sc = Tokenizer->v->NextTokenSequence(Tokenizer,
                                                 Content, LookAhead3,
                                                 (PBYTE) "]]>", 3,
                                                 &TokenLength);
            assert(sc != E_WOULD_BLOCK); /* can't happen */
            CHECKSC("find ]]>");

            Tokenizer->v->SetIterator(Tokenizer, LookAhead3, Content);
            goto Out;     
        } else
            CHECKSETSC(FALSE, E_PARSE_ERROR, "unsupported !extension");
    } else if( IsASCII(*LookAhead, '-')) {
        /* Looks like a comment */
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
        CHECKSC("advance past -");
        CHECKSETSC(IsASCII(*Content, '-'), E_PARSE_ERROR, "expected dash");
        /* Valid comment */
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
        CHECKSC("advance comment");

        /* Search for --> */
        do {
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
            CHECKSC("advance comment2");
            if (!IsASCII(*LookAhead, '-'))
                continue;

            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
            CHECKSC("advance comment3");
            if (!IsASCII(*LookAhead, '-'))
                continue;

            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
            CHECKSC("advance comment4");
            if (!IsASCII(*LookAhead, '>'))
                return E_PARSE_ERROR;
        } while (TRUE);
        /* Comment processed, drop it */
        Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);
        *Type = TYPE_COMMENT;
        goto Out;
    } else {
        /* Opening */
        StartElement = 1;
    }

    EatWHITESPACE(LookAhead);
    /* Parse the element name */
    Tokenizer->v->SetIterator(Tokenizer, LookAhead, LookAhead2);

    sc = Tokenizer->v->AllocateAndCopyNextToken(Tokenizer, 
                                                LookAhead,
                                                LookAhead2, 
                                                (const PBYTE )" \t\r\n/>",
                                                6,
                                                &ElementName,
                                                &ElementNameLength);
    if (FAILED(sc) || sc == S_FALSE) goto err_eat_macro;

    *Elem = (PSAXATTRIBUTE) Heap->v->Alloc(Heap, 
                                           HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("sax"),
                                           sizeof(SAXATTRIBUTE), 0);
    EatWHITESPACE(LookAhead2);
    Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);

    tmp = _tcssplit(ElementName, _T(":"));
    if (tmp) {
        /* Empty prefix is not allowed */
        CHECKSETSC(_tcslen(ElementName), E_PARSE_ERROR,
                   "empty elem prefix w/ colon");
        (*Elem)->Prefix = ElementName;
        (*Elem)->Name = tmp;
    } else {
        (*Elem)->Prefix = _T("");
        (*Elem)->Name = ElementName;
    }
    (*Elem)->Xns = NULL;

    /* Parse attributes if appropriate */
    if (StartElement) {
        do {
            sc = Tokenizer->v->AllocateAndCopyNextToken(Tokenizer, 
                                                        LookAhead,
                                                        LookAhead2, 
                                                        (const PBYTE) " =/>", 4,
                                                        &AttributeName,
                                                        &AttributeNameLength);
            CHECKSC("next  =/>");

            EatWHITESPACE(LookAhead2);
            Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);

            if (IsASCII(*LookAhead2, '=')) {
                    CHECKSETSC(AttrCount < MAXATTRIBUTES, E_OUT_OF_STRUCTURES,
                               "too many attributes");
                Attr = &AttributesTmp[AttrCount];
                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
                CHECKSC("advance attr");

                EatWHITESPACE(LookAhead2);
                /* Be sure '"' follows */
                CHECKSETSC(IsASCII(*LookAhead2, '"'), E_PARSE_ERROR,
                           "missing closing quote");
                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
                    
                CHECKSC("advance quote");
                Tokenizer->v->SetIterator(Tokenizer,
                                          LookAhead2, LookAhead);
                    
                sc = Tokenizer->v->AllocateAndCopyNextToken(Tokenizer, 
                                                            LookAhead,
                                                            LookAhead2, 
                                                            (const PBYTE) "\"", 
                                                            1, 
                                                            &Token,
                                                            &TokenLength);
                CHECKSC("next quote");

                /* Advance past the attribute */
                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
                CHECKSC("advance attr");

                tmp = _tcssplit(AttributeName, _T(":"));
                if (tmp) {
                    /* Empty prefix is not allowed */
                    CHECKSETSC(_tcslen(AttributeName), E_PARSE_ERROR,
                               "empty attr prefix w/ colon");
                    Attr->Prefix = AttributeName;
                    Attr->Name = tmp;
                } else {
                    Attr->Prefix = _T("");
                    Attr->Name = AttributeName;
                }
                Attr->Token = TOKEN_invalid;
                Attr->Xns = NULL;

                if (TokenLength > 0) {
                    Attr->Value = Token;
                } else {
                    Attr->Value = _T("");
                    if (Token)
                        Heap->v->Free(Heap, 0, (PTR)Token);
                }
                Token = NULL;
                AttrCount++;
                AttributeName = NULL;
                AttributeNameLength = TokenLength = 0;

                EatWHITESPACE(LookAhead2);
                Tokenizer->v->SetIterator(Tokenizer, LookAhead2, LookAhead);
            }
            if (AttributeName) {
                Heap->v->Free(Heap, 0, (PTR) AttributeName);
                AttributeName = NULL;
                AttributeNameLength = 0;
            }
        } while (! IsASCII(*LookAhead2, '/')
                 && ! IsASCII(*LookAhead2, '>'));
        /* Pull in two bytes for /> so as to make sure we don't get
         * inconvenient E_WOULD_BLOCK after calling Start
         */
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 2, NULL);
        CHECKSC("advance close 2");

        /* Copy the AttributesTmp to Attributes */
        *Attributes = (PSAXATTRIBUTE) Heap->v->Alloc(Heap, 
                                          HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("sax"),
                                          sizeof(SAXATTRIBUTE) * AttrCount, 0);
        for (i = 0; i < AttrCount; i++) {
            (*Attributes[i]).Prefix = AttributesTmp[i].Prefix;
            (*Attributes[i]).Name = AttributesTmp[i].Name;
            (*Attributes[i]).Value = AttributesTmp[i].Value;
            (*Attributes[i]).Token = TOKEN_invalid;
            (*Attributes[i]).Xns = NULL;
        }
        *AttributeCount = AttrCount;
    } else {
        /* Pull in one byte for > so as to make sure we don't get
         * inconvenient E_WOULD_BLOCK before calling End
         */
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead2, 1, NULL);
        CHECKSC("advance close 1");
    }

    /* The content pointer points at either the element right bracket, or a slash */

    if (IsASCII(*LookAhead, '/')) {
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
        assert(sc != E_WOULD_BLOCK); /* can't happen, see above */
        CHECKSC("advance / 2");
        CHECKSETSC(IsASCII(*LookAhead, '>'), E_PARSE_ERROR, "missing >");
        CHECKSETSC(EndElement == 0, E_PARSE_ERROR, "unexpected slash");
        EndElement = 1;
        (*Elem)->Token = TOKEN_invalid;
    } else if(StartInstruction) {
        if (IsASCII(*Content, '?')) {
            sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead,
                                             1, NULL);
            assert(sc != E_WOULD_BLOCK); /* can't happen, see above */
            CHECKSC("advance ? 2");
            CHECKSETSC(IsASCII(*LookAhead, '>'), E_PARSE_ERROR, "expected >");
        }
//        StartInstruction = 0;
        EndInstruction = 1;
        StartElement = 0;
        EndElement = 0;
    }

    sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
    assert(sc != E_WOULD_BLOCK); /* can't happen, see above */
    if (FAILED(sc) && !(sc == E_END_OF_DATA))
        goto err_eat_macro;

    if (EndInstruction) {
        *Type = TYPE_INSTRUCTION;
    } else {
        if (EndElement) {
            if (StartElement) 
                *Type = TYPE_EMPTY_ELEMENT;
            else 
                *Type = TYPE_END_ELEMENT;
        } else {
            *Type = TYPE_START_ELEMENT;
        }
    }

    /* Done with this element, drop data */
    Tokenizer->v->SetIterator(Tokenizer, LookAhead, Content);

#define err_eat_macro zzz
    /* Get rid of any trailing newlines that might be present in the buffer */
    EatWHITESPACE(Content);

err_eat_macro:
    sc = S_OK;                  /* ignore errors at tail */
#undef err_eat_macro

err_eat_macro:
METHOD_CLEANUP;
//assert(sc == S_OK || sc == E_WOULD_BLOCK); 
    return sc;
}

#endif // _MINIMIZE

/* Override default name for CreatePrinter in v-table */
#define SaxParserCreatePrinter SaxCreatePrinter
#include <s_saxparser_v.c>

PRIVATE struct _SAXPARSER SaxParser = {&SaxParserVtbl, 1};

/* SAX parser */
PIUNKNOWN saxCobMain(void)
{
    return (PIUNKNOWN) &SaxParser;
}
