/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* XML Printer */
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <string.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <netif.h>
#include <web/typeinfo.h>

#define pPrinter(_i_) ((SaxPrinter *)(_i_))
#define iPrinter(_p_) ((PISAXHANDLER)(_p_))

enum TOKEN_TYPE {
    NONE = 0,
    ELEMENT_START,
    INSTRUCTION_START,
};

typedef struct _SaxPrinter SaxPrinter;
struct _SaxPrinter {
    struct ISaxHandlerVtbl const *   v;
    UINT   Refs;

    PIFILE  OutputFile;
    UINT previousTokenType;
};

#define w(_s_) \
    MACRO_BEGIN \
    const char *s=(_s_);\
    sc = ReadWriteFile(File, TRUE, UINT64_MAX, (PBYTE) s, strlen(s), NULL); \
    CHECKSC("Write"); \
    MACRO_END


#define wt(_s_) \
    MACRO_BEGIN \
    const _TCHAR *s=(_s_);\
    BYTE c; \
    for (; *s; s++) { \
      c = (BYTE) *s; /* XXX UTF8 */ \
      sc = ReadWriteFile(File, TRUE, UINT64_MAX, &c, 1, NULL); \
      CHECKSC("Write"); \
    } \
    MACRO_END

#define wtn(_s_, _l_) \
    MACRO_BEGIN \
    const _TCHAR *s=(_s_);\
    BYTE c; \
    for (; *s && _l_ > 0; s++, _l_--) { \
      c = (BYTE) *s; /* XXX UTF8 */ \
      sc = ReadWriteFile(File, TRUE, UINT64_MAX, &c, 1, NULL); \
      CHECKSC("Write"); \
    } \
    MACRO_END

/* Helper */
PRIVATE SCODE SaxPrinterStart(PISAXHANDLER iThis, BOOL IsInstruction,
                              PSAXATTRIBUTE Elem,
                              PSAXATTRIBUTE Attr, UINT Count)

{
    METHOD_DECLARE(SaxPrinter, Start);
    UINT i;
    const _TCHAR *Prefix, *Name;
    PIFILE File = This->OutputFile;

    if (Elem->Prefix)
        Prefix = Elem->Prefix;
    else if (Elem->Xns)
        Prefix = Elem->Xns->name;
    else
        Prefix = _T("");

    Name = Elem->Name;

    if (This->previousTokenType == ELEMENT_START)
        w(">");
    if (This->previousTokenType == INSTRUCTION_START)
        w("?>");

    if (IsInstruction) {
        w("<?");
        This->previousTokenType = INSTRUCTION_START;
    } else {
        w("<");
        This->previousTokenType = ELEMENT_START;
    }

    if (*Prefix > 0) {
        wt(Prefix);
        w(":");
    }

    wt(Name);

    if (FAILED(sc))
        return sc;
    for (i = 0; i < Count; i++, Attr++) {
        UINT ValueLength = _tcslen(Attr->Value);

        if (Attr->Xns == XMLNS_DECLARE)
            Prefix = _T("xmlns");
        else if (Attr->Prefix)
            Prefix = Attr->Prefix;
        else if (Attr->Xns)
            Prefix = Attr->Xns->name;
        else
            Prefix = _T("");

        w(" ");
        if (*Prefix > 0) {
            wt(Prefix);
            /* Special case for default ns declaration: no colon */
            if (Attr->Xns != XMLNS_DECLARE || Attr->Name[0] != _EOS)
                w(":");
        }

        wt(Attr->Name);
        w("=\"");
    
        wt(Attr->Value);
        w("\"");
    }
    METHOD_CLEANUP;

    return S_OK;
}

/***************** Methods **********************/
PRIVATE SCODE MCT
SaxPrinterQueryInterface(PISAXHANDLER This, REFIID IidQuery, void** Object)
{
    return GenericQueryInterface((PIUNKNOWN) This, IidQuery, Object,
                                 &IID_ISaxHandler);
}


PRIVATE UINT MCT
SaxPrinterAddRef(PISAXHANDLER iThis)
{
    SaxPrinter *This = (SaxPrinter *) iThis;
    return AtomicInc(&This->Refs);
}

PRIVATE UINT MCT 
SaxPrinterRelease(PISAXHANDLER iThis)
{
    UINT Refs;
    
    SaxPrinter *This = pPrinter(iThis) ;
    PIHEAP Heap = CurrentHeap();
    
    TRACE("saxprinter.c function : SaxPrinterRelease");
    
    Refs = AtomicDec(&This->Refs);
    if (Refs == 0) {
        This->OutputFile->v->Release(This->OutputFile);
        //Release any memory associated with the structure pointers;
        Heap->v->Free(Heap, 0, (PTR) This);
    }
    return Refs;
}

PRIVATE SCODE SaxPrinterStartDocument(PISAXHANDLER This)
{
    SaxPrinter * x = pPrinter(This);
    UnusedParameter(x);
    TRACE("saxprinter.c function : SaxPrinterStartDocument");
    return S_OK;
}

PRIVATE SCODE SaxPrinterEndDocument(PISAXHANDLER This)
{
    SaxPrinter * x = pPrinter(This);
    TRACE("saxprinter.c function : SaxPrinterEndDocument");

    /* XXX SAXParser expects at least one more char after /soap:envelope> */
    return ReadWriteFile(x->OutputFile, TRUE, UINT64_MAX,
                         (PUINT8) "\r\n", 2, NULL);
}

PRIVATE SCODE SaxPrinterStartElement(PISAXHANDLER Sax, PSAXATTRIBUTE Elem,
                                     PSAXATTRIBUTE Attr, UINT Count)

{
    return SaxPrinterStart(Sax, FALSE, Elem, Attr, Count);
}

PRIVATE SCODE SaxPrinterEndElement(PISAXHANDLER iThis, PSAXATTRIBUTE Elem)
{
    METHOD_DECLARE(SaxPrinter, EndElement);
    const _TCHAR *Prefix;
    PIFILE File = This->OutputFile;

    //TRACE("saxprinter.c function : SaxPrinterEndElement");

    if (Elem->Prefix)
        Prefix = Elem->Prefix;
    else if (Elem->Xns)
        Prefix = Elem->Xns->name;
    else
        Prefix = _T("");

    if (This->previousTokenType == ELEMENT_START) {
        w("/>");
    } else {
        w("</");
        if (*Prefix) {
            wt(Prefix);
            w(":");
        }
        wt(Elem->Name);
        w(">");
    }

    This->previousTokenType = NONE;
    METHOD_CLEANUP;
    return S_OK;
}

PRIVATE SCODE
SaxPrinterAttachment(PISAXHANDLER iThis, PIBUFFER Content, UINT Flags)
{
    METHOD_DECLARE(SaxPrinter, Attachment);
    PIFILE File = This->OutputFile;
    UINT count;
    UINT64 Size64;
    char intbuf[12];
    PIBUFFER Output;

    if (This->previousTokenType == ELEMENT_START)
        w(">");
    else if (This->previousTokenType == INSTRUCTION_START)
        w("?>");
        
    This->previousTokenType = NONE;

    Flags &= ARGUMENT_DESCR_ENCODING_MASK;
    if (Flags == ARGUMENT_DESCR_ENC_CDATA)
        w("<![CDATA[");
    else if (Flags == ARGUMENT_DESCR_ENC_BIN) {
        w("<![BIN ");
        sc = Content->v->GetSize(Content, &Size64);
        CHECKSC("GetSize");
        CHECKSETSC(Uint64High32Bits(Size64) == 0, E_INVALID_PARAMETER,
                   "buffer is ridiculously large");
        count = Uint64Low32Bits(Size64);
        sprintf(intbuf, "%d", count);
        w(intbuf);
        w(" [");
    }

    /* optimize: if File is also IBuffer, use Append, if ep, XXX SendBuffer */
    sc = File->v->QueryInterface(File, &IID_IBuffer, (void **) &Output);
    if (SUCCEEDED(sc)) {
        sc = Output->v->Append(Output, Content);
        CHECKSC("Output->Append");
    } else {
        sc = Content->v->ReadWriteFile(Content, File, TRUE,
                                       UINT64_MAX, 0, UINT_MAX, &count);
        CHECKSC("Content->ReadWriteFile");
    }

    if (Flags == ARGUMENT_DESCR_ENC_CDATA || Flags == ARGUMENT_DESCR_ENC_BIN)
        w("]]>");

    METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE
SaxPrinterText(PISAXHANDLER iThis, const _TCHAR *Text, UINT Length,
               PIBUFFER Content, UINT Flags)
{
    METHOD_DECLARE(SaxPrinter, Text);
    PIFILE File = This->OutputFile;

    sc = S_OK;
    if (Content)
        return SaxPrinterAttachment(iThis, Content, Flags);

    if (!Text || 0 == Length) {
        assert(0 == Length);
        return sc;
    }

    if (This->previousTokenType == ELEMENT_START)
        w(">");
    else if (This->previousTokenType == INSTRUCTION_START)
        w("?>");
        
    This->previousTokenType = NONE;

    wtn(Text, Length);
METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE SaxPrinterStartInstruction(PISAXHANDLER Sax, PSAXATTRIBUTE Elem,
                                         PSAXATTRIBUTE Attr, UINT Count)

{
    return SaxPrinterStart(Sax, TRUE, Elem, Attr, Count);
}

PRIVATE SCODE SaxPrinterEndInstruction(PISAXHANDLER This, PSAXATTRIBUTE Elem)
{
    SaxPrinter * x = pPrinter(This);

    /* Instructions cannot be nested */
    ReadWriteFile(x->OutputFile, TRUE, UINT64_MAX, (PUINT8) "?>", 2, NULL);

    x->previousTokenType = NONE;
    return S_OK;
}

PRIVATE SCODE SaxPrinterReset(PISAXHANDLER This, PIFILE file)
{
    SaxPrinter * x = pPrinter(This);

    x->previousTokenType = NONE;

    if (file) {
        x->OutputFile->v->Release(x->OutputFile);
        file->v->AddRef(file);
        x->OutputFile = file;
    }
    return x->OutputFile->v->SetSize(x->OutputFile, UINT64_ZERO);
}

PRIVATE SCODE SaxPrinterGetContent(PISAXHANDLER This, PIFILE *pFile)
{
    SaxPrinter * x = pPrinter(This);

    x->OutputFile->v->AddRef(x->OutputFile);
    *pFile = x->OutputFile;
    return S_OK;
}

#include "s_saxprinter_v.c"

/**** Constructor  ****/
SCODE MCT SaxCreatePrinter(PISAXPARSER Parser, PIFILE OutputFile,
                           UINT IndentLevel, PISAXHANDLER *pPrinter)
{
    SaxPrinter *   x = NULL;
    PIHEAP          Heap = CurrentHeap();
    
    UnusedParameter(Parser);
    UnusedParameter(IndentLevel);

    TRACE("saxprinter.c function: SAXPrinterCallBacksCreate");

    x = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("SaxPrinter"), sizeof(*x), 0);
    if (x == NULL) return E_NOT_ENOUGH_MEMORY;

    x->v     = &SaxPrinterVtbl;
    x->Refs  = 1;

    OutputFile->v->AddRef(OutputFile);
    x->OutputFile = OutputFile;

    /* Reset */
    x->previousTokenType = NONE;
    OutputFile->v->SetSize(OutputFile, UINT64_ZERO);

    *pPrinter = (PISAXHANDLER) x;
    return S_OK;
}
