/* Copyright (c) Microsoft Corporation. All rights reserved. */

/*
 * Test for the EB63 board buttons SW1..SW4
 */

#ifdef arm

#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>

#include <arm/at91m63200.h>
#include <drivers/drivers.h>

#include "sensor.h"

#ifdef _MSC_VER
#pragma warning(disable:4054)   /* cast from function ptr to data ptr (/Za) */
#endif

#if 0
#if (PLATFORM==2)
#define BUTTS 0xf0180
#define BUTT2 0x80
#else // 2B
#define BUTTS 0xf0140
#define BUTT2 0x40
#endif
#endif


BOOL TimeToDie = FALSE;

/* The first three buttons are multiplexed. 
 * Increase nBUTTONS to 4 to include the 4th button, 
 * or decrease to 3 to use only the multiplexed ones.
 * The 4th button only reports button press, the first 
 * three also report release. 
 */
#define nBUTTONS 3
struct _BUTTONS {
    CONDITION ButtonPress;
    UINT ButtonsChanged;
    struct _BUTTON_STATE {
        TIME LastOnAt;
        TIME LastOffAt;
        BOOL On;
    } ButtonState[nBUTTONS];
} TheButtons
  = {CONDITION_INITIALIZER, 0,};

UINT nInterrupts = 0;
UINT nDebounced = 0;
TIME t1,t2,t3;

/* ISR for Buttons with a direct IRQ.
 */
BOOL MCT ButtonIsr(void *This, PBOOL pNotMyInterrupt)
{
    UINT ButtonNo = (UINT)This;
    TIME Now = CurrentTime(), Last;
    static const TIME DEBOUNCE_TIME = Int64Initializer(0,TIME_MILLIS(10));

    UnusedParameter(pNotMyInterrupt);

    nInterrupts++;

    /* Debounce the hw */
    Last = (TheButtons.ButtonState[ButtonNo].On) ?
        TheButtons.ButtonState[ButtonNo].LastOnAt :
        TheButtons.ButtonState[ButtonNo].LastOffAt;
    t1 = Now;
    t2 = Last;
    Last = TimeSubtract(Now,Last);
    t3 = Last;
    if (TimeLess(Last,DEBOUNCE_TIME)) {
        nDebounced++;// = Int64ToInt32(Last);
        return FALSE;
    }

    /* We are the Isr, so we are atomic */
    TheButtons.ButtonsChanged |= 1<<ButtonNo;

    /* Record the event time */
    if (TheButtons.ButtonState[ButtonNo].On)
        TheButtons.ButtonState[ButtonNo].LastOffAt = Now;
    else
        TheButtons.ButtonState[ButtonNo].LastOnAt = Now;

    /* Flip the state bit */
    TheButtons.ButtonState[ButtonNo].On = !TheButtons.ButtonState[ButtonNo].On;

    /* Wake up the thread, if necessary */
    return Condition_InterruptSignal(&TheButtons.ButtonPress);
}

/* ISR for Buttons that are multiplexed
 */
BOOL MCT PioIsr(void *This, PBOOL pNotMyInterrupt)
{
    struct _Pio *Pio = (struct _Pio *)This;
    BOOL rv;
    UINT Pins;

    /* See what ISR-enabled button was pressed
     * NB: This clears the interrupt also.
     */
    Pins = Pio->IntrStatus & Pio->IntrMask;

    /* sanity check
     */
    if (Pins == 0) {
        *pNotMyInterrupt = TRUE;
        return FALSE;
    }

    /* There might be more than one pending, dispatch all we got
     */
    rv = FALSE;

    if (Pins & PIOB_PB3) {
        rv = ButtonIsr((void*)0,pNotMyInterrupt);
    }
    if (Pins & PIOB_PB4) {
        rv |= ButtonIsr((void*)1,pNotMyInterrupt);
    }
    if (Pins & PIOB_PB5) {
        rv |= ButtonIsr((void*)2,pNotMyInterrupt);
    }

    return rv;
}

/* This thread does all the non-ISR work
 * BUGBUG Use constraints to keep latency bounded.
 */
#define nLEDS 8
void THREAD_LINKAGE ButtonThread(THREAD_ARGUMENT arg)
{
    PSENSORDEMO demo = (PSENSORDEMO)arg;
    struct _BUTTONS *Butts = &TheButtons;
    UINT LastSeen = 0, ButtonNo, LedIdx = 0;
    TIME_CONSTRAINT TimeConstraint;

    _tprintf(_TEXT("ButtonThread started (%x)\r\n"),(ADDRESS)Butts);
    
    /* Construct a very urgent TimeConstraint and do an
     * initial BeginConstraint
     */
    Int32ToInt64(TimeConstraint.Start, 0);/* can start immediately after wait*/
    Int32ToInt64(TimeConstraint.Estimate, 0); /* does not really matter */
    Int32ToInt64(TimeConstraint.Deadline,0); /* immediately if not sooner */
    TimeConstraint.Criticality = TRUE;
    (void) BeginConstraint(FALSE, &TimeConstraint, NULL);

    for (;!TimeToDie;) {

        Int32ToInt64(TimeConstraint.Start, 0);
        Int32ToInt64(TimeConstraint.Estimate, 0);
        Int32ToInt64(TimeConstraint.Deadline,0);
        TimeConstraint.Criticality = TRUE;

        _tprintf(_TEXT("ButtonThread waiting (%x)\r\n"),(ADDRESS)Butts);
        (void) Condition_WaitAndBeginConstraint(&Butts->ButtonPress,
                                                NULL, NULL,
                                                TRUE, &TimeConstraint, NULL);
        _tprintf(_TEXT("ButtonThread started again (%x)\r\n"),(ADDRESS)Butts);

        /* Pick up what happened */
        do {
            LastSeen = Butts->ButtonsChanged;
        } while (!AtomicCmpAndSwap(&Butts->ButtonsChanged,
                                   LastSeen,
                                   0));
        /* Anything else.. later */

        printf("Pio: %x\r\n", PioB->PinDataStatus);

        Mutex_Lock(&demo->Mutex);

        for (ButtonNo = 0; ButtonNo < nBUTTONS; ButtonNo++) {
            if (LastSeen & (1<<ButtonNo)) {
                TIME tOn, tOff;
                tOn = Butts->ButtonState[ButtonNo].LastOnAt;
                tOff = Butts->ButtonState[ButtonNo].LastOffAt;
                if (Butts->ButtonState[ButtonNo].On) {
                    _tprintf(_TEXT("Button%d press at %lld00 (%d,%d)"),
                             1+ButtonNo,tOn,nInterrupts,nDebounced);
                    /* Toggle the bit */
                    demo->State ^= 1 << ButtonNo;
                } else {
                    _tprintf(_TEXT("Button%d release at %lld00 (%d,%d)"),
                             1+ButtonNo,tOff,nInterrupts,nDebounced);
                    /* Nevermind releases */
                }
                printf(" [%lld %lld %lld]\r\n",t1,t2,t3);
            }
        }
        printf(_TEXT("ButtonThread: demo->State=%x\n"), demo->State);

        /* Set the LEDs, for shits and giggles.
         * Leave the last LED alone, it indicates successful initialization. 
         * Only fiddle with the LED array if the buttons were actually used
         * (otherwise the button thread destroyes recorder LED settings...).
         */
        if (LastSeen) {
            for (LedIdx = 0; LedIdx < nLEDS - 1; ++LedIdx) {
                if (demo->State & (1 << LedIdx))
                    PioB->ClearData = PIOB_PB8 << LedIdx;
                else
                    PioB->SetData = PIOB_PB8 << LedIdx;
            }
        }

        Mutex_Unlock(&demo->Mutex);

#if 0
        /* Don't re-enable this code without removing the demo 
         * structure deallocation from the ButtonThread method exit. 
         * That line assumes the button thread only terminates when 
         * 0 == demo->Refs
         */
        /* On release of this button terminate the program */
#define DIE_BUTTON (3-1)
        if ((LastSeen & (1<<DIE_BUTTON)) &&
            (!Butts->ButtonState[DIE_BUTTON].On))
            TimeToDie = TRUE;
#endif

    }

    /* Disable button interrupts */
#ifdef arm
    _tprintf(_TEXT("Disabling Button interrupts (%x)\r\n"),(ADDRESS)Butts);
    RemoveDevice((PTR)3,31);
    PioB->IntrDisable = PIOB_PB3 | PIOB_PB4 | PIOB_PB5;
    RemoveDevice((PTR)PioB,14);
#endif
    _tprintf(_TEXT("ButtonThread terminated (%x)\r\n"),(ADDRESS)Butts);

    /* Finish clean-up of component structure. 
     * XXX Nasty - assumes current heap and 0 == demo->Refs 
     */
    Mutex_Destroy(&demo->Mutex);
    CurrentHeap()->v->Free(CurrentHeap(), 0, demo);
}

void StartButtonThread(PSENSORDEMO demo)
{
#ifdef arm
    SCODE sc;
    PIPROCESS pPrc = CurrentProcess();

    Condition_Init(&TheButtons.ButtonPress);
    sc = pPrc->v->CreateThread(pPrc, 0,
                               ButtonThread,
                               (THREAD_ARGUMENT)demo,
                               2048, /* use 2k for stack */
                               NULL,
                               NULL);

    _tprintf(_TEXT("Initializing Button 4\r\n"));
    PioA->Disable = PIOA_IRQ0;
    PioA->FilterEnable = PIOA_IRQ0;
    sc = AddDevice ((PTR)3, (PTR) ButtonIsr, 0, 31, NULL);

    _tprintf(_TEXT("Initializing Buttons 1,2,3\r\n"));
    PioB->Disable = PIOB_PB3 | PIOB_PB4 | PIOB_PB5;
    PioB->FilterEnable = PIOB_PB3 | PIOB_PB4 | PIOB_PB5;
    sc = PioB->IntrStatus; /* clear pending */
    sc = AddDevice ((PTR)PioB, (PTR) PioIsr, DEVICE_FLAGS_LEVEL_TRIGGER, 14, NULL);
    PioB->IntrEnable = PIOB_PB3 | PIOB_PB4 | PIOB_PB5; 

    _tprintf(_TEXT("All Buttons initialized\r\n"));
#else
    _tprintf(_TEXT("StartButtonThread: Only implemented for ARM platforms\n"));
#endif
}

void StopButtonThread(PSENSORDEMO demo)
{
    UnusedParameter(demo); 

    /* It will go away in it's own good time */
    TimeToDie = TRUE; 
}

#endif /* arm */
