/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Sensor demo implementation. 
 */

#include "instigator.h"


/* IUnknown */
PRIVATE SCODE SensorDemoInstigatorQueryInterface (
    PIINSTIGATOR pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                                 &IID_IInstigator);
}
    
PRIVATE UINT SensorDemoInstigatorAddRef (
    PIINSTIGATOR pThis)
{
    return AtomicInc(&pInstigator(pThis)->RefCnt);
}
    
/* Deallocate the instigator memory.
 * 
 * The task structure is finalized by the planning library, 
 * not by the instigator
 */
PRIVATE UINT SensorDemoInstigatorRelease (
    PIINSTIGATOR pThis)
{
    PINSTIGATOR This = pInstigator(pThis);
    UINT Refs = AtomicDec(&This->RefCnt);
    if (Refs == 0) {
        PIHEAP pHeap = CurrentHeap();
        /* Finalize object */
        if (This->ProducerUri)
            pHeap->v->Free(pHeap, 0, This->ProducerUri);
        if (This->ConsumerUri)
            pHeap->v->Free(pHeap, 0, This->ConsumerUri);
        if (This->ProducerReservationFullName)
            pHeap->v->Free(pHeap, 0, This->ProducerReservationFullName);
        if (This->ConsumerReservationFullName)
            pHeap->v->Free(pHeap, 0, This->ConsumerReservationFullName);
        if (This->FeedbackUri)
            pHeap->v->Free(pHeap, 0, This->FeedbackUri);
        This->Task = NULL;
        pHeap->v->Free(pHeap, 0, (void *) This);
    }
    return Refs;
}
    
PRIVATE SCODE SensorDemoInstigatorInitialize(
    PIINSTIGATOR iThis, INT argc, _TCHAR **argv)
{
    METHOD_DECLARE(Instigator, Initialize);
    _TCHAR **av;

    This->RefCnt = 1;

    sc = ParseCmdLine(&argc,&argv);
    CHECKSC("parse command line");

    av = argv;
    while (argc > 1 && av[1][0] == _T('-')) {
        switch(av[1][1]) {
        case _T('t'):
            This->method = _T("WasteTime");
            break;
        default:
            CHECKSETSC(FALSE, E_INVALID_PARAMETER, "unknown option");
        }
        argc--;
        av++;
    }

    CHECKSETSC(argc <= 4, E_INVALID_PARAMETER, "too many arguments");

    if (argc >= 2)
        This->ProducerUri = _tcsdup(av[1]);
    else
        This->ProducerUri =  _tcsdup(_T("http://127.0.0.1/COB/sensor.cob"));

    if (argc >= 3)
        This->ConsumerUri = _tcsdup(av[2]);
    else
        This->ConsumerUri = _tcsdup(This->ProducerUri);

    if (argc >= 4)
        This->FeedbackUri = _tcsdup(av[3]);
    else
        This->FeedbackUri = _tcsdup(_T("http://127.0.0.1/COB/wsres.cob"));

    printf("SENSORMAIN: producer=%s, consumer=%s, feedback=%s method=%x\n", 
        This->ProducerUri, This->ConsumerUri, This->FeedbackUri, This->method);

METHOD_CLEANUP;
    FreeCmdLine(&argc, &argv);
    if (sc == E_INVALID_PARAMETER) 
        printf("Usage: sensormain.exe [-t] [<sensor-url> [<recorder-url>]]\n");
    return sc;
}

/* Replicate here for linkage reasons -> ws.c */
/* Assumes the struct and string were allocated in one call */
PRIVATE INT WsEndpointDelete(PTR pItem, PTR pArg)
{
    PIHEAP pHeap = (PIHEAP) pArg;
    pHeap->v->Free(pHeap, 0, pItem);
    return 0;
}


PRIVATE void WsContextDestroy(PWSCONTEXT pCtx)
{
    PIHEAP Heap = CurrentHeap();

    if (pCtx->To)
        WsEndpointDelete(pCtx->To, Heap);
    if (pCtx->Action)
        Heap->v->Free(Heap, 0, pCtx->Action);
    if (pCtx->ReplyTo)
        WsEndpointDelete(pCtx->ReplyTo, Heap);
    if (pCtx->taskName)
        Heap->v->Free(Heap, 0, pCtx->taskName);
    if (pCtx->ReservationName)
        Heap->v->Free(Heap, 0, pCtx->ReservationName);
    if (pCtx->TaskName)
        Heap->v->Free(Heap, 0, pCtx->TaskName);
    Heap->v->Free(Heap, 0, pCtx);
}


/* Start (or restart...) the producer.
 *
 * Here's how this works... The instigator code (_tmain) prepares
 * the reservation using the reservation builder API, and then 
 * hands the completed structure to the planning library using the
 * ActivateReservation message. 
 *
 * The planning library invokes the provided instigator callback 
 * directly, but also stores a copy of the reservation, and 
 * periodically checks in on it. The checks consist of (a) making 
 * sure that a recurring reservation is kept alive (because hosts
 * may throw out recurring reservations that aren't periodically 
 * revived with a heartbeat), and (b) looking for improved 
 * reservation parameters using feedback statistics. 
 
 * The planner's copy of the reservation is updated as feedback 
 * becomes available, and the instigator callback is re-invoked when 
 * it is necessary to renew the host. The host receives the updated 
 * parameters at the same time
 */
PRIVATE SCODE MCT SensorDemoProducerDispatch(
    PTR Arg, PIRESERVATIONBUILDER ProducerReservation)
{
    PIINSTIGATOR iThis = (PIINSTIGATOR) Arg;
    METHOD_DECLARE(Instigator, ProducerDispatch);

    PISOAPFACTORY soap;
    PWSCONTEXT ctx;
    PISENSOR pxy;

	printf("instigator.c:ProducerDispatch: this=0x%x res=0x%x\n", 
		This, ProducerReservation);

    sc = ProducerReservation->v->CreateContext(ProducerReservation,
                                               This->FeedbackUri,
                                               &ctx);
    CHECKSC("create producer context object");

    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &soap);
    CHECKSC("bind to SOAP COB");

    sc = soap->v->CreateProxy3(soap, NULL, ctx, _T("sensor"),
                               0, (PIUNKNOWN*) &pxy);
    CHECKSC("create proxy");
    printf("sensor-main: Created proxy: pxy=%x sc=%x\n", (UINT) pxy, sc);

    if (This->method) {
        DBG(("WasteTime method=%x\n", This->method));
        sc = pxy->v->Main(pxy, This->ConsumerUri);
        printf("Do-some-work, uri=%s sc=%x\n", This->ConsumerUri, sc);
    } else {
        DBG(("Read and deliver\n"));
        sc = pxy->v->ReadAndDeliver(pxy, 0,
                                    This->ConsumerUri,
                                    This->ConsumerReservationFullName);
        printf("Read-and-deliver, uri=%s sc=%x\n", This->ConsumerUri, sc);
    }

METHOD_CLEANUP;
    if (pxy)
        pxy->v->Release(pxy);
    if (ctx)
        WsContextDestroy(ctx);
    if (soap)
        soap->v->Release(soap);

    return sc;
}
    

PRIVATE SCODE MCT SensorDemoConsumerDispatch(
    PTR Arg, PIRESERVATIONBUILDER ConsumerReservation)
{
    PIINSTIGATOR iThis = (PIINSTIGATOR) Arg;
    METHOD_DECLARE(Instigator, ConsumerDispatch);

    PISOAPFACTORY soap;
    PWSCONTEXT ctx;
    PISENSOR pxy;

	printf("instigator.c:ConsumerDispatch: this=0x%x res=0x%x\n", 
		This, ConsumerReservation);

    sc = ConsumerReservation->v->
        CreateContext(ConsumerReservation, This->FeedbackUri, &ctx);
    CHECKSC("create consumer context object");

    /* Clear the wsa:RelatedTo field. 
       This dispatch is only intended to establish a reservation. 
       By clearing the RelatedTo field we indicate that no continuation
       should be constructed.
       (The continuation appears later, when the producer invokes the 
       consumer.) */
    free(ctx->taskName);
    ctx->taskName = NULL;

    /* Get a proxy */

    sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"), NAME_SPACE_READ,
                      &IID_ISoapFactory, (void **) &soap);
    CHECKSC("bind to SOAP COB");

    sc = soap->v->CreateProxy3(soap, NULL, ctx, _T("sensor"),
                               0, (PIUNKNOWN*) &pxy);
    CHECKSC("create proxy");
    printf("sensor-main: Created proxy: pxy=%x sc=%x\n", (UINT) pxy, sc);

    /* This is kind of dumb... but we need to invoke something on the 
       consumer, just to make sure that the reservation is established. 
       The invocation that the reservation is piggybacked with has *nothing*
       to do with claiming the reservation. The reservation won't be claimed
       until the producer starts invoking ReadAndDeliver.

       Got that? */
    sc = pxy->v->DoNoWork(pxy);
    CHECKSC("establish consumer reservation");
	printf("established consumer reservation--------------------------------------------\n");

METHOD_CLEANUP;
    if (pxy)
        pxy->v->Release(pxy);
    if (ctx)
        WsContextDestroy(ctx);
    if (soap)
        soap->v->Release(soap);

    return sc;
}


PRIVATE SCODE 
SensorDemoInstigatorDefineProducerReservation(
    PIINSTIGATOR iThis, PTIME OutDeadline)
{
    METHOD_DECLARE(Instigator, DefineProducerReservation);
    PIRESERVATIONBUILDER ProducerReservation;
    _TCHAR *ReservationName;
    TIME ProducerPeriod, ProducerDeadline;

    if (! Int64Equal(*OutDeadline, TIME_ORIGIN))
        return E_INVALID_PARAMETER;

    /* Set up the producer reservation */

    sc = This->Task->v->CreateReservation(This->Task, 
        This->ProducerUri, This->FeedbackUri, 
        PRODUCER_RESERVATION_NAME,
        &ProducerReservation);
    CHECKSC("create producer reservation");

    ProducerDeadline = CurrentTime();
    ProducerDeadline = Int64SubtractInt32(
        ProducerDeadline, TIME_RELATIVE(TIME_MILLIS(9000)));

    sc = ProducerReservation->v->
        SetParameters(ProducerReservation, 
                      ProducerDeadline, 
                      TIME_RELATIVE(TIME_MILLIS(20)), 
                      TIME_RELATIVE(TIME_MILLIS(5)));
    CHECKSC("set producer reservation parameters");

    Int32ToInt64(ProducerPeriod, TIME_RELATIVE(TIME_SECONDS(PRODUCER_PERIOD)));
    sc = ProducerReservation->v->SetRecurrence(ProducerReservation, 
        ProducerPeriod, PRODUCER_REPEATS);
    CHECKSC("set producer reservation trigger");

    /* Require 95% confidence before overriding the existing estimate */

    ReservationName = (_TCHAR *) 
        ProducerReservation->v->GetFullName(ProducerReservation);
    This->ProducerReservationFullName = _tcsdup(ReservationName);

    sc = ProducerReservation->v->SetDispatch(ProducerReservation, 
        (PTR) SensorDemoProducerDispatch, (PTR) This);
    CHECKSC("set producer dispatch function and data");

    sc = ProducerReservation->v->SetAdaptivity(
        ProducerReservation, ReservationName, 95, 
        (PTR) SensorDemoAdaptation, (PTR) This);
    free(ReservationName);
    CHECKSC("set producer reservation adaptivity");

    *OutDeadline = ProducerDeadline;

METHOD_CLEANUP;
    if (Int64Equal(*OutDeadline, TIME_ORIGIN))
        assert(FAILED(sc));
    return sc;
}


PRIVATE SCODE 
SensorDemoInstigatorDefineConsumerReservation(
    PIINSTIGATOR iThis, TIME ProducerDeadline, TIME TransmissionTime)
{
    METHOD_DECLARE(Instigator, DefineConsumerReservation);
    PIRESERVATIONBUILDER ConsumerReservation;
    _TCHAR *ReservationName;
    TIME ConsumerPeriod, ConsumerDeadline;

    /* Set up the consumer reservation */

    sc = This->Task->v->CreateReservation(This->Task, 
        This->ConsumerUri, This->FeedbackUri, 
        CONSUMER_RESERVATION_NAME,
        &ConsumerReservation);
    CHECKSC("create consumer reservation");

    ConsumerDeadline = CurrentTime();
    ConsumerDeadline = 
        Int64SubtractInt32(
            ConsumerDeadline, 
            TIME_RELATIVE(TIME_MILLIS(9500)));

    sc = ConsumerReservation->v->
            SetParameters(ConsumerReservation, 
                          ConsumerDeadline, 
                          TIME_RELATIVE(TIME_MILLIS(20)), 
                          TIME_RELATIVE(TIME_MILLIS(5)));
    CHECKSC("set consumer reservation parameters");

    Int32ToInt64(ConsumerPeriod, TIME_RELATIVE(TIME_SECONDS(CONSUMER_PERIOD)));
    sc = ConsumerReservation->v->SetRecurrence(ConsumerReservation, 
        ConsumerPeriod, CONSUMER_REPEATS);
    CHECKSC("set consumer reservation trigger");

    /* Require 95% confidence before overriding the existing estimate */

    ReservationName = (_TCHAR *) 
        ConsumerReservation->v->GetFullName(ConsumerReservation);
    This->ConsumerReservationFullName = _tcsdup(ReservationName);

    sc = ConsumerReservation->v->
        SetDispatch(
            ConsumerReservation, 
            (PTR) SensorDemoConsumerDispatch, 
            (PTR) This);
    CHECKSC("set consumer dispatch function and data");

    sc = ConsumerReservation->v->
        SetAdaptivity(
            ConsumerReservation, ReservationName, 95, 
            (PTR) SensorDemoAdaptation, (PTR) This);
    free(ReservationName);
    CHECKSC("set consumer reservation duration adaptivity");

METHOD_CLEANUP;
    return sc;
}

PRIVATE SCODE 
SensorDemoInstigatorDefineTask(PIINSTIGATOR iThis)
{
    METHOD_DECLARE(Instigator, DefineTask);
    TIME ProducerDeadline = TIME_ORIGIN; 
    TIME TransmissionTime = TIME_ORIGIN; 

    /* Set up the producer reservation and get back the producer deadline */

    sc = SensorDemoInstigatorDefineProducerReservation(iThis, &ProducerDeadline);
    CHECKSC("define producer reservation");

    /* Set up an initial guess of the transmission time */

    TransmissionTime = 
        Int64SubtractInt32(
            TransmissionTime, 
            WILDLY_INACCURATE_TRANSMISSION_TIME_GUESS);

    /* Combine the producer deadline and transmission time to 
       define the consumer deadline */

    sc = SensorDemoInstigatorDefineConsumerReservation(
        iThis, ProducerDeadline, TransmissionTime);
    CHECKSC("define consumer reservation");

METHOD_CLEANUP;
    return sc;
}
    
PRIVATE SCODE SensorDemoInstigatorRun(PIINSTIGATOR iThis)
{
    METHOD_DECLARE(Instigator, Run);
    PIPLANNER Planner;

    sc = BindToObject(CurrentNameSpace(), _T("COB/planner.cob"), NAME_SPACE_READ,
                      &IID_IPlanner, (void **) &Planner);
    CHECKSC("create planner");

    sc = Planner->v->CreateTask(Planner, iThis, &This->Task);
    CHECKSC("create task");

    sc = iThis->v->DefineTask(iThis);
    CHECKSC("create reservations");

    sc = This->Task->v->Activate(This->Task, iThis);
    CHECKSC("activate task");

METHOD_CLEANUP;
    return FAILED(sc) ? -1 : 0;
}
    
#include "s_SensorDemoInstigator_v.c"
    
PRIVATE PIINSTIGATOR SensorDemoInstigatorCreate(INT argc, _TCHAR *argv[])
{
    METHOD_DECLARE_NOTHIS(sensormain.c, InstigatorCreate);
    PIHEAP Heap = CurrentHeap();
    PIINSTIGATOR Instigator;

    Instigator = (PIINSTIGATOR) Heap->v->Alloc(Heap, 
        HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("sensormain.c:InstigatorCreate"),
        sizeof(Instigator), 0);
    if (!Instigator)
        return NULL;

    Instigator->v = &SensorDemoInstigatorVtbl;

    sc = Instigator->v->Initialize(Instigator, argc, argv); 
    CHECKSC("parse producer/consumer/feedback URLs");

METHOD_CLEANUP;
    if (FAILED(sc) && Instigator) {
        Instigator->v->Release(Instigator);
        Instigator = NULL;
    }
    return Instigator;
}


int _tmain(INT argc, _TCHAR *argv[])
{
    METHOD_DECLARE_NOTHIS(instigator.c, _tmain);

    PIINSTIGATOR Instigator = SensorDemoInstigatorCreate(argc, argv);
    CHECKSETSC(Instigator, E_FAIL, "create instigator");

    sc = Instigator->v->Run(Instigator);
    CHECKSC("run instigator");

METHOD_CLEANUP;
    if (Instigator)
        Instigator->v->Release(Instigator);
    return FAILED(sc) ? -1 : 0;
}
    
