/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Sensor demo implementation. 
 *
 * Instigator first (_tmain)
 * Then demo component setup,
 * then sensor interface implementation, 
 * then recorder interface implementation.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <assert.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <web/proxy.h>
#include <web/udpsoapserv.h>
#include <arm/at91m63200.h>
#include "sensor.h"
#include "../soap/soap.h"

#define _MANUAL_COBMAIN 1
#include "s_SensorDemo_v.c"

// Instigator implementation

#define vloff offsetof(struct via, ListNode)
#define rloff offsetof(struct reservation, ListNode)    

PRIVATE SCODE SensorDemoDestroy(IUnknown *iThis)
{
    PSENSOR This = pSensor(iThis);
    
#if defined(arm) && 0
    StopButtonThread((PSENSORDEMO) This);
#endif

    return S_OK;
}

#if 0
// Sensor interface implementation

PRIVATE SCODE MCT SensorQueryInterface(PISENSOR Sensor, REFIID Iid, void** Unk)
{
    SCODE sc;
    PSENSOR This = pSensor(Sensor);

    sc = GenericQueryInterface((IUnknown *) Sensor, Iid, Unk, &IID_ISensor);
//    DBG(("SensorQI sc=%x sensor=%x unk=%x\n", sc, (ADDRESS) This, (ADDRESS) *Unk));

    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IRecorder, Iid)) {
        Sensor->v->AddRef(Sensor);
        *Unk = (void*) &This->Recorderv;
        sc = S_OK;
//        DBG(("SensorQI Rec-from-Sens &this->vs=%x &this->vr=%x unk=%x offset=%d\n", 
//               (ADDRESS) &This->vs, (ADDRESS) &This->vr, (ADDRESS) *Unk, offsetof(RECORDER, vr)));
    }
    if ((sc == E_NO_INTERFACE) && UuidCmp(&IID_IProgram, Iid)) {
        Sensor->v->AddRef(Sensor);
        *Unk = (void*) &This->Sensorv;
        sc = S_OK;
    }

    return sc;
}
        
PRIVATE UINT MCT SensorAddRef(PISENSOR Sensor)
{
    UINT Refs;
    PSENSOR This = pSensor(Sensor);

    Refs = AtomicInc(&This->Refs);

//    DBG(("SensorAddRef: Refs=%d\n", Refs));
    return (Refs);
}

PRIVATE UINT MCT SensorRelease(PISENSOR Sensor)
{
    UINT Refs;
    PSENSOR This = pSensor(Sensor);

    Refs = AtomicDec(&This->Refs);

    if (Refs == 0)
        SensorDemoDestroy((IUnknown *) This);

    return Refs;
}
#endif

/* Read sensor data into parameter buffer */
#define nBUTTONS 4
#define nLEDS 8
SCODE SensorRead(PISENSOR Sensor, 
                 /*in*/UINT32 Element, 
                 /*out*/_TCHAR* Data, 
                 /*in*/UINT DataSize)
{
    SCODE sc = S_OK;
    PSENSOR This = pSensor(Sensor);
    UINT DataIdx; /* , LedIdx = 0; */

    /* Check and clear the data buffer */
    if (sizeof(UINT) > DataSize) 
        return E_BUFFER_TOO_SMALL;
    for (DataIdx = 0; DataIdx < DataSize; ++DataIdx)
        Data[DataIdx] = 0;

    DBGME(1, printf(
        "sensor.c:SensorRead: buffer size=%d, data=%x\n", DataSize, *(UINT*)Data));

    /* Lock, copy and unlock button state */
    Mutex_Lock(&This->Mutex);
    *((UINT*)Data) = This->State;
    Mutex_Unlock(&This->Mutex);

    printf("sensor.c:SensorRead: sc=%x, data=%x\n", sc, *(UINT*)Data);
    
    return sc;
}

/* Read sensor data and deliver to parameter URL.

   The last parameter must be the full name (i.e. path): 
   the task name, reservation name and sequence number */
SCODE SensorReadAndDeliver(PISENSOR Sensor, 
                           /*in*/UINT32 Element, 
                           /*in*/const _TCHAR* DestinationUrl,
                           /*in*/const _TCHAR* DestinationReservation)
{
    static PIRECORDER pxy = NULL;

    SCODE sc = S_OK;
    PISOAPFACTORY soapfactory = NULL;
    PWSCONTEXT pCtx = NULL;
    PEndpointReference pTo;
    _TCHAR *consumeruri = (_TCHAR*) DestinationUrl;
    _TCHAR *consumerrsrv = (_TCHAR *) DestinationReservation;
    _TCHAR Data[sizeof(UINT) << 1];
    UINT BufIdx;
    for (BufIdx = 0; BufIdx < sizeof(Data); ++BufIdx)
        Data[BufIdx] = '\0';

    printf("sensor.c:ReadAndDeliver: element=0x%x dest-url=%s dest-res=%s\n", 
           Element, DestinationUrl, DestinationReservation);

    /* Read the sensor data */
    sc = Sensor->v->Read(Sensor, Element, Data, sizeof(Data));
    if (FAILED(sc)) return sc;

    if (!pxy) {
        /* Prepare context object */
        pCtx = (PWSCONTEXT) malloc(sizeof(*pCtx));
        pTo = (PEndpointReference) malloc(sizeof(*pTo));
        if (!pTo || !pCtx)
            goto NoMem;
        memset(pCtx, 0, sizeof(WSCONTEXT));
        memset(pTo, 0, sizeof(*pTo));
        pCtx->To = pTo;
        pTo->Address = _tcsdup(consumeruri);
        if (consumerrsrv && _tcslen(consumerrsrv) > 0)
            pCtx->taskName = _tcsdup(consumerrsrv);
        else
            pCtx->taskName = NULL;
        pCtx->TaskName = NULL;

        /* Construct recorder proxy */
        sc = BindToObject(CurrentNameSpace(),
                          _T("COB/soap.cob"), NAME_SPACE_READ,
                          &IID_ISoapFactory, (void **) &soapfactory);
        if (FAILED(sc)) goto NoSoap;
        sc = soapfactory->v->
            CreateProxy3(soapfactory, NULL, pCtx, _T("recorder"),
                         0, (PIUNKNOWN*) &pxy);
        if (FAILED(sc)) goto NoProxy;
        DBGME(1, printf(
                      "sensor.c:ReadAndDeliver: Created proxy: pxy=%x sc=%x\n",
                      (UINT) pxy, sc));
    }

    /* Deliver the data */
    sc = pxy->v->Write(pxy, _T("Joe Bob"), Data, sizeof(Data));
    DBGME(1, printf(
        "sensor.c:ReadAndDeliver: Delivered data=%x\n", *(UINT*)Data));

Out:
    //if (pxy)
    //    pxy->v->Release(pxy);
    if (soapfactory)
        soapfactory->v->Release(soapfactory);

//    DBG(("SensorReadAndDeliver sc=%x\n", sc));
    return sc;

NoProxy:
    DBG(("CreateProxy %s failed sc=%x\n", pCtx->To->Address, sc));
    goto Out;

NoSoap:
    DBG(("Could not bind to soap.cob sc=%x\n", sc));
    goto Out;
NoMem:
    sc = E_NOT_ENOUGH_MEMORY;
    DBG(("sensor.c out of mem sc=%x\n", sc));
    goto Out;
}

PRIVATE volatile UINT xxx = 0;
PRIVATE volatile UINT iter = 0;
PRIVATE SCODE SensorDoSomeWork(PISENSOR Sensor, const _TCHAR* arg)
{
    SCODE sc = S_FALSE;
    UINT i;
    /* A parameter of ~10K works OK on the eb63 */
    UINT loopCount = _ttoi(arg);
    UINT doubleCount = 2*loopCount;

    DBGME(1, printf("sensor.c:DoSomeWork\n"));

    /* XXX encrypt a megabyte or some such */
    if (iter++ > 256+64)
        for (i = 0; i < loopCount; i++) /* 50% of the below */
            xxx *= i;
    else
        for (i = 0; i < doubleCount; i++)
            xxx *= i;
    
    return sc;
}

PRIVATE SCODE 
SensorDoNoWork(PISENSOR Sensor)
{
    return S_OK;
}

void SkipListTest(void);

PRIVATE SCODE SensorMain(PISENSOR Sensor, const _TCHAR* arg)
{
    fprintf(stderr, "qwerty!\n");
    //SkipListTest();
    fprintf(stderr, "asdfgh!\n");
    DBGT((_T("** SensorMain(%s)\n"), arg));
    return SensorDoSomeWork(Sensor, arg);
}

// Recorder interface implementation

PRIVATE SCODE MCT RecorderQueryInterface(PIRECORDER iThis,
                                         REFIID iid, void** ppObject)
{
    IUnknown *unk = (IUnknown *) pRecorder(iThis);
    return unk->v->QueryInterface(unk, iid, ppObject);
}

PRIVATE UINT MCT RecorderAddRef(PIRECORDER iThis)
{
    IUnknown *unk = (IUnknown *) pRecorder(iThis);
    return unk->v->AddRef(unk);
}

PRIVATE UINT MCT RecorderRelease(PIRECORDER iThis)
{
    IUnknown *unk = (IUnknown *) pRecorder(iThis);
    return unk->v->Release(unk);
}

void THREAD_LINKAGE RecorderClearLedThread(THREAD_ARGUMENT arg) 
{
    UINT LedIdx = 0;

    SleepUntil(Int64SubtractInt32(CurrentTime(), TIME_RELATIVE(TIME_SECONDS(5))));

    /* Leave the last LED alone - it indicates correct initialization. 
     * (See startup script.) */
    for (LedIdx = 0; LedIdx < nLEDS - 1; ++LedIdx) {
#ifdef arm
        PioB->SetData = PIOB_PB8 << LedIdx;
#endif
    }
}

SCODE RecorderWrite(PIRECORDER Recorder, 
                    /*in*/const _TCHAR* Source, 
                    /*in*/const _TCHAR* Data, 
                    /*in*/UINT DataSize)
{
    UINT LedIdx = 0;
#ifndef arm 
    static UINT32 Mask = 0;
#endif

    printf("RecorderWrite recorder=%x source=%s data=%x\n", 
        (ADDRESS) Recorder, Source, *(UINT*)Data);
    PrintTime("RecorderWrite time", CurrentTime());

    for (LedIdx = 0; LedIdx < nLEDS; ++LedIdx) {
        if (*(UINT*)Data & (1 << LedIdx)) {
#ifdef arm 
            PioB->ClearData = PIOB_PB8 << LedIdx;
#else
            Mask |= 1 << LedIdx;
#endif
        } else {
#ifdef arm
            PioB->SetData = PIOB_PB8 << LedIdx;
#else
            Mask &= ~(1 << LedIdx);
#endif
        }
    }

//    DBG(("RecorderWrite: state %d\n", Mask));

    /* Construct slave thread for the actual execution. */
    CurrentProcess()->v->CreateThread(CurrentProcess(), 0,
                                      RecorderClearLedThread,
                                      NULL,
                                      2048, /* use 2k for stack */
                                      NULL,
                                      NULL);

    return S_OK;
}

typedef struct _WARGS {
    PIRECORDER Recorder;
    const _TCHAR* Source;
    const _TCHAR* Data;
    UINT DataSize;
    const _TCHAR* Reservation;
} WARGS, *PWARGS;

SCODE RecorderAsyncWrite(PIRECORDER Recorder, 
                    /*in*/const _TCHAR* Source, 
                    /*in*/const _TCHAR* Data, 
                    /*in*/UINT DataSize, 
                    /*in*/const _TCHAR* Reservation)
{
    SCODE sc;

    /* Deallocated by the slave thread. */
    PWARGS args = (PWARGS) CurrentHeap()->v->
        Alloc(CurrentHeap(), HEAP_DEBUG_NAME(_T("sensor-demo")), sizeof(WARGS), 0);
    args->Source = Source;
    args->Data = Data;
    args->DataSize = DataSize;
    args->Reservation = Reservation;

#if 0
    /* Construct slave thread for the actual execution. */
    sc = CurrentProcess()->v->CreateThread(CurrentProcess(),
                               0,
                               RecorderAsyncWriteThread,
                               (THREAD_ARGUMENT)args,
                               2048, /* use 2k for stack */
                               NULL,
                               NULL);
#else
    sc = E_NOT_IMPLEMENTED;
#endif

//    DBG(("RecorderAsyncWrite: thread creation sc=%x\n", sc));

    return sc;
}

#if 0
/* Sensor demo component */
PIUNKNOWN CreateSensorDemo(void)
{
    PSENSOR demo;
    PIHEAP pHeap = CurrentHeap();

    demo = (PSENSOR) pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME(_T("sensor-demo")), sizeof(SENSOR), 0);
    if (! demo)
        return NULL;

    demo->super.super.v = (PTR) &SensorSensorDemoVtbl.v;
    demo->super.super.RefCnt = 1;
    demo->Recorderv = &RecorderSensorDemoVtbl.v;
    Mutex_Init(&demo->Mutex);
    demo->State = 0;

    return (PIUNKNOWN) demo;
}
#endif

PIUNKNOWN sensorCobMain(void)
{
    PIUNKNOWN demo = GenericCobInit((void*) &sensor_CobDesc, NULL);

#if defined(arm)
    if (demo)
        StartButtonThread((PSENSOR) demo);
#endif
    return demo;
}



