/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Typeinfo handling */
#include "diagnostics.h"
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <mmlite.h>
#include <netif.h>
#include <winsock.h>
#include <assert.h>
#include "soapif.h"
#include <rpc.h>
#include "soap.h"
#include <sampleif.h>
#include <typeinfo.h>

char* TYPEINFO_ELEMENT_STRINGS[] = {
    "interface-group",
    "interface", 
    "Envelope", 
    "Body", 
    "Header",
    "Fault",
    "faultcode",
    "faultstring",
    "detail",
    "reply", 
    "method", 
    "arg", 
    "stack", 
    "frame", 
    "response", 
    "WSDL_PART", 
    "definitions", 
    "types", 
    "binding", 
    "portType", 
    "operation", 
    "input", 
    "output",
    ":",
    " ",
    "xmlns",
    "<",
    ">",
    "\"",
    "=",
    "encodingStyle",
    "</",
    "\0",
    "struct-group",
    "struct",
    "field"
};

char ** const_strings = TYPEINFO_ELEMENT_STRINGS;

const struct ISaxHandlerVtbl  SAXTypeInfoVtbl;

PRIVATE SCODE MCT TypeInfoProcessAttribute(PSAXCALLBACKS Sax_Object,
                                   char * prefix, UINT prefixLength,
                                   char * name, UINT nameLength, 
                                   char * value, UINT valueLength);
PSTRUCT_DESCR FindStruct(PTYPEINFO ptypeinfo, char *name, UINT nameLength, UINT*index);

/* SOAP/SAX Typeinfo construction/destruction */
SCODE MCT
SAXTypeInfoCreate(PISAXHANDLER * Saxcallbacks, PTYPEINFO ptypeinfo, PIHEAP Heap)
{
    PSAXCALLBACKS This;

    TRACE("SAX CallBacks Create");

    This = Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY, sizeof(SAXCALLBACKS), 0);
    if (This == NULL) return E_NOT_ENOUGH_MEMORY;
    
    This->v     = &SAXTypeInfoVtbl;
    This->Refs  = 1;
    This->Heap = Heap;
    This->indexCurrentArgument = -1;
    This->indexCurrentMethod = -1;
    This->pCurrentInterface = NULL;
    This->prpcinfo = NULL;
    This->ptypeinfo = ptypeinfo;
    This->Elem_Length = 0;
    This->Elem_or_Attr = NULL;
    This->filename = NULL;
    This->previousTokenType = TOKEN_NOT_SEEN;
    This->currentLevel = 0;
    This->version = 0;
    This->char_set = 0;
    This->flag = 0;
    *Saxcallbacks = (PISAXHANDLER) This;
    return S_OK;
}

PRIVATE PIHEAP __pheap;

SCODE InitTypeInfo_FromFile(PTYPEINFO *pRetval)
{
    PTYPEINFO pTypeInfo = NULL;
    PINAMESPACE NameSpace = CurrentNameSpace();
    PIUNKNOWN   Unk = NULL;
    SCODE sc;
    _TCHAR*       InputFileName = _T("fs/Type_Info_Schema.xml");
    PISAXHANDLER TypeInfoHandler;
    PISAXPARSER     SAXParser;
    PITOKENIZER     tokenizer;
    PITOKENIZERFACTORY  Factory     = NULL;
    PIHEAP Heap = _GetHeap(_T("TypeInfo"), 4000);

    __pheap = Heap;
    DBG(("Constructing SAX parser\n"));
    sc = NameSpace->v->Bind(NameSpace, _T("COB/sax.cob"), 0, &Unk);
    if (FAILED(sc)) goto err_parser_bind;

    DBG(("Bound SAX parser COB\n"));

    sc = Unk->v->QueryInterface(Unk, &IID_ISAXPARSER, (void**) &SAXParser);
    Unk->v->Release(Unk);
    if (FAILED(sc)) goto err_parser_query;

    DBG(("Queried for SAX parser interface\n"));

    pTypeInfo = (PTYPEINFO) Heap->v->Alloc(Heap, 
                                           HEAP_ZERO_MEMORY, 
                                           sizeof(TYPEINFO),
                                           0);
    if (pTypeInfo == NULL) {
        sc = E_NOT_ENOUGH_MEMORY;
        goto err_typeinfo_alloc;
    }
    SAXTypeInfoCreate(&TypeInfoHandler, pTypeInfo, Heap);

    /* Open the interface definition file for input*/
    DBG(("Opening xml file for interface definitions\n"));
    sc = BindToObject(NameSpace, InputFileName, NAME_SPACE_READ,
                      &IID_IFile, (void**) &pTypeInfo->ifaceDefnFile);
    if (FAILED(sc)) goto err_file;
    
    DBG(("HTTPSessionCreate: Binding to tokenizer COB\n"));
    sc = NameSpace->v->Bind(NameSpace, _T("COB/tokenizer.cob"), 
                            0, (PIUNKNOWN*) &Unk);
    if (FAILED(sc)) goto err_tok_bind;

    DBG(("HTTPSessionCreate: Querying for tokenizer factory\n"));

    sc = Unk->v->QueryInterface(Unk, &IID_ITOKENIZERFACTORY, (void**) &Factory);
    Unk->v->Release(Unk);
    if (FAILED(sc)) goto err_tok_bind;

    DBG(("HTTPSessionCreate: Creating tokenizer\n"));

    sc = Factory->v->Create(Factory, (PIUNKNOWN) (pTypeInfo)->ifaceDefnFile, &tokenizer);

    Factory->v->Release(Factory);
    
    if (FAILED(sc)) goto err_tok_create;
    if (sc == S_FALSE) {    /* XXX??? */
        sc = E_FAIL;
        goto err_tok_create;
    }

    DBG(("Parsing and echoing\n"));
    sc = SAXParser->v->Parse(SAXParser, tokenizer, TypeInfoHandler);
    if (FAILED(sc)) {
        DBG(("Typeinfo parsing failed sc=%x\n", sc));
        goto err_parse;
    }

    pTypeInfo->ifaceDefnFile->v->Release(pTypeInfo->ifaceDefnFile);
    tokenizer->v->Release(tokenizer);
    SAXParser->v->Release(SAXParser);
    *pRetval = pTypeInfo;
    return S_OK;

err_tok_create:
err_tok_bind:
    pTypeInfo->ifaceDefnFile->v->Release(pTypeInfo->ifaceDefnFile);
err_file:
    Heap->v->Free(Heap, 0, pTypeInfo);
err_typeinfo_alloc:
    SAXParser->v->Release(SAXParser);
err_parser_query:
err_parser_bind:
err_parse:
    return sc;
}

PRIVATE SCODE MCT
SAXTypeInfoQueryInterface(PISAXHANDLER This, REFIID IidQuery, void** Object)
{
    PIUNKNOWN Unk;
   
    TRACE("SAXCallbacksQueryInterface");

    if (UuidCmp(IidQuery, &IID_ISaxHandler))
        Unk = (PIUNKNOWN) This;
    else 
        Unk = NULL;

    if (Unk != NULL)
        (Unk)->v->AddRef(Unk);

    *Object = Unk;
    
    return (Unk == NULL ? E_NO_INTERFACE : S_OK);
}


PRIVATE UINT MCT
SAXTypeInfoAddRef(PISAXHANDLER Sax)
{
    UINT RefCnt;
    PSAXCALLBACKS This = (PSAXCALLBACKS) Sax;
    
    TRACE("SAXCallbacksAddRef");

    RefCnt = AtomicInc(&This->Refs);
    return (RefCnt);
}

SCODE DestroyTypeInfo(PTYPEINFO pTypeInfo)
{
    PIHEAP Heap = __pheap;
    Heap->v->Free(Heap, 0, pTypeInfo->ifacelist->groupname);
    Heap->v->Free(Heap, 0, pTypeInfo->ifacelist->name);
    Heap->v->Free(Heap, 0, pTypeInfo);
    _PutHeap(Heap);
    return S_OK;
}

PRIVATE UINT MCT 
SAXTypeInfoRelease(PISAXHANDLER Sax)
{
    UINT Refs;
    PSAXCALLBACKS This = (PSAXCALLBACKS) Sax ;
    PIHEAP Heap = This->Heap;
    
    TRACE("SAXCallbacksRelease");
   
    Refs = AtomicDec(&This->Refs);
    if (Refs > 0) return (Refs);
    assert(This->Refs == 0);

    //Release any memory associated with the structure pointers.
    Heap->v->Free(Heap, 0, (PTR) This);
    return 0;
}

PRIVATE SCODE MCT
SAXTypeInfoStartDocument(PISAXHANDLER Sax)
{
    TRACE("SOAPSAXCallbacksStartDocument");
    return S_OK;
}


PRIVATE SCODE MCT
SAXTypeInfoEndDocument(PISAXHANDLER Sax)
{
    TRACE("SOAPSAXCallbacksEndDocument");
    
    return S_OK;
}

PRIVATE SCODE MCT
TypeInfoUpdateElementName(PSAXCALLBACKS Sax_Obj, char * Prefix, UINT PrefixLength,
                          char *ElementName, UINT ElementNameLength)
//TypeInfoUpdateElementName(PSAXCALLBACKS Sax_Obj, char * name, UINT Length)
{
    UINT len = 0;
    PIHEAP Heap = Sax_Obj->Heap;
    
    //TRACE("typeinfo.c function : UpdateElementName");

    if (Sax_Obj->Elem_or_Attr != NULL) {
        Heap->v->Free(Heap, 0, (PTR) Sax_Obj->Elem_or_Attr);
        Sax_Obj->Elem_Length = 0;
    }

    Sax_Obj->Elem_or_Attr = Heap->v->Alloc(Heap, 0, (PrefixLength + ElementNameLength + 2)* sizeof(char), 0);

    if( Sax_Obj->Elem_or_Attr == NULL) {   
        DBG(("Error: Insufficient memory\n"));
    } else {
        //DBG(("typeinfoHandler.c Function : UpdateElementName --> Memory Allocation Successful\n"));
        if(Prefix != NULL) {
            if(PrefixLength == 0) PrefixLength = strlen(Prefix);
            while(len < PrefixLength) {
                Sax_Obj->Elem_or_Attr[len] = Prefix[len];
                len++;
            }
            Sax_Obj->Elem_or_Attr[len++] = ':';
            while(len - PrefixLength - 1< ElementNameLength) {
                Sax_Obj->Elem_or_Attr[len] = ElementName[len - PrefixLength - 1];
                len++;
            }
        } else {
            while(len < ElementNameLength) {
                Sax_Obj->Elem_or_Attr[len] = ElementName[len];
                len++;
            }
        }
    }
    Sax_Obj->Elem_or_Attr[len] = '\0';
    Sax_Obj->Elem_Length = len;
    return S_OK;
}

PRIVATE SCODE MCT
SAXTypeInfoStartElement(PISAXHANDLER Sax, char * Prefix, UINT PrefixLength,
                        char * ElementName, UINT ElementNameLength)
                        
//SAXTypeInfoStartElement(PISAXCALLBACKS Sax, char * Name, UINT Length,
  //                      UINT NameSpaceDeclared, char* pElement)
{

    PSAXCALLBACKS Sax_Object = pSAXCallbacks(Sax);
    
    //TRACE("SOAPSAXCallbacksStartElement");
    
    TypeInfoUpdateElementName(Sax_Object, Prefix, PrefixLength,
                              ElementName, ElementNameLength);
    return S_OK;
}

/* After gettign all the typeinfo for the method, do some post-processing */
PRIVATE void FillMethod(PMETHOD_DESCR metd, UINT MyIndex)
{
    UINT argix, Offset;
    UINT16 SaveCount;
    PARGUMENT_DESCR argd;
    ARGUMENT_DESCR_FLAGS Flags;

    assert(MyIndex <= _UI8_MAX - 3);
    metd->Ordinal = (UINT8) MyIndex + 3; /* XXX 3 for IUnknown */

    Offset = sizeof(ADDRESS); /* One for this-arg */
    SaveCount = 1; /* One based so as to treat the zero value as "none". */

    for (argix = 0; argix < metd->argcount; argix++) {
        argd = &metd->parglist[argix];
        Flags = argd->Flags;

        /* Figure out where in the frame the method resides */
        /* XXX Take into account alignments and INT64 */
        if (Flags & ARGUMENT_DESCR_FLAGS_RETURN)
            continue;

        if (argd->argtype == ARG_INT64
            && !(Flags & ARGUMENT_DESCR_FLAGS_OUT)) {
#if 0 /* definitely not on i386. XXX parameterize correctly */
            Offset = __ROUND(Offset,sizeof(INT64));
#endif
            argd->FrameOffset = (UINT16) Offset;
            Offset += sizeof(INT64);
        } else {
            argd->FrameOffset = (UINT16) Offset;
            Offset += sizeof(INT);
        }

        if (!(Flags
              & (ARGUMENT_DESCR_FLAGS_SIZE_OF|ARGUMENT_DESCR_FLAGS_LENGTH_OF))
            && (Flags & ARGUMENT_DESCR_FLAGS_OUT))
        {
            argd->SaveIndex = SaveCount;
            SaveCount++;
            /* For arrays add one slot for size */
            if ((argd->LengthIsArg || (Flags & ARGUMENT_DESCR_FLAGS_ARRAY)))
                SaveCount++;
        }
    }
///tmp;
    assert(metd->FrameSize == Offset);
    assert(Offset <= _UI16_MAX);
    metd->FrameSize = (UINT16) Offset;
    assert(SaveCount <= _UI16_MAX);
    metd->SaveCount = (UINT16) SaveCount;
}

PRIVATE void FillStruct(PSTRUCT_DESCR metd)
{
    UINT argix, Offset;
    PARGUMENT_DESCR argd;

    Offset = 0; /* No this here! */

    for (argix = 0; argix < metd->itemCount; argix++) {
        argd = &metd->itemList[argix];

        /* Figure out where in the frame the method resides */
        /* XXX Take into account alignments and INT64 */
        argd->FrameOffset = (UINT16) Offset;
        if (!(argd->Flags & ARGUMENT_DESCR_FLAGS_RETURN)) {
            Offset += sizeof(UINT);
        }
    }

    // XXX Account for nested structs and aligments first...
    //metd->FrameSize = Offset;
}

PRIVATE SCODE MCT
SAXTypeInfoEndElement(PISAXHANDLER Sax, 
                      char *Prefix, UINT PrefixLength, 
                      char *ElementName, UINT NameLength)
{
    PSAXCALLBACKS Sax_Object = pSAXCallbacks(Sax);
    char** tiStrings = TYPEINFO_ELEMENT_STRINGS;
    PINTERFACE_DESCR ifd = &Sax_Object->ptypeinfo->ifacelist[Sax_Object->currentinterface];
    PSTRUCT_DESCR structd = &Sax_Object->ptypeinfo->structList[Sax_Object->currentStructIndex];

    //TRACE("SAXTypeInfoEndElement");

    if(!strncmp(ElementName, tiStrings[TYPEINFO_STRUCT], NameLength)) {
        FillStruct(structd);
        if( Sax_Object->currentStructIndex + 1 < Sax_Object->ptypeinfo->structCount)
            Sax_Object->currentStructIndex++;
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_ITEM], NameLength)) {
        if (Sax_Object->currentItemIndex + 1 < structd->itemCount)
            Sax_Object->currentItemIndex++;
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_IFACEGROUP], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_IFACE], NameLength)) {
        if( Sax_Object->currentinterface + 1 < Sax_Object->ptypeinfo->interfacecount)
            Sax_Object->currentinterface++;
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_ENVELOPE], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_BODY], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_REPLY], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_METHOD], NameLength)) {
        FillMethod(&ifd->pmethodlist[Sax_Object->indexCurrentMethod],
                     Sax_Object->indexCurrentMethod);
        if(Sax_Object->indexCurrentMethod + 1 < ifd->methodcount)
            Sax_Object->indexCurrentMethod++;
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_ARG], NameLength)) {
        PMETHOD_DESCR metd = &ifd->pmethodlist[Sax_Object->indexCurrentMethod];
        if (Sax_Object->indexCurrentArgument + 1 < metd->argcount)
            Sax_Object->indexCurrentArgument++;
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_STACK], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_FRAME], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_RESPONSE], NameLength)){
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_WSDL_PART], NameLength)){
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_DEFINITIONS],NameLength)){
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_TYPES], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_BINDING], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_PORT_TYPE], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_OPERATION], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_INPUT], NameLength)) {
    } else if(!strncmp(ElementName, tiStrings[TYPEINFO_OUTPUT], NameLength))  {
    } else {
    }
    return S_OK;
}


PRIVATE SCODE MCT   
SAXTypeInfoCharacters(PISAXHANDLER Callbacks, char * Characters, UINT ValueLength)
{
    PSAXCALLBACKS Sax_Object = pSAXCallbacks(Callbacks);
    return TypeInfoProcessAttribute(Sax_Object, 
                                    NULL, 0, 
                                    Sax_Object->Elem_or_Attr, Sax_Object->Elem_Length,
                                    Characters, ValueLength);
}

PRIVATE SCODE MCT
SAXTypeInfoAttribute(PISAXHANDLER Callbacks, 
                     char * Prefix, UINT PrefixLength,
                     char * AttributeName, UINT NameLength,
                     char * AttributeValue, UINT ValueLength)
{
    PSAXCALLBACKS Sax_Object = pSAXCallbacks(Callbacks);

    return TypeInfoProcessAttribute(Sax_Object, 
                                    Prefix, PrefixLength,
                                    AttributeName, NameLength,
                                    AttributeValue, ValueLength);
}

PRIVATE ARGTYPE SetType(char* value, UINT valueLength) {
    if (!strncmp(value, "UINT32", valueLength)
        || !strncmp(value, "UINT", valueLength)) {
        return ARG_INT;
    } else if (!strncmp(value, "PSEUDOFLOAT", valueLength)) {
        return ARG_FLOAT;
    } else if (!strncmp(value, "dateTime", valueLength)) {
        return ARG_DATETIME;
    } else if (!strncmp(value, "BOOL", valueLength)) {
        return ARG_BOOL;
    } else if (!strncmp(value, "char*", valueLength)) {
        return ARG_STRING;
    } else if (!strncmp(value, "INT64", valueLength)) {
        return ARG_INT64;
    } else if (!strncmp(value, "IID", valueLength)) {
        return ARG_GUID;
    } else if (!strncmp(value, "URI", valueLength)) {
        return ARG_URI;
    } else if (!strncmp(value, "void*", valueLength)
               || !strncmp(value, "PTR", valueLength)) {
        return ARG_BLOB;
    } else if (!strncmp(value, "UINT*", valueLength)) {
        return ARG_INT_ARRAY;
    } else if (!strncmp(value, "PSEUDOFLOAT*", valueLength)) {
        return ARG_FLOAT_ARRAY;
    } else if (!strncmp(value, "char**", valueLength)) {
        return ARG_STRING_ARRAY;
    } else if (valueLength > 1 && value[valueLength-1] == '*') {
        return ARG_STRUCT_ARRAY; /* "foo*" */
    }
    return ARG_STRUCT;          /* "foo" */
}

PRIVATE SCODE TypeInfoArgument(PSAXCALLBACKS Sax_Object, PARGUMENT_DESCR argd,
                               char* name, UINT nameLength,
                               char* value, UINT valueLength)
{
    PIHEAP Heap = Sax_Object->Heap;

    if (!strncmp(name, "name", nameLength)) {
        Allocate_And_Copy(Heap, &argd->name, value, valueLength);
        if (valueLength > _UI8_MAX)  /* length to fit into UINT8 */
            return E_BUFFER_TOO_SMALL;
        argd->nameLength = (UINT8) valueLength;
    } else if (!strncmp(name, "type", nameLength)) {
        argd->argtype = SetType(value, valueLength);
        if (argd->argtype == ARG_STRUCT || argd->argtype == ARG_STRUCT_ARRAY) {
            UINT len = valueLength;
            if (argd->argtype == ARG_STRUCT_ARRAY)
                len--;
            argd->structd = FindStruct(Sax_Object->ptypeinfo, value,len, NULL);
            if (!argd->structd)
                return E_ENTRY_NOT_FOUND;
        }
        /* Set ByRef argument as appropriate */
        switch(argd->argtype) {
        case ARG_INT:
        case ARG_FLOAT:
        case ARG_INT64:
        case ARG_BOOL:
        case ARG_THIS:
            break;
        case ARG_INT_ARRAY:
        case ARG_FLOAT_ARRAY:
        case ARG_STRING_ARRAY:
        case ARG_STRUCT_ARRAY:
            argd->Flags |= ARGUMENT_DESCR_FLAGS_ARRAY;
            /*FALLTHROUGH*/
        default:
            /* Represented as pointers */
            argd->Flags |= ARGUMENT_DESCR_FLAGS_BYREF;
            /* XXX Except arrays and structs within structs! (unless...) */
        }
    } else if (!strncmp(name, "direction", nameLength)) {
        if (!strncmp(value, "in", valueLength)) {
            argd->Flags |= ARGUMENT_DESCR_FLAGS_IN;
        } else if (!strncmp(value, "out", valueLength)) {
            argd->Flags |= ARGUMENT_DESCR_FLAGS_OUT;
        } else if (!strncmp(value, "inout", valueLength)) {
            argd->Flags |= ARGUMENT_DESCR_FLAGS_IN |  ARGUMENT_DESCR_FLAGS_OUT;
        } else {
            argd->Flags |= ARGUMENT_DESCR_FLAGS_IN;
        }
    } else if (nameLength == 4 && !strncmp(name, "size", nameLength)) {
        argd->ArraySize = _atoin(value, valueLength);
    } else if (!strncmp(name, "size-is-pos", nameLength)) {
        argd->SizeIsArg = _atoin(value, valueLength);
    } else if (!strncmp(name, "length-is-pos", nameLength)) {
        argd->LengthIsArg = _atoin(value, valueLength);
    } else if (!strncmp(name, "is-return-value", nameLength)) {
        if (!strncmp(value, "true", valueLength))
            argd->Flags |= ARGUMENT_DESCR_FLAGS_RETURN;
    } else if (!strncmp(name, "size-of", nameLength)) {
        argd->Flags |= ARGUMENT_DESCR_FLAGS_SIZE_OF;
    } else if (!strncmp(name, "length-of", nameLength)) {
        argd->Flags |= ARGUMENT_DESCR_FLAGS_LENGTH_OF;
    }

    return S_OK;
}

PRIVATE SCODE MCT TypeInfoProcessAttribute(PSAXCALLBACKS Sax_Object,
                                   char * prefix, UINT prefixLength,
                                   char * name, UINT nameLength, 
                                   char * value, UINT valueLength)
{
    //name : Attribute Name ; value : Attribute Value
    PIHEAP Heap = Sax_Object->Heap;
    char** tiStrings = TYPEINFO_ELEMENT_STRINGS;
    char* elementname = Sax_Object->Elem_or_Attr;
    PINTERFACE_DESCR ifd = &Sax_Object->ptypeinfo->ifacelist[Sax_Object->currentinterface];
    PSTRUCT_DESCR structd = &Sax_Object->ptypeinfo->structList[Sax_Object->currentStructIndex];
    SCODE sc;
    
    if(strcmp(elementname, tiStrings[TYPEINFO_STRUCT_GROUP]) == 0 ) { 
        if(strncmp(name, "structcount", nameLength) == 0) {
            Sax_Object->ptypeinfo->structCount = _atoin(value, valueLength);
            Sax_Object->ptypeinfo->structList = Heap->v->Alloc(Heap,
                HEAP_ZERO_MEMORY, 
                sizeof(STRUCT_DESCR) * Sax_Object->ptypeinfo->structCount, 0);
            if(Sax_Object->ptypeinfo->structList == NULL)
                return E_NOT_ENOUGH_MEMORY;
            Sax_Object->currentStructIndex = 0;
        }
    } else if(strcmp(elementname, tiStrings[TYPEINFO_STRUCT]) == 0 ) { 
        if(strncmp(name, "name", nameLength) == 0) {
            Allocate_And_Copy(Heap, &structd->name, value, valueLength);
            assert(valueLength <= _UI8_MAX);
            structd->nameLength = (UINT8) valueLength;
        } else if(strncmp(name, "fieldcount", nameLength) == 0) {
            structd->itemCount = _atoin(value, valueLength);
            structd->itemList = Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY, 
                                    sizeof(STRUCT_ITEM_DESCR) * structd->itemCount, 0);
            if(structd->itemList == NULL)
                return E_NOT_ENOUGH_MEMORY;
            Sax_Object->currentItemIndex = 0;
        } else if(strncmp(name, "framesize", nameLength) == 0) {
            structd->FrameSize = _atoin(value, valueLength);
        }
    } else if(strcmp(elementname, tiStrings[TYPEINFO_ITEM]) == 0 ) { 
        PSTRUCT_ITEM_DESCR itemd = &structd->itemList[Sax_Object->currentItemIndex];
        sc = TypeInfoArgument(Sax_Object, itemd, name, nameLength, value, valueLength);
        if (FAILED(sc))
            return sc;
    } else if(strcmp(elementname, tiStrings[TYPEINFO_IFACEGROUP]) == 0 ) {
        if(strncmp(name, "name", nameLength) == 0) {
            Allocate_And_Copy(Heap, &ifd->groupname, value, valueLength);
            assert(valueLength <= _UI8_MAX);
            ifd->groupnamelength = (UINT8) valueLength;
        } if(strncmp(name, "interfacecount", nameLength) == 0) {
            Sax_Object->ptypeinfo->interfacecount = _atoin(value, valueLength);
            Sax_Object->currentinterface = 0;
            Sax_Object->ptypeinfo->ifacelist = Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY, 
                                    sizeof(struct _INTERFACE_DESCR) * Sax_Object->ptypeinfo->interfacecount, 0);
            if(Sax_Object->ptypeinfo->ifacelist == NULL)
                return E_NOT_ENOUGH_MEMORY;
        }
    } else if(strcmp(elementname, tiStrings[TYPEINFO_IFACE]) == 0 ) {
        if(strncmp(name, "name", nameLength)==0) {
            Allocate_And_Copy(Heap, &ifd->name, value, valueLength);
            assert(valueLength <= _UI8_MAX);
            ifd->length = (UINT8) valueLength;
        } else if (strncmp(name, "iid", nameLength) == 0) {
            ParseIid(value, &ifd->iid);
        } else if (strncmp(name, "methodcount", nameLength) == 0) {
            ifd->methodcount = _atoin(value, valueLength);
            ifd->pmethodlist = Heap->v->Alloc(Heap, 0,
                           sizeof(struct _METHOD_DESCR) * ifd->methodcount, 0);
            memset(ifd->pmethodlist, 0,
                   sizeof(struct _METHOD_DESCR) * ifd->methodcount);
            Sax_Object->indexCurrentMethod = 0;
        } else {
        }
#if 0
    } else if(strcmp(elementname, tiStrings[TYPEINFO_ENVELOPE]) == 0 ) {
        if(strncmp(name, "namespace", nameLength)==0) {
            Allocate_And_Copy(Heap, &ifd->envelope.nameSpace, value, valueLength);
            ifd->envelope.nameSpaceLength= valueLength;
        } else if (strncmp(name, "namespaceURL", nameLength)==0) {
            Allocate_And_Copy(Heap, &ifd->envelope.nameSpaceURL, value, valueLength);
            ifd->envelope.nameSpaceURLLength = valueLength;
        } else if (strncmp(name, "encodingstyle", nameLength)==0) {
            Allocate_And_Copy(Heap, &ifd->envelope.encodingStyle,value, valueLength);
            ifd->envelope.encodingStyleLength = valueLength;
        } else {
        }
    } else if(strcmp(elementname, tiStrings[TYPEINFO_BODY]) == 0 ) {
        if(strncmp(name, "namespace", nameLength)==0) {
            Allocate_And_Copy(Heap, &ifd->body.nameSpace, value, valueLength);
            ifd->body.nameSpaceLength= valueLength;
        } else if (strncmp(name, "namespaceURL", nameLength)==0) {
            Allocate_And_Copy(Heap, &ifd->body.nameSpaceURL, value, valueLength);
            ifd->body.nameSpaceURLLength = valueLength;
        } else if (strcmp(name, "encodingstyle")==0) {
            Allocate_And_Copy(Heap, &ifd->body.encodingStyle, value, valueLength);
            ifd->body.encodingStyleLength = valueLength;
        } else {
        }
    } else if(strcmp(elementname, tiStrings[TYPEINFO_REPLY]) == 0 ) {
        //Is this necessary?
        if(strncmp(name, "namespace", nameLength)==0) {
            Allocate_And_Copy(Heap, &ifd->reply.nameSpace, value, valueLength);
            ifd->reply.nameSpaceLength= valueLength;
        } else if (strncmp(name, "replyname", nameLength)==0) {
            Allocate_And_Copy(Heap, &ifd->reply.methodName, value, valueLength);
            ifd->reply.methodNameLength = valueLength;
        } else if (strncmp(name, "encodingstyle", nameLength)==0) {
            Allocate_And_Copy(Heap, &ifd->reply.encodingStyle, value, valueLength);
            ifd->reply.encodingStyleLength = valueLength;
        } else {
        }
#endif
    } else if(strcmp(elementname, tiStrings[TYPEINFO_METHOD]) == 0 ) {
        PMETHOD_DESCR metd = &ifd->pmethodlist[Sax_Object->indexCurrentMethod];
        if(strncmp(name, "name", nameLength)==0) {
            Allocate_And_Copy(Heap, &metd->name, 
                              value, valueLength);
            assert(valueLength <= _UI8_MAX);
            metd->length = (UINT8) valueLength;
        } else if(strncmp(name, "responsename", nameLength)==0) {
            Allocate_And_Copy(Heap, &metd->responsename, value, valueLength);
            assert(valueLength <= _UI8_MAX);
            metd->responselength = (UINT8) valueLength;
        } else if (strncmp(name, "argcount", nameLength)==0) {
            PARGUMENT_DESCR argd;
            UINT size;

            metd->argcount= _atoin(value, valueLength);

            size = sizeof(struct _ARGUMENT_DESCR) * metd->argcount;
            argd = Heap->v->Alloc(Heap, 0, size, 0);
            memset(argd, 0, size);

            metd->parglist = argd;
            Sax_Object->indexCurrentArgument = 0;
        } else {
        }
    } else if(strcmp(elementname, tiStrings[TYPEINFO_ARG]) == 0 ) {
        PMETHOD_DESCR metd = &ifd->pmethodlist[Sax_Object->indexCurrentMethod];
        PARGUMENT_DESCR argd = &metd->parglist[Sax_Object->indexCurrentArgument];
        sc = TypeInfoArgument(Sax_Object, argd, name, nameLength, value, valueLength);
        if (FAILED(sc))
            return sc;
    } else if(strcmp(elementname, tiStrings[TYPEINFO_STACK]) == 0 ) {
        PMETHOD_DESCR metd = &ifd->pmethodlist[Sax_Object->indexCurrentMethod];
        if(strncmp(name, "size", nameLength)==0) 
            metd->stacksize = _atoin(value, valueLength);
    } else if(strcmp(elementname, tiStrings[TYPEINFO_FRAME]) == 0 ) {
        PMETHOD_DESCR metd = &ifd->pmethodlist[Sax_Object->indexCurrentMethod];
        if(strncmp(name, "size", nameLength)==0)
            metd->FrameSize = _atoin(value, valueLength);
    } else if(strcmp(elementname, tiStrings[TYPEINFO_RESPONSE]) == 0 ) {
        TRACE("TYPEINFO_RESPONSE no longer supported");
#if 0
        PMETHOD_DESCR metd = &ifd->pmethodlist[Sax_Object->indexCurrentMethod];
        if(strncmp(name, "name", nameLength)==0) {
            Allocate_And_Copy(Heap, &metd->response.ResponseName, 
                              value, valueLength);
            metd->response.ResponseLength = valueLength;
        } else if (strncmp(name, "type", nameLength) == 0) {
            if (strncmp(value, "UINT32", valueLength) == 0) {
                metd->response.ResponseType = RES_INTEGER;    
            } else if (strncmp(value, "char*", valueLength) == 0) {
                metd->response.ResponseType = RES_STRING;    
            } else {
            }
        } else {
        }
#endif 
    } else if(strcmp(elementname, tiStrings[TYPEINFO_WSDL_PART]) == 0 ) {
    } else if(strcmp(elementname, tiStrings[TYPEINFO_DEFINITIONS]) == 0 ) {
    } else if(strcmp(elementname, tiStrings[TYPEINFO_TYPES]) == 0 ) {
    } else if(strcmp(elementname, tiStrings[TYPEINFO_BINDING]) == 0 ) {
    } else if(strcmp(elementname, tiStrings[TYPEINFO_OPERATION]) == 0 ) {
    } else if(strcmp(elementname, tiStrings[TYPEINFO_INPUT]) == 0 ) {
    } else if(strcmp(elementname, tiStrings[TYPEINFO_OUTPUT]) == 0 ) {
    } else {
    }

    return S_OK;
}

PRIVATE SCODE MCT   SAXTypeInfoEndInstruction(PISAXHANDLER piscs, char * InstructionName, UINT InNameLen)
{
    TRACE("SAXCallbacksEndInstruction");
    return 0;
}

PRIVATE SCODE MCT   SAXTypeInfoStartInstruction(PISAXHANDLER piscs, char * InName , UINT InNameLen)
{
    TRACE("SAXCallbacksStartInstruction");
    return 0;
}

const struct ISaxHandlerVtbl  SAXTypeInfoVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    SAXTypeInfoQueryInterface,
    SAXTypeInfoAddRef,
    SAXTypeInfoRelease,
    SAXTypeInfoStartDocument,
    SAXTypeInfoEndDocument,
    SAXTypeInfoStartElement,
    SAXTypeInfoEndElement,
    SAXTypeInfoCharacters,
    SAXTypeInfoAttribute,
    SAXTypeInfoStartInstruction,
    SAXTypeInfoEndInstruction
};
