/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <string.h>
#include <limits.h>
#include <stdlib.h>
#include <mmlite.h>
#include <mmmacros.h>
#include <netif.h>
#include <tokenizerif.h>
#include <winsock.h>
#include <net/mbuf.h>
#include <heaps/dbg_heap.h>

#include <tokenizerutils.h>

/* Tokenizer concrete classe */

typedef struct mbuf BUFFER, * PBUFFER;

/* Manipulation of Content and LookAhead is somewhat spooky. 
 *
 * First and Last are primarily kept for memory management.
 * The content buffers are kept in a single-linked list, 
 * starting with First and proceeding (through ->Next) to 
 * Last.
 *
 * Each tokenizer can only keep track of one traversal. This 
 * is by design. Currently no more than one traversal can be 
 * done through a sequence of mbufs, but with proper reference 
 * counting of mbufs there is nothing that prevents more than 
 * one tokenizer being active in the sequence at a given time.
 * Don't know if that would be useful for anything though.
 */
typedef struct _Tokenizer {
    const struct ITokenizerVtbl* v;
    UINT            Refs;
    UINT64          Position; /* File offset when Source is an IFile */
    PIUNKNOWN       Source;
    PBUFFER         First;

#define CONTENT_BUFFER_COUNT 4
    struct {
        /* First one is special: keeps data around. Anything before is dead */
        PBYTE           Content;    /* keep data up to here, block start */
        PBUFFER         ContentBuffer;
        INT             ContentLength;
    } LookAhead [CONTENT_BUFFER_COUNT];

    BOOL            FiniteContent; /* Length set per protocol spec */
    PIHEAP          Heap;
    enum Encoding {ENC_8BIT, ENC_UTF8,
                   ENC_UTF16_LITTLE, ENC_UTF16_BIG} Encoding;
} Tokenizer, TokenizerFile, *PTOKENIZER;

#include <s_tokenizer_v.c>
#define TokenizerFileDestroy NULL
#include <s_tokenizerfile_v.c>

/* This macro takes a pointer to either the Content or the 
 * LookAhead pointer and changes it into the corresponding 
 * buffer. (the buffer is the next item in the struct) */
#define PointerToBuffer(PointerToPointer) \
    ((PBUFFER*) (PointerToPointer + 1))
/* Same for ContentLength */
#define PointerToLength(PointerToPointer) \
    ((INT*) (PointerToPointer + 2))

#define TOKENIZER_BUFFER_SIZE 0x1000

SCODE MCT       TokenizerEndpointBufferCreate(PITOKENIZER);
SCODE MCT       TokenizerFileBufferCreate(PITOKENIZER);

/* IUnknown implementation */
PRIVATE SCODE MCT TokenizerQueryInterface(
    PITOKENIZER This,
    REFIID Iid,
    void**pResult)
{
    return GenericQueryInterface((PIUNKNOWN) This, Iid,
                                 pResult, &IID_ITokenizer);
}

PRIVATE UINT MCT TokenizerAddRef(PITOKENIZER Tokenizer)
{
    UINT RefCnt;
    PTOKENIZER This = (PTOKENIZER) Tokenizer;

    RefCnt = AtomicInc(&This->Refs);
    return RefCnt;
}

PRIVATE SCODE TokenizerDestroy(IUnknown *iThis)
{
    PTOKENIZER This = (PTOKENIZER) iThis;
    PIHEAP Heap = This->Heap;

    RELEASE((PIBUFFER) This->First);
    RELEASE(This->Source);

    Heap->v->Free(Heap, 0, (PTR) This);
    _PutHeap(Heap);

    return S_OK;
}

PRIVATE UINT MCT TokenizerRelease(PITOKENIZER Tokenizer)
{
    UINT Refs;
    PTOKENIZER This = (PTOKENIZER) Tokenizer;

    Refs = AtomicDec(&This->Refs);
    if (Refs == 0)
        TokenizerDestroy((IUnknown *) This);

    return Refs;
}


/* Tokenizer implementation */

/*
 * Pre:  LookAhead == This->ContentLimit && LookAhead <  This->BufferLimit.
 *       LookAhead is not at the end of the request */
PRIVATE SCODE MCT TokenizerLoad(PITOKENIZER iThis) /* Endpoint Load */
{
    METHOD_DECLARE(Tokenizer, Load);
    PIENDPOINT  ep          = (PIENDPOINT) This->Source;
    PIBUFFER    NewBuffer   = NULL;
#if 1
    TRACE("TokenizerEndpointLoad");

    /* The size can be anything here */
    sc = ep->v->RecvBuffer(ep, &NewBuffer, TOKENIZER_BUFFER_SIZE,
                           0,0,0,0);
    CHECKSC("ReceiveBuffer");

#if defined(_DEBUG) && defined(i386) && !defined(_MINIMIZE)
    /* Spit out the serialized data into file for examination */
    if (NewBuffer) {
        PIFILE file;
        UINT wrote;
        sc = BindToObject(CurrentNameSpace(), _T("fs/httpconn-in.txt"),
                          NAME_SPACE_CREATE | NAME_SPACE_WRITE,
                          &IID_IFile, (void **) &file);
        assert(SUCCEEDED(sc));
        if (SUCCEEDED(sc)) {
            NewBuffer->v->ReadWriteFile(NewBuffer, file, TRUE,
                                        UINT64_MAX, 0, UINT_MAX, &wrote);
            file->v->Release(file);
        }
        sc = S_OK;
    }
#endif

#else
    UINT size;
    TRACE("TokenizerEndpointLoad");

    /* Allocate the new content buffer */
    sc = ep->v->AllocateBuffer(ep, UINT_MAX, &NewBuffer);
    CHECKSC("AllocBuffer");
    assert(((PBUFFER)NewBuffer)->m_hdr.mh_next == NULL);

    size = M_TRAILINGSPACE(((PBUFFER)NewBuffer));
    size = __min(size, This->LookAhead[0].ContentLength);

    /* Pour in the content */
    sc = ep->v->RecvFrom(ep, 
                         ((PBUFFER)NewBuffer)->m_hdr.mh_data, 
                         size,
                         (UINT32*) &size,
                         0, NULL, 0, NULL);
    CHECKSC("RecvFrom");
    ((PBUFFER)NewBuffer)->m_hdr.mh_len = size;

    if (sc == S_FALSE && size != 0)
        sc = S_OK;
#endif

    /* The new buffer is the first one... */
    if (This->First == NULL) {
        This->First = (PBUFFER) NewBuffer;
        NewBuffer = NULL;
    } else if (NewBuffer) { /* ... or we plug it into the existing chain */
        PIBUFFER f = (PIBUFFER) This->First;
        sc = f->v->Append(f, NewBuffer);
        CHECKSC("Append");
    }

METHOD_CLEANUP;
    RELEASE(NewBuffer); 
    if (sc == STATUS_WIN32_ERROR(WSAETIMEDOUT)) /* messy error code */
        sc = E_TIMED_OUT;
    if (sc == STATUS_WIN32_ERROR(WSAEWOULDBLOCK)) { /* messy error code */
        if (!This->First) {
            /* Allocate empty buffer so as to make tokenizer useable */
            sc = ep->v->AllocateBuffer(ep, 0, &NewBuffer);
            CHECKSC("AllocBuffer");
            This->First = (PBUFFER) NewBuffer;
        }
        sc = E_WOULD_BLOCK;
    }
    return sc;
}


PRIVATE SCODE MCT TokenizerFileLoad(PITOKENIZER iThis)
{
    METHOD_DECLARE(Tokenizer, FileLoad);
    PIBUFFER     NewBuffer   = NULL;
    PIENDPOINTFACTORY ef;
    UINT size = TOKENIZER_BUFFER_SIZE;

    //TRACE("TokenizerFileBufferCreate");
    if (!This->Source)          /* IBuffer: The input is used directly */
        return S_FALSE;

    /* Create IBuffer and dump some file contents in there */
    sc = BindToObject(CurrentNameSpace(), _T("COB/protocol.cob"),
                      NAME_SPACE_READ, &IID_IEndpointFactory, (void**) &ef);
    CHECKSC("bind to protocol.cob");

    sc = ef->v->AllocateBuffer(ef, TOKENIZER_BUFFER_SIZE, NULL, &NewBuffer);
    RELEASE(ef);
    CHECKSC("AllocateBuffer");

    sc = NewBuffer->v->ReadWriteFile(NewBuffer, (PIFILE) This->Source, FALSE,
                                     This->Position, 0, size, &size);
    CHECKSC("Read to buffer");

    /* Next time read from where we got */
    This->Position = Uint64AddUint32(This->Position, size);

    if (size != 0)
        sc = S_OK;
    if (sc == S_FALSE)
        goto Out;

    /* The new buffer is the first one... */
    if (This->First == NULL) {
        This->First = (PBUFFER) NewBuffer;
        NewBuffer = NULL;
    } else { /* ... or we plug it into the existing chain */
        PIBUFFER f = (PIBUFFER) This->First;
        sc = f->v->Append(f, NewBuffer);
        CHECKSC("Append");
    }

METHOD_CLEANUP;
    RELEASE(NewBuffer); 
    return sc;
}


/* The rest should be independent of file or endpoint */


PRIVATE void MCT TokenizerBufferDestroy(PITOKENIZER Tokenizer)
{
    PTOKENIZER  This    = (PTOKENIZER) Tokenizer;
    PBUFFER     Dead    = NULL;
    UINT i;

    //TRACE("TokenizerBufferDestroy");
#if 1
    while (This->First != This->LookAhead[0].ContentBuffer) {
        Dead = This->First;
        This->First = Dead->m_hdr.mh_next;
        //DBG(("TokenizerBufferDestroy:\n"));
        //DBG(("\tFreeing content buffer at %x\n", (ADDRESS) Dead));
        //DBG(("\tFreeing content at %x\n", (ADDRESS) Dead->m_hdr.mh_data));

        Dead->m_hdr.mh_next = NULL;
        ((PIBUFFER)Dead)->v->Release((PIBUFFER) Dead);

        for (i = 1; i < CONTENT_BUFFER_COUNT; i++)
            if (This->LookAhead[i].ContentBuffer == Dead) {
                This->LookAhead[i].ContentBuffer = NULL;
                This->LookAhead[i].Content = NULL;
            }
    }

    if (This->LookAhead[0].ContentBuffer == NULL)
        This->First = NULL;
#endif
}


PRIVATE PIUNKNOWN MCT TokenizerGetSource(PITOKENIZER Tokenizer)
{
    return ((PTOKENIZER) Tokenizer)->Source;
}


PRIVATE PBYTE* MCT TokenizerGetContent(PITOKENIZER Tokenizer)
{
    return &(((PTOKENIZER) Tokenizer)->LookAhead[0].Content); 
}


PRIVATE PBYTE* MCT TokenizerGetLookAhead(PITOKENIZER Tokenizer)
{
    return &(((PTOKENIZER) Tokenizer)->LookAhead[1].Content); 
}

PRIVATE PBYTE* MCT TokenizerGetLookAhead2(PITOKENIZER Tokenizer)
{
    return &(((PTOKENIZER) Tokenizer)->LookAhead[2].Content); 
}

PRIVATE PBYTE* MCT TokenizerGetLookAhead3(PITOKENIZER Tokenizer)
{
    return &(((PTOKENIZER) Tokenizer)->LookAhead[3].Content); 
}

PRIVATE void MCT TokenizerSetIterator(PITOKENIZER Tokenizer,
                              const PBYTE* Source, PBYTE* Target)
{
    *Target = *Source;
    *(PointerToBuffer(Target)) = *(PointerToBuffer(Source));
    *(PointerToLength(Target)) = *(PointerToLength(Source));    
}


/* Calculate distance from one to the other. LookAhead better be ahead. */
PRIVATE SCODE MCT TokenizerDistance(PITOKENIZER Tokenizer, 
                            PBYTE* Content, PBYTE* LookAhead,
                            UINT* Distance)
{
    PBUFFER     TmpBuffer          = *(PointerToBuffer(Content));
    PBUFFER*    LookAheadBuffer    = PointerToBuffer(LookAhead);
    PBYTE       Tmp                = *Content;

    *Distance = 0;

    while (TmpBuffer != *LookAheadBuffer) {

        /* Add the distance fraction */
        if (*Distance == 0) 
            *Distance += TmpBuffer->m_hdr.mh_len 
                         - (Tmp - TmpBuffer->m_hdr.mh_data);
        else *Distance += TmpBuffer->m_hdr.mh_len;
        
        /* Jump to the next buffer */
        TmpBuffer = TmpBuffer->m_hdr.mh_next;
        Tmp = TmpBuffer->m_hdr.mh_data;
    }

    *Distance += *LookAhead - Tmp;

    return S_OK;
}

/* Delete chunk out of buffer. Return chunk in Slice if requested.
 * Content must be the leading lookahead.
 * If _Which_ is Content, discard anything before.
 * _Which_ may not be LookAhead3 since it is used here.
 * Any lookaheads after the split will become invalid.
 */
PRIVATE SCODE MCT TokenizerCarveBuffer(PITOKENIZER iThis, 
                               PBYTE* Which, UINT Length,
                               PIBUFFER* Slice)
{
    METHOD_DECLARE(Tokenizer, CarveBuffer);
    PBUFFER m;
    PIBUFFER b1, b2, b3;
    UINT i, dist;
    PBYTE* LookAhead3 = iThis->v->GetLookAhead3(iThis);
    PBYTE* Content = iThis->v->GetContent(iThis);

    CHECKSETSC(Which != LookAhead3, E_INVALID_PARAMETER,
               " - incorrect iterator");

    /* Clean up any stale data from the very beginning */
    m = *(PointerToBuffer(Content));
    dist = *Content - m->m_hdr.mh_data;
    b1 = (PIBUFFER) m;
    b1->v->Truncate(b1, dist);
    TokenizerBufferDestroy(iThis); /* Don't leak dead leading buffers */

    if (Which) {
        /* Treat UINT_MAX as special value indicating "as much as possible" */
        if (Length == UINT_MAX) {
            INT *pLength = PointerToLength(Which);
            Length = *pLength;
        }
    } else {
        /* NULL iterator: Truncate from end */
        INT Amount = Length;
        b1->v->Truncate(b1, -Amount);
        return S_OK;
    }

    /* Pull in enough bytes to tail */
    iThis->v->SetIterator(iThis, Which, LookAhead3);
    sc = iThis->v->AdvanceSymbol(iThis, LookAhead3, Length, NULL);
    CHECKSC("Advance buffer length");

    /* If this is not the leading lookahead we may need a three-way split,
     * or reverse split.
     */

    sc = iThis->v->Distance(iThis, Content, Which, &dist);
    CHECKSC("Distance");

    if (dist) { /* Leading non-carved data */
        b2 = b1->v->Split(b1, dist);
    } else {
        b2 = b1;
        b1 = NULL;
    }

    /* b1 is head, b2 is carved, b3 is tail */

    if (Slice) {
        b3 = b2->v->Split(b2, Length);
    } else {
        b2->v->Truncate(b2, Length);
        b3 = b2;
        b2 = NULL;
    }

    /* Put head and new tail back together */
    if (b1 && b3) {
        PBYTE old = *Which;

        sc = b1->v->Append(b1, b3);
        CHECKSC("Append");
        b3->v->Release(b3);

        /* Content cannot have moved but the lookahead may have */
        iThis->v->SetIterator(iThis, Content, Which);
        sc = iThis->v->AdvanceSymbol(iThis, Which, dist, NULL);
        CHECKSC("Advance buffer length");

        /* Copy to other lookaheads as needed */
        for (i = 0; i < CONTENT_BUFFER_COUNT; i++) {
            if (This->LookAhead[i].Content == old) {
                This->LookAhead[i].Content = *Which;
                This->LookAhead[i].ContentBuffer = *(PointerToBuffer(Which));
                This->LookAhead[i].ContentLength = *(PointerToLength(Which));
            }
        }

    } else if (b3) {
        /* Since this was the beginning we can set all the iterators to
         * point to the new beginnig (some still zeroed below as intended).
         */
        m = (PBUFFER) b3;
        This->First = (PBUFFER) b3;
        for (i = 0; i < CONTENT_BUFFER_COUNT; i++) {
            This->LookAhead[i].Content = m->m_hdr.mh_data;
            This->LookAhead[i].ContentBuffer = m;
            /* The length of LookAhead3 is correct, use it for all */
            This->LookAhead[i].ContentLength = *(PointerToLength(LookAhead3));
        }
    } else {
        assert(FALSE);
    }

#ifdef _DEBUG
    /* Make sure noone uses the old trailing ones */
    for (i = 1; i < CONTENT_BUFFER_COUNT; i++) {
        if (Which >= &This->LookAhead[i].Content)
            continue;
        This->LookAhead[i].Content = NULL;
        This->LookAhead[i].ContentBuffer = NULL;
    }
#endif

    if (Slice)
        *Slice = b2;

METHOD_CLEANUP;
    return sc;
}

/* This method has to work equally well for both the Content and 
 * LookAhead iterators */
PRIVATE SCODE MCT TokenizerAdvanceSymbol(PITOKENIZER Tokenizer,
                                         PBYTE* Iterator, 
                                         UINT Steps, UINT *AdvancedSteps)
{
    SCODE           sc = S_OK;
    PBUFFER*        Buffer      = PointerToBuffer(Iterator), m;
    INT*            pLength     = PointerToLength(Iterator);
    UINT            Cnt, Follows;
    PTOKENIZER  This    = (PTOKENIZER) Tokenizer;

    if ((INT) Steps > *pLength) { /* Don't overrun end of message */
        if(AdvancedSteps)
            *AdvancedSteps = 0;
        return E_END_OF_DATA;
    }

    Cnt = 0;
    m = *Buffer;
    for (;;) {
        Follows = m->m_hdr.mh_data + m->m_hdr.mh_len - *Iterator;
        if (Steps < Follows) {
            *Iterator += Steps;
            Cnt += Steps;
            *pLength -= Steps;
            Steps = 0;
        } else {
            *Iterator += Follows;
            Cnt += Follows;
            Steps -= Follows;
            *pLength -= Follows;
            Follows = 0;
        }

        /* Are we at the end of the buffer? */
        if (Follows == 0 && *pLength > 0) {
            /* Has the next buffer been constructed already? */
            if (m->m_hdr.mh_next == NULL) {
                sc = Tokenizer->v->Load(Tokenizer);
                if (FAILED(sc) || sc == S_FALSE)
                    break;
                continue;     /* Load might have extended current */
            }

            /* Jump over to the next buffer */
            m = m->m_hdr.mh_next;
            *Iterator = m->m_hdr.mh_data;

            /* Do any previous buffers need to be deallocated? */
            TokenizerBufferDestroy(Tokenizer);
            continue;
        }
        if (Steps == 0)
            break;
    }

    *Buffer = m;

    if (AdvancedSteps)
        *AdvancedSteps = Cnt;
    return sc;
}

PRIVATE SCODE MCT TokenizerNextToken(PITOKENIZER Tokenizer,
                             PBYTE* Content, 
                             const PBYTE Delim, UINT NumDelims,
                             UINT* Length)
{
    SCODE sc; 
    UINT cnt;
    BYTE found = 0;

    *Length = 0;
    while (1) {
        for (cnt = 0; cnt < NumDelims; ++cnt)
            if (IsASCII(*Content, Delim[cnt])) 
                found = 1;
        if (found) break;
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, Content, 1, NULL);
        if (FAILED(sc) || sc == S_FALSE) {
            *Length = 0;
            return sc;
        }
        ++*Length;
    }

    return S_OK;
}

PRIVATE SCODE MCT TokenizerNextTokenSequence(PITOKENIZER Tokenizer,
                             PBYTE* Content, PBYTE* LookAhead,
                             const PBYTE Delim, UINT NumDelims,
                             UINT* Length)
{
    METHOD_DECLARE_NOTHIS(Tokenizer, NextTokenSequence);
    UINT d;

    assert(NumDelims > 0);

    *Length = 0;
    sc = S_OK;

    for (;;) {
        d = 0;

        if (IsASCII(*Content, Delim[d])) {
            Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead);
            do {
                d++;
                sc = Tokenizer->v->AdvanceSymbol(Tokenizer, LookAhead, 1, NULL);
                if (sc == S_FALSE)
                    sc = E_END_OF_DATA;
                CHECKSC("AdvanceLA");
                if (d == NumDelims)
                    goto Out;   /* Success! */
            } while (IsASCII(*LookAhead, Delim[d]));
        }
        sc = Tokenizer->v->AdvanceSymbol(Tokenizer, Content, 1, NULL);
        if (sc == S_FALSE)
            sc = E_END_OF_DATA;
        CHECKSC("AdvanceC");        
        ++*Length;
    }

METHOD_CLEANUP;
    return sc;
}

_TCHAR * utf8ncpytot(_TCHAR * pDst, const BYTE *pSrc, size_t Count)
{
#ifdef _UNICODE
    _TCHAR * r = pDst;
    UINT c, c2, c3, c4, c5, c6;

    for ( ; Count-- > 0 &&  *pSrc != _T('\0'); pSrc++, pDst++) {
        /* Decode the awful mess that somehow got standardized... UTF-8 */
        c = *pSrc;
        if (c <= 127)           /* 1 byte */
            goto gotit;
        if (c < 192)            /* error */
            goto Error;
        c2 = *pSrc++;
        if ((c2 & 0xc0) != 0xc0)
            goto Error;
        if (c <= 223) {  /* 2 bytes */
            c = (c - 192) * 64 + (c2 - 128);
            goto gotit;
        }
        c3 = *pSrc++;
        if ((c3 & 0xc0) != 0xc0)
            goto Error;
        if (c  <= 239) { /* 3 bytes */
            c = (c - 224) * 4096 + (c2 - 128) * 64 + (c3 - 128);
            goto gotit;
        }
        c4 = *pSrc++;
        if ((c4 & 0xc0) != 0xc0)
            goto Error;
        if (c  <= 247) { /* 4 bytes */
            c = (c - 240) * 262144 + (c2 - 128) * 4096 + (c3 - 128) * 64 + (c4 - 128);
            goto gotit;
        }
        c5 = *pSrc++;
        if ((c5 & 0xc0) != 0xc0)
            goto Error;
        if (c  <= 251) { /* 5 bytes */
            c = (c - 248) * 16777216 + (c2 - 128) * 262144 + (c3 - 128) * 4096 + (c4 - 128) * 64 + (c5 - 128);
            goto gotit;
        }
        c6 = *pSrc++;
        if ((c6 & 0xc0) != 0xc0)
            goto Error;
        if (c  <= 253) { /* 6 bytes */
            c = (c - 252) * 1073741824 + (c2 - 128) * 16777216 + (c3 - 128) * 262144 + (c4 - 128) * 4096 + (c5 - 128) * 64 + (c6 - 128);
            goto gotit;
        }
        /* error */
        *pDst = _T('\0');
        break;
    gotit:
        /* Now turn that into UTF-16 */
        if (c < 0x10000) {
            /* Fits into 16 bits */
            *pDst = (_TCHAR) c;
        } else if (c > 0x10FFFF) {
            /* More than 20 bits invalid */
            goto Error;
        } else {
            /* In this case the original string is 4 bytes or more */
            *pDst++ = (_TCHAR)(0xd800 | (c & 0x3f));
            *pDst = (_TCHAR)(0xdc00 | (c >> 10));
        }
    }

    return r;

 Error:
    *pDst = _T('\0');
    return NULL;
#else
    return strncpy(pDst, (const char *)pSrc, Count);
#endif
}

PRIVATE SCODE MCT TokenizerCopyToken(PITOKENIZER Tokenizer,
                             const PBYTE* Content, const PBYTE* LookAhead,
                             _TCHAR *Buffer)
{
    PBYTE           Tmp             = *Content;
    PBUFFER  TmpBuffer       = *(PointerToBuffer(Content));
    PBUFFER  LastBuffer      = *PointerToBuffer(LookAhead);
    UINT m, n;

    if (*Content == *LookAhead) {
        Buffer[0] = _T('\0');
        return S_OK;
    }

    n = 0;
    while (TmpBuffer != LastBuffer) {
        PBYTE TmpBufferEnd = TmpBuffer->m_hdr.mh_data + TmpBuffer->m_hdr.mh_len;
        m = TmpBufferEnd - Tmp;
        utf8ncpytot(Buffer + n, Tmp, TmpBufferEnd - Tmp);
        TmpBuffer = TmpBuffer->m_hdr.mh_next;
        Tmp = TmpBuffer->m_hdr.mh_data;
        n += m;
    }

    assert(TmpBuffer == LastBuffer);

    m = *LookAhead - Tmp;
    utf8ncpytot(Buffer + n, (const BYTE *) Tmp, m);
    (Buffer + n)[m] = _T('\0');
    
    return S_OK;
}


PRIVATE SCODE MCT TokenizerCopyTokenA(PITOKENIZER Tokenizer,
                             const PBYTE* Content, const PBYTE* LookAhead,
                             char *Buffer)
{
    PBYTE           Tmp             = *Content;
    PBUFFER  TmpBuffer       = *(PointerToBuffer(Content));
    PBUFFER  LastBuffer      = *PointerToBuffer(LookAhead);
    UINT m, n;

    if (*Content == *LookAhead) {
        Buffer[0] = '\0';
        return S_OK;
    }

    n = 0;
    while (TmpBuffer != LastBuffer) {
        PBYTE TmpBufferEnd = TmpBuffer->m_hdr.mh_data + TmpBuffer->m_hdr.mh_len;
        m = TmpBufferEnd - Tmp;
        strncpy((char*) (Buffer + n), (const char*) Tmp, TmpBufferEnd - Tmp);
        TmpBuffer = TmpBuffer->m_hdr.mh_next;
        Tmp = TmpBuffer->m_hdr.mh_data;
        n += m;
    }

    assert(TmpBuffer == LastBuffer);

    m = *LookAhead - Tmp;
    strncpy((char*) (Buffer + n), (const char*) Tmp, m);
    ((char*) (Buffer + n))[m] = '\0';
    
    return S_OK;
}


PRIVATE SCODE MCT TokenizerAllocateAndCopyNextToken(PITOKENIZER Tokenizer,
        const PBYTE* Content, PBYTE* LookAhead,
        const PBYTE Delim, UINT NumDelims,
        _TCHAR** Buffer, UINT* Length)
{
    SCODE       sc;
    SCODE       TokenFound = S_OK;
    PTOKENIZER  This = (PTOKENIZER) Tokenizer;
    PIHEAP      Heap = This->Heap;

    /* Delimit the token */
    Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead);
    
    sc = Tokenizer->v->NextToken(Tokenizer, LookAhead, Delim, NumDelims, Length);
	if (sc == S_FALSE)
		sc = E_END_OF_DATA;
    if (FAILED(sc)) return sc;
    TokenFound = sc;

    /* Allocate space for the token */
    *Buffer = Heap->v->Alloc(Heap,
                             HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("tok:token"),
                             (*Length + 1) * sizeof(_TCHAR), 0);
    if (*Buffer == NULL) {
        *Length = 0;
        return E_NOT_ENOUGH_MEMORY;
    }

        /* Copy token into allocated space */
    sc = Tokenizer->v->CopyToken(Tokenizer, Content, LookAhead, *Buffer);
    if (FAILED(sc)) {
        *Length = 0;
        Heap->v->Free(Heap, 0, Buffer);
        return sc;
    }

    return TokenFound;
}

PRIVATE SCODE MCT TokenizerAllocateAndCopyNextTokenA(PITOKENIZER Tokenizer,
        const PBYTE* Content, PBYTE* LookAhead,
        const PBYTE Delim, UINT NumDelims,
        char** Buffer, UINT* Length)
{
    SCODE       sc;
    SCODE       TokenFound = S_OK;
    PTOKENIZER  This = (PTOKENIZER) Tokenizer;
    PIHEAP      Heap = This->Heap;

    /* Delimit the token */
    Tokenizer->v->SetIterator(Tokenizer, Content, LookAhead);
    
    sc = Tokenizer->v->NextToken(Tokenizer, LookAhead, Delim, NumDelims, Length);
    if (FAILED(sc)) return sc;
    TokenFound = sc;

    /* Allocate space for the token */
    *Buffer = Heap->v->Alloc(Heap,
                             HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("tok:token"),
                             *Length + 1, 0);
    if (*Buffer == NULL) {
        *Length = 0;
        return E_NOT_ENOUGH_MEMORY;
    }

    /* Copy token into allocated space */
    sc = Tokenizer->v->CopyTokenA(Tokenizer, Content, LookAhead, *Buffer);
    if (FAILED(sc)) {
        *Length = 0;
        Heap->v->Free(Heap, 0, Buffer);
        return sc;
    }

    return TokenFound;
}


/* BUGBUG Match not unsigned int */
PRIVATE SCODE MCT TokenizerCmpToken(PITOKENIZER Tokenizer, 
                             const PBYTE* Iterator, const PBYTE External, 
                             UINT TotalToCheck, UINT* Match)
{
    SCODE           sc;
    PBUFFER         Buffer;
    PBYTE           Tmp;
    UINT            TotalChecked;
    UINT            NumInBuffer;

 Again:
    Buffer = *PointerToBuffer(Iterator);
    Tmp = NULL;
    TotalChecked = 0;

    assert(*Iterator >= Buffer->m_hdr.mh_data);
    assert(*Iterator <=  Buffer->m_hdr.mh_data +  Buffer->m_hdr.mh_len);

    Tmp = Buffer->m_hdr.mh_data + Buffer->m_hdr.mh_len;

    *Match = 0;
    NumInBuffer = Tmp - *Iterator;
    Tmp = *Iterator;
    do {
        UINT NumToCheck;
        if (TotalToCheck - TotalChecked <= NumInBuffer) 
            NumToCheck = TotalToCheck - TotalChecked;
        else
            NumToCheck = NumInBuffer;
        if (NumToCheck > 0) {
            *Match = _strnicmp((char*) Tmp, 
                               (const char*) External + TotalChecked, 
                               NumToCheck);
            TotalChecked += NumToCheck;
            if (*Match != 0 || TotalChecked >= TotalToCheck) 
                break;
        }
        if (Buffer->m_hdr.mh_next == NULL) {
            sc = Tokenizer->v->Load(Tokenizer);
            if (sc == S_FALSE)
                sc = E_END_OF_DATA;
            if (FAILED(sc)) { *Match = -1; return sc; }
            /* We might have extended the current buffer. XXX Perf finesse. */
            goto Again;
        }
        Buffer = Buffer->m_hdr.mh_next;
        Tmp = Buffer->m_hdr.mh_data;
        NumInBuffer = Buffer->m_hdr.mh_len;
    } while (1);

    assert(*Match != 0 || TotalChecked == TotalToCheck);

    return (*Match == 0 ? S_OK : S_FALSE);
}

PRIVATE SCODE TokenizerSetContentLength(PITOKENIZER iThis, INT Length)
{
    METHOD_DECLARE(Tokenizer, SetContentLength);
    BOOL WasFinite = This->FiniteContent;
    INT OldLength = This->LookAhead[0].ContentLength;
    PBYTE* Iterator = iThis->v->GetContent(iThis);
    PBUFFER* Buffer = PointerToBuffer(Iterator);

    assert (OldLength >= 0);
    sc = S_OK;

    if (WasFinite && OldLength > 0) {
        /* Drop any unprocessed remains from old packet */
        sc = iThis->v->AdvanceSymbol(iThis, &This->LookAhead[0].Content,
                                     OldLength,
                                     NULL);
        if (sc != E_END_OF_DATA)
            CHECKSC("Zap stale content");
    }

    This->LookAhead[0].ContentLength = Length;
    This->FiniteContent = (Length == INT_MAX) ? FALSE : TRUE;

    /* Make sure we are not left pointing past the buffer */
    sc = iThis->v->AdvanceSymbol(iThis, &This->LookAhead[0].Content, 0, NULL);
    if (sc == E_END_OF_DATA)
        sc = S_FALSE;
    CHECKSC("Advance to next network packet");

    METHOD_CLEANUP;
    return sc;
}

PRIVATE PIHEAP TokenizerGetHeap(PITOKENIZER Tokenizer)
{
    PTOKENIZER  This = (PTOKENIZER) Tokenizer;

    return This->Heap;
}

/*** Tokenizer factory ***/
typedef struct _TokenizerFactory {
    const struct ITokenizerFactoryVtbl* v;
    UINT            Refs;
} TokenizerFactory;

#include <s_tokenizerfactory_v.c>

PRIVATE SCODE MCT TokenizerFactoryQueryInterface(PITOKENIZERFACTORY This,
                                         REFIID Iid, void**pResult)
{
    return GenericQueryInterface((PIUNKNOWN) This, Iid,
                                 pResult, &IID_ITokenizerFactory);
}

PRIVATE UINT MCT TokenizerFactoryAddRef(PITOKENIZERFACTORY Tokenizer)
{
    UINT RefCnt;
    TokenizerFactory * This = (TokenizerFactory *) Tokenizer;

    RefCnt = AtomicInc(&This->Refs);
    return RefCnt;
}

PRIVATE SCODE TokenizerFactoryDestroy(IUnknown *iThis)
{
    return S_OK;
}

PRIVATE UINT MCT TokenizerFactoryRelease(PITOKENIZERFACTORY Factory)
{
    UINT                Refs;
    TokenizerFactory *   This    = (TokenizerFactory *) Factory;

    Refs = AtomicDec(&This->Refs);

    if (Refs > 0)
        return Refs;

    TokenizerFactoryDestroy((IUnknown *) This);
    return 0;
}

PRIVATE SCODE MCT TokenizerFactoryCreate(
    PITOKENIZERFACTORY Factory,
    PIUNKNOWN Source,
    PITOKENIZER* Tokenizer)
{
    SCODE       sc;
    PIHEAP      Heap        = _GetHeap(_T("Tokenizer"), 4000);
    PTOKENIZER  This;
    PIENDPOINT  Endpoint    = NULL;
    PIFILE      File        = NULL;
    PBUFFER m;

    //TRACE("TokenizerCreate");
    UnusedParameter(Factory);

    //DBG(("TokenizerCreate: Allocating tokenizer structure\n"));
    This = Heap->v->Alloc(Heap, HEAP_ZERO_MEMORY |HEAP_DEBUG_NAME("tok"),
                          sizeof(*This), 0);
    if (This == NULL) return E_NOT_ENOUGH_MEMORY;

    This->Heap = Heap;

    assert(This->Source == NULL);
    sc = Source->v->QueryInterface(Source, &IID_IEndpoint, (void**) &Endpoint);
    
    if (SUCCEEDED(sc)) {
        /* Endpoint-specific initialization */
        UINT32 RcvTimeo;

        //DBG(("TokenizerCreate: Endpoint-specific initialization\n"));

        /* Set recv timeout to 35 seconds, the NT way */
        RcvTimeo = 1000000; //35000;
        sc = Endpoint->v->SetSockOpt(Endpoint,
                                     SOL_SOCKET,SO_RCVTIMEO,
                                     (PBYTE)&RcvTimeo,
                                     sizeof RcvTimeo);
        if (FAILED(sc)) {
            Endpoint->v->Release(Endpoint);
            goto err_no_timeo;
        }

        This->v = &TokenizerVtbl;
        This->First = NULL;
        This->Source = (PIUNKNOWN) Endpoint; /* Casting from now on... */
    } else if (sc == E_NO_INTERFACE) {
        sc = Source->v->QueryInterface(Source, &IID_IBuffer, (void**) &File);
        if (SUCCEEDED(sc)) {
            /* IBuffer specific initialization */
            DBG(("TokenizerCreate: IBuffer-specific initialization\n"));

            This->v = &TokenizerFileVtbl;
            This->First = (PBUFFER) File;
            This->Source = NULL;
        } else if (sc == E_NO_INTERFACE) {
            sc = Source->v->QueryInterface(Source, &IID_IFile, (void**) &File);
            if (FAILED(sc)) goto err_no_source;

            /* File specific initialization */
            DBG(("TokenizerCreate: File-specific initialization\n"));

            This->v = &TokenizerFileVtbl;
            This->First = NULL;
            This->Source = (PIUNKNOWN) File; /* Casting from now on... */
        } else goto err_no_source;            
    } else goto err_no_source;

    /* Source-independent initialization */
    This->LookAhead[0].ContentLength = INT_MAX;
    if (This->Source)
        sc = This->v->Load((PITOKENIZER) This);
    if (sc == S_FALSE) sc = E_END_OF_DATA; /* Tokenizer without content no good... */
    if (FAILED(sc) && sc != E_WOULD_BLOCK) goto err_buf_alloc;

    //DBG(("TokenizerCreate: Source-independent initialization\n"));

    m = This->First;
    /* Throw away any empty leading buffers */
    while (!m->m_hdr.mh_len && m->m_hdr.mh_next)
        m = m->m_hdr.mh_next;

    This->LookAhead[0].ContentBuffer = m;
    This->LookAhead[0].Content = m->m_hdr.mh_data;

    This->Refs = 1;

    *Tokenizer = (PITOKENIZER) This;

    //DBG(("TokenizerCreate: Creation successful\n"));

    return S_OK;

err_no_timeo:
err_buf_alloc:
err_no_source:
    RELEASE(This->Source);
    RELEASE((PIBUFFER)This->First);
    Heap->v->Free(Heap, 0, (PTR) This);
    _PutHeap(Heap);
    return sc;
}

/* Tokenizer COB entry point */
PIUNKNOWN 
tokenizerCobMain()
{
    static TokenizerFactory Factory = {&TokenizerFactoryVtbl, 1};

    TRACE("TokenizerCobMain");

    return (PIUNKNOWN) &Factory;
}
