/* Copyright (c) Microsoft Corporation. All rights reserved. */

/*********************************************************************/
/* simbmap.c                                                         */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <mmlite.h>
#include <framebuffer.h>

#include <limits.h>
#include <mmmacros.h>
#include <netif.h>

PRIVATE PIFRAMEBUFFER pDisplay = NULL;
PRIVATE UINT8 *FrameBufferBase = NULL;

SCODE ScreenInit(void)
{
    METHOD_DECLARE_NOTHIS(bmp.c, ScreenInit);
    UINT Size;

    sc = BindToObject(CurrentNameSpace(), _T("display"),
                      NAME_SPACE_READ | NAME_SPACE_WRITE,
                      &IID_IFrameBuffer,
                      (void**)&pDisplay);
    CHECKSC("Bind(display)");

    /* Set it the way we want it
     */
    sc = pDisplay->v->SetMode(pDisplay,FB_MODE_320x200_in_320x400_8bpp_PALETTE);
    CHECKSC("SetMode");

    /* Map it
     */
    Size = 0;
    sc = pDisplay->v->Map(pDisplay,(PTR*)&FrameBufferBase,&Size);
    CHECKSC("Map");

    memset(FrameBufferBase,0,Size);
METHOD_CLEANUP;
    return sc;
}

void ScreenTerminate(void)
{
    if (pDisplay) {
        pDisplay->v->Unmap(pDisplay,FrameBufferBase);
        pDisplay->v->Release(pDisplay);
        pDisplay = NULL;
        FrameBufferBase = NULL;
    }
}

void ScreenBlit(UINT16 w, UINT16 h, UINT16 s, UINT32* Data)
{
    if (FrameBufferBase) {
        memcpy(FrameBufferBase,Data,s*4*h);
    }
}

SCODE ScreenBlit2(PIBUFFER im, UINT off, UINT size)
{
    UINT64 off64;

    Uint32ToUint64(off64, off);

    if (!FrameBufferBase)
        return E_OBJECT_NOT_INITIALIZED;

    return im->v->ReadAt(im, off64, FrameBufferBase, size, &size);
}

#if defined(_MSC_VER)
#pragma pack(1)
#endif

typedef struct tagRGB {
    UINT8    rgbRed;
    UINT8    rgbGreen;
    UINT8    rgbBlue;
} RGB;
typedef RGB * PRGB;



void ScreenPalette( PRGB Data, UINT nEntries)
{
    if (pDisplay) {
        pDisplay->v->SetPalette(pDisplay,
                                (UINT8*)Data,
                                nEntries * 3,
                                PIXEL_TYPE_R8G8B8);
    }
}

/*****************************/

typedef struct tagBITMAPINFOHEADER
    {
    UINT32 biSize;
    UINT32 biWidth;
    UINT32 biHeight;
    UINT16 biPlanes;
    UINT16 biBitCount;
    UINT32 biCompression;
    UINT32 biSizeImage;
    UINT32 biXPelsPerMeter;
    UINT32 biYPelsPerMeter;
    UINT32 biClrUsed;
    UINT32 biClrImportant;
    } 	BITMAPINFOHEADER;

typedef struct tagBITMAPINFOHEADER *PBITMAPINFOHEADER;

#if defined(_MSC_VER)
#pragma pack(1)
#endif

typedef struct tagBITMAPFILEHEADER
    {
    UINT16 bfType;
    UINT32 bfSize;
    UINT16 bfReserved1;
    UINT16 bfReserved2;
    UINT32 bfOffBits;
    } _PACKED	BITMAPFILEHEADER;

typedef struct tagBITMAPFILEHEADER *PBITMAPFILEHEADER;

typedef struct tagRGBQUAD {
    UINT8    rgbBlue;
    UINT8    rgbGreen;
    UINT8    rgbRed;
    UINT8    rgbReserved;
} RGBQUAD;
typedef RGBQUAD * PRGBQUAD;

UINT32 Buffer[128];

SCODE GetFile(const _TCHAR *filename, PIBUFFER *pBuf, /*inout*/ UINT *pSize)
{
    METHOD_DECLARE_NOTHIS(bmp.c, GetFile);
    PIFILE file = NULL;
    PIBUFFER im = NULL;
    PIENDPOINTFACTORY epf = NULL;
    PINAMESPACE ns = CurrentNameSpace();
    UINT64 size64;
    UINT size;

    sc = BindToObject(ns, filename, NAME_SPACE_READ, &IID_IFile,
                      (void **) &file);
    CHECKSC("open file");

    /* Get size hint */
    sc = file->v->GetSize(file, &size64);
    CHECKSC("GetSize file");

    size = Uint64Low32Bits(size64);
    CHECKSETSC(Uint64High32Bits(size64) == 0, E_INVALID_PARAMETER,
               "file too large for IBuffer");

    sc = BindToObject(ns, _T("COB/protocol.cob"), NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void **) &epf);
    CHECKSC("get epf");

    sc = epf->v->AllocateBuffer(epf, size, NULL, &im);
    CHECKSC("AllocateBuffer");

    sc = im->v->ReadWriteFile(im, file, FALSE, UINT64_ZERO, 0, size, pSize);
    CHECKSC("read file into IBuffer");
    CHECKSETSC(*pSize == size, E_FAIL, "file read size mismatch");

    *pBuf = im;
    im = NULL;

METHOD_CLEANUP;
    RELEASE(file);
    RELEASE(im);
    RELEASE(epf);
    return sc;
}

SCODE ShowBmp(PIBUFFER im)
{
    METHOD_DECLARE_NOTHIS(bmp.c, ShowBmp);
    PBITMAPFILEHEADER bfh;
    PBITMAPINFOHEADER bih;
    PRGB              pRgb, pRgb2;
    PRGBQUAD          pRgbQuad;
    size_t            size, PaletteSize, off;

    sc = im->v->GetDataPointer(im, 0, (void*) &bfh, &size);
    CHECKSC("GetDataPointer for headers");
    if (size < sizeof *bfh + sizeof *bih) {
        sc = im->v->MakeContiguous(im, 0, sizeof *bfh + sizeof *bih);
        CHECKSC("Make headers contiguous");

        sc = im->v->GetDataPointer(im, 0, (void*) &bfh, &size);
        CHECKSC("GetDataPointer for headers");
        assert(size >= sizeof *bfh + sizeof *bih);
    }
    bih = (void *) (bfh + 1);

    if ((bih->biWidth != 320) ||
        (bih->biHeight != 200)) {
        _tprintf(_T("Uhmm %dW x %dH x %dD wont work but..\n"), 
                 bih->biWidth, bih->biHeight, bih->biBitCount);
    }

    /* Palette
     */
    off = sizeof(BITMAPFILEHEADER) + bih->biSize;

    PaletteSize = bih->biClrUsed;
    if (PaletteSize == 0) 
        PaletteSize = 1 << bih->biBitCount;

    sc = im->v->GetDataPointer(im, off, (void*) &pRgbQuad, &size);
    CHECKSC("Get palette pointer");
    if (size < PaletteSize * sizeof(RGBQUAD)) {
        sc = im->v->MakeContiguous(im, off, PaletteSize * sizeof(RGBQUAD));
        CHECKSC("Make paletter contiguous");

        sc = im->v->GetDataPointer(im, off, (void*) &pRgbQuad, &size);
        CHECKSC("Get palette pointer");
        CHECKSETSC(size >= PaletteSize * sizeof(RGBQUAD), E_FAIL,
                   "Palette size mismatch");
    }

    pRgb2 = pRgb = (PRGB) pRgbQuad;
    for (size = 0; size < PaletteSize; size++) {
        RGB Rgb;
        Rgb.rgbRed = pRgbQuad->rgbRed;
        Rgb.rgbGreen = pRgbQuad->rgbGreen;
        Rgb.rgbBlue = pRgbQuad->rgbBlue;
        *pRgb = Rgb;

        pRgb++, pRgbQuad++;
    }

#if 1
{
    UINT8* flippy = malloc(bih->biSizeImage);
    UINT8 *src;
    UINT32 line, cnt;
    if (flippy != NULL) {
        UINT64 off64;
        off = bfh->bfOffBits;
        Uint32ToUint64(off64, off);
        im->v->ReadAt(im, off64, flippy, bih->biSizeImage, &size);
        CHECKSC("read to flippy");
        assert(size == bih->biSizeImage);
        line = (bih->biWidth * bih->biBitCount) / 8;
        cnt = line * bih->biHeight;
        src = flippy+cnt-line;
        while (cnt) {
            im->v->WriteAt(im, off64, src, line, &size);
            cnt -= line;
            src -= line;
            off += line;
            Uint32ToUint64(off64, off);
        }
        free(flippy);
    } else {
        _tprintf(_T("%d Outtamem2?\n"),bih->biSizeImage);
    }
}
#endif

    /* Show it
     */
    size = ((bih->biBitCount*bih->biWidth) + 31) / 32;
    sc = ScreenInit();
    CHECKSC("ScreenInit");

    ScreenPalette(pRgb2, PaletteSize);
    ScreenBlit2(im, bfh->bfOffBits, size * bih->biHeight * 4);
    ScreenTerminate();
METHOD_CLEANUP;
    return sc; 
}

#if 0
/*** main entry point ***/
int _tmain(INT argc,_TCHAR **argv)
{
    METHOD_DECLARE_NOTHIS(bmp.c, _tmain);
    PIBUFFER im = NULL;
    UINT size;

    /* check args
     */
    argc = 0;
    (void) ParseCmdLine( &argc, &argv);
    if (argc < 2) {
        _tprintf(_T("Usage:%s <bitmap_filename>\n"), argv[0]);
        goto Out;
    }

    sc = GetFile(argv[1], &im, &size);
    CHECKSC("GetFile");

    sc = ShowBmp(im);
    CHECKSC("ShowBmp");

    _tprintf(_T("done ok.\r\n"));

METHOD_CLEANUP;
    RELEASE(im);

    (void) FreeCmdLine( &argc, &argv);
    return 0;
}
#endif
