/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>
#include <string.h>
#include <util/list-double2.h>
#include <web/tokenizerutils.h>
#include "ws.h"

/* Memory management */

/* Constructs a dynamically allocated EndpointReference instance */
PEndpointReference WsEndpointNew(const _TCHAR *Address)
{
    PEndpointReference pEndpoint;
    PIHEAP pHeap = CurrentHeap();
    UINT BufSz; 
    if (!Address) return NULL;
    /* Single heap allocation for struct and string. 
     * Destructor relies on this */
    BufSz = sizeof(EndpointReference) + sizeof(_TCHAR) * (_tcslen(Address) + 1);
    pEndpoint = (PEndpointReference) pHeap->v->
        Alloc(pHeap, HEAP_DEBUG_NAME("ws:Endpoint"), BufSz, 0);
    if (!pEndpoint) return NULL;
    memset(pEndpoint, 0, BufSz);
    pEndpoint->Address = (_TCHAR *) (pEndpoint + 1);
    _tcscpy(pEndpoint->Address, Address);
    ListNodeInit(eprloff, pEndpoint);
    return pEndpoint;
}

/* Assumes the struct and string were allocated in one call */
INT WsEndpointDelete(PTR pItem, PTR pArg)
{
    PIHEAP pHeap = (PIHEAP) pArg;
    pHeap->v->Free(pHeap, 0, pItem);
    return 0;
}

/* ReferenceProperty and ReferenceParameter are not handled here */
PTR WsEndpointCopy(PTR pItem, PTR pArg)
{
    PEndpointReference pOldEp = (PEndpointReference) pItem, pNewEp;
    PIHEAP pHeap = (PIHEAP) pArg;
    UINT lenAddr, lenPortType, lenSvcName, lenSvcPort, blocklen;
    _TCHAR *nextfield = NULL;

    if (pItem == 0) 
        return 0;

    /* Allocate a block of memory for struct and strings */
    lenAddr = pOldEp->Address ? _tcslen(pOldEp->Address) + 1 : 0;
    lenPortType = pOldEp->PortType ? _tcslen(pOldEp->PortType) + 1 : 0;
    lenSvcName = pOldEp->ServiceName ? _tcslen(pOldEp->ServiceName) + 1 : 0;
    lenSvcPort = pOldEp->ServicePort ? _tcslen(pOldEp->ServicePort) + 1 : 0;
    blocklen = sizeof(EndpointReference) + 
        (lenAddr + lenPortType + lenSvcName + lenSvcPort) * sizeof(_TCHAR);

    pNewEp = pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:Endpoint"), blocklen, 0);
    if (!pNewEp) return NULL;
    nextfield = (_TCHAR *) (pNewEp + 1);

    /* Copy simple fields */
    memcpy(pNewEp, pOldEp, sizeof(EndpointReference));

    if (lenAddr > 0) {
        pNewEp->Address = nextfield;
        _tcscpy(pNewEp->Address, pOldEp->Address);
        nextfield += lenAddr;
    }
    if (lenPortType > 0) {
        pNewEp->PortType = nextfield;
        _tcscpy(pNewEp->PortType, pOldEp->PortType);
        nextfield += lenPortType;
    }
    if (lenSvcName > 0) {
        pNewEp->ServiceName = nextfield;
        _tcscpy(pNewEp->ServiceName, pOldEp->ServiceName);
        nextfield += lenSvcName;
    }
    if (lenSvcPort > 0) {
        pNewEp->ServicePort = nextfield;
        _tcscpy(pNewEp->ServicePort, pOldEp->ServicePort);
        nextfield += lenSvcPort;
    }
    ListNodeInit(eprloff, pNewEp);

    return pNewEp;
}

PRIVATE INT WsTokenizer(const _TCHAR *str)
{
    _TCHAR *pt = NULL;

    pt = _tcschr(str, ' ');
    if (pt == NULL) return -1;
    else return pt - str;
}

PQNameItem WsQNameItemNew(const _TCHAR *str)
{
    PQNameItem pQName = NULL, pTmp = NULL;
    PIHEAP pHeap = CurrentHeap();
    UINT BufSz;
    INT idx;
    INT len;
    INT slen;

    if (str == NULL) return NULL;
    len = WsTokenizer(str);
    if (len == -1) slen = _tcslen(str);
    else slen = len;

    BufSz = sizeof(struct QNameItem) + sizeof(_TCHAR) * (slen + 1);
    pQName = (PQNameItem) pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:QNameItem"), BufSz, 0);
    if (!pQName) return NULL;
    memset(pQName, 0, BufSz);
    pQName->Value = (_TCHAR *) (pQName + 1);
    _tcsncpy(pQName->Value, str, slen);
    ListNodeInit(qnloff, pQName);
    if (len == -1) return pQName; 
    idx = len + 1;

    while (1) {
        len = WsTokenizer(str + idx);
        if (len == -1) slen = _tcslen(str + idx);
        else slen = len;

        BufSz = sizeof(struct QNameItem) + sizeof(_TCHAR) * (slen + 1);
        pTmp = (PQNameItem) pHeap->v->
            Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:QNameItem"), BufSz, 0);
        if (!pTmp) return pQName;
        memset(pTmp, 0, BufSz);
        pTmp->Value = (_TCHAR *) (pTmp + 1);
        _tcsncpy(pTmp->Value, str + idx, slen);
        ListAddLast(&pQName, qnloff, pTmp);
        if (len == -1) return pQName;
        idx += (len + 1);
    }
}

INT WsQNameItemDelete(PTR pItem, PTR pArg)
{
    PIHEAP pHeap = (PIHEAP) pArg;
    PQNameItem pQName = (PQNameItem) pItem;
    pHeap->v->Free(pHeap, 0, pQName);
    return 0;
}

PTR WsQNameItemCopy(PTR pItem, PTR pArg)
{
    PQNameItem pOld = (PQNameItem) pItem, pNew;
    PIHEAP pHeap = (PIHEAP) pArg;
    UINT lenValue, blocklen;
    _TCHAR *nextfield = NULL;

    if (pItem == 0) 
        return 0;

    /* Allocate a block of memory for struct and strings */
    lenValue = pOld->Value ? _tcslen(pOld->Value) + 1 : 0;
    blocklen = sizeof(*pNew) + lenValue * sizeof(_TCHAR);

    pNew = pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:QNameItem"), blocklen, 0);
    if (!pNew) return NULL;
    nextfield = (_TCHAR *) (pNew + 1);

    /* Copy simple fields */
    memcpy(pNew, pOld, sizeof(*pNew));

    if (lenValue > 0) {
        pNew->Value = nextfield;
        _tcscpy(pNew->Value, pOld->Value);
        nextfield += lenValue;
    }
    return pNew;
}

INT WsQNameItemCmp(PTR p1, PTR p2)
{
    PQNameItem pItem = (PQNameItem) p1;
    _TCHAR *str = (_TCHAR *) p2;

    return _tcscmp(pItem->Value, str);
}

PScopes WsScopesNew(const _TCHAR *matchby, const PQNameItem qnames)
{
    PScopes pScopes = NULL;
    PIHEAP pHeap = CurrentHeap();
    UINT BufSz;

    if (matchby == NULL) {
        if (qnames == NULL) return NULL;
        BufSz = sizeof(struct Scopes);
        pScopes = (PScopes) pHeap->v->
            Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:Scopes"), BufSz, 0);
        if (!pScopes) return NULL;
    } else {
        BufSz = sizeof(struct Scopes) + sizeof(_TCHAR) * (_tcslen(matchby) + 1);
        pScopes = (PScopes) pHeap->v->
            Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:Scopes"), BufSz, 0);
        if (!pScopes) return NULL;
        pScopes->MatchBy = (_TCHAR *) (pScopes + 1);
        _tcscpy(pScopes->MatchBy, matchby);
        if (qnames == NULL) return pScopes;
    }

    pScopes->Items = ListCopy(&qnames, qnloff, WsQNameItemCopy, WsQNameItemDelete, pHeap);
    ListNodeInit(scloff, pScopes);
    return pScopes;
}

INT WsScopesDelete(PTR pItem, PTR pArg)
{
    PIHEAP pHeap = (PIHEAP) pArg;
    PScopes pScopes = (PScopes) pItem;
    if (pScopes->Items != NULL) 
        ListWalk(&pScopes->Items, qnloff, LIST_WALK_ITERATE, WsQNameItemDelete, pHeap);
    pHeap->v->Free(pHeap, 0, pScopes);
    return 0;
}

PTR WsScopesCopy(PTR pItem, PTR pArg)
{
    PScopes pOld = (PScopes) pItem, pNew;
    PIHEAP pHeap = (PIHEAP) pArg;
    UINT lenMatchBy, blocklen;
    _TCHAR *nextfield = NULL;

    if (pItem == 0) 
        return 0;

    /* Allocate a block of memory for struct and strings */
    lenMatchBy = pOld->MatchBy ? _tcslen(pOld->MatchBy) + 1 : 0;
    blocklen = sizeof(*pNew) + lenMatchBy * sizeof(_TCHAR);

    pNew = pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:Scopes"), blocklen, 0);
    if (!pNew) return NULL;
    nextfield = (_TCHAR *) (pNew + 1);

    /* Copy simple fields */
    memcpy(pNew, pOld, sizeof(*pNew));

    if (lenMatchBy > 0) {
        pNew->MatchBy = nextfield;
        _tcscpy(pNew->MatchBy, pOld->MatchBy);
        nextfield += lenMatchBy;
    }

    /* Copy Items (PQNameItem) */
    pNew->Items = ListCopy(&pOld->Items, qnloff, WsQNameItemCopy, WsQNameItemDelete, pHeap);
    return pNew;
}


PStringItem WsStringItemNew(const _TCHAR *str)
{
    PStringItem pString = NULL, pTmp = NULL;
    PIHEAP pHeap = CurrentHeap();
    UINT BufSz;
    INT idx;
    INT len;
    INT slen;

    if (str == NULL) return NULL;
    len = WsTokenizer(str);
    if (len == -1) slen = _tcslen(str);
    else slen = len;

    BufSz = sizeof(struct StringItem) + sizeof(_TCHAR) * (slen + 1);
    pString = (PStringItem) pHeap->v->Alloc(pHeap, HEAP_DEBUG_NAME("ws:StringItem"), BufSz, 0);
    if (!pString) return NULL;
    memset(pString, 0, BufSz);
    pString->Value = (_TCHAR *) (pString + 1);
    _tcsncpy(pString->Value, str, slen);
    ListNodeInit(sloff, pString);
    if (len == -1) return pString; 
    idx = len + 1;

    while (1) {
        len = WsTokenizer(str + idx);
        if (len == -1) slen = _tcslen(str + idx);
        else slen = len;

        BufSz = sizeof(struct StringItem) + sizeof(_TCHAR) * (slen + 1);
        pTmp = (PStringItem) pHeap->v->Alloc(pHeap, HEAP_DEBUG_NAME("ws:StringItem"), BufSz, 0);
        if (!pTmp) return pString;
        memset(pTmp, 0, BufSz);
        pTmp->Value = (_TCHAR *) (pTmp + 1);
        _tcsncpy(pTmp->Value, str + idx, slen);
        ListAddLast(&pString, sloff, pTmp);
        if (len == -1) return pString;
        idx += (len + 1);
    }
}

INT WsStringItemDelete(PTR pItem, PTR pArg)
{
    PIHEAP pHeap = (PIHEAP) pArg;
    PStringItem pString = (PStringItem) pItem;
    pHeap->v->Free(pHeap, 0, pString);
    return 0;
}

PTR WsStringItemCopy(PTR pItem, PTR pArg)
{
    PStringItem pOld = (PStringItem) pItem, pNew;
    PIHEAP pHeap = (PIHEAP) pArg;
    UINT lenValue, blocklen;
    _TCHAR *nextfield = NULL;

    if (pItem == 0) 
        return 0;

    /* Allocate a block of memory for struct and strings */
    lenValue = pOld->Value ? _tcslen(pOld->Value) + 1 : 0;
    blocklen = sizeof(*pNew) + lenValue * sizeof(_TCHAR);

    pNew = pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:StringItem"), blocklen, 0);
    if (!pNew) return NULL;
    nextfield = (_TCHAR *) (pNew + 1);

    /* Copy simple fields */
    memcpy(pNew, pOld, sizeof(*pNew));

    if (lenValue > 0) {
        pNew->Value = nextfield;
        _tcscpy(pNew->Value, pOld->Value);
        nextfield += lenValue;
    }
    return pNew;
}

/* Advert constructor */
PWSADVERT WsAdvertNew(
    PEndpointReference pEndpoint,   //const _TCHAR *Address,
    PQNameItem pTypes,              //const _TCHAR *Types,
    PScopes pScopes,                //const _TCHAR *Scopes,
    PStringItem pXAddrs,             //const _TCHAR *XAddrs,
    UINT MetadataVersion)
{
    PWSADVERT pAdvert;
    PIHEAP pHeap = CurrentHeap();
    UINT BufSz = 0;

    if (!pEndpoint) return NULL;

    BufSz = sizeof(WSADVERT);
    pAdvert = (PWSADVERT) pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:Advert"), BufSz, 0);
    if (!pAdvert) return NULL;
    pAdvert->Endpoint = WsEndpointCopy(pEndpoint, pHeap);
    pAdvert->Types = ListCopy(&pTypes, qnloff, WsQNameItemCopy, WsQNameItemDelete, pHeap);
    pAdvert->Scopes = ListCopy(&pScopes, scloff, WsScopesCopy, WsScopesDelete, pHeap);
    pAdvert->XAddrs = ListCopy(&pXAddrs, sloff, WsStringItemCopy, WsStringItemDelete, pHeap);
    pAdvert->MetadataVersion = MetadataVersion;
    ListNodeInit(aloff, pAdvert);

    return pAdvert;
}

/* Advert destructor. */
INT WsAdvertDelete(PTR pItem, PTR pArg) 
{
    PIHEAP pHeap = (PIHEAP) pArg;
    PWSADVERT pAdvert = (PWSADVERT) pItem;

    if (pAdvert->Endpoint) 
        WsEndpointDelete(pAdvert->Endpoint, pHeap);
    if (pAdvert->Types) 
        ListWalk(&pAdvert->Types, qnloff, LIST_WALK_ITERATE, WsQNameItemDelete, pHeap);
    if (pAdvert->Scopes)
        ListWalk(&pAdvert->Scopes, scloff, LIST_WALK_ITERATE, WsScopesDelete, pHeap);
    if (pAdvert->XAddrs)
        ListWalk(&pAdvert->XAddrs, sloff, LIST_WALK_ITERATE, WsStringItemDelete, pHeap);
    assert(ListNodeIsAlone(aloff, pAdvert));
    pHeap->v->Free(pHeap, 0, pAdvert);
    return 0;
}

PTR WsAdvertCopy(PTR pItem, PTR pArg)
{
    PWSADVERT pOld = (PWSADVERT) pItem, pNew;
    PIHEAP pHeap = (PIHEAP) pArg;

    if (pItem == 0) 
        return 0;

    /* Allocate a block of memory for struct and strings */
    pNew = pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:Advert"), sizeof(*pNew), 0);
    if (!pNew) return NULL;
    
    /* Copy simple fields */
    memcpy(pNew, pOld, sizeof(*pNew));

    /* Copy Items (PQNameItem) */
    pNew->Endpoint = WsEndpointCopy(pOld->Endpoint, pHeap);
    pNew->Types = ListCopy(&pOld->Types, qnloff, WsQNameItemCopy, WsQNameItemDelete, pHeap);
    pNew->Scopes = ListCopy(&pOld->Scopes, scloff, WsScopesCopy, WsScopesDelete, pHeap);
    pNew->XAddrs = ListCopy(&pOld->XAddrs, sloff, WsStringItemCopy, WsStringItemDelete, pHeap);
    return pNew;
}

INT WsAdvertEndpointCmp(PTR p1, PTR p2)
{
    PWSADVERT pAd1 = (PWSADVERT) p1;
    PEndpointReference pEp = (PEndpointReference) p2;

    return _tcscmp(pAd1->Endpoint->Address, pEp->Address);
}

/* ProbeMatch constructor */
PProbeMatch WsProbeMatchNew(
    PEndpointReference pEndpoint,   //const _TCHAR *Address,
    PQNameItem pTypes,              //const _TCHAR *Types,
    PScopes pScopes,                //const _TCHAR *Scopes,
    PStringItem pXAddrs,             //const _TCHAR *XAddrs,
    UINT MetadataVersion)
{
    PProbeMatch pMatch;
    PIHEAP pHeap = CurrentHeap();
    UINT BufSz = 0;

    if (!pEndpoint) return NULL;

    BufSz = sizeof(struct ProbeMatch);
    pMatch = (PProbeMatch) pHeap->v->
        Alloc(pHeap, HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("ws:ProbeMatch"), BufSz, 0);
    if (!pMatch) return NULL;

    pMatch->EndpointReference = WsEndpointCopy(pEndpoint, pHeap);
    pMatch->Types = ListCopy(&pTypes, qnloff, WsQNameItemCopy, WsQNameItemDelete, pHeap);
    pMatch->Scopes = ListCopy(&pScopes, scloff, WsScopesCopy, WsScopesDelete, pHeap);
    pMatch->XAddrs = ListCopy(&pXAddrs, sloff, WsStringItemCopy, WsStringItemDelete, pHeap);
    pMatch->MetadataVersion = MetadataVersion;
    ListNodeInit(pmloff, pMatch);
    return pMatch;
}

/* ProbeMatch destructor. */
INT WsProbeMatchDelete(PTR pItem, PTR pArg) 
{
    PIHEAP pHeap = (PIHEAP) pArg;
    PProbeMatch pMatch = (PProbeMatch) pItem;

    if (pMatch->EndpointReference) 
        WsEndpointDelete(pMatch->EndpointReference, pHeap);
    if (pMatch->Types) 
        ListWalk(&pMatch->Types, qnloff, LIST_WALK_ITERATE, WsQNameItemDelete, pHeap);
    if (pMatch->Scopes)
        ListWalk(&pMatch->Scopes, scloff, LIST_WALK_ITERATE, WsScopesDelete, pHeap);
    if (pMatch->XAddrs)
        ListWalk(&pMatch->XAddrs, sloff, LIST_WALK_ITERATE, WsStringItemDelete, pHeap);
    pHeap->v->Free(pHeap, 0, pMatch);
    return 0;
}


#if 0
/* Constructs a dynamically allocated EndpointReference instance
 * from the To header field of the SOAP message the current 
 * thread is processing
 */
PEndpointReference WsEndpointNewFromSoapContext(void)
{
    /* XXX BUGBUG not implemented... */
    DebugBreak();
    return NULL;
}

/* Constructs a dynamically allocated WSPROPERTY instance */
PWSPROPERTY WsPropertyNew(const _TCHAR *QName)
{
    PWSPROPERTY pProp;
    PIHEAP pHeap = CurrentHeap();
    UINT BufSz = sizeof(WSPROPERTY) + sizeof(_TCHAR) * (_tcslen(QName) + 1);

    DBG(("wsaddr.c:WsPropertyNew: qname='%s'\n", QName));

    /* Single heap allocation for struct and string. 
     * Destructor relies on this */
    pProp = (PWSPROPERTY) pHeap->v->
        Alloc(pHeap, HEAP_DEBUG_NAME("wsdiscov.c:WsPropertyNew"), BufSz, 0);
    if (!pProp) return NULL;
    pProp->QName = (_TCHAR *) (pProp + 1);
    _tcscpy(pProp->QName, QName);
    ListNodeInit(ploff, pProp);
    return pProp;
}

/* Assumes the struct and string were allocated in one call */
INT WsPropertyDelete(PTR pItem, PTR pArg) 
{
    PIHEAP pHeap = (PIHEAP) pArg;
    pHeap->v->Free(pHeap, 0, pItem);
    return 0;
}

/* WSPROPERTY copy constructor. 
 * Copies are shallow; connected list nodes are not copied. 
 * Use ListWalk to deep copy 
 */
PTR WsPropertyCopy(PWSPROPERTY pProp)
{
    PIHEAP pHeap = CurrentHeap();
    UINT BufSz = sizeof(pProp) + sizeof(_TCHAR) * (_tcslen(pProp->QName) + 1);
    /* Single allocation - see constructor, destructor */
    PWSPROPERTY pCopy = (PWSPROPERTY) pHeap->v->
        Alloc(pHeap, HEAP_DEBUG_NAME("wsdiscov.c:WsPropertyCopy"), BufSz, 0);
    if (!pCopy) return NULL;
    _tcscpy(pCopy->QName, pProp->QName);
    ListNodeInit(ploff, pCopy);
    return pCopy;
}

/* Creates a list of properties from a whitespace-delimited 
 * input string. Example input:
 *
 *      "i:printer i:scanner i:camera"
 */
PWSPROPERTY WsPropertyParse(const _TCHAR *Input)
{
    /* Surely the *intent* is const... */
    _TCHAR *Properties = (_TCHAR *) Input; 
    /* Parsing goes backwards through the input argument */
    PWSPROPERTY pFirst = NULL, pCur = NULL;
    UINT Len, End, Start;
    _TCHAR *QName = NULL;
    _TCHAR Tmp = _EOS;
    PIHEAP pHeap = CurrentHeap();

    DBG(("wsaddr.c:WsPropertyParse: input='%s'\n", Input));

    if (!Properties) return NULL;

    Len = _tcslen(Properties);
    End = Len;
    Start = End;

    if (0 >= Len) return NULL;

    /* Cut off trailing whitespace */

    /* Both Start and End point past the last unprocessed character. 
     * Walk End leftwards to the leftmost trailing whitespace. 
     * Set Start to End. 
     * Walk Start leftwards to the leftmost contiguous non-whitespace.
     * Snarf the token.
     * Repeat until Start is 0.
     */
    do {
        /* Walk End leftwards... */
        while (0 < End && IsWhitespace(Properties + (End - 1)))
            --End;

        /* Set Start to End... */
        Start = End;

        /* Any more non-whitespace stuff? */
        if (0 >= Start)
            return pFirst;

        /* Walk Start leftwards... */
        while (0 < Start && !IsWhitespace(Properties + (Start - 1)))
            --Start;

        /* Snarf... */
        QName = Properties + Start;
        Tmp = End < Len ? Properties[End] : _EOS;
        if (_EOS != Tmp) 
            Properties[End] = _EOS;
        pCur = WsPropertyNew(QName);
        if (_EOS != Tmp)
            Properties[End] = Tmp;
        if (!pCur) goto Cleanup;
        if (pFirst)
            ListAddFirst(&pFirst, ploff, pCur);
        pFirst = pCur;

        /* Repeat... */
        End = Start;
    } while (0 < Start);

    return pFirst;

Cleanup:
    if (pFirst)
        ListWalk(pFirst, ploff, LIST_WALK_REMOVE, WsPropertyDelete, pHeap);
    return NULL;
}

/* Collapses a list of properties into a single whitespace-delimited 
 * string holding all the properties. For example, 
 *      "i:Printer i:Scanner i:Camera"
 *
 * The caller is responsible for deallocating the string from the heap.
 *
 * Returns NULL if out of memory or if the argument is null. So check 
 * your argument... 
 */
_TCHAR * WsPropertyCollapseList(PWSPROPERTY pProp)
{
    PWSPROPERTY pCur = NULL;
    UINT BufSz = 0;
    _TCHAR *pBuf = NULL, *pOffset;
    PIHEAP pHeap = CurrentHeap();

    DBG(("wsaddr.c:WsPropertyCollapseList prop=0x%x\n", pProp));

    if (!pProp) return NULL;

    pCur = pProp;
    do {
        BufSz += _tcslen(pCur->QName) + 1;
        pCur = ListNodeNext(ploff, pCur);
    } while (pCur != pProp);

    pBuf = (_TCHAR *) pHeap->v->
        Alloc(pHeap, HEAP_DEBUG_NAME("wsdiscov.c:WsPropertyCollapseList"), BufSz, 0);
    if (!pBuf) return NULL;

    pOffset = pBuf;
    pCur = pProp;
    do {
        _tcscat(pOffset, pCur->QName);
        pOffset += _tcslen(pCur->QName);
        pOffset[0] = ' '; ++pOffset;
        pCur = ListNodeNext(ploff, pCur);
    } while (pCur != pProp);

    assert(pBuf + BufSz == pOffset);

    return pBuf;
}
#endif
