/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Implements WsReservation concrete class -jvh 2004 */

//#define _DEBUG 1
//#define USE_SAMPLER 1

#include <stddef.h> /* for offsetof() */
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <mmlite.h>
#include <util/list-double2.h>
#include <wsif.h>
#include <soapif.h>
#include <wsifhelp.h>
#include <web/cont.h>
#include <base/cobdesc.h>

extern void TimeFormat(TIME t, _TCHAR *buf, UINT Len);

/*** reservation flags ***/
/* Refcount */
#define RESERVATION_FLAGS_REFCOUNT_MASK     0x0000ffff
/* User settable */
#define RESERVATION_FLAGS_PRODUCER          0x00010000
#define RESERVATION_FLAGS_CONSUMER          0x00020000
#define RESERVATION_FLAGS_CASCADE_CANCEL    0x00040000
#define RESERVATION_FLAGS_NEEDS_TRIGGER     0x00080000
#define RESERVATION_FLAGS_MASK              0x00ff0000
/* Internal state */
#define RESERVATION_FLAGS_TRIGGERED         0x01000000
#define RESERVATION_FLAGS_TERMINATED        0x02000000
#define RESERVATION_FLAGS_HAS_ALL_MESSAGES  0x10000000
#define RESERVATION_FLAGS_RECURRING         0x20000000
#define RESERVATION_FLAGS_COPY_MASK (RESERVATION_FLAGS_MASK|0xf0000000)

#define resoloff offsetof(struct resource, ListNode)
#define trloff offsetof(struct trigger, ListNode)

#define rloff offsetof(struct reservation, ListNode)
#define rgoff offsetof(struct reservation, GlobalLink)
#define rroff offsetof(struct reservation, RepeatLink)

#define soff offsetof(struct sample, ListNode)

PRIVATE PTR ResourceCopy(PTR pItem, PTR pArg)
{
    assert(FALSE);
    return NULL;                /* XXX NYI */
}

PRIVATE INT ResourceDelete(PTR pItem, PTR pArg)
{
    assert(FALSE);
    return 0;                /* XXX NYI */
}

PRIVATE PTR TriggerCopy(PTR pItem, PTR pArg)
{
    PIHEAP pHeap = pArg;
    Ptrigger trig = pItem, trnew;
    UINT wlen, olen, blocklen;
    _TCHAR *nextfield = NULL;

    if (0 == pItem)
        return 0;

    /* Allocate a block of memory for struct and strings */
    wlen = trig->what ? _tcslen(trig->what) + 1 : 0;
    olen = trig->otherwise ? _tcslen(trig->otherwise) + 1 : 0;
    olen *= sizeof(_TCHAR);
    blocklen = sizeof(*trnew) + (wlen + olen) * sizeof(_TCHAR);

    trnew = pHeap->v->Alloc(pHeap, HEAP_DEBUG_NAME("soap:trigger"),
                            blocklen, 0);
    if (!trnew) return NULL;
    nextfield = (_TCHAR*) (trnew + 1);

    /* Copy simple fields */
    memcpy(trnew, trig, sizeof(*trnew));

    if (0 < wlen) {
        trnew->what = nextfield;
        _tcscpy(trnew->what, trig->what);
        nextfield += wlen;
    } else {
        trnew->what = NULL;
    }
    if (0 < olen) {
        trnew->otherwise = (PTR/*XXX*/)nextfield;
        _tcscpy(trnew->otherwise, trig->otherwise);
        nextfield += olen;
    } else {
        trnew->otherwise = NULL;
    }

    return trnew;
}

INT TriggerDelete(PTR pItem, PTR pArg)
{
    PIHEAP pHeap = pArg;
    pHeap->v->Free(pHeap, 0, pItem);
    return 0;
}

PRIVATE INT ReservationRelease(PTR pItem, PTR pArg);

PRIVATE INT ReservationDelete(PTR pItem, PTR pArg)
{
    PIHEAP pHeap = pArg;
    Preservation res = pItem;
    if (res->ResourceList)
        ListDelete(&res->ResourceList, resoloff, ResourceDelete, pHeap);
    if (res->TriggerList)
        ListDelete(&res->TriggerList, trloff, TriggerDelete, pHeap);
    if (res->SubReservationList)
        ListDelete(res->SubReservationList, rloff, ReservationRelease, pHeap);
    /* XXX Free the name and contact strings - but watch out 
     * for deserializer-allocated ones. 
     * The copy code currently allocates the strings with the struct so 
     * additional Frees are not needed... not so for the deserializer...
     */
    RELEASE(res->Continuation);
    pHeap->v->Free(pHeap, 0, res);
    return 0;
}

PRIVATE INT ReservationRelease(PTR pItem, PTR pArg)
{
    PIHEAP pHeap = pArg;
    Preservation res = pItem;
    UINT Refs;

    res->Flags--;
    Refs = res->Flags & RESERVATION_FLAGS_REFCOUNT_MASK;
    if (Refs)
        return Refs;

    if (res->CloneOf) {
        ReservationRelease(res->CloneOf, pArg);
        pHeap->v->Free(pHeap, 0, res);
        return 0;
    }

    return ReservationDelete(pItem, pArg);
}

PRIVATE INT ReservationAddRef(PTR pItem, PTR pArg)
{
    Preservation res = pItem;

    res->Flags++;
    return 0;
}

PRIVATE PTR ReservationCopyCommon(PTR pItem, PTR pArg, BOOL DeepCopy)
{
    PIHEAP pHeap = pArg;
    Preservation res = pItem, rnew;
    UINT len = 0;

    if (0 == pItem)
        return 0;

    /* Allocate a block of memory for struct and strings */
    if (DeepCopy) {
        /* Extend buffer to also contain string copies */
        len += res->name ? _tcslen(res->name) + 1 : 0;
        len += res->contact ? _tcslen(res->contact) + 1 : 0;
        len *= sizeof(_TCHAR);
    }
    len += sizeof(*res);

    /* The memory management needs to be re-designed.
     * The deserializer uses a temporary heap, everything is cleaned up
     * in one swoop, nothing needs cleaning. 
     * But, reservations that are allocated here need to be cleaned up. 
     * So there has to be some way of telling them apart...
     * (Reservations do need to be cleaned up as soon as they can be, 
     * because presumably tasks should be expected to occasionally be 
     * very long-lived.)
     * 
     * Note that the name and contact buffers, plus possibly various 
     * other auxiliary memory to do with the reservations, may never 
     * get deallocated... 
     */

    rnew = (Preservation) pHeap->v->Alloc(pHeap, 
                       HEAP_DEBUG_NAME("soap:res2") | HEAP_ZERO_MEMORY,
                       len, 0);
    if (!rnew) return NULL;

    /* Copy the structure */
    memcpy(rnew, res, sizeof(*res));
    /* Set refcount to 1 */
    rnew->Flags &= RESERVATION_FLAGS_COPY_MASK;
    rnew->Flags++;

    /* Fixup pointers */
    ListNodeInit(rloff, rnew);
    ListNodeInit(rgoff, rnew);
    ListNodeInit(rroff, rnew);

    /* Crummy inefficient... */
    if (DeepCopy) {
        if (res->name) {
            rnew->name = (_TCHAR *) (rnew + 1);
            _tcscpy(rnew->name, res->name);
        }
        if (res->contact) {
            rnew->contact = rnew->name + _tcslen(res->name) + 1;
            _tcscpy(rnew->contact, res->contact);
        }
    }

    return rnew;
}

PRIVATE PTR ReservationCopyShallow(PTR pItem, PTR pArg)
{
    Preservation res = pItem, rnew;

    rnew = ReservationCopyCommon(res, pArg, 0);
    if (!rnew)
        return NULL;

    /* Copy flags */
    rnew->Flags = res->Flags & RESERVATION_FLAGS_COPY_MASK;
    /* Set to triggered when not waiting for trigger */
    if (!(res->Flags & RESERVATION_FLAGS_NEEDS_TRIGGER))
        rnew->Flags |= RESERVATION_FLAGS_TRIGGERED;

    /* Copy continuation when appropriate */
    if (res->Flags & RESERVATION_FLAGS_HAS_ALL_MESSAGES)
        res->Continuation->v->AddRef(res->Continuation);
    else
        rnew->Continuation = NULL;

    /* Track refs by keeping the originator res alive */
    if (res->CloneOf) {
        /* Take ref to original clonee so as to avoid long chains */
        ReservationAddRef(res->CloneOf, pArg);
        rnew->CloneOf = res->CloneOf;
    } else {
        ReservationAddRef(res, pArg);
        rnew->CloneOf = res;
    }

    return rnew;
}

PRIVATE PTR ReservationCopy(PTR pItem, PTR pArg)
{
    Preservation res = pItem, rnew;

    rnew = ReservationCopyCommon(res, pArg, 1);
    if (!rnew)
        return NULL;

    /* Set recount to 1 */
    rnew->Flags = 1 | (res->Flags & ~RESERVATION_FLAGS_REFCOUNT_MASK);

    /* Copy complex fields */
    rnew->ResourceList = ListCopy(&res->ResourceList, resoloff,
                                   ResourceCopy, ResourceDelete, pArg);
    rnew->TriggerList = ListCopy(&res->TriggerList, trloff,
                                  TriggerCopy, TriggerDelete, pArg); 

    rnew->SubReservationList = ListCopy(&res->SubReservationList, rloff,
                                         ReservationCopy, ReservationDelete,
                                         pArg);

    return rnew;
}

UINT32 WsResRandUint32(void)
{
    static UINT32 RandomData = 0x81fa590b;
    RandomData = RandomData*43 + Uint64ToUint32(CurrentTime()) + (RandomData >> 16);
    return RandomData;
}
    
_TCHAR *CreateUniqueTaskName(void)
{
    TIME current = CurrentTime();
#define BUFLEN 256
    _TCHAR buf[BUFLEN];
    UINT i;
    TimeFormat(current, buf, BUFLEN);
    for (i = 0; i < BUFLEN; ++i)
        if ('\0' == buf[i])
            break;
    if (10 > BUFLEN - i)
        return NULL;
    _stprintf(buf + i, _T("%d"), WsResRandUint32());
    return _tcsdup(buf); /* XXX LEAK Look out for memory leaks here... */
}
    
_TCHAR *CreateFullReservationName(
    const _TCHAR *taskName, 
    const _TCHAR *resName, 
    const _TCHAR *Suffix)
{
    UINT tlen = _tcslen(taskName);
    UINT rlen = _tcslen(resName) + 1;
    UINT slen = (NULL != Suffix) ? _tcslen(Suffix) : 0;
    UINT len = tlen + rlen + slen + 1;
    _TCHAR *buf = (_TCHAR *) malloc(len * sizeof(_TCHAR));
    if (!buf)
        return NULL;
    _tcscpy(buf, taskName);
    _tcscat(buf, _T("/"));
    _tcscat(buf, resName);
    if (Suffix)
        _tcscat(buf, Suffix);
    return buf;
}

/* reservation state

TRIGGERED_FROM

TRIGGERED
HAS_MESSAGE (WorkItem != NULL)
TERMINATED
HAS_ALL_MESSAGES //msg applies to repeats (vs. only this instance)

init:
if (!From)
    IsT = TRUE;

msg:
if (HasM || Term)
    fail;
HasM = TRUE;
if (IsT)
    run;

trig:
if (Term)
    fail;
if (HasM)
    run;

done:
    Term = TRUE;
    execute triggers;

0          IsT     HasM   IsT|HasM   Term
init       waitM   waitT    run      term
   --!From-->  -----msg------>  -done-->
   --trig--->  ------------------done-->
   -----------msg-->  --trig->
                      -----------done-->
   ------------------------------done-->

run state includes sleep waitin for Start time.

-----------
1. lost msg, skipped reservation
2. should failed msg propagate as [done] or [trig]?
3. missed reservations

trigger from:
1. done running OK
2. failed
3. started running
4. time arrived
4b. automatically (same as 4 I think)
5. seqno skipped

 */

/* A task is a contextually related collection of temporally unrelated
 * reservations. It represents one application's scheduling state.
 */
typedef struct _TASK {
    UINT RefCnt;
    Preservation taskres;       /* list of reservations in this task */
    ListNodeDeclare(_TASK, Link);
#define toff offsetof(struct _TASK, Link)
#ifdef USE_SAMPLER
    PISAMPLING Sampler;
#else
    PIWSRESERVATION Feedback;
#endif
    _TCHAR name[1]; /* must be last */
} *PTASK;

typedef struct _WsRes {
    //CONST_VTBL struct IWsReservationVtbl *v;
    struct _DefCob super;

    MUTEX mx;
    PTASK tasks;
    Preservation allres;        /* global list of reservations, time sorted */
    PISTATISTICS stats;         /* for instigator node, cached */
} WsRes, *PWSRESERVATION;

#include "s_WsRes_v.c"

/* Convert between interface and instance pointer and vice versa */
#define pReservation(_i_) ((PWSRESERVATION)((_i_)))
#define iReservation(_p_) ((PIWSRESERVATION)((_p_)))

PRIVATE PTASK TaskCreate(const _TCHAR *name, Preservation res,
                         const _TCHAR *feedbackURL, PIHEAP Heap)
{
    METHOD_DECLARE_NOTHIS(wsres.c, TaskCreate);
    PTASK task = NULL;
    PISOAPFACTORY fact = NULL;
    UINT nlen = name ? _tcslen(name) : 0; /* +1 extra in struct already */

    task = (PTASK) Heap->v->Alloc(Heap,
                               HEAP_DEBUG_NAME("soap:task") | HEAP_ZERO_MEMORY,
                               sizeof(*task) + nlen * sizeof(_TCHAR), 0);
    CHECKSETSC(task, E_NOT_ENOUGH_MEMORY, "alloc");

    task->RefCnt = 1;
#if !defined(USE_SAMPLER)
    if (feedbackURL) {
        sc = BindToObject(CurrentNameSpace(),
                          _T("COB/soap.cob"), NAME_SPACE_READ,
                          &IID_ISoapFactory, (void **) &fact);
        CHECKSC("bind to SOAP COB");

        sc = fact->v->CreateProxy(fact, NULL,
                                  feedbackURL, _T("res"),
                                  0,
                                  (PIUNKNOWN*) &task->Feedback);
        CHECKSC("create proxy");
    }
#endif
    _tcscpy(task->name, name);
    task->name[nlen] = '\0';    /* in case NULL */
    task->taskres = res;

METHOD_CLEANUP;
    if (FAILED(sc) && task) {
        Heap->v->Free(Heap, 0, task);
        task = NULL;
    }
    RELEASE(fact);
    return task;
}

PRIVATE UINT TaskAddRef(PTASK task)
{
    return AtomicInc(&task->RefCnt);
}

PRIVATE UINT TaskRelease(PTASK task, PTR pArg)
{
    PIHEAP Heap = pArg;
    UINT RefCnt = AtomicDec(&task->RefCnt);
    
    if (RefCnt > 0)
        return RefCnt;

    ListDelete(&task->taskres, rloff, ReservationRelease, pArg);
#ifdef USE_SAMPLER
    RELEASE(task->Sampler);
#else
    RELEASE(task->Feedback);
#endif
    Heap->v->Free(Heap, 0, task);
    return 0;
}

/* Compare RestartBy of two reservations, returns like strcmp: s1-s2 */
PRIVATE INT ReservationCmp(PTR r1, PTR r2)
{
    Preservation res1 = r1, res2 = r2;
    TIME by1, by2, r;

    /* Compare deadline-duration times */
    by1 = Int64Add(res1->deadline, res1->duration);
    by2 = Int64Add(res2->deadline, res2->duration);
    r = Int64Subtract(by1, by2);
    if (Int64IsNegative(r))
        return -1;
    else if (Int64IsZero(r))
        return 0;
    else
        return 1;
        
}

/* Compare ..., returns like strcmp: s1-s2 */
PRIVATE INT ReservationNameCmp(PTR r1, PTR name)
{
    Preservation res1 = r1;
    size_t len;
    int r;

    /* Compare name before slash */
    len = _tcscspn(name, _T("/\\"));
    r = _tcsncmp(res1->name, name, len);
    /* If the task.name is longer than the segment then <name> comes first */
    if (!r && res1->name[len] != _T('\0'))
        return 1;
    else
        return r;
}

/* Called for each reservation in new task. Insert each into global list */
PRIVATE INT ReservationInsert(PTR r, PTR ghead)
{
    Preservation res = r, res2;
    TIME t, xx;
    PLIST_HEAD head = ghead;

    /* Insert into previous list in deadline-duration order */
    ListWalk(ghead, rgoff, LIST_WALK_ADD | LIST_WALK_SORTED,
             ReservationCmp, res);

    /* XXX half-baked code follows XXX */

    /* Simple admission check: make sure no overlap with neighbors
     * XXX periodic not dealt with.
     */
    res2 = ListNodeNext(rgoff, res);
    if (res == res2 || res2 == *head) /* only or last reservation? */
        return S_OK;

    /* Check for overlap with next one. Get its earliest possible start. */
    if (TimeLess(res2->Start, res->deadline)) {
        /* Potential overlap with next one. Adjust slack down */
        t = Int64Subtract(res2->Start, res->deadline);/* relative */
        DBG(("Needs %lld slack between new=%s next=%s\n",
             t, res->name, res2->name));
        /* XXX better division needed. Take it all out of new for now when p */
        if (TimeLess(res->tolerance, t)) { /* note neg numbers */
            xx = TimeSubtract(res->tolerance, t); /* enough */
            res->tolerance = xx;
        } else {
            xx = Int64Subtract(t, res->tolerance);
            t = xx;
            res->tolerance = TIME_ORIGIN;
            if (TimeLess(res2->tolerance, t)) {
                xx = TimeSubtract(res2->tolerance, t); /* ok */
                res2->tolerance = xx;
                xx = TimeSubtract(res2->Start, t); /* adj forward */
                res2->Start = xx;
            } else {
                DBG(("*** Insufficient slack new=%s next=%s deficit=%lld\n",
                     res->name, res2->name, t));
                return E_ALREADY_COMMITTED; /* XXX add new code */
            }
        }
    }

    if (*head == res)
        return S_OK;            /* first reservation */

    /* Check with previous */
    res2 = ListNodePrev(rgoff, res);
    if (TimeLess(res->Start, res2->deadline)) {
        /* Potential overlap with prev one. Adjust slack down */
        t = Int64Subtract(res->Start, res2->deadline);/* relative */
        DBG(("Needs %lld slack between new=%s prev=%s\n",
             t, res->name, res2->name));
        /* XXX better division needed. Take it all out of new for now when p */
        if (TimeLess(res->tolerance, t)) { /* note neg numbers */
            xx = TimeSubtract(res->tolerance, t); /* enough */
            res->tolerance = xx;
            xx = TimeSubtract(res->Start, t); /* adjust forward */
            res->Start = xx;
        } else {
            xx = TimeSubtract(res->Start, res->tolerance);
            res->Start = xx;
            xx = Int64Subtract(t, res->tolerance);
            t = xx;
            res->tolerance = TIME_ORIGIN;
            if (TimeLess(res2->tolerance, t)) {
                xx = TimeSubtract(res2->tolerance, t); /* ok */
                res2->tolerance = xx;
            } else {
                DBG(("*** Insufficient slack new=%s prev=%s deficit=%lld\n",
                     res->name, res2->name, t));
                return E_ALREADY_COMMITTED; /* XXX add new code */
            }
        }
    }

    return S_OK;
}

PRIVATE INT ReservationNormalize(PTR r, PTR pArg)
{
    Preservation res = r;
    TIME t, t2, cur;

    cur = CurrentTime();

    PrintTime("Normalize dl", res->deadline);
    PrintTime("Normalize dur", res->duration);
    PrintTime("Normalize est", res->tolerance);
    PrintTime("Normalize now", cur);

    if (TimeIsRelative(res->deadline) /* absolute (does not apply to subres) */
        || TimeIsPositive(res->duration) /* 0 or relative */
        || TimeIsPositive(res->tolerance)) /* 0 or relative */
    {
        printf("Normalize bad input time\n");
        return E_INVALID_PARAMETER;
    }

    t = Int64Add(res->deadline, res->duration);
    if (TimeLess(t, cur)) {       /* already missed */
#if 0
        printf("Normalize already missed -- bumping forward...\n");
        res->deadline = cur;
        t = t = Int64Add(res->deadline, res->duration);
#else
        PrintTime("Normalize already missed by", Int64Subtract(t, cur));
        return E_INVALID_PARAMETER;
#endif
    }

    /* Earliest allowed start */
    t2 = Int64Add(t, res->tolerance);
    if (TimeLess(t2, cur)) {
        /* Adjust slack down */
        res->tolerance = Int64Subtract(cur, t); /* XXX BAD IDEA for later reps */
        t2 = cur;
    }
    res->Start = t2;

    /* State intialization */
    if (!(res->Flags & RESERVATION_FLAGS_NEEDS_TRIGGER))
        res->Flags |= RESERVATION_FLAGS_TRIGGERED;

    if (res->seqno == 0)
        res->seqno = 1;

    /* Set refcount to 1 initially */
    if ((res->Flags & RESERVATION_FLAGS_REFCOUNT_MASK) == 0)
        res->Flags++;

    return S_OK;
}

struct TriggerResolveArg {
    Preservation *reslist;
    Preservation res;
};

/* fixup triggers */
PRIVATE INT TriggerResolve(PTR pItem, PTR pArg)
{
    Ptrigger trig = pItem;
    struct TriggerResolveArg *a = pArg;
    Preservation res;

    assert(trig->WhatNode == NULL);
    assert(trig->what && trig->what[0]);

    res = ListWalk(a->reslist, rloff, LIST_WALK_SEARCH,
                   ReservationNameCmp, (PTR) trig->what);
    if (!res)
        return (INT) E_NAME_NOT_FOUND;

    trig->WhatNode = res;
    res->Flags |= RESERVATION_FLAGS_NEEDS_TRIGGER;
    if (res == a->res)
        res->Flags |= RESERVATION_FLAGS_RECURRING |RESERVATION_FLAGS_TRIGGERED;

    assert(trig->OtherwiseNode == NULL);
    if (trig->otherwise && trig->otherwise[0]) {
        res = ListWalk(a->reslist, rloff, LIST_WALK_SEARCH,
                       ReservationNameCmp, (PTR) trig->otherwise);
        trig->OtherwiseNode = res;
        if (!res)
            return (INT) E_NAME_NOT_FOUND;
    }

    return (INT) S_OK; /* XXX check results */
}

/* figure out which reservations need triggers and fixup triggers */
PRIVATE INT ReservationTriggerResolve(PTR pItem, PTR pArg)
{
    Preservation res = pItem;
    struct TriggerResolveArg a;
    a.reslist = pArg;
    a.res = res;

    /* State intialization */
    res->Flags &= RESERVATION_FLAGS_MASK | RESERVATION_FLAGS_REFCOUNT_MASK;

    return (INT) ListWalk(&res->TriggerList, trloff, LIST_WALK_SCODE,
                          TriggerResolve, (PTR)&a);
}

#if 0
/* _tcscmp on names */
PRIVATE INT TaskNameCmp(PTR t1, PTR t2)
{
    PTASK task1 = t1, task2 = t2;

    return _tcscmp(task1->name, task2->name);
}
#endif

PRIVATE INT TaskNameCmp2(PTR t1, PTR name)
{
    PTASK task1 = t1;
    size_t len;
    int r;

    /* Compare name before slash */
    len = _tcscspn(name, _T("/\\"));
    r = _tcsncmp(task1->name, name, len);
    /* If the task.name is longer than the segment then <name> comes first */
    if (!r && task1->name[len] != _T('\0'))
        return 1;
    else
        return r;
}

PRIVATE INT TriggerNameCmp(PTR t1, PTR t2)
{
    Ptrigger trig1 = t1;
    Ptrigger trig2 = t2;

    return _tcscmp(trig1->what, trig2->what);
}

PRIVATE INT TriggerAdjust(PTR t1, PTR r2)
{
    METHOD_DECLARE_NOTHIS(wsres.c, TriggerAdjust);
    Ptrigger trig1 = t1;        /* new */
    Ptrigger trig2;             /* old */
    Preservation res2 = r2;

    sc = S_OK;

    trig2 = ListWalk(&res2->TriggerList, trloff, LIST_WALK_SEARCH,
                     TriggerNameCmp, trig1);
    

    if (res2) {
        DBG(("TriggerAdjust: found '%s' in '%s'\n",
             trig1->what, res2->name));
        /* Some fields are not allowed to be changed */
        if (trig1->otherwise && (!trig2->otherwise
                              || _tcscmp(trig1->otherwise, trig2->otherwise)))
            CHECKSETSC(FALSE, E_INVALID_PARAMETER,
                       "contact change not allowed");
        /* Copy some fields */
        trig2->offset = trig1->offset;
        trig2->maxCount = trig1->maxCount; /* most crucial! */
    } else {
        DBG(("TriggerAdjust: DID NOT find '%s' in '%s'\n",
             trig1->what, res2->name));
        CHECKSETSC(FALSE, E_ENTRY_NOT_FOUND, "conflicting triggers");
    }

    METHOD_CLEANUP;
    return (INT) sc;
}

/* Loop over task checking for pre-existing res. Insert if not, else adjust */
PRIVATE INT ReservationAdjust(PTR r1, PTR t2)
{
    METHOD_DECLARE_NOTHIS(wsres.c, ReservationAdjust);
    Preservation res1 = r1;     /* new */
    Preservation res2;          /* old */
    PTASK task = t2;

    sc = S_OK;

    res2 = ListWalk(&task->taskres, rloff, LIST_WALK_SEARCH,
                    ReservationNameCmp, res1->name);

    if (res2) {
        DBG(("ReservationAdjust: found '%s' in '%s'\n",
             res1->name, task->name));
        /* Check state */
        assert(!(res2->Flags & RESERVATION_FLAGS_TERMINATED)); /* XXX fix it up */
        /* Some fields are not allowed to be changed */
        // XXX resources;
        if (res1->contact && (!res2->contact
                              || _tcscmp(res1->contact, res2->contact)))
            CHECKSETSC(FALSE, E_INVALID_PARAMETER,
                       "contact change not allowed");
        // XXX admission check;

        /* Check and copy triggers, omitted ones are left in place */
        sc = (SCODE) ListWalk(&res1->TriggerList, trloff, LIST_WALK_SCODE,
                              TriggerAdjust, res2);
        CHECKSC("TriggerAdjust");

        /* Copy new values into old res */
        if (res1->seqno == res2->seqno) {
            res2->deadline = res1->deadline;
            res2->Start = res1->Start;
        } else if (res2->seqno > res1->seqno) {
            CHECKSETSC(FALSE, E_INVALID_PARAMETER,
                       "adjustment in the past (old seqno)");
            /* ... could as well project deadline into future */
        } else {
            /* project deadline into past: the planner is ahead of us */
            Ptrigger trig = res2->TriggerList; /* XXX find right one! XXX */
            TIME cur, t;

            assert(trig && ListNodeIsAlone(trloff, trig)); /* XXX right trig */

            t = trig->offset;
            t = Int64TimesInt32(t, res1->seqno - res2->seqno);
            /* Set deadline for current seqno so that when it gets adjusted
             * forward, it will get to the input specification at the given
             * seqno.
             */
            res2->deadline = Int64Subtract(res1->deadline, t);
            /* Re-normalize start time
             * -> Earliest allowed start
             */
            t = Int64Add(res2->deadline, res1->duration);
            t = Int64Add(t, res1->tolerance);
            cur = CurrentTime();
            if (TimeLess(t, cur))
                res2->Start = cur;
            else
                res2->Start = t;
        }
        res2->tolerance = res1->tolerance ;
        res2->duration = res1->duration ;
        
    } else {
        DBG(("ReservationAdjust: NEW Res '%s' in '%s'\n",
             res1->name, task->name));
        res2 = ReservationCopy(res1, CurrentHeap());
        CHECKSETSC(res2, E_NOT_ENOUGH_MEMORY, "ReservationCopy new res");

        ListAddLast(&task->taskres, rloff, res2);
        TaskAddRef(task);

        sc = (SCODE) ReservationTriggerResolve(res2, &task->taskres);
        CHECKSC("ReservationTriggerResolve new res");

#if 0
        sc = (SCODE) ReservationInsert(res2, &_wsres.allres);
        CHECKSC2("ReservationInsert new res");
    METHOD_CLEANUP2;
        if (res2)
            ReservationDelete(res2, CurrentHeap());
#endif
    }

    METHOD_CLEANUP;
    return (INT) sc;
}

/* METHOD: Provide new reservation group */
PRIVATE SCODE WsRes_task (
    PIWSRESERVATION iThis,
    PICONTINUATION iCont,
    const _TCHAR * name,
    Preservation reslistarg,
    const _TCHAR *feedbackURL)
{
    METHOD_DECLARE(WsRes, Task);
    PTASK task = NULL;
    PIHEAP Heap = CurrentHeap();
    BOOL GotMutex = FALSE;
    BOOL NewTask = FALSE;

    DBG(("WsRes_task(%x, %s)\n", (ADDRESS) This, name));

    sc = (SCODE) ListWalk(&reslistarg, rloff, LIST_WALK_SCODE,
                          ReservationNormalize, NULL);
    CHECKSC("Normalize");

    Mutex_Lock(&This->mx);
    GotMutex = TRUE;

    /* Check for pre-existing reservation with same name and adjust */
    task = ListWalk(&This->tasks,toff, LIST_WALK_SEARCH, TaskNameCmp2, (PTR) name);
    if (!task) {
        task = TaskCreate(name, NULL, feedbackURL, Heap);
        CHECKSETSC(task, E_NOT_ENOUGH_MEMORY, "TaskCreate");
        NewTask = TRUE;
        DBG(("WsRes_task: new task %x %s\n", (ADDRESS) task, name));
    }

    sc = (SCODE) ListWalk(&reslistarg, rloff, LIST_WALK_SCODE,
                          ReservationAdjust, task);
    CHECKSC("ReservationAdjust task");

#if 0
    sc = (SCODE) ListWalk(&reslist, rloff, LIST_WALK_SCODE,
                          ReservationTriggerResolve, &reslist);
    CHECKSC("TriggerResolve");

    reslist = NULL;

    /* Create new task and try to schedule it against pre-existing */
    reslist = ListCopy(&reslistarg, rloff,
                       ReservationCopy, ReservationDelete, Heap);
    CHECKSETSC(reslist, E_NOT_ENOUGH_MEMORY, "CopyReservation");

    sc = (SCODE) ListWalk(&task->taskres, rloff, LIST_WALK_SCODE,
                          ReservationInsert, &This->allres);
    CHECKSC("ReservationInsert");
#endif

    if (NewTask) {
        /* Insert new task in task list */
        ListAddLast(&This->tasks, toff, task);
        task = NULL;
    }

    METHOD_CLEANUP;
    if (GotMutex)
        Mutex_Unlock(&This->mx);
    if (task && NewTask)
        TaskRelease(task, Heap);

    DBG(("END WsRes_task(%x, %s)\n", (ADDRESS) This, name));

    return sc;
}

/* METHOD: Adjust existing reservations within task */
PRIVATE SCODE WsRes_adjust (
    PIWSRESERVATION iThis,
    PICONTINUATION iCont,
    const _TCHAR * name,
    Preservation reslistarg)
{
    return WsRes_task(iThis, iCont, name, reslistarg, NULL);
#if 0
    METHOD_DECLARE(WsRes, adjust);
    assert(FALSE);
    sc = E_NOT_IMPLEMENTED;
    METHOD_CLEANUP;
    return sc;
#endif
 }

PRIVATE Preservation FindRes(PWSRESERVATION This, const _TCHAR *name,
                             UINT *pSeq, PTASK *pTask)
{
    Preservation res;
    PTASK task;
    UINT seqno = 0;

    //assert(Mutex_Locked(&This->mx));

    task = ListWalk(&This->tasks,toff, LIST_WALK_SEARCH,
                    TaskNameCmp2, (PTR) name);
    if (!task)
        return NULL;

    name = _tcspbrk(name, _T("/\\"));
    if (!name)
        return NULL;
    name++;

    res = ListWalk(&task->taskres, rloff, LIST_WALK_SEARCH,
                   ReservationNameCmp, (PTR) name);
    if (!res)
        return NULL;

    name = _tcspbrk(name, _T("/\\"));
    if (name) {
        name++;
        /* Match sequence number with recurring reservation */
        if (!_tcscmp(name, _T("*")))
            seqno = UINT_MAX;   /* special "all" value */
        else if (isdigit(name[0]))
            seqno = _ttoi(name);
    }
    if (pSeq)
        *pSeq = seqno;
    if (pTask)
        *pTask = task;
    return res;
}

PRIVATE SCODE WsResDestroy(IUnknown *iThis)
{
    METHOD_DECLARE(WsRes, Destroy);
    sc = S_OK;

    /* Finalize object */
    RELEASE(This->stats);

METHOD_CLEANUP;
    return sc; 
}

#if 0
PRIVATE SCODE WsRes_QueryInterface (
    PIWSRESERVATION pThis,
    REFIID iid,
    /*out*/ void** ppObject)
{
    return GenericQueryInterface((IUnknown *)pThis, iid, ppObject,
                                 &IID_IWsReservation);
}

PRIVATE UINT WsRes_AddRef (
    PIWSRESERVATION pThis)
{
    return AtomicInc(&pReservation(pThis)->RefCnt);
}

PRIVATE UINT WsRes_Release (
    PIWSRESERVATION pThis)
{
    PWSRESERVATION This = pReservation(pThis);
    UINT Refs = AtomicDec(&This->RefCnt);
    if (Refs)
        return Refs;

    /* Finalize object */
    WsResDestroy((IUnknown *) This);

    return 0;
}
#endif

/* METHOD: Provide continuation to reservation */
PRIVATE SCODE WsRes_Match(
    PIWSRESERVATION iThis,
    PICONTINUATION iCont,
    const _TCHAR *name,
    /*out*/ Preservation pRes)
{
    METHOD_DECLARE(WsRes, Match);
    Preservation res = NULL;
    UINT seqno;

    DBG(("WsRes_match(%x, %s)\n", (ADDRESS) This, name));

    sc = S_OK;
    Mutex_Lock(&This->mx);
    res = FindRes(This, name, &seqno, NULL);
    CHECKSETSC(res, E_ENTRY_NOT_FOUND, "FindRes");

    /* Can't do same twice... */
    if (res->Continuation) {
        if ((seqno == UINT_MAX)
            || (res->Flags & RESERVATION_FLAGS_HAS_ALL_MESSAGES)
            || !(res->Flags & RESERVATION_FLAGS_RECURRING)
            || (res->seqno == seqno))
        {
            CHECKSETSC(FALSE, E_ALREADY_EXISTS, "res already matched");
        } else {
            DBG(("WsRes_match: needs to split\n"));
            CHECKSETSC(FALSE, E_ENTRY_NOT_FOUND, "res already matched");
        }
    }

    /* Match sequence number with recurring reservation */
    if (seqno == UINT_MAX) {
        res->Flags |= RESERVATION_FLAGS_HAS_ALL_MESSAGES;
    } else if (seqno > res->seqno) {
        /* XXX might have multiple reservation with same name, different seq */
        assert(res->Flags & RESERVATION_FLAGS_RECURRING);
        DBG(("WsRes_Match seqno=%d SKIPPED from %d\n", seqno, res->seqno));
        res->seqno = seqno;
    } else if (seqno && seqno < res->seqno) {
        CHECKSETSC(FALSE, E_ENTRY_NOT_FOUND, "out of sequence");
    }

    iCont->v->AddRef(iCont);
    res->Continuation = iCont;

    /* If the reservation is not ready to run,let caller know */
    CHECKSETSC((res->Flags & RESERVATION_FLAGS_TRIGGERED),
               E_OBJECT_NOT_INITIALIZED, "res not triggered");

    /* Return copy of reservation to caller */
    /* XXX subelements not refcounted and could be freed ! */
    *pRes = *res;

    METHOD_CLEANUP;
    if (FAILED(sc))
        DBG(("%s: resname=\"%s\"\n", _method_name, name));
    Mutex_Unlock(&This->mx);

    DBG(("END WsRes_match(%x, %s)\n", (ADDRESS) This, name));
    return sc;
}

struct RunTriggerArg {
    PWSRESERVATION wsres;
    PTASK task;
    SCODE how;
};

PRIVATE INT RunTrigger(PTR pItem, PTR pArg)
{
    METHOD_DECLARE_NOTHIS(wsres.c, RunTrigger);
    Ptrigger trig = (Ptrigger) pItem;
    struct RunTriggerArg *a = pArg;
    SCODE how = a->how;
    PTASK task = a->task;
    Preservation res = trig->WhatNode;
    UINT Flags;
    PICONTINUATION cont = res->Continuation;

    DBG(("Runtrigger %s %x seq=%d\n",
         trig->what, (ADDRESS) trig->WhatNode, res->seqno));
    sc = S_OK;
    if (res->seqno >= trig->maxCount) {
        res = trig->OtherwiseNode;
        if (!res) {
            DBG(("RunTrigger reached max and no <otherwise>, end\n"));
            /* XXX do stuff + release res */
            goto Out;
        }
    }
    Flags = res->Flags;
    /* Should expect trigger and be in initial state */
    assert(Flags & RESERVATION_FLAGS_NEEDS_TRIGGER);
    assert(!(Flags & RESERVATION_FLAGS_TRIGGERED));

    assert(!TimeIsPositive(trig->offset));

    if (TimeIsRelative(trig->offset)) {
        /* Bump reservation forward in time as specified by trigger */
        res->Start = Int64Subtract(res->Start, trig->offset);
        res->deadline = Int64Subtract(res->deadline, trig->offset);
        PrintTime("RunTrigger bump by", trig->offset);
    }

    if (Flags & RESERVATION_FLAGS_RECURRING) {
        Preservation rold = res;

        if (Flags & RESERVATION_FLAGS_TERMINATED) {
            res->Flags &= ~RESERVATION_FLAGS_TERMINATED;
            res->seqno++;
#if 0
            // cont->Execute now addrefs
            if (cont)
                cont->v->AddRef(cont);
#endif
        } else {
            res = ReservationCopyShallow(res, CurrentHeap());
            assert(res);

            /* Insert instance copy into global list in dl-duration order */
            ListWalk(&a->wsres->allres, rgoff,
                     LIST_WALK_ADD | LIST_WALK_SORTED, ReservationCmp, res);
            /* Insert copy into task list */
            ListAddLast(&task->taskres, rloff, res);
            TaskAddRef(task);

            /* XXX move old recurring one into correct spot when appropriate */
            rold->seqno++;
        }

        assert(!(Flags & RESERVATION_FLAGS_TRIGGERED));
    }

    res->Flags |= RESERVATION_FLAGS_TRIGGERED;

    if (cont) {
        sc = cont->v->Execute(cont, FALSE, res);
        CHECKSC("cont->Execute");
    }

    METHOD_CLEANUP;
    return (INT) sc;
}

/* METHOD: Signal reservation completed, run triggers etc. */
PRIVATE SCODE WsRes_Done(
    PIWSRESERVATION iThis,
    PICONTINUATION iCont,
    const _TCHAR *name,
    SCODE how)
{
    METHOD_DECLARE(WsRes, Done);
    Preservation res;
    UINT seqno;
    PTASK task;
    struct RunTriggerArg a;
    TIME TimeRequested = INT64_ZERO, TimeTaken = INT64_ZERO;
    struct sample _sample, *sample = &_sample;
    BOOL TaskNeedsRelease = FALSE;

    DBG(("WsRes_done(%x, %s)\n", (ADDRESS) This, name));

    sc = S_OK;
    Mutex_Lock(&This->mx);

    res = FindRes(This, name, &seqno, &task);
    CHECKSETSC(res, E_ENTRY_NOT_FOUND, "FindRes");

    /* Copy the sequence number of the reservation, for the sampler */
    seqno = res->seqno;

    res->Flags |= RESERVATION_FLAGS_TERMINATED;
    //if (!(res->Flags & RESERVATION_FLAGS_HAS_ALL_MESSAGES))
    //    RELEASE_NULL(res->Continuation);

    if (res->Flags & RESERVATION_FLAGS_NEEDS_TRIGGER)
        res->Flags &= ~RESERVATION_FLAGS_TRIGGERED;        

    a.how = how;
    a.task = task;
    a.wsres = This;

    /* Prepare for sampling by copying the estimated duration from 
       the 'res' structure, to make sure sampling is not affected 
       by running triggers */
    if (res->contact) {
        TimeRequested = res->duration;
    }

    /* If this is a recurring reservation, bump it forward */
    /* Note: This overwrites the 'res' structure! */
    sc = (SCODE) ListWalk(&res->TriggerList, trloff, LIST_WALK_SCODE,
                          RunTrigger, (PTR) &a);
    CHECKSC("ListWalk/RunTrigger");

    METHOD_CLEANUP;

#ifdef USE_SAMPLER
    /* Lazy load the sampler */
    if (!task->Sampler) {
        sc = BindToObject(CurrentNameSpace(), 
                          _T("COB/sampling.cob"), NAME_SPACE_READ, 
                          &IID_ISampling, (void **) &task->Sampler);
        DBG(("wsres.c:WsRes_Done: Loaded sampling.cob, sampler=%x sc=%x\n", 
             (ADDRESS) task->Sampler, sc));
    }
#endif

    /* If the reservation is done get rid of it */
    if (res) {
        if (res->Flags & RESERVATION_FLAGS_TERMINATED) {
            DBG(("WsRes_Done unlinking res %x %s:%d\n",
                 res,res->name,res->seqno));
#if 0 /* XXX tmp disabled ReservationInsert */
            ListRemoveItem(&This->allres, rgoff, res);
#endif
            ListRemoveItem(&task->taskres, rloff, res);
            //assert((res->Flags & RESERVATION_FLAGS_REFCOUNT_MASK) == 1);
            TaskNeedsRelease = TRUE;
        } else
            /* make sure next iteration doesn't delete res */
            ReservationAddRef(res, NULL);
    }

    Mutex_Unlock(&This->mx);

    if (SUCCEEDED(sc) && res->contact) {
        PCONTINFO cont = pCONT(iCont);

        /* End the scheduler constraint and extract the time taken during 
           its execution. Flip the sign, because the scheduler considers 
           positive time values to be relative, but the rest of the system 
           considers negative values to be relative */
        sc = EndConstraint(&TimeTaken);
        //assert(SUCCEEDED(sc));  /* why would it fail? */
        sc = S_OK;
        TimeTaken = Int64Subtract(TIME_ORIGIN, TimeTaken);

#ifdef USE_SAMPLER
        /* Hand over the new sample */
        if (task->Sampler) {
            sc = task->Sampler->v->AddSample(task->Sampler, name, seqno,
                                             cont->BeginTime, CurrentTime(),
                                             TimeRequested, TimeTaken,
                                             res->contact);
            DBG(("WsRes_Done: Invoked sampler, sc=%x name=%s seqno=%d\n", 
                 sc, name, seqno));
        }
#else
        /* Create sampling entry */
        sample->name = (_TCHAR *) name;
        sample->seqno = seqno;
        ListNodeInit(soff, sample);
        sample->sendTime = cont->sendTime;
        sample->BeginTime = cont->BeginTime;
        sample->EndTime = CurrentTime();
        sample->Estimated = TimeRequested;
        sample->Consumed = TimeTaken;
        sample->status = S_OK;
        sample->message = NULL;

        sc = task->Feedback->v->feedback(task->Feedback, task->name, sample);
        printf("WsRes_Done: sent feedback, sc=%x name=%s seqno=%d\n", 
               sc, name, seqno);
#endif
    }

    Mutex_Lock(&This->mx);
    /* Delete res and task when res is zombie and we are last here */
    if (res && ReservationRelease(res, CurrentHeap()) == 0) {
        if (!task->taskres && !ListNodeIsAlone(toff, task)) {
            ListRemoveItem(&This->tasks, toff, task);
            if (TaskNeedsRelease)
                TaskRelease(task, CurrentHeap());
        }
        CHECK_HEAP();
    }
    Mutex_Unlock(&This->mx);

    DBG(("END WsRes_done(%x, %s)\n", (ADDRESS) This, name));
    return sc;
}

/* tmp XXX - these are created from <pattern> or automaticaly */
#define RES_DEADLINE    CurrentTime()
#define RES_DEADLINE_OFF TIME_RELATIVE(TIME_MILLIS(1000))
#define RES_DURATION    TIME_RELATIVE(TIME_MILLIS(200))
#define RES_TOLERANCE   TIME_RELATIVE(TIME_MILLIS(20))
#define RES_PERIOD      TIME_RELATIVE(TIME_SECONDS(3))
#define RES_REPEATS     100
#define RES_CONFIDENCE  95

/* Loop over new pattern actions and create real reservations */
PRIVATE INT ReservationDefine(PTR r1, PTR tb)
{
    METHOD_DECLARE_NOTHIS(wsres.c, ReservationDefine);
    Preservation patres = r1;
    PITASKBUILDER Task = tb;
    PIRESERVATIONBUILDER Reservation;
    _TCHAR *ReservationName;
    TIME Period;

    sc = Task->v->CreateReservation(Task, 
                                    _T("HostUri"), _T("FeedbackUri"),
                                    _T("planexp1-reserve"),
                                    &Reservation);
    CHECKSC("create reservation");

    Int32ToInt64(Period, RES_PERIOD);
    sc = Reservation->v->SetRecurrence(Reservation, 
        Period, RES_REPEATS);
    CHECKSC("set reservation trigger");

    ReservationName = (_TCHAR *) Reservation->v->GetFullName(Reservation), 
    sc = Reservation->v->SetAdaptivity(
        Reservation, ReservationName, RES_CONFIDENCE, NULL, NULL);
    free(ReservationName);
    CHECKSC("set reservation adaptivity");

    METHOD_CLEANUP;
    return (INT) sc;
}

/* METHOD: Define new behavior pattern. */
PRIVATE SCODE WsRes_behavior(
    PIWSRESERVATION iThis,
    const _TCHAR *name,
    Preservation reslistarg,
    Psampling sampling)
{
    METHOD_DECLARE(WsRes, behavior);
    PIPLANNER Planner;
    PITASKBUILDER Task;

    sc = BindToObject(CurrentNameSpace(), _T("COB/planner.cob"), NAME_SPACE_READ,
                      &IID_IPlanner, (void **) &Planner);
    CHECKSC("create planner");

    sc = Planner->v->CreateTask(Planner, NULL, &Task);
    CHECKSC("create task");

    sc = (SCODE) ListWalk(&reslistarg, rloff, LIST_WALK_SCODE,
                          ReservationDefine, Task);
    CHECKSC("build actions");


METHOD_CLEANUP;
    return sc;
}

/* METHOD: Activate pattern at given time. */
PRIVATE SCODE WsRes_run(
    PIWSRESERVATION iThis,
    const _TCHAR *patternName,
    TIME deadline,
    const _TCHAR *deadlineAction,
    UINT confidence) /* ppm */
{
    METHOD_DECLARE(WsRes, Run);
    sc = E_NOT_IMPLEMENTED;

METHOD_CLEANUP;
    return sc;
}

#define TRANSMISSION_TIME_SAMPLE_POOL _T("This shouldn't be hard-coded")

/* METHOD: Receive feedback on processed actions. */
PRIVATE SCODE WsRes_feedback(
    PIWSRESERVATION iThis,
    const _TCHAR *taskName,
    Psample SampleList)
{
    METHOD_DECLARE(WsRes, feedback);
    Psample sample = SampleList;
    TIME delay;
    PISTATISTICS stats = NULL;

#if 0 /* Caching is pointlessly complicated so forget about it */
    stats = This->stats;
    if (!stats) {
#endif
        sc = BindToObject(CurrentNameSpace(),
                          _T("COB/statistics.cob"), NAME_SPACE_READ,
                          &IID_IStatistics, (void **) &stats);
        CHECKSC("bind statistics.cob");
#if 0
        Mutex_Lock(&This->mx);
        if (This->stats) {
            stats->v->Release(stats);
            stats = This->stats;
        } else
            This->stats = stats;
        Mutex_Unlock(&This->mx);
    }
#endif

    CHECKSETSC(sample, E_INVALID_PARAMETER, "empty sample list");

    assert(ListNodeIsAlone(soff, SampleList)); /* XXX handle more later */

    /* Subtract backwards so as to get negative (relative) values */
    delay = Int64Subtract(sample->BeginTime, sample->EndTime);
    PrintTime("----ProcessDelay", delay);

    delay = Int64Subtract(sample->BeginTime, sample->EndTime);
    PrintTime("----VirtualDelay", sample->Consumed);

    delay = Int64Subtract(sample->sendTime, sample->BeginTime);
    PrintTime("----MessageDelay", delay);

    if (!_tcsstr(sample->name, _T("producer"))) { /* XXX DEMO: skip prod */
        sc = stats->v->AddSample(stats,
                                 TRANSMISSION_TIME_SAMPLE_POOL, sample->seqno,
                                 TIME_ORIGIN, TIME_ORIGIN,
                                 TIME_ORIGIN, delay);
        CHECKSC("stats.AddSample trasmission");
    }
    sc = stats->v->AddSample(stats,
                             sample->name, sample->seqno,
                             sample->BeginTime, sample->EndTime,
                             sample->Estimated, sample->Consumed);
    CHECKSC("stats.AddSample");

METHOD_CLEANUP;
    RELEASE(stats);             /* !CACHE */
    return sc;
}

/* HEADER METHOD: Signal reservation completed, run triggers etc. */
PRIVATE SCODE WsRes_sendTime(
    PIWSRESERVATION iThis,
    PICONTINUATION iCont,
    TIME time)
{
    METHOD_DECLARE(WsRes, sendTime);
    PCONTINFO cont = pCONT(iCont);
    PWSCONTEXT ctx;

    sc = S_OK;
    ctx = cont->pContext;
    CHECKSETSC(ctx, E_OBJECT_NOT_INITIALIZED,
               "res:sendTime w/o preceding wsa:To");

    ctx->sendTime = time;

METHOD_CLEANUP;
    return sc;
}

/* HEADER METHOD: set correlation info */
PRIVATE SCODE WsRes_taskName(
    PIWSRESERVATION iThis,
    PICONTINUATION iCont,
    const _TCHAR *Value)
{
    METHOD_DECLARE(WsRes, taskName);
    PCONTINFO cont = pCONT(iCont);
    PWSCONTEXT ctx;

    sc = S_OK;
    ctx = cont->pContext;
    CHECKSETSC(ctx && ctx->taskName, E_OBJECT_NOT_INITIALIZED,
               "res:sendTime w/o preceding wsa:To");

    ctx->taskName = _tcsdup(Value);

METHOD_CLEANUP;
    return sc;
}

#if 0
struct _WsRes _wsres = {
    &WsReservationVtbl,
    1,
    MUTEX_INITIALIZER,

    NULL,
    NULL
};

PIUNKNOWN wsresCobMain(void)
{
    return (PIUNKNOWN) iReservation(&_wsres);
}
#endif
