/* Copyright (c) Microsoft Corporation. All rights reserved. */

#define DBGLVL 0
#include "ws.h"
#include <netif.h>

typedef struct WSROUTING {
    struct IWsRoutingVtbl const *v;
    UINT Refs;
} WSROUTING, *PWSROUTING;

#define pWsRouting(_x_) ((PWSROUTING)(_x_))

PRIVATE SCODE Allocate_And_Copy(PIHEAP Heap, _TCHAR** destn,
                                const _TCHAR* source, UINT size)
{
    *destn = (_TCHAR*) Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("soap:alloc&copy"),
                                    (size + 1)*sizeof(_TCHAR), 0);
    if (!*destn) {
        DBG(("typeinfo.c::Allocate_And_Copy: alloc failed.\n"));
        return E_NOT_ENOUGH_MEMORY;
    }
    _tcsncpy(*destn, source, size);
    (*destn)[size] = _T('\0');
    return S_OK;
}

typedef struct _REGISTRARY {
    UINT vid;
    PIENDPOINT pEndPoint;
    struct _REGISTRARY *next;
} *PREGISTRARY, *PITEM;

PREGISTRARY EndPointPool = NULL;

PRIVATE PITEM Search(PREGISTRARY registrary, UINT vid)
{
    PREGISTRARY reg = registrary;

    if (reg == NULL)
        return NULL;

    while (reg->vid != vid) {
        reg = reg->next;
        if (reg == NULL)
            return NULL;
    }

    return reg;
}

PRIVATE SCODE Register(PREGISTRARY *registrary, PIENDPOINT ep, UINT vid,
                       PIHEAP Heap)
{
    PREGISTRARY reg = *registrary;
    
    if (Search(*registrary, vid)) {
        TRACE("duplicate Vid");
        DebugBreak();
    }

    reg = Heap->v->Alloc(Heap, HEAP_DEBUG_NAME("soap:rout")| HEAP_ZERO_MEMORY,
                         sizeof(*reg), 0);
    if (reg == NULL) 
        return E_NOT_ENOUGH_MEMORY;

    reg->vid = vid;
    reg->pEndPoint = ep;
    if (ep)
        ep->v->AddRef(ep);

    reg->next = *registrary;
    *registrary = reg;
    
    return S_OK;
}

PRIVATE void Unregister(PREGISTRARY *registrary, UINT vid, PIHEAP Heap)
{
    PREGISTRARY *prev = registrary, reg = *prev;

    while (reg->vid != vid) {
        prev = &reg->next;
        reg = *prev;
    }
    assert(reg);

    *prev = reg->next;
    if (reg->pEndPoint)
        reg->pEndPoint->v->Release(reg->pEndPoint);
    Heap->v->Free(Heap, 0, reg);
}

PRIVATE UINT VidCounter = 1000000;
#define vloff offsetof(struct via, ListNode)


PRIVATE SCODE MCT WsRouting_QueryInterface(
    /*in*/ PIWSROUTING pAddr,
    /*in*/ REFIID Iid,
    /*out*/ void** ppUnk)
{
    SCODE sc;
    sc = GenericQueryInterface((IUnknown *)pAddr, Iid, ppUnk, &IID_IWsRouting);
    DBG(("WsRoutingQueryInterface: sc=%x unk=%x\n", sc, (ADDRESS) *ppUnk));
    return sc;
}

PRIVATE UINT MCT WsRouting_AddRef(
    /*in*/ PIWSROUTING pAddr)
{
    PWSROUTING pThis = pWsRouting(pAddr);
    UINT Refs = AtomicInc(&pThis->Refs);

    DBG(("WsRoutingAddRef: Refs=%d\n", Refs));

    return Refs;
}

PRIVATE UINT MCT WsRouting_Release(
    /*in*/ PIWSROUTING pAddr)
{
    UINT Refs;
    PWSROUTING pThis = pWsRouting(pAddr);

    Refs = AtomicDec(&pThis->Refs);

    if (Refs)
        return Refs;

    DBG(("WsRoutingRelease: refs=%d\n", Refs));

    return 0;
}

PRIVATE SCODE WsRouting_path(PIWSROUTING iThis, PICONTINUATION iCont,
                             const _TCHAR *action, const _TCHAR *to,
                             Pvia fwd,
                             Pvia rev,
                             const _TCHAR *from, const _TCHAR *id,
                             const _TCHAR *relatesTo)
{
    PCONTINFO cont = pCONT(iCont);
    PWSROUTING This = pWsRouting(iThis);
    PIHEAP Heap = CurrentHeap();
    SCODE sc;
    Pvia via;
    PITEM pItem;

    if (fwd && !ListNodeIsAlone(vloff, fwd)) {
        UINT vid;

        TRACE("FORWARD INTERMEDIARY");
        vid = VidCounter++;
        if (vid == 0)
            vid = VidCounter++;

        sc = Register(&EndPointPool, NULL, vid, Heap);
        if (FAILED(sc))
            return sc;

        /* do routing */
        via = ListRemoveFirst(&fwd, vloff);

        /* Make a reverse entry with continuation in it (vid) */
        /* Reuse the descriptor for a new reverse entry */
        via->target = NULL;
        via->vid = vid;

        ListAddFirst(&rev, vloff, via);
        goto RouteIt;
    } else if (fwd) {
        _TCHAR *pdest = NULL;
        PISOAPFACTORY soapfact;

        TRACE("ULTIMATE RECEIVER");
        /* Drop first and only entry */

        via = ListRemoveFirst(&fwd, vloff);
        /* URL required here */
        if (!via->target)
            return E_SCHEMA_CHECK;

        sc = BindToObject(CurrentNameSpace(), _T("COB/soap.cob"),
                          NAME_SPACE_READ, &IID_ISoapFactory,
                          (void **) &soapfact);
        if (FAILED(sc))
            return sc;
        sc = soapfact->v->ParseURL(soapfact, via->target,
                                   NULL/*flags*/, NULL/*host*/,
                                   NULL/*hostlen*/, NULL/*port*/,
                                   (const _TCHAR **)&pdest, NULL,
                                   NULL, NULL, NULL);

        if (SUCCEEDED(sc)) {    
#if 1
            DBG(("WsRouting_path: Local = %s -> %s\n", via->target, pdest));
            cont->To = pdest;
#else
            if (cont->ShouldFreeURL)
                free((void*) cont->reqLocalURL);
            sc = Allocate_And_Copy(Heap, (void*) &cont->reqLocalURL, 
                                   pdest, _tcslen(pdest) + 1);
            assert(sc == S_OK); /* XXX */
            cont->ShouldFreeURL = TRUE;
            DBG(("WsRouting_path: reqLocalURL = %s\n", This->reqLocalURL));
#endif

            if (via->reservation) {
                sc = Allocate_And_Copy(Heap, (void*) &cont->resname, 
                                       via->reservation,
                                       _tcslen(via->reservation) + 1);
                assert(sc == S_OK); /* XXX */
            }
        }
#if 0
        sc = This->Serializer->v->StartMessage(cont->Serializer,
                                               cont->envelopeNS,
                                               GetNamespace("soapenc"),
                                               cont->namespaces,
                                               rev,
                                               NULL, 
                                               NULL);
    This->MsgStarted = TRUE;
#endif
    } else if (!rev) {
        /* No reverse path, no forward path. Bad routing. */
        TRACE("wsroute: isLocal no path");
        return E_PATH_NOT_FOUND;
    } else if (!ListNodeIsAlone(vloff, rev)) {
        TRACE("REVERSE INTERMEDIARY");
        via = ListRemoveFirst(&rev, vloff);

        /* lookup the vid in the endpoint pool and set active endpoint.
           If not found, return error */
        assert(via->vid);
        pItem = Search(EndPointPool, via->vid);
        if (pItem) {
            DBG(("wsroute:isLocalCall vid not matched (XXX IGNORED)!\n"));
            Unregister(&EndPointPool, via->vid, Heap);
        }

        goto RouteIt;
    } else {
        TRACE("ORIGINAL SENDER");

        via = ListRemoveFirst(&rev, vloff);
        assert(via->vid);
        pItem = Search(EndPointPool, via->vid);
        if (pItem)
            Unregister(&EndPointPool, via->vid, Heap);
#if 0
        else
            return E_PATH_NOT_FOUND;
#endif
        sc = S_OK;
    }

 SavePath:
    assert(cont->pFwdVia == NULL);
    assert(cont->pRevVia == NULL);

    cont->pFwdVia = fwd;
    cont->pRevVia = rev;
    return sc;

 RouteIt:
/* Disabled - addressing replaces routing 
    sc = This->Serializer->v->StartMessage(This->Serializer,
                                           This->envelopeNS,
                                           GetNamespace(_T("soapenc")),
                                           This->namespaces,
                                           fwd,
                                           rev, 
                                           NULL);
*/
//    This->ultimateReceiver = FALSE;
    goto SavePath;
}

#include <s_WsRouting_v.c>

PIUNKNOWN WsRouteCobMain(void)
{
    PRIVATE struct WSROUTING rt = {&WsRoutingVtbl, 1};
    return (PIUNKNOWN) &rt;
}
