/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Microsoft MMLITE June 1995 --jvh
 * NT server for remote IFile using serial lines or sockets.
 */

#undef _DEBUG
#include <import_windows.h>
#include <mmlite.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <mmmacros.h>

#include <tchar.h>
#include <wchar.h>

#define hostfsd_debug_level 1
#define EPRINT(_a_) DBG(_a_)

/* An object for NT files/dirs
 */
typedef struct _NTFILE {
    const struct IFileVtbl *v;
    UINT RefCnt;
    HANDLE Handle;
} *PNTFILE;
#define pNF(_ifile_) ((PNTFILE)(_ifile_))
#define iNF(_file_) ((PIFILE)(_file_))

/* This is the same as above but different v-table type */
typedef struct _NTNS {
    const struct INameSpaceVtbl *v;
    UINT RefCnt;
    HANDLE Handle;
    _TCHAR Prefix[MAX_PATH];    /* relative path */
} *PNTNS;
#define pNS(_ins_) ((PNTNS)(_ins_))
#define iNS(_ns_) ((PINAMESPACE)(_ns_))


SCODE error_code_NT2MMLITE(SCODE sc, char *where)
{
    if (sc > 0) {
        /* Assume this is a Win32 error. */
        sc = STATUS_WIN32_ERROR(sc);
    }
    return sc;
}

#define HostFsd_NTBegin() (0)
#define HostFsd_NTEnd(_a_)

extern const struct IFileVtbl NtFileVtbl; /* forward */
extern const struct IFileVtbl NtSerialLineVtbl; /* forward */
extern const struct INameSpaceVtbl NtNsVtbl; /* ditto */

static PIFILE AllocNtFile(HANDLE handle, BOOL isSerial)
{
    PNTFILE File;

    File = (PNTFILE) malloc(sizeof(*File));
    if (!File)
        return NULL;

    memset(File, 0, sizeof(*File));

    if (isSerial)
        File->v = &NtSerialLineVtbl;
    else
        File->v = &NtFileVtbl;
    File->RefCnt = 1;
    File->Handle = handle;

    return iNF(File);
}

PRIVATE PNTNS AllocNtNs(const _TCHAR *name)
{
    PNTNS Ns;

    Ns = (PNTNS) malloc(sizeof(*Ns));

    if (Ns) {
        Ns->v = &NtNsVtbl;
        Ns->RefCnt = 1;
        Ns->Handle = INVALID_HANDLE_VALUE; /* no such thing for dirs */

        _tcscpy(Ns->Prefix, name);
    }

    return Ns;
}

PRIVATE UINT MCT NtFileAddRef(PIFILE This)
{
    TRACE(("NtFileADDREF"));
    return AtomicInc(&pNF(This)->RefCnt);
}

PRIVATE UINT MCT NtNsAddRef(PINAMESPACE This)
{
    TRACE(("NSaddref"));
    return AtomicInc(&pNS(This)->RefCnt);
}

PRIVATE SCODE MCT NtFileQueryInterface(PIFILE This, REFIID pIid, void* *ppNew)
{
    return GenericQueryInterface((PIUNKNOWN) This, pIid, ppNew, &IID_IFile);
}

PRIVATE SCODE MCT NtNsQueryInterface(PINAMESPACE This, REFIID pIid, void* *ppNew)
{
    return GenericQueryInterface((PIUNKNOWN) This, pIid, ppNew, &IID_INameSpace);
}

PRIVATE UINT MCT NtFileRelease(PIFILE pThis)
{
    SCODE sc = S_OK;
    PNTFILE File = pNF(pThis);
    INT Refs;
    BOOL ok;
    UINT State;

    Refs = AtomicDec(&File->RefCnt);
    DBG(("NtFileRELE%d", Refs));
    if (Refs == 0) {
        State = HostFsd_NTBegin();
        ok = CloseHandle(File->Handle);
        if (!ok) {
            sc = GetLastError();
            HostFsd_NTEnd(State);
            EPRINT(("NtFileRelease(%x) failed sc=%x Handle=%x\n",
                   File, sc, File->Handle));
        } else {
            HostFsd_NTEnd(State);
        }
        free(File);
    }

    return Refs;
}

PRIVATE UINT MCT NtNsRelease(PINAMESPACE pThis)
{
    PNTNS Ns = pNS(pThis);
    INT Refs;

    Refs = AtomicDec(&Ns->RefCnt);
    DBG(("NtNsRELE%d", Refs));
    if (Refs == 0) {
        free(Ns);
    }

    return Refs;
}

PRIVATE SCODE MCT NtFileGetSize(PIFILE This, PUINT64 pSize);

PRIVATE SCODE MCT
NtFileReadAt(PIFILE This, UINT64 Position, 
             BYTE *Buffer, UINT ByteCount, UINT * pSizeRead)
{
    PNTFILE File = pNF(This);
    SCODE sc = S_OK, sc1;
    OVERLAPPED ol;              /* NT type */
    UINT64 Size64;
    BOOL r;
    UINT State;

    *(UINT64*)&ol.Offset = Position;
    ol.hEvent = NULL;

#if 0
    DBG(("NtFileReadAt: read pos %I64x (x%x.%x) count x%x file %x\n",
         Position, ol.OffsetHigh, ol.Offset, ByteCount, (UINT32) File));
#endif

    State = HostFsd_NTBegin();
    r = ReadFile(File->Handle, Buffer, ByteCount, pSizeRead, &ol);
    if (r == FALSE)
        sc = GetLastError();
    HostFsd_NTEnd(State);
 
    if ((sc == S_OK) && (*pSizeRead < ByteCount)) {
        if (File->v == &NtSerialLineVtbl && *pSizeRead == 0)
            sc = E_TIMED_OUT;
        else
            sc = S_FALSE;
    } else if (sc == ERROR_HANDLE_EOF) {
        /* at or past EOF ? we need to know */
        sc1 = NtFileGetSize(This, &Size64);
        sc = S_FALSE;
        if (!FAILED(sc1) && (Position > Size64))
            sc = E_INVALID_PARAMETER;
    }
    else
        sc = error_code_NT2MMLITE(sc, "NtFileReadAt");

#if 0
    DBG(("NtFileReadAt: read(x%x) => x%x  sc=x%x\n",
            (UINT32) File, *pSizeRead, sc));
#endif

    return sc;
}

PRIVATE SCODE MCT
NtFileWriteAt(PIFILE This, UINT64 Position, 
               const BYTE *Buffer, UINT ByteCount, UINT *pSizeWritten)
{
    SCODE sc = S_OK;
    PNTFILE File = pNF(This);
    OVERLAPPED ol;              /* NT type */
    UINT State;

    *(UINT64*)&ol.Offset = Position;
    ol.hEvent = NULL;

    DBG(("NtFileWriteAt: write pos %I64x (x%x.%x) count x%x file %x\n",
         Position, ol.OffsetHigh, ol.Offset, ByteCount, (UINT32) File));

    State = HostFsd_NTBegin();
    if (WriteFile(File->Handle, Buffer, ByteCount, pSizeWritten, &ol) ==FALSE)
        sc = GetLastError();
    HostFsd_NTEnd(State);
        
    DBG(("NtFileWriteAt: write(x%x) => x%x  sc=x%x\n",
         (UINT32) File, *pSizeWritten, sc));

    return error_code_NT2MMLITE(sc, "NtFileWriteAt");
}

/* Get/Set a file's size
 */
PRIVATE SCODE MCT NtFileGetSize(PIFILE This, PUINT64 pSize)
{
    SCODE sc = S_OK;
    PNTFILE File = pNF(This);
    DWORD sizeLow, sizeHigh;
    UINT State;

    sizeHigh = -1;
    State = HostFsd_NTBegin();
    sizeLow = GetFileSize(File->Handle, &sizeHigh);
    if (sizeLow == 0xffffffff) {
        sc = GetLastError();
        HostFsd_NTEnd(State);
    } else {
        HostFsd_NTEnd(State);
        *pSize = (UINT64)(sizeLow) | (((UINT64)sizeHigh) << 32);
    }
    return error_code_NT2MMLITE(sc, "NtFileGetSize");
}

PRIVATE SCODE MCT NtFileSetSize(PIFILE This, UINT64 Size)
{
    SCODE sc = S_OK;
    PNTFILE File = pNF(This);
    UINT State;

    /* We do not truncate files (should ?)
     */
    sc = STATUS_WIN32_ERROR(  87);

    /* Cache probe ?
     */
    if (Size == (UINT64)(~0)) {
        EPRINT(("simfiled: Cache probe unimplemented\n"));
#if 0
        if (Conn->CacheTimestamp != CacheTimestamp) {
            sc = /*E_MEDIA_CHANGED*/ STATUS_WIN32_ERROR(1110);
            Conn->CacheTimestamp = CacheTimestamp;
        }
#endif
    } else {
        State = HostFsd_NTBegin();
        /* First seek to the right position and then set the size */
        if (SetFilePointer(File->Handle, (UINT32) Size, NULL, FILE_BEGIN)
            == 0xffffffff)
            sc = GetLastError();
        else if (!SetEndOfFile(File->Handle))
            sc = GetLastError();
        else
            sc = S_OK;
        HostFsd_NTEnd(State);
        sc = error_code_NT2MMLITE(sc, "NtFileSetSize");
    }
    return sc;
}

PRIVATE SCODE MCT NtSerialLineSetSize(PIFILE This, UINT64 Size)
{
    SCODE sc = S_OK;
    PNTFILE File = pNF(This);
    UINT State;
    COMMTIMEOUTS Timeouts;
    DCB ControlBlock;
    UINT32 *p = (UINT32*)&Size;
    
    switch (*p) {
    case 0: /* set recv timeout */
        /* Setup receive timeouts (none on xmit)
        * First char might take a long time.
        */
        Timeouts.ReadIntervalTimeout = 0;
        Timeouts.ReadTotalTimeoutConstant = p[1] / TIME_MILLIS(1);
        Timeouts.ReadTotalTimeoutMultiplier = 0;
        Timeouts.WriteTotalTimeoutMultiplier = 0;
        Timeouts.WriteTotalTimeoutConstant = 0;        

        State = HostFsd_NTBegin();
        if (!SetCommTimeouts( File->Handle, &Timeouts)) {
            sc = error_code_NT2MMLITE(GetLastError(), "NtSerialLineSetSize");
            DBG(("NtSerialLineSetSize: SetCommTimeouts() failed (x%x), ignoring..\n",
                 GetLastError()));
        }
        HostFsd_NTEnd(State);

        break;
    case 1: /* set xmit timeout */
        //Int32ToInt64(this->XmitTimeout,TIME_RELATIVE((INT32)p[1]));
        break;
    default: /* set baudrate */
    /* Get what we wont change
     */
        ControlBlock.DCBlength = sizeof(ControlBlock);
        if (!GetCommState(File->Handle,&ControlBlock)) {
            EPRINT(("GetCommState() failed on (x%x)\n",GetLastError()));
            /* might be a named pipe, ignore */
        }

        /* No parity, no flow..
        */
        ControlBlock.BaudRate = p[0];
        ControlBlock.fBinary = TRUE;     /* no EOF check */
        ControlBlock.fParity = FALSE;
        ControlBlock.fOutxCtsFlow = FALSE;
        ControlBlock.fOutxDsrFlow = FALSE;
        ControlBlock.fDtrControl = DTR_CONTROL_ENABLE;
        ControlBlock.fDsrSensitivity = FALSE;
        ControlBlock.fTXContinueOnXoff = TRUE;
        ControlBlock.fOutX = FALSE;
        ControlBlock.fInX = FALSE;
        ControlBlock.fErrorChar = FALSE;
        ControlBlock.fNull = FALSE;
        ControlBlock.fRtsControl = RTS_CONTROL_ENABLE;
        ControlBlock.fAbortOnError = FALSE;
        ControlBlock.ByteSize = 8;
        ControlBlock.Parity = NOPARITY;
        ControlBlock.StopBits = ONESTOPBIT;
        if (!SetCommState(File->Handle,&ControlBlock)) {
            EPRINT(("SetCommState() failed on (x%x)\n",GetLastError()));
            /* might be a named pipe, ignore */
        }

        break;
    }

    return sc;
}

PRIVATE SCODE MCT Unimp()
{
    return E_NOT_IMPLEMENTED;
}

const struct IFileVtbl NtFileVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NtFileQueryInterface, 
    NtFileAddRef, 
    NtFileRelease, 
    NtFileReadAt, 
    NtFileWriteAt, 
    NtFileSetSize, 
    NtFileGetSize
};

const struct IFileVtbl NtSerialLineVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NtFileQueryInterface, 
    NtFileAddRef, 
    NtFileRelease, 
    NtFileReadAt, 
    NtFileWriteAt, 
    NtSerialLineSetSize, 
    NtFileGetSize
};

static HANDLE MakePipe(
    const _TCHAR *name)
{
    HANDLE Pipe;

#define DEFAULT_SIZE (4*1024)
    Pipe = CreateNamedPipe(name,
                           PIPE_ACCESS_DUPLEX,
                           PIPE_TYPE_BYTE |
                           PIPE_READMODE_BYTE |
                           PIPE_WAIT,
                           PIPE_UNLIMITED_INSTANCES,
                           DEFAULT_SIZE,     // output buffer size
                           DEFAULT_SIZE,     // input buffer size
                           INFINITE,   // default timeout
                           NULL       // security attrib
                           );
    if (Pipe == INVALID_HANDLE_VALUE)
        return NULL;

    _tprintf(_T("Waiting for connection on pipe %s\n"), name);

    if (ConnectNamedPipe(Pipe,NULL))
        return Pipe;

    CloseHandle(Pipe);
    return NULL;
}

/*------------------ "methods" ------------------------------*/
#define hostfsd_may_write TRUE

SCODE NtFileOpen(const _TCHAR *Prefix, const _TCHAR *filename, UINT flags, PTR *ppNew)
{
    SCODE sc = S_OK;
    UINT openflags = 0, createflags = 0, Handle;
    BOOL isdir = FALSE;
    HANDLE file = NULL;
    UINT State;

    if (flags & NAME_SPACE_FAILIFEXIST)         /* 0x20 */
        createflags = CREATE_NEW;
    else if (flags & NAME_SPACE_CREATE)         /* 0x10 */
        createflags = OPEN_ALWAYS;
    else
        createflags = OPEN_EXISTING;

    if (flags & NAME_SPACE_READ)                /* 1 */
        openflags |= GENERIC_READ;
    if (flags & NAME_SPACE_WRITE)               /* 2 */
        openflags |= GENERIC_WRITE;

    if (flags & NAME_SPACE_EXTENSION) {         /* mkdir */
        isdir = TRUE;
        if (!(openflags & GENERIC_WRITE)) {
            TRACE(("Can't create directory without NAME_SPACE_WRITE\n"));
            sc = STATUS_WIN32_ERROR(5);
            goto replystuff;
        }
    }

    if (!hostfsd_may_write
        && (openflags != GENERIC_READ || createflags != OPEN_EXISTING))
    {
        DBG(("Rejected \"%ls\" flags=x%x because server is Read Only\n",
                   filename, flags));
        sc = E_ACCESS_DENIED;
        goto replystuff;
    }

#if 0
    /* Check for \\.\ escape */
    if (_tcsncmp(filename, _T("...."), 4) == 0) {
        filename[0] = filename[1] = filename[3] = _T('\\');
        filename[2] = _T('.');
    }
#endif

    Handle = 0;
    if (isdir) {
        State = HostFsd_NTBegin();
        /* In the end it might be better to have a separate RPC for mkdir */
        if (CreateDirectory(filename, NULL) == FALSE) {
            sc = GetLastError();
            HostFsd_NTEnd(State);
            EPRINT(("NOT! "));
        } else {
            HostFsd_NTEnd(State);
            *ppNew = AllocNtNs(filename);
        }
    } else {
        BOOL isSerial = FALSE;
        State = HostFsd_NTBegin();
        if (filename[0] == L'C' && filename[1] == L'O' && filename[2] == L'M'
            && filename[3] >= L'0' && filename[3] <= L'9' ) {
            /* open a serial port */
            file = CreateFile( filename, GENERIC_READ | GENERIC_WRITE,
                       0, /* exclusive access */
                       NULL, /* no security attributes */
                       OPEN_EXISTING,
                       0, /* Not overlapped */
                       NULL  /* hTemplate - must be NULL for comms */
                       );
            isSerial = TRUE;
        } else if ((flags & NAME_SPACE_CREATE)
                   && filename[0] == L'\\' && filename[1] == L'\\'
                   && !_tcsncmp(filename+2+_tcscspn(filename+2, _T("\\")),
                               _T("\\pipe\\"), 6)) {
            /* Named pipe \\host\pipe\foo */
            file = MakePipe(filename);
        } else {
            file = CreateFile(filename,
                               openflags,
                               FILE_SHARE_READ|FILE_SHARE_WRITE,
                               NULL,
                               createflags,
                               FILE_ATTRIBUTE_NORMAL,
                               NULL);
        }
        if (file == INVALID_HANDLE_VALUE) {
            sc = GetLastError();
            HostFsd_NTEnd(State);
            EPRINT(("NOT! "));
        } else {
            HostFsd_NTEnd(State);
            *ppNew = AllocNtFile(file, isSerial);
            
            if (isSerial) { 
                PIFILE File = *ppNew;
                if (File) {
                    /* set the default Baud rate */
                    sc = File->v->SetSize(File, 38400);
                } else {
                    sc = E_INVALID_PARAMETER;
                }
            }
        }
    }

    EPRINT(("Opened \"%s\" %s as file x%x handle x%x sc=x%x flags=x%x\n",
               filename,
               (openflags & GENERIC_WRITE)
               ? (isdir ? "New directory" : "WRITABLE")
               : "read only",
               (UINT) Handle, file, sc, flags));
 replystuff:
    return error_code_NT2MMLITE(sc, "do_open");
    }

SCODE NtDelete(const _TCHAR *Prefix, const _TCHAR *filename)
{
    BOOL r;
    SCODE sc = S_OK;
    UINT State;

    if (!hostfsd_may_write) {
        r = FALSE;
        sc = E_ACCESS_DENIED;
        goto Done;
    }

    /* try first to delete as file, then as dir.
     * yes, this mangles error reporting a bit.
     */

    State = HostFsd_NTBegin();
    r = DeleteFile(filename);
    if (!r) {
        sc = GetLastError();
        if (sc == ERROR_ACCESS_DENIED) {
            r = RemoveDirectory(filename);
            if (!r)
                sc = GetLastError();
            else
                sc = S_OK;
        }
    }
    HostFsd_NTEnd(State);

 Done:
    EPRINT(("%sDeleted \"%ls\" sc=x%x\n",
            (r) ? "" : "NOT! ",
            filename,
            sc));

    return error_code_NT2MMLITE(sc, "do_delete");
}

/* 
 */
SCODE findnext_helper(const _TCHAR *prefix, const _TCHAR *previous,
                      WIN32_FIND_DATA *search)
{
    _TCHAR path[MAX_PATH * 2];
    HANDLE dir;
    SCODE sc = S_OK;
    UINT State;

    _tcscpy(path,prefix);
    if (prefix[0] != _EOS)
        _tcscat(path, _T("\\"));
    _tcscat(path, _T("*"));

    /* Assume error */
    sc = error_code_NT2MMLITE(ERROR_FILE_NOT_FOUND,"do_findnext");

    State = HostFsd_NTBegin();
    dir = FindFirstFile(path,search);
    DBG(("findnext_helper FFW %s->%x %x\n", path, dir, GetLastError()));

    if (dir == INVALID_HANDLE_VALUE) {
        HostFsd_NTEnd(State);
        return sc;
    }

    if (previous[0] == _EOS)  /* This is a FindFirst */
        sc = S_OK;
    else {
        do {
            if (_tcscmp(previous,search->cFileName) == 0) { /* Found previous */
                WIN32_FIND_DATA Search2;
                if (FindNextFile(dir,search)) { /* Found next */
                    /* FindNext is supposed to return S_FALSE when it returns the last entry
                     * See if this is the last one and if so return S_FALSE + name
                     */
                    Search2 = *search; /* make copy so as to not mess up returned values */
                    if (FindNextFile(dir, &Search2))
                        sc = S_OK;
                    else
                        sc = S_FALSE;
                } else {                          /* Previous was last */
                    /* Search beyond end. Say this is last but return empty string */
                    search->cFileName[0] = _EOS;
                    sc = S_FALSE;
                }
                break;
            }
        } while (FindNextFile(dir,search));
    }

    FindClose(dir);
    HostFsd_NTEnd(State);
    return sc;
}

PRIVATE SCODE ComposePath(_TCHAR *NewPath, UINT MaxLen,
                          const _TCHAR *Prefix, const _TCHAR *Name,
                          UINT *pLengthInBytes)
{
    UINT len1, len2;

    len1 = _tcslen(Prefix);
    len2 = Name ? _tcslen(Name) : 0;
    if ((len1+len2+2) > MaxLen) {
        /* So what should we do? */
        DBG(("ComposePath \"%s\" \"%s\" too long %d+%d+2 > %d\n",
             Prefix, Name ? Name : _T("NULL"), len1, len2, MaxLen));
        *pLengthInBytes = 0;
        return E_INVALID_PARAMETER; /* reject it altogether */
    }

    /* Combine Prefix and Name  "Prefix/Name" or "Prefix" or "Name" or "" */
    NewPath[0] = _EOS;
    if (len1) {
        _tcscpy(NewPath, Prefix);
        /* add NT separator */
        if (len2) {
            NewPath[len1++] = _T('\\');
            /* beware: no EOS */
        }
    }            
    if (len2)
        _tcscpy(NewPath+len1, Name);

    /* Remove trailing slash */
    len1 += len2;
    if (len1 > 0 && IS_PATH_SEPARATOR(NewPath[len1 - 1]))
        NewPath[--len1] = _EOS; /* NB: account for shorter strlen */

    DBG(("ComposePath \"%s\" \"%s\" -> \"%s\" => %d\n",
         Prefix, Name ? Name : _T("NULL"), NewPath, len1+1));
    *pLengthInBytes = len1 + 1; /* _EOS counts! */
    return NO_ERROR;
}

PRIVATE UINT MCT NtNSRelease(PINAMESPACE This)
{
    PNTNS Ns = (PNTNS) This;
    INT Refs;

    Refs = AtomicDec(&Ns->RefCnt);
    DBG(("NSrele%x ", Refs));
    if (Refs == 0)
        free((PTR) Ns);
    return Refs;
}

/* Give "inetaddr:inetport/remotename", get IFile.  :inetport is optional */
PRIVATE SCODE MCT NtNsBind(PINAMESPACE This, const _TCHAR *Name, NAME_SPACE_FLAGS Flags, PIUNKNOWN *Object)
{
    PNTNS Ns = (PNTNS) This, NewNs;
    UINT len;
    SCODE sc = NO_ERROR;
    PNTFILE File;

    DBG(("NtNSBind(x%x \"%s\" x%x)\n", (ADDRESS) This, Name, Flags));
    /* Check flags. Parse name. Open remote file. */
    if (Object == NULL)
        return E_INVALID_PARAMETER;
    *Object = NULL;
    if (Name == NULL)
        return E_INVALID_PARAMETER;

    Ns->v->AddRef((PINAMESPACE) Ns);
    
    len = _tcslen(Name);
    if (Name[0] == _EOS) {      /* DIR start directory or this directory */
        *Object = (PIUNKNOWN) Ns;       /* May be NULL */
        DBG(("NtNSBind Ns: Name=\"%s\" Ns=x%x sc=x%x\n",
                 Name, (ADDRESS) Ns, sc));
    } else if (IS_PATH_SEPARATOR(Name[len - 1])) { /* DIR */
        /* 
         * BUGBUG: mind "/" root directory and "//" UNC quotes etc.
         * BUT, they are not seen in practice.
         */
        /* Make sure the remembered prefix is big enough.  Make dynamic! */
        if (_tcslen(Ns->Prefix) + len + 2 > MAX_PATH) {
            sc = E_INVALID_PARAMETER;
            goto ErrorA;
        }

        NewNs = AllocNtNs(_T(""));
        if (!NewNs) {
            goto ErrorA;
        }

        sc = ComposePath(NewNs->Prefix, MAX_PATH, Ns->Prefix, Name, &len);
        Ns->v->Release((PINAMESPACE) Ns);
        if (FAILED(sc))
            goto ErrorB;
        DBG(("NtNSBind subdir: NewNs=%x Prefix=\"%s\"\n",
                 (ADDRESS) NewNs, NewNs->Prefix));

        /* Check if it really exists on remote or create it. */
        if (Flags & NAME_SPACE_CREATE) {
            /* Create it! */
            sc = NtFileOpen(_T(""), NewNs->Prefix, Flags | NAME_SPACE_EXTENSION, &Ns);
            if (FAILED(sc)) {
                DBG(("NtNSBind: create subdir \"%s\" failed %x\n",
                         NewNs->Prefix, sc));
                goto ErrorB;
            }
        } else {
            /* Does it exist then? */
            _TCHAR tmpbuf[2];
            UINT tmp;
            sc = NewNs->v->FindNext((PINAMESPACE) NewNs, _T(""), _T(""),
                                    tmpbuf, sizeof(tmpbuf), &tmp);
            if (FAILED(sc)) {
                DBG(("NtNSBind: nonexisting subdir \"%s\" rejected %x\n",
                         NewNs->Prefix, sc));
                goto ErrorB;
            }
        }

        *Object = (PIUNKNOWN) NewNs;
        return NO_ERROR;
    ErrorA:
        Ns->v->Release((PINAMESPACE) Ns);
        *Object = NULL;
        return sc;
    ErrorB:
        NewNs->v->Release((PINAMESPACE) NewNs);
        *Object = NULL;
        return sc;
    } else {                    /* FILE */
        sc = NtFileOpen(Ns->Prefix, Name, Flags, &File);
        DBG(("NtNSBind File: Name=%s Ns=x%x file=x%x fla=%x sc=x%x\n",
                 Name, (ADDRESS) Ns, (ADDRESS) File, Flags, sc));
        *Object = (PIUNKNOWN) File;     /* may be NULL */
        Ns->v->Release((PINAMESPACE) Ns);               /* not returned */
    }
    return sc;
}

PRIVATE SCODE MCT NtNsFindNext(PINAMESPACE This, const _TCHAR *Prefix,
         const _TCHAR *Previous,_TCHAR *Buffer, UINT BufSize, UINT *StrLen_out)
{
    SCODE sc;
    PNTNS Ns = (PNTNS) This;
    UINT len = 0;
    UINT16 offset_of_previous = 0;
    WIN32_FIND_DATA SearchData;

    if (Previous == NULL)
      Previous = _T("");

    DBG(("NtNSFindNext(Ns=x%x Ns->Prefix=\"%s\" Prefix=\"%s\" Previ=\"%s\" Bs=%d)\n",
         (ADDRESS) Ns, Ns->Prefix, Prefix ? Prefix : _T("NULL"), Previous,
         BufSize));

    /* at least the _EOS must fit in Buffer */
    if (BufSize < sizeof(_TCHAR) || Buffer == NULL) {
        DBG(("bad NtNSFindNext, args=x%x,\"%s\",\"%s\",x%x,%d,x%x\n",
             (ADDRESS) This, Prefix ? Prefix : _T("NULL"), Previous,
             (ADDRESS)Buffer, BufSize, (ADDRESS)StrLen_out));
        return E_INVALID_PARAMETER;
    }

    if (!Prefix)
        Prefix = _T("");
    sc = findnext_helper(Prefix, Previous, &SearchData);

    if (FAILED(sc))
        goto Done;

    /* XXX return special status when buf too small */
    len = BufSize/sizeof(_TCHAR);
    _tcsncpy(Buffer, SearchData.cFileName, len);
    Buffer[len-1] = _EOS;
    if (StrLen_out)
      *StrLen_out = _tcslen(Buffer);
 Done:
    return sc;
}

PRIVATE SCODE MCT NtNsUnregister(PINAMESPACE This, const _TCHAR * Name)
{
    PNTNS Ns = (PNTNS) This;
    UINT len;
    SCODE sc;

    DBG(("NtNSUnregister(x%x \"%s\"+\"%s\")\n",
             (ADDRESS) This, Ns->Prefix, Name));

    /* Make sure it fits
     */
    len = _tcslen(Name);
    if (len + _tcslen(Ns->Prefix) + 2 > MAX_PATH)
        return E_INVALID_PARAMETER;

    sc = NtDelete(Ns->Prefix, Name);

    return sc;
}

const struct INameSpaceVtbl NtNsVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    NtNsQueryInterface,
    NtNsAddRef,
    NtNsRelease,
    (_INameSpace_Register_METHOD_TYPE) Unimp,
    NtNsUnregister,
    NtNsBind,
    NtNsFindNext,
    (_INameSpace_GetCapabilities_METHOD_TYPE) Unimp
};

PINAMESPACE InitNtFile()
{
    return (PINAMESPACE) AllocNtNs(_T(""));
}
