/* Copyright (c) Microsoft Corporation. All rights reserved. */
#define DBGLVL 0

#include <import_windows.h>
#include <mmlite.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <mmmacros.h>

#include <tchar.h>
#include <wchar.h>

#ifndef HEAP_CREATE_ENABLE_EXECUTE
/* Required on ia64 but "cleverly" hidden on other platforms */
#define HEAP_CREATE_ENABLE_EXECUTE 0x00040000
#endif

typedef struct {
    ADDR_SIZE Reserve;  /* amount of address space reserved for the heap */
    ADDR_SIZE Commit;   /* of that, how much is actually committed */
    ADDR_SIZE Used;     /* of that, how much is actually allocated */
    ADDR_SIZE MaxUsed;
    UINT Allocs;        /* total number of alloc operations */
} HEAP_STATUS;

typedef struct _WINHEAP {
    const struct IHeapVtbl *v;
    UINT RefCnt;
    HANDLE h;
    BOOL IsSubHeap;
    HEAP_STATUS Status;
} *PWINHEAP;
#define pWH(_i_) ((PWINHEAP) (_i_))

#define IncUsedCnt(_phi_,_size_) \
{ \
      (_phi_)->Status.Used += (_size_); \
      if ((_phi_)->Status.Used > (_phi_)->Status.MaxUsed) \
          (_phi_)->Status.MaxUsed = (_phi_)->Status.Used; \
      if ((_phi_)->Status.Used > (_phi_)->Status.Commit) \
          (_phi_)->Status.Commit = (_phi_)->Status.Used; \
}
#define DecUsedCnt(_phi_,_size_) ((_phi_)->Status.Used -= (_size_))

/*
 * Exported methods
 */
PRIVATE SCODE CWinHeapQueryInterface(PIHEAP iThis, REFIID pIid, void* *ppUnk)
{
    return GenericQueryInterface((IUnknown *)iThis,pIid,ppUnk,&IID_IHeap);
}

PRIVATE UINT CWinHeapAddRef(PIHEAP iThis)
{
    PWINHEAP This = pWH(iThis);
    return AtomicInc(&This->RefCnt);
}

PRIVATE UINT CWinHeapRelease(PIHEAP iThis)
{
    PWINHEAP This = pWH(iThis);
    UINT Refs;

    Refs = AtomicDec(&This->RefCnt);

    if (Refs == 0) {
        BOOL Sub = This->IsSubHeap;
        HANDLE h = This->h;
        /* In this implementation, the heap header is allocated from the
         * heap itself.
         */
        iThis->v->Free(iThis,0,This);
        if (Sub)
            HeapDestroy(h);
    }

    return Refs;
}

PRIVATE PTR CWinHeapAlloc(PIHEAP iThis, UINT Flags, UINT Size, UINT Alignment)
{
    PWINHEAP This = pWH(iThis);
    PTR rv;

    rv = HeapAlloc(This->h, Flags & HEAP_FLAGS_MASK, Size);
    if (!rv) {
        DBG(("CWinHeapAlloc: Alloc size=%d flags=%x failed sc=%d\n",
             Size, Flags & HEAP_FLAGS_MASK, GetLastError()));
        return NULL;
    }

    IncUsedCnt(This, Size);

    return rv;
}

PRIVATE PTR CWinHeapRealloc(PIHEAP iThis, UINT Flags, PTR pMem, UINT NewSize, UINT Alignment)
{
    PWINHEAP This = pWH(iThis);
    PTR rv;
    UINT OldSize = iThis->v->Size(iThis, Flags , pMem);

    rv = HeapReAlloc(This->h, Flags & HEAP_FLAGS_MASK, pMem, NewSize);
    if (!rv)
        return NULL;

    DecUsedCnt(This, OldSize);
    IncUsedCnt(This, NewSize);

    return rv;
}

PRIVATE BOOL CWinHeapFree(PIHEAP iThis, UINT Flags, PTR pMem)
{
    PWINHEAP This = pWH(iThis);
    UINT OldSize = iThis->v->Size(iThis, Flags, pMem);

    DecUsedCnt(This, OldSize);

    return HeapFree(This->h, Flags & HEAP_FLAGS_MASK, pMem);
}

PRIVATE ADDR_SIZE CWinHeapSize(PIHEAP iThis, UINT Flags, PTR pMem)
{
    PWINHEAP This = pWH(iThis);

    return HeapSize(This->h, Flags & HEAP_FLAGS_MASK, pMem);
}

PRIVATE SCODE CWinHeapValidate(PIHEAP iThis, UINT Flags, PTR pMem)
{
    PWINHEAP This = pWH(iThis);

    return HeapValidate(This->h, Flags, pMem) ? S_OK : E_FAIL;
}

PRIVATE SCODE CWinHeapStatus(PIHEAP iThis, ADDR_SIZE *pReserve, ADDR_SIZE *pCommit, ADDR_SIZE *pUsed, ADDR_SIZE *pMaxUsed)
{
    PWINHEAP This = pWH(iThis);

    if (pReserve)
        *pReserve = This->Status.Reserve;
    if (pCommit)
        *pCommit = This->Status.Commit;
    if (pUsed)
        *pUsed = This->Status.Used;
    if (pMaxUsed) {
        do {
            *pMaxUsed = This->Status.MaxUsed;
/* BUGBUG Replace with new AtomicCmpAndSwapAddress() */
/* BUGBUG Replace with new AtomicCmpAndSwapAddress() */
        } while (!AtomicCmpAndSwap((UINT*) &This->Status.MaxUsed,
                                   (UINT) *pMaxUsed,
                                   (UINT) This->Status.Used));
    }

    return S_OK;
}

PRIVATE PTR CWinHeapExtract(PIHEAP iThis, UINT Flags, PTR pMem, UINT Size)
{
    PWINHEAP This = pWH(iThis);

    return NULL;
}

const struct IHeapVtbl CWinHeapVtbl;

PRIVATE SCODE CWinHeapCreateHeap(PIHEAP pHeap,
                                 UINT Flags, UINT InitSize, UINT MaxSize,
                                 PIHEAP *ppHeap)
{
    HANDLE h;
    PWINHEAP NewHeap;

    Flags &= Flags & HEAP_FLAGS_MASK;
    Flags |= HEAP_CREATE_ENABLE_EXECUTE;

    h = HeapCreate(Flags | HEAP_CREATE_ENABLE_EXECUTE, InitSize, 0);

    if (h == INVALID_HANDLE_VALUE)
        return E_INVALID_PARAMETER;

    NewHeap = HeapAlloc(h, HEAP_ZERO_MEMORY, sizeof(*NewHeap));
    if (!NewHeap) {
        HeapDestroy(h);
        return E_NOT_ENOUGH_MEMORY;
    }

    NewHeap->v = &CWinHeapVtbl;
    NewHeap->RefCnt = 1;
    NewHeap->h = h;
    NewHeap->IsSubHeap = TRUE;

    NewHeap->Status.Reserve = MaxSize;
    NewHeap->Status.Commit = MaxSize;

    *ppHeap = (PIHEAP) NewHeap;
    return S_OK;
}

const struct IHeapVtbl CWinHeapVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER
    CWinHeapQueryInterface,
    CWinHeapAddRef,
    CWinHeapRelease,
    CWinHeapAlloc,
    CWinHeapRealloc,
    CWinHeapFree,
    CWinHeapSize,
    CWinHeapValidate,
    CWinHeapExtract,
    CWinHeapStatus,
    CWinHeapCreateHeap
};

PTR *CheckDelayedFree(void)
{
    static PTR pDelayedFree = NULL;
    return &pDelayedFree;
}

 /* Create the system heap.  Left to machdep code because messy otherwise.
  * The heap should account for all available memory, carved out of memory
  * that is already used or unavailable (like the 640K-1meg hole on PCs).
  * It does not have to account for all the memory in the system.
  * Returns a PIHEAP pointer.
  */
PIHEAP MachineHeapCreate( void )
{
    HANDLE h = HeapCreate(HEAP_CREATE_ENABLE_EXECUTE, 0, 0);
    PWINHEAP This = HeapAlloc(h, HEAP_ZERO_MEMORY, sizeof(*This));

    This->v = &CWinHeapVtbl;
    This->RefCnt = 1;
    This->h = h;

    /* In reality the heap is extensible but tends to get very fragmented */
    /* Size tuned for Doom with account to broken system heap */
    This->Status.Reserve = 2*1024*1024;
    This->Status.Commit = 0;

    return (PIHEAP) This;
}

PIHEAP RTLCALLTYPE CreateHeapFrom(ADDRESS Mem,
                      UINT Flags, ADDR_SIZE InitSize, ADDR_SIZE MaxSize)
{
    PIHEAP pHeap = CurrentHeap(), NewHeap;
    SCODE sc;

    UnusedParameter(Mem);
    sc = pHeap->v->CreateHeap(pHeap, Flags, InitSize, MaxSize, &NewHeap);
    if (FAILED(sc))
        return NULL;
    return NewHeap;
}
