/* Copyright (c) Microsoft Corporation. All rights reserved. */

#define _WIN32_WINNT 0x0500 /* For SignalObjectAndWait */
#include <import_windows.h>

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <mmmacros.h>
#include "../base/process.h"    /* for CurrentNameSpace */

CONDITION SleepCond = CONDITION_INITIALIZER;

SCODE SleepUntil(TIME t)
{
    /* Wait on default condition variable */
    return Condition_Wait(&SleepCond, NULL, &t);
}

/* Same as above but pops its own stack. Used in thread creation only. */
PRIVATE SCODE __stdcall _stdcall_SleepUntil(TIME t)
{
    /* Wait on default condition variable */
    return Condition_Wait(&SleepCond, NULL, &t);
}

typedef struct _THDINFO *PTHDINFO;
typedef struct _THDINFO {
    const struct IThreadVtbl *v;
    UINT RefCnt;               /* Reference count (see below)               */

    HANDLE Handle;
    PIPROCESS StartedIn;       /* initially created in this process/module  */
    PBYTE  StackBase;          /* Stack info                                */
    UINT   StackSize;
    THREAD_FUNCTION pStart;
    THREAD_ARGUMENT Arg;
    SCODE LastError;
    PTR RpcContinuation;
    BOOL Suspended;
} THDINFO;
#define pTH(pThd)  ((PTHDINFO)  (pThd))

/* Tls slot that points to "current thread" */
DWORD ThreadTlsId = 0;

PRIVATE SCODE MCT ThreadQueryInterface( 
            IThread *pThis,
            /* [in] */ REFIID pIid,
            /* [out] */ void **ppUnk)
{
    return GenericQueryInterface((IUnknown *)pThis,pIid,ppUnk,&IID_IThread);
}
        
PRIVATE UINT MCT ThreadAddRef( 
            IThread *pThis)
{
    UINT RefCnt;

    RefCnt = AtomicInc(&pTH(pThis)->RefCnt);

    return RefCnt;
}

INT CheckStackUsage(IThread *pThis)
{
    return 0;
}
        
PRIVATE UINT MCT ThreadRelease ( 
            IThread *pThis)
{
    UINT RefCnt;

    RefCnt = AtomicDec(&pTH(pThis)->RefCnt);

#ifdef _DEBUG
    printf("ThreadRelease (0x%x)->%d stack usage: %d/%d\n",
           (ADDRESS)pThis, RefCnt,
           CheckStackUsage(pThis), pTH(pThis)->StackSize);
#endif

    if (RefCnt == 0) {
        PIPROCESS pPrc;
        PIHEAP pIHeap = CurrentHeap();
        HANDLE h = pTH(pThis)->Handle;
        BOOL ok;

        pPrc = pTH(pThis)->StartedIn;
        pIHeap->v->Free(pIHeap, 0, pThis);
        /* If ThreadExit did not zero StartedIn, this must be a thread that
         * never ran and thus this is from normal thread context so no
         * need to worry about locks etc. In any case the ref needs to go.
         */
        if (pPrc) {
#ifdef _DEBUG
            printf("Thread (0x%x) dead. DID NOT RUN.\n", (ADDRESS)pThis);
#endif
            pPrc->v->Release(pPrc);
            /* Also get rid of Windows thread */
            TerminateThread(h, 0);
        } else {
#ifdef _DEBUG
            printf("Thread (0x%x) dead.\n", (ADDRESS)pThis);
#endif
        }
        ok = CloseHandle(h);
        assert(ok);
    }

    return RefCnt;
}

/* Disagreeable call this one.
 */
PRIVATE SCODE MCT ThreadGetProcess ( 
            IThread *pThis,
            /* [out] */ IProcess **ppIProcess)
{
    if (ppIProcess) {
        IProcess *pPrc = pTH(pThis)->StartedIn;
        pPrc->v->AddRef(pPrc);
        *ppIProcess = pPrc;
    }

    return NO_ERROR;
}

/* Get partially initialized thread going.
 */
PRIVATE SCODE ThreadStartWithConstraint (
    PITHREAD pThis,
    PTIME_CONSTRAINT pcval,
    THREAD_FUNCTION pStart,
    THREAD_ARGUMENT Arg,
    PTR StackTop,
    UINT ArgCount)
{
    PTHDINFO This = pTH(pThis);

    UnusedParameter(ArgCount);

    This->pStart = pStart;
    This->Arg = Arg;
    assert(This->Suspended);
    if (!This->Suspended)
        return E_OBJECT_NOT_INITIALIZED; /* invalid state */
    This->Suspended = FALSE;

    if (StackTop) {
        /* Caller pushed some frames. Push more on top */
        CONTEXT cxt;
        UINT *stk = StackTop;

        /* Where to return after pStart(): Arg is usually ThreadExit(void) */
        *(--stk) = (ADDRESS) Arg;
        cxt.ContextFlags = CONTEXT_CONTROL;
        GetThreadContext(This->Handle, &cxt);
        cxt.Ebp = 0; /* to terminate stack trace */

#if 1
        /* Push extra frame that calls __stdcall TlsSetValue(TlsId, pThis)
         * This is needed so that CurrentThread() works.
         * The args are conveniently popped by the callee--for once stupidcall
         * is useful!
         */
        cxt.Eip = (ADDRESS) TlsSetValue;
        *(--stk) = (ADDRESS) pThis;
        *(--stk) = ThreadTlsId;
        *(--stk) = (ADDRESS) pStart; /* then execute at pStart(void) */
#else
        cxt.Eip = (ADDRESS) pStart;
#endif
#if 1
        /* Make sure we don't run before StartTime arrives */
        if (pcval) {
            /* Extra frame calls SleepUntil */
            stk -= 2;           /* 64bit arg */
            *(TIME*) stk = pcval->Start;
            *(--stk) = cxt.Eip;
            cxt.Eip = (ADDRESS) _stdcall_SleepUntil;
        }
#endif
        cxt.Esp = (ADDRESS) stk;
        SetThreadContext(This->Handle, &cxt);
    }

    /* Let scheduler know about this new thread
     */
    pThis->v->AddRef(pThis);     /* One ref for the running thread */

    ResumeThread(This->Handle);

    return S_OK;
}

PRIVATE UINT ThreadGetValue(PITHREAD pTh, THREAD_VALUE Index)
{
    switch (Index) {
    case THREAD_VALUE_LAST_ERROR:
        return pTH(pTh)->LastError;
    case THREAD_VALUE_RPC_CONTINUATION:
        return (UINT) pTH(pTh)->RpcContinuation;
    default:
        return 0;
    }
}

PRIVATE void ThreadSetValue(PITHREAD pTh, THREAD_VALUE Index, UINT Value)
{
    switch (Index) {
    case THREAD_VALUE_LAST_ERROR:
        pTH(pTh)->LastError = Value;
        break;
    case THREAD_VALUE_RPC_CONTINUATION:
        pTH(pTh)->RpcContinuation = (PTR) Value;
        break;
    default:
        ;
    }
}

#include <s_CThread_v.c>

DWORD WINAPI NT_ThdProc(LPVOID param)
{
    PTHDINFO This = (PTHDINFO) param;

    TlsSetValue(ThreadTlsId, param);

    This->pStart(This->Arg);

    ThreadExit();
    return 0;
}

#include <winternl.h>

SCODE MCT ThreadCreate( 
            IProcess *pProcess,
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg,
            ADDR_SIZE StackSize,
            UINT32 Flags,
            /*optional*/     PIVMVIEW pVmView,
            /*optional,out*/ PTR *pStackTop,
            /*optional,out*/ IThread **ppNewThread)
{
    HANDLE h;
    PTHDINFO th;
    PIHEAP pIHeap = CurrentHeap();

    Flags = 0;

    th = (PTHDINFO) pIHeap->v->Alloc(pIHeap,
                                     HEAP_DEBUG_NAME("IThread"),
                                     sizeof(THDINFO),
                                     0);
    if (th == NULL)
        goto ErrorExit;

    memset(th,0,sizeof(*th));

    th->v = &CThreadVtbl;
    th->RefCnt = (ppNewThread && !pStackTop) ? 2 : 1;
    th->StartedIn = pProcess;
    th->pStart = pStart;
    th->Arg = Arg;

    /* Take a reference on the process for this thread.
     */
    if (pProcess)
        (void) pProcess->v->AddRef(pProcess);

    if (pStackTop) {
        Flags = CREATE_SUSPENDED;
        th->Suspended = TRUE;
    }

    h = CreateThread(NULL, 0 /*StackSize*/, NT_ThdProc, th, Flags, NULL);
    assert(h != INVALID_HANDLE_VALUE);
    if (h == NULL) {
        DWORD err = GetLastError();
        assert(h != NULL);
        return STATUS_WIN32_ERROR(err);
    }

    if (pStackTop) {
        BOOL ok;
        CONTEXT cxt;
        cxt.ContextFlags = CONTEXT_CONTROL;
        ok = GetThreadContext(h, &cxt);
        assert(ok);
        *pStackTop = (PTR) cxt.Esp;
        /* Clear top spot so as to get stack traces terminated properly */
        *(UINT *)cxt.Esp = 0;
    }

    th->Handle = h;

    if (ppNewThread)
        *ppNewThread = (PITHREAD) th;

    return S_OK;

 ErrorExit:
    DBG(("Failed CreateThread(%x)\n", pStart));
    return E_FAIL;
}


extern PIHEAP pTheHeap;
extern PINAMESPACE pTheNameSpace;

PITHREAD RTLCALLTYPE CurrentThread(void)
{
    /* does not take a ref */
    return (PITHREAD)TlsGetValue(ThreadTlsId);
}

PIPROCESS RTLCALLTYPE CurrentProcess(void)
{
    /* does not take a ref */
    return ((PTHDINFO)CurrentThread())->StartedIn;
}

PIVMVIEW RTLCALLTYPE CurrentVmView(void)
{
    /* does not take a ref */
    return NULL;
}

PIHEAP RTLCALLTYPE CurrentHeap(void)
{
#if 0
    /* does not take a ref */
    pPR(((PTHDINFO)CurrentThread())->StartedIn)->pHeap;
#else
    return pTheHeap;
#endif
}

PINAMESPACE RTLCALLTYPE CurrentNameSpace(void)
{
    PTHDINFO th = (PTHDINFO)CurrentThread();
    if (th)
        return pPR(th->StartedIn)->pNameSpace;
    else
        return pTheNameSpace;
}


/* Mutex is a concrete class (doesn't have a vtbl)
 */
typedef struct _MX { /* Internal definition */
    HANDLE event;   /* Owner, or NULL if free. LSB set if there are waiters */
} MX, *PMX;

#define BAD_MUTEX 0xFFFFBAD0

#define pMX(pMutex) ((PMX)(pMutex))

PRIVATE void InitMx(PMX mx)
{
    HANDLE event;
    if (mx->event)
        return;

    event = CreateEvent(NULL, FALSE, TRUE, NULL);
    if (AtomicCmpAndSwap((PUINT) &mx->event, 0, (UINT) event))
        return;
    /* Already inited */
    CloseHandle(event);
}

void Mutex_Init(PMUTEX pMutex)
{
    PMX mx = pMX(pMutex);
    mx->event = NULL;
}

void Mutex_Destroy(PMUTEX pMutex)
{
    PMX mx = pMX(pMutex);

    if (mx->event && mx->event != (PTR) BAD_MUTEX)
        CloseHandle(mx->event);
    mx->event = (PTR) BAD_MUTEX;
}

BOOL Mutex_TryLock(PMUTEX pMutex)
{
    PMX mx = pMX(pMutex);
    InitMx(mx);
    if (WaitForSingleObject(mx->event, 0) == WAIT_TIMEOUT)
        return FALSE;
    return TRUE;
}

void Mutex_Lock(PMUTEX pMutex)
{
    PMX mx = pMX(pMutex);
    InitMx(mx);
    WaitForSingleObject(mx->event, INFINITE);
}

void Mutex_Unlock(PMUTEX pMutex)
{
    PMX mx = pMX(pMutex);
    SetEvent(mx->event);
}

/* Return TRUE if the current thread owns the mutex.
 */
BOOL Mutex_Locked(const MUTEX *pMutex)
{
    TRACE("Mutex_Locked NYI XXX"); /* do we really need this? XXX */
    return FALSE;
}


typedef struct _COND { /* Internal definition */
    HANDLE event;
    UINT WaitCount;
} COND, *PCOND;

#define BAD_COND (0x8FFBAD01)

/* Use volatile here to ensure that the compiler does NOT optimize/move
 * any loads/stores. Conditions are used/modified from ISRs etc,
 * we don't want any values cached in registers.
 */
#define pCD(pCond) ((COND *)(pCond))

/* ==== Interface functions
 */

PRIVATE void InitCd(PCOND cd)
{
    HANDLE event;
    if (cd->event)
        return;

    event = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (AtomicCmpAndSwap((PUINT) &cd->event, 0, (UINT) event))
        return;
    /* Already inited */
    CloseHandle(event);
}

void Condition_Signal(PCONDITION pCondition)
{
    PCOND cd = pCD(pCondition);
    InitCd(cd);
    SetEvent(cd->event);
}


void Condition_Broadcast(PCONDITION pCondition)
{
    PCOND cd = pCD(pCondition);
    UINT WaitCount;
    InitCd(cd);

    for (WaitCount = cd->WaitCount; WaitCount; WaitCount--)
        SetEvent(cd->event);
}

SCODE Condition_Wait(PCONDITION pCondition, PMUTEX pMutex, PTIME pTimeout)
{
    PCOND cd = pCD(pCondition);
    PMX mx = pMX(pMutex);
    DWORD millis;
    DWORD rv;
    InitCd(cd);

    if (pTimeout && *pTimeout != TIME_FOREVER)
        if (*pTimeout < 0)      /* relative time */
            millis = (DWORD) (-*pTimeout / 10000);
        else {
            TIME t = CurrentTime();
            if (*pTimeout <= t) /* abs time in the past */
                millis = 0;
            else                /* in future, subtract current time */
                millis = (DWORD) ((*pTimeout - t) / 10000);
        }
    else
        millis = INFINITE;

    AtomicInc(&cd->WaitCount);
    if (mx && millis) {
        rv = SignalObjectAndWait(mx->event, cd->event, millis, FALSE);
    } else
        rv = WaitForSingleObject(cd->event, millis);
    AtomicDec(&cd->WaitCount);
    if (mx && millis)
        WaitForSingleObject(mx->event, INFINITE);
    if (rv == WAIT_TIMEOUT)
        return E_TIMED_OUT;
    else if (rv == WAIT_FAILED) {
        rv = GetLastError();
        DebugBreak();
        return E_TIMED_OUT;
    } else
        return S_OK;
}

SCODE Condition_WaitAndBeginConstraint(PCONDITION pCondition,
                           PMUTEX pMutex, PTIME pTimeout, BOOL EndPrevious,
                           PTIME_CONSTRAINT pTimeConstraint, PTIME pTimeTaken)
{
    SCODE sc;
    UnusedParameter(EndPrevious);
    UnusedParameter(pTimeTaken);
    sc = Condition_Wait(pCondition, pMutex, pTimeout);
    if (SUCCEEDED(sc) && pTimeConstraint)
        SleepUntil(pTimeConstraint->Start);
    return sc;
}

SCODE 
RTLCALLTYPE 
BeginConstraint(
        BOOL EndPrevious,
        PTIME_CONSTRAINT TimeConstraint,
        PTIME pTimeTaken
        )
{
    SleepUntil(TimeConstraint->Start);
    return S_OK;
}

SCODE 
RTLCALLTYPE 
EndConstraint(
        PTIME pTimeTaken
        )
{
    /* Not implemented */
    return S_OK;
}

void Condition_Init(PCONDITION pCondition)
{
    PCOND cd = pCD(pCondition);

    cd->event = NULL;
    cd->WaitCount = 0;
}

void Condition_Destroy(PCONDITION pCondition)
{
    PCOND cd = pCD(pCondition);

    if (cd->event)
        CloseHandle(cd->event);

    cd->event = (HANDLE) BAD_COND;
}

/* Signal a condition, but from special context (a device ISR).
 * Only assumption that can be made here is *all* interrupts are disabled.
 * [Relaxing this assumption is not that hard though]
 * Returns TRUE if reschedule should happen.  Caller is supposed to return
 * that value to its caller; ultimately get back to the interrupt dispatcher.
 */
extern void LogIt(UINT Key, UINT Info);

BOOL Condition_InterruptSignal(PCONDITION pCondition)
{
    PCOND cd = pCD(pCondition);
    InitCd(cd);
    SetEvent(cd->event);

    return TRUE;
}


#include <../src/base/fred.h>   /* XXX */
#include <base/cons.h>
#include <heaps/dbg_heap.h>

extern PUBLIC SCODE MCT BaseInitializeClassRegistry(void);
#if __SUPPORTS_COB_NAMESPACE
extern BOOL CobInit(void);
#endif

#define _USE_DEBUG_HEAP

/* xxx globals xxx */
PIHEAP pTheHeap = NULL;
PINAMESPACE pTheNameSpace = NULL;
#ifdef _USE_DEBUG_HEAP
PIHEAP pDebugHeap = NULL;
#endif

/*
 * Base initialization.  This is the first MI function ever invoked.
 */
void BaseInit(void)
{
    PIPROCESS TheProcess;
    static int BootCount = 0;

    BootCount++;
    DBG(("BaseInit: %sEntered\n",(BootCount>1) ? "Re" : ""));

    /* Initialize the memory allocator.
     */
    DBG(("BaseInit: Initializing heap\n"));
#if 0
    sc = CobPreLookup(_T("ff_heap.cob"), (PIUNKNOWN *) &HeapFactory);
    if (FAILED(sc)) {
        DBG(("CobPreLookup failed"));
        BaseDelete();
    }
#endif

    pTheHeap = MachineHeapCreate();

#ifdef _USE_DEBUG_HEAP
    pTheHeap = pDebugHeap = CreateDebugHeap(_T("system"), 0, 0, 8);
#if 0
 {
    PIHEAPFACTORY HeapFactory;
    SCODE sc;
    sc = CobPreLookup(_T("dbg_heap.cob"), (PIUNKNOWN *) &HeapFactory);
    if (SUCCEEDED(sc)) {
        sc = HeapFactory->v->CreateNested(HeapFactory, pTheHeap,
                                          0, 0, 8, &pDebugHeap);
        if (SUCCEEDED(sc))            
            pTheHeap = pDebugHeap;
    }
 }
#endif
#endif

    /* Initialize the name space.
     */
    DBG(("BaseInit: Initializing Namespace\n"));
    pTheNameSpace = NameSpaceNew(NULL);

#if __SUPPORTS_COB_NAMESPACE
    /* Initialize class registry
     */
    BaseInitializeClassRegistry();

    /* Initialize the COB demand-loading namespace
     */
    CobInit();
#endif

    /* Initialize the loader.
     * Also makes CurrentHeap() work for the thread.
     */
    DBG(("BaseInit: Initializing Loader\n"));
    TheProcess = InitializeIModule(pTheHeap, pTheNameSpace);

    ThreadTlsId = TlsAlloc();

    /* Create the first thread. Must succeed.
     * This is simpler than reusing the current thread.
     */
    DBG(("BaseInit: Creating First Thread\n"));
    (void) TheProcess->v->CreateThread(
                         TheProcess,
                         0,
                         FirstThread,
                         NULL,
                         0,/*default stacksize*/
                         NULL,
                         NULL);
}

void BaseDeleteClassRegistry (void);
void CobMgrExit(void);

/*
 * Base finalization
 */
void BaseDelete(void)
{
    BaseDeleteClassRegistry();
    CobMgrExit();

    pTheNameSpace->v->Release(pTheNameSpace);
    ThreadRelease(CurrentThread());

#ifdef _USE_DEBUG_HEAP
    /* Catch some leaks if any */
    pDebugHeap->v->Release(pDebugHeap);
#endif

    DBG(("I have nothing left to do.  Thanks for the ride.\n"));
    TerminateThread(GetCurrentThread(), 0);
}

void ThreadExit(void)
{
    PITHREAD pThis = CurrentThread();
    PIPROCESS pPrc;

    DBG(("ThreadExit %x\n", (UINT) pThis));
    CHECK_HEAP();
    /* Release process here so the logic stays the same as native
     * and ThreadRelease can distinguish self case and not leak handle
     */
    pPrc = pTH(pThis)->StartedIn;
    pPrc->v->Release(pPrc);
    pTH(pThis)->StartedIn = NULL; /* We no longer have a ref... */

    ThreadRelease(pThis);
    CHECK_HEAP();

    TerminateThread(GetCurrentThread(), 0);
}

/*
 * Implement BaseGetLastError() and BaseSetLastError() for MMLite programs.
 */

SCODE RTLCALLTYPE BaseGetLastError(void)
{
    PITHREAD pTh = CurrentThread();
    return pTH(pTh)->LastError;
}

void RTLCALLTYPE BaseSetLastError(SCODE Error)
{
    PITHREAD pTh = CurrentThread();
    pTH(pTh)->LastError = Error;
}

/* Say we are a simulator
 */
UINT CurrentProcessorOEM(void)
{
    return ~0;
}

LONG NT_NtSetTimerResolution(
    IN ULONG DesiredTime,
    IN BOOLEAN SetResolution,
    OUT PULONG ActualTime)
{
    HANDLE h;
    static LONG (_stdcall *pFunc)(
        IN ULONG DesiredTime,
        IN BOOLEAN SetResolution,
        OUT PULONG ActualTime) = NULL;

    if (pFunc == NULL) {
        h = GetModuleHandle(_T("ntdll.dll"));
        assert(h != INVALID_HANDLE_VALUE);
        pFunc = (LONG (_stdcall *)(
            IN ULONG DesiredTime,
            IN BOOLEAN SetResolution,
            OUT PULONG ActualTime)) GetProcAddress(h, "NtSetTimerResolution");
        assert(pFunc != NULL);
    }
    return (*pFunc)(DesiredTime, SetResolution, ActualTime);
}

/* Internal NT Stuff */
typedef struct _KSYSTEM_TIME {
    UINT32 LowPart;
    INT32 High1Time;
    INT32 High2Time;
} KSYSTEM_TIME, *PKSYSTEM_TIME;

typedef struct _KUSER_SHARED_DATA {

    /* Current low 32-bit of tick count and tick count multiplier.
     * N.B. The tick count is updated each time the clock ticks.
     */
    volatile ULONG TickCountLow;
    UINT32 TickCountMultiplier;

    /* Current 64-bit interrupt time in 100ns units. */
    volatile KSYSTEM_TIME InterruptTime;

    /* Current 64-bit system time in 100ns units. */
    volatile KSYSTEM_TIME SystemTime;

    /* Current 64-bit time zone bias. */
    volatile KSYSTEM_TIME TimeZoneBias;
} KUSER_SHARED_DATA, *PKUSER_SHARED_DATA;

#define MM_SHARED_USER_DATA_VA      0x7FFE0000
#define USER_SHARED_DATA ((KUSER_SHARED_DATA * const)MM_SHARED_USER_DATA_VA)

__int64 SysToIntTimeOffset = UINT64_ZERO;

/* Initializes the offset between SystemTime and InterruptTime */
void InitTime(void)
{
    union {
        KSYSTEM_TIME SysTime;
        __int64 CurrTime;
    } it, st;

    /* Wait for the system time to change. This ensures that I read
     * SystemTime and InterruptTime when they're most in sync, just after 
     * they've both been updated.
     * (Based on knowledge that SystemTime is updated _after_ InterruptTime)
     */
    st.SysTime.LowPart   = USER_SHARED_DATA->SystemTime.LowPart;
    while (st.SysTime.LowPart == USER_SHARED_DATA->SystemTime.LowPart);

    /* Read system time */
    do {
        st.SysTime.High1Time = USER_SHARED_DATA->SystemTime.High1Time;
        st.SysTime.LowPart   = USER_SHARED_DATA->SystemTime.LowPart;
    } while (st.SysTime.High1Time != USER_SHARED_DATA->SystemTime.High2Time);

    /* Read interrupt time */
    do {
        it.SysTime.High1Time = USER_SHARED_DATA->InterruptTime.High1Time;
        it.SysTime.LowPart   = USER_SHARED_DATA->InterruptTime.LowPart;
    } while (it.SysTime.High1Time != USER_SHARED_DATA->InterruptTime.High2Time);

    /* Save offset between the two */
    SysToIntTimeOffset = st.CurrTime - it.CurrTime;
#ifdef _DEBUG
    printf("SysToIntTimeOffset=0x%016llx\n",SysToIntTimeOffset);
#endif

    {
        /* ask for maximum clock granularity */
        PRIVATE ULONG TimeDeltaIs;
        NT_NtSetTimerResolution( 1, TRUE, &TimeDeltaIs);
        DBG(("Timer resolution is %d00 ns (%d Hz)\n",
             TimeDeltaIs, 10000000 / TimeDeltaIs));
    }
}

/* Return the current time in 100ns units since Jan 1, 1601
 * This code reads the interrupt time and then converts to absolute time
 * so it's accurate to within one timer interrupt
 */
__int64 CurrentTime(void)
{
    union {
        KSYSTEM_TIME SysTime;
        __int64 CurrTime;
    } ts;

    do {
        ts.SysTime.High1Time = USER_SHARED_DATA->InterruptTime.High1Time;
        ts.SysTime.LowPart   = USER_SHARED_DATA->InterruptTime.LowPart;
    } while (ts.SysTime.High1Time != USER_SHARED_DATA->InterruptTime.High2Time);

    return ts.CurrTime + SysToIntTimeOffset;
}

/* Returns the speed of the current processor.
 * BUGBUG Fix Brad' MHz code to keep more bits of the freq.
 */
extern int MHz(int Iterations);

extern UINT64 CurrentProcessorSpeed(void)
{
#if 1
    return 1000000000;          /* XXX */
#else
    static UINT64 Speed = 0;

    if (Speed == 0) {
        int m = MHz(3);
        if (m > 0)
            Speed = 1000000 * m;
    }
    return Speed;
#endif
}

extern void __stdcall DebugSynNotifyLoad(char *pbData, const char *pszDllPath);
extern void __stdcall DebugSynNotifyUnload(char *pbData);

/* Notify debugger of load so that it can find the symbols
 */
UINT DebuggerNotifyLoad(const _TCHAR *ModuleName,
                        ADDRESS LoadStart,
                        ADDRESS LoadEnd,
                        PTR Unused)
{
#ifdef _UNICODE
    char *_name = _wcstos(ModuleName);
#else
#define _name ModuleName
#endif

    DebugSynNotifyLoad((PTR) LoadStart, _name);

#ifdef _UNICODE
    free(_name);
#else
#undef _name
#endif

    return 0;
}

void DebuggerNotifyUnload(const _TCHAR *ModuleName, ADDRESS LoadStart,
                          UINT Token)
{
    UnusedParameter(Token);
    UnusedParameter(ModuleName);

    DebugSynNotifyUnload((PTR) LoadStart);
}
