/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <import_windows.h>
#define sockaddr WSAsockaddr
#define WIN32
#include <winsock2.h>
#undef sockaddr

#include <stdio.h>
#include <stdlib.h>
#include <mmlite.h>
#include <netif.h>
#include <mmmacros.h>
#include <base/cobdesc.h>

/* struct soackaddr conversions. WSA uses old 4.3 BSD format */
void sa2wsa(struct WSAsockaddr *wa, const struct sockaddr *sa)
{
    memcpy(wa, sa, sizeof(*wa));
    wa->sa_family = sa->sa_family;
}

void wsa2sa(struct sockaddr *sa, const struct WSAsockaddr *wa, UINT Size)
{
    memcpy(sa, wa, sizeof(*sa));

    assert(wa->sa_family < 256);
    sa->sa_family = (UINT8) wa->sa_family;

    assert(Size < 256);
    sa->sa_len = (UINT8) Size;
}

/*
 * The IEndpoint object
 */

typedef struct socket {
    struct CUnknown super;

    SOCKET so;
    PCONDITION pCond;
} CEndpoint, *PISOCKET;

#include <s_CEndpoint_v.c>

/* Destructor */
PRIVATE SCODE CEndpointDestroy(PIUNKNOWN iThis)
{
    struct socket *so = (struct socket *)iThis;

    /*
     * Close down (just in case we aren't already).
     */
    (void) closesocket(so->so);

    /*
     * Free ourselves.
     */
    free(so);

    return S_OK;
}

#if 0
/*
 * Add a reference.
 */
PRIVATE UINT MCT CEndpointAddRef(PIENDPOINT pThis)
{
    INT Refs;
    struct socket *so = (struct socket *)pThis;

    Refs = AtomicInc(&((PISOCKET)pThis)->RefCnt);
    return Refs;
}

/*
 * Release a reference.
 * When last ref is gone, destructor.
 */
PRIVATE UINT MCT CEndpointRelease(PIENDPOINT pThis)
{
    INT Refs;
    struct socket *so = (struct socket *)pThis;

    Refs = AtomicDec(&((PISOCKET)pThis)->RefCnt);

    if (Refs == 0) {
        /*
         * Close down (just in case we aren't already).
         */
        (void) closesocket(so->so);

        /*
         * Free ourselves.
         */
        free(so);
    }

    return Refs;
}
        
PRIVATE SCODE MCT CEndpointQueryInterface(PIENDPOINT pThis, REFIID pIid, void* *ppUnk)
{
    if (UuidCmp(&IID_IEndpoint, pIid)
        || UuidCmp(&IID_IFile, pIid)
        || UuidCmp(&IID_IUnknown, pIid)) {
        CEndpointAddRef(pThis);
        *ppUnk = (PIUNKNOWN) pThis;
        return NO_ERROR;
    }
    *ppUnk = NULL;
    return E_NO_INTERFACE;
}
#endif

PRIVATE SCODE MCT CEndpointSetSize(IEndpoint *This, UINT64 Size)
{
    UnusedParameter(This);
    UnusedParameter(Size);
    return E_NOT_IMPLEMENTED;
}
        
PRIVATE SCODE MCT CEndpointGetSize(IEndpoint *This, UINT64 *pSize)
{
    UnusedParameter(This);
    UnusedParameter(pSize);
    return E_NOT_IMPLEMENTED;
}

PIENDPOINT make_socket(SOCKET s)
{
    PIHEAP Heap = CurrentHeap();

    struct socket *so = Heap->v->Alloc(Heap,
                                  HEAP_ZERO_MEMORY | HEAP_DEBUG_NAME("socket"),
                                       sizeof(*so), 0);
    if (so) {
        so->super.v = (struct IUnknownVtbl *) &CEndpointVtbl.v;
        so->super.RefCnt = 1;
        so->so = s;
    }

    return (PIENDPOINT) so;
}

PRIVATE SCODE MCT CEndpointAccept(PIENDPOINT pThis, PIENDPOINT *ppThat,
                    struct sockaddr *pAddr, UINT32 AddrSize, UINT32 *pAddrSize)
{
    struct socket *so = (struct socket *)pThis;
    SOCKET s2;
    struct WSAsockaddr wsa;
    int salen = sizeof(wsa);

    s2 = WSAAccept(so->so, &wsa, &salen, NULL, 0);
    if (s2 == INVALID_SOCKET)
        return STATUS_WIN32_ERROR(WSAGetLastError());

    *ppThat = make_socket(s2);

    if (pAddr)
        wsa2sa(pAddr, &wsa, salen);

    if (pAddrSize)
        *pAddrSize = salen;

    return S_OK;
}

SCODE MCT CEndpointBind(PIENDPOINT pThis, const struct sockaddr *Name, UINT32 NameLen)
{
    struct socket *so = (struct socket *)pThis;
    int rv;
    struct WSAsockaddr wsa;

    sa2wsa(&wsa, Name);
    rv = bind(so->so, &wsa, NameLen);
    if (rv)
        return STATUS_WIN32_ERROR(WSAGetLastError());
    else
        return S_OK;
}

PRIVATE SCODE MCT CEndpointClose(PIENDPOINT pThis)
{
    struct socket *so = (struct socket *)pThis;
    INT rv;

    rv = closesocket(so->so);

    if(rv)
        return STATUS_WIN32_ERROR(rv);

    return S_OK;
}

PRIVATE SCODE MCT CEndpointConnect(PIENDPOINT pThis, const struct sockaddr *Name, UINT32 NameLen)
{
    struct socket *so = (struct socket *)pThis;
    struct WSAsockaddr wsa;
    int rv;

    sa2wsa(&wsa, Name);
    rv = WSAConnect(so->so, &wsa, NameLen, NULL, NULL, NULL, NULL);

    if (rv)
        return STATUS_WIN32_ERROR(WSAGetLastError());
    else
        return S_OK;
}

PRIVATE SCODE MCT CEndpointGetName(PIENDPOINT pThis, struct sockaddr *Name,
                          UINT32 NameSize, UINT *pNameLen, 
                          ENDPOINT_NAME_TYPE WhichName)
{
    struct socket *so = (struct socket *)pThis;
    int rv;
    struct WSAsockaddr wsa;

    switch(WhichName) {

    case ENDPOINT_ADDR_MINE:
        rv = getsockname(so->so, &wsa, &NameSize);
        break;

    case ENDPOINT_ADDR_PEER:
        rv = getpeername(so->so, &wsa, &NameSize);
        break;

    default:
        return E_INVALID_PARAMETER;
    }

    if (rv)
        return STATUS_WIN32_ERROR(WSAGetLastError());

    wsa2sa(Name, &wsa, NameSize);

    if (pNameLen)
        *pNameLen = NameSize;

    return S_OK;
}

PRIVATE SCODE MCT CEndpointGetSockOpt(PIENDPOINT pThis, INT32 Level,
                    UINT32 OptName, BYTE *OptVal, UINT32 OptValSize, UINT *pOptLen)
{
    struct socket *so = (struct socket *)pThis;
    int rv;

    rv = getsockopt(so->so, Level, OptName, OptVal, &OptValSize);

    if (rv)
        return STATUS_WIN32_ERROR(WSAGetLastError());

    if (pOptLen)
        *pOptLen = OptValSize;
    return S_OK;
}

PRIVATE SCODE MCT CEndpointIoctl(PIENDPOINT pThis, UINT32 cmd, UINT8 *data)
{
    struct socket *so = (struct socket *)pThis;
    int rv;

    rv = ioctlsocket(so->so, cmd, (PTR) data);

    if (rv)
        return STATUS_WIN32_ERROR(WSAGetLastError());
    else
        return S_OK;
}

PRIVATE SCODE MCT CEndpointListen(PIENDPOINT pThis, INT32 BackLog)
{
    struct socket *so = (struct socket *)pThis;
    int rv;

    rv = listen(so->so, BackLog);

    if (rv)
        return STATUS_WIN32_ERROR(WSAGetLastError());
    else
        return S_OK;
}

/*
 * Receive from and return the origin
 */
PRIVATE SCODE MCT CEndpointRecvFrom(PIENDPOINT pThis,
                   BYTE *pBuffer, UINT32 Size,
                   UINT32 *pRetSize, UINT32 Flags, struct sockaddr *pFrom,
                   UINT32 FromSize, UINT32 *pFromSize)
{

    struct socket *so = (struct socket *)pThis;
    int rv;
    WSABUF wb;
    struct WSAsockaddr wsa;

    wb.len = Size;
    wb.buf = pBuffer;

    rv = WSARecvFrom(so->so, &wb, 1, pRetSize, &Flags,
                     pFrom ? &wsa : NULL,
                     pFrom ? &FromSize : NULL,
                     NULL, NULL);
    if (rv)
        return STATUS_WIN32_ERROR(rv);

    if (pFrom)
        wsa2sa(pFrom, &wsa, FromSize);

    if (pFromSize)
        *pFromSize = FromSize;

    return S_OK;
}

/*
 * Receive data from peer.
 */
PRIVATE SCODE MCT CEndpointReadAt(PIENDPOINT pThis, UINT64 Flags, BYTE *Buffer,
                          UINT Len, UINT *pRetSize)
{
    return CEndpointRecvFrom(pThis, Buffer, Len, pRetSize, (UINT32) Flags,
                             NULL, 0, NULL);
}

/*
 * Send to a given destination
 */
PRIVATE SCODE MCT CEndpointSendTo(PIENDPOINT pThis,
                 const BYTE *pBuffer, UINT32 Size,
                 UINT32 *pRetSize, UINT32 Flags, const struct sockaddr *pTo,
                 UINT32 ToSize)
{
    struct socket *so = (struct socket *)pThis;
    int rv;
    WSABUF wb;
    struct WSAsockaddr wsa;

    wb.len = Size;
    wb.buf = (void *) pBuffer;

    if (pTo)
        sa2wsa(&wsa, pTo);

    rv = WSASendTo(so->so, &wb, 1, pRetSize, Flags,
                   pTo ? &wsa : NULL, ToSize,
                   NULL, NULL);
    if (rv == 0)
        return S_OK;
    else
        return STATUS_WIN32_ERROR(rv);
}

/*
 * Send data to peer.
 */
PRIVATE SCODE MCT CEndpointWriteAt(PIENDPOINT pThis, UINT64 Flags,
                                  const BYTE *Buffer, UINT Len, UINT *pRetSize)
{
    return CEndpointSendTo(pThis, Buffer, Len, pRetSize, (UINT32) Flags,
                           NULL, 0);
}

PRIVATE SCODE MCT CEndpointSetSockOpt(PIENDPOINT pThis,
             UINT32 Level, UINT32 OptName, const BYTE *OptVal, UINT32 OptLen)
{
    struct socket *so = (struct socket *)pThis;
    int rv;

    if (Level == SOL_SOCKET && OptName == SO_RCVTIMEO) {
        /* Since SO_RCVTIMEO is documented as unimplemented, try this */
        COMMTIMEOUTS ct;
#if 0
        int tmp;
        rv = ioctlsocket(so->so, FIONBIO, &tmp);
        if(rv == SOCKET_ERROR)
            return STATUS_WIN32_ERROR(WSAGetLastError());
#endif

        rv = GetCommTimeouts((HANDLE) so->so, &ct);
        if (rv)
            return STATUS_WIN32_ERROR(rv);

        ct.ReadIntervalTimeout = *(UINT32*)OptVal;
        ct.ReadTotalTimeoutMultiplier = *(UINT32*)OptVal;
        ct.ReadTotalTimeoutConstant = *(UINT32*)OptVal;

        rv = SetCommTimeouts((HANDLE) so->so, &ct);
    } else {
        rv = setsockopt(so->so, Level, OptName, OptVal, OptLen);
    }

    if (rv == 0)
        return S_OK;
    else
        return STATUS_WIN32_ERROR(rv);
}

/*
 * Disable sends and/or receives on an Endpoint.
 *
 * Valid values for "How":
 *    0 - subsequent receives on the Endpoint will be disallowed.
 *    1 - subsequent sends are disallowed.
 *    2 - both 0&1
 */
PRIVATE SCODE MCT CEndpointShutdown(PIENDPOINT pThis, UINT32 How)
{
    struct socket *so = (struct socket *) pThis;
    int rv;

    rv = shutdown(so->so, How);

    if (rv == 0)
        return S_OK;
    else
        return STATUS_WIN32_ERROR(rv);
}

#include <net/mbuf.h>

SCODE CEndpointAllocateBuffer (
    /*in*/ PIENDPOINT pThis,
    /*in*/ UINT32 DataSize,
    /*out*/ PIBUFFER * NewBuffer)
{
    struct socket *so = (struct socket *) pThis;
    struct mbuf *m;

    UnusedParameter(so);

    m = m_gethdr(M_WAIT, MT_DATA);
    if (m == NULL)
        return E_NOT_ENOUGH_MEMORY;
    m->m_pkthdr.len = 0;
    m->m_pkthdr.rcvif = (struct ifnet *)0;    

    if (DataSize >= MINCLSIZE) {
        MCLGET(m, M_WAIT);
        if (!(m->m_flags & M_EXT)) {
            m_free(m);
            return E_NOT_ENOUGH_MEMORY;
        }
    }

    /* XXX if datagram to be sent, reserve space for headers */

    *NewBuffer = (PIBUFFER) m;
    return S_OK;
}

PRIVATE SCODE CEndpointSendBuffer (
    /*in*/ PIENDPOINT pThis,
    /*in*/ PIBUFFER iBuffer,
    /*in*/ UINT32 Flags,
    /*in*/ const struct sockaddr * pTo,
    /*in*/ UINT32 ToSize)
{
    struct socket *so = (struct socket *) pThis;
    int rv, i;
    struct mbuf *m = (struct mbuf *) iBuffer;
#if 0
    assert(!pTo);

    for (i = 0; m && i < 20; i++) {
        rv = WriteFile((HANDLE) so->so, m->m_hdr.mh_data, m->m_hdr.mh_len, &i, NULL);
        assert(rv == TRUE);
        assert(i == m->m_hdr.mh_len);
        if (rv != TRUE)
            return STATUS_WIN32_ERROR(WSAGetLastError());
        m = m->m_hdr.mh_next;
    }
    return S_OK;
#else
#define MAXWB 50
    WSABUF wb[MAXWB];
    struct WSAsockaddr wsa;
    DWORD Transferred;

    if (pTo)
        sa2wsa(&wsa, pTo);

    for (i = 0; m && i < MAXWB; i++) {
        wb[i].len = m->m_hdr.mh_len;
        wb[i].buf = m->m_hdr.mh_data;

        m = m->m_hdr.mh_next;
    }

    if (m)
        return E_OUT_OF_STRUCTURES;

    rv = WSASendTo(so->so, wb, i, &Transferred, Flags,
                   pTo ? &wsa : NULL, ToSize,
                   NULL, NULL);
    if (rv == 0)
        return S_OK;
    else
        return STATUS_WIN32_ERROR(WSAGetLastError());
#endif
}

#define MAX_SIZE 0x10000

//#define TEST_NONBLOCK 1
//#define MAX_SIZE 1

PRIVATE SCODE CEndpointRecvBuffer (
    PIENDPOINT iThis,
    /*out*/ PIBUFFER* piBuffer,
    UINT32 Size,
    UINT32 Flags,
    /*out*/ struct sockaddr * pFrom,
    UINT32 FromSize,
    /*out*/ UINT *pFromLength)
{
    struct socket *so = (struct socket *) iThis;
    struct mbuf *m;
    PIBUFFER im;
    WSABUF wb[MAXWB];
    int rv, i;
    SCODE sc;
    UINT64 Size64;
    struct WSAsockaddr wsa;
#ifdef TEST_NONBLOCK
    static int block = 10;
#endif

    if (Size > MAX_SIZE)
      Size = MAX_SIZE;
    Size64 = Size;

    /* Since we don't know how much data there is, we first allocate a big
     * buffer and later trim it to size.
     */
   
    sc = CEndpointAllocateBuffer(NULL, Size, &im);
    assert(SUCCEEDED(sc));

    im->v->SetSize(im, Size64);

    m = (struct mbuf *) im;
    for (i = 0; m && i < MAXWB; i++) {
        wb[i].len = m->m_hdr.mh_len;
        wb[i].buf = m->m_hdr.mh_data;

        m = m->m_hdr.mh_next;
    }
    assert(!m);

#ifdef TEST_NONBLOCK
    block--;
    if (block >= 0) {
        Size = 0;
        wb[0].len = 0;
        rv = recv(so->so, wb[0].buf, 0, 0);
        sc = STATUS_WIN32_ERROR(WSAGetLastError());

        sc = STATUS_WIN32_ERROR(WSAEWOULDBLOCK);
        goto NoData;        
    }
    block = 2;
#endif

    rv = WSARecvFrom(so->so, wb, i, &Size, &Flags,
                     pFrom ? &wsa : NULL,
                     pFrom ? &FromSize : NULL,
                     NULL, NULL);
    if (rv) {
		sc = WSAGetLastError();
        sc = STATUS_WIN32_ERROR(sc);
        goto NoData;
    }

    if (pFrom) {
        wsa2sa(pFrom, &wsa, FromSize);

        if (pFromLength)
            *pFromLength = FromSize;
    }

    if (Size == 0) {
        sc = S_FALSE;
        *piBuffer = NULL;
        goto NoData;
    }

    Size64 = Size;
    im->v->SetSize(im, Size64);

    *piBuffer = im;
    return S_OK;

 NoData:
    im->v->Release(im);
    return sc;
}

/* XXX copy from xpshim.c */
typedef struct _COND { /* Internal definition */
    HANDLE event;
    UINT WaitCount;
} COND, *PCOND;
#define pCD(pCond) ((COND *)(pCond))

PRIVATE SCODE CEndpointSetCondition (
    PIENDPOINT pThis,
    PCONDITION pCond)
{
    struct socket *This = (struct socket *) pThis;
    PCONDITION old;
    int rv;

    old = (PCONDITION) AtomicSwap((PUINT) &This->pCond, (UINT) pCond);

    /* Signal old condition as the status changed for this socket */
    if (old)
        Condition_Broadcast(old);

    /* Make sure this condition is initialized */
    Condition_Signal(pCond);

    rv = WSAEventSelect(This->so, pCD(pCond)->event,
                        FD_READ|FD_WRITE|FD_ACCEPT|FD_CLOSE);
    if (rv != 0)
        rv = WSAGetLastError();
    assert(rv == 0);

    return S_OK;
}

extern SCODE MCT CEndpointBind(PIENDPOINT, const struct sockaddr *, UINT32);
extern SCODE ProtocolStart(void);
extern void mbinit(void);

typedef struct _NETFACTORY {
    struct _DefCob super;
    BOOL Started;
} CEndpointFactory, *PNETFACTORY;

#define _MANUAL_COBMAIN 1
#define CEndpointFactoryDestroy NULL
#include <s_CEndpointFactory_v.c>

#define pNF(_i_) ((PNETFACTORY) (_i_))

#if 0
/*
 * Add a reference.
 */
PRIVATE UINT MCT CEndpointFactoryAddRef(PIENDPOINTFACTORY pThis)
{
    PNETFACTORY This = pNF(pThis);
    return AtomicInc(&This->RefCnt);
}

/*
 * Release a reference.
 * When last ref is gone, destructor.
 */
PRIVATE UINT MCT CEndpointFactoryRelease(PIENDPOINTFACTORY pThis)
{
    PNETFACTORY This = pNF(pThis);
    return AtomicDec(&This->RefCnt);
}

/*
 * Query support.
 * We inherit solely from IUnknown.
 */
PRIVATE SCODE MCT CEndpointFactoryQueryInterface(PIENDPOINTFACTORY pThis,
                                                 REFIID pIid,
                                                 void* *ppUnk)
{
    SCODE sc = GenericQueryInterface((PIUNKNOWN) pThis, pIid, ppUnk,
                                     &IID_IEndpointFactory);
    if (sc == E_NO_INTERFACE)
        sc = GenericQueryInterface((PIUNKNOWN) pThis, pIid, ppUnk,
                                   &IID_IProgram);
    return sc;
}
#endif

/*
 * Build a new Endpoint object.
 */
PRIVATE SCODE MCT EndpointObjNew(PIENDPOINTFACTORY pThis,
                                 UINT32 Family, UINT32 Type, UINT32 Protocol,
                                 PIENDPOINT *ppNew)
{
    struct socket *so;
    SOCKET s;

    s = WSASocket(Family, Type, Protocol, NULL, 0, 0);
    if (s == INVALID_SOCKET)
        return STATUS_WIN32_ERROR(WSAGetLastError());

    so = (struct socket *) make_socket(s);
    if (!so) {
        closesocket(s);
        return E_NOT_ENOUGH_MEMORY;
    }

    *ppNew = (PIENDPOINT)so;

    return S_OK;
}

SCODE MCT CEndpointFactoryCreateEndpointPair(PIENDPOINTFACTORY pThis,
                                             PIENDPOINT1 *ppNew1, PIENDPOINT1 *ppNew2);

/*
 * Wait for one or more Endpoints previously built by this factory to
 * undergo some notable event (like becoming readable, writeable, etc).
 *
 * Since we want to be able to wait on multiple endpoints simultaneously,
 * but can only wait on one condition at once, we use the same condition
 * for all waiting and re-check event occurrence on all wakeups.
 */
PRIVATE SCODE MCT WaitForEndpoints(PIENDPOINTFACTORY pThis, PIENDPOINT *Endpts,
                                   ENDPOINT_REQ_TYPES *EventTypes, UINT NumEndpts,
                                   TIME Timeout)
{
    fd_set rs;
    fd_set ws;
    fd_set xs;
    UINT i;
    int rv;
    ENDPOINT_REQ_TYPES types = 0;
    struct timeval *timo = NULL, tval = {0,0};

    if (Timeout != TIME_FOREVER) {
        timo = &tval;
        if (Timeout != TIME_ORIGIN)
            return E_NOT_IMPLEMENTED; /* not handled in this implementation */
    }
    if (NumEndpts > FD_SETSIZE)
        return E_INVALID_PARAMETER;

    FD_ZERO(&rs);
    FD_ZERO(&ws);
    FD_ZERO(&xs);
    for (i = 0; i < NumEndpts; i++) {
        ENDPOINT_REQ_TYPES req = EventTypes[i];
        struct socket *so  = (struct socket *) Endpts[i];

        if (!so)
            continue;

        if (req & ENDPOINT_READ)
            FD_SET(so->so, &rs);
        if (req & ENDPOINT_WRITE)
            FD_SET(so->so, &ws);
        if (req & ENDPOINT_EXCEPTION)
            FD_SET(so->so, &xs);
        types |= req;
    }
    rv = select(NumEndpts,
                (types & ENDPOINT_READ) ? &rs : NULL,
                (types & ENDPOINT_WRITE) ? &ws : NULL,
                (types & ENDPOINT_EXCEPTION) ? &xs : NULL,
                timo);
    if (rv == 0)
        return E_TIMED_OUT;
    if (rv == SOCKET_ERROR) {
        rv = WSAGetLastError();
        assert(FALSE);
        rv = STATUS_WIN32_ERROR(rv);
        return rv;
    }

    for (i = 0; i < NumEndpts; i++) {
        struct socket *so  = (struct socket *) Endpts[i];

        if (!so)
            continue;

        /* Init OUT bits */
        EventTypes[i] &= ~(ENDPOINT_ISREADABLE|ENDPOINT_ISWRITEABLE|ENDPOINT_ISEXCEPTIONAL);

        if (FD_ISSET(so->so, &rs))
            EventTypes[i] |= ENDPOINT_ISREADABLE;
        if (FD_ISSET(so->so, &ws))
            EventTypes[i] |= ENDPOINT_ISWRITEABLE;
        if (FD_ISSET(so->so, &xs))
            EventTypes[i] |= ENDPOINT_ISEXCEPTIONAL;
    }

    return S_OK;
}

SCODE CEndpointAllocateBuffer (
    PIENDPOINT pThis,
    UINT32 DataSize,
    PIBUFFER * NewBuffer);

PRIVATE SCODE CEndpointFactoryAllocateBuffer (
    PIENDPOINTFACTORY pThis,
    UINT32 DataSize,
    /*optional*/ PINETDRIVER recvif,
    /*out*/ PIBUFFER * NewBuffer)
{
    SCODE sc;
    PIBUFFER im;

    sc = CEndpointAllocateBuffer(NULL, DataSize, &im);
    if (FAILED(sc))
        return sc;

    *NewBuffer = im;
    return sc;
}

PRIVATE SCODE MCT 
CEndpointFactoryEtherInput(PIENDPOINTFACTORY pThis, PINETDRIVER ifp,
                           struct ether_header *eh, PIBUFFER m)
{
    DebugBreak();
    return E_NOT_IMPLEMENTED;
}


PRIVATE SCODE MCT
CEndpointFactoryEtherOutput(PIENDPOINTFACTORY pThis, PINETDRIVER ifp,
                            PIBUFFER m, struct sockaddr *dst,
                            struct route_in *rti)
{
    DebugBreak();
    return E_NOT_IMPLEMENTED;    
}


PRIVATE void MCT 
CEndpointFactoryNetLock(PIENDPOINTFACTORY pThis)
{
    DebugBreak();
}


PRIVATE void MCT 
CEndpointFactoryNetUnlock(PIENDPOINTFACTORY pThis)
{
    DebugBreak();
}

PRIVATE SCODE MCT 
CEndpointFactoryIfAttach(PIENDPOINTFACTORY pThis, PINETDRIVER ifp)
{
    DebugBreak();
    return E_NOT_IMPLEMENTED;
}

static MUTEX NetMx = MUTEX_INITIALIZER;

PRIVATE void MCT
CEndpointFactorySignalCondition(PIENDPOINTFACTORY pThis, PCONDITION pCond)
{
    /* Barrier */
    Mutex_Lock(&NetMx);
    Mutex_Unlock(&NetMx);

    Condition_Signal(pCond);
}

PRIVATE SCODE MCT
CEndpointFactoryWaitCondition(PIENDPOINTFACTORY pThis,
                              PCONDITION pCond, PMUTEX pMutex,
                              PTIME pTimeOut)
{
    SCODE sc;

    Mutex_Lock(&NetMx);
    if (pMutex)
        Mutex_Unlock(pMutex);

    sc = Condition_Wait(pCond, &NetMx, pTimeOut);
    Mutex_Unlock(&NetMx);


    if (pMutex)
        Mutex_Lock(pMutex);

    return sc;
}

PRIVATE SCODE MCT
CEndpointFactoryMain(PIENDPOINTFACTORY pThis, const _TCHAR *cmdline)
{
    SCODE sc;

    DBG(("CEndpointFactoryMain did not do anything\n"));
    sc = S_OK;

    return sc;
}

struct _NETFACTORY TheNetFactory;
#if 0
{
    &CEndpointFactoryVtbl,
    1,
    FALSE
};
#endif

/*
 * Entry point: Create an Endpoint factory and return it to cobmgr
 */

PIUNKNOWN protocolCobMain(void)
{
    WSADATA wsadata;
    if (WSAStartup(MAKEWORD(2,2), &wsadata) != 0)
        DebugBreak();

    mbinit();

    TRACE("ProtocolCobMain");
    
    return GenericCobInit((void *) &protocol_CobDesc, (void *) &TheNetFactory);
}

/*********************** Leftovers from protocol stack for u_mbuf.c *******/
/*
 * This is pretty ugly, should rip spls out of the code and do it right.
 */
struct {
    MUTEX Mutex;
    PITHREAD Owner;
    INT Nested;
}NetGlobalLock = {0,};

void RTLCALLTYPE NetLock(void)
{
    PITHREAD Who;

    Who = CurrentThread();

    if (Who != NetGlobalLock.Owner){
        Mutex_Lock(&NetGlobalLock.Mutex);
        NetGlobalLock.Owner = Who;
    }

    NetGlobalLock.Nested++;
}

void RTLCALLTYPE NetUnlock(void)
{
    assert (CurrentThread() == NetGlobalLock.Owner);

    NetGlobalLock.Nested--;
    if (NetGlobalLock.Nested == 0){
        NetGlobalLock.Owner = NULL;
        Mutex_Unlock(&NetGlobalLock.Mutex);
    }
}

/* XXX bogus values */
INT max_linkhdr = 16;            /* largest link-level header */
INT max_protohdr = 16;           /* largest protocol header */

void
m_reclaim(void)
{
}
