/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <mmlite.h>
#include <stdlib.h>
#include <tchar.h>
#include <crypto.h>

/* Utils
 */
#define CHUNKSIZE 32 /* maximum size */
static UINT32 Buffer[CHUNKSIZE/sizeof(UINT32)]; /* aligned !! */
static UINT32 Buffer2[CHUNKSIZE/sizeof(UINT32)]; /* aligned !! */

PICYPHER
GetCypher(
          _TCHAR *Algorithm,
          UINT Arg1
          )
{
    SCODE sc;
    PICYPHER Cypher;
    PICYPHERCONSTRUCTOR Factory;
    PIUNKNOWN pUnk;

    sc = BindToObject(CurrentNameSpace(), Algorithm,
                      NAME_SPACE_READ, &IID_ICypherConstructor,
                      (void *) &Factory);
    if (FAILED(sc)) {
        _tprintf(_TEXT("Could not find the %s factory sc=%x\r\n"),
                 Algorithm, sc);
        return NULL;
    }

    sc = Factory->v->Constructor(Factory,&pUnk,Algorithm,Arg1);
    (void) Factory->v->Release(Factory);

    if (FAILED(sc)) {
        _tprintf(_TEXT("Constructor failed sc=%x\r\n"), sc);
        return NULL;
    }

    sc = pUnk->v->QueryInterface(pUnk,&IID_ICypher,(void**)&Cypher);
    pUnk->v->Release(pUnk);

    if (FAILED(sc)) {
        _tprintf(_TEXT("%s is not a cypher sc=%x\r\n"), Algorithm, sc);
        return NULL;
    }

    return Cypher;
}

BOOL
CypherTest(
           PICYPHER Cypher,
           _TCHAR *FileIn,
           _TCHAR *FileOut,
           CYPHER_OPS Operation
       )
{
    BOOL ok = FALSE;
    FILE *Fin, *Fout;
    size_t n, nRecords;
    UINT nBytes, BufferSize, off, consumed;
    SCODE sc;

    Fin = Fout = NULL;

    _tprintf(_TEXT("Opening %s for reading\r\n"), FileIn);
    Fin = _tfopen(FileIn, _T("rb"));
    if (Fin == NULL) {
        _tprintf(_TEXT("Could not open %s for reading.\r\n"), FileIn);
        goto Out;
    }

    _tprintf(_TEXT("Opening %s for writing\r\n"), FileOut);
    Fout = _tfopen(FileOut, _T("wb"));
    if (Fout == NULL) {
        _tprintf(_TEXT("Could not open %s for writing.\r\n"), FileOut);
        goto Out;
    }

    /* See what buffer size we should use
     * NB: Do check that the output buffer wont overflow.
     */
    BufferSize = nBytes = CHUNKSIZE;
    sc = Cypher->v->GetBlockLengths(Cypher,
                                    &Operation,
                                    &BufferSize,
                                    &nBytes);
    if (FAILED(sc)) {
        _tprintf(_TEXT("CypherGetBlockLengths failure (sc=%x nBytes=%d)\r\n"), 
                 sc, nBytes);
        goto Out;
    }
    if ((Operation & CYPHER_SIZE_FIXED) == 0) {
        BufferSize = nBytes = CHUNKSIZE;
    }
    

    /* ok, do it.
     */
    off = 0;
    for (nRecords = 1; ; nRecords++) {
        n = fread((BYTE*)Buffer + off,1,BufferSize-off,Fin);
        if (n == 0 && off == 0) {
            break;
        } else {
            _tprintf(_TEXT("R%d "), n);
        }
        if (n == 0)
            Operation |= CYPHER_LAST;
            
        if (n < BufferSize-off && (Operation & CYPHER_SIZE_FIXED)) {
            memset((BYTE*) Buffer + off + n, 0, BufferSize-off-n);
            n = BufferSize-off;
            Operation |= CYPHER_LAST;
        }

        nBytes = 0;
        sc = Cypher->v->Cypher(Cypher,
                               Operation | ((n == 0) ? CYPHER_LAST : 0),
                               (BYTE*) Buffer,
                               n+off,
                               &consumed,
                               (BYTE*) Buffer2,
                               sizeof Buffer2,
                               &nBytes);
        if (FAILED(sc) || (nBytes == 0)) {
            _tprintf(_TEXT("Cypher failure (sc=%x nBytes=%d).\r\n"), 
                     sc, nBytes);
            goto Out;
        }
        off = n+off-consumed;
        if (off)
            memcpy(Buffer, (BYTE*)Buffer + consumed, off);
        //memset((BYTE*)Buffer + off, 'z', BufferSize-off);

        n = fwrite(Buffer2,nBytes,1,Fout);
        if (n != 1) {
            _tprintf(_TEXT("Failed to write %d bytes, record %d (to %s).\r\n"), 
                     nBytes, nRecords, FileOut);
            goto Out;
        } else {
            _tprintf(_TEXT("W%d "), nBytes);
        }
    }

    ok = TRUE;

 Out:
    if (Fout)
        fclose(Fout);
    if (Fin)
        fclose(Fin);
    return ok;
}

BOOL
Test(
     _TCHAR *Algorithm,
     UINT Arg1,
     _TCHAR *FileIn,
     _TCHAR *FileOut,
     _TCHAR *File3,
     _TCHAR *Key
     )
{
    PICYPHER Cypher;
    BOOL ok = FALSE;
    UINT KeyLen;
    SCODE sc;
    BYTE *AlignedKey;

    _tprintf(_TEXT("Looking for cypher %s\r\n"), Algorithm);
    Cypher = GetCypher(Algorithm,Arg1);
    if (Cypher == NULL)
        return FALSE;

    KeyLen = _tcslen(Key);

    /* Unfortunately, the KEY must be word-aligned too.
     * Else the ARM gets confused and screws up.
     * malloc() makes sure its aligned.
     */
    AlignedKey = malloc(KeyLen);
    if (AlignedKey == NULL) {
        _tprintf(_TEXT("malloc(%d) FAILED\r\n"), KeyLen);
        goto Out;
    }

    memcpy(AlignedKey,Key,KeyLen);

    _tprintf(_TEXT("Setting key (length=%d)\r\n"), KeyLen);
    sc = Cypher->v->SetKey(Cypher,
                           (const BYTE *) AlignedKey,
                           KeyLen,
                           CYPHER_ENCRYPT|CYPHER_DECRYPT);

    free(AlignedKey);

    if (FAILED(sc)) {
        _tprintf(_TEXT("SetKey failed sc=%x\r\n"), sc);
        goto Out;
    }

    _tprintf(_TEXT("Encrypting file %s into file %s\r\n"), FileIn,FileOut);
    ok = CypherTest(Cypher,FileIn,FileOut,CYPHER_ENCRYPT);
    if (!ok)
        goto Out;

    _tprintf(_TEXT("Decrypting file %s into file %s\r\n"), FileOut, File3);
    ok = CypherTest(Cypher,FileOut,File3,CYPHER_DECRYPT);

 Out:
    Cypher->v->Release(Cypher);

    return ok;
}

/* Trivial crypto test
 */
int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    UINT Arg1;
    BOOL ok = FALSE;
    
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_TEXT("No command line args (%x)\r\n"), sc);
        goto Out;
    }

    _tprintf(_TEXT("  Test %s\n\r"), argv[0]);

    if (argc < 7) {
        _tprintf(_TEXT("Not enough command line args. Usage:\r\n"));
        _tprintf(_TEXT("%s CypherName KeyStrength infilename outfilename key\r\n"),
                 argv[0]);
        goto Out;
    }

    Arg1 = _ttoi(argv[2]);
    ok = Test(argv[1],Arg1,argv[3],argv[4],argv[5],argv[6]);

 Out:
    _tprintf(_TEXT("Test %s %s\r\n"), argv[0], 
             (ok) ? _TEXT("PASSED") : _TEXT("FAILED"));
    (void) FreeCmdLine(&argc,&argv);
    exit((ok)?EXIT_SUCCESS:EXIT_FAILURE);
}

