/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <mmlite.h>
#include <stdlib.h>
#include <tchar.h>
#include <crypto.h>

#include <mmmacros.h>
#include <netif.h>
#include <limits.h>

/* Utils
 */

PICYPHER
GetCypher(
          _TCHAR *Algorithm,
          UINT Arg1
          )
{
    SCODE sc;
    PICYPHER Cypher;
    PICYPHERCONSTRUCTOR Factory;
    PIUNKNOWN pUnk;

    sc = BindToObject(CurrentNameSpace(), Algorithm,
                      NAME_SPACE_READ, &IID_ICypherConstructor,
                      (void *) &Factory);
    if (FAILED(sc)) {
        _tprintf(_TEXT("Could not find the %s factory sc=%x\r\n"),
                 Algorithm, sc);
        return NULL;
    }

    sc = Factory->v->Constructor(Factory,&pUnk,Algorithm,Arg1);
    (void) Factory->v->Release(Factory);

    if (FAILED(sc)) {
        _tprintf(_TEXT("Constructor failed sc=%x\r\n"), sc);
        return NULL;
    }

    sc = pUnk->v->QueryInterface(pUnk,&IID_ICypher,(void**)&Cypher);
    pUnk->v->Release(pUnk);

    if (FAILED(sc)) {
        _tprintf(_TEXT("%s is not a cypher sc=%x\r\n"), Algorithm, sc);
        return NULL;
    }

    return Cypher;
}

BOOL
CypherTest(
           PICYPHER Cypher,
           _TCHAR *FileIn,
           _TCHAR *FileOut,
           CYPHER_OPS Operation
       )
{
    METHOD_DECLARE_NOTHIS(cryp2.c, CypherTest);
    PIFILE Fin = NULL, Fout = NULL;
    UINT nBytes;
    PIBUFFER m = NULL;
    PIENDPOINTFACTORY epf = NULL;

    Fin = Fout = NULL;

    _tprintf(_TEXT("Opening %s for reading\r\n"), FileIn);
    sc = BindToObject(CurrentNameSpace(), FileIn,
                      NAME_SPACE_READ,
                      &IID_IFile, (void **) &Fin);
    CHECKSC("open input");

    _tprintf(_TEXT("Opening %s for writing\r\n"), FileOut);
    sc = BindToObject(CurrentNameSpace(), FileOut,
                      NAME_SPACE_WRITE | NAME_SPACE_CREATE,
                      &IID_IFile, (void **) &Fout);
    CHECKSC("open output");

    sc = Fout->v->SetSize(Fout, UINT64_ZERO);
    CHECKSC("truncate output file");

    /* Create iBuffer */
    sc = BindToObject(CurrentNameSpace(), _T("COB/protocol.cob"),
                      NAME_SPACE_READ,
                      &IID_IEndpointFactory, (void **) &epf);
    CHECKSC("bind to COB/protocol.cob");

    sc = epf->v->AllocateBuffer(epf, 0, NULL, &m);
    CHECKSC("AllocateBuffer");

    /* Read in data */
    sc = m->v->ReadWriteFile(m, Fin, FALSE,
                             UINT64_ZERO, 0, UINT_MAX,
                             &nBytes);
    CHECKSC("read input using ReadWriteFile");

    /* ok, do it.
     */
    printf("Applying cypher\n");
    sc = m->v->ApplyCypher(m, Cypher, Operation);
    CHECKSC("ApplyCypher");

    /* Output result */
    sc = m->v->ReadWriteFile(m, Fout, TRUE,
                             UINT64_ZERO, 0, UINT_MAX,
                             &nBytes);
    CHECKSC("write output using ReadWriteFile");

METHOD_CLEANUP;
    RELEASE(Fout);
    RELEASE(Fin);
    RELEASE(m);
    RELEASE(epf);

    return SUCCEEDED(sc);
}

BOOL
Test(
     _TCHAR *Algorithm,
     UINT Arg1,
     _TCHAR *FileIn,
     _TCHAR *FileOut,
     _TCHAR *File3,
     _TCHAR *Key
     )
{
    PICYPHER Cypher;
    BOOL ok = FALSE;
    UINT KeyLen;
    SCODE sc;
    BYTE *AlignedKey;

    _tprintf(_TEXT("Looking for cypher %s\r\n"), Algorithm);
    Cypher = GetCypher(Algorithm,Arg1);
    if (Cypher == NULL)
        return FALSE;

    KeyLen = _tcslen(Key);

    /* Unfortunately, the KEY must be word-aligned too.
     * Else the ARM gets confused and screws up.
     * malloc() makes sure its aligned.
     */
    AlignedKey = malloc(KeyLen);
    if (AlignedKey == NULL) {
        _tprintf(_TEXT("malloc(%d) FAILED\r\n"), KeyLen);
        goto Out;
    }

    memcpy(AlignedKey,Key,KeyLen);

    _tprintf(_TEXT("Setting key (length=%d)\r\n"), KeyLen);
    sc = Cypher->v->SetKey(Cypher,
                           (const BYTE *) AlignedKey,
                           KeyLen,
                           CYPHER_ENCRYPT|CYPHER_DECRYPT);

    free(AlignedKey);

    if (FAILED(sc)) {
        _tprintf(_TEXT("SetKey failed sc=%x\r\n"), sc);
        goto Out;
    }

    _tprintf(_TEXT("Encrypting file %s into file %s\r\n"), FileIn,FileOut);
    ok = CypherTest(Cypher,FileIn,FileOut,CYPHER_ENCRYPT);
    if (!ok)
        goto Out;

    _tprintf(_TEXT("Decrypting file %s into file %s\r\n"), FileOut, File3);
    ok = CypherTest(Cypher,FileOut,File3,CYPHER_DECRYPT);

 Out:
    Cypher->v->Release(Cypher);

    return ok;
}

/* Trivial crypto test
 */
int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    UINT Arg1;
    BOOL ok = FALSE;
    
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_TEXT("No command line args (%x)\r\n"), sc);
        goto Out;
    }

    _tprintf(_TEXT("  Test %s\n\r"), argv[0]);

    if (argc < 7) {
        _tprintf(_TEXT("Not enough command line args. Usage:\r\n"));
        _tprintf(_TEXT("%s CypherName KeyStrength infilename outfilename key\r\n"),
                 argv[0]);
        goto Out;
    }

    Arg1 = _ttoi(argv[2]);
    ok = Test(argv[1],Arg1,argv[3],argv[4],argv[5],argv[6]);

 Out:
    _tprintf(_TEXT("Test %s %s\r\n"), argv[0], 
             (ok) ? _TEXT("PASSED") : _TEXT("FAILED"));
    (void) FreeCmdLine(&argc,&argv);
    exit((ok)?EXIT_SUCCESS:EXIT_FAILURE);
}

