/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Computes the appropriate constants to tune the Delay() API.
 * The result should be put in the platform-specific code.
 */
#include <mmlite.h>
#include <tchar.h>
#include <stdio.h>

extern void MyDelay(INT32 LoopCount);
#define TUNEDELAY 1
#include "pdelay.c"

extern UINT TuneDelayMultiplier;
extern UINT TuneDelayShifter;
extern UINT TuneDelayOverhead;

UINT TunePoint           = 100;

/* Automagically adjust Delay().
 * Reduce/Increase the Spin factor until we get to within +-10% of the
 * intended value.   Tuning values are kept in machdep code, so that they
 * can be set to a more accurate initial value.
 */
void AutotuneDelay(void)
{
    TIME Ts, Te;
    UINT Td, PrevCnt[2];

    /* This small array of past values avoids oscillating loops
     */
    PrevCnt[1] = 0;

    /* Repeat until we are satisfied
     */
    do {
        /* This is what we use this time around.
         */
        PrevCnt[0] = PrevCnt[1];
        PrevCnt[1] = TuneDelayMultiplier;

        /* Do it
         */
        Ts = CurrentTime();
        MyDelay(TunePoint);
        Te = CurrentTime();

        /* How long did it take.
         */
        Td = Uint64ToUint32(TimeSubtract(Te,Ts));
        Td -= TimingOverhead;

        /* Are we within limits ?
         */
#define Nine 10 /*9*/
        if (Td < (Nine*TunePoint)) {
            /* Spin factor too small
             */
            TuneDelayMultiplier++;

        } else
#define Eleven 10 /*11*/
        if (Td > (Eleven*TunePoint)) {
            /* Spin factor too large
             */
            if (TuneDelayMultiplier)
                TuneDelayMultiplier--;
        }
        /* else we more or less got it.
         */
        if (Verbose)
        _tprintf(_T("Tunafish: %d %d %d (%d)\r\n"),
                 PrevCnt[0], PrevCnt[1], TuneDelayMultiplier,Td);

    } while (PrevCnt[1] != TuneDelayMultiplier && PrevCnt[0] != TuneDelayMultiplier);
}

/* A somewhat more accurate timing of a small MyDelay()
 */
UINT Times8DelayOf(UINT uSeconds)
{
    TIME Ts, Te;
    UINT Tmin, Tcur, i;

    /* Repeat ten times and keep the non-zero minimum
     */
    Tmin = ~0;
    for (i = 0; i < 10; i++) {
        /* Do it 8 times, DONT shift out at the end (keep bits)
         */
        Ts = CurrentTime();
        MyDelay(uSeconds);
        MyDelay(uSeconds);
        MyDelay(uSeconds);
        MyDelay(uSeconds);
        MyDelay(uSeconds);
        MyDelay(uSeconds);
        MyDelay(uSeconds);
        MyDelay(uSeconds);
        Te = CurrentTime();

        Te = TimeSubtract(Te,Ts);
        Tcur = Uint64ToUint32(Te) - TimingOverhead;

        if (Tcur && Tcur < Tmin)
            Tmin = Tcur;
    }

    return Tmin;
}

/* The Delay() function, no matter what the implementation, must behave as
 * a linear function of the input parameter:
 *     T = f(uSec) = C + L * uSec
 * If we measure it at two known points we can compute both constants:
 *     T1 = C + L * uSec1       and     T2 = C + L * uSec2
 * Solving the system gives:
 *     L = (T2 - T1) / (uSec2 - uSec1)
 *     C = (T2 + T1 - L * (uSec2 + uSec1)) / 2
 * NB: All units are in uSec in these formulas, we'll use TIME below.
 *
 * Our problem here is twofold.  
 * First we must determine L and derive from it the two values for
 * TuneDelayMultiplier and TuneDelayShifter.
 * Second we must determine C and derive from it the correct value
 * for TuneDelayOverhead.  This comes second because the first two
 * values must be known to derive TuneDelayOverhead = f(Delay(0)).
 */
BOOL AlgebraicDelay(void)
{
    UINT T1, T2;
#define uSec2 181   /* both primes, their diff is 114=2^1 3^1 19^1 */
#define uSec1 67
    UINT64 C, L, temp;

    /* Zero-point
     */
    TuneDelayMultiplier = 1;
    TuneDelayShifter = 0;
    TuneDelayOverhead = 0;

    /* Take the samples (NB: T granularity = 100ns/8)
     */
    T1 = Times8DelayOf(uSec1);
    T2 = Times8DelayOf(uSec2);
    if (Verbose)
        _tprintf(_T("T2=f(%d)=%d T1=f(%d)=%d\r\n"), 
                 uSec2, T2, uSec1, T1);

#if 0
    /* yes it was more readable in 32 bits.. sigh. */
    L = ((T2 - T1) * 1000 * 1024) / ((uSec2 - uSec1) * 80);

    C = ((T2 + T1) * 1000 * 1024) / 80;
    C = C - (L * ((uSec2 + uSec1)));
    C = C / 2;
#endif
    /* Compute L in nanoseconds (the 1000) and with a few extra bits (the 1024)
     * NB: the 80 = 10 * 8 maps T to uSecs
     */
    Uint32ToUint64(L,(T2 - T1));
    L = Uint64TimesUint32(L,1000 * 1024);
    L = Uint64DividedByUint32(L,((uSec2 - uSec1) * 80));

    /* Same with C
     */
    Uint32ToUint64(C,(T2+T1));
    C = Uint64TimesUint32(C,1000 * 1024);
    C = Uint64DividedByUint32(C,80);
    temp = Uint64TimesUint32(L,(uSec2 + uSec1));
    C = Uint64Subtract(C,temp);
    C = Uint64RShift(C,1);

    if (Verbose) {
        _tprintf(_T("L = %ld %lx (%ld.~%d %lx.%x)\r\n"), L, L,
                 Uint64RShift(L,10), Uint64ToUint32(L) & 1023,
                 Uint64RShift(L,10), Uint64ToUint32(L) & 1023);
        _tprintf(_T("C = %ld %lx (%ld.~%d %lx.%x)\r\n"), C, C,
                 Uint64RShift(C,10), Uint64ToUint32(C) & 1023,
                 Uint64RShift(C,10), Uint64ToUint32(C) & 1023);
    }

    _tprintf(_T("Per-loop cost is ~%ld ns\r\n"), Uint64RShift(L,10));
    _tprintf(_T("Call-setup cost is ~%ld ns\r\n"), Uint64RShift(C,10));

    /* Bad timer protection
     */
    if (Int64IsZero(L)) {
        _tprintf(_T("Like I said.. it did not work.\r\n"));
        TuneDelayMultiplier = 0;
        return FALSE;
    }

    /* In conclusion..
     */
    TuneDelayShifter = 10; /* the 1024 above */

    L = Uint64RShift(L,10);
    Uint32ToUint64(temp,1000);
    temp = Uint64LShift(temp,TuneDelayShifter);
    temp = Uint64DividedByUint64(temp,L);
    TuneDelayMultiplier = Uint64ToUint32(temp);

    C = Uint64RShift(C,10);
    temp = Uint64TimesUint32(C,TuneDelayMultiplier);
    temp = Uint64DividedByUint32(temp,1000);
    TuneDelayOverhead = Uint64ToUint32(temp);

    /* Say what we came up with
     */
    _tprintf(_T("Algebra says: %d %d %d\r\n"),
             TuneDelayMultiplier,
             TuneDelayShifter,
             TuneDelayOverhead);
    return TRUE;
}

/*
 */
int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;

    /* Arguments
     */
    sc = ParseCmdLine(&argc, &argv);
    if (!FAILED(sc) && (argc > 1))
        Verbose = !Verbose;
    (void) FreeCmdLine(&argc,&argv);

    /* First thing first
     */
    ComputeOverhead();
    if (TimingOverhead == 0) {
        _tprintf(_T("The timer is too coarse, this will not work.\r\n"));
    }

    /* Get a solid start
     */
    if (AlgebraicDelay()) {

        /* Fine tune from there
         */
        AutotuneDelay();

        /* Verify
         */
        TestResult(0,NULL);

        /* Say what we came up with
         */
        _tprintf(_T("/* Suggested implementation for Delay():\r\n */\r\n"));
        _tprintf(_T("void Delay(UINT uSec)\r\n") \
                 _T("{\r\n") \
                 _T("    INT nLoops = (INT)uSec;\r\n") \
                 _T("    nLoops = (nLoops * %d) - %d;\r\n") \
                 _T("    nLoops = nLoops >> %d;\r\n") \
                 _T("    TuneDelayLoop(nLoops);\r\n") \
                 _T("}\r\n"),
                 TuneDelayMultiplier, TuneDelayOverhead, TuneDelayShifter);

        _tprintf(_T("[This test ran @x%x]\r\n"),(ADDRESS)_tmain);
    }
    return 0;
}

