/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <arm\at91m55800.h>

/* This is specific to the EB55 board and its AD/DA converters.
 * The M55800A processor has 2x4 AD converters and 2 DA converters, 10 bits.
 * Builtin connections on the board are as follows.
 * - AD0 links to DA1
 * - AD1 links to a temperature sensor near the 32KHz crystal
 * - AD2 links to VDDCORE via a 10Kohm resistor, reading VDDCORE/2
 * - AD4 links to DA0
 * - AD3,5,6,7 are free.
 * - AD0-trigger links to PB6 internally, to PA4 externally
 * - AD1-trigger links to PB7 internally, to PA7 externally
 * Reference voltages are 2.5V via AnalogDevices REF192GS chips.
 */

/* Who needs floats
 */
typedef union _DECIMAL {
    struct {
        UINT16 DecimalPart;
        INT16  IntegerPart;
    } v;
    UINT32 u;
    INT32 i;
} DECIMAL;
#define Point(_i_,_d_) {{_d_,_i_}}

extern DECIMAL SubtractDecimals(DECIMAL a, DECIMAL b);
extern void PrintDecimal(DECIMAL a);
extern void TestDecimals(void);
#include "dec.c"

/* Setup the ADC the way we want it
 */
void Initialize(void)
{
    /* Enable clock to the peripherals
     */
    ThePmc->PerClockEnable = PMCPC_ADC0 | PMCPC_DAC1;

    /* Dont know how long this should be, really
     */
    Delay(1000);

    /* Use ADC0 with software triggers, max slow.
     */
    Adc0->Mode = ADCM_SW_TRIGGER|ADCM_10_BITS|ADCM_PRESCALE(63);

    /* Dont enable any channels, they work one at a time anyways
     */

    /* DAC
     */
    Dac1->Mode = DACM_SW_TRIGGER | DACM_10_BITS;
}

#define PIPELINE 1

/* Change sampling speed
 */
void SetPrescaler(UINT Scaler)
{
    UINT32 Enabled = Adc0->ChannelStatus;

    /* Disable all channels
     */
    Adc0->ChannelDisable = Enabled;

    /* Dont know how long this should be, really
     */
    Delay(1000);

    /* Set up new speed
     */
    Adc0->Mode = ADCM_SW_TRIGGER|ADCM_10_BITS|ADCM_PRESCALE(Scaler);

    /* Re-enable the channels that were
     */
    Adc0->ChannelEnable = Enabled;

    /* Dont know how long this should be, really
     */
    Delay(1000);

#if PIPELINE
    /* Start one conversion
     */
    Adc0->Control = ADCC_START;
#endif
}

/* Enable/disable a channel
 */
void EnableChannel(UINT Channel)
{
    Adc0->ChannelEnable = ADCS_CHANNEL(Channel);
#if PIPELINE
    /* Start one conversion
     */
    Adc0->Control = ADCC_START;
#endif
}

void DisableChannel(UINT Channel)
{
    Adc0->ChannelDisable = ADCS_CHANNEL(Channel);
}

/* Disable and minimize power waste
 */
void Finalize(void)
{
    /* Disable all channels via reset
     */
    Adc0->Control = ADCC_RESET;
    Dac1->Control = DACC_RESET;

    /* Dont know how long this should be, really
     */
    Delay(1000);

    /* Disable clock to the peripheral
     */
    ThePmc->PerClockDisable = PMCPC_ADC0 | PMCPC_DAC1;
}

/* Take one sample from the given channel
 */
#define INVALID_SAMPLE (~0)

UINT32 TakeSample(int Channel)
{
#if PIPELINE
    UINT32 v;

    /* Wait till done, make sure its not broken
     */
    v = 0;
    while ((Adc0->Status & ADCS_CHANNEL(Channel)) == 0) {
        if (++v > 0x10000) {
            _tprintf(_T("ADC%d did not settle\r\n"),Channel);
            return INVALID_SAMPLE;
        }
    }

    /* Return the converted 10-bit value
     */
    v = Adc0->Data[Channel] & 0x3ff;

    /* Get another started
     * NB: this clears the done bit
     */
    Adc0->Control = ADCC_START;

    /* Done
     */
    return v;
#else
    UINT32 v;

    /* Start the conversion
     * NB: this clears the done bit
     */
    Adc0->Control = ADCC_START;

    /* Wait till done, make sure its not broken
     */
    v = 0;
    while ((Adc0->Status & ADCS_CHANNEL(Channel)) == 0) {
        if (++v > 0x10000) {
            _tprintf(_T("ADC%d did not settle\r\n"),Channel);
            return INVALID_SAMPLE;
        }
    }

    /* Return the converted 10-bit value
     */
    return Adc0->Data[Channel] & 0x3ff;
#endif
}

/* Set the output of the DAC
 */
BOOL DacOutput(UINT Value)
{
    UINT32 v;

    /* Wait till available, make sure its not broken
     */
    v = 0;
    while ((Dac1->Status & 2/*DACM_READY*/) == 0) {
        if (++v > 0x10000) {
            _tprintf(_T("DAC%d did not settle\r\n"),1);
            return FALSE;
        }
    }

    /* Transfer it to the holding register
     */
    Dac1->DataHold = Value & 0x3ff;

    /* Done ok
     */
    return TRUE;
}

/* Pause..
 */
void SleepOneSec(void)
{
    TIME t;
    Int32ToInt64(t, TIME_RELATIVE(TIME_SECONDS(1)));
    SleepUntil(t);
}

/*
 */
void LoopbackTest(void)
{
    int i, j, nFail;
    UINT32 v, avg, s;

    /* We'll be using the looped one
     */
    EnableChannel(0);

#define BOUND 2
    _tprintf(_T("Checking that ADC0 stays within +-%d of the DAC1 value.\r\n"),
             BOUND);

    /* One per second for a minute
     */
    nFail = 0;
    for (i = 1; i <= 60*1; i++) {

        /* Slope of 17 to cover the whole range in about a minute
         */
        s = i * 17;

        /* Set the output, skip if it doesnt take
         */
        if (!DacOutput(s)) {
            nFail++;
            continue;
        }

        /* Says it settles in 5us, give it 50
         */
        Delay(50);

        /* Must read multiple times for this 'successive approximation' 
         * thing to take.. sigh.  And even then it swings around it.
         * So drop a couple readings and take the rounded-up average
         */
        v = TakeSample(0);
        v = TakeSample(0);
        avg = 0;
        for (j = 0; j < 8; j++) {
            v = TakeSample(0);
            avg += v;
        }
        avg /= 4;
        if (avg & 1)
            avg++;
        avg /= 2;

        /* Check its what we want, within +-2 (sigh)
         */
        if ((avg > (s+BOUND)) || (avg < (s-BOUND))) {
            nFail++;
            _tprintf(_T("Error: DAC(%d) !~= %d. ADC: %d"),
                     s, avg, v);
            for (j = 0; j < 5; j++)
                _tprintf(_T(", %d"),TakeSample(0));
            _tprintf(_T("\r\n"));
        } else
            _tprintf(_T(". "));

        /* Not too fast
         */
        SleepOneSec();
    }

    if (nFail)
        _tprintf(_T("\r\nOut of bounds %d times out of %d\r\n"),nFail,i-1);


    DisableChannel(0);
}

/* Main driving program
 */
int _tmain (int argc, _TCHAR **argv)
{
    /* Get started
     */
    Initialize();

    /* Test the loopback
     */
    LoopbackTest();

    /* Close up shop
     */
    Finalize();

    return 0;    
}





