/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <arm\at91m55800.h>

/* This is specific to the EB55 board and its AD/DA converters.
 * The M55800A processor has 2x4 AD converters and 2 DA converters, 10 bits.
 * Builtin connections on the board are as follows.
 * - AD0 links to DA1
 * - AD1 links to a temperature sensor near the 32KHz crystal
 * - AD2 links to VDDCORE via a 10Kohm resistor, reading VDDCORE/2
 * - AD4 links to DA0
 * - AD3,5,6,7 are free.
 * - AD0-trigger links to PB6 internally, to PA4 externally
 * - AD1-trigger links to PB7 internally, to PA7 externally
 * Reference voltages are 2.5V via AnalogDevices REF192GS chips.
 */

/* Who needs floats
 */
typedef union _DECIMAL {
    struct {
        UINT16 DecimalPart;
        INT16  IntegerPart;
    } v;
    UINT32 u;
    INT32 i;
} DECIMAL;
#define Point(_i_,_d_) {{_d_,_i_}}

extern DECIMAL SubtractDecimals(DECIMAL a, DECIMAL b);
extern void PrintDecimal(DECIMAL a);
extern void TestDecimals(void);
#include "dec.c"

/* Setup the ADC the way we want it
 */
void Initialize(void)
{
    /* Enable clock to the peripheral
     */
    ThePmc->PerClockEnable = PMCPC_ADC0;

    /* Dont know how long this should be, really
     */
    Delay(1000);

    /* Use ADC0 with software triggers, do not enable sleep mode
     * [triggers even more fluctuautions in the reading.. umbug]
     * Set prescaler to stay within 1.1 MHz max clock (/max)
     * BUGBUG read system clock and make this adaptable.
     */
    Adc0->Mode = ADCM_SW_TRIGGER|ADCM_10_BITS|ADCM_PRESCALE(63);

    /* Dont enable any channels, they work one at a time anyways
     */
}

#define PIPELINE 1

/* Change sampling speed
 */
void SetPrescaler(UINT Scaler)
{
    UINT32 Enabled = Adc0->ChannelStatus;

    /* Disable all channels
     */
    Adc0->ChannelDisable = Enabled;

    /* Dont know how long this should be, really
     */
    Delay(1000);

    /* Set up new speed
     */
    Adc0->Mode = ADCM_SW_TRIGGER|ADCM_10_BITS|ADCM_PRESCALE(Scaler);

    /* Re-enable the channels that were
     */
    Adc0->ChannelEnable = Enabled;

    /* Dont know how long this should be, really
     */
    Delay(1000);

#if PIPELINE
    /* Start one conversion
     */
    Adc0->Control = ADCC_START;
#endif
}

/* Enable/disable a channel
 */
void EnableChannel(UINT Channel)
{
    Adc0->ChannelEnable = ADCS_CHANNEL(Channel);
#if PIPELINE
    /* Start one conversion
     */
    Adc0->Control = ADCC_START;
#endif
}

void DisableChannel(UINT Channel)
{
    Adc0->ChannelDisable = ADCS_CHANNEL(Channel);
}

/* Disable and minimize power waste
 */
void Finalize(void)
{
    /* Disable all channels via reset
     */
    Adc0->Control = ADCC_RESET;

    /* Dont know how long this should be, really
     */
    Delay(1000);

    /* Disable clock to the peripheral
     */
    ThePmc->PerClockDisable = PMCPC_ADC0;
}

/* Take one sample from the given channel
 */
#define INVALID_SAMPLE (~0)

UINT32 TakeSample(int Channel)
{
#if PIPELINE
    UINT32 v;

    /* Wait till done, make sure its not broken
     */
    v = 0;
    while ((Adc0->Status & ADCS_CHANNEL(Channel)) == 0) {
        if (++v > 0x10000) {
            _tprintf(_T("ADC%d did not settle\r\n"),Channel);
            return INVALID_SAMPLE;
        }
    }

    /* Return the converted 10-bit value
     */
    v = Adc0->Data[Channel] & 0x3ff;

    /* Get another started
     * NB: this clears the done bit
     */
    Adc0->Control = ADCC_START;

    /* Done
     */
    return v;
#else
    UINT32 v;

    /* Start the conversion
     * NB: this clears the done bit
     */
    Adc0->Control = ADCC_START;

    /* Wait till done, make sure its not broken
     */
    v = 0;
    while ((Adc0->Status & ADCS_CHANNEL(Channel)) == 0) {
        if (++v > 0x10000) {
            _tprintf(_T("ADC%d did not settle\r\n"),Channel);
            return INVALID_SAMPLE;
        }
    }

    /* Return the converted 10-bit value
     */
    return Adc0->Data[Channel] & 0x3ff;
#endif
}

/* Sample and convert the temperature sensor
 */
DECIMAL ReadTemperature(UINT nPrecharge)
{
    DECIMAL r;

    do {
        r.u = TakeSample(1);
    } while (--nPrecharge > 0);

    if (r.u != INVALID_SAMPLE) {
        /* The LM61B's Vs is VDDA == 3.3V (I measured 3.29)
         * It has a DC offset of 600mV (e.g. Vo at 0C).
         * "Output is +300mV to +1600mV" for the -30C..+100C temperature range.
         * The slope is 10mV/C, accuracy is +-3C, range -25C..85C
         * [should get 24.3C with Vo=0.843V]
         */
        static DECIMAL Offset = Point(0,39322); /* funny 0.6 */

        r.u = (r.u * (/*x2.5*/0x20000+0x8000)) / 1023;
        r   = SubtractDecimals(r,Offset);
        r.u *= 100;
    }

    /* Return value
     */
    return r;
}

/* Some checks on the temperature sensor
 */
void TestTemp(void)
{
    int i;
    DECIMAL v;

    EnableChannel(1);

    /* Read and print the temperature a few times
     */
#define NL 4
    _tprintf(_T("Making %d readings of the temperature sensor:\r\n"),10*NL);
    for (i = 1; i <= 10*NL; i++) {
        v = ReadTemperature(1);
        _tprintf(_T("  "));
        PrintDecimal(v);
        if ((i % NL) == 0)
            _tprintf(_T("\r\n"));
    }

    DisableChannel(1);
}

/* Sample temp many times and report range of values
 */
static UINT BestTempSpeed = 59;
static DECIMAL BestDeltaT = Point(0x7fff,0xffff);

void TempRange(UINT Speed)
{
    DECIMAL MaxT, MinT, v;
    int i;

    /* Set the ADC speed
     */
    SetPrescaler(Speed);

    /* Take min and max values, no drops
     */
    MinT.u = (UINT)(~0)>>1;
    MaxT.i = 0;

    /* 10k times seems enough
     */
    for (i = 0; i < 10000; i++) {
        v = ReadTemperature(1);
        if (v.u == INVALID_SAMPLE) {
            return;
        }
        if (v.i < MinT.i)
            MinT = v;
        if (v.i > MaxT.i)
            MaxT = v;
    }

    /* Compute range
     */
    v = SubtractDecimals(MaxT,MinT);

    /* Say what we got
     */
    _tprintf(_T("[prescaler=%d] %d samples, range "), Speed, i);
    PrintDecimal(MinT);
    _tprintf(_T(".."));
    PrintDecimal(MaxT);
    _tprintf(_T(" ["));
    PrintDecimal(v);
    _tprintf(_T("]\r\n"));

    /* Is this the most accurate so far
     */
    if (v.i < BestDeltaT.i) {
        BestTempSpeed = Speed;
        BestDeltaT = v;
    }
}

/* Check stability of software conversion across a number of speeds
 */

void Test2Temp(void)
{
    UINT Speed;

    EnableChannel(1);

    /* Make a bunch of measures, see what range we get
     */
    _tprintf(_T("Estimating the best prescaler value for the ADC:\r\n"));
    for (Speed = 15; /* max speed */
         Speed < 64;
         Speed++) {
        TempRange(Speed);
    }

    /* Say what
     */
    _tprintf(_T("Minimum range was at prescaler=%d ["), BestTempSpeed);
    PrintDecimal(BestDeltaT);
    _tprintf(_T("].  But ...\r\n"));

    /* Do it again for grins
     */
    TempRange(BestTempSpeed);

    DisableChannel(1);
}

/* ~~Continuous temperature reading
 */
#define PAUSE 1
void SleepOneSec(void)
{
    TIME t;
    Int32ToInt64(t, TIME_RELATIVE(TIME_SECONDS(PAUSE)));
    SleepUntil(t);
}

int nMinutes = 1;
void Termo(UINT Speed)
{
    int i;
    DECIMAL v;

    EnableChannel(1);

    /* Set selected speed
     */
    SetPrescaler(Speed);

    /* Say what we'll do, it can take a while
     */
    _tprintf(_T("Reading the temperature sensor at %d second intervals for %d minute(s):\r\n"),
             PAUSE,nMinutes);
    for (i = 0; i < 60*nMinutes; i++){
        /* Back to the start of the line
         */
        _tprintf(_T("\rT="));

        /* Read and print
         */
        v = ReadTemperature(5);
        PrintDecimal(v);

        /* Cleanup rest of previous line
         */
        _tprintf(_T("        "));

        /* Let her read it
         */
        SleepOneSec();
    }
    _tprintf(_T("\r\n"));

    DisableChannel(1);
}

/* Main driving program
 */
int _tmain (INT argc, _TCHAR **argv)
{
    BOOL doTemp = TRUE, doTemp2 = TRUE;
    BOOL doTermo = TRUE;

    /* We got options
     */
    (void) ParseCmdLine(&argc, &argv);
    while (argc > 1) {
        if (argv[1][1] == _T('t'))
            doTemp = !doTemp;
        if (argv[1][1] == _T('T'))
            doTemp2 = !doTemp2;
        if (argv[1][1] == _T('c')) {
            if (argv[1][2] != 0)
                nMinutes = _ttoi(&argv[1][2]);
            else
                doTermo = !doTermo;
        }
        argc--, argv++;
    }
    (void) FreeCmdLine(&argc,&argv);

    /* Make sure our trivia math didnt break
     */
    TestDecimals();

    /* Get set
     */
    Initialize();

    /* Do what we are asked to do
     */
    if (doTemp) TestTemp();
    if (doTemp2) Test2Temp();
    if (doTermo) Termo(BestTempSpeed);
        
    /* Close up shop
     */
    Finalize();

    return 0;    
}





