/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <arm\at91m55800.h>

/* This is specific to the EB55 board and its AD/DA converters.
 * The M55800A processor has 2x4 AD converters and 2 DA converters, 10 bits.
 * Builtin connections on the board are as follows.
 * - AD0 links to DA1
 * - AD1 links to a temperature sensor near the 32KHz crystal
 * - AD2 links to VDDCORE via a 10Kohm resistor, reading VDDCORE/2
 * - AD4 links to DA0
 * - AD3,5,6,7 are free.
 * - AD0-trigger links to PB6 internally, to PA4 externally
 * - AD1-trigger links to PB7 internally, to PA7 externally
 * Reference voltages are 2.5V via AnalogDevices REF192GS chips.
 */

/* Who needs floats
 */
typedef union _DECIMAL {
    struct {
        UINT16 DecimalPart;
        INT16  IntegerPart;
    } v;
    UINT32 u;
    INT32 i;
} DECIMAL;
#define Point(_i_,_d_) {{_d_,_i_}}

extern DECIMAL SubtractDecimals(DECIMAL a, DECIMAL b);
extern void PrintDecimal(DECIMAL a);
extern void TestDecimals(void);
#include "dec.c"

/* Setup the ADC the way we want it
 */
void Initialize(void)
{
    /* Enable clock to the peripheral
     */
    ThePmc->PerClockEnable = PMCPC_ADC0;

    /* Dont know how long this should be, really
     */
    Delay(1000);

    /* Use ADC0 with software triggers, do not enable sleep mode
     * [triggers even more fluctuautions in the reading.. umbug]
     * Set prescaler to stay within 1.1 MHz max clock (/max)
     * BUGBUG read system clock and make this adaptable.
     */
    Adc0->Mode = ADCM_SW_TRIGGER|ADCM_10_BITS|ADCM_PRESCALE(63);

    /* Dont enable any channels, they work one at a time anyways
     */
}

#define PIPELINE 1

/* Change sampling speed
 */
void SetPrescaler(UINT Scaler)
{
    UINT32 Enabled = Adc0->ChannelStatus;

    /* Disable all channels
     */
    Adc0->ChannelDisable = Enabled;

    /* Dont know how long this should be, really
     */
    Delay(1000);

    /* Set up new speed
     */
    Adc0->Mode = ADCM_SW_TRIGGER|ADCM_10_BITS|ADCM_PRESCALE(Scaler);

    /* Re-enable the channels that were
     */
    Adc0->ChannelEnable = Enabled;

    /* Dont know how long this should be, really
     */
    Delay(1000);

#if PIPELINE
    /* Start one conversion
     */
    Adc0->Control = ADCC_START;
#endif
}

/* Enable/disable a channel
 */
void EnableChannel(UINT Channel)
{
    Adc0->ChannelEnable = ADCS_CHANNEL(Channel);
#if PIPELINE
    /* Start one conversion
     */
    Adc0->Control = ADCC_START;
#endif
}

void DisableChannel(UINT Channel)
{
    Adc0->ChannelDisable = ADCS_CHANNEL(Channel);
}

/* Disable and minimize power waste
 */
void Finalize(void)
{
    /* Disable all channels via reset
     */
    Adc0->Control = ADCC_RESET;

    /* Dont know how long this should be, really
     */
    Delay(1000);

    /* Disable clock to the peripheral
     */
    ThePmc->PerClockDisable = PMCPC_ADC0;
}

/* Take one sample from the given channel
 */
#define INVALID_SAMPLE (~0)

UINT32 TakeSample(int Channel)
{
#if PIPELINE
    UINT32 v;

    /* Wait till done, make sure its not broken
     */
    v = 0;
    while ((Adc0->Status & ADCS_CHANNEL(Channel)) == 0) {
        if (++v > 0x10000) {
            _tprintf(_T("ADC%d did not settle\r\n"),Channel);
            return INVALID_SAMPLE;
        }
    }

    /* Return the converted 10-bit value
     */
    v = Adc0->Data[Channel] & 0x3ff;

    /* Get another started
     * NB: this clears the done bit
     */
    Adc0->Control = ADCC_START;

    /* Done
     */
    return v;
#else
    UINT32 v;

    /* Start the conversion
     * NB: this clears the done bit
     */
    Adc0->Control = ADCC_START;

    /* Wait till done, make sure its not broken
     */
    v = 0;
    while ((Adc0->Status & ADCS_CHANNEL(Channel)) == 0) {
        if (++v > 0x10000) {
            _tprintf(_T("ADC%d did not settle\r\n"),Channel);
            return INVALID_SAMPLE;
        }
    }

    /* Return the converted 10-bit value
     */
    return Adc0->Data[Channel] & 0x3ff;
#endif
}

/* Sample and convert to get the VDDCORE value in volts
 */
DECIMAL Vddcore(void)
{
    DECIMAL r;

    r.u = TakeSample(2);

    if (r.u != INVALID_SAMPLE) {

        /* The value we read is a fraction of 2.5 V,
         * it represents half the measure we want.
         */
        r.u = (r.u * 5 * 0x10000) / 1023;
    }

    /* Return value
     */
    return r;
}

/* Sample many times and print the range of values for VDDCORE
 */
static UINT BestVddSpeed = 0;
static DECIMAL BestDeltaVdd = Point(0x7fff,0xffff);

void VddRange(UINT Speed)
{
    DECIMAL MinVdd, MaxVdd, v;
    int i;

    /* Set the ADC speed
     */
    SetPrescaler(Speed);

    /* Take min and max values, no drops
     */
    MinVdd.u = (UINT)(~0)>>1;
    MaxVdd.i = 0;

    /* 10k times seems enough
     */
    for (i = 0; i < 10000; i++) {
        v = Vddcore();
        if (v.i < MinVdd.i)
            MinVdd = v;
        if (v.i > MaxVdd.i)
            MaxVdd = v;
    }

    /* Compute range
     */
    v = SubtractDecimals(MaxVdd,MinVdd);

    /* Say what we got
     */
    _tprintf(_T("[prescaler=%d] %d samples, range "), Speed, i);
    PrintDecimal(MinVdd);
    _tprintf(_T(".."));
    PrintDecimal(MaxVdd);
    _tprintf(_T(" ["));
    PrintDecimal(v);
    _tprintf(_T("]\r\n"));

    /* Is this the most accurate so far
     */
    if (v.i < BestDeltaVdd.i) {
        BestVddSpeed = Speed;
        BestDeltaVdd = v;
    }
}

/* Some checks on VDD
 */
void TestVdd(void)
{
    int i;
    DECIMAL v, o;
    UINT Speed;

    EnableChannel(2);

    /* Print the first 10 changing values of VDDCORE
     */
#define NREADS 10
    _tprintf(_T("Making %d reads of VDDCORE:\r\n"),NREADS);
    o.u = 0;
    for (i = 0; i < NREADS;) {
        v = Vddcore();
        if (v.u != o.u) {
            o = SubtractDecimals(v,o);
            _tprintf(_T("Vddcore = "));
            PrintDecimal(v);
            _tprintf(_T(" (delta="));
            PrintDecimal(o);
            _tprintf(_T(")\r\n"));
            o = v;
            i++;
        }
    }

    /* Make a bunch of measures, see what range we get
     */
    _tprintf(_T("Estimating the best prescaler value for the ADC:\r\n"));
    for (Speed = 15; /* max speed */
         Speed < 64;
         Speed++) {
        VddRange(Speed);
    }

    /* Say what was most accurate
     */
    _tprintf(_T("Minimum range was at prescaler=%d ["), BestVddSpeed);
    PrintDecimal(BestDeltaVdd);
    _tprintf(_T("].  But ...\r\n"));

    /* Do it again for grins
     */
    VddRange(BestVddSpeed);

    DisableChannel(2);
}

/* Main driving program
 */
int _tmain (int argc, _TCHAR **argv)
{
    /* Make sure our trivia math didnt break
     */
    TestDecimals();

    /* Get set
     */
    Initialize();

    /* Test
     */
    TestVdd();
        
    /* Close up shop
     */
    Finalize();

    return 0;    
}





