/* Copyright (c) Microsoft Corporation. All rights reserved. */

/*
 * Test for the EB63 board buttons SW1..SW4 (and LEDs)
 */

#define DEVICE_ORDER LITTLE_ENDIAN
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>

#include <arm/at91m63200.h>
#include <drivers/drivers.h>


#ifdef _MSC_VER
#pragma warning(disable:4054)   /* cast from function ptr to data ptr (/Za) */
#endif

void setled(int ledno, BOOL on)
{
    if (on)
        PioB->ClearData = _bs32(PIOB_PB8 << (ledno - 1));
    else
        PioB->SetData = _bs32(PIOB_PB8 << (ledno - 1));
}

BOOL TimeToDie = FALSE;

/*
 */
#define nBUTTONS 4
struct _BUTTONS {
    CONDITION ButtonPress;
    UINT ButtonsChanged;
    struct _BUTTON_STATE {
        TIME LastOnAt;
        TIME LastOffAt;
        BOOL On;
    } ButtonState[nBUTTONS];
} TheButtons
  = {CONDITION_INITIALIZER, 0,};

UINT nInterrupts = 0;
UINT nDebounced = 0;
TIME t1,t2,t3;

/* ISR for Buttons with a direct IRQ.
 */
BOOL MCT ButtonIsr(void *This, PBOOL pNotMyInterrupt)
{
    UINT ButtonNo = (UINT)This;
    TIME Now = CurrentTime(), Last;
    static const TIME DEBOUNCE_TIME = Int64Initializer(0,TIME_MILLIS(10));

    UnusedParameter(pNotMyInterrupt);

    nInterrupts++;

    /* Debounce the hw */
    Last = (TheButtons.ButtonState[ButtonNo].On) ?
        TheButtons.ButtonState[ButtonNo].LastOnAt :
        TheButtons.ButtonState[ButtonNo].LastOffAt;
    t1 = Now;
    t2 = Last;
    Last = TimeSubtract(Now,Last);
    t3 = Last;
    if (TimeLess(Last,DEBOUNCE_TIME)) {
        nDebounced++;// = Int64ToInt32(Last);
        return FALSE;
    }

    /* We are the Isr, so we are atomic */
    TheButtons.ButtonsChanged |= 1<<ButtonNo;

    /* Record the event time */
    if (TheButtons.ButtonState[ButtonNo].On)
        TheButtons.ButtonState[ButtonNo].LastOffAt = Now;
    else
        TheButtons.ButtonState[ButtonNo].LastOnAt = Now;

    /* Flip the state bit */
    TheButtons.ButtonState[ButtonNo].On = !TheButtons.ButtonState[ButtonNo].On;

    /* Wake up the thread, if necessary */
    return Condition_InterruptSignal(&TheButtons.ButtonPress);
}

/* ISR for Buttons that are multiplexed
 */
BOOL MCT PioIsr(void *This, PBOOL pNotMyInterrupt)
{
    struct _Pio *Pio = (struct _Pio *)This;
    BOOL rv;
    UINT Pins;

    /* See what ISR-enabled button was pressed
     * NB: This clears the interrupt also.
     */
    Pins = Pio->IntrStatus & Pio->IntrMask;
    Pins = _bs32(Pins);

    /* sanity check
     */
    if (Pins == 0) {
        *pNotMyInterrupt = TRUE;
        return FALSE;
    }

    /* There might be more than one pending, dispatch all we got
     */
    rv = FALSE;

    if (Pins & PIOB_PB3) {
        rv = ButtonIsr((void*)0,pNotMyInterrupt);
    }
    if (Pins & PIOB_PB4) {
        rv |= ButtonIsr((void*)1,pNotMyInterrupt);
    }
    if (Pins & PIOB_PB5) {
        rv |= ButtonIsr((void*)2,pNotMyInterrupt);
    }

    return rv;
}

/* This thread does all the non-ISR work
 * BUGBUG Use constraints to keep latency bounded.
 */
void THREAD_LINKAGE ButtonThread(THREAD_ARGUMENT arg)
{
    struct _BUTTONS *Butts = (struct _BUTTONS *)arg;
    UINT LastSeen = 0, ButtonNo;
    UINT32 Value32;
    TIME_CONSTRAINT TimeConstraint;

    _tprintf(_TEXT("ButtonThread started (%x)\r\n"),(ADDRESS)Butts);
    
    /* Construct a very urgent TimeConstraint and do an
     * initial BeginConstraint
     */
    Int32ToInt64(TimeConstraint.Start, 0);/* can start immediately after wait*/
    Int32ToInt64(TimeConstraint.Estimate, 0); /* does not really matter */
    Int32ToInt64(TimeConstraint.Deadline,0); /* immediately if not sooner */
    TimeConstraint.Criticality = TRUE;
    (void) BeginConstraint(FALSE, &TimeConstraint, NULL);

    for (;!TimeToDie;) {

        Int32ToInt64(TimeConstraint.Start, 0);
        Int32ToInt64(TimeConstraint.Estimate, 0);
        Int32ToInt64(TimeConstraint.Deadline,0);
        TimeConstraint.Criticality = TRUE;

        (void) Condition_WaitAndBeginConstraint(&Butts->ButtonPress,
                                                NULL, NULL,
                                                TRUE, &TimeConstraint, NULL);

        /* Pick up what happened */
        do {
            LastSeen = Butts->ButtonsChanged;
        } while (!AtomicCmpAndSwap(&Butts->ButtonsChanged,
                                   LastSeen,
                                   0));
        /* Anything else.. later */

        Value32 = PioB->PinDataStatus;
        Value32 = _bs32(Value32);
        printf("Pio: %x\r\n", Value32);
        for (ButtonNo = 0; ButtonNo < nBUTTONS; ButtonNo++) {
            if (LastSeen & (1<<ButtonNo)) {
                TIME tOn, tOff;
                tOn = Butts->ButtonState[ButtonNo].LastOnAt;
                tOff = Butts->ButtonState[ButtonNo].LastOffAt;
                if (Butts->ButtonState[ButtonNo].On) {
                    _tprintf(_TEXT("Button%d press at %lld00 (%d,%d)"),
                             1+ButtonNo,tOn,nInterrupts,nDebounced);
                    setled(1+ButtonNo,TRUE);
                } else {
                    _tprintf(_TEXT("Button%d release at %lld00 (%d,%d)"),
                             1+ButtonNo,tOff,nInterrupts,nDebounced);
                    setled(1+ButtonNo,FALSE);
                }
                printf(" [%lld %lld %lld]\r\n",t1,t2,t3);
            }
        }

        /* On release of this button terminate the program */
#define DIE_BUTTON (3-1)
        if ((LastSeen & (1<<DIE_BUTTON)) &&
            (!Butts->ButtonState[DIE_BUTTON].On))
            TimeToDie = TRUE;
    }

    /* Disable button interrupts */
    RemoveDevice((PTR)3,31);
    PioB->IntrDisable = _bs32(PIOB_PB3 | PIOB_PB4 | PIOB_PB5);
    RemoveDevice((PTR)PioB,14);

    _tprintf(_TEXT("ButtonThread terminated (%x)\r\n"),(ADDRESS)Butts);
}

/* Main driver program
 */
int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    PIPROCESS pPrc = CurrentProcess();

    UnusedParameter(argc);
    UnusedParameter(argv);

    _tprintf(_TEXT("Initializing Button 4\r\n"));
    Condition_Init(&TheButtons.ButtonPress);
    sc = pPrc->v->CreateThread(pPrc,
                               0,
                               ButtonThread,
                               (THREAD_ARGUMENT)&TheButtons,
                               2048, /* use 2k for stack */
                               NULL,
                               NULL);

    PioA->Disable = _bs32(PIOA_IRQ0);
    PioA->FilterEnable = _bs32(PIOA_IRQ0);
    sc = AddDevice ((PTR)3, (PTR) ButtonIsr, 0, 31, NULL);

    _tprintf(_TEXT("Initializing Buttons 1,2,3\r\n"));
    PioB->Disable = _bs32(PIOB_PB3 | PIOB_PB4 | PIOB_PB5);
    PioB->FilterEnable = _bs32(PIOB_PB3 | PIOB_PB4 | PIOB_PB5);
    sc = PioB->IntrStatus; /* clear pending */
    sc = AddDevice ((PTR)PioB, (PTR) PioIsr, DEVICE_FLAGS_LEVEL_TRIGGER, 14, NULL);
    PioB->IntrEnable = _bs32(PIOB_PB3 | PIOB_PB4 | PIOB_PB5);

    /* Let this thread die, the other one has the refcount
     */

    return 0;
}

