/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>

#ifdef _MSC_VER
#pragma warning(disable:4146) /* unary minus..result still unsigned */
#endif

/* 
    Various algorithms optimized for TM1 for finding the leftmost or 
    rightmost 1 bit in a 32-bit word. Each returns the bit number of
    the leftmost or rightmost 1, counting from the least-significant
    bit (bit 0). If no bits are set, each function returns -1.

    Note: None of these are fully tested!
 */

/* Rightmost 1. This is easy, and takes only 8 cycles */
int Rightmost1(UINT32 Val)
{
    UINT32 Mask;
 
    if (Val==0) return -1;

    Mask = Val & -Val;
    return ((Mask & 0xFFFF0000)!=0) << 4
         | ((Mask & 0xFF00FF00)!=0) << 3
         | ((Mask & 0xF0F0F0F0)!=0) << 2
         | ((Mask & 0xCCCCCCCC)!=0) << 1
         | ((Mask & 0xAAAAAAAA)!=0);
}

/* Leftmost 1 algorithms. Much tougher */

/* Brute force method - compiles to 42 cycles */
#define TEST(_val_,_bit_) if ((_val_>>_bit_) & 1) return _bit_
int Leftmost1_1(UINT32 Val)
{
    TEST(Val,31);
    TEST(Val,30);
    TEST(Val,29);
    TEST(Val,28);
    TEST(Val,27);
    TEST(Val,26);
    TEST(Val,25);
    TEST(Val,24);
    TEST(Val,23);
    TEST(Val,22);
    TEST(Val,21);
    TEST(Val,20);
    TEST(Val,19);
    TEST(Val,18);
    TEST(Val,17);
    TEST(Val,16);
    TEST(Val,15);
    TEST(Val,14);
    TEST(Val,13);
    TEST(Val,12);
    TEST(Val,11);
    TEST(Val,10);
    TEST(Val,9);
    TEST(Val,8);
    TEST(Val,7);
    TEST(Val,6);
    TEST(Val,5);
    TEST(Val,4);
    TEST(Val,3);
    TEST(Val,2);
    TEST(Val,1);
    TEST(Val,0);
    return -1;
}

/* Binary search without lookup tables */
/* 21 cycles! */
int Leftmost1_2(UINT32 Val)
{
    int Count;

    if (Val==0) return -1;

    Count=0;
    if (Val & 0xFFFF0000) Count += 16;
    if (Val & 0xFFFF0000) Val >>= 16;

    if (Val & 0x0000FF00) Count += 8;
    if (Val & 0x0000FF00) Val >>= 8;

    if (Val & 0x000000F0) Count += 4;
    if (Val & 0x000000F0) Val >>= 4;

    if (Val & 0x0000000C) Count += 2;
    if (Val & 0x0000000C) Val >>= 2;

    if (Val & 0x00000002) Count += 1;
    return Count;
}

static signed char HighBit[16] =
{ 
    -1, /* 0x00 - no bits set */
    0,  /* 0x01 */
    1,  /* 0x02 */
    1,  /* 0x03 */
    2,  /* 0x04 */
    2,  /* 0x05 */
    2,  /* 0x06 */
    2,  /* 0x07 */
    3,  /* 0x08 */
    3,  /* 0x09 */
    3,  /* 0x0A */
    3,  /* 0x0B */
    3,  /* 0x0C */
    3,  /* 0x0D */
    3,  /* 0x0E */
    3   /* 0x0F */
};

/* 13 cycles using a 16 byte lut, 8 data accesses (but all in cache) */
int Leftmost1_3(UINT32 Val)
{
    int Ret[8];
    Ret[0]=HighBit[(Val    )&0x0F];
    Ret[1]=HighBit[(Val>> 4)&0x0F];
    Ret[2]=HighBit[(Val>> 8)&0x0F];
    Ret[3]=HighBit[(Val>>12)&0x0F];
    Ret[4]=HighBit[(Val>>16)&0x0F];
    Ret[5]=HighBit[(Val>>20)&0x0F];
    Ret[6]=HighBit[(Val>>24)&0x0F];
    Ret[7]=HighBit[ Val>>28      ];

    if ((Val>> 4)&0x0F)  Ret[0] = Ret[1] +  4;
    if ((Val>> 8)&0x0F)  Ret[0] = Ret[2] +  8;
    if ((Val>> 12)&0x0F) Ret[0] = Ret[3] + 12;
    if ((Val>> 16)&0x0F) Ret[0] = Ret[4] + 16;
    if ((Val>> 20)&0x0F) Ret[0] = Ret[5] + 20;
    if ((Val>> 24)&0x0F) Ret[0] = Ret[6] + 24;
    if ( Val>> 28      ) Ret[0] = Ret[7] + 28;
    return Ret[0];
}

/* 12 cycles using a 16 byte lut, 8 data accesses (but all in cache) */
/* This is the best one I think */
int Leftmost1_4(UINT32 Val)
{
    int Ret[8];
    Ret[0]=HighBit[(Val    )&0x0F];
    Ret[1]=HighBit[(Val>> 4)&0x0F];
    Ret[2]=HighBit[(Val>> 8)&0x0F];
    Ret[3]=HighBit[(Val>>12)&0x0F];
    Ret[4]=HighBit[(Val>>16)&0x0F];
    Ret[5]=HighBit[(Val>>20)&0x0F];
    Ret[6]=HighBit[(Val>>24)&0x0F];
    Ret[7]=HighBit[ Val>>28      ];

    if (Val>>  4) Ret[0] = Ret[1] +  4;
    if (Val>>  8) Ret[0] = Ret[2] +  8;
    if (Val>> 12) Ret[0] = Ret[3] + 12;
    if (Val>> 16) Ret[0] = Ret[4] + 16;
    if (Val>> 20) Ret[0] = Ret[5] + 20;
    if (Val>> 24) Ret[0] = Ret[6] + 24;
    if (Val>> 28) return Ret[7] + 28;
    return Ret[0];
}


int _tmain()
{
    return 0;
}
