/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * What is often intended for "Hard Real Time System" is a system
 * composed of a (fixed) set of tasks, scheduled in a statically
 * predefined manner.  Each task is simply a function that is invoked
 * repeatedly, with a certain frequency.  The appeal in this approach
 * is its simplicity, which provides a totally deterministic and
 * easily predictable execution model.
 *
 * This example shows how to build such a scheduling behaviour on top
 * of our primitives.  Added functionality in our system include the
 * ability to run more than one such task-set in parallel (say the
 * CPU doubles in speed) and/or to run one or more background tasks
 * (logging, maintainance, etc etc) as time-sharing threads.
 *
 * The basic idea is trivial: a single thread wakes up every
 * NN milliseconds and executes all tasks, one after the other.
 * A simple extension assigns different frequencies to each task
 * (this subsumes the previous case if all frequencies are identical).
 *
 * There are many possible design options, and/or additions to what
 * this example provides. Here are a few ideas for consideration.
 * - The description of a task typically include its execution time.
 * This information is used to verify that the system is performing
 * normally.  After each task invocation the system time is checked
 * to verify that it completed in time.  This could be done by the
 * individual tasks, or by the task dispatcher.  When the task 
 * takes less time then it was allocated it spins idle.
 * - "Overload" is something that in these systems just cant happen.
 * A simple way to add overload management is to check the time as
 * above, and if/when we run out of time in a given period .. skip
 * the rest of the tasks.  The tasks are prioritized by the order
 * in which they appear in the list. This prioritization is crude,
 * but it will nonetheles suffice in most cases.
 * - A more complicated scheme might assign priorities to tasks,
 * to provide for sets-of-tasks to be skipped during overload.
 * - Multiple levels of fixed-priorities can be implemented by 
 * assigning a separate thread to each priority level, plus a central
 * synchronization thread that coordinates them.
 * - Using more than one thread to dispatch tasks could be used
 * in multiprocessor systems. This could be done either by modifying
 * the dispatcher to use multiple threads, or more simply by running
 * multiple copies of this program (one each with the task-set for
 * one processor).
 */

#include <mmlite.h>
#include <stdio.h>

#define TIME_SLICE 10 /* milliseconds */
typedef void (* TASK)(int);
typedef struct {
    char *Name;
    TASK Code;
    UINT Period; /* In milliseconds */
    UINT Skipto;
} TASK_DESCRIPTION, *PTASK_DESCRIPTION;

/*forward*/
void Task1(int Me), Task2(int Me);

TASK_DESCRIPTION TaskList[] = {
 { "First Task", Task1, 10, },
 { "Second Task", Task2, 20, }
};
#define NTasks ((sizeof TaskList)/(sizeof TaskList[0]))

#define TaskName(_id_) TaskList[_id_].Name

int _tmain()
{
    TIME SleepTime = CurrentTime(), TimeSlice;
    int id;

    Int32ToInt64(TimeSlice,TIME_MILLIS(TIME_SLICE));
    for (;;) {
        SleepTime = TimeAdd(SleepTime,TimeSlice);
        SleepUntil(SleepTime);/* absolute time */

        for (id = 0; id < NTasks; id++) {
            TaskList[id].Skipto += TIME_SLICE;
            if (TaskList[id].Skipto == TaskList[id].Period) {
                TaskList[id].Code(id);
                TaskList[id].Skipto = 0;
            }
        }
    }
}

/* Actual tasks */

void Task1(int Me)
{
    printf("%s runs..\n", TaskName(Me));
}

void Task2(int Me)
{
    printf("%s runs..\n", TaskName(Me));
}
