/*
* EX.C
* Tests IHeap::Extract() for various failures. 
* This will allocate a hunk of memory, then will attempt to extract a hunk 
* at a specified offset up or downstream of the previously allocated block. 
* A tester can pass flags to allocate / extract, and the wise tester will fool 
* around with various kinds of overlaps. 
* BUGBUG: Since we don't know for sure if a particular piece of memory is free, 
* it is hard to know if we should be able to get a piece at a particular address. 
* So, we could in principle ask for a free piece and be denied. Is such a case a bug?
* Only the heap knows for sure. 
* Copyright (c) Microsoft Corporation. All rights reserved.
*/
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "heaputils.h"
#include "..\utilities\utility.h"

void EXStatus(PTR pAlloc, UINT32 nAllocSize, UINT32 nAllocFlags, PTR pExtract, UINT32 nExtractSize, UINT32 nExtractFlags, UINT32 nExtractOffset);

void EXUsage(const _TCHAR * s);

BOOL ShouldExtractOK(PTR pAlloc, INT32 nExtractOffset, UINT32 nAllocSize, UINT32 nExtractSize);

int _tmain(INT argc, _TCHAR **argv)
{
    IHeap * pHeap = 0; 
    PTR pAlloc = 0, pExtract = 0; 
    const _TCHAR * sArg = 0;
    _TCHAR * stop = 0;
    UINT32 nAllocSize = 0x10000, nAllocFlags = 0x0, nExtractSize = 0x8000, nExtractFlags = 0x0; 
    INT32 nExtractOffset = 0;
    ParseCmdLine(&argc, &argv); 
    OptionalBreak(argc, argv); 
    TellAboutOptionalBreak();

    if(argc < 2)
        EXUsage(argv[0]);

    sArg = GetArgFromCommandLine(argc, argv, _T("asize")); 
    if(sArg && *sArg)
        nAllocSize = _tcstoul(sArg, &stop, 10); 

    sArg = GetArgFromCommandLine(argc, argv, _T("xsize")); 
    if(sArg && *sArg)
        nExtractSize = _tcstoul(sArg, &stop, 10); 

    sArg = GetArgFromCommandLine(argc, argv, _T("xflags")); 
    if(sArg && *sArg)
        nExtractFlags = _tcstoul(sArg, &stop, 16); 

    sArg = GetArgFromCommandLine(argc, argv, _T("aflags")); 
    if(sArg && *sArg)
        nAllocFlags = _tcstoul(sArg, &stop, 16); 

    sArg = GetArgFromCommandLine(argc, argv, _T("xoffset")); 
    if(sArg && *sArg)
        nExtractOffset = _tcstol(sArg, &stop, 10); 

    FreeCmdLine(&argc, &argv); 
    
    pHeap = CurrentHeap(); 
    
    //first alloc a hunk with params user passed in (or defaults ..)
    pAlloc = pHeap->v->Alloc(pHeap, nAllocFlags, nAllocSize, 0); 

    //we might want to test without allocating a block first so only fail if memory wanted
    if(!pAlloc && nAllocSize > 0)
    {
        _tprintf(_T("\n\r\aAllocation failed, quitting")); 
        return SetAndReturn(1); 
    }

    //now extract. Basically we are testing overlap - try to extract over an allocated piece
    pExtract = pHeap->v->Extract(pHeap, nExtractSize, (BYTE*)pAlloc + nExtractOffset, nExtractFlags); 
    EXStatus(pAlloc, nAllocSize, nAllocFlags, pExtract, nExtractSize, nExtractFlags, nExtractOffset); 
    
    _tprintf(_T("\n\r[EXTRACT] returned [0x%.8x](%s)"), (UINT32)pExtract, pExtract ? _T("(good)") : _T("(null)")); 

    //We got memory, but did we set to zero if requested?
    if(pExtract && (nExtractFlags & HEAP_ZERO_MEMORY))
        if(!MemoryAllZeroes(pExtract, nExtractSize, 0))
            error(_T("_tmain"), _T(__FILE__), __LINE__, pExtract ? _T("IHeap::Extract() succeeded, should have failed") : _T("IHeap::Extract() returned a pointer, but failed to respect HEAP_ZERO_MEM flag"));


    //is there a disagreement between what we expected and what we got?
    if(XOR(pExtract, ShouldExtractOK(pAlloc, nExtractOffset, nAllocSize, nExtractSize)))
    {
        //should never succeed if failure expected
        if(pExtract)
            error(_T("_tmain"), _T(__FILE__), __LINE__, pExtract ? _T("IHeap::Extract() succeeded, should have failed") : _T("IHeap::Extract() failed, should have succeeded"));
        //we cannot be sure a call MUST succeed ... but we can flag as a possible bug
        else
            anomaly(_T("_tmain"), _T(__FILE__), __LINE__, pExtract ? _T("IHeap::Extract() failed, might have succeeded (maybe not ...)") : _T("IHeap::Extract() failed, should have succeeded"));

        pHeap->v->Free(pHeap, 0x0, pAlloc); 
        return SetAndReturn(1); 
    }
          
    pHeap->v->Free(pHeap, 0x0, pAlloc); 
    
    return SetAndReturn(0); 
}

void EXUsage(const _TCHAR *s)
{
    _tprintf(_T("\n\rUsage:%s -asize <decimal bytes> -xsize <decimal bytes> -xflags <hex flag> -xoffset <[+/-]decimal bytes>"), s); 
    _tprintf(_T("\n\r\tWhere -asize is size of original fragment Allocated, -xsize is size of eXtracted fragment,")); 
    _tprintf(_T("\n\r\t-xflags are flags for eXtraction, and -xoffset is offset into/before allocated fragment to start extracting")); 
}

void EXStatus(PTR pAlloc, UINT32 nAllocSize, UINT32 nAllocFlags, PTR pExtract, UINT32 nExtractSize, UINT32 nExtractFlags, UINT32 nExtractOffset)
{
    _TCHAR sAFlagsDesc[256], sXFlagsDesc[256]; 
    GetFlagsDescription(sXFlagsDesc, sizeof(sXFlagsDesc)/sizeof(sXFlagsDesc[0]), nExtractFlags, EXTRACT);
    GetFlagsDescription(sAFlagsDesc, sizeof(sAFlagsDesc)/sizeof(sAFlagsDesc[0]), nAllocFlags, ALLOCATE);
    
    _tprintf(_T("\n\rOriginal block:[0x%.8x] \n\r\tOriginal Block Size:[0x%.8x],"),(UINT32)pAlloc, nAllocSize);
    _tprintf(_T("\n\r\tOriginal Allocation Flags:[%s], \n\rExtracted Block:[0x%.8x], \n\r\tExtracted Block Size:[0x%.8x], \n\r\tExtract Flags:[%s], \n\r\tExtract Offset:[%d]"), 
         sAFlagsDesc, (UINT32)pExtract, nExtractSize, sXFlagsDesc, nExtractOffset);
}

/*
* ShouldExtractOK() tries to predict if an extraction should fail based on overlap. 
* This can return false positives (return TRUE when extraction fails) because the 
* heap has more stringent requirements for allocation than we know about. 
* However, every time this function returns FALSE and extraction SUCCEEDS we have a bug. 
*/
BOOL ShouldExtractOK(PTR pAlloc, INT32 nExtractOffset, UINT32 nAllocSize, UINT32 nExtractSize)
{
    BOOL b0, b1;

    //no overlap .. should test corner cases better here ... 
    //but null ptr should never overlap anything, no?
    if(!pAlloc) return TRUE; 

    //do we overlap allocated block?
    //BOOLEANs used to make this easier to follow in the debugger ... 
    //..don't need sign problems here ... 
    b0 = (INT32)(nExtractOffset + (INT32)nExtractSize) < 0; //extract block misses left end of alloc
    b1 = (INT32)(nExtractOffset + (INT32)nExtractSize) < (INT32)nAllocSize; //misses right end ... 
    if(b0 && b1)
        return TRUE; 

    return FALSE;
}
