/*
* Heaputils.h - contains some generic heap testing utility function prototypes. 
* Copyright (c) 2002 Microsoft Corp. All rights reserved. 
* a-ericf
*/

#include "heaputils.h"
/*
* GetAlignment() - Gets (likely) alignment of memory. This function 
* can return an alignment LARGER than actual. For example, if memory 
* is aligned on a 2 - byte boundary, it might (by accident) also be 
* aligned on an 8-byte boundary. This fxn will return 8 in such 
* a case ... it will return the LARGEST candidate alignment boundary 
* ... returns -1 if fails to find alignment (!?!)
* BUGBUG : only good up to 128K alignment bounds
* BUGBUG : not efficiently coded but this is test code no?
*/
UINT GetAlignment(const void * p)
{
    ADDR_SIZE idx; 
    for(idx = 1; idx < ((ADDR_SIZE)32 * 4096); idx *= 2)
        if(((UINT)p % idx) != 0)
            return idx / 2;

    return ~0;
}


/*
* MemoryAllZeroes()
* Just test to see if ZERO_MEM flag worked ok
* Returns T/F if mem is /is not all 0s
* Returns T on NULL pointer or 0 len input
*/
BOOL MemoryAllZeroes(const PTR pMem, UINT nLenOverall, UINT nOffset)
{
    char * p = (char *)pMem; 
    unsigned idx;
    //not interested in null ptrs, so not an error case
    if(!p) return TRUE; 
    for(idx = nOffset; idx < nLenOverall; ++idx)
        if(p[idx]) return FALSE; 

        return TRUE; 
}

/*
* StuffMemory()
* For testing if memory is corrupted. 
* We skip the first byte, and stuff the rest with psuedo random junk values. 
* We calc a chksum byte and poke it into byte 0
* Later we can test the chksum byte with MemoryNotThrashed(), below
*/

void StuffMemory(PTR pMem, UINT nLen)
{
    UINT8 c;
    UINT8 * p = (UINT8 *)pMem;
    unsigned idx;

    //seed c
    TIME t = CurrentTime(); 
    c = (UINT8)Int64ToInt32(t);
        
    if(!nLen) return; 

    if(!p) return; 
    
    *p = 0; 
   
    //trap special case where we only have enuf memory for the ck byte anyway .. 
    // *p is already 0 which is correct cksum
    if(nLen == 1) 
        return; 
    

    for(idx = 1; idx < nLen; ++idx)
    {
        c ^= p[idx - 1]; //just generate a few weird values is all ... 
        p[0] ^= p[idx] = c; //cksum goes into byte 0
    }
}

/*
* MemoryNotThrashed()
* Test the ck byte for corruption. See StuffMemory() for description. 
* This function tests the check byte in a memory block set up by StuffMemory()
* 
*/
BOOL MemoryNotThrashed(const PTR pMem, UINT nLen)
{
    UINT8 * p = (UINT8 *)pMem;
    UINT8 c;
    unsigned idx;
    
    //NULL ptr is always OK
    if(!p) 
        return TRUE;
    
    //zero len is always OK
    if(!nLen) 
        return TRUE; 
    
    //len == 1 means 1st byte must be 0
    if(nLen == 1) 
        return *p ? FALSE : TRUE; 
    
    for(idx = 1, c = 0; idx < nLen; ++idx)
        c ^= p[idx]; 

    return (c == p[0]) ? TRUE : FALSE; 
}

/*
* anomaly()
* Report anomaly to user. Does not stop script, but flags situation
* for user attention. 
* Prints file name, line, and brief message onscreen. 
*/
void anomaly(const _TCHAR *sFunc, const _TCHAR * sFile, UINT nLine, const _TCHAR * szTxt)
{
    _TCHAR szOut[256]; 
    _stprintf(szOut, _T("\n\rAnomaly (perhaps a bug) in file %s, function %s, line %d, reason:%s\n\r"),  sFile, sFunc, nLine, szTxt);
    _tprintf(szOut); 
}
 
/*
* error()
* Report error to user. Does not stop script, but flags situation
* for user attention. 
* Prints file name, line, and brief message onscreen. 
*/       
void error(const _TCHAR *sFunc, const _TCHAR * sFile, UINT nLine, const _TCHAR * szTxt)
{
    _TCHAR szOut[256]; 
    _stprintf(szOut, _T("\n\rERROR in file %s, function %s, line %d, reason:%s"),  sFile, sFunc, nLine, szTxt);
    _tprintf(szOut); 
}

/*
* params() - Reports parameters used in heapwise operation. 
* Used to generate visible debug spew for user. 
* Does not imply an error, just provides visible feedback for easy debugging. 
* Preconditions: 
*   - pHeap points to heap
*   - pAllocBlock points to (re)allocated block
*   - nFlags are flags used in (re)allocation
*   - nSize and nAllocAlign are flags and alignment resp. 
*/
void params(const IHeap * pHeap, const void * pAllocBlock, UINT nFlags, UINT nSize, UINT nAllocAlign)
{
    _TCHAR szTxt[256]; 
    _stprintf(szTxt, _T("\n\rParams:\r\nHeap:0x%x Ptr:0x%x Flags:0x%x Size:%d Alignment: %d\n\r"), (unsigned)pHeap, (unsigned)pAllocBlock, nFlags, nSize, nAllocAlign); 
    _tprintf(szTxt);
}

/*
* CheckSize()
* Checks for correctness of allocation given a particular set of sizes. 
* Returns TRUE if allocation is expected to succeed. 
* This test is very permissive since there is no way to know size of a 
* particular block
* Given all the uncertainties of forecasting successful allocation based
* on sizes, perhaps this function should be deprecated and forgotten .. 
*/
BOOL CheckSize(IHeap * p, UINT nSize, UINT nAlign)
{
    UINT nReallyNeeded = nSize + (nAlign - (nSize % nAlign));
    ADDR_SIZE reserved, committed, used, maxused;
    
    if(p->v->Status(p, &reserved, &committed, &used, &maxused) != S_OK)
    {
        _TCHAR szTxt[32]; 
        _stprintf(szTxt, _T("\n\rError calling Status() function")); 
        error(_T("CheckSize()"), _T(__FILE__), __LINE__, szTxt);
        DebugBreak();
        return FALSE; 
    }
    
    nReallyNeeded = nSize + (nAlign - (nSize % nAlign));

   if(nReallyNeeded <= PhysicalMemoryAvailable())
        return TRUE;
        
    return FALSE;
}
        
/*
* CheckAlignment()
* Returns FALSE if allocation is expected to fail, based on alignment
* Tests to see if flags and aligment conflict in obvious ways
*/
BOOL CheckAlignment(const PTR pBlock, UINT nFlags, UINT nAlign)
{
    ADDR_SIZE idx; 
    //step through all legal alignment sizes
    //is align size legal?
    for(idx = ((ADDR_SIZE)4096 * 32); idx; idx /= 2)
        if(nAlign == idx) 
            break; 
    
    if(nAlign != idx)
        return FALSE; 

    //alignsize does not line up on cache boundaries
    if(_DCACHELINESIZE % nAlign)
        return FALSE;

    //Block allocced with cache align does not align on cache bounds
    if(nFlags & HEAP_CACHE_ALIGN)
        if(((unsigned)pBlock) % _DCACHELINESIZE)
            return FALSE;
    
    return TRUE;
}

/*
* Returns physical mem available. Not the same as largest block allocatable
*/
UINT PhysicalMemoryAvailable(void)
{
    struct MACHINE_INFO mach;
    PIPROCESS pPrc = CurrentProcess(); 

    SCODE  hr = pPrc->v->MachineInfo(pPrc, MACHINE_KIND_GENERIC, (struct MACHINE_INFO *)&mach); 
 
    if(SUCCEEDED(hr))
        return mach.AvailablePhysicalMemory; 
    return -1; 
}

