/*
* Prime_realloc_and_free.c 
* Omnibus test suite for heap functions - Alloc(), Realloc(), Free(), Validate(), and Status()
* Most test suites are designed with only one or two functions in mind, but the heap functions
* are interdependent. For example, reallocation success/failure depends on (among other things) 
* the alignment and flags specified during the original allocation of the block in question. 
* It is possible to specify during allocation that a block is not to be moved, and during reallocation
* to specify an alignment setting that would *require* a move. Such a reallocation must fail. 
* So, with heavy heart, I have stuck all these functions together here in a hugely nested loop ...
* Copyright (c) Microsoft Corporation All rights reserved
*/
#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include "heaptest.h"
#include "..\utilities\utility.h"

//We have a stock of smallish and primish numbers to use as allocation sizes
//yes, 1 is not prime, but since they are multiplied, we want the primes themselves in there too
//there are few nonprime nos in here .. fiddling around with boundaries etc
int aPrimes[] = {1, 13, 17, 2048, 4097, 0x10000};

const unsigned  NUM_ENTRIES                 = sizeof(aPrimes)/sizeof(aPrimes[0]); 
const UINT      DEFAULT_ALIGNMENT           = 8; 
const UINT      HEAPSIZEPAD                 = HEAPSIZE/0x100; 

#define HEAP_BOGUS_FLAG  (~(HEAP_CACHE_ALIGN | HEAP_NO_COPY | HEAP_REALLOC_IN_PLACE_ONLY | HEAP_ZERO_MEMORY))


UINT aFlags[] = {
    HEAP_CACHE_ALIGN, 
    HEAP_NO_COPY, 
    HEAP_REALLOC_IN_PLACE_ONLY, 
    HEAP_ZERO_MEMORY, 
    HEAP_BOGUS_FLAG
};


/*
Test to make sure memory allocation does not have mem leaks by allocating and freeing odd (prime) sized 
blocks and testing available memory
Test for different alignments by changing ALIGNMENT const above
Also tests allocation and borrows a lot from alloc() test.

todo: probably have taken the brute force all at once approach a little too far with this suite.
todo: what will instantiation of different types of heaps do to this code? Need to test FirstFit Bitmap and Null heaps. 
todo: Need to pass different heap sizes in on cmd line
a-ericf 11-15-01
*/

BOOL ReallocationShouldSucceed(IHeap *pHeap, const PTR pAllocBlock, const PTR pReallocBlock, UINT nAllocSize, UINT nReallocSize, UINT nAllocFlags, UINT nReallocFlags, UINT nAllocAlign, UINT nReallocAlign);
BOOL AllocationShouldSucceed(IHeap *, const PTR, UINT, UINT, UINT);

BOOL CheckReallocAlignment(const PTR pAllocMem, const PTR pReallocMem, UINT nAllocFlags, UINT nReallocFlags, UINT nAllocAlign, UINT nReallocAlign);
BOOL CheckAllocFlags(UINT, UINT);
BOOL CheckReallocFlags(UINT nAllocFlags, UINT nReallocFlags);
BOOL Usage(int, _TCHAR **);

int _tmain(INT argc, _TCHAR **argv)
{
    PIHEAPFACTORY HeapFactory;
    int nFailures;
    IHeap * pHeap;
    PTR pHeapBlock;
    short fSizeFactor; 
    SCODE sc;

    //allocation flags - use several vars for different combos
    //terminology - n means iNteger, a or r means Alloc or Realloc, 
    //i, j, k are for nested loops so the flags can be OR'd together in combination
    UINT naFlagsi, naFlagsj, naFlagsk, nAllocAlign, n, q;
    UINT nrFlagsi, nrFlagsj, nrFlagsk, nReallocAlign;
    unsigned idx = 0;
    
    //_tprintf(_T("\n\rTesting with %s heap"), HEAPTYPE);
    //ParseCmdLine(&argc, &argv);
    //XXX need to add code to get ptr to a particular type of heap ... 

    pHeapBlock = CurrentHeap()->v->Alloc(CurrentHeap(), 0, HEAPSIZE, DEFAULT_ALIGNMENT); 
    if(!pHeapBlock)
    {
        _tprintf(_T("\n\rERROR: failed to get a heap block")); 
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1);
    }

    sc = BindToObject(CurrentNameSpace(), _T("COB/ff_heap.cob"),
                      NAME_SPACE_READ, &IID_IHeap,
                      (void**) &HeapFactory);
    if(FAILED(sc))
    {
        _tprintf(_T("\n\rERROR: failed to bind to COB/ff_heap.cob")); 
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1);
    }

    sc = HeapFactory->v->Create(HeapFactory,
                                (ADDRESS)pHeapBlock, 0, HEAPSIZE/2,
                                HEAPSIZE - HEAPSIZEPAD, &pHeap);
    HeapFactory->v->Release(HeapFactory);
    if(FAILED(sc)) {
        _tprintf(_T("\n\rERROR: cannot get that type of heap"));
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1); 
    }
    
    //pHeap = ref cryp1.c for bindtoobject 
    /*
    Brute force testing
    Idea here is to test realloc(). Since realloc()'s success depends in part on what happened during alloc(), 
    we are testing both together. Roughly it goes like this:
        for(a combination of sizes of alloc'd memory [must be different from realloc sizes])
            for(realloc sizes from zero to [larger than alloc'd])
                for(each possible flag combination, for both alloc and realloc)
                    for(each possible alignment combination for alloc and realloc)
                        {
                            ptr = alloc(alloc_parms); 
                            bool bOK = AllocationShouldSucceedWithTheseParms(alloc_parms) 
                            if( ptr XOR bOK) //either both should be true, or neither
                                error();

                            realloc_ptr = realloc(ptr, realloc_parms);
                            bOK = AllocationShouldSucceedWithTheseParms(realloc_parms) 
                            if( realloc_ptr XOR bOK)
                                error(); 

                            clean up pointers; 
                            continue;
                        }
    */
    _tprintf(_T("\n\rNo 0-len reallocs for now"));
    //yeeeee-haaaw! ten deep! a personal record :-o)
    nFailures = 0;
    for(n = 0; n < sizeof(aPrimes)/sizeof(aPrimes[0]); ++n)
        for(q = 0; q < sizeof(aPrimes)/sizeof(aPrimes[0]); ++q)
            for(fSizeFactor = 0; fSizeFactor < 225; fSizeFactor += 75) //check reallocation of blocks of zero, same size, larger 
                for(naFlagsi = 0; naFlagsi < sizeof(aFlags)/sizeof(aFlags[0]); ++naFlagsi) //force every different combination of flags for ALLOCATION
                    for(naFlagsj = naFlagsi; naFlagsj < sizeof(aFlags)/sizeof(aFlags[0]); ++naFlagsj)
                        for(naFlagsk = naFlagsj; naFlagsk < sizeof(aFlags)/sizeof(aFlags[0]); ++naFlagsk)
                            for(nrFlagsi = 0; nrFlagsi < sizeof(aFlags)/sizeof(aFlags[0]); ++nrFlagsi) //force flags for RE-allocation
                                for(nrFlagsj = nrFlagsi; nrFlagsj < sizeof(aFlags)/sizeof(aFlags[0]); ++nrFlagsj)
                                    for(nrFlagsk = nrFlagsj; nrFlagsk < sizeof(aFlags)/sizeof(aFlags[0]); ++nrFlagsk)
                                        for(nAllocAlign = 1; nAllocAlign < 32; nAllocAlign *= 2) //fiddle around with alignment a little bit .. 
                                            for(nReallocAlign = 1; nReallocAlign < 32; nReallocAlign *= 2)
                                                {
                                                    //allocated and reallocated blocks, respectively
                                                    char * pReallocBlock = 0; 
                                                    char * pAllocBlock = 0;

                                                    //sizes for call to Status() before and after allocations
                                                    ADDR_SIZE nStatusMaxUsed0, nStatusMaxUsed1, nStatusUsed0, nStatusUsed1;
                                                    
                                                    BOOL bAllocationShouldSucceed;
                                                    
                                                    UINT nAllocSize = aPrimes[n] * aPrimes[q];
                                                    
                                                    //we set this later using fSizeFactor
                                                    UINT nReallocSize;
                                                    //two sets of flags, one for alloc, one for realloc, to get all 
                                                    //possible combinations
                                                    UINT naFlags = aFlags[naFlagsi] | aFlags[naFlagsj] | aFlags[naFlagsk];
                                                    UINT nrFlags = aFlags[nrFlagsi] | aFlags[nrFlagsj] | aFlags[nrFlagsk];
                                                    
                                                    ++idx;

                                                    sc = pHeap->v->Status(pHeap, 0, 0, &nStatusUsed0, &nStatusMaxUsed0);
                                                    Outcome(STATUS, sc, _T(__FILE__), 0, UINT64_ZERO);
                                                    if(FAILED(sc))
                                                        error(_T("\n\r_tmain()"), _T(__FILE__), __LINE__, _T("Status failed - why?")); 
                                                
                                                    bAllocationShouldSucceed = AllocationShouldSucceed(pHeap, 0, nAllocSize, naFlags, nAllocAlign);
                                                    pAllocBlock = pHeap->v->Alloc(pHeap, naFlags, nAllocSize, nAllocAlign);

                                                    sc = pHeap->v->Validate(pHeap, 0, pAllocBlock); 
                                                    Outcome(VALIDATE, sc, _T(__FILE__), 0, UINT64_ZERO);
                                                    if(FAILED(sc))
                                                    {
                                                        _TCHAR szTxt[64]; 
                                                        _stprintf(szTxt, _T("Validate() failed"));
                                                        error(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                        params(pHeap, pAllocBlock, naFlags, nAllocSize, nAllocAlign);
                                                    }

                                                    sc = pHeap->v->Status(pHeap, 0, 0, &nStatusUsed1, &nStatusMaxUsed1);
                                                    Outcome(STATUS, sc, _T(__FILE__), 0, UINT64_ZERO);
                                                        error(_T("\n\r_tmain()"), _T(__FILE__), __LINE__, _T("Status failed - why?")); 

                                                    if((nStatusUsed1 - nStatusUsed0) < nAllocSize)
                                                    {
                                                        _TCHAR szTxt[64]; 
                                                        _stprintf(szTxt, _T("Allocation succeeded, but size consumed < size requested"));
                                                        anomaly(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                        params(pHeap, pAllocBlock, naFlags, nAllocSize, nAllocAlign);
                                                    }
                                                    
                                                    //make sure mem-zero was respected
                                                    if(naFlags & HEAP_ZERO_MEMORY)
                                                    {
                                                        if(!MemoryAllZeroes(pAllocBlock, nAllocSize, 0))
                                                        {
                                                            _TCHAR szTxt[64]; 
                                                            _stprintf(szTxt, _T("Zero-init memory had nonzero data"));
                                                            error(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                            params(pHeap, pAllocBlock, naFlags, nAllocSize, nAllocAlign);
                                                            ++nFailures;
                                                        }
                                                        else 
                                                            ;
                                                    }

                                                    
                                                    //poke a bunch of crud in and checksum to test for corruption later
                                                    //test for null ptr OK
                                                    StuffMemory(pAllocBlock, nAllocSize);


                                                    //either both passed, or neither ... otherwise an anomaly
                                                    if(XOR(bAllocationShouldSucceed, (BOOL)pAllocBlock))
                                                    {
                                                        ++nFailures;
                                                        _tprintf(_T("\n\rIteration:[%d], Failures:[%d]"), idx, nFailures);
                                                        
                                                        if(!pAllocBlock)  //this means allocation should have succeeded ... 
                                                        {
                                                            _TCHAR szTxt[64]; 
                                                            _stprintf(szTxt, _T("Expected success, got a NULL!"));
                                                            error(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                            params(pHeap, pAllocBlock, naFlags, nAllocSize, nAllocAlign);
                                                            DebugBreak();
                                                        }
                                                        else
                                                        {
                                                            _TCHAR szTxt[64]; 
                                                            _stprintf(szTxt, _T("Expected failure, got a pointer!"));
                                                            error(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                            params(pHeap, pAllocBlock, naFlags, nAllocSize, nAllocAlign);
                                                            if(!CheckSize(pHeap, nAllocSize, nAllocAlign))
                                                                _tprintf(_T("\n\rCheckSize() expected to fail!")); 

                                                            if(!CheckAlignment(pAllocBlock,naFlags, nAllocAlign))
                                                            {
                                                                _tprintf(_T("\n\rCheckAlignment() expected to fail!")); 

                                                                if(!CheckAllocFlags(naFlags, nAllocAlign))
                                                                    _tprintf(_T("\n\rCheckAllocFlags() expected to fail!")); 
                                                                else
                                                                    _tprintf(_T("\n\rFalse alarm from CheckAlignment()")); 
                                                            }
                                                            else
                                                                //Alignment was OK test flags in their own right
                                                                if(!CheckAllocFlags(naFlags, nAllocAlign))
                                                                    _tprintf(_T("\n\rCheckAllocFlags() expected to fail!")); 
                                                                
                                                        }
                                                    }
                                                    else
                                                        ;

                                                

                                                    //this basically means we have realloc sizes of 0, less than original, greater than original
                                                    nReallocSize = (nAllocSize * (fSizeFactor/50));
                                                    _tprintf(_T("\n\r[%d]:%d:%d"), idx, nAllocSize, nReallocSize);

                                                    //we are skipping 0-len reallocs for now 
                                                    if(nReallocSize)
                                                        pReallocBlock = pHeap->v->Realloc(pHeap, nrFlags, pAllocBlock, nReallocSize, nReallocAlign); 
                                                    else
                                                    {
                                                        //_tprintf(_T("\n\rNo 0-len reallocs"));
                                                        if(pAllocBlock)
                                                            pHeap->v->Free(pHeap, 0, pAllocBlock);
                                                        continue; 
                                                    }

                                                    sc = pHeap->v->Validate(pHeap, 0, pReallocBlock); 
                                                    Outcome(VALIDATE, sc, _T(__FILE__), 0, UINT64_ZERO);
                                                    if(FAILED(sc))
                                                    {
                                                        _TCHAR szTxt[64]; 
                                                        _stprintf(szTxt, _T("Validate() failed"));
                                                        error(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                        params(pHeap, pReallocBlock, naFlags, nAllocSize, nAllocAlign);
                                                        DebugBreak();
                                                    }
 
                                                    //now we look to see if it toasted mem that should have been 0-init
                                                    //BUGBUG this test produces false positives ... out for now
                                                    /*
                                                    if(nrFlags & HEAP_ZERO_MEMORY)
                                                        if(!MemoryAllZeroes(pReallocBlock, nReallocSize, nAllocSize))
                                                         {
                                                            _TCHAR szTxt[64]; 
                                                            _stprintf(szTxt, _T("\n\rERROR:Zero-init, reallocated memory had nonzero data"));
                                                            //error(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                            params(pHeap, pReallocBlock, nrFlags, nReallocSize, nReallocAlign);
                                                            ++nFailures;
                                                            
                                                        }
                                                        */

                                                    if(!(nrFlags & HEAP_NO_COPY))
                                                        if(!MemoryNotThrashed(pReallocBlock, nAllocSize))
                                                        {
                                                            _TCHAR szTxt[64]; 
                                                            _stprintf(szTxt, _T("Memory was corrupted during reallocation"));
                                                            error(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                            params(pHeap, pReallocBlock, nrFlags, nReallocSize, nReallocAlign);
                                                            ++nFailures;
                                                            
                                                        }
                                                    
                                                    bAllocationShouldSucceed = ReallocationShouldSucceed(pHeap, pAllocBlock, pReallocBlock, nAllocSize, nReallocSize, naFlags, nrFlags,  nAllocAlign, nReallocAlign); 

                                                    if(XOR(bAllocationShouldSucceed, (BOOL)pReallocBlock))
                                                    {
                                                        if(!pReallocBlock)  //this means reallocation should have succeeded ... 
                                                        {
                                                            _TCHAR szTxt[64]; 
                                                            _stprintf(szTxt, _T("Expected successful reallocation, got a NULL!"));
                                                            error(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                            params(pHeap, pReallocBlock, nrFlags, nReallocSize, nReallocAlign);
                                                            ++nFailures;
                                                            _tprintf(_T("\n\rIteration:[%d], Failures:[%d]"), idx, nFailures);
                                                        }
                                                        //realloc should have failed, but succeeded ... why?
                                                        else
                                                        {
                                                            //maybe size was broken
                                                            if(!CheckSize(pHeap, nReallocSize, nReallocAlign))
                                                            {
                                                                _TCHAR szTxt[64]; 
                                                                _stprintf(szTxt, _T("CheckSize() expected to fail, succeeded"));
                                                                anomaly(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                                params(pHeap, pReallocBlock, naFlags, nAllocSize, nAllocAlign);
                                                                ++nFailures;
                                                                _tprintf(_T("\n\rIteration:[%d], Failures:[%d]"), idx, nFailures);
                                                            }  
                                                            else
                                                                ;
    
                                                            //make sure : flags are good? 
                                                            if(!CheckReallocFlags(naFlags, nrFlags))
                                                            {
                                                                //some 'bad' flags combos can really be good, depends on alignment ... 
                                                                if(!CheckReallocAlignment(pAllocBlock, pReallocBlock, naFlags, nrFlags, nAllocAlign, nReallocAlign))
                                                                {
                                                                    _TCHAR szTxt[64]; 
                                                                    _stprintf(szTxt, _T("\n\rCheckReallocFlags() expected to fail, succeeded"));
                                                                    error(_T("_tmain()"), _T(__FILE__), __LINE__, szTxt);
                                                                    params(pHeap, pReallocBlock, naFlags, nAllocSize, nAllocAlign);
                                                                    ++nFailures;
                                                                    _tprintf(_T("\n\rIteration:[%d], Failures:[%d]"), idx, nFailures);
                                                                }
                                                                else
                                                                    ;
                                                            }
                                                            else
                                                                ;
                                                        }           
                                                    }
                                                    if(pAllocBlock)
                                                        pHeap->v->Free(pHeap, 0, pAllocBlock);
                                                }
    _tprintf(_T("\n\rTotal Iterations:[%d], Failures:[%d]"), idx++, nFailures);

    FreeCmdLine(&argc, &argv);

    pHeap->v->Release(pHeap); 
                                                    
    return SetAndReturn(0);;
}

/*
Returns either true or false depending on whether we think allocation should be OK, XOR this with the actual result to 
determine if there is a gap between what we expect and what we get. 
*/
BOOL AllocationShouldSucceed(IHeap *pHeap, const PTR pAllocBlock, UINT nSize, UINT nFlags, UINT nAlign)
{
     //should we flunk this allocation based on size?
    if(CheckSize(pHeap, nSize, nAlign))
        //size is OK, what about alignment?
        if(CheckAlignment(pAllocBlock, nFlags, nAlign))
            //alignment is OK, what about flags?
            if(CheckAllocFlags(nFlags, nAlign))
                return TRUE;   //expect allocation to succeed... 
            else
                return FALSE;   //fail because flags no good ... 
        else
            return FALSE;    //alignment no good, expect to fail 
    else
        return FALSE;    //size is no good ...
}

/*
Stub : see AllocationShouldSucceed() above
*/
BOOL ReallocationShouldSucceed(IHeap *pHeap, const PTR pAllocBlock, const PTR pReallocBlock, UINT nAllocSize, UINT nReallocSize, UINT nAllocFlags, UINT nReallocFlags, UINT nAllocAlign, UINT nReallocAlign)
{
    if(!pAllocBlock) return FALSE;

    if(CheckSize(pHeap, nReallocSize, nReallocAlign))
        if(CheckReallocAlignment(pAllocBlock, pReallocBlock, nAllocFlags, nReallocFlags, nAllocAlign, nReallocAlign))
            if(CheckReallocFlags(nAllocFlags, nReallocFlags))
                return TRUE; 
            else
                return FALSE; 
        else
            return FALSE; 
    else
        return FALSE;
}

BOOL CheckReallocAlignment(const PTR pAllocMem, const PTR pReallocMem, UINT nAllocFlags, UINT nReallocFlags, UINT nAllocAlign, UINT nReallocAlign)
{
    //was not realloc'd in place ... 
    if((nReallocFlags & HEAP_REALLOC_IN_PLACE_ONLY) && (pAllocMem != pReallocMem))
        return FALSE; 

    //realloc and alloc flags differ on cache align
    if((nReallocFlags & HEAP_CACHE_ALIGN) ^ (nAllocFlags & HEAP_CACHE_ALIGN))
        //wanted cache align on realloc
        if(nReallocFlags & HEAP_CACHE_ALIGN)
            if((unsigned)pAllocMem%_DCACHELINESIZE)
                return FALSE;
            else
            ; 
        else
            ; //null
    else
    {
        //consensus on cache alignment between alloc and realloc, what was it? 
        if(nAllocFlags & HEAP_CACHE_ALIGN)  //if ONE was set, BOTH were at this point
        {
            //did we really get cache-aligned memory?
            if((unsigned)pReallocMem % _DCACHELINESIZE)
                return FALSE; 
        }
        else
        {
            //neither was set ... did original pointer have nice allocation size?
            if((unsigned)pAllocMem%nReallocAlign)
                return FALSE;
            else
                ;
                
        }
           
    }

    //we care about this because it might not be possible to allocate 
    //with proper alignment AND without moving memory
    if(nAllocAlign < nReallocAlign)
    {
        //we said realloc in place but with bigger alignment ... 
        //if the original memory was aligned according to the realloc request
        //this should pass else fail
        if(nReallocFlags & HEAP_REALLOC_IN_PLACE_ONLY)
            if((unsigned)pAllocMem%nReallocAlign)
                return FALSE; 
    }
    
    //check against vanilla allocation settings
    if(!CheckAlignment(pReallocMem, nReallocFlags, nReallocAlign))
        return FALSE;
    
    return TRUE; //can't think of any reason to flunk it .. 
}

BOOL CheckReallocFlags(UINT nAllocFlags, UINT nReallocFlags)
{
    if(nReallocFlags & HEAP_CACHE_ALIGN)
        if(!(nAllocFlags & HEAP_CACHE_ALIGN))
            return FALSE; 

    //everything else should be OK -- ?
   
    return TRUE; 
}


/*
Basically a stub now in case jvh changes any flags ... always returns TRUE
*/
BOOL CheckAllocFlags(UINT nFlags, UINT nAlign)
{
    if(nFlags & HEAP_ZERO_MEMORY)
    {
        ; //all combinations are OK with this flag
    }

    if(nFlags & HEAP_CACHE_ALIGN)
    {
        ; //all combinations are OK with this flag
    }

    if(nFlags & HEAP_NO_COPY)
    {
        ; //all combinations are OK with this flag
    }

    if(nFlags & HEAP_REALLOC_IN_PLACE_ONLY)
    {
        ; //all combinations are OK with this flag
    }
    return TRUE; 
}


UINT Usage(int argc, _TCHAR **argv)
{
    if(argc != 2)
    {
        _tprintf(_T("\n\rUsage:<prog.exe> [NULL | FIRSTFIT | BITMAP] \n\r. Leaving out heap type gets you the default heap.")); 
        return 0; 
    }

    else
    {
        _tprintf(_T("\n\rGonna see if I can get a %s heap"), argv[1]); 
        return argc - 1; 
    }
}

