/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Trivia tests for a palette-based framebuffer */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <framebuffer.h>

UINT8 *FB;
UINT32 Hlen, Hvis, Vvis;

SCODE Initialize(PIFRAMEBUFFER FrameBuffer, _TCHAR *Name)
{
    SCODE sc;
    UINT32 Length;
    FRAMEBUFFER_MODE Mode;
    _TCHAR *Msg;

    FB = NULL;
    Length = 0;

    Msg = _T("get mode");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_MODE,&Mode);
    if (FAILED(sc) || 
        (Mode != FB_MODE_640x480_in_768x480_8bpp_PALETTE)) {
        Msg = _T("set mode");
        sc = FrameBuffer->v->SetMode(FrameBuffer,FB_MODE_640x480_in_768x480_8bpp_PALETTE);
        if (FAILED(sc)) goto Bad;
    }

    Msg = _T("map");
    sc = FrameBuffer->v->Map(FrameBuffer,(PTR*)&FB,&Length);
    if (FAILED(sc)) goto Bad;

    Msg = _T("get Hlen");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_HORIZONTAL_LENGTH,&Hlen);
    if (FAILED(sc)) goto Bad;

    Msg = _T("get Hvis");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_HORIZONTAL_VISIBLE_LENGTH,&Hvis);
    if (FAILED(sc)) goto Bad;

    Msg = _T("get Vvis");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_VERTICAL_VISIBLE_LENGTH,&Vvis);
    if (FAILED(sc)) goto Bad;

    return S_OK;
 Bad:
    _tprintf(_T("Could not %s %s sc=%x\n"), Msg, Name, sc);
    return sc;
}

/* All tests have the same interface
 */
#define DECLARE_TEST(_n_,_doc_) \
_TCHAR TestDoc ## _n_[] = _T(_doc_); \
void test ## _n_(PIFRAMEBUFFER Fb)

/* Show palette in a 256x256 square, centered, identical pixels per 1 row
 */
DECLARE_TEST(0,
"Show palette in a 256x256 square, centered, identical pixels per 1 row\n")
{
    UINT i, j, o;

    _tprintf(TestDoc0);
    for (i = 0; i < 256; i++) {
        o = ((Hvis-256)/2) + ((((Vvis-256)/2) + i) * Hlen); /* center at (192,112) */
        for (j = 0; j < 256; j++)
            FB[o+j] = (UINT8) i;
    }
}

/* Show palette in a 512x256 rect, centered, identical pixels per 2 columns
 */
DECLARE_TEST(1,
"Show palette in a 512x256 rect, centered, identical pixels per 2 columns\n")
{
    UINT i, j, o;

    _tprintf(TestDoc1);
    for (i = 0; i < 256; i++) {
        o = ((Hvis-512)/2) + ((((Vvis-256)/2) + i) * Hlen); /* center at (64,112) */
        for (j = 0; j < 512; j += 2) {
            FB[o+j] = (UINT8) (j >> 1);
            FB[o+j+1] = (UINT8) (j >> 1);
        }
    }
}

/* Show palette in a 20x13 matrix, centered, blocks of 32x32 identical pixels
 */
int zero = 0;
void dotest2(void)
{
    UINT i, j, k, n, o, p, c;

    p = zero;
    for (i = 0; i < 13; i++) {
        o = 0 + ((32 + (i * 32)) * Hlen); /* center at (0,32) */
        for (n = 0; n < 32; n++) {
            for (j = 0; j < Hvis; j += 32) {
                c = p + (j/32);
                if (c > 255) c = 255;
                for (k = 0; k < 32; k++)
                    FB[o+j+k] = (UINT8) c;
            }
            o += Hlen;
        }
        p += 20;
    }
}

DECLARE_TEST(2,
"Show palette in a 20x13 matrix, centered, blocks of 32x32 identical pixels\n")
{
    _tprintf(TestDoc2);
    dotest2();
}

/* Repeatedly do the above, for 10 seconds.
 * Count the number of iterations (371).
 */
DECLARE_TEST(3,
"Show palette in a 20x13 matrix, centered, blocks of 32x32 identical pixels\n"\
_T("Repeat for 10 seconds, count the number of iterations.\n"))
{
    TIME Now, EndTime;
    UINT32 nReps = 0;

    _tprintf(TestDoc3);

    Now = CurrentTime();
    Int32ToInt64(EndTime,TIME_SECONDS(10));
    EndTime = TimeAdd(Now,EndTime);

    for (; TimeLess(Now,EndTime);nReps++) {
        dotest2();
        Now = CurrentTime();
    }
    _tprintf(_T("that was %u reps in 10 secs.\n"),nReps);
}


/* Same, flip blocks each frame
 */
DECLARE_TEST(4,
"Show palette in a 20x13 matrix, centered, blocks of 32x32 identical pixels\n"\
_T("Flip blocks, repeat, count the number of iterations in 10 seconds.\n"))
{
    TIME Now, EndTime;
    UINT32 nReps = 0;

    _tprintf(TestDoc4);

    Now = CurrentTime();
    Int32ToInt64(EndTime,TIME_SECONDS(10));
    EndTime = TimeAdd(Now,EndTime);

    for (; TimeLess(Now,EndTime);nReps++) {
        dotest2();
        zero ^= 1;
        Now = CurrentTime();
    }
    _tprintf(_T("that was %u reps in 10 secs.\n"),nReps);
}

/* Same, flip blocks each frame, modulo 8
 */
DECLARE_TEST(5,
"Show palette in a 20x13 matrix, centered, blocks of 32x32 identical pixels\n"\
_T("Flip blocks mod8, repeat, count the number of iterations in 10 seconds.\n"))
{
    TIME Now, EndTime;
    UINT32 nReps = 0;

    _tprintf(TestDoc5);

    Now = CurrentTime();
    Int32ToInt64(EndTime,TIME_SECONDS(10));
    EndTime = TimeAdd(Now,EndTime);

    for (; TimeLess(Now,EndTime);nReps++) {
        dotest2();
        zero = (zero + 1) & 7;
        Now = CurrentTime();
    }
    _tprintf(_T("that was %u reps in 10 secs.\n"),nReps);
}

/* Same, flip blocks each frame, modulo 4
 */
DECLARE_TEST(6,
"Show palette in a 20x13 matrix, centered, blocks of 32x32 identical pixels\n"\
_T("Flip blocks mod4, repeat, count the number of iterations in 10 seconds.\n"))
{
    TIME Now, EndTime;
    UINT32 nReps = 0;

    _tprintf(TestDoc6);

    Now = CurrentTime();
    Int32ToInt64(EndTime,TIME_SECONDS(10));
    EndTime = TimeAdd(Now,EndTime);

    for (; TimeLess(Now,EndTime);nReps++) {
        dotest2();
        zero = (zero + 1) & 3;
        Now = CurrentTime();
    }
    zero = 0;
    dotest2();
    _tprintf(_T("that was %u reps in 10 secs.\n"),nReps);
}

static const unsigned char gamma[256] = 
{8,12,16,19,22,24,27,29,31,34,36,38,40,41,43,45,47,49,50,52,
53,55,57,58,60,61,63,64,65,67,68,70,71,72,74,75,76,77,79,80,
81,82,84,85,86,87,88,90,91,92,93,94,95,96,98,99,100,101,102,103,
104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,
124,125,126,127,128,129,130,131,132,133,134,135,135,136,137,138,139,140,141,142,
143,143,144,145,146,147,148,149,150,150,151,152,153,154,155,155,156,157,158,159,
160,160,161,162,163,164,165,165,166,167,168,169,169,170,171,172,173,173,174,175,
176,176,177,178,179,180,180,181,182,183,183,184,185,186,186,187,188,189,189,190,
191,192,192,193,194,195,195,196,197,197,198,199,200,200,201,202,202,203,204,205,
205,206,207,207,208,209,210,210,211,212,212,213,214,214,215,216,216,217,218,219,
219,220,221,221,222,223,223,224,225,225,226,227,227,228,229,229,230,231,231,232,
233,233,234,235,235,236,237,237,238,238,239,240,240,241,242,242,243,244,244,245,
246,246,247,247,248,249,249,250,251,251,252,253,253,254,254,255};

/* Show palette in a 20x13 matrix, centered, blocks of 32x32 identical pixels
 * with gamma-correction.
 */
DECLARE_TEST(7,
"Show gamma-corrected palette in a 20x13 matrix, centered, blocks of 32x32 identical pixels\n")
{
    UINT i, j, k, n, o, p, c;

    _tprintf(TestDoc7);

    p = zero;
    for (i = 0; i < 13; i++) {
        o = 0 + ((32 + (i * 32)) * Hlen); /* center at (0,32) */
        for (n = 0; n < 32; n++) {
            for (j = 0; j < Hvis; j += 32) {
                c = p + (j/32);
                if (c > 255) c = 255;
                c = gamma[c];
                for (k = 0; k < 32; k++)
                    FB[o+j+k] = (UINT8) c;
            }
            o += Hlen;
        }
        p += 20;
    }
}

/* Palette contents
 */
typedef struct {
    UINT8 r,g,b,x;
} RGBQUAD;

const RGBQUAD DefaultPalette[256] = 

{{0xff,0xff,0xff,0x0}, {0x0,0xff,0xff,0x0}, {0xff,0x0,0xff,0x0}, {0x0,0x0,0xff,0x0}, {0xff,0xff,0x0,0x0}, {0x0,0xff,0x0,0x0}, {0xff,0x0,0x0,0x0}, {0x80,0x80,0x80,0x0},
{0xa0,0xa0,0xa4,0x0}, {0xff,0xfb,0xf0,0x0}, {0xa0,0xc0,0xc0,0x0}, {0x80,0xc0,0xc0,0x0}, {0x60,0xc0,0xc0,0x0}, {0x40,0xc0,0xc0,0x0}, {0x20,0xc0,0xc0,0x0}, {0x0,0xc0,0xc0,0x0},
{0xe0,0xa0,0xc0,0x0}, {0xc0,0xa0,0xc0,0x0}, {0xa0,0xa0,0xc0,0x0}, {0x80,0xa0,0xc0,0x0}, {0x60,0xa0,0xc0,0x0}, {0x40,0xa0,0xc0,0x0}, {0x20,0xa0,0xc0,0x0}, {0x0,0xa0,0xc0,0x0},
{0xe0,0x80,0xc0,0x0}, {0xc0,0x80,0xc0,0x0}, {0xa0,0x80,0xc0,0x0}, {0x80,0x80,0xc0,0x0}, {0x60,0x80,0xc0,0x0}, {0x40,0x80,0xc0,0x0}, {0x20,0x80,0xc0,0x0}, {0x0,0x80,0xc0,0x0},
{0xe0,0x60,0xc0,0x0}, {0xc0,0x60,0xc0,0x0}, {0xa0,0x60,0xc0,0x0}, {0x80,0x60,0xc0,0x0}, {0x60,0x60,0xc0,0x0}, {0x40,0x60,0xc0,0x0}, {0x20,0x60,0xc0,0x0}, {0x0,0x60,0xc0,0x0},
{0xe0,0x40,0xc0,0x0}, {0xc0,0x40,0xc0,0x0}, {0xa0,0x40,0xc0,0x0}, {0x80,0x40,0xc0,0x0}, {0x60,0x40,0xc0,0x0}, {0x40,0x40,0xc0,0x0}, {0x20,0x40,0xc0,0x0}, {0x0,0x40,0xc0,0x0},
{0xe0,0x20,0xc0,0x0}, {0xc0,0x20,0xc0,0x0}, {0xa0,0x20,0xc0,0x0}, {0x80,0x20,0xc0,0x0}, {0x60,0x20,0xc0,0x0}, {0x40,0x20,0xc0,0x0}, {0x20,0x20,0xc0,0x0}, {0x0,0x20,0xc0,0x0},
{0xe0,0x0,0xc0,0x0}, {0xc0,0x0,0xc0,0x0}, {0xa0,0x0,0xc0,0x0}, {0x80,0x0,0xc0,0x0}, {0x60,0x0,0xc0,0x0}, {0x40,0x0,0xc0,0x0}, {0x20,0x0,0xc0,0x0}, {0x0,0x0,0xc0,0x0},
{0xe0,0xe0,0x80,0x0}, {0xc0,0xe0,0x80,0x0}, {0xa0,0xe0,0x80,0x0}, {0x80,0xe0,0x80,0x0}, {0x60,0xe0,0x80,0x0}, {0x40,0xe0,0x80,0x0}, {0x20,0xe0,0x80,0x0}, {0x0,0xe0,0x80,0x0},
{0xe0,0xc0,0x80,0x0}, {0xc0,0xc0,0x80,0x0}, {0xa0,0xc0,0x80,0x0}, {0x80,0xc0,0x80,0x0}, {0x60,0xc0,0x80,0x0}, {0x40,0xc0,0x80,0x0}, {0x20,0xc0,0x80,0x0}, {0x0,0xc0,0x80,0x0},
{0xe0,0xa0,0x80,0x0}, {0xc0,0xa0,0x80,0x0}, {0xa0,0xa0,0x80,0x0}, {0x80,0xa0,0x80,0x0}, {0x60,0xa0,0x80,0x0}, {0x40,0xa0,0x80,0x0}, {0x20,0xa0,0x80,0x0}, {0x0,0xa0,0x80,0x0},
{0xe0,0x80,0x80,0x0}, {0xc0,0x80,0x80,0x0}, {0xa0,0x80,0x80,0x0}, {0x80,0x80,0x80,0x0}, {0x60,0x80,0x80,0x0}, {0x40,0x80,0x80,0x0}, {0x20,0x80,0x80,0x0}, {0x0,0x80,0x80,0x0},
{0xe0,0x60,0x80,0x0}, {0xc0,0x60,0x80,0x0}, {0xa0,0x60,0x80,0x0}, {0x80,0x60,0x80,0x0}, {0x60,0x60,0x80,0x0}, {0x40,0x60,0x80,0x0}, {0x20,0x60,0x80,0x0}, {0x0,0x60,0x80,0x0},
{0xe0,0x40,0x80,0x0}, {0xc0,0x40,0x80,0x0}, {0xa0,0x40,0x80,0x0}, {0x80,0x40,0x80,0x0}, {0x60,0x40,0x80,0x0}, {0x40,0x40,0x80,0x0}, {0x20,0x40,0x80,0x0}, {0x0,0x40,0x80,0x0},
{0xe0,0x20,0x80,0x0}, {0xc0,0x20,0x80,0x0}, {0xa0,0x20,0x80,0x0}, {0x80,0x20,0x80,0x0}, {0x60,0x20,0x80,0x0}, {0x40,0x20,0x80,0x0}, {0x20,0x20,0x80,0x0}, {0x0,0x20,0x80,0x0},
{0xe0,0x0,0x80,0x0}, {0xc0,0x0,0x80,0x0}, {0xa0,0x0,0x80,0x0}, {0x80,0x0,0x80,0x0}, {0x60,0x0,0x80,0x0}, {0x40,0x0,0x80,0x0}, {0x20,0x0,0x80,0x0}, {0x0,0x0,0x80,0x0},
{0xe0,0xe0,0x40,0x0}, {0xc0,0xe0,0x40,0x0}, {0xa0,0xe0,0x40,0x0}, {0x80,0xe0,0x40,0x0}, {0x60,0xe0,0x40,0x0}, {0x40,0xe0,0x40,0x0}, {0x20,0xe0,0x40,0x0}, {0x0,0xe0,0x40,0x0},
{0xe0,0xc0,0x40,0x0}, {0xc0,0xc0,0x40,0x0}, {0xa0,0xc0,0x40,0x0}, {0x80,0xc0,0x40,0x0}, {0x60,0xc0,0x40,0x0}, {0x40,0xc0,0x40,0x0}, {0x20,0xc0,0x40,0x0}, {0x0,0xc0,0x40,0x0},
{0xe0,0xa0,0x40,0x0}, {0xc0,0xa0,0x40,0x0}, {0xa0,0xa0,0x40,0x0}, {0x80,0xa0,0x40,0x0}, {0x60,0xa0,0x40,0x0}, {0x40,0xa0,0x40,0x0}, {0x20,0xa0,0x40,0x0}, {0x0,0xa0,0x40,0x0},
{0xe0,0x80,0x40,0x0}, {0xc0,0x80,0x40,0x0}, {0xa0,0x80,0x40,0x0}, {0x80,0x80,0x40,0x0}, {0x60,0x80,0x40,0x0}, {0x40,0x80,0x40,0x0}, {0x20,0x80,0x40,0x0}, {0x0,0x80,0x40,0x0},
{0xe0,0x60,0x40,0x0}, {0xc0,0x60,0x40,0x0}, {0xa0,0x60,0x40,0x0}, {0x80,0x60,0x40,0x0}, {0x60,0x60,0x40,0x0}, {0x40,0x60,0x40,0x0}, {0x20,0x60,0x40,0x0}, {0x0,0x60,0x40,0x0},
{0xe0,0x40,0x40,0x0}, {0xc0,0x40,0x40,0x0}, {0xa0,0x40,0x40,0x0}, {0x80,0x40,0x40,0x0}, {0x60,0x40,0x40,0x0}, {0x40,0x40,0x40,0x0}, {0x20,0x40,0x40,0x0}, {0x0,0x40,0x40,0x0},
{0xe0,0x20,0x40,0x0}, {0xc0,0x20,0x40,0x0}, {0xa0,0x20,0x40,0x0}, {0x80,0x20,0x40,0x0}, {0x60,0x20,0x40,0x0}, {0x40,0x20,0x40,0x0}, {0x20,0x20,0x40,0x0}, {0x0,0x20,0x40,0x0},
{0xe0,0x0,0x40,0x0}, {0xc0,0x0,0x40,0x0}, {0xa0,0x0,0x40,0x0}, {0x80,0x0,0x40,0x0}, {0x60,0x0,0x40,0x0}, {0x40,0x0,0x40,0x0}, {0x20,0x0,0x40,0x0}, {0x0,0x0,0x40,0x0},
{0xe0,0xe0,0x0,0x0}, {0xc0,0xe0,0x0,0x0}, {0xa0,0xe0,0x0,0x0}, {0x80,0xe0,0x0,0x0}, {0x60,0xe0,0x0,0x0}, {0x40,0xe0,0x0,0x0}, {0x20,0xe0,0x0,0x0}, {0x0,0xe0,0x0,0x0},
{0xe0,0xc0,0x0,0x0}, {0xc0,0xc0,0x0,0x0}, {0xa0,0xc0,0x0,0x0}, {0x80,0xc0,0x0,0x0}, {0x60,0xc0,0x0,0x0}, {0x40,0xc0,0x0,0x0}, {0x20,0xc0,0x0,0x0}, {0x0,0xc0,0x0,0x0},
{0xe0,0xa0,0x0,0x0}, {0xc0,0xa0,0x0,0x0}, {0xa0,0xa0,0x0,0x0}, {0x80,0xa0,0x0,0x0}, {0x60,0xa0,0x0,0x0}, {0x40,0xa0,0x0,0x0}, {0x20,0xa0,0x0,0x0}, {0x0,0xa0,0x0,0x0},
{0xe0,0x80,0x0,0x0}, {0xc0,0x80,0x0,0x0}, {0xa0,0x80,0x0,0x0}, {0x80,0x80,0x0,0x0}, {0x60,0x80,0x0,0x0}, {0x40,0x80,0x0,0x0}, {0x20,0x80,0x0,0x0}, {0x0,0x80,0x0,0x0},
{0xe0,0x60,0x0,0x0}, {0xc0,0x60,0x0,0x0}, {0xa0,0x60,0x0,0x0}, {0x80,0x60,0x0,0x0}, {0x60,0x60,0x0,0x0}, {0x40,0x60,0x0,0x0}, {0x20,0x60,0x0,0x0}, {0x0,0x60,0x0,0x0},
{0xe0,0x40,0x0,0x0}, {0xc0,0x40,0x0,0x0}, {0xa0,0x40,0x0,0x0}, {0x80,0x40,0x0,0x0}, {0x60,0x40,0x0,0x0}, {0x40,0x40,0x0,0x0}, {0x20,0x40,0x0,0x0}, {0x0,0x40,0x0,0x0},
{0xe0,0x20,0x0,0x0}, {0xc0,0x20,0x0,0x0}, {0xa0,0x20,0x0,0x0}, {0x80,0x20,0x0,0x0}, {0x60,0x20,0x0,0x0}, {0x40,0x20,0x0,0x0}, {0xa6,0xca,0xf0,0x0}, {0xc0,0xdc,0xc0,0x0},
{0xc0,0xc0,0xc0,0x0}, {0x0,0x80,0x80,0x0}, {0x80,0x0,0x80,0x0}, {0x0,0x0,0x80,0x0}, {0x80,0x80,0x0,0x0}, {0x0,0x80,0x0,0x0}, {0x80,0x0,0x0,0x0}, {0x0,0x0,0x0,0x0}};


RGBQUAD CurrentPalette[256];

/* Dump the palette
 */
DECLARE_TEST(8,
"Dump the palette\n")
{
    UINT i;
    UINT8 Entry[4];
    SCODE sc;
    UINT32 v;

    _tprintf(TestDoc8);

    /* Do it all-at-once first
     */
    sc = Fb->v->GetPalette(Fb,(UINT8*)CurrentPalette,sizeof CurrentPalette,PIXEL_TYPE_R8G8B8X8);
    if (FAILED(sc)) {
        _tprintf(_T("GetPalette failed sc=%x\n"),sc);
        return;
    }

    /* Now verify entry-by-entry
     */
    for (i = 0; i < 256; i++) {
        sc = Fb->v->GetPaletteEntry(Fb,i,Entry,4,PIXEL_TYPE_R8G8B8X8);
        if (FAILED(sc)) {
            _tprintf(_T("GetPaletteEntry(%d) failed sc=%x\n"),i,sc);
            return;
        }
        if (memcmp(Entry,&CurrentPalette[i],4) != 0) {
            _tprintf(_T("Palette[%3d] inconsistent %08x != %08x\n"),i,
                     *(UINT*)Entry,
                     *(UINT*)(&CurrentPalette[i]));
        }
    }

    /* Dump it
     */
    for (i = 0; i < 256; i++) {
        v = *(UINT32*)(&CurrentPalette[i]);
        _tprintf(_T(" %08x"),v);
        if ((i & 7) == 7)
            _tprintf(_T("\n"));
    }
}

/* Set the palette
 */
DECLARE_TEST(9,
"Load the default Windows Palette\n")
{
    UINT i;
    UINT32 v;
    SCODE sc;

    /* Say what
     */
    _tprintf(TestDoc9);
    for (i = 0; i < 256; i++) {

        v = *(UINT32*)(&DefaultPalette[i]);
        _tprintf(_T(" %08x"),v);
        if ((i & 7) == 7)
            _tprintf(_T("\n"));
    }

    /* Load it
     */
    sc = Fb->v->SetPalette(Fb,(UINT8*)DefaultPalette,sizeof DefaultPalette,PIXEL_TYPE_R8G8B8X8);
    if (FAILED(sc)) {
        _tprintf(_T("SetPalette failed sc=%x\n"),sc);
        return;
    }

    /* Verify
     */
    sc = Fb->v->GetPalette(Fb,(UINT8*)CurrentPalette,sizeof CurrentPalette,PIXEL_TYPE_R8G8B8X8);
    if (FAILED(sc)) {
        _tprintf(_T("GetPalette failed sc=%x\n"),sc);
        return;
    }

    if (memcmp(DefaultPalette,CurrentPalette,sizeof DefaultPalette) == 0)
        return;

    /* Something is wrong, say what.
     */
    for (i = 0; i < 256; i++) {

        if (memcmp(&DefaultPalette[i],&CurrentPalette[i],4) != 0) {
            _tprintf(_T("Palette[%3d] is wrong %08x (not %08x)\n"),i,
                     *(UINT*)(&CurrentPalette[i]),
                     *(UINT*)(&DefaultPalette[i]));
        }

    }
}

/* Set a b&w palette
 */
DECLARE_TEST(10,
"Set a b&w palette\n")
{
    UINT i;
    UINT32 v;
    RGBQUAD e;
    SCODE sc;

    _tprintf(TestDoc10);

    e.x = 0;
    for (i = 0; i < 256; i++) {

        e.r = (UINT8)i;
        e.g = (UINT8)i;
        e.b = (UINT8)i;

        sc = Fb->v->SetPaletteEntry(Fb,i,(UINT8*) &e, sizeof e,PIXEL_TYPE_R8G8B8X8);
        if (FAILED(sc)) {
            _tprintf(_T("SetPaletteEntry(%d) failed sc=%x\n"),i,sc);
            return;
        }

        v = *(UINT32*)&e;
        _tprintf(_T(" %08x"),v);
        if ((i & 7) == 7)
            _tprintf(_T("\n"));
    }

}

/* Set the palette in reverse
 */
DECLARE_TEST(11,
"Load the default Windows Palette reversed\n")
{
    UINT i;
    UINT32 v;
    SCODE sc;

    _tprintf(TestDoc11);

    for (i = 0; i < 256; i++) {

        sc = Fb->v->SetPaletteEntry(Fb,
                                    i,
                                    (UINT8*) &DefaultPalette[255-i],
                                    sizeof DefaultPalette[0],
                                    PIXEL_TYPE_R8G8B8X8);
        if (FAILED(sc)) {
            _tprintf(_T("SetPaletteEntry(%d) failed sc=%x\n"),i,sc);
            return;
        }

        v = *(UINT32*)(&DefaultPalette[255-i]);
        _tprintf(_T(" %08x"),v);
        if ((i & 7) == 7)
            _tprintf(_T("\n"));
    }

}

/* Check pure colors
 */
DECLARE_TEST(12,
"Set palette[1]=pure red, palette[2]=pure green, palette[3]=pure blue\n")
{
    UINT i;
    SCODE sc;
    RGBQUAD e;

    _tprintf(TestDoc12);

    /* red */
    i = 1;
    e.x = 0;
    e.r = 0xff; e.g = 0;  e.b = 0;

    sc = Fb->v->SetPaletteEntry(Fb,i,(UINT8*) &e, sizeof e,PIXEL_TYPE_R8G8B8X8);
    if (FAILED(sc)) {
        _tprintf(_T("SetPaletteEntry(%d) failed sc=%x\n"),i,sc);
        return;
    }

    /* green */
    i = 2;
    e.x = 0;
    e.r = 0; e.g = 0xff;  e.b = 0;

    sc = Fb->v->SetPaletteEntry(Fb,i,(UINT8*) &e, sizeof e,PIXEL_TYPE_R8G8B8X8);
    if (FAILED(sc)) {
        _tprintf(_T("SetPaletteEntry(%d) failed sc=%x\n"),i,sc);
        return;
    }

    /* blue */
    i = 3;
    e.x = 0;
    e.r = 0; e.g = 0;  e.b = 0xff;

    sc = Fb->v->SetPaletteEntry(Fb,i,(UINT8*) &e, sizeof e,PIXEL_TYPE_R8G8B8X8);
    if (FAILED(sc)) {
        _tprintf(_T("SetPaletteEntry(%d) failed sc=%x\n"),i,sc);
        return;
    }

    /* Now show it
     */
    test2(Fb);
}

/* Set a b&w palette in reverse
 */
DECLARE_TEST(13,
"Set a reversed b&w palette\n")
{
    UINT i;
    UINT32 v;
    RGBQUAD e;
    SCODE sc;

    _tprintf(TestDoc13);

    e.x = 0;
    for (i = 0; i < 256; i++) {

        e.r = (UINT8)(255-i);
        e.g = (UINT8)(255-i);
        e.b = (UINT8)(255-i);

        sc = Fb->v->SetPaletteEntry(Fb,i,(UINT8*) &e, sizeof e,PIXEL_TYPE_R8G8B8X8);
        if (FAILED(sc)) {
            _tprintf(_T("SetPaletteEntry(%d) failed sc=%x\n"),i,sc);
            return;
        }

        v = *(UINT32*)&e;
        _tprintf(_T(" %08x"),v);
        if ((i & 7) == 7)
            _tprintf(_T("\n"));
    }

}

/*
 */
#define default 0

typedef struct {
    void (*TestCase)(PIFRAMEBUFFER Fb);
    _TCHAR * TestDoc;
} TEST, *PTEST;
#define test(_n_) { test ## _n_, TestDoc ## _n_ }

#define MAXTESTS 16
TEST AllTests[MAXTESTS] = {
    test(0), test(1), test(2), test(3), test(4),
    test(5), test(6), test(7), test(8), test(9),
    test(10), test(11), test(12), test(13),
    {NULL,NULL}
};

int _tmain(int argc, _TCHAR **argv)
{
    SCODE sc;
    PIFRAMEBUFFER FrameBuffer = NULL;
    UINT which;

    /* Parse args
     */
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_T("No command line args (%x)\n"), sc);
        return -1;
    }

    if (argc < 2) {
        _tprintf(_T("Usage: %s framebuffername [testno]\n"), argv[0]);
    Usages:
        _tprintf(_T("testno -> Action:\n"));
        for (which = 0; which < MAXTESTS; which++)
            if (AllTests[which].TestDoc != NULL)
                _tprintf(_T("%2d -> %s"),which,AllTests[which].TestDoc);
        goto Out;
    }

    /* Usages
     */
    if (_tcscmp(_T("?"),argv[1]) == 0) {
        goto Usages;
    }

    which = default;
    if (argc > 2)
        which = (_ttoi(argv[2]) & (MAXTESTS-1));

    /* Find it
     */
    sc = BindToObject(CurrentNameSpace(),
                      argv[1],
                      NAME_SPACE_READ | NAME_SPACE_WRITE,
                      &IID_IFrameBuffer,
                      (void *) &FrameBuffer);
    if (FAILED(sc)) {
        _tprintf(_T("Could not bind to %s sc=%x\n"), argv[1], sc);
        goto Out;
    }

    /* Map it and get all the necessary info
     */
    sc = Initialize(FrameBuffer,argv[1]);
    if (FAILED(sc)) {
        _tprintf(_T("Could not initialize %s sc=%x\n"), argv[1], sc);
        goto Out;
    }

    /* perform the test
     */
    if (AllTests[which].TestCase)
        AllTests[which].TestCase(FrameBuffer);
    else
        _tprintf(_T("No such test\n"));

    /* Done
     */
 Out:
    if (FrameBuffer) {
        if (FB != NULL)
            FrameBuffer->v->Unmap(FrameBuffer,FB);
        FrameBuffer->v->Release(FrameBuffer);
    }
    (void) FreeCmdLine(&argc,&argv);
    return 0;
}
