/* Copyright (c) Microsoft Corporation. All rights reserved. */

/*********************************************************************/
/* playwav.c                                                          */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <tchar.h>
#include <riff.h>
#include <mmlite.h>

/* utility */
FILE *GetWaveFile(_TCHAR *FileName, UINT *nFileSize,
          UINT *nSamplesPerSec, UINT *nChannels, UINT *nBytesPerSample)
{
    WAVE_CHUNK  wh;
    RIFF_HEADER rh;
    UINT        c;
    FILE       *fpin = NULL;
    _TCHAR     *msg = NULL;

#define ERR(s) {msg = _T(s);goto ErrorExit;}

    /* open test data file */
    fpin = _tfopen(FileName,_T("rb"));
    if(fpin==NULL) ERR("unable to open test audio data file\n");

    /* parse file header
     */
    c = fread((void*)&rh, 1, sizeof(rh), fpin);
    if (c != sizeof(rh) ||
    rh.ckid != RIFF_CHUNK_ID)
        ERR("Short read");

    c = fread((void*)&wh, 1, sizeof(wh), fpin);
    if (c != sizeof(wh)) ERR("Short read");
    if (wh.ckid != WAVE_CHUNK_ID) ERR("Bad riff file");
    if (wh.fmtid != WAVEFMT_CHUNK_ID) ERR("Not a WAVE file");
    if (wh.fmt.wFormatTag != WAVE_FORMAT_PCM) ERR("Not a PCM file");
    if (wh.cksize != sizeof(PCMWAVEFORMAT)) ERR("Bad PCM format file");

    c = fread((void*)&rh, 1, sizeof(rh), fpin);
    if (c != sizeof(rh) ||
    rh.ckid != DATA_CHUNK_ID)
    ERR("missing RIFF data");

    *nFileSize = rh.cksize;

    /* Return other params */
    *nSamplesPerSec = wh.fmt.nSamplesPerSec;
    *nChannels = wh.fmt.nChannels;
    *nBytesPerSample = wh.fmt.wBitsPerSample >> 3;

    return fpin;

 ErrorExit:
    if (fpin != NULL)
    fclose(fpin);
    _tprintf(_T("%s\r\n"), msg);
    return NULL;
}

PIFILE SetSampleRate(_TCHAR *AudioName,
                   UINT nSampleRate,
                   UINT nBitsPerSample,
                   UINT nChannels)
{
    UINT64 Aurgh;
    PIFILE AudioOut;
    SCODE sc;

    sc = BindToObject(CurrentNameSpace(),
                      AudioName,
                      NAME_SPACE_WRITE,
                      &IID_IFile,
                      (void *)&AudioOut);
    if (FAILED(sc)) {
        _tprintf(_T("Output codec %s not found [sc=%x]\r\n"),AudioName,sc);
        return NULL;
    }

    /* funnything: rate=low32 bps=32..47 channels=48..55 cmd=56..63 */
    Int64FromHighAndLow(Aurgh,
                        (0<<24)|((nChannels&0xff)<<16)|(nBitsPerSample&0xffff),
                        nSampleRate);

    sc = AudioOut->v->SetSize(AudioOut,Aurgh);
    if (FAILED(sc)) {
        _tprintf(_T("SetSize[%s] failed with sc=%x\r\n"),AudioName,sc);
        AudioOut->v->Release(AudioOut);
        AudioOut = NULL;
    }

    return AudioOut;
}

struct _FILE_GUTS {
        char *_ptr;
        int   _cnt;
        char *_base;
        int   _flag;
        void *_file;
        int   _bufsize;
#ifdef __USE_IFILE
    UINT64 Seek;
#endif
};

#define nSamples 1024
typedef UINT32 BUFFER[nSamples]; /* max typesize => max buffer size */
BUFFER DataBuffer[4];

/*** main entry point ***/
int _tmain(INT argc,_TCHAR **argv)
{
    UINT                 nBytesRead;
    UINT                 nSampleRate, nChannels, nBytesPerSample;
    PIFILE               AudioOut = NULL;
    FILE                *DataIn = NULL;
    int                  idx = 0;

    argc = 0;
    (void) ParseCmdLine( &argc, &argv);
    if (argc < 3) {
        _tprintf(_T("Usage:%s <wav filename> <codec name>\n"), argv[0]);
        goto cleanup;
    }

    /* say hello
     */
    _tprintf(_T("[%s - Play .WAV file %s on %s]\n"), argv[0], argv[1], argv[2]);

    /* input data file
     */
    DataIn = GetWaveFile(argv[1],
                         &nBytesRead,
                         &nSampleRate,
                         &nChannels,
                         &nBytesPerSample);
    if (DataIn == NULL)
        /* spoken already */
    goto cleanup;

    _tprintf(_T("%s holds %d bytes worth of %d-Bit %d Hz %s data.\n"),
             argv[1],
             nBytesRead,
             nBytesPerSample*8,
             nSampleRate,
             (nChannels == 1) ? _T("mono") : _T("stereo")
             );

    /* get audio driver, the easy way
     */
    AudioOut = SetSampleRate(argv[2],
                             nSampleRate,
                             nBytesPerSample*8,
                             nChannels);
    if (AudioOut == NULL) {
        _tprintf(_T("%s cant take it, sorry."),argv[2]);
        goto cleanup;
    }

    /*triplehacko*/
    /* but. There is some SetOSHandle?() standard way to do this */
    ((struct _FILE_GUTS *)stderr)->_file = AudioOut;

    /* loop till done
     */
    _tprintf(_T("playwav: playing ..."));
    for (;;) {
        size_t n;
        n = fread(DataBuffer[idx],
                  nBytesPerSample*nChannels,
                  nSamples,
                  DataIn);

        if (n == 0)
            break;

        /* munge.. */

        n = fwrite(DataBuffer[idx],
                   n*nBytesPerSample*nChannels,
                   1,
                   stderr);

        idx = ++idx & 0x3;
    }
    _tprintf(_T("done.\r\n"));

    /* release audio driver interfaces */
 cleanup:
    if (AudioOut != NULL)
        AudioOut->v->Release(AudioOut);
    if (DataIn != NULL)
        fclose(DataIn);
    (void) FreeCmdLine( &argc, &argv);

    /* say goodbye */
    _tprintf(_T("[playwav - bye]\n"));
    return 0;
}

