#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <assert.h>
#include "..\utilities\utility.h"

/*
* CT.EXE - CreateThread() testing. 
* Provides access to CreateThread() api through command line. 
* Args: 
*   -start <0|1> sets start address of function, 0 = do not set, !0 = set to ThreadProc
*   -arg <string> is argument to pass to new thread
*   -stack <0xnumbytes> is size of stack for new threads
*   -threads <num threads> is number of threads to kick off
*   -valid_thread <0|1> sets whether a valid thread ptr or NULL is passed to CreateThread()
* Copyright 2002 Microsoft Corporation All Rights Reserved
*/

/*
* Threads are stored in a list for easy handling. This is just a singly linked list node
* 
*/
typedef struct thread_list
{
    IThread * pThread; 
    struct thread_list * pNext; 
}thread_list;

/*
* 'Care package' of stuff to pass to threads as params. There is only one, each ThreadProc 
* gets a ptr to it.
* The nCount var is set to the number of threads created. 
* Each ThreadProc dereferences the ptr and decrements nCount once. When _tmain() sees nCount
* go to 0 it knows all threads are done.
* The mutex is used to protect nCount. 
* There is a text field just for fun
*/
typedef struct thread_care_package
{
    UINT32 nCount; 
    MUTEX  mtx; 
    _TCHAR sMsg[1024];
}thread_care_package; 


void CTUsage(const _TCHAR *s);
void ThreadFunc(void * p);
void CleanupThreadList(struct thread_list * pHead);

#define ARGLEN (1024)

int _tmain(int argc, _TCHAR ** argv);
int _tmain(int argc, _TCHAR ** argv)
{
    THREAD_FUNCTION pStart; 
    THREAD_ARGUMENT Arg; 
    ADDR_SIZE   Stacksize; 
    UINT32 nThreads, idx;
    IProcess * pProc = 0;
    SCODE sc;
    BOOL bValidThread;
    IThread ** ppT;
    const _TCHAR * s;
    thread_list * pList = 0, * pCur, * pHead;
    _TCHAR ThrdArgs[ARGLEN] = {0,};
    thread_care_package tcp = {0,};

    ParseCmdLine(&argc, &argv); 
    OptionalBreak(argc, argv); 
    TellAboutOptionalBreak();

    if(argc < 2)
       CTUsage(argv[0]); 

    //set start address arg, default is 0
    s = GetArgFromCommandLine(argc, argv, _T("start")); 
    if(!s)
        pStart = ThreadFunc; 
    else
        if(!*s)
            pStart = ThreadFunc; 
        else
        {
            _TCHAR * stop = 0;
        
            if(!_tcstoul(s, &stop, 10))
                pStart = 0; 
            else
                pStart = ThreadFunc; 
        }

    //args for thread ... just a text message to pass to ThreadProc
    s = GetArgFromCommandLine(argc, argv, _T("arg")); 
    if(!s)
        Arg = 0; 
    else
        if(!*s)
            Arg = 0; 
        else
        {
            _tcscpy(ThrdArgs, s); 
            Arg = ThrdArgs; 
        }

    //stack size for threads. Zero uses system default. 
    //less than about 400 bytes will thrash the system without fail
    //no way to test for stack size. 
    s = GetArgFromCommandLine(argc, argv, _T("stack")); 
    if(!s)
        Stacksize = 0; 
    else
        if(!*s)
            Stacksize = 0; 
        else
        {
            _TCHAR * stop = 0;
            Stacksize = _tcstoul(s, &stop, 16); 
        }

    //set number of threads to create
    s = GetArgFromCommandLine(argc, argv, _T("threads")); 
    if(!s)
        nThreads = 1; 
    else
        if(!*s)
            nThreads = 1; 
        else
        {
            _TCHAR * stop = 0;
  
            nThreads = _tcstoul(s, &stop, 10);
            if(!nThreads)
                nThreads = 1; 
        }


    //do we pass in a nice thread ptr or null?
    s = GetArgFromCommandLine(argc, argv, _T("valid_thread")); 
    //default to using nice thread
    if(!s)
        bValidThread = 1; 
    else
        if(!*s)
            bValidThread = 1;
        else
        {
            _TCHAR * stop = 0;
            bValidThread = _tcstoul(s, &stop, 10); 
        }

    FreeCmdLine(&argc, &argv);
   
    pProc = CurrentProcess(); 
    _tprintf(_T("\n\rCreateThread([Process:0x%.8x], [Start Address 0x%.8x], [Thread Args:%s], [Stacksize:%d], [Threads:%d])"), 
        (UINT32)pProc, 
        (UINT32)pStart, 
        ThrdArgs ? ThrdArgs : _T("\n\r(null)"), 
        (UINT32)Stacksize, 
        nThreads);
       
    //avoid interlacing output from threadprocs if we can ... 
    fflush(stdout);

    //set up mutex to protect counter
    Mutex_Init(&tcp.mtx); 
    if(Arg)
        _tcscpy(tcp.sMsg, Arg); 
    else
        *tcp.sMsg = 0; 

    //for(each thread)
    //  add to list;
    //  lock mutex
    //  create thread
    //  report
    //  unlock mutex
    for(idx = 0, tcp.nCount = nThreads;
     idx < nThreads; 
     ++idx)
    {
        pCur = malloc(sizeof(thread_list));
        assert(pCur);
        pCur->pNext = 0; 
        pCur->pThread = 0;
        
        if(idx == 0)
            pHead = pList = pCur;

        Mutex_Lock(&tcp.mtx); 
        
        if(bValidThread)
            ppT = &pCur->pThread; 
        else
            ppT = 0;

        //thread creation at last ... 
        sc = pProc->v->CreateThread(pProc, pStart, &tcp, Stacksize, 0, (IThread **)ppT);
        
        CreateThreadOutcome(CREATETHREAD, sc, pStart, pCur->pThread, ThrdArgs, Stacksize);
        if(FAILED(sc))
        {
            CleanupThreadList(pHead);
            Mutex_Unlock(&tcp.mtx);
            return SetAndReturn(1); 
        }
        
        Mutex_Unlock(&tcp.mtx);

        if(idx != 0)
            pList->pNext = pCur; 
    }


    //for(EVER)
    //  lock mutex
    //  get count and store it
    //  unlock mutex
    //  if count is 0 break
    for(;;)
    {
        UINT n; 
        Mutex_Lock(&tcp.mtx); 
        n = tcp.nCount;
        Mutex_Unlock(&tcp.mtx);
        if(!n)
            break; 
    }

    CleanupThreadList(pHead);

    Mutex_Destroy(&tcp.mtx);

    return SetAndReturn(0);
}


/*
* ThreadFunc - basic ThreadProc for testing
* Gets void pointer to thread_care_package with mutex, text, and count variable. 
* Prints text if not null
* Gets mutex/decrements count var/releases mutex/exits
*/
void ThreadFunc(void * p)
{
    thread_care_package * pTCP = (thread_care_package *)p; 
   _tprintf(_T("\n\rThread Function called, thread is [0x%.8x]"), (UINT32)CurrentThread()); 
   fflush(stdout);
 
    if(pTCP)
    {
        Mutex_Lock(&pTCP->mtx); 
        _tprintf(_T("\n\rThread parms are [%s]"), pTCP->sMsg); 
        pTCP->nCount--; 
        Mutex_Unlock(&pTCP->mtx); 
    }
}

void CTUsage(const _TCHAR *s)
{
    _tprintf(_T("\n\rUsage:%s [-start <0|1>] [-args <string>] [-stack <stack size base 10>] -threads [numthreads base 10] -valid_thread [0|1]"), s); 
    fflush(stdout);
}

/*
* Iterates list and frees memory
*/
void CleanupThreadList(struct thread_list * pHead)
{
    struct thread_list * pList, * pCur;
    for(pList = pHead; pList; pList = pCur)
    {
        pCur = pList->pNext; 
        free(pList);
    }
}