/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
* Tests LoadImage
* TestModule - pass in module name, binds and interrogates for basic info. 
* Params:
*   -module:name of module, e.g. corn.cob, go.exe
*   -buflen:
*   -func:Function within module. Can be expressed as a string e.g. Foo or an ordinal e.g. 2
*   -arg:Argument to pass to module, as ptr to text
*   -wait:Should we wait for module to complete? 0 does not wait, or !0 does
*   -flags:flags to pass to LoadImage() : LOAD_IMAGE_CALL_ENTRY_POINT | LOAD_IMAGE_CREATE_THREAD 
* are 0x1 and 0x2. 
*/

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "..\utilities\utility.h"

#define NAMEBUFLEN (1024)


void Error(void);
void Status(const _TCHAR * sMod, const _TCHAR * sFunc, UINT32 nFlags, UINT32 nOrdinal, INT32 nExit, UINT32 nBufLen, BOOL bWait);
void TMUsage(const _TCHAR * s);

int _tmain(INT argc, _TCHAR ** argv)
{
    SCODE sc; 
    const _TCHAR * sMod = 0, * sFunc = 0, * sWait = 0, * sFlags = 0, * sArgs = 0, * sNameBufLen = 0;
    UINT32 nNameBufLen = NAMEBUFLEN * sizeof(_TCHAR);
    _TCHAR * sName = 0;
    _TCHAR * sArgBuf = 0;
    _TCHAR * sFuncName = 0;
    IModule * pMod = 0; 
    UINT nOrdinal = 0; 
    UINT32 nFlags = 0;
    INT nExit = 0;
    BOOL bWait;
    ADDRESS FuncAddr = 0, EntryPoint = 0;
    _TCHAR * stop = 0;
    IHeap * pHeap = 0;
    PIPROCESS pSpawn = NULL;

    ParseCmdLine(&argc, &argv);
    OptionalBreak(argc, argv);
    TellAboutOptionalBreak();

    if(argc < 2)
    {
        TMUsage(argv[0]); 
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1); 
    }

    //get module name
    sMod = GetArgFromCommandLine(argc, argv, _T("module")); 
    if(!sMod)
    {
        TMUsage(argv[0]); 
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1);
    }
    if(!*sMod)
    {
        TMUsage(argv[0]); 
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1);
    }

 
    //get size of buffer ... 
    sNameBufLen = GetArgFromCommandLine(argc, argv, _T("buflen")); 
    if(sNameBufLen)
        nNameBufLen = _tcstoul(sNameBufLen, &stop, 16); 

    pHeap = CurrentHeap(); 
    
    //everything passed into the modules must be duped because we do not know 
    //if this host process will still be around when module dies
    sName = (_TCHAR *)pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, nNameBufLen, 0); 
    if(!sName)
    {
        Error(); 
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1);
    }
    
    //argument buffer
    sArgBuf = (_TCHAR *)pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, nNameBufLen, 0); 
    if(!sArgBuf)
    {
        Error(); 
        pHeap->v->Free(pHeap, 0, sName); 
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1); 
    }

    //name of function buffer
    sFuncName = (_TCHAR *)pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, nNameBufLen, 0); 
    if(!sFuncName)
    {
        Error(); 
        pHeap->v->Free(pHeap, 0, sName); 
        pHeap->v->Free(pHeap, 0, sArgBuf); 
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1);
    }

    sFunc = GetArgFromCommandLine(argc, argv, _T("func")); 
    if(sFunc && *sFunc)
    {
        nOrdinal = _tcstoul(sFunc, &stop, 10); 
        //non numeric input
        if(stop == sFunc)
        {
            nOrdinal = -1; //we will use the func name stored in sFuncname
            _tcscpy(sFuncName, sFunc); 
        }
        else
            ; //we are using the ordinal from _tcstoul() [might be 0]
    }
    else //nothing at all, use defaults
    {
        *sFuncName = 0; 
        nOrdinal = -1; 
    }

    //text to pass to module
    sArgs = GetArgFromCommandLine(argc, argv,  _T("args")); 
    /*copy out 'cuz possible cmd line storage goes away b4 child process done with it ... */
    if(sArgs)
        _tcscpy(sArgBuf, sArgs); 
    else
        *sArgBuf = 0; 
        

    //do we wait or not? boolean
    sWait = GetArgFromCommandLine(argc, argv,  _T("wait"));

    if(sWait)
        bWait = _tcstoul(sWait, &stop, 10); 
    else
        bWait = FALSE; 

    //flags. See header for comments
    sFlags = GetArgFromCommandLine(argc, argv, _T("flags")); 
    if(!sFlags)
        nFlags = LOAD_IMAGE_CALL_ENTRY_POINT | LOAD_IMAGE_CREATE_THREAD; 
    else
    {
        if(!*sFlags)
            nFlags = LOAD_IMAGE_CALL_ENTRY_POINT | LOAD_IMAGE_CREATE_THREAD;
        else
            nFlags = _tcstoul(sFlags, &stop, 16); 
    }

    Status(sMod, sFuncName, nFlags, nOrdinal, _tcstoul(sArgBuf, &stop, 10), nNameBufLen, bWait); 

    //finally ... 
    sc = CurrentProcess()->v->LoadImage(CurrentProcess(), sMod, sArgBuf, nFlags, &pMod); 

    //reporting function
    Outcome(LOADIMAGE, sc, sMod, nFlags, UINT64_ZERO); 

    if(FAILED(sc))
    {
        pHeap->v->Free(pHeap, 0, sName); 
        pHeap->v->Free(pHeap, 0, sArgBuf); 
        pHeap->v->Free(pHeap, 0, sFuncName);
        FreeCmdLine(&argc, &argv);
        return SetAndReturn(1);
    }

    sc = pMod->v->QueryInterface(pMod, &IID_IProcess, (void **) &pSpawn);
    if (FAILED(sc))
        _tprintf(_T("QI for IProcess failed!\n"));

    if(bWait)
    {
        _tprintf(_T("\n\r[%s]Waiting for process to finish ... "), argv[0]); 
        sc = pSpawn->v->WaitFor(pSpawn, TIME_FOREVER); 
        _tprintf(_T("\n\r[%s]WaitFor() returned [%s]"), argv[0], 
            ErrorDescription(sc)); 
    }
    else
        _tprintf(_T("\n\r[%s]NOT waiting for process to finish ...."), argv[0]);

    sc = pMod->v->GetFunctionAddress(pMod, sFuncName, nOrdinal, &FuncAddr); 
    GFAOutcome(GETFUNCTIONADDRESS, sc, sFuncName, nOrdinal, FuncAddr);
    sc = pMod->v->GetArgs(pMod, sArgBuf, sizeof(sArgBuf)); 
    _tprintf(_T("\n\r[%s]GetArgs() returned [%s], args are:[%s]"), argv[0], ErrorDescription(sc), sArgs); 

    sc = pMod->v->GetName(pMod, sName, sizeof(sName)); 
    _tprintf(_T("\n\r[%s]GetName() returned [%s], name is:[%s]"), argv[0], ErrorDescription(sc), sName); 

    sc = pMod->v->GetEntryPoint(pMod, &EntryPoint); 
    _tprintf(_T("\n\r[%s]GetEntryPoint() returned [%s], name [%s], entry point is:[0x%.8x]"), 
        argv[0], 
        ErrorDescription(sc), 
        sName, 
        (UINT32)EntryPoint); 

    nExit = pSpawn->v->GetValue(pSpawn, PROCESS_VALUE_EXIT_CODE); 
    _tprintf(_T("\n\r[%s]GetExitCode() returned exit code is:[%d]"),
             argv[0], nExit); 

    FreeCmdLine(&argc, &argv);
    
    pHeap->v->Free(pHeap, 0, sName);
    pHeap->v->Free(pHeap, 0, sFuncName);
    pHeap->v->Free(pHeap, 0, sArgBuf);
    
    pMod->v->Release(pMod);
    pSpawn->v->Release(pSpawn);
    _tprintf(_T("\n\r[%s] exiting ... "), argv[0]);
    return SetAndReturn(0);
}

/*
*Basic reporting function .. tells luser what values are being passed in .. 
*/
void Status(const _TCHAR * sMod, const _TCHAR * sFunc, UINT32 nFlags, UINT32 nOrdinal, INT32 nExit, UINT32 nBufLen, BOOL bWait)
{
    _TCHAR sNull[] = _T("(null)");
    _tprintf(_T("\n\rModule: [%s], Function:[%s], Ordinal:[%d], Exit Code:[%d], Buffer len:[%d bytes], Wait:[%s]"), 
        sMod ? sMod : sNull, 
        sFunc ? sFunc : sNull, 
        nOrdinal, 
        nExit,
        nBufLen,
        bWait ? _T("TRUE"):_T("FALSE"));
}

void TMUsage(const _TCHAR * s)
{
    _tprintf(_T("\n\rUsage:%s -module <module_name> [-func [<funcname>|<ordinal>]] [-wait N] [-flags <01b|10b>] -buflen <0xNUM>"), s); 
    _tprintf(_T("\n\r<module_name> is name of module to load")); 
    _tprintf(_T("\n\r<funcname> is name of exported function to find, or <ordinal> is it's numeric (base 10) ordinal")); 
    _tprintf(_T("\n\r'-wait N' means wait until process is finished if N is nonzero")); 
    _tprintf(_T("\n\r'-flags N' where N is in 0-3 sets flags to pass to LoadImage()")); 
    _tprintf(_T("\n\r'-buflen N' sets the length of the buffer used to GetName(), GetFunctionAddress(), and GetArgs() to N bytes"));
}

void Error(void)
{
    _tprintf(_T("\n\rError - probably memory allocation - bailing out early."));
}
