/* Copyright (c) Microsoft Corporation. All rights reserved. */

/* Simple tests for RAM memory.  Detects shorts on the address/data lines
 */
#include <stdio.h>
#include <mmlite.h>
#include <tchar.h>
#include <stdlib.h>
#include <string.h>

_TCHAR Usage[] = _T("Usage: %s HexStartAddress DecimalByteCount\r\n");

/* Run a pattern over the entire memory range, fill it up first
 * then checking the actual content in a second pass.
 * Location N should hold the value "PATTERN+N".
 * This is done three times, accessing memory by 32/16/8 bits.
 * Can either stop at the first occurrence of an error, or keep
 * going [which is useful to find *which* lines are actually shorted]
 */
SCODE
MemoryTest1(
           ADDRESS Start,
           UINT nBytes,
           UINT Pattern,
           BOOL StopAtFirstError
)
{
    UINT32 *pW32, A32;
    UINT16 *pW16, A16;
    UINT8 *pW8, A8;
    UINT i;

    /* Test by words, write then readback
     */
#define pattern(_i_)  (Pattern + (_i_))
    _tprintf(_T("By 32 bits at [x%x..x%x] with base pattern %x\r\n"),
             Start,Start+nBytes,Pattern);
    pW32 = (UINT32 *) Start;
    for (i = 0; i < nBytes/4; i++) {
        pW32[i] = pattern(i);
    }

    for (i = 0; i < nBytes/4; i++) {
        A32 = pW32[i];
        if (A32 != pattern(i)) {
            _tprintf(_T("Word[%d] mismatch, got x%x (!= x%x)\r\n"),
                        i, A32, pattern(i));
            if (StopAtFirstError)
                return E_FAIL;
        }
    }
#undef pattern

    /* Test by shorts, write then readback
     */
#define pattern(_i_)  (UINT16)((~Pattern + (_i_)))
    _tprintf(_T("By 16 bits at [x%x..x%x] with base pattern %x\r\n"),
             Start,Start+nBytes,~Pattern);
    pW16 = (UINT16 *) Start;
    for (i = 0; i < nBytes/2; i++) {
        pW16[i] = pattern(i);
    }

    for (i = 0; i < nBytes/2; i++) {
        A16 = pW16[i];
        if (A16 != pattern(i)) {
            _tprintf(_T("Short[%d] mismatch, got x%x (!= x%x)\r\n"),
                        i, A16, pattern(i));
            if (StopAtFirstError)
                return E_FAIL;
        }
    }
#undef pattern


    /* Test by bytes, write then readback
     */
#define pattern(_i_)  (UINT8)((Pattern + (_i_)))
    _tprintf(_T("By  8 bits at [x%x..x%x] with base pattern %x\r\n"),
             Start,Start+nBytes,Pattern);
    pW8 = (UINT8 *) Start;
    for (i = 0; i < nBytes; i ++) {
        pW8[i] = pattern(i);
    }

    for (i = 0; i < nBytes; i ++) {
        A8 = pW8[i];
        if (A8 != pattern(i)) {
            _tprintf(_T("Address %x Byte[%d] mismatch, got x%x (!= x%x)\r\n"),
                        (int)&pW8[i],i, A8, pattern(i));
            if (StopAtFirstError)
                return E_FAIL;
        }
    }
#undef pattern

    return S_OK;
}

/* Run a fixed string around the memory range, copy it then read it back.
 */
static UINT16 Random(void)
{
    TIME Now = CurrentTime();

    return (UINT16)((Uint64Low32Bits(Now) >> 3) ^ 31);
}

SCODE
MemoryTest2(
           ADDRESS Start,
           UINT nBytes,
           UINT Pattern,
           BOOL StopAtFirstError
)
{
    UINT i, l, nTests;
    UINT Offset;
    _TCHAR *Where;

    l = _tcslen(Usage) + 1;

    _tprintf(_T("By strings..."));

    nTests = 0;
    for (i = 0; i < 10240; i++) {
        Offset = Random();

        if ((Offset < nBytes) &&
            ((Offset + l + 3) < nBytes)) {

            nTests += 2;

            Where = (_TCHAR*)Start + Offset;
            _tcsncpy(Where, Usage, l);
            if (_tcsncmp(Where, Usage, l) != 0) {
                _tprintf(_T("Offset %x string mismatch, got '%s'\r\n"),
                         Offset,Where);
                if (StopAtFirstError)
                    return E_FAIL;
            }

            /* Repeat, 3 bytes into it
             */
            Where += 3;
            _tcsncpy(Where, Usage, l);
            if (_tcsncmp(Where, Usage, l) != 0) {
                _tprintf(_T("Offset3 %x string mismatch, got '%s'\r\n"),
                         Offset,Where);
                if (StopAtFirstError)
                    return E_FAIL;
            }

        }
    }
    _tprintf(_T("%d tests executed\r\n"), nTests);
    return S_OK;
}

/* Put random values in random places
 */
SCODE
MemoryTest3(
           ADDRESS Start,
           UINT nBytes,
           UINT Pattern,
           BOOL StopAtFirstError
)
{
    UINT i, j, k, nTests32, nTests16, nTests8;
#define STRIDE 16
    UINT Offset[STRIDE];
    UINT16 Seed[STRIDE];
    UINT32 *pW32, A32;
    UINT16 *pW16, A16;
    UINT8 *pW8, A8;

    _tprintf(_T("By word/short/bytes..."));

    pW32 = (UINT32 *)Start;
    pW16 = (UINT16 *)Start;
    pW8  = (UINT8  *)Start;
    nTests32 = 0;
    nTests16 = 0;
    nTests8 = 0;
    for (i = 0; i < 10240; i++) {
        /* Generate the offset/value pairs
         */
        for (j = 0; j < STRIDE; j++) {
        Retry:
            Offset[j] = Random();
            if ((Offset[j] < nBytes) &&
                ((Offset[j] + 8) < nBytes)) {
                for (k = 0; k < j; k++)
                    if ((Offset[j] & ~3) == (Offset[k] & ~3))
                        goto Retry;
                Seed[j] = Random();
                continue;
            }
            goto Retry;
        }

        /* Insert them values
         */
        for (j = 0; j < STRIDE; j++) {
            k = Offset[j];
            A16 = Seed[j];
            A32 = A16 | (A16 << 16);
            A8 = (UINT8) A16;

            switch (k & 3) {
            case 0:
                nTests32++;
                pW32[ k / 4 ] = A32;
                break;
            case 1:
            case 3:
                nTests8++;
                pW8[ k ] = A8;
                break;
            case 2:
                nTests16++;
                pW16[ k / 2 ] = A16;
                break;
            }
        }

        /* Check
         */
        for (j = 0; j < STRIDE; j++) {
            k = Offset[j];
            A16 = Seed[j];
            A32 = A16 | (A16 << 16);
            A8 = (UINT8) A16;

            switch (k & 3) {
            case 0:
                if (pW32[ k / 4 ] != A32)
                    goto Error;
                break;
            case 1:
            case 3:
                if (pW8[ k ] != A8)
                    goto Error;
                break;
            case 2:
                if (pW16[ k / 2 ] != A16)
                    goto Error;
                break;
            }
            continue;
        Error:
            _tprintf(_T("Offset %x mismatch, got %x not %x\r\n"),
                     k, pW32[ k / 4 ], A32);
            if (StopAtFirstError)
                return E_FAIL;
        }

    }

    _tprintf(_T("%d+%d+%d tests executed\r\n"), nTests32, nTests16, nTests8);
    return S_OK;
}

/* Run all the above tests
 */
SCODE
MemoryTest(
           ADDRESS Start,
           UINT nBytes,
           UINT Pattern,
           BOOL StopAtFirstError
)
{
    SCODE sc = S_OK;
    if (!FAILED(sc)) sc = MemoryTest1(Start,nBytes,Pattern,StopAtFirstError);
    if (!FAILED(sc)) sc = MemoryTest2(Start,nBytes,Pattern,StopAtFirstError);
    if (!FAILED(sc)) sc = MemoryTest3(Start,nBytes,Pattern,StopAtFirstError);
    return sc;
}

/* Driving code. Gets arguments, tries four shuffled patterns.
 */
int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    ADDRESS StartAddress = ~0;
    PTR pMem;
    UINT nBytes = 0;
    _TCHAR *me;
    PIHEAP pHeap = CurrentHeap();

    sc = ParseCmdLine(&argc, &argv);
    me = argv[0];
    if (argc < 3)
        goto Usage;

    StartAddress = _tcstoul(argv[1],NULL,16);
    nBytes       = _tcstoul(argv[2],NULL,10);

    /* Try to play nice
     */
    pMem = pHeap->v->Extract(pHeap,0,(PTR) StartAddress, nBytes);
    if (pMem == NULL) {
        _tprintf(_T("Warning: that memory range might be in use.\r\n"));
    }

    /* Perform the tests
     */
    if (!FAILED(sc)) sc = MemoryTest(StartAddress,nBytes,0x5511eeaa,FALSE);
    if (!FAILED(sc)) sc = MemoryTest(StartAddress,nBytes,0x1155aaee,FALSE);
    if (!FAILED(sc)) sc = MemoryTest(StartAddress,nBytes,0xaaee1155,FALSE);
    if (!FAILED(sc)) sc = MemoryTest(StartAddress,nBytes,0xeeaa5511,FALSE);

    /* Let it go if we took it
     */
    if (pMem != NULL) {
        pHeap->v->Free(pHeap,0, pMem);
    }

 Done:
    FreeCmdLine(&argc,&argv);
    return 0;
 Usage:
    _tprintf(Usage,me);
    goto Done;
}
