/* Copyright (c) Microsoft Corporation. All rights reserved. */
/******************************************************************/
/* constrn2.c - a BeginConstraint/EndConstraint test program      */
/******************************************************************/

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

#ifndef RELATIVE_TIME_MILLIS
#define RELATIVE_TIME_MILLIS(time) (time * -10000)
#endif

#define ITERATIONS      20              /* Loop iterations*/
#define NUM_THREADS     5               /* Number of test threads*/
const struct {
    int Rate;
    int Load;
} Tparm[NUM_THREADS] = {
    { 5, 1000 },
    { 3, 1000 },
    { 7, 1000 },
    { 2, 1000 },
    { 6, 1000 }
};


/* prototypes */
void THREAD_LINKAGE AnyThread(THREAD_ARGUMENT arg);

/* globals */
volatile BOOL bRunFlag = FALSE;
volatile INT nThreadsTerminated = 0;
CONDITION OneThreadIsGone;

struct {
    TIME dtime[ITERATIONS];
    TIME stime[ITERATIONS];
    TIME etime[ITERATIONS];
    TIME elapsed[ITERATIONS];
    BOOL MissedDeadline[ITERATIONS];
} Stats[NUM_THREADS];


/* main() */
int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    PIPROCESS pPrc;
    int i, j;
    int NumThreads = NUM_THREADS;
    int nMissed = 0;

    /* arguments ?*/
    sc = ParseCmdLine(&argc, &argv);
_tprintf(_TEXT("sc %x argc %d argv %x\n\r"), sc, argc, (ADDRESS) argv);
    if (!FAILED(sc) && argc > 1) {
_tprintf(_TEXT("%s %s\n\r"), argv[0], argv[2]);
        NumThreads = _ttoi(argv[2]);
        if (NumThreads <= 0 ||
            NumThreads > NUM_THREADS)
            NumThreads = NUM_THREADS;
    }

    _tprintf(_TEXT("Will run this test with %d concurrent threads\n\r"),
             NumThreads);

    /* Init*/
    Condition_Init(&OneThreadIsGone);

    /* Get current process*/
    pPrc = CurrentProcess();

    /* Spawn all threads*/
    for (j = 0; j < NumThreads; j++) {
        sc = pPrc->v->CreateThread(pPrc, 0,
                                   AnyThread,
                                   (THREAD_ARGUMENT)j,
                                   0,NULL,NULL);
        if(sc != S_OK) {
            _tprintf(_TEXT("Unable to create Thread #%d\n\r"),j+1);
            exit(EXIT_FAILURE);
        }
    }

    /* Set threads running*/
    bRunFlag = TRUE;

    /* Wait for all threads to finish*/
    while(nThreadsTerminated<NumThreads) {
        Condition_Wait(&OneThreadIsGone, NULL, NULL);
    }

    /* Dump all thread's timing data*/
    for (j = 0; j < NumThreads; j++) {
        _tprintf(_TEXT("\n\r***** Thread #%d Data (rate %dms load %d) *****\n\r"),
                 j+1,Tparm[j].Rate,Tparm[j].Load);
        for(i=0;i<ITERATIONS;i++) {
            char *missed = (Stats[j].MissedDeadline[i]) ? "MISSED deadline" :"";
            _tprintf(_TEXT(" Iter#%d Deadline: %lld  Start %lld +%lld= End %lld  Run %lld  Margin %lld %hs\n\r"),i+1,
                     Stats[j].dtime[i],
                     Stats[j].stime[i],
                     TimeSubtract(Stats[j].etime[i],Stats[j].stime[i]),
                     Stats[j].etime[i],
                     Stats[j].elapsed[i],
                     TimeSubtract(Stats[j].dtime[i],Stats[j].etime[i]),
                     missed);
            if (Stats[j].MissedDeadline[i])
                nMissed++;
        }
    }

    if (nMissed)
        _tprintf(_TEXT("%d deadline(s) were missed (out of %d total)\n\r"),
                 nMissed, NumThreads * ITERATIONS);
    _tprintf(_TEXT("This test %hs\n\r"), (nMissed==0) ? "PASSED" : "FAILED");

    (void) FreeCmdLine(&argc,&argv);

    exit ((nMissed == 0)?EXIT_SUCCESS:EXIT_FAILURE);
}

/* FirstThread, SecondThread, ... */
void THREAD_LINKAGE AnyThread(THREAD_ARGUMENT arg)
{
    TIME_CONSTRAINT tc;
    TIME est;
    TIME temptime;
    int i,j,x,me;
    SCODE sc;

    me = (int) arg;

    /* Wait for start flag*/
    while(bRunFlag==FALSE);

    /* Set start time to now + MyRate*/
    Int32ToInt64(temptime,TIME_MILLIS(Tparm[me].Rate));
    tc.Start = Int64Add(CurrentTime(),temptime);

    /* Set deadline to now + MyRate +2ms*/
    Int32ToInt64(temptime,TIME_MILLIS(2));
    tc.Deadline = Int64Add(tc.Start,temptime);

    /* Set initial estimate to 1 msec*/
    Int32ToInt64(tc.Estimate,TIME_MILLIS(1));

    /* Set criticality to NONCRITICAL*/
    tc.Criticality = _NONCRITICAL;

    for(i=0;i<ITERATIONS;i++) {
        /* Get deadline for stats*/
        Stats[me].dtime[i] = tc.Deadline;

        /* Set Loop Constraint*/
        sc = BeginConstraint(FALSE,&tc,NULL);
        Stats[me].stime[i] = CurrentTime();
        if(sc != S_OK) {
#ifdef _DEBUG
            _tprintf(_TEXT("Thread #%d BeginConstraint not S_OK on loop #%d\n\r"),me+1,i+1);
#endif
            Stats[me].MissedDeadline[i] = TRUE;
            /*DebugBreak();*/
        }

        /* do some busy work*/
        x = 1;
        for(j=0;j<Tparm[me].Load;j++) x = (x + x)/2;
        /* end of busy work*/

        /* save end time*/
        Stats[me].etime[i] = CurrentTime();

        /* End loop constraint*/
        sc = EndConstraint(&est);
        if(sc != S_OK) {
#ifdef _DEBUG
            _tprintf(_TEXT("Thread #%d EndConstraint not S_OK on loop #%d\n\r"),me+1,i+1);
#endif
            Stats[me].MissedDeadline[i] = TRUE;
        }

        Stats[me].elapsed[i] = est;

        /* Add My Rate to deadline*/
        Int32ToInt64(temptime,TIME_MILLIS(Tparm[me].Rate));
        tc.Deadline = Int64Add(tc.Deadline,temptime);

#if 1/* fornow make it simple */
#if 0
        Int32ToInt64(est,TIME_MILLIS(1));
#else
#ifdef _DEBUG
        Int32ToInt64(est,TIME_MICROS(800));/*actually, about 700 */
#else
        Int32ToInt64(est,TIME_MICROS(400));/*actually, about 200 */
#endif
#endif
#endif

        /* Set estimate to returned estimate value*/
        tc.Estimate = est;

        /* Set start time to deadline - (estimate * 2)*/
        tc.Start = Int64Subtract(tc.Deadline,Int64TimesInt32(est,2));
    }

    AtomicInc((PUINT)&nThreadsTerminated);
    Condition_Signal(&OneThreadIsGone);
}

