/* Copyright (c) Microsoft Corporation. All rights reserved. */
/******************************************************************/
/* constrnt.c - a BeginConstraint/EndConstraint test program      */
/******************************************************************/

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

#ifndef RELATIVE_TIME_MILLIS
#define RELATIVE_TIME_MILLIS(time) (time * -10000)
#endif

/* prototypes */


/* main() */
int _tmain(int argc, _TCHAR **argv)
{
    TIME_CONSTRAINT tc;
    TIME estimated, actual;
    TIME start,end,temptime;
    TIME stime[20];
    TIME etime[20];
    int i, j, x = 1;
    SCODE sc;
    BOOL Passed = TRUE;

    UnusedParameter(argc);
    UnusedParameter(argv);

    /* Establish initial constraint.
     * This takes us out of time-sharing mode, and lets us enter
     * correctly the loop code below, which uses constraint chaining 
     */

    /* Set:
     *  start time to now + 5ms
     *  deadline to now + 6ms
     *  initial estimate to 0.5ms
     *  criticality to NONCRITICAL
     */

    Int32ToInt64(temptime,TIME_MILLIS(5));
    tc.Start = Int64Add(CurrentTime(),temptime);

    Int32ToInt64(estimated,TIME_MICROS(500));
    tc.Estimate = estimated;
    actual = estimated;

    Int32ToInt64(temptime,TIME_MILLIS(1));
    tc.Deadline = Int64Add(tc.Start,temptime);

    tc.Criticality = _NONCRITICAL;


    /* Make the initial call.  This constraint will only cover
     * the short timespan to the first loop constraint. E.g. it
     * does not account for any actual computation load.
     * That will come out of the first round of the loop.
     */
    sc = BeginConstraint(FALSE,&tc,NULL);
    if(sc != S_OK)
    {
        _tprintf(_TEXT("Initial BeginConstraint not S_OK(x%x)\n\r"), sc);
        Passed = FALSE;
    }

    /* Get current time as start time for the test.
     * This also gives us the reference for the 5ms ticker.
     */
    start = CurrentTime();
    tc.Deadline = start;

    /* Enter the loop. It will immediately chain constraints.
     */
    for(i=0;i<20;i++)
    {
            /* Compute next loop constraint
         * Recompute our estimated time. Use MAX of runtimes.
         */
        if (TimeLess(estimated,actual)) {

            /* We took longer than anticipated.  Correct the estimate,
             * but do not overshoot (sanity check).
             */
            TIME MaxEstimate;
            Int32ToInt64(MaxEstimate,TIME_MILLIS(2));

            if (TimeLess(actual,MaxEstimate))
            estimated = actual;
        }
        tc.Estimate = estimated;

        /* We want to run periodically, every 5msecs.
         * So the next deadline is 5msecs past the previous one.
         */
        Int32ToInt64(temptime,TIME_MILLIS(5));
        tc.Deadline = Int64Add(tc.Deadline,temptime);

        /* We could just say the start time is the previous deadline.
         * But to make it slightly more interesting, we'll set the
         * start time to the deadline - (estimate * 2).
         */
        tc.Start = Int64Subtract(tc.Deadline,Int64TimesInt32(estimated,2));

        /* Set Loop Constraint.  This will sleep for a while.
         */
        sc = BeginConstraint(TRUE,&tc,&actual);
        if(sc != S_OK)
        {
            _tprintf(_TEXT("BeginConstraint not S_OK(x%x) on loop #%d\n\r"),
                   sc, i+1);
            Passed = FALSE;
        }

        /* Save current time for statistics
         */
        stime[i] = CurrentTime();

        /* Save (previous) time taken for statistics
         */
        if (i > 0)
            etime[i-1] = actual;


        /* Do our computational work
         */
        x = 1;
        for(j=0;j<2500;j++) x = (x + x) / 2;

    }

    /* Get end time for totals
     */
    end = CurrentTime();

    /* Terminate the last constraint and finish up with statistics
     */
    EndConstraint(&etime[i-1]);


    /* Calculate overall time taken
     */
    end = Int64Subtract(end,start);

    /* Print start and elapsed times for each loop iteration
     */
    for(i=0;i<20;i++)
    {
        _tprintf(_TEXT("Iter: %d  Start: %lld  Elapsed: %lld\n\r"),
             i+1, stime[i], etime[i]);
    }

    /* Print total time taken
     */
    _tprintf(_TEXT("\n\rTotal time to run 20 iterations: %lld\n\r\n\r"),
         end);

    /* Notify of the outcome
     */
    if (Passed) {
        _tprintf(_TEXT("Constraint test PASSED.\n\r"));
        exit(EXIT_SUCCESS);
    } else {
        _tprintf(_TEXT("Constraint test FAILED.\n\r"));
        exit(EXIT_FAILURE);
    }

    /* Make sure X does not get thrown away by the compiler
     */
    return(x);
}
