/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <tchar.h>

/* Debug */
BOOL hpState(IHeap *pHeap)
{
    ADDR_SIZE r, c, u, m;
    SCODE sc;

    sc = pHeap->v->Status(pHeap, &r, &c, &u, &m);
    if (FAILED(sc)) {
        _tprintf(_TEXT("HeapStatus FAILED, sc = x%x\n\r"), sc);
        return FALSE;
    }
    _tprintf(_TEXT("Heap @%x: Commit x%x Reserve x%x Used x%x MaxUsed x%x\n\r"),
             (ADDRESS) pHeap, c, r, u, m);
    return TRUE;
}

/* Maximum number of successive calls to Alloc() */
#define MAX_NUM_ALLOCS   (0x400)

#if 1
#define ZEROIT HEAP_ZERO_MEMORY
#else
#define ZEROIT 0
#endif

BOOL
TestOne(IHeap *pHeap, PTR *ptr)
{
    UINT ui;
    TIME st, et;
    
    _tprintf(_TEXT("Testing repeated Alloc/Realloc.\n\r"));
    if (!hpState(pHeap))
        return FALSE;
    st = CurrentTime();
    for (ui=0; ui <= 50; ui++) {
        ptr[ui] = pHeap->v->Alloc(pHeap, ZEROIT, 33 + ui, 0);
        if (ptr[ui] == NULL) {
            _tprintf(_TEXT("heaptest: Alloc1(%d) FAILED\n\r"), 33+ui);
            return FALSE;
        }

        if (pHeap->v->Size(pHeap, 0, ptr[ui]) < 33 + ui) {
            _tprintf(_TEXT("heaptest: Size FAILED\n\r"));
            return FALSE;
        }

        ptr[ui] = pHeap->v->Realloc(pHeap, ZEROIT, ptr[ui],
                                    66 + ui, 0);
        if (ptr[ui] == NULL) {
            _tprintf(_TEXT("heaptest: Realloc() FAILED\n\r"));
            return FALSE;
        }

        if (pHeap->v->Size(pHeap, 0, ptr[ui]) < 66 + ui) {
            _tprintf(_TEXT("heaptest: Size() FAILED\n\r"));
            return FALSE;
        }

        if (FAILED(pHeap->v->Validate(pHeap, 0, ptr[ui])) ||
            FAILED(pHeap->v->Validate(pHeap, 0, NULL))) {
            _tprintf(_TEXT("heaptest: Validate() FAILED\n\r"));
            return FALSE;
        }
    }
    
    for (ui=0; ui <= 50; ui++) {
        if (!pHeap->v->Free(pHeap, 0, ptr[ui])) {
            _tprintf(_TEXT("heaptest: Free() FAILED\n\r"));
            return FALSE;
        }

        if (FAILED(pHeap->v->Validate(pHeap, 0, NULL))) {
            _tprintf(_TEXT("heaptest: Validate() FAILED\n\r"));
            return FALSE;
        }
    }
    et = CurrentTime();
    if (!hpState(pHeap))
        return FALSE;
    et = TimeSubtract(et,st);
    et = Int64DividedByInt32(et,10);
    _tprintf(_TEXT("heaptest: test one completed in %lld usec\n\r"), et);
    return TRUE;
}

BOOL
TestTwo(IHeap *pHeap, PTR *ptr, UINT *SpaceAvailable)
{
    UINT Count;
    TIME st, et;

    _tprintf(_TEXT("Test Alloc until heap exhaustion.\n\r"));
    if (!hpState(pHeap))
        return FALSE;
    *SpaceAvailable = 0;

    st = CurrentTime();
    for (Count = 1; Count < MAX_NUM_ALLOCS; Count++) {
        ptr[Count] = pHeap->v->Alloc(pHeap, 0, Count, 0);
        if (ptr[Count] == 0) {
            _tprintf(_TEXT("heaptest: Alloc(#%d) failed, with %d bytes allocated\n\r"),
                     Count, (Count * (Count + 1)) / 2 );
            break;
        }
        *SpaceAvailable += Count;
    }

    for (Count-- ; Count > 0; Count--) {
        if (!pHeap->v->Free(pHeap, 0, ptr[Count])) {
            _tprintf(_TEXT("heaptest: Free() 0x%x FAILED\n\r"), Count);
            return FALSE;
        }
    }
    et = CurrentTime();
    if (!hpState(pHeap))
        return FALSE;
    et = TimeSubtract(et,st);
    et = Int64DividedByInt32(et,10);
    _tprintf(_TEXT("heaptest: test two completed in %lld usec\n\r"), et);
    return TRUE;
}

int _tmain()
{
    PTR *ptr = NULL;
    ADDR_SIZE blkSize = MAX_NUM_ALLOCS*sizeof(PTR);
    UINT SpaceAvailable;
    int ret = EXIT_SUCCESS;
    SCODE sc;

    IHeap *pHeap = CurrentHeap(), *pTestHeap = NULL;

    /* Allocate an array of pointers as workspace
     */
    ptr = (PTR *)pHeap->v->Alloc(pHeap, 0, blkSize, 0);
    if (ptr == NULL) {
        _tprintf(_TEXT("heaptest:  Alloc(%d) FAILED\n\r"), blkSize);
        ret = EXIT_FAILURE;
        goto Out;
    }
    /* Double check it
     */
    if (pHeap->v->Size(pHeap, 0, ptr) < blkSize) {
        _tprintf(_TEXT("heaptest: Size FAILED\n\r"));
        ret = EXIT_FAILURE;
        goto Out;
    }

    /* Run the tests on the main heap
     */

    _tprintf(_TEXT("Testing default process heap.\n\r"));
    if (!TestOne(pHeap, ptr)) {
        ret = EXIT_FAILURE;
        goto Out;
    }
    if (!TestTwo(pHeap, ptr, &SpaceAvailable)) {
        ret = EXIT_FAILURE;
        goto Out;
    }

    /* Ditto, on a newly created heap.
     * Make the heap as big as TestTwo could allocate.
     */
    /* XXX Guess that the fragmentation allows us to get 1/2 of the available heap */
    SpaceAvailable /= 2;
    _tprintf(_TEXT("Testing a new heap size %d.\n\r"), SpaceAvailable);
    sc = pHeap->v->CreateHeap(pHeap, 0, 0, SpaceAvailable, &pTestHeap);
    if (FAILED(sc)) {
        _tprintf(_TEXT("CreateHeap new heap FAILED avail=%x\n\r"),
                 SpaceAvailable);
        ret = EXIT_FAILURE;
        goto Out;
    }
    if (!TestOne(pTestHeap, ptr)) {
        ret = EXIT_FAILURE;
        goto Out;
    }
    if (!TestTwo(pTestHeap, ptr, &SpaceAvailable)) {
        ret = EXIT_FAILURE;
        goto Out;
    }

Out:
    /* Destroy the temp heap
     */
    if (pTestHeap)
        pTestHeap->v->Release(pTestHeap);

    /* Release workspace
     */
    if (ptr)
        if (!pHeap->v->Free(pHeap, 0, ptr)) {
            _tprintf(_TEXT("heaptest: Free() FAILED\n\r"));
            ret = EXIT_FAILURE;
        }

    /* Done.
     */
    if (ret == EXIT_SUCCESS)
        _tprintf(_TEXT("Tests of heap functions PASSED.\n\r"));
    else
        _tprintf(_TEXT("Tests of heap functions FAILED.\n\r"));
    exit(ret);
}
