/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <assert.h>

/* Maximum number of successive calls to Alloc() */
#define MAX_NUM_ALLOCS   (0xf0) /* 0x100 */

/* Maximum heap size
 * This is defined by the above, since we'll allocate N blocks each 
 * of size N words, then shrink them in half, then realloc them to
 * be of size 2N. Account for overhead and the sum of the series gives
 * 
 *      Size = N(N+1)/2 + ON
 *
 * where O is the per-block overhead. Assume O=2 words.
 */
#define MAXSIZE (((MAX_NUM_ALLOCS+5) * MAX_NUM_ALLOCS) * 4)

/* Minimum (arbitrary) size of created heap */
#define MINSIZE (4 * 1024)

/* Find the size of the largest free block
 * Dont go beyond ENOUGHMEM (screws up stats)
 */
BOOL FindMaxSize(ADDR_SIZE * pResult, ADDR_SIZE EnoughMem)
{
    IHeap *pPH = CurrentHeap();
    ADDR_SIZE Avail, Used;
    SCODE sc;
    PTR pBlk;

    sc = pPH->v->Status(pPH, NULL, &Avail, &Used, NULL);
    if (FAILED(sc)) {
        _tprintf(_TEXT("heaptstc:  Heap::Status FAILED sc = x%x\n\r"), sc);
        return FALSE;
    }

    /* Find out now how much of that memory is contiguous
     */
    Avail -= Used;    
    if (Avail > EnoughMem)
        Avail = EnoughMem;
    for (;;) {
        if (Avail < MINSIZE) {
            _tprintf(_TEXT("heaptstc: FAILED, not enough memory (%d)\n\r"),
                     Avail);
            return FALSE;
        }
        pBlk = pPH->v->Alloc(pPH, 0, Avail, 0);
        if (pBlk != NULL) {
            /* Found the maximum contig free block
             * Assume noone else will take it...
             */
            (void) pPH->v->Free(pPH,0,pBlk);
            break;
        }
        /* Too big, shrink
         */
        Avail -= MINSIZE;
    }

    *pResult = Avail;
    return TRUE;
}

/* Reverse the formula above for size = f(number of allocs);
 */
UINT FindMaxAllocs( ADDR_SIZE HeapSize )
{
    /* Assuming O == 2 we get S = N(N+5)/2 or  N^2+5N-2S=0
     * therefore the positive N is at N = (sqrt(25+8S)-5)/2
     * We dont have sqrt(), but we dont need to be fast.
     */
    UINT x, lo, hi, i;

    HeapSize = 25 + (HeapSize * 8);

    /* Find highest bit set, add one and sqrt -> upper limit.
     */
    i = 0;
    x = HeapSize;
    for (; x; i++)
        x >>= 1;
    if (i & 1) i++;
    i >>= 1;

    /* Binsearch-multiply until we get it. Dumb, yes.
     */
    hi = 1 << i;
    lo = 1 << (i -3);
    assert( (hi * hi >= HeapSize)
           && (lo * lo <= HeapSize));
    for (;;) {
        x = (hi + lo) / 2;
        i = x * x;
        if (i > HeapSize)
            hi = x;
        else if (i < HeapSize)
            lo = x;
        else
            lo = hi = x;
        if ((hi - lo) < 2)
            break;
    }
#if 0
    /* this is the right number. But the alignment screws it up
     */
    return (lo - 5) / 2;
#else
    return ((lo - 5) / 2) - (_DCACHELINESIZE-1);
#endif
}

void el(TIME ts, PIHEAP pHeap)
{
    TIME te = CurrentTime();
    te = TimeSubtract(te,ts);
    te = Int64DividedByInt32(te,10);
    _tprintf(_TEXT("heaptstc: %lld usecs so far..\n\r"), te);
#if 0
    /* debug dump internal structs */ // XXX most heaps barf
    if (pHeap)
        (void) pHeap->v->Validate(pHeap, 0, (PTR)1);
#endif
}

/* Heap cache-alignment test:  Test use of HEAP_CACHE_ALIGN flag with
 * IHeap::Alloc().
 */
int _tmain()
{
    PTR *ptr;
    ADDR_SIZE blkSize = (1+MAX_NUM_ALLOCS)*sizeof(PTR);
    int i, initsize;
    UINT alignment, align_mask;
    IHeap *pHeap;
    IHeap *pPH = CurrentHeap();
    ADDR_SIZE MaxSize;
    int MaxAllocs;
    TIME ts;
    SCODE sc;

    ts = CurrentTime();

    /* Allocate an array of pointers as workspace
     */
    ptr = (PTR *)pPH->v->Alloc(pPH, 0, blkSize, 0);
    if (ptr == NULL) {
        _tprintf(_TEXT("heaptest:  Alloc() FAILED\n\r"));
        exit(EXIT_FAILURE);
    }

    /* Allocate minimum-size cache-aligned block.  The block size
     * equals the cache-line size.
     */
    ptr[0] = (PTR)pPH->v->Alloc(pPH, HEAP_CACHE_ALIGN, (UINT)1, 0);
    alignment = pPH->v->Size(pPH, 0, ptr[0]);  /* cache-line size */
    if (alignment == _DCACHELINESIZE) {
        _tprintf(_TEXT("heaptstc:  Cache-line size is %d bytes.\n\r"),
                 alignment);
    } else {
        _tprintf(_TEXT("heaptstc:  WARNING: Cache-line size is %d bytes and not %d!\n\r"),
                 alignment,_DCACHELINESIZE);
    }
    align_mask = alignment - 1;
    if (alignment & align_mask) {
        _tprintf(_TEXT("heaptstc:  FAILED %d is not a power of two.\n\r"),
                alignment);
        exit(EXIT_FAILURE);
    }
    if (!pPH->v->Free(pPH, 0, ptr[0])) {
        _tprintf(_TEXT("heaptstc:  FAILED Unable to free aligned block.\n\r"));
        exit(EXIT_FAILURE);
    }

    /* Find out a sensible maximum size to use.
     * With that goes a maximum number of allocs.
     */
    if (!FindMaxSize(&MaxSize, MAXSIZE))
        exit(EXIT_FAILURE);
    el(ts, pPH);
    MaxAllocs = FindMaxAllocs(MaxSize/8);
    assert(MaxAllocs <= MAX_NUM_ALLOCS);

    _tprintf(_TEXT("heaptstc:  testing with MaxSize %d and MaxAllocs %d.\n\r"),
             MaxSize, MaxAllocs);

    /* Run test of alternating aligned and unaligned allocs and reallocs.
     * Use a variety of heaps created with different values of InitialSize.
     */
    for (initsize = MaxSize; initsize >= 0; initsize -= MaxSize>>3) {

        /* Create a new heap.
         */
        _tprintf(_TEXT("heaptstc:  Creating heap with InitialSize = %d bytes.\n\r"),
                         initsize);
        sc = pPH->v->CreateHeap(pPH, 0, initsize, MaxSize, &pHeap);
        if (FAILED(sc)) {
            _tprintf(_TEXT("heaptstc:  FAILED to create heap sc=%x.\n\r"), sc);
            exit(EXIT_FAILURE);
        }

        /* If loop count i is odd, allocate a cache-aligned block of
         * memory.  Otherwise, allocate a block that is not cache-aligned.
         */
        el(ts, NULL);
        _tprintf(_TEXT("heaptstc:  Alloc() mix of aligned/unaligned blocks.\n\r"));
        for (i = 1; i < MaxAllocs; ++i) {
            if (i & 1) {
                /* Allocate cache-aligned block and verify its alignment.
                 */
                ptr[i] = (PTR)pHeap->v->Alloc(pHeap,
                                           HEAP_ZERO_MEMORY | HEAP_CACHE_ALIGN,
                                           i << 2,
                                           0);
                if (!ptr[i]) {
                    _tprintf(_TEXT("heaptstc:  Aligned Alloc() %d FAILED.\n\r"), i);
                    exit(EXIT_FAILURE);
                }
                alignment = ((UINT)ptr[i]) | pHeap->v->Size(pHeap, 0, ptr[i]);
                if (alignment & align_mask) {
                    _tprintf(_TEXT("heaptstc: FAILED Alloc() %d is badly aligned.\n\r"), i);
                    exit(EXIT_FAILURE);
                }
            } else {
                /* Allocate a block that is not cache-aligned.
                 */
                ptr[i] = (PTR)pHeap->v->Alloc(pHeap, HEAP_ZERO_MEMORY, i<<2,0);
                if (!ptr[i]) {
                    _tprintf(_TEXT("heaptstc:  Unaligned Alloc() %d FAILED.\n\r"), i);
                    exit(EXIT_FAILURE);
                }
            }
        }

        /* Reallocate allocated blocks to shrink them by roughly one half
         * If loop count i is odd, reallocate a cache-aligned block of
         * memory.  Otherwise, allocate a block that is not cache-aligned.
         */
        el(ts,pHeap);
        _tprintf(_TEXT("heaptstc:  Use Realloc() to shrink alloc'd blocks.\n\r"));
        for (i = 1; i < MaxAllocs; ++i) {
            if (i & 1) {
                /* ReAlloc() cache-aligned block and verify its alignment.
                 */
                ptr[i] = (PTR)pHeap->v->Realloc(pHeap,
                                           HEAP_ZERO_MEMORY | HEAP_CACHE_ALIGN,
                                           ptr[i],
                                           i<<1,
                                           0);
                if (!ptr[i]) {
                    _tprintf(_TEXT("heaptstc:  Aligned shrink %d FAILED.\n\r"), i);
                    exit(EXIT_FAILURE);
                }
                alignment = ((UINT)ptr[i]) | pHeap->v->Size(pHeap, 0, ptr[i]);
                if (alignment & align_mask) {
                    _tprintf(_TEXT("heaptstc:  FAILED Realloc() %d is badly aligned (x%x).\n\r"), i, alignment);
                    exit(EXIT_FAILURE);
                }
            } else {
                /* Realloc() a block that is not cache-aligned.
                 */
                ptr[i] = (PTR)pHeap->v->Realloc(pHeap, HEAP_ZERO_MEMORY, ptr[i], i<<1, 0);
                if (!ptr[i]) {
                    _tprintf(_TEXT("heaptstc:  Unaligned shrink %d FAILED.\n\r"), i);
                    exit(EXIT_FAILURE);
                }
            }
        }

        /* Reallocate allocated blocks to expand them by a factor of roughly of
         * four.  If loop count i is odd, reallocate a cache-aligned block
         * memory.  Otherwise, allocate a block that is not cache-aligned.
         */
        el(ts,pHeap);
        _tprintf(_TEXT("heaptstc: Use Realloc() to expand alloc'd blocks.\n\r"));
        for (i = 1; i < MaxAllocs; ++i) {
            if (i & 1) {
                /* Realloc() cache-aligned block and verify its alignment.
                 */
                ptr[i] = (PTR)pHeap->v->Realloc(pHeap,
                                           HEAP_ZERO_MEMORY | HEAP_CACHE_ALIGN,
                                           ptr[i],
                                           i<<3,
                                           0);
                if (!ptr[i]) {
                    _tprintf(_TEXT("heaptstc:  Aligned expand %d FAILED.\n\r"),
                             i);
                    exit(EXIT_FAILURE);
                }
                alignment = ((UINT)ptr[i]) | pHeap->v->Size(pHeap, 0, ptr[i]);
                if (alignment & align_mask) {
                    _tprintf(_TEXT("heaptstc:  FAILED Realloc() %d is badly aligned (x%x).\n\r"), i, alignment);
                    exit(EXIT_FAILURE);
                }
            } else {
                /* Realloc() a block that is not cache-aligned.
                 */
                ptr[i] = (PTR)pHeap->v->Realloc(pHeap, HEAP_ZERO_MEMORY, ptr[i], i<<3, 0);
                if (!ptr[i]) {
                    _tprintf(_TEXT("heaptstc: Unaligned expand %d FAILED.\n\r"),
                             i);
                    exit(EXIT_FAILURE);
                }
            }
        }

        el(ts,pHeap);
        /* Free all the aligned and unaligned blocks alloc'd above.
         */
        ptr[i] = NULL;    /* <--- I hope you allocated space for this */
        for (i = 1; i < MaxAllocs && ptr[i]; ++i) {
            if (!pHeap->v->Free(pHeap, 0, ptr[i])) {
                _tprintf(_TEXT("heaptstc:  Free() %d FAILED.\n\r"), i);
                exit(EXIT_FAILURE);
            }
        }

        /* Destroy the temp heap.
         */
        _tprintf(_TEXT("heaptstc:  About to release heap 0x%x\n\r"),
                 (ADDRESS) pHeap);
        pHeap->v->Release(pHeap);
    }

    el(ts,NULL);
    /* Release workspace.
     */
    if (!pPH->v->Free(pPH, 0, ptr)) {
        _tprintf(_TEXT("heaptest: Free() FAILED\n\r"));
        exit(EXIT_FAILURE);
    }

    /* Done.
     */
    _tprintf(_TEXT("Cache-aligned heap test PASSED.\n\r\n\r"));
    el(ts,NULL);
    exit(EXIT_SUCCESS);
}

