/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Test operation of IHeap::Extract().
 */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <tchar.h>

/* Test-case control parameters */
UINT NUM_STEPS =        (128);
#define STEP_SIZE      (1024)
#define STEP_MARGIN    (128)
#define HEAP_SIZE      (NUM_STEPS*STEP_SIZE+3*1024)
#define ALLOC_SIZE     (HEAP_SIZE - 3*1024)


PIHEAP CreateSubHeap(PIHEAP Heap, UINT Flags, UINT InitialSize, UINT MaxSize)
{
    PIHEAP NewHeap;
    SCODE sc;
    sc = Heap->v->CreateHeap(Heap, Flags, InitialSize, MaxSize, &NewHeap);
    if (FAILED(sc))
        return NULL;
    return NewHeap;
}

/* Adjust our params to match the available free memory
 * Reduces NUM_STEPS until we think its enough.
 */
BOOL AdjustNumSteps(PIHEAP pHeap)
{
    PTR Ptr;
    PIHEAP TestHeap;
    
    for (;NUM_STEPS > 2;NUM_STEPS--) {

        /* Alloc them pointers, check
         */
        Ptr = (PTR *)pHeap->v->Alloc(pHeap, 0, NUM_STEPS*sizeof(PTR), 0);
        if (Ptr == NULL)
            continue; /* Boy we are tight */

        /* Create the work heap, then destroy it.
         */
        TestHeap = CreateSubHeap(pHeap, 0, 0, HEAP_SIZE);
        if (TestHeap != NULL)
            TestHeap->v->Release(TestHeap);

        /* Free them pointers
         */
        (void) pHeap->v->Free(pHeap, 0, Ptr);

        /* Check if we got a valid heap
         */
        if (TestHeap != NULL) {
            /* Got a good one
             */
            return TRUE;
        }

        /* Nope, keep trying
         */
    }

    /* Nope */
    return FALSE;
}

int _tmain()
{
    UINT ui;
    ADDR_SIZE size;
    BOOL status;
    PTR *ptr;
    IHeap *pPH = CurrentHeap(), *pHeap;
    char *pbase, *pblk, *pstart, *pend, *pout;

    _tprintf(_TEXT("heaptstx: checking contiguous free memory...\n\r"));
    if (!AdjustNumSteps(pPH)) {
        _tprintf(_TEXT("heaptstx: Not enough memory to run.\n\r"));
        goto Done;
    }
    _tprintf(_TEXT("heaptstx: NUM_STEPS=%d HEAP_SIZE=%d.\n\r"),
             NUM_STEPS, HEAP_SIZE);

    /* Allocate an array of pointers as workspace.
     */
    ptr = (PTR *)pPH->v->Alloc(pPH, 0, NUM_STEPS*sizeof(PTR), 0);
    if (!ptr) {
        _tprintf(_TEXT("heaptstx:  Alloc1() FAILED\n\r"));
        exit(EXIT_FAILURE);
    }

    /* Create a new heap.
     */
    pHeap = CreateSubHeap(pPH, 0, 0, HEAP_SIZE);
    if (!pHeap) {
        _tprintf(_TEXT("heaptstx: CreateHeap() FAILED\n\r"));
        exit(EXIT_FAILURE);
    }

    /* Find base address of large, contiguous block within heap.
     */
    pbase = (char *)pHeap->v->Alloc(pHeap, 0, ALLOC_SIZE, 0);
    if (!pbase) {
        _tprintf(_TEXT("heaptstx: Alloc2(%d) FAILED\n\r"), ALLOC_SIZE);
        exit(EXIT_FAILURE);
    }
    pHeap->v->Free(pHeap, 0, pbase);

    /* Attempt to extract a series of blocks of size STEP_SIZE
     * or less from the heap.  All Extract() calls should SUCCEED.
     */
    pblk = pbase;
    for (ui = 0; ui < (UINT) NUM_STEPS; ++ui, pblk += STEP_SIZE) {
        pstart = pblk + (~ui & 15);
        pend = pblk + STEP_SIZE - STEP_MARGIN - ((~ui >> 4) & 15);
        size = pend - pstart;
        ptr[ui] = pHeap->v->Extract(pHeap, 0, (PTR)pstart, size);
        if (!ptr[ui]) {
            _tprintf(_TEXT("heaptstx: Extract() FAILED\n\r"));
            pPH->v->Free(pPH, 0, ptr);
            pHeap->v->Release(pHeap);
            exit(EXIT_FAILURE);
        }
        size = pHeap->v->Size(pHeap, 0, ptr[ui]);
        if ((char *)ptr[ui] > pstart || (char *)ptr[ui] + size < pend) {
            _tprintf(_TEXT("heaptstx: Size FAILED\n\r"));
            pPH->v->Free(pPH, 0, ptr);
            pHeap->v->Release(pHeap);
            exit(EXIT_FAILURE);
        }
    }

    /* Attempt to extract a series of blocks of size STEP_SIZE
     * or less from the heap.  All Extract() calls should FAIL.
     */
    pblk = pbase;
    for (ui = 0; ui < NUM_STEPS; ++ui, pblk += STEP_SIZE) {
        pstart = pblk + (ui & 15);
        pend = pblk + STEP_SIZE - STEP_MARGIN - ((ui >> 4) & 15);
        size = pend - pstart;
        pout = (char *)pHeap->v->Extract(pHeap, 0, (PTR)pstart, size);
        if (pout) {
            _tprintf(_TEXT("heaptstx: Extract() FAILED to fail\n\r"));
            pPH->v->Free(pPH, 0, ptr);
            pHeap->v->Release(pHeap);
            exit(EXIT_FAILURE);
        }
        status = pHeap->v->Free(pHeap, 0, ptr[ui]);
        if (!status) {
            _tprintf(_TEXT("heaptstx: Free() FAILED\n\r"));
            pPH->v->Free(pPH, 0, ptr);
            pHeap->v->Release(pHeap);
            exit(EXIT_FAILURE);
        }
    }

    /* Destroy the temp heap.
     */
    pHeap->v->Release(pHeap);

    /* Release workspace alloc'd from process heap.
     */
    if (!pPH->v->Free(pPH, 0, ptr)) {
        _tprintf(_TEXT("heaptstx: Free() FAILED\n\r"));
        exit(EXIT_FAILURE);
    }

    /* Done.
     */
 Done:
    _tprintf(_TEXT("This test PASSED.\n\r"));
    exit(EXIT_SUCCESS);
}
