/* Tests for src/base/kconstraint2.c
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#define SIM 1

#include <stdio.h>
#include <mmlite.h>
#include <base/mutex2.h>
#include <base/thread2.h>
#include <base/kconstraint2.h>
#include <util/test.h>
#include <util/fake_sched.h>

PRIVATE KCONSTRAINT kc1;
PRIVATE KCONSTRAINT kc2;
PRIVATE KCONSTRAINT kc3;
PRIVATE PKCONSTRAINT pkc1 = &kc1;
PRIVATE PKCONSTRAINT pkc2 = &kc2;
PRIVATE PKCONSTRAINT pkc3 = &kc3;
PRIVATE THDINFO Thd1;
PRIVATE THDINFO Thd2;
PRIVATE THDINFO Thd3;
PRIVATE PTHDINFO pThd1 = &Thd1;
PRIVATE PTHDINFO pThd2 = &Thd2;
PRIVATE PTHDINFO pThd3 = &Thd3;
PRIVATE MUTEX mMutex1;
PRIVATE PMUTEX pMX1 = &mMutex1;
struct IScheduler* pTheScheduler;

BOOL KConstraint_TestInitializer(void);
BOOL Test_KConstraint_Init(void);
BOOL Test_KConstraint_SchedObject(void);
BOOL Test_KConstraint_InitialRestartBy(void);
BOOL Test_KConstraint_LastRefresh(void);
BOOL Test_KConstraint_LocalUrgency(void);
BOOL Test_KConstraint_CurrentUrgency(void);
BOOL Test_KConstraint_Pending(void);
BOOL Test_KConstraint_StackNext(void);
BOOL Test_KConstraint_Push(void);
BOOL Test_KConstraint_Pop(void);
BOOL Test_KConstraint_StartInheritance(void);
BOOL Test_KConstraint_RecalculateInheritance(void);
BOOL Test_KConstraint_UrgencyGT(void);
BOOL Test_KConstraint_UrgencyGTEQ(void);

void ClearKConstraintStack(PKCONSTRAINT pkc);

void THREAD_LINKAGE Mutex_LockThread(THREAD_ARGUMENT arg);

int _tmain(void) {
  TestFunctionPair KConstraintTests[] = {
    {KConstraint_TestInitializer,            ""},
    {Test_KConstraint_Init,                   "Init"},
    {Test_KConstraint_SchedObject,            "SchedObject"},
    {Test_KConstraint_InitialRestartBy,       "InitialRestartBy"},
    {Test_KConstraint_LastRefresh,            "LastRefresh"},
    {Test_KConstraint_LocalUrgency,           "LocalUrgency"},
    {Test_KConstraint_CurrentUrgency,         "CurrentUrgency"},
    {Test_KConstraint_Pending,                "Pending"},
    {Test_KConstraint_StackNext,              "StackNext"},
    {Test_KConstraint_Push,                   "Push"},
    {Test_KConstraint_Pop,                    "Pop"},
    {Test_KConstraint_StartInheritance,       "StartInheritance"},
    {Test_KConstraint_RecalculateInheritance, "RecalculateInheritance"},
    {Test_KConstraint_UrgencyGT,              "UrgencyGT"},
    {Test_KConstraint_UrgencyGTEQ,            "UrgencyGTEQ"}
  };

  pTheNameSpace = CurrentNameSpace();

  if (pTheNameSpace->v->Bind(pTheNameSpace, _TEXT("cb_sched2"),
			     0, (PIUNKNOWN*)&pTheScheduler) != S_OK) {
    printf("Failed to retrieve scheduler.");
    return 1;
  }

  RunTestSuite("Kernel Constraint Test Suite", KConstraintTests,
	       GetTestFunctionArraySize(KConstraintTests));

  return 0;
}
    
BOOL KConstraint_TestInitializer(void)
{
  KConstraint_Init(pkc1, NULL);
  KConstraint_Init(pkc2, NULL);
  KConstraint_Init(pkc3, NULL);

  ClearKConstraintStack(pkc1);
  ClearKConstraintStack(pkc2);
  ClearKConstraintStack(pkc3);

  Mutex_Init(pMX1);

  Thread_Init(pThd1);
  Thread_Init(pThd2);
  Thread_Init(pThd3);

  return TRUE;
}

BOOL Test_KConstraint_Init(void)
{
  BOOL success = TRUE;

  TIME tNow = CurrentTime();

  success &= compare_integer(NULL, KConstraint_GetSchedObject(pkc1),
			     "sched object is initially NULL");

  success &=
    compare_integer(TRUE, !TimeLess(tNow,
				    KConstraint_GetInitialRestartBy(pkc1)),
		    "initial restart by is later or equal to current time");

  success &= compare_integer(TRUE,
			     TimeEqual(KConstraint_GetInitialRestartBy(pkc1),
				       KConstraint_GetLastRefresh(pkc1)),
			     "initial restart by == last refresh");

  success &=
    compare_integer(TRUE,
		    TimeEqual(Urgency_GetRestartBy(
				  KConstraint_GetLocalUrgency(pkc1)),
			      Urgency_GetRestartBy(
				  KConstraint_GetCurrentUrgency(pkc1))),
		    "local restart by == current restart by");

  success &= compare_integer(Urgency_GetCriticality(
				 KConstraint_GetLocalUrgency(pkc1)),
			     Urgency_GetCriticality(
				 KConstraint_GetCurrentUrgency(pkc1)),
			     "local criticality == current criticality");

  success &= compare_integer(TRUE, Urgency_IsTimeSharing(
				 KConstraint_GetLocalUrgency(pkc1)),
			     "local urgency is initially timesharing");

  success &=
    compare_integer(TRUE, !TimeLess(tNow,
				    Urgency_GetRestartBy(
					KConstraint_GetLocalUrgency(pkc1))),
		    "local restart by is later than or equal to current time");

  success &= compare_integer(NULL, KConstraint_GetPending(pkc1),
			     "pending initially NULL");

  success &= compare_integer(NULL, KConstraint_GetStackNext(pkc1),
			     "stack next initially NULL");

  success &=
    compare_integer(TRUE, Node_IsAlone(KConstraint_GetSchedQueueNode(pkc1)),
		    "sched queue node is initially alone");

  success &=
    compare_integer(TRUE, Node_IsAlone(KConstraint_GetWaitQueueNode(pkc1)),
		    "wait queue node is initially alone");

  success &=
    compare_integer(pkc1, Node_GetData(KConstraint_GetSchedQueueNode(pkc1)),
		    "sched queue node data == pkc1");

  success &=
    compare_integer(pkc1, Node_GetData(KConstraint_GetWaitQueueNode(pkc1)),
		    "wait queue node data == pkc1");

  return success;
}

BOOL Test_KConstraint_SchedObject(void)
{
  BOOL success = TRUE;

  KConstraint_SetSchedObject(pkc1, (PTR)0x2222);

  success &= compare_integer(0x2222, KConstraint_GetSchedObject(pkc1),
			     "set sched object == get sched object");

  return success;
}

BOOL Test_KConstraint_InitialRestartBy(void)
{
  BOOL success = TRUE;

  TIME tNow = CurrentTime();

  KConstraint_SetInitialRestartBy(pkc1, tNow);

  success &=
    compare_integer(TRUE,
		    TimeEqual(KConstraint_GetInitialRestartBy(pkc1),
			      tNow),
		    "set initial restart by == get initial restart by");
    
  return success;
}

BOOL Test_KConstraint_LastRefresh(void)
{
  BOOL success = TRUE;
  
  TIME tNow = CurrentTime();
  
  KConstraint_SetLastRefresh(pkc1, tNow);
  
  success &= compare_integer(TRUE, TimeEqual(KConstraint_GetLastRefresh(pkc1),
			     tNow),
			     "set last refresh == get last refresh");

  return success;
}

BOOL Test_KConstraint_LocalUrgency(void)
{
  BOOL success = TRUE;

  TIME tNow = CurrentTime();
    
  KConstraint_SetRestartBy(pkc1, tNow);
  KConstraint_SetCriticality(pkc1, _CRITICAL);

  success &=
    compare_integer(TRUE, TimeEqual(Urgency_GetRestartBy(
					KConstraint_GetLocalUrgency(pkc1)),
				    tNow),
		    "set local restartby = get local restart by");

  success &= compare_integer(TRUE, Urgency_IsCritical(
				 KConstraint_GetLocalUrgency(pkc1)),
			     "set local criticality == get local criticality");

  return success;
}

BOOL Test_KConstraint_CurrentUrgency(void)
{
  BOOL success = TRUE;

  TIME tNow = CurrentTime();
  KConstraint_SetRestartBy(pkc1, tNow);
  KConstraint_SetCriticality(pkc1, _CRITICAL);

  Urgency_Copy(KConstraint_GetCurrentUrgency(pkc1),
	       KConstraint_GetLocalUrgency(pkc1));

  success &= compare_integer(TRUE, TimeEqual(KConstraint_GetRestartBy(pkc1),
					     tNow),
			     "set current restartby = get current restart by");

  success &=
    compare_integer(_CRITICAL, KConstraint_GetCriticality(pkc1),
		    "set current criticality == get current criticality");

  return success;
}

BOOL Test_KConstraint_Pending(void)
{
  BOOL success = TRUE;

  success &= compare_integer(FALSE, KConstraint_HasPending(pkc1),
			     "has pending FALSE initially");

  KConstraint_SetPending(pkc1, (PTIME_CONSTRAINT)0x2222);

  success &= compare_integer(TRUE, KConstraint_HasPending(pkc1),
			     "has pending is TRUE after being set");

  success &= compare_integer(0x2222, KConstraint_GetPending(pkc1),
			     "set pending == get pending");

  return success;
}

BOOL Test_KConstraint_StackNext(void)
{
  BOOL success = TRUE;

  success &= compare_integer(FALSE, KConstraint_HasStackNext(pkc1),
			     "has stack next is FALSE initially");

  KConstraint_SetStackNext(pkc1, (PKCONSTRAINT)0x2222);

  success &= compare_integer(TRUE, KConstraint_HasStackNext(pkc1),
			     "has stack next is TRUE after being set");

  success &= compare_integer(0x2222, KConstraint_GetStackNext(pkc1),
			     "set stack next == get stack next");

  return success;
}

BOOL Test_KConstraint_Push(void)
{
  BOOL success = TRUE;

  PKCONSTRAINT pkcStackHead = NULL;

  success &= compare_integer(FALSE, KConstraint_HasStackNext(pkc1),
			     "pkc1 is alone");

  pkcStackHead = KConstraint_Push(pkcStackHead, pkc1);

  success &= compare_integer(pkc1, pkcStackHead,
			     "pkc1 is new stack head");

  success &= compare_integer(FALSE, KConstraint_HasStackNext(pkc1),
			     "last node is NULL-terminated");

  pkcStackHead = KConstraint_Push(pkcStackHead, pkc2);

  success &= compare_integer(pkc2, pkcStackHead,
			     "pkc2 is new stack head");

  success &= compare_integer(pkc1, KConstraint_GetStackNext(pkc2),
			     "pkc1 is next after pkc2");

  success &= compare_integer(FALSE, KConstraint_HasStackNext(pkc1),
			     "stack is still NULL-terminated");

  return success;
}

BOOL Test_KConstraint_Pop(void)
{
  BOOL success = TRUE;

  PKCONSTRAINT pkcStackHead = NULL;

  success &= compare_integer(FALSE, KConstraint_HasStackNext(pkc1),
			     "pkc1 is alone");

  pkcStackHead = KConstraint_Push(pkcStackHead, pkc1);
  pkcStackHead = KConstraint_Push(pkcStackHead, pkc2);
  pkcStackHead = KConstraint_Push(pkcStackHead, pkc3);

  pkcStackHead = KConstraint_Pop(pkcStackHead);

  success &= compare_integer(pkc2, pkcStackHead,
			     "new stack head is old stack head's next");

  success &= compare_integer(FALSE, KConstraint_HasStackNext(pkc3),
			     "popped kconstraint is alone");

  pkcStackHead = KConstraint_Pop(pkcStackHead);

  success &= compare_integer(pkc1, pkcStackHead,
			     "new stack head is old stack head's next");

  success &= compare_integer(FALSE, KConstraint_HasStackNext(pkc2),
			     "popped kconstraint is alone");

  pkcStackHead = KConstraint_Pop(pkcStackHead);

  success &= compare_integer(NULL, pkcStackHead,
			     "new stack head is old stack head's next");

  success &= compare_integer(FALSE, KConstraint_HasStackNext(pkc1),
			     "popped kconstraint is alone");

  return success;
}

BOOL Test_KConstraint_StartInheritance(void)
{
  BOOL success = TRUE;

  return success;
}

BOOL Test_KConstraint_RecalculateInheritance(void)
{
  BOOL success = TRUE;

  TIME tNow = CurrentTime();
  TIME t1;
  TIME t2;

  Thread_SetCurrentConstraint(pThd1, pkc1);
  Thread_SetCurrentConstraint(pThd2, pkc2);
  Thread_SetCurrentConstraint(pThd3, pkc3);
  KConstraint_SetSchedObject(pkc1, pThd1);
  KConstraint_SetSchedObject(pkc2, pThd2);
  KConstraint_SetSchedObject(pkc3, pThd3);

  Thread_MutexListAdd(pThd1, pMX1);
  Mutex_WaiterQueueAddConstraint(pMX1, pkc2);
  Thread_SetWaiting(pThd2, THREAD_WAIT_MUTEX, pMX1);

  KConstraint_SetRestartBy(pkc1, tNow);
  KConstraint_SetCriticality(pkc1, _CRITICAL);

  KConstraint_RecalculateInheritance(pkc1);

  success &= compare_integer(TRUE,
			     Urgency_EQ(KConstraint_GetLocalUrgency(pkc1),
				       KConstraint_GetCurrentUrgency(pkc1)),
			     "resets current urgency to local urgency");

  Int32ToInt64(t1, 0x222222);
  Int32ToInt64(t2, 0x111111);

  KConstraint_SetRestartBy(pkc1, t1);
  KConstraint_SetCriticality(pkc1, _TIMESHARING);
  KConstraint_SetRestartBy(pkc2, t2);
  KConstraint_SetCriticality(pkc2, _CRITICAL);

  KConstraint_RecalculateInheritance(pkc2);
  KConstraint_RecalculateInheritance(pkc1);

  success &= compare_integer(TRUE, TimeEqual(KConstraint_GetRestartBy(pkc1),
					     t2),
			     "pkc1 inherits restartby from pkc2");

  success &= compare_integer(_CRITICAL, KConstraint_GetCriticality(pkc1),
			     "pkc2 inherits criticality from pkc2");

  return success;
}

BOOL Test_KConstraint_UrgencyGTEQ(void)
{
  BOOL success = TRUE;

  PURGENCY pu1;
  PURGENCY pu2;

  TIME t1;
  TIME t2;

  pu1 = KConstraint_GetCurrentUrgency(pkc1);
  pu2 = KConstraint_GetCurrentUrgency(pkc2);

  Int32ToInt64(t1, 2222);
  Int32ToInt64(t2, 222222);

  Urgency_SetTimeSharing(pu1);
  Urgency_SetTimeSharing(pu2);

  Urgency_SetRestartBy(pu1, t1);
  Urgency_SetRestartBy(pu2, t1);

  success &= compare_integer(TRUE, KConstraint_UrgencyGTEQ(pkc1, pkc2),
			     "timesharing w/ same restartbys are >=");

  Urgency_SetCriticality(pu2, _CRITICAL);

  success &= compare_integer(FALSE, KConstraint_UrgencyGTEQ(pkc1, pkc2),
			     "timesharing !>= critical w/ same restartby");

  Urgency_SetCriticality(pu1, _CRITICAL);

  success &= compare_integer(TRUE, KConstraint_UrgencyGTEQ(pkc1, pkc2),
			     "critical w/ same restartbys are GTEQ each other");

  success &= compare_integer(TRUE, KConstraint_UrgencyGTEQ(pkc2, pkc1),
			     "critical w/ same restartbys are GTEQ each other");

  Urgency_SetRestartBy(pu2, t2);

  success &= compare_integer(TRUE, KConstraint_UrgencyGTEQ(pkc1, pkc2),
			     "critical w/ sooner restartby >= later restartby");

  Urgency_SetCriticality(pu1, _TIMESHARING);
  Urgency_SetCriticality(pu2, _TIMESHARING);

  success &=
    compare_integer(TRUE, KConstraint_UrgencyGTEQ(pkc1, pkc2),
		    "timesharing w/ sooner restartby != later restartby");

  success &= compare_integer(TRUE, KConstraint_UrgencyGTEQ(pkc1, pkc1),
			     "urgency >= itself");

  return success;
}

BOOL Test_KConstraint_UrgencyGT(void)
{
  BOOL success = TRUE;

  PURGENCY pu1;
  PURGENCY pu2;

  TIME t1;
  TIME t2;

  Int32ToInt64(t1, 2222);
  Int32ToInt64(t2, 222222);

  pu1 = KConstraint_GetCurrentUrgency(pkc1);
  pu2 = KConstraint_GetCurrentUrgency(pkc2);

  Urgency_SetTimeSharing(pu1);
  Urgency_SetTimeSharing(pu2);

  Urgency_SetRestartBy(pu1, t1);
  Urgency_SetRestartBy(pu2, t1);

  success &= compare_integer(FALSE, KConstraint_UrgencyGT(pkc1, pkc2),
			     "timesharing w/ same restartbys are !>");

  Urgency_SetCriticality(pu2, _CRITICAL);

  success &= compare_integer(FALSE, KConstraint_UrgencyGT(pkc1, pkc2),
			     "timesharing !> critical w/ same restartby");

  Urgency_SetCriticality(pu1, _CRITICAL);

  success &= compare_integer(FALSE, KConstraint_UrgencyGT(pkc1, pkc2),
			     "critical w/ same restartbys are !> each other");

  Urgency_SetRestartBy(pu2, t2);

  success &= compare_integer(TRUE, KConstraint_UrgencyGT(pkc1, pkc2),
			     "critical w/ sooner restartby > later restartby");

  success &= compare_integer(FALSE, KConstraint_UrgencyGT(pkc2, pkc1),
			     "critical w/ later restartby !> later restartby");

  Urgency_SetCriticality(pu1, _TIMESHARING);
  Urgency_SetCriticality(pu2, _TIMESHARING);

  success &= compare_integer(TRUE, KConstraint_UrgencyGT(pkc1, pkc2),
			     "timesharing w/ sooner restartby > later restartby");

  success &= compare_integer(FALSE, KConstraint_UrgencyGT(pkc1, pkc1),
			     "urgency !> itself");


  return success;
}

void ClearKConstraintStack(
    PKCONSTRAINT pkc
    )
{
  PKCONSTRAINT pkcCurrent;
  PKCONSTRAINT pkcNext;

  pkcCurrent = pkc;

  while (pkcCurrent != NULL) {
    pkcNext = KConstraint_GetStackNext(pkcCurrent);
    KConstraint_SetStackNext(pkcCurrent, NULL);
    pkcCurrent = pkcNext;
  }
}

