/* Tests for src/base/logger.c
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <util/test.h>
#include <util/util.h>
#include <base/schedul.h>
#include <schedulif.h>

#define LOGGING_ENABLED 1
#include <util/logger.h>

#ifdef LOGGING_ENABLED

#define TEST_LOG_SIZE 128
LOG_ENTRY TestLogBuffer[TEST_LOG_SIZE];

DECLARE_LOG(TestLog);

DECLARE_LOG_EVENT_TYPE(RESCHEDULE             , "rit Cur=", TRUE );
DECLARE_LOG_EVENT_TYPE(NO_RESCHEDULE          , "rin Cxt=", TRUE );
DECLARE_LOG_EVENT_TYPE(RESCHEDULE_MAIN        , "rl  Cur=", TRUE );
DECLARE_LOG_EVENT_TYPE(IDLE_THREAD            , "enIdle  ", TRUE );
DECLARE_LOG_EVENT_TYPE(END_IDLE_THREAD        , "exIdle  ", TRUE );
DECLARE_LOG_EVENT_TYPE(DEAD_THREAD            , "dead    ", FALSE);
DECLARE_LOG_EVENT_TYPE(SLEEP_UNTIL            , "enSleep ", TRUE );
DECLARE_LOG_EVENT_TYPE(OUT                    , "OUT     ", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_CONDITION_WAIT   , "enCdWait", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_CONDITION_WAIT    , "exCdWait", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_CONDITION_SIGNAL , "enCdSign", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_CONDITION_SIGNAL  , "exCdSign", TRUE );
DECLARE_LOG_EVENT_TYPE(DRAIN_DEFERRED         , "drainDef", FALSE);
DECLARE_LOG_EVENT_TYPE(PUT_READY_QUEUE        , "addReady", TRUE );
DECLARE_LOG_EVENT_TYPE(POP_READY_QUEUE        , "delReady", TRUE );
DECLARE_LOG_EVENT_TYPE(PUT_SLEEP_QUEUE        , "addSleep", TRUE );
DECLARE_LOG_EVENT_TYPE(POP_SLEEP_QUEUE        , "delSleep", TRUE );
DECLARE_LOG_EVENT_TYPE(THREAD_STATE           , "ThdState", FALSE);
DECLARE_LOG_EVENT_TYPE(ADD_CONSTRAINT         , "AddKC   ", TRUE );
DECLARE_LOG_EVENT_TYPE(ADD_THREAD             , "AddThd  ", FALSE);
DECLARE_LOG_EVENT_TYPE(PREEMPT                , "Preempt ", TRUE );
DECLARE_LOG_EVENT_TYPE(RESCHEDULE_INTERRUPT   , "ReschInt", TRUE );
DECLARE_LOG_EVENT_TYPE(REMOVE_CONSTRAINT      , "RemoveKC", FALSE);
DECLARE_LOG_EVENT_TYPE(CONSTRAINT_STACK_FREE  , "KCStackF", FALSE);
DECLARE_LOG_EVENT_TYPE(ENTER_WAKE_THREADS     , "enWake  ", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_WAKE_REALTIME     , "exWakeRT", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_WAKE_TIMESHARE    , "exWakeTS", TRUE );
DECLARE_LOG_EVENT_TYPE(ENTER_UPDATE_RESTARTBY , "enUpRsBy", TRUE );
DECLARE_LOG_EVENT_TYPE(EXIT_UPDATE_RESTARTBY  , "exUpRsBy", TRUE );
DECLARE_LOG_EVENT_TYPE(NO_UPDATE_RESTARTBY    , "noUpRsBy", TRUE );
DECLARE_LOG_EVENT_TYPE(PREEMPT_DEAD           , "PreDead ", FALSE);
DECLARE_LOG_EVENT_TYPE(PREEMPT_RESCHEDULING   , "PreResch", FALSE);
DECLARE_LOG_EVENT_TYPE(PREEMPT_NO_OTHERS      , "PreNoOth", FALSE);
DECLARE_LOG_EVENT_TYPE(PREEMPT_STILL_URGENT   , "PreStill", FALSE);

#endif /* LOGGING_ENABLED */

BOOL Log_TestInitializer(void);
BOOL Test_Log_Init(void);
BOOL Test_Log_WriteEntry(void);
BOOL Test_Log_Print(void);

int _tmain(void) {

  TestFunctionPair LoggerTests[] = {
    {Log_TestInitializer,           ""},
    {Test_Log_Init,                 "Init"      },
    {Test_Log_WriteEntry,           "WriteEntry"},
    {Test_Log_Print,                "Print"     }
  };

  RunTestSuite("Logger Test Suite", LoggerTests,
	       GetTestFunctionArraySize(LoggerTests));

  return 0;
}

BOOL Log_TestInitializer(void)
{
  Log_Init(pTestLog, TEST_LOG_SIZE, TestLogBuffer);
  return TRUE;
}

BOOL Test_Log_Init(void) 
{
  BOOL success = TRUE;
  
  success &= compare_integer(TestLogBuffer, pTestLog->pleBuffer,
			     "Buffer is non-null");
  success &= compare_integer(TEST_LOG_SIZE, pTestLog->BufferSize,
			     "BufferSize should be TEST_LOG_SIZE");
  success &= compare_integer(0, pTestLog->EntryIndex,
			     "EntryIndex should be 0");
  success &= compare_integer(NULL, (UINT)pTestLog->pleBuffer[0].pletType,
			     "First entry is NULL");

  return success;
}

BOOL Test_Log_WriteEntry(void)
{
  BOOL success = TRUE;
  int i;

  /* Test Log_WriteEntry function */
  Log_WriteEntry(pTestLog, LOG_RESCHEDULE, 0x1234);
  success &= compare_integer((UINT)LOG_RESCHEDULE,
			     (UINT)pTestLog->pleBuffer[0].pletType,
			     "entry[0] should be LOG_RESCHEDULE");

  success &= compare_integer(1, pTestLog->EntryIndex,
			     "EntryIndex should be 1");

  Log_Init(pTestLog, TEST_LOG_SIZE, TestLogBuffer);
  Log_WriteEntry(pTestLog, LOG_OUT, 0xffff);
  success &= compare_integer((UINT)LOG_OUT,
			     (UINT)pTestLog->pleBuffer[0].pletType,
			     "entry[0] should be LOG_OUT");

  success &= compare_integer(1, pTestLog->EntryIndex,
			     "EntryIndex should be 1");
    
  Log_WriteEntry(pTestLog, LOG_RESCHEDULE,           0x1000);
  success &= compare_integer((UINT)LOG_RESCHEDULE,
			     (UINT)pTestLog->pleBuffer[1].pletType,
			     "entry[1] should be LOG_RESCHEDULE");
  success &= compare_integer(2, pTestLog->EntryIndex,
			     "EntryIndex should be 2");

  MilliSleep(1);
  Log_WriteEntry(pTestLog, LOG_NO_RESCHEDULE,        0x1001);
  success &= compare_integer((UINT)LOG_NO_RESCHEDULE,
			     (UINT)pTestLog->pleBuffer[2].pletType,
			     "entry[2] should be LOG_NO_RESCHEDULE");

  MilliSleep(2);
  Log_WriteEntry(pTestLog, LOG_RESCHEDULE_MAIN,      0x1002);
  success &= compare_integer((UINT)LOG_RESCHEDULE_MAIN,
			     (UINT)pTestLog->pleBuffer[3].pletType,
			     "entry[3] should be LOG_RESCHEDULE_MAIN");

  MilliSleep(3);
  Log_WriteEntry(pTestLog, LOG_IDLE_THREAD,          0x1003);
  success &= compare_integer((UINT)LOG_IDLE_THREAD,
			     (UINT)pTestLog->pleBuffer[4].pletType,
			     "entry[4] should be LOG_IDLE_THREAD");

  MilliSleep(4);
  Log_WriteEntry(pTestLog, LOG_END_IDLE_THREAD,      0x1004);
  success &= compare_integer((UINT)LOG_END_IDLE_THREAD,
			     (UINT)pTestLog->pleBuffer[5].pletType,
			     "entry[5] should be LOG_END_IDLE_THREAD");

  MilliSleep(5);
  Log_WriteEntry(pTestLog, LOG_DEAD_THREAD,          0x1005);
  success &= compare_integer(NULL,
			     (UINT)pTestLog->pleBuffer[6].pletType,
			     "entry[6]==NULL b/c LOG_DEAD_THREAD disabled");

  MilliSleep(6);
  Log_WriteEntry(pTestLog, LOG_SLEEP_UNTIL,          0x1006);
  success &= compare_integer((UINT)LOG_SLEEP_UNTIL,
			     (UINT)pTestLog->pleBuffer[6].pletType,
			     "entry[6] should be LOG_SLEEP_UNTIL");

  MilliSleep(7);
  Log_WriteEntry(pTestLog, LOG_OUT,                  0x1007);
  success &= compare_integer((UINT)LOG_OUT,
			     (UINT)pTestLog->pleBuffer[7].pletType,
			     "entry[7] should be LOG_OUT");

  MilliSleep(8);
  Log_WriteEntry(pTestLog, LOG_ENTER_CONDITION_WAIT, 0x1008);
  success &= compare_integer((UINT)LOG_ENTER_CONDITION_WAIT,
			     (UINT)pTestLog->pleBuffer[8].pletType,
			     "entry[8] should be LOG_ENTER_CONDITION_WAIT");

  MilliSleep(9);
  Log_WriteEntry(pTestLog, LOG_DRAIN_DEFERRED,       0x1009);
  success &= compare_integer(NULL,
			     (UINT)pTestLog->pleBuffer[9].pletType,
			     "entry[9]==NULL, LOG_DRAIN_DEFERRED disabled");

  MilliSleep(10);
  Log_WriteEntry(pTestLog, LOG_PUT_READY_QUEUE,      0x1010);
  success &= compare_integer((UINT)LOG_PUT_READY_QUEUE,
			     (UINT)pTestLog->pleBuffer[9].pletType,
			     "entry[9] should be LOG_PUT_READY_QUEUE");

  MilliSleep(11);
  Log_Print(pTestLog);
  success &= compare_integer(10, pTestLog->EntryIndex,
			     "EntryIndex should be 10");

  success &= compare_integer(NULL, (UINT)pTestLog->pleBuffer[11].pletType,
			     "entry[11] should still be NULL");

  for (i = 0; i < TEST_LOG_SIZE; i++) {
    Log_WriteEntry(pTestLog, LOG_RESCHEDULE,       0x1000);
  }

  success &= compare_integer(LOG_RESCHEDULE,
			     (UINT)pTestLog->pleBuffer[12].pletType,
			     "entry[12] should now be LOG_RESCHEDULE");

  success &= compare_integer(11, pTestLog->EntryIndex,
			     "EntryIndex should have wrapped around to 11");


  /* The first log entry, LOG_UNUSED 0xffff, should be overwritten */
  Log_Print(pTestLog);
    
  return success;
}

BOOL Test_Log_Print(void)
{
  BOOL success = TRUE;

  /* Test Log_Print on initialized log */
  printf("The following log should be empty.\n");
  Log_Print(pTestLog);

  success &= compare_integer(0, pTestLog->EntryIndex,
			     "EntryIndex should still be 0");

  Log_WriteEntry(pTestLog, LOG_RESCHEDULE, 0x1234);
  success &= compare_integer(1, pTestLog->EntryIndex,
			     "EntryIndex should be 1");

  printf("The following entry should be '%s'.\n", LOG_RESCHEDULE->szName);
  Log_Print(pTestLog);

  success &= compare_integer(1, pTestLog->EntryIndex,
			     "EntryIndex should still be 1");

  return success;
}
