/* Tests for src/base/mutex2.c
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#define SIM 1

#include <stdio.h>
#include <mmlite.h>
#include <base/mutex2.h>
#include <base/thread2.h>
#include <base/kconstraint2.h>
#include <util/test.h>
#include <util/fake_sched.h>

PRIVATE MUTEX mMutex1;
PRIVATE PMUTEX pMX1 = &mMutex1;
PRIVATE PITHREAD pThdCurrent = NULL;
PRIVATE PTR pFakeSchedObject = (PTR)0x2222;
PRIVATE PITHREAD threads[3];
PRIVATE BOOL lock_signals[3];
PRIVATE BOOL unlock_signals[3];
PRIVATE BOOL end_signals[3];
struct IScheduler* pTheScheduler;

BOOL Mutex_TestInitializer(void);
BOOL Test_Mutex_Init(void);
BOOL Test_Mutex_Bad(void);
BOOL Test_Mutex_GetSchedObject(void);
BOOL Test_Mutex_WaiterQueueConstraint(void);
BOOL Test_Mutex_Destroy(void);
BOOL Test_Mutex_TryLock(void);
BOOL Test_MutexLocks(void);

void THREAD_LINKAGE Mutex_LockThread(THREAD_ARGUMENT arg);

int _tmain(void) {
    TestFunctionPair MutexTests[] = {
      {Mutex_TestInitializer,            ""},
      {Test_Mutex_Init,                 "Init"},
      {Test_Mutex_Bad,                  "Bad"},
      {Test_Mutex_GetSchedObject,       "GetSchedObject"},
      {Test_Mutex_WaiterQueueConstraint,"WaiterQueueConstraint"},
      {Test_Mutex_Destroy,              "Destroy"},
      {Test_Mutex_TryLock,              "TryLock"},
      {Test_MutexLocks,                 "Lock/Unlock/Locked"}
    };

    pTheNameSpace = CurrentNameSpace();

    if (pTheNameSpace->v->Bind(pTheNameSpace, _TEXT("cb_sched2"),
			       0, (PIUNKNOWN*)&pTheScheduler) != S_OK) {
      printf("Failed to retrieve scheduler.");
      return 1;
    }

    pThdCurrent = CurrentThread();

    RunTestSuite("Mutex Test Suite", MutexTests,
		 GetTestFunctionArraySize(MutexTests));

    return 0;
}
    
BOOL Mutex_TestInitializer(void)
{
    Mutex_Init(pMX1);
    return TRUE;
}

BOOL Test_Mutex_Init(void)
{
    BOOL success = TRUE;

    success &= compare_integer(NULL, Mutex_GetSchedObject(pMX1),
			       "sets sched object to NULL");

    success &= compare_integer(TRUE, Mutex_WaiterQueueIsEmpty(pMX1),
			       "empties waiter queue");

    success &= compare_integer(pMX1, Node_GetData(Mutex_GetListNode(pMX1)),
			       "node data points to mutex.");

    return success;
}

BOOL Test_Mutex_GetSchedObject(void)
{
    BOOL success = TRUE;

    Mutex_SetSchedObject(pMX1, pFakeSchedObject);

    success &= compare_integer(pFakeSchedObject, Mutex_GetSchedObject(pMX1),
			       "get sched object = set sched object");

    return success;
}

BOOL Test_Mutex_Bad(void)
{
    BOOL success = TRUE;

    success &= compare_integer(FALSE, Mutex_IsBad(pMX1),
			       "mutexes are not born bad");

    Mutex_SetBad(pMX1);

    success &= compare_integer(TRUE, Mutex_IsBad(pMX1),
			       "but they can be pushed over the edge");

    return success;
}

BOOL Test_Mutex_WaiterQueueConstraint(void)
{
    BOOL success = TRUE;
    KCONSTRAINT kc1;
    KCONSTRAINT kc2;
    PKCONSTRAINT pkc1 = &kc1;
    PKCONSTRAINT pkc2 = &kc2;
    PKCONSTRAINT pkcTemp;
    TIME t1;
    TIME t2;

    Int32ToInt64(t1, 0x111111);
    Int32ToInt64(t2, 0x222222);

    KConstraint_Init(pkc1, NULL);
    KConstraint_Init(pkc2, NULL);
    KConstraint_SetRestartBy(pkc1, t1);
    KConstraint_SetRestartBy(pkc2, t2);
    KConstraint_SetCriticality(pkc1, _CRITICAL);
    KConstraint_SetCriticality(pkc2, _CRITICAL);
    Urgency_Copy(KConstraint_GetCurrentUrgency(pkc1),
		 KConstraint_GetLocalUrgency(pkc1));
    Urgency_Copy(KConstraint_GetCurrentUrgency(pkc2),
		 KConstraint_GetLocalUrgency(pkc2));

    Mutex_WaiterQueueAddConstraint(pMX1, pkc2);

    success &= compare_integer(FALSE, Mutex_WaiterQueueIsEmpty(pMX1),
			       "adding constraint makes queue non-empty");

    success &=
      compare_integer(pkc2,
		      Node_GetData( List_GetTail(Mutex_GetWaiterQueue(pMX1))),
		      "adding constraint to empty queue makes it the tail");

    success &= compare_integer(
	KConstraint_GetWaitQueueNode(pkc2),
	List_GetTail(Mutex_GetWaiterQueue(pMX1)),
	"queue uses wait queue node and not sched queue node");

    Mutex_WaiterQueueAddConstraint(pMX1, pkc1);

    success &=
      compare_integer(KConstraint_GetWaitQueueNode(pkc2),
		      Node_GetNext(KConstraint_GetWaitQueueNode(pkc1)),
		      "kconstraints are sorted by urgency");

    success &= compare_integer(NULL, Node_GetNext(
				   KConstraint_GetWaitQueueNode(pkc2)),
			       "waiter queue is NULL-terminated");

    pkcTemp = Mutex_WaiterQueueRemoveMostUrgentConstraint(pMX1);

    success &= compare_integer(pkc1, pkcTemp,
			       "most urgent waiter is found");

    success &= compare_integer(NULL, Node_GetNext(
				   KConstraint_GetWaitQueueNode(pkc1)),
			       "most urgent waiter is removed");

    Mutex_WaiterQueueRemoveConstraint(pMX1, pkc2);

    success &= compare_integer(TRUE, Mutex_WaiterQueueIsEmpty(pMX1),
			       "removing all waiters makes queue empty");

    return success;
}

BOOL Test_Mutex_Destroy(void)
{
    BOOL success = TRUE;

    Mutex_Destroy(pMX1);

    success &= compare_integer(TRUE, Mutex_IsBad(pMX1),
			       "destroying a mutex makes it bad");

    return success;
}

BOOL Test_Mutex_TryLock(void)
{
    BOOL success = TRUE;

    success &= compare_integer(TRUE, Mutex_TryLock(pMX1),
			       "trying a lock on an untaken mutex returns true");

    success &= compare_integer(FALSE, Mutex_TryLock(pMX1),
			       "trying a lock on a taken mutex returns false");
    Mutex_Unlock(pMX1);

    return success;
}


void THREAD_LINKAGE Mutex_LockThread(THREAD_ARGUMENT arg)
{
    int me = (int)arg;
    end_signals[me] = FALSE;
    lock_signals[me] = FALSE;
    unlock_signals[me] = FALSE;
    threads[me] = CurrentThread();
    while (!end_signals[me]) {

      while (!lock_signals[me]) {
	if (end_signals[me] == TRUE) {
	  goto Exit;
	}
	MilliSleep(1);
      }

      lock_signals[me] = FALSE;
      Mutex_Lock(pMX1);

      while (!unlock_signals[me]) {
	if (end_signals[me] == TRUE) {
	  goto Exit;
	}
	MilliSleep(1);
      }

      unlock_signals[me] = FALSE;
      Mutex_Unlock(pMX1);
    }
 Exit:
    threads[me] = NULL;
}

BOOL Test_MutexLocks(void)
{
    BOOL success = TRUE;

    success &= compare_integer(NULL, Mutex_GetSchedObject(pMX1),
			       "no sched object");

    success &= compare_integer(TRUE, Mutex_WaiterQueueIsEmpty(pMX1),
			       "mutex waiter queue is still empty");


    threads[0] = NULL;
    threads[1] = NULL;
    threads[2] = NULL;
    CreateAThread(Mutex_LockThread, (THREAD_ARGUMENT)0);

    while (threads[0] == NULL) {
	MilliSleep(1);
    }

    printf("Started 0\n");
    
    CreateAThread(Mutex_LockThread, (THREAD_ARGUMENT)1);

    while (threads[1] == NULL) {
	MilliSleep(1);
    }

    printf("Started 1\n");

    CreateAThread(Mutex_LockThread, (THREAD_ARGUMENT)2);
    
    while (threads[2] == NULL) {
	MilliSleep(1);
    }

    printf("Started 2\n");

    lock_signals[0] = TRUE;

    while (lock_signals[0] == TRUE) {
	MilliSleep(1);
    }

    printf("Locked 0\n");

    success &= compare_integer(threads[0], Mutex_GetSchedObject(pMX1),
			       "first thread locked mutex");

    success &= compare_integer(TRUE, Mutex_WaiterQueueIsEmpty(pMX1),
			       "mutex waiter queue is still empty");

    lock_signals[1] = TRUE;

    MilliSleep(1);

    while (lock_signals[1] == TRUE) {
	MilliSleep(1);
    }

    success &= compare_integer(TRUE, Thread_IsReady(threads[0]),
			       "first thread is still ready");

    success &= compare_integer(FALSE, Mutex_WaiterQueueIsEmpty(pMX1),
			       "mutex waiter queue now holds second thread");

    success &= compare_integer(TRUE, Thread_IsWaitingForMutex(threads[1]),
			       "second thread is waiting for mutex");

    lock_signals[2] = TRUE;

    MilliSleep(1);

    while (lock_signals[2] == TRUE) {
	MilliSleep(1);
    }

    success &= compare_integer(TRUE, Thread_IsWaitingForMutex(threads[2]),
			       "third thread is waiting for mutex");

    unlock_signals[0] = TRUE;

    MilliSleep(1);

    while (unlock_signals[0] == TRUE) {
	MilliSleep(1);
    }

    success &= compare_integer(TRUE, Thread_IsReady(threads[0]),
			       "first thread is ready, waiting for relock");

    success &= compare_integer(TRUE, Thread_IsReady(threads[1]),
			       "second thread has woken up and is ready");

    success &= compare_integer(threads[1], Mutex_GetSchedObject(pMX1),
			       "second thread has acquired mutex lock");

    unlock_signals[1] = TRUE;

    MilliSleep(1);

    while (unlock_signals[1] == TRUE) {
	MilliSleep(1);
    }

    success &= compare_integer(threads[2], Mutex_GetSchedObject(pMX1),
			       "third thread has acquired mutex lock");

    unlock_signals[2] = TRUE;

    MilliSleep(1);

    while (unlock_signals[2] == TRUE) {
	MilliSleep(1);
    }

    success &= compare_integer(NULL, Mutex_GetSchedObject(pMX1),
			       "mutex is not locked by any thread");

    end_signals[0] = TRUE;

    MilliSleep(1);

    while (threads[0] != NULL) {
	MilliSleep(1);
    }

    end_signals[1] = TRUE;

    MilliSleep(1);

    while (threads[1] != NULL) {
	MilliSleep(1);
    }

    end_signals[2] = TRUE;

    MilliSleep(1);

    while (threads[2] != NULL) {
	MilliSleep(1);
    }

    return success;
}
