/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Test of INameSpace
 */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <assert.h>


#define BUFSIZE     (32)     /* maximum number of characters in buffer */
#define MAXNUMFOOS  (12)     /* maximum number of foos to be released */



/* IFoo interface -- Define a simple COM object that stores a name (a text
 * string) and exports SetName() and GetName() methods.  We will test the
 * INameSpace interface by registering, binding, finding, and
 * unregistering IFoo objects in a namespace.
 */

/* IFoo's external interface
 */
typedef struct _IFOO {
    struct _IFooVtbl *v;
} IFOO, *PIFOO;

/* IFoo's virtual function table
 */
typedef struct _IFooVtbl {
    __VTABLE_COMPATIBILITY_FILLER_DECLARATION

    SCODE (*QueryInterface)(PIFOO, REFIID, void **);
    UINT (*AddRef)(PIFOO);
    UINT (*Release)(PIFOO);
    SCODE (*SetName)(PIFOO, _TCHAR *);
    SCODE (*GetName)(PIFOO, _TCHAR *, UINT, PUINT);
} IFooVtbl;

/* IFoo's internal structure
 */
typedef struct _FOO {
    struct  _IFooVtbl *Vtbl;
    UINT    Sign;           /* debugging signature "AFOO" */
    UINT    Refs;           /* ref count for object */
    UINT    NameLen;        /* length of foo's name */
    _TCHAR   Name[BUFSIZE];  /* foo's name */
} FOO, *PFOO;


/* IFoo's virtual function table
 */
struct _IFooVtbl FooObjVtbl;

/* Interface identifiers for COM objects
 */
const struct IID IID_IFoo       = {0xF000F000,0xF000,0xF000,{0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0}};


/* Foo debugging aids -- Set and check signature field in FOO structure.
 */
const BYTE FooSign[] = "AFOOIAM";   /* only sizeof(UINT) bytes are used */


static INLINE void SetFooSign(PFOO pFoo)
{
    pFoo->Sign = *((PUINT)FooSign);
}


static INLINE BOOL ChkFooSign(PFOO pFoo)
{
    return (pFoo->Sign == *((PUINT)FooSign));
}


/* Generic string length function.  Returns number of characters in
 * NULL-terminated text string, not counting the NULL itself.
 */
static INLINE UINT GetStringLength(_TCHAR *pStr)
{
    _TCHAR *pNext;

    for (pNext = pStr; *pNext != _EOS; ++pNext) {
    }
    return (pNext - pStr);
}


/* **** IFoo constructor
 */
SCODE CreateFoo(PIFOO *ppIFoo)
{
    IHeap *pHeap = CurrentHeap();
    PFOO pFoo = (PFOO)pHeap->v->Alloc(pHeap, 0, sizeof(FOO), 0);

    if (!pFoo) {
        _tprintf(_TEXT("nametest:  FAILED to create FOO object.\n\r"));
        return (E_NOT_ENOUGH_MEMORY);    /* FAILED */
    }
    pFoo->Vtbl = &FooObjVtbl;
    SetFooSign(pFoo);
    pFoo->Refs = 1;
    memcpy(pFoo->Name, _TEXT("<no name>"), 10*sizeof(_TCHAR));
    pFoo->NameLen = 9;   /* don't count _EOS */
    *ppIFoo = (PIFOO)pFoo;
    return (S_OK);    /* okay */
}


/* **** IFoo::SetName()
 * This function sets the foo object's name to the specified string.
 */
SCODE FooObjSetName(PIFOO pThis, _TCHAR *pName)
{
    UINT NameLen;
    PFOO pFoo = (PFOO)pThis;

    if (!ChkFooSign(pFoo)) {
        _tprintf(_TEXT("nametest:  Bad Foo object passed as arg\n\r"));
        return (E_INVALID_PARAMETER);
    }
    NameLen = GetStringLength(pName);
    if (NameLen >= BUFSIZE) {
        NameLen = BUFSIZE - 1;
    }
    memcpy(pFoo->Name, pName, (NameLen+1)*sizeof(_TCHAR));
    pFoo->Name[BUFSIZE-1] = _EOS;
    pFoo->NameLen = NameLen;
    return (S_OK);
}


/* **** IFoo::GetName()
 * This function returns a string containing the foo object's name.
 */
SCODE FooObjGetName(PIFOO pThis, _TCHAR *pName, UINT BufSize, PUINT pNameLen)
{
    PFOO pFoo = (PFOO)pThis;
    UINT NameLen = pFoo->NameLen;

    if (!ChkFooSign(pFoo)) {
        _tprintf(_TEXT("nametest:  Bad Foo object passed as arg\n\r"));
        return (E_INVALID_PARAMETER);
    }
    if (NameLen >= BufSize) {
        NameLen = BufSize - 1;
    }
    memcpy(pName, pFoo->Name, (NameLen+1)*sizeof(_TCHAR));
    pName[BufSize-1] = _EOS;
    *pNameLen = NameLen;
    return (S_OK);
}


/* **** IFoo::AddRef()
 */
UINT FooObjAddRef(PIFOO pThis)
{
    PFOO pFoo = (PFOO)pThis;
    UINT refs;

    if (!ChkFooSign(pFoo)) {
        _tprintf(_TEXT("nametest:  Bad Foo object passed as arg\n\r"));
        DebugBreak();
    }
    refs = AtomicInc(&pFoo->Refs);
    return (refs);
}


/* **** IFoo::Release()
 */
UINT FooObjRelease(PIFOO pThis)
{
    PFOO pFoo = (PFOO)pThis;
    UINT refs;

    if (!ChkFooSign(pFoo)) {
        _tprintf(_TEXT("nametest:  Bad Foo object passed as arg\n\r"));
        DebugBreak();
    }
    refs = AtomicSub(&pFoo->Refs, 1);

    if (refs == 0) {
        IHeap *pHeap = CurrentHeap();
        (void) pHeap->v->Free(pHeap, 0, pThis);
    }

    return (refs);
}


/* **** IFoo::QueryInterface()
 */
SCODE FooObjQueryInterface(PIFOO pThis, REFIID pIid, void **ppNew)
{
    if (UuidCmp(&IID_IFoo, pIid) || UuidCmp(&IID_IUnknown, pIid)) {
        *ppNew = (void *)pThis;
        FooObjAddRef(pThis);
        return NO_ERROR;
    }
    _tprintf(_TEXT("NameTest: Foo::QueryInterface FAILED\n\r"));
    *ppNew = NULL;
    return E_NO_INTERFACE;
}


/* IFoo virtual function table
 */
struct _IFooVtbl FooObjVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    FooObjQueryInterface,
    FooObjAddRef,
    FooObjRelease,
    FooObjSetName,
    FooObjGetName
};



/* Test Program
 */

/* Test Routine:  Create a Foo object, set its name to the string
 * pRealName, and register it in the specified namespace under
 * the name contained in the string pNickName.
 */
BOOL SetUpAFoo(PINAMESPACE pNS, _TCHAR *pNickName, _TCHAR *pRealName)
{
    SCODE sc;
    PIFOO pFoo;

    sc = CreateFoo(&pFoo);
    if (FAILED(sc)) return FALSE;

    pFoo->v->SetName(pFoo, pRealName);
    sc = pNS->v->Register(pNS, pNickName, (PIUNKNOWN)pFoo,
                                             NAME_SPACE_FAILIFEXIST, NULL);
    if (sc != S_OK) {
        _tprintf(_TEXT("nametest:  FAILED to register foo object %s.\n\r"),
                                                               pNickName);
        return FALSE;
    }
    pFoo->v->Release(pFoo);    /* now only reference is in namespace */
    return TRUE;
}


/* Test Routine:  List the names of all foo objects for which entries
 * exist in the namespace pointed to by pNS.
 */
BOOL ListAllNames(PINAMESPACE pNS)
{
    SCODE sc1, sc2, sc3, sc4;
    _TCHAR pNickName[BUFSIZE] = { _TEXT("") }, pRealName[BUFSIZE];
    PIUNKNOWN pUnk;
    PIFOO pFoo = NULL;
    UINT Len;

#if defined(__ARMCC_VERSION)
    /* Compilerbug workaround */
    pNickName[0] = _EOS;
#endif

    do {
        /* Copy name of next namespace entry into string pNickName.
         */
        sc1 = pNS->v->FindNext(pNS, NULL, pNickName, pNickName,
                                             BUFSIZE*sizeof(_TCHAR), &Len);
        if (FAILED(sc1)) {
            _tprintf(_TEXT("nametest:  FAILED -- no entries in namespace.\n\r"));
            return FALSE;
        }

        /* Find namespace entry that matches string pNickName.
         */
        sc2 = pNS->v->Bind(pNS, pNickName, 0, &pUnk);
        if (sc2 != S_OK) {
            _tprintf(_TEXT("nametest:  FAILED to bind to foo named %s.\n\r"), pNickName);
            return FALSE;
        }

        /* Trade in IUnknown pointer for IFoo pointer, then release IUnknown pointer.
         */
        sc3 = pUnk->v->QueryInterface(pUnk, &IID_IFoo, (void **)&pFoo);
        if (sc3 == E_NO_INTERFACE) {
            _tprintf(_TEXT("nametest:  QI() FAILED on foo object.\n\r"));
            return FALSE;
        }
        pUnk->v->Release(pUnk);   /* discard ref thru pointer pUnk */

        /* Print "real name" string from IFoo object and release IFoo pointer.
         */
        sc4 = pFoo->v->GetName(pFoo, pRealName, BUFSIZE, &Len);
        if (sc4 != S_OK) {
            _tprintf(_TEXT("nametest:  FAILED to GetName sc=x%x.\n\r"), sc4);
            return FALSE;
        }
        _tprintf(_TEXT("  ==> %s's real name is %s.\n\r"), pNickName, pRealName);
        pFoo->v->Release(pFoo);    /* discard ref thru pointer pFoo */
    } while (sc1 == S_OK);

    return TRUE;
}


/* Test Routine:  Create a foo object and set its name to the string
 * pRealName.  Next, in the namespace pointed to by pNS, replace the
 * object pointer in the entry whose name matches the string pNickName
 * with a pointer to the newly created object.
 */
BOOL ReplaceAFoo(PINAMESPACE pNS, _TCHAR *pNickName, _TCHAR *pRealName)
{
    SCODE sc;
    UINT nRefs;
    PIUNKNOWN pUnk;
    PIFOO pNewFoo;

    /* Get a pointer to the foo object currently registered under the name
     * pNickName.
     */
    sc = pNS->v->Bind(pNS, pNickName, 0, &pUnk);
    if (sc != S_OK) {
        _tprintf(_TEXT("nametest:  FAILED to bind to foo named %s.\n\r"), pNickName);
        return FALSE;
    }

    /* Create a foo object and set its name to the string pRealName.
     */
    sc = CreateFoo(&pNewFoo);
    if (FAILED(sc)) return FALSE;

    pNewFoo->v->SetName(pNewFoo, pRealName);

    /* Test the NAME_SPACE_FAILIFEXIST flag before replacing registered foo.
     */
    sc = pNS->v->Register(pNS, pNickName, (PIUNKNOWN)pNewFoo,
                                             NAME_SPACE_FAILIFEXIST, NULL);
    if (sc != E_ALREADY_EXISTS) {
        _tprintf(_TEXT("nametest: FAILED Illegally registered FOO object named %s.\n\r"),
                                                                 pNickName);
        return FALSE;
    }

    /* This time, register the new foo object in place of the old one.
     */
    sc = pNS->v->Register(pNS, pNickName, (PIUNKNOWN)pNewFoo, 0, NULL);
    if (sc != S_ALREADY_EXISTS) {
        _tprintf(_TEXT("nametest:  FAILED to replace foo named %s.\n\r"),
                                                             pRealName);
        return FALSE;
    }
    pNewFoo->v->Release(pNewFoo);    /* now only reference is in namespace */

    /* Release local reference to old foo object.  Assuming that the
     * namespace correctly released the old foo object, its reference
     * count should go to zero.  Verify this.
     */
    nRefs = pUnk->v->Release(pUnk);
    if (nRefs) {
        _tprintf(_TEXT("nametest:  Namespace FAILED to release foo named %s.\n\r"),
                                                              pNickName);
        return FALSE;
    }
    return TRUE;
}


/* Test Routine:  Unregister a foo object with the name contained in the
 * string pNickName from the namespace pointed to by pNS.
 */
BOOL RemoveAFoo(PINAMESPACE pNS, _TCHAR *pNickName)
{
    SCODE sc;
    UINT nRefs;
    PIUNKNOWN pUnk;

    /* First, call Bind() to get a local reference to the foo object.
     */
    sc = pNS->v->Bind(pNS, pNickName, 0, &pUnk);
    if (sc != S_OK) {
        _tprintf(_TEXT("nametest:  FAILED to bind to foo named %s.\n\r"),
                 pNickName);
        return FALSE;
    }

    /* Now unregister the foo object from the namespace.
     */
    sc = pNS->v->Unregister(pNS, pNickName);
    if (sc != S_OK) {
        _tprintf(_TEXT("nametest:  FAILED to unregister foo named %s.\n\r"),
                 pNickName);
        return FALSE;
    }

    /* If the namespace correctly released the foo object, this routine
     * should hold only remaining reference to the object.  Verify this.
     */
    nRefs = pUnk->v->Release(pUnk);
    if (nRefs) {
        _tprintf(_TEXT("nametest:  Namespace FAILED to release foo named %s.\n\r"),
                                                                 pNickName);
        return FALSE;
    }

    return TRUE;
}


/* Test Routine:  Close the specified namespace, which may still contain
 * entries for some number of foo objects.  Make sure that all such object
 * references are released before the namespace itself is released.
 */
BOOL CloseNameSpace(PINAMESPACE pNS, _TCHAR *pNSName)
{
    SCODE sc;
    UINT Len;
    int i, n;
    UINT nRefs;
    _TCHAR pNickName[BUFSIZE], pRealName[BUFSIZE];
    PIUNKNOWN pUnk;
    PINAMESPACE pSubNS;
    PIFOO pFoo[MAXNUMFOOS];

    /* Get pointer, pSubNS, to sub-namespace named pNSName, which is
     * currently registered in namespace pNS.
     */
    sc = pNS->v->Bind(pNS, pNSName, 0, &pUnk);
    if (sc != S_OK) {
        _tprintf(_TEXT("nametest:  FAILED to bind to namespace named %s.\n\r"),
                 pNSName);
        return FALSE;
    }
    sc = pUnk->v->QueryInterface(pUnk, &IID_INameSpace, (void **)&pSubNS);
    if (sc == E_NO_INTERFACE) {
        _tprintf(_TEXT("nametest:  QI() FAILED on sub-namespace.\n\r"));
        return FALSE;
    }
    pUnk->v->Release(pUnk);    /* discard ref thru pointer pUnk */

    /* Get pointers to all the foos located in sub-namespace pSubNS.
     */
    sc = pSubNS->v->FindNext(pSubNS, NULL, NULL, pNickName,
                             BUFSIZE*sizeof(_TCHAR), &Len);
    if (FAILED(sc) || !Len) {
        _tprintf(_TEXT("nametest: FAILED No foos registered in namespace.\n\r"));
        return FALSE;
    }
    for (n = 0; n < MAXNUMFOOS;) {
        sc = pSubNS->v->Bind(pSubNS, pNickName, 0, &pUnk);
        if (sc != S_OK) {
            _tprintf(_TEXT("nametest:  FAILED to bind to foo named %s.\n\r"), pNickName);
            return FALSE;
        }
        sc = pUnk->v->QueryInterface(pUnk, &IID_IFoo, (void **)&pFoo[n]);
        if (sc == E_NO_INTERFACE) {
            _tprintf(_TEXT("nametest:  QI() FAILED on foo object.\n\r"));
            return FALSE;
        }
        pUnk->v->Release(pUnk);    /* discard ref thru pointer pUnk */

        n++;

        sc = pSubNS->v->FindNext(pSubNS, NULL, pNickName, pNickName,
                                             BUFSIZE*sizeof(_TCHAR), &Len);
        if (sc != S_OK) {
            break;    /* reached last entry in namespace? */
        }
    }

    /* Now that we have pointers to all the foos located in the
     * sub-namespace, unregister the sub-namespace.
     */
    sc = pNS->v->Unregister(pNS, pNSName);
    if (sc != S_OK) {
        _tprintf(_TEXT("nametest:  FAILED to unregister namespace named %s.\n\r"),
                                                             pNSName);
        return FALSE;
    }

    /* If the unregister call above released the sub-namespace, we should
     * hold the last reference to the sub-namespace object.  Verify this.
     */
    nRefs = pSubNS->v->Release(pSubNS);
    if (nRefs) {
        _tprintf(_TEXT("nametest:  Namespace FAILED to release namespace named %s.\n\r"),
                                                             pNSName);
        return FALSE;
    }

    /* When the sub-namespace was released, it should have released all
     * object references contained in its entries.  Verify this.
     */
    for (i = 0; i < n; ++i) {
        nRefs = pFoo[i]->v->Release(pFoo[i]);
        if (nRefs) {
            sc = pFoo[i]->v->GetName(pFoo[i], pRealName, BUFSIZE*sizeof(_TCHAR), &Len);
            /* BUGBUG check retvals even if failed already ? */
            _tprintf(_TEXT("nametest:  Namespace FAILED to release foo named %s.\n\r"),
                                                             pRealName);
            return FALSE;
        }
    }

    return TRUE;
}


/* Main Test Program.  Create a namespace for foo objects.  Register foo
 * objects in the namespace, bind them, find them, and unregister them.
 */
int _tmain()
{
    SCODE sc;
    PIUNKNOWN pUnk = NULL;
    PINAMESPACE pNS = NULL, pSysNS, pFooNS = NULL;

    /* Get pointer to system namespace and confirm its validity.
     */
    pSysNS = CurrentNameSpace();
    assert(pSysNS != NULL);
    sc = pSysNS->v->QueryInterface(pSysNS, &IID_INameSpace, (void **)&pNS);
    if (sc == E_NO_INTERFACE) {
        _tprintf(_TEXT("nametest:  QI() FAILED on system namespace.\n\r"));
        goto ErrorExit;
    }
    pNS->v->Release(pNS);

    /* Create a namespace object in which to register objects of class
     * FOO.  Register the new namespace, which is itself an object, in the
     * system namespace under the name "Foo Namespace".
     */
    _tprintf(_TEXT("nametest:  Create a namespace for named foo objects.\n\r"));
    sc = pSysNS->v->Bind(pSysNS, _TEXT("Foo Namespace"),
                         NAME_SPACE_CREATE | NAME_SPACE_FAILIFEXIST, &pUnk);
    if (sc != S_OK) {
        _tprintf(_TEXT("nametest:  FAILED to create namespace for foos.\n\r"));
        goto ErrorExit;
    }
    pUnk->v->Release(pUnk);   /* now only foo namespace has reference */

    /* Verify that the name "Foo Namespace" has been successfully
     * registered in the system namespace.
     */
    sc = pSysNS->v->Bind(pSysNS, _TEXT("Foo Namespace"), 0, &pUnk);
    if (sc != S_OK) {
        _tprintf(_TEXT("nametest:  FAILED to bind to first namespace.\n\r"));
        goto ErrorExit;
    }

    /* Verify that the "Foo Namespace" object exposes a namespace interface.
     */
    sc = pUnk->v->QueryInterface(pUnk, &IID_INameSpace, (void **)&pFooNS);
    pUnk->v->Release(pUnk);    /* release obj ref thru pointer pUnk */
    if (sc == E_NO_INTERFACE) {
        _tprintf(_TEXT("nametest:  QI() FAILED on first namespace.\n\r"));
        goto ErrorExit;
    }

    /* Create a number of Foo objects and register them in the foo namespace.
     */
    if (!SetUpAFoo(pFooNS, _TEXT("Ed"), _TEXT("Edward")))
        goto ErrorExit;
    if (!SetUpAFoo(pFooNS, _TEXT("Chuck"), _TEXT("Charles")))
        goto ErrorExit;
    if (!SetUpAFoo(pFooNS, _TEXT("Jim"), _TEXT("James")))
        goto ErrorExit;
    if (!SetUpAFoo(pFooNS, _TEXT("Gerry"), _TEXT("Gerald")))
        goto ErrorExit;
    if (!SetUpAFoo(pFooNS, _TEXT("Sam"), _TEXT("Samuel")))
        goto ErrorExit;
    if (!SetUpAFoo(pFooNS, _TEXT("Ted"), _TEXT("Theodore")))
        goto ErrorExit;
    if (!SetUpAFoo(pFooNS, _TEXT("Max"), _TEXT("Maximillian")))
        goto ErrorExit;

    _tprintf(_TEXT("nametest:  List all names in new namespace:\n\r"));
    if (!ListAllNames(pFooNS))
        goto ErrorExit;

    _tprintf(_TEXT("nametest:  Replace Edward with Edmund and list again:\n\r"));
    ReplaceAFoo(pFooNS, _TEXT("Ed"), _TEXT("Edmund"));
    if (!ListAllNames(pFooNS))
        goto ErrorExit;

    _tprintf(_TEXT("nametest:  Replace Maximillian with Maxwell and list again:\n\r"));
    if (!ReplaceAFoo(pFooNS, _TEXT("Max"), _TEXT("Maxwell")))
        goto ErrorExit;
    if (!ListAllNames(pFooNS))
        goto ErrorExit;

    _tprintf(_TEXT("nametest:  Replace Gerald with Gerard and list again:\n\r"));
    if (!ReplaceAFoo(pFooNS, _TEXT("Gerry"), _TEXT("Gerard")))
        goto ErrorExit;
    if (!ListAllNames(pFooNS))
        goto ErrorExit;

    _tprintf(_TEXT("nametest:  Unregister Ed, Jim, Max and Sam and list again:\n\r"));
    if (!RemoveAFoo(pFooNS, _TEXT("Ed")))
        goto ErrorExit;
    if (!RemoveAFoo(pFooNS, _TEXT("Jim")))
        goto ErrorExit;
    if (!RemoveAFoo(pFooNS, _TEXT("Max")))
        goto ErrorExit;
    if (!RemoveAFoo(pFooNS, _TEXT("Sam")))
        goto ErrorExit;
    if (!ListAllNames(pFooNS))
        goto ErrorExit;

    /* Release the local reference to the "Foo Namespace" object thru the
     * pointer pFooNS.  The sole remaining ref will be in system namespace.
     */
    pFooNS->v->Release(pFooNS);
    /* BUGBUG verify count */
    pFooNS = NULL;

    /* Close the "Foo Namespace" object and verify that all of its foo
     * object entries are automatically released as well.
     */
    if (!CloseNameSpace(pSysNS, _TEXT("Foo Namespace")))
        goto ErrorExit;

    _tprintf(_TEXT("Name space tests PASSED.\n\r"));
    exit(EXIT_SUCCESS);

 ErrorExit:
    _tprintf(_TEXT("Name space tests FAILED.\n\r"));
    if (pFooNS) {
        pFooNS->v->Release(pFooNS);
        (void) CloseNameSpace(pSysNS, _TEXT("Foo Namespace"));
    }
    exit(EXIT_FAILURE);
    return 1;/*amusme*/
}
