/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "../utilities/utility.h"

/* Test for Condition_InterruptSignal
 * Hard to do a good job of this, and impossible to do it in a portable way.
 *
 * The CANT_THINK_OF_A_WAY conditional should NEVER be used.
 * It might or might not be able to catch some bugs, and it might get into
 * other bugs since Condition_InterruptSignal is called from non-ISR.
 */

#if defined(__TCS__)
#define MMIO(offset)    (((volatile unsigned int *)_MMIO_base)[(offset) >> 2])
#define TIMER_TMODULUS(_n_)     (0x100C00 + (_n_*0x20))
#define TIMER_TVALUE(_n_)       (0x100C04 + (_n_*0x20))
#define TIMER_TCTL(_n_)         (0x100C08 + (_n_*0x20))

volatile unsigned int *_MMIO_base;
#else
#define CANT_THINK_OF_A_WAY 1
#endif


UINT nThreads = 1;/*main*/

SCODE CreateAThread( 
            THREAD_FUNCTION pStart,
            THREAD_ARGUMENT Arg)
{
    PIPROCESS pPrc = CurrentProcess();
    SCODE sc;

    AtomicInc(&nThreads);
    sc = pPrc->v->CreateThread(pPrc, 0,
                               pStart,
                               Arg,
                               0, NULL, NULL);
    return sc;
}


/* Only returns for the last thread.
 */
void MaybeTerminate(void)
{
    if (AtomicDec(&nThreads) != 0)
        ThreadExit();
}

/* Other utils
 */
void Microsleep(int me)
{
    TIME t;

#if defined(_DEBUG)
    _tprintf(_TEXT("Thread %d microsleeps...\n\r"), me);
#else
    UnusedParameter(me);
#endif
    Int32ToInt64(t,TIME_RELATIVE(TIME_MICROS(500)));
    SleepUntil(t);
}


/*-----------------------*/
BOOL Passed = TRUE;
volatile int Step = 0;

CONDITION MyCondition;
volatile BOOL HasInterrupted = FALSE;

BOOL MCT MyInterruptServiceRoutine( PTR pThis, BOOL *pNotMyInterrupt )
{
    UnusedParameter(pThis);
    UnusedParameter(pNotMyInterrupt);
    HasInterrupted = TRUE;
    return Condition_InterruptSignal( & MyCondition );
}

void ArmMyInterrupt(void)
{
#if defined(CANT_THINK_OF_A_WAY)
    BOOL xx = FALSE;
    HasInterrupted = FALSE;
    (void) MyInterruptServiceRoutine( &MyCondition, &xx);

#elif defined(__TCS__)
    /* Do it using Interrupt 6, system timer 1 [as in timer 0,1,2].
     * Cant quite use software interrupts cuz they are edge-triggered,
     * and there does not seem to be a generic way to dispatch them.
     */
#define USETIMER 1
#define IRQTIMER (5+USETIMER)
#define MY_IRQ IRQTIMER
    HasInterrupted = FALSE;

    MMIO(TIMER_TMODULUS(USETIMER)) = 0xffffffff;

    MMIO(TIMER_TVALUE(USETIMER))   = ~10;
#define TIMER_SOURCE_PRESCALE       0x100
#define TIMER_STOP  0x00
#define TIMER_RUN   0x01
#define TIMER_PRESCALE(_n_)         ((_n_ & 0x0F)<<16)
    MMIO(TIMER_TCTL(USETIMER))     = TIMER_RUN|TIMER_SOURCE_PRESCALE|TIMER_PRESCALE(1);

#endif
}

#if defined(__TCS__)
void InitMmioBase(void)
{
    TM_MACHINE_INFO tm;
    SCODE sc;
    PIPROCESS pPrc = CurrentProcess();

    sc = pPrc->v->MachineInfo( pPrc,
                               MACHINE_KIND_TRIMEDIA,
                               (PMACHINE_INFO) &tm );
    if (FAILED(sc)) {
        _tprintf(_TEXT("Trimedia machinfo failed ?!? x%x\n\r"), sc);
        DebugBreak();
        ThreadExit(); /* we really dont want to go on */
    }
    _MMIO_base = (volatile unsigned int *)tm.MMIOBase;
}
#endif

/* Test code proper
 */
void THREAD_LINKAGE FirstThread(THREAD_ARGUMENT arg)
{
    int me;
    SCODE sc;
    TIME t;

    me = (int)arg;

    if (me == 0) {

        _tprintf(_TEXT("Will test Condition_InterruptSignal\n\r"));

        /* Initialize and start our "device" 
         */
        Condition_Init(&MyCondition);

#if defined(CANT_THINK_OF_A_WAY)
        /* dont */
#else
        sc = AddDevice( &MyCondition, (PTR) MyInterruptServiceRoutine, 0, MY_IRQ, 0);
        if (FAILED(sc)) {
            _tprintf(_TEXT("AddDevice failed with sc=x%x\n\r"), sc);
            Passed = FALSE;
            goto TestComplete;
        }
#endif

        /* Start a helper thread
         */
        (void) CreateAThread(FirstThread, (THREAD_ARGUMENT) 1 );
    }

    _tprintf(_TEXT("Thread %d is x%x\n\r"), me, (ADDRESS) CurrentThread());

    if (me == 0) {
        /* Micro-sleep to make sure th1 is waiting
         */
        while (Step == 0)
            Microsleep(me);

        /* Trigger an interrupt, make sure th1 gets it
         */
        ArmMyInterrupt();

        while (Step != 2 && Passed)
            Microsleep(me);
        if (!Passed)
            goto TestComplete;

        /* Second test
         */
        ArmMyInterrupt();
        while (!HasInterrupted)
            Microsleep(me);
        Step = 3;

        /* We are done, let the other thread complete
         */

    } else {
        /* First test: we go to wait BEFORE the interrupt
         */
        Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(50)));
        Step = 1;
        sc = Condition_Wait( &MyCondition, NULL, &t);
        if (sc == E_TIMED_OUT) {
            _tprintf(_TEXT("Thread %d did not get signalled\n\r"), me);
            Passed = FALSE;
            goto TestComplete;
        }

        /* Second test: we go to wait AFTER the interrupt
         */
        Step = 2;
        while (Step == 2)
            Microsleep(me);
        
        Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(1)));
        sc = Condition_Wait( &MyCondition, NULL, &t);
        if (sc == E_TIMED_OUT) {
            _tprintf(_TEXT("Condition was not signalled !\n\r"));
            Passed = FALSE;
            /* goto TestComplete; */
        }
        
        /* Done
         */
    }

 TestComplete:
    /* All threads should be done
     */
    MaybeTerminate();

#if defined(CANT_THINK_OF_A_WAY)
        /* dont */
#else
    /* Remove our ISR
     */
    sc = RemoveDevice( &MyCondition, MY_IRQ );
    if (FAILED(sc)) {
        _tprintf(_TEXT("RemoveDevice failed with sc=x%x\n\r"), sc);
        Passed = FALSE;
    }
#endif

    /* So how did it go ?
     */
    if (Passed)
        _tprintf(_TEXT("This test PASSED\n\r"));
    else {
        _tprintf(_TEXT("This test FAILED\n\r"));
    }

    /* NB: Two threads call this.  
     * Only the last one actually terminates the process,
     * therefore setting the exit code to the proper last-seen value
     */
    SetAndReturn(Passed?EXIT_SUCCESS:EXIT_FAILURE);
}

int _tmain(int argc, _TCHAR **argv)
{
#if defined(__TCS__)
    InitMmioBase();
#endif

    UnusedParameter(argc);
    UnusedParameter(argv);

    FirstThread(0);
    return(0);
}
