/* Tests for src\base\tconstraint2.c
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <mmlite.h>
#include <base/tconstraint2.h>
#include <base/urgency2.h>
#include <util/test.h>

/* Global variables for testing */

PRIVATE TIME_CONSTRAINT tc1;
PRIVATE TIME_CONSTRAINT tc2;
PRIVATE TIME_CONSTRAINT tc3;
PRIVATE PTIME_CONSTRAINT ptc1 = &tc1;
PRIVATE PTIME_CONSTRAINT ptc2 = &tc2;
PRIVATE PTIME_CONSTRAINT ptc3 = &tc3;

/* Function prototypes for tests */

BOOL TConstraint_TestInitializer(void);
BOOL Test_TConstraint_Init(void);
BOOL Test_TConstraint_StartTime(void);
BOOL Test_TConstraint_EstimateTime(void);
BOOL Test_TConstraint_DeadlineTime(void);
BOOL Test_TConstraint_Criticality(void);
BOOL Test_TConstraint_EndPrevious(void);
BOOL Test_TConstraint_TimeTaken(void);

int _tmain(void) {
  TestFunctionPair TConstraintTestFunctions[] = {
    {TConstraint_TestInitializer,            ""},
    {Test_TConstraint_Init,                  "TConstraint_Init"},
    {Test_TConstraint_StartTime,             "TConstraint_StartTime"},
    {Test_TConstraint_EstimateTime,          "TConstraint_EstimateTime"},
    {Test_TConstraint_DeadlineTime,          "TConstraint_DeadlineTime"},
    {Test_TConstraint_Criticality,           "TConstraint_Criticality"},
    {Test_TConstraint_EndPrevious,           "TConstraint_EndPrevious"},
    {Test_TConstraint_TimeTaken,             "TConstraint_TimeTaken"}
  };

  RunTestSuite("Time Constraint Test Suite", TConstraintTestFunctions,
	       GetTestFunctionArraySize(TConstraintTestFunctions));

  return 0;
}

BOOL TConstraint_TestInitializer(void)
{
  TConstraint_Init(ptc1);
  TConstraint_Init(ptc2);
  TConstraint_Init(ptc3);

  return TRUE;
}

BOOL Test_TConstraint_Init(void)
{
  BOOL success = TRUE;

  TIME tNow = CurrentTime();
  TIME tZero;

  Int32ToInt64(tZero, 0);

  success &= compare_integer(TRUE,
			     TimeEqual(TConstraint_GetStartTime(ptc1),
				       TConstraint_GetDeadlineTime(ptc1)),
			     "start time == deadline time");

  success &= compare_integer(TRUE,
			     TimeLess(TConstraint_GetStartTime(ptc1), tNow),
			     "start time < current time");

  success &= compare_integer(TRUE,
			     TimeEqual(TConstraint_GetEstimateTime(ptc1),
				       tZero),
			     "estimate time is originally 0");

  success &= compare_integer(_TIMESHARING, TConstraint_GetCriticality(ptc1),
			     "timesharing initially");

  success &= compare_integer(FALSE, TConstraint_GetEndPrevious(ptc1),
			     "no end previous initially");

  success &= compare_integer(NULL, TConstraint_GetTimeTaken(ptc1),
			     "time taken is NULL initially");

  return success;
}

BOOL Test_TConstraint_StartTime(void)
{
  BOOL success = TRUE;

  TIME t1;

  Int32ToInt64(t1, 2222);

  TConstraint_SetStartTime(ptc1, t1);

  success &= compare_integer(TRUE, TimeEqual(TConstraint_GetStartTime(ptc1),
					     t1),
			     "set start time == get start time");

  return success;
}

BOOL Test_TConstraint_DeadlineTime(void)
{
  BOOL success = TRUE;

  TIME t1;

  Int32ToInt64(t1, 2222);

  TConstraint_SetDeadlineTime(ptc1, t1);

  success &= compare_integer(TRUE,
			     TimeEqual(TConstraint_GetDeadlineTime(ptc1),
				       t1),
			     "set deadline time == get deadline time");

  return success;
}

BOOL Test_TConstraint_EstimateTime(void)
{
  BOOL success = TRUE;

  TIME t1;

  Int32ToInt64(t1, 2222);

  TConstraint_SetEstimateTime(ptc1, t1);

  success &= compare_integer(TRUE,
			     TimeEqual(TConstraint_GetEstimateTime(ptc1),
				       t1),
			     "set estimate time == get estimate time");

  return success;
}

BOOL Test_TConstraint_Criticality(void)
{
  BOOL success = TRUE;

  success &= compare_integer(FALSE, TConstraint_IsCritical(ptc1),
			     "initially not critical");

  success &= compare_integer(TRUE, TConstraint_IsTimeSharing(ptc1),
			     "initially timesharing");

  TConstraint_SetCritical(ptc1);

  success &= compare_integer(TRUE, TConstraint_IsCritical(ptc1),
			     "get critical == set critical");

  success &= compare_integer(TRUE, TConstraint_IsTimeSharing(ptc1),
			     "setting critical does not affect timesharing");

  TConstraint_ClearTimeSharing(ptc1);

  success &= compare_integer(FALSE, TConstraint_IsTimeSharing(ptc1),
			     "get timesharing == set timesharing");

  success &= compare_integer(TRUE, TConstraint_IsCritical(ptc1),
			     "setting timesharing does not affect critical");

  TConstraint_ClearCritical(ptc1);

  success &= compare_integer(FALSE, TConstraint_IsCritical(ptc1),
			     "get critical == set critical");

  success &= compare_integer(FALSE, TConstraint_IsTimeSharing(ptc1),
			     "setting critical does not affect timesharing");

  TConstraint_SetTimeSharing(ptc1);

  success &= compare_integer(TRUE, TConstraint_IsTimeSharing(ptc1),
			     "get timesharing == set timesharing");

  success &= compare_integer(FALSE, TConstraint_IsCritical(ptc1),
			     "setting timesharing does not affect critical");

  return success;
}

BOOL Test_TConstraint_EndPrevious(void)
{
  BOOL success = TRUE;

  TConstraint_SetCritical(ptc1);
  TConstraint_SetEndPrevious(ptc1);

  success &= compare_integer(TRUE, TConstraint_GetEndPrevious(ptc1),
			     "get end previous == set end previous");

  success &= compare_integer(TRUE, TConstraint_IsCritical(ptc1),
			     "setting end previous does not affect criticality");

  TConstraint_ClearEndPrevious(ptc1);

  success &= compare_integer(FALSE, TConstraint_GetEndPrevious(ptc1),
			     "get end previous == set end previous");

  success &= compare_integer(TRUE, TConstraint_IsCritical(ptc1),
			     "clearing end previous does not affect criticality");

  return success;
}

BOOL Test_TConstraint_TimeTaken(void)
{
  BOOL success = TRUE;

  TIME tNow;

  TConstraint_SetTimeTaken(ptc1, &tNow);

  success &= compare_integer(&tNow, TConstraint_GetTimeTaken(ptc1),
			     "set timetaken = get timetaken");

  return success;
}
