/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * This module times context switches between threads.
 *
 * It works by using a rotating ring of mutex acquisitions among the
 * participating threads to force context switches.  One more mutex must
 * be in the ring than the number of threads in order to prevent deadlock.
 *
 * Michael B. Jones (mbj) -- May 22, 1996
 */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "../utilities/utility.h"

#define MAXNOTHREADS    100
UINT NThreads = 2;

volatile int ThreadCount[MAXNOTHREADS];
MUTEX Mutexes [MAXNOTHREADS+1];
IThread *Thread[MAXNOTHREADS];

volatile BOOL Ready[MAXNOTHREADS] = { FALSE };
volatile BOOL OkToCount = FALSE;
volatile BOOL ThatsAllFolks = FALSE;

void THREAD_LINKAGE ThreadFunc(THREAD_ARGUMENT arg)
{
    UINT Me = (UINT) arg;
    UINT Mine = MAXNOTHREADS, Next;

    Mine = Me;

    Mutex_Lock(&Mutexes[Mine]);

    Ready[Me] = TRUE;

    for (;;) {
        if (ThatsAllFolks) {
            Ready[Me] = FALSE;
            break;
        }
        if (! OkToCount)
            continue;

        ThreadCount[Me]++;
        Next = Mine + 1;
        if (Next > NThreads) Next = 0; /* Range is 0 .. NThreads inclusive */

/*      _tprintf(_TEXT("%d: Lock %d\n\r"), Me, Next);*/
        Mutex_Lock(&Mutexes[Next]); /* Forces wait for next thread */
/*      _tprintf(_TEXT("%d: Unlock %d\n\r"), Me, Mine);*/
        Mutex_Unlock(&Mutexes[Mine]); /* Allows previous thread to proceed */
        Mine = Next;            /* Advance lock held number */
    }

    /* did we leave one locked behind ? */
    if (Mine < MAXNOTHREADS)
        Mutex_Unlock(&Mutexes[Mine]);

}

void Barrier( BOOL ReadyOrNot)
{
    UINT i;
    TIME t;
    for (;;) {

        /* Loop over all threads, or exit earlier if we discover
         * a thread that is not yet ready.
         */
        for (i = 0; i < NThreads; i++)
            if (Ready[i] == ReadyOrNot)
                break;

        /* If all threads are ready, then exit loop.
         */
        if (i == NThreads)
            break;


        /* Otherwise, sleep for a bit and check again.
         */
        Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(100)));
        SleepUntil(t);
    }
}

INT __cdecl _tmain(INT argc, _TCHAR * argv[])
{
    INT TestTime = 3;
    UINT i;
    SCODE sc;
    UINT TotalCount = 0;
    PIPROCESS pPrc;

    sc = ParseCmdLine(&argc,&argv);

    /*_tprintf(_TEXT("argc == x%x, argv == x%x\n\r"), argc, argv);*/

    switch (argc) {
    case 3:
        TestTime = _ttoi(argv[2]);
    case 2:
        NThreads = _ttoi(argv[1]);
    case 1:
    case 0:
        break;
    default:
        _tprintf(_TEXT("usage: tcswitch [NThreads [TestTime]]\n\r"));
#if 0
        return SetAndReturn(EXIT_SUCCESS);
#endif
    }

    if (NThreads > MAXNOTHREADS) {

        _tprintf(_TEXT("tcswitch: FAILED Can't create more than %d threads\n\r"),
                 MAXNOTHREADS);
        return SetAndReturn(EXIT_FAILURE);
    }

    _tprintf(_TEXT("tcswitch: Test %d threads for %d seconds\n\r"),
             NThreads, TestTime);

    /* Create all those threads */

    pPrc = CurrentProcess();

    for (i = 0; i < NThreads; i++) {

        Ready[i] = FALSE;
        Mutex_Init(&Mutexes[i]);

        sc = pPrc->v->CreateThread(pPrc, 0, ThreadFunc, (THREAD_ARGUMENT)i,
                                   0, NULL, &Thread[i]);
        if (FAILED(sc)) {
            _tprintf(_TEXT("tcswitch: FAILED error 0x%x creating thread #%d\n\r"), sc, i);
            /* XXX Release those already created */
            return SetAndReturn(EXIT_FAILURE);
        }

    }
    Mutex_Init(&Mutexes[NThreads]); /* One more than thread count */

    _tprintf(_TEXT("tcswitch: All %d threads created\n\r"), NThreads);

    /* Wait until all threads are running */

    Barrier( FALSE );


    /* Enable counting, and wait for TestTime seconds */

    _tprintf(_TEXT("tcswitch: All %d threads ready; sleeping for %d seconds\n\r"),
             NThreads, TestTime);
    OkToCount = TRUE;

    {
        TIME t;
        Int32ToInt64(t,TIME_RELATIVE(TIME_SECONDS(TestTime)));
        SleepUntil(t);
    }

    /* Disable counting, and kill all the threads */

    ThatsAllFolks = TRUE;
    OkToCount = FALSE;

    _tprintf(_TEXT("tcswitch: terminating threads\n\r"));

    /* Wait to make sure that they are gone */

    Barrier( TRUE );

    for (i = 0; i < NThreads; i++) {
        if (Thread[i])
            Thread[i]->v->Release(Thread[i]);
    }
        
    _tprintf(_TEXT("tcswitch: terminated all %d threads\n\r"), NThreads);

    /* Show how many iterations each thread has run */

    for (i = 0; i < NThreads; i++) {
        _tprintf(_TEXT("ThreadCount[%2d] = %7d\n\r"), i, ThreadCount[i]);
        TotalCount += ThreadCount[i];
    }

    _tprintf(_TEXT("tcswitch: %u context switches among %d threads in %d seconds\n\r"),
             TotalCount, NThreads, TestTime);
    _tprintf(_TEXT("tcswitch: %u nanoseconds per context switch\n\r"),
             (TIME_SECONDS(TestTime) / TotalCount) * 100);
    _tprintf(_TEXT("tcswitch: test PASSED.\n\r"));

    (void) FreeCmdLine(&argc,&argv);

    return SetAndReturn(EXIT_SUCCESS);
}
