/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
* Test_drive.exe opens a script file as input and reads it for test suites and params. 
* Format follows
# this is a comment. It can appear anywhere but must have a '#' on the left end. 

#Next line calls foo.exe and passes the parms to it ... just like running from cmd. line.   
#Note that foo.exe should call IModule::SetExitCode() to 0 for success or 1 for failure. 
foo.exe -i BAR 
#more comments and code here ... 
*/

#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <ctype.h>
#include <assert.h>
#include <mmlite.h>
#include "../utilities/utility.h"

#define EXEBUF  (1024)
#define PARMBUF (1024)

void TDUsage(const _TCHAR *); 
_TCHAR * TrimTrailingWhitespace(_TCHAR * s);

_TCHAR sBuf[1024];
_TCHAR sExe[EXEBUF] = {0,};
_TCHAR sParms[PARMBUF] = {0,}; 

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE       sc;
    FILE * fIn = 0, * fOut = stdout; 
    const _TCHAR * sArg = 0; 
    BOOL bVerbose = FALSE;
    IProcess * pProc = CurrentProcess();


    ParseCmdLine(&argc, &argv); 
    OptionalBreak(argc, argv); 
    TellAboutOptionalBreak();

    if(argc < 2)
        TDUsage(argv[0]);

    sArg = GetArgFromCommandLine(argc, argv, _T("i")); 
    if(sArg && *sArg)
    {
        //this little side step eliminates some const/non-const issues 
        _tcscpy(sBuf, sArg); 
        TrimTrailingWhitespace(sBuf);
        fIn = _tfopen(sBuf, _T("r")); 
        if(!fIn)
        {
            _ftprintf(fOut, _T("\nHARNESS:ERROR:Cannot open [%s] for input, quitting ..."), sBuf); 
            fflush(fOut);
            FreeCmdLine(&argc, &argv); 
            return SetAndReturn(1);
        }
        else
            _ftprintf(fOut, _T("\nTaking input from [%s]"), sBuf); 
    }
    else
    {
       _ftprintf(fOut, _T("\nNeed a valid input file, quitting ... "));
       FreeCmdLine(&argc, &argv); 
       return SetAndReturn(1);
    }

    sArg = GetArgFromCommandLine(argc, argv, _T("o")); 
    if(sArg && *sArg)
    {
        _tcscpy(sBuf, sArg); 
        TrimTrailingWhitespace(sBuf);
        fOut = _tfopen(sBuf, _T("w+"));
        if(!fOut)
        {
            _tprintf(_T("\nHARNESS:xERROR:Cannot open [%s] for output, quitting ..."), sBuf); 
            if(fIn != stdin)
                fclose(fIn);
            FreeCmdLine(&argc, &argv); 
            return SetAndReturn(1);
        }
        else
            _ftprintf(stdout, _T("\nSending output to [%s]"), sBuf); 
    }
    else
    {
        _ftprintf(stdout, _T("\nSending output to the console ..."));
        fOut = stdout; 
    }

    sArg = GetArgFromCommandLine(argc, argv, _T("v"));

    if(sArg)
        bVerbose = TRUE; 
    else
        bVerbose = FALSE; 
  

    //our little utility buf starts clean ... 
    *sBuf = (_TCHAR)0; 

    for(;_fgetts(sBuf, sizeof(sBuf)/sizeof(sBuf[0]), fIn);)
    {
        IModule * pMod = 0;
        _TCHAR * s; 

        *sParms = (_TCHAR)0; 
        *sExe = (_TCHAR)0; 
        _ftprintf(fOut, _T("\n"));
        if(_tcslen(sBuf))
        {
            _ftprintf(fOut, sBuf);
            fflush(fOut); 
        }
        
        //eat up leading whitespace
        for(s = sBuf; *s; ++s)
            if(!_istspace(*s))
                break; 

        //skip comment lines
        if(*sBuf == *_T("#"))
            continue; 

        //eat up leading whitespace
        for(s = sBuf; *s; ++s)
            if(!_istspace(*s))
                break; 


        //look for end of module name, e.g., FOO.EXE 
        for( ; *s; ++s)
            if(_istspace(*s) || !*s)
                break; 

        if(*s)
        {
            _tcsncpy(sExe, sBuf, (s - sBuf) * sizeof(_TCHAR));
            _tcscpy(sParms, s);
        }
        else
            /* can't parse, probably just whitespace */
            continue; 

        sc = pProc->v->LoadImage(pProc, sExe, sParms, LOAD_IMAGE_CALL_ENTRY_POINT, (IModule **)&pMod); 
        
        if(FAILED(sc))
            _ftprintf(fOut, _T("\n#LoadImage() failed, returned [%s(0x%.8x)]"), ErrorDescription(sc), sc); 
        else
        {
            INT nExitCode; 
            PIPROCESS pSpawn;

            sc = pMod->v->QueryInterface(pMod, &IID_IProcess, (void **) &pSpawn);
            if(FAILED(sc))
                _ftprintf(fOut, _T("\n#Something bad wrong hear perfesser - QI() FAILED?!?")); 

            nExitCode = pSpawn->v->GetValue(pSpawn, PROCESS_VALUE_EXIT_CODE); 
            pSpawn->v->Release(pSpawn);
            if(nExitCode) {
                _ftprintf(fOut, _T("\n#Nonzero exit code [%d](usually indicates error)"), nExitCode); 
                if(bVerbose)
                    _ftprintf(fOut, _T("\a"));
            } else
                _ftprintf(fOut, _T("\n#Zero exit code [%d](usually indicates success)"), nExitCode); 

            pMod->v->Release(pMod);
        }
    }

    if(fIn != stdin)
        fclose(fIn); 

    if(fOut != stdout)
        fclose(fOut); 
    
    FreeCmdLine(&argc, &argv); 
    return SetAndReturn(0);
}

void TDUsage(const _TCHAR * s)
{
    _tprintf(_T("\nUsage:%s -i <infile> [-o <outfile>]"), s); 
    _tprintf(_T("\nOutput defaults to the console ... ")); 
}

_TCHAR * TrimTrailingWhitespace(_TCHAR * s)
{
    _TCHAR * p = 0;
    if(!s) return s; 
    if(!*s) return s; 

    for(p = &s[_tcslen(s)]; p >= s && (!*p || _istspace(*p)); p--)
        *p = (_TCHAR)0;

    return s; 
}


