/* Tests for src\base\urgency2.c
 * Copyright (c) Microsoft Corporation. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <mmlite.h>
#include <base/urgency2.h>
#include <util/test.h>

/* Global variables for testing */

PRIVATE URGENCY u1;
PRIVATE URGENCY u2;
PRIVATE PURGENCY pu1 = &u1;
PRIVATE PURGENCY pu2 = &u2;

/* Function prototypes for tests */

BOOL Urgency_TestInitializer(void);
BOOL Test_Urgency_Init(void);
BOOL Test_Urgency_Copy(void);
BOOL Test_Urgency_RestartBy(void);
BOOL Test_Urgency_Criticality(void);
BOOL Test_Urgency_TimeSharing(void);
BOOL Test_Urgency_EQ(void);
BOOL Test_Urgency_GTEQ(void);
BOOL Test_Urgency_GT(void);

int _tmain(void) {
    TestFunctionPair UrgencyTestFunctions[] = {
      {Urgency_TestInitializer,            ""},
      {Test_Urgency_Init,                  "Urgency_Init"},
      {Test_Urgency_Copy,                  "Urgency_Copy"},
      {Test_Urgency_RestartBy,             "Urgency_RestartBy"},
      {Test_Urgency_Criticality,           "Urgency_Criticality"},
      {Test_Urgency_TimeSharing,           "Urgency_TimeSharing"},
      {Test_Urgency_EQ,                    "Urgency_EQ"},
      {Test_Urgency_GTEQ,                  "Urgency_GTEQ"},
      {Test_Urgency_GT,                    "Urgency_GT"}
    };

    RunTestSuite("Urgency Test Suite", UrgencyTestFunctions,
		 GetTestFunctionArraySize(UrgencyTestFunctions));

    return 0;
}

BOOL Urgency_TestInitializer(void)
{
    Urgency_Init(pu1);
    Urgency_Init(pu2);

    return TRUE;
}

BOOL Test_Urgency_Init(void)
{
    BOOL success = TRUE;

    TIME tNow = CurrentTime();

    success &= compare_integer(TRUE, TimeLess(Urgency_GetRestartBy(pu1),
					      tNow),
			       "RestartBy is less than current time");

    success &= compare_integer(_TIMESHARING, Urgency_GetCriticality(pu2),
			       "timesharing initially");

    return success;
}

BOOL Test_Urgency_Copy(void)
{
    BOOL success = TRUE;
    TIME tNew;

    Int32ToInt64(tNew, 2222);
    Urgency_SetRestartBy(pu1, tNew);
    Urgency_SetCritical(pu1);

    Urgency_Copy(pu2, pu1);

    success &= compare_integer(TRUE, TimeEqual(Urgency_GetRestartBy(pu1),
					       Urgency_GetRestartBy(pu2)),
			       "dest restartby == source restartby");

    success &= compare_integer(Urgency_GetCriticality(pu1),
			       Urgency_GetCriticality(pu2),
			       "dest criticality == source criticality");

    return success;
}

BOOL Test_Urgency_RestartBy(void)
{
    BOOL success = TRUE;
    TIME tNew;

    Int32ToInt64(tNew, 2222);
    Urgency_SetRestartBy(pu1, tNew);

    success &= compare_integer(TRUE, TimeEqual(Urgency_GetRestartBy(pu1),
					       tNew),
			       "set RestartBy == get RestartBy");

    return success;
}

BOOL Test_Urgency_Criticality(void)
{
    BOOL success = TRUE;

    Urgency_SetCritical(pu1);

    success &= compare_integer(TRUE, Urgency_IsCritical(pu1),
			       "urgency is critical");


    Urgency_ClearCritical(pu1);

    success &= compare_integer(FALSE, Urgency_IsCritical(pu1),
			       "urgency is not critical");

    Urgency_SetTimeSharing(pu1);

    return success;
}

BOOL Test_Urgency_TimeSharing(void)
{
    BOOL success = TRUE;

    Urgency_SetTimeSharing(pu1);

    success &= compare_integer(TRUE, Urgency_IsTimeSharing(pu1),
			       "urgency is timesharing");

    Urgency_ClearTimeSharing(pu1);

    success &= compare_integer(FALSE, Urgency_IsTimeSharing(pu1),
			       "urgency is not timesharing");

    return success;
}

BOOL Test_Urgency_EQ(void)
{
    BOOL success = TRUE;

    TIME t1;
    TIME t2;

    Int32ToInt64(t1, 2222);
    Int32ToInt64(t2, 222222);

    Urgency_SetTimeSharing(pu1);
    Urgency_SetTimeSharing(pu2);

    Urgency_SetRestartBy(pu1, t1);
    Urgency_SetRestartBy(pu2, t1);

    success &= compare_integer(TRUE, Urgency_EQ(pu1, pu2),
			       "timesharing w/ same restartbys are equal");

    Urgency_SetCritical(pu2);

    success &= compare_integer(FALSE, Urgency_EQ(pu1, pu2),
			       "critical != timesharing w/ same restartby");

    Urgency_SetCritical(pu1);

    success &= compare_integer(TRUE, Urgency_EQ(pu1, pu2),
			       "critical w/ same restartbys are equal");

    Urgency_SetRestartBy(pu2, t2);

    success &= compare_integer(FALSE, Urgency_EQ(pu1, pu2),
			       "critical w/ sooner restartby != later restartby");

    Urgency_SetTimeSharing(pu1);
    Urgency_SetTimeSharing(pu2);

    success &= compare_integer(FALSE, Urgency_EQ(pu1, pu2),
			       "timesharing w/ sooner restartby != later restartby");

    success &= compare_integer(TRUE, Urgency_EQ(pu1, pu1),
			       "urgency equals itself");

    return success;
}

BOOL Test_Urgency_GTEQ(void)
{
    BOOL success = TRUE;

    TIME t1;
    TIME t2;

    Int32ToInt64(t1, 2222);
    Int32ToInt64(t2, 222222);

    Urgency_SetTimeSharing(pu1);
    Urgency_SetTimeSharing(pu2);

    Urgency_SetRestartBy(pu1, t1);
    Urgency_SetRestartBy(pu2, t1);

    success &= compare_integer(TRUE, Urgency_GTEQ(pu1, pu2),
			       "timesharing w/ same restartbys are >=");

    Urgency_SetCriticality(pu2, _CRITICAL);

    success &= compare_integer(FALSE, Urgency_GTEQ(pu1, pu2),
			       "timesharing !>= critical w/ same restartby");

    Urgency_SetCriticality(pu1, _CRITICAL);

    success &= compare_integer(TRUE, Urgency_GTEQ(pu1, pu2),
			       "critical w/ same restartbys are GTEQ each other");

    success &= compare_integer(TRUE, Urgency_GTEQ(pu2, pu1),
			       "critical w/ same restartbys are GTEQ each other");

    Urgency_SetRestartBy(pu2, t2);

    success &= compare_integer(TRUE, Urgency_GTEQ(pu1, pu2),
			       "critical w/ sooner restartby >= later restartby");

    Urgency_SetCriticality(pu1, _TIMESHARING);
    Urgency_SetCriticality(pu2, _TIMESHARING);

    success &= compare_integer(TRUE, Urgency_GTEQ(pu1, pu2),
			       "timesharing w/ sooner restartby != later restartby");

    success &= compare_integer(TRUE, Urgency_GTEQ(pu1, pu1),
			       "urgency >= itself");

    return success;
}

BOOL Test_Urgency_GT(void)
{
    BOOL success = TRUE;

    TIME t1;
    TIME t2;

    Int32ToInt64(t1, 2222);
    Int32ToInt64(t2, 222222);

    Urgency_SetTimeSharing(pu1);
    Urgency_SetTimeSharing(pu2);

    Urgency_SetRestartBy(pu1, t1);
    Urgency_SetRestartBy(pu2, t1);

    success &= compare_integer(FALSE, Urgency_GT(pu1, pu2),
			       "timesharing w/ same restartbys are !>");

    Urgency_SetCriticality(pu2, _CRITICAL);

    success &= compare_integer(FALSE, Urgency_GT(pu1, pu2),
			       "timesharing !> critical w/ same restartby");

    Urgency_SetCriticality(pu1, _CRITICAL);

    success &= compare_integer(FALSE, Urgency_GT(pu1, pu2),
			       "critical w/ same restartbys are !> each other");

    Urgency_SetRestartBy(pu2, t2);

    success &= compare_integer(TRUE, Urgency_GT(pu1, pu2),
			       "critical w/ sooner restartby > later restartby");

    success &= compare_integer(FALSE, Urgency_GT(pu2, pu1),
			       "critical w/ later restartby !> later restartby");

    Urgency_SetCriticality(pu1, _TIMESHARING);
    Urgency_SetCriticality(pu2, _TIMESHARING);

    success &= compare_integer(TRUE, Urgency_GT(pu1, pu2),
			       "timesharing w/ sooner restartby > later restartby");

    success &= compare_integer(FALSE, Urgency_GT(pu1, pu1),
			       "urgency !> itself");

    return success;
}
