/*
CB.C - Principally for testing Condition_Broadcast() hence the name. 

  Basic idea is to create a user-specified number of threads, line them up on the same
  condition variable('starting line'), then turn 'em loose with either a Condition_Broadcast()
  or a Condition_Signal(), then check to see if they got the signal ... 
  The threads can optionally wait for a timed interval, or forever. 
  They share an optional mutex, which serves the dual purpose of protecting the console 
  print outputs, and of testing basic mutex locking functionality. 
  The test ensures that the condition signalling APIs work, but the user must inspect output 
  with some knowledge of how the suite works to really benefit. 
  Try running cb.exe with the -mutex and -broadcast flags set to 1 and 0 in various combos. 

  Tests Condition_Broadcast(), Lock_Mutex(), Unlock_Mutex(), Condition_Signal(), and atomic funcs. 
Drives thread proc in wt.c.

Copyright (c) Microsoft Corporation All rights reserved. 

Eric Fowler 
*/

#include <mmlite.h>
#include <stdio.h>
#include <tchar.h>
#include <stdlib.h>
#include <assert.h>
#include "../utilities/utility.h"
#include "wt.h"

/* DONT put these on stack */
wait_care_pkg wcp = {0,};   //'care package' of shared data for thread proc
MUTEX       mtx_console;
TIME_CONSTRAINT constraint;  /* for constraint based waits if req'd */
CONDITION   cv; 

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE       sc;
    _TCHAR      *stop; 
    UINT32      nThreads;
    UINT32      nStart = 0, nFinish = 0; //flags for starting and finishing 
    UINT32      nStackSize = DEFAULT_STACK_SIZE;
    UINT32      idx = 0;
    BOOL        bBroadcast; 
    //temporary for storing cmd line variables for processing
    const _TCHAR      * sArg = 0;

    IProcess * pProc = CurrentProcess();

    memset(&constraint, 0, sizeof(constraint));
    ParseCmdLine(&argc, &argv); 

    OptionalBreak(argc, argv);

    if(argc < 2)
    {
        CBUsage(argv[0]); 
        _tprintf(_T("\n\rMain:Proceeding with default values")); 
    }

    //------how many threads are we using?-------
    nThreads = DEFAULT_THREAD_NUMBER; 

    sArg = GetArgFromCommandLine(argc, argv, _T("threads")); 
    if(sArg)
        if(*sArg)
            nThreads = _tcstoul(sArg, &stop, 10); 


    //-------default stack size, or do we change it?-------------
    nStackSize = DEFAULT_STACK_SIZE; 
    sArg = GetArgFromCommandLine(argc, argv, _T("stack")); 
    if(sArg)
        if(*sArg)
        {
            nStackSize = _tcstoul(sArg, &stop, 10); 
            if(nStackSize)
                CurrentProcess()->v->Init(CurrentProcess(), nStackSize,
                                          NULL, NULL, NULL);
        }



    //------does user want a mutex?--------
    wcp.pmtx_console = NULL; 
    sArg = GetArgFromCommandLine(argc, argv, _T("mutex")); 
    if(sArg)
    {
        if(*sArg)
        {
            if(_tcstoul(sArg, &stop, 10))
            {
                //yes he does ... give him one
                Mutex_Init(&mtx_console);
                CheckLock(&mtx_console,  FALSE, _T(__FILE__), __LINE__);
                wcp.pmtx_console = &mtx_console;
            }
            else
                //null mutex
                wcp.pmtx_console = NULL; 
        }
    }

    //------are we broadcasting a signal, or just signalling in a loop?--------

    bBroadcast = FALSE;
    sArg = GetArgFromCommandLine(argc, argv, _T("broadcast")); 
    if(sArg)
    {
        if(*sArg)
        {
            if(_tcstoul(sArg, &stop, 10))
                //Broadcast signals 
                bBroadcast = TRUE;
            else
                bBroadcast = FALSE;
        }
    }

    //---set timeouts for thread waits-------
    wcp.tm = TIME_FOREVER; 

    //note that time is a 64 bit value
    //we are taking only 32 bits from cmd line and converting ... 
    sArg = GetArgFromCommandLine(argc, argv, _T("time")); 
    if(sArg)
    {
        if(*sArg)
        {
            //time in millisecs, relative not absolute
            Int32ToInt64(wcp.tm, TIME_RELATIVE(TIME_MILLIS(_tcstol(sArg, &stop, 10))));
        }
        else
            ; //make stupid autoindent happy :-(
    }
    else
        ; 
 
    
    //we load up with defaults
    //if user specified start, end, criticality, etc., we will overwrite later
    //START is already set to zero near top of _tmain()
    Int32ToInt64(constraint.Estimate, DEFAULT_ESTIMATE);
    Int32ToInt64(constraint.Deadline, DEFAULT_DEADLINE);
    constraint.Criticality = _NONCRITICAL;
    wcp.bEndPrevious    = FALSE;

    //are we using constraints?
    wcp.pCons = NULL; 
    sArg = GetArgFromCommandLine(argc, argv, _T("constraint"));
    if(sArg)
    {
        if(*sArg)
            if(_tcstoul(sArg, &stop, 10))
               wcp.pCons = &constraint; 
            else
               wcp.pCons = NULL; 
        else
            ; 
    }


    //more constraint stuff ... start time ... 
    sArg = GetArgFromCommandLine(argc, argv, _T("start"));
    if(sArg)
        if(*sArg)
        {
            //time in millisecs, relative not absolute
            Int32ToInt64(constraint.Start, TIME_RELATIVE(TIME_MILLIS(_tcstol(sArg, &stop, 10))));
        }
    //constraint.estimate .... 
    sArg = GetArgFromCommandLine(argc, argv, _T("estimate"));
    if(sArg)
        if(*sArg)
        {
            //time in millisecs, relative not absolute
            Int32ToInt64(constraint.Estimate, TIME_RELATIVE(TIME_MILLIS(_tcstol(sArg, &stop, 10))));
        }
   
    //constraint.deadline
    sArg = GetArgFromCommandLine(argc, argv, _T("deadline"));
    if(sArg)
        if(*sArg)
        {
            //time in millisecs, relative not absolute
            Int32ToInt64(constraint.Deadline, TIME_RELATIVE(TIME_MILLIS(_tcstol(sArg, &stop, 10)))); 
        }
   
    sArg = GetArgFromCommandLine(argc, argv, _T("endprev"));
    if(sArg)
        if(*sArg)
            wcp.bEndPrevious = _tcstoul(sArg, &stop, 10);
   
    sArg = GetArgFromCommandLine(argc, argv, _T("record"));
    if(sArg)
        if(*sArg)
            wcp.bRecordTimeTaken = _tcstoul(sArg, &stop, 10);
        else
            ; //make stupid autoindent happy :-(
    else
        ; 

    nStart = nThreads;
 
    //initialize condition var to pass to thread proc
    //we have already initialized (possibly NULL) mutex above
    Condition_Init(&cv); 
    
    //initialize care package of parms for the threads ... we share a mutex, two 
    //counters to tell how many threads are in various states .. 
    wcp.pcv = &cv; 
    wcp.nStart = nStart;
    wcp.nFinish = nFinish;

    //wcp.pmtx_console has already been set ... 

    CBStatus(&wcp, bBroadcast, nStackSize);
 
    //create a set of threads
    //these will wait on the mutex and C.V. in the thread proc
    for(idx = 0; idx < nThreads; ++idx)
    {
        sc = pProc->v->CreateThread(pProc, 0, WaitThreadProc, &wcp,
                                    0, NULL, NULL); 
        if(FAILED(sc)) {
            _tprintf(_T("CreateThread failed! x%x\r\n"),sc);
            return SetAndReturn(1);
        }
    }

    //wait until the thread is started
    //WaitThreadProc() will decrement this variable by 1
    //When we get to zero all threads have kicked off and are (probably) 
    //waiting on condition variable
    for(;!AtomicCmpAndSwap(&wcp.nStart, 0, 0);)
        ;

    SafePrint(_T("\n\rMain sees threads are started"), wcp.pmtx_console);
    
    //wake up our threads by user-specified method
    if(bBroadcast)
    {
        //user wanted broadcast
        //Provided user checks to see all threads woke up (screen output), 
        //this is a fair test that C_B(&cv) will wake ALL threads waiting on 
        //an event. 
        Condition_Broadcast(wcp.pcv); 
    
        //wait until the threads finish
        //Each thread proc increments *pFinish once AFTER wait is complete
        //When *pFinish == nThreads all threads are awake and heading for exit
        for(;!AtomicCmpAndSwap(&wcp.nFinish, nThreads, nThreads);)
            /*null*/;
    }
    //narrowcast wakeups
    else
        //signal repeatedly until all threads are known to be past the checkpoint
        //Note that this signals nThreads times. If we have nThreads waiting on 
        //the condition var, this must wake up NO LESS THAN one thread per click
        //or threads will be 'marooned' on the condition variable and will never 
        //finish ... user must check thread output to get results of this test,
        //if N threads are started, all N must report passing the starting line
        //This is NOT a test that Condition_Signal() wakes NO MORE THAN one thread
        //per tick ... that is below. 
        for(idx = 0; !AtomicCmpAndSwap(&wcp.nFinish, nThreads, nThreads); ++idx)
        {
            if(wcp.pmtx_console)
            {
                CheckLock(&mtx_console,  FALSE, _T(__FILE__), __LINE__);
                Mutex_Lock(wcp.pmtx_console);
                CheckLock(&mtx_console,  TRUE, _T(__FILE__), __LINE__);
            }
 
            _tprintf(_T("\n\rMain:Signaling, %d threads have finished waiting so far ... "), wcp.nFinish + 1);
            Condition_Signal(wcp.pcv);
            
            if(wcp.pmtx_console)
            {
                CheckLock(&mtx_console,  TRUE, _T(__FILE__), __LINE__);
                Mutex_Unlock(wcp.pmtx_console);
                CheckLock(&mtx_console,  FALSE, _T(__FILE__), __LINE__);
            }
        }

        //This is a test that Condition_Signal() wakes NO MORE THAN one thread per 
        //tick. 
        if(idx < nThreads)
        {
            _TCHAR sBuf[256] = {0,};
            
            _stprintf(sBuf, _T("\n\rMain:ANOMALY:It took [%d] calls to Condition_Signal() to wake up %d threads?!?"), idx, nThreads);
            SafePrint(sBuf, wcp.pmtx_console);
        }

    SafePrint(_T("\n\rMain sees threads are finished"), wcp.pmtx_console);
  
    Condition_Destroy(wcp.pcv);

    if(wcp.pmtx_console)
    {
        CheckLock(&mtx_console, FALSE, _T(__FILE__), __LINE__);
        Mutex_Destroy(wcp.pmtx_console);
        CheckLock(&mtx_console, FALSE, _T(__FILE__), __LINE__);
    }

    FreeCmdLine(&argc, &argv); 
    return SetAndReturn(0);
}

/*
Basic usage screen. 
*/
void CBUsage(const _TCHAR * s)
{
    _tprintf(_T("\n\rUsage:%s [-mutex [0|1]][-timeout <decimal millisecs>] \n\r\t"), s);
    _tprintf(_T("[-threads <num threads>] [-stacksize <stacksize>][-broadcast 0/1]\n\r\t"));
    _tprintf(_T("[-constraint [0|1]] [-start <time in millisecs>][-estimate<time in millisecs>]\r\n\t"));
    _tprintf(_T("[-deadline <time in millisecs>][-endprev [0|1]][-record [0|1]]"));
    _tprintf(_T("\n\rWhere -mutex 0/1 uses a null/valid mutex, -timeout sets timeout in milliseconds, -threads set number of threads,\n\r\t"));
    _tprintf(_T("-stacksize sets number of bytes to use in thread stack, -broadcast sends signal as Condition_Broadcast() or Condition_Signal(),\r\n\t"));
    _tprintf(_T("-constraint uses Condition_WaitAndBeginConstraint() instead of Condition_Wait()\r\n\t"));
    _tprintf(_T("-start, -estimate, and -deadline set start, estimate, and deadlines for constrained waits, and\r\n\t"));
    _tprintf(_T("-record turns on recording of elapsed time, for constrained waits, and"));
    _tprintf(_T("\r\n\t-endprev forces non-nesting of constrained wait when turned on."));

    TellAboutOptionalBreak();
}

/*
Basic status report for user. 
*/
void CBStatus(const wait_care_pkg * p, BOOL bBC, UINT32 nStackSize)
{
    _tprintf(_T("\n\rMutex 0x%.8x"), (UINT32)p->pmtx_console);
    _tprintf(_T("\n\rThreads %d"), (UINT32)p->nStart);
    if(bBC)
        _tprintf(_T("\n\rBroadcasting signals ... "));
    else
        _tprintf(_T("\n\r'Narrowcasting' signals .... "));

    if(nStackSize != DEFAULT_STACK_SIZE)
        _tprintf(_T("\n\rSetting thread stack size to %d bytes"), nStackSize);
    else
        _tprintf(_T("\n\rThread stack size remains at default"));


    if(p->pCons)
    {
        _tprintf(
            _T(
            "\n\rUsing Condition_WaitAndBeginConstraint():\r\n\tStart:[%d] Estimate:[%d] Deadline:[%d], End Previous Constraint(t/f):[%d], Record Elapsed Time(t/f):[%d]"),
            Int64ToInt32(p->pCons->Start), Int64ToInt32(p->pCons->Estimate), Int64ToInt32(p->pCons->Deadline),
            (UINT32)p->bEndPrevious,
            (UINT32)p->bRecordTimeTaken
            );
    }
    else
        _tprintf(_T("\n\rUsing Condition_Wait() only")); 
}

/*
Sanity checking of TryLock and Locked functions
Compares status of mutex against expected status (passed in). 
If they don't match, it squawks. 
Should leave mutex in same state, but can get and release a lock in the process. 
Missing test case : Does TryLock() always get a mutex if it is available? 
If we call TryLock() inside the bLocked == FALSE branch here, we run the risks that 
(1) some other thread owns the mutex, we just don't know it; or (2) someone else gets it
between our call to CheckLock() and the call to TryLock() in CheckLock().  
Forcing that test could be quite involved, since it is hard to know which thread holds 
the mutex ... perhaps a mutex-mutex will be involved.
Output from this function is not protected from interlacing on screen becuz console is 
protected with the mutex we are testing ... 
*/
void CheckLock(MUTEX * pMTX, BOOL bLocked, const _TCHAR *sFile, unsigned nLine)
{
    if(!pMTX) return; // no need to test null mutex!
        
    /* If we expect the mutex to be locked, it is 'cuz we locked it. So, these tests 
     * can never fail if bLocked == TRUE.
     */
    if(bLocked)
    {
        if(!Mutex_Locked(pMTX))
        {
            _tprintf(_T("\n\rANOMALY: mutex [0x%.8x] is supposed to be locked\n\r\t%s, line [%d]"), (UINT32)pMTX, sFile, nLine); 
            assert(0);
        }
        
        if(Mutex_TryLock(pMTX))
        {
            _tprintf(_T("\n\rANOMALY: mutex [0x%.8x] is supposed to be locked\n\r\t%s, line [%d]"), (UINT32)pMTX, sFile, nLine); 
            Mutex_Unlock(pMTX); 
            assert(0);
        }
    }

    /* In the case of bLocked == FALSE, it just means OUR THREAD
     * (the one that called into this function) did not hold the lock when we called into here.
     * It does not mean someone *else* does not hold it, or won't get it. 
     * Still need to test case of Mutex_TryLock() failing to acquire a free mutex ... but IMO 
     * that is not an important case. 
     */
    else
        if(Mutex_Locked(pMTX))
            _tprintf(_T("\n\rMutex [0x%.8x] is supposed to be free\n\r\t%s, line [%d].\n\r\tProbably not an error unless very frequent"), (UINT32)pMTX, sFile, nLine); 
}

/*
A reporting function that uses a mutex to protect the console from overlapped writes. 
A null mutex is not a problem ... ignored. 
*/
void SafePrint(const _TCHAR * sMsg, MUTEX * pMTX)
{
    if(!sMsg)
        return; 

    if(!*sMsg)
        return; 

    if(pMTX)
    {
        CheckLock(pMTX,  FALSE, _T(__FILE__), __LINE__);
        Mutex_Lock(pMTX);
        CheckLock(pMTX,  TRUE, _T(__FILE__), __LINE__);
    }

    _tprintf(sMsg); fflush(stdout); 

    if(pMTX)
    {
       CheckLock(pMTX,  TRUE, _T(__FILE__), __LINE__);
       Mutex_Unlock(pMTX);
       CheckLock(pMTX,  FALSE, _T(__FILE__), __LINE__);
    }
}
