/*
* WaitThread - a thread process that takes a ptr to a condition variable and an integer. 
* Waits on the CV and when it fires, decrements the integer. 
* Can optionally take a mutex. This has the dual effect of testing mutex functions
* and protecting the console from interlaced writes, which harmlessly but annoyingly 
* garbles the output.
* This function touches two shared variables but only through atomic functions. 
* Precondition: pV points to a wait_care_pkg that has been populated ...
* Copyright (c) Microsoft Corporation All rights reserved 
*/

#include "wt.h"
#include "../utilities/utility.h"
#include <stdio.h>
void WaitThreadProc(THREAD_ARGUMENT pV)
{
    UINT32 n;
    SCODE sc;
    TIME tm_taken = {0,};

    wait_care_pkg * p = (wait_care_pkg *)pV;
    
//    DebugBreak();

    //the mutex is really just locking the console, ensuring debug spew is not
    //interlaced when printed to the screen ... p->pStart is "protected" in that 
    //it is accessed only thru atomic functions
    //Pass null pmtx_console to skip this part and get interlaced output
    if(p->pmtx_console) 
    {
        CheckLock(p->pmtx_console,  FALSE, _T(__FILE__), __LINE__);
        Mutex_Lock(p->pmtx_console); 
        CheckLock(p->pmtx_console,  TRUE, _T(__FILE__), __LINE__);
    }
   
    
    _tprintf(_T("\n\rWPT:Thread @%x [%d] starting, mutex @%.8x "), 
             CurrentThread(), p->nStart - 1, (UINT32)p->pmtx_console); 
    
    //strangely enuf the compiler is rejecting trinary operators 
    //p->pmtx_console ? *p->pmtx_console : 0 is being tossed ... 
    //so we use an ugly workaround 
//af: a MUTEX is a struct not an integer. 
//printf is happy to push it on the stack but the comparison fails as it should.
    
    if(p->pmtx_console)
        _tprintf(_T("thread @%.8x owns the mutex"), *p->pmtx_console); 
    else
        _tprintf(_T("thread @%.8x owns the mutex"), 0); 
    
 
    //This sets a flag tested in main that says we are about to start our wait
    //Not mutex protected, accessed only by atomic funcs 
    //When p->nStart goes to 0 all threads have started and are (presumably) waiting
    //for the signal in next line ... [not strictly true becuz a context switch between
    //next 3 lines could break that, but not likely nor really dangerous .. ]
    n = AtomicDec(&p->nStart);

    //we own the console mutex here, if it is not null; and we own it coming out, too ...
    CheckLock(p->pmtx_console,  TRUE, _T(__FILE__), __LINE__);
    
    //are we using constraint or not? Choose the right function ... 
    if(p->pCons)
        sc = Condition_WaitAndBeginConstraint(p->pcv, p->pmtx_console, &p->tm, p->bEndPrevious, p->pCons,  &tm_taken); 
    else
        sc = Condition_Wait(p->pcv, p->pmtx_console, &p->tm); 


    CheckLock(p->pmtx_console, TRUE, _T(__FILE__), __LINE__);

    //this sets a flag tested in main that just says we have finished our wait
    //this var is NOT protected by the mutex, only by atomicity, but we want it 
    //touched here so the printf() does not lie about sequence
    n = AtomicInc(&p->nFinish);

    //don't need to take mutex here, Condition_Wait*() got it for us
    _tprintf(_T("\n\rWTP:Thread [%d] done waiting, returned [0x%.8x(%s)]"), n - 1, sc, ErrorDescription(sc)); 
    if(p->pCons)
        ReportConstrainedResults(p, &tm_taken);

    fflush(stdout);

    if(p->pmtx_console)
    {
        CheckLock(p->pmtx_console,  TRUE, _T(__FILE__), __LINE__);
        Mutex_Unlock(p->pmtx_console);
        CheckLock(p->pmtx_console,  FALSE, _T(__FILE__), __LINE__);
    }
}

/*
Report relevant stuff for constrained wait. 
Preconditions:
    - BOTH pointers are non-NULL
    - Does not take mutex so caller must protect console.
*/
void ReportConstrainedResults(const wait_care_pkg * p, const TIME * pTM)
{
    _tprintf(_T("\n\rConstraints:\r\n\tStart:[%d msecs] Estimate:[%d msecs] Deadline:[%d msecs]")\
        _T("End Previous(t/f):[%d] Report Elapsed Time(t/f):[%d] Elapsed Time:[%d msecs]"), 
        Int64ToInt32(p->pCons->Start), Int64ToInt32(p->pCons->Estimate), Int64ToInt32(p->pCons->Deadline), 
        p->bEndPrevious,  
        p->bRecordTimeTaken, 
        Int64ToInt32(*pTM));
}
