/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Simple test program for FRAM board from Texas A&M University.
 * Hardware design: Richard Pittman
 * Hardware testing: Richard Pittman and Oscar Alameida
 * Software tests: Richard Pittman and Alessandro Forin
 */

#include <mmlite.h>
#include <tchar.h>
#include <stdio.h>
#include <stdlib.h>
#include <arm/at91m63200.h>

#define START_ADDRESS2 ((char *)0x20000000)
#define START_ADDRESS3 ((char *)0x03000000)
#define FRAM_SIZE      (256*1024)

char *FramStartAddress = START_ADDRESS3;
int  FramChipSelect = 3;

#define USE_PB15 1
#if USE_PB15
#define BlipInit()  \
        PioB->SetData = PIOB_PB15;
#define Blip() { \
        PioB->ClearData = PIOB_PB15; \
        PioB->SetData = PIOB_PB15; \
}
#else
#define BlipInit()
#define Blip()
#endif

/* Sleep for a few milliseconds
 */
PRIVATE void MilliSleep(int mSecs)
{
    TIME t;
    Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(mSecs)));
    SleepUntil(t);
}

BOOL Test(UINT Verbose)
{
    volatile UINT32 *p32, v32, t;
    volatile UINT16 *p16, v16a, v16b;
    volatile UINT8 *p8, v8a, v8b, v8c, v8d;
    int i;
    UINT Pass, Fail;

#define reset() \
    p32 = (volatile UINT32 *)FramStartAddress; \
    p16 = (volatile UINT16 *)FramStartAddress; \
    p8  = (volatile UINT8  *)FramStartAddress;

    /* First trivia test: make sure we can read it all
     * and that we read the same things in 32/16/8 bit chunks
     */
    reset();
    Pass = Fail = 0;

    for (i = 0; i < (FRAM_SIZE/4); i++) {
        v32 = *p32++;
        v16a = *p16++;
        v16b = *p16++;
        v8a = *p8++;
        v8b = *p8++;
        v8c = *p8++;
        v8d = *p8++;
        if (Verbose > 1)
            _tprintf(_TEXT("[%d] %x {%x %x} {%x %x %x %x}\n"), i,
                   v32,
                   v16a, v16b,
                   v8a, v8b, v8c, v8d);

        t = (v16b << 16) | v16a;
        if (t != v32) {
        if (Verbose)
            _tprintf(_TEXT("\tBUG16 @%x: %x != %x\n"), p32-1, v32, t);
            Fail++;
        } else Pass++;

        t = (v8d << 24) | (v8c << 16) | (v8b << 8) | v8a;
        if (t != v32) {
        if (Verbose)
            _tprintf(_TEXT("\tBUG8  @%x: %x != %x\n"), p32-1, v32, t);
            Fail++;
        } else Pass++;

        if (Fail > 100) {
        if (Verbose)
            _tprintf(_TEXT("That looks pretty busted uhu?\n"));
            goto Out;
        }

    }

 Out:
    _tprintf(_TEXT("Test0 %s Pass=%d Fail=%d\n"),
             (Fail == 0) ? _TEXT("Ok") : _TEXT("FAILED"), Pass, Fail);
    return (Fail == 0);
}

BOOL Test1(UINT Verbose)
{
    volatile UINT32 *p32, v32, t;
    volatile UINT16 *p16, v16a, v16b;
    volatile UINT8 *p8, v8a, v8b, v8c, v8d;
    volatile UINT32 val;
    int i;
    UINT Pass, Fail;

    /* Write 32bit values and check that what we write takes
     * Values are 1...N sequential
     */
    reset();
    Pass = Fail = 0;
    val = 1;

    for (i = 0; i < (FRAM_SIZE/4); i++) {
		// *p32 = (UINT32)(i);
		*p32 = val;
        v32 = *p32++;
        v16a = *p16++;
        v16b = *p16++;
        v8a = *p8++;
        v8b = *p8++;
        v8c = *p8++;
        v8d = *p8++;
		
        if (Verbose)
            _tprintf(_TEXT("[%x] %x {%x %x} {%x %x %x %x}\n"), i,
                   v32,
                   v16a, v16b,
                   v8a, v8b, v8c, v8d);

        t = (v16b << 16) | v16a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG16 @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

        t = (v8d << 24) | (v8c << 16) | (v8b << 8) | v8a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG8  @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

		if (val != v32) {
			_tprintf(_TEXT("\t[%x]BUG32  @%x: %x != %x\n"), val, p32-1, val, v32);
            Fail++;
		} else Pass++;

        if (Fail > 100) {
            _tprintf(_TEXT("That looks pretty busted uhu?\n"));
            goto Out;
        }
		val = val+1;
		if (val == 0) val = 1;
    }

 Out:
    _tprintf(_TEXT("Test1 %s Pass=%d Fail=%d\n"),
             (Fail == 0) ? _TEXT("Ok") : _TEXT("FAILED"), Pass, Fail);
    return (Fail == 0);
}

BOOL Test2(UINT Verbose)
{
    volatile UINT32 *p32, v32, t;
    volatile UINT16 *p16, v16a, v16b;
    volatile UINT8 *p8, v8a, v8b, v8c, v8d;
    volatile UINT32 val;
    int i;
    UINT Pass, Fail;

    /* Write 32bit values and check that what we write takes
     * Values are 2^N sequential
     */
    reset();
    Pass = Fail = 0;
	val = 1;

	for (i = 0; i < (FRAM_SIZE/4); i++) {
		// *p32 = (UINT32)(i);
		*p32 = val;
        v32 = *p32++;
        v16a = *p16++;
        v16b = *p16++;
        v8a = *p8++;
        v8b = *p8++;
        v8c = *p8++;
        v8d = *p8++;
		
        if (Verbose)
            _tprintf(_TEXT("[%x] %x {%x %x} {%x %x %x %x}\n"), i,
                   v32,
                   v16a, v16b,
                   v8a, v8b, v8c, v8d);

        t = (v16b << 16) | v16a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG16 @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

        t = (v8d << 24) | (v8c << 16) | (v8b << 8) | v8a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG8  @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

		if (val != v32) {
			_tprintf(_TEXT("\t[%x]BUG32  @%x: %x != %x\n"), val, p32-1, val, v32);
            Fail++;
		} else Pass++;

        if (Fail > 150) {
            _tprintf(_TEXT("That looks pretty busted uhu?\n"));
            goto Out;
        }
		val = val<<1;
		if (val == 0) val = 1;
    }
 Out:
    _tprintf(_TEXT("Test2 %s Pass=%d Fail=%d\n"),
             (Fail == 0) ? _TEXT("Ok") : _TEXT("FAILED"), Pass, Fail);
    return (Fail == 0);
}

BOOL Test3(UINT Verbose)
{
    volatile UINT32 *p32, v32, t;
    volatile UINT16 *p16, v16a, v16b;
    volatile UINT8 *p8, v8a, v8b, v8c, v8d;
    volatile UINT32 val;
    int i;
    UINT Pass, Fail;

    /* Write 32bit values and check that what we write takes
     * Values are ~2^N sequential
     */
    reset();
    Pass = Fail = 0;

	val = 0xfffffffe;
	for (i = 0; i < (FRAM_SIZE/4); i++) {
		//  *p32 = (UINT32)(i);
		*p32 = val;
        v32 = *p32++;
        v16a = *p16++;
        v16b = *p16++;
        v8a = *p8++;
        v8b = *p8++;
        v8c = *p8++;
        v8d = *p8++;
		
        if (Verbose)
            _tprintf(_TEXT("[%x] %x {%x %x} {%x %x %x %x}\n"), i,
                   v32,
                   v16a, v16b,
                   v8a, v8b, v8c, v8d);

        t = (v16b << 16) | v16a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG16 @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

        t = (v8d << 24) | (v8c << 16) | (v8b << 8) | v8a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG8  @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

		if (val != v32) {
			_tprintf(_TEXT("\t[%x]BUG32  @%x: %x != %x\n"), val, p32-1, val, v32);
            Fail++;
		} else Pass++;

        if (Fail > 200) {
            _tprintf(_TEXT("That looks pretty busted uhu?\n"));
            goto Out;
        }
		val = (val<<1)+1;
		if (val == 0xffffffff) val = 0xfffffffe;
    }

 Out:
    _tprintf(_TEXT("Test3 %s Pass=%d Fail=%d\n"),
             (Fail == 0) ? _TEXT("Ok") : _TEXT("FAILED"), Pass, Fail);
    return (Fail == 0);
}

BOOL Test4(UINT Verbose)
{
    volatile UINT32 *p32, v32, t;
    volatile UINT32 val;
    volatile UINT16 *p16, v16a, v16b;
    volatile UINT8 *p8, v8a, v8b, v8c, v8d;
	UINT32 val1, val2, val3, val4, val5;
    int i;
    UINT Pass, Fail;

    /* Write 32bit values and check that what we write takes
     * Base values are 1..N sequential; each step writes 6 times;
     * The base values is repeated into the higher nibbles
     */
    reset();
    Pass = Fail = 0;

	val = 1;
	for (i = 0; i < (FRAM_SIZE/4); i++) {
		val1 = (val<<4)+val;
		val2 = (val1<<4)+val1;
		val3 = (val2<<4)+val2;
		val4 = (val3<<4)+val3;
		val5 = (val4<<4)+val4;
		// *p32 = (UINT32)(i);
		*p32 = val1;
		*p32 = val2;
		*p32 = val3;
		*p32 = val4;
		*p32 = val5;
		*p32 = val;
		
        v32 = *p32++;
        v16a = *p16++;
        v16b = *p16++;
        v8a = *p8++;
        v8b = *p8++;
        v8c = *p8++;
        v8d = *p8++;
		
        if (Verbose)
            _tprintf(_TEXT("[%x] %x {%x %x} {%x %x %x %x}\n"), i,
                   v32,
                   v16a, v16b,
                   v8a, v8b, v8c, v8d);

        t = (v16b << 16) | v16a;
        if (t != val) {
			_tprintf(_TEXT("\t[[%x][%x][%x][%x][%x]]\n"),val1,val2,val3,val4,val5);
            _tprintf(_TEXT("\t[%x]BUG16 @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

        t = (v8d << 24) | (v8c << 16) | (v8b << 8) | v8a;
        if (t != val) {
			_tprintf(_TEXT("\t[[%x][%x][%x][%x][%x]]\n"),val1,val2,val3,val4,val5);
            _tprintf(_TEXT("\t[%x]BUG8  @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

		if (val != v32) {
			_tprintf(_TEXT("\t[[%x][%x][%x][%x][%x]]\n"),val1,val2,val3,val4,val5);
			_tprintf(_TEXT("\t[%x]BUG32  @%x: %x != %x\n"), val, p32-1, val, v32);
            Fail++;
		} else Pass++;

        if (Fail > 250) {
            _tprintf(_TEXT("That looks pretty busted uhu?\n"));
            goto Out;
        }
		val = val+1;
		if (val == 0) val = 1;
    }

 Out:
    _tprintf(_TEXT("Test4 %s Pass=%d Fail=%d\n"),
             (Fail == 0) ? _TEXT("Ok") : _TEXT("FAILED"), Pass, Fail);
    return (Fail == 0);
}

BOOL Test5(UINT Verbose)
{
    volatile UINT32 *p32, v32, t;
    volatile UINT32 val;
    volatile UINT16 *p16, v16a, v16b;
    volatile UINT8 *p8, v8a, v8b, v8c, v8d;
    int i;
    UINT Pass, Fail;

    /* Write 32bit values and check that what we write takes
     * Values are 1..N
     * They are all written first, then checked.
     */
    reset();

    memset((void*)p32,0,FRAM_SIZE);
	val = 1;
	for (i = 0; i < (FRAM_SIZE/4); i++) {
        Blip();
        *p32++ = val;
		val = val+1;
		if (val == 0) val = 1;
    }

    reset();
    Pass = Fail = 0;

	val = 1;
	for (i = 0; i < (FRAM_SIZE/4); i++) {
        Blip();
        v32 = *p32++;
        v16a = *p16++;
        v16b = *p16++;
        Blip();
        v8a = *p8++;
        v8b = *p8++;
        v8c = *p8++;
        v8d = *p8++;
        Blip();
		
        if (Verbose)
            _tprintf(_TEXT("[%x] %x {%x %x} {%x %x %x %x}\n"), i,
                   v32,
                   v16a, v16b,
                   v8a, v8b, v8c, v8d);

        t = (v16b << 16) | v16a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG16 @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

        t = (v8d << 24) | (v8c << 16) | (v8b << 8) | v8a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG8  @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

		if (val != v32) {
			_tprintf(_TEXT("\t[%x]BUG32  @%x: %x != %x\n"), val, p32-1, val, v32);
            Fail++;
		} else Pass++;

        if (Fail > 100) {
            _tprintf(_TEXT("That looks pretty busted uhu?\n"));
            goto Out;
        }
		val = val+1;
		if (val == 0) val = 1;
    }

 Out:
    _tprintf(_TEXT("Test5 %s Pass=%d Fail=%d\n"),
             (Fail == 0) ? _TEXT("Ok") : _TEXT("FAILED"), Pass, Fail);
    return (Fail == 0);
}

BOOL Test6(UINT Verbose)
{
    volatile UINT32 *p32, v32, t;
    volatile UINT32 val;
    volatile UINT16 *p16, v16a, v16b;
    volatile UINT8 *p8, v8a, v8b, v8c, v8d;
    int i;
    UINT Pass, Fail;

    /* Write 32bit values and check that what we write takes
     * Values are alternating AAAAAAAA and 55555555
     * They are all written first, then checked.
     */
    reset();

    memset((void*)p32,0,FRAM_SIZE);
	val = 0xaaaaaaaa;
	for (i = 0; i < (FRAM_SIZE/4); i++) {
        *p32++ = val;
        Blip();


        val = val+1;
		if (val == 0) val = 1;

		val = ~val;
    }

    reset();
    Pass = Fail = 0;

	val = 0xaaaaaaaa;
	for (i = 0; i < (FRAM_SIZE/4); i++) {
        Blip();
        v32 = *p32++;
        v16a = *p16++;
        v16b = *p16++;
        Blip();
        v8a = *p8++;
        v8b = *p8++;
        v8c = *p8++;
        v8d = *p8++;
        Blip();
		
        if (Verbose)
            _tprintf(_TEXT("[%x] %x {%x %x} {%x %x %x %x}\n"), i,
                   v32,
                   v16a, v16b,
                   v8a, v8b, v8c, v8d);

        t = (v16b << 16) | v16a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG16 @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

        t = (v8d << 24) | (v8c << 16) | (v8b << 8) | v8a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG8  @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

		if (val != v32) {
			_tprintf(_TEXT("\t[%x]BUG32  @%x: %x != %x\n"), val, p32-1, val, v32);
            Fail++;
		} else Pass++;

        if (Fail > 100) {
            _tprintf(_TEXT("That looks pretty busted uhu?\n"));
            goto Out;
        }
		val = ~val;
    }

 Out:
    _tprintf(_TEXT("Test6 %s Pass=%d Fail=%d\n"),
             (Fail == 0) ? _TEXT("Ok") : _TEXT("FAILED"), Pass, Fail);
    return (Fail == 0);
}

BOOL Test7(UINT Verbose)
{
    volatile UINT32 *p32, v32, t;
    volatile UINT32 val;
    volatile UINT16 *p16, v16a, v16b;
    volatile UINT8 *p8, v8a, v8b, v8c, v8d;
    int i;
    UINT Pass, Fail;

    /* Write 32bit values and check that what we write takes
     * Values are alternating AAAAAAAA and 55555555
     * They are all written first, then checked.
     */
    reset();

    memset((void*)p32,0,FRAM_SIZE);
	val = 0xaaaaaaaa;
	for (i = 0; i < (FRAM_SIZE/4); i++) {
        *p32++ = val;
        Blip();


        val = val+1;
		if (val == 0) val = 1;

		val = ~val;
    }

    reset();
    Pass = Fail = 0;

	val = 0xaaaaaaaa;
	for (i = 0; i < (FRAM_SIZE/4); i++) {
        Blip();
        v32 = *p32++;
        v16a = *p16++;
        v16b = *p16++;
        Blip();
        v8a = *p8++;
        v8b = *p8++;
        v8c = *p8++;
        v8d = *p8++;
        Blip();
		
        if (Verbose)
            _tprintf(_TEXT("[%x] %x {%x %x} {%x %x %x %x}\n"), i,
                   v32,
                   v16a, v16b,
                   v8a, v8b, v8c, v8d);

        t = (v16b << 16) | v16a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG16 @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

        t = (v8d << 24) | (v8c << 16) | (v8b << 8) | v8a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG8  @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

		if (val != v32) {
			_tprintf(_TEXT("\t[%x]BUG32  @%x: %x != %x\n"), val, p32-1, val, v32);
            Fail++;
		} else Pass++;

        if (Fail > 100) {
            _tprintf(_TEXT("That looks pretty busted uhu?\n"));
            goto Out;
        }
		val = ~val;
    }

 Out:
    _tprintf(_TEXT("Test6 %s Pass=%d Fail=%d\n"),
             (Fail == 0) ? _TEXT("Ok") : _TEXT("FAILED"), Pass, Fail);
    return (Fail == 0);
}

BOOL Test8(UINT Verbose)
{
    volatile UINT32 *p32, v32, t;
    volatile UINT32 val;
    volatile UINT16 *p16, v16a, v16b;
    volatile UINT8 *p8, v8a, v8b, v8c, v8d;
    int i;
    UINT Pass, Fail;

    /* Write 32bit values and check that what we write takes
     * Values are alternating AAAAAAAA and 55555555
     * They are  written, then checked, then the whole array is checked.
     */
    reset();
    Pass = Fail = 0;

	val = 0xaaaaaaaa;
	for (i = 0; i < (FRAM_SIZE/4); i++) {
        *p32 = val;
        Blip();
        v32 = *p32++;
        v16a = *p16++;
        v16b = *p16++;
        Blip();
        v8a = *p8++;
        v8b = *p8++;
        v8c = *p8++;
        v8d = *p8++;
        Blip();
		
        if (Verbose)
            _tprintf(_TEXT("[%x] %x {%x %x} {%x %x %x %x}\n"), i,
                   v32,
                   v16a, v16b,
                   v8a, v8b, v8c, v8d);

        t = (v16b << 16) | v16a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG16 @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

        t = (v8d << 24) | (v8c << 16) | (v8b << 8) | v8a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x]BUG8  @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

		if (val != v32) {
			_tprintf(_TEXT("\t[%x]BUG32  @%x: %x != %x\n"), val, p32-1, val, v32);
            Fail++;
		} else Pass++;

        if (Fail > 100) {
            _tprintf(_TEXT("That looks pretty busted uhu?\n"));
            goto Out;
        }
		val = ~val;
    }

    reset();

	val = 0xaaaaaaaa;
	for (i = 0; i < (FRAM_SIZE/4); i++) {
        Blip();
        v32 = *p32++;
        v16a = *p16++;
        v16b = *p16++;
        Blip();
        v8a = *p8++;
        v8b = *p8++;
        v8c = *p8++;
        v8d = *p8++;
        Blip();
		
        if (Verbose)
            _tprintf(_TEXT("[%x] %x {%x %x} {%x %x %x %x}\n"), i,
                   v32,
                   v16a, v16b,
                   v8a, v8b, v8c, v8d);

        t = (v16b << 16) | v16a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x].BUG16 @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

        t = (v8d << 24) | (v8c << 16) | (v8b << 8) | v8a;
        if (t != val) {
            _tprintf(_TEXT("\t[%x].BUG8  @%x: %x != %x\n"), val, p32-1, val, t);
            Fail++;
        } else Pass++;

		if (val != v32) {
			_tprintf(_TEXT("\t[%x].BUG32  @%x: %x != %x\n"), val, p32-1, val, v32);
            Fail++;
		} else Pass++;

        if (Fail > 100) {
            _tprintf(_TEXT("That looks pretty busted uhu?\n"));
            goto Out;
        }
		val = ~val;
    }

 Out:
    _tprintf(_TEXT("Test8 %s Pass=%d Fail=%d\n"),
             (Fail == 0) ? _TEXT("Ok") : _TEXT("FAILED"), Pass, Fail);
    return (Fail == 0);
}


void PowerOn(void)
{
    PioB->SetData   = PIOB_PB16;
}

void PowerOff(void)
{
    PioB->ClearData   = PIOB_PB16;
}

void PowerInit(int ChipSelect)
{
    _tprintf(_TEXT("Turning power on..."));

    /* Make sure the CS is right */
    if (ChipSelect == 2) {
        FramStartAddress = START_ADDRESS2;
    } else {
        FramStartAddress = START_ADDRESS3;
    }

    FramChipSelect = ChipSelect;
    TheEbi->ChipSelect[FramChipSelect] = (ADDRESS)FramStartAddress | 0x2f3d; //0x2edd?;

    /* We'll be using PB16 */
    PioB->Enable    = PIOB_PB16;
    PioB->OutEnable = PIOB_PB16;
}

int _tmain(int argc, _TCHAR **argv)
{
    UINT Verbose = 0;
    BOOL Powerup = FALSE;
    int cs = 3, max = 5;
    int i;
    BOOL AllOk = TRUE;

    ParseCmdLine(&argc, &argv);

    if (argc > 1) {
        if (argv[1][0] == 'p') {
            Powerup = TRUE;
            if (argv[1][1] == '2')
                cs = 2;
            /* else stays at 3 */
        }
        else if (argv[1][0] == 'n') {
            max = _ttoi(argv[1]+1);
        } else
            Verbose = 1;
    }
    if (argc > 2) Verbose++;

    FreeCmdLine(&argc, &argv);

    if (Powerup) {
        /* Turn on power to the board */
        PowerInit(cs);

        MilliSleep(3000);

        PowerOn();

        MilliSleep(3000);

    }
    
    _tprintf(_TEXT("Testing FRAM at address %x, ChipSelect %d, power at PB16\n"), 
           FramStartAddress, FramChipSelect);

    AllOk &= Test8(Verbose);

#if 0
    AllOk &= Test(Verbose);
    AllOk &= Test1(Verbose);
    AllOk &= Test2(Verbose);
    AllOk &= Test3(Verbose);
    AllOk &= Test4(Verbose);
    AllOk &= Test5(Verbose);
    AllOk &= Test6(Verbose);
    AllOk &= Test7(Verbose);
    AllOk &= Test8(Verbose);
#endif

    for (i = 0; i < max-1; i++) {
        PowerOff();
        MilliSleep(1000);
        PowerOn();
        MilliSleep(1000);

        AllOk &= Test(FALSE);
        AllOk &= Test1(FALSE);
        AllOk &= Test2(FALSE);
        AllOk &= Test3(FALSE);
        AllOk &= Test4(FALSE);
        AllOk &= Test5(FALSE);
        AllOk &= Test6(FALSE);
        AllOk &= Test7(FALSE);
        AllOk &= Test8(FALSE);
    }

    if (AllOk)
        _tprintf(_TEXT("Hurrah! all tests passed!!\n"));

    return 0;
}
