/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <mmlite.h>
#include <mmmacros.h>
#include <base/int64.h>

#include "../src/games/sntp/sntp.h"

/* Time format conversion tests */

extern UINT64 TranslateAbsoluteMMLiteTimeToSntp(TIME mmlTime);
extern TIME TranslateSntpTimeToAbsoluteMMLite(UINT64 sntpTime);

typedef struct _CONVERSION {
    UINT64 sntp; 
    UINT64 mml;
} CONVERSION;

/* 
    Use w32time/? to find out about translating Windows/SNTP time formats
    to human-readable text

    Helpful web script:
    http://www.spiders.com/portfolio/demos/countdown/index.jsp
 */
CONVERSION Convert[] = {
      /* SNTP format */                         /* MMLite (Windows) format */
/* January 1 1900 00:00:00 */
    { Int64Initializer(0x0, 0x0),               Int64Initializer(0x014F373B, 0xFDE04000) }, 
/* January 20 1968 03:14:08 */
    { Int64Initializer(0x80000000, 0x0),        Int64Initializer(0x019b827b, 0xfde04000) },
/* January 1 2005 00:00:00 */
    { Int64Initializer(0xC5806680, 0x0),        Int64Initializer(0x01c4ef94, 0xd6228000) }
};

#define NUM_CONVERSIONS sizeof(Convert)/sizeof(CONVERSION)

 /* 10^7 decimal, or 1 second */
#define MML_THRESHOLD 0x989680
    
int TestSntpToMml(UINT64 sntp, UINT64 correct)
{
    int failed = 0;
    TIME translated;
    UINT64 result, distance, threshold;
    
    translated = TranslateSntpTimeToAbsoluteMMLite(sntp);
    result = *(UINT64*) &translated;
    Int32ToInt64(threshold, MML_THRESHOLD);

    DBG(("Translated SNTP=0x%llx to MML=0x%llx, correct=0x%llx", 
        sntp, translated, correct));

    /* Make sure translated >= correct */
    if (Uint64Less(result, correct)) {
        UINT64 tmp = result;
        result = correct;
        correct = tmp;
    }
    /* Check that translated - correct < 1 second */
    distance = Uint64Subtract(result, correct);
    DBG((", error=0x%llx", distance));
    if (Uint64Less(distance, threshold)) {
        DBG(("... SUCCESS\n")); 
    } else { 
        DBG(("... FAILED\n")); 
        ++failed; 
    } 
    return -failed;
}

/* < 1 second */
#define SNTP_THRESHOLD 0xffffffff

int TestMmlToSntp(UINT64 mml, UINT64 correct) 
{
    int failed = 0;
    UINT64 translated, threshold, distance;
    
    translated = TranslateAbsoluteMMLiteTimeToSntp(*(PTIME) &mml);
    Int32ToInt64(threshold, SNTP_THRESHOLD); 

    DBG(("Translated MML=0x%llx to SNTP=0x%llx, correct=0x%llx", 
        mml, translated, correct));
    if (Uint64Less(translated, correct)) {
        UINT64 tmp = translated;
        translated = correct;
        correct = tmp;
    }
    distance = Uint64Subtract(translated, correct);
    DBG((", error=0x%llx", distance));
    if (Uint64Less(distance, threshold)) {
        DBG(("... SUCCESS\n")); 
    } else { 
        DBG(("... FAILED\n")); 
        ++failed; 
    } 
    return -failed;
}

int TestConversion(void)
{
    int failed = 0, i;

    DBG(("NUM_CONVERSIONS=%d\n", NUM_CONVERSIONS));

    for (i = 0; i < NUM_CONVERSIONS; ++i) {
        failed += TestMmlToSntp(Convert[i].mml, Convert[i].sntp);
    }

    for (i = 0; i < NUM_CONVERSIONS; ++i) {
        failed += TestSntpToMml(Convert[i].sntp, Convert[i].mml);
    }

    return -failed;
}
    
/* Component construction */
    
extern PIUNKNOWN CreateComponent(void);

#define PrintStructure(__pcli__) \
    DBG(("SNTP client structure at 0x%x:\n", (PTR) __pcli__)); \
    for (i = 0; i < sizeof(SntpClient)/sizeof(UINT32); ++i) \
        DBG(("\t0x%x\n", ((UINT32*) __pcli__)[i])); \
    
int TestCreateSntpClientComponent(void) 
{
    PIUNKNOWN punk = CreateComponent();
    SntpClient *pcli = (SntpClient *) punk;
    int i, failed = 0;

    UnusedParameter(pcli);

    DBG(("tsntp.c:TestCreateSntpClientComponent\n"));

    PrintStructure(pcli);

    assert(INVALID_SOCKET == pcli->Socket);
    assert(SNTP_CLIENT_MIN_PERIOD == pcli->Period);

    DBG(("client->ServerName=%s at 0x%x\n", 
        pcli->ServerName, (PTR) pcli->ServerName));

    return failed;
}
    
/* Networking tests */

extern SCODE MCT CreateSntpRequest(PISNTPCLIENT icli, UINT32 *Buffer, UINT Len) ;

int TestCreateSntpRequest(void)
{
    METHOD_DECLARE_NOTHIS(tsntp.c, TestCreateSntpRequest);
    SntpClient *pcli = (SntpClient *) CreateComponent();
    PISNTPCLIENT icli = (PISNTPCLIENT) pcli;
    int failed = 0, i;
    UINT32 buffer[SNTP_REQUEST_WORDS];
    UINT64 transmit, compare, distance, threshold, zero;
    
    Int32ToInt64(zero, 0x0);
    Int32ToInt64(threshold, SNTP_THRESHOLD);

    UnusedParameter(zero);

    DBG(("tsntp.c:TestCreateSntpRequest\n"));

    /* Create and pack request buffer */
    sc = CreateSntpRequest(icli, buffer, SNTP_REQUEST_BYTES);
    CHECKSC("failed to create buffer");

    /* Print field values word-by-word */
    DBG(("SNTP request buffer:\n"));
    for (i = 0; i < SNTP_REQUEST_WORDS; ++i)
        DBG(("\t0x%x\n", buffer[i]));

    /* Check transmit timestamp */
    compare = TranslateAbsoluteMMLiteTimeToSntp(CurrentTime());
    transmit = *(UINT64 *) (buffer + 10);
    distance = Uint64Subtract(compare, transmit);
    if (Uint64Less(distance, threshold)) {
        DBG(("bad transmit timestamp, tt=0x%llx now=0x%llx\n", 
             transmit, compare));
        --failed;
        goto Out;
    }

    /* Check other timestamps */
    assert(Int64Equal(zero, *(UINT64 *) (buffer + 4)));
    assert(Int64Equal(zero, *(UINT64 *) (buffer + 6)));
    assert(Int64Equal(zero, *(UINT64 *) (buffer + 8)));

METHOD_CLEANUP;
    if (FAILED(sc) && ! failed)
        --failed;
    return failed;
}
    
extern SCODE MCT SntpClientSendRequest(PISNTPCLIENT iThis, const _TCHAR *Server);
extern SCODE MCT SntpClientReceiveReply(
    PISNTPCLIENT iThis, UINT64 *ot, UINT64 *rt, UINT64 *tt, UINT64 *dt, PTIME stmp);

int TestSntpRequestReply(void)
{
    METHOD_DECLARE_NOTHIS(tsntp.c, TestSntpRequestReply);
    SntpClient *pcli = (SntpClient *) CreateComponent();
    PISNTPCLIENT icli = (PISNTPCLIENT) pcli;
    int failed = 0;
    UINT64 ot, rt, tt, dt;
    TIME stamp; 
    
    DBG(("tsntp.c:TestSntpRequestReply\n"));

    DBG(("Parameter checks...\n"));

    sc = SntpClientSendRequest(NULL, DEFAULT_TIME_SERVER);
    if (E_INVALID_PARAMETER != sc) {
        DBG(("Failed test: No check for null client\n"));
        --failed;
    }

    sc = SntpClientSendRequest(icli, NULL);
    if (E_INVALID_PARAMETER != sc) {
        DBG(("Failed test: No check for null server name\n"));
        --failed;
    }

    DBG(("Sending request...\n"));

    sc = SntpClientSendRequest(icli, DEFAULT_TIME_SERVER);
    CHECKSC("failed to send");

    assert(pcli->SockAddr.sin_family = AF_INET);
    assert(pcli->SockAddr.sin_port = htons(123));

    DBG(("Receiving reply...\n"));

    sc = SntpClientReceiveReply(icli, &ot, &rt, &tt, &dt, &stamp);
    CHECKSC("failed to receive");

    if (!failed)
        DBG(("tsntp.c:TestSntpRequestReply: success\n"));

METHOD_CLEANUP;
    return failed;
}
    
/* Known values from one of the early test runs against the MS server... 
   
   (RT-OT) = -0x88DB61A3F
   (TT-DT) = -0x897ABC530

   Sum of those two: -0x112561DF6F
   Divided by two to get the resulting offset: -0x892B0EFB7
   The sign indicates the client is ahead of the server.
   In two's complement: 0xfffffff76d4f1049
*/
#if 0
#define TEST_OT     Int64Initializer(0xc544decf, 0x363399d7)
#define TEST_RT     Int64Initializer(0xc544dec6, 0xa87d7f98)
#define TEST_TT     Int64Initializer(0xc544dec6, 0xa898cfa1)
#define TEST_DT     Int64Initializer(0xc544decf, 0x404494d1)
#define CORRECT     Int64Initializer(0xfffffff7, 0x6d4f1049)
#endif

extern SCODE MCT ComputeSntpFormatOffset(
    UINT64 ot, UINT64 rt, UINT64 tt, UINT64 dt, 
    PUINT64 result, PBOOL negative);

int TestComputeSntpFormatOffset(void)
{
    METHOD_DECLARE_NOTHIS(tsntp.c, TestComputeSntpFormatOffset);
    int failed = 0;
    UINT64 testOt, testRt, testTt, testDt;
    UINT64 offset, correct, shiftedOffset, shiftedCorrect;
    INT64 signedCorrect;
    BOOL negative, corneg;

    /* Test 1... */

    /* Used to be TEST_OT etc... then ARM/gcc... */
    Int64FromHighAndLow(testOt, 0xc544decf, 0x363399d7);
    Int64FromHighAndLow(testRt, 0xc544dec6, 0xa87d7f98);
    Int64FromHighAndLow(testTt, 0xc544dec6, 0xa898cfa1);
    Int64FromHighAndLow(testDt, 0xc544decf, 0x404494d1);
    Int64FromHighAndLow(signedCorrect, 0xfffffff7, 0x6d4f1049);
    signedCorrect = Int64TimesInt32(signedCorrect, -1);
    correct = *(UINT64*) &signedCorrect;
    corneg = TRUE;

    sc = ComputeSntpFormatOffset(testOt, testRt, testTt, testDt, &offset, &negative);
    CHECKSC("compute offset");

    /* Don't care about the lowest bits */
    shiftedOffset = Uint64RShift(offset, 3);
    shiftedCorrect = Uint64RShift(correct, 3);

    if (! Int64Equal(shiftedOffset, shiftedCorrect) || negative != corneg) {
        DBG(("Failed to compute offset, result=0x%llx neg=%d, correct=0x%llx neg=%d\n", 
            offset, negative, correct, corneg));
        --failed;
    }

    /* Test 2... */

    Int64FromHighAndLow(testOt, 0x000000f0, 0x322ca2f2); // 0x 000000f0322ca2f2
    Int64FromHighAndLow(testRt, 0xc5477ae2, 0x81fdbdb7); // 0x c5477ae281fdbdb7
    Int64FromHighAndLow(testTt, 0xc5477ae2, 0x821921e8); // 0x c5477ae2821921e8
    Int64FromHighAndLow(testDt, 0x000000f1, 0xaf0464e7); // 0x 000000f1af0464e7
    /* Term1: C54779F24FD11AC5
       Term2: C54779F0D314BD01
       Term1 / 2: 62A3BCF927E88D62 
       Term2 / 2: 62A3BCF8698A5E80
       (Term1 + Term2) / 2: C54779F19172EBE2 */
    Int64FromHighAndLow(correct, 0xC54779F1, 0x9172EBE2);
    corneg = FALSE;

    sc = ComputeSntpFormatOffset(testOt, testRt, testTt, testDt, &offset, &negative);
    CHECKSC("compute offset");

    /* Don't care about the lowest bits */
    shiftedOffset = Uint64RShift(offset, 3);
    shiftedCorrect = Uint64RShift(correct, 3);

    if (! Int64Equal(shiftedOffset, shiftedCorrect) || negative != corneg) {
        DBG(("Failed to compute offset, result=0x%llx neg=%d, correct=0x%llx neg=%d\n", 
            offset, negative, correct, corneg));
        --failed;
    }

    /* Should add more tests for the "65 bit arithmetic" in ComputeSntpFormatOffset */
    
METHOD_CLEANUP;
    if (!failed)
        DBG(("tsntp.c:TestComputeSntpFormatOffset: success\n"));
    return -failed;
}
    
int TestStartStop(void) 
{
    METHOD_DECLARE_NOTHIS(tsntp.c, TestSntpRequestReply);
    SntpClient *pcli = (SntpClient *) CreateComponent();
    PISNTPCLIENT icli = (PISNTPCLIENT) pcli;
    int i;
    TIME time;

    sc = icli->v->Start(icli, NULL, 60);
    CHECKSC("start SNTP client");

    /* Let it do > 2 periods */
    for (i = 0; i < 12 + (INT) (pcli->Period << 1); i += 10) {
        TIME later;
        Int32ToInt64(later, TIME_RELATIVE(TIME_SECONDS(10)));
        SleepUntil(later);
        sc = icli->v->CurrentTime(icli, &time);
        DBG(("Current time: \n\t  sntp=0x%llx \n\tsystem=0x%llx\n", 
            time, CurrentTime()));
    }

    /* Enough... */
    sc = icli->v->Stop(icli);
    CHECKSC("stop SNTP client");

METHOD_CLEANUP;
    return 0;
}
    
int _tmain(INT argc, _TCHAR *argv[])
{
    int failed = 0;

    {
        int i;
        SCODE sc = ParseCmdLine(&argc, &argv);
        if (FAILED(sc)) {
            printf("tsntp.c:_tmain: failed to parse command line\n");
            return -1;
        }

        printf("tsntp.c:_tmain: argc=%d", argc);
        for (i = 0; i < argc; ++i)
            printf(" argv[%d]=%s", i, argv[i]);
        printf("\n");
    }

    if (2 > argc) {

        failed += TestConversion();
        failed += TestCreateSntpClientComponent();
        failed += TestCreateSntpRequest();
        failed += TestSntpRequestReply();
        failed += TestComputeSntpFormatOffset();

    } else if (0 == _tcscmp(_T("start"), argv[1])) {

        TestStartStop();
    
    } else printf("tsntp.c:_tmain: bad input\n");

    if (failed)
        DBG(("Failed %d tests\n", failed));

    FreeCmdLine(&argc, &argv);
    return 0;
}
