/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>

#if 0
#define VERBOSE(x) _tprintf x
#else
#define VERBOSE(x)
#endif

UINT32 Random(UINT32 LastBlock)
{
    static UINT32 next = 1;

    next = next * 1103515245 + 12345;

    if (next <= LastBlock) return next;

    return (UINT32)(next % LastBlock);
    
}

void DumpSector(BYTE *SectorData, UINT32 SectorSize, _TCHAR *BannerMessage)
{
    UINT32 i, j;
    _tprintf(_TEXT("Dump of %d bytes of sector data @%x. %s\n\r"),
             SectorSize, (ADDRESS)SectorData, BannerMessage);

    for (i = 0; i < SectorSize; ) {
        _tprintf(_TEXT("[%4d]"), i);
        for (j = 1; j < 32; j++) {
            _tprintf(_TEXT(" %02x"), SectorData[i]);
            if (++i >= SectorSize) break;
        }
        _tprintf(_TEXT("\n\r"));
    }
}

SCODE ReadWriteTest(PIFILE Disk, UINT SectorSize, UINT MaxLoops)
{
  SCODE sc = E_NOT_ENOUGH_MEMORY;
  UINT i = 0, nRead, nWritten;
  UINT32 SectorNumber, DiskSize;
  UINT64 Offset64;
  BYTE *SectorPattern, *SectorContent, *SectorCheck, *SectorZero;
  SCODE ErrorCode = S_OK;

  /* Get buffers
   */
  SectorPattern = SectorContent = SectorCheck = SectorZero = NULL;
  SectorPattern = malloc(SectorSize);
  SectorContent = malloc(SectorSize);
  SectorCheck = malloc(SectorSize);
  SectorZero = malloc(SectorSize);
  if ((SectorPattern == NULL) ||
      (SectorContent == NULL) ||
      (SectorCheck == NULL) ||
      (SectorZero == NULL)) {
      _tprintf(_TEXT("FAILED to allocate 4 buffers of %d bytes....\n\r"), SectorSize);
      goto Out;
  }

  /* How big is the disk ?
   */
  sc = Disk->v->GetSize(Disk,&Offset64);
  if (FAILED(sc)) {
      _tprintf(_TEXT("FAILED to GetSize on disk (sc=x%x).\n\r"), sc);
      goto Out;
  }

  Offset64 = Uint64DividedByUint32(Offset64,SectorSize);
  /* Hopefully that fits in 32bits */
  DiskSize = Uint64ToUint32(Offset64);
  _tprintf(_TEXT("Will test across %d sectors.\n\r"), DiskSize);

  /* Initialize our write pattern
   */
  for (i = 0; i < SectorSize; i++)
      SectorPattern[i] = (BYTE) Random(0xff);

  /* Pick a random sector, read, write, check, restore.
   * Repeat.
   */
  for (i = 0; (i < MaxLoops) && (ErrorCode == S_OK); i++) {

      /* Show progress
       */
      if (0 == (i % 1000)) _tprintf(_TEXT("."));

      /* Pick a sector, any sector
       */
      SectorNumber = Random(DiskSize);
      VERBOSE((_TEXT("%d "), SectorNumber));
      Uint32ToUint64(Offset64,SectorNumber);
      Offset64 = Uint64TimesUint32(Offset64,SectorSize);

      /* Read current content
       */
      VERBOSE((_TEXT("Reading sector %d\n"), SectorNumber));
      nRead = 0;
      sc = Disk->v->ReadAt(Disk, Offset64, SectorContent, SectorSize, &nRead);
      if (FAILED(sc) || (nRead != SectorSize)) {
          _tprintf(_TEXT("FAILED to read sector %d (sc=x%x nR=%d).\n\r"),
                   SectorNumber, sc, nRead);
          sc = (FAILED(sc)) ? sc : E_DISK_OPERATION_FAILED;
          goto Out;
      }


      /* Write test pattern
       */
      VERBOSE((_TEXT("Writing to sector %d\n"), SectorNumber));
      nWritten = 0;
      sc = Disk->v->WriteAt(Disk, Offset64, SectorPattern, SectorSize, &nWritten);
      if (FAILED(sc) || (nWritten != SectorSize)) {
          _tprintf(_TEXT("FAILED to write test pattern at sector %d (sc=x%x nW=%d).\n\r"),
                   SectorNumber, sc, nWritten);
          sc = (FAILED(sc)) ? sc : E_DISK_OPERATION_FAILED;
          goto Out;
      }

      /* Read sector zero to create interference, e.g. flush any one-sector cache.
       */
      VERBOSE((_TEXT("Re-reading sector 0\n")));
      nRead = 0;
      sc = Disk->v->ReadAt(Disk, UINT64_ZERO, SectorZero, SectorSize, &nRead);
      if (FAILED(sc) || (nRead != SectorSize)) {
          _tprintf(_TEXT("FAILED to re-read sector 0 (sc=x%x nR=%d).\n\r"),
                   sc, nRead);
          sc = (FAILED(sc)) ? sc : E_DISK_OPERATION_FAILED;

          ErrorCode = sc;
          goto RestoreIt;
      }

      /* Read it back
       */
      VERBOSE((_TEXT("Checking sector %d\n"), SectorNumber));
      nRead = 0;
      sc = Disk->v->ReadAt(Disk, Offset64, SectorCheck, SectorSize, &nRead);
      if (FAILED(sc) || (nRead != SectorSize)) {
          _tprintf(_TEXT("FAILED to read-back sector %d (sc=x%x nR=%d).\n\r"),
                   SectorNumber, sc, nRead);
          sc = (FAILED(sc)) ? sc : E_DISK_OPERATION_FAILED;

          ErrorCode = sc;
          goto RestoreIt;
      }

      /* Check
       */
      if (memcmp(SectorCheck,SectorPattern,SectorSize) != 0) {
          _tprintf(_TEXT("Read-back sector %d does not compare with test pattern.\n\r"),
                   SectorNumber);
          DumpSector(SectorPattern, SectorSize, _TEXT("Content should have been:"));
          DumpSector(SectorCheck, SectorSize, _TEXT("Actual content is:"));

          ErrorCode = E_PARSE_ERROR;
          goto RestoreIt;
      }

      /* Write back original content
       */
  RestoreIt:
      VERBOSE((_TEXT("Restoring sector %d\n"), SectorNumber));
      nWritten = 0;
      sc = Disk->v->WriteAt(Disk, Offset64, SectorContent, SectorSize, &nWritten);
      if (FAILED(sc) || (nWritten != SectorSize)) {
          _tprintf(_TEXT("FAILED to restore sector %d (sc=x%x nW=%d).  DISK CORRUPTED!!!\n\r"),
                   SectorNumber, sc, nWritten);
          sc = (FAILED(sc)) ? sc : E_DISK_OPERATION_FAILED;
          goto Out;
      }
  }

 Out:
  if (FAILED(ErrorCode) && !FAILED(sc))
      sc = ErrorCode;

  if (SectorPattern) free(SectorPattern);
  if (SectorContent) free(SectorContent);
  if (SectorCheck) free(SectorCheck);
  if (SectorZero) free(SectorZero);

  if (FAILED(sc))
      _tprintf(_TEXT("Terminating test after %d of %d iterations.\n\r"), i, MaxLoops);
  
  return sc;
}

SCODE DiskTest(_TCHAR *DiskName, UINT SectorSize, UINT nLoops)
{
  PINAMESPACE pNs = CurrentNameSpace();
  PIFILE pD = NULL;
  PIUNKNOWN pUnk;
  SCODE sc;

  sc = pNs->v->Bind(pNs, DiskName, NAME_SPACE_WRITE | NAME_SPACE_READ, &pUnk);
  if (FAILED(sc)) {
    _tprintf(_TEXT("FAILED to bind to %s.\n\r"), DiskName);
    goto Out;
  }
  
  sc = pUnk->v->QueryInterface(pUnk, &IID_IFile, (void **)&pD);
  pUnk->v->Release(pUnk);
  if (FAILED(sc)) {
    _tprintf(_TEXT("%s not a IFile.\n\r"), DiskName);
    goto Out;
  }
  
  /* Now test it */
  sc = ReadWriteTest(pD, SectorSize, nLoops);
  if (FAILED(sc))
    _tprintf(_TEXT("ReadWriteTest failed with sc=x%x.\n\r"), sc);

  pD->v->Release(pD);
 Out:
  return sc;
}

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    int r;
    _TCHAR *DiskName = _TEXT("disk");
    UINT SectorSize = 512; /* one FAT sector worth */
    UINT nLoops = 10000;
    
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_TEXT("No command line args (%x), test FAILED\n\r"), sc);
        return (EXIT_FAILURE);
    }

    if (argc > 1) {
        DiskName = argv[1];
    }
    _tprintf(_TEXT("Will run this test on disk '%s'\n\r"), DiskName);

    if (argc > 2) {
      SectorSize = _ttoi(argv[2]);
    }
    _tprintf(_TEXT("Will use a sector size of %d bytes.\n\r"), SectorSize);

    if (argc > 3) {
        nLoops = _ttoi(argv[3]);
    }
    _tprintf(_TEXT("Will do %d random read/write operations.\n\r"), nLoops);

    sc = DiskTest(DiskName,SectorSize,nLoops);

    if (FAILED(sc)) {
      _tprintf(_TEXT("Test FAILED. Usage: %s [disk_name [sector_size [number_of_loops]]]\n\r"),
               argv[0]);
      r = (EXIT_FAILURE);
    } else {
        _tprintf(_TEXT("Test PASSED ok.\n\r"));
        r = (EXIT_SUCCESS);
    }

    (void) FreeCmdLine(&argc,&argv);

    return r;
}
