/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* Tracks the mouse, stops when all 3 buttons are pressed at once.
 */
#include <stdio.h>
#include <mmlite.h>
#include <stdlib.h>
#include <tchar.h>
#include <framebuffer.h>

UINT8 *FB = NULL;
UINT32 fW, fWv, fHv;

SCODE InitializeFrameBuffer(_TCHAR* FbName)
{
    PIFRAMEBUFFER FrameBuffer;
    SCODE sc;
    UINT32 Length;
    _TCHAR *Msg;

    Msg = _T("bind to");
    sc = BindToObject(CurrentNameSpace(),
                      FbName,
                      NAME_SPACE_READ | NAME_SPACE_WRITE,
                      &IID_IFrameBuffer,
                      (void *) &FrameBuffer);
    if (FAILED(sc)) goto Bad;

    FB = NULL;
    Length = 0;
    Msg = _T("map");
    sc = FrameBuffer->v->Map(FrameBuffer,(PTR*)&FB,&Length);
    if (FAILED(sc)) goto Bad;

    Msg = _T("get Hlen");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_HORIZONTAL_LENGTH,&fW);
    if (FAILED(sc)) goto Bad;

    Msg = _T("get Hvis");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_HORIZONTAL_VISIBLE_LENGTH,&fWv);
    if (FAILED(sc)) goto Bad;

    Msg = _T("get Vvis");
    sc = FrameBuffer->v->GetProperty(FrameBuffer,FB_PROP_VERTICAL_VISIBLE_LENGTH,&fHv);
    if (FAILED(sc)) goto Bad;

    /* BUGBUG strictly speaking we cant do this
     */
    FrameBuffer->v->Release(FrameBuffer);

    return S_OK;
 Bad:
    _tprintf(_T("Could not %s %s sc=%x\n"), Msg, FbName, sc);
    return sc;
}

#define cW 8
#define cH 12
#define O 255
#define B 0

UINT8 PointerBitmap[cW*cH] = {
    B, O, O, O, O, O, O, O,
    B, B, O, O, O, O, O, O,
    B, B, B, O, O, O, O, O,
    B, B, B, B, O, O, O, O,
    B, B, B, B, B, O, O, O,
    B, B, B, B, B, B, O, O,
    B, B, B, B, O, O, O, O,
    B, O, B, B, O, O, O, O,
    O, O, O, B, O, O, O, O,
    O, O, O, B, B, O, O, O,
    O, O, O, B, B, O, O, O,
    O, O, O, O, O, O, O, O
};

/* Obvious way
 */
UINT8 Bg[cW*cH];

void getbg0(int x0, int y0)
{
    int x,y;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++)
            Bg[x + (y * cW)] = FB[(x + x0) + ((y + y0) * fW)];;
}

void setbg0(int x0, int y0)
{
    int x,y;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++)
            FB[(x + x0) + ((y + y0) * fW)] = Bg[x + (y * cW)];
}

void paintc0(int x0, int y0)
{
    int x,y;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++)
            FB[(x + x0) + ((y + y0) * fW)] = PointerBitmap[x + (y * cW)];
}

/* Using XOR
 */
void getbg1(int x0, int y0) {return;}
#define setbg1 paintc1

void paintc1(int x0, int y0)
{
    int x,y;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++)
            FB[(x + x0) + ((y + y0) * fW)] ^= ~PointerBitmap[x + (y * cW)];
}

/* Non-transparent overlay
 */
#define getbg2 getbg0
#define setbg2 setbg0

void paintc2(int x0, int y0)
{
    int x,y;
    UINT8 c;

    for (x = 0; x < cW; x++)
        for (y = 0; y < cH; y++) {
            c = PointerBitmap[x + (y * cW)];
            if (c == B)
                FB[(x + x0) + ((y + y0) * fW)] = c;
        }
}



/* Select which is which
 */
typedef struct {
    void (*GetBackground)(int x, int y);
    void (*SetBackground)(int x, int y);
    void (*PaintPointer)(int x, int y);
} PAINTER;

#define nPAINTERS 3
PAINTER Pointers[nPAINTERS] = {
    { getbg0, setbg0, paintc0 },
    { getbg1, setbg1, paintc1 },
    { getbg2, setbg2, paintc2 },
};

PAINTER *Pointer = &Pointers[2];

typedef struct _REPORT {
    UINT8 Buttons;
    INT8 x;
    INT8 y;
    INT8 Wheel;
} REPORT, *PREPORT;

#if _DEBUG
_TCHAR *Positive(int i)
{
    return (i > 0) ? _T("+") : _T("");
}
#endif

void THREAD_LINKAGE MouseThread(THREAD_ARGUMENT arg)
{
    PIFILE Mouse = (PIFILE )arg;
    REPORT R;
    SCODE sc;
    UINT nBytes;
    int x, x1, y, y1;

    x = fWv / 2;
    y = fHv / 2;
    Pointer->GetBackground(x,y);
    Pointer->PaintPointer(x,y);

    for (;;) {

        /* Wait for report
         */
        nBytes = 0;
        sc = Mouse->v->ReadAt(Mouse,
                              UINT64_ZERO,
                              (BYTE*)&R,
                              sizeof R,
                              &nBytes);
        if (FAILED(sc) || (nBytes != sizeof R))
            break;

        /* 3-button-press is termination
         */
        if (R.Buttons == 7)
            break;

        /* Tentatively update pointer position
         */
        x1 = x + R.x;
        y1 = y + R.y;

        /* Box it
         */
        if ((x1 < 0) || (x1 > (int)fWv))
            x1 = x;
        if ((y1 < 0) || (y1 > (int)fHv))
            y1 = y;

        /* Valid move?
         */
        if ((x != x1) || (y != y1)) {
            /* Yes, restore previous background
             */
            Pointer->SetBackground(x,y);
            /* Preserve new bg
             */
            Pointer->GetBackground(x1,y1);
            /* Set pointer at new position
             */
            Pointer->PaintPointer(x1,y1);
            /* Move
             */
            x = x1;
            y = y1;
        }
#if _DEBUG
        else 
            _tprintf(_T("buttons %x position (%s%d,%s%d) wheel %s%d      \r"),
                     R.Buttons,
                     Positive(R.x), R.x,
                     Positive(R.y), R.y,
                     Positive(R.Wheel), R.Wheel);
#endif
    }
    Pointer->SetBackground(x,y);

    _tprintf(_T("MouseThread done.\n"));
}

int _tmain(INT argc, _TCHAR **argv)
{
    SCODE sc;
    INT RetVal = EXIT_FAILURE;
    _TCHAR *MouseName = _T("mouse0");
    _TCHAR *FbName = _T("fb0");
    PIFILE Mouse = NULL;
    int c;
    
    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_T("No command line args (%x), test FAILED\n"), sc);
        exit(RetVal);
    }

    if (argc > 1) {
        MouseName = argv[1];
    }

    if (argc > 2) {
        c = _ttoi(argv[2]);
        if ((c >= 0) && (c < nPAINTERS))
            Pointer = &Pointers[c];
    }

    if (argc > 3) {
        FbName = argv[3];
    }

    sc = BindToObject(CurrentNameSpace(), MouseName,
                      NAME_SPACE_READ,
                      &IID_IFile,
                      (void *)&Mouse);
    if (FAILED(sc)) {
        _tprintf(_T("Could not open mouse file %s\n"), MouseName);
        goto Done;
    }

    sc = InitializeFrameBuffer(FbName);
    if (FAILED(sc)) {
        _tprintf(_T("Could not init framebuffer %s\n"), FbName);
        goto Done;
    }

    _tprintf(_T("Tracking %s. Press all 3 buttons to terminate\n"),
             MouseName);

    MouseThread((THREAD_ARGUMENT)Mouse);

    RetVal = EXIT_SUCCESS;
Done:
    if (Mouse)
        Mouse->v->Release(Mouse);

    (void) FreeCmdLine(&argc,&argv);
    exit(RetVal);
    return 0;
}
