/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbctl.c
 *
 * Purpose:
 *
 *    Utility to prform control operations on a hub
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tchar.h>
#define _DEBUG 1
#define DBGLVL 0
#include <diagnostics.h>
#include <drivers\usb.h>
#include <usbmaster.h>

/* BUGBUG This needs to be added to :GetUsbAddress()
 */
#define MAXCTLPKTSIZE 64

typedef struct {
    _TCHAR *HubName;
    _TCHAR *DeviceName;
    _TCHAR *Operation;
    UINT8   UsbAddress;
    UINT8   UpstreamPort;
    UINT8   Mps;
    UINT8   PortNumber;
    UINT16  Feature;
    UINT8   fSetIt;
} DEVARGS, *PDEVARGS;

#define MAX_PATH 512
UINT8 Buffer[MAX_PATH];


/* Utilities
 */
#if _DEBUG
PRIVATE void PrintUrbResults(PUSBM_REQUEST_BLOCK pUrb, _TCHAR *Banner)
{
    int i;

    _tprintf(Banner);
    for (i = 0; i < pUrb->ActualDataLength; i++)
        _tprintf(_T("%2x "), pUrb->Data[i]);
    _tprintf(_T("\n"));
}

PRIVATE void PrintConfigurationInformation(PDEVARGS Args)
{
    _tprintf(_T("\nWill do a '%s' on HUB '%s' (connected to HUB '%s')\n"),
             Args->Operation,Args->DeviceName,Args->HubName);

    _tprintf(_T("\tUsbAddress %d\n"), Args->UsbAddress);
    _tprintf(_T("\tUpstreamPort %d\n"), Args->UpstreamPort);
    _tprintf(_T("\tMps %d\n"), Args->Mps);
    _tprintf(_T("\tPortNumber %d\n"), Args->PortNumber);
    _tprintf(_T("\tFeature %d\n"), Args->Feature);
    _tprintf(_T("\tfSetIt %d\n"), Args->fSetIt);
}
#endif


/* Change a port's feature, either setting it or clearing it
 */
PRIVATE SCODE SetOrClearPortFeature(
    PIUSBHUB Hub,
    PDEVARGS Args)
{
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
    CONDITION Done;


    DBGME(0,_tprintf(_T("usbctl:%sFeature(%d,%x,%d,%d)\n"),
                     (Args->fSetIt) ? _T("Set"):_T("Clear"), 
                     Args->PortNumber, Args->Feature,
                     Args->UpstreamPort,Args->UsbAddress));

    Condition_Init(&Done);

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &Done;
    Urb.DeviceAddress = Args->UsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_CONTROL;
    Urb.SetupData[0] = USB_RT_CLASS | USB_RT_OTHER; //x23, 20 for hub
    Urb.SetupData[1] = (Args->fSetIt) ? USB_R_SET_FEATURE : USB_R_CLEAR_FEATURE;
    Set2(&Urb.SetupData[2],Args->Feature);
    Urb.SetupData[4] = Args->PortNumber;
    Urb.SetupLength = 8;
    Urb.MaximumPacketSize = Args->Mps;

    sc = Hub->v->SubmitUrb(Hub,
                           &Urb,
                           Args->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("usbctl:?etFe SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);
    sc = (Urb.Flags & URB_FLAG_ERROR) ? E_FAIL : S_OK;
    DBGME(0,_tprintf(_T("usbctl:?etFe(%d) ok. sc=x%x\n"),
                     Args->UsbAddress,sc));

    return sc;
}
#define SetPortFeature(h,a)   { a->PortNumber += 1; \
                                a->fSetIt = TRUE; \
                                PrintConfigurationInformation(a); \
                                SetOrClearPortFeature(h,a); }
#define ClearPortFeature(h,a) { a->PortNumber += 1; \
                                a->fSetIt = FALSE; \
                                PrintConfigurationInformation(a); \
                                SetOrClearPortFeature(h,a); }
#define SetHubFeature(h,a)    { a->PortNumber = 0; \
                                a->fSetIt = TRUE; \
                                PrintConfigurationInformation(a); \
                                SetOrClearPortFeature(h,a); }
#define ClearHubFeature(h,a)  { a->PortNumber = 0; \
                                a->fSetIt = FALSE; \
                                PrintConfigurationInformation(a); \
                                SetOrClearPortFeature(h,a); }

PRIVATE SCODE VerifyTheMps(PIUSBHUB Hub, PDEVARGS Args)
{
    PUSB_DEVICE_struct Descriptor;
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
    CONDITION Done;

    Condition_Init(&Done);

#define maxdesc 64
    Descriptor = (PUSB_DEVICE_struct) Buffer;

    DBGME(0,_tprintf(_T("VerifyTheMps(%d)\n"),Args->UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &Done;
    Urb.DeviceAddress = Args->UsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb.SetupData[0] = USB_RT_DTH;
    Urb.SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb.SetupData[3] = USB_DT_DEVICE;
    Urb.SetupData[6] = maxdesc;
    Urb.SetupLength = 8;
    Urb.ExpectedDataLength = maxdesc;
    Urb.Data = (UINT8 *)Descriptor;
    Urb.MaximumPacketSize = Args->Mps;

    sc = Hub->v->SubmitUrb(Hub,
                           &Urb,
                           Args->UpstreamPort);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("VerifyTheMps: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    sc = Condition_Wait(Urb.Done,NULL,NULL);

    if (FAILED(sc))
        return sc;

    if ((Urb.Flags & URB_FLAG_ERROR) ||
        (Urb.ActualDataLength < sizeof *Descriptor))
        return E_FAIL;

    PrintUrbResults(&Urb,_T("Device Descriptor:"));

    /* Adjust the control Mps if necessary */
    if (Args->Mps != Descriptor->MaxPacketSize0) {
        DBGME(2,_tprintf(_T("usbctl: Mps changed to %d (was %d)\n"),
                         Descriptor->MaxPacketSize0, Args->Mps));
        Args->Mps = Descriptor->MaxPacketSize0;
    }

    return S_OK;
#undef maxdesc
}

PRIVATE SCODE doDev(PDEVARGS Args)
{
    UINT UsbAddress32, Port32;
    PIUSBDRIVER Driver = NULL;
    SCODE sc;
    PIUSBHUB Hub = NULL;
    PINAMESPACE Ns = CurrentNameSpace();

    /* And it must be connected to a HUB..
     */
    sc = BindToObject(Ns,
                      Args->HubName,
                      NAME_SPACE_READ,
                      &IID_IUsbHub,
                      (void **)&Hub);
    if (FAILED(sc)) {
        _tprintf(_TEXT("%s is not an IUsbHub (%x)\n\r"), 
                 Args->HubName, sc);
        goto Out;
    }

    /* But it only needs to be a UsbDriver
     */
    sc = BindToObject(Ns,
                      Args->DeviceName,
                      NAME_SPACE_READ,
                      &IID_IUsbDriver,
                      (void **)&Driver);
    if (FAILED(sc)) {
        _tprintf(_T("%s: cant find or not a UsbDriver, sc=x%x\n"),
                 Args->DeviceName,sc);
        goto Out;
    }

    sc = Driver->v->GetUsbAddress(Driver,&UsbAddress32,&Port32);
    if (FAILED(sc)) {
        _tprintf(_T("%s has no UsbAddress, sc=x%x\n"),Args->DeviceName,sc);
        goto Out;
    }

    Args->UsbAddress   = (UINT8) UsbAddress32;
    Args->UpstreamPort = (UINT8) Port32;
    /* BUGBUG See note at top */
    Args->Mps          =  MAXCTLPKTSIZE;
    sc = VerifyTheMps(Hub,Args);

    /* Port ops
     */
    if (_tcsncmp(Args->Operation,_T("up"),2) == 0) {
        Args->Feature = USB_FS_PORT_POWER;
        SetPortFeature(Hub,Args);
    }

    if (_tcsncmp(Args->Operation,_T("down"),4) == 0) {
        Args->Feature = USB_FS_PORT_POWER;
        ClearPortFeature(Hub,Args);
    }

    /* HUB ops
     */
    if (_tcsncmp(Args->Operation,_T("hup"),2) == 0) {
        Args->Feature = USB_FS_HUB_LOCAL_POWER;
        SetHubFeature(Hub,Args);
    }

    if (_tcsncmp(Args->Operation,_T("hdown"),4) == 0) {
        Args->Feature = USB_FS_HUB_LOCAL_POWER;
        ClearHubFeature(Hub,Args);
    }


 Out:
    if (Hub)
        Hub->v->Release(Hub);
    if (Driver)
        Driver->v->Release(Driver);
    return sc;
}

int _tmain(int argc, _TCHAR **argv)
{
    SCODE sc;
    DEVARGS Args;

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_TEXT("No command line args (%x)\n\r"), sc);
        return -1;
    }

    /* There must be a few arguments: the HUB to work on, where its connected,
     * what to do, to which port
     */
    if (argc < 5) {
        _tprintf(_TEXT("Usage:%s UsbHubName ConnectionHubName up|down|hup|hdown port\n\r"), argv[0]);
        goto Out;
    }

    memset(&Args,0,sizeof Args);

    Args.DeviceName = argv[1];
    Args.HubName    = argv[2];
    Args.Operation  = argv[3];
    Args.PortNumber = _ttoi(argv[4]);

    sc = doDev(&Args);

 Out:
    (void) FreeCmdLine(&argc,&argv);
    return 0;
}
