/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbdir.c
 *
 * Purpose:
 *
 *    Utility to print the details of (a) USB device tree.
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <stdio.h>
#include <string.h>
#include <tchar.h>
//#define _DEBUG 1
//#define DBGLVL 0
#include <diagnostics.h>
#include <drivers\usb.h>
#include <usbmaster.h>

#include "vendors.h"

#define MAX_PATH 512
_TCHAR Buffer[MAX_PATH];
_TCHAR AttributesBuffer[128];

/* Strings are a pain, but we need them.
 *
 * According to specs, we should be able to ask for *any* string and
 * receive a STALL back if the string does not exist.
 * But of course that is not what happens, there are *some* that break
 * the specs, and we are talking about USB2.0 devices here.
 * Some HC will not react well to, say, returning six NAKs in a row
 * instead of STALL. So unles I can fix those HCs I shall avoid asking
 * for strings that are not mentioned in any descriptor.
 */
UINT32 StringIdsReferenced[256/32];
#define ClearAllStringReferences() { \
    int i; \
    for (i = 0; i < (256/32); i++) StringIdsReferenced[i] = 0; \
}
#define MarkStringReferenced(_n_) { \
    StringIdsReferenced[(_n_)/32] |= (1 << ((_n_) & 31)); \
}
#define StringWasReferenced(_n_) \
    ((StringIdsReferenced[(_n_)/32] & (1 << ((_n_) & 31))) != 0)

/* Utilities
 */
#if _DEBUG
PRIVATE void PrintUrbResults(PUSBM_REQUEST_BLOCK pUrb, _TCHAR *Banner)
{
    int i;

    _tprintf(Banner);
    for (i = 0; i < pUrb->ActualDataLength; i++)
        _tprintf(_T("%2x "), pUrb->Data[i]);
    _tprintf(_T("\n"));
}
#endif

PRIVATE _TCHAR *VendorName( UINT16 VendorId)
{
    int i;

    for (i = 0; i < nVendorIdTable; i++)
        if (VendorIdTable[i].VendorId == VendorId)
            return VendorIdTable[i].VendorName;
    return _T("");
}

PRIVATE _TCHAR *CountryName( UINT8 CountryCode)
{
    switch (CountryCode) {
    case USB_CC_Arabic:    return _T("Arabic");
    case USB_CC_Belgian:    return _T("Belgian");
    case USB_CC_Canadian_Bilingual:    return _T("Canadian_Bilingual");
    case USB_CC_Canadian_French:    return _T("Canadian_French");
    case USB_CC_CzechRepublic:    return _T("CzechRepublic");
    case USB_CC_Danish:    return _T("Danish");
    case USB_CC_Finnish:    return _T("Finnish");
    case USB_CC_French:    return _T("French");
    case USB_CC_German:    return _T("German");
    case USB_CC_Greek:    return _T("Greek");
    case USB_CC_Hebrew:    return _T("Hebrew");
    case USB_CC_Hungary:    return _T("Hungary");
    case USB_CC_ISO:    return _T("ISO");
    case USB_CC_Italian:    return _T("Italian");
    case USB_CC_Japan_Katakana:    return _T("Japan_Katakana");
    case USB_CC_Korean:    return _T("Korean");
    case USB_CC_LatinAmerican:    return _T("LatinAmerican");
    case USB_CC_Dutch:    return _T("Dutch");
    case USB_CC_Norwegian:    return _T("Norwegian");
    case USB_CC_Persian_Farsi:    return _T("Persian_Farsi");
    case USB_CC_Poland:    return _T("Poland");
    case USB_CC_Portuguese:    return _T("Portuguese");
    case USB_CC_Russia:    return _T("Russia");
    case USB_CC_Slovakia:    return _T("Slovakia");
    case USB_CC_Spanish:    return _T("Spanish");
    case USB_CC_Swedish:    return _T("Swedish");
    case USB_CC_Swiss_French:    return _T("Swiss_French");
    case USB_CC_Swiss_German:    return _T("Swiss_German");
    case USB_CC_Switzerland:    return _T("Switzerland");
    case USB_CC_Taiwan:    return _T("Taiwan");
    case USB_CC_Turkish_Q:    return _T("Turkish_Q");
    case USB_CC_UK:    return _T("UK");
    case USB_CC_US:    return _T("US");
    case USB_CC_Yugoslavia:    return _T("Yugoslavia");
    case USB_CC_Turkish_F:    return _T("Turkish_F");
    default:
        break;
    }
    return _T("");
}

PRIVATE _TCHAR *InterfaceClassName( UINT8 ClassCode)
{
    switch (ClassCode) {
    case USB_DC_MULTICLASS:
        return _T("Multiclass");
    case USB_DC_AUDIO:
        return _T("Audio");
    case USB_DC_CDC:
        return _T("CommunicationDevice");
    case USB_DC_HID:
        return _T("HumanInterfaceDevice");
    case USB_DC_PHYS:
        return _T("PhysicalDevice");
    case USB_DC_PRINTER:
        return _T("Printer");
    case USB_DC_STORAGE:
        return _T("Storage");
    case USB_DC_HUB:
        return _T("Hub");
    case USB_DC_CDC_DATA:
        return _T("CommunicationDevice data");
    case USB_DC_SMART_CARD:
        return _T("SmartCard");
    case USB_DC_SECURE_CONTENT:
        return _T("SecureContent");
    case USB_DC_VENDOR_SPECIFIC:
        return _T("Vendor-specific");
    default:
        return _T("");
    }
}

PRIVATE _TCHAR *InterfaceSubclassName( UINT8 ClassCode, UINT8 SubclassCode)
{
    switch (ClassCode) {
    case USB_DC_MULTICLASS:
        break;
    case USB_DC_AUDIO:
        switch (SubclassCode) {
        case USB_DS_AUDIO_CONTROL:
            return _T("AudioControl");
        case USB_DS_AUDIO_STREAMING:
            return _T("AudioStreaming");
        case USB_DS_MIDI_STREAMING:
            return _T("MidiStreaming");
        default:
            break;
        }
    case USB_DC_CDC:
        switch (SubclassCode) {
        case USB_DS_DIRECT_LINE:
            return _T("DirectLine");
        case USB_DS_ABSTRACT:
            return _T("Abstract");
        case USB_DS_TELEPHONE:
            return _T("Telephone");
        case USB_DS_MULTICHANNEL:
            return _T("MultiChannel");
        case USB_DS_CAPI:
            return _T("CAPI");
        case USB_DS_ETHERNET:
            return _T("Ethernet");
        case USB_DS_ATM:
            return _T("ATM");
        default:
            break;
        }
    case USB_DC_HID:
        if (SubclassCode == USB_DS_BOOT)
            return _T("Boot");
        break;
    case USB_DC_PHYS:
        break;
    case USB_DC_PRINTER:
#if 0
        /* Already done at Class level, this would be a repeat */
        if (SubclassCode == USB_DS_PRINTER)
            return _T("printer");
#endif
        break;    
    case USB_DC_STORAGE:
        switch (SubclassCode) {
        case USB_DS_RBC:
            return _T("RBC");
        case USB_DS_ATAPI:
            return _T("ATAPI");
        case USB_DS_QIC:
            return _T("QIC");
        case USB_DS_UFI:
            return _T("UFI");
        case USB_DS_SFF_8070i:
            return _T("SFF_8070i");
        case USB_DS_SCSI:
            return _T("SCSI");
        default:
            break;
        }
    case USB_DC_HUB:
    case USB_DC_CDC_DATA:
    case USB_DC_SMART_CARD:
    case USB_DC_SECURE_CONTENT:
    case USB_DC_VENDOR_SPECIFIC:
    default:
        break;
    }
    return _T("");
}

PRIVATE _TCHAR *InterfaceProtocolName(UINT8 ClassCode, UINT8 SubclassCode, UINT8 ProtocolCode)
{
    switch (ClassCode) {
    case USB_DC_MULTICLASS:
    case USB_DC_AUDIO:
        break;
    case USB_DC_CDC:
        if (ProtocolCode == USB_DP_V_25ter)
            return _T("V.25ter (AT,Hayes compatible)");
        break;
    case USB_DC_HID:
        if (SubclassCode == USB_DS_BOOT) {
            switch (ProtocolCode) {
            case USB_DP_KEYBOARD:
                return _T("Keyboard");
            case USB_DP_MOUSE:
                return _T("Mouse");
            default:
                break;
            }
        }
        break;
    case USB_DC_PHYS:
        break;
    case USB_DC_PRINTER:
        if (SubclassCode == USB_DS_PRINTER) {
            switch (ProtocolCode) {
            case USB_DP_UNIDIRECTIONAL:
                return _T("Unidirectional");
            case USB_DP_BIDIRECTIONAL:
                return _T("Bi-directional");
            case USB_DP_IEEE_1284_4:
                return _T("IEEE-1284.4 Bi-directional");
            default:
                break;
            }
        }
        break;
    case USB_DC_STORAGE:
        switch (ProtocolCode) {
        case USB_DP_CBI:
            return _T("CBI with command completion");
        case USB_DP_CBI_NO_CC:
            return _T("CBI without command completion");
        case USB_DS_BULK_ONLY:
            return _T("Bulk-Only");
        default:
            break;
        }
        break;
    case USB_DC_HUB:
        /* NB: This is a bit of a mess, really. */
        switch (ProtocolCode) {
        case USB_DP_SINGLE_TT:
            return _T("Single TT");
        case USB_DP_MULTIPLE_TT:
            return _T("Multiple TT");
        default:
            break;
        }
        break;
    case USB_DC_CDC_DATA:
        switch (ProtocolCode) {
        case USB_DP_ISDN:
            return _T("ISDN");
        case USB_DP_HDLC:
            return _T("HDLC");
        case USB_DP_TRANSPARENT:
            return _T("Transparent");
        case USB_DP_Q921M:
            return _T("Q.921 Management");
        case USB_DP_Q921:
            return _T("Q.921 Data");
        case USB_DP_Q921TM:
            return _T("Q.921 TEI-multiplexor");
        default:
            break;
        }
    case USB_DC_SMART_CARD:
    case USB_DC_SECURE_CONTENT:
    case USB_DC_VENDOR_SPECIFIC:
    default:
        break;
    }
    return _T("");
}

PRIVATE _TCHAR *AudioControlSubtypeName(UINT8 SubType)
{
    switch (SubType) {
    case USB_DST_HEADER: 
        return _T("Header");
    case USB_DST_INPUT_TERMINAL:
        return _T("Input Terminal");
    case USB_DST_OUTPUT_TERMINAL:
        return _T("Output Terminal");
    case USB_DST_MIXER_UNIT:
        return _T("Mixer Unit");
    case USB_DST_SELECTOR_UNIT:
        return _T("Selector Unit");
    case USB_DST_FEATURE_UNIT:
        return _T("Feature Unit");
    case USB_DST_PROCESSING_UNIT:
        return _T("Processing Unit");
    case USB_DST_EXTENSION_UNIT:
        return _T("Extension Unit");
    default:
        break;
    }
    return _T("");
}

PRIVATE _TCHAR *AudioStreamingSubtypeName(UINT8 SubType)
{
    switch (SubType) {
    case USB_DST_AS_GENERAL:
        return _T("General");
    case USB_DST_FORMAT_TYPE:
        return _T("Format Type");
    case USB_DST_FORMAT_SPECIFIC:
        return _T("Format Specific");
    default:
        break;
    }
    return _T("");
}


PRIVATE void SprintConfigurationAttributes(_TCHAR *String, UINT8 Attributes)
{
    String[0] = 0;
    if (Attributes & USB_AT_SELF_POWERED)
        _tcscat(String,_T("SelfPowered "));
    if (Attributes & USB_AT_REMOTE_WAKEUP)
        _tcscat(String,_T("RemoteWakeup"));
}

PRIVATE void SprintEndpointAttributes(_TCHAR *String, UINT8 Attributes)
{
    String[0] = 0;
    switch (Attributes & USB_AE_MASK) {
    case USB_AE_CONTROL:
        _tcscat(String,_T("Control"));
        break;
    case USB_AE_ISOCHRONOUS:
        _tcscat(String,_T("Isochronous"));
        break;
    case USB_AE_BULK:
        _tcscat(String,_T("BulkData"));
        break;
    case USB_AE_INTERRUPT:
        _tcscat(String,_T("Interrupt"));
        break;
    }

    switch (Attributes & USB_AE_SMASK) {
    case USB_AE_ASYNCHRONOUS:
        _tcscat(String,_T(" Asynchronous"));
        break;
    case USB_AE_ADAPTIVE:
        _tcscat(String,_T(" Adaptive"));
        break;
    case USB_AE_SYNCHRONOUS:
        _tcscat(String,_T(" Syncronous"));
        break;
    }

    switch (Attributes & USB_AE_UMASK) {
    case USB_AE_FEEDBACK:
        _tcscat(String,_T(" Feedback"));
        break;
    case USB_AE_IMPL_FEEDBACK:
        _tcscat(String,_T(" FeedbackData"));
        break;
    }
}

PRIVATE void SprintEndpointMps(_TCHAR *String, UINT Mps)
{
    String[0] = 0;
    switch (Mps & ~USB_MP_MASK) {
    case USB_MP_2xuf:
        _tcscat(String,_T("2xMicroFrame"));
        break;
    case USB_MP_3xuf:
        _tcscat(String,_T("3xMicroFrame"));
        break;
    }
}

PRIVATE void PrintDeviceDescriptor(PUSB_DEVICE_struct Descriptor)
{
    _tprintf(_T("Device descriptor:\n"));
    if (Descriptor->DescriptorType != USB_DT_DEVICE)
        _tprintf(_T("\tDescriptorType %d (?)\n"), Descriptor->DescriptorType);
    if (Descriptor->Length != sizeof(USB_DEVICE_DESCRIPTOR))
        _tprintf(_T("\tLength %d (?)\n"), Descriptor->Length);

    _tprintf(_T("\tUSB Version %d%d.%d%d\n"),
             Descriptor->BcdUSB[1]>>4, Descriptor->BcdUSB[1] & 0xf,
             Descriptor->BcdUSB[0]>>4, Descriptor->BcdUSB[0] & 0xf);
    _tprintf(_T("\tClass %d %s\n"),
             Descriptor->DeviceClass,
             InterfaceClassName(Descriptor->DeviceClass));/*NB: abuse */
    _tprintf(_T("\tSubclass %d\n"), Descriptor->DeviceSubClass);
    _tprintf(_T("\tProtocol %d\n"), Descriptor->DeviceProtocol);
    _tprintf(_T("\tNumber of Configurations %d\n"), Descriptor->NumConfigurations);
    _tprintf(_T("\tMaximum Control Packet Size %d\n"), Descriptor->MaxPacketSize0);
    _tprintf(_T("\tVendor Identifier %d %s\n"),
             Get2(Descriptor->IdVendor),
             VendorName((UINT16)Get2(Descriptor->IdVendor)));
    _tprintf(_T("\tProduct Identifier 0x%x\n"),Get2(Descriptor->IdProduct));
    _tprintf(_T("\tRelease Number %d%d.%d%d\n"),
             Descriptor->BcdDevice[1]>>4, Descriptor->BcdDevice[1] & 0xf,
             Descriptor->BcdDevice[0]>>4, Descriptor->BcdDevice[0] & 0xf);
    if (Descriptor->ixManufacturer != 0) {
        _tprintf(_T("\tManufacturer String Index %d\n"), 
                 Descriptor->ixManufacturer);
        MarkStringReferenced(Descriptor->ixManufacturer);
    }
    if (Descriptor->ixProduct != 0) {
        _tprintf(_T("\tProduct String Index %d\n"), Descriptor->ixProduct);
        MarkStringReferenced(Descriptor->ixProduct);
    }
    if (Descriptor->ixSerialNumber != 0) {
        _tprintf(_T("\tSerial Number String Index %d\n"),
                 Descriptor->ixSerialNumber);
        MarkStringReferenced(Descriptor->ixSerialNumber);
    }
}

PRIVATE void PrintDeviceQualifierDescriptor(PUSB_DEVICE_QUALIFIER_struct Descriptor)
{
    _tprintf(_T("Device Qualifier descriptor:\n"));
    if (Descriptor->DescriptorType != USB_DT_QUALIF)
        _tprintf(_T("\tDescriptorType %d (?)\n"), Descriptor->DescriptorType);
    if (Descriptor->Length != sizeof(USB_DEVICE_QUALIFIER_DESCRIPTOR))
        _tprintf(_T("\tLength %d (?)\n"), Descriptor->Length);

    _tprintf(_T("\tUSB Version %d%d.%d%d\n"),
             Descriptor->BcdUSB[1]>>4, Descriptor->BcdUSB[1] & 0xf,
             Descriptor->BcdUSB[0]>>4, Descriptor->BcdUSB[0] & 0xf);
    _tprintf(_T("\tClass %d %s\n"),
             Descriptor->DeviceClass,
             InterfaceClassName(Descriptor->DeviceClass));/*NB: abuse */
    _tprintf(_T("\tSubclass %d\n"), Descriptor->DeviceSubClass);
    _tprintf(_T("\tProtocol %d\n"), Descriptor->DeviceProtocol);
    _tprintf(_T("\tNumber of Configurations %d\n"), Descriptor->NumConfigurations);
    _tprintf(_T("\tMaximum Control Packet Size %d\n"), Descriptor->MaxPacketSize0);
}

PRIVATE void PrintConfigurationDescriptor(PUSB_CONFIGURATION_struct Descriptor)
{
    _tprintf(_T("Configuration Descriptor:\n"));
    if (Descriptor->DescriptorType != USB_DT_CONFIG)
        _tprintf(_T("\tDescriptorType %d (?)\n"), Descriptor->DescriptorType);
    if (Descriptor->Length != sizeof(USB_CONFIGURATION_DESCRIPTOR))
        _tprintf(_T("\tLength %d (?)\n"), Descriptor->Length);
    _tprintf(_T("\tTotal Length of Configuration Data %d\n"), Get2(Descriptor->TotalLength));
    _tprintf(_T("\tNumber of Interfaces %d\n"), Descriptor->NumInterfaces);
    _tprintf(_T("\tConfiguration Value %d\n"), Descriptor->ConfigurationValue);
    if (Descriptor->ixConfiguration != 0) {
        _tprintf(_T("\tConfiguration String Index %d\n"), 
                 Descriptor->ixConfiguration);
        MarkStringReferenced(Descriptor->ixConfiguration);
    }
    SprintConfigurationAttributes(AttributesBuffer,Descriptor->Attributes);
    _tprintf(_T("\tAttributes 0x%x %s\n"), Descriptor->Attributes,AttributesBuffer);
    _tprintf(_T("\tMaximum Power: %d mA\n"),Descriptor->MaxPower*2);
}

UINT8 LastInterfaceSubclass = 0;

PRIVATE void PrintInterfaceDescriptor(PUSB_INTERFACE_struct Descriptor)
{
    _tprintf(_T("Interface Descriptor:\n"));
    if (Descriptor->DescriptorType != USB_DT_IFACE)
        _tprintf(_T("\tDescriptorType %d (?)\n"), Descriptor->DescriptorType);
    if (Descriptor->Length != sizeof(USB_INTERFACE_DESCRIPTOR))
        _tprintf(_T("\tLength %d (?)\n"), Descriptor->Length);
    _tprintf(_T("\tInterface Number %d\n"),Descriptor->InterfaceNumber);
    _tprintf(_T("\tAlternate Setting %d\n"),Descriptor->AlternateSetting);
    _tprintf(_T("\tNumber of Endpoints %d\n"),Descriptor->NumEndpoints);
    _tprintf(_T("\tInterface Class %d %s\n"),
             Descriptor->InterfaceClass,
             InterfaceClassName(Descriptor->InterfaceClass));
    _tprintf(_T("\tInterface Subclass %d %s\n"),
             Descriptor->InterfaceSubClass,
             InterfaceSubclassName(Descriptor->InterfaceClass,
                                   Descriptor->InterfaceSubClass));

    /* Remember for audio later */
    LastInterfaceSubclass = Descriptor->InterfaceSubClass;

    _tprintf(_T("\tInterface Protocol %d %s\n"),
             Descriptor->InterfaceProtocol,
             InterfaceProtocolName(Descriptor->InterfaceClass,
                                   Descriptor->InterfaceSubClass,
                                   Descriptor->InterfaceProtocol));
    if (Descriptor->ixInterface != 0) {
        _tprintf(_T("\tInterface String Index %d\n"),Descriptor->ixInterface);
        MarkStringReferenced(Descriptor->ixInterface);
    }
}

PRIVATE void PrintEndpointDescriptor( PUSB_ENDPOINT_struct Descriptor)
{
    _tprintf(_T("Endpoint Descriptor:\n"));
    if (Descriptor->DescriptorType != USB_DT_ENDP)
        _tprintf(_T("\tDescriptorType %d (?)\n"), Descriptor->DescriptorType);
    if (Descriptor->Length != sizeof(USB_ENDPOINT_DESCRIPTOR))
        _tprintf(_T("\tLength %d (?)\n"), Descriptor->Length);
    _tprintf(_T("\tEndpoint Address 0x%x: %d %sput\n"),
             Descriptor->EndpointAddress,
             Descriptor->EndpointAddress & USB_EA_ADDRESS_MASK,
             (Descriptor->EndpointAddress & USB_EA_IN) ? _T("in") : _T("out"));
    SprintEndpointAttributes(AttributesBuffer,Descriptor->Attributes);
    _tprintf(_T("\tAttributes 0x%x %s\n"),Descriptor->Attributes,AttributesBuffer);
    SprintEndpointMps(AttributesBuffer,Get2(Descriptor->MaxPacketSize));
    _tprintf(_T("\tMaximum Packet Size %d %s\n"),
             USB_MP_MASK & Get2(Descriptor->MaxPacketSize),
             AttributesBuffer);
    _tprintf(_T("\tPolling Interval %d ms.\n"),Descriptor->Interval);
    if (Descriptor->Length > sizeof(USB_ENDPOINT_DESCRIPTOR)) {
        UINT8 *Data = (UINT8*)Descriptor;
        int i;
        _tprintf(_T("\tNon-standard info:"));
        for (i = sizeof(USB_ENDPOINT_DESCRIPTOR);
             i < Descriptor->Length;
             i++)
            _tprintf(_T(" %x"),Data[i]);
        _tprintf(_T("\n"));        
    }
}

PRIVATE void PrintHidDescriptor(PUSB_HID_struct Descriptor)
{
    _tprintf(_T("HID Descriptor:\n"));
    if (Descriptor->DescriptorType != USB_DT_HID)
        _tprintf(_T("\tDescriptorType %d (?)\n"), Descriptor->DescriptorType);
    if (Descriptor->Length != sizeof(USB_HID_DESCRIPTOR))
        _tprintf(_T("\tLength %d (?)\n"), Descriptor->Length);
    _tprintf(_T("\tHID Version %d%d.%d%d\n"),
             Descriptor->BcdHID[1]>>4, Descriptor->BcdHID[1] & 0xf,
             Descriptor->BcdHID[0]>>4, Descriptor->BcdHID[0] & 0xf);
    if (Descriptor->CountryCode != 0)
        _tprintf(_T("\tCountry Code %d %s\n"),
                 Descriptor->CountryCode,
                 CountryName(Descriptor->CountryCode));
    _tprintf(_T("\tNumber of Descriptors %d\n"),Descriptor->NumDescriptors);
    _tprintf(_T("\tReport Descriptor Type %d\n"),Descriptor->ReportDescriptorType);
    _tprintf(_T("\tReport Descriptor Length %d\n"),
             Get2(Descriptor->ReportDescriptorLength));
}

PRIVATE void PrintAudioControlInterfaceDescriptor(UINT8 *DescriptorData, UINT Length)
{
    UINT i;
    PUSB_AC_struct AcDescriptor = (PUSB_AC_struct)DescriptorData;
    PUSB_INPUT_TERMINAL_struct ItDescriptor =
        (PUSB_INPUT_TERMINAL_struct)DescriptorData;
    PUSB_OUTPUT_TERMINAL_struct OtDescriptor =
        (PUSB_OUTPUT_TERMINAL_struct)DescriptorData;
    PUSB_MIXER_UNIT_struct MuDescriptor =
        (PUSB_MIXER_UNIT_struct)DescriptorData;
    PUSB_SELECTOR_UNIT_struct SuDescriptor =
        (PUSB_SELECTOR_UNIT_struct)DescriptorData;
    PUSB_FEATURE_UNIT_struct FuDescriptor =
        (PUSB_FEATURE_UNIT_struct)DescriptorData;

    _tprintf(_T("Audio Control Interface %s Descriptor:\n"),
             AudioControlSubtypeName(AcDescriptor->DescriptorSubType));
    if (AcDescriptor->DescriptorType != USB_DT_AUDIO_IF)
        _tprintf(_T("\tDescriptorType %d (?)\n"), AcDescriptor->DescriptorType);

    switch (AcDescriptor->DescriptorSubType) {
    case USB_DST_HEADER:
        if (AcDescriptor->Length != sizeof(USB_AC_DESCRIPTOR))
            _tprintf(_T("\tLength %d\n"), AcDescriptor->Length);
        _tprintf(_T("\tClass Spec Version %d%d.%d%d\n"),
                 AcDescriptor->BcdADC[1]>>4, AcDescriptor->BcdADC[1] & 0xf,
                 AcDescriptor->BcdADC[0]>>4, AcDescriptor->BcdADC[0] & 0xf);
        _tprintf(_T("\tTotal Length %d\n"),
                 Get2(AcDescriptor->TotalLength));
        _tprintf(_T("\tIn Collection %d\n"),AcDescriptor->InCollection);
        for (i = 0; i <= (AcDescriptor->Length-sizeof(USB_AC_DESCRIPTOR)); i++) {
            _tprintf(_T("\tInterface Number %d\n"),
                     AcDescriptor->InterfaceNumber[i]);
        }
        break;

    case USB_DST_INPUT_TERMINAL:
        if (ItDescriptor->Length != sizeof(USB_INPUT_TERMINAL_DESCRIPTOR))
            _tprintf(_T("\tLength %d (?)\n"), ItDescriptor->Length);
        _tprintf(_T("\tTerminalID %d\n"), ItDescriptor->TerminalID);
        _tprintf(_T("\tTerminalType %x\n"), 
                 Get2(ItDescriptor->TerminalType));
        _tprintf(_T("\tAssociatedTerminal %d\n"), ItDescriptor->AssociatedTerminal);
        _tprintf(_T("\tNumberOfChannels %d\n"), ItDescriptor->NumberOfChannels);
        _tprintf(_T("\tChannelConfig %x\n"), 
                 Get2(ItDescriptor->ChannelConfig));
        _tprintf(_T("\tChannelNames %d\n"), ItDescriptor->ChannelNames);
        if (ItDescriptor->ixTerminal) {
            _tprintf(_T("\tTerminal String Index %d\n"),
                     ItDescriptor->ixTerminal);
            MarkStringReferenced(ItDescriptor->ixTerminal);
        }
        break;

    case USB_DST_OUTPUT_TERMINAL:
        if (OtDescriptor->Length != sizeof(USB_OUTPUT_TERMINAL_DESCRIPTOR))
            _tprintf(_T("\tLength %d (?)\n"), OtDescriptor->Length);
        _tprintf(_T("\tTerminalID %d\n"), OtDescriptor->TerminalID);
        _tprintf(_T("\tTerminalType %x\n"), 
                 Get2(OtDescriptor->TerminalType));
        _tprintf(_T("\tAssociatedTerminal %d\n"), OtDescriptor->AssociatedTerminal);
        _tprintf(_T("\tSourceID %d\n"), OtDescriptor->SourceID);
        if (OtDescriptor->ixTerminal) {
            _tprintf(_T("\tTerminal String Index %d\n"),
                     OtDescriptor->ixTerminal);
            MarkStringReferenced(OtDescriptor->ixTerminal);
        }
        break;

    case USB_DST_MIXER_UNIT:
        _tprintf(_T("\tLength %d\n"), MuDescriptor->Length);
        _tprintf(_T("\tUnitID %d\n"), MuDescriptor->UnitID);
        _tprintf(_T("\tNumber of Pins %d\n"), MuDescriptor->NumberOfPins);
        i = MuDescriptor->NumberOfPins;
        while (i-- > 0) {
            _tprintf(_T("\tSourceID %d\n"), MuDescriptor->Pins[0].SourceID);
            _tprintf(_T("\tNumber of Channels %d\n"),
                     MuDescriptor->Pins[0].NumberOfChannels);
            _tprintf(_T("\tChannelConfig %x\n"), 
                     Get2(MuDescriptor->Pins[0].ChannelConfig));
            _tprintf(_T("\tAssociatedTerminal %d\n"),
                     MuDescriptor->Pins[0].AssociatedTerminal);
            _tprintf(_T("\tTerminal %d\n"), MuDescriptor->Pins[0].Terminal);

            MuDescriptor = (PUSB_MIXER_UNIT_struct)
                ((ADDRESS)MuDescriptor + sizeof(struct _USB_MIXER_PIN_struct));
        }
        if (MuDescriptor->ixMixer) {
            _tprintf(_T("\tMixer String Index String Index %x\n"), 
                     MuDescriptor->ixMixer);
            MarkStringReferenced(MuDescriptor->ixMixer);
        }
        break;

    case USB_DST_SELECTOR_UNIT:
        _tprintf(_T("\tLength %d\n"), SuDescriptor->Length);
        _tprintf(_T("\tUnitID %d\n"), SuDescriptor->UnitID);
        _tprintf(_T("\tNumber of Pins %d\n"), SuDescriptor->NumberOfPins);
        i = SuDescriptor->NumberOfPins;
        while (i-- > 0) {
            _tprintf(_T("\tSourceID %d\n"), SuDescriptor->Pins[0].SourceID);

            SuDescriptor = (PUSB_SELECTOR_UNIT_struct)
                ((ADDRESS)SuDescriptor + sizeof(struct _USB_SELECTOR_PIN_struct));
        }
        if (SuDescriptor->ixSelector) {
            _tprintf(_T("\tSelector String Index String Index %x\n"), 
                     SuDescriptor->ixSelector);
            MarkStringReferenced(SuDescriptor->ixSelector);
        }
        break;

    case USB_DST_FEATURE_UNIT:
        _tprintf(_T("\tLength %d\n"), FuDescriptor->Length);
        _tprintf(_T("\tUnitID %d\n"), FuDescriptor->UnitID);
        _tprintf(_T("\tNumber of Controls %d\n"), FuDescriptor->NumberOfControls);
        i = FuDescriptor->NumberOfControls;
        while (i-- > 0) {
            _tprintf(_T("\tSourceID %d\n"),
                     Get2(FuDescriptor->Controls[0].Bitmap));

            FuDescriptor = (PUSB_FEATURE_UNIT_struct)
                ((ADDRESS)FuDescriptor + sizeof(struct _USB_FEATURE_CONTROL_struct));
        }
        if (FuDescriptor->ixFeature) {
            _tprintf(_T("\tFeature String Index String Index %x\n"),
                     FuDescriptor->ixFeature);
            MarkStringReferenced(FuDescriptor->ixFeature);
        }
        break;

    case USB_DST_PROCESSING_UNIT:
    case USB_DST_EXTENSION_UNIT:
    default:
        _tprintf(_T("\t"));
        for (i = 0; i < Length; i++)
            _tprintf(_T("%x "), DescriptorData[i]);
        _tprintf(_T("\n"));
    }
}

PRIVATE void PrintSamplingFrequencies(UINT8 *SfDescriptor, UINT NumFrequencies)
{
    UINT i;
#define Get3(f) (((f)[0]) | ((f)[1]<<8) | ((f)[2]<<16))

    /* Continuous? */
    if (SfDescriptor[0] == 0) {
        _tprintf(_T("\tContinuous Sampling Frequencies: [%d .. %d]:\n"),
                 Get3(SfDescriptor+1), Get3(SfDescriptor+4));
        return;
    }

    _tprintf(_T("\t%d Sampling Frequencies:\n"),SfDescriptor[0]);
    for (i = SfDescriptor[0]; i > 0; i--) {
        _tprintf(_T("\t  %d\n"), Get3(SfDescriptor+1));
        SfDescriptor += 3;
    }
}

PRIVATE void PrintAudioStreamingInterfaceDescriptor(UINT8 *DescriptorData, UINT Length)
{
    UINT i;
    PUSB_AS_GENERAL_struct AsDescriptor = (PUSB_AS_GENERAL_struct)DescriptorData;
    PUSB_AS_FORMAT_TYPE1_struct Fm1Descriptor =
        (PUSB_AS_FORMAT_TYPE1_struct)DescriptorData;
    PUSB_AS_FORMAT_TYPE2_struct Fm2Descriptor =
        (PUSB_AS_FORMAT_TYPE2_struct)DescriptorData;

    _tprintf(_T("Audio Streaming Interface %s Descriptor:\n"),
             AudioStreamingSubtypeName(AsDescriptor->DescriptorSubType));
    if (AsDescriptor->DescriptorType != USB_DT_AUDIO_IF)
        _tprintf(_T("\tDescriptorType %d (?)\n"), AsDescriptor->DescriptorType);

    switch (AsDescriptor->DescriptorSubType) {

    case USB_DST_AS_GENERAL:
        if (AsDescriptor->Length != sizeof(USB_AS_GENERAL_DESCRIPTOR))
            _tprintf(_T("\tLength %d (?)\n"), AsDescriptor->Length);
        _tprintf(_T("\tTerminalLink %d\n"),AsDescriptor->TerminalLink);
        _tprintf(_T("\tDelay %d\n"),AsDescriptor->Delay);
        _tprintf(_T("\tFormat Tag %d\n"),
                 Get2(AsDescriptor->FormatTag));
        break;

    case USB_DST_FORMAT_TYPE:
        switch (Fm1Descriptor->FormatType) {
        case 1:
        case 3:
            _tprintf(_T("\tNumberOfChannels %d\n"),Fm1Descriptor->NumberOfChannels);
            _tprintf(_T("\tSubframeSize %d\n"),Fm1Descriptor->SubframeSize);
            _tprintf(_T("\tBitResolution %d\n"),Fm1Descriptor->BitResolution);
            PrintSamplingFrequencies((UINT8*)&Fm1Descriptor->SamplingFrequencyType,
                                     (Length - 8) / 3);
            return;
        case 2:
            _tprintf(_T("\tMaxBitRate %d\n"),
                     Get2(Fm2Descriptor->MaxBitRate));
            _tprintf(_T("\tSamplesPerFrame %d\n"),
                     Get2(Fm2Descriptor->SamplesPerFrame));
            PrintSamplingFrequencies((UINT8*)&Fm2Descriptor->SamplingFrequencyType,
                                     (Length - 9) / 3);
            return;
        }
        /* fall through */

    case USB_DST_FORMAT_SPECIFIC:
    default:
        _tprintf(_T("\t"));
        for (i = 0; i < Length; i++)
            _tprintf(_T("%x "), DescriptorData[i]);
        _tprintf(_T("\n"));
    }
}

PRIVATE void PrintAudioInterfaceDescriptor(UINT8 *DescriptorData, UINT Length)
{
    if (LastInterfaceSubclass == USB_DS_AUDIO_CONTROL)
        PrintAudioControlInterfaceDescriptor(DescriptorData,Length);
    if (LastInterfaceSubclass == USB_DS_AUDIO_STREAMING)
        PrintAudioStreamingInterfaceDescriptor(DescriptorData,Length);
}

PRIVATE void PrintAudioEndpointDescriptor(PUSB_AS_ENDPOINT_struct Descriptor)
{
    _tprintf(_T("Audio Endpoint Descriptor:\n"));

    if (Descriptor->Length != sizeof(USB_AS_ENDPOINT_DESCRIPTOR))
        _tprintf(_T("\tLength %d (?)\n"), Descriptor->Length);
    if (Descriptor->DescriptorType != USB_DT_AUDIO_EP)
        _tprintf(_T("\tDescriptorType %d (?)\n"), Descriptor->DescriptorType);
    if (Descriptor->DescriptorSubType != USB_DST_AS_GENERAL)
        _tprintf(_T("\tDescriptorSubType %d (?)\n"), Descriptor->DescriptorSubType);
    _tprintf(_T("\tAttributes %x\n"),Descriptor->Attributes);
    _tprintf(_T("\tLock Delay Units %x\n"),Descriptor->LockDelayUnits);
    _tprintf(_T("\tLock Delay %d\n"),
             Get2(Descriptor->LockDelay));
}

PRIVATE void PrintHubDescriptor(PUSB_HUB_struct Descriptor)
{
    _tprintf(_T("HUB Descriptor:\n"));
    if (Descriptor->DescriptorType != USB_DT_HUB)
        _tprintf(_T("\tDescriptorType %d (?)\n"), Descriptor->DescriptorType);
    if (Descriptor->Length != sizeof(USB_HUB_DESCRIPTOR))
        /* NB: A HUB with more than 8 ports would do that.
         * There is no such thing, far as I know.
         */
        _tprintf(_T("\tLength %d (?)\n"), Descriptor->Length);
    _tprintf(_T("\tNumberOfPorts %d\n"), Descriptor->NumberOfPorts);
    _tprintf(_T("\tHubCharacteristics %x\n"),
             Get2(Descriptor->HubCharacteristics));
    _tprintf(_T("\tPowerOnGood %d ms.\n"), Descriptor->PowerOnGood*2);
    _tprintf(_T("\tControlCurrent %d mA.\n"), Descriptor->ControlCurrent);
    _tprintf(_T("\tDeviceRemovable %x\n"), Descriptor->DeviceRemovable);
    _tprintf(_T("\tCompat1.0 %x\n"), Descriptor->Compat10);
}

PRIVATE void PrintDescriptor( UINT8 *DescriptorData, UINT Length)
{
    UINT i;

    switch (DescriptorData[1]) {
    case USB_DT_DEVICE:
        PrintDeviceDescriptor((PUSB_DEVICE_struct) DescriptorData);
        break;
    case USB_DT_CONFIG:
        PrintConfigurationDescriptor((PUSB_CONFIGURATION_struct) DescriptorData);
        break;
    case USB_DT_IFACE:
        PrintInterfaceDescriptor((PUSB_INTERFACE_struct) DescriptorData);
        break;
    case USB_DT_ENDP:
        PrintEndpointDescriptor((PUSB_ENDPOINT_struct) DescriptorData);
        break;
    case USB_DT_QUALIF:
        PrintDeviceQualifierDescriptor((PUSB_DEVICE_QUALIFIER_struct) DescriptorData);
        break;
    case USB_DT_AUDIO_IF: /* polymorphic */
        PrintAudioInterfaceDescriptor(DescriptorData, Length);
        break;
    case USB_DT_AUDIO_EP:
        PrintAudioEndpointDescriptor((PUSB_AS_ENDPOINT_struct) DescriptorData);
        break;
    case USB_DT_HID:
        PrintHidDescriptor((PUSB_HID_struct) DescriptorData);
        break;
    case USB_DT_HUB:
        PrintHubDescriptor((PUSB_HUB_struct) DescriptorData);
        break;
    case USB_DT_STRING:
    case USB_DT_REPORT:
    case USB_DT_PHYS:
    default:
        _tprintf(_T("\t"));
        for (i = 0; i < Length; i++)
            _tprintf(_T("%x "), DescriptorData[i]);
        _tprintf(_T("\n"));
    }
}

PRIVATE SCODE do1Str(PIUSBHUB Hub, PUSBM_REQUEST_BLOCK pUrb, UINT8 UsbAddress, UINT8 Port)
{
    SCODE sc;
    CONDITION Done;

    Condition_Init(&Done);

    DBGME(0,_tprintf(_T("do1Str(%x,%d,%d)\n"),(ADDRESS)pUrb,UsbAddress,Port));

    pUrb->Done = &Done;
    pUrb->DeviceAddress = UsbAddress;
    pUrb->Pid = USB_PID_SETUP;
    pUrb->Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    pUrb->SetupData[0] = USB_RT_DTH;
    pUrb->SetupData[1] = USB_R_GET_DESCRIPTOR;
    pUrb->SetupData[3] = USB_DT_STRING;
    pUrb->SetupLength = 8;
    pUrb->Data = (UINT8*) Buffer;
    pUrb->MaximumPacketSize = 8; /* BUGBUG */

    sc = Hub->v->SubmitUrb(Hub,
                           pUrb,
                           Port);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("do1Str: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    Condition_Wait(pUrb->Done,NULL,NULL);
#undef maxdesc
    return S_OK;
}

PRIVATE SCODE doStr(PIUSBHUB Hub, UINT8 UsbAddress, UINT8 Port)
{
    SCODE sc;
    int i;
    BOOL GotSome = FALSE;
    struct USBM_REQUEST_BLOCK Urb;

    memset(&Urb,0,sizeof Urb);

    for (i = 0; i < 256; i++)
    {
        if (!StringWasReferenced(i))
            continue;

        Urb.SetupData[2] = i;
        Urb.SetupData[4] = 9;
        Urb.SetupData[5] = 4;
        Set2(&Urb.SetupData[6],MAX_PATH);
        Urb.ExpectedDataLength = MAX_PATH;

        sc = do1Str(Hub,&Urb,UsbAddress,Port);
        if (!FAILED(sc) && (0 == (Urb.Flags & URB_FLAG_ERROR)))
        {
            int j, n = (Urb.ActualDataLength) ? Urb.Data[0] : 0;

            if (!GotSome)
                _tprintf(_T("Strings:\n"));
            GotSome = TRUE;

#if _DEBUG
            _tprintf(_T("[%d (%d)]"), i, n);
            PrintUrbResults(&Urb,_T("Get String:"));
#endif
            _tprintf(_T("\t[%d] \""),i);
            for (j = 2; j < n; j += 2)
                _puttchar(Urb.Data[j]);
            _tprintf(_T("\"\n"));
        }
    }

    return S_OK;
}

PRIVATE SCODE doDesc(PIUSBHUB Hub, UINT8 UsbAddress, UINT8 Port)
{
    PUSB_DEVICE_struct Descriptor;
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
    CONDITION Done;

    Condition_Init(&Done);

#define maxdesc 64
    Descriptor = (PUSB_DEVICE_struct) Buffer;

    DBGME(0,_tprintf(_T("doDesc(%d)\n"),UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &Done;
    Urb.DeviceAddress = UsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb.SetupData[0] = USB_RT_DTH;
    Urb.SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb.SetupData[3] = USB_DT_DEVICE;
    Urb.SetupData[6] = maxdesc;
    Urb.SetupLength = 8;
    Urb.ExpectedDataLength = maxdesc;
    Urb.Data = (UINT8 *)Descriptor;
    Urb.MaximumPacketSize = 8; /* BUGBUG */

    sc = Hub->v->SubmitUrb(Hub,
                           &Urb,
                           Port);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("doDesc: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);

#if _DEBUG
    PrintUrbResults(&Urb,_T("Device Descriptor:"));
#endif
    PrintDeviceDescriptor(Descriptor);

    return S_OK;
#undef maxdesc
}

PRIVATE SCODE doRep(
    PIUSBHUB Hub,
    UINT8 UsbAddress,
    UINT8 Port,
    UINT8 Interface,
    UINT16 RepLen);

PRIVATE SCODE doConf(
    PIUSBHUB Hub, 
    UINT8 UsbAddress,
    UINT8 Port)
{
    int i;
#define maxifaces 8
    struct {
        UINT8 InterfaceIndex;
        UINT16 ReportLength;
    } Interfaces[maxifaces];
    int nIfaces = -1;
    PUSB_CONFIGURATION_struct Descriptor;
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
    CONDITION Done;

    Condition_Init(&Done);

#define maxdesc 512
    Descriptor = (PUSB_CONFIGURATION_struct) Buffer;

    DBGME(0,_tprintf(_T("doConf(%d)\n"),UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &Done;
    Urb.DeviceAddress = UsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb.SetupData[0] = USB_RT_DTH;
    Urb.SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb.SetupData[3] = USB_DT_CONFIG;
    Set2(&Urb.SetupData[6],maxdesc);
    Urb.SetupLength = 8;
    Urb.ExpectedDataLength = maxdesc;
    Urb.Data = (UINT8 *)Descriptor;
    Urb.MaximumPacketSize = 8; /* BUGBUG */

    sc = Hub->v->SubmitUrb(Hub,
                           &Urb,
                           Port);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("doConfDesc: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);

#if _DEBUG
    PrintUrbResults(&Urb,_T("Configuration Descriptor:"));
#endif
    /* It must be a config one we got first.
     */
    PrintConfigurationDescriptor(Descriptor);

    memset(Interfaces,0,sizeof Interfaces);
    for (i = 9; i < Urb.ActualDataLength;) {
        PrintDescriptor(Urb.Data+i,Urb.Data[i]);

        /* If an interface get the number (specs say must be first)
         */
        if (Urb.Data[i+1] == USB_DT_IFACE) {
            if (++nIfaces < maxifaces)
                Interfaces[nIfaces].InterfaceIndex = Urb.Data[i+2];
        }

        /* Look for the HID report descriptor length, if any
         */
        if (Urb.Data[i+1] == USB_DT_HID) {
            if (nIfaces < maxifaces)
                Interfaces[nIfaces].ReportLength = Get2(&Urb.Data[i+7]);
        }

        /* Do not crash on screwy data */
        i += (Urb.Data[i]) ? Urb.Data[i] : Urb.ActualDataLength;
    }

    for (i = 0; i <= nIfaces; i++) {
        if (i >= maxifaces) break;
        if (Interfaces[i].ReportLength != 0)
            sc = doRep(Hub,UsbAddress,Port,
                       Interfaces[i].InterfaceIndex,
                       Interfaces[i].ReportLength);
    }

    return S_OK;
#undef maxifaces
#undef maxdesc
}

PRIVATE INT32 pn(UINT8 *p, int nb)
{
    /* Simplistic!! */
    switch (nb) {
    case 1: return (INT32) (INT8) p[0];
    case 2: return (INT32) (INT16) (p[0] | (p[1] << 8));
    case 4: return (INT32) (p[0] | (p[1]<<8) | (p[2]<<16) | (p[3]<<24));
    default:
        return 0; /* shouldnthappen! */
    }
}

PRIVATE _TCHAR *UsagePageName(INT32 Page)
{
    switch (Page) {
    case 0x01: return _T("Generic Desktop");
    case 0x02: return _T("Simulation Controller");
    case 0x03: return _T("Virtual Reality Controller");
    case 0x04: return _T("Sport Controller");
    case 0x05: return _T("Game Controller");
    case 0x06: return _T("Generic Device");
    case 0x07: return _T("Keyboard");
    case 0x08: return _T("LEDs");
    case 0x09: return _T("Button");
    case 0x0A: return _T("Ordinal");
    case 0x0B: return _T("Telephony");
    case 0x0C: return _T("Consumer");
    case 0x0D: return _T("Digitizer");
    case 0x0F: return _T("Physical Interface");
    case 0x10: return _T("Unicode");
    case 0x14: return _T("Alphanumeric Display");
    case 0x40: return _T("Medical Instrument");
    case 0x80: return _T("80-Monitor");
    case 0x81: return _T("81-Monitor");
    case 0x82: return _T("82-Monitor");
    case 0x83: return _T("83-Monitor");
    case 0x84: return _T("84-Power");
    case 0x85: return _T("85-Power");
    case 0x86: return _T("86-Power");
    case 0x87: return _T("87-Power");
    case 0x8C: return _T("Bar Code Scanner");
    case 0x8D: return _T("Scale");
    case 0x8E: return _T("Magnetic Stripe Reader");
    case 0x8F: return _T("Reservedf for Point of Sale");
    case 0x90: return _T("Camera Control");
    case 0x91: return _T("Arcade");
    case -1:   return _T("Vendor Defined");
    default:
        break;
    }
    _stprintf(AttributesBuffer,_T("x%x"),Page);
    return AttributesBuffer;
}

PRIVATE _TCHAR *UsageName(INT32 Page, INT32 Usage)
{
    switch (Page) {
    case 0x01: /* Generic Desktop */
        switch (Usage) {
        case 0x01: return _T("Pointer");
        case 0x02: return _T("Mouse");
        case 0x04: return _T("Joystick");
        case 0x05: return _T("Game Pad");
        case 0x06: return _T("Keyboard");
        case 0x07: return _T("Keypad");
        case 0x08: return _T("Multi-axis Controller");
        case 0x30: return _T("X");
        case 0x31: return _T("Y");
        case 0x32: return _T("Z");
        case 0x33: return _T("Rx");
        case 0x34: return _T("Ry");
        case 0x35: return _T("Rz");
        case 0x36: return _T("Slider");
        case 0x37: return _T("Dial");
        case 0x38: return _T("Wheel");
        case 0x39: return _T("Hat switch");
        case 0x3A: return _T("Counted Buffer");
        case 0x3B: return _T("Byte Count");
        case 0x3C: return _T("Motion Wakeup");
        case 0x3D: return _T("Start");
        case 0x3E: return _T("Select");
        case 0x3F: return _T("Reserved");
        case 0x40: return _T("Vx");
        case 0x41: return _T("Vy");
        case 0x42: return _T("Vz");
        case 0x43: return _T("Vbrx");
        case 0x44: return _T("Vbry");
        case 0x45: return _T("Vbrz");
        case 0x46: return _T("Vno");
        case 0x47: return _T("Feature Notification");
        case 0x80: return _T("System Control");
        case 0x81: return _T("System Power Down");
        case 0x82: return _T("System Sleep");
        case 0x83: return _T("System Wake Up");
        case 0x84: return _T("System Context Menu");
        case 0x85: return _T("System Main Menu");
        case 0x86: return _T("System App Menu");
        case 0x87: return _T("System Menu Help");
        case 0x88: return _T("System Menu Exit");
        case 0x89: return _T("System Menu Select");
        case 0x8A: return _T("System Menu Right");
        case 0x8B: return _T("System Menu Left");
        case 0x8C: return _T("System Menu Up");
        case 0x8D: return _T("System Menu Down");
        case 0x8E: return _T("System Cold Restart");
        case 0x8F: return _T("System Warm Restart");
        case 0x90: return _T("D-pad Up");
        case 0x91: return _T("D-pad Down");
        case 0x92: return _T("D-pad Right");
        case 0x93: return _T("D-pad Left");
        case 0xA0: return _T("System Dock");
        case 0xA1: return _T("System Undock");
        case 0xA2: return _T("System Setup");
        case 0xA3: return _T("System Break");
        case 0xA4: return _T("System Debugger Break");
        case 0xA5: return _T("Application Break");
        case 0xA6: return _T("Application Debugger Break");
        case 0xA7: return _T("System Speaker Mute");
        case 0xA8: return _T("System Hibernate");
        case 0xB0: return _T("System Display Invert");
        case 0xB1: return _T("System Display Internal");
        case 0xB2: return _T("System Display External");
        case 0xB3: return _T("System Display Both");
        case 0xB4: return _T("System Display Dual");
        case 0xB5: return _T("System Display Toggle Int/Ext");
        case 0xB6: return _T("System Display Swap Primary/Secondary");
        case 0xB7: return _T("System Display LCD Autoscale");
        default:
            break;
        }
        break;
    case 0x02: /* Simulation */
        switch (Usage) {
        case 0x01: return _T("Flight Simulation");
        case 0x02: return _T("Automobile Simulation");
        case 0x03: return _T("Tank Simulation");
        case 0x04: return _T("Spaceship Simulation");
        case 0x05: return _T("Submarine Simulation");
        case 0x06: return _T("Sailing Simulation");
        case 0x07: return _T("Motorcycle Simulation");
        case 0x08: return _T("Sports Simulation");
        case 0x09: return _T("Airplane Simulation");
        case 0x0A: return _T("Helicopter Simulation");
        case 0x0B: return _T("Magic Carpet Simulation");
        case 0x0C: return _T("Bicycle Simulation");
        case 0x20: return _T("Flight Control Stick");
        case 0x21: return _T("Flight Stick");
        case 0x22: return _T("Cyclic Control");
        case 0x23: return _T("Cyclic Trim");
        case 0x24: return _T("Flight Yoke");
        case 0x25: return _T("Track Control");
        case 0xB0: return _T("Aileron");
        case 0xB1: return _T("Aileron Trim");
        case 0xB2: return _T("Anti-Torque Control");
        case 0xB3: return _T("Autopilot Enable");
        case 0xB4: return _T("Chaff Release");
        case 0xB5: return _T("Collective Control");
        case 0xB6: return _T("Dive Brake");
        case 0xB7: return _T("Electronic Countermeasures");
        case 0xB8: return _T("Elevator");
        case 0xB9: return _T("Elevator Trim");
        case 0xBA: return _T("Rudder");
        case 0xBB: return _T("Throttle");
        case 0xBC: return _T("Flight Communications");
        case 0xBD: return _T("Flare Release");
        case 0xBE: return _T("Landing Gear");
        case 0xBF: return _T("Toe Brake");
        case 0xC0: return _T("Trigger");
        case 0xC1: return _T("Weapons Arm");
        case 0xC2: return _T("Weapons Select");
        case 0xC3: return _T("Wing Flaps");
        case 0xC4: return _T("Accelerator");
        case 0xC5: return _T("Brake");
        case 0xC6: return _T("Clutch");
        case 0xC7: return _T("Shifter");
        case 0xC8: return _T("Steering");
        case 0xC9: return _T("Turret Direction");
        case 0xCA: return _T("Barrel Elevation");
        case 0xCB: return _T("Dive Plane");
        case 0xCC: return _T("Ballast");
        case 0xCD: return _T("Bicycle Crank");
        case 0xCE: return _T("Handle Bars");
        case 0xCF: return _T("Front Brake");
        case 0xD0: return _T("Rear Brake");
        default:
            break;
        }
        break;
    case 0x03: /* VR Controls */
        switch (Usage) {
        case 0x01: return _T("Belt");
        case 0x02: return _T("Body Suit");
        case 0x03: return _T("Flexor");
        case 0x04: return _T("Glove");
        case 0x05: return _T("Head Tracker");
        case 0x06: return _T("Head Mounted Display");
        case 0x07: return _T("Hand Tracker");
        case 0x08: return _T("Oculometer");
        case 0x09: return _T("Vest");
        case 0x0A: return _T("Animatronic Device");
        case 0x20: return _T("Stereo Enable");
        case 0x21: return _T("Display Enable");
        default:
            break;
        }
        break;
    case 0x04: /* Sports */
        switch (Usage) {
        case 0x01: return _T("Baseball Bat");
        case 0x02: return _T("Golf Club");
        case 0x03: return _T("Rowing Machine");
        case 0x04: return _T("Treadmill");
        case 0x30: return _T("Oar");
        case 0x31: return _T("Slope");
        case 0x32: return _T("Rate");
        case 0x33: return _T("Stick Speed");
        case 0x34: return _T("Stick Face Angle");
        case 0x35: return _T("Stick Heel/Toe");
        case 0x36: return _T("Stick Follow Through");
        case 0x37: return _T("Stick Tempo");
        case 0x38: return _T("Stick Type");
        case 0x39: return _T("Stick Height");
        case 0x50: return _T("Putter");
        case 0x51: return _T("1 Iron");
        case 0x52: return _T("2 Iron");
        case 0x53: return _T("3 Iron");
        case 0x54: return _T("4 Iron");
        case 0x55: return _T("5 Iron");
        case 0x56: return _T("6 Iron");
        case 0x57: return _T("7 Iron");
        case 0x58: return _T("8 Iron");
        case 0x59: return _T("9 Iron");
        case 0x5A: return _T("10 Iron");
        case 0x5B: return _T("11 Iron");
        case 0x5C: return _T("Sand Wedge");
        case 0x5D: return _T("Loft Wedge");
        case 0x5E: return _T("Power Wedge");
        case 0x5F: return _T("1 Wood");
        case 0x60: return _T("3 Wood");
        case 0x61: return _T("5 Wood");
        case 0x62: return _T("7 Wood");
        case 0x63: return _T("9 Wood");
        default:
            break;
        }
        break;
    case 0x05: /* Game Pad */
        switch (Usage) {
        case 0x01: return _T("3D Game Controller");
        case 0x02: return _T("Pinball Device");
        case 0x03: return _T("Gun Device");
        case 0x20: return _T("Point of View");
        case 0x21: return _T("Turn Right/Left");
        case 0x22: return _T("Pitch Forward/Backward");
        case 0x23: return _T("Roll Right/Left");
        case 0x24: return _T("Move Right/Left");
        case 0x25: return _T("Move Forward/Backward");
        case 0x26: return _T("Move Up/Down");
        case 0x27: return _T("Lean Right/Left");
        case 0x28: return _T("Lean Forward/Backward");
        case 0x29: return _T("Height of POV");
        case 0x2A: return _T("Flipper");
        case 0x2B: return _T("Secondary Flipper");
        case 0x2C: return _T("Bump");
        case 0x2D: return _T("New Game");
        case 0x2E: return _T("Shoot Ball");
        case 0x2F: return _T("Player");
        case 0x30: return _T("Gun Bolt");
        case 0x31: return _T("Gun Clip");
        case 0x32: return _T("Gun Selector");
        case 0x33: return _T("Gun Single Shot");
        case 0x34: return _T("Gun Burst");
        case 0x35: return _T("Gun Automatic");
        case 0x36: return _T("Gun Safety");
        case 0x37: return _T("Gamepad Fire/Jump");
        case 0x39: return _T("Gamepad Trigger");
        default:
            break;
        }
        break;
    case 0x06: /* Generic Device */
        switch (Usage) {
        case 0x20: return _T("Battery Strength");
        case 0x21: return _T("Wireless Channel");
        case 0x22: return _T("Wireless ID");
        default:
            break;
        }
        break;
    case 0x07: /* keyboard */
        break;
    case 0x08: /* LEDs */
        switch (Usage) {
        case 0x01: return _T("Num Lock");
        case 0x02: return _T("Caps Lock");
        case 0x03: return _T("Scroll Lock");
        case 0x04: return _T("Compose");
        case 0x05: return _T("Kana");
        case 0x06: return _T("Power");
        case 0x07: return _T("Shift");
        case 0x08: return _T("Do Not Disturb");
        case 0x09: return _T("Mute");
        case 0x0A: return _T("Tone Enable");
        case 0x0B: return _T("High Cut Filter");
        case 0x0C: return _T("Low Cut Filter");
        case 0x0D: return _T("Equalizer Enable");
        case 0x0E: return _T("Sound Field On");
        case 0x0F: return _T("Surround On");
        case 0x10: return _T("Repeat");
        case 0x11: return _T("Stereo");
        case 0x12: return _T("Sampling Rate Detect");
        case 0x13: return _T("Spinning");
        case 0x14: return _T("CAV");
        case 0x15: return _T("CLV");
        case 0x16: return _T("Recording Format Detect");
        case 0x17: return _T("Off-Hook");
        case 0x18: return _T("Ring");
        case 0x19: return _T("Message Waiting");
        case 0x1A: return _T("Data Mode");
        case 0x1B: return _T("Battery Operation");
        case 0x1C: return _T("Battery OK");
        case 0x1D: return _T("Battery Low");
        case 0x1E: return _T("Speaker");
        case 0x1F: return _T("Head Set");
        case 0x20: return _T("Hold");
        case 0x21: return _T("Microphone");
        case 0x22: return _T("Coverage");
        case 0x23: return _T("Night Mode");
        case 0x24: return _T("Send Calls");
        case 0x25: return _T("Call Pickup");
        case 0x26: return _T("Conference");
        case 0x27: return _T("Stand-by");
        case 0x28: return _T("Camera On");
        case 0x29: return _T("Camera Off");
        case 0x2A: return _T("On-Line");
        case 0x2B: return _T("Off-Line");
        case 0x2C: return _T("Busy");
        case 0x2D: return _T("Ready");
        case 0x2E: return _T("Paper-Out");
        case 0x2F: return _T("Paper-Jam");
        case 0x30: return _T("Remote");
        case 0x31: return _T("Forward");
        case 0x32: return _T("Reverse");
        case 0x33: return _T("Stop");
        case 0x34: return _T("Rewind");
        case 0x35: return _T("Fast Forward");
        case 0x36: return _T("Play");
        case 0x37: return _T("Pause");
        case 0x38: return _T("Record");
        case 0x39: return _T("Error");
        case 0x3A: return _T("Usage Selected Indicator");
        case 0x3B: return _T("Usage In Use Indicator");
        case 0x3C: return _T("Usage Multi Mode Indicator");
        case 0x3D: return _T("Indicator On");
        case 0x3E: return _T("Indicator Flash");
        case 0x3F: return _T("Indicator Slow Blink");
        case 0x40: return _T("Indicator Fast Blink");
        case 0x41: return _T("Indicator Off");
        case 0x42: return _T("Flash On Time");
        case 0x43: return _T("Slow Blink On Time");
        case 0x44: return _T("Slow Blink Off Time");
        case 0x45: return _T("Fast Blink On Time");
        case 0x46: return _T("Fast Blink Off Time");
        case 0x47: return _T("Usage Indicator Color");
        case 0x48: return _T("Indicator Red");
        case 0x49: return _T("Indicator Green");
        case 0x4A: return _T("Indicator Amber");
        case 0x4B: return _T("Generic Indicator");
        case 0x4C: return _T("System Suspend");
        case 0x4D: return _T("External Power Connected");
        default:
            break;
        }
        break;
    case 0x09: /* Button */
        _stprintf(AttributesBuffer,_T("Button %d"),Usage);
        return AttributesBuffer;
    case 0x0a:
        break;
    case 0x0b: /* Telephony */
        switch (Usage) {
        case 0x01: return _T("Phone");
        case 0x02: return _T("Answering Machine");
        case 0x03: return _T("Message Controls");
        case 0x04: return _T("Handset");
        case 0x05: return _T("Headset");
        case 0x06: return _T("Telephony Key Pad");
        case 0x07: return _T("Programmable Button");
        case 0x20: return _T("Hook Switch");
        case 0x21: return _T("Flash");
        case 0x22: return _T("Feature");
        case 0x23: return _T("Hold");
        case 0x24: return _T("Redial");
        case 0x25: return _T("Transfer");
        case 0x26: return _T("Drop");
        case 0x27: return _T("Park");
        case 0x28: return _T("Forward Calls");
        case 0x29: return _T("Alternate Function");
        case 0x2A: return _T("Line");
        case 0x2B: return _T("Speaker Phone");
        case 0x2C: return _T("Conference");
        case 0x2D: return _T("Ring Enable");
        case 0x2E: return _T("Ring Select");
        case 0x2F: return _T("Phone Mute");
        case 0x30: return _T("Caller ID");
        case 0x31: return _T("Send");
        case 0x50: return _T("Speed Dial");
        case 0x51: return _T("Store Number");
        case 0x52: return _T("Recall Number");
        case 0x53: return _T("Phone Directory");
        case 0x70: return _T("Voice Mail");
        case 0x71: return _T("Screen Calls");
        case 0x72: return _T("Do Not Disturb");
        case 0x73: return _T("Message");
        case 0x74: return _T("Answer On/Off");
        case 0x90: return _T("Inside Dial Tone");
        case 0x91: return _T("Outside Dial Tone");
        case 0x92: return _T("Inside Ring Tone");
        case 0x93: return _T("Outside Ring Tone");
        case 0x94: return _T("Priority Ring Tone");
        case 0x95: return _T("Inside Ringback");
        case 0x96: return _T("Priority Ringback");
        case 0x97: return _T("Line Busy Tone");
        case 0x98: return _T("Reorder Tone");
        case 0x99: return _T("Call Waiting Tone");
        case 0x9A: return _T("Confirmation Tone 1");
        case 0x9B: return _T("Confirmation Tone 2");
        case 0x9C: return _T("Tones Off");
        case 0x9D: return _T("Outside Ringback");
        case 0x9E: return _T("Ringer");
        case 0xB0: return _T("Key 0");
        case 0xB1: return _T("Key 1");
        case 0xB2: return _T("Key 2");
        case 0xB3: return _T("Key 3");
        case 0xB4: return _T("Key 4");
        case 0xB5: return _T("Key 5");
        case 0xB6: return _T("Key 6");
        case 0xB7: return _T("Key 7");
        case 0xB8: return _T("Key 8");
        case 0xB9: return _T("Key 9");
        case 0xBA: return _T("Key Star");
        case 0xBB: return _T("Key Pound");
        case 0xBC: return _T("Key A");
        case 0xBD: return _T("Key B");
        case 0xBE: return _T("Key C");
        case 0xBF: return _T("Key D");
        default:
            break;
        }
        break;
    case 0x0c:
#if 0
        /* Sorry, but this is just too big at 400+ strings!*/
        switch (Usage) {
        case 0x01: return _T("Consumer Control");
        case 0x02: return _T("Numeric Key Pad");
        case 0x03: return _T("Programmable Buttons");
        case 0x04: return _T("Microphone");
        case 0x05: return _T("Headphone");
        case 0x06: return _T("Graphic Equalizer");
        case 0x20: return _T("+10");
        case 0x21: return _T("+100");
        case 0x22: return _T("AM/PM");
        case 0x30: return _T("Power");
        case 0x31: return _T("Reset");
        case 0x32: return _T("Sleep");
        case 0x33: return _T("Sleep After");
        case 0x34: return _T("Sleep Mode");
        case 0x35: return _T("Illumination");
        case 0x36: return _T("Function Buttons");
        case 0x40: return _T("Menu");
        case 0x41: return _T("Menu Pick");
        case 0x42: return _T("Menu Up");
        case 0x43: return _T("Menu Down");
        case 0x44: return _T("Menu Left");
        case 0x45: return _T("Menu Right");
        case 0x46: return _T("Menu Escape");
        case 0x47: return _T("Menu Value Increase");
        case 0x48: return _T("Menu Value Decrease");
        case 0x60: return _T("Data On Screen");
        case 0x61: return _T("Closed Caption");
        case 0x62: return _T("Closed Caption Select");
        case 0x63: return _T("VCR/TV");
        case 0x64: return _T("Broadcast Mode");
        case 0x65: return _T("Snapshot");
        case 0x66: return _T("Still");
        case 0x80: return _T("Selection");
        case 0x81: return _T("Assign Selection");
        case 0x82: return _T("Mode Step");
        case 0x83: return _T("Recall Last");
        case 0x84: return _T("Enter Channel");
        case 0x85: return _T("Order Movie");
        case 0x86: return _T("Channel");
        case 0x87: return _T("Media Selection");
        case 0x88: return _T("Media Select Computer");
        case 0x89: return _T("Media Select TV");
        case 0x8A: return _T("Media Select WWW");
        case 0x8B: return _T("Media Select DVD");
        case 0x8C: return _T("Media Select Telephone");
        case 0x8D: return _T("Media Select Program Guide");
        case 0x8E: return _T("Media Select Video Phone");
        case 0x8F: return _T("Media Select Games");
        case 0x90: return _T("Media Select Messages");
        case 0x91: return _T("Media Select CD");
        case 0x92: return _T("Media Select VCR");
        case 0x93: return _T("Media Select Tuner");
        case 0x94: return _T("Quit");
        case 0x95: return _T("Help");
        case 0x96: return _T("Media Select Tape");
        case 0x97: return _T("Media Select Cable");
        case 0x98: return _T("Media Select Satellite");
        case 0x99: return _T("Media Select Security");
        case 0x9A: return _T("Media Select Home");
        case 0x9B: return _T("Media Select Call");
        case 0x9C: return _T("Channel Increment");
        case 0x9D: return _T("Channel Decrement");
        case 0x9E: return _T("Media Select SAP");
        case 0xA0: return _T("VCR Plus");
        case 0xA1: return _T("Once");
        case 0xA2: return _T("Daily");
        case 0xA3: return _T("Weekly");
        case 0xA4: return _T("Monthly");
        case 0xB0: return _T("Play");
        case 0xB1: return _T("Pause");
        case 0xB2: return _T("Record");
        case 0xB3: return _T("Fast Forward");
        case 0xB4: return _T("Rewind");
        case 0xB5: return _T("Scan Next Track");
        case 0xB6: return _T("Scan Previous Track");
        case 0xB7: return _T("Stop");
        case 0xB8: return _T("Eject");
        case 0xB9: return _T("Random Play");
        case 0xBA: return _T("Select Disc");
        case 0xBB: return _T("Enter Disc");
        case 0xBC: return _T("Repeat");
        case 0xBD: return _T("Tracking");
        case 0xBE: return _T("Track Normal");
        case 0xBF: return _T("Slow Tracking");
        case 0xC0: return _T("Frame Forward");
        case 0xC1: return _T("Frame Back");
        case 0xC2: return _T("Mark");
        case 0xC3: return _T("Clear Mark");
        case 0xC4: return _T("Repeat From Mark");
        case 0xC5: return _T("Return To Mark");
        case 0xC6: return _T("Search Mark Forward");
        case 0xC7: return _T("Search Mark Backwards");
        case 0xC8: return _T("Counter Reset");
        case 0xC9: return _T("Show Counter");
        case 0xCA: return _T("Tracking Increment");
        case 0xCB: return _T("Tracking Decrement");
        case 0xCC: return _T("Stop/Eject");
        case 0xCD: return _T("Play/Pause");
        case 0xCE: return _T("Play/Skip");
        case 0xE0: return _T("Volume");
        case 0xE1: return _T("Balance");
        case 0xE2: return _T("Mute");
        case 0xE3: return _T("Bass");
        case 0xE4: return _T("Treble");
        case 0xE5: return _T("Bass Boost");
        case 0xE6: return _T("Surround Mode");
        case 0xE7: return _T("Loudness");
        case 0xE8: return _T("MPX");
        case 0xE9: return _T("Volume Increment");
        case 0xEA: return _T("Volume Decrement");
        case 0xEB:-return _T("EF Reserved");
        case 0xF0: return _T("Speed Select");
        case 0xF1: return _T("Playback Speed");
        case 0xF2: return _T("Standard Play");
        case 0xF3: return _T("Long Play");
        case 0xF4: return _T("Extended Play");
        case 0xF5: return _T("Slow");
        case 0x100: return _T("Fan Enable");
        case 0x101: return _T("Fan Speed");
        case 0x102: return _T("Light Enable");
        case 0x103: return _T("Light Illumination Level");
        case 0x104: return _T("Climate Control Enable");
        case 0x105: return _T("Room Temperature");
        case 0x106: return _T("Security Enable");
        case 0x107: return _T("Fire Alarm");
        case 0x108: return _T("Police Alarm");
        case 0x109: return _T("Proximity");
        case 0x10A: return _T("Motion");
        case 0x10B: return _T("Duress Alarm");
        case 0x10C: return _T("Holdup Alarm");
        case 0x10D: return _T("Medical Alarm");
        case 0x150: return _T("Balance Right");
        case 0x151: return _T("Balance Left");
        case 0x152: return _T("Bass Increment");
        case 0x153: return _T("Bass Decrement");
        case 0x154: return _T("Treble Increment");
        case 0x155: return _T("Treble Decrement");
        case 0x160: return _T("Speaker System");
        case 0x161: return _T("Channel Left");
        case 0x162: return _T("Channel Right");
        case 0x163: return _T("Channel Center");
        case 0x164: return _T("Channel Front");
        case 0x165: return _T("Channel Center Front");
        case 0x166: return _T("Channel Side");
        case 0x167: return _T("Channel Surround");
        case 0x168: return _T("Channel Low Frequency Enhancement");
        case 0x169: return _T("Channel Top");
        case 0x16A: return _T("Channel Unknown");
        case 0x170: return _T("Sub-channel");
        case 0x171: return _T("Sub-channel Increment");
        case 0x172: return _T("Sub-channel Decrement");
        case 0x173: return _T("Alternate Audio Increment");
        case 0x174: return _T("Alternate Audio Decrement");
        case 0x180: return _T("Application Launch Buttons");
        case 0x181: return _T("AL Launch Button Configuration Tool");
        case 0x182: return _T("AL Programmable Button Configuration");
        case 0x183: return _T("AL Consumer ControlConfiguration");
        case 0x184: return _T("AL Word Processor");
        case 0x185: return _T("AL Text Editor");
        case 0x186: return _T("AL Spreadsheet");
        case 0x187: return _T("AL Graphics Editor");
        case 0x188: return _T("AL Presentation App");
        case 0x189: return _T("AL Database App");
        case 0x18A: return _T("AL Email Reader");
        case 0x18B: return _T("AL Newsreader");
        case 0x18C: return _T("AL Voicemail");
        case 0x18D: return _T("AL Contacts/Address Book");
        case 0x18E: return _T("AL Calendar/Schedule");
        case 0x18F: return _T("AL Task/Project Manager");
        case 0x190: return _T("AL Log/Journal/Timecard");
        case 0x191: return _T("AL Checkbook/Finance");
        case 0x192: return _T("AL Calculator");
        case 0x193: return _T("AL A/V Capture/Playback");
        case 0x194: return _T("AL Local Machine Browser");
        case 0x195: return _T("AL LAN/WAN Browser");
        case 0x196: return _T("AL Internet Browser");
        case 0x197: return _T("AL Remote Networking/ISP Connect");
        case 0x198: return _T("AL Network Conference");
        case 0x199: return _T("AL Network Chat");
        case 0x19A: return _T("AL Telephony/Dialer");
        case 0x19B: return _T("AL Logon");
        case 0x19C: return _T("AL Logoff");
        case 0x19D: return _T("AL Logon/Logoff");
        case 0x19E: return _T("AL Terminal Lock/Screensaver");
        case 0x19F: return _T("AL Control Panel");
        case 0x1A0: return _T("AL Command Line Processor/Run");
        case 0x1A1: return _T("AL Process/Task Manager");
        case 0x1A2: return _T("AL Select Task/Application");
        case 0x1A3: return _T("AL Next Task/Application");
        case 0x1A4: return _T("AL Previous Task/Application");
        case 0x1A5: return _T("AL Preemptive Halt Task/Application");
        case 0x1A6: return _T("AL Integrated Help Center");
        case 0x1A7: return _T("AL Documents");
        case 0x1A8: return _T("AL Thesaurus");
        case 0x1A9: return _T("AL Dictionary");
        case 0x1AA: return _T("AL Desktop");
        case 0x1AB: return _T("AL Spell Check");
        case 0x1AC: return _T("AL Grammar Check");
        case 0x1AD: return _T("AL Wireless Status");
        case 0x1AE: return _T("AL Keyboard Layout");
        case 0x1AF: return _T("AL Virus Protection");
        case 0x1B0: return _T("AL Encryption");
        case 0x1B1: return _T("AL Screen Saver");
        case 0x1B2: return _T("AL Alarms");
        case 0x1B3: return _T("AL Clock");
        case 0x1B4: return _T("AL File Browser");
        case 0x1B5: return _T("AL Power Status");
        case 0x200: return _T("Generic GUI Application Controls");
        case 0x201: return _T("AC New");
        case 0x202: return _T("AC Open");
        case 0x203: return _T("AC Close");
        case 0x204: return _T("AC Exit");
        case 0x205: return _T("AC Maximize");
        case 0x206: return _T("AC Minimize");
        case 0x207: return _T("AC Save");
        case 0x208: return _T("AC Print");
        case 0x209: return _T("AC Properties");
        case 0x21A: return _T("AC Undo");
        case 0x21B: return _T("AC Copy");
        case 0x21C: return _T("AC Cut");
        case 0x21D: return _T("AC Paste");
        case 0x21E: return _T("AC Select All");
        case 0x21F: return _T("AC Find");
        case 0x220: return _T("AC Find and Replace");
        case 0x221: return _T("AC Search");
        case 0x222: return _T("AC Go To");
        case 0x223: return _T("AC Home");
        case 0x224: return _T("AC Back");
        case 0x225: return _T("AC Forward");
        case 0x226: return _T("AC Stop");
        case 0x227: return _T("AC Refresh");
        case 0x228: return _T("AC Previous Link");
        case 0x229: return _T("AC Next Link");
        case 0x22A: return _T("AC Bookmarks");
        case 0x22B: return _T("AC History");
        case 0x22C: return _T("AC Subscriptions");
        case 0x22D: return _T("AC Zoom In");
        case 0x22E: return _T("AC Zoom Out");
        case 0x22F: return _T("AC Zoom");
        case 0x230: return _T("AC Full Screen View");
        case 0x231: return _T("AC Normal View");
        case 0x232: return _T("AC View Toggle");
        case 0x233: return _T("AC Scroll Up");
        case 0x234: return _T("AC Scroll Down");
        case 0x235: return _T("AC Scroll");
        case 0x236: return _T("AC Pan Left");
        case 0x237: return _T("AC Pan Right");
        case 0x238: return _T("AC Pan");
        case 0x239: return _T("AC New Window");
        case 0x23A: return _T("AC Tile Horizontally");
        case 0x23B: return _T("AC Tile Vertically");
        case 0x23C: return _T("AC Format");
        case 0x23D: return _T("AC Edit");
        case 0x23E: return _T("AC Bold");
        case 0x23F: return _T("AC Italics");
        case 0x240: return _T("AC Underline");
        case 0x241: return _T("AC Strikethrough");
        case 0x242: return _T("AC Subscript");
        case 0x243: return _T("AC Superscript");
        case 0x244: return _T("AC All Caps");
        case 0x245: return _T("AC Rotate");
        case 0x246: return _T("AC Resize");
        case 0x247: return _T("AC Flip horizontal");
        case 0x248: return _T("AC Flip Vertical");
        case 0x249: return _T("AC Mirror Horizontal");
        case 0x24A: return _T("AC Mirror Vertical");
        case 0x24B: return _T("AC Font Select");
        case 0x24C: return _T("AC Font Color");
        case 0x24D: return _T("AC Font Size");
        case 0x24E: return _T("AC Justify Left");
        case 0x24F: return _T("AC Justify Center H");
        case 0x250: return _T("AC Justify Right");
        case 0x251: return _T("AC Justify Block H");
        case 0x252: return _T("AC Justify Top");
        case 0x253: return _T("AC Justify Center V");
        case 0x254: return _T("AC Justify Bottom");
        case 0x255: return _T("AC Justify Block V");
        case 0x256: return _T("AC Indent Decrease");
        case 0x257: return _T("AC Indent Increase");
        case 0x258: return _T("AC Numbered List");
        case 0x259: return _T("AC Restart Numbering");
        case 0x25A: return _T("AC Bulleted List");
        case 0x25B: return _T("AC Promote");
        case 0x25C: return _T("AC Demote");
        case 0x25D: return _T("AC Yes");
        case 0x25E: return _T("AC No");
        case 0x25F: return _T("AC Cancel");
        case 0x260: return _T("AC Catalog");
        case 0x261: return _T("AC Buy/Checkout");
        case 0x262: return _T("AC Add to Cart");
        case 0x263: return _T("AC Expand");
        case 0x264: return _T("AC Expand All");
        case 0x265: return _T("AC Collapse");
        case 0x266: return _T("AC Collapse All");
        case 0x267: return _T("AC Print Preview");
        case 0x268: return _T("AC Paste Special");
        case 0x269: return _T("AC Insert Mode");
        case 0x26A: return _T("AC Delete");
        case 0x26B: return _T("AC Lock");
        case 0x26C: return _T("AC Unlock");
        case 0x26D: return _T("AC Protect");
        case 0x26E: return _T("AC Unprotect");
        case 0x26F: return _T("AC Attach Comment");
        case 0x270: return _T("AC Delete Comment");
        case 0x271: return _T("AC View Comment");
        case 0x272: return _T("AC Select Word");
        case 0x273: return _T("AC Select Sentence");
        case 0x274: return _T("AC Select Paragraph");
        case 0x275: return _T("AC Select Column");
        case 0x276: return _T("AC Select Row");
        case 0x277: return _T("AC Select Table");
        case 0x278: return _T("AC Select Object");
        case 0x279: return _T("AC Redo/Repeat");
        case 0x27A: return _T("AC Sort");
        case 0x27B: return _T("AC Sort Ascending");
        case 0x27C: return _T("AC Sort Descending");
        case 0x27D: return _T("AC Filter");
        case 0x27E: return _T("AC Set Clock");
        case 0x27F: return _T("AC View Clock");
        case 0x280: return _T("AC Select Time Zone");
        case 0x281: return _T("AC Edit Time Zones");
        case 0x282: return _T("AC Set Alarm");
        case 0x283: return _T("AC Clear Alarm");
        case 0x284: return _T("AC Snooze Alarm");
        case 0x285: return _T("AC Reset Alarm");
        case 0x286: return _T("AC Synchronize");
        case 0x287: return _T("AC Send/Receive");
        case 0x288: return _T("AC Send To");
        case 0x289: return _T("AC Reply");
        case 0x28A: return _T("AC Reply All");
        case 0x28B: return _T("AC Forward Msg");
        case 0x28C: return _T("AC Send");
        case 0x28D: return _T("AC Attach File");
        case 0x28E: return _T("AC Upload");
        case 0x28F: return _T("AC Download (Save Target As)");
        case 0x290: return _T("AC Set Borders");
        case 0x291: return _T("AC Insert Row");
        case 0x292: return _T("AC Insert Column");
        case 0x293: return _T("AC Insert File");
        case 0x294: return _T("AC Insert Picture");
        case 0x295: return _T("AC Insert Object");
        case 0x296: return _T("AC Insert Symbol");
        case 0x297: return _T("AC Save and Close");
        case 0x298: return _T("AC Rename");
        case 0x299: return _T("AC Merge");
        case 0x29A: return _T("AC Split");
        case 0x29B: return _T("AC Disribute Horizontally");
        case 0x29C: return _T("AC Distribute Vertically");
        default:
            break;
        }
#endif
        break;
    case 0x0d: /* Digitizer */
        switch (Usage) {
        case 0x01: return _T("Digitizer");
        case 0x02: return _T("Pen");
        case 0x03: return _T("Light Pen");
        case 0x04: return _T("Touch Screen");
        case 0x05: return _T("Touch Pad");
        case 0x06: return _T("White Board");
        case 0x07: return _T("Coordinate Measuring Machine");
        case 0x08: return _T("3D Digitizer");
        case 0x09: return _T("Stereo Plotter");
        case 0x0A: return _T("Articulated Arm");
        case 0x0B: return _T("Armature");
        case 0x0C: return _T("Multiple Point Digitizer");
        case 0x0D: return _T("Free Space Wand");
        case 0x20: return _T("Stylus");
        case 0x21: return _T("Puck");
        case 0x22: return _T("Finger");
        case 0x30: return _T("Tip Pressure");
        case 0x31: return _T("Barrel Pressure");
        case 0x32: return _T("In Range");
        case 0x33: return _T("Touch");
        case 0x34: return _T("Untouch");
        case 0x35: return _T("Tap");
        case 0x36: return _T("Quality");
        case 0x37: return _T("Data Valid");
        case 0x38: return _T("Transducer Index");
        case 0x39: return _T("Tablet Function Keys");
        case 0x3A: return _T("Program Change Keys");
        case 0x3B: return _T("Battery Strength");
        case 0x3C: return _T("Invert");
        case 0x3D: return _T("X Tilt");
        case 0x3E: return _T("Y Tilt");
        case 0x3F: return _T("Azimuth");
        case 0x40: return _T("Altitude");
        case 0x41: return _T("Twist");
        case 0x42: return _T("Tip Switch");
        case 0x43: return _T("Secondary Tip Switch");
        case 0x44: return _T("Barrel Switch");
        case 0x45: return _T("Eraser");
        case 0x46: return _T("Tablet Pick");
        default:
            break;
        }
        break;
    case 0x14: /* Alphanum Display */
        switch (Usage) {
        case 0x01: return _T("Alphanumeric Display");
        case 0x20: return _T("Display Attributes Report");
        case 0x21: return _T("ASCII Character Set");
        case 0x22: return _T("Data Read Back");
        case 0x23: return _T("Font Read Back");
        case 0x24: return _T("Display Control Report");
        case 0x25: return _T("Clear Display");
        case 0x26: return _T("Display Enable");
        case 0x27: return _T("Screen Saver Delay");
        case 0x28: return _T("Screen Saver Enable");
        case 0x29: return _T("Vertical Scroll");
        case 0x2A: return _T("Horizontal Scroll");
        case 0x2B: return _T("Character Report");
        case 0x2C: return _T("Display Data");
        case 0x2D: return _T("Display Status");
        case 0x2E: return _T("Stat Not Ready");
        case 0x2F: return _T("Stat Ready");
        case 0x30: return _T("Err Not a loadable character");
        case 0x31: return _T("Err Font data cannot be read");
        case 0x32: return _T("Cursor Position Report");
        case 0x33: return _T("Row");
        case 0x34: return _T("Column");
        case 0x35: return _T("Rows");
        case 0x36: return _T("Columns");
        case 0x37: return _T("Cursor Pixel Positioning");
        case 0x38: return _T("Cursor Mode");
        case 0x39: return _T("Cursor Enable");
        case 0x3A: return _T("Cursor Blink");
        case 0x3B: return _T("Font Report");
        case 0x3C: return _T("Font Data");
        case 0x3D: return _T("Character Width");
        case 0x3E: return _T("Character Height");
        default:
            break;
        }
        break;
    case 0x40: /* Medical Instrument */
        switch (Usage) {
        case 0x01: return _T("Medical Ultrasound");
        case 0x20: return _T("VCR/Acquisition");
        case 0x21: return _T("Freeze/Thaw");
        case 0x22: return _T("Clip Store");
        case 0x23: return _T("Update");
        case 0x24: return _T("Next");
        case 0x25: return _T("Save");
        case 0x26: return _T("Print");
        case 0x27: return _T("Microphone Enable");
        case 0x40: return _T("Cine");
        case 0x41: return _T("Transmit Power");
        case 0x42: return _T("Volume");
        case 0x43: return _T("Focus");
        case 0x44: return _T("Depth");
        case 0x60: return _T("Soft Step - Primary");
        case 0x61: return _T("Soft Step - Secondary");
        case 0x70: return _T("Depth Gain Compensation");
        case 0x80: return _T("Zoom Select");
        case 0x81: return _T("Zoom Adjust");
        case 0x82: return _T("Spectral Doppler Mode Select");
        case 0x83: return _T("Spectral Doppler Adjust");
        case 0x84: return _T("Color Doppler Mode Select");
        case 0x85: return _T("Color Doppler Adjust");
        case 0x86: return _T("Motion Mode Select");
        case 0x87: return _T("Motion Mode Adjust");
        case 0x88: return _T("2-D Mode Select");
        case 0x89: return _T("2-D Mode Adjust");
        case 0xA0: return _T("Soft Control Select");
        default:
            break;
        }
        break;
    default:
        break;
    }
    _stprintf(AttributesBuffer,_T("x%x"),Usage);
    return AttributesBuffer;
}

PRIVATE _TCHAR *CollectionName(INT32 Collection)
{
    switch (Collection) {
    case 0x00: return _T("Physical");
    case 0x01: return _T("Application");
    case 0x02: return _T("Logical");
    case 0x03: return _T("Report");
    case 0x04: return _T("Named Array");
    case 0x05: return _T("Usage Switch");
    case 0x06: return _T("Usage Modifier");
    default:
        break; 
    }
    _stprintf(AttributesBuffer,_T("x%x"),Collection);
    return AttributesBuffer;
}

PRIVATE _TCHAR *ParseIof(INT32 Format)
{
    _stprintf(AttributesBuffer,_T("%s, "),
              (Format & 0x1) ? _T("Constant") : _T("Data"));
    _tcscat(AttributesBuffer,
              (Format & 0x2) ? _T("Var, ") : _T("Array, "));
    _tcscat(AttributesBuffer,
              (Format & 0x4) ? _T("Rel") : _T("Abs"));
    if (Format & 0x8)
        _tcscat(AttributesBuffer,_T(", Wrap"));
    if (Format & 0x10)
        _tcscat(AttributesBuffer,_T(", -Lin"));
    if (Format & 0x20)
        _tcscat(AttributesBuffer,_T(", -Pref"));
    if (Format & 0x40)
        _tcscat(AttributesBuffer,_T(", -0State"));
    if (Format & 0x80)
        _tcscat(AttributesBuffer,_T(", Volat"));
    if (Format & 0x100)
        _tcscat(AttributesBuffer,_T(", BufBytes"));
    return AttributesBuffer;
}

PRIVATE void ParseHidReportDescriptor(UINT8 *Descriptor, UINT16 Length)
{
    int i, Index, Indentation = 0;
    int InSize, OutSize, FeaSize;
    int Rc, Rs;
    int UsagePage = 0;
    UINT8 c, size;
    INT32 v;

    _tprintf(_T("Report Descriptor:\n"));

    FeaSize = InSize = OutSize = 0;
    Rc = Rs = 0;
    for (Index = 0; Index < Length; Index++)
    {
        c = Descriptor[Index];

        /* Long item?
         */
        if (c == 0xfe) {
            /* Yes, but spec-undefined at this point
             */
            size = Descriptor[Index+1];
            _tprintf(_T("\tLongItem Tag %x Data"),Descriptor[Index+2]);
            for (i = 0; i < size; i++)
                _tprintf(_T(" %x"),Descriptor[Index+4+i]);
            goto Next;
        }

        /* Short Item 
         */
        size = c & 3;
        if (size == 3)
            size++;

        v = pn(Descriptor+Index+1,size);

        _tprintf(_T("\t"));
        for (i = 0; i < Indentation; i++)
            _tprintf(_T(" "));

#if _DEBUG
        _tprintf(_T("%x "),c);
        if (size)
            _tprintf(_T("%d "),v);
#endif

        switch (c & 0xfc) {
            /* Global Items
             */
        case 0x04:
            UsagePage = v;
            _tprintf(_T("USAGE_PAGE (%s)"),UsagePageName(v));
            break;
        case 0x14:
            _tprintf(_T("LOGICAL_MINIMUM (%d)"),v);
            break;
        case 0x24:
            _tprintf(_T("LOGICAL_MAXIMUM (%d)"),v);
            break;
        case 0x34:
            _tprintf(_T("PHYSICAL_MINIMUM (%d)"),v);
            break;
        case 0x44:
            _tprintf(_T("PHYSICAL_MAXIMUM (%d)"),v);
            break;
        case 0x54:
            _tprintf(_T("UNIT_EXPONENT (%d)"),v);
            break;
        case 0x64:
            _tprintf(_T("UNIT (x%x)"),v);
            break;
        case 0x74:
            _tprintf(_T("REPORT_SIZE (%d bits)"),v);
            break;
        case 0x84:
            _tprintf(_T("REPORT_ID (%u)"),v);
            break;
        case 0x94:
            _tprintf(_T("REPORT_COUNT (%u)"),v);
            break;
        case 0xa4:
            _tprintf(_T("PUSH"));
            break;
        case 0xb4:
            _tprintf(_T("POP"));
            break;

            /* Local Items
             */
        case 0x08:
            _tprintf(_T("USAGE (%s)"),UsageName(UsagePage,v));
            break;
        case 0x18:
            _tprintf(_T("USAGE_MINIMUM (%d)"),v);
            break;
        case 0x28:
            _tprintf(_T("USAGE_MAXIMUM (%d)"),v);
            break;
        case 0x38:
            _tprintf(_T("DESIGNATOR_INDEX (%d)"),v);
            break;
        case 0x48:
            _tprintf(_T("DESIGNATOR_MINIMUM (%d)"),v);
            break;
        case 0x58:
            _tprintf(_T("DESIGNATOR_MAXIMUM (%d)"),v);
            break;
        case 0x78:
            _tprintf(_T("STRING_INDEX (%d)"),v);
            break;
        case 0x88:
            _tprintf(_T("STRING_MINIMUM (%d)"),v);
            break;
        case 0x98:
            _tprintf(_T("STRING_MAXIMUM (%d)"),v);
            break;
        case 0xa8:
            _tprintf(_T("DELIMITER"));
            break;

            /* Main Items
             */
        case 0x80:
            _tprintf(_T("INPUT (%s)"),ParseIof(v));
            break;
        case 0x90:
            _tprintf(_T("OUTPUT (%s)"),ParseIof(v));
            break;
        case 0xb0:
            _tprintf(_T("FEATURE (%s)"),ParseIof(v));
            break;
        case 0xa0:
            _tprintf(_T("COLLECTION (%s)"),CollectionName(v));
            Indentation++;
            break;
        case 0xc0:
            _tprintf(_T("END_COLLECTION"));
            Indentation--;
            break;

        default:
            _tprintf(_T("unknown?"));
            break;
        }

        Index += size;
        if (c == 0x75)
            Rs = Descriptor[Index];
        if (c == 0x95)
            Rc = Descriptor[Index];
        if (c == 0x81)
            InSize += (Rs * Rc);
        if (c == 0x91)
            OutSize += (Rs * Rc);
        if (c == 0xb1)
            FeaSize += (Rs * Rc);
    Next:
        _tprintf(_T("\n"));
    }
    _tprintf(_T("\t--Input Report Length %d bits (%d bytes)\n"), InSize,InSize/8);
    _tprintf(_T("\t--Output Report Length %d bits (%d bytes)\n"), OutSize,OutSize/8);
    _tprintf(_T("\t--Feature Report Length %d bits (%d bytes)\n"),FeaSize,FeaSize/8);
}

PRIVATE SCODE doRep(
    PIUSBHUB Hub,
    UINT8 UsbAddress,
    UINT8 Port,
    UINT8 Interface,
    UINT16 RepLen)
{
    UINT8 *Descriptor;
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
    CONDITION Done;

    Condition_Init(&Done);

    Descriptor = (UINT8 *) Buffer;

    DBGME(0,_tprintf(_T("doRep(%d)\n"),UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &Done;
    Urb.DeviceAddress = UsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb.SetupData[0] = USB_RT_DTH|USB_RT_IFACE;
    Urb.SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb.SetupData[3] = USB_DT_REPORT;
    Urb.SetupData[4] = Interface;
    Urb.SetupData[6] = (UINT8) RepLen; /* BUGBUG */
    Urb.SetupLength = 8;
    Urb.ExpectedDataLength = (UINT8) RepLen; /* BUGBUG */
    Urb.Data = (UINT8 *)Descriptor;
    Urb.MaximumPacketSize = 8; /* BUGBUG */

    sc = Hub->v->SubmitUrb(Hub,
                           &Urb,
                           Port);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("doRep: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);

#if _DEBUG
    PrintUrbResults(&Urb,_T("Report Descriptor:"));
#endif
    ParseHidReportDescriptor(Descriptor,Urb.ActualDataLength);
    return S_OK;
}

PRIVATE SCODE doDev(PIUSBHUB Hub, PINAMESPACE Ns, _TCHAR *Name)
{
    UINT8 UsbAddress, Port;
    UINT UsbAddress32, Port32;
    PIUSBDRIVER Driver;
    PIUSBHUB Hub2;
    SCODE sc;

    _tprintf(_T("\nUsbDevice: '%s'\n"),Name);

    sc = BindToObject(Ns,
                      Name,
                      NAME_SPACE_READ,
                      &IID_IUsbDriver,
                      (void **)&Driver);
    if (FAILED(sc)) {
        _tprintf(_T("%s: cant find or not a UsbDriver, sc=x%x\n"),Name,sc);
        return sc;
    }

    sc = Driver->v->GetUsbAddress(Driver,&UsbAddress32,&Port32);
    if (FAILED(sc)) {
        _tprintf(_T("%s has no UsbAddress, sc=x%x\n"),Name,sc);
        return sc;
    }
    UsbAddress = (UINT8) UsbAddress32;
    Port       = (UINT8) Port32;

    DBGME(0,_tprintf(_T("doDev(%s) @UsbAddress %d\n"),Name,UsbAddress));

    ClearAllStringReferences();
    sc = doDesc(Hub,UsbAddress,Port);
    sc = doConf(Hub,UsbAddress,Port);
    /*...more as needed;
     */
    sc = doStr(Hub,UsbAddress,Port);

    sc = Driver->v->QueryInterface(Driver,
                                   &IID_IUsbHub,
                                   (void **) &Hub2);
    Driver->v->Release(Driver);
    if (!FAILED(sc) && Hub2) {
        Hub2->v->Release(Hub2);
        return S_OK;
    }
    return S_NO_MORE_ENTRIES;
}

typedef struct _WORK {
    struct _WORK *Next;
    PINAMESPACE Ns;
    PIUSBHUB Hub;
    _TCHAR *Name;
} WORK;

PRIVATE WORK *MoreWork(PIUSBHUB Hub, PINAMESPACE Ns, _TCHAR *Name)
{
    WORK *Item = malloc(sizeof *Item);

    DBGME(0,_tprintf(_T("MoreWork for %s\n"),Name));

    Item->Next = NULL;
    Item->Name = _tcsdup(Name);
    Item->Ns = Ns;
    Item->Hub = Hub;
    Hub->v->AddRef(Hub);
    Ns->v->AddRef(Ns);
    return Item;
}

PRIVATE void LessWork(WORK *Item)
{
    DBGME(0,_tprintf(_T("done with %s\n"),Item->Name));
    Item->Hub->v->Release(Item->Hub);
    Item->Ns->v->Release(Item->Ns);
    free(Item->Name);
    free(Item);
}
		
int _tmain(int argc, _TCHAR **argv)
{
    WORK *MoreDevs, *Item;
    SCODE sc;
    UINT Length;
    PINAMESPACE pNs = NULL;
    PIUSBHUB Hub = NULL;

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_TEXT("No command line args (%x)\n\r"), sc);
        return -1;
    }

    /* There must be an argument, the HUB where to start from..
     */
    if (argc < 2) {
        _tprintf(_TEXT("Usage:%s UsbHubName\n\r"), argv[0]);
        goto Out;
    }

    /* And it must be a HUB..
     */
    sc = BindToObject(CurrentNameSpace(),
                      argv[1],
                      NAME_SPACE_READ,
                      &IID_IUsbHub,
                      (void **)&Hub);
    if (FAILED(sc)) {
        _tprintf(_TEXT("%s is not an IUsbHub (%x)\n\r"), argv[1], sc);
        goto Out;
    }

    /* ..Who exports a namespace
     */
    sc = Hub->v->QueryInterface(Hub,
                                &IID_INameSpace,
                                (void **) &pNs);


    if (FAILED(sc)) {
        _tprintf(_TEXT("%s does not implement INameSpace (%x)\n\r"), argv[1],sc);
        goto Out;
    }
    pNs->v->Release(pNs);
    pNs = NULL;

    MoreDevs = MoreWork(Hub, CurrentNameSpace(), argv[1]);
    Hub->v->Release(Hub);
    Hub = NULL;
    do {
        sc = doDev( MoreDevs->Hub, MoreDevs->Ns, MoreDevs->Name);

        /* Is this a HUB ?
         */
        if (sc == S_OK) {
            /* Yes, get it
             */
            DBGME(0,_tprintf(_T("%s is a Hub!\n"),MoreDevs->Name));
            sc = BindToObject(MoreDevs->Ns,
                              MoreDevs->Name,
                              NAME_SPACE_READ,
                              &IID_IUsbHub,
                              (void **)&Hub);
            if (FAILED(sc)) {
                _tprintf(_TEXT("%s is confused (%x)\n\r"), MoreDevs->Name, sc);
                goto Next;
            }

            /* It must export a namespace
             */
            sc = Hub->v->QueryInterface(Hub,
                                        &IID_INameSpace,
                                        (void **) &pNs);

            if (FAILED(sc)) {
                _tprintf(_TEXT("%s is soo confused (%x)\n\r"), MoreDevs->Name, sc);
                Hub->v->Release(Hub);
                Hub = NULL;
                goto Next;
            }

            /* Get started with the listing
             */
            Length = 0;
            sc = pNs->v->FindNext(pNs,
                                 _T(""),
                                 NULL,
                                 Buffer,
                                 MAX_PATH-sizeof(_TCHAR),
                                 &Length);

            Buffer[Length] = 0;
            DBGME(0,_tprintf(_T("FindFirst(sc=%x len=%d) gets '%s'\n"),
                             sc, Length, Buffer));

            /* Loop while we find some
             */
            while (!FAILED(sc) && (Length > 0)) {
			
                /* Say what
                 */
				Item = MoreWork(Hub,pNs,Buffer);
				Item->Next = MoreDevs->Next;
				MoreDevs->Next = Item;


                /* Next ?
                 */
                Length = 0;
                sc = pNs->v->FindNext(pNs,
                                     _T(""),
                                     Buffer,
                                     Buffer,
                                     MAX_PATH-sizeof(_TCHAR),
                                     &Length);
                Buffer[Length] = 0;
                DBGME(0,_tprintf(_T("FindNext(sc=%x len=%d) gets '%s'\n"),
                             sc, Length, Buffer));
            }
    
            /* Done here
             */				
            Hub->v->Release(Hub);
            Hub = NULL;
            pNs->v->Release(pNs);
            pNs = NULL;
        }

    Next:
        Item = MoreDevs;
        MoreDevs = MoreDevs->Next;
        LessWork(Item);
    } while (MoreDevs != NULL);

 Out:
    (void) FreeCmdLine(&argc,&argv);
    if (Hub)
        Hub->v->Release(Hub);
    if (pNs)
        pNs->v->Release(pNs);
    return 0;
}
