/* Copyright (c) Microsoft Corporation. All rights reserved. */
/* --------------------------------------------------------------------------
 *
 * Module:
 *
 *    usbiso.c
 *
 * Purpose:
 *
 *    Utility to test I/O from isochronous endpoints
 *
 * Author:
 *    A. Forin (sandrof)
 *
 * --------------------------------------------------------------------------
 */

#include <mmlite.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tchar.h>
#define _DEBUG 1
#define DBGLVL 0
#include <assert.h>
#include <diagnostics.h>
#include <drivers\usb.h>
#include <usbmaster.h>

/* BUGBUG This needs to be added to :GetUsbAddress()
 */
#define MAXCTLPKTSIZE 64

typedef struct {
    UINT8   EndpointAddress;
    UINT8   Attributes;
    UINT16  MaxPacketSize;
    UINT8   Interval;
} MYENDP, *PMYENDP;

typedef struct {
    _TCHAR *HubName;
    _TCHAR *DeviceName;
    UINT32  nLoops;
    UINT8   UsbAddress;
    UINT8   Port;
    UINT8   Mps;
    UINT8   Configuration;

    UINT8   Interface;
    UINT8   Setting;
    UINT8   InputEndpoint;
    /* This is for one config+iface+setting */
    UINT8   NumEndpoints;
#define MaxEndpoints 4
    MYENDP  Endpoint[MaxEndpoints];
} DEVARGS, *PDEVARGS;

#define LINES    480
#define COLUMNS  640
#define MAX_PATH (LINES*COLUMNS)
UINT8 DataBuffer[MAX_PATH];

#define ONE_BUFFERFULL (COLUMNS * 10)

#define NO_DATA_IN_BUFFER 0xe

BOOL CheckBuffer(void)
{
    int i;
    for (i = 0; i < ONE_BUFFERFULL; i++)
        if (DataBuffer[i] != NO_DATA_IN_BUFFER)
            return TRUE;
    return FALSE;
}

/* Utilities
 */
#if _DEBUG
PRIVATE void PrintUrbResults(PUSBM_REQUEST_BLOCK pUrb, _TCHAR *Banner)
{
    int i, n;

    n = pUrb->ActualDataLength;
    if (n > pUrb->ExpectedDataLength) {
        _tprintf(_T("ERROR: actual %d > expected %d!\n"),
                 n, pUrb->ExpectedDataLength);
        n = pUrb->ExpectedDataLength;
    }
    _tprintf(Banner);
    for (i = 0; i < n; i++) {
        _tprintf(_T("%2x "), pUrb->Data[i]);
        if (n > 0x1000) {
            _tprintf(_T("..."));
            break;
        }
    }
    _tprintf(_T("\n"));
}

PRIVATE void PrintConfigurationInformation(PDEVARGS Args)
{
    int i;
    _tprintf(_T("\nConfiguration used for %s (hub %s):\n"),
             Args->DeviceName, Args->HubName);
    _tprintf(_T("\tUsbAddress %d\n"), Args->UsbAddress);
    _tprintf(_T("\tPort %d\n"), Args->Port);
    _tprintf(_T("\tMps %d\n"), Args->Mps);
    _tprintf(_T("\tConfiguration %d\n"), Args->Configuration);
    _tprintf(_T("\tInterface %d\n"), Args->Interface);
    _tprintf(_T("\tSetting %d\n"), Args->Setting);
    _tprintf(_T("\tnLoops %d\n"), Args->nLoops);
    _tprintf(_T("\tNumEndpoints %d\n"), Args->NumEndpoints);

    for (i = 0; i < Args->NumEndpoints; i++) {
        _tprintf(_T("\t[%2d] EndpointAddress %x\n"), i,
                 Args->Endpoint[i].EndpointAddress);
        _tprintf(_T("\t      Attributes %x\n"), Args->Endpoint[i].Attributes);
        _tprintf(_T("\t      MaxPacketSize %d\n"), Args->Endpoint[i].MaxPacketSize);
        _tprintf(_T("\t      Interval %d\n"), Args->Endpoint[i].Interval);
    }
}
#endif


PRIVATE SCODE VerifyTheMps(PIUSBHUB Hub, PDEVARGS Args)
{
    PUSB_DEVICE_struct Descriptor;
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
    CONDITION Done;

    Condition_Init(&Done);

#define maxdesc 64
    Descriptor = (PUSB_DEVICE_struct) DataBuffer;

    DBGME(0,_tprintf(_T("VerifyTheMps(%d)\n"),Args->UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &Done;
    Urb.DeviceAddress = Args->UsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb.SetupData[0] = USB_RT_DTH;
    Urb.SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb.SetupData[3] = USB_DT_DEVICE;
    Urb.SetupData[6] = maxdesc;
    Urb.SetupLength = 8;
    Urb.ExpectedDataLength = maxdesc;
    Urb.Data = (UINT8 *)Descriptor;
    Urb.MaximumPacketSize = Args->Mps;

    sc = Hub->v->SubmitUrb(Hub,
                           &Urb,
                           Args->Port);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("VerifyTheMps: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    sc = Condition_Wait(Urb.Done,NULL,NULL);

    if (FAILED(sc))
        return sc;

    if ((Urb.Flags & URB_FLAG_ERROR) ||
        (Urb.ActualDataLength < sizeof *Descriptor))
        return E_FAIL;

    DBGME(0,PrintUrbResults(&Urb,_T("Device Descriptor:")));

    /* Adjust the control Mps if necessary */
    if (Args->Mps != Descriptor->MaxPacketSize0) {
        DBGME(2,_tprintf(_T("usbiso: Mps changed to %d (was %d)\n"),
                         Descriptor->MaxPacketSize0, Args->Mps));
        Args->Mps = Descriptor->MaxPacketSize0;
    }

#if 0
    /* And check we are talking to the right camera */
    if ((Get2(Descriptor->IdVendor) != 0xVVVV) ||
        (Get2(Descriptor->IdProduct) != 0xPPPP)) {
        _tprintf(_T("usbiso: Vendor x%x Product x%x wont go too far.\n"),
                 Get2(Descriptor->IdVendor),
                 Get2(Descriptor->IdProduct) );
    }
#endif

    return S_OK;
#undef maxdesc
}

PRIVATE SCODE GetConfigurationInformation(
    PIUSBHUB Hub, 
    PDEVARGS Args)
{
    int i;
    PUSB_CONFIGURATION_struct Descriptor;
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
    CONDITION Done;
    int MatchLevel = 0;
    int NextEndpoint = 0;

    Condition_Init(&Done);

#define maxdesc 512
    assert(maxdesc <= MAX_PATH);
    Descriptor = (PUSB_CONFIGURATION_struct) DataBuffer;

    DBGME(0,_tprintf(_T("GetConfigurationInformation(%d)\n"),Args->UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &Done;
    Urb.DeviceAddress = Args->UsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_IN|URB_FLAG_CONTROL;
    Urb.SetupData[0] = USB_RT_DTH;
    Urb.SetupData[1] = USB_R_GET_DESCRIPTOR;
    Urb.SetupData[3] = USB_DT_CONFIG;
    Set2(&Urb.SetupData[6],maxdesc);
    Urb.SetupLength = 8;
    Urb.ExpectedDataLength = maxdesc;
    Urb.Data = (UINT8 *)Descriptor;
    Urb.MaximumPacketSize = Args->Mps;

    sc = Hub->v->SubmitUrb(Hub,
                           &Urb,
                           Args->Port);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("GetConfigurationInformationDesc: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    Condition_Wait(Urb.Done,NULL,NULL);

    DBGME(0,PrintUrbResults(&Urb,_T("Configuration Descriptor:")));

    /* It *must* be a config one we got first.
     */
    if ((Descriptor->Length != sizeof(USB_CONFIGURATION_DESCRIPTOR)) ||
        (Descriptor->DescriptorType != USB_DT_CONFIG))
        goto Bad;

    for (i = 0; i < Urb.ActualDataLength;) {

        Descriptor = (PUSB_CONFIGURATION_struct)(Urb.Data+i);

        /* If the correct configuration look further into it, else skip
         */
        if (Descriptor->DescriptorType == USB_DT_CONFIG) {
            if (Descriptor->ConfigurationValue == Args->Configuration)
                MatchLevel = 1;
            else
                MatchLevel = 0;
        }

        /* Skip till the next configuration if we have to
         */
        if (MatchLevel < 1)
            goto Next;

        /* If an interface check the number
         */
        if (Descriptor->DescriptorType == USB_DT_IFACE) {
            PUSB_INTERFACE_struct Iface = (PUSB_INTERFACE_struct)Descriptor;

            if ((Iface->InterfaceNumber == Args->Interface) &&
                (Iface->AlternateSetting == Args->Setting)) {

                /* Trim of we must, warn its an issue */
                Args->NumEndpoints = Iface->NumEndpoints;
                if (Args->NumEndpoints > MaxEndpoints) {
                    DBGME(2,_tprintf(_T("usbiso: Ep# trimmed to %d (was %d)\n"),
                                     MaxEndpoints, Args->NumEndpoints));
                    Args->NumEndpoints = MaxEndpoints;
                    NextEndpoint = 0;
                }

                MatchLevel = 2;
            } else {
                MatchLevel = 1;
            }
        }

        /* Skip till the next interface if we have to
         */
        if (MatchLevel < 2)
            goto Next;

        if (Descriptor->DescriptorType == USB_DT_ENDP) {
            PUSB_ENDPOINT_struct Endp = (PUSB_ENDPOINT_struct)Descriptor;

            Args->Endpoint[NextEndpoint].EndpointAddress = Endp->EndpointAddress;
            Args->Endpoint[NextEndpoint].Attributes = Endp->Attributes;
            Args->Endpoint[NextEndpoint].MaxPacketSize = Get2(Endp->MaxPacketSize);
            Args->Endpoint[NextEndpoint].Interval = Endp->Interval;
            NextEndpoint++;

            /* Well, are we done yet?
             */
            if (NextEndpoint == Args->NumEndpoints) {
                MatchLevel = 3;
                break;
            }
        }

        /* Do not crash on screwy data */
    Next:
        i += (Descriptor->Length) ? Descriptor->Length : Urb.ActualDataLength;
    }

    if (MatchLevel == 3)
        return S_OK;

 Bad:
    return E_FAIL;
#undef maxdesc
}

/* This picks both the config and the interface
 */
PRIVATE SCODE SelectConfiguration(
    PIUSBHUB Hub, 
    PDEVARGS Args)
{
    struct USBM_REQUEST_BLOCK Urb;
    SCODE sc;
    CONDITION Done;

    Condition_Init(&Done);

    DBGME(0,_tprintf(_T("SelectConfiguration(%d)\n"),Args->UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &Done;
    Urb.DeviceAddress = Args->UsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_CONTROL;
    Urb.SetupData[1] = USB_R_SET_CONFIG;
    Urb.SetupData[2] = Args->Configuration;
    Urb.SetupLength = 8;
    Urb.MaximumPacketSize = Args->Mps;

    sc = Hub->v->SubmitUrb(Hub,
                           &Urb,
                           Args->Port);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("SelectConfigurationDesc: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    sc = Condition_Wait(Urb.Done,NULL,NULL);

    if (FAILED(sc))
        return sc;
    if ((Urb.Flags & URB_FLAG_ERROR))
        return E_FAIL;

    DBGME(0,_tprintf(_T("SelectConfiguration(%d) ok.\n"),Args->Configuration));

    /* Now set the interface
     */
    DBGME(0,_tprintf(_T("SelectInterface(%d)\n"),Args->UsbAddress));

    memset(&Urb,0,sizeof Urb);
    Urb.Done = &Done;
    Urb.DeviceAddress = Args->UsbAddress;
    Urb.Pid = USB_PID_SETUP;
    Urb.Flags = URB_FLAG_CONTROL;
    Urb.SetupData[0] = USB_RT_IFACE;
    Urb.SetupData[1] = USB_R_SET_IFACE;
    Urb.SetupData[2] = Args->Setting;
    Urb.SetupData[4] = Args->Interface;
    Urb.SetupLength = 8;
    Urb.MaximumPacketSize = Args->Mps;

    sc = Hub->v->SubmitUrb(Hub,
                           &Urb,
                           Args->Port);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("SelectInterfaceDesc: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    sc = Condition_Wait(Urb.Done,NULL,NULL);

    if (FAILED(sc))
        return sc;
    if ((Urb.Flags & URB_FLAG_ERROR))
        return E_FAIL;

    DBGME(0,_tprintf(_T("SelectInterface(%d,%d) ok.\n"),
                     Args->Interface, Args->Setting));

    return S_OK;
}

/* Reset the input iso pipe
 */
PRIVATE SCODE ResetPipe(
    PIUSBHUB Hub, 
    PDEVARGS Args,
    PUSBM_REQUEST_BLOCK Urb)
{
    MYENDP *Endp;
    SCODE sc;
    
    DBGME(0,_tprintf(_T("ResetPipe(%d)\n"),Args->UsbAddress));

    /* Pick up the endpoint
     */
    Endp = &Args->Endpoint[Args->InputEndpoint];

    /* Reset the pipe
     */
    Urb->DeviceAddress = Args->UsbAddress;
    Urb->EndpointAddress = 0;
    Urb->Pid = USB_PID_SETUP;
    Urb->Flags = URB_FLAG_CONTROL;
    Urb->SetupData[0] = USB_RT_ENDP;
    Urb->SetupData[1] = USB_R_CLEAR_FEATURE;
    Set2(&Urb->SetupData[2],USB_FS_ENDP_STALL);
    Urb->SetupData[4] = Endp->EndpointAddress;
    Urb->SetupLength = 8;
    Urb->MaximumPacketSize = Args->Mps;

    sc = Hub->v->SubmitUrb(Hub,
                           Urb,
                           Args->Port);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("ResetPipe: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    sc = Condition_Wait(Urb->Done,NULL,NULL);

    if (FAILED(sc))
        return sc;
    if ((Urb->Flags & URB_FLAG_ERROR))
        return E_FAIL;

    DBGME(0,_tprintf(_T("ResetPipe(%x) ok.\n"), Endp->EndpointAddress));

    return S_OK;
}

PRIVATE SCODE InitializeCamera(PIUSBHUB Hub, PDEVARGS Args)
{
    SCODE sc;

    DBGME(0,_tprintf(_T("InitializeCamera(%d)\n"),Args->UsbAddress));

    /* Add here vendor-specific commands to get the camera started.
     */
    sc = S_OK;

    DBGME(1,_tprintf(_T("InitializeCamera(%d) ok.\n"),Args->UsbAddress));

    return sc;
}

PRIVATE SCODE FinalizeCamera(PIUSBHUB Hub, PDEVARGS Args)
{
    SCODE sc;

    DBGME(0,_tprintf(_T("FinalizeCamera(%d)\n"),Args->UsbAddress));

    /* Add here vendor-specific commands to stop the camera.
     */
    sc = S_OK;

    DBGME(1,_tprintf(_T("FinalizeCamera(%d) ok.\n"),Args->UsbAddress));

    return sc;
}

/* State for URBs that are posted.
 */
#define nURB 16
typedef struct _BUFFER_STATE {
    struct USBM_REQUEST_BLOCK Urb;
    UINT16                    BytesAvailable;
} BUFFER_STATE, *PBUFFER_STATE;

static BUFFER_STATE ActiveUrbs[nURB];

/* This function is called for each descriptor that is completed. 
 * It should synchronize to frame boundaries and display frames
 * that are complete.
 */
PRIVATE void CheckFrame( UINT8 Buffer)
{
    PBUFFER_STATE Bs;
    UINT8 *Data;
    static int Offset = 0;

    Bs = &ActiveUrbs[Buffer & (nURB-1)];

    DBGME(1,_tprintf(_T("CheckFrame: Buffer[%d] %d bytes\r"),
                     Buffer, Bs->BytesAvailable));

    /* Check if we have enough data
     * NB: A camera might fragment one scanline in various ways
     */
#define BYTES_PER_PACKET 256
#define PACKETS_PER_FRAME 300
#define FRAMEBUFFER_LINESIZE 768
    if (Bs->BytesAvailable >= BYTES_PER_PACKET) {
        Data = Bs->Urb.Data;

        DBGME(0,PrintUrbResults(&Bs->Urb,_T("Content:")));
            
#if 0
        /* Enable this to view the frame */
#define FrameBuffer ((UINT8*)0xb0008000)
        memcpy(FrameBuffer+Offset,Data,BYTES_PER_LINE);
#endif
        Offset += FRAMEBUFFER_LINESIZE;
            if (Offset > (FRAMEBUFFER_LINESIZE*PACKETS_PER_FRAME))
                Offset = 0;
    }
    else {
        /* Live the wrong count on the screen */
        DBGME(1,_tprintf(_T("\n")));
    }
}

PRIVATE SCODE FrameTest(
    PIUSBHUB Hub, 
    PDEVARGS Args)
{
    PBUFFER_STATE Bs;
    PUSBM_REQUEST_BLOCK Urb;
    SCODE sc;
    MYENDP *Endp;
    UINT8 *Data;
    UINT8 NextMsec;
    UINT8 LastBuf;
    CONDITION Done;
    UINT32 i;
    UINT16 Mps;

    Condition_Init(&Done);
    memset(DataBuffer,NO_DATA_IN_BUFFER,sizeof DataBuffer);

    /* Pick up the endpoint
     */
    Endp = &Args->Endpoint[Args->InputEndpoint];
    Mps  = Endp->MaxPacketSize & USB_MP_MASK;

    /* Reset the pipe
     */
    Urb = &ActiveUrbs[0].Urb;
    memset(Urb,0,sizeof *Urb);
    Urb->Done = &Done;
    sc = ResetPipe(Hub,Args,Urb);
    if (FAILED(sc))
        return sc;

    /* Init the unchanging parts
     */
    DBGME(0,_tprintf(_T("FrameTest(%d) DataBuffer @%x\n"),
                     Args->UsbAddress,(ADDRESS)DataBuffer));

    Bs = &ActiveUrbs[0];
    Data = DataBuffer;
    assert(MAX_PATH >= (nURB * Mps));

    for (i = 0; i < nURB; i++) {
        memset(Bs,0,sizeof *Bs);
        Urb = &Bs->Urb;
        Urb->Done = &Done;
        Urb->DeviceAddress = Args->UsbAddress;
        Urb->EndpointAddress = Endp->EndpointAddress & USB_EA_ADDRESS_MASK;
        Urb->Pid = USB_PID_IN;
        Urb->Flags = URB_FLAG_IN | URB_FLAG_ISOCHRONOUS;
        Urb->SetupLength = 0;
        Urb->MaximumPacketSize = Mps;
        Urb->ExpectedDataLength = Mps;
        Urb->Data = Data;

        Data += Mps;
        Bs++;
    }

    /* Post the first URB, this tells us where we are in the USB schedule
     */
    Bs = &ActiveUrbs[0];
    Urb = &Bs->Urb;
    Urb->Interval = 255;
    sc = Hub->v->SubmitUrb(Hub,
                           Urb,
                           Args->Port);
    if (FAILED(sc)) {
        DBGME(3,_tprintf(_T("FrameTest0: SubmitUrb failed sc=x%x\n"),sc));
        return sc;
    }

    NextMsec = Urb->Interval + 1;

    /* Post the remaining URBs
     */
    Bs++;
    for (i = 1; i < nURB; i++) {
        Urb = &Bs->Urb;
        Urb->Interval = (NextMsec++) & (nURB-1);
        sc = Hub->v->SubmitUrb(Hub,
                               Urb,
                               Args->Port);
        if (FAILED(sc)) {
            DBGME(3,_tprintf(_T("FrameTest1: SubmitUrb failed sc=x%x\n"),sc));
            return sc;
        }
        Bs++;
    }

    /* Now go steady-state:
     * - Wait until at leat one URB is complete
     * - Repost it with the updated NextMsec
     * - Check if the input makes any sense
     */
    Bs = &ActiveUrbs[0];
    Urb = &Bs->Urb;
    LastBuf = 0;
    for (i = 0; i < Args->nLoops; i++) {

        sc = Condition_Wait(&Done,NULL,NULL);
        if (FAILED(sc))
            return sc;

        while (Urb->Flags & URB_FLAG_COMPLETE) {

            if (Urb->Flags & URB_FLAG_ERROR) {
                DBGME(3,_tprintf(_T("\nFrameTest: fs=x%x\n"),Urb->FinalStatus));
                return E_FAIL;
            }

            Bs->BytesAvailable = Urb->ActualDataLength;

            Urb->Flags = URB_FLAG_IN | URB_FLAG_ISOCHRONOUS;
            Urb->Interval = (NextMsec++) & (nURB-1);
            sc = Hub->v->SubmitUrb(Hub,
                                   Urb,
                                   Args->Port);
            if (FAILED(sc)) {
                DBGME(3,_tprintf(_T("\nFrameTest2: SubmitUrb failed sc=x%x\n"),sc));
                return sc;
            }

            if (++Bs >= &ActiveUrbs[nURB])
                Bs = &ActiveUrbs[0];
            Urb = &Bs->Urb;

            CheckFrame(LastBuf++);
        }
    }

    /* Tail end: wait until all URBs are completed
     * (we could also put a big sleep of 16msec here)
     */
#if 1
 Again:
    Bs = &ActiveUrbs[0];
    for (i = 0; i < nURB; i++) {

        Urb = &Bs->Urb;
        if ((Urb->Flags & URB_FLAG_COMPLETE) == 0) {
            DBGME(0,_tprintf(_T("FrameTest: Tail URB %d not complete.\n"),i));
            sc = Condition_Wait(&Done,NULL,NULL);
            if (FAILED(sc))
                return sc;
            goto Again;
        }

        Bs++;
    }
#else
    {
        TIME t;
        Int32ToInt64(t,TIME_RELATIVE(TIME_MILLIS(16)));
        SleepUntil(t);
    }
#endif

    DBGME(1,_tprintf(_T("\nFrameTest(%d) ok.\n"),Args->Configuration));
    return S_OK;
}

PRIVATE SCODE IsoTest(PDEVARGS Args)
{
    UINT UsbAddress32, Port32;
    PIUSBDRIVER Driver = NULL;
    SCODE sc;
    PIUSBHUB Hub = NULL;
    PINAMESPACE Ns = CurrentNameSpace();
    PINAMESPACE HubNs = NULL;

    _tprintf(_T("\nTesting UsbDevice: '%s' on HUB '%s'\n"),
             Args->DeviceName,Args->HubName);

    /* Must be a namespace
     */
    sc = BindToObject(Ns,
                      Args->HubName,
                      NAME_SPACE_READ,
                      &IID_INameSpace,
                      (void **)&HubNs);
    if (FAILED(sc)) {
        _tprintf(_TEXT("%s is not an INameSpace (%x)\n\r"), 
                 Args->HubName, sc);
        goto Out;
    }


    /* And it must be a HUB..
     */
    sc = BindToObject(Ns,
                      Args->HubName,
                      NAME_SPACE_READ,
                      &IID_IUsbHub,
                      (void **)&Hub);
    if (FAILED(sc)) {
        _tprintf(_TEXT("%s is not an IUsbHub (%x)\n\r"), 
                 Args->HubName, sc);
        goto Out;
    }

    /* And the device is attached there
     */
    sc = BindToObject(HubNs,
                      Args->DeviceName,
                      NAME_SPACE_READ,
                      &IID_IUsbDriver,
                      (void **)&Driver);
    if (FAILED(sc)) {
        _tprintf(_T("%s: cant find or not a UsbDriver, sc=x%x\n"),
                 Args->DeviceName,sc);
        goto Out;
    }

    sc = Driver->v->GetUsbAddress(Driver,&UsbAddress32,&Port32);
    if (FAILED(sc)) {
        _tprintf(_T("%s has no UsbAddress, sc=x%x\n"),Args->DeviceName,sc);
        goto Out;
    }

    Args->UsbAddress = (UINT8) UsbAddress32;
    Args->Port       = (UINT8) Port32;
    /* BUGBUG See note at top */
    Args->Mps        =  MAXCTLPKTSIZE;
    sc = VerifyTheMps(Hub,Args);
    if (FAILED(sc)) goto Out;

    DBGME(0,_tprintf(_T("IsoTest(%s) @UsbAddress %d Port %d Mps %d\n"),
                     Args->DeviceName,Args->UsbAddress,Args->Port,Args->Mps));

    sc = GetConfigurationInformation(Hub,Args);
    if (FAILED(sc)) goto Out;
    DBGME(1,PrintConfigurationInformation(Args));

    sc = SelectConfiguration(Hub,Args);
    if (FAILED(sc)) goto Out;

    sc = InitializeCamera(Hub,Args);
    if (FAILED(sc)) goto Out;

    sc = FrameTest(Hub,Args);
    if (FAILED(sc)) goto Out;

    sc = FinalizeCamera(Hub,Args);
    if (FAILED(sc)) goto Out;

 Out:
    if (FAILED(sc))
        DBGME(3,_tprintf(_T("\nTest FAILED sc==x%x\n"),sc));
    
    if (HubNs)
        HubNs->v->Release(HubNs);
    if (Hub)
        Hub->v->Release(Hub);
    if (Driver)
        Driver->v->Release(Driver);
    return sc;
}

int _tmain(int argc, _TCHAR **argv)
{
    SCODE sc;
    _TCHAR *Tail;
    DEVARGS Args;

    sc = ParseCmdLine(&argc, &argv);
    if (FAILED(sc)) {
        _tprintf(_TEXT("No command line args (%x)\n\r"), sc);
        return -1;
    }

    /* There must be at least an argument, the full path of the device
     */
    if (argc < 2) {
    Usage:
        _tprintf(_TEXT("Usage:%s IsochDeviceName\n\r"), argv[0]);
        goto Out;
    }

    memset(&Args,0,sizeof Args);

    /* Split the argument "<hub>\Vid...." in two, we need the HUB as well
     */
    Args.HubName    = argv[1];
    Args.DeviceName = argv[1];
    for (;;) {
        Tail = _tcschr(Args.DeviceName,_T('\\'));
        if (Tail == NULL)
            Tail = _tcschr(Args.DeviceName,_T('/'));
        if (Tail == NULL)
            break;
        Args.DeviceName = Tail + 1;
    }
    if (Args.DeviceName == argv[1]) {
        goto Usage;
    }
    Args.DeviceName[-1] = 0;

    /* Optional arguments
     */
    Args.Configuration = 1;
    Args.Interface = 0;
    Args.Setting = 0;
    Args.nLoops = 10000;
    Args.InputEndpoint = 1;

    if (argc > 2)
        Args.Configuration = _ttoi(argv[2]);

    if (argc > 3)
        Args.Interface = _ttoi(argv[3]);

    if (argc > 4)
        Args.Setting = _ttoi(argv[4]);

    if (argc > 5)
        Args.nLoops = _ttoi(argv[5]);


    sc = IsoTest(&Args);

 Out:
    (void) FreeCmdLine(&argc,&argv);
    return 0;
}
