/*
* Utility.c contains several utility functions for test suites ... 
* Reporting functions for the most part. 
* Copyright (c) Microsoft Corporation All rights reserved
*/

#include <tchar.h>
#include <stdlib.h>
#include <stdio.h>
#include <mmlite.h>

#include <assert.h>
#include "utility.h"

/*
* The ops table
* To add one do the following:
*   - Add it here, AT THE END OF THIS TABLE, e.g., "MY_NEW_OP" after "QUERYINTERFACE"
*   - Add #define to utility.h, so that #define value == array index
*   - If flags are different, create an array of type flag_desc_map, and populate it
*   (See flags_map for an example)
*   - Touch up switch stmt. in   to include your op, and return array of your flags

*/
_TCHAR * OPS[] = 
{
    _TEXT("BIND"), 
    _TEXT("READ"), 
    _TEXT("WRITE"), 
    _TEXT("GETSIZE"), 
    _TEXT("SETSIZE"),
    _TEXT("UNREGISTER"), 
    _TEXT("LOADIMAGE"),
    _TEXT("CREATETHREAD"), 
    _TEXT("SETEXITCODE"), 
    _TEXT("REGISTER"), 
    _TEXT("QUERYINTERFACE"),
    _TEXT("VALIDATE"),
    _TEXT("STATUS"),
    _TEXT("FINDNEXT"), 
    _TEXT("GETFUNCTIONADDRESS"),
    _TEXT("EXTRACT"),
    _TEXT("ALLOCATE")
}; 

/*
* Parses command lines of the form:
* FOO.EXE -w -r -c -x -sr -sw -bogus -size 100
* Returns 
*   
*/
UINT FetchFlagsFromCommandLine(UINT argc, _TCHAR ** argv, UINT64 * pSize)
{
    UINT i = 0; 
    UINT nRet = 0;

    for(i = 1; i < argc; ++i)
    {
        if(*argv[i] != '-' && *argv[i] != '/')
        {
            continue;
        }

        if(!_tcsicmp(argv[i] + 1,_TEXT("r")))
        {
            nRet |= NAME_SPACE_READ;
            continue;
        }

        if(!_tcsicmp(argv[i] + 1, _TEXT("w")))
        {
            nRet |= NAME_SPACE_WRITE;
            continue; 
        }

        if(!_tcsicmp(argv[i] + 1, _TEXT("c")))
        {
            nRet |= NAME_SPACE_CREATE;
            continue; 
        }

        if(!_tcsicmp(argv[i] + 1, _TEXT("x")))
        {
            nRet |= NAME_SPACE_FAILIFEXIST;
            continue;
        }

        if(!_tcsicmp(argv[i] + 1, _TEXT("sw")))
        {
            nRet |= NAME_SPACE_SHARE_WRITE;
            continue;
        }
        
        if(!_tcsicmp(argv[i] + 1, _TEXT("sr")))
        {
            nRet |= NAME_SPACE_SHARE_READ;
            continue;
        }

        if(!_tcsicmp(argv[i] + 1, _TEXT("bogus")))
        {
            nRet |= NAME_SPACE_BOGUS_FLAG;
            continue;
        }

        if(!_tcsicmp(argv[i] + 1, _TEXT("size")))
        {
            _TCHAR *stop = 0;
            if(pSize)
                if(argc > i)
                    Uint32ToUint64(*pSize, _tcstoul(argv[i+1], &stop, 16));
                    
             continue;
        }
    }
    return nRet; 
}

/*
* Reporting function for various operations. As long as op is in the ops table
* this should give user some visual output ... 
* Preconditions:
*   - op is in op table in utility.h
*   - SCODE is legit
*   - sFile has file name 
*   - flags are flags
*   - nBytes is a generic size descriptor, usually bytes allocated if testing heap
*/
void Outcome(BYTE op, SCODE sc, const _TCHAR * sFile, UINT flags, UINT64 nBytes)
{
    _TCHAR sFlagsDesc[256]; 
    GetFlagsDescription(sFlagsDesc, sizeof(sFlagsDesc), flags, op);
    _tprintf(_T("\n\rOP:[%s] returned '%s' [0x%.8x] on file or object [%s]\n\r\tFlags '%s' [0x%.8x] Bytes [0x%.8x]"), 
            OPS[op], ErrorDescription(sc), sc, sFile, sFlagsDesc, flags, Uint64Low32Bits(nBytes)); 
}

void CreateThreadOutcome(BYTE op, SCODE sc, THREAD_FUNCTION pStart, IThread * pThread, const _TCHAR * sArgs, ADDR_SIZE stacksize)
{
    _tprintf(
        _T("\n\rOP:[%s] returned '%s' [0x%.8x] Thread [0x%.8x] Entry point [0x%.8x] Stack size [%d]\n\r\tThread args [%s]"), 
        OPS[op], 
        ErrorDescription(sc), 
        sc, 
        (UINT)pThread, 
        (UINT32)pStart, //suppress ARM compiler warning
        (UINT)stacksize, 
        sArgs ? sArgs : _T("null")
        ); 
    fflush(stdout);
}

void SetExitCodeOutcome(BYTE op, SCODE sc)
{
    _tprintf(_T("\n\r[%s] returned [0x%.8x], [%s]"), OPS[op], sc, ErrorDescription(sc));
}


/*
#define NAME_SPACE_READ 0x1
#define NAME_SPACE_WRITE 0x2
#define NAME_SPACE_SHARE_READ 0x4
#define NAME_SPACE_SHARE_WRITE 0x8
#define NAME_SPACE_CREATE 0x10
#define NAME_SPACE_FAILIFEXIST 0x20
#define NAME_SPACE_EXTENSION 0x800
#define NAME_SPACE_NOFOLLOW 0x1000
#define NAME_SPACE_PREFIX 0x2000
#define NAME_SPACE_APPEND 0x40000
*/
flags_desc_map file_flags[] = 
{
    {NAME_SPACE_READ, _TEXT("NAME_SPACE_READ")}, 
    {NAME_SPACE_WRITE, _TEXT("NAME_SPACE_WRITE")}, 
    {NAME_SPACE_CREATE, _TEXT("NAME_SPACE_CREATE")}, 
    {NAME_SPACE_FAILIFEXIST, _TEXT("NAME_SPACE_FAILIFEXIST")}, 
    {NAME_SPACE_SHARE_READ, _TEXT("NAME_SPACE_SHARE_READ")}, 
    {NAME_SPACE_SHARE_WRITE, _TEXT("NAME_SPACE_SHARE_WRITE")}, 
    {NAME_SPACE_EXTENSION,  _TEXT("NAME_SPACE_EXTENSION")},
    {NAME_SPACE_NOFOLLOW,  _TEXT("NAME_SPACE_NOFOLLOW")},
    {NAME_SPACE_PREFIX,  _TEXT("NAME_SPACE_PREFIX")},
    {NAME_SPACE_APPEND,  _TEXT("NAME_SPACE_APPEND")},
    {NAME_SPACE_BOGUS_FLAG, _TEXT("NAME_SPACE_BOGUS_FLAG")}, 
    {0,0}
};

flags_desc_map process_flags[] = 
{
    {LOAD_IMAGE_CALL_ENTRY_POINT, _TEXT("LOAD_IMAGE_CALL_ENTRY_POINT")}, 
    {LOAD_IMAGE_CREATE_THREAD,  _TEXT("LOAD_IMAGE_CREATE_THREAD")}, 
    {0,0}
};

flags_desc_map null_flag[] = 
{
    {0, _TEXT("No flags for this function")}
}; 


flags_desc_map alloc_flags[] = 
{
    {HEAP_ZERO_MEMORY, _TEXT("HEAP_ZERO_MEMORY")}, 
    {HEAP_REALLOC_IN_PLACE_ONLY, _TEXT("HEAP_REALLOC_IN_PLACE_ONLY")}, 
    {HEAP_CACHE_ALIGN, _TEXT("HEAP_CACHE_ALIGN")}, 
    {HEAP_NO_COPY, _TEXT("HEAP_NO_COPY")}, 
    {HEAP_NO_SERIALIZE, _TEXT("HEAP_NO_SERIALIZE")}, 
    {0, 0}
};

/*
* Different flags are interpreted different according to operation
* We just check ops table to see which flags map we will use
* To add an operation, see comments in op table at header. 
* You will need to add a case here for your new op. 
*/
const flags_desc_map * GetDescMap(BYTE op)
{
    switch(op)
    {
    case    BIND:
    case    READ:
    case    WRITE:
    case    GETSIZE:
    case    SETSIZE:
    case    UNREGISTER:
    case    REGISTER:
        return file_flags;

    case    LOADIMAGE:
        return process_flags; 
    
    case    QUERYINTERFACE:
    case    VALIDATE:
    case    STATUS:
    case    FINDNEXT:
        return null_flag; 

    case    EXTRACT:
    case    ALLOCATE:
        return alloc_flags;

//  case    MY_NEW_OP:
//        foo(); 
//        return new_op_flags;

    

    default:
        return 0; 
    }
}

/*
* Gets a text description of flags, e.g. 'F_READ | F_WRITE'
* Finds right flags map by operation performed ...  
*/
void GetFlagsDescription(_TCHAR * sFlags, UINT nLen, UINT flags, BYTE op)
{
    BOOL bHit = FALSE;
    UINT nTotal = 0; 
    unsigned idx; 
    
    const flags_desc_map * f = GetDescMap(op);

    memset(sFlags, 0, nLen * sizeof(sFlags[0]));

    for(idx = 0; f[idx].sDesc; ++idx)
    {
        if(f[idx].flag & flags)
        {
            _TCHAR * sNew = f[idx].sDesc; 
            //total so far + len of new + pipe char + null
            if(nTotal + _tcslen(sNew) + 1 + 1 >= nLen)
                break; 
            _tcscat(sFlags, sNew); 
            _tcscat(sFlags, _T("|")); 
            bHit = TRUE;
        }
    }

    if(!bHit)
        _tcsncpy(sFlags, _T("(zero)"), nLen);
    
    //stomp trailing pipe char
    if(sFlags[_tcslen(sFlags) - 1] == '|')
        sFlags[_tcslen(sFlags) - 1] = '\0'; 
}
        
/*
* Map of all standard error descriptions
*/
typedef struct err_desc_map
{
    SCODE sc; 
    _TCHAR * sDesc; 
}err_desc_map;

err_desc_map a[] =  
{
    {S_OK,_TEXT("S_OK")},
    {S_FALSE,_TEXT("S_FALSE")},
    {E_UNEXPECTED_ERROR,_TEXT("E_UNEXPECTED_ERROR")},
    {E_OBJECT_NOT_INITIALIZED,_TEXT("E_OBJECT_NOT_INITIALIZED")},
    {E_TIMED_OUT,_TEXT("E_TIMED_OUT")},
    {E_ALREADY_RESERVED,_TEXT("E_ALREADY_RESERVED")},
    {E_NOT_RESERVED,_TEXT("E_NOT_RESERVED")},
    {E_ALREADY_COMMITTED,_TEXT("E_ALREADY_COMMITTED")},
    {E_NOT_COMMITTED,_TEXT("E_NOT_COMMITTED")},
    {E_TOO_MANY_CONTIGUOUS,_TEXT("E_TOO_MANY_CONTIGUOUS")},
    {E_RESTART,_TEXT("E_RESTART")},
    {S_ALREADY_EXISTS,_TEXT("S_ALREADY_EXISTS")},
    {S_BUFFER_TOO_SMALL,_TEXT("S_BUFFER_TOO_SMALL")},
    {E_NOT_IMPLEMENTED,_TEXT("E_NOT_IMPLEMENTED")},
    {E_NO_INTERFACE,_TEXT("E_NO_INTERFACE")},
    {E_FAIL,_TEXT("E_FAIL")},
    {E_FILE_NOT_FOUND,_TEXT("E_FILE_NOT_FOUND")},
    {E_PATH_NOT_FOUND,_TEXT("E_PATH_NOT_FOUND")},
    {E_ACCESS_DENIED,_TEXT("E_ACCESS_DENIED")},
    {E_INVALID_HANDLE,_TEXT("E_INVALID_HANDLE")},
    {E_NOT_ENOUGH_MEMORY,_TEXT("E_NOT_ENOUGH_MEMORY")},
    {E_WRITE_PROTECT,_TEXT("E_WRITE_PROTECT")},
    {E_BAD_UNIT,_TEXT("E_BAD_UNIT")},
    {E_NOT_READY,_TEXT("E_NOT_READY")},
    {E_CRC,_TEXT("E_CRC")},
    {E_DMA_FAILURE,_TEXT("E_DMA_FAILURE")},
    {E_SEEK,_TEXT("E_SEEK")},
    {E_SECTOR_NOT_FOUND,_TEXT("E_SECTOR_NOT_FOUND")},
    {E_OUT_OF_STRUCTURES,_TEXT("E_OUT_OF_STRUCTURES")},
    {E_INVALID_PARAMETER,_TEXT("E_INVALID_PARAMETER")},
    {E_BUFFER_TOO_SMALL,_TEXT("E_BUFFER_TOO_SMALL")},
    {E_MOD_NOT_FOUND,_TEXT("E_MOD_NOT_FOUND")},
    {E_PROC_NOT_FOUND,_TEXT("E_PROC_NOT_FOUND")},
    {E_INVALID_ORDINAL,_TEXT("E_INVALID_ORDINAL")},
    {E_ALREADY_EXISTS,_TEXT("E_ALREADY_EXISTS")},
    {E_INVALID_EXE_SIGNATURE,_TEXT("E_INVALID_EXE_SIGNATURE")},
    {E_EXE_MARKED_INVALID,_TEXT("E_EXE_MARKED_INVALID")},
    {E_BAD_EXE_FORMAT,_TEXT("E_BAD_EXE_FORMAT")},
    {E_LOCKED,_TEXT("E_LOCKED")},
    {E_INVALID_ADDRESS,_TEXT("E_INVALID_ADDRESS")},
    {E_IO_PENDING,_TEXT("E_IO_PENDING")},
    {E_NOACCESS,_TEXT("E_NOACCESS")},
    {E_MEDIA_CHANGED,_TEXT("E_MEDIA_CHANGED")},
    {E_NO_MEDIA_IN_DRIVE,_TEXT("E_NO_MEDIA_IN_DRIVE")},
    {E_DLL_INIT_FAILED,_TEXT("E_DLL_INIT_FAILED")},
    {E_FLOPPY_ID_MARK_NOT_FOUND,_TEXT("E_FLOPPY_ID_MARK_NOT_FOUND")},
    {E_FLOPPY_WRONG_CYLINDER,_TEXT("E_FLOPPY_WRONG_CYLINDER")},
    {E_FLOPPY_UNKNOWN_ERROR,_TEXT("E_FLOPPY_UNKNOWN_ERROR")},
    {E_FLOPPY_BAD_REGISTERS,_TEXT("E_FLOPPY_BAD_REGISTERS")},
    {E_DISK_RECALIBRATE_FAILED,_TEXT("E_DISK_RECALIBRATE_FAILED")},
    {E_DISK_OPERATION_FAILED,_TEXT("E_DISK_OPERATION_FAILED")},
    {E_DISK_RESET_FAILED,_TEXT("E_DISK_RESET_FAILED")},
    {E_INTERNAL_ERROR,_TEXT("E_INTERNAL_ERROR")},
    {E_ENTRY_NOT_FOUND,_TEXT("E_ENTRY_NOT_FOUND")},
    {E_UNRECOGNIZED_MEDIA,_TEXT("E_UNRECOGNIZED_MEDIA")},
    {E_SIG_NOT_FOUND,_TEXT("E_SIG_NOT_FOUND")},
    {E_DEVICE_IN_USE,_TEXT("E_DEVICE_IN_USE")},
    {S_NO_MORE_ENTRIES,_TEXT("S_NO_MORE_ENTRIES")},
    {E_NAME_NOT_FOUND,_TEXT("E_NAME_NOT_FOUND")},
    {E_OBJECT_NOT_FOUND,_TEXT("E_OBJECT_NOT_FOUND")},
    {E_INVALID_METHOD,_TEXT("E_INVALID_METHOD")},
    {0x80070050 ,      _TEXT("File already exists")}
}; 

/*
* Maps SCODE onto text description
*/
const _TCHAR * ErrorDescription(SCODE sc)
{
    int idx;
    for(idx = 0 ; idx < sizeof(a)/sizeof(a[0]); ++idx)
        if(a[idx].sc == sc)
            return a[idx].sDesc; 
        return _TEXT("Unmapped Error Code");
}


/*
Fetches command line parameter if preceded by a '-' or '/', so typing 
FOOBAR.EXE -arg BARFOO 
will return "BARFOO" if called as GetArgFromCommandLine(argc, argv, "arg")
Returns pointer into params, does not allocate mem for reply ... copy out of response
ptr before calling FreeCmdLine() !!
Returns NULL if arg not found or if arg empty

*/
const _TCHAR * GetArgFromCommandLine(UINT argc, _TCHAR **argv, const _TCHAR * sFlag)
{
    unsigned i;
    for(i = 1; i < argc; ++i)
    {
        if(*argv[i] != (_TCHAR)'-' && *argv[i] != (_TCHAR)'/')
            continue; 

        if(!_tcsicmp(argv[i] + 1, sFlag))
        {
            if(argc > i)
                return argv[i + 1];
            else
                return 0;
        }
    }
    return 0;
}

    
//for returning processor types in IModule::MachineInfo()
_TCHAR * PK[] = 
{
   _TEXT("IA32"),
   _TEXT("MIPS"),
   _TEXT("MIPS_TIGON"),
   _TEXT("PPC"),
   _TEXT("H8"),
   _TEXT("ARM"), 
   _TEXT("M68K"), 
   _TEXT("ALPHA"),
   _TEXT("TRIMEDIA")
};


const _TCHAR * ProcessorKind(UINT32 nProc)
{
    if(nProc < sizeof(PK)/sizeof(PK[0]))
        return PK[nProc];
    else
        return _T("No such processor"); 
}


const _TCHAR * ProcessorFlavor(UINT32 nFlavor)
{
    switch(nFlavor)
    {
    case MACHINE_KIND_GENERIC:
        return _T("Generic"); 

    case    MACHINE_KIND_TRIMEDIA:
        return _T("Trimedia");

    default:
        return _T("No such kind of machine"); 
    }
}
 
/*
Easy way to force a debug break from command line. 
User can pass '-break 1' on command line to force a DebugBreak() 
at the point this is called.
*/   
void OptionalBreak(int argc, _TCHAR ** argv)
{
    const _TCHAR * sArg; 
    _TCHAR * stop = 0;
    
    sArg = GetArgFromCommandLine(argc, argv, _T("break")); 
    
    if(sArg && *sArg)
        if(_tcstoul(sArg, &stop, 10))
            DebugBreak(); 
}

void TellAboutOptionalBreak()
{
    _tprintf(_T("\n\rPass '-break 1' to break into debugger"));
}

void NSOutcome(BYTE op, SCODE sc, const _TCHAR * sPref, const _TCHAR * sPrev, const _TCHAR * sOutput, UINT nBuflen, UINT nStrLenOut)
{
   _tprintf(_T("\n\rOP:[%s] returned '%s' [0x%.8x] on prefix [%s], previous object [%s], output [%s]\n\r\tBuffer length going in [%d], bytes output [%d]"), 
        OPS[op], ErrorDescription(sc), sc, sPref, sPrev, sOutput, nBuflen, nStrLenOut);
}

void GFAOutcome(BYTE op, SCODE sc, const _TCHAR * sFuncName, INT nOrdinal, ADDRESS addr)
{
    _tprintf(_T("\n\rOP:[%s] returned '%s' [0x%.8x] after seeking function [%s], ordinal [%d], address [0x%.8x]"), 
        OPS[op], 
        ErrorDescription(sc), 
        sc, 
        sFuncName, 
        nOrdinal, 
        (UINT32)addr);  //suppress ARM compiler warning
}

