/* Copyright (c) Microsoft Corporation. All rights reserved. */

#include <stdio.h>
#include <string.h>
#include <mmlite.h>
#include <assert.h>
#include <soapif.h>
#include <util/list-double2.h>
#include <sampleif.h>
#include "../../src/web/ws/wsdiscov.h"

SCODE Setup(PIWSDISCOVERY *ppDisco);
SCODE TestWsAdvertNew(void);
SCODE TestWsDiscoveryNameSpaceRegister(PIWSDISCOVERY pDisco);
SCODE TestWsDiscoveryNameSpaceUnregister(PIWSDISCOVERY pDisco);
SCODE TestWsDiscoveryNameSpaceBind(PIWSDISCOVERY pDisco);
SCODE TestWsDiscoveryNameSpaceFindNext(PIWSDISCOVERY pDisco);
SCODE TestWsDiscoveryNameSpaceGetCapabilities(PIWSDISCOVERY pDisco);

extern PIUNKNOWN CreateWsDiscovery(void);

int _tmain(INT argc, _TCHAR *argv[])
{
    SCODE sc;
    PIWSDISCOVERY pDisco;

    sc = Setup(&pDisco);
    if (FAILED(sc)) goto Failed;

    sc = TestWsAdvertNew();
    if (FAILED(sc)) goto Failed;

    sc = TestWsDiscoveryNameSpaceRegister(pDisco);
    if (FAILED(sc)) goto Failed;

    sc = TestWsDiscoveryNameSpaceBind(pDisco);
    if (FAILED(sc)) goto Failed;

    sc = TestWsDiscoveryNameSpaceUnregister(pDisco);
    if (FAILED(sc)) goto Failed;

    sc = TestWsDiscoveryNameSpaceGetCapabilities(pDisco);
    if (FAILED(sc)) goto Failed;

    sc = TestWsDiscoveryNameSpaceFindNext(pDisco);
    if (FAILED(sc)) goto Failed;

    return 0;

Failed:
    DBG(("WS-Discovery tests failed\n"));
    return -1;
}

SCODE Setup(PIWSDISCOVERY *ppDisco)
{
    SCODE sc = BindToObject(
        CurrentNameSpace(), 
        _T("COB/wsdiscov.cob"), 
        NAME_SPACE_READ, 
        &IID_IWsDiscovery, 
        (void**) ppDisco);
    if (FAILED(sc)) {
        DBG(("Failed to load wsdiscov.cob\n"));
        return sc;
    }

    return sc;
}

extern PWSADVERT WsAdvertNew(
    const _TCHAR *Address, 
    const _TCHAR *Types, 
    const _TCHAR *Scopes, 
    UINT MetadataVersion);

SCODE TestWsAdvertNew(void)
{
    DBG(("Testing WsAdvertNew with null parameters\n"));
    {
        PWSADVERT pAdvert = WsAdvertNew(NULL, NULL, NULL, 0);
        if (pAdvert) {
            DBG(("FAILED: WsAdvertNew parameters not checked\n"));
            return E_FAIL;
        }
    }

    DBG(("Testing WsAdvertNew with valid parameters, #1\n"));
    {
        PWSADVERT pAdvert;
        const _TCHAR *EndpointAddress = _T("uuid:0");
        const _TCHAR *PrinterTypeName = _T("i:Printer");
        const _TCHAR *ScannerTypeName = _T("j:Scanner");
        const _TCHAR *Types = _T("i:Printer j:Scanner");
        const _TCHAR *UpstairsScope = _T("ldap:///upstairs");
        const _TCHAR *DownstairsScope = _T("ldap:///downstairs");
        const _TCHAR *Scopes = _T("ldap:///upstairs ldap:///downstairs");
        const UINT MetadataVersion = 42;

        pAdvert = WsAdvertNew(EndpointAddress, Types, Scopes, MetadataVersion);
        if (!pAdvert) {
            DBG(("FAILED: WsAdvertNew\n"));
            return E_FAIL;
        }

        DBG(("#1, checking endpoint\n"));
        if (_tcscmp(EndpointAddress, pAdvert->Endpoint->Address)) {
            DBG(("FAILED: Endpoint addresses:\n\t%s\n\t%s\n", 
                 EndpointAddress, pAdvert->Endpoint->Address));
            return E_FAIL;
        }

        DBG(("#1, checking types\n"));
        {
            PWSPROPERTY Printer = pAdvert->Types;
            PWSPROPERTY Scanner = ListNodeNext(ploff, Printer);
            PWSPROPERTY Check = ListNodeNext(ploff, Scanner);
            if (_tcscmp(Printer->QName, PrinterTypeName)) {
                DBG(("FAILED: Types:\n\t%s\n\t%s\n", 
                     Printer->QName, PrinterTypeName));
                return E_FAIL;
            }
            if (_tcscmp(Scanner->QName, ScannerTypeName)) {
                DBG(("FAILED: Types:\n\t%s\n\t%s\n",
                    Scanner->QName, ScannerTypeName));
                return E_FAIL;
            }
            if (Check != Printer) {
               DBG(("FAILED: Malformed type list\n"));
               return E_FAIL;
            }  
        }

        DBG(("#1, checking scopes\n"));
        {
            PWSPROPERTY Upstairs = pAdvert->Scopes;
            PWSPROPERTY Downstairs = ListNodeNext(ploff, Upstairs);
            PWSPROPERTY Check = ListNodeNext(ploff, Downstairs);
            if (_tcscmp(Upstairs->QName, UpstairsScope)) {
                DBG(("FAILED: Scopes:\n\t%s\n\t%s\n", 
                     Upstairs->QName, UpstairsScope));
                return E_FAIL;
            }
            if (_tcscmp(Downstairs->QName, DownstairsScope)) {
                DBG(("FAILED: Scopes:\n\t%s\n\t%s\n", 
                     Downstairs->QName, DownstairsScope));
                return E_FAIL;
            }
            if (Check != Upstairs) {
                DBG(("FAILED: Malformed scope list\n"));
                return E_FAIL;
            }
        }

        DBG(("#1, checking metadata version\n"));
        if (pAdvert->MetadataVersion != MetadataVersion) {
            DBG(("FAILED: Metadata versions:\n\t%d\n\t%d\n", 
                 MetadataVersion, pAdvert->MetadataVersion));
            return E_FAIL;
        }

        DBG(("#1, success\n"));

        return S_OK;
    }
}

typedef struct TESTTARGET {
    struct IWsTargetServiceVtbl *v;
    UINT Refs;
    UINT Index;
} TESTTARGET, *PTESTTARGET;

SCODE TestTargetQueryInterface(PIWSTARGETSERVICE Target, REFIID Iid, void** Unk) 
{
    *Unk = (PIUNKNOWN) &Target->v;
    return S_OK;
}

TESTTARGET Target;
PTESTTARGET pTarget = &Target;

#define TESTTARGET_ENDPOINT _T("uuid:0")
#define TESTTARGET_TYPE _T("i:TestTarget")
#define TESTTARGET_SCOPE NULL
#define TESTTARGET_METADATAVERSION 1

SCODE TestTargetCloneAdvert(PIWSTARGETSERVICE pTarget, PWSADVERT *ppAdvert)
{
    *ppAdvert = WsAdvertNew(
        TESTTARGET_ENDPOINT, 
        TESTTARGET_TYPE, 
        TESTTARGET_SCOPE, 
        TESTTARGET_METADATAVERSION);
    if (!*ppAdvert) return E_FAIL;
    (*ppAdvert)->Target = pTarget;
    ListNodeInit(aloff, *ppAdvert);
    return S_OK;
}

struct IWsTargetServiceVtbl TestTargetVtbl = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    TestTargetQueryInterface, 
    NULL,
    NULL,
    TestTargetCloneAdvert
};

SCODE TestWsDiscoveryNameSpaceRegister(PIWSDISCOVERY pDisco)
{
    SCODE sc;
    PWSDISCOVERY pDiscoThis = pD(pDisco);
    PINAMESPACE pNs = (PINAMESPACE) &pDiscoThis->vns;

    DBG(("TestWsDiscoveryNameSpaceRegister this=0x%x\n", 
        pDiscoThis));

    pTarget->v = &TestTargetVtbl;
    pTarget->Refs = 1;

#define BEEF _T("deadbeef")
    DBG(("#1, ns=%x name=%s unk=%x flags=%x xxx=%x\n", 
         pNs, BEEF, pTarget, 0, NULL));
    sc = pNs->v->Register(pNs, BEEF, (PIUNKNOWN) pTarget, 0, NULL);
    if (SUCCEEDED(sc)) {
        DBG(("FAILED: Non-null pName parameter allowed\n"));
        return sc;
    }

    DBG(("#2, ns=%x name=%s unk=%x flags=%x xxx=%x\n", 
         pNs, NULL, NULL, 0, NULL));
    sc = pNs->v->Register(pNs, NULL, NULL, 0, NULL);
    if (SUCCEEDED(sc)) {
        DBG(("FAILED: Null target service parameter allowed\n"));
        return sc;
    }

    DBG(("#3, ns=%x name=%s unk=%x flags=%x xxx=%x\n", 
         pNs, NULL, pTarget, 1, NULL));
    sc = pNs->v->Register(pNs, NULL, (PIUNKNOWN) pTarget, 1, NULL);
    if (SUCCEEDED(sc)) {
        DBG(("FAILED: Non-zero flags parameter allowed\n"));
        return sc;
    }

    DBG(("#4, ns=%x name=%s unk=%x flags=%x xxx=%x\n", 
         pNs, NULL, pTarget, 0, NULL));
    sc = pNs->v->Register(pNs, NULL, (PIUNKNOWN) pTarget, 0, NULL);
    if (FAILED(sc)) {
        DBG(("FAILED: Registration should have succeeded, sc=%x\n", sc));
        return sc;
    }

    {
        DBG(("#4: Checking adverts list\n"));
        if (pDiscoThis->Adverts != ListNodeNext(aloff, pDiscoThis->Adverts)) {
            DBG(("FAILED: Malformed adverts list\n"));
            return E_FAIL;
        }
        DBG(("#4: Checking advert endpoint address\n"));
        if (_tcscmp(TESTTARGET_ENDPOINT, pDiscoThis->Adverts->Endpoint->Address)) {
            DBG(("FAILED: Endpoints \n\t%s\n\t%s\n", 
                 TESTTARGET_ENDPOINT, pDiscoThis->Adverts->Endpoint->Address));
            return E_FAIL;
        }
        DBG(("#4: Checking advert target service\n"));
        if (pDiscoThis->Adverts->Target != (PIWSTARGETSERVICE) &pTarget->v) {
            DBG(("FAILED: Targets 0x%x 0x%x\n", 
                pTarget, pDiscoThis->Adverts->Target));
            return E_FAIL;
        }
        DBG(("Checking advert types\n"));
        if (_tcscmp(TESTTARGET_TYPE, pDiscoThis->Adverts->Types->QName)) {
            DBG(("FAILED: Types \n\t%s\n\t%s\n", 
                 TESTTARGET_TYPE, pDiscoThis->Adverts->Types->QName));
            return E_FAIL;
        }
    }

    DBG(("SUCCEEDED (Test...Register): Registered %s\n", TESTTARGET_ENDPOINT));

    return S_OK;
}

SCODE TestWsDiscoveryNameSpaceUnregister(PIWSDISCOVERY pDisco) 
{
    PWSDISCOVERY pDiscoThis = pD(pDisco);
    PINAMESPACE pNs = (PINAMESPACE) &pDiscoThis->vns;

#define BAD_ENDPOINT _T("blah")
    SCODE sc = pNs->v->Unregister(pNs, BAD_ENDPOINT);
    if (SUCCEEDED(sc)) {
        DBG(("FAILED: Successfully unregistered %s\n", BAD_ENDPOINT));
        return E_FAIL;
    }
    sc = pNs->v->Unregister(pNs, TESTTARGET_ENDPOINT);
    if (FAILED(sc)) {
        DBG(("FAILED: Could not unregister %s\n(Did Test...Register succeed?)\n", 
             TESTTARGET_ENDPOINT));
        return E_FAIL;
    }

    DBG(("SUCCEEDED (Test...Unregister): Unregistered %s\n", TESTTARGET_ENDPOINT));

    return S_OK;
}

SCODE TestWsDiscoveryNameSpaceBind(PIWSDISCOVERY pDisco)
{
    SCODE sc;
    PWSDISCOVERY pDiscoThis = pD(pDisco);
    PINAMESPACE pNs = (PINAMESPACE) &pDiscoThis->vns;
    PIUNKNOWN pUnk = NULL;

    sc = pNs->v->Bind(pNs, TESTTARGET_ENDPOINT, 0, &pUnk);
    if (FAILED(sc)) {
        DBG(("FAILED: Could not bind to test target service, sc=%x\n", sc));
        return sc;
    }
    if (pUnk != (PIUNKNOWN) &pTarget->v) {
        DBG(("FAILED: Test target service pointers %x, %x\n", pUnk, &pTarget->v));
        return E_FAIL;
    }

    DBG(("SUCCEEDED (Test...Bind): Bound 0x%x\n", pUnk));

    return S_OK;
}

SCODE TestWsDiscoveryNameSpaceGetCapabilities(PIWSDISCOVERY pDisco)
{
    PWSDISCOVERY pDiscoThis = pD(pDisco);
    PINAMESPACE pNs = (PINAMESPACE) &pDiscoThis->vns;
    NAME_SPACE_FLAGS Flags = -1;
    
    pNs->v->GetCapabilities(pNs, NULL, &Flags);
    if (0 != Flags) {
        DBG(("FAILED: GetCapabilities returned %d, should return 0\n", Flags));
        return E_FAIL;
    }

    DBG(("SUCCEEDED (Test...GetCapabilities): No capabilities ;)\n"));

    return S_OK;
}

#define nTESTTARGETS 3
TESTTARGET Target1, Target2, Target3;
PTESTTARGET pTargets[] = { &Target1, &Target2, &Target3 };
_TCHAR *TestTargetEp[] = { _T("uuid:1"), _T("uuid:2"), _T("uuid:3") };
_TCHAR *TestTargetT[] = { 
    _T("i:Printer"), _T("i:Printer i:Scanner"), _T("i:Printer i:Scanner i:Camera") };
_TCHAR *TestTargetS[] = { 
    _T("ldap:///upstairs"), _T("ldap:///downstairs"), _T("ldap:///upstairs ldap:///downstairs") };
UINT TestTargetMd[] = { 0, 1, 2 };

SCODE TestTargetNastyCloneAdvert(PIWSTARGETSERVICE pTarget, PWSADVERT *ppAdvert)
{
    PTESTTARGET pTest = (PTESTTARGET) pTarget;
    *ppAdvert = NULL;
    *ppAdvert = WsAdvertNew(
        TestTargetEp[pTest->Index], 
        TestTargetT[pTest->Index], 
        TestTargetS[pTest->Index], 
        TestTargetMd[pTest->Index]);
    return ppAdvert ? S_OK : E_FAIL;
}

struct IWsTargetServiceVtbl TestTargetVtbl2 = {
    __VTABLE_COMPATIBILITY_FILLER_INITIALIZER

    TestTargetQueryInterface, 
    NULL,
    NULL,
    TestTargetNastyCloneAdvert
};

#define TESTFINDNEXTBUFSZ 1024
PRIVATE _TCHAR TestFindNextBuffer[TESTFINDNEXTBUFSZ];

SCODE TestWsDiscoveryNameSpaceFindNext(PIWSDISCOVERY pDisco)
{
    SCODE sc;
    PWSDISCOVERY pDiscoThis = pD(pDisco);
    PINAMESPACE pNs = (PINAMESPACE) &pDiscoThis->vns;
    UINT i, BufSz = TESTFINDNEXTBUFSZ, Len;
    _TCHAR *pBuffer = TestFindNextBuffer, *pPrefix, *pPrevious;

    DBG(("TestWsDiscoveryNameSpaceFindNext: this=0x%x\n", pDiscoThis));

    for (i = 0; i < nTESTTARGETS; ++i) {
        pTargets[i]->v = &TestTargetVtbl2; /* &TestTargetVtbl; to reproduce heap corruption problem XXX BUGBUG */
        pTargets[i]->Refs = 1;
        pTargets[i]->Index = i;
        
        DBG(("TestWsDiscoveryNameSpaceFindNext: registering %d\n", i)); 
        sc = pNs->v->Register(pNs, NULL, (PIUNKNOWN) pTargets[i], 0, NULL);
        if (FAILED(sc)) {
            DBG(("FAILED (Test...FindNext 1): Registering index=%d, sc=%x\n", i, sc));
            return sc;
        }
    }

    Len = 0;
    sc = pNs->v->FindNext(pNs, NULL, NULL, pBuffer, BufSz, &Len);
    if (FAILED(sc)) {
        DBG(("FAILED (Test...FindNext 2): Default /ws/, sc=%x, buffer=%s, size=%d, length=%d\n",
             sc, pBuffer, BufSz, Len));
        return sc;
    }

    Len = 0;
    pPrefix = _T("/ws");
    sc = pNs->v->FindNext(pNs, pPrefix, NULL, pBuffer, BufSz, &Len);
    if (SUCCEEDED(sc)) {
        DBG(("FAILED (Test...FindNext 3): sc=%x, prefix=%s, buffer=%s, size=%d, length=%d\n",
             sc, pPrefix, pBuffer, BufSz, Len));
        return sc;
    }

    Len = 0;
    pPrefix = _T("/ws/");
    sc = pNs->v->FindNext(pNs, pPrefix, NULL, pBuffer, BufSz, &Len);
    if (FAILED(sc) || WSD_NSID_SZ != Len || 0 != _tcscmp(WSD_NSID, pBuffer)) {
        DBG(("FAILED (Test...FindNext 4): sc=%x, prefix=%s, buffer=%s, size=%d, length=%d\n",
             sc, pPrefix, pBuffer, BufSz, Len));
        return sc;
    }

    Len = 0;
    pPrefix = _T("/ws/discovery/");
    sc = pNs->v->FindNext(pNs, pPrefix, NULL, pBuffer, BufSz, &Len);
    if (FAILED(sc) || WSDE_NSID_SZ != Len || 0 != _tcscmp(WSDE_NSID, pBuffer)) {
        DBG(("FAILED (Test...FindNext 5): sc=%x, prefix=%s, buffer=%s, size=%d, length=%d\n",
             sc, pPrefix, pBuffer, BufSz, Len));
        return sc;
    }

    Len = 0;
    pPrefix = _T("/ws/discovery/");
    pPrevious = _T("endpoints");
    sc = pNs->v->FindNext(pNs, pPrefix, pPrevious, pBuffer, BufSz, &Len);
    if (FAILED(sc) || WSDT_NSID_SZ != Len || 0 != _tcscmp(WSDT_NSID, pBuffer)) {
        DBG(("FAILED (Test...FindNext 6): \
             \nsc=%x, prefix=%s, previous=%s, buffer=%s, size=%d, length=%d\n",
             sc, pPrefix, pPrevious, pBuffer, BufSz, Len));
        return sc;
    }

    Len = 0;
    pPrefix = _T("/ws/discovery/");
    pPrevious = _T("types");
    sc = pNs->v->FindNext(pNs, pPrefix, pPrevious, pBuffer, BufSz, &Len);
    if (FAILED(sc) || WSDS_NSID_SZ != Len || 0 != _tcscmp(WSDS_NSID, pBuffer)) {
        DBG(("FAILED (Test...FindNext 7): \
             \nsc=%x, prefix=%s, previous=%s, buffer=%s, size=%d, length=%d\n",
             sc, pPrefix, pPrevious, pBuffer, BufSz, Len));
        return sc;
    }

    Len = 0;
    pPrefix = _T("/ws/discovery/");
    pPrevious = _T("scopes");
    sc = pNs->v->FindNext(pNs, pPrefix, pPrevious, pBuffer, BufSz, &Len);
    if (S_NO_MORE_ENTRIES != sc || 0 != Len || '\0' != pBuffer[0]) {
        DBG(("FAILED (Test...FindNext 8): \
             \nsc=%x, prefix=%s, previous=%s, buffer=%s, size=%d, length=%d\n",
             sc, pPrefix, pPrevious, pBuffer, BufSz, Len));
        return sc;
    }

    return S_OK;
}