/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Tool to create Secure Extension images for eMIPS.
 */

#include "se.h"

#define VERBOSE(x) /* x */

extern void SeBuildHeader(PSE_HEADER Header, BOOL MustSwap,
                          UINT32 FirstChunkSize, UINT32 SecondChunkSize);
extern BOOL SeCreateDigest(PSE_HEADER Header, BOOL MustSwap,
                           void *FirstChunk, UINT32 FirstChunkSize,
                           void *SecondChunk, UINT32 SecondChunkSize,
                           void **DigestStart, UINT32 *DigestSize);

/* Operations for the "software side" of things, e.g. executable-image specific codes.
 */
typedef struct _SOFTWARE_SIDE {
    BOOL (*IsMyImage)
        (void *ImageStart, UINT32 ImageSize, BOOL *IsBigEndian);
    BOOL (*PatchImage)
        (void *ImageStart, UINT32 ImageSize, PSE_HEADER SeHeader);
} SOFTWARE_SIDE, *PSOFTWARE_SIDE;

extern PSOFTWARE_SIDE AllSwOps[]; /* forward, NULL-terminated */

/* Operations for the "hardware side" of things, e.g. configuration-file specific codes.
 */
typedef struct _HARDWARE_SIDE {
    BOOL (*IsMyImage)
        (void *ImageStart, UINT32 ImageSize);
    BOOL (*PatchImage)
        (void *ImageStart, UINT32 ImageSize, PSE_HEADER SeHeader);
} HARDWARE_SIDE, *PHARDWARE_SIDE;

extern PHARDWARE_SIDE AllHwOps[]; /* forward, NULL-terminated */

/* Utilities
 */
static UINT32 SwapWord(UINT32 w)
{
   return (((w & 0xff) << 24) |
           ((w & 0xff00) << 8) |
           ((w & 0xff0000) >> 8) |
           ((w & 0xff000000) >> 24));
}

static UINT16 SwapHalf(UINT16 h)
{
   return (((h & 0xff) << 8) |
           ((h & 0xff00) >> 8));
}

#define GetWord(_swap_,_w_)  ((_swap_) ? SwapWord(_w_) : (_w_))
#define GetHalf(_swap_,_h_)  ((_swap_) ? SwapHalf(_h_) : (_h_))
#define SetWord(_swap_,_a_,_v_) _a_ = ((_swap_) ? SwapWord(_v_) : (_v_))
#define SetHalf(_swap_,_a_,_v_) _a_ = ((_swap_) ? SwapHalf(_v_) : (_v_))

UINT32 ReadFile(char *FileName,
                void **FileBits)
{
    UINT32 FileSize = 0;
    FILE *File = NULL;

    /* Sanity */
    *FileBits = NULL;

    /* Open & check
     */
    File = fopen(FileName, "rb");
    if (File == NULL) {
        printf("Could not open %s for reading.\n", FileName);
        goto Out;
    }

    /* How big is it
     */
    fseek(File,0,SEEK_END);
    FileSize = ftell(File);
    fseek(File,0,SEEK_SET);

    /* Read it in
     */
    if (FileSize > 0) {
        char *Bits = malloc(FileSize);

        if (Bits == NULL) {
            printf("Outtamem??(%d bytes)\n", FileSize);
            FileSize = 0;
            goto Out;
        }

        if (1 != fread(Bits,FileSize,1,File)) {
            free(Bits);
            printf("Short read on %s (!all %d bytes)\n",FileName,FileSize);
            FileSize = 0;
            goto Out;
        }

        /* All is well
         */
        *FileBits = Bits;
    }

    /* Done
     */
 Out:
    if (File) fclose(File);
    return FileSize;
}

/* Main program
 */
BOOL BuildSeImage(char *OutputImageName,
                  char *SwImageName, UINT32 SwProperties,
                  char *HwImageName, UINT32 HwProperties,
                  UINT32 DigestProperties)
{
    BOOL r = FALSE, MustSwap = FALSE;
    PSOFTWARE_SIDE SwOps = NULL;
    PHARDWARE_SIDE HwOps = NULL;
    SE_HEADER SeHeader;
    FILE *WorkFile = NULL;
    int i;
    char *SwImageStart = NULL, *HwImageStart = NULL, *DigestStart = NULL;
    UINT32 SwImageSize, HwImageSize, DigestSize;

    /* Read the SW file in memory
     */
    SwImageSize = ReadFile(SwImageName, &SwImageStart);
    if (SwImageSize == 0) {
        printf("WARNING!! Could not read %s. Using empty file.\n", SwImageName);
        goto ReadHwImage;
    }

    /* See if we know this image
     */
    for (i = 0; (SwOps = AllSwOps[i]) != NULL; i++)
        if (SwOps->IsMyImage(SwImageStart, SwImageSize, &MustSwap))
            break;
    if (SwOps == NULL) {
        printf("Executable image %s not recognized, fatal.\n", SwImageName);
        goto Out;
    }

    /* Read the HW file in memory
     */
 ReadHwImage:
    HwImageSize = ReadFile(HwImageName, &HwImageStart);
    if (HwImageSize == 0) {
        printf("WARNING!! Could not read %s. Using empty file.\n", HwImageName);
        goto BuildHeader;
    }

    /* See if we know this image
     */
    for (i = 0; (HwOps = AllHwOps[i]) != NULL; i++)
        if (HwOps->IsMyImage(HwImageStart, HwImageSize))
            break;
    if (HwOps == NULL) {
        printf("Configuration bitfile %s not recognized, fatal.\n", HwImageName);
        goto Out;
    }

    /* Build the SE header out of what we know
     */
 BuildHeader:
    SeBuildHeader(&SeHeader, MustSwap, SwImageSize, HwImageSize);
    SeHeader.SwProperties = GetWord(MustSwap,SwProperties);
    SeHeader.HwProperties = GetWord(MustSwap,HwProperties);
    SeHeader.DigestProperties = GetWord(MustSwap,DigestProperties);

    /* Patch the images before we write them out
     */
    if (SwOps && !SwOps->PatchImage(SwImageStart, SwImageSize, &SeHeader)) {
        printf("SwPatchImage failed? fatal.\n");
        goto Out;
    }
    if (HwOps && !HwOps->PatchImage(HwImageStart, HwImageSize, &SeHeader)) {
        printf("HwPatchImage failed? fatal.\n");
        goto Out;
    }

    /* Now create the security digest
     * NB: Updates SeHeader.DigestSize
     */
    if (!SeCreateDigest(&SeHeader, MustSwap,
                        SwImageStart, SwImageSize,
                        HwImageStart, HwImageSize,
                        &DigestStart, &DigestSize)) {
        printf("CreateDigest failed? fatal.\n");
        goto Out;
    }

    /* Create the output file
     */
    WorkFile = fopen(OutputImageName, "wb");
    if (WorkFile == NULL) {
        printf("Could not open/create %s for writing\n", OutputImageName);
        goto Out;
    }

    /* Write out the pieces
     * Roundup each section to the next 32-byte boundary
     */
    if (SwImageSize) {
        fwrite(SwImageStart, SwImageSize, 1, WorkFile);
        SwImageSize = (SwImageSize + SECTION_ROUNDUP - 1) & ~(SECTION_ROUNDUP - 1);
        fseek(WorkFile, SwImageSize, SEEK_SET);
    }

    if (HwImageSize) {
        fwrite(HwImageStart, HwImageSize, 1, WorkFile);
        HwImageSize = (HwImageSize + SECTION_ROUNDUP - 1) & ~(SECTION_ROUNDUP - 1);
        fseek(WorkFile, SwImageSize + HwImageSize, SEEK_SET);
    }

    fwrite(DigestStart, DigestSize, 1, WorkFile);
    DigestSize = (DigestSize + SECTION_ROUNDUP - 1) & ~(SECTION_ROUNDUP - 1);
    fseek(WorkFile, SwImageSize + HwImageSize + DigestSize, SEEK_SET);

    fwrite(&SeHeader, sizeof SeHeader, 1, WorkFile);

    /* Done with file
     */
    fclose(WorkFile); WorkFile = NULL;

    /* All is well.
     */
    r = TRUE;
 Out:
    if (SwImageStart) free(SwImageStart);
    if (HwImageStart) free(HwImageStart);
    if (DigestStart) free(DigestStart);
    if (WorkFile) fclose(WorkFile);
    return r;
}

int main(int argc, char **argv)
{
    UINT32 HwFlags = 0;

 Retest:
    if (argc < 4) {
        printf("Usage: %s [-ph<hex-flags>] Output-SE-ImageName SoftwareImageName HwImageName\n", argv[0]);
        return -1;
    }

    if (_strnicmp(argv[1],"-ph",3) == 0) {
        HwFlags = strtoul(argv[1]+3,0,16);
        argc--, argv++;
        goto Retest;
    }

    return (BuildSeImage(argv[1], argv[2], 0, argv[3], HwFlags, 0) ? 0 : -1);
}

/* =============== SE ==============
 * SE header functions
 */
void SeBuildHeader(PSE_HEADER Header, BOOL MustSwap,
                   UINT32 FirstChunkSize, UINT32 SecondChunkSize)
{
    memset(Header,0,sizeof *Header);
    Header->MySize = GetWord(MustSwap,sizeof *Header);
    Header->MagicNumber = (MustSwap) ? SE_MAGIC_LE : SE_MAGIC_BE; /* A PC is.. */
    Header->SwImageOffset = 0;
    if (FirstChunkSize > 0) {
        Header->SwImageSize = GetWord(MustSwap,FirstChunkSize);
        FirstChunkSize = (FirstChunkSize + SECTION_ROUNDUP - 1) & ~(SECTION_ROUNDUP - 1);
    }
    if (SecondChunkSize > 0) {
        Header->HwImageOffset = GetWord(MustSwap,FirstChunkSize);
        Header->HwImageSize = GetWord(MustSwap,SecondChunkSize);
        SecondChunkSize = (SecondChunkSize + SECTION_ROUNDUP - 1) & ~(SECTION_ROUNDUP - 1);
    }
    SecondChunkSize += FirstChunkSize;
    Header->DigestOffset = GetWord(MustSwap,SecondChunkSize);
    /* size later */
}

/* TBD */
typedef struct _SE_DIGEST {
    char Bytes[32];
} SE_DIGEST, *PSE_DIGEST;

UINT32 Sum32(BOOL MustSwap, UINT32 *Start, UINT32 nWords, UINT32 Seed)
{
    UINT32 v;
    while (nWords-- > 0) {
        v = *Start++;
        v = GetWord(MustSwap,v);
        Seed += v;
    }
    return Seed;
}

BOOL SeCreateDigest(PSE_HEADER Header, BOOL MustSwap,
                    void *FirstChunk, UINT32 FirstChunkSize,
                    void *SecondChunk, UINT32 SecondChunkSize,
                    void **DigestStart, UINT32 *DigestSize)
{
    PSE_DIGEST Digest = malloc(sizeof *Digest);
    UINT32 Sum = 0;

    memset(Digest,0,sizeof *Digest);
    *DigestStart = Digest;
    *DigestSize  = sizeof *Digest;

    Header->DigestSize = GetWord(MustSwap,sizeof *Digest);

    Sum = Sum32( MustSwap, (UINT32 *)FirstChunk, FirstChunkSize / 4, Sum);
    VERBOSE(printf("Sum=%x\n",Sum));
    Sum = Sum32( MustSwap, (UINT32 *)SecondChunk, SecondChunkSize / 4, Sum);
    VERBOSE(printf("Sum=%x\n",Sum));
    Sum = Sum32( MustSwap, (UINT32 *)Header, sizeof(*Header) / 4, Sum);
    VERBOSE(printf("Sum=%x\n",Sum));

    Sum = GetWord(MustSwap,Sum);
    memcpy(Digest,&Sum,4);

    return TRUE;
}


/* Executable-specific functions start here
 */

/* ======================   ELF ========================
 * Describe the structure of the ELF object file format.
 * [Just as much of it as we need to understand]
 */

/* Size of file identifier in ELF header */
#define EI_NIDENT       16

/* Identification indexes for e_ident field in ELF header */
#define EI_MAG0     0         /* file ID byte 0 */
#define EI_MAG1     1         /* file ID byte 1 */
#define EI_MAG2     2         /* file ID byte 2 */
#define EI_MAG3     3         /* file ID byte 3 */
#define EI_CLASS    4         /* file class (capacity) */
#define EI_DATA     5         /* data encoding */
#define EI_VERSION  6         /* file version */
#define EI_PAD      7         /* start of padding bytes */
#define EI_NIDENT  16         /* size of e_ident[] */

/* "Magic number" in e_ident field of ELF header */
#define ELFMAG0     '\x7f'    /* e_ident[0] */
#define ELFMAG1     'E'       /* e_ident[1] */
#define ELFMAG2     'L'       /* e_ident[2] */
#define ELFMAG3     'F'       /* e_ident[3] */

/* Data encoding in e_ident[5] of ELF header */
#define ELFDATANONE   0       /* invalid data encoding */
#define ELFDATA2LSB   1       /* little-endian format */
#define ELFDATA2MSB   2       /* big-endian format */

#if (BYTE_ORDER==BIG_ENDIAN)
#define MYBYTEORDER ELFDATA2MSB
#else
#define MYBYTEORDER ELFDATA2LSB
#endif

/* Object file type in e_type field of ELF header */
#define ET_NONE     0         /* no file type */
#define ET_REL      1         /* relocatble file */
#define ET_EXEC     2         /* executable file */
#define ET_DYN      3         /* shared object file */
#define ET_CORE     4         /* core file */
#define ET_LOPROC   0xff00    /* processor-specific */
#define ET_HIPROC   0xffff    /* processor-specific */
/* Our processor-specific file type */
#define ET_SE       0xff02

/* Required architecture in e_machine field of ELF header */
#define EM_NONE     0         /* no machine */
#define EM_M32      1         /* AT&T WE 32100 */
#define EM_SPARC    2         /* SPARC */
#define EM_386      3         /* Intel 80386 */
#define EM_68K      4         /* Motorola 68000 */
#define EM_88K      5         /* Motorola 88000 */
#define EM_860      7         /* Intel 80860 */
#define EM_MIPS     8         /* MIPS RS3000 */
#define EM_ARM      40        /* Advanced RISC Machines ARM */

/* File header for ELF object file format */
typedef struct {
    UINT8 e_ident[EI_NIDENT];  /* identify file as object file */
    Elf32_Half    e_type;        /* file type */
    Elf32_Half    e_machine;     /* required architecture */
    Elf32_Word    e_version;     /* object file version */
    Elf32_Addr    e_entry;       /* entry point address */
    Elf32_Off     e_phoff;       /* program header table's file offset */
    Elf32_Off     e_shoff;       /* section header table's file offset */
    Elf32_Word    e_flags;       /* processor-specific flags */
    Elf32_Half    e_ehsize;      /* ELF header's size in bytes */
    Elf32_Half    e_phentsize;   /* size of entry in program header table */
    Elf32_Half    e_phnum;       /* no. of entries in program header table */
    Elf32_Half    e_shentsize;   /* size of entry in section header table */
    Elf32_Half    e_shnum;       /* no. of entries in section header table */
    Elf32_Half    e_shstrndx;    /* section hdr index of string table */
} Elf32_Ehdr;

/*
 * Check if we recognize an image as a valid ELF image
 */
BOOL IsElfImage(void *ImageStart, UINT32 ImageSize, BOOL *IsBigEndian)
{
    Elf32_Ehdr *Elf = (Elf32_Ehdr *)ImageStart;
    Elf32_Half m;
    BOOL MustSwap;
    char *s;

    /* Is this a valid ELF file?
     */
    MustSwap = (Elf->e_ident[EI_DATA] != MYBYTEORDER);
    m = GetHalf(MustSwap,Elf->e_machine);
    s = (char *)Elf->e_ident;
    if ((*s != ELFMAG0) || 
        (memcmp(s + 1, "ELF", 3) != 0)) {
        printf("Not a valid ELF file.\n");
        return FALSE;
    }

    /* Do we believe this machine type (optional)
     */
    if (m != EM_MIPS) {
        printf("Machine type in ELF image is not MIPS (x%x? != x%x). Proceeding anyways.\n",
               m,EM_MIPS);
    }

    /* Is the version compatible with our build ?
     */
#if defined(MMLITE_VERSION)
    if (GetWord(MustSwap,Elf->e_version) > MMLITE_VERSION) {
        printf("Error -- Incompatible MMLite version!\n");
        return FALSE;
    }
#endif

    /* All is well
     */
    *IsBigEndian = MustSwap;
    return TRUE;
}

BOOL PatchElfImage(void *ImageStart, UINT32 ImageSize, PSE_HEADER SeHeader)
{
    Elf32_Ehdr *Elf = (Elf32_Ehdr *)ImageStart;
    BOOL MustSwap = (Elf->e_ident[EI_DATA] != MYBYTEORDER);

    /* Easy :-) */
    Elf->e_type = GetHalf(MustSwap,ET_SE);
    return TRUE;
}

SOFTWARE_SIDE ElfImage = { IsElfImage, PatchElfImage };

/* List of all supported sw image types
 */
PSOFTWARE_SIDE AllSwOps[] = {
    &ElfImage,
    NULL
};

/* Configuration Bitfile-specific functions start here
 */

/* ======================   ACE ========================
 */
/*
 * Check if we recognize an image as a valid Xilinx System-ACE image
 */
BOOL IsAceImage(void *ImageStart, UINT32 ImageSize)
{
    static char AceHeader[] = 
"0000000000000000000000000000000000000000000000000000000000000000\n"
"Automatically generated. DO NOT MODIFY\n"
"Xilinx, Inc.\n"
"xmpmutil(tm) Multi-Function MPM Utility Version 1.01 (c) 2001 Xilinx, Inc.\n";

    if ((ImageSize < 513) || /* at least one by of config data */
        (memcmp(ImageStart,AceHeader,(sizeof AceHeader) - 1) != 0)) {
        printf("Not a valid ACE file.\n");
        return FALSE;
    }
    return TRUE;
}

BOOL PatchAceImage(void *ImageStart, UINT32 ImageSize, PSE_HEADER SeHeader)
{
    /* Nothing to do */
    return TRUE;
}

HARDWARE_SIDE AceImage = { IsAceImage, PatchAceImage };

/* List of all supported hw image types
 */
PHARDWARE_SIDE AllHwOps[] = {
    &AceImage,
    NULL
};

