/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Tool to dump Secure Extension images for eMIPS.
 */

#include "se.h"

/* Utilities
 */
static UINT32 SwapWord(UINT32 w)
{
   return (((w & 0xff) << 24) |
           ((w & 0xff00) << 8) |
           ((w & 0xff0000) >> 8) |
           ((w & 0xff000000) >> 24));
}

static UINT16 SwapHalf(UINT16 h)
{
   return (((h & 0xff) << 8) |
           ((h & 0xff00) >> 8));
}

#define GetWord(_swap_,_w_)  ((_swap_) ? SwapWord(_w_) : (_w_))
#define GetHalf(_swap_,_h_)  ((_swap_) ? SwapHalf(_h_) : (_h_))
#define SetWord(_swap_,_a_,_v_) _a_ = ((_swap_) ? SwapWord(_v_) : (_v_))
#define SetHalf(_swap_,_a_,_v_) _a_ = ((_swap_) ? SwapHalf(_v_) : (_v_))

UINT32 ReadFile(char *FileName,
                void **FileBits)
{
    UINT32 FileSize = 0;
    FILE *File = NULL;

    /* Sanity */
    *FileBits = NULL;

    /* Open & check
     */
    File = fopen(FileName, "rb");
    if (File == NULL) {
        printf("Could not open %s for reading.\n", FileName);
        goto Out;
    }

    /* How big is it
     */
    fseek(File,0,SEEK_END);
    FileSize = ftell(File);
    fseek(File,0,SEEK_SET);

    /* Read it in
     */
    if (FileSize > 0) {
        char *Bits = malloc(FileSize);

        if (Bits == NULL) {
            printf("Outtamem??(%d bytes)\n", FileSize);
            FileSize = 0;
            goto Out;
        }

        if (1 != fread(Bits,FileSize,1,File)) {
            free(Bits);
            printf("Short read on %s (!all %d bytes)\n",FileName,FileSize);
            FileSize = 0;
            goto Out;
        }

        /* All is well
         */
        *FileBits = Bits;
    }

    /* Done
     */
 Out:
    if (File) fclose(File);
    return FileSize;
}

void RawDump( UINT32 *Start, UINT32 nWords, char *Banner)
{
    if (Banner) printf(Banner);
    while (nWords-- > 0) {
        printf("%08x ", *Start++);
        if ((nWords & 7) == 0) printf("\n");
    }
}

char *ListSwProperties(UINT32 p)
{
    return "";
}

char *ListHwProperties(UINT32 p)
{
    static char buffer[128];

    buffer[0] = 0;
    if (p & SE_HW_PRIVILEGED)
        strcat(buffer,"PRIVILEDGED ");
    if (p & SE_HW_PERIPHERAL) 
        strcat(buffer,"PERIPHERAL ");
    if (p & SE_HW_PRIORITY_MASK) {
        strcat(buffer,"PRIORITY=");
        switch (p & SE_HW_PRIORITY_MASK) {
        case SE_HW_PRIORITY_1:
            strcat(buffer,"1"); break;
        case SE_HW_PRIORITY_2:
            strcat(buffer,"2"); break;
        case SE_HW_HIGH_PRIORITY:
            strcat(buffer,"HIGH "); break;
        default:
            strcat(buffer,"?? "); break;
        }
    }

    return buffer;
}

char *ListDigestProperties(UINT32 p)
{
    return "";
}

int main(int argc, char **argv)
{
    UINT32 FileSize, ChunkSize, x;
    char *FileData, *ChunkStart;
    PSE_HEADER SeHeader;
    BOOL MustSwap;

    if (argc < 2) {
        printf("Usage: %s SeImageFileName\n", argv[0]);
        return -1;
    }

    FileSize = ReadFile(argv[1], &FileData);
    if ((FileSize == 0) || (FileData == NULL)) {
        printf("Nothing to do.\n");
        return -2;
    }

    SeHeader = ((PSE_HEADER)(FileData + FileSize)) - 1;

    if (SeHeader->MagicNumber == SE_MAGIC_LE)
        MustSwap = TRUE;
    else
    if (SeHeader->MagicNumber == SE_MAGIC_BE)
        MustSwap = FALSE;
    else {
        printf("SeHeader: Bad MagicNumber x%x\n", SeHeader->MagicNumber);
        return -3;
    }
    printf("SeHeader is %sEndian:\n", (MustSwap) ? "Big" : "Little");
    printf("\tSwImageOffset: x%x\n", GetWord(MustSwap,SeHeader->SwImageOffset));
    printf("\tSwImageSize: x%x\n", GetWord(MustSwap,SeHeader->SwImageSize));
    x = GetWord(MustSwap,SeHeader->SwProperties);
    printf("\tSwProperties: x%x %s\n", x, ListSwProperties(x));
    printf("\tHwImageOffset: x%x\n", GetWord(MustSwap,SeHeader->HwImageOffset));
    printf("\tHwImageSize: x%x\n", GetWord(MustSwap,SeHeader->HwImageSize));
    x = GetWord(MustSwap,SeHeader->HwProperties);
    printf("\tHwProperties: x%x %s\n", x, ListHwProperties(x));
    printf("\tDigestOffset: x%x\n", GetWord(MustSwap,SeHeader->DigestOffset));
    printf("\tDigestSize: x%x\n", GetWord(MustSwap,SeHeader->DigestSize));
    x = GetWord(MustSwap,SeHeader->DigestProperties);
    printf("\tDigestProperties: x%x %s\n", x, ListDigestProperties(x));
    printf("\tMySize: x%x\n", GetWord(MustSwap,SeHeader->MySize));
    {
        union {
            UINT32 AsWord;
            char AsBytes[5];
        } u;
        u.AsWord = GetWord(MustSwap,SeHeader->MagicNumber);
        u.AsBytes[4] = 0;
        printf("\tMagicNumber: (x%x) %x '%s'\n", SeHeader->MagicNumber, u.AsWord, u.AsBytes);
    }
    
    ChunkStart = FileData + GetWord(MustSwap,SeHeader->SwImageOffset);
    ChunkSize = GetWord(MustSwap,SeHeader->SwImageSize);
    ChunkSize = min(ChunkSize, 512);
    RawDump( (UINT32 *)ChunkStart, ChunkSize / 4, "SwImage:\n");
    printf("...\n");

    ChunkStart = FileData + GetWord(MustSwap,SeHeader->HwImageOffset);
    ChunkSize = GetWord(MustSwap,SeHeader->HwImageSize);
    ChunkSize = min(ChunkSize, 512);
    RawDump( (UINT32 *)ChunkStart, ChunkSize / 4, "HwImage:\n");
    printf("...\n");

    ChunkStart = FileData + GetWord(MustSwap,SeHeader->DigestOffset);
    ChunkSize = GetWord(MustSwap,SeHeader->DigestSize);
    RawDump( (UINT32 *)ChunkStart, ChunkSize / 4, "Digest:\n");
    {
        union {
            UINT32 AsWord;
            char AsBytes[4];
        } u;
        memcpy(&u,ChunkStart,4);
        printf("Checksum: %x\n", GetWord(MustSwap,u.AsWord));
    }

    printf("Ok.\n");
    return 0;
}

