/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Poor Man's Version control for AIF images.
 *
 * Uses the reserved fields in the header, which is probably non-standard.
 */

#include <windows.h>
#include <stdio.h>

#if 0
typedef unsigned long  UINT32;
typedef   signed long   INT32;
typedef unsigned short UINT16;
typedef   signed short  INT16;
typedef unsigned char  UINT8;
#endif

/* AIF header, 32 words total
 */
typedef struct _AIF_HEADER {
    UINT32  BlDecompress;
    UINT32  BlSelfReloc;
    UINT32  BlDebugInit;
    UINT32  BlEntryPoint;   /* if non exec, just EntryPoint */
    UINT32  ExitInstruction;
    UINT32  TextSize;
    UINT32  DataSize;
    UINT32  DebugSize;
    UINT32  BssSize;
    UINT32  DebugType;
    UINT32  ImageBase;
    UINT32  WorkSpace;
    UINT32  AddressingMode;
    UINT32  DataBase;
    UINT32  Reserved[2];
#define VersionNumber Reserved[1]
    UINT32  DebugInitInstruction;
    UINT32  ZeroInitCode[15];
} AIF_HEADER;

#define EV_CURRENT 1

/* File copy utility, sigh.
 */
BOOL FileCopy(FILE *fIn, FILE *fOut)
{
  int in, out, n, w;
#define CHUNK (64*1024)
  static UINT32 Buffer[CHUNK/sizeof(UINT32)];

  /* Bypass the buffering inside the STDIO library, sigh.
   */
  fflush(fOut);
  in = _fileno(fIn);
  out = _fileno(fOut);

  /* urgh.. */
  _lseek(in,sizeof(AIF_HEADER),SEEK_SET);
  _lseek(out,sizeof(AIF_HEADER),SEEK_SET);

  /* Loop while we got bytes
   */
  while ((n = _read(in,Buffer,CHUNK)) > 0) {
    w = _write(out,Buffer,n);
    if (w != n) {
      printf("Error: FileCopy, write of %d bytes got %d\n",n,w);
      return FALSE;
    }
  }
  return TRUE;
}


/* Main driving program
 */
int main(int argc, char **argv)
{
  char *pgm = argv[0];
  AIF_HEADER Header;
  UINT32 ImageVersion = EV_CURRENT;
  BOOL JustPrint = TRUE, Quiet = FALSE, Copy = FALSE;
  FILE *fIn, *fOut;
  UINT n, RetVal = 1;

  /* Parse arguments, if any
   */
  while ((argc > 1) && (argv[1][0] == '-')) {
    switch (argv[1][1]) {
    case 'V':
      JustPrint = !JustPrint;
      ImageVersion = atoi(argv[2]);
      argv += 2;
      argc -= 2;
      break;
    case 'q':
      Quiet = TRUE;
      argv += 1;
      argc -= 1;
      break;
    default:
      printf("Unknown option '-%c'\n",argv[1][1]);
      goto Usage;
    }
  }

  /* Make sure we got a file to work with
   */
  if (argc == 1) {
  Usage:
    printf("Usage: %s [-q] [-V DecimalVersionNumber] InAifFilename [OutAifFilename]\n", pgm);
    printf(" Prints the version number of an ARM AIF executable image\n");
    printf(" -q keeps quiet, for use in automated builds.\n");
    printf(" -V sets the version number of the image. [Default: %d]\n",
           EV_CURRENT);
    return RetVal;
  }

  /* Inplace or with copy ?
   */
  if (argc > 2)
    Copy = TRUE;

  /* Open the input file(s), bail on error
   */
  fIn = fopen(argv[1],
              (Copy || JustPrint) ? "rb" : "r+b");
  if (fIn == NULL) {
    printf("Error: could not open file %s for %s\n",
           argv[1],
           (Copy || JustPrint) ? "reading" : "reading+writing");
    return RetVal;
  }

  /* Open the output file, bail on error
   */
  if (Copy) {
    fOut = fopen(argv[2],"wb+");
    if (fOut == NULL) {
      printf("Error: could not open file %s for writing\n", argv[2]);
      goto Done;
    }
  } else
    fOut = fIn;

  /* Read in the header, make sure its big enough
   */
  n = fread(&Header,1,sizeof Header,fIn);
  if (n != sizeof Header) {
    printf("File %s is too short for an AIF image (%d < %d min)\n",
           argv[1], n, sizeof Header);
    goto Done;
  }

  /* If we intend to set the version do so now.
   */
  if (!JustPrint) {
    /* Change the header info we use for versioning
     */
    Header.VersionNumber = ImageVersion;

    if (Copy) {
      /* Write the header
       */
      n = fwrite(&Header,1,sizeof Header,fOut);
      if (n != sizeof Header) {
        printf("Error: writing %d bytes to file %s returned %d.\n",
               sizeof Header, argv[2], n);
        goto Done;
      }

      if (!FileCopy(fIn,fOut))
        goto Done;

    } else {

      /* Seek back to the beginning
       */
      n = fseek(fIn,0,SEEK_SET);
      if (n != 0) {
        printf("Error: failed to seek back to start of file, so weird...\n");
        goto Done;
      }

      /* Rewrite it
       */
      n = fwrite(&Header,1,sizeof Header,fIn);
      if (n != sizeof Header) {
        printf("Error: writing %d bytes to file %s returned %d.\n",
               sizeof Header, argv[1], n);
        goto Done;
      }

    }

    /* Say we did it ok
     */
    if (!Quiet)
      printf("AIF image file %s set to ", argv[Copy ? 2 : 1]);
  }

  /* Either way, say what the version is
   */
  if (!Quiet)
    printf("Version %d decimal, 0x%x hex\n",
           Header.VersionNumber, Header.VersionNumber);

  /* Done ok
   */
  RetVal = 0;

 Done:
  fclose(fIn);
  if (fOut != fIn)
      fclose(fOut);

  return RetVal;
}
