/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Poor Man's Version control for ELF images.
 *
 * Uses the e_version fields in the header, which is probably non-standard.
 */

#include <windows.h>
#include <stdio.h>

#if 0
typedef unsigned long  UINT32;
typedef   signed long   INT32;
#endif
typedef unsigned short UINT16;
typedef   signed short  INT16;
typedef unsigned char  UINT8;

/* Types defined in the ELF specification
 */
typedef UINT32 Elf32_Addr, Elf32_Off, Elf32_Word;
typedef INT32 Elf32_Sword;
typedef UINT16 Elf32_Half;

/* Identification indexes for e_ident field in ELF header
 */
#define EI_MAG0     0         /* file ID byte 0 */
#define EI_MAG1     1         /* file ID byte 1 */
#define EI_MAG2     2         /* file ID byte 2 */
#define EI_MAG3     3         /* file ID byte 3 */
#define EI_CLASS    4         /* file class (capacity) */
#define EI_DATA     5         /* data encoding */
#define EI_VERSION  6         /* file version */
#define EI_PAD      7         /* start of padding bytes */
#define EI_NIDENT  16         /* size of e_ident[] */

/* Data encoding in e_ident[5] of ELF header
 */
#define ELFDATANONE   0       /* invalid data encoding */
#define ELFDATA2LSB   1       /* little-endian format */
#define ELFDATA2MSB   2       /* big-endian format */

/* File header for ELF object file format
 */
typedef struct {
    UINT8 e_ident[EI_NIDENT];  /* identify file as object file */
    Elf32_Half    e_type;        /* file type */
    Elf32_Half    e_machine;     /* required architecture */
    Elf32_Word    e_version;     /* object file version */
    Elf32_Addr    e_entry;       /* entry point address */
    Elf32_Off     e_phoff;       /* program header table's file offset */
    Elf32_Off     e_shoff;       /* section header table's file offset */
    Elf32_Word    e_flags;       /* processor-specific flags */
    Elf32_Half    e_ehsize;      /* ELF header's size in bytes */
    Elf32_Half    e_phentsize;   /* size of entry in program header table */
    Elf32_Half    e_phnum;       /* no. of entries in program header table */
    Elf32_Half    e_shentsize;   /* size of entry in section header table */
    Elf32_Half    e_shnum;       /* no. of entries in section header table */
    Elf32_Half    e_shstrndx;    /* section hdr index of string table */
} Elf32_Ehdr;

#define EV_CURRENT 1

/* This is under windows, therefore our byteorder is LITTLE ENDIAN
 */
#define MYBYTEORDER ELFDATA2LSB

/* Utilities
 */
UINT32 SwapWord(UINT32 w)
{
   return (((w & 0xff) << 24) |
           ((w & 0xff00) << 8) |
           ((w & 0xff0000) >> 8) |
           ((w & 0xff000000) >> 24));
}

#define GetWord(_swap_,_w_)  ((_swap_) ? SwapWord(_w_) : (_w_))

/* File copy utility, sigh.
 */
BOOL FileCopy(FILE *fIn, FILE *fOut)
{
  int in, out, n, w;
#define CHUNK (64*1024)
  static UINT32 Buffer[CHUNK/sizeof(UINT32)];

  /* Bypass the buffering inside the STDIO library, sigh.
   */
  fflush(fOut);
  in = _fileno(fIn);
  out = _fileno(fOut);

  /* urgh.. */
  _lseek(in,sizeof(Elf32_Ehdr),SEEK_SET);
  _lseek(out,sizeof(Elf32_Ehdr),SEEK_SET);

  /* Loop while we got bytes
   */
  while ((n = _read(in,Buffer,CHUNK)) > 0) {
    w = _write(out,Buffer,n);
    if (w != n) {
      printf("Error: FileCopy, write of %d bytes got %d\n",n,w);
      return FALSE;
    }
  }
  return TRUE;
}


/* Main driving program
 */
int main(int argc, char **argv)
{
  char *pgm = argv[0];
  Elf32_Ehdr Header;
  BOOL MustSwap;
  UINT32 ImageVersion = EV_CURRENT;
  BOOL JustPrint = TRUE, Quiet = FALSE, Copy = FALSE;
  FILE *fIn, *fOut;
  UINT n, RetVal = 1;

  /* Parse arguments, if any
   */
  while ((argc > 1) && (argv[1][0] == '-')) {
    switch (argv[1][1]) {
    case 'V':
      JustPrint = !JustPrint;
      ImageVersion = atoi(argv[2]);
      argv += 2;
      argc -= 2;
      break;
    case 'q':
      Quiet = TRUE;
      argv += 1;
      argc -= 1;
      break;
    default:
      printf("Unknown option '-%c'\n",argv[1][1]);
      goto Usage;
    }
  }

  /* Make sure we got a file to work with
   */
  if (argc == 1) {
  Usage:
    printf("Usage: %s [-q] [-V DecimalVersionNumber] InElfFilename [OutElfFilename]\n", pgm);
    printf(" Prints the version number of an ELF32 executable image\n");
    printf(" -q keeps quiet, for use in automated builds.\n");
    printf(" -V sets the version number of the image. [Default: %d]\n",
           EV_CURRENT);
    return RetVal;
  }

  /* Inplace or with copy ?
   */
  if (argc > 2)
    Copy = TRUE;

  /* Open the input file(s), bail on error
   */
  fIn = fopen(argv[1],
              (Copy || JustPrint) ? "rb" : "r+b");
  if (fIn == NULL) {
    printf("Error: could not open file %s for %s\n",
           argv[1],
           (Copy || JustPrint) ? "reading" : "reading+writing");
    return RetVal;
  }

  /* Open the output file, bail on error
   */
  if (Copy) {
    fOut = fopen(argv[2],"wb+");
    if (fOut == NULL) {
      printf("Error: could not open file %s for writing\n", argv[2]);
      goto Done;
    }
  } else
    fOut = fIn;

  /* Read in the header, make sure its big enough
   */
  n = fread(&Header,1,sizeof Header,fIn);
  if (n != sizeof Header) {
    printf("File %s is too short for an ELF image (%d < %d min)\n",
           argv[1], n, sizeof Header);
    goto Done;
  }

  /* Check its a valid file, see if we must byteswap
   */
  MustSwap = (Header.e_ident[EI_DATA] != MYBYTEORDER);

  /* If we intend to set the version do so now.
   */
  if (!JustPrint) {
    /* Change the header info we use for versioning
     */
    Header.e_version = GetWord(MustSwap,ImageVersion);

    if (Copy) {
      /* Write the header
       */
      n = fwrite(&Header,1,sizeof Header,fOut);
      if (n != sizeof Header) {
        printf("Error: writing %d bytes to file %s returned %d.\n",
               sizeof Header, argv[2], n);
        goto Done;
      }

      if (!FileCopy(fIn,fOut))
        goto Done;

    } else {

      /* Seek back to the beginning
       */
      n = fseek(fIn,0,SEEK_SET);
      if (n != 0) {
        printf("Error: failed to seek back to start of file, so weird...\n");
        goto Done;
      }

      /* Rewrite it
       */
      n = fwrite(&Header,1,sizeof Header,fIn);
      if (n != sizeof Header) {
        printf("Error: writing %d bytes to file %s returned %d.\n",
               sizeof Header, argv[1], n);
        goto Done;
      }

    }

    /* Say we did it ok
     */
    if (!Quiet)
      printf("ELF image file %s set to ", argv[Copy ? 2 : 1]);
  }

  /* Either way, say what the version is
   */
  ImageVersion = GetWord(MustSwap,Header.e_version);
  if (!Quiet)
    printf("Version %d decimal, 0x%x hex\n", ImageVersion, ImageVersion);

  /* Done ok
   */
  RetVal = 0;

 Done:
  fclose(fIn);
  if (fOut != fIn)
      fclose(fOut);

  return RetVal;
}
