/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Poor Man's Version control for LIFE images.
 *
 * Uses the reserved fields in the header, which is probably non-standard.
 */

#include <windows.h>
#include <stdio.h>

#if 0
typedef unsigned long  UINT32;
typedef   signed long   INT32;
typedef unsigned short UINT16;
typedef   signed short  INT16;
#endif
typedef unsigned char  UINT8;


#define C4ToNum(_c4_,_n_) (_n_) = ((_c4_)[3] | ((_c4_)[2] << 8) | \
                   ((_c4_)[1] << 16) | ((_c4_)[0] << 24))
#define C3ToNum(_c4_,_n_) (_n_) = (((_c4_)[2] << 0) | \
                   ((_c4_)[1] << 8) | ((_c4_)[0] << 16))
#define C2ToNum(_c4_,_n_) (_n_) = (((_c4_)[1] << 0) | ((_c4_)[0] << 8))
#define NumToC4(_n_,_c4_) { (_c4_)[3] = ((_n_) >>  0) & 0xff; \
                            (_c4_)[2] = ((_n_) >>  8) & 0xff; \
                            (_c4_)[1] = ((_n_) >> 16) & 0xff; \
                            (_c4_)[0] = ((_n_) >> 24) & 0xff; \
              }

typedef struct _LIFE_HEADER {
    UINT8 Signature[8];  /*  "LIFE_Obj"  */
    UINT8 Version[2];
    UINT8 VersionStr[3];
    UINT8 MachineStr[3];
    UINT8 Check[4];
    UINT8 Flags[2];
    UINT8 StartSection;
    UINT8 StartHigh[4];
    UINT8 StartLow[4];
    UINT8 BinaryOffset[4];
    UINT8 SourceOffset[4];
    UINT8 Length[4];
    UINT8 StrTableOffset[4];
    UINT8 StrTableSize[3];
    UINT8 SctTableOffset[4];
    UINT8 SctTableSize[1];
    UINT8 SymTableOffset[4];
    UINT8 SymTableSize[3];
    UINT8 HistoryOffset[4];
    UINT8 HistorySize[4];
    UINT8 SrcTableOffset[4];
    UINT8 SrcTableSize[2];
    UINT8 ScatterTableOffset[4];
    UINT8 ScatterTableSize[4];
#if 0
    UINT8 pad[44]; /* that makes it 128 bytes total */
#else
    UINT8 Undocumented[12]; /* only the first 4 really, but lets play safe */
    UINT8 VersionNumber[4];
    UINT8 pad[40-12]; /* that makes it 128 bytes total */
#endif
} LIFE_HEADER;

#define EV_CURRENT 1

/* File copy utility, sigh.
 */
BOOL FileCopy(FILE *fIn, FILE *fOut)
{
  int in, out, n, w;
#define CHUNK (64*1024)
  static UINT32 Buffer[CHUNK/sizeof(UINT32)];

  /* Bypass the buffering inside the STDIO library, sigh.
   */
  fflush(fOut);
  in = _fileno(fIn);
  out = _fileno(fOut);

  /* urgh.. */
  _lseek(in,sizeof(LIFE_HEADER),SEEK_SET);
  _lseek(out,sizeof(LIFE_HEADER),SEEK_SET);

  /* Loop while we got bytes
   */
  while ((n = _read(in,Buffer,CHUNK)) > 0) {
    w = _write(out,Buffer,n);
    if (w != n) {
      printf("Error: FileCopy, write of %d bytes got %d\n",n,w);
      return FALSE;
    }
  }
  return TRUE;
}


/* Main driving program
 */
int main(int argc, char **argv)
{
  char *pgm = argv[0];
  LIFE_HEADER Header;
  UINT32 ImageVersion = EV_CURRENT;
  BOOL JustPrint = TRUE, Quiet = FALSE, Copy = FALSE;
  FILE *fIn, *fOut;
  UINT n, RetVal = 1;

  /* Parse arguments, if any
   */
  while ((argc > 1) && (argv[1][0] == '-')) {
    switch (argv[1][1]) {
    case 'V':
      JustPrint = !JustPrint;
      ImageVersion = atoi(argv[2]);
      argv += 2;
      argc -= 2;
      break;
    case 'q':
      Quiet = TRUE;
      argv += 1;
      argc -= 1;
      break;
    default:
      printf("Unknown option '-%c'\n",argv[1][1]);
      goto Usage;
    }
  }

  /* Make sure we got a file to work with
   */
  if (argc == 1) {
  Usage:
    printf("Usage: %s [-q] [-V DecimalVersionNumber] InAifFilename [OutAifFilename]\n", pgm);
    printf(" Prints the version number of TriMedia LIFE executable images\n");
    printf(" -q keeps quiet, for use in automated builds.\n");
    printf(" -V sets the version number of the image. [Default: %d]\n",
           EV_CURRENT);
    return RetVal;
  }

  /* Inplace or with copy ?
   */
  if (argc > 2)
    Copy = TRUE;

  /* Open the input file(s), bail on error
   */
  fIn = fopen(argv[1],
              (Copy || JustPrint) ? "rb" : "r+b");
  if (fIn == NULL) {
    printf("Error: could not open file %s for %s\n",
           argv[1],
           (Copy || JustPrint) ? "reading" : "reading+writing");
    return RetVal;
  }

  /* Open the output file, bail on error
   */
  if (Copy) {
    fOut = fopen(argv[2],"wb+");
    if (fOut == NULL) {
      printf("Error: could not open file %s for writing\n", argv[2]);
      goto Done;
    }
  } else
    fOut = fIn;

  /* Read in the header, make sure its big enough
   */
  n = fread(&Header,1,sizeof Header,fIn);
  if (n != sizeof Header) {
    printf("File %s is too short for an LIFE image (%d < %d min)\n",
           argv[1], n, sizeof Header);
    goto Done;
  }

  /* If we intend to set the version do so now.
   */
  if (!JustPrint) {
    /* Change the header info we use for versioning
     */
    NumToC4(ImageVersion,Header.VersionNumber);

    if (Copy) {
      /* Write the header
       */
      n = fwrite(&Header,1,sizeof Header,fOut);
      if (n != sizeof Header) {
        printf("Error: writing %d bytes to file %s returned %d.\n",
               sizeof Header, argv[2], n);
        goto Done;
      }

      if (!FileCopy(fIn,fOut))
        goto Done;

    } else {

      /* Seek back to the beginning
       */
      n = fseek(fIn,0,SEEK_SET);
      if (n != 0) {
        printf("Error: failed to seek back to start of file, so weird...\n");
        goto Done;
      }

      /* Rewrite it
       */
      n = fwrite(&Header,1,sizeof Header,fIn);
      if (n != sizeof Header) {
        printf("Error: writing %d bytes to file %s returned %d.\n",
               sizeof Header, argv[1], n);
        goto Done;
      }

    }

    /* Say we did it ok
     */
    if (!Quiet)
      printf("LIFE image file %s set to ", argv[Copy ? 2 : 1]);
  }

  /* Either way, say what the version is
   */
  C4ToNum(Header.VersionNumber,ImageVersion);
  if (!Quiet)
    printf("Version %d decimal, 0x%x hex\n",ImageVersion,ImageVersion);

  /* Done ok
   */
  RetVal = 0;

 Done:
  fclose(fIn);
  if (fOut != fIn)
      fclose(fOut);

  return RetVal;
}
