/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * This program takes a BOOT image and builds a BIN image,
 * it just adds a jump instruction at the start of the file.
 */
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include <windows.h>
#include <assert.h>

#define T_ARM 0
#define T_X86 1
#define T_PPC 2

#define _bs32(_n_)  ((((_n_)&0xff)<<24) | \
                     (((_n_)&0xff00)<<8) | \
                     (((_n_)&0xff0000)>>8) | \
                     (((_n_)&0xff000000)>>24))
int t_type;

void quit(
    int     exit_code,
    char    *message,
    void    *arg);
int copy_them(
    int     infile,
    int     outfile);

int __cdecl main(
    int     argc,
    char    **argv)
{
    int fi, fdos, fo, ret;

    if ((argc < 4)|| (argv[1][0] != '-'))
        goto Usage;

    if (_stricmp(argv[1],"-arm") == 0)
        t_type = T_ARM;
    else if (_stricmp(argv[1],"-x86") == 0)
        t_type = T_X86;
    else if (_stricmp(argv[1],"-ppc") == 0)
        t_type = T_PPC;
    else
        goto Usage;

    fi = open(argv[2], O_RDONLY|O_BINARY, 0);
    if (fi < 0)
        quit(1, "Could not open %s for reading\n", argv[2]);
    fo = open(argv[3], O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0644);
    if (fo < 0)
        quit(2, "Could not open %s for writing\n", argv[3]);

    ret = copy_them(fi, fo);

    close(fi); close(fo);
    return(ret);
 Usage:
    quit(-1, "Usage: %s -[arm|x86|ppc] exe_image output_bin_image\n",
         argv[0]);
    return 0;
}

void quit(
    int     exit_code,
    char    *message,
    void    *arg)
{
    fprintf(stderr, message, arg);
    exit(exit_code);
}

typedef struct {
    DWORD        adwConfounder[2];   // Random confounder for this block
    DWORD        dwSize;             // Size of this block (header and contents)
    DWORD        dwEntryPoint;       // Offset to entry point or beginning of content
    DWORD        dwType;             // See BOOT_INFO_TYPE_*
#define BOOT_INFO_TYPE_0BL '0Blp'
    DWORD        dwSizeOfTemp;       // Maximum size of temporary storage required
    DWORD        dwAddrOfTemp;       // Address where this block will place temporaries
    DWORD        dwAddrOfThisOrNext; // Address where this block should be loaded (or next block is stored)
} BOOT_INFO_BASE;

int file_copy(
    int     infile,
    int     outfile,
    unsigned long int   bytes);
 
int copy_them(
    int     infile,
    int     outfile)
{
#define MAX32HEADER sizeof(BOOT_INFO_BASE)
    BOOT_INFO_BASE   *f32 = (void *) malloc(MAX32HEADER);
    int ret;
    char *msg;
    UINT32 jumpy = 0;

    assert(f32);

    /*
     * Read&check in the BOOT file header
     */
    msg = "Bad BOOT image";
    ret = _read(infile, f32, MAX32HEADER);
    if (ret != MAX32HEADER) goto out;

printf("hdr: s=%x e=%x t=%x ts=%x ta=%x n=%x\n",
       f32->dwSize,
       f32->dwEntryPoint,
       f32->dwType,
       f32->dwSizeOfTemp,
       f32->dwAddrOfTemp,
       f32->dwAddrOfThisOrNext);

    if (t_type == T_PPC)
        jumpy = _bs32(BOOT_INFO_TYPE_0BL);
    else
        jumpy = BOOT_INFO_TYPE_0BL;
    if (f32->dwType != jumpy)
        goto out;

    /*
     * Fix the out header
     */
    jumpy = f32->dwEntryPoint;

    switch (t_type) {
    case T_ARM:
        /* branch forward to, PC-relative */
        jumpy = 0xEA000000 | ((jumpy >> 2) - 8);
        *(UINT32*)f32 = jumpy;
        break;
    case T_X86:
        {
            char *pcode = (char *)f32;
            *pcode++ = 0xe9; /* jump near */
            *(UINT32*)pcode = jumpy - 5;
        }
        break;
    case T_PPC:
        /* branch forward to, PC-relative */
        jumpy = 0x48000000 | (_bs32(jumpy) & 0x03ffffff);
        /* PPC is BE code */
        jumpy = _bs32(jumpy);
        *(UINT32*)f32 = jumpy;
        break;
    default:
        quit(-99,"Shouldnt be here?!?!?\n", 0);
        return -1;
    }

    /*
     * Now do the writing, header first
     */
    msg = "Could not write out header";
    ret = write(outfile, f32, MAX32HEADER);
    if (ret != MAX32HEADER) goto out;

    // now copy the rest of the EXE image
    msg = "Could not copy the image";
    ret = file_copy( infile, outfile, _filelength(infile) - MAX32HEADER);
    if (ret) goto out; 

    msg = NULL;
    ret = 0;
out:
    if (msg != NULL)
        fprintf(stderr, "join_them: %s\n", msg);
    return ret;
}

int file_copy(
    int     infile,
    int     outfile,
    unsigned long int   bytes)
{
    unsigned char   *buf;
    unsigned int copy_size;
    int wsize, ret;
#define _64k (64*1024)
    copy_size = bytes < _64k ? bytes : _64k;

    buf = (unsigned char *) malloc( copy_size);

    do {
        wsize = read(infile, buf, copy_size);
        ret = wsize;
        if (ret <= 0)
            break;

        ret = write(outfile, buf, wsize);
        if (ret != wsize)
            break;

        bytes -= wsize;
        if (bytes < copy_size)
            copy_size = bytes;
    } while (bytes > 0);
    return bytes;
}
