/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * This program takes an EXE NT image and a DOS EXE
 * image and builds a joint DOS EXE image.
 * In this DOS format the image can be loaded
 * either from DOS or from a bootstrap loader.
 */
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include <windows.h>
#include <assert.h>

#ifndef TARGET_ALIGN
#define TARGET_ALIGN    512
#endif

unsigned int    t_align = TARGET_ALIGN;

void quit(
    int     exit_code,
    char    *message,
    void    *arg);
int join_them(
    int     infile,
    int     dos_infile,
    int     outfile);

int __cdecl main(
    int     argc,
    char    **argv)
{
    int fi, fdos, fo, ret;

    if (argc < 3)
        quit(-1, "Usage: %s exe32_nt_image exe16_dos_image [output_image]\n",
             argv[0]);
    if (argc < 4) argv[3] = "a.exe";

    fi = open(argv[1], O_RDONLY|O_BINARY, 0);
    if (fi < 0)
        quit(1, "Could not open %s for reading\n", argv[1]);
    fdos = open(argv[2], O_RDONLY|O_BINARY, 0);
    if (fdos < 0)
        quit(1, "Could not open %s for reading\n", argv[2]);
    fo = open(argv[3], O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0644);
    if (fo < 0)
        quit(2, "Could not open %s for writing\n", argv[3]);

    ret = join_them(fi, fdos, fo);

    close(fi); close(fo);
    return(ret);
}

void quit(
    int     exit_code,
    char    *message,
    void    *arg)
{
    fprintf(stderr, message, arg);
    exit(exit_code);
}

typedef struct {
    IMAGE_DOS_HEADER    dos;
    char                dos_warning[0x40];
    IMAGE_NT_HEADERS    nt;
} f32_t;
typedef struct {
    IMAGE_DOS_HEADER    dos;
} f16_t;

int check_image( f32_t  *h);

int file_copy(
    int     infile,
    int     outfile,
    unsigned long int   bytes);
 
int join_them(
    int     infile,
    int     dos_infile,
    int     outfile)
{
#define MAX32HEADER 4096
    f32_t   *f32 = (void *) malloc(MAX32HEADER);
    f16_t   f16;
    int ret;
    unsigned long int image_size, exe16size, rndexe16;
    char *msg;

    assert(f32);

    /*
     * Read and verify DOS image header
     */
    msg = "Bad EXE16 image";
    ret = read(dos_infile, &f16.dos, sizeof(f16.dos));
    if (ret != sizeof(f16.dos)) goto out;

    /* Read in the exe32 file */
    msg = "Bad EXE32 image";
    ret = _read(infile, f32, MAX32HEADER);
    if (ret != MAX32HEADER) goto out;

    /*
     * How big is the file going to be ?
     */
    exe16size = _filelength(dos_infile);
    rndexe16 = ((exe16size + t_align - 1) & ~(t_align - 1));
    image_size = _filelength(infile) + rndexe16;

    /*
     * Fix the out header
     */
    f16.dos.e_cblp = (unsigned short) ((unsigned long)(image_size) % 512);
    f16.dos.e_cp = (unsigned short) ((unsigned long)(image_size) / 512);
    if (f16.dos.e_cblp)
        f16.dos.e_cp++;
printf("last page has x%x bytes, x%x pages\n", f16.dos.e_cblp, f16.dos.e_cp);

    /* Fix exe32 header */
    assert(sizeof(f32_t) - sizeof(IMAGE_NT_HEADERS) == 128);
    if (f32->dos.e_lfanew != 128) {
        /* Our DOS Extender assumes this and offsets into the NT header.
         * It is easier to fix the image here than to change the 16-bit
         * assembly.
         * The second DOS program is never used so it is irrelevant
         */
        UINT32 Offset = f32->dos.e_lfanew - 128;
        UINT HeaderSize = *(UINT32*) ((char*)&f32->nt.OptionalHeader.SizeOfHeaders + Offset);

        fprintf(stderr,
            "mk86boot: Truncating 32-bit image's DOS header by %d to to 128 bytes (hdr=%d)\n", Offset, HeaderSize);
        /* Move padding.   SizeOfHeaders includes the 128 byte DOS stuff */
        memcpy(&f32->nt,
               (char *)&f32->nt + Offset,
               HeaderSize - Offset - 128);
        /* Write back corrected offset (not that anyone uses it...) */
        f32->dos.e_lfanew = 128;
    }
    assert(f32->nt.OptionalHeader.SizeOfHeaders <= MAX32HEADER);
    assert(_filelength(infile) >= MAX32HEADER);

#if 0
    // BSS info, hopefully less than 64k !
    f16.dos.e_minalloc = f32->nt.OptionalHeader.SizeOfUninitializedData;
#else
    assert(f32->nt.OptionalHeader.SizeOfUninitializedData == 0);
#endif

    /*
     * Now do the writing, header first
     */
    msg = "Could not write out header";
    ret = write(outfile, &f16.dos, sizeof f16.dos);
    if (ret != sizeof f16.dos) goto out;

    // now copy the EXE image
    msg = "Could not copy the EXE image";
    ret = file_copy( dos_infile, outfile, exe16size - sizeof(f16.dos));
    if (ret) goto out; 

    // pad outfile up as needed
    (void) _lseek(outfile, rndexe16, SEEK_SET);

    // now write the f32 beginnings
    msg = "Could not write out f32 header";
    ret = _write(outfile, f32, MAX32HEADER);
    if (ret != MAX32HEADER) goto out;

#if 0
    // and finally the entire NT image
    msg = "Seek failed";
    ret = _lseek(infile, 0, SEEK_SET);
    if (ret < 0) goto out;
#endif
    
    msg = "Could not copy the image";
    ret = file_copy( infile, outfile, _filelength(infile) - MAX32HEADER);
    if (ret) goto out; 

    msg = NULL;
    ret = 0;
out:
    if (msg != NULL)
        fprintf(stderr, "join_them: %s\n", msg);
    return ret;
}

int check_image( f32_t  *h)
{
    return 0;
}

int file_copy(
    int     infile,
    int     outfile,
    unsigned long int   bytes)
{
    unsigned char   *buf;
    unsigned int copy_size;
    int wsize, ret;
#define _64k (64*1024)
    copy_size = bytes < _64k ? bytes : _64k;

    buf = (unsigned char *) malloc( copy_size);

    do {
        wsize = read(infile, buf, copy_size);
        ret = wsize;
        if (ret <= 0)
            break;

        ret = write(outfile, buf, wsize);
        if (ret != wsize)
            break;

        bytes -= wsize;
        if (bytes < copy_size)
            copy_size = bytes;
    } while (bytes > 0);
    return bytes;
}
