/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include <windows.h>
#include "protos.h"

#define dos_align 512

#pragma pack(1)

/*
 * Information in a directory entry
 */
typedef struct {
    char name[8];          /* file name */
    char ext[3];           /* file extension */
    char attr;             /* attribute byte */
#define FAT_ATTR_RONLY      0x01
#define FAT_ATTR_HIDDEN     0x02
#define FAT_ATTR_SYS        0x04
#define FAT_ATTR_LABEL      0x08    /* optional entry, only in rootdir */
#define FAT_ATTR_DIR        0x10
#define FAT_ATTR_ARCHIVE    0x20    /* e.g. dirty bit */
#define FAT_ATTR_xxx        0xc0
    char reserved[10];     /* mbz */
    char time[2];          /* time stamp */
    char date[2];          /* date stamp */
    short start;           /* starting cluster number */
    int  size;             /* size of the file */
} fat_directory_record;

/* Turn a string into a FAT 8.3 name
 */
static void StrToFName(char *ncp, char *cp)
{
    int len = 0;
    char c, *p, *sep;

    /* First strip CP of path leaders
     */
    p = cp;
    sep = NULL;
    do {
        c = *p++;
        if (c == '/' || c == '\\')
            sep = p;
    } while (c != 0);
    if (sep)
        cp = sep;

    /* Now do our thing
     */
    while ((c = *cp) != 0) {

        if ((c == '.') && (len < 9)) {
            while (len < 8) {
                *ncp++ = ' ';
                len++;
            }
            cp++;
            continue;
        }

        if (++len > 11)
            return;

        /* Upcase it.  Shouldn't, really.
         */
        if (c >= 'a' && c <= 'z')
            c -= 'a' - 'A';

        *ncp++ = c;
        cp++;
    }
    
    /* Pad it
     */
    while (len++ < 11)
        *ncp++ = ' ';

}

/*
 * Information contained in sector zero of a FAT filesystem
 */
typedef struct {
    char   bootid;         /* bootable or not */
#define BIOS_BOOTABLE   128

    char   beghead;        /* beginning head, sector, cylinder */
    char   begsect;        /* begcyl is a 10-bit number. High 2 bits */
    char   begcyl;         /*     are in begsect. */

    char   systid;         /* filesystem type */
#define SYS_FAT 6
#define SYS_EXT 5

    char   endhead;        /* ending head, sector, cylinder */
    char   endsect;        /* endcyl is a 10-bit number.  High 2 bits */
    char   endcyl;         /*     are in endsect. */

    int    offset;
    int    n_sectors;
}  bios_partition_info;

typedef struct {
    char   partitions[4*sizeof(bios_partition_info)];
    /* alignment !! */
    char  magic[2];
#define BIOS_LABEL_MAGIC    ((short)0xaa55)
} bios_label;


struct {
    char   jump[3];        /* Jump to boot code */
    char   banner[8];      /* OEM name & version */
    short  secsiz;         /* Bytes per sector */
    char   clsiz;          /* Cluster size in sectors */
    short  nrsvsect;       /* Number of reserved (boot) sectors */
    char   nfat;           /* Number of FAT tables hopefully 2 */
    short  dirents;        /* Number of root directory entries */
    short  psect;          /* Total sectors on disk */
    char   descr;          /* Media descriptor=first byte of FAT */
    short  fatlen;         /* Sectors in FAT */
    short  nsect;          /* Sectors/track */
    short  nheads;         /* Heads */
    int    nhs;            /* number of hidden sectors */
    int    bigsect;        /* big total sectors */
    char   driveno;        /* drive number */
    char   reserved;
    char   bsigrec;        /* ext boot signature record (x29) */
    char   volid[4];       /* 32bit volume id */
    char   volume_label[11];   /* */
    char   res1[8];        /* actually, "FATnn   " nn=12/16 */
    char   bootcode[384]; /* boot code */
    bios_label label;       /* right at the end */
} TheLabel;

/* Initialize the FAT fs label
 */
static void SetLabel(int n_files, int fatsize, int secsize, int nsecs)
{
    memset(&TheLabel, 0, sizeof TheLabel);

    TheLabel.secsiz = secsize;
    TheLabel.clsiz = 1;
    TheLabel.nrsvsect = 1;
    TheLabel.nfat = 1;
    TheLabel.dirents = n_files;
    if (nsecs > 4087)
        TheLabel.bigsect = nsecs;
    else
        TheLabel.psect = nsecs;

    /* TheLabel.descr = ?; */
    TheLabel.fatlen = fatsize / dos_align;
    TheLabel.bsigrec = 0x29;

    *(short*)(TheLabel.label.magic) = BIOS_LABEL_MAGIC;

    if (sizeof TheLabel != dos_align)
        quit(-1, "INTERNAL ERROR: LABEL SIZE x%x != 0x200\n",
             (void*)sizeof TheLabel);
}

/* The FAT table
 */
unsigned short *TheFAT;

static void
FatSetEntry( int FBits, uint Entry, unsigned short Value)
{
    if (FBits == 16) {
        TheFAT[Entry] = Value;
        return;
    }

    Value &= 0xfff;
    FBits = (Entry * 3) / 4;/* 12/16 == 3/4 */

    switch (Entry & 0x3) {
    case 0:
        TheFAT[FBits] = (unsigned short)((TheFAT[FBits] & 0xf000) | Value);
        break;
    case 1:
        TheFAT[FBits] = (unsigned short)((TheFAT[FBits] & 0x0fff) | (Value << 12));
        FBits++;
        TheFAT[FBits] = (unsigned short)((TheFAT[FBits] & 0xff00) | (Value >>  4));
        break;
    case 2:
        TheFAT[FBits] = (unsigned short)((TheFAT[FBits] & 0x00ff) | (Value <<  8));
        FBits++;
        TheFAT[FBits] = (unsigned short)((TheFAT[FBits] & 0xfff0) | (Value >>  8));
        break;
    case 3:
        TheFAT[FBits] = (unsigned short)((TheFAT[FBits] & 0x000f) | (Value <<  4));
        break;
    }
}

/* Allocate FAT entries, sequentially
 */
unsigned short FatSequential( unsigned short firstcl, int fsize)
{
    int b, clsiz;

    clsiz = TheLabel.psect;
    if (clsiz == 0)
        clsiz = TheLabel.bigsect;
    b = (clsiz > 4087) ? 16 : 12;
    clsiz = (TheLabel.secsiz * TheLabel.clsiz);

    if (TheFAT == NULL) {
        TheFAT = malloc( TheLabel.fatlen * dos_align );
        memset(TheFAT, 0, TheLabel.fatlen * dos_align );
        FatSetEntry( b, 0, 0xffff); /* circa */
        FatSetEntry( b, 1, 0xffff); /* circa */
    }

    while (fsize > clsiz) {
        FatSetEntry( b, (uint) firstcl, (unsigned short)(firstcl + 1));
        firstcl++;
        fsize -= clsiz;
    }

    /* Terminate list
     */
    FatSetEntry( b, firstcl++, 0xffff);
    return firstcl;
}


/* This does the real work.  Appends the given list of files
 * to make up the FAT filesystem (cluster space). While going,
 * creates the root directory, and marks clusters as in use
 * in the FAT table.
 */
static char *StickThemIn( int argc, char **argv, int fout)
{
    int f, ro;
    uint fl;
    fat_directory_record *root;
    char *rt;
    unsigned short next_cluster = 2;

    fl = ((argc * sizeof(*root)) + dos_align - 1) & ~(dos_align - 1);
    root = malloc( fl );
    memset(root, 0, fl);
    rt = (char *)root;

    while (argc-- > 0) {

        /* Open next file
         */
        if (strcmp(*argv,"-ro") == 0) {
            if (argc-- == 0) /* trailing '-ro' with no file name */
                break;
            argv++;
            ro = 1;
        }
        else
            ro = 0;

        f = _open(*argv,O_RDONLY|O_BINARY, 0);
        if (f < 0)
            quit(3, "Could not open %s for reading\n", *argv);

        /* Set metadata
         */
        StrToFName(root->name, *argv);
        root->attr = FAT_ATTR_ARCHIVE;
        if (ro)
            root->attr |= FAT_ATTR_ARCHIVE;

        fl = _filelength(f);
        root->start = next_cluster;
        root->size = fl;
printf("File %s at x%x.x%x\n", *argv, next_cluster, fl);

        next_cluster = FatSequential(next_cluster, fl);

        /* Append it
         */
        if (file_copy( f, fout, fl) != 0)
            quit(5, "Error appending file %s\n", *argv);

        /* Pad if necessary
         */
        fl = (fl & (t_align - 1));
        if (fl) {
            fl = t_align - fl;
printf("Pad %s to x%x\n", *argv, fl);
#if 0
            (void) _lseek(fout, fl, SEEK_CUR);
#else
            (void) file_zero(fout, fl);
#endif
        }

        /* Move on to next file
         */
        _close(f);
        argv++;
        root++;
    }
    return rt;
}

/* How much overhead
 */
static uint FatFsGoop(uint data_size, uint n_files, uint *pfats)
{
    uint bsiz, fsiz, rsiz;

    /* one for boot */
    bsiz = dos_align;

    /* one for FAT (conservative!) */
    fsiz = ( ((data_size / t_align) * 2) + dos_align - 1 ) & ~(dos_align - 1);

    /* one for root */
    rsiz = ((n_files * 32) + t_align - 1 ) & ~(t_align - 1);

printf("Boot+FAT+Root+Data = %d + %d + %d + %d == %d\n",
       bsiz, fsiz, rsiz, data_size,
       bsiz + fsiz + rsiz + data_size);

    *pfats = fsiz;
    return bsiz + fsiz + rsiz;
}

/* Main control and only entry point
 */
char *mkfat(
        int     argc,
        char    **argv)
{
    char *fname;
    int f;
    uint datsiz, metasiz, fats;
    char *rt;

    /* How much data do we have
     */
    datsiz = SizeOfTheseFiles(argc, argv);

printf("SizeOfFiles is x%x\n", datsiz);

    /* Compute metadata sizes
     */
    metasiz = FatFsGoop(datsiz, argc, &fats);

    /* Initialize label
     */
    SetLabel( argc, fats, t_align, (metasiz + datsiz) / t_align);

    /* Create output file
     */
    fname = MakeAName();
    if (fname == NULL)
        quit(6, "Could not generate a suitable temporary filename\n", NULL);

    f = _open(fname, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0644);
    if (f < 0)
        quit(6, "Could not create a temporary file [%s]\n", fname);

    /* Leave room for metadata
     */
    (void) _lseek(f, metasiz, SEEK_SET);

    /* Copy all files, create root directory
     */
    rt = StickThemIn(argc,argv,f);

    /* Write out metadata
     */
    (void) _lseek(f, 0, SEEK_SET);
    _write(f, &TheLabel, dos_align);
    _write(f, TheFAT, fats);
    _write(f, rt, metasiz - (dos_align + fats));

    /* Done
     */
    _close(f);

    return fname;
}
