/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Build an executable that contains an in-memory filesystem.
 *
 * This program takes a 32bit EXE NT image (a MMLite kernel),
 * and an arbitrary file as input.  It produces a new 32bit
 * EXE NT image that concatenates the two files.  While copying
 * it, the original EXE is modified in the appropriate spot to
 * point to the catenated 'arbitrary file', and to record the size.
 * The resulting MMLite kernel image now will use that
 * 'arbitrary file' as its root filesystem during boot.
 *
 * Optionally, this program can be given a list of files
 * which it merges into a (FAT or ROM) filesystem image.
 * This file in turn is then used in the procedure above.
 *
 *
 **** Example1:
 *
 *      mkmemfs mmlite.exe m32.exe filesyst
 *
 * Produce a bootable image 'mmlite.exe' from the kernel image 'm32.exe'
 * and the filesystem image 'filesyst'. No check is made on 'filesyst'.
 *
 *
 **** Example2:
 *
 *      mkmemfs mmlite.exe m32.exe init.exe init.dat shell.exe ...
 *
 * Take the list of filenames 'init.exe init.dat ...' and produce an
 * intermediate FAT filesystem image file 'unnamed'.  Then execute
 *      mkmemfs mmlite.exe m32.exe 'unnamed'
 *
 *
 **** Example3:
 *
 *      mkmemfs -rom mmlite.exe m32.exe init.exe init.dat shell.exe ...
 *
 * Same as above, except the filesystem has a ROM format (see mkrom.c)
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include <windows.h>
#include "protos.h"

#define _64k (64*1024)

/* Better be a power of 2 or else
 */
#ifndef MY_SECTOR_SIZE
#define MY_SECTOR_SIZE  512/*128*/
#endif

uint    t_align = MY_SECTOR_SIZE;

#define FS_ISA_FAT 1
#define FS_ISA_ROM 2

uint    fs_type = FS_ISA_FAT;

uint    debug = 0;
uint    unicode = 1;
uint    byteswap = 0;

void ParseArgs(
        int     *pargc,
        char    ***pargv)
{
        int argc = *pargc;
        char **argv = *pargv;

        while (argc-- > 1) {
            argv++;
            if (argv[0][0] == '-') {
                (*pargc)--;
                (*pargv)++;

                if (strcmp(argv[0]+1, "rom") == 0) {
                    fs_type = FS_ISA_ROM;
                    t_align = 0x1000; /* current common pagesize */
                } else if (strcmp(argv[0]+1, "debug") == 0) {
                    debug = 1;
                } else if (strcmp(argv[0]+1, "ascii") == 0) {
                    unicode = 0;
                } else if (strcmp(argv[0]+1, "unicode") == 0) {
                    unicode = 1;
                } else if (strcmp(argv[0]+1, "align") == 0) {
                    (*pargc)--;
                    (*pargv)++;
                    argc--;
                    argv++;
                    t_align = atoi(argv[0]);
                    if ((t_align & (~(t_align - 1))) != t_align) {
                        quit(-1, "Alignment not a power of 2\n", argv[0]);
                    }

                } else if (strcmp(argv[0]+1, "bigendian") == 0) {
                    byteswap = 1;
                } else
                    quit (-1, "Bad arg %s\n", argv[0]);
            }
            else
                break;
        }
}

int __cdecl main(int argc, char **argv)
{
    int ret;
    int dirt = 0;
    char * fileSystemImage;
    int newArgc;
    char ** newArgv;
    
    ParseArgs( &argc, &argv);
    
    if (argc < 4) {
        quit(-1, 
             "Example uses:\n\t%s boot_imagename mmlite_image filesystem_image\n"\
             "\t%s boot_imagename mmlite_image <list of filenames to put in the FAT filesystem>\n"\
             "\t%s -rom boot_imagename mmlite_image <list of filenames to put in the ROM filesystem>",
                     argv[0]);
    }                
    
    /* Do we have to build the filesystem
     */
    if (argc >= 4) {
        if (FALSE == CreateNewCmdArguments (argc-3, argv+3, &newArgc, &newArgv)) {
            quit (-1, "Exiting...\n", 0);
        }
        fileSystemImage = mkfs(newArgc, newArgv);
          
        FreeCmdArguments (newArgv);
          
        if (fileSystemImage == NULL) {
            quit(4, "Could not build the filesystem\n", NULL);
        }
        dirt = 1;
    } else {
        printf("Not making fs\n");
        fileSystemImage = argv[3];
    }
    
    ret = catfiles(argv[1], argv[2], fileSystemImage);
    
    if (dirt) {
        if (debug)
          printf("Filesystem image preserved in %s\n", fileSystemImage);
        else
          _unlink(fileSystemImage);
    }
    
    return ret;
}

/*
 */
char *mkfs(
        int     argc,
        char    **argv)
{
    switch (fs_type) {
    case FS_ISA_FAT:
        return mkfat(argc, argv);
    case FS_ISA_ROM:
        if (debug)
          printf("Making ROMFS with %s strings\n",
                 unicode ? "unicode" : "ascii");
        return mkrom(argc, argv);
    default:
        return NULL;
    }
}



int catfiles( char *fout, char *fkern, char *fsys)
{
        int fk, fs, fo, ret;

        fo = _open(fout, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0644);
        if (fo < 0)
                quit(2, "Could not open %s for writing\n", fout);

        fk = _open(fkern, O_RDONLY|O_BINARY, 0);
        if (fk < 0)
                quit(1, "Could not open %s for reading\n", fkern);

        fs = _open(fsys, O_RDONLY|O_BINARY, 0);
        if (fs < 0)
                quit(1, "Could not open %s for reading\n", fsys);
/* xxx peek into it to see sector size ?? */

        ret = join_them(fk, fs, fo);

        _close(fk); _close(fo); _close(fs);
        return(ret);
}

void quit(
        int             exit_code,
        char    *message,
        void    *arg)
{
        fprintf(stderr, message, arg, arg, arg, arg);
        exit(exit_code);
}

int join_them(
        int             kern_infile,
        int             fs_infile,
        int             outfile)
{
        IMAGE_DOS_HEADER DosHeader;
        struct {
            IMAGE_NT_HEADERS PeHeader;
            /* Big 'nuf to hold 4 section headers */
            unsigned char Data[608 - sizeof(IMAGE_NT_HEADERS)];
        } h;
        PIMAGE_SECTION_HEADER SectionHdr;
        int ret;
        uint image_size, exe32size, rndexe32, fs_infile_length;
        char *msg;

        /*
         * Read and verify NT image header
         */
        msg = "Unable to read DOS header";
        ret = _read(kern_infile, &DosHeader, sizeof DosHeader);
        if (ret != sizeof(DosHeader)) goto out;

        if (DosHeader.e_magic != 0x5a4d) {
            msg = "Bad magic in DOS header";
            ret = -1;
            goto out;
        }

        msg = "Unable to read PE header";
        /* Skip forward to NT header */
        ret = _lseek(kern_infile, DosHeader.e_lfanew, SEEK_SET);
        if (ret == -1) goto out;

        ret = _read(kern_infile, &h, sizeof h);
        if (ret != sizeof(h)) goto out;

        /*
         * How big is the file going to be ?
         */
        exe32size = _filelength(kern_infile);
        rndexe32 = ((exe32size + t_align - 1) & ~(t_align - 1));
        fs_infile_length = _filelength(fs_infile);
        image_size = fs_infile_length + rndexe32;

        /*
         * Fix the out header
         */
        DosHeader.e_cblp = (unsigned short) ((unsigned long)(image_size) % 512);
        DosHeader.e_cp = (unsigned short) ((unsigned long)(image_size) / 512);
        if (DosHeader.e_cblp)
                DosHeader.e_cp++;

        /*
         * Fix the last section header to account for the extra data
         */
        SectionHdr = IMAGE_FIRST_SECTION(&h.PeHeader);
        SectionHdr += h.PeHeader.FileHeader.NumberOfSections - 1;

        SectionHdr->SizeOfRawData += (image_size - rndexe32);
        /* Do NOT fix the Misc.VirtualSize field, in case we need to know */

        /*
         * Do not update the PE image size field.  This allows the kernel boot code to
         * map only the kernel image instead of the entire boot image.  The file system
         * will be addressed through unmapped memory.  
         */
#if 1
        h.PeHeader.OptionalHeader.SizeOfImage += (image_size - rndexe32);
#endif

        /*
         * Now do the writing, header first
         */
        msg = "Could not write out DOS header";
        ret = _write(outfile, &DosHeader, sizeof DosHeader);
        if (ret != sizeof DosHeader) goto out;

        msg = "Could not seek when writing out header";
        ret = _lseek(outfile, DosHeader.e_lfanew, SEEK_SET);
        if (ret == -1) goto out;

        msg = "Could not write out PE header";
        ret = _write(outfile, &h, sizeof h);
        if (ret != sizeof h) goto out;

        /*
         * Now copy the EXE image
         */
        msg = "Could not copy the EXE image";
        ret = file_check_copy( kern_infile,
                               outfile,
                               exe32size - DosHeader.e_lfanew - sizeof h,
                               fs_infile_length,
                               h.PeHeader.OptionalHeader.ImageBase + rndexe32,
                               fs_type);
        if (ret) goto out; 

        printf("ImageBase is %x\n", h.PeHeader.OptionalHeader.ImageBase);

        /*
         * pad outfile up as needed
         */
        (void) _lseek(outfile, rndexe32, SEEK_SET);

        /*
         * and finally the entire filesystem image
         */
        msg = "Seek failed";
        ret = _lseek(fs_infile, 0, SEEK_SET);
        if (ret < 0) goto out;
        
        msg = "Could not copy the image";
        ret = file_copy( fs_infile, outfile, fs_infile_length);
        if (ret) goto out; 

        return 0;
out:
        if (msg != NULL)
                fprintf(stderr, "join_them: %s\n", msg);
        return ret;
}

unsigned char   *buf = NULL;

int file_zero(
        int     outfile,
        uint    bytes)
{
        unsigned int zero_size;
        int ret;

        if (buf == NULL)
            buf = (unsigned char *) malloc( _64k);
        memset(buf,0,_64k);

        do {
                zero_size = bytes < _64k ? bytes : _64k;

                ret = _write(outfile, buf, zero_size);
                if (ret != zero_size)
                        break;

                bytes -= zero_size;

        } while (bytes > 0);
        return bytes;
}

int file_copy(
        int     infile,
        int     outfile,
        uint    bytes)
{
        unsigned int copy_size;
        int wsize, ret;
        copy_size = bytes < _64k ? bytes : _64k;

        if (buf == NULL)
            buf = (unsigned char *) malloc( _64k);

        do {
                wsize = _read(infile, buf, copy_size);
                ret = wsize;
                if (ret <= 0)
                        break;

                ret = _write(outfile, buf, wsize);
                if (ret != wsize)
                        break;

                bytes -= wsize;
                if (bytes < copy_size)
                        copy_size = bytes;
        } while (bytes > 0);
        return bytes;
}

/* Same, but look for hot sexy spot
 */
int file_check_copy(
        int     infile,
        int     outfile,
        uint    bytes,
        uint    fssize,
        uint    fpointer,
        uint    ftype)
{
        unsigned int copy_size;
        int wsize, ret, prev_match, found = 0;

/*printf("check_copy x%x x%x x%x %x\n", bytes, fssize, fpointer, ftype);*/
        copy_size = bytes < _64k ? bytes : _64k;

        if (buf == NULL)
            buf = (unsigned char *) malloc( _64k);

        prev_match = 0;
        do {
                wsize = _read(infile, buf, copy_size);
                ret = wsize;
                if (ret <= 0)
                        break;

                if (!found)
                found = check( prev_match, (uint *)buf, wsize, &prev_match,
                              fssize, fpointer, ftype);

                ret = _write(outfile, (uint *) buf, wsize);
                if (ret != wsize)
                        break;

                bytes -= wsize;
                if (bytes < copy_size)
                        copy_size = bytes;
        } while (bytes > 0);
        if (!found)
            printf("\n\tWARNING: did not find MemFs header signature!\n\n");
        return bytes;
}

__inline uint _bs32(uint a)
{
    if (byteswap) {
        a =  ((((a)&0xff)<<24) | (((a)&0xff00)<<8) |
              (((a)&0xff0000)>>8) | (((a)&0xff000000)>>24));
    }
    return a;
}

int check( int nmatch, uint *data, int nbytes, int *pmatch,
            uint fssize, uint fpointer, uint ftype)
{
    /* !!! Keep compat with kernel
     */ 
    static const uint patt[] =
    { 0x11131719, 0x01020304, 0x04030201, 0x19171311 };
#define SIGNATURE_SIZE (sizeof patt/sizeof patt[0])

    /* This code assumes that nbytes is a multiple of sizeof *data.
     */
    while (nbytes > 0) {

        switch (nmatch) {
        case SIGNATURE_SIZE:
            *data++ = _bs32(ftype);
            nmatch++;
            break;

        case SIGNATURE_SIZE + 1:
            *data++ = _bs32(fssize);
            nmatch++;
            break;

        case SIGNATURE_SIZE + 2:
            *data++ = _bs32(fpointer);
            nmatch++;
            break;

        case SIGNATURE_SIZE + 3:
            /* All done
             */
            return 1;

        default:
            if (*data++ == patt[nmatch])
                nmatch++;
            else
                nmatch = 0;
            break;
        }

        nbytes -= sizeof *data;
    }

    *pmatch = nmatch;
    return 0;
}

/*
 */
int SizeOfTheseFiles( int argc, char **argv)
{
    int s = 0;
    int f;

    while (argc-- > 0) {

        if (strcmp(*argv,"-ro") == 0) {
            if (argc-- == 0) /* trailing '-ro' with no file name */
                break;
            argv++;
        }
        f = _open(*argv,O_RDONLY|O_BINARY, 0);
        if (f < 0)
            quit(3, "Could not open %s for reading\n", *argv);
        s += (_filelength(f) + t_align - 1) & ~(t_align - 1);
        _close(f);
        argv++;
    }
    return s;
}

/*
 */
char *MakeAName()
{
    static char TheName[128];
    int f, n = 0;

    do {
        TheName[0] = 0;
        sprintf(TheName, "%d.flp", n++);
        /* good nuf */
        f = _open(TheName, O_RDONLY, 0);
        if (f < 0)
            return TheName;

        _close(f);
    } while (n < 9999);
    return NULL;
}


/*
    This Function searches for duplicate arguments in the given
    argument list.

    Returns 
        TRUE and new argc/argv pair if no duplicate arguments w/ different option found
    else
        FALSE
*/    

int CreateNewCmdArguments (int argc, char **argv, int * newArgc, char ***newArgv) 
{
    int i, j;
    char ** myArgv;
    int myArgc;
    int uniqueItem;
    int diffOption;
    char *orgOption;
    char *newOption;

    *newArgc = 0;
    *newArgv = NULL;

    myArgv = malloc (argc * sizeof (char **));
    myArgc= 0;
    for (i = 0; i < argc; i++) {
        uniqueItem = TRUE;
        if (argv[i][0] != '-') { /* always keep option argument */
            for (j = 0; j < myArgc; j++) {
                if (0 == strcmp(myArgv[j], argv[i])) {
                    /* 
                       we found duplicate argument.  we can silently ignore 
                       it if it is preceded by the same option (if any)
                       as the original
                    */      
                    orgOption = newOption = NULL;
                    if (j >= 1)
                        if (myArgv[j-1][0] == '-')
                            orgOption = myArgv[j-1];
                    if (i >= 1)
                        if (argv[i-1][0] == '-')
                            newOption = argv[i-1];

                    diffOption = TRUE;
                    if ((orgOption == NULL) && (newOption == NULL))
                        diffOption = FALSE;
                    if ((orgOption != NULL) && (newOption != NULL))
                        if (0 == strcmp (orgOption, newOption)) {
                            diffOption = FALSE;
                            myArgc--;               /* yank out the duplicate option */
                        }
                            
                    if (diffOption == TRUE) {
                        printf ("\n\nError: Duplicate file %s with different options\n\n", argv[i]);
                        FreeCmdArguments (myArgv);
                        return FALSE;
                    }
                    printf ("Warning: Ignoring duplicate argument %s\n", argv[i]);

                    uniqueItem = FALSE;
                    break;
                }
            }
        }    
        if (uniqueItem == TRUE) {
            myArgv[myArgc] = argv[i];
            myArgc++;
        }            
    }
    *newArgc = myArgc;
    *newArgv = myArgv;

    return TRUE;
}

/*
    This function frees the command line argument list created by 
    CreateNewCmdArguments
 */            
void FreeCmdArguments (char **argv) 
{
    if (argv)       
        free (argv);        
}
