/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include <windows.h>
#include "protos.h"


extern uint unicode;            /* BOOL really */
extern uint byteswap;           /* BOOL really */

__inline uint _bs32(uint a)
{
    if (byteswap) {
        a =  ((((a)&0xff)<<24) | (((a)&0xff00)<<8) |
              (((a)&0xff0000)>>8) | (((a)&0xff000000)>>24));
    }
    return a;
}

typedef struct {
    uint        FileOffset;
    uint        FileSize;

} rom_direntry;

struct {
    uint        Signature;
#define ROM_SIGNATURE 0xabaddeed
    uint        TotalSize;
    uint        BlockSize; /* must be 1page for VM mapping */
    uint        FilesOffset;
    uint        StringsOffset;
    uint        NumberOfFiles;
} rom_label;

/* Initialize ROMFS' label
 */
static void SetLabel( uint nf, uint ms, uint bs, uint ts)
{
    rom_label.Signature = _bs32(ROM_SIGNATURE);
    rom_label.TotalSize = _bs32(ms + ts);
    rom_label.BlockSize = _bs32(bs);
    rom_label.FilesOffset = _bs32(ms);
    rom_label.StringsOffset = _bs32((nf * sizeof(rom_direntry)) + sizeof rom_label);
    rom_label.NumberOfFiles = _bs32(nf);
    
}

/* How big string space will be
 */
static int SizeOfTheseStrings( int argc, char **argv)
{
    int s = 0;

    /* BUGBUG should only count last component of filepath
     */
    while (argc-- > 0) {
        /* UNICODE !!! */
        if (unicode)
            s += (strlen(argv[0]) + 1) * 2;
        else
            s += (strlen(argv[0]) + 1);
        argv++;
    }
    return s;
}

/* Strip off leading path component, if any, from an ascii string.
 * [NOT doing this would be cute, but misleading]
 */
char *basename(char *fname)
{
    char *p, *sep, c;

    sep = NULL;
    p = fname;
    for ( c = 1; c; p++) {
        c = *p;
        if (c == '\\' || c == '/')
            sep = p;
    }
    return (sep ? sep + 1 : fname);
}

/* Turn an ASCII filepath into a UNICODE filename
 */
char *AsciiToUnicode( char *dest, char *fname)
{
    unsigned short *ucode;

    ucode = (unsigned short *) dest;
    while (*fname)
        *ucode++ = (unsigned short)*fname++;
    *ucode++ = 0;
    return (char *)ucode;
}

/* Same as above but keep it ascii */
char *AsciiToAscii( char *dest, char *fname)
{
    while (*fname)
        *dest++ = *fname++;
    *dest++ = 0;
    return dest;
}

/* This does the real work.
 * Append files to the filesystem, mark root entries
 * and add to string space as we go.
 * Returns a pointer to the start of the FS (metadata).
 */
static char *StickThemIn( int argc, char **argv, int fout, uint overh)
{
    int f, ro;
    uint fl;
    char *rt, *strings;
    rom_direntry *dir;
    uint next_offset;

    rt = malloc( overh );
    memset(rt, 0, overh);

    memcpy(rt, &rom_label, sizeof rom_label);
    dir = (rom_direntry *) (rt + sizeof rom_label);
    strings = (char *)(dir + argc);

    next_offset = overh;
    while (argc-- > 0) {

        /* Open next file
         */
        if (strcmp(*argv,"-ro") == 0) {
            if (argc-- == 0) /* trailing '-ro' with no file name */
                break;
            argv++;
            ro = 1;
        }
        else
            ro = 0;

        f = _open(*argv,O_RDONLY|O_BINARY, 0);
        if (f < 0)
            quit(3, "Could not open %s for reading\n", *argv);

        /* Set metadata
         */
        if (unicode)
            strings = AsciiToUnicode(strings, basename(*argv));
        else
            strings = AsciiToAscii(strings, basename(*argv));
        fl = _filelength(f);

        dir->FileOffset = _bs32(next_offset);
        dir->FileSize = fl;
        if (ro)
            dir->FileSize |= 0x80000000; /* Read only file */
        dir->FileSize = _bs32(dir->FileSize);

printf("File %s at x%x.x%x\n", *argv, next_offset, fl);

        /* Append it
         */
        if (file_copy( f, fout, fl) != 0)
            quit(5, "Error appending file %s\n", *argv);

        /* Pad if necessary
         */
        next_offset += fl;
        fl = (fl & (t_align - 1));
        if (fl) {
            fl = t_align - fl;
printf("Pad %s to x%x\n", argv[0], fl);
#if 0
            (void) _lseek(fout, fl, SEEK_CUR);
#else
            (void) file_zero(fout, fl);
#endif
            next_offset += fl;
        }

        /* Move on to next file
         */
        _close(f);
        argv++;
        dir++;
    }
    /* Truly pad at end, or else kernel complains
     */
    if (fl) {
        (void) _lseek( fout, 0-fl, SEEK_CUR);
        _write(fout, rt, fl);
    }
    return rt;
}

/* How much overhead
 */
static uint RomFsGoop(uint n_files, uint sizeof_strings)
{
    uint s;
    s = (n_files * sizeof(rom_direntry)) + sizeof rom_label + sizeof_strings;
    return (s + t_align - 1) & ~(t_align - 1);
}

/* Main control and only entry point
 */
char *mkrom(
        int     argc,
        char    **argv)
{
    char *fname;
    int f;
    uint strsiz, metasiz, fsize;
    char *rt;

    /* How big is file space
     */
    fsize = SizeOfTheseFiles(argc, argv);

    /* How big is string space
     */
    strsiz = SizeOfTheseStrings(argc, argv);

    /* Say what
     */
    fprintf(stderr, "SizeOf* is Files x%x Strings x%x\n", fsize, strsiz);

    /* Compute metadata size
     */
    metasiz = RomFsGoop(argc, strsiz);

    /* Initialize label
     */
    SetLabel( argc, metasiz, t_align, fsize);

    /* Create output file
     */
    fname = MakeAName();
    if (fname == NULL)
        quit(6, "Could not generate a suitable temporary filename\n", NULL);

    f = _open(fname, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0644);
    if (f < 0)
        quit(6, "Could not create a temporary file [%s]\n", fname);

    /* Leave room for metadata
     */
    (void) _lseek( f, metasiz, SEEK_SET);

    /* Copy all files, create root directory
     */
    rt = StickThemIn(argc,argv,f, metasiz);

    /* Write out metadata
     */
    (void) _lseek( f, 0, SEEK_SET);
    _write(f, rt, metasiz);

    /* Done
     */
    _close(f);

    return fname;
}
