/* Copyright (c) Microsoft Corporation. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "import.h"

typedef unsigned char u_char;

#define DEBUG 0
int verbose = 0;

void quit(int status,char *Format, ...)
{
    va_list marker;
    va_start (marker,Format);
    vprintf (Format, marker);
    va_end(marker);

    printf("\n");/*just in case..  */
    exit(status);
}

/* ++ ************ Import list stuff ************ */

/* For sorting and searching, comparison functions.
 */
int CompareRtlByName(const void *node1, const void *node2)
{
    PRTL_ENTRY
        p1 = (PRTL_ENTRY) node1,
        p2 = (PRTL_ENTRY) node2;

    return strcmp(p1->RtlName, p2->RtlName);
}

int CompareRtlByOrdinal(const void *node1, const void *node2)
{
    PRTL_ENTRY
        p1 = (PRTL_ENTRY) node1,
        p2 = (PRTL_ENTRY) node2;

    return p1->Ordinal - p2->Ordinal;
}

int CompareRtlByRef(const void *node1, const void *node2)
{
    PRTL_ENTRY
        p1 = (PRTL_ENTRY) node1,
        p2 = (PRTL_ENTRY) node2;

    return p1->RefAsSymbol - p2->RefAsSymbol;
}

/* Debug
 */
void PrintEntries( char *msg, PRTL_ENTRY pEntries, int NumEntries)
{
    int i;
    printf("%s the RTL contains these %d entries:\n", msg, NumEntries);
    for (i = 0; i < NumEntries; i++, pEntries++)
        printf("%d: (%s %d) %x\n", i, 
               pEntries->RtlName, pEntries->Ordinal, pEntries->RefAsSymbol);
}

/* Utilities
 */
char *SkipBlanks(char *Line)
{
    char c;
    while ((c = *Line) &&
           ((c == ' ') || (c == '\t') || (c == '\n')))
        Line++;
    return (c) ? Line : NULL;
}

char *SkipToBlank(char *Line)
{
    char c;
    while ((c = *Line) &&
           ((c != ' ') && (c != '\t') && (c != '\n')))
        Line++;
    return (c) ? Line : NULL;
}

/* Parse a line of a .DEF file.
 */
#define T_ERROR    0
#define T_BLANKS   1
#define T_LIBNAME  2
#define T_UUID     3
#define T_DESCRIPT 4
#define T_PROPS    5
#define T_EXPORTS  6
#define T_NAME     7
typedef char *NAME;
NAME TokName[] = {
    "error", "blanks", "libname", "uuid",
    "description", "properties", "exports", "name"
};

int ParseDefline(char *Line, char **TokenStart)
{
    char c, *p, *q;
    int tok;

    Line = SkipBlanks(Line);
    if (Line == NULL)
        return T_BLANKS;
    /* Comments ? */
    if (Line[0] == ';')
        return T_BLANKS;
    if ((Line[0] == '/') && (Line[1] == '*')) {
        /* Multiline comments are not worth it.. just use the semicolon */
        return T_BLANKS;
    }

    /* got somehting..*/
    p = SkipToBlank(Line);
    if (p) {
        c = *p;
        *p = 0;
    }

    if (_stricmp(Line,"library") == 0) {
        tok = T_LIBNAME;
        goto Keyword;
    }

    if (_stricmp(Line,"interface") == 0) {
        tok = T_UUID;
        goto Keyword;
    }

    if (_stricmp(Line,"description") == 0) {
        tok = T_DESCRIPT;
        goto Keyword;
    }

    if (_stricmp(Line,"code") == 0) {
        tok = T_PROPS;
        goto Keyword;
    }

    if (_stricmp(Line,"exports") == 0) {
        return T_EXPORTS;
    }

    /* Must have an ordinal */
    if (!p) goto Fatal;
    *p = c;
    *TokenStart = Line;
    return T_NAME;

 Keyword:
    if (!p) goto Fatal;
    q = SkipBlanks(p+1);
    if (q == NULL) goto Fatal;
    *TokenStart = q;
    return tok;

 Fatal:
    if (p) *p = c;
    printf("Syntax error in line '%s'\n", Line);
    return T_ERROR;
}


/* Read a text file describing the DLL functions, and create
 * an array of descriptors for them.
 */
void ReadRtlFile( FILE *File, PRTL pRet)
{
    PRTL_ENTRY pEntries;
    int NumEntries, nAlloced, n, LineNo;
    char buff[256];
    char *Name, *Token, *p;
    int DoingExports = 0;

    /* start with a bunch */
#define ABUNCH 128
    nAlloced = ABUNCH;
    pEntries = (PRTL_ENTRY) malloc(nAlloced * sizeof(RTL_ENTRY));
    if (pEntries == NULL)
        quit(3,"Malloc(%d) failed", nAlloced * sizeof(RTL_ENTRY));

    /* Read one line at a time. BUGBUG should allow comments.
     * Or better yet, use the same format as a .DEF file
     */
    NumEntries = 0;
    LineNo = 0;
    while (fgets(buff,255,File) != NULL) {
        LineNo++;
        Token = NULL;
        n = ParseDefline(buff,&Token);

#if DEBUG
        printf("%s -> %s\n", TokName[n], Token);
#endif

        switch (n) {
        case T_ERROR:
            quit(1,"Fatal -- pilot error at line %d",LineNo);
        case T_BLANKS:
            continue;
        case T_LIBNAME:
            *SkipToBlank(Token) = 0;
            pRet->DllName = (char *)malloc(1+strlen(Token));
            strcpy(pRet->DllName,Token);
            continue;
        case T_UUID:
            *SkipToBlank(Token) = 0;
            pRet->Uuid = (char *)malloc(1+strlen(Token));
            strcpy(pRet->DllName,Token);
            continue;
        case T_DESCRIPT:
        case T_PROPS:
            continue;
        case T_EXPORTS:
            DoingExports = 1;
            continue;

        case T_NAME:
            break;

        default:
            quit(n,"Fatal -- Internal error at line %d",LineNo);
        }
        
        if (!DoingExports)
            quit(1,"Fatal -- Missing or misplaced EXPORTS directive");

        /* Extend if necessary (~~never).
         */
        if (NumEntries == nAlloced) {
            nAlloced += ABUNCH;
            pEntries = (PRTL_ENTRY) realloc( pEntries, nAlloced);
            if (pEntries == NULL)
                quit(4, "Realloc(%d) failed", nAlloced * sizeof(RTL_ENTRY));
        }

        /* NULL terminate */
        p = Token;        
        Token = SkipToBlank(Token);
        /* Scanner has checked there is more stuff after Token */
        *Token++ = 0;

        /* Copy to a new string and remember
         */
        Name = (char *) malloc(1+strlen(p));
        if (Name == NULL)
            quit(4, "Malloc(Name) failed for %s", Token);
        strcpy(Name,p);
        pEntries[NumEntries].RtlName = Name;

        Token = SkipBlanks(Token);
        if (Token == NULL)
            quit(3,"Fatal -- Missing ordinal for entry %s",Name);
        Name = Token;
        Token = SkipToBlank(Name);
        if (Token)
            *Token = 0; /* ignore the rest of the line */

        if (Name[0] == '@')
            Name++;

        n = atoi(Name);

        /* Fill up descriptor
         */
        pEntries[NumEntries].Ordinal = n;
        pEntries[NumEntries].RefAsSymbol = NULL;

        NumEntries++;
    }


    pRet->nEntries = NumEntries;
    pRet->pEntries = pEntries;
#undef ABUNCH
}

/* Search function.  Returns the descriptor if any, NULL otherwise.
 */
PRTL_ENTRY IsaRtl( PRTL pRtl, char *Name)
{
    RTL_ENTRY SrchEl;
    PRTL_ENTRY Entry;

    SrchEl.RtlName = Name;
    Entry = (PRTL_ENTRY) bsearch( &SrchEl,
                                  pRtl->pEntries,
                                  pRtl->nEntries,
                                  sizeof(RTL_ENTRY),
                                  CompareRtlByName );
    return Entry;
}

/* More debug
 */
static void CheckRtls(PRTL pRtl)
{
    PRTL_ENTRY Entry;
    char *Name;

    PrintEntries("Before sorting", pRtl->pEntries, pRtl->nEntries);

    qsort((void *)pRtl->pEntries, pRtl->nEntries, sizeof(RTL_ENTRY), CompareRtlByOrdinal);
    PrintEntries("\nAfter ordinal sorting", pRtl->pEntries, pRtl->nEntries);

    qsort((void *)pRtl->pEntries, pRtl->nEntries, sizeof(RTL_ENTRY), CompareRtlByName);
    PrintEntries("\nAfter alphabetical sorting", pRtl->pEntries, pRtl->nEntries);

    /* Look for something */
    Name = "CurrentTime";
    Entry = IsaRtl(pRtl, Name);
    if (Entry == NULL)
        printf("RTL '%s' not in the list\n", Name);
    else
        printf("RTL '%s' found with ordinal %d\n", 
               Entry->RtlName, Entry->Ordinal);
}

/* Read the RTL definition file, and sort the descriptors.
 */
RTL GetRtl(FILE *file)
{
    RTL Ret;

    memset(&Ret,0,sizeof Ret);
    ReadRtlFile(file,&Ret);
    printf("Parsed .def file for library %s uuid %s\n",
           Ret.DllName, Ret.Uuid);
#if DEBUG
    CheckRtls(&Ret);
#endif
    qsort((void *)Ret.pEntries, Ret.nEntries, sizeof(RTL_ENTRY),
          CompareRtlByName);
    return Ret;
}

/* Destructor */
void DestroyRtl(PRTL pRtl)
{
    if (pRtl->nEntries) {
        while (pRtl->nEntries--)
            free(pRtl->pEntries[pRtl->nEntries].RtlName);
        free(pRtl->pEntries);
    }
    if (pRtl->DllName)
        free(pRtl->DllName);
    if (pRtl->Uuid)
        free(pRtl->Uuid);
    free(pRtl);
}

/* -- ************ RTL list stuff ************ */

#if DEBUG
/* TESTING */

int main(int argc, char **argv)
{
    FILE * f;
    RTL Rtl;

    if (argc < 2)
        quit(1, "Usage: %s DLL-definition-file", argv[0]);

    f = fopen(argv[1], "r");
    if (f == NULL)
        quit(2, "Could not open %s for reading", argv[1]);

    Rtl = GetRtl(f);

    fclose(f);
    return 0;
}
#else
PRTL ParseDllDeffile(char *Filename)
{
    FILE * f;
    PRTL pRtl;

    f = fopen(Filename, "r");
    if (f == NULL)
        quit(2, "Could not open %s for reading", Filename);

    pRtl = (PRTL) malloc(sizeof(RTL));
    *pRtl = GetRtl(f);
    pRtl->Index = 1;/* unless stated otherwise */

    fclose(f);
    return pRtl;
}
#endif
