/* Copyright (c) Microsoft Corporation. All rights reserved. */
/*
 * Edit files according to command line.
 * This program does all kinds of random little shell programming
 * needed in the build environment.
 */

#include <windows.h>
#include <stdio.h>
#include <process.h>

void quit(char *message)
{
    fprintf(stderr, "sedit: %s", message);
    exit(100);
}

char *fetchfile(FILE *f, int size)
{
    int i;
    char *mem = malloc(size);
    if (!mem)
        quit("malloc");
    i = fread(mem, 1, size, f);
    if (i != size)
        quit("fread");

    return mem;
}

char *fetchstdin(int *pSize)
{
#define CHUNK 65536
    int i, size = 0, c;
    char *mem = malloc(CHUNK);
    if (!mem)
        quit("malloc");

    for(;;) {
        for (i = 0; i < CHUNK; i++) {
            c = getchar();
            if (c == EOF) {
                *pSize = size;
                return mem;
            }
            mem[size] = (char) c;
            size++;
        }

        mem = realloc(mem, size + CHUNK);
        if (!mem)
            quit("realloc");
    };
}

void writeit(FILE *f, char *mem, int size)
{
    int i;

    i = fwrite(mem, 1, size, f);
    if (i != size)
        quit("fwrite");
}


void skip(char *blob, int *pSize, int skipcount)
{
    if (*pSize < skipcount)
        quit("File is too short for skipcount");
    /* Handle -skip */
    *pSize -= skipcount;
    memmove(blob, &blob[skipcount], *pSize);
}

char *edit(char *blob, int *pSize, char *zapstring, char *substitute,
           BOOL ZapCR, BOOL ZapLine, BOOL OnlyBeginning, UINT tabwidth)
{
    int size = *pSize, totalsize = *pSize;
    int i, zlen = 0, sublen = 0;
    int BeginPos = 0, LineStartPos = 0;
    BOOL AtBeginning = TRUE;

    if (zapstring)
        zlen = strlen(zapstring);
    if (substitute)
        sublen = strlen(substitute);

    for (i = 0; i < size; i++) {
        /* Handle -nocr */
        if (ZapCR && blob[i] == '\r') {
            /* zap by moving all remaining bytes up one position */
            memmove(&blob[i], &blob[i+1], size-i-1);
            size--;
            i--; /* maybe there is another one right next to it */
            continue;
        }

        if (blob[i] == '\n' || blob[i] == '\r')
            LineStartPos = i+1;

        /* Handle detabbing: -t */
        if (tabwidth > 0 && blob[i] == '\t') {
            int spaces = tabwidth - (i - LineStartPos) % tabwidth;
            /* If we have to bother with making the buffer bigger,
             * make it at least double the size.
             */
            if (totalsize < size + spaces) {
                totalsize += spaces + totalsize;
                blob = realloc(blob, totalsize);
                if (!blob)
                    quit("Can't realloc in detabify");
            }
            /* Insert spaces and delete tab */
            memmove(&blob[i+spaces-1], &blob[i], size-i);
            memset(&blob[i], ' ', spaces);
            size += spaces-1;
        }

        /* If -b was given, only match at beginning of line ignoring spaces */
        if (OnlyBeginning) {
            if (blob[i] == '\n') {
                AtBeginning = TRUE;
                BeginPos = i+1;
                continue;
            }
            if (!AtBeginning)
                continue;
            if (blob[i] != ' ' && blob[i] != '\t')
                AtBeginning = FALSE;
        } else {
            BeginPos = i;
        }

        /* Handle -z */
        if (zlen && i + zlen <= size && !strncmp(&blob[i], zapstring, zlen)) {
            /* Handle -s */
            if (substitute)
                memmove(&blob[BeginPos], substitute, sublen);

            /* zap by moving all remaining bytes up */
            memmove(&blob[BeginPos+sublen], &blob[i+zlen],size-i-zlen);
            size = size + sublen - zlen - (i - BeginPos);
            i = BeginPos;

            /* Handle -d */
            if (ZapLine) {
                /* Zap to end of line, or end of file, whichever is first */
                int j;
                for (j = 0; i + sublen + j < size; j++)
                    if (blob[i+sublen+j] == '\n')
                        break;
                if (j > 0) {    /* guard against immediate nl */
                    memmove(&blob[i+sublen], &blob[i+sublen+j], size-i-sublen-j);
                    size -= j;
                }
            }
            i--;
        }
    }

    *pSize = size;
    return blob;
}


/* Main driving program
 */
int main(int argc, char **argv)
{
    char *pgm = argv[0], *blob;
    BOOL JustPrint = TRUE, Verbose = FALSE;
    BOOL ZapCR = FALSE, ZapLine = FALSE, Beginning = FALSE;
    FILE *fp;
    INT n, size;
    char *zapstring = NULL;
    char *substitute = NULL;
    char *Execute = NULL;
    UINT skipcount = 0, tabwidth = 0;

    /* Parse arguments, if any
     */
    while ((argc > 1) && (argv[1][0] == '-')) {
        if (!strcmp(&argv[1][1], "nocr")) {
            ZapCR++;
            argv += 1; argc -= 1;
            continue;
        }
        if (!strcmp(&argv[1][1], "skip")) {
            skipcount = atoi(argv[2]);
            argv += 2; argc -= 2;
            continue;
        }
        if (!strcmp(&argv[1][1], "t")) {
            tabwidth = atoi(argv[2]);
            argv += 2; argc -= 2;
            continue;
        }
        if (!strcmp(&argv[1][1], "exec")) {
            Execute = argv[2];
            argv += 2; argc -= 2;
            break;
        }
        switch (argv[1][1]) {
        case 'V':
            JustPrint = !JustPrint;
            argv += 2; argc -= 2;
            break;
        case 'b':
            Beginning = TRUE;
            argv += 1; argc -= 1;
            break;
        case 'v':
            Verbose = TRUE;
            argv += 1; argc -= 1;
            break;
        case 'd':
            ZapLine = TRUE;
            /*FALLTHROUGH*/
        case 'z':
            zapstring = argv[2];
            if (Verbose)
                fprintf(stderr, "sedit: %s '%s'\n", ZapLine ? "Deleting" : "Zapping", zapstring);
            argv += 2; argc -= 2;
            break;
        case 's':
            substitute = argv[2];
            if (zapstring == NULL)
                goto Usage;
            if (strlen(substitute) > strlen(zapstring))
                quit("substitute must be shorter than zapstring (lazybones)\n");
            if (Verbose)
                fprintf(stderr, "sedit: substituting with '%s'\n", substitute);
            argv += 2; argc -= 2;
            break;
        default:
            fprintf(stderr, "sedit: Unknown option '-%c'\n",argv[1][1]);
            goto Usage;
        }
    }

    /* Make sure we got a file to work with
     */
    if (argc > 3 && !Execute) {
    Usage:
        printf("Usage: %s [-z zapstring]|[-d zapstring] [-s substitution] [-nocr] [-skip count] [-t number] [Filename1[ Filename2]]|[-exec prog args]\n", pgm);
        printf(" Simple file editing, in place or with stdin/stdout\n");
        printf(" -v verbose.\n");
        printf(" -z sets string to be deleted\n");
        printf(" -d sets string to be deleted until end-of line\n");
        printf(" -s sets substitution for string that was deleted\n");
        printf(" -b match -z/-d only at beginning of line (skip spaces)\n");
        printf(" -nocr delete all carriage returns\n");
        printf(" -skip delete this many bytes from the beginning\n");
        printf(" -t sets tab width and replaces all tabs with spaces\n");
        printf(" 0 files: stdin/stdout\n");
        printf(" 1 file: in-place\n");
        printf(" 2 files: read File1, write File2\n");
        printf("\n");
        printf(" -exec executes the rest of the command line as a command line\n"
               "       after running the substitution on the command line. e.g.\n"
               "       sedit -z \\ -s / -exec \\foo\\bar -arg 1\n");
        quit("nothing done");
    }

    if (Execute) {
        char **args = calloc(argc+1,sizeof(char *));
        int i, size = strlen(Execute);
        args[0] = edit(strdup(Execute), &size, zapstring, substitute, ZapCR, ZapLine, Beginning, tabwidth);
        args[0][size] = 0;

        for (i = 1; i < argc; i++) {
            size = strlen(argv[i]);
            args[i] = edit(strdup(argv[i]), &size, zapstring, substitute, ZapCR, ZapLine, Beginning, tabwidth);
            args[i][size] = 0;
        }
        args[argc] = NULL;

        if (Verbose) {
            fprintf(stderr, "sedit: Executing [argc=%d]: ", argc);            
            for (i = 0; i < argc; i++)
                fprintf(stderr, "%s ", args[i]);
            fprintf(stderr, "\n");            
        }

        i = _spawnvp(_P_WAIT, argv[0], args);
        if (i == -1) {
            fprintf(stderr, "sedit: exec failed rv=%d: %s argc=%d)\n",
                    i, argv[0], argc);
            quit("can't run program");
        }
        return i;
    }

    if (argc == 1) {
        blob = fetchstdin(&size);
    } else {
        /* Open the input file(s), bail on error
         */
        fp = fopen(argv[1],
                   (argc == 3 || JustPrint) ? "rb" : "r+b");
        if (fp == NULL) {
            printf("Error: could not open file %s for %s\n", argv[1],
                   (argc == 3 || JustPrint) ? "reading" : "reading+writing");
            quit("fopen for reading");
        }

        /* Find out size of input */
        if (fseek(fp, 0, SEEK_END) != 0)
            quit("fseek");
  
        size = ftell(fp);
        if (Verbose)
            fprintf(stderr, "sedit: Size is %d\n", size);

        if (fseek(fp, 0, SEEK_SET) != 0)
            quit("fseek2");

        blob = fetchfile(fp, size);
        fclose(fp);
    }

    n = size;
    if (skipcount)
        skip(blob, &size, skipcount);
    blob = edit(blob, &size, zapstring, substitute, ZapCR, ZapLine, Beginning,
                tabwidth);
    n = size - n;
    if (Verbose)
        fprintf(stderr, "sedit: Edit changed size by %d bytes\n", n);

    if (argc == 1) {
        writeit(stdout, blob, size);
    } else {
        /* Open the output file, bail on error
         */
        if (argc == 3)
            fp = fopen(argv[2],"wb+");
        else                          /* in place */
            fp = fopen(argv[1],"wb+");

        if (fp == NULL) {
            printf("Error: could not open file %s for writing\n",
                   argc == 3 ? argv[2] : argv[1]);
            quit("fopen for writing");
        }

        writeit(fp, blob, size);
        fclose(fp);
    }

    return 0;
}
